Package: lavaan
Title: Latent Variable Analysis
Version: 0.6-21
Authors@R: c(person(given   = "Yves", family = "Rosseel",
                    role    = c("aut", "cre"),
                    email   = "Yves.Rosseel@UGent.be",
                    comment = c(ORCID = "0000-0002-4129-4477")),
             person(given = c("Terrence","D."), family = "Jorgensen",
                    role = "aut",
                    email = "TJorgensen314@gmail.com",
                    comment = c(ORCID = "0000-0001-5111-6773")),
             person(given = c("Luc"), family = "De Wilde",
                    role = "aut",
                    email = "Luc.DeWilde@UGent.be"),
             person(given = "Daniel", family = "Oberski",
                    role = "ctb",
                    email = "daniel.oberski@gmail.com"),
             person(given = "Jarrett", family = "Byrnes", 
                    role = "ctb",
                    email = "byrnes@nceas.ucsb.edu"),
             person(given = "Leonard", family = "Vanbrabant",
                    role = "ctb",
                    email = "info@restriktor.org"),
             person(given = "Victoria", family = "Savalei",
                    role = "ctb",
                    email = "vsavalei@ubc.ca"),
             person(given = "Ed", family = "Merkle",
                    role = "ctb",
                    email = "merklee@missouri.edu"),
             person(given = "Michael", family = "Hallquist",
                    role = "ctb",
                    email = "michael.hallquist@gmail.com"),
             person(given = "Mijke", family = "Rhemtulla",
                    role = "ctb",
                    email = "mrhemtulla@ucdavis.edu"),
             person(given = "Myrsini", family = "Katsikatsou",
                    role = "ctb",
                    email = "mirtok2@gmail.com"),
             person(given = "Mariska", family = "Barendse",
                    role = "ctb",
                    email = "m.t.barendse@gmail.com"),
             person(given = c("Nicholas"), family = "Rockwood",
                    role = "ctb",
                    email = "nrockwood@rti.org"),
             person(given = "Florian", family = "Scharf",
                    role = "ctb",
                    email = "florian.scharf@uni-leipzig.de"),
             person(given = "Han", family = "Du",
                    role = "ctb",
                    email = "hdu@psych.ucla.edu"),
             person(given = "Haziq", family = "Jamil",
                    role = "ctb",
                    email = "haziq.jamil@ubd.edu.bn",
                    comment = c(ORCID = "0000-0003-3298-1010")),
             person(given = "Franz", family = "Classe",
                    role  = "ctb",
                    email = "classe@dji.de")
            )
Description: Fit a variety of latent variable models, including confirmatory
   factor analysis, structural equation modeling and latent growth curve models.
Depends: R(>= 3.4)
Imports: methods, stats4, stats, utils, graphics, MASS, mnormt,
        pbivnorm, numDeriv, quadprog
BugReports: https://github.com/yrosseel/lavaan/issues
License: GPL (>= 2)
LazyData: yes
ByteCompile: true
URL: https://lavaan.ugent.be
NeedsCompilation: no
Packaged: 2025-12-20 16:39:17 UTC; yves
Author: Yves Rosseel [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4129-4477>),
  Terrence D. Jorgensen [aut] (ORCID:
    <https://orcid.org/0000-0001-5111-6773>),
  Luc De Wilde [aut],
  Daniel Oberski [ctb],
  Jarrett Byrnes [ctb],
  Leonard Vanbrabant [ctb],
  Victoria Savalei [ctb],
  Ed Merkle [ctb],
  Michael Hallquist [ctb],
  Mijke Rhemtulla [ctb],
  Myrsini Katsikatsou [ctb],
  Mariska Barendse [ctb],
  Nicholas Rockwood [ctb],
  Florian Scharf [ctb],
  Han Du [ctb],
  Haziq Jamil [ctb] (ORCID: <https://orcid.org/0000-0003-3298-1010>),
  Franz Classe [ctb]
Maintainer: Yves Rosseel <Yves.Rosseel@UGent.be>
Repository: CRAN
Date/Publication: 2025-12-21 08:40:02 UTC
Built: R 4.6.0; ; 2026-04-25 23:29:07 UTC; unix

Package: lavaan
Title: Latent Variable Analysis
Version: 0.7-2
Authors@R: c(person(given   = "Yves", family = "Rosseel",
                    role    = c("aut", "cre"),
                    email   = "Yves.Rosseel@UGent.be",
                    comment = c(ORCID = "0000-0002-4129-4477")),
             person(given = c("Terrence","D."), family = "Jorgensen",
                    role = "aut",
                    email = "TJorgensen314@gmail.com",
                    comment = c(ORCID = "0000-0001-5111-6773")),
             person(given = c("Luc"), family = "De Wilde",
                    role = "aut",
                    email = "Luc.DeWilde@UGent.be"),
             person(given = "Daniel", family = "Oberski",
                    role = "ctb",
                    email = "daniel.oberski@gmail.com"),
             person(given = "Jarrett", family = "Byrnes", 
                    role = "ctb",
                    email = "byrnes@nceas.ucsb.edu"),
             person(given = "Leonard", family = "Vanbrabant",
                    role = "ctb",
                    email = "info@restriktor.org"),
             person(given = "Victoria", family = "Savalei",
                    role = "ctb",
                    email = "vsavalei@ubc.ca"),
             person(given = "Ed", family = "Merkle",
                    role = "ctb",
                    email = "merklee@missouri.edu"),
             person(given = "Michael", family = "Hallquist",
                    role = "ctb",
                    email = "michael.hallquist@gmail.com"),
             person(given = "Mijke", family = "Rhemtulla",
                    role = "ctb",
                    email = "mrhemtulla@ucdavis.edu"),
             person(given = "Myrsini", family = "Katsikatsou",
                    role = "ctb",
                    email = "mirtok2@gmail.com"),
             person(given = "Mariska", family = "Barendse",
                    role = "ctb",
                    email = "m.t.barendse@gmail.com"),
             person(given = c("Nicholas"), family = "Rockwood",
                    role = "ctb",
                    email = "nrockwood@rti.org"),
             person(given = "Florian", family = "Scharf",
                    role = "ctb",
                    email = "florian.scharf@uni-leipzig.de"),
             person(given = "Han", family = "Du",
                    role = "ctb",
                    email = "hdu@psych.ucla.edu"),
             person(given = "Haziq", family = "Jamil",
                    role = "ctb",
                    email = "haziq.jamil@ubd.edu.bn",
                    comment = c(ORCID = "0000-0003-3298-1010")),
             person(given = "Franz", family = "Classe",
                    role  = "ctb",
                    email = "classe@dji.de")
            )
Description: Fit a variety of latent variable models, including confirmatory
   factor analysis, structural equation modeling and latent growth curve models.
Depends: R(>= 3.4)
Imports: methods, stats4, stats, utils, graphics, MASS, mnormt,
        pbivnorm, numDeriv, quadprog
BugReports: https://github.com/yrosseel/lavaan/issues
License: GPL (>= 2)
LazyData: yes
LazyDataCompression: xz
ByteCompile: true
URL: https://lavaan.ugent.be
NeedsCompilation: no
Packaged: 2026-07-16 15:01:49 UTC; yves
Author: Yves Rosseel [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4129-4477>),
  Terrence D. Jorgensen [aut] (ORCID:
    <https://orcid.org/0000-0001-5111-6773>),
  Luc De Wilde [aut],
  Daniel Oberski [ctb],
  Jarrett Byrnes [ctb],
  Leonard Vanbrabant [ctb],
  Victoria Savalei [ctb],
  Ed Merkle [ctb],
  Michael Hallquist [ctb],
  Mijke Rhemtulla [ctb],
  Myrsini Katsikatsou [ctb],
  Mariska Barendse [ctb],
  Nicholas Rockwood [ctb],
  Florian Scharf [ctb],
  Han Du [ctb],
  Haziq Jamil [ctb] (ORCID: <https://orcid.org/0000-0003-3298-1010>),
  Franz Classe [ctb]
Maintainer: Yves Rosseel <Yves.Rosseel@UGent.be>
Repository: CRAN
Date/Publication: 2026-07-16 21:00:07 UTC

# only when we have src code
# useDynLib("lavaan")

importFrom("graphics",
           "abline", "axis", "box", "hist", "lines", "pairs", "par", "plot.default",
           "polygon", "text", "arrows",
           # generics
           "plot")

importFrom("grDevices", "dev.off", "png", "col2rgb")

importFrom("methods",
           "is", "new", "slot", "slotNames", ".hasSlot",
           "setClass", "representation", "setGeneric",
           "setRefClass", "setMethod",
           # generics,
           "show", "signature")

importFrom("stats",
           "as.formula", "complete.cases", "cor", "cov", "cov2cor", "cov.wt",
           "dnorm", "lm.fit", "lm.wfit", "mad", "na.omit", "nlminb", "optim",
           "pchisq", "plogis", "pnorm", "qchisq", "qnorm", "quantile", "rnorm",
           "runif", "sd", "terms", "uniroot", "var", "weighted.mean", "median",
           "aggregate", "dlogis", "qlogis", "optimize", "lm", "setNames",
           # generics
           "coef", "residuals", "resid", "fitted.values", "fitted",
           "predict", "update", "anova", "vcov")

importFrom("utils",
           "combn", "modifyList", "packageDescription", "read.table",
           "str", "write.table", "write.csv", "packageVersion")

importFrom("quadprog",
           "solve.QP")

importFrom("mnormt",
           "sadmvn")

importFrom("pbivnorm",
           "pbivnorm")

# AIC and friends...
if(getRversion() >= "2.13.0") {
    importFrom("stats4",AIC, BIC, logLik, nobs)
    export(logLik, nobs) # not AIC, BIC?
} else {
    importFrom("stats4",AIC, BIC, logLik)
    export(logLik, BIC, nobs) # not AIC?
}


# export ordinary functions defined in this package
export("lavaan", "cfa", "sem", "growth",
       "lavaanList", "cfaList", "semList",
       #"fsr", "twostep",
       "sam", "efa",
       # new name            # old name
       "lavParTable",        "lavaanify",
       "lavNames",           "lavaanNames",
       "lavParseModelString", "lav_parse_options", # "parseModelString",
       "lavInspect",         "inspect",
       "lavTech",
       "lavListInspect",     "lavListTech",
       "lavResiduals",
       "lavEffects",
       # utilities
       "lav_getcov", "lav_char2num", "lav_cor2cov",
       "getCov", "char2num", "cor2cov",                 # old names
       "lavOptions",
#      "lav_use_lavaanC",
       "modindices", "modificationIndices", "modificationindices",
       "standardizedSolution", "standardizedsolution",
       "lavParameterEstimates", "parameterEstimates", "parameterestimates",
       "parameterTable", "parametertable", "parTable", "partable",
       "varTable", "vartable",
       "fitMeasures", "fitmeasures",
       "lavInspectSampleCov",                          "inspectSampleCov",
       "lavBootstrap",       "bootstrapLavaan",
       "lavSimulateData",    "simulateData",
       "estfun.lavaan", "lavScores",
       "lavTables",
       "lavTablesFitCp", "lavTablesFitCf", "lavTablesFitCm",
       "lavExport",
       "lavTest", "lavTestLRT", "lavTestWald", "lavTestScore",
       "lavMatrixRepresentation",
       "lav_mplus_lavaan", "mplus2lavaan",
       "lav_mplus_syntax_model", "mplus2lavaan.modelSyntax",
       #"lavData",
       "lavPredict",
       "lavPredictY",
       "lavPredictY_cv",
       "lavResidualsY",
       "lavCor",

       # API functions

       # lav_matrix
       "lav_matrix_vec",
       "lav_matrix_vecr",
       "lav_matrix_vech",
       "lav_matrix_vechr",
       "lav_matrix_vechu",
       "lav_matrix_vechru",
       "lav_matrix_vech_idx",
       "lav_matrix_vech_row_idx",
       "lav_matrix_vech_col_idx",
       "lav_matrix_vechr_idx",
       "lav_matrix_vechu_idx",
       "lav_matrix_vechru_idx",
       "lav_matrix_diag_idx",
       "lav_matrix_diagh_idx",
       "lav_matrix_antidiag_idx",
       "lav_matrix_vech_reverse",
       "lav_matrix_vechru_reverse",
       "lav_matrix_upper2full",
       "lav_matrix_vechr_reverse",
       "lav_matrix_vechu_reverse",
       "lav_matrix_lower2full",
       "lav_matrix_duplication",
       "lav_matrix_duplication_pre",
       "lav_matrix_duplication_post",
       "lav_matrix_duplication_pre_post",
       "lav_matrix_duplication_ginv",
       "lav_matrix_duplication_ginv_pre",
       "lav_matrix_duplication_ginv_post",
       "lav_matrix_duplication_ginv_pre_post",
       "lav_matrix_commutation",
       "lav_matrix_commutation_pre",
       "lav_matrix_commutation_post",
       "lav_matrix_commutation_pre_post",
       "lav_matrix_commutation_mn_pre",
       "lav_matrix_symmetric_sqrt",
       "lav_matrix_orthogonal_complement",
       "lav_matrix_bdiag",
       "lav_matrix_trace",
       "lav_matrix_cov",

       # lav_partable
       "lav_partable_independence",   # used by semTools!
       "lav_partable_unrestricted",
       "lav_partable_npar",
       "lav_partable_ndat",
       "lav_partable_df",
       "lav_partable_labels",
       "lav_partable_from_lm",
       "lav_partable_complete",
       "lav_partable_attributes",
       "lav_partable_merge",
       "lav_partable_add",
       "lav_partable_constraints_def",
       "lav_partable_constraints_ceq",
       "lav_partable_constraints_ciq",

       # lav_constraints
       "lav_constraints_parse",

       # lav_func
       "lav_func_gradient_complex",
       "lav_func_gradient_simple",
       "lav_func_jacobian_complex",
       "lav_func_jacobian_simple",    # used by semTools!

       # lav_model
       "lav_model_get_parameters",
       "lav_model_set_parameters",
       "lav_model_implied",
       "lav_model_vcov_se",

       # lav_data
       "lav_data_update",

       # lav_samplestats
       "lav_samplestats_from_data",

       # estimation
       "lav_export_estimation",

       # plotting utilities
       "lav_label_code",
       "lav_model_plotinfo",
       "lav_plotinfo_positions",
       "lav_plotinfo_tikzcode",
       "lav_plotinfo_rgraph",
       "lav_plotinfo_svgcode",
       "lav_plot",

       # shortened function names
       "lav_mat_dup_ginv_pre_post",
       "lav_mat_ortho_complement",
       "lav_func_grad_complex",
       "lav_mat_dup_ginv_post",
       "lav_func_grad_simple",
       "lav_mat_vech_row_idx",
       "lav_mat_vech_col_idx",
       "lav_mat_antidiag_idx",
       "lav_mat_dup_pre_post",
       "lav_mat_dup_ginv_pre",
       "lav_mat_com_pre_post",
       "lav_pt_unrestricted",
       "lav_pt_independence",
       "lav_mat_vechru_idx",
       "lav_mat_upper2full",
       "lav_mat_lower2full",
       "lav_mat_com_mn_pre",
       "lav_samp_from_data",
       "lav_mat_vechru_rev",
       "lav_mat_vechr_idx",
       "lav_mat_vechu_idx",
       "lav_mat_diagh_idx",
       "lav_pt_attributes",
       "lav_mat_vechr_rev",
       "lav_mat_vechu_rev",
       "lav_mat_vech_idx",
       "lav_mat_diag_idx",
       "lav_mat_dup_post",
       "lav_mat_dup_ginv",
       "lav_mat_com_post",
       "lav_mat_sym_sqrt",
       "lav_mat_vech_rev",
       "lav_mat_dup_pre",
       "lav_mat_com_pre",
       "lav_pt_complete",
       "lav_mat_vechru",
       "lav_pt_con_def",
       "lav_pt_con_ceq",
       "lav_pt_con_ciq",
       "lav_pt_from_lm",
       "lav_con_parse",
       "lav_mat_vechr",
       "lav_mat_vechu",
       "lav_mat_bdiag",
       "lav_mat_trace",
       "lav_pt_labels",
       "lav_mat_vecr",
       "lav_mat_vech",
       "lav_pt_merge",
       "lav_mat_vec",
       "lav_mat_dup",
       "lav_mat_com",
       "lav_mat_cov",
       "lav_pt_ndat",
       "lav_pt_npar",
       "lav_pt_add",
       "lav_pt_df"

       # deprecated functions
       #"vech", "vech.reverse", "vechru", "vechru.reverse", "lower2full",
       #"vechr", "vechr.reverse", "vechu", "vechu.reverse", "upper2full",
       #"duplicationMatrix", "commutationMatrix", "sqrtSymmetricMatrix"
      )

# export Classes
exportClasses(
              "lavaan",
              "lavaanList"
             )

# export Methods
exportMethods(
              "predict",
              "update",
              "anova",
              "coef",
              "residuals", "resid",
              "vcov",
              "logLik",
              "fitted.values", "fitted",
              "summary"
             )

S3method(print, lavaan.data.frame, lav_dataframe_print)
S3method(print, lavaan.list, lav_lavaanlist_print)
S3method(print, lavaan.matrix, lav_mat_print)
S3method(print, lavaan.matrix.symmetric, lav_mat_sym_print)
S3method(print, lavaan.vector, lav_vector_print)
S3method(print, lavaan.parameterEstimates, lav_parameterestimates_print)
S3method(print, lavaan.effects, lav_effects_print)
S3method(print, lavaan.fitMeasures, lav_fitmeasures_print)
S3method(print, lavaan.residuals.summary, lav_residuals_summary_print)
S3method(print, lavaan.summary, lav_summary_print)
S3method(print, lavaan.tables.fit.Cf, lav_tables_fit_cf_print)
S3method(print, lavaan.tables.fit.Cp, lav_tables_fit_cp_print)
S3method(print, lavaan.tables.fit.Cm, lav_tables_fit_cm_print)
S3method(print, lavaan.efa, lav_efa_print)
S3method(print, efaList.summary, lav_efalist_summary_print)
S3method(print,    efaList, lav_efalist_print)
S3method(predict,  efaList, lav_efalist_predict)
S3method(summary, efaList, lav_efalist_summary)
S3method(fitMeasures, efaList, lav_efalist_fitmeasures)
S3method(fitmeasures, efaList, lav_efalist_fitmeasures)
S3method(pairs, lavaan, lav_lavaan_pairs)
S3method(inspect,    lavaan, lav_lavaan_inspect)
S3method(inspect,    lavaanList, lav_lavaanlist_inspect)
S3method(lavInspect, lavaan, lav_lavaan_lavinspect)
S3method(lavTech,    lavaan, lav_lavaan_lavtech)
S3method(lavInspect, lavaanList, lav_lavaanlist_lavinspect)
S3method(lavTech,    lavaanList, lav_lavaanlist_lavtech)
if(getRversion() >= "3.6.0") {
    S3method(sandwich::estfun, lavaan)
}

# lavaan

lavaan is a free, open source R package for latent variable analysis. You can
use lavaan to estimate a large variety of multivariate statistical models,
including path analysis, confirmatory factor analysis, structural equation
modeling and growth curve models.

The lavaan package is developed to provide useRs, researchers and teachers a
free open-source, but commercial-quality package for latent variable modeling.
The long-term goal of lavaan is to implement all the state-of-the-art
capabilities that are currently available in commercial packages. However,
lavaan is still under development, and much work still needs to be done.

## Installation

Install the stable version from CRAN:

```R
install.packages("lavaan")
```

Or the development version from GitHub:

```R
# install.packages("remotes")
remotes::install_github("yrosseel/lavaan")
```

## Usage

To get a first impression of how lavaan works in practice, consider the
following example of a SEM model (the Political Democracy Example from 
Bollen's 1989 book):

```R
library(lavaan)

model <- '
   # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + y2 + y3 + y4
     dem65 =~ y5 + y6 + y7 + y8
   # regressions
     dem60 ~ ind60
     dem65 ~ ind60 + dem60
   # residual covariances
     y1 ~~ y5
     y2 ~~ y4 + y6
     y3 ~~ y7
     y4 ~~ y8
     y6 ~~ y8
'
fit <- sem(model, data = PoliticalDemocracy)
summary(fit)
```

More information can be found on the website: https://lavaan.org

# lavaan

lavaan is a free, open source R package for latent variable analysis. You can
use lavaan to estimate a large variety of multivariate statistical models,
including path analysis, confirmatory factor analysis, structural equation
modeling and growth curve models.

The lavaan package is developed to provide useRs, researchers and teachers a
free open-source, but commercial-quality package for latent variable modeling.
The long-term goal of lavaan is to implement all the state-of-the-art
capabilities that are currently available in commercial packages. However,
lavaan is still under development, and much work still needs to be done.

## Job ad

As part of an NWO-funded project for Dutch open-science infrastructural
development, we are looking for a (second) 
**researcher/junior postdoc for the lavaan project**. 
The job ad can be found on the University of [Twente](https://utwentecareers.nl/en/vacancies/2525/researcher-in-methods-and-statistics-specializing-in-programming-and-structural-equation-modelling/).

The deadline is *31 July 2026*.

The proposal for the full project
can be found on [Zenodo](https://doi.org/10.5281/zenodo.19552779). The post-doc
position above is part of Work Package 2 (WP2).


## Installation

Install the stable version from CRAN:

```R
install.packages("lavaan")
```

Or the development version from GitHub:

```R
# install.packages("remotes")
remotes::install_github("yrosseel/lavaan")
```

## Usage

To get a first impression of how lavaan works in practice, consider the
following example of a SEM model (the Political Democracy Example from 
Bollen's 1989 book):

```R
library(lavaan)

model <- '
   # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + y2 + y3 + y4
     dem65 =~ y5 + y6 + y7 + y8
   # regressions
     dem60 ~ ind60
     dem65 ~ ind60 + dem60
   # residual covariances
     y1 ~~ y5
     y2 ~~ y4 + y6
     y3 ~~ y7
     y4 ~~ y8
     y6 ~~ y8
'
fit <- sem(model, data = PoliticalDemocracy)
summary(fit)
```

More information can be found on the website: https://lavaan.org

# How does lavaan work?

library(lavaan)

# PART 1: from syntax to matrices

model <- 'f =~ x1 + a*x2 + 3*x3'

# parsing the syntax
lavParseModelString(model, as.data.frame. = TRUE)
# creates a 'FLAT' initial parameter table
FLAT <- lavParseModelString(model)
lavNames(FLAT)
lavNames(FLAT, "lv")

# lavParTable()
# - first creates FLAT
# - needs information from the data (eg number of groups)
# - builds the initial parameter table
# - the parameter table is the core representation of the model in lavaan
lavParTable(model)

# if sem/cfa is used, more parameters are set free
lavParTable(model, auto = TRUE)

# example: equality constraints using labels
model <- 'f =~ x1 + a*x2 + a*x3'
lavParTable(model, auto = TRUE)

# alternative for 'simple' equality constraints
# (will become the default soon)
lavParTable(model, auto = TRUE, ceq.simple = TRUE)

# explicit equality constraints
model <- 'f =~ x1 + a*x2 + b*x3; a == b'
lavParTable(model, auto = TRUE, ceq.simple = TRUE)

# multiple groups/blocks
model <- 'f =~ x1 + c(a1,a2)*x2 + c(b1, b2)*x3'
lavParTable(model, auto = TRUE, ngroups = 2)

# matrix representation: LISREL-all-y
model <- 'f =~ x1 + x2 + x3'
PT <- lavParTable(model, auto = TRUE, as.data.frame. = TRUE)
PT
# map every parameter to a matrix element
MAT <- as.data.frame(lavaan:::lav_lisrel(PT))
cbind(PT, MAT)

# alternative matrix representation: RAM
MAT <- as.data.frame(lavaan:::lav_ram(PT))
cbind(PT, MAT)

# first create default lavoptions
fit <- sem(model = 'f =~x1 + x2 + x3', data = HolzingerSwineford1939)
lavoptions <- fit@Options

# create 'Model' object (S4 class)
# (often called lavmodel in internal functions)
Model <- lavaan:::lav_model(PT, lavoptions = lavoptions)
# another representation of the model, more suitable for computations
Model@nmat
Model@isSymmetric
Model@dimNames

# the matrix representation is in GLIST (group list)
# GLIST may contain the matrices of multiple groups, levels, blocks, ...
MLIST <- Model@GLIST

# MLIST is the matrix representation of a single block
# MLIST is used to compute model-based statistics, e.g.,
# the model-implied variance-covariance matrix
lavaan:::lav_lisrel_sigma(MLIST)




# PART 2: lavaan() workflow:
# 18 steps in total (many of them can be skipped)

# 1. lavParseModelString + ov.names + ngroups + nlevels ...

# 2. set/check options (eg meanstructure = TRUE)
#   lav_options_set() # lav_options.R

# 3. check the (raw) data, or the sample statistics
#   lav_lavdata()
lavdata <- lavaan:::lav_lavdata(data = HolzingerSwineford1939,
                            ov.names = c("x1", "x2", "x3"))
slotNames(lavdata)
lavdata@ov.names
lavdata@ngroups
lavdata@X[[1]] # the raw data in group/block 1

# 4. lavpartable: create the parameter table
# needs to know: how many groups, how many categories, ...
lavpartable <- lavParTable(model = FLAT, auto = TRUE)

# 4b. compute and store partable attributes (ov.names, ov.names.x, ...)
lavpta <- lav_partable_attributes(lavpartable)
lavpta$vnames$ov

# 5. lavsamplestats
# compute sample statistics (cov, mean, Gamma, ...)
lavsamplestats <- lavaan:::lav_samplestats_from_data(lavdata,
                                                     lavoptions = lavoptions)
slotNames(lavsamplestats)
lavsamplestats@cov[[1]] # observed covariance matrix first group/block
lavsamplestats@icov[[1]] # inverse observed covariance matrix
lavsamplestats@cov.log.det[[1]] # log determinant covariance matrix

# 6. lavh1
# summary statistics of the 'unrestricted' (h1) model
# with complete data, this is trivial (cov, mean)
# when data is missing, we need to estimate cov/mean using EM
lavh1 <- lavaan:::lav_h1_implied_logl(lavdata = lavdata,
                             lavsamplestats   = lavsamplestats,
                             lavoptions       = lavoptions)
lavh1$implied

# 7. parameter bounds (needs lavh1$implied)
# only used if bounded estimation is requested
lavoptions$bounds <- "standard"
lavoptions$optim.bounds <- list(lower = c("ov.var", "loadings"),
                                upper = c("ov.var", "loadings"),
                                min.reliability.marker = 0.1)
lavpartable <- lavaan:::lav_partable_add_bounds(partable = lavpartable,
                lavh1 = lavh1, lavdata = lavdata, lavsamplestats = lavsamplestats,
                lavoptions = lavoptions)
# remove bounds again to save space
lavpartable$lower <- NULL
lavpartable$upper <- NULL

# 8. compute starting values
lavpartable$start <- lavaan:::lav_start(start.method = lavoptions$start,
                               lavpartable     = lavpartable,
                               lavsamplestats  = lavsamplestats)
lavpartable

# 9. lavmodel: create internal model representation (with GLIST)
lavmodel <- lavaan:::lav_model(lavpartable      = lavpartable,
                               lavoptions       = lavoptions)
lavmodel@GLIST

# 10. lavcache: compute some additional summary statistis
# only used when estimator = "PML", and "MML" (for now)
lavcache <- list()

# 11. estimation
# - default: lav_model_estimate() + nlminb() (quasi-Newton optimization)
# - lav_optim_gn(): Gauss-Newton optimization
# - lav_optim_noniter(): non-iterative procedures
# - lav_mvnorm_cluster_em_h0: EM for multilevel models
lavaan:::lav_verbose(TRUE) # be verbose
x <- try(lavaan:::lav_model_estimate(lavmodel        = lavmodel,
                            lavpartable     = lavpartable,
                            lavsamplestats  = lavsamplestats,
                            lavdata         = lavdata,
                            lavoptions      = lavoptions,
                            lavcache        = lavcache))
lavaan:::lav_verbose(FALSE) # switch off verbose

# store parameters in lavmodel
lavmodel <- lav_model_set_parameters(lavmodel, x = as.numeric(x))

# store parameters in @ParTable$est
lavpartable$est <- lav_model_get_parameters(lavmodel = lavmodel,
                                            type = "user", extra = TRUE)
lavpartable

# 12. lavimplied + lavloglik
# store model implied statistics in @implied
# if likelihood-based method, store also loglik in @loglik
lavimplied <- lav_model_implied(lavmodel)
lavloglik <- lavaan:::lav_model_loglik(lavdata        = lavdata,
                              lavsamplestats = lavsamplestats,
                              lavimplied     = lavimplied,
                              lavmodel       = lavmodel,
                              lavoptions     = lavoptions)

# 13. compute 'vcov': the variance matrix of the (free) parameters
# this is needed to compute standard errors
VCOV <- lavaan:::lav_model_vcov(lavmodel        = lavmodel,
                       lavsamplestats  = lavsamplestats,
                       lavoptions      = lavoptions,
                       lavdata         = lavdata,
                       lavpartable     = lavpartable,
                       lavcache        = lavcache,
                       lavimplied      = lavimplied,
                       lavh1           = lavh1)
VCOV

# prepare lavvcov slot
lavvcov <- list(se = lavoptions$se, information = lavoptions$information,
                vcov = VCOV)

# store standard errors in parameter table
lavpartable$se <- lavaan:::lav_model_vcov_se(lavmodel = lavmodel,
                                    lavpartable = lavpartable,
                                    VCOV = VCOV)

# 14. compute global test statistic (chi-square)
# trivial for standard test (=N * F_ML)
# more work for 'robust' test statistics (eg test = "satorra-bentler")
TEST <- lavaan:::lav_model_test(lavmodel            = lavmodel,
                       lavpartable         = lavpartable,
                       lavsamplestats      = lavsamplestats,
                       lavimplied          = lavimplied,
                       lavh1               = lavh1,
                       lavoptions          = lavoptions,
                       x                   = x,
                       VCOV                = VCOV,
                       lavdata             = lavdata,
                       lavcache            = lavcache,
                       lavloglik           = lavloglik)
lavtest <- TEST

# 14bis. lavfit
# store 'fit'information
# no longer used, but if I remove it, a dozen (old) packages break...

# 15. fit baseline model
fit.indep <- try(lavaan:::lav_object_independence(object = NULL,
                                         lavsamplestats = lavsamplestats,
                                         lavdata        = lavdata,
                                         lavcache       = lavcache,
                                         lavoptions     = lavoptions,
                                         lavh1          = lavh1), silent = TRUE)

# 16. rotation
# only needed if efa() blocks are defined
# lavmodel <- lav_model_efa_rotate(lavmodel   = lavmodel,
#                                  lavoptions = lavoptions)

# 17. create lavaan object

# don't run, as some pieces have not been created...
# lavaan <- new("lavaan",
#               version      = as.character(packageVersion("lavaan")),
#               call         = mc,                  # match.call
#               timing       = timing,              # list
#               Options      = lavoptions,          # list
#               ParTable     = lavpartable,         # list
#               pta          = lavpta,              # list
#               Data         = lavdata,             # S4 class
#               SampleStats  = lavsamplestats,      # S4 class
#               Model        = lavmodel,            # S4 class
#               Cache        = lavcache,            # list
#               Fit          = lavfit,              # S4 class
#               boot         = lavboot,             # list
#               optim        = lavoptim,            # list
#               implied      = lavimplied,          # list
#               loglik       = lavloglik,           # list
#               vcov         = lavvcov,             # list
#               test         = lavtest,             # list
#               h1           = lavh1,               # list
#               baseline     = lavbaseline,         # list
#               internal     = list(),              # empty list
#               external     = list()               # empty list
# )

# 18. post-fitting check of parameters
# lavInspect(lavaan, "post.check")

# the sem/cfa/growth function just set some
# options to user-friendly settings:

# default options for sem/cfa call
# mc$int.ov.free     = TRUE
# mc$int.lv.free     = FALSE
# #mc$auto.fix.first  = !std.lv
# mc$auto.fix.first  = TRUE # (re)set in lav_options_set
# mc$auto.fix.single = TRUE
# mc$auto.var        = TRUE
# mc$auto.cov.lv.x   = TRUE
# mc$auto.cov.y      = TRUE
# mc$auto.th         = TRUE
# mc$auto.delta      = TRUE
# mc$auto.efa        = TRUE


# PART 3: extractor functions
fit <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939)

lavParameterEstimates(fit)
# = subset of parTable(fit), but with additional columsn (z, pvalues, ...)

lavParameterEstimates(fit, output = "text")
# this is a big part of the summary() output

# summary()
# first creates summary output as a list:
out <- summary(fit)
names(out)
class(out)
# with a specific print function
out

# fit indices
fitMeasures(fit)
fitMeasures(fit, output = "matrix")
fitMeasures(fit, output = "text")

# lavInspect/lavTech
lavInspect(fit, "est")
lavTech(fit, "est")


# PART 4: lavaan slots

fit <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939)
class(fit)
slotNames(fit)

# 1. lavaan version used to create this object
fit@version

# 2. user-specified call
fit@call

# 3. timings of several substeps
unlist(fit@timing)

# 4. options used for this object
unlist(fit@Options)

# 5. the parameter table
fit@ParTable  # list
parTable(fit) # return as data.frame

# 6. the parameter table attributes
names(fit@pta)
fit@pta$vnames$ov
fit@pta$vidx$ov
fit@pta$nfac
fit@pta$nblocks

# 7. Data slot (S4)
fit@Data # has its own print function
slotNames(fit@Data)
as.data.frame(fit@Data@ov)
str(fit@Data)

# 8. SampleStats (S4)
fit@SampleStats
slotNames(fit@SampleStats)
fit@SampleStats@cov[[1]] # list with element per group

# 9. Model (S4)
slotNames(fit@Model)
fit@Model@x.free.idx # parameter index in parameter table
fit@Model@m.free.idx # parameter index in model matrix

# 10. Cache (list)
# cached information, only used for estimator PML and MML (for now)

# 11. Fit
# deprecated, only kept to avoid breaking some (old) packages (eg rsem)

# 12. boot (list)
# only used if bootstrapping was used
fitb <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939,
            se = "bootstrap", bootstrap = 100L, verbose = TRUE)
head(fitb@boot$coef)

# 13. optim -- info about (iterative) optimization process
str(fit@optim)

# 14. loglik -- loglikelihood information (ML estimation only)
unlist(fit@loglik)

# 15. implied -- implied sample statistics (per group)
fit@implied

# 16. vcov -- variance-covariance matrix of the (free) model parameters
fit@vcov

# 17. test -- global test statistics
fit2 <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939,
            test = "satorra-bentler")
names(fit2@test)
fit2@test$satorra.bentler

# 18. h1 -- sample statistics + logl unrestricted/saturated model
# often similar to info in @SampleStats, but not if missing data,
# multilevel data, ...
fit@h1$implied # this is what is used for lavInspect(fit, "implied")
fit@h1$logl

# 19. baseline -- information about the baseline model (needed for CFI/TLI)
names(fit@baseline)
as.data.frame(fit@baseline$partable)
fit@baseline$test$standard

# 20. internal -- list to store specials flags/info, for internal use only

# 21. external -- list to store specials flags/info if you are an external
# developer


# PART 5: source code structure

# 5 main functions:

# 1) xxx_lavaan.R
# 2) xxx_lavaanList.R
# 3) xxx_efa.R (new in 0.6-13)
# 4) xxx_sam.R
# 5) xxx_fsr.R (only for repro reasons; access via lavaan:::fsr())

# files that start with ctr_  contain contributed code
# written by others (although we are slowly converting them to lav_
# files)

# 00class.R contains S4 class definitions
# 00generic.R defines 'generic' functions (that can be specialized)
# fitMeasures(), lavInspect(), lavTech()

# zzz_OLDNAMES.R contains aliases for (very old) function names
# that are still used by some packages
# eg computeExpectedInformation <- lav_model_information_expected

# zzz.R traditionally contains the (in)famous startup message

# most files start with the lav_ prefix
# the second term often refers the type of object for which the file
# contains functions, for example
# lav_matrix.R
# lav_partable_subset.R
# lav_model_estimate.R
# lav_object_post_check.R

# but sometimes, it refers to what is created, or what is done
# lav_test.R  # creates @test slot
# lav_print.R # prints various objects

# for ML, an important set of functions are:
# lav_mvnorm.R
# lav_mvnorm_h1.R
# lav_mvnorm_missing.R
# lav_mvnorm_missing_h1.R
# lav_mvnorm_cluster.R
# lav_mvnorm_cluster_missing.R

# the standard ML discrepancy function is found at the top of
# lav_objective.R

# sometimes, lavaan needs to do trivial things (like regression), but
# needs a bit more info than what we get from standard functions (like lm):
# lav_uvreg.R          # univariate regression
# lav_uvord.R          # univariate probit regression
# lav_mvreg.R          # multivariate regression
# lav_mvreg_cluster.R  # multivariate twolevel regression

# during iterative estimation, we need to compute the value of the
# objective function (i.e., the discrepancy function) many times

# lav_model_estimate.R # runs the iterations
# and defines the following objective function:

   # function to be minimized
    objective_function <- function(x, verbose = FALSE, infToMax = FALSE,
                                   debug = FALSE) {
        # 2. unpack
        if(lavmodel@eq.constraints) {
            x <- as.numeric(lavmodel@eq.constraints.K %*% x) +
                            lavmodel@eq.constraints.k0
        }

        # 1. unscale
        x <- x / parscale

        # update GLIST (change `state') and make a COPY!
        GLIST <- lav_model_x2glist(lavmodel, x = x)

        fx <- lav_model_objective(lavmodel       = lavmodel,
                                  GLIST          = GLIST,
                                  lavsamplestats = lavsamplestats,
                                  lavdata        = lavdata,
                                  lavcache       = lavcache,
                                  verbose        = verbose)

        # only for PML: divide by N (to speed up convergence)
        if(estimator == "PML") {
            fx <- fx / lavsamplestats@ntotal
        }

        if(debug || verbose) {
            cat("  objective function  = ",
                sprintf("%18.16f", fx), "\n", sep="")
        }
        if(debug) {
            cat("Current free parameter values =\n"); print(x); cat("\n")
        }

        if(lavoptions$optim.partrace) {
            PENV$PARTRACE <- rbind(PENV$PARTRACE, c(fx, x))
        }

        # for L-BFGS-B
        #if(infToMax && is.infinite(fx)) fx <- 1e20
        if(!is.finite(fx)) {
            fx.group <- attr(fx, "fx.group")
            fx <- 1e20
            attr(fx, "fx.group") <- fx.group # only for lav_model_fit()
        }

        fx
    }

# lav_model_objective() can be found in lav_model_objective.R
# 1) compute model implied summary statistics (for each group)
#    using eg lav_model_sigma()
# 2) compute value discrepancy function
#    eg lav_model_objective_gls() or lav_model_objective_ml() # see lav_objective.R
# 3) if multiple groups, combine the values using group weights
# 4) return value (fx)

# lavaan parameter table
#
# initial version: YR 22/05/2009
#  major revision: YR 02/11/2010: - FLATTEN the model syntax and turn it into a
#                                   data.frame, with a "modifiers" attribute
#                                 - add default elements here
#                                 - check for duplicate elements
#                                 - allow for every possible model...
#                                 - since 0.4-5
#                                 - the end result is a full description of
#                                   a model (but no matrix representation)
# - 14 Jan 2014: merge 02lavaanUser.R with lav_partable.R
#                move syntax-based code to lav_syntax.R
# - 26 April 2016: handle multiple 'blocks' (levels, classes, groups, ...)
# - 24 March 2019: handle efa sets
# - 23 May   2020: support for random slopes

lav_model_pt  <- function(
                          model = NULL,
                          meanstructure = FALSE,
                          int_ov_free = FALSE,
                          int_lv_free = FALSE,
                          marker_int_zero = FALSE,
                          orthogonal = FALSE,
                          orthogonal_y = FALSE,
                          orthogonal_x = FALSE,
                          orthogonal_efa = FALSE,
                          std_lv = FALSE,
                          correlation = FALSE,
                          composites = TRUE,
                          composites_cov_free = FALSE,
                          effect_coding = "",
                          conditional_x = FALSE,
                          fixed_x = FALSE,
                          parameterization = "delta",
                          constraints = NULL,
                          ceq_simple = FALSE,
                          auto = FALSE,
                          model_type = "sem",
                          auto_fix_first = FALSE,
                          marker = NULL,
                          auto_fix_single = FALSE,
                          auto_var = FALSE,
                          auto_cov_lv_x = FALSE,
                          auto_cov_x = FALSE,
                          auto_cov_y = FALSE,
                          auto_th = FALSE,
                          auto_delta = FALSE,
                          auto_efa = FALSE,
                          var_table = NULL,
                          ngroups = 1L,
                          nthresholds = NULL,
                          group_equal = NULL,
                          group_partial = NULL,
                          group_w_free = FALSE,
                          debug = FALSE,
                          warn = TRUE,
                          as_data_frame = TRUE) {
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  if (!missing(warn)) {
    current_warn <- lav_warn()
    if (lav_warn(warn))
      on.exit(lav_warn(current_warn), TRUE)
  }

  # check if model is already flat or a full parameter table
  if (is.list(model) && !is.null(model$lhs)) {
    if (is.null(model$mod.idx)) {
      lav_msg_warn(gettext("input already looks like a parameter table"))
      return(lav_pt_set_cache(model))
    } else {
      flat <- model
    }
  } else {
    # parse the model syntax and flatten the user-specified model
    # return a data.frame, where each line is a model element (rhs, op, lhs)
    flat <- lavParseModelString(
      model_syntax = model, debug = FALSE
    )
  }
  # user-specified *modifiers* are returned as an attribute
  tmp_mod <- attr(flat, "modifiers")
  attr(flat, "modifiers") <- NULL
  # user-specified *constraints* are returned as an attribute
  tmp_con <- attr(flat, "constraints")
  attr(flat, "constraints") <- NULL

  # ov_names_data?
  ov_names_data <- attr(flat, "ovda")

  # extra constraints?
  if (!is.null(constraints) && any(nchar(constraints) > 0L)) {
    flat2 <- lavParseModelString(model_syntax = constraints, warn = lav_warn())
    con2 <- attr(flat2, "constraints")
    rm(flat2)
    tmp_con <- c(tmp_con, con2)
  }
  if (length(tmp_con) > 0L) {
    # add 'user' column
    tmp_con <- lapply(tmp_con, function(x) {
      x$user <- 1L
      x
    })
    # any explicit (in)equality constraints? (ignoring := definitions)
    tmp_con_nondef_flag <- (sum(sapply(tmp_con, "[[", "op")
                            %in% c("==", "<", ">")) > 0L)
    # any explicit equality constraints?
    if (tmp_con_nondef_flag) {
      ceq_simple <- FALSE
    }
  }

  if (lav_debug()) {
    cat("[lavaan DEBUG]: flat (flattened user model):\n")
    print(flat)
    cat("[lavaan DEBUG]: tmp_mod (modifiers):\n")
    print(str(tmp_mod))
    cat("[lavaan DEBUG]: tmp_con (constraints):\n")
    print(str(tmp_con))
  }

  # bogus var_table? (if data.type == "none")
  if (!is.null(var_table)) {
    if (!is.list(var_table) || is.null(var_table$name)) {
      lav_msg_stop(gettext(
        "varTable is not a list or does not contain variable names."))
    }
    if (all(var_table$nobs == 0)) {
      var_table <- NULL
    }
  }

  # check for wrongly specified variances/covariances/intercepts
  # of exogenous variables in model syntax (if fixed.x=TRUE)
  if (fixed_x && lav_warn()) { # we ignore the groups here!
    # we only call this function for the warning message
    tmp <- lav_pt_vnames(flat, "ov.x", force_warn = TRUE)
    rm(tmp)
  }

  # check if group.equal is non-empty, but ngroups = 1L
  # fixme: triggers this if mimic="Mplus"!
  # if(ngroups == 1L && length(group.equal) > 0L) {
  #    warning("lavaan WARNING: group.equal= argument",
  #    " has no effect if no groups are specified.")
  # }

  # auto=TRUE?
  auto_cov_x_explicit <- auto_cov_x # was auto.cov.x requested explicitly?
  if (auto) { # mimic sem/cfa auto behavior
    if (model_type == "sem") {
      int_ov_free <- TRUE
      int_lv_free <- FALSE
      auto_fix_first <- !std_lv
      auto_fix_single <- TRUE
      auto_var <- TRUE
      auto_cov_lv_x <- TRUE
      auto_cov_y <- TRUE
      auto_th <- TRUE
      auto_delta <- TRUE
      auto_efa <- TRUE
    } else if (model_type == "growth") {
      int_ov_free <- FALSE
      int_lv_free <- TRUE
      auto_fix_first <- !std_lv
      auto_fix_single <- TRUE
      auto_var <- TRUE
      auto_cov_lv_x <- TRUE
      auto_cov_y <- TRUE
      auto_th <- TRUE
      auto_delta <- TRUE
      auto_efa <- TRUE
    }
  }

  # auto.cov.x implies auto.cov.lv.x; but the covariances between the
  # exogenous latent variables and the observed exogenous covariates cannot
  # be represented if conditional.x = TRUE (the covariates are conditioned
  # out); in that case, auto.cov.x is switched off again -- with a warning,
  # but only if the user asked for auto.cov.x = TRUE explicitly
  if (auto_cov_x) {
    auto_cov_lv_x <- TRUE
    if (conditional_x) {
      auto_cov_x <- FALSE
      if (auto_cov_x_explicit) {
        lav_msg_warn(gettext(
          "auto.cov.x = TRUE has no effect if conditional.x = TRUE;
          covariances between exogenous latent variables and observed
          exogenous covariates will not be added."))
      }
    }
  }

  # check for meanstructure
  if (any(flat$op == "~1")) {
    meanstructure <- TRUE
  }

  # if there is no data (no var_table) and no explicit nthresholds= argument,
  # but the syntax specifies thresholds (the "|" operator), derive the number
  # of thresholds per ordered variable from the syntax. This is needed to set
  # up ordered variables at blocks/levels where the thresholds are not given
  # explicitly (e.g. the between level of a two-level categorical model, which
  # inherits its thresholds from the within level), for instance when
  # simulating data from a model without a dataset (sample.nobs= only).
  if (is.null(var_table) && is.null(nthresholds) && any(flat$op == "|")) {
    th_idx <- which(flat$op == "|")
    # count the *distinct* thresholds per variable; the same threshold may
    # appear more than once in 'flat' (e.g. once per group in a multiple-group
    # model), and should not be counted twice
    tmp_th_pairs <- unique(data.frame(lhs = flat$lhs[th_idx],
                                      rhs = flat$rhs[th_idx],
                                      stringsAsFactors = FALSE))
    tmp_th_tab <- table(tmp_th_pairs$lhs)
    nthresholds <- as.integer(tmp_th_tab)
    names(nthresholds) <- names(tmp_th_tab)
  }

  # check for block identifiers in the syntax (op = ":")
  n_block_flat <- length(which(flat$op == ":"))
  # this is NOT the number of blocks (eg group 1: level 1: -> 1 block)

  # for each non-empty `block' in n_block_flat, produce a USER
  if (n_block_flat > 0L) {
    # make sure flat is a data.frame
    flat <- as.data.frame(flat, stringsAsFactors = FALSE)

    # what are the block lhs labels?
    blocks_lhs_all <- tolower(flat$lhs[flat$op == ":"])
    tmp_blocks_lhs <- unique(blocks_lhs_all)

    # if we have group and level, check that group comes first!
    if ("group" %in% tmp_blocks_lhs && "level" %in% tmp_blocks_lhs) {
      group_idx <- which(tmp_blocks_lhs == "group")
      level_idx <- which(tmp_blocks_lhs == "level")
      if (group_idx > level_idx) {
        lav_msg_stop(gettext(
          "levels must be nested within groups (not the other way around)."))
      }
    }

    # block op == ":" indices
    block_op_idx <- which(flat$op == ":")

    # check for wrong spelled 'group' lhs
    if (length(grep("group", tmp_blocks_lhs, fixed = TRUE)) > 1L) {
      lav_msg_warn(gettext("ambiguous block identifiers for group:"),
        lav_msg_view(tmp_blocks_lhs[grep("group", tmp_blocks_lhs, fixed = TRUE)],
                    "none"))
    }

    # no empty :rhs fields allowed!
    if (any(nchar(flat$rhs[block_op_idx]) == 0L)) {
      empty_idx <- nchar(flat$rhs[block_op_idx]) == 0L
      txt <- paste(flat$lhs[block_op_idx][empty_idx], ":")
      lav_msg_stop(gettext(
        "syntax contains block identifiers with missing numbers/labels: "), txt)
    }

    # check for ngroups (ngroups is based on the data!)
    if ("group" %in% tmp_blocks_lhs) {
      # how many group blocks?
      group_block_idx <- flat$op == ":" & flat$lhs == "group"
      n_group_flat <- length(unique(flat$rhs[group_block_idx]))

      if (n_group_flat > 0L && n_group_flat != ngroups) {
        lav_msg_stop(gettextf(
          "syntax defines %1$s groups; data (or argument ngroups)
          suggests %2$s groups", n_group_flat, ngroups))
      }
    }

    # figure out how many 'blocks' we have, and store indices/block_labels
    tmp_block_rhs <- rep("0", length(tmp_blocks_lhs))
    block_id <- 0L
    block_info <- vector("list", length = n_block_flat) # too large
    block_op_idx1 <- c(block_op_idx, nrow(flat) + 1L) # add addition row
    for (block_op in seq_len(n_block_flat)) {
      # fill block_rhs value(s)
      block_lhs <- flat$lhs[block_op_idx1[block_op]]
      block_rhs <- flat$rhs[block_op_idx1[block_op]]
      tmp_block_rhs[which(block_lhs == tmp_blocks_lhs)] <- block_rhs

      # another block identifier?
      if (block_op_idx1[block_op + 1L] - block_op_idx1[block_op] == 1L) {
        next
      }

      # we have a 'block'
      block_id <- block_id + 1L

      # select flat rows for this block
      tmp_idx <- seq.int(
        block_op_idx1[block_op] + 1L,
        block_op_idx1[block_op + 1L] - 1L
      )

      # store info in block_info
      block_info[[block_id]] <- list(
        lhs = tmp_blocks_lhs, # always the same
        rhs = tmp_block_rhs, # for this block
        idx = tmp_idx
      )
    }
    block_info <- block_info[seq_len(block_id)]

    # new in 0.6-12
    # check for blocks with the same block_rhs combination
    # (perhaps added later?)
    # - merge the indices
    # - remove the duplicated blocks
    block_labels <- sapply(lapply(block_info, "[[", "rhs"),
      paste,
      collapse = "."
    )
    nblocks <- length(unique(block_labels))
    if (nblocks < length(block_labels)) {
      # it would appear we have duplicated block_labels -> merge
      dup_idx <- which(duplicated(block_labels))
      for (i in seq_along(dup_idx)) {
        this_dup_idx <- dup_idx[i]
        orig_idx <- which(block_labels == block_labels[this_dup_idx])[1]
        block_info[[orig_idx]]$idx <- c(
          block_info[[orig_idx]]$idx,
          block_info[[this_dup_idx]]$idx
        )
      }
      block_info <- block_info[-dup_idx]
    }

    # split the flat data.frame per `block', create tmp_list
    # for each `block', and rbind them together, adding block columns
    for (block in seq_len(nblocks)) {
      tmp_block_rhs <- block_info[[block]]$rhs
      block_lhs <- block_info[[block]]$lhs[length(tmp_blocks_lhs)] # last one
      block_idx <- block_info[[block]]$idx

      flat_block <- flat[block_idx, ]
      # rm 'block' column (if any) in flat_block
      flat_block$block <- NULL

      # new in 0.6-7: check for random slopes, add them here
      if (block_lhs == "level" &&
        block > 1L && # FIXME: multigroup, multilevel
        !is.null(flat$rv) &&
        any(nchar(flat$rv) > 0L)) {
        lv_names_rv <- unique(flat$rv[nchar(flat$rv) > 0L])
        for (i in seq_along(lv_names_rv)) {
          # add phantom latent variable
          tmp <- flat_block[1, ]
          tmp$lhs <- lv_names_rv[i]
          tmp$op <- "=~"
          tmp$rhs <- lv_names_rv[i]
          tmp$mod.idx <- max(flat$mod.idx) + i
          # the modifier columns below are only present in the FLAT
          # matrix of the original parser; the newer parsers use the
          # 'modifiers' attribute only (see tmp_mod below)
          if (!is.null(flat_block$fixed)) tmp$fixed <- "0"
          if (!is.null(flat_block$start)) tmp$start <- ""
          if (!is.null(flat_block$lower)) tmp$lower <- ""
          if (!is.null(flat_block$upper)) tmp$upper <- ""
          if (!is.null(flat_block$label)) tmp$label <- ""
          if (!is.null(flat_block$prior)) tmp$prior <- ""
          if (!is.null(flat_block$efa)) tmp$efa <- ""
          tmp$rv <- lv_names_rv[i]
          flat_block <- rbind(flat_block, tmp, deparse.level = 0L)
          tmp_mod <- c(tmp_mod, list(list(fixed = 0)))
        }
      }

      # new in 0.6-8: if multilevel, use 'global' ov.names.x
      if (fixed_x && block_lhs == "level") {
        tmp_ov_names_x <- lav_pt_vnames(flat, "ov.x") # global
        ov_names_x_block <- lav_pt_vnames(flat_block, "ov.x")
        if (length(ov_names_x_block) > 0L) {
          idx <- which(!ov_names_x_block %in% tmp_ov_names_x)
          if (length(idx) > 0L) {
            # warn!
            lav_msg_warn(gettextf(
              "the variable(s) [%s] are exogenous at one level, but endogenous
              at another level. These variables will be treated as endogenous,
              and their variances/intercepts will be freely estimated.
              To remove this warning, use fixed.x = FALSE.",
              lav_msg_view(ov_names_x_block[idx], "none")))
            ov_names_x_block <- ov_names_x_block[-idx]
          }
        }
      } else {
        ov_names_x_block <- NULL
      }

      # new in 0.6-12: if multilevel and conditional.x, make sure
      # that 'split' exogenous covariates become 'y' variables
      if (conditional_x && block_lhs == "level") {
        if (ngroups == 1L) {
          other_block_names <- lav_pt_vnames(flat, "ov",
            block = seq_len(nblocks)[-block]
          )
        } else {
          nlevels_pt <- nblocks %/% ngroups
          this_group <- ceiling(block / nlevels_pt)
          blocks_within_group <- (this_group - 1L) * nlevels_pt +
            seq_len(nlevels_pt)
          other_block_names <- lav_pt_vnames(flat, "ov",
            block = blocks_within_group[blocks_within_group != block]
          )
        }
        ov_names_x_block <- lav_pt_vnames(flat_block, "ov.x")
        if (length(ov_names_x_block) > 0L) {
          idx <- which(ov_names_x_block %in% other_block_names)
          if (length(idx) > 0L) {
            ov_names_x_block <- ov_names_x_block[-idx]
          }
        }
      } else {
        ov_names_x_block <- NULL
      }

      list_block <- lav_pt_flat(flat_block,
        blocks = tmp_blocks_lhs,
        block_id = block,
        meanstructure = meanstructure,
        int_ov_free = int_ov_free, int_lv_free = int_lv_free,
        orthogonal = orthogonal, orthogonal_y = orthogonal_y,
        orthogonal_x = orthogonal_x, orthogonal_efa = orthogonal_efa,
        std_lv = std_lv, correlation = correlation, composites = composites,
        composites_cov_free = composites_cov_free,
        conditional_x = conditional_x, fixed_x = fixed_x,
        parameterization = parameterization,
        auto_fix_first = auto_fix_first, marker = marker,
        auto_fix_single = auto_fix_single,
        auto_var = auto_var, auto_cov_lv_x = auto_cov_lv_x,
        auto_cov_x = auto_cov_x,
        auto_cov_y = auto_cov_y, auto_th = auto_th,
        auto_delta = auto_delta, auto_efa = auto_efa,
        var_table = var_table, group_equal = NULL,
        group_w_free = group_w_free, ngroups = 1L,
        nthresholds = nthresholds,
        ov_names_x_block = ov_names_x_block
      )
      list_block <- as.data.frame(list_block, stringsAsFactors = FALSE)

      # add block columns with current values in block_rhs
      for (b in seq_along(tmp_blocks_lhs)) {
        block_lhs <- tmp_blocks_lhs[b]
        block_rhs <- tmp_block_rhs[b]
        list_block[block_lhs] <- rep(block_rhs, length(list_block$lhs))
      }

      if (!exists("tmp_list")) {
        tmp_list <- list_block
      } else {
        list_block$id <- list_block$id + max(tmp_list$id)
        tmp_list <- rbind(tmp_list, list_block)
      }
    }
    tmp_list <- as.list(tmp_list)

    # convert block columns to integers if possible
    for (b in seq_along(tmp_blocks_lhs)) {
      block_lhs <- tmp_blocks_lhs[b]
      block_rhs <- tmp_block_rhs[b]
      tmp <- try(scan(
        text = tmp_list[[block_lhs]], what = integer(),
        quiet = TRUE
      ), silent = TRUE)
      if (inherits(tmp, "integer")) {
        tmp_list[[block_lhs]] <- tmp
      }
    }
  } else {
    tmp_list <- lav_pt_flat(flat,
      blocks = "group",
      meanstructure = meanstructure,
      int_ov_free = int_ov_free, int_lv_free = int_lv_free,
      orthogonal = orthogonal, orthogonal_y = orthogonal_y,
      orthogonal_x = orthogonal_x, orthogonal_efa = orthogonal_efa,
      std_lv = std_lv, correlation = correlation, composites = composites,
      composites_cov_free = composites_cov_free,
      conditional_x = conditional_x, fixed_x = fixed_x,
      parameterization = parameterization,
      auto_fix_first = auto_fix_first, marker = marker,
      auto_fix_single = auto_fix_single,
      auto_var = auto_var, auto_cov_lv_x = auto_cov_lv_x,
      auto_cov_x = auto_cov_x,
      auto_cov_y = auto_cov_y, auto_th = auto_th,
      auto_delta = auto_delta, auto_efa = auto_efa,
      var_table = var_table, group_equal = group_equal,
      group_w_free = group_w_free,
      ngroups = ngroups, nthresholds = nthresholds
    )
  }
  if (lav_debug()) {
    cat("[lavaan DEBUG]: parameter tmp_list without MODIFIERS:\n")
    print(as.data.frame(tmp_list, stringsAsFactors = FALSE))
  }

  # check ordinal variables
  categorical <- FALSE
  ov_ord <- lav_object_vnames(tmp_list, "ov.ord")
  all_ord <- all(lav_object_vnames(tmp_list, "ov") %in% ov_ord)
  if (length(ov_ord) > 0L) {
    categorical <- TRUE
    ord_var_idx <- which(tmp_list$op == "~~" &
                         tmp_list$lhs == tmp_list$rhs &
                         tmp_list$lhs %in% ov_ord &
                         tmp_list$user == 1L)
    if (parameterization == "delta" && length(ord_var_idx) > 0L) {
      lav_msg_warn(gettextf("variances of ordered variables are ignored when
      parameterization = \"delta\"; please remove them from the model syntax
      or use parameterization = \"theta\"; variables involved are: %s",
      paste(tmp_list$lhs[ord_var_idx], collapse = " ")))
      # force them to be nonfree and set ustart to 1 (new in 0.6-20)
      # later, after we have processes the modifiers
    }
  } # ov_ord

  # handle multilevel-specific constraints
  multilevel <- FALSE
  nlevels <- 1L
  if (!is.null(tmp_list$level)) {
    nlevels <- lav_pt_nlevels(tmp_list)
    if (nlevels > 1L) {
      multilevel <- TRUE
    }
  }
  if (multilevel && any(tmp_list$op == "~1")) {
    # fix ov intercepts for all within ov that also appear at level 2
    # FIXME: not tested with > 2 levels
    ov_names <- lav_pt_vnames(tmp_list, "ov") ## all names
    level_values <- lav_pt_level_values(tmp_list)
    other_names <- tmp_list$lhs[tmp_list$op == "~1" &
      tmp_list$level %in% level_values[-1L] &
      tmp_list$lhs %in% ov_names]
    fix_names_idx <- which(tmp_list$op == "~1" &
      tmp_list$level %in% level_values[1L] &
      tmp_list$lhs %in% other_names)
    if (length(fix_names_idx) > 0L) {
      tmp_list$free[fix_names_idx] <- 0L
      tmp_list$ustart[fix_names_idx] <- 0
    }
  }
  if (multilevel && any(tmp_list$op == "|")) {
    # fix ALL thresholds at level 1
    level_values <- lav_pt_level_values(tmp_list)
    th_idx <- which(tmp_list$op == "|" &
      tmp_list$level %in% level_values[1L])
    tmp_list$free[th_idx] <- 0L
    tmp_list$ustart[th_idx] <- 0

    # fix ALL scaling parameters at higher levels
    scale_idx <- which(tmp_list$op == "~*~" &
      tmp_list$level %in% level_values[-1L])
    tmp_list$free[scale_idx] <- 0L
    tmp_list$ustart[scale_idx] <- 1

    # theta parameterization: the residual variances of the ordinal
    # variables are fixed to 1 at the WITHIN level only (setting the
    # scale of the latent responses); at the between level they are free
    if (parameterization == "theta") {
      ord_bvar_idx <- which(tmp_list$op == "~~" &
        tmp_list$lhs == tmp_list$rhs &
        tmp_list$lhs %in% ov_ord &
        tmp_list$level %in% level_values[-1L] &
        tmp_list$user == 0L)
      if (length(ord_bvar_idx) > 0L) {
        tmp_list$free[ord_bvar_idx] <- 1L
        tmp_list$ustart[ord_bvar_idx] <- as.numeric(NA)
      }
    }
  }

  # apply user-specified modifiers
  warn_about_single_label <- FALSE
  if (length(tmp_mod)) {
    for (el in seq_along(tmp_mod)) {
      idx <- which(tmp_list$mod.idx == el) # for each group

      # 0.5-21: check if idx exists
      # perhaps the corresponding element was duplicated, and removed
      if (length(idx) == 0L) {
        next
      }

      tmp_mod_fixed <- tmp_mod[[el]]$fixed
      tmp_mod_start <- tmp_mod[[el]]$start
      tmp_mod_lower <- tmp_mod[[el]]$lower
      tmp_mod_upper <- tmp_mod[[el]]$upper
      tmp_mod_label <- tmp_mod[[el]]$label
      tmp_mod_prior <- tmp_mod[[el]]$prior
      tmp_mod_efa <- tmp_mod[[el]]$efa
      tmp_mod_rv <- tmp_mod[[el]]$rv

      # check for single argument if multiple groups
      if (ngroups > 1L && length(idx) > 1L) {
        # Ok, this is not very consistent:
        # A) here we force same behavior across groups
        if (length(tmp_mod_fixed) == 1L) {
          tmp_mod_fixed <- rep(tmp_mod_fixed, ngroups)
        }
        if (length(tmp_mod_start) == 1L) {
          tmp_mod_start <- rep(tmp_mod_start, ngroups)
        }
        if (length(tmp_mod_lower) == 1L) {
          tmp_mod_lower <- rep(tmp_mod_lower, ngroups)
        }
        if (length(tmp_mod_upper) == 1L) {
          tmp_mod_upper <- rep(tmp_mod_upper, ngroups)
        }
        if (length(tmp_mod_prior) == 1L) {
          tmp_mod_prior <- rep(tmp_mod_prior, ngroups)
        }
        if (length(tmp_mod_efa) == 1L) {
          tmp_mod_efa <- rep(tmp_mod_efa, ngroups)
        }
        if (length(tmp_mod_rv) == 1L) {
          tmp_mod_rv <- rep(tmp_mod_rv, ngroups)
        }

        # new in 0.6-7 (proposal):
        # - always recycle modifiers, including labels
        # - if ngroups > 1 AND group.label= is empty, produce a warning
        #   (as this is a break from < 0.6-6)
        if (length(tmp_mod_label) == 1L) {
          tmp_mod_label <- rep(tmp_mod_label, ngroups)
          if (is.null(group_equal) || length(group_equal) == 0L) {
            warn_about_single_label <- TRUE
          }
        }

        # < 0.6-7 code:
        # B) here we do NOT! otherwise, it would imply an equality
        #                    constraint...
        #    except if group.equal="loadings"!
        # if(length(tmp_mod_label) == 1L) {
        #    if("loadings" %in% group.equal ||
        #       "composite.loadings" %in% group.equal) {
        #        tmp_mod_label <- rep(tmp_mod_label, ngroups)
        #    } else {
        #        tmp_mod_label <- c(tmp_mod_label, rep("", (ngroups-1L)) )
        #    }
        # }
      }

      # check for wrong number of arguments if multiple groups
      nidx <- length(idx)
      if ((!is.null(tmp_mod_fixed) && nidx != length(tmp_mod_fixed)) ||
        (!is.null(tmp_mod_start) && nidx != length(tmp_mod_start)) ||
        (!is.null(tmp_mod_lower) && nidx != length(tmp_mod_lower)) ||
        (!is.null(tmp_mod_upper) && nidx != length(tmp_mod_upper)) ||
        (!is.null(tmp_mod_prior) && nidx != length(tmp_mod_prior)) ||
        (!is.null(tmp_mod_efa) && nidx != length(tmp_mod_efa)) ||
        (!is.null(tmp_mod_rv) && nidx != length(tmp_mod_rv)) ||
        (!is.null(tmp_mod_label) && nidx != length(tmp_mod_label))) {
        el_idx <- which(tmp_list$mod.idx == el)[1L]
        lav_msg_stop(gettextf(
          "wrong number of arguments in modifier (%s) of element",
          lav_msg_view(tmp_mod_label, "none")),
          tmp_list$lhs[el_idx], tmp_list$op[el_idx], tmp_list$rhs[el_idx]
        )
      }

      # apply modifiers
      if (!is.null(tmp_mod_fixed)) {
        # two options: constant or NA
        na_idx <- which(is.na(tmp_mod_fixed))
        not_na_idx <- which(!is.na(tmp_mod_fixed))

        # constant
        tmp_list$ustart[idx][not_na_idx] <- tmp_mod_fixed[not_na_idx]
        tmp_list$free[idx][not_na_idx] <- 0L

        # NA* modifier
        tmp_list$free[idx][na_idx] <- 1L # eg factor loading
        tmp_list$ustart[idx][na_idx] <- as.numeric(NA)
      }
      if (!is.null(tmp_mod_start)) {
        tmp_list$ustart[idx] <- tmp_mod_start
      }
      if (!is.null(tmp_mod_prior)) {
        # do we already have a `prior' column? if not, create one
        if (is.null(tmp_list$prior)) {
          tmp_list$prior <- character(length(tmp_list$lhs))
        }
        tmp_list$prior[idx] <- tmp_mod_prior
      }
      if (!is.null(tmp_mod_efa)) {
        # do we already have a `efa' column? if not, create one
        if (is.null(tmp_list$efa)) {
          tmp_list$efa <- character(length(tmp_list$lhs))
        }
        tmp_list$efa[idx] <- tmp_mod_efa
      }
      if (!is.null(tmp_mod_rv)) {
        # do we already have a `rv' column? if not, create one
        if (is.null(tmp_list$rv)) {
          tmp_list$rv <- character(length(tmp_list$lhs))
        }
        tmp_list$rv[idx] <- tmp_mod_rv

        tmp_list$free[idx] <- 0L
        tmp_list$ustart[idx] <- as.numeric(NA) #
      }
      if (!is.null(tmp_mod_lower)) {
        # do we already have a `lower' column? if not, create one
        if (is.null(tmp_list$lower)) {
          tmp_list$lower <- rep(-Inf, length(tmp_list$lhs))
        }
        tmp_list$lower[idx] <- as.numeric(tmp_mod_lower)
      }
      if (!is.null(tmp_mod_upper)) {
        # do we already have a `upper' column? if not, create one
        if (is.null(tmp_list$upper)) {
          tmp_list$upper <- rep(Inf, length(tmp_list$lhs))
        }
        tmp_list$upper[idx] <- as.numeric(tmp_mod_upper)
      }
      if (!is.null(tmp_mod_label)) {
        tmp_list$label[idx] <- tmp_mod_label
      }
    }
  }
  # remove mod.idx column
  tmp_list$mod.idx <- NULL

  # categorical: check for nonfree variances if parameterization = "delta"
  # we already gave warning; here, we force them to be nonfree
  if (categorical && parameterization == "delta") {
    ord_var_idx <- which(tmp_list$op == "~~" &
                         tmp_list$lhs == tmp_list$rhs &
                         tmp_list$lhs %in% ov_ord &
                         tmp_list$user == 1L)
    if (length(ord_var_idx) > 0L) {
      # force them to be nonfree and set ustart to 1 (new in 0.6-20)
      tmp_list$free[ord_var_idx]   <- rep(0L, length(ord_var_idx))
      tmp_list$ustart[ord_var_idx] <- rep(1,  length(ord_var_idx))
    }
  } # categorical


  # warning about single label in multiple group setting?
  if (warn_about_single_label) {
    lav_msg_warn(gettext(
      "using a single label per parameter in a multiple group setting implies
      imposing equality constraints across all the groups; If this is not
      intended, either remove the label(s), or use a vector of labels (one for
      each group); See the Multiple groups section in the man page of
      model.syntax."
    ))
  }

  # if lower/upper values were added, fix non-free values to ustart values
  # new in 0.6-6
  if (!is.null(tmp_list$lower)) {
    fixed_idx <- which(tmp_list$free == 0L)
    if (length(fixed_idx) > 0L) {
      tmp_list$lower[fixed_idx] <- tmp_list$ustart[fixed_idx]
    }
  }
  if (!is.null(tmp_list$upper)) {
    fixed_idx <- which(tmp_list$free == 0L)
    if (length(fixed_idx) > 0L) {
      tmp_list$upper[fixed_idx] <- tmp_list$ustart[fixed_idx]
    }
  }

  # if rv column is present, add rv_names to ALL rows where they are used
  if (!is.null(tmp_list$rv)) {
    rv_names <- unique(tmp_list$rv[nchar(tmp_list$rv) > 0L])
    for (i in seq_along(rv_names)) {
      lhs_idx <- which(tmp_list$lhs == rv_names[i] &
        tmp_list$op == "=~")
      tmp_list$rv[lhs_idx] <- rv_names[i]
    }
  }

  if (lav_debug()) {
    cat("[lavaan DEBUG]: parameter tmp_list with MODIFIERS:\n")
    print(as.data.frame(tmp_list, stringsAsFactors = FALSE))
  }

  # get 'virtual' parameter labels
  if (n_block_flat > 1L) {
    blocks <- tmp_blocks_lhs
  } else {
    blocks <- "group"
  }
  label <- lav_pt_labels(
    partable = tmp_list,
    blocks = blocks,
    group_equal = group_equal,
    group_partial = group_partial
  )

  if (lav_debug()) {
    cat("[lavaan DEBUG]: parameter tmp_list with LABELS:\n")
    tmp <- tmp_list
    tmp$label <- label
    print(as.data.frame(tmp, stringsAsFactors = FALSE))
  }

  # handle EFA equality constraints
  # YR 14 Jan 2020: 0.6-6 does no longer impose 'explicit' constraints
  #                 if we only need to fix a parameter to 0/1
  # Note: we should also check if they are really needed:
  #       eg., if all the factor-loadings of the 'second' set (time/group)
  #       are constrained to be equal to the factor-loadings of the first
  #       set, no further constraints are needed
  if (auto_efa && !is.null(tmp_list$efa)) {
    tmp_list <- lav_pt_efa_con(
      list_1 = tmp_list,
      orthogonal_efa = orthogonal_efa,
      group_equal = group_equal
    )
  } # auto_efa

  # EFA/ESEM: do not use the 'simple equality constraint' shortcut when the
  # model contains EFA blocks. The rotation machinery adds its own (efa)
  # constraints, so ceq.simple.only is never TRUE here; collapsing the user
  # equality constraints would then drop them entirely (neither '==' rows nor a
  # shared free index), leaving the vcov rank-deficient. Keep them as explicit
  # '==' constraints instead (the pre-ceq.simple behaviour).
  if (ceq_simple && !is.null(tmp_list$efa) &&
      any(nchar(tmp_list$efa) > 0L)) {
    ceq_simple <- FALSE
  }

  # handle user-specified equality constraints
  # lavaan 0.6-11:
  #     two settings:
  #       1) simple equality constraints ONLY -> back to basics: only
  #          duplicate 'free' numbers; no longer explicit == rows with plabels
  #       2) mixture of simple and other (explicit) constraints
  #          treat them together as we did in <0.6-11
  tmp_list$plabel <- paste(".p", tmp_list$id, ".", sep = "")
  eq_labels <- unique(label[duplicated(label)])
  eq_id <- integer(length(tmp_list$lhs))
  for (eq_label in eq_labels) {
    tmp_con_idx <- length(tmp_con)
    all_idx <- which(label == eq_label) # all same-label parameters
    ref_idx <- all_idx[1L] # the first one only
    other_idx <- all_idx[-1L] # the others
    eq_id[all_idx] <- ref_idx

    # new in 0.6-6: make sure lower/upper constraints are equal too
    if (!is.null(tmp_list$lower) &&
      length(unique(tmp_list$lower[all_idx])) > 0L) {
      non_inf <- which(is.finite(tmp_list$lower[all_idx]))
      if (length(non_inf) > 0L) {
        smalles_val <- min(tmp_list$lower[all_idx][non_inf])
        tmp_list$lower[all_idx] <- smalles_val
      }
    }
    if (!is.null(tmp_list$upper) &&
      length(unique(tmp_list$upper[all_idx])) > 0L) {
      non_inf <- which(is.finite(tmp_list$upper[all_idx]))
      if (length(non_inf) > 0L) {
        largest_val <- max(tmp_list$upper[all_idx][non_inf])
        tmp_list$upper[all_idx] <- largest_val
      }
    }

    # two possibilities:
    #   1. all_idx contains a fixed parameter: in this case,
    #      we fix them all (hopefully to the same value)
    #   2. all_idx contains only free parameters

    # 1. all_idx contains a fixed parameter
    if (any(tmp_list$free[all_idx] == 0L)) {
      # which one is fixed?
      fixed_all <- all_idx[tmp_list$free[all_idx] == 0L]
      # only pick the first
      fixed_idx <- fixed_all[1]

      # sanity check: are all ustart values equal?
      ustart1 <- tmp_list$ustart[fixed_idx]
      if (all(is.na(tmp_list$ustart[fixed_all]))) {
        # nothing to do; ustart values have not been set yet
      } else if (!all(ustart1 == tmp_list$ustart[fixed_all])) {
        lav_msg_warn(gettext(
          "equality constraints involve fixed parameters with different values;
          only the first one will be used"))
      }

      # make them all fixed
      tmp_list$ustart[all_idx] <- tmp_list$ustart[fixed_idx]
      tmp_list$free[all_idx] <- 0L # not free anymore, since it must
      # be equal to the 'fixed' parameter
      # (Note: Mplus ignores this)
      eq_id[all_idx] <- 0L # remove from eq_id list

      # new in 0.6-8 (for efa + user-specified eq constraints)
      if (any(tmp_list$user[all_idx] %in% c(7L, 77L))) {
        # if involved in an efa block, store in tmp_con anyway
        # we may need it for the rotated solution
        for (o in other_idx) {
          tmp_con_idx <- tmp_con_idx + 1L
          tmp_con[[tmp_con_idx]] <- list(
            op = "==",
            lhs = tmp_list$plabel[ref_idx],
            rhs = tmp_list$plabel[o],
            user = 2L
          )
        }
      }
    } else {
      # 2. all_idx contains only free parameters
      # old system:
      # - add tmp_con entry
      # - in 0.6-11: only if tmp_con is not empty
      if (!ceq_simple) {
        for (o in other_idx) {
          tmp_con_idx <- tmp_con_idx + 1L
          tmp_con[[tmp_con_idx]] <- list(
            op = "==",
            lhs = tmp_list$plabel[ref_idx],
            rhs = tmp_list$plabel[o],
            user = 2L
          )
        }
      } else {
        # new system:
        # - set $free elements to zero, and later to ref id
        tmp_list$free[other_idx] <- 0L # all but the first are non-free
        # but will get a duplicated number
      }

      # just to trick semTools, also add something in the label
      # colum, *if* it is empty
      # update: 0.6-11 we keep this, because it shows the plabels
      #         when eg group.equal = "loadings"
      for (i in all_idx) {
        if (nchar(tmp_list$label[i]) == 0L) {
          tmp_list$label[i] <- tmp_list$plabel[ref_idx]
        }
      }
    } # all free
  } # eq in eq_labels
  if (lav_debug()) {
    print(tmp_con)
  }



  # handle constraints (if any) (NOT per group, but overall - 0.4-11)
  if (length(tmp_con) > 0L) {
    n_con <- length(tmp_con)
    tmp_idx <- length(tmp_list$id) + seq_len(n_con)
    # grow tmp_list with length(tmp_con) extra rows
    tmp_list <- lapply(tmp_list, function(x) {
      if (is.character(x)) {
        c(x, rep("", n_con))
      } else {
        c(x, rep(NA, n_con))
      }
    })

    # fill in some columns
    tmp_list$id[tmp_idx] <- tmp_idx
    tmp_list$lhs[tmp_idx] <- unlist(lapply(tmp_con, "[[", "lhs"))
    tmp_list$op[tmp_idx] <- unlist(lapply(tmp_con, "[[", "op"))
    tmp_list$rhs[tmp_idx] <- unlist(lapply(tmp_con, "[[", "rhs"))
    tmp_list$user[tmp_idx] <- unlist(lapply(tmp_con, "[[", "user"))

    # zero is nicer?
    tmp_list$free[tmp_idx] <- rep(0L, n_con)
    tmp_list$exo[tmp_idx] <- rep(0L, n_con)
    tmp_list$block[tmp_idx] <- rep(0L, n_con)

    if (!is.null(tmp_list$group)) {
      if (is.character(tmp_list$group)) {
        tmp_list$group[tmp_idx] <- rep("", n_con)
      } else {
        tmp_list$group[tmp_idx] <- rep(0L, n_con)
      }
    }
    if (!is.null(tmp_list$level)) {
      if (is.character(tmp_list$level)) {
        tmp_list$level[tmp_idx] <- rep("", n_con)
      } else {
        tmp_list$level[tmp_idx] <- rep(0L, n_con)
      }
    }
    if (!is.null(tmp_list$class)) {
      if (is.character(tmp_list$class)) {
        tmp_list$class[tmp_idx] <- rep("", n_con)
      } else {
        tmp_list$class[tmp_idx] <- rep(0L, n_con)
      }
    }
  }

  # check defined variables (:=)
  def_idx <- which(tmp_list$op == ":=")
  if (length(def_idx) > 0L) {
    # check if the lhs is unique (new in 0.6-20)
    def_lhs <- tmp_list$lhs[def_idx]
    dup_idx <- which(duplicated(def_lhs))
    if (length(dup_idx) > 0L) {
      # warn or stop? warn for now
      lav_msg_warn(gettextf("at least one defined variable (using the :=
                             operator) has been duplicated, and will be
                             overwritten by the last one: %s",
                            paste(def_lhs[dup_idx], collapse = " ")))
    }
    # put lhs of := elements in label column
    tmp_list$label[def_idx] <- def_lhs
  }


  # handle effect_coding related equality constraints
  if (is.logical(effect_coding) && effect_coding) {
    effect_coding <- c("loadings", "intercepts")
  } else if (!is.character(effect_coding)) {
    lav_msg_stop(gettext("effect.coding argument must be a character string"))
  }
  # in ordinal models, do integer coding
  if (all_ord) effect_coding <- c(effect_coding, "thresholds")
  if (any(c("loadings", "intercepts") %in% effect_coding)) {
    tmp <- list()
    # for each block
    nblocks <- lav_pt_nblocks(tmp_list)
    for (b in seq_len(nblocks)) {

      # which group?
      this_group <- floor(b / nlevels + 0.5)

      # lv's for this block/set
      lv_names <- unique(tmp_list$lhs[tmp_list$op == "=~" &
        tmp_list$block == b])

      if (length(lv_names) == 0L) {
        next
      }

      for (lv in lv_names) {
        # ind_names
        ind_names <- tmp_list$rhs[tmp_list$op == "=~" &
          tmp_list$block == b &
          tmp_list$lhs == lv]

        if ("loadings" %in% effect_coding &&
            (!"loadings" %in% group_equal || this_group == 1L)) {
          # factor loadings indicators of this lv
          loadings_idx <- which(tmp_list$op == "=~" &
            tmp_list$block == b &
            tmp_list$rhs %in% ind_names &
            tmp_list$lhs == lv)

          # all free?
          if (length(loadings_idx) > 0L &&
            all(tmp_list$free[loadings_idx] > 0L)) {
            # add eq constraint
            plabel <- tmp_list$plabel[loadings_idx]

            # Note:  we write them as
            # .p1. == 3 - .p2. - .p3.
            # instead of
            # 3 ==  .p1.+.p2.+.p3.
            # as this makes it easier to translate things to
            # JAGS/stan

            tmp_lhs <- plabel[1]
            if (length(loadings_idx) > 1L) {
              tmp_rhs <- paste(length(loadings_idx), "-",
                paste(plabel[-1], collapse = "-"),
                sep = ""
              )
            } else {
              tmp_rhs <- length(loadings_idx)
            }

            tmp$lhs <- c(tmp$lhs, tmp_lhs)
            tmp$op <- c(tmp$op, "==")
            tmp$rhs <- c(tmp$rhs, tmp_rhs)
            tmp$block <- c(tmp$block, 0L)
            tmp$user <- c(tmp$user, 2L)
            tmp$ustart <- c(tmp$ustart, as.numeric(NA))
          }
        } # loadings

        if ("thresholds" %in% effect_coding) {
          thresholds_idx <- which(tmp_list$op == "|" &
                                  tmp_list$block == b &
                                  tmp_list$lhs %in% ind_names)

          nlevs <- table(tmp_list$lhs[thresholds_idx]) + 1
          kmax <- max(nlevs)

          if (!"thresholds" %in% group_equal || this_group == 1L) {
            # all free?
            if (length(thresholds_idx) > 0L && all(tmp_list$free[thresholds_idx] > 0L)) {
              for (ind in ind_names) {
                # 1) fix bottom and top per ov
                trows <- which(tmp_list$lhs[thresholds_idx] == ind)
                thisidx <- thresholds_idx[trows[1]]
                tmp_list$free[thisidx] <- 0L
                tmp_list$ustart[thisidx] <- round(.5 + kmax / nlevs[ind], 3)
                tmp_list$user[thisidx] <- 2L

                if (nlevs[ind] > 2L) {
                  thisidx <- thresholds_idx[trows[length(trows)]]
                  tmp_list$free[thisidx] <- 0L
                  tmp_list$ustart[thisidx] <- round(.5 + kmax * (nlevs[ind] - 1) /
                                                    nlevs[ind], 3)
                  tmp_list$user[thisidx] <- 2L
                }
              }
            }
          }

          for (ind in ind_names) {
            # 2) free intercepts of variables with > 2 levels,
            #                     only if automatically fixed
            intercept_idx <- which(tmp_list$op == "~1" &
                tmp_list$block == b &
                tmp_list$lhs == ind)
            if (nlevs[ind] > 2L && length(intercept_idx) > 0L &&
                tmp_list$user[intercept_idx] == 0L) {
              tmp_list$free[intercept_idx] <- 1L
            }

            # 3) free variances/deltas of variables with > 2 levels,
            #                           only if automatically fixed
            varop <- ifelse(parameterization == "delta", "~*~", "~~")
            var_idx <- which(tmp_list$op == varop &
                          tmp_list$block == b &
                          tmp_list$lhs == ind &
                          tmp_list$lhs == tmp_list$rhs)
            if (length(var_idx) > 0L && tmp_list$user[var_idx] == 0L) {
                 tmp_list$free[var_idx] <- 1L
            }
          }
          # 4) fix latent variance if all variables have 2 levels
          if (all(nlevs == 2L)) {
            lv_var_idx <- which(tmp_list$op == "~~" &
                                tmp_list$block == b &
                                tmp_list$lhs == lv &
                                tmp_list$lhs == tmp_list$rhs)
            if (length(lv_var_idx) > 0L && tmp_list$user[lv_var_idx] == 0L) {
              tmp_list$free[lv_var_idx] <- 0L
              tmp_list$ustart[lv_var_idx] <- 1L
            }
          }
        } # thresholds

        if ("intercepts" %in% effect_coding &&
            (!"intercepts" %in% group_equal || this_group == 1L)) {
          # intercepts for indicators of this lv
          intercepts_idx <- which(tmp_list$op == "~1" &
            tmp_list$block == b &
            tmp_list$lhs %in% ind_names)

          # all free?
          if (length(intercepts_idx) > 0L &&
            (all(tmp_list$free[intercepts_idx] > 0L) ||
              "thresholds" %in% effect_coding)) {
            # 1) add eq constraint
            plabel <- tmp_list$plabel[intercepts_idx]

            tmp_lhs <- plabel[1]
            if (length(intercepts_idx) > 1L) {
              tmp_rhs <- paste("0-",
                paste(plabel[-1], collapse = "-"),
                sep = ""
              )
            } else {
              tmp_rhs <- 0L
            }

            tmp$lhs <- c(tmp$lhs, tmp_lhs)
            tmp$op <- c(tmp$op, "==")
            tmp$rhs <- c(tmp$rhs, tmp_rhs)
            tmp$block <- c(tmp$block, 0L)
            tmp$user <- c(tmp$user, 2L)
            tmp$ustart <- c(tmp$ustart, as.numeric(NA))

            # 2) release latent mean
            lv_int_idx <- which(tmp_list$op == "~1" &
              tmp_list$block == b &
              tmp_list$lhs == lv)
            # free only if automatically added
            if (length(lv_int_idx) > 0L &&
              tmp_list$user[lv_int_idx] == 0L) {
              tmp_list$free[lv_int_idx] <- 1L
            }
          }
        } else if ("intercepts" %in% effect_coding &&
                   "intercepts" %in% group_equal &&
                   this_group > 1L) {
          ## add equality constraints of the newly-freed intercepts
          intercepts_idx <- which(tmp_list$op == "~1" &
                                  tmp_list$block == b &
                                  tmp_list$lhs %in% ind_names)
          intercepts_idx_g1 <- which(tmp_list$op == "~1" &
                                     tmp_list$block == (1L + b%%2) &
                                     tmp_list$lhs %in% ind_names)
          ncon <- length(intercepts_idx)

          tmp$lhs <- c(tmp$lhs, tmp_list$plabel[intercepts_idx_g1])
          tmp$op <- c(tmp$op, rep("==", ncon))
          tmp$rhs <- c(tmp$rhs, tmp_list$plabel[intercepts_idx])
          tmp$block <- c(tmp$block, rep(0L, ncon))
          tmp$user <- c(tmp$user, rep(2L, ncon))
          tmp$ustart <- c(tmp$ustart, rep(as.numeric(NA), ncon))
        } # intercepts
      } # lv
    } # blocks

    tmp_list <- lav_pt_merge(tmp_list, tmp)
  }

  # marker.int.zero
  if (meanstructure && marker_int_zero) {
    # for each block
    nblocks <- lav_pt_nblocks(tmp_list)
    for (b in seq_len(nblocks)) {
      # lv's for this block/set
      lv_names <- lav_pt_vnames(tmp_list,
        type = "lv.regular",
        block = b
      )
      lv_marker <- lav_pt_vnames(tmp_list,
        type = "lv.marker",
        block = b
      )
      ov_num <- lav_pt_vnames(tmp_list,
        type = "ov.num",
        block = b
      )
      # new in 0.7-1: only consider ov_num markers
      lv_marker <- lv_marker[lv_marker %in% ov_num]

      if (length(lv_marker) == 0L) {
        next
      }

      # fix marker intercepts to zero
      marker_idx <- which(tmp_list$op == "~1" &
        tmp_list$lhs %in% lv_marker & tmp_list$block == b &
        tmp_list$user == 0L)
      tmp_list$free[marker_idx] <- 0L
      tmp_list$ustart[marker_idx] <- 0

      # free latent means
      lv_idx <- which(tmp_list$op == "~1" &
        tmp_list$lhs %in% lv_names & tmp_list$block == b &
        tmp_list$user == 0L)
      tmp_list$free[lv_idx] <- 1L
      tmp_list$ustart[lv_idx] <- as.numeric(NA)
    } # block
  }


  # mg.lv.variances
  if (ngroups > 1L && "mg.lv.variances" %in% effect_coding) {
    tmp <- list()

    # do not include 'EFA' lv's
    if (!is.null(tmp_list$efa)) {
      lv_names <- unique(tmp_list$lhs[tmp_list$op == "=~" &
        !nchar(tmp_list$efa) > 0L])
    } else {
      lv_names <- unique(tmp_list$lhs[tmp_list$op == "=~"])
    }
    group_values <- lav_pt_group_values(tmp_list)

    for (lv in lv_names) {
      # factor variances
      lv_var_idx <- which(tmp_list$op == "~~" &
        tmp_list$lhs == lv &
        tmp_list$rhs == tmp_list$lhs &
        tmp_list$lhs == lv)

      # all free (but the first?)
      if (length(lv_var_idx) > 0L &&
        all(tmp_list$free[lv_var_idx][-1] > 0L)) {
        # 1) add eq constraint
        plabel <- tmp_list$plabel[lv_var_idx]

        tmp_lhs <- plabel[1]
        if (length(lv_var_idx) > 1L) {
          tmp_rhs <- paste(length(lv_var_idx), "-",
            paste(plabel[-1], collapse = "-"),
            sep = ""
          )
        } else {
          tmp_rhs <- length(lv_var_idx)
        }

        tmp$lhs <- c(tmp$lhs, tmp_lhs)
        tmp$op <- c(tmp$op, "==")
        tmp$rhs <- c(tmp$rhs, tmp_rhs)
        tmp$block <- c(tmp$block, 0L)
        tmp$user <- c(tmp$user, 2L)
        tmp$ustart <- c(tmp$ustart, as.numeric(NA))

        # 2) free lv variances first group
        lv_var_g1_idx <- which(tmp_list$op == "~~" &
          tmp_list$group == group_values[1] &
          tmp_list$lhs == lv &
          tmp_list$rhs == tmp_list$lhs &
          tmp_list$lhs == lv)
        # free only if automatically added
        if (length(lv_var_g1_idx) > 0L &&
          tmp_list$user[lv_var_g1_idx] == 0L) {
          tmp_list$free[lv_var_g1_idx] <- 1L
        }
      }
    } # lv

    tmp_list <- lav_pt_merge(tmp_list, tmp)
  }

  # mg.lv.efa.variances
  if (ngroups > 1L && "mg.lv.efa.variances" %in% effect_coding) {
    tmp <- list()

    # only 'EFA' lv's
    if (!is.null(tmp_list$efa)) {
      lv_names <- unique(tmp_list$lhs[tmp_list$op == "=~" &
        nchar(tmp_list$efa) > 0L])
    } else {
      lv_names <- character(0L)
    }
    group_values <- lav_pt_group_values(tmp_list)

    for (lv in lv_names) {
      # factor variances
      lv_var_idx <- which(tmp_list$op == "~~" &
        tmp_list$lhs == lv &
        tmp_list$rhs == tmp_list$lhs &
        tmp_list$lhs == lv)

      # all free (but the first?)
      if (length(lv_var_idx) > 0L &&
        all(tmp_list$free[lv_var_idx][-1] > 0L)) {
        # 1) add eq constraint
        plabel <- tmp_list$plabel[lv_var_idx]

        tmp_lhs <- plabel[1]
        if (length(lv_var_idx) > 1L) {
          tmp_rhs <- paste(length(lv_var_idx), "-",
            paste(plabel[-1], collapse = "-"),
            sep = ""
          )
        } else {
          tmp_rhs <- length(lv_var_idx)
        }

        tmp$lhs <- c(tmp$lhs, tmp_lhs)
        tmp$op <- c(tmp$op, "==")
        tmp$rhs <- c(tmp$rhs, tmp_rhs)
        tmp$block <- c(tmp$block, 0L)
        tmp$user <- c(tmp$user, 2L)
        tmp$ustart <- c(tmp$ustart, as.numeric(NA))

        # 2) free lv variances first group
        lv_var_g1_idx <- which(tmp_list$op == "~~" &
          tmp_list$group == group_values[1] &
          tmp_list$lhs == lv &
          tmp_list$rhs == tmp_list$lhs &
          tmp_list$lhs == lv)
        # free only if automatically added
        if (length(lv_var_g1_idx) > 0L &&
          tmp_list$user[lv_var_g1_idx] == 0L) {
          tmp_list$free[lv_var_g1_idx] <- 1L
        }
      }
    } # lv

    tmp_list <- lav_pt_merge(tmp_list, tmp)
  }


  # count free parameters
  idx_free <- which(tmp_list$free > 0L)
  tmp_list$free[idx_free] <- seq_along(idx_free)

  # new in 0.6-11: add free counter to this element (as in < 0.5-18)
  # unless we have other constraints (in which case ceq_simple = FALSE)
  if (ceq_simple) {
    idx_equal <- which(eq_id > 0)
    tmp_list$free[idx_equal] <- tmp_list$free[eq_id[idx_equal]]
  }

  # new in 0.6-14: add 'da' entries to reflect data-based order of ov's
  # now via attribute "ovda"
  attr(tmp_list, "ovda") <- ov_names_data

  # backwards compatibility...
  if (!is.null(tmp_list$unco)) {
    tmp_list$unco[idx_free] <- seq_along(sum(tmp_list$free > 0L))
  }

  if (lav_debug()) {
    cat("[lavaan DEBUG] lavParTable\n")
    print(as.data.frame(tmp_list))
  }

  # data.frame?
  if (as_data_frame) {
    tmp_list <- as.data.frame(tmp_list, stringsAsFactors = FALSE)
    attr(tmp_list, "ovda") <- ov_names_data
  } else {
    tmp_list <- lav_pt_set_cache(tmp_list) # add cached "pta" data
  }

  tmp_list
}
lavaanify <- function(                             # nolint
                      model = NULL,
                      meanstructure = FALSE,
                      int_ov_free = FALSE,
                      int_lv_free = FALSE,
                      marker_int_zero = FALSE,
                      orthogonal = FALSE,
                      orthogonal_y = FALSE,
                      orthogonal_x = FALSE,
                      orthogonal_efa = FALSE,
                      std_lv = FALSE,
                      correlation = FALSE,
                      composites = TRUE,
                      composites_cov_free = FALSE,
                      effect_coding = "",
                      conditional_x = FALSE,
                      fixed_x = FALSE,
                      parameterization = "delta",
                      constraints = NULL,
                      ceq_simple = FALSE,
                      auto = FALSE,
                      model_type = "sem",
                      auto_fix_first = FALSE,
                      marker = NULL,
                      auto_fix_single = FALSE,
                      auto_var = FALSE,
                      auto_cov_lv_x = FALSE,
                      auto_cov_x = FALSE,
                      auto_cov_y = FALSE,
                      auto_th = FALSE,
                      auto_delta = FALSE,
                      auto_efa = FALSE,
                      var_table = NULL,
                      ngroups = 1L,
                      nthresholds = NULL,
                      group_equal = NULL,
                      group_partial = NULL,
                      group_w_free = FALSE,
                      debug = FALSE,
                      warn = TRUE,
                      as_data_frame = TRUE,
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  lav_model_pt(
              model = model,
              meanstructure = meanstructure,
              int_ov_free = int_ov_free,
              int_lv_free = int_lv_free,
              marker_int_zero = marker_int_zero,
              orthogonal = orthogonal,
              orthogonal_y = orthogonal_y,
              orthogonal_x = orthogonal_x,
              orthogonal_efa = orthogonal_efa,
              std_lv = std_lv,
              correlation = correlation,
              composites = composites,
              composites_cov_free = composites_cov_free,
              effect_coding = effect_coding,
              conditional_x = conditional_x,
              fixed_x = fixed_x,
              parameterization = parameterization,
              constraints = constraints,
              ceq_simple = ceq_simple,
              auto = auto,
              model_type = model_type,
              auto_fix_first = auto_fix_first,
              marker = marker,
              auto_fix_single = auto_fix_single,
              auto_var = auto_var,
              auto_cov_lv_x = auto_cov_lv_x,
              auto_cov_x = auto_cov_x,
              auto_cov_y = auto_cov_y,
              auto_th = auto_th,
              auto_delta = auto_delta,
              auto_efa = auto_efa,
              var_table = var_table,
              ngroups = ngroups,
              nthresholds = nthresholds,
              group_equal = group_equal,
              group_partial = group_partial,
              group_w_free = group_w_free,
              debug = debug,
              warn = warn,
              as_data_frame = as_data_frame)
}

# synonym (thin forwarder; see lav_alias in 00alias.R)
lavParTable <- lav_alias("lavaanify")

# loglikelihood clustered/twolevel data -- conditional.x = TRUE

# YR: first version around Sept 2021
# YR 03 Jul 2026: shared per-unit sigma-gradient kernel + per-cluster-size
#                 Sigma.j cache (refactoring; no change in behavior)
# YR 07 Jul 2026: conversion sigma blocks now go through the shared
#                 lav_cl_* scatter/gather core (refactoring; no change in
#                 behavior)

# take model-implied mean+variance matrices, and reorder/augment them
# to facilitate computing of (log)likelihood in the two-level case
#
# NOTE: see the top of lav_mvnorm_cluster.R for the definition of the
# 'implied' vs '2l' representations and the invariants of the two
# conversion functions (tilde padding, ordering, the dual parameter-map/
# gradient-adjoint role of 2l2implied); everything there applies here
# too, with these conditional.x differences:
# - the mean parts are (residual) intercepts + slopes: beta.w/beta.b/
#   beta.z instead of mu.w/mu.b/mu.z, with the same folding convention
#   (only beta.w + beta.b is identified for the 'both' variables; their
#   sum is beta.wb)
# - the y block contains only the 'y' variables (ov.y.idx): the
#   exogenous x variables (within.x.idx, between.x.idx) are removed
#   from the tilde universe together with the between-only y variables
#   (between.y.idx, the 'z' block)

# when conditional.x = TRUE:
# - sigma.w and sigma.b: same dimensions, level-1 'Y' variables only
# - sigma.zz: level-2 variables only
# - sigma.yz: cov(level-1, level-2)
# - beta.w: beta y within part
# - beta.b: beta y between part
# - beta.z: beta z (between-only)
lav_mvreg_cl_implied22l <- function(lp = NULL,
                                    implied = NULL,
                                    res_int_w = NULL,
                                    res_int_b = NULL,
                                    res_pi_w = NULL,
                                    res_pi_b = NULL,
                                    res_sigma_w = NULL,
                                    res_sigma_b = NULL) {
  if (!is.null(implied)) {
    # FIXME: only for single-group analysis!
    res_sigma_w <- implied$res.cov[[1]]
    res_int_w <- implied$res.int[[1]]
    res_pi_w <- implied$res.slopes[[1]]

    res_sigma_b <- implied$res.cov[[2]]
    res_int_b <- implied$res.int[[2]]
    res_pi_b <- implied$res.slopes[[2]]
  }

  # within/between idx
  within_x_idx <- lp$within.x.idx[[1]]
  between_y_idx <- lp$between.y.idx[[2]]
  between_x_idx <- lp$between.x.idx[[2]]

  # ov.idx per level
  ov_idx <- lp$ov.idx

  # 'tilde' matrices: ALL variables within and between
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))

  # only 'y'
  ov_y_idx <- lp$ov.y.idx

  # two levels only (for now)
  ov_y_idx1 <- ov_y_idx[[1]]
  ov_y_idx2 <- ov_y_idx[[2]]

  # INT.W.tilde
  int_w_tilde <- matrix(0, p_tilde, 1L)
  int_w_tilde[ov_y_idx1, 1L] <- res_int_w

  # PI.W.tilde
  pi_w_tilde <- matrix(0, p_tilde, ncol(res_pi_w))
  pi_w_tilde[ov_y_idx1, ] <- res_pi_w

  beta_w_tilde <- rbind(t(int_w_tilde), t(pi_w_tilde))

  # INT.B.tilde
  int_b_tilde <- matrix(0, p_tilde, 1L)
  int_b_tilde[ov_y_idx2, 1L] <- res_int_b

  # PI.B.tilde
  pi_b_tilde <- matrix(0, p_tilde, ncol(res_pi_b))
  pi_b_tilde[ov_y_idx2, ] <- res_pi_b

  beta_b_tilde <- rbind(t(int_b_tilde), t(pi_b_tilde))

  # remove all 'x' (and the between-only y) columns from the tilde universe
  rm_idx <- c(within_x_idx, between_x_idx, between_y_idx)

  # sigma blocks: shared scatter/gather core (lav_mvnorm_cluster_kernels.R)
  out_sigma <- lav_cl_sigma_22l(
    sigma_w = res_sigma_w, sigma_b = res_sigma_b, p_tilde = p_tilde,
    l1_idx = ov_y_idx1, l2_idx = ov_y_idx2,
    z_idx = between_y_idx, rm_idx = rm_idx
  )
  sigma_w <- out_sigma$sigma.w
  sigma_b <- out_sigma$sigma.b
  sigma_zz <- out_sigma$sigma.zz
  sigma_yz <- out_sigma$sigma.yz

  # beta blocks
  y_idx <- seq_len(p_tilde)
  if (length(rm_idx) > 0L) {
    y_idx <- y_idx[-rm_idx]
  }
  beta_w <- beta_w_tilde[, y_idx, drop = FALSE]
  beta_b <- beta_b_tilde[, y_idx, drop = FALSE]
  if (length(between_y_idx) > 0L) {
    beta_z <- beta_b_tilde[, between_y_idx, drop = FALSE]
  } else {
    beta_z <- matrix(0, 0L, 0L)
  }

  # beta.wb # FIXME: not correct if some 'x' are split (overlap)
  # but because we ALWAYS treat split-x as 'y', this is not a problem
  beta_wb <- rbind(beta_w, beta_b[-1, , drop = FALSE])
  beta_wb[1, ] <- beta_wb[1, , drop = FALSE] + beta_b[1, , drop = FALSE]

  list(
    sigma.w = sigma_w, sigma.b = sigma_b, sigma.zz = sigma_zz,
    sigma.yz = sigma_yz, beta.w = beta_w, beta.b = beta_b, beta.z = beta_z,
    beta.wb = beta_wb
  )
}


# recreate implied matrices from 2L matrices
lav_mvreg_cl_2l2implied <- function(lp,
                                    sigma_w = NULL,
                                    sigma_b = NULL,
                                    sigma_zz = NULL,
                                    sigma_yz = NULL,
                                    beta_w = NULL,
                                    beta_b = NULL,
                                    beta_z = NULL) {
  # within/between idx
  # within_x_idx <- lp$within.x.idx[[1]]
  between_y_idx <- lp$between.y.idx[[2]]
  # between_x_idx <- lp$between.x.idx[[2]]

  # ov.idx per level
  ov_idx <- lp$ov.idx

  # 'tilde' matrices: ALL variables within and between
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))

  # only 'y'
  ov_y_idx <- lp$ov.y.idx

  # two levels only (for now)
  ov_y_idx1 <- ov_y_idx[[1]]
  ov_y_idx2 <- ov_y_idx[[2]]

  # Sigma.W.tilde
  sigma_w_tilde <- matrix(0, p_tilde, p_tilde)
  sigma_w_tilde[ov_y_idx1, ov_y_idx1] <- sigma_w

  # INT.W.tilde
  int_w_tilde <- matrix(0, p_tilde, 1L)
  int_w_tilde[ov_y_idx1, 1L] <- beta_w[1L, ]

  # PI.W.tilde
  pi_w_tilde <- matrix(0, p_tilde, nrow(beta_w) - 1L)
  pi_w_tilde[ov_y_idx1, ] <- t(beta_w[-1L, ])

  # INT.B.tilde
  int_b_tilde <- matrix(0, p_tilde, 1L)
  int_b_tilde[ov_y_idx1, 1L] <- beta_b[1L, ]

  # PI.B.tilde
  pi_b_tilde <- matrix(0, p_tilde, nrow(beta_b) - 1L)
  pi_b_tilde[ov_y_idx1, ] <- t(beta_b[-1L, ])

  if (length(between_y_idx) > 0L) {
    int_b_tilde[between_y_idx, 1L] <- beta_z[1L, ]
    pi_b_tilde[between_y_idx, ] <- t(beta_z[-1L, ])
  }

  # Sigma.B: shared scatter/gather core (lav_mvnorm_cluster_kernels.R)
  res_sigma_b <- lav_cl_sigma_2l2(
    sigma_b = sigma_b, sigma_zz = sigma_zz, sigma_yz = sigma_yz,
    p_tilde = p_tilde, y_idx = ov_y_idx1, z_idx = between_y_idx,
    l2_idx = ov_y_idx2
  )

  res_sigma_w <- sigma_w_tilde[ov_y_idx1, ov_y_idx1, drop = FALSE]
  res_int_w <- int_w_tilde[ov_y_idx1, , drop = FALSE]
  res_pi_w <- pi_w_tilde[ov_y_idx1, , drop = FALSE]

  res_int_b <- int_b_tilde[ov_y_idx2, , drop = FALSE]
  res_pi_b <- pi_b_tilde[ov_y_idx2, , drop = FALSE]

  implied <- list(
    res.cov = list(res_sigma_w, res_sigma_b),
    res.int = list(res_int_w, res_int_b),
    res.slopes = list(res_pi_w, res_pi_b)
  )

  # Note: cov.x and mean.x must be added by the caller
  implied
}


# reconstruct the pooled-within residual covariance matrix S.PW from the
# stored sample statistics (shared by loglik and dlogl)
lav_mvreg_cl_spw <- function(ylp = NULL,
                             lp = NULL,
                             beta_wb = NULL) {
  cluster_size <- lp$cluster.size[[2]]

  wb1_diff <- ylp[[2]]$sample.wb - beta_wb
  y1y1_wb_res <- (ylp[[2]]$sample.YYres.wb1 +
    t(wb1_diff) %*% ylp[[2]]$sample.XX.wb1 %*% (wb1_diff))

  # this one is weighted -- not the same as crossprod(Y2w.res)
  wb2_diff <- ylp[[2]]$sample.wb2 - beta_wb
  y2y2w_res <- (ylp[[2]]$sample.YYres.wb2 +
    ylp[[2]]$sample.YresX.wb2 %*% (wb2_diff) +
    t(wb2_diff) %*% t(ylp[[2]]$sample.YresX.wb2) +
    t(wb2_diff) %*% ylp[[2]]$sample.XX.wb2 %*% (wb2_diff))

  (y1y1_wb_res - y2y2w_res) / sum(cluster_size - 1)
}

# level-2 residual crossproducts for one cluster-size group, from the
# stored per-cluster-size sample statistics (shared by loglik and dlogl)
lav_mvreg_cl_unit_stats <- function(ylp = NULL,
                                    clz = NULL,
                                    beta_wb = NULL,
                                    beta_z = NULL,
                                    between_y = FALSE) {
  y2_diff <- ylp[[2]]$sample.clz.Y2.B[[clz]] - beta_wb
  xx_y2_diff <- ylp[[2]]$sample.clz.Y2.XX[[clz]] %*% y2_diff
  y2yc_yy <- ylp[[2]]$sample.clz.Y2.res[[clz]] +
    crossprod(y2_diff, xx_y2_diff)

  if (!between_y) {
    return(list(
      y2_diff = y2_diff, xx_y2_diff = xx_y2_diff,
      y2yc_yy = y2yc_yy
    ))
  }

  zz_diff <- ylp[[2]]$sample.clz.ZZ.B[[clz]] - beta_z
  y2yc_zz <- (ylp[[2]]$sample.clz.ZZ.res[[clz]] +
    t(zz_diff) %*% ylp[[2]]$sample.clz.ZZ.XX[[clz]] %*% (zz_diff))
  y2yc_yz <- (ylp[[2]]$sample.clz.YZ.res[[clz]] +
    ylp[[2]]$sample.clz.YresXZ[[clz]] %*% zz_diff + # zero?
    t(y2_diff) %*% ylp[[2]]$sample.clz.XWZres[[clz]] +
    t(y2_diff) %*% ylp[[2]]$sample.clz.YZ.XX[[clz]] %*% zz_diff)

  list(
    y2_diff = y2_diff, xx_y2_diff = xx_y2_diff, zz_diff = zz_diff,
    y2yc_yy = y2yc_yy, y2yc_yz = y2yc_yz, y2yc_zz = y2yc_zz
  )
}

# per-unit gradient of -2*logl with respect to the SIGMA blocks
# (shared by the gradient -- one unit per cluster size, ns = n_s[clz] --
#  and the cluster-wise scores -- one unit per cluster, ns = 1)
lav_mvreg_cl_grad_sigma_kernel <- function(nj = NULL,
                                           ns = 1,
                                           y2yc_yy = NULL,
                                           y2yc_yz = NULL,
                                           y2yc_zz = NULL,
                                           sigma_j_inv = NULL,
                                           sigma_ji_yz_zi = NULL,
                                           sigma_zi_zy_ji = NULL,
                                           sigma_ji_yz = NULL,
                                           sigma_zz_inv = NULL,
                                           sigma_yz = NULL,
                                           sigma_yz_zi = NULL) {
  if (is.null(y2yc_zz)) { # no between-y variables
    # common part
    j_yyj <- nj * sigma_j_inv %*% y2yc_yy %*% sigma_j_inv

    # SIGMA.W (between part)
    g_sigma_w1 <- (ns * sigma_j_inv) - j_yyj

    return(list(
      j_yzj = j_yyj,
      sigma_w1 = lav_mat_vech_dd(g_sigma_w1),
      sigma_b = lav_mat_vech_dd(nj * g_sigma_w1)
    ))
  }

  # common parts
  zz_zi_yz_ji <- y2yc_zz %*% sigma_zi_zy_ji
  ji_yz_zi <- sigma_j_inv %*% y2yc_yz %*% sigma_zz_inv

  j_yzj_yy <- sigma_j_inv %*% y2yc_yy %*% sigma_j_inv
  j_yzj_yz <- tcrossprod(ji_yz_zi, sigma_ji_yz)
  j_yzj_zz <- sigma_ji_yz_zi %*% zz_zi_yz_ji

  j_yzj <- nj * (j_yzj_yy + j_yzj_zz - j_yzj_yz - t(j_yzj_yz))

  # SIGMA.W (between part)
  g_sigma_w1 <- (ns * sigma_j_inv) - j_yzj

  # SIGMA.ZZ
  yz1 <- zz_zi_yz_ji %*% sigma_yz
  yz2 <- crossprod(y2yc_yz, sigma_ji_yz)
  tmp <- (t(sigma_yz) %*% g_sigma_w1 %*% sigma_yz
    - (1 / nj * y2yc_zz + t(yz1) + yz1 - t(yz2) - yz2))
  g_sigma_zz <- ((ns * sigma_zz_inv) +
    nj * sigma_zz_inv %*% tmp %*% sigma_zz_inv)

  # SIGMA.ZY
  tmp1 <- crossprod(zz_zi_yz_ji, sigma_zz_inv)
  tmp2 <- ns * sigma_ji_yz_zi
  tmp3 <- ji_yz_zi
  tmp4 <- j_yzj %*% sigma_yz_zi
  g_sigma_yz <- 2 * nj * (tmp1 - tmp2 - tmp3 + tmp4)

  list(
    j_yzj = j_yzj,
    sigma_w1 = lav_mat_vech_dd(g_sigma_w1),
    sigma_b = lav_mat_vech_dd(nj * g_sigma_w1),
    sigma_zz = lav_mat_vech_dd(g_sigma_zz),
    sigma_yz = lav_mat_vec(g_sigma_yz)
  )
}

lav_mvreg_cl_loglik_samp_2l <- function(ylp = NULL,
                                        lp = NULL,
                                        res_sigma_w = NULL,
                                        res_int_w = NULL,
                                        res_pi_w = NULL,
                                        res_sigma_b = NULL,
                                        res_int_b = NULL,
                                        res_pi_b = NULL,
                                        out = NULL, # 2l
                                        sinv_method = "eigen",
                                        log2pi = FALSE,
                                        minus_two = TRUE) {
  # map implied to 2l matrices
  if (is.null(out)) {
    out <- lav_mvreg_cl_implied22l(
      lp = lp, implied = NULL,
      res_sigma_w = res_sigma_w,
      res_int_w = res_int_w, res_pi_w = res_pi_w,
      res_sigma_b = res_sigma_b,
      res_int_b = res_int_b, res_pi_b = res_pi_b
    )
  }
  sigma_w <- out$sigma.w
  sigma_b <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz
  beta_z <- out$beta.z
  beta_wb <- out$beta.wb # out= must be a complete implied22l output

  # log 2*pi
  log_2pi <- log(2 * pi)

  # Lp
  cluster_size <- lp$cluster.size[[2]]
  cluster_sizes <- lp$cluster.sizes[[2]]
  ncluster_sizes <- lp$ncluster.sizes[[2]]
  n_s <- lp$cluster.size.ns[[2]]

  # dependent 'y' level-2 ('Z') only variables?
  between_y_idx <- lp$between.y.idx[[2]]
  between_y <- length(between_y_idx) > 0L

  # reconstruct S.PW
  s_pw <- lav_mvreg_cl_spw(ylp = ylp, lp = lp, beta_wb = beta_wb)

  # common parts:
  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w,
    logdet = TRUE, sinv_method = sinv_method
  )
  sigma_w_logdet <- attr(sigma_w_inv, "logdet")
  if (between_y) {
    sigma_zz_inv <- lav_mat_sym_inverse(
      s = sigma_zz,
      logdet = TRUE, sinv_method = sinv_method
    )
    sigma_zz_logdet <- attr(sigma_zz_inv, "logdet")
    sigma_yz_zi <- sigma_yz %*% sigma_zz_inv
    sigma_zi_zy <- t(sigma_yz_zi)
    sigma_b_z <- sigma_b - sigma_yz %*% sigma_zi_zy
  } else {
    sigma_b_z <- sigma_b
  }

  # min 2* logliklihood
  dist_1 <- numeric(ncluster_sizes)
  logdet <- numeric(ncluster_sizes)
  const <- numeric(ncluster_sizes)
  for (clz in seq_len(ncluster_sizes)) {
    # cluster size
    nj <- cluster_sizes[clz]

    # data between
    stats <- lav_mvreg_cl_unit_stats(
      ylp = ylp, clz = clz,
      beta_wb = beta_wb, beta_z = beta_z, between_y = between_y
    )
    y2yc_yy <- stats$y2yc_yy

    # construct sigma.j
    sigma_j <- (nj * sigma_b_z) + sigma_w
    sigma_j_inv <- lav_mat_sym_inverse(
      s = sigma_j,
      logdet = TRUE, sinv_method = sinv_method
    )
    sigma_j_logdet <- attr(sigma_j_inv, "logdet")

    if (between_y) {
      sigma_ji_yz_zi <- sigma_j_inv %*% sigma_yz_zi

      # part 1 -- zz
      vinv_11 <- sigma_zz_inv + nj * (sigma_zi_zy %*% sigma_ji_yz_zi)
      q_zz <- sum(vinv_11 * stats$y2yc_zz)

      # part 2 -- yz
      q_yz <- -nj * sum(sigma_ji_yz_zi * stats$y2yc_yz)
    } else {
      q_zz <- q_yz <- sigma_zz_logdet <- 0
    }

    # part 5 -- yyc
    q_yyc <- -nj * sum(sigma_j_inv * y2yc_yy)

    if (log2pi) {
      p <- nj * nrow(sigma_w) + nrow(sigma_zz)
      const[clz] <- p * log_2pi
    }
    logdet[clz] <- sigma_zz_logdet + sigma_j_logdet
    dist_1[clz] <- q_zz + 2 * q_yz - q_yyc
  }
  # q.yya + q.yyb
  q_w <- sum(cluster_size - 1) * sum(sigma_w_inv * s_pw)
  # logdet within part
  l_w <- sum(cluster_size - 1) * sigma_w_logdet

  # -2*times logl (without the constant) (for optimization)
  loglik <- sum(logdet * n_s) + sum(dist_1) + q_w + l_w

  if (log2pi) {
    loglik <- loglik + sum(const * n_s)
  }

  # functions below compute -2 * logl
  if (!minus_two) {
    loglik <- loglik / (-2)
  }

  loglik
}

# first derivative -2*logl wrt Beta.W, Beta.B, Sigma.W, Sigma.B
# two orderings of the two-level conditional.x statistic vector coexist:
# - the lav_mvreg_cl_* kernels use, per level,
#       [ res.int, vec(res.slopes), vech(res.cov) ]
# - lav_model_delta() and the *LS machinery use, per level,
#       [ vecr(cbind(res.int, res.slopes)), vech(res.cov) ]
#   (one [intercept, slopes] set per variable, cfr. lav_model_wls_est)
# this returns the permutation p such that x.kernel[p] == x.delta;
# for a matrix: B.delta == B.kernel[p, p]
lav_mvreg_cl_stat_perm <- function(res_int_w = NULL, res_pi_w = NULL,
                                   res_int_b = NULL, res_pi_b = NULL) {
  perm_l <- function(res_int, res_pi) {
    nvar <- NROW(res_int)
    nexo <- NCOL(res_pi)
    pstar <- nvar * (nvar + 1L) / 2L
    c(
      as.vector(t(matrix(seq_len(nvar * (1L + nexo)), nrow = nvar))),
      nvar * (1L + nexo) + seq_len(pstar)
    )
  }
  p_w <- perm_l(res_int_w, res_pi_w)
  p_b <- perm_l(res_int_b, res_pi_b)
  c(p_w, length(p_w) + p_b)
}

lav_mvreg_cl_dlogl_2l_samp <- function(ylp = NULL,
                                       lp = NULL,
                                       res_sigma_w = NULL,
                                       res_int_w = NULL,
                                       res_pi_w = NULL,
                                       res_sigma_b = NULL,
                                       res_int_b = NULL,
                                       res_pi_b = NULL,
                                       out = NULL, # 2l
                                       return_list = FALSE,
                                       sinv_method = "eigen") {
  # map implied to 2l matrices
  if (is.null(out)) {
    out <- lav_mvreg_cl_implied22l(
      lp = lp, implied = NULL,
      res_sigma_w = res_sigma_w,
      res_int_w = res_int_w, res_pi_w = res_pi_w,
      res_sigma_b = res_sigma_b,
      res_int_b = res_int_b, res_pi_b = res_pi_b
    )
  }
  sigma_w <- out$sigma.w
  sigma_b <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz
  beta_w <- out$beta.w
  beta_b <- out$beta.b
  beta_z <- out$beta.z
  beta_wb <- out$beta.wb # out= must be a complete implied22l output

  # Lp
  cluster_size <- lp$cluster.size[[2]]
  cluster_sizes <- lp$cluster.sizes[[2]]
  ncluster_sizes <- lp$ncluster.sizes[[2]]
  n_s <- lp$cluster.size.ns[[2]]

  within_x_idx <- lp$within.x.idx[[1]]
  between_y_idx <- lp$between.y.idx[[2]]
  between_y <- length(between_y_idx) > 0L

  w1_idx <- seq_len(length(within_x_idx) + 1L)
  b1_idx <- c(1L, seq_len(nrow(beta_wb))[-w1_idx])

  # reconstruct S.PW
  s_pw <- lav_mvreg_cl_spw(ylp = ylp, lp = lp, beta_wb = beta_wb)

  # common parts:
  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w,
    sinv_method = sinv_method
  )

  g_beta_w <- matrix(0, ncluster_sizes, length(beta_w))
  g_beta_b <- matrix(0, ncluster_sizes, length(beta_b))
  g_sigma_w1_1 <- matrix(0, ncluster_sizes, length(lav_mat_vech(sigma_w)))
  g_sigma_b_1 <- matrix(0, ncluster_sizes, length(lav_mat_vech(sigma_b)))

  if (between_y) {
    g_beta_z <- matrix(0, ncluster_sizes, length(beta_z))
    g_sigma_zz_1 <- matrix(0, ncluster_sizes, length(lav_mat_vech(sigma_zz)))
    g_sigma_yz_1 <- matrix(0, ncluster_sizes, length(sigma_yz))

    sigma_zz_inv <- lav_mat_sym_inverse(
      s = sigma_zz,
      sinv_method = sinv_method
    )
    sigma_yz_zi <- sigma_yz %*% sigma_zz_inv
    sigma_b_z <- sigma_b - sigma_yz %*% t(sigma_yz_zi)

    # Sigma.j quantities per unique cluster size
    cache <- lav_mvn_cl_sigma_j_cache(
      cluster_sizes = cluster_sizes,
      sigma_b_z = sigma_b_z, sigma_w_1 = sigma_w,
      sigma_yz_zi = sigma_yz_zi, sinv_method = sinv_method
    )

    for (clz in seq_len(ncluster_sizes)) {
      # cluster size
      nj <- cluster_sizes[clz]

      # data between (for this cluster-size group)
      stats <- lav_mvreg_cl_unit_stats(
        ylp = ylp, clz = clz,
        beta_wb = beta_wb, beta_z = beta_z, between_y = TRUE
      )
      y2_diff <- stats$y2_diff
      xx_y2_diff <- stats$xx_y2_diff
      zz_diff <- stats$zz_diff

      # Sigma.j quantities (cached per cluster size)
      cj <- cache[[clz]]
      sigma_j_inv <- cj$sigma_j_inv
      sigma_ji_yz_zi <- cj$sigma_ji_yz_zi
      sigma_zi_zy_ji <- cj$sigma_zi_zy_ji
      sigma_ji_yz <- sigma_j_inv %*% sigma_yz

      # SIGMA blocks (shared kernel; ns = n_s[clz])
      g <- lav_mvreg_cl_grad_sigma_kernel(
        nj = nj, ns = n_s[clz],
        y2yc_yy = stats$y2yc_yy, y2yc_yz = stats$y2yc_yz,
        y2yc_zz = stats$y2yc_zz,
        sigma_j_inv = sigma_j_inv, sigma_ji_yz_zi = sigma_ji_yz_zi,
        sigma_zi_zy_ji = sigma_zi_zy_ji, sigma_ji_yz = sigma_ji_yz,
        sigma_zz_inv = sigma_zz_inv, sigma_yz = sigma_yz,
        sigma_yz_zi = sigma_yz_zi
      )
      g_sigma_w1_1[clz, ] <- g$sigma_w1
      g_sigma_b_1[clz, ] <- g$sigma_b
      g_sigma_zz_1[clz, ] <- g$sigma_zz
      g_sigma_yz_1[clz, ] <- g$sigma_yz

      # BETA.Z
      a <- (sigma_zz_inv + nj * (sigma_zi_zy_ji %*% sigma_yz_zi)) # symm!
      m_b <- nj * (sigma_zi_zy_ji)
      tmp_z <- (ylp[[2]]$sample.clz.ZZ.XX[[clz]] %*% zz_diff %*% a -
        (t(ylp[[2]]$sample.clz.YresXZ[[clz]]) +
          t(ylp[[2]]$sample.clz.YZ.XX[[clz]]) %*% y2_diff) %*% t(m_b))
      g_beta_z[clz, ] <- as.vector(-2 * tmp_z)

      # BETA.W (between part only) + BETA.B
      tmp <- (ylp[[2]]$sample.clz.XWZres[[clz]] +
        ylp[[2]]$sample.clz.YZ.XX[[clz]] %*% zz_diff)
      out_b <- tmp %*% sigma_zi_zy_ji - xx_y2_diff %*% sigma_j_inv
      out_w <- out_b + xx_y2_diff %*% sigma_w_inv
      tmp_b <- out_b[b1_idx, , drop = FALSE]
      tmp_w <- out_w[w1_idx, , drop = FALSE]
      g_beta_b[clz, ] <- as.vector(2 * nj * tmp_b)
      g_beta_w[clz, ] <- as.vector(2 * nj * tmp_w)
    } # clz

    d_beta_w1 <- matrix(colSums(g_beta_w), nrow(beta_w), ncol(beta_w))
    d_beta_b <- matrix(colSums(g_beta_b), nrow(beta_b), ncol(beta_b))
    d_sigma_w1 <- lav_mat_vech_rev(colSums(g_sigma_w1_1))
    d_sigma_b <- lav_mat_vech_rev(colSums(g_sigma_b_1))

    # z
    d_beta_z <- matrix(colSums(g_beta_z), nrow(beta_z), ncol(beta_z))
    d_sigma_zz <- lav_mat_vech_rev(colSums(g_sigma_zz_1))
    d_sigma_yz <- matrix(colSums(g_sigma_yz_1), nrow(sigma_yz), ncol(sigma_yz))
  # between.y.idx

  } else { # no between.y.idx

    # Sigma.j inverses per unique cluster size
    cache <- lav_mvn_cl_sigma_j_cache(
      cluster_sizes = cluster_sizes,
      sigma_b_z = sigma_b, sigma_w_1 = sigma_w,
      sinv_method = sinv_method
    )

    for (clz in seq_len(ncluster_sizes)) {
      # cluster size
      nj <- cluster_sizes[clz]

      # data between (for this cluster-size group)
      stats <- lav_mvreg_cl_unit_stats(
        ylp = ylp, clz = clz,
        beta_wb = beta_wb, between_y = FALSE
      )
      xx_y2_diff <- stats$xx_y2_diff

      # Sigma.j (cached per cluster size)
      sigma_j_inv <- cache[[clz]]$sigma_j_inv

      # SIGMA blocks (shared kernel; ns = n_s[clz])
      g <- lav_mvreg_cl_grad_sigma_kernel(
        nj = nj, ns = n_s[clz],
        y2yc_yy = stats$y2yc_yy,
        sigma_j_inv = sigma_j_inv
      )
      g_sigma_w1_1[clz, ] <- g$sigma_w1
      g_sigma_b_1[clz, ] <- g$sigma_b

      # BETA.W (between part only) + BETA.B
      out_b <- -1 * xx_y2_diff %*% sigma_j_inv
      out_w <- out_b + xx_y2_diff %*% sigma_w_inv
      tmp_b <- out_b[b1_idx, , drop = FALSE]
      tmp_w <- out_w[w1_idx, , drop = FALSE]
      g_beta_b[clz, ] <- as.vector(2 * nj * tmp_b)
      g_beta_w[clz, ] <- as.vector(2 * nj * tmp_w)
    } # cl

    d_beta_w1 <- matrix(colSums(g_beta_w), nrow(beta_w), ncol(beta_w))
    d_beta_b <- matrix(colSums(g_beta_b), nrow(beta_b), ncol(beta_b))
    d_sigma_w1 <- lav_mat_vech_rev(colSums(g_sigma_w1_1))
    d_sigma_b <- lav_mat_vech_rev(colSums(g_sigma_b_1))

    # z
    d_beta_z <- matrix(0, 0L, 0L)
    d_sigma_zz <- matrix(0, 0L, 0L)
    d_sigma_yz <- matrix(0, 0L, 0L)
  } # no-between-y

  # Sigma.W (bis)
  d_sigma_w2 <- sum(cluster_size - 1) * (sigma_w_inv
  - sigma_w_inv %*% s_pw %*% sigma_w_inv)
  tmp <- d_sigma_w2 * 2
  diag(tmp) <- diag(d_sigma_w2)
  d_sigma_w2 <- tmp

  d_sigma_w <- d_sigma_w1 + d_sigma_w2

  # beta.w (bis)
  d_beta_w2 <- -2 * (ylp[[2]]$sample.XX.wb1 %*%
      (ylp[[2]]$sample.wb - beta_wb))[w1_idx, , drop = FALSE] %*% sigma_w_inv

  d_beta_w <- d_beta_w1 + d_beta_w2

  # rearrange
  dimplied <- lav_mvreg_cl_2l2implied(lp,
    sigma_w = d_sigma_w, sigma_b = d_sigma_b,
    sigma_zz = d_sigma_zz, sigma_yz = d_sigma_yz,
    beta_w = d_beta_w, beta_b = d_beta_b, beta_z = d_beta_z
  )

  if (return_list) {
    return(dimplied)
  }

  # as a single vector
  out <- c(
    drop(dimplied$res.int[[1]]),
    lav_mat_vec(dimplied$res.slopes[[1]]),
    lav_mat_vech(dimplied$res.cov[[1]]),
    drop(dimplied$res.int[[2]]),
    lav_mat_vec(dimplied$res.slopes[[2]]),
    lav_mat_vech(dimplied$res.cov[[2]])
  )
  out
}

# cluster-wise scores -2*logl wrt Beta.W, Beta.B, Sigma.W, Sigma.B
lav_mvreg_cl_sc_2l <- function(y1 = NULL,
                               ylp = NULL,
                               lp = NULL,
                               res_sigma_w = NULL,
                               res_int_w = NULL,
                               res_pi_w = NULL,
                               res_sigma_b = NULL,
                               res_int_b = NULL,
                               res_pi_b = NULL,
                               out = NULL, # 2l
                               sinv_method = "eigen") {
  # map implied to 2l matrices
  if (is.null(out)) {
    out <- lav_mvreg_cl_implied22l(
      lp = lp, implied = NULL,
      res_sigma_w = res_sigma_w,
      res_int_w = res_int_w, res_pi_w = res_pi_w,
      res_sigma_b = res_sigma_b,
      res_int_b = res_int_b, res_pi_b = res_pi_b
    )
  }
  sigma_w <- out$sigma.w
  sigma_b <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz
  beta_w <- out$beta.w
  beta_b <- out$beta.b
  beta_z <- out$beta.z
  beta_wb <- out$beta.wb # out= must be a complete implied22l output

  # Lp
  nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]
  cluster_sizes <- lp$cluster.sizes[[2]]
  cluster_idx <- lp$cluster.idx[[2]]

  within_x_idx <- lp$within.x.idx[[1]]
  # between_idx <- lp$between.idx[[2]]
  between_y_idx <- lp$between.y.idx[[2]]
  between_x_idx <- lp$between.x.idx[[2]]

  y1_idx <- lp$ov.y.idx[[1]]
  x1_idx <- c(within_x_idx, between_x_idx) # in that order

  # residuals for 'Y'
  y1_wb <- y1[, y1_idx, drop = FALSE]

  if (length(x1_idx) > 0L) {
    exo_wb <- cbind(1, y1[, x1_idx, drop = FALSE])
    y1_wb_hat <- exo_wb %*% beta_wb
    y1_wb_res <- y1_wb - y1_wb_hat
  } else {
    y1_wb_res <- y1_wb
  }

  # residuals 'Y' (level 2)
  y2 <- ylp[[2]]$Y2
  if (length(x1_idx) > 0L) {
    exo_wb2 <- cbind(1, y2[, x1_idx, drop = FALSE])
    y2w_res <- y2[, y1_idx, drop = FALSE] - exo_wb2 %*% beta_wb
  } else {
    exo_wb2 <- matrix(1, nrow(y2), 1L)
    y2w_res <- y2[, y1_idx, drop = FALSE]
  }

  # residual 'Z' (level 2)
  if (length(between_y_idx) > 0L) {
    if (length(between_x_idx) > 0L) {
      exo_z_1 <- cbind(1, y2[, between_x_idx, drop = FALSE])
      y2_z <- y2[, between_y_idx, drop = FALSE]
      y2z_res <- y2_z - exo_z_1 %*% beta_z
    } else {
      y2z_res <- y2[, between_y_idx, drop = FALSE]
    }
  }

  # common parts:
  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w,
    sinv_method = sinv_method
  )

  g_beta_w1 <- matrix(0, nclusters, length(beta_w))
  g_beta_b <- matrix(0, nclusters, length(beta_b))
  g_sigma_w1_1 <- matrix(0, nclusters, length(lav_mat_vech(sigma_w)))
  g_sigma_b_1 <- matrix(0, nclusters, length(lav_mat_vech(sigma_b)))


  if (length(between_y_idx) > 0L) {
    g_beta_z <- matrix(0, nclusters, length(beta_z))
    g_sigma_zz_1 <- matrix(0, nclusters, length(lav_mat_vech(sigma_zz)))
    g_sigma_yz_1 <- matrix(0, nclusters, length(sigma_yz))

    sigma_zz_inv <- lav_mat_sym_inverse(
      s = sigma_zz,
      sinv_method = sinv_method
    )
    sigma_yz_zi <- sigma_yz %*% sigma_zz_inv
    sigma_b_z <- sigma_b - sigma_yz %*% t(sigma_yz_zi)

    # Sigma.j quantities per unique cluster size (was: per cluster)
    cache <- lav_mvn_cl_sigma_j_cache(
      cluster_sizes = cluster_sizes,
      sigma_b_z = sigma_b_z, sigma_w_1 = sigma_w,
      sigma_yz_zi = sigma_yz_zi, sinv_method = sinv_method
    )
    size_idx <- match(cluster_size, cluster_sizes)

    for (cl in seq_len(nclusters)) {
      # cluster size
      nj <- cluster_size[cl]

      # data within for the cluster (centered)
      yc <- y2w_res[cl, ]

      # data between
      zc <- y2z_res[cl, ]
      y2yc_yy <- tcrossprod(y2w_res[cl, ])
      y2yc_zz <- tcrossprod(y2z_res[cl, ])
      y2yc_yz <- tcrossprod(y2w_res[cl, ], y2z_res[cl, ])

      # Sigma.j quantities (cached per cluster size)
      cj <- cache[[size_idx[cl]]]
      sigma_j_inv <- cj$sigma_j_inv
      sigma_ji_yz_zi <- cj$sigma_ji_yz_zi
      sigma_zi_zy_ji <- cj$sigma_zi_zy_ji
      sigma_ji_yz <- sigma_j_inv %*% sigma_yz

      # SIGMA blocks (shared kernel; ns = 1)
      g <- lav_mvreg_cl_grad_sigma_kernel(
        nj = nj, ns = 1,
        y2yc_yy = y2yc_yy, y2yc_yz = y2yc_yz, y2yc_zz = y2yc_zz,
        sigma_j_inv = sigma_j_inv, sigma_ji_yz_zi = sigma_ji_yz_zi,
        sigma_zi_zy_ji = sigma_zi_zy_ji, sigma_ji_yz = sigma_ji_yz,
        sigma_zz_inv = sigma_zz_inv, sigma_yz = sigma_yz,
        sigma_yz_zi = sigma_yz_zi
      )
      g_sigma_w1_1[cl, ] <- g$sigma_w1
      g_sigma_b_1[cl, ] <- g$sigma_b
      g_sigma_zz_1[cl, ] <- g$sigma_zz
      g_sigma_yz_1[cl, ] <- g$sigma_yz

      # BETA.Z
      # here, we avoid the (sample.z - beta.z) approach
      exo_z <- cbind(1, y2[cl, between_x_idx, drop = FALSE])
      tmp1 <- (sigma_zz_inv + nj * (sigma_zi_zy_ji %*% sigma_yz_zi)) %*% zc
      tmp2 <- nj * (sigma_zi_zy_ji) %*% yc
      tmp_z <- crossprod(exo_z, drop(tmp1 - tmp2))
      g_beta_z[cl, ] <- as.vector(-2 * tmp_z)

      # BETA.W (between part only)
      exo2_w <- cbind(1, y2[cl, within_x_idx, drop = FALSE])
      tmp2 <- (zc %*% sigma_zi_zy_ji -
        yc %*% sigma_j_inv +
        yc %*% sigma_w_inv)
      g_beta_w1[cl, ] <- as.vector(2 * nj * crossprod(exo2_w, tmp2))

      # BETA.B
      exo2_b <- cbind(1, y2[cl, between_x_idx, drop = FALSE])
      tmp <- (zc %*% sigma_zi_zy_ji - yc %*% sigma_j_inv)
      g_beta_b[cl, ] <- as.vector(2 * nj * crossprod(exo2_b, tmp))
    } # cl
  # between.y.idx

  } else { # no between.y.idx

    # Sigma.j inverses per unique cluster size (was: per cluster)
    cache <- lav_mvn_cl_sigma_j_cache(
      cluster_sizes = cluster_sizes,
      sigma_b_z = sigma_b, sigma_w_1 = sigma_w,
      sinv_method = sinv_method
    )
    size_idx <- match(cluster_size, cluster_sizes)

    for (cl in seq_len(nclusters)) {
      # cluster size
      nj <- cluster_size[cl]

      # data within for the cluster (centered)
      yc <- y2w_res[cl, ]

      # data between
      y2yc_yy <- tcrossprod(y2w_res[cl, ])

      # Sigma.j (cached per cluster size)
      sigma_j_inv <- cache[[size_idx[cl]]]$sigma_j_inv

      # SIGMA blocks (shared kernel; ns = 1)
      g <- lav_mvreg_cl_grad_sigma_kernel(
        nj = nj, ns = 1,
        y2yc_yy = y2yc_yy,
        sigma_j_inv = sigma_j_inv
      )
      g_sigma_w1_1[cl, ] <- g$sigma_w1
      g_sigma_b_1[cl, ] <- g$sigma_b

      # BETA.W (between part only)
      exo2_w <- cbind(1, y2[cl, within_x_idx, drop = FALSE])
      tmp2 <- (-yc %*% sigma_j_inv + yc %*% sigma_w_inv)
      g_beta_w1[cl, ] <- as.vector(2 * nj * crossprod(exo2_w, tmp2))

      # BETA.B
      exo2_b <- cbind(1, y2[cl, between_x_idx, drop = FALSE])
      tmp <- -yc %*% sigma_j_inv
      g_beta_b[cl, ] <- as.vector(2 * nj * crossprod(exo2_b, tmp))
    } # cl
  } # no-between-y

  # beta.w (bis)
  y1_wb_res_i <- y1_wb_res %*% sigma_w_inv
  w1_idx <- seq_len(length(within_x_idx) + 1L)
  a1_idx <- rep(w1_idx, times = ncol(y1_wb_res_i))
  b1_idx <- rep(seq_len(ncol(y1_wb_res_i)), each = length(w1_idx))
  tmp_1 <- exo_wb[, a1_idx, drop = FALSE] * y1_wb_res_i[, b1_idx, drop = FALSE]
  g_beta_w2 <- -2 * rowsum.default(tmp_1, cluster_idx,
    reorder = FALSE,
    na.rm = TRUE
  )
  g_beta_w <- g_beta_w1 + g_beta_w2

  # Sigma.W (bis)
  y1a_res <- y1_wb_res - y2w_res[cluster_idx, , drop = FALSE]
  y1a_res_i <- y1a_res %*% sigma_w_inv
  idx1 <- lav_mat_vech_col_idx(nrow(sigma_w))
  idx2 <- lav_mat_vech_row_idx(nrow(sigma_w))
  sw2 <- matrix(lav_mat_vech(sigma_w_inv),
    nrow = nclusters,
    length(lav_mat_vech(sigma_w_inv)), byrow = TRUE
  )
  sw2 <- sw2 * (cluster_size - 1)
  tmp_1 <- y1a_res_i[, idx1, drop = FALSE] * y1a_res_i[, idx2, drop = FALSE]
  tmp2_1 <- rowsum.default(tmp_1, cluster_idx, reorder = FALSE, na.rm = TRUE)
  g_sigma_w2 <- 2 * (sw2 - tmp2_1)
  diagh_idx <- lav_mat_diagh_idx(nrow(sigma_w))
  g_sigma_w2[, diagh_idx] <- g_sigma_w2[, diagh_idx, drop = FALSE] / 2
  g_sigma_w <- g_sigma_w1_1 + g_sigma_w2



  # rearrange columns to Res.Int.W, Res.Pi.W, Res.Sigma.W,
  #                      Res.Int.B, Res.Pi.B, Res.Sigma.B

  # ov.idx per level
  ov_idx <- lp$ov.idx

  # 'tilde' matrices: ALL variables within and between
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))
  p_tilde_star <- p_tilde * (p_tilde + 1) / 2
  b_tilde <- lav_mat_vech_rev(seq_len(p_tilde_star))

  # only 'y'
  ov_y_idx <- lp$ov.y.idx

  # two levels only (for now)
  ov_y_idx1 <- ov_y_idx[[1]]
  ov_y_idx2 <- ov_y_idx[[2]]

  # WITHIN (is easy)
  beta_w_idx <- matrix(seq_along(beta_w), nrow(beta_w), ncol(beta_w))
  beta_b_idx <- matrix(seq_along(beta_b), nrow(beta_b), ncol(beta_b))

  res_int_w <- g_beta_w[, beta_w_idx[1L, ], drop = FALSE]
  res_pi_w <- g_beta_w[, lav_mat_vecr(beta_w_idx[-1L, ]), drop = FALSE]
  res_sigma_w <- g_sigma_w

  # Sigma.B
  sigma_b_tilde <- matrix(0, nclusters, p_tilde_star)
  col_idx <- lav_mat_vech(b_tilde[ov_y_idx1, ov_y_idx1, drop = FALSE])
  sigma_b_tilde[, col_idx] <- g_sigma_b_1

  # Int.B
  beta_b_tilde <- matrix(seq_len(nrow(beta_b) * p_tilde), nrow(beta_b), p_tilde)
  int_b <- matrix(0, nclusters, p_tilde)
  int_b[, ov_y_idx1] <- g_beta_b[, beta_b_idx[1L, ]]

  # Pi.B
  pi_b <- matrix(0, nclusters, p_tilde * (nrow(beta_b) - 1L))
  col_idx <- lav_mat_vecr(beta_b_tilde[-1L, ov_y_idx1, drop = FALSE])
  pi_b[, col_idx] <- g_beta_b[, lav_mat_vecr(beta_b_idx[-1L, ]),
                                                                drop = FALSE]

  if (length(between_y_idx) > 0L) {
    # Sigma.B: add yz/zz parts
    col_idx <- lav_mat_vec(b_tilde[ov_y_idx1, between_y_idx, drop = FALSE])
    sigma_b_tilde[, col_idx] <- g_sigma_yz_1
    col_idx <- lav_mat_vech(b_tilde[between_y_idx, between_y_idx,
      drop = FALSE
    ])
    sigma_b_tilde[, col_idx] <- g_sigma_zz_1

    # Int.B: add z-part
    beta_z_idx <- matrix(seq_along(beta_z), nrow(beta_z), ncol(beta_z))
    int_b[, between_y_idx] <- g_beta_z[, beta_z_idx[1L, ], drop = FALSE]

    # Pi.B: add beta.z
    col_idx <- lav_mat_vecr(beta_b_tilde[-1L, between_y_idx, drop = FALSE])
    pi_b[, col_idx] <-
      g_beta_z[, lav_mat_vecr(beta_z_idx[-1L, ]), drop = FALSE]
  }

  # only extract ov.y.idx2 for BETWEEN
  col_idx <- lav_mat_vech(b_tilde[ov_y_idx2, ov_y_idx2, drop = FALSE])
  res_sigma_b <- sigma_b_tilde[, col_idx, drop = FALSE]

  res_int_b <- int_b[, ov_y_idx2, drop = FALSE]

  col_idx <- lav_mat_vecr(beta_b_tilde[-1, ov_y_idx2])
  res_pi_b <- pi_b[, col_idx, drop = FALSE]

  scores <- cbind(
    res_int_w, res_pi_w, res_sigma_w,
    res_int_b, res_pi_b, res_sigma_b
  )

  scores
}

# first-order information: outer crossprod of scores per cluster
lav_mvreg_cl_info_firstorder <- function(y1 = NULL,
                                         ylp = NULL,
                                         lp = NULL,
                                         res_sigma_w = NULL,
                                         res_int_w = NULL,
                                         res_pi_w = NULL,
                                         res_sigma_b = NULL,
                                         res_int_b = NULL,
                                         res_pi_b = NULL,
                                         divide_by_two = FALSE,
                                         sinv_method = "eigen") {
  scores <- lav_mvreg_cl_sc_2l(
    y1 = y1,
    ylp = ylp,
    lp = lp,
    res_sigma_w = res_sigma_w,
    res_int_w = res_int_w,
    res_pi_w = res_pi_w,
    res_sigma_b = res_sigma_b,
    res_int_b = res_int_b,
    res_pi_b = res_pi_b,
    sinv_method = sinv_method
  )

  # divide by 2 (if we want scores wrt objective function)
  if (divide_by_two) {
    scores <- scores / 2
  }

  # unit information
  information <- crossprod(scores) / lp$nclusters[[2]]

  information
}

lav_plotinfo_svgcode <- function(plotinfo,
                         outfile = "",
                         sloped_labels = TRUE,
                         standalone = FALSE,
                         stroke_width = 2L,
                         font_size = 20L,
                         idx_font_size = 15L,
                         dy = 5L,
                         mlovcolors = c("lightgreen", "lightblue"),
                         lightness = 1,
                         font_family = "Latin Modern Math, arial, Arial, sans",
                         italic = TRUE,
                         auto_subscript = TRUE
                         ) {
  textattr <- paste0('fill="black" font-size="', font_size,
              '" font-family="', font_family, '" ',
              ifelse(italic, 'font-style="italic"', ''))
  tmpcol <- col2rgb(mlovcolors)
  wovcol <- paste(as.hexmode(tmpcol[, 1L]), collapse = "")
  bovcol <- paste(as.hexmode(tmpcol[, 2L]), collapse = "")
  node_elements_svg <- function(nodetiepe, noderadius, waar, stroke_width) {
    # define form, color and anchors for a node
    localradius <- noderadius
    if (nodetiepe == "varlv") localradius <- noderadius * .8
    ovxy <- localradius * sqrt(0.5)
    cvxy <- localradius * c(0.5, sqrt(0.75))
    constxy <- cvxy
    drawit <- switch(nodetiepe,
                     lv = ,
                     varlv = paste0('<circle cx="', waar[1], '" cy="', waar[2],
                                    '" r="', localradius,
                                    '" stroke-width="', stroke_width,
                                    '" stroke="black" fill="white"/>'),
                     ov = paste0('<rect width="', 2 * ovxy, '" height="',
                                 2 * ovxy, '" x="', waar[1] - ovxy, '" y="',
                                 waar[2] - ovxy,
                                 '" stroke-width="', stroke_width,
                                 '" stroke="black" fill="white" />'),
                     wov =  paste0('<rect width="', 2 * ovxy, '" height="',
                                   2 * ovxy, '" x="', waar[1] - ovxy, '" y="',
                                   waar[2] - ovxy, '" rx="', ovxy / 3, '" ry="',
                                   ovxy / 3, '" stroke-width="', stroke_width,
                                   '" stroke="black" fill="#', wovcol,
                                   '" />'),
                     bov =  paste0('<rect width="', 2 * ovxy, '" height="',
                                   2 * ovxy, '" x="', waar[1] - ovxy, '" y="',
                                   waar[2] - ovxy, '" rx="', ovxy / 3, '" ry="',
                                   ovxy / 3, '" stroke-width="', stroke_width,
                                   '" stroke="black" fill="#', bovcol,
                                   '" />'),
                     cv = paste0('<polygon points="',
                                 waar[1] - cvxy[1], ',', waar[2] - cvxy[2], ' ',
                                 waar[1] + cvxy[1], ',', waar[2] - cvxy[2], ' ',
                                 waar[1] + localradius, ',', waar[2], ' ',
                                 waar[1] + cvxy[1], ',', waar[2] + cvxy[2], ' ',
                                 waar[1] - cvxy[1], ',', waar[2] + cvxy[2], ' ',
                                 waar[1] - localradius, ',', waar[2],
                                 '" stroke-width="', stroke_width,
                                 '" stroke="black" fill="none" />'),
                     const = paste0('<polygon points="',
                                    waar[1], ',', waar[2] - localradius, ' ',
                                    waar[1] + constxy[2], ',',
                                    waar[2] + constxy[1], ' ',
                                    waar[1] - constxy[2], ',',
                                    waar[2] + constxy[1],
                                    '" stroke-width="', stroke_width,
                                    '" stroke="black" fill="none" />')
    )
    n <- c(waar[1], switch(nodetiepe,
                           lv = , varlv = , const = waar[2] - localradius,
                           ov = , wov = , bov = waar[2] - ovxy,
                           cv = waar[2] - cvxy[2]))
    s <- c(waar[1], switch(nodetiepe,
                           lv = , varlv = waar[2] + localradius,
                           ov = , wov = , bov = waar[2] + ovxy,
                           cv = waar[2]  + cvxy[2],
                           const = waar[2] + constxy[1]))
    e <- switch(nodetiepe,
                lv = , varlv = , cv = waar + c(localradius, 0),
                ov = , wov = , bov = waar + c(ovxy, 0),
                const = waar + c(constxy[2], constxy[1]))
    w <- switch(nodetiepe,
                lv = , varlv = , cv = waar + c(-localradius, 0),
                ov = , wov = , bov = waar + c(-ovxy, 0),
                const = waar + c(-constxy[2], constxy[1]))
    ne <- switch(nodetiepe,
                 lv = , varlv = , ov = , wov = ,
                 bov = waar + ovxy * c(1, -1),
                 cv = waar + c(cvxy[1], -cvxy[2]),
                 const = e)
    nw <- switch(nodetiepe,
                 lv = , varlv = , ov = , wov = ,
                 bov = waar + ovxy * c(-1, -1),
                 cv = waar + c(-cvxy[1], -cvxy[2]),
                 const = w)
    se <- switch(nodetiepe,
                 lv = , varlv = , ov = , wov = ,
                 bov = waar + ovxy * c(1, 1),
                 cv = waar + cvxy,
                 const = e)
    sw <- switch(nodetiepe,
                 lv = , varlv = , ov = , wov = ,
                 bov = waar + ovxy * c(-1, 1),
                 cv = waar + c(-cvxy[1L], cvxy[2L]),
                 const = w)
    list(drawit = drawit, n = n, ne = ne, e = e,
         se = se, s = s, sw = sw, w = w, nw = nw)
  }
  get_file_extension <- function(path) {
    if (path == "") return("")
    delen <- strsplit(path, ".", fixed = TRUE)[[1]]
    if (length(delen) > 1L) return(tolower(delen[length(delen)]))
    ""
  }
  if (is.character(outfile) && outfile != "") {
    stopifnot(standalone || get_file_extension(outfile) == "svg",
              !standalone || get_file_extension(outfile) %in% c("htm", "html"))
  }
  mlrij <- plotinfo$mlrij
  if (is.null(mlrij))
    lav_msg_stop(gettext(
      "plotinfo hasn't been processed by lav_plotinfo_positions!"))
  if (outfile == "") outfile <- stdout()
  if (is.character(outfile)) {
    zz <- file(outfile, open = "w")
    closezz <- TRUE
  } else {
    zz <- outfile
    closezz <- FALSE
  }
  nodes <- plotinfo$nodes
  edges <- plotinfo$edges
  nodedist <- 100
  noderadius <- 0.3
  rijen <- max(nodes$rij)
  kolommen <- max(nodes$kolom)
  nodes$rij <- nodes$rij + 1
  nodes$kolom <- nodes$kolom + 1
  if (lightness != 1) {
    mlrij <- lightness * mlrij
    nodes$kolom <- lightness * nodes$kolom
    nodes$rij <- lightness * nodes$rij
    edges$controlpt.kol <- lightness * edges$controlpt.kol
    edges$controlpt.rij <- lightness * edges$controlpt.rij
  }
  if (standalone) {
    writeLines(c(
      '<!DOCTYPE html>',
      '<html>',
      '<body>',
      '<h2>SVG diagram created by lav_plot R package</h2>'),
      zz)
  }
  writeLines(c(
    paste0('<svg width="', lightness * (kolommen + 3) * nodedist, '" height="',
           lightness * (rijen + 3) * nodedist,
           '" version="1.1" xmlns="http://www.w3.org/2000/svg"',
           ' xmlns:xlink="http://www.w3.org/1999/xlink">'),
    '<rect width="100%" height="100%" fill="white" />',
    '<defs>',
    '  <marker id="arr" markerWidth="6" markerHeight="6"',
    '          refX="6" refY="2.5" orient="auto">',
    '    <path d="M 0 0 L 6 2.5 L 0 5 L 2 2.5 z" fill="black" />',
    '  </marker>',
    '  <marker id="sarr" markerWidth="6" markerHeight="6"',
    '          refX="0" refY="2.5" orient="auto">',
    '    <path d="M 0 2.5 L 6 0 L 4 2.5 L 6 5 z" fill="black" />',
    '  </marker>',
    '</defs>'),
     zz)
  plot_edge <- function(van, naar, label = "", dubbel = FALSE,
                        control = NA_real_, below = FALSE,
                        id = 0) {
    labele <- lav_label_code(label,
                               idx_font_size = idx_font_size,
                               dy = dy,
                               auto_subscript = auto_subscript)$svg
    theta <- atan2(naar[2] - van[2], naar[1] - van[1])
    if (is.na(control[1L])) { # line
      if (van[1L] <= naar[1L]) {
        writeLines(paste0('<path id="L', id, '" d="M ', van[1L],
                          ' ', van[2L], ' L ', naar[1L], " ", naar[2L],
                        '" stroke-width="', stroke_width, '" stroke="black" ',
                        ifelse(dubbel, 'marker-start="url(#sarr)" ', ''),
                        'marker-end="url(#arr)" />'), zz)
      } else {
        writeLines(paste0('<path d="M ', van[1L],
                          ' ', van[2L], ' L ', naar[1L], " ", naar[2L],
                          '" stroke-width="', stroke_width, '" stroke="black" ',
                          ifelse(dubbel, 'marker-start="url(#sarr)" ', ''),
                          'marker-end="url(#arr)" />'), zz)
        writeLines(paste0('<path id="L', id, '" d="M ', naar[1L],
                          ' ', naar[2L], ' L ', van[1L], " ", van[2L],
                          '" stroke-width="0" stroke="none" fill="none" />'),
                   zz)
      }
      midden <- (van + naar) * 0.5
    } else {  # path Q (quadratic B\'{e}zier)
      if (van[1L] <= naar[1L]) {
        writeLines(paste0('<path id="L', id, '" d="M ', van[1L], ' ',
                          van[2L], ' Q ', control[1L], ' ', control[2L],
                          ' ', naar[1L], " ", naar[2L],
                          '" stroke-width="', stroke_width,
                          '" stroke="black" fill="none" ',
                          ifelse(dubbel, 'marker-start="url(#sarr)" ', ''),
                          'marker-end="url(#arr)" />'), zz)
      } else {
        writeLines(paste0('<path d="M ', van[1L], ' ',
                          van[2L], ' Q ', control[1L], ' ', control[2L],
                          ' ', naar[1L], " ", naar[2L],
                          '" stroke-width="', stroke_width,
                          '" stroke="black" fill="none" ',
                          ifelse(dubbel, 'marker-start="url(#sarr)" ', ''),
                          'marker-end="url(#arr)" />'), zz)
        writeLines(paste0('<path id="L', id, '" d="M ', naar[1L], ' ',
                          naar[2L], ' Q ', control[1L], ' ', control[2L],
                          ' ', van[1L], " ", van[2L],
                          '" stroke-width="0" stroke="none" fill="none" />'),
                   zz)
      }
      midden <- 0.25 * (van + naar) + 0.5 * control
    }
    if (label != "") {
      if (sloped_labels) {
        writeLines(
          c(paste0('<text ', textattr, ' text-anchor="middle">'),
            paste0('<textPath xlink:href="#L', id, '" startOffset="50%">',
                   labele, '</textPath>'),
            '</text>'), zz)
      } else {
        if (below) {
        if (theta >= 0 && theta < pi / 2) {
          extra <- 'dy="30"'
        } else if (theta >= pi / 2) {
          extra <- 'dy="30" text-anchor="end"'
        } else if (theta < -pi / 2) {
          extra <- 'dy="30"'
        } else {
          extra <- 'dy="0" text-anchor="end"'
        }
      } else {
        if (theta >= 0 && theta < pi / 2) {
          extra <- 'text-anchor="end"'
        } else if (theta >= pi / 2) {
          extra <- ' '
        } else if (theta < -pi / 2) {
          extra <- 'text-anchor="end"'
        } else {
          extra <- ' '
        }
        writeLines(paste0('<text x="', midden[1L], '" y="', midden[2L],
                          '" ', textattr, ' ', extra, '>', labele, '</text>'),
                   zz)
        }
      }
    }
  }
  plot_var <- function(waar, noderadius, label = "", side = "n") {
    labele <- lav_label_code(label,
                               idx_font_size = idx_font_size,
                               dy = dy,
                               auto_subscript = auto_subscript)$svg
    thetarange <- c(pi / 6, 11 * pi / 6)
    if (side == "s") thetarange <- thetarange + 3 * pi / 2
    if (side == "e") thetarange <- thetarange + pi
    if (side == "n") thetarange <- thetarange + pi / 2
    localradius <- noderadius * 0.8
    middelpt <- switch(side,
                       n = c(0, -localradius),
                       w = c(-localradius, 0),
                       s = c(0, localradius),
                       e = c(localradius, 0))
    middelpt <- middelpt + waar
    # cirkelsegment
    straal <- localradius
    xs <- middelpt[1] + cos(thetarange) * straal
    ys <- middelpt[2] + sin(thetarange) * straal
    writeLines(paste0(
      '<path d="M ', xs[1L], ' ', ys[1L], ' A ', straal, ' ', straal,
      ' 0 1,1 ', xs[2L], ' ', ys[2L], '" stroke-width="', stroke_width,
      '" stroke="black" fill="none" ',
      'marker-start="url(#sarr)" marker-end="url(#arr)" />'
    ), zz)
    # label
    if (label != "") {
      writeLines(paste0('<text x="', middelpt[1L], '" y="', middelpt[2L],
                        '" text-anchor="middle" ', textattr, '>', labele,
                        '</text>'), zz)
    }
  }
  plot_node <- function(waar, tiepe, label = "") {
    labele <- lav_label_code(label,
                               idx_font_size = idx_font_size,
                               dy = dy,
                               auto_subscript = auto_subscript)$svg
    elems <- node_elements_svg(tiepe, nodedist * noderadius, waar, stroke_width)
    writeLines(c(
      elems$drawit,
      paste0('<text x="', waar[1], '" y="', waar[2], '" ',
              textattr, ' dominant-baseline="central" text-anchor="middle">',
             labele, '</text>')
    ), zz)
  }

  if (mlrij > 0L) {
    mlrij <- mlrij + lightness
    writeLines(paste0('<path d="M 1 ', mlrij * nodedist, ' L ',
                       (max(nodes$kolom) + lightness) * nodedist,
                      ' ', mlrij * nodedist, '" stroke="black"/>'),
               zz)
  }
  for (j in seq.int(nrow(edges))) {
    if (edges$naar[j] != edges$van[j]) {
      van <- which(nodes$id == edges$van[j])
      naar <- which(nodes$id == edges$naar[j])
      adrvan <- c(nodedist * nodes$kolom[van], nodedist * nodes$rij[van])
      elems <- node_elements_svg(nodes$tiepe[van], nodedist * noderadius,
                                 adrvan, stroke_width)
      adrvan <- elems[[edges$vananker[j]]]
      adrnaar <- c(nodedist * nodes$kolom[naar], nodedist * nodes$rij[naar])
      elems <- node_elements_svg(nodes$tiepe[naar], nodedist * noderadius,
                                 adrnaar, stroke_width)
      adrnaar <- elems[[edges$naaranker[j]]]
      if (is.na(edges$controlpt.rij[j])) {
        plot_edge(adrvan, adrnaar, edges$label[j],
                  dubbel = (edges$tiepe[j] == "~~"),
                  below = edges$labelbelow[j], id = j)
      } else {
        controlpt <- nodedist * c(edges$controlpt.kol[j] + 1,
                                  edges$controlpt.rij[j] + 1)
        plot_edge(adrvan, adrnaar, edges$label[j],
                  dubbel = (edges$tiepe[j] == "~~"),
                  below = edges$labelbelow[j],
                  control = controlpt,
                  id = j
        )
      }
    } else {
      van <- which(nodes$id == edges$van[j])
      adrvan <- c(nodedist * nodes$kolom[van], nodedist * nodes$rij[van])
      elems <- node_elements_svg(nodes$tiepe[van], nodedist * noderadius,
        adrvan, stroke_width)
      adrvan <- elems[[edges$vananker[j]]]
      plot_var(adrvan, noderadius * nodedist, edges$label[j], edges$vananker[j])
    }
  }
  for (j in seq.int(nrow(nodes))) {
    plot_node(nodedist * c(nodes$kolom[j], nodes$rij[j]),
              nodes$tiepe[j],
              nodes$naam[j])
  }
  writeLines("</svg>", zz)
  if (standalone) writeLines(c("</body>", "</html>"), zz)
  if (closezz) close(zz)
  invisible(NULL)
}

# constructor of the matrix lavoptions$representation
#
# initial version: YR 22/11/2010
# - YR 14 Jan 2014: moved to lav_model.R
# - YR 18 Nov 2014: more efficient handling of linear equality constraints
# - YR 02 Dec 2014: allow for bare-minimum parameter tables
# - YR 25 Jan 2017: collect options in lavoptions
# - YR 12 Mar 2021: add lavpta as argument; create model attributes (ma)
# - YR 21 Jan 2025: add composites

# construct MATRIX lavoptions$representation of the model
lav_model <- function(lavpartable = NULL,
                      lavoptions = NULL,
                      th_idx = list()) {
  # handle bare-minimum partables
  lavpartable <- lav_pt_complete(lavpartable)
  lavpta <- lav_pt_attributes(lavpartable)
  lavpartable <- lav_pt_set_cache(lavpartable, lavpta)

  # global info from user model
  nblocks <- lav_pt_nblocks(lavpartable)
  ngroups <- lav_pt_ngroups(lavpartable)
  meanstructure <- any(lavpartable$op == "~1")
  correlation <- lavoptions$correlation
  if (is.null(correlation)) {
    correlation <- FALSE
  }
  # partial correlation structure: only these observed variables have unit
  # variance; the others keep a free residual variance and are treated like
  # 'numeric' variables (i.e. added to num.idx below). Empty -> all ov.
  correlation_ov <- lavoptions$.correlation.ov
  if (is.null(correlation_ov)) {
    correlation_ov <- character(0L)
  }
  composites_option <- lavoptions$composites
  if (is.null(composites_option)) {
    composites_option <- TRUE
  }
  composites <- any(lavpartable$op == "<~") && composites_option
  group_w_free <- any(lavpartable$lhs == "group" & lavpartable$op == "%")
  multilevel <- FALSE
  if (!is.null(lavpartable$level)) {
    nlevels <- lav_pt_nlevels(lavpartable)
    if (nlevels > 1L) {
      multilevel <- TRUE
    }
  } else {
    nlevels <- 1L
  }
  categorical <- any(lavpartable$op == "|")
  if (categorical) {
    meanstructure <- TRUE

    # handle th.idx if length(th.idx) != nblocks
    if (nblocks != length(th_idx)) {
      if (multilevel) {
        # per BLOCK: the levels may span different variable sets, so
        # rebuild th.idx from the parameter table (in the block's
        # variable order: ordinal variables contribute their thresholds,
        # numeric variables a 0 placeholder); if conditional.x, the
        # exogenous covariates are not part of the variable set
        conditional_x_flag <- any(lavpartable$exo > 0L &
                                  lavpartable$op == "~")
        th_idx <- vector("list", nblocks)
        for (b in seq_len(nblocks)) {
          ov_names_b <- lav_pt_vnames(lavpartable,
            if (conditional_x_flag) "ov.nox" else "ov", block = b)
          th_lhs_b <- lavpartable$lhs[lavpartable$op == "|" &
                                      lavpartable$block == b]
          th_idx_b <- integer(0L)
          for (v in seq_along(ov_names_b)) {
            nth_v <- sum(th_lhs_b == ov_names_b[v])
            if (nth_v > 0L) {
              th_idx_b <- c(th_idx_b, rep(v, nth_v))
            } else {
              th_idx_b <- c(th_idx_b, 0L)
            }
          }
          th_idx[[b]] <- th_idx_b
        }
      } else {
        th_idx <- rep(th_idx, each = nblocks)
      }
    }
  }

  nefa <- lav_pt_nefa(lavpartable)
  if (nefa > 0L) {
    efa_values <- lav_pt_efa_values(lavpartable)
    # early check of *named* per-set target/target_mask lists (issue #516),
    # so a name mismatch is caught before the model is estimated
    if (!is.null(lavoptions$rotation) &&
        any(lavoptions$rotation == c("target.strict", "pst"))) {
      lav_model_efa_check_target(
        ropts = lavoptions$rotation.args,
        lavpartable = lavpartable, efa_values = efa_values
      )
    }
  }

  # check for simple equality constraints
  eq_simple <- any(lavpartable$free > 0L & duplicated(lavpartable$free))
  if (eq_simple) {
    # just like in <0.5-18, add (temporary) 'unco' column
    # so we can fill in x.unco.idx
    lavpartable$unco <- integer(length(lavpartable$id))
    idx_free <- which(lavpartable$free > 0L)
    lavpartable$unco[idx_free] <- seq_along(idx_free)
  }

  # handle variable definitions and (in)equality constraints
  tmp_con <- lav_con_parse(
    partable = lavpartable,
    constraints = NULL
  )

  # handle *linear* equality constraints special
  if (tmp_con$ceq.linear.only.flag) {
    con_jac <- tmp_con$ceq.JAC
    con_lambda <- numeric(nrow(tmp_con$ceq.JAC))
    attr(con_jac, "inactive.idx") <- integer(0L)
    attr(con_jac, "ceq.idx") <- seq_len(nrow(tmp_con$ceq.JAC))
  } else {
    con_jac <- matrix(0, 0, 0)
    con_lambda <- numeric(0)
  }

  # select model matrices
  if (lavoptions$representation == "LISREL") {
    tmp_rep <- lav_lisrel(lavpartable, target = NULL, extra = TRUE,
                          allow_composites = composites)
  } else if (lavoptions$representation == "RAM") {
    tmp_rep <- lav_ram(lavpartable, target = NULL, extra = TRUE)
  } else {
    lav_msg_stop(gettextf(
      "%1$s argument must be either %2$s or %3$s",
      "representation", "LISREL", "RAM"))
  }
  if (lav_debug()) print(tmp_rep)

  # FIXME: check for non-existing parameters
  bad_idx <-
    which((tmp_rep$mat == "" | is.na(tmp_rep$row) | is.na(tmp_rep$col)) &
          !lavpartable$op %in% c("==", "<", ">", ":=", "|~"))

  if (length(bad_idx) > 0L) {
    this_formula <- paste(lavpartable$lhs[bad_idx[1]],
      lavpartable$op[bad_idx[1]],
      lavpartable$rhs[bad_idx[1]],
      sep = " "
    )
    if (lavoptions$representation == "LISREL") {
      lav_msg_stop(gettextf(
        "a model parameter is not defined in the LISREL representation %s.
        Upgrade to latent variables or consider using representation = 'RAM'.",
        this_formula))
    } else {
      lav_msg_stop(
        gettextf("parameter is not defined: %s", this_formula)
      )
    }
  }

  # prepare nG-sized slots
  tmp_ng <- sum(unlist(attr(tmp_rep, "mmNumber")))
  tmp_glist <- vector(mode = "list", tmp_ng)
  names(tmp_glist) <- unlist(attr(tmp_rep, "mmNames"))
  dim_names <- vector(mode = "list", length = tmp_ng)
  is_symmetric <- logical(tmp_ng)
  mm_size <- integer(tmp_ng)

  m_free_idx <- m_user_idx <- vector(mode = "list", length = tmp_ng)
  x_free_idx <- x_unco_idx <- x_user_idx <- vector(
    mode = "list",
    length = tmp_ng
  )

  # prepare nblocks-sized slots
  nvar <- integer(nblocks)
  nmat <- unlist(attr(tmp_rep, "mmNumber"))
  num_idx <- vector("list", length = nblocks)
  nexo <- integer(nblocks)
  ov_x_dummy_ov_idx <- vector(mode = "list", length = nblocks)
  ov_x_dummy_lv_idx <- vector(mode = "list", length = nblocks)
  ov_y_dummy_ov_idx <- vector(mode = "list", length = nblocks)
  ov_y_dummy_lv_idx <- vector(mode = "list", length = nblocks)
  ov_efa_idx <- vector(mode = "list", length = nblocks)
  lv_efa_idx <- vector(mode = "list", length = nblocks)

  offset <- 0L
  # keep track of ov.names across blocks
  for (g in 1:nblocks) {
    # observed and latent variables for this block
    ov_names <- lav_pt_vnames(lavpartable, "ov", block = g)
    ov_names_nox <- lav_pt_vnames(lavpartable, "ov.nox", block = g)
    ov_names_x <- lav_pt_vnames(lavpartable, "ov.x", block = g)
    ov_num <- lav_pt_vnames(lavpartable, "ov.num", block = g)
    if (lavoptions$conditional.x) {
      if (nlevels > 1L) {
        if (ngroups == 1L) {
          other_block_names <- lav_pt_vnames(lavpartable, "ov",
            block = seq_len(nblocks)[-g]
          )
        } else {
          # TEST ME!
          # which group is this?
          this_group <- ceiling(g / nlevels)
          blocks_within_group <- (this_group - 1L) * nlevels + seq_len(nlevels)
          other_block_names <- lav_pt_vnames(lavpartable, "ov",
            block = blocks_within_group[-g]
          )
        }


        if (length(ov_names_x) > 0L) {
          idx <- which(ov_names_x %in% other_block_names)
          if (length(idx) > 0L) {
            ov_names_nox <- unique(c(ov_names_nox, ov_names_x[idx]))
            ov_names_x <- ov_names_x[-idx]
            ov_names <- ov_names_nox
          }
        }
      }
      nvar[g] <- length(ov_names_nox)
      if (correlation) {
        if (length(correlation_ov) > 0L) {
          # partial: non-correlation variables keep a free residual variance
          num_idx[[g]] <- which(!(ov_names_nox %in% correlation_ov))
        } else {
          num_idx[[g]] <- integer(0L)
        }
      } else {
        num_idx[[g]] <- which(ov_names_nox %in% ov_num)
      }
    } else {
      nvar[g] <- length(ov_names)
      if (correlation) {
        if (length(correlation_ov) > 0L) {
          # partial: non-correlation variables keep a free residual variance
          num_idx[[g]] <- which(!(ov_names %in% correlation_ov))
        } else {
          num_idx[[g]] <- integer(0L)
        }
      } else {
        num_idx[[g]] <- which(ov_names %in% ov_num)
      }
    }
    nexo[g] <- length(ov_names_x)

    if (nefa > 0L) {
      lv_names <- lav_pt_vnames(lavpartable, "lv", block = g)
    }

    # model matrices for this block
    mm_number <- attr(tmp_rep, "mmNumber")[[g]]
    mm_names <- attr(tmp_rep, "mmNames")[[g]]
    mm_symmetric <- attr(tmp_rep, "mmSymmetric")[[g]]
    mm_dim_names <- attr(tmp_rep, "mmDimNames")[[g]]
    mm_rows <- attr(tmp_rep, "mmRows")[[g]]
    mm_cols <- attr(tmp_rep, "mmCols")[[g]]

    for (mm in 1:mm_number) {
      # offset in tmp.glist
      offset <- offset + 1L

      # matrix size, symmetric, dim.names
      if (mm_symmetric[mm]) {
        tmp_n <- mm_rows[mm]
        mm_size <- as.integer(tmp_n * (tmp_n + 1) / 2)
      } else {
        mm_size <- as.integer(mm_rows[mm] * mm_cols[mm])
      }
      mm_size[offset] <- mm_size
      is_symmetric[offset] <- mm_symmetric[mm]
      dim_names[[offset]] <- mm_dim_names[[mm]]

      # select elements for this matrix
      idx <- which(lavpartable$block == g & tmp_rep$mat == mm_names[mm])

      # create empty `pattern' matrix
      # FIXME: one day, we may want to use sparse matrices...
      #        but they should not slow things down!
      tmp <- matrix(0L,
        nrow = mm_rows[mm],
        ncol = mm_cols[mm]
      )

      # 1. first assign free values only, to get vector index
      #    -> to be used in lav_model_objective
      tmp[cbind(tmp_rep$row[idx], tmp_rep$col[idx])] <- lavpartable$free[idx]
      if (mm_symmetric[mm]) {
        # NOTE: we assume everything is in the UPPER tri!
        tmp_tt <- t(tmp)
        tmp[lower.tri(tmp)] <- tmp_tt[lower.tri(tmp_tt)]
      }
      m_free_idx[[offset]] <- which(tmp > 0)
      x_free_idx[[offset]] <- tmp[which(tmp > 0)]

      # 2. if simple equality constraints, unconstrained free parameters
      #    -> to be used in lav_model_gradient
      if (eq_simple) {
        tmp[cbind(
          tmp_rep$row[idx],
          tmp_rep$col[idx]
        )] <- lavpartable$unco[idx]
        if (mm_symmetric[mm]) {
          # NOTE: we assume everything is in the UPPER tri!
          tmp_tt <- t(tmp)
          tmp[lower.tri(tmp)] <- tmp_tt[lower.tri(tmp_tt)]
        }
        # m.unco.idx[[offset]] <-     which(tmp > 0)
        x_unco_idx[[offset]] <- tmp[which(tmp > 0)]
      } else {
        # m.unco.idx[[offset]] <- m.free.idx[[offset]]
        x_unco_idx[[offset]] <- x_free_idx[[offset]]
      }

      # 3. general mapping between user and tmp.glist
      tmp[cbind(tmp_rep$row[idx], tmp_rep$col[idx])] <- lavpartable$id[idx]
      if (mm_symmetric[mm]) {
        tmp_tt <- t(tmp)
        tmp[lower.tri(tmp)] <- tmp_tt[lower.tri(tmp_tt)]
      }
      m_user_idx[[offset]] <- which(tmp > 0)
      x_user_idx[[offset]] <- tmp[which(tmp > 0)]

      # 4. now assign starting/fixed values
      # create empty matrix
      # FIXME: again, we may want to use sparse matrices here...
      tmp <- matrix(0.0,
        nrow = mm_rows[mm],
        ncol = mm_cols[mm]
      )
      tmp[cbind(tmp_rep$row[idx], tmp_rep$col[idx])] <- lavpartable$start[idx]
      if (mm_symmetric[mm]) {
        tmp_tt <- t(tmp)
        tmp[lower.tri(tmp)] <- tmp_tt[lower.tri(tmp_tt)]
      }

      # 4b. override with cov.x (if conditional.x = TRUE)
      # new in 0.6-1
      # shouldn't be needed, if lavpartable$start contains cov.x values
      # if(mm.names[mm] == "cov.x") {
      #    tmp <- cov.x[[g]]
      # }
      # 4c. override with mean.x (if conditional.x = TRUE)
      # new in 0.6-1
      # shouldn't be needed, if lavpartable$start contains mean.x values
      # if(mm.names[mm] == "mean.x") {
      #    tmp <- as.matrix(mean.x[[g]])
      # }

      # representation specific stuff
      if (lavoptions$representation == "LISREL" &&
        mm_names[mm] == "lambda") {
        ov_dummy_names_nox <- attr(tmp_rep, "ov.dummy.names.nox")[[g]]
        ov_dummy_names_x <- attr(tmp_rep, "ov.dummy.names.x")[[g]]
        ov_dummy_names <- c(ov_dummy_names_nox, ov_dummy_names_x)
        # define dummy latent variables
        if (length(ov_dummy_names)) {
          # in this case, lv.names will be extended with the dummys
          tmp_lv_names <- mm_dim_names$psi[[1]]
          row_tmp_idx <- match(ov_dummy_names, ov_names)
          col_tmp_idx <- match(ov_dummy_names, tmp_lv_names)
          # Fix lambda values to 1.0
          tmp[cbind(row_tmp_idx, col_tmp_idx)] <- 1.0

          ov_x_dummy_ov_idx[[g]] <- match(ov_dummy_names_x, ov_names)
          ov_x_dummy_lv_idx[[g]] <- match(ov_dummy_names_x, tmp_lv_names)
          ov_y_dummy_ov_idx[[g]] <- match(ov_dummy_names_nox, ov_names)
          ov_y_dummy_lv_idx[[g]] <- match(ov_dummy_names_nox, tmp_lv_names)
        }
      }

      # representation specific
      if (lavoptions$representation == "LISREL" && mm_names[mm] == "delta") {
        # only categorical values are listed in the lavpartable
        # but all remaining values should be 1.0
        idx <- which(tmp[, 1L] == 0.0)
        tmp[idx, 1L] <- 1.0
      }

      # representation specific
      if (lavoptions$representation == "RAM" && mm_names[mm] == "ov.idx") {
        tmp[1, ] <- attr(tmp_rep, "ov.idx")[[g]]
      }

      # assign matrix to tmp.glist
      tmp_glist[[offset]] <- tmp
    } # mm

    # efa related info
    if (nefa > 0L) {
      ov_efa_idx[[g]] <- vector("list", length = nefa)
      lv_efa_idx[[g]] <- vector("list", length = nefa)
      for (set in seq_len(nefa)) {
        # determine ov idx for this set
        ov_efa <-
          unique(lavpartable$rhs[lavpartable$op == "=~" &
            lavpartable$block == g &
            lavpartable$efa == efa_values[set]])
        ov_efa_idx[[g]][[set]] <- match(ov_efa, ov_names)

        lv_efa <-
          unique(lavpartable$lhs[lavpartable$op == "=~" &
            lavpartable$block == g &
            lavpartable$efa == efa_values[set]])
        lv_efa_idx[[g]][[set]] <- match(lv_efa, lv_names)
      }
      names(ov_efa_idx[[g]]) <- efa_values
      names(lv_efa_idx[[g]]) <- efa_values
    } # efa

    # set variances composites (new in 0.6-20)
    if (composites) {
      mm_in_group <- 1:nmat[g] + cumsum(c(0L, nmat))[g]
      tmp_glist[mm_in_group] <-
        lav_lisrel_comp_set_intresvar(tmp_glist[mm_in_group])
    }
  } # g

  # fixed.x parameters?
  # fixed.x <- any(lavpartable$exo > 0L & lavpartable$free == 0L)
  # if(categorical) {
  #    fixed.x <- TRUE
  # }

  # dirty hack to mimic MUML
  if (!is.null(lavoptions$tech.muml.scale)) {
    lav_msg_warn(gettext("using muml scale in group 2"))

    # find matrix
    lambda_idx <- which(names(tmp_glist) == "lambda")[2L]

    # find rows/cols
    b_names <- paste0("b", ov_names) ## ad-hoc assumption!!!
    tmp_cols <- match(b_names, tmp_lv_names)
    tmp_rows <- seq_len(nvar[2])
    stopifnot(length(tmp_cols) == length(tmp_rows))
    tmp_glist[[lambda_idx]][cbind(tmp_rows, tmp_cols)] <-
      lavoptions$tech.muml.scale
  }

  # which free parameters are observed variances?
  ov_names <- lav_pt_vnames(lavpartable, "ov")
  x_free_var_idx <- lavpartable$free[lavpartable$free &
    # !duplicated(lavpartable$free) &
    lavpartable$lhs %in% ov_names &
    lavpartable$op == "~~" &
    lavpartable$lhs == lavpartable$rhs]

  rv_lv <- rv_ov <- list()
  if (multilevel) {
    # store information about random slopes (if any)
    lv_names <- lav_pt_vnames(lavpartable, "lv")
    # we should also add the 'split' both-level observed variables: a
    # random slope on such a variable multiplies its *latent*
    # within-cluster component (latent decomposition), regardless of
    # its exogenous (fixed.x) status
    # FIXME: make this work for multiple work multilevel
    level_values <- lav_pt_level_values(lavpartable)
    ov1 <- lav_object_vnames(lavpartable, "ov", level = level_values[1])
    ov2 <- lav_object_vnames(lavpartable, "ov", level = level_values[2])
    ov12 <- ov2[ov2 %in% ov1]
    lv_names <- c(lv_names, ov12)

    # RV LV
    rv_idx <- which(nchar(lavpartable$rv) > 0L &
      lavpartable$level == level_values[1] &
      lavpartable$rhs %in% lv_names)
    if (length(rv_idx)) {
      rv_lv <- lapply(rv_idx, function(x) {
        c(lavpartable$lhs[x], lavpartable$rhs[x])
      })
      names(rv_lv) <- lavpartable$rv[rv_idx]
    }

    # RV OV
    rv_idx <- which(nchar(lavpartable$rv) > 0L &
      lavpartable$level == level_values[1] &
      !lavpartable$rhs %in% lv_names)
    if (length(rv_idx)) {
      rv_ov <- lapply(rv_idx, function(x) {
        c(lavpartable$lhs[x], lavpartable$rhs[x])
      })
      names(rv_ov) <- lavpartable$rv[rv_idx]
    }
  } # multilevel

  # new in 0.6-9: model properties
  modprop <- lav_model_properties(
    glist = tmp_glist,
    lavpartable = lavpartable,
    nmat = nmat,
    m_free_idx = m_free_idx
  )

  # GLS: the gradient reconstructs the (possibly rescaled) V11 block of
  # the normal-theory weight matrix without forming WLS.V, so it needs
  # this flag (see lav_model_omega)
  estimator_args <- lavoptions$estimator.args
  if (lavoptions$estimator == "GLS") {
    estimator_args$gls.v11.mplus <- lavoptions$gls.v11.mplus
  }

  tmp_model <- new("lavModel",
    GLIST = tmp_glist,
    dimNames = dim_names,
    isSymmetric = is_symmetric,
    mmSize = mm_size,
    representation = lavoptions$representation,
    modprop = modprop,
    meanstructure = meanstructure,
    correlation = correlation,
    composites = composites,
    categorical = categorical,
    multilevel = multilevel,
    link = lavoptions$link,
    nblocks = nblocks,
    ngroups = ngroups, # breaks rsem????
    nefa = nefa,
    group.w.free = group_w_free,
    nmat = nmat,
    nvar = nvar,
    num.idx = num_idx,
    th.idx = th_idx,
    nx.free = max(lavpartable$free),
    nx.unco = if (is.null(lavpartable$unco)) {
      max(lavpartable$free)
    } else {
      max(lavpartable$unco)
    },
    nx.user = max(lavpartable$id),
    m.free.idx = m_free_idx,
    x.free.idx = x_free_idx,
    x.free.var.idx = x_free_var_idx,
    # m.unco.idx=m.unco.idx,
    x.unco.idx = x_unco_idx,
    m.user.idx = m_user_idx,
    x.user.idx = x_user_idx,
    x.def.idx = which(lavpartable$op == ":="),
    x.ceq.idx = which(lavpartable$op == "=="),
    x.cin.idx = which(lavpartable$op == ">" | lavpartable$op == "<"),
    ceq.simple.only = tmp_con$ceq.simple.only,
    ceq.simple.K = tmp_con$ceq.simple.K,
    eq.constraints = tmp_con$ceq.linear.only.flag,
    eq.constraints.K = tmp_con$ceq.JAC.NULL,
    eq.constraints.k0 = tmp_con$ceq.rhs.NULL,
    def.function = tmp_con$def.function,
    ceq.function = tmp_con$ceq.function,
    ceq.JAC = tmp_con$ceq.JAC,
    ceq.rhs = tmp_con$ceq.rhs,
    ceq.jacobian = tmp_con$ceq.jacobian,
    ceq.linear.idx = tmp_con$ceq.linear.idx,
    ceq.nonlinear.idx = tmp_con$ceq.nonlinear.idx,
    cin.simple.only = tmp_con$cin.simple.only,
    cin.function = tmp_con$cin.function,
    cin.JAC = tmp_con$cin.JAC,
    cin.rhs = tmp_con$cin.rhs,
    cin.jacobian = tmp_con$cin.jacobian,
    cin.linear.idx = tmp_con$cin.linear.idx,
    cin.nonlinear.idx = tmp_con$cin.nonlinear.idx,
    con.jac = con_jac,
    con.lambda = con_lambda,
    nexo = nexo,
    fixed.x = lavoptions$fixed.x,
    conditional.x = lavoptions$conditional.x,
    parameterization = lavoptions$parameterization,
    ov.x.dummy.ov.idx = ov_x_dummy_ov_idx,
    ov.x.dummy.lv.idx = ov_x_dummy_lv_idx,
    ov.y.dummy.ov.idx = ov_y_dummy_ov_idx,
    ov.y.dummy.lv.idx = ov_y_dummy_lv_idx,
    ov.efa.idx = ov_efa_idx,
    lv.efa.idx = lv_efa_idx,
    rv.lv = rv_lv,
    rv.ov = rv_ov,
    estimator = lavoptions$estimator,
    estimator.args = estimator_args
  )

  if (lav_debug()) {
    cat("lavaan debug: lavaanModel\n")
    print(str(tmp_model))
    print(tmp_model@GLIST)
  }

  tmp_model
}

# for backwards compatibility
# tmp.model <- lav_model

# check if a fitted model is admissible
lav_object_post_check <- function(object) {
  stopifnot(inherits(object, "lavaan"))
  lavpartable <- object@ParTable
  lavmodel <- object@Model
  lavdata <- object@Data

  var_ov_ok <- var_lv_ok <- result_ok <- TRUE
  var_na <- FALSE

  # 1a. check for negative variances ov
  var_idx <- which(lavpartable$op == "~~" &
    lavpartable$lhs %in% lav_object_vnames(object, "ov") &
    lavpartable$lhs == lavpartable$rhs)
  if (any(is.na(lavpartable$est[var_idx]))) {
    # perhaps estimator = "IV" + stage 1 only
    var_na <- TRUE
  } else if (length(var_idx) > 0L && any(lavpartable$est[var_idx] < 0.0)) {
    result_ok <- var_ov_ok <- FALSE
    lav_msg_warn(gettext("some estimated ov variances are negative"))
  }

  # 1b. check for negative variances lv
  var_idx <- which(lavpartable$op == "~~" &
    lavpartable$lhs %in% lav_object_vnames(object, "lv") &
    lavpartable$lhs == lavpartable$rhs)
  if (any(is.na(lavpartable$est[var_idx]))) {
    # perhaps estimator = "IV" + stage 1 only
    var_na <- TRUE
  } else if (length(var_idx) > 0L && any(lavpartable$est[var_idx] < 0.0)) {
    result_ok <- var_lv_ok <- FALSE
    lav_msg_warn(gettext("some estimated lv variances are negative"))
  }

  # 2. is cov.lv (PSI) positive definite? (only if we did not already warn
  # for negative variances)
  if (!var_na && var_lv_ok &&
      length(lav_object_vnames(lavpartable, type = "lv.regular")) > 0L) {
    eta <- lavTech(object, "cov.lv")
    for (g in 1:lavdata@ngroups) {
      if (nrow(eta[[g]]) == 0L) next
      txt_group <- if (lavdata@ngroups > 1L) gettextf("in group %s", g) else ""
      eigvals <- eigen(eta[[g]], symmetric = TRUE, only.values = TRUE)$values
      if (any(eigvals < -1 * .Machine$double.eps^(3 / 4))) {
        lav_msg_warn(gettextf(
          "covariance matrix of latent variables is not positive definite %s;
          use lavInspect(fit, \"cov.lv\") to investigate.", txt_group
        ))
        result_ok <- FALSE
      }
    }
  }

  # 3. is THETA positive definite (but only for numeric variables)
  # and if we have not already warned for negative ov variances
  if (!var_na && var_ov_ok) {
    mm_theta <- lavTech(object, "theta")
    for (g in 1:lavdata@ngroups) {
      num_idx <- lavmodel@num.idx[[g]]
      if (length(num_idx) > 0L) {
        txt_group <- ""
        if (lavdata@ngroups > 1L) txt_group <- gettextf("in group %s", g)
        eigvals <- eigen(mm_theta[[g]][num_idx, num_idx, drop = FALSE],
          symmetric = TRUE,
          only.values = TRUE
        )$values
        if (any(eigvals < -1 * .Machine$double.eps^(3 / 4))) {
          lav_msg_warn(gettextf(
            "the covariance matrix of the residuals of the observed variables
            (theta) is not positive definite %s; use lavInspect(fit, \"theta\")
            to investigate.", txt_group))
          result_ok <- FALSE
        }
      }
    }
  }

  result_ok
}

# All code below is written by Myrsini Katsikatsou (Feb 2015)

# The following function refers to PLRT for nested models and equality
# constraints. Namely, it is developed to test either of the following
# hypotheses:
# a) H0 states that some parameters are equal to 0
# b) H0 states that some parameters are equal to some others.
# Note that for the latter I haven't checked if it is ok when equality
# constraints are imposed on parameters that refer to different groups in a
# multi-group analysis.
# All the code below has been developed for a single-group analysis.

# Let fit_objH0 and fit_objH1 be the outputs of lavaan() function when we fit
# a model under the null hypothesis and under the alternative, respectively.
# The argument equal_constr is logical (T/F) and it is TRUE if  equality
# constraints are imposed on subsets of the parameters.

# The main idea of the code below is that we consider the parameter vector
# under the alternative H1 evaluated at the values derived under H0 and for
# these values we should evaluate the Hessian, the variability matrix (denoted
# by J) and Godambe matrix.

lav_pml_test_plrt <- function(fit_obj_h0, fit_obj_h1) {
  # sanity check, perhaps we misordered H0 and H1 in the function call??
  if (fit_obj_h1@test[[1]]$df > fit_obj_h0@test[[1]]$df) {
    tmp <- fit_obj_h0
    fit_obj_h0 <- fit_obj_h1
    fit_obj_h1 <- tmp
  }

  plrt <- 2 * (fit_obj_h1@optim$logl - fit_obj_h0@optim$logl)

  # create a new object 'objH1_h0': the object 'H1', but where
  # the parameter values are from H0
  obj_h1_h0 <- lav_test_diff_m10(m1 = fit_obj_h1, m0 = fit_obj_h0, test = FALSE)

  # EqMat # YR: from 0.6-2, use lav_test_diff_a() (again)
  #             this should allow us to test models that are
  #             nested in the covariance matrix sense, but not
  #             in the parameter (table) sense
  eq_mat <- lav_test_diff_a(m1 = fit_obj_h1, m0 = fit_obj_h0)
  # same basis as used inside lav_test_diff_a (also covers ceq.simple.only
  # and equality constraints coexisting with inequality constraints/bounds)
  eq_basis_h1 <- lav_con_eq_basis(obj_h1_h0@Model)
  if (!is.null(eq_basis_h1)) {
    eq_mat <- eq_mat %*% t(eq_basis_h1)
  }
  # if (equal_constr == TRUE) {
  #    EqMat <- fit_objH0@Model@ceq.JAC
  # } else {
  #    PT0 <- fit_objH0@ParTable
  #    PT1 <- fit_objH1@ParTable
  #    h0.par.idx <- which(PT1$free > 0 & !(PT0$free > 0))
  #    tmp.ind <- PT1$free[ h0.par.idx ]
  #
  #    no.par0 <- length(tmp.ind)
  #    tmp.ind2 <- cbind(1:no.par0, tmp.ind ) # matrix indices
  #    EqMat <- matrix(0, nrow=no.par0, ncol=fit_objH1@Model@nx.free)
  #    EqMat[tmp.ind2] <- 1
  # }

  # DEBUG YR -- eliminate the constraints also present in H1
  #          -- if we do this, there is no need to use MASS::ginv later
  # JAC0 <- fit_objH0@Model@ceq.JAC
  # JAC1 <- fit_objH1@Model@ceq.JAC
  # unique.idx <- which(apply(JAC0, 1, function(x) {
  #                    !any(apply(JAC1, 1, function(y) { all(x == y) })) }))
  # if(length(unique.idx) > 0L) {
  #    EqMat <- EqMat[unique.idx,,drop = FALSE]
  # }

  # Observed information (= for PML, this is Hessian / N)
  hes_theta0 <- lavTech(obj_h1_h0, "information.observed")

  # handle possible constraints in H1 (and therefore also in objH1_h0)
  inv_hes_theta0 <-
    lav_model_info_augment_invert(
      lavmodel = obj_h1_h0@Model,
      information = hes_theta0,
      inverted = TRUE
    )

  # the estimated variability matrix is given (=unit information first order)
  j_theta0 <- lavTech(obj_h1_h0, "first.order")

  # the Inverse of the G matrix
  inv_g <- inv_hes_theta0 %*% j_theta0 %*% inv_hes_theta0

  minv_gt_m <- eq_mat %*% inv_g %*% t(eq_mat)
  minv_ht_m <- eq_mat %*% inv_hes_theta0 %*% t(eq_mat)
  # Inv_MinvHtM <- solve(MinvHtM)
  inv_minv_ht_m <- MASS::ginv(minv_ht_m)
  tmp_prod <- minv_gt_m %*% inv_minv_ht_m
  tmp_prod2 <- tmp_prod %*% tmp_prod
  sum_eig <- sum(diag(tmp_prod))
  sum_eigsq <- sum(diag(tmp_prod2))

  fsma_plrt <- (sum_eig / sum_eigsq) * plrt
  adj_df <- (sum_eig * sum_eig) / sum_eigsq
  pvalue <- 1 - pchisq(fsma_plrt, df = adj_df)

  list(FSMA.PLRT = fsma_plrt, adj.df = adj_df, pvalue = pvalue)
}


# for testing: this is the 'original' (using m.el.idx and x.el.idx)
lav_pml_test_plrt2 <- function(fit_obj_h0, fit_obj_h1) {
  if (fit_obj_h1@test[[1]]$df > fit_obj_h0@test[[1]]$df) {
    tmp <- fit_obj_h0
    fit_obj_h0 <- fit_obj_h1
    fit_obj_h1 <- tmp
  }

  if (fit_obj_h0@Model@eq.constraints) {
    equal_constr <- TRUE
  } else {
    equal_constr <- FALSE
  }

  nsize <- fit_obj_h0@SampleStats@ntotal
  plrt <- 2 * nsize * (fit_obj_h0@optim$fx - fit_obj_h1@optim$fx)
  npar <- fit_obj_h1@optim$npar
  my_m_el_idx2 <- fit_obj_h1@Model@m.free.idx
  my_x_el_idx2 <- fit_obj_h1@Model@x.free.idx
  my_m_el_idx <- my_m_el_idx2
  my_x_el_idx <- my_x_el_idx2

  # my_m_el_idx2 <- fit_objH1@Model@m.free.idx
  # my_m_el_idx2 gives the POSITION index of the free parameters within each
  # parameter matrix under H1 model.
  # The index numbering restarts from 1 when we move to a new parameter matrix.
  # Within each matrix the index numbering "moves" columnwise.

  # my_x_el_idx2 <- fit_objH1@Model@x.free.idx
  # my_x_el_idx2 ENUMERATES the free parameters within each parameter matrix.
  # The numbering continues as we move from one parameter matrix to the
  # next one.

  # In the case of the symmetric matrices, Theta and Psi,in some functions below
  # we need to give as input my_m_el_idx2 and my_x_el_idx2 after
  # we have eliminated the information about the redundant parameters
  # (those placed above the main diagonal).
  # That's why I do the following:

  # my_m_el_idx <- my_m_el_idx2
  # my_x_el_idx <- my_x_el_idx2
  # Psi, the variance - covariance matrix of factors
  # if( length(my_x_el_idx2[[3]])!=0 & any(table(my_x_el_idx2[[3]])>1)) {
  #  nfac <- ncol(fit_objH1@Model@GLIST$lambda) #number of factors
  #  tmp  <- matrix(c(1:(nfac^2)), nrow= nfac, ncol= nfac )
  #  tmp_keep <- tmp[lower.tri(tmp, diag=TRUE)]
  #  my_m_el_idx[[3]] <- my_m_el_idx[[3]][my_m_el_idx[[3]] %in% tmp_keep]
  #  my_x_el_idx[[3]] <- unique( my_x_el_idx2[[3]] )
  # }

  # for Theta, the variance-covariance matrix of measurement errors
  # if( length(my_x_el_idx2[[2]])!=0 & any(table(my_x_el_idx2[[2]])>1)) {
  #  nvar <- fit_objH1@Model@nvar #number of indicators
  #  tmp  <- matrix(c(1:(nvar^2)), nrow= nvar, ncol= nvar )
  #  tmp_keep <- tmp[lower.tri(tmp, diag=TRUE)]
  #  my_m_el_idx[[2]] <- my_m_el_idx[[2]][my_m_el_idx[[2]] %in% tmp_keep]
  #  my_x_el_idx[[2]] <- unique( my_x_el_idx2[[2]] )
  # }

  # below the commands to find the row-column indices of the Hessian that
  # correspond to the parameters to be tested equal to 0
  # tmp.ind contains these indices
  # MY.m.el.idx2.H0 <- fit_objH0@Model@m.free.idx
  # tmp.ind <- c()
  # for(i in 1:6) {
  #   tmp.ind <- c(tmp.ind ,
  #               my_x_el_idx2[[i]] [!(my_m_el_idx2[[i]]  %in%
  #                                    MY.m.el.idx2.H0[[i]] )  ]  )
  # }
  # next line added by YR
  # tmp.ind <- unique(tmp.ind)

  # YR: use partable to find which parameters are restricted in H0
  #     (this should work in multiple groups too)
  # h0.par.idx <- which(   PT.H1.extended$free[PT.H1.extended$user < 2] > 0  &
  #                     !(PT.H0.extended$free[PT.H0.extended$user < 2] > 0)   )
  # tmp.ind <- PT.H1.extended$free[ h0.par.idx ]
  # print(tmp.ind)
  if (length(my_x_el_idx2[[3]]) != 0 && any(table(my_x_el_idx2[[3]]) > 1)) {
    nfac <- ncol(fit_obj_h1@Model@GLIST$lambda)
    tmp <- matrix(c(1:(nfac * nfac)), nrow = nfac, ncol = nfac)
    tmp_keep <- tmp[lower.tri(tmp, diag = TRUE)]
    my_m_el_idx[[3]] <- my_m_el_idx[[3]][my_m_el_idx[[3]] %in% tmp_keep]
    my_x_el_idx[[3]] <- unique(my_x_el_idx2[[3]])
  }

  if (length(my_x_el_idx2[[2]]) != 0 && any(table(my_x_el_idx2[[2]]) > 1)) {
    nvar <- fit_obj_h1@Model@nvar
    tmp <- matrix(c(1:(nvar * nvar)), nrow = nvar, ncol = nvar)
    tmp_keep <- tmp[lower.tri(tmp, diag = TRUE)]
    my_m_el_idx[[2]] <- my_m_el_idx[[2]][my_m_el_idx[[2]] %in% tmp_keep]
    my_x_el_idx[[2]] <- unique(my_x_el_idx2[[2]])
  }
  my_m_el_idx2_h0 <- fit_obj_h0@Model@m.free.idx

  tmp_ind <- c()
  for (i in 1:6) {
    tmp_ind <- c(tmp_ind, my_x_el_idx2[[i]][!(my_m_el_idx2[[i]] %in%
      my_m_el_idx2_h0[[i]])])
  }
  tmp_ind <- unique(tmp_ind)

  # if the models are nested because of equality constraints among the
  # parameters, we need to construct the matrix of derivatives of function
  # g(theta) with respect to theta where g(theta) is the function that
  # represents the equality constraints. g(theta) is
  # an rx1 vector where r are the equality constraints. In the null hypothesis
  # we test H0: g(theta)=0. The matrix of derivatives is of dimension:
  # nrows= number of free non-redundant parameters under H0, namely
  # NparH0 <- fit_objH0[[1]]@optim$npar , and ncols= number of free
  # non-redundant parameters under H1, namely
  #    NparH1 <- fit_objH0[[1]]@optim$npar.
  # The matrix of derivatives of g(theta) is composed of 0's, 1's, -1's, and in
  # the rows that refer to odd number of parameters that are equal there is
  # one -2.
  # The 1's, -1's (and possibly -2) are the contrast coefficients of the
  # parameters.
  # The sum of the rows should be equal to 0.
  # if(equal_constr==TRUE) {
  #    EqMat <- fit_objH0@Model@ceq.JAC
  # } else {
  #   no.par0 <- length(tmp.ind)
  #   tmp.ind2 <- cbind(1:no.par0, tmp.ind)
  #   EqMat <- matrix(0, nrow = no.par0, ncol = Npar)
  # EqMat[tmp.ind2] <- 1
  # }

  if (equal_constr == TRUE) {
    eq_mat <- fit_obj_h0@Model@ceq.JAC
  } else {
    no_par0 <- length(tmp_ind)
    tmp_ind2 <- cbind(1:no_par0, tmp_ind)
    eq_mat <- matrix(0, nrow = no_par0, ncol = npar)
    eq_mat[tmp_ind2] <- 1
  }

  obj <- fit_obj_h0

  # Compute the sum of the eigenvalues and the sum of the squared eigenvalues
  # so that the adjustment to PLRT can be applied.
  # Here a couple of functions (e.g. lav_pml_inspect_hessian) which are
  # modifications of lavaan functions (e.g. getHessian) are needed. These are
  # defined in the end of the file.

  # the quantity below follows the same logic as getHessian of lavaan 0.5-18
  # and it actually gives N*Hessian. That's why the command following the
  # command below.
  # NHes.theta0 <- lav_pml_inspect_hessian (object = obj@Model,
  #                           samplestats = obj@SampleStats ,
  #                           x = obj@Data@X ,
  #                           estimator = "PML",
  #                           lavcache = obj@Cache,
  #                           my_m_el_idx = my_m_el_idx,
  #                           my_x_el_idx = my_x_el_idx,
  #                           my_m_el_idx2 = my_m_el_idx2,
  #                              # input for lav_pml_object_x2glist
  #                           my_x_el_idx2 = my_x_el_idx2,
  #                              # input for lav_pml_object_x2glist
  #                           Npar = Npar,
  #                           equal_constr=equal_constr)
  nhes_theta0 <- lav_pml_inspect_hessian(
    object = obj@Model, samplestats = obj@SampleStats,
    x = obj@Data@X, estimator = "PML", lavcache = obj@Cache,
    my_m_el_idx = my_m_el_idx, my_x_el_idx = my_x_el_idx,
    my_m_el_idx2 = my_m_el_idx2, my_x_el_idx2 = my_x_el_idx2,
    npar = npar, equal_constr = equal_constr
  )
  hes_theta0 <- nhes_theta0 / nsize
  # Inv.Hes.theta0 <- solve(Hes.theta0)
  inv_hes_theta0 <- MASS::ginv(hes_theta0)

  nj_theta0 <- lav_pml_object_information_firstorder(
    object = obj, my_m_el_idx = my_m_el_idx,
    my_x_el_idx = my_x_el_idx, equal_constr = equal_constr
  )
  j_theta0 <- nj_theta0 / (nsize * nsize)


  inv_g <- inv_hes_theta0 %*% j_theta0 %*% inv_hes_theta0
  minv_gt_m <- eq_mat %*% inv_g %*% t(eq_mat)
  minv_ht_m <- eq_mat %*% inv_hes_theta0 %*% t(eq_mat)
  # Inv_MinvHtM <- solve(MinvHtM)    #!!! change names
  inv_minv_ht_m <- MASS::ginv(minv_ht_m)
  tmp_prod <- minv_gt_m %*% inv_minv_ht_m # !!! change names
  tmp_prod2 <- tmp_prod %*% tmp_prod
  sum_eig <- sum(diag(tmp_prod))
  sum_eigsq <- sum(diag(tmp_prod2))


  fsma_plrt <- (sum_eig / sum_eigsq) * plrt
  adj_df <- (sum_eig * sum_eig) / sum_eigsq
  pvalue <- 1 - pchisq(fsma_plrt, df = adj_df)
  list(FSMA.PLRT = fsma_plrt, adj.df = adj_df, pvalue = pvalue)
}




###############################################################################
# auxiliary functions used above, they are all copy from the corresponding
# functions of lavaan where parts no needed were deleted and some parts were
# modified.
# I mark the modifications with comments.


# library(lavaan)

# To run an example for the functions below the following input is needed.
# obj <- fit.objH0[[i]]
# object <- obj@Model
# samplestats = obj@SampleStats
# X = obj@Data@X
# estimator = "PML"
# lavcache = obj@Cache
# my_m_el_idx = my_m_el_idx
# my_x_el_idx = my_x_el_idx
# my_m_el_idx2 = my_m_el_idx2 # input for lav_pml_object_x2glist
# my_x_el_idx2 = my_x_el_idx2 # input for lav_pml_object_x2glist
# Npar = Npar
# equal_constr =TRUE

lav_pml_inspect_hessian <- function(object, samplestats, x,
                         estimator = "PML", lavcache,
                         my_m_el_idx, my_x_el_idx,
                         my_m_el_idx2, my_x_el_idx2,
                                 # input for lav_pml_object_x2glist
                         npar, # Npar is the number of parameters under H1
                         equal_constr) { # takes TRUE/ FALSE
  if (equal_constr) { # !!! added line
  }
  hessian <- matrix(0, npar, npar) #

  # !!!! MYfunction below
  x_1 <- lav_pml_inspect_parameters(
    object = object,
    glist = NULL, n = npar, # N the number of parameters to consider
    my_m_el_idx = my_m_el_idx,
    my_x_el_idx = my_x_el_idx
  )

  for (j in 1:npar) {
    h_j <- 1e-05
    x_left <- x_left2 <- x_right <- x_right2 <- x_1
    x_left[j] <- x_1[j] - h_j
    x_left2[j] <- x_1[j] - 2 * h_j
    x_right[j] <- x_1[j] + h_j
    x_right2[j] <- x_1[j] + 2 * h_j
    # !!!! MYfunction below : lav_pml_inspect_grad and
    #                         lav_pml_object_x2glist
    g_left <- lav_pml_inspect_grad(
      object = object,
      glist = lav_pml_object_x2glist(
        object = object, x = x_left,
        my_m_el_idx = my_m_el_idx2,
        my_x_el_idx = my_x_el_idx2
      ),
      samplestats = samplestats, x = x,
      lavcache = lavcache, estimator = "PML",
      my_m_el_idx = my_m_el_idx,
      my_x_el_idx = my_x_el_idx,
      equal_constr = equal_constr
    )

    g_left2 <- lav_pml_inspect_grad(
      object = object,
      glist = lav_pml_object_x2glist(
        object = object, x = x_left2,
        my_m_el_idx = my_m_el_idx2,
        my_x_el_idx = my_x_el_idx2
      ),
      samplestats = samplestats, x = x,
      lavcache = lavcache, estimator = "PML",
      my_m_el_idx = my_m_el_idx,
      my_x_el_idx = my_x_el_idx,
      equal_constr = equal_constr
    )

    g_right <- lav_pml_inspect_grad(
      object = object,
      glist = lav_pml_object_x2glist(
        object = object, x = x_right,
        my_m_el_idx = my_m_el_idx2,
        my_x_el_idx = my_x_el_idx2
      ),
      samplestats = samplestats, x = x,
      lavcache = lavcache, estimator = "PML",
      my_m_el_idx = my_m_el_idx,
      my_x_el_idx = my_x_el_idx,
      equal_constr = equal_constr
    )

    g_right2 <- lav_pml_inspect_grad(
      object = object,
      glist = lav_pml_object_x2glist(
        object = object, x = x_right2,
        my_m_el_idx = my_m_el_idx2,
        my_x_el_idx = my_x_el_idx2
      ),
      samplestats = samplestats, x = x,
      lavcache = lavcache, estimator = "PML",
      my_m_el_idx = my_m_el_idx,
      my_x_el_idx = my_x_el_idx,
      equal_constr = equal_constr
    )

    hessian[, j] <- (g_left2 - 8 * g_left + 8 * g_right - g_right2) / (12 * h_j)
  }
  hessian <- (hessian + t(hessian)) / 2
  # (-1) * Hessian
  hessian
}
#############################################################################




##################################  lav_pml_inspect_parameters
# different input arguments: my_m_el_idx, my_x_el_idx
lav_pml_inspect_parameters <-
  function(object, glist = NULL, n, # N the number of parameters to consider
           my_m_el_idx, my_x_el_idx) {
  if (is.null(glist)) {
    glist <- object@GLIST
  }

  x <- numeric(n)

  for (mm in seq_along(object@GLIST)) { # mm<-1
    m_idx <- my_m_el_idx[[mm]] # !!!!! different here and below
    x_idx <- my_x_el_idx[[mm]]
    x[x_idx] <- glist[[mm]][m_idx]
  }
  x
}
#############################################################################




#############################  lav_pml_inspect_grad
# the difference are the input arguments my_m_el_idx, my_x_el_idx
# used  in  lavaan:::lav_model_delta
lav_pml_inspect_grad <-
  function(object, glist, samplestats = NULL, x = NULL,
           lavcache = NULL, estimator = "PML",
           my_m_el_idx, my_x_el_idx, equal_constr) {
  if (equal_constr) { # added line
  }
  num_idx <- object@num.idx
  th_idx <- object@th.idx
  if (is.null(glist)) {
    glist <- object@GLIST
  }
  sigma_hat <-
    lav_model_sigma(object, glist = glist, extra = (estimator == "ML"))
  mu_hat <- lav_model_mu(object, glist = glist)
  th <- lav_model_th(object, glist = glist)
  g <- 1
  d1 <- lav_pml_dploglik_dimplied(
    sigma_hat = sigma_hat[[g]], mu_hat = mu_hat[[g]],
    th = th[[g]], th_idx = th_idx[[g]], num_idx = num_idx[[g]],
    x = x[[g]], lavcache = lavcache[[g]]
  )

  # !?  if(equal_constr) { #delete the following three commented lines, wrong
  #     Delta <- lavaan:::lav_model_delta (lavmodel= object, GLIST. = GLIST)
  #  } else {
  delta <- lav_model_delta(
    lavmodel = object, glist = glist,
    m_el_idx = my_m_el_idx,
    x_el_idx = my_x_el_idx
  )
  # }

  # !!! that was before: as.numeric(t(d1) %*% Delta[[g]])/samplestats@nobs[[g]]
  as.numeric(t(d1) %*% delta[[g]])
           # !!! modified to follow current computeGradient() function of lavaan
  # !!! which gives minus the gradient of PL-loglik
  }

###############################################################################


##################################  lav_pml_object_x2glist
# difference in input arguments my_m_el_idx, my_x_el_idx

lav_pml_object_x2glist <- function(object, x = NULL, my_m_el_idx, my_x_el_idx) {
  glist <- object@GLIST
  for (mm in seq_along(glist)) {
    m_el_idx <- my_m_el_idx[[mm]]
    x_el_idx <- my_x_el_idx[[mm]]
    glist[[mm]][m_el_idx] <- x[x_el_idx]
  }
  glist
}
############################################################################


##### lav_pml_object_information_firstorder function
# difference from corresponding of lavaan: I use lav_pml_model_vcov_firstorder
lav_pml_object_information_firstorder <-    # nolint
  function(object, my_m_el_idx, my_x_el_idx, equal_constr) {
  nacov <- lav_pml_model_vcov_firstorder(
    lavmodel = object@Model,
    lavsamplestats = object@SampleStats,
    lavdata = object@Data,
    estimator = "PML",
    my_m_el_idx = my_m_el_idx,
    my_x_el_idx = my_x_el_idx,
    equal_constr = equal_constr
  )
  if (equal_constr) { # added lines
  }
  b0 <- attr(nacov, "B0")
  # !!!! Note below that I don't multiply  with nsize
  # !!! so what I get is J matrix divided by n
  # if (object@Options$estimator == "PML") {
  #    B0 <- B0 * object@SampleStats@ntotal
  # }
  # !!!!!!!!!!!!!!!!!!! added the following lines so that the output of
  # !!! lav_pml_object_information_firstorder is in line with that of
  #                    lavaan 0.5-18 getVariability
  # !! what's the purpose of the following lines?
  if (object@Options$estimator == "PML") {
    b0 <- b0 * object@SampleStats@ntotal
  }

  b0
  }

##############################################################################
# example
# obj <- fit.objH0[[i]]
# object <- obj@Model
# samplestats = obj@SampleStats
# X = obj@Data@X
# estimator = "PML"
# lavcache = obj@Cache
# my_m_el_idx = my_m_el_idx
# my_x_el_idx = my_x_el_idx
# my_m_el_idx2 = my_m_el_idx2 # input for lav_pml_object_x2glist
# my_x_el_idx2 = my_x_el_idx2 # input for lav_pml_object_x2glist
# Npar = Npar
# equal_constr =TRUE


lav_pml_model_vcov_firstorder <- function(lavmodel, lavsamplestats = NULL,
                                lavdata = NULL, lavcache = NULL,
                                estimator = "PML",
                                my_m_el_idx, my_x_el_idx,
                                equal_constr) { # equal_constr takes TRUE/FALSE
  if (equal_constr) { # added lines
  }
  b0_group <- vector("list",
                     lavsamplestats@ngroups) # in my case list of length 1

  # !?   if (equal_constr) {
  #      ###the following three lines are commented because they are wrong
  #      Delta <- lavaan:::lav_model_delta(lavmodel, GLIST. = NULL)
  #   } else {
  delta <- lav_model_delta(lavmodel,
    glist = NULL,
    m_el_idx = my_m_el_idx, # !!!!! different here and below
    x_el_idx = my_x_el_idx
  )
  #  }
  sigma_hat <- lav_model_sigma(lavmodel)
  mu_hat <- lav_model_mu(lavmodel)
  th <- lav_model_th(lavmodel)
  g <- 1

  sc <- lav_pml_dploglik_dimplied(
    sigma_hat = sigma_hat[[g]], th = th[[g]],
    mu_hat = mu_hat[[g]], th_idx = lavmodel@th.idx[[g]],
    num_idx = lavmodel@num.idx[[g]],
    x = lavdata@X[[g]], lavcache = lavcache,
    scores = TRUE, negative = FALSE
  )
  group_sc <- sc %*% delta[[g]]
  b0_group[[g]] <- lav_mat_crossprod(group_sc)
  # !!!! B0.group[[g]] <- B0.group[[g]]/lavsamplestats@ntotal
  #  !!! skip so that the result is in line with the 0.5-18 version of lavaan

  b0 <- b0_group[[1]]

  m_e <- b0

  eigvals <- eigen(m_e, symmetric = TRUE, only.values = TRUE)$values
  if (any(eigvals < -1 * .Machine$double.eps^(3 / 4))) {
    lav_msg_warn(gettext(
      "matrix based on first order outer product of the derivatives is not
      positive definite; the standard errors may not be trustworthy"))
  }
  nvar_cov <- MASS::ginv(m_e)

  attr(nvar_cov, "B0") <- b0
  attr(nvar_cov, "B0.group") <- b0_group
  nvar_cov
}

# - 0.6-13: fix multiple-group UG^2 bug (reported by Gronneberg, Foldnes and
#           Moss) when Satterthwaite = TRUE, ngroups > 1, and eq constraints.
#           Use ug2.old.approach = TRUE to get the old result
# - 0.7-1:  the traces are computed by the streamed worker
#           lav_test_ug_trace_stream() (lav_test_utils.R) unless the U or
#           UGamma matrices themselves are needed; shared helpers for the
#           E.inv-recompute check, failure entries and labels

lav_test_sb <- function(lavobject = NULL,
                                     lavsamplestats = NULL,
                                     lavmodel = NULL,
                                     lavimplied = NULL,
                                     lavoptions = NULL,
                                     lavdata = NULL,
                                     test_unscaled = NULL,
                                     e_inv = NULL,
                                     delta = NULL,
                                     wls_v = NULL,
                                     m_gamma = NULL,
                                     test = "satorra.bentler",
                                     method = "original",
                                     ug2_old_approach = FALSE,
                                     return_u = FALSE,
                                     return_ugamma = FALSE,
                                     gamma_full = NULL) {
  test_1 <- list()

  # if a full (cross-group) Gamma is supplied, we must use the 'original'
  # method (the only path that consumes a single assembled Gamma matrix)
  if (!is.null(gamma_full)) {
    method <- "original"
  }

  if (!is.null(lavobject)) {
    lavsamplestats <- lavobject@SampleStats
    lavmodel <- lavobject@Model
    lavoptions <- lavobject@Options
    lavimplied <- lavobject@implied
    lavdata <- lavobject@Data
    test_1$standard <- lavobject@test[[1]]
  } else {
    test_1$standard <- test_unscaled
  }
  npar <- lavmodel@nx.free

  # ug2.old.approach
  if (missing(ug2_old_approach)) {
    if (!is.null(lavoptions$ug2.old.approach)) {
      ug2_old_approach <- lavoptions$ug2.old.approach
    } else {
      ug2_old_approach <- FALSE
    }
  }

  # E.inv ok?
  lavoptions <- lav_test_change_options(lavoptions)
  e_inv_recompute <- attr(lavoptions, "recompute")
  attr(lavoptions, "recompute") <- NULL
  if (!is.null(e_inv) && !is.null(wls_v) && !is.null(delta)) {
    e_inv_recompute <- FALSE # user-provided
  }

  # check test
  if (!all(test %in% c(
    "satorra.bentler",
    "scaled.shifted",
    "mean.var.adjusted",
    "mean.var.adjusted.corrected",
    "scaled.shifted.corrected"
  ))) {
    lav_msg_warn(gettext(
      "test must be one of `satorra.bentler', `scaled.shifted',
      `mean.var.adjusted', `mean.var.adjusted.corrected' or
      `scaled.shifted.corrected'; will use `satorra.bentler' only"))
    test <- "satorra.bentler"
  }

  # corrected trace (Hayakawa 2018) tests need the U matrix and casewise data
  corrected_trace <- any(test %in% c(
    "mean.var.adjusted.corrected",
    "scaled.shifted.corrected"
  ))
  if (corrected_trace) {
    lav_test_hayakawa_check(
      lavoptions = lavoptions, lavdata = lavdata, lavmodel = lavmodel
    )
  }

  if (return_u || corrected_trace) {
    method <- "original"
  }

  # check method
  if (!all(method %in% c("original", "orthogonal.complement", "ABA"))) {
    lav_msg_warn(gettext("method must be one of `original', `ABA',
                         `orthogonal.complement'; will use `ABA'"))
    method <- "original"
  }

  # do we have E.inv, Delta, WLS.V?
  if (npar > 0L &&
    (is.null(e_inv) || is.null(delta) || is.null(wls_v) || e_inv_recompute)) {
    if (lavoptions$information.expected.mplus && lavoptions$estimator == "ML") {
      m_e <- lav_model_info_expected_mlm(
        lavmodel = lavmodel,
        augmented = FALSE, inverted = FALSE,
        lavsamplestats = lavsamplestats, extra = TRUE
      )
    } else {
      m_e <- lav_model_info(
        lavmodel = lavmodel,
        lavimplied = lavimplied,
        lavsamplestats = lavsamplestats, lavdata = lavdata,
        lavoptions = lavoptions, extra = TRUE
      )
    }
    e_inv <- try(lav_model_info_augment_invert(lavmodel,
      information = m_e, inverted = TRUE
    ), silent = TRUE)
    if (inherits(e_inv, "try-error")) {
      if (return_ugamma) {
        lav_msg_warn(gettext(
          "could not invert information matrix needed for UGamma"))
        return(NULL)
      } else if (return_u) {
        lav_msg_warn(gettext(
          "could not invert information matrix needed for UfromUGamma"))
        return(NULL)
      } else {
        na_out <- lav_test_scaled_na(test_1$standard,
          test1 = test[1],
          ngroups = lavdata@ngroups
        )
        test_1$standard <- na_out$standard
        test_1[[test[1]]] <- na_out$failed
        lav_msg_warn(gettext("could not invert information matrix needed for
                             robust test statistic"))
        return(test_1)
      }
    }
    delta <- attr(m_e, "Delta")
    wls_v <- attr(m_e, "WLS.V")
  }

  # catch df == 0
  if ((test_1$standard$df == 0L || test_1$standard$df < 0) &&
    !return_u && !return_ugamma) {
    na_out <- lav_test_scaled_na(test_1$standard,
      test1 = test[1],
      na_standard = FALSE, shift = FALSE
    )
    test_1[[test[1]]] <- na_out$failed
    return(test_1)
  }

  # Gamma
  if (is.null(m_gamma)) {
    # stored NACOV, else recompute with the fit-time settings
    if (!is.null(lavobject)) {
      m_gamma <- lav_gamma_used(lavobject)
    } else {
      m_gamma <- lav_gamma_used(
        lavdata = lavdata,
        lavoptions = lavoptions,
        lavsamplestats = lavsamplestats,
        lavh1 = NULL,
        lavimplied = NULL
      )
    }
  }

  if (lavoptions$information.expected.mplus && lavmodel@categorical) {
    for (g in 1:lavsamplestats@ngroups) {
      ng <- lavsamplestats@nobs[[g]]
      m_gamma[[g]] <- m_gamma[[g]] / ng * (ng - 1L)
    }
  }

  # ngroups
  ngroups <- lavsamplestats@ngroups

  # mean and variance adjusted?
  satterthwaite <- FALSE
  if (any(test %in% c(
    "mean.var.adjusted", "scaled.shifted",
    "mean.var.adjusted.corrected", "scaled.shifted.corrected"
  ))) {
    satterthwaite <- TRUE
  }

  if (npar == 0) {
    # catch npar == 0 (eg baseline model if correlation structure)
    u_all <- as.numeric(NA)
    # Gamma_g / fg: the Cov(sqrt(ntotal) s_g) stacking convention (see the
    # SCALING CONVENTIONS note in lav_samplestats_gamma.R)
    gamma_f <- lav_gamma_rescale_ntotal(m_gamma,
      nobs = lavsamplestats@nobs, ntotal = lavsamplestats@ntotal
    )
    gamma_all <- lav_mat_bdiag(gamma_f)
    ug <- gamma_all
    trace_ugamma <- sum(diag(gamma_all))
    trace_ugamma2 <- sum(ug * t(ug))
    out <- list(
      trace.UGamma = trace_ugamma, trace.UGamma2 = trace_ugamma2,
      UGamma = ug, UfromUGamma = u_all
    )
  } else if (method == "original") {
    out <- lav_test_sb_trace_original(
      m_gamma = m_gamma,
      delta = delta, wls_v = wls_v, e_inv = e_inv,
      ngroups = ngroups, nobs = lavsamplestats@nobs,
      ntotal = lavsamplestats@ntotal, return_u = return_u || corrected_trace,
      return_ugamma = return_ugamma,
      ug2_old_approach = ug2_old_approach,
      satterthwaite = satterthwaite,
      gamma_full = gamma_full
    )
  } else if (method == "orthogonal.complement") {
    out <- lav_test_sb_trace_complement(
      m_gamma = m_gamma,
      delta = delta, wls_v = wls_v, lavmodel = lavmodel,
      ngroups = ngroups, nobs = lavsamplestats@nobs,
      ntotal = lavsamplestats@ntotal,
      return_ugamma = return_ugamma,
      ug2_old_approach = ug2_old_approach,
      satterthwaite = satterthwaite
    )
  } else if (method == "ABA") {
    out <- lav_test_sb_trace_aba(
      m_gamma = m_gamma,
      m_delta = delta, wls_v = wls_v, e_inv = e_inv,
      ngroups = ngroups, nobs = lavsamplestats@nobs,
      ntotal = lavsamplestats@ntotal,
      return_ugamma = return_ugamma,
      ug2_old_approach = ug2_old_approach,
      satterthwaite = satterthwaite
    )
  } else {
    lav_msg_stop(gettextf("method `%s' not supported", method))
  }
  trace_ugamma <- out$trace.UGamma
  trace_ugamma2 <- out$trace.UGamma2

  # corrected traces (Hayakawa 2018) from the casewise moment vectors
  if (corrected_trace) {
    u_mat <- out$UfromUGamma
    if (is.list(u_mat)) {
      u_mat <- u_mat[[1]] # single group only (checked above)
    }
    if (is.matrix(u_mat)) {
      trace_c <- lav_test_hayakawa_trace2(
        u = u_mat, m_y = lavdata@X[[1]],
        meanstructure = lavmodel@meanstructure
      )
      trace_ugamma_c <- trace_c$trace.UGamma
      trace_ugamma2_c <- trace_c$trace.UGamma2
    } else {
      trace_ugamma_c <- trace_ugamma2_c <- as.numeric(NA)
    }
    if (!is.finite(trace_ugamma2_c) || trace_ugamma2_c <= 0 ||
      !is.finite(trace_ugamma_c) || trace_ugamma_c <= 0) {
      lav_msg_warn(gettext(
        "corrected trace estimates are not positive; the corrected
        adjusted test statistics will be NA."))
      trace_ugamma_c <- trace_ugamma2_c <- as.numeric(NA)
    }
  }

  # Mplus-flavored labels?
  mplus_flag <- lavoptions$information.expected.mplus

  if ("satorra.bentler" %in% test) {
    # same df
    df_scaled <- test_1$standard$df

    # scaling factor
    scaling_factor <- trace_ugamma / df_scaled
    if (scaling_factor < 0) scaling_factor <- as.numeric(NA)

    # scaled test statistic per group
    stat_group <- test_1$standard$stat.group / scaling_factor

    # scaled test statistic global
    stat <- sum(stat_group)

    test_1$satorra.bentler <-
      list(
        test = "satorra.bentler",
        stat = stat,
        stat.group = stat_group,
        df = df_scaled,
        pvalue = 1 - pchisq(stat, df_scaled),
        trace.UGamma = trace_ugamma,
        scaling.factor = scaling_factor,
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test,
        label = lav_test_scaled_label(
          "satorra.bentler",
          lavoptions$estimator, mplus_flag
        )
      )
  }

  if ("mean.var.adjusted" %in% test) {
    if (mplus_flag) {
      df_scaled <- floor(trace_ugamma^2 / trace_ugamma2 + 0.5)
    } else {
      # more precise, fractional df
      df_scaled <- trace_ugamma^2 / trace_ugamma2
    }

    # scaling factor
    scaling_factor <- trace_ugamma / df_scaled
    if (scaling_factor < 0) scaling_factor <- as.numeric(NA)

    # scaled test statistic per group
    stat_group <- test_1$standard$stat.group / scaling_factor

    # scaled test statistic global
    if (ug2_old_approach) {
      stat <- sum(stat_group)
    } else {
      stat <- test_1$standard$stat / scaling_factor
    }

    test_1$mean.var.adjusted <-
      list(
        test = "mean.var.adjusted",
        stat = stat,
        stat.group = stat_group,
        df = df_scaled,
        pvalue = 1 - pchisq(stat, df_scaled),
        trace.UGamma = trace_ugamma,
        trace.UGamma2 = trace_ugamma2,
        scaling.factor = scaling_factor,
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test,
        label = lav_test_scaled_label(
          "mean.var.adjusted",
          lavoptions$estimator, mplus_flag
        )
      )
  }

  if ("scaled.shifted" %in% test) {
    # this is the T3 statistic as used by Mplus 6 and higher
    # see 'Simple Second Order Chi-Square Correction' 2010
    # www.statmodel.com

    # same df
    df_scaled <- test_1$standard$df

    # scaling factor
    fg <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal
    a <- sqrt(df_scaled / trace_ugamma2)
    if (isTRUE(a < 0) || is.nan(a)) a <- as.numeric(NA)
    scaling_factor <- 1 / a
    if (isTRUE(scaling_factor < 0)) scaling_factor <- as.numeric(NA)

    if (ug2_old_approach) {
      # scaling factor
      shift_parameter <- fg * (df_scaled - a * trace_ugamma)
      # scaled test statistic per group
      stat_group <- (test_1$standard$stat.group * a + shift_parameter)
      # scaled test statistic global
      stat <- sum(stat_group)
    } else {
      shift_parameter <- df_scaled - a * trace_ugamma
      stat <- test_1$standard$stat * a + shift_parameter
      stat_group <- test_1$standard$stat.group * a + fg * shift_parameter
    }

    test_1$scaled.shifted <-
      list(
        test = "scaled.shifted",
        stat = stat,
        stat.group = stat_group,
        df = df_scaled,
        pvalue = 1 - pchisq(stat, df_scaled),
        trace.UGamma = trace_ugamma,
        trace.UGamma2 = trace_ugamma2,
        scaling.factor = scaling_factor,
        shift.parameter = shift_parameter,
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test,
        label = lav_test_scaled_label(
          "scaled.shifted",
          lavoptions$estimator, mplus_flag
        )
      )
  }

  if ("mean.var.adjusted.corrected" %in% test) {
    # Hayakawa (2018): mean and variance adjusted test with the unbiased
    # (Srivastava 2005; Himeno & Yamada 2014) estimator of tr(UGamma^2)
    df_scaled <- trace_ugamma_c^2 / trace_ugamma2_c

    # scaling factor
    scaling_factor <- trace_ugamma2_c / trace_ugamma_c
    if (isTRUE(scaling_factor < 0)) scaling_factor <- as.numeric(NA)

    # scaled test statistic per group and global
    stat_group <- test_1$standard$stat.group / scaling_factor
    stat <- test_1$standard$stat / scaling_factor

    test_1$mean.var.adjusted.corrected <-
      list(
        test = "mean.var.adjusted.corrected",
        stat = stat,
        stat.group = stat_group,
        df = df_scaled,
        pvalue = 1 - pchisq(stat, df_scaled),
        trace.UGamma = trace_ugamma_c,
        trace.UGamma2 = trace_ugamma2_c,
        trace.UGamma2.naive = trace_ugamma2,
        scaling.factor = scaling_factor,
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test,
        label =
          "mean and variance adjusted correction (corrected trace)"
      )
  }

  if ("scaled.shifted.corrected" %in% test) {
    # scaled and shifted test with the corrected estimator of tr(UGamma^2)
    # (not in Hayakawa 2018, but the same substitution)
    df_scaled <- test_1$standard$df

    fg <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal
    a <- sqrt(df_scaled / trace_ugamma2_c)
    if (isTRUE(a < 0) || is.nan(a)) a <- as.numeric(NA)
    scaling_factor <- 1 / a
    if (isTRUE(scaling_factor < 0)) scaling_factor <- as.numeric(NA)

    shift_parameter <- df_scaled - a * trace_ugamma_c
    stat <- test_1$standard$stat * a + shift_parameter
    stat_group <- test_1$standard$stat.group * a + fg * shift_parameter

    test_1$scaled.shifted.corrected <-
      list(
        test = "scaled.shifted.corrected",
        stat = stat,
        stat.group = stat_group,
        df = df_scaled,
        pvalue = 1 - pchisq(stat, df_scaled),
        trace.UGamma = trace_ugamma_c,
        trace.UGamma2 = trace_ugamma2_c,
        trace.UGamma2.naive = trace_ugamma2,
        scaling.factor = scaling_factor,
        shift.parameter = shift_parameter,
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test,
        label = "simple second-order correction (corrected trace)"
      )
  }

  if (return_ugamma) {
    test_1$UGamma <- out$UGamma
  }

  if (return_u) {
    test_1$UfromUGamma <- out$UfromUGamma
  }

  test_1
}

# using the `classical' formula
# UG = Gamma * [V - V Delta E.inv Delta' V']
#
# if only the traces are needed (no return_u/return_ugamma, no cross-group
# gamma_full), they are computed by the streamed worker
# lav_test_ug_trace_stream() without ever forming U or U %*% Gamma
# (this matters in the categorical setting -- and whenever the number of
# variables grows -- where pstar is large)
lav_test_sb_trace_original <- function(m_gamma = NULL,
                                                    delta = NULL,
                                                    wls_v = NULL,
                                                    e_inv = NULL,
                                                    ngroups = NULL,
                                                    nobs = NULL,
                                                    ntotal = NULL,
                                                    return_u = FALSE,
                                                    return_ugamma = FALSE,
                                                    ug2_old_approach = FALSE,
                                                    satterthwaite = FALSE,
                                                    gamma_full = NULL) {
  # gamma_full: an already-assembled (full) Gamma matrix for the *stacked*
  # statistics of all groups. When supplied (e.g. by sam() with across-group
  # constraints, where Gamma has nonzero cross-group blocks), it is used as-is
  # instead of block-diagonalizing the per-group m_gamma list. This forces the
  # full (non per-group) computation path.
  fg <- unlist(nobs) / ntotal

  # fast path: only the traces are needed
  if (!return_u && !return_ugamma && is.null(gamma_full)) {
    out <- lav_test_ug_trace_stream(
      m_gamma = m_gamma, delta = delta, wls_v = wls_v, e_inv = e_inv,
      fg = fg, satterthwaite = satterthwaite,
      old_approach = ug2_old_approach
    )
    return(list(
      trace.UGamma = out$trace.UGamma, trace.UGamma2 = out$trace.UGamma2,
      UGamma = as.numeric(NA), UfromUGamma = as.numeric(NA)
    ))
  }

  # explicit path: the U and/or UGamma matrices themselves are needed
  # (lavInspect, Hayakawa corrected traces, FMG eigenvalues, gamma_full)
  if (ug2_old_approach) {
    # this is what we did <0.6-13: everything per group
    ufrom_ugamma <- ug <- vector("list", ngroups)
    trace_ugamma <- trace_ugamma2 <- rep(as.numeric(NA), ngroups)
    for (g in 1:ngroups) {
      gamma_g <- m_gamma[[g]] / fg[g]
      if (is.matrix(wls_v[[g]])) {
        wls_vg <- wls_v[[g]] * fg[g]
      } else {
        wls_vg <- diag(wls_v[[g]]) * fg[g]
      }

      u <- (wls_vg - wls_vg %*% delta[[g]] %*% e_inv %*%
        t(delta[[g]]) %*% wls_vg)
      trace_ugamma[g] <- sum(u * gamma_g)

      if (return_u) {
        ufrom_ugamma[[g]] <- u
      }

      ug <- NULL
      if (satterthwaite || return_ugamma) {
        ug_group <- u %*% gamma_g
        trace_ugamma2[g] <- sum(ug_group * t(ug_group))
        ug[[g]] <- ug_group
      }
    } # g
    # sum over groups
    trace_ugamma <- sum(trace_ugamma)
    trace_ugamma2 <- sum(trace_ugamma2)
    u_all <- ufrom_ugamma # group-specific
  } else {
    stacked <- lav_test_ug_stack(
      m_gamma = m_gamma, delta = delta, wls_v = wls_v,
      nobs = nobs, ntotal = ntotal, gamma_full = gamma_full
    )
    u_all <- stacked$v_all - stacked$v_all %*% stacked$delta_all %*%
      e_inv %*% t(stacked$delta_all) %*% stacked$v_all
    ug <- u_all %*% stacked$gamma_all

    trace_ugamma <- sum(u_all * stacked$gamma_all)
    trace_ugamma2 <- sum(ug * t(ug))
  }

  list(
    trace.UGamma = trace_ugamma, trace.UGamma2 = trace_ugamma2,
    UGamma = ug, UfromUGamma = u_all
  )
}

# using the orthogonal complement of Delta: Delta.c
# UG = [ (Delta.c' W Delta.c)^{-1} (Delta.c' Gamma Delta.c)
lav_test_sb_trace_complement <- function(m_gamma = NULL,
                                                      delta = NULL,
                                                      wls_v = NULL,
                                                      lavmodel = NULL,
                                                      ngroups = NULL,
                                                      nobs = NULL,
                                                      ntotal = NULL,
                                                      return_ugamma = FALSE,
                                                      ug2_old_approach = FALSE,
                                                      satterthwaite = FALSE) {
  # handle equality constraints
  # FIXME: inequality constraints are ignored!
  eq_basis <- lav_con_eq_basis(lavmodel)

  # this is what we did <0.6-13: everything per group
  # does not work when ngroups > 1 + equality constraints
  if (ug2_old_approach) {
    ug <- vector("list", ngroups)
    trace_ugamma <- trace_ugamma2 <- rep(as.numeric(NA), ngroups)
    for (g in 1:ngroups) {
      fg <- nobs[[g]] / ntotal
      gamma_g <- m_gamma[[g]] / fg
      delta_g <- delta[[g]]
      if (is.matrix(wls_v[[g]])) {
        wls_vg <- wls_v[[g]] * fg
      } else {
        wls_vg <- diag(wls_v[[g]]) * fg
      }

      if (!is.null(eq_basis)) {
        delta_g <- delta_g %*% eq_basis
      }

      # orthogonal complement of Delta.g
      delta_c <- lav_mat_ortho_complement(delta_g)

      ### FIXME: compute WLS.W directly, instead of using solve(WLS.V)

      tmp1 <- solve(t(delta_c) %*% solve(wls_vg) %*% delta_c)
      tmp2 <- t(delta_c) %*% gamma_g %*% delta_c

      trace_ugamma[g] <- sum(tmp1 * tmp2)
      ug <- NULL
      if (satterthwaite || return_ugamma) {
        ug_group <- tmp1 %*% tmp2
        trace_ugamma2[g] <- sum(ug_group * t(ug_group))
        ug[[g]] <- ug_group
      }
    }
    # sum over groups
    trace_ugamma <- sum(trace_ugamma)
    trace_ugamma2 <- sum(trace_ugamma2)
  } else {
    trace_ugamma2 <- ug <- as.numeric(NA)
    stacked <- lav_test_ug_stack(
      m_gamma = m_gamma, delta = delta, wls_v = wls_v,
      nobs = nobs, ntotal = ntotal
    )
    delta_all <- stacked$delta_all
    if (!is.null(eq_basis)) {
      delta_all <- delta_all %*% eq_basis
    }

    # orthogonal complement of Delta
    delta_c <- lav_mat_ortho_complement(delta_all)

    tmp1 <- solve(t(delta_c) %*% solve(stacked$v_all) %*% delta_c)
    tmp2 <- t(delta_c) %*% stacked$gamma_all %*% delta_c

    ug <- tmp1 %*% tmp2

    trace_ugamma <- sum(tmp1 * tmp2)
    trace_ugamma2 <- sum(ug * t(ug))
  }

  list(
    trace.UGamma = trace_ugamma, trace.UGamma2 = trace_ugamma2,
    UGamma = ug
  )
}

# using the ABA form
# UG = Gamma %*% [V - V %*% Delta %*% E.inv %*% tDelta %*% V]
#    = Gamma %*% V  - Gamma %*% V %*% Delta %*% E.inv %*% tDelta %*% V
#    = Gamma %*% A1 - Gamma %*% A1 %*% Delta %*% E.inv %*% tDelta %*% A1
# (define AGA1 := A1 %*% Gamma %*% A1)
# Note this is not identical to 'B1', (model-based) first-order information
#
#    = A1.inv %*% A1 %*% Gamma %*% A1 -
#      A1.inv %*% A1 %*% Gamma %*% A1 %*% Delta %*% E.inv %*% tDelta %*% A1
#
#    = A1.inv %*% AGA1 -
#      A1.inv %*% AGA1 %*% Delta %*% E.inv %*% tDelta %*% A1
#
# if only the trace is needed, we can use reduce the rhs (after the minus)
# to AGA1 %*% Delta %*% E.inv %*% tDelta (eliminating A1 and A1.inv)

# we write it like this to highlight the connection with MLR
#
lav_test_sb_trace_aba <- function(m_gamma = NULL,
                                               m_delta = NULL,
                                               wls_v = NULL,
                                               e_inv = NULL,
                                               ngroups = NULL,
                                               nobs = NULL,
                                               ntotal = NULL,
                                               return_ugamma = FALSE,
                                               ug2_old_approach = FALSE,
                                               satterthwaite = FALSE) {
  # per-group computation: either the <0.6-13 approach (also for the
  # UGamma^2 trace), or (default) for the first trace only
  if (ug2_old_approach || !(satterthwaite || return_ugamma)) {
    ug <- vector("list", ngroups)
    trace_ugamma <- trace_ugamma2 <- rep(as.numeric(NA), ngroups)

    for (g in 1:ngroups) {
      fg <- nobs[[g]] / ntotal
      gamma_g <- m_gamma[[g]] / fg
      delta_g <- m_delta[[g]]

      # diagonal wls_v? we check for this since 0.5-17
      diagonal <- FALSE
      if (is.matrix(wls_v[[g]])) {
        aa1 <- wls_v[[g]] * fg
        aga1 <- aa1 %*% gamma_g %*% aa1
      } else {
        diagonal <- TRUE
        a1 <- wls_v[[g]] * fg # numeric vector!
        aga1 <- gamma_g * tcrossprod(a1)
      }

      # note: we have aga1 at the end, to avoid ending up with
      # a transposed matrix (both parts are non-symmetric)
      if (diagonal) {
        ug_group <- t(gamma_g * a1) -
          (delta_g %*% tcrossprod(e_inv, delta_g) %*% aga1)
      } else {
        ug_group <- (gamma_g %*% aa1) -
          (delta_g %*% tcrossprod(e_inv, delta_g) %*% aga1)
      }

      trace_ugamma[g] <- sum(diag(ug_group))
      if (ug2_old_approach && satterthwaite) {
        trace_ugamma2[g] <- sum(ug_group * t(ug_group))
      }
    } # g
    # sum over groups
    trace_ugamma <- sum(trace_ugamma)
    if (ug2_old_approach && satterthwaite) {
      trace_ugamma2 <- sum(trace_ugamma2)
    } else {
      trace_ugamma2 <- as.numeric(NA)
    }
    ug <- as.numeric(NA)
  } else {
    # for trace_ugamma2, we can no longer compute the trace per group
    stacked <- lav_test_ug_stack(
      m_gamma = m_gamma, delta = m_delta, wls_v = wls_v,
      nobs = nobs, ntotal = ntotal
    )
    aga1 <- stacked$v_all %*% stacked$gamma_all %*% stacked$v_all

    ug <- (stacked$gamma_all %*% stacked$v_all) -
      (stacked$delta_all %*% tcrossprod(e_inv, stacked$delta_all) %*% aga1)

    trace_ugamma <- sum(diag(ug))
    trace_ugamma2 <- sum(ug * t(ug))
  }

  if (!return_ugamma) {
    ug <- NULL
  }

  list(
    trace.UGamma = trace_ugamma, trace.UGamma2 = trace_ugamma2,
    UGamma = ug
  )
}

# utility functions needed to compute various (robust) fit measures:
#
# - lav_fit_catml_dwls (for 'robust' RMSEA/CFI if data is categorical)
# - lav_fit_fiml_corrected (correct RMSEA/CFI if data is incomplete)

# compute scaling-factor (c.hat3) for fit.dwls, using fit.catml ingredients
# see:
#     Savalei, V. (2021) Improving Fit Indices In SEM with categorical data.
#     Multivariate Behavioral Research, 56(3), 390-407.
#

# YR Dec 2022: first version
# YR Jan 2023: catml_dwls should check if the input 'correlation' matrix
#              is positive-definite (or not)
# YR Jul 2026: replace check_pd = TRUE/FALSE by nonpd = "na"/"refit"/"smooth"
#              (what to do when an input correlation matrix is not
#              positive definite):
#                "na"     -> return NA for the robust quantities (default)
#                "refit"  -> smooth the input matrix, and re-estimate the
#                            model parameters using estimator = "catML"
#                            (the former check_pd = FALSE behavior)
#                "smooth" -> smooth the input matrix, but keep the (DWLS)
#                            parameter estimates (as in lavaan <= 0.6-13)
#              if the input matrices are positive definite, all three
#              choices are identical: the ML ingredients are evaluated at
#              the (unchanged) DWLS estimates

lav_fit_catml_dwls <- function(lavobject, nonpd = "na") {
  # empty list
  empty_list <- list(
    XX3 = as.numeric(NA), df3 = as.numeric(NA),
    c.hat3 = as.numeric(NA), XX3.scaled = as.numeric(NA),
    XX3.null = as.numeric(NA), df3.null = as.numeric(NA),
    c.hat3.null = as.numeric(NA)
  )

  # limitations
  if (!lavobject@Model@categorical ||
    lavobject@Options$conditional.x ||
    length(unlist(lavobject@pta$vnames$ov.num)) > 0L) {
    return(empty_list)
  } else {
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
  }

  # check if input matrix (or matrices) are all positive definite
  if (nonpd == "na") {
    for (g in seq_len(lavdata@ngroups)) {
      cor_1 <- lavsamplestats@cov[[g]]
      ev <- eigen(cor_1, symmetric = TRUE, only.values = TRUE)$values
      if (any(ev < .Machine$double.eps^(1 / 2))) {
        # non-pd! return NA
        # should we give a warning here? (not for now)
        # warning("lavaan WARNING: robust RMSEA/CFI could not be computed
        #   because the input correlation matrix is not positive-definite")
        return(empty_list)
      }
    }
  }

  # evaluate the ML ingredients using estimator = "catML": a plain
  # (re)evaluation at the DWLS estimates if the input matrices are
  # positive definite; if not, the input matrices are smoothed first, and
  # the model parameters are re-estimated (nonpd = "refit") or kept at
  # the DWLS values (nonpd = "smooth")
  fit_catml <- try(
    lav_object_catml(lavobject, allow_refit = (nonpd != "smooth")),
    silent = TRUE
  )
  if (inherits(fit_catml, "try-error")) {
    return(empty_list)
  }

  xx3 <- fit_catml@test[[1]]$stat
  df3 <- fit_catml@test[[1]]$df


  # compute 'k'
  v <- lavTech(fit_catml, "wls.v") # NT-ML weight matrix

  w_dwls <- lavTech(lavobject, "wls.v") # DWLS weight matrix
  gamma <- lavTech(lavobject, "gamma") # acov of polychorics
  delta <- lavTech(lavobject, "delta")
  e_inv <- lavTech(lavobject, "inverted.information")

  fg <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal

  # Fixme: as we only need the trace, perhaps we could do this
  # group-specific? (see lav_test_sb_trace_original)
  v_g <- v
  w_dwls_g <- w_dwls
  gamma_f <- gamma
  delta_g <- delta
  for (g in seq_len(lavdata@ngroups)) {
    ntotal <- nrow(gamma[[g]])
    nvar <- lavobject@Model@nvar[[g]]
    pstar <- nvar * (nvar - 1) / 2
    rm_idx <- seq_len(ntotal - pstar)

    # reduce
    delta_g[[g]] <- delta[[g]][-rm_idx, , drop = FALSE]
    # reduce and weight: Gamma_g / fg paired with fg-weighted V blocks (see
    # the SCALING CONVENTIONS note in lav_samplestats_gamma.R)
    w_dwls_g[[g]] <- fg[g] * w_dwls[[g]][-rm_idx, -rm_idx]
    v_g[[g]] <- fg[g] * v[[g]] # should already have the right dims
    gamma_f[[g]] <- 1 / fg[g] * gamma[[g]][-rm_idx, -rm_idx]
  }
  # create 'big' matrices
  w_dwls_all <- lav_mat_bdiag(w_dwls_g)
  v_all <- lav_mat_bdiag(v_g)
  gamma_all <- lav_mat_bdiag(gamma_f)
  delta_all <- do.call("rbind", delta_g)

  # compute trace
  wi_u_all <- diag(nrow(w_dwls_all)) -
              delta_all %*% e_inv %*% t(delta_all) %*% w_dwls_all
  ks <- sum(diag(t(wi_u_all) %*% v_all %*% wi_u_all %*% gamma_all))

  # convert to lavaan 'scaling.factor'
  c_hat3 <- ks / df3
  xx3_scaled <- xx3 / c_hat3

  # baseline model
  xx3_null <- fit_catml@baseline$test[[1]]$stat
  if (is.null(xx3_null)) {
    xx3_null <- as.numeric(NA)
    df3_null <- as.numeric(NA)
    kbs <- as.numeric(NA)
    c_hat3_null <- as.numeric(NA)
  } else {
    df3_null <- fit_catml@baseline$test[[1]]$df
    kbs <- sum(diag(gamma_all))
    c_hat3_null <- kbs / df3_null
  }

  # return values
  list(
    XX3 = xx3, df3 = df3, c.hat3 = c_hat3, XX3.scaled = xx3_scaled,
    XX3.null = xx3_null, df3.null = df3_null, c.hat3.null = c_hat3_null
  )
}


# compute ingredients to compute FIML-Corrected RMSEA/CFI
# see:
#     Zhang X, Savalei V. (2022). New computations for RMSEA and CFI
#     following FIML and TS estimation with missing data. Psychological Methods.

lav_fit_fiml_corrected <- function(lavobject, baseline_model,
                                   version = "V3") {
  version <- toupper(version)
  if (!version %in% c("V3", "V6")) {
    lav_msg_stop(gettext("only FIML-C(V3) and FIML-C(V6) are available."))
  }

  # empty list
  empty_list <- list(
    XX3 = as.numeric(NA), df3 = as.numeric(NA),
    c.hat3 = as.numeric(NA), XX3.scaled = as.numeric(NA),
    XX3.null = as.numeric(NA), df3.null = as.numeric(NA),
    c.hat3.null = as.numeric(NA)
  )

  # limitations
  if (lavobject@Options$conditional.x ||
    lavobject@Data@nlevels > 1L ||
    is.null(lavobject@h1$implied$cov[[1]])) {
    return(empty_list)
  } else {
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats

    h1 <- lavTech(lavobject, "h1", add.labels = TRUE)
    cov_tilde <- lapply(h1, "[[", "cov")
    mean_tilde <- lapply(h1, "[[", "mean")
    sample_nobs <- unlist(lavsamplestats@nobs)
  }

  # 'refit' using 'tilde' (=EM/saturated) sample statistics
  # re-attach the data-based ov order (ov_order = "data"); parTable() has
  # stripped the "ovda" attribute, so without this fit_tilde would be built
  # in model order while lavobject (and its delta/information used below) is
  # in data order, yielding an order-dependent scaling factor. Harmless
  # no-op when the data order already equals the model order.
  pt_tilde <- parTable(lavobject)
  attr(pt_tilde, "ovda") <- lavobject@Data@ov.names[[1]]
  fit_tilde <- try(lavaan(
    model = pt_tilde,
    sample_cov = cov_tilde,
    sample_mean = mean_tilde,
    sample_nobs = sample_nobs,
    sample.cov.rescale = FALSE,
    information = "observed",
    optim.method = "none",
    se = "none",
    test = "standard",
    baseline = FALSE,
    fit.by.level = FALSE,
    check.post = FALSE
  ), silent = TRUE)
  if (inherits(fit_tilde, "try-error")) {
    return(empty_list)
  }

  xx3 <- fit_tilde@test[[1]]$stat
  df3 <- fit_tilde@test[[1]]$df

  # compute 'k'

  # V3/V6: always use h1.information = "unstructured"!!
  lavobject@Options$h1.information <- c("unstructured", "unstructured")
  lavobject@Options$observed.information <- c("h1", "h1")
  fit_tilde@Options$h1.information <- c("unstructured", "unstructured")
  fit_tilde@Options$observed.information <- c("h1", "h1")

  wm <- wm_g <- lav_model_h1_info_observed(lavobject)
  wc <- wc_g <- lav_model_h1_info_observed(fit_tilde)

  if (version == "V3") {
    jm <- jm_g <- lav_model_h1_info_firstorder(lavobject)
    gamma_f <- vector("list", length = lavdata@ngroups)
  }
  delta <- lavTech(lavobject, "delta")
  e_inv <- lavTech(lavobject, "inverted.information")
  wmi <- wmi_g <- try(lapply(wm, lav_mat_sym_inverse),
    silent = TRUE
  )
  if (inherits(wmi, "try-error")) {
    return(empty_list)
  }

  fg <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal
  # Fixme: as we only need the trace, perhaps we could do this
  # group-specific? (see lav_test_sb_trace_original)
  for (g in seq_len(lavdata@ngroups)) {
    # group weight
    wc_g[[g]] <- fg[g] * wc[[g]]
    wm_g[[g]] <- fg[g] * wm[[g]]
    wmi_g[[g]] <- 1 / fg[g] * wmi[[g]]

    # gamma
    if (version == "V3") {
      jm_g[[g]] <- fg[g] * jm[[g]]
      gamma_g <- wmi[[g]] %*% jm[[g]] %*% wmi[[g]]
      gamma_f[[g]] <- 1 / fg[g] * gamma_g
    }
  }
  # create 'big' matrices
  wc_all <- lav_mat_bdiag(wc_g)
  wm_all <- lav_mat_bdiag(wm_g)
  wmi_all <- lav_mat_bdiag(wmi_g)
  delta_all <- do.call("rbind", delta)

  e_comp <- t(delta_all) %*% wc_all %*% delta_all
               # VS: or grab from fit.tilde, with observed.info="h1"


  # compute trace
  if (version == "V3") {
    gamma_all <- lav_mat_bdiag(gamma_f)
    # VS: Simplification of k.fimlc to minimize matrix multiplication
    #                                                 of big matrices
    jm_all <- lav_mat_bdiag(jm_g)

    # VS: tr11 is also used for baseline
    # VS: tr(AB) = sum(A*t(B)) is more efficient

    tr11 <- sum(wc_all * gamma_all)
    tr12 <- sum((t(delta_all) %*% jm_all %*% wmi_all %*%
                 wc_all %*% delta_all) * e_inv)
    tr22 <- sum((t(delta_all) %*% jm_all %*% delta_all %*% e_inv)
             * t(t(delta_all) %*% wc_all %*% delta_all %*% e_inv))

    k_fimlc <- tr11 - 2 * tr12 + tr22
  } else {
    # V6
    tr1 <- sum(wc_all * wmi_all)
    k_fimlc <- tr1 - sum(e_comp * e_inv)
  }

  # convert to lavaan 'scaling.factor'
  c_hat3 <- k_fimlc / df3
  xx3_scaled <- xx3 / c_hat3

  # collect temp results
  out <- list(
    XX3 = xx3, df3 = df3,
    c.hat3 = c_hat3, XX3.scaled = xx3_scaled,
    XX3.null = as.numeric(NA), df3.null = as.numeric(NA),
    c.hat3.null = as.numeric(NA)
  )

  # baseline model
  if (!is.null(baseline_model)) {
    fit_b <- baseline_model
  } else {
    fit_b <- try(lav_object_independence(lavobject), silent = TRUE)
  }

  if (inherits(fit_b, "try-error")) {
    return(out)
  }

  # 'refit' using 'tilde' (=EM/saturated) sample statistics
  fit_b_tilde <- try(lavaan(
    model = parTable(fit_b),
    sample_cov = cov_tilde,
    sample_mean = mean_tilde,
    sample_nobs = sample_nobs,
    sample.cov.rescale = FALSE,
    information = "observed",
    optim.method = "none",
    se = "none",
    test = "standard",
    baseline = FALSE,
    fit.by.level = FALSE,
    check.post = FALSE
  ), silent = TRUE)
  if (inherits(fit_b_tilde, "try-error")) {
    return(out)
  }

  xx3_null <- fit_b_tilde@test[[1]]$stat
  df3_null <- fit_b_tilde@test[[1]]$df

  fit_b@Options$h1.information <- c("unstructured", "unstructured")
  fit_b@Options$observed.information <- c("h1", "h1")
  fit_b_tilde@Options$h1.information <- c("unstructured", "unstructured")
  fit_b_tilde@Options$observed.information <- c("h1", "h1")

  e_inv_b <- lavTech(fit_b, "inverted.information")
  delta_b <- lavTech(fit_b, "Delta")
  delta_b_all <- do.call("rbind", delta_b)

  e_comp_b <- t(delta_b_all) %*% wc_all %*% delta_b_all #or grab from fitB.tilde

  # V3 or V6?
  if (version == "V3") {
    tr12b <-
      sum((t(delta_b_all) %*% jm_all %*% wmi_all %*% wc_all %*% delta_b_all) *
        e_inv_b)
    tr22b <-
      sum((t(delta_b_all) %*% jm_all %*% delta_b_all %*% e_inv_b) *
        t(t(delta_b_all) %*% wc_all %*% delta_b_all %*% e_inv_b))
    kb_fimlc <- tr11 - 2 * tr12b + tr22b
  } else {
    # V6
    kb_fimlc <- tr1 - sum(e_comp_b * e_inv_b)
  }

  # convert to lavaan 'scaling.factor'
  c_hat3_null <- kb_fimlc / df3_null

  # return values
  list(
    XX3 = xx3, df3 = df3, c.hat3 = c_hat3, XX3.scaled = xx3_scaled,
    XX3.null = xx3_null, df3.null = df3_null, c.hat3.null = c_hat3_null
  )
}

# classic Wald test
#
# NOTE: does not handle redundant constraints yet!
#
# - YR 11 Jul 2026: add the scaled and adjusted versions of the Wald test
#                   (Satorra, 2000); see lav_test_satorra2000.R. Note that
#                   the main statistic is already the generalized ('robust')
#                   Wald test whenever the object's se is sandwich-based,
#                   as it is based on the object's vcov.
#

lavTestWald <- function(object, constraints = NULL, verbose = FALSE) { # nolint
  # check object
  object <- lav_object_check_version(object)

  if (!missing(verbose)) {
    current_verbose <- lav_verbose()
    if (lav_verbose(verbose))
      on.exit(lav_verbose(current_verbose), TRUE)
  }
  if (object@optim$npar > 0L && !object@optim$converged) {
    lav_msg_stop(gettext("model did not converge"))
  }

  if (is.null(constraints) || all(nchar(constraints) == 0L)) {
    lav_msg_stop(gettext("constraints are empty"))
  }

  # extract slots
  lavoptions <- object@Options
  lavmodel <- object@Model
  lavpartable <- data.frame(object@ParTable)

  # remove == constraints from parTable
  eq_idx <- which(lavpartable$op == "==")
  if (length(eq_idx) > 0L) {
    lavpartable <- lavpartable[-eq_idx, ]
  }
  partable <- as.list(lavpartable)

  # parse constraints
  flat <- lavParseModelString(constraints, parser = lavoptions$parser)
  con <- attr(flat, "constraints")
  list_1 <- list()
  if (length(con) > 0L) {
    lhs <- unlist(lapply(con, "[[", "lhs"))
    op <- unlist(lapply(con, "[[", "op"))
    rhs <- unlist(lapply(con, "[[", "rhs"))
    list_1$lhs <- c(list_1$lhs, lhs)
    list_1$op <- c(list_1$op, op)
    list_1$rhs <- c(list_1$rhs, rhs)
  } else {
    lav_msg_stop(gettext(
      "no equality constraints found in constraints argument"))
  }

  # theta = free parameters only
  theta <- lav_model_get_parameters(lavmodel)

  # build constraint function
  ceq_function <- lav_pt_con_ceq(
    partable = partable,
    con = list_1, debug = FALSE
  )
  # compute jacobian restrictions
  jac <- try(lav_func_jacobian_complex(func = ceq_function, x = theta),
    silent = TRUE
  )
  if (inherits(jac, "try-error")) { # eg. pnorm()
    jac <- lav_func_jacobian_simple(func = ceq_function, x = theta)
  }

  # check for linear redundant rows in JAC
  out <- lav_mat_rref(t(jac))
  ran_k <- length(out$pivot)
  if (ran_k < nrow(jac)) {
    lav_msg_warn(gettext(
      "Jacobian of constraints is rank deficient. Some constraints",
      " may be redundant, and have been removed."))
  }
  jac <- jac[out$pivot, , drop = FALSE]

  if (lav_verbose()) {
    cat("Restriction matrix (jacobian):\n")
    print(jac)
    cat("\n")
  }

  # linear restriction
  theta_r <- ceq_function(theta)
  theta_r <- theta_r[out$pivot]

  if (lav_verbose()) {
    cat("Restricted theta values:\n")
    print(theta_r)
    cat("\n")
  }

  # get VCOV
  # VCOV <- vcov(object, labels = FALSE)
  # avoid S4 dispatch
  vcov_1 <- lav_inspect_vcov(object,
    standardized = FALSE,
    free_only = TRUE,
    add_labels = FALSE,
    add_class = FALSE,
    remove_duplicated = FALSE
  )

  # the constraint jacobian is in the compact (nx.free) space; when
  # ceq.simple.only, vcov_1 is in the larger 'unco' space, so reduce it
  vcov_1 <- lav_model_vcov_unco_to_free(lavmodel, vcov_1)

  # restricted vcov
  vcov_r <- jac %*% vcov_1 %*% t(jac)

  # fixme: what if VCOV.r is singular?
  # Wald test statistic
  # (this is the generalized/robust Wald test of Satorra (2000) whenever
  # the object's se is sandwich-based, as vcov_1 is then a sandwich)
  wald <- as.numeric(t(theta_r) %*% solve(vcov_r) %*% theta_r)

  # df
  wald_df <- ran_k

  # p-value based on chisq
  wald_pvalue <- 1 - pchisq(wald, df = wald_df)

  # prepare output
  out <- list(
    stat = wald, df = wald_df, p.value = wald_pvalue,
    se = lavoptions$se
  )

  # scaled, adjusted and generalized ('robust') versions (Satorra, 2000);
  # these need both the standard (normal-theory) and a sandwich vcov;
  # not for rotated (EFA/ESEM) solutions, as the stored vcov is
  # post-processed in that case (see step 16)
  se_flavor <- lav_test_robust_se_flavor(object)
  if (!is.null(se_flavor) &&
    !(lavmodel@nefa > 0L && lavoptions$rotation != "none")) {
    # helper: (re)compute the vcov for a given se setting, in the same
    # (nx.free) parameter space as the constraint jacobian
    wald_vcov <- function(se) {
      lavoptions2 <- lavoptions
      lavoptions2$se <- se
      lavoptions2$check.vcov <- FALSE
      lavh1 <- object@h1
      if (length(lavh1) == 0L) {
        lavh1 <- NULL
      }
      vv <- try(
        lav_model_vcov(
          lavmodel = lavmodel,
          lavsamplestats = object@SampleStats,
          lavoptions = lavoptions2,
          lavdata = object@Data,
          lavpartable = object@ParTable,
          lavcache = object@Cache,
          lavimplied = object@implied,
          lavh1 = lavh1
        ),
        silent = TRUE
      )
      if (inherits(vv, "try-error") || is.null(vv)) {
        return(NULL)
      }
      lav_model_vcov_unco_to_free(lavmodel, vv)
    }

    # sandwich vcov (reuse the object's vcov if it is already of this
    # flavor) and normal-theory vcov
    if (lavoptions$se == se_flavor) {
      vcov_rob <- vcov_1
    } else {
      vcov_rob <- wald_vcov(se_flavor)
    }
    if (lavoptions$se == "standard") {
      vcov_nt <- vcov_1
    } else {
      vcov_nt <- wald_vcov("standard")
    }

    if (!is.null(vcov_rob) && !is.null(vcov_nt)) {
      m1 <- jac %*% vcov_nt %*% t(jac)   # A J A' / n
      m2 <- jac %*% vcov_rob %*% t(jac)  # A J B J A' / n
      # the standard (normal-theory) Wald statistic
      wald_nt <- as.numeric(t(theta_r) %*% MASS::ginv(m1) %*% theta_r)
      s2000 <- lav_test_satorra2000(
        stat = wald_nt, df = wald_df, m1 = m1, m2 = m2,
        v = theta_r, n = 1
      )
      out$stat.standard <- wald_nt
      out$stat.scaled <- s2000$stat.scaled
      out$p.value.scaled <- s2000$p.value.scaled
      out$scaling.factor <- s2000$scaling.factor
      out$stat.adjusted <- s2000$stat.adjusted
      out$df.adjusted <- s2000$df.adjusted
      out$p.value.adjusted <- s2000$p.value.adjusted
      out$stat.robust <- s2000$stat.robust
      out$p.value.robust <- s2000$p.value.robust
    }
  }

  out
}

# adapt marker (bad.marker.crit)
#
# By default, lavaan sets the metric of a latent variable by fixing the
# factor loading of the FIRST indicator to 1.0. This works well, except when
# the first indicator happens to be a poor item (i.e., it correlates weakly
# with the other indicators of the same factor). In that case, fixing its
# loading to 1.0 often leads to (serious) convergence problems, and the user
# has no idea why.
#
# When bad.marker.crit > 0 (the default is 0.1, and auto.fix.first = TRUE),
# lavaan inspects -- after the unrestricted (h1) sample statistics are
# available -- whether the first indicator of each latent variable is a poor
# item. If so, a warning is issued and another (better) indicator is used as
# the marker instead. If the first indicator is good enough, nothing happens.
#
# The 'quality' of an indicator is measured by its corrected item-total
# correlation: the correlation between the indicator and the sum score of the
# OTHER indicators of the same factor. This is computed from the (pooled)
# unrestricted (h1) covariance matrix, so the behavior is consistent whether
# or not we have missing data, categorical data, etc.
#
# Composites (op == "<~") get a similar treatment, but with a different
# quality measure: the metric of a composite is set by fixing the WEIGHT of
# the first indicator to 1.0. If the ML estimate of that weight is (near)
# zero, the remaining (free) weights must diverge to infinity to reach the
# optimum, and the model never converges. The implied (relative) weights can
# be gauged before estimation: under the composite model, the covariances of
# the indicators of one composite with ANY outside variable are proportional
# to a single vector l = S_jj w (S_jj = the within-block covariance matrix).
# We estimate l as the dominant (rank-1) left singular vector of the
# cross-correlation block between the indicators and all other observed
# variables, solve w = S_jj^{-1} l, and rescale so max |w| = 1. If the first
# indicator's relative weight is below the threshold (and a clearly better
# indicator exists), we switch the marker to the indicator with the largest
# absolute weight.
#
# lav_pt_marker_adapt() returns NULL if no change is needed, or a list
# with:
#   - marker: a named character vector (lv -> new indicator) for the latent
#             variables whose marker should be switched
#   - info:   a data.frame (lv, type, old, new, r.old, r.new) for the
#             warning; type is "factor" or "composite"; for factors r.* are
#             corrected item-total correlations, for composites relative
#             weights
#
# 'threshold' is the value below which the first indicator is considered a
# 'poor' marker (in absolute value of its corrected item-total correlation,
# or of its relative weight for composites); it is set by the
# 'bad.marker.crit' option (default 0.1)
lav_pt_marker_adapt <- function(lavpartable = NULL,
                                lavh1 = NULL,
                                lavoptions = NULL,
                                lavdata = NULL,
                                threshold = 0.1) {

  # we need the unrestricted (h1) covariance matrix/matrices
  implied <- lavh1$implied
  if (is.null(implied)) {
    return(NULL)
  }
  cov_list <- implied[["cov"]]
  if (is.null(cov_list)) {
    # conditional.x: use the residual covariance matrix
    cov_list <- implied[["res.cov"]]
  }
  if (is.null(cov_list) || length(cov_list) == 0L) {
    return(NULL)
  }

  # variable names per block (from the dimnames, with a single-level fallback)
  block_names <- lapply(seq_along(cov_list), function(b) {
    vn <- rownames(cov_list[[b]])
    if (is.null(vn) && lavdata@nlevels == 1L &&
        b <= length(lavdata@ov.names)) {
      vn <- lavdata@ov.names[[b]]
    }
    vn
  })

  # observed indicators (ordered indicators are fine: for those, the h1 'cov'
  # already holds the latent (polychoric) correlations)
  ov_names <- lav_pt_vnames(lavpartable, type = "ov")
  # exclude efa factors (composites are handled separately below:
  # lv.regular only contains =~ definitions)
  lv_regular <- lav_pt_vnames(lavpartable, type = "lv.regular")
  if (!is.null(lavpartable$efa)) {
    lv_efa <- unique(lavpartable$lhs[lavpartable$op == "=~" &
                                     nchar(lavpartable$efa) > 0L])
    lv_regular <- lv_regular[!lv_regular %in% lv_efa]
  }
  # composites
  lv_comp <- unique(lavpartable$lhs[lavpartable$op == "<~"])
  if (length(lv_regular) == 0L && length(lv_comp) == 0L) {
    return(NULL)
  }

  # corrected item-total correlations, given a covariance submatrix C
  cit_from_cov <- function(C) {
    d <- diag(C)
    R <- rowSums(C)
    S <- sum(C)
    cov_rest <- R - d
    var_rest <- S - 2 * R + d
    out <- cov_rest / sqrt(d * var_rest)
    out[!is.finite(out)] <- NA_real_
    out
  }

  new_marker <- character(0L)
  info <- list()

  for (lv in lv_regular) {
    lv_rows <- which(lavpartable$op == "=~" & lavpartable$lhs == lv)
    # only consider one block (the structure is identical across blocks)
    b1 <- lavpartable$block[lv_rows]
    lv_rows <- lv_rows[b1 == b1[1L]]
    ind <- lavpartable$rhs[lv_rows]

    # only observed indicators, and at least two of them
    keep <- ind %in% ov_names
    lv_rows <- lv_rows[keep]
    ind <- ind[keep]
    if (length(ind) < 2L) {
      next
    }

    # only adapt the 'clean' default situation: exactly one indicator has a
    # fixed loading (free == 0), it is the FIRST indicator, and its value is
    # 1.0; otherwise the user set up the scaling and we leave it alone
    fixed <- which(lavpartable$free[lv_rows] == 0L)
    if (length(fixed) != 1L || fixed != 1L ||
        !isTRUE(lavpartable$ustart[lv_rows][1L] == 1)) {
      next
    }
    cur <- ind[1L]

    # collect corrected item-total correlations across all blocks that
    # contain all the indicators of this factor
    cit_blocks <- list()
    for (b in seq_along(cov_list)) {
      vn <- block_names[[b]]
      if (is.null(vn) || !all(ind %in% vn)) {
        next
      }
      idx <- match(ind, vn) # the h1 cov matrices may have no dimnames
      C <- cov_list[[b]][idx, idx, drop = FALSE]
      cit_blocks[[length(cit_blocks) + 1L]] <- cit_from_cov(C)
    }
    if (length(cit_blocks) == 0L) {
      next
    }
    cit <- colMeans(do.call(rbind, cit_blocks), na.rm = TRUE)
    names(cit) <- ind
    if (all(is.na(cit))) {
      next
    }

    r <- abs(cit)
    r_cur <- r[[1L]]
    best <- which.max(r)
    # switch only if the first indicator is poor AND a clearly better
    # indicator exists
    if (!is.na(r_cur) && r_cur < threshold &&
        ind[best] != cur && r[[best]] >= threshold) {
      new_marker[lv] <- ind[best]
      info[[length(info) + 1L]] <- data.frame(
        lv = lv, type = "factor", old = cur, new = ind[best],
        r.old = round(cit[[1L]], 3), r.new = round(cit[[best]], 3),
        stringsAsFactors = FALSE
      )
    }
  }

  # composites: check the implied relative weight of the first indicator
  for (lv in lv_comp) {
    lv_rows <- which(lavpartable$op == "<~" & lavpartable$lhs == lv)
    # only consider one block (the structure is identical across blocks)
    b1 <- lavpartable$block[lv_rows]
    lv_rows <- lv_rows[b1 == b1[1L]]
    ind <- lavpartable$rhs[lv_rows]

    # only observed indicators, and at least two of them
    keep <- ind %in% ov_names
    lv_rows <- lv_rows[keep]
    ind <- ind[keep]
    if (length(ind) < 2L) {
      next
    }

    # only adapt the 'clean' default situation: exactly one indicator has a
    # fixed weight (free == 0), it is the FIRST indicator, and its value is
    # 1.0; otherwise the user set up the scaling and we leave it alone
    fixed <- which(lavpartable$free[lv_rows] == 0L)
    if (length(fixed) != 1L || fixed != 1L ||
        !isTRUE(lavpartable$ustart[lv_rows][1L] == 1)) {
      next
    }
    cur <- ind[1L]

    # implied relative weights, collected across all blocks that contain
    # all the indicators of this composite plus at least one other variable
    w_blocks <- list()
    for (b in seq_along(cov_list)) {
      vn <- block_names[[b]]
      if (is.null(vn) || !all(ind %in% vn)) {
        next
      }
      in_idx <- match(ind, vn) # the h1 cov matrices may have no dimnames
      out_idx <- setdiff(seq_along(vn), in_idx)
      if (length(out_idx) == 0L) {
        next
      }
      C <- cov_list[[b]]
      d <- diag(C)
      if (any(!is.finite(d)) || any(d <= 0)) {
        next
      }
      # scale-free: work with correlations
      isd <- 1 / sqrt(d)
      r_jj <- C[in_idx, in_idx, drop = FALSE] *
              tcrossprod(isd[in_idx])
      r_jo <- C[in_idx, out_idx, drop = FALSE] *
              tcrossprod(isd[in_idx], isd[out_idx])
      # dominant direction of the cross-correlations (~ S_jj %*% w)
      sv <- try(svd(r_jo, nu = 1L, nv = 0L), silent = TRUE)
      if (inherits(sv, "try-error") || sv$d[1L] < .Machine$double.eps) {
        next
      }
      w <- try(solve(r_jj, sv$u[, 1L]), silent = TRUE)
      if (inherits(w, "try-error") || any(!is.finite(w))) {
        next
      }
      w_blocks[[length(w_blocks) + 1L]] <- abs(w) / max(abs(w))
    }
    if (length(w_blocks) == 0L) {
      next
    }
    r <- colMeans(do.call(rbind, w_blocks), na.rm = TRUE)
    names(r) <- ind
    if (all(is.na(r))) {
      next
    }

    r_cur <- r[[1L]]
    best <- which.max(r)
    # switch only if the first indicator has a (near) zero implied weight
    # AND a clearly better indicator exists
    if (!is.na(r_cur) && r_cur < threshold &&
        ind[best] != cur && r[[best]] >= threshold) {
      new_marker[lv] <- ind[best]
      info[[length(info) + 1L]] <- data.frame(
        lv = lv, type = "composite", old = cur, new = ind[best],
        r.old = round(r[[1L]], 3), r.new = round(r[[best]], 3),
        stringsAsFactors = FALSE
      )
    }
  }

  if (length(new_marker) == 0L) {
    return(NULL)
  }

  list(marker = new_marker, info = do.call(rbind, info))
}

# the bivariate two-level linear/ordinal model, univariate parameters fixed
#
#     y_a,ij  = mu_a + b_a,j + e_a,ij           (continuous)
#     y*_c,ij = w_c,ij + b_c,j                  (ordinal, probit)
#
#     Cov(e_a, w_c) = cw (within),  Cov(b_a, b_c) = cb (between)
#     Var(e_a) = vw_a, Var(w_c) = 1, Var(b_a) = vb_a, Var(b_c) = vb_c
#
# YR 2026 (two-level WLS, phase 3)
#
# This is stage 2 of the stage-wise unrestricted two-level model for a
# mixed continuous/ordinal pair: the univariate parameters (mu_a, vw_a,
# vb_a from lav_uvreg_2l_fit; tau_c, vb_c from lav_uvord_2l_fit) are held
# fixed, and theta = c(cw, cb) is estimated by ML with ONE-dimensional
# Gauss-Hermite quadrature:
#
# the probit means depend on the cluster effects only through the scalar
#
#     gamma_j = b_c,j - (cw/vw_a) b_a,j
#
# because, conditional on (b_a, b_c) and y_a,
#
#     y*_c,ij | . ~ N( gamma_j + (cw/vw_a)(y_a,ij - mu_a),  1 - cw^2/vw_a )
#
# with independence over i. Hence
#
#     L_j = f(y_a,j) * int prod_i P(y_c,ij | gamma) phi(gamma; m_j, v_j)
#
# where f(y_a,j) is the (fixed-parameter) univariate two-level Gaussian
# density of y_a in cluster j, and gamma | y_a,j is normal with
#
#     m_j = c_g nj r_aj / v_aj,  v_j = Var(gamma) - c_g^2 nj / v_aj
#     c_g = cb - (cw/vw_a) vb_a,  Var(gamma) = vb_c + (cw/vw_a)^2 vb_a
#                                              - 2 (cw/vw_a) cb
#     v_aj = vw_a + nj vb_a,  r_aj = ybar_a,j - mu_a
#
# The gradient is posterior-weighted, with both a direct term (cw enters
# the conditional probit) and a chain term through the per-cluster node
# positions gamma_jq = m_j + sqrt(v_j) x_q.

lav_bvmix_2l_cov_twostep_fit <- function(fit_y1 = NULL, # continuous
                                         fit_y2 = NULL, # ordinal
                                         ngh = 21L,
                                         optim_method = "nlminb1",
                                         control = list(),
                                         y1_name = NULL, y2_name = NULL,
                                         output = "theta") {
  # create cache environment
  cache <- lav_bvmix_2l_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, ngh = ngh
  )

  # bounds
  cw_bound <- 0.999 * sqrt(cache$vw_a)
  cb_bound <- 0.999 * sqrt(cache$vb_a * cache$vb_c)
  cb_fixed <- (cb_bound < 1e-08)
  if (cb_fixed) {
    cache$theta[2] <- 0
    lower <- c(-cw_bound, 0)
    upper <- c(+cw_bound, 0)
  } else {
    lower <- c(-cw_bound, -cb_bound)
    upper <- c(+cw_bound, +cb_bound)
  }

  # optim
  minfns <- lav_uvbv_optim_fns(optim_method, lav_bvmix_2l_min_fns())

  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  # try 1
  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb, lower = lower, upper = upper,
    cache = cache
  )

  # try 2: start from independence
  if (optim$convergence != 0L) {
    optim <- nlminb(
      start = c(0, 0), objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control_nlminb, lower = lower, upper = upper,
      cache = cache
    )
  }

  if (optim$convergence != 0L) {
    if (!is.null(y1_name) && !is.null(y2_name)) {
      lav_msg_warn(gettextf(
        "estimation of the two-level polyserial parameters did not converge
        for variables %1$s and %2$s", y1_name, y2_name))
    } else {
      lav_msg_warn(gettext(
        "estimation of the two-level polyserial parameters did not always
        converge"))
    }
  }

  if (output == "cache") {
    cache$theta <- optim$par
    return(cache)
  }

  # theta = c(cw, cb)
  optim$par
}

# prepare cache environment
# fit_y1: lav_uvreg_2l_fit() object (continuous)
# fit_y2: lav_uvord_2l_fit() object (ordinal)
lav_bvmix_2l_init_cache <- function(fit_y1 = NULL,
                                    fit_y2 = NULL,
                                    ngh = 21L,
                                    parent = parent.frame()) {
  stopifnot(length(fit_y1$y) == length(fit_y2$y))
  y_a <- fit_y1$y
  y_c <- fit_y2$y
  cluster_idx <- fit_y1$cluster_idx
  nobs <- length(y_a)

  # fixed univariate parameters
  mu_a <- fit_y1$theta[fit_y1$mu_idx]
  vw_a <- fit_y1$theta[fit_y1$wvar_idx]
  vb_a <- fit_y1$theta[fit_y1$bvar_idx]
  tau_c <- fit_y2$theta[fit_y2$th_idx]
  vb_c <- fit_y2$theta[fit_y2$bvar_idx]
  nth <- length(tau_c)

  # covariates: residualize the continuous side (the pair model is then
  # the no-covariate model with mu_a = 0), and compute the fixed linear
  # predictor eta_c of the ordinal side
  nexo_a <- fit_y1$nexo
  nexo_c <- fit_y2$nexo
  z_a <- z_c <- NULL
  if (nexo_a > 0L) {
    theta_ma <- fit_y1$theta[c(1L, 3L + seq_len(nexo_a))]
    y_a <- fit_y1$y - drop(fit_y1$z %*% theta_ma)
    mu_a <- 0
    z_a <- fit_y1$z[, -1, drop = FALSE] # slope columns only
  }
  eta_c <- 0
  if (nexo_c > 0L) {
    beta_c <- fit_y2$theta[fit_y2$nth + 1L + seq_len(nexo_c)]
    z_c <- fit_y2$z
    eta_c <- drop(z_c %*% beta_c)
  }

  # cluster structure (reuse the stage-1 sufficient statistics of y_a,
  # unless we residualized)
  if (nexo_a > 0L) {
    cs_a <- lav_2l_cluster_stats(y = y_a, cluster_idx = cluster_idx)
  } else {
    cs_a <- fit_y1$cs
  }
  njs <- cs_a$njs
  nclusters <- cs_a$nclusters
  cl_sorted <- sort(unique(cluster_idx))
  cl_pos <- match(cluster_idx, cl_sorted)

  # cluster means of the a-side slope design columns (for the chain
  # through m_j: d r_aj / d beta_r = -zbar_a[j, r])
  zbar_a <- NULL
  if (nexo_a > 0L) {
    zbar_a <- rowsum.default(z_a, group = cluster_idx,
                             reorder = TRUE) / njs
  }

  # fixed per-cluster quantities
  v_aj <- vw_a + njs * vb_a
  r_aj <- cs_a$ybar[, 1] - mu_a

  # fixed constant: log f(y_a,j) (does not depend on cw/cb)
  logf_a <- -0.5 * lav_2l_gauss_m2ll(
    cs = cs_a, mu = mu_a,
    sigma_w = matrix(vw_a, 1, 1), sigma_b = matrix(vb_a, 1, 1)
  )

  # per-item pieces
  dev_a <- y_a - mu_a
  o12 <- lav_uvord_o12(y = y_c, nth = nth)
  y1 <- matrix(seq_len(nth), nobs, nth, byrow = TRUE) == y_c
  y2 <- matrix(seq_len(nth), nobs, nth, byrow = TRUE) == (y_c - 1L)

  # standardized GH nodes
  gh <- lav_2l_gh_xw(ngh = ngh)

  # starting values: crude split of the total correlation
  r_tot <- suppressWarnings(cor(y_a, y_c))
  if (!is.finite(r_tot)) {
    r_tot <- 0
  }
  cw0 <- max(min(r_tot * sqrt(vw_a), 0.9 * sqrt(vw_a)),
             -0.9 * sqrt(vw_a))
  cb_max <- 0.9 * sqrt(vb_a * vb_c)
  cb0 <- max(min(r_tot * sqrt(vb_a * vb_c), cb_max), -cb_max)
  theta <- c(cw0, cb0)

  list2env(
    list(
      y_c = y_c, cluster_idx = cluster_idx, cl_pos = cl_pos,
      nobs = nobs, n = nobs, nclusters = nclusters, njs = njs,
      mu_a = mu_a, vw_a = vw_a, vb_a = vb_a,
      tau_c = tau_c, vb_c = vb_c, nth = nth,
      nexo_a = nexo_a, nexo_c = nexo_c,
      nexo_wa = fit_y1$nexo_w, nexo_ba = fit_y1$nexo_b,
      nexo_wc = fit_y2$nexo_w, nexo_bc = fit_y2$nexo_b,
      z_a = z_a, z_c = z_c, zbar_a = zbar_a, eta_c = eta_c,
      v_aj = v_aj, r_aj = r_aj, logf_a = logf_a, cs_a = cs_a,
      dev_a = dev_a, o1 = o12$o1, o2 = o12$o2, y1 = y1, y2 = y2,
      gh = gh, ngh = ngh,
      theta = theta
    ),
    parent = parent
  )
}

# the (cw, cb)-dependent per-cluster quantities
lav_bvmix_2l_gamma_pars <- function(cache = NULL) {
  with(cache, {          # nolint start
    cw <- theta[1]
    cb <- theta[2]
    a_sl <- cw / vw_a
    sd_c2 <- 1 - cw * cw / vw_a
    if (sd_c2 <= 0) {
      return(NULL)
    }
    sd_c <- sqrt(sd_c2)
    c_g <- cb - a_sl * vb_a
    var_g <- vb_c + a_sl * a_sl * vb_a - 2 * a_sl * cb
    m_j <- c_g * njs * r_aj / v_aj
    v_gj <- var_g - c_g * c_g * njs / v_aj
    if (any(v_gj <= 0)) {
      return(NULL)
    }
    out <- list(
      a_sl = a_sl, sd_c = sd_c, c_g = c_g, var_g = var_g,
      m_j = m_j, v_gj = v_gj, s_gj = sqrt(v_gj)
    )
    return(out)
  })                     # nolint end
}

lav_bvmix_2l_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    gp <- lav_bvmix_2l_gamma_pars(cache)
    if (is.null(gp)) {
      loglik <- -Inf
      return(loglik)
    }

    th_s <- tau_c / gp$sd_c
    # per item, fixed over nodes (eta_c = fixed ordinal linear predictor)
    base_i <- (gp$a_sl * dev_a + eta_c) / gp$sd_c

    ll <- matrix(0, nclusters, ngh)
    for (q in seq_len(ngh)) {
      gamma_j <- gp$m_j + gp$s_gj * gh$x[q]
      offset_i <- gamma_j[cl_pos] / gp$sd_c + base_i
      z12 <- lav_uvord_z12(y = y_c, th = th_s, o1 = o1, o2 = o2,
                           offset = offset_i)
      pi_i <- pnorm(z12$z1) - pnorm(z12$z2)
      large_idx <- which(z12$z2 > 1)
      if (length(large_idx) > 0L) {
        pi_i[large_idx] <- (pnorm(z12$z2[large_idx], lower.tail = FALSE) -
          pnorm(z12$z1[large_idx], lower.tail = FALSE))
      }
      pi_i[pi_i < .Machine$double.eps] <- .Machine$double.eps
      ll[, q] <- rowsum.default(log(pi_i), group = cluster_idx,
                                reorder = TRUE)
    }

    lwll <- sweep(ll, 2L, gh$logw, "+")
    mmax <- apply(lwll, 1L, max)
    logl_j <- mmax + log(rowSums(exp(lwll - mmax)))
    pw <- exp(lwll - logl_j)

    loglik <- sum(logl_j + logf_a)
    return(loglik)
  })                     # nolint end
}

lav_bvmix_2l_grad_cache <- function(cache = NULL) {
  sc <- lav_bvmix_2l_sc_cache(cache = cache)
  colSums(sc[, c("cw", "cb"), drop = FALSE])
}

# derivatives of the model ingredients (a_sl, sd_c, c_g, var_g, v_aj,
# r_aj, dev) with respect to every parameter, for the chain rule:
# columns of the returned patterns:
#   mu_a, vw_a, vb_a, vb_c, cw, cb
lav_bvmix_2l_dpats <- function(cache = NULL, gp = NULL) {
  with(cache, {          # nolint start
    cw <- theta[1]
    cb <- theta[2]
    a_sl <- gp$a_sl
    sd_c <- gp$sd_c

    pats <- list(
      mu_a = list(da = 0, dsd = 0, dcg = 0, dvarg = 0,
                  dvw = 0, dvb = 0, dr = -1, ddev = -1),
      vw_a = list(da = -a_sl / vw_a,
                  dsd = cw * cw / (2 * vw_a * vw_a * sd_c),
                  dcg = a_sl * vb_a / vw_a,
                  dvarg = (-a_sl / vw_a) * (2 * a_sl * vb_a - 2 * cb),
                  dvw = 1, dvb = 0, dr = 0, ddev = 0),
      vb_a = list(da = 0, dsd = 0, dcg = -a_sl, dvarg = a_sl * a_sl,
                  dvw = 0, dvb = 1, dr = 0, ddev = 0),
      vb_c = list(da = 0, dsd = 0, dcg = 0, dvarg = 1,
                  dvw = 0, dvb = 0, dr = 0, ddev = 0),
      cw = list(da = 1 / vw_a, dsd = -cw / (vw_a * sd_c),
                dcg = -vb_a / vw_a,
                dvarg = (2 * a_sl * vb_a - 2 * cb) / vw_a,
                dvw = 0, dvb = 0, dr = 0, ddev = 0),
      cb = list(da = 0, dsd = 0, dcg = 1, dvarg = -2 * a_sl,
                dvw = 0, dvb = 0, dr = 0, ddev = 0)
    )
    return(pats)
  })                     # nolint end
}

# cluster-wise scores of the loglik w.r.t. ALL parameters
# columns: mu_a, vw_a, vb_a, tau_c (nth), vb_c, cw, cb
# assumes the loglik has been evaluated at cache$theta (pw available)
lav_bvmix_2l_sc_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    gp <- lav_bvmix_2l_gamma_pars(cache)
    a_sl <- gp$a_sl
    sd_c <- gp$sd_c
    c_g <- gp$c_g

    pats <- lav_bvmix_2l_dpats(cache, gp = gp)
    pnames <- names(pats)
    np <- length(pnames)

    # per-cluster chain pieces: gamma_jq = m_j + s_gj x_q, with
    # m_j = c_g nj r_aj / v_aj and v_gj = var_g - c_g^2 nj / v_aj
    dm <- matrix(0, nclusters, np)
    dsg <- matrix(0, nclusters, np)
    for (p in seq_len(np)) {
      pt <- pats[[p]]
      dvaj <- pt$dvw + njs * pt$dvb
      dm[, p] <- pt$dcg * njs * r_aj / v_aj +
        c_g * njs * pt$dr / v_aj -
        c_g * njs * r_aj * dvaj / (v_aj * v_aj)
      dvg <- pt$dvarg -
        njs * (2 * c_g * pt$dcg * v_aj - c_g * c_g * dvaj) /
          (v_aj * v_aj)
      dsg[, p] <- dvg / (2 * gp$s_gj)
    }

    th_s <- tau_c / sd_c
    base_i <- (a_sl * dev_a + eta_c) / sd_c

    sc <- matrix(0, nclusters, nth + np + nexo_a + nexo_c)
    tau_idx <- 3L + seq_len(nth) # after mu_a, vw_a, vb_a
    par_idx <- c(1L, 2L, 3L, 3L + nth + 1L, 3L + nth + 2L, 3L + nth + 3L)
    # (mu_a, vw_a, vb_a, vb_c, cw, cb) positions in the output
    asl_idx <- nth + np + seq_len(nexo_a)
    csl_idx <- nth + np + nexo_a + seq_len(nexo_c)

    for (q in seq_len(ngh)) {
      gamma_j <- gp$m_j + gp$s_gj * gh$x[q]
      offset_i <- gamma_j[cl_pos] / sd_c + base_i
      z12 <- lav_uvord_z12(y = y_c, th = th_s, o1 = o1, o2 = o2,
                           offset = offset_i)
      z1 <- z12$z1
      z2 <- z12$z2
      pi_i <- pnorm(z1) - pnorm(z2)
      large_idx <- which(z2 > 1)
      if (length(large_idx) > 0L) {
        pi_i[large_idx] <- (pnorm(z2[large_idx], lower.tail = FALSE) -
          pnorm(z1[large_idx], lower.tail = FALSE))
      }
      pi_i[pi_i < .Machine$double.eps] <- .Machine$double.eps
      inv_pi <- 1 / pi_i
      p1 <- dnorm(z1)
      p2 <- dnorm(z2)

      # thresholds (direct): dz/dtau_k = y_k / sd_c
      g_tau <- rowsum.default(
        (y1 * p1 - y2 * p2) * (inv_pi / sd_c),
        group = cluster_idx, reorder = TRUE
      )

      # per-node cluster aggregates for the generic direct/chain terms
      # (note: z1/z2 include the +-100 offsets of the extreme categories,
      #  but there p1/p2 = dnorm(+-100) are exactly zero)
      s0 <- rowsum.default((p1 - p2) * inv_pi,
                           group = cluster_idx, reorder = TRUE)
      s1 <- rowsum.default((p1 * z1 - p2 * z2) * inv_pi,
                           group = cluster_idx, reorder = TRUE)
      sdev <- rowsum.default(dev_a * (p1 - p2) * inv_pi,
                             group = cluster_idx, reorder = TRUE)
      # gamma: dz/dgamma = -1/sd_c
      g_gam <- -s0 / sd_c

      pwq <- pw[, q]
      sc[, tau_idx] <- sc[, tau_idx] + pwq * g_tau

      for (p in seq_len(np)) {
        pt <- pats[[p]]
        # direct term: dz_i = (-da dev_i - a ddev)/sd_c - z_i dsd/sd_c
        direct <- (-pt$da * sdev - a_sl * pt$ddev * s0) / sd_c -
          (pt$dsd / sd_c) * s1
        # chain term through the node positions
        dgam <- dm[, p] + gh$x[q] * dsg[, p]
        sc[, par_idx[p]] <- sc[, par_idx[p]] +
          pwq * (direct + g_gam * dgam)
      }

      # slope columns (covariates)
      if (nexo_a > 0L) {
        # a-side slopes: like mu_a, but with design column z_a[, r]:
        # direct: ddev_i = -z_a[, r]; chain: d r_aj = -zbar_a[j, r]
        sza <- rowsum.default(z_a * ((p1 - p2) * inv_pi),
                              group = cluster_idx, reorder = TRUE)
        for (r in seq_len(nexo_a)) {
          direct_r <- a_sl * sza[, r] / sd_c
          dgam_r <- -c_g * njs * zbar_a[, r] / v_aj
          sc[, asl_idx[r]] <- sc[, asl_idx[r]] +
            pwq * (direct_r + g_gam * dgam_r)
        }
      }
      if (nexo_c > 0L) {
        # c-side slopes: dz_i = -z_c[, r] / sd_c (direct only)
        szc <- rowsum.default(z_c * ((p1 - p2) * inv_pi),
                              group = cluster_idx, reorder = TRUE)
        sc[, csl_idx] <- sc[, csl_idx] - pwq * szc / sd_c
      }
    }

    # the (fixed-parameter) Gaussian factor f(y_a,j): mu_a, vw_a, vb_a
    dm2ll_a <- lav_2l_gauss_dm2ll(
      cs = cs_a, mu = mu_a,
      sigma_w = matrix(vw_a, 1, 1), sigma_b = matrix(vb_a, 1, 1),
      params = list(
        list(dmu = 1, dsw = NULL, dsb = NULL),
        list(dmu = NULL, dsw = matrix(1, 1, 1), dsb = NULL),
        list(dmu = NULL, dsw = NULL, dsb = matrix(1, 1, 1))
      )
    )
    sc[, 1:3] <- sc[, 1:3] + (-0.5 * dm2ll_a)

    # ... and its a-side slope scores: rowsum(z_a * m) with m the
    # (univariate) GLS-weighted residuals
    if (nexo_a > 0L) {
      h_j <- njs * vb_a / v_aj
      m_uni <- (dev_a - (h_j * r_aj)[cl_pos]) / vw_a
      sc[, asl_idx] <- sc[, asl_idx] +
        rowsum.default(z_a * m_uni, group = cluster_idx, reorder = TRUE)
    }

    colnames(sc) <- c(
      "mu_a", "vw_a", "vb_a", paste0("tau_c", seq_len(nth)),
      "vb_c", "cw", "cb",
      if (nexo_a > 0L) paste0("a_", lav_uvreg_2l_slope_names(
        nexo_wa, nexo_ba)) else character(0L),
      if (nexo_c > 0L) paste0("c_", lav_uvreg_2l_slope_names(
        nexo_wc, nexo_bc)) else character(0L)
    )
    return(sc)
  })                     # nolint end
}

# cluster-wise scores at given values - no cache
lav_bvmix_2l_sc <- function(fit_y1 = NULL, fit_y2 = NULL,
                            cw = NULL, cb = NULL, ngh = 21L) {
  cache <- lav_bvmix_2l_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, ngh = ngh
  )
  cache$theta <- c(cw, cb)
  tmp <- lav_bvmix_2l_loglik_cache(cache = cache) # populates pw
  lav_bvmix_2l_sc_cache(cache = cache)
}

# total loglik at given values - no cache
lav_bvmix_2l_logl <- function(fit_y1 = NULL, fit_y2 = NULL,
                              cw = NULL, cb = NULL, ngh = 21L) {
  cache <- lav_bvmix_2l_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, ngh = ngh
  )
  cache$theta <- c(cw, cb)
  lav_bvmix_2l_loglik_cache(cache = cache)
}

# nlminb objective/gradient (see lav_uvbv_common.R)
lav_bvmix_2l_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_bvmix_2l_loglik_cache,
    grad_fun = lav_bvmix_2l_grad_cache
  )
}

# data.frame utilities
# Y.R. 11 April 2013

# - 10 nov 2019: * removed lav_dataframe_check_vartype(), as we can simply use
#                  sapply(lapply(frame, class), "[", 1L) (unused anyway)
#                * removed lav_dataframe_check_ordered() as we can simply use
#                  any(sapply(frame[, ov.names], inherits, "ordered"))

# construct vartable, but allow 'ordered/factor' argument to intervene
# we do NOT change the data.frame
lav_dataframe_vartable <- function(frame = NULL, ov_names = NULL,
                                   ov_names_x = NULL,
                                   ordered = NULL,
                                   factor = NULL,
                                   as_data_frame = FALSE,
                                   allow_empty_cell = FALSE) {
  if (missing(ov_names)) {
    var_names <- names(frame)
  } else {
    ov_names <- unlist(ov_names, use.names = FALSE)
    ov_names_x <- unlist(ov_names_x, use.names = FALSE)
    var_names <- unique(c(ov_names, ov_names_x))
  }
  nvar <- length(var_names)
  var_idx <- match(var_names, names(frame))


  nobs <- integer(nvar)
  type <- character(nvar)
  user <- integer(nvar)
  exo <- ifelse(var_names %in% ov_names_x, 1L, 0L)
  mean <- numeric(nvar)
  var <- numeric(nvar)
  nlev <- integer(nvar)
  lnam <- character(nvar)
  for (i in seq_len(nvar)) {
    x <- frame[[var_idx[i]]]

    type_x <- class(x)[1L]

    # correct for matrix with 1 column
    if (inherits(x, "matrix") && (is.null(dim(x)) ||
      (!is.null(dim) && ncol(x) == 1L))) {
      type_x <- "numeric"
    }

    # correct for integers
    if (inherits(x, "integer")) {
      type_x <- "numeric"
    }

    # handle the 'labelled' type from the haven package
    # - if the variable name is not in 'ordered', we assume
    #   it is numeric (for now) 11 March 2018
    # - haven >= 2.0 renamed the class to 'haven_labelled' (and
    #   'haven_labelled_spss' for SPSS user-defined missings); the old
    #   'labelled' class only matched pre-2.0 haven, so labelled numeric
    #   variables were left with type = "haven_labelled" (which, e.g.,
    #   sent lavPredict() down the numerical EBM path). Detect a labelled
    #   *numeric* variable directly instead (July 2026).
    if (is.numeric(x) &&
        inherits(x, c("labelled", "haven_labelled")) &&
        !(var_names[i] %in% ordered)) {
      type_x <- "numeric"
    }

    # handle ordered/factor
    if (!is.null(ordered) && var_names[i] %in% ordered) {
      type_x <- "ordered"
      if (allow_empty_cell) {
        if (inherits(x, 'factor')) {
          nlev[i] <- nlevels(x)
          lnam[i] <- paste(levels(x), collapse = "|")
        } else {
          nlev[i] <- max(as.numeric(x), na.rm = TRUE)
          lnam[i] <- paste(1:nlev[i], collapse = "|")
        }
      } else {
        lev <- sort(unique(x)) # we assume integers!
        nlev[i] <- length(lev)
        lnam[i] <- paste(lev, collapse = "|")
      }
      user[i] <- 1L
    } else if (!is.null(factor) && var_names[i] %in% factor) {
      type_x <- "factor"
      lev <- sort(unique(x)) # we assume integers!
      nlev[i] <- length(lev)
      lnam[i] <- paste(lev, collapse = "|")
      user[i] <- 1L
    } else {
      nlev[i] <- nlevels(x)
      lnam[i] <- paste(levels(x), collapse = "|")
    }

    type[i] <- type_x
    nobs[i] <- sum(!is.na(x))
    mean[i] <- ifelse(type_x == "numeric", mean(x, na.rm = TRUE),
      as.numeric(NA)
    )
    var[i] <- ifelse(type_x == "numeric", var(x, na.rm = TRUE),
      as.numeric(NA)
    )
  }

  var_1 <- list(
    name = var_names, idx = var_idx, nobs = nobs, type = type, exo = exo,
    user = user, mean = mean, var = var, nlev = nlev, lnam = lnam
  )

  if (as_data_frame) {
    var_1 <- as.data.frame(var_1,
      stringsAsFactors = FALSE,
      row.names = seq_along(var_1$name)
    )
    class(var_1) <- c("lavaan.data.frame", "data.frame")
  }

  var_1
}

# model objective

lav_model_objective <- function(lavmodel = NULL,
                                glist = NULL,
                                lavsamplestats = NULL,
                                lavdata = NULL,
                                lavcache = NULL) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  # shortcut for data.type == "none" or estimator == "none"
  if (lavmodel@estimator == "none" || length(lavsamplestats@cov) == 0L) {
    fx <- as.numeric(NA)
    attr(fx, "fx.group") <- rep(as.numeric(NA), lavsamplestats@ngroups)
    return(fx)
  }

  # random slopes: if the rs cache is missing (unusual call path),
  # rebuild it from the raw data
  if ((length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L) &&
      (is.null(lavcache) || is.null(lavcache[[1]]$rs))) {
    if (is.null(lavcache)) {
      lavcache <- vector("list", length = lavdata@ngroups)
    }
    for (g in seq_len(lavdata@ngroups)) {
      rs_info <- lav_mvn_cl_rs_info(lavmodel = lavmodel,
                                    lavdata = lavdata)
      rs_stats <- lav_mvn_cl_rs_stats(
        y1 = lavdata@X[[g]], lp = lavdata@Lp[[g]], rs_info = rs_info
      )
      lavcache[[g]]$rs <- list(info = rs_info, stats = rs_stats)
    }
  }

  meanstructure <- lavmodel@meanstructure
  estimator <- lavmodel@estimator
  categorical <- lavmodel@categorical
  correlation <- lavmodel@correlation
  group_w_free <- lavmodel@group.w.free
  # fixed.x <- lavmodel@fixed.x
  conditional_x <- lavmodel@conditional.x
  num_idx <- lavmodel@num.idx
  th_idx <- lavmodel@th.idx
  estimator_args <- lavmodel@estimator.args

  # partial correlation structure (GLS): a subset of the observed variables
  # is standardized to unit variance. The objective is then the canonical
  # GLS quadratic form in the partial moment space (consistent with the
  # analytic gradient), not the closed-form Bentler-Savalei expression.
  partial_cor <- correlation && any(lengths(num_idx) > 0L)

  # do we need WLS.est?
  if (estimator %in% c("ULS", "WLS", "DWLS", "NTRLS", "DLS") ||
      (estimator == "GLS" &&
        (partial_cor || conditional_x || group_w_free))) {
    lavimplied <- lav_model_implied(lavmodel, glist = glist)
    # check for COV with negative diagonal elements?
    # (note: all blocks, not groups -- multilevel has 2 blocks per group)
    for (g in seq_len(lavmodel@nblocks)) {
      cov_1 <- if (lavmodel@conditional.x) {
        lavimplied$res.cov[[g]]
      } else {
        lavimplied$cov[[g]]
      }
      d_cov <- diag(cov_1)
      if (anyNA(cov_1) || any(d_cov < 0)) {
        # return NA
        fx <- as.numeric(NA)
        attr(fx, "fx.group") <- rep(as.numeric(NA), lavsamplestats@ngroups)
        return(fx)
      }
    }
    wls_est <- lav_model_wls_est(
      lavmodel = lavmodel, glist = glist,
      lavimplied = lavimplied
    ) # ,
    # cov.x = lavsamplestats@cov.x)
    if (estimator == "NTRLS") {
      sigma_hat <- lav_model_sigma(
        lavmodel = lavmodel, glist = glist,
        extra = TRUE
      )
      mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
    }
    if (estimator == "DLS" && estimator_args$dls.GammaNT == "model") {
      sigma_hat <- lav_model_sigma(
        lavmodel = lavmodel, glist = glist,
        extra = FALSE
      )
      mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
    }
    if (lav_debug()) print(wls_est)
  } else if (estimator %in% c("ML", "GLS", "PML", "FML", "REML", "catML") &&
    lavdata@nlevels == 1L) {
    # compute moments for all groups
    # if(conditional.x) {
    #    Sigma.hat <- lav_model_cond2joint_sigma(lavmodel = lavmodel,
    #                     GLIST = GLIST, lavsamplestats = lavsamplestats,
    #                     extra = (estimator %in% c("ML", "REML","NTRLS")))
    # } else {
    sigma_hat <- lav_model_sigma(
      lavmodel = lavmodel, glist = glist,
      extra = (estimator %in% c(
        "ML", "REML",
        "NTRLS", "catML"
      ))
    )
    # }

    # if (estimator == "REML") {
    #   LAMBDA <- lav_model_lambda(lavmodel = lavmodel, GLIST = GLIST)
    # }

    # ridge?
    if (lavsamplestats@ridge > 0.0) {
      for (g in 1:lavsamplestats@ngroups) {
        diag(sigma_hat[[g]]) <- diag(sigma_hat[[g]]) +
          lavsamplestats@ridge
      }
    }
    if (lav_debug()) print(sigma_hat)

    if (meanstructure) {
      # if(conditional.x) {
      #    Mu.hat <- lav_model_cond2joint_mu(lavmodel = lavmodel, GLIST = GLIST,
      #                           lavsamplestats = lavsamplestats)
      # } else {
      mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
      # }
    }

    if (categorical) {
      th <- lav_model_th(lavmodel = lavmodel, glist = glist)
    }

    if (conditional_x) {
      pi0 <- lav_model_pi(lavmodel = lavmodel, glist = glist)
    }

    if (group_w_free) {
      gw <- lav_model_gw(lavmodel = lavmodel, glist = glist)
    }
  } else if (estimator == "MML") {
    th <- lav_model_th(lavmodel = lavmodel, glist = glist)
    mm_theta <- lav_model_theta(lavmodel = lavmodel, glist = glist)
    gw <- lav_model_gw(lavmodel = lavmodel, glist = glist)
  }

  fx <- 0.0
  fx_group <- numeric(lavsamplestats@ngroups)
  logl_group <- rep(as.numeric(NA), lavsamplestats@ngroups)

  for (g in 1:lavsamplestats@ngroups) {
    # incomplete data and fiml?
    if (lavsamplestats@missing.flag && estimator != "Bayes") {
      if (estimator == "ML" && lavdata@nlevels == 1L) {
        # FIML
        if (!attr(sigma_hat[[g]], "po")) {
          fx <- as.numeric(Inf)
          attr(fx, "fx.group") <- rep(as.numeric(Inf), lavsamplestats@ngroups)
          return(fx)
        }
        # check if h1 is defined (eg zero coverage)
        if (is.null(lavsamplestats@missing.h1[[g]]$h1)) {
          #this.h1 <- lav_mvn_mi_loglik_samp(
          #  Yp = lavsamplestats@missing[[g]],
          #  Mu = Mu.hat[[g]], Sigma = Sigma.hat[[g]],
          #  log2pi = FALSE, minus.two = TRUE) / lavsamplestats@nobs[[g]]
          #this.h1 <- this.h1 * 0.9999999 # avoid perfect fit
          this_h1 <- NULL #for now
        } else {
          this_h1 <- lavsamplestats@missing.h1[[g]]$h1
        }
        if (conditional_x) {
          group_fx <- lav_model_objective_fiml_res(
            sigma_hat = sigma_hat[[g]],
            mu_hat = mu_hat[[g]],
            pi0 = pi0[[g]],
            yp = lavsamplestats@missing[[g]],
            h1 = this_h1, n = lavsamplestats@nobs[[g]]
          )
        } else {
          group_fx <- lav_model_objective_fiml(
            sigma_hat = sigma_hat[[g]],
            mu_hat = mu_hat[[g]],
            yp = lavsamplestats@missing[[g]],
            h1 = this_h1, n = lavsamplestats@nobs[[g]]
          )
        }
      } else if (estimator == "ML" && lavdata@nlevels > 1L) {
        # FIML twolevel
        group_fx <- lav_model_objective_2l(
          lavmodel = lavmodel,
          glist = glist,
          y1 = lavdata@X[[g]],
          lp = lavdata@Lp[[g]],
          mp = lavdata@Mp[[g]],
          lavsamplestats = lavsamplestats,
          lavcache_group = if (is.null(lavcache)) NULL else lavcache[[g]],
          group = g
        )
      } else {
        lav_msg_stop(gettextf(
          "this estimator: `%s' can not be used with incomplete data and
          the missing=\"ml\" option", estimator))
      }
    } else if (estimator == "ML" || estimator == "Bayes" ||
      estimator == "catML") {
      # complete data
      # ML and friends
      if (lavdata@nlevels > 1L) {
        if (estimator %in% c("catML", "Bayes")) {
          lav_msg_stop(gettext("multilevel data not supported for estimator"),
                       estimator)
          }
        group_fx <- lav_model_objective_2l(
          lavmodel = lavmodel,
          glist = glist,
          lp = lavdata@Lp[[g]],
          mp = NULL, # complete data
          lavsamplestats = lavsamplestats,
          lavcache_group = if (is.null(lavcache)) NULL else lavcache[[g]],
          group = g
        )
      } else if (conditional_x) {
        group_fx <- lav_model_objective_ml_res(
          sigma_hat        = sigma_hat[[g]],
          mu_hat           = mu_hat[[g]],
          pi0              = pi0[[g]],
          res_cov          = lavsamplestats@res.cov[[g]],
          res_int          = lavsamplestats@res.int[[g]],
          res_slopes       = lavsamplestats@res.slopes[[g]],
          res_cov_log_det  = lavsamplestats@res.cov.log.det[[g]],
          cov_x            = lavsamplestats@cov.x[[g]],
          mean_x           = lavsamplestats@mean.x[[g]]
        )
      } else {
        group_fx <- lav_model_objective_ml(
          sigma_hat        = sigma_hat[[g]],
          mu_hat           = mu_hat[[g]],
          data_cov         = lavsamplestats@cov[[g]],
          data_mean        = lavsamplestats@mean[[g]],
          data_cov_log_det = lavsamplestats@cov.log.det[[g]],
          meanstructure    = meanstructure
        )
      }


      ### GLS #### (0.6-10: not using WLS function any longer)
    } else if (estimator == "GLS") {
      if (partial_cor || conditional_x || group_w_free) {
        # canonical GLS quadratic form; needed when the trace shortcut
        # below does not apply:
        # - partial_cor: partial moment space
        # - conditional.x: residual moment space (WLS.obs/WLS.V are in
        #   the [res.int|res.slopes, vech(res.cov)] metric, while
        #   sigma_hat/data_cov would mix residual and joint moments)
        # - group.w.free: the group weight discrepancy must be included
        group_fx <- lav_model_objective_wls(
          wls_est = wls_est[[g]],
          wls_obs = lavsamplestats@WLS.obs[[g]],
          wls_v = lavsamplestats@WLS.V[[g]]
        )
        attr(group_fx, "WLS.est") <- wls_est[[g]]
      } else {
        group_fx <- lav_model_objective_gls(
          sigma_hat        = sigma_hat[[g]],
          mu_hat           = mu_hat[[g]],
          data_cov         = lavsamplestats@cov[[g]],
          data_cov_inv     = lavsamplestats@icov[[g]],
          data_mean        = lavsamplestats@mean[[g]],
          meanstructure    = meanstructure,
          correlation      = correlation
        )
      }
    } else if (estimator == "WLS" ||
      estimator == "DLS" ||
      estimator == "NTRLS") {
      # full weight matrix
      if (estimator == "WLS") {
        wls_v <- lavsamplestats@WLS.V[[g]]
      } else if (estimator == "DLS") {
        if (estimator_args$dls.GammaNT == "sample") {
          wls_v <- lavsamplestats@WLS.V[[g]]
        } else {
          wls_v <- lav_dls_wls_v_g(
            m_cov         = sigma_hat[[g]],
            m_mean        = mu_hat[[g]],
            nacov_g       = lavsamplestats@NACOV[[g]],
            dls_a         = estimator_args$dls.a,
            x_idx         = lavsamplestats@x.idx[[g]],
            fixed_x       = lavmodel@fixed.x,
            conditional_x = lavmodel@conditional.x,
            meanstructure = lavmodel@meanstructure
          )
        }
      } else if (estimator == "NTRLS") {
        # WLS.V <- lav_samp_gamma_inverse_nt(
        #             m_icov = attr(Sigma.hat[[g]],"inv")[,,drop=FALSE],
        #             m_cov            = Sigma.hat[[g]][,,drop=FALSE],
        #             m_mean           = Mu.hat[[g]],
        #             x_idx          = c(10000,10001), ### FIXME!!!!
        #             fixed_x        = fixed.x,
        #             conditional_x  = conditional.x,
        #             meanstructure  = meanstructure,
        #             slopestructure = conditional.x)
        wls_v <- lav_mvn_info_expected(
          sigma_1 = sigma_hat[[g]],
          x_idx = lavsamplestats@x.idx[[g]],
          meanstructure = lavmodel@meanstructure
        )
        # DEBUG!!!!
        # WLS.V <- 2*WLS.V
      }

      group_fx <- lav_model_objective_wls(
        wls_est = wls_est[[g]],
        wls_obs = lavsamplestats@WLS.obs[[g]],
        wls_v = wls_v
      )
      attr(group_fx, "WLS.est") <- wls_est[[g]]
    } else if (estimator == "DWLS" || estimator == "ULS") {
      # diagonal weight matrix
      group_fx <- lav_model_objective_dwls(
        wls_est = wls_est[[g]],
        wls_obs = lavsamplestats@WLS.obs[[g]],
        wls_vd = lavsamplestats@WLS.VD[[g]]
      )
      attr(group_fx, "WLS.est") <- wls_est[[g]]
    } else if (estimator == "PML") {
      # Pairwise maximum likelihood
      if (lavdata@nlevels > 1L) {
        # group.fx <- lav_model_objective_pml.2L(lavmodel       = lavmodel,
        #                             GLIST          = GLIST,
        #                             Lp             = lavdata@Lp[[g]],
        #                             lavsamplestats = lavsamplestats,
        #                             group          = g)
        group_fx <- 0 # for now
        attr(group_fx, "logl") <- 0
      } else if (conditional_x) {
        group_fx <- lav_model_objective_pml(
          sigma_hat = sigma_hat[[g]],
          mu_hat = mu_hat[[g]],
          th = th[[g]],
          pi0 = pi0[[g]],
          th_idx = th_idx[[g]],
          num_idx = num_idx[[g]],
          x = lavdata@X[[g]],
          exo = lavdata@eXo[[g]],
          wt = lavdata@weights[[g]],
          lavcache = lavcache[[g]],
          missing = lavdata@missing
        )
      } else {
        group_fx <- lav_model_objective_pml(
          sigma_hat = sigma_hat[[g]],
          mu_hat = mu_hat[[g]],
          th = th[[g]],
          pi0 = NULL,
          th_idx = th_idx[[g]],
          num_idx = num_idx[[g]],
          x = lavdata@X[[g]],
          exo = NULL,
          wt = lavdata@weights[[g]],
          lavcache = lavcache[[g]],
          missing = lavdata@missing
        )
      }
      logl_group[g] <- attr(group_fx, "logl")
    } else if (estimator == "FML") {
      # Full maximum likelihood (underlying multivariate normal)
      group_fx <- lav_model_objective_fml(
        sigma_hat = sigma_hat[[g]],
        th = th[[g]],
        th_idx = th_idx[[g]],
        num_idx = num_idx[[g]],
        x = lavdata@X[[g]],
        lavcache = lavcache[[g]]
      )
    } else if (estimator == "MML") {
      # marginal maximum likelihood
      group_fx <- lav_model_objective_mml(
        lavmodel = lavmodel,
        glist = glist,
        mm_theta = mm_theta[[g]],
        th = th[[g]],
        group = g,
        lavdata = lavdata,
        sample_mean = lavsamplestats@mean[[g]],
        sample_mean_x = lavsamplestats@mean.x[[g]],
        lavcache = lavcache
      )
    } else if (estimator == "REML") {
      # restricted/residual maximum likelihood
      group_fx <- lav_model_objective_reml(
        sigma_hat = sigma_hat[[g]],
        mu_hat = mu_hat[[g]],
        data_cov = lavsamplestats@cov[[g]],
        data_mean = lavsamplestats@mean[[g]],
        data_cov_log_det = lavsamplestats@cov.log.det[[g]],
        meanstructure = meanstructure,
        group = g,
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata
      )
    } else {
      lav_msg_stop(gettext("unsupported estimator:"), estimator)
    }

    if (estimator %in% c("ML", "REML", "NTRLS", "catML")) {
      if (lavdata@nlevels == 1L) {
        group_fx <- 0.5 * group_fx ## FIXME
      }
    } else if (estimator == "PML" || estimator == "FML" ||
      estimator == "MML") {
      # do nothing
    } else if (estimator == "DLS") {
      if (estimator_args$dls.FtimesNminus1) {
        group_fx <- 0.5 * (lavsamplestats@nobs[[g]] - 1) /
            lavsamplestats@nobs[[g]] * group_fx
      } else {
        group_fx <- 0.5 * group_fx
      }
    } else {
      group_fx <- 0.5 * (lavsamplestats@nobs[[g]] - 1) /
          lavsamplestats@nobs[[g]] * group_fx
    }

    fx_group[g] <- group_fx
  } # g

  if (lavsamplestats@ngroups > 1) {
    ## FIXME: if group.w.free, should we use group.w or nobs???
    ##  - if we use estimated group.w, gradient changes!!!!
    ##  - but, if group models are misspecified, the group weights
    ##    will be affected too... which is unwanted (I think)
    # if(group.w.free) {
    # nobs <- unlist(GW) * lavsamplestats@ntotal
    #   nobs <- exp(unlist(GW))
    # } else {
    if (estimator == "PML") {
      # no weighting needed! (since N_g is part of the logl per group)
      fx <- sum(fx_group)
  } else if (lavdata@nlevels > 1L &&
    !(estimator %in% c("WLS", "DWLS", "ULS"))) {
    # no weighting needed! (implicit in obj, which is based on loglik)
    fx <- sum(fx_group)
    } else {
      nobs <- unlist(lavsamplestats@nobs)
      # }
      fx <- weighted.mean(fx_group, w = nobs)
    }
  } else { # single group
    fx <- fx_group[1]
  }

  # penalty for group.w + ML
  if (group_w_free && estimator %in% c(
    "ML", "MML", "FML", "PML",
    "REML", "catML"
  )) {
    # obs.prop <- unlist(lavsamplestats@group.w)
    # est.prop <- unlist(GW)
    # if(estimator %in% c("WLS", "GLS", ...) {
    #    # X2 style discrepancy measures (aka GLS/WLS!!)
    #    fx.w <- sum ( (obs.prop-est.prop)^2/est.prop )
    # } else {
    #    # G2 style discrepancy measures (aka ML)
    #    # deriv is here -2 * (obs.prop - est.prop)
    # fx.w <- sum(obs.prop * log(obs.prop/est.prop) )
    # }

    # poisson kernel
    obs_freq <- unlist(lavsamplestats@group.w) * lavsamplestats@ntotal
    est_freq <- exp(unlist(gw))
    fx_w <- -1 * sum(obs_freq * log(est_freq) - est_freq)
    # divide by N (to be consistent with the rest of lavaan)
    fx_w <- fx_w / lavsamplestats@ntotal

    fx_sat <- sum(ifelse(obs_freq > 0, obs_freq * log(obs_freq), 0) - obs_freq)
    fx_sat <- fx_sat / lavsamplestats@ntotal

    # saturated - poisson
    # fx.w <- sum(obs.freq * log(obs.freq/est.freq))
    # does not work without constraints? --> need lagrange multiplier

    fx <- fx + (fx_w + fx_sat)
  }

  fx_value <- as.numeric(fx)

  attr(fx, "fx.group") <- fx_group
  if (estimator == "PML") {
    attr(fx, "logl.group") <- logl_group
    attr(fx, "fx.pml") <- fx_value
  }

  fx
}

# utility functions needed for lav_cfa_*

# compute THETA and PSI, given lambda using either ULS or GLS
# this function assumes:
# - THETA is diagonal
# - PSI is unrestricted
# - we assume W = S^{-1}
#
# YR 17 oct 2022: - add lower/upper bounds for theta (only to compute PSI)
#                 - use 'lambda' correction to ensure PSI is positive definite
# YR 02 feb 2023: - add psi.mapping.ML argument
lav_cfa_lambda2thetapsi <- function(lambda = NULL, s = NULL, s_inv = NULL,
                                    gls = FALSE, psi_mapping_ml = FALSE,
                                    nobs = 20L) {
  mm_lambda <- as.matrix(lambda)
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)

  if (gls) {
    # see Browne, 1974 section 4 case II
    if (is.null(s_inv)) {
      m_w <- solve(s)
    } else {
      m_w <- s_inv
    }
    t_lw <- crossprod(mm_lambda, m_w)
    m <- solve(t_lw %*% mm_lambda, t_lw) # GLS mapping
    # d <- W %*% mm_lambda %*% M # symmetric
    d <- crossprod(m, t_lw)
    # theta <- solve(W*W - d*d, diag(W %*% S %*% W - d %*% S %*% d))
    theta <- try(solve(m_w * m_w - d * d, diag(m_w - d)), # because W == S^{-1}
      silent = TRUE
    )
    if (inherits(theta, "try-error")) {
      # what to do?
      lav_msg_warn(gettext(
        "problem computing THETA values; trying pace algorithm"))
      theta <- lav_efa_pace(s = s, nfactors = nfac, theta_only = TRUE)
    }
  } else {
    # see Hagglund 1982, section 4
    m <- solve(crossprod(mm_lambda), t(mm_lambda)) # ULS mapping function
    d <- mm_lambda %*% m
    theta <- try(solve(diag(nvar) - d * d, diag(s - (d %*% s %*% d))),
      silent = TRUE
    )
    if (inherits(theta, "try-error")) {
      # what to do?
      lav_msg_warn(gettext(
        "problem computing THETA values; trying pace algorithm"))
      theta <- lav_efa_pace(s = s, nfactors = nfac, theta_only = TRUE)
    }
  }
  theta_nobounds <- theta

  # ALWAYS check bounds for theta (only to compute PSI)!
  theta_bounds <- TRUE
  if (theta_bounds) {
    diag_s <- diag(s)
    # lower bound
    lower_bound <- diag_s * 0 # * 0.01
    too_small_idx <- which(theta < lower_bound)
    if (length(too_small_idx) > 0L) {
      theta[too_small_idx] <- lower_bound[too_small_idx]
    }

    # upper bound
    upper_bound <- diag_s * 1 # * 0.99
    too_large_idx <- which(theta > upper_bound)
    if (length(too_large_idx) > 0L) {
      theta[too_large_idx] <- upper_bound[too_large_idx]
    }
  }

  # psi
  diag_theta <- diag(theta, nvar)
  lambda <- try(lav_mat_sym_diff_smallest_root(s, diag_theta),
    silent = TRUE
  )
  if (inherits(lambda, "try-error")) {
    lav_msg_warn(gettext("failed to compute lambda"))
    s_min_theta <- s - diag_theta # and hope for the best
  } else {
    cutoff <- 1 + 1 / (nobs - 1)
    if (lambda < cutoff) {
      lambda_star <- lambda - 1 / (nobs - 1)
      s_min_theta <- s - lambda_star * diag_theta
    } else {
      s_min_theta <- s - diag_theta
    }
  }

  # just like local SAM
  if (psi_mapping_ml) {
    ti <- 1 / theta
    zero_theta_idx <- which(abs(theta) < 0.01) # be conservative
    if (length(zero_theta_idx) > 0L) {
      ti[zero_theta_idx] <- 1
    }
    m <- solve(t(mm_lambda) %*% diag(ti, nvar) %*% mm_lambda) %*%
      t(mm_lambda) %*% diag(ti, nvar)
    mm_psi <- m %*% s_min_theta %*% t(m) # ML
  } else {
    mm_psi <- m %*% s_min_theta %*% t(m) # ULS/GLS
  }

  # we take care of the bounds later!
  list(lambda = mm_lambda, theta = theta_nobounds, psi = mm_psi)
}

# compute PSI, given lambda and theta using either ULS, GLS, ML
# this function assumes:
# - THETA is diagonal
# - PSI is unrestricted
#
# YR 08 Mar 2023: - first version
lav_cfa_lambdatheta2psi <- function(lambda = NULL, theta = NULL, # vector!
                                    s = NULL, s_inv = NULL,
                                    mapping = "ML", nobs = 20L) {
  mm_lambda <- as.matrix(lambda)
  nvar <- nrow(mm_lambda)

  # ALWAYS check bounds for theta to compute PSI
  diag_s <- diag(s)
  # lower bound
  lower_bound <- diag_s * 0 # * 0.01
  too_small_idx <- which(theta < lower_bound)
  if (length(too_small_idx) > 0L) {
    theta[too_small_idx] <- lower_bound[too_small_idx]
  }

  # upper bound
  upper_bound <- diag_s * 1 # * 0.99
  too_large_idx <- which(theta > upper_bound)
  if (length(too_large_idx) > 0L) {
    theta[too_large_idx] <- upper_bound[too_large_idx]
  }

  # psi
  diag_theta <- diag(theta, nvar)
  lambda <- try(lav_mat_sym_diff_smallest_root(s, diag_theta),
    silent = TRUE
  )
  if (inherits(lambda, "try-error")) {
    lav_msg_warn(gettext("failed to compute lambda"))
    s_min_theta <- s - diag_theta # and hope for the best
  } else {
    cutoff <- 1 + 1 / (nobs - 1)
    if (lambda < cutoff) {
      lambda_star <- lambda - 1 / (nobs - 1)
      s_min_theta <- s - lambda_star * diag_theta
    } else {
      s_min_theta <- s - diag_theta
    }
  }

  # mapping matrix
  if (mapping == "ML") {
    ti <- 1 / theta
    zero_theta_idx <- which(abs(theta) < 0.01) # be conservative
    if (length(zero_theta_idx) > 0L) {
      ti[zero_theta_idx] <- 1
    }
    m <- solve(t(mm_lambda) %*% diag(ti, nvar) %*% mm_lambda) %*%
         t(mm_lambda) %*% diag(ti, nvar)
  } else if (mapping == "GLS") {
    if (is.null(s_inv)) {
      s_inv <- try(solve(s), silent = TRUE)
    }
    if (inherits(s_inv, "try-error")) {
      m <- tcrossprod(solve(crossprod(mm_lambda)), mm_lambda)
    } else {
      m <- solve(t(mm_lambda) %*% s_inv %*% mm_lambda) %*%
           t(mm_lambda) %*% s_inv
    }
  } else if (mapping == "ULS") {
    m <- tcrossprod(solve(crossprod(mm_lambda)), mm_lambda)
  }

  # compute PSI
  mm_psi <- m %*% s_min_theta %*% t(m)

  mm_psi
}

# compute theta elements for a 1-factor model
lav_cfa_theta_spearman <- function(s, bounds = "wide") {
  p <- ncol(s)
  out <- numeric(p)
  r <- cov2cor(s)
  for (p_idx in seq_len(p)) {
    x <- r[, p_idx][-p_idx]
    aa <- lav_mat_vech(tcrossprod(x), diagonal = FALSE)
    ss <- lav_mat_vech(r[-p_idx, -p_idx, drop = FALSE], diagonal = FALSE)
    h2 <- mean(aa / ss) # communaliteit
    if (bounds == "standard") {
      h2[h2 < 0] <- 0
      h2[h2 > 1] <- 1
    } else if (bounds == "wide") {
      h2[h2 < -0.05] <- -0.05 # corresponds to lower bound ov.var "wide"
      h2[h2 > +1.20] <- +1.20 # corresponds to upper bound ov.var "wide"
    }
    out[p_idx] <- (1 - h2) * s[p_idx, p_idx]
  }
  out
}

# the bivariate two-level Gaussian model, univariate parameters fixed
#
#     (y_a, y_b)_ij = mu + b_j + e_ij
#     b_j ~ N(0, Sigma_b),  e_ij ~ N(0, Sigma_w)
#
#     Sigma_w = | vw_a  cw   |     Sigma_b = | vb_a  cb   |
#               | cw    vw_b |               | cb    vb_b |
#
# YR 2026 (two-level WLS, phase 2)
#
# This is stage 2 of the stage-wise unrestricted two-level model for a
# pair of continuous variables: the univariate parameters (mu, vw, vb of
# both variables) are held fixed at their stage-1 values, and only the
# two covariances theta = c(cw, cb) are estimated ('twostep', cfr.
# lav_bvreg.R for the single-level case; note that here a pair has TWO
# free parameters).
#
# lav_bvreg_2l_sc() returns the cluster-wise scores with respect to ALL
# parameters (both fixed univariate and the two covariances), as needed
# for the stage-wise sandwich (A21 blocks).

lav_bvreg_2l_cov_twostep_fit <- function(fit_y1 = NULL,
                                         fit_y2 = NULL,
                                         optim_method = "nlminb1",
                                         control = list(),
                                         y1_name = NULL, y2_name = NULL,
                                         output = "theta") {
  # create cache environment
  cache <- lav_bvreg_2l_init_cache(fit_y1 = fit_y1, fit_y2 = fit_y2)

  # degenerate between level? (one of the between variances ~ 0)
  cb_bound <- 0.999 * sqrt(cache$vb1 * cache$vb2)
  cw_bound <- 0.999 * sqrt(cache$vw1 * cache$vw2)
  cb_fixed <- (cb_bound < 1e-08)

  # optim
  minfns <- lav_uvbv_optim_fns(optim_method, lav_bvreg_2l_min_fns())

  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  if (cb_fixed) {
    cache$theta[2] <- 0
    lower <- c(-cw_bound, 0)
    upper <- c(+cw_bound, 0)
  } else {
    lower <- c(-cw_bound, -cb_bound)
    upper <- c(+cw_bound, +cb_bound)
  }

  # try 1
  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb, lower = lower, upper = upper,
    cache = cache
  )

  # try 2: start from zero
  if (optim$convergence != 0L) {
    optim <- nlminb(
      start = c(0, 0), objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control_nlminb, lower = lower, upper = upper,
      cache = cache
    )
  }

  if (optim$convergence != 0L) {
    if (!is.null(y1_name) && !is.null(y2_name)) {
      lav_msg_warn(gettextf(
        "estimation of the two-level covariances did not converge for
        variables %1$s and %2$s", y1_name, y2_name))
    } else {
      lav_msg_warn(gettext(
        "estimation of the two-level covariances did not always converge"))
    }
  }

  if (output == "cache") {
    cache$theta <- optim$par
    return(cache)
  }

  # theta = c(cw, cb)
  optim$par
}

# prepare cache environment
# fit_y1/fit_y2: lav_uvreg_2l_fit() objects of the two variables
#
# with covariates (fit$nexo > 0) the pair likelihood is evaluated on the
# RESIDUALS e = y - z theta_m (all mean parameters fixed at their stage-1
# values): the pair model is then the no-covariate model with mu = 0
lav_bvreg_2l_init_cache <- function(fit_y1 = NULL,
                                    fit_y2 = NULL,
                                    parent = parent.frame()) {
  stopifnot(length(fit_y1$y) == length(fit_y2$y))
  cluster_idx <- fit_y1$cluster_idx

  # fixed univariate parameters
  mu1 <- fit_y1$theta[1]; vw1 <- fit_y1$theta[2]; vb1 <- fit_y1$theta[3]
  mu2 <- fit_y2$theta[1]; vw2 <- fit_y2$theta[2]; vb2 <- fit_y2$theta[3]

  # residualize (covariates)
  e1 <- fit_y1$y
  if (fit_y1$nexo > 0L) {
    theta_m1 <- fit_y1$theta[c(1L, 3L + seq_len(fit_y1$nexo))]
    e1 <- fit_y1$y - drop(fit_y1$z %*% theta_m1)
    mu1 <- 0
  }
  e2 <- fit_y2$y
  if (fit_y2$nexo > 0L) {
    theta_m2 <- fit_y2$theta[c(1L, 3L + seq_len(fit_y2$nexo))]
    e2 <- fit_y2$y - drop(fit_y2$z %*% theta_m2)
    mu2 <- 0
  }

  # pair sufficient statistics
  cs <- lav_2l_cluster_stats(
    y = cbind(e1, e2),
    cluster_idx = cluster_idx
  )

  # ANOVA-type starting values for (cw, cb)
  nobs <- cs$nobs
  nclusters <- cs$nclusters
  njs <- cs$njs
  # within cross-products: wvech columns are (11, 21, 22)
  cw0 <- cs$wtot[2] / (nobs - nclusters)
  csb <- sum(njs * (cs$ybar[, 1] - mean(e1)) *
               (cs$ybar[, 2] - mean(e2))) / (nclusters - 1)
  n0 <- (nobs - sum(njs^2) / nobs) / (nclusters - 1)
  cb0 <- (csb - cw0) / n0
  # keep the starting values admissible
  cw_max <- 0.95 * sqrt(vw1 * vw2)
  cb_max <- 0.95 * sqrt(vb1 * vb2)
  cw0 <- max(min(cw0, cw_max), -cw_max)
  cb0 <- max(min(cb0, cb_max), -cb_max)
  if (!is.finite(cb0)) {
    cb0 <- 0
  }

  theta <- c(cw0, cb0)

  list2env(
    list(
      cs = cs, n = nobs, cluster_idx = cluster_idx,
      e = cbind(e1, e2),
      z1 = fit_y1$z, z2 = fit_y2$z,
      nexo1 = fit_y1$nexo, nexo2 = fit_y2$nexo,
      nexo_w1 = fit_y1$nexo_w, nexo_b1 = fit_y1$nexo_b,
      nexo_w2 = fit_y2$nexo_w, nexo_b2 = fit_y2$nexo_b,
      mu1 = mu1, vw1 = vw1, vb1 = vb1,
      mu2 = mu2, vw2 = vw2, vb2 = vb2,
      theta = theta
    ),
    parent = parent
  )
}

lav_bvreg_2l_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    cw <- theta[1]
    cb <- theta[2]
    sigma_w <- matrix(c(vw1, cw, cw, vw2), 2, 2)
    sigma_b <- matrix(c(vb1, cb, cb, vb2), 2, 2)
    m2ll_j <- lav_2l_gauss_m2ll(
      cs = cs, mu = c(mu1, mu2),
      sigma_w = sigma_w, sigma_b = sigma_b
    )
    loglik <- -0.5 * sum(m2ll_j)
    return(loglik)
  })                     # nolint end
}

# the (dmu, dSw, dSb) patterns; order:
#   1 mu_y1, 2 wvar_y1, 3 bvar_y1, 4 mu_y2, 5 wvar_y2, 6 bvar_y2,
#   7 wcov, 8 bcov
lav_bvreg_2l_params <- function() {
  e11 <- matrix(c(1, 0, 0, 0), 2, 2)
  e22 <- matrix(c(0, 0, 0, 1), 2, 2)
  e12 <- matrix(c(0, 1, 1, 0), 2, 2)
  list(
    mu_y1 = list(dmu = c(1, 0), dsw = NULL, dsb = NULL),
    wvar_y1 = list(dmu = NULL, dsw = e11, dsb = NULL),
    bvar_y1 = list(dmu = NULL, dsw = NULL, dsb = e11),
    mu_y2 = list(dmu = c(0, 1), dsw = NULL, dsb = NULL),
    wvar_y2 = list(dmu = NULL, dsw = e22, dsb = NULL),
    bvar_y2 = list(dmu = NULL, dsw = NULL, dsb = e22),
    wcov = list(dmu = NULL, dsw = e12, dsb = NULL),
    bcov = list(dmu = NULL, dsw = NULL, dsb = e12)
  )
}

lav_bvreg_2l_grad_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    cw <- theta[1]
    cb <- theta[2]
    sigma_w <- matrix(c(vw1, cw, cw, vw2), 2, 2)
    sigma_b <- matrix(c(vb1, cb, cb, vb2), 2, 2)
    dm2ll <- lav_2l_gauss_dm2ll(
      cs = cs, mu = c(mu1, mu2),
      sigma_w = sigma_w, sigma_b = sigma_b,
      params = lav_bvreg_2l_params()[c("wcov", "bcov")]
    )
    return(-0.5 * colSums(dm2ll))
  })                     # nolint end
}

# nlminb objective/gradient (see lav_uvbv_common.R)
lav_bvreg_2l_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_bvreg_2l_loglik_cache,
    grad_fun = lav_bvreg_2l_grad_cache
  )
}

# cluster-wise scores (of the loglik) w.r.t. ALL parameters
# (mu_y1, wvar_y1, bvar_y1, mu_y2, wvar_y2, bvar_y2, wcov, bcov
#  [, y1 slopes, y2 slopes])
lav_bvreg_2l_sc <- function(fit_y1 = NULL, fit_y2 = NULL,
                            cw = NULL, cb = NULL) {
  cache <- lav_bvreg_2l_init_cache(fit_y1 = fit_y1, fit_y2 = fit_y2)
  sigma_w <- matrix(c(cache$vw1, cw, cw, cache$vw2), 2, 2)
  sigma_b <- matrix(c(cache$vb1, cb, cb, cache$vb2), 2, 2)
  dm2ll <- lav_2l_gauss_dm2ll(
    cs = cache$cs, mu = c(cache$mu1, cache$mu2),
    sigma_w = sigma_w, sigma_b = sigma_b,
    params = lav_bvreg_2l_params()
  )
  sc <- -0.5 * dm2ll
  colnames(sc) <- names(lav_bvreg_2l_params())

  # slope columns (covariates): rowsum(z * m) with m the GLS-weighted
  # residuals
  if (cache$nexo1 > 0L || cache$nexo2 > 0L) {
    m <- lav_2l_gauss_mresid(
      e = cache$e, cluster_idx = cache$cluster_idx,
      sigma_w = sigma_w, sigma_b = sigma_b
    )
    if (cache$nexo1 > 0L) {
      sc_sl1 <- rowsum.default(
        cache$z1[, -1, drop = FALSE] * m[, 1],
        group = cache$cluster_idx, reorder = TRUE
      )
      colnames(sc_sl1) <- paste0(
        "y1_", lav_uvreg_2l_slope_names(cache$nexo_w1, cache$nexo_b1)
      )
      sc <- cbind(sc, sc_sl1)
    }
    if (cache$nexo2 > 0L) {
      sc_sl2 <- rowsum.default(
        cache$z2[, -1, drop = FALSE] * m[, 2],
        group = cache$cluster_idx, reorder = TRUE
      )
      colnames(sc_sl2) <- paste0(
        "y2_", lav_uvreg_2l_slope_names(cache$nexo_w2, cache$nexo_b2)
      )
      sc <- cbind(sc, sc_sl2)
    }
  }
  sc
}

# total loglik at given values - no cache
lav_bvreg_2l_logl <- function(fit_y1 = NULL, fit_y2 = NULL,
                              cw = NULL, cb = NULL) {
  cache <- lav_bvreg_2l_init_cache(fit_y1 = fit_y1, fit_y2 = fit_y2)
  cache$theta <- c(cw, cb)
  lav_bvreg_2l_loglik_cache(cache = cache)
}

lav_step14_test <- function(lavoptions = NULL,
                                   lavmodel = NULL,
                                   lavsamplestats = NULL,
                                   lavdata = NULL,
                                   lavpartable = NULL,
                                   lavcache = NULL,
                                   lavimplied = NULL,
                                   lavh1 = NULL,
                                   x = NULL,
                                   vcov = NULL,
                                   lavloglik = NULL) {
  # # # # # # # # # # #
  # #  14. lavtest # #
  # # # # # # # # # # #

  # if lavoptions$test != "none" and x converged
  #   compute lavtest via lav_model_test(...)
  # else
  #   lavtest <- list(list(test = "none", stat = NA,
  #                        stat.group = rep(NA, lavdata@ngroups),
  #                        df = NA, refdistr = "unknown", pvalue = NA))
  lavtest <- NULL
  if (!(length(lavoptions$test) == 1L && lavoptions$test == "none") &&
    attr(x, "converged")) {
    if (lav_verbose()) {
      cat("computing TEST for test(s) =", lavoptions$test, "...")
    }
    lavtest <- lav_model_test(
      lavmodel = lavmodel,
      lavpartable = lavpartable,
      lavsamplestats = lavsamplestats,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavoptions = lavoptions,
      x = x,
      vcov_1 = vcov,
      lavdata = lavdata,
      lavcache = lavcache,
      lavloglik = lavloglik
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }
  } else {
    lavtest <- list(list(
      test = "none", stat = NA,
      stat.group = rep(NA, lavdata@ngroups), df = NA,
      refdistr = "unknown", pvalue = NA
    ))
  }

  lavtest
}

lav_step14_fit <- function(lavpartable = NULL,
                                  lavmodel = NULL,
                                  lavimplied = NULL,
                                  x = NULL,
                                  vcov = NULL,
                                  lavtest = NULL) {
  # # # # # # # # # # # #
  # #  14bis. lavfit  # # -> remove if the offending packages are fixed!!
  # # # # # # # # # # # #

  lavfit <- lav_model_fit(
    lavpartable = lavpartable,
    lavmodel = lavmodel,
    lavimplied = lavimplied,
    x = x,
    vcov_1 = vcov,
    test = lavtest
  )
  # lavfit <- new("Fit")

  lavfit
}

# varcov solution is:
# theta_2 = solve(t(Delta2) %*% W2 %*% Delta2) %*% t(Delta2) %*% W2 %*% s_vech

# we need to compute the Jacobian of theta_2 wrt the elements of s_vech
#
# four options:
# - ULS
# - GLS
# - 2RLS
# - RLS

# based on lav_utils_wls_linearization(), but only for ULS and H
lav_sem_miiv_utils_jacb_uls <- function(sample_cov = NULL, delta2 = NULL) {
  nvar <- nrow(sample_cov)
  pstar <- nvar * (nvar + 1L) / 2L

  jac_cov <- delta2
  # nc <- ncol(jac_cov)
  meanstructure <- FALSE
  if (nrow(delta2) > pstar) {
    meanstructure <- TRUE
  }

  # split delta2/svec
  if (meanstructure) {
    mean_idx <- seq_len(nvar)
    jac_mean <- delta2[mean_idx, , drop = FALSE]
    jac_cov <- delta2[-mean_idx, , drop = FALSE]
  }

  # vech
  w <- rep(1.0, pstar)
  w[lav_mat_diagh_idx(nvar)] <- 0.5

  q_cov <- w * jac_cov
  a <- crossprod(jac_cov, q_cov)
  if (meanstructure) {
    a <- a + crossprod(jac_mean)
  }

  # tQ only needed if H is requested
  t_q <- t(q_cov)
  if (meanstructure) t_q <- cbind(t(jac_mean), t_q)

  r <- chol(a)
  ainv  <- chol2inv(r)
  h <- ainv %*% t_q

  h
}

# this is the GLS version where W2 = t(D) %*% (S.inv %x% S.inv) %*% D
lav_sem_miiv_utils_jacb_gls <- function(sample_cov = NULL, delta2 = NULL,
                                        theta2 = NULL) {
  nvar  <- nrow(sample_cov)
  pstar <- nvar * (nvar + 1L) / 2L
  s_vech <- lav_mat_vech(sample_cov)

  # vech weights
  w <- rep(1.0, pstar)
  w[lav_mat_diagh_idx(nvar)] <- 0.5

  # only covariance block
  meanstructure_flag <- FALSE
  if (nrow(delta2) > pstar) {
    meanstructure_flag <- TRUE
    delta2 <- delta2[-seq_len(nvar), , drop = FALSE]
  }

  nc <- ncol(delta2)
  c_s <- chol(sample_cov)
  si <- chol2inv(c_s)

  # q_cov[,j] = w * vech(Si %*% Vj %*% Si),  Vj = vech_inv(delta2[,j])
  # equivalent to (1/2) W2 %*% delta2
  q_cov <- matrix(0.0, pstar, nc)
  for (j in seq_len(nc)) {
    v_j <- lav_mat_vech_rev(delta2[, j])
    si_vj <- si %*% v_j
    q_cov[, j] <- w * lav_mat_vech(si_vj %*% si)
  }

  # M_half = (1/2) M = t(delta2) %*% q_cov
  m_half <- crossprod(delta2, q_cov)
  b_half <- drop(crossprod(q_cov, s_vech))

  # theta2: GLS estimate
  # (scaling cancels: M_half^{-1} b_half = M^{-1} t(delta2) W2 s_vech)
  r <- chol(m_half)
  if (is.null(theta2)) {
    theta2 <- backsolve(r, forwardsolve(t(r), b_half))
  }

  # error term and GLS correction q_mat = Si %*% e_mat %*% Si
  e_mat <- lav_mat_vech_rev(s_vech - delta2 %*% theta2)
  q_mat <- si %*% e_mat %*% si

  # correction_cov[,j] = w * vech(Tj + t(Tj)),  Tj = Si %*% Vj %*% q_mat
  # equivalent to (1/2) correction %*% delta2
  correction_cov <- matrix(0.0, pstar, nc)
  for (j in seq_len(nc)) {
    v_j              <- lav_mat_vech_rev(delta2[, j])
    t_j              <- si %*% v_j %*% q_mat
    correction_cov[, j] <- w * lav_mat_vech(t_j + t(t_j))
  }

  # out = M^{-1} %*% t(delta2) %*% (W2 - correction)
  #     = M_half^{-1} %*% t(q_cov - correction_cov)
  out <- backsolve(r, forwardsolve(t(r), t(q_cov - correction_cov)))

  # meanstructure: prepend zero block for mean parameters
  if (meanstructure_flag) {
    out <- cbind(matrix(0.0, nrow = nrow(out), ncol = nvar), out)
  }

  out
}


# 2RLS solution, where W2 = t(D) %*% (Sigma.inv %x% Sigma.inv) %*% D
# and Sigma is based on the ULS estimate of theta2
lav_sem_miiv_utils_jacb_2rls <- function(sample_cov = NULL, delta2 = NULL,
                                         theta2 = NULL) {

  nvar  <- nrow(sample_cov)
  pstar <- nvar * (nvar + 1L) / 2L
  s_vech <- lav_mat_vech(sample_cov)

  # vech weights: 1 for off-diagonal positions, 0.5 for diagonal positions
  w <- rep(1.0, pstar)
  w[lav_mat_diagh_idx(nvar)] <- 0.5

  # only covariance block
  meanstructure_flag <- FALSE
  if (nrow(delta2) > pstar) {
    meanstructure_flag <- TRUE
    delta2 <- delta2[-seq_len(nvar), , drop = FALSE]
  }
  nc <- ncol(delta2)

  # step 1: ULS estimate of theta_uls
  q_uls <- w * delta2
  m_uls <- crossprod(delta2, q_uls)
  r_uls <- chol(m_uls)
  theta_uls <-
    backsolve(r_uls, forwardsolve(t(r_uls), drop(crossprod(q_uls, s_vech))))

  # sigma to be used for GLS estimate
  sigma <- lav_mat_vech_rev(delta2 %*% theta_uls)
  c_si <- chol(sigma)
  sigma_inv <- chol2inv(c_si)

  # step 2: GLS estimate of theta using sigma_inv
  # q_cov[,j] = w * vech(Si Vj Si)  =  (1/2) w2 delta2[,j]
  q_cov <- matrix(0.0, pstar, nc)
  for (j in seq_len(nc)) {
    v_j <- lav_mat_vech_rev(delta2[, j])
    q_cov[, j] <- w * lav_mat_vech(sigma_inv %*% v_j %*% sigma_inv)
  }
  m_sigma_half <- crossprod(delta2, q_cov)
  r_sigma <- chol(m_sigma_half)
  if (is.null(theta2)) {
    theta2 <- backsolve(r_sigma,
      forwardsolve(t(r_sigma), drop(crossprod(q_cov, s_vech))))
  }

  # step 3: correction matrix (as column operations, no c_mat formed)
  # c_cov[,j] = w * vech(Tj + t(Tj)),  Tj = Si Vj q_mat
  #           = (1/2) c_mat delta2[,j]
  e_mat <- lav_mat_vech_rev(s_vech - delta2 %*% theta2)
  q_mat <- sigma_inv %*% e_mat %*% sigma_inv

  c_cov <- matrix(0.0, pstar, nc)
  for (j in seq_len(nc)) {
    v_j <- lav_mat_vech_rev(delta2[, j])
    t_j <- sigma_inv %*% v_j %*% q_mat
    c_cov[, j] <- w * lav_mat_vech(t_j + t(t_j))
  }

  # step 4: Jacobian
  #   out = M_sigma_half^{-1} (t(q_cov) - t(c_cov) delta2 m_uls^{-1} t(q_uls))
  # build rhs = t(q_cov) - t(c_cov) delta2 m_uls^{-1} t(q_uls)  (nc x pstar)
  # Step A: CtD = t(c_cov) delta2  (nc x nc)
  ct_d  <- crossprod(c_cov, delta2)
  mi_ct_d <- backsolve(r_uls, forwardsolve(t(r_uls), t(ct_d)))
  # Step B: rhs = t(q_cov) - MiCtD^T t(q_uls)  (nc x pstar)
  rhs <- t(q_cov) - t(mi_ct_d) %*% t(q_uls)
  # Step C: out = M_sigma_half^{-1} rhs
  out <- backsolve(r_sigma, forwardsolve(t(r_sigma), rhs))

  # meanstructure: prepend zero block for mean parameters
  if (meanstructure_flag) {
    out <- cbind(matrix(0.0, nrow = nrow(out), ncol = nvar), out)
  }

  out
}

# RLS solution
lav_sem_miiv_utils_jacb_rls <- function(sample_cov = NULL, delta2 = NULL,
                                        theta2 = NULL) {
  nvar  <- nrow(sample_cov)
  pstar <- nvar * (nvar + 1L) / 2L
  s_vech <- lav_mat_vech(sample_cov)

  # vech weights: 0.5 for diagonal positions, 1 for off-diagonal positions
  w <- rep(1.0, pstar)
  w[lav_mat_diagh_idx(nvar)] <- 0.5

  # only covariance block
  meanstructure_flag <- FALSE
  if (nrow(delta2) > pstar) {
    meanstructure_flag <- TRUE
    delta2 <- delta2[-seq_len(nvar), , drop = FALSE]
  }

  nc <- ncol(delta2)

  # Pre-compute vech_reverse of each column of delta2 once; used in the
  # final block (and in the loop when theta2 is not supplied).
  v_list <- vector("list", nc)
  for (j in seq_len(nc)) v_list[[j]] <- lav_mat_vech_rev(delta2[, j])

  q_cov <- matrix(0.0, pstar, nc)

  # using final theta2 estimate
  new_sigma <- lav_mat_vech_rev(delta2 %*% theta2)
  c_si       <- chol(new_sigma)
  sigma_inv <- chol2inv(c_si)
  for (j in seq_len(nc)) {
    q_cov[, j] <- w * lav_mat_vech(sigma_inv %*% v_list[[j]] %*% sigma_inv)
  }
  m_mat <- crossprod(delta2, q_cov)
  r_mat <- chol(m_mat)

  # correction columns: c_cov[,j] = w * vech(Tj + t(Tj)),  Tj = Si Vj q_mat
  # where q_mat = Si e_mat Si.
  # JW_theta2 = -c_cov  (dW_e2_from_v(delta2[,j]) = -c_cov[,j], derived above)
  e_mat <- lav_mat_vech_rev(s_vech - delta2 %*% theta2)
  q_mat <- sigma_inv %*% e_mat %*% sigma_inv
  c_cov <- matrix(0.0, pstar, nc)
  for (j in seq_len(nc)) {
    t_j <- sigma_inv %*% v_list[[j]] %*% q_mat
    c_cov[, j] <- w * lav_mat_vech(t_j + t(t_j))
  }

  # Jacobian:
  #   A           = m_mat_inv t(delta2) JW_theta2
  #               = -solve(m_mat, crossprod(delta2, c_cov))
  #   IminusA_inv = solve(I - A)
  #   out         = IminusA_inv %*% solve(m_mat, t(q_cov))
  dt_c <- crossprod(delta2, c_cov)
  a <- -backsolve(r_mat, forwardsolve(t(r_mat), dt_c))
  iminus_a_inv <- solve(diag(nc) - a)
  minv_t_q <- backsolve(r_mat, forwardsolve(t(r_mat), t(q_cov)))
  out     <- iminus_a_inv %*% minv_t_q

  # meanstructure: prepend zero block for mean parameters
  if (meanstructure_flag) {
    out <- cbind(matrix(0.0, nrow = nrow(out), ncol = nvar), out)
  }

  out
}


# we also need to compute the Jacobian of theta_2 wrt the elements theta_1
#
# four options:
# - ULS
# - GLS
# - 2RLS
# - RLS

# ULS and GLS: only delta2 depends on theta1
lav_sem_miiv_utils_jaca_uls_gls <- function(lavmodel = NULL,  # nolint
                                            lavpartable = NULL,
                                            lavh1 = NULL,
                                            free_directed_idx = integer(0L),
                                            free_undirected_idx = integer(0L),
                                            iv_varcov_method = "ULS") {
  nblocks <- lavmodel@nblocks

  # delta across all blocks
  delta_block <- lav_model_delta(lavmodel = lavmodel)

  # augment lavpartable to include model matrices and row/col indices
  mm_info <- lav_lisrel(lavpartable, target = NULL, extra = FALSE)

  jac_a <- matrix(0, length(free_undirected_idx), length(free_directed_idx))
  for (b in seq_len(nblocks)) {
    # s_vech
    sample_cov <- lavh1$implied$cov[[b]]
    nvar <- nrow(sample_cov)
    s_vech <- lav_mat_vech(sample_cov)

    # delta2
    delta2 <- delta_block[[b]][, free_undirected_idx, drop = FALSE]
    if (lavmodel@meanstructure) {
      delta2 <- delta2[-seq_len(nvar), , drop = FALSE]
    }

    # w2
    if (iv_varcov_method == "GLS") {
      s_inv <- solve(sample_cov)
    } else {
      s_inv <- diag(1, nrow = nrow(sample_cov))
    }
    w2_22 <- 0.5 * lav_mat_dup_pre_post(s_inv %x% s_inv)
    w2 <- w2_22

    # MLIST
    mm_in_group <- seq_len(lavmodel@nmat[b]) + cumsum(c(0, lavmodel@nmat))[b]
    mlist <- lavmodel@GLIST[mm_in_group]
    if (!is.null(mlist$beta)) {
      mlist$IB.inv <- lav_lisrel_ibinv(mlist)
    }

    m_mat <- t(delta2) %*% w2 %*% delta2
    theta2 <- solve(m_mat, t(delta2) %*% w2 %*% s_vech)
    m_inv <- solve(m_mat)
    w2e <- w2 %*% (s_vech - delta2 %*% theta2)
    delta2tw2 <- t(delta2) %*% w2

    # container for Jacobian for this block
    for (k in seq_along(free_directed_idx)) {
      d_idx <- match(free_directed_idx[k], lavpartable$free)[1]
      # correct block?
      if (lavpartable$block[d_idx] != b) {
        next
      }
      d_mat <- mm_info$mat[d_idx]
      d_row <- mm_info$row[d_idx]
      d_col <- mm_info$col[d_idx]
      d_deltak <- matrix(0, nrow(delta2), ncol(delta2))
      # we only need to fill the columns of dDeltak that correspond to the
      # psi elements in free.undirected.idx
      for (i in seq_along(free_undirected_idx)) {
        un_idx <- match(free_undirected_idx[i], lavpartable$free)[1]
        un_mat <- mm_info$mat[un_idx]
        if (un_mat == "theta") {
          next
        }
        un_row <- mm_info$row[un_idx]
        un_col <- mm_info$col[un_idx]

        if (d_mat == "lambda") {
          tmp <- lav_lisrel_d2sigma_lambda_psi(
            mlist = mlist,
            i = d_row, j = d_col, k = un_row, l = un_col
          )
        } else if (d_mat == "beta") {
          tmp <- lav_lisrel_d2sigma_beta_psi(
            mlist = mlist,
            i = d_row, j = d_col, k = un_row, l = un_col
          )
        }
        d_deltak[, i] <- lav_mat_vech(tmp)
      }
      jac_a[, k] <-
        m_inv %*% (t(d_deltak) %*% w2e - delta2tw2 %*% d_deltak %*% theta2)
    }
  }
  jac_a
}

# 2RLS
# - first step: only Delta_2 depends on theta1
# - second step: W_2 also depends on theta1!
lav_sem_miiv_utils_jaca_2rls <- function(lavmodel = NULL,
                                         lavpartable = NULL,
                                         lavh1 = NULL,
                                         free_directed_idx = integer(0L),
                                         free_undirected_idx = integer(0L)) {
  nblocks <- lavmodel@nblocks

  # delta across all blocks
  delta_block <- lav_model_delta(lavmodel = lavmodel)

  # augment lavpartable to include model matrices and row/col indices
  mm_info <- lav_lisrel(lavpartable, target = NULL, extra = FALSE)

  # first get jac_a for ULS (step 1)
  jac_uls <- lav_sem_miiv_utils_jaca_uls_gls(
    lavmodel = lavmodel,
    lavpartable = lavpartable, lavh1 = lavh1,
    free_directed_idx = free_directed_idx,
    free_undirected_idx = free_undirected_idx, iv_varcov_method = "ULS"
  )

  jac_a <- matrix(0, length(free_undirected_idx), length(free_directed_idx))
  for (b in seq_len(nblocks)) {
    # s_vech
    sample_cov <- lavh1$implied$cov[[b]]
    nvar <- nrow(sample_cov)
    s_vech <- lav_mat_vech(sample_cov)

    # delta2
    delta2 <- delta_block[[b]][, free_undirected_idx, drop = FALSE]
    if (lavmodel@meanstructure) {
      delta2 <- delta2[-seq_len(nvar), , drop = FALSE]
    }
    # w2
    s_inv <- diag(1, nrow = nrow(sample_cov))
    w2_uls <- 0.5 * lav_mat_dup_pre_post(s_inv %x% s_inv)

    # MLIST
    mm_in_group <- seq_len(lavmodel@nmat[b]) + cumsum(c(0, lavmodel@nmat))[b]
    mlist <- lavmodel@GLIST[mm_in_group]
    if (!is.null(mlist$beta)) {
      mlist$IB.inv <- lav_lisrel_ibinv(mlist)
    }

    # step 1
    m_uls <- t(delta2) %*% w2_uls %*% delta2
    m_uls_inv <- solve(m_uls)
    theta_uls <- drop(m_uls_inv %*% t(delta2) %*% w2_uls %*% s_vech)

    # step 2
    new_sigma <- lav_mat_vech_rev(delta2 %*% theta_uls)
    sigma_inv <- solve(new_sigma)
    w2 <- 0.5 * lav_mat_dup_pre_post(sigma_inv %x% sigma_inv)
    m_mat <- t(delta2) %*% w2 %*% delta2
    m_mat_inv <- solve(m_mat)
    theta2 <- drop(m_mat_inv %*% t(delta2) %*% w2 %*% s_vech)
    e2 <- s_vech - delta2 %*% theta2
    e_mat <- lav_mat_vech_rev(e2)

    # pre-compute
    minv_delta2t <- m_mat_inv %*% t(delta2)
    w2_e2 <- w2 %*% e2
    delta2t_w2 <- t(delta2) %*% w2
    sigma_inv_e <- sigma_inv %*% e_mat

    for (k in seq_along(free_directed_idx)) {
      d_idx <- match(free_directed_idx[k], lavpartable$free)[1]
      # correct block?
      if (lavpartable$block[d_idx] != b) {
        next
      }
      d_mat <- mm_info$mat[d_idx]
      d_row <- mm_info$row[d_idx]
      d_col <- mm_info$col[d_idx]
      d_deltak <- matrix(0, nrow(delta2), ncol(delta2))
      # we only need to fill the columns of dDeltak that correspond to the
      # psi elements in free.undirected.idx
      for (i in seq_along(free_undirected_idx)) {
        un_idx <- match(free_undirected_idx[i], lavpartable$free)[1]
        un_mat <- mm_info$mat[un_idx]
        if (un_mat == "theta") {
          next
        }
        un_row <- mm_info$row[un_idx]
        un_col <- mm_info$col[un_idx]

        if (d_mat == "lambda") {
          tmp <- lav_lisrel_d2sigma_lambda_psi(
            mlist = mlist,
            i = d_row, j = d_col, k = un_row, l = un_col
          )
        } else if (d_mat == "beta") {
          tmp <- lav_lisrel_d2sigma_beta_psi(
            mlist = mlist,
            i = d_row, j = d_col, k = un_row, l = un_col
          )
        }
        d_deltak[, i] <- lav_mat_vech(tmp)
      }

      # d vech(Sigma)/d theta1[k]: product rule on unvech(Delta2 theta_uls)
      v_k <- d_deltak %*% theta_uls + delta2 %*% jac_uls[, k]
      d_sigma_k <- lav_mat_vech_rev(v_k)

      # d_sigma_inv = -sigma.inv dSigma_k sigma.inv
      d_sigma_inv <- -sigma_inv %*% d_sigma_k %*% sigma_inv

      # the tricky one:
      # (dw2_k) e2 = (1/2)
      # D^T vec(sigma_inv e_mat dsigma_inv_k + dsigma_invk e_mat sigma_inv)
      sym_part <- sigma_inv_e %*% d_sigma_inv + d_sigma_inv %*% t(sigma_inv_e)
      dw2_e2 <-
        0.5 * drop(lav_mat_dup_pre(as.matrix(as.vector(sym_part))))

      # term 1: direct delta2 variation in final WLS
      term1 <-
        m_mat_inv %*% (t(d_deltak) %*% w2_e2 -
                                           delta2t_w2 %*% d_deltak %*% theta2)

      # term 2: W2 variation via Sigma = unvech(delta2 theta_uls)
      term2 <- minv_delta2t %*% dw2_e2

      jac_a[, k] <- term1 + term2
    }
  }
  jac_a
}

# RLS
# - first step: only Delta_2 depends on theta1
# - second step: iteratively updating Sigma -> W_2 also depends on theta1!
lav_sem_miiv_utils_jaca_rls <- function(lavmodel = NULL,
                                        lavpartable = NULL,
                                        lavh1 = NULL,
                                        free_directed_idx = integer(0L),
                                        free_undirected_idx = integer(0L)) {
  nblocks <- lavmodel@nblocks

  # delta across all blocks
  delta_block <- lav_model_delta(lavmodel = lavmodel)

  # augment lavpartable to include model matrices and row/col indices
  mm_info <- lav_lisrel(lavpartable, target = NULL, extra = FALSE)

  jac_a <- matrix(0, length(free_undirected_idx), length(free_directed_idx))
  for (b in seq_len(nblocks)) {
    # s_vech
    sample_cov <- lavh1$implied$cov[[b]]
    nvar <- nrow(sample_cov)
    s_vech <- lav_mat_vech(sample_cov)

    # delta2
    delta2 <- delta_block[[b]][, free_undirected_idx, drop = FALSE]
    if (lavmodel@meanstructure) {
      delta2 <- delta2[-seq_len(nvar), , drop = FALSE]
    }

    # w2
    s_inv <- diag(1, nrow = nrow(sample_cov))
    w2_uls <- 0.5 * lav_mat_dup_pre_post(s_inv %x% s_inv)

    # MLIST
    mm_in_group <- seq_len(lavmodel@nmat[b]) + cumsum(c(0, lavmodel@nmat))[b]
    mlist <- lavmodel@GLIST[mm_in_group]
    if (!is.null(mlist$beta)) {
      mlist$IB.inv <- lav_lisrel_ibinv(mlist)
    }

    # step 1
    m_uls <- t(delta2) %*% w2_uls %*% delta2
    m_uls_inv <- solve(m_uls)
    theta_uls <- drop(m_uls_inv %*% t(delta2) %*% w2_uls %*% s_vech)

    # step 2
    theta2 <- theta_uls
    for (i in seq_len(200)) {
      old_x <- theta2
      new_sigma <- lav_mat_vech_rev(delta2 %*% theta2)
      sigma_inv <- solve(new_sigma)
      w2 <- 0.5 * lav_mat_dup_pre_post(sigma_inv %x% sigma_inv)
      m_mat <- t(delta2) %*% w2 %*% delta2
      theta2 <- drop(solve(m_mat, t(delta2) %*% w2 %*% s_vech))
      if (sum((old_x - theta2)^2) < 1e-12 * (1 + sum(theta2^2))) break
    }

    # final quantities
    new_sigma <- lav_mat_vech_rev(delta2 %*% theta2)
    sigma_inv <- solve(new_sigma)
    w2 <- 0.5 * lav_mat_dup_pre_post(sigma_inv %x% sigma_inv)
    m_mat <- t(delta2) %*% w2 %*% delta2
    m_mat_inv <- solve(m_mat)
    e2 <- s_vech - delta2 %*% theta2
    e_mat <- lav_mat_vech_rev(e2)
    sigma_inv_e <- sigma_inv %*% e_mat

    minv_delta2t <- m_mat_inv %*% t(delta2)
    w2_e2 <- w2 %*% e2
    delta2t_w2 <- t(delta2) %*% w2

    # helper: compute (dW)e2 given d vech(Sigma) as a vector v
    d_w_e2_from_v <- function(v) {
      d_sigma <- lav_mat_vech_rev(v)
      d_sigma_inv <- -sigma_inv %*% d_sigma %*% sigma_inv
      sym_part <- sigma_inv_e %*% d_sigma_inv + d_sigma_inv %*% t(sigma_inv_e)
      dw2_e2 <-
        0.5 * drop(lav_mat_dup_pre(as.matrix(as.vector(sym_part))))
      dw2_e2
    }

    # A = M2^{-1} delta2^T J_W^(theta2): q2 x q2
    # column j of J_W^(theta2): dSigma = unvech(delta2[:,j])
    jw_theta2 <- matrix(0.0, nrow(delta2), ncol(delta2))
    for (j in seq_len(ncol(delta2))) {
      jw_theta2[, j] <- d_w_e2_from_v(delta2[, j])
    }
    a <- minv_delta2t %*% jw_theta2
    iminus_a_inv <- solve(diag(ncol(delta2)) - a)

    for (k in seq_along(free_directed_idx)) {
      d_idx <- match(free_directed_idx[k], lavpartable$free)[1]
      # correct block?
      if (lavpartable$block[d_idx] != b) {
        next
      }
      d_mat <- mm_info$mat[d_idx]
      d_row <- mm_info$row[d_idx]
      d_col <- mm_info$col[d_idx]
      d_deltak <- matrix(0, nrow(delta2), ncol(delta2))
      # we only need to fill the columns of dDeltak that correspond to the
      # psi elements in free.undirected.idx
      for (i in seq_along(free_undirected_idx)) {
        un_idx <- match(free_undirected_idx[i], lavpartable$free)[1]
        un_mat <- mm_info$mat[un_idx]
        if (un_mat == "theta") {
          next
        }
        un_row <- mm_info$row[un_idx]
        un_col <- mm_info$col[un_idx]

        if (d_mat == "lambda") {
          tmp <- lav_lisrel_d2sigma_lambda_psi(
            mlist = mlist,
            i = d_row, j = d_col, k = un_row, l = un_col
          )
        } else if (d_mat == "beta") {
          tmp <- lav_lisrel_d2sigma_beta_psi(
            mlist = mlist,
            i = d_row, j = d_col, k = un_row, l = un_col
          )
        }
        d_deltak[, i] <- lav_mat_vech(tmp)
      }

      # term 1: direct delta2 variation in final WLS
      term1 <-
        m_mat_inv %*% (t(d_deltak) %*% w2_e2 -
                                            delta2t_w2 %*% d_deltak %*% theta2)

      # term 2: W variation from dSigma^(Delta2) = unvech(dDelta2_k %*% theta2)
      # (the dSigma^(theta2) part is handled implicitly by (I-A)^{-1})
      v_k <- as.vector(d_deltak %*% theta2)
      term2 <- minv_delta2t %*% d_w_e2_from_v(v_k)

      jac_a[, k] <- iminus_a_inv %*% (term1 + term2)
    }
  }
  jac_a
}


# jac_k: d theta1 / d svec
# see Fisher & Bollen (2020) section 2.4

# per equation, one block only!
lav_sem_miiv_utils_jack_eqs <- function(eqs = NULL, # one block only
                                        block = 1L,
                                        lavmodel = NULL,
                                        lavpartable = NULL,
                                        free_directed_idx = integer(0L)) {
  # continuous data only (for now)
  nvar <- lavmodel@nvar[block]
  if (lavmodel@categorical) {
    pstar <- nvar * (nvar - 1L) / 2
  } else {
    pstar <- nvar * (nvar + 1L) / 2
  }
  ntheta1 <- length(free_directed_idx)

  b <- block

  # K matrix
  if (lavmodel@categorical) {
    nc <- ncol(eqs[[1]][[1]]$k_mat)
    k_mat <- matrix(0.0, nrow = ntheta1, ncol = nc)
  } else if (lavmodel@meanstructure) {
    k_mat <- matrix(0.0, nrow = ntheta1, ncol = (nvar + pstar))
  } else {
    k_mat <- matrix(0.0, nrow = ntheta1, ncol = pstar)
  }

  # collect k_mat matrices from each equation
  for (j in seq_along(eqs[[b]])) {
    # this equation
    eq <- eqs[[b]][[j]]
    tmp <- lavpartable$free[eq$pt] # free index
    tmp <- tmp[!tmp == 0]
    free_idx <- match(tmp, free_directed_idx)
    nx <- nrow(eq$k_mat)
    if (length(free_idx) > 0L && nx > 0L) {
      if (all(free_idx > 0L)) {
        k_mat[free_idx, ] <- eq$k_mat
      } else {
        # remove non-free elements
        zero_idx <- which(free_idx == 0L)
        free_idx <- free_idx[-zero_idx]
        if (length(free_idx) > 0) {
          k_mat[free_idx, ] <- eq$k_mat[-zero_idx, , drop = FALSE]
        }
      }
    }
    if (lavmodel@meanstructure) {
      tmp <- lavpartable$free[eq$ptint]
      tmp <- tmp[!tmp == 0]
      free_int_idx <- match(tmp, free_directed_idx)
      if (length(free_int_idx) > 0L && free_int_idx > 0L) {
        k_mat[free_int_idx, ] <- eq$k_mat_int
      }
    }
  } # eq

  k_mat
}

# per-equation overidentification test based on Browne's residual-based
# statistic, using the asymptotic covariance (ACOV/Gamma) of the sample
# (co)variances or correlations. This replaces the standard Sargan test when
# the latter is not valid, in particular when the moments are polychoric
# correlations (categorical data), but also for non-normal continuous data.
#
# For an equation  y = beta' x + e  with instruments z (the moment conditions
# being cov(z, e) = 0), the overidentifying restrictions are
#   g = s_zy - S_zx %*% beta = 0   (a q-vector, q = #instruments)
# The (reduced) residual-based statistic is
#   T = N * g' [W - W D (D'W D)^{-1} D'W] g,   W = Gamma_g^{-1},  D = S_zx
# where Gamma_g is the ACOV of g, obtained from the ACOV of the relevant
# sample moments via the delta method (g is linear in those moments with
# d g_k / d s_{z_k y} = 1 and d g_k / d s_{z_k x_j} = -beta_j). This is
# algebraically equivalent to Browne's residual test applied to the single
# equation viewed as a covariance/correlation structure that is saturated
# except for the q - p overidentifying restrictions. Under the classical
# homoskedastic-normal weight Gamma_g = resvar * S_zz, T reduces exactly to
# the standard Sargan statistic (and D'W g = 0, so the projection drops out).
#
# Arguments:
#   y_idx, x_idx, i_idx : column indices (into 'sample_cov') of the outcome,
#                         the regressors x, and the instruments z
#   beta                : the (estimated) directed slopes (length = #x)
#   sample_cov          : the (saturated) covariance/correlation matrix over
#                         all variables in this block
#   nacov               : the ACOV of the sample statistics (Gamma); for
#                         categorical data this is the polychoric ACOV over
#                         [thresholds, correlations]
#   nth                 : number of leading entries in 'nacov' that precede the
#                         (co)variance/correlation block (e.g. thresholds);
#                         0 for the continuous case
#   moment_idx_mat      : symmetric integer matrix mapping a variable pair
#                         (a, b) to its position within the moment block of
#                         'nacov' (vech order, off-diagonal for correlations)
#   nobs                : number of observations for this block
lav_sem_miiv_browne_test <- function(y_idx = integer(0L), x_idx = integer(0L),
                                     i_idx = integer(0L), beta = numeric(0L),
                                     sample_cov = NULL, nacov = NULL,
                                     nth = 0L, moment_idx_mat = NULL,
                                     nobs = 0L) {
  q <- length(i_idx)
  p <- length(x_idx)
  df <- q - p
  out <- c(stat = as.numeric(NA), df = df, pvalue = as.numeric(NA))
  if (df <= 0L) {
    return(out)
  }

  s_zy <- sample_cov[i_idx, y_idx, drop = FALSE]
  s_zx <- sample_cov[i_idx, x_idx, drop = FALSE]
  g <- drop(s_zy - s_zx %*% beta)

  # indices into 'nacov' of the moments that enter g: the q instrument-outcome
  # moments, followed by the q * p instrument-regressor moments
  p_y <- nth + moment_idx_mat[cbind(i_idx, rep.int(y_idx, q))]
  p_x <- nth + as.vector(moment_idx_mat[cbind(rep(i_idx, times = p),
                                              rep(x_idx, each = q))])
  p_all <- c(p_y, p_x)
  gam_p <- nacov[p_all, p_all, drop = FALSE]

  # delta-method Jacobian of g w.r.t. those moments
  jmat <- matrix(0, q, length(p_all))
  jmat[seq_len(q), seq_len(q)] <- diag(q)
  for (j in seq_len(p)) {
    cols <- q + (j - 1L) * q + seq_len(q)
    jmat[cbind(seq_len(q), cols)] <- -beta[j]
  }
  gamma_g <- jmat %*% gam_p %*% t(jmat)

  # T = N g' [W - W D (D'W D)^{-1} D'W] g
  w <- lav_mat_sym_inverse(gamma_g)
  wg <- drop(w %*% g)
  q1 <- sum(g * wg)
  q2 <- 0
  if (p > 0L) {
    dtwg <- crossprod(s_zx, wg)
    dtwd <- crossprod(s_zx, w %*% s_zx)
    q2 <- drop(crossprod(dtwg, lav_mat_sym_solve_spd(dtwd, dtwg)))
  }
  stat <- as.numeric(nobs * (q1 - q2))

  out["stat"] <- stat
  out["pvalue"] <- pchisq(stat, df, lower.tail = FALSE)
  out
}

# Bootstrap inference for rotated EFA/ESEM models.
#
# YR/Claude - 2026
#
# Ordinary (or other) bootstrap inference is not straightforward for models
# with rotated exploratory factors: the optimizer works with the *unrotated*
# solution (with echelon constraints), so the raw bootstrap draws (optim$x) do
# not correspond to the rotated parameters reported to the user. Moreover, even
# the rotated solution is only identified up to a signed permutation of the
# factors, so the rotated factors of each bootstrap sample must be aligned
# (re-ordered and reflected) to the original (full-sample) rotated solution
# before their variability is meaningful.
#
# This file implements that machinery. It is driven from
# lav_lavaan_step16_rotation.R when se = "bootstrap" and the model contains
# (rotated) exploratory factors. See github issue #376.

# extract, per block and per EFA set, the (rotated) loading submatrix involved
# in that set. Returns a list (per block) of lists (per set); entries for
# single-factor sets (or empty sets) are NULL (nothing to align there).
lav_efa_bootstrap_lambda_list <- function(lavmodel = NULL, glist = NULL) {
  if (is.null(glist)) {
    glist <- lavmodel@GLIST
  }
  out <- vector("list", lavmodel@nblocks)
  for (g in seq_len(lavmodel@nblocks)) {
    mm_in_group <- seq_len(lavmodel@nmat[g]) + cumsum(c(0, lavmodel@nmat))[g]
    mlist <- glist[mm_in_group]
    set_list <- vector("list", lavmodel@nefa)
    for (set in seq_len(lavmodel@nefa)) {
      ov_idx <- lavmodel@ov.efa.idx[[g]][[set]]
      lv_idx <- lavmodel@lv.efa.idx[[g]][[set]]
      if (length(ov_idx) == 0L || length(lv_idx) < 2L) {
        next
      }
      set_list[[set]] <- mlist$lambda[ov_idx, lv_idx, drop = FALSE]
    }
    out[[g]] <- set_list
  }
  out
}

# given a (rotated) fitted bootstrap model, align its rotated EFA factors to the
# reference solution 'ref_lambda' (a lambda-list as returned by
# lav_efa_bootstrap_lambda_list()), and return the aligned rotated parameter
# vector (free parameters only, in parameter-table order).
#
# Alignment is a signed permutation T per EFA set; it is applied to *all*
# factor-related model matrices (lambda, psi, beta, alpha) exactly as a rotation
# would be (see lav_model_efa_rotate_x()), so that every parameter connected to
# the rotated factors (loadings, factor (co)variances, structural coefficients,
# factor means) is transformed consistently. This makes the approach valid for
# ESEM (factors embedded in a larger structural model), not just plain EFA.
lav_efa_bootstrap_align_coef <- function(lavmodel = NULL, lavpartable = NULL,
                                         ref_lambda = NULL) {
  glist <- lavmodel@GLIST
  for (g in seq_len(lavmodel@nblocks)) {
    mm_in_group <- seq_len(lavmodel@nmat[g]) + cumsum(c(0, lavmodel@nmat))[g]
    mlist <- glist[mm_in_group]

    # build the block-level signed permutation hg (identity outside the EFA
    # factors, so only the exploratory factors are touched)
    hg <- diag(ncol(mlist$lambda))
    changed <- FALSE
    for (set in seq_len(lavmodel@nefa)) {
      ref <- ref_lambda[[g]][[set]]
      if (is.null(ref)) {
        next
      }
      ov_idx <- lavmodel@ov.efa.idx[[g]][[set]]
      lv_idx <- lavmodel@lv.efa.idx[[g]][[set]]
      lambda_b <- mlist$lambda[ov_idx, lv_idx, drop = FALSE]
      tt <- lav_efa_align_signed_perm(lambda = lambda_b, lambda_ref = ref)
      hg[lv_idx, lv_idx] <- tt
      changed <- TRUE
    }

    if (changed) {
      # hg is an orthonormal signed permutation (solve(t(hg)) == hg), so the
      # general rotation transform reduces to lambda %*% hg etc. Reuse the same
      # helper as lav_model_efa_rotate_x() for a single, consistent transform.
      mlist <- lav_model_efa_apply_hg(mlist, hg)
      glist[mm_in_group] <- mlist
    }
  } # block

  lavmodel@GLIST <- glist
  est_user <- lav_model_get_parameters(
    lavmodel = lavmodel, type = "user", extra = TRUE
  )
  free_idx <- which(lavpartable$free > 0L & !duplicated(lavpartable$free))
  est_user[free_idx]
}

# run the (ordinary) bootstrap for a rotated EFA/ESEM model.
#
# 'lavmodel'/'lavpartable' are the *unrotated* model and parameter table (so
# each bootstrap sample is refit and rotated from scratch), while 'efa_ref'
# carries the reference rotated solution and the rotated free-parameter layout
# (see lav_efa_bootstrap_reference()). Returns the (r x npar.rotated) matrix of
# aligned rotated bootstrap coefficients, with the usual error.idx / seed /
# nonadmissible attributes.
lav_efa_bootstrap_run <- function(lavmodel = NULL, lavpartable = NULL,
                                  lavsamplestats = NULL, lavoptions = NULL,
                                  lavdata = NULL, efa_ref = NULL) {
  # number of bootstrap draws
  r <- lavoptions$bootstrap
  if (is.list(r)) {
    r <- r$R
  }
  if (is.null(r)) {
    r <- 1000L
  }
  r <- as.integer(r)

  coef_boot <- lav_bootstrap_internal(
    object = NULL,
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavpartable = lavpartable,
    lavoptions = lavoptions,
    lavdata = lavdata,
    r = r,
    check_post = lavoptions$check.post,
    type = "ordinary",
    fun = "coef",
    efa_reference = efa_ref
  )

  coef_boot
}

# turn the (aligned, rotated) bootstrap coefficient matrix into a vcov + se,
# using the *rotated* model and parameter table. Returns a list with elements
# 'se' (for lavpartable$se), 'vcov' (the rotated bootstrap vcov) and 'coef' (the
# raw bootstrap coefficients, for the @boot slot). Mirrors the scaling used in
# lav_model_nvcov_bootstrap().
lav_efa_bootstrap_vcov <- function(coef_boot = NULL, lavmodel = NULL,
                                   lavpartable = NULL, lavoptions = NULL) {
  coef_orig <- coef_boot

  # drop failed runs
  error_idx <- attr(coef_boot, "error.idx")
  if (length(error_idx) > 0L) {
    coef_boot <- coef_boot[-error_idx, , drop = FALSE] # drops attributes
  }

  nboot <- nrow(coef_boot)
  if (is.null(nboot) || nboot < 2L) {
    lav_msg_stop(gettext(
      "not enough successful bootstrap runs to compute standard errors."))
  }

  # the bootstrap directly estimates Cov(coef), which IS the target vcov of the
  # (rotated) parameters. Use the same (nboot-1)/nboot scaling as the rest of
  # lavaan (see lav_model_nvcov_bootstrap()).
  vcov <- cov(coef_boot) * (nboot - 1) / nboot

  # standard errors for the (rotated) free parameters + defined parameters
  se <- lav_model_vcov_se(
    lavmodel = lavmodel, lavpartable = lavpartable,
    vcov = vcov, boot = coef_orig, lavoptions = lavoptions
  )

  list(se = se, vcov = vcov, coef = coef_orig)
}

# methods
setMethod(
  "show", "lavaanList",
  function(object) {
    # check object
    object <- lav_object_check_version(object)
    # show only basic information
    lav_lavaanlist_short_summary(object, print = TRUE)
  }
)

lav_lavaanlist_short_summary <- function(object, print = TRUE) {
  txt <- sprintf(
    "lavaanList (%s) -- based on %d datasets (%d converged)\n",
    object@version,
    object@meta$ndat,
    sum(object@meta$ok)
  )

  if (print) {
    cat(txt)
  }

  invisible(txt)
}

setMethod(
  "summary", "lavaanList",
  function(object, header = TRUE,
           estimates = TRUE,
           print = TRUE,
           nd = 3L,
           simulate.args = list(est.bias = TRUE,                     # nolint
                                se.bias  = TRUE,
                                se.bias.med = TRUE,
                                prop.sig = TRUE,
                                coverage = TRUE,
                                level    = 0.95,
                                trim     = 0),
           ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("summary"))
        )
      }
    }
    lav_lavaanlist_summary(object,
      header = header, estimates = estimates,
      print = print, nd = nd, simulate_args = simulate.args
    )
  }
)

lav_lavaanlist_summary <- function(object,
                                   header = TRUE,
                                   estimates = TRUE,
                                   simulate_args = list(est.bias = TRUE,
                                                        se.bias  = TRUE,
                                                        se.bias.med = TRUE,
                                                        prop.sig = TRUE,
                                                        coverage = TRUE,
                                                        level    = 0.95,
                                                        trim     = 0),
                                   zstat = TRUE,
                                   pvalue = TRUE,
                                   print = TRUE,
                                   nd = 3L) {
  # check object
  object <- lav_object_check_version(object)

  out <- list()

  if (header) {
    out$header <- lav_lavaanlist_short_summary(object, print = print)

    # if(print) {
    #    # show only basic information
    #    lav_lavaanlist_short_summary(object)
    # }
  }
  if (print) {
    output <- "text"
  } else {
    output <- "data.frame"
  }

  if (estimates && "partable" %in% object@meta$store.slots) {
    pe <- lavParameterEstimates(object,
      se = FALSE,
      remove_system_eq = FALSE, remove_eq = FALSE,
      remove_ineq = FALSE, remove_def = FALSE,
      remove_nonfree = FALSE, remove_unused = FALSE,
      remove_step1 = FALSE, # in case we used sam()
      # zstat = FALSE, pvalue = FALSE, ci = FALSE,
      standardized = FALSE,
      output = output
    )

    # scenario 1: simulation
    if (!is.null(object@meta$lavSimulate)) {

      # default behavior
      sim_args <- list(est.bias = TRUE,
                       se.bias  = TRUE,
                       se.bias.med = TRUE,
                       prop.sig = TRUE,
                       coverage = TRUE,
                       level    = 0.95,
                       trim     = 0)
      sim_args <- modifyList(sim_args, simulate_args)
      #if (!sim.args$est.bias) {
      #  sim.args$se.bias <- FALSE
      #}
      if (!sim_args$se.bias) {
        sim_args$prop.sig <- FALSE
      }

      pe$est.true <- object@meta$est.true
      nel <- length(pe$est.true)

      # always compute EST
      est <- lav_lavaanlist_pt(object, what = "est", type = "all")

      # sometimes compute SE
      if (sim_args$se.bias || sim_args$prop.sig || sim_args$coverage) {
        se <- lav_lavaanlist_pt(object, what = "se", type = "all")
      }

      # est.bias?
      if (sim_args$est.bias) {
        ave_1 <- apply(est, 1L, mean, na.rm = TRUE, trim = sim_args$trim)

        # remove things like equality constraints
        if (length(ave_1) > nel) {
          ave_1 <- ave_1[seq_len(nel)]
        }
        ave_1[!is.finite(ave_1)] <- as.numeric(NA)
        pe$est.ave <- ave_1
        pe$est.bias <- pe$est.ave - pe$est.true

        # FIXME: should we also add bias^2 and MSE?
        # and what about relative bias?
      }


      # SE?
      if (sim_args$se.bias) {
        se_obs_1 <- apply(est, 1L, lav_sample_trimmed_sd, na_rm = TRUE,
                        trim = sim_args$trim)
        if (length(se_obs_1) > nel) {
          se_obs_1 <- se_obs_1[seq_len(nel)]
        }
        se_obs_1[!is.finite(se_obs_1)] <- as.numeric(NA)
        pe$se.obs <- se_obs_1
        se_ave <- apply(se, 1L, mean, na.rm = TRUE, trim = sim_args$trim)
        if (length(se_ave) > nel) {
          se_ave <- se_ave[seq_len(nel)]
        }
        se_ave[!is.finite(se_ave)] <- as.numeric(NA)
        pe$se.ave <- se_ave
        se_obs <- se_obs_1
        se_obs[se_obs < .Machine$double.eps ^ (1 / 3)] <- as.numeric(NA)
        pe$se.bias <- pe$se.ave / se_obs # use ratio!
        pe$se.bias[!is.finite(pe$se.bias)] <- as.numeric(NA)

        # robust (median-based) SE summary. The SE sampling distribution of
        # (non)linear defined parameters can be heavy-tailed (eg ratios with
        # near-zero denominators), so the mean-based se.ave/se.bias may be
        # dominated by a few outliers; the median is robust. Use the original
        # (untrimmed) 'se' here, before the prop.sig block alters it below.
        if (sim_args$se.bias.med) {
          se_med <- apply(se, 1L, median, na.rm = TRUE)
          if (length(se_med) > nel) {
            se_med <- se_med[seq_len(nel)]
          }
          se_med[!is.finite(se_med)] <- as.numeric(NA)
          pe$se.med <- se_med
          pe$se.bias.med <- pe$se.med / se_obs # use ratio!
          pe$se.bias.med[!is.finite(pe$se.bias.med)] <- as.numeric(NA)
        }
      }

      if (sim_args$prop.sig) {
        se[se < sqrt(.Machine$double.eps)] <- as.numeric(NA)
        wald_1 <- est / se
        wald <- apply(wald_1, 1L, mean, na.rm = TRUE,
                      trim = sim_args$trim)
        wald[!is.finite(wald)] <- as.numeric(NA)
        pe$wald <- wald
        pval <- 2 * (1 - pnorm(abs(wald_1)))
        propsig <- apply(pval, 1L, function(x) {
                        x_ok <- x[is.finite(x)]
                        nx <- length(x_ok)
                        sum(x_ok < (1 - sim_args$level)) / nx
                       })
        propsig[!is.finite(propsig)] <- as.numeric(NA)
        pe$prop.sig <- propsig
      }

      if (sim_args$coverage) {
        # 1. symmetric (Wald-type) confidence intervals, based on the
        #    standard error: est +/- z * se
        # next three lines based on confint.lm
        a <- (1 - sim_args$level) / 2
        a <- c(a, 1 - a)
        fac <- qnorm(a)
        ci_lower_1 <- est + fac[1] * se
        ci_upper_1 <- est + fac[2] * se
        ci_lower <- apply(ci_lower_1, 1L, mean, na.rm = TRUE,
                          trim = sim_args$trim)
        ci_upper <- apply(ci_upper_1, 1L, mean, na.rm = TRUE,
                          trim = sim_args$trim)
        ci_lower[!is.finite(ci_lower)] <- as.numeric(NA)
        ci_upper[!is.finite(ci_upper)] <- as.numeric(NA)
        pe$ci.lower <- ci_lower
        pe$ci.upper <- ci_upper
        # columnwise comparison
        inside_flag <- (ci_lower_1 <= pe$est.true) & (ci_upper_1 >= pe$est.true)
        coverage <- apply(inside_flag, 1L, mean, na.rm = TRUE)
        coverage[!is.finite(coverage)] <- as.numeric(NA)
        pe$coverage <- coverage

        # 2. (possibly asymmetric) confidence intervals, based on the
        #    bootstrap or the Monte Carlo method; only available if the
        #    per-dataset ci.lower/ci.upper were stored (see lavaanList).
        #    These intervals are computed at the 95% level (see lavaanList).
        has_asym_ci <-
          "ci.lower" %in% names(object@ParTableList[[1]]) &&
          "ci.upper" %in% names(object@ParTableList[[1]])
        if (has_asym_ci) {
          if (!isTRUE(all.equal(sim_args$level, 0.95))) {
            lav_msg_note(gettext(
              "the bootstrap/Monte Carlo confidence intervals were stored at
               the 95% level; the requested `level' only affects the
               symmetric (SE-based) intervals."))
          }
          ci_lower_a1 <- lav_lavaanlist_pt(object, what = "ci.lower",
                                                 type = "all")
          ci_upper_a1 <- lav_lavaanlist_pt(object, what = "ci.upper",
                                                 type = "all")
          # average bounds
          ci_lower_a <- apply(ci_lower_a1, 1L, mean, na.rm = TRUE,
                              trim = sim_args$trim)
          ci_upper_a <- apply(ci_upper_a1, 1L, mean, na.rm = TRUE,
                              trim = sim_args$trim)
          if (length(ci_lower_a) > nel) ci_lower_a <- ci_lower_a[seq_len(nel)]
          if (length(ci_upper_a) > nel) ci_upper_a <- ci_upper_a[seq_len(nel)]
          ci_lower_a[!is.finite(ci_lower_a)] <- as.numeric(NA)
          ci_upper_a[!is.finite(ci_upper_a)] <- as.numeric(NA)
          # columnwise comparison
          ci_lower_a1 <- ci_lower_a1[seq_len(nel), , drop = FALSE]
          ci_upper_a1 <- ci_upper_a1[seq_len(nel), , drop = FALSE]
          inside_a <- (ci_lower_a1 <= pe$est.true) &
                      (ci_upper_a1 >= pe$est.true)
          coverage_a <- apply(inside_a, 1L, mean, na.rm = TRUE)
          coverage_a[!is.finite(coverage_a)] <- as.numeric(NA)
          # use a method-specific suffix (.boot or .mc) for the columns
          ci_sfx <- if (identical(object@Options$se, "bootstrap")) {
            "boot"
          } else {
            "mc"
          }
          pe[[paste0("ci.lower.", ci_sfx)]] <- ci_lower_a
          pe[[paste0("ci.upper.", ci_sfx)]] <- ci_upper_a
          pe[[paste0("coverage.", ci_sfx)]] <- coverage_a
        }
      }

      # if sam(), should we keep or remove the step1 values?
      # keep them for now

      # scenario 2: bootstrap
    } else if (!is.null(object@meta$lavBootstrap)) {
      # print the average value for est
      est <- lav_lavaanlist_pt(object, what = "est", type = "all")
      pe$est.ave <- rowMeans(est, na.rm = TRUE)

      # scenario 3: multiple imputation
    } else if (!is.null(object@meta$lavMultipleImputation)) {
      # pool est: take the mean
      est <- lav_lavaanlist_pt(object, what = "est", type = "all")
      m <- NCOL(est)
      pe$est <- rowMeans(est, na.rm = TRUE)

      # pool se

      # between-imputation variance
      # B.var <- apply(EST, 1L, var)
      est1 <- rowMeans(est, na.rm = TRUE)
      est2 <- rowMeans(est^2, na.rm = TRUE)
      b_var <- (est2 - est1 * est1) * m / (m - 1)

      # within-imputation variance
      se <- lav_lavaanlist_pt(object, what = "se", type = "all")
      w_var <- rowMeans(se^2, na.rm = TRUE)

      # total variance: T.var = W.var + B.var + B.var/m
      pe$se <- sqrt(w_var + b_var + (b_var / m))

      tmp_se <- ifelse(pe$se == 0.0, NA, pe$se)
      if (zstat) {
        pe$z <- pe$est / tmp_se
        if (pvalue) {
          pe$pvalue <- 2 * (1 - pnorm(abs(pe$z)))
        }
      }

      # scenario 4: multiple groups/sets
    } else if (!is.null(object@meta$lavMultipleGroups)) {
      # show individual estimates, for each group
      est <- lav_lavaanlist_pt(object, what = "est", type = "all")
      est <- as.list(as.data.frame(est))
      names(est) <- object@meta$group.label
      attr_1 <- attributes(pe)
      names_1 <- c(names(pe), names(est))
      pe <- c(pe, est)
      attributes(pe) <- attr_1
      names(pe) <- names_1
    } else {
      # scenario 5: just a bunch of fits, using different datasets
      # print the average value for est
      est <- lav_lavaanlist_pt(object, what = "est", type = "all")
      pe$est.ave <- rowMeans(est, na.rm = TRUE)

      # more?
    }

    # remove ==,<,>
    rm_idx <- which(pe$op %in% c("==", "<", ">"))
    if (length(rm_idx) > 0L) {
      pe <- pe[-rm_idx, ]
    }

    out$pe <- pe

    if (print) {
      # print pe?
      print(pe, nd = nd)
    }
  } else {
    cat("available slots (per dataset) are:\n")
    print(object@meta$store.slots)
  }

  invisible(out)
}

setMethod(
  "coef", "lavaanList",
  function(object, type = "free", labels = TRUE, ...) {
    # check object
    object <- lav_object_check_version(object)
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("coef"))
        )
      }
    }
    lav_lavaanlist_pt(
      object = object, what = "est", type = type,
      labels = labels
    )
  }
)

lav_lavaanlist_pt <- function(object, what = "est",
                                    type = "free", labels = TRUE) {
  # check object
  object <- lav_object_check_version(object)
  if ("partable" %in% object@meta$store.slots) {
    if (what %in% names(object@ParTableList[[1]])) {
      out <- sapply(object@ParTableList, "[[", what)
    } else {
      lav_msg_stop(gettextf(
        "column `%s' not found in the first element of the ParTableList slot.",
        what))
    }
  } else {
    lav_msg_stop(gettext("no ParTable slot stored in lavaanList object"))
  }

  if (type == "user" || type == "all") {
    type <- "user"
    idx <- seq_along(object@ParTable$lhs)
  } else if (type == "free") {
    idx <- which(object@ParTable$free > 0L & !duplicated(object@ParTable$free))
  } else {
    lav_msg_stop(gettext("argument `type' must be one of free or user"))
  }

  out <- out[idx, , drop = FALSE]

  if (labels) {
    rownames(out) <- lav_pt_labels(object@ParTable, type = type)
  }

  out
}

lav_step12_implied <- function(lavoptions = NULL,
                                      lavmodel = NULL) {
  # # # # # # # # # # # #
  # #  12. lavimplied # #
  # # # # # # # # # # # #

  # if lavoptions$implied compute lavimplied via lav_model_implied
  lavimplied <- list()
  if (lavoptions$implied) {
    if (lav_verbose()) {
      cat("lavimplied  ...")
    }
    lavimplied <- lav_model_implied(lavmodel)
    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  lavimplied
}

lav_step12_loglik <- function(lavoptions = NULL,
                                     lavdata = NULL,
                                     lavsamplestats = NULL,
                                     lavh1 = NULL,
                                     lavimplied = NULL,
                                     lavmodel = NULL,
                                     lavcache = NULL) {
  # # # # # # # # # # # #
  # #  12. lavloglik  # #
  # # # # # # # # # # # #

  # only when missing = "ml" (or "ml.x") and zero coverage
  if (lavoptions$missing %in% c("ml", "ml.x") &&
      lavoptions$model.type == "unrestricted" &&
      length(lavh1) == 0L) {
    lavh1 <- list(implied = lavimplied,
                  logl = list())
  }

  # if lavoptions$loglik compute lavloglik via lav_model_loglik
  lavloglik <- list()
  if (lavoptions$loglik) {
    if (lav_verbose()) {
      cat("lavloglik   ...")
    }
    lavloglik <- lav_model_loglik(
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      lavh1 = lavh1,
      lavimplied = lavimplied,
      lavmodel = lavmodel,
      lavoptions = lavoptions,
      lavcache = lavcache
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  lavloglik
}

lav_step04_pt <- function(slot_par_table = NULL,
                                       model = NULL,
                                       flat_model = NULL,
                                       lavoptions = NULL,
                                       lavdata = NULL,
                                       constraints = NULL,
                                       marker = NULL) {
  # # # # # # # # # # # #
  # #  4. lavpartable # #
  # # # # # # # # # # # #

  # if slot_par_table not null
  #   copy slot_par_table to lavpartable
  # else
  #   if model is character or formula
  #     create a temporary variable tmp.data.ov equal to lavdata@ov
  #     if estimator "catML"
  #       set meanstructure to FALSE
  #       set the member type in the temporary variable tmp.data.ov to a
  #         numeric vector with all zeroes
  #     create lavpartable via function lavParTable (=lav_model_pt)
  #                     using the temporary variable for parameter varTable
  #   else
  #     if model is lavaan object
  #       set lavpartable = parTable(model)
  #     else
  #       if model is a list
  #         set lavpartable to
  #           as.list(lav_pt_complete(as.list(flat_model)))
  #       else
  #         *** error ***
  # if slot_par_table is NULL check lavpartable via lav_partable_check
  # if lavoptions$optim.method is "em" and there are variances specified in
  #   partable with free = 0L and
  #    starting value ustart 0, set ustart for these variances to
  #    lavoptions$em.args$zerovar_offset

  if (!is.null(slot_par_table)) {
    lavpartable <- lav_pt_set_cache(slot_par_table)
  } else if (is.character(model) ||
    inherits(model, "formula") ||
  # model was already a flat_model
  (is.list(model) && !is.null(model$mod.idx) &&
   !is.null(attr(model, "modifiers")))) {
    if (lav_verbose()) {
      cat("lavpartable        ...")
    }
    # check flat_model before we proceed
    if (lav_debug()) {
      print(as.data.frame(flat_model))
    }
    # catch ~~ of fixed.x covariates if fixed.x = TRUE
    # --> done inside lav_model_pt!

    # if(lavoptions$fixed.x) {
    #    tmp <- lav_pt_vnames(flat_model, type = "ov.x",
    #                               ov.x.fatal = FALSE, warn = TRUE)
    # tmp <- try(lav_pt_vnames(flat_model, type = "ov.x",
    #                                         ov.x.fatal = TRUE),
    #           silent = TRUE)
    # if(inherits(tmp, "try-error")) {
    #    warning("lavaan WARNING: syntax contains parameters involving ",
    #      "exogenous covariates; switching to fixed.x = FALSE")
    #    lavoptions$fixed.x <- FALSE
    # }
    # }
    # if(lavoptions$conditional.x) {
    #    tmp <- lav_pt_vnames(flat_model,
    #                  type = "ov.x", ov.x.fatal = TRUE)
    # }
    tmp_data_ov <- lavdata@ov
    if (lavoptions$estimator == "catML") {
      lavoptions$meanstructure <- FALSE
      tmp_data_ov$type <- rep("numeric", length(tmp_data_ov$type))
    }
    lavpartable <-
      lavParTable(
        model = flat_model,
        constraints = constraints,
        var_table = tmp_data_ov,
        ngroups = lavdata@ngroups,
        meanstructure = lavoptions$meanstructure,
        int_ov_free = lavoptions$int.ov.free,
        int_lv_free = lavoptions$int.lv.free,
        marker_int_zero = lavoptions$marker.int.zero,
        orthogonal = lavoptions$orthogonal,
        orthogonal_x = lavoptions$orthogonal.x,
        orthogonal_y = lavoptions$orthogonal.y,
        orthogonal_efa = lavoptions$rotation.args$orthogonal,
        conditional_x = lavoptions$conditional.x,
        fixed_x = lavoptions$fixed.x,
        std_lv = lavoptions$std.lv,
        correlation = if (length(lavoptions$.correlation.ov) > 0L) {
          lavoptions$.correlation.ov
        } else {
          lavoptions$correlation
        },
        composites = lavoptions$composites,
        composites_cov_free = identical(lavoptions$composites.cov, "free"),
        effect_coding = lavoptions$effect.coding,
        ceq_simple = lavoptions$ceq.simple,
        parameterization = lavoptions$parameterization,
        auto_fix_first = lavoptions$auto.fix.first,
        marker = marker,
        auto_fix_single = lavoptions$auto.fix.single,
        auto_var = lavoptions$auto.var,
        auto_cov_lv_x = lavoptions$auto.cov.lv.x,
        auto_cov_x = lavoptions$auto.cov.x,
        auto_cov_y = lavoptions$auto.cov.y,
        auto_th = lavoptions$auto.th,
        auto_delta = lavoptions$auto.delta,
        auto_efa = lavoptions$auto.efa,
        group_equal = lavoptions$group.equal,
        group_partial = lavoptions$group.partial,
        group_w_free = lavoptions$group.w.free,
        as_data_frame = FALSE
      )
    lavpartable <- lav_pt_set_cache(lavpartable)
    if (lav_verbose()) {
      cat(" done.\n")
    }
  } else if (inherits(model, "lavaan")) {
    lavpartable <- lav_pt_set_cache(as.list(parTable(model)), model@pta)
  } else if (is.list(model)) {
    # we already checked this when creating flat_model
    # but we may need to complete it
    lavpartable <- as.list(flat_model) # in case model is a data.frame
    # complete table
    lavpartable <- as.list(lav_pt_complete(lavpartable))
    lavpartable <- lav_pt_set_cache(lavpartable)
  } else {
    lav_msg_stop(gettextf(
      "model [type = %s] is not of type character or list", class(model)))
    }
  if (lav_debug()) {
    print(as.data.frame(lavpartable))
  }

  # at this point, we should check if the partable is complete
  # or not; this is especially relevant if the lavaan() function
  # was used, but the user has forgotten some variances/intercepts...
  if (is.null(slot_par_table)) {
    junk <- lav_pt_check(lavpartable,
      categorical = lavoptions$.categorical
    )
    rm(junk)
  }

  # if optim.method = "em" was set by default (two-level + missing =
  # "ml"; see lav_options_set), quietly fall back to "nlminb" when the
  # EM optimizer does not support the data configuration (single-group
  # two-level only, no conditional.x); this must happen *before* the
  # zero-variance offset tweak below
  if (lavoptions$optim.method == "em" &&
      isTRUE(lavoptions$.optim.em.fallback) &&
      (lavdata@nlevels == 1L ||
       lavdata@ngroups > 1L ||
       lavoptions$conditional.x)) {
    lavoptions$optim.method <- "nlminb"
  }

  # for EM only (for now), force fixed-to-zero (residual) variances
  # to be slightly larger than zero
  if (lavoptions$optim.method == "em") {
    zero_var_idx <- which(lavpartable$op == "~~" &
      lavpartable$lhs == lavpartable$rhs &
      lavpartable$free == 0L &
      lavpartable$ustart == 0)
    if (length(zero_var_idx) > 0L) {
      lavpartable$ustart[zero_var_idx] <- lavoptions$em.args$zerovar_offset
    }
    lavpartable <- lav_pt_set_cache(lavpartable, NULL, force = TRUE)
  }

  list(
    lavoptions  = lavoptions,
    lavpartable = lavpartable
  )
}

# initial version YR 02/08/2010

# YR 28 Jan 2017: add lavOptions(), lav_options_default()
# LDW 16 Apr 2024: move lavOptions and lav_options_default to separate file

# help functions for lav_options_set ####
# help function to determine estimator 'group'
lav_options_estimatorgroup <- function(estimator) {
  goal <- switch(estimator,
    ml = ,
    mlf = ,
    mlm = ,
    mlmv = ,
    mlmvs = ,
    mlr = "ML",
    catml = "catML",
    dwls = ,
    wlsm = ,
    wlsmv = ,
    wlsmvs = "DWLS",
    uls = ,
    ulsm = ,
    ulsmv = ,
    ulsmvs = "ULS",
    none = "none",
    toupper(estimator)
  )
  goal
}
lav_options_checkinterval <- function(x, nm, num2int) {
  if (num2int) x <- as.integer(x)
  oks <- vapply(x, function(x1) {
    (x1 > nm$bounds[1] || (x1 == nm$bounds[1] && nm$first.in)) &&
      (x1 < nm$bounds[2] || (x1 == nm$bounds[2] && nm$last.in))
  }, TRUE)
  all(oks)
}
lav_options_checkvalues <- function(optname, optvalue, chr) {
  optvalid <- names(chr)
  if (is.null(optvalid)) optvalid <- chr
  if (any(optvalid == "empty.string")) {
    optvalid[optvalid == "empty.string"] <- ""
  }
  optvals <- gsub("[_-]", ".", tolower(optvalue))
  optvalsok <- match(optvals, optvalid)
  if (any(is.na(optvalsok))) {
    lav_msg_stop(sprintf(
      ngettext(
        length(optvalue[is.na(optvalsok)]),
        "invalid value in %1$s option: %2$s.",
        "invalid values in %1$s option: %2$s."
      ),
      optname,
      lav_msg_view(optvalue[is.na(optvalsok)], log_sep = "none")
    ))
  }
  as.vector(chr[optvalsok])
}
lav_options_check <- function(opts, opt_check, subname) {
  opt_names <- names(opts)
  hiddens <- startsWith(opt_names, ".")
  if (any(hiddens)) { # remove hidden options temporarily
    opts_hidden <- opts[hiddens]
    opts <- opts[!hiddens]
    opt_names <- opt_names[!hiddens]
  }
  check_names <- names(opt_check)
  match_opt <- match(opt_names, check_names)
  if (any(is.na(match_opt))) {
    lav_msg_stop(gettextf(
      "Some option(s) unknown: %s !",
      lav_msg_view(opt_names[is.na(match_opt)], log_sep = "none")
    ))
  }
  for (j in seq_along(opts)) {
    opt_name <- opt_names[j]
    opt_value <- opts[[j]]
    opt_check1 <- opt_check[[match_opt[j]]]
    if (!is.null(attr(opt_check1, "SUB"))) {
      opts[[j]] <- lav_options_check(
        opt_value, opt_check1,
        paste0(opt_name, "$")
      )
      next
    }
    # check length of option value
    if (length(opt_value) < opt_check1$oklen[1]) {
      lav_msg_stop(gettextf(
        "Length of option '%1$s' value must be at least %2$s.",
        paste0(subname, opt_name), opt_check1$oklen[1]
      ))
    }
    if (length(opt_value) > abs(opt_check1$oklen[2])) {
      if (opt_check1$oklen[2] > 0L) {
        lav_msg_stop(gettextf(
          "Length of option '%1$s' value must be maximum %2$s.",
          paste0(subname, opt_name), opt_check1$oklen[2]
        ))
      } else {
        lav_msg_warn(gettextf(
          "Length of option '%1$s' value should be maximum %2$s.
          Only first %3$s elements used.",
          paste0(subname, opt_name), -opt_check1$oklen[2],
          -opt_check1$oklen[2]
        ))
      }
    }
    if (is.null(opt_check1$bl)) opt_check1$bl <- FALSE
    if (!is.null(opt_check1$chr) || !is.null(opt_check1$nm) ||
      opt_check1$bl) {
      if (!opt_check1$bl || !is.logical(opt_value)) {
        if (!is.null(opt_check1$nm) && is.numeric(opt_value)) {
          num2int <- FALSE
          if (!is.null(opt_check1$num2int)) num2int <- opt_check1$num2int
          if (!lav_options_checkinterval(opt_value, opt_check1$nm, num2int)) {
            lav_msg_stop(gettextf(
              "Value(s) of option %1$s out of range (%2$s)!",
              paste0(subname, opt_name),
              paste0(
                opt_check1$nm$bounds[1],
                if (opt_check1$nm$first.in) " <= " else " < ",
                "x",
                if (opt_check1$nm$last.in) " <= " else " < ",
                opt_check1$nm$bounds[2]
              )
            ))
          }
        }
        if (!is.null(opt_check1$chr) && is.character(opt_value)) {
          opt_value <- lav_options_checkvalues(
            opt_name, opt_value,
            opt_check1$chr
          )
          opts[[j]] <- opt_value
        }
      }
    }
  }
  if (any(hiddens)) { # add hidden options
    opts <- modifyList(opts, opts_hidden)
  }
  opts
}
# this function collects and checks the user-provided options/arguments,
# and fills in the "default" values, or changes them in an attempt to
# produce a consistent set of values...
#
# returns a list with the named options
lav_options_set <- function(opt = NULL) {
  # check the presence of necessary hidden options ####
  if (is.null(opt$.categorical) || is.null(opt$.multilevel) ||
    is.null(opt$.clustered)) {
    lav_msg_fixme(
      ".categorical, .multilevel and .clustered must be present"
    )
  }

  # get opt_default and opt_check ####
  if (!exists("opt_check", lavaan_cache_env)) lav_options_default()
  opt_check <- get("opt_check", lavaan_cache_env)

  if (lav_debug()) {
    cat("lavaan DEBUG: lavaanOptions IN\n")
    str(opt)
    opt$optim.partrace <- TRUE
  }

  # options OPT for which there also exist OPT.args and opt$OPT is a list
  # -> split in opt$OPT and opt$OPT.args
  welke <- which(paste0(names(opt_check), ".args") %in% names(opt_check))
  for (j in welke) {
    optname <- names(opt_check)[j]
    optname_args <- paste0(optname, ".args")
    if (is.list(opt[[optname]])) {
      if (optname == "bootstrap" && !is.null(opt[[optname]][["R"]])) {
        # special case for bootstrap, instead of "bootstrap" replications
        # argument can be "R"
        opt[[optname]][[optname]] <- opt[[optname]][["R"]]
        opt[[optname]][["R"]] <- NULL
      }
      opt[[optname_args]] <- opt[[optname]]
      opt[[optname_args]][[optname]] <- NULL
      opt[[optname]] <- opt[[optname]][[optname]]
    }
  }

  if (is.list(opt$rotation.args)) {
    opt$rotation.args <- lav_snake_case(opt$rotation.args)
  }
  if (is.list(opt$em.args)) {
    opt$em.args <- lav_snake_case(opt$em.args)
  }
  if (is.list(opt$em.h1.args)) {
    opt$em.h1.args <- lav_snake_case(opt$em.h1.args)
  }

  # 'correlation' may be a character vector of (observed) variable names:
  # in that case, only those variables are standardized to unit variance
  # (a 'partial' correlation structure). We store the names in a hidden
  # option (.correlation.ov) and treat 'correlation' as a logical TRUE, so
  # the rest of the machinery (and all `if (correlation)` checks) work as
  # before. An empty .correlation.ov means 'all observed variables'.
  if (is.null(opt$.correlation.ov)) {
    opt$.correlation.ov <- character(0L)
  }
  if (is.character(opt$correlation)) {
    opt$.correlation.ov <- opt$correlation
    opt$correlation <- TRUE
  }

  # check options with definitions ####
  opt <- lav_options_check(opt, opt_check, "")

  # check option 'start'
  if (is.character(opt$start) && all(opt$start != c("default", "simple"))) {
    lav_msg_stop(gettext(
      "start option must be 'default', 'simple', a fitted object,
      a vector of parameter values, or a parameter table"
    ))
  }

  # first of all: set estimator ####
  if (opt$estimator == "default") {
    if (opt$.categorical) {
      opt$estimator <- "wlsmv"
    } else {
      opt$estimator <- "ml"
    }
  }

  # defaults for opt$sample.cov.rescale
  if (opt$sample.cov.rescale == "default") {
    opt$sample.cov.rescale <- switch(opt$estimator,
      dls = TRUE,
      fabin2 = ,
      fabin3 = ,
      mgm = ,
      js = ,
      jsa = ,
      bentler1982 = TRUE,
      iv = TRUE,
      "default"
    )
  }

  # option defaults specific for mimic=...
  opt <- lav_options_mimic(opt)

  # store opt$estimator as estimator.orig in upper case
  opt$estimator.orig <- toupper(opt$estimator)

  # rename names of test statistics if needed, check for invalid values ####
  opt$test <- lav_test_rename(opt$test, check = TRUE)

  # same for standard.test
  opt$standard.test <- lav_test_rename(opt$standard.test, check = TRUE)

  # same for scaled.test
  opt$scaled.test <- lav_test_rename(opt$scaled.test, check = TRUE)

  # rename names of se values, check illegal combinations se/estimator ####
  # pass-through function: may change value of information
  # for backwards compatibility (eg if se = "expected")
  opt <- lav_options_check_se(opt)

  # do.fit FALSE implies se="none" and test="none" (unless not default) ####
  if (!opt$do.fit) {
    if (opt$se == "default") opt$se <- "none"
    if (opt$test[1] == "default") opt$test <- "none"
  }

  # marker.int.fixed ####
  if (opt$marker.int.zero) {
    opt$meanstructure <- TRUE
    opt$int.ov.free <- TRUE
    if ((is.logical(opt$effect.coding) && opt$effect.coding) ||
      (is.character(opt$effect.coding) && nchar(opt$effect.coding) > 0L)) {
      lav_msg_stop(gettext(
        "effect coding cannot be combined with marker.int.zero = TRUE option"
      ))
    }
    if (opt$std.lv) {
      lav_msg_stop(gettext(
        "std.lv = TRUE cannot be combined with marker.int.zero = TRUE"
      ))
    }
  }

  # group.equal and group.partial ####
  if (length(opt$group.equal) > 0L && opt$group.equal[1] == "none") {
    opt$group.equal <- character(0)
  } else if (is.null(opt$group.equal) || all(nchar(opt$group.equal) == 0L)) {
    opt$group.equal <- character(0)
  } else if ("all" %in% opt$group.equal) {
    # "all": constrain all (free) parameters to be equal across groups
    # (expand to the full set of parameter types; any other elements that
    #  were given together with "all" become redundant)
    opt$group.equal <- c("loadings", "composite.weights", "intercepts",
                         "means", "thresholds", "regressions", "residuals",
                         "residual.covariances", "lv.variances",
                         "lv.covariances")
  }

  if (is.null(opt$group.partial) || all(nchar(opt$group.partial) == 0L)) {
    opt$group.partial <- character(0)
  } else if (length(opt$group.partial) == 0) {
    # nothing to do
  } else {
    # strip white space
    opt$group.partial <- gsub("[[:space:]]+", "", opt$group.partial)
  }

  # if categorical, and group.equal contains "intercepts", also add
  # thresholds (and vice versa)
  # not any longer since 0.6-20
  # if (opt$.categorical && any("intercepts" == opt$group.equal)) {
  #  opt$group.equal <- unique(c(opt$group.equal, "thresholds"))
  # }
  # if (opt$.categorical && any("thresholds" == opt$group.equal)) {
  #  opt$group.equal <- unique(c(opt$group.equal, "intercepts"))
  # }

  # clustered ####
  # brute-force override (for now)
  if (opt$.clustered && !opt$.multilevel) {
    opt$meanstructure <- TRUE

    # single-level clustered continuous ML: the pairwise / available.cases
    # missing-data methods are not supported (they were silently fed into the
    # cluster-robust machinery, producing nonsense estimates and a failing
    # baseline model, or crashing deep inside with a cryptic "NA/NaN/Inf in
    # foreign function call" error); the categorical (D)WLS + cluster path
    # handles these via muthen1984() and is left untouched
    if (!opt$.categorical &&
        any(opt$missing == c("pairwise", "available.cases"))) {
      lav_msg_stop(gettextf(
        "missing = %1$s is not supported for clustered (cluster =) continuous
         models; use missing = %2$s or missing = %3$s instead.",
        dQuote(opt$missing), dQuote("ml"), dQuote("listwise")))
    }

    if (opt$estimator == "mlr") {
      opt$estimator <- "ml"
      opt$test <- "yuan.bentler.mplus"
      opt$se <- "robust.cluster"
    } else if (opt$estimator == "mlm") {
      opt$estimator <- "ml"
      opt$test <- "satorra.bentler"
      opt$se <- "robust.cluster.sem"
    } else if (opt$.categorical && opt$estimator != "pml") {
      opt$test <- "satorra.bentler"
      opt$se <- "robust.cluster.sem"
    }

    # test ####
    if (length(opt$test) == 1L && opt$test == "default") {
      opt$test <- "yuan.bentler.mplus"
    } else if (all(opt$test %in% c(
      "none", "standard",
      "satorra.bentler",
      "yuan.bentler", "yuan.bentler.mplus"
    ))) {
      # nothing to do
    } else if (opt$se == "robust") {
      opt$test <- "yuan.bentler.mplus"
    } else {
      lav_msg_stop(
        gettextf(
          "`test' argument must be one of %s in the clustered case",
          lav_msg_view(c(
            "none", "yuan.bentler", "yuan.bentler.mplus",
            "satorra.bentler"
          ), log_sep = "or")
        )
      )
    }

    # se ####
    if (opt$se == "default") {
      opt$se <- "robust.cluster"
    } else if (any(opt$se == c(
      "none", "robust.cluster",
      "robust.cluster.sem"
    ))) {
      # nothing to do
    } else if (opt$se == "robust") {
      opt$se <- "robust.cluster"
    }

    # information ####
    if (opt$information[1] == "default") {
      if (opt$se == "robust.cluster" && opt$estimator == "ml") {
        opt$information[1] <- "observed"
      } else {
        opt$information[1] <- "expected"
      }
    }
    if (length(opt$information) > 1L && opt$information[2] == "default") {
      if (opt$se == "robust.cluster") {
        opt$information[2] <- "observed"
      } else {
        opt$information[2] <- "expected"
      }
    }
  }

  # cluster-robust SEs were explicitly requested, but no cluster= variable
  # was given: the requested estimator is undefined without cluster data
  # (before, lavaan silently computed a non-cluster-robust sandwich that did
  # not match any documented estimator)
  if (!opt$.clustered &&
      any(opt$se == c("robust.cluster", "robust.cluster.sem"))) {
    lav_msg_stop(gettextf(
      "se = %s requires clustered data; please provide the cluster= argument.",
      dQuote(opt$se)))
  }

  # composites ####
  # resolve composites.cov = "default": the composite-indicator (co)variances
  # (the 'T' matrix) are fixed to their sample values for single-level models,
  # but estimated freely for multilevel models (where the same indicators may
  # serve as composite indicators at more than one level, and fixing T at every
  # level would over-constrain the model).
  if (identical(opt$composites.cov, "default")) {
    if (opt$.multilevel) {
      opt$composites.cov <- "free"
    } else {
      opt$composites.cov <- "fixed"
    }
  }

  # multilevel ####
  # brute-force override (for now)
  if (opt$.multilevel) {
    opt$meanstructure <- TRUE

    # no estimator supports sampling weights in the two-level case (yet);
    # before 0.7-1 the ML machinery silently ignored them
    if (!is.null(opt$.sampling.weights) && opt$.sampling.weights) {
      lav_msg_stop(gettext(
        "sampling weights are not supported for two-level models (yet)."))
    }

    # only listwise deletion and full-information ML can handle missing
    # values in the two-level case; the other methods (two.stage,
    # robust.two.stage, pairwise, available.cases, doubly.robust) are
    # single-level methods (before 0.7-1 these crashed deep inside the
    # saturated-model EM with a cryptic error)
    if (any(opt$missing == c(
      "two.stage", "robust.two.stage", "pairwise",
      "available.cases", "doubly.robust"
    ))) {
      lav_msg_stop(gettextf(
        "missing = %1$s is not supported for two-level models; use
         missing = %2$s or missing = %3$s instead.",
        dQuote(opt$missing), dQuote("ml"), dQuote("listwise")))
    }

    # two-level + conditional.x: the expected information would need the
    # Delta matrix in the unconditional statistic space (the parameter-space
    # kernel lav_mvn_cl_info_expected_delta works with unconditional
    # moments); until that conversion exists, only observed/first.order
    # information is supported (the default)
    if (isTRUE(opt$conditional.x) && any(opt$information == "expected")) {
      lav_msg_stop(gettext(
        "information = \"expected\" is not supported (yet) for two-level
         models with conditional.x = TRUE; use \"observed\" (the default)
         or \"first.order\"."))
    }

    # ditto for a user-requested unstructured h1: the unstructured
    # (unconditional) h1 information cannot be combined with the
    # conditional.x Delta (before 0.7-1 this crashed with a cryptic
    # "non-conformable arguments" error); note that the internal
    # yuan.bentler.mplus route (which never mixes the two spaces) does
    # not pass through this check
    if (isTRUE(opt$conditional.x) &&
        any(opt$h1.information == "unstructured")) {
      lav_msg_stop(gettext(
        "h1.information = \"unstructured\" is not supported (yet) for
         two-level models with conditional.x = TRUE."))
    }

    # two-level least-squares estimation? (WLS/WLSM(V)/ULS(M,MV)/DWLS)
    multilevel_wls <-
      lav_options_estimatorgroup(opt$estimator) %in% c("WLS", "DWLS", "ULS")

    if (multilevel_wls) {
      # two-level (D)WLS: continuous, complete data (for now)
      if (!opt$missing %in% c("default", "listwise")) {
        lav_msg_stop(gettextf(
          "missing = %s is not supported for two-level (D)WLS estimation;
          only complete data (missing = \"listwise\") is supported (for now).",
          dQuote(opt$missing)))
      }
      if (isTRUE(opt$conditional.x) && !opt$.categorical) {
        lav_msg_stop(gettext(
          "conditional.x = TRUE is not supported for two-level (D)WLS
          estimation with continuous-only data (yet)."))
      }
      if (opt$.categorical) {
        # categorical two-level (D)WLS:
        # - the stage-wise estimation works in the theta parameterization
        #   (the within-level latent response variances are fixed to 1)
        if (opt$parameterization == "default") {
          opt$parameterization <- "theta"
        } else if (opt$parameterization != "theta") {
          lav_msg_stop(gettext(
            "two-level (D)WLS estimation with categorical data requires
            parameterization = \"theta\"."))
        }
        # - no (continuous) h1 model; the unrestricted statistics come
        #   from the stage-wise (univariate + bivariate) estimation
        opt$h1 <- FALSE
      } else {
        # the sample 'statistics' are the h1 (saturated) estimates
        opt$h1 <- TRUE
      }

      # test
      if (length(opt$test) == 1L && opt$test == "default") {
        # ok, will be set later
      } else if (all(opt$test %in% c(
        "none", "standard", "satorra.bentler",
        "scaled.shifted", "mean.var.adjusted"
      ))) {
        # nothing to do
      } else {
        lav_msg_stop(gettextf(
          "`test' argument must be one of %s in the two-level (D)WLS case",
          lav_msg_view(c(
            "none", "standard", "satorra.bentler",
            "scaled.shifted", "mean.var.adjusted"
          ), log_sep = "or")
        ))
      }

      # se
      if (opt$se == "default") {
        # ok, will be set later
      } else if (any(opt$se == c(
        "none", "standard", "robust.sem", "bootstrap"
      ))) {
        # nothing to do (bootstrap uses the cluster bootstrap)
      } else if (opt$se == "robust") {
        opt$se <- "robust.sem"
      } else {
        lav_msg_stop(gettextf(
          "`se' argument must be one of %s in the two-level (D)WLS case",
          lav_msg_view(c("none", "standard", "robust.sem", "bootstrap"),
            log_sep = "or"
          )
        ))
      }

      # information
      if (opt$information[1] == "default") {
        opt$information[1] <- "expected"
      }
      if (length(opt$information) > 1L && opt$information[2] == "default") {
        opt$information[2] <- "expected"
      }
    } else {
      # test
      if (length(opt$test) == 1L && opt$test == "default") {
        # ok, will be set later
      } else if (all(opt$test %in% c(
        "none", "standard", "yuan.bentler",
        "yuan.bentler.mplus"
      ))) {
        # nothing to do
      } else {
        lav_msg_stop(gettextf(
          "`test' argument must be one of %s in the multilevel case",
          lav_msg_view(c(
            "none", "standard", "yuan.bentler",
            "yuan.bentler.mplus"
          ), log_sep = "or")
        ))
      }

      # se
      if (opt$se == "default") {
        # ok, will be set later
      } else if (any(opt$se == c(
        "none", "standard", "robust.huber.white", "sandwich", "bootstrap"
      ))) {
        # nothing to do (bootstrap uses the cluster bootstrap; see lavBootstrap)
      } else if (opt$se == "robust") {
        opt$se <- "robust.huber.white"
      } else {
        lav_msg_stop(gettextf(
          "`se' argument must be one of %s in the multilevel case",
          lav_msg_view(c("none", "standard", "robust.huber.white", "bootstrap"),
            log_sep = "or"
          )
        ))
      }

      # information
      if (opt$information[1] == "default") {
        opt$information[1] <- "observed"
      }
      if (length(opt$information) > 1L && opt$information[2] == "default") {
        opt$information[2] <- "observed"
      }
    }
  }

  # random slopes (rv() modifier) ####
  if (is.null(opt$.rv)) {
    opt$.rv <- FALSE
  }
  if (opt$.rv) {
    # random slopes are only available for two-level models with
    # continuous data, complete data, estimator ML (for now)
    if (!opt$.multilevel) {
      lav_msg_stop(gettext(
        "random slopes (rv() modifier) are only supported for two-level
         models (with a cluster= argument)."))
    }
    if (opt$.categorical) {
      lav_msg_stop(gettext(
        "random slopes (rv() modifier) are not supported for categorical
         data."))
    }
    if (!is.null(opt$.sampling.weights) && opt$.sampling.weights) {
      lav_msg_stop(gettext(
        "random slopes (rv() modifier) are not supported in combination
         with sampling weights."))
    }
    if (!opt$estimator %in% c("ml", "mlr", "default")) {
      lav_msg_stop(gettextf(
        "random slopes (rv() modifier) require estimator = %s or %s.",
        dQuote("ML"), dQuote("MLR")))
    }
    if (opt$missing %in% c("ml.x", "two.stage", "robust.two.stage")) {
      lav_msg_stop(gettextf(
        "random slopes (rv() modifier) support missing = %s or
         missing = %s only.", dQuote("listwise"), dQuote("ml")))
    }
    if (isTRUE(opt$conditional.x)) {
      lav_msg_stop(gettext(
        "random slopes (rv() modifier) are not supported in combination
         with conditional.x = TRUE."))
    }
    # note: fixed.x = FALSE may simply mean that the model contains no
    # observed exogenous covariates at all (e.g., all random slopes
    # involve latent covariates); if an *observed* covariate does carry
    # a random slope while fixed.x = FALSE, lav_mvn_cl_rs_info() stops
    # with a specific message (the covariate must be exogenous/fixed.x)
    # no saturated (h1) model exists for random-slope models: no
    # chi-square test statistic, no baseline model, no (incremental)
    # fit indices; only logl/AIC/BIC are available
    # note: we do keep h1 = TRUE: the (standard, no-random-slope)
    # saturated two-level moments provide good starting values; but the
    # h1 loglikelihood is NOT comparable to the (conditional-on-x)
    # loglikelihood of the random-slope model
    opt$baseline <- FALSE
    opt$fit.by.level <- FALSE
    # note: estimator = "MLR" implies test = "yuan.bentler.mplus" (set
    # in lav_options_mimic); this is not an explicit user request, so
    # we do not warn about it
    if (!all(opt$test %in% c("default", "none")) &&
        !(opt$estimator == "mlr" &&
          all(opt$test %in% c("yuan.bentler", "yuan.bentler.mplus")))) {
      lav_msg_warn(gettext(
        "test statistics are not available (yet) for models with random
         slopes; test set to \"none\"."))
    }
    opt$test <- "none"
    if (opt$se == "robust") {
      opt$se <- "robust.huber.white"
    }
    if (!opt$se %in% c("default", "none", "standard",
                       "robust.huber.white")) {
      lav_msg_stop(gettextf(
        "`se' argument must be one of %s for models with random slopes.",
        lav_msg_view(c("none", "standard", "robust.huber.white"),
                     log_sep = "or")))
    }
    # standard errors: numerical hessian of the observed-data loglikelihood
    opt$observed.information[1] <- "hessian"
    if (length(opt$observed.information) > 1L) {
      opt$observed.information[2] <- "hessian"
    }
    # the gradient is analytic (adjoint pass; numerical fallback);
    # optim.gradient = "numerical" remains available as an escape hatch
    # optimizer: nlminb (default) or em
    if (opt$optim.method == "gn") {
      lav_msg_warn(gettext(
        "optim.method = \"gn\" is not available for models with random
         slopes; switching to nlminb."))
      opt$optim.method <- "nlminb"
    }
  }

  # em.h1.args: resolve "default" values
  # note: the single-level (missing data) EM converges when the change in
  # the parameter values is smaller than tol (tol was 1e-06 < 0.6-9), while
  # the multilevel EM converges when the change in the loglikelihood is
  # smaller than tol
  if (identical(opt$em.h1.args$max_iter, "default")) {
    if (opt$.multilevel) {
      opt$em.h1.args$max_iter <- 5000L
    } else {
      opt$em.h1.args$max_iter <- 500L
    }
  }
  if (identical(opt$em.h1.args$tol, "default")) {
    if (opt$.multilevel) {
      opt$em.h1.args$tol <- 1e-04
    } else {
      opt$em.h1.args$tol <- 1e-05
    }
  }

  # missing ####
  if (opt$missing == "default") {
    opt$missing <- "listwise"
  } else if (opt$missing == "ml") {
    if (opt$.categorical) {
      lav_msg_stop(gettextf(
        "missing = %s not available in the categorical setting",
        dQuote(opt$missing)
      ))
    }
    # for the (continuous) least-squares estimators, ML is not available;
    # interpret missing = "ml"/"fiml" as a generic request to handle missing
    # values, and silently switch to two-stage estimation
    if (any(opt$estimator == c("uls", "gls", "wls", "dls"))) {
      opt$missing <- "two.stage"
    } else if (any(opt$estimator == c(
      "mlm", "mlmv", "wlsm", "wlsmv", "ulsm", "ulsmv", "pml"
    ))) {
      lav_msg_stop(gettextf(
        "missing=%1$s is not allowed for estimator %2$s",
        dQuote(opt$missing),
        dQuote(lav_options_estimatorgroup(opt$estimator.orig))
      ))
    }
  } else if (opt$missing == "ml.x") {
    if (opt$.categorical) {
      lav_msg_stop(gettextf(
        "missing = %s not available in the categorical setting",
        dQuote(opt$missing)
      ))
    }
    # see missing = "ml" above: handle missing values via two-stage
    if (any(opt$estimator == c("uls", "gls", "wls", "dls"))) {
      opt$missing <- "two.stage"
    } else if (any(opt$estimator == c(
      "mlm", "mlmv", "wlsm", "wlsmv", "ulsm", "ulsmv", "pml"
    ))) {
      lav_msg_stop(gettextf(
        "missing=%1$s is not allowed for estimator %2$s",
        dQuote(opt$missing),
        dQuote(lav_options_estimatorgroup(opt$estimator.orig))
      ))
    }
  } else if (opt$missing == "two.stage") {
    if (opt$.categorical) {
      lav_msg_stop(gettextf(
        "missing = %s not available in the categorical setting",
        dQuote(opt$missing)
      ))
    }
    # uls/gls/wls/dls are allowed (continuous data only): they use the
    # two-stage EM moments for estimation and the robust.sem sandwich for SEs
    if (any(opt$estimator == c(
      "mlm", "mlmv", "wlsm", "wlsmv",
      "ulsm", "ulsmv", "pml", "mml"
    ))) {
      lav_msg_stop(gettextf(
        "missing=%1$s is not allowed for estimator %2$s",
        dQuote(opt$missing),
        dQuote(lav_options_estimatorgroup(opt$estimator.orig))
      ))
    }
  } else if (opt$missing == "robust.two.stage") {
    if (opt$.categorical) {
      lav_msg_stop(gettextf(
        "missing = %s not available in the categorical setting",
        dQuote(opt$missing)
      ))
    }
    if (any(opt$estimator == c(
      "mlm", "mlmv", "wlsm", "wlsmv",
      "ulsm", "ulsmv", "pml", "mml"
    ))) {
      lav_msg_stop(gettextf(
        "missing=%1$s is not allowed for estimator %2$s",
        dQuote(opt$missing),
        dQuote(lav_options_estimatorgroup(opt$estimator.orig))
      ))
    }
  } else if (opt$missing == "doubly.robust") {
    if (opt$estimator != "pml") {
      lav_msg_stop(gettextf(
        "missing=%s option only available for estimator PML",
        dQuote(opt$missing)
      ))
    }
  }

  # check missing ####
  if (any(opt$missing == c("ml", "ml.x")) &&
    opt$se %in% c("robust.sem", "robust.sem.nt")) {
    lav_msg_warn(gettextf(
      "missing will be set to %1$s for se = %2$s.",
      dQuote("listwise"), dQuote(opt$se)
    ))
    opt$missing <- "listwise"
  }
  if (any(opt$missing == c("ml", "ml.x")) &&
    any(opt$test %in% c(
      "satorra.bentler",
      "mean.var.adjusted", "scaled.shifted",
      "mean.var.adjusted.corrected", "scaled.shifted.corrected"
    ))) {
    lav_msg_warn(gettextf(
      "missing will be set to %s for satorra.bentler style test",
      dQuote("listwise")
    ))
    opt$missing <- "listwise"
  }

  # checks if missing = "two.stage" or "robust.two.stage" ####
  if (any(opt$missing == c("two.stage", "robust.two.stage"))) {
    opt$meanstructure <- TRUE
    # the (continuous) least-squares estimators do not have a dedicated
    # two-stage SE machinery; instead, they reuse the standard robust.sem
    # sandwich, with the two-stage NACOV of the EM moments (computed in
    # lav_samplestats); only the (point) estimation uses the EM moments
    ls_two_stage <- any(opt$estimator == c("uls", "gls", "wls", "dls"))
    # se
    if (ls_two_stage) {
      if (opt$se == "default") {
        opt$se <- "robust.sem"
      } else if (!any(opt$se == c("none", "robust.sem", "robust.sem.nt"))) {
        lav_msg_warn(gettextf(
          "se will be set to %1$s if missing = %2$s and estimator = %3$s",
          dQuote("robust.sem"), dQuote(opt$missing),
          dQuote(lav_options_estimatorgroup(opt$estimator.orig))
        ))
        opt$se <- "robust.sem"
      }
    } else if (opt$se == "default") {
      if (opt$missing == "two.stage") {
        opt$se <- "two.stage"
      } else {
        opt$se <- "robust.two.stage"
      }
    } else if (opt$missing == "two.stage" &&
      opt$se == "two.stage") {
      # nothing to do
    } else if (opt$missing == "robust.two.stage" &&
      opt$se == "robust.two.stage") {
      # nothing to do
    } else {
      lav_msg_warn(gettextf(
        "se will be set to %1$s if missing = %2$s",
        dQuote(opt$missing), dQuote(opt$missing)
      ))
      opt$se <- opt$missing
    }

    # the information/observed.information overrides below are specific to the
    # ML two-stage SE machinery; the least-squares + robust.sem path keeps the
    # estimator's own (default) information settings
    if (!ls_two_stage) {
      # information
      if (opt$information[1] == "default") {
        # for both two.stage and robust.two.stage
        opt$information[1] <- "observed"
      } else if (opt$information[1] == "first.order") {
        lav_msg_warn(gettextf(
          "information will be set to %1$s if missing = %2$s",
          dQuote("observed"), dQuote(opt$missing)
        ))
        opt$information[1] <- "observed"
      }

      # observed.information (ALWAYS "h1" for now)
      opt$observed.information[1] <- "h1"
      opt$observed.information[2] <- "h1"

      # new in 0.6-9: ALWAYS h1.information = "unstructured"
      opt$h1.information <- c("unstructured", "unstructured")

      if (length(opt$information) > 1L && opt$information[2] == "default") {
        # for both two.stage and robust.two.stage
        opt$information[2] <- "observed"
      }
    }

    # test
    if (length(opt$test) > 1L) {
      lav_msg_warn(gettextf(
        "test= argument can only contain a single element if missing = %s
        (taking the first)", dQuote(opt$missing)
      ))
      opt$test <- opt$test[1]
    }

    if (length(opt$test) == 1L && opt$test == "default") {
      opt$test <- "satorra.bentler"
    } else if (length(opt$test) == 1L && any(
      opt$test == c("satorra", "sb", "satorra.bentler", "satorra-bentler")
    )) {
      opt$test <- "satorra.bentler"
    } else {
      lav_msg_warn(gettextf(
        "test will be set to %1$s if missing = %2$s",
        dQuote("satorra.bentler"), dQuote(opt$missing)
      ))
      opt$test <- "satorra.bentler"
    }
  }

  # meanstructure ####
  if (is.logical(opt$meanstructure)) {
    if (opt$meanstructure == FALSE) {
      if (any(opt$missing == c("ml", "ml.x", "two.stage"))) {
        lav_msg_warn(gettextf(
          "missing argument %s forces meanstructure = TRUE",
          opt$missing
        ))
      }
    }
  } else if (opt$meanstructure == "default") {
    # by default: no meanstructure!
    if (opt$estimator == "pml") {
      opt$meanstructure <- TRUE
    } else {
      opt$meanstructure <- FALSE
    }
  }

  # force meanstructure = TRUE if either int.lv.free or int.ov.free is TRUE
  # but only if lavaan() was used
  if (opt$cmd == "lavaan" && (opt$int.lv.free || opt$int.ov.free)) {
    opt$meanstructure <- TRUE
  }

  # bootstrap ####
  if (is.numeric(opt$bootstrap)) {
    opt$bootstrap <- list(R = as.integer(opt$bootstrap))
  }
  if (opt$se == "bootstrap") {
    opt$information[1] <- "observed"
    if (length(opt$information) > 1L && opt$information[2] == "default") {
      opt$information[2] <- "observed"
    }
  }

  # specific per estimator (group) ####
  opt <- switch(opt$estimator,
    ml = ,
    mlf = ,
    mlm = ,
    mlmv = ,
    mlmvs = ,
    mlr = lav_options_est_ml(opt),
    gls = lav_options_est_gls(opt),
    ntrls = lav_options_est_ntrls(opt),
    catml = lav_options_est_catml(opt),
    wls = lav_options_est_wls(opt),
    dls = lav_options_est_dls(opt),
    dwls = ,
    wlsm = ,
    wlsmv = ,
    wlsmvs = lav_options_est_dwls(opt),
    uls = ,
    ulsm = ,
    ulsmv = ,
    ulsmvs = lav_options_est_uls(opt),
    pml = lav_options_est_pml(opt),
    fml = lav_options_est_fml(opt),
    reml = lav_options_est_reml(opt),
    mml = lav_options_est_mml(opt),
    fabin2 = ,
    fabin3 = ,
    mgm = ,
    js = ,
    jsa = ,
    bentler1982 = lav_options_est_fabin(opt),
    iv = lav_options_est_iv(opt),
    rbm = lav_options_est_rbm(opt),
    lav_options_est_none(opt) # estimator = none
  )

  # after code specific to estimator types                          ####
  # optim.method - if still "default" at this point -> set to "nlminb";
  # new in 0.7-1: for two-level models with missing = "ml" (and an
  # ML-family estimator), the EM algorithm is the default optimizer
  # (more robust; this mimics Mplus); if the EM optimizer turns out not
  # to support the model (e.g., multiple groups, conditional.x), we
  # quietly fall back to "nlminb" later (see lav_lavaan_step11_optim)
  if (opt$optim.method == "default") {
    if (opt$.multilevel && opt$missing %in% c("ml", "ml.x") &&
        lav_options_estimatorgroup(opt$estimator) == "ML") {
      opt$optim.method <- "em"
      opt$.optim.em.fallback <- TRUE
    } else {
      opt$optim.method <- "nlminb"
    }
  }

  # special stuff for categorical
  if (opt$.categorical) {
    opt$meanstructure <- TRUE # Mplus style
    if (lav_options_estimatorgroup(opt$estimator) == "ML") {
      lav_msg_stop(gettext(
        "estimator ML for ordered data is not supported yet; use WLSMV instead."
      ))
    }
  }

  # link
  if (opt$link == "logit") {
    if (opt$estimator != "mml") {
      lav_msg_warn(gettextf(
        "link will be set to %1$s for estimator = %2$s",
        dQuote("probit"), dQuote(opt$estimator)
      ))
    }
  }

  # likelihood approach (wishart or normal) + sample.cov.rescale
  if (!any(lav_options_estimatorgroup(opt$estimator) ==
    c("ML", "REML", "PML", "FML", "NTRLS", "catML"))) {
    # if(opt$likelihood != "default") {
    #    lav_msg_stop(gettext(
    #    "likelihood argument is only relevant if estimator = ML"))
    # }
    if (opt$sample.cov.rescale == "default") {
      opt$sample.cov.rescale <- FALSE
    } # else {
    #    lav_msg_warn(gettext(
    #    "sample.cov.rescale argument is only relevant if estimator = ML"))
    # }
  } else { # ml and friends
    if (any(lav_options_estimatorgroup(opt$estimator) == c("PML", "FML"))) {
      opt$likelihood <- "normal"
    } else if (opt$likelihood == "default") {
      opt$likelihood <- "normal"
    }

    if (opt$sample.cov.rescale == "default") {
      opt$sample.cov.rescale <- FALSE
      if (opt$likelihood == "normal") {
        opt$sample.cov.rescale <- TRUE
      }
    }
  }

  # se information
  if (opt$information[1] == "default") {
    if (any(opt$missing == c("ml", "ml.x")) ||
      any(opt$se == c("robust.huber.white", "first.order"))) {
      # nchar(opt$constraints) > 0L) {
      opt$information[1] <- "observed"
    } else {
      opt$information[1] <- "expected"
    }
  }

  # first.order information can not be used with robust
  if (opt$information[1] == "first.order" &&
    any(opt$se == c("robust.huber.white", "robust.sem", "robust.sem.nt"))) {
    lav_msg_stop(gettextf(
      "information must be either %s if robust standard errors are requested.",
      lav_msg_view(c("expected", "observed"), log_sep = "or")
    ))
  }

  # test information
  if (length(opt$information) == 1L) {
    opt$information <- rep(opt$information, 2L)
  }
  if (opt$information[2] == "default") {
    if (any(opt$missing == c("ml", "ml.x")) ||
      any(opt$se == c("robust.huber.white", "first.order"))) {
      # nchar(opt$constraints) > 0L) {
      opt$information[2] <- "observed"
    } else {
      opt$information[2] <- "expected"
    }
  }

  # first.order information cannot be used with robust
  if (opt$information[2] == "first.order" &&
    any(opt$test %in% c(
      "satorra.bentler", "yuan.bentler",
      "yuan.bentler.mplus",
      "mean.var.adjusted", "scaled.shifted",
      "mean.var.adjusted.corrected", "scaled.shifted.corrected"
    ))) {
    lav_msg_stop(gettextf(
      "information must be either %s if robust test statistics are requested.",
      lav_msg_view(c("expected", "observed"), log_sep = "or")
    ))
  }


  if (length(opt$observed.information) == 1L) {
    opt$observed.information <- rep(opt$observed.information, 2L)
  }

  if (all(opt$observed.information[2] != c("hessian", "h1"))) {
    if (opt$observed.information[2] == "default") {
      if (any(opt$test %in% c(
        "satorra.bentler",
        "yuan.bentler",
        "yuan.bentler.mplus",
        "mean.var.adjusted",
        "mean.var.adjusted.corrected",
        "scaled.shifted",
        "scaled.shifted.corrected"
      ))) {
        if (length(opt$test) > 1L) {
          opt$observed.information[2] <- "h1" # CHANGED in 0.6-6!
          if (any(opt$test == "yuan.bentler.mplus")) {
            lav_msg_warn(gettext(
              "observed.information for ALL test statistics is set to h1."
            ))
          }
        } else {
          if (opt$estimator == "PML" ||
            opt$test[1] == "yuan.bentler.mplus") {
            opt$observed.information[2] <- "hessian"
          } else {
            opt$observed.information[2] <- "h1" # CHANGED in 0.6-6!
          }
        }
      } else {
        # default is "hessian"
        opt$observed.information[2] <- "hessian"
      }
    }
  }

  if (length(opt$h1.information) == 1L) {
    opt$h1.information <- rep(opt$h1.information, 2L)
  }

  if (opt$h1.information.meat == "default") {
    opt$h1.information.meat <- opt$h1.information[1]
  }

  # check information if estimator is uls/wls and friends
  if (any(lav_options_estimatorgroup(opt$estimator) ==
    c("ULS", "WLS", "DWLS"))) {
    #if (opt$information[1] != "expected") {
    #  lav_msg_warn(gettextf(
    #    "information will be set to %1$s for estimator = %2$s",
    #    dQuote("expected"), dQuote(opt$estimator)
    #  ))
    #  opt$information <- rep.int("expected", 2L)
    #}
    opt$h1.information <- rep.int("unstructured", 2L) # should always be fixed
  }


  # omega information
  if (opt$omega.information == "default") {
    opt$omega.information <- opt$information[2] # test version!
  }

  if (opt$omega.h1.information == "default") {
    # opt$omega.h1.information <- opt$h1.information[2] # test version!
    opt$omega.h1.information <- "unstructured"
  }

  if (opt$omega.h1.information.meat == "default") {
    opt$omega.h1.information.meat <- opt$omega.h1.information
  }

  # conditional.x
  if (is.character(opt$conditional.x)) { # = "default"
    if (opt$.categorical) {
      opt$conditional.x <- TRUE
    } else {
      opt$conditional.x <- FALSE
    }
  }

  # if conditional.x, always use a meanstructure
  if (opt$conditional.x) {
    opt$meanstructure <- TRUE
    # auto.cov.x has no effect if conditional.x = TRUE: the exogenous
    # covariates are conditioned out, so their covariances with exogenous
    # latent variables cannot be represented
    opt$auto.cov.x <- FALSE
  }

  # fixed.x
  if (is.logical(opt$fixed.x)) {
    # if(opt$conditional.x && opt$fixed.x == FALSE && !opt$.multilevel) {
    if (opt$conditional.x && opt$fixed.x == FALSE) {
      lav_msg_stop(gettext(
        "fixed.x = FALSE is not supported when conditional.x = TRUE."
      ))
    }
    if (opt$fixed.x && is.character(opt$start) && opt$start == "simple") {
      lav_msg_warn(gettextf(
        "start = %s implies fixed.x = FALSE", dQuote(opt$start)
      ))
      opt$fixed.x <- FALSE
    }
  } else if (opt$fixed.x == "default") {
    if (opt$conditional.x) {
      opt$fixed.x <- TRUE
    } else {
      opt$fixed.x <- FALSE
    }
  }

  # meanstructure again
  if (any(opt$missing == c("ml", "ml.x")) || opt$model.type == "growth") {
    opt$meanstructure <- TRUE
  }
  if (any(c("intercepts", "means") %in% opt$group.equal)) {
    opt$meanstructure <- TRUE
  }
  # if(opt$se == "robust.huber.white" ||
  #   opt$se == "robust.sem" ||
  #   opt$test == "satorra.bentler" ||
  #   opt$test == "mean.var.adjusted" ||
  #   opt$test == "scaled.shifted" ||
  #   opt$test == "yuan.bentler") {
  #    opt$meanstructure <- TRUE
  # }
  if (!is.logical(opt$meanstructure)) {
    lav_msg_fixme("meanstructure must be logical at this point!")
  }

  # zero cell frequencies
  if (is.character(opt$zero.add)) { # = "default"
    opt$zero.add <- c(0.5, 0.0)
    # FIXME: TODO: mimic EQS , LISREL (0.0, 0.0)
  } else {
    if (length(opt$zero.add) == 1L) {
      opt$zero.add <- c(opt$zero.add, opt$zero.add)
    }
  }

  if (is.character(opt$zero.keep.margins)) { # = "default"
    opt$zero.keep.margins <- FALSE
  }


  # parameterization
  if (opt$parameterization == "default") {
    # for now, default is always delta
    opt$parameterization <- "delta"
  }

  # std.lv vs auto.fix.first # new in 0.6-5 (used to be in sem/cfa/growth)
  if (opt$std.lv) {
    opt$auto.fix.first <- FALSE
  }

  # std.lv vs effect.coding # new in 0.6-4
  if (is.logical(opt$effect.coding)) {
    if (opt$effect.coding) {
      opt$effect.coding <- c("loadings", "intercepts")
    } else {
      opt$effect.coding <- ""
    }
  }

  # if we use effect coding for the factor loadings, we don't need/want
  # std.lv = TRUE
  if (any("loadings" == opt$effect.coding)) {
    if (opt$std.lv) {
      lav_msg_stop(gettextf(
        "std.lv is set to FALSE but effect.coding contains %s",
        dQuote("loadings")
      ))
    }
    # shut off auto.fix.first
    opt$auto.fix.first <- FALSE
  }

  # test again

  # unless test = "none", always add test = "standard" as the
  # first entry
  # NO: this breaks lavaan.survey pval.pFsum, which has the following check:
  # if (!lavInspect(lavaan.fit, "options")$test %in% c("satorra.bentler",
  #    "mean.var.adjusted", "Satterthwaite")) {
  #    lav_msg_stop(
  #    gettext("Please refit the model with Satorra-Bentler (MLM)"),
  #    gettext(" or Satterthwaite (MLMVS) adjustment."))
  # }
  # if(! (length(opt$test) == 1L && opt$test == "none") ) {
  #    opt$test <- c("standard", opt$test)
  #    opt$test <- unique(opt$test)
  # }

  # add standard.test to test (if not already there)
  if (opt$standard.test != "standard") {
    if (length(opt$test) == 1L && opt$test[1] == "standard") {
      opt$test <- unique(c(opt$test, opt$standard.test))
    } else {
      opt$test <- unique(c(opt$standard.test, opt$test))
    }
  }

  # add scaled.test to test (if not already there)
  if (opt$scaled.test != "standard") {
    if (length(opt$test) == 1L && opt$test[1] == "standard") {
      opt$test <- unique(c(opt$test, opt$scaled.test))
    } else {
      opt$test <- unique(c(opt$scaled.test, opt$test))
    }
  }

  # where does "standard" appear in the opt$test vector?
  if (opt$test[1] != "none") {
    standard_idx <- which(opt$test == "standard")[1]
    if (is.na(standard_idx)) {
      # "standard" is not in the opt$test vector at all,
      # so add it
      opt$test <- c("standard", opt$test)
    } else if (length(standard_idx) > 0L && standard_idx != 1L) {
      # make sure "standard" comes first
      opt$test <- c("standard", opt$test[-standard_idx])
    }
  }

  # final check
  wrong_idx <- which(!opt$test %in% c(
    "none", "standard", "satorra.bentler",
    "yuan.bentler", "yuan.bentler.mplus",
    "yuan.chan", # SAM-only (sam.method = "global"); ignored elsewhere
    "mean.var.adjusted", "scaled.shifted",
    "mean.var.adjusted.corrected", "scaled.shifted.corrected",
    "browne.residual.adf", "browne.residual.nt",
    "browne.residual.nt.model",
    "browne.residual.adf.model",
    "bollen.stine"
  ))
  fmg_idx <- which(vapply(opt$test, lav_test_fmg_is_fmg, logical(1L)))
  wrong_idx <- setdiff(wrong_idx, fmg_idx)
  if (length(wrong_idx) > 0L) {
    lav_msg_stop(gettextf(
      "invalid option(s) for test argument: %1$s. Possible options are: %2$s.",
      lav_msg_view(opt$test[wrong_idx]),
      lav_msg_view(c(
        "none", "standard", "browne.residual.adf",
        "browne.residual.nt", "browne.residual.adf.model",
        "browne.residual.nt.model", "satorra.bentler",
        "yuan.bentler", "yuan.bentler.mplus",
        "mean.var.adjusted", "scaled.shifted",
        "mean.var.adjusted.corrected", "scaled.shifted.corrected",
        "bollen.stine"
      ), log_sep = "or")
    ))
  }

  # bounds
  if (is.null(opt$bounds)) {
    if (length(opt$optim.bounds) > 0L) {
      opt$bounds <- "user"
    } else {
      opt$bounds <- "none" # for now
    }
  } else if (is.logical(opt$bounds)) {
    if (opt$bounds) {
      opt$bounds <- "wide" # default for most estimators
    } else {
      opt$bounds <- "none"
    }
  }

  # optim.bounds
  if (length(opt$optim.bounds) > 0L) {
    # opt$bounds should be "default", or "user" (or "none")
    if (any(opt$bounds == c("default", "none", "user"))) {
      opt$bounds <- "user"
    } else {
      lav_msg_stop(
        gettext("bounds and optim.bounds arguments can not be used together;
                remove the bounds= argument or set it to \"user\".")
      )
    }
  }

  # handle different 'profiles'
  if (opt$bounds == "none") {
    opt$optim.bounds <- list(
      lower = character(0L),
      upper = character(0L)
    )
  } else if (opt$bounds == "user") {
    if (length(opt$optim.bounds) == 0L) {
      lav_msg_stop(gettextf(
        "bounds= is %s but optim.bounds= argument is empty", dQuote("user")
      ))
    }
  } else if (opt$bounds == "default" || opt$bounds == "wide") {
    opt$optim.bounds <- list(
      lower = c("ov.var", "lv.var", "loadings", "covariances"),
      upper = c("ov.var", "lv.var", "loadings", "covariances"),
      lower.factor = c(1.05, 1.0, 1.1, 1.0),
      upper.factor = c(1.20, 1.3, 1.1, 1.0),
      min.reliability.marker = 0.1,
      min.var.lv.endo = 0.005
    )
  } else if (opt$bounds == "wide.zerovar") {
    opt$optim.bounds <- list(
      lower = c("ov.var", "lv.var", "loadings", "covariances"),
      upper = c("ov.var", "lv.var", "loadings", "covariances"),
      lower.factor = c(1.00, 1.0, 1.1, 1.0),
      upper.factor = c(1.20, 1.3, 1.1, 1.0),
      min.reliability.marker = 0.1,
      min.var.lv.endo = 0.005
    )
  } else if (opt$bounds == "standard") {
    opt$optim.bounds <- list(
      lower = c("ov.var", "lv.var", "loadings", "covariances"),
      upper = c("ov.var", "lv.var", "loadings", "covariances"),
      lower.factor = c(1.0, 1.0, 1.0, 0.999),
      upper.factor = c(1.0, 1.0, 1.0, 0.999),
      min.reliability.marker = 0.1,
      min.var.lv.endo = 0.005
    )
  } else if (opt$bounds == "pos.var") {
    opt$optim.bounds <- list(
      lower = c("ov.var", "lv.var"),
      lower.factor = c(1, 1),
      min.reliability.marker = 0.0,
      min.var.lv.exo = 0.0,
      min.var.lv.endo = 0.0
    )
  } else if (opt$bounds == "pos.ov.var") {
    opt$optim.bounds <- list(
      lower = c("ov.var"),
      lower.factor = 1
    )
  } else if (opt$bounds == "pos.lv.var") {
    opt$optim.bounds <- list(
      lower = c("lv.var"),
      lower.factor = 1,
      min.reliability.marker = 0.0,
      min.var.lv.exo = 0.0,
      min.var.lv.endo = 0.0
    )
  }

  # rotations.args
  if (!is.list(opt$rotation.args)) {
    lav_msg_stop(gettext("rotation.args should be a list."))
  }

  # transform names rotation.args to snake_case
  opt$rotation.args <- lav_snake_case(opt$rotation.args)

  # force orthogonal for some rotation algorithms
  if (any(opt$rotation == c(
    "varimax", "entropy", "mccammon",
    "tandem1", "tandem2"
  ))) {
    opt$rotation.args$orthogonal <- TRUE
  }

  # if target, check target matrix, and set order.lv.by to = "none"
  if (opt$rotation == "target.strict" || opt$rotation == "pst") {
    target <- opt$rotation.args$target
    if (is.null(target)) {
      lav_msg_stop(gettext("rotation target matrix is NULL"))
    }
    if (is.list(target)) {
      if (!all(sapply(target, is.matrix))) {
        lav_msg_stop(gettext("the target list contains
                              elements that are not a matrix"))
      }
    } else if (!is.matrix(target)) {
      lav_msg_stop(gettext("rotation target matrix is not a matrix"))
    }
    opt$rotation.args$order_lv_by <- "none"
  }

  if (opt$rotation == "pst") {
    target_mask <- opt$rotation.args$target_mask
    if (is.null(target_mask) || length(target_mask) == 0L) {
      # lav_msg_stop(gettext("rotation target_mask matrix is NULL"))
      if (is.matrix(target)) {
        tmp <- matrix(1L, nrow = nrow(target), ncol = ncol(target))
        tmp[target != 0] <- 0L # ignore these (non-zero) elements
        opt$rotation.args$target_mask <- target_mask <- tmp
      } else if (is.list(target)) {
        # lapply preserves the names (if any) of the target list
        out <- lapply(target, function(tt) {
          tmp <- matrix(1L, nrow = nrow(tt), ncol = ncol(tt))
          tmp[tt != 0] <- 0L # ignore these (non-zero) elements
          tmp
        })
        opt$rotation.args$target_mask <- target_mask <- out
      }
    }
    if (is.list(target_mask)) {
      if (!all(sapply(target_mask, is.matrix))) {
        lav_msg_stop(gettext("the target_mask list contains
                              elements that are not a matrix"))
      }
    } else if (!is.matrix(target_mask)) {
      lav_msg_stop(gettext("rotation target_mask matrix is not a matrix"))
    }
    if (is.list(target) && !is.list(target_mask)) {
      lav_msg_stop(gettext("target is a list, but target_mask is not a list"))
    }
    if (is.list(target_mask) && !is.list(target)) {
      lav_msg_stop(gettext("target_mask is a list, but target is not a list"))
    }
    if (is.list(target) && is.list(target_mask)) {
      if (length(target) != length(target_mask)) {
        lav_msg_stop(gettext("length(target) != length(target_mask)"))
      }
      # a *named* list is keyed by the efa block labels; a *nameless* list
      # is interpreted per group -- both lists must use the same convention
      target_named <- !is.null(names(target)) && any(nzchar(names(target)))
      mask_named <- !is.null(names(target_mask)) &&
        any(nzchar(names(target_mask)))
      if (target_named != mask_named) {
        lav_msg_stop(gettext(
          "target and target_mask must either both be named lists (with the efa block labels as names), or both be nameless lists (one element per group)."))
      }
      if (target_named && !setequal(names(target), names(target_mask))) {
        lav_msg_stop(gettext(
          "the names of the target list do not match the names of the target_mask list."))
      }
    }
  }

  # if NAs, force opt$rotation to be 'pst' and create target_mask
  if (opt$rotation == "target.strict") {
    # matrix
    warn_flag <- FALSE
    if (is.matrix(target) && anyNA(target)) {
      warn_flag <- TRUE
      opt$rotation <- "pst"
      target_mask <- matrix(1, nrow = nrow(target), ncol = ncol(target))
      target_mask[is.na(target)] <- 0
      opt$rotation.args$target_mask <- target_mask

      # list
    } else if (is.list(target) && any(sapply(target, anyNA))) {
      warn_flag <- TRUE
      opt$rotation <- "pst"
      # create a mask for *every* element (all-ones if no NA values);
      # lapply preserves the names (if any) of the target list
      opt$rotation.args$target_mask <- lapply(target, function(tt) {
        tmp <- matrix(1, nrow = nrow(tt), ncol = ncol(tt))
        tmp[is.na(tt)] <- 0
        tmp
      })
    }
    if (warn_flag) {
      lav_msg_warn(gettext(
        "switching to PST rotation as target matrix contains NA values"
      ))
    }
  }

  # set row_weights
  opt$rotation.args$row_weights <- tolower(opt$rotation.args$row_weights)
  if (opt$rotation.args$row_weights == "default") {
    # the default is "none", except for varimax and promax
    if (any(opt$rotation == c("varimax", "promax"))) {
      opt$rotation.args$row_weights <- "kaiser"
    } else {
      opt$rotation.args$row_weights <- "none"
    }
  }

  # override if bifactor
  if (any(opt$rotation == c(
    "bi-geomin", "bigeomin", "bi-quartimin",
    "biquartimin"
  ))) {
    opt$rotation.args$order_lv_by <- "none"
  }

  # no standard errors for promax (for now)...
  if (opt$rotation == "promax") {
    opt$se <- "none"
    opt$rotation.args$algorithm <- "promax"
    opt$rotation.args$rstarts <- 0L
  }

  # correlation
  if (opt$correlation) {
    # standardize
    opt$std.ov <- TRUE
    # if ML, switch to GLS
    if (opt$estimator == "ml") {
      # lav_msg_warn(gettext(
      #        "GLS should be used for correlation structures instead of ML."))
      opt$estimator <- "gls"
    }
    if (opt$missing == "ml") {
      lav_msg_stop(gettext(
        "correlation structures only work for complete data (for now)."
      ))
    }
    if (opt$.multilevel) {
      lav_msg_stop(gettext(
        "correlation structures only work for single-level data."
      ))
    }
    if (opt$conditional.x) {
      lav_msg_stop(gettext(
        "correlation structures only work for conditional.x = FALSE (for now)."
      ))
    }
    if (opt$representation == "RAM") {
      lav_msg_stop(gettext(
        "correlation structures only work for representation = \"LISREL\"."
      ))
    }
  }

  # sample.cov.robust
  # new in 0.6-17
  # sample.cov.robust cannot be used if:
  # - data is missing (for now),
  # - sampling weights are used
  # - estimator is (D)WLS
  # - multilevel
  # - conditional.x
  if (opt$sample.cov.robust) {
    if (opt$missing != "listwise") {
      lav_msg_stop(gettext(
        "sample.cov.robust = TRUE does not work (yet) if data is missing."
      ))
    }
    if (opt$.categorical) {
      lav_msg_stop(gettext(
        "sample.cov.robust = TRUE does not work (yet) if data is categorical"
      ))
    }
    if (opt$.clustered || opt$.multilevel) {
      lav_msg_stop(gettext(
        "sample.cov.robust = TRUE does not work (yet) if data is clustered"
      ))
    }
    if (opt$conditional.x) {
      lav_msg_stop(gettext(
        "sample.cov.robust = TRUE does not work (yet) if conditional.x = TRUE"
      ))
    }
    if (all(lav_options_estimatorgroup(opt$estimator) != c("ML", "GLS"))) {
      lav_msg_stop(gettext(
        "sample.cov.robust = TRUE does not work (yet)
        if estimator is not GLS or ML"
      ))
    }
  }

  opt$estimator <- lav_options_estimatorgroup(opt$estimator)

  # group.w.free
  # if(opt$group.w.free && opt$.categorical) {
  #    lav_msg_stop(gettext(
  #    "group.w.free = TRUE is not supported (yet) in the categorical setting."
  #    ))
  # }

  # in order not to break semTools and blavaan, we restore categorical:
  opt$categorical <- opt$.categorical

  if (lav_debug()) {
    cat("lavaan DEBUG: lavaanOptions OUT\n")
    str(opt)
  }

  opt
}

lav_step11_estoptim <- function(lavdata = NULL,
                                       lavmodel = NULL,
                                       lavcache = NULL,
                                       lavsamplestats = NULL,
                                       lavh1 = NULL,
                                       lavoptions = NULL,
                                       lavpartable = NULL) {
  # # # # # # # # # # # # # #
  # #  11. est + lavoptim # #
  # # # # # # # # # # # # # #

  # if lavoptions$do.fit and lavoptions$estimator not "none" and
  #   lavmodel$nx.free > 0
  #   select case lavoptions$optim.method
  #   case "noniter"
  #     try x <- lav_optim_noniter(...)
  #   case "em"
  #     if nlevels < 2L *** error ***
  #     try x <- lav_mvn_cl_em_h0(...)
  #   case "gn"
  #     try x <- lav_optim_gn(...)
  #   case else
  #     set 1 in lavoptions$optim.attempts if it wasn't specified
  #     try x <- lav_model_est(...)
  #     if not successful and optim.attempts > 1L
  #       try x <- lav_optim_estimate(...) with
  #         options$optim.parscale = "standardized"
  #       if not successful and optim.attempts > 2L
  #         try x <- lav_optim_estimate(...) with start = "simple"
  #         if not successful and optim.attempts > 3L
  #           try x <- lav_optim_estimate(...) with
  #             options$optim.parscale = "standardized" and start = "simple"
  #   end select
  #   if x not successfully computed
  #     ** warning **
  #     set starting values and appropriate attributes in x
  #   in case of non-linear constraints: store final con.jac and
  #     con.lambda in lavmodel
  #   store parameters in lavmodel
  #   store parameters in partable$est
  # else
  #   initialize x and attributes (iterations, converged, warn.txt,
  #     control, dx) of x
  #   try fx <- lav_model_objective
  #   if not successful
  #     fx = NA_real_
  #     attribute fx.group of fx = NA_real_
  #   store fx in attribute "fx" of x
  #   set lavpartable$est to starting values
  # if lavoptions$optim.force.converged set attribute converged of x to TRUE
  # store optimization info in lavoptim

  x <- NULL
  if (lavoptions$do.fit && lavoptions$estimator != "none" &&
    lavmodel@nx.free > 0L) {
    if (lav_verbose()) {
      cat("lavoptim           ... start:\n")
    }

    # the EM optimizer only supports two-level models with a single
    # group, an ML-family estimator, no conditional.x, and no
    # (nonlinear or inequality) constraints; if optim.method = "em" was
    # merely the default (missing = "ml"; see lav_options_set), quietly
    # fall back to "nlminb"; if the user asked for it explicitly, warn
    if (lavoptions$optim.method == "em" &&
        (lavdata@nlevels == 1L ||
         lavdata@ngroups > 1L ||
         lavmodel@conditional.x ||
         lavmodel@estimator != "ML" ||
         length(lavmodel@ceq.nonlinear.idx) > 0L ||
         length(lavmodel@cin.linear.idx) > 0L ||
         length(lavmodel@cin.nonlinear.idx) > 0L)) {
      if (!isTRUE(lavoptions$.optim.em.fallback)) {
        lav_msg_warn(gettext(
          "optim.method = \"em\" is not available for this model;
           using optim.method = \"nlminb\" instead."))
      }
      lavoptions$optim.method <- "nlminb"
    }

    # random slopes? check if the EM version can handle this model
    rv_flag <- length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L
    if (lavoptions$optim.method == "em" && rv_flag) {
      em_reason <- lav_mvn_cl_rs_em_ok(lavcache[[1]]$rs)
      if (length(em_reason) > 0L) {
        if (!isTRUE(lavoptions$.optim.em.fallback)) {
          lav_msg_warn(gettextf(
            "optim.method = \"em\" is not available for this random-slope
             model (%s); using optim.method = \"nlminb\" instead.",
            paste(em_reason, collapse = "; ")))
        }
        lavoptions$optim.method <- "nlminb"
      }
    }

    # non-iterative methods (fabin, miiv, ...)
    if (lavoptions$optim.method == "noniter") {
      x <- try(
        lav_optim_noniter(
          lavmodel = lavmodel,
          lavsamplestats = lavsamplestats,
          lavh1 = lavh1,
          lavdata = lavdata,
          lavpartable = lavpartable,
          lavoptions = lavoptions
        ),
        silent = FALSE
      )
      # EM for multilevel models
    } else if (lavoptions$optim.method == "em") {
      # multilevel only for now
      stopifnot(lavdata@nlevels > 1L)

      handoff_flag <- !isFALSE(lavoptions$em.args$nlminb_handoff)
      em_fx_tol <- lavoptions$em.args$fx_tol
      if (!is.numeric(em_fx_tol)) {
        em_fx_tol <- 1e-08
      }
      em_dx_tol <- lavoptions$em.args$dx_tol
      if (!is.numeric(em_dx_tol)) {
        em_dx_tol <- 1e-04
      }

      # run the EM iterations, starting from the current parameter
      # values of lavmodel_start
      run_em <- function(lavmodel_start, fx_tol) {
        if (rv_flag) {
          # random slopes: dedicated EM algorithm
          try(
            lav_mvn_cl_rs_em_h0(
              lavsamplestats = lavsamplestats,
              lavdata = lavdata,
              lavpartable = lavpartable,
              lavmodel = lavmodel_start,
              lavoptions = lavoptions,
              lavcache = lavcache,
              fx_tol = fx_tol,
              dx_tol = em_dx_tol,
              max_iter = lavoptions$em.args$max_iter,
              mstep_verbose = isTRUE(lavoptions$em.args$verbose),
              acceleration = lavoptions$em.args$acceleration
            ),
            silent = TRUE
          )
        } else {
          try(
            lav_mvn_cl_em_h0(
              lavsamplestats = lavsamplestats,
              lavdata = lavdata,
              lavimplied = NULL,
              lavpartable = lavpartable,
              lavmodel = lavmodel_start,
              lavoptions = lavoptions,
              fx_tol = fx_tol,
              dx_tol = em_dx_tol,
              max_iter = lavoptions$em.args$max_iter,
              mstep_verbose = isTRUE(lavoptions$em.args$verbose),
              acceleration = lavoptions$em.args$acceleration,
              fused = lavoptions$em.args$fused
            ),
            silent = TRUE
          )
        }
      }

      # EM-stall -> nlminb hand-off (new in 0.7-1): the EM iterations
      # typically stop because the loglikelihood stalls (fx_tol) before
      # the gradient becomes small (dx_tol); in that case, refine the
      # solution with nlminb, warm-started at the EM values; if nlminb
      # gets lost in parameter space (non-finite values, a worse
      # solution) or does not converge, we keep the EM solution;
      # returns list(x = , accepted = )
      run_handoff <- function(x) {
        accept <- FALSE
        lavmodel_em <- lav_model_set_parameters(lavmodel,
          x = as.numeric(x))
        dx_em <- try(lav_model_grad(
          lavmodel = lavmodel_em,
          lavsamplestats = lavsamplestats,
          lavdata = lavdata,
          lavcache = lavcache
        ), silent = TRUE)
        # random slopes: if the M-step model used the variance floor,
        # the EM fixed point is not the optimum of the exact model:
        # always run the hand-off
        force_handoff <- isTRUE(attr(x, "rs.floored"))
        if (!inherits(dx_em, "try-error") && all(is.finite(dx_em)) &&
            (force_handoff || max(abs(dx_em)) >= em_dx_tol)) {
          if (lav_verbose()) {
            cat("EM stalled (max|grad| = ", max(abs(dx_em)),
                "); handing off to nlminb\n")
          }
          # objective at the EM solution (for the accept/reject check)
          fx_em <- try(lav_model_objective(
            lavmodel = lavmodel_em,
            lavsamplestats = lavsamplestats,
            lavdata = lavdata,
            lavcache = lavcache
          ), silent = TRUE)
          lavoptions2 <- lavoptions
          lavoptions2$optim.method <- "nlminb"
          # do NOT pin saturated blocks at their h1 values: the EM
          # values are (at least as) good, and pinning would overwrite
          # the warm start
          lavoptions2$optim.fix.saturated <- FALSE
          x2 <- try(
            lav_model_est(
              lavmodel = lavmodel_em, # warm start at the EM solution
              lavpartable = lavpartable,
              lavh1 = lavh1,
              lavsamplestats = lavsamplestats,
              lavdata = lavdata,
              lavoptions = lavoptions2,
              lavcache = lavcache
            ),
            silent = TRUE
          )
          accept <- (!inherits(x2, "try-error") &&
            isTRUE(attr(x2, "converged")) &&
            all(is.finite(x2)) &&
            is.finite(attr(x2, "fx")[1]) &&
            (inherits(fx_em, "try-error") || !is.finite(fx_em[1]) ||
             attr(x2, "fx")[1] <= fx_em[1] + 1e-10))
          if (accept) {
            x <- x2
          }
          if (lav_verbose()) {
            if (accept) {
              cat("nlminb hand-off accepted (fx = ",
                  attr(x2, "fx")[1], ")\n")
            } else {
              cat("nlminb hand-off rejected; keeping the EM solution\n")
            }
          }
        }
        list(x = x, accepted = accept)
      }

      # early hand-off (new in 0.7-1): when the nlminb hand-off is
      # active, the EM phase does not need to iterate until fx_tol --
      # it only needs to get close enough for a reliable warm start;
      # run the EM with a loosened tolerance, and let nlminb finish
      # (a safety net below reverts to the strict tolerance if nlminb
      # fails)
      early_flag <- handoff_flag &&
        !isFALSE(lavoptions$em.args$early_handoff) &&
        em_fx_tol < 1e-03
      x <- run_em(lavmodel, if (early_flag) 1e-03 else em_fx_tol)

      if (handoff_flag && !inherits(x, "try-error")) {
        out_h <- run_handoff(x)
        x <- out_h$x
        if (early_flag && !out_h$accepted) {
          # safety net: nlminb could not improve on the loose EM
          # solution; finish the EM iterations with the strict
          # tolerance (warm start at the current values), and try the
          # hand-off once more (this reproduces the non-early
          # behavior)
          if (lav_verbose()) {
            cat("early hand-off failed; resuming EM with fx_tol = ",
                em_fx_tol, "\n")
          }
          x_strict <- run_em(
            lav_model_set_parameters(lavmodel, x = as.numeric(x)),
            em_fx_tol
          )
          if (!inherits(x_strict, "try-error")) {
            x <- run_handoff(x_strict)$x
          }
        }
      }

      # Gauss-Newton (Fisher scoring with Levenberg-Marquardt damping);
      # the default for DLS
    } else if (lavoptions$optim.method == "gn") {
      x <- try(
        lav_optim_gn(
          lavmodel = lavmodel,
          lavsamplestats = lavsamplestats,
          lavdata = lavdata,
          lavpartable = lavpartable,
          lavoptions = lavoptions
        ),
        silent = TRUE
      )

      # reduced-bias M-estimation (RBM); optim.method is nlminb, so this is
      # keyed on the estimator.args marker instead
    } else if (!is.null(lavoptions$estimator.args$rbm.method)) {
      x <- try(
        lav_model_est_rbm(
          lavmodel = lavmodel,
          lavpartable = lavpartable,
          lavsamplestats = lavsamplestats,
          lavdata = lavdata,
          lavoptions = lavoptions,
          lavcache = lavcache
        ),
        silent = TRUE
      )

      # Quasi-Newton
    } else {
      # for backwards compatibility (<0.6)
      if (is.null(lavoptions$optim.attempts)) {
        lavoptions$optim.attempts <- 1L
      }

      # try 1
      if (lav_verbose()) {
        cat("attempt 1 -- default options\n")
      }
      x <- try(
        lav_model_est(
          lavmodel = lavmodel,
          lavpartable = lavpartable,
          lavh1 = lavh1,
          lavsamplestats = lavsamplestats,
          lavdata = lavdata,
          lavoptions = lavoptions,
          lavcache = lavcache
        ),
        silent = TRUE
      )

      # try 2: optim.parscale = "standardize" (new in 0.6-7)
      if (lavoptions$optim.attempts > 1L &&
        lavoptions$rstarts == 0L &&
        (inherits(x, "try-error") || !attr(x, "converged"))) {
        lavoptions2 <- lavoptions
        lavoptions2$optim.parscale <- "standardized"
        if (lav_verbose()) {
          str(x)
          cat("attempt 2 -- optim.parscale = \"standardized\"\n")
        }
        x <- try(
          lav_model_est(
            lavmodel = lavmodel,
            lavpartable = lavpartable,
            lavh1 = lavh1,
            lavsamplestats = lavsamplestats,
            lavdata = lavdata,
            lavoptions = lavoptions2,
            lavcache = lavcache
          ),
          silent = TRUE
        )
      }

      # try 3: start = "simple"
      if (lavoptions$optim.attempts > 2L &&
        lavoptions$rstarts == 0L &&
        (inherits(x, "try-error") || !attr(x, "converged"))) {
        if (lav_verbose()) {
          str(x)
          cat("attempt 3 -- start = \"simple\"\n")
        }
        x <- try(
          lav_model_est(
            lavmodel = lavmodel,
            lavpartable = lavpartable,
            lavh1 = lavh1,
            lavsamplestats = lavsamplestats,
            lavdata = lavdata,
            lavoptions = lavoptions,
            start = "simple",
            lavcache = lavcache
          ),
          silent = TRUE
        )
      }

      # try 4: start = "simple" + optim.parscale = "standardize"
      if (lavoptions$optim.attempts > 3L &&
        lavoptions$rstarts == 0L &&
        (inherits(x, "try-error") || !attr(x, "converged"))) {
        lavoptions2 <- lavoptions
        lavoptions2$optim.parscale <- "standardized"
        if (lav_verbose()) {
          str(x)
          cat(
            "attempt 4 -- optim.parscale = \"standardized\" + ",
            "start = \"simple\"\n"
          )
        }
        x <- try(
          lav_model_est(
            lavmodel = lavmodel,
            lavpartable = lavpartable,
            lavh1 = lavh1,
            lavsamplestats = lavsamplestats,
            lavdata = lavdata,
            lavoptions = lavoptions2,
            start = "simple",
            lavcache = lavcache
          ),
          silent = TRUE
        )
      }


      # random starts? -- new in 0.6-18
      # run this even if we already have a converged solution
      # perhaps we find a better solution?
      if (lavoptions$rstarts > 0L) {
        x_rstarts <- vector("list", length = lavoptions$rstarts)
        if (lav_verbose()) {
          str(x)
          cat("trying again with random starts (", lavoptions$rstarts,
            " in total):\n",
            sep = ""
          )
        }
        for (i in seq_len(lavoptions$rstarts)) {
          if (lav_verbose()) {
            cat("-- random start run: ", i, "\n")
          }
          x_rstarts[[i]] <-
            try(
              lav_model_est(
                lavmodel = lavmodel,
                lavpartable = lavpartable,
                lavh1 = lavh1,
                lavsamplestats = lavsamplestats,
                lavdata = lavdata,
                lavoptions = lavoptions,
                start = "random",
                lavcache = lavcache
              ),
              silent = TRUE
            )
        }

        # pick best solution (if any)
        x_converged <- vector("list", length = 0L)
        fx_rstarts <- numeric(0L)
        ok_flag <- sapply(x_rstarts, function(x) {
            if (inherits(x, "try-error")) {
              FALSE
            } else {
              attr(x, "converged")
            }
          })
        if (sum(ok_flag) > 0L) {
          x_converged <- x_rstarts[ok_flag]
        }
        if (length(x_converged) > 0L) {
          fx_rstarts <- sapply(x_converged, "attr", "fx")
          x_best <- x_converged[[which.min(fx_rstarts)]]
          fx_best <- attr(x_best, "fx")[1]

          # if we did not find a converged solution, use x.best
          if (inherits(x, "try-error") || !attr(x, "converged")) {
            x <- x_best


            # if we already had a converged solution, only replace
            # if fx.best is better than attr(x, "fx")[1]
          } else {
            if (fx_best < attr(x, "fx")[1]) {
              x <- x_best
            }
          }
        }

        attr(x, "x.rstarts") <- x_rstarts
      } # random starts
    }

    # optimization failed with error
    if (inherits(x, "try-error")) {
      warn_txt <- gettext("Model estimation FAILED! Returning starting values.")
      x <- lav_model_get_parameters(
        lavmodel = lavmodel,
        type = "free"
      ) # starting values
      attr(x, "iterations") <- 0L
      attr(x, "converged") <- FALSE
      attr(x, "warn.txt") <- warn_txt
      attr(x, "control") <- lavoptions$control
      attr(x, "dx") <- numeric(0L)
      fx <- as.numeric(NA)
      attr(fx, "fx.group") <- as.numeric(NA)
      attr(x, "fx") <- fx
    }

    # if a warning was produced, say it here
    warn_txt <- attr(x, "warn.txt")
    if (!is.null(warn_txt) && nchar(warn_txt) > 0L) {
      lav_msg_warn(gettext(warn_txt))
    }

    # in case of non-linear constraints: store final con.jac and con.lambda
    # in lavmodel
    if (!is.null(attr(x, "con.jac"))) {
      lavmodel@con.jac <- attr(x, "con.jac")
    }
    if (!is.null(attr(x, "con.lambda"))) {
      lavmodel@con.lambda <- attr(x, "con.lambda")
    }

    # store parameters in lavmodel
    lavmodel <- lav_model_set_parameters(lavmodel, x = as.numeric(x))

    # store parameters in @ParTable$est
    lavpartable$est <- lav_model_get_parameters(
      lavmodel = lavmodel,
      type = "user", extra = TRUE
    )

    if (lav_verbose()) {
      cat("lavoptim    ... done.\n")
    }
  } else {
    x <- numeric(0L)
    attr(x, "iterations") <- 0L
    attr(x, "converged") <- FALSE
    attr(x, "warn.txt") <- ""
    attr(x, "control") <- lavoptions$control
    attr(x, "dx") <- numeric(0L)
    fx <- try(lav_model_objective(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavcache = lavcache
    ), silent = TRUE)
    if (!inherits(fx, "try-error")) {
      attr(x, "fx") <- fx
    } else {
      fx <- as.numeric(NA)
      attr(fx, "fx.group") <- as.numeric(NA)
      attr(x, "fx") <- fx
    }

    lavpartable$est <- lavpartable$start
  }

  # should we fake/force convergence? (eg. to enforce the
  # computation of a test statistic)
  if (lavoptions$optim.force.converged) {
    attr(x, "converged") <- TRUE
  }

  # store optimization info in lavoptim
  lavoptim <- list()
  x2 <- x
  attributes(x2) <- NULL
  lavoptim$x <- x2
  lavoptim$dx <- attr(x, "dx")
  lavoptim$npar <- length(x)
  lavoptim$iterations <- attr(x, "iterations")
  lavoptim$converged <- attr(x, "converged")
  lavoptim$warn.txt <- attr(x, "warn.txt")
  lavoptim$parscale <- attr(x, "parscale")
  lavoptim$partrace <- attr(x, "partrace")
  fx_copy <- fx <- attr(x, "fx")
  attributes(fx) <- NULL
  lavoptim$fx <- fx
  lavoptim$fx.group <- attr(fx_copy, "fx.group")
  if (!is.null(attr(fx_copy, "logl.group"))) {
    lavoptim$logl.group <- attr(fx_copy, "logl.group")
    lavoptim$logl <- sum(lavoptim$logl.group)
  } else {
    lavoptim$logl.group <- as.numeric(NA)
    lavoptim$logl <- as.numeric(NA)
  }
  lavoptim$control <- attr(x, "control")
  if (!is.null(attr(x, "x.rstarts"))) {
    lavoptim$x.rstarts <- attr(x, "x.rstarts")
  }

  lavpartable <- lav_pt_set_cache(lavpartable, force = TRUE)
  list(
    lavoptim = lavoptim, lavmodel = lavmodel, lavpartable = lavpartable,
    x = x,
    # lavoptions may have been updated (e.g., the em -> nlminb
    # fallback): return it, so that the fitted object records the
    # optimizer that was actually used
    lavoptions = lavoptions
  )
}

# 'robust' mean and (co)variance matrix using Huber weights
#
# see Yuan & Hayashi (2010). Fitting Data to Model: SEM Diagnosis using two
#     scatter plots. Psychological Methods, 15(4), 335-351
#
# this function is based on the 'robmusig' function from K.H. Yuan's website:
# https://www.nd.edu/~kyuan/SEMdiagnosis
# see file CFA.r lines 46--96
#
lav_cov_huber <- function(y = NULL, prob = 0.95, max_it = 200L, tol = 1e-07) {
  y <- as.matrix(y)
  names_1 <- colnames(y)
  y <- unname(y)
  n <- nrow(y)
  p <- ncol(y)

  # tuning parameters for Huber's weight
  chip <- qchisq(prob, p)
  ck <- sqrt(chip)
  cbeta <- (p * pchisq(chip, p + 2L) + chip * (1 - prob)) / p

  # initial values
  this_mu <- colMeans(y, na.rm = TRUE)
  this_sigma <- cov(y, use = "pairwise.complete.obs")

  for (i in seq_len(max_it)) {
    # store old
    old_mu <- this_mu
    old_sigma <- this_sigma

    # squared Mahalanobis distance
    inv_sigma <- solve(this_sigma)
    y_c <- t(t(y) - this_mu)
    mdist2 <- rowSums((y_c %*% inv_sigma) * y_c)
    mdist <- sqrt(mdist2)

    # Huber weights
    wt <- ifelse(mdist <= ck, 1, ck / mdist)

    # weighted mean
    this_mu <- apply(y, 2L, weighted.mean, w = wt, na.rm = TRUE)

    # weighted cov
    y_c <- t(t(y) - this_mu)
    this_sigma <- crossprod(y_c * wt) / (n * cbeta)
    # question: why N, and not sum(wt)?

    # check progress
    diff_mu <- abs(this_mu - old_mu)
    diff_sigma <- abs(this_sigma - old_sigma)
    crit <- max(c(max(diff_mu), max(diff_sigma)))
    if (crit < tol) {
      break
    }
    if (i == max_it) {
      lav_msg_warn(gettext(
        "maximum number of iterations has been reached, without convergence."))
    }
  }

  names(this_mu) <- names_1
  colnames(this_sigma) <- rownames(this_sigma) <- names_1

  res <- list(Mu = this_mu, Sigma = this_sigma, niter = i, wt = wt)

  res
}

.onAttach <- function(libname, pkgname) {
  version <- read.dcf(
    file = system.file("DESCRIPTION", package = pkgname),
    fields = "Version"
  )[1]
  packageStartupMessage(
    "This is ", paste(pkgname, version), "\n",
    pkgname, " is FREE software! Please report any bugs."
  )
}

lav_samp_wls_obs <- function(mean_g, cov_g, var_g,
                                    th_g, th_idx_g,
                                    res_int_g, res_cov_g, res_var_g, res_th_g,
                                    res_slopes_g,
                                    group_w_g,
                                    categorical = FALSE,
                                    conditional_x = FALSE,
                                    meanstructure = FALSE,
                                    correlation = FALSE,
                                    num_idx = integer(0L),
                                    slopestructure = FALSE,
                                    group_w_free = FALSE) {
  # WLS.obs
  if (categorical) {
    # order of elements is important here:
    # 1. thresholds + (negative) means (interleaved)
    # 2. slopes (if any)
    # 3. variances (if any)
    # 4. covariance matrix (no diagonal!)

    # NOTE: prior to 0.5-17, we had this:
    # TH[ov.types == "numeric"] <- -1*TH[ov.types == "numeric"]
    # which is WRONG if we have more than one threshold per variable
    # (thanks to Sacha Epskamp for spotting this!)
    if (conditional_x) {
      th <- res_th_g
      th[th_idx_g == 0] <- -1 * th[th_idx_g == 0]

      nvar <- length(res_var_g)
      num_idx <- which(!seq_len(nvar) %in% th_idx_g)

      wls_obs <- c(
        th,
        lav_mat_vec(res_slopes_g),
        res_var_g[num_idx],
        lav_mat_vech(res_cov_g, diagonal = FALSE)
      )
    } else {
      th <- th_g
      th[th_idx_g == 0] <- -1 * th[th_idx_g == 0]

      nvar <- length(var_g)
      num_idx <- which(!seq_len(nvar) %in% th_idx_g)

      wls_obs <- c(
        th,
        var_g[num_idx],
        lav_mat_vech(cov_g, diagonal = FALSE)
      )
    }
  } else {
    # CONTINUOUS:
    diag_1 <- TRUE
    if (correlation) {
      diag_1 <- FALSE
    }
    if (conditional_x) {
      if (meanstructure) {
        if (slopestructure) {
          # order = vec(Beta), where first row are intercepts
          # cbind(res.int, res.slopes) is t(Beta)
          # so we need vecr
          wls_obs <- c(
            lav_mat_vecr(cbind(
              res_int_g,
              res_slopes_g
            )),
            lav_mat_vech(res_cov_g, diagonal = diag_1)
          )
        } else {
          wls_obs <- c(
            res_int_g,
            lav_mat_vech(res_cov_g, diagonal = diag_1)
          )
        }
      } else {
        if (slopestructure) {
          wls_obs <- c(
            lav_mat_vecr(res_slopes_g),
            lav_mat_vech(res_cov_g, diagonal = diag_1)
          )
        } else {
          wls_obs <- lav_mat_vech(res_cov_g, diagonal = diag_1)
        }
      }
    } else {
      if (correlation && length(num_idx) > 0L) {
        # partial correlation structure: keep the variances of the
        # non-correlation variables (num_idx), drop only the correlation
        # variables' diagonal (mirrors the categorical layout above)
        v_obs <- c(var_g[num_idx], lav_mat_vech(cov_g, diagonal = FALSE))
      } else {
        v_obs <- lav_mat_vech(cov_g, diagonal = diag_1)
      }
      if (meanstructure) {
        wls_obs <- c(mean_g, v_obs)
      } else {
        wls_obs <- v_obs
      }
    }
  }

  # group.w.free?
  if (group_w_free) {
    wls_obs <- c(group_w_g, wls_obs)
  }

  wls_obs
}

# a partial implementation of the Bentler (1982) non-iterative method for CFA
#
# Bentler, P. M. (1982). Confirmatory factor-analysis via noniterative
# estimation - a fast, inexpensive method. Journal of Marketing Research,
# 19(4), 417-424. https://doi.org/10.1177/002224378201900403
#
#
# YR 03 Feb 2023: - first version in lavaan: simple setting only,
#                   no constraints, no 'fixed' (but nonzero) values,
#                   no correlated residuals (ie diagonal-theta only!)
# YR 23 Apr 2023: - quadprog is not needed if we have no (in)equality
#                   constraints

lav_cfa_bentler1982 <- function(s,
                                marker_idx = NULL,
                                lambda_nonzero_idx = NULL,
                                gls = FALSE,
                                bounds = TRUE,
                                min_reliability_marker = 0.1,
                                quadprog = FALSE,
                                nobs = 20L) { # for cutoff
  # dimensions
  nvar <- ncol(s)
  nfac <- length(marker_idx)

  # lambda structure
  m_b <- matrix(0, nvar, nfac)
  lambda_marker_idx <- (seq_len(nfac) - 1L) * nvar + marker_idx
  m_b[lambda_marker_idx] <- 1L
  m_b[lambda_nonzero_idx] <- 1L

  # partition sample covariance matrix: marker vs non-marker
  s_xx <- s[marker_idx, marker_idx, drop = FALSE]
  s_yx <- s[-marker_idx, marker_idx, drop = FALSE]
  s_yy <- s[-marker_idx, -marker_idx, drop = FALSE]
  p <- nvar - nfac
  b_y <- m_b[-marker_idx, , drop = FALSE]

  # check for p = 0?

  # phase 1: initial estimate for Sigma.yx

  # phase 2: using GLS/ULS to obtain PSI and Theta
  if (gls) {
    m_w <- try(solve(s_yy), silent = TRUE)
    if (inherits(m_w, "try-error")) {
      lav_msg_warn(gettext("could not invert S.yy; switching to ULS"))
      m_w <- diag(p)
    }
    ws_yx <- m_w %*% s_yx
    xy_sws_yx <- crossprod(s_yx, ws_yx)
    g <- ws_yx %*% solve(xy_sws_yx) %*% t(ws_yx)
  } else {
    ip <- diag(p)
    xy_ss_yx <- crossprod(s_yx)
    g <- s_yx %*% solve(xy_ss_yx) %*% t(s_yx)
  }

  # only needed if theta.y is not diagonal:
  # q <- 6 # all free
  # # dimension P: q x p*p where q is the number of free elements theta.y
  # theta.fy <- function(x) {
  #     theta.y <- matrix(0, p, p)
  #     # insert 'free' parameters only
  #     diag(theta.y) <- x
  #     lav_mat_vec(theta.y)
  # }
  # P <- t(numDeriv::jacobian(func = theta.fy, x = rep(1, q)))

  # tmp1 <- P %*% ((W %x% W) - (G %x% G)) %*% t(P)
  # NOTE:
  # if only the 'diagonal' element of Theta are free (as usual), then we
  # can write tmp1 as
  if (gls) {
    tmp1 <- m_w * m_w - g * g
  } else {
    tmp1 <- ip - g * g
  }

  # only needed if fixed values
  # Theta.F <- matrix(0, p, p) # all free
  # tmp2 <- W %*% (S.yy - Theta.F) %*% W - G %*% (S.yy - Theta.F) %*% G
  if (gls) {
    tmp2 <- m_w %*% s_yy %*% m_w - g %*% s_yy %*% g
  } else {
    tmp2 <- s_yy - g %*% s_yy %*% g
  }

  # Theta.f    <- as.numeric(solve(tmp1) %*% P %*% lav_mat_vec(tmp2))
  # Note:
  # if only the 'diagonal' element of Theta are free (as usual), then we
  # can write Theta.f as
  theta_f <- solve(tmp1, diag(tmp2))
  theta_f_nobounds <- theta_f # store unbounded Theta.f values

  # ALWAYS apply standard bounds to proceed
  too_small_idx <- which(theta_f < 0)
  if (length(too_small_idx) > 0L) {
    theta_f[too_small_idx] <- 0
  }
  too_large_idx <- which(theta_f > diag(s_yy))
  if (length(too_large_idx) > 0L) {
    theta_f[too_large_idx] <- diag(s_yy)[too_large_idx] * 1
  }

  # create diagonal matrix with Theta.f elements on diagonal
  theta_yhat <- diag(theta_f, p)

  # force (S.yy - Theta.yhat) to be positive definite
  lambda <- try(lav_mat_sym_diff_smallest_root(s_yy, theta_yhat),
    silent = TRUE
  )
  if (inherits(lambda, "try-error")) {
    lav_msg_warn(gettext("failed to compute lambda"))
    s_min_theta <- s_yy - theta_yhat # and hope for the best
  } else {
    cutoff <- 1 + 1 / (nobs - 1)
    if (lambda < cutoff) {
      lambda_star <- lambda - 1 / (nobs - 1)
      s_min_theta <- s_yy - lambda_star * theta_yhat
    } else {
      s_min_theta <- s_yy - theta_yhat
    }
  }

  # estimate Phi
  if (gls) {
    tmp1 <- xy_sws_yx
    tmp2 <- t(ws_yx) %*% s_min_theta %*% ws_yx
  } else {
    tmp1 <- xy_ss_yx
    tmp2 <- t(s_yx) %*% s_min_theta %*% s_yx
  }
  mm_psi <- tmp1 %*% solve(tmp2, tmp1)
  psi_nobounds <- mm_psi

  # ALWAYS apply bounds to proceed
  lower_bounds_psi <- diag(s_xx) - (1 - min_reliability_marker) * diag(s_xx)
  toolow_idx <- which(diag(mm_psi) < lower_bounds_psi)
  if (length(toolow_idx) > 0L) {
    diag(mm_psi)[toolow_idx] <- lower_bounds_psi[toolow_idx]
  }
  too_large_idx <- which(diag(mm_psi) > diag(s_xx))
  if (length(too_large_idx) > 0L) {
    diag(mm_psi)[too_large_idx] <- diag(s_xx)[too_large_idx] * 1
  }

  # in addition, force PSI to be PD
  mm_psi <- lav_mat_sym_force_pd(mm_psi, tol = 1e-04)

  # residual variances markers
  theta_x <- diag(s_xx - mm_psi)

  # create theta vector
  theta_nobounds <- numeric(nvar)
  theta_nobounds[marker_idx] <- theta_x
  theta_nobounds[-marker_idx] <- theta_f_nobounds

  # compute LAMBDA for non-marker items

  if (quadprog) {
    # only really needed if we need to impose (in)equality constraints
    # (TODO)
    dmat <- lav_mat_bdiag(rep(list(mm_psi), p))
    dvec <- as.vector(t(s_yx))
    eq_idx <- which(t(b_y) != 1) # these must be zero (row-wise!)
    rmat <- diag(nrow(dmat))[eq_idx, , drop = FALSE]
    bvec <- rep(0, length(eq_idx)) # optional, 0=default
    out <- try(quadprog::solve.QP(
      Dmat = dmat, dvec = dvec, Amat = t(rmat),
      meq = length(eq_idx), bvec = bvec
    ), silent = TRUE)
    if (inherits(out, "try-error")) {
      lav_msg_warn(gettext("solve.QP failed to find a solution"))
      lambda_1 <- matrix(0, nvar, nfac)
      lambda_1[marker_idx, ] <- diag(nfac)
      lambda_1[lambda_nonzero_idx] <- as.numeric(NA)
      theta <- numeric(nvar)
      theta[marker_idx] <- theta_x
      theta[-marker_idx] <- theta_f
      return(list(
        lambda = lambda_1, theta = theta_nobounds,
        psi = psi_nobounds
      ))
    } else {
      lambda_y <- matrix(out$solution,
        nrow = p, ncol = nfac,
        byrow = TRUE
      )
      # zap almost zero elements
      lambda_y[abs(lambda_y) < sqrt(.Machine$double.eps)] <- 0
    }
  } else {
    # simple version
    #LAMBDA.y <- t(t(S.yx) / diag(PSI)) * B.y # works only if no crossloadings
    lambda_y <- t(solve(mm_psi, t(s_yx))) * b_y
  }


  # assemble matrices
  mm_lambda <- matrix(0, nvar, nfac)
  mm_lambda[marker_idx, ] <- diag(nfac)
  mm_lambda[-marker_idx, ] <- lambda_y

  list(lambda = mm_lambda, theta = theta_nobounds, psi = psi_nobounds)
}

# internal function to be used inside lav_optim_noniter
# return 'x', the estimated vector of free parameters
lav_cfa_bentler1982_internal <- function(lavobject = NULL, # convenience
                                         # internal slot
                                         lavmodel = NULL,
                                         lavsamplestats = NULL,
                                         lavpartable = NULL,
                                         lavdata = NULL,
                                         lavoptions = NULL,
                                         gls = TRUE,
                                         min_reliability_marker = 0.1,
                                         quadprog = FALSE,
                                         nobs = 20L) {
  lavpta <- NULL
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))
    # extract slots
    lavmodel <- lavobject@Model
    lavsamplestats <- lavobject@SampleStats
    lavpartable <- lav_pt_set_cache(lavobject@ParTable, lavobject@pta)
    lavpta <- lavobject@pta
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
  }
  if (is.null(lavpta)) {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }
  # no structural part!
  if (any(lavpartable$op == "~")) {
    lav_msg_stop(gettext("bentler1982 estimator only available for CFA models"))
  }
  # no BETA matrix! (i.e., no higher-order factors)
  if (!is.null(lavmodel@GLIST$beta)) {
    lav_msg_stop(gettext("bentler1982 estimator not available
                 for models that require a BETA matrix"))
  }
  # no std.lv = TRUE for now
  if (lavoptions$std.lv) {
    lav_msg_stop(gettext(
      "bentler1982 estimator not available if std.lv = TRUE"))
  }

  nblocks <- lav_pt_nblocks(lavpartable)
  stopifnot(nblocks == 1L) # for now
  b <- 1L
  sample_cov <- lavsamplestats@cov[[b]]
  nvar <- nrow(sample_cov)
  marker_idx <- lavpta$vidx$lv.marker[[b]]
  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")
  lambda_nonzero_idx <- lavmodel@m.free.idx[[lambda_idx]]
  # only diagonal THETA for now...
  # because if we have correlated residuals, we should remove the
  # corresponding variables as instruments before we estimate lambda...
  # (see MIIV)
  theta_idx <- which(names(lavmodel@GLIST) == "theta") # usually '2'
  m_theta <- lavmodel@m.free.idx[[theta_idx]]
  nondiag_idx <- m_theta[!m_theta %in% lav_mat_diag_idx(nvar)]
  if (length(nondiag_idx) > 0L) {
    lav_msg_warn(gettext(
      "this implementation of FABIN does not handle correlated residuals yet!"))
  }

  if (!missing(gls)) {
    gls_flag <- gls
  } else {
    gls_flag <- FALSE
    if (!is.null(lavoptions$estimator.args$GLS) &&
      lavoptions$estimator.args$GLS) {
      gls_flag <- TRUE
    }
  }

  if (missing(quadprog) &&
    !is.null(lavoptions$estimator.args$quadprog)) {
    quadprog <- lavoptions$estimator.args$quadprog
  }

  # run bentler1982 non-iterative CFA algorithm
  out <- lav_cfa_bentler1982(
    s = sample_cov, marker_idx = marker_idx,
    lambda_nonzero_idx = lambda_nonzero_idx,
    gls = gls_flag,
    min_reliability_marker = 0.1,
    quadprog = quadprog,
    nobs = lavsamplestats@ntotal
  )
  mm_lambda <- out$lambda
  mm_theta <- diag(out$theta, nvar)
  mm_psi <- out$psi

  # store matrices in lavmodel@GLIST
  lavmodel@GLIST$lambda <- mm_lambda
  lavmodel@GLIST$theta <- mm_theta
  lavmodel@GLIST$psi <- mm_psi

  # extract free parameters only
  x <- lav_model_get_parameters(lavmodel)

  # apply bounds (if any)
  if (!is.null(lavpartable$lower)) {
    lower_x <- lavpartable$lower[lavpartable$free > 0]
    too_small_idx <- which(x < lower_x)
    if (length(too_small_idx) > 0L) {
      x[too_small_idx] <- lower_x[too_small_idx]
    }
  }
  if (!is.null(lavpartable$upper)) {
    upper_x <- lavpartable$upper[lavpartable$free > 0]
    too_large_idx <- which(x > upper_x)
    if (length(too_large_idx) > 0L) {
      x[too_large_idx] <- upper_x[too_large_idx]
    }
  }

  x
}

# numerical derivatives using complex numbers
# see Squire & Trapp 1998, siam rev 40(1) 110-112
# or Ridout, MS (2009), the american statistician 63(1) 66-74

# it would seem that you can choose h to be fairly small, without
# sacrificing accuracy due to rounding errors

# YR 17 July 2012

lav_func_grad_complex <- function(func, x,                        # nolint start
                                      h = .Machine$double.eps, ...,
                                      fallback_simple = TRUE) {       #nolint end
  f0 <- try(func(x * (0 + 1i), ...), silent = TRUE)
  if (!is.complex(f0)) {
    if (fallback_simple) {
      dx <- lav_func_grad_simple(func = func, x = x, h = sqrt(h), ...)
      return(dx)
    } else {
      lav_msg_stop(gettext(
        "function does not return a complex value")) # eg abs()
    }
  }
  if (inherits(f0, "try-error")) {
    if (fallback_simple) {
      dx <- lav_func_grad_simple(func = func, x = x, h = sqrt(h), ...)
      return(dx)
    } else {
      lav_msg_stop(gettext(
        "function does not support non-numeric (complex) argument"))
    }
  }
  if (length(f0) != 1L) {
    lav_msg_stop(gettext(
      "function is not scalar and returns more than one element"))
  }

  nvar <- length(x)

  # determine 'h' per element of x
  h <- pmax(h, abs(h * x))

  # get exact h, per x
  tmp <- x + h
  h <- (tmp - x)

  # simple 'forward' method
  dx <- rep(as.numeric(NA), nvar)
  for (p in seq_len(nvar)) {
    dx[p] <- Im(func(x + h * 1i * (seq.int(nvar) == p), ...)) / h[p]
  }

  dx
}

# as a backup, if func() is not happy about non-numeric arguments
lav_func_grad_simple <- function(func, x,
                                     h = sqrt(.Machine$double.eps), ...) {
  # check current point, see if it is a scalar function
  f0 <- func(x, ...)
  if (length(f0) != 1L) {
    lav_msg_stop(gettext(
      "function is not scalar and returns more than one element"))
  }

  nvar <- length(x)

  # determine 'h' per element of x
  h <- pmax(h, abs(h * x))

  # get exact h, per x
  tmp <- x + h
  h <- (tmp - x)

  # simple 'forward' method
  dx <- rep(as.numeric(NA), nvar)
  for (p in seq_len(nvar)) {
    dx[p] <- (func(x + h * (seq.int(nvar) == p), ...) - func(x, ...)) / h[p]
  }

  dx
}

lav_func_jacobian_complex <- function(func, x,
                                      h = .Machine$double.eps, ...,
                                      fallback_simple = TRUE) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  f0 <- try(func(x * (0 + 1i), ...), silent = TRUE)
  if (!is.complex(f0)) {
    if (fallback_simple) {
      dx <- lav_func_jacobian_simple(func = func, x = x, h = sqrt(h), ...)
      return(dx)
    } else {
      lav_msg_stop(gettext(
        "function does not return a complex value")) # eg abs()
    }
  }
  if (inherits(f0, "try-error")) {
    if (fallback_simple) {
      dx <- lav_func_jacobian_simple(func = func, x = x, h = sqrt(h), ...)
      return(dx)
    } else {
      lav_msg_stop(gettext(
        "function does not support non-numeric (complex) argument"))
    }
  }
  nres <- length(f0)
  nvar <- length(x)

  # determine 'h' per element of x
  h <- pmax(h, abs(h * x))

  # get exact h, per x
  tmp <- x + h
  h <- (tmp - x)

  # simple 'forward' method
  dx <- matrix(as.numeric(NA), nres, nvar)
  for (p in seq_len(nvar)) {
    dx[, p] <- Im(func(x + h * 1i * (seq.int(nvar) == p), ...)) / h[p]
  }

  dx
}

lav_func_jacobian_simple <- function(func, x,
                                     h = sqrt(.Machine$double.eps), ...) {
  f0 <- func(x, ...)
  nres <- length(f0)
  nvar <- length(x)

  # determine 'h' per element of x
  h <- pmax(h, abs(h * x))

  # get exact h, per x
  tmp <- x + h
  h <- (tmp - x)

  # simple 'forward' method
  dx <- matrix(as.numeric(NA), nres, nvar)
  for (p in seq_len(nvar)) {
    dx[, p] <- (func(x + h * (seq.int(nvar) == p), ...) - func(x, ...)) / h[p]
  }

  dx
}

# Hessian computation for a (possibly vector-valued) function func: R^k -> R^m
# Returns a list of length m, where the i-th element is the (k x k)
# symmetric Hessian matrix of the i-th component of func at x.
#
# The inner Jacobian is computed using the complex-step method (high
# accuracy), while the outer step uses forward finite differences.
# A small symmetrization step compensates for finite-difference asymmetry.
#
# Components whose Jacobian row is exactly zero at x (i.e. the function
# is locally constant for that component) are returned with a zero
# Hessian. Without this, forward-FD noise in the Jacobian can be
# amplified to O(1) when divided by the outer step.
lav_func_hessian_complex <- function(func, x,
                                     h = sqrt(.Machine$double.eps), ...,
                                     fallback_simple = TRUE) {
  # Jacobian at x
  j0 <- try(lav_func_jacobian_complex(func = func, x = x, ...,
                                      fallback_simple = fallback_simple),
            silent = TRUE)
  if (inherits(j0, "try-error")) {
    return(j0)
  }
  nres <- nrow(j0)
  nvar <- length(x)

  # outer step per element of x
  h_vec <- pmax(h, abs(h * x))
  # exact h per element (avoid floating-point asymmetry)
  tmp <- x + h_vec
  h_vec <- tmp - x

  # detect components whose function is (locally) constant: their
  # Jacobian row is essentially zero. We will zero their Hessians too.
  # The threshold accommodates numerical noise from the fallback simple
  # forward FD, which has O(sqrt(eps)) accuracy per element.
  j0_row_max <- apply(abs(j0), 1, max)
  constant_idx <- which(j0_row_max < .Machine$double.eps^(1 / 3))

  # build Hessians by finite-differencing the Jacobian
  h_list <- vector("list", nres)
  for (i in seq_len(nres)) {
    h_list[[i]] <- matrix(0, nrow = nvar, ncol = nvar)
  }
  for (j in seq_len(nvar)) {
    x_plus <- x
    x_plus[j] <- x[j] + h_vec[j]
    jp <- try(lav_func_jacobian_complex(func = func, x = x_plus, ...,
                                        fallback_simple = fallback_simple),
              silent = TRUE)
    if (inherits(jp, "try-error")) {
      return(jp)
    }
    for (i in seq_len(nres)) {
      h_list[[i]][, j] <- (jp[i, ] - j0[i, ]) / h_vec[j]
    }
  }

  # symmetrize (small numerical asymmetry from finite differences)
  for (i in seq_len(nres)) {
    h_list[[i]] <- 0.5 * (h_list[[i]] + t(h_list[[i]]))
  }

  # zero out Hessians for locally-constant components
  for (i in constant_idx) {
    h_list[[i]][] <- 0
  }

  h_list
}

# 'simple' version (forward finite differences only), used as a fallback
# when the complex-step Jacobian cannot be used (e.g. functions that
# do not accept complex arguments). See lav_func_hessian_complex() for
# the rationale behind zeroing Hessians of locally-constant components.
lav_func_hessian_simple <- function(func, x,
                                    h = .Machine$double.eps^(1 / 4), ...) {
  j0 <- lav_func_jacobian_simple(func = func, x = x, ...)
  nres <- nrow(j0)
  nvar <- length(x)

  h_vec <- pmax(h, abs(h * x))
  tmp <- x + h_vec
  h_vec <- tmp - x

  # detect components whose function is (locally) constant (see complex
  # variant for rationale)
  j0_row_max <- apply(abs(j0), 1, max)
  constant_idx <- which(j0_row_max < .Machine$double.eps^(1 / 3))

  h_list <- vector("list", nres)
  for (i in seq_len(nres)) {
    h_list[[i]] <- matrix(0, nrow = nvar, ncol = nvar)
  }
  for (j in seq_len(nvar)) {
    x_plus <- x
    x_plus[j] <- x[j] + h_vec[j]
    jp <- lav_func_jacobian_simple(func = func, x = x_plus, ...)
    for (i in seq_len(nres)) {
      h_list[[i]][, j] <- (jp[i, ] - j0[i, ]) / h_vec[j]
    }
  }

  for (i in seq_len(nres)) {
    h_list[[i]] <- 0.5 * (h_list[[i]] + t(h_list[[i]]))
  }

  for (i in constant_idx) {
    h_list[[i]][] <- 0
  }

  h_list
}

lav_deriv_cov2cor_b <- function(m_cov = NULL) {
  nvar <- nrow(m_cov)
  ds_inv <- 1 / diag(m_cov)
  m_r <- cov2cor(m_cov)
  m_a <- -m_r %x% (0.5 * diag(ds_inv))
  m_b <- (0.5 * diag(ds_inv)) %x% -m_r
  m_dd <- diag(lav_mat_vec(diag(nvar)))
  a2 <- m_a %*% m_dd
  b2 <- m_b %*% m_dd
  out <- a2 + b2 + diag(lav_mat_vec(tcrossprod(sqrt(ds_inv))))
  m_d <- lav_mat_dup(nvar)
  out_vech <- 0.5 * (t(m_d) %*% out %*% m_d)
  out_vech
}

# quick and dirty (FIXME!!!) way to get
# surely there must be a more elegant way?
# see lav_deriv_cov2cor_b, if no num.idx...
# dCor/dCov
lav_deriv_cov2cor <- function(cov_1 = NULL, num_idx = NULL) {
  # dCor/dvar1 = - cov / (2*var1 * sqrt(var1) * sqrt(var2))
  # dCor/dvar2 = - cov / (2*var2 * sqrt(var1) * sqrt(var2))
  # dCor/dcov  =  1/(sqrt(var1) * sqrt(var2))

  # diagonal: diag(lav_mat_vech(tcrossprod(1/delta)))

  nvar <- ncol(cov_1)
  pstar <- nvar * (nvar + 1) / 2
  delta <- sqrt(diag(cov_1))
  if (length(num_idx) > 0L) {
    delta[num_idx] <- 1.0
  }

  a <- cov_1 * -1 / (2 * delta * delta * tcrossprod(delta))
  if (length(num_idx) > 0L) {
    a[num_idx, ] <- 0
    a[cbind(num_idx, num_idx)] <- 1
  }
  a2 <- diag(nvar) %x% t(a)

  out <- diag(pstar)
  diag(out) <- lav_mat_vech(tcrossprod(1 / delta))
  var_idx <- lav_mat_diagh_idx(nvar)
  dup <- lav_mat_dup(nvar)
  out[, var_idx] <- t(dup) %*% a2[, lav_mat_diag_idx(nvar)]

  if (length(num_idx) > 0L) {
    var_idx <- var_idx[-num_idx]
  }
  out[var_idx, var_idx] <- 0

  out
}

# Reduced Bias M-estimation for SEM.

mod <- "
  # latent variables 
  ind60 =~ x1 + x2 + x3 
  dem60 =~ y1 + y2 + y3 + y4 
  dem65 =~ y5 + y6 + y7 + y8 
   
  # regressions
  dem60 ~ ind60 
  dem65 ~ ind60 + dem60 
  
  # residual covariances 
  y1 ~~ y5
  y2 ~~ y4 + y6 
  y3 ~~ y7 
  y4 ~~ y8
  y6 ~~ y8
"

# Maximum likelihood fit
fit_ML   <- sem(model = mod, data = PoliticalDemocracy)

# Implicit RBM fit (default)
fit_iRBM <- sem(model = mod, data = PoliticalDemocracy,
                estimator = list(estimator = "rbm", rbm.method = "implicit")) 

# Explicit RBM fit
fit_eRBM <- sem(model = mod, data = PoliticalDemocracy,
                estimator = list(estimator = "rbm", rbm.method = "explicit"))

# Compare
tab <- data.frame(
    ML = coef(fit_ML),
    iRBM = coef(fit_iRBM),
    eRBM = coef(fit_eRBM)
)
rownames(tab) <- names(coef(fit_ML))
print(round(tab, 3))

# merge two parameter tables
# - but allow different number of columns
lav_pt_merge <- function(pt1 = NULL, pt2 = NULL,        # nolint start
                               remove_duplicated = FALSE,
                               from_last = FALSE,
                               warn = TRUE) {                 # nolint end
  if (!missing(warn)) {
    current_warn <- lav_warn()
    if (lav_warn(warn))
      on.exit(lav_warn(current_warn), TRUE)
  }
  # check for empty pt2
  if (is.null(pt2) || length(pt2) == 0L) {
    return(pt1)
  }

  pt1 <- as.data.frame(pt1, stringsAsFactors = FALSE)
  pt2 <- as.data.frame(pt2, stringsAsFactors = FALSE)

  # check minimum requirements: lhs, op, rhs
  stopifnot(
    !is.null(pt1$lhs), !is.null(pt1$op), !is.null(pt1$rhs),
    !is.null(pt2$lhs), !is.null(pt2$op), !is.null(pt2$rhs)
  )

  # both should have block (or not)
  if (is.null(pt1$block) && is.null(pt2$block)) {
    pt1$block <- rep(1L, length(pt1$lhs))
    pt2$block <- rep(1L, length(pt2$lhs))
    tmp <- rbind(
      pt1[, c("lhs", "op", "rhs", "block")],
      pt2[, c("lhs", "op", "rhs", "block")]
    )
  } else {
    if (is.null(pt1$block) && !is.null(pt2$block)) {
      pt1$block <- rep(1L, length(pt1$lhs))
    } else if (is.null(pt2$block) && !is.null(pt1$block)) {
      pt2$block <- rep(1L, length(pt2$lhs))
    }
    tmp <- rbind(
      pt1[, c("lhs", "op", "rhs", "block")],
      pt2[, c("lhs", "op", "rhs", "block")]
    )
  }

  # if missing columns, provide default values of the right type
  # (numeric/integer/character)

  # group
  if (is.null(pt1$group) && !is.null(pt2$group)) {
    pt1$group <- rep(1L, length(pt1$lhs))
  } else if (is.null(pt2$group) && !is.null(pt1$group)) {
    pt2$group <- rep(1L, length(pt2$lhs))
  }

  # level
  if (is.null(pt1$level) && !is.null(pt2$level)) {
    pt1$level <- rep(1L, length(pt1$lhs))
  } else if (is.null(pt2$level) && !is.null(pt1$level)) {
    pt2$level <- rep(1L, length(pt2$lhs))
  }

  # user
  if (is.null(pt1$user) && !is.null(pt2$user)) {
    pt1$user <- rep(0L, length(pt1$lhs))
  } else if (is.null(pt2$user) && !is.null(pt1$user)) {
    pt2$user <- rep(0L, length(pt2$lhs))
  }

  # free
  if (is.null(pt1$free) && !is.null(pt2$free)) {
    pt1$free <- rep(0L, length(pt1$lhs))
  } else if (is.null(pt2$free) && !is.null(pt1$free)) {
    pt2$free <- rep(0L, length(pt2$lhs))
  }

  # ustart -- set to zero!!
  if (is.null(pt1$ustart) && !is.null(pt2$ustart)) {
    pt1$ustart <- rep(0, length(pt1$lhs))
  } else if (is.null(pt2$ustart) && !is.null(pt1$ustart)) {
    pt2$ustart <- rep(0, length(pt2$lhs))
  }

  # exo
  if (is.null(pt1$exo) && !is.null(pt2$exo)) {
    pt1$exo <- rep(0L, length(pt1$lhs))
  } else if (is.null(pt2$exo) && !is.null(pt1$exo)) {
    pt2$exo <- rep(0L, length(pt2$lhs))
  }

  # label
  if (is.null(pt1$label) && !is.null(pt2$label)) {
    pt1$label <- rep("", length(pt1$lhs))
  } else if (is.null(pt2$label) && !is.null(pt1$label)) {
    pt2$label <- rep("", length(pt2$lhs))
  }

  # plabel
  if (is.null(pt1$plabel) && !is.null(pt2$plabel)) {
    pt1$plabel <- rep("", length(pt1$lhs))
  } else if (is.null(pt2$plabel) && !is.null(pt1$plabel)) {
    pt2$plabel <- rep("", length(pt2$lhs))
  }

  # efa
  if (is.null(pt1$efa) && !is.null(pt2$efa)) {
    pt1$efa <- rep("", length(pt1$lhs))
  } else if (is.null(pt2$efa) && !is.null(pt1$efa)) {
    pt2$efa <- rep("", length(pt2$lhs))
  }

  # start
  if (is.null(pt1$start) && !is.null(pt2$start)) {
    pt1$start <- rep(as.numeric(NA), length(pt1$lhs))
  } else if (is.null(pt2$start) && !is.null(pt1$start)) {
    pt2$start <- rep(as.numeric(NA), length(pt2$lhs))
  }

  # est
  if (is.null(pt1$est) && !is.null(pt2$est)) {
    pt1$est <- rep(0, length(pt1$lhs))
  } else if (is.null(pt2$est) && !is.null(pt1$est)) {
    pt2$est <- rep(0, length(pt2$lhs))
  }


  # check for duplicated elements
  if (remove_duplicated) {
    # if fromLast = TRUE, idx is in pt1
    # if fromLast = FALSE, idx is in pt2
    idx <- which(duplicated(tmp, fromLast = from_last))

    if (length(idx)) {
      lav_msg_warn(
        gettext("duplicated parameters are ignored:"),
        paste(apply(tmp[idx, c("lhs", "op", "rhs")], 1,
          paste,
          collapse = " "
        ), collapse = "\n")
      )
      if (from_last) {
        pt1 <- pt1[-idx, ]
      } else {
        idx <- idx - nrow(pt1)
        pt2 <- pt2[-idx, ]
      }
    }
  } else if (!is.null(pt1$start) && !is.null(pt2$start)) {
    # copy start values from pt1 to pt2
    for (i in seq_along(pt1$lhs)) {
      idx <- which(pt2$lhs == pt1$lhs[i] &
        pt2$op == pt1$op[i] &
        pt2$rhs == pt1$rhs[i] &
        pt2$block == pt1$block[i])

      pt2$start[idx] <- pt1$start[i]
    }
  }

  # nicely merge, using 'id' column (if it comes first)
  if (is.null(pt1$id) && !is.null(pt2$id)) {
    nid <- max(pt2$id)
    pt1$id <- (nid + 1L):(nid + nrow(pt1))
  } else if (is.null(pt2$id) && !is.null(pt1$id)) {
    nid <- max(pt1$id)
    pt2$id <- (nid + 1L):(nid + nrow(pt2))
  }

  new_1 <- base::merge(pt1, pt2, all = TRUE, sort = FALSE)

  # make sure group/block/level are zero (or "") if
  # op %in% c("==", "<", ">", ":=")
  op_idx <- which(new_1$op %in% c("==", "<", ">", ":="))
  if (length(op_idx) > 0L) {
    if (!is.null(new_1$block)) {
      # ALWAYS integer
      new_1$block[op_idx] <- 0L
    }
    if (!is.null(new_1$group)) {
      if (is.character(new_1$level)) {
        new_1$group[op_idx] <- ""
      } else {
        new_1$group[op_idx] <- 0L
      }
    }
    if (!is.null(new_1$level)) {
      if (is.character(new_1$level)) {
        new_1$level[op_idx] <- ""
      } else {
        new_1$level[op_idx] <- 0L
      }
    }
  }

  new_1
}

lav_step16_rotation <- function(lavoptions = NULL,
                                       lavmodel = NULL,
                                       lavpartable = NULL,
                                       lavh1 = NULL,
                                       lavdata = NULL,
                                       x = NULL,
                                       lavvcov = NULL,
                                       vcov = NULL,
                                       lavcache = NULL,
                                       lavimplied = NULL,
                                       lavsamplestats = NULL) {
  # # # # # # # # # # #
  # #  16. rotation # #
  # # # # # # # # # # #

  # if lavmodel@nefa > 0L and lavoptions$rotation not "none"
  #   store unrotated  solution in partable (column est.unrotated)
  #   rotate lavmodel via lav_model_efa_rotate and overwrite column est
  #     in partable
  #   if lavoptions$se not in none, bootstrap, external, twostep
  #     if lavoptions$rotation.se == "delta"
  #       re-compute vcov with delta rule (*)
  #       re-compute SE and store them in lavpartable (*)
  #     else if lavoptions$rotation.se == "bordered"
  #       create 'new' partable where the user = 7/77 parameters are free (*)
  #
  # (*) code too complicated to summarize here

  # bootstrap results for the rotated solution (filled in below if
  # se = "bootstrap"); NULL means 'use whatever step 13 produced'
  lavboot <- NULL

  if (lavmodel@nefa > 0L &&
    (lavoptions$rotation != "none")) {
    # store unrotated solution in partable
    lavpartable$est.unrotated <- lavpartable$est
    lavpartable$se.unrotated <- lavpartable$se

    # rotate, and create new lavmodel
    if (lav_verbose()) {
      cat(
        "rotating EFA factors using rotation method =",
        toupper(lavoptions$rotation), "..."
      )
    }
    x_unrotated <- as.numeric(x)
    lavmodel_unrot <- lavmodel
    # keep an unrotated copy of the parameter table: the bootstrap (if any)
    # refits each sample from the unrotated model and rotates it (see #376)
    lavpartable_unrotated <- lavpartable
    efa_out <- lav_model_efa_rotate(
      lavmodel = lavmodel,
      lavoptions = lavoptions
    )

    # adapt partable:
    # - change 'free' column to reflect that user = 7/77 parameters are free
    # - save unrotated free column in free.unrotated
    lavpartable$free.unrotated <- lavpartable$free
    user7_idx <- which((lavpartable$user == 7L | lavpartable$user == 77L) &
      lavpartable$free == 0L)
    lavpartable$free[user7_idx] <- 1L
    lavpartable$free[lavpartable$free > 0L] <-
      seq_len(sum(lavpartable$free > 0L))
    # avoid cin.simple entries for these user=7 parameters
    if (!is.null(lavpartable$lower)) {
      lavpartable$lower[user7_idx] <- -Inf
    }
    if (!is.null(lavpartable$upper)) {
      lavpartable$upper[user7_idx] <- +Inf
    }

    # create 'rotated' lavmodel, reflecting the 'new' free parameters
    lavmodel <- lav_model(
      lavpartable = lavpartable,
      lavoptions = lavoptions,
      th_idx = lavmodel@th.idx
    )

    # add rotated information
    lavmodel@H <- efa_out$H
    lavmodel@lv.order <- efa_out$lv.order
    lavmodel@GLIST <- efa_out$GLIST

    # add con.jac information (if any)
    lavmodel@con.lambda <- lavmodel_unrot@con.lambda
    if (nrow(lavmodel_unrot@con.jac) > 0L) {
      # the re-parsed constraint set of the ROTATED model may have MORE
      # equality rows than the unrotated one (equality rows involving the
      # temporarily-fixed user=7 EFA identification parameters were
      # removed as zero rows by lav_con_parse, and reappear here), so the
      # cin/inactive indices and con.lambda of the unrotated fit -- which
      # are aligned with the unrotated row layout -- must be rebuilt for
      # the new layout, not copied verbatim (copying left con.lambda
      # shorter than con.jac, and the cin/inactive indices off by the
      # difference in equality rows)
      nceq_new <- nrow(lavmodel@ceq.JAC)
      ncin <- nrow(lavmodel@cin.JAC)
      con_jac <- rbind(lavmodel@ceq.JAC, lavmodel@cin.JAC)
      old_cin_idx <- attr(lavmodel_unrot@con.jac, "cin.idx")
      old_ina_idx <- attr(lavmodel_unrot@con.jac, "inactive.idx")
      old_lambda <- lavmodel_unrot@con.lambda
      attr(con_jac, "ceq.idx") <- seq_len(nceq_new)
      attr(con_jac, "cin.idx") <- nceq_new + seq_len(ncin)
      # inactive rows are inequality rows; re-express their positions
      attr(con_jac, "inactive.idx") <-
        nceq_new + which(old_cin_idx %in% old_ina_idx)
      lavmodel@con.jac <- con_jac
      # equality multipliers of the rotated fit are unknown (and unused:
      # they only enter the decoupled lambda block of the bordered
      # matrix); the inequality rows keep their multipliers
      lavmodel@con.lambda <- c(
        numeric(nceq_new),
        old_lambda[old_cin_idx]
      )
    }

    # overwrite parameters in @ParTable$est
    lavpartable$est <- lav_model_get_parameters(
      lavmodel = lavmodel,
      type = "user", extra = TRUE
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }

    # bootstrap standard errors for the rotated solution (see github #376).
    # We cannot reuse the analytic machinery below: each bootstrap sample is
    # refit from the unrotated model, rotated, and its factors aligned (sign +
    # order) to the original rotated solution before its variability is used.
    if (lavoptions$se == "bootstrap") {
      if (lav_verbose()) {
        cat("computing bootstrap VCOV for rotated solution ...")
      }

      # reference rotated loadings (per block, per set) for factor alignment,
      # plus the rotated free-parameter layout
      ref_lambda <- lav_efa_bootstrap_lambda_list(lavmodel = lavmodel)
      free_idx <- which(lavpartable$free > 0L & !duplicated(lavpartable$free))
      efa_ref <- list(
        ref_lambda = ref_lambda,
        npar = length(free_idx),
        names = lav_pt_labels(lavpartable, type = "free")
      )

      # run the bootstrap on the unrotated model
      coef_boot <- lav_efa_bootstrap_run(
        lavmodel = lavmodel_unrot, lavpartable = lavpartable_unrotated,
        lavsamplestats = lavsamplestats, lavoptions = lavoptions,
        lavdata = lavdata, efa_ref = efa_ref
      )

      # always warn for failed / nonadmissible runs
      nfailed <- length(attr(coef_boot, "error.idx"))
      if (nfailed > 0L) {
        lav_msg_warn(gettextf(
          "%s bootstrap runs failed or did not converge.", nfailed))
      }
      notok <- length(attr(coef_boot, "nonadmissible"))
      if (notok > 0L) {
        lav_msg_warn(gettextf(
          "%s bootstrap runs resulted in nonadmissible solutions.", notok))
      }

      # vcov + se for the rotated parameters
      tmp_boot <- lav_efa_bootstrap_vcov(
        coef_boot = coef_boot, lavmodel = lavmodel, lavpartable = lavpartable,
        lavoptions = lavoptions
      )
      lavpartable$se <- tmp_boot$se
      lavboot <- list(coef = tmp_boot$coef)

      # store the rotated bootstrap vcov (strip attributes but 'dim')
      vcov1 <- tmp_boot$vcov
      attributes(vcov1) <- attributes(vcov1)["dim"]
      lavvcov <- list(
        se = lavoptions$se,
        information = lavoptions$information[1],
        vcov = vcov1
      )

      if (lav_verbose()) {
        cat(" done.\n")
      }
    }

    # VCOV rotated parameters
    if (!lavoptions$se %in% c("none", "bootstrap", "external", "two.step")) {
      if (lav_verbose()) {
        cat(
          "computing VCOV for se =", lavoptions$se,
          "and rotation.se =", lavoptions$rotation.se, "..."
        )
      }

      # use delta rule to recompute vcov
      if (lavoptions$rotation.se == "delta") {
        # Jacobian
        jac <- numDeriv::jacobian(
          func = lav_model_efa_rotate_x,
          x = x_unrotated, lavmodel = lavmodel_unrot,
          init_rot = lavmodel@H, lavoptions = lavoptions,
          type = "user", extra = FALSE,
          method.args = list(eps = 0.0050),
          method = "simple"
        ) # important!

        # force VCOV to be pd, before we transform (not very elegant)
        vcov_in <- lav_mat_sym_force_pd(lavvcov$vcov,
          tol = 1e-10
        )
        # apply Delta rule
        vcov_user <- jac %*% vcov_in %*% t(jac)

        # re-compute SE and store them in lavpartable
        tmp <- diag(vcov_user)
        min_idx <- which(tmp < 0)
        if (length(min_idx) > 0L) {
          tmp[min_idx] <- as.numeric(NA)
        }
        tmp <- sqrt(tmp)
        # catch near-zero SEs  (was ^(1/2) < 0.6)
        zero_idx <- which(tmp < .Machine$double.eps^(1 / 3))
        if (length(zero_idx) > 0L) {
          tmp[zero_idx] <- 0.0
        }
        lavpartable$se <- tmp

        # store rotated VCOV
        # lavvcov$vcov.unrotated <- lavvcov$vcov
        if (lavmodel@ceq.simple.only) {
          free_idx <- which(lavpartable$free > 0L &
            !duplicated(lavpartable$free))
        } else {
          free_idx <- which(lavpartable$free > 0L)
        }
        lavvcov$vcov <- vcov_user[free_idx, free_idx, drop = FALSE]

        # rotation.se = "bordered" is the default
      } else if (lavoptions$rotation.se == "bordered") {
        # create 'border' for augmented information matrix
        x_rot <- lav_model_get_parameters(lavmodel)
        jac <- numDeriv::jacobian(
          func = lav_model_efa_rotate_border_x,
          x = x_rot, lavmodel = lavmodel,
          lavoptions = lavoptions,
          lavpartable = lavpartable,
          # method.args = list(eps = 0.0005),
          # method = "simple")
          method = "Richardson"
        )
        # store JAC
        lavmodel@ceq.efa.JAC <- jac

        # no other constraints
        if (nrow(lavmodel@con.jac) == 0L) {
          lavmodel@con.jac <- jac
          attr(lavmodel@con.jac, "inactive.idx") <- integer(0L)
          attr(lavmodel@con.jac, "ceq.idx") <- seq_len(nrow(jac))
          attr(lavmodel@con.jac, "cin.idx") <- integer(0L)
          lavmodel@con.lambda <- rep(0, nrow(jac))

          # other constraints
        } else {
          old_jac <- lavmodel@con.jac
          inactive_idx <- attr(old_jac, "inactive.idx")
          ceq_idx <- attr(old_jac, "ceq.idx")
          cin_idx <- attr(old_jac, "cin.idx")
          lambda <- lavmodel@con.lambda
          nbord <- nrow(jac)

          # prepend the rotation border to the EXISTING con.jac rows (NOT
          # to rbind(ceq.JAC, cin.JAC)): since 0.6-19 the con.jac of a
          # ceq.simple/bounds model is SYNTHESIZED (eg the orthogonal
          # complement of ceq.simple.K), so its rows -- which is what the
          # ceq.idx/cin.idx/inactive.idx attributes and con.lambda are
          # aligned with -- do not correspond to ceq.JAC/cin.JAC rows;
          # the old rbind left con.lambda (and the offset attributes)
          # misaligned with the merged matrix. For nonlinear constraints
          # the existing con.jac is also the at-the-solution Jacobian,
          # which is the correct linearization (ceq.JAC/cin.JAC are
          # evaluated at parse time).
          con_jac_1 <- rbind(jac, old_jac)
          attr(con_jac_1, "cin.idx") <- cin_idx + nbord
          attr(con_jac_1, "ceq.idx") <- c(seq_len(nbord), ceq_idx + nbord)
          attr(con_jac_1, "inactive.idx") <- inactive_idx + nbord

          lavmodel@con.jac <- con_jac_1
          lavmodel@con.lambda <- c(rep(0, nbord), lambda)
        }

        # compute VCOV, taking 'rotation constraints' into account
        vcov <- lav_model_vcov(
          lavmodel = lavmodel,
          lavsamplestats = lavsamplestats,
          lavoptions = lavoptions,
          lavdata = lavdata,
          lavpartable = lavpartable,
          lavcache = lavcache,
          lavimplied = lavimplied,
          lavh1 = lavh1
        )

        # compute SE and store them in lavpartable
        tmp <- lav_model_vcov_se(
          lavmodel = lavmodel,
          lavpartable = lavpartable, vcov = vcov,
          lavoptions = lavoptions
        )
        lavpartable$se <- tmp

        # store rotated VCOV in lavvcov
        tmp_attr <- attributes(vcov)
        vcov1 <- vcov
        attributes(vcov1) <- tmp_attr["dim"]
        # lavvcov$vcov.unrotated <- lavvcov$vcov
        lavvcov$vcov <- vcov1
      } # bordered

      if (lav_verbose()) {
        cat(" done.\n")
      }
    } # vcov
  } # efa

  list(
    lavpartable = lavpartable,
    lavmodel    = lavmodel,
    lavvcov     = lavvcov,
    lavboot     = lavboot
  )
}

# compute model implied statistics
# per block

# YR 7 May 2022: add cov.x and mean.x if conditional.x (so that we do
#                no longer depend on SampleStats)

lav_model_implied <- function(lavmodel = NULL, glist = NULL, delta = TRUE, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  stopifnot(inherits(lavmodel, "lavModel"))

  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  # model-implied variance/covariance matrix ('sigma hat')
  sigma_hat <- lav_model_sigma(
    lavmodel = lavmodel, glist = glist,
    delta = delta
  )

  # model-implied mean structure ('mu hat')
  if (lavmodel@meanstructure) {
    mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
  } else {
    mu_hat <- vector("list", length = lavmodel@nblocks)
  }

  # if conditional.x, slopes, cov.x, mean.x
  if (lavmodel@conditional.x) {
    slopes <- lav_model_pi(lavmodel = lavmodel, glist = glist)

    # per block, because for some blocks, cov.x may not exist
    cov_x <- vector("list", lavmodel@nblocks)
    mean_x <- vector("list", lavmodel@nblocks)
    for (b in seq_len(lavmodel@nblocks)) {
      mm_in_block <- (seq_len(lavmodel@nmat[b]) +
        cumsum(c(0, lavmodel@nmat))[b])
      mlist <- lavmodel@GLIST[mm_in_block]
      cov_x_idx <- which(names(mlist) == "cov.x")
      if (length(cov_x_idx) > 0L) {
        cov_x[[b]] <- mlist[[cov_x_idx]]
      } else {
        cov_x[[b]] <- matrix(0, 0L, 0L)
      }
      mean_x_idx <- which(names(mlist) == "mean.x")
      if (length(mean_x_idx) > 0L) {
        mean_x[[b]] <- mlist[[mean_x_idx]]
      } else {
        mean_x[[b]] <- matrix(0, 0L, 1L)
      }
    }
  } else {
    slopes <- vector("list", length = lavmodel@nblocks)
  }

  # if categorical, model-implied thresholds
  if (lavmodel@categorical) {
    th <- lav_model_th(lavmodel = lavmodel, glist = glist)
  } else {
    th <- vector("list", length = lavmodel@nblocks)
  }

  if (lavmodel@group.w.free) {
    w_idx <- which(names(lavmodel@GLIST) == "gw")
    gw <- unname(glist[w_idx])
    gw <- lapply(gw, as.numeric)
  } else {
    gw <- vector("list", length = lavmodel@nblocks)
  }

  if (lavmodel@conditional.x) {
    implied <- list(
      res.cov = sigma_hat, res.int = mu_hat,
      res.slopes = slopes, cov.x = cov_x, mean.x = mean_x,
      res.th = th, group.w = gw
    )
  } else {
    implied <- list(cov = sigma_hat, mean = mu_hat, th = th, group.w = gw)
  }

  implied
}

# convert 'conditional.x = TRUE' to 'conditional.x = FALSE'
lav_model_implied_cond2uncond <- function(lavimplied) {
  # check for res.cov
  if (is.null(lavimplied$res.cov[[1]])) {
    # already unconditional
    return(lavimplied)
  } else {
    nblocks <- length(lavimplied$res.cov)
  }

  cov_1 <- vector("list", length = nblocks)
  mean_1 <- vector("list", length = nblocks)

  # reconstruct COV/MEAN per block
  for (b in seq_len(nblocks)) {
    res_sigma <- lavimplied$res.cov[[b]]
    res_slopes <- lavimplied$res.slopes[[b]]
    res_int <- lavimplied$res.int[[b]]
    s_xx <- lavimplied$cov.x[[b]]
    m_x <- lavimplied$mean.x[[b]]

    s_yx <- res_slopes %*% s_xx
    s_xy <- t(s_yx)
    s_yy <- res_sigma + tcrossprod(s_yx, res_slopes)
    cov_1[[b]] <- rbind(cbind(s_yy, s_yx), cbind(s_xy, s_xx))

    mu_y <- as.vector(res_int + res_slopes %*% m_x)
    mu_x <- as.vector(m_x)
    mean_1[[b]] <- matrix(c(mu_y, mu_x), ncol = 1L)
  }

  # we ignore res.th for now, as we do not support categorical data
  # in the two-level setting anyway
  implied <- list(
    cov = cov_1, mean = mean_1, th = lavimplied$res.th,
    group.w = lavimplied$group.w
  )

  implied
}

# STEP 2 of the Muthen (1984) three-stage estimator: correlations
#
# given the univariate fits of step 1, estimate the correlation of each
# variable pair (pearson, polyserial or polychoric) with the univariate
# parameters held fixed at their step-1 values ('twostep')
#
# YR/LDW 2026: refactored; fixed the swapped y1_name/y2_name in the
#              polychoric branch (affected warning messages only)

lav_samp_step2 <- function(uni = NULL,
                                  wt = NULL,
                                  ov_names = NULL, # error message only
                                  # polychoric and empty cells
                                  zero_add = c(0.5, 0.0),
                                  zero_keep_margins = TRUE,
                                  zero_cell_warn = FALSE,
                                  # keep track of tables with zero cells?
                                  zero_cell_tables = TRUE) {
  nvar <- length(uni)
  cor_1 <- diag(nvar)

  if (zero_cell_tables) {
    zero_var1 <- character(0L)
    zero_var2 <- character(0L)
  }

  # one-by-one (for now)
  for (j in seq_len(nvar - 1L)) {
    for (i in (j + 1L):nvar) {
      ord_i <- !is.null(uni[[i]]$th_idx)
      ord_j <- !is.null(uni[[j]]$th_idx)
      if (!ord_i && !ord_j) {
        # pearson correlation
        rho <- lav_bvreg_cor_twostep_fit(
          fit_y1 = uni[[i]], # linear
          fit_y2 = uni[[j]], # linear
          wt = wt,
          y1_name = ov_names[i],
          y2_name = ov_names[j]
        )
      } else if (!ord_i && ord_j) {
        # polyserial
        rho <- lav_bvmix_cor_twostep_fit(
          fit_y1 = uni[[i]], # linear
          fit_y2 = uni[[j]], # ordinal
          wt = wt,
          y1_name = ov_names[i],
          y2_name = ov_names[j]
        )
      } else if (!ord_j && ord_i) {
        # polyserial
        rho <- lav_bvmix_cor_twostep_fit(
          fit_y1 = uni[[j]], # linear
          fit_y2 = uni[[i]], # ordinal
          wt = wt,
          y1_name = ov_names[j],
          y2_name = ov_names[i]
        )
      } else {
        # polychoric correlation
        rho <- lav_bvord_cor_twostep_fit(
          fit_y1 = uni[[j]], # ordinal
          fit_y2 = uni[[i]], # ordinal
          wt = wt,
          zero_add = zero_add,
          zero_keep_margins = zero_keep_margins,
          zero_cell_warn = zero_cell_warn,
          zero_cell_flag = zero_cell_tables,
          y1_name = ov_names[j],
          y2_name = ov_names[i]
        )
        if (zero_cell_tables) {
          if (attr(rho, "zero.cell.flag")) {
            zero_var1 <- c(zero_var1, ov_names[j])
            zero_var2 <- c(zero_var2, ov_names[i])
          }
          attr(rho, "zero.cell.flag") <- NULL
        }
      }
      cor_1[i, j] <- cor_1[j, i] <- rho

      # check for near 1.0 correlations
      if (abs(cor_1[i, j]) > 0.99) {
        lav_msg_warn(gettextf(
          "correlation between variables %1$s and %2$s is (nearly) 1.0",
          ov_names[i], ov_names[j]))
      }
    }
  }

  # keep track of tables with zero cells
  if (zero_cell_tables) {
    zero_cell_tables <- cbind(zero_var1, zero_var2)
    attr(cor_1, "zero.cell.tables") <- zero_cell_tables
  }

  cor_1
}

# This code is written by YR (using lavaan components), but based on
# research code written by Mariska Barendse (Groningen/Amsterdam, NL)
#
# September 2013
#
# Three fit indices for the PML estimator (if all categorical, no exo)
# - Cp(max)
# - CF
# - CM

# FIXME: how to handle multiple groups??

# Mariska Barendse Cp statistic
# lav_tables_fit_Cp <- function(object, alpha = 0.05) {
#
#    out <- lavTablesFit(object, statistic = "G2", p.value = TRUE)
#
#    # Bonferroni adjusted p-value
#    ntests <- length(out$lhs)
#    out$alpha.adj <- alpha / ntests
#    #out$pval <- pchisq(out$G2, df=out$df, lower.tail = FALSE)
#
#    # remove G2.h0.pval
#    #out$G2.h0.pval <- NULL
#
#    out
# }

lavTablesFitCp <- function(object, alpha = 0.05) {   # nolint
  lavdata <- object@Data

  if (!any(lavdata@ov$type == "ordered")) {
    return(list(
      G2 = as.numeric(NA), df = as.numeric(NA),
      p.value = as.numeric(NA), p.value.Bonferroni = as.numeric(NA)
    ))
  }

  tf <- lavTables(object,
    dimension = 2L, type = "table",
    statistic = "G2", p_value = TRUE
  )

  # Bonferroni adjusted p-value
  ntests <- length(tf$lhs)
  tf$alpha.adj <- alpha / ntests

  out <- subset(tf, tf$G2.pval < tf$alpha.adj)

  # find largest G2
  max_idx <- which(tf$G2 == max(tf$G2))

  extra <- list(
    G2 = unname(tf$G2[max_idx]), df = unname(tf$df[max_idx]),
    lhs = tf$lhs[max_idx],
    rhs = tf$rhs[max_idx],
    group = tf$group[max_idx],
    p.value = unname(tf$G2.pval[max_idx]),
    ntests = ntests,
    p.value.Bonferroni = unname(tf$G2.pval[max_idx] * length(tf$lhs))
  )

  attr(out, "CpMax") <- extra

  class(out) <- c("lavaan.tables.fit.Cp", "lavaan.data.frame", "data.frame")

  out
}

lav_tables_fit_cp_print <- function(x, ...) {
  cat("CP-values that are significant at a Bonferroni",
      "adjusted level of significance\n")
  tmp <- x
  class(tmp) <- c("lavaan.data.frame", "data.frame")
  print(tmp)
}

# Mariska Barendse CF statistic
lavTablesFitCf <- function(object) {    # nolint
  # check object class
  if (!inherits(object, "lavaan")) {
    lav_msg_stop(gettext("object must be an object of class lavaan"))
  }
  lavdata <- object@Data
  lavpta <- object@pta
  lavmodel <- object@Model
  lavcache <- object@Cache
  implied <- object@implied

  cf_group <- rep(as.numeric(NA), lavdata@ngroups)
  df_group <- rep(as.numeric(NA), lavdata@ngroups)

  # check if all ordered
  if (!any(lavdata@ov$type == "ordered")) {
    cf <- as.numeric(NA)
    attr(cf, "CF.group") <- cf_group
    attr(cf, "DF.group") <- df_group
    return(cf)
  }

  # ord var in this group
  ov_ord <- unique(unlist(lavpta$vnames$ov.ord))
  ov_idx <- which(ov_ord %in% lavdata@ov$name)
  ov_nlev <- lavdata@ov$nlev[ov_idx]

  sigma_hat <- if (lavmodel@conditional.x) implied$res.cov else implied$cov
  th <- if (lavmodel@conditional.x) implied$res.th else implied$th
  df_1 <- prod(ov_nlev) - object@optim$npar - 1L

  for (g in seq_len(lavdata@ngroups)) {
    f_group <- lav_model_objective_fml(
      sigma_hat = sigma_hat[[g]],
      th = th[[g]],
      th_idx = lavmodel@th.idx[[g]],
      num_idx = lavmodel@num.idx[[g]],
      x = lavdata@X[[g]],
      lavcache = lavcache[[g]]
    )
    cf_group[g] <- 2 * lavdata@nobs[[g]] * f_group
  }

  # check for negative values
  cf_group[cf_group < 0] <- 0.0

  # global test statistic
  cf <- sum(cf_group)

  attr(cf, "CF.group") <- cf_group
  attr(cf, "DF") <- df_1
  attr(cf, "rpat.observed") <- sapply(lavdata@Rp, "[[", "npatterns")
  attr(cf, "rpat.total") <- sapply(lavdata@Rp, "[[", "total.patterns")
  attr(cf, "rpat.empty") <- sapply(lavdata@Rp, "[[", "empty.patterns")

  class(cf) <- c("lavaan.tables.fit.Cf", "numeric")

  cf
}

lav_tables_fit_cf_print <- function(x, ...) {
  cat("Total response patterns: ", attr(x, "rpat.total"), "\n")
  cat("Observed response patterns: ", attr(x, "rpat.observed"), "\n")
  cat("Empty response patterns: ", attr(x, "rpat.empty"), "\n")
  cat("Cf results may be biased because of large numbers of empty",
      "cells in the multivariate contingency table\n")
  cat("Cf-value, overall:\n")
  cf <- unclass(x)
  attributes(cf) <- NULL
  print(cf)
  cf_group <- attr(x, "CF.group")
  if (length(cf_group) > 1L) {
    cat("Cf-value, per group:\n")
    print(cf_group)
  }
  cat("Degrees of freedom\n")
  print(attr(x, "DF"))
}

lavTablesFitCm <- function(object) {  # nolint
  lavdata <- object@Data
  lavoptions <- object@Options

  cf_h0 <- lavTablesFitCf(object)

  # fit unrestricted model
  h1 <- lav_object_cor(lavdata,
    estimator = lavoptions$estimator,
    se = "none", test = "none", output = "lavaan"
  )
  cf_h1 <- lavTablesFitCf(h1)

  cf_h0_group <- attr(cf_h0, "CF.group")
  cf_h1_group <- attr(cf_h1, "CF.group")
  df_h0 <- attr(cf_h0, "DF")
  df_h1 <- attr(cf_h1, "DF")

  attributes(cf_h0) <- NULL
  attributes(cf_h1) <- NULL

  cm_1 <- cf_h0 - cf_h1
  attr(cm_1, "CM.group") <- cf_h0_group - cf_h1_group
  attr(cm_1, "DF") <- df_h0 - df_h1

  class(cm_1) <- c("lavaan.tables.fit.Cm", "numeric")

  cm_1
}


lav_tables_fit_cm_print <- function(x, ...) {
  # cat("The percentage of empty cells\n") #weet niet goed want FML werkt niet
  # cat("CM results may be a little biased because of large numbers of empty ",
  #      "cells in the multivariate contingency table\n")
  cat("Cm-value, overall:\n")
  cm_1 <- unclass(x)
  attributes(cm_1) <- NULL
  print(cm_1)
  cm_group <- attr(x, "CM.group")
  if (length(cm_group) > 1L) {
    cat("Cm-value, per group:\n")
    print(cm_group)
  }
  cat("Degrees of freedom:\n")
  print(attr(x, "DF"))
}

# extract info from model
lav_model_plotinfo <- function(model = NULL, infile = NULL, varlv = FALSE) {
  if (is.null(model) == is.null(infile)) {
    lav_msg_stop(gettext("either model or infile must be specified"))
  }
  edge_label <- function(label, fix) {
    if (is.null(fix)) fix <- ""
    if (label == "" && fix == "") {
      return("")
    }
    if (label == "") {
      return(fix)
    }
    if (fix == "") {
      return(label)
    }
    paste(label, fix, sep = "=")
  }
  if (!is.null(infile)) {
    if (!file.exists(infile)) {
      lav_msg_stop(gettextf("Infile %s not found!", infile))
    }
    model <- readLines(infile)
  }
  if (inherits(model, "lavaan")) {
    model <- as.data.frame(model@ParTable)
    model <- model[model$user == 1, ]
    model$fixed <- round(model$est, 3L)
    if (max(model$block) == 1L) model$block <- rep(0L, length(model$block))
    model$est <- NULL
  }
  if (is.list(model) && !is.null(model$op) && !is.null(model$lhs) &&
    !is.null(model$rhs) && !is.null(model$label) &&
    !is.null(model$fixed)) {
    tbl <- as.data.frame(model)
  } else if (is.character(model)) {
    tbl <- lavParseModelString(
      model_syntax = model,
      as_data_frame = TRUE
    )
  } else {
    lav_msg_stop(gettext(
      "model, or content of infile, not in interpretable form!"
    ))
  }
  if (is.null(tbl$block)) {
    tbl$block <- 0L
  } else if (all(tbl$block == tbl$block[1L])) {
    tbl$block <- 0L
  } else {
    blockmin <- min(tbl$block)
    tbl$block[tbl$block == blockmin] <- 999L
    tbl$block[tbl$block != 999L] <- 2L
    tbl$block[tbl$block == 999L] <- 1L
  }
  create_node_id <- function(blok, naam) {
    if (blok == 0L) {
      naam
    } else {
      paste(blok, naam, sep = ".")
    }
  }
  maxedges <- nrow(tbl)
  maxnodes <- 2 * maxedges
  nodes <- data.frame(
    id = character(maxnodes),
    naam = character(maxnodes),
    tiepe = character(maxnodes), # ov, lv, varlv, cv, wov, bov, const
    # cv: composite; wov = within; bov = between; const = intercept
    blok = integer(maxnodes)
  )
  edges <- data.frame(
    id = integer(maxedges),
    label = character(maxedges),
    van = character(maxedges),
    naar = character(maxedges),
    tiepe = character(maxedges) # lavaan operator, behalve ~~ self -> "~~~"
  )
  curnode <- 0L
  curedge <- 0L
  for (i in seq.int(nrow(tbl))) {
    if (tbl$op[i] == "=~") {
      #### =~ : is manifested by ####
      # lhs node
      jl <- match(create_node_id(tbl$block[i], tbl$lhs[i]),
                  nodes$id, nomatch = 0L)
      if (jl == 0L) {
        curnode <- curnode + 1L
        jl <- curnode
        nodes$naam[curnode] <- tbl$lhs[i]
        nodes$tiepe[curnode] <- "lv"
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        nodes$tiepe[jl] <- "lv"
      }
      # rhs node
      jr <- match(create_node_id(tbl$block[i], tbl$rhs[i]),
                  nodes$id, nomatch = 0L)
      nodetype <- "ov"
      if (length(unique(tbl$block[tbl$rhs == tbl$rhs[i]])) > 1L) {
        nodetype <- switch(tbl$block[i],
          "wov",
          "bov"
        )
      }
      if (jr == 0L) {
        curnode <- curnode + 1L
        jr <- curnode
        nodes$naam[curnode] <- tbl$rhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jr] == "") nodes$tiepe[jr] <- nodetype
      }
      # edge
      curedge <- curedge + 1L
      edges$id[curedge] <- curedge
      edges$label[curedge] <- edge_label(tbl$label[i], tbl$fixed[i])
      edges$van[curedge] <- nodes$id[jl]
      edges$naar[curedge] <- nodes$id[jr]
      edges$tiepe[curedge] <- tbl$op[i]
    } else if (tbl$op[i] == "<~") {
      #### <~ : is a result of ####
      # lhs node
      jl <- match(create_node_id(tbl$block[i], tbl$lhs[i]),
                  nodes$id, nomatch = 0L)
      if (jl == 0L) {
        curnode <- curnode + 1L
        jl <- curnode
        nodes$naam[curnode] <- tbl$lhs[i]
        nodes$tiepe[curnode] <- "cv"
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        nodes$tiepe[jl] <- "cv"
      }
      # rhs node
      jr <- match(create_node_id(tbl$block[i], tbl$rhs[i]),
                  nodes$id, nomatch = 0L)
      nodetype <- "ov"
      if (length(unique(tbl$block[tbl$rhs == tbl$rhs[i]])) > 1L) {
        nodetype <- switch(tbl$block[i],
          "wov",
          "bov"
        )
      }
      if (jr == 0L) {
        curnode <- curnode + 1L
        jr <- curnode
        nodes$naam[curnode] <- tbl$rhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jr] == "") nodes$tiepe[jr] <- nodetype
      }
      # edge
      curedge <- curedge + 1L
      edges$id[curedge] <- curedge
      edges$label[curedge] <- edge_label(tbl$label[i], tbl$fixed[i])
      edges$van[curedge] <- nodes$id[jr]
      edges$naar[curedge] <- nodes$id[jl]
      edges$tiepe[curedge] <- tbl$op[i]
    } else if (tbl$op[i] == "~") {
      #### ~ : is regressed on ####
      # lhs node
      jl <- match(create_node_id(tbl$block[i], tbl$lhs[i]),
                                 nodes$id, nomatch = 0L)
      nodetype <- "ov"
      if (length(unique(tbl$block[tbl$rhs == tbl$rhs[i]])) > 1L) {
        nodetype <- switch(tbl$block[i],
          "wov",
          "bov"
        )
      }
      if (jl == 0L) {
        curnode <- curnode + 1L
        jl <- curnode
        nodes$naam[curnode] <- tbl$lhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jl] == "") nodes$tiepe[jl] <- nodetype
      }
      # rhs node
      jr <- match(create_node_id(tbl$block[i], tbl$rhs[i]),
                  nodes$id, nomatch = 0L)
      if (jr == 0L) {
        curnode <- curnode + 1L
        jr <- curnode
        nodes$naam[curnode] <- tbl$rhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jr] == "") nodes$tiepe[jr] <- nodetype
      }
      # edge
      curedge <- curedge + 1L
      edges$id[curedge] <- curedge
      edges$label[curedge] <- edge_label(tbl$label[i], tbl$fixed[i])
      edges$van[curedge] <- nodes$id[jr]
      edges$naar[curedge] <- nodes$id[jl]
      edges$tiepe[curedge] <- tbl$op[i]
    } else if (tbl$op[i] == "~1") {
      #### ~1 : intercept ####
      # lhs node
      jl <- match(create_node_id(tbl$block[i], tbl$lhs[i]),
                  nodes$id, nomatch = 0L)
      nodetype <- "ov"
      if (length(unique(tbl$block[tbl$rhs == tbl$lhs[i] |
        tbl$lhs == tbl$lhs[i]])) > 1L) {
        nodetype <- switch(tbl$block[i],
          "wov",
          "bov"
        )
      }
      if (jl == 0L) {
        curnode <- curnode + 1L
        jl <- curnode
        nodes$naam[curnode] <- tbl$lhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jl] == "") nodes$tiepe[jl] <- nodetype
      }
      # rhs node
      jr <- 0L
      curnode <- curnode + 1L
      jr <- curnode
      nodes$naam[curnode] <- paste0("1van", tbl$lhs[i])
      nodes$tiepe[curnode] <- "const"
      nodes$blok[curnode] <- tbl$block[i]
      nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                          nodes$naam[curnode])
      # edge
      curedge <- curedge + 1L
      edges$id[curedge] <- curedge
      edges$label[curedge] <- edge_label(tbl$label[i], tbl$fixed[i])
      edges$van[curedge] <- nodes$id[jr]
      edges$naar[curedge] <- nodes$id[jl]
      edges$tiepe[curedge] <- "~"
    } else if (tbl$op[i] == "~~") {
      #### ~~ : is correlated with ####
      # lhs node
      jl <- match(create_node_id(tbl$block[i], tbl$lhs[i]),
                  nodes$id, nomatch = 0L)
      nodetype <- "ov"
      if (length(unique(tbl$block[tbl$rhs == tbl$lhs[i] |
        tbl$lhs == tbl$lhs[i]])) > 1L) {
        nodetype <- switch(tbl$block[i],
          "wov",
          "bov"
        )
      }
      if (jl == 0L) {
        curnode <- curnode + 1L
        jl <- curnode
        nodes$naam[curnode] <- tbl$lhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jl] == "") nodes$tiepe[jl] <- nodetype
      }
      # rhs node
      jr <- match(create_node_id(tbl$block[i], tbl$rhs[i]),
                  nodes$id, nomatch = 0L)
      nodetype <- "ov"
      if (length(unique(tbl$block[tbl$rhs == tbl$lhs[i] |
        tbl$lhs == tbl$lhs[i]])) > 1L) {
        nodetype <- switch(tbl$block[i],
          "wov",
          "bov"
        )
      }
      if (jr == 0L) {
        curnode <- curnode + 1L
        jr <- curnode
        nodes$naam[curnode] <- tbl$rhs[i]
        nodes$tiepe[curnode] <- nodetype
        nodes$blok[curnode] <- tbl$block[i]
        nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                            nodes$naam[curnode])
      } else {
        if (nodes$tiepe[jr] == "") nodes$tiepe[jr] <- nodetype
      }
      # edge
      curedge <- curedge + 1L
      edges$id[curedge] <- curedge
      edges$label[curedge] <- edge_label(tbl$label[i], tbl$fixed[i])
      if (varlv && jl == jr) { # prepare for handling varlv
        edges$label[curedge] <- paste(tbl$label[i],
          if (is.null(tbl$fixed)) "" else tbl$fixed[i],
          sep = "=")
      }
      edges$van[curedge] <- nodes$id[jr]
      edges$naar[curedge] <- nodes$id[jl]
      edges$tiepe[curedge] <- ifelse(jl == jr, "~~~", tbl$op[i])
    }
  }
  # aanpassingen voor varlv
  if (varlv && any(edges$tiepe == "~~~")) {
    #       wijzigen varianties
    welke <- which(edges$tiepe == "~~~")
    varlvnodes <- rep(0L, length(welke))
    lvnodes <- rep(0L, length(welke))
    for (ji in seq_along(welke)) {
      j <- welke[ji]
      curnode <- curnode + 1L
      nodes$naam[curnode] <- gsub("=.*$", "", edges$label[j])
      if (nodes$naam[curnode] == "")
        nodes$naam[curnode] <- paste0("var", edges$van[j])
      nodes$tiepe[curnode] <- "varlv"
      nodes$blok[curnode] <- nodes$blok[which(nodes$id == edges$van[j])[[1L]]]
      nodes$id[curnode] <- create_node_id(nodes$blok[curnode],
                                          nodes$naam[curnode])
      varlvnodes[ji] <- curnode
      lvnodes[ji] <- edges$van[j]
      edges$van[j] <- nodes$id[curnode]
      edges$tiepe[j] <- "~."
      edges$label[j] <- ifelse(grepl("=", edges$label[j], fixed = TRUE),
        gsub("^.*=", "", edges$label[j]), ""
      )
    }
    #       wijzigen covarianties
    for (j1 in seq_along(welke)) {
      for (j2 in seq_along(welke)) {
        if (j1 != j2 && any(edges$tiepe == "~~" &
          edges$van == lvnodes[j1] & edges$naar == lvnodes[j2])) {
          edg <- which(edges$tiepe == "~~" &
            edges$van == lvnodes[j1] & edges$naar == lvnodes[j2])[[1L]]
          edges$van[edg] <- varlvnodes[j1]
          edges$naar[edg] <- varlvnodes[j2]
        }
      }
    }
  }
  nodes <- nodes[seq.int(curnode), ]
  edges <- edges[seq.int(curedge), ]
  edges$label <- trimws(edges$label)
  if (any(grepl("[<>&]", c(edges$label, nodes$naam)))) {
    lav_msg_warn(gettext(
      "some labels contain '<', '>' or '&', which can result in errors!"
    ))
  }
  list(nodes = nodes, edges = edges)
}

# utility functions for the sam() function
# YR 4 April 2023

# clamped small-sample correction factor alpha/(n - 1), forced into [0, 1]
lav_sam_alpha_n1 <- function(alpha_correction = 0L, n = 20L) {
  alpha_n1 <- alpha_correction / (n - 1)
  if (!is.finite(alpha_n1)) {
    alpha_n1 <- 0
  }
  min(max(alpha_n1, 0), 1)
}

# row/column index pairs and labels for the elements of (eta* %x% eta*),
# where lv_names already includes the leading "..int.." intercept entry;
# the first nfac labels are the plain (first-order) lv names
lav_sam_lvnames2 <- function(lv_names) {
  nfac <- length(lv_names)
  idx1 <- rep(seq_len(nfac), each = nfac)
  idx2 <- rep(seq_len(nfac), times = nfac)
  names2 <- paste(lv_names[idx1], lv_names[idx2], sep = ":")
  names2[seq_len(nfac)] <- lv_names
  list(idx1 = idx1, idx2 = idx2, names = names2)
}

# is this a sam object created with sam.method = "local", "fsr" or "cfsr"?
lav_sam_local_flag <- function(object) {
  sam_method <- object@internal$sam.method
  !is.null(sam_method) && sam_method %in% c("local", "fsr", "cfsr")
}

# the stored step-2 structural fit of a local SAM object (with the two-step
# corrected test/baseline injected, if available); NULL for global sam
# objects, and for sam objects created by older versions of lavaan
lav_sam_struc_object <- function(object) {
  object@internal$sam.struc.fit.object
}

# construct 'mapping matrix' M using either "ML", "GLS" or "ULS" method
#
# by construction, M %*% LAMBDA = I (the identity matrix)
lav_sam_mapping_mat <- function(mm_lambda = NULL, mm_theta = NULL,
                                   s = NULL, s_inv = NULL,
                                   method = "ML") {

  method <- toupper(method)

  # catch empty columns in LAMBDA (eg higher order!)
  lambda_orig <- mm_lambda
  empty_idx <- which(apply(mm_lambda, 2L, function(x) all(x == 0)))
  if (length(empty_idx) > 0L) {
    mm_lambda <- lambda_orig[, -empty_idx, drop = FALSE]
  }

  # ULS
  # M == solve( t(LAMBDA) %*% LAMBDA ) %*% t(LAMBDA)
  #   == MASS:::ginv(LAMBDA)
  if (method == "ULS") {
    m <- try(tcrossprod(solve(crossprod(mm_lambda)), mm_lambda),
      silent = TRUE
    )
    if (inherits(m, "try-error")) {
      lav_msg_warn(gettext(
        "cannot invert crossprod(LAMBDA); using generalized inverse"))
      m <- MASS::ginv(mm_lambda)
    }


    # GLS
    # M == solve( t(LAMBDA) %*% S.inv %*% LAMBDA ) %*% t(LAMBDA) %*% S.inv
  } else if (method == "GLS") {
    if (is.null(s_inv)) {
      s_inv <- try(solve(s), silent = TRUE)
    }
    if (inherits(s_inv, "try-error")) {
      lav_msg_warn(gettext("S is not invertible; switching to ULS method"))
      m <- lav_sam_mapping_mat(mm_lambda = mm_lambda, method = "ULS")
    } else {
      t_lsinv <- t(mm_lambda) %*% s_inv
      t_lsinv_l <- t_lsinv %*% mm_lambda
      m <- try(solve(t_lsinv_l, t_lsinv), silent = TRUE)
      if (inherits(m, "try-error")) {
        lav_msg_warn(gettext("problem constructing mapping matrix;
                              switching to generalized inverse"))
        m <- MASS::ginv(t_lsinv_l) %*% t_lsinv
      }
    }

    # ML
    # M == solve(t(LAMBDA) %*% THETA.inv %*% LAMBDA) %*% t(LAMBDA) %*% THETA.inv
  } else if (method == "ML") {
    # Problem: if THETA has zero elements on the diagonal, we cannot invert

    # As we do not have access to Sigma(.inv), we cannot
    # use the trick as in lavPredict() (where we replace THETA.inv by
    # Sigma.inv)

    # old method (<0.6-16): remove rows/cols with zero values
    # on the diagonal of THETA, invert the submatrix and
    # set the '0' diagonal elements to one. This resulted in (somewhat)
    # distorted results.

    # new in 0.6-16: we use the Wall & Amemiya (2000) method using
    # the so-called 'T' transformation

    # new in 0.6-18: if we cannot use the marker method (eg growth models),
    # use the 'old' method anyway: remove zero rows/cols and invert submatrix

    zero_theta_idx <- which(abs(diag(mm_theta)) < 1e-4) # be conservative
    if (length(zero_theta_idx) == 0L) {
      # ok, no zero diagonal elements: try to invert THETA
      if (lav_mat_is_diagonal(mm_theta)) {
        theta_inv <- diag(1 / diag(mm_theta), nrow = nrow(mm_theta))
      } else {
        theta_inv <- try(solve(mm_theta), silent = TRUE)
        if (inherits(theta_inv, "try-error")) {
          theta_inv <- NULL
        }
      }
    } else {
      # see if we can use marker method
      marker_idx <- lav_utils_get_marker(mm_lambda = mm_lambda, std_lv = TRUE)
      if (any(is.na(marker_idx))) {
        theta_inv <- try(lav_mat_sym_inverse(mm_theta), silent = TRUE)
        if (inherits(theta_inv, "try-error")) {
          theta_inv <- NULL
        } else {
          diag(theta_inv)[zero_theta_idx] <- 1
        }
      } else {
        # try tmat method later on
        theta_inv <- NULL
      }
    }

    # could we invert THETA?
    if (!is.null(theta_inv)) {
      # ha, all is good; compute M the usual way
      t_lti <- t(mm_lambda) %*% theta_inv
      t_lti_l <- t_lti %*% mm_lambda
      m <- try(solve(t_lti_l, t_lti), silent = TRUE)
      if (inherits(m, "try-error")) {
        lav_msg_warn(gettext(
          "problem constructing ML mapping matrix; switching to ULS"))
        m <- lav_sam_mapping_mat(mm_lambda = mm_lambda, method = "ULS")
      }
    } else {
      # use W&A2000's method using the 'T' transformation
      m <- try(lav_sam_mapping_mat_tmat(
        mm_lambda = mm_lambda,
        mm_theta = mm_theta
      ), silent = TRUE)
      if (inherits(m, "try-error")) {
        lav_msg_warn(gettext(
          "problem constructing ML mapping matrix; switching to ULS"))
        m <- lav_sam_mapping_mat(mm_lambda = mm_lambda, method = "ULS")
      }
    }
  } # ML

  # empty.idx?
  if (length(empty_idx) > 0L) {
    m_full <- m
    m <- matrix(0, nrow = ncol(lambda_orig), ncol = ncol(m_full))
    m[-empty_idx, ] <- m_full
  }

  m
}

# use 'T' transformation to create the 'Bartlett/ML' mapping matrix
# see Wall & Amemiya (2000) eq (7)
# see also Fuller 1987 page 357 (where T is called H), and page 364

# although Fuller/W&A always assumed that THETA is diagonal,
# their method seems to work equally well for non-diagonal THETA
#
# in our implementation:
# - we do NOT reorder the rows of LAMBDA
# - if std.lv = TRUE, we first rescale to 'create' marker indicators
#   and then rescale back at the end
#
lav_sam_mapping_mat_tmat <- function(mm_lambda = NULL,
                                        mm_theta = NULL,
                                        marker_idx = NULL,
                                        std_lv = NULL) {

  mm_lambda <- as.matrix.default(mm_lambda)

  # catch empty columns in LAMBDA (eg higher order!)
  lambda_orig <- mm_lambda
  empty_idx <- which(apply(mm_lambda, 2L, function(x) all(x == 0)))
  if (length(empty_idx) > 0L) {
    mm_lambda <- lambda_orig[, -empty_idx, drop = FALSE]
  }

  # nvar <- nrow(mm_lambda)
  # nfac <- ncol(mm_lambda)

  # do we have marker.idx?
  if (is.null(marker_idx)) {
    # 'marker' indicator has a single non-zero element in a row
    marker_idx <- lav_utils_get_marker(mm_lambda = mm_lambda, std_lv = TRUE)
    if (any(is.na(marker_idx))) {
      lav_msg_stop(gettext("no clear markers in LAMBDA matrix"))
    }
  }

  # std.lv TRUE or FALSE?
  if (is.null(std_lv)) {
    std_lv <- FALSE
    if (any(diag(mm_lambda[marker_idx, , drop = FALSE]) != 1)) {
      std_lv <- TRUE
    }
  }

  # if std.lv = TRUE, rescale
  if (std_lv) {
    marker <- mm_lambda[marker_idx, , drop = FALSE]
    marker_inv <- 1 / diag(marker)
    mm_lambda <- t(t(mm_lambda) * marker_inv)
  }

  # compute 'T' matrix
  tmat <- lav_sam_tmat(
    mm_lambda = mm_lambda, mm_theta = mm_theta,
    marker_idx = marker_idx
  )

  # ML mapping matrix
  m <- tmat[marker_idx, , drop = FALSE]

  if (std_lv) {
    m <- m * marker_inv
  }

  # empty.idx?
  if (length(empty_idx) > 0L) {
    m_full <- m
    m <- matrix(0, nrow = ncol(lambda_orig), ncol = ncol(m_full))
    m[-empty_idx, ] <- m_full
  }

  m
}

# create 'T' matrix (tmat) for T-transformation
#
# Notes: - here we assume that LAMBDA has unity markers (no std.lv = TRUE)
#        - TMAT is NOT symmetric!
#        - Yc %*% t(TMAT) transforms the data in such a way that we get:
#           1)  Bartlett factor scores in the marker columns
#           2) 'V' values in the non-marker columns, where:
#               V = Yc - Yc[,marker.idx] %*% t(LAMBDA)
#
lav_sam_tmat <- function(mm_lambda = NULL,
                         mm_theta = NULL,
                         marker_idx = NULL) {
  mm_lambda <- as.matrix.default(mm_lambda)
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)

  # do we have marker.idx?
  if (is.null(marker_idx)) {
    # 'marker' indicator has a single 1 element in a row
    marker_idx <- lav_utils_get_marker(mm_lambda = mm_lambda, std_lv = FALSE)
    if (any(is.na(marker_idx))) {
      lav_msg_stop(gettext("no clear markers in LAMBDA matrix"))
    }
  }

  # construct 'm_c' matrix
  c2 <- diag(nvar)
  c2[, marker_idx] <- -1 * mm_lambda
  m_c <- c2[-marker_idx, , drop = FALSE]

  # compute Sigma.ve and Sigma.vv
  sigma_ve <- m_c %*% mm_theta
  # Sigma.vv <- m_c %*% THETA %*% t(m_c)
  sigma_vv <- sigma_ve %*% t(m_c)

  # construct 'Gamma' (=gamma_1) (and gamma2) matrix
  # Gamma <- (t(Sigma.ve) %*% solve(Sigma.vv))[marker.idx,, drop = FALSE]
  gamma_1 <- try(t(solve(sigma_vv, sigma_ve)[, marker_idx, drop = FALSE]),
    silent = TRUE
  )
  if (inherits(gamma_1, "try-error")) {
    tmp <- t(sigma_ve) %*% MASS::ginv(sigma_vv)
    gamma_1 <- tmp[marker_idx, , drop = FALSE]
  }
  gamma2 <- matrix(0, nfac, nvar)
  gamma2[, -marker_idx] <- gamma_1
  gamma2[, marker_idx] <- diag(nfac)

  # transformation matrix 'T' (we call it here 'Tmat')
  tmat <- matrix(0, nvar, nvar)
  tmat[-marker_idx, ] <- m_c
  tmat[marker_idx, ] <- -gamma2 %*% c2

  tmat
}


# compute VETA
# - if alpha.correction == 0     -> same as local SAM (or MOC)
# - if alpha.correction == (N-1) -> same as FSR+Bartlett
lav_sam_veta <- function(m = NULL, s = NULL, mm_theta = NULL,
                         alpha_correction = 0L, lambda_correction = TRUE,
                         n = 20L, dummy_lv_idx = integer(0L), extra = FALSE) {

  # catch empty rows in M (higher-order?)
  m_orig <- m
  empty_idx <- which(apply(m_orig, 1L, function(x) all(x == 0)))
  if (length(empty_idx) > 0L) {
    m <- m_orig[-empty_idx, , drop = FALSE]
  }

  # MSM
  msm <- m %*% s %*% t(m)

  # MTM
  mtm <- m %*% mm_theta %*% t(m)

  # note: MTM no longer needs to be regularized (force-pd) here:
  # lav_mat_sym_diff_smallest_root() handles a singular (or even zero)
  # MTM natively, by regressing out its null space (which includes the
  # rows/columns of any dummy lvs, and any empty theta elements)
  if (all(abs(mtm) < sqrt(.Machine$double.eps))) {
    # all zero? certainly no need for alpha or lambda.correction
    alpha_correction <- 0L
    lambda_correction <- FALSE
  }

  # apply small sample correction (if requested)
  alpha <- alpha_correction
  if (alpha_correction > 0) {
    alpha_n1 <- lav_sam_alpha_n1(alpha_correction, n)
    mtm <- (1 - alpha_n1) * mtm
  }

  lambda <- lambda_star <- +Inf
  if (lambda_correction) {
    # use Fuller (1987) approach to ensure VETA is positive
    lambda <- try(lav_mat_sym_diff_smallest_root(msm, mtm),
      silent = TRUE
    )
    if (inherits(lambda, "try-error")) {
      lav_msg_warn(gettext("failed to compute lambda"))
      veta <- msm - mtm # and hope for the best
    } else {
      cutoff <- 1 + 1 / (n - 1)
      if (lambda < cutoff) {
        lambda_star <- lambda - 1 / (n - 1)
        veta <- msm - lambda_star * mtm
      } else {
        veta <- msm - mtm
      }
    }
  } else {
    veta <- msm - mtm
  }

  # empty.idx?
  if (length(empty_idx) > 0L) {
    msm_full <- msm
    mtm_full <- mtm
    veta_full <- veta
    nfac_orig <- nrow(m_orig)

    msm <- mtm <- veta <- matrix(0, nrow = nfac_orig, ncol = nfac_orig)
    msm[-empty_idx, -empty_idx] <- msm_full
    mtm[-empty_idx, -empty_idx] <- mtm_full
    veta[-empty_idx, -empty_idx] <- veta_full
  }

  # extra attributes?
  if (extra) {
    attr(veta, "lambda") <- lambda
    attr(veta, "alpha") <- alpha
    attr(veta, "lambda.star") <- lambda_star
    attr(veta, "MSM") <- msm
    attr(veta, "MTM") <- mtm
  }

  veta
}

# compute EETA = E(Eta) = M %*% [YBAR - NU]
lav_sam_eeta <- function(m = NULL, ybar = NULL, mm_nu = NULL) {
  m %*% (ybar - mm_nu)
}

# compute veta including quadratic/interaction terms
lav_sam_veta2 <- function(fs = NULL, m = NULL,
                          veta = NULL, eeta = NULL, mm_theta = NULL,
                          lv_names = NULL,
                          lv_int_names = NULL,
                          dummy_lv_names = character(0L),
                          alpha_correction = 0L,
                          lambda_correction = TRUE,
                          lambda1 = 1,
                          fs_outlier_idx = integer(0L),
                          return_fs = FALSE,
                          return_cov_iveta2 = TRUE,
                          extra = FALSE) {

  # small utility function: var() divided by N
  varn <- function(x, n) {
    var(x, use = "pairwise.complete.obs") * (n - 1) / n
  }

  if (length(lv_int_names) == 0L) {
    lav_msg_stop(gettext("lv.int.names is empty: no lv quadratic/interaction
                         terms are provided"))
  }

  if (is.null(lv_names)) {
    lv_names <- paste("eta", seq_len(ncol(fs)), sep = "")
  }

  # MTM
  mtm <- m %*% mm_theta %*% t(m)

  # note: MTM no longer needs to be regularized (force-pd) here:
  # lav_mat_sym_diff_smallest_root() handles a singular var.error matrix
  # natively, by regressing out its null space (which includes the
  # rows/columns of any dummy lvs); this also keeps MTM identical to the
  # (raw) version used in lav_sam_gamma_add()

  # augment to include intercept
  fs <- cbind(1, fs)
  n <- nrow(fs)
  mtm <- lav_mat_bdiag(0, mtm)
  veta <- lav_mat_bdiag(0, veta)
  eeta <- c(1, eeta)
  lv_names <- c("..int..", lv_names)
  nfac <- ncol(fs)

  tmp <- lav_sam_lvnames2(lv_names)
  idx1 <- tmp$idx1
  idx2 <- tmp$idx2
  names_1 <- tmp$names

  fs2 <- fs[, idx1, drop = FALSE] * fs[, idx2, drop = FALSE]

  k_nfac <- lav_mat_com(nfac, nfac)
  ik <- diag(nfac * nfac) + k_nfac

  eeta <- as.matrix(drop(eeta))
  vetak_mtm <- veta %x% mtm

  # normal version (for now):
  gamma_me22 <- ik %*% (mtm %x% mtm)

  # ingredients (normal ME case)
  var_fs2 <- varn(fs2, n)
  var_etak_me <- (tcrossprod(eeta) %x% mtm + vetak_mtm)
  var_mek_eta <- lav_mat_com_pre_post(var_etak_me)
  var_me2 <- gamma_me22

  cov_etak_me_mek_eta <- lav_mat_com_post(var_etak_me)
  cov_mek_eta_etak_me <- t(cov_etak_me_mek_eta)

  var_error <- (var_etak_me + var_mek_eta + cov_etak_me_mek_eta
    + cov_mek_eta_etak_me + var_me2)

  # select only what we need
  colnames(var_fs2) <- rownames(var_fs2) <- names_1
  colnames(var_error) <- rownames(var_error) <- names_1
  colnames(fs2) <- names_1
  lv_keep <- c(lv_names[-1], lv_int_names)
  var_fs2 <- var_fs2[lv_keep, lv_keep, drop = FALSE]
  var_error <- var_error[lv_keep, lv_keep, drop = FALSE]
  fs_mean <- colMeans(fs2, na.rm = TRUE)
  names(fs_mean) <- names_1
  fs2_mean <- fs_mean # all of them
  fs_mean <- fs_mean[lv_keep]

  # apply small sample correction (if requested)
  alpha <- alpha_correction
  alpha_n1 <- 0
  if (alpha_correction > 0) {
    alpha_n1 <- lav_sam_alpha_n1(alpha_correction, n)
    var_error <- (1 - alpha_n1) * var_error
  }

  lambda <- +Inf
  lambda_star <- 1
  if (lambda_correction) {
    # use Fuller (1987) approach to ensure VETA2 is positive
    lambda <- try(lav_mat_sym_diff_smallest_root(
      var_fs2,
      var_error
    ), silent = TRUE)
    if (inherits(lambda, "try-error")) {
      lav_msg_warn(gettext("failed to compute lambda"))
      veta2 <- var_fs2 - var_error # and hope for the best
    } else {
      cutoff <- 1 + 2 / n # be more conservative for VETA2
      if (lambda < cutoff) {
        lambda_star <- max(c(0, lambda - ncol(var_fs2) / (n - 1)))
        veta2 <- var_fs2 - lambda_star * var_error
      } else {
        veta2 <- var_fs2 - var_error
      }
    }
  } else {
    veta2 <- var_fs2 - var_error
  }

  # new in 0.6-20: to compute Gamma.eta
  # the casewise contributions are defined so that their average over the
  # observations equals *exactly* the augmented summary statistics
  # (EETA2, VETA2) that are used in the second step; this requires the
  # effective *first-order* lambda1 inside the decomposition (because
  # EETA2 and var.error are built from the lambda1-corrected first-order
  # VETA), while the *second-order* lambda.star multiplies the whole
  # correction term:
  #   ieeta2_i = fs2[i, ] - lambda1 * vec(MTM)
  #   (the terms involving the mean factor scores cancel out exactly)
  # and
  #   iveta2_i = tcrossprod(fs2[i, ] - fs2.mean) - lambda2.eff * tmp_i
  # where lambda2.eff = lambda.star * (1 - alpha/(n-1)) is the effective
  # multiplier of the *unscaled* var.error (the alpha correction rescales
  # var.error before the lambda step), and
  #       tmp_i = ((F_i - lambda1 * MTM) %x% MTM) +
  #               (MTM %x% (F_i - lambda1 * MTM)) +
  #               lav_mat_com_post((F_i - lambda1 * MTM) %x% MTM) +
  #               lav_mat_com_pre((F_i - lambda1 * MTM) %x% MTM) +
  #               (I + K) %*% (MTM %x% MTM)
  # with F_i = tcrossprod(f_i); both selected in the rows/columns we keep
  # note: cov(iveta2_1) is invariant to the value of lambda1 (the lambda1
  # terms are constant across the observations); but lambda1 does matter
  # for the jacobian of the average casewise contribution computed in
  # lav_sam_gamma_add()
  # rewritten 12 June 2026 to avoid the loop over the observations: every
  # kept element is a linear function of the second-order factor scores
  # fs2[i, ]; elementwise, for kept elements a = (r1, s1) and b = (r2, s2)
  # of (eta* %x% eta*):
  #   tmp_i[a, b] = F_i[r1, r2] MTM[s1, s2] + MTM[r1, r2] F_i[s1, s2] +
  #                 F_i[r1, s2] MTM[s1, r2] + F_i[s1, r2] MTM[r1, s2] +
  #                 (1 - 2 * lambda1) * (MTM[r1, r2] MTM[s1, s2] +
  #                                      MTM[s1, r2] MTM[r1, s2])
  # where F_i[r, s] = fs2[i, (r - 1) * nfac + s]
  if (return_cov_iveta2) {
    lambda2_eff <- lambda_star * (1 - alpha_n1)
    keep_idx <- match(lv_keep, names_1)
    nkeep <- length(keep_idx)
    # row/column pairs of the kept elements of (eta* %x% eta*)
    i1k <- idx1[keep_idx]
    i2k <- idx2[keep_idx]

    # casewise contributions to E(eta* %x% eta*)[keep]
    part1 <- t(t(fs2[, keep_idx, drop = FALSE]) -
               lambda1 * lav_mat_vec(mtm)[keep_idx])

    # casewise contributions to vech(Var(eta* %x% eta*)[keep, keep])
    # vech row (pa) and column (pb) indices over the kept elements
    tmp_mat <- matrix(seq_len(nkeep), nkeep, nkeep)
    pa <- tmp_mat[lower.tri(tmp_mat, diag = TRUE)]
    pb <- t(tmp_mat)[lower.tri(tmp_mat, diag = TRUE)]
    r1 <- i1k[pa]; s1 <- i2k[pa]
    r2 <- i1k[pb]; s2 <- i2k[pb]

    # fs2 columns holding F_i[r1, r2], F_i[s1, s2], F_i[r1, s2], F_i[s1, r2]
    col_11 <- (r1 - 1L) * nfac + r2
    col_22 <- (s1 - 1L) * nfac + s2
    col_12 <- (r1 - 1L) * nfac + s2
    col_21 <- (s1 - 1L) * nfac + r2

    b_11 <- mtm[cbind(r1, r2)]
    b_22 <- mtm[cbind(s1, s2)]
    b_12 <- mtm[cbind(r1, s2)]
    b_21 <- mtm[cbind(s1, r2)]

    fs2kc <- t(t(fs2[, keep_idx, drop = FALSE]) - fs2_mean[keep_idx])
    part2 <- (fs2kc[, pa, drop = FALSE] * fs2kc[, pb, drop = FALSE] -
              lambda2_eff *
              (t(t(fs2[, col_11, drop = FALSE]) * b_22) +
               t(t(fs2[, col_22, drop = FALSE]) * b_11) +
               t(t(fs2[, col_12, drop = FALSE]) * b_21) +
               t(t(fs2[, col_21, drop = FALSE]) * b_12) +
               rep((1 - 2 * lambda1) * (b_11 * b_22 + b_21 * b_12),
                   each = n)))

    iveta2_1 <- cbind(part1, part2)
    cov_iveta2 <- cov(iveta2_1) * (n - 1) / n
  }

  # extra attributes?
  if (extra) {
    attr(veta2, "lambda") <- lambda
    attr(veta2, "alpha") <- alpha
    attr(veta2, "lambda.star") <- lambda_star
    attr(veta2, "MSM") <- var_fs2
    attr(veta2, "MTM") <- var_error
    attr(veta2, "FS.mean") <- fs_mean
  }
  if (return_fs) {
    attr(veta2, "FS") <- fs2[, lv_keep, drop = FALSE]
  }
  if (return_cov_iveta2) {
    attr(veta2, "cov.iveta2") <- cov_iveta2
  }

  veta2
}

lav_sam_eeta2 <- function(eeta = NULL, veta = NULL, lv_names = NULL,
                          lv_int_names = NULL) {
  if (length(lv_int_names) == 0L) {
    lav_msg_stop(gettext("lv.int.names is empty: no lv quadratic/interaction
                         terms are provided"))
  }

  if (is.null(lv_names)) {
    lv_names <- paste("eta", seq_len(ncol(veta)), sep = "")
  }

  nfac <- nrow(veta)
  idx1 <- rep(seq_len(nfac), each = nfac)
  idx2 <- rep(seq_len(nfac), times = nfac)
  names_1 <- c(lv_names, paste(lv_names[idx1], lv_names[idx2], sep = ":"))

  # E(\eta %x% \eta)
  eeta2 <- lav_mat_vec(veta) + eeta %x% eeta

  # add 1st order
  eeta2_aug <- c(eeta, eeta2)

  # select only what we need
  names(eeta2_aug) <- names_1
  lv_keep <- c(lv_names, lv_int_names)
  eeta2_aug <- eeta2_aug[lv_keep]

  eeta2_aug
}

# compute var(fs2) including quadratic/interaction terms
lav_sam_fs2 <- function(fs = NULL, lv_names = NULL, lv_int_names = NULL) {
  varn <- function(x, n) {
    var(x) * (n - 1) / n
  }

  if (length(lv_int_names) == 0L) {
    lav_msg_stop(gettext("lv.int.names is empty: no lv quadratic/interaction
                         terms are provided"))
  }

  if (is.null(lv_names)) {
    lv_names <- paste("eta", seq_len(ncol(fs)), sep = "")
  }

  # augment to include intercept
  fs <- cbind(1, fs)
  n <- nrow(fs)
  lv_names <- c("..int..", lv_names)
  nfac <- ncol(fs)

  idx1 <- rep(seq_len(nfac), each = nfac)
  idx2 <- rep(seq_len(nfac), times = nfac)

  names_1 <- paste(lv_names[idx1], lv_names[idx2], sep = ":")

  fs2 <- fs[, idx1, drop = FALSE] * fs[, idx2, drop = FALSE]
  var_fs2 <- varn(fs2, n)

  # select only what we need
  colnames(var_fs2) <- rownames(var_fs2) <- names_1
  lv_main <- paste(lv_names[-1], "..int..", sep = ":")
  lv_keep <- c(lv_main, lv_int_names)
  var_fs2 <- var_fs2[lv_keep, lv_keep, drop = FALSE]

  var_fs2
}

# create consistent lavaan object, based on (filled in) PT
lav_sam_step3_joint <- function(fit = NULL, pt_1 = NULL, sam_method = "local") {
  lavoptions <- fit@Options

  lavoptions_joint <- lavoptions
  lavoptions_joint$optim.method <- "none"
  lavoptions_joint$optim.parscale <- "none"
  lavoptions_joint$start <- "default"
  lavoptions_joint$optim.force.converged <- TRUE
  lavoptions_joint$check.gradient <- FALSE
  lavoptions_joint$check.start <- FALSE
  lavoptions_joint$check.post <- FALSE
  lavoptions_joint$rotation <- "none"
  lavoptions_joint$se <- "none"
  lavoptions_joint$store.vcov <- FALSE # we do this manually

  if (sam_method %in% c("local", "fsr", "cfsr")) {
    lavoptions_joint$baseline <- FALSE
    lavoptions_joint$fit.by.level <- FALSE
    lavoptions_joint$sample.icov <- FALSE
    #lavoptions.joint$h1 <- TRUE # we need this if we re-use the sam object
    lavoptions_joint$test <- "none"
    lavoptions_joint$estimator <- "none"
  } else {
    # global: the joint computes the core (base) test(s); "yuan.chan" is not a
    # core lavaan test (it is added afterwards by lav_sam_global_test()), so we
    # strip it here and make sure the unscaled "standard" base is present.
    joint_test <- lavoptions$test[lavoptions$test != "yuan.chan"]
    if (length(joint_test) == 0L || !"standard" %in% joint_test) {
      joint_test <- unique(c("standard", joint_test))
    }
    lavoptions_joint$test <- joint_test
    lavoptions_joint$estimator <- lavoptions$estimator
  }

  # set ustart values
  pt_1$ustart <- pt_1$est # as this is used if optim.method == "none"

  # the step-0 dummy fit of the local methods skips the inverse of S
  # (sample.icov = FALSE: the inverse may not exist when N < nvar), but the
  # joint-model computations of some estimators need S^-1 (eg the GLS
  # information/objective). Backfill it here when it is missing; if S cannot
  # be inverted it simply stays missing, exactly as before.
  lavsamplestats_joint <- fit@SampleStats
  if (fit@Data@nlevels == 1L && !fit@Model@categorical &&
      !lavsamplestats_joint@missing.flag) { # core lavaan skips icov for
                                            # missing = "ml" on purpose
    for (g in seq_len(lavsamplestats_joint@ngroups)) {
      if (is.null(lavsamplestats_joint@icov[[g]]) &&
          !is.null(lavsamplestats_joint@cov[[g]])) {
        # quietly: the inverse is optional here (only some estimators need
        # it), so a non-PD S must not surface a warning at this point
        current_warn <- lav_warn()
        lav_warn(FALSE)
        out <- try(lav_samp_icov(
          cov_1 = lavsamplestats_joint@cov[[g]], ridge = 1e-05,
          x_idx = lavsamplestats_joint@x.idx[[g]],
          ngroups = fit@Data@ngroups, g = g
        ), silent = TRUE)
        lav_warn(current_warn)
        if (!inherits(out, "try-error")) {
          lavsamplestats_joint@icov[[g]] <- out$icov
          lavsamplestats_joint@cov.log.det[[g]] <- out$cov.log.det
        }
      }
    }
  }

  joint <- lavaan::lavaan(pt_1,
    slot_options = lavoptions_joint,
    slot_sample_stats = lavsamplestats_joint,
    slot_data = fit@Data,
    verbose = FALSE
  )
  joint
}

# the (two-step robust) STRUCTURAL fit object of a local SAM model.
#
# The structural model FIT.PA is fitted with se = "robust.sem" using the
# step-1 induced NACOV (Gamma.eta) of the latent (co)variances/means. Its
# Satorra-Bentler scaled chi-square and the robust fit measures derived from it
# are the correct 'two-step robust' fit measures.
#
# For a single group, and for multiple groups WITHOUT across-group constraints
# in the measurement model (Case A), the per-group Gamma.eta list passed to
# FIT.PA already yields the correct robust test, and we simply reproduce it.
# For multiple groups WITH across-group measurement constraints (Case B), the
# step-1 variability couples the groups, so the relevant Gamma has nonzero
# cross-group blocks (step1$Gamma.eta.full) that a per-group NACOV list cannot
# represent. We then recompute the Satorra-Bentler scaled chi-square ourselves
# -- for the model AND the baseline -- using that full Gamma.
#
# The corrected FIT.PA is stored in the @internal slot of the final sam
# object (element sam.struc.fit.object), so that fitMeasures(), lavResiduals()
# and modindices() can delegate to it (issue #517).
lav_sam_struc_fit_object <- function(fit_pa = NULL, step1 = NULL) {
  # do we have a (local) Gamma.eta and a non-saturated structural model?
  have_gamma <- !is.null(step1$Gamma.eta) &&
    length(step1$Gamma.eta) > 0L && !is.null(step1$Gamma.eta[[1]])
  df_struc <- if (length(fit_pa@test) > 0L) fit_pa@test[[1]]$df else 0L
  has_baseline <- !is.null(fit_pa@baseline) &&
    !is.null(fit_pa@baseline$partable)

  if (!have_gamma || df_struc < 1L || !has_baseline) {
    return(fit_pa)
  }

  out <- try(
    {
      fp <- fit_pa
      ng <- fp@Data@ngroups
      ntot <- fp@SampleStats@ntotal
      fg <- unlist(fp@SampleStats@nobs) / ntot
      meanstr <- fp@Model@meanstructure

      # full Gamma of the stacked structural statistics (cross-group blocks
      # for Case B; block-diagonal otherwise)
      if (isTRUE(step1$caseB)) {
        gamma_all <- ntot * step1$Gamma.eta.full
      } else {
        gamma_all <- lav_mat_bdiag(lapply(
          seq_len(ng), function(g) step1$Gamma.eta[[g]] / fg[g]
        ))
      }

      # model: Satorra-Bentler scaled test using the full Gamma
      test_model <- lav_test_sb(
        lavobject = fp, test = "satorra.bentler",
        gamma_full = gamma_all
      )

      # baseline: refit the independence model on the latent (co)variances,
      # then apply the same Satorra-Bentler correction with the full Gamma
      fit_base <- lavaan::lavaan(
        model = fp@baseline$partable,
        sample.cov = step1$VETA,
        sample.mean = if (meanstr) step1$EETA else NULL,
        sample.nobs = as.list(unlist(fp@SampleStats@nobs)),
        nacov = step1$Gamma.eta,
        slot_options = fp@Options
      )
      test_base <- lav_test_sb(
        lavobject = fit_base, test = "satorra.bentler",
        gamma_full = gamma_all
      )

      # inject the corrected scaled tests
      fp@test[["satorra.bentler"]] <- test_model[["satorra.bentler"]]
      base_idx <- which(vapply(fp@baseline$test,
        function(x) x$test, character(1L)) == "satorra.bentler")
      if (length(base_idx) > 0L) {
        fp@baseline$test[[base_idx[1]]] <- test_base[["satorra.bentler"]]
      }
      fp@external$sam.struc.corrected <- TRUE
      fp
    },
    silent = TRUE
  )

  if (inherits(out, "try-error")) {
    return(fit_pa) # the untouched original
  }
  out
}

# (two-step robust) fit measures for the structural part of a local SAM model,
# extracted from the (corrected) structural fit object
lav_sam_struc_fit <- function(fit_pa = NULL) {
  measures <- c("chisq", "df", "pvalue", "cfi", "rmsea", "srmr")

  fallback <- function() {
    out <- try(fitMeasures(fit_pa, measures), silent = TRUE)
    if (inherits(out, "try-error")) {
      out <- "(unable to obtain fit measures)"
      names(out) <- "warning"
    }
    out
  }

  if (!isTRUE(fit_pa@external$sam.struc.corrected)) {
    return(fallback())
  }

  # use the fitMeasures() labels for the SCALED/ROBUST measures, so the
  # printed header makes clear these are the two-step corrected values
  # (chisq.scaled / pvalue.scaled / cfi.robust / rmsea.robust), exactly as
  # fitMeasures() reports them. (The naive fallback() above keeps the plain
  # chisq/pvalue/cfi/rmsea labels.)
  scaled_labels <- c(
    "chisq.scaled", "df", "pvalue.scaled",
    "cfi.robust", "rmsea.robust", "srmr"
  )
  fm <- try(fitMeasures(fit_pa, scaled_labels), silent = TRUE)
  if (inherits(fm, "try-error")) {
    return(fallback())
  }
  fm <- as.numeric(fm)
  names(fm) <- scaled_labels
  fm
}

# Yuan & Chan (2002) rescaled GLOBAL test statistic for sam.method = "global".
#
# The 'joint' object already carries the base statistic T = n F[S, Sigma(theta,
# gamma.hat)] -- the discrepancy of the FULL (observed-variable) model evaluated
# at the plugged-in SAM estimates -- and the correct reference df = p* - q_total
# (all measurement AND structural parameters counted) in joint@test[[1]]. What is
# missing is the scaling correction tau.hat = tr(Q Gamma)/(p* - q) of eq. (22),
# which accounts for (a) the non-normality of the data (via the ADF Gamma) and
# (b) the fact that the measurement parameters gamma were estimated SEPARATELY in
# step 1 (segregated estimation), not jointly with theta.
#
# Key simplification that lets us reuse lav_test_sb() verbatim. With
#   H.theta = sigma.dot.theta (sigma.dot.theta' W sigma.dot.theta)^-1
#                             sigma.dot.theta' W   (the W-projector onto the
#                                                   structural columns of Delta),
#   D       = I - sigma.dot.gamma P   (P = d gamma.hat / d stats, eq. 12a),
# Yuan & Chan's U (eq. 19) gives  I - sigma.dot U = (I - H.theta) D, hence
#   Q = (I - sigma.dot U)' W (I - sigma.dot U) = D' U.struc D ,
# with U.struc = W - W Delta.theta (Delta.theta' W Delta.theta)^-1 Delta.theta' W
# -- exactly the Satorra-Bentler residual-weight matrix, but built from the
# STRUCTURAL columns of Delta only. By the cyclic trace property
#   tr(Q Gamma) = tr( U.struc . (D Gamma D') ).
# So we hand lav_test_sb():
#   - delta   = Delta.theta              (structural columns of the full Delta)
#   - e_inv   = (sum_g f_g Delta.theta_g' W_g Delta.theta_g)^-1
#   - wls_v   = W                        (the estimator weight of the full model)
#   - gamma_full = D Gamma D'            (the 'measurement-residualized' Gamma)
#   - test_unscaled = the standard (unscaled) full-model test  (T and df)
# and read off the satorra.bentler-scaled statistic, which IS eq. (22).
#
# P (the step-1 measurement influence) is the SAME jacobian already used by the
# twostep.robust SEs (lav_sam_step1_local_jac(p_only = TRUE)). Multigroup support
# (Case A block-diagonal, Case B cross-group) follows from the stacked assembly:
# D and Gamma.tilde are built in the stacked statistics space, so the cross-group
# coupling that across-group measurement constraints induce in P is carried
# through automatically.
#
# Returns a list with two components, to be stored in the joint object:
#   $test          -> joint@test          (standard + the YC-scaled test)
#   $baseline.test -> joint@baseline$test (standard + a scaled baseline test)
# The baseline (independence) model has no measurement model, so the Yuan-Chan
# correction there reduces to the ordinary Satorra-Bentler scaling; we obtain it
# by refitting the baseline with test = "satorra.bentler". Both scaled tests are
# needed so that the robust/scaled fit measures (CFI, RMSEA) can be computed. On
# any failure the original (unscaled) test list is returned unchanged.
# 'test' is the internal SB-family base used to drive lav_test_sb (eq. 22 is the
# satorra.bentler-type rescaling); the resulting entry is renamed to "yuan.chan"
# (and given a Yuan-Chan label) in both the model and the baseline test lists.
lav_sam_global_test <- function(joint = NULL, step1 = NULL, step2 = NULL,
                                fit = NULL, test = "satorra.bentler") {
  test_orig <- joint@test

  # rename a satorra.bentler entry (and its $test field) to "yuan.chan"
  rename_yc <- function(test_list) {
    idx <- which(names(test_list) == "satorra.bentler")
    if (length(idx) == 0L) {
      return(test_list)
    }
    e <- test_list[[idx[1]]]
    e$test  <- "yuan.chan"
    e$label <- "Yuan-Chan (2002) correction"
    names(test_list)[idx[1]] <- "yuan.chan"
    test_list[[idx[1]]] <- e
    test_list
  }
  fallback  <- list(test = test_orig, baseline.test = NULL)

  # conditional.x: not supported (yet). The machinery below runs, but with
  # NON-NORMAL (eg skewed or binary) covariates the scaled statistic is badly
  # anti-conservative (empirical rejection rates up to ~0.3 at the 5% level
  # in simulations, with either random or fixed covariate values; the
  # simultaneous sem() satorra.bentler test on the same data is accurate, and
  # so is the sam yuan.chan test with multivariate normal covariates -- the
  # interplay between the unconditionally fitted step-1 blocks and the
  # conditional full-model discrepancy is not yet correctly captured by the
  # D Gamma D' residualization). Report the honest unscaled test instead.
  # (The corrected STRUCTURAL test of sam.method = "local" does not have
  # this problem: it is accurate in the same simulations.)
  if (joint@Model@conditional.x) {
    lav_msg_warn(gettext(
      "the Yuan & Chan (2002) scaled test statistic (test = \"yuan.chan\") is
       not available (yet) when conditional.x = TRUE; the standard (unscaled)
       test is reported instead."))
    return(fallback)
  }

  # locate the standard (unscaled) full-model test = base statistic T
  std_idx <- which(vapply(test_orig, function(x) x$test, character(1L)) ==
                   "standard")
  if (length(std_idx) == 0L) {
    return(fallback) # no usable base statistic (eg test = "none")
  }
  std <- test_orig[[std_idx[1]]]
  if (is.null(std$df) || anyNA(std$df) || std$df < 1L) {
    return(fallback) # saturated full model: no global test
  }

  out <- try(
    {
      ngroups <- joint@Data@ngroups
      ntot    <- joint@SampleStats@ntotal
      fg      <- unlist(joint@SampleStats@nobs) / ntot
      step1_free_idx <- step1$step1.free.idx
      step2_free_idx <- step2$step2.free.idx

      # ingredients of the full (joint) model:
      # stored NACOV, else recompute with the fit-time settings. The
      # (unbiased) ADF Gamma is unavailable for the fixed.x / conditional.x
      # setting (lav_samp_gamma() with unbiased = TRUE stops), so fall back
      # to the *biased* ADF Gamma there (gamma.unbiased = FALSE) -- exactly
      # the Gamma that simultaneous sem(test = "satorra.bentler") uses for
      # such models. Asymptotically equivalent; the simple (no-x) cases keep
      # the unbiased Gamma so their values are unchanged.
      gamma <- tryCatch(lav_gamma_used(joint), error = function(e) NULL)
      if (is.null(gamma) || is.null(gamma[[1]])) {
        opts <- joint@Options
        opts$gamma.unbiased <- FALSE
        gamma <- lav_gamma_used(
          lavdata        = joint@Data,
          lavoptions     = opts,
          lavsamplestats = joint@SampleStats,
          lavh1          = joint@h1,
          lavimplied     = joint@implied
        )
      }
      delta <- lavTech(joint, "Delta")
      wls_v <- lavTech(joint, "WLS.V")

      # P = d(gamma.mm)/d(stats), reusing the twostep.robust jacobian; align its
      # rows with step1.free.idx (P rows are in ascending free-parameter order)
      p <- lav_sam_step1_local_jac(step1 = step1, fit = fit, p_only = TRUE)
      p_row_idx <- match(step1_free_idx, attr(p, "free.idx"))
      if (anyNA(p_row_idx)) {
        lav_msg_stop(gettext(
          "internal error: unable to align the step 1 jacobian (P) for the
           global test statistic"))
      }
      p <- p[p_row_idx, , drop = FALSE]

      # per group: structural and measurement columns of Delta, the weight W,
      # and the SB-convention Gamma (Cov of sqrt(N) * per-group statistics)
      delta_theta_list <- vector("list", ngroups)
      delta_gamma_list <- vector("list", ngroups)
      gamma_sb_list    <- vector("list", ngroups)
      a_theta <- matrix(0, length(step2_free_idx), length(step2_free_idx))
      for (g in seq_len(ngroups)) {
        delta_theta_list[[g]] <- delta[[g]][, step2_free_idx, drop = FALSE]
        delta_gamma_list[[g]] <- delta[[g]][, step1_free_idx, drop = FALSE]
        gamma_sb_list[[g]]    <- gamma[[g]] / fg[g]
        # structural information E.theta = sum_g f_g Delta.theta_g' W_g Delta.theta_g
        wd <- lav_sam_wls_delta(wls_v[[g]], delta_theta_list[[g]])
        a_theta <- a_theta + fg[g] * crossprod(delta_theta_list[[g]], wd)
      }
      e_theta_inv <- solve(a_theta)

      # D (stacked) = I - Delta.gamma P ; Gamma.tilde = D Gamma D'
      delta_gamma_all <- do.call(rbind, delta_gamma_list)
      nstat_all <- nrow(delta_gamma_all)
      d_mat <- diag(nstat_all) - delta_gamma_all %*% p
      gamma_sb_all <- lav_mat_bdiag(gamma_sb_list)
      gamma_tilde  <- d_mat %*% gamma_sb_all %*% t(d_mat)

      # reuse lav_test_sb: structural-only Delta/E.inv/W + residualized Gamma
      # (gamma_full forces the assembled 'original' path)
      lav_test_sb(
        lavmodel       = joint@Model,
        lavsamplestats = joint@SampleStats,
        lavoptions     = joint@Options,
        lavimplied     = joint@implied,
        lavdata        = joint@Data,
        test_unscaled  = std,
        e_inv          = e_theta_inv,
        delta          = delta_theta_list,
        wls_v          = wls_v,
        m_gamma        = gamma, # ignored when gamma_full is set; avoids recompute
        gamma_full     = gamma_tilde,
        test           = test
      )
    },
    silent = TRUE
  )

  if (inherits(out, "try-error")) {
    # the Yuan-Chan correction reuses the same step-1 jacobian (P) as the
    # robust SEs, so it shares their current limitations. Equality constraints
    # (within-block and across-group) and exogenous covariates (fixed.x, and
    # conditional.x with the biased ADF Gamma) ARE supported. Fall back to the
    # unscaled test, with a note.
    lav_msg_warn(gettext(
      "the Yuan & Chan (2002) scaled test statistic (test = \"yuan.chan\") is
       not available for this model; the standard (unscaled) test is reported
       instead."))
    return(fallback) # fall back to the unscaled standard test
  }

  # name the scaled entry "yuan.chan" (the SAM-global analogue of the
  # simultaneous "satorra.bentler"); @test = list(standard, yuan.chan)
  out <- rename_yc(out)

  # baseline (independence) model: no measurement model -> the Yuan-Chan
  # correction reduces to the ordinary Satorra-Bentler scaling. Refit the
  # baseline with the matching scaled test so the robust CFI/RMSEA can be
  # computed (and so fitMeasures()/summary() do not choke on a model that has a
  # scaled test while its baseline does not).
  baseline_test <- NULL
  if (!is.null(joint@baseline$partable)) {
    refit_baseline <- function(gamma_unbiased = NULL) {
      opts <- fit@Options
      opts$se       <- "none"
      opts$test     <- test
      opts$baseline <- FALSE
      opts$fit.by.level <- FALSE
      opts$estimator <- joint@Model@estimator
      if (!is.null(gamma_unbiased)) {
        opts$gamma.unbiased <- gamma_unbiased
      }
      fit_base <- lavaan::lavaan(
        model             = joint@baseline$partable,
        slot_data         = joint@Data,
        slot_sample_stats = joint@SampleStats,
        slot_options      = opts,
        verbose           = FALSE
      )
      rename_yc(fit_base@test)
    }
    baseline_test <- try(refit_baseline(), silent = TRUE)
    if (inherits(baseline_test, "try-error")) {
      # the unbiased ADF Gamma (the sam default) is not available for eg
      # fixed.x models: retry with the biased ADF Gamma, exactly as the
      # model test above does (lav_gamma_used() fallback). Without a scaled
      # baseline the scaled/robust incremental fit measures would be NA.
      baseline_test <- try(refit_baseline(gamma_unbiased = FALSE),
                           silent = TRUE)
    }
    if (inherits(baseline_test, "try-error")) {
      baseline_test <- NULL
    }
  }

  list(test = out, baseline.test = baseline_test)
}

lav_sam_table <- function(joint = NULL, step1 = NULL, fit_pa = NULL,
                          cmd = NULL, lavoptions = NULL,
                          mm_args = list(), struc_args = list(),
                          sam_method = "local",
                          local_options = list(), global_options = list()) {
  mm_fit <- step1$MM.FIT

  sam_mm_table <- data.frame(
    Block = seq_along(step1$mm.list),
    Latent = sapply(mm_fit, function(x) {
      paste(unique(unlist(x@pta$vnames$lv)), collapse = ",")
    }),
    Nind = sapply(mm_fit, function(x) {
      length(unique(unlist(x@pta$vnames$ov)))
    }),
    Chisq = sapply(mm_fit, function(x) {
      x@test[[1]]$stat
    }),
    Df = sapply(mm_fit, function(x) {
      x@test[[1]]$df
    })
  )
  class(sam_mm_table) <- c("lavaan.data.frame", "data.frame")

  # extra info for @internal slot
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    # the structural fit object, with the (two-step robust) corrected test
    # and baseline injected; for multigroup models with across-group
    # measurement constraints (Case B) these are recomputed with the full
    # cross-group Gamma. The object is stored in the @internal slot, so that
    # fitMeasures(), lavResiduals() and modindices() can delegate to it
    # (issue #517)
    sam_struc_fit_object <- lav_sam_struc_fit_object(
      fit_pa = fit_pa, step1 = step1
    )
    # (two-step robust) fit measures for the structural part
    sam_struc_fit <- lav_sam_struc_fit(fit_pa = sam_struc_fit_object)
    sam_mm_rel <- step1$REL
  } else {
    sam_struc_fit <- paste0("no local fit measures available for",
                    "structural part if sam.method is global")
    names(sam_struc_fit) <- "warning"
    sam_struc_fit_object <- NULL
    sam_mm_rel <- numeric(0L)
  }


  sam_1 <- list(
    sam.cmd = cmd,
    sam.method = sam_method,
    sam.local.options = local_options,
    sam.global.options = global_options,
    sam.mm.list = step1$mm.list,
    sam.mm.estimator = mm_fit[[1]]@Model@estimator,
    sam.mm.args = mm_args,
    sam.mm.ov.names = lapply(mm_fit, function(x) {
      x@pta$vnames$ov
    }),
    sam.mm.table = sam_mm_table,
    sam.mm.rel = sam_mm_rel,
    sam.struc.estimator = fit_pa@Model@estimator,
    sam.struc.args = struc_args,
    sam.struc.fit = sam_struc_fit,
    sam.struc.fit.object = sam_struc_fit_object,
    sam.lavoptions = lavoptions
  )
  sam_1
}

lav_sam_get_cov_ybar <- function(fit = NULL, local_options = list(
                                  M.method = "ML",
                                  lambda.correction = TRUE,
                                  alpha.correction = 0L,
                                  twolevel.method = "h1"
                                )) {

  # local.twolevel.method
  local_twolevel_method <- tolower(local_options[["twolevel.method"]])
  if (!local_twolevel_method %in% c("h1", "anova", "mean")) {
    lav_msg_stop(gettext(
      "local option twolevel.method should be one of h1, anova or mean."))
  }

  lavpta <- fit@pta
  nlevels <- lavpta$nlevels
  nblocks <- lavpta$nblocks

  # do we need H1?
  if (nlevels > 1L && local_twolevel_method == "h1") {
    h1 <- lav_h1_implied_logl(
      lavdata = fit@Data,
      lavsamplestats = fit@SampleStats,
      lavoptions = fit@Options
    )
    h1implied <- h1$implied
  } else {
    h1implied <- fit@h1$implied
  }

  # containers
  cov_list  <- vector("list", nblocks)
  ybar_list <- vector("list", nblocks)
  slopes_list <- vector("list", nblocks)

  # label
  if (nblocks > 1L) {
    names(cov_list)  <- fit@Data@block.label
    names(ybar_list) <- fit@Data@block.label
    names(slopes_list) <- fit@Data@block.label
  }

  # collect COV/YBAR per block
  for (b in seq_len(nblocks)) {

    # get sample statistics for this block
    if (nlevels > 1L) {
      # block order is group-major: (g1l1, g1l2, g2l1, g2l2, ...)
      this_level <- (b - 1L) %% nlevels + 1L
      this_group <- floor((b - 1L) / nlevels) + 1L
      if (this_level > 2L) {
        lav_msg_stop(gettext("level 3 not supported (yet)."))
      }

      if (local_twolevel_method == "h1") {
        # note: the h1 estimates are already reduced to the
        # level-specific variables -- no need to subset
        cov_1 <- h1implied$cov[[(this_group - 1L) * nlevels + this_level]]
        ybar <- h1implied$mean[[(this_group - 1L) * nlevels + this_level]]
      } else {
        ylp <- fit@SampleStats@YLp[[this_group]][[2]]
        if (this_level == 1L) {
          # "anova" and "mean" use the same within statistics
          cov_1 <- ylp$Sigma.W
          ybar <- ylp$Mu.W
        } else if (local_twolevel_method == "anova") {
          cov_1 <- ylp$Sigma.B
          ybar <- ylp$Mu.B
        } else { # "mean"
          s_pw <- ylp$Sigma.W
          nj <- ylp$s
          y2 <- ylp$Y2
          # grand mean
          mu_y <- ylp$Mu.W + ylp$Mu.B
          y2c <- t(t(y2) - mu_y) # MUST be centered
          yb <- crossprod(y2c) / nrow(y2c)
          cov_1 <- yb - 1 / nj * s_pw
          ybar <- ylp$Mu.B
        }

        # reduce to the level-specific variables
        ov_idx <- fit@Data@Lp[[this_group]]$ov.idx[[this_level]]
        cov_1 <- cov_1[ov_idx, ov_idx, drop = FALSE]
        ybar <- ybar[ov_idx]
      }

      # single level
    } else {
      this_group <- b
      if (fit@Model@conditional.x) {
        ybar <- h1implied$res.int[[b]]
        cov_1  <- h1implied$res.cov[[b]]
        slopes_list[[b]] <- h1implied$res.slopes[[b]]
      } else {
        ybar <- h1implied$mean[[b]] # EM version if missing="ml"
        cov_1  <- h1implied$cov[[b]]
      }
    } # single level

    cov_list[[b]] <- cov_1
    ybar_list[[b]] <- ybar
  }

  list(COV = cov_list, YBAR = ybar_list, SLOPES = slopes_list)
}

# automatically generate mm.list; we create measurement blocks so that:
# - unlinked factors become a singleton
# - linked factor are joined into a single measurement block
#   based: cross-loadings, or correlated residuals between item errors
lav_sam_get_mmlist <- function(lavobject) {

  lavmodel <- lavobject@Model
  lavpta <- lavobject@pta
  nblocks <- lavpta$nblocks

  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")

  mm_list <- vector("list", length = nblocks)

  glist <- lavTech(lavobject, "partable")
  for (b in seq_len(nblocks)) {
    mm_in_block <- (seq_len(lavmodel@nmat[b]) +
        cumsum(c(0, lavmodel@nmat))[b])
    mlist <- glist[mm_in_block]
    cc <- t(mlist$lambda) %*% mlist$theta %*% mlist$lambda

    # note: CC contains dummy lv's, higher-order, etc... and they
    # must be removed

    # get ALL lv names (including dummy ov.x/ov.y)
    ov_names <- lavmodel@dimNames[[lambda_idx[b]]][[1L]]
    lv_names <- lavmodel@dimNames[[lambda_idx[b]]][[2L]]
    names_1 <- lv_names

    # needs to removed:
    # - all lv.ind variables
    # - all higher-order latent variables
    # - all dummy lv's
    rm_idx <- c(match(lavpta$vnames$lv.ind[[b]], lv_names),
                match(lavpta$vnames$lv.interaction[[b]], lv_names),
                which(lv_names %in% ov_names))
    if (length(rm_idx) > 0L) {
      cc <- cc[-rm_idx, -rm_idx, drop = FALSE]
      names_1 <- lv_names[-rm_idx]
    }

    # cluster membership
    membership <- lav_graph_get_connected_nodes(cc)

    out <- split(names_1, names_1[membership])
    names(out) <- paste("block", seq_along(out), sep = "")

    mm_list[[b]] <- out
  }

  mm_list
}

# labels for the elements of the WLS.obs statistics vector in the
# categorical case: 1) thresholds and (negative) means, interleaved per
# variable, 2) [conditional.x only] the slopes of the y-variables on the
# exogenous covariates, column-major (lav_mat_vec(SLOPES): all y for x1,
# then x2, ...), 3) variances (numeric variables only), 4) correlations
# (vech, no diagonal); used to map the statistics of a measurement block
# into the statistics vector of the joint model
lav_sam_wls_obs_labels <- function(ov_names, th_idx, ov_names_x = NULL) {
  nvar <- length(ov_names)

  # variable of each th element: ordered variables have th_idx > 0; the
  # th_idx == 0 elements are the (negative) means of the numeric
  # variables, appearing in variable order
  v_of <- th_idx
  num_var_idx <- which(!seq_len(nvar) %in% th_idx)
  v_of[th_idx == 0L] <- num_var_idx

  th_labels <- character(length(th_idx))
  count <- integer(nvar)
  for (k in seq_along(th_idx)) {
    v <- v_of[k]
    count[v] <- count[v] + 1L
    th_labels[k] <- paste0(ov_names[v], "|t", count[v])
  }

  # slopes (conditional.x only), column-major over the nvar x nexo matrix
  slope_labels <- character(0L)
  if (length(ov_names_x) > 0L) {
    slope_labels <- paste0(rep(ov_names, times = length(ov_names_x)), "~",
                           rep(ov_names_x, each = nvar))
  }

  # variances (numeric variables only)
  var_labels <- character(0L)
  if (length(num_var_idx) > 0L) {
    var_labels <- paste0(ov_names[num_var_idx], "|var")
  }

  # correlations: lav_mat_vech(cov, diagonal = FALSE), column-major;
  # use a canonical (sorted) pair label, as the relative order of two
  # variables may differ between the joint model and a measurement block
  cor_labels <- character(0L)
  if (nvar > 1L) {
    row_idx <- matrix(seq_len(nvar), nvar, nvar)[lower.tri(diag(nvar))]
    col_idx <- matrix(seq_len(nvar), nvar, nvar,
                      byrow = TRUE)[lower.tri(diag(nvar))]
    cor_labels <- vapply(seq_along(row_idx), function(k) {
      paste(sort(c(ov_names[row_idx[k]], ov_names[col_idx[k]])),
            collapse = "~~")
    }, character(1L))
  }

  c(th_labels, slope_labels, var_labels, cor_labels)
}

# conditional.x: permutation of the conditional structural-moment vector
# c( vecr(cbind(EETA, RS_eta)), vech(VETA) ) -- with the variables in
# VETA (measurement) order and the covariates in data (ov.names.x) order --
# to the structural model's internal order (target_eta = its ov.nox,
# target_x = its ov.x). This is the layout of FIT.PA's WLS.obs: the NACOV
# passed to lavaan() is consumed as-is (no name matching), so Gamma.eta must
# be permuted with this index before the step-2 handoff.
lav_sam_condx_perm_idx <- function(eta_names, x_names,
                                   target_eta, target_x) {
  ord <- match(target_eta, eta_names)
  ord_x <- match(target_x, x_names)
  if (length(ord) != length(eta_names) || anyNA(ord) ||
      length(ord_x) != length(x_names) || anyNA(ord_x)) {
    lav_msg_stop(gettext(
      "internal error: unable to align the conditional structural moments
      with the structural model (Gamma.eta permutation)"))
  }
  p <- length(eta_names)
  q <- length(x_names)
  mean_block <- unlist(lapply(ord, function(j) {
    (j - 1L) * (q + 1L) + c(1L, 1L + ord_x)
  }))
  pos <- matrix(0L, p, p)
  pos[lower.tri(pos, diag = TRUE)] <- seq_len(p * (p + 1L) / 2L)
  pos[upper.tri(pos)] <- t(pos)[upper.tri(pos)]
  sub <- pos[ord, ord, drop = FALSE]
  vech_perm <- sub[lower.tri(sub, diag = TRUE)]
  c(mean_block, p * (q + 1L) + vech_perm)
}

# merge the per-block mm clusters (as returned by lav_sam_get_mmlist())
# into a single mm.list, transitively joining clusters that share a latent
# variable or an observed indicator; needed for multilevel models, where
# the within and between factors of the same indicators must end up in the
# same (two-level) measurement block
lav_sam_mmlist_merge_blocks <- function(lavobject, mm_list_per_block) {
  pt_1 <- lavobject@ParTable
  nblocks <- lavobject@pta$nblocks
  block_values <- lav_pt_block_values(pt_1)

  # collect all clusters: a set of lv names + their observed indicators
  clusters <- list()
  for (b in seq_len(nblocks)) {
    for (cl in mm_list_per_block[[b]]) {
      lv <- unlist(cl)
      ind_idx <- which(pt_1$op == "=~" & pt_1$block == block_values[b] &
                       pt_1$lhs %in% lv)
      clusters <- c(clusters,
                    list(list(lv = lv, ov = unique(pt_1$rhs[ind_idx]))))
    }
  }

  # merge until no two clusters share an lv or an ov
  changed <- TRUE
  while (changed) {
    changed <- FALSE
    i <- 1L
    while (i < length(clusters)) {
      j <- i + 1L
      while (j <= length(clusters)) {
        if (any(clusters[[i]]$lv %in% clusters[[j]]$lv) ||
            any(clusters[[i]]$ov %in% clusters[[j]]$ov)) {
          clusters[[i]]$lv <- unique(c(clusters[[i]]$lv, clusters[[j]]$lv))
          clusters[[i]]$ov <- unique(c(clusters[[i]]$ov, clusters[[j]]$ov))
          clusters[[j]] <- NULL
          changed <- TRUE
        } else {
          j <- j + 1L
        }
      }
      i <- i + 1L
    }
  }

  lapply(clusters, "[[", "lv")
}

lav_sam_veta_pt <- function(lavobject, block = 1L) {

  lavmodel <- lavobject@Model

  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")

  glist <- lavTech(lavobject, "partable")
  b <- block
  mm_in_block <- (seq_len(lavmodel@nmat[b]) + cumsum(c(0, lavmodel@nmat))[b])
  mlist <- glist[mm_in_block]
  mm_psi    <- (mlist$psi    != 0) + 0L
  nr <- nrow(mlist$psi)
  if (!is.null(mlist$beta)) {
    mm_beta <- (mlist$beta != 0) + 0L
    tmp <- -mm_beta
    tmp[lav_mat_diag_idx(nr)] <- 1
    ib_inv <- solve(tmp)
  } else {
    ib_inv <- diag(nr)
  }
  veta <- ib_inv %*% mm_psi %*% t(ib_inv)
  # get ALL lv names (including dummy ov.x/ov.y)
  lv_names <- lavmodel@dimNames[[lambda_idx[b]]][[2L]]
  colnames(veta) <- rownames(veta) <- lv_names

  veta
}

lav_sam_veta_con <- function(s = NULL, mm_lambda = NULL, mm_theta = NULL,
                             l_veta = NULL, local_m_method = "ML",
                             tol = 1e-07, max_iter = 100L) {

  # first do GLS/ULS
  m <- ncol(mm_lambda)
  if (local_m_method == "ULS") {
    m_w <- diag(m)
  } else {
    m_w <- solve(s)
  }

  wl <- m_w %*% mm_lambda
  t_lw  <- t(wl)
  t_lwl <- t_lw %*% mm_lambda
  tmp1 <- t(l_veta) %*% (t_lwl %x% t_lwl) %*% l_veta
  tmp2 <- t(l_veta) %*% lav_mat_vec(t_lw %*% (s - mm_theta) %*% wl)
  out <- solve(tmp1, tmp2)
  veta_init <- matrix(l_veta %*% out, m, m)

  if (local_m_method != "ML") {
    # we are done!
    return(veta_init)
  }

  veta_new <- veta_init
  for (i in seq_len(max_iter)) {
    sigma_ml <- mm_lambda %*% veta_new %*% t(mm_lambda) + mm_theta
    m_w <- solve(sigma_ml)
    wl <- m_w %*% mm_lambda
    t_lw  <- t(wl)
    t_lwl <- t_lw %*% mm_lambda
    tmp1 <- t(l_veta) %*% (t_lwl %x% t_lwl) %*% l_veta
    tmp2 <- t(l_veta) %*% lav_mat_vec(t_lw %*% (s - mm_theta) %*% wl)
    out <- solve(tmp1, tmp2)
    veta_ml <- matrix(l_veta %*% out, m, m)
    rmsea <- sqrt(sum((veta_new - veta_ml)^2))
    #cat("i = ", i, " rmsea = ", rmsea, "\n")
    veta_new <- veta_ml
    if (rmsea < tol) {
      break
    }
  }

  veta_new
}

# print 'blocks' of test statistics
# - blocks with 'scaling.factors' come first (in 'two columns')
# - then come the 'single-column' test statistics (eg browne.residual.adf)
# - print additional information (eg information matrix, h1.information, ...)
#   if they deviate from what is used for the standard errors

# this is used by the summary() function and lavTest(, output = "text")

# YR 13 Feb 2026: if test[[1]] has stat = NA, skip if there are other tests

lav_test_print <- function(object, nd = 3L) {
  # object is list of tests
  test <- object

  # empty list?
  if (is.null(test) || length(test) == 0L || !is.list(test)) {
    return(character(0L))
  }

  # test = "none"?
  if (test[[1]]$test == "none") {
    return(character(0L))
  }

  # remove empty first test (stat = NA) if multiple tests are available)
  if (length(test) > 1L && is.na(test[[1]]$stat)) {
    test <- test[-1]
  }

  # meta data
  info <- attr(object, "info")
  ngroups <- info$ngroups
  group_label <- info$group.label
  information <- info$information
  h1_information <- info$h1.information
  observed_information <- info$observed.information

  # num format
  num_format <- paste("%", max(8L, nd + 5L), ".", nd, "f", sep = "")

  # header
  cat("Model Test User Model:\n")

  # locate 'robust' tests (here: having a scaling factor)
  has_no_scaling <- unname(sapply(
    (lapply(test, "[[", "scaling.factor")),
    is.null
  ))
  robust_idx <- which(!has_no_scaling)
  non_robust_idx <- which(has_no_scaling)
  scaled_idx <- 1L
  if (length(robust_idx) > 0L) {
    scaled_idx <- which(names(test) == test[[robust_idx[1]]]$scaled.test)
    if (length(scaled_idx) == 0L) {
      scaled_idx <- 1L
    }
    # remove 'scaled.test', because it is shown together with robust
    non_robust_idx <- non_robust_idx[-scaled_idx]
  }
  blocks <- c(robust_idx, non_robust_idx)
  # n_blocks <- length(blocks)

  # print out blocks
  for (block in blocks) {
    # one or two-columns for this block?
    if (length(robust_idx) > 0L && block %in% robust_idx) {
      twocolumn <- TRUE
    } else {
      twocolumn <- FALSE
    }

    if (!twocolumn) {
      # print label
      c1 <- c2 <- c3 <- character(0L)
      if (!is.null(test[[block]]$label)) {
        c1 <- c(c1, test[[block]]$label)
        c2 <- c(c2, "")
        c3 <- c(c3, "")
      }
      if (is.na(test[[block]]$df) || test[[block]]$df == 0L) {
        c1 <- c(c1, c("Test statistic", "Degrees of freedom"))
        c2 <- c(c2, c(
          sprintf(num_format, test[[block]]$stat),
          ifelse(test[[block]]$df %% 1 == 0, # integer
            test[[block]]$df,
            sprintf(num_format, test[[block]]$df)
          )
        ))
        c3 <- c(c3, c("", ""))
      } else {
        plabel <- "P-value"
        if (!is.null(test[[block]]$refdistr)) {
          if (test[[block]]$refdistr == "chisq") {
            plabel <- "P-value (Chi-square)"
          } else if (test[[block]]$refdistr == "unknown") {
            plabel <- "P-value (Unknown)"
          } else if (test[[block]]$refdistr == "bootstrap") {
            plabel <- "P-value (Bollen-Stine bootstrap)"
          }
        }
        c1 <- c(c1, c("Test statistic", "Degrees of freedom", plabel))
        c2 <- c(c2, c(
          sprintf(num_format, test[[block]]$stat),
          ifelse(test[[block]]$df %% 1 == 0, # integer
            test[[block]]$df,
            sprintf(num_format, test[[block]]$df)
          ),
          sprintf(num_format, test[[block]]$pvalue)
        ))
        c3 <- c(c3, c("", "", ""))
      }

      # two-column
    } else {
      # print label
      c1 <- c2 <- c3 <- character(0L)
      if (!is.null(test[[scaled_idx]]$label)) {
        c1 <- c(c1, test[[scaled_idx]]$label)
        c2 <- c(c2, "")
        c3 <- c(c3, "")
      }
      if (is.na(test[[block]]$df) || test[[block]]$df == 0L) {
        c1 <- c(c1, c("Test Statistic", "Degrees of freedom"))
        c2 <- c(
          c2,
          c(
            sprintf(num_format, test[[scaled_idx]]$stat),
            ifelse(test[[scaled_idx]]$df %% 1 == 0, # integer
              test[[scaled_idx]]$df,
              sprintf(num_format, test[[scaled_idx]]$df)
            )
          )
        )
        c3 <- c(
          c3,
          c(
            sprintf(num_format, test[[block]]$stat),
            ifelse(test[[block]]$df %% 1 == 0, # integer
              test[[block]]$df,
              sprintf(num_format, test[[block]]$df)
            )
          )
        )
      } else {
        if (!is.null(test[[scaled_idx]]$refdistr)) {
          if (test[[scaled_idx]]$refdistr == "chisq") {
            plabel <- "P-value (Chi-square)"
          } else if (test[[scaled_idx]]$refdistr == "unknown") {
            plabel <- "P-value (Unknown)"
          } else {
            plabel <- "P-value"
          }
        }
        c1 <- c(c1, c(
          "Test Statistic", "Degrees of freedom", plabel,
          "Scaling correction factor"
        ))
        c2 <- c(
          c2,
          c(
            sprintf(num_format, test[[scaled_idx]]$stat),
            ifelse(test[[scaled_idx]]$df %% 1 == 0, # integer
              test[[scaled_idx]]$df,
              sprintf(num_format, test[[scaled_idx]]$df)
            ),
            sprintf(num_format, test[[scaled_idx]]$pvalue), ""
          )
        )
        c3 <- c(
          c3,
          c(
            sprintf(num_format, test[[block]]$stat),
            ifelse(test[[block]]$df %% 1 == 0, # integer
              test[[block]]$df,
              sprintf(num_format, test[[block]]$df)
            ),
            sprintf(num_format, test[[block]]$pvalue),
            sprintf(num_format, test[[block]]$scaling.factor)
          )
        )

        if (test[[block]]$test %in% c(
          "scaled.shifted", "scaled.shifted.corrected"
        )) {
          if (ngroups == 1L ||
            length(test[[block]]$shift.parameter) == 1L) {
            c1 <- c(c1, "Shift parameter")
            c2 <- c(c2, "")
            c3 <- c(
              c3,
              sprintf(num_format, test[[block]]$shift.parameter)
            )
          } else {
            c1 <- c(c1, "Shift parameter for each group:")
            c2 <- c(c2, "")
            c3 <- c(c3, "")
            for (g in 1:ngroups) {
              c1 <- c(c1, sprintf("    %-38s", group_label[[g]]))
              c2 <- c(c2, "")
              c3 <- c(c3, sprintf(
                num_format,
                test[[block]]$shift.parameter[g]
              ))
            }
          }
        } # shift

        # which correction factor?
        c1 <- c(c1, paste("  ", test[[block]]$label, sep = ""))
        c2 <- c(c2, "")
        c3 <- c(c3, "")
      }
    }

    # if twocolumn, label first row
    if (twocolumn && block == blocks[1]) {
      c1 <- c("", c1)
      c2 <- c("Standard", c2)
      c3 <- c("Scaled", c3)
    } else {
      # empty row
      c1 <- c("", c1)
      c2 <- c("", c2)
      c3 <- c("", c3)
    }

    # if information type is different from 'se', print it
    if (length(information) > 1L &&
      information[1] != information[2]) {
      c1 <- c(c1, "Information")
      tmp_txt <- information[2]
      c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
        substring(tmp_txt, 2),
        sep = ""
      ))
      c3 <- c(c3, "")
    }
    # if h1.information type is different from 'se', print it
    if (length(h1_information) > 1L &&
      h1_information[1] != h1_information[2]) {
      c1 <- c(c1, "Information saturated (h1) model")
      tmp_txt <- h1_information[2]
      c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
        substring(tmp_txt, 2),
        sep = ""
      ))
      c3 <- c(c3, "")
    }
    # if observed.information type is different from 'se', print it
    if (length(observed_information) > 1L &&
      information[2] == "observed" &&
      (observed_information[1] !=
        observed_information[2])) {
      c1 <- c(c1, "Observed information based on")
      tmp_txt <- observed_information[2]
      c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
        substring(tmp_txt, 2),
        sep = ""
      ))
      c3 <- c(c3, "")
    }


    # format c1/c2/c3 (note: fitMeasures uses 35/16/8)
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (twocolumn) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)

    # multiple groups?
    ngroups <- ngroups
    if (ngroups > 1L && !is.null(test[[block]]$stat.group)) {
      c1 <- c2 <- c3 <- character(ngroups)
      for (g in 1:ngroups) {
        tmp <- sprintf("  %-40s", group_label[[g]])
        c1[g] <- format(tmp, width = 43L)
        if (!twocolumn) {
          tmp <- sprintf(num_format, test[[block]]$stat.group[g])
          c2[g] <- format(tmp,
            width = 8L + max(0, (nd - 3L)) * 4L,
            justify = "right"
          )
        } else {
          tmp <- sprintf(num_format, test[[block]]$stat.group[g])
          c2[g] <- format(tmp,
            width = 8L + max(0, (nd - 3L)) * 4L,
            justify = "right"
          )
          tmp <- sprintf(num_format, test[[block]]$stat.group[g])
          c3[g] <- format(tmp, width = 8L + nd, justify = "right")
        }
      }
      if (twocolumn) {
        m <- cbind(c1, c2, c3, deparse.level = 0)
      } else {
        m <- cbind(c1, c2, deparse.level = 0)
      }
      colnames(m) <- rep("", ncol(m))
      rownames(m) <- rep(" ", nrow(m))
      cat("  Test statistic for each group:\n")
      write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
    }
  } # blocks

  # invisible(M)
}

# the univariate two-level Gaussian (random-intercept) model
#
#     y_ij = mu + xw_ij' bw + xb_j' bb + b_j + e_ij
#     b_j ~ N(0, vb),  e_ij ~ N(0, vw)
#
# YR 2026 (two-level WLS, phase 2; covariates added in phase 7)
#
# parameter vector: theta = c(mu, vw, vb[, bw, bb])
# the slope parameters (bw = within-level covariates, bb = between-level
# covariates) are APPENDED, so that the positions of (mu, vw, vb) do not
# depend on the presence of covariates
#
# this is stage 1 of the stage-wise unrestricted two-level model for a
# continuous variable (cfr. lav_uvreg.R for the single-level case)
#
# fit object (list) contract -- see lav_uvreg_2l_fit():
#   theta, mu_idx (1), wvar_idx (2), bvar_idx (3), wslope_idx, bslope_idx,
#   nexo_w, nexo_b, nexo, between, y, x_w, x_b, z, resid,
#   cluster_idx, cs (no covariates only), logl, converged

lav_uvreg_2l_fit <- function(y = NULL,
                             x_w = NULL, # within-level covariates
                             x_b = NULL, # between-level covariates
                             cluster_idx = NULL,
                             between = TRUE, # vb free? (FALSE for
                                             # within-only variables)
                             optim_method = "nlminb1",
                             control = list(),
                             output = "list") {
  # create cache environment
  cache <- lav_uvreg_2l_init_cache(y = y, x_w = x_w, x_b = x_b,
                                   cluster_idx = cluster_idx,
                                   between = between)
  nexo <- cache$nexo

  # optim
  minfns <- lav_uvbv_optim_fns(
    optim_method, lav_uvreg_2l_min_fns(covariates = (nexo > 0L))
  )

  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  lower <- c(-Inf, .Machine$double.eps^0.75, 0, rep(-Inf, nexo))
  upper <- c(+Inf, +Inf, +Inf, rep(+Inf, nexo))
  if (!between) {
    # pin vb at zero
    lower[3] <- upper[3] <- 0
    cache$theta[3] <- 0
  }

  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb,
    lower = lower, upper = upper,
    cache = cache
  )

  if (output == "cache") {
    return(cache)
  }

  # loglik (and residuals) at the optimum
  if (nexo > 0L) {
    cache$theta <- optim$par
    logl <- lav_uvreg_2l_x_loglik_cache(cache = cache)
    resid <- cache$e
  } else {
    logl <- -0.5 * sum(lav_2l_gauss_m2ll(
      cs = cache$cs, mu = optim$par[1],
      sigma_w = matrix(optim$par[2], 1, 1),
      sigma_b = matrix(optim$par[3], 1, 1)
    ))
    resid <- NULL
  }

  list(
    theta = optim$par,
    mu_idx = 1L, wvar_idx = 2L, bvar_idx = 3L,
    wslope_idx = if (cache$nexo_w > 0L) 3L + seq_len(cache$nexo_w)
                 else integer(0L),
    bslope_idx = if (cache$nexo_b > 0L) 3L + cache$nexo_w +
                   seq_len(cache$nexo_b)
                 else integer(0L),
    nexo_w = cache$nexo_w, nexo_b = cache$nexo_b, nexo = nexo,
    between = between,
    y = cache$y,
    x_w = cache$x_w, x_b = cache$x_b, z = cache$z,
    resid = resid,
    cluster_idx = cache$cluster_idx,
    cs = cache$cs,
    logl = logl,
    converged = (optim$convergence == 0L)
  )
}

# prepare cache environment
lav_uvreg_2l_init_cache <- function(y = NULL,
                                    x_w = NULL,
                                    x_b = NULL,
                                    cluster_idx = NULL,
                                    between = TRUE,
                                    parent = parent.frame()) {
  y <- as.numeric(y)
  if (anyNA(y)) {
    lav_msg_stop(gettext(
      "missing values are not supported (yet) in lav_uvreg_2l_fit()."))
  }
  nexo_w <- if (is.null(x_w)) 0L else NCOL(x_w)
  nexo_b <- if (is.null(x_b)) 0L else NCOL(x_b)
  nexo <- nexo_w + nexo_b

  if (nexo == 0L) {
    cs <- lav_2l_cluster_stats(y = y, cluster_idx = cluster_idx)

    # ANOVA-type starting values
    nobs <- cs$nobs
    nclusters <- cs$nclusters
    njs <- cs$njs
    mu0 <- mean(y)
    vw0 <- sum(cs$wtot) / (nobs - nclusters)
    msb <- sum(njs * (cs$ybar[, 1] - mu0)^2) / (nclusters - 1)
    n0 <- (nobs - sum(njs^2) / nobs) / (nclusters - 1)
    vb0 <- max((msb - vw0) / n0, 0.01 * vw0)
    if (!between) {
      vb0 <- 0
      vw0 <- sum((y - mu0)^2) / nobs
    }
    theta <- c(mu0, vw0, vb0)

    return(list2env(
      list(
        y = y, x_w = NULL, x_b = NULL, z = NULL,
        cluster_idx = cluster_idx, cs = cs,
        nexo_w = 0L, nexo_b = 0L, nexo = 0L,
        n = nobs, theta = theta
      ),
      parent = parent
    ))
  }

  # with covariates
  if (!between && nexo_b > 0L) {
    lav_msg_stop(gettext(
      "a within-only variable cannot be regressed on between-level
      covariates."))
  }
  z <- cbind(1, x_w, x_b)
  storage.mode(z) <- "double"
  if (anyNA(z)) {
    lav_msg_stop(gettext(
      "missing values are not supported (yet) in lav_uvreg_2l_fit()."))
  }
  nobs <- length(y)
  njs <- tabulate(cluster_idx)
  nclusters <- length(njs)

  # OLS starting values for the mean parameters; ANOVA-type starting
  # values for (vw, vb) from the OLS residuals
  theta_m0 <- qr.coef(qr(z), y)
  theta_m0[is.na(theta_m0)] <- 0
  r <- y - drop(z %*% theta_m0)
  rbar <- drop(rowsum.default(r, group = cluster_idx, reorder = TRUE)) / njs
  sr2 <- drop(rowsum.default(r * r, group = cluster_idx, reorder = TRUE))
  vw0 <- sum(sr2 - njs * rbar^2) / (nobs - nclusters)
  msb <- sum(njs * rbar^2) / (nclusters - 1)
  n0 <- (nobs - sum(njs^2) / nobs) / (nclusters - 1)
  vb0 <- max((msb - vw0) / n0, 0.01 * vw0)
  if (!between) {
    vb0 <- 0
    vw0 <- sum(r * r) / nobs
  }
  theta <- c(theta_m0[1], vw0, vb0, theta_m0[-1])

  list2env(
    list(
      y = y, x_w = x_w, x_b = x_b, z = z,
      cluster_idx = cluster_idx, njs = njs, nclusters = nclusters,
      nexo_w = nexo_w, nexo_b = nexo_b, nexo = nexo,
      n = nobs, theta = theta
    ),
    parent = parent
  )
}

lav_uvreg_2l_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    mu <- theta[1]
    vw <- theta[2]
    vb <- theta[3]
    m2ll_j <- lav_2l_gauss_m2ll(
      cs = cs, mu = mu,
      sigma_w = matrix(vw, 1, 1), sigma_b = matrix(vb, 1, 1)
    )
    loglik <- -0.5 * sum(m2ll_j)
    return(loglik)
  })                     # nolint end
}

# the (dmu, dSw, dSb) patterns of theta = c(mu, vw, vb)
lav_uvreg_2l_params <- function() {
  list(
    list(dmu = 1, dsw = NULL, dsb = NULL),
    list(dmu = NULL, dsw = matrix(1, 1, 1), dsb = NULL),
    list(dmu = NULL, dsw = NULL, dsb = matrix(1, 1, 1))
  )
}

lav_uvreg_2l_grad_cache <- function(cache = NULL) {
  with(cache, {           # nolint start
    mu <- theta[1]
    vw <- theta[2]
    vb <- theta[3]
    dm2ll <- lav_2l_gauss_dm2ll(
      cs = cs, mu = mu,
      sigma_w = matrix(vw, 1, 1), sigma_b = matrix(vb, 1, 1),
      params = lav_uvreg_2l_params()
    )
    return(-0.5 * colSums(dm2ll))
  })                      # nolint end
}

# --- with covariates ---
#
# per-cluster -2 loglik (univariate, mean structure mu_ij = z_ij' theta_m):
# with e_ij = y_ij - mu_ij, ebar_j the cluster residual means, sww_j the
# within-cluster residual scatter and Vj = vw + nj vb,
#
#   -2 logl_j = nj log(2 pi) + (nj - 1) log(vw) + log(Vj)
#               + sww_j / vw + nj ebar_j^2 / Vj
#
# scores: the GLS-weighted residuals m_ij = [Sigma_j^{-1} e_j]_ij have the
# closed form m_ij = (e_ij - h_j ebar_j) / vw with h_j = nj vb / Vj, and
# the score of the loglik w.r.t. any mean parameter with observation-level
# design column z is sum_ij z_ij m_ij

lav_uvreg_2l_x_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    vw <- theta[2]
    vb <- theta[3]
    if (vw <= 0 || vb < 0) {
      return(-Inf)
    }
    theta_m <- theta[c(1L, 3L + seq_len(nexo))]
    e <- y - drop(z %*% theta_m)
    ebar <- drop(rowsum.default(e, group = cluster_idx,
                                reorder = TRUE)) / njs
    se2 <- drop(rowsum.default(e * e, group = cluster_idx,
                               reorder = TRUE))
    sww <- se2 - njs * ebar^2
    vj <- vw + njs * vb
    m2ll_j <- njs * log(2 * pi) + (njs - 1) * log(vw) + log(vj) +
      sww / vw + njs * ebar^2 / vj
    loglik <- -0.5 * sum(m2ll_j)
    return(loglik)
  })                     # nolint end
}

# assumes the loglik has been evaluated at cache$theta (e, ebar, sww, vj)
lav_uvreg_2l_x_grad_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    vw <- theta[2]
    vb <- theta[3]
    h <- njs * vb / vj
    m <- (e - (h * ebar)[cluster_idx]) / vw
    g_mean <- drop(crossprod(z, m))
    g_vw <- -0.5 * sum((njs - 1) / vw + 1 / vj - sww / vw^2 -
                         njs * ebar^2 / vj^2)
    g_vb <- -0.5 * sum(njs / vj - njs^2 * ebar^2 / vj^2)
    return(c(g_mean[1], g_vw, g_vb, g_mean[-1]))
  })                     # nolint end
}

# cluster-wise scores, covariate case; columns (mu, vw, vb, slopes)
lav_uvreg_2l_x_sc <- function(y = NULL, z = NULL, cluster_idx = NULL,
                              theta = NULL, nexo = NULL) {
  njs <- tabulate(cluster_idx)
  vw <- theta[2]
  vb <- theta[3]
  theta_m <- theta[c(1L, 3L + seq_len(nexo))]
  e <- y - drop(z %*% theta_m)
  ebar <- drop(rowsum.default(e, group = cluster_idx, reorder = TRUE)) / njs
  se2 <- drop(rowsum.default(e * e, group = cluster_idx, reorder = TRUE))
  sww <- se2 - njs * ebar^2
  vj <- vw + njs * vb
  h <- njs * vb / vj
  m <- (e - (h * ebar)[cluster_idx]) / vw
  sc_mean <- rowsum.default(z * m, group = cluster_idx, reorder = TRUE)
  sc_vw <- -0.5 * ((njs - 1) / vw + 1 / vj - sww / vw^2 -
                     njs * ebar^2 / vj^2)
  sc_vb <- -0.5 * (njs / vj - njs^2 * ebar^2 / vj^2)
  cbind(sc_mean[, 1], sc_vw, sc_vb, sc_mean[, -1, drop = FALSE])
}

# slope column names (within covariates first, then between covariates)
lav_uvreg_2l_slope_names <- function(nexo_w = 0L, nexo_b = 0L) {
  c(
    if (nexo_w > 0L) paste0("wsl", seq_len(nexo_w)) else character(0L),
    if (nexo_b > 0L) paste0("bsl", seq_len(nexo_b)) else character(0L)
  )
}

# cluster-wise scores (of the loglik) at given (or fitted) parameter values
lav_uvreg_2l_sc <- function(fit_y = NULL, theta = NULL) {
  if (is.null(theta)) {
    theta <- fit_y$theta
  }
  if (fit_y$nexo > 0L) {
    sc <- lav_uvreg_2l_x_sc(
      y = fit_y$y, z = fit_y$z, cluster_idx = fit_y$cluster_idx,
      theta = theta, nexo = fit_y$nexo
    )
    colnames(sc) <- c(
      "mu", "wvar", "bvar",
      lav_uvreg_2l_slope_names(fit_y$nexo_w, fit_y$nexo_b)
    )
    return(sc)
  }
  dm2ll <- lav_2l_gauss_dm2ll(
    cs = fit_y$cs, mu = theta[1],
    sigma_w = matrix(theta[2], 1, 1), sigma_b = matrix(theta[3], 1, 1),
    params = lav_uvreg_2l_params()
  )
  sc <- -0.5 * dm2ll
  colnames(sc) <- c("mu", "wvar", "bvar")
  sc
}

# nlminb objective/gradient (see lav_uvbv_common.R)
lav_uvreg_2l_min_fns <- function(covariates = FALSE) {
  if (covariates) {
    lav_uvbv_min_fns(
      logl_fun = lav_uvreg_2l_x_loglik_cache,
      grad_fun = lav_uvreg_2l_x_grad_cache
    )
  } else {
    lav_uvbv_min_fns(
      logl_fun = lav_uvreg_2l_loglik_cache,
      grad_fun = lav_uvreg_2l_grad_cache
    )
  }
}

# summary information for a single (lavaan) efa model
#
# workflow:
# - summary() first calls lav_efalist_summary()
# - for each model, lav_efalist_summary() calls lav_object_summary() with
#   efa = TRUE and efa.args
# - for each model, lav_object_summary() calls
#   lav_efa_summary(object, efa.args = efa.args) to populate the $efa slot


# efa summary for a single lavaan object
lav_efa_summary <- function(object,
                            efa_args = list(
                              lambda = TRUE,
                              theta = TRUE,
                              psi = TRUE,
                              eigenvalues = TRUE,
                              sumsq_table = TRUE,
                              lambda_structure = FALSE,
                              fs_determinacy = FALSE,
                              se = FALSE,
                              zstat = FALSE,
                              pvalue = FALSE
                            )) {
  stopifnot(inherits(object, "lavaan"))
  efa_args <- lav_snake_case(efa_args)

  nblocks <- object@Model@nblocks
  orthogonal_flag <- object@Options$rotation.args$orthogonal

  # get standardized solution
  mm_lambda <- mm_theta <- mm_psi <- NULL
  std <- lavTech(object, "std",
    add.class = TRUE, add.labels = TRUE,
    list.by.group = FALSE
  )
  lambda_idx <- which(names(std) == "lambda")
  theta_idx <- which(names(std) == "theta")
  psi_idx <- which(names(std) == "psi")

  # LAMBDA
  mm_lambda <- std[lambda_idx]
  names(mm_lambda) <- NULL

  # THETA
  mm_theta <- std[theta_idx]
  # make THETA diagonal
  mm_theta <- lapply(seq_len(nblocks), function(b) {
    tmp <- diag(mm_theta[[b]])
    class(tmp) <- c("lavaan.vector", "numeric")
    tmp
  })

  # PSI
  mm_psi <- std[psi_idx]
  names(mm_psi) <- NULL

  # eigenvalues correlation matrix
  std_ov <- object@Options$rotation.args$std_ov
  cov_1 <- object@h1$implied$cov # h1
  if (std_ov) {
    cov_1 <- lapply(cov_1, cov2cor)
  }
  eigvals <- NULL
  if (efa_args$eigenvalues) {
    eigvals <- lapply(seq_len(nblocks), function(b) {
      tmp <- eigen(cov_1[[b]], only.values = TRUE)$values
      names(tmp) <- paste("ev", seq_len(nrow(mm_lambda[[b]])), sep = "")
      class(tmp) <- c("lavaan.vector", "numeric")
      tmp
    })
  }

  fs_determinacy <- NULL
  # Note: these 'determinacy' values are only properly defined for the
  #       'regression' factor scores! (If we would apply the same formulas
  #       for Bartlett factor scores, we would obtain 1's!
  if (efa_args$fs_determinacy) {
    fs_determinacy <- lapply(seq_len(nblocks), function(b) {
      cor_1 <- cov2cor(cov_1[[b]]) # just in case
      cor_inv <- try(solve(cor_1), silent = TRUE)
      if (inherits(cor_inv, "try-error")) {
        return(rep(as.numeric(NA), nrow(mm_psi[[b]])))
      }
      fs <- mm_lambda[[b]] %*% mm_psi[[b]] # factor structure
      out <- sqrt(diag(t(fs) %*% cor_inv %*% fs))
      class(out) <- c("lavaan.vector", "numeric")
      out
    })
  }

  # sum-of-squares table
  sumsq_table <- NULL
  if (efa_args$sumsq_table) {
    sumsq_table <- lapply(seq_len(nblocks), function(b) {
      # nvar <- nrow(mm_lambda[[b]])
      nfactor <- ncol(mm_lambda[[b]])

      # sum of squares:
      # - if orthogonal, this is really the sum of the squared factor
      #   loadings
      # - if oblique, we need to take the correlation into account
      sumsq <- diag(mm_psi[[b]] %*% crossprod(mm_lambda[[b]]))

      # reorder
      if (nfactor > 1L) {
        # determine order
        order_idx <- sort.int(sumsq, decreasing = TRUE, index.return = TRUE)$ix
        # re-order from large to small
        sumsq <- sumsq[order_idx]
      }

      # Proportion 'explained' (= proportion of total sumsq)
      #   note: sum(sumsq) == sum(communalities)
      propexpl <- sumsq / sum(sumsq)

      # Proportion var (= sumsq/nvar)
      propvar <- sumsq / nrow(mm_lambda[[b]])

      # Cumulative var
      cumvar <- cumsum(propvar)

      # construct table
      tmp <- rbind(sumsq, propexpl, propvar, cumvar)

      # total + colnames
      if (nfactor > 1L) {
        # add total column
        tmp <- cbind(tmp, rowSums(tmp))
        tmp[4, ncol(tmp)] <- tmp[3, ncol(tmp)]
        colnames(tmp) <-
          c(
            colnames(mm_lambda[[b]])[order_idx],
            "total"
          )
      } else {
        colnames(tmp) <- colnames(mm_lambda[[b]])[1]
      }

      # rownames
      if (nfactor == 1L) {
        ssq_label <- "Sum of squared loadings"
      } else if (orthogonal_flag) {
        ssq_label <- "Sum of sq (ortho) loadings"
      } else {
        ssq_label <- "Sum of sq (obliq) loadings"
      }
      rownames(tmp) <- c(
        ssq_label,
        "Proportion of total",
        "Proportion var",
        "Cumulative var"
      )

      # class
      class(tmp) <- c("lavaan.matrix", "matrix")

      tmp
    })
  } # sumsq_table

  # (factor) structure coefficients
  if (efa_args$lambda_structure) {
    lambda_structure <- lapply(seq_len(nblocks), function(b) {
      tmp <- mm_lambda[[b]] %*% mm_psi[[b]]
      class(tmp) <- c("lavaan.matrix", "matrix")
      tmp
    })
  } else {
    lambda_structure <- NULL
  }

  # standard errors (if any)
  lambda_se <- theta_se <- psi_se <- NULL
  lambda_zstat <- theta_zstat <- psi_zstat <- NULL
  lambda_pval <- theta_pval <- psi_pval <- NULL
  if (object@Options$se != "none") {
    se <- lavTech(object, "std.se",
      add.class = TRUE, add.labels = TRUE,
      list.by.group = FALSE
    )

    se_flag <- (efa_args$se || efa_args$zstat || efa_args$pvalue)

    # ALWAYS use lambda.se
    if (efa_args$lambda) {
      lambda_se <- se[lambda_idx]
      names(lambda_se) <- NULL
    }

    # theta.se
    if (se_flag && efa_args$theta) {
      theta_se <- se[theta_idx]
      # make theta.se diagonal
      theta_se <- lapply(seq_len(nblocks), function(b) {
        tmp <- diag(theta_se[[b]])
        class(tmp) <- c("lavaan.vector", "numeric")
        tmp
      })
    }

    # ALWAYS use psi.se
    if (efa_args$psi) {
      psi_se <- se[psi_idx]
      names(psi_se) <- NULL
    }

    # compute zstat
    if (efa_args$zstat || efa_args$pvalue) {
      if (efa_args$lambda) {
        lambda_zstat <- lapply(seq_len(nblocks), function(b) {
          tmp_se <- lambda_se[[b]]
          tmp_se[tmp_se < sqrt(.Machine$double.eps)] <-
            as.numeric(NA)
          tmp <- mm_lambda[[b]] / tmp_se
          class(tmp) <- c("lavaan.matrix", "matrix")
          tmp
        })
      }
      if (efa_args$theta) {
        theta_zstat <- lapply(seq_len(nblocks), function(b) {
          tmp_se <- theta_se[[b]]
          tmp_se[tmp_se < sqrt(.Machine$double.eps)] <-
            as.numeric(NA)
          tmp <- mm_theta[[b]] / tmp_se
          class(tmp) <- c("lavaan.vector", "numeric")
          tmp
        })
      }
      if (efa_args$psi) {
        psi_zstat <- lapply(seq_len(nblocks), function(b) {
          tmp_se <- psi_se[[b]]
          tmp_se[tmp_se < sqrt(.Machine$double.eps)] <-
            as.numeric(NA)
          tmp <- mm_psi[[b]] / tmp_se
          class(tmp) <- c(
            "lavaan.matrix.symmetric",
            "matrix"
          )
          tmp
        })
      }
    }

    # compute pval
    if (efa_args$pvalue) {
      if (efa_args$lambda) {
        lambda_pval <- lapply(seq_len(nblocks), function(b) {
          tmp <- 2 * (1 - pnorm(abs(lambda_zstat[[b]])))
          class(tmp) <- c("lavaan.matrix", "matrix")
          tmp
        })
      }
      if (efa_args$theta) {
        theta_pval <- lapply(seq_len(nblocks), function(b) {
          tmp <- 2 * (1 - pnorm(abs(theta_zstat[[b]])))
          class(tmp) <- c("lavaan.vector", "numeric")
          tmp
        })
      }
      if (efa_args$psi) {
        psi_pval <- lapply(seq_len(nblocks), function(b) {
          tmp <- 2 * (1 - pnorm(abs(psi_zstat[[b]])))
          class(tmp) <- c(
            "lavaan.matrix.symmetric",
            "matrix"
          )
          tmp
        })
      }
    }
  } # se/zstat/pvalue

  # block.label
  block_label <- object@Data@block.label

  # we remove them here; we may have needed them for other parts
  if (!efa_args$lambda) {
    mm_lambda <- NULL
  }
  if (!efa_args$theta) {
    mm_theta <- NULL
  }
  if (!efa_args$psi) {
    mm_psi <- NULL
  }
  if (!efa_args$se) {
    # always keep lambda.se and psi.se (for the signif stars)
    theta_se <- NULL
  }
  if (!efa_args$zstat) {
    lambda_zstat <- theta_zstat <- psi_zstat <- NULL
  }

  res <- list(
    nblocks = nblocks,
    block.label = block_label,
    std_ov = std_ov,
    eigvals = eigvals,
    sumsq_table = sumsq_table,
    orthogonal = object@Options$rotation.args$orthogonal,
    lambda_structure = lambda_structure,
    fs_determinacy = fs_determinacy,
    lambda = mm_lambda,
    theta = mm_theta,
    psi = mm_psi,
    lambda.se = lambda_se,
    lambda.zstat = lambda_zstat,
    lambda.pvalue = lambda_pval,
    psi.se = psi_se,
    psi.zstat = psi_zstat,
    psi.pvalue = psi_pval,
    theta.se = theta_se,
    theta.zstat = theta_zstat,
    theta.pvalue = theta_pval
  )

  res
}


# summary efaList
lav_efalist_summary <- function(object, nd = 3L, cutoff = 0.3, dot_cutoff = 0.1,
                            alpha_level = 0.01,
                            lambda = TRUE, theta = TRUE, psi = TRUE,
                            fit_table = TRUE, fs_determinacy = FALSE,
                            eigenvalues = TRUE, sumsq_table = TRUE,
                            lambda_structure = FALSE, se = FALSE,
                            zstat = FALSE, pvalue = FALSE, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  # kill object$loadings if present
  object[["loadings"]] <- NULL

  # unclass the object
  y <- unclass(object)

  # construct efa.args
  efa_args <- list(
    lambda = lambda, theta = theta, psi = psi,
    eigenvalues = eigenvalues, sumsq_table = sumsq_table,
    lambda_structure = lambda_structure,
    fs_determinacy = fs_determinacy,
    se = se, zstat = zstat, pvalue = pvalue
  )

  # extract useful info from first model
  out <- lav_object_summary(y[[1]],
    header = TRUE, estimates = FALSE,
    efa = FALSE
  )

  # header information
  lavaan_version <- out$header$lavaan.version
  converged_flag <- all(sapply(y, lavInspect, "converged"))

  # estimator
  estimator <- out$optim$estimator
  estimator_args <- out$optim$estimator.args

  # rotation
  rotation <- out$rotation$rotation
  rotation_args <- out$rotation$rotation.args

  # data
  lavdata <- out$data

  # main part: lav_object_summary information per model
  res <- lapply(y, lav_object_summary,
    header = FALSE,
    fit_measures = FALSE, estimates = TRUE, efa = TRUE,
    efa_args = efa_args
  )

  # number of factors (for ALL blocks)
  nfactors <- sapply(y, function(x) x@pta$nfac[[1]])

  # fit.measures
  table_1 <- NULL
  if (fit_table) {
    # first, create standard table
    fit <- fitMeasures(object, fit.measures = "default")
    names_1 <- rownames(fit)
    idx <- integer(0L)

    # AIC/BIC
    if (all(c("aic", "bic", "bic2") %in% names_1)) {
      this_idx <- match(c("aic", "bic", "bic2"), names_1)
      idx <- c(idx, this_idx)
    }

    # chi-square
    if (all(c("chisq.scaled", "df.scaled", "pvalue.scaled") %in% names_1)) {
      this_idx <- match(
        c("chisq.scaled", "df.scaled", "pvalue.scaled"),
        names_1
      )
      idx <- c(idx, this_idx)
    } else {
      this_idx <- match(c("chisq", "df", "pvalue"), names_1)
      idx <- c(idx, this_idx)
    }

    # CFI
    if ("cfi.robust" %in% names_1 && !all(is.na(fit["cfi.robust", ]))) {
      this_idx <- match("cfi.robust", names_1)
      idx <- c(idx, this_idx)
    } else if ("cfi.scaled" %in% names_1) {
      this_idx <- match("cfi.scaled", names_1)
      idx <- c(idx, this_idx)
    } else if ("cfi" %in% names_1) {
      this_idx <- match("cfi", names_1)
      idx <- c(idx, this_idx)
    }

    # RMSEA
    if ("rmsea.robust" %in% names_1 && !all(is.na(fit["rmsea.robust", ]))) {
      this_idx <- match("rmsea.robust", names_1)
      idx <- c(idx, this_idx)
    } else if ("rmsea.scaled" %in% names_1) {
      this_idx <- match("rmsea.scaled", names_1)
      idx <- c(idx, this_idx)
    } else if ("rmsea" %in% names_1) {
      this_idx <- match("rmsea", names_1)
      idx <- c(idx, this_idx)
    }

    # table with fitmeasures
    if (length(idx) > 0L) {
      table_1 <- t(fit[idx, , drop = FALSE])
      tmp <- names_1[idx]
      # strip '.scaled'
      tmp <- gsub(".scaled", "", tmp, fixed = TRUE)
      # strip '.robust'
      tmp <- gsub(".robust", "", tmp, fixed = TRUE)
      # rename "bic2" -> "sabic"
      bic2_idx <- which(tmp == "bic2")
      if (length(bic2_idx) > 0L) {
        tmp[bic2_idx] <- "sabic"
      }
      colnames(table_1) <- tmp
    } else {
      table_1 <- matrix(0, nrow = nfactors, ncol = 0L)
    }
    rownames(table_1) <- paste("nfactors = ", nfactors, sep = "")
    class(table_1) <- c("lavaan.matrix", "matrix")
  }

  # create return object
  out <- list(
    lavaan.version = lavaan_version,
    converged.flag = converged_flag,
    estimator = estimator,
    estimator.args = estimator_args,
    rotation = rotation,
    rotation.args = rotation_args,
    lavdata = lavdata,
    fit_table = table_1,
    nfactors = nfactors,
    model.list = res
  )

  # add nd, cutoff, dot_cutoff, ... as attributes (for printing)
  attr(out, "nd") <- nd
  attr(out, "cutoff") <- cutoff
  attr(out, "dot_cutoff") <- dot_cutoff
  attr(out, "alpha_level") <- alpha_level

  # create class
  class(out) <- c("efaList.summary", "list")

  out
}

# casewise residuals

lav_residuals_casewise <- function(object, labels = labels) {

  # check object
  object <- lav_object_check_version(object)

  # check if we have full data
  if (object@Data@data.type != "full") {
    lav_msg_stop(gettext("casewise residuals not available if sample statistics
                         were used for fitting the model"))
  }
  # check if we have categorical data
  if (object@Model@categorical) {
    lav_msg_stop(gettext(
      "casewise residuals not available if data is categorical"))
  }

  g_1 <- object@Data@ngroups

  x_1 <- object@Data@X
  if (object@Model@categorical) {
    # add 'eXo' columns to X
    x_1 <- lapply(seq_len(object@Data@ngroups), function(g) {
      ret <- cbind(x_1[[g]], object@Data@eXo[[g]])
      ret
    })
  }
  m <- lav_predict_yhat(object)
  # Note: if M has already class lavaan.matrix, print goes crazy
  # with Error: C stack usage is too close to the limit
  out_1 <- lapply(seq_len(g_1), function(x) {
    out <- x_1[[x]] - m[[x]]
    class(out) <- c("lavaan.matrix", "matrix")
    out
  })

  if (labels) {
    for (g in 1:g_1) {
      colnames(out_1[[g]]) <- object@pta$vnames$ov[[g]]
    }
  }

  if (g_1 == 1) {
    out_1 <- out_1[[1]]
  } else {
    names(out_1) <- unlist(object@Data@group.label)
  }

  out_1
}

# loglikelihood clustered/twolevel data

# YR: first version around Feb 2017
# YR 28 Oct 2024: [EM steps:] if fx.delta is NA, check if we have (near)-perfect
#                 correlations, and provide an informative warning
# YR 07 Jul 2026: representation invariants documented below; shared
#                 scatter/gather core (lav_cl_* in the kernels file);
#                 out= hooks to skip re-conversion (refactoring; no change
#                 in behavior)


# ---------------------------------------------------------------------------
# the 'implied' and '2l' representations of the model-implied moments
# ---------------------------------------------------------------------------
#
# two representations are used throughout the two-level machinery
# (this file, lav_mvnorm_cluster_kernels.R, lav_mvnorm_cluster_missing*.R,
#  and -- for conditional.x = TRUE -- lav_mvreg_cluster.R):
#
# 1) the 'implied' representation: one mean vector + covariance matrix PER
#    LEVEL, exactly as produced by lav_model_implied() and consumed by the
#    rest of lavaan (fitted, residuals, lavInspect, h1, delta method, ...):
#      - mean[[1]], cov[[1]]: the level-1 variables, i.e. ov.idx[[1]]
#        (within-only + shared 'both' variables)
#      - mean[[2]], cov[[2]]: the level-2 variables, i.e. ov.idx[[2]]
#        ('both' + between-only variables)
#
# 2) the '2l' representation: the y/z partition in which the marginal
#    likelihood of two-level data is naturally expressed, where
#    y = the level-1 variables and z = the between-only variables:
#      - sigma.w  : Var(e_ij)  (y variables; within residual)
#      - sigma.b  : Var(u_j)   (y variables; only the 'both' block is
#                   nonzero)
#      - sigma.zz : Var(z_j)
#      - sigma.yz : Cov(u_j, z_j) (only the 'both' rows are nonzero)
#      - mu.y     : E(y) = mu.w + mu.b
#      - mu.z     : E(z)
#      - mu.w/mu.b: the separate within/between mean parts (see below)
#
# lav_mvn_cl_implied22l() maps 1) -> 2); lav_mvn_cl_2l2implied() maps
# 2) -> 1). both maps pass through p.tilde x p.tilde 'tilde' matrices,
# where p.tilde = the number of distinct variables across both levels:
# each level's matrices are first scattered into the tilde universe, and
# the y/z blocks are then gathered from it.
#
# invariants -- read this before touching any of the code below:
#
# * naming: list elements in the implied representation are Capitalized
#   dot.case (Mu.W, Sigma.B); elements in the 2l representation are
#   lowercase dot.case (mu.y, sigma.zz) -- the case tells you which space
#   you are in
#
# * ordering: the tilde universe puts the between-only variables LAST;
#   ov.idx[[1]] is strictly increasing and equals the ascending complement
#   of between.idx, so the y block of the 2l matrices -- gathered as
#   [-between.idx] from the tilde universe -- is ordered exactly like
#   ov.idx[[1]]; the z block is ordered like between.idx.
#   ov.idx[[2]] need NOT be increasing (it follows the level-2 variable
#   order, e.g. c(1, 3, 2)); this is harmless because the conversions only
#   use it symmetrically (scatter and gather with the same index vector),
#   so its internal order cancels out
#
# * mean folding: for the 'both' variables only the SUM mu.w + mu.b is
#   identified; implied22l() therefore folds the two parts into
#   mu.y = mu.w + mu.b, and its mu.w/mu.b outputs follow the convention
#   mu.w[both.idx] = 0 and mu.b[within.idx] = 0
#
# * 2l2implied() is used in TWO distinct roles:
#   (a) as the parameter map back (EM updates, final h1 estimates):
#       mu_w = and mu_b = are given explicitly (mu_y = NULL), and are
#       scattered back to the per-level mean vectors
#   (b) as the gradient ADJOINT (both maps are linear!): given the
#       gradient of -2*logl wrt the 2l matrices, the same gather/scatter
#       -- transposed -- yields the gradient wrt the implied matrices;
#       in this role mu_y = d(-2logl)/d(mu.y) is given, and BOTH mu.w
#       and mu.b receive the full mu.y block, because wrt the implied
#       representation d/d(mu.w[both]) = d/d(mu.b[both]) = d/d(mu.y[both])
#       (only their sum enters the likelihood).
#       do NOT zero mu.w[both.idx] inside 2l2implied(): that would break
#       role (b), producing non-symmetric (numerically differentiated)
#       Hessians -- this regressed twice before, in 0.6-5 and 0.6-18;
#       zeroing mu.w[both.idx] is a post-estimation normalization only
#
# * round trip: on the sigma parts, 2l2implied(implied22l(x)) == x
#   exactly (pure scatter/gather, no arithmetic); on the mean parts the
#   round trip reproduces x only after x itself has been normalized to
#   the folding convention above
#   (property tests: ~/semtest/claude_tests/cl_repr/)
# ---------------------------------------------------------------------------

# take model-implied mean+variance matrices, and reorder/augment them
# to facilitate computing of (log)likelihood in the two-level case

# when conditional.x = FALSE:
# - sigma.w and sigma.b: same dimensions, level-1 variables only
# - sigma.zz: level-2 variables only
# - sigma.yz: cov(level-1, level-2)
# - mu.y: level-1 variables only (mu.w + mu.b)
# - mu.w: y within  part
# - mu.b: y between part
# - mu.z: level-2 variables only
lav_mvn_cl_implied22l <- function(lp = NULL,
                                  implied = NULL,
                                  mu_w = NULL,
                                  mu_b = NULL,
                                  sigma_w = NULL,
                                  sigma_b = NULL) {
  if (!is.null(implied)) {
    # FIXME: only for single-group analysis!
    sigma_w <- implied$cov[[1]]
    mu_w <- implied$mean[[1]]

    sigma_b <- implied$cov[[2]]
    mu_b <- implied$mean[[2]]
  }

  # within/between.idx
  between_idx <- lp$between.idx[[2]]
  within_idx <- lp$within.idx[[2]]
  both_idx <- lp$both.idx[[2]]

  # ov.idx per level
  ov_idx <- lp$ov.idx

  # 'tilde' matrices: ALL variables within and between
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))

  # sigma blocks: shared scatter/gather core (lav_mvnorm_cluster_kernels.R)
  out <- lav_cl_sigma_22l(
    sigma_w = sigma_w, sigma_b = sigma_b, p_tilde = p_tilde,
    l1_idx = ov_idx[[1]], l2_idx = ov_idx[[2]],
    z_idx = between_idx, rm_idx = between_idx
  )

  # Mu.W.tilde
  mu_w_tilde <- numeric(p_tilde)
  mu_w_tilde[ov_idx[[1]]] <- mu_w

  # Mu.B.tilde
  mu_b_tilde <- numeric(p_tilde)
  mu_b_tilde[ov_idx[[2]]] <- mu_b

  # mean folding (see the header invariants): mu.y = mu.w + mu.b
  mu_wb_tilde <- numeric(p_tilde)
  mu_wb_tilde[within_idx] <- mu_w_tilde[within_idx]
  mu_wb_tilde[both_idx] <- (mu_b_tilde[both_idx] +
    mu_w_tilde[both_idx])

  # set Mu.W[both.idx] to zero (after we added to WB)
  mu_w_tilde[both_idx] <- 0
  # get Mu.B[both.idx] from WB
  mu_b_tilde[both_idx] <- mu_wb_tilde[both_idx]

  # map to vectors needed for loglik
  if (length(within_idx) > 0L) {
    mu_b_tilde[within_idx] <- 0
  }
  if (length(between_idx) > 0L) {
    mu_z <- mu_b_tilde[between_idx]
    mu_y <- mu_wb_tilde[-between_idx]
    mu_w_1 <- mu_w_tilde[-between_idx]
    mu_b_1 <- mu_b_tilde[-between_idx]
  } else {
    mu_z <- numeric(0L)
    mu_y <- mu_wb_tilde
    mu_w_1 <- mu_w_tilde
    mu_b_1 <- mu_b_tilde
  }

  list(
    sigma.w = out$sigma.w, sigma.b = out$sigma.b, sigma.zz = out$sigma.zz,
    sigma.yz = out$sigma.yz, mu.z = mu_z, mu.y = mu_y, mu.w = mu_w_1,
    mu.b = mu_b_1
  )
}

lav_mvn_cl_2l2implied <- function(lp,
                                  sigma_w = NULL,
                                  sigma_b = NULL,
                                  sigma_zz = NULL,
                                  sigma_yz = NULL,
                                  mu_z = NULL,
                                  mu_y = NULL,
                                  mu_w = NULL,
                                  mu_b = NULL) {
  # between.idx
  between_idx <- lp$between.idx[[2]]
  within_idx <- lp$within.idx[[2]]
  # both_idx <- lp$both.idx[[2]]

  # ov.idx per level
  ov_idx <- lp$ov.idx

  # 'tilde' matrices: ALL variables within and between
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))

  # gradient-adjoint role (see the header invariants): mu_y is given, and
  # BOTH mu.w and mu.b receive the full mu.y block; do NOT zero
  # mu.w[both.idx] here (non-symmetric Hessians; regressed in 0.6-5 and
  # 0.6-18) -- that is a post-estimation normalization only
  if (!is.null(mu_y)) {
    mu_b <- mu_y
    mu_w_tilde <- numeric(p_tilde)
    mu_w_tilde[ov_idx[[1]]] <- mu_y
    mu_w <- mu_w_tilde[ov_idx[[1]]]
  }

  # Mu.W
  mu_w_tilde_1 <- numeric(p_tilde)
  mu_w_tilde_1[ov_idx[[1]]] <- mu_w
  mu_w_1 <- mu_w_tilde_1[ov_idx[[1]]]

  # Mu.B
  mu_b_tilde <- numeric(p_tilde)
  mu_b_tilde[ov_idx[[1]]] <- mu_b
  mu_b_tilde[between_idx] <- mu_z
  if (length(within_idx) > 0) {
      mu_b_tilde[within_idx] <- 0
  }
  mu_b_1 <- mu_b_tilde[ov_idx[[2]]]


  # Sigma.W
  sigma_w_1 <- sigma_w

  # Sigma.B: shared scatter/gather core (lav_mvnorm_cluster_kernels.R)
  sigma_b_1 <- lav_cl_sigma_2l2(
    sigma_b = sigma_b, sigma_zz = sigma_zz, sigma_yz = sigma_yz,
    p_tilde = p_tilde, y_idx = ov_idx[[1]], z_idx = between_idx,
    l2_idx = ov_idx[[2]]
  )

  list(Mu.W = mu_w_1, Mu.B = mu_b_1, Sigma.W = sigma_w_1, Sigma.B = sigma_b_1)
}


# Mu.W, Mu.B, Sigma.W, Sigma.B are the model-implied statistics
# (not yet reordered); alternatively, a pre-converted 2l list (as returned
# by lav_mvn_cl_implied22l) may be passed as out= to avoid re-converting
lav_mvn_cl_loglik_samp_2l <- function(ylp = NULL,
                                      lp = NULL,
                                      mu_w = NULL,
                                      sigma_w = NULL,
                                      mu_b = NULL,
                                      sigma_b = NULL,
                                      sinv_method = "eigen",
                                      log2pi = FALSE,
                                      minus_two = TRUE,
                                      out = NULL) {
  # map implied to 2l matrices (unless the caller already did)
  if (is.null(out)) {
    out <- lav_mvn_cl_implied22l(
      lp = lp, mu_w = mu_w, mu_b = mu_b,
      sigma_w = sigma_w, sigma_b = sigma_b
    )
  }
  mu_y <- out$mu.y
  mu_z <- out$mu.z
  sigma_w_1 <- out$sigma.w
  sigma_b_1 <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # Lp
  # nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]
  between_idx <- lp$between.idx[[2]]
  cluster_sizes <- lp$cluster.sizes[[2]]
  ncluster_sizes <- lp$ncluster.sizes[[2]]
  cluster_size_ns <- lp$cluster.size.ns[[2]]

  # Y1 samplestats
  if (length(between_idx) > 0L) {
    s_pw <- ylp[[2]]$Sigma.W[-between_idx, -between_idx, drop = FALSE]
  } else {
    s_pw <- ylp[[2]]$Sigma.W
  }

  # Y2 samplestats
  cov_d <- ylp[[2]]$cov.d
  mean_d <- ylp[[2]]$mean.d

  # common parts:
  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w_1,
    logdet = TRUE, sinv_method = sinv_method
  )
  sigma_w_logdet <- attr(sigma_w_inv, "logdet")
  attr(sigma_w_inv, "logdet") <- NULL

  if (length(between_idx) > 0L) {
    sigma_zz_inv <- lav_mat_sym_inverse(
      s = sigma_zz,
      logdet = TRUE, sinv_method = sinv_method
    )
    sigma_zz_logdet <- attr(sigma_zz_inv, "logdet")
    attr(sigma_zz_inv, "logdet") <- NULL
    sigma_yz_zi <- sigma_yz %*% sigma_zz_inv
    sigma_zi_zy <- t(sigma_yz_zi)
    sigma_b_z <- sigma_b_1 - sigma_yz %*% sigma_zi_zy
  } else {
    sigma_zz_logdet <- 0
    sigma_b_z <- sigma_b_1
  }

  # min 2* logliklihood
  l <- numeric(ncluster_sizes) # logdet
  m_b <- numeric(ncluster_sizes) # between qf
  for (clz in seq_len(ncluster_sizes)) {
    # cluster size
    nj <- cluster_sizes[clz]

    # data between
    y2yc <- (cov_d[[clz]] + tcrossprod(mean_d[[clz]] - c(mu_z, mu_y)))

    # FIXME: avoid reorder/b.idx, so we can use between.idx
    if (length(between_idx) > 0L) {
      b_idx <- seq_along(lp$between.idx[[2]])
      y2yc_zz <- y2yc[b_idx, b_idx, drop = FALSE]
      y2yc_yz <- y2yc[-b_idx, b_idx, drop = FALSE]
      y2yc_yy <- y2yc[-b_idx, -b_idx, drop = FALSE]
    } else {
      y2yc_yy <- y2yc
    }

    # construct sigma.j
    sigma_j <- (nj * sigma_b_z) + sigma_w_1
    sigma_j_inv <- lav_mat_sym_inverse(
      s = sigma_j,
      logdet = TRUE, sinv_method = sinv_method
    )
    sigma_j_logdet <- attr(sigma_j_inv, "logdet")
    attr(sigma_j_inv, "logdet") <- NULL

    # check: what if sigma.j is non-pd? should not happen
    # (new in 0.7-1: return +Inf (-2*logl) or NA (logl) right away, instead
    #  of silently propagating NaN; matches the missing-data version)
    if (is.na(sigma_j_logdet)) {
      if (minus_two) {
        return(+Inf)
      } else {
        return(as.numeric(NA))
      }
    }

    # logdet -- between only
    l[clz] <- (sigma_zz_logdet + sigma_j_logdet)

    if (length(between_idx) > 0L) {
      # part 1 -- zz
      sigma_ji_yz_zi <- sigma_j_inv %*% sigma_yz_zi
      vinv_11 <- sigma_zz_inv + nj * (sigma_zi_zy %*% sigma_ji_yz_zi)
      q_zz <- sum(vinv_11 * y2yc_zz)

      # part 2 -- yz
      q_yz <- -nj * sum(sigma_ji_yz_zi * y2yc_yz)
    } else {
      q_zz <- q_yz <- 0
    }

    # part 5 -- yyc
    q_yyc <- -nj * sum(sigma_j_inv * y2yc_yy)

    # qf -- between only
    m_b[clz] <- q_zz + 2 * q_yz - q_yyc
  }
  # q.yya + q.yyb
  # the reason why we multiply the trace by 'N - nclusters' is
  # S.PW has been divided by 'N - nclusters'
  q_w <- sum(cluster_size - 1) * sum(sigma_w_inv * s_pw)
  # logdet within part
  l_w <- sum(cluster_size - 1) * sigma_w_logdet

  # -2*times logl (without the constant)
  loglik <- sum(l * cluster_size_ns) + sum(m_b * cluster_size_ns) + q_w + l_w

  # functions below compute -2 * logl
  if (!minus_two) {
    loglik <- loglik / (-2)
  }

  # constant
  # Note: total 'N' = (nobs * #within vars) + (nclusters * #between vars)
  if (log2pi) {
    log_2pi <- log(2 * pi)
    n_within <- length(c(lp$both.idx[[2]], lp$within.idx[[2]]))
    n_between <- length(lp$between.idx[[2]])
    p <- lp$nclusters[[1]] * n_within + lp$nclusters[[2]] * n_between
    constant <- -(p * log_2pi) / 2
    loglik <- loglik + constant
  }

  # loglik.x (only if loglik is requested)
  if (length(unlist(lp$ov.x.idx)) > 0L && log2pi && !minus_two) {
    loglik <- loglik - ylp[[2]]$loglik.x
  }

  loglik
}


# first derivative -2*logl wrt Mu.W, Mu.B, Sigma.W, Sigma.B
lav_mvn_cl_dlogl_2l_samp <- function(ylp = NULL,
                                     lp = NULL,
                                     mu_w = NULL,
                                     sigma_w = NULL,
                                     mu_b = NULL,
                                     sigma_b = NULL,
                                     return_list = FALSE,
                                     sinv_method = "eigen") {
  lav_mvn_cl_grad_engine(
    ylp = ylp, lp = lp,
    mu_w = mu_w, sigma_w = sigma_w,
    mu_b = mu_b, sigma_b = sigma_b,
    score_mode = FALSE, return_list = return_list,
    sinv_method = sinv_method
  )
}

# cluster-wise scores -2*logl wrt Mu.W, Mu.B, Sigma.W, Sigma.B
lav_mvn_cl_sc_2l <- function(y1 = NULL,
                             ylp = NULL,
                             lp = NULL,
                             mu_w = NULL,
                             sigma_w = NULL,
                             mu_b = NULL,
                             sigma_b = NULL,
                             sinv_method = "eigen") {
  lav_mvn_cl_grad_engine(
    ylp = ylp, y1 = y1, lp = lp,
    mu_w = mu_w, sigma_w = sigma_w,
    mu_b = mu_b, sigma_b = sigma_b,
    score_mode = TRUE,
    sinv_method = sinv_method
  )
}


# first-order information: outer crossprod of scores per cluster
lav_mvn_cl_info_firstorder <- function(y1 = NULL,
                                       ylp = NULL,
                                       lp = NULL,
                                       mu_w = NULL,
                                       sigma_w = NULL,
                                       mu_b = NULL,
                                       sigma_b = NULL,
                                       x_idx = NULL,
                                       divide_by_two = FALSE,
                                       sinv_method = "eigen") {
  # n <- NROW(y1)

  scores <- lav_mvn_cl_sc_2l(
    y1 = y1,
    ylp = ylp,
    lp = lp,
    mu_w = mu_w,
    sigma_w = sigma_w,
    mu_b = mu_b,
    sigma_b = sigma_b,
    sinv_method = sinv_method
  )

  # divide by 2 (if we want scores wrt objective function)
  if (divide_by_two) {
    scores <- scores / 2
  }

  # unit information
  information <- crossprod(scores) / lp$nclusters[[2]]

  # if x.idx, set rows/cols to zero
  information <- lav_mvn_cl_zero_x_idx(information,
    lp = lp, mu_w = mu_w, mu_b = mu_b, x_idx = x_idx
  )

  information
}

# expected information 'h1' model
# order: mu.w within, vech(sigma.w) within, mu.b between, vech(sigma.b) between
# mu.w rows/cols that are splitted within/between are forced to zero
lav_mvn_cl_info_expected <- function(lp = NULL,
                                     mu_w = NULL,
                                     sigma_w = NULL,
                                     mu_b = NULL,
                                     sigma_b = NULL,
                                     x_idx = integer(0L),
                                     sinv_method = "eigen") {
  # translate to internal matrices
  out <- lav_mvn_cl_implied22l(
    lp = lp,
    mu_w = mu_w, mu_b = mu_b,
    sigma_w = sigma_w, sigma_b = sigma_b
  )
  # mu_y <- out$mu.y
  # mu_z <- out$mu.z
  sigma_w_1 <- out$sigma.w
  sigma_b_1 <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # create Delta.W.tilde, Delta.B.tilde
  ov_idx <- lp$ov.idx
  nw <- length(ov_idx[[1]])
  nb <- length(ov_idx[[2]])
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))
  p_tilde_star <- p_tilde * (p_tilde + 1) / 2
  npar <- p_tilde + p_tilde_star
  vidx <- lav_cl_vech_blocks(
    p_tilde = p_tilde, l1_idx = ov_idx[[1]], l2_idx = ov_idx[[2]]
  )
  w_idx <- vidx$l1
  b_idx <- vidx$l2

  delta_w_tilde <- matrix(0, npar, npar)
  delta_b_tilde <- matrix(0, npar, npar)
  delta_w_tilde[
    c(ov_idx[[1]], w_idx + p_tilde),
    c(ov_idx[[1]], w_idx + p_tilde)
  ] <- diag(nw + nw * (nw + 1) / 2)
  delta_b_tilde[
    c(ov_idx[[2]], b_idx + p_tilde),
    c(ov_idx[[2]], b_idx + p_tilde)
  ] <- diag(nb + nb * (nb + 1) / 2)
  delta_w_tilde <- cbind(delta_w_tilde, matrix(0, npar, npar))
  delta_b_tilde <- cbind(matrix(0, npar, npar), delta_b_tilde)

  nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]

  # sum over the cluster sizes of n_s * t(Delta.j) I(Omega.j) Delta.j
  information_j <- lav_mvn_cl_info_omega_j_loop(
    lp = lp, sigma_w_1 = sigma_w_1, sigma_b_1 = sigma_b_1,
    sigma_zz = sigma_zz, sigma_yz = sigma_yz,
    delta_w_tilde = delta_w_tilde, delta_b_tilde = delta_b_tilde
  )

  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w, logdet = FALSE,
    sinv_method = sinv_method
  )
  # create Sigma.W.inv.tilde
  sigma_w_inv_tilde <- matrix(0, p_tilde, p_tilde)
  sigma_w_inv_tilde[ov_idx[[1]], ov_idx[[1]]] <- sigma_w_inv

  i11_w <- sigma_w_inv_tilde
  i22_w <- lav_mvn_kron_dup_half(sigma_w_inv_tilde)
  i_w <- lav_mat_bdiag(i11_w, i22_w)
  information_w <- (nobs - nclusters) *
    (t(delta_w_tilde) %*% i_w %*% delta_w_tilde)

  # unit information
  information_tilde <- 1 / lp$nclusters[[2]] * (information_w + information_j)

  # force zero for means both.idx in within part
  information_tilde[lp$both.idx[[2]], ] <- 0
  information_tilde[, lp$both.idx[[2]]] <- 0

  # if x.idx, set rows/cols to zero
  if (length(x_idx) > 0L) {
    xw_idx <- c(
      x_idx,
      p_tilde + lav_mat_vech_which_idx(n = p_tilde, idx = x_idx)
    )
    xb_idx <- npar + xw_idx
    all_idx <- c(xw_idx, xb_idx)
    information_tilde[all_idx, ] <- 0
    information_tilde[, all_idx] <- 0
  }

  # remove redundant rows/cols
  ok_idx <- c(
    ov_idx[[1]],
    w_idx + p_tilde,
    npar + ov_idx[[2]],
    npar + b_idx + p_tilde
  )

  information <- information_tilde[ok_idx, ok_idx, drop = FALSE]

  information
}


# expected information -- delta
# for non-saturated models only
lav_mvn_cl_info_expected_delta <- function(lp = NULL,
                                           delta = NULL,
                                           mu_w = NULL,
                                           sigma_w = NULL,
                                           mu_b = NULL,
                                           sigma_b = NULL,
                                           sinv_method = "eigen") {
  # translate to internal matrices
  out <- lav_mvn_cl_implied22l(
    lp = lp,
    mu_w = mu_w, mu_b = mu_b,
    sigma_w = sigma_w, sigma_b = sigma_b
  )
  # mu_y <- out$mu.y
  # mu_z <- out$mu.z
  sigma_w_1 <- out$sigma.w
  sigma_b_1 <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # Delta -- this group
  npar <- NCOL(delta)

  # create Delta.W.tilde, Delta.B.tilde
  ov_idx <- lp$ov.idx
  nw <- length(ov_idx[[1]])
  nw_star <- nw * (nw + 1) / 2
  nb <- length(ov_idx[[2]])

  delta_w <- delta[1:(nw + nw_star), , drop = FALSE]
  delta_b <- delta[-(1:(nw + nw_star)), , drop = FALSE]

  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))
  p_tilde_star <- p_tilde * (p_tilde + 1) / 2
  delta_w_tilde_mu <- matrix(0, p_tilde, npar)
  delta_w_tilde_sigma <- matrix(0, p_tilde_star, npar)
  delta_b_tilde_mu <- matrix(0, p_tilde, npar)
  delta_b_tilde_sigma <- matrix(0, p_tilde_star, npar)

  delta_w_tilde_mu[ov_idx[[1]], ] <- delta_w[1:nw, ]
  delta_b_tilde_mu[ov_idx[[2]], ] <- delta_b[1:nb, ]

  # correct Delta to reflect Mu.W[ both.idx ] is added to Mu.B[ both.idx ]
  # changed in 0.6-5
  delta_b_tilde_mu[lp$both.idx[[2]], ] <-
    (delta_b_tilde_mu[lp$both.idx[[2]], ] +
      delta_w_tilde_mu[lp$both.idx[[2]], ])
  delta_w_tilde_mu[lp$both.idx[[2]], ] <- 0


  vidx <- lav_cl_vech_blocks(
    p_tilde = p_tilde, l1_idx = ov_idx[[1]], l2_idx = ov_idx[[2]]
  )
  delta_w_tilde_sigma[vidx$l1, ] <- delta_w[-(1:nw), ]
  delta_b_tilde_sigma[vidx$l2, ] <- delta_b[-(1:nb), ]

  delta_w_tilde <- rbind(delta_w_tilde_mu, delta_w_tilde_sigma)
  delta_b_tilde <- rbind(delta_b_tilde_mu, delta_b_tilde_sigma)

  nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]

  # sum over the cluster sizes of n_s * t(Delta.j) I(Omega.j) Delta.j
  information_j <- lav_mvn_cl_info_omega_j_loop(
    lp = lp, sigma_w_1 = sigma_w_1, sigma_b_1 = sigma_b_1,
    sigma_zz = sigma_zz, sigma_yz = sigma_yz,
    delta_w_tilde = delta_w_tilde, delta_b_tilde = delta_b_tilde
  )

  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w_1, logdet = FALSE,
    sinv_method = sinv_method
  )
  i11_w <- sigma_w_inv
  i22_w <- lav_mvn_kron_dup_half(sigma_w_inv)
  i_w <- lav_mat_bdiag(i11_w, i22_w)

  # force zero for means both.idx in within part
  # changed in 0.6-5
  i_w[lp$both.idx[[2]], ] <- 0
  i_w[, lp$both.idx[[2]]] <- 0

  information_w <- (nobs - nclusters) * (t(delta_w) %*% i_w %*% delta_w)

  # unit information
  information <- 1 / lp$nclusters[[2]] * (information_w + information_j)


  information
}


# observed information
# order: mu.w within, vech(sigma.w) within, mu.b between, vech(sigma.b) between
# mu.w rows/cols that are splitted within/between are forced to zero
#
# numerical approximation (for now)
lav_mvn_cl_info_observed <- function(lp = NULL,
                                     ylp = NULL,
                                     mu_w = NULL,
                                     sigma_w = NULL,
                                     mu_b = NULL,
                                     sigma_b = NULL,
                                     x_idx = integer(0L),
                                     sinv_method = "eigen") {
  lav_mvn_cl_info_obs_engine(
    lp = lp,
    mu_w = mu_w, sigma_w = sigma_w,
    mu_b = mu_b, sigma_b = sigma_b,
    x_idx = x_idx, sinv_method = sinv_method,
    dlogl_fn = lav_mvn_cl_dlogl_2l_samp,
    dlogl_args = list(ylp = ylp)
  )
}

# estimate ML estimates of Mu.W, Mu.B, Sigma.W, Sigma.B
# using the EM algorithm
#
# per cluster-SIZE
#
lav_mvn_cl_em_sat <- function(ylp = NULL,
                              lp = NULL,
                              tol = 1e-04, # = em.h1.args$tol
                              max_iter = 5000L, # = em.h1.args$max_iter
                              min_variance = 1e-05,
                              acceleration = "none") {
  # lavdata
  between_idx <- lp$between.idx[[2]]
  # within_idx <- lp$within.idx[[2]]
  y2 <- ylp[[2]]$Y2

  # starting values: zero means, identity covariance matrices
  ov_idx <- lp$ov.idx
  sigma_w_1 <- diag(length(ov_idx[[1]]))
  sigma_b_1 <- diag(length(ov_idx[[2]]))
  mu_w_1 <- numeric(length(ov_idx[[1]]))
  mu_b_1 <- numeric(length(ov_idx[[2]]))

  # report initial fx
  fx <- lav_mvn_cl_loglik_samp_2l(
    ylp = ylp, lp = lp,
    mu_w = mu_w_1, sigma_w = sigma_w_1,
    mu_b = mu_b_1, sigma_b = sigma_b_1,
    sinv_method = "eigen", log2pi = TRUE, minus_two = FALSE
  )

  # if verbose, report
  if (lav_verbose()) {
    cat(
      "EM iter:", sprintf("%3d", 0),
      " fx =", sprintf("%17.10f", fx),
      "\n"
    )
  }

  # translate to internal matrices
  out <- lav_mvn_cl_implied22l(
    lp = lp,
    mu_w = mu_w_1, sigma_w = sigma_w_1, mu_b = mu_b_1, sigma_b = sigma_b_1
  )
  # mu_y <- out$mu.y
  mu_z <- out$mu.z
  mu_w <- out$mu.w
  mu_b <- out$mu.b
  sigma_w <- out$sigma.w
  sigma_b <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # mu.z and sigma.zz can be computed beforehand
  if (length(between_idx) > 0L) {
    z <- y2[, between_idx, drop = FALSE]
    mu_z <- colMeans(z, na.rm = TRUE)
    sigma_zz <- cov(z, use = "pairwise.complete.obs") *
                              (lp$nclusters[[2]] - 1L) / lp$nclusters[[2]]
    # sigma.zz <- 1/Lp$nclusters[[2]] * crossprod(Z) - tcrossprod(mu.z)
    # Y1Y1 <- Y1Y1[-between.idx, -between.idx, drop=FALSE]
  }

  # pack/unpack the internal parameter vector (structural zeros stay zero
  # under linear extrapolation, so we simply pack the full matrices)
  n_w <- length(mu_w)
  n_b <- length(mu_b)
  n_z <- length(mu_z)
  em_pack <- function(mu_w, mu_b, sigma_w, sigma_b, sigma_yz) {
    c(mu_w, mu_b, lav_mat_vech(sigma_w), lav_mat_vech(sigma_b),
      as.vector(sigma_yz))
  }
  em_unpack <- function(theta) {
    i <- 0L
    mu_w <- theta[i + seq_len(n_w)]; i <- i + n_w
    mu_b <- theta[i + seq_len(n_b)]; i <- i + n_b
    sigma_w <- lav_mat_vech_rev(theta[i + seq_len(n_w * (n_w + 1) / 2)])
    i <- i + n_w * (n_w + 1) / 2
    sigma_b <- lav_mat_vech_rev(theta[i + seq_len(n_b * (n_b + 1) / 2)])
    i <- i + n_b * (n_b + 1) / 2
    sigma_yz <- matrix(theta[i + seq_len(n_w * n_z)], n_w, n_z)
    list(mu_w = mu_w, mu_b = mu_b, sigma_w = sigma_w, sigma_b = sigma_b,
         sigma_yz = sigma_yz)
  }

  # one EM step: theta -> theta' (mu.z and sigma.zz stay fixed)
  em_step <- function(theta) {
    th <- em_unpack(theta)
    estep <- lav_mvn_cl_em_estepb( # Y1 = Y1,
      ylp = ylp,
      lp = lp,
      sigma_w = th$sigma_w,
      sigma_b = th$sigma_b,
      mu_w = th$mu_w,
      mu_b = th$mu_b,
      sigma_yz = th$sigma_yz,
      sigma_zz = sigma_zz,
      mu_z = mu_z
    )

    # mstep
    sigma_w_new <- estep$sigma.w

    # check for (near-zero) variances at the within level, and set
    # them to min.variance
    zero_var <- which(diag(estep$sigma.w) < min_variance)
    if (length(zero_var) > 0L) {
      sigma_w_new[, zero_var] <- 0
      sigma_w_new[zero_var, ] <- 0
      diag(sigma_w_new)[zero_var] <- min_variance
    }

    em_pack(estep$mu.w, estep$mu.b, sigma_w_new, estep$sigma.b,
            estep$sigma.yz)
  }

  # loglikelihood at theta -- the state is already in 2l form, so we can
  # skip the 2l -> implied -> 2l round trip (mu.y = mu.w + mu.b, with
  # mu.b[within-only] structurally zero; see the header invariants)
  em_logl <- function(theta) {
    th <- em_unpack(theta)
    out <- list(
      sigma.w = th$sigma_w, sigma.b = th$sigma_b,
      sigma.zz = sigma_zz, sigma.yz = th$sigma_yz,
      mu.z = mu_z, mu.y = th$mu_w + th$mu_b,
      mu.w = th$mu_w, mu.b = th$mu_b
    )
    lav_mvn_cl_loglik_samp_2l(
      ylp = ylp, lp = lp,
      sinv_method = "eigen", log2pi = TRUE, minus_two = FALSE,
      out = out
    )
  }

  # EM iterations
  theta <- em_pack(mu_w, mu_b, sigma_w, sigma_b, sigma_yz)
  if (acceleration %in% c("squarem", "qn")) {
    accel_fn <- if (acceleration == "qn") lav_em_qn else lav_em_squarem
    out_em <- accel_fn(
      theta = theta, step_fn = em_step, logl_fn = em_logl,
      fx0 = fx, tol = tol, max_iter = max_iter
    )
    theta <- out_em$theta
    fx <- out_em$fx
    converged <- out_em$converged
  } else {
    fx_old <- fx
    converged <- FALSE
    for (i in 1:max_iter) {
      theta <- em_step(theta)
      fx <- em_logl(theta)

      # fx.delta
      fx_delta <- fx - fx_old

      # check if fx.delta is finite
      if (!is.finite(fx_delta)) {
        # not good ... something is very wrong; perhaps near-singular
        # matrices?
        th <- em_unpack(theta)
        cat("\n")
        cat("FATAL problem: dumping Sigma.W and Sigma.B matrices:\n\n")
        cat("Sigma.W:\n")
        print(th$sigma_w)
        cat("\n")
        cat("Sigma.B:\n")
        print(th$sigma_b)
        cat("\n")
        lav_msg_stop(gettext(
          "EM steps of the saturated (H1) model failed; some matrices may
           be singular; please check your data for (near-)perfect
           correlations."))
      }

      # what if fx.delta is negative?
      if (fx_delta < 0) {
        lav_msg_warn(gettext(
          "logl decreased during EM steps of the saturated (H1) model"))
      }

      if (lav_verbose()) {
        cat(
          "EM iter:", sprintf("%3d", i),
          " fx =", sprintf("%17.10f", fx),
          " fx.delta =", sprintf("%9.8f", fx_delta),
          "\n"
        )
      }

      # convergence check
      if (fx_delta < tol) {
        converged <- TRUE
        break
      } else {
        fx_old <- fx
      }
    }
  } # EM iterations

  # warning?
  if (!converged) {
    lav_msg_warn(gettext(
      "Maximum number of iterations reached when computing the sample
       moments of the saturated (H1) model using EM; increase the max_iter
       element of the em.h1.args= argument to increase the number of
       iterations"))
  }

  # final implied matrices
  th <- em_unpack(theta)
  implied2 <- lav_mvn_cl_2l2implied(
    lp = lp,
    sigma_w = th$sigma_w, sigma_b = th$sigma_b,
    sigma_zz = sigma_zz, sigma_yz = th$sigma_yz,
    mu_z = mu_z,
    mu_y = NULL, mu_w = th$mu_w, mu_b = th$mu_b
  )

  list(
    Sigma.W = implied2$Sigma.W, Sigma.B = implied2$Sigma.B,
    Mu.W = implied2$Mu.W, Mu.B = implied2$Mu.B, logl = fx
  )
}


# EM for the fitted (h0) model: E-step = lav_mvn_cl_em_estepb() (or the
# missing-data engine), M-step = fit the two-group model to the expected
# complete-data moments
lav_mvn_cl_em_h0 <- function(lavsamplestats = NULL,
                             lavdata = NULL,
                             lavimplied = NULL,
                             lavpartable = NULL,
                             lavmodel = NULL,
                             lavoptions = NULL,
                             verbose_x = FALSE,
                             fx_tol = 1e-08,
                             dx_tol = 1e-05,
                             max_iter = 5000,
                             mstep_max_iter = 10000L,
                             mstep_rel_tol = 1e-10,
                             mstep_verbose = FALSE, # = em.args$verbose
                             acceleration = "none",
                             fused = TRUE) { # = em.args$fused
  if (is.null(fused)) {
    fused <- TRUE
  }
  # single group only for now
  stopifnot(lavdata@ngroups == 1L)

  # lavdata
  lp <- lavdata@Lp[[1]] # first group only (for now)
  ov_names_l <- lavdata@ov.names.l[[1]] # first group only (for now)
  # y1 <- lavdata@X[[1]] # first group only
  ylp <- lavsamplestats@YLp[[1]] # first group only

  between_idx <- lp$between.idx[[2]]
  y2 <- ylp[[2]]$Y2

  # missing data? use the missing-data E-step machinery (the same
  # engine as the saturated h1 EM: in the E-step, we integrate over both
  # the cluster random effects and the missing values); new in 0.7-1
  missing_flag <- lavsamplestats@missing.flag
  if (missing_flag) {
    y1 <- lavdata@X[[1]] # first group only (for now)
    mp <- lavdata@Mp[[1]]
    loglik_x <- ylp[[2]]$loglik.x
    if (is.null(loglik_x)) {
      loglik_x <- 0
    }
    mi_engine <- lav_mvn_cl_mi_em_engine(
      y1 = y1, y2 = y2, lp = lp, mp = mp, loglik_x = loglik_x
    )
  }

  # initial values
  x_current <- lav_model_get_parameters(lavmodel)

  # implied
  if (is.null(lavimplied)) {
    lavimplied <- lav_model_implied(lavmodel)
  }

  # TODO: what if current 'starting' parameters imply a non-pd sigma.b?

  # report initial fx
  if (missing_flag) {
    fx <- lav_mvn_cl_mi_loglik_samp_2l(
      y1 = y1, y2 = y2, lp = lp, mp = mp,
      mu_w = lavimplied$mean[[1]], sigma_w = lavimplied$cov[[1]],
      mu_b = lavimplied$mean[[2]], sigma_b = lavimplied$cov[[2]],
      loglik_x = loglik_x, log2pi = TRUE, minus_two = FALSE
    )
  } else {
    fx <- lav_mvn_cl_loglik_samp_2l(
      ylp = ylp, lp = lp,
      mu_w = lavimplied$mean[[1]], sigma_w = lavimplied$cov[[1]],
      mu_b = lavimplied$mean[[2]], sigma_b = lavimplied$cov[[2]],
      sinv_method = "eigen", log2pi = TRUE, minus_two = FALSE
    )
  }

  # if verbose, report
  if (lav_verbose()) {
    cat(
      "EM iter:", sprintf("%3d", 0),
      " fx =", sprintf("%17.10f", fx),
      "\n"
    )
  }

  # initial state for the (complete-data) plain EM iterations below
  if (!missing_flag) {
    # translate to internal matrices
    out <- lav_mvn_cl_implied22l(
      lp = lp,
      mu_w = lavimplied$mean[[1]], sigma_w = lavimplied$cov[[1]],
      mu_b = lavimplied$mean[[2]], sigma_b = lavimplied$cov[[2]]
    )
    # mu_y <- out$mu.y
    mu_z <- out$mu.z
    mu_w <- out$mu.w
    mu_b <- out$mu.b
    sigma_w <- out$sigma.w
    sigma_b <- out$sigma.b
    sigma_zz <- out$sigma.zz
    sigma_yz <- out$sigma.yz

    # mu.z and sigma.zz can be computed beforehand
    if (length(between_idx) > 0L) {
      z <- y2[, between_idx, drop = FALSE]
      mu_z <- colMeans(y2)[between_idx]
      sigma_zz <- cov(z) * (lp$nclusters[[2]] - 1L) / lp$nclusters[[2]]
      # sigma.zz <- 1/Lp$nclusters[[2]] * crossprod(Z) - tcrossprod(mu.z)
      # Y1Y1 <- Y1Y1[-between.idx, -between.idx, drop=FALSE]
    }
  }

  # M-step model: a two-group version of the parameter table (template)
  mstep_partable <- lavpartable
  # if a group column exists, delete it (it will be overriden anyway)
  mstep_partable$group <- NULL
  level_idx <- which(names(mstep_partable) == "level")
  names(mstep_partable)[level_idx] <- "group"
  mstep_partable$est <- NULL
  mstep_partable$se <- NULL
  free_idx <- which(lavpartable$free > 0L)
  mstep_fixed_x <- any(lavpartable$exo == 1L)

  # M-step: fit the two-group model to the expected complete-data moments
  em_mstep <- function(implied, x_start) {
    local_partable <- mstep_partable
    # give current values as starting values
    local_partable$ustart[free_idx] <- x_start
    # unless em.args$verbose = TRUE, suppress the (per M-step) optimizer
    # output, even when the global verbose = TRUE (so that only the
    # EM/SQUAREM cycles are shown)
    if (!mstep_verbose) {
      current_verbose <- lav_verbose()
      if (lav_verbose(FALSE)) {
        on.exit(lav_verbose(current_verbose), TRUE)
      }
    }
    lavaan(local_partable,
      sample_cov = list(
        within = implied$Sigma.W,
        between = implied$Sigma.B
      ),
      sample_mean = list(
        within = implied$Mu.W,
        between = implied$Mu.B
      ),
      sample_nobs = lp$nclusters,
      sample.cov.rescale = FALSE,
      control = list(
        iter.max = mstep_max_iter,
        rel.tol = mstep_rel_tol
      ),
      fixed.x = mstep_fixed_x,
      estimator = "ML",
      warn = FALSE, # no warnings
      check.start = FALSE,
      check.post = FALSE,
      check.gradient = FALSE,
      check.vcov = FALSE,
      baseline = FALSE,
      fit.by.level = FALSE,
      h1 = FALSE,
      se = "none",
      test = "none"
    )
  }

  # one EM step as a map in the model parameter space: x -> x'
  # (E-step at the model-implied moments of x, then the M-step fit);
  # if logl = TRUE (missing data only), the observed-data loglikelihood
  # at the *input* x is attached as the "logl" attribute (a byproduct
  # of the E-step)
  em_step <- function(x, logl = FALSE) {
    lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
    implied_x <- lav_model_implied(lavmodel2)
    out2 <- lav_mvn_cl_implied22l(
      lp = lp,
      mu_w = implied_x$mean[[1]], sigma_w = implied_x$cov[[1]],
      mu_b = implied_x$mean[[2]], sigma_b = implied_x$cov[[2]]
    )
    if (missing_flag) {
      # E-step + saturated update = the expected complete-data moments
      theta <- mi_engine$pack(
        out2$mu.w, out2$mu.b, out2$sigma.w, out2$sigma.b,
        out2$mu.z, out2$sigma.zz, out2$sigma.yz
      )
      theta_new <- mi_engine$step(theta, logl = logl)
      implied <- mi_engine$implied(theta_new)
    } else {
      estep <- lav_mvn_cl_em_estepb(
        ylp = ylp,
        lp = lp,
        sigma_w = out2$sigma.w,
        sigma_b = out2$sigma.b,
        mu_w = out2$mu.w,
        mu_b = out2$mu.b,
        sigma_yz = out2$sigma.yz,
        sigma_zz = out2$sigma.zz,
        mu_z = out2$mu.z
      )
      implied <- lav_mvn_cl_2l2implied(
        lp = lp,
        sigma_w = estep$sigma.w, sigma_b = estep$sigma.b,
        sigma_zz = out2$sigma.zz, sigma_yz = estep$sigma.yz,
        mu_z = out2$mu.z,
        mu_y = NULL, mu_w = estep$mu.w, mu_b = estep$mu.b
      )
    }
    rownames(implied$Sigma.W) <- ov_names_l[[1]]
    rownames(implied$Sigma.B) <- ov_names_l[[2]]
    local_fit <- em_mstep(implied, x_start = x)
    x_new <- local_fit@optim$x
    if (logl && missing_flag) {
      attr(x_new, "logl") <- attr(theta_new, "logl")
    }
    x_new
  }

  # loglikelihood at x
  em_logl <- function(x) {
    lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
    implied_x <- lav_model_implied(lavmodel2)
    if (missing_flag) {
      lav_mvn_cl_mi_loglik_samp_2l(
        y1 = y1, y2 = y2, lp = lp, mp = mp,
        mu_w = implied_x$mean[[1]], sigma_w = implied_x$cov[[1]],
        mu_b = implied_x$mean[[2]], sigma_b = implied_x$cov[[2]],
        loglik_x = loglik_x, log2pi = TRUE, minus_two = FALSE
      )
    } else {
      lav_mvn_cl_loglik_samp_2l(
        ylp = ylp,
        lp = lp, mu_w = implied_x$mean[[1]], sigma_w = implied_x$cov[[1]],
        mu_b = implied_x$mean[[2]], sigma_b = implied_x$cov[[2]],
        sinv_method = "eigen", log2pi = TRUE, minus_two = FALSE
      )
    }
  }

  if (acceleration %in% c("squarem", "qn")) {
    # same dual stopping rule as the plain EM iterations below: the
    # (signed!) change in the loglikelihood is smaller than fx_tol
    # (this also stops when the -- inexact -- M-step no longer makes
    # progress), or the gradient is small
    em_conv <- function(theta_old, theta, fx_old, fx) {
      if ((fx - fx_old) < fx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: fx.delta < ", fx_tol, "\n")
        }
        return(TRUE)
      }
      lavmodel2 <- lav_model_set_parameters(lavmodel, x = theta)
      dx <- lav_model_grad(lavmodel2,
        lavdata = lavdata,
        lavsamplestats = lavsamplestats
      )
      if (max(abs(dx)) < dx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: max.dx < ", dx_tol, "\n")
        }
        return(TRUE)
      }
      FALSE
    }
    accel_fn <- if (acceleration == "qn") lav_em_qn else lav_em_squarem
    out_em <- accel_fn(
      theta = x_current, step_fn = em_step, logl_fn = em_logl,
      fx0 = fx, conv_fn = em_conv, max_iter = max_iter,
      logl_dec = 0 # inexact M-step: only accept monotone extrapolations
    )
    x <- out_em$theta
    fx <- out_em$fx
    # a stall (the M-step no longer improves the loglikelihood) counts as
    # convergence, just like the signed fx.delta check in the plain
    # iterations below
    em_converged <- out_em$converged || out_em$stalled
    em_iterations <- out_em$fpeval
  } else if (missing_flag) {
    # plain EM iterations (missing data) -- driven by the em_step() map;
    # if fused = TRUE, the loglikelihood is obtained as a byproduct of
    # the E-step (looking one EM step ahead), avoiding the separate
    # em_logl() evaluations; same iterates, same stopping rule
    fx_old <- fx
    x <- x_current
    if (fused) {
      x_next <- em_step(x, logl = TRUE)
    }
    for (i in 1:max_iter) {
      if (fused) {
        x <- x_next
        attr(x, "logl") <- NULL
        x_next <- em_step(x, logl = TRUE)
        fx <- attr(x_next, "logl") # logl at the current x
      } else {
        x <- em_step(x)
        fx <- em_logl(x)
      }

      # fx.delta
      fx_delta <- fx - fx_old

      # derivatives
      lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
      dx <- lav_model_grad(lavmodel2,
        lavdata = lavdata,
        lavsamplestats = lavsamplestats
      )
      max_dx <- max(abs(dx))

      if (lav_verbose()) {
        cat(
          "EM iter:", sprintf("%3d", i),
          " fx =", sprintf("%17.10f", fx),
          " fx.delta =", sprintf("%9.8f", fx_delta),
          " max.dx = ", sprintf("%9.8f", max_dx),
          "\n"
        )
      }

      # stopping rule check
      if (fx_delta < fx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: fx.delta < ", fx_tol, "\n")
        }
        break
      } else {
        fx_old <- fx
        if (verbose_x) {
          print(round(x, 3))
        }
      }

      # second stopping rule check -- derivatives
      if (max_dx < dx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: max.dx < ", dx_tol, "\n")
        }
        break
      }
    } # EM iterations

    em_converged <- i < max_iter
    em_iterations <- i
  } else {
    # EM iterations
    fx_old <- fx
    for (i in 1:max_iter) {
      # E-step
      estep <- lav_mvn_cl_em_estepb(
        ylp = ylp,
        lp = lp,
        sigma_w = sigma_w,
        sigma_b = sigma_b,
        mu_w = mu_w,
        mu_b = mu_b,
        sigma_yz = sigma_yz,
        sigma_zz = sigma_zz,
        mu_z = mu_z
      )

      # back to model-implied dimensions
      implied <- lav_mvn_cl_2l2implied(
        lp = lp,
        sigma_w = estep$sigma.w, sigma_b = estep$sigma.b,
        sigma_zz = sigma_zz, sigma_yz = estep$sigma.yz,
        mu_z = mu_z,
        mu_y = NULL, mu_w = estep$mu.w, mu_b = estep$mu.b
      )
      rownames(implied$Sigma.W) <- ov_names_l[[1]]
      rownames(implied$Sigma.B) <- ov_names_l[[2]]

      # M-step
      local_fit <- em_mstep(implied, x_start = x_current)

      implied2 <- local_fit@implied
      # translate to internal matrices -- once; reused by the loglik below
      # and as the state for the next E-step
      out <- lav_mvn_cl_implied22l(
        lp = lp,
        mu_w = implied2$mean[[1]], sigma_w = implied2$cov[[1]],
        mu_b = implied2$mean[[2]], sigma_b = implied2$cov[[2]]
      )
      fx <- lav_mvn_cl_loglik_samp_2l(
        ylp = ylp, lp = lp,
        sinv_method = "eigen", log2pi = TRUE, minus_two = FALSE,
        out = out
      )

      # fx.delta
      fx_delta <- fx - fx_old

      # derivatives
      lavmodel <- lav_model_set_parameters(lavmodel, x = local_fit@optim$x)
      dx <- lav_model_grad(lavmodel,
        lavdata = lavdata,
        lavsamplestats = lavsamplestats
      )
      max_dx <- max(abs(dx))

      if (lav_verbose()) {
        cat(
          "EM iter:", sprintf("%3d", i),
          " fx =", sprintf("%17.10f", fx),
          " fx.delta =", sprintf("%9.8f", fx_delta),
          " mstep.iter =", sprintf(
            "%3d",
            lavInspect(local_fit, "iterations")
          ),
          " max.dx = ", sprintf("%9.8f", max_dx),
          "\n"
        )
      }

      # stopping rule check
      if (fx_delta < fx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: fx.delta < ", fx_tol, "\n")
        }
        break
      } else {
        fx_old <- fx
        x_current <- local_fit@optim$x
        if (verbose_x) {
          print(round(x_current, 3))
        }
      }

      # second stopping rule check -- derivatives
      if (max_dx < dx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: max.dx < ", dx_tol, "\n")
        }
        break
      }

      # state for the next E-step (converted above, right after the M-step)
      mu_z <- out$mu.z
      mu_w <- out$mu.w
      mu_b <- out$mu.b
      sigma_w <- out$sigma.w
      sigma_b <- out$sigma.b
      sigma_zz <- out$sigma.zz
      sigma_yz <- out$sigma.yz
    } # EM iterations

    x <- local_fit@optim$x
    em_converged <- i < max_iter
    em_iterations <- i
  }

  # add attributes
  if (em_converged) {
    attr(x, "converged") <- TRUE
    attr(x, "warn.txt") <- ""
  } else {
    attr(x, "converged") <- FALSE
    attr(x, "warn.txt") <- paste("maximum number of iterations (",
      max_iter, ") ",
      "was reached without convergence.\n",
      sep = ""
    )
  }
  attr(x, "iterations") <- em_iterations
  attr(x, "control") <- list(
    em.args = list(
      max_iter = max_iter,
      fx_tol = fx_tol,
      dx_tol = dx_tol,
      acceleration = acceleration
    )
  )
  attr(fx, "fx.group") <- fx # single group for now
  attr(x, "fx") <- fx

  x
}

# get the random effects (here: expected values for cluster means)
# and optionally a standard error
lav_mvn_cl_em_estep_ranef <- function(ylp = NULL,
                                      lp = NULL,
                                      sigma_w = NULL,
                                      sigma_b = NULL,
                                      sigma_yz = NULL,
                                      sigma_zz = NULL,
                                      mu_z = NULL,
                                      mu_w = NULL,
                                      mu_b = NULL,
                                      se = FALSE) {
  # sample stats
  # nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]
  between_idx <- lp$between.idx[[2]]

  y2 <- ylp[[2]]$Y2

  nvar_y <- ncol(sigma_w)
  # nvar_z <- ncol(sigma_zz)

  mb_j <- matrix(0, nrow = nclusters, ncol = nvar_y)
  se_j <- matrix(0, nrow = nclusters, ncol = nvar_y)

  mu_y <- mu_w + mu_b

  if (length(between_idx) > 0L) {
    sigma_1 <- cbind(sigma_yz, sigma_b)
    mu <- c(mu_z, mu_y)
  } else {
    sigma_1 <- sigma_b
    mu <- mu_y
  }

  # E-step
  for (cl in seq_len(nclusters)) {
    nj <- cluster_size[cl]

    # data
    if (length(between_idx) > 0L) {
      # z comes first!
      b_j <- c(
        y2[cl, between_idx],
        y2[cl, -between_idx]
      )
      # ybar_j <- y2[cl, -between_idx]
    } else {
      b_j <- y2[cl, ]
    }

    sigma_j <- sigma_w + nj * sigma_b
    if (length(between_idx) > 0L) {
      omega_j <- rbind(
        cbind(sigma_zz, t(sigma_yz)),
        cbind(sigma_yz, 1 / nj * sigma_j)
      )
    } else {
      omega_j <- 1 / nj * sigma_j
    }
    omega_j_inv <- solve(omega_j)

    # E(v|y)
    ev <- as.numeric(mu_b + (sigma_1 %*% omega_j_inv %*% (b_j - mu)))
    mb_j[cl, ] <- ev

    if (se) {
      # Cov(v|y)
      covv <- sigma_b - (sigma_1 %*% omega_j_inv %*% t(sigma_1))

      # force symmetry
      covv <- (covv + t(covv)) / 2

      covv_diag <- diag(covv)
      nonzero_idx <- which(covv_diag > 0)

      se_j[cl, ] <- numeric(length(covv_diag))
      se_j[cl, nonzero_idx] <- sqrt(covv_diag[nonzero_idx])
    }
  }

  if (se) {
    attr(mb_j, "se") <- se_j
  }

  mb_j
}

# per cluster
# NOTE: this is the reference implementation of lav_mvn_cl_em_estepb()
#       (which loops over the cluster SIZES and is the one actually used
#        by lav_mvn_cl_em_sat/lav_mvn_cl_em_h0); it is kept for
#       verification purposes, but no longer called anywhere
lav_mvn_cl_em_estep <- function( # Y1           = NULL,
                                ylp = NULL,
                                lp = NULL,
                                sigma_w = NULL,
                                sigma_b = NULL,
                                sigma_yz = NULL,
                                sigma_zz = NULL,
                                mu_z = NULL,
                                mu_w = NULL,
                                mu_b = NULL) {
  # sample stats
  nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]
  # cluster_idx <- lp$cluster.idx[[2]]
  # within_idx <- lp$within.idx[[2]]
  between_idx <- lp$between.idx[[2]]
  # both_idx <- lp$both.idx[[2]]

  y2 <- ylp[[2]]$Y2
  y1y1 <- ylp[[2]]$Y1Y1

  nvar_y <- ncol(sigma_w)
  nvar_z <- ncol(sigma_zz)

  cw2_j <- matrix(0, nrow = nvar_y, ncol = nvar_y)
  cb_j <- matrix(0, nrow = nvar_y, ncol = nvar_y)
  mw_j <- matrix(0, nrow = nclusters, ncol = nvar_y)
  mb_j <- matrix(0, nrow = nclusters, ncol = nvar_y)
  zy_j <- matrix(0, nrow = nvar_z, ncol = nvar_y)

  mu_y <- mu_w + mu_b

  if (length(between_idx) > 0L) {
    sigma_1 <- cbind(sigma_yz, sigma_b)
    mu <- c(mu_z, mu_y)
    y1y1 <- y1y1[-between_idx, -between_idx, drop = FALSE]
  } else {
    sigma_1 <- sigma_b
    mu <- mu_y
  }

  # E-step
  for (cl in seq_len(nclusters)) {
    nj <- cluster_size[cl]

    # data
    if (length(between_idx) > 0L) {
      # z comes first!
      b_j <- c(
        y2[cl, between_idx],
        y2[cl, -between_idx]
      )
      ybar_j <- y2[cl, -between_idx]
    } else {
      ybar_j <- b_j <- y2[cl, ]
    }

    sigma_j <- sigma_w + nj * sigma_b
    if (length(between_idx) > 0L) {
      omega_j <- rbind(
        cbind(sigma_zz, t(sigma_yz)),
        cbind(sigma_yz, 1 / nj * sigma_j)
      )
    } else {
      omega_j <- 1 / nj * sigma_j
    }
    omega_j_inv <- solve(omega_j)

    # E(v|y)
    ev <- as.numeric(mu_b + (sigma_1 %*% omega_j_inv %*% (b_j - mu)))

    # Cov(v|y)
    covv <- sigma_b - (sigma_1 %*% omega_j_inv %*% t(sigma_1))

    # force symmetry
    covv <- (covv + t(covv)) / 2

    # E(vv|y) = Cov(v|y) + E(v|y)E(v|y)^T
    evv <- covv + tcrossprod(ev)

    # store for this cluster
    mw_j[cl, ] <- ybar_j - ev
    mb_j[cl, ] <- ev
    cw2_j <- cw2_j + nj * (evv - tcrossprod(ybar_j, ev)
      - tcrossprod(ev, ybar_j))
    cb_j <- cb_j + evv

    # between only
    if (length(between_idx) > 0L) {
      zy_j <- zy_j + tcrossprod(y2[cl, between_idx], ev)
    }
  }

  m_w <- 1 / nobs * colSums(mw_j * cluster_size)
  m_b <- 1 / nclusters * colSums(mb_j)
  c_b <- 1 / nclusters * cb_j
  c_w <- 1 / nobs * (y1y1 + cw2_j)
  # end of E-step

  # make symmetric (not needed here?)
  # C.b <- (C.b + t(C.b))/2
  # C.w <- (C.w + t(C.w))/2

  # between only
  if (length(between_idx) > 0L) {
    a <- 1 / nclusters * zy_j - tcrossprod(mu_z, m_b)
  }

  sigma_w <- c_w - tcrossprod(m_w)
  sigma_b <- c_b - tcrossprod(m_b)
  mu_w <- m_w
  mu_b <- m_b

  if (length(between_idx) > 0L) {
    sigma_yz <- t(a)
  }

  list(
    sigma.w = sigma_w, sigma.b = sigma_b, mu.w = mu_w, mu.b = mu_b,
    sigma.yz = sigma_yz, sigma.zz = sigma_zz, mu.z = mu_z
  )
}

# per cluster SIZE
lav_mvn_cl_em_estepb <- function( # Y1           = NULL, # not used!
                                 ylp = NULL,
                                 lp = NULL,
                                 sigma_w = NULL,
                                 sigma_b = NULL,
                                 sigma_yz = NULL,
                                 sigma_zz = NULL,
                                 mu_z = NULL,
                                 mu_w = NULL,
                                 mu_b = NULL) {
  # sample stats
  nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]
  # cluster_idx <- lp$cluster.idx[[2]]
  between_idx <- lp$between.idx[[2]]
  cluster_sizes <- lp$cluster.sizes[[2]]
  ncluster_sizes <- lp$ncluster.sizes[[2]]
  n_s <- lp$cluster.size.ns[[2]]

  y2 <- ylp[[2]]$Y2
  y1y1 <- ylp[[2]]$Y1Y1

  nvar_y <- ncol(sigma_w)
  nvar_z <- ncol(sigma_zz)

  mu_y <- mu_w + mu_b

  if (length(between_idx) > 0L) {
    sigma_1 <- cbind(sigma_yz, sigma_b)
    mu <- c(mu_z, mu_y)
    y1y1 <- y1y1[-between_idx, -between_idx, drop = FALSE]
  } else {
    sigma_1 <- sigma_b
    mu <- mu_y
  }

  # per cluster SIZE
  cw2_s <- matrix(0, nrow = nvar_y, ncol = nvar_y)
  cb_s <- matrix(0, nrow = nvar_y, ncol = nvar_y)
  mw_s <- matrix(0, nrow = ncluster_sizes, ncol = nvar_y)
  mb_s <- matrix(0, nrow = ncluster_sizes, ncol = nvar_y)
  zy_s <- matrix(0, nvar_z, nvar_y)

  # E-step
  for (clz in seq_len(ncluster_sizes)) {
    # cluster size
    nj <- cluster_sizes[clz]

    # data
    if (length(between_idx) > 0L) {
      # z comes first!
      b_j <- cbind(
        y2[cluster_size == nj, between_idx, drop = FALSE],
        y2[cluster_size == nj, -between_idx, drop = FALSE]
      )
      ybar_j <- y2[cluster_size == nj, -between_idx, drop = FALSE]
    } else {
      ybar_j <- b_j <- y2[cluster_size == nj, , drop = FALSE]
    }

    sigma_j <- sigma_w + nj * sigma_b
    if (length(between_idx) > 0L) {
      omega_j <- rbind(
        cbind(sigma_zz, t(sigma_yz)),
        cbind(sigma_yz, 1 / nj * sigma_j)
      )
    } else {
      omega_j <- 1 / nj * sigma_j
    }
    omega_j_inv <- solve(omega_j)
    sigma_1_j_inv <- sigma_1 %*% omega_j_inv

    # E(v|y)
    b_jc <- t(t(b_j) - mu)
    tmp <- b_jc %*% t(sigma_1_j_inv)
    ev <- t(t(tmp) + mu_b)

    # Cov(v|y)
    covv <- n_s[clz] * (sigma_b - (sigma_1_j_inv %*% t(sigma_1)))

    # force symmetry
    covv <- (covv + t(covv)) / 2

    # E(vv|y) = Cov(v|y) + E(v|y)E(v|y)^T
    evv <- covv + crossprod(ev)

    # store for this cluster SIZE
    mw_s[clz, ] <- nj * colSums(ybar_j - ev)
    mb_s[clz, ] <- colSums(ev)
    cw2_s <- cw2_s + nj * (evv - crossprod(ybar_j, ev)
      - crossprod(ev, ybar_j))
    cb_s <- cb_s + evv

    # between only
    if (length(between_idx) > 0L) {
      zy_s <- zy_s + crossprod(y2[cluster_size == nj, between_idx,
        drop = FALSE
      ], ev)
    }
  } # cluster-sizes

  m_ws <- 1 / nobs * colSums(mw_s)
  m_bs <- 1 / nclusters * colSums(mb_s)
  c_bs <- 1 / nclusters * cb_s
  c_ws <- 1 / nobs * (y1y1 + cw2_s)

  # between only
  if (length(between_idx) > 0L) {
    as_1 <- 1 / nclusters * zy_s - tcrossprod(mu_z, m_bs)
  }

  sigma_w <- c_ws - tcrossprod(m_ws)
  sigma_b <- c_bs - tcrossprod(m_bs)
  mu_w <- m_ws
  mu_b <- m_bs
  if (length(between_idx) > 0L) {
    sigma_yz <- t(as_1)
  }

  list(
    sigma.w = sigma_w, sigma.b = sigma_b, mu.w = mu_w, mu.b = mu_b,
    sigma.yz = sigma_yz, sigma.zz = sigma_zz, mu.z = mu_z
  )
}

# the Multivariate normal distribution, unrestricted (h1), missing values

# 1) loglikelihood --> same as h0 but where Mu and Sigma are unrestricted
# 2) 3) 4) 5)      --> (idem)

# YR 26 Mar 2016: first version
# YR 20 Jan 2017: added _h1_omega_sw()

# here, we estimate Mu and Sigma from Y with missing values, assuming normality
# this is a rewrite of the 'estimate.moments.EM' function in <= 0.5-22
lav_mvn_mi_h1_est_moments <- function(y = NULL,
                                                   mp = NULL,
                                                   yp = NULL,
                                                   wt = NULL,
                                                   sinv_method = "eigen",
                                                   max_iter = 500L,
                                                   tol = 1e-05,
                                                   non_pd_action = "warn",
                                                   non_pd_tol = 1e-05,
                                                   acceleration = "none") {
  # check input
  y <- as.matrix(y)
  p <- NCOL(y)
  if (!is.null(wt)) {
    n <- sum(wt)
  } else {
    n <- NROW(y)
  }

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }
  if (is.null(yp)) {
    yp <- lav_samp_mi_patterns(y = y, mp = mp, wt = wt)
  }

  # covariances with zero coverage (perhaps planned?)
  zero_coverage_flag <- FALSE
  zero_coverage_idx <- which(mp$coverage == 0) # as a vector
  if (length(zero_coverage_idx) > 0L) {
    zero_coverage_flag <- TRUE
    # this is a problem; the current implementation of the E-step does not
    # take this into account; for now, we need to switch to nlminb() instead
    lav_msg_warn(
      gettext("some covariances have zero coverage; the current implementation
               of the EM algorithm will ignore this!"))
  }

  if (is.null(max_iter)) {
    max_iter <- 500L
  }
  if (is.null(tol)) {
    tol <- 1e-05
  }
  if (is.null(non_pd_action)) {
    non_pd_action <- "warn" # backwards compatibility
  }
  if (is.null(non_pd_tol)) {
    non_pd_tol <- 1e-05
  }
  if (is.null(acceleration)) {
    acceleration <- "none" # backwards compatibility
  }

  # remove empty cases
  n_full <- n
  if (length(mp$empty.idx) > 0L) {
    if (!is.null(wt)) {
      n <- n - sum(wt[mp$empty.idx])
    } else {
      n <- n - length(mp$empty.idx)
    }
  }

  # verbose?
  if (lav_verbose()) {
    cat("\n")
    cat("lav_mvn_mi_h1_est_moments: start EM steps\n")
  }

  # starting values; zero covariances to guarantee a pd matrix
  if (!is.null(wt)) {
    tmp <- na.omit(cbind(wt, y))
    if (nrow(tmp) > 2L) {
      y_tmp <- tmp[, -1, drop = FALSE]
      wt_tmp <- tmp[, 1]
      out <- stats::cov.wt(y_tmp, wt = wt_tmp, method = "ML")
      mu0 <- out$center
      var0 <- diag(out$cov)
    } else {
      mu0 <- base::.colMeans(y, m = n_full, n = p, na.rm = TRUE)
      yc <- t(t(y) - mu0)
      var0 <- base::.colMeans(yc * yc, m = n_full, n = p, na.rm = TRUE)
    }
  } else {
    mu0 <- base::.colMeans(y, m = n_full, n = p, na.rm = TRUE)
    yc <- t(t(y) - mu0)
    var0 <- base::.colMeans(yc * yc, m = n_full, n = p, na.rm = TRUE)
  }
  # sanity check
  bad_idx <- which(!is.finite(var0) | var0 == 0)
  if (length(bad_idx) > 0L) {
    var0[bad_idx] <- 1
  }
  bad_idx <- which(!is.finite(mu0))
  if (length(bad_idx) > 0L) {
    mu0[bad_idx] <- 0
  }
  sigma0 <- diag(x = var0, nrow = p)
  mu <- mu0
  sigma_1 <- sigma0

  # report
  if (lav_verbose()) {
    # fx0 <- lav_model_objective_fiml(Sigma.hat=Sigma, Mu.hat=Mu, M=Yp)
    fx0 <- lav_mvn_mi_loglik_samp(
      yp = yp,
      mu = mu, sigma_1 = sigma_1,
      log2pi = FALSE,
      minus_two = TRUE
    ) / n
    cat(
      "  EM iteration:", sprintf("%4d", 0),
      " fx = ", sprintf("%15.10f", fx0),
      "\n"
    )
  }

  # one EM step: theta -> theta'
  em_step <- function(theta) {
    mu <- theta[seq_len(p)]
    sigma_1 <- lav_mat_vech_rev(theta[-seq_len(p)])

    # E-step
    estep <- lav_mvn_mi_estep(
      y = y, mp = mp, wt = wt,
      mu = mu, sigma_1 = sigma_1,
      sinv_method = sinv_method
    )
    t1 <- estep$T1
    t2 <- estep$T2

    # M-step
    mu <- t1 / n
    sigma_1 <- t2 / n - tcrossprod(mu)

    # check if Sigma is near-pd (+ poor fix)
    ev <- eigen(sigma_1, symmetric = TRUE, only.values = TRUE)
    evtol <- 1e-6 # FIXME!
    if (any(ev$values < evtol)) {
      # too.small <- which( ev$values < tol )
      # ev$values[too.small] <- tol
      # ev$values <- ev$values + tol
      # Sigma <- ev$vectors %*% diag(ev$values) %*% t(ev$vectors)

      # ridge
      diag(sigma_1) <- diag(sigma_1) + max(diag(sigma_1)) * 1e-08
    }

    c(mu, lav_mat_vech(sigma_1))
  }

  # loglikelihood at theta
  em_logl <- function(theta) {
    lav_mvn_mi_loglik_samp(
      yp = yp,
      mu = theta[seq_len(p)],
      sigma_1 = lav_mat_vech_rev(theta[-seq_len(p)]),
      log2pi = FALSE,
      minus_two = FALSE
    )
  }

  # EM steps
  converged <- FALSE
  if (acceleration %in% c("squarem", "qn")) {
    accel_fn <- if (acceleration == "qn") lav_em_qn else lav_em_squarem
    out_em <- accel_fn(
      theta = c(mu, lav_mat_vech(sigma_1)),
      step_fn = em_step, logl_fn = em_logl,
      conv = "param", # same criterion as the plain EM steps below
      tol = tol, max_iter = max_iter
    )
    mu <- out_em$theta[seq_len(p)]
    sigma_1 <- lav_mat_vech_rev(out_em$theta[-seq_len(p)])
    converged <- out_em$converged
  } else {
    theta_old <- c(mu, lav_mat_vech(sigma_1))
    for (i in 1:max_iter) {
      theta <- em_step(theta_old)
      mu <- theta[seq_len(p)]
      sigma_1 <- lav_mat_vech_rev(theta[-seq_len(p)])

      # max absolute difference in parameter values
      mm_delta <- max(abs(theta - theta_old))

      # report fx
      if (lav_verbose()) {
        # fx <- lav_model_objective_fiml(Sigma.hat=Sigma, Mu.hat=Mu, M=Yp)
        fx <- lav_mvn_mi_loglik_samp(
          yp = yp,
          mu = mu, sigma_1 = sigma_1,
          log2pi = FALSE,
          minus_two = TRUE
        ) / n
        cat(
          "  EM iteration:", sprintf("%4d", i),
          " fx = ", sprintf("%15.10f", fx),
          " delta par = ", sprintf("%9.8f", mm_delta),
          "\n"
        )
      }

      # convergence check: using parameter values:
      if (mm_delta < tol) {
        converged <- TRUE
        break
      }

      # again
      theta_old <- theta
    }
  } # EM iterations

  if (lav_verbose()) {
    cat("\nSigma:\n")
    print(sigma_1)
    cat("\nMu:\n")
    print(mu)
    cat("\n")
  }

  # compute fx if we haven't already
  if (acceleration %in% c("squarem", "qn") || !lav_verbose()) {
    # fx <- lav_model_objective_fiml(Sigma.hat = Sigma, Mu.hat = Mu, M = Yp)
    fx <- lav_mvn_mi_loglik_samp(
      yp = yp,
      mu = mu, sigma_1 = sigma_1,
      log2pi = FALSE,
      minus_two = TRUE
    ) / n
  }

  # warning?
  if (!converged) {
    lav_msg_warn(
      gettext("Maximum number of iterations reached when computing the sample
              moments using EM; increase the max_iter element of the
              em.h1.args= argument to increase the number of iterations")
    )
  }

  # check if the EM estimated variance-covariance matrix is (near) singular
  if (non_pd_action != "none") {
    ev <- eigen(sigma_1, symmetric = TRUE, only.values = TRUE)$values
    if (any(ev < non_pd_tol)) {
      near_zero_idx <- which(diag(sigma_1) <= non_pd_tol)
      if (length(near_zero_idx) > 0L) {
        var_names <- colnames(y)
        if (is.null(var_names)) {
          var_names <- paste0("V", seq_len(NCOL(y)))
        }
        msg <- gettextf(
          "The EM estimated variance-covariance matrix (Sigma) is (near)
           singular: the variances of some variables are (near) zero: %s.",
          lav_msg_view(var_names[near_zero_idx]))
      } else {
        msg <- gettextf(
          "The EM estimated variance-covariance matrix (Sigma) is (near)
           singular: its smallest eigenvalue is smaller than %s.",
          format(non_pd_tol))
      }
      if (non_pd_action == "stop") {
        lav_msg_stop(paste(msg, gettext(
          "Set the non_pd_action element of the em.h1.args= argument to
           \"warn\" or \"none\" to continue anyway.")))
      } else {
        lav_msg_warn(paste(msg, gettext(
          "This may cause numerical instabilities; interpret the results
           with caution.")))
      }
    }
  }

  list(Sigma = sigma_1, Mu = mu, fx = fx)
}

# if we cannot use the EM algorithm (zero coverage?), we can always use
# plain FIML and nlminb() instead
#
# single level only
lav_mvn_mi_h1_est_moments_chol <- function(lavdata = NULL,
                                                        lavsamplestats = NULL,
                                                        lavoptions = NULL,
                                                        group = 1L) {
  # not for multilevel
  stopifnot(lavdata@nlevels == 1L)

  # no ov.names.x in lavdata
  lavdata@ov.names.x <- vector("list", length = lavdata@ngroups)

  # construct unrestricted partable (using chol parameterization)
  # for this group only
  lavpartable <- lav_pt_unrestricted_chol(
                             lavdata = lavdata, lavoptions = lavoptions,
                             lavpta = NULL, group = group)

  lavoptions2 <- lavoptions
  lavoptions2$estimator <- "ML"
  lavoptions2$missing <- "ml"
  lavoptions2$se <- "none"
  lavoptions2$test <- "none"
  lavoptions2$do.fit <- TRUE
  lavoptions2$optim.method <- "nlminb"
  lavoptions2$h1 <- FALSE
  lavoptions2$implied <- TRUE
  lavoptions2$loglik <- TRUE
  lavoptions2$baseline <- FALSE
  lavoptions2$fit.by.level <- FALSE
  lavoptions2$fixed.x <- FALSE # even if model uses fixed.x=TRUE
  lavoptions2$model.type <- "unrestricted"
  lavoptions2$optim.attempts <- 4L
  lavoptions2$check.gradient <- FALSE
  lavoptions2$optim.force.converged <- TRUE # for now...
  lavoptions2$control <- list(rel.tol = 1e-7)
  lavoptions2$start <- "simple" # add this point, we have no lavh1 yet!
  fit <- lavaan(lavpartable,
    slot_options = lavoptions2,
    slot_sample_stats = lavsamplestats,
    slot_data = lavdata,
    warn = FALSE
  )

  out <- list(Sigma = fit@implied$cov[[1]],
              Mu = fit@implied$mean[[1]],
              fx = fit@optim$fx)

  out
}

# compute N times ACOV(Mu, vech(Sigma))
# in the literature: - `Omega_{SW}'
#                    - `Gamma for incomplete data'
#                    - (N times the) sandwich estimator for acov(mu,vech(Sigma))
lav_mvn_mi_h1_omega_sw <- function(y = NULL,
                                           mp = NULL,
                                           wt = NULL,
                                           cluster_idx = NULL,
                                           yp = NULL,
                                           sinv_method = "eigen",
                                           mu = NULL,
                                           sigma_1 = NULL,
                                           x_idx = integer(0L),
                                           sigma_inv = NULL,
                                           information = "observed") {
  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # sample stats per pattern
  if (is.null(yp) && (information == "observed" || is.null(sigma_1))) {
    yp <- lav_samp_mi_patterns(y = y, mp = mp, wt = wt)
  }

  # Sigma and Mu
  if (is.null(sigma_1) || is.null(mu)) {
    out <- lav_mvn_mi_h1_est_moments(y = y, mp = mp, yp = yp)
    mu <- out$Mu
    sigma_1 <- out$Sigma
  }

  # information matrices
  info <- lav_mvn_mi_info_both(
    y = y, mp = mp, mu = mu,
    wt = wt, cluster_idx = cluster_idx,
    sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
    sigma_inv = sigma_inv, information = information
  )

  a <- info$Abeta
  a_inv <- lav_mat_sym_inverse(
    s = a, logdet = FALSE,
    sinv_method = sinv_method
  )
  m_b <- info$Bbeta

  # sandwich
  sw <- a_inv %*% m_b %*% a_inv

  sw
}

# casewise Mahalanobis-distance diagnostics for models with ordered
# categorical (and possibly mixed continuous/ordinal) indicators, following
# Mansolf & Reise (2017)
#
# the M-distances defined for the continuous case (the leverage distance of
# the factor scores, the residual-based distance, and the distance of the
# fitted values) are generalized to categorical data by taking their
# expectation over the region of the (latent) multivariate normal response
# vector y* that is consistent with the observed response pattern:
#
#     d*_i = E[ d(y*) | y* in R_i ],   with  y* ~ N(mu, Sigma(theta))
#
# where R_i is the rectangle defined by the model-implied thresholds and the
# observed categories of case i; observed continuous coordinates fix the
# corresponding elements of y* (so the expectation is conditional on them),
# and missing coordinates leave the corresponding region unbounded (they are
# integrated out); as a result, when ALL variables are continuous and
# observed, the region degenerates to a single point and the distances
# reduce exactly to their continuous counterparts
#
# the expectation is approximated by Monte Carlo integration, using a
# (vectorized) GHK importance sampler: no burn-in or mixing issues (as with
# a Gibbs sampler), and a deterministic computational cost
#
# YR + Claude, July 2026

# main entry: compute the expected (squared) M-distances for all cases
#
# type = "lv"    -> leverage distance of the (Bartlett/regression) factor
#                   scores (d_f* in Mansolf & Reise, 2017)
# type = "resid" -> residual-based distance (d_r*), based on the casewise
#                   residuals of the latent responses
# type = "yhat"  -> distance of the fitted (model-predicted) values
#
# returns a list with one element per group; each element is a list with
#  - d2:    the expected squared distances E[d^2] (length n)
#  - d:     the expected distances E[d] (length n); note that for the
#           Monte Carlo based values E[d] != sqrt(E[d^2])
#  - resid: (only if type = "resid") the n x p matrix of expected casewise
#           residuals of the latent responses
lav_predict_mdist_cat <- function(lavobject = NULL, # for convenience
                                  # sub objects
                                  lavmodel = NULL, lavdata = NULL,
                                  lavimplied = NULL,
                                  # optional new data
                                  data_obs = NULL, exo = NULL,
                                  # options
                                  type = "lv", method = "ML",
                                  ndraws = 2000L) {
  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavimplied <- lavobject@implied
  } else {
    stopifnot(!is.null(lavmodel), !is.null(lavdata))
  }

  if (lavdata@nlevels > 1L) {
    lav_msg_stop(gettext(
      "casewise Mahalanobis distances are not available (yet) for
      multilevel data with categorical variables."))
  }

  type <- tolower(type)
  stopifnot(type %in% c("lv", "resid", "yhat"))

  # Bartlett (ML) or regression (EB/EBM) factor-score mapping?
  bartlett <- tolower(method) %in% c("ml", "bartlett", "bartlet")

  ndraws <- as.integer(ndraws)
  if (length(ndraws) != 1L || is.na(ndraws) || ndraws < 10L) {
    lav_msg_stop(gettext("mdist.draws must be a single integer >= 10."))
  }

  # data
  if (is.null(data_obs)) {
    data_obs <- lavdata@X
  }
  if (is.null(exo)) {
    exo <- lavdata@eXo
  }

  # model-implied moments for the latent responses y*
  # (on the 'marginal' scale, i.e., including the delta scaling, so that
  #  they match the sample polychoric correlations and thresholds)
  if (is.null(lavimplied) || length(lavimplied) == 0L) {
    lavimplied <- lav_model_implied(lavmodel)
  }
  cond_x <- lavmodel@conditional.x
  if (cond_x) {
    sigma_hat <- lavimplied$res.cov
    int_hat <- lavimplied$res.int
    slopes_hat <- lavimplied$res.slopes
    th_hat <- lavimplied$res.th
  } else {
    sigma_hat <- lavimplied$cov
    int_hat <- lavimplied$mean
    th_hat <- lavimplied$th
  }

  mm_lambda <- lav_model_lambda(lavmodel = lavmodel,
    remove_dummy_lv = FALSE, use_wmat = TRUE)
  if (cond_x) {
    # all moments are conditional on the exogenous covariates, so the
    # latent (co)variances must be too: V(eta|x), not V(eta)
    # (this only affects the regression-based factor-score mapping; the
    #  Bartlett mapping does not involve the latent covariance matrix)
    veta <- lav_model_vetax(lavmodel = lavmodel)
  } else {
    veta <- lav_model_veta(lavmodel = lavmodel, remove_dummy_lv = FALSE)
  }
  mm_idx <- lav_model_group_mm_indices(lavmodel@nmat)

  out <- vector("list", length = lavdata@ngroups)
  for (g in seq_len(lavdata@ngroups)) {
    p <- ncol(sigma_hat[[g]])
    n <- nrow(data_obs[[g]])

    # delta-scale the factor loadings, so that
    # Sigma = LAMBDA VETA LAMBDA' + DELTA THETA DELTA
    # with Sigma the (delta-scaled) model-implied covariance matrix of y*
    mlist <- lavmodel@GLIST[mm_idx[[g]]]
    lambda_g <- mm_lambda[[g]]
    if (!is.null(mlist$delta)) {
      lambda_g <- mlist$delta[, 1L] * lambda_g
    }

    # the 'regular' latent variables (drop the dummy lv's, if any)
    lv_dummy_idx <- c(
      lavmodel@ov.y.dummy.lv.idx[[g]],
      lavmodel@ov.x.dummy.lv.idx[[g]]
    )
    reg_lv_idx <- seq_len(ncol(lambda_g))
    if (length(lv_dummy_idx) > 0L) {
      reg_lv_idx <- reg_lv_idx[-lv_dummy_idx]
    }

    # model-implied mean of y*: a single vector, or (if conditional.x)
    # an n x p matrix with case-specific conditional means
    if (cond_x && !is.null(exo[[g]]) && ncol(exo[[g]]) > 0L) {
      mu_g <- t(as.numeric(int_hat[[g]]) +
                slopes_hat[[g]] %*% t(exo[[g]]))
    } else if (length(int_hat[[g]]) > 0L) {
      mu_g <- as.numeric(int_hat[[g]])
    } else {
      mu_g <- numeric(p)
    }

    out[[g]] <- lav_predict_mdist_cat_group(
      x = data_obs[[g]], sigma = sigma_hat[[g]], mu = mu_g,
      lambda = lambda_g, veta = veta[[g]],
      th = th_hat[[g]], th_idx = lavmodel@th.idx[[g]],
      num_idx = lavmodel@num.idx[[g]], reg_lv_idx = reg_lv_idx,
      type = type, bartlett = bartlett, ndraws = ndraws
    )
  } # g

  out
}

# single group
lav_predict_mdist_cat_group <- function(x = NULL, sigma = NULL, mu = NULL,
                                        lambda = NULL, veta = NULL,
                                        th = NULL, th_idx = NULL,
                                        num_idx = NULL, reg_lv_idx = NULL,
                                        type = "lv", bartlett = TRUE,
                                        ndraws = 2000L) {
  n <- nrow(x)
  p <- ncol(sigma)
  resid_flag <- (type == "resid")

  # drop factors without any (direct) observed indicator (e.g. higher-order
  # factors): their column in LAMBDA is empty, so they do not span any
  # dimension of y*, and keeping them would miscount the rank of the
  # residual space below
  zero_col <- which(apply(lambda, 2L, function(v) all(v == 0)))
  if (length(zero_col) > 0L) {
    keep <- seq_len(ncol(lambda))[-zero_col]
    lambda <- lambda[, keep, drop = FALSE]
    veta <- veta[keep, keep, drop = FALSE]
    reg_lv_idx <- match(intersect(reg_lv_idx, keep), keep)
  }
  nfac <- ncol(lambda)

  sigma_inv <- lav_predict_solve(sigma)

  # factor-score coefficient matrix (Bartlett or regression), mapping the
  # centered latent responses to (centered) factor scores
  if (bartlett) {
    fsc <- lav_predict_solve(t(lambda) %*% sigma_inv %*% lambda) %*%
      t(lambda) %*% sigma_inv
  } else {
    fsc <- veta %*% t(lambda) %*% sigma_inv
  }

  # for each distance type, reduce the computation to a single quadratic
  # form: d^2(y*_c) = y*_c' Q y*_c, with y*_c the centered latent response
  # vector; this mirrors the (continuous) computations in
  # lav_predict_internal(), but pushed through to the y* metric
  proj <- NULL
  if (type == "lv") {
    fsc_r <- fsc[reg_lv_idx, , drop = FALSE]
    if (nrow(fsc_r) == 0L) {
      # no regular latent variables
      return(list(d2 = rep(as.numeric(NA), n),
                  d = rep(as.numeric(NA), n), resid = NULL))
    }
    fs_cov <- fsc_r %*% sigma %*% t(fsc_r)
    qmat <- t(fsc_r) %*% solve(fs_cov) %*% fsc_r
  } else if (type == "resid") {
    proj <- diag(p) - lambda %*% fsc
    omega_e <- proj %*% sigma %*% t(proj)
    # omega_e is rank-deficient (rank p - nfac); project on the range space
    # (Yuan & Zhong, 2008)
    eig <- eigen(omega_e, symmetric = TRUE)
    a <- eig$vectors[, seq_len(p - nfac), drop = FALSE]
    qmat <- t(proj) %*% a %*% solve(t(a) %*% omega_e %*% a) %*%
      t(a) %*% proj
  } else { # yhat
    pmat <- lambda %*% fsc
    omega_h <- pmat %*% sigma %*% t(pmat)
    eig <- eigen(omega_h, symmetric = TRUE)
    a <- eig$vectors[, seq_len(nfac), drop = FALSE]
    qmat <- t(pmat) %*% a %*% solve(t(a) %*% omega_h %*% a) %*%
      t(a) %*% pmat
  }

  # per-case integration region: for the ordinal variables, the rectangle
  # implied by the thresholds and the observed category; observed continuous
  # variables are 'fixed'; missing values are left unbounded
  lower <- matrix(-Inf, nrow = n, ncol = p)
  upper <- matrix(+Inf, nrow = n, ncol = p)
  fixed <- matrix(FALSE, nrow = n, ncol = p)
  ord_idx <- unique(th_idx[th_idx > 0L])
  for (v in ord_idx) {
    tv <- th[th_idx == v]
    lo_v <- c(-Inf, tv)
    up_v <- c(tv, +Inf)
    k <- x[, v]
    ok <- which(!is.na(k))
    lower[ok, v] <- lo_v[k[ok]]
    upper[ok, v] <- up_v[k[ok]]
  }
  for (v in num_idx) {
    fixed[, v] <- !is.na(x[, v])
  }

  # center everything at the model-implied mean
  if (is.matrix(mu)) {
    mu_case <- mu
  } else {
    mu_case <- matrix(mu, nrow = n, ncol = p, byrow = TRUE)
  }
  lower_c <- lower - mu_case
  upper_c <- upper - mu_case
  xc <- x - mu_case

  # the integral is identical for cases sharing the same response pattern
  # (and, if conditional.x, the same covariate values): compute once per
  # unique combination; key on the exact (round-tripping) hexadecimal
  # representation so that only bitwise-identical rows are merged
  key_mat <- cbind(x, mu_case)
  case_key <- do.call(paste, c(
    lapply(seq_len(ncol(key_mat)), function(j) sprintf("%a", key_mat[, j])),
    sep = "\r"
  ))
  rep_idx <- which(!duplicated(case_key))
  u_of_case <- match(case_key, case_key[rep_idx])

  # the Cholesky factor of (the permuted) Sigma only depends on which
  # coordinates are fixed; cache it per missingness pattern of the
  # continuous variables
  chol_cache <- new.env(parent = emptyenv())

  n_u <- length(rep_idx)
  d2_u <- rep(as.numeric(NA), n_u)
  d_u <- rep(as.numeric(NA), n_u)
  if (resid_flag) {
    resid_u <- matrix(as.numeric(NA), nrow = n_u, ncol = p)
  }

  for (u in seq_len(n_u)) {
    i <- rep_idx[u]

    # empty cases: no information at all
    if (all(is.na(x[i, ]))) {
      next
    }

    fix_i <- which(fixed[i, ])
    samp_i <- seq_len(p)
    if (length(fix_i) > 0L) {
      samp_i <- samp_i[-fix_i]
    }
    nf <- length(fix_i)
    ns <- length(samp_i)
    # variable order: fixed (observed continuous) coordinates first, so
    # that the GHK recursion below samples exactly from the conditional
    # distribution given the observed continuous values (their importance
    # weight contribution is constant across draws, and cancels)
    ord <- c(fix_i, samp_i)

    ckey <- paste0("f", paste(fix_i, collapse = ","))
    lmat <- chol_cache[[ckey]]
    if (is.null(lmat)) {
      lmat <- try(t(chol(sigma[ord, ord, drop = FALSE])), silent = TRUE)
      if (inherits(lmat, "try-error")) {
        lav_msg_stop(gettext(
          "the model-implied covariance matrix of the latent responses is
          not positive definite; Mahalanobis distances cannot be
          computed."))
      }
      chol_cache[[ckey]] <- lmat
    }

    # degenerate region: all coordinates observed and continuous; the
    # distances reduce to their exact continuous counterparts
    if (ns == 0L) {
      zc <- xc[i, ]
      d2_u[u] <- sum((zc %*% qmat) * zc)
      d_u[u] <- sqrt(d2_u[u])
      if (resid_flag) {
        resid_u[u, ] <- drop(proj %*% zc)
      }
      next
    }

    lo_i <- lower_c[i, ord]
    up_i <- upper_c[i, ord]

    # GHK sampler: y*_c = L eta with eta iid standard normal; sample the
    # eta's recursively from their (univariate) truncated conditionals, and
    # accumulate the importance (log-)weights
    emat <- matrix(0, nrow = ndraws, ncol = p)
    if (nf > 0L) {
      eta_f <- forwardsolve(lmat[seq_len(nf), seq_len(nf), drop = FALSE],
                            xc[i, fix_i])
      emat[, seq_len(nf)] <- matrix(eta_f, nrow = ndraws, ncol = nf,
                                    byrow = TRUE)
    }
    logw <- numeric(ndraws)
    for (j in (nf + 1L):p) {
      if (j > 1L) {
        m <- drop(emat[, seq_len(j - 1L), drop = FALSE] %*%
                  lmat[j, seq_len(j - 1L)])
      } else {
        m <- numeric(ndraws)
      }
      ljj <- lmat[j, j]
      lo_j <- (lo_i[j] - m) / ljj
      up_j <- (up_i[j] - m) / ljj
      plo <- pnorm(lo_j)
      pup <- pnorm(up_j)
      logw <- logw + log(pup - plo)
      u01 <- plo + runif(ndraws) * (pup - plo)
      # guard against numerical under/overflow in the extreme tails: keep
      # u01 strictly inside (0, 1), and the draw inside its bounds
      u01 <- pmin(pmax(u01, 1e-12), 1 - 1e-12)
      emat[, j] <- pmin(pmax(qnorm(u01), lo_j), up_j)
    }

    # normalized importance weights (log-scale to avoid underflow when the
    # response pattern has a very small probability)
    mw <- max(logw)
    if (is.finite(mw)) {
      w <- exp(logw - mw)
    } else {
      # the pattern probability underflows completely; fall back to equal
      # weights (the pattern is extremely outlying anyway)
      w <- rep(1, ndraws)
    }
    wsum <- sum(w)

    # draws of the centered y*, back-permuted to the original order
    zmat_ord <- tcrossprod(emat, lmat)
    zmat <- zmat_ord
    zmat[, ord] <- zmat_ord

    dsq <- rowSums((zmat %*% qmat) * zmat)
    d2_u[u] <- sum(w * dsq) / wsum
    d_u[u] <- sum(w * sqrt(dsq)) / wsum
    if (resid_flag) {
      resid_u[u, ] <- colSums(w * tcrossprod(zmat, proj)) / wsum
    }
  } # u

  # map the per-pattern values back to the cases
  ret <- list(
    d2 = d2_u[u_of_case],
    d = d_u[u_of_case],
    resid = NULL
  )
  if (resid_flag) {
    ret$resid <- resid_u[u_of_case, , drop = FALSE]
  }

  ret
}

# takes a model in lavaan syntax and the user's data and returns the covariance
# matrix of observed variables.  Useful so that the user can do things like
# diagnose errors in the cov matrix, use cov2cor to look at the correlation
# matrix, try and invert the sample covariance matrix, etc.

# update 5/27/2011 JEB
# changelog: using sem and inspect to get output.
#            This way, all arguments such as groups, etc, can be used
# update 3 june 2011 YR: removed se="none" (since now implied by do.fit=FALSE)
# update 13 dec 2011 YR: changed name (to avoid confusion with the
#                        model-implied cov)
lavInspectSampleCov <- function(model, data, ...) {        # nolint
  fit <- sem(model, data = data, ..., do.fit = FALSE)
  lavInspect(fit, "sampstat")
}

# return Gauss-Hermite quadrature rule for given order (n)
# return list: x = nodes, w = quadrature weights
#

# As noted by Wilf (1962, chapter 2, ex 9), the nodes are given by
# the eigenvalues of the Jacobi matrix; weights are given by the squares of the
# first components of the (normalized) eigenvectors, multiplied by sqrt(pi)
#
# (This is NOT identical to Golub & Welsch, 1968: as they used a specific
#  method tailored for tridiagonal symmetric matrices)
#
# TODO: look at internet site
#  github.com/ajt60gaibb/FastGaussQuadrature.jl/blob/master/src/gausshermite.jl
# featuring the work of Ignace Bogaert (UGent)
#
# approximation of the integral of 'f(x) * exp(-x*x)' from -inf to +inf
# by sum( f(x_i) * w_i )
#
# CHECK: sum(w_i) should be always sqrt(pi) = 1.772454
lav_integration_gauss_hermite_xw <- function(n = 21L, revert = FALSE) {  # nolint
  # force n to be an integer
  n <- as.integer(n)
  stopifnot(n > 0L)

  if (n == 1L) {
    x <- 0
    w <- sqrt(pi)
  } else {
    # construct symmetric, tridiagonal Jacobi matrix
    # diagonal = 0, -1/+1 diagonal is sqrt(1:(n-1)/2)
    u <- sqrt(seq.int(n - 1L) / 2) # upper diagonal of J
    jn <- matrix(0, n, n)
    didx <- lav_mat_diag_idx(n)
    jn[(didx + 1)[-n]] <- u
    # Jn[(didx-1)[-1]] <- u # only lower matrix is used anyway

    # eigen decomposition
    # FIXME: use specialized function for tridiagonal symmetric matrix
    ev <- eigen(jn, symmetric = TRUE)
    x <- ev$values
    tmp <- ev$vectors[1L, ]
    w <- sqrt(pi) * tmp * tmp
  }

  # revert? (minus to plus)
  if (revert) {
    x <- -x
  }

  list(x = x, w = w)
}

# generate GH points + weights
lav_integration_gauss_hermite <- function(n = 21L,
                                          dnorm = FALSE,
                                          mean = 0, sd = 1,
                                          ndim = 1L,
                                          revert = TRUE,
                                          prune = FALSE) {
  xw <- lav_integration_gauss_hermite_xw(n = n, revert = revert)

  # dnorm kernel?
  if (dnorm) {
    # scale/shift x
    x <- xw$x * sqrt(2) * sd + mean

    # scale w
    w <- xw$w / sqrt(pi)
  } else {
    x <- xw$x
    w <- xw$w
  }

  if (ndim > 1L) {
    # cartesian product
    x <- as.matrix(expand.grid(rep(list(x), ndim), KEEP.OUT.ATTRS = FALSE))
    w <- as.matrix(expand.grid(rep(list(w), ndim), KEEP.OUT.ATTRS = FALSE))
    w <- apply(w, 1, prod)
  } else {
    x <- as.matrix(x)
    w <- as.matrix(w)
  }

  # prune?
  if (is.logical(prune) && prune) {
    # always divide by N=21
    lower_limit <- xw$w[1] * xw$w[floor((n + 1) / 2)] / 21
    keep_idx <- which(w > lower_limit)
    w <- w[keep_idx]
    x <- x[keep_idx, , drop = FALSE]
  } else if (is.numeric(prune) && prune > 0) {
    lower_limit <- quantile(w, probs = prune)
    keep_idx <- which(w > lower_limit)
    w <- w[keep_idx]
    x <- x[keep_idx, , drop = FALSE]
  }

  list(x = x, w = w)
}

# old not-exported functions which are temporarily kept to enable
# publishing lavaan 0.7-1 on CRAN
lav_warning_nonexported <- function(funcname) {
  return(invisible(NULL))
  lav_msg_warn(gettextf(
    "The function %s is an internal function of lavaan and NOT meant to be
    called directly from outside lavaan. This function will be removed in
    the near future.", funcname
  ))
}

lav_partable_ngroups <- function(partable) {
  lav_warning_nonexported("lav_partable_ngroups")
  lav_pt_ngroups(partable)
}
lav_mvnorm_loglik_samplestats <- function(
  sample.mean = NULL,
  sample.cov = NULL,
  sample.nobs = NULL,
  Mu = NULL,
  Sigma = NULL,
  x.idx = integer(0L),
  x.mean = NULL,
  x.cov = NULL,
  Sinv.method = "eigen",
  Sigma.inv = NULL
) {
  lav_warning_nonexported("lav_mvnorm_loglik_samplestats")
  lav_mvn_loglik_samp(
    sample_mean = sample.mean,
    sample_cov = sample.cov,
    sample_nobs = sample.nobs,
    mu = Mu,
    sigma_1 = Sigma,
    x_idx = x.idx,
    x_mean = x.mean,
    x_cov = x.cov,
    sinv_method = Sinv.method,
    sigma_inv = Sigma.inv
  )
}
lav_mvnorm_missing_loglik_samplestats <- function(
  Yp = NULL,
  Mu = NULL,
  Sigma = NULL,
  x.idx = integer(0L),
  x.mean = NULL,
  x.cov = NULL,
  Sinv.method = "eigen",
  log2pi = TRUE,
  minus.two = FALSE
) {
  lav_warning_nonexported("lav_mvnorm_missing_loglik_samplestats")
  lav_mvn_mi_loglik_samp(
    yp = Yp,
    mu = Mu,
    sigma_1 = Sigma,
    x_idx = x.idx,
    x_mean = x.mean,
    x_cov = x.cov,
    sinv_method = Sinv.method,
    log2pi = log2pi,
    minus_two = minus.two
  )
}

lav_mvnorm_cluster_implied22l <- function(
  lp = NULL,
  implied = NULL,
  Mu.W = NULL,
  Mu.B = NULL,
  Sigma.W = NULL,
  Sigma.B = NULL
) {
  lav_warning_nonexported("lav_mvnorm_cluster_implied22l")
  lav_mvn_cl_implied22l(
    lp = lp,
    implied = implied,
    mu_w = Mu.W,
    mu_b = Mu.B,
    sigma_w = Sigma.W,
    sigma_b = Sigma.B
  )
}

lav_matrix_symmetric_inverse_update <- function(
  S.inv,
  rm.idx = integer(0L),
  logdet = FALSE,
  S.logdet = NULL
) {
  lav_warning_nonexported("lav_matrix_symmetric_inverse_update")
  lav_mat_sym_inverse_update(
    s_inv = S.inv,
    rm_idx = rm.idx,
    logdet = logdet,
    s_logdet = S.logdet
  )
}

lav_mvnorm_dmvnorm <- function(Y = NULL,
                               wt = NULL,
                               Mu = NULL,
                               Sigma = NULL,
                               Sigma.inv = NULL,
                               Sinv.method = "eigen",
                               x.idx = integer(0L),
                               x.mean = NULL,
                               x.cov = NULL,
                               log = TRUE) {
  lav_warning_nonexported("lav_matrix_symmetric_inverse_update")
  lav_mvn_dmvnorm(
    y = Y,
    wt = wt,
    mu = Mu,
    sigma_1 = Sigma,
    sigma_inv = Sigma.inv,
    sinv_method = Sinv.method,
    x_idx = x.idx,
    x_mean = x.mean,
    x_cov = x.cov,
    log = log
  )
}

lav_partable_subset_measurement_model <- function(
  PT = NULL,
  lv.names = NULL,
  add.lv.cov = TRUE,
  add.ind.predictors = FALSE,
  add.idx = FALSE,
  idx.only = FALSE
) {
  lav_warning_nonexported("lav_partable_subset_measurement_model")
  lav_pt_subset_mm(
    pt_1 = PT,
    lv_names = lv.names,
    add_lv_cov = add.lv.cov,
    add_ind_predictors = add.ind.predictors,
    add_idx = add.idx,
    idx_only = idx.only
  )
}

lav_partable_subset_structural_model <- function(
  PT = NULL,
  add.idx = FALSE,
  idx.only = FALSE,
  add.exo.cov = FALSE,
  fixed.x = FALSE,
  conditional.x = FALSE,
  free.fixed.var = FALSE,
  meanstructure = FALSE
) {
  lav_warning_nonexported("lav_partable_subset_structural_model")
  lav_pt_subset_sm(
    pt_1 = PT,
    add_idx = add.idx,
    idx_only = idx.only,
    add_exo_cov = add.exo.cov,
    fixed_x = fixed.x,
    conditional_x = conditional.x,
    free_fixed_var = free.fixed.var,
    meanstructure = meanstructure)
}

# model-implied statistics + Jacobian for TWO-LEVEL WLS with categorical
# data (the model-side counterpart of lav_samp_wls_2l_cat)
#
# YR 2026 (two-level WLS, phase 5/6)
#
# The unrestricted (stage-wise) statistics live on the scale where the
# WITHIN-level latent responses have variance 1. The structured model
# (theta parameterization) fixes the within RESIDUAL variances to 1, so
# its implied within variances sigma**_w,pp = 1 + (common part) exceed 1:
# a different scale convention. The model-implied statistics are
# therefore STANDARDIZED by the within-level scale factors
#
#     delta_p = 1 / sqrt(sigma**_w,pp)      (ordinal variables)
#     delta_p = 1                           (continuous variables)
#
# applied to the thresholds, the within covariance matrix, the between
# means AND the between covariance matrix (note: the between matrix is
# scaled by the WITHIN scale factors -- a cross-block standardization).
#
# statistic vector layout (per group; matches lav_samp_wls_2l_cat):
#   1. th: thresholds (ordinal, standardized); 0 for continuous variables
#   2. within means of the within-only (continuous) variables
#   3. within variances of the continuous variables
#   4. within covariances/'correlations' (vech, no diagonal), standardized
#   5. between means (continuous; standardized 0 for ordinal) -- between
#      variables only
#   6. vech(Sigma_b, diagonal = TRUE), standardized -- between variables
#      only
#
# The Jacobian is assembled as H %*% Delta_raw, where Delta_raw contains
# the raw (unstandardized) derivative rows
#   [ th ; mean_w ; vech(Sigma_w) ; mean_b ; vech(Sigma_b) ]
# and H is the (implied-value dependent) sparse standardization map.
# Delta_raw is obtained from the standard continuous two-level Delta
# machinery (a 'continuous view' of the model) plus the threshold rows
# (exact finite differences: th is affine in the free parameters).

# a 'continuous view' of the model: same GLIST/parameters, but the
# categorical machinery switched off, so that lav_model_delta() and
# lav_model_implied() return the raw means/covariances per block
lav_m07_model_nocat <- function(lavmodel = NULL) {
  lavmodel2 <- lavmodel
  lavmodel2@categorical <- FALSE
  lavmodel2@correlation <- FALSE
  lavmodel2@th.idx <- lapply(lavmodel@th.idx, function(x) integer(0L))
  # no implied standardization:
  # - under parameterization = "theta", lav_model_set_parameters() stores
  #   the per-block scaling values 1/sqrt(diag(Sigma*)) in the GLIST
  #   'delta' matrices; switch to "delta" and reset those matrices to 1,
  #   so that all scaling is the identity and the implied moments (and
  #   their Jacobian) come out unstandardized (raw)
  lavmodel2@parameterization <- "delta"
  delta_mm <- which(names(lavmodel2@GLIST) == "delta")
  for (m in delta_mm) {
    lavmodel2@GLIST[[m]][] <- 1
  }
  lavmodel2
}

# raw thresholds per block: the tau parameters themselves, without any
# scaling (lav_model_th applies the delta/theta standardization);
# entries for numeric variables (th.idx == 0) are set to zero
lav_m07_th_raw <- function(lavmodel = NULL, glist = NULL) {
  if (is.null(glist)) {
    glist <- lavmodel@GLIST
  }
  mm_idx <- lav_model_group_mm_indices(lavmodel@nmat)
  out <- vector("list", lavmodel@nblocks)
  for (b in seq_len(lavmodel@nblocks)) {
    th_idx_b <- lavmodel@th.idx[[b]]
    th_b <- numeric(length(th_idx_b))
    mlist <- glist[mm_idx[[b]]]
    if (!is.null(mlist$tau) && any(th_idx_b > 0L)) {
      th_b[th_idx_b > 0L] <- as.numeric(mlist$tau)
    }
    out[[b]] <- th_b
  }
  out
}

# raw model-implied moments per block (no delta/correlation scaling:
# delta = FALSE switches off the implied standardization)
lav_m07_implied_raw <- function(lavmodel = NULL, glist = NULL) {
  imp <- lav_model_implied(lavmodel, glist = glist, delta = FALSE)
  if (lavmodel@conditional.x) {
    # lav_model_implied() does not forward delta = FALSE to the slopes
    imp$res.slopes <- lav_model_pi(lavmodel = lavmodel, glist = glist,
                                   delta = FALSE)
  }
  imp$th.raw <- lav_m07_th_raw(lavmodel = lavmodel, glist = glist)
  imp
}

# per-group block info: the within (b1) variable layout, the ordinal
# positions, and the between (b2) -> within (b1) variable mapping
lav_m07_block_info <- function(lavmodel = NULL, g = NULL,
                               nw = NULL, nb = NULL) {
  b1 <- (g - 1L) * 2L + 1L
  b2 <- (g - 1L) * 2L + 2L

  # ordinal variables at the within level (level-1 threshold rows exist,
  # fixed to zero, with the same variable-major structure as level 2)
  th_idx_1 <- lavmodel@th.idx[[b1]]
  ord_1 <- sort(unique(th_idx_1[th_idx_1 > 0L]))

  # between -> within variable mapping
  if (nb == nw) {
    w2b <- seq_len(nw)
  } else {
    # from the dimNames of the (ov x ov) theta matrices per block
    mm_idx <- lav_model_group_mm_indices(lavmodel@nmat)
    nm <- names(lavmodel@GLIST)
    t1 <- mm_idx[[b1]][which(nm[mm_idx[[b1]]] == "theta")]
    t2 <- mm_idx[[b2]][which(nm[mm_idx[[b2]]] == "theta")]
    ov1 <- lavmodel@dimNames[[t1]][[1]]
    ov2 <- lavmodel@dimNames[[t2]][[1]]
    if (is.null(ov1) || is.null(ov2)) {
      lav_msg_stop(gettext(
        "cannot map the between-level variables to the within level
        (no dimNames); this should not happen."))
    }
    w2b <- match(ov2, ov1)
    if (anyNA(w2b)) {
      lav_msg_stop(gettext(
        "between-only variables are not supported (yet) for two-level
        (D)WLS estimation with categorical data."))
    }
  }

  between_flag <- seq_len(nw) %in% w2b
  list(
    b1 = b1, b2 = b2, th_idx_1 = th_idx_1, ord_1 = ord_1,
    w2b = w2b, between_flag = between_flag
  )
}

# model-implied (standardized) statistic vector; one vector per GROUP
lav_m07_wls_est <- function(lavmodel = NULL, glist = NULL) {
  nlevels <- 2L
  ngroups <- lavmodel@nblocks %/% nlevels
  conditional_x <- lavmodel@conditional.x
  imp <- lav_m07_implied_raw(lavmodel, glist = glist)

  out <- vector("list", ngroups)
  for (g in seq_len(ngroups)) {
    if (conditional_x) {
      sigma_w <- imp$res.cov[[(g - 1L) * 2L + 1L]]
      sigma_b <- imp$res.cov[[(g - 1L) * 2L + 2L]]
    } else {
      sigma_w <- imp$cov[[(g - 1L) * 2L + 1L]]
      sigma_b <- imp$cov[[(g - 1L) * 2L + 2L]]
    }
    nvar <- NROW(sigma_w)
    nb <- NROW(sigma_b)
    bi <- lav_m07_block_info(lavmodel, g = g, nw = nvar, nb = nb)

    if (conditional_x) {
      mean_w <- imp$res.int[[bi$b1]]
      mu_b <- imp$res.int[[bi$b2]]
      pi_w <- imp$res.slopes[[bi$b1]]
      pi_b <- imp$res.slopes[[bi$b2]]
    } else {
      mean_w <- imp$mean[[bi$b1]]
      mu_b <- imp$mean[[bi$b2]]
      pi_w <- matrix(0, nvar, 0L)
      pi_b <- matrix(0, nb, 0L)
    }
    th_raw2 <- imp$th.raw[[bi$b2]]
    th_idx_2 <- lavmodel@th.idx[[bi$b2]]

    ord <- (seq_len(nvar) %in% bi$ord_1)

    # within scale factors
    delta_1 <- ifelse(ord, 1 / sqrt(diag(sigma_w)), 1.0)

    # 1. thresholds (standardized); 0 entries for continuous variables
    # (the k-th ordinal threshold at level 1 corresponds to the k-th
    #  threshold parameter at level 2)
    th_std <- numeric(length(bi$th_idx_1))
    ord_pos_1 <- which(bi$th_idx_1 > 0L)
    ord_pos_2 <- which(th_idx_2 > 0L)
    th_std[ord_pos_1] <- th_raw2[ord_pos_2] *
      delta_1[bi$th_idx_1[ord_pos_1]]

    # 2. within means/intercepts of the within-only (continuous) variables
    mu_w <- mean_w[!bi$between_flag]

    # 3. standardized within slopes (vec, column-major over (var, exo))
    piw_std <- as.numeric(delta_1 * pi_w)

    # 4. within variances of the continuous variables
    vw <- diag(sigma_w)[!ord]

    # 5. standardized within covariances (vech, no diagonal)
    sw_std <- diag(delta_1) %*% sigma_w %*% diag(delta_1)
    cw <- lav_mat_vech(sw_std, diagonal = FALSE)

    # 6./7./8. standardized between means/intercepts + slopes +
    # covariance matrix (all scaled by the WITHIN scale factors)
    delta_b <- delta_1[bi$w2b]
    mub_std <- delta_b * mu_b
    pib_std <- as.numeric(delta_b * pi_b)
    sb_std <- diag(delta_b, nrow = nb) %*% sigma_b %*%
      diag(delta_b, nrow = nb)

    out[[g]] <- c(
      th_std, mu_w, piw_std, vw, cw, mub_std, pib_std,
      lav_mat_vech(sb_std, diagonal = TRUE)
    )
  }

  out
}

# the standardization map H (per group): rows = statistics, columns = raw
# rows. The raw row layout follows the continuous-view lav_model_delta():
#
#   conditional.x = FALSE:
#     [ th (nth2) ; mean_w (nvar) ; vech(Sigma_w) ;
#       mean_b (nb) ; vech(Sigma_b) ]
#   conditional.x = TRUE (per block: interleaved intercept + slope rows,
#   one set per variable):
#     [ th (nth2) ; {int_p, slopes_p (nq1)} x nvar ; vech(Sigma_w) ;
#       {int_k, slopes_k (nq2)} x nb ; vech(Sigma_b) ]
#
# (the unconditional case is the nq1 = nq2 = 0 special case)
lav_m07_hmat <- function(sigma_w = NULL, sigma_b = NULL, mu_b = NULL,
                         pi_w = NULL, pi_b = NULL,
                         th_raw2 = NULL, th_idx_1 = NULL,
                         th_idx_2 = NULL, w2b = NULL) {
  nvar <- NROW(sigma_w)
  nb <- NROW(sigma_b)
  if (is.null(pi_w)) {
    pi_w <- matrix(0, nvar, 0L)
  }
  if (is.null(pi_b)) {
    pi_b <- matrix(0, nb, 0L)
  }
  nq1 <- NCOL(pi_w)
  nq2 <- NCOL(pi_b)
  between_flag <- seq_len(nvar) %in% w2b
  ord <- (seq_len(nvar) %in% th_idx_1[th_idx_1 > 0L])
  num_idx <- which(!ord)
  nth1 <- length(th_idx_1) # output th rows (within-block layout)
  nth2 <- length(th_idx_2) # raw th rows (between-block tau vector)
  pstar_d <- nvar * (nvar + 1L) / 2L # vech with diagonal (within)
  pstar <- nvar * (nvar - 1L) / 2L # vech without diagonal
  pstar_b <- nb * (nb + 1L) / 2L # vech with diagonal (between)

  delta_1 <- ifelse(ord, 1 / sqrt(diag(sigma_w)), 1.0)
  # d delta_p / d sigma_w,pp = -0.5 * delta_p^3 (ordinal only)
  ddelta <- ifelse(ord, -0.5 * delta_1^3, 0.0)

  # raw row indices
  r_th <- seq_len(nth2)
  r_int1 <- nth2 + (seq_len(nvar) - 1L) * (1L + nq1) + 1L
  # slope rows of variable p: r_int1[p] + 1..nq1
  r_sw <- nth2 + nvar * (1L + nq1) + seq_len(pstar_d)
  r_int2 <- nth2 + nvar * (1L + nq1) + pstar_d +
    (seq_len(nb) - 1L) * (1L + nq2) + 1L
  r_sb <- nth2 + nvar * (1L + nq1) + pstar_d + nb * (1L + nq2) +
    seq_len(pstar_b)
  nraw <- nth2 + nvar * (1L + nq1) + pstar_d + nb * (1L + nq2) + pstar_b

  # position of sigma[i,j] (i >= j) within vech(, diagonal = TRUE)
  vech_pos <- matrix(0L, nvar, nvar)
  vech_pos[lower.tri(vech_pos, diag = TRUE)] <- seq_len(pstar_d)
  vech_pos[upper.tri(vech_pos)] <- t(vech_pos)[upper.tri(vech_pos)]
  diag_pos <- diag(vech_pos)
  # ... and for the between block
  vech_pos_b <- matrix(0L, nb, nb)
  vech_pos_b[lower.tri(vech_pos_b, diag = TRUE)] <- seq_len(pstar_b)
  vech_pos_b[upper.tri(vech_pos_b)] <- t(vech_pos_b)[upper.tri(vech_pos_b)]

  nstat <- nth1 + sum(!between_flag) + nvar * nq1 + length(num_idx) +
    pstar + nb + nb * nq2 + pstar_b
  h <- matrix(0, nstat, nraw)
  row <- 0L

  # 1. thresholds: th*_k = delta_p th_k
  ord_pos_2 <- which(th_idx_2 > 0L)
  kk <- 0L
  for (k in seq_len(nth1)) {
    row <- row + 1L
    p <- th_idx_1[k]
    if (p > 0L) {
      kk <- kk + 1L
      h[row, r_th[ord_pos_2[kk]]] <- delta_1[p]
      h[row, r_sw[diag_pos[p]]] <- th_raw2[ord_pos_2[kk]] * ddelta[p]
    }
    # continuous entries: fixed at zero (zero row)
  }

  # 2. within means/intercepts of the within-only (continuous) variables
  for (p in which(!between_flag)) {
    row <- row + 1L
    h[row, r_int1[p]] <- 1
  }

  # 3. standardized within slopes: pi*_pq = delta_p pi_pq
  # (vec, column-major: all variables for exo 1, then exo 2, ...)
  for (q in seq_len(nq1)) {
    for (p in seq_len(nvar)) {
      row <- row + 1L
      h[row, r_int1[p] + q] <- delta_1[p]
      h[row, r_sw[diag_pos[p]]] <- pi_w[p, q] * ddelta[p]
    }
  }

  # 4. within variances of continuous variables
  for (p in num_idx) {
    row <- row + 1L
    h[row, r_sw[diag_pos[p]]] <- 1
  }

  # 5. standardized within covariances (i > j)
  for (j in seq_len(nvar - 1L)) {
    for (i in (j + 1L):nvar) {
      row <- row + 1L
      s_ij <- sigma_w[i, j]
      h[row, r_sw[vech_pos[i, j]]] <- delta_1[i] * delta_1[j]
      h[row, r_sw[diag_pos[i]]] <- h[row, r_sw[diag_pos[i]]] +
        s_ij * delta_1[j] * ddelta[i]
      h[row, r_sw[diag_pos[j]]] <- h[row, r_sw[diag_pos[j]]] +
        s_ij * delta_1[i] * ddelta[j]
    }
  }

  # 6. standardized between means/intercepts (between variables only)
  for (k in seq_len(nb)) {
    row <- row + 1L
    p <- w2b[k]
    h[row, r_int2[k]] <- delta_1[p]
    h[row, r_sw[diag_pos[p]]] <- mu_b[k] * ddelta[p]
  }

  # 7. standardized between slopes: pi*_kq = delta_{w2b(k)} pi_kq
  for (q in seq_len(nq2)) {
    for (k in seq_len(nb)) {
      row <- row + 1L
      p <- w2b[k]
      h[row, r_int2[k] + q] <- delta_1[p]
      h[row, r_sw[diag_pos[p]]] <- pi_b[k, q] * ddelta[p]
    }
  }

  # 8. standardized between covariances (vech with diagonal, i >= j)
  for (jj in seq_len(nb)) {
    for (ii in jj:nb) {
      row <- row + 1L
      pi_1 <- w2b[ii]; pj_1 <- w2b[jj]
      s_ij <- sigma_b[ii, jj]
      h[row, r_sb[vech_pos_b[ii, jj]]] <- delta_1[pi_1] * delta_1[pj_1]
      h[row, r_sw[diag_pos[pi_1]]] <- h[row, r_sw[diag_pos[pi_1]]] +
        s_ij * delta_1[pj_1] * ddelta[pi_1]
      h[row, r_sw[diag_pos[pj_1]]] <- h[row, r_sw[diag_pos[pj_1]]] +
        s_ij * delta_1[pi_1] * ddelta[pj_1]
    }
  }

  h
}

# Jacobian of the standardized statistic vector; one matrix per GROUP
# (rows match lav_m07_wls_est, columns match lav_model_delta)
lav_m07_delta <- function(lavmodel = NULL, glist = NULL) {
  nlevels <- 2L
  ngroups <- lavmodel@nblocks %/% nlevels

  # raw continuous-style Delta: per group rbind of
  # [mean_w; vech(Sigma_w); mean_b; vech(Sigma_b)]
  lavmodel2 <- lav_m07_model_nocat(lavmodel)
  glist2 <- glist
  if (!is.null(glist2)) {
    # reset the delta matrices here as well (see lav_m07_model_nocat)
    delta_mm <- which(names(glist2) == "delta")
    for (m in delta_mm) {
      glist2[[m]][] <- 1
    }
  }
  delta_raw <- lav_model_delta(lavmodel = lavmodel2, glist = glist2)

  # raw implied (for the H map)
  imp <- lav_m07_implied_raw(lavmodel, glist = glist)

  # threshold Jacobian: the raw thresholds are affine in the free
  # parameters, so one-sided finite differences are exact.
  # IMPORTANT: the base thresholds must be evaluated at the SAME
  # parameter vector as the perturbed ones (x0 from the lavmodel object,
  # which may differ from the current glist state!)
  x0 <- lav_model_get_parameters(lavmodel)
  lavmodel_0 <- lav_model_set_parameters(lavmodel, x = x0)
  th_base <- lav_m07_th_raw(lavmodel = lavmodel_0,
                            glist = lavmodel_0@GLIST)
  nx <- length(x0)
  th_jac <- vector("list", lavmodel@nblocks)
  nth_b <- sapply(th_base, length)
  for (b in seq_len(lavmodel@nblocks)) {
    th_jac[[b]] <- matrix(0, nth_b[b], nx)
  }
  for (k in seq_len(nx)) {
    x1 <- x0
    x1[k] <- x1[k] + 1
    lavmodel_k <- lav_model_set_parameters(lavmodel, x = x1)
    th1 <- lav_m07_th_raw(lavmodel = lavmodel_k,
                          glist = lavmodel_k@GLIST)
    for (b in seq_len(lavmodel@nblocks)) {
      if (nth_b[b] > 0L) {
        th_jac[[b]][, k] <- th1[[b]] - th_base[[b]]
      }
    }
  }

  conditional_x <- lavmodel@conditional.x
  delta_out <- vector("list", ngroups)
  for (g in seq_len(ngroups)) {
    if (conditional_x) {
      sigma_w <- imp$res.cov[[(g - 1L) * 2L + 1L]]
      sigma_b <- imp$res.cov[[(g - 1L) * 2L + 2L]]
    } else {
      sigma_w <- imp$cov[[(g - 1L) * 2L + 1L]]
      sigma_b <- imp$cov[[(g - 1L) * 2L + 2L]]
    }
    nvar <- NROW(sigma_w)
    nb <- NROW(sigma_b)
    bi <- lav_m07_block_info(lavmodel, g = g, nw = nvar, nb = nb)
    if (conditional_x) {
      mu_b <- imp$res.int[[bi$b2]]
      pi_w <- imp$res.slopes[[bi$b1]]
      pi_b <- imp$res.slopes[[bi$b2]]
    } else {
      mu_b <- imp$mean[[bi$b2]]
      pi_w <- pi_b <- NULL
    }
    th_raw2 <- imp$th.raw[[bi$b2]]
    th_idx_2 <- lavmodel@th.idx[[bi$b2]]

    # stack the raw rows: [th (b2); int/slope rows W; vech Sw;
    #                      int/slope rows B; vech Sb]
    draw <- rbind(th_jac[[bi$b2]], delta_raw[[g]])

    h <- lav_m07_hmat(
      sigma_w = sigma_w, sigma_b = sigma_b, mu_b = mu_b,
      pi_w = pi_w, pi_b = pi_b,
      th_raw2 = th_raw2, th_idx_1 = bi$th_idx_1,
      th_idx_2 = th_idx_2, w2b = bi$w2b
    )
    delta_out[[g]] <- h %*% draw
  }

  delta_out
}

#
# added by procedure shortening function names
#
                                                                 # nolint start

lav_matrix_duplication_ginv_pre_post <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_dup_ginv_pre_post", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_ginv_pre_post)
  eval(sc, parent.frame())
}
lav_matrix_orthogonal_complement <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_ortho_complement", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_ortho_complement)
  eval(sc, parent.frame())
}
lav_func_gradient_complex <- function(
                                      func,
                                      x,
                                      h = .Machine$double.eps,
                                      ...,
                                      fallback.simple = TRUE) {
  lav_deprecated("lav_func_grad_complex", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_func_grad_complex)
  eval(sc, parent.frame())
}
lav_matrix_duplication_ginv_post <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_dup_ginv_post", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_ginv_post)
  eval(sc, parent.frame())
}
lav_func_gradient_simple <- function(
                                     func,
                                     x,
                                     h = sqrt(.Machine$double.eps),
                                     ...) {
  lav_deprecated("lav_func_grad_simple", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_func_grad_simple)
  eval(sc, parent.frame())
}
lav_matrix_vech_row_idx <- function(n = 1L, diagonal = TRUE) {
  lav_deprecated("lav_mat_vech_row_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vech_row_idx)
  eval(sc, parent.frame())
}
lav_matrix_vech_col_idx <- function(n = 1L, diagonal = TRUE) {
  lav_deprecated("lav_mat_vech_col_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vech_col_idx)
  eval(sc, parent.frame())
}
lav_matrix_antidiag_idx <- function(n = 1L) {
  lav_deprecated("lav_mat_antidiag_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_antidiag_idx)
  eval(sc, parent.frame())
}
lav_matrix_duplication_pre_post <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_dup_pre_post", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_pre_post)
  eval(sc, parent.frame())
}
lav_matrix_duplication_ginv_pre <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_dup_ginv_pre", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_ginv_pre)
  eval(sc, parent.frame())
}
lav_matrix_commutation_pre_post <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_com_pre_post", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_com_pre_post)
  eval(sc, parent.frame())
}
lav_partable_unrestricted <- function(
                                      lavobject = NULL,
                                      lavdata = NULL,
                                      lavpta = NULL,
                                      lavoptions = NULL,
                                      lavsamplestats = NULL,
                                      lavh1 = NULL,
                                      sample.cov = NULL,
                                      sample.mean = NULL,
                                      sample.slopes = NULL,
                                      sample.th = NULL,
                                      sample.th.idx = NULL,
                                      sample.cov.x = NULL,
                                      sample.mean.x = NULL) {
  lav_deprecated("lav_pt_unrestricted", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_unrestricted)
  eval(sc, parent.frame())
}
lav_partable_independence <- function(
                                      lavobject = NULL,
                                      lavdata = NULL,
                                      lavpta = NULL,
                                      lavoptions = NULL,
                                      lavsamplestats = NULL,
                                      lavh1 = NULL,
                                      sample.cov = NULL,
                                      sample.mean = NULL,
                                      sample.slopes = NULL,
                                      sample.th = NULL,
                                      sample.th.idx = NULL,
                                      sample.cov.x = NULL,
                                      sample.mean.x = NULL) {
  lav_deprecated("lav_pt_independence", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_independence)
  eval(sc, parent.frame())
}
lav_matrix_vechru_idx <- function(n = 1L, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechru_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechru_idx)
  eval(sc, parent.frame())
}
lav_matrix_upper2full <- function(x, diagonal = TRUE) {
  lav_deprecated("lav_mat_upper2full", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_upper2full)
  eval(sc, parent.frame())
}
lav_matrix_lower2full <- function(x, diagonal = TRUE) {
  lav_deprecated("lav_mat_lower2full", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_lower2full)
  eval(sc, parent.frame())
}
lav_matrix_commutation_mn_pre <- function(
                                          A,
                                          m = 1L,
                                          n = 1L) {
  lav_deprecated("lav_mat_com_mn_pre", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_com_mn_pre)
  eval(sc, parent.frame())
}
lav_samplestats_from_data <- function(
                                      lavdata = NULL,
                                      lavoptions = NULL,
                                      WLS.V = NULL,
                                      NACOV = NULL) {
  lav_deprecated("lav_samp_from_data", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_samp_from_data)
  eval(sc, parent.frame())
}
lav_matrix_vechru_reverse <- function(x, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechru_rev", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechru_rev)
  eval(sc, parent.frame())
}
lav_matrix_vechr_idx <- function(n = 1L, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechr_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechr_idx)
  eval(sc, parent.frame())
}
lav_matrix_vechu_idx <- function(n = 1L, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechu_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechu_idx)
  eval(sc, parent.frame())
}
lav_matrix_diagh_idx <- function(n = 1L) {
  lav_deprecated("lav_mat_diagh_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_diagh_idx)
  eval(sc, parent.frame())
}
lav_partable_attributes <- function(partable, pta = NULL) {
  lav_deprecated("lav_pt_attributes", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_attributes)
  eval(sc, parent.frame())
}
lav_matrix_vechr_reverse <- function(x, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechr_rev", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechr_rev)
  eval(sc, parent.frame())
}
lav_matrix_vechu_reverse <- function(x, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechu_rev", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechu_rev)
  eval(sc, parent.frame())
}
lav_matrix_vech_idx <- function(n = 1L, diagonal = TRUE) {
  lav_deprecated("lav_mat_vech_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vech_idx)
  eval(sc, parent.frame())
}
lav_matrix_diag_idx <- function(n = 1L) {
  lav_deprecated("lav_mat_diag_idx", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_diag_idx)
  eval(sc, parent.frame())
}
lav_matrix_duplication_post <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_dup_post", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_post)
  eval(sc, parent.frame())
}
lav_matrix_duplication_ginv <- function(n = 1L) {
  lav_deprecated("lav_mat_dup_ginv", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_ginv)
  eval(sc, parent.frame())
}
lav_matrix_commutation_post <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_com_post", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_com_post)
  eval(sc, parent.frame())
}
lav_matrix_symmetric_sqrt <- function(S = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_sym_sqrt", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_sym_sqrt)
  eval(sc, parent.frame())
}
lav_matrix_vech_reverse <- function(x, diagonal = TRUE) {
  lav_deprecated("lav_mat_vech_rev", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vech_rev)
  eval(sc, parent.frame())
}
lav_matrix_duplication_pre <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_dup_pre", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup_pre)
  eval(sc, parent.frame())
}
lav_matrix_commutation_pre <- function(A = matrix(0, 0, 0)) {
  lav_deprecated("lav_mat_com_pre", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_com_pre)
  eval(sc, parent.frame())
}
lav_partable_complete <- function(partable = NULL, start = TRUE) {
  lav_deprecated("lav_pt_complete", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_complete)
  eval(sc, parent.frame())
}
lav_matrix_vechru <- function(S, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechru", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechru)
  eval(sc, parent.frame())
}
lav_partable_constraints_def <- function(
                                         partable,
                                         con = NULL,
                                         debug = FALSE,
                                         txtOnly = FALSE,
                                         warn = TRUE) {
  lav_deprecated("lav_pt_con_def", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_con_def)
  eval(sc, parent.frame())
}
lav_partable_constraints_ceq <- function(
                                         partable,
                                         con = NULL,
                                         debug = FALSE,
                                         txtOnly = FALSE) {
  lav_deprecated("lav_pt_con_ceq", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_con_ceq)
  eval(sc, parent.frame())
}
lav_partable_constraints_ciq <- function(
                                         partable,
                                         con = NULL,
                                         debug = FALSE,
                                         txtOnly = FALSE) {
  lav_deprecated("lav_pt_con_ciq", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_con_ciq)
  eval(sc, parent.frame())
}
lav_partable_from_lm <- function(
                                 object,
                                 est = FALSE,
                                 label = FALSE,
                                 as.data.frame. = FALSE) {
  lav_deprecated("lav_pt_from_lm", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_from_lm)
  eval(sc, parent.frame())
}
lav_constraints_parse <- function(
                                  partable = NULL,
                                  constraints = NULL,
                                  theta = NULL,
                                  debug = FALSE) {
  lav_deprecated("lav_con_parse", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_con_parse)
  eval(sc, parent.frame())
}
lav_matrix_vechr <- function(S, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechr", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechr)
  eval(sc, parent.frame())
}
lav_matrix_vechu <- function(S, diagonal = TRUE) {
  lav_deprecated("lav_mat_vechu", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vechu)
  eval(sc, parent.frame())
}
lav_matrix_bdiag <- function(...) {
  lav_deprecated("lav_mat_bdiag", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_bdiag)
  eval(sc, parent.frame())
}
lav_matrix_trace <- function(..., check = TRUE) {
  lav_deprecated("lav_mat_trace", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_trace)
  eval(sc, parent.frame())
}
lav_partable_labels <- function(
                                partable,
                                blocks = c("group", "level"),
                                group.equal = "",
                                group.partial = "",
                                type = "user") {
  lav_deprecated("lav_pt_labels", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_labels)
  eval(sc, parent.frame())
}
# used in bain
getParameterLabels <- lav_alias("lav_partable_labels")

lav_matrix_vecr <- function(A) {
  lav_deprecated("lav_mat_vecr", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vecr)
  eval(sc, parent.frame())
}
lav_matrix_vech <- function(S, diagonal = TRUE) {
  lav_deprecated("lav_mat_vech", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vech)
  eval(sc, parent.frame())
}
lav_partable_merge <- function(
                               pt1 = NULL,
                               pt2 = NULL,
                               remove.duplicated = FALSE,
                               fromLast = FALSE,
                               warn = TRUE) {
  lav_deprecated("lav_pt_merge", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_merge)
  eval(sc, parent.frame())
}
lav_matrix_vec <- function(A) {
  lav_deprecated("lav_mat_vec", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_vec)
  eval(sc, parent.frame())
}
lav_matrix_duplication <- function(n = 1L) {
  lav_deprecated("lav_mat_dup", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_dup)
  eval(sc, parent.frame())
}
lav_matrix_commutation <- function(m = 1L, n = 1L) {
  lav_deprecated("lav_mat_com", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_com)
  eval(sc, parent.frame())
}
lav_matrix_cov <- function(Y, Mu = NULL) {
  lav_deprecated("lav_mat_cov", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_mat_cov)
  eval(sc, parent.frame())
}
lav_partable_ndat <- function(partable) {
  lav_deprecated("lav_pt_ndat", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_ndat)
  eval(sc, parent.frame())
}
lav_partable_npar <- function(partable) {
  lav_deprecated("lav_pt_npar", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_npar)
  eval(sc, parent.frame())
}
lav_partable_add <- function(partable = NULL, add = list()) {
  lav_deprecated("lav_pt_add", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_add)
  eval(sc, parent.frame())
}
lav_partable_df <- function(partable) {
  lav_deprecated("lav_pt_df", times = 0L) # --> for now no warning
  sc <- sys.call()
  names(sc) <- lav_snake_case(names(sc))
  sc[[1L]] <- quote(lavaan::lav_pt_df)
  eval(sc, parent.frame())
}                                                             # nolint end

# handle bare-minimum partables
# add some additional columns
lav_pt_complete <- function(partable = NULL, start = TRUE) {
  # check if we have a data.frame
  # if so, check for columns that are 'factor' and convert them to 'character'
  ovda <- attr(partable, "ovda")
  if (is.data.frame(partable)) {
    fac_idx <- sapply(partable, is.factor)
    partable[fac_idx] <- lapply(partable[fac_idx], as.character)
  }

  # check if we have lhs, op, rhs
  stopifnot(!is.null(partable$lhs),
    !is.null(partable$op),
    !is.null(partable$rhs))

  # number of elements
  tmp_n <- length(partable$lhs)
  if (!is.data.frame(partable)) {
    # check for equal column length
    nel <- sapply(partable, length)
    short_idx <- which(nel < tmp_n)
    long_idx <- which(nel > tmp_n)
    if (length(long_idx) > 0L) {
      lav_msg_warn(gettext("partable columns have unequal length"))
    }
    if (length(short_idx) > 0L) {
      # try to extend them in a 'natural' way
      for (i in short_idx) {
        too_short <- tmp_n - nel[i]
        if (is.integer(partable[[i]])) {
          partable[[i]] <- c(partable[[i]],
            integer(too_short))
        } else if (is.numeric(partable[[i]])) {
          partable[[i]] <- c(partable[[i]],
            numeric(too_short))
        } else {
          partable[[i]] <- c(partable[[i]],
            character(too_short))
        }
      }
    }
  }

  # create new id column
  # if(is.null(partable$id)) {
  partable$id <- seq_len(tmp_n)
  # }

  # add user column
  if (is.null(partable$user)) {
    partable$user <- rep(1L, tmp_n)
  } else {
    partable$user <- as.integer(partable$user)
  }

  # add block column
  if (is.null(partable$block)) {
    partable$block <- rep(1L, tmp_n)
  } else {
    partable$block <- as.integer(partable$block)
  }

  # add group column
  if (is.null(partable$group)) {
    partable$group <- rep(1L, tmp_n)
  } else {
    # partable$group <- as.integer(partable$group) # maybe labels?
  }

  # add free column
  if (is.null(partable$free)) {
    partable$free <- seq_len(tmp_n)
    # 0.6-11: check for simple equality constraints
    #         note: this is perhaps only a subset (eg SAM!) of a larger
    #         table, and we have to renumber the 'free' column
  } else if (is.integer(partable$free) &&
    any(partable$free > 0L)   &&
    !any(partable$op == "==") &&
    !is.null(partable$label)  &&
    !is.null(partable$plabel) &&
    any(duplicated(partable$free[partable$free > 0L]))) {
    dup_idx <- which(partable$free > 0L & duplicated(partable$free))
    all_idx <- which(partable$free %in% unique(partable$free[dup_idx]))
    eq_labels <- unique(partable$free[all_idx])
    eq_id <- integer(length(partable$lhs))
    eq_id[all_idx] <- partable$free[all_idx]
    partable$free[dup_idx] <- 0L
    idx_free <- which(partable$free > 0L)
    partable$free <- rep(0L, tmp_n)
    partable$free[idx_free] <- seq_along(idx_free)
    for (eq_label in eq_labels) {
      all_idx <- which(eq_id == eq_label)
      ref_idx <- all_idx[1L]
      other_idx <- all_idx[-1L]
      partable$free[other_idx] <- partable$free[ref_idx]
    }
  } else {
    # treat non-zero as 'free'
    free_idx <- which(as.logical(partable$free))
    partable$free <- rep(0L, tmp_n)
    if (length(free_idx) > 0L) {
      partable$free[free_idx] <- seq_along(free_idx)
    }
  }

  # add ustart column
  if (is.null(partable$ustart)) {
    # do we have something else? start? est?
    if (!is.null(partable$start)) {
      partable$ustart <- as.numeric(partable$start)
    } else if (!is.null(partable$est)) {
      partable$ustart <- as.numeric(partable$est)
    } else {
      partable$ustart <- rep(as.numeric(NA), tmp_n)
      non_free <- which(!partable$free)
      if (length(non_free)) {
        partable$ustart[non_free] <- 0
      }
    }
  } else {
    partable$ustart <- as.numeric(partable$ustart)
  }

  # add exo column
  if (is.null(partable$exo)) {
    partable$exo <- rep(0, tmp_n)
  } else {
    partable$exo <- as.integer(partable$exo)
  }

  # add label column
  if (is.null(partable$label)) {
    partable$label <- rep("", tmp_n)
  } else {
    partable$label <- as.character(partable$label)
  }

  # order them nicely: id lhs op rhs group
  idx <- match(c("id", "lhs", "op", "rhs", "user", "block", "group",
    "free", "ustart", "exo", "label"),
  names(partable))
  tmp <- partable[idx]
  partable <- c(tmp, partable[-idx])

  # add start column
  if (start) {
    if (is.null(partable$start)) {
      partable$start <- lav_start(start_method = "simple",
        lavpartable = partable)
    }
  }
  attr(partable, "ovda") <- ovda
  attr(partable, "vnames") <- lav_pt_vnames(partable, "*")
  partable
}

lav_step01_ovnames_initflat <- function(slot_par_table     = NULL,
                                               model            = NULL,
                                               dotdotdot_parser = "new") {
  # if slotPartable not NULL copy to flat_model
  # else
  #   if model is of type character
  #     parse model to flat_model
  #   else if model is a formula (** warning **)
  #     transform "~ x1 + x2 + x3" to character "f =~ x1 + x2 + x3",
  #         and parse to flat_model
  #     transform "y =~ x1 + x2 + x3" to character and parse to flat_model
  #     something else : *** error ***
  #   else if model is a lavaan object
  #     extract flat_model from @parTable
  #   else if model is a list
  #     if bare minimum present (columns lhs, op, rhs, free)
  #        flat_model = model
  #        replace column block by column group (if present)
  #     else
  #        --> *** error ***
  #   else
  #     --> ***error***
  #

  # 1a. get ov.names and ov.names.x (per group) -- needed for lav_lavdata()
  if (!is.null(slot_par_table)) {
    flat_model <- slot_par_table
  } else if (is.character(model)) {
    if (is.null(dotdotdot_parser)) {
      flat_model <- lavParseModelString(model)
    } else {
      flat_model <- lavParseModelString(model, parser = dotdotdot_parser)
    }
  } else if (inherits(model, "formula")) {
    # two typical cases:
    # 1. regression type formula
    # 2. no quotes, e.g. f =~ x1 + x2 + x3
    #    TODO: this isn't a valid formula !!!
    tmp <- as.character(model)
    if (tmp[1] == "~" && length(tmp) == 2L) {
      # looks like an unquoted single factor model f =~ something
      lav_msg_warn(
        gettext("model seems to be a formula;
                please enclose the model syntax between quotes"))
      # create model and hope for the best
      model_bis <- paste("f =", paste(tmp, collapse = " "), sep = "")
      flat_model <- lavParseModelString(model_bis)
    } else if (tmp[1] == "~" && length(tmp) == 3L) {
      # looks like a (unquoted) regression formula
      lav_msg_warn(
        gettext("model seems to be a formula;
                please enclose the model syntax between quotes"))
      # create model and hope for the best
      model_bis <- paste(tmp[2], tmp[1], tmp[3])
      flat_model <- lavParseModelString(model_bis)
    } else {
      lav_msg_stop(
        gettext("model seems to be a formula;
                please enclose the model syntax between quotes"))
    }
  } else if (inherits(model, "lavaan")) {
    # hm, a lavaan model; let's try to extract the parameter table
    # and see what happens
    flat_model <- parTable(model)
  } else if (is.list(model)) {
    # a list! perhaps a full parameter table, or an initial flat model,
    # or something else...

  # 1. flat_model already (output of lavParseModelString)?
  if (!is.null(model$lhs) && !is.null(model$op) &&
      !is.null(model$rhs) && !is.null(model$mod.idx) &&
    !is.null(attr(model, "modifiers"))) {
    flat_model <- model
    # look for the bare minimum columns: lhs - op - rhs
    } else if (!is.null(model$lhs) && !is.null(model$op) &&
      !is.null(model$rhs) && !is.null(model$free)) {
      # ok, we have something that looks like a parameter table
      flat_model <- model

      # fix semTools issue here? for auxiliary() which does not use
      # block column yet
      if (!is.null(flat_model$block)) {
        nn <- length(flat_model$lhs)
        if (length(flat_model$block) != nn) {
          flat_model$block <- flat_model$group
        }
        if (any(is.na(flat_model$block))) {
          flat_model$block <- flat_model$group
        }
      } else if (!is.null(flat_model$group)) {
        flat_model$block <- flat_model$group
      }
    } else {
      bare_minimum <- c("lhs", "op", "rhs", "free")
      missing_idx <- is.na(match(bare_minimum, names(model)))
      lav_msg_stop(
        gettextf("model is a list, but not a parameterTable?
                 missing column(s) in parameter table: [%s]",
                  lav_msg_view(bare_minimum[missing_idx], "none")))
    }
  } else {
    lav_msg_stop(gettext("model is NULL or not a valid type for it!"))
  }

  # Ok, we got a flattened model; usually this a flat_model object, but it
  # could also be an already lavaanified parTable, or a bare-minimum list with
  # lhs/op/rhs/free elements
  flat_model
}

# this function is only needed for backwards compatibility: it falls back
# to the old (<0.6-20) way of handling composites (defined by the '<~' operator)
# we do this by substituting the '<~' operator by '~', and by creating
# phantom latent variables for the LHS, and setting the residual variance
# to zero
#
# For example: "f <~ 1*income + occup + educ" is replaced by:
# "f =~ 0; f ~~ 0*f; f ~ 1*income + occup + educ"
#
lav_step01_ovnames_composites <- function(flat_model = NULL) {

  # get composite idx
  c_idx <- which(flat_model$op == "<~")
  if (length(c_idx) == 0L) {
    return(flat_model)
  }

  # flat_model info
  flat_names <- names(flat_model)
  c_names <- unique(flat_model$lhs[c_idx])
  nc <- length(c_names)
  mod_val <- max(flat_model$mod.idx)

  # check block numbers
  max_block <- max(flat_model$block)
  if (max_block > 1L) {
    lav_msg_stop(gettext(
      "composites = FALSE is not supported when multiple blocks are present;",
      " manually replace f <~ rhs by f =~ 0; f ~~ 0*f; f ~ rhs"))
  }

  block <- 1L

  # replace '<~' by '~'
  flat_model$op[c_idx] <- "~"

  # add phantom latent variables
  flat_model$lhs     <- c(flat_model$lhs, c_names)
  flat_model$op      <- c(flat_model$op,  rep("=~", nc))
  flat_model$rhs     <- c(flat_model$rhs, c_names)
  flat_model$fixed   <- c(flat_model$fixed, rep("", nc))
  flat_model$mod.idx <- c(flat_model$mod.idx, rep(0L, nc))
  flat_model$block   <- c(flat_model$block, rep(block, nc))

  # fix residual variances
  flat_model$lhs     <- c(flat_model$lhs, c_names)
  flat_model$op      <- c(flat_model$op,  rep("~~", nc))
  flat_model$rhs     <- c(flat_model$rhs, c_names)
  flat_model$fixed   <- c(flat_model$fixed, rep("0", nc)) # just for show
  flat_model$mod.idx <- c(flat_model$mod.idx, mod_val + seq_len(nc))
  flat_model$block   <- c(flat_model$block, rep(block, nc))

  # extend the 'other' columns
  other <- flat_names[!flat_names %in% c("lhs", "op", "rhs",
                                         "fixed", "mod.idx", "block")]
  for (o in other) {
    if (is.character(flat_model[[o]])) {
      flat_model[[o]] <- c(flat_model[[o]], rep("", nc * 2))
    } else if (is.integer(flat_model[[o]])) {
      flat_model[[o]] <- c(flat_model[[o]], rep(0L, nc * 2))
    } else {
      flat_model[[o]] <- c(flat_model[[o]], rep(as.numeric(NA), nc * 2))
    }
  }

  # add fixed values to the modifier attribute
  attr(flat_model, "modifiers") <- c(attr(flat_model, "modifiers"),
                                     rep(list(list(fixed = 0)), nc))

  flat_model
}

lav_step01_ovnames_ovorder <- function(flat_model = NULL,
                                              ov_order   = "model",
                                              data       = NULL,
                                              sample_cov = NULL,
                                              slot_data   = NULL) {
  # set ov_order in lowercase, check if it is "data" or "model",
  #  if not *** error ***
  # if ov_order == "data"
  #   try adapt flat_model via lav_partable_ov_from_data
  #   (** warning ** if this fails)

  # new in 0.6-14
  # if ov_order = "data", it would seem we need to intervene here;
  # ldw 1/3/2024:
  # we do this by adding an attribute "ovda" to flat_model and partable
  ov_order <- tolower(ov_order)
  if (ov_order == "data") {
    flat_model_orig <- flat_model
    try(
      flat_model <- lav_pt_ov_from_data(flat_model,
        data = data,
        sample_cov = sample_cov,
        slot_data = slot_data
      ),
      silent = TRUE
    )
    if (inherits(flat_model, "try-error")) {
      lav_msg_warn(gettext("ov_order = \"data\" setting failed;
                           switching back to ov_order = \"model\""))
      flat_model <- flat_model_orig
    }
  } else if (ov_order != "model") {
    lav_msg_stop(gettext(
      "ov_order= argument should be \"model\" (default) or \"data\""))
  }

  flat_model
}

lav_step01_ovnames_group <- function(flat_model = NULL,
                                            ngroups    = 1L) {
  # if "group :" appears in flat_model
  #   tmp.group.values: set of names in corresponding right hand sides
  #   copy flat_model without attributes and call lav_model_pt,
  #     store result in tmp.lav
  #   extract ov.names, ov.names.y, ov.names.x, lv.names from tmp.lav
  #     via lav_partable_vnames
  # else
  #   if flat_model$group not NULL and more than 1 group.value
  #     extract group.values via lav_partable_group_values
  #     extract, for each group.value,
  #       ov.names, ov.names.y, ov.names.x, lv.names from flat_model
  #       via lav_partable_vnames
  #   else
  #     extract ov.names, ov.names.y, ov.names.x, lv.names from flat_model
  #     via lav_partable_vnames
  #
  #     TODO: call lav_pt_vnames only ones and not for each type

  flat_model_2 <- NULL
  tmp_lav <- NULL
  group_values <- NULL
  ov_names  <- character(0L)
  if (any(flat_model$op == ":" & tolower(flat_model$lhs) == "group")) {
    # here, we only need to figure out:
    # - ngroups
    # - ov's per group
    #
    # - FIXME: we need a more efficient way, avoiding
    #          lav_model_pt/lav_partable_vnames
    #
    group_idx <- which(flat_model$op == ":" &
                       tolower(flat_model$lhs) == "group")
    # replace by 'group' (in case we got 'Group'):
    flat_model$lhs[group_idx] <- "group"
    tmp_group_values <- unique(flat_model$rhs[group_idx])
    tmp_ngroups <- length(tmp_group_values)

    flat_model_2 <- flat_model
    attr(flat_model_2, "modifiers") <- NULL
    attr(flat_model_2, "constraints") <- NULL
    tmp_lav <-
      lav_model_pt(flat_model_2, ngroups = tmp_ngroups, warn = FALSE)
    ov_names <- ov_names_y <- ov_names_x <- lv_names <- vector("list",
      length = tmp_ngroups
    )
    attr(tmp_lav, "vnames") <- lav_pt_vnames(tmp_lav, type = "*")
    for (g in seq_len(tmp_ngroups)) {
      ov_names[[g]] <- unique(unlist(lav_pt_vnames(tmp_lav,
        type = "ov", group = tmp_group_values[g]
      )))
      ov_names_y[[g]] <- unique(unlist(lav_pt_vnames(tmp_lav,
        type = "ov.nox", group = tmp_group_values[g]
      )))
      ov_names_x[[g]] <- unique(unlist(lav_pt_vnames(tmp_lav,
        type = "ov.x", group = tmp_group_values[g]
      )))
      lv_names[[g]] <- unique(unlist(lav_pt_vnames(tmp_lav,
        type = "lv", group = tmp_group_values[g]
      )))
    }
  } else if (!is.null(flat_model$group)) {
    # user-provided full partable with group column!
    attr(flat_model, "vnames") <- lav_pt_vnames(flat_model, type = "*")
    ngroups <- lav_pt_ngroups(flat_model)
    if (ngroups > 1L) {
      group_values <- lav_pt_group_values(flat_model)
      ov_names <- ov_names_y <- ov_names_x <- lv_names <- vector("list",
        length = ngroups
      )
      for (g in seq_len(ngroups)) {
        # collapsed over levels (if any)
        ov_names[[g]] <- unique(unlist(lav_pt_vnames(flat_model,
          type = "ov", group = group_values[g]
        )))
        ov_names_y[[g]] <- unique(unlist(lav_pt_vnames(flat_model,
          type = "ov.nox", group = group_values[g]
        )))
        ov_names_x[[g]] <- unique(unlist(lav_pt_vnames(flat_model,
          type = "ov.x", group = group_values[g]
        )))
        lv_names[[g]] <- unique(unlist(lav_pt_vnames(flat_model,
          type = "lv", group = group_values[g]
        )))
      }
    } else {
      ov_names   <- lav_pt_vnames(flat_model, type = "ov")
      ov_names_y <- lav_pt_vnames(flat_model, type = "ov.nox")
      ov_names_x <- lav_pt_vnames(flat_model, type = "ov.x")
      lv_names   <- lav_pt_vnames(flat_model, type = "lv")
    }
  } else {
    # collapse over levels (if any)
    attr(flat_model, "vnames") <- lav_pt_vnames(flat_model, type = "*")
    ov_names   <- unique(unlist(lav_pt_vnames(flat_model, type = "ov")))
    ov_names_y <- unique(unlist(lav_pt_vnames(flat_model,
      type = "ov.nox")))
    ov_names_x <- unique(unlist(lav_pt_vnames(flat_model, type = "ov.x")))
    lv_names   <- unique(unlist(lav_pt_vnames(flat_model, type = "lv")))
  }

  # sanity check (new in 0.6-8): do we have any ov.names?
  # detect early
  if (length(unlist(ov_names)) == 0L) {
    lav_msg_stop(
      gettext("ov.names is empty: model does not refer to any observed
              variables; check your syntax."))
  }

  list(
    flat_model   = flat_model,
    ov_names     = ov_names,
    ov_names_x   = ov_names_x,
    ov_names_y   = ov_names_y,
    lv_names     = lv_names,
    group_values = group_values,
    ngroups      = ngroups
  )
}

lav_step01_ovnames_checklv <- function(
    lv_names    = character(0L),
    ov_names    = character(0L),
    data        = NULL,
    sample_cov  = NULL,
    dotdotdot   = NULL,
    slot_options = NULL) {

  # latent variable names should not appear in the subset of the data
  #   that is formed by merging ov+lv names --> **warning**
  # latent interactions are not supported ---> *** error ***

  if (is.null(data) && is.null(sample_cov)) {
    return(invisible(NULL))
  }

  # handle for lv.names that are also observed variables (new in 0.6-6)
  lv_lv_names <- unique(unlist(lv_names))
  ov_ov_names <- unique(unlist(ov_names))
  if (length(lv_lv_names) > 0L) {

    # get data-based variable names
    data_names <- character(0L)
    if (!is.null(data)) {
      data_names <- names(data)
    } else if (!is.null(sample_cov)) {
      data_names <- rownames(sample_cov)
    }

    # create subset of variable names, based on the model
    model_names <- unique(c(ov_ov_names, lv_lv_names))
    subset_names <- data_names[data_names %in% model_names]
    bad_idx <- which(lv_lv_names %in% subset_names)

    if (length(bad_idx) > 0L) {
      if (!is.null(dotdotdot$check.lv.names) &&
        !dotdotdot$check.lv.names) {
        # ignore it, user switched this check off -- new in 0.6-7
      } else {
        lav_msg_warn(gettextf(
          "Some latent variable names collide with observed variable names in
           the dataset: %s. Please provide alternative names for the latent
           variables, or switch off this check using check.lv.names = FALSE",
           paste(lv_lv_names[bad_idx], collapse = " "))
        )
      }
    }
  }

  # sanity check: we do not support latent interaction yet (using the :)
  lv_int_idx <- grep(":", lv_lv_names, fixed = TRUE)
  if (length(lv_int_idx) > 0L) {
    if (!is.null(dotdotdot$check.lv.interaction) &&
      !dotdotdot$check.lv.interaction) {
      # ignore, user (or sam) switched this check off - new in 0.6-16
    } else if (!is.null(slot_options) && !slot_options$check.lv.interaction) {
      # ignore
    } else {
      lav_msg_stop(gettextf(
        "Interaction terms involving latent variables (%s) are not supported.
         Either use the sam() function, or consider using the modsem package.",
         lv_lv_names[lv_int_idx[1]]))
    }
  }

  invisible(NULL)
}

lav_step01_ovnames_namesl <- function(data = NULL,
                                      cluster      = NULL,
                                      flat_model   = NULL,
                                      group_values = NULL,
                                      ngroups      = 1L) {
  # if "level :" appears in flat_model
  #   if data not NULL, cluster must not be NULL, if it is: *** error ***
  #   compute tmp.group.values and tmp.level.values from flat_model
  #   there should be at least 2 levels, if not *** error ***
  #   copy flat_model without attributes and lav_model_pt -> tmp.lav
  #   check at least 2 levels for tmp.lav, if not *** error ***
  #   compute ov.names.l per group and per level (via lav_partable_vnames
  #     on tmp.lav)
  # else
  #   if lav_pt_nlevels(flat_model) > 0
  #   if data not NULL, cluster must not be NULL, if it is: *** error ***
  #     compute ov.names.l per group and per level (via lav_partable_vnames
  #     on flat_model)
  #   else
  #     there are no levels (ov.names.l = list())

  # handle ov.names.l
  if (any(flat_model$op == ":" & tolower(flat_model$lhs) == "level")) {
    # check for cluster argument
    if (!is.null(data) && is.null(cluster)) {
      lav_msg_stop(gettext("cluster argument is missing."))
    }

    # here, we only need to figure out:
    # - nlevels
    # - ov's per level
    # - FIXME: we need a more efficient way, avoiding lav_model_pt/vnames

    group_idx <- which(flat_model$op == ":" & flat_model$lhs == "group")
    tmp_group_values <- unique(flat_model$rhs[group_idx])
    tmp_ngroups <- max(c(length(tmp_group_values), 1))

    level_idx <- which(flat_model$op == ":" &
                       tolower(flat_model$lhs) == "level")
    # replace by "level" (in case we got 'Level')
    flat_model$lhs[level_idx] <- "level"
    tmp_level_values <- unique(flat_model$rhs[level_idx])
    tmp_nlevels <- length(tmp_level_values)

    # we need at least 2 levels (for now)
    if (tmp_nlevels < 2L) {
      lav_msg_stop(
        gettext("when data is clustered, you must specify a model
                for each level in the model syntax (for now);
                see example(Demo.twolevel)")
      )
    }

    flat_model_2 <- flat_model
    attr(flat_model_2, "modifiers") <- NULL
    attr(flat_model_2, "constraints") <- NULL
    tmp_lav <-
      lav_model_pt(flat_model_2, ngroups = tmp_ngroups, warn = FALSE)
    # check for empty levels
    if (max(tmp_lav$level) < 2L) {
      lav_msg_stop(
        gettext("at least one level has no model syntax;
                you must specify a model for each level in the model syntax;
                see example(Demo.twolevel)")
      )
    }
    ov_names_l <- vector("list", length = tmp_ngroups) # per group

    for (g in seq_len(tmp_ngroups)) {
      ov_names_l[[g]] <- vector("list", length = tmp_nlevels)
      for (l in seq_len(tmp_nlevels)) {
        if (tmp_ngroups > 1L) {
          ov_names_l[[g]][[l]] <-
            unique(unlist(lav_pt_vnames(tmp_lav,
              type = "ov",
              group = tmp_group_values[g],
              level = tmp_level_values[l]
            )))
        } else {
          ov_names_l[[g]][[l]] <-
            unique(unlist(lav_pt_vnames(tmp_lav,
              type = "ov",
              level = tmp_level_values[l]
            )))
        }
      } # levels
    } # groups
  } else {
    # perhaps model is already a parameter table
    nlevels <- lav_pt_nlevels(flat_model)
    if (nlevels > 1L) {
      # check for cluster argument (only if we have data)
      if (!is.null(data) && is.null(cluster)) {
        lav_msg_stop(gettext("cluster argument is missing."))
      }

      ngroups <- lav_pt_ngroups(flat_model)
      group_values <- lav_pt_group_values(flat_model)
      ov_names_l <- vector("list", length = ngroups)
      for (g in 1:ngroups) {
        # note: lav_object_vnames() will return a list if any level:
        ov_names_l[[g]] <-
          lav_object_vnames(flat_model, "ov", group = group_values[g])
      }
    } else {
      # no level: in model syntax
      ov_names_l <- list()
    }
  }

  list(
    flat_model = flat_model,
    ov_names_l = ov_names_l
  )
}

lav_step01_ovnames_ordered <- function(ordered    = NULL,
                                              flat_model = NULL,
                                              data       = NULL) {
  # interpretation and check ordered parameter, modify if needed

  # sanity check ordered argument (just in case, add lhs variables names)
  if (!is.null(ordered)) { # new in 0.6-4
    if (is.logical(ordered) && ordered) { # ordered = TRUE
      # assume the user means: ordered = names(Data)
      ordered <- lav_object_vnames(flat_model, "ov.nox")
                                          # new in 0.6-6: changed from ov
    } else if (is.logical(ordered) && !ordered) {
      ordered <- character(0L)
    } else if (!is.character(ordered)) {
      lav_msg_stop(gettext("ordered argument must be a character vector"))
    } else if (length(ordered) == 1L && nchar(ordered) == 0L) {
      ordered <- character(0L)
    } else {
      # check if all names in "ordered" occur in the dataset?
      if (!is.null(data)) {
        if (inherits(data, "data.frame")) {
          data_names <- names(data)
        } else if (inherits(data, "matrix")) {
          data_names <- colnames(data)
        }
        missing_idx <- which(!ordered %in% data_names)
        if (length(missing_idx) > 0L) { # FIXME: warn = FALSE has no eff
          lav_msg_warn(gettextf(
            "ordered variable(s): %s could not be found
            in the data and will be ignored",
             paste(ordered[missing_idx], collapse = " ")))
        }
      }
    }
  }

  # add the variable names that were treated as ordinal
  # in the model syntax
  ordered <- unique(c(ordered, lav_object_vnames(flat_model, "ov.ord")))

  ordered
}

# LDW 26 Mar 2024: use option settings and store in cache environment

lavaan_cache_env <- new.env(parent = emptyenv())
# The variables stored in lavaan_cache_env environment all have a prefix
# to avoid conflicts. The prefixes used at the moment of writing
# (April 2026) are:
#                      for options: opt_
#    for trace debugging functions: trc_ and trace_
#     for the lavaan syntax parser: mdl_
# for tracking deprecated warnings: dpr_


# # function to get or set the switch to use c++ code in lavaanC
# lav_use_lavaanC <- uselavaanC <- function(x) {
#   if (missing(x)) {
#     if (!exists("opt.lavaanC", lavaan_cache_env)) {
#       assign("opt.lavaanC", requireNamespace("lavaanC", quietly = TRUE),
#               lavaan_cache_env)
#     }
#     return(get("opt.lavaanC", lavaan_cache_env))
#   } else {
#     if (!is.logical(x) || length(x) != 1L)
#       lav_msg_stop(gettext("'x' must be a scalar logical"))
#     if (x) {
#       if (!requireNamespace("lavaanC", quietly = TRUE)) {
#         lav_msg_warn(gettext("cannot use lavaanC, package not found."))
#         assign("opt.lavaanC", FALSE, lavaan_cache_env)
#         return(invisible(NULL))
#       }
#     }
#     assign("opt.lavaanC", x, lavaan_cache_env)
#     return(invisible(NULL))
#   }
# }

# functions to handle warn/debug/verbose options
#             (no longer in 'standard' options)
# if x not present returns the current value of opt_warn/debug/verbose
# if x present
#   if x different from current value, assign x to current value and return TRUE
#   else return FALSE
lav_warn <- function(x) {
  optwarn <- get0("opt_warn", lavaan_cache_env, ifnotfound = TRUE)
  if (missing(x)) {
    optwarn
  } else {
    setwarn <- as.logical(x)
    if (setwarn != optwarn) {
      if (setwarn) {
        # because default TRUE, removing value is the same as setting to TRUE
        rm("opt_warn", envir = lavaan_cache_env)
      } else {
        assign("opt_warn", FALSE, lavaan_cache_env)
      }
      TRUE
    } else {
      FALSE
    }
  }
}
lav_debug <- function(x) {
  optdebug <- get0("opt_debug", lavaan_cache_env, ifnotfound = FALSE)
  if (missing(x)) {
    optdebug
  } else {
    setdebug <- as.logical(x)
    if (setdebug != optdebug) {
      if (setdebug) {
        assign("opt_debug", TRUE, lavaan_cache_env)
      } else {
        # because default FALSE, removing value is the same as setting to FALSE
        rm("opt_debug", envir = lavaan_cache_env)
      }
      TRUE
    } else {
      FALSE
    }
  }
}
lav_verbose <- function(x) {
  optverbose <- get0("opt_verbose", lavaan_cache_env, ifnotfound = FALSE)
  if (missing(x)) {
    optverbose
  } else {
    setverbose <- as.logical(x)
    if (setverbose != optverbose) {
      if (setverbose) {
        assign("opt_verbose", TRUE, lavaan_cache_env)
      } else {
        # because default FALSE, removing value is the same as setting to FALSE
        rm("opt_verbose", envir = lavaan_cache_env)
      }
      TRUE
    } else {
      FALSE
    }
  }
}

# set the default options (including unspecified values "default")
lav_options_default <- function() {
  if (exists("opt_default", lavaan_cache_env)) {
    opt <- get("opt_default", lavaan_cache_env)
    return(opt)
  }
  # ---------------- preparation -----------------
  opt_default <- list()
  opt_check <- list()
  elm <- function(
    name = NULL,          # name of option, if length 2 first is sublist name
    dflt = NULL,          # default value
    chr = NULL,           # valid strings (names) and replacement values
    nm = NULL,            # valid numeric interval
    bl = FALSE,           # logical OK?
    oklen = c(1L, 1L),    # length > 1 OK, second negative to have a warning
    #                and not an error when length greater then abs(oklen[2])
    num2int = FALSE       # should numerical values be converted to int?
  ) {
    stopifnot(any(length(name) == 1:2))
    stopifnot(is.null(nm) || (length(nm) == 1 && grepl("^[][].*,.*[][]$", nm)))
    stopifnot(length(bl) == 1, is.logical(bl))
    stopifnot(is.null(chr) || is.character(chr))
    stopifnot(length(oklen) == 2, oklen[1] <= abs(oklen[2]))
    stopifnot(length(num2int) == 1, is.logical(num2int))

    # prepare list to store for checking option
    list2store <- list(oklen = oklen)
    if (!is.null(chr)) {
      if (is.null(names(chr))) names(chr) <- chr
      list2store$chr <- chr
    }
    if (!is.null(nm)) {
      first_in <- grepl("^\\[", nm)
      last_in <- grepl("\\]$", nm)
      elems <- as.numeric(strsplit(gsub("[][ ]", "", nm), ",")[[1]])
      if (num2int) {
        elems[elems == -Inf] <- -2e9
        elems[elems == Inf] <- 2e9
        elems <- as.integer(elems)
      }
      list2store$nm <- list(bounds = elems,
                            first.in = first_in, last.in = last_in)
    }
    if (bl) list2store$bl <- TRUE
    if (num2int) list2store$num2int <- TRUE

    # store default and list for checking
    if (length(name) == 1) name <- c("", name)
    if (name[1] != "") {
      if (is.null(opt_default[[name[1]]])) { # make sure sublists exist
        opt_default[[name[1]]] <<- list()
        sublist <- list()
        attr(sublist, "SUB") <- TRUE # indicate as sublist
        opt_check[[name[1]]] <<- sublist
      }
      opt_default[[name[1]]][[name[2]]] <<- dflt
      if (is.null(dflt)) opt_default[[name[1]]][name[2]] <<- list(NULL)
      opt_check[[name[1]]][[name[2]]] <<- list2store
    } else {
      opt_default[[name[2]]] <<- dflt
      if (is.null(dflt)) opt_default[name[2]] <<- list(NULL)
      opt_check[[name[2]]] <<- list2store
    }
    NULL
  }
  elmdup <- function(
    name = NULL, # name(s) of option
    from = NULL  # name(s) of option to duplicate
  ) {
    if (length(name) == 1) name <- c("", name)
    if (length(from) == 1) from <- c("", from)
    if (from[1] != "") {
      from_default <- opt_default[[from[1]]][[from[2]]]
      from_check <- opt_check[[from[1]]][[from[2]]]
    } else {
      from_default <- opt_default[[from[2]]]
      from_check <- opt_check[[from[2]]]
    }
    if (name[1] != "") {
      if (is.null(opt_default[[name[1]]])) { # make sure sublists exist
        opt_default[[name[1]]] <<- list()
        sublist <- list()
        attr(sublist, "SUB") <- TRUE # indicate as sublist
        opt_check[[name[1]]] <<- sublist
      }
      opt_default[[name[1]]][[name[2]]] <<- from_default
      opt_check[[name[1]]][[name[2]]] <<- from_check
    } else {
      opt_default[[name[2]]] <<- from_default
      opt_check[[name[2]]] <<- from_check
    }
  }
  # ------------------------- store options --------------------------
  elm("cmd", "lavaan", chr = c(lavaan = "lavaan", cfa = "cfa",
    growth = "growth", sem = "sem"))
  elm("model.type", "sem", chr = c(lavaan = "lavaan", cfa = "cfa",
            growth = "growth", sem = "sem", efa = "efa", path = "path",
            unrestricted = "unrestricted"))

  # global
  elm("mimic", "lavaan", chr = c(default = "lavaan", lavaan = "lavaan",
                                 regression = "lm", lisrel = "EQS",
                                 eqs = "EQS", lm = "lm", mplus = "Mplus"
  ))
  elm("gls.v11.mplus", "default", chr = "default", bl = TRUE)
  elm("gamma.vcov.mplus", "default", chr = "default", bl = TRUE)
  elm("gamma.wls.mplus", "default", chr = "default", bl = TRUE)
  elm("information.expected.mplus", "default", chr = "default", bl = TRUE)
  elm("rmsea.scaled.mplus", "default", chr = "default", bl = TRUE)

  # model modifiers
  elm("meanstructure", "default", chr = "default", bl = TRUE)
  elm("int.ov.free", FALSE, bl = TRUE)
  elm("int.lv.free", FALSE, bl = TRUE)
  elm("marker.int.zero", FALSE, bl = TRUE) # fix marker intercepts
  # free lv means
  elm("conditional.x", "default", chr = "default", bl = TRUE)
  elm("fixed.x", "default", chr = "default", bl = TRUE)
  elm("orthogonal", FALSE, bl = TRUE)
  elm("orthogonal.x", FALSE, bl = TRUE)
  elm("orthogonal.y", FALSE, bl = TRUE)
  elm("std.lv", FALSE, bl = TRUE)
  elm("correlation", FALSE, bl = TRUE) # correlation structure
  elm("effect.coding", FALSE, chr = c("",
                                      "loadings", "intercepts",
                                      "mg.lv.efa.variances", "mg.lv.variances",
                                      "mg.lv.means", "mg.lv.intercepts"),
      bl = TRUE, oklen = c(0L, 6L))
  elm("ceq.simple", FALSE, bl = TRUE) # treat simple eq cons special?
  elm("parameterization", "default", c(
    "default", "mml", "delta", "theta"))
  elm("auto.fix.first", FALSE, bl = TRUE)
  # if > 0 (and auto.fix.first = TRUE), switch to another marker if the first
  # indicator is a poor item; the value is the threshold below which a marker
  # indicator is considered 'poor' (in absolute value of its corrected
  # item-total correlation)
  elm("bad.marker.crit", 0.1, nm = "[0, 1]")
  elm("auto.fix.single", FALSE, bl = TRUE)
  elm("auto.var", FALSE, bl = TRUE)
  elm("auto.cov.lv.x", FALSE, bl = TRUE)
  # if TRUE, also add covariances between exogenous latent variables and
  # observed exogenous covariates (and imply auto.cov.lv.x = TRUE)
  elm("auto.cov.x", FALSE, bl = TRUE)
  elm("auto.cov.y", FALSE, bl = TRUE)
  elm("auto.th", FALSE, bl = TRUE)
  elm("auto.delta", FALSE, bl = TRUE)
  elm("auto.efa", FALSE, bl = TRUE)
  elm("composites", TRUE, bl = TRUE)
  # should the composite-indicator (co)variances (the 'T' matrix) be estimated
  # as free parameters, or fixed to their sample values?
  #   "fixed"   - fix T to the sample (co)variances (default for single-level)
  #   "free"    - estimate T as free parameters (needed for multilevel, where
  #               the same indicators may be composite indicators at >1 level)
  #   "default" - resolved later: "free" if multilevel, "fixed" otherwise
  elm("composites.cov", "default", chr = c("default", "fixed", "free"))

  # rotation
  elm("rotation", "geomin", chr = c(crawfer = "cf", crawford.ferguson = "cf",
    crawfordferguson = "cf", cf = "cf",
    varimax = "varimax", quartimax = "quartimax", orthomax = "orthomax",
    oblimin = "oblimin", quartimin = "quartimin", geomin = "geomin",
    entropy = "entropy", mccammon = "mccammon", infomax = "infomax",
    tandem1 = "tandem1", tandem2 = "tandem2", none = "none", promax = "promax",
    oblimax = "oblimax", bentler = "bentler", simplimax = "simplimax",
    target.strict = "target.strict", target = "pst", pst = "pst",
    cf.quartimax = "cf-quartimax", cf.varimax = "cf-varimax",
    cf.equamax = "cf-equamax", cf.parsimax = "cf-parsimax",
    cf.facparsim = "cf-facparsim", bi.quartimin = "biquartimin",
    biquartimin = "biquartimin", bi.geomin = "bigeomin", bigeomin = "bigeomin"
  ))
  elm("rotation.se", "bordered", chr = c("delta", "bordered"))

  # rotation-args sublist
  elm(c("rotation.args", "orthogonal"), FALSE, bl = TRUE)
  elm(c("rotation.args", "row_weights"), "default", chr = c(
    default = "default", kaiser = "kaiser", none = "none",
    cureton.mulaik = "cm", cm = "cm"))
  elm(c("rotation.args", "std_ov"), TRUE, bl = TRUE)
  elm(c("rotation.args", "geomin_epsilon"), 0.001, nm = "]0, 1.00]")
  # was 0.01 < 0.6-10
  elm(c("rotation.args", "orthomax_gamma"), 1, nm = "[0, 1]")
  elm(c("rotation.args", "cf_gamma"), 0, nm = "[0, 1]")
  elm(c("rotation.args", "oblimin_gamma"), 0, nm = "[0, 1000]")
  elm(c("rotation.args", "promax_kappa"), 4, nm = "[0, 1000]")
  elm(c("rotation.args", "target"), matrix(0, 0L, 0L), oklen = c(0, 1e+32))
  elm(c("rotation.args", "target_mask"), matrix(0, 0L, 0L), oklen = c(0, 1e+32))
  elm(c("rotation.args", "rstarts"), 30L, nm = "[0, 1e+07]")
  elm(c("rotation.args", "algorithm"), "gpa", chr = c("gpa", "pairwise"))
  elm(c("rotation.args", "reflect"), TRUE, bl = TRUE)
  elm(c("rotation.args", "order_lv_by"), "index",
      chr = c("sumofsquares", "index", "none"))
  elm(c("rotation.args", "gpa_tol"), 1e-05, nm = "]0, 0.01]")
  elm(c("rotation.args", "tol"), 1e-08, nm = "]0, 0.01]")
  elm(c("rotation.args", "warn"), FALSE, bl = TRUE)
  elm(c("rotation.args", "verbose"), FALSE, bl = TRUE)
  elm(c("rotation.args", "jac_init_rot"), TRUE, bl = TRUE)
  elm(c("rotation.args", "max_iter"), 10000L, nm = "[0, 1e+12]")
  elm(c("rotation.args", "mg_agreement"), FALSE, bl = TRUE)
  elm(c("rotation.args", "mg_agreement_weight"), 0.5, nm = "[0, 1]")
  elm(c("rotation.args", "mg_agreement_algorithm"), "pairwise",
    chr = c("pairwise", "lg"))
  elm(c("rotation.args", "mg_agreement_method"), "procrustes",
    chr = c("procrustes", "loading.alignment"))

  # full data
  elm("std.ov", FALSE, bl = TRUE)
  elm("missing", "default", chr = c(
    default = "default", ml = "ml", direct = "ml",
    ml.x = "ml.x", direct.x = "ml.x", fiml.x = "ml.x", fiml = "ml",
    two.stage = "two.stage", twostage = "two.stage", two.step = "two.stage",
    twostep = "two.stage", robust.two.stage = "robust.two.stage",
    robust.twostage = "robust.two.stage",
    robust.two.step = "robust.two.stage",
    robust.twostep = "robust.two.stage",
    two.stage.robust = "robust.two.stage",
    twostage.robust = "robust.two.stage",
    two.step.robust = "robust.two.stage",
    twostep.robust = "robust.two.stage",
    listwise = "listwise", pairwise = "pairwise",
    available.cases = "available.cases", doubly.robust = "doubly.robust"))
  elm("sampling.weights.normalization", "group", chr = c(
    "total", "group", "none"))
  elm("sampling.weights.type", "design", chr = c("design", "frequency"))
  elm("samplestats", TRUE, bl = TRUE)

  # summary data
  elm("sample.cov.rescale", "default", bl = TRUE)
  elm("sample.cov.robust", FALSE, bl = TRUE)
  elm("sample.icov", TRUE, bl = TRUE)
  elm("ridge", FALSE, bl = TRUE)
  elm("ridge.constant", "default", chr = "default", nm = "[0, Inf[")

  # multiple groups !!! group.label and group.partial capitals OK !!!
  elm("group.label", NULL, oklen = c(0L, 100L)) # no checks
  elm("group.equal", "", chr =
        c("", "none", "all", "loadings", "intercepts", "means",
          "composite.loadings", "composite.weights",
          "regressions", "residuals", "residual.covariances", "thresholds",
          "lv.variances", "lv.covariances"), oklen = c(0L, 100L))
  elm("group.partial", "", oklen = c(0L, 100L)) # no checks
  elm("group.w.free", FALSE, bl = TRUE)

  # clusters
  elm("level.label", NULL, oklen = c(0L, 100L)) # no checks

  # estimation
  elm("estimator", "default", chr = c(
    none = "none", default = "default", wlsmv = "wlsmv", ml = "ml", mlr = "mlr",
    mlf = "mlf", mlm = "mlm", mlmv = "mlmv", mlmvs = "mlmvs", gls = "gls",
    wls = "wls", wlsm = "wlsm", uls = "uls", ulsm = "ulsm", ulsmv = "ulsmv",
    pml = "pml", dls = "dls", ntrls = "ntrls", catml = "catml",
    dwls = "dwls", wlsmvs = "wlsmvs", ulsmvs = "ulsmvs", fml = "fml",
    umn = "fml", reml = "reml", mml = "mml", fabin = "fabin2",
    fabin2 = "fabin2", fabin3 = "fabin3", mgm = "mgm", guttman = "mgm",
    gutman = "mgm", gutmann = "mgm", guttman1952 = "mgm",
    js = "js", jsa = "jsa", james.stein = "js",
    james.stein.aggregated = "jsa", bentler = "bentler1982",
    bentler1982 = "bentler1982", miiv = "iv", iv = "iv",
    miiv.2sls = "iv", rbm = "rbm"
  ))
  elmdup("estimator.orig", "estimator")

  elm("estimator.args", list(), oklen = c(0L, 100L))
  elm("likelihood", "default", chr = c("default", "normal", "wishart"))
  elm("link", "default", chr = c("default", "logit", "probit"))
  elm("representation", "default", chr = c(
    default = "LISREL", lisrel = "LISREL", ram = "RAM"))
  elm("do.fit", TRUE, bl = TRUE)
  elm("bounds", "none", chr = c(
    "none", "default", "standard", "user", "wide", "wide.zerovar", "pos.var",
    "pos.ov.var", "pos.lv.var"))      # new in 0.6-6
  elm("rstarts", 0L, nm = "[0, 10000]", num2int = TRUE) # new in 0.6-18

  # inference
  elm("se", "default", chr = c(
    default = "default", none = "none", standard = "standard",
    robust.huber.white = "robust.huber.white", robust = "robust",
    robust.cluster = "robust.cluster",
    robust.cluster.sem = "robust.cluster.sem",
    sandwich = "robust.huber.white", robust.sem = "robust.sem",
    robust.sem.nt = "robust.sem.nt",
    two.stage = "two.stage", robust.two.stage = "robust.two.stage",
    bootstrap = "bootstrap", boot = "bootstrap", first.order = "first.order",
    robust.mlm.nt = "robust.sem.nt",
    robust.mlm = "robust.sem", robust.mlr = "robust.huber.white",
    observed = "observed", expected = "expected"),
    oklen = c(1L, -1L)
  )
  elm("test", "default", oklen = c(1L, 100L))
                  # checks for 'test' are in lav_test_rename !!!

  # delta method for defined parameter SEs (:=)
  # FALSE = first-order delta method (default)
  # TRUE  = second-order delta method
  elm("se.delta.second.order", FALSE, bl = TRUE)

  # SE method for defined parameters (:=)
  # "default" = delta method (respects se.delta.second.order)
  # "monte.carlo" = Monte Carlo method (Preacher & Selig 2012)
  elm("se.def", "default", chr = c(
    default = "default",
    delta = "default",
    monte.carlo = "monte.carlo",
    mc = "monte.carlo",
    MC = "monte.carlo"))

  # Monte Carlo settings (used when se.def = "monte.carlo")
  elm("monte.carlo", list(R = 20000L, seed = NULL), oklen = c(0L, 100L))

  # information (se + test)
  elm("information", c("default", "default"), chr = c(
    "default", "expected", "observed", "first.order"),
    oklen = c(1L, 2L))
  elm("h1.information", c("structured", "structured"), chr = c(
    "structured", "unstructured"), oklen =  c(1L, 2L))
  # multilevel h1 estimation with missing data (new in 0.7-1):
  # "em" (mimics Mplus) or "fiml" (cholesky + nlminb; old behavior)
  elm("h1.missing.method", "em", chr = c("em", "fiml"))
  elm("observed.information", c("hessian", "default"), chr = c(
    "default", "hessian", "h1"), oklen =  c(1L, 2L))

  # information se only
  elm("information.meat", "default",
      chr = c(default = "first.order", first.order = "first.order"))
  elm("h1.information.meat", "default", chr = c(
    "default", "structured", "unstructured"))

  # information for 'Omega' (yuan-benter test only)
  # ("first.order" is what omega.information becomes when information[2] is
  #  "first.order", eg estimator = "MLF"; it must be accepted here so that
  #  already-processed options can be fed back into lav_options_set(), as
  #  sam() does)
  elm("omega.information", "default", chr = c(
    "default", "expected", "observed", "first.order"
  ))
  elm("omega.h1.information", "default", chr = c(
    "default", "structured", "unstructured"
  ))
  elm("omega.information.meat", "default", chr = c(
    default = "first.order", first.order = "first.order"
  ))
  elm("omega.h1.information.meat", "default", chr = c(
    "default", "structured", "unstructured"
  ))

  # test statistic for fit measures
  elm("standard.test", "standard", oklen = c(1L, 100L))

  # test statistic for scaling
  elm("scaled.test", "standard", oklen = c(1L, 100L))

  # old approach trace.UGamma2
  elm("ug2.old.approach", FALSE, bl = TRUE)

  # bootstrap
  elm("bootstrap", list(R = 1000L))

  # gamma
  elm("gamma.n.minus.one", FALSE, bl = TRUE)
  elm("gamma.unbiased", FALSE, bl = TRUE)

  # optimization
  elm("control", list(), oklen = c(0L, 100L))
  elm("optim.method", "default", chr = c(
    "nlminb", "nlminb0", "nlminb1", "nlminb2",
    "bfgs", "l.bfgs.b", "gn", "default", "noniter", "none", "em"
  )) # gn for DLS, nlminb rest
  elm("optim.attempts", 4L, nm = "[1, 4]")
  elm("optim.force.converged", FALSE, bl = TRUE)
  # multilevel: fix parameters of saturated blocks at their h1 values
  # during optimization (new in 0.7-1)
  elm("optim.fix.saturated", TRUE, bl = TRUE)
  elm("optim.gradient", "analytic", chr = c(
    analytic = "analytic", analytical = "analytic",
    numeric = "numerical", numerical = "numerical"
  ))
  elm("optim.init_nelder_mead", FALSE, bl = TRUE)
  elm("optim.var.transform", "none", chr = c(
    "none", "sqrt"
  ))
  elm("optim.parscale", "none", chr = c(
    none = "none", st = "standardized", stand = "standardized",
    standardize = "standardized", standardized = "standardized"
  ))
  elm("optim.partrace", FALSE, bl = TRUE)
  elm("optim.dx.tol", 1e-03, nm = "]0, 0.01]") # not too strict
  elm("optim.bounds", list(), oklen = c(0L, 100L))
  # em-args sublist (EM algorithm for the h0 model; optim.method = "em")
  elm(c("em.args", "max_iter"), 10000L, nm = "[100, 1e8]", num2int = TRUE)
  elm(c("em.args", "fx_tol"), 1e-08, nm = "]0, 0.01]")
  elm(c("em.args", "dx_tol"), 1e-04, nm = "]0, 0.01]")
  elm(c("em.args", "zerovar_offset"), 0.0001, nm = "]0, 0.01]")
  # accelerate the EM iterations?
  elm(c("em.args", "acceleration"), "squarem",
      chr = c("squarem", "qn", "none"))
  # compute the loglikelihood as a byproduct of the E-step?
  # (two-level + missing data, plain iterations only)
  elm(c("em.args", "fused"), TRUE, bl = TRUE)
  # if the EM iterations end with a non-negligible gradient (a 'stall'),
  # refine the solution with nlminb, warm-started at the EM values?
  # (the EM solution is kept if nlminb fails to converge, or ends up
  # at a worse solution)
  elm(c("em.args", "nlminb_handoff"), TRUE, bl = TRUE)
  # if the nlminb hand-off is active, stop the EM phase early (using a
  # loosened fx_tol) and let nlminb finish? (a safety net reverts to
  # the strict fx_tol if nlminb fails)
  elm(c("em.args", "early_handoff"), TRUE, bl = TRUE)
  # if verbose = TRUE, also show the optimizer output of each (inner)
  # M-step? (off by default, so only the EM/SQUAREM cycles are shown)
  elm(c("em.args", "verbose"), FALSE, bl = TRUE)

  # em-h1-args sublist (EM algorithm for the unrestricted (h1) model)
  # "default" values are resolved later: single-level (missing data) uses
  # max_iter = 500L and tol = 1e-05 (tol checks the change in the parameter
  # values), multilevel uses max_iter = 5000L and tol = 1e-04 (tol checks
  # the change in the loglikelihood)
  elm(c("em.h1.args", "max_iter"), "default", chr = "default",
      nm = "[10, 1e7]", num2int = TRUE)
  elm(c("em.h1.args", "tol"), "default", chr = "default", nm = "]0, 0.01]")
  # minimum value a variance can take (multilevel only)
  elm(c("em.h1.args", "min_variance"), 1e-05, nm = "]0, 0.01]")
  elm(c("em.h1.args", "warn"), TRUE, bl = TRUE)
  # what to do if an EM estimated covariance matrix is (near) singular?
  elm(c("em.h1.args", "non_pd_action"), "warn", # "stop" breaks jmv
      chr = c("stop", "warn", "none"))
  elm(c("em.h1.args", "non_pd_tol"), 1e-05, nm = "]0, 0.01]")
  # accelerate the EM iterations?
  elm(c("em.h1.args", "acceleration"), "squarem",
      chr = c("squarem", "qn", "none"))
  # compute the loglikelihood as a byproduct of the E-step?
  # (two-level + missing data, plain iterations only)
  elm(c("em.h1.args", "fused"), TRUE, bl = TRUE)
  # gn-args sublist (Gauss-Newton; optim.method = "gn")
  # (< 0.7-1 these lived at the top level, as optim.gn.max.iter,
  #  optim.gn.tol.x and optim.gn.stephalf.max; the latter is gone: the
  #  Levenberg-Marquardt damping replaces the step-halving)
  elm(c("gn.args", "max_iter"), 200L, nm = "[100, 1e8]", num2int = TRUE)
  elm(c("gn.args", "tol_x"), 1e-05, nm = "]0, 0.01]")
  elm(c("gn.args", "tol_g"), 1e-06, nm = "]0, 0.01]")

  # numerical integration
  elm("integration.ngh", 21L, nm = "[1, 1000]", num2int = TRUE)

  # parallel
  elm("parallel", "no", chr = c(
    "no", "multicore", "snow"
  ))
  #maxcpu <- max(1L, parallel::detectCores() - 1L) # crashes if NA is returned
  maxcpu <- max(1L, parallel::detectCores() - 1L, na.rm = TRUE)
  # CRAN policy: never use more than 2 cores by default; users may request
  # more (up to maxcpu) explicitly.
  elm("ncpus", min(2L, maxcpu), nm = paste0("[1,", maxcpu, "]"))
  elm("cl", NULL, oklen = c(0L, 1L))
  elm("iseed", NULL, oklen = c(0L, 1L))

  # categorical
  elm("zero.add", c(0.5, 0.0), chr = "default",
      nm = "[0, 1]", oklen = c(1L, -2L))
  elm("zero.keep.margins", "default", chr = "default", bl = TRUE)
  elm("zero.cell.warn", FALSE, bl = TRUE) # since 0.6-1
  elm("allow.empty.cell", FALSE, bl = TRUE) # since 0.6-19
  elm("cat.wls.w", TRUE, bl = TRUE) # since 0.6-18

  # starting values (char values checked in lav_options_set())
  elm("start", "default", oklen = c(1L, 1000L))

  # sanity checks
  elm("check.start", TRUE, bl = TRUE)
  elm("check.sigma.pd", "chol", chr = c("chol", "eigen")) # stored in env
  elm("check.post", TRUE, bl = TRUE)
  elm("check.gradient", TRUE, bl = TRUE)
  elm("check.vcov", TRUE, bl = TRUE)
  elm("check.lv.names", TRUE, bl = TRUE)
  elm("check.lv.interaction", TRUE, bl = TRUE)
  elm("check.delta.cat.mediator", TRUE, bl = TRUE)

  # more models/info
  elm("h1", TRUE, bl = TRUE)
  elm("baseline", TRUE, bl = TRUE)
  elm("baseline.type", "independence", chr = c(indep = "independence",
    independence = "independence", nested = "nested"))
  elm("baseline.conditional.x.free.slopes", TRUE, bl = TRUE)
  elm("baseline.fixed.x.free.cov", TRUE, bl = TRUE)
  elm("fit.by.level", TRUE, bl = TRUE)
  elm("implied", TRUE, bl = TRUE)
  elm("loglik", TRUE, bl = TRUE)

  # storage of information
  elm("store.vcov", "default", chr = "default", bl = TRUE)

  # internal
  elm("parser", "open", chr = c(old = "old", orig = "old", new = "new",
          open = "open", c.r = "c.r", cr = "c.r", classic = "old"))

  # categorical
  elm("categorical", "default", chr = "default", bl = TRUE)

  # sort list and sublists
  for (nm in names(opt_default)) {
      if (is.list(opt_default[[nm]]))
        opt_default[[nm]] <- opt_default[[nm]][sort(names(opt_default[[nm]]))]
    }
  opt_default <- opt_default[sort(names(opt_default))]
  # ------------- store info in lavaan environment ---------------
  assign("opt_default", opt_default, lavaan_cache_env)
  assign("opt_check", opt_check, lavaan_cache_env)
  assign("opt_check_sigma_pd", "chol", lavaan_cache_env)

  # return defaults
  opt_default
}

# public function
lavOptions <- function(x = NULL, default = NULL, mimic = "lavaan") { # nolint
  lavoptions <- lav_options_default()

  # selection only
  if (!is.null(x)) {
    if (is.character(x)) {
      # lower case and dots only
      x <- lav_option_names(x)

      # check if x is in names(lavoptions)
      not_ok <- which(!x %in% names(lavoptions))
      if (length(not_ok) > 0L) {
        lav_msg_warn(gettextf(
          "option(s) %s not available", lav_msg_view(x[not_ok]))
        )
        x <- x[-not_ok]
      }

      # return requested option(s)
      if (length(x) == 0L) {
        default
      } else {
        lavoptions[x]
      }
    } else {
      lav_msg_stop(gettext("x= must be of type character"))
    }
  } else {
    lavoptions
  }
}

# lav_partable_names
#
# YR. 29 june 2013
#  - as separate file; used to be in utils-user.R
#  - lav_partable_names (aka 'vnames') allows multiple options in 'type'
#    returning them all as a list (or just a vector if only 1 type is needed)

# public version
lav_object_vnames <- function(object, type = "ov", ...) {
  if (inherits(object, "lavaan") || inherits(object, "lavaanList")) {
    # check object
    object <- lav_object_check_version(object)
    partable <- object@ParTable
  } else if (inherits(object, "list") ||
    inherits(object, "data.frame")) {
    partable <- object
  } else if (inherits(object, "character")) {
    # just a model string?
    partable <- lavParseModelString(object)
  }
  lav_pt_vnames(partable, type = type, ...)
}
lavNames <- lav_alias("lav_object_vnames") # synonym #nolint

# return variable names in a partable
# - the 'type' argument determines the status of the variable: observed,
#   latent, endo/exo/...; default = "ov", but most used is type = "all"
#   LDW 30/1/24: there is no default and most used is a single type: "all" is
#                rarely used in other code
# - the 'group' argument either selects a single group (if group is an integer)
#   or returns a list per group
# - the 'level' argument either selects a single level (if level is an integer)
#   or returns a list per level
#   LDW 30/1/24: 'level' argument not explicitly tested !?
# - the 'block' argument either selects a single block (if block is an integer)
#   or returns a list per block
#   LDW 30/1/24: 'block' argument not explicitly tested !?
#   LDW 29/2/24: ov.order = "data" via attribute "ovda"
# - YR 11/02/25: add type = "lv.ho" for higher-order latent variables
lav_pt_vnames <- function(partable, type = NULL, ...,
                                force_warn = FALSE, ov_x_fatal = FALSE) {
  # This function derives the names of some types of variable (as specified
  # in type) from a 'partable'. The 'warn' parameter needs no explanation.
  # The ov.x.fatal parameter implies, when set to TRUE, that the function
  # issues a 'stop' when there are exogenous variables present in variance/
  # covariance or intercept formulas.
  # The call of this function can also contain extra parameters (...) which
  # have to be the name(s) of blockvariable(s) to be used to select names.
  # If more than 1 blockvariable given, all must be satisfied to select!
  # The 'partable' must be a list with minimum members lhs, op, rhs.
  # Other members of 'partable' used (if present): block, ustart, free, exo,
  #                                                user, efa, rv, 'blockname'
  # If the 'partable' contains an attribute vnames and the type is not "*",
  # the 'type' elements of this attribute are used.

  # ----- lav_pt_vnames ---- common ----------------------------------
  # sanity check
  stopifnot(is.list(partable), !missing(type))
  # this is a special function where the default is to suppress warnings,
  # overwritten if parameter force.warn TRUE (used in lav_partable function)
  current_warn <- lav_warn()
  if (force_warn) {
    if (lav_warn(TRUE)) on.exit(lav_warn(current_warn))
  } else {
    if (lav_warn(FALSE)) on.exit(lav_warn(current_warn))
  }
  # check for empty table
  if (length(partable$lhs) == 0) {
    return(character(0L))
  }
  # dotdotdot
  dotdotdot <- list(...)
  type_list <- c(
    "ov", # observed variables (ov)
    "ov.x", # (pure) exogenous observed variables
    "ov.nox", # non-exogenous observed variables
    "ov.model", # modeled observed variables (joint vs cond)
    "ov.y", # (pure) endogenous variables (dependent only)
    "ov.num", # numeric observed variables
    "ov.ord", # ordinal observed variables
    "ov.ind",  # observed indicators of latent variables
    "ov.cind", # observed indicators of composites (new in 0.6-20)
    "ov.orphan", # lonely observed intercepts/variances
    "ov.interaction", # interaction terms (with colon)
    "ov.efa", # indicators involved in efa

    "th", # thresholds ordinal only
    "th.mean", # thresholds ordinal + numeric variables

    "lv", # latent variables
    "lv.regular", # latent variables (defined by =~ only)
    "lv.formative", # latent variables (defined by <~ only) (old style)
    "lv.composite", # latent variables (defined by <~ only) (new style)
    "lv.x", # (pure) exogenous variables
    "lv.y", # (pure) endogenous variables
    "lv.nox", # non-exogenous latent variables
    "lv.nonnormal", # latent variables with non-normal indicators
    "lv.interaction", # interaction terms
    "lv.efa", # latent variables involved in efa
    "lv.rv", # random slopes, random variables
    "lv.ind", # latent indicators (higher-order cfa)
    "lv.ho",  # higher-order latent variables
    "lv.marker", # marker indicator per lv

    "eqs.y", # y's in regression
    "eqs.x" # x's in regression
  )
  if (type[1L] != "all" && type[1L] != "*" && !all(type %in% type_list)) {
    wrongtypes <- type[!(type %in% type_list)]
    lav_msg_stop(sprintf(
      ngettext(length(wrongtypes),
      "type = %s is not a valid option",
      "type = %s are not valid options"),
      lav_msg_view(wrongtypes, "none", FALSE)))
  }
  return_value <- NULL
  if (type[1L] != "*" && !is.null(attr(partable, "vnames"))) {
    # ----- lav_pt_vnames ---- cached data --------------------------
    # uncomment/comment following line to enable/disable trace
    # lav_trace(paste("cached:", paste(type, collapse = ",")))

    if (type[1L] == "all") {
      return_value <- attr(partable, "vnames")
    } else {
      return_value <- attr(partable, "vnames")[type]
    }
  }
  # ----- lav_pt_vnames ---- common ----------------------------------
  if (type[1L] == "all" || type[1L] == "*") {
    type <- type_list
  }
  # ALWAYS need `block' column -- create one if missing
  if (is.null(partable$block)) {
    partable$block <- rep(1L, length(partable$lhs))
  }
  # per default, use full partable
  block_select <- lav_pt_block_values(partable)
  # check for ... selection argument(s)
  ndotdotdot <- length(dotdotdot)
  if (ndotdotdot > 0L) {
    dot_names <- names(dotdotdot)
    row_select <- rep(TRUE, length(partable$lhs))
    for (dot in seq_len(ndotdotdot)) {
      # selection variable?
      block_var <- dot_names[dot]
      block_val <- dotdotdot[[block_var]]
      # do we have this 'block.var' in partable?
      if (is.null(partable[[block_var]])) {
        # for historical reasons, treat "group = 1" special
        if (block_var == "group" && block_val == 1L) {
          partable$group <- rep(1L, length(partable$lhs))
          # remove block == 0
          idx <- which(partable$block == 0L)
          if (length(idx) > 0L) {
            partable$group[idx] <- 0L
          }
          row_select <- (row_select &
            partable[[block_var]] %in% block_val)
        } else {
          lav_msg_stop(gettextf(
            "selection variable '%s' not found in the parameter table.",
            block_var))
        }
      } else {
        if (!all(block_val %in% partable[[block_var]])) {
          lav_msg_stop(gettextf(
            "%1$s column does not contain value `%2$s'", block_var, block_val))
        }
        row_select <- (row_select &
          !partable$op %in% c("==", "<", ">", ":=") &
          partable[[block_var]] %in% block_val)
      }
    } # dot
    block_select <- unique(partable$block[row_select])
    if (length(block_select) == 0L) {
      lav_msg_warn(gettext("no blocks selected."))
    }
  }
  if (is.null(return_value)) {
    # ----- lav_pt_vnames ---- no cache ------------------------

    # uncomment/comment following line to enable/disable trace
    # lav_trace(paste("computed:", paste(type, collapse = ",")))

    # random slope names, if any (new in 0.6-7)
    if (!is.null(partable$rv) && any(nchar(partable$rv) > 0L)) {
      rv_names <- unique(partable$rv[nchar(partable$rv) > 0L])
    } else {
      rv_names <- character(0L)
    }

    # output: list per block
    return_value <- lapply(type, function(x) {
      vector("list", length = length(block_select))
    })
    names(return_value) <- type

    for (b in block_select) {
      # indices for this block
      block_ind <- partable$block == b

      # always compute lv.names
      lv_names <- unique(partable$lhs[block_ind &
        (partable$op == "=~" | partable$op == "<~")])
      # including random slope names
      lv_names2 <- unique(c(lv_names, rv_names))

      # determine lv interactions
      int_names <- unique(partable$rhs[block_ind &
        grepl(":", partable$rhs, fixed = TRUE)])
      n_int <- length(int_names)
      if (n_int > 0L) {
        ok_idx <- logical(n_int)
        for (iv in seq_len(n_int)) {
          tmp_names <- strsplit(int_names[iv], ":", fixed = TRUE)[[1L]]
          # three scenario's:
          # - both variables are latent (ok)
          # - both variables are observed (ignore)
          # - only one latent (warn??) -> upgrade observed to latent
          # thus if at least one is in lv.names, we treat it as a
          # latent interaction
          if (any(tmp_names %in% lv_names)) {
            ok_idx[iv] <- TRUE
          }
        }
        lv_interaction <- int_names[ok_idx]
        lv_names <- c(lv_names, lv_interaction)
        lv_names2 <- c(lv_names2, lv_interaction)
      } else {
        lv_interaction <- character(0L)
      }
      if (length(type) == 1L) {
        # ----- lav_pt_vnames ---- no cache ----- 1 type -----------
        # store lv
        if ("lv" == type) {
          # check if FLAT for random slopes
          # if (!is.null(partable$rv) && any(nchar(partable$rv) > 0L) &&
          #    !is.null(partable$block)) {
          #    return.value$lv[[b]] <- lv.names2
          # } else {
          # here, they will be 'defined' at level 2 as regular =~ lvs
          return_value$lv[[b]] <- lv_names
          # }
          next
        }

        # regular latent variables ONLY (ie defined by =~ only)
        if ("lv.regular" == type) {
          out <- unique(partable$lhs[block_ind &
            partable$op == "=~" &
            partable$lhs != partable$rhs & # no phantom
            !partable$lhs %in% rv_names])
          return_value$lv.regular[[b]] <- out
          next
        }

        # interaction terms involving latent variables (only)
        if ("lv.interaction" == type) {
          return_value$lv.interaction[[b]] <- lv_interaction
          next
        }

        # formative latent variables: phantom lv + zero residual
        if ("lv.formative" == type) {
          out <- unique(partable$lhs[block_ind &
            partable$lhs == partable$rhs &
            partable$op == "=~"])
          rm_idx <- integer(0L)
          for (i in seq_along(out)) {
            var_idx <- which(block_ind &
                             partable$op == "~~" &
                             partable$lhs == out[i] &
                             partable$lhs == partable$rhs)
            if (length(var_idx) == 0L) {
              next
            }
            if (!is.null(partable$mod.idx) && !is.null(partable$fixed)) {
              if (partable$fixed[var_idx] != "0") {
                rm_idx <- c(rm_idx, i)
              }
            } else if (!is.null(partable$free) && !is.null(partable$ustart)) {
              if (partable$free[var_idx] > 0L ||
                  partable$ustart[var_idx] != 0) {
                rm_idx <- c(rm_idx, i)
              }
            }
          }
          if (length(rm_idx) > 0L) {
            out <- out[-rm_idx]
          }
          return_value$lv.formative[[b]] <- out
          next
        }

        # composites defined by "<~"
        if ("lv.composite" == type) {
          out <- unique(partable$lhs[block_ind &
            partable$op == "<~"])
          return_value$lv.composite[[b]] <- out
          next
        }

        # lv's involved in efa
        if (any(type == c("lv.efa", "ov.efa"))) {
          if (is.null(partable$efa)) {
            out <- character(0L)
          } else {
            set_names <- lav_pt_efa_values(partable)
            out <- unique(partable$lhs[partable$op == "=~" &
              block_ind &
              partable$efa %in% set_names])
          }
          if (type == "ov.efa") ov_efa <- out
          if (type == "lv.efa") {
            return_value$lv.efa[[b]] <- out
            next
          }
        }

        # lv's that are random slopes
        if ("lv.rv" == type) {
          if (is.null(partable$rv)) {
            out <- character(0L)
          } else {
            out <- unique(partable$lhs[partable$op == "=~" &
              block_ind &
              partable$lhs %in% rv_names])
          }
          return_value$lv.rv[[b]] <- out
          next
        }

        # lv's that are indicators of a higher-order factor
        if ("lv.ind" == type) {
          out <- unique(partable$rhs[block_ind &
            partable$op == "=~" &
            partable$rhs %in% lv_names &
            partable$lhs != partable$rhs]) # no phantom lv's
          return_value$lv.ind[[b]] <- out
          next
        }

        # higher-order latent variables
        if ("lv.ho" == type) {
          out_ind <- unique(partable$rhs[block_ind &
            partable$op == "=~" &
            partable$rhs %in% lv_names &
            partable$lhs != partable$rhs]) # no phantom lv's
          out <- unique(partable$lhs[block_ind &
            partable$op == "=~" &
            partable$lhs %in% lv_names &
            partable$rhs %in% out_ind])
          return_value$lv.ho[[b]] <- out
          next
        }

        # eqs.y
        if (!(any(type == c("lv", "lv.regular")))) {
          eqs_y <- unique(partable$lhs[block_ind &
            partable$op == "~"])
        }

        # store eqs.y
        if ("eqs.y" == type) {
          return_value$eqs.y[[b]] <- eqs_y
          next
        }

        # eqs.x
        if (!(any(type == c("lv", "lv.regular", "lv.x")))) {
          eqs_x <- unique(partable$rhs[block_ind &
            (partable$op == "~" |
              partable$op == "<~")])
        }

        # store eqs.x
        if ("eqs.x" == type) {
          return_value$eqs.x[[b]] <- eqs_x
          next
        }

        # v.ind -- indicators of latent variables
        if (!(any(type == c("lv", "lv.regular")))) {
          v_ind <- unique(partable$rhs[block_ind &
            partable$op == "=~"])
        }

        # v.cind -- indicators of composites
        if (!(any(type == c("lv", "lv.regular")))) {
          v_cind <- unique(partable$rhs[block_ind &
            partable$op == "<~"])
        }

        # ov.*
        if (!(any(type == c("lv", "lv.regular", "lv.x", "lv.y")))) {
          # 1. indicators, which are not latent variables themselves
          ov_ind <- v_ind[!v_ind %in% lv_names2]
          # 1b. indicator of composites
          ov_cind <- v_cind[!v_cind %in% lv_names2]
          # 2. dependent ov's
          ov_y <- eqs_y[!eqs_y %in% c(lv_names2, ov_ind, ov_cind)]
          # 3. independent ov's
          if (lav_pt_nlevels(partable) > 1L && b > 1L) {
            # NEW in 0.6-8: if an 'x' was an 'y' in a previous level,
            #               treat it as 'y'
            tmp_eqs_y <- unique(partable$lhs[partable$op == "~"]) # all blocks
            ov_x <- eqs_x[!eqs_x %in% c(lv_names2, ov_ind, ov_cind, tmp_eqs_y)]
          } else {
            ov_x <- eqs_x[!eqs_x %in% c(lv_names2, ov_ind, ov_cind, ov_y)]
          }
          # new in 0.6-12: if we have interaction terms in ov.x, check
          # if some terms are in eqs.y; if so, remove the interaction term
          # from ov.x
          int_idx <- which(grepl(":", ov_x, fixed = TRUE))
          bad_idx <- integer(0L)
          for (iv in int_idx) {
            tmp_names <- strsplit(ov_x[iv], ":", fixed = TRUE)[[1L]]
            if (any(tmp_names %in% eqs_y)) {
              bad_idx <- c(bad_idx, iv)
            }
          }
          if (length(bad_idx) > 0L) {
            ov_y <- unique(c(ov_y, ov_x[bad_idx]))
            # it may be removed later, but needed to construct ov.names
            ov_x <- ov_x[-bad_idx]
          }
        }

        # observed variables
        # easy approach would be: everything that is not in lv.names,
        # but the main purpose here is to 'order' the observed variables
        # according to 'type' (indicators, ov.y, ov.x, orphans)
        if (!(any(type == c("lv", "lv.regular", "lv.x", "lv.y")))) {
          # 4. orphaned covariances
          ov_cov <- c(
            partable$lhs[block_ind &
              partable$op == "~~" &
              !partable$lhs %in% lv_names2],
            partable$rhs[block_ind &
              partable$op == "~~" &
              !partable$rhs %in% lv_names2]
          )
          # 5. orphaned intercepts/thresholds
          ov_int <- partable$lhs[block_ind &
            (partable$op == "~1" |
              partable$op == "|") &
            !partable$lhs %in% lv_names2]

          ov_tmp <- c(ov_ind, ov_cind, ov_y, ov_x)
          ov_extra <- unique(c(ov_cov, ov_int)) # must be in this order!
          # so that
          # lav_partable_independence
          # retains the same order
          ov_names <- c(ov_tmp, ov_extra[!ov_extra %in% ov_tmp])
        }

        # store ov?
        if ("ov" == type) {
          return_value$ov[[b]] <- ov_names
          next
        }

        if ("ov.ind" == type) {
          return_value$ov.ind[[b]] <- ov_ind
          next
        }

        if ("ov.cind" == type) {
          return_value$ov.cind[[b]] <- ov_cind
          next
        }

        if ("ov.interaction" == type) {
          ov_int_names <- ov_names[grepl(":", ov_names, fixed = TRUE)]
          n_int <- length(ov_int_names)
          if (n_int > 0L) {
            ov_names_noint <- ov_names[!ov_names %in% ov_int_names]

            ok_idx <- logical(n_int)
            for (iv in seq_len(n_int)) {
              tmp_names <- strsplit(ov_int_names[iv], ":", fixed = TRUE)[[1L]]

              # two scenario's:
              # - both variables are in ov.names.noint (ok)
              # - at least one variable is NOT in ov.names.noint (ignore)
              if (all(tmp_names %in% ov_names_noint)) {
                ok_idx[iv] <- TRUE
              }
            }
            ov_interaction <- ov_int_names[ok_idx]
          } else {
            ov_interaction <- character(0L)
          }

          return_value$ov.interaction[[b]] <- ov_interaction
          next
        }

        if ("ov.efa" == type) {
          ov_efa_1 <- partable$rhs[partable$op == "=~" &
            block_ind &
            partable$rhs %in% ov_ind &
            partable$lhs %in% ov_efa]
          return_value$ov.efa[[b]] <- unique(ov_efa_1)
          next
        }


        # exogenous `x' covariates
        if (any(type == c(
          "ov.x", "ov.nox", "ov.model",
          "th.mean", "lv.nonnormal"
        ))) {
          # correction: is any of these ov.names.x mentioned as a variance,
          #             covariance, or intercept?
          # this should trigger a warning in lav_model_pt()
          if (is.null(partable$user)) { # FLAT!
            partable$user <- rep(1L, length(partable$lhs))
          }
          vars <- c(
            partable$lhs[block_ind &
              partable$op == "~1" &
              partable$user == 1],
            partable$lhs[block_ind &
              partable$op == "~~" &
              partable$user == 1],
            partable$rhs[block_ind &
              partable$op == "~~" &
              partable$user == 1]
          )
          idx_no_x <- which(ov_x %in% vars)
          if (length(idx_no_x)) {
            if (ov_x_fatal) {
              lav_msg_stop(gettextf(
                "model syntax contains variance/covariance/intercept formulas
                involving (an) exogenous variable(s): [%s];  Please remove them
                and try again.", lav_msg_view(ov_x[idx_no_x], "none")))
            }
            lav_msg_warn(gettextf(
                "model syntax contains variance/covariance/intercept formulas
                involving (an) exogenous variable(s): [%s]; these variables will
                now be treated as random introducing additional free parameters.
                If you wish to treat those variables as fixed, remove these
                formulas from the model syntax. Otherwise, consider adding the
                fixed.x = FALSE option.", lav_msg_view(ov_x[idx_no_x], "none")))
            ov_x <- ov_x[-idx_no_x]
          }
          ov_tmp_x <- ov_x

          # extra
          if (!is.null(partable$exo)) {
            ov_cov <- c(
              partable$lhs[block_ind &
                partable$op == "~~" &
                partable$exo == 1L],
              partable$rhs[block_ind &
                partable$op == "~~" &
                partable$exo == 1L]
            )
            ov_int <- partable$lhs[block_ind &
              partable$op == "~1" &
              partable$exo == 1L]
            ov_extra <- unique(c(ov_cov, ov_int))
            ov_tmp_x <- c(ov_tmp_x, ov_extra[!ov_extra %in% ov_tmp_x])
          }

          ov_names_x <- ov_tmp_x
        }

        # store ov.x?
        if ("ov.x" == type) {
          return_value$ov.x[[b]] <- ov_names_x
          next
        }

        # story ov.orphan?
        if ("ov.orphan" == type) {
          return_value$ov.orphan[[b]] <- ov_extra
          next
        }

        # ov's without ov.x
        if (any(type == c(
          "ov.nox", "ov.model",
          "th.mean", "lv.nonnormal"
        ))) {
          ov_names_nox <- ov_names[!ov_names %in% ov_names_x]
        }

        # store ov.nox
        if ("ov.nox" == type) {
          return_value$ov.nox[[b]] <- ov_names_nox
          next
        }

        # store ov.model
        if ("ov.model" == type) {
          # if no conditional.x, this is just ov
          # else, this is ov.nox
          if (any(block_ind & partable$op == "~" &
            partable$exo == 1L)) {
            return_value$ov.model[[b]] <- ov_names_nox
          } else {
            return_value$ov.model[[b]] <- ov_names
          }
          next
        }

        # ov's strictly ordered
        if (any(type == c(
          "ov.ord", "th", "th.mean",
          "ov.num", "lv.nonnormal"
        ))) {
          tmp <- unique(partable$lhs[block_ind &
            partable$op == "|"])
          ord_names <- ov_names[ov_names %in% tmp]
        }

        if ("ov.ord" == type) {
          return_value$ov.ord[[b]] <- ord_names
          next
        }

        # ov's strictly numeric
        if (any(type == c("ov.num", "lv.nonnormal"))) {
          ov_num <- ov_names[!ov_names %in% ord_names]
        }

        if ("ov.num" == type) {
          return_value$ov.num[[b]] <- ov_num
          next
        }

        # nonnormal lv's
        if ("lv.nonnormal" == type) {
          # regular lv's
          lv_reg <- unique(partable$lhs[block_ind &
                           partable$op == "=~" &
                           partable$lhs != partable$rhs])
          if (length(lv_reg) > 0L) {
            out <- unlist(lapply(lv_reg, function(x) {
              # get indicators for this lv
              tmp_ind <- unique(partable$rhs[block_ind &
                partable$op == "=~" &
                partable$lhs == x])
              if (!all(tmp_ind %in% ov_num)) {
                x
              } else {
                character(0)
              }
            }), use.names = FALSE)
            return_value$lv.nonnormal[[b]] <- out
          } else {
            return_value$lv.nonnormal[[b]] <- character(0)
          }
          next
        }

        if (any(c("th", "th.mean") == type)) {
          tmp_th_lhs <- partable$lhs[block_ind &
            partable$op == "|"]
          tmp_th_rhs <- partable$rhs[block_ind &
            partable$op == "|"]
        }

        # threshold
        if ("th" == type) {
          if (length(ord_names) > 0L) {
            # return in the right order (following ord.names!)
            out <- unlist(lapply(ord_names, function(x) {
              idx <- which(x == tmp_th_lhs)
              tmp_th <- unique(paste(tmp_th_lhs[idx], "|",
                tmp_th_rhs[idx],
                sep = ""
              ))
              # make sure the th's are in increasing order
              # sort(tmp.th)
              # NO!, don't do that; t10 will be before t2
              # fixed in 0.6-1 (bug report from Myrsini)
              # in 0.6-12, we do this anyway like this:

              # get var name
              tmp_th1 <- sapply(
                strsplit(tmp_th, split = "\\|t"),
                "[[", 1
              )
              # get number, and sort
              tmp_th2 <- as.character(sort(as.integer(sapply(
                strsplit(tmp_th, split = "\\|t"), "[[", 2
              ))))
              # paste back together in the right order
              paste(tmp_th1, tmp_th2, sep = "|t")
            }), use.names = FALSE)
          } else {
            out <- character(0L)
          }
          return_value$th[[b]] <- out
          next
        }

        # thresholds and mean/intercepts of numeric variables
        if ("th.mean" == type) {
          # if fixed.x -> use ov.names.nox
          # else -> use ov.names
          if (is.null(partable$exo) || all(partable$exo == 0L)) {
            tmp_ov_names <- ov_names
          } else {
            tmp_ov_names <- ov_names_nox
          }
          if (length(tmp_ov_names) > 0L) {
            # return in the right order (following ov.names.nox!)
            out <- unlist(lapply(tmp_ov_names, function(x) {
              if (x %in% ord_names) {
                idx <- which(x == tmp_th_lhs)
                tmp_th <- unique(
                  paste(tmp_th_lhs[idx], "|",
                    tmp_th_rhs[idx],
                    sep = ""
                  ),
                  use.names = FALSE
                )
                # make sure the th's are in increasing order
                # get var name
                tmp_th1 <- sapply(
                  strsplit(tmp_th, split = "\\|t"),
                  "[[", 1
                )
                # get number, and sort
                tmp_th2 <- as.character(sort(as.integer(sapply(
                  strsplit(tmp_th, split = "\\|t"), "[[", 2
                ))))
                # paste back together in the right order
                paste(tmp_th1, tmp_th2, sep = "|t")
              } else {
                x
              }
            }))
          } else {
            out <- character(0L)
          }
          return_value$th.mean[[b]] <- out
          next
        }

        # exogenous lv's
        if (any(c("lv.x", "lv.nox") == type)) {
          tmp <- lv_names[!lv_names %in% c(v_ind, eqs_y)]
          lv_names_x <- lv_names[lv_names %in% tmp]
        }

        if ("lv.x" == type) {
          return_value$lv.x[[b]] <- lv_names_x
          next
        }

        # dependent ov (but not also indicator or x)
        if ("ov.y" == type) {
          tmp <- eqs_y[!eqs_y %in% c(v_ind, eqs_x, lv_names)]
          return_value$ov.y[[b]] <- ov_names[ov_names %in% tmp]
          next
        }

        # dependent lv (but not also indicator or x)
        if ("lv.y" == type) {
          tmp <- eqs_y[!eqs_y %in% c(v_ind, eqs_x) &
            eqs_y %in% lv_names]
          return_value$lv.y[[b]] <- lv_names[lv_names %in% tmp]
          next
        }

        # non-exogenous latent variables
        if ("lv.nox" == type) {
          return_value$lv.nox[[b]] <- lv_names[!lv_names %in% lv_names_x]
          next
        }

        # marker indicator (if any) for each lv
        if ("lv.marker" == type) {
          # default: "" per lv
          out <- character(length(lv_names))
          names(out) <- lv_names
          for (l in seq_along(lv_names)) {
            this_lv_name <- lv_names[l]
            # try to see if we can find a 'marker' indicator for this factor
            # here defined as: has fixed-to-one factor loading
            marker_idx <- which(block_ind &
              partable$op == "=~" &
              partable$lhs == this_lv_name &
              partable$rhs %in% v_ind &
              partable$ustart == 1L &
              partable$free == 0L)
            if (length(marker_idx) > 0L) {
              # we may have multiple potential markers
              potential_markers <- partable$rhs[marker_idx]
              valid_markers <- character(0L)
              for (m in seq_along(potential_markers)) {
                this_marker_idx <- marker_idx[m]
                # check if 'other' loadings are fixed to zero
                other_idx <- which(block_ind &
                  partable$op == "=~" &
                  partable$lhs != this_lv_name &
                  partable$rhs == partable$rhs[this_marker_idx] &
                  partable$free == 0L)
                if (length(other_idx) == 0L) {
                  # simple structure, or one factor
                  valid_markers <- c(valid_markers,
                                     partable$rhs[this_marker_idx])
                } else if (all(partable$ustart[other_idx] == 0)) {
                  valid_markers <- c(valid_markers,
                                     partable$rhs[this_marker_idx])
                }
              }
              if (length(valid_markers) > 0L) {
                out[l] <- valid_markers[1L] # pick the first one
              }
            }
          } # l
          return_value$lv.marker[[b]] <- out
        }
      } else {
        # ----- lav_pt_vnames ---- no cache ----- more than 1 type ------
        # store lv
        if (any("lv" == type)) {
          # check if FLAT for random slopes
          # if (!is.null(partable$rv) && any(nchar(partable$rv) > 0L) &&
          #    !is.null(partable$block)) {
          #    return.value$lv[[b]] <- lv.names2
          # } else {
          # here, they will be 'defined' at level 2 as regular =~ lvs
          return_value$lv[[b]] <- lv_names
          # }
        }

        # regular latent variables ONLY (ie defined by =~ only)
        if (any("lv.regular" == type)) {
          out <- unique(partable$lhs[block_ind &
            partable$op == "=~" &
            partable$lhs != partable$rhs & # no phantom
            !partable$lhs %in% rv_names])
          return_value$lv.regular[[b]] <- out
        }


        # interaction terms involving latent variables (only)
        if (any("lv.interaction" == type)) {
          return_value$lv.interaction[[b]] <- lv_interaction
        }

        # formative latent variables: phantom lv + zero residual
        if (any("lv.formative" == type)) {
          out <- unique(partable$lhs[block_ind &
            partable$lhs == partable$rhs &
            partable$op == "=~"])
          rm_idx <- integer(0L)
          for (i in seq_along(out)) {
            var_idx <- which(block_ind &
                             partable$op == "~~" &
                             partable$lhs == out[i] &
                             partable$lhs == partable$rhs)
            if (length(var_idx) == 0L) {
              next
            }
            if (!is.null(partable$mod.idx) && !is.null(partable$fixed)) {
              if (partable$fixed[var_idx] != "0") {
                rm_idx <- c(rm_idx, i)
              }
            } else if (!is.null(partable$free) && !is.null(partable$ustart)) {
              if (partable$free[var_idx] > 0L ||
                  partable$ustart[var_idx] != 0) {
                rm_idx <- c(rm_idx, i)
              }
            }
          }
          if (length(rm_idx) > 0L) {
            out <- out[-rm_idx]
          }
          return_value$lv.formative[[b]] <- out
        }

        # composite latent variables (ie defined by <~)
        if (any("lv.composite" == type)) {
          out <- unique(partable$lhs[block_ind &
            partable$op == "<~"])
          return_value$lv.composite[[b]] <- out
        }

        # lv's involved in efa
        if (any(type %in% c("lv.efa", "ov.efa"))) {
          if (is.null(partable$efa)) {
            out <- character(0L)
          } else {
            set_names <- lav_pt_efa_values(partable)
            out <- unique(partable$lhs[partable$op == "=~" &
              block_ind &
              partable$efa %in% set_names])
          }
          if (any(type == "ov.efa")) ov_efa <- out
          if (any(type == "lv.efa")) return_value$lv.efa[[b]] <- out
        }

        # lv's that are random slopes
        if (any("lv.rv" == type)) {
          if (is.null(partable$rv)) {
            out <- character(0L)
          } else {
            out <- unique(partable$lhs[partable$op == "=~" &
              block_ind &
              partable$lhs %in% rv_names])
          }
          return_value$lv.rv[[b]] <- out
        }

        # lv's that are indicators of a higher-order factor
        if (any("lv.ind" == type)) {
          out <- unique(partable$rhs[block_ind &
            partable$op == "=~" &
            partable$rhs %in% lv_names &
            partable$lhs != partable$rhs]) # no phantom lv's
          return_value$lv.ind[[b]] <- out
        }

        # higher-order latent variables
        if (any("lv.ho" == type)) {
          out_ind <- unique(partable$rhs[block_ind &
            partable$op == "=~" &
            partable$rhs %in% lv_names &
            partable$lhs != partable$rhs]) # no phantom lv's
          out <- unique(partable$lhs[block_ind &
            partable$op == "=~" &
            partable$lhs %in% lv_names &
            partable$rhs %in% out_ind])
          return_value$lv.ho[[b]] <- out
        }

        # eqs.y
        eqs_y <- unique(partable$lhs[block_ind &
          partable$op == "~"])

        # store eqs.y
        if (any("eqs.y" == type)) {
          return_value$eqs.y[[b]] <- eqs_y
        }

        # eqs.x
        eqs_x <- unique(partable$rhs[block_ind &
          (partable$op == "~" |
            partable$op == "<~")])

        # store eqs.x
        if (any("eqs.x" == type)) {
          return_value$eqs.x[[b]] <- eqs_x
        }

        # v.ind -- indicators of latent variables
        v_ind <- unique(partable$rhs[block_ind &
          partable$op == "=~"])

        # v.cind -- indicators of composites
        v_cind <- unique(partable$rhs[block_ind &
          partable$op == "<~"])


        # ov.*
        # 1. indicators, which are not latent variables themselves
        ov_ind <- v_ind[!v_ind %in% lv_names2]
        # 1a. indicators of composites
        ov_cind <- v_cind[!v_cind %in% lv_names2]
        # 2. dependent ov's
        ov_y <- eqs_y[!eqs_y %in% c(lv_names2, ov_ind, ov_cind)]
        # 3. independent ov's
        if (lav_pt_nlevels(partable) > 1L && b > 1L) {
          # NEW in 0.6-8: if an 'x' was an 'y' in a previous level,
          #               treat it as 'y'
          tmp_eqs_y <- unique(partable$lhs[partable$op == "~"]) # all blocks
          ov_x <- eqs_x[!eqs_x %in% c(lv_names2, ov_ind, ov_cind, tmp_eqs_y)]
        } else {
          ov_x <- eqs_x[!eqs_x %in% c(lv_names2, ov_ind, ov_cind, ov_y)]
        }
        # new in 0.6-12: if we have interaction terms in ov.x, check
        # if some terms are in eqs.y; if so, remove the interaction term
        # from ov.x
        int_idx <- which(grepl(":", ov_x, fixed = TRUE))
        bad_idx <- integer(0L)
        for (iv in int_idx) {
          tmp_names <- strsplit(ov_x[iv], ":", fixed = TRUE)[[1L]]
          if (any(tmp_names %in% eqs_y)) {
            bad_idx <- c(bad_idx, iv)
          }
        }
        if (length(bad_idx) > 0L) {
          ov_y <- unique(c(ov_y, ov_x[bad_idx]))
          # it may be removed later, but needed to construct ov.names
          ov_x <- ov_x[-bad_idx]
        }

        # observed variables
        # easy approach would be: everything that is not in lv.names,
        # but the main purpose here is to 'order' the observed variables
        # according to 'type' (indicators, ov.y, ov.x, orphans)

        # 4. orphaned covariances
        ov_cov <- c(
          partable$lhs[block_ind &
            partable$op == "~~" &
            !partable$lhs %in% lv_names2],
          partable$rhs[block_ind &
            partable$op == "~~" &
            !partable$rhs %in% lv_names2]
        )
        # 5. orphaned intercepts/thresholds
        ov_int <- partable$lhs[block_ind &
          (partable$op == "~1" |
            partable$op == "|") &
          !partable$lhs %in% lv_names2]

        ov_tmp <- c(ov_ind, ov_cind, ov_y, ov_x)
        ov_extra <- unique(c(ov_cov, ov_int)) # must be in this order!
        # so that
        # lav_partable_independence
        # retains the same order
        ov_names <- c(ov_tmp, ov_extra[!ov_extra %in% ov_tmp])

        # store ov?
        if (any("ov" == type)) {
          return_value$ov[[b]] <- ov_names
        }

        if (any("ov.ind" == type)) {
          return_value$ov.ind[[b]] <- ov_ind
        }

        if (any("ov.cind" == type)) {
          return_value$ov.cind[[b]] <- ov_cind
        }

        if (any("ov.interaction" == type)) {
          ov_int_names <- ov_names[grepl(":", ov_names, fixed = TRUE)]
          n_int <- length(ov_int_names)
          if (n_int > 0L) {
            ov_names_noint <- ov_names[!ov_names %in% ov_int_names]

            ok_idx <- logical(n_int)
            for (iv in seq_len(n_int)) {
              tmp_names <- strsplit(ov_int_names[iv], ":", fixed = TRUE)[[1L]]

              # two scenario's:
              # - both variables are in ov.names.noint (ok)
              # - at least one variable is NOT in ov.names.noint (ignore)
              if (all(tmp_names %in% ov_names_noint)) {
                ok_idx[iv] <- TRUE
              }
            }
            ov_interaction <- ov_int_names[ok_idx]
          } else {
            ov_interaction <- character(0L)
          }

          return_value$ov.interaction[[b]] <- ov_interaction
        }

        if (any("ov.efa" == type)) {
          ov_efa_1 <- partable$rhs[partable$op == "=~" &
            block_ind &
            partable$rhs %in% ov_ind &
            partable$lhs %in% ov_efa]
          return_value$ov.efa[[b]] <- unique(ov_efa_1)
        }


        # exogenous `x' covariates
        if (any(type %in% c(
          "ov.x", "ov.nox", "ov.model",
          "th.mean", "lv.nonnormal"
        ))) {
          # correction: is any of these ov.names.x mentioned as a variance,
          #             covariance, or intercept?
          # this should trigger a warning in lav_model_pt()
          if (is.null(partable$user)) { # FLAT!
            partable$user <- rep(1L, length(partable$lhs))
          }
          vars <- c(
            partable$lhs[block_ind &
              partable$op == "~1" &
              partable$user == 1],
            partable$lhs[block_ind &
              partable$op == "~~" &
              partable$user == 1],
            partable$rhs[block_ind &
              partable$op == "~~" &
              partable$user == 1]
          )
          idx_no_x <- which(ov_x %in% vars)
          if (length(idx_no_x)) {
            if (ov_x_fatal) {
              lav_msg_stop(gettextf(
                "model syntax contains variance/covariance/intercept formulas
                involving (an) exogenous variable(s): [%s];  Please remove them
                and try again.", lav_msg_view(ov_x[idx_no_x], "none")))
            }
              lav_msg_warn(gettextf(
                "model syntax contains variance/covariance/intercept formulas
                involving (an) exogenous variable(s): [%s]; these variables will
                now be treated as random introducing additional free parameters.
                If you wish to treat those variables as fixed, remove these
                formulas from the model syntax. Otherwise, consider adding the
                fixed.x = FALSE option.", lav_msg_view(ov_x[idx_no_x], "none")))
            ov_x <- ov_x[-idx_no_x]
          }
          ov_tmp_x <- ov_x

          # extra
          if (!is.null(partable$exo)) {
            ov_cov <- c(
              partable$lhs[block_ind &
                partable$op == "~~" &
                partable$exo == 1L],
              partable$rhs[block_ind &
                partable$op == "~~" &
                partable$exo == 1L]
            )
            ov_int <- partable$lhs[block_ind &
              partable$op == "~1" &
              partable$exo == 1L]
            ov_extra <- unique(c(ov_cov, ov_int))
            ov_tmp_x <- c(ov_tmp_x, ov_extra[!ov_extra %in% ov_tmp_x])
          }

          ov_names_x <- ov_tmp_x
        }

        # store ov.x?
        if (any("ov.x" == type)) {
          return_value$ov.x[[b]] <- ov_names_x
        }

        # story ov.orphan?
        if (any("ov.orphan" == type)) {
          return_value$ov.orphan[[b]] <- ov_extra
        }

        # ov's without ov.x
        if (any(type %in% c(
          "ov.nox", "ov.model",
          "th.mean", "lv.nonnormal"
        ))) {
          ov_names_nox <- ov_names[!ov_names %in% ov_names_x]
        }

        # store ov.nox
        if (any("ov.nox" == type)) {
          return_value$ov.nox[[b]] <- ov_names_nox
        }

        # store ov.model
        if (any("ov.model" == type)) {
          # if no conditional.x, this is just ov
          # else, this is ov.nox
          if (any(block_ind & partable$op == "~" &
            partable$exo == 1L)) {
            return_value$ov.model[[b]] <- ov_names_nox
          } else {
            return_value$ov.model[[b]] <- ov_names
          }
        }

        # ov's strictly ordered
        if (any(type %in% c(
          "ov.ord", "th", "th.mean",
          "ov.num", "lv.nonnormal"
        ))) {
          tmp <- unique(partable$lhs[block_ind &
            partable$op == "|"])
          ord_names <- ov_names[ov_names %in% tmp]
        }

        if (any("ov.ord" == type)) {
          return_value$ov.ord[[b]] <- ord_names
        }

        # ov's strictly numeric
        if (any(type %in% c("ov.num", "lv.nonnormal"))) {
          ov_num <- ov_names[!ov_names %in% ord_names]
        }

        if (any("ov.num" == type)) {
          return_value$ov.num[[b]] <- ov_num
        }

        # nonnormal lv's
        if (any("lv.nonnormal" == type)) {
          # regular lv's
          lv_reg <- unique(partable$lhs[block_ind &
                           partable$op == "=~" &
                           partable$lhs != partable$rhs])
          if (length(lv_reg) > 0L) {
            out <- unlist(lapply(lv_reg, function(x) {
              # get indicators for this lv
              tmp_ind <- unique(partable$rhs[block_ind &
                partable$op == "=~" &
                partable$lhs == x])
              if (!all(tmp_ind %in% ov_num)) {
                x
              } else {
                character(0)
              }
            }), use.names = FALSE)
            return_value$lv.nonnormal[[b]] <- out
          } else {
            return_value$lv.nonnormal[[b]] <- character(0)
          }
        }

        if (any(c("th", "th.mean") %in% type)) {
          tmp_th_lhs <- partable$lhs[block_ind &
            partable$op == "|"]
          tmp_th_rhs <- partable$rhs[block_ind &
            partable$op == "|"]
        }

        # threshold
        if (any("th" == type)) {
          if (length(ord_names) > 0L) {
            # return in the right order (following ord.names!)
            out <- unlist(lapply(ord_names, function(x) {
              idx <- which(x == tmp_th_lhs)
              tmp_th <- unique(paste(tmp_th_lhs[idx], "|",
                tmp_th_rhs[idx],
                sep = ""
              ))
              # make sure the th's are in increasing order
              # sort(tmp.th)
              # NO!, don't do that; t10 will be before t2
              # fixed in 0.6-1 (bug report from Myrsini)
              # in 0.6-12, we do this anyway like this:

              # get var name
              tmp_th1 <- sapply(
                strsplit(tmp_th, split = "\\|t"),
                "[[", 1
              )
              # get number, and sort
              tmp_th2 <- as.character(sort(as.integer(sapply(
                strsplit(tmp_th, split = "\\|t"), "[[", 2
              ))))
              # paste back together in the right order
              paste(tmp_th1, tmp_th2, sep = "|t")
            }), use.names = FALSE)
          } else {
            out <- character(0L)
          }
          return_value$th[[b]] <- out
        }

        # thresholds and mean/intercepts of numeric variables
        if (any("th.mean" == type)) {
          # if fixed.x -> use ov.names.nox
          # else -> use ov.names
          if (is.null(partable$exo) || all(partable$exo == 0L)) {
            tmp_ov_names <- ov_names
          } else {
            tmp_ov_names <- ov_names_nox
          }
          if (length(tmp_ov_names) > 0L) {
            # return in the right order (following ov.names.nox!)
            out <- unlist(lapply(tmp_ov_names, function(x) {
              if (x %in% ord_names) {
                idx <- which(x == tmp_th_lhs)
                tmp_th <- unique(paste(tmp_th_lhs[idx], "|",
                  tmp_th_rhs[idx],
                  sep = ""
                ))
                # make sure the th's are in increasing order
                # get var name
                tmp_th1 <- sapply(
                  strsplit(tmp_th, split = "\\|t"),
                  "[[", 1
                )
                # get number, and sort
                tmp_th2 <- as.character(sort(as.integer(sapply(
                  strsplit(tmp_th, split = "\\|t"), "[[", 2
                ))))
                # paste back together in the right order
                paste(tmp_th1, tmp_th2, sep = "|t")
              } else {
                x
              }
            }), use.names = FALSE)
          } else {
            out <- character(0L)
          }
          return_value$th.mean[[b]] <- out
        }

        # exogenous lv's
        if (any(c("lv.x", "lv.nox") %in% type)) {
          tmp <- lv_names[!lv_names %in% c(v_ind, eqs_y)]
          lv_names_x <- lv_names[lv_names %in% tmp]
        }

        if (any("lv.x" == type)) {
          return_value$lv.x[[b]] <- lv_names_x
        }

        # dependent ov (but not also indicator or x)
        if (any("ov.y" == type)) {
          tmp <- eqs_y[!eqs_y %in% c(v_ind, eqs_x, lv_names)]
          return_value$ov.y[[b]] <- ov_names[ov_names %in% tmp]
        }

        # dependent lv (but not also indicator or x)
        if (any("lv.y" == type)) {
          tmp <- eqs_y[!eqs_y %in% c(v_ind, eqs_x) &
            eqs_y %in% lv_names]
          return_value$lv.y[[b]] <- lv_names[lv_names %in% tmp]
        }

        # non-exogenous latent variables
        if (any("lv.nox" == type)) {
          return_value$lv.nox[[b]] <- lv_names[!lv_names %in% lv_names_x]
        }

        # marker indicator (if any) for each lv
        if (any("lv.marker" == type)) {
          # default: "" per lv
          out <- character(length(lv_names))
          names(out) <- lv_names
          for (l in seq_along(lv_names)) {
            this_lv_name <- lv_names[l]
            # try to see if we can find a 'marker' indicator for this factor
            # here defined as: has a fixed-to-one factor loading
            marker_idx <- which(block_ind &
              partable$op == "=~" &
              partable$lhs == this_lv_name &
              partable$rhs %in% v_ind &
              partable$ustart == 1L &
              partable$free == 0L)
            if (length(marker_idx) > 0L) {
              # we may have multiple potential markers
              potential_markers <- partable$rhs[marker_idx]
              valid_markers <- character(0L)
              for (m in seq_along(potential_markers)) {
                this_marker_idx <- marker_idx[m]
                # check if 'other' loadings are fixed to zero
                other_idx <- which(block_ind &
                  partable$op == "=~" &
                  partable$lhs != this_lv_name &
                  partable$rhs == partable$rhs[this_marker_idx] &
                  partable$free == 0L)
                if (length(other_idx) == 0L) {
                  # simple structure, or one factor
                  valid_markers <- c(valid_markers,
                                     partable$rhs[this_marker_idx])
                } else if (all(partable$ustart[other_idx] == 0)) {
                  valid_markers <- c(valid_markers,
                                     partable$rhs[this_marker_idx])
                }
              }
              if (length(valid_markers) > 0L) {
                out[l] <- valid_markers[1L] # pick the first one
              }
            }
          } # l
          return_value$lv.marker[[b]] <- out
        }
      }
    } # b

    # ----- lav_pt_vnames ---- no cache ------------------------
    # new in 0.6-14: if 'da' operator, change order! (for ov_order = "data")
    # now via attribute "ovda"
    ov_names_data <- attr(partable, "ovda")
    if (!is.null(ov_names_data)) {
      return_value <- lapply(return_value, function(x) {
        for (b in seq_along(x)) {
          # match against the base ov name: threshold-type names look like
          # "y1|t1", so strip the "|..." suffix before matching (plain ov
          # names are unaffected). This keeps "th" ordering consistent with
          # "ov" ordering under ov_order = "data" (order() is stable, so
          # t1, t2, ... keep their relative order within a variable).
          base_name <- sub("\\|.*$", "", x[[b]])
          m <- match(base_name, ov_names_data)
          target_idx <- which(!is.na(m))
          if (length(target_idx) > 1L) {
            x[[b]][target_idx] <- x[[b]][target_idx][order(m[target_idx])]
          }
        }
        x
      })
    }
  }
  # ----- lav_pt_vnames ---- common ------ 1 type --------
  # to mimic old behaviour, if length(type) == 1L
  if (length(type) == 1L) {
    return_value <- return_value[[type]]
    # to mimic old behaviour, if specific block is requested
    if (ndotdotdot == 0L) {
      if (type == "lv.marker") {
        return_value <- unlist(return_value)
        # no unique(), as unique() drops attributes, and reduces
        # c("", "", "") to a single ""
        # (but, say for 2 groups, you get 2 copies)
        # as this is only for 'display', we leave it like that
      } else {
        return_value <- unique(unlist(return_value))
      }
    } else if (length(block_select) == 1L) {
      return_value <- return_value[[block_select]]
    } else {
      return_value <- return_value[block_select]
    }
  }
  # ----- lav_pt_vnames ---- common ------------------------
  return_value
}

# this code is written by Michael Hallquist
# First draft of parser to convert Mplus model syntax to lavaan model syntax

# idea: build parTable and run model from mplus syntax
# then perhaps write export function: parTable2Mplus
# and/or parTable2lavaan

lav_mplus_trim <- function(string) {
  string_trim <- sapply(string, function(x) {
    x <- sub("^\\s*", "", x, perl = TRUE)
    x <- sub("\\s*$", "", x, perl = TRUE)
    x
  }, USE.NAMES = FALSE)
  string_trim
}

# small utility function to join strings in a regexp loop
lav_mplus_join_regex <- function(cmd, arg_expand, matches,
                  iterator, match_length = "match.length") {
  if (iterator == 1 && matches[iterator] > 1) {
    pre <- substr(cmd, 1, matches[iterator] - 1)
  } else {
    pre <- ""
  }

  # if this is not the final match, then get sub-string between the end of
  #                               this match and the beginning of the next
  # otherwise, match to the end of the command
  post_end <- ifelse(iterator < length(matches),
               matches[iterator + 1] - 1, nchar(cmd))
  post <- substr(cmd, matches[iterator] + attr(matches,
               match_length)[iterator], post_end)

  cmd_expand <- paste(pre, arg_expand, post, sep = "")
  cmd_expand
}

# expand Mplus hyphen syntax (will also expand constraints with hyphens)
lav_mplus_expand_cmd <- function(cmd, alpha_start = TRUE) {
  # use negative lookahead and negative lookbehind to eliminate possibility of
  #  hyphen being used as a negative starting value (e.g., x*-1)
  # also avoid match of anything that includes a decimal point, such as a
  #  floating-point starting value -10.5*x1

  # if alphaStart==TRUE, then require that the matches before and after hyphens
  #  begin with alpha character
  # this is used for variable names, whereas the more generic expansion works
  #  for numeric constraints and such

  # need to do a better job of this so that u1-u20* is supported...
  #         I don't think the regexp below is general enough

  # if (alphaStart) {
  #  hyphens <- gregexpr("[_A-Za-z]+\\w*\\s*-\\s*[_A-Za-z]+\\w*",
  #                      cmd, perl=TRUE)[[1]]
  # } else {
  #  hyphens <- gregexpr(
  # "(?!<(\\*|\\.))\\w+(?!(\\*|\\.))\\s*-\\s*(?!<(\\*|\\.))\\w+(?!(\\*|\\.))",
  #  cmd, perl=TRUE)[[1]]
  # }

  # hyphens <- gregexpr(
  # "(?!<(\\*|\\.))\\w+(?!(\\*|\\.))\\s*-\\s*(?!<(\\*|\\.))\\w+(?!(\\*|\\.))",
  #  cmd, perl=TRUE)[[1]]

  # support trailing @XXX. Still still fail on Trait1-Trait3*XXX
  hyphens <- gregexpr(
    "(?!<(\\*|\\.))\\w+(?!(\\*|\\.))\\s*-\\s*(?!<(\\*|\\.))\\w+(?!(\\*|\\.))(@[\\d\\.\\-]+)?", # nolint
     cmd, perl = TRUE)[[1]]

  # Promising, but this is still failing in the case of x3*1 -4.25*x4
  # On either side of a hyphen, require alpha character followed
  # by alphanumeric
  # This enforces that neither side of the hyphen can be a number
  # Alternatively, match digits on either side alone
  # hyphens <- gregexpr(
  # "([A-z]+\\w*\\s*-\\s*[A-z]+\\w*(@[\\d\\.-]+)?|\\d+\\s*-\\s*\\d+)",
  #  cmd, perl=TRUE)[[1]]

  if (hyphens[1L] > 0) {
    cmd_expand <- c()
    ep <- 1

    for (v in seq_along(hyphens)) {
      # match one keyword before and after hyphen
      argsplit <- strsplit(substr(cmd, hyphens[v], hyphens[v] +
        attr(hyphens, "match.length")[v] - 1), "\\s*-\\s*", perl = TRUE)[[1]]

      v_pre <- argsplit[1]
      v_post <- argsplit[2]

      v_post_suffix <-                   # will be empty string if not present
        sub("^([^@]+)(@[\\d\\-.]+)?$", "\\2", v_post, perl = TRUE)
      v_post <- sub("@[\\d\\-.]+$", "", v_post, perl = TRUE) # trim @ suffix

      # If v_pre and v_post contain leading alpha characters,
      #   verify that these prefixes match.
      # Otherwise, there is nothing to expand, as in the case of
      #   MODEL CONSTRAINT: e1e2=e1-e2_n.
      v_pre_alpha <- sub("\\d+$", "", v_pre, perl = TRUE)
      v_post_alpha <- sub("\\d+$", "", v_post, perl = TRUE)

      # only enforce prefix match if we have leading alpha characters
      #  (i.e., not simple numeric 1 - 3 syntax)
      if (length(v_pre_alpha) > 0L && length(v_post_alpha) > 0L) {
        # if alpha prefixes do match, assume that the hyphen is not for
        #  expansion (e.g., in subtraction case)
        if (v_pre_alpha != v_post_alpha) {
          return(cmd)
        }
      }

      # the basic positive lookbehind blows up with pure numeric
      #   constraints (1 - 3) because no alpha char precedes digit
      # can use an non-capturing alternation grouping to allow for digits only
      #  or the final digits after alphas (as in v_post.num)
      v_pre_num <- as.integer(sub(
        "\\w*(?<=[A-Za-z_])(\\d+)$", "\\1", v_pre, perl = TRUE))
         # use positive lookbehind to avoid greedy \w+
         # match -- capture all digits

      v_post_match <- regexpr("^(?:\\w*(?<=[A-Za-z_])(\\d+)|(\\d+))$",
                      v_post, perl = TRUE)
      stopifnot(v_post_match[1L] > 0)

      # match may be under capture[1] or capture[2]
      # because of alternation above
      which_capture <- which(attr(v_post_match, "capture.start") > 0)

      v_post_num <- as.integer(substr(v_post, attr(v_post_match,
         "capture.start")[which_capture],
         attr(v_post_match, "capture.start")[which_capture] +
           attr(v_post_match, "capture.length")[which_capture] - 1))
      v_post_prefix <- substr(v_post, 1, attr(v_post_match,
         "capture.start")[which_capture] - 1)
         # just trusting that pre and post match

      if (is.na(v_pre_num) || is.na(v_post_num)) lav_msg_stop(
        gettext("Cannot expand variables:"), v_pre, ", ", v_post)
      v_expand <- paste(v_post_prefix, v_pre_num:v_post_num, v_post_suffix,
         sep = "", collapse = " ")

      # for first hyphen, there may be non-hyphenated syntax
      #    preceding the initial match
      cmd_expand[ep] <- lav_mplus_join_regex(cmd, v_expand, hyphens, v)

      # This won't really work because the cmd.expand element may contain
      # other variables that are at the beginning or end, prior to hyphen stuff
      # This is superseded by logic above where @ is included in hyphen match,
      # then trapped as suffix I don't think it will work yet for this Mplus
      # syntax: y1-y10*5 -- the 5 wouldn't propagate
      # handle the case of @ fixed values or * starting values used in a list
      # example: Trait1-Trait3@1
      ## if (grepl("@|\\*", cmd.expand[ep], perl=TRUE)) {
      ##   exp_split <- strsplit(cmd.expand[ep], "\\s+", perl=TRUE)[[1]]
      ##   suffixes <- sub("^([^@\\*]+)([@*][\\d\\.-]+)?$", "\\2",
      #                exp_split, perl=TRUE)
      ##   variables <- sub("^([^@\\*]+)([@*][\\d\\.-]+)?$", "\\1",
      #                exp_split, perl=TRUE)
      ##   suffixes <- suffixes[suffixes != ""]
      ##   if (length(unique(suffixes)) > 1L) {
      ##     browser()

      ##     #stop("Don't know how to interpret syntax: ", cmd)
      ##   } else {
      ##     variables <- paste0(variables, suffixes[1])
      ##     cmd.expand[ep] <- paste(variables, collapse=" ")
      ##   }
      ## }

      ep <- ep + 1
    }
    paste(cmd_expand, collapse = "")
  } else {
    cmd # no hyphens to expand
  }
}


# handle starting values and fixed parameters on rhs
lav_mplus_cmd_fix_start <- function(cmd) {
  cmd_parse <- c()
  ep <- 1L

  # support ESEM-like syntax: F BY a1* a2*
  # The easy path: putting in 1s before we proceed on parsing
  # Mar2023 bugfix: support parenthesis after * in case a parameter
  #                      constraint comes next
  cmd <- gsub("([A-z]+\\w*)\\s*\\*(?=\\s+\\(?[A-z]+|\\s*$)", "\\1*1",
             cmd, perl = TRUE)

  if ((fixed_starts <- gregexpr("[\\w\\.\\-$]+\\s*([@*])\\s*[\\w\\.\\-]+",
                                cmd, perl = TRUE)[[1]])[1L] > 0) {
       # shouldn't it be \\*, not * ?! Come back to this.
    for (f in seq_along(fixed_starts)) {
      # capture above obtains the fixed/start character (@ or *),
      # whereas match obtains the full regex match
      opchar <- substr(cmd, attr(fixed_starts, "capture.start")[f],
      attr(fixed_starts, "capture.start")[f] + attr(fixed_starts,
         "capture.length")[f] - 1)

      # match arguments around asterisk/at symbol
      argsplit <- strsplit(substr(cmd, fixed_starts[f], fixed_starts[f] +
        attr(fixed_starts, "match.length")[f] - 1), paste0("\\s*",
        ifelse(opchar == "*", "\\*", opchar), "\\s*"), perl = TRUE)[[1]]
      v_pre <- argsplit[1]
      v_post <- argsplit[2]

      if (suppressWarnings(is.na(as.numeric(v_pre)))) {
        # fixed.starts value post-multiplier
        var <- v_pre
        val <- v_post
      } else if (suppressWarnings(is.na(as.numeric(v_post)))) {
        # starting value pre-multiplier
        var <- v_post
        val <- v_pre
      } else {
        lav_msg_stop(
          gettext("Cannot parse Mplus fixed/starts values specification:"),
          v_pre, v_post)
      }

      if (opchar == "@") {
        cmd_parse[ep] <- lav_mplus_join_regex(cmd,
          paste0(val, "*", var, sep = ""), fixed_starts, f)
        ep <- ep + 1L
      } else {
        cmd_parse[ep] <- lav_mplus_join_regex(cmd,
          paste0("start(", val, ")*", var, sep = ""), fixed_starts, f)
        ep <- ep + 1L
      }
    }
    paste(cmd_parse, collapse = "")
  } else {
    cmd
  }
}

lav_mplus_cmd_con <- function(cmd) {
  # Allow cmd to have newlines embedded. In this case, split on newlines, and
  #  loop over and parse each chunk
  # Dump leading and trailing newlines, which contain no information about
  #  constraints, but may add dummy elements to vector after strsplit
  # Maybe return LHS and RHS parsed command where constraints only appear on
  #  the RHS, whereas the LHS contains only parameters.
  # Example: LHS is v1 v2 v3 and RHS is con1*v1 con2*v2 con3*v3

  cmd_split <- strsplit(cmd, "\n")[[1]]

  # drop empty lines (especially leading newline)
  cmd_split <- if (length(empty_pos <- which(cmd_split == "")) > 0L) {
    cmd_split[-1 * empty_pos]
  } else {
    cmd_split
  }

  # Create a version of the command with no modifiers (constraints,
  #   starting values, etc.) specifications.
  # This is useful for syntax that uses the params on the LHS and with a
  #   modified RHS. Example: v1 ~~ conB*v1
  cmd_nomodifiers <- paste0(gsub("(start\\([^\\)]+\\)\\*|[\\d\\-\\.]+\\*)",
   "", cmd_split, perl = TRUE), collapse = " ") # peel off premultiplication
  cmd_nomodifiers <- gsub("\\([^\\)]+\\)", "", cmd_nomodifiers, perl = TRUE)

  cmd_tojoin <- c()
  # will store all chunks divided by newlines, which will be joined at the end.

  # iterate over each newline segment
  for (n in seq_along(cmd_split)) {
    # in principle, now that we respect newlines, parens should only be of
    # length 1, since Mplus syntax dictates newlines for each use of
    # parentheses for constraints
    if ((parens <- gregexpr("(?<!start)\\(([^\\)]+)\\)", cmd_split[n],
     perl = TRUE)[[1L]])[1L] > 0) { # match parentheses, but not start()
      # the syntax chunk after all parentheses have been matched
      cmd_expand <- c()

      for (p in seq_along(parens)) {
        # string within the constraint parentheses
        constraints <- substr(cmd_split[n], attr(parens, "capture.start")[p],
         attr(parens,
          "capture.start")[p] + attr(parens, "capture.length")[p] - 1)

        # Divide constraints on spaces to determine number of constraints to
        # parse. Use lav_mplus_trim to avoid problem of user including
        # leading/trailing spaces within parentheses.
        con_split <- strsplit(lav_mplus_trim(constraints), "\\s+",
         perl = TRUE)[[1]]

        # if Mplus uses a purely numeric constraint, then add ".con" prefix
        #   to be consistent with R naming.
        con_split <- sapply(con_split, function(x) {
          if (!suppressWarnings(is.na(as.numeric(x)))) {
            make.names(paste0(".con", x))
          } else {
            x
          }
        })

        # determine the parameters that precede the parentheses (either first
        #  character for p == 1 or character after preceding parentheses)
        prestr_start <- ifelse(p > 1, attr(parens, "capture.start")[p - 1] +
          attr(parens, "capture.length")[p - 1] + 1, 1)

        # obtain the parameters that precede the parentheses, divide into
        #  arguments on spaces
        # use lav_mplus_trim here because first char after prestrStart for
        #  p > 1 will probably be a space
        precmd_split <- strsplit(lav_mplus_trim(substr(cmd_split[n],
           prestr_start, parens[p] - 1)), "\\s+", perl = TRUE)[[1]]

        # peel off any potential LHS arguments, such as F1 BY
        precmd_lhsop <- which(tolower(precmd_split) %in% c("by", "with", "on"))
        if (any(precmd_lhsop)) {
          lhsop <- paste0(precmd_split[1:precmd_lhsop[1L]], " ",
                          collapse = " ")
          # join lhs and op as a single string, add trailing space so that
          #  paste with expanded RHS is right.
          rhs <- precmd_split[(precmd_lhsop + 1):length(precmd_split)]
        } else {
          lhsop <- ""
          rhs <- precmd_split
        }

        if (length(con_split) > 1L) {
          # several constraints listed within parentheses.
          #  Example: F1 BY X1 X2 X3 X4 (C2 C3 C4)
          # thus, backwards match the constraints to parameters

          # restrict parameters to backwards match to be of the same length
          #  as number of constraints
          rhs_backmatch <- rhs[(length(rhs) -
                           length(con_split) + 1):length(rhs)]

          rhs_expand <- c()

          # check that no mean or scale markers are part of the rhs param
          #  to expand
          if ((pre_mark_match <- regexpr("^\\s*[\\[\\{]", rhs_backmatch[1L],
                  perl = TRUE))[1L] > 0) {
            pre_mark <- substr(rhs_backmatch[1L], pre_mark_match[1L],
              pre_mark_match[1L] + attr(pre_mark_match,
                "match.length")[1L] - 1)
            rhs_backmatch[1L] <- substr(rhs_backmatch[1L],
              pre_mark_match[1L] + attr(pre_mark_match,
                 "match.length")[1L], nchar(rhs_backmatch[1L]))
          } else {
            pre_mark <- ""
          }

          if ((post_mark_match <- regexpr("[\\]\\}]\\s*$",
             rhs_backmatch[length(rhs_backmatch)], perl = TRUE))[1L] > 0) {
            post_mark <- substr(rhs_backmatch[length(rhs_backmatch)],
             post_mark_match[1L], nchar(rhs_backmatch[length(rhs_backmatch)]))
            rhs_backmatch[length(rhs_backmatch)] <-
              substr(rhs_backmatch[length(rhs_backmatch)], 1,
                       post_mark_match[1L] - 1)
          } else {
            post_mark <- ""
          }


          # pre-multiply each parameter with each corresponding constraint
          for (i in seq_along(rhs_backmatch)) {
            rhs_expand[i] <- paste0(con_split[i], "*", rhs_backmatch[i])
          }

          # join rhs as string and add back in mean/scale operator, if present
          rhs_expand <- paste0(pre_mark, paste(rhs_expand,
             collapse = " "), post_mark)

          # if there were params that preceded the backwards match,
          #  then add these back to the syntax
          # append this syntax to the parsed command, cmd.expand
          if (length(rhs) - length(con_split) > 0L) {
            cmd_expand <- c(cmd_expand, paste(lhsop, paste(rhs[1:(length(rhs) -
                            length(con_split))], collapse = " "), rhs_expand))
          } else {
            cmd_expand <- c(cmd_expand, paste0(lhsop, rhs_expand))
          }
        } else {
          # should be able to reduce redundancy with above

          # all parameters on the right hand side are to be equated
          # thus, pre-multiply each parameter by the constraint

          # check that no mean or scale markers are part of the
          #   rhs param to expand
          # DUPE CODE FROM ABOVE. Make Function?!
          if ((pre_mark_match <-
            regexpr("^\\s*[\\[\\{]", rhs[1L], perl = TRUE))[1L] > 0) {
            pre_mark <- substr(rhs[1L], pre_mark_match[1L], pre_mark_match[1L] +
              attr(pre_mark_match, "match.length")[1L] - 1)
            rhs[1L] <- substr(rhs[1L], pre_mark_match[1L] + attr(pre_mark_match,
               "match.length")[1L], nchar(rhs[1L]))
          } else {
            pre_mark <- ""
          }

          if ((post_mark_match <-
             regexpr("[\\]\\}]\\s*$", rhs[length(rhs)], perl = TRUE))[1L] > 0) {
            post_mark <- substr(rhs[length(rhs)], post_mark_match[1L],
             nchar(rhs[length(rhs)]))
            rhs[length(rhs)] <- substr(rhs[length(rhs)], 1,
             post_mark_match[1L] - 1)
          } else {
            post_mark <- ""
          }


          rhs_expand <- c()
          for (i in seq_along(rhs)) {
            rhs_expand[i] <- paste0(con_split[1L], "*", rhs[i])
          }

          # join rhs as string
          rhs_expand <- paste0(pre_mark,
            paste(rhs_expand, collapse = " "), post_mark)

          cmd_expand <- c(cmd_expand, paste0(lhsop, rhs_expand))
        }
      }

      cmd_tojoin[n] <- paste(cmd_expand, collapse = " ")
    } else {
      cmd_tojoin[n] <- cmd_split[n]
    } # no parens
  }

  # eliminate newlines in this function so that they
  # don't mess up \\s+ splits downstream
  to_return <- paste(cmd_tojoin, collapse = " ")
  attr(to_return, "noModifiers") <- cmd_nomodifiers

  to_return
}

lav_mplus_cmd_growth <- function(cmd) {
  # can assume that any spaces between tscore and variable were stripped by
  #  lav_mplus_cmd_fix_start

  # verify that this is not a random slope
  if (any(tolower(strsplit(cmd, "\\s+", perl = TRUE)[[1]]) %in%
                   c("on", "at"))) {
    lav_msg_stop(gettext(
      "lavaan does not support random slopes or individually varying
      growth model time scores"))
  }

  cmd_split <- strsplit(cmd, "\\s*\\|\\s*", perl = TRUE)[[1]]
  if (!length(cmd_split) == 2) {
    lav_msg_stop(gettext("Unknown growth syntax:"), cmd)
  }

  lhs <- cmd_split[1]
  lhs_split <- strsplit(lhs, "\\s+", perl = TRUE)[[1]]

  rhs <- cmd_split[2]
  rhs_split <- strsplit(rhs, "(\\*|\\s+)", perl = TRUE)[[1]]

  if (length(rhs_split) %% 2 != 0) {
    lav_msg_stop(gettext(
      "Number of variables and number of tscores does not match:"), rhs)
  }
  tscores <- as.numeric(rhs_split[seq_along(rhs_split) %% 2 != 0])
             # pre-multipliers

  vars <- rhs_split[seq_along(rhs_split) %% 2 == 0]

  cmd_expand <- c()

  for (p in 0:(length(lhs_split) - 1)) {
    if (p == 0) {
      # intercept
      cmd_expand <- c(cmd_expand, paste(lhs_split[(p + 1)], "=~",
       paste("1*", vars, sep = "", collapse = " + ")))
    } else {
      cmd_expand <- c(cmd_expand, paste(lhs_split[(p + 1)], "=~",
       paste(tscores^p, "*", vars, sep = "", collapse = " + ")))
    }
  }

  cmd_expand
}

# function to wrap long lines at a certain width,
#  splitting on + symbols to be consistent with R syntax
lav_mplus_cmd_wrap <- function(cmd, width = 90, exdent = 5) {
  result <- lapply(cmd, function(line) {
    if (nchar(line) > width) {
      split <- c()
      spos <- 1L

      plus_match <- gregexpr("+", line, fixed = TRUE)[[1]]
#      mpos <- 1L

      if (plus_match[1L] > 0L) {
        # split after plus symbol
        chars_remain <- nchar(line)
        while (chars_remain > 0L) {
#          to_process <- substr(line, nchar(line) - chars_remain + 1,
#           nchar(line))
          offset <- nchar(line) - chars_remain + 1

          if (nchar(remainder <- substr(line, offset, nchar(line))) <=
                              (width - exdent)) {
            # remainder of line fits within width -- no need
            #            to continue wrapping
            split[spos] <- remainder
            chars_remain <- 0
          } else {
            wrap_at <- which(plus_match < (width + offset - exdent))
            wrap_at <- wrap_at[length(wrap_at)] # at the final +

            split[spos] <- substr(line, offset, plus_match[wrap_at])
            chars_remain <- chars_remain - nchar(split[spos])
            spos <- spos + 1
          }
        }

        # remove leading and trailing chars
        split <- lav_mplus_trim(split)

        # handle exdent
        split <- sapply(seq_along(split), function(x) {
          if (x > 1) {
            paste0(paste(rep(" ", exdent), collapse = ""), split[x])
          } else {
            split[x]
          }
        })

        split
      } else {
        strwrap(line, width = width, exdent = exdent)
           # convention strwrap when no + present
      }
    } else {
      line
    }
  })

  # bind together multi-line expansions into single vector
  unname(do.call(c, result))
}

lav_mplus_syntax_con <- function(syntax) {
  # should probably pass in model syntax along with some tracking
  #  of which parameter labels are defined.

  # convert MODEL CONSTRAINT section to lavaan model syntax
  syntax <- paste(lapply(lav_mplus_trim(strsplit(syntax, "\n")), function(x) {
    if (length(x) == 0L && is.character(x)) "" else x
  }), collapse = "\n")

  # replace ! with # for comment lines.
  #  Also strip newline and replace with semicolon
  syntax <- gsub("(\\s*)!(.+)\n", "\\1#\\2;", syntax, perl = TRUE)

  # split into vector of strings
  # need to peel off leading or trailing newlines -- leads to parsing
  #  confusion downstream otherwise
  syntax_split <- gsub("(^\n|\n$)", "",
                unlist(strsplit(syntax, ";")), perl = TRUE)

  constraint_out <- c()

  # TODO: Handle PLOT and LOOP syntax for model constraints.
  # TODO: Handle DO loop convention

  # first parse new parameters defined in MODEL CONSTRAINT into a vector
  new_parameters <- c() # parameters that are defined in constraint section
  if (length(new_con_lines <- grep("^\\s*NEW\\s*\\([^\\)]+\\)",
        syntax_split, perl = TRUE, ignore.case = TRUE)) > 0L) {
    for (cmd in syntax_split[new_con_lines]) {
      # process new constraint definition
      new_con <- regexpr("^\\s*NEW\\s*\\(([^\\)]+)\\)", cmd,
        perl = TRUE, ignore.case = TRUE)
      if (new_con[1L] == -1)
        lav_msg_stop(gettext("Unable to parse names of new constraints"))
      new_con <- substr(cmd, attr(new_con, "capture.start"),
       attr(new_con, "capture.start") + attr(new_con, "capture.length") - 1L)
      new_con <- lav_mplus_expand_cmd(new_con) # allow for hyphen expansion
      new_parameters <- c(new_parameters, strsplit(lav_mplus_trim(new_con),
          "\\s+", perl = TRUE)[[1L]])
    }

    syntax_split <- syntax_split[-1L * new_con_lines] # drop out these lines
    parameters_undefined <- new_parameters
    # to be used below to handle ambiguity of equation versus definition
  }

  for (cmd in syntax_split) {
    if (grepl("^\\s*#", cmd, perl = TRUE)) { # comment line
      constraint_out <- c(constraint_out, gsub("\n", "", cmd, fixed = TRUE))
      # drop any newlines
    } else if (grepl("^\\s+$", cmd, perl = TRUE)) {
      # do nothing, just a space line
    } else {
      # constraint proper
      cmd <- gsub("**", "^", cmd, fixed = TRUE) # handle exponent

      # lower case the math operations supported by Mplus to be
      #        consistent with R
      # match all math operators, then lower case each and rejoin string
      maths <- gregexpr(
        "(SQRT|LOG|LOG10|EXP|ABS|SIN|COS|TAN|ASIN|ACOS|ATAN)\\s*\\(",
         cmd, perl = TRUE)[[1L]]
      if (maths[1L] > 0) {
        maths_replace <- c()
        ep <- 1

        for (i in seq_along(maths)) {
          operator <- tolower(substr(cmd, attr(maths, "capture.start")[i],
           attr(maths, "capture.start")[i] +
             attr(maths, "capture.length")[i] - 1))
          maths_replace[ep] <- lav_mplus_join_regex(cmd, operator, maths, i,
             match_length = "capture.length")
          # only match operator, not opening (
          ep <- ep + 1
        }
        cmd <- paste(maths_replace, collapse = "")
      }

      # equating some lhs and rhs: could reflect definition of new parameter
      if ((equals <- regexpr("=", cmd, fixed = TRUE))[1L] > 0) {
        lhs <- lav_mplus_trim(substr(cmd, 1, equals - 1))
        rhs <- lav_mplus_trim(substr(cmd, equals +
          attr(equals, "match.length"), nchar(cmd)))

        # possibility of lhs or rhs containing the single variable to be equated
        if (regexpr("\\s+", lhs, perl = TRUE)[1L] > 0L) {
          def <- rhs
          body <- lhs
        } else if (regexpr("\\s+", rhs, perl = TRUE)[1L] > 0L) {
          def <- lhs
          body <- rhs
        } else {
          # warning("Can't figure out which side of constraint
          #         defines a parameter")
          # this would occur for simple rel5 = rel2 sort of syntax
          def <- lhs
          body <- rhs
        }

        # must decide whether this is a new parameter (:=)
        #      or equation of existing labels (==)
        # alternatively, could be zero, as in  0 = x + y
        # this is tricky, because mplus doesn't differentiate
        #  definition from equation
        # consequently, could confuse the issue as in ex5.20
        # NEW(rel2 rel5 stan3 stan6);
        # rel2 = lam2**2*vf1/(lam2**2*vf1 + ve2);
        # rel5 = lam5**2*vf2/(lam5**2*vf2 + ve5);
        # rel5 = rel2;

        # for now, only define a new constraint if it's not already defined
        # otherwise equate
        if (def %in% new_parameters && def %in% parameters_undefined) {
          constraint_out <- c(constraint_out, paste(def, ":=", body))
          parameters_undefined <-
            parameters_undefined[!parameters_undefined == def]
        } else {
          constraint_out <- c(constraint_out, paste(def, "==", body))
        }
      } else {
        # inequality constraints -- paste as is
        constraint_out <- c(constraint_out, cmd)
      }
    }
  }

  wrap <- paste(lav_mplus_cmd_wrap(constraint_out, width = 90, exdent = 5),
     collapse = "\n")
  wrap
}

lav_mplus_syntax_model <- function(syntax) {
  if (is.character(syntax)) {
    if (length(syntax) > 1L) {
      syntax <- paste(syntax, collapse = "\n")
    } # concatenate into a long string separated by newlines
  } else {
    lav_msg_stop(gettext(
      "lav_mplus_syntax_model accepts a single character string or
      character vector containing all model syntax"))
  }

  # because this is now exposed as a function in the package, handle the case of
  #   the user passing in full .inp file as text
  # we should only be interested in the MODEL and MODEL CONSTRAINT sections
  by_line <- strsplit(syntax, "\r?\n", perl = TRUE)[[1]]
  input_headers <- grep(
    "^\\s*(title:|data.*:|variable:|define:|analysis:|model.*:|output:|savedata:|plot:|montecarlo:)", # nolint
     by_line, ignore.case = TRUE, perl = TRUE)
  con_syntax <- c()
  if (length(input_headers) > 0L) {
    # warning("lav_mplus_syntax_model is intended to accept only the model
    #  section, not an entire .inp file. For the .inp file case,
    #  use lav_mplus_lavaan")
    parsed_syntax <- lav_mplus_text_sections(by_line, "local")

    # handle model constraint
    if ("model.constraint" %in% names(parsed_syntax)) {
      con_syntax <- strsplit(lav_mplus_syntax_con(
                  parsed_syntax$model.constraint), "\n")[[1]]
    }

    # just keep model syntax before continuing
    syntax <- parsed_syntax$model
  }

  # initial strip of leading/trailing whitespace, which can interfere with
  #   splitting on spaces
  # strsplit generates character(0) for empty strings, which causes problems in
  #   paste because paste actually includes it as a literal
  # example: paste(list(character(0), "asdf", character(0)), collapse=" ")
  # thus, use lapply to convert these to empty strings first
  syntax <- paste(lapply(lav_mplus_trim(strsplit(syntax, "\n")), function(x) {
    if (length(x) == 0L && is.character(x)) "" else x
  }), collapse = "\n")

  # replace ! with # for comment lines. Also strip newline and replace
  #  with semicolon
  syntax <- gsub("(\\s*)!(.+)\n*", "\\1#\\2;", syntax, perl = TRUE)

  # new direction: retain newlines in parsed syntax until after
  #  constraints have been parsed

  # delete newlines
  # syntax <- gsub("\n", "", syntax, fixed=TRUE)

  # replace semicolons with newlines prior to split (divide into commands)
  # syntax <- gsub(";", "\n", syntax, fixed=TRUE)

  # split into vector of strings
  # syntax.split <- unlist( strsplit(syntax, "\n") )
  syntax_split <- lav_mplus_trim(unlist(strsplit(syntax, ";")))

  # format of parTable to mimic.
  # 'data.frame':	34 obs. of  12 variables:
  #  $ id    : int  1 2 3 4 5 6 7 8 9 10 ...
  #  $ lhs   : chr  "ind60" "ind60" "ind60" "dem60" ...
  #  $ op    : chr  "=~" "=~" "=~" "=~" ...
  #  $ rhs   : chr  "x1" "x2" "x3" "y1" ...
  #  $ user  : int  1 1 1 1 1 1 1 1 1 1 ...
  #  $ group : int  1 1 1 1 1 1 1 1 1 1 ...
  #  $ free  : int  0 1 2 0 3 4 5 0 6 7 ...
  #  $ ustart: num  1 NA NA 1 NA NA NA 1 NA NA ...
  #  $ exo   : int  0 0 0 0 0 0 0 0 0 0 ...
  #  $ label : chr  "" "" "" "" ...
  #  $ eq.id : int  0 0 0 0 0 0 0 0 0 0 ...
  #  $ unco  : int  0 1 2 0 3 4 5 0 6 7 ...

  # vector of lavaan syntax
  lavaan_out <- c()

  for (cmd in syntax_split) {
    if (grepl("^\\s*#", cmd, perl = TRUE)) { # comment line
      lavaan_out <- c(lavaan_out, gsub("\n", "", cmd, fixed = TRUE))
       # drop any newlines (otherwise done by lav_mplus_cmd_con)
    } else if (grepl("^\\s*$", cmd, perl = TRUE)) {
      # do nothing, just a space or blank line
    } else {
      # hyphen expansion
      cmd <- lav_mplus_expand_cmd(cmd)

      # parse fixed parameters and starting values
      cmd <- lav_mplus_cmd_fix_start(cmd)

      # parse any constraints here (avoid weird logic below)
      cmd <- lav_mplus_cmd_con(cmd)

      if ((op <- regexpr("\\s+(by|on|with|pwith)\\s+", cmd, ignore.case = TRUE,
       perl = TRUE))[1L] > 0) { # regressions, factors, covariances

        lhs <- substr(cmd, 1, op - 1)
        # using op takes match.start which will omit spaces before operator
        rhs <- substr(cmd, op + attr(op, "match.length"), nchar(cmd))
        operator <- tolower(substr(cmd, attr(op, "capture.start"),
           attr(op, "capture.start") + attr(op, "capture.length") - 1))

        if (operator == "by") {
          lav_operator <- "=~"
        } else if (operator == "with" || operator == "pwith") {
          lav_operator <- "~~"
        } else if (operator == "on") {
          lav_operator <- "~"
        }

        # handle parameter combinations
        lhs_split <- strsplit(lhs, "\\s+")[[1]] # lav_mplus_trim(

        # handle pwith syntax
        if (operator == "pwith") {
          # TODO: Figure out if pwith can be paired with constraints?

          rhs_split <- strsplit(rhs, "\\s+")[[1]] # lav_mplus_trim(
          if (length(lhs_split) != length(rhs_split)) {
            browser()
            lav_msg_stop(gettext(
              "PWITH command does not have the same number of arguments on
              the left and right sides."))
          }

          cmd <- sapply(seq_along(lhs_split), function(i) {
            paste(lhs_split[i], lav_operator, rhs_split[i])
          })
        } else {
          # insert plus signs on the rhs as long as it isn't preceded or
          #    followed by a plus already
          rhs <- gsub("(?<!\\+)\\s+(?!\\+)", " + ", rhs, perl = TRUE)

          if (length(lhs_split) > 1L) {
            # expand using possible combinations
            cmd <- sapply(lhs_split, function(larg) {
              pair <- paste(larg, lav_operator, rhs)
              pair
            })
          } else {
            cmd <- paste(lhs, lav_operator, rhs)
          }
        }
          } else if ((means_scales <-
        regexpr("^\\s*([\\[\\{])([^\\]\\}]+)[\\]\\}]\\s*$", cmd,
         ignore.case = TRUE, perl = TRUE))[1L] > 0) {
        # intercepts/means or scales
        # first capture is the operator: [ or {
        operator <- substr(cmd, attr(means_scales, "capture.start")[1L],
        attr(means_scales, "capture.start")[1L] + attr(means_scales,
           "capture.length")[1L] - 1)

        params <- substr(cmd, attr(means_scales, "capture.start")[2L],
         attr(means_scales, "capture.start")[2L] +
           attr(means_scales, "capture.length")[2L] - 1)

        # obtain parameters with no modifiers specified for LHS
        params_no_modifiers <- sub(
          "^\\s*[\\[\\{]([^\\]\\}]+)[\\]\\}]\\s*$", "\\1",
           attr(cmd, "noModifiers"), perl = TRUE)

        means_scales_split <- strsplit(params, "\\s+")[[1]] # lav_mplus_trim(
        means_scales_no_modifiers_split <-                                 # nolint
          strsplit(params_no_modifiers, "\\s+")[[1]] # lav_mplus_trim(

        if (operator == "[") {
          # Tricky syntax shift (and corresponding kludge). For means, need to
          #    put constraint on RHS as pre-multiplier of 1 (e.g., x1 ~ 5*1).
          # But lav_mplus_cmd_con returns constraints
          #     multiplied by parameters
          cmd <- sapply(means_scales_split, function(v) {
            # shift pre-multiplier
            if ((premult <- regexpr("([^\\*]+\\*[^\\*]+)\\*([^\\*]+)",
             v, perl = TRUE))[1L] > 0) { # double modifier: label and constraint
              modifier <- substr(v, attr(premult, "capture.start")[1L],
               attr(premult, "capture.start")[1L] +
                 attr(premult, "capture.length")[1L] - 1)
              param_name <- substr(v, attr(premult, "capture.start")[2L],
               attr(premult, "capture.start")[2L] +
                 attr(premult, "capture.length")[2L] - 1)
              paste0(param_name, " ~ ", modifier, "*1")
            } else if ((premult <-
              regexpr("([^\\*]+)\\*([^\\*]+)", v, perl = TRUE))[1L] > 0) {
              modifier <- substr(v, attr(premult, "capture.start")[1L],
                attr(premult, "capture.start")[1L] +
                  attr(premult, "capture.length")[1L] - 1)
              param_name <- substr(v, attr(premult, "capture.start")[2L],
                attr(premult, "capture.start")[2L] +
                   attr(premult, "capture.length")[2L] - 1)
              paste0(param_name, " ~ ", modifier, "*1")
            } else {
              paste(v, "~ 1")
            }
          })
        } else if (operator == "{") {
          # only include constraints on RHS
          cmd <- sapply(seq_along(means_scales_split),
                 function(v) {
                   paste(means_scales_no_modifiers_split[v],
                   "~*~", means_scales_split[v])
                 })
        } else {
          lav_msg_stop(gettext("What's the operator?!"))
        }
                 } else if (grepl("|", cmd, fixed = TRUE)) {
        # expand growth modeling language
        cmd <- lav_mplus_cmd_growth(cmd)
      } else { # no operator, no means, must be variance.
        # cat("assuming vars: ", cmd, "\n")

        vars_lhs <- strsplit(attr(cmd, "noModifiers"), "\\s+")[[1]]
        # lav_mplus_trim(
        vars_rhs <- strsplit(cmd, "\\s+")[[1]] # lav_mplus_trim(

        cmd <- sapply(seq_along(vars_lhs),
             function(v) paste(vars_lhs[v], "~~", vars_rhs[v]))
      }

      # handle threshold substitution: $ -> |
      cmd <- gsub("$", "|", cmd, fixed = TRUE)

      # if we have both starting/fixed values and constraints, these must be
      #    handled by separate commands.
      # starting and fixed values are already handled in the pipeline by this
      #    point, so should be evident in the command
      # bfi BY lab1*start(1)*bfi_1 ==> bfi BY lab1*bfi_1 + start(1)*bfi_1
      double_asterisks <- grepl(
        "\\s*[\\w\\(\\)\\.]+\\*[\\w\\(\\)\\.]+\\*[\\w\\(\\)\\.]+",
         cmd, perl = TRUE)

      if (isTRUE(double_asterisks[1])) {
        ss <- strsplit(cmd, "*", fixed = TRUE)[[1]]
        if (length(ss) != 3) {
          lav_msg_warn(gettext("problem interpreting double asterisk syntax:"),
                       cmd) # sanity check on my logic
        } else {
          cmd <- paste0(ss[1], "*", ss[3], " + ", ss[2], "*", ss[3])
        }
      }

      lavaan_out <- c(lavaan_out, cmd)
    }
  }

  # new threshold syntax shifts things to the form:
  # VAR | t1 + t2 + t3 (left to write ordering)
  # Parameter labels, fixed values, and starting values are tacked on in
  # the usual way, like
  # VAR | 5*t1 + start(1.5)*t2 + par_label*t3 (left to write ordering)

  thresh_lines <- grep("^\\s*[A-z]+\\w*\\|\\d+", lavaan_out, perl = TRUE)
  if (length(thresh_lines) > 0L) {
    thresh_vars <- unname(sub("^\\s*([A-z]+\\w*).*", "\\1",
     lavaan_out[thresh_lines], perl = TRUE))
    thresh_split <- split(thresh_lines, thresh_vars)
    drop_elements <- c()
    for (i in seq_along(thresh_split)) {
      this_set <- lavaan_out[thresh_split[[i]]]
      tnum <- as.integer(sub("^\\s*[A-z]+\\w*\\|(\\d+)\\s*.*", "\\1", this_set))
      this_set <- this_set[order(tnum)]
      # ensure that threshold numbering matches ascending order
      this_set <- sub("[^~]+\\s*~\\s*", "",
      this_set, perl = TRUE) # drop variable and ~

      # convert to new t1, t2 syntax by combining modifiers
      #    with threshold numbers
      this_set <- sapply(seq_along(this_set), function(j) {
        # gsub("[^~]+\\s*~\\s*([\\w\\.\\-]+\\*)*1",
        #   paste0("\\1t", j), this_set[j], perl=TRUE)
        gsub("([\\w\\.\\-]+\\*)*1", paste0("\\1t", j), this_set[j], perl = TRUE)
      })

      new_str <- paste(names(thresh_split)[i], "|",
       paste(this_set, collapse = " + "))
      # replace in model string on the first line having relevant syntax
      lavaan_out[thresh_split[[i]][1]] <- new_str
      drop_elements <- c(drop_elements, thresh_split[[i]][-1])
    }
    lavaan_out <- lavaan_out[-drop_elements]
  }


  # tack on constraint syntax, if included
  lavaan_out <- c(lavaan_out, con_syntax)

  # for now, include a final lav_mplus_trim call since some arguments have
  #   leading/trailing space stripped.
  wrap <- paste(lav_mplus_cmd_wrap(lavaan_out, width = 90, exdent = 5),
   collapse = "\n") # lav_mplus_trim(
  wrap
}

lav_mplus_lavaan <- function(inpfile, run = TRUE) {
  stopifnot(length(inpfile) == 1L)
  stopifnot(grepl("\\.inp$", inpfile, ignore.case = TRUE))
  if (!file.exists(inpfile)) {
    lav_msg_stop(gettext("Could not find file:"), inpfile)
  }

  # for future consideration. For now, require a .inp file
  #  if (length(inpfile) == 1L && grepl("\\.inp$", inpfile)) {
  #    if (!file.exists(inpfile)) { stop("Could not find file: ", inpfile) }
  #    inpfile.text <- scan(inpfile, what="character", sep="\n",
  #         strip.white=FALSE, blank.lines.skip=FALSE, quiet=TRUE)
  #  } else {
  #    #assume that inpfile itself is syntax (e.g., in a character vector)
  #    inpfile.text <- inpfile
  #  }

  inpfile_text <- scan(inpfile, what = "character", sep = "\n",
           strip.white = FALSE, blank.lines.skip = FALSE, quiet = TRUE)
  sections <- lav_mplus_text_sections(inpfile_text, inpfile)

  mplus_inp <- list()

  mplus_inp$title <- lav_mplus_trim(paste(sections$title, collapse = " "))
  mplus_inp$data <- lav_mplus_text_fields(sections$data, required = "file")
  mplus_inp$variable <- lav_mplus_text_fields(sections$variable,
       required = "names")
  mplus_inp$analysis <- lav_mplus_text_fields(sections$analysis)

  meanstructure <- "default" # lavaan default
  if (!is.null(mplus_inp$analysis$model)) {
    if (tolower(mplus_inp$analysis$model) == "nomeanstructure") {
      meanstructure <- FALSE
    } # explicitly disable mean structure
  }

  information <- "default" # lavaan default
  if (!is.null(mplus_inp$analysis$information)) {
    information <- tolower(mplus_inp$analysis$information)
  }

  estimator <- "default"
  if (!is.null(est <- mplus_inp$analysis$estimator)) {
    # no memory of what this is up to....
    if (toupper(est) == "MUML") lav_msg_warn(gettext(
      "Mplus does not support MUML estimator. Using default instead."))
    estimator <- est

    # march 2013: handle case where categorical data are specified,
    #    but ML-based estimator requested.
    # use WLSMV instead
    if (!is.null(mplus_inp$variable$categorical) &&
        toupper(substr(mplus_inp$analysis$estimator, 1, 2)) == "ML") {
      lav_msg_warn(gettext(
        "Lavaan does not yet support ML-based estimation for categorical data.
        Reverting to WLSMV"))
      estimator <- "WLSMV"
    }
  }

  # expand hyphens in variable names and split into vector
  #   that will be the names for read.table
  mplus_inp$variable$names <-
    strsplit(lav_mplus_expand_cmd(mplus_inp$variable$names), "\\s+",
             perl = TRUE)[[1]]

  # expand hyphens in categorical declaration
  if (!is.null(mplus_inp$variable$categorical))
    mplus_inp$variable$categorical <- strsplit(
      lav_mplus_expand_cmd(mplus_inp$variable$categorical), "\\s+",
      perl = TRUE)[[1]]

  # convert mplus syntax to lavaan syntax
  mplus_inp$model <- lav_mplus_syntax_model(sections$model)

  # handle model constraint
  if ("model.constraint" %in% names(sections)) {
    mplus_inp$model.constraint <-
      lav_mplus_syntax_con(sections$model.constraint)
    mplus_inp$model <-
      paste(mplus_inp$model, mplus_inp$model.constraint, sep = "\n")
  }

  # read mplus data (and handle missing spec)
  mplus_inp$data <- lav_mplus_path_data(mplus_inp, inpfile)

  # handle bootstrapping specification
  se <- "default"
  bootstrap <- 1000L
  test <- "default"
  if (!is.null(mplus_inp$analysis$bootstrap)) {
    boot_type <- "standard"
    # check whether standard versus residual bootstrap is specified
    if ((boot_match <- regexpr("\\((\\w+)\\)", mplus_inp$analysis$bootstrap,
     perl = TRUE)) > 0L) {
      boot_type <- tolower(substr(mplus_inp$analysis$bootstrap,
        attr(boot_match, "capture.start"), attr(boot_match, "capture.start") +
          attr(boot_match, "capture.length") - 1L))
    }

    if (boot_type == "residual") test <- "Bollen.Stine"

    se <- "bootstrap"

    if ((nboot_match <- regexpr("^\\s*(\\d+)",
    mplus_inp$analysis$bootstrap, perl = TRUE)) > 0L) {
      bootstrap <- as.numeric(substr(mplus_inp$analysis$bootstrap,
        attr(nboot_match, "capture.start"), attr(nboot_match, "capture.start") +
          attr(nboot_match, "capture.length") - 1L))
    }
  }

  if (run) {
    fit <- sem(mplus_inp$model, data = mplus_inp$data, meanstructure =
      meanstructure, mimic = "Mplus", estimator = estimator, test = test,
      se = se, bootstrap = bootstrap, information = information)
    fit@external <- list(mplus.inp = mplus_inp)
  } else {
    fit <- mplus_inp # just return the syntax outside of a lavaan object
  }

  fit
}


lav_mplus_text_fields <- function(section_text, required) {
  if (is.null(section_text)) {
    return(NULL)
  }

  # The parser breaks down when there is a line with a trailing comment because
  #    then splitting on semicolon will combine it with the following line
  # Thus, trim off trailing comments before initial split
  section_text <- gsub("\\s*!.*$", "", section_text, perl = TRUE)
  section_split <- strsplit(paste(section_text, collapse = " "), ";",
       fixed = TRUE)[[1]] # split on semicolons
  section_divide <- list()

  for (cmd in section_split) {
    if (grepl("^\\s*!.*", cmd, perl = TRUE)) next # skip comment lines
    if (grepl("^\\s+$", cmd, perl = TRUE)) next # skip blank lines

    # mplus is apparently tolerant of specifications that don't include IS/ARE/=
    # example: usevariables x1-x10;
    # thus, split on spaces and assume that first element is lhs,
    #    drop second element if IS/ARE/=, and assume remainder is rhs

    # but if user uses equals sign, then spaces will not always
    #           be present (e.g., usevariables=x1-x10)
    if ((leading_equals <- regexpr(
      "^\\s*[A-Za-z]+[A-Za-z_-]*\\s*(=)", cmd[1L], perl = TRUE))[1L] > 0) {
      cmd_name <- lav_mplus_trim(substr(cmd[1L], 1,
         attr(leading_equals, "capture.start") - 1))
      cmd_args <- lav_mplus_trim(substr(cmd[1L],
        attr(leading_equals, "capture.start") + 1, nchar(cmd[1L])))
    } else {
      cmd_spacesplit <-
        strsplit(lav_mplus_trim(cmd[1L]), "\\s+", perl = TRUE)[[1L]]

      if (length(cmd_spacesplit) < 2L) {
        # for future: make room for this function to parse things
        #     like just TECH13 (no rhs)
      } else {
        cmd_name <- lav_mplus_trim(cmd_spacesplit[1L])
        if (length(cmd_spacesplit) > 2L && tolower(cmd_spacesplit[2L]) %in%
                    c("is", "are")) {
          cmd_args <- paste(cmd_spacesplit[3L:length(cmd_spacesplit)],
               collapse = " ") # remainder, removing is/are
        } else {
          cmd_args <- paste(cmd_spacesplit[2L:length(cmd_spacesplit)],
               collapse = " ") # is/are not used, so just join rhs
        }
      }
    }

    section_divide[[make.names(tolower(cmd_name))]] <- cmd_args
  }

  if (!missing(required)) {
    stopifnot(all(required %in% names(section_divide)))
  }
  section_divide
}

# helper function
lav_mplus_path_splitted <- function(abspath) {
  # function to split path into path and filename
  # code adapted from R.utils filePath command
  if (!is.character(abspath)) lav_msg_stop(gettext(
    "Path not a character string"))
  if (nchar(abspath) < 1 || is.na(abspath)) lav_msg_stop(gettext(
    "Path is missing or of zero length"))

  components <- strsplit(abspath, split = "[\\/]")[[1]]
  lcom <- length(components)

  stopifnot(lcom > 0)

  # the file is the last element in the list.
  # In the case of length == 1, this will extract the only element.
  rel_filename <- components[lcom]
  absolute <- FALSE

  if (lcom == 1) {
    dirpart <- NA_character_
  } else if (lcom > 1) {
    # drop the file from the list (the last element)
    components <- components[-lcom]
    dirpart <- do.call("file.path", as.list(components))

    # if path begins with C:, /, //, or \\, then treat as absolute
    if (grepl("^([A-Z]{1}:|/|//|\\\\)+.*$", dirpart, perl = TRUE))
        absolute <- TRUE
  }

  list(directory = dirpart, filename = rel_filename, absolute = absolute)
}

lav_mplus_path_data <- function(mplus_inp, inpfile) {
  # handle issue of lav_mplus_lavaan being called with an absolute path,
  #   whereas mplus has only a local data file
  inpfile_split <- lav_mplus_path_splitted(inpfile)
  datfile_split <- lav_mplus_path_splitted(mplus_inp$data$file)

  # if inp file target directory is non-empty, but mplus data is without
  # directory, then append inp file directory to mplus data. This ensures that
  # R need not be in the working directory to read the dat file. But if mplus
  # data has an absolute directory, don't append

  # if mplus data directory is present and absolute, or if no directory
  #    in input file, just use filename as is
  if (!is.na(datfile_split$directory) && datfile_split$absolute) {
    dat_file <- mplus_inp$data$file
    # else: just use mplus data filename if it has absolute path
  } else if (is.na(inpfile_split$directory)) {
    dat_file <- mplus_inp$data$file
  } else {
    # just use mplus data filename if inp file is missing path (working dir)
    dat_file <- file.path(inpfile_split$directory, mplus_inp$data$file)
  } # dat file path is relative or absent, and inp file directory is present

  if (!file.exists(dat_file)) {
    lav_msg_warn(gettext("Cannot find data file:"), dat_file)
    return(NULL)
  }

  # handle missing is/are:
  miss_list <- NULL
  if (!is.null(miss_spec <- mplus_inp$variable$missing)) {
    expand_miss_vec <- function(miss_str) {
      # sub-function to obtain a vector of all missing values
      #   within a set of parentheses
      miss_split <- strsplit(miss_str, "\\s+")[[1L]]
      miss_vals <- c()
      for (f in miss_split) {
        if ((hyphen_pos <- regexpr("\\d+(-)\\d+", f, perl = TRUE))[1L] > -1L) {
          # expand hyphen
          pre_hyphen <- substr(f, 1, attr(hyphen_pos, "capture.start") - 1)
          post_hyphen <- substr(f,
              attr(hyphen_pos, "capture.start") + 1, nchar(f))
          miss_vals <- c(miss_vals, as.character(seq(pre_hyphen, post_hyphen)))
        } else {
          # append to vector
          miss_vals <- c(miss_vals, f)
        }
      }
      as.numeric(miss_vals)
    }

    if (miss_spec == "." || miss_spec == "*") { # case 1: MISSING ARE|=|IS .;
      na_strings <- miss_spec
    } else if ((all_match <- regexpr("\\s*ALL\\s*\\(([^\\)]+)\\)", miss_spec,
     perl = TRUE))[1L] > -1L) { # case 2: use of ALL with parens
      miss_str <- lav_mplus_trim(substr(miss_spec,
        attr(all_match, "capture.start"),
        attr(all_match, "capture.start") +
          attr(all_match, "capture.length") - 1L))
      na_strings <- expand_miss_vec(miss_str)
    } else { # case 3: specific missing values per variable
      # process each element
      miss_blocks <- gregexpr("(?:(\\w+)\\s+\\(([^\\)]+)\\))+", miss_spec,
         perl = TRUE)[[1]]
      miss_list <- list()

      if (miss_blocks[1L] > -1L) {
        for (i in seq_along(miss_blocks)) {
          vname <- substr(miss_spec, attr(miss_blocks,
            "capture.start")[i, 1L], attr(miss_blocks, "capture.start")[i, 1L] +
               attr(miss_blocks, "capture.length")[i, 1L] - 1L)
          vmiss <- substr(miss_spec, attr(miss_blocks, "capture.start")[i, 2L],
           attr(miss_blocks, "capture.start")[i, 2L] + attr(miss_blocks,
              "capture.length")[i, 2L] - 1L)

          vname_hyphen <- regexpr("(\\w+)-(\\w+)", vname, perl = TRUE)[1L]
          if (vname_hyphen > -1L) {
            # lookup against variable names
            vstart <- which(mplus_inp$variable$names == substr(vname,
               attr(vname_hyphen, "capture.start")[1L], attr(vname_hyphen,
                                   "capture.start")[1L]
                 + attr(vname_hyphen, "capture.length")[1L] - 1L))
            vend <- which(mplus_inp$variable$names ==
              substr(vname, attr(vname_hyphen, "capture.start")[2L],
            attr(vname_hyphen, "capture.start")[2L] +
              attr(vname_hyphen, "capture.length")[2L] - 1L))
            if (length(vstart) == 0L || length(vend) == 0L) {
              lav_msg_stop(gettext("Unable to lookup missing variable list: "),
                           vname)
            }
            # I suppose start or finish could be mixed up
            if (vstart > vend) {
              vstart_orig <- vstart
              vstart <- vend
              vend <- vstart_orig
            }
            vname <- mplus_inp$variable$names[vstart:vend]
          }

          miss_vals <- expand_miss_vec(vmiss)

          for (j in seq_along(vname)) {
            miss_list[[vname[j]]] <- miss_vals
          }
        }
      } else {
        lav_msg_stop(gettext("I don't understand this missing specification:"),
                     miss_spec)
      }
    }
  } else {
    na_strings <- "NA"
  }

  if (!is.null(miss_list)) {
    dat <- read.table(dat_file, header = FALSE,
      col.names = mplus_inp$variable$names, colClasses = "numeric")
    # loop over variables in missList and set missing values to NA
    dat[, names(miss_list)] <- lapply(names(miss_list), function(vmiss) {
      dat[which(dat[, vmiss] %in% miss_list[[vmiss]]), vmiss] <- NA
      dat[, vmiss]
    })

    names(dat) <- mplus_inp$variable$names # loses these from the lapply
  } else {
    dat <- read.table(dat_file, header = FALSE,
      col.names = mplus_inp$variable$names, na.strings = na_strings,
      colClasses = "numeric")
  }


  # TODO: support covariance/mean+cov inputs

  # store categorical variables as ordered factors
  if (!is.null(mplus_inp$variable$categorical)) {
    dat[, c(mplus_inp$variable$categorical)] <-
      lapply(dat[, c(mplus_inp$variable$categorical), drop = FALSE], ordered)
  }

  dat
}


lav_mplus_text_sections <- function(inpfile_text, filename) {
  input_headers <- grep(
    "^\\s*(title:|data.*:|variable:|define:|analysis:|model.*:|output:|savedata:|plot:|montecarlo:)", # nolint
     inpfile_text, ignore.case = TRUE, perl = TRUE)

  stopifnot(length(input_headers) > 0L)

  mplus_sections <- list()

  for (h in seq_along(input_headers)) {
    section_end <- ifelse(h < length(input_headers), input_headers[h + 1] - 1,
        length(inpfile_text))
    section <- inpfile_text[input_headers[h]:section_end]
    section_name <- lav_mplus_trim(sub("^([^:]+):.*$", "\\1", section[1L],
     perl = TRUE)) # obtain text before the colon

    # dump section name from input syntax
    section[1L] <- sub("^[^:]+:(.*)$", "\\1", section[1L], perl = TRUE)

    mplus_sections[[make.names(tolower(section_name))]] <- section
  }

  mplus_sections
}

# This code was contributed by Myrsini Katsikatsou (LSE) -- September 2016
#
# lav_pml_bivprob_unicondprob()
# lav_pml_object_uni_pairwise_prob()
# lav_pml_th_rho_generalised()
# pairwiseExpProbVec_GivenObs_UncMod()

lav_pml_bivprob_unicondprob <- function(biv_prob, nvar,
                                                 idx_pairs,
                                                 idx_y1,
                                                 idx_gy2,
                                                 idx_cat_y1_split,
                                                 idx_cat_y2_split) {
  biv_prob_split <- split(biv_prob, idx_pairs)
  lngth <- 2 * length(biv_prob)
  idx_vec_el <- 1:lngth
  prob_y1gy2 <- rep(NA, lngth)
  no_pairs <- nvar * (nvar - 1) / 2
  idx2_pairs <- combn(nvar, 2)

  for (k in 1:no_pairs) {
    y2sums <- tapply(biv_prob_split[[k]], idx_cat_y2_split[[k]], sum)
    y2sums_mult <- y2sums[idx_cat_y2_split[[k]]]
    y1gy2 <- biv_prob_split[[k]] / y2sums_mult
    tmp_idx_vec_el <- idx_vec_el[(idx_y1 == idx2_pairs[1, k]) &
      (idx_gy2 == idx2_pairs[2, k])]
    prob_y1gy2[tmp_idx_vec_el] <- y1gy2
  }

  for (k in 1:no_pairs) {
    y1sums <- tapply(biv_prob_split[[k]], idx_cat_y1_split[[k]], sum)
    y1sums_mult <- y1sums[idx_cat_y1_split[[k]]]
    y2gy1 <- biv_prob_split[[k]] / y1sums_mult
    reordered_y2gy1 <- y2gy1[order(idx_cat_y1_split[[k]])]
    tmp_idx_vec_el <- idx_vec_el[(idx_y1 == idx2_pairs[2, k]) &
      (idx_gy2 == idx2_pairs[1, k])]
    prob_y1gy2[tmp_idx_vec_el] <- reordered_y2gy1
  }

  prob_y1gy2
}

# The input of the function is a lavobject, which, in turn, is the output of the
# sem function having specified estimator="PML", missing="available.cases"

# The output of the function is a list of two lists: the pairwiseProbGivObs
# list and the univariateProbGivObs list. Each of the two lists consists of
# G matrices where G is the number of groups in a multigroup analysis.
# If G=1 each of the lists contains only one matrix that can be called as
# pairwiseProbGivObs[[1]], and univariateProbGivObs[[1]].

# Each of the matrices in the pairwiseProbGivObs list is of dimension:
# nrow=sample size, ncol=sum of the number of response categories for all
# pairs of variables (i.e. the length of the vector pxixj.ab where
# i<j=1,...,p, a=1,...,Ci, b=1,...,Cj;
# a which is the index for the response category for yi variable runs the
# fastest, then b which is the index for the response category for yj variable,
# then j, and last i.)
# The cells in a matrix of the pairwiseProbGivObs list have the value 0 except
# for those cells that correspond to the pairs of variables where both variables
# are missing. Those cells have the value of the bivariate conditional
# probability for the given pair for all their response categories.
# The bivariate probabilities are computed as follows:
# the information in the observed variables is summarised in a factor score
# for each individual and the bivariate probability given the estimated factor
# scores is computed.


# Each of the matrices in the univariateProbGivObs list is of dimension:
# nrow=sample size, ncol=sum of the number of response categories for all
# variables. The columns are indexed with i and a, where i=1,...,p, and
# a=1,...,Ci, the response categories for yi variable; a runs faster than i.
# The cells in a matrix of the univariateProbGivObs list have the value 0
# except for those cells that correspond to variables with missing values.
# Those cells have the value of the univariate conditional probability for the
# given variable for all its response categories. The univariate conditional
# probabilities are computed as follows:
# given that the bivariate conditional probabilities have been computed we sum
# over the response categories of each variable at a time (i.e. we compute the
# marginals).

# Version 3 - first compute univariate and then bivariate probabilities

lav_pml_object_uni_pairwise_prob <- function(lavobject) {  # nolint
  # compute yhat where yaht=nu + Lamda*eta + K*x where the parameter
  # estimates are used and the factor scores for eta
  # Below yhat is a list if lavobject@Data@ngroups >1, it is a list of
  # G matrices where G the number of groups and the matrices are of
  # dimension nrow=sample size and ncol=number of items.
  # If lavobject@Data@ngroups=1 then yhat is a matrix.
  yhat <- lavPredict(object = lavobject, type = "yhat")

  # compute bivariate probabilities
  ngroups <- lavobject@Data@ngroups
  univariate_prob <- vector("list", length = ngroups)
  pairwise_prob <- vector("list", length = ngroups)
  # save the indices of the Theta matrices for the groups stored in GLIST
  idx_theta_mat <- which(names(lavobject@Model@GLIST) == "theta")

  for (g in seq_len(ngroups)) { # g<-1

    if (ngroups > 1L) {
      yhat_group <- yhat[[g]]
    } else {
      yhat_group <- yhat
    }

    nsize <- lavobject@Data@nobs[[g]]
    nvar <- lavobject@Model@nvar[[g]]
    data_1 <- lavobject@Data@X[[g]]
    th <- lavobject@Fit@TH[[g]]
    th_idx <- lavobject@Model@th.idx[[g]]
    theta <- lavobject@Model@GLIST[idx_theta_mat[g]]$theta
    error_stddev <- diag(theta)^0.5

    # for the computation of the univariate probabilities
    nlev <- lavobject@Data@ov$nlev
    idx_uniy <- rep(1:nvar, times = nlev)

    # indices vectors for the computation of bivariate probabilities
    idx_pairs_yiyj <- combn(1:nvar, 2)
    no_biv_resp_cat_yiyj <- sapply(seq_len(ncol(idx_pairs_yiyj)),
     function(x) {
      prod(nlev[idx_pairs_yiyj[, x]])
    })
    idx_y1 <- unlist(
      mapply(rep, idx_pairs_yiyj[1, ], each = no_biv_resp_cat_yiyj)
    )
    idx_y2 <- unlist(
      mapply(rep, idx_pairs_yiyj[2, ], each = no_biv_resp_cat_yiyj)
    )


    univariate_prob[[g]] <- matrix(0, nrow = nsize, ncol = sum(nlev))
    pairwise_prob[[g]] <- matrix(0,
      nrow = nsize,
      ncol = length(lavobject@Cache[[g]]$bifreq)
    )

    idx_miss_var_casewise <- apply(data_1, 1, function(x) {
      which(is.na(x))
    })

    for (i in 1:nsize) {
      idx_miss_var <- idx_miss_var_casewise[[i]]
      no_miss_var <- length(idx_miss_var)

      if (no_miss_var > 0L) {
        # compute the univariate probabilities
        th_list <- split(th, th_idx)
        tmp_th <- th_list[idx_miss_var]
        tmp_lower_th <- unlist(lapply(tmp_th, function(x) {
          c(-Inf, x)
        }))
        tmp_upper_th <- unlist(lapply(tmp_th, function(x) {
          c(x, Inf)
        }))

        idx_items <- rep(c(1:no_miss_var), times = nlev[idx_miss_var])
        tmp_mean <- yhat_group[i, idx_miss_var]
        tmp_mean_extended <- tmp_mean[idx_items]
        tmp_stddev <- error_stddev[idx_miss_var]
        tmp_stddev_extended <- tmp_stddev[idx_items]

        tmp_uni_prob <- pnorm((tmp_upper_th - tmp_mean_extended) /
          tmp_stddev_extended) -
          pnorm((tmp_lower_th - tmp_mean_extended) /
            tmp_stddev_extended)
        idx_columns_uni <- which(idx_uniy %in% idx_miss_var)
        univariate_prob[[g]][i, idx_columns_uni] <- tmp_uni_prob

        # compute the bivariate probabilities
        if (no_miss_var > 1L) {
          idx_pairs_miss <- combn(idx_miss_var, 2)
          no_pairs <- ncol(idx_pairs_miss)
          idx_pairs_v2 <- combn(no_miss_var, 2)
          idx_columns <- unlist(lapply(1:no_pairs, function(x) {
            which((idx_y1 == idx_pairs_miss[1, x]) &
              (idx_y2 == idx_pairs_miss[2, x]))
          }))

          if (all(theta[t(idx_pairs_miss)] == 0)) {
            # items independence given eta
            tmp_uni_prob_list <- split(tmp_uni_prob, idx_items)
            pairwise_prob[[g]][i, idx_columns] <-
              unlist(lapply(1:no_pairs, function(x) {
                c(outer(
                  tmp_uni_prob_list[[idx_pairs_v2[1, x]]],
                  tmp_uni_prob_list[[idx_pairs_v2[2, x]]]
                ))
              }))
          } else { # when correlation between measurement errors

            tmp_th_idx <- th_idx[th_idx %in% idx_miss_var]
            # recode so that it is always 1,1,..,1, 2,...,2, etc.
            tmp_th_idx_recoded <- rep(c(1:no_miss_var),
                                  times = table(tmp_th_idx))
            tmp_th <- th[th_idx %in% idx_miss_var]

            tmp_ind_vec <- lav_pml_longvec_ind(
              no_x = no_miss_var,
              all_thres = tmp_th,
              index_var_of_thres = tmp_th_idx_recoded
            )

            tmp_th_rho_vec <- lav_pml_th_rho_generalised(
              no_x = no_miss_var,
              th = tmp_th,
              th_idx = tmp_th_idx_recoded,
              cov_xixj = theta[t(idx_pairs_miss)],
              mean_x = yhat_group[i, idx_miss_var],
              stddev_x = error_stddev[idx_miss_var]
            )

            tmp_biv_prob <- lav_pml_expprob_vec(
              ind_vec = tmp_ind_vec,
              th_rho_vec = tmp_th_rho_vec
            )

            pairwise_prob[[g]][i, idx_columns] <- tmp_biv_prob
          } # end of else of if( all( Theta[t(idx.pairsMiss)]==0 ) )
          # which checks item local independence
        } # end of if( noMissVar>1L )

        # cat(i,  "\n")
      } # end of if(noMissVar>0L)
    } # end of for(i in 1:nsize)
  } # end of for(g in seq_len(lavobject@Data@ngroups))

  list(
    univariateProbGivObs = univariate_prob,
    pairwiseProbGivObs = pairwise_prob
  )
} # end of the function lav_pml_object_uni_pairwise_prob

##################################################################



# lav_pml_th_rho_generalised function is defined as follows
lav_pml_th_rho_generalised <- function(no_x, th, th_idx,
                                      cov_xixj, mean_x, stddev_x) {
  all_std_thres <- (th - mean_x[th_idx]) / stddev_x[th_idx]
  id_pairs <- utils::combn(no_x, 2)
  cor_xixj <- cov_xixj / (stddev_x[id_pairs[1, ]] * stddev_x[id_pairs[2, ]])

  lav_pml_longvec_th_rho(
    no_x = no_x,
    all_thres = all_std_thres,
    index_var_of_thres = th_idx,
    rho_xixj = cor_xixj
  )
}

# lav_pml_th_rho_generalised is a generalisation  of the function
# lav_pml_longvec_th_rho . The latter assumes that all y* follow standard
# normal so the thresholds are automatically the standardised ones.
# lav_pml_th_rho_generalised does not assume that, each of y*'s can follow
# a normal distribution with mean mu and standard deviation sigma.
# lav_pml_th_rho_generalised has the following input arguments:
# no.x (same as in lav_pml_longvec_th_rho),
# TH (similar to the TH in lav_pml_longvec_th_rho but here they are the
#  unstandardised thresholds, i.e. of the normal distribution with
#  mean mu and standard deviation sigma)
# th.idx (same as index.var.of.thres in lav_pml_longvec_th_rho)
# cov.xixj which are the polychoric covariances of the pairs of underlying
# variables provided in a similar fashion as rho.xixj in lav_pml_longvec_th_rho)
# mean.x  is a vector including the means of y*'s provided
#   in the order mean.x1, mean.x2, ...., mean.xp
# stddev.x  is a vector including the standard deviations of y*'s provided
#   in the order stddev.x1, stddev.x2, ...., stddev.xp

# The output of the new function is similar to that of lav_pml_longvec_th_rho
# #############################################



# lavobject is the output of lavaan function where either the unconstrained
# or a hypothesized model has been fitted
lav_pml_object_uni_pairwise_unconstr <- function(lavobject) {  # nolint
  ngroups <- lavobject@Data@ngroups
  th <- lavobject@implied$th # these are the standardized thresholds
  # mean and variance of y* have been taken into account
  th_idx <- lavobject@SampleStats@th.idx
  sigma_hat <- lavobject@implied$cov

  univariate_prob <- vector("list", length = ngroups)
  pairwise_prob <- vector("list", length = ngroups)

  for (g in 1:ngroups) {
    sigma_hat_g <- sigma_hat[[g]]
    # is Sigma.hat always a correlation matrix?
    cor_hat_g <- cov2cor(sigma_hat_g)
    cors <- cor_hat_g[lower.tri(cor_hat_g)]
    if (any(abs(cors) > 1)) {
      lav_msg_warn(gettext(
      "some model-implied correlations are larger than 1.0"))
    }
    nvar <- nrow(sigma_hat_g)
    mean_1 <- rep(0, nvar)
    th_g <- th[[g]]
    th_idx_g <- th_idx[[g]]

    nlev <- lavobject@Data@ov$nlev

    # create index vector to keep track which variable each column of
    # univariateProb matrix refers to
    idx_uniy <- rep(1:nvar, times = nlev)

    # create index vector to keep track which variables each column of
    # pairwiseProb matrix refers to
    idx_pairs_yiyj <- combn(1:nvar, 2)
    no_biv_resp_cat_yiyj <- sapply(seq_len(ncol(idx_pairs_yiyj)),
     function(x) {
      prod(nlev[idx_pairs_yiyj[, x]])
    })
    idx_y1 <- unlist(
      mapply(rep, idx_pairs_yiyj[1, ], each = no_biv_resp_cat_yiyj)
    )
    idx_y2 <- unlist(
      mapply(rep, idx_pairs_yiyj[2, ], each = no_biv_resp_cat_yiyj)
    )

    data_1 <- lavobject@Data@X[[g]]
    nsize <- nrow(data_1)

    # create the lists of matrices
    univariate_prob[[g]] <- matrix(0, nrow = nsize, ncol = sum(nlev))
    pairwise_prob[[g]] <- matrix(0,
      nrow = nsize,
      ncol = length(lavobject@Cache[[g]]$bifreq)
    )

    idx_miss_var_casewise <- apply(data_1, 1, function(x) {
      which(is.na(x))
    })

    for (i in 1:nsize) {
      idx_miss_var <- idx_miss_var_casewise[[i]]
      no_miss_var <- length(idx_miss_var)

      if (no_miss_var > 0L) {
        # compute the denominator of the conditional probability
        th_var <- lapply(1:nvar, function(x) c(-Inf, th_g[th_idx_g == x], +Inf))
        lower <- sapply(1:nvar, function(x) th_var[[x]][data_1[i, x]])
        upper <- sapply(1:nvar, function(x) th_var[[x]][data_1[i, x] + 1L])
        lower_denom <- lower[-idx_miss_var]
        upper_denom <- upper[-idx_miss_var]
        mean_i <- mean_1[-idx_miss_var]
        corhat_i <- cor_hat_g[-idx_miss_var, -idx_miss_var, drop = FALSE]
        denom <- sadmvn(lower_denom, upper_denom, mean = mean_i,
           varcov = corhat_i)[1]
      } # end of if( noMissVar>0L )

      if (no_miss_var == 1L) { # only univariate probabilities for one item
        # compute the numerator
        th_miss_var <- c(-Inf, th_g[th_idx_g == idx_miss_var], +Inf)
        # for all response categories of the missing item
        no_cat <- nlev[idx_miss_var]
        numer <- sapply(1:no_cat, function(x) {
          lower[idx_miss_var] <- th_miss_var[x]
          upper[idx_miss_var] <- th_miss_var[x + 1L]
          sadmvn(lower, upper, mean = mean_1, varcov = cor_hat_g)[1]
        })
        idx_columns_uni <- which(idx_uniy %in% idx_miss_var)
        univariate_prob[[g]][i, idx_columns_uni] <- numer / denom
      } # end of if( noMissVar==1L )

      if (no_miss_var > 1L) {
        # compute the bivariate probabilities and based on them
        # calculate the univariate ones

        # form all possible pairs of items with missing values
        idx_pairs_miss <- combn(idx_miss_var, 2)
        no_pairs <- ncol(idx_pairs_miss)
        for (j in 1:no_pairs) {
          idx_missy1y2 <- idx_pairs_miss[, j]
          idx_missy1 <- idx_missy1y2[1]
          idx_missy2 <- idx_missy1y2[2]
          idx_miss_rest_items <- idx_miss_var[!(idx_miss_var %in% idx_missy1y2)]
          th_missy1 <- c(-Inf, th_g[th_idx_g == idx_missy1], +Inf)
          th_missy2 <- c(-Inf, th_g[th_idx_g == idx_missy2], +Inf)
          no_cat_missy1 <- nlev[idx_missy1]
          no_cat_missy2 <- nlev[idx_missy2]
          no_biv_resp_cat <- no_cat_missy1 * no_cat_missy2
          mat_biv_resp_cat <- matrix(1:no_biv_resp_cat,
            nrow = no_cat_missy1,
            ncol = no_cat_missy2
          )

          numer <- sapply(1:no_biv_resp_cat, function(x) {
            idx_y1_cat <- which(mat_biv_resp_cat == x, arr.ind = TRUE)[1]
            idx_y2_cat <- which(mat_biv_resp_cat == x, arr.ind = TRUE)[2]
            lower[idx_missy1y2] <-
              c(th_missy1[idx_y1_cat], th_missy2[idx_y2_cat])
            upper[idx_missy1y2] <-
              c(th_missy1[idx_y1_cat + 1L], th_missy2[idx_y2_cat + 1L])
            lower_tmp <- lower
            upper_tmp <- upper
            mean_tmp <- mean_1
            cor_hat_g_tmp <- cor_hat_g
            if (length(idx_miss_rest_items) > 0) {
              lower_tmp <- lower[-idx_miss_rest_items]
              upper_tmp <- upper[-idx_miss_rest_items]
              mean_tmp <- mean_1[-idx_miss_rest_items]
              cor_hat_g_tmp <-
                    cor_hat_g[-idx_miss_rest_items, -idx_miss_rest_items]
            }
            sadmvn(lower_tmp, upper_tmp,
              mean = mean_tmp, varcov = cor_hat_g_tmp
            )[1]
          })

          idx_columns <- which((idx_y1 == idx_missy1) &
            (idx_y2 == idx_missy2))
          tmp_biv <- numer / denom
          pairwise_prob[[g]][i, idx_columns] <- tmp_biv

          # compute the univariateProb based on the above bivariate
          # probabilities
          if (j == 1L) {
            univariate_prob[[g]][i, which(idx_uniy %in% idx_missy1)] <-
              apply(mat_biv_resp_cat, 1, function(x) {
                sum(tmp_biv[x])
              })

            univariate_prob[[g]][i, which(idx_uniy %in% idx_missy2)] <-
              apply(mat_biv_resp_cat, 2, function(x) {
                sum(tmp_biv[x])
              })
          }

          if (j > 1L && j < no_miss_var) {
            univariate_prob[[g]][i, which(idx_uniy %in% idx_missy2)] <-
              apply(mat_biv_resp_cat, 2, function(x) {
                sum(tmp_biv[x])
              })
          }
        } # end of for(j in 1:no.pairs ) #no.pairs is that of missing items
      } # end of if( noMissVar>1L )
    } # end of for(i in 1:nsize)
  } # end of for(g in 1:ngroups)

  list(
    univariateProbGivObs = univariate_prob,
    pairwiseProbGivObs = pairwise_prob
  )
} # end of function

# FMG (Foldnes, Moss, Gronneberg 2024) test statistics
# Improved goodness-of-fit tests using eigenvalue methods
#
# References:
# - Foldnes, N., Moss, J., & Gronneberg, S. (2024). Improved goodness of fit
#   procedures for structural equation models. Structural Equation Modeling.
# - Du, H., & Bentler, P. M. (2022). 40-Year Old Unbiased Distribution Free
#   Estimator Reliably Improves SEM Statistics for Nonnormal Data.
# - Wu, H., & Lin, J. (2016). A Scaled F Distribution as an Approximation to
#   the Distribution of Test Statistics in Covariance Structure Analysis.

# =====================================================
# Parser (ported from semTests split_input)
# =====================================================

#' Check if test string is an FMG test
#' @param test Character string specifying the test
#' @return Logical indicating if this is an FMG test
#' @keywords internal
lav_test_fmg_is_fmg <- function(test) {
  test <- tolower(test)
  patterns <- c(
    "^fmg($|_)", "^peba", "^pols", "^pall", "^all$",
    "^all_", "^sb_", "^ss_", "^sf_", "^std_"
  )
  any(vapply(patterns, function(p) grepl(p, test), logical(1)))
}

lav_test_fmg_is_preset <- function(test) {
  grepl("^fmg($|_)", tolower(test))
}

lav_test_fmg_resolve_preset <- function(test, nested = FALSE,
                                        ngroups = 1L) {
  if (!lav_test_fmg_is_preset(test)) {
    return(test)
  }

  test <- tolower(test)
  splitted <- strsplit(test, "_")[[1L]]
  if (splitted[1L] != "fmg" || length(splitted) > 3L) {
    lav_msg_stop(gettextf(
      "FMG preset does not support test= %1$s.",
      dQuote(test)
    ))
  }

  unbiased <- isTRUE(nested)
  chisq <- if (isTRUE(nested) && ngroups > 1L) "ml" else "rls"

  if (length(splitted) > 1L) {
    suffix <- splitted[-1L]
    if (!all(suffix %in% c("ug", "ml", "rls")) ||
        sum(suffix %in% c("ml", "rls")) > 1L ||
        sum(suffix == "ug") > 1L) {
      lav_msg_stop(gettextf(
        "FMG preset does not support test= %1$s.",
        dQuote(test)
      ))
    }
    if ("ug" %in% suffix) {
      unbiased <- TRUE
    }
    if ("ml" %in% suffix) {
      chisq <- "ml"
    } else if ("rls" %in% suffix) {
      chisq <- "rls"
    }
  }

  method <- if (isTRUE(nested)) "pall" else "peba4"
  paste0(
    method,
    if (isTRUE(unbiased)) "_ug" else "",
    "_", chisq
  )
}

#' Parse FMG test string into components
#'
#' Format: {method}{param}_{ug?}_{chisq?}
#' Examples: peba4_ug_rls, eba2, pols2_ml, sb_ug
#'
#' @param string Character string specifying the test
#' @return List with method, param, unbiased, chisq
#' @keywords internal
lav_test_fmg_parse <- function(string) {
  string <- tolower(string)
  splitted <- strsplit(string, "_")[[1]]

  # Defaults
  method <- NULL
  param <- 2L # default for j (eba/peba) or gamma (pols)
  unbiased <- FALSE
  unbiased_explicit <- FALSE
  chisq <- "default"
  chisq_explicit <- FALSE

  type <- splitted[1]

  # Parse unbiased and chisq from suffix parts
  if (length(splitted) == 3L) {
    unbiased <- (splitted[2] == "ug")
    unbiased_explicit <- TRUE
    chisq <- splitted[3]
    chisq_explicit <- TRUE
  } else if (length(splitted) == 2L) {
    if (splitted[2] %in% c("rls", "ml")) {
      chisq <- splitted[2]
      chisq_explicit <- TRUE
    } else if (splitted[2] == "ug") {
      unbiased <- TRUE
      unbiased_explicit <- TRUE
    }
  }

  # Parse method and parameter
  if (startsWith(type, "peba")) {
    method <- "peba"
    param_str <- substring(type, 5)
    if (nchar(param_str) > 0L) param <- as.integer(param_str)
  } else if (startsWith(type, "pols")) {
    method <- "pols"
    param_str <- substring(type, 5)
    if (nchar(param_str) > 0L) param <- as.numeric(param_str)
  } else if (type == "pall") {
    method <- "pall"
  } else if (type == "all") {
    method <- "all"
  } else if (type %in% c("sb", "ss", "sf", "std")) {
    method <- type
  } else {
    lav_msg_stop(gettextf(
      "FMG tests do not support test= %1$s.",
      dQuote(type)
    ))
  }

  list(
    method = method,
    param = param,
    unbiased = unbiased,
    chisq = chisq,
    unbiased_explicit = unbiased_explicit,
    chisq_explicit = chisq_explicit
  )
}

lav_test_fmg_resolve_unbiased <- function(parsed, lavoptions = NULL) {
  if (isTRUE(parsed$unbiased_explicit)) {
    return(parsed$unbiased)
  }
  if (!is.null(lavoptions$gamma.unbiased)) {
    return(isTRUE(lavoptions$gamma.unbiased))
  }
  FALSE
}

lav_test_fmg_resolve_chisq <- function(parsed, lavoptions = NULL) {
  if (isTRUE(parsed$chisq_explicit)) {
    return(parsed$chisq)
  }

  scaled_test <- "standard"
  if (!is.null(lavoptions$scaled.test)) {
    scaled_test <- lavoptions$scaled.test[1L]
  }

  if (scaled_test %in% c("default", "standard", "none")) {
    return("ml")
  } else if (scaled_test == "browne.residual.nt.model") {
    return("rls")
  }

  lav_msg_stop(gettextf(
    "FMG tests require scaled.test= %1$s or %2$s; found %3$s.",
    dQuote("standard"), dQuote("browne.residual.nt.model"),
    dQuote(scaled_test)
  ))
}

lav_test_fmg_check_mi <- function(lavoptions = NULL,
                                       context = "FMG tests") {
  if (is.null(lavoptions) || is.null(lavoptions$missing)) {
    return(invisible(TRUE))
  }

  missing <- lavoptions$missing[1L]
  if (isTRUE(missing == "listwise")) {
    return(invisible(TRUE))
  }

  lav_msg_stop(gettextf(
    "%1$s require missing= %2$s; found %3$s.",
    context, dQuote("listwise"), dQuote(missing)
  ))
}

lav_test_fmg_check_estimator <- function(lavoptions = NULL,
                                         context = "FMG tests") {
  if (is.null(lavoptions)) {
    return(invisible(TRUE))
  }

  estimator <- lavoptions$estimator
  estimator_orig <- lavoptions$estimator.orig
  if (isTRUE(estimator == "ML") &&
      isTRUE(estimator_orig %in% c("ML", "MLM"))) {
    return(invisible(TRUE))
  }

  reported <- if (!is.null(estimator_orig)) estimator_orig else estimator
  lav_msg_stop(gettextf(
    "%1$s require estimator= %2$s or %3$s; found %4$s.",
    context, dQuote("ML"), dQuote("MLM"), dQuote(reported)
  ))
}

lav_test_fmg_ugamma_eigenvalues <- function(ugamma, df,
                                            truncate_negative = TRUE) {
  eigs <- Re(eigen(ugamma, only.values = TRUE)$values)
  lambdas_raw <- sort(eigs, decreasing = TRUE)[seq_len(df)]
  lambdas <- lambdas_raw
  truncated <- rep(FALSE, length(lambdas))

  if (isTRUE(truncate_negative)) {
    truncated <- lambdas < 0
    lambdas[truncated] <- 0
  }

  list(
    values = lambdas,
    raw = lambdas_raw,
    truncated = truncated,
    n_truncated = sum(truncated),
    min_raw = min(lambdas_raw)
  )
}

lav_test_fmg_label <- function(parsed, unbiased = FALSE, chisq = "ml") {
  param <- format(parsed$param, trim = TRUE, scientific = FALSE)
  method_label <- switch(parsed$method,
    "peba" = paste0("pEBA-", param),
    "pols" = paste0("pOLS-", param),
    "pall" = "PALL",
    "all" = "ALL",
    "sb" = "SB",
    "ss" = "scaled-shifted",
    "sf" = "scaled-F",
    "std" = "standard chi-square",
    parsed$method
  )
  chisq_label <- if (chisq == "rls") "RLS" else toupper(chisq)
  gamma_label <- if (isTRUE(unbiased)) "unbiased" else "biased"

  paste0(
    method_label,
    " p-value test (base: ", chisq_label,
    ", gamma: ", gamma_label, ")"
  )
}

lav_test_fmg_chisq_equivalent <- function(pvalue, df) {
  if (is.na(pvalue) || is.na(df) || df <= 0L) {
    return(as.numeric(NA))
  }
  pvalue <- max(0, min(1, pvalue))
  stats::qchisq(pvalue, df = df, lower.tail = FALSE)
}

lav_test_fmg_stat_group_equivalent <- function(stat_group = NULL,
                                               source_stat = NULL,
                                               stat = NULL) {
  if (is.null(stat_group)) {
    return(NULL)
  }
  if (length(stat_group) == 1L) {
    return(stat)
  }
  if (is.na(source_stat) || source_stat == 0 || is.infinite(stat)) {
    return(rep(stat / length(stat_group), length(stat_group)))
  }

  stat_group * stat / source_stat
}

lav_test_fmg_browne_nt_model <- function(lavobject = NULL,
                                         lavdata = NULL,
                                         lavsamplestats = NULL,
                                         lavmodel = NULL,
                                         lavpartable = NULL,
                                         lavoptions = NULL,
                                         lavh1 = NULL,
                                         lavimplied = NULL) {
  out <- try(
    lav_test_browne(
      lavobject = lavobject,
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      lavmodel = lavmodel,
      lavpartable = lavpartable,
      lavoptions = lavoptions,
      lavh1 = lavh1,
      lavimplied = lavimplied,
      adf = FALSE,
      model_based = TRUE
    ),
    silent = TRUE
  )
  if (!inherits(out, "try-error")) {
    return(out)
  }

  if (!is.null(lavobject)) {
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
    lavmodel <- lavobject@Model
    lavpartable <- lavobject@ParTable
    lavoptions <- lavobject@Options
    lavh1 <- lavobject@h1
    lavimplied <- lavobject@implied
  }

  if (!is.logical(lavoptions$gamma.n.minus.one)) {
    n_minus_one <- !(lavoptions$estimator == "ML" &&
      lavoptions$likelihood == "normal")
  } else {
    n_minus_one <- lavoptions$gamma.n.minus.one
  }

  delta <- lav_model_delta(lavmodel)
  gamma <- lav_object_gamma(
    lavobject = NULL,
    lavdata = lavdata,
    lavoptions = lavoptions,
    lavsamplestats = lavsamplestats,
    lavh1 = lavh1,
    lavimplied = lavimplied,
    adf = FALSE,
    model_based = TRUE
  )
  wls_obs <- lavsamplestats@WLS.obs
  wls_est <- lav_model_wls_est(lavmodel)
  nobs <- lavsamplestats@nobs
  ntotal <- lavsamplestats@ntotal
  ngroups <- length(wls_obs)
  stat_group <- numeric(ngroups)

  # NOTE: not the packing flags (@eq.constraints/@ceq.simple.only) -- they
  # are both FALSE when equality constraints coexist with inequality
  # constraints or bounds (see lav_con_eq_basis)
  eq_basis <- lav_con_eq_basis(lavmodel)
  lineq_flag <- !is.null(eq_basis)

  if (!lineq_flag) {
    for (g in seq_len(ngroups)) {
      RES <- wls_obs[[g]] - wls_est[[g]]
      delta_c <- lav_mat_ortho_complement(delta[[g]])
      t_dgd <- crossprod(delta_c, gamma[[g]]) %*% delta_c
      t_dgd_inv <- lav_mat_sym_inverse(t_dgd)
      Ng <- if (n_minus_one) nobs[[g]] - 1L else nobs[[g]]
      t_res_delta_c <- crossprod(RES, delta_c)
      stat_group[g] <-
        Ng * drop(t_res_delta_c %*% t_dgd_inv %*% t(t_res_delta_c))
    }
    STAT <- sum(stat_group)
  } else {
    res_all <- do.call("c", wls_obs) - do.call("c", wls_est)
    delta_all <- do.call("rbind", delta)
    delta_g <- delta_all %*% eq_basis
    gamma_inv_weighted <- vector("list", ngroups)
    for (g in seq_len(ngroups)) {
      Ng <- if (n_minus_one) nobs[[g]] - 1L else nobs[[g]]
      gamma_inv_temp <- try(solve(gamma[[g]]), silent = TRUE)
      if (inherits(gamma_inv_temp, "try-error")) {
        gamma_inv_temp <- MASS::ginv(gamma[[g]])
      }
      gamma_inv_weighted[[g]] <- gamma_inv_temp * Ng / ntotal
    }
    GI <- lav_mat_bdiag(gamma_inv_weighted)
    t_dgid <- t(delta_g) %*% GI %*% delta_g
    t_dgid_inv <- MASS::ginv(t_dgid)
    q1 <- drop(t(res_all) %*% GI %*% res_all)
    q2 <- drop(t(res_all) %*%
      GI %*% delta_g %*% t_dgid_inv %*% t(delta_g) %*% GI %*%
      res_all)
    STAT <- ntotal * (q1 - q2)
    stat_group <- STAT * unlist(nobs) / ntotal
  }

  if (!is.null(lavobject)) {
    DF <- lavobject@test[[1]]$df
  } else {
    DF <- lav_pt_df(lavpartable)
    if (!lavmodel@cin.simple.only && nrow(lavmodel@con.jac) > 0L) {
      ceq_idx <- attr(lavmodel@con.jac, "ceq.idx")
      if (length(ceq_idx) > 0L) {
        DF <- DF + qr(lavmodel@con.jac[ceq_idx, , drop = FALSE])$rank
      }
    } else if (lavmodel@ceq.simple.only) {
      DF <- lav_pt_ndat(lavpartable) - max(lavpartable$free)
    }
  }

  list(
    test = "browne.residual.nt.model",
    stat = STAT,
    stat.group = stat_group,
    df = DF,
    refdistr = "chisq",
    pvalue = 1 - stats::pchisq(STAT, DF),
    label = "Browne's residual (NT model-based) test"
  )
}

# =====================================================
# Main entry point
# =====================================================

#' Compute FMG test statistics
#'
#' @param lavobject A lavaan object
#' @param lavsamplestats Sample statistics (extracted from lavobject if NULL)
#' @param lavmodel Model object (extracted from lavobject if NULL)
#' @param lavdata Data object (extracted from lavobject if NULL)
#' @param lavoptions Options (extracted from lavobject if NULL)
#' @param test_unscaled Unscaled test from lavobject@test[[1]]
#' @param test Character string specifying the test (e.g., "peba4_ug_rls")
#' @return List with test results including stat, df, pvalue
#' @keywords internal
lav_test_fmg <- function(lavobject = NULL,
                         lavsamplestats = NULL,
                         lavmodel = NULL,
                         lavimplied = NULL,
                         lavdata = NULL,
                         lavoptions = NULL,
                         test_unscaled = NULL,
                         test_chisq = NULL,
                         e_inv = NULL,
                         delta = NULL,
                         wls_v = NULL,
                         gamma = NULL,
                         test = "peba4") {
  # Extract from lavobject if provided
  if (!is.null(lavobject)) {
    lavsamplestats <- lavobject@SampleStats
    lavmodel <- lavobject@Model
    lavoptions <- lavobject@Options
    lavdata <- lavobject@Data
    if (is.null(test_unscaled)) {
      test_unscaled <- lavobject@test[[1]]
    }
  }

  lav_test_fmg_check_estimator(lavoptions, context = "FMG tests")
  lav_test_fmg_check_mi(lavoptions, context = "FMG tests")

  # Parse test string
  parsed <- lav_test_fmg_parse(test)
  chisq <- lav_test_fmg_resolve_chisq(parsed, lavoptions = lavoptions)
  unbiased <- lav_test_fmg_resolve_unbiased(parsed, lavoptions = lavoptions)
  label <- lav_test_fmg_label(parsed, unbiased = unbiased, chisq = chisq)

  if (is.null(test_chisq)) {
    test_chisq <- test_unscaled
    if (chisq == "rls") {
      if (is.null(lavobject)) {
        return(NULL)
      }
      test_chisq <- lav_test_fmg_browne_nt_model(lavobject = lavobject)
    }
  }
  chisq_stat <- test_chisq$stat

  # Get df
  df <- test_unscaled$df

  # Handle df == 0 case
  if (df == 0L || df < 0L) {
    return(list(
      test = test,
      stat = as.numeric(NA),
      source.stat = chisq_stat,
      source.stat.group = test_chisq$stat.group,
      df = df,
      pvalue = as.numeric(NA),
      refdistr = "chisq",
      method = parsed$method,
      param = parsed$param,
      unbiased = unbiased,
      chisq.type = chisq,
      label = label
    ))
  }

  # Get ugamma matrix (with unbiased option)
  ugamma <- lav_test_fmg_ugamma(
    lavobject = lavobject,
    lavsamplestats = lavsamplestats,
    lavmodel = lavmodel,
    lavimplied = lavimplied,
    lavdata = lavdata,
    lavoptions = lavoptions,
    test_unscaled = test_unscaled,
    e_inv = e_inv,
    delta = delta,
    wls_v = wls_v,
    gamma = gamma,
    unbiased = unbiased
  )

  if (is.null(ugamma)) {
    return(list(
      test = test,
      stat = as.numeric(NA),
      source.stat = chisq_stat,
      source.stat.group = test_chisq$stat.group,
      df = df,
      pvalue = as.numeric(NA),
      refdistr = "chisq",
      method = parsed$method,
      param = parsed$param,
      unbiased = unbiased,
      chisq.type = chisq,
      label = label
    ))
  }

  # Compute eigenvalues (first df eigenvalues)
  eig <- lav_test_fmg_ugamma_eigenvalues(ugamma, df)
  lambdas <- eig$values

  # Compute p-value based on method
  pvalue <- switch(parsed$method,
    "peba" = lav_test_fmg_peba(chisq_stat, lambdas, j = parsed$param),
    "pols" = lav_test_fmg_pols(chisq_stat, lambdas, gamma = parsed$param),
    "pall" = lav_test_fmg_pall(chisq_stat, lambdas),
    "all" = lav_test_fmg_all(chisq_stat, lambdas),
    "sb" = lav_test_fmg_sb(chisq_stat, lambdas),
    "ss" = lav_test_fmg_ss(chisq_stat, lambdas, df),
    "sf" = lav_test_fmg_scaled_f(chisq_stat, lambdas),
    "std" = 1 - stats::pchisq(chisq_stat, df),
    lav_test_fmg_peba(chisq_stat, lambdas, j = 4L) # default
  )

  # Ensure p-value is in valid range
  pvalue <- max(0, min(1, pvalue))
  stat <- lav_test_fmg_chisq_equivalent(pvalue, df)
  stat_group <- lav_test_fmg_stat_group_equivalent(
    stat_group = test_chisq$stat.group,
    source_stat = chisq_stat,
    stat = stat
  )

  # Return TEST structure
  list(
    test = test,
    stat = stat,
    stat.group = stat_group,
    source.stat = chisq_stat,
    source.stat.group = test_chisq$stat.group,
    df = df,
    pvalue = pvalue,
    refdistr = "chisq",
    method = parsed$method,
    param = parsed$param,
    unbiased = unbiased,
    chisq.type = chisq,
    UGamma.eigenvalues = lambdas,
    UGamma.eigenvalues.raw = eig$raw,
    UGamma.eigenvalues.truncated = eig$n_truncated,
    label = label
  )
}

# =====================================================
# ugamma computation
# =====================================================

#' Get ugamma matrix with optional unbiased gamma
#'
#' @param lavobject A lavaan object
#' @param unbiased Logical: use Du & Bentler (2022) unbiased gamma?
#' @return ugamma matrix
#' @keywords internal
lav_test_fmg_ugamma <- function(lavobject = NULL,
                                lavsamplestats = NULL,
                                lavmodel = NULL,
                                lavimplied = NULL,
                                lavdata = NULL,
                                lavoptions = NULL,
                                test_unscaled = NULL,
                                e_inv = NULL,
                                delta = NULL,
                                wls_v = NULL,
                                gamma = NULL,
                                unbiased = FALSE) {
  if (!is.null(lavobject)) {
    lavsamplestats <- lavobject@SampleStats
    lavmodel <- lavobject@Model
    lavimplied <- lavobject@implied
    lavoptions <- lavobject@Options
    lavdata <- lavobject@Data
    if (is.null(test_unscaled)) {
      test_unscaled <- lavobject@test[[1]]
    }
  }

  ngroups <- lavdata@ngroups

  if (unbiased) {
    # Recompute gamma with unbiased = TRUE
    gamma <- vector("list", ngroups)
    for (g in seq_len(ngroups)) {
      gamma[[g]] <- lav_samp_gamma(
        m_y = lavdata@X[[g]],
        meanstructure = lavoptions$meanstructure,
        unbiased = TRUE
      )
    }
  } else {
    if (is.null(gamma)) {
      # stored NACOV, else recompute with the fit-time settings
      if (!is.null(lavobject)) {
        gamma <- lav_gamma_used(lavobject)
      } else {
        gamma <- lav_gamma_used(
          lavdata = lavdata,
          lavoptions = lavoptions,
          lavsamplestats = lavsamplestats,
          lavh1 = NULL,
          lavimplied = lavimplied
        )
      }
    }
    if (!is.list(gamma)) {
      gamma <- list(gamma)
    }
  }

  # Get U matrix and ugamma using existing infrastructure
  out <- lav_test_sb(
    lavobject = lavobject,
    lavsamplestats = lavsamplestats,
    lavmodel = lavmodel,
    lavimplied = lavimplied,
    lavdata = lavdata,
    lavoptions = lavoptions,
    test_unscaled = test_unscaled,
    e_inv = e_inv,
    delta = delta,
    wls_v = wls_v,
    m_gamma = gamma,
    method = "original",
    return_ugamma = TRUE,
    return_u = TRUE
  )

  if (is.null(out)) {
    return(NULL)
  }

  # If unbiased, recompute ugamma with new gamma
  if (unbiased) {
    U <- out$UfromUGamma
    if (is.null(U)) {
      return(NULL)
    }

    # Rescale gamma by group weights
    fg <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal
    Gamma_scaled <- vector("list", length(gamma))
    for (g in seq_along(gamma)) {
      Gamma_scaled[[g]] <- gamma[[g]] / fg[g]
    }
    Gamma_all <- lav_mat_bdiag(Gamma_scaled)
    ugamma <- U %*% Gamma_all
  } else {
    ugamma <- out$UGamma # field name as returned by lav_test_sb()
  }

  ugamma
}

# =====================================================
# P-value methods
# =====================================================

#' Pure R Imhof p-value for quadratic forms
#'
#' Computes P[Q > q] where Q = sum(lambda_j * chi^2(h_j, delta_j)).
#' This is a small R port of the Imhof integration used by CompQuadForm,
#' using stats::integrate().
#'
#' @param q Observed quadratic-form value
#' @param lambda Eigenvalues/weights
#' @param h Degrees of freedom per component
#' @param delta Noncentrality parameters per component
#' @param epsabs Absolute integration tolerance
#' @param epsrel Relative integration tolerance
#' @param limit Maximum number of subdivisions
#' @return List with Qq and abserr
#' @keywords internal
lav_test_fmg_imhof <- function(q, lambda, h = rep(1, length(lambda)),
                               delta = rep(0, length(lambda)),
                               epsabs = 1e-6, epsrel = 1e-6,
                               limit = 10000L) {
  lambda <- as.numeric(lambda)
  h <- as.numeric(h)
  delta <- as.numeric(delta)

  keep <- abs(lambda) > .Machine$double.eps
  lambda <- lambda[keep]
  h <- h[keep]
  delta <- delta[keep]

  if (length(lambda) == 0L) {
    return(list(Qq = ifelse(q < 0, 1, 0), abserr = 0))
  }

  if (length(lambda) == 1L) {
    if (lambda > 0) {
      Qq <- stats::pchisq(q / lambda,
        df = h, ncp = delta,
        lower.tail = FALSE
      )
    } else if (q < 0) {
      Qq <- stats::pchisq(q / lambda, df = h, ncp = delta)
    } else {
      Qq <- 0
    }
    return(list(Qq = Qq, abserr = 0))
  }

  integrand <- function(u) {
    vapply(u, function(ui) {
      if (ui == 0) {
        return(0.5 * (sum((h + delta) * lambda) - q))
      }

      lu <- lambda * ui
      lu2 <- lu * lu

      theta <- 0.5 * sum(h * atan(lu) + delta * lu / (1 + lu2)) -
        0.5 * q * ui
      log_rho <- sum(0.25 * h * log1p(lu2) +
        0.5 * delta * lu2 / (1 + lu2))

      sin(theta) / (ui * exp(log_rho))
    }, numeric(1L))
  }

  out <- stats::integrate(
    integrand,
    lower = 0,
    upper = Inf,
    rel.tol = epsrel,
    abs.tol = epsabs,
    subdivisions = limit
  )

  list(Qq = 0.5 + out$value / pi, abserr = out$abs.error)
}

#' Penalized EBA p-value (Foldnes et al. 2024)
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @param j Number of blocks
#' @return p-value
#' @keywords internal
lav_test_fmg_peba <- function(chisq, lambdas, j = 4L) {
  m <- length(lambdas)
  if (m == 0L) {
    return(as.numeric(NA))
  }

  k <- ceiling(m / j)
  eig <- c(lambdas, rep(NA_real_, k * j - m))
  dim(eig) <- c(k, j)
  eig_means <- colMeans(eig, na.rm = TRUE)
  eig_mean <- mean(lambdas)
  repeated <- rep(eig_means, each = k)[seq_len(m)]
  penalized <- (repeated + eig_mean) / 2

  lav_test_fmg_imhof(chisq, penalized)$Qq
}

#' Penalized OLS p-value (Foldnes et al. 2024)
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @param gamma Penalization parameter
#' @return p-value
#' @keywords internal
lav_test_fmg_pols <- function(chisq, lambdas, gamma = 2) {
  m <- length(lambdas)
  if (m == 0L) {
    return(as.numeric(NA))
  }

  if (m == 1L) {
    lambda_hat <- lambdas
  } else {
    x <- seq_along(lambdas)
    beta1_hat <- 1 / gamma * stats::cov(x, lambdas) / stats::var(x)
    beta0_hat <- mean(lambdas) - beta1_hat * mean(x)
    lambda_hat <- pmax(beta0_hat + beta1_hat * x, 0)
  }

  lav_test_fmg_imhof(chisq, lambda_hat)$Qq
}

#' Penalized all eigenvalues (for nested models)
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @return p-value
#' @keywords internal
lav_test_fmg_pall <- function(chisq, lambdas) {
  if (length(lambdas) == 0L) {
    return(as.numeric(NA))
  }

  penalized <- lambdas / 2 + mean(lambdas) / 2
  lav_test_fmg_imhof(chisq, penalized)$Qq
}

#' All eigenvalues exact p-value
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @return p-value
#' @keywords internal
lav_test_fmg_all <- function(chisq, lambdas) {
  if (length(lambdas) == 0L) {
    return(as.numeric(NA))
  }

  lav_test_fmg_imhof(chisq, lambdas)$Qq
}

#' Satorra-Bentler via eigenvalues
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @return p-value
#' @keywords internal
lav_test_fmg_sb <- function(chisq, lambdas) {
  m <- length(lambdas)
  if (m == 0L) {
    return(as.numeric(NA))
  }

  1 - stats::pchisq(chisq * m / sum(lambdas), df = m)
}

#' Scaled and shifted p-value
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @param df Degrees of freedom
#' @return p-value
#' @keywords internal
lav_test_fmg_ss <- function(chisq, lambdas, df) {
  if (length(lambdas) == 0L || df == 0L) {
    return(as.numeric(NA))
  }

  tr_ug <- sum(lambdas)
  tr_ug2 <- sum(lambdas^2)

  if (tr_ug2 < .Machine$double.eps) {
    return(as.numeric(NA))
  }

  a <- sqrt(df / tr_ug2)
  b <- df - sqrt(df * tr_ug^2 / tr_ug2)
  t3 <- chisq * a + b

  1 - stats::pchisq(t3, df)
}

#' Scaled F p-value (Wu & Lin 2016)
#'
#' @param chisq Chi-square statistic
#' @param lambdas Eigenvalues of ugamma
#' @return p-value
#' @keywords internal
lav_test_fmg_scaled_f <- function(chisq, lambdas) {
  if (length(lambdas) == 0L) {
    return(as.numeric(NA))
  }

  s1 <- sum(lambdas)
  s2 <- sum(lambdas^2)
  s3 <- sum(lambdas^3)
  denom <- 2 * s1 * s2^2 - s1^2 * s3 + 2 * s2 * s3

  if (denom > 0) {
    d1f3 <- s1 * (s1^2 * s2 - 2 * s2^2 + 4 * s1 * s3) / denom
    d2f3 <- (s1^2 * s2 + 2 * s2^2) / (s3 * s1 - s2^2) + 6
    if (d2f3 < 6) d2f3 <- Inf
    cf3 <- s1 * (s1^2 * s2 - 2 * s2^2 + 4 * s1 * s3) /
      (s1^2 * s2 - 4 * s2^2 + 6 * s1 * s3)
  } else {
    d1f3 <- Inf
    d2f3 <- s1^2 / s2 + 4
    cf3 <- s1 * (s1^2 + 2 * s2) / (s1^2 + 4 * s2)
  }

  1 - stats::pf(chisq / cf3, d1f3, d2f3)
}

# =====================================================
# Nested model implementation
# =====================================================

#' FMG test for nested model comparison
#'
#' @param m0 Restricted (null) model
#' @param m1 Unrestricted (alternative) model
#' @param test Test specification string (e.g., "pall_ug_ml")
#' @return List with test results
#' @keywords internal
lav_test_fmg_nested <- function(m0, m1, test = "pall") {
  # Ensure m0 has more df than m1
  if (m0@test[[1]]$df < m1@test[[1]]$df) {
    tmp <- m0
    m0 <- m1
    m1 <- tmp
  }

  lav_test_fmg_check_estimator(m0@Options, context = "FMG nested tests")
  lav_test_fmg_check_estimator(m1@Options, context = "FMG nested tests")
  if (!isTRUE(m0@Options$estimator.orig == m1@Options$estimator.orig)) {
    lav_msg_stop(gettextf(
      "FMG nested tests require both models to use the same estimator; found %1$s and %2$s.",
      dQuote(m0@Options$estimator.orig),
      dQuote(m1@Options$estimator.orig)
    ))
  }
  lav_test_fmg_check_mi(m0@Options, context = "FMG nested tests")
  lav_test_fmg_check_mi(m1@Options, context = "FMG nested tests")

  # Parse test string
  parsed <- lav_test_fmg_parse(test)
  chisq <- lav_test_fmg_resolve_chisq(parsed, lavoptions = m1@Options)
  unbiased <- lav_test_fmg_resolve_unbiased(parsed, lavoptions = m1@Options)
  label <- lav_test_fmg_label(parsed, unbiased = unbiased, chisq = chisq)
  if (!parsed$method %in% c("pall", "all", "peba", "pols")) {
    lav_msg_stop(gettextf(
      "FMG nested tests only support test= %1$s; found %2$s.",
      lav_msg_view(c("pall", "all", "peba", "pols"), "or"),
      dQuote(parsed$method)
    ))
  }

  # Get df difference
  df0 <- m0@test[[1]]$df
  df1 <- m1@test[[1]]$df
  df <- df0 - df1

  if (df == 0L) {
    lav_msg_stop(gettext(
      "Cannot test models with the same degrees of freedom."
    ))
  }

  # Get chi-square difference
  if (chisq == "rls") {
    chisq0 <- lav_test_fmg_browne_nt_model(lavobject = m0)$stat
    chisq1 <- lav_test_fmg_browne_nt_model(lavobject = m1)$stat
  } else {
    chisq0 <- m0@test[[1]]$stat
    chisq1 <- m1@test[[1]]$stat
  }
  chisq_diff <- chisq0 - chisq1

  # Get ugamma for nested comparison (Satorra 2000 projection)
  ugamma <- lav_test_fmg_ugamma_nested(m0, m1, unbiased = unbiased)

  if (is.null(ugamma)) {
    return(list(
      test = test,
      stat = as.numeric(NA),
      source.stat = chisq_diff,
      df = df,
      pvalue = as.numeric(NA),
      label = label
    ))
  }

  # Compute first df eigenvalues
  eig <- lav_test_fmg_ugamma_eigenvalues(ugamma, df)
  lambdas <- eig$values

  # Compute p-value (pall is recommended for nested)
  pvalue <- switch(parsed$method,
    "pall" = lav_test_fmg_pall(chisq_diff, lambdas),
    "all"  = lav_test_fmg_all(chisq_diff, lambdas),
    "peba" = lav_test_fmg_peba(chisq_diff, lambdas, j = parsed$param),
    "pols" = lav_test_fmg_pols(chisq_diff, lambdas, gamma = parsed$param)
  )
  pvalue <- max(0, min(1, pvalue))
  stat <- lav_test_fmg_chisq_equivalent(pvalue, df)

  list(
    test = test,
    stat = stat,
    source.stat = chisq_diff,
    df = df,
    pvalue = pvalue,
    chisq.type = chisq,
    unbiased = unbiased,
    UGamma.eigenvalues = lambdas,
    UGamma.eigenvalues.raw = eig$raw,
    UGamma.eigenvalues.truncated = eig$n_truncated,
    label = label
  )
}

#' Compute ugamma for nested models (Satorra 2000 projection)
#'
#' @param m0 Restricted model
#' @param m1 Unrestricted model
#' @param unbiased Use unbiased gamma?
#' @return ugamma matrix
#' @keywords internal
lav_test_fmg_ugamma_nested <- function(m0, m1, unbiased = FALSE) {
  lavdata <- m1@Data
  ngroups <- lavdata@ngroups

  # Get gamma from m1 (with optional unbiased)
  if (unbiased) {
    gamma <- vector("list", ngroups)
    for (g in seq_len(ngroups)) {
      gamma[[g]] <- lav_samp_gamma(
        m_y = lavdata@X[[g]],
        meanstructure = m1@Options$meanstructure,
        unbiased = TRUE
      )
    }
  } else {
    gamma <- lavTech(m1, "gamma")
    if (is.null(gamma)) {
      gamma <- lavTech(m0, "gamma")
    }
    if (is.null(gamma)) {
      lav_msg_stop(gettextf(
        paste(
          "Could not calculate the gamma matrix.",
          "Use estimator= %1$s or test= %2$s when fitting your lavaan model."
        ),
        dQuote("MLM"), dQuote("satorra.bentler")
      ))
    }
    if (!is.list(gamma)) gamma <- list(gamma)
  }

  wls_v <- lavTech(m1, "WLS.V")
  PI <- lav_model_delta(m1@Model)
  p_inv <- lavTech(m1, "inverted.information")

  if (is.null(p_inv)) {
    return(NULL)
  }

  A <- lav_test_diff_a(m1, m0, method = "delta", reference = "H1")

  # Handle equality constraints (same orthonormal basis as used inside
  # lav_test_diff_a; also covers equality constraints coexisting with
  # inequality constraints/bounds)
  eq_basis_m1 <- lav_con_eq_basis(m1@Model)
  if (!is.null(eq_basis_m1)) {
    A <- A %*% t(eq_basis_m1)
  }

  # Safety check: remove zero rows/columns
  APA <- A %*% p_inv %*% t(A)
  col_sums <- colSums(APA)
  row_sums <- rowSums(APA)
  empty_idx <- which(abs(col_sums) < .Machine$double.eps^0.5 &
    abs(row_sums) < .Machine$double.eps^0.5)
  if (length(empty_idx) > 0L) {
    A <- A[-empty_idx, , drop = FALSE]
  }

  if (nrow(A) == 0L) {
    return(NULL)
  }

  # PAAPAAP projection
  PAAPAAP <- p_inv %*% t(A) %*% MASS::ginv(A %*% p_inv %*% t(A)) %*%
    A %*% p_inv

  # Build global matrices
  fg <- unlist(m1@SampleStats@nobs) / m1@SampleStats@ntotal
  Gamma_f <- vector("list", length(gamma))
  for (g in seq_along(gamma)) {
    Gamma_f[[g]] <- gamma[[g]] / fg[g]
  }
  Gamma_all <- lav_mat_bdiag(Gamma_f)

  V_f <- wls_v
  for (g in seq_along(wls_v)) {
    V_f[[g]] <- fg[g] * wls_v[[g]]
  }
  V_all <- lav_mat_bdiag(V_f)

  PI_all <- do.call(rbind, PI)

  # U_global (eq. 22 in Satorra 2000)
  U_all <- V_all %*% PI_all %*% PAAPAAP %*% t(PI_all) %*% V_all
  U_all %*% Gamma_all
}

lav_samp_icov <- function(cov_1 = NULL, ridge = 0.0, x_idx = integer(0L),
                                 ngroups = 1L, g = 1L) {

  c_s <- tryCatch(chol(cov_1), error = function(e) NULL)

  # what if this fails...
  # ridge exogenous part only (if any); this may help for GLS (but not ML)
  if (is.null(c_s)) {
    if (length(x_idx) > 0L && ridge > 0) {
      # maybe, we can fix it by gently ridging the exo variances
      ridge_eps <- ridge
      diag(cov_1)[x_idx] <- diag(cov_1)[x_idx] + ridge_eps

      # try again
      c_s <- tryCatch(chol(cov_1), error = function(e) NULL)

      if (is.null(c_s)) {
        # fatal stop after all
        lav_msg_stop(gettext(
          "sample covariance matrix is not positive-definite"))
      } else {
        icov <- chol2inv(c_s)
        d <- diag(c_s)
        cov_log_det <- 2 * sum(log(d))
        # give a warning
        if (ngroups > 1) {
          lav_msg_warn(gettextf(
            "sample covariance matrix in group: %s is not
            positive-definite", g))
        } else {
          lav_msg_warn(gettext(
            "sample covariance matrix is not positive-definite"))
        }
      }
    } else {
      # fatal stop
      lav_msg_stop(gettext("sample covariance matrix is not positive-definite"))
    }
  } else {
    icov <- chol2inv(c_s)
    d <- diag(c_s)
    cov_log_det <- 2 * sum(log(d))
  }

  list(icov = icov, cov.log.det = cov_log_det)
}

# compute sample statistics for the unrestricted (h1) model
# and also the logl (if available)
lav_h1_implied_logl <- function(lavdata = NULL,
                                lavsamplestats = NULL,
                                lavpartable = NULL, # multilevel + missing
                                lavoptions = NULL) {
  lavpta <- NULL
  if (!is.null(lavpartable)) {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }

  # single-level case
  if (lavdata@nlevels == 1L) {

    # missing data
    if (lavsamplestats@missing.flag) {
      if (lavoptions$conditional.x) {
        implied <- list(
          res.cov = lavsamplestats@res.cov,
          res.int = lavsamplestats@res.int,
          res.slopes = lavsamplestats@res.slopes,
          cov.x = lavsamplestats@cov.x,
          mean.x = lavsamplestats@mean.x,
          res.th = lavsamplestats@res.th,
          group.w = lavsamplestats@group.w
        )
        # insert the EM-based estimates (obtained by partitioning the
        # joint EM moments; see lav_samp_from_data)
        for (g in 1:lavdata@ngroups) {
          if (!is.null(lavsamplestats@missing.h1[[g]]$res.cov)) {
            implied$res.cov[[g]] <- lavsamplestats@missing.h1[[g]]$res.cov
            implied$res.int[[g]] <- lavsamplestats@missing.h1[[g]]$res.int
            implied$res.slopes[[g]] <-
              lavsamplestats@missing.h1[[g]]$res.slopes
          }
          # else: zero coverage; keep the pairwise-based sample statistics
        }
      } else {
        implied <- list(
          cov = lavsamplestats@cov,
          mean = lavsamplestats@mean,
          th = lavsamplestats@th,
          group.w = lavsamplestats@group.w
        )
        # insert ML estimates
        for (g in 1:lavdata@ngroups) {
          # zero coverage?
          if (any(lav_mat_vech(lavdata@Mp[[g]]$coverage,
                                  diagonal = FALSE) == 0L)) {
            out <- lav_mvn_mi_h1_est_moments_chol(
              lavdata = lavdata, lavsamplestats = lavsamplestats,
              lavoptions = lavoptions, group = g)
            implied$cov[[g]] <- out$Sigma
            implied$mean[[g]] <- out$Mu
          } else {
            # regular EM estimates
            implied$cov[[g]] <- lavsamplestats@missing.h1[[g]]$sigma
            implied$mean[[g]] <- lavsamplestats@missing.h1[[g]]$mu
          }
        }
      }

    # complete data
    } else {
      if (lavoptions$conditional.x) {
        implied <- list(
          res.cov = lavsamplestats@res.cov,
          res.int = lavsamplestats@res.int,
          res.slopes = lavsamplestats@res.slopes,
          cov.x = lavsamplestats@cov.x,
          mean.x = lavsamplestats@mean.x,
          res.th = lavsamplestats@res.th,
          group.w = lavsamplestats@group.w
        )
      } else {
        implied <- list(
          cov = lavsamplestats@cov,
          mean = lavsamplestats@mean,
          th = lavsamplestats@th,
          group.w = lavsamplestats@group.w
        )
      }
    } # complete data

    logl <- lav_h1_logl(
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      h1_implied = implied,
      lavoptions = lavoptions
    )
  } else {
    # estimate Mu.B, Mu.W, Sigma.B and Sigma.W for unrestricted model
    ngroups <- lavdata@ngroups
    nlevels <- lavdata@nlevels

    # EM control arguments (fall back to the multilevel defaults if
    # lavoptions is incomplete, e.g. when refitting objects created by
    # older versions)
    em_h1_tol <- lavoptions$em.h1.args$tol
    if (!is.numeric(em_h1_tol)) {
      em_h1_tol <- 1e-04
    }
    em_h1_max_iter <- lavoptions$em.h1.args$max_iter
    if (!is.numeric(em_h1_max_iter)) {
      em_h1_max_iter <- 5000L
    }
    em_h1_min_variance <- lavoptions$em.h1.args$min_variance
    if (!is.numeric(em_h1_min_variance)) {
      em_h1_min_variance <- 1e-05
    }
    non_pd_action <- lavoptions$em.h1.args$non_pd_action
    if (is.null(non_pd_action)) {
      non_pd_action <- "warn" # backwards compatibility
    }
    non_pd_tol <- lavoptions$em.h1.args$non_pd_tol
    if (!is.numeric(non_pd_tol)) {
      non_pd_tol <- 1e-05
    }
    em_h1_accel <- lavoptions$em.h1.args$acceleration
    if (is.null(em_h1_accel)) {
      em_h1_accel <- "none" # backwards compatibility
    }
    em_h1_fused <- lavoptions$em.h1.args$fused
    if (is.null(em_h1_fused)) {
      em_h1_fused <- TRUE
    }
    implied <- list(
      cov = vector("list", length = ngroups * nlevels),
      mean = vector("list", length = ngroups * nlevels)
    )
    loglik_group <- numeric(lavdata@ngroups)

    for (g in 1:lavdata@ngroups) {
      if (lav_verbose()) {
        cat("\n\nfitting unrestricted (H1) model in group ", g, "\n")
      }
      if (lavsamplestats@missing.flag &&
          (is.null(lavoptions$h1.missing.method) || # backwards compatibility
           lavoptions$h1.missing.method == "em")) {
        # missing data: EM algorithm (new in 0.7-1; this mimics Mplus)
        out_1 <- lav_mvn_cl_mi_em_sat(
          y1 = lavdata@X[[g]],
          y2 = lavsamplestats@YLp[[g]][[2]]$Y2,
          lp = lavdata@Lp[[g]],
          mp = lavdata@Mp[[g]],
          loglik_x = lavsamplestats@YLp[[g]][[2]]$loglik.x,
          tol = em_h1_tol,
          max_iter = em_h1_max_iter,
          min_variance = em_h1_min_variance,
          acceleration = em_h1_accel,
          fused = em_h1_fused
        )
      } else if (lavsamplestats@missing.flag) {
        # missing data: h1.missing.method = "fiml"
        # fit the unrestricted model using a cholesky parameterization
        # and nlminb() (this was the default in 0.6-* and 0.7-1)

        # 1. first a few EM iteration faking complete data
        # Y1 <- lavdata@X[[g]]
        # cluster.idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
        # Y2.complete <- unname(as.matrix(aggregate(Y1,
        #                by = list(cluster.idx),
        #                FUN = function(x) {
        #                    if( all(is.na(x)) ) { # all elements are NA
        #                        as.numeric(0)     # in this cluster
        #                    } else {
        #                        mean(x, na.rm = TRUE)
        #                    }
        #                })[,-1]))
        # YLp = lavsamplestats@YLp[[g]]
        # YLp[[2]]$Y2 <- Y2.complete
        # OUT <- lav_mvn_cl_em_sat(
        #    YLp      = YLp,
        #    Lp       = lavdata@Lp[[g]],
        #    verbose  = TRUE, # for now
        #    tol      = 1e-04,
        #    min.variance = 1e-05,
        #    max.iter = 5L)
        ## create tmp lav1, only for this group
        # implied$cov[[(g-1)*nlevels + 1L]] <- OUT$Sigma.W
        # implied$cov[[(g-1)*nlevels + 2L]] <- OUT$Sigma.B
        # implied$mean[[(g-1)*nlevels + 1L]] <- OUT$Mu.W
        # implied$mean[[(g-1)*nlevels + 2L]] <- OUT$Mu.B
        # loglik.group[g] <- OUT$logl
        # lavh1 <- list(implied = implied, logl = sum(loglik.group))
        # lavpartable <- lav_pt_unrestricted(lavdata = lavdata,
        #     lavsamplestats = lavsamplestats, lavoptions = lavoptions,
        #     lavpta = lavpta, lavh1 = lavh1)
        # lavpartable$lower <- rep(-Inf, length(lavpartable$lhs))
        # var.idx <- which(lavpartable$free > 0L &
        #                 lavpartable$op == "~~" &
        #                 lavpartable$lhs == lavpartable$rhs)
        # lavpartable$lower[var.idx] <- 1e-05

        lavpartable <- lav_pt_unrestricted_chol(
          lavdata = lavdata,
          lavoptions = lavoptions, lavpta = lavpta
        )

        lavoptions2 <- lavoptions
    lavoptions2$estimator <- "ML"
        lavoptions2$se <- "none"
        lavoptions2$test <- "none"
        lavoptions2$do.fit <- TRUE
    lavoptions2$optim.method <- "nlminb"
        lavoptions2$h1 <- FALSE
    lavoptions2$implied <- TRUE
    lavoptions2$loglik <- TRUE
        lavoptions2$baseline <- FALSE
        lavoptions2$fit.by.level <- FALSE
        lavoptions2$fixed.x <- FALSE # even if model uses fixed.x=TRUE
        lavoptions2$model.type <- "unrestricted"
        lavoptions2$optim.attempts <- 4L
        lavoptions2$check.gradient <- FALSE
        lavoptions2$optim.force.converged <- TRUE # for now...
        lavoptions2$control <- list(rel.tol = 1e-7)
        # FIT <- lavaan(lavpartable, slot_options = lavoptions2,
        #          slot_sample_stats = lavsamplestats,
        #          slot_data = lavdata, sloth1 = lavh1)
        fit <- lavaan(lavpartable,
          slot_options = lavoptions2,
          slot_sample_stats = lavsamplestats,
          slot_data = lavdata,
          warn = FALSE
        )
        out_1 <- list(
          Sigma.W = fit@implied$cov[[1]],
          Sigma.B = fit@implied$cov[[2]],
          Mu.W = fit@implied$mean[[1]],
          Mu.B = fit@implied$mean[[2]],
          logl = fit@loglik$loglik
        )
        # if(lavoptions$fixed.x) {
        #   OUT$logl <- OUT$logl - lavsamplestats@YLp[[g]][[2]]$loglik.x
        # }
      } else {
        # complete data
        out_1 <- lav_mvn_cl_em_sat(
          ylp = lavsamplestats@YLp[[g]],
          lp = lavdata@Lp[[g]],
          tol = em_h1_tol,
          min_variance = em_h1_min_variance,
          max_iter = em_h1_max_iter,
          acceleration = em_h1_accel
        )
      }
      if (lav_verbose()) {
        cat("\n")
      }

      # check for (near) singular within/between covariance matrices;
      # this typically happens when a (within or between) variance
      # approaches zero, but the full matrix may also be (near) singular
      if (non_pd_action != "none") {
        gtxt <- if (ngroups > 1L) {
          gettextf(" in group %s", g)
        } else {
          ""
        }
        for (l in 1:2) {
          if (l == 1L) {
            sigma_l <- out_1$Sigma.W
            level_txt <- "within"
          } else {
            sigma_l <- out_1$Sigma.B
            level_txt <- "between"
          }
          if (NROW(sigma_l) == 0L) {
            next
          }
          msg <- NULL
          near_zero_idx <- which(diag(sigma_l) <= non_pd_tol)
          if (length(near_zero_idx) > 0L) {
            msg <- gettextf(
              "H1 estimation resulted in a %1$s covariance matrix%2$s with
               (near) zero variances for some of the observed variables:
               %3$s.", level_txt, gtxt,
              lav_msg_view(lavdata@ov.names.l[[g]][[l]][near_zero_idx]))
          } else {
            ev <- eigen(sigma_l, symmetric = TRUE, only.values = TRUE)$values
            if (any(ev < non_pd_tol)) {
              msg <- gettextf(
                "H1 estimation resulted in a (near) singular %1$s covariance
                 matrix%2$s: its smallest eigenvalue is smaller than %3$s.",
                level_txt, gtxt, format(non_pd_tol))
            }
          }
          if (!is.null(msg)) {
            if (non_pd_action == "stop") {
              lav_msg_stop(paste(msg, gettext(
                "Set the non_pd_action element of the em.h1.args= argument
                 to \"warn\" or \"none\" to continue anyway.")))
            } else {
              lav_msg_warn(msg)
            }
          }
        }
      }

      # new in 0.6-18: ensure Mu.W[both.idx] is zero (post-estimation!)
      # (not correctly; fixed in 0.6-19...)
      both_idx   <- lavdata@Lp[[g]]$both.idx[[2]]
      within_idx <- lavdata@Lp[[g]]$within.idx[[2]]
      ov_idx     <- lavdata@Lp[[g]]$ov.idx
      p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))
      mu_w_tilde <- mu_b_tilde <- mu_wb_tilde <- numeric(p_tilde)
      mu_w_tilde[ov_idx[[1]]] <- out_1$Mu.W
      mu_b_tilde[ov_idx[[2]]] <- out_1$Mu.B
      mu_wb_tilde[both_idx] <- (mu_b_tilde[both_idx] + mu_w_tilde[both_idx])
      mu_w_tilde[both_idx] <- 0
      mu_b_tilde[both_idx] <- mu_wb_tilde[both_idx]
      out_1$Mu.W <- mu_w_tilde[ov_idx[[1]]]
      out_1$Mu.B <- mu_b_tilde[ov_idx[[2]]]

      # store in implied
      implied$cov[[(g - 1) * nlevels + 1L]] <- out_1$Sigma.W
      implied$cov[[(g - 1) * nlevels + 2L]] <- out_1$Sigma.B
      implied$mean[[(g - 1) * nlevels + 1L]] <- out_1$Mu.W
      implied$mean[[(g - 1) * nlevels + 2L]] <- out_1$Mu.B

      # store logl per group
      loglik_group[g] <- out_1$logl
    }

    logl <- list(loglik = sum(loglik_group), loglik.group = loglik_group)
  }

  list(implied = implied, logl = logl)
}

# various rotation criteria and their gradients
# YR 05 April 2019: initial version
# YR 14 June  2019: add more rotation criteria
# YR 22 Dec   2025: add multigroup rotation criteria

# references:
#
# Bernaards, C. A., & Jennrich, R. I. (2005). Gradient projection algorithms
# and software for arbitrary rotation criteria in factor analysis. Educational
# and Psychological Measurement, 65(5), 676-696.
# old website: http://web.archive.org/web/20180708170331/http://www.stat.ucla.edu/research/gpa/splusfunctions.net # nolint
#
# Browne, M. W. (2001). An overview of analytic rotation in exploratory factor
# analysis. Multivariate behavioral research, 36(1), 111-150.
#
# Mulaik, S. A. (2010). Foundations of factor analysis (Second Edition).
# Boca Raton: Chapman and Hall/CRC.
#
# De Roover, K., & Vermunt, J. K. (2019). On the exploratory road to unraveling
# factor loading non-invariance: A new multigroup rotation approach. Structural
# Equation Modeling: A Multidisciplinary Journal, 26(6), 905-923.

# Note: this is YR's implementation, not a copy of the GPArotation
#       package
#
# Why did I write my own functions (and not use the GPArotation):
# - to better understand what is going on
# - to have direct access to the gradient functions
# - to avoid yet another dependency
# - to simplify further experiments


# Orthomax family (Harman, 1960)
#
# gamma = 0   -> quartimax
# gamma = 1/2 -> biquartimax
# gamma = 1/P -> equamax
# gamma = 1   -> varimax
#
lav_mat_rotate_orthomax <- function(mm_lambda = NULL, orthomax_gamma = 1,
                                       ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda
  # center L2 column-wise
  c_l2 <- t(t(l2) - orthomax_gamma * colMeans(l2))
  out <- -1 * sum(l2 * c_l2) / 4

  if (grad) {
    attr(out, "grad") <- -1 * mm_lambda * c_l2
  }

  out
}

# Crawford-Ferguson (1970) family
#
# combine penalization for 1) row complexity, and 2) column complexity
# if combined with orthogonal rotation, this is equivalent to the
# orthomax family:
#
#     quartimax -> gamma = 0 (only row complexity)
#     varimax   -> gamma = 1/nrow
#     equamax   -> gamma = ncol/(2*nrow)
#     parsimax  -> gamma = (ncol - 1)/(nrow + ncol - 2)
#     factor parsimony -> gamma = 1 (only column complexity)
#
# the Crawford-Ferguson family is also equivalent to the oblimin family
# if the latter is restricted to orthogonal rotation
#
lav_mat_rotate_cf <- function(mm_lambda = NULL, cf_gamma = 0, ...,
                                 grad = FALSE) {
  # check if gamma is between 0 and 1?
  n_row <- nrow(mm_lambda)
  n_col <- ncol(mm_lambda)
  row1 <- matrix(1.0, n_col, n_col)
  diag(row1) <- 0.0
  col1 <- matrix(1.0, n_row, n_row)
  diag(col1) <- 0.0

  l2 <- mm_lambda * mm_lambda
  lr <- l2 %*% row1
  lc <- col1 %*% l2

  f1 <- sum(l2 * lr) / 4
  f2 <- sum(l2 * lc) / 4

  out <- (1 - cf_gamma) * f1 + cf_gamma * f2

  if (grad) {
    attr(out, "grad") <- ((1 - cf_gamma) * mm_lambda * lr) +
      (cf_gamma * mm_lambda * lc)
  }

  out
}

# Oblimin family (Carroll, 1960; Harman, 1976)
#
# quartimin   -> gamma = 0
# biquartimin -> gamma = 1/2
# covarimin   -> gamma = 1
#
# if combined with orthogonal rotation, this is equivalent to the
# orthomax family (they have the same optimizers):
#
# gamma = 0   -> quartimax
# gamma = 1/2 -> biquartimax
# gamma = 1   -> varimax
# gamma = P/2 -> equamax
#
lav_mat_rotate_oblimin <- function(mm_lambda = NULL, oblimin_gamma = 0, ...,
                                      grad = FALSE) {
  n_row <- nrow(mm_lambda)
  n_col <- ncol(mm_lambda)
  row1 <- matrix(1.0, n_col, n_col)
  diag(row1) <- 0.0

  l2 <- mm_lambda * mm_lambda
  lr <- l2 %*% row1
  jp <- matrix(1, n_row, n_row) / n_row

  # see Jennrich (2002, p. 11)
  tmp <- (diag(n_row) - oblimin_gamma * jp) %*% lr

  # same as t( t(L2) - gamma * colMeans(L2) ) %*% ROW1

  out <- sum(l2 * tmp) / 4

  if (grad) {
    attr(out, "grad") <- mm_lambda * tmp
  }

  out
}


# quartimax criterion
# Carroll (1953); Saunders (1953) Neuhaus & Wrigley (1954); Ferguson (1954)
# we use here the equivalent 'Ferguson, 1954' variant
# (See Mulaik 2010, p. 303)
lav_mat_rotate_quartimax <- function(mm_lambda = NULL, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda
  out <- -1 * sum(l2 * l2) / 4

  if (grad) {
    attr(out, "grad") <- -1 * mm_lambda * l2
  }

  out
}


# varimax criterion
# Kaiser (1958, 1959)
#
# special case of the Orthomax family (Harman, 1960), where gamma = 1
# see Jennrich (2001, p. 296)
lav_mat_rotate_varimax <- function(mm_lambda = NULL, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda
  # center L2 column-wise
  c_l2 <- t(t(l2) - colMeans(l2))
  out <- -1 * abs(sum(l2 * c_l2)) / 4 # abs needed?

  if (grad) {
    attr(out, "grad") <- -1 * mm_lambda * c_l2
  }

  out
}

# quartimin criterion (part of Carroll's oblimin family
lav_mat_rotate_quartimin <- function(mm_lambda = NULL, ..., grad = FALSE) {
  n_col <- ncol(mm_lambda)
  row1 <- matrix(1.0, n_col, n_col)
  diag(row1) <- 0.0

  l2 <- mm_lambda * mm_lambda
  lr <- l2 %*% row1

  out <- sum(l2 * lr) / 4

  if (grad) {
    attr(out, "grad") <- mm_lambda * lr
  }

  out
}

# Browne's (2001) version of Yates (1984) geomin criterion
#
# we use the exp/log trick as in Bernaard & Jennrich (2005, p. 687)
lav_mat_rotate_geomin <- function(mm_lambda = NULL, geomin_epsilon = 0.01,
                                     ..., grad = FALSE) {
  n_col <- ncol(mm_lambda)

  l2 <- mm_lambda * mm_lambda
  l2 <- l2 + geomin_epsilon

  if (geomin_epsilon < sqrt(.Machine$double.eps)) {
    # Yates's original formula
    tmp <- apply(l2, 1, prod)^(1 / n_col)
  } else {
    tmp <- exp(rowSums(log(l2)) / n_col)
  }

  out <- sum(tmp)

  if (grad) {
    attr(out, "grad") <- (2 / n_col) * mm_lambda / l2 * tmp
  }

  out
}

# simple entropy
# seems to only work for orthogonal rotation
lav_mat_rotate_entropy <- function(mm_lambda = NULL, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda

  # handle zero elements -> replace by '1', so log(1) == 0
  l2[l2 == 0] <- 1

  out <- -1 * sum(l2 * log(l2)) / 2

  if (grad) {
    attr(out, "grad") <- -mm_lambda * log(l2) - mm_lambda
  }

  out
}

# McCammon's (1966) Minimum Entropy Criterion
#
# for p-vector x, where x > 0 and sum(x) = 1, we have
# - entropy(x) == 0, if there is only one 1, and all zeroes
# - entropy(x) == max == log(p) if all elements are 1/p
# - entropy(x) is similar as complexity(x), but also measure of equality
#   of elements of x
#
# works only ok with orthogonal rotation!

lav_mat_rotate_mccammon <- function(mm_lambda = NULL, ..., grad = FALSE) {
  n_col <- ncol(mm_lambda)
  n_row <- nrow(mm_lambda)
  l2 <- mm_lambda * mm_lambda

  # entropy function (Browne, 2001, eq 9)
  f_entropy <- function(x) {
    -1 * sum(ifelse(x > 0, x * log(x), 0))
  }

  # sums of rows/columns/all
  sum_j <- colSums(l2)
  sum_1 <- sum(l2)


  q1 <- f_entropy(t(l2) / sum_j) # encouraging columns with few large,
  # and many small elements
  q2 <- f_entropy(sum_j / sum_1) # encouraging equal column sums

  # minimize
  out <- log(q1) - log(q2)

  if (grad) { # See Bernaards and Jennrich 2005 page 685+686
    h_1 <- -(log(t(t(l2) / sum_j)) + 1)
    g1 <- t(t(h_1) / sum_j - rowSums(t(l2 * h_1) / (sum_j * sum_j)))

    h <- -(log(sum_j / sum_1) + 1)
    alpha <- as.numeric(h %*% sum_j) / (sum_1 * sum_1) # paper divides by
    # sum.., not sum..^2??
    g2 <- matrix(h / sum_1 - alpha, n_row, n_col, byrow = TRUE)

    attr(out, "grad") <- 2 * mm_lambda * (g1 / q1 - g2 / q2)
  }

  out
}


# Infomax
# McKeon (1968, unpublished) and Browne (2001)
# Treat mm_lambda^2 as a contingency table, and use simplicity function based
# on tests for association; most effective was LRT for association
# (see Agresti, 1990, eq 3.13) which is maximized for max simplicity
#
# McKeon: criterion may be regarded as a measure of information about row
# categories conveyed by column categories (and vice versa); hence infomax

# - favors perfect cluster
# - discourages general factor
# - both for orthogonal and oblique rotation

#
# Note: typo in Browne (2001), see last  paragraph of Bernaards and
# Jennrich (2005) page 684
lav_mat_rotate_infomax <- function(mm_lambda = NULL, ..., grad = FALSE) {
  n_col <- ncol(mm_lambda)
  n_row <- nrow(mm_lambda)
  l2 <- mm_lambda * mm_lambda

  # entropy function (Browne, 2001, eq 9)
  f_entropy <- function(x) {
    -1 * sum(ifelse(x > 0, x * log(x), 0))
  }

  # sums of rows/columns/all
  sumi <- rowSums(l2)
  sum_j <- colSums(l2)
  sum_1 <- sum(l2)

  q1 <- f_entropy(l2 / sum_1) # Bernaards & Jennrich version!! (Browne
  # divides by sum.j, like in McCammon)
  q2 <- f_entropy(sum_j / sum_1)
  q3 <- f_entropy(sumi / sum_1)

  # minimize
  out <- log(n_col) + q1 - q2 - q3

  if (grad) {
    h <- -(log(l2 / sum_1) + 1)
    alpha <- sum(l2 * h) / (sum_1 * sum_1)
    g1 <- h / sum_1 - alpha

    hj <- -(log(sum_j / sum_1) + 1)
    alphaj <- as.numeric(hj %*% sum_j) / (sum_1 * sum_1)
    g2 <- matrix(hj, n_row, n_col, byrow = TRUE) / sum_1 - alphaj

    hi <- -(log(sumi / sum_1) + 1)
    alphai <- as.numeric(sumi %*% hi) / (sum_1 * sum_1)
    g3 <- matrix(hi, n_row, n_col) / sum_1 - alphai

    attr(out, "grad") <- 2 * mm_lambda * (g1 - g2 - g3)
  }

  out
}

# oblimax
# Harman, 1976; Saunders, 1961
#
# for orthogonal rotation, oblimax is equivalent to quartimax
lav_mat_rotate_oblimax <- function(mm_lambda = NULL, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda

  # minimize version
  out <- -log(sum(l2 * l2)) + 2 * log(sum(l2))

  if (grad) {
    attr(out, "grad") <- (-4 * l2 * mm_lambda / (sum(l2 * l2))
      + 4 * mm_lambda / (sum(l2)))
  }

  out
}

# Bentler's Invariant Pattern Simplicity
# Bentler (1977)
#
#
lav_mat_rotate_bentler <- function(mm_lambda = NULL, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda

  l2t_l2 <- crossprod(l2)
  l2t_l2_inv <- lav_mat_sym_inverse(s = l2t_l2, logdet = TRUE)
  l2t_l2_logdet <- attr(l2t_l2_inv, "logdet")

  diag_1 <- diag(l2t_l2)
  diag_inv <- diag(1 / diag_1)
  diag_logdet <- sum(log(diag_1)) # add small constant?

  # minimize version
  out <- -(l2t_l2_logdet - diag_logdet) / 4

  if (grad) {
    attr(out, "grad") <- -mm_lambda * (l2 %*% (l2t_l2_inv - diag_inv))
  }

  out
}


# The Tandem criteria
# Comrey (1967)
#
# only for sequential use:
# - tandem1 is used to determine the number of factors
#   (it removes the minor factors)
# - tandem2 is used for final rotation
#
lav_mat_rotate_tandem1 <- function(mm_lambda, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda
  ll <- tcrossprod(mm_lambda)
  ll2 <- ll * ll

  # minimize version
  out <- -1 * sum(l2 * (ll2 %*% l2))

  if (grad) {
    tmp1 <- 4 * mm_lambda * (ll2 %*% l2)
    tmp2 <- 4 * (ll * (l2 %*% t(l2))) %*% mm_lambda
    attr(out, "grad") <- -tmp1 - tmp2
  }

  out
}

lav_mat_rotate_tandem2 <- function(mm_lambda, ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda
  ll <- tcrossprod(mm_lambda)
  ll2 <- ll * ll

  # minimize version
  out <- sum(l2 * ((1 - ll2) %*% l2))

  if (grad) {
    tmp1 <- 4 * mm_lambda * ((1 - ll2) %*% l2)
    tmp2 <- 4 * (ll * tcrossprod(l2, l2)) %*% mm_lambda
    attr(out, "grad") <- tmp1 - tmp2
  }

  out
}


# simplimax
# Kiers (1994)
#
# oblique rotation method
# designed to rotate so that a given number 'k' of small loadings are
# as close to zero as possible
#
# may be viewed as partially specified target rotation with
# dynamically chosen weights
#
lav_mat_rotate_simplimax <- function(mm_lambda = NULL, k = nrow(mm_lambda),
                                        ..., grad = FALSE) {
  l2 <- mm_lambda * mm_lambda

  # 'k' smallest element of L2
  small_element <- sort(l2)[k]

  # which elements are smaller than (or equal than) 'small.element'?
  id <- sign(l2 <= small_element)

  # minimize version
  out <- sum(l2 * id)

  if (grad) {
    attr(out, "grad") <- 2 * id * mm_lambda
  }

  out
}


# target rotation
# Harman, 1976
#
# mm_lambda is rotated toward a specified target matrix 'target'
#
# Note: 'target' must be fully specified; if there are any NAs
#        use lav_mat_rotate_pst() instead
#
lav_mat_rotate_target <- function(mm_lambda = NULL, target = NULL,
                                     ..., grad = FALSE) {
  # squared difference
  diff_1 <- mm_lambda - target
  diff2 <- diff_1 * diff_1

  out <- sum(diff2, na.rm = TRUE)

  if (grad) {
    tmp <- 2 * diff_1
    # change NAs to zero
    tmp[is.na(tmp)] <- 0
    attr(out, "grad") <- tmp
  }

  out
}

# partially specified target rotation
#
# Browne 1972a, 1972b
#
# a pre-specified weight matrix W with ones/zeroes determines
# which elements of (mm_lambda - target) are used by the rotation criterion
#
# if 'target' contains NAs, they should correspond to '0' values in the
# target_mask matrix
#
lav_mat_rotate_pst <- function(mm_lambda = NULL, target = NULL,
                                  target_mask = NULL, ..., grad = FALSE) {
  # mask target+mm_lambda
  target <- target_mask * target
  mm_lambda <- target_mask * mm_lambda

  # squared difference
  diff_1 <- mm_lambda - target
  diff2 <- diff_1 * diff_1

  # minimize
  out <- sum(diff2, na.rm = TRUE)

  if (grad) {
    tmp <- 2 * diff_1
    # change NAs to zero
    tmp[is.na(tmp)] <- 0
    attr(out, "grad") <- tmp
  }

  out
}


# bi-quartimin
#
# Jennrich & Bentler 2011
#
lav_mat_rotate_biquartimin <- function(mm_lambda, ..., grad = FALSE) {
  # see Matlab code page 549
  stopifnot(ncol(mm_lambda) > 1L)

  # remove first column
  lambda_group <- mm_lambda[, -1, drop = FALSE]

  # apply quartimin on the 'group' part
  out <- lav_mat_rotate_quartimin(lambda_group, ..., grad = grad)

  if (grad) {
    tmp <- attr(out, "grad")
    attr(out, "grad") <- cbind(0, tmp)
  }

  out
}


# bi-geomin
#
# Jennrich & Bentler 2012
#
lav_mat_rotate_bigeomin <- function(mm_lambda, geomin_epsilon = 0.01, ...,
                                       grad = FALSE) {
  stopifnot(ncol(mm_lambda) > 1L)

  # remove first column
  lambda_group <- mm_lambda[, -1, drop = FALSE]

  # apply geomin on the 'group' part
  out <- lav_mat_rotate_geomin(lambda_group,
    geomin_epsilon = geomin_epsilon, ...,
    grad = grad
  )

  if (grad) {
    tmp <- attr(out, "grad")
    attr(out, "grad") <- cbind(0, tmp)
  }

  out
}

# multiple group rotation + agreement
lav_mat_rotate_mg_agreement <- function(lambda_list, method_fname = "geomin",
                                   method_args = list(),
                                   mg_agreement_method = "procrustes",
                                   w = 0.5, scale = TRUE) {
  ngroups <- length(lambda_list)

  # 0. rescale, so all columns have the same (average) sum-of-squares
  if (scale) {
    ss_list <- lapply(lambda_list, function(x) colSums(x * x))
    ss_ave <- Reduce("+", ss_list) / length(ss_list)
    for (g in seq_len(ngroups)) {
      lambda_list[[g]] <- t(t(lambda_list[[g]]) /
        sqrt(ss_list[[g]]) * sqrt(ss_ave))
    }
  }

  # 1. simple structure crit for each group
  q_group <- numeric(ngroups)
  for (g in seq_len(ngroups)) {
    q_group[g] <- do.call(method_fname, c(
      list(mm_lambda = lambda_list[[g]]),
      method_args, list(grad = FALSE)
    ))
  }
  # FIXME: should we 'weight' the groups, based on sample size?
  q_mg <- sum(q_group)

  # 2. agreement (generalized procrustes) across all groups
  if (mg_agreement_method == "procrustes") {
    a_mg <- 0
    pairs <- utils::combn(seq_len(ngroups), 2L)
    for (p in seq_len(ncol(pairs))) {
      g1 <- pairs[1L, p]
      g2 <- pairs[2L, p]
      diff <- (lambda_list[[g1]] - lambda_list[[g2]])
      diff2 <- diff * diff
      # FIXME: should we 'weight' the groups, based on sample size?
      a_mg <- a_mg + sum(diff2)
    }
  } else {
    lav_msg_stop(gettext(
      "only mg.agreement.method procrustes is supported for now"))
  }

  out <- (w * a_mg) + ((1 - w) * q_mg)

  attr(out, "Q.mg") <- q_mg
  attr(out, "A.mg") <- a_mg

  out
}

# gradient check
ilav_mat_rotate_grad_test <- function(crit = NULL, ...,
                                         mm_lambda = NULL,
                                         n_row = 20L, n_col = 5L) {
  # test matrix
  if (is.null(mm_lambda)) {
    mm_lambda <- matrix(rnorm(n_row * n_col), n_row, n_col)
  }

  ff <- function(x, ...) {
    lambda <- matrix(x, n_row, n_col)
    crit(lambda, ..., grad = FALSE)
  }

  gq1 <- matrix(
    numDeriv::grad(func = ff, x = as.vector(mm_lambda), ...),
    n_row, n_col
  )
  gq2 <- attr(crit(mm_lambda, ..., grad = TRUE), "grad")

  if (lav_verbose()) {
    print(list(mm_lambda = mm_lambda, GQ1 = gq1, GQ2 = gq2))
  }

  all.equal(gq1, gq2, tolerance = 1e-07)
}

ilav_mat_rotate_grad_test_all <- function() {
  # Orthomax family with various values for gamma
  for (gamma in seq(0, 1, 0.2)) {
    check <- ilav_mat_rotate_grad_test(
      crit = lav_mat_rotate_orthomax,
      gamma = gamma
    )
    if (is.logical(check) && check) {
      cat(
        "orthomax + gamma = ", sprintf("%3.1f", gamma),
        ": OK\n"
      )
    } else {
      cat(
        "orthomax + gamma = ", sprintf("%3.1f", gamma),
        ": FAILED\n"
      )
    }
  }

  # Crawford-Ferguson with various values for gamma
  for (gamma in seq(0, 1, 0.2)) {
    check <- ilav_mat_rotate_grad_test(
      crit = lav_mat_rotate_cf,
      gamma = gamma
    )
    if (is.logical(check) && check) {
      cat(
        "Crawford-Ferguson + gamma = ", sprintf("%3.1f", gamma),
        ": OK\n"
      )
    } else {
      cat(
        "Crawford-Ferguson + gamma = ", sprintf("%3.1f", gamma),
        ": FAILED\n"
      )
    }
  }

  # Oblimin family with various values for gamma
  for (gamma in seq(0, 1, 0.2)) {
    check <- ilav_mat_rotate_grad_test(
      crit = lav_mat_rotate_oblimin,
      gamma = gamma
    )
    if (is.logical(check) && check) {
      cat(
        "Oblimin + gamma = ", sprintf("%3.1f", gamma),
        ": OK\n"
      )
    } else {
      cat(
        "Oblimin + gamma = ", sprintf("%3.1f", gamma),
        ": FAILED\n"
      )
    }
  }

  # quartimax
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_quartimax)
  if (is.logical(check) && check) {
    cat("quartimax: OK\n")
  } else {
    cat("quartimax: FAILED\n")
  }

  # varimax
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_varimax)
  if (is.logical(check) && check) {
    cat("varimax: OK\n")
  } else {
    cat("varimax: FAILED\n")
  }

  # quartimin
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_quartimin)
  if (is.logical(check) && check) {
    cat("quartimin: OK\n")
  } else {
    cat("quartimin: FAILED\n")
  }

  # geomin
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_geomin)
  if (is.logical(check) && check) {
    cat("geomin: OK\n")
  } else {
    cat("geomin: FAILED\n")
  }

  # simple entropy
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_entropy)
  if (is.logical(check) && check) {
    cat("entropy: OK\n")
  } else {
    cat("entropy: FAILED\n")
  }

  # McCammon entropy criterion
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_mccammon)
  if (is.logical(check) && check) {
    cat("McCammon: OK\n")
  } else {
    cat("McCammon: FAILED\n")
  }

  # infomax
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_infomax)
  if (is.logical(check) && check) {
    cat("infomax: OK\n")
  } else {
    cat("infomax: FAILED\n")
  }

  # oblimax
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_oblimax)
  if (is.logical(check) && check) {
    cat("oblimax: OK\n")
  } else {
    cat("oblimax: FAILED\n")
  }

  # bentler
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_bentler)
  if (is.logical(check) && check) {
    cat("bentler: OK\n")
  } else {
    cat("bentler: FAILED\n")
  }

  # simplimax
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_simplimax)
  if (is.logical(check) && check) {
    cat("simplimax: OK\n")
  } else {
    cat("simplimax: FAILED\n")
  }

  # tandem1
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_tandem1)
  if (is.logical(check) && check) {
    cat("tandem1: OK\n")
  } else {
    cat("tandem1: FAILED\n")
  }

  # tandem2
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_tandem2)
  if (is.logical(check) && check) {
    cat("tandem2: OK\n")
  } else {
    cat("tandem2: FAILED\n")
  }

  # bi-quartimin
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_biquartimin)
  if (is.logical(check) && check) {
    cat("biquartimin: OK\n")
  } else {
    cat("biquartimin: FAILED\n")
  }

  # bi-quartimin
  check <- ilav_mat_rotate_grad_test(crit = lav_mat_rotate_bigeomin)
  if (is.logical(check) && check) {
    cat("bigeomin: OK\n")
  } else {
    cat("bigeomin: FAILED\n")
  }
}

# chi-square test statistic:
# comparing the current model versus the saturated/unrestricted model
# TDJ 9 April 2024: Add a (hidden) function to update the @test slot
#                   when the user provides a custom h1 model.  Called by
#                   lav_object_summary and lav_fit(_check_baseline)

lavTest <- function(lavobject, test = "standard",               # nolint
                    scaled_test = "standard",
                    output = "list", drop_list_single = TRUE,
                  ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # check object
  lavobject <- lav_object_check_version(lavobject)

  # check output
  output_valid <- c("list", "text")
  if (!any(output == output_valid)) {
      lav_msg_stop(gettextf(
        "%1$s argument must be either %2$s",
        "output", lav_msg_view(output_valid, "or")
      ))
  }

  # extract 'test' slot
  test_1 <- lavobject@test

  # backwards compatibility:
  # if (length(TEST) > 0L && is.null(names(TEST))) {
  #   names(TEST) <- sapply(TEST, "[[", "test")
  # }

  # which test?
  if (!missing(test)) {
    # check 'test'
    if (!is.character(test)) {
      lav_msg_stop(
        gettextf("%s should be a character string.", "test"))
    } else {
      test <- lav_test_rename(test, check = TRUE)
      # always include "standard"
      test <- unique(c("standard", test))
    }

    # check scaled_test
    if (!missing(scaled_test)) {
      if (!is.character(scaled_test)) {
        lav_msg_stop(
          gettextf("%s should be a character string.", "scaled_test"))
      } else {
        scaled_test <- lav_test_rename(scaled_test, check = TRUE)
      }

      # merge
      test <- unique(c(test, scaled_test))
    }

    # "standard" must always be first
    standard_idx <- which(test == "standard")
    if (length(standard_idx) > 0L && standard_idx != 1L) {
      test <- c("standard", test[-standard_idx])
    }

    if (test[1] == "none") {
      return(list())
    } else if (any(test %in% c("bootstrap", "bollen.stine"))) {
      lav_msg_stop(gettext(
      "please use lavBootstrap() to obtain a bootstrap based test statistic."
      ))
    }

    # check if we already have it:
    if (all(test %in% names(test_1))) {
      info_attr <- attr(test_1, "info")
      test_idx <- which(names(test_1) %in% test)
      test_1 <- test_1[test_idx]
      attr(test_1, "info") <- info_attr
    } else {
      # redo ALL of them, even if already have some in TEST
      # later, we will allow to also change the options (like information)
      # and this should be reflected in the 'info' attribute

      # fill-in test in Options slot
      lavobject@Options$test <- test

      # fill-in scaled_test in Options slot
      lavobject@Options$scaled.test <- scaled_test

      # get requested test statistics
      test_1 <- lav_model_test(lavobject = lavobject)
    }
  }

  if (output == "list") {
    # remove 'info' attribute
    attr(test_1, "info") <- NULL

    # select only those that were requested (eg remove standard)
    test_idx <- which(names(test_1) %in% test)
    test_1 <- test_1[test_idx]

    # if only 1 test, drop outer list
    if (length(test_1) == 1L && drop_list_single) {
      test_1 <- test_1[[1]]
    }

    return(test_1)
  } else {
    lav_test_print(test_1)
  }

  invisible(test_1)
}

# synonyms for the test statistics: canonical name -> accepted aliases
lav_test_synonyms <- list(
  standard = c("chisq", "chi", "chi-square", "chi.square"),
  satorra.bentler = c(
    "satorra", "sb", "satorra-bentler",
    "m.adjusted", "m", "mean.adjusted", "mean-adjusted"
  ),
  yuan.bentler = c("yuan", "yb", "yuan-bentler"),
  yuan.bentler.mplus = c("yuan-bentler.mplus", "yuan-bentler-mplus"),
  yuan.chan = c("yuan-chan", "yuan_chan", "yuanchan", "yc"),
  mean.var.adjusted = c(
    "mean-var-adjusted", "mv", "second.order",
    "satterthwaite", "mv.adjusted"
  ),
  scaled.shifted = c("mplus6", "scale.shift", "scaled-shifted"),
  mean.var.adjusted.corrected = c(
    "mean-var-adjusted-corrected",
    "mv.adjusted.corrected", "mvc", "hayakawa"
  ),
  scaled.shifted.corrected = c("scaled-shifted-corrected", "ssc"),
  bollen.stine = c("bootstrap", "boot", "bollen-stine"),
  browne.residual.adf = c(
    "browne", "residual", "residuals", "browne.residual",
    "browne.residuals", "residual-based", "residual.based",
    "browne.residuals.adf"
  ),
  browne.residual.nt = c(
    "browne.residuals.nt",
    "browne.nt.residuals", "browne.nt.residual"
  ),
  browne.residual.adf.model = c(
    "browne.residuals.adf.model",
    "browne.residual.model.adf", "browne.residuals.model.adf"
  ),
  browne.residual.nt.model = c(
    "browne.residuals.nt.model",
    "rls", "browne.rls", "nt.rls", "nt-rls", "ntrls"
  )
)

# allow for 'flexible' names for the test statistics
# 0.6-13: if multiple names, order them in such a way
#         that the 'scaled' variants appear after the others
lav_test_rename <- function(test, check = FALSE) {
  test <- tolower(test)

  for (canonical in names(lav_test_synonyms)) {
    target_idx <- which(test %in% lav_test_synonyms[[canonical]])
    if (length(target_idx) > 0L) {
      test[target_idx] <- canonical
    }
  }

  if (length(target_idx <- which(vapply(
    test, lav_test_fmg_is_preset, logical(1L)
  ))) > 0L) {
    test[target_idx] <- vapply(
      test[target_idx],
      lav_test_fmg_resolve_preset,
      character(1L)
    )
  }


  # check?
  if (check) {
    fmg_idx <- which(vapply(test, lav_test_fmg_is_fmg, logical(1L)))

    # report unknown values
    bad_idx <- which(!test %in% c(
      "standard", "none", "default",
      "satorra.bentler",
      "yuan.bentler",
      "yuan.bentler.mplus",
      "yuan.chan",
      "mean.adjusted",
      "mean.var.adjusted",
      "mean.var.adjusted.corrected",
      "scaled.shifted",
      "scaled.shifted.corrected",
      "bollen.stine",
      "browne.residual.nt",
      "browne.residual.nt.model",
      "browne.residual.adf",
      "browne.residual.adf.model"
    ))
    bad_idx <- setdiff(bad_idx, fmg_idx)
    if (length(bad_idx) > 0L) {
      lav_msg_stop(sprintf(
        ngettext(
          length(test[bad_idx]),
          "invalid value in %1$s argument: %2$s.",
          "invalid values in %1$s argument: %2$s."
        ),
        "test", lav_msg_view(test[bad_idx], log_sep = "none")
      ))
    }

    # if 'default' is included, length(test) must be 1
    if (length(test) > 1L && any("default" == test)) {
      lav_msg_stop(
        gettextf("if test= argument contains \"%s\", it cannot contain
                 additional elements", "default"))
    }

    # if 'none' is included, length(test) must be 1
    if (length(test) > 1L && any("none" == test)) {
      lav_msg_stop(
        gettextf("if test= argument contains \"%s\" it cannot contain
                 additional elements", "none"))
    }
  }

  # reorder: first nonscaled, then scaled
  nonscaled_idx <- which(test %in% c(
    "standard", "none", "default",
    "bollen.stine",
    "browne.residual.nt",
    "browne.residual.nt.model",
    "browne.residual.adf",
    "browne.residual.adf.model"
  ))
  scaled_idx <- which(test %in% c(
    "satorra.bentler",
    "yuan.bentler",
    "yuan.bentler.mplus",
    "yuan.chan",
    "mean.adjusted",
    "mean.var.adjusted",
    "mean.var.adjusted.corrected",
    "scaled.shifted",
    "scaled.shifted.corrected"
  ))
  fmg_idx <- which(vapply(test, lav_test_fmg_is_fmg, logical(1L)))
  test <- c(test[nonscaled_idx], test[scaled_idx], test[fmg_idx])

  test
}

lav_model_test <- function(lavobject = NULL,
                           lavmodel = NULL,
                           lavpartable = NULL,
                           lavsamplestats = NULL,
                           lavimplied = NULL,
                           lavh1 = list(),
                           lavoptions = NULL,
                           x = NULL,
                           vcov_1 = NULL,
                           lavcache = NULL,
                           lavdata = NULL,
                           lavloglik = NULL,
                           test_ugamma_eigvals = FALSE) {
  # lavobject?
  if (!is.null(lavobject)) {
    lavmodel <- lavobject@Model
    lavpartable <- lav_pt_set_cache(lavobject@ParTable, lavobject@pta)
    lavsamplestats <- lavobject@SampleStats
    lavimplied <- lavobject@implied
    lavh1 <- lavobject@h1
    lavoptions <- lavobject@Options
    x <- lavobject@optim$x
    fx <- lavobject@optim[["fx"]]
    fx_group <- lavobject@optim[["fx.group"]]
    attr(fx, "fx.group") <- fx_group
    attr(x, "fx") <- fx
    vcov_1 <- lavobject@vcov$vcov
    lavcache <- lavobject@Cache
    lavdata <- lavobject@Data
    lavloglik <- lavobject@loglik
  }

  # backwards compatibility
  if (is.null(lavoptions$scaled.test)) {
    lavoptions$scaled.test <- "standard"
  }

  test <- lavoptions$test

  # "yuan.chan" is a SAM-only (sam.method = "global") rescaled test statistic;
  # it is computed in lav_sam_global_test(), not here. If it ever reaches the
  # core test machinery (eg sem(test = "yuan.chan")), drop it with a note.
  if (any(test == "yuan.chan")) {
    lav_msg_warn(gettext(
      "test = \"yuan.chan\" is only available for sam(sam.method = \"global\");
       it will be ignored here."))
    test <- test[test != "yuan.chan"]
    if (length(test) == 0L) {
      test <- "standard"
    }
    lavoptions$test <- test
  }

  test_1 <- list()

  # degrees of freedom, handling equality constraints (note: we ignore
  # inequality constraints, active or not! we use the rank of con.jac,
  # even if the constraints are nonlinear)
  df <- lav_test_df(lavpartable = lavpartable, lavmodel = lavmodel)

  # shortcut: return empty list if one of the conditions below is true:
  # - test == "none"
  # - df < 0
  # - estimator == "MML"
  if (test[1] == "none" || df < 0L || lavoptions$estimator == "MML") {
    test_1[[1]] <- list(
      test = test[1],
      stat = as.numeric(NA),
      stat.group = as.numeric(NA),
      df = df,
      refdistr = "unknown",
      pvalue = as.numeric(NA)
    )

    if (length(test) > 1L) {
      test_1[[2]] <- list(
        test = test[2],
        stat = as.numeric(NA),
        stat.group = as.numeric(NA),
        df = df,
        refdistr = "unknown",
        pvalue = as.numeric(NA)
      )
    }

    attr(test_1, "info") <- lav_test_info_attr(
      lavdata = lavdata, lavoptions = lavoptions
    )

    return(test_1)
  }


  ######################
  ## TEST == STANDARD ##
  ######################

  # get chisq value, per group

  # PML
  if (lavoptions$estimator == "PML" && test[1] != "none") {
    # attention!
    # if the thresholds are saturated (ie, nuisance parameters)
    # we should use the lav_pml_plrt() function.
    #
    # BUT, if the thresholds are structured (eg equality constraints)
    # then we MUST use the lav_pml_plrt2() function.
    #
    # This was not done automatically < 0.6-6
    #


    thresholds_structured <- FALSE
    # check
    th_idx <- which(lavpartable$op == "|")
    if (any(lavpartable$free[th_idx] == 0L)) {
      thresholds_structured <- TRUE
    }

    eq_idx <- which(lavpartable$op == "==")
    if (length(eq_idx) > 0L) {
      th_labels <- lavpartable$plabel[th_idx]
      eq_labels <- unique(c(
        lavpartable$lhs[eq_idx],
        lavpartable$rhs[eq_idx]
      ))
      if (any(th_labels %in% eq_labels)) {
        thresholds_structured <- TRUE
      }
    }

    # switch between lav_pml_plrt() and lav_pml_plrt2()
    if (thresholds_structured) {
      pml_plrt <- lav_pml_plrt2
    } else {
      pml_plrt <- lav_pml_plrt
    }

    pml <- pml_plrt(
      lavobject = NULL,
      lavmodel = lavmodel,
      lavdata = lavdata,
      lavoptions = lavoptions,
      x = x,
      vcov_1 = vcov_1,
      lavcache = lavcache,
      lavsamplestats = lavsamplestats,
      lavpartable = lavpartable
    )
    # get chi.group from PML, since we compare to `unrestricted' model,
    # NOT observed data
    chisq_group <- pml$PLRTH0Sat.group

    # twolevel (ML: LRT versus the h1 model; the least-squares estimators
    # fall through to the fx-based statistic below)
  } else if (lavdata@nlevels > 1L &&
    !(lavoptions$estimator %in% c("WLS", "DWLS", "ULS"))) {
    if (length(lavh1) > 0L) {
      # LRT
      chisq_group <- -2 * (lavloglik$loglik.group - lavh1$logl$loglik.group)
    } else {
      chisq_group <- rep(as.numeric(NA), lavdata@ngroups)
    }
  } else if (lavdata@missing == "ml" && any(lavdata@Mp[[1]]$coverage == 0)) {
    if (length(lavh1) > 0L) {
      chisq_group <- -2 * (lavloglik$loglik.group - lavh1$logl$loglik.group)
    } else {
      chisq_group <- rep(as.numeric(NA), lavdata@ngroups)
    }
  } else if (lavoptions$estimator %in% c("IV")) {
    # no 'standard' chi-square statistic
    chisq_group <- rep(as.numeric(NA), lavdata@ngroups)
  } else {
    # get fx.group
    fx <- attr(x, "fx")
    fx_group <- attr(fx, "fx.group")

    # always compute `standard' test statistic
    ## FIXME: the NFAC is now implicit in the computation of fx...
    nfac <- 2 * unlist(lavsamplestats@nobs)
    if (lavoptions$estimator == "ML" && lavoptions$likelihood == "wishart") {
      # first divide by two
      nfac <- nfac / 2
      nfac <- nfac - 1
      nfac <- nfac * 2
    } else if (lavoptions$estimator == "DLS") {
      nfac <- nfac / 2
      nfac <- nfac - 1
      nfac <- nfac * 2
    }

    chisq_group <- fx_group * nfac
  }

  # check for negative values
  chisq_group[chisq_group < 0] <- 0.0

  # global test statistic
  chisq <- sum(chisq_group)

  # reference distribution: always chi-square, except for the
  # non-robust version of ULS and PML
  if (lavoptions$estimator %in% c("ULS", "DWLS", "PML")) {
    refdistr <- "unknown"
    pvalue <- as.numeric(NA)
  } else {
    refdistr <- "chisq"

    # pvalue  ### FIXME: what if df=0? NA? or 1? or 0?
    # this is not trivial, since
    # 1 - pchisq(0, df=0) = 1
    # but
    # 1 - pchisq(0.00000000001, df=0) = 0
    # and
    # 1 - pchisq(0, df=0, ncp=0) = 0
    #
    # This is due to different definitions of limits (from the left,
    # or from the right)
    #
    # From 0.5-17 onwards, we will use NA if df=0, to be consistent
    if (df == 0) {
      pvalue <- as.numeric(NA)
    } else {
      pvalue <- 1 - pchisq(chisq, df)
    }
  }

  test_1[["standard"]] <- list(
    test = "standard",
    stat = chisq,
    stat.group = chisq_group,
    df = df,
    refdistr = refdistr,
    pvalue = pvalue
  )

  if (length(test) == 1L && test == "standard") {
    # we are done
    attr(test_1, "info") <- lav_test_info_attr(
      lavdata = lavdata, lavoptions = lavoptions
    )
    return(test_1)
  } else {
    # strip 'standard' from test list
    if (length(test) > 1L) {
      standard_idx <- which(test == "standard")
      if (length(standard_idx) > 0L) {
        test <- test[-standard_idx]
      }
    }
  }

  ######################
  ## additional tests ## # new in 0.6-5
  ######################

  # first: sort tests, so that the unscaled tests come first, and then the
  # scaled (otherwise, we don't find the base test)
  unscaled_idx <- which(test %in% c(
      "browne.residual.adf",
      "browne.residual.adf.model",
      "browne.residual.nt",
      "browne.residual.nt.model"
    ))
  if (length(unscaled_idx) > 0L) {
    test <- c(test[unscaled_idx], test[-unscaled_idx])
  }

  for (this_test in test) {
    if (lav_test_fmg_is_fmg(this_test)) {
      parsed <- lav_test_fmg_parse(this_test)
      chisq <- lav_test_fmg_resolve_chisq(parsed, lavoptions = lavoptions)
      unscaled_test <- test_1[[1]]
      if (chisq == "rls") {
        unscaled_test <- lav_test_fmg_browne_nt_model(
          lavobject = lavobject,
          lavdata = lavdata,
          lavsamplestats = lavsamplestats,
          lavmodel = lavmodel,
          lavpartable = lavpartable,
          lavoptions = lavoptions,
          lavh1 = lavh1,
          lavimplied = lavimplied
        )
        test_1[[unscaled_test$test]] <- unscaled_test
      }

      test_1[[this_test]] <- lav_test_fmg(
        lavobject = lavobject,
        lavsamplestats = lavsamplestats,
        lavmodel = lavmodel,
        lavdata = lavdata,
        lavoptions = lavoptions,
        lavimplied = lavimplied,
        test_unscaled = test_1[[1]],
        test_chisq = unscaled_test,
        e_inv = attr(vcov_1, "E.inv"),
        delta = attr(vcov_1, "Delta"),
        wls_v = attr(vcov_1, "WLS.V"),
        gamma = attr(vcov_1, "Gamma"),
        test = this_test
      )
    } else if (lavoptions$estimator == "PML") {
      if (this_test == "mean.var.adjusted") {
        label <- "mean+var adjusted correction (PML)"
        test_1[[this_test]] <-
          list(
            test = this_test,
            stat = pml$stat,
            stat.group = test_1[[1]]$stat.group * pml$scaling.factor,
            df = pml$df,
            pvalue = pml$p.value,
            scaling.factor = 1 / pml$scaling.factor,
            scaled.test.stat = test_1[[1]]$stat,
            scaled.test = test_1[[1]]$test,
            label = label,
            shift.parameter = as.numeric(NA),
            trace.UGamma = as.numeric(NA),
            trace.UGamma4 = as.numeric(NA),
            trace.UGamma2 = as.numeric(NA),
            UGamma.eigenvalues = as.numeric(NA)
          )
      } else {
        lav_msg_warn(gettextf("test option %s not available for estimator PML",
                              this_test))
      }
    } else if (this_test %in% c(
      "browne.residual.adf",
      "browne.residual.adf.model",
      "browne.residual.nt",
      "browne.residual.nt.model"
    )) {
      adf <- TRUE
      if (this_test %in% c(
        "browne.residual.nt",
        "browne.residual.nt.model"
      )) {
        adf <- FALSE
      }
      model_based <- FALSE
      if (this_test %in% c(
        "browne.residual.adf.model",
        "browne.residual.nt.model"
      )) {
        model_based <- TRUE
      }

      out <- lav_test_browne(
        lavobject = NULL,
        lavdata = lavdata,
        lavsamplestats = lavsamplestats,
        lavmodel = lavmodel,
        lavpartable = lavpartable,
        lavoptions = lavoptions,
        lavh1 = lavh1,
        lavimplied = lavimplied,
        adf = adf,
        model_based = model_based
      )
      test_1[[this_test]] <- out
    } else if (this_test %in% c(
      "satorra.bentler",
      "mean.var.adjusted",
      "mean.var.adjusted.corrected",
      "scaled.shifted",
      "scaled.shifted.corrected"
    )) {
      # which test statistic shall we scale?
      unscaled_test <- lav_test_scaled_base(
        test_1 = test_1, lavoptions = lavoptions, test = test
      )

      out <- lav_test_sb(
        lavobject = NULL,
        lavsamplestats = lavsamplestats,
        lavmodel = lavmodel,
        lavimplied = lavimplied,
        lavdata = lavdata,
        lavoptions = lavoptions,
        test_unscaled = unscaled_test,
        e_inv = attr(vcov_1, "E.inv"),
        delta = attr(vcov_1, "Delta"),
        wls_v = attr(vcov_1, "WLS.V"),
        m_gamma = attr(vcov_1, "Gamma"),
        test = this_test,
        method = "original", # since 0.6-13
        return_ugamma = FALSE
      )
      test_1[[this_test]] <- out[[this_test]]
    } else if (this_test %in% c(
      "yuan.bentler",
      "yuan.bentler.mplus"
    )) {
      # which test statistic shall we scale?
      unscaled_test <- lav_test_scaled_base(
        test_1 = test_1, lavoptions = lavoptions, test = test
      )

      out <- lav_test_yb(
        lavobject = NULL,
        lavsamplestats = lavsamplestats,
        lavmodel = lavmodel,
        lavdata = lavdata,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        lavoptions = lavoptions,
        test_unscaled = unscaled_test,
        e_inv = attr(vcov_1, "E.inv"),
        b0_group = attr(vcov_1, "B0.group"),
        test = this_test,
        mimic = lavoptions$mimic,
        # method         = "default",
        return_ugamma = FALSE
      )
      test_1[[this_test]] <- out[[this_test]]
    } else if (this_test == "bollen.stine") {
      # check if we have bootstrap lavdata
      boot_test <- attr(vcov_1, "BOOT.TEST")
      if (is.null(boot_test)) {
        if (!is.null(lavoptions$bootstrap)) {
          r <- as.integer(lavoptions$bootstrap$R)
        } else {
          r <- 1000L
        }
        boot_type <- "bollen.stine"
        boot_test <-
          lav_bootstrap_internal(
            object = NULL,
            lavmodel = lavmodel,
            lavsamplestats = lavsamplestats,
            lavpartable = lavpartable,
            lavoptions = lavoptions,
            lavdata = lavdata,
            r = r,
            type = boot_type,
            fun = "test"
          )

        # new in 0.6-12: always warn for failed and nonadmissible
        error_idx <- attr(boot_test, "error.idx")
        nfailed <- length(attr(boot_test, "error.idx")) # zero if NULL
        if (nfailed > 0L) {
          lav_msg_warn(gettextf(
            "%d bootstrap runs failed or did not converge.", nfailed
          ))
        }

        notok <- length(attr(boot_test, "nonadmissible")) # zero if NULL
        if (notok > 0L) {
          lav_msg_warn(gettextf(
            "%d bootstrap runs resulted in nonadmissible solutions.", notok
          ))
        }

        if (length(error_idx) > 0L) {
          # new in 0.6-13: we must still remove them!
          boot_test <- boot_test[-error_idx, , drop = FALSE]
          # this also drops the attributes
        }

        boot_test <- drop(boot_test)
      }

      # bootstrap p-value
      boot_larger <- sum(boot_test > chisq)
      boot_length <- length(boot_test)
      pvalue_boot <- boot_larger / boot_length

      test_1[[this_test]] <- list(
        test = this_test,
        stat = chisq,
        stat.group = chisq_group,
        df = df,
        pvalue = pvalue_boot,
        refdistr = "bootstrap",
        boot.T = boot_test,
        boot.larger = boot_larger,
        boot.length = boot_length
      )
    }
  } # additional tests

  # add additional information as an attribute, needed for independent
  # printing
  attr(test_1, "info") <- lav_test_info_attr(
    lavdata = lavdata, lavoptions = lavoptions
  )

  test_1
}

lav_update_test_custom_h1 <- function(lav_obj_h0, lav_obj_h1) {
  stopifnot(inherits(lav_obj_h0, "lavaan"))
  stopifnot(inherits(lav_obj_h1, "lavaan"))

  ## this breaks if object not nested in (df >=) h1, so check df
  stopifnot(lav_obj_h0@test[[1]]$df >= lav_obj_h1@test[[1]]$df)

  ## remove any other (potentially hidden) h1 model from BOTH objects
  lav_obj_h0@external$h1.model <- NULL
  lav_obj_h1@external$h1.model <- NULL

  ## save old @test slot as template
  ## (so the @test[[1]]$df don't change while looping over tests to update)
  new_test <- lav_obj_h0@test

  ## assemble a call to lavTestLRT()
  lrt_call_template <- list(quote(lavTestLRT), object = quote(lav_obj_h0),
                          quote(lav_obj_h1)) # in ...

  ## can only update tests available in both objects
  test_names0 <- names(lav_obj_h0@test)
  test_names1 <- names(lav_obj_h1@test)
  test_names <- intersect(test_names0, test_names1)

  ## loop over those tests
  for (tn in test_names) {
    lrt_call <- lrt_call_template
    ## conditional arguments:
    if (tn == "standard") {
      lrt_call$method <- "standard"
    } else if (tn %in% c("scaled.shifted", "mean.var.adjusted")) {
      if (lav_obj_h0@Options$estimator == "PML") {
        lrt_call$method <- "mean.var.adjusted.PLRT"
      } else {
        lrt_call$method <- "satorra.2000"
      }
      lrt_call$scaled.shifted <- tn == "scaled.shifted"
    } else if (tn %in% c("satorra.bentler",
                         "yuan.bentler", "yuan.bentler.mplus")) {
      lrt_call$test <- tn
    } else if (grepl(pattern = "browne", x = tn, fixed = TRUE)) {
      lrt_call$type <- tn
    } else {
      #TODO?
      #   - if (tn %in% c("bootstrap", "bollen.stine")) next
      #   - any other possibilities in @test?
    }

    ## get new test
    if (lav_obj_h0@test[[1]]$df == lav_obj_h1@test[[1]]$df) {
      ## suppress warning about == df
      anova_1 <- suppressWarnings(eval(as.call(lrt_call)))
    } else {
      ## maybe some other informative warning would be important to see
      anova_1 <- eval(as.call(lrt_call))
    }


    ## replace old @test[[tn]] values
    new_test[[tn]]$stat.group <- NULL # avoid wrong stats in summary() header?
    new_test[[tn]]$stat   <- anova_1["lav_obj_h0", "Chisq diff"]
    new_test[[tn]]$df     <- anova_1["lav_obj_h0", "Df diff"]
    new_test[[tn]]$pvalue <- anova_1["lav_obj_h0", "Pr(>Chisq)"]
    if (!is.null(new_test[[tn]]$scaling.factor)) {
      new_test[[tn]]$scaling.factor <- attr(anova_1, "scale")[2] #1st row is NA
    }
    if (!is.null(new_test[[tn]]$shift.parameter)) {
      new_test[[tn]]$shift.parameter <- attr(anova_1, "shift")[2] #1st row is NA
    } else {
      ## unless scaled.shifted, RMSEA is calculated from $standard$stat and
      ## df == sum($trace.UGamma).  Reverse-engineer from $scaling.factor:
      new_test[[tn]]$trace.UGamma <- new_test[[tn]]$df *
                                                   new_test[[tn]]$scaling.factor
    }
    ## should not be necessary to replace $trace.UGamma2
    ## nor to replace $scaling.factor.h0/h1
  } # end loop over tests

  ## assign updated @test slot and return
  lav_obj_h0@test <- new_test
  lav_obj_h0
}

# step 1 in SAM: fitting the measurement blocks

lav_sam_step1 <- function(cmd = "sem", mm_list = NULL, mm_args = list(),
                          fit = NULL, sam_method = "local") {
  lavoptions <- fit@Options
  lavpta <- fit@pta
  pt_1 <- fit@ParTable
  nblocks <- lavpta$nblocks
  ngroups <- lavpta$ngroups

  if (lav_verbose()) {
    cat("Fitting the measurement part:\n")
  }

  # total number of free parameters
  if (fit@Model@ceq.simple.only) {
    npar <- fit@Model@nx.unco
    pt_free <- pt_1$free
    pt_free[pt_free > 0] <- seq_len(npar)
  } else {
    npar <- fit@Model@nx.free
    pt_free <- pt_1$free
  }
  if (npar < 1L) {
    lav_msg_stop(gettext("model does not contain any free parameters"))
  }

  # do we have at least 1 'regular' (measured) latent variable?
  lv_names <- unique(unlist(fit@pta$vnames$lv.regular))
  if (length(lv_names) == 0L) {
    lav_msg_warn(gettext("structural model does not contain any (measured)
                          latent variables; consider using sem() instead"))
  }

  # check for higher-order factors (supported since 0.6-20):
  # remove the lower-order factors from lv_names
  lv_ind_names <- unique(unlist(fit@pta$vnames$lv.ind))
  if (length(lv_ind_names) > 0L) {
    lv_names <- lv_names[!lv_names %in% lv_ind_names]
  }

  # how many measurement models?
  if (!is.null(mm_list)) {
    n_mmblocks <- length(mm_list)
    # check each measurement block
    for (b in seq_len(n_mmblocks)) {
      # check if we can find all lv names in LV.names
      if (!all(unlist(mm_list[[b]]) %in% lv_names)) {
        tmp <- unlist(mm_list[[b]])
        lav_msg_stop(gettext("mm.list contains unknown latent variable(s):"),
          lav_msg_view(tmp[!tmp %in% lv_names], "none"))
      }
      # make list per block
      if (!is.list(mm_list[[b]])) {
        mm_list[[b]] <- rep(list(mm_list[[b]]), nblocks)
      } else {
        if (length(mm_list[[b]]) != nblocks) {
          lav_msg_stop(gettextf(
            "mm.list block %1$s has length %2$s but nblocks = %3$s",
            b, length(mm_list[[b]]), nblocks))
        }
      }
    }
  } else {
    mm_list_per_block <- lav_sam_get_mmlist(fit) # we get a list PER BLOCK
    if (fit@Data@nlevels > 1L) {
      # multilevel: join the within and between clusters that share
      # observed indicators into a single (two-level) measurement block
      mm_list <- lav_sam_mmlist_merge_blocks(fit, mm_list_per_block)
    } else {
      # for now, we only keep the first block
      mm_list <- mm_list_per_block[[1]]
    }
    n_mmblocks <- length(mm_list)
  }

  # adjust options for measurement models
  lavoptions_mm <- lavoptions
  # "yuan.chan" is a SAM-global test for the JOINT model only; the measurement
  # blocks use the ordinary (standard) test
  if (any(lavoptions_mm$test == "yuan.chan")) {
    lavoptions_mm$test <- "standard"
  }
  lavoptions_mm$optim.bounds <- NULL
  if (lavoptions$se %in% c("none", "bootstrap")) {
    lavoptions_mm$se <- "none"
  } else {
    # single-level clustered data: use the cluster-robust variants, so that
    # the step 1 standard errors (and Sigma.11) reflect the clustering;
    # for the Gamma (NACOV) based sandwich (se = "robust.sem"), no switch
    # is needed: the NACOV is computed with the cluster correction whenever
    # the data contains a cluster variable
    clustered_flag <- fit@Data@nlevels == 1L && length(fit@Data@cluster) > 0L
    # categorical?
    # (PML first: it is a categorical estimator, but se = "robust.sem" is
    # not valid for PML -- it needs the huber-white sandwich)
    if (lavoptions$estimator.orig == "PML") {
      lavoptions_mm$se <- "robust.huber.white"
    } else if (fit@Model@categorical) {
      lavoptions_mm$se <- "robust.sem"
    } else if (lavoptions$estimator.orig == "MLM") {
      lavoptions_mm$se <- "robust.sem"
    } else if (lavoptions$estimator.orig == "MLR") {
      lavoptions_mm$se <- if (clustered_flag) "robust.cluster"
                          else "robust.huber.white"
    } else if (clustered_flag) {
      lavoptions_mm$se <- "robust.cluster"
    } else {
      lavoptions_mm$se <- "standard" # may be overridden later
    }
  }
  # note: we keep the tests, as we need them for the summary info about MM
  lavoptions_mm$check.post <- FALSE # neg lv variances may be overridden
  lavoptions_mm$check.gradient <- FALSE # too sensitive in large model (global)
  lavoptions_mm$baseline <- FALSE
  lavoptions_mm$fit.by.level <- FALSE
  lavoptions_mm$bounds <- "wide.zerovar"

  # conditional.x: for CONTINUOUS data the measurement models are always
  # fitted unconditionally -- this does not affect the measurement parameters
  # (if the covariates act on the structural part only). For CATEGORICAL
  # data, however, the latent-response scale differs between the marginal
  # (Var(y*) = 1) and the conditional (Var(y*|x) = 1) parameterization, and
  # the marginal polychorics are misspecified when the covariates are
  # non-normal (eg binary). The measurement blocks are then fitted on the
  # same conditional scale as the joint model: conditional.x = TRUE, with
  # saturated eta ~ x regressions (added via lav_pt_subset_mm() below).
  if (fit@Model@categorical && lavoptions$conditional.x) {
    lavoptions_mm$conditional.x <- TRUE
  } else {
    lavoptions_mm$conditional.x <- FALSE
  }

  # override with user-specified mm.args
  lavoptions_mm <- modifyList(lavoptions_mm, mm_args)

  # create MM slot_options
  slot_options_mm <- lav_options_set(lavoptions_mm)

  # we assume the same number/names of lv's per group!!!
  mm_fit <- vector("list", n_mmblocks) # fitted object

  # for joint model later
  if (!lavoptions$se %in% c("none", "bootstrap")) {
    sigma_11_1 <- matrix(0, npar, npar)
    colnames(sigma_11_1) <- rownames(sigma_11_1) <-
      lav_pt_labels(fit@ParTable, type = "free")
  }
  step1_free_idx <- integer(0L)
  block_mm_idx  <- vector("list", length = n_mmblocks)
  block_ptm_idx <- vector("list", length = n_mmblocks)

  # NOTE: we should explicitly add zero-constrained LV covariances
  # to PT, and keep them zero in PTM
  add_lv_cov <- cmd != "lavaan"

  # if data.type == "moment", we need the h1 sample statistics (per group)
  if (fit@Data@data.type == "moment") {
    h1_all <- lavTech(fit, "h1", add.labels = TRUE)
  }

  # fit mm model for each measurement block
  for (mm in seq_len(n_mmblocks)) {
    if (lav_verbose()) {
      cat(
        "  block ", mm, "[",
        paste(mm_list[[mm]], collapse = " "), "]\n"
      )
    }

    # create parameter table for this measurement block only
    ptm <- lav_pt_subset_mm(
      pt_1 = pt_1,
      add_lv_cov = add_lv_cov,
      add_idx = TRUE,
      lv_names = mm_list[[mm]],
      add_exo_reg = slot_options_mm$conditional.x,
      ov_names_x = fit@pta$vnames$ov.x
    )
    mm_idx <- attr(ptm, "idx")
    attr(ptm, "idx") <- NULL
    ptm$est <- NULL
    ptm$se <- NULL
    block_mm_idx[[mm]] <- mm_idx

    # check for categorical in PTM in this mm-block
    # (a fully continuous block is also fitted with conditional.x = TRUE
    # when the joint model is categorical + conditional.x: its statistics
    # must live in the same conditional space as the joint model)
    if (!any(ptm$op == "|")) {
      slot_options_mm$categorical <- FALSE
      slot_options_mm$.categorical <- FALSE
    }

    # update slot_data for this measurement block
    # (under conditional.x, lavData keeps the exogenous covariates OUT of
    # ov.names/X -- they live in ov.names.x/eXo -- so the block ov.names
    # must be the indicators only)
    ov_names_block <- lapply(1:ngroups, function(g) {
      if (slot_options_mm$conditional.x) {
        unique(unlist(lav_pt_vnames(ptm, type = "ov.nox", group = g)))
      } else {
        unique(unlist(lav_pt_vnames(ptm, type = "ov", group = g)))
      }
    })
    slot_data_block <- lav_data_update_subset(fit@Data,
      ov_names = ov_names_block
    )
    if (!slot_options_mm$conditional.x) {
      # get rid of ov.names.x
      slot_data_block@ov.names.x <-
        lapply(seq_len(nblocks), function(x) character(0L))
      slot_data_block@eXo <-
        lapply(seq_len(nblocks), function(x) NULL)
    } else {
      # keep ALL exogenous covariates for the conditional block
      # (lav_data_update_subset() dropped them, as they are not part of
      # ov_names_block); also restore their entries in the ov table
      slot_data_block@ov.names.x <- fit@Data@ov.names.x
      slot_data_block@eXo <- fit@Data@eXo
      ov_keep_idx <- which(fit@Data@ov$name %in%
        unique(c(unlist(ov_names_block), unlist(fit@Data@ov.names.x))))
      slot_data_block@ov <- lapply(fit@Data@ov, "[", ov_keep_idx)
    }

    # if data.type == "moment", (re)create sample.cov and sample.nobs
    if (fit@Data@data.type == "moment") {
      mm_sample_mean <- NULL
      if (ngroups == 1L) {
        mm_sample_cov <- h1_all[[1L]]$cov
        if (fit@Model@meanstructure) {
          mm_sample_mean <- h1_all[[1L]]$mean
        }
        mm_sample_nobs <- fit@SampleStats@nobs[[1L]]
      } else {
        mm_sample_cov <- lapply(seq_len(ngroups), function(x) {
          h1_all[[x]]$cov[ov_names_block[[x]], ov_names_block[[x]]]
        })
        if (fit@Model@meanstructure) {
          mm_sample_mean <- lapply(seq_len(ngroups), function(x) {
            h1_all[[x]]$mean[ov_names_block[[x]]]
          })
        }
        mm_sample_nobs <- fit@SampleStats@nobs
      }
    }

    # handle single block 1-factor CFA with (only) two indicators
    # (under conditional.x the block variables include the exogenous
    # covariates -- count the indicators only)
    n_ind_block <- length(unlist(ov_names_block))
    if (slot_options_mm$conditional.x) {
      n_ind_block <- length(unlist(lapply(1:ngroups, function(g) {
        unique(unlist(lav_pt_vnames(ptm, type = "ov.nox", group = g)))
      })))
    }
    if (n_ind_block == 2L && ngroups == 1L) {
      lambda_idx <- which(ptm$op == "=~")
      # check if both factor loadings are fixed
      # (note: this assumes std.lv = FALSE)
      if (any(ptm$free[lambda_idx] != 0)) {
        ptm$free[lambda_idx] <- 0L
        ptm$ustart[lambda_idx] <- 1
        ptm$start[lambda_idx] <- 1
        # adjust free counter
        free_idx <- which(as.logical(ptm$free))
        if (length(free_idx) > 0L) {
          ptm$free[free_idx] <- seq_along(free_idx)
        }
        lav_msg_warn(gettextf(
          "measurement block [%1$s] (%2$s) contains only two indicators;
          -> fixing both factor loadings to unity",
          mm, lav_msg_view(mm_list[[mm]], "none")))
      }
    }

    # fit this measurement model only
    # (question: can we re-use even more slots?)
    if (fit@Data@data.type == "full") {
      fit_mm_block <- lavaan(
        model = ptm, slot_data = slot_data_block,
        slot_options = slot_options_mm, debug = FALSE, verbose = FALSE
      )
    } else if (fit@Data@data.type == "moment") {
      slot_options_mm$sample.cov.rescale <- FALSE
      fit_mm_block <- lavaan(
        model = ptm, slot_data = slot_data_block,
        sample_cov = mm_sample_cov, sample_mean = mm_sample_mean,
        sample_nobs = mm_sample_nobs,
        slot_options = slot_options_mm, debug = FALSE, verbose = FALSE
      )
    }

    # check convergence
    if (!lavInspect(fit_mm_block, "converged")) {
      # warning for now, but this is not good!
      lav_msg_warn(gettextf(
        "measurement model for %s did not converge!",
        lav_msg_view(mm_list[[mm]], "none")))
    }

    # check that this measurement block is identified on its own: in the
    # SAM approach every block is estimated in isolation, so a block with
    # more free parameters than sample statistics (negative degrees of
    # freedom) cannot be estimated, even if the full (joint) model is
    # identified (eg a factor with three indicators plus a residual
    # covariance between two of them). Without this check the block fit
    # produces garbage and downstream computations fail cryptically.
    blk_df <- fit_mm_block@test[[1]]$df
    if (!is.null(blk_df) && !is.na(blk_df) && blk_df < 0L) {
      lav_msg_stop(gettextf(
        "measurement block %1$s is not identified on its own: it has more
         free parameters than sample statistics (df = %2$d). The SAM
         approach estimates each measurement block in isolation, so every
         block must be identified by itself, even if the full model is
         identified. Consider simplifying the measurement block (eg
         removing residual covariances), combining blocks via mm.list, or
         using sem() instead.",
        lav_msg_view(mm_list[[mm]], "none"), blk_df))
    }

    # store fitted measurement model
    mm_fit[[mm]] <- fit_mm_block

    # fill in point estimates measurement block (including slack values)
    ptm <- mm_fit[[mm]]@ParTable
    # note: the row-numbers in PT that correspond to the rows in PTM
    # are given by mm_idx
    ptm_idx <- which((ptm$free > 0L | ptm$op %in% c(":=", "<", ">")) &
      ptm$user != 3L)
    block_ptm_idx[[mm]] <- ptm_idx
    pt_1$est[mm_idx[ptm_idx]] <- ptm$est[ptm_idx]

    # if categorical, add non-free residual variances
    if (fit_mm_block@Model@categorical || fit_mm_block@Model@correlation) {
      extra_idx <- which(ptm$op %in% c("~~", "~*~") &
        ptm$lhs == ptm$rhs &
        ptm$user == 0L &
        ptm$free == 0L &
        ptm$ustart == 1)
      if (length(extra_idx) > 0L) {
        pt_1$est[mm_idx[extra_idx]] <- ptm$est[extra_idx]
      }
    }
    # if EFA, add user=7 values (but do not add to ptm.idx)
    user7_idx <- which(ptm$user == 7L)
    if (length(user7_idx)) {
      pt_1$est[mm_idx[user7_idx]] <- ptm$est[user7_idx]
    }

    # add step1.free.idx
    par_idx <- pt_free[mm_idx[ptm_idx]]
    # store (ordered) indices in step1.free.idx
    step1_free_idx <- c(step1_free_idx, sort.int(par_idx)) # all combined

    # fill in standard errors measurement block
    if (!lavoptions$se %in% c("none", "bootstrap")) {
      ptm_free <- ptm$free
      if (fit_mm_block@Model@ceq.simple.only) {
        ptm_free[ptm_free > 0] <- seq_len(fit_mm_block@Model@nx.unco)
      }

      ptm_se_idx <- which((ptm$free > 0L) & ptm$user != 3L) # no :=, <, >
      pt_1$se[mm_idx[ptm_se_idx]] <- ptm$se[ptm_se_idx]

      # fill in variance matrix for this measurement block
      sigma_11 <- mm_fit[[mm]]@vcov$vcov
      if (is.null(sigma_11)) {
        # the block vcov could not be computed (eg the block information
        # matrix could not be inverted, typically an identification issue
        # that the df >= 0 check above cannot catch)
        lav_msg_stop(gettextf(
          "the variance matrix of measurement block %s could not be
           computed (its information matrix could not be inverted); the
           block may not be identified on its own (empirically). Consider
           simplifying the measurement block, combining blocks via
           mm.list, or using sem() instead.",
          lav_msg_view(mm_list[[mm]], "none")))
      }
      keep_idx <- ptm_free[ptm_idx]
      sigma_11_1[par_idx, par_idx] <-
        sigma_11[keep_idx, keep_idx, drop = FALSE]
    }
  } # measurement block

  # only keep 'measurement part' parameters in Sigma.11
  if (!lavoptions$se %in% c("none", "bootstrap")) {
    sigma_11_1 <- sigma_11_1[step1_free_idx, step1_free_idx, drop = FALSE]
  } else {
    sigma_11_1 <- NULL
  }

  # create STEP1 list
  step1 <- list(
    MM.FIT = mm_fit, Sigma.11 = sigma_11_1,
    step1.free.idx = step1_free_idx,
    block.mm.idx = block_mm_idx,
    block.ptm.idx = block_ptm_idx,
    PT.free = pt_free,
    mm.list = mm_list, PT = pt_1
  )

  step1
}

# STEP 0: process full model, without fitting
lav_sam_step0 <- function(cmd = "sem", model = NULL, data = NULL,
                          se = "twostep", sam_method = "local",
                          dotdotdot = NULL) {

  # create dotdotdot0 for dummy fit
  dotdotdot0 <- dotdotdot

  # 'parse' the model, so we can inspect a few features; accept the same
  # model= input types as a regular call to lavaan(): model syntax (character),
  # a formula, a fitted lavaan object, a flat model, or a parameter table
  # (issue #514)
  if (is.null(dotdotdot$parser)) {
    useparser <- lav_options_default()$parser
  } else {
    useparser <- dotdotdot$parser
  }
  flat_model <- lav_step01_ovnames_initflat(
    model = model, dotdotdot_parser = useparser
  )

  # composites (<~) have no reflective measurement model, so there is no
  # measurement block that step 1 could estimate in isolation
  if (any(flat_model$op == "<~")) {
    lav_msg_stop(gettext(
      "composites (defined by the <~ operator) are not supported by the
       SAM approach (yet); consider using sem() instead."))
  }

  # remove do.fit option if present
  dotdotdot0$do.fit <- NULL

  if (sam_method %in% c("local", "fsr", "cfsr")) {
    dotdotdot0$sample.icov <- FALSE # if N < nvar
  }
  dotdotdot0$se                   <- "none"
  dotdotdot0$test                 <- "none"
  dotdotdot0$verbose              <- FALSE # no output for this 'dummy' FIT
  dotdotdot0$ceq.simple           <- TRUE # if not the default yet
  dotdotdot0$check.lv.interaction <- FALSE # we allow for it
  # note: setting cat.wls.w = FALSE (no weight matrix if categorical)
  # would break the computation of twostep standard errors

  # single-level clustered data (cluster = , no level: syntax): the classic
  # twostep correction uses the model-based (expected) information for the
  # second-step piece, which is not robust to clustering; switch to the
  # robust (sandwich) variant, which picks up the cluster-robust Gamma
  multilevel_flag <- any(flat_model$op == ":" &
                         tolower(flat_model$lhs) == "level")
  # if model= was a parameter table, the "level:" rows are absent, but the
  # level column is available
  if (!multilevel_flag && !is.null(flat_model$level)) {
    multilevel_flag <-
      length(unique(flat_model$level[flat_model$level > 0L])) > 1L
  }
  if (se == "twostep" && !multilevel_flag && !is.null(dotdotdot$cluster)) {
    se <- "twostep.robust"
  }

  if (se %in% c("local", "ij", "twostep.robust")) {
    dotdotdot0$sample.icov <- TRUE
    # keep the moments of the exogenous covariates in the sample statistics
    # (their sampling variability is part of Gamma) -- but not when
    # conditional.x = TRUE, where the x-moments are fixed by design and
    # fixed.x = FALSE is not supported (see also the retry below, for the
    # settings where conditional.x = TRUE only emerges during the options
    # processing, eg categorical data + exogenous covariates)
    if (!isTRUE(dotdotdot$conditional.x)) {
      dotdotdot0$fixed.x <- FALSE
    }
    # ij/twostep.robust need the stored observed Gamma (NACOV); the
    # "force.model" ov_order sentinel keeps it in model order (it is only
    # honored when NACOV is supplied; see lav_lavaan_step00_init()). se =
    # "local" builds Gamma.eta in (memory-lean) influence form and computes the
    # observed Gamma on demand only when needed (eg categorical), so it needs
    # neither -- and the default ov_order is already "model".
    if (se %in% c("ij", "twostep.robust")) {
      dotdotdot0$NACOV <- TRUE
      dotdotdot0$ov_order <- "force.model" # avoid data ordering...
    }
  }

  # any lv interaction terms?
  if (length(lav_object_vnames(flat_model, "lv.interaction")) > 0L) {
    dotdotdot0$meanstructure <- TRUE
  }

  # initial processing of the model, no fitting
  fit <- tryCatch(
    do.call(cmd,
      args = c(list(
        model  = flat_model,
        data   = data,
        do.fit = FALSE
      ), dotdotdot0)
    ),
    error = function(e) e
  )
  if (inherits(fit, "error")) {
    fit0_error <- fit
    fit <- NULL
    if (identical(dotdotdot0$fixed.x, FALSE) ||
        isTRUE(dotdotdot0$gamma.unbiased)) {
      # fixed.x = FALSE (set above for se = local/ij/twostep.robust) and the
      # unbiased Gamma (the sam() default) are both unsupported when
      # conditional.x = TRUE, which may only emerge during the options
      # processing (eg categorical data + exogenous covariates, where
      # conditional.x = TRUE is the default): retry with
      # conditional.x-compatible settings, and keep the result only if
      # conditional.x is indeed TRUE
      dotdotdot0$fixed.x <- NULL
      dotdotdot0$gamma.unbiased <- FALSE
      fit <- tryCatch(
        do.call(cmd,
          args = c(list(
            model  = flat_model,
            data   = data,
            do.fit = FALSE
          ), dotdotdot0)
        ),
        error = function(e) NULL
      )
      if (!is.null(fit) && !fit@Model@conditional.x) {
        fit <- NULL # not the conditional.x situation: keep the original error
      }
    }
    if (is.null(fit)) {
      lav_msg_stop(gettextf(
        "initial processing of the model failed: %s",
        conditionMessage(fit0_error)))
    }
  }

  # restore options

  # do.fit
  fit@Options$do.fit <- TRUE

  # conditional.x: the unbiased Gamma does not exist -- make sure no
  # downstream fit (measurement blocks, structural part, lav_gamma_used())
  # ever requests it (the sam() default is gamma.unbiased = TRUE)
  if (fit@Model@conditional.x && isTRUE(fit@Options$gamma.unbiased)) {
    fit@Options$gamma.unbiased <- FALSE
  }

  # sample.icov
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    fit@Options$sample.icov <- TRUE
  }

  # se
  # PML: the two-step (robust) corrections need ingredients (eg the h1
  # information, the block influence P) that are not available for the
  # pairwise likelihood; only the naive (structural-fit) standard errors
  # can be computed, and only for the local methods
  if (identical(fit@Options$estimator.orig, "PML")) {
    if (sam_method == "global") {
      lav_msg_stop(gettext(
        "sam(sam.method = \"global\") does not support estimator PML (yet);
         use sam.method = \"local\" or sem() instead."))
    }
    # notes:
    # - se = "local" IS available for PML: it uses the casewise
    #   structural-space Gamma.eta (see lav_sam_gamma_eta_pml(); single
    #   group, complete data, all-ordinal -- guarded there);
    # - "twostep.huber.white" is in the fallback list DELIBERATELY, even
    #   though casewise PML scores exist (lav_sc_pml()): for
    #   sam.method = "local" the joint-PML-given-theta1 estimator that the
    #   huber.white sandwich linearizes is NOT the local (VETA-fit) step-2
    #   estimator (their variance estimates differ substantially), so the
    #   sandwich describes the wrong estimator; use se = "local" instead.
    if (se %in% c("twostep", "twostep.robust", "twostep.huber.white",
                  "local.nt")) {
      lav_msg_warn(gettextf(
        "se = \"%s\" is not available (yet) for estimator PML; naive
         standard errors are reported instead
         (tip: se = \"local\" is available for PML).", se))
      se <- "naive"
    }
  }
  if (fit@Model@categorical && se == "twostep") {
    # for categorical data, the classic ('global') two-step correction uses
    # the model-based information matrix, which underestimates the standard
    # errors for the (D)WLS estimator. Use the robust (Yuan & Chan, 2002)
    # correction instead. This needs the 'P' matrix (the influence of the
    # measurement parameters on the sample statistics), which is available for
    # all sam methods (the measurement blocks are always fitted in step 1).
    se <- "twostep.robust"
  }
  if (fit@Model@conditional.x && se == "twostep" &&
      sam_method %in% c("local", "fsr", "cfsr")) {
    # conditional.x + local sam method: the classic two-step correction is
    # built on the joint information matrix, so it linearizes the
    # joint-model-given-theta1 estimator. Under conditional.x that estimator
    # is NOT asymptotically equivalent to the local step-2 estimator (the
    # structural fit is saturated in the latent-on-x slopes and cannot pool
    # information across statistics the way the joint estimator does), and
    # the joint-information formula underestimates the sampling variability
    # (badly so for binary covariates). Use the robust variant instead: under
    # conditional.x it is computed as the structural-space sandwich (see the
    # tsrobust_condx_flag note in lav_sam_step2_se.R), which linearizes the
    # actual two-step estimator (and coincides with se = "local"). For
    # sam.method = "global" step 2 IS the joint estimator, and the classic
    # formula still applies.
    se <- "twostep.robust"
  }
  fit@Options$se <- se

  # test
  if (!is.null(dotdotdot$test)) {
    fit@Options$test <- lav_test_rename(dotdotdot$test)
  } else if (sam_method == "global") {
    # global SAM: the Yuan & Chan (2002) rescaled GLOBAL statistic is the
    # default robust test (the analogue of test = "satorra.bentler" in sem()).
    fit@Options$test <- "yuan.chan"
  } else {
    fit@Options$test <- "standard"
  }

  # adjust parameter table:
  pt_1 <- fit@ParTable

  # check parameter table
  pt_1$est <- pt_1$se <- NULL
  # est equals ustart by default (except exo values)
  pt_1$est <- pt_1$ustart
  if (any(pt_1$exo > 0L)) {
    pt_1$est[pt_1$exo > 0L] <- pt_1$start[pt_1$exo > 0L]
  }
  # thresholds: fill in the sample-based starting values. Thresholds of
  # observed (dummy) variables that are not part of any measurement block
  # (eg an ordered endogenous variable in the structural part) are never
  # (re)estimated in step 1 or step 2 and keep these h1 values; leaving
  # them NA breaks every computation that evaluates the joint model
  # matrices (information, Gamma.eta) before the final assembly. Block
  # thresholds are simply overwritten by the step-1 estimates.
  th_idx <- which(pt_1$op == "|" & is.na(pt_1$est))
  if (length(th_idx) > 0L) {
    pt_1$est[th_idx] <- pt_1$start[th_idx]
  }

  # clear se values (needed here?) only for global approach to compute SE
  pt_1$se <- rep(as.numeric(NA), length(pt_1$lhs))
  pt_1$se[pt_1$free == 0L & !is.na(pt_1$ustart)] <- 0.0

  fit@ParTable <- pt_1


  fit
}

# Factor extraction method(s)
# YR Feb 2020
#
# - ULS_corner only (for now)
# - just to get better starting values for ESEM

# YR July 2020
# - adding generic function lav_efa_extraction, using eigenvalue based
#   approach; ML and ULS
# - 'corner' is an option

lav_efa_extraction <- function(s, nfactors = 1L,
                               method = "ULS", # or ML
                               corner = FALSE,
                               reflect = FALSE, order_lv_by = "none",
                               min_var = 0.0001) {
  stopifnot(is.matrix(s))
  s <- unname(s)
  method <- tolower(method)

  # extract variances
  s_var <- diag(s)

  # force S to be pd (eg if we have polychoric correlations)
  s <- lav_mat_sym_force_pd(s, tol = 1e-08)

  # convert to correlation matrix (ULS is not scale invariant!)
  r <- cov2cor(s)

  # optim.method
  if (method == "uls") {
    min_objective <- lav_efa_uls_min_objective
    min_gradient <- lav_efa_uls_min_grad
    cache <- lav_efa_uls_init_cache(r = r, nfactors = nfactors)
  } else if (method == "ml") {
    min_objective <- lav_efa_ml_min_objective
    min_gradient <- lav_efa_ml_min_grad
    cache <- lav_efa_ml_init_cache(r = r, nfactors = nfactors)
  } else {
    lav_msg_stop(gettext("method must be uls or ml (for now)"))
  }
  min_hessian <- NULL

  # optimize
  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = if (lav_verbose()) {
      1L
    } else {
      0L
    },
    abs.tol = (.Machine$double.eps * 10)
  )
  if (lav_verbose()) {
    cat("\n")
    cat("factor extraction iterations using ", method, ":\n")
  }
  out <- nlminb(
    start = cache$theta, objective = min_objective,
    gradient = min_gradient, hessian = min_hessian,
    control = control_nlminb, lower = min_var, upper = +1,
    cache = cache
  )

  # extract LAMBDA/THETA
  if (method == "uls") {
    mm_theta <- diag(out$par * out$par)

    # compute LAMBDA
    a <- r
    diag(a) <- diag(a) - (out$par * out$par)
    ev <- eigen(a, symmetric = TRUE)
    omega_1 <- ev$vectors[, 1:nfactors]
    gamma_1 <- ev$values[1:nfactors]

    # LAMBDA <- Omega.1 %*% diag(sqrt(gamma.1))
    mm_lambda <- t(t(omega_1) * sqrt(gamma_1))

    # rescale if the input matrix was not a correlation matrix
    mm_lambda <- sqrt(s_var) * mm_lambda
    diag(mm_theta) <- s_var * diag(mm_theta)
  } else if (method == "ml") {
    mm_theta <- diag(out$par * out$par)

    # compute LAMBDA
    psi <- out$par
    a <- t(psi * cache$r_inv) * psi
    ev <- eigen(a, symmetric = TRUE)
    omega_1 <- ev$vectors[, 1L + cache$nvar - seq_len(cache$nfactors),
      drop = FALSE
    ]
    gamma_1 <- ev$values[1L + cache$nvar - seq_len(cache$nfactors)]

    # LAMBDA <- diag(psi) %*% Omega.1 %*%sqrt(solve(Gamma.1)-diag(nfactors))
    tmp1 <- psi * omega_1
    mm_lambda <- t(t(tmp1) * sqrt((1 / gamma_1) - 1))

    # rescale if the input matrix was not a correlation matrix
    mm_lambda <- sqrt(s_var) * mm_lambda
    diag(mm_theta) <- s_var * diag(mm_theta)
  }

  # corner?
  if (corner) {
    # rotate to echelon pattern (see echelon() in GPArotation package)
    head_1 <- mm_lambda[seq_len(nfactors), , drop = FALSE]
    post <- try(solve(head_1, t(chol(tcrossprod(head_1)))), silent = TRUE)
    okflag <- FALSE
    if (inherits(post, "try-error")) { # new in 0.6-18
      # this will happen if we have identical elements in the columns
      # of HEAD (perhaps the data is artificial?)
      # -> add some fuzz and try again
      sd_1 <- sqrt(mean(abs(head_1))) * 1e-04
      fuzz <- matrix(rnorm(nfactors * nfactors, 0, sd_1), nfactors, nfactors)
      head2 <- head_1 + fuzz
      post <- try(solve(head2, t(chol(tcrossprod(head2)))), silent = TRUE)
      if (!inherits(post, "try-error")) {
        okflag <- TRUE
      }
    } else {
      okflag <- TRUE
    }
    if (okflag) {
      mm_lambda <- mm_lambda %*% post
    } else {
      lav_msg_warn(gettext(
        "rotation of initial factor solution to echelon pattern failed."))
    }
  }

  # ALWAYS change the sign so that largest element in the column is positive
  # neg.max <- apply(LAMBDA, 2, function(x) { sign(x[which.max(abs(x))]) })
  # neg.idx <- which(neg.max < 0)
  # if(length(neg.idx) > 0L) {
  #    LAMBDA[, neg.idx] <- -1 * LAMBDA[, neg.idx, drop = FALSE]
  # }

  # ALWAYS change the sign so that diag(LAMBDA) is positive
  neg_idx <- which(diag(mm_lambda) < 0)
  if (length(neg_idx) > 0L) {
    mm_lambda[, neg_idx] <- -1 * mm_lambda[, neg_idx, drop = FALSE]
  }

  # reflect (optional) + reorder the columns
  mm_lambda <- lav_efa_reflect_and_reorder(mm_lambda, reflect, order_lv_by)

  list(LAMBDA = mm_lambda, THETA = mm_theta)
}



lav_efa_uls_init_cache <- function(r = NULL,
                                          nfactors = 1L,
                                          parent = parent.frame()) {
  r_inv <- solve(r)
  nvar <- ncol(r)

  # starting values for diagonal elements of THETA
  # using Joreskog (1966) suggestion:
  theta_init <- (1 - nfactors / (2 * nvar)) * 1 / diag(r_inv)
  theta <- sqrt(theta_init)

  out <- list2env(
    list(
      r = r, nfactors = nfactors,
      theta = theta
    ),
    parent = parent
  )
  out
}

# x is here the sqrt() of theta!
lav_efa_uls_min_objective <- function(x, cache = NULL) {
  cache$theta <- x
  with(cache, {    # nolint start
    a <- r
    diag(a) <- diag(a) - (theta * theta)
    ev <- eigen(a, symmetric = TRUE, only.values = TRUE)
    gamma_2 <- ev$values[-seq_len(nfactors)]

    res <- 0.5 * sum(gamma_2 * gamma_2)
    return(res)
  })               # nolint end
}

lav_efa_uls_min_grad <- function(x, cache = NULL) {
  # check if x has changed
  if (!all(x == cache$theta)) {
    cache$theta <- x
    # nothing to do
  }
  with(cache, {       # nolint start
    a <- r
    diag(a) <- diag(a) - (theta * theta)
    ev <- eigen(a, symmetric = TRUE)
    omega_2 <- ev$vectors[, -seq_len(nfactors)]
    gamma_2 <- ev$values[-seq_len(nfactors)]

    res <- -2 * theta * colSums(t(omega_2 * omega_2) * gamma_2)
    return(res)
  })                  # nolint end
}


# ML
lav_efa_ml_init_cache <- function(r = NULL,
                                  nfactors = 1L,
                                  parent = parent.frame()) {
  r_inv <- solve(r)
  nvar <- ncol(r)

  # starting values for diagonal elements of THETA
  # using Joreskog (1966) suggestion:
  theta_init <- (1 - nfactors / (2 * nvar)) * 1 / diag(r_inv)
  theta <- sqrt(theta_init)

  out <- list2env(
    list(
      r = r, nfactors = nfactors, r_inv = r_inv,
      nvar = nvar, # for ML only
      theta = theta
    ),
    parent = parent
  )
  out
}


# x is here the sqrt of theta
lav_efa_ml_min_objective <- function(x, cache = NULL) {
  cache$theta <- x
  with(cache, {       # nolint start
    psi <- theta
    # A <- diag(psi) %*% R.inv %*% diag(psi)
    a <- t(r_inv * psi) * psi
    ev <- eigen(a, symmetric = TRUE, only.values = TRUE)
    gamma_2 <- ev$values[(nvar - nfactors):1L]

    res <- sum(log(gamma_2) + 1 / gamma_2 - 1)
    return(res)
  })                  # nolint end
}

lav_efa_ml_min_grad <- function(x, cache = NULL) {
  # check if x has changed
  if (!all(x == cache$theta)) {
    cache$theta <- x
    # nothing to do
  }
  with(cache, {          # nolint start
    psi <- theta
    # A <- diag(psi) %*% solve(S) %*% diag(psi)
    a <- t(r_inv * psi) * psi
    ev <- eigen(a, symmetric = TRUE)

    omega_2 <- ev$vectors[, (nvar - nfactors):1L, drop = FALSE]
    gamma_2 <- ev$values[(nvar - nfactors):1L]

    res <- colSums(t(omega_2 * omega_2) * (1 - 1 / gamma_2))
    return(res)
  })                     # nolint end
}




# ULS estimation
#
# - but resulting in a upper-corner all zeroes LAMBDA matrix
# - not using eigenvalues/vectors, but minimizing the residuals
#   directly

# - should give the same results as MINRES (after an orthogonal transformation)
# - unless there are heywood cases; this function allows for negative variances!

lav_efa_extraction_uls_corner <- function(s, nfactors = 1L, reflect = TRUE,
                                          order_lv_by = "none") {
  stopifnot(is.matrix(s))
  s <- unname(s)
  nvar <- nrow(s)

  # extract variances
  s_var <- diag(s)

  # convert to correlation matrix (ULS is not scale invariant!)
  r <- cov2cor(s)
  # R.inv <- solve(R)

  # eigenvalue decomposition (to get starting values for LAMBDA)
  ev <- eigen(r, symmetric = TRUE)

  # extract first nfac components (assuming no measurement error)
  pc <- (ev$vectors[, seq_len(nfactors), drop = FALSE] %*%
         diag(sqrt(ev$values[seq_len(nfactors)])))

  # rotate to echelon pattern (see echelon() in GPArotation package)
  head_1 <- pc[seq_len(nfactors), , drop = FALSE]
  mm_lambda <- pc %*% solve(head_1, t(chol(tcrossprod(head_1))))

  mm_theta <- diag(nvar)
  if (nfactors > 1L) {
    corner_idx <- which(row(mm_lambda) < nfactors &
                        col(mm_lambda) > row(mm_lambda))
    lambda_idx <- seq_len(nvar * nfactors)[-corner_idx]
    mm_lambda[corner_idx] <- 0 # to make them exactly zero
  } else {
    corner_idx <- integer(0L)
    lambda_idx <- seq_len(nvar)
  }

  # optim.method
  min_objective <- lav_efa_uls_corner_min_objective
  min_gradient <- lav_efa_uls_corner_min_grad
  min_hessian <- NULL

  # create cache environment
  cache <- lav_efa_uls_corner_init_cache(
    mm_lambda = mm_lambda,
    lambda_idx = lambda_idx, r = r
  )

  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = if (lav_verbose()) {
      1L
    } else {
      0L
    },
    abs.tol = (.Machine$double.eps * 10)
  )

  # optimize
  out <- nlminb(
    start = cache$theta, objective = min_objective,
    gradient = min_gradient, hessian = min_hessian,
    control = control_nlminb, lower = -1, upper = +1,
    cache = cache
  )

  mm_lambda[lambda_idx] <- out$par
  diag(mm_theta) <- 1 - diag(tcrossprod(mm_lambda))

  # rescale if the input matrix was not a correlation matrix
  mm_lambda <- sqrt(s_var) * mm_lambda
  diag(mm_theta) <- s_var * diag(mm_theta)

  # reflect (optional) + reorder the columns
  mm_lambda <- lav_efa_reflect_and_reorder(mm_lambda, reflect, order_lv_by)

  list(LAMBDA = mm_lambda, THETA = mm_theta)
}


lav_efa_uls_corner_init_cache <- function(mm_lambda = NULL,
                                          lambda_idx = NULL,
                                          r = NULL,
                                          parent = parent.frame()) {
  theta <- mm_lambda[lambda_idx]
  out <- list2env(
    list(
      mm_lambda = mm_lambda,
      lambda_idx = lambda_idx,
      r = r,
      theta = theta
    ),
    parent = parent
  )
  out
}

lav_efa_uls_corner_min_objective <- function(x, cache = NULL) { # nolint
  cache$theta <- x
  with(cache, {               # nolint start
    mm_lambda[lambda_idx] <- theta
    res1 <- lav_mat_vech(r - tcrossprod(mm_lambda), diagonal = FALSE)
    res2 <- res1 * res1
    return(sum(res2))
  })                          # nolint end
}

lav_efa_uls_corner_min_grad <- function(x, cache = NULL) {
  # check if x has changed
  if (!all(x == cache$theta)) {
    cache$theta <- x
    # nothing to do
  }
  with(cache, {                            # nolint start
    mm_lambda[lambda_idx] <- theta
    sigma_1 <- tcrossprod(mm_lambda)
    diag(sigma_1) <- 1 # diagonal is ignored
    tmp <- -2 * (r - sigma_1) %*% mm_lambda
    return(tmp[lambda_idx])
  })                                       # nolint end
}

lav_model_grad_mml <- function(lavmodel = NULL,
                                   mm_theta = NULL,
                                   th = NULL,
                                   glist = NULL,
                                   group = 1L,
                                   lavdata = NULL,
                                   sample_mean = NULL,
                                   sample_mean_x = NULL,
                                   lavcache = NULL) {
  if (lavmodel@link == "logit") {
    lav_msg_stop(gettext("logit link not implemented yet; use probit"))
  }

  # shortcut
  ov_y_dummy_ov_idx <- lavmodel@ov.y.dummy.ov.idx[[group]]
  ov_x_dummy_ov_idx <- lavmodel@ov.x.dummy.ov.idx[[group]]
  ov_y_dummy_lv_idx <- lavmodel@ov.y.dummy.lv.idx[[group]]
  ov_x_dummy_lv_idx <- lavmodel@ov.x.dummy.lv.idx[[group]]
  # ov_dummy_idx <- c(ov_y_dummy_ov_idx, ov_x_dummy_ov_idx)
  lv_dummy_idx <- c(ov_y_dummy_lv_idx, ov_x_dummy_lv_idx)
  th_idx <- lavmodel@th.idx[[group]]
  num_idx <- lavmodel@num.idx[[group]]
  ord_idx <- unique(th_idx[th_idx > 0L])


  # data for this group
  x_1 <- lavdata@X[[group]]
  nobs <- nrow(x_1)
  nvar <- ncol(x_1)
  exo <- lavdata@eXo[[group]]

  # MLIST (for veta and yhat)
  mm_in_group <- 1:lavmodel@nmat[group] + cumsum(c(0, lavmodel@nmat))[group]
  mlist <- glist[mm_in_group]

  # quadrature points
  gh <- lavcache[[group]]$GH
  n_gh <- nrow(gh$x)
  nfac <- ncol(gh$x)

  # compute VETAx (latent lv only)
  # VETAx <- lav_lisrel_vetax(MLIST = MLIST, lv.dummy.idx = lv.dummy.idx)
  vetax <- lav_lisrel_vetax(mlist = mlist)
  # check for negative values?
  if (any(diag(vetax) < 0)) {
    lav_msg_warn(gettext("--- VETAx contains negative values"))
    print(vetax)
    return(0)
  }

  # cholesky?
  # if(is.null(lavmodel@control$cholesky)) {
  cholesky <- TRUE
  # } else {
  #    CHOLESKY <- as.logical(lavmodel@control$cholesky)
  # if(nfac > 1L && !CHOLESKY) {
  #    warning("lavaan WARNING: CHOLESKY is OFF but nfac > 1L")
  # }
  # }

  if (!cholesky) {
    # we should still 'scale' the factors, if std.lv=FALSE
    eta_sd <- sqrt(diag(vetax))
  } else {
    # cholesky takes care of scaling
    eta_sd <- rep(1, nfac)
    tchol_veta <- try(chol(vetax), silent = TRUE)
    if (inherits(tchol_veta, "try-error")) {
      lav_msg_warn(gettext("--- VETAx not positive definite"))
      print(vetax)
      return(0)
    }
    if (!is.null(mlist$alpha) || !is.null(mlist$gamma)) {
      eetax <- lav_lisrel_eetax(
        mlist = mlist, exo = exo, n = nobs,
        sample_mean = sample_mean,
        ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
        ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
        ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
        ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
      )
      # if(length(lv.dummy.idx) > 0L) {
      #    EETAx <- EETAx[,-lv.dummy.idx,drop=FALSE]
      # }
    }
  }

  # prepare common stuff
  # fix Lambda?
  mm_lambda <- lav_lisrel_lambda(
    mlist = mlist,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # fix ALPHA
  mm_alpha <- mlist$alpha
  if (is.null(mm_alpha)) {
    mm_alpha <- numeric(nfac)
  } else if (length(lv_dummy_idx)) {
    mm_alpha <- mm_alpha[-lv_dummy_idx, , drop = FALSE]
  }

  # Beta?
  mm_beta <- mlist$beta
  if (is.null(mm_beta)) {
    lambda__ib_inv <- mm_lambda
  } else {
    tmp <- -mm_beta
    nr <- nrow(mm_beta)
    i <- seq_len(nr)
    tmp[cbind(i, i)] <- 1
    ib_inv <- solve(tmp)
    lambda__ib_inv <- mlist$lambda %*% ib_inv ## no need to FIX???
    if (length(lv_dummy_idx) > 0L) {
      lambda__ib_inv <- lambda__ib_inv[, -lv_dummy_idx, drop = FALSE]
    }

    # fix BETA
    if (length(lv_dummy_idx)) {
      mm_beta <- mlist$beta[-lv_dummy_idx, -lv_dummy_idx, drop = FALSE]
    }
    tmp <- -mm_beta
    nr <- nrow(mm_beta)
    i <- seq_len(nr)
    tmp[cbind(i, i)] <- 1
    ib_inv <- solve(tmp)
  }

  # fix GAMMA
  mm_gamma <- mlist$gamma
  if (is.null(mm_gamma)) {
    alpha_gamma_e_xo <- matrix(as.numeric(mm_alpha), nobs, nfac, byrow = TRUE)
  } else if (length(lv_dummy_idx)) {
    mm_gamma <- mm_gamma[-lv_dummy_idx, , drop = FALSE]
    alpha_gamma_e_xo <- sweep(exo %*% t(mm_gamma),
      MARGIN = 2, STATS = as.numeric(mm_alpha), FUN = "+"
    )
  }

  # Delta
  ## DD <- lavcache[[group]]$DD
  dd <- lav_model_grad_dd(lavmodel, g_list = glist, group = group)

  ## FIXME!!! do this analytically...
  x <- lav_model_get_parameters(lavmodel = lavmodel, glist = mlist)
  d_vetadx <- function(x, lavmodel = lavmodel, g = 1L) {
    glist <- lav_model_x2glist(lavmodel, x = x, type = "free")
    vetax <- lav_model_vetax(lavmodel, glist = glist)[[g]]
    if (cholesky) {
      s <- chol(vetax) ### FIXME or t(chol())????
    } else {
      s <- diag(sqrt(diag(vetax)))
    }
    s
  }
  delta_s <- lav_func_jacobian_simple(func = d_vetadx,
               x = x, lavmodel = lavmodel, g = group)
  dd$S <- delta_s

  # compute dL/dx for each node
  # dLdx <-  matrix(0, nGH, lavmodel@nx.free)
  d_fyp <- matrix(0, nobs, lavmodel@nx.free)
  sum_log_fy <- matrix(0, nrow = n_gh, ncol = nobs)
  for (q in 1:n_gh) {
    # contribution to dFYp for this q
    d_fyp_q <- matrix(0, nobs, lavmodel@nx.free)

    # current value(s) for ETA
    eta <- ksi <- gh$x[q, , drop = FALSE]

    # rescale/unwhiten
    if (cholesky) {
      eta <- eta %*% tchol_veta
    } else {
      # no unit scale? (un-standardize)
      eta <- sweep(eta, MARGIN = 2, STATS = eta_sd, FUN = "*")
    }

    # eta_i = alpha + BETA eta_i + GAMMA eta_i + error
    #
    # - direct effect of BETA is already in VETAx, and hence tchol.VETA
    # - need to add alpha, and GAMMA eta_i
    if (!is.null(mlist$alpha) || !is.null(mlist$gamma)) {
      eta <- sweep(eetax, MARGIN = 2, STATS = eta, FUN = "+")
    }

    # again, compute yhat for this node (eta)
    if (lavmodel@conditional.x) {
      yhat <- lav_lisrel_eyetax(
        mlist = mlist, exo = exo,
        eta = eta, sample_mean = sample_mean,
        ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
        ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
        ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
        ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
      )
    } else {
      yhat <- lav_lisrel_eyetax3(
        mlist = mlist,
        eta = eta, sample_mean = sample_mean,
        mean_x = sample_mean_x,
        ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[group]],
        ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[group]],
        ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[group]],
        ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[group]]
      )
    }

    # compute fy.var, for this node (eta): P(Y_i =  y_i | eta_i, x_i)
    log_fy_var <- lav_predict_fy_internal(
      x = x_1, yhat = yhat,
      th = th, mm_theta = mm_theta,
      num_idx = num_idx, th_idx = th_idx,
      link = lavmodel@link, log_1 = TRUE
    )

    # if log, fy is just the sum of log.fy.var
    log_fy <- apply(log_fy_var, 1L, sum)

    # store log likelihoods for this node
    sum_log_fy[q, ] <- log_fy

    # FY
    fy <- exp(log_fy_var) ### FIXME log/exp/log/...
    lik_eta <- apply(fy, 1, prod)
    # fyp <- LIK.eta * GH$w[q]

    ######### dFY_p ###########################################
    # note, dFYp is actually 1/FY[,p] * dFYp

    pre <- matrix(0, nobs, nvar)
    if (length(num_idx) > 0L) {
      tmp <- x_1[, num_idx, drop = FALSE] - yhat[, num_idx, drop = FALSE]
      theta_var <- diag(mm_theta)[num_idx]
      pre[, num_idx] <- sweep(tmp, MARGIN = 2, STATS = 1 / theta_var, FUN = "*")
    }

    if (length(ord_idx) > 0L) {
      for (p in ord_idx) {
        # just in case we need theta[v,v] after all...
        sd_v_inv <- 1 / sqrt(mm_theta[p, p])

        # lav_probit
        y <- x_1[, p]
        th_y <- th[th_idx == p]
        th_y_1 <- c(-Inf, th_y, Inf)
        ncat <- length(th_y) + 1L
        nth <- ncat - 1L
        y1 <- matrix(1:nth, nobs, nth, byrow = TRUE) == y
        y2 <- matrix(1:nth, nobs, nth, byrow = TRUE) == (y - 1L)
        z1 <- pmin(100, th_y_1[y + 1L] - yhat[, p])
        z2 <- pmax(-100, th_y_1[y + 1L - 1L] - yhat[, p])
        p1 <- dnorm(z1)
        p2 <- dnorm(z2)
        # probits = p1 - p2

        pre[, p] <- -1 * (p1 - p2) * sd_v_inv * (1 / fy[, p])

        # [nobx * n.th]
        # dth <- -1 * (Y2*p2 - Y1*p1) * sd.v.inv
        dth <- -1 * (y2 * p2 - y1 * p1) * sd_v_inv * (1 / fy[, p])
        d_fyp_q <- d_fyp_q +
          (dth %*% dd$tau[which(th_idx == p), , drop = FALSE])
      }
    }

    if (length(num_idx) > 0L) {
      # THETA (num only)
      dsigma2 <- sweep(0.5 * pre[, num_idx, drop = FALSE] *
        pre[, num_idx, drop = FALSE],
        MARGIN = 2,
        STATS = 1 / (2 * theta_var), FUN = "-"
      )
      d_fyp_q <- d_fyp_q + (dsigma2 %*% dd$theta)

      # NU (num only)
      dnu <- pre[, num_idx, drop = FALSE]
      d_fyp_q <- d_fyp_q + (dnu %*% dd$nu)
    }

    # LAMBDA
    if (nrow(eta) == 1L) {
      dlambda <- pre %*% eta
      ### FIXME!!!!!
    } else {
      dlambda <- matrix(apply(pre, 2, function(x) x * eta), nobs, )
      # dlambda <- sweep(PRE, MARGIN=1, STATS=eta, FUN="*")
    }
    d_fyp_q <- d_fyp_q + (dlambda %*% dd$lambda)

    # PSI
    # if(nrow(ksi) == 1L) {
    dpsi <- pre %*% kronecker(mm_lambda[, , drop = FALSE], ksi)
    # } else {
    #    dpsi <- PRE * kronecker(LAMBDA[,,drop=FALSE], ksi)
    # }
    d_fyp_q <- d_fyp_q + (dpsi %*% dd$S)

    # KAPPA
    if (length(ov_y_dummy_ov_idx) > 0L) {
      dkappa <- matrix(apply(
        pre[, ov_y_dummy_ov_idx, drop = FALSE], 2,
        function(x) x * exo
      ), nobs, )
      d_fyp_q <- d_fyp_q + (dkappa %*% dd$kappa)
    }

    # GAMMA
    if (!is.null(exo)) {
      dgamma <- matrix(apply(
        pre %*% lambda__ib_inv, 2,
        function(x) x * exo
      ), nobs, )
      d_fyp_q <- d_fyp_q + (dgamma %*% dd$gamma)
    }

    # BETA
    if (!is.null(mm_beta)) {
      # tmp <- kronecker(LAMBDA, ALPHA.GAMMA.eXo) %*%
      #         t( kronecker(t(IB.inv), IB.inv) )
      # dbeta <- apply(matrix(as.numeric(PRE) * tmp, nobs, ), 1, sum)
      dbeta <- matrix(apply(
        pre %*% lambda__ib_inv, 2,
        function(x) x * alpha_gamma_e_xo
      ), nobs, )
      d_fyp_q <- d_fyp_q + (dbeta %*% dd$beta)
    }

    d_fyp <- d_fyp + ((lik_eta * gh$w[q]) * d_fyp_q)
  }

  lik <- as.numeric(t(gh$w) %*% exp(sum_log_fy))
  # avoid underflow
  idx <- which(lik < exp(-600))
  if (length(idx) > 0L) {
    lik[idx] <- exp(-600)
  }

  d_fyp <- 1 / lik * d_fyp

  dx <- apply(d_fyp, 2, sum)

  # integration
  # dx <- apply(as.numeric(GH$w) * dLdx, 2, sum)

  # minimize
  dx <- -1 * dx

  dx
}

# Evaluate 'expr' using a local, fixed RNG state, and restore the global
# '.Random.seed' afterwards. This keeps RNG-using helpers (eg the jacobian
# linearity checks below) from disturbing the global stream -- which would,
# for example, change the seed picked by a subsequent se = "bootstrap" run.
lav_with_local_seed <- function(expr, seed = 1234L) {
  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    old_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    on.exit(assign(".Random.seed", old_seed, envir = .GlobalEnv,
                   inherits = FALSE), add = TRUE)
  } else {
    # no '.Random.seed' existed: remove the one set.seed() creates below
    on.exit(rm(".Random.seed", envir = .GlobalEnv, inherits = FALSE),
            add = TRUE)
  }
  set.seed(seed)
  force(expr)
}

lav_con_parse <- function(partable = NULL, constraints = NULL,
                                  theta = NULL,
                                  debug = FALSE) {
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  # just in case we do not have a $free column in partable
  if (is.null(partable$free)) {
    partable$free <- seq_along(partable$lhs)
  }

  # from the partable: free parameters
  if (!is.null(theta)) {
    # nothing to do
  } else if (!is.null(partable$est)) {
    theta <- partable$est[partable$free > 0L]
  } else if (!is.null(partable$start)) {
    theta <- partable$start[partable$free > 0L]
  } else {
    theta <- rep(0, length(partable$lhs))
  }

  # note: avoid zero values in theta (so jac is not all null)
  # (eg, regression coefficients are often zero in partable$start)
  # we perturb with small random numbers, but isolate the RNG state
  # (see lav_with_local_seed() above)
  zero_theta_idx <- which(theta == 0)
  if (length(zero_theta_idx) > 0L) {
    theta[zero_theta_idx] <- lav_with_local_seed(
      rnorm(length(zero_theta_idx), 0, 0.1))
  }

  # number of free (but possibly constrained) parameters
  npar <- length(theta)

  # parse the constraints
  if (is.null(constraints)) {
    list_1 <- NULL
  } else if (!is.character(constraints)) {
    lav_msg_stop(gettext("constraints should be a string"))
  } else {
    flat <- lavParseModelString(constraints)
    con <- attr(flat, "constraints")
    list_1 <- list()
    if (length(con) > 0L) {
      lhs <- unlist(lapply(con, "[[", "lhs"))
      op <- unlist(lapply(con, "[[", "op"))
      rhs <- unlist(lapply(con, "[[", "rhs"))
      list_1$lhs <- c(list_1$lhs, lhs)
      list_1$op <- c(list_1$op, op)
      list_1$rhs <- c(list_1$rhs, rhs)
    } else {
      lav_msg_stop(gettext("no constraints found in constraints argument"))
    }
  }

  # simple equality constraints?
  ceq_simple <- FALSE
  if (!is.null(partable$unco)) {
    ceq_simple <- TRUE
  }

  # simple inequality constraints?
  cin_simple <- TRUE
  if (any(partable$op %in% c(">", "<"))) {
    cin_simple <- FALSE
  }
  if (!is.null(list_1) && any(list_1$op %in% c(">", "<"))) {
    cin_simple <- FALSE
  }
  if (is.null(partable$upper) && is.null(partable$lower)) {
    cin_simple <- FALSE
  } else {
    # only check free parameters!
    free_idx <- which(partable$free > 0L)
    if (all(partable$lower[free_idx] == -Inf) &&
        all(partable$upper[free_idx] == +Inf)) {
      cin_simple <- FALSE
    }
  }

  # variable definitions
  def_function <- lav_pt_con_def(partable,
    con = list_1,
    debug = debug
  )

  # construct ceq/ciq functions
  ceq_function <- lav_pt_con_ceq(partable,
    con = list_1,
    debug = debug
  )
  # linear or nonlinear?
  ceq_linear_idx <- lav_con_linear_idx(
    func = ceq_function,
    npar = npar
  )
  ceq_nonlinear_idx <- lav_con_nonlinear_idx(
    func = ceq_function,
    npar = npar
  )

  # inequalities
  cin_function <- lav_pt_con_ciq(partable,
    con = list_1,
    debug = debug
  )

  # linear or nonlinear?
  cin_linear_idx <- lav_con_linear_idx(
    func = cin_function,
    npar = npar
  )
  cin_nonlinear_idx <- lav_con_nonlinear_idx(
    func = cin_function,
    npar = npar
  )

  # Jacobians
  if (!is.null(body(ceq_function))) {
    ceq_jac <- try(lav_func_jacobian_complex(
      func = ceq_function,
      x = theta
    ), silent = TRUE)
    if (inherits(ceq_jac, "try-error")) { # eg. pnorm()
      ceq_jac <- lav_func_jacobian_simple(func = ceq_function, x = theta)
    }

    # constants
    # do we have a non-zero 'rhs' elements? FIXME!!! is this reliable??
    ceq_rhs <- -1 * ceq_function(numeric(npar))

    # evaluate constraints
    ceq_theta <- ceq_function(theta)
  } else {
    ceq_jac <- matrix(0, nrow = 0L, ncol = npar)
    ceq_rhs <- numeric(0L)
    ceq_theta <- numeric(0L)
  }

  if (!is.null(body(cin_function))) {
    cin_jac <- try(lav_func_jacobian_complex(
      func = cin_function,
      x = theta
    ), silent = TRUE)
    if (inherits(cin_jac, "try-error")) { # eg. pnorm()
      cin_jac <- lav_func_jacobian_simple(func = cin_function, x = theta)
    }

    # constants
    # do we have a non-zero 'rhs' elements? FIXME!!! is this reliable??
    cin_rhs <- -1 * cin_function(numeric(npar))

    # evaluate constraints
    cin_theta <- cin_function(theta)
  } else {
    cin_jac <- matrix(0, nrow = 0L, ncol = npar)
    cin_rhs <- numeric(0L)
    cin_theta <- numeric(0L)
  }

  # check for empty/unused constraints (new in 0.7-1)
  if (nrow(ceq_jac) > 0L) {
    if (all(ceq_jac == 0)) {
      ceq_jac <- matrix(0, nrow = 0L, ncol = npar)
      ceq_rhs <- numeric(0L)
      ceq_theta <- numeric(0L)
      ceq_linear_idx <- integer(0L)
      ceq_nonlinear_idx <- integer(0L)
    } else {
      zero_idx <- which(apply(ceq_jac, 1, function(x) all(x == 0)))
      if (length(zero_idx) > 0L) {
        ceq_jac <- ceq_jac[-zero_idx, , drop = FALSE]
        ceq_rhs <- ceq_rhs[-zero_idx]
        ceq_theta <- ceq_theta[-zero_idx]
        # hm, how to handle these? indices no longer match rows of ceq.JAC!
        ceq_linear_idx <- ceq_linear_idx[!ceq_linear_idx %in% zero_idx]
        ceq_nonlinear_idx <- ceq_nonlinear_idx[!ceq_nonlinear_idx %in% zero_idx]
      }
    }
  }
  if (nrow(cin_jac) > 0L) {
    if (all(cin_jac == 0)) {
      cin_jac <- matrix(0, nrow = 0L, ncol = npar)
      cin_rhs <- numeric(0L)
      cin_theta <- numeric(0L)
      cin_linear_idx <- integer(0L)
      cin_nonlinear_idx <- integer(0L)
    } else {
      zero_idx <- which(apply(cin_jac, 1, function(x) all(x == 0)))
      if (length(zero_idx) > 0L) {
        cin_jac <- cin_jac[-zero_idx, , drop = FALSE]
        cin_rhs <- cin_rhs[-zero_idx]
        cin_theta <- cin_theta[-zero_idx]
        # hm, how to handle these? indices no longer match rows of cin.JAC!
        cin_linear_idx <- cin_linear_idx[!cin_linear_idx %in% zero_idx]
        cin_nonlinear_idx <- cin_nonlinear_idx[!cin_nonlinear_idx %in% zero_idx]
      }
    }
  }

  # shortcut flags
  ceq_linear_flag <- length(ceq_linear_idx) > 0L
  ceq_nonlinear_flag <- length(ceq_nonlinear_idx) > 0L
  ceq_flag <- ceq_linear_flag || ceq_nonlinear_flag

  cin_linear_flag <- length(cin_linear_idx) > 0L
  cin_nonlinear_flag <- length(cin_nonlinear_idx) > 0L
  cin_flag <- cin_linear_flag || cin_nonlinear_flag
  if (cin_simple) {
    cin_flag <- FALSE
  }

  # ceq_only_flag <- ceq_flag && !cin_flag
  cin_only_flag <- cin_flag && !ceq_flag

  ceq_linear_only_flag <- (ceq_linear_flag &&
    !ceq_nonlinear_flag &&
    !cin_flag && !cin_simple)

  ceq_simple_only <- ceq_simple && !ceq_flag && !cin_flag
  cin_simple_only <- cin_simple && !ceq_linear_flag

  # additional info if ceq.linear.flag
  if (ceq_linear_only_flag) {
    ## NEW: 18 nov 2014: handle general *linear* constraints
    ##
    ## see Nocedal & Wright (2006) 15.3
    ## - from x to x.red:
    ##       x.red <- MASS::ginv(Q2) %*% (x - Q1 %*% solve(t(R)) %*% b)
    ##   or
    ##       x.red <- as.numeric((x - b %*% qr.coef(QR,diag(npar))) %*% Q2)
    ##
    ## - from x.red to x
    ##       x <- as.numeric(Q1 %*% solve(t(R)) %*% b + Q2 %*% x.red)
    ##   or
    ##       x <- as.numeric(b %*% qr.coef(QR, diag(npar))) +
    ##                       as.numeric(Q2 %*% x.red)
    ##
    ## we write eq.constraints.K = Q2
    ##          eq.constraints.k0 = b %*% qr.coef(QR, diag(npar)))

    # compute range+null space of the jacobian (JAC) of the constraint
    # matrix
    # JAC <- lav_func_jacobian_complex(func = ceq.function,
    #           x = lavpartable$start[lavpartable$free > 0L]
    qr_1 <- qr(t(ceq_jac))
    ran_k <- qr_1$rank
    q_1 <- qr.Q(qr_1, complete = TRUE)
    # Q1 <- Q[,1:ranK, drop = FALSE]         # range space
    # Q2 <- Q[,-seq_len(ranK), drop = FALSE] # null space
    # R <- qr.R(QR)
    ceq_jac_null <- q_1[, -seq_len(ran_k), drop = FALSE]

    if (all(ceq_rhs == 0)) {
      ceq_rhs_null <- numeric(npar)
    } else {
      tmp <- qr.coef(qr_1, diag(npar))
      na_idx <- which(is.na(rowSums(tmp))) # catch NAs
      if (length(na_idx) > 0L) {
        tmp[na_idx, ] <- 0
      }
      ceq_rhs_null <- as.numeric(ceq_rhs %*% tmp)
    }
  } else {
    ceq_jac_null <- matrix(0, 0L, 0L)
    ceq_rhs_null <- numeric(0L)
  }

  # if simple equalities only, create 'K' matrix
  ceq_simple_k <- matrix(0, 0, 0)
  if (ceq_simple_only) {
    n_unco <- max(partable$unco)
    n_free <- max(partable$free)
    ceq_simple_k <- matrix(0, nrow = n_unco, ncol = n_free)
    #####
    #####     FIXME !
    #####
    idx_free <- partable$free[partable$free > 0]
    for (k in 1:n_unco) {
      c <- idx_free[k]
      ceq_simple_k[k, c] <- 1
    }
  }

  # dummy jacobian 'function'
  ceq_jacobian <- function() NULL
  cin_jacobian <- function() NULL


  out <- list(
    def.function = def_function,
    ceq.function = ceq_function,
    ceq.JAC = ceq_jac,
    ceq.jacobian = ceq_jacobian,
    ceq.rhs = ceq_rhs,
    ceq.theta = ceq_theta,
    ceq.linear.idx = ceq_linear_idx,
    ceq.nonlinear.idx = ceq_nonlinear_idx,
    ceq.linear.flag = ceq_linear_flag,
    ceq.nonlinear.flag = ceq_nonlinear_flag,
    ceq.flag = ceq_flag,
    ceq.linear.only.flag = ceq_linear_only_flag,
    ceq.JAC.NULL = ceq_jac_null,
    ceq.rhs.NULL = ceq_rhs_null,
    ceq.simple.only = ceq_simple_only,
    ceq.simple.K = ceq_simple_k,
    cin.function = cin_function,
    cin.JAC = cin_jac,
    cin.jacobian = cin_jacobian,
    cin.rhs = cin_rhs,
    cin.theta = cin_theta,
    cin.linear.idx = cin_linear_idx,
    cin.nonlinear.idx = cin_nonlinear_idx,
    cin.linear.flag = cin_linear_flag,
    cin.nonlinear.flag = cin_nonlinear_flag,
    cin.flag = cin_flag,
    cin.only.flag = cin_only_flag,
    cin.simple.only = cin_simple_only
  )

  out
}

lav_con_linear_idx <- function(func = NULL, npar = NULL) {
  if (is.null(func) || is.null(body(func))) {
    return(integer(0L))
  }

  # compare the jacobian at two (distinct) evaluation points; isolate the
  # RNG state so we do not disturb the global stream (see lav_with_local_seed)
  a0min_a1 <- lav_with_local_seed({
    a0 <- lav_func_jacobian_complex(func = func, x = rnorm(npar)) # point 1
    a1 <- lav_func_jacobian_complex(func = func, x = rnorm(npar)) # point 2
    a0 - a1
  })
  linear <- apply(a0min_a1, 1, function(x) all(x == 0))
  which(linear)
}

lav_con_nonlinear_idx <- function(func = NULL, npar = NULL) {
  if (is.null(func) || is.null(body(func))) {
    return(integer(0L))
  }

  # compare the jacobian at two (distinct) evaluation points; isolate the
  # RNG state so we do not disturb the global stream (see lav_with_local_seed)
  a0min_a1 <- lav_with_local_seed({
    a0 <- lav_func_jacobian_complex(func = func, x = rnorm(npar)) # point 1
    a1 <- lav_func_jacobian_complex(func = func, x = rnorm(npar)) # point 2
    a0 - a1
  })
  linear <- apply(a0min_a1, 1, function(x) all(x == 0))
  which(!linear)
}

# check if the equality constraints are 'simple' (a == b)
lav_con_check_simple <- function(lavmodel = NULL) {
  ones <- (lavmodel@ceq.JAC == 1 | lavmodel@ceq.JAC == -1)
  simple <- all(lavmodel@ceq.rhs == 0) &&
    all(apply(lavmodel@ceq.JAC != 0, 1, sum) == 2) &&
    all(apply(ones, 1, sum) == 2) &&
    length(lavmodel@ceq.nonlinear.idx) == 0

  # TRUE or FALSE
  simple
}

# orthonormal basis 'K' of the equality-constrained (tangent) space, in the
# column space of lav_model_delta(): post-multiplying Delta by K restricts it
# to the directions that satisfy the equality constraints.
#
# This plays the role of eq.constraints.K, but is also available when
# equality constraints coexist with inequality constraints or with parameter
# bounds: in that case BOTH lavmodel@eq.constraints and
# lavmodel@ceq.simple.only are FALSE (they are *packing* flags, meaning
# "only equality constraints"), eq.constraints.K is not computed, and code
# that gates the K-reduction on those flags would silently drop the equality
# constraints (wrong statistics with unchanged df).
#
# Returns NULL when the model has no equality constraints. For nonlinear
# equality constraints the basis is the null space of the Jacobian
# (a local, asymptotically correct linearization).
lav_con_eq_basis <- function(lavmodel = NULL) {
  if (lavmodel@eq.constraints) {
    # only linear equality constraints: precomputed orthonormal null space
    out <- lavmodel@eq.constraints.K
  } else if (lavmodel@ceq.simple.only) {
    # ceq.simple.K is a 0/1 duplication matrix (t(K) %*% K != I);
    # orthonormalize so that K / t(K) round-trips are consistent
    out <- qr.Q(qr(lavmodel@ceq.simple.K))
  } else if (nrow(lavmodel@ceq.JAC) > 0L) {
    # equality constraints together with inequality constraints/bounds
    out <- lav_mat_ortho_complement(t(lavmodel@ceq.JAC))
  } else {
    out <- NULL
  }
  out
}

# classify the inequality constraints at a converged solution: which rows
# are INACTIVE (not binding)? A constraint is only binding when it is
# satisfied with (near) equality AND its Lagrange multiplier exerts actual
# force on the objective gradient (strict complementarity). Both tests are
# needed: an estimate that merely lands within the slack of a bound, with a
# vanishing multiplier, is an interior solution -- treating the constraint
# as a binding equality would (wrongly) collapse its standard error to
# zero, with a discontinuous jump to the free standard error just across
# the slack threshold.
#
# The multiplier is compared on the scale of the force it exerts on the
# gradient, |lambda| * ||jac row||, which is invariant to the scaling of
# the constraint function (for c * g(x) >= 0, lambda scales with 1/c and
# the jacobian row with c). At an interior optimum the (post-hoc)
# multiplier is gradient-level noise, many orders of magnitude below the
# force of even a weakly binding constraint (which is proportional to how
# deep the unconstrained optimum violates the bound).
#
# arguments:
# - con0: the constraint values c(ceq, cin) at the solution
# - lambda: the Lagrange multipliers, aligned with con0
# - jac: the constraint Jacobian, rows aligned with con0
# - cin_flag: logical, TRUE for the inequality rows
lav_con_cin_inactive_idx <- function(con0 = NULL, lambda = NULL,
                                     jac = NULL, cin_flag = NULL,
                                     slack = 1e-05) {
  # measured post-hoc multiplier noise at interior optima is <= ~1e-7
  # (gradient-level), while the force of even a weakly binding constraint
  # is proportional to how deep the unconstrained optimum violates the
  # bound (~1e-4 for a violation of 1e-4); 1e-6 separates the two with a
  # comfortable margin on both sides
  force_tol <- 1e-06
  # clearly interior: strictly satisfied beyond the slack
  interior <- con0 > slack
  # no force: multiplier (in scale-free force units) vanishing; this also
  # covers a (wrong-sided) negative multiplier
  row_norm <- sqrt(rowSums(jac * jac))
  no_force <- (lambda * row_norm) < force_tol
  which(cin_flag & (interior | no_force))
}

lav_con_r2k <- function(lavmodel = NULL) {
  # constraint matrix
  m_r <- NULL
  if (!is.null(lavmodel)) {
    m_r <- lavmodel@ceq.JAC
  }
  stopifnot(!is.null(m_r))

  npar_full <- NCOL(m_r)
  # npar_red <-
  npar_full - NROW(m_r)

  m_k <- diag(npar_full)
  for (i in seq_len(NROW(m_r))) {
    idx1 <- which(m_r[i, ] == 1)
    idx2 <- which(m_r[i, ] == -1)
    m_k[idx2, idx1] <- 1
  }

  # remove redundant columns
  neg_idx <- which(colSums(m_r) < 0)
  m_k <- m_k[, -neg_idx]

  m_k
}

lav_con_lambda_pre <- function(lavobject = NULL, method = "Don") {
  # compute factor 'pre' so that pre %*% g = lambda
  method <- tolower(method)

  # note: drop = FALSE, so a single-row constraint matrix stays a matrix
  r <- lavobject@Model@con.jac[, , drop = FALSE]
  if (is.null(r) || length(r) == 0L) {
    return(numeric(0L))
  }

  info <- lavTech(lavobject, "information.first.order")
  npar <- nrow(info)

  # Don 1985
  if (method == "don") {
    r_plus <- MASS::ginv(r)

    # construct augmented matrix
    z <- rbind(
      cbind(info, t(r)),
      cbind(r, matrix(0, nrow = nrow(r), ncol = nrow(r)))
    )
    z_plus <- MASS::ginv(z)
    p_star <- z_plus[1:npar, 1:npar]
    pre <- t(r_plus) %*% (diag(npar) - info %*% p_star)

    # Bentler EQS manual
  } else if (method == "bentler") {
    info_inv <- solve(info)
    pre <- solve(r %*% info_inv %*% t(r)) %*% r %*% info_inv
  }

  pre
}

# export `lavaan' lav model description to third-party software
#

lav_export <- function(object, target = "lavaan", prefix = "sem",
                      dir_name = "lav_export", export = TRUE, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)

  stopifnot(inherits(object, "lavaan"))
  # check object
  object <- lav_object_check_version(object)
  target <- tolower(target)

  # check for conditional.x = TRUE
  # if(object@Model@conditional.x) {
  #    stop("lavaan ERROR: this function is not (yet)
  #           available if conditional.x = TRUE")
  # }

  ngroups <- object@Data@ngroups

  # 2. create syntax file
  if (target == "lavaan") {
    out <- lav_export_lavaan(object)
  } else if (target == "mplus") {
    # a single combined data file (with GRP/CLUS columns when needed)
    data_file <- paste(prefix, ".", target, ".raw", sep = "")
    out <- lav_export_mplus(object, data_file = data_file)
  } else if (target == "lisrel") {
    out <- lav_export_lisrel(object)
  } else if (target == "eqs") {
    out <- lav_export_eqs(object)
  } else if (target == "openmx") {
    out <- lav_export_openmx(object)
  } else {
    lav_msg_stop(gettextf("target %s has not been implemented yet", target))
  }

  # export to file?
  if (export) {
    if (!dir.exists(dir_name)) {
      dir.create(path = dir_name, recursive = TRUE)
    }
    input_file <- paste(dir_name, "/", prefix, ".", target, ".in", sep = "")
    cat(out, file = input_file, sep = "")

    # write data (if available)
    if (identical(object@Data@data.type, "full")) {
      data_combined <- lav_export_data(object)
      write.table(data_combined,
        file = paste(dir_name, "/", data_file, sep = ""),
        na = "-999999",
        col.names = FALSE, row.names = FALSE, quote = FALSE
      )
    } else if (identical(object@Data@data.type, "moment")) {
      data_1 <- object@SampleStats@cov[[1L]]
      write.table(data_1,
        file = paste(dir_name, "/", data_file, sep = ""),
        na = "-999999",
        col.names = FALSE, row.names = FALSE, quote = FALSE
      )
    } else {
      lav_msg_warn(gettext("no data available"))
    }
    return(invisible(out))
  }

  out
}

# build a single combined data frame from a fitted object, with columns in the
# order expected by the generated input: observed variables, [weight], [GRP],
# [CLUS]. Variable names are sanitized (dots -> underscores).
lav_export_data <- function(object) {
  ngroups <- object@Data@ngroups
  nlevels <- object@Data@nlevels
  # single-level clustered data (cluster= but no level: structure): the cluster
  # column is needed for Mplus TYPE=COMPLEX (see lav_export_mplus_header)
  clustered <- nlevels == 1L && length(object@Data@cluster) > 0L
  ov_names <- lav_pt_vnames(object@ParTable, "ov")
  weight_name <- object@Data@sampling.weights

  parts <- vector("list", ngroups)
  clus_offset <- 0L
  for (g in seq_len(ngroups)) {
    X <- object@Data@X[[g]]
    colnames(X) <- object@Data@ov.names[[g]]
    df <- as.data.frame(X, stringsAsFactors = FALSE)
    eXo <- object@Data@eXo[[g]]
    if (!is.null(eXo) && ncol(eXo) > 0L) {
      colnames(eXo) <- object@Data@ov.names.x[[g]]
      newc <- setdiff(colnames(eXo), colnames(df))
      if (length(newc)) {
        df <- cbind(df, as.data.frame(eXo[, newc, drop = FALSE]))
      }
    }
    # reorder observed columns to match the partable ov order
    df <- df[, ov_names, drop = FALSE]
    if (length(weight_name) && !is.null(object@Data@weights[[g]])) {
      df[[weight_name]] <- object@Data@weights[[g]]
    }
    if (ngroups > 1L) df[["GRP"]] <- g
    if (nlevels > 1L || clustered) {
      # cluster.idx restarts at 1 within each group; offset so that cluster ids
      # are unique across groups (Mplus requires this)
      cidx <- object@Data@Lp[[g]]$cluster.idx[[2L]]
      df[["CLUS"]] <- cidx + clus_offset
      clus_offset <- clus_offset + max(cidx)
    }
    parts[[g]] <- df
  }

  do.call(rbind, parts)
}
lavExport <- lav_alias("lav_export") # synonym #nolint


lav_export_check <- function(lav) {
  if (inherits(lav, "lavaan")) {
    lav <- lav@ParTable
  } else if (is.list(lav)) {
    # nothing to do
  } else {
    lav_msg_stop(gettext("lav must be of class `lavaan' or a parTable"))
  }

  # check syntax
  if (is.null(lav$ustart)) lav$ustart <- lav$est

  # check if free is missing
  if (is.null(lav$free)) lav$free <- rep(0L, length(lav$ustart))

  # check if label is missing
  if (is.null(lav$label)) lav$label <- rep("", length(lav$ustart))

  # check if group is missing
  if (is.null(lav$group)) lav$group <- rep(1L, length(lav$ustart))

  # if eq.id not all zero, create labels instead
  # if(!is.null(lav$eq.id) && !all(lav$eq.id == 0L)) {
  #    lav$label <- paste("p",as.character(lav$eq.id), sep="")
  #    lav$label[lav$label == "p0"] <- ""
  # }

  lav
}

# Regenerate (compact) lavaan model syntax from a fitted lavaan object
# (or a parameter table).
#
# Design goals:
#  - the regenerated syntax, when refitted with the same data and the same
#    fitting options, reproduces an equivalent model (same free/fixed pattern,
#    same equality constraints, same fixed values)
#  - the syntax is as compact as possible: rhs terms that share the same lhs and
#    operator are combined on a single line (eg `f =~ x1 + x2 + x3`)
#  - multiple groups are written using `group:` blocks; multilevel models using
#    `level:` blocks (nested within `group:` blocks if both are present)
lav_export_lavaan <- function(object, ...) {
  # accept a fitted lavaan object or a parameter table
  if (inherits(object, "lavaan")) {
    pt        <- as.data.frame(object@ParTable, stringsAsFactors = FALSE)
    ngroups   <- object@Data@ngroups
    nlevels   <- object@Data@nlevels
  } else if (is.list(object)) {
    pt <- as.data.frame(object, stringsAsFactors = FALSE)
    ngroups   <- if (is.null(pt$group)) 1L else max(pt$group)
    nlevels   <- if (is.null(pt$level)) 1L else max(pt$level)
  } else {
    lav_msg_stop(gettext("object must be of class \"lavaan\" or a parameter table"))
  }
  nn <- length(pt$lhs)

  # ensure the columns we rely on exist
  if (is.null(pt$free))   pt$free   <- rep(0L, nn)
  if (is.null(pt$ustart)) {
    pt$ustart <- if (!is.null(pt$est)) pt$est else rep(as.numeric(NA), nn)
  }
  if (is.null(pt$label))  pt$label  <- rep("", nn)
  if (is.null(pt$plabel)) pt$plabel <- rep("", nn)
  if (is.null(pt$block))  pt$block  <- rep(1L, nn)
  if (is.null(pt$group))  pt$group  <- ifelse(pt$block == 0L, 0L, 1L)
  if (is.null(pt$level))  pt$level  <- ifelse(pt$block == 0L, 0L, 1L)
  pt$label[is.na(pt$label)] <- ""

  # operator categories
  con_op <- c("==", "<", ">", ":=")
  is_con <- pt$op %in% con_op

  # ---- 1. resolve equality classes -> clean display labels -----------------
  display_label <- lav_export_lavaan_labels(pt, is_con)

  # ---- 2. emit the model, block by block -----------------------------------
  body <- character(0L)
  for (g in seq_len(ngroups)) {
    if (ngroups > 1L) {
      # use the numeric block index: group labels may contain characters that
      # are not valid in the block header syntax (eg "Grant-White")
      body <- c(body, "", paste0("group: ", g))
    }
    for (l in seq_len(nlevels)) {
      if (nlevels > 1L) {
        body <- c(body, paste0("level: ", l))
      }
      block_idx <- which(!is_con & pt$group == g & pt$level == l)
      body <- c(body,
                lav_export_lavaan_block(pt, block_idx, display_label))
    }
  }

  # ---- 3. constraints and defined parameters (block-independent) ------------
  con_lines <- lav_export_lavaan_constraints(pt, is_con, display_label)
  if (length(con_lines)) {
    body <- c(body, "", con_lines)
  }

  header <- "# model syntax autogenerated by lavExport()\n"
  out <- paste0(header, paste(body, collapse = "\n"), "\n")
  class(out) <- c("lavaan.character", "character")
  out
}

# format a fixed numeric value compactly but without loss of precision
lav_export_num <- function(x) {
  if (is.na(x)) {
    return("NA")
  }
  out <- formatC(x, digits = 17L, format = "g")
  # round-trip check: use the shortest representation that recovers x
  for (d in 1:16) {
    short <- formatC(x, digits = d, format = "g")
    if (isTRUE(all.equal(as.numeric(short), x))) {
      out <- short
      break
    }
  }
  trimws(out)
}

# determine, for each parameter row, the clean label to display (or "")
# parameters that are constrained equal (via a shared label or via a pairwise
# `==` constraint) get a single shared display label
lav_export_lavaan_labels <- function(pt, is_con) {
  nn <- length(pt$lhs)
  par_idx <- which(!is_con)

  # union-find over parameter rows (indexed by position in pt)
  parent <- seq_len(nn)
  find <- function(i) {
    while (parent[i] != i) {
      parent[i] <<- parent[parent[i]]
      i <- parent[i]
    }
    i
  }
  union <- function(i, j) {
    ri <- find(i)
    rj <- find(j)
    if (ri != rj) parent[rj] <<- ri
  }

  # lookup: plabel -> row, (non-empty) label -> rows
  plab2row <- integer(0L)
  if (any(nzchar(pt$plabel))) {
    idx <- which(nzchar(pt$plabel))
    plab2row <- idx
    names(plab2row) <- pt$plabel[idx]
  }
  resolve <- function(token) {
    # return parameter row(s) referenced by a token (plabel or label)
    if (token %in% names(plab2row)) {
      return(unname(plab2row[token]))
    }
    which(!is_con & pt$label == token)
  }

  # (a) union rows that share the same non-empty label
  lab_tab <- pt$label[par_idx]
  for (lab in unique(lab_tab[nzchar(lab_tab)])) {
    rows <- par_idx[lab_tab == lab]
    if (length(rows) > 1L) {
      for (k in 2:length(rows)) union(rows[1L], rows[k])
    }
  }

  # (b) union rows linked by a pairwise equality constraint between two
  #     parameter references (plabel or label); these are folded into a shared
  #     label and the `==` line itself is dropped later
  token_re <- "^[A-Za-z.][A-Za-z0-9._]*$"
  eq_idx <- which(pt$op == "==")
  for (i in eq_idx) {
    lhs <- pt$lhs[i]
    rhs <- pt$rhs[i]
    if (grepl(token_re, lhs) && grepl(token_re, rhs)) {
      lrow <- resolve(lhs)
      rrow <- resolve(rhs)
      if (length(lrow) && length(rrow)) {
        union(lrow[1L], rrow[1L])
      }
    }
  }

  # build classes
  root <- vapply(seq_len(nn), find, integer(1L))
  display_label <- rep("", nn)

  # collect names already in use (variable names + user labels) to avoid
  # collisions when generating fresh labels
  is_plabel_style <- function(x) grepl("^\\.", x)
  user_lab <- unique(pt$label[!is_con & nzchar(pt$label) &
                              !is_plabel_style(pt$label)])
  reserved <- unique(c(pt$lhs, pt$rhs, user_lab,
                       pt$lhs[pt$op == ":="]))
  gen_counter <- 0L
  gen_name <- function() {
    repeat {
      gen_counter <<- gen_counter + 1L
      cand <- paste0("p", gen_counter)
      if (!cand %in% reserved) {
        reserved <<- c(reserved, cand)
        return(cand)
      }
    }
  }

  for (r in unique(root[par_idx])) {
    members <- which(root == r & !is_con)
    if (length(members) == 0L) next
    # prefer an existing user label
    member_user_lab <- pt$label[members]
    member_user_lab <- member_user_lab[nzchar(member_user_lab) &
                                       !is_plabel_style(member_user_lab)]
    if (length(member_user_lab)) {
      lab <- member_user_lab[1L]
    } else if (length(members) > 1L) {
      lab <- gen_name()
    } else {
      lab <- ""
    }
    display_label[members] <- lab
  }

  display_label
}

# emit the (compact) model lines for one block
lav_export_lavaan_block <- function(pt, block_idx, display_label) {
  if (length(block_idx) == 0L) {
    return(character(0L))
  }

  out <- character(0L)

  # modifier-decorated rhs term for a single row; returns NA to drop the term
  term <- function(i) {
    rhs <- pt$rhs[i]
    lab <- display_label[i]
    if (pt$free[i] == 0L) {
      if (is.na(pt$ustart[i])) {
        return(NA_character_) # eg fixed.x (co)variance/mean: refit re-adds it
      }
      val <- lav_export_num(pt$ustart[i])
      if (nzchar(lab)) {
        # fixed value *and* a referenced label: stack both modifiers
        return(paste0(val, "*", lab, "*", rhs))
      }
      return(paste0(val, "*", rhs))
    }
    if (nzchar(lab)) {
      return(paste0(lab, "*", rhs))
    }
    rhs
  }

  # ordered operator groups for stable, readable output
  op_order <- c("=~", "<~", "~", "~1", "~~", "|", "~*~")
  ops <- intersect(op_order, unique(pt$op[block_idx]))
  ops <- c(ops, setdiff(unique(pt$op[block_idx]), op_order))

  for (op in ops) {
    rows <- block_idx[pt$op[block_idx] == op]
    if (op == "~1") {
      # one line per lhs: `x1 ~ <mod>1`
      for (lhs in unique(pt$lhs[rows])) {
        i <- rows[pt$lhs[rows] == lhs][1L]
        lab <- display_label[i]
        if (pt$free[i] == 0L) {
          if (is.na(pt$ustart[i])) next
          mod <- paste0(lav_export_num(pt$ustart[i]), "*")
          if (nzchar(lab)) mod <- paste0(mod, lab, "*")
        } else {
          mod <- if (nzchar(lab)) paste0(lab, "*") else ""
        }
        out <- c(out, paste0(lhs, " ~ ", mod, "1"))
      }
      next
    }
    # group by lhs and combine rhs terms with `+`
    for (lhs in unique(pt$lhs[rows])) {
      lrows <- rows[pt$lhs[rows] == lhs]
      if (op == "~~") {
        # list the variance (diagonal) before any covariances
        diag_first <- order(pt$rhs[lrows] != lhs)
        lrows <- lrows[diag_first]
      }
      terms <- vapply(lrows, term, character(1L))
      terms <- terms[!is.na(terms)]
      if (length(terms) == 0L) next
      out <- c(out, paste0(lhs, " ", op, " ", paste(terms, collapse = " + ")))
    }
  }

  out
}

# determine the defined parameters (:=) and constraints (==, <, >) to emit, with
# all references to plabels / old labels rewritten to the clean display labels.
# pairwise equality constraints between two parameters are dropped (they are
# folded into shared labels). returns a data.frame with columns lhs, op, rhs.
lav_export_constraints_df <- function(pt, is_con, display_label) {
  con_idx <- which(is_con)
  empty <- data.frame(lhs = character(0L), op = character(0L),
                      rhs = character(0L), stringsAsFactors = FALSE)
  if (length(con_idx) == 0L) {
    return(empty)
  }

  # token translation map: plabel/old-label -> display label
  trans <- character(0L)
  has_disp <- which(nzchar(display_label))
  if (length(has_disp)) {
    if (any(nzchar(pt$plabel[has_disp]))) {
      pidx <- has_disp[nzchar(pt$plabel[has_disp])]
      trans[pt$plabel[pidx]] <- display_label[pidx]
    }
    if (any(nzchar(pt$label[has_disp]))) {
      lidx <- has_disp[nzchar(pt$label[has_disp])]
      trans[pt$label[lidx]] <- display_label[lidx]
    }
  }
  rewrite <- function(expr) {
    if (length(trans) == 0L || is.na(expr) || !nzchar(expr)) {
      return(expr)
    }
    for (from in names(trans)) {
      to <- trans[[from]]
      if (identical(from, to)) next
      pat <- paste0("(?<![A-Za-z0-9._])", lav_export_escape(from),
                    "(?![A-Za-z0-9._])")
      expr <- gsub(pat, to, expr, perl = TRUE)
    }
    expr
  }

  # which `==` rows are pairwise equalities already folded into labels?
  token_re <- "^[A-Za-z.][A-Za-z0-9._]*$"
  is_param_ref <- function(token) {
    token %in% pt$plabel[!is_con] ||
      token %in% pt$label[!is_con & nzchar(pt$label)]
  }

  lhs <- rhs <- op <- character(0L)
  for (i in con_idx) {
    o <- pt$op[i]
    if (o == "==" &&
        grepl(token_re, pt$lhs[i]) && grepl(token_re, pt$rhs[i]) &&
        is_param_ref(pt$lhs[i]) && is_param_ref(pt$rhs[i])) {
      next # folded into a shared label
    }
    op  <- c(op, o)
    lhs <- c(lhs, if (o == ":=") pt$lhs[i] else rewrite(pt$lhs[i]))
    rhs <- c(rhs, rewrite(pt$rhs[i]))
  }
  data.frame(lhs = lhs, op = op, rhs = rhs, stringsAsFactors = FALSE)
}

# emit lavaan-syntax lines for defined parameters and constraints
lav_export_lavaan_constraints <- function(pt, is_con, display_label) {
  con_df <- lav_export_constraints_df(pt, is_con, display_label)
  if (nrow(con_df) == 0L) {
    return(character(0L))
  }
  paste0(con_df$lhs, " ", con_df$op, " ", con_df$rhs)
}

# escape regex metacharacters in a literal token (eg the dots in plabels)
lav_export_escape <- function(x) {
  gsub("([.\\\\+*?\\[\\]^$(){}=!<>|:-])", "\\\\\\1", x, perl = TRUE)
}

lav_export_lisrel <- function(lav) {
  lav <- lav_export_check(lav)
  lav_msg_stop(gettext("this function needs revision"))
}

lav_export_eqs <- function(lav) {
  lav <- lav_export_check(lav)
  lav_msg_stop(gettext("this function needs revision"))
}

lav_export_openmx <- function(lav) {
  lav <- lav_export_check(lav)
  lav_msg_stop(gettext("this function needs revision"))
}

# auxiliary (aux=) variable handling and, more generally, the "extra observed
# variable" channel: observed variables that are read from the data and enter
# the sample statistics, but are NOT part of the user-specified model and never
# show up in the model-implied summary statistics. This channel is carried by
# the lavData@aux / lavData@ov.names.aux slots. It is currently used for:
#
#  - auxiliary variables (aux=), to make the missing-at-random assumption more
#    plausible under missing data (see below); and
#
#  - external instruments for estimator = "IV": instruments supplied with the
#    |~ operator that are not otherwise part of the model. They must be read
#    from the data so that their covariances with the model variables are
#    available to the 2SLS estimator, but they must not enter the model-implied
#    moments / degrees of freedom (see lav_iv_external_names() below and the
#    augmented-covariance handling in lav_sem_miiv.R).
#
# auxiliary variables are observed variables that are NOT part of the
# user-specified model. They are used to make the missing-at-random (MAR)
# assumption more plausible under missing data. Two mechanisms are supported:
#
#  - missing = "two.stage" / "robust.two.stage": the auxiliary variables are
#    included in the first-stage EM run that estimates the saturated moments
#    the model is fitted to (see lav_samplestats.R); the model estimates are
#    therefore affected. (Handled via the lavData@aux channel + the EM code.)
#
#  - missing = "ml" / "fiml" (full-information ML): the auxiliary variables
#    are added to the model as a block of "saturated correlates" (Graham,
#    2003): each auxiliary variable is freely correlated with every other
#    auxiliary variable, with every model observed variable, and has a free
#    mean. The (enlarged) model is then estimated with casewise FIML, so the
#    auxiliary variables affect both the point estimates and the standard
#    errors, while leaving the model degrees of freedom unchanged. The
#    auxiliary parameters are nuisance parameters and are hidden from the
#    default output. (Handled by the functions below.)

# clean up the user-supplied aux= names: drop names that are not found in the
# data, that are model (observed) variables, or that are binary or
# ordered/categorical (not supported). Returns a character vector of the
# (continuous) auxiliary variable names that can be used.
lav_aux_clean_names <- function(aux      = NULL,
                                data     = NULL,
                                ov_names = NULL,
                                ordered  = NULL) {
  if (is.null(aux)) {
    return(character(0L))
  }
  if (!is.character(aux)) {
    lav_msg_stop(gettext("aux= argument must be a character vector."))
  }
  aux <- unique(aux[nchar(aux) > 0L])
  if (length(aux) == 0L) {
    return(character(0L))
  }

  # we need raw data
  if (is.null(data) || !(is.data.frame(data) || is.matrix(data))) {
    return(character(0L))
  }
  data_names <- if (is.data.frame(data)) names(data) else colnames(data)

  # 1) drop aux names that are not found in the data
  bad_idx <- which(!aux %in% data_names)
  if (length(bad_idx) > 0L) {
    lav_msg_warn(gettextf(
      "auxiliary (aux) variable(s) not found in the data and ignored: %s",
      paste(aux[bad_idx], collapse = " ")))
    aux <- aux[-bad_idx]
  }
  if (length(aux) == 0L) {
    return(character(0L))
  }

  # 2) drop aux names that are model (observed) variables
  model_ov <- unique(unlist(ov_names))
  bad_idx <- which(aux %in% model_ov)
  if (length(bad_idx) > 0L) {
    lav_msg_note(gettextf(
      "auxiliary (aux) variable(s) are also model variables and ignored
       as auxiliary: %s", paste(aux[bad_idx], collapse = " ")))
    aux <- aux[-bad_idx]
  }
  if (length(aux) == 0L) {
    return(character(0L))
  }

  # 3) drop binary or ordered/categorical aux variables
  cat_flag <- logical(length(aux))
  for (i in seq_along(aux)) {
    if (!is.null(ordered) && aux[i] %in% ordered) {
      cat_flag[i] <- TRUE
      next
    }
    vv <- if (is.data.frame(data)) data[[aux[i]]] else data[, aux[i]]
    if (is.factor(vv) || is.ordered(vv) || is.character(vv) ||
        is.logical(vv)) {
      cat_flag[i] <- TRUE
    } else {
      # numeric: binary?
      if (length(unique(vv[!is.na(vv)])) <= 2L) {
        cat_flag[i] <- TRUE
      }
    }
  }
  if (any(cat_flag)) {
    lav_msg_warn(gettextf(
      "auxiliary (aux) variable(s) are binary or ordered/categorical;
       this is not supported and they are removed: %s",
      paste(aux[cat_flag], collapse = " ")))
    aux <- aux[!cat_flag]
  }

  aux
}

# append the 'saturated correlates' block for the auxiliary variables to a
# flat model (the parsed-syntax object produced by lavParseModelString). The
# new entries are all free parameters (mod.idx = 0, no modifier):
#   aux_i ~~ aux_j   (auxiliary variances and covariances)
#   aux_i ~ 1        (auxiliary means)
#   aux_i ~~ ov_k    (auxiliary correlated with every model observed variable)
# lavParTable() expands these per group/block, just like ordinary syntax.
lav_aux_satcor_flat <- function(flat_model = NULL,
                                aux        = character(0L),
                                ov_names   = character(0L)) {
  if (length(aux) == 0L) {
    return(flat_model)
  }
  ov <- setdiff(unique(unlist(ov_names)), aux)
  na <- length(aux)

  lhs <- character(0L)
  op  <- character(0L)
  rhs <- character(0L)

  # aux ~~ aux (variances + covariances)
  for (i in seq_len(na)) {
    for (j in i:na) {
      lhs <- c(lhs, aux[i]); op <- c(op, "~~"); rhs <- c(rhs, aux[j])
    }
  }
  # aux ~ 1 (means)
  lhs <- c(lhs, aux); op <- c(op, rep("~1", na)); rhs <- c(rhs, rep("", na))
  # aux ~~ model observed variables
  for (a in aux) {
    for (v in ov) {
      lhs <- c(lhs, a); op <- c(op, "~~"); rhs <- c(rhs, v)
    }
  }
  nn <- length(lhs)

  # append to each column of flat_model (handle the known columns explicitly,
  # pad any other columns by type)
  for (cn in names(flat_model)) {
    val <- switch(cn,
      lhs     = lhs,
      op      = op,
      rhs     = rhs,
      block   = rep(1L, nn),
      mod.idx = rep(0L, nn),
      NULL
    )
    if (is.null(val)) {
      old <- flat_model[[cn]]
      if (is.character(old)) {
        val <- rep("", nn)
      } else if (is.integer(old)) {
        val <- rep(0L, nn)
      } else {
        val <- rep(as.numeric(NA), nn)
      }
    }
    flat_model[[cn]] <- c(flat_model[[cn]], val)
  }
  # modifiers/constraints attributes are unchanged (all new rows are free)
  attr(flat_model, "aux") <- aux

  flat_model
}

# augment the baseline (independence) parameter table when auxiliary variables
# are present: the model observed variables stay mutually uncorrelated (as in
# the independence model), but the auxiliary variables become a saturated block
# (freely correlated with each other and with every model observed variable).
# We start from the ordinary independence partable (which already contains the
# variances and means in the correct variable order -- crucial, so that the
# saturated reference stays aligned with the sample statistics) and merge in the
# free auxiliary covariances. This matches the baseline used by semTools'
# auxiliary() so that CFI/TLI/NFI are computed correctly.
lav_aux_satcor_baseline <- function(indep_partable = NULL,
                                    aux            = character(0L),
                                    ov_names       = character(0L),
                                    ngroups        = 1L) {
  if (length(aux) == 0L) {
    return(indep_partable)
  }
  ov <- setdiff(unique(unlist(ov_names)), aux)
  na <- length(aux)
  blk <- character(0L)
  # auxiliary covariances (off-diagonal; variances already in independence pt)
  if (na > 1L) {
    for (i in 1:(na - 1L)) {
      for (j in (i + 1L):na) {
        blk <- c(blk, paste(aux[i], "~~", aux[j]))
      }
    }
  }
  # auxiliary ~~ model observed variables
  for (a in aux) {
    for (v in ov) {
      blk <- c(blk, paste(a, "~~", v))
    }
  }
  if (length(blk) == 0L) {
    return(indep_partable)
  }

  sat_pt <- lavaanify(paste(blk, collapse = "\n"), ngroups = ngroups)
  sat_pt <- sat_pt[, c("lhs", "op", "rhs", "user", "block", "group")]

  pt_df <- as.data.frame(indep_partable, stringsAsFactors = FALSE)
  merged <- lav_partable_merge(pt_df, sat_pt,
    remove.duplicated = TRUE, warn = FALSE
  )
  n_new <- nrow(merged) - nrow(pt_df)
  if (n_new > 0L) {
    new_rows <- seq_len(n_new) + nrow(pt_df)
    merged$free[new_rows]   <- seq_len(n_new) + max(pt_df$free)
    merged$plabel[new_rows] <- paste0(".p", new_rows, ".")
  }
  as.list(merged)
}

# two-stage estimation with auxiliary variables (Savalei & Bentler, 2009):
# the stage-1 saturated moments are estimated over the augmented set
# [model variables, auxiliary variables]; the model is then fitted to the
# model-variable subset. For the (robust) standard errors, the stage-1
# asymptotic covariance Omega must therefore be computed over the augmented
# set and then reduced to the model-variable moments. This function returns
# the indices, within the augmented moment vector [mu (p_aug),
# vech(Sigma) (p_aug*(p_aug+1)/2)], that correspond to the model-variable
# moments [mu_model, vech(Sigma_model)] (the model variables come first, the
# auxiliary variables last). Selecting this sub-block of Omega gives the
# correct covariance of the model-variable stage-1 estimates (the sub-matrix
# of the inverse information, NOT the inverse of the model-variable
# information).
lav_aux_moment_idx <- function(p_model = 0L, p_aug = 0L) {
  mean_idx <- seq_len(p_model)
  ri <- lav_mat_vech_row_idx(p_aug) # incl. diagonal
  ci <- lav_mat_vech_col_idx(p_aug)
  vech_keep <- which(ri <= p_model & ci <= p_model)
  c(mean_idx, p_aug + vech_keep)
}

# detect 'external' instruments for estimator = "IV": instruments supplied via
# the |~ operator (the rhs of a |~ row) that are NOT part of the model observed
# variables. These are returned so that they can be routed through the
# extra-observed-variable channel (ov_names_aux -> lavData@aux): read from the
# data, but kept out of the model-implied summary statistics. Instruments that
# are already model variables are handled by the ordinary IV machinery (they
# are looked up in the model covariance matrix) and are not returned here.
# 'flat_model' is the parsed-syntax object (with $op/$lhs/$rhs); 'ov_names' are
# the model observed variable names (a vector or a per-group list); 'data' is
# the raw data.frame/matrix (used only to check that the instruments exist).
lav_iv_external_names <- function(flat_model = NULL,
                                  ov_names   = NULL,
                                  data       = NULL) {
  if (is.null(flat_model$op) || !any(flat_model$op == "|~")) {
    return(character(0L))
  }
  iv_idx <- which(flat_model$op == "|~" & nchar(flat_model$rhs) > 0L)
  if (length(iv_idx) == 0L) {
    return(character(0L))
  }
  iv_rhs <- unique(flat_model$rhs[iv_idx])

  # drop instruments that are already model observed variables
  model_ov <- unique(unlist(ov_names))
  ext <- iv_rhs[!iv_rhs %in% model_ov]
  if (length(ext) == 0L) {
    return(character(0L))
  }

  # the remaining instruments must be found in the data
  if (is.null(data) || !(is.data.frame(data) || is.matrix(data))) {
    lav_msg_warn(gettextf(
      "external instrument(s) require raw data and are ignored: %s",
      paste(ext, collapse = " ")))
    return(character(0L))
  }
  data_names <- if (is.data.frame(data)) names(data) else colnames(data)
  bad_idx <- which(!ext %in% data_names)
  if (length(bad_idx) > 0L) {
    lav_msg_warn(gettextf(
      "external instrument(s) not found in the data and ignored: %s",
      paste(ext[bad_idx], collapse = " ")))
    ext <- ext[-bad_idx]
  }

  ext
}

# given a (fitted) lavaan parameter table and the auxiliary variable names,
# return a logical vector marking the rows that belong to the auxiliary
# (saturated-correlates) block (i.e. any row that involves an auxiliary
# variable). Used to hide these nuisance parameters from the default output.
lav_aux_idx <- function(partable = NULL, aux = character(0L)) {
  if (length(aux) == 0L) {
    return(rep(FALSE, length(partable$lhs)))
  }
  (partable$lhs %in% aux) | (partable$rhs %in% aux)
}

# shared computational engines for the two-level (clustered) mvnorm/mvreg
# machinery
#
# the gradient of -2*logl and the cluster-wise scores share (almost) all
# of their algebra; the engines below compute both, controlled by a
# score_mode= switch:
#   - score_mode = FALSE: one 'unit' per cluster SIZE, aggregate (dlogl)
#   - score_mode = TRUE : one 'unit' per cluster, cluster-wise rows (scores)
#
# YR/CC 03 July 2026: first version (refactoring; no change in behavior)


# 0) shared scatter/gather core of the implied <-> 2l conversions
#    (see the representation invariants at the top of lav_mvnorm_cluster.R)
#
#    the sigma blocks of lav_mvn_cl_implied22l/lav_mvreg_cl_implied22l and
#    their inverses only differ in which index sets partition the tilde
#    universe:
#      mvnorm: l1_idx/l2_idx = ov.idx[[1]]/[[2]], z_idx = between.idx,
#              rm_idx = between.idx
#      mvreg : l1_idx/l2_idx = ov.y.idx[[1]]/[[2]], z_idx = between.y.idx,
#              rm_idx = within.x.idx + between.x.idx (+ between.y.idx)

# implied -> 2l: scatter the per-level covariance matrices into the tilde
# universe and gather the y/z blocks
lav_cl_sigma_22l <- function(sigma_w = NULL,
                             sigma_b = NULL,
                             p_tilde = NULL,
                             l1_idx = NULL,
                             l2_idx = NULL,
                             z_idx = NULL,
                             rm_idx = NULL) {
  # scatter into the tilde universe
  sigma_w_tilde <- matrix(0, p_tilde, p_tilde)
  sigma_w_tilde[l1_idx, l1_idx] <- sigma_w
  sigma_b_tilde <- matrix(0, p_tilde, p_tilde)
  sigma_b_tilde[l2_idx, l2_idx] <- sigma_b

  # gather the y/z blocks
  y_idx <- seq_len(p_tilde)
  if (length(rm_idx) > 0L) {
    y_idx <- y_idx[-rm_idx]
  }
  if (length(z_idx) > 0L) {
    sigma_zz <- sigma_b_tilde[z_idx, z_idx, drop = FALSE]
    sigma_yz <- sigma_b_tilde[y_idx, z_idx, drop = FALSE]
  } else {
    sigma_zz <- matrix(0, 0L, 0L)
    sigma_yz <- matrix(0, length(y_idx), 0L)
  }

  list(
    sigma.w = sigma_w_tilde[y_idx, y_idx, drop = FALSE],
    sigma.b = sigma_b_tilde[y_idx, y_idx, drop = FALSE],
    sigma.zz = sigma_zz, sigma.yz = sigma_yz
  )
}

# 2l -> implied: rebuild the level-2 covariance matrix from its y/z blocks
# (the within matrix needs no reassembly: it is the y block itself)
lav_cl_sigma_2l2 <- function(sigma_b = NULL,
                             sigma_zz = NULL,
                             sigma_yz = NULL,
                             p_tilde = NULL,
                             y_idx = NULL,
                             z_idx = NULL,
                             l2_idx = NULL) {
  sigma_b_tilde <- matrix(0, p_tilde, p_tilde)
  sigma_b_tilde[y_idx, y_idx] <- sigma_b
  if (length(z_idx) > 0L) {
    sigma_b_tilde[y_idx, z_idx] <- sigma_yz
    sigma_b_tilde[z_idx, y_idx] <- t(sigma_yz)
    sigma_b_tilde[z_idx, z_idx] <- sigma_zz
  }
  sigma_b_tilde[l2_idx, l2_idx, drop = FALSE]
}

# vech/vec positions (in the tilde vech space) of the per-level and y/z
# blocks; shared by the scores/information rearrangement code below
lav_cl_vech_blocks <- function(p_tilde = NULL,
                               l1_idx = NULL,
                               l2_idx = NULL,
                               z_idx = integer(0L)) {
  b_tilde <- lav_mat_vech_rev(seq_len(p_tilde * (p_tilde + 1) / 2))
  out <- list(
    l1 = lav_mat_vech(b_tilde[l1_idx, l1_idx, drop = FALSE]),
    l2 = lav_mat_vech(b_tilde[l2_idx, l2_idx, drop = FALSE])
  )
  if (length(z_idx) > 0L) {
    out$yz <- lav_mat_vec(b_tilde[l1_idx, z_idx, drop = FALSE])
    out$zz <- lav_mat_vech(b_tilde[z_idx, z_idx, drop = FALSE])
  }
  out
}


# 1) per-cluster-size cache of Sigma.j = nj*Sigma.b.z + Sigma.w quantities
#
#    Sigma.j only depends on the cluster SIZE nj, so the scores engine
#    (looping over clusters) can reuse one inverse per unique size
lav_mvn_cl_sigma_j_cache <- function(cluster_sizes = NULL,
                                     sigma_b_z = NULL,
                                     sigma_w_1 = NULL,
                                     sigma_yz_zi = NULL,
                                     sinv_method = "eigen") {
  lapply(cluster_sizes, function(nj) {
    sigma_j <- (nj * sigma_b_z) + sigma_w_1
    sigma_j_inv <- lav_mat_sym_inverse(
      s = sigma_j,
      logdet = FALSE, sinv_method = sinv_method
    )
    if (!is.null(sigma_yz_zi)) {
      sigma_ji_yz_zi <- sigma_j_inv %*% sigma_yz_zi
      list(
        sigma_j_inv = sigma_j_inv,
        sigma_ji_yz_zi = sigma_ji_yz_zi,
        sigma_zi_zy_ji = t(sigma_ji_yz_zi)
      )
    } else {
      list(sigma_j_inv = sigma_j_inv)
    }
  })
}

# 2) rearrange cluster-wise score blocks (g_muy, g_muz, g_sigma_*) to the
#    Mu.W, vech(Sigma.W), Mu.B, vech(Sigma.B) column layout
#    (shared by the complete-data and missing-data scores)
lav_mvn_cl_scores_2implied <- function(lp = NULL,
                                       g_muy = NULL,
                                       g_muz = NULL,
                                       g_sigma_w = NULL,
                                       g_sigma_b = NULL,
                                       g_sigma_yz = NULL,
                                       g_sigma_zz = NULL) {
  nclusters <- NROW(g_muy)
  between_idx <- lp$between.idx[[2]]
  ov_idx <- lp$ov.idx
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))

  # Mu.W (for within-only)
  mu_w_tilde <- matrix(0, nclusters, p_tilde)
  mu_w_tilde[, ov_idx[[1]]] <- g_muy
  mu_w_tilde[, lp$both.idx[[2]]] <- 0 # ZERO!!!
  mu_w <- mu_w_tilde[, ov_idx[[1]], drop = FALSE]

  # Mu.B
  mu_b_tilde <- matrix(0, nclusters, p_tilde)
  mu_b_tilde[, ov_idx[[1]]] <- g_muy
  if (length(between_idx) > 0L) {
    mu_b_tilde[, between_idx] <- g_muz
  }
  mu_b <- mu_b_tilde[, ov_idx[[2]], drop = FALSE]

  # Sigma.W
  sigma_w <- g_sigma_w

  # Sigma.B
  p_tilde_star <- p_tilde * (p_tilde + 1) / 2
  vidx <- lav_cl_vech_blocks(
    p_tilde = p_tilde, l1_idx = ov_idx[[1]], l2_idx = ov_idx[[2]],
    z_idx = between_idx
  )

  sigma_b_tilde <- matrix(0, nclusters, p_tilde_star)
  sigma_b_tilde[, vidx$l1] <- g_sigma_b
  if (length(between_idx) > 0L) {
    sigma_b_tilde[, vidx$yz] <- g_sigma_yz
    sigma_b_tilde[, vidx$zz] <- g_sigma_zz
  }
  sigma_b <- sigma_b_tilde[, vidx$l2, drop = FALSE]

  cbind(mu_w, sigma_w, mu_b, sigma_b)
}

# 3) zero out the rows/cols of a (mu.w, vech(sigma.w), mu.b, vech(sigma.b))
#    information matrix that correspond to fixed exogenous variables
lav_mvn_cl_zero_x_idx <- function(information,
                                  lp = NULL,
                                  mu_w = NULL,
                                  mu_b = NULL,
                                  x_idx = NULL) {
  if (length(x_idx) == 0L) {
    return(information)
  }

  nw <- length(as.vector(mu_w))
  nw_star <- nw * (nw + 1) / 2
  nb <- length(as.vector(mu_b))
  ov_idx <- lp$ov.idx

  x_idx_w <- which(ov_idx[[1]] %in% x_idx)
  if (length(x_idx_w) > 0L) {
    xw_idx <- c(
      x_idx_w,
      nw + lav_mat_vech_which_idx(n = nw, idx = x_idx_w)
    )
  } else {
    xw_idx <- integer(0L)
  }
  x_idx_b <- which(ov_idx[[2]] %in% x_idx)
  if (length(x_idx_b) > 0L) {
    xb_idx <- c(
      x_idx_b,
      nb + lav_mat_vech_which_idx(n = nb, idx = x_idx_b)
    )
  } else {
    xb_idx <- integer(0L)
  }

  all_idx <- c(xw_idx, nw + nw_star + xb_idx)

  information[all_idx, ] <- 0
  information[, all_idx] <- 0

  information
}

# 4) the per-cluster-size Omega.j loop of the expected information:
#    sum over cluster sizes of n_s * t(Delta.j) I(Omega.j) Delta.j
#    (shared by lav_mvn_cl_info_expected and _info_expected_delta)
lav_mvn_cl_info_omega_j_loop <- function(lp = NULL,
                                         sigma_w_1 = NULL,
                                         sigma_b_1 = NULL,
                                         sigma_zz = NULL,
                                         sigma_yz = NULL,
                                         delta_w_tilde = NULL,
                                         delta_b_tilde = NULL) {
  cluster_sizes <- lp$cluster.sizes[[2]]
  ncluster_sizes <- lp$ncluster.sizes[[2]]
  n_s <- lp$cluster.size.ns[[2]]
  between_idx <- lp$between.idx[[2]]
  ov_idx <- lp$ov.idx
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))
  npar <- NCOL(delta_w_tilde)

  information_j <- matrix(0, npar, npar)
  for (clz in seq_len(ncluster_sizes)) {
    # cluster size
    nj <- cluster_sizes[clz]

    # Delta.j -- changes per cluster(size)
    # this is why we can not write info = t(delta) info.sat delta
    delta_j <- delta_b_tilde + 1 / nj * delta_w_tilde

    # compute Sigma.j
    sigma_j <- sigma_w_1 + nj * sigma_b_1
    if (length(between_idx) > 0L) {
      omega_j <- matrix(0, p_tilde, p_tilde)
      omega_j[-between_idx, -between_idx] <- 1 / nj * sigma_j
      omega_j[-between_idx, between_idx] <- sigma_yz
      omega_j[between_idx, -between_idx] <- t(sigma_yz)
      omega_j[between_idx, between_idx] <- sigma_zz
    } else {
      omega_j <- 1 / nj * sigma_j
    }
    omega_j_inv <- solve(omega_j)

    i11_j <- omega_j_inv
    i22_j <- lav_mvn_kron_dup_half(omega_j_inv)
    i_j <- lav_mat_bdiag(i11_j, i22_j)
    info_j <- t(delta_j) %*% i_j %*% delta_j

    information_j <- information_j + n_s[clz] * info_j
  }

  information_j
}

# 5) numerical observed information: jacobian of an analytic gradient
#    function (shared by the complete-data and missing-data versions)
#
#    dlogl_fn is called with (lp, mu_w, sigma_w, mu_b, sigma_b, sinv_method,
#    return_list = FALSE) plus the extra arguments in dlogl_args, and must
#    return the gradient of -2*logl
lav_mvn_cl_info_obs_engine <- function(lp = NULL,
                                       mu_w = NULL,
                                       sigma_w = NULL,
                                       mu_b = NULL,
                                       sigma_b = NULL,
                                       x_idx = integer(0L),
                                       sinv_method = "eigen",
                                       dlogl_fn = NULL,
                                       dlogl_args = list()) {
  nw <- length(as.vector(mu_w))
  nw_star <- nw * (nw + 1) / 2
  nb <- length(as.vector(mu_b))
  nb_star <- nb * (nb + 1) / 2

  ov_idx <- lp$ov.idx
  p_tilde <- length(unique(c(ov_idx[[1]], ov_idx[[2]])))

  # Mu.W (for within-only)
  mu_w_tilde <- numeric(p_tilde)
  mu_w_tilde[ov_idx[[1]]] <- mu_w

  # local function -- gradient
  grad <- function(x) {
    # Mu.W (for within-only)
    mu_w_tilde2 <- numeric(p_tilde)
    mu_w_tilde2[ov_idx[[1]]] <- x[1:nw]
    mu_w_tilde2[lp$both.idx[[2]]] <- mu_w_tilde[lp$both.idx[[2]]]
    mu_w2 <- mu_w_tilde2[ov_idx[[1]]]

    sigma_w2 <- lav_mat_vech_rev(x[nw + 1:nw_star])
    mu_b2 <- x[nw + nw_star + 1:nb]
    sigma_b2 <- lav_mat_vech_rev(x[nw + nw_star + nb + 1:nb_star])

    dx <- do.call(dlogl_fn, c(
      list(
        lp = lp, mu_w = mu_w2, sigma_w = sigma_w2,
        mu_b = mu_b2, sigma_b = sigma_b2,
        return_list = FALSE, sinv_method = sinv_method
      ),
      dlogl_args
    ))

    # dx is for -2*logl
    -1 / 2 * dx
  }

  # start.x
  start_x <- c(
    as.vector(mu_w), lav_mat_vech(sigma_w),
    as.vector(mu_b), lav_mat_vech(sigma_b)
  )

  # total information
  information <- -1 * numDeriv::jacobian(func = grad, x = start_x)

  # unit information
  information <- information / lp$nclusters[[2]]

  # if x.idx, set rows/cols to zero
  lav_mvn_cl_zero_x_idx(information,
    lp = lp, mu_w = mu_w, mu_b = mu_b,
    x_idx = x_idx
  )
}


# 6) gradient/scores engine, complete data
#
#    score_mode = FALSE: gradient of -2*logl wrt Mu.W, vech(Sigma.W),
#                        Mu.B, vech(Sigma.B) (loop over cluster sizes)
#    score_mode = TRUE : cluster-wise scores, one row per cluster
lav_mvn_cl_grad_engine <- function(ylp = NULL,
                                   y1 = NULL,
                                   lp = NULL,
                                   mu_w = NULL,
                                   sigma_w = NULL,
                                   mu_b = NULL,
                                   sigma_b = NULL,
                                   score_mode = FALSE,
                                   return_list = FALSE,
                                   sinv_method = "eigen") {
  # map implied to 2l matrices
  out <- lav_mvn_cl_implied22l(
    lp = lp, mu_w = mu_w, mu_b = mu_b,
    sigma_w = sigma_w, sigma_b = sigma_b
  )
  mu_y <- out$mu.y
  mu_z <- out$mu.z
  sigma_w_1 <- out$sigma.w
  sigma_b_1 <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # Lp
  nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]
  cluster_idx <- lp$cluster.idx[[2]]
  between_idx <- lp$between.idx[[2]]
  cluster_sizes <- lp$cluster.sizes[[2]]
  ncluster_sizes <- lp$ncluster.sizes[[2]]
  cluster_size_ns <- lp$cluster.size.ns[[2]]

  # Y1 (pooled-within sample statistics)
  if (length(between_idx) > 0L) {
    s_pw <- ylp[[2]]$Sigma.W[-between_idx, -between_idx, drop = FALSE]
  } else {
    s_pw <- ylp[[2]]$Sigma.W
  }

  # units: cluster sizes (gradient) or clusters (scores)
  if (score_mode) {
    nunits <- nclusters
    unit_nj <- cluster_size
    unit_clz <- match(cluster_size, cluster_sizes)

    # Y1 within part (centered)
    if (length(between_idx) > 0L) {
      y1w <- y1[, -between_idx, drop = FALSE]
    } else {
      y1w <- y1
    }
    y1w_cm <- t(t(y1w) - mu_y)

    # Y2
    y2 <- ylp[[2]]$Y2
    # NOTE: ORDER mu.b must match Y2
    mu_b_2 <- numeric(ncol(y2))
    if (length(between_idx) > 0L) {
      mu_b_2[-between_idx] <- mu_y
      mu_b_2[between_idx] <- mu_z
    } else {
      mu_b_2 <- mu_y
    }
    y2_cm <- t(t(y2) - mu_b_2)
  } else {
    nunits <- ncluster_sizes
    unit_nj <- cluster_sizes
    unit_clz <- seq_len(ncluster_sizes)

    # Y2 samplestats (aggregated per cluster size)
    cov_d <- ylp[[2]]$cov.d
    mean_d <- ylp[[2]]$mean.d
  }

  # common parts:
  sigma_w_inv <- lav_mat_sym_inverse(
    s = sigma_w_1,
    logdet = FALSE, sinv_method = sinv_method
  )

  # unit-wise gradient blocks
  g_muy <- matrix(0, nunits, length(mu_y))
  g_sigma_w_1 <- matrix(0, nunits, length(lav_mat_vech(sigma_w_1)))
  g_sigma_b_1 <- matrix(0, nunits, length(lav_mat_vech(sigma_b_1)))
  g_muz <- matrix(0, nunits, length(mu_z))
  g_sigma_zz_1 <- matrix(0, nunits, length(lav_mat_vech(sigma_zz)))
  g_sigma_yz_1 <- matrix(0, nunits, length(lav_mat_vec(sigma_yz)))

  if (length(between_idx) > 0L) {
    sigma_zz_inv <- lav_mat_sym_inverse(
      s = sigma_zz,
      logdet = FALSE, sinv_method = sinv_method
    )
    sigma_yz_zi <- sigma_yz %*% sigma_zz_inv
    sigma_zi_zy <- t(sigma_yz_zi)
    sigma_b_z <- sigma_b_1 - sigma_yz %*% sigma_zi_zy

    # Sigma.j quantities per unique cluster size
    cache <- lav_mvn_cl_sigma_j_cache(
      cluster_sizes = cluster_sizes,
      sigma_b_z = sigma_b_z, sigma_w_1 = sigma_w_1,
      sigma_yz_zi = sigma_yz_zi, sinv_method = sinv_method
    )

    b_idx <- seq_along(between_idx) # z-first layout (gradient mode)
    for (u in seq_len(nunits)) {
      # cluster size
      nj <- unit_nj[u]

      # Sigma.j quantities (cached per cluster size)
      cj <- cache[[unit_clz[u]]]
      sigma_j_inv <- cj$sigma_j_inv
      sigma_ji_yz_zi <- cj$sigma_ji_yz_zi
      sigma_zi_zy_ji <- cj$sigma_zi_zy_ji

      # level-2 vectors + crossproducts
      if (score_mode) {
        yc <- y2_cm[u, -between_idx]
        zc <- y2_cm[u, between_idx]

        y2yc <- tcrossprod(y2_cm[u, ])
        y2yc_zz <- y2yc[between_idx, between_idx, drop = FALSE]
        y2yc_yz <- y2yc[-between_idx, between_idx, drop = FALSE]
        y2yc_yy <- y2yc[-between_idx, -between_idx, drop = FALSE]
      } else {
        zyc <- mean_d[[u]] - c(mu_z, mu_y)
        yc <- zyc[-b_idx]
        zc <- zyc[b_idx]

        y2yc <- (cov_d[[u]] + tcrossprod(mean_d[[u]] - c(mu_z, mu_y)))
        y2yc_zz <- y2yc[b_idx, b_idx, drop = FALSE]
        y2yc_yz <- y2yc[-b_idx, b_idx, drop = FALSE]
        y2yc_yy <- y2yc[-b_idx, -b_idx, drop = FALSE]
      }

      # common parts
      j_yzj <- nj * (sigma_j_inv %*%
        (sigma_yz_zi %*% y2yc_zz %*% t(sigma_yz_zi)
          - y2yc_yz %*% t(sigma_yz_zi)
          - t(y2yc_yz %*% t(sigma_yz_zi)) + y2yc_yy)
        %*% sigma_j_inv)

      z1 <- y2yc_zz %*% t(sigma_ji_yz_zi) %*% sigma_yz
      yz1 <- t(y2yc_yz) %*% sigma_j_inv %*% sigma_yz


      # Mu.Z
      g_muz[u, ] <- -2 * as.numeric(
        (sigma_zz_inv + nj * (sigma_zi_zy_ji %*% sigma_yz_zi)) %*% zc
          - nj * sigma_zi_zy_ji %*% yc
      )

      # MU.Y
      g_muy[u, ] <- 2 * nj * as.numeric(zc %*% sigma_zi_zy_ji -
        yc %*% sigma_j_inv)

      # SIGMA.W
      if (score_mode) {
        # data within for the cluster (centered by mu.y)
        y1m <- y1w_cm[cluster_idx == u, , drop = FALSE]
        g_sigma_w <- ((nj - 1) * sigma_w_inv
          - sigma_w_inv %*% (crossprod(y1m) - nj * y2yc_yy) %*% sigma_w_inv
          + sigma_j_inv - j_yzj)
      } else {
        # between part only; the pooled-within part is added below
        g_sigma_w <- sigma_j_inv - j_yzj
      }
      g_sigma_w_1[u, ] <- lav_mat_vech_dd(g_sigma_w)

      # SIGMA.B
      g_sigma_b <- nj * (sigma_j_inv - j_yzj)
      g_sigma_b_1[u, ] <- lav_mat_vech_dd(g_sigma_b)

      # SIGMA.ZZ
      g_sigma_zz <- (sigma_zz_inv + nj * sigma_zz_inv %*% (
        t(sigma_yz) %*% (sigma_j_inv - j_yzj) %*% sigma_yz
          - (1 / nj * y2yc_zz + t(z1) + z1 - t(yz1) - yz1)) %*%
        sigma_zz_inv)
      g_sigma_zz_1[u, ] <- lav_mat_vech_dd(g_sigma_zz)

      # SIGMA.ZY
      g_sigma_yz <- 2 * nj * (
        (sigma_j_inv %*%
          (sigma_yz_zi %*% y2yc_zz - sigma_yz - y2yc_yz)
          + j_yzj %*% sigma_yz) %*% sigma_zz_inv)

      g_sigma_yz_1[u, ] <- lav_mat_vec(g_sigma_yz)
    }
  # between.idx
  } else { # no level-2 variables

    # Sigma.j quantities per unique cluster size
    cache <- lav_mvn_cl_sigma_j_cache(
      cluster_sizes = cluster_sizes,
      sigma_b_z = sigma_b_1, sigma_w_1 = sigma_w_1,
      sinv_method = sinv_method
    )

    for (u in seq_len(nunits)) {
      # cluster size
      nj <- unit_nj[u]

      # Sigma.j (cached per cluster size)
      sigma_j_inv <- cache[[unit_clz[u]]]$sigma_j_inv

      # level-2 vectors + crossproducts
      if (score_mode) {
        yc <- y2_cm[u, ]
        y2yc_yy <- tcrossprod(y2_cm[u, ])
      } else {
        yc <- mean_d[[u]] - mu_y
        y2yc_yy <- (cov_d[[u]] + tcrossprod(mean_d[[u]] - mu_y))
      }

      # common part
      j_yyj <- nj * sigma_j_inv %*% y2yc_yy %*% sigma_j_inv

      # MU.Y
      g_muy[u, ] <- -2 * nj * as.numeric(yc %*% sigma_j_inv)

      # SIGMA.W
      if (score_mode) {
        # data within for the cluster (centered by mu.y)
        y1m <- y1w_cm[cluster_idx == u, , drop = FALSE]
        g_sigma_w <- ((nj - 1) * sigma_w_inv
          - sigma_w_inv %*% (crossprod(y1m) - nj * y2yc_yy) %*% sigma_w_inv
          + sigma_j_inv - j_yyj)
      } else {
        # between part only; the pooled-within part is added below
        g_sigma_w <- sigma_j_inv - j_yyj
      }
      g_sigma_w_1[u, ] <- lav_mat_vech_dd(g_sigma_w)

      # SIGMA.B
      g_sigma_b <- nj * (sigma_j_inv - j_yyj)
      g_sigma_b_1[u, ] <- lav_mat_vech_dd(g_sigma_b)
    }
  }

  # scores: rearrange columns to Mu.W, Mu.B, Sigma.W, Sigma.B
  if (score_mode) {
    return(lav_mvn_cl_scores_2implied(
      lp = lp,
      g_muy = g_muy, g_muz = g_muz,
      g_sigma_w = g_sigma_w_1, g_sigma_b = g_sigma_b_1,
      g_sigma_yz = g_sigma_yz_1, g_sigma_zz = g_sigma_zz_1
    ))
  }

  # gradient: weighted sum over the cluster-size groups

  # level-1
  d_mu_y <- colSums(g_muy * cluster_size_ns)
  d_sigma_w1 <- lav_mat_vech_rev(colSums(g_sigma_w_1 *
    cluster_size_ns))
  d_sigma_b <- lav_mat_vech_rev(colSums(g_sigma_b_1 *
    cluster_size_ns))

  # level-2
  if (length(between_idx) > 0L) {
    d_mu_z <- colSums(g_muz * cluster_size_ns)
    d_sigma_zz <- lav_mat_vech_rev(colSums(g_sigma_zz_1 *
      cluster_size_ns))
    d_sigma_yz <- matrix(
      colSums(g_sigma_yz_1 * cluster_size_ns),
      nrow(sigma_yz), ncol(sigma_yz)
    )
  } else {
    d_mu_z <- numeric(0L)
    d_sigma_zz <- matrix(0, 0L, 0L)
    d_sigma_yz <- matrix(0, 0L, 0L)
  }

  # Sigma.W (bis): pooled-within part
  d_sigma_w2 <- (lp$nclusters[[1]] - nclusters) * (sigma_w_inv
  - sigma_w_inv %*% s_pw %*% sigma_w_inv)
  tmp <- d_sigma_w2 * 2
  diag(tmp) <- diag(d_sigma_w2)
  d_sigma_w2 <- tmp

  d_sigma_w <- d_sigma_w1 + d_sigma_w2

  # rearrange
  dout <- lav_mvn_cl_2l2implied(
    lp = lp,
    sigma_w = d_sigma_w, sigma_b = d_sigma_b,
    sigma_yz = d_sigma_yz, sigma_zz = d_sigma_zz,
    mu_y = d_mu_y, mu_z = d_mu_z
  )

  if (return_list) {
    dout
  } else {
    c(
      dout$Mu.W, lav_mat_vech(dout$Sigma.W),
      dout$Mu.B, lav_mat_vech(dout$Sigma.B)
    )
  }
}


# 7) gradient/scores engine, missing data (FIML)
#
#    score_mode = FALSE: gradient of -2*logl wrt Mu.W, vech(Sigma.W),
#                        Mu.B, vech(Sigma.B)
#    score_mode = TRUE : cluster-wise scores, one row per cluster
#
#    (the gradient equals the column sums of the cluster-wise scores;
#     both used to be separate, almost identical, functions)
lav_mvn_cl_mi_grad_engine <- function(y1 = NULL,
                                      y2 = NULL,
                                      lp = NULL,
                                      mp = NULL,
                                      mu_w = NULL,
                                      sigma_w = NULL,
                                      mu_b = NULL,
                                      sigma_b = NULL,
                                      score_mode = FALSE,
                                      return_list = FALSE,
                                      sinv_method = "eigen") {
  # map implied to 2l matrices
  out <- lav_mvn_cl_implied22l(
    lp = lp, mu_w = mu_w, mu_b = mu_b,
    sigma_w = sigma_w, sigma_b = sigma_b
  )
  mu_y <- out$mu.y
  mu_z <- out$mu.z
  sigma_w_1 <- out$sigma.w
  sigma_b_1 <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # Lp
  nclusters <- lp$nclusters[[2]]
  between_idx <- lp$between.idx[[2]]
  cluster_idx <- lp$cluster.idx[[2]]
  both_idx <- lp$both.idx[[2]]

  # sigma.w
  sigma_w_inv <- solve.default(sigma_w_1)
  sigma_b_1 <- sigma_b_1[both_idx, both_idx, drop = FALSE] # only both part

  # y
  ny <- ncol(sigma_w_1)
  if (length(between_idx) > 0L) {
    y1w <- y1[, -between_idx, drop = FALSE]
  } else {
    y1w <- y1
  }
  y1w_c <- t(t(y1w) - mu_y)
  pij_1 <- matrix(0, nrow(y1w_c), ny)

  # z
  nz <- length(between_idx)
  if (nz > 0L) {
    z <- y2[, between_idx, drop = FALSE]
    z_c <- t(t(z) - mu_z)
    sigma_yz <- sigma_yz[both_idx, , drop = FALSE] # only both part
    sigma_zy <- t(sigma_yz)
    sigma_zz_inv <- solve.default(sigma_zz)
    sigma_zi_zy <- sigma_zz_inv %*% sigma_zy
    sigma_b_z <- sigma_b_1 - sigma_yz %*% sigma_zi_zy
    gz <- z_c %*% sigma_zz_inv # for complete cases only
  }

  # containters per cluster
  alist_1 <- rep(list(matrix(0, ny, ny)), nclusters)

  # cluster-wise gradient blocks
  g_muy <- matrix(0, nclusters, length(mu_y))
  g_sigma_w <- matrix(0, nclusters, length(lav_mat_vech(sigma_w_1)))
  g_sigma_b <- matrix(0, nclusters, length(lav_mat_vech(out$sigma.b)))
  g_muz <- matrix(0, nclusters, length(mu_z))
  g_sigma_zz <- matrix(0, nclusters, length(lav_mat_vech(sigma_zz)))
  g_sigma_yz <- matrix(0, nclusters, length(lav_mat_vec(out$sigma.yz)))

  # Z per missing pattern
  if (nz > 0L) {
    zp_1 <- mp$Zp
    zpat2j <- integer(nclusters) # which pattern per cluster
    sigma_b_z_1 <- vector("list", length = zp_1$npatterns + 1L) # +1 for empty
    zizy_1 <- rep(list(matrix(
      0, nrow(sigma_zy),
      ncol(sigma_zy)
    )), zp_1$npatterns + 1L)
    zip_1 <- rep(list(matrix(
      0, nrow(sigma_zz),
      ncol(sigma_zz)
    )), zp_1$npatterns + 1L)
    for (p in seq_len(zp_1$npatterns)) {
      z_na_idx <- which(!zp_1$pat[p, ])
      j_idx <- zp_1$case.idx[[p]] # cluster indices with this pattern
      zpat2j[j_idx] <- p

      if (length(z_na_idx) > 0L) {
        zp_inv <- lav_mat_sym_inverse_update(
          s_inv = sigma_zz_inv, rm_idx = z_na_idx,
          logdet = FALSE
        )
        zip_1[[p]][-z_na_idx, -z_na_idx] <- zp_inv
        gz[j_idx, -z_na_idx] <- z_c[j_idx, -z_na_idx] %*% zp_inv
        z_g_zy <- zp_inv %*% sigma_zy[-z_na_idx, , drop = FALSE]
        zizy_1[[p]][-z_na_idx, ] <-
          zp_inv %*% sigma_zy[-z_na_idx, , drop = FALSE]
        yziy <- sigma_yz[, -z_na_idx, drop = FALSE] %*% z_g_zy
        sigma_b_z_1[[p]] <- (sigma_b_1 - yziy)
      } else {
        # complete case
        zizy_1[[p]] <- sigma_zi_zy
        zip_1[[p]] <- sigma_zz_inv
        sigma_b_z_1[[p]] <- sigma_b_z
      }
    } # p

    # add empty patterns (if any)
    if (length(zp_1$empty.idx) > 0L) {
      zpat2j[zp_1$empty.idx] <- p + 1L
      sigma_b_z_1[[p + 1L]] <- sigma_b_1
    }

    gz[is.na(gz)] <- 0
    gj <- gz %*% sigma_zy
  }

  # Y per missing pattern
  wip <- rep(list(matrix(0, ny, ny)), mp$npatterns)
  mpi <- integer(nrow(y1))
  for (p in seq_len(mp$npatterns)) {
    na_idx <- which(!mp$pat[p, ])
    j_idx <- mp$j.idx[[p]]
    j1_idx <- mp$j1.idx[[p]]
    tab <- integer(nclusters)
    tab[j1_idx] <- mp$j.freq[[p]]

    if (length(na_idx) > 0L) {
      mpi[mp$case.idx[[p]]] <- p
      wp_inv <- lav_mat_sym_inverse_update(
        s_inv = sigma_w_inv, rm_idx = na_idx,
        logdet = FALSE
      )
      wip[[p]][-na_idx, -na_idx] <- wp_inv
      pij_1[mp$case.idx[[p]], -na_idx] <-
        y1w_c[mp$case.idx[[p]], -na_idx] %*% wp_inv

      for (j in j1_idx) {
        alist_1[[j]] <-
          alist_1[[j]] + (wip[[p]] * tab[j])
      }
    } else {
      # complete case
      pij_1[mp$case.idx[[p]], ] <-
        y1w_c[mp$case.idx[[p]], ] %*% sigma_w_inv
      wip[[p]] <- sigma_w_inv
      for (j in j1_idx) {
        alist_1[[j]] <-
          alist_1[[j]] + (sigma_w_inv * tab[j])
      }
    }
  } # p

  # fully-missing-within cases (empty.idx) belong to a cluster but
  # contribute nothing to the loglik; flag them (-1L) so they are not
  # counted as complete cases when computing dx.sigma.w below
  mpi[mp$empty.idx] <- -1L

  pj <- rowsum.default(pij_1[, , drop = FALSE],
    cluster_idx,
    reorder = FALSE, na.rm = TRUE
  )

  # per cluster
  both_diag_idx <- lav_mat_diag_idx(length(both_idx))
  for (j in seq_len(nclusters)) {
    a_j_full <- alist_1[[j]]
    a_j <- a_j_full[both_idx, both_idx, drop = FALSE]
    p_j <- as.matrix(pj[j, ])
    pb_j <- as.matrix(pj[j, both_idx]) # only both.idx part
    if (nz > 0L) {
      sigma_b_z <- sigma_b_z_1[[zpat2j[j]]]
    } else {
      sigma_b_z <- sigma_b_1
    }

    ibza_j <- sigma_b_z %*% a_j
    ibza_j[both_diag_idx] <- ibza_j[both_diag_idx] + 1

    # one inverse per cluster (used to be 3 solve.default() calls)
    ibza_j_inv <- solve.default(ibza_j)

    ibza_j_inv_bz <- ibza_j_inv %*% sigma_b_z
    ibza_j_inv_bz_p <- ibza_j_inv_bz %*% pb_j
    a_ibza_j_inv_bz <- a_j %*% ibza_j_inv_bz
    a_ibza_j_inv_bz_p <- a_ibza_j_inv_bz %*% pb_j

    a_ibza_j_inv <- a_j %*% ibza_j_inv
    p_ibza_j_inv <- t(crossprod(pb_j, ibza_j_inv))

    # only if we have between-only variables
    if (nz > 0L) {
      g_j <- as.matrix(gj[j, ])
      zij <- as.matrix(gz[j, ])
      zizy <- zizy_1[[zpat2j[j]]]
      zip <- zip_1[[zpat2j[j]]]

      ibza_j_inv_zizy <- ibza_j_inv %*% t(zizy)
      ibza_j_inv_g <- ibza_j_inv %*% g_j
      a_ibza_j_inv_g <- a_j %*% ibza_j_inv_g
      a_ibza_j_inv_zizy <- a_j %*% ibza_j_inv_zizy
      zizy_a_ibza_j_inv_g <- zizy %*% a_ibza_j_inv_g
      p_ibza_j_inv_zizy <- crossprod(pb_j, ibza_j_inv_zizy)
      ggbzpp <- 2 * a_ibza_j_inv_g + a_ibza_j_inv_bz_p - pb_j
      zijzizyp <- (2 * zij - zizy %*% pb_j)

      ###########
      # dx.mu.z #
      ###########
      tmp <- 2 * (t(p_ibza_j_inv_zizy) - zij - zizy_a_ibza_j_inv_g)
      g_muz[j, ] <- drop(tmp)

      ###############
      # dx.sigma.zz #
      ###############
      tmp1 <- (zip + zizy %*% a_ibza_j_inv_zizy # logdet
        - tcrossprod(zij) # ZA
        - tcrossprod(zizy_a_ibza_j_inv_g)) # ZB-1

      d <- (t((2 * zizy_a_ibza_j_inv_g + zizy %*% a_ibza_j_inv_bz_p)
      %*% p_ibza_j_inv_zizy)
      + zijzizyp %*% p_ibza_j_inv_zizy
        - 2 * tcrossprod(zizy_a_ibza_j_inv_g, zij))
      tmp2 <- (d + t(d)) / 2
      tmp <- tmp1 + tmp2
      # symmetry correction
      g_sigma_zz[j, ] <- lav_mat_vech_dd(tmp)

      ###############
      # dx.sigma.yz #
      ###############
      t0 <- -2 * a_ibza_j_inv_zizy

      t1 <- (-2 * tcrossprod(p_ibza_j_inv, g_j)
        - 1 * tcrossprod(p_ibza_j_inv, sigma_b_z %*% pb_j)
        + 2 * tcrossprod(a_ibza_j_inv_g, g_j)) %*% a_ibza_j_inv_zizy
      t2 <- -ggbzpp %*% p_ibza_j_inv_zizy
      t3 <- -tcrossprod(p_ibza_j_inv, zijzizyp)
      t4 <- 2 * tcrossprod(a_ibza_j_inv_g, zij)
      tmp <- t0 + t1 + t2 + t3 + t4
      tmp2 <- matrix(0, nrow(out$sigma.yz), ncol(out$sigma.yz))
      tmp2[both_idx, ] <- tmp
      g_sigma_yz[j, ] <- lav_mat_vec(tmp2)

      ##############
      # dx.sigma.b #
      ##############
      c <- tcrossprod(ggbzpp, p_ibza_j_inv)
      tmp <- t(a_ibza_j_inv) - tcrossprod(a_ibza_j_inv_g) + (c + t(c)) / 2
      # symmetry correction
      zz <- 2 * tmp
      diag(zz) <- diag(tmp)
      zz2 <- matrix(0, nrow(out$sigma.b), ncol(out$sigma.b))
      zz2[both_idx, both_idx] <- zz
      g_sigma_b[j, ] <- lav_mat_vech(zz2)

      # for dx.sigma.w
      part1_b <- -1 * (ibza_j_inv_g %*%
        (2 * t(ibza_j_inv_bz_p) + t(g_j) -
          t(g_j) %*% a_ibza_j_inv_bz)
        + ibza_j_inv_bz + tcrossprod(ibza_j_inv_bz_p))
      part2_b <- 2 * (ibza_j_inv_g + ibza_j_inv_bz_p) # vector
    } else {
      ##############
      # dx.sigma.b #
      ##############
      bzpp <- a_ibza_j_inv_bz_p - pb_j
      c <- tcrossprod(bzpp, p_ibza_j_inv)
      tmp <- t(a_ibza_j_inv) + (c + t(c)) / 2
      # symmetry correction
      zz <- 2 * tmp
      diag(zz) <- diag(tmp)
      zz2 <- matrix(0, nrow(out$sigma.b), ncol(out$sigma.b))
      zz2[both_idx, both_idx] <- zz
      g_sigma_b[j, ] <- lav_mat_vech(zz2)

      part1_b <- -1 * (ibza_j_inv_bz + tcrossprod(ibza_j_inv_bz_p))
      part2_b <- 2 * ibza_j_inv_bz_p # vector
    }

    ##############
    # dx.sigma.w #
    ##############

    part1 <- matrix(0, ny, ny)
    part1[both_idx, both_idx] <- part1_b

    part2 <- matrix(0, ny, 1L)
    part2[both_idx, 1L] <- part2_b

    ij_index <- which(cluster_idx == j)
    pij <- pij_1[ij_index, , drop = FALSE]

    which_compl <- which(mpi[ij_index] == 0L)
    which_incompl <- which(mpi[ij_index] > 0L)

    ap2 <- rep(list(sigma_w_inv %*% part2), length(ij_index))
    ap1a_a <- ap1a_b <- matrix(0, ny, ny)
    if (length(which_compl) > 0L) {
      tmp <- (sigma_w_inv %*% part1 %*% sigma_w_inv)
      ap1a_a <- tmp * length(which_compl)
    }
    if (length(which_incompl) > 0L) {
      p_idx <- mpi[ij_index][which_incompl]
      tmp <- lapply(wip[p_idx], function(x) {
        x %*% part1 %*% x
      })
      ap1a_b <- Reduce("+", tmp)
      ap2[which_incompl] <-
        lapply(wip[p_idx], function(x) {
          x %*% part2
        })
    }
    t1 <- ap1a_a + ap1a_b
    t2 <- (do.call("cbind", ap2) - t(pij)) %*% pij

    aa_wj <- t1 + t2

    tmp <- a_j_full + (aa_wj + t(aa_wj)) / 2
    # symmetry correction
    g_sigma_w[j, ] <- lav_mat_vech_dd(tmp)

    ###########
    # dx.mu.y #
    ###########
    tmp <- numeric(ny)
    if (nz > 0L) {
      tmp[both_idx] <- ibza_j_inv_g + ibza_j_inv_bz_p
    } else {
      tmp[both_idx] <- ibza_j_inv_bz_p
    }
    gbzpp <- a_j_full %*% tmp - p_j
    g_muy[j, ] <- drop(2 * gbzpp)
  } # j

  # scores: rearrange columns to Mu.W, Mu.B, Sigma.W, Sigma.B
  if (score_mode) {
    return(lav_mvn_cl_scores_2implied(
      lp = lp,
      g_muy = g_muy, g_muz = g_muz,
      g_sigma_w = g_sigma_w, g_sigma_b = g_sigma_b,
      g_sigma_yz = g_sigma_yz, g_sigma_zz = g_sigma_zz
    ))
  }

  # gradient: sum over the clusters
  dx_mu_y <- colSums(g_muy)
  dx_sigma_w <- lav_mat_vech_rev(colSums(g_sigma_w))
  dx_sigma_b <- lav_mat_vech_rev(colSums(g_sigma_b))
  if (nz > 0L) {
    dx_mu_z <- colSums(g_muz)
    dx_sigma_zz <- lav_mat_vech_rev(colSums(g_sigma_zz))
    dx_sigma_yz <- matrix(
      colSums(g_sigma_yz),
      nrow(out$sigma.yz), ncol(out$sigma.yz)
    )
  } else {
    dx_mu_z <- numeric(0L)
    dx_sigma_zz <- matrix(0, 0L, 0L)
    dx_sigma_yz <- matrix(0, nrow(out$sigma.yz), 0L)
  }

  # rearrange
  dout <- lav_mvn_cl_2l2implied(
    lp = lp,
    sigma_w = dx_sigma_w, sigma_b = dx_sigma_b,
    sigma_yz = dx_sigma_yz, sigma_zz = dx_sigma_zz,
    mu_y = dx_mu_y, mu_z = dx_mu_z
  )

  if (return_list) {
    dout
  } else {
    c(
      dout$Mu.W, lav_mat_vech(dout$Sigma.W),
      dout$Mu.B, lav_mat_vech(dout$Sigma.B)
    )
  }
}

# print only (standardized) loadings
lav_efa_print <- function(x, nd = 3L, cutoff = 0.3,
                             dot_cutoff = 0.1, alpha_level = 0.01, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  # unclass
  y <- unclass(x)

  if (!y$header$optim.converged) {
    cat("** WARNING ** Optimizer did not end normally\n")
    cat("** WARNING ** Estimates below are most likely unreliable\n")
  }

  # loadings per block
  for (b in seq_len(y$efa$nblocks)) {
    cat("\n")
    if (length(y$efa$block.label) > 0L) {
      cat(y$efa$block.label[[b]], ":\n\n", sep = "")
    }
    mm_lambda <- unclass(y$efa$lambda[[b]])
    lav_print_loadings(mm_lambda,
      nd = nd, cutoff = cutoff,
      dot_cutoff = dot_cutoff,
      alpha_level = alpha_level,
      x_se = y$efa$lambda.se[[b]]
    )
    cat("\n")
  }

  invisible(mm_lambda)
}

# print efaList
lav_efalist_print <- function(x, nd = 3L, cutoff = 0.3,
                          dot_cutoff = 0.1, alpha_level = 0.01, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  # unclass
  y <- unclass(x)

  # kill loadings element if present
  y[["loadings"]] <- NULL

  nfits <- length(y)
  res_1 <- vector("list", nfits)
  for (ff in seq_len(nfits)) {
    res <- lav_object_summary(y[[ff]],
      fit_measures = FALSE,
      estimates = FALSE,
      modindices = FALSE,
      efa = TRUE,
      efa_args = list(
        lambda           = TRUE,
        theta            = FALSE,
        psi              = FALSE,
        eigenvalues      = FALSE,
        sumsq_table      = FALSE,
        lambda_structure = FALSE,
        fs_determinacy   = FALSE,
        se               = FALSE,
        zstat            = FALSE,
        pvalue           = FALSE
      )
    )
    res_1[[ff]] <- lav_efa_print(res,
      nd = nd, cutoff = cutoff,
      dot_cutoff = dot_cutoff,
      alpha_level = alpha_level, ...
    )
  }

  invisible(res_1)
}


# print summary efaList
lav_efalist_summary_print <- function(x, nd = 3L, cutoff = 0.3,
                                  dot_cutoff = 0.1, alpha_level = 0.01,
                                  ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  # unclass
  y <- unclass(x)

  # get nd, if it is stored as an attribute
  nd_1 <- attr(y, "nd")
  if (!is.null(nd_1) && is.numeric(nd_1)) {
    nd <- as.integer(nd_1)
  }
  # get cutoff, if it is stored as an attribute
  ct <- attr(y, "cutoff")
  if (!is.null(ct) && is.numeric(ct)) {
    cutoff <- ct
  }
  # get dot_cutoff, if it is stored as an attribute
  dc <- attr(y, "dot_cutoff")
  if (!is.null(dc) && is.numeric(dc)) {
    dot_cutoff <- dc
  }
  # get alpha_level, if it is stored as an attribute
  al <- attr(y, "alpha_level")
  if (!is.null(al) && is.numeric(al)) {
    alpha_level <- al
  }

  cat("This is ",
    sprintf("lavaan %s", x$lavaan.version),
    " -- running exploratory factor analysis\n",
    sep = ""
  )

  # everything converged?
  if (!x$converged.flag) {
    cat("lavaan WARNING: not all models did converge!\n")
  }
  cat("\n")


  # estimator
  c1 <- c("Estimator")
  # second column
  tmp_est <- toupper(x$estimator)
  if (tmp_est == "DLS") {
    dls_first_letter <- substr(
      x$estimator.args$dls.GammaNT,
      1L, 1L
    )
    tmp_est <- paste("DLS-", toupper(dls_first_letter), sep = "")
  }
  c2 <- tmp_est

  # additional estimator args
  if (!is.null(x$estimator.args) &&
    length(x$estimator.args) > 0L) {
    if (x$estimator == "DLS") {
      c1 <- c(c1, "Estimator DLS value for a")
      c2 <- c(c2, x$estimator.args$dls.a)
    }
  }

  # rotation method
  c1 <- c(c1, "Rotation method")
  if (x$rotation == "none") {
    mm <- toupper(x$rotation)
  } else if (x$rotation.args$orthogonal) {
    mm <- paste(toupper(x$rotation), " ", "ORTHOGONAL",
      sep = ""
    )
  } else {
    mm <- paste(toupper(x$rotation), " ", "OBLIQUE",
      sep = ""
    )
  }
  c2 <- c(c2, mm)

  if (x$rotation != "none") {
    # method options
    if (x$rotation == "geomin") {
      c1 <- c(c1, "Geomin epsilon")
      c2 <- c(c2, x$rotation.args$geomin_epsilon)
    } else if (x$rotation == "orthomax") {
      c1 <- c(c1, "Orthomax gamma")
      c2 <- c(c2, x$rotation.args$orthomax_gamma)
    } else if (x$rotation == "cf") {
      c1 <- c(c1, "Crawford-Ferguson gamma")
      c2 <- c(c2, x$rotation.args$cf_gamma)
    } else if (x$rotation == "oblimin") {
      c1 <- c(c1, "Oblimin gamma")
      c2 <- c(c2, x$rotation.args$oblimin_gamma)
    } else if (x$rotation == "promax") {
      c1 <- c(c1, "Promax kappa")
      c2 <- c(c2, x$rotation.args$promax_kappa)
    }

    # rotation algorithm
    c1 <- c(c1, "Rotation algorithm (rstarts)")
    tmp <- paste(toupper(x$rotation.args$algorithm),
      " (", x$rotation.args$rstarts, ")",
      sep = ""
    )
    c2 <- c(c2, tmp)

    # Standardized metric (or not)
    c1 <- c(c1, "Standardized metric")
    if (x$rotation.args$std_ov) {
      c2 <- c(c2, "TRUE")
    } else {
      c2 <- c(c2, "FALSE")
    }

    # Row weights
    c1 <- c(c1, "Row weights")
    tmp_txt <- x$rotation.args$row_weights
    c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
      substring(tmp_txt, 2),
      sep = ""
    ))
  }

  # format c1/c2
  c1 <- format(c1, width = 33L)
  c2 <- format(c2,
    width = 18L + max(0, (nd - 3L)) * 4L,
    justify = "right"
  )

  # create character matrix
  m <- cbind(c1, c2, deparse.level = 0)
  colnames(m) <- rep("", ncol(m))
  rownames(m) <- rep(" ", nrow(m))

  # print
  write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)

  # data
  if (!is.null(x$lavdata)) {
    cat("\n")
    lav_data_print_short(x$lavdata, nd = nd)
  }

  # number of models
  nfits <- length(x$model.list)

  # number of factors
  nfactors <- x$nfactors

  # fit measures
  if (!is.null(x$fit_table)) {
    cat("\n")
    if (nfits > 1L) {
      cat("Overview models:\n")
    } else {
      cat("Fit measures:\n")
    }
    print(x$fit_table, nd = nd, shift = 2L)
  }

  # eigenvalues
  if (!is.null(x$model.list[[1]]$efa$eigvals[[1]])) {
    cat("\n")
    if (x$model.list[[1]]$efa$std_ov) {
      cat("Eigenvalues correlation matrix:\n")
    } else {
      cat("Eigenvalues covariance matrix:\n")
    }
    for (b in seq_len(x$model.list[[1]]$efa$nblocks)) {
      cat("\n")
      if (length(x$model.list[[1]]$efa$block.label) > 0L) {
        cat(x$model.list[[1]]$efa$block.label[[b]], ":\n\n", sep = "")
      }
      print(x$model.list[[1]]$efa$eigvals[[b]], nd = nd, shift = 2L)
    } # blocks
  }

  # print summary for each model
  for (f in seq_len(nfits)) {
    res <- x$model.list[[f]]
    attr(res, "nd") <- nd
    attr(res, "cutoff") <- cutoff
    attr(res, "dot_cutoff") <- dot_cutoff
    attr(res, "alpha_level") <- alpha_level

    if (nfits > 1L) {
      if (f == 1L) {
        cat("\n")
      }
      cat("Number of factors: ", nfactors[f], "\n")
    }
    # lav_summary_print() prints the $efa element (only) or res
    print(res)
  }

  invisible(y)
}

# sam: fit a SEM in two (or more) steps: Structural After Measurement
#
# sam_method = "local"  : compute (corrected) VETA/EETA per block, and use
#                         these summary statistics to fit the structural part
# sam_method = "global" : fix the measurement parameters, and estimate the
#                         structural parameters in the full (joint) model
# sam_method = "fsr"    : naive (uncorrected) factor score regression
# sam_method = "cfsr"   : Croon-corrected factor score regression

sam <- function(model = NULL,
                data = NULL,
                aux = NULL,
                cmd = "sem",
                se = "twostep",
                mm_list = NULL,
                mm_args = list(bounds = "wide.zerovar"),
                struc_args = list(estimator = "ML"),
                sam_method = "local", # or "global", or "fsr", or "cfsr"
                ..., # common options
                local_options = list(
                  M.method = "ML", # mapping matrix
                  lambda.correction = TRUE,
                  alpha.correction = 0L, # 0 -> (N-1)
                  twolevel.method = "h1"
                ),
                # h1, anova, mean
                global_options = list(), # not used for now
                bootstrap = list(R = 1000L, type = "ordinary",
                                      show.progress = FALSE),
                output = "lavaan",
                bootstrap_args = bootstrap) {
  # capture the user's call; it is stored in the @call slot of the returned
  # object (if output = "lavaan"), so that update() and getCall() work
  mc <- match.call()
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, TRUE)

  # auxiliary variables: forward via dotdotdot, so they reach the
  # underlying measurement-block (and structural) lavaan() calls
  if (!is.null(aux)) {
    dotdotdot$aux <- aux
  }

  # "bootstrap" is the new way to specify arguments, replacing bootstrap_args
  if (!missing(bootstrap_args)) {
    lav_msg_warn(gettext(
      "'bootstrap_args' is deprecated; please use 'bootstrap' instead."))
    bootstrap <- bootstrap_args
  }
  # plugin compatibility with sem()/cfa(): bootstrap = <number of draws>
  if (!is.list(bootstrap) && is.numeric(bootstrap) &&
      length(bootstrap) == 1L && is.finite(bootstrap)) {
    bootstrap <- list(R = as.integer(bootstrap))
  }
  if (!is.list(bootstrap)) {
    lav_msg_stop(gettext(
      "the bootstrap= argument must be a list (eg list(R = 1000L)) or a
       single number (the number of bootstrap draws)."))
  }

  # check model= argument
  has_sam_object_flag <- FALSE
  if (inherits(model, "lavaan") && !is.null(model@internal$sam.method)) {
    has_sam_object_flag <- TRUE
  }

  # check sam_method
  sam_method <- tolower(sam_method)
  if (!sam_method %in% c("local", "global", "fsr", "cfsr")) {
    lav_msg_stop(gettextf(
      "unknown option for sam_method: [%s]; available options are
       local, global, fsr and cfsr.", sam_method))
  }

  # ------------- handling of warn/debug/verbose switches ----------
  if (length(dotdotdot) > 0L) {
    if (!is.null(dotdotdot$debug)) {
     current_debug <- lav_debug()
      if (lav_debug(dotdotdot$debug))
        on.exit(lav_debug(current_debug), TRUE)
      dotdotdot$debug <- NULL
      if (lav_debug()) {
        dotdotdot$warn <- TRUE       # force warnings if debug
        dotdotdot$verbose <- TRUE    # force verbose if debug
      }
    }
    if (!is.null(dotdotdot$warn)) {
      current_warn <- lav_warn()
      if (lav_warn(dotdotdot$warn))
        on.exit(lav_warn(current_warn), TRUE)
      dotdotdot$warn <- NULL
    }
    if (!is.null(dotdotdot$verbose)) {
      current_verbose <- lav_verbose()
      if (lav_verbose(dotdotdot$verbose))
        on.exit(lav_verbose(current_verbose), TRUE)
      dotdotdot$verbose <- NULL
    }
    # check for orthogonal= argument
    if (!is.null(dotdotdot$orthogonal) &&
        dotdotdot$orthogonal &&
        sam_method != "global") {
      lav_msg_warn(gettext(
        "local sam does not support orthogonal = TRUE -> switching to
         global sam"))
     sam_method <- "global"
    }
  } # length(dotdotdot) > 0L

  # check output= argument
  output <- tolower(output)
  if (!output %in% c("list", "list.step1.only", "lavaan")) {
    lav_msg_stop(gettext(
      "output should be \"lavaan\", \"list\" or \"list.step1.only\"."))
  }

  # check se= argument
  if (!missing(se)) {
    se <- tolower(se)
    # aliases
    if (se %in% c("two-step", "two_step", "two.step")) {
      se <- "twostep"
    } else if (se %in% c("two-step-robust", "two-step.robust",
                        "two_step_robust", "two.step.robust",
                        "twostep.robust")) {
      se <- "twostep.robust"
    } else if (se %in% c("twostep.hw", "twostep.huber.white",
                         "two-step.huber.white", "two.step.huber.white",
                         "twostep.huber-white")) {
      se <- "twostep.huber.white"
    } else if (se %in% c("ij", "local")) {
      se <- "local"
    } else if (se %in% c("robust", "robust.sem", "robust.huber.white",
                         "robust.cluster", "robust.cluster.sem")) {
      # plugin compatibility with sem()/cfa(): map the robust se= variants
      # to the SAM analogue (the robust two-step correction); clustering is
      # picked up automatically when cluster= is set
      lav_msg_note(gettextf(
        "se = \"%s\" is not a sam() option; using the SAM analogue
         se = \"twostep.robust\" instead.", se))
      se <- "twostep.robust"
    }
    # check if valid
    if (!se %in% c("standard", "naive", "twostep", "local", "local.nt",
                   "twostep.robust", "twostep.huber.white",
                   "bootstrap", "none")) {
      lav_msg_stop(gettextf(
        "invalid se= argument (%1$s) for sam(); valid options are %2$s.",
        se,
        lav_msg_view(c("twostep", "twostep.robust", "twostep.huber.white",
                       "local", "local.nt",
                       "naive", "standard", "bootstrap", "none"),
                     log_sep = "or")))
    }
    # check for local
    if (se %in% c("local", "local.nt")) {
      if (!sam_method %in% c("local", "fsr", "cfsr")) { # for now
        lav_msg_stop(gettext(
          "local se only available if sam_method is local, fsr, or cfsr"))
      }
    }
  }
  # default is twostep

  # check for gamma.unbiased
  if (is.null(dotdotdot$gamma.unbiased)) {
     # put in TRUE in dotdotdot # lavaan default is still FALSE
     # but not for clustered data or conditional.x = TRUE: the unbiased Gamma
     # is not available there; the (biased) cluster-robust/conditional Gamma
     # is used instead
     dotdotdot$gamma.unbiased <- is.null(dotdotdot$cluster) &&
       !isTRUE(dotdotdot$conditional.x)
  }


  ###############################################
  # STEP 0: process full model, without fitting #
  ###############################################
  if (has_sam_object_flag) {
    fit <- model
    # restore options
    fit@Options <- fit@internal$sam.lavoptions
    # extract other arguments from FIT@internal, unless specified as arguments
    if (missing(mm_list)) {
      mm_list <- fit@internal$sam.mm.list
    } else {
      mm_list <- mm_list
    }
    if (missing(mm_args)) {
      mm_args <- fit@internal$sam.mm.args
    } else {
      mm_args <- mm_args
    }
    if (missing(struc_args)) {
      struc_args <- fit@internal$sam.struc.args
    } else {
      struc_args <- struc_args
    }
    if (missing(sam_method)) {
      sam_method <- fit@internal$sam.method
    } # else: sam_method already holds the validated sam_method
    if (missing(local_options)) {
      local_options <- fit@internal$sam.local.options
    } else {
      local_options <- local_options
    }
    if (missing(global_options)) {
      global_options <- fit@internal$sam.global.options
    }
    if (missing(se)) {
      se <- fit@Options$se
    } else {
      fit@Options$se <- se
    }
    if (missing(cmd)) {
      cmd <- fit@internal$sam.cmd
    }
    # remove @internal slot
    fit@internal <- list()
  } else {
    fit <- lav_sam_step0(
      cmd = cmd, model = model, data = data, se = se,
      sam_method = sam_method, dotdotdot = dotdotdot
    )

    # check for data.type == "none"
    if (fit@Data@data.type == "none") {
      # we are done; perhaps we only wished to create a FIT object?
      return(fit)
      #lav_msg_stop(gettext("no data or sample statistics are provided."))
    }

    # check if we have categorical data
    if (fit@Model@categorical) {
      # switch to M.method = "ULS"
      local_options[["M.method"]] <- "ULS"
      # if sam_method = "global", force estimator to DWLS in struc par
      if (sam_method == "global" &&
          !is.null(struc_args[["estimator"]]) &&
          struc_args[["estimator"]] == "ML") {
          struc_args[["estimator"]] <- "DWLS"
      }
    }

    # check if we have latent interactions
    lv_interaction_flag <- FALSE
    if (length(unlist(fit@pta$vnames$lv.interaction)) > 0L) {
      lv_interaction_flag <- TRUE
      if (!se %in% c("none", "bootstrap")) {
        se <- "local"
        fit@Options$se <- se
      }
    }

    # check for multiple blocks: multigroup (single level) is supported for
    # the local approach as long as the measurement model has no across-group
    # constraints (checked later in lav_sam_step1_local_jac_mg()); multilevel
    # local SEs are supported for the single-group continuous complete-data
    # setting (see lav_sam_gamma_eta_2l()), but not the normal-theory variant
    if (fit@Data@nlevels > 1L && se == "local.nt") {
      lav_msg_stop(gettext("se = \"local.nt\" not available if multiple
                            levels are involved; use se = \"local\"."))
    }

    # single-level clustered data: the normal-theory Gamma used by
    # se = "local.nt" does not account for clustering
    if (se == "local.nt" && fit@Data@nlevels == 1L &&
        length(fit@Data@cluster) > 0L) {
      lav_msg_warn(gettext("se = \"local.nt\" uses the normal-theory Gamma,
        which does not account for clustering; consider se = \"local\"
        instead."))
    }

    # twostep.huber.white (casewise-score sandwich): current scope is
    # continuous ML, single level, no equality constraints, no clustering,
    # complete data or missing = "ml"; single- or multigroup. Outside that
    # scope, fall back to the closest supported flavour. (PML is handled
    # earlier, in lav_sam_step0(): although casewise PML scores exist, the
    # joint-score linearization does not describe the local step-2
    # estimator for PML -- see the note there.)
    if (se == "twostep.huber.white") {
      hw_fallback <- NULL
      if (fit@Data@nlevels > 1L) {
        hw_fallback <- "twostep" # no casewise scores for multilevel (yet)
      } else if (fit@Model@conditional.x) {
        hw_fallback <- "twostep.robust" # no scores under conditional.x (yet)
      } else if (fit@Model@categorical) {
        hw_fallback <- "twostep.robust" # untested for (D)WLS scores (yet)
      } else if (fit@Model@eq.constraints || fit@Model@ceq.simple.only ||
        nrow(fit@Model@ceq.JAC) > 0L) {
        # the block-to-joint free-parameter mapping of the casewise influence
        # does not handle equality constraints (yet); the extra ceq.JAC check
        # catches equality constraints coexisting with inequality
        # constraints/bounds, where both packing flags are FALSE
        hw_fallback <- "twostep.robust"
      } else if (length(fit@Data@cluster) > 0L) {
        # the casewise meat would need within-cluster score aggregation (yet)
        hw_fallback <- "twostep.robust"
      }
      if (!is.null(hw_fallback)) {
        lav_msg_warn(gettextf(
          "se = \"twostep.huber.white\" is not available (yet) for this
           setting; using se = \"%s\" instead.", hw_fallback))
        se <- hw_fallback
        fit@Options$se <- se
      }
    }
  }

  lavoptions <- lavInspect(fit, "options")
  if (lav_verbose()) {
    cat("This is sam using sam_method = ", sam_method, ".\n", sep = "")
  }

  ##############################################
  # STEP 1: fit each measurement model (block) #
  ##############################################
  step1 <- lav_sam_step1(
    cmd = cmd, mm_list = mm_list, mm_args = mm_args,
    fit = fit, sam_method = sam_method
  )

  ##################################################
  # STEP 1b: compute Var(eta) and E(eta) per block #
  #          only needed for local approach!       #
  ##################################################
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    # default local_options
    local_opt <- list(
      M.method = "ML",
      lambda.correction = TRUE,
      alpha.correction = 0L,
      twolevel.method = "h1"
    )
    local_options <- modifyList(local_opt, local_options,
      keep.null = FALSE
    )

    # collect COV/YBAR sample statistics per block from FIT
    out <- lav_sam_get_cov_ybar(fit = fit, local_options = local_options)
    step1$COV  <- out$COV
    step1$YBAR <- out$YBAR
    if (fit@Model@conditional.x) {
      step1$SLOPES <- out$SLOPES # res.slopes of y on x
    }

    # compute EETA/VETA
    step1 <- lav_sam_step1_local(
      step1 = step1, fit = fit,
      sam_method = sam_method,
      local_options = local_options,
      return_cov_iveta2 = (se %in% c("local", "local.nt", "twostep.robust"))
    )
  }

  ##################################################
  # STEP 1c: jacobian of vech(VETA) = f(vech(S))   #
  #          Gamma.eta = NACOV of vech(VETA), the   #
  #          ingredient for (1) the local SEs and   #
  #          (2) the corrected two-step STRUCTURAL  #
  #          test (Satorra-Bentler), which is the   #
  #          DEFAULT test for sam.method = local    #
  #          REGARDLESS of the requested SE. So we  #
  #          compute Gamma.eta for every analytic   #
  #          SE of the local family (incl. the      #
  #          default se = "twostep" and "naive").   #
  #          For SEs that do not strictly need it   #
  #          (twostep/twostep.robust/naive) a        #
  #          failure is NON-fatal: we just skip the #
  #          corrected test. local/local.nt require #
  #          it (re-raise on failure).              #
  ##################################################
  if (se %in% c("local", "local.nt", "twostep", "twostep.robust",
                "twostep.huber.white", "naive")) {
    gamma_eta_required <- se %in% c("local", "local.nt")
    ge_try <- tryCatch({
      gamma_eta <- vector("list", length = fit@Data@ngroups)
      if (lv_interaction_flag) {
        for (g in seq_len(fit@Data@ngroups)) { # group or block
          # initial Gamma.eta
          gamma_eta_init <- step1$COV.IVETA2[[g]]
          # compute 'additional variability' due to step1
          gamma_eta_add <- lav_sam_gamma_add(step1 = step1, fit = fit, group = g)
          gamma_eta[[g]] <- gamma_eta_init + gamma_eta_add
        }
      } else if (fit@Data@nlevels > 1L) {
        # two-level: Gamma.eta per level (one entry per pseudo-group of the
        # structural refit) from the cluster-sandwich Gamma of the h1
        # estimates; the cross-level covariance of the stacked statistics
        # goes through the Case B (full-covariance) sandwich in step 2
        tmp_2l <- lav_sam_gamma_eta_2l(step1 = step1, fit = fit)
        gamma_eta <- tmp_2l$Gamma.eta
        step1$JAC <- tmp_2l$JAC
        step1$Gamma.eta.full <- tmp_2l$Gamma.eta.full
        step1$caseB <- TRUE
      } else if (identical(fit@Options$estimator.orig, "PML")) {
        # PML: theta1.hat is not a function of the saturated statistics
        # (the pairwise likelihood loads on the full bivariate tables), so
        # the influence route below does not apply; build the casewise
        # joint-influence Gamma.eta instead (see lav_sam_gamma_eta_pml())
        gamma_eta <- lav_sam_gamma_eta_pml(step1 = step1, fit = fit)
      } else {
        jac <- lav_sam_step1_local_jac(step1 = step1, fit = fit)

        # build Gamma.eta = JAC %*% Gamma %*% t(JAC) in a memory-lean form that
        # never materializes the p* x p* observed Gamma (see
        # lav_sam_gamma_eta_g()). Falls back to the explicit observed Gamma for
        # settings where that form does not apply (categorical, conditional.x,
        # missing data, ...) and for se = "local.nt" (normal-theory Gamma). The
        # ADF/influence path applies to all SEs except the normal-theory one.
        use_influence <- (se != "local.nt")
        if (use_influence) {
          for (g in seq_len(fit@Data@ngroups)) {
            ge_g <- lav_sam_gamma_eta_g(fit = fit, jac_g = jac[[g]], g = g)
            if (is.null(ge_g)) {
              use_influence <- FALSE
              break
            }
            gamma_eta[[g]] <- ge_g
          }
        }

        gamma <- NULL
        if (!use_influence) {
          if (se == "local.nt") {
            gamma <- lav_object_gamma(lavobject = fit, adf = FALSE)
          } else {
            # stored NACOV, else recompute with the fit-time settings. The
            # (unbiased) ADF Gamma is unavailable for fixed.x / conditional.x
            # (lav_samp_gamma() stops), which the default se = "twostep" can
            # hit (its default is fixed.x = TRUE); fall back to the *biased*
            # ADF Gamma there, exactly as the global yuan.chan test does.
            gamma <- tryCatch(lav_gamma_used(fit), error = function(e) NULL)
            if (is.null(gamma) || is.null(gamma[[1]])) {
              opts <- fit@Options
              opts$gamma.unbiased <- FALSE
              gamma <- lav_gamma_used(
                lavdata = fit@Data, lavoptions = opts,
                lavsamplestats = fit@SampleStats, lavh1 = fit@h1,
                lavimplied = fit@implied
              )
            }
          }
          for (g in seq_len(fit@Data@ngroups)) {
            gamma_eta[[g]] <- jac[[g]] %*% gamma[[g]] %*% t(jac[[g]])
          }
        }
        step1$JAC <- jac

        # Case B (across-group constraints in the measurement model): the
        # per-group Gamma.eta blocks above are incomplete, because vech(VETA_g)
        # also depends on the other groups' sample statistics. Build the full
        # (cross-group) covariance of the stacked structural statistics so that
        # step 2 can compute a proper cross-group sandwich. Cov(vech(S_g)) =
        # gamma[[g]] / n_g, so the actual covariance of the stacked structural
        # statistics is jac.full %*% bdiag(gamma_g / n_g) %*% t(jac.full). This
        # rare path needs the explicit per-group observed Gamma.
        if (isTRUE(attr(jac, "caseB"))) {
          jac_full <- attr(jac, "jac.full")
          nobs_g <- unlist(fit@SampleStats@nobs)
          if (is.null(gamma)) {
            gamma <- fit@SampleStats@NACOV
            if (is.null(gamma) || is.null(gamma[[1]])) {
              gamma <- lavTech(fit, "gamma")
            }
          }
          gscaled <- lapply(seq_len(fit@Data@ngroups),
                            function(g) gamma[[g]] / nobs_g[g])
          gamma_obs_full <- lav_mat_bdiag(gscaled)
          step1$Gamma.eta.full <- jac_full %*% gamma_obs_full %*% t(jac_full)
          step1$caseB <- TRUE
        }
      } # no lv-interaction
      step1$Gamma.eta <- gamma_eta
      TRUE
    }, error = function(e) e)
    if (inherits(ge_try, "error")) {
      if (gamma_eta_required) {
        # the local / local.nt SEs cannot be computed without Gamma.eta
        lav_msg_stop(gettextf(
          "Gamma.eta (needed for se = %s) could not be computed: %s",
          dQuote(se, q = FALSE), conditionMessage(ge_try)))
      }
      # twostep / twostep.robust / naive: the SEs do not need Gamma.eta, so a
      # failure is non-fatal -- we simply do not report the corrected structural
      # test (the naive structural test is reported instead).
      step1$Gamma.eta <- NULL
    }
  }

  if (output == "list.step1.only") {
    # stop here, return interim results
    return(step1)
  }

  # If the structural model uses only a SUBSET of the variables in VETA, restrict
  # VETA / EETA / Gamma.eta to that sub-block. A bi-factor's specific factors, for
  # instance, are pure measurement factors with no structural role: they are part
  # of VETA (and hence of Gamma.eta, the NACOV of vech(VETA)) but absent from the
  # structural model, so the structural fit's Delta/WLS.V would not conform with
  # the full Gamma.eta. For ordinary models every latent appears in the structural
  # model, so 'keep' is everything and this is a no-op.
  if (sam_method %in% c("local", "fsr", "cfsr") &&
      !is.null(step1$VETA) && !is.null(colnames(step1$VETA[[1]])) &&
      !isTRUE(step1$caseB)) {
    keep <- tryCatch({
      struc_pt <- lav_pt_subset_sm(step1$PT, add_exo_cov = TRUE,
                    fixed_x = FALSE,
                    conditional_x = fit@Options$conditional.x,
                    free_fixed_var = TRUE,
                    meanstructure = fit@Options$meanstructure)
      struc_vars <- unique(c(struc_pt$lhs, struc_pt$rhs))
      which(colnames(step1$VETA[[1]]) %in% struc_vars)
    }, error = function(e) integer(0L))
    p_full <- ncol(step1$VETA[[1]])
    if (length(keep) > 0L && length(keep) < p_full) {
      meanstr <- length(step1$EETA[[1]]) > 0L
      # vech positions of the kept sub-block within the full vech(VETA)
      pos <- matrix(0L, p_full, p_full)
      pos[lower.tri(pos, diag = TRUE)] <- seq_len(p_full * (p_full + 1L) / 2L)
      pos[upper.tri(pos)] <- t(pos)[upper.tri(pos)]
      sub_pos <- pos[keep, keep, drop = FALSE]
      vech_keep <- sub_pos[lower.tri(sub_pos, diag = TRUE)]
      # Gamma.eta row order per group is c(EETA (means), vech(VETA));
      # under conditional.x it is c(vecr(cbind(EETA, RS_eta)), vech(VETA)),
      # i.e. (q+1) interleaved entries per variable, then vech(VETA)
      res_slopes_attr <- attr(step1$VETA, "res.slopes")
      if (!is.null(res_slopes_attr)) {
        q_x <- ncol(res_slopes_attr[[1]])
        ge_mean <- unlist(lapply(keep, function(j) {
          (j - 1L) * (q_x + 1L) + seq_len(q_x + 1L)
        }))
        ge_keep <- c(ge_mean, p_full * (q_x + 1L) + vech_keep)
      } else {
        ge_keep <- if (meanstr) c(keep, p_full + vech_keep) else vech_keep
      }
      for (g in seq_along(step1$VETA)) {
        step1$VETA[[g]] <- step1$VETA[[g]][keep, keep, drop = FALSE]
        if (meanstr) {
          step1$EETA[[g]] <- step1$EETA[[g]][keep]
        }
        # conditional.x: keep only the slopes of the retained variables
        # (the exogenous covariates are unaffected)
        if (!is.null(res_slopes_attr)) {
          res_slopes_attr[[g]] <- res_slopes_attr[[g]][keep, , drop = FALSE]
        }
        if (!is.null(step1$Gamma.eta) && !is.null(step1$Gamma.eta[[g]])) {
          step1$Gamma.eta[[g]] <-
            step1$Gamma.eta[[g]][ge_keep, ge_keep, drop = FALSE]
        }
      }
      if (!is.null(res_slopes_attr)) {
        attr(step1$VETA, "res.slopes") <- res_slopes_attr
      }
    }
  }

  # conditional.x: permute Gamma.eta to the structural model's internal order.
  # VETA/EETA are matched by name in the step-2 lavaan() call, but the NACOV
  # is consumed as-is, and the structural model's internal variable order
  # (y-variables first) generally differs from the measurement (VETA) order.
  # The permutation mirrors the lav_pt_subset_sm() call of lav_sam_step2().
  if (fit@Model@conditional.x &&
      sam_method %in% c("local", "fsr", "cfsr") &&
      !is.null(step1$Gamma.eta) && !is.null(step1$Gamma.eta[[1]])) {
    perm_ok <- tryCatch({
      struc_pt <- lav_pt_subset_sm(step1$PT, add_exo_cov = TRUE,
                    fixed_x = fit@Options$fixed.x,
                    conditional_x = TRUE,
                    free_fixed_var = TRUE,
                    meanstructure = fit@Options$meanstructure)
      rs <- attr(step1$VETA, "res.slopes")
      perm <- lav_sam_condx_perm_idx(
        eta_names  = colnames(step1$VETA[[1]]),
        x_names    = colnames(rs[[1]]),
        target_eta = unique(unlist(lav_pt_vnames(struc_pt, type = "ov.nox"))),
        target_x   = unique(unlist(lav_pt_vnames(struc_pt, type = "ov.x"))))
      for (g in seq_along(step1$Gamma.eta)) {
        step1$Gamma.eta[[g]] <- step1$Gamma.eta[[g]][perm, perm, drop = FALSE]
      }
      TRUE
    }, error = function(e) FALSE)
    if (!isTRUE(perm_ok)) {
      if (se %in% c("local", "local.nt")) {
        lav_msg_stop(gettextf(
          "Gamma.eta (needed for se = %s) could not be aligned with the
          structural model under conditional.x = TRUE.",
          dQuote(se, q = FALSE)))
      }
      step1$Gamma.eta <- NULL # no corrected structural test
    }
  }

  ####################################
  # STEP 2: estimate structural part #
  ####################################

  step2 <- lav_sam_step2(
    step1 = step1, fit = fit,
    sam_method = sam_method, struc_args = struc_args
  )

  # make sure step1.free.idx and step2.free.idx are disjoint
  both_idx <- which(step1$step1.free.idx %in% step2$step2.free.idx)
  if (length(both_idx) > 0L) {
    step1$step1.free.idx <- step1$step1.free.idx[-both_idx]
    # STEP1$Sigma.11[both.idx,] <- 0
    # STEP1$Sigma.11[,both.idx] <- 0
    if (!is.null(step1$Sigma.11)) { # maybe se = "none"...
      step1$Sigma.11 <- step1$Sigma.11[-both_idx, -both_idx]
    }
  }

  if (output == "list" && lavoptions$se == "none") {
    return(c(step1, step2))
  }

  ################################################################
  # Step 3: assemble results in a 'dummy' JOINT model for output #
  ################################################################
  if (lav_verbose()) {
    cat("Creating JOINT lavaan object ... ")
  }
  joint <- lav_sam_step3_joint(
    fit = fit, pt_1 = step2$PT,
    sam_method = sam_method
  )
  # fill information from FIT.PA
  joint@Options$optim.method <- step2$FIT.PA@Options$optim.method
  joint@Model@estimator <- fit@Options$estimator # could be DWLS!
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    joint@optim <- step2$FIT.PA@optim
    joint@test <- step2$FIT.PA@test
    # keep @Options$test consistent with the (structural) tests that ended up
    # in @test, just like the global branch below does (the joint itself was
    # fitted with test = "none")
    joint@Options$test <- unname(vapply(joint@test, `[[`,
                                        character(1L), "test"))
  } else {
    # global SAM: if requested (the default for sam.method = "global"), augment
    # the standard full-model test with the Yuan & Chan (2002) rescaled GLOBAL
    # chi-square, which corrects the full-model fit statistic for the separate
    # estimation of the measurement parameters in step 1 and for non-normality.
    # @test then holds both "standard" and "yuan.chan", exactly as sem() holds
    # "standard" and "satorra.bentler". See lav_sam_global_test().
    if (any(fit@Options$test == "yuan.chan")) {
      yc_test <- lav_sam_global_test(
        joint = joint, step1 = step1, step2 = step2, fit = fit
      )
      joint@test <- yc_test$test
      if (!is.null(yc_test$baseline.test)) {
        joint@baseline$test <- yc_test$baseline.test
      }
    }
    # the joint was fitted internally with "yuan.chan" stripped from its test
    # option (lavaan's core cannot compute it). Report exactly the test(s) that
    # ended up in @test, so @Options$test stays consistent with @test (as sem()
    # does for "satorra.bentler"). If the Yuan-Chan test could not be computed
    # (lav_sam_global_test() warned and fell back), this honestly reports just
    # "standard" instead of misleadingly claiming "yuan.chan".
    joint@Options$test <- unname(vapply(joint@test, `[[`, character(1L), "test"))
  }
  # fill in vcov/se information from step 1
  if (!lavoptions$se %in% c("none", "bootstrap")) {
    joint@Options$se <- lavoptions$se # naive/twostep/none/local/bootstrap
    if (joint@Model@ceq.simple.only) {
      vcov_all <- matrix(
        0, joint@Model@nx.unco,
        joint@Model@nx.unco
      )
    } else {
      vcov_all <- matrix(
        0, joint@Model@nx.free,
        joint@Model@nx.free
      )
    }
    vcov_all[step1$step1.free.idx, step1$step1.free.idx] <- step1$Sigma.11
    joint@vcov <- list(
      se = lavoptions$se,
      information = lavoptions$information[1],
      vcov = vcov_all
    )
    # no need to fill @ParTable$se, as step1 SE values should already
    # be in place
  }

  if (lav_verbose()) {
    cat("done.\n")
  }

  # assemble the extra information for the @internal slot
  # (needed for output = "lavaan", and for se = "bootstrap", where
  #  lav_bootstrap_internal() needs a full sam object)
  if (output == "lavaan" || lavoptions$se == "bootstrap") {
    sam_1 <- lav_sam_table(
      joint = joint, step1 = step1,
      fit_pa = step2$FIT.PA,
      cmd = cmd, lavoptions = fit@Options,
      mm_args = mm_args,
      struc_args = struc_args,
      sam_method = sam_method,
      local_options = local_options,
      global_options = global_options
    )
  }

  ##############################################
  # Step 4: compute standard errors for step 2 #
  ##############################################

  if (lavoptions$se == "bootstrap") {
    if (lav_verbose()) {
      cat("computing VCOV for      se = bootstrap ...")
    }
    # construct temporary sam object, so that lav_bootstrap_internal() can
    # use it
    sam_object <- joint
    sam_object@internal <- sam_1
    boot_out <- lav_sam_step2_se_bootstrap(
      sam_object = sam_object, bootstrap = bootstrap
    )
    joint@boot$coef <- boot_out$boot.coef
    joint@Options$bootstrap <- list(R = boot_out$R)
    vcov_1 <- list(se = "bootstrap", VCOV = boot_out$VCOV)

    if (lav_verbose()) {
      cat("done.\n")
    }

  # analytic twostep/standard/naive/local
  } else {
    vcov_1 <- lav_sam_step2_se(
      fit = fit, joint = joint, step1 = step1,
      step2 = step2, local_options = local_options
    )
  }

  # fill in twostep standard errors
  if (lavoptions$se != "none") {
    pt_1 <- joint@ParTable
    joint@Options$se <- vcov_1$se
    joint@vcov$se <- vcov_1$se
    if (lavoptions$se == "bootstrap") {
      joint@vcov$vcov <- vcov_1$VCOV
    } else {
      # sanity check: the step-2 vcov must match the number of structural
      # parameters; a mismatch means the structural vcov could not be
      # computed properly (typically an identification problem that
      # slipped through, eg an empirically under-identified structural
      # part) -- give a clear error instead of a cryptic assignment
      # failure
      n2 <- length(step2$step2.free.idx)
      if (is.null(vcov_1$VCOV) || is.null(dim(vcov_1$VCOV)) ||
          !all(dim(vcov_1$VCOV) == c(n2, n2))) {
        lav_msg_stop(gettext(
          "the variance matrix of the structural (step 2) parameters could
           not be computed; the structural part may not be identified given
           the (fixed) measurement part. Consider simplifying the
           structural part, or using sem() instead."))
      }
      joint@vcov$vcov[step2$step2.free.idx, step2$step2.free.idx] <- vcov_1$VCOV
    }

    # Monte Carlo samples for defined parameter CIs (Preacher & Selig 2012)
    mc_coef <- NULL
    if (sum(pt_1$op == ":=") > 0L && lavoptions$se != "bootstrap" &&
      lav_model_vcov_se_mc_active(lavoptions)) {
      mc_coef <- lav_model_vcov_mc(lavmodel = joint@Model,
                                   vcov = joint@vcov$vcov,
                                   lavoptions = lavoptions)
      joint@internal$monte.carlo <- list(coef = mc_coef)
    }

    pt_1$se <- lav_model_vcov_se(
      lavmodel = joint@Model,
      lavpartable = pt_1,
      vcov = joint@vcov$vcov,
      mc = mc_coef,
      lavoptions = lavoptions
    )

    # surface SEs for latent (residual) variances that are 'fixed' in the
    # joint model but data-derived (eg std.lv = TRUE: the residual variance of
    # an endogenous factor = total - explained). These are free = 0 in the
    # joint model, so lav_model_vcov_se() set their se to 0; recover the proper
    # (delta-method) se from the joint vcov. See lav_sam_step2_se_lv_var().
    lv_var_se <- lav_sam_step2_se_lv_var(joint = joint)
    nz_idx <- which(lv_var_se > 0)
    if (length(nz_idx) > 0L) {
      pt_1$se[nz_idx] <- lv_var_se[nz_idx]
    }

    joint@ParTable <- pt_1
  }



  ##################
  # Step 5: Output #
  ##################

  # assemble pieces to assemble final lavaan object
  if (output == "lavaan") {
    if (lav_verbose()) {
      cat("Assembling results for output ... ")
    }
    res <- joint
    mc_keep <- joint@internal$monte.carlo
    res@internal <- sam_1
    if (!is.null(mc_keep)) {
      res@internal$monte.carlo <- mc_keep
    }
    # replace the internal 'joint' lavaan call by the original sam() call,
    # so that update() and getCall() work (issue #514)
    res@call <- mc
    if (lav_verbose()) {
      cat("done.\n")
    }
  } else {
    res <- c(step1, step2, vcov_1)
  }

  if (lav_verbose()) {
    cat("End of sam.\n")
  }

  res
}

# two-level WLS/DWLS/ULS: sample 'statistics' and their asymptotic
# covariance matrix
#
# YR 2026 (two-level DWLS, phase 1: continuous, complete data)
#
# For two-level data the (per-group) unrestricted statistic vector is the
# saturated (h1) model estimate:
#
#     s = c( Mu.W, vech(Sigma.W), Mu.B, vech(Sigma.B) )
#
# in the same order as (1) the h1 parameter vector used by the
# lav_mvn_cl_* information/score functions, and (2) the per-group Delta
# matrix (lav_model_delta() rbinds the within and between blocks, each
# block laid out as means first, then vech of the covariance matrix).
#
# Note that s contains structurally-fixed entries: the within-level means
# of variables that live at both levels are zero by convention (their mean
# goes to Mu.B), and the model fixes the corresponding intercepts to zero
# as well. These entries have zero rows/columns in Gamma and in Delta, and
# get zero weight; their residuals are identically zero.
#
# The same mechanism handles fixed.x = TRUE: the lav_mvn_cl_* kernels zero
# the rows/columns of the entries that only involve exogenous covariates
# (their means and their (co)variances, at each level), just as for the
# two-level robust ML standard errors; those entries drop out of the
# sandwich (zero rows/columns in Gamma), while the y-x covariances remain
# part of s.
#
# The asymptotic covariance matrix of s treats the CLUSTERS as the
# independent units (sandwich):
#
#     acov(s) = (1/J) * I1^{-1} J1 I1^{-1}
#
# with I1 the (per-cluster, expected) h1 information and J1 the first-order
# information from the cluster-wise scores. To keep all downstream
# consumers (which scale by the total number of OBSERVATIONS) working
# unchanged, we store
#
#     NACOV = Gamma = nobs * acov(s)
#
# so that acov(s) = Gamma/nobs exactly, whatever the cluster structure.

# Gamma (= NACOV = nobs * acov(s)) for ONE group of a two-level dataset,
# evaluated at the h1 (saturated) estimates; the shared kernel for
# lav_samp_wls_2l() below and the post-fit lav_object_gamma()
lav_samp_gamma_2l_g <- function(lavsamplestats = NULL,
                                lavh1 = NULL,
                                lavdata = NULL,
                                g = 1L) {
  nlevels <- lavdata@nlevels
  lp <- lavdata@Lp[[g]]
  nobs <- lavsamplestats@nobs[[g]]
  nclusters <- lp$nclusters[[2]]

  # h1 (saturated) estimates for this group
  mu_w <- lavh1$implied$mean[[(g - 1) * nlevels + 1L]]
  sigma_w <- lavh1$implied$cov[[(g - 1) * nlevels + 1L]]
  mu_b <- lavh1$implied$mean[[(g - 1) * nlevels + 2L]]
  sigma_b <- lavh1$implied$cov[[(g - 1) * nlevels + 2L]]

  # h1 information matrices (per cluster) at the h1 estimates: I1 (the
  # bread) and the first-order J1 from the cluster-wise scores. With
  # missing data (fiml) we use the pattern-based kernels, and (as in
  # lav_model_h1_information, and as in the single-level missing-data
  # Gamma) the bread is the observed -- rather than the expected --
  # information; the h1 estimates then come from EM.
  if (lavdata@missing %in% c("ml", "ml.x")) {
    i1 <- lav_mvn_cl_mi_info_observed(
      y1 = lavdata@X[[g]],
      y2 = lavsamplestats@YLp[[g]][[2]]$Y2,
      lp = lp,
      mp = lavdata@Mp[[g]],
      mu_w = mu_w, sigma_w = sigma_w,
      mu_b = mu_b, sigma_b = sigma_b,
      x_idx = lavsamplestats@x.idx[[g]]
    )
    j1 <- lav_mvn_cl_mi_info_firstorder(
      y1 = lavdata@X[[g]],
      y2 = lavsamplestats@YLp[[g]][[2]]$Y2,
      lp = lp,
      mp = lavdata@Mp[[g]],
      mu_w = mu_w, sigma_w = sigma_w,
      mu_b = mu_b, sigma_b = sigma_b,
      x_idx = lavsamplestats@x.idx[[g]],
      divide_by_two = TRUE
    )
  } else {
    i1 <- lav_mvn_cl_info_expected(
      lp = lp,
      mu_w = mu_w, sigma_w = sigma_w,
      mu_b = mu_b, sigma_b = sigma_b,
      x_idx = lavsamplestats@x.idx[[g]]
    )
    j1 <- lav_mvn_cl_info_firstorder(
      y1 = lavdata@X[[g]],
      ylp = lavsamplestats@YLp[[g]],
      lp = lp,
      mu_w = mu_w, sigma_w = sigma_w,
      mu_b = mu_b, sigma_b = sigma_b,
      x_idx = lavsamplestats@x.idx[[g]],
      divide_by_two = TRUE
    )
  }

  # invert I1, dropping the structurally-fixed (zero-information)
  # entries (cfr. lav_model_h1_acov)
  diag_i1 <- diag(i1)
  keep <- which(abs(diag_i1) > sqrt(.Machine$double.eps) * max(abs(diag_i1)))
  i1_inv <- matrix(0, nrow(i1), ncol(i1))
  inv_keep <- try(
    lav_mat_sym_inverse(i1[keep, keep, drop = FALSE]),
    silent = TRUE
  )
  if (inherits(inv_keep, "try-error")) {
    lav_msg_stop(gettextf(
      "could not invert the h1 information matrix in group %s", g))
  }
  i1_inv[keep, keep] <- inv_keep

  # sandwich acov of s (clusters are the units), rescaled to the
  # 'Gamma/nobs' convention
  acov <- (i1_inv %*% j1 %*% i1_inv) / nclusters
  gamma_1 <- nobs * acov
  # enforce symmetry (numerical)
  gamma_1 <- (gamma_1 + t(gamma_1)) / 2

  attr(gamma_1, "keep") <- keep
  gamma_1
}

# WLS/DWLS/ULS weight matrices from a two-level Gamma (NACOV); shared by
# the continuous and the categorical two-level builders. `keep` indexes the
# non-(structurally-fixed) entries of the statistic vector.
lav_samp_wls_v_2l <- function(gamma_1 = NULL, keep = NULL,
                              estimator = "DWLS", nclusters = NULL,
                              group = 1L) {
  wls_v <- wls_vd <- NULL
  ndat <- nrow(gamma_1)
  if (estimator == "WLS") {
    if (nclusters < length(keep)) {
      lav_msg_stop(gettextf(
        "estimator WLS for two-level data needs more clusters (%1$s) than
        sample statistics (%2$s) in group %3$s; use estimator WLSMV
        (or ULSMV) instead.", nclusters, length(keep), group))
    }
    w_full <- matrix(0, ndat, ndat)
    w_keep <- try(
      lav_mat_sym_inverse(gamma_1[keep, keep, drop = FALSE]),
      silent = TRUE
    )
    if (inherits(w_keep, "try-error")) {
      lav_msg_stop(gettextf(
        "could not invert Gamma (acov of the two-level sample statistics)
        in group %s; use estimator WLSMV (or ULSMV) instead.", group))
    }
    w_full[keep, keep] <- w_keep
    wls_v <- w_full
  } else if (estimator == "DWLS") {
    dacov <- diag(gamma_1)
    idacov <- ifelse(dacov > 0, 1 / dacov, 0)
    wls_v <- diag(idacov, nrow = length(idacov))
    wls_vd <- idacov
  } else if (estimator == "ULS") {
    wls_v <- diag(ndat)
    wls_vd <- rep(1, ndat)
  } else {
    lav_msg_stop(gettextf(
      "estimator %s is not supported for two-level WLS estimation.",
      estimator))
  }
  list(wls_v = wls_v, wls_vd = wls_vd)
}

# fill the WLS.obs/WLS.V/WLS.VD/NACOV slots of lavsamplestats for
# two-level data; called after the h1 model has been estimated (step 06)
lav_samp_wls_2l <- function(lavsamplestats = NULL,
                            lavh1 = NULL,
                            lavdata = NULL,
                            lavoptions = NULL) {
  stopifnot(lavdata@nlevels == 2L)
  estimator <- lavoptions$estimator
  ngroups <- lavsamplestats@ngroups
  nlevels <- lavdata@nlevels

  wls_obs <- vector("list", length = ngroups)
  wls_v <- vector("list", length = ngroups)
  wls_vd <- vector("list", length = ngroups)
  nacov <- vector("list", length = ngroups)

  for (g in seq_len(ngroups)) {
    lp <- lavdata@Lp[[g]]
    nclusters <- lp$nclusters[[2]]

    # h1 (saturated) estimates for this group
    mu_w <- lavh1$implied$mean[[(g - 1) * nlevels + 1L]]
    sigma_w <- lavh1$implied$cov[[(g - 1) * nlevels + 1L]]
    mu_b <- lavh1$implied$mean[[(g - 1) * nlevels + 2L]]
    sigma_b <- lavh1$implied$cov[[(g - 1) * nlevels + 2L]]

    # the statistic vector
    wls_obs[[g]] <- c(
      mu_w, lav_mat_vech(sigma_w, diagonal = TRUE),
      mu_b, lav_mat_vech(sigma_b, diagonal = TRUE)
    )

    # Gamma (cluster sandwich at the h1 estimates; shared kernel)
    gamma_1 <- lav_samp_gamma_2l_g(
      lavsamplestats = lavsamplestats,
      lavh1 = lavh1,
      lavdata = lavdata,
      g = g
    )
    keep <- attr(gamma_1, "keep")
    attr(gamma_1, "keep") <- NULL
    nacov[[g]] <- gamma_1

    # weight matrices
    w <- lav_samp_wls_v_2l(
      gamma_1 = gamma_1, keep = keep,
      estimator = estimator, nclusters = nclusters, group = g
    )
    wls_v[[g]] <- w$wls_v
    if (!is.null(w$wls_vd)) {
      wls_vd[[g]] <- w$wls_vd
    }
  }

  # fill the slots
  lavsamplestats@WLS.obs <- wls_obs
  lavsamplestats@WLS.V <- wls_v
  lavsamplestats@WLS.VD <- wls_vd
  if (!lavsamplestats@NACOV.user) {
    lavsamplestats@NACOV <- nacov
  }

  lavsamplestats
}

# lavEffects(): compute (and provide standard errors for) various 'effects'
# that are functions of the estimated model parameters.
#
# In this first version, we focus on the classic LISREL-style 'total' and
# 'indirect' effects among the variables that appear in the structural part
# of the model (the BETA matrix in the LISREL representation). Both observed
# and latent variables that are involved in a regression are included.
#
# All effects are derived from the reduced-form matrix (I - BETA)^{-1}:
#
#   - direct   effect of j on i:   BETA[i, j]
#   - total    effect of j on i:   ((I - BETA)^{-1} - I)[i, j]
#   - indirect effect of j on i:   total[i, j] - direct[i, j]
#
# Standard errors are obtained either by the delta method (se_def = "delta")
# or by the Monte Carlo method (se_def = "monte.carlo", the default), both of
# which only need the point estimates and the (co)variance matrix of the free
# parameters.
#
# Future versions may add standardized effects, average treatment effects, and
# support for conditional.x = TRUE models (effects of exogenous covariates that
# are stored in the GAMMA matrix).

# YR 2026 -- first version

lavEffects <- function(object,
                       effects = c("total", "indirect"),
                       se_def = NULL,
                       level = 0.95,
                       monte_carlo = NULL,
                       boot_ci_type = "perc",
                       zstat = TRUE,
                       pvalue = TRUE,
                       ci = TRUE,
                       standardized = FALSE,
                       cov_std = TRUE,
                       add_class = TRUE,
                       output = "data.frame",
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)

  # check object
  stopifnot(inherits(object, "lavaan"))

  # are bootstrap draws available?
  boot_available <- !is.null(object@boot$coef) &&
    NROW(object@boot$coef) > 0L

  # resolve the 'se_def' argument: if the user did not specify it, honor the
  # choices that were made when the model was fitted
  if (is.null(se_def)) {
    if (identical(object@Options$se, "bootstrap") && boot_available) {
      # se = "bootstrap" -> use the available bootstrap draws
      se_def <- "bootstrap"
    } else if (identical(object@Options$se.def, "monte.carlo")) {
      # the user explicitly requested the Monte Carlo method for the
      # defined parameters of the original model
      se_def <- "monte.carlo"
    } else {
      # lavEffects() default
      se_def <- "monte.carlo"
    }
  }

  # check/expand 'effects' argument
  effects_choices <- c("total", "indirect", "direct")
  if (!is.character(effects) || length(effects) == 0L) {
    lav_msg_stop(gettext(
      "The \"effects\" argument must be a character vector with one or more of:
       \"total\", \"indirect\", \"direct\"."))
  }
  bad_idx <- which(!effects %in% effects_choices)
  if (length(bad_idx) > 0L) {
    lav_msg_stop(gettextf(
      "Unknown effect(s) in the \"effects\" argument: %s; valid choices are:
       \"total\", \"indirect\", \"direct\".",
      paste(dQuote(effects[bad_idx]), collapse = ", ")))
  }
  # keep canonical order, drop duplicates
  effects <- effects_choices[effects_choices %in% effects]

  # check 'se_def' argument
  se_def <- tolower(as.character(se_def)[1L])
  # allow a few aliases
  se_def <- switch(se_def,
    "mc" = , "montecarlo" = , "monte_carlo" =, "monte-carlo" = "monte.carlo",
    "boot" = "bootstrap",
    se_def)
  if (!se_def %in% c("monte.carlo", "delta", "bootstrap", "none")) {
    lav_msg_stop(gettextf(
      "Unknown \"se_def\" argument: %s; valid choices are \"monte.carlo\",
       \"delta\", \"bootstrap\" or \"none\".", dQuote(se_def)))
  }
  if (identical(se_def, "bootstrap") && !boot_available) {
    lav_msg_warn(gettext(
      "se_def = \"bootstrap\" was requested, but no bootstrap draws are
       available in the fitted object; standard errors are not available.
       Refit the model with se = \"bootstrap\" (or use bootstrapLavaan())."))
    se_def <- "none"
  }

  # check 'boot_ci_type' argument (only relevant when se_def = "bootstrap")
  boot_ci_type <- as.character(boot_ci_type)[1L]
  if (!boot_ci_type %in% c("norm", "basic", "perc", "bca.simple", "bca")) {
    lav_msg_stop(gettextf(
      "Unknown \"boot_ci_type\" argument: %s; valid choices are \"norm\",
       \"basic\", \"perc\", \"bca.simple\" or \"bca\".", dQuote(boot_ci_type)))
  }

  # resolve the Monte Carlo settings (R, seed): if not specified, inherit the
  # 'monte.carlo' list that was used when the model was fitted
  if (is.null(monte_carlo)) {
    monte_carlo <- object@Options$monte.carlo
  }
  mc_r <- 20000L
  mc_seed <- NULL
  if (!is.null(monte_carlo)) {
    if (!is.null(monte_carlo$R)) {
      mc_r <- as.integer(monte_carlo$R)
    }
    if (!is.null(monte_carlo$seed)) {
      mc_seed <- monte_carlo$seed
    }
  }

  # check 'output' argument
  output <- tolower(as.character(output)[1L])
  if (!output %in% c("data.frame", "list")) {
    lav_msg_stop(gettextf(
      "Unknown \"output\" argument: %s; valid choices are \"data.frame\"
       or \"list\".", dQuote(output)))
  }

  lavmodel <- object@Model
  lavdata  <- object@Data

  # only the LISREL representation is supported (the default)
  if (lavmodel@representation != "LISREL") {
    lav_msg_stop(gettextf(
      "lavEffects() is only available for the LISREL representation (not %s).",
      dQuote(lavmodel@representation)))
  }

  nblocks <- lavmodel@nblocks
  nmat    <- lavmodel@nmat
  glist_names <- names(lavmodel@GLIST)

  # build the per-block 'specs' that describe, for each block, which effects
  # are structurally non-zero and must be reported.
  #
  # The structural model is  eta = B eta + Gamma x + zeta, so we work with an
  # augmented effect matrix whose rows are the (endogenous) BETA variables
  # 'eta' and whose columns are the predictors: the eta variables themselves
  # plus, when conditional.x = TRUE, the exogenous covariates 'x' (stored in
  # the GAMMA matrix). When conditional.x = FALSE there is no GAMMA matrix and
  # the columns are simply the eta variables (square case).
  specs <- vector("list", nblocks)
  has_beta <- logical(nblocks)
  for (b in seq_len(nblocks)) {
    # which GLIST matrices belong to block b?
    mm_in_block <- seq_len(nmat[b]) + cumsum(c(0L, nmat))[b]
    beta_idx <- mm_in_block[glist_names[mm_in_block] == "beta"]
    if (length(beta_idx) == 0L) {
      has_beta[b] <- FALSE
      next
    }
    has_beta[b] <- TRUE
    beta_idx <- beta_idx[1L]

    b0 <- lavmodel@GLIST[[beta_idx]]
    nr <- nrow(b0)
    row_names_b <- lavmodel@dimNames[[beta_idx]][[1L]]
    if (is.null(row_names_b)) {
      row_names_b <- as.character(seq_len(nr))
    }
    beta_m_idx <- lavmodel@m.free.idx[[beta_idx]]
    beta_x_idx <- lavmodel@x.free.idx[[beta_idx]]

    # structural adjacency among eta: a[i, j] = TRUE if there is a (possible)
    # direct edge j -> i (the element is free, or fixed to a non-zero value)
    a <- (b0 != 0)
    if (length(beta_m_idx) > 0L) {
      a[beta_m_idx] <- TRUE
    }
    diag(a) <- FALSE

    # reachability among eta (transitive closure): t1[i, j] = TRUE if j
    # influences i through one or more directed paths
    t1 <- lav_effects_reachable(a)

    # GAMMA (exogenous covariates), if conditional.x = TRUE
    gamma_idx <- mm_in_block[glist_names[mm_in_block] == "gamma"]
    g0 <- NULL
    gamma_m_idx <- gamma_x_idx <- integer(0L)
    x_names_b <- character(0L)
    nx <- 0L
    if (length(gamma_idx) > 0L) {
      gamma_idx <- gamma_idx[1L]
      g0 <- lavmodel@GLIST[[gamma_idx]]
      nx <- ncol(g0)
      x_names_b <- lavmodel@dimNames[[gamma_idx]][[2L]]
      if (is.null(x_names_b)) {
        x_names_b <- as.character(seq_len(nx))
      }
      gamma_m_idx <- lavmodel@m.free.idx[[gamma_idx]]
      gamma_x_idx <- lavmodel@x.free.idx[[gamma_idx]]
    }

    nr_col <- nr + nx
    col_names_b <- c(row_names_b, x_names_b)

    # augmented structural patterns (rows = eta, cols = eta then x)
    #   direct  : [ a | ax ]
    #   total   : [ t1 | ((I|t1) %*% ax) ]   (reflexive reach * Gamma)
    #   indirect: [ (a %*% t1) | (t1 %*% ax) ]
    ind_ee <- ((a %*% t1) > 0L)
    diag(ind_ee) <- FALSE
    if (nx > 0L) {
      ax <- (g0 != 0)
      if (length(gamma_m_idx) > 0L) {
        ax[gamma_m_idx] <- TRUE
      }
      refl <- t1
      diag(refl) <- TRUE # reflexive reachability (I | t1)
      tot <- cbind(t1, (refl %*% ax) > 0L)
      dir <- cbind(a,  ax)
      ind <- cbind(ind_ee, (t1 %*% ax) > 0L)
    } else {
      tot <- t1
      dir <- a
      ind <- ind_ee
    }

    # per effect type, the (i, j) cells (and linear indices) to report
    pat <- list(total = tot, indirect = ind, direct = dir)
    cells <- list()
    for (e in effects) {
      P <- pat[[e]]
      ij <- which(P, arr.ind = TRUE) # col 1 = i (row/outcome), 2 = j (col/pred)
      if (nrow(ij) == 0L) {
        cells[[e]] <- NULL
        next
      }
      # order by outcome (i), then predictor (j), following variable order
      o <- order(ij[, 1L], ij[, 2L])
      ij <- ij[o, , drop = FALSE]
      cells[[e]] <- list(
        i   = ij[, 1L],
        j   = ij[, 2L],
        lin = ij[, 1L] + (ij[, 2L] - 1L) * nr # linear idx into nr x nr_col mat
      )
    }

    specs[[b]] <- list(
      nr          = nr,         # number of rows (eta / outcomes)
      nr_col      = nr_col,     # number of columns (eta + x / predictors)
      row.names   = row_names_b,
      col.names   = col_names_b,
      b0          = b0,
      beta_m_idx  = beta_m_idx,
      beta_x_idx  = beta_x_idx,
      g0          = g0,
      gamma_m_idx = gamma_m_idx,
      gamma_x_idx = gamma_x_idx,
      cells       = cells
    )
  }

  if (!any(has_beta)) {
    lav_msg_warn(gettext(
      "The model contains no regressions (empty BETA matrix); there are no
       indirect or total effects to compute."))
    empty <- lav_effects_empty_df(effects, zstat, pvalue)
    return(empty)
  }

  # build the (ordered) skeleton data.frame and remember, for each row, the
  # block and the linear cell index, so that the effect-evaluation function
  # returns its values in exactly the same order
  skeleton <- lav_effects_skeleton(specs, effects, lavdata, lavmodel)
  if (nrow(skeleton$df) == 0L) {
    empty <- lav_effects_empty_df(effects, zstat, pvalue)
    return(empty)
  }
  eff_df <- skeleton$df
  row_block <- skeleton$block
  row_eff   <- skeleton$effect
  row_lin   <- skeleton$lin

  # the master effect-evaluation function: free parameter vector -> vector of
  # all reported effects (in the order of eff_df). Written to also work with a
  # complex 'x' (needed for the complex-step jacobian).
  eff_func <- lav_effects_func(specs, effects, nblocks)

  # point estimates
  x_hat <- lav_model_get_parameters(lavmodel = lavmodel, type = "free")
  est <- eff_func(x_hat)
  eff_df$est <- as.numeric(est)

  # standard errors / confidence intervals
  se <- rep(as.numeric(NA), length(est))
  ci_lower <- ci_upper <- rep(as.numeric(NA), length(est))

  a <- (1 - level) / 2
  nboot <- NULL # for the print header (bootstrap only)

  if (se_def == "bootstrap") {
    # bootstrap-based standard errors and percentile confidence intervals,
    # using the bootstrap draws stored in the fitted object
    boot <- try(lav_inspect_boot(object, add_labels = FALSE,
      add_class = FALSE), silent = TRUE)
    error_idx <- attr(boot, "error.idx")
    if (length(error_idx) > 0L) {
      boot <- boot[-error_idx, , drop = FALSE] # drops attributes
    }
    if (inherits(boot, "try-error") || is.null(boot) || NROW(boot) == 0L) {
      lav_msg_warn(gettext(
        "Could not extract the bootstrap draws from the fitted object;
         standard errors are not available."))
    } else {
      nboot <- NROW(boot)
      boot_eff <- apply(boot, 1L, eff_func)
      if (length(est) == 1L) {
        boot_eff <- matrix(boot_eff, ncol = 1L)
      } else {
        boot_eff <- t(boot_eff)
      }
      boot_eff[!is.finite(boot_eff)] <- as.numeric(NA)
      # bootstrap standard error, using the same MLE-normalised covariance
      # (cov * (nboot - 1)/nboot) that lavaan uses elsewhere for bootstrap SEs
      se <- apply(boot_eff, 2L, sd, na.rm = TRUE) * sqrt((nboot - 1) / nboot)
      # bootstrap confidence interval; the per-quantity computation is shared
      # with parameterEstimates() via lav_bootstrap_ci() (see lav_bootstrap.R)
      if (identical(boot_ci_type, "bca")) {
        # adjusted bootstrap percentile method: the acceleration constants are
        # obtained from the empirical influence values of the effects
        design <- lav_bootstrap_bca_design(object, error_idx)
        stopifnot(nrow(design) == nrow(boot_eff))
        acc <- lav_bootstrap_acceleration(boot_eff, design)
        eff_ci <- lav_bootstrap_ci(boot_t = boot_eff, t0 = est,
          boot_ci_type = "bca", level = level, acc = acc)
      } else {
        # For the "norm" method, the bias is evaluated at the mean of the
        # bootstrapped (free) parameters, exactly as in parameterEstimates().
        bias <- if (identical(boot_ci_type, "norm")) {
          as.numeric(eff_func(colMeans(boot, na.rm = TRUE))) - est
        } else {
          NULL
        }
        eff_ci <- lav_bootstrap_ci(boot_t = boot_eff, t0 = est,
          boot_ci_type = boot_ci_type, level = level, se = se, bias = bias)
      }
      ci_lower <- eff_ci[, 1L]
      ci_upper <- eff_ci[, 2L]
    }
  } else if (se_def %in% c("delta", "monte.carlo")) {
    # (co)variance matrix of the free parameters
    vcov <- try(unclass(lavInspect(object, "vcov")), silent = TRUE)
    if (inherits(vcov, "try-error") || is.null(vcov)) {
      lav_msg_warn(gettext(
        "Could not extract the variance-covariance matrix of the parameters;
         standard errors are not available."))
    } else {
      # reduce to the 'free' (compact) space if needed (ceq.simple.only)
      vcov <- lav_model_vcov_unco_to_free(lavmodel, vcov)

      if (se_def == "delta") {
        jac <- try(lav_func_jacobian_complex(func = eff_func, x = x_hat),
          silent = TRUE)
        if (inherits(jac, "try-error")) {
          jac <- lav_func_jacobian_simple(func = eff_func, x = x_hat)
        }
        eff_cov <- jac %*% vcov %*% t(jac)
        diag_cov <- diag(eff_cov)
        diag_cov[diag_cov < 0] <- as.numeric(NA)
        se <- sqrt(diag_cov)
        # symmetric normal-theory confidence interval
        ci_lower <- est + qnorm(a) * se
        ci_upper <- est + qnorm(1 - a) * se
      } else { # monte.carlo
        mc <- lav_effects_mc_draws(x_hat, vcov, r = mc_r, seed = mc_seed)
        mc_eff <- apply(mc, 1L, eff_func)
        if (length(est) == 1L) {
          mc_eff <- matrix(mc_eff, ncol = 1L)
        } else {
          mc_eff <- t(mc_eff)
        }
        mc_eff[!is.finite(mc_eff)] <- as.numeric(NA)
        se <- apply(mc_eff, 2L, sd, na.rm = TRUE)
        # percentile confidence interval
        mc_ci <- apply(mc_eff, 2L, quantile, probs = c(a, 1 - a),
          na.rm = TRUE, names = FALSE)
        ci_lower <- mc_ci[1L, ]
        ci_upper <- mc_ci[2L, ]
      }
    }
  }
  eff_df$se <- as.numeric(se)

  if (zstat || pvalue) {
    z <- eff_df$est / eff_df$se
    if (zstat) {
      eff_df$z <- z
    }
    if (pvalue) {
      eff_df$pvalue <- 2 * (1 - pnorm(abs(z)))
    }
  }

  if (ci && se_def != "none") {
    eff_df$ci_lower <- as.numeric(ci_lower)
    eff_df$ci_upper <- as.numeric(ci_upper)
  }

  if (output == "list") {
    return(lav_effects_as_list(eff_df, specs, effects, lavmodel,
      row_block = row_block, row_eff = row_eff, row_lin = row_lin))
  }

  # standardized effects? Add std.lv/std.all/std.nox/std.user columns (point
  # estimates), just like the standardized= argument of parameterEstimates().
  std_types <- lav_effects_std_types(object, standardized)
  if (length(std_types$types) > 0L) {
    eff_df <- lav_effects_add_standardized(eff_df, object, lavmodel, specs,
      row_block = row_block, row_lin = row_lin,
      std_types = std_types$types, ov_std_user = std_types$ov_std_user,
      cov_std = cov_std)
  }

  # attributes for printing
  attr(eff_df, "se_def") <- se_def
  attr(eff_df, "level")  <- if (ci) level else NULL
  attr(eff_df, "R")      <- if (se_def == "monte.carlo") mc_r else NULL
  attr(eff_df, "nboot")  <- nboot
  attr(eff_df, "boot_ci_type") <- if (se_def == "bootstrap") boot_ci_type
    else NULL
  attr(eff_df, "ngroups") <- lavdata@ngroups

  if (add_class) {
    class(eff_df) <- c("lavaan.effects", "lavaan.data.frame", "data.frame")
  }

  eff_df
}

# resolve the 'standardized' argument (as in parameterEstimates()) into a
# vector of standardization types and an optional set of user-specified
# observed variables to standardize (std.user)
lav_effects_std_types <- function(object, standardized) {
  ov_std_user <- NULL
  if (is.logical(standardized)) {
    if (isTRUE(standardized)) {
      types <- c("std.lv", "std.all")
      if (length(lavNames(object, "ov.x")) && object@Options$fixed.x) {
        types <- c(types, "std.nox")
      }
    } else {
      types <- character(0L)
    }
  } else {
    standardized <- as.character(standardized)
    std_set <- c("std.lv", "std.all", "std.nox")
    if (length(standardized) > 0L &&
        !any(tolower(standardized) %in% std_set)) {
      # interpret 'standardized' as a vector of observed variable names
      ov_all <- lavNames(object, "ov")
      bad <- standardized[!(standardized %in% ov_all)]
      if (length(bad) > 0L) {
        lav_msg_warn(gettextf(
          "standardized= argument contains unknown observed variable(s): %s",
          paste(bad, collapse = ", ")))
      }
      ov_std_user <- standardized[standardized %in% ov_all]
      if (length(ov_std_user) == 0L) {
        types <- character(0L)
      } else {
        types <- c("std.lv", "std.user")
      }
    } else {
      types <- tolower(standardized)
      if ("std.nox" %in% types) {
        if (length(lavNames(object, "ov.x")) == 0L) {
          lav_msg_warn(gettext(
            "`std.nox' values are identical to `std.all' values in the
             absence of exogenous covariates."))
        } else if (!object@Options$fixed.x) {
          lav_msg_warn(gettext(
            "`std.nox' values are identical to `std.all' values when
             fixed.x = FALSE."))
        }
      }
    }
  }
  list(types = types, ov_std_user = ov_std_user)
}

# the (model-implied) standard-deviation vector used to standardize the BETA
# variables of a block, for a given standardization 'type'. Variables that are
# not standardized under this type get a SD of 1.
lav_effects_std_sd_vec <- function(var_names, sd_all, type, lv_names,
                                   ovx_names, ov_std_user) {
  sdv <- sd_all
  if (type == "std.lv") {
    # only the (true) latent variables are standardized
    sdv[!(var_names %in% lv_names)] <- 1
  } else if (type == "std.nox") {
    # all variables except the exogenous observed 'x'
    sdv[var_names %in% ovx_names] <- 1
  } else if (type == "std.user") {
    # latents + user-specified observed variables
    keep <- (var_names %in% lv_names) | (var_names %in% ov_std_user)
    sdv[!keep] <- 1
  }
  # type == "std.all": standardize everything (sdv = sd_all)
  sdv
}

# add standardized effect columns (point estimates) to the effects data.frame.
# The standardized effect of j on i equals effect[i,j] * sd_j / sd_i, where the
# SDs are the model-implied standard deviations of the BETA variables under the
# requested standardization type.
lav_effects_add_standardized <- function(eff_df, object, lavmodel, specs,
                                         row_block, row_lin, std_types,
                                         ov_std_user = NULL, cov_std = TRUE) {
  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  glist_names <- names(lavmodel@GLIST)
  partable <- object@ParTable

  # model-implied (co)variances of the BETA (eta) variables, per block
  veta <- lav_model_veta(lavmodel = lavmodel, glist = lavmodel@GLIST)

  # per block: 'all' SDs for the row (eta) and column (eta + exogenous x)
  # variables, plus the variable classifications needed by each type
  sd_row_all <- vector("list", nblocks)
  sd_col_all <- vector("list", nblocks)
  lv_names <- vector("list", nblocks)
  ovx_names <- vector("list", nblocks)
  for (b in seq_len(nblocks)) {
    sp <- specs[[b]]
    if (is.null(sp)) {
      next
    }
    v2 <- diag(veta[[b]])
    v2[v2 < 0] <- as.numeric(NA)
    sd_eta <- sqrt(v2)
    sd_row_all[[b]] <- sd_eta
    if (sp$nr_col > sp$nr) {
      # exogenous covariate SDs (from cov.x), in the GAMMA column order
      mm_in_block <- seq_len(nmat[b]) + cumsum(c(0L, nmat))[b]
      covx_idx <- mm_in_block[glist_names[mm_in_block] == "cov.x"]
      if (length(covx_idx) > 0L) {
        sd_x <- sqrt(diag(lavmodel@GLIST[[covx_idx[1L]]]))
      } else {
        sd_x <- rep(1, sp$nr_col - sp$nr)
      }
      sd_col_all[[b]] <- c(sd_eta, sd_x)
    } else {
      sd_col_all[[b]] <- sd_eta
    }
    lv_names[[b]] <- lav_pt_vnames(partable, "lv", block = b)
    if (lavmodel@fixed.x) {
      ovx_names[[b]] <- lav_pt_vnames(partable, "ov.x", block = b)
    } else {
      # if fixed.x = FALSE, the exogenous 'x' are treated as random
      # variables like any other, and std.nox reduces to std.all
      ovx_names[[b]] <- character(0L)
    }
  }

  nr_rows <- nrow(eff_df)
  for (ty in std_types) {
    # per-block SD vectors (row and column) for this standardization type
    sd_row <- vector("list", nblocks)
    sd_col <- vector("list", nblocks)
    for (b in seq_len(nblocks)) {
      if (is.null(specs[[b]])) {
        next
      }
      sd_row[[b]] <- lav_effects_std_sd_vec(specs[[b]]$row.names,
        sd_row_all[[b]], ty, lv_names[[b]], ovx_names[[b]], ov_std_user)
      sd_col[[b]] <- lav_effects_std_sd_vec(specs[[b]]$col.names,
        sd_col_all[[b]], ty, lv_names[[b]], ovx_names[[b]], ov_std_user)
    }
    std_col <- rep(as.numeric(NA), nr_rows)
    for (r in seq_len(nr_rows)) {
      b <- row_block[r]
      nr <- specs[[b]]$nr
      lin <- row_lin[r]
      i <- ((lin - 1L) %% nr) + 1L
      j <- ((lin - 1L) %/% nr) + 1L
      std_col[r] <- eff_df$est[r] * sd_col[[b]][j] / sd_row[[b]][i]
    }
    eff_df[[ty]] <- std_col
  }

  eff_df
}

# print method for the object returned by lavEffects()
lav_effects_print <- function(x, ..., nd = 3L) {
  # if the object has been subsetted (and lost the columns/attributes that
  # this method relies on), fall back to a plain data.frame print
  if (is.null(x$effect) || is.null(x$lhs) || is.null(x$est)) {
    y <- x
    class(y) <- "data.frame"
    print(y, ...)
    return(invisible(x))
  }

  # number formats (mirrors lav_parameterestimates_print)
  num_format  <- paste("%", max(8L, nd + 5L), ".", nd, "f", sep = "")
  char_format <- paste("%", max(8L, nd + 5L), "s", sep = "")

  se_def  <- attr(x, "se_def")
  level   <- attr(x, "level")
  ngroups <- attr(x, "ngroups")
  if (is.null(ngroups)) {
    ngroups <- 1L
  }

  # header
  cat("\nEffects:\n\n")
  if (!is.null(se_def)) {
    c1 <- c("Standard errors")
    c2 <- c(switch(se_def, monte.carlo = "Monte Carlo", delta = "Delta",
      bootstrap = "Bootstrap", none = "None", se_def))
    if (identical(se_def, "monte.carlo") && !is.null(attr(x, "R"))) {
      c1 <- c(c1, "Monte Carlo draws")
      c2 <- c(c2, format(attr(x, "R")))
    }
    if (identical(se_def, "bootstrap") && !is.null(attr(x, "nboot"))) {
      c1 <- c(c1, "Bootstrap draws")
      c2 <- c(c2, format(attr(x, "nboot")))
    }
    if (identical(se_def, "bootstrap") && !is.null(attr(x, "boot_ci_type"))) {
      c1 <- c(c1, "Bootstrap CI type")
      c2 <- c(c2, attr(x, "boot_ci_type"))
    }
    if (!is.null(level) && !identical(se_def, "none")) {
      c1 <- c(c1, "Confidence level")
      c2 <- c(c2, format(level))
    }
    c1 <- format(c1, width = 33L)
    c2 <- format(c2, width = 14L + max(0, (nd - 3L)) * 4L, justify = "right")
    m1 <- cbind(c1, c2, deparse.level = 0)
    colnames(m1) <- rep("", ncol(m1))
    rownames(m1) <- rep(" ", nrow(m1))
    write.table(m1, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  effect_title <- c(total = "Total Effects", indirect = "Indirect Effects",
    direct = "Direct Effects")

  # group/level handling
  group_vec <- if (!is.null(x$group)) x$group else rep(1L, nrow(x))
  ugroups <- unique(group_vec)

  for (g in ugroups) {
    if (ngroups > 1L) {
      cat("\n\nGroup ", which(ugroups == g), " [", as.character(g), "]:\n",
        sep = "")
    }
    g_idx <- which(group_vec == g)

    for (e in c("total", "indirect", "direct")) {
      e_idx <- g_idx[x$effect[g_idx] == e]
      if (length(e_idx) == 0L) {
        next
      }
      cat("\n", effect_title[e], ":\n", sep = "")

      sub <- x[e_idx, , drop = FALSE]
      # left label: "outcome ~ predictor"
      labels <- paste(format(sub$lhs), sub$op, format(sub$rhs))

      # numeric columns to show
      cols <- list()
      cols[["Estimate"]] <- sprintf(num_format, sub$est)
      if (!is.null(sub$se)) {
        cols[["Std.Err"]] <- sprintf(num_format, sub$se)
      }
      if (!is.null(sub$z)) {
        cols[["z-value"]] <- sprintf(num_format, sub$z)
      }
      if (!is.null(sub$pvalue)) {
        cols[["P(>|z|)"]] <- sprintf(num_format, sub$pvalue)
      }
      if (!is.null(sub$ci_lower)) {
        cols[["ci_lower"]] <- sprintf(num_format, sub$ci_lower)
      }
      if (!is.null(sub$ci_upper)) {
        cols[["ci_upper"]] <- sprintf(num_format, sub$ci_upper)
      }
      # standardized columns (if any)
      for (sc in c("std.lv", "std.all", "std.nox", "std.user")) {
        if (!is.null(sub[[sc]])) {
          cols[[sc]] <- sprintf(num_format, sub[[sc]])
        }
      }

      m <- do.call(cbind, cols)
      if (!is.null(sub$se)) {
        # blank out se/z/p/ci when se is NA (no SE available)
        na_se <- which(is.na(sub$se))
        if (length(na_se) > 0L) {
          for (cn in intersect(c("Std.Err", "z-value", "P(>|z|)",
            "ci_lower", "ci_upper"), colnames(m))) {
            m[na_se, cn] <- sprintf(char_format, "")
          }
        }
        # blank out se/z/p for fixed effects (se == 0); keep the ci columns
        # (which equal the point estimate)
        zero_se <- which(!is.na(sub$se) & sub$se == 0)
        if (length(zero_se) > 0L) {
          for (cn in intersect(c("Std.Err", "z-value", "P(>|z|)"),
            colnames(m))) {
            m[zero_se, cn] <- sprintf(char_format, "")
          }
        }
      }
      rownames(m) <- paste("   ", format(labels, justify = "left"))
      print.default(m, quote = FALSE, right = TRUE)
    }
  }
  cat("\n")

  invisible(x)
}

# transitive closure (reachability) of a boolean adjacency matrix, where
# a[i, j] = TRUE means there is a direct edge j -> i. Returns t1 with
# t1[i, j] = TRUE if j reaches i via one or more directed edges. The diagonal
# is forced to FALSE (we are not interested in self-loops).
lav_effects_reachable <- function(a) {
  n <- nrow(a)
  r <- (a != 0)
  if (n > 1L) {
    # Warshall: r[i, j] |= r[i, k] & r[k, j]
    for (k in seq_len(n)) {
      r <- r | (r[, k, drop = FALSE] %*% r[k, , drop = FALSE] > 0L)
    }
  }
  diag(r) <- FALSE
  r
}

# build the master effect-evaluation function:
# free parameter vector 'x' -> numeric (or complex) vector of all reported
# effects, in the order: block, then effect type, then (outcome, predictor)
lav_effects_func <- function(specs, effects, nblocks) {
  function(x) {
    out <- vector(mode = if (is.complex(x)) "complex" else "numeric",
      length = 0L)
    for (b in seq_len(nblocks)) {
      sp <- specs[[b]]
      if (is.null(sp)) {
        next
      }
      b <- sp$b0
      if (is.complex(x)) {
        storage.mode(b) <- "complex"
      }
      if (length(sp$beta_m_idx) > 0L) {
        b[sp$beta_m_idx] <- x[sp$beta_x_idx]
      }
      diag(b) <- 0
      nr <- sp$nr
      i_b_inv <- solve(diag(nr) - b)

      # eta -> eta effects (square block)
      total_ee <- i_b_inv - diag(nr)
      if (is.null(sp$g0)) {
        total  <- total_ee
        direct <- b
      } else {
        # eta <- Gamma x effects: total = (I-B)^-1 Gamma, direct = Gamma,
        # indirect = total - direct
        g <- sp$g0
        if (is.complex(x)) {
          storage.mode(g) <- "complex"
        }
        if (length(sp$gamma_m_idx) > 0L) {
          g[sp$gamma_m_idx] <- x[sp$gamma_x_idx]
        }
        total  <- cbind(total_ee, i_b_inv %*% g)
        direct <- cbind(b, g)
      }
      indirect <- total - direct

      for (e in effects) {
        cell <- sp$cells[[e]]
        if (is.null(cell)) {
          next
        }
        MAT <- switch(e,
          total    = total,
          indirect = indirect,
          direct   = direct)
        out <- c(out, MAT[cell$lin])
      }
    }
    out
  }
}

# draw r samples from MVN(x_hat, vcov), saving/restoring the RNG state
lav_effects_mc_draws <- function(x_hat, vcov, r = 20000L, seed = NULL) {
  r <- as.integer(r)
  stopifnot(r > 0L)

  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    saved_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    on.exit(assign(".Random.seed", saved_seed, envir = .GlobalEnv), add = TRUE)
  } else {
    on.exit({
      if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
        rm(".Random.seed", envir = .GlobalEnv)
      }
    }, add = TRUE)
  }
  if (!is.null(seed)) {
    set.seed(seed)
  }

  vcov_sym <- 0.5 * (vcov + t(vcov))
  mc <- lav_mvrnorm(n = r, mu = x_hat, sigma_1 = vcov_sym,
    check_symmetry = FALSE)
  if (!is.matrix(mc)) {
    mc <- matrix(mc, nrow = r)
  }
  mc
}

# build the (ordered) skeleton data.frame: one row per reported effect, in the
# same order as lav_effects_func() returns its values
lav_effects_skeleton <- function(specs, effects, lavdata, lavmodel) {
  nblocks <- lavmodel@nblocks
  ngroups <- lavdata@ngroups
  nlevels <- lavdata@nlevels

  eff_col <- character(0L)
  lhs_col <- character(0L)
  rhs_col <- character(0L)
  blk_col <- integer(0L)
  lin_col <- integer(0L)

  for (b in seq_len(nblocks)) {
    sp <- specs[[b]]
    if (is.null(sp)) {
      next
    }
    for (e in effects) {
      cell <- sp$cells[[e]]
      if (is.null(cell)) {
        next
      }
      n_e <- length(cell$lin)
      eff_col <- c(eff_col, rep(e, n_e))
      lhs_col <- c(lhs_col, sp$row.names[cell$i]) # outcome (eta)
      rhs_col <- c(rhs_col, sp$col.names[cell$j]) # predictor (eta or x)
      blk_col <- c(blk_col, rep(b, n_e))
      lin_col <- c(lin_col, cell$lin)
    }
  }

  df <- data.frame(
    effect = eff_col,
    lhs    = lhs_col,
    op     = rep("~", length(lhs_col)),
    rhs    = rhs_col,
    stringsAsFactors = FALSE
  )

  # group/level/block columns (only when relevant)
  if (nblocks > 1L) {
    if (ngroups > 1L) {
      group_idx <- ((blk_col - 1L) %/% max(nlevels, 1L)) + 1L
      glabels <- lavdata@group.label
      if (length(glabels) >= max(group_idx)) {
        df$group <- glabels[group_idx]
      } else {
        df$group <- group_idx
      }
    }
    if (nlevels > 1L) {
      level_idx <- ((blk_col - 1L) %% nlevels) + 1L
      llabels <- lavdata@level.label
      if (length(llabels) >= max(level_idx)) {
        df$level <- llabels[level_idx]
      } else {
        df$level <- level_idx
      }
    }
    if (ngroups <= 1L && nlevels <= 1L) {
      df$block <- blk_col
    }
  }

  list(df = df, effect = eff_col, block = blk_col, lin = lin_col)
}

# an empty result (no effects) with the expected columns
lav_effects_empty_df <- function(effects, zstat, pvalue) {
  df <- data.frame(
    effect = character(0L),
    lhs    = character(0L),
    op     = character(0L),
    rhs    = character(0L),
    est    = numeric(0L),
    se     = numeric(0L),
    stringsAsFactors = FALSE
  )
  if (zstat) {
    df$z <- numeric(0L)
  }
  if (pvalue) {
    df$pvalue <- numeric(0L)
  }
  df$ci_lower <- numeric(0L)
  df$ci_upper <- numeric(0L)
  class(df) <- c("lavaan.effects", "lavaan.data.frame", "data.frame")
  df
}

# convert the result to a list of matrices (one per block and per effect type),
# using the per-row block/effect/linear-cell bookkeeping from the skeleton
lav_effects_as_list <- function(eff_df, specs, effects, lavmodel,
                                row_block, row_eff, row_lin) {
  nblocks <- lavmodel@nblocks
  out <- list()
  for (b in seq_len(nblocks)) {
    sp <- specs[[b]]
    if (is.null(sp)) {
      next
    }
    nr <- sp$nr
    nr_col <- sp$nr_col
    dn <- list(sp$row.names, sp$col.names)
    block_list <- list()
    for (e in effects) {
      rows <- which(row_block == b & row_eff == e)
      if (length(rows) == 0L) {
        next
      }
      m_est <- matrix(0, nr, nr_col, dimnames = dn)
      m_se  <- matrix(as.numeric(NA), nr, nr_col, dimnames = dn)
      m_est[row_lin[rows]] <- eff_df$est[rows]
      if (!is.null(eff_df$se)) {
        m_se[row_lin[rows]] <- eff_df$se[rows]
      }
      block_list[[e]] <- list(est = m_est, se = m_se)
    }
    if (nblocks > 1L) {
      out[[paste0("block.", b)]] <- block_list
    } else {
      out <- block_list
    }
  }
  out
}

# handle EFA equality constraints
# YR 14 Jan 2020: 0.6-6 does no longer impose 'explicit' constraints
#                 if we only need to fix a parameter to 0/1
# Note: we should also check if they are really needed:
#       eg., if all the factor-loadings of the 'second' set (time/group)
#       are constrained to be equal to the factor-loadings of the first
#       set, no further constraints are needed
lav_pt_efa_con <- function(list_1 = NULL,
                                         orthogonal_efa = FALSE,
                                         group_equal = character(0L)) {
  # for each set, for each block
  nblocks <- lav_pt_nblocks(list_1)
  set_names <- lav_pt_efa_values(list_1)
  nsets <- length(set_names)

  for (b in seq_len(nblocks)) {
    for (s in seq_len(nsets)) {
      # lv's for this block/set
      lv_nam_efa <- unique(list_1$lhs[list_1$op == "=~" &
        list_1$block == b &
        list_1$efa == set_names[s]])
      if (length(lv_nam_efa) == 1L) {
        # nothing to do (warn?)
        next
      }

      # equality constraints on ALL factor loadings in this set?
      # two scenario's:
      # 1. eq constraints within the same block, perhaps time1/time2/
      # 2. eq constraints across groups (group.equal = "loadings")
      # --> no constraints are needed

      # store labels (if any)
      fix_to_zero <- TRUE

      # 1. within block/group
      if (s == 1L) {
        set_idx <- which(list_1$op == "=~" &
          list_1$block == b &
          list_1$lhs %in% lv_nam_efa)
        label_set1 <- list_1$label[set_idx]
      } else {
        # collect labels for this set, if any
        set_idx <- which(list_1$op == "=~" &
          list_1$block == b &
          list_1$lhs %in% lv_nam_efa)

        # user-provided labels (if any)
        this_label_set <- list_1$label[set_idx]

        # same as in reference set?
        if (all(nchar(this_label_set) > 0L) &&
          all(this_label_set %in% label_set1)) {
          fix_to_zero <- FALSE
        }
      }

      # 2. across groups
      if (b == 1L) {
        set_idx <- which(list_1$op == "=~" &
          list_1$block == b &
          list_1$lhs %in% lv_nam_efa)
        label_group1 <- list_1$label[set_idx]
      } else {
        if ("loadings" %in% group_equal) {
          fix_to_zero <- FALSE
        } else {
          # collect labels for this set, if any
          set_idx <- which(list_1$op == "=~" &
            list_1$block == b &
            list_1$lhs %in% lv_nam_efa)

          # user-provided labels (if any)
          this_label_set <- list_1$label[set_idx]

          # same as in reference set?
          if (all(nchar(this_label_set) > 0L) &&
            all(this_label_set %in% label_group1)) {
            fix_to_zero <- FALSE
          }
        }
      }

      # 1. echelon pattern
      nfac <- length(lv_nam_efa)
      for (f in seq_len(nfac)) {
        if (f == 1L) {
          next
        }
        nzero <- (f - 1L)
        ind_idx <- which(list_1$op == "=~" &
          list_1$block == b &
          list_1$lhs %in% lv_nam_efa[f])
        if (length(ind_idx) < nzero) {
          lav_msg_stop(gettextf(
            "efa factor %s has not enough indicators for echelon pattern",
            lv_nam_efa[f]))
        }

        # fix to zero
        if (fix_to_zero) {
          list_1$free[ind_idx[seq_len(nzero)]] <- 0L
          list_1$ustart[ind_idx[seq_len(nzero)]] <- 0
          list_1$user[ind_idx[seq_len(nzero)]] <- 7L
        } else {
          list_1$user[ind_idx[seq_len(nzero)]] <- 77L
        }
      }

      # 2. covariances constrained to zero (only if oblique rotation)
      if (!orthogonal_efa) {
        # skip if user == 1 (user-override!)
        cov_idx <- which(list_1$op == "~~" &
          list_1$block == b &
          list_1$user == 0L &
          list_1$lhs %in% lv_nam_efa &
          list_1$rhs %in% lv_nam_efa &
          list_1$lhs != list_1$rhs)

        # fix to zero
        if (fix_to_zero) {
          list_1$free[cov_idx] <- 0L
          list_1$ustart[cov_idx] <- 0
          list_1$user[cov_idx] <- 7L
        } else {
          list_1$user[cov_idx] <- 77L
        }
      }
    } # sets
  } # blocks

  list_1
}

# machinery for the 'robust' versions of restricted tests (score and Wald):
# the scaled, (mean and variance) adjusted, and generalized (asymptotically
# distribution-free) test statistics
#
# reference:
#
# Satorra, A. (2000). Scaled and adjusted restricted tests in multi-sample
# analysis of moment structures. In Heijmans, R.D.H., Pollock, D.S.G. &
# Satorra, A. (Eds.), Innovations in multivariate statistical analysis:
# A festschrift for Heinz Neudecker (pp. 233-247). London: Kluwer.
#
# YR 11 July 2026: initial version

# which sandwich 'flavor' can we use to robustify a restricted test for
# this object?
#
# - if the se's are already sandwich-based, use that flavor
# - otherwise, if a scaled test statistic was requested, the ingredients
#   are available anyway: use Gamma (robust.sem) if we have it, or the
#   casewise scores (robust.huber.white) if we have raw data
# - otherwise, return NULL (only the standard test is available)
lav_test_robust_se_flavor <- function(object) {
  lavoptions <- object@Options

  sandwich_se <- c(
    "robust.sem", "robust.sem.nt", "robust.cluster.sem",
    "robust.huber.white", "robust.cluster"
  )
  if (lavoptions$se %in% sandwich_se) {
    return(lavoptions$se)
  }

  scaled_tests <- c(
    "satorra.bentler", "yuan.bentler", "yuan.bentler.mplus",
    "mean.var.adjusted", "scaled.shifted",
    "mean.var.adjusted.corrected", "scaled.shifted.corrected"
  )
  if (!any(lavoptions$test %in% scaled_tests)) {
    return(NULL)
  }

  gamma <- object@SampleStats@NACOV
  if (length(gamma) > 0L && !is.null(gamma[[1]])) {
    return("robust.sem")
  }
  if (object@Model@estimator %in% c("ML", "PML") &&
    length(object@Data@X) > 0L) {
    return("robust.huber.white")
  }

  NULL
}

# the (unit-scale) 'meat' of the sandwich covariance matrix, in the same
# parameter space (unco when ceq.simple.only) and on the same scale as
# lavTech(object, "information.*"):
#
#   B = sum_g f_g Delta_g' V_g Gamma_g V_g Delta_g   (robust.sem flavors)
#   B = first-order information (casewise scores)    (robust.huber.white)
#
# so that n * Avar(thetahat) = J %*% B %*% J, with J the (constrained)
# inverse of the unit information; returns NULL if the ingredients are
# not available for this fit
lav_test_meat <- function(object, se = NULL) {
  lavoptions <- object@Options
  lavmodel <- object@Model

  if (is.null(se)) {
    se <- lav_test_robust_se_flavor(object)
  }
  if (is.null(se)) {
    return(NULL)
  }

  # the h1 slot may be empty; downstream functions recompute it if NULL
  lavh1 <- object@h1
  if (length(lavh1) == 0L) {
    lavh1 <- NULL
  }

  if (se %in% c("robust.sem", "robust.sem.nt", "robust.cluster.sem")) {
    # Gamma needed
    gamma <- object@SampleStats@NACOV
    if (length(gamma) == 0L || is.null(gamma[[1]])) {
      return(NULL)
    }
    nvcov <- try(
      lav_model_nvcov_robust_sem(
        lavmodel = lavmodel,
        lavsamplestats = object@SampleStats,
        lavdata = object@Data,
        lavcache = object@Cache,
        lavimplied = object@implied,
        lavh1 = lavh1,
        lavoptions = lavoptions,
        attr_delta = FALSE,
        attr_t_dvgvd = TRUE
      ),
      silent = TRUE
    )
    if (inherits(nvcov, "try-error")) {
      return(NULL)
    }
    b_meat <- attr(nvcov, "tDVGVD")
  } else {
    # casewise (or clusterwise) scores
    if (!lavmodel@estimator %in% c("ML", "PML")) {
      return(NULL)
    }
    # allow information.meat/h1.information.meat overrides, as in
    # lav_model_nvcov_robust_sandwich()
    lavoptions2 <- lavoptions
    if (!is.null(lavoptions$information.meat)) {
      lavoptions2$information <- lavoptions$information.meat
    }
    if (!is.null(lavoptions$h1.information.meat)) {
      lavoptions2$h1.information <- lavoptions$h1.information.meat
    }
    b_meat <- try(
      lav_model_info_firstorder(
        lavmodel = lavmodel,
        lavsamplestats = object@SampleStats,
        lavdata = object@Data,
        lavimplied = object@implied,
        lavh1 = lavh1,
        lavcache = object@Cache,
        lavoptions = lavoptions2,
        check_pd = FALSE,
        extra = FALSE,
        augmented = FALSE,
        inverted = FALSE
      ),
      silent = TRUE
    )
    if (inherits(b_meat, "try-error")) {
      return(NULL)
    }
  }

  b_meat
}

# scaled, adjusted, and generalized ('robust') versions of a restricted
# test statistic (score or Wald), following Satorra (2000, section 17.3)
#
# arguments:
#   stat : the standard (normal-theory) test statistic
#   df   : its degrees of freedom (m = number of tested restrictions)
#   m1   : A J A'      (m x m, 'bread-only' middle matrix)
#   m2   : A J B J A'  (m x m, sandwich middle matrix)
#          (equivalently: n * A Avar(thetahat) A')
#   v, n : the generalized statistic is n * t(v) %*% solve(m2) %*% v
#          (score test: v = A J h with n the sample size; Wald test:
#           v = a(thetahat), with m1/m2 based on vcov and n = 1)
#
# the trace quantities follow from eq. (22)-(23):
#   tr(U Gamma) = tr[ (A J B J A') (A J A')^{-1} ] = tr(m2 %*% solve(m1))
lav_test_satorra2000 <- function(stat = NULL, df = NULL,
                                 m1 = NULL, m2 = NULL,
                                 v = NULL, n = 1) {
  out <- list(
    scaling.factor = as.numeric(NA),
    stat.scaled = as.numeric(NA),
    df.scaled = df,
    p.value.scaled = as.numeric(NA),
    stat.adjusted = as.numeric(NA),
    df.adjusted = as.numeric(NA),
    p.value.adjusted = as.numeric(NA),
    stat.robust = as.numeric(NA),
    df.robust = df,
    p.value.robust = as.numeric(NA)
  )

  m1_inv <- try(MASS::ginv(m1), silent = TRUE)
  if (!inherits(m1_inv, "try-error")) {
    ug <- m2 %*% m1_inv
    tr_ug <- sum(diag(ug))
    tr_ug2 <- sum(ug * t(ug))

    # scaled
    if (is.finite(tr_ug) && tr_ug > 0) {
      out$scaling.factor <- tr_ug / df
      out$stat.scaled <- stat / out$scaling.factor
      out$p.value.scaled <- 1 - pchisq(out$stat.scaled, df = df)

      # adjusted (for mean and variance), with fractional df
      if (is.finite(tr_ug2) && tr_ug2 > 0) {
        df_adj <- tr_ug * tr_ug / tr_ug2
        out$stat.adjusted <- stat * df_adj / tr_ug
        out$df.adjusted <- df_adj
        out$p.value.adjusted <- 1 - pchisq(out$stat.adjusted, df = df_adj)
      }
    }
  }

  # generalized ('robust')
  if (!is.null(v)) {
    m2_inv <- try(MASS::ginv(m2), silent = TRUE)
    if (!inherits(m2_inv, "try-error")) {
      out$stat.robust <- as.numeric(n * t(v) %*% m2_inv %*% v)
      out$p.value.robust <- 1 - pchisq(out$stat.robust, df = df)
    }
  }

  out
}

# compute the loglikelihood of the data, given the current values of the
# model parameters
lav_model_loglik <- function(lavdata = NULL,
                             lavsamplestats = NULL,
                             lavh1 = NULL,
                             lavimplied = NULL,
                             lavmodel = NULL,
                             lavoptions = NULL,
                             lavcache = NULL) {
  ngroups <- lavdata@ngroups

  logl_group <- rep(as.numeric(NA), ngroups)

  # should compute logl, or return NA?
  logl_ok <- FALSE
  if (lavoptions$estimator %in% c("ML", "MML")) {
    # check if everything is numeric, OR if we have exogenous
    # factor with 2 levels only
    # if(all(lavdata@ov$type == "numeric")) {
    logl_ok <- TRUE
    # } else {
    if (lavoptions$fixed.x == FALSE) {
      exo_idx <- which(lavdata@ov$exo == 1L)
      for (i in exo_idx) {
        if (lavdata@ov$nlev[i] > 1L) {
          logl_ok <- FALSE
        }
      }
    }
    # nlevels + fiml
    # if(lavdata@nlevels > 1L && lavsamplestats@missing.flag) {
    #    logl.ok <- FALSE
    # }
  }

  # lavsamplestats filled in? (not if no data, or samplestats = FALSE)
  if (length(lavsamplestats@ntotal) == 0L ||
      (!is.null(lavoptions$samplestats) && !lavoptions$samplestats)) {
    logl_ok <- FALSE
  }

  # catch all-zero Sigma (new in 0.6-20)
  nblocks <- lavmodel@nblocks
  for (b in seq_len(nblocks)) {
    # except for level-2, where Sigma could be all zero (0.7-1)
    if (lavdata@nlevels > 1L && ((b %% lavdata@nlevels) != 1)) {
      next
    }
    if (lavmodel@conditional.x) {
      if (all(lavimplied$res.cov[[b]] == 0)) {
        logl_ok <- FALSE
      }
    } else {
      if (all(lavimplied$cov[[b]] == 0)) {
        logl_ok <- FALSE
      }
    }
  }


  if (logl_ok) {
    for (g in seq_len(ngroups)) {
      if (lavdata@nlevels > 1L &&
          (length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L)) {
        # random slopes (rv() modifier): conditional (on the exogenous
        # covariates) observed-data loglikelihood
        rs <- lavcache[[g]]$rs
        if (is.null(rs)) {
          rs_info <- lav_mvn_cl_rs_info(
            lavmodel = lavmodel, lavdata = lavdata
          )
          rs_stats <- lav_mvn_cl_rs_stats(
            y1 = lavdata@X[[g]], lp = lavdata@Lp[[g]],
            rs_info = rs_info
          )
          rs <- list(info = rs_info, stats = rs_stats)
        }
        logl_group[g] <- lav_mvn_cl_rs_m2ll(
          lavmodel = lavmodel, rs = rs,
          log2pi = TRUE, minus_two = FALSE
        )
      } else if (lavdata@nlevels > 1L) {
        # here, we assume only 2 levels, at [[1]] and [[2]]
        if (lavmodel@conditional.x) {
          res_sigma_w <- lavimplied$res.cov[[(g - 1) * 2 + 1]]
          res_int_w <- lavimplied$res.int[[(g - 1) * 2 + 1]]
          res_pi_w <- lavimplied$res.slopes[[(g - 1) * 2 + 1]]

          res_sigma_b <- lavimplied$res.cov[[(g - 1) * 2 + 2]]
          res_int_b <- lavimplied$res.int[[(g - 1) * 2 + 2]]
          res_pi_b <- lavimplied$res.slopes[[(g - 1) * 2 + 2]]
        } else {
          sigma_w <- lavimplied$cov[[(g - 1) * 2 + 1]]
          mu_w <- lavimplied$mean[[(g - 1) * 2 + 1]]
          sigma_b <- lavimplied$cov[[(g - 1) * 2 + 2]]
          mu_b <- lavimplied$mean[[(g - 1) * 2 + 2]]
        }

        if (lavsamplestats@missing.flag) {
          if (lavmodel@conditional.x) {
            # TODO
            logl_group[g] <- as.numeric(NA)
          } else {
            logl_group[g] <-
              lav_mvn_cl_mi_loglik_samp_2l(
                y1 = lavdata@X[[g]],
                y2 = lavsamplestats@YLp[[g]][[2]]$Y2,
                lp = lavdata@Lp[[g]],
                mp = lavdata@Mp[[g]],
                mu_w = mu_w, sigma_w = sigma_w,
                mu_b = mu_b, sigma_b = sigma_b,
                loglik_x = lavsamplestats@YLp[[g]][[2]]$loglik.x,
                log2pi = TRUE, minus_two = FALSE
              )
          }
        } else {
          # complete case
          if (lavmodel@conditional.x) {
            logl_group[g] <-
              lav_mvreg_cl_loglik_samp_2l(
                ylp          = lavsamplestats@YLp[[g]],
                lp           = lavdata@Lp[[g]],
                res_sigma_w  = res_sigma_w,
                res_int_w    = res_int_w,
                res_pi_w     = res_pi_w,
                res_sigma_b  = res_sigma_b,
                res_int_b    = res_int_b,
                res_pi_b     = res_pi_b,
                sinv_method  = "eigen",
                log2pi       = TRUE,
                minus_two    = FALSE
              )
          } else {
            logl_group[g] <-
              lav_mvn_cl_loglik_samp_2l(
                ylp          = lavsamplestats@YLp[[g]],
                lp           = lavdata@Lp[[g]],
                mu_w         = mu_w,
                sigma_w      = sigma_w,
                mu_b         = mu_b,
                sigma_b      = sigma_b,
                sinv_method  = "eigen",
                log2pi       = TRUE,
                minus_two    = FALSE
              )
          }
        } # complete
        # end multilevel
      } else if (lavsamplestats@missing.flag) {
        if (lavmodel@conditional.x) {
          # the likelihood is already conditional on x; no x-marginal
          # correction needed
          logl_group[g] <- lav_mvreg_mi_loglik_samp(
            yp         = lavsamplestats@missing[[g]],
            res_int    = lavimplied$res.int[[g]],
            res_slopes = lavimplied$res.slopes[[g]],
            res_cov    = lavimplied$res.cov[[g]]
          )
        } else {
          x_idx <- lavsamplestats@x.idx[[g]]
          x_mean <- x_cov <- NULL
          if (length(x_idx) > 0L) {
            x_mean <- lavh1$implied$mean[[g]][x_idx]
            x_cov <- lavh1$implied$cov[[g]][x_idx, x_idx, drop = FALSE]
          }
          logl_group[g] <- lav_mvn_mi_loglik_samp(
            yp     = lavsamplestats@missing[[g]],
            mu     = lavimplied$mean[[g]],
            sigma_1  = lavimplied$cov[[g]],
            x_idx  = lavsamplestats@x.idx[[g]],
            x_mean = x_mean, # not needed? should be part of Sigma
            x_cov  = x_cov
          ) # not needed at all!
        }
      } else { # single-level, complete data
        if (lavoptions$conditional.x) {
          # FIXME: use lavh1
          logl_group[g] <- lav_mvreg_loglik_samp(
            sample_res_int    = lavsamplestats@res.int[[g]],
            sample_res_slopes = lavsamplestats@res.slopes[[g]],
            sample_res_cov    = lavsamplestats@res.cov[[g]],
            sample_mean_x     = lavsamplestats@mean.x[[g]],
            sample_cov_x      = lavsamplestats@cov.x[[g]],
            sample_nobs       = lavsamplestats@nobs[[g]],
            res_int           = lavimplied$res.int[[g]],
            res_slopes        = lavimplied$res.slopes[[g]],
            res_cov           = lavimplied$res.cov[[g]],
            sinv_method       = "eigen"
          )
        } else {
          if (lavoptions$meanstructure) {
            mu <- lavimplied$mean[[g]]
          } else {
            mu <- lavsamplestats@mean[[g]]
          }
          logl_group[g] <- lav_mvn_loglik_samp(
            sample_mean = lavsamplestats@mean[[g]],
            sample_cov  = lavsamplestats@cov[[g]],
            sample_nobs = lavsamplestats@nobs[[g]],
            mu          = mu,
            sigma_1     = lavimplied$cov[[g]],
            x_idx       = lavsamplestats@x.idx[[g]],
            x_mean      = lavsamplestats@mean.x[[g]],
            x_cov       = lavsamplestats@cov.x[[g]],
            sinv_method = "eigen",
            sigma_inv   = NULL
          )
        }
      } # complete
    } # g
  } # logl.ok is TRUE

  # logl
  logl <- sum(logl_group)

  # number of parameters, taking into account any equality constraints
  npar <- lavmodel@nx.free
  ceq_simple_only <- lavmodel@ceq.simple.only
  cin_simple_only <- lavmodel@cin.simple.only

  if (ceq_simple_only) {
    # nothing to do
  } else if (!cin_simple_only && nrow(lavmodel@con.jac) > 0L) {
    ceq_idx <- attr(lavmodel@con.jac, "ceq.idx")
    if (length(ceq_idx) > 0L) {
      neq <- qr(lavmodel@con.jac[ceq_idx, , drop = FALSE])$rank
      npar <- npar - neq
    }
  }

  # logl
  logl <- sum(logl_group)

  if (logl_ok) {
    # AIC
    aic <- lav_fit_aic(logl = logl, npar = npar)

    # BIC
    bic <- lav_fit_bic(logl = logl, npar = npar, n = lavsamplestats@ntotal)

    # BIC2
    bic2 <- lav_fit_sabic(
      logl = logl, npar = npar,
      n = lavsamplestats@ntotal
    )
  } else {
    aic <- bic <- bic2 <- as.numeric(NA)
  }

  out <- list(
    loglik = logl,
    loglik.group = logl_group,
    npar = npar,
    ntotal = lavsamplestats@ntotal,
    AIC = aic,
    BIC = bic,
    BIC2 = bic2,
    estimator = lavoptions$estimator,
    conditional.x = lavoptions$conditional.x,
    fixed.x = lavoptions$fixed.x
  )

  out
}

# class definitions
#
# initial version: YR 25/03/2009
# added ModelSyntax: YR 02/08/2010
# deleted ModelSyntax: YR 01/11/2010 (using flattened model syntax now)
# ldw 20/11/2023: replace 'representation()' by 'slots='

setClass("lavData",
  slots = c(
    data.type = "character", # "full", "moment" or "none"
    group = "character", # group variable
    ngroups = "integer", # number of groups
    group.label = "character", # group labels
    block.label = "character", # block labels
    cluster = "character", # cluster variable(s)
    nlevels = "integer", # number of levels
    level.label = "character", # level labels
    std.ov = "logical", # standardize observed variables?
    nobs = "list", # effective number of observations
    norig = "list", # original number of observations
    ov.names = "list", # variable names (per group)
    ov.names.x = "list", # exo variable names (per group)
    ov.names.l = "list", # names per level
    ov.names.aux = "list", # auxiliary variable names (per group)
    # ov.types          = "list",           # variable types (per group)
    # ov.idx            = "list",           # column indices (all observed variables)
    ordered = "character", # ordered variables
    weights = "list", # sampling weights (per group)
    sampling.weights = "character", # sampling weights variable
    ov = "list", # variable table
    ov.aux = "list", # variable table for auxiliary/instrument variables
    case.idx = "list", # case indices per group
    missing = "character", # "listwise" or not?
    Mp = "list", # if not complete, missing patterns
    # we need this here, to get nobs right!
    Rp = "list", # response patterns (categorical only)
    Lp = "list", # level patterns
    eXo = "list", # local copy exo only
    aux = "list", # local copy auxiliary variables only (per group)
    X = "list" # local copy
  )
)


setClass("lavSampleStats", # sample moments
  slots = c(
    var                = "list", # observed variances (per group)
    cov                = "list", # observed var/cov matrix (per group)
    mean               = "list", # observed mean vector (per group)
    th                 = "list", # thresholds for non-numeric var (per group)
    th.idx             = "list", # th index (0 for numeric)
    th.names           = "list", # threshold names

    res.cov            = "list", # residual var/cov matrix (if conditional.x)
    res.var            = "list", # residual variances
    res.th             = "list", # residual thresholds
    res.th.nox         = "list", # residual thresholds ignoring x
    res.slopes         = "list", # slopes exo (if conditional.x)
    res.int            = "list", # intercepts (if conditional.x)

    mean.x             = "list", # mean exo
    cov.x              = "list", # variance/covariance exo
    bifreq             = "list", # bivariate frequency tables
    group.w            = "list", # group weight

    nobs               = "list", # effective number of obs (per group)
    ntotal             = "numeric", # total number of obs (all groups)
    ngroups            = "integer", # number of groups
    x.idx              = "list", # x.idx if fixed.x = TRUE

    icov               = "list", # inverse of observed cov (per group)
    cov.log.det        = "list", # log det of observed cov (per group)
    res.icov           = "list",
    res.cov.log.det    = "list",
    # ridge.constant   = "numeric",        # ridge constant (per group)
    # ridge.constant.x = "numeric",        # ridge constant (per group) for eXo
    ridge              = "numeric",
    WLS.obs            = "list", # all relevant observed stats in a vector
    WLS.V              = "list", # weight matrix for GLS/WLS
    WLS.VD             = "list", # diagonal of weight matrix only
    NACOV              = "list", # N times the asymptotic covariance matrix
    NACOV.user         = "logical", # user-specified NACOV?

    missing.flag       = "logical", # missing patterns?
    missing            = "list", # missingness information
    missing.h1         = "list", # h1 model

    YLp                = "list", # cluster/level information

    zero.cell.tables   = "list" # bivariate tables with empty cells
  )
)

setClass("lavModel", # MATRIX representation of the sem model
  slots = c(
    GLIST              = "list", # list of all model matrices (for all groups)
    dimNames           = "list", # dim names for the model matrices
    isSymmetric        = "logical", # model matrix symmetric?
    mmSize             = "integer", # model matrix size (unique only)

    representation     = "character", # stub, until we define more classes
    modprop            = "list", # model properties
    meanstructure      = "logical",
    correlation        = "logical",
    composites         = "logical",
    categorical        = "logical",
    multilevel         = "logical",
    group.w.free       = "logical",
    link               = "character",
    nblocks            = "integer",
    ngroups            = "integer", # only for rsem!! (which uses rsem:::computeDelta)
    nefa               = "integer",
    nmat               = "integer",
    nvar               = "integer",
    num.idx            = "list",
    th.idx             = "list",
    nx.free            = "integer",
    nx.unco            = "integer",
    nx.user            = "integer",
    m.free.idx         = "list",
    x.free.idx         = "list",
    # m.unco.idx        = "list",           # always the same as m.free.idx
    x.unco.idx         = "list",
    m.user.idx         = "list",
    x.user.idx         = "list",
    x.def.idx          = "integer",
    x.ceq.idx          = "integer",
    x.cin.idx          = "integer",
    x.free.var.idx     = "integer",
    ceq.simple.only    = "logical",
    ceq.simple.K       = "matrix",
    eq.constraints     = "logical",
    eq.constraints.K   = "matrix",
    eq.constraints.k0  = "numeric",
    def.function       = "function",
    ceq.function       = "function",
    ceq.jacobian       = "function",
    ceq.JAC            = "matrix",
    ceq.rhs            = "numeric",
    ceq.linear.idx     = "integer",
    ceq.nonlinear.idx  = "integer",
    cin.simple.only    = "logical",
    cin.function       = "function",
    cin.jacobian       = "function",
    cin.JAC            = "matrix",
    cin.rhs            = "numeric",
    cin.linear.idx     = "integer",
    cin.nonlinear.idx  = "integer",
    ceq.efa.JAC        = "matrix",
    con.jac            = "matrix",
    con.lambda         = "numeric",
    nexo               = "integer",
    conditional.x      = "logical",
    fixed.x            = "logical",
    parameterization   = "character",
    ov.x.dummy.ov.idx  = "list",
    ov.x.dummy.lv.idx  = "list",
    ov.y.dummy.ov.idx  = "list",
    ov.y.dummy.lv.idx  = "list",
    ov.efa.idx         = "list",
    lv.efa.idx         = "list",
    rv.ov              = "list",
    rv.lv              = "list",
    H                  = "list",
    lv.order           = "list",
    estimator          = "character",
    estimator.args     = "list"
  )
)

setClass("Fit",
  slots = c(
    npar               = "integer", # number of free parameters
    # ndat              = "integer",
    # df                = "integer",
    x                  = "numeric", # x
    partrace           = "matrix", # parameter trace
    start              = "numeric", # starting values (only for other packages)
    est                = "numeric", # estimated values (only for other packages)
    se                 = "numeric", # standard errors
    fx                 = "numeric",
    fx.group           = "numeric",
    logl               = "numeric",
    logl.group         = "numeric",
    iterations         = "integer", # number of iterations
    converged          = "logical",
    control            = "list",
    Sigma.hat          = "list",
    Mu.hat             = "list",
    TH                 = "list",
    test               = "list"
  )
)

setClass("lavaan",
  slots = c(
    version            = "character", # lavaan version
    call               = "call", # matched call
    timing             = "list", # timing information
    Options            = "list", # lavOptions
    ParTable           = "list", # parameter table user-specified model
    pta                = "list", # parameter table attributes
    Data               = "lavData", # full data
    SampleStats        = "lavSampleStats", # sample statistics
    Model              = "lavModel", # internal matrix representation
    Cache              = "list", # housekeeping stuff
    Fit                = "Fit", # fitted results
    boot               = "list", # bootstrap results
    optim              = "list", # optimizer results
    loglik             = "list", # loglik values and info
    implied            = "list", # model implied moments
    vcov               = "list", # vcov
    test               = "list", # test
    h1                 = "list", # unrestricted model results
    baseline           = "list", # baseline model results
    internal           = "list", # optional slot, for internal use
    external           = "list" # optional slot, for add-on packages
  )
)

setClass("lavaanList",
  slots = c(
    version            = "character", # lavaan version
    call               = "call", # matched call
    Options            = "list", # lavOptions
    ParTable           = "list",
    pta                = "list",
    Data               = "lavData", # from first dataset (ngroups!)
    Model              = "lavModel", # based on first dataset
    meta               = "list",
    timingList         = "list",
    ParTableList       = "list",
    DataList           = "list",
    SampleStatsList    = "list",
    CacheList          = "list",
    vcovList           = "list",
    testList           = "list",
    optimList          = "list",
    impliedList        = "list",
    h1List             = "list",
    loglikList         = "list",
    baselineList       = "list",
    internalList       = "list",
    funList            = "list",
    external           = "list" # optional slot, for add-on packages
  )
)

# the univariate (weighted) linear model
#
# - scores/gradient/hessian
# - including the residual variance!
#
# YR 30 Dec 2019 (replacing the old lav_ols.R routines)
# YR/LDW 2026: refactored (see lav_uvbv_common.R for the shared cache/minfns
#              design); fixed a dormant bug in lav_uvreg_update_fit() (yhat
#              lost the slopes; never triggered, as all callers pass
#              intercept-only models)
#
# The parameter vector is theta = c(beta, evar), with beta = c(int, slopes)
# and evar the residual variance; yhat = X1 %*% beta with X1 = [1, X].
#
# fit object (list) contract -- see lav_uvreg_fit():
#   theta, nexo, int_idx, slope_idx, beta_idx, var_idx, y, wt, x, yhat
# The yhat values are those of the *last* evaluation inside nlminb;
# lav_uvreg_update_fit() recomputes them for new parameter values.

lav_uvreg_fit <- function(y = NULL,
                          x = NULL,
                          wt = NULL,
                          optim_method = "nlminb",
                          control = list(),
                          output = "list") {
  # check weights
  wt <- lav_uvbv_wt_check(y = y, wt = wt)

  # optim.method
  minfns <- lav_uvbv_optim_fns(optim_method, lav_uvreg_min_fns())

  # create cache environment
  cache <- lav_uvreg_init_cache(y = y, x = x, wt = wt)

  # optimize -- only changes from defaults
  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb, cache = cache
  )

  if (output == "cache") {
    return(cache)
  }

  # return results as a list (to be compatible with lav_bvord.R)
  list(
    theta = optim$par,
    nexo = cache$nexo,
    int_idx = cache$int_idx,
    slope_idx = cache$slope_idx,
    beta_idx = cache$beta_idx,
    var_idx = cache$var_idx,
    y = cache$y,
    wt = cache$wt,
    x = cache$x1[, -1L, drop = FALSE],
    yhat = cache$yhat
  )
}

# prepare cache environment
lav_uvreg_init_cache <- function(y = NULL,
                                 x = NULL,
                                 wt = rep(1, length(y)),
                                 parent = parent.frame()) {
  # y
  y <- as.vector(y)

  # X
  if (is.null(x)) {
    nexo <- 0L
    x1 <- matrix(1, length(y), 1)
  } else {
    x <- unname(x)
    nexo <- ncol(x)
    x1 <- cbind(1, x, deparse.level = 0)

    # new in 0.6-17: check if X is full rank
    if (!anyNA(x)) {
      if (qr(x)$rank < ncol(x)) {
        lav_msg_stop(gettext("matrix of exogenous covariates is rank deficient!
                    (i.e., some x variables contain redundant information)"))
      }
    }
  }

  # nobs
  if (is.null(wt)) {
    n <- length(y)
  } else {
    n <- sum(wt)
  }


  # indices of free parameters
  int_idx <- 1L
  slope_idx <- seq_len(nexo) + 1L
  beta_idx <- c(int_idx, slope_idx)
  var_idx <- 1L + nexo + 1L

  # starting values + crossprod
  if (any(is.na(y)) || any(is.na(x1))) {
    missing_idx <- which(apply(cbind(y, x1), 1, function(x) any(is.na(x))))
    y_tmp <- y[-missing_idx]
    x1_tmp <- x1[-missing_idx, , drop = FALSE]
    wt_tmp <- wt[-missing_idx]
    fit_lm <- stats::lm.wfit(y = y_tmp, x = x1_tmp, w = wt_tmp)
    theta_evar <- sum(fit_lm$residuals * wt_tmp * fit_lm$residuals) /
                  sum(wt_tmp)

    lav_crossprod <- lav_mat_crossprod
  } else {
    fit_lm <- stats::lm.wfit(y = y, x = x1, w = wt)
    theta_evar <- sum(fit_lm$residuals * wt * fit_lm$residuals) / sum(wt)

    lav_crossprod <- base::crossprod
  }
  theta_beta <- unname(fit_lm$coefficients)
  theta <- c(theta_beta, theta_evar)

  list2env(
    list(
      y = y, x1 = x1, wt = wt, n = n,
      int_idx = int_idx, beta_idx = beta_idx,
      var_idx = var_idx, slope_idx = slope_idx, nexo = nexo,
      lav_crossprod = lav_crossprod,
      theta = theta
    ),
    parent = parent
  )
}

# compute total (log)likelihood
lav_uvreg_loglik <- function(y = NULL,
                             x = NULL,
                             wt = rep(1, length(y)),
                             cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvreg_fit(y = y, x = x, wt = wt, output = "cache")
  }
  lav_uvreg_loglik_cache(cache = cache)
}

lav_uvreg_loglik_cache <- function(cache = NULL) {
  with(cache, {    # nolint start
    # free parameters
    beta <- theta[beta_idx]
    evar <- theta[var_idx]

    yhat <- drop(x1 %*% beta)
    logliki <- dnorm(y, mean = yhat, sd = sqrt(evar), log = TRUE)

    # total weighted log-likelihood
    loglik <- sum(wt * logliki, na.rm = TRUE)

    return(loglik)
  })               # nolint end
}

# casewise scores
lav_uvreg_sc <- function(y = NULL,
                             x = NULL,
                             wt = rep(1, length(y)),
                             cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvreg_fit(y = y, x = x, wt = wt, output = "cache")
  }
  lav_uvreg_sc_cache(cache = cache)
}

lav_uvreg_sc_cache <- function(cache = NULL) {
  with(cache, {   # nolint start
    res <- y - yhat
    resw <- res * wt
    evar2 <- evar * evar

    scores_beta <- 1 / evar * x1 * resw
    scores_evar <- -wt / (2 * evar) + 1 / (2 * evar2) * res * resw

    return(cbind(scores_beta, scores_evar, deparse.level = 0))
  })              # nolint end
}

# gradient
lav_uvreg_grad <- function(y = NULL,
                               x = NULL,
                               wt = rep(1, length(y)),
                               cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvreg_fit(y = y, x = x, wt = wt, output = "cache")
  }
  lav_uvreg_grad_cache(cache = cache)
}

lav_uvreg_grad_cache <- function(cache = NULL) {
  with(cache, {    # nolint start
    res <- y - yhat
    resw <- res * wt
    evar2 <- evar * evar

    dx_beta <- colSums(1 / evar * x1 * resw, na.rm = TRUE)
    dx_var <- sum(-wt / (2 * evar) + 1 / (2 * evar2) * res * resw, na.rm = TRUE)

    return(c(dx_beta, dx_var))
  })               # nolint end
}

# compute total Hessian
# note: this evaluates at the *current* cache state (as left behind by the
# last nlminb evaluation) -- the beta/evar/yhat/res quantities must already
# be present in the cache
lav_uvreg_hessian <- function(y = NULL,
                              x = NULL,
                              wt = rep(1, length(y)),
                              cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvreg_fit(y = y, x = x, wt = wt, output = "cache")
  }
  lav_uvreg_hessian_cache(cache = cache)
}

lav_uvreg_hessian_cache <- function(cache = NULL) {
  with(cache, {      # nolint start
    dx2_beta <- -1 / evar * lav_crossprod(x1 * wt, x1)
    dx_beta_var <- -1 / (evar2) * lav_crossprod(x1, resw)

    sq_evar <- sqrt(evar)
    sq_evar6 <- sq_evar * sq_evar * sq_evar * sq_evar * sq_evar * sq_evar
    dx2_var <- (sum(wt, na.rm = TRUE) / (2 * evar2) -
      1 / sq_evar6 * sum(resw * res, na.rm = TRUE))

    hessian <- rbind(cbind(dx2_beta, dx_beta_var, deparse.level = 0),
      cbind(t(dx_beta_var), dx2_var, deparse.level = 0),
      deparse.level = 0
    )
    return(hessian)
  })                    # nolint end
}

# nlminb objective/gradient/hessian (see lav_uvbv_common.R)
lav_uvreg_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_uvreg_loglik_cache,
    grad_fun = lav_uvreg_grad_cache,
    hessian_fun = lav_uvreg_hessian_cache
  )
}

# update fit object with new parameters
lav_uvreg_update_fit <- function(fit_y = NULL,
                                 evar_new = NULL, beta_new = NULL) {
  if (is.null(evar_new) && is.null(beta_new)) {
    return(fit_y)
  }

  if (!is.null(evar_new)) {
    fit_y$theta[fit_y$var_idx] <- evar_new
  }
  if (!is.null(beta_new)) {
    fit_y$theta[fit_y$beta_idx] <- beta_new
  }

  beta <- fit_y$theta[fit_y$beta_idx]
  x <- fit_y$x # was fit_y$X (always NULL): yhat lost the slopes
  x1 <- cbind(1, x, deparse.level = 0)

  fit_y$yhat <- drop(x1 %*% beta)

  fit_y
}

# main function used by various bootstrap related functions
# this function draws the bootstrap samples, and estimates the
# free parameters for each bootstrap sample
#
# return COEF matrix of size r x npar (r = number of bootstrap samples)
#
# Ed. 9 mar 2012
#
# Notes: - faulty runs are simply ignored (with a warning)
#        - default r=1000
#
# Updates: - now we have a separate @Data slot, we only need to transform once
#            for the bollen.stine bootstrap (13 dec 2011)
#          - bug fix: we need to 'update' the fixed.x variances/covariances
#            for each bootstrap draw!!
#
# Question: if fixed.x=TRUE, should we not keep X fixed, and bootstrap Y
#           only, conditional on X?? How to implement the conditional part?

# YR 27 Aug 2022: - add keep_idx argument
#                 - always return 'full' set of bootstrap results, including
#                   failed runs (as NAs)
#                 - idx nonadmissible/error solutions as an attribute
#                 - thanks to keep_idx, it is easy to replicate/investigate
#                   these cases if needed

# YR 10 Nov 2024: - detect sam object

lavBootstrap <- function(object,                                   # nolint
                            r = 1000L,
                            type = "ordinary",
                            verbose = FALSE,
                            fun = "coef",
                            # return.boot = FALSE, # no use, as boot stores
                            #                      # sample indices differently
                            keep_idx = FALSE,
                            parallel = c("no", "multicore", "snow"),
                            ncpus = min(2L, max(1L, parallel::detectCores() - 2L, na.rm = TRUE)),
                            cl = NULL,
                            iseed = NULL,
                            h0_rmsea = NULL,
                            ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)

  # check object
  object <- lav_object_check_version(object)

  # checks
  type_1 <- tolower(type) # overwritten if nonparametric
  stopifnot(
    inherits(object, "lavaan"),
    type_1 %in% c(
      "nonparametric", "ordinary",
      "bollen.stine", "parametric", "yuan"
    )
  )
  if (!missing(verbose)) {
    current_verbose <- lav_verbose()
    if (lav_verbose(verbose))
      on.exit(lav_verbose(current_verbose), TRUE)
  }
  if (type_1 == "nonparametric") {
    type_1 <- "ordinary"
  }
  if (missing(parallel)) {
    parallel <- "no"
  }
  parallel <- match.arg(parallel)

  # check if options$se is not bootstrap, otherwise, we get an infinite loop
  if (object@Options$se == "bootstrap") {
    object@Options$se <- "standard"
  }

  # check if options$test is not bollen.stine
  if ("bollen.stine" %in% object@Options$test) {
    object@Options$test <- "standard"
  }

  # check for conditional.x = TRUE
  if (object@Model@conditional.x) {
    lav_msg_stop(gettext(
      "this function is not (yet) available if conditional.x = TRUE"))
  }

  lavoptions <- list(
    parallel = parallel, ncpus = ncpus, cl = cl,
    iseed = iseed
  )

  tocall <- c(list(as.name("lav_bootstrap_internal")),
    list(object = object, lavoptions = lavoptions, r = r,
    show_progress = verbose, type = type_1, fun = fun,
    keep_idx = keep_idx, h0_rmsea = h0_rmsea), dotdotdot)
  out <- eval(as.call(tocall))

  # new in 0.6-12: always warn for failed and nonadmissible runs
  nfailed <- length(attr(out, "error.idx")) # zero if NULL
  if (nfailed > 0L) {
    lav_msg_warn(gettextf(
      "%s bootstrap runs failed or did not converge.", nfailed))
  }

  notok <- length(attr(out, "nonadmissible")) # zero if NULL
  if (notok > 0L) {
     lav_msg_warn(gettextf(
      "%s bootstrap runs resulted in nonadmissible solutions.", notok))
  }

  out
}

# we need an internal version to be called from VCOV and lav_model_test
# when there is no lavaan object yet!
lav_bootstrap_internal <- function(object = NULL,
                                   lavdata = NULL,
                                   lavmodel = NULL,
                                   lavsamplestats = NULL,
                                   lavoptions = NULL,
                                   lavpartable = NULL,
                                   r = 1000L,
                                   show_progress = FALSE,
                                   type = "ordinary",
                                   fun = "coef",
                                   check_post = TRUE,
                                   keep_idx = FALSE,
                                   # return.boot     = FALSE,
                                   h0_rmsea = NULL,
                                   efa_reference = NULL,
                                   ...) {
  # warning: avoid use of 'options', 'sample' (both are used as functions
  # below...
  # options -> opt
  # sample -> samp

  # object slots
  fun_orig <- fun
  has_sam_object_flag <- FALSE
  if (!is.null(object)) {
    stopifnot(inherits(object, "lavaan"))
    # check for sam object
    if (!is.null(object@internal$sam.method)) {
      has_sam_object_flag <- TRUE
    }
    lavdata_1 <- object@Data
    lavmodel_1 <- object@Model
    lavsamplestats_1 <- object@SampleStats
    lavoptions_1 <- object@Options
    if (!is.null(lavoptions)) {
      lavoptions_1$parallel <- lavoptions$parallel
      lavoptions_1$ncpus <- lavoptions$ncpus
      lavoptions_1$cl <- lavoptions$cl
      lavoptions_1$iseed <- lavoptions$iseed
    }
    lavpartable_1 <- object@ParTable
    fun <- match.fun(fun)
    t0 <- fun(object, ...)
    t_star <- matrix(as.numeric(NA), r, length(t0))
    colnames(t_star) <- names(t0)
  } else {
    # internal version!
    lavdata_1 <- lavdata
    lavmodel_1 <- lavmodel
    lavsamplestats_1 <- lavsamplestats
    lavoptions_1 <- lavoptions
    lavpartable_1 <- lavpartable
    if (fun == "coef") {
      if (!is.null(efa_reference)) {
        # rotated EFA/ESEM: the bootstrap draws live in the (larger) *rotated*
        # parameter space, not the unrotated optimizer space (see #376)
        t_star <- matrix(as.numeric(NA), r, efa_reference$npar)
        colnames(t_star) <- efa_reference$names
      } else {
        t_star <- matrix(as.numeric(NA), r, lavmodel_1@nx.free)
      }
      lavoptions_1$test <- "none"
    } else if (fun == "test") {
      t_star <- matrix(as.numeric(NA), r, 1L)
      lavoptions_1$test <- "standard"
    } else if (fun == "coeftest") {
      t_star <- matrix(as.numeric(NA), r, lavmodel_1@nx.free + 1L)
      lavoptions_1$test <- "standard"
    }
  }

  # always shut off some options:
  current_verbose <- lav_verbose()
  if (missing(show_progress)) {
    show_progress <- current_verbose
  }
  if (lav_verbose(FALSE)) on.exit(lav_verbose(current_verbose))
  lavoptions_1$check.start <- FALSE
  lavoptions_1$check.post <- FALSE
  lavoptions_1$optim.attempts <- 1L # can save a lot of time
  lavoptions_1$fit.by.level <- FALSE # no partially saturated models

  # if internal or fun == "coef", we can shut off even more
  if (is.null(object) || (is.character(fun_orig) && fun_orig == "coef")) {
    lavoptions_1$baseline <- FALSE
    lavoptions_1$h1 <- FALSE
    lavoptions_1$loglik <- FALSE
    lavoptions_1$implied <- FALSE
    lavoptions_1$store.vcov <- FALSE
    lavoptions_1$se <- "none"
    if (fun_orig == "coef") {
      lavoptions_1$test <- "none"
    }
    # the (MI)IV estimator needs the (unrestricted) h1 moments to find the
    # instruments and estimate each equation, so keep h1 for IV refits
    if (identical(lavoptions_1$estimator, "IV")) {
      lavoptions_1$h1 <- TRUE
    }
  }

  # clustered/multilevel data?
  # if so, we resample whole (level-2) clusters with replacement, keeping
  # all the level-1 observations within a selected cluster together, instead
  # of resampling individual observations (the 'cluster bootstrap'; see eg
  # Field & Welsh, 2007). This applies both to single-level models with
  # cluster-robust standard errors (cluster = ) and to two-level models.
  clustered_flag <- length(lavdata_1@cluster) > 0L
  if (clustered_flag) {
    # only the (nonparametric) ordinary bootstrap makes sense here
    if (type != "ordinary") {
      lav_msg_stop(gettextf(
        "the %s bootstrap is not available for clustered/multilevel data;
         only the (nonparametric) ordinary bootstrap is supported.",
        dQuote(type)))
    }
    # for now, a single cluster (level-2) variable only
    if (length(lavdata_1@cluster) > 1L) {
      lav_msg_stop(gettext(
        "the cluster bootstrap is currently only available for a single
         cluster (level-2) variable."))
    }
    lav_msg_note(gettext(
      "clustered/multilevel data detected: the bootstrap resamples whole
       clusters (not individual observations)."))
  }

  # bollen.stine, yuan, or parametric: we need the Sigma.hat values
  if (type == "bollen.stine" || type == "parametric" || type == "yuan") {
    sigma_hat <- lav_model_sigma(lavmodel = lavmodel_1)
    mu_hat <- lav_model_mu(lavmodel = lavmodel_1)
  }

  # can we use the original data, or do we need to transform it first?
  if (type == "bollen.stine" || type == "yuan") {
    # check if data is continuous
    if (lavmodel_1@categorical) {
      lav_msg_stop(gettext(
        "bollen.stine/yuan bootstrap not available for categorical/ordinal data"
        ))
    }
    # check if data is complete
    if (lavoptions_1$missing != "listwise") {
      lav_msg_stop(gettext(
        "bollen.stine/yuan bootstrap not available for missing data"))
    }
    data_x <- vector("list", length = lavdata_1@ngroups)
  } else {
    data_x <- lavdata_1@X
  }

  # if bollen.stine, transform data here
  if (type == "bollen.stine") {
    for (g in 1:lavsamplestats_1@ngroups) {
      sigma_sqrt <- lav_mat_sym_sqrt(sigma_hat[[g]])
      s_inv_sqrt <- lav_mat_sym_sqrt(lavsamplestats_1@icov[[g]])

      # center (needed???)
      x_1 <- scale(lavdata_1@X[[g]], center = TRUE, scale = FALSE)

      # transform
      x_1 <- x_1 %*% s_inv_sqrt %*% sigma_sqrt

      # add model-based mean
      if (lavmodel_1@meanstructure) {
        x_1 <- scale(x_1, center = (-1 * mu_hat[[g]]), scale = FALSE)
      }

      # transformed data
      data_x[[g]] <- x_1
    }

    # if yuan, transform data here
  } else if (type == "yuan") {
    # page numbers refer to Yuan et al, 2007
    # Define a function to find appropriate value of a
    # (p. 272); code supplied 16 jun 2016 by Cheng & Wu
    search_a <- function(f0, d, p) {
      if (f0 == 0) {
        a0 <- 0
        return(a0)
      }
      max_a <- 1 / (1 - min(d)) - 1e-3
      # starting value; Yuan p. 272
      a0 <- min(sqrt(2 * f0 / sum((d - 1)^2)), max_a)

      # See Yuan p. 280
      for (i in 1:50) {
        dia <- a0 * d + (1 - a0)
        g1 <- -sum(log(dia)) + sum(dia) - p
        dif <- g1 - f0
        if (abs(dif) < 1e-6) {
          return(a0)
        }
        g2 <- a0 * sum((d - 1)^2 / dia)
        a0 <- min(max(a0 - dif / g2, 0), max_a)
      }
      # if search fails to converge in 50 iterations
      lav_msg_warn(gettext("yuan bootstrap search for `a` did not converge.
                           h0_rmsea may be too large."))
      a0
    }

    # Now use g.a within each group
    for (g in 1:lavsamplestats_1@ngroups) {
      s <- lavsamplestats_1@cov[[g]]
      # test is in Fit slot
      ghat <- object@test[[1]]$stat.group[[g]]
      df <- object@test[[1]]$df
      sigmahat <- sigma_hat[[g]]
      nmv <- nrow(sigmahat)
      n <- nrow(lavdata_1@X[[g]])

      # Calculate tauhat_1, middle p. 267.
      # Yuan et al note that tauhat_1 could be negative;
      # if so, we need to let S.a = Sigmahat. (see middle p 275)
      ifelse(length(h0_rmsea) == 0,
        tau_hat <- (ghat - df) / (n - 1), # middle p 267
        tau_hat <- df * (h0_rmsea * h0_rmsea)
      ) # middle p 273

      if (tau_hat >= 0) {
        # from Cheng and Wu
        el_1 <- t(chol(sigmahat))
        ese <- forwardsolve(el_1, t(forwardsolve(el_1, s)))
        d <- eigen(ese, symmetric = TRUE, only.values = TRUE)$values
        if ("a" %in% names(list(...))) {
          a <- list(...)$a
        } else {
          # Find a to minimize g.a
          a <- search_a(tau_hat, d, nmv)
        }
        # Calculate S_a (p. 267)
        s_a <- a * s + (1 - a) * sigmahat
      } else {
        s_a <- sigmahat
      }

      # Transform the data (p. 263)
      s_a_sqrt <- lav_mat_sym_sqrt(s_a)
      s_inv_sqrt <- lav_mat_sym_sqrt(lavsamplestats_1@icov[[g]])

      x_1 <- lavdata_1@X[[g]]
      x_1 <- x_1 %*% s_inv_sqrt %*% s_a_sqrt

      # transformed data
      data_x[[g]] <- x_1
    }
  }

  # run bootstraps
  fn <- function(b) {
    # create bootstrap sample, and generate new 'data' object
    if (type == "bollen.stine" || type == "ordinary" || type == "yuan") {
      # take a bootstrap sample for each group
      boot_idx_1 <- vector("list", length = lavdata_1@ngroups)
      # for clustered data: (relabeled) cluster ids for the resampled data
      boot_clus_1 <- if (clustered_flag) {
        vector("list", length = lavdata_1@ngroups)
      } else {
        NULL
      }
      # Note: we generate the bootstrap indices separately for each group, to
      #       ensure the group sizes (or, when clustered, the number of
      #       clusters per group) do not change!
      for (g in 1:lavdata_1@ngroups) {
        if (clustered_flag) {
          # cluster bootstrap: resample whole (level-2) clusters, keeping all
          # observations within a selected cluster together. The number of
          # clusters is held fixed; the total N may vary across samples.
          cl_idx <- lavdata_1@Lp[[g]]$cluster.idx[[2L]]
          cl_rows <- split(seq_along(cl_idx), cl_idx) # row indices / cluster
          nclus <- length(cl_rows)
          stopifnot(nclus > 1L)
          cl_boot <- sample.int(nclus, replace = TRUE)
          boot_idx <- unlist(cl_rows[cl_boot], use.names = FALSE)
          # relabel: each resampled cluster instance gets a fresh id, so a
          # cluster drawn more than once is treated as several distinct
          # clusters
          new_clus <- rep.int(seq_along(cl_boot), lengths(cl_rows[cl_boot]))
          boot_clus_1[[g]] <- matrix(new_clus, ncol = 1L)
        } else {
          # exclude empty cases (all-NA rows) from the resampling pool: they
          # carry no information and are excluded from the analysis N, but are
          # kept in @X for bookkeeping. Resampling them would make the
          # effective N vary across bootstrap samples (see github issue #520).
          empty_idx <- lavdata_1@Mp[[g]]$empty.idx
          if (length(empty_idx) > 0L) {
            pool <- seq_len(nrow(lavdata_1@X[[g]]))[-empty_idx]
          } else {
            pool <- seq_len(nrow(lavdata_1@X[[g]]))
          }
          stopifnot(length(pool) > 1L)
          boot_idx <- pool[sample.int(length(pool), replace = TRUE)]
        }
        boot_idx_1[[g]] <- boot_idx
        data_x[[g]] <- data_x[[g]][boot_idx, , drop = FALSE]
      }
      new_data <- lav_data_update(
        lavdata = lavdata_1, new_x = data_x,
        boot_idx = boot_idx_1,
        boot_clus = boot_clus_1,
        lavoptions = lavoptions_1
      )
    } else { # parametric! (using sign-invariant method for reproducibility)
      for (g in 1:lavdata_1@ngroups) {
        data_x[[g]] <- lav_mvrnorm(
          n = lavdata_1@nobs[[g]],
          sigma_1 = sigma_hat[[g]],
          mu = mu_hat[[g]]
        )
      }
      new_data <- lav_data_update(
        lavdata = lavdata_1, new_x = data_x,
        lavoptions = lavoptions_1
      )
    }

    # show progress?
    if (show_progress) {
      cat(if (interactive()) "\r" else "",
          "  ... bootstrap draw number:", sprintf("%4d", b))
    }
    boot_sample_stats <- try(lav_samp_from_data(
      lavdata       = new_data,
      lavoptions    = lavoptions_1
    ), silent = TRUE)
    if (inherits(boot_sample_stats, "try-error")) {
      if (show_progress) {
        cat("     FAILED: creating sample statistics\n")
        cat(boot_sample_stats[1])
      }
      out <- as.numeric(NA)
      attr(out, "nonadmissible.flag") <- TRUE
      if (keep_idx) {
        attr(out, "BOOT.idx") <- boot_idx_1
      }
      return(out)
    }
    if (has_sam_object_flag) {
      # also need h1
      booth1 <- lav_h1_implied_logl(lavdata = new_data,
        lavsamplestats = boot_sample_stats, lavpartable = lavpartable_1,
        lavoptions = lavoptions_1)
    }

    # do we need to update Model slot? only if we have fixed exogenous
    # covariates, as their variances/covariances are stored in GLIST
    if (lavmodel_1@fixed.x &&
        length(lav_pt_vnames(lavpartable_1, "ov.x")) > 0L) {
      model_boot <- NULL
    } else {
      model_boot <- lavmodel_1
    }

    # fit model on bootstrap sample
    if (has_sam_object_flag) {
      new_object <- object
      new_object@Data <- new_data
      new_object@SampleStats <- boot_sample_stats
      new_object@h1 <- booth1
      # what about lavoptions?
      fit_boot <- suppressWarnings(try(sam(new_object, se = "none"),
                                       silent = FALSE)) # show what is wrong
    } else {
      fit_boot <- suppressWarnings(try(lavaan(
        slot_options = lavoptions_1,
        slot_par_table = lavpartable_1,
        slot_model = model_boot,
        slot_sample_stats = boot_sample_stats,
        slot_data = new_data
      ), silent = FALSE))
    }
    if (inherits(fit_boot, "try-error")) {
      if (show_progress) {
        cat("     FAILED: with ERROR message\n")
      }
      out <- as.numeric(NA)
      attr(out, "nonadmissible.flag") <- TRUE
      if (keep_idx) {
        attr(out, "BOOT.idx") <- boot_idx_1
      }
      return(out)
    }
    if (!fit_boot@optim$converged) {
      if (show_progress) {
        cat("     FAILED: no convergence\n")
      }
      out <- as.numeric(NA)
      attr(out, "nonadmissible.flag") <- TRUE
      if (keep_idx) {
        attr(out, "BOOT.idx") <- boot_idx_1
      }
      return(out)
    }

    # extract information we need
    if (is.null(object)) { # internal use only!
      if (fun == "coef") {
        if (!is.null(efa_reference)) {
          # rotated EFA/ESEM: extract the *rotated* parameters and align the
          # factors (sign + order) to the original solution (see #376)
          out <- try(lav_efa_bootstrap_align_coef(
            lavmodel = fit_boot@Model, lavpartable = fit_boot@ParTable,
            ref_lambda = efa_reference$ref_lambda
          ), silent = TRUE)
          if (inherits(out, "try-error")) {
            if (show_progress) {
              cat("     FAILED: aligning rotated bootstrap solution\n")
            }
            out <- as.numeric(NA)
            attr(out, "nonadmissible.flag") <- TRUE
            if (keep_idx) {
              attr(out, "BOOT.idx") <- boot_idx_1
            }
            return(out)
          }
        } else {
          out <- fit_boot@optim$x
        }
      } else if (fun == "test") {
        out <- fit_boot@test[[1L]]$stat
      } else if (fun == "coeftest") {
        out <- c(fit_boot@optim$x, fit_boot@test[[1L]]$stat)
      }
    } else { # general use
      out <- try(as.numeric(fun(fit_boot, ...)), silent = TRUE)
    }
    if (inherits(out, "try-error")) {
      if (show_progress) {
        cat("     FAILED: applying fun to fit.boot\n")
      }
      out <- as.numeric(NA)
      attr(out, "nonadmissible.flag") <- TRUE
      if (keep_idx) {
        attr(out, "BOOT.idx") <- boot_idx_1
      }
      return(out)
    }

    # check if the solution is admissible
    admissible_flag <- suppressWarnings(lavInspect(fit_boot, "post.check"))
    attr(out, "nonadmissible.flag") <- !admissible_flag

    if (show_progress) {
      cat(
        "   OK -- niter = ",
        sprintf("%3d", fit_boot@optim$iterations), " fx = ",
        sprintf("%11.9f", fit_boot@optim$fx),
        # proposed by issue #519 Shu Fai Cheung
        if (admissible_flag) " " else "n", "",
        if (interactive()) "" else "\n"
      )
    }

    if (keep_idx) {
      # add BOOT.idx (for all groups)
      attr(out, "BOOT.idx") <- boot_idx_1
    }

    out
  } # end-of-fn

  # get parallelization options
  parallel <- lavoptions_1$parallel[1]
  ncpus <- lavoptions_1$ncpus
  cl <- lavoptions_1[["cl"]] # often NULL
  iseed <- lavoptions_1[["iseed"]] # often NULL

  # the next 8 lines are borrowed from the boot package
  have_mc <- have_snow <- FALSE
  if (parallel != "no" && ncpus > 1L) {
    if (parallel == "multicore") {
      have_mc <- .Platform$OS.type != "windows"
    } else if (parallel == "snow") {
      have_snow <- TRUE
    }
    if (!have_mc && !have_snow) ncpus <- 1L
    loadNamespace("parallel") # before recording seed!
  }

  # iseed:
  # this follows a proposal of Shu Fai Cheung (see github issue #240)
  # - iseed is used for both serial and parallel
  # - if iseed is not set, iseed is generated + .Random.seed created/updated
  #     -> tmp.seed <- NA
  # - if iseed is set: don't touch .Random.seed (if it exists)
  #     -> tmp.seed <- .Random.seed (if it exists)
  #     -> tmp.seed <- NULL (if it does not exist)
  if (is.null(iseed)) {
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      runif(1)
    }
    # identical(temp.seed, NA): Will not change .Random.seed in GlobalEnv
    temp_seed <- NA
    iseed <- runif(1, 0, 999999999)
  } else {
    if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      temp_seed <-
        get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    } else {
      # is.null(temp.seed): Will remove .Random.seed in GlobalEnv
      #                     if serial.
      #                     If parallel, .Random.seed will not be touched.
      temp_seed <- NULL
    }
  }
  if (!(ncpus > 1L && (have_mc || have_snow))) { # Only for serial
    set.seed(iseed)
  }


  # this is adapted from the boot function in package boot
  rr <- r
  if (show_progress) {
    cat("\n")
  }
  res <- if (ncpus > 1L && (have_mc || have_snow)) {
    if (have_mc) {
      rngkind_old <- RNGkind() # store current kind
      RNGkind("L'Ecuyer-CMRG") # to allow for reproducible results
      set.seed(iseed)
      parallel::mclapply(seq_len(rr), fn, mc.cores = ncpus)
    } else if (have_snow) {
      list(...) # evaluate any promises
      if (is.null(cl)) {
        cl <- parallel::makePSOCKcluster(rep("localhost", ncpus))
        # # No need for
        # if(RNGkind()[1L] == "L'Ecuyer-CMRG")
        # clusterSetRNGStream() always calls `RNGkind("L'Ecuyer-CMRG")`
        parallel::clusterSetRNGStream(cl, iseed = iseed)
        res <- parallel::parLapply(cl, seq_len(rr), fn)
        parallel::stopCluster(cl)
        res
      } else {
        parallel::parLapply(cl, seq_len(rr), fn)
      }
    }
  } else {
    lapply(seq_len(rr), fn)
  }
  if (show_progress) {
    cat("\n")
  }
  # restore old RNGkind()
  if (ncpus > 1L && have_mc) {
    RNGkind(rngkind_old[1], rngkind_old[2], rngkind_old[3])
  }

  # fill in container
  t_star[] <- do.call("rbind", res)

  # handle errors
  error_idx <- which(sapply(res, function(x) is.na(x[1L])))
  attr(t_star, "error.idx") <- error_idx # could be integer(0L)

  # handle nonadmissible solutions
  if (check_post) {
    notok <- which(sapply(res, attr, "nonadmissible.flag"))
    if (length(error_idx) > 0L) {
      notok <- notok[-which(notok %in% error_idx)]
    }
    attr(t_star, "nonadmissible") <- notok
  }

  # store iseed
  attr(t_star, "seed") <- iseed

  # handle temp.seed
  if (!is.null(temp_seed) && !identical(temp_seed, NA)) {
    assign(".Random.seed", temp_seed, envir = .GlobalEnv)                   # nolint start
  } else if (is.null(temp_seed) && !(ncpus > 1L && (have_mc || have_snow))) {
    # serial
    rm(.Random.seed, pos = 1)
  } else if (is.null(temp_seed) && (ncpus > 1L && have_mc)) {
    # parallel/multicore only
    rm(.Random.seed, pos = 1) # because set used set.seed()                 # nolint end
  }

  # store BOOT.idx per group
  if (keep_idx) {
    boot_idx_1 <- vector("list", length = lavsamplestats_1@ngroups)
    for (g in 1:lavsamplestats_1@ngroups) {
      # note that failed runs (NULL) are removed (for now)
      boot_idx_1[[g]] <- do.call(
        "rbind",
        lapply(res, function(x) attr(x, "BOOT.idx")[[g]])
      )
    }
    attr(t_star, "boot.idx") <- boot_idx_1
  }

  #    # No use, as boot package stores the sample indices differently
  #    # See boot:::boot.array() versus lav_bootstrap_indices()
  #    if(return.boot) {
  #        # mimic output boot function
  #
  #        if(is.null(object)) {
  #            stop("lavaan ERROR: return.boot = TRUE requires a full
  #                  lavaan object")
  #        }
  #
  #        # we start with ordinary only for now
  #        stopifnot(type == "ordinary")
  #
  #        if(! type %in% c("ordinary", "parametric")) {
  #            stop("lavaan ERROR: only ordinary and parametric bootstrap are
  #                  supported if return.boot = TRUE")
  #        } else {
  #            sim <- type
  #        }
  #
  #        statistic. <- function(data, idx) {
  #                         data.boot <- data[idx,]
  #                         fit.boot <- update(object, data = data.boot)
  #                         out <- try(fun(fit.boot, ...), silent = TRUE)
  #                         if(inherits(out, "try-error")) {
  #                             out <- rep(as.numeric(NA), length(t0))
  #                         }
  #                         out
  #                     }
  #        attr(t.star, "seed") <- NULL
  #        attr(t.star, "nonadmissible") <- NULL
  #        out <- list(t0 = t0, t = t.star, r = RR,
  #                    data = lavInspect(object, "data"),
  #                    seed = iseed, statistic = statistic.,
  #                    sim = sim, call = mc)
  #
  #        #if(sim == "parametric") {
  #        #    ran.gen. <- function() {} # TODO
  #        #    out <- c(out, list(ran.gen = ran.gen, mle = mle))
  #        #} else if(sim == "ordinary") {
  #            stype <- "i"
  #            strata <- rep(1, nobs(object))
  #            weights <- 1/tabulate(strata)[strata]
  #            out <- c(out, list(stype = stype, strata = strata,
  #                                weights = weights))
  #        #}
  #
  #        class(out) <- "boot"
  #        return(out)
  #    }
  #
  t_star
}

# create matrix with indices to reconstruct the bootstrap samples
# per group
# (originally needed for BCa confidence intervals)
#
# rows are the (r) bootstrap runs
# columns are the (N) observations
#
# simple version: no strata, no weights
#
lav_bootstrap_indices <- function(r = 0L,
                                  nobs = list(0L), # per group
                                  parallel = "no",
                                  ncpus = 1L,
                                  cl = NULL,
                                  iseed = NULL,
                                  merge_groups = FALSE,
                                  return_freq = FALSE) {
  # iseed must be set!
  stopifnot(!is.null(iseed))

  if (return_freq && !merge_groups) {
    lav_msg_stop(gettext("return.freq only available if merge.groups = TRUE"))
  }

  if (is.integer(nobs)) {
    nobs <- list(nobs)
  }

  # number of groups
  ngroups <- length(nobs)

  # mimic 'random' sampling from lav_bootstrap_internal:

  # the next 7 lines are borrowed from the boot package
  have_mc <- have_snow <- FALSE
  parallel <- parallel[1]
  if (parallel != "no" && ncpus > 1L) {
    if (parallel == "multicore") {
      have_mc <- .Platform$OS.type != "windows"
    } else {
      if (parallel == "snow") have_snow <- TRUE
    }
    if (!have_mc && !have_snow) ncpus <- 1L
    loadNamespace("parallel") # before recording seed!
  }
  temp_seed <- NULL
  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    temp_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  }
  if (!(ncpus > 1L && (have_mc || have_snow))) { # Only for serial
    set.seed(iseed)
  }

  # fn() returns indices per group
  fn <- function(b) {
    boot_idx_1 <- vector("list", length = ngroups)
    offset_1 <- cumsum(c(0, unlist(nobs)))
    for (g in 1:ngroups) {
      stopifnot(nobs[[g]] > 1L)
      boot_idx <- sample.int(nobs[[g]], replace = TRUE)
      if (merge_groups) {
        boot_idx_1[[g]] <- boot_idx + offset_1[g]
      } else {
        boot_idx_1[[g]] <- boot_idx
      }
    }
    boot_idx_1
  }

  rr <- r
  res <- if (ncpus > 1L && (have_mc || have_snow)) {
    if (have_mc) {
      rngkind_old <- RNGkind() # store current kind
      RNGkind("L'Ecuyer-CMRG") # to allow for reproducible results
      set.seed(iseed)
      parallel::mclapply(seq_len(rr), fn, mc.cores = ncpus)
    } else if (have_snow) {
      # list(...) # evaluate any promises
      if (is.null(cl)) {
        cl <- parallel::makePSOCKcluster(rep("localhost", ncpus))
        parallel::clusterSetRNGStream(cl, iseed = iseed)
        res <- parallel::parLapply(cl, seq_len(rr), fn)
        parallel::stopCluster(cl)
        res
      } else {
        parallel::parLapply(cl, seq_len(rr), fn)
      }
    }
  } else {
    lapply(seq_len(rr), fn)
  }

  # restore old RNGkind()
  if (ncpus > 1L && have_mc) {
    RNGkind(rngkind_old[1], rngkind_old[2], rngkind_old[3])
  }

  # handle temp.seed
  if (!is.null(temp_seed) && !identical(temp_seed, NA)) {            # nolint start
    assign(".Random.seed", temp_seed, envir = .GlobalEnv)
  } else if (is.null(temp_seed) && !(ncpus > 1L && (have_mc || have_snow))) {
    # serial
    rm(.Random.seed, pos = 1)
  } else if (is.null(temp_seed) && (ncpus > 1L && have_mc)) {
    # parallel/multicore only
    rm(.Random.seed, pos = 1) # because set used set.seed()          # nolint end
  }


  # assemble IDX
  boot_idx_1 <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    # FIXME: handle failed runs
    boot_idx_1[[g]] <- do.call("rbind", lapply(res, "[[", g))
  }

  # merge groups
  if (merge_groups) {
    out <- do.call("cbind", boot_idx_1)
  } else {
    out <- boot_idx_1
  }

  # NOTE: the order of the indices is different from the boot package!
  # we fill in the matrix 'row-wise' (1 row = sample(N, replace = TRUE)),
  # while boot fills in the matrix 'column-wise'
  # this also explains why we get different results with return.boot = TRUE
  # despite using the same iseed

  # return frequencies instead?
  if (return_freq && merge_groups) {
    out <- t(apply(out, 1L, tabulate, ncol(out)))
  }

  out
}


# -------------------------------------------------------------------------- #
# Shared bootstrap confidence-interval helpers
#
# These helpers are used by both parameterEstimates() and lavEffects() so that
# the bootstrap confidence interval machinery lives in a single place.
# -------------------------------------------------------------------------- #

# local copy of 'norm.inter' from the boot package (not exported there).
# Given a vector 't' of (bootstrap) realizations and a vector 'alpha' of
# probabilities, it returns, per requested alpha, cbind(round(rk, 2), q) where
# 'q' is the interpolated order statistic (the boot-package interpolation on
# the normal quantile scale). Used for the 'basic', 'perc' and 'bca' methods.
lav_bootstrap_norm_inter <- function(t, alpha) {
  t <- t[is.finite(t)]
  tmp_r <- length(t)
  rk <- (tmp_r + 1) * alpha
  if (!all(rk > 1 & rk < tmp_r)) {
    lav_msg_warn(gettext("Extreme order statistics used as endpoints."))
  }
  k <- trunc(rk)
  inds <- seq_along(k)
  out <- inds
  kvs <- k[k > 0 & k < tmp_r]
  tstar <- sort(t, partial = sort(union(c(1, tmp_r), c(kvs, kvs + 1))))
  ints <- (k == rk)
  if (any(ints)) out[inds[ints]] <- tstar[k[inds[ints]]]
  out[k == 0] <- tstar[1L]
  out[k == tmp_r] <- tstar[tmp_r]
  not <- function(v) xor(rep(TRUE, length(v)), v)
  temp <- inds[not(ints) & k != 0 & k != tmp_r]
  temp1 <- qnorm(alpha[temp])
  temp2 <- qnorm(k[temp] / (tmp_r + 1))
  temp3 <- qnorm((k[temp] + 1) / (tmp_r + 1))
  tk <- tstar[k[temp]]
  tk1 <- tstar[k[temp] + 1L]
  out[temp] <- tk + (temp1 - temp2) / (temp3 - temp2) * (tk1 - tk)
  cbind(round(rk, 2), out)
}

# Compute bootstrap confidence interval bounds for a set of 'p' quantities.
#
# Arguments:
#   boot_t : R x p matrix of bootstrap realizations of the quantities (may
#            contain NA/Inf values; these are ignored). Not used for the
#            "norm" method.
#   t0     : length-p vector of point estimates.
#   boot_ci_type : one of "norm", "basic", "perc", "bca.simple", "bca".
#   level  : the confidence level.
#   se     : length-p vector of standard errors (required for "norm").
#   bias   : length-p vector of bias estimates (only for "norm"); if NULL, it
#            is computed as colMeans(boot_t) - t0.
#   acc    : length-p vector of acceleration constants (only for "bca"); for
#            "bca.simple" the acceleration is taken to be zero.
#
# Returns a p x 2 matrix with the lower and upper confidence bounds. For
# quantities with no bootstrap variability (or with se = 0), the bounds default
# to the point estimate.
lav_bootstrap_ci <- function(boot_t = NULL, t0, boot_ci_type = "perc",
                             level = 0.95, se = NULL, bias = NULL,
                             acc = NULL) {
  boot_ci_type <- match.arg(boot_ci_type,
    c("norm", "basic", "perc", "bca.simple", "bca"))
  p <- length(t0)
  ci <- cbind(t0, t0)
  dimnames(ci) <- NULL

  if (boot_ci_type == "norm") {
    stopifnot(!is.null(se))
    if (is.null(bias)) {
      bias <- colMeans(boot_t, na.rm = TRUE) - t0
    }
    fac <- qnorm(c((1 - level) / 2, 1 - (1 - level) / 2))
    ci <- (t0 - bias) + se %o% fac
    return(ci)
  }

  stopifnot(!is.null(boot_t))
  boot_t <- as.matrix(boot_t)

  if (boot_ci_type == "basic") {
    alpha <- (1 + c(level, -level)) / 2
    qq <- apply(boot_t, 2L, lav_bootstrap_norm_inter, alpha)
    ci <- 2 * ci - t(qq[c(3L, 4L), , drop = FALSE])
  } else if (boot_ci_type == "perc") {
    alpha <- (1 + c(-level, level)) / 2
    qq <- apply(boot_t, 2L, lav_bootstrap_norm_inter, alpha)
    ci <- t(qq[c(3L, 4L), , drop = FALSE])
  } else if (boot_ci_type %in% c("bca.simple", "bca")) {
    alpha <- (1 + c(-level, level)) / 2
    zalpha <- qnorm(alpha)
    if (is.null(acc)) {
      acc <- numeric(p) # bca.simple: acceleration = 0
    }
    for (i in seq_len(p)) {
      t <- boot_t[, i]
      t <- t[is.finite(t)]
      # no bootstrap variability? keep the point estimate
      if (length(t) < 2L || var(t) == 0) {
        next
      }
      w <- qnorm(sum(t < t0[i]) / length(t))
      a <- acc[i]
      adj_alpha <- pnorm(w + (w + zalpha) / (1 - a * (w + zalpha)))
      qq <- lav_bootstrap_norm_inter(t, adj_alpha)
      ci[i, ] <- qq[, 2L]
    }
  }

  ci
}

# Compute the acceleration constants for the BCa method, using empirical
# influence values obtained by regressing the bootstrap realizations 'boot_t'
# (R x p) on the bootstrap frequency design matrix 'design' (R x k, including
# the intercept column). Returns a length-p vector of acceleration constants.
lav_bootstrap_acceleration <- function(boot_t, design) {
  boot_t <- as.matrix(boot_t)
  tmp_lm <- stats::lm.fit(x = design, y = boot_t)
  tmp_beta <- tmp_lm$coefficients
  # lm.fit() returns 'coefficients' as a plain vector when y has a single
  # column; coerce to a matrix so the indexing below always works
  if (!is.matrix(tmp_beta)) {
    tmp_beta <- matrix(tmp_beta, ncol = ncol(boot_t))
  }
  tmp_beta <- unname(tmp_beta)[-1L, , drop = FALSE]
  tmp_ll <- rbind(0, tmp_beta)
  apply(tmp_ll, 2L, function(x) {
    tmp_l <- x - mean(x)
    sum(tmp_l^3) / (6 * sum(tmp_l^2)^1.5)
  })
}

# Build the design matrix used to compute empirical influence values for the
# BCa method: an intercept column followed by the bootstrap frequency matrix,
# with one reference case per group removed (to avoid collinearity, since the
# frequencies within a group sum to the group size). Returns an
# R_successful x k matrix, where R_successful is the number of non-failed
# bootstrap runs. Performs the BCa prerequisite checks and stops with an
# informative message if they are not met. 'object' is a fitted lavaan object
# that was fitted with se = "bootstrap"; 'error_idx' are the indices of the
# failed bootstrap runs (as stored in the attribute of lav_inspect_boot()).
lav_bootstrap_bca_design <- function(object, error_idx = integer(0L)) {
  lavoptions <- object@Options
  nobs   <- object@SampleStats@nobs
  ntotal <- object@SampleStats@ntotal

  # BCa is not available for the cluster bootstrap: the frequency-matrix
  # reconstruction below assumes individual observations are resampled with
  # a fixed total N, which is not the case when whole clusters are resampled.
  if (length(object@Data@cluster) > 0L) {
    lav_msg_stop(gettext(
      "BCa confidence intervals are not available for clustered/multilevel
       data (cluster bootstrap)."))
  }

  # number of bootstrap runs requested (the 'bootstrap' option is a list)
  r <- lavoptions$bootstrap
  if (is.list(r)) {
    r <- r$R
  }

  # we need enough (successful) bootstrap runs
  n_ok <- r - length(error_idx)
  if (n_ok < ntotal) {
    lav_msg_stop(gettextf(
      "BCa confidence intervals require more (successful) bootstrap runs
       (%1$s) than the number of observations (%2$s).", n_ok, ntotal))
  }

  # does not work with sampling weights (yet)
  if (!is.null(object@Data@weights[[1]])) {
    lav_msg_stop(gettext(
      "BCa confidence intervals not available in the presence of sampling
       weights."))
  }

  # we need the seed that was used for the bootstrap
  boot_seed <- attr(lav_inspect_boot(object), "seed")
  if (is.null(boot_seed)) {
    lav_msg_stop(gettext("Seed not available in the bootstrap object."))
  }

  # reconstruct the frequency matrix (one column per observation)
  tmp_freq <- lav_bootstrap_indices(
    r = r, nobs = nobs, parallel = lavoptions$parallel[1],
    ncpus = lavoptions$ncpus, cl = lavoptions[["cl"]],
    iseed = boot_seed, return_freq = TRUE, merge_groups = TRUE
  )
  if (length(error_idx) > 0L) {
    tmp_freq <- tmp_freq[-error_idx, , drop = FALSE]
  }

  # remove the first case of each group, then add the intercept column
  first_idx <- sapply(object@Data@case.idx, "[[", 1L)
  cbind(1, tmp_freq[, -first_idx, drop = FALSE])
}

# EFA: exploratory factor analysis
#
# EFA is implemented as a special version of ESEM
# - it is therefore a wrapper around the lavaan() function to simplify
#   the input
# - a lavaan model is generated with a single 'block' that can be rotated
# - the 'default' output produces output that is more in line with traditional
#   EFA software (in R) like factanal() and fa() from the psych package

# YR 20 Sept 2022 - first version

efa <- function(data = NULL,
                nfactors = 1L,
                sample_cov = NULL,
                sample_nobs = NULL,
                rotation = "geomin",
                rotation_args = list(),
                ov_names = NULL,
                bounds = "pos.var",
                ...,
                output = "efa") {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, TRUE)
  # rotation.args deprecation handling
  if (!missing(rotation_args)) {
    lav_deprecated_args("rotation", "rotation_args")
  }

  if (is.list(rotation)) {
    rotation_args <- modifyList(list(), rotation)
    rotation_args <- lav_snake_case(rotation_args)
    rotation <- rotation[[1L]]
  }

  # twolevel?
  twolevel_flag <- !is.null(dotdotdot$cluster)

  # multiple groups?
  group_flag <- !is.null(dotdotdot$group)

  # sampling weights?
  sampling_weights_flag <- !is.null(dotdotdot$sampling.weights)

  # if data= argument is used, convert to data.frame (eg matrix, tibble, ...)
  if (!is.null(data) && !inherits(data, "lavMoments")) {
    data <- as.data.frame(data)
  }

  # handle ov_names
  if (!is.null(data) && inherits(data, "lavMoments")) {
    if ("sample.cov" %in% names(data)) {
      ov_names <- rownames(data$sample.cov)
      if (is.null(ov_names)) {
        ov_names <- colnames(data$sample.cov)
      }
    } else {
      lav_msg_stop(gettext(
        "When data= is of class lavMoments, it must contain sample.cov"))
    }

  } else if (!is.null(data) && inherits(data, "data.frame")) {
    if (length(ov_names) > 0L) {
      names_1 <- ov_names
      if (twolevel_flag) {
        names_1 <- c(names_1, dotdotdot$cluster)
      }
      if (group_flag) {
        names_1 <- c(names_1, dotdotdot$group)
      }
      if (sampling_weights_flag) {
        names_1 <- c(names_1, dotdotdot$sampling.weights)
      }
      data <- data[, names_1, drop = FALSE]
    } else {
      ov_names <- names(data)
      if (twolevel_flag) {
        if (!all(dotdotdot$cluster %in% ov_names)) {
          lav_msg_stop(gettextf(
            "cluster variable(s) %s not found in data.",
            lav_msg_view(dotdotdot$cluster[!dotdotdot$cluster %in% ov_names])))
        }
        ov_names <- setdiff(ov_names, dotdotdot$cluster)
      }
      if (group_flag) {
        if (!all(dotdotdot$group %in% ov_names)) {
          lav_msg_stop(gettextf(
            "group variable(s) %s not found in data.",
            lav_msg_view(dotdotdot$group[!dotdotdot$group %in% ov_names])))
        }
        ov_names <- setdiff(ov_names, dotdotdot$group)
      }
      if (sampling_weights_flag) {
        if (!all(dotdotdot$sampling.weights %in% ov_names)) {
          lav_msg_stop(gettextf(
            "sampling.weights variable(s) %s not found in data.",
            lav_msg_view(dotdotdot$sampling.weights[
              !dotdotdot$sampling.weights %in% ov_names])))
        }
        ov_names <- setdiff(ov_names, dotdotdot$sampling.weights)
      }
    }
  } else if (!is.null(sample_cov)) {
    ov_names <- rownames(sample_cov)
    if (is.null(ov_names)) {
      ov_names <- colnames(sample_cov)
    }
  }
  # ov_names?
  if (length(ov_names) == 0L) {
    lav_msg_stop(gettext(
      "could not extract variable names from data or sample_cov"))
  }

  # determine the block structure (blocks are ordered group-major, level-minor,
  # ie block = (g - 1L) * nlevels + level)
  nlevels <- if (twolevel_flag) 2L else 1L
  if (group_flag && !is.null(data) && inherits(data, "data.frame")) {
    ngroups <- length(unique(data[[dotdotdot$group]]))
  } else if (!is.null(sample_cov) && is.list(sample_cov) &&
    !inherits(sample_cov, "lavMoments")) {
    ngroups <- length(sample_cov)
  } else {
    ngroups <- 1L
  }
  nblocks <- ngroups * nlevels

  # normalize 'nfactors' into a list of per-block (integer) vectors, one element
  # per candidate model; a scalar or plain vector keeps its original meaning,
  # ie 'candidate models, the same factor count in every block'
  if (is.list(nfactors)) {
    nfactors_list <- lapply(nfactors, function(x) {
      x <- as.integer(x)
      if (length(x) == 1L) {
        x <- rep.int(x, nblocks)
      }
      if (length(x) != nblocks) {
        lav_msg_stop(gettextf("each element of nfactors must have length 1 or
                              the number of blocks (= %1$s); found length %2$s.",
          nblocks, length(x)))
      }
      x
    })
  } else {
    nfactors_list <- lapply(as.integer(nfactors), function(k) {
      rep.int(k, nblocks)
    })
  }
  # labels for the resulting fits ('nf2', or 'nf1_2' if it varies per block)
  nfactors_labels <- vapply(nfactors_list, function(x) {
    if (length(unique(x)) == 1L) {
      paste0("nf", x[1L])
    } else {
      paste0("nf", paste(x, collapse = "_"))
    }
  }, character(1L))

  # check nfactors
  all_nfactors <- unlist(nfactors_list)
  if (any(all_nfactors < 1L)) {
    lav_msg_stop(gettext("nfactors must be greater than zero."))
  } else {
    # check for maximum number of factors
    # Fixme: can we do this more efficiently? also holds for categorical?
    nvar <- length(ov_names)
    p_star <- nvar * (nvar + 1) / 2
    nfac_max <- 0L
    for (nfac in seq_len(nvar)) {
      # compute number of free parameters
      npar <- nfac * nvar + nfac * (nfac + 1L) / 2 + nvar - nfac^2
      if (npar > p_star) {
        nfac_max <- nfac - 1L
        break
      }
    }
    if (any(all_nfactors > nfac_max)) {
      lav_msg_stop(gettextf("when nvar = %1$s the maximum number of factors
                            is %2$s", nvar, nfac_max))
    }
  }

  # output
  output <- tolower(output)
  if (!output %in% c("lavaan", "efa")) {
    lav_msg_stop(gettext("output= must be either \"lavaan\" or \"efa\""))
  }
  if (output == "lavaan" && length(nfactors_list) > 1L) {
    lav_msg_stop(gettext("when output = \"lavaan\", nfactors must be a
                         single (integer) number."))
  }

  # fit models
  nfits <- length(nfactors_list)
  out <- vector("list", length = nfits)
  for (f in seq_len(nfits)) {
    # generate model syntax
    model_syntax <- lav_syntax_efa(
      ov_names = ov_names,
      nfactors = nfactors_list[[f]],
      nlevels = nlevels,
      ngroups = ngroups
    )
    # call lavaan (using sem())
    fit <- do.call("sem",
      args = c(
        list(
          model = model_syntax,
          data = data,
          sample_cov = sample_cov,
          sample_nobs = sample_nobs,
          rotation = rotation,
          rotation.args = rotation_args,
          bounds = bounds,
          cmd = "efa"
        ),
        dotdotdot
      )
    )

    if (output == "efa") {
      fit@Options$model.type <- "efa"
    }

    out[[f]] <- fit
  }

  # class
  if (nfits == 1L && output == "lavaan") {
    out <- out[[1]]
  } else {
    names(out) <- nfactors_labels
    # add loadings element to the end of the list
    # so we an use the non-generic but useful loadings() function
    # from the stats package
    out$loadings <- lav_efa_get_loadings(out)
    class(out) <- c("efaList", "list")
  }

  out
}

# create `full' parameter table, containing (almost) all parameters
# that could be free
#
# main motivation: univariate scores tests (modification indices)
#
lav_pt_full <- function(partable = NULL,
                              strict_exo = FALSE,
                              free = FALSE, start = FALSE) {
  # check minimum requirements: lhs, op, rhs
  stopifnot(
    !is.null(partable$lhs),
    !is.null(partable$op),
    !is.null(partable$rhs)
  )

  # lavpta?
  lavpta <- lav_pt_attributes(partable)

  # meanstructure
  if (!is.null(lavpta$meanstructure)) {
    meanstructure <- lavpta$meanstructure
  } else {
    # old object
    meanstructure <- any(partable$op == "~1")
  }

  # number of blocks
  # nblocks <- lavpta$nblocks
  ngroups <- lavpta$ngroups
  nlevels <- lavpta$nlevels

  lhs <- rhs <- op <- character(0L)
  block <- group <- level <- integer(0L)

  # new in 0.6-3:
  group_values <- lav_pt_group_values(partable)
  level_values <- lav_pt_level_values(partable)
  if (is.character(group_values[1])) {
    group <- character(0L)
  }
  if (is.character(level_values[1L])) {
    level <- character(0L)
  }

  # block number
  b <- 0L
  for (g in 1:ngroups) {
    for (l in 1:nlevels) {
      # block
      b <- b + 1L

      ov_names <- lavpta$vnames$ov[[b]]
      ov_names_nox <- lavpta$vnames$ov.nox[[b]]
      ov_names_x <- lavpta$vnames$ov.x[[b]]
      ov_names_ind <- lavpta$vnames$ov.ind[[b]]
      ov_names_ord <- lavpta$vnames$ov.ord[[b]]

      lv_names <- lavpta$vnames$lv[[b]]

      # eqs.y, eqs.x
      eqs_names <- unique(c(
        lavpta$vnames$eqs.y[[b]],
        lavpta$vnames$eqs.x[[b]]
      ))
      if (length(eqs_names) > 0L) {
        eqs_y <- eqs_names
        if (strict_exo) {
          x_idx <- which(eqs_names %in% ov_names_x)
          if (length(x_idx) > 0L) {
            eqs_y <- eqs_names[-x_idx]
          }
        }
        eqs_x <- eqs_names
      } else {
        eqs_y <- character(0L)
        eqs_x <- character(0L)
      }

      # 1 "=~"
      l_lhs <- rep(lv_names, each = length(ov_names_nox))
      l_rhs <- rep(ov_names_nox, times = length(lv_names))

      # remove factor ~ eqs.y combinations, if any
      # because they also appear as a regression
      bad_idx <- which(l_lhs %in% lv_names &
        l_rhs %in% eqs_y)
      if (length(bad_idx) > 0L) {
        l_lhs <- l_lhs[-bad_idx]
        l_rhs <- l_rhs[-bad_idx]
      }
      l_op <- rep("=~", length(l_lhs))

      # 2a. "~~" ov ## FIXME: ov.names.nox or ov.names??
      # if(strict.exo) {
      ov <- ov_names_nox
      # } else {
      #    OV <- ov.names
      # }
      nx <- length(ov)
      idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
      ov_lhs <- rep(ov, each = nx)[idx] # fill upper.tri
      ov_rhs <- rep(ov, times = nx)[idx]
      ov_op <- rep("~~", length(ov_lhs))

      # remove dummy indicators that correlate with 'proper'
      # indicators; new in 0.6-14; fixed in 0.6-16
      ov_other <- ov_names[!ov_names %in% c(
        ov_names_ind, ov_names_x,
        eqs_x, eqs_y
      )]
      if (length(ov_other) > 0L) {
        bad_idx <- which((ov_lhs %in% ov_names &
          ov_rhs %in% ov_other) |
          (ov_lhs %in% ov_other &
            ov_rhs %in% ov_names))
        if (length(bad_idx) > 0L) {
          ov_lhs <- ov_lhs[-bad_idx]
          ov_rhs <- ov_rhs[-bad_idx]
          ov_op <- ov_op[-bad_idx]
        }
      }

      # exo ~~
      if (!strict_exo && length(ov_names_x) > 0L) {
        ov <- ov_names_x
        nx <- length(ov)
        idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
        more_lhs <- rep(ov, each = nx)[idx] # fill upper.tri
        more_rhs <- rep(ov, times = nx)[idx]
        ov_lhs <- c(ov_lhs, more_lhs)
        ov_rhs <- c(ov_rhs, more_rhs)
        ov_op <- c(ov_op, rep("~~", length(more_lhs)))
      }

      # 2b. "~~" lv
      nx <- length(lv_names)
      idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
      lv_lhs <- rep(lv_names, each = nx)[idx] # fill upper.tri
      lv_rhs <- rep(lv_names, times = nx)[idx]
      lv_op <- rep("~~", length(lv_lhs))

      # 3 regressions?
      r_lhs <- r_rhs <- r_op <- character(0)
      if (length(eqs_names) > 0L) {
        r_lhs <- rep(eqs_y, each = length(eqs_x))
        r_rhs <- rep(eqs_x, times = length(eqs_y))

        # remove self-arrows
        idx <- which(r_lhs == r_rhs)
        if (length(idx) > 0L) {
          r_lhs <- r_lhs[-idx]
          r_rhs <- r_rhs[-idx]
        }

        # remove indicator ~ factor if they exist
        bad_idx <- which(r_lhs %in% ov_names_ind &
          r_rhs %in% lv_names)
        if (length(bad_idx) > 0L) {
          r_lhs <- r_lhs[-bad_idx]
          r_rhs <- r_rhs[-bad_idx]
        }

        r_op <- rep("~", length(r_rhs))
      }

      # 4. intercepts
      int_lhs <- int_rhs <- int_op <- character(0)
      if (meanstructure) {
        if (strict_exo) {
          int_lhs <- c(ov_names_nox, lv_names)
        } else {
          int_lhs <- c(ov_names, lv_names)
        }
        int_rhs <- rep("", length(int_lhs))
        int_op <- rep("~1", length(int_lhs))
      }

      # 5. thresholds
      th_lhs <- th_rhs <- th_op <- character(0)
      if (length(ov_names_ord) > 0L) {
        th_names <- lavpta$vnames$th[[b]]
        tmp <- strsplit(th_names, "\\|")
        th_lhs <- sapply(tmp, function(x) x[1])
        th_rhs <- sapply(tmp, function(x) x[2])
        th_op <- rep("|", length(th_lhs))
      }

      # 6. scaling parameters
      delta_lhs <- delta_rhs <- delta_op <- character(0)
      if (ngroups > 1L && length(ov_names_ord) > 0L) {
        delta_lhs <- ov_names_ord
        delta_rhs <- ov_names_ord
        delta_op <- rep("~*~", length(delta_lhs))
      }

      # combine
      this_lhs <- c(
        l_lhs, ov_lhs, lv_lhs, r_lhs, int_lhs, th_lhs,
        delta_lhs
      )
      this_rhs <- c(
        l_rhs, ov_rhs, lv_rhs, r_rhs, int_rhs, th_rhs,
        delta_rhs
      )
      this_op <- c(
        l_op, ov_op, lv_op, r_op, int_op, th_op,
        delta_op
      )
      n_el <- length(this_lhs)

      lhs <- c(lhs, this_lhs)
      rhs <- c(rhs, this_rhs)
      op <- c(op, this_op)
      block <- c(block, rep(b, n_el))
      group <- c(group, rep(group_values[g], n_el))
      level <- c(level, rep(level_values[l], n_el))
    } # level
  } # group

  list_1 <- data.frame(
    lhs = lhs, op = op, rhs = rhs, block = block,
    group = group, level = level, stringsAsFactors = FALSE
  )

  if (free) {
    list_1$free <- rep(0L, nrow(list_1))
  }

  if (start) {
    list_1$start <- rep(0, nrow(list_1))
  }

  list_1
}

# various ways to compute a (scaled) difference chi-square test statistic

# - 0.6-13: fix multiple-group UG^2 bug in Satorra.2000 (reported by
#           Gronneberg, Foldnes and Moss) when Satterthwaite = TRUE and
#           ngroups > 1L (use old_approach = TRUE to get the old result)

# All equality-constraint reductions below use lav_con_eq_basis(): one
# orthonormal basis (eq.constraints.K, orthonormalized ceq.simple.K, or the
# ceq.JAC null space when equality constraints coexist with inequality
# constraints/bounds -- where the @eq.constraints/@ceq.simple.only packing
# flags are both FALSE), so K / t(K) round-trips are consistent everywhere.

lav_test_diff_satorra2000 <- function(m1, m0, h1 = TRUE, a_method = "delta",
                                      m_a = NULL,
                                      satterthwaite = FALSE,
                                      scaled_shifted = FALSE,
                                      old_approach = FALSE) {
  if (scaled_shifted) {
    satterthwaite <- TRUE
  }

  # extract information from m1 and m2
  t1 <- m1@test[[1]]$stat
  r1 <- m1@test[[1]]$df

  t0 <- m0@test[[1]]$stat
  r0 <- m0@test[[1]]$df

  # m = difference between the df's
  m <- r0 - r1

  # check for identical df setting
  if (m == 0L) {
    return(list(
      t_delta = (t0 - t1), scaling.factor = as.numeric(NA),
      df_delta = m, a = as.numeric(NA), b = as.numeric(NA)
    ))
  }

  # check for (near) identical test statistics (despite m > 0)
  if (abs(t1 - t0) < sqrt(.Machine$double.eps)) {
    lav_msg_warn(gettext("the test statistic of the restricted model is (nearly)
                         identical to the test statistic of the full model;
                         check your models."))
  }


  # bail out here, if m == 0 (but we should catch this earlier)
  # if(m < 1L) {
  #    txt <- paste("Can not compute (scaled) difference test when ",
  #                 "the degrees of freedom (df) are the same for both ",
  #                 "models:\n",
  #                 "Df model 1 = ", r1, ", and Df model 2 = ", r0, "\n",
  #                 sep = "")
  #            stop(lav_txt2message(txt, header = "lavaan ERROR:"))
  # }

  m_gamma <- lavTech(m1, "Gamma") # the same for m1 and m0
  # check for NULL
  if (is.null(m_gamma)) {
    lav_msg_stop(gettext(
      "can not compute Gamma matrix; perhaps missing = \"ml\"?"))
  }

  if (h1) {
    wls_v <- lavTech(m1, "WLS.V")
    m_pi <- lav_model_delta(m1@Model)
    p <- lavTech(m1, "information")
    # needed? (yes, if h1 already has eq constraints)
    p_inv <- lav_model_info_augment_invert(m1@Model,
      information = p,
      inverted = TRUE
    )
    # compute 'A' matrix
    # NOTE: order of parameters may change between h1 and H0, so be
    # careful!
    if (is.null(m_a)) {
      m_a <- lav_test_diff_a(m1, m0, method = a_method, reference = "H1")
      # take into account equality constraints m1
      if (a_method == "delta") {
        eq_basis_m1 <- lav_con_eq_basis(m1@Model)
        if (!is.null(eq_basis_m1)) {
          m_a <- m_a %*% t(eq_basis_m1)
        }
      }
      if (lav_debug()) print(m_a)
    }
  } else {
    lav_msg_stop(gettext("not ready yet"))

    wls_v <- lavTech(m0, "WLS.V")
    m_pi <- lav_model_delta(m0@Model)
    p <- lavTech(m0, "information")
    # needed?
    p_inv <- lav_model_info_augment_invert(m0@Model,
      information = p,
      inverted = TRUE
    )

    # compute 'A' matrix
    # NOTE: order of parameters may change between H1 and H0, so be
    # careful!
    if (is.null(m_a)) {
      # m1, m0 OR m0, m1 (works for delta, but not for exact)
      m_a <- lav_test_diff_a(m1, m0, method = a_method, reference = "H0")
      # take into account equality constraints m1
      eq_basis_m0 <- lav_con_eq_basis(m0@Model)
      if (!is.null(eq_basis_m0)) {
        m_a <- m_a %*% t(eq_basis_m0)
      }
      if (lav_debug()) print(m_a)
    }
  }

  # compute tr UG per group
  ngroups <- m1@SampleStats@ngroups

  # safety check: m_a %*% p_inv %*% t(m_a) should NOT contain all-zero
  # rows/columns
  # FIXME: is this really needed? As we use ginv later on
  apa <- m_a %*% p_inv %*% t(m_a)
  c_sums <- colSums(apa)
  r_sums <- rowSums(apa)
  empty_idx <- which(abs(c_sums) < .Machine$double.eps^0.5 &
    abs(r_sums) < .Machine$double.eps^0.5)
  if (length(empty_idx) > 0) {
    m_a <- m_a[-empty_idx, , drop = FALSE]
  }
  if (nrow(m_a) == 0L) {
    # oops... abort!
    return(list(
      t_delta = (t0 - t1), scaling.factor = as.numeric(NA),
      df_delta = m, a = as.numeric(NA), b = as.numeric(NA)
    ))
  }

  # paapaap
  paapaap <- p_inv %*% t(m_a) %*% MASS::ginv(m_a %*%
                               p_inv %*% t(m_a)) %*% m_a %*% p_inv

  # compute scaling factor
  fg <- unlist(m1@SampleStats@nobs) / m1@SampleStats@ntotal

  # both traces live in the npar x npar space: with M = paapaap
  # (symmetric) and K_g = Pi_g' V_g Gamma_g V_g Pi_g, we have
  # U = V Pi M Pi' V, so that (Satorra 2000, eq. 23)
  #   tr(U Gamma)     = sum_g fg_g tr(M K_g)
  #   tr((U Gamma)^2) = tr((M K)^2), K = sum_g fg_g K_g
  # this replaces the stacked pstar x pstar computation; the fg * Gamma_g
  # with UNweighted V blocks convention is trace-equivalent to the
  # Gamma_g / fg + fg-weighted-V convention used elsewhere (see the
  # SCALING CONVENTIONS note in lav_samplestats_gamma.R)
  k_group <- vector("list", length = ngroups)
  trace_ugamma_group <- numeric(ngroups)
  for (g in 1:ngroups) {
    vp <- wls_v[[g]] %*% m_pi[[g]] # pstar x npar
    k_group[[g]] <- crossprod(vp, m_gamma[[g]] %*% vp)
    trace_ugamma_group[g] <- sum(paapaap * k_group[[g]])
  }
  trace_ugamma <- sum(fg * trace_ugamma_group)

  trace_ugamma2 <- as.numeric(NA)
  if (satterthwaite) {
    if (old_approach) {
      # this is what we did <0.6-13: also the second trace per group
      trace_ugamma2_group <- numeric(ngroups)
      for (g in 1:ngroups) {
        mk <- paapaap %*% k_group[[g]]
        trace_ugamma2_group[g] <- sum(mk * t(mk))
      }
      trace_ugamma2 <- sum(fg * trace_ugamma2_group)
    } else {
      # global approach (not group-specific)
      k_all <- fg[1] * k_group[[1]]
      for (g in seq_len(ngroups - 1L) + 1L) {
        k_all <- k_all + fg[g] * k_group[[g]]
      }
      mk <- paapaap %*% k_all
      trace_ugamma2 <- sum(mk * t(mk))
    }
  }

  if (satterthwaite && !scaled_shifted) {
    cd <- trace_ugamma2 / trace_ugamma
    df_delta <- trace_ugamma^2 / trace_ugamma2
    t_delta <- (t0 - t1) / cd
    a <- as.numeric(NA)
    b <- as.numeric(NA)
  } else if (satterthwaite && scaled_shifted) {
    a <- sqrt(m / trace_ugamma2)
    # b <- m - sqrt(m * trace_ugamma^2 / trace_ugamma2)
    b <- m - a * trace_ugamma
    df_delta <- m
    t_delta <- (t0 - t1) * a + b
    cd <- as.numeric(NA)
  } else {
    cd <- 1 / m * trace_ugamma
    df_delta <- m
    t_delta <- (t0 - t1) / cd
    a <- as.numeric(NA)
    b <- as.numeric(NA)
  }

  list(
    t_delta = t_delta, scaling.factor = cd, df_delta = df_delta,
    trace.ugamma = trace_ugamma, trace.ugamma2 = trace_ugamma2,
    a = a, b = b
  )
}

lav_test_diff_sb2001 <- function(m1, m0, test = 2) {
  # extract information from m1 and m2
  t1 <- m1@test[[1]]$stat
  r1 <- m1@test[[1]]$df
  c1 <- m1@test[[test]]$scaling.factor

  ## check for situations when scaling.factor would be NA
  if (r1 == 0) {
    ## saturated model
    c1 <- 1 # canceled out by 0 when calculating "cd"

  } else if (r1 > 0 && isTRUE(all.equal(t1, 0))) {
    ## perfect fit
    c1 <- 0 # cancels out r1 when calculating "cd"
  }

  t0 <- m0@test[[1]]$stat
  r0 <- m0@test[[1]]$df
  c0 <- m0@test[[test]]$scaling.factor

  # m = difference between the df's
  m <- r0 - r1

  # check for identical df setting
  if (m == 0L) {
    return(list(
      T.delta = (t0 - t1), scaling.factor = as.numeric(NA),
      df.delta = m
    ))
  }

  # compute c_d
  cd <- (r0 * c0 - r1 * c1) / m

  # warn if cd is negative
  if (cd < 0) {
    lav_msg_warn(gettext("scaling factor is negative"))
    cd <- as.numeric(NA)
  }

  # compute scaled difference test
  t_delta <- (t0 - t1) / cd

  list(T.delta = t_delta, scaling.factor = cd, df.delta = m)
}

lav_test_diff_sb2010 <- function(m1, m0, test = 2,
                                             h1 = FALSE) {
  ### FIXME: check if models are nested at the parameter level!!!

  # extract information from m1 and m2
  t1 <- m1@test[[1]]$stat
  r1 <- m1@test[[1]]$df
  # c1 <- m1@test[[test]]$scaling.factor
  # if (r1 == 0) { # saturated model
  #   c1 <- 1
  # }

  t0 <- m0@test[[1]]$stat
  r0 <- m0@test[[1]]$df
  c0 <- m0@test[[test]]$scaling.factor
  if (r0 == 0) { # should never happen
    c0 <- 1
  }

  # m = difference between the df's
  m <- r0 - r1

  # check for identical df setting
  if (m == 0L) {
    return(list(
      T.delta = (t0 - t1), scaling.factor = as.numeric(NA),
      df.delta = m
    ))
  }

  # generate `M10' model
  if (h1) {
    # M0 with M1 parameters
    m01 <- lav_test_diff_m10(m0, m1, test = TRUE)
    c01 <- m01@test[[test]]$scaling.factor

    # check if vcov is positive definite (new in 0.6)
    # if not, we may get negative values
    eigvals <- eigen(lavTech(m01, "information"),
      symmetric = TRUE, only.values = TRUE
    )$values
    if (any(eigvals < -1 * .Machine$double.eps^(3 / 4))) {
      lav_msg_warn(gettext(
        "information matrix of the M01 model is not positive definite."
      ))
      # "                  As a result, the scale-factor can not be computed.")
      # cd <- as.numeric(NA)
    } # else {
    # compute c_d
    # cd.01 <- (r0 * c01 - r1 * c0) / m ???
    cd <- (r0 * c0 - r1 * c01) / m
    # }
  } else {
    # M1 with M0 parameters (as in Satorra & Bentler 2010)
    m10 <- lav_test_diff_m10(m1, m0, test = TRUE)
    c10 <- m10@test[[test]]$scaling.factor

    # check if vcov is positive definite (new in 0.6)
    # if not, we may get negative values
    eigvals <- eigen(lavTech(m10, "information"),
      symmetric = TRUE, only.values = TRUE
    )$values
    if (any(eigvals < -1 * .Machine$double.eps^(3 / 4))) {
      lav_msg_warn(gettext(
        "information matrix of the M10 model is not positive definite."
      ))
      # "                  As a result, the scale-factor can not be computed.")
      # cd <- as.numeric(NA)
    } # else {
    # compute c_d
    cd <- (r0 * c0 - r1 * c10) / m
    # }
  }

  # compute scaled difference test
  t_delta <- (t0 - t1) / cd

  list(
    T.delta = t_delta, scaling.factor = cd, df.delta = m,
    T.delta.unscaled = (t0 - t1)
  )
}

# create a new model 'm10', where we use model 'm1', but we
# inject it with the values of 'm0'
lav_test_diff_m10 <- function(m1, m0, test = FALSE) {
  # switch of verbose/se/test
  options_1 <- m1@Options
  # switch of optim.gradient check
  options_1$check.gradient <- FALSE

  # should we compute se/test statistics?
  if (!test) {
    options_1$se <- "none"
    options_1$test <- "none"
  }

  pt_m0 <- lav_pt_set_cache(m0@ParTable, m0@pta)
  pt_m1 <- lav_pt_set_cache(m1@ParTable, m1@pta)

  # `extend' PT.M1 partable to include all `fixed-to-zero parameters'
  pt_m1_full <- lav_pt_full(
    partable = pt_m1,
    free = TRUE, start = TRUE
  )
  pt_m1_extended <- lav_pt_merge(pt_m1, pt_m1_full,
    remove_duplicated = TRUE, warn = FALSE
  )

  # remove most columns
  pt_m1_extended$start <- NULL # new in 0.6-4! (otherwise, they are used)
  pt_m1_extended$est <- NULL
  pt_m1_extended$se <- NULL

  # in addition, use 'NA' for free parameters in ustart column
  free_par_idx <- which(pt_m1_extended$free > 0L)
  pt_m1_extended$ustart[free_par_idx] <- as.numeric(NA)

  # `extend' PT.M0 partable to include all `fixed-to-zero parameters'
  pt_m0_full <- lav_pt_full(
    partable = pt_m0,
    free = TRUE, start = TRUE
  )
  pt_m0_extended <- lav_pt_merge(pt_m0, pt_m0_full,
    remove_duplicated = TRUE, warn = FALSE
  )
  # remove most columns, but not 'est'
  pt_m0_extended$ustart <- NULL
  pt_m0_extended$start <- NULL
  pt_m0_extended$se <- NULL


  # FIXME:
  # - check if H0 does not contain additional parameters...

  options_1$optim.method <- "none"
  options_1$optim.force.converged <- TRUE
  options_1$baseline <- FALSE
  options_1$fit.by.level <- FALSE
  options_1$h1 <- TRUE # needed after all (yuan.benter.mplus)
  options_1$start <- pt_m0_extended # new in 0.6!
  m10 <- lavaan(
    model = pt_m1_extended,
    slot_options = options_1,
    slot_sample_stats = m1@SampleStats,
    slot_data = m1@Data,
    slot_cache = m1@Cache,
    verbose = FALSE
  )

  m10
}

# compute the `A' matrix: the jacobian of the constraint function a(\delta)
# (see Satorra 2000)
#
#
#
lav_test_diff_a <- function(m1, m0, method = "delta", reference = "H1") {
  # FIXME!!!!

  if (method == "exact") {
    if (reference == "H1") {
      af <- lav_test_diff_af_h1(m1 = m1, m0 = m0)
      xx <- m1@optim$x
    } else { # evaluate under H0
      lav_msg_stop(gettext("not ready yet"))
      # af <- .test_compute_partable_A_diff_h0(m1 = m1, m0 = m0)
      xx <- m0@optim$x
    }
    m_a <- try(lav_func_jacobian_complex(func = af, x = xx), silent = TRUE)
    if (inherits(m_a, "try-error")) {
      m_a <- lav_func_jacobian_simple(func = af, x = xx)
    }
  } else if (method == "delta") {
    # use a numeric approximation of `A'
    delta1_list <- lav_model_delta(m1@Model)
    delta0_list <- lav_model_delta(m0@Model)
    delta1 <- do.call(rbind, delta1_list)
    delta0 <- do.call(rbind, delta0_list)

    # take into account equality constraints m0
    # note: delta has one column per non-collapsed free parameter; the basis
    # K maps it to the constrained space, so we post-multiply by K (NOT t(K))
    eq_basis_m0 <- lav_con_eq_basis(m0@Model)
    if (!is.null(eq_basis_m0)) {
      delta0 <- delta0 %*% eq_basis_m0
    }

    # take into account equality constraints m1
    eq_basis_m1 <- lav_con_eq_basis(m1@Model)
    if (!is.null(eq_basis_m1)) {
      delta1 <- delta1 %*% eq_basis_m1
    }

    # H <- solve(t(Delta1) %*% Delta1) %*% t(Delta1) %*% Delta0
    h <- MASS::ginv(delta1) %*% delta0
    m_a <- t(lav_mat_ortho_complement(h))
  }

  m_a
}


# for each parameter in H1 (m1), see if we have somehow constrained
# this parameter under H0 (m0)
#
# since we work 'under H0', we need to use the labels/constraints/def
# as they appear in H0. Unfortunately, the order of the parameters, and
# even the (p)labels may be different in the two models...
#
# Therefore, we will attempt to:
#   - change the 'order' of the 'free' column in m0, so that they map to
#     to the 'x' that we will provide from H1
#   - the plabels used in "==" constraints must be renamed, if necessary
#
lav_test_diff_af_h1 <- function(m1, m0) {
  pt_m0 <- lav_pt_set_cache(parTable(m0), m0@pta)
  pt_m1 <- lav_pt_set_cache(parTable(m1), m1@pta)

  # select .p*. parameters only
  m0_p_idx <- grep(".p", pt_m0$plabel, fixed = TRUE)
  np0 <- length(m0_p_idx)
  m1_p_idx <- grep(".p", pt_m1$plabel, fixed = TRUE)
  np1 <- length(m1_p_idx)

  # check if parameter space is the same
  if (np0 != np1) {
    lav_msg_stop(gettext(
      "unconstrained parameter set is not the same in m0 and m1"))
  }

  # split partable in 'parameter' and 'constraints' section
  pt_m0_part1 <- pt_m0[m0_p_idx, ]
  pt_m0_part2 <- pt_m0[-m0_p_idx, ]

  pt_m1_part1 <- pt_m1[m1_p_idx, ]
  pt_m1_part2 <- pt_m1[-m1_p_idx, ]

  # figure out relationship between m0 and m1
  p1_id <- lav_pt_map_id_p1_in_p2(pt_m0_part1, pt_m1_part1)
  p0_free_idx <- which(pt_m0_part1$free > 0)

  # change 'free' order in m0
  # NOTE: this only works all the free parameters in h0 are also free
  # in h1 (and if not, they will become fixed in h0)
  pt_m0_part1$free[p0_free_idx] <-
    pt_m1_part1$free[pt_m0_part1$id[p1_id][p0_free_idx]]

  # paste back
  pt_m0 <- rbind(pt_m0_part1, pt_m0_part2)
  pt_m1 <- rbind(pt_m1_part1, pt_m1_part2)

  # `extend' PT.M1 partable to include all `fixed-to-zero parameters'
  pt_m1_full <- lav_pt_full(
    partable = pt_m1,
    free = TRUE, start = TRUE
  )
  pt_m1_extended <- lav_pt_merge(pt_m1, pt_m1_full,
    remove_duplicated = TRUE, warn = FALSE
  )

  # `extend' PT.M0 partable to include all `fixed-to-zero parameters'
  pt_m0_full <- lav_pt_full(
    partable = pt_m0,
    free = TRUE, start = TRUE
  )
  pt_m0_extended <- lav_pt_merge(pt_m0, pt_m0_full,
    remove_duplicated = TRUE, warn = FALSE
  )

  p1 <- pt_m1_extended
  np1 <- length(p1$lhs)
  p0 <- pt_m0_extended
  np0 <- length(p0$lhs)

  con_function <- function() NULL
  formals(con_function) <- alist(.x. = , ... = )
  body_txt <- paste("{\nout <- numeric(0L)\n", sep = "")


  # first handle def + == constraints
  # but FIRST, remove == constraints that also appear in H1!!!

  # remove equivalent eq constraints from p0
  p0_1 <- p0

  p0_eq_idx <- which(p0$op == "==")
  p1_eq_idx <- which(p1$op == "==")
  p0_remove_idx <- integer(0L)
  if (length(p0_eq_idx) > 0L) {
    for (i in seq_along(p0_eq_idx)) {
      # e0 in p0
      e0 <- p0_eq_idx[i]
      lhs <- p0$lhs[e0]
      rhs <- p0$rhs[e0]

      # do we have an equivalent constraint in H1?
      # NOTE!! the (p)labels may differ

      # SO, we will use an 'empirical' approach: if we fill in (random)
      # values, and work out the constraint, do we get identical values?
      # if yes, constraint is equivalent, and we should NOT add it here

      if (length(p1_eq_idx) > 0) {
        # generate random parameter values
        xx1 <- rnorm(length(m1_p_idx))
        xx0 <- xx1[p1_id]

        con_h0_value <- m0@Model@ceq.function(xx0)[i]
        con_h1_values <- m1@Model@ceq.function(xx1)

        if (con_h0_value %in% con_h1_values) {
          p0_remove_idx <- c(p0_remove_idx, e0)
        }
      }
    }
  }
  if (length(p0_remove_idx) > 0L) {
    p0_1 <- p0_1[-p0_remove_idx, ]
  }

  # only for the UNIQUE equality constraints in H0, generate syntax
  defcon_txt <- lav_pt_con_ceq(p0_1, txt_only = TRUE)
  body_txt <- paste(body_txt, defcon_txt, "\n", sep = "")


  # for each parameter in p1, we 'check' is it is fixed to a constant in p0
  ncon <- length(which(p0_1$op == "=="))
  for (i in seq_len(np1)) {
    # p in p1
    lhs <- p1$lhs[i]
    op <- p1$op[i]
    rhs <- p1$rhs[i]
    group <- p1$group[i]

    # ignore '==', '<', '>' and ':=' for now
    if (op == "==" || op == ">" || op == "<" || op == ":=") next

    # search for corresponding parameter in p0
    p0_idx <- which(p0$lhs == lhs & p0$op == op & p0$rhs == rhs &
      p0$group == group)
    if (length(p0_idx) == 0L) {
      lav_msg_stop(
        gettextf("parameter in H1 not found in H0: %s",
        paste(lhs, op, rhs, "(group = ", group, ")", sep = " ")
      ))
    }

    # 4 possibilities: p is free/fixed in p1, p is free/fixed in p0
    if (p1$free[i] == 0L) {
      if (p0$free[p0_idx] == 0L) {
        # match, nothing to do
      } else {
        lav_msg_warn(
          gettextf("fixed parameter in H1 is free in H0: %s",
          paste("\"", lhs, " ", op, " ", rhs,
            "\" (group = ", group, ")",
            sep = ""
          )
        ))
      }
    } else {
      if (p0$free[p0_idx] == 0L) {
        # match, this is a constrained parameter in H0
        ncon <- ncon + 1L
        body_txt <- paste(body_txt,
          "out[", ncon, "] = .x.[", p1$free[i], "] - ",
          p0$ustart[p0_idx], "\n",
          sep = ""
        )
        next
      } else {
        # match, nothing to do
      }
    }
  }


  # wrap function
  body_txt <- paste(body_txt, "return(out)\n}\n", sep = "")
  body(con_function) <- parse(file = "", text = body_txt)

  con_function
}

lav_step03_data <- function(slot_data = NULL,
                                   lavoptions = NULL,
                                   ov_names = NULL,
                                   ov_names_y = NULL,
                                   group = NULL,
                                   data = NULL,
                                   cluster = NULL,
                                   ov_names_x = NULL,
                                   ov_names_l = NULL,
                                   ov_names_aux = NULL,
                                   ordered = NULL,
                                   sampling_weights = NULL,
                                   sample_cov = NULL,
                                   sample_mean = NULL,
                                   sample_th = NULL,
                                   sample_nobs = NULL,
                                   slot_par_table = NULL,
                                   ngroups = NULL,
                                   dotdotdot = NULL,
                                   flat_model = NULL,
                                   model = NULL,
                                   nacov = NULL,
                                   wls_v = NULL) {
  # # # # # # # # # # #
  # #  3. lavdata  # #
  # # # # # # # # # # #

  # if slot_data not null
  #   copy slot_data to lavdata
  # else
  #   create lavdata via function lav_lavdata, setting ov.names to ov.names.y
  #     if lavoptions$conditional.x
  # if lavdata$data.type is "none"
  #   set lavoptions$do.fit to FALSE
  #   if flat_model$est not null set lavoptions$start to "est", else set
  #     it to "simple"
  #   set lavoptions$se and lavoptions$test to "none"
  # else
  #   if lavdata$data.type is "moment"
  #     if estimator one of MLM, MLMV, MLR, ULSM, ULSMV, ULSMVS and NACOV
  #       is NULL: *** error ***
  #     if estimator one of WLS, WLSM, WLSMV, WLSMVS, DWLS and wls_v is
  #       NULL: *** error ***
  #     if lavoptions$se = bootstrap: *** error ***
  # if slotPartable not NULL and model is lavaan-object, check equality
  #   ngroups and lavdata$ngroups
  #                                       --> *** error *** if not

  if (!is.null(slot_data)) {
    lavdata <- slot_data
  } else {
    if (lav_verbose()) {
      cat("lavdata            ...")
    }
    # FIXME: ov.names should always contain both y and x!
    tmp_ov_names <- if (lavoptions$conditional.x) {
      ov_names_y
    } else {
      ov_names
    }
    lavdata <- lav_lavdata(
      data = data,
      group = group,
      cluster = cluster,
      ov_names = tmp_ov_names,
      ov_names_x = ov_names_x,
      ov_names_l = ov_names_l,
      ov_names_aux = ov_names_aux,
      ordered = ordered,
      sampling_weights = sampling_weights,
      sample_cov = sample_cov,
      sample_mean = sample_mean,
      sample_th = sample_th,
      sample_nobs = sample_nobs,
      lavoptions = lavoptions
    )

    if (lav_verbose()) {
      cat(" done.\n")
    }
  }
  # what have we learned from the data?
  if (lavdata@data.type == "none") {
    lavoptions$do.fit <- FALSE
    # check if 'model' was a fitted parameter table
    lavoptions$start <- ifelse(is.null(flat_model$est), "simple", "est")
    lavoptions$se <- "none"
    lavoptions$test <- "none"
  } else if (lavdata@data.type == "moment") {
    # check user-specified options first
    if (!is.null(dotdotdot$estimator)) {
      if (is.list(dotdotdot$estimator)) {
        estimator <- toupper(dotdotdot$estimator[[1]])
      } else {
        estimator <- toupper(dotdotdot$estimator)
      }
      if (any(estimator == c(
        "MLM", "MLMV", "MLR", "MLR",
        "ULSM", "ULSMV", "ULSMVS"
      )) &&
        is.null(nacov)) {
        lav_msg_stop(gettextf(
          "estimator %s requires full data or user-provided NACOV", estimator))
      } else if (any(estimator == c(
        "WLS", "WLSM", "WLSMV",
        "WLSMVS", "DWLS"
      )) &&
        is.null(wls_v)) {
        lav_msg_stop(gettextf(
          "estimator %s requires full data or user-provided wls_v and NACOV",
          estimator))
      }
    }
    # catch here some options that will not work with moments
    if (lavoptions$se == "bootstrap") {
      lav_msg_stop(gettext("bootstrapping requires full data"))
    }
    # robust (sandwich) standard errors need the raw-data Gamma/NACOV;
    # robust.sem/robust.sem.nt can be rescued by a user-provided NACOV=,
    # but robust.huber.white/first.order need the casewise scores (raw data)
    if (any(lavoptions$se == c("robust.sem", "robust.sem.nt")) &&
      is.null(nacov)) {
      lav_msg_stop(gettextf(
        "se = %s requires full data or a user-provided NACOV= argument.",
        dQuote(lavoptions$se)))
    } else if (any(lavoptions$se == c("robust.huber.white", "first.order"))) {
      lav_msg_stop(gettextf(
        "se = %s requires full data.", dQuote(lavoptions$se)))
    }
    # more needed?
  }
  # sanity check
  if (!is.null(slot_par_table) || inherits(model, "lavaan")) {
    if (ngroups != lavdata@ngroups) {
      lav_msg_stop(gettext(
        "mismatch between number of groups in data
        and number of groups in model."))
    }
  }
  if (lav_verbose()) {
    print(lavdata)
  }
  if (lav_debug()) {
    print(str(lavdata))
  }
  # if lavdata@nlevels > 1L, adapt start option (for now)
  # until we figure out how to handle groups+blocks
  # if(lavdata@nlevels > 1L) {
  #   lavoptions$start <- "simple"
  # }

  list(
    lavdata    = lavdata,
    lavoptions = lavoptions
  )
}

# the bivariate two-level ordinal probit model, univariate parameters fixed
#
#     y*_a,ij = w_a,ij + b_a,j     y*_c,ij = w_c,ij + b_c,j
#
#     (w_a, w_c) ~ N(0, | 1     rho_w |)    (b_a, b_c) ~ N(0, | vb_a  cb   |)
#                       | rho_w 1     |                       | cb    vb_c |
#
# YR 2026 (two-level WLS, phase 3)
#
# This is stage 2 of the stage-wise unrestricted two-level model for a
# pair of ordinal variables: the univariate parameters (thresholds tau_a,
# tau_c and between variances vb_a, vb_c) are held fixed at their stage-1
# values (lav_uvord_2l_fit), and theta = c(rho_w, cb) -- the WITHIN
# polychoric correlation and the BETWEEN covariance -- is estimated by ML
# with two-dimensional Gauss-Hermite quadrature over the cluster effects:
#
#     b_a = s_a u1,   b_c = (cb/s_a) u1 + w u2,  w = sqrt(vb_c - cb^2/vb_a)
#
# (the Cholesky of Sigma_b; u1, u2 iid standard normal nodes). The
# per-node conditional kernel is the shifted-threshold bivariate probit
# rectangle probability (pbinorm), exactly as in the exo branch of the
# single-level lav_bvord.R.
#
# Computations are organized per response PATTERN (category pair): the
# conditional kernel takes only ncat_a * ncat_c distinct values per node,
# and the cluster log-likelihoods follow from the (nclusters x npatterns)
# count matrix -- the cost is (nearly) independent of the number of
# observations.
#
# lav_bvord_2l_sc() returns cluster-wise scores with respect to ALL
# parameters (tau_a, tau_c, vb_a, vb_c, rho_w, cb): the fixed-parameter
# columns are the cross-derivatives needed for the stage-wise sandwich.

lav_bvord_2l_cor_twostep_fit <- function(fit_y1 = NULL,
                                         fit_y2 = NULL,
                                         ngh = 13L,
                                         optim_method = "nlminb1",
                                         control = list(),
                                         y1_name = NULL, y2_name = NULL,
                                         output = "theta") {
  # create cache environment
  cache <- lav_bvord_2l_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, ngh = ngh
  )

  # bounds
  cb_bound <- 0.999 * sqrt(cache$vb1 * cache$vb2)
  cb_fixed <- (cb_bound < 1e-08)
  if (cb_fixed) {
    cache$theta[2] <- 0
    lower <- c(-0.999, 0)
    upper <- c(+0.999, 0)
  } else {
    lower <- c(-0.999, -cb_bound)
    upper <- c(+0.999, +cb_bound)
  }

  # optim
  minfns <- lav_uvbv_optim_fns(
    optim_method, lav_bvord_2l_min_fns(covariates = cache$nexo_any)
  )

  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  # try 1
  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb, lower = lower, upper = upper,
    cache = cache
  )

  # try 2: start from independence
  if (optim$convergence != 0L) {
    optim <- nlminb(
      start = c(0, 0), objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control_nlminb, lower = lower, upper = upper,
      cache = cache
    )
  }

  if (optim$convergence != 0L) {
    if (!is.null(y1_name) && !is.null(y2_name)) {
      lav_msg_warn(gettextf(
        "estimation of the two-level polychoric parameters did not converge
        for variables %1$s and %2$s", y1_name, y2_name))
    } else {
      lav_msg_warn(gettext(
        "estimation of the two-level polychoric parameters did not always
        converge"))
    }
  }

  if (output == "cache") {
    cache$theta <- optim$par
    return(cache)
  }

  # theta = c(rho_w, cb)
  optim$par
}

# prepare cache environment
# fit_y1/fit_y2: lav_uvord_2l_fit() objects of the two variables
lav_bvord_2l_init_cache <- function(fit_y1 = NULL,
                                    fit_y2 = NULL,
                                    ngh = 13L,
                                    parent = parent.frame()) {
  stopifnot(length(fit_y1$y) == length(fit_y2$y))
  y1 <- fit_y1$y
  y2 <- fit_y2$y
  cluster_idx <- fit_y1$cluster_idx
  nclusters <- length(unique(cluster_idx))
  nobs <- length(y1)

  # fixed univariate parameters
  th1 <- fit_y1$theta[fit_y1$th_idx]
  vb1 <- fit_y1$theta[fit_y1$bvar_idx]
  th2 <- fit_y2$theta[fit_y2$th_idx]
  vb2 <- fit_y2$theta[fit_y2$bvar_idx]

  nth1 <- length(th1)
  nth2 <- length(th2)
  ncat1 <- nth1 + 1L
  ncat2 <- nth2 + 1L
  npat <- ncat1 * ncat2

  # covariates? (per-observation kernels instead of per-pattern ones)
  nexo_a <- fit_y1$nexo
  nexo_c <- fit_y2$nexo
  nexo_any <- (nexo_a > 0L || nexo_c > 0L)

  cl_sorted <- sort(unique(cluster_idx))
  cl_pos <- match(cluster_idx, cl_sorted)

  cmat <- NULL
  yp1 <- yp2 <- NULL
  y1_1 <- y1_2 <- y2_1 <- y2_2 <- NULL
  z_a <- z_c <- NULL
  idx1a <- idx2a <- idx1c <- idx2c <- NULL

  if (!nexo_any) {
    # response patterns: pattern p = (k, l) at index k + (l-1)*ncat1
    yp1 <- rep(seq_len(ncat1), times = ncat2)
    yp2 <- rep(seq_len(ncat2), each = ncat1)

    # cluster x pattern count matrix
    pid <- y1 + (y2 - 1L) * ncat1
    cmat <- unclass(table(
      factor(cl_pos, levels = seq_len(nclusters)),
      factor(pid, levels = seq_len(npat))
    ))
    dimnames(cmat) <- NULL
    storage.mode(cmat) <- "double"

    # pattern-level unshifted rectangle bounds
    o12_1 <- lav_uvord_o12(y = yp1, nth = nth1)
    o12_2 <- lav_uvord_o12(y = yp2, nth = nth2)
    z12_1 <- lav_uvord_z12(y = yp1, th = th1, o1 = o12_1$o1,
                           o2 = o12_1$o2)
    z12_2 <- lav_uvord_z12(y = yp2, th = th2, o1 = o12_2$o1,
                           o2 = o12_2$o2)

    # pattern-level threshold indicator matrices
    y1_1 <- matrix(seq_len(nth1), npat, nth1, byrow = TRUE) == yp1
    y1_2 <- matrix(seq_len(nth1), npat, nth1, byrow = TRUE) == (yp1 - 1L)
    y2_1 <- matrix(seq_len(nth2), npat, nth2, byrow = TRUE) == yp2
    y2_2 <- matrix(seq_len(nth2), npat, nth2, byrow = TRUE) == (yp2 - 1L)
  } else {
    # observation-level unshifted rectangle bounds, with the fixed linear
    # predictors eta = z beta absorbed (z1 = tau - eta + o1, ...)
    eta_a <- 0
    if (nexo_a > 0L) {
      beta_a <- fit_y1$theta[fit_y1$nth + 1L + seq_len(nexo_a)]
      z_a <- fit_y1$z
      eta_a <- drop(z_a %*% beta_a)
    }
    eta_c <- 0
    if (nexo_c > 0L) {
      beta_c <- fit_y2$theta[fit_y2$nth + 1L + seq_len(nexo_c)]
      z_c <- fit_y2$z
      eta_c <- drop(z_c %*% beta_c)
    }
    o12_1 <- lav_uvord_o12(y = y1, nth = nth1)
    o12_2 <- lav_uvord_o12(y = y2, nth = nth2)
    z12_1 <- lav_uvord_z12(y = y1, th = th1, o1 = o12_1$o1,
                           o2 = o12_1$o2, offset = eta_a)
    z12_2 <- lav_uvord_z12(y = y2, th = th2, o1 = o12_2$o1,
                           o2 = o12_2$o2, offset = eta_c)

    # threshold-score index sets: z1 involves tau_y (y <= nth), z2
    # involves tau_{y-1} (y >= 2)
    idx1a <- which(y1 <= nth1)
    idx2a <- which(y1 >= 2L)
    idx1c <- which(y2 <= nth2)
    idx2c <- which(y2 >= 2L)
  }

  # 2-D tensor grid of standardized GH nodes
  gh <- lav_2l_gh_xw(ngh = ngh)
  u1 <- rep(gh$x, times = ngh)
  u2 <- rep(gh$x, each = ngh)
  logw <- rep(gh$logw, times = ngh) + rep(gh$logw, each = ngh)

  # starting values: crude total correlation, split within/between
  r_tot <- suppressWarnings(cor(y1, y2))
  if (!is.finite(r_tot)) {
    r_tot <- 0
  }
  rho0 <- max(min(r_tot, 0.80), -0.80)
  cb0 <- r_tot * sqrt(vb1 * vb2)
  cb0 <- max(min(cb0, 0.95 * sqrt(vb1 * vb2)), -0.95 * sqrt(vb1 * vb2))
  theta <- c(rho0, cb0)

  list2env(
    list(
      cluster_idx = cluster_idx,
      nobs = nobs, n = nobs, nclusters = nclusters, npat = npat,
      th1 = th1, th2 = th2, nth1 = nth1, nth2 = nth2,
      vb1 = vb1, vb2 = vb2, s1 = sqrt(vb1), s2 = sqrt(vb2),
      cmat = cmat,
      z1a0 = z12_1$z1, z2a0 = z12_1$z2,
      z1c0 = z12_2$z1, z2c0 = z12_2$z2,
      y1_1 = y1_1, y1_2 = y1_2, y2_1 = y2_1, y2_2 = y2_2,
      nexo_a = nexo_a, nexo_c = nexo_c, nexo_any = nexo_any,
      nexo_wa = fit_y1$nexo_w, nexo_ba = fit_y1$nexo_b,
      nexo_wc = fit_y2$nexo_w, nexo_bc = fit_y2$nexo_b,
      z_a = z_a, z_c = z_c,
      y1v = y1, y2v = y2,
      idx1a = idx1a, idx2a = idx2a, idx1c = idx1c, idx2c = idx2c,
      u1 = u1, u2 = u2, logw = logw, nq = ngh * ngh, ngh = ngh,
      theta = theta
    ),
    parent = parent
  )
}

# full-grid (pattern x node) rectangle bounds and probabilities; all
# quantities are stored in the cache (pattern index moves fastest)
lav_bvord_2l_grid_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    rho_w <- theta[1]
    cb <- theta[2]

    # Cholesky of Sigma_b
    l21 <- cb / s1
    l22 <- sqrt(max(vb2 - cb * cb / vb1, 1e-12))
    b_a <- s1 * u1
    b_c <- l21 * u1 + l22 * u2

    ba_g <- rep(b_a, each = npat)
    bc_g <- rep(b_c, each = npat)
    z1a_g <- rep(z1a0, times = nq) - ba_g
    z2a_g <- rep(z2a0, times = nq) - ba_g
    z1c_g <- rep(z1c0, times = nq) - bc_g
    z2c_g <- rep(z2c0, times = nq) - bc_g

    p_g <- pbinorm(
      upper_x = z1a_g, upper_y = z1c_g,
      lower_x = z2a_g, lower_y = z2c_g, rho = rho_w
    )
    p_g[p_g < .Machine$double.eps] <- .Machine$double.eps
    inv_p_g <- 1 / p_g

    return(invisible(NULL))
  })                     # nolint end
}

lav_bvord_2l_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    tmp <- lav_bvord_2l_grid_cache(cache)

    # cluster x node log-likelihood contributions
    ll <- cmat %*% matrix(log(p_g), npat, nq)

    lwll <- sweep(ll, 2L, logw, "+")
    mmax <- apply(lwll, 1L, max)
    logl_j <- mmax + log(rowSums(exp(lwll - mmax)))
    pw <- exp(lwll - logl_j)

    loglik <- sum(logl_j)
    return(loglik)
  })                     # nolint end
}

lav_bvord_2l_grad_cache <- function(cache = NULL) {
  sc <- lav_bvord_2l_sc_cache(cache = cache, free_only = TRUE)
  colSums(sc)
}

# cluster-wise scores of the loglik; assumes the loglik has been
# evaluated at cache$theta (grid + pw available)
# free_only = FALSE: columns tau_a (nth1), tau_c (nth2), vb_a, vb_c,
#                    rho_w, cb
# free_only = TRUE:  columns rho_w, cb (for the optimizer)
lav_bvord_2l_sc_cache <- function(cache = NULL, free_only = FALSE) {
  cache$free_only <- free_only
  with(cache, {          # nolint start
    rho_w <- theta[1]
    cb <- theta[2]
    r <- sqrt(1 - rho_w * rho_w)

    l21 <- cb / s1
    l22 <- sqrt(max(vb2 - cb * cb / vb1, 1e-12))

    # derivatives of the node positions w.r.t. (vb_a, vb_c, cb)
    # b_a = s1 u1; b_c = (cb/s1) u1 + l22 u2
    dbc_dcb <- u1 / s1 - u2 * cb / (vb1 * l22)

    # rectangle derivatives w.r.t. the c-bounds (needed for cb)
    d_z1c_g <- dnorm(z1c_g) * (pnorm((z1a_g - rho_w * z1c_g) / r) -
                               pnorm((z2a_g - rho_w * z1c_g) / r))
    d_z2c_g <- dnorm(z2c_g) * (pnorm((z1a_g - rho_w * z2c_g) / r) -
                               pnorm((z2a_g - rho_w * z2c_g) / r))
    g_c <- cmat %*% matrix(-(d_z1c_g - d_z2c_g) * inv_p_g, npat, nq)

    # within correlation: four-corner densities
    phi_rho_g <- (lav_dbinorm(z1a_g, z1c_g, rho_w) -
                  lav_dbinorm(z2a_g, z1c_g, rho_w) -
                  lav_dbinorm(z1a_g, z2c_g, rho_w) +
                  lav_dbinorm(z2a_g, z2c_g, rho_w))
    g_rho <- cmat %*% matrix(phi_rho_g * inv_p_g, npat, nq)

    sc_rho <- rowSums(pw * g_rho)
    sc_cb <- rowSums(pw * sweep(g_c, 2L, dbc_dcb, "*"))

    if (free_only) {
      sc <- cbind(sc_rho, sc_cb)
      colnames(sc) <- c("rho_w", "cb")
      return(sc)
    }

    # remaining node-position derivatives
    dba_dvba <- u1 / (2 * s1)
    dbc_dvba <- -u1 * cb / (2 * vb1 * s1) +
      u2 * (cb * cb / (vb1 * vb1)) / (2 * l22)
    dbc_dvbc <- u2 / (2 * l22)

    # rectangle derivatives w.r.t. the a-bounds
    d_z1a_g <- dnorm(z1a_g) * (pnorm((z1c_g - rho_w * z1a_g) / r) -
                               pnorm((z2c_g - rho_w * z1a_g) / r))
    d_z2a_g <- dnorm(z2a_g) * (pnorm((z1c_g - rho_w * z2a_g) / r) -
                               pnorm((z2c_g - rho_w * z2a_g) / r))
    g_a <- cmat %*% matrix(-(d_z1a_g - d_z2a_g) * inv_p_g, npat, nq)

    npar <- nth1 + nth2 + 4L
    sc <- matrix(0, nclusters, npar)

    # thresholds: dP/dtau_ak = y1a_k d_z1a - y2a_k d_z2a
    # (the pattern-level indicator columns recycle down the grid columns)
    m_z1a <- matrix(d_z1a_g * inv_p_g, npat, nq)
    m_z2a <- matrix(d_z2a_g * inv_p_g, npat, nq)
    m_z1c <- matrix(d_z1c_g * inv_p_g, npat, nq)
    m_z2c <- matrix(d_z2c_g * inv_p_g, npat, nq)
    for (k in seq_len(nth1)) {
      g_k <- cmat %*% (y1_1[, k] * m_z1a - y1_2[, k] * m_z2a)
      sc[, k] <- rowSums(pw * g_k)
    }
    for (k in seq_len(nth2)) {
      g_k <- cmat %*% (y2_1[, k] * m_z1c - y2_2[, k] * m_z2c)
      sc[, nth1 + k] <- rowSums(pw * g_k)
    }

    sc[, nth1 + nth2 + 1L] <-
      rowSums(pw * (sweep(g_a, 2L, dba_dvba, "*") +
                    sweep(g_c, 2L, dbc_dvba, "*")))
    sc[, nth1 + nth2 + 2L] <- rowSums(pw * sweep(g_c, 2L, dbc_dvbc, "*"))
    sc[, nth1 + nth2 + 3L] <- sc_rho
    sc[, nth1 + nth2 + 4L] <- sc_cb

    colnames(sc) <- c(
      paste0("tau_a", seq_len(nth1)), paste0("tau_c", seq_len(nth2)),
      "vb_a", "vb_c", "rho_w", "cb"
    )
    return(sc)
  })                     # nolint end
}

# --- with covariates (per-observation kernels, one node at a time) ---

lav_bvord_2l_x_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    rho_w <- theta[1]
    cb <- theta[2]

    # Cholesky of Sigma_b
    l21 <- cb / s1
    l22 <- sqrt(max(vb2 - cb * cb / vb1, 1e-12))
    b_a <- s1 * u1
    b_c <- l21 * u1 + l22 * u2

    ll <- matrix(0, nclusters, nq)
    for (q in seq_len(nq)) {
      p_i <- pbinorm(
        upper_x = z1a0 - b_a[q], upper_y = z1c0 - b_c[q],
        lower_x = z2a0 - b_a[q], lower_y = z2c0 - b_c[q], rho = rho_w
      )
      p_i[p_i < .Machine$double.eps] <- .Machine$double.eps
      ll[, q] <- drop(rowsum.default(log(p_i), group = cluster_idx,
                                     reorder = TRUE))
    }

    lwll <- sweep(ll, 2L, logw, "+")
    mmax <- apply(lwll, 1L, max)
    logl_j <- mmax + log(rowSums(exp(lwll - mmax)))
    pw <- exp(lwll - logl_j)

    loglik <- sum(logl_j)
    return(loglik)
  })                     # nolint end
}

lav_bvord_2l_x_grad_cache <- function(cache = NULL) {
  sc <- lav_bvord_2l_x_sc_cache(cache = cache, free_only = TRUE)
  colSums(sc)
}

# cluster-wise scores, covariate case; assumes the loglik has been
# evaluated at cache$theta (pw available)
# free_only = FALSE: columns tau_a (nth1), tau_c (nth2), vb_a, vb_c,
#                    rho_w, cb, a-slopes, c-slopes
# free_only = TRUE:  columns rho_w, cb (for the optimizer)
lav_bvord_2l_x_sc_cache <- function(cache = NULL, free_only = FALSE) {
  cache$free_only <- free_only
  with(cache, {          # nolint start
    rho_w <- theta[1]
    cb <- theta[2]
    r <- sqrt(1 - rho_w * rho_w)

    l21 <- cb / s1
    l22 <- sqrt(max(vb2 - cb * cb / vb1, 1e-12))
    b_a <- s1 * u1
    b_c <- l21 * u1 + l22 * u2

    # derivatives of the node positions w.r.t. (vb_a, vb_c, cb)
    dbc_dcb <- u1 / s1 - u2 * cb / (vb1 * l22)
    dba_dvba <- u1 / (2 * s1)
    dbc_dvba <- -u1 * cb / (2 * vb1 * s1) +
      u2 * (cb * cb / (vb1 * vb1)) / (2 * l22)
    dbc_dvbc <- u2 / (2 * l22)

    if (free_only) {
      npar <- 2L
      rho_col <- 1L
      cb_col <- 2L
    } else {
      npar <- nth1 + nth2 + 4L + nexo_a + nexo_c
      rho_col <- nth1 + nth2 + 3L
      cb_col <- nth1 + nth2 + 4L
      asl_col <- nth1 + nth2 + 4L + seq_len(nexo_a)
      csl_col <- nth1 + nth2 + 4L + nexo_a + seq_len(nexo_c)
    }
    sc <- matrix(0, nclusters, npar)

    for (q in seq_len(nq)) {
      z1a <- z1a0 - b_a[q]
      z2a <- z2a0 - b_a[q]
      z1c <- z1c0 - b_c[q]
      z2c <- z2c0 - b_c[q]
      p_i <- pbinorm(upper_x = z1a, upper_y = z1c,
                     lower_x = z2a, lower_y = z2c, rho = rho_w)
      p_i[p_i < .Machine$double.eps] <- .Machine$double.eps
      inv_p <- 1 / p_i
      pwq <- pw[, q]

      # rectangle derivatives w.r.t. the c-bounds
      d_z1c <- dnorm(z1c) * (pnorm((z1a - rho_w * z1c) / r) -
                             pnorm((z2a - rho_w * z1c) / r))
      d_z2c <- dnorm(z2c) * (pnorm((z1a - rho_w * z2c) / r) -
                             pnorm((z2a - rho_w * z2c) / r))
      kc <- -(d_z1c - d_z2c) * inv_p # d logp / d(b_c shift)
      g_c <- drop(rowsum.default(kc, group = cluster_idx,
                                 reorder = TRUE))

      # within correlation: four-corner densities
      phi_rho <- (lav_dbinorm(z1a, z1c, rho_w) -
                  lav_dbinorm(z2a, z1c, rho_w) -
                  lav_dbinorm(z1a, z2c, rho_w) +
                  lav_dbinorm(z2a, z2c, rho_w)) * inv_p
      g_rho <- drop(rowsum.default(phi_rho, group = cluster_idx,
                                   reorder = TRUE))

      sc[, rho_col] <- sc[, rho_col] + pwq * g_rho
      sc[, cb_col] <- sc[, cb_col] + pwq * g_c * dbc_dcb[q]

      if (free_only) {
        next
      }

      # rectangle derivatives w.r.t. the a-bounds
      d_z1a <- dnorm(z1a) * (pnorm((z1c - rho_w * z1a) / r) -
                             pnorm((z2c - rho_w * z1a) / r))
      d_z2a <- dnorm(z2a) * (pnorm((z1c - rho_w * z2a) / r) -
                             pnorm((z2c - rho_w * z2a) / r))
      ka <- -(d_z1a - d_z2a) * inv_p
      g_a <- drop(rowsum.default(ka, group = cluster_idx,
                                 reorder = TRUE))

      # thresholds: z1 involves tau_y (y <= nth), z2 involves tau_{y-1}
      kern_a <- matrix(0, nobs, nth1)
      kern_a[cbind(idx1a, y1v[idx1a])] <- (d_z1a * inv_p)[idx1a]
      kern_a[cbind(idx2a, y1v[idx2a] - 1L)] <-
        kern_a[cbind(idx2a, y1v[idx2a] - 1L)] - (d_z2a * inv_p)[idx2a]
      kern_c <- matrix(0, nobs, nth2)
      kern_c[cbind(idx1c, y2v[idx1c])] <- (d_z1c * inv_p)[idx1c]
      kern_c[cbind(idx2c, y2v[idx2c] - 1L)] <-
        kern_c[cbind(idx2c, y2v[idx2c] - 1L)] - (d_z2c * inv_p)[idx2c]
      cl_th <- rowsum.default(cbind(kern_a, kern_c),
                              group = cluster_idx, reorder = TRUE)
      sc[, seq_len(nth1 + nth2)] <- sc[, seq_len(nth1 + nth2)] +
        pwq * cl_th

      # between variances (node-position chain)
      sc[, nth1 + nth2 + 1L] <- sc[, nth1 + nth2 + 1L] +
        pwq * (g_a * dba_dvba[q] + g_c * dbc_dvba[q])
      sc[, nth1 + nth2 + 2L] <- sc[, nth1 + nth2 + 2L] +
        pwq * g_c * dbc_dvbc[q]

      # slopes: dz/dbeta_r = -z[, r] (both bounds) -> same kernels as
      # the node-position shifts
      if (nexo_a > 0L) {
        sc[, asl_col] <- sc[, asl_col] +
          pwq * rowsum.default(z_a * ka, group = cluster_idx,
                               reorder = TRUE)
      }
      if (nexo_c > 0L) {
        sc[, csl_col] <- sc[, csl_col] +
          pwq * rowsum.default(z_c * kc, group = cluster_idx,
                               reorder = TRUE)
      }
    }

    if (free_only) {
      colnames(sc) <- c("rho_w", "cb")
    } else {
      colnames(sc) <- c(
        paste0("tau_a", seq_len(nth1)), paste0("tau_c", seq_len(nth2)),
        "vb_a", "vb_c", "rho_w", "cb",
        if (nexo_a > 0L) paste0("a_", lav_uvreg_2l_slope_names(
          nexo_wa, nexo_ba)) else character(0L),
        if (nexo_c > 0L) paste0("c_", lav_uvreg_2l_slope_names(
          nexo_wc, nexo_bc)) else character(0L)
      )
    }
    return(sc)
  })                     # nolint end
}

# cluster-wise scores at given values - no cache
lav_bvord_2l_sc <- function(fit_y1 = NULL, fit_y2 = NULL,
                            rho_w = NULL, cb = NULL, ngh = 13L) {
  cache <- lav_bvord_2l_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, ngh = ngh
  )
  cache$theta <- c(rho_w, cb)
  if (cache$nexo_any) {
    tmp <- lav_bvord_2l_x_loglik_cache(cache = cache) # populates pw
    return(lav_bvord_2l_x_sc_cache(cache = cache))
  }
  tmp <- lav_bvord_2l_loglik_cache(cache = cache) # populates pw
  lav_bvord_2l_sc_cache(cache = cache)
}

# total loglik at given values - no cache
lav_bvord_2l_logl <- function(fit_y1 = NULL, fit_y2 = NULL,
                              rho_w = NULL, cb = NULL, ngh = 13L) {
  cache <- lav_bvord_2l_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, ngh = ngh
  )
  cache$theta <- c(rho_w, cb)
  if (cache$nexo_any) {
    return(lav_bvord_2l_x_loglik_cache(cache = cache))
  }
  lav_bvord_2l_loglik_cache(cache = cache)
}

# nlminb objective/gradient (see lav_uvbv_common.R)
lav_bvord_2l_min_fns <- function(covariates = FALSE) {
  if (covariates) {
    lav_uvbv_min_fns(
      logl_fun = lav_bvord_2l_x_loglik_cache,
      grad_fun = lav_bvord_2l_x_grad_cache
    )
  } else {
    lav_uvbv_min_fns(
      logl_fun = lav_bvord_2l_loglik_cache,
      grad_fun = lav_bvord_2l_grad_cache
    )
  }
}

# compute logl for the unrestricted (h1) model -- per group
lav_h1_logl <- function(lavdata = NULL,
                        lavsamplestats = NULL,
                        h1_implied = NULL,
                        lavoptions = NULL) {
  # number of groups
  ngroups <- lavdata@ngroups

  logl_group <- rep(as.numeric(NA), ngroups)

  # should compute logl, or return NA?
  logl_ok <- FALSE
  if (lavoptions$estimator %in% c("ML", "MML")) {
    # check if everything is numeric, OR if we have exogenous
    # factor with 2 levels only
    if (all(lavdata@ov$type == "numeric")) {
      logl_ok <- TRUE
    } else {
      not_idx <- which(lavdata@ov$type != "numeric")
      for (i in not_idx) {
        if (lavdata@ov$type[i] == "factor" &&
          lavdata@ov$exo[i] == 1L &&
          lavdata@ov$nlev[i] == 2L) {
          logl_ok <- TRUE
        } else {
          logl_ok <- FALSE
          break
        }
      }
    }
  }

  # lavsamplestats filled in? (not if no data, or samplestats = FALSE)
  if (length(lavsamplestats@ntotal) == 0L ||
      (!is.null(lavoptions$samplestats) && !lavoptions$samplestats)) {
    logl_ok <- FALSE
  }

  # new in 0.6-9 (so SAM can handle N<P)
  if (!is.null(lavoptions$sample.icov) && !lavoptions$sample.icov) {
    logl_ok <- FALSE
  }

  if (logl_ok) {
    for (g in seq_len(ngroups)) {
      if (lavdata@nlevels > 1L) {
        current_verbose <- lav_verbose()
        if (lav_verbose(FALSE))
            on.exit(lav_verbose(current_verbose), TRUE)
        # EM control arguments (fall back to the multilevel defaults if
        # lavoptions is incomplete)
        em_h1_tol <- lavoptions$em.h1.args$tol
        if (!is.numeric(em_h1_tol)) {
          em_h1_tol <- 1e-04
        }
        em_h1_max_iter <- lavoptions$em.h1.args$max_iter
        if (!is.numeric(em_h1_max_iter)) {
          em_h1_max_iter <- 5000L
        }
        em_h1_min_variance <- lavoptions$em.h1.args$min_variance
        if (!is.numeric(em_h1_min_variance)) {
          em_h1_min_variance <- 1e-05
        }
        em_h1_accel <- lavoptions$em.h1.args$acceleration
        if (is.null(em_h1_accel)) {
          em_h1_accel <- "none" # backwards compatibility
        }
        out_1 <- lav_mvn_cl_em_sat(
          ylp = lavsamplestats@YLp[[g]],
          lp = lavdata@Lp[[g]],
          tol = em_h1_tol,
          min_variance = em_h1_min_variance,
          max_iter = em_h1_max_iter,
          acceleration = em_h1_accel
        )
        lav_verbose(current_verbose)
        # store logl per group
        logl_group[g] <- out_1$logl
      } else if (lavsamplestats@missing.flag) {
        if (lavoptions$conditional.x) {
          # the likelihood is already conditional on x; no x-marginal
          # correction needed
          logl_group[g] <-
            lav_mvreg_mi_loglik_samp(
              yp         = lavsamplestats@missing[[g]],
              res_int    = h1_implied$res.int[[g]],
              res_slopes = h1_implied$res.slopes[[g]],
              res_cov    = h1_implied$res.cov[[g]]
            )
        } else {
          logl_group[g] <-
            lav_mvn_mi_loglik_samp(
              yp     = lavsamplestats@missing[[g]],
              #Mu     = lavsamplestats@missing.h1[[g]]$mu,
              mu     = h1_implied$mean[[g]],
              #Sigma  = lavsamplestats@missing.h1[[g]]$sigma,
              sigma_1  = h1_implied$cov[[g]],
              x_idx  = lavsamplestats@x.idx[[g]],
              x_mean = lavsamplestats@mean.x[[g]],
              x_cov  = lavsamplestats@cov.x[[g]]
            )
        }
      } else { # single-level, complete data
        # all we need is: logdet of covariance matrix, nobs and nvar
        if (lavoptions$conditional.x) {
          logl_group[g] <-
            lav_mvn_h1_loglik_samp(
              sample_cov_logdet =
                lavsamplestats@res.cov.log.det[[g]],
              sample_nvar =
                NCOL(lavsamplestats@res.cov[[g]]),
              sample_nobs = lavsamplestats@nobs[[g]]
            )
        } else {
          logl_group[g] <-
            lav_mvn_h1_loglik_samp(
              sample_cov_logdet = lavsamplestats@cov.log.det[[g]],
              sample_nvar       = NCOL(lavsamplestats@cov[[g]]),
              sample_nobs       = lavsamplestats@nobs[[g]],
              x_idx             = lavsamplestats@x.idx[[g]],
              x_cov             = lavsamplestats@cov.x[[g]]
            )
        }
      } # complete
    } # g
  } # logl.ok is TRUE

  out <- list(
    loglik = sum(logl_group),
    loglik.group = logl_group
  )

  out
}

# model gradient

lav_model_grad <- function(lavmodel = NULL,
                               glist = NULL,
                               lavsamplestats = NULL,
                               lavdata = NULL,
                               lavcache = NULL,
                               type = "free",
                               group_weight = TRUE,
                               delta = NULL,
                               m_el_idx = NULL,
                               x_el_idx = NULL,
                               ceq_simple = FALSE) {
  nmat <- lavmodel@nmat
  estimator <- lavmodel@estimator
  representation <- lavmodel@representation
  meanstructure <- lavmodel@meanstructure
  categorical <- lavmodel@categorical
  group_w_free <- lavmodel@group.w.free
  # fixed_x <- lavmodel@fixed.x
  conditional_x <- lavmodel@conditional.x
  num_idx <- lavmodel@num.idx
  th_idx <- lavmodel@th.idx
  nx_free <- lavmodel@nx.free
  estimator_args <- lavmodel@estimator.args

  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  # random slopes? (rv() modifier): analytic gradient (adjoint pass +
  # a small Jacobian of the implied map); falls back to a numerical
  # approximation of the gradient of the objective if the analytic
  # path is not available
  if (length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L) {
    if (type != "free") {
      lav_msg_fixme(
        "gradient with type != \"free\" is not available for models
         with random slopes")
    }
    x0 <- lav_model_get_parameters(lavmodel, glist = glist)

    # the rs cache
    rs <- NULL
    if (!is.null(lavcache)) {
      rs <- lavcache[[1]]$rs
    }
    if (is.null(rs) && !is.null(lavdata)) {
      rs_info <- try(lav_mvn_cl_rs_info(lavmodel = lavmodel,
                                        lavdata = lavdata),
                     silent = TRUE)
      if (!inherits(rs_info, "try-error")) {
        rs <- list(
          info = rs_info,
          stats = lav_mvn_cl_rs_stats(
            y1 = lavdata@X[[1]], lp = lavdata@Lp[[1]],
            rs_info = rs_info
          )
        )
      }
    }
    if (!is.null(rs)) {
      lavmodel2 <- lav_model_set_parameters(lavmodel, x = x0)
      dx <- try(lav_mvn_cl_rs_grad(lavmodel = lavmodel2, rs = rs),
                silent = TRUE)
      if (!inherits(dx, "try-error") && !is.null(dx)) {
        # match the objective scale (-2 loglik / (2 N))
        return(dx / (2 * lavsamplestats@ntotal))
      }
    }

    # numerical fallback
    obj_f <- function(x) {
      lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
      as.numeric(lav_model_objective(
        lavmodel = lavmodel2,
        lavsamplestats = lavsamplestats, lavdata = lavdata,
        lavcache = lavcache
      ))
    }
    dx <- lav_func_grad_simple(func = obj_f, x = x0)
    return(dx)
  }

  if (estimator == "REML") lav_msg_warn(gettext(
    "analytical gradient not implemented; use numerical approximation"))

  # group.weight
  # FIXME --> block.weight
  if (group_weight) {
    if (estimator %in% c("ML", "PML", "FML", "MML", "REML", "NTRLS", "catML")) {
      group_w <- (unlist(lavsamplestats@nobs) / lavsamplestats@ntotal)
    } else if (estimator == "DLS") {
      if (estimator_args$dls.FtimesNminus1) {
        group_w <- ((unlist(lavsamplestats@nobs) - 1) / lavsamplestats@ntotal)
      } else {
        group_w <- (unlist(lavsamplestats@nobs) / lavsamplestats@ntotal)
      }
    } else {
      # FIXME: double check!
      group_w <- ((unlist(lavsamplestats@nobs) - 1) / lavsamplestats@ntotal)
    }
  } else {
    group_w <- rep(1.0, lavmodel@nblocks)
  }

  # GLS: for single-level complete-data unconditional GLS, the weight
  # matrix is a simple function of the sample moments:
  #   WLS.V = bdiag(S.inv,  0.5 t(D) (S.inv %x% S.inv) D)
  # so the gradient never needs the full WLS.V matrix:
  # - gls_omega_flag: use the same Omega ('POST') approach as ML
  #   (see lav_model_omega); no Delta needed at all
  # - gls_stream_flag: keep the Delta-based *LS route, but compute
  #   WLS.V %*% diff without forming WLS.V (composites, type != "free")
  # excluded (not this simple Kronecker sandwich, or broken/rare):
  # multilevel, conditional.x, (partial) correlation structures,
  # group.w.free, missing data
  gls_stream_flag <- (estimator == "GLS" &&
    !lavmodel@multilevel &&
    !conditional_x &&
    !lavmodel@correlation &&
    !group_w_free &&
    !lavsamplestats@missing.flag)
  gls_omega_flag <- (gls_stream_flag &&
    type == "free" &&
    !lavmodel@composites)

  # ULS/DWLS (continuous): the weight matrix is diagonal (the identity
  # for ULS), so the same Omega approach applies, with
  # omega = [unvech(WLS.VD) *] (S - Sigma) with a doubled diagonal --
  # so that 0.5 t(D) vec(omega) == WLS.VD * vech(S - Sigma) -- and
  # omega_mu = [WLS.VD_mean *] (ybar - mu); no Delta needed at all
  uls_omega_flag <- (estimator %in% c("ULS", "DWLS") &&
    type == "free" &&
    !categorical &&
    !lavmodel@multilevel &&
    !conditional_x &&
    !lavmodel@correlation &&
    !lavmodel@composites &&
    !group_w_free &&
    !lavsamplestats@missing.flag &&
    (estimator == "ULS" || !is.null(lavsamplestats@WLS.VD[[1]])))

  # do we need WLS.est?
  if (estimator %in% c("WLS", "NTRLS", "DLS") ||
    (estimator == "GLS" && !gls_omega_flag) ||
    (estimator %in% c("ULS", "DWLS") && !uls_omega_flag)) {
    # always compute WLS.est
    wls_est <- lav_model_wls_est(lavmodel = lavmodel, glist = glist) # ,
    # cov.x = lavsamplestats@cov.x)
  }

  if (estimator %in% c("ML", "PML", "FML", "REML", "NTRLS", "catML") ||
    gls_omega_flag || uls_omega_flag) {
    # compute moments for all groups
    # if(conditional.x) {
    #    Sigma.hat <- lav_model_cond2joint_sigma(lavmodel = lavmodel,
    #                     GLIST = GLIST,
    #                     extra = (estimator %in% c("ML", "REML","NTRLS")))
    # } else {
    sigma_hat <- lav_model_sigma(
      lavmodel = lavmodel, glist = glist,
      extra = (estimator %in% c(
        "ML", "REML",
        "NTRLS", "catML"
      ))
    )
    # }

    if (meanstructure) {
      # if(conditional.x) {
      #    Mu.hat <- lav_model_mu(lavmodel = lavmodel, GLIST = GLIST)
      # } else {
      mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
      # }
    }

    if (categorical) {
      th <- lav_model_th(lavmodel = lavmodel, glist = glist)
    }

    if (conditional_x) {
      pi0 <- lav_model_pi(lavmodel = lavmodel, glist = glist)
    } else if (estimator == "PML") {
      pi0 <- vector("list", length = lavmodel@nblocks)
    }

    # if (group_w_free) {
    #   gw <- lav_model_gw(lavmodel = lavmodel, glist = glist)
    # }
  } else if (estimator == "DLS" && estimator_args$dls.GammaNT == "model") {
    sigma_hat <- lav_model_sigma(
      lavmodel = lavmodel, glist = glist,
      extra = FALSE
    )
    mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
  } else if (estimator == "MML") {
    th <- lav_model_th(lavmodel = lavmodel, glist = glist)
    mm_theta <- lav_model_theta(lavmodel = lavmodel, glist = glist)
    # gw <- lav_model_gw(lavmodel = lavmodel, glist = glist)
  }

  # four approaches (FIXME!!!! merge this!)
  # - ML approach: using Omega (and Omega.mu)
  #    Omega = 'POST' = Sigma.inv %*% (S - Sigma) %*% t(Sigma.inv)
  #   (still 2x faster than Delta method)
  # - WLS/DWLS/GLS: using Delta + WLS.V; support for fixed.x, conditional.x
  # - (ML)/NTRLS: using Delta, no support for fixed.x, conditional.x
  # - PML/FML/MML: custom

  # composites?
  composites_flag <- lavmodel@composites

  # 1. ML approach (also GLS/ULS since 0.7-1, see the omega flags above)
  if (((estimator == "ML" || estimator == "REML" || estimator == "catML") &&
    lavdata@nlevels == 1L && !composites_flag &&
    !lavmodel@conditional.x) || gls_omega_flag || uls_omega_flag) {
  correlation <- lavmodel@correlation
    if (meanstructure) {
      omega <- lav_model_omega(
        sigma_hat = sigma_hat, mu_hat = mu_hat,
        lavsamplestats = lavsamplestats,
        estimator = estimator,
        estimator_args = estimator_args,
        meanstructure = TRUE,
        conditional_x = conditional_x,
    correlation = correlation, num_idx = num_idx
      )
      omega_mu <- attr(omega, "mu")
    } else {
      omega <- lav_model_omega(
        sigma_hat = sigma_hat, mu_hat = NULL,
        lavsamplestats = lavsamplestats,
        estimator = estimator,
        estimator_args = estimator_args,
        meanstructure = FALSE,
        conditional_x = conditional_x,
    correlation = correlation, num_idx = num_idx
      )
      omega_mu <- vector("list", length = lavmodel@nblocks)
    }

    # compute DX (for all elements in every model matrix)
    dx_1 <- vector("list", length = length(glist))

    for (g in 1:lavmodel@nblocks) {
      # which mm belong to group g?
      mm_in_group <- 1:nmat[g] + cumsum(c(0, nmat))[g]
      mm_names <- names(glist[mm_in_group])

      if (representation == "LISREL") {
        dx_group <- lav_lisrel_df_dmlist(
          glist[mm_in_group],
          omega[[g]],
          omega_mu[[g]]
        )

        # FIXME!!!
        # add empty gamma
        if (lavmodel@conditional.x) {
          dx_group$gamma <- lavmodel@GLIST$gamma
        }

        # only save what we need
        dx_1[mm_in_group] <- dx_group[mm_names]
      } else if (representation == "RAM") {
        dx_group <- lav_ram_df(
          glist[mm_in_group],
          omega[[g]],
          omega_mu[[g]]
        )
        # only save what we need
        dx_1[mm_in_group] <- dx_group[mm_names]
      } else {
        lav_msg_stop(gettext(
          "only LISREL and RAM representation has been implemented for now"))
      }

      # weight by group
      # (also for a single group when group_w != 1; for ML the single-group
      #  weight is always nobs/ntotal == 1, but for GLS it is (nobs-1)/ntotal)
      if (lavmodel@nblocks > 1L || group_w[g] != 1.0) {
        for (mm in mm_in_group) {
          dx_1[[mm]] <- group_w[g] * dx_1[[mm]]
        }
      }
    }

    # extract free parameters

    if (type == "free") {
      if (lavmodel@ceq.simple.only) { # new in 0.6-11
        dx <- numeric(lavmodel@nx.unco)
        for (g in 1:lavmodel@nblocks) {
          mm_in_group <- 1:nmat[g] + cumsum(c(0, nmat))[g]
          for (mm in mm_in_group) {
            m_free_idx <- lavmodel@m.free.idx[[mm]]
            x_unco_idx <- lavmodel@x.unco.idx[[mm]]
            dx[x_unco_idx] <- dx_1[[mm]][m_free_idx]
          }
        }
        if (ceq_simple) {
          dx <- drop(crossprod(lavmodel@ceq.simple.K, dx))
        }
      } else {
        dx <- numeric(nx_free)
        for (g in 1:lavmodel@nblocks) {
          mm_in_group <- 1:nmat[g] + cumsum(c(0, nmat))[g]
          for (mm in mm_in_group) {
            m_free_idx <- lavmodel@m.free.idx[[mm]]
            x_free_idx <- lavmodel@x.free.idx[[mm]]
            dx[x_free_idx] <- dx_1[[mm]][m_free_idx]
          }
        }
      }
    } else {
      dx <- dx_1
      # handle equality constraints
      ### FIXME!!!! TODO!!!!
    }
  } else # ML

  # 1a. ML approach for composites: use Delta directly. The existing
  # per-matrix lav_lisrel_df_dmlist() path does not handle composites
  # (it ignores wmat / WTW.inv). We reuse the standard ML "POST" trick:
  #   dF/dphi = crossprod(Delta, POST)  where
  #     POST = c(-2 * Omega.mu, -D' vec(Omega))
  # This works for any model where lav_model_delta() can produce an
  # analytical Jacobian -- including composites since 0.6-23.
  if ((estimator == "ML" || estimator == "REML" || estimator == "catML") &&
      lavdata@nlevels == 1L && composites_flag &&
      !lavmodel@conditional.x) {
    correlation <- lavmodel@correlation
    if (meanstructure) {
      omega <- lav_model_omega(
        sigma_hat = sigma_hat, mu_hat = mu_hat,
        lavsamplestats = lavsamplestats,
        estimator = estimator,
        meanstructure = TRUE,
        conditional_x = conditional_x,
        correlation = correlation, num_idx = num_idx
      )
      omega_mu <- attr(omega, "mu")
    } else {
      omega <- lav_model_omega(
        sigma_hat = sigma_hat, mu_hat = NULL,
        lavsamplestats = lavsamplestats,
        estimator = estimator,
        meanstructure = FALSE,
        conditional_x = conditional_x,
        correlation = correlation, num_idx = num_idx
      )
      omega_mu <- vector("list", length = lavmodel@nblocks)
    }

    delta <- lav_model_delta(lavmodel = lavmodel, glist = glist)

    if (lavmodel@ceq.simple.only) {
      dx <- numeric(lavmodel@nx.unco)
    } else {
      dx <- numeric(nx_free)
    }

    # lavaan's ML objective is 0.5 * F_ML (lav_model_objective.R: group.fx
     # <- 0.5 * group.fx for ML/REML/NTRLS/catML), so dF/dphi has a 0.5
     # factor relative to the textbook ML loss.
    for (g in 1:lavmodel@nblocks) {
      post_sigma_1 <- -0.5 * lav_mat_dup_pre(
        matrix(omega[[g]], ncol = 1L))
      if (meanstructure) {
        post_mu <- -1 * as.numeric(omega_mu[[g]])
        post <- c(post_mu, post_sigma_1)
      } else {
        post <- post_sigma_1
      }
      # Delta rows: [gw (if group.w.free) | mu (if meanstructure) | vech(Sigma)]
      # group weight is overwritten below via the explicit Poisson term, so
      # pad POST with a leading zero when needed (gradient there is set later)
      if (lavmodel@group.w.free) {
        post <- c(0, post)
      }
      group_dx <- as.numeric(crossprod(delta[[g]], post))
      dx <- dx + group_w[g] * group_dx
    }

    if (lavmodel@ceq.simple.only && ceq_simple) {
      dx <- drop(crossprod(lavmodel@ceq.simple.K, dx))
    }
  } else # ML + composites

  # 2. using Delta - *LS family
  if (estimator %in% c("WLS", "DWLS", "ULS", "GLS", "NTGLS", "DLS")) {
    if (type != "free") {
      if (is.null(delta)) {
        lav_msg_fixme("Delta should be given if type != free")
      }
      # stop("FIXME: WLS gradient with type != free needs fixing!")
    } else {
      delta <- lav_model_delta(lavmodel = lavmodel, glist = glist)
    }

    # note: WLS.obs/WLS.est/Delta/WLS.V(D) and group_w are all per GROUP
    # (for multilevel models, the within/between blocks are stacked per
    #  group in all of them; nblocks == ngroups for single-level models)
    for (g in 1:lavsamplestats@ngroups) {
      # diff <- as.matrix(lavsamplestats@WLS.obs[[g]]  - WLS.est[[g]])
      # group.dx <- -1 * ( t(Delta[[g]]) %*% lavsamplestats@WLS.V[[g]] %*% diff)
      # 0.5-17: use crossprod twice; treat DWLS/ULS special
      if (estimator == "WLS" ||
        estimator == "GLS" ||
        estimator == "DLS" ||
        estimator == "NTRLS") {
        # full weight matrix
        diff <- lavsamplestats@WLS.obs[[g]] - wls_est[[g]]

        if (estimator == "GLS" && gls_stream_flag) {
          # compute WLS.V %*% diff without forming WLS.V
          v11_scale <- 1.0
          if (meanstructure && isTRUE(estimator_args$gls.v11.mplus)) {
            v11_scale <- lavsamplestats@nobs[[g]] /
              (lavsamplestats@nobs[[g]] - 1)
          }
          wv_diff <- lav_samp_wls_v_nt_prod(
            m_icov = lavsamplestats@icov[[g]],
            x = diff,
            meanstructure = meanstructure,
            fixed_x = lavmodel@fixed.x,
            x_idx = lavsamplestats@x.idx[[g]],
            v11_scale = v11_scale
          )
          group_dx <- -1 * crossprod(delta[[g]], wv_diff)

          # full weight matrix
        } else if (estimator == "GLS" || estimator == "WLS") {
          wls_v <- lavsamplestats@WLS.V[[g]]
          group_dx <- -1 * crossprod(
            delta[[g]],
            crossprod(wls_v, diff)
          )
        } else if (estimator == "DLS") {
          if (estimator_args$dls.GammaNT == "sample") {
            wls_v <- lavsamplestats@WLS.V[[g]] # for now
          } else {
            wls_v <- lav_dls_wls_v_g(
              m_cov         = sigma_hat[[g]],
              m_mean        = mu_hat[[g]],
              nacov_g       = lavsamplestats@NACOV[[g]],
              dls_a         = estimator_args$dls.a,
              x_idx         = lavsamplestats@x.idx[[g]],
              fixed_x       = lavmodel@fixed.x,
              conditional_x = lavmodel@conditional.x,
              meanstructure = lavmodel@meanstructure
            )
          }
          group_dx <- -1 * crossprod(
            delta[[g]],
            crossprod(wls_v, diff)
          )
        } else if (estimator == "NTRLS") {
          stopifnot(!conditional_x)
          # WLS.V <- lav_samp_gamma_inverse_nt(
          #         m_icov = attr(Sigma.hat[[g]],"inv")[,,drop=FALSE],
          #         m_cov            = Sigma.hat[[g]][,,drop=FALSE],
          #         m_mean           = Mu.hat[[g]],
          #         x_idx          = lavsamplestats@x.idx[[g]],
          #         fixed_x        = fixed.x,
          #         conditional_x  = conditional.x,
          #         meanstructure  = meanstructure,
          #         slopestructure = conditional.x)

          s <- lavsamplestats@cov[[g]]
          sigma_1 <- sigma_hat[[g]]
          sigma_inv <- attr(sigma_1, "inv")
          nvar <- NROW(sigma_1)

          if (meanstructure) {
            mean_1 <- lavsamplestats@mean[[g]]
            mu <- mu_hat[[g]]
            post_sigma_1 <- lav_mat_dup_pre(
              matrix(
                (sigma_inv %*% (s - sigma_1) %*% t(sigma_inv)) %*%
                  (diag(nvar) + (s - sigma_1) %*% sigma_inv) +
                  (sigma_inv %*% tcrossprod(mean_1 - mu) %*% sigma_inv),
                ncol = 1
              )
            )
            post_mu <- as.numeric(2 * sigma_inv %*% (mean_1 - mu))
            post <- c(post_mu, post_sigma_1)
          } else {
            post <- lav_mat_dup_pre(
              matrix((sigma_inv %*% (s - sigma_1) %*% t(sigma_inv)) %*%
                (diag(nvar) + (s - sigma_1) %*% sigma_inv), ncol = 1)
            )
          }

          group_dx <- as.numeric(-1 * crossprod(delta[[g]], post))
        }
      } else if (estimator == "DWLS" || estimator == "ULS") {
        # diagonal weight matrix
        diff <- lavsamplestats@WLS.obs[[g]] - wls_est[[g]]
        group_dx <- -1 * crossprod(
          delta[[g]],
          lavsamplestats@WLS.VD[[g]] * diff
        )
      }

      group_dx <- group_w[g] * group_dx
      if (g == 1) {
        dx <- group_dx
      } else {
        dx <- dx + group_dx
      }
    } # g

    if (type == "free") {
      # *LS Delta has one column per *unconstrained* parameter (nx.unco); with
      # simple equality constraints (ceq.simple, e.g. across-group equal
      # loadings) reduce the gradient to the free parameters (nx.free), exactly
      # as the ML path above. Without this the gradient is too long and breaks
      # the optimizer (parscale length mismatch) for ceq.simple models -- this
      # is what blocked categorical (DWLS) multigroup SAM with across-group
      # measurement constraints.
      if (lavmodel@ceq.simple.only && ceq_simple) {
        dx <- drop(crossprod(lavmodel@ceq.simple.K, dx))
      }
    } else {
      # make a GLIST
      dx <- lav_model_x2glist(
        lavmodel = lavmodel, x = dx,
        type = "custom", set_delta = FALSE,
        m_el_idx = m_el_idx,
        x_el_idx = x_el_idx
      )
    }

  # WLS
  # ML + conditional.x
  } else if (estimator %in% c("ML", "catML") && lavmodel@conditional.x &&
    lavdata@nlevels == 1L) {
    if (type != "free") {
      if (is.null(delta)) {
        lav_msg_fixme("Delta should be given if type != free")
      }
      # stop("FIXME: WLS gradient with type != free needs fixing!")
    } else {
      delta <- lav_model_delta(lavmodel = lavmodel, glist = glist)
    }

    for (g in 1:lavmodel@nblocks) {
      # missing data (FIML)?
      if (lavsamplestats@missing.flag) {
        sigma_1 <- sigma_hat[[g]]
        sigma_inv <- attr(sigma_1, "inv")

        # dlogl with respect to (vec(Beta), vech(res.cov)) -- the same
        # stat order as the conditional.x DELTA rows
        post <- c(
          lav_mvreg_mi_dlogl_dbeta_samp(
            yp = lavsamplestats@missing[[g]],
            res_int = mu_hat[[g]], res_slopes = pi0[[g]],
            res_cov = sigma_1, res_cov_inv = sigma_inv
          ),
          lav_mvreg_mi_dlogl_dvechrescov_samp(
            yp = lavsamplestats@missing[[g]],
            res_int = mu_hat[[g]], res_slopes = pi0[[g]],
            res_cov = sigma_1, res_cov_inv = sigma_inv
          )
        )

        # the objective is sum_g (nobs_g/ntotal) * 0.5 * (-2*logl_g/nobs_g
        # - h1_g), so dF/dtheta = -(1/ntotal) * sum_g dlogl_g/dtheta;
        # note: as in the unconditional missing-data path, the 1/ntotal
        # factor is applied here directly and group_w is NOT used (for
        # missing data, the callers -- optimization, hessian -- use
        # group_weight = FALSE)
        group_dx <- as.numeric(
          -1 / lavsamplestats@ntotal * crossprod(delta[[g]], post)
        )

        if (g == 1) {
          dx <- group_dx
        } else {
          dx <- dx + group_dx
        }
        next
      }

      # augmented mean.x + cov.x matrix
      mean_x <- lavsamplestats@mean.x[[g]]
      cov_x <- lavsamplestats@cov.x[[g]]
      c3 <- rbind(
        c(1, mean_x),
        cbind(mean_x, cov_x + tcrossprod(mean_x))
      )

      sigma_1 <- sigma_hat[[g]]
      mu_g <- mu_hat[[g]]
      pi_g <- pi0[[g]]
      sigma_inv <- attr(sigma_1, "inv")
      nvar <- NROW(sigma_1)
      s <- lavsamplestats@res.cov[[g]]

      # beta
      obs <- t(cbind(
        lavsamplestats@res.int[[g]],
        lavsamplestats@res.slopes[[g]]
      ))
      est <- t(cbind(mu_g, pi_g))
      # obs.beta <- c(lavsamplestats@res.int[[g]],
      #              lav_mat_vec(lavsamplestats@res.slopes[[g]]))
      # est.beta <- c(Mu.g,  lav_mat_vec(PI.g))
      # beta.COV <- C3 %x% Sigma.inv

      # a <- t(obs.beta - est.beta)
      # b <- as.matrix(obs.beta - est.beta)
      # K <- lav_mat_com(m = nvar, n = nvar)
      # AB <- (K %x% diag(NROW(C3)*NROW(C3))) %*%
      #          (diag(nvar) %x% lav_mat_vec(C3) %x% diag(nvar))
      # K <- lav_mat_com(m = nvar, n = NROW(C3))
      # AB <- ( diag(NROW(C3)) %x% K %x% diag(nvar) ) %*%
      #        (lav_mat_vec(C3) %x% diag( nvar * nvar) )

      # POST.beta <- 2 *  beta.COV %*% (obs.beta - est.beta)
      d_beta <- c3 %*% (obs - est) %*% sigma_inv
      # NOTE: the vecr here, unlike lav_mvreg_dlogl_beta
      #       this is because DELTA has used vec(t(BETA)),
      #       instead of vec(BETA)
      # POST.beta <- 2 * lav_mat_vecr(d.BETA)
      # NOT any longer, since 0.6-1!!!
      post_beta <- 2 * lav_mat_vec(d_beta)

      # POST.sigma1 <- lav_mat_dup_pre(
      #        (Sigma.inv %x% Sigma.inv)  %*% t(AB)  %*% (t(a) %x% b) )

      # Sigma
      # POST.sigma2 <- lav_mat_dup_pre(
      #                 matrix( lav_mat_vec(
      #          Sigma.inv %*% (S - Sigma) %*% t(Sigma.inv)), ncol = 1L))
      w_tilde <- s + t(obs - est) %*% c3 %*% (obs - est)
      d_sigma <- (sigma_inv - sigma_inv %*% w_tilde %*% sigma_inv)
      d_vech_sigma <- as.numeric(lav_mat_dup_pre(
        as.matrix(lav_mat_vec(d_sigma))
      ))
      post_sigma <- -1 * d_vech_sigma

      # POST <- c(POST.beta, POST.sigma1 + POST.sigma2)
      post <- c(post_beta, post_sigma)

      group_dx <- as.numeric(-1 * crossprod(delta[[g]], post))

      # because we still use obj/2, we need to divide by 2!
      group_dx <- group_dx / 2 # fixed in 0.6-1

      group_dx <- group_w[g] * group_dx
      if (g == 1) {
        dx <- group_dx
      } else {
        dx <- dx + group_dx
      }
    } # g

    if (type == "free") {
      # nothing to do
    } else {
      # make a GLIST
      dx <- lav_model_x2glist(
        lavmodel = lavmodel, x = dx,
        type = "custom", set_delta = FALSE,
        m_el_idx = m_el_idx,
        x_el_idx = x_el_idx
      )
    }

  # ML + conditional.x
  } else if (estimator == "ML" && lavdata@nlevels > 1L) {
    if (type != "free") {
      lav_msg_fixme("type != free in lav_model_grad for
                    estimator ML for nlevels > 1")
    } else {
      delta <- lav_model_delta(lavmodel = lavmodel, glist = glist)
    }

    # for each upper-level group....
    for (g in 1:lavmodel@ngroups) {
      if (!lavsamplestats@missing.flag) { # complete data
        if (lavmodel@conditional.x) {
          dx_1 <- lav_mvreg_cl_dlogl_2l_samp(
            ylp = lavsamplestats@YLp[[g]],
            lp = lavdata@Lp[[g]],
            res_sigma_w = sigma_hat[[(g - 1) * 2 + 1]],
            res_int_w = mu_hat[[(g - 1) * 2 + 1]],
            res_pi_w = pi0[[(g - 1) * 2 + 1]],
            res_sigma_b = sigma_hat[[(g - 1) * 2 + 2]],
            res_int_b = mu_hat[[(g - 1) * 2 + 2]],
            res_pi_b = pi0[[(g - 1) * 2 + 2]],
            sinv_method = "eigen"
          )
          # reorder from the kernel statistic order to the Delta row order
          perm <- lav_mvreg_cl_stat_perm(
            res_int_w = mu_hat[[(g - 1) * 2 + 1]],
            res_pi_w = pi0[[(g - 1) * 2 + 1]],
            res_int_b = mu_hat[[(g - 1) * 2 + 2]],
            res_pi_b = pi0[[(g - 1) * 2 + 2]]
          )
          dx_1 <- dx_1[perm]
        } else {
          dx_1 <- lav_mvn_cl_dlogl_2l_samp(
            ylp = lavsamplestats@YLp[[g]],
            lp = lavdata@Lp[[g]],
            mu_w = mu_hat[[(g - 1) * 2 + 1]],
            sigma_w = sigma_hat[[(g - 1) * 2 + 1]],
            mu_b = mu_hat[[(g - 1) * 2 + 2]],
            sigma_b = sigma_hat[[(g - 1) * 2 + 2]],
            sinv_method = "eigen"
          )
        }
      } else {
        # missing data
        if (lavmodel@conditional.x) {
          lav_msg_stop(gettext("gradient for twolevel + conditional.x + fiml
                  is not ready; use optim.gradient = \"numerical\""))
        } else {
          dx_1 <- lav_mvn_cl_mi_dlogl_2l_samp(
            y1 = lavdata@X[[g]],
            y2 = lavsamplestats@YLp[[g]][[2]]$Y2,
            lp = lavdata@Lp[[g]],
            mp = lavdata@Mp[[g]],
            mu_w = mu_hat[[(g - 1) * 2 + 1]],
            sigma_w = sigma_hat[[(g - 1) * 2 + 1]],
            mu_b = mu_hat[[(g - 1) * 2 + 2]],
            sigma_b = sigma_hat[[(g - 1) * 2 + 2]],
            sinv_method = "eigen"
          )
        }
      }

      group_dx <- as.numeric(dx_1 %*% delta[[g]])

      # group weights (if any)
      group_dx <- group_w[g] * group_dx
      if (g == 1) {
        dx <- group_dx
      } else {
        dx <- dx + group_dx
      }
    } # g

    # divide by 2 * N
    dx <- dx / (2 * lavsamplestats@ntotal)

    # Delta has one column per *unconstrained* parameter (nx.unco); with
    # simple equality constraints (ceq.simple, e.g. across-level/across-group
    # equal loadings) reduce the gradient to the free parameters (nx.free),
    # exactly as the single-level paths above. Without this the gradient is too
    # long and breaks the optimizer (parscale length mismatch) for ceq.simple
    # multilevel models.
    if (lavmodel@ceq.simple.only && ceq_simple) {
      dx <- drop(crossprod(lavmodel@ceq.simple.K, dx))
    }

  # cat("dx1 (numerical) = \n"); print( zapsmall(dx1) )
  # cat("dx  (analytic)  = \n"); print( zapsmall(dx ) )
  # ML + two-level
  } else if (estimator == "PML" || estimator == "FML" ||
    estimator == "MML") {
    if (type != "free") {
      lav_msg_fixme("type != free in lav_model_grad for estimator PML")
    } else {
      delta <- lav_model_delta(lavmodel = lavmodel, glist = glist)
    }

    for (g in 1:lavmodel@nblocks) {
      # print(GLIST)
      # print(lav_model_get_parameters(lavmodel = lavmodel, glist = GLIST))
      # print(Sigma.hat[[g]])
      # print(TH[[g]])
      # cat("*****\n")

      # compute partial derivative of logLik with respect to
      # thresholds/means, slopes, variances, correlations
      if (estimator == "PML") {
        if (lavdata@nlevels > 1L) {
          lav_msg_stop(gettext(
            "PL gradient + multilevel not implemented;
            try optim.gradient = \"numerical\""))
        } else if (conditional_x) {
          d1 <- lav_pml_dploglik_dimplied(
            sigma_hat = sigma_hat[[g]],
            mu_hat = mu_hat[[g]],
            th = th[[g]],
            th_idx = th_idx[[g]],
            num_idx = num_idx[[g]],
            x = lavdata@X[[g]],
            lavcache = lavcache[[g]],
            exo = lavdata@eXo[[g]],
            wt = lavdata@weights[[g]],
            pi0 = pi0[[g]],
            missing = lavdata@missing
          )
        } else {
          d1 <- lav_pml_dploglik_dimplied(
            sigma_hat = sigma_hat[[g]],
            mu_hat = mu_hat[[g]],
            th = th[[g]],
            th_idx = th_idx[[g]],
            num_idx = num_idx[[g]],
            x = lavdata@X[[g]],
            lavcache = lavcache[[g]],
            exo = NULL,
            wt = lavdata@weights[[g]],
            pi0 = NULL,
            missing = lavdata@missing
          )
        } # not conditional.x

        # chain rule (fmin)
        group_dx <-
          as.numeric(t(d1) %*% delta[[g]])
    # PML
    } else if (estimator == "FML") {
        d1 <- lav_pml_fml_dploglik_dimplied(
          sigma_hat = sigma_hat[[g]],
          th = th[[g]],
          th_idx = th_idx[[g]],
          num_idx = num_idx[[g]],
          x = lavdata@X[[g]],
          lavcache = lavcache[[g]]
        )

        # chain rule (fmin)
        group_dx <-
          as.numeric(t(d1) %*% delta[[g]]) / lavsamplestats@nobs[[g]]
      } else if (estimator == "MML") {
        group_dx <-
          lav_model_grad_mml(
            lavmodel = lavmodel,
            glist = glist,
            mm_theta = mm_theta[[g]],
            th = th[[g]],
            group = g,
            lavdata = lavdata,
            sample_mean = lavsamplestats@mean[[g]],
            sample_mean_x = lavsamplestats@mean.x[[g]],
            lavcache = lavcache
          )
      }

      # group weights (if any)
      # note (fixed July 2026): NOT for PML. The PML objective is the plain
      # SUM of the group objectives (each -logPL_g already carries its own
      # N_g; see lav_model_objective(): "no weighting needed!"), so the
      # gradient must be the plain sum of the (total) group gradients as
      # well. The n_g/N weighting made the analytic gradient inconsistent
      # with the objective for multigroup PML and -- through the
      # hessian-based information -- inflated all multigroup PML standard
      # errors by a factor sqrt(N/n_g). Single-group fits are unaffected
      # (group_w = 1).
      if (estimator != "PML") {
        group_dx <- group_w[g] * group_dx
      }
      if (g == 1) {
        dx <- group_dx
      } else {
        dx <- dx + group_dx
      }
    } # g
  } else {
    lav_msg_stop(gettext(
      "no analytical gradient available for estimator"), estimator)
  }


  # group.w.free for ML
  if (lavmodel@group.w.free &&
    estimator %in% c("ML", "MML", "FML", "PML", "REML", "catML")) {
    # est.prop <- unlist( lav_model_gw(lavmodel = lavmodel, GLIST = GLIST) )
    # obs.prop <- unlist(lavsamplestats@group.w)
    # FIXME: G2 based -- ML and friends only!!
    # dx.GW <- - (obs.prop - est.prop)

    # poisson version
    est_freq <- exp(unlist(lav_model_gw(lavmodel = lavmodel, glist = glist)))
    obs_freq <- unlist(lavsamplestats@group.w) * lavsamplestats@ntotal
    dx_gw <- -(obs_freq - est_freq)
    # divide by N (to be consistent with the rest of lavaan)
    dx_gw <- dx_gw / lavsamplestats@ntotal

    # remove last element (fixed LAST group to zero)
    # dx.GW <- dx.GW[-length(dx.GW)]

    # fill in in dx
    gw_mat_idx <- which(names(lavmodel@GLIST) == "gw")
    gw_x_idx <- unlist(lavmodel@x.free.idx[gw_mat_idx])
    dx[gw_x_idx] <- dx_gw
  }

  # dx is 1xnpar matrix of LIST (type != "free")
  if (is.matrix(dx)) {
    dx <- as.numeric(dx)
  }

  dx
}

# for testing purposes only
lav_model_delta_numerical <- function(lavmodel = NULL, glist = NULL, g = 1L) {

  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  compute_moments <- function(x) {
    glist <- lav_model_x2glist(lavmodel = lavmodel, x = x, type = "free")
    sigma_hat <- lav_model_sigma(lavmodel = lavmodel, glist = glist)
    s_vec <- lav_mat_vech(sigma_hat[[g]])
    if (lavmodel@meanstructure) {
      mu_hat <- lav_model_mu(lavmodel = lavmodel, glist = glist)
      out <- c(mu_hat[[g]], s_vec)
    } else {
      out <- s_vec
    }
    out
  }

  x <- lav_model_get_parameters(lavmodel = lavmodel,
                                glist = glist, type = "free")
  delta <- lav_func_jacobian_complex(func = compute_moments, x = x)

  delta
}


lav_model_ddelta_dx <- function(lavmodel = NULL, glist = NULL,
   target = "lambda", ceq_simple = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  representation <- lavmodel@representation
  nmat <- lavmodel@nmat
  nblocks <- lavmodel@nblocks
  th_idx <- lavmodel@th.idx

  # number of columns in DELTA + m.el.idx/x.el.idx
  type <- "free"
  # if(type == "free") {
  if (lavmodel@ceq.simple.only) {
    n_col <- lavmodel@nx.unco
  } else {
    n_col <- lavmodel@nx.free
  }
  m_el_idx <- x_el_idx <- vector("list", length = length(glist))
  for (mm in seq_along(glist)) {
    m_el_idx[[mm]] <- lavmodel@m.free.idx[[mm]]
    if (lavmodel@ceq.simple.only) {
      x_el_idx[[mm]] <- lavmodel@x.unco.idx[[mm]]
    } else {
      x_el_idx[[mm]] <- lavmodel@x.free.idx[[mm]]
    }
    # handle symmetric matrices
    if (lavmodel@isSymmetric[mm]) {
      # since we use 'x.free.idx', only symmetric elements
      # are duplicated (not the equal ones, only in x.free.free)
      dix <- duplicated(x_el_idx[[mm]])
      if (any(dix)) {
        m_el_idx[[mm]] <- m_el_idx[[mm]][!dix]
        x_el_idx[[mm]] <- x_el_idx[[mm]][!dix]
      }
    }
  }
  # } else {
  #    n_col <- sum(unlist(lapply(x.el.idx, function(x) length(unique(x)))))
  # }

  # compute Delta per group
  delta <- vector("list", length = nblocks)
  for (g in 1:nblocks) {
    mm_in_group <- 1:nmat[g] + cumsum(c(0, nmat))[g]
    delta_group <- NULL
    for (mm in mm_in_group) {
      mname <- names(lavmodel@GLIST)[mm]

      # skip empty ones
      if (!length(m_el_idx[[mm]])) next

      # get Delta columns for this model matrix
      if (representation == "LISREL") {
        if (target == "lambda") {
         mm_delta <- lav_lisrel_dlambda_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "th") {
          mm_delta <- lav_lisrel_dth_dx(
             mlist = glist[mm_in_group], m = mname, th_idx = th_idx[[g]],
            idx = m_el_idx[[mm]],
            delta = TRUE
          )
        } else if (target == "mu") {
          mm_delta <- lav_lisrel_dmu_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "nu") {
          mm_delta <- lav_lisrel_dnu_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "tau") {
          mm_delta <- lav_lisrel_dtau_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "theta") {
          mm_delta <- lav_lisrel_dtheta_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "gamma") {
          mm_delta <- lav_lisrel_dgamma_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "beta") {
          mm_delta <- lav_lisrel_dbeta_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "alpha") {
          mm_delta <- lav_lisrel_dalpha_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "psi") {
          mm_delta <- lav_lisrel_dpsi_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]]
          )
        } else if (target == "sigma") {
          mm_delta <- lav_lisrel_dsigma_dx(
            mlist = glist[mm_in_group],
            m = mname,
            idx = m_el_idx[[mm]],
            delta = TRUE
          )
        } else {
          lav_msg_stop(gettextf("target %s not implemented yet", target))
        }

        # initialize?
        if (is.null(delta_group)) {
          delta_group <- matrix(0, nrow = nrow(mm_delta), ncol = n_col)
        }
        delta_group[, x_el_idx[[mm]]] <- mm_delta
      }
    } # mm

    if (type == "free" && ceq_simple && lavmodel@ceq.simple.only) {
      delta_group <- delta_group %*% lavmodel@ceq.simple.K
    }

    delta[[g]] <- delta_group
  } # g

  delta
}


# Compute the Delta matrix: Jacobian of the model-implied moments wrt the
# free parameters, per block.
#
# Supported:
#   - LISREL representation, full surface: continuous + categorical +
#     correlation, conditional.x on/off, parameterization "delta" or
#     "theta", composites, group.w.free. Both the standard "free" indexing
#     and caller-supplied m.el.idx./x.el.idx. (e.g. lav_pml_test_plrt(),
#     modindices()) are handled.
#   - RAM representation, classical mean- and covariance structures only.
#     RAM with categorical / correlation / conditional.x / composites is
#     not supported and produces an error.
#
# This is the housekeeping wrapper: it loops over blocks, extracts the
# block-specific MLIST and free-element indices, and dispatches the actual
# computation to the representation-specific worker.
lav_model_delta <- function(lavmodel = NULL, glist = NULL,
                            m_el_idx = NULL, x_el_idx = NULL,
                            ceq_simple = FALSE) {

  # two-level + categorical: custom standardized Jacobian, stacked per
  # group (see lav_model_wls_2l_cat.R); 'free' parameters only
  if (lavmodel@multilevel && lavmodel@categorical &&
    is.null(m_el_idx)) {
    return(lav_m07_delta(lavmodel = lavmodel, glist = glist))
  }

  representation <- lavmodel@representation
  if (representation == "RAM") {
    if (lavmodel@conditional.x || lavmodel@categorical ||
        lavmodel@correlation   || lavmodel@composites) {
      lav_msg_stop(gettext(
        "lav_model_delta(): RAM representation does not support
         conditional.x, categorical, correlation, or composites."))
    }
  } else if (representation != "LISREL") {
    lav_msg_stop(gettextf(
      "lav_model_delta(): representation %s not supported.",
      representation))
  }

  # state or final?
  if (is.null(glist)) {
    glist_1 <- lavmodel@GLIST
  } else {
    glist_1 <- glist
  }

  # which (model element, x column) lists do we use?
  #  - 'free' (default): use lavmodel@m.free.idx + x.free.idx (or x.unco.idx
  #     when ceq.simple.only); n_col = nx.free (or nx.unco). The worker writes
  #     Delta with nx.unco columns when ceq.simple.only; the caller maps to
  #     nx.free via ceq.simple.K.
  #  - 'custom': caller passes m.el.idx./x.el.idx. directly (e.g. modindices
  #     after lav_object_extended()); n_col = max column index.
  use_custom <- !is.null(m_el_idx) || !is.null(x_el_idx)
  if (use_custom) {
    m_idx <- m_el_idx
    x_idx <- x_el_idx
    nx    <- max(unlist(x_idx), 0L)
  } else if (lavmodel@ceq.simple.only) {
    m_idx <- lavmodel@m.free.idx
    x_idx <- lavmodel@x.unco.idx
    nx    <- lavmodel@nx.unco
  } else {
    m_idx <- lavmodel@m.free.idx
    x_idx <- lavmodel@x.free.idx
    nx    <- lavmodel@nx.free
  }

  delta <- vector("list", length = lavmodel@nblocks)
  for (b in seq_len(lavmodel@nblocks)) {
    mm_in_group <-
      seq_len(lavmodel@nmat[b]) + cumsum(c(0, lavmodel@nmat))[b]

    if (representation == "LISREL") {
      delta[[b]] <- lav_lisrel_dimplied_dx(
        mlist            = glist_1[mm_in_group],
        m_free_idx       = m_idx[mm_in_group],
        x_free_idx       = x_idx[mm_in_group],
        nx_free          = nx,
        meanstructure    = lavmodel@meanstructure,
        categorical      = lavmodel@categorical,
        correlation      = lavmodel@correlation,
        conditional_x    = lavmodel@conditional.x,
        num_idx          = lavmodel@num.idx[[b]],
        th_idx           = lavmodel@th.idx[[b]],
        group_w_free     = lavmodel@group.w.free,
        parameterization = lavmodel@parameterization
      )
    } else { # RAM
      delta[[b]] <- lav_ram_dimplied_dx(
        mlist         = glist_1[mm_in_group],
        m_free_idx    = m_idx[mm_in_group],
        x_free_idx    = x_idx[mm_in_group],
        nx_free       = nx,
        meanstructure = lavmodel@meanstructure,
        group_w_free  = lavmodel@group.w.free
      )
    }
  }

  # if multilevel, rbind levels within group
  if (lavmodel@multilevel) {
    delta_tmp <- vector("list", length = lavmodel@ngroups)
    for (g in 1:lavmodel@ngroups) {
      delta_tmp[[g]] <- rbind(
        delta[[(g - 1) * 2 + 1]],
        delta[[(g - 1) * 2 + 2]]
      )
    }
    delta <- delta_tmp
  }

  delta
}



lav_model_omega <- function(sigma_hat = NULL, mu_hat = NULL,
                         lavsamplestats = NULL, estimator = "ML",
                         estimator_args = list(),
                         meanstructure = FALSE, conditional_x = FALSE,
             correlation = FALSE, num_idx = NULL) {
  # nblocks
  nblocks <- length(sigma_hat)

  omega <- vector("list", length = nblocks)
  omega_mu <- vector("list", length = nblocks)

  for (g in 1:nblocks) {
    # ML
    if (estimator %in% c("ML", "REML", "catML")) {
      if (attr(sigma_hat[[g]], "po") == FALSE) {
        # FIXME: WHAT IS THE BEST THING TO DO HERE??
        # CURRENTLY: stop
        lav_msg_warn(gettext(
          "lav_model_grad: Sigma.hat is not positive definite\n"))
        sigma_hat_inv <- MASS::ginv(sigma_hat[[g]])
      } else {
        sigma_hat_inv <- attr(sigma_hat[[g]], "inv")
      }

      if (!lavsamplestats@missing.flag) { # complete data
        if (meanstructure) {
          if (conditional_x) {
            diff <- lavsamplestats@res.int[[g]] - mu_hat[[g]]
            w_tilde <- lavsamplestats@res.cov[[g]] + tcrossprod(diff)
          } else {
            diff <- lavsamplestats@mean[[g]] - mu_hat[[g]]
            w_tilde <- lavsamplestats@cov[[g]] + tcrossprod(diff)
          }
          # Browne 1995 eq 4.55
          omega_mu[[g]] <- t(t(diff) %*% sigma_hat_inv)
          omega[[g]] <-
            (sigma_hat_inv %*% (w_tilde - sigma_hat[[g]]) %*%
              sigma_hat_inv)
        } else {
          if (conditional_x) {
            w_tilde <- lavsamplestats@res.cov[[g]]
          } else {
            w_tilde <- lavsamplestats@cov[[g]]
          }
          omega[[g]] <-
            (sigma_hat_inv %*% (w_tilde - sigma_hat[[g]]) %*%
              sigma_hat_inv)
        }
      } else { # missing data
        m <- lavsamplestats@missing[[g]]

        nvar <- ncol(lavsamplestats@cov[[g]])
        omega_1 <- matrix(0, nvar, nvar)
        omega_mu_1 <- matrix(0, nvar, 1)

        for (p in seq_along(m)) {
          sx <- m[[p]][["SY"]]
          mx <- m[[p]][["MY"]]
          nobs <- m[[p]][["freq"]]
          var_idx <- m[[p]][["var.idx"]]

          sigma_inv <- try(chol2inv(chol(
                          sigma_hat[[g]][var_idx, var_idx, drop = FALSE])),
                          silent = TRUE)
          if (inherits(sigma_inv, "try-error")) {
            sigma_inv <- MASS::ginv(
                          sigma_hat[[g]][var_idx, var_idx, drop = FALSE])
          }
          mu <- mu_hat[[g]][var_idx]
          w_tilde <- sx + tcrossprod(mx - mu)

          omega_mu_1[var_idx, 1] <-
            (omega_mu_1[var_idx, 1] + nobs / lavsamplestats@ntotal *
              t(t(mx - mu) %*% sigma_inv))

          omega_1[var_idx, var_idx] <-
            (omega_1[var_idx, var_idx] + nobs / lavsamplestats@ntotal *
              (sigma_inv %*%
                (w_tilde - sigma_hat[[g]][var_idx, var_idx, drop = FALSE]) %*%
                sigma_inv))
        }
        omega_mu[[g]] <- omega_mu_1
        omega[[g]] <- omega_1
      } # missing

      # GLS
    } else if (estimator == "GLS") {
      # rides the same 'POST' convention as ML; the caller weights each
      # group by (nobs - 1)/ntotal (see group_w in lav_model_grad), so
      # omega itself is the unweighted S.inv (S - Sigma) S.inv
      w_inv <- lavsamplestats@icov[[g]]
      m_w <- lavsamplestats@cov[[g]]
      omega[[g]] <- w_inv %*% (m_w - sigma_hat[[g]]) %*% w_inv
      if (meanstructure) {
        diff <- as.matrix(lavsamplestats@mean[[g]] - mu_hat[[g]])
        omega_mu[[g]] <- w_inv %*% diff
        if (isTRUE(estimator_args$gls.v11.mplus)) {
          # V11 was rescaled by n/(n-1) (see lav_samp_from_data)
          omega_mu[[g]] <- omega_mu[[g]] * lavsamplestats@nobs[[g]] /
            (lavsamplestats@nobs[[g]] - 1)
        }
      }

      # ULS
    } else if (estimator == "ULS") {
      # identity weight matrix: the 'POST' convention requires
      # 0.5 t(D) vec(omega) == vech(S - Sigma), so double the diagonal
      omega[[g]] <- lavsamplestats@cov[[g]] - sigma_hat[[g]]
      diag(omega[[g]]) <- 2 * diag(omega[[g]])
      if (meanstructure) {
        omega_mu[[g]] <- as.matrix(
          lavsamplestats@mean[[g]] - mu_hat[[g]])
      }

      # DWLS (continuous)
    } else if (estimator == "DWLS") {
      # diagonal weight matrix (the @WLS.VD vector): the 'POST'
      # convention requires 0.5 t(D) vec(omega) == WLS.VD * vech(S -
      # Sigma), so weight the residuals elementwise and double the
      # diagonal; this also reproduces the structural zero weights of
      # the x/x combinations when fixed.x = TRUE
      vd <- lavsamplestats@WLS.VD[[g]]
      nvar <- NCOL(sigma_hat[[g]])
      if (meanstructure) {
        vd_mean <- vd[seq_len(nvar)]
        vd_cov <- vd[-seq_len(nvar)]
      } else {
        vd_cov <- vd
      }
      omega[[g]] <- lav_mat_vech_rev(vd_cov) *
        (lavsamplestats@cov[[g]] - sigma_hat[[g]])
      diag(omega[[g]]) <- 2 * diag(omega[[g]])
      if (meanstructure) {
        omega_mu[[g]] <- as.matrix(
          vd_mean * (lavsamplestats@mean[[g]] - mu_hat[[g]]))
      }
    }

    # new in 0.6-18
    # the diagonal of omega is the gradient wrt the (fixed) unit variances;
    # zero it out for the correlation variables only. For a 'partial'
    # correlation structure, num_idx holds the non-correlation variables
    # (which keep a free residual variance), so we zero the complement.
    if (correlation) {
      ni <- num_idx[[g]]
      if (length(ni) > 0L) {
        cor_idx <- seq_len(nrow(omega[[g]]))[-ni]
      } else {
        cor_idx <- seq_len(nrow(omega[[g]]))
      }
      diag(omega[[g]])[cor_idx] <- 0
    }
  } # g

  if (meanstructure) attr(omega, "mu") <- omega_mu

  omega
}

lav_model_grad_dd <- function(lavmodel, g_list = NULL, group = 1L) {
  if (is.null(g_list)) g_list <- lavmodel@GLIST

  #### FIX th + mu!!!!!
  delta_lambda <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "lambda")[[group]]
  delta_tau <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "tau")[[group]]
  delta_nu <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "nu")[[group]]
  delta_theta <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "theta")[[group]]
  delta_beta <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "beta")[[group]]
  delta_psi <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "psi")[[group]]
  delta_alpha <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "alpha")[[group]]
  delta_gamma <-
    lav_model_ddelta_dx(lavmodel, glist = g_list, target = "gamma")[[group]]

  ov_y_dummy_ov_idx <- lavmodel@ov.y.dummy.ov.idx[[group]]
  ov_x_dummy_ov_idx <- lavmodel@ov.x.dummy.ov.idx[[group]]
  ov_y_dummy_lv_idx <- lavmodel@ov.y.dummy.lv.idx[[group]]
  ov_x_dummy_lv_idx <- lavmodel@ov.x.dummy.lv.idx[[group]]
  ov_dummy_idx <- c(ov_y_dummy_ov_idx, ov_x_dummy_ov_idx)
  lv_dummy_idx <- c(ov_y_dummy_lv_idx, ov_x_dummy_lv_idx)
  num_idx <- lavmodel@num.idx[[group]]

  # fix Delta's...
  mm_in_group <- 1:lavmodel@nmat[group] + cumsum(c(0, lavmodel@nmat))[group]
  m_list <- g_list[mm_in_group]

  dd <- list()
  nvar <- lavmodel@nvar
  nfac <- ncol(m_list$lambda) - length(lv_dummy_idx)

  # dd$theta
  theta_idx <- lav_mat_diagh_idx(nvar)
  dd$theta <- delta_theta[theta_idx, , drop = FALSE]
  if (length(ov_dummy_idx) > 0L) {
    psi_idx <- lav_mat_diagh_idx(ncol(m_list$psi))[lv_dummy_idx]
    dd$theta[ov_dummy_idx, ] <- delta_psi[psi_idx, , drop = FALSE]
  }
  # num only? FIXME or just all of them?
  dd$theta <- dd$theta[num_idx, , drop = FALSE]

  # dd$nu
  dd$nu <- delta_nu
  if (length(ov_dummy_idx) > 0L) {
    dd$nu[ov_dummy_idx, ] <- delta_alpha[lv_dummy_idx, ]
  }
  dd$nu <- dd$nu[num_idx, , drop = FALSE] # needed?

  # dd$lambda
  nr <- nvar
  nc <- nfac
  lambda_idx <- nr * ((1:nc) - 1L) + rep(1:nvar, each = nc)
  dd$lambda <- delta_lambda[lambda_idx, , drop = FALSE]
  if (length(ov_dummy_idx) > 0L) {
    nr <- nrow(m_list$beta)
    nc <- nfac # only the first 1:nfac columns
    # beta_idx <- rep(nr*((1:nc) - 1L), each=length(lv_dummy_idx)) +
    #                                 rep(lv_dummy_idx, times=nc) ## FIXME
    beta_idx <- rep(nr * ((1:nc) - 1L),
           times = length(lv_dummy_idx)) + rep(lv_dummy_idx, each = nc)

    # l_idx <- inr*((1:nc) - 1L) + rep(ov_dummy_idx, each=nc) ## FIXME
    # l_idx <- rep(nr*((1:nc) - 1L), each=length(ov_dummy_idx)) +
    #                                rep(ov_dummy_idx, times=nc)
    l_idx <- rep(nr * ((1:nc) - 1L),
                times = length(ov_dummy_idx)) + rep(ov_dummy_idx, each = nc)
    dd$lambda[match(l_idx, lambda_idx), ] <-
                delta_beta[beta_idx, , drop = FALSE]
  }

  # dd$KAPPA
  dd$kappa <- delta_gamma
  if (length(ov_dummy_idx) > 0L) {
    nr <- nrow(m_list$gamma)
    nc <- ncol(m_list$gamma)
    kappa_idx <- nr * ((1:nc) - 1L) + rep(lv_dummy_idx, each = nc)
    dd$kappa <- dd$kappa[kappa_idx, , drop = FALSE]
  }

  # dd$GAMMA
  if (!is.null(m_list$gamma)) {
    nr <- nrow(m_list$gamma)
    nc <- ncol(m_list$gamma)
    lv_idx <- 1:nfac
    # MUST BE ROWWISE!
    gamma_idx <- rep(nr * ((1:nc) - 1L),
                 times = length(lv_idx)) + rep(lv_idx, each = nc)
    dd$gamma <- delta_gamma[gamma_idx, , drop = FALSE]
  }

  # dd$BETA
  if (!is.null(m_list$beta)) {
    nr <- nc <- nrow(m_list$beta)
    lv_idx <- 1:nfac
    # MUST BE ROWWISE!
    beta_idx <- rep(nr * ((1:nfac) - 1L),
                times = nfac) + rep(lv_idx, each = nfac)
    dd$beta <- delta_beta[beta_idx, , drop = FALSE]
  }

  ## dd$psi
  dd$psi <- delta_psi
  if (length(lv_dummy_idx) > 0L) {
    nr <- nc <- nrow(m_list$psi)
    lv_idx <- 1:nfac
    # MUST BE ROWWISE!
    psi_idx <- rep(nr * ((1:nfac) - 1L),
               times = nfac) + rep(lv_idx, each = nfac)

    dd$psi <- dd$psi[psi_idx, , drop = FALSE]
  }

  ## dd$tau
  if (!is.null(m_list$tau)) {
    dd$tau <- delta_tau
  }

  dd
}

# the entries of the lavaan and lavaanList objects have changed over time
# this function will check if the lavaan/lavaanList object is up to date, and
# adapt, if not.
#
# this may be useful if an older (say 0.5) lavaan object was saved, and
# passed to a function like lavPredict() in, say, lavaan 0.6-21.

# notes:
# - pre<0.5 lavaan objects are no longer supported
# - @Fit slot is ignored (as not used anymore)

# YR 16 Oct 2025 + LDW 22 Oct 2025

lav_object_check_version <- function(object = NULL) {

  is_lavaan_object <- inherits(object, "lavaan")
  if (!is_lavaan_object) {
    # check if lavaanList object, if not return input object
    #if (!inherits(object, "lavaanList")) return(object)
    return(object)
  }

  # flag: check or not?
  check_not_needed_flag <- TRUE

  # do we have a version slot?
  if (.hasSlot(object, "version")) {
    has_version_flag <- TRUE
    lavobject_version <- object@version[1] # lavaan.mi has two
    lavaanpkg_version <- read.dcf(
      file = system.file("DESCRIPTION", package = "lavaan"),
      fields = "Version"
    )[1]
    if (lavobject_version != lavaanpkg_version) {
      check_not_needed_flag <- FALSE
    }
  } else {
    # <0.6
    has_version_flag <- FALSE
    check_not_needed_flag <- FALSE
  }

  # check needed?
  if (check_not_needed_flag) {
    return(object)
  }

  # ok, we have potentially an older (saved) lavaan or lavaanList object
  # check needed slots, and if missing, add them
  suppressWarnings(lavobject <- object)
  ngroups <- lav_pt_ngroups(lavobject@ParTable)
  nblocks <- lav_pt_nblocks(lavobject@ParTable)
  nlevels <- lav_pt_nlevels(lavobject@ParTable)

  if (!has_version_flag) { # pre 0.6 object!
    # 0.5-10 (25 Oct 2012)
    if (!.hasSlot(lavobject@Data, "group")) {
      lavobject@Data@group <- character(0L)
    }

    # 0.5-11 (19 dec 2012)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "bifreq")) {
        lavobject@SampleStats@bifreq <- vector("list", length = ngroups)
      }
      if (!.hasSlot(lavobject, "Cache")) {
        lavobject@Cache <- list()
      }
    }

    # 0.5-12 (8 March 2013)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "ridge")) {
        lavobject@SampleStats@ridge <- 0
      }
    }

    # 0.5-14 (21 July 2013)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Model, "ov.x.dummy.ov.idx")) {
        lavobject@Model@ov.x.dummy.ov.idx <- vector("list", length = nblocks)
        lavobject@Model@ov.x.dummy.lv.idx <- vector("list", length = nblocks)
        lavobject@Model@ov.y.dummy.ov.idx <- vector("list", length = nblocks)
        lavobject@Model@ov.y.dummy.lv.idx <- vector("list", length = nblocks)
      }
      if (!.hasSlot(lavobject@SampleStats, "mean.x")) {
        lavobject@SampleStats@mean.x <- vector("list", length = ngroups)
        for (g in seq_len(ngroups)) {
          if (!is.null(lavobject@SampleStats@x.idx[[g]])) {
            lavobject@SampleStats@mean.x[[g]] <-
              lavobject@SampleStats@mean[[g]][lavobject@SampleStats@x.idx[[g]]]
          }
        }
      }
      if (!.hasSlot(lavobject, "pta")) {
        lavobject@pta <- list()
      }
    }

    # 0.5-15 (15 Nov 2013)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "Rp")) {
        lavobject@Data@Rp <- vector("list", length = ngroups)
      }
    }

    # 0.5-16 (7 March 2014)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "group.w")) {
        lavobject@SampleStats@group.w <-  vector("list", length = ngroups)
        for (g in seq_len(ngroups)) {
          lavobject@SampleStats@group.w[[g]] <-
            lavobject@SampleStats@nobs[[g]] / lavobject@SampleStats@ntotal
        }
      }
      if (!.hasSlot(lavobject@Model, "group.w.free")) {
        lavobject@Model@group.w.free <- FALSE
      }
      if (!.hasSlot(lavobject@Model, "parameterization")) {
        lavobject@Model@parameterization <- "delta"
      }
      if (!.hasSlot(lavobject@Model, "link")) {
        lavobject@Model@link <- "default"
      }
    }

    # 0.5-17 (30 Sept 2014)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "WLS.VD")) {
        lavobject@SampleStats@WLS.VD <- vector("list", length = ngroups)
      }
    }

    # 0.5-18 (18 Nov 2014)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Model, "eq.constraints.k0")) {
        lavobject@Model@eq.constraints.k0 <- numeric(0L)
      }
      if (!.hasSlot(lavobject@Model, "ceq.linear.idx")) {
        lavobject@Model@ceq.linear.idx <- integer(0L)
        lavobject@Model@ceq.nonlinear.idx <- integer(0L)
        lavobject@Model@cin.linear.idx <- integer(0L)
        lavobject@Model@cin.nonlinear.idx <- integer(0L)
      }
    }

    # 0.5-18 (13 Jan 2015)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Model, "ceq.JAC")) {
        lavobject@Model@ceq.JAC <- matrix(0, nrow = 0L,
                                            ncol = lavobject@Model@nx.free)
        lavobject@Model@ceq.rhs <- numeric(0L)
        lavobject@Model@cin.JAC <- matrix(0, nrow = 0L,
                                            ncol = lavobject@Model@nx.free)
        lavobject@Model@cin.rhs <- numeric(0L)
      }
    }

    # 0.5-19 (30 Jul 2015)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject, "boot")) {
        # construct partial optim list
        optim_list <- list(x = lavobject@Fit@x,
                           dx = numeric(0L),
                           npar = lavobject@Fit@npar,
                           iterations = lavobject@Fit@iterations,
                           converged = lavobject@Fit@converged,
                           warn.txt = "",
                           parscale = rep(1, lavobject@Fit@npar),
                           fx = lavobject@Fit@fx,
                           fx.group = lavobject@Fit@fx.group,
                           logl.group = lavobject@Fit@logl.group,
                           control = lavobject@Fit@control)
        lavobject@boot <- vector("list", 0L)
        lavobject@optim <- optim_list
        lavobject@implied <- lav_model_implied(lavobject@Model)
        lavobject@vcov <- list(se = lavobject@Options$se[1],
                               information = lavobject@Options$information[1],
                               vcov = matrix(0, 0, 0)) # for now
        lavobject@test <- lavTest(lavobject)
        lavobject@external <- vector("list", 0L)
      }
    }

    # 0.5-19: est/se move to @ParTable
    if (!is.null(lavobject@Fit@est) && is.null(lavobject@ParTable$est)) {
      lavobject@ParTable$est <- lavobject@Fit@est
    }
    if (!is.null(lavobject@Fit@se) && is.null(lavobject@ParTable$se)) {
      lavobject@ParTable$se <- lavobject@Fit@se
    }


    # 0.5-21 (16 Dec 2015)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Model, "conditional.x")) {
        lavobject@Model@conditional.x <- FALSE
      }
    }

    # 0.5-21 (5 Jan 2016)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "x.idx")) {
        lavobject@SampleStats@x.idx <- rep(list(integer(0L)), ngroups)
      }
    }

    # 0.5-21 (8 Jan 2016)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "res.cov")) {
        lavobject@SampleStats@res.cov <- vector("list", ngroups)
        lavobject@SampleStats@res.var <- vector("list", ngroups)
        lavobject@SampleStats@res.th <- vector("list", ngroups)
        lavobject@SampleStats@res.th.nox <- vector("list", ngroups)
        lavobject@SampleStats@res.slopes <- vector("list", ngroups)
        lavobject@SampleStats@res.int <- vector("list", ngroups)
        lavobject@SampleStats@res.icov <- vector("list", ngroups)
        lavobject@SampleStats@res.icov.log.det <- vector("list", ngroups)
      }
    }

    # 0.5-21 (28 Mar 2016)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "NACOV.user")) {
        lavobject@SampleStats@NACOV.user <- FALSE
      }
    }

    #### 0.5-21, 3 Jul 2016, Class lavaanList is added ####

    # 0.5-23 (25 Jan 2017)
    if (!.hasSlot(lavobject@Model, "estimator")) {
      lavobject@Model@estimator <- lavobject@Options$estimator
    }

    # 0.5-23 (30 Jan 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "cluster")) {
        lavobject@Data@cluster <- character(0L)
        lavobject@Data@ordered <- character(0L)
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "cluster")) {
          lavobject@DataList[[j]]@cluster <- character(0L)
          lavobject@DataList[[j]]@ordered <- character(0L)
        }
      }
    }

    # 0.5-23 (7 Feb 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "zero.cell.tables")) {
        lavobject@SampleStats@zero.cell.tables <- vector("list", ngroups)
      }
    } else {
      for (j in seq_along(lavobject@SampleStatsList)) {
        if (!.hasSlot(lavobject@SampleStatsList[[j]], "zero.cell.tables")) {
          lavobject@SampleStatsList[[j]]@zero.cell.tables <-
                                                   vector("list", ngroups)
        }
      }
    }

    # 0.5-23 (21 Feb 2017)
    if (!.hasSlot(lavobject@Model, "nblocks")) {
      lavobject@Model@nblocks <- nblocks
    }
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "level.label")) {
        lavobject@Data@level.label <- as.character(seq.int(nlevels))
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "level.label")) {
          lavobject@DataList[[j]]@level.label <- as.character(seq.int(nlevels))
        }
      }
    }
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "block.label")) {
        if (nlevels <= 1L) {
          if (ngroups <= 1L) {
            lavobject@Data@block.label <- character(0L)
          } else {
            lavobject@Data@block.label <- lavobject@Data@group.label
          }
        } else {
          if (ngroups <= 1L) {
            lavobject@Data@block.label <- lavobject@Data@level.label
          } else {
            lavobject@Data@block.label <-
              paste(rep(lavobject@Data@group.label,
                        each = length(lavobject@Data@level.label)),
                    rep(lavobject@Data@level.label,
                        times = length(lavobject@Data@group.label)),
                    sep = "."
              )
          }
        }
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "block.label")) {
          if (nlevels <= 1L) {
            if (ngroups <= 1L) {
              lavobject@DataList[[j]]@block.label <- character(0L)
            } else {
              lavobject@DataList[[j]]@block.label <-
                lavobject@DataList[[j]]@group.label
            }
          } else {
            if (ngroups <= 1L) {
              lavobject@DataList[[j]]@block.label <-
                lavobject@DataList[[j]]@level.label
            } else {
              lavobject@DataList[[j]]@block.label <-
                paste(rep(lavobject@DataList[[j]]@group.label,
                          each = length(lavobject@DataList[[j]]@level.label)),
                      rep(lavobject@DataList[[j]]@level.label,
                          times = length(lavobject@DataList[[j]]@group.label)),
                      sep = "."
                )
            }
          }
        }
      }
    }

    # 0.5-23 (24 Feb 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "nlevels")) {
        lavobject@Data@nlevels <- nlevels
        lavobject@Data@Lp <- vector("list", ngroups)
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "nlevels")) {
          lavobject@DataList[[j]]@nlevels <- nlevels
          lavobject@DataList[[j]]@Lp <- vector("list", ngroups)
        }
      }
    }
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@SampleStats, "YLp")) {
        lavobject@SampleStats@YLp <- vector("list", ngroups)
      }
    } else {
      for (j in seq_along(lavobject@SampleStatsList)) {
        if (!.hasSlot(lavobject@SampleStatsList[[j]], "YLp")) {
          lavobject@SampleStatsList[[j]]@YLp <- vector("list", ngroups)
        }
      }
    }

    # 0.6-1 (8 Mar 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject, "h1")) {
        lavobject@h1 <- lav_h1_implied_logl(lavdata = lavobject@Data,
          lavsamplestats = lavobject@SampleStats,
          lavpartable = lavobject@ParTable,
          lavoptions = lavobject@Options)
        lavobject@baseline <- lav_step15_baseline(
          lavoptions = lavobject@Options,
          lavsamplestats = lavobject@SampleStats,
          lavdata = lavobject@Data,
          lavcache = lavobject@Cache,
          lavpartable = lavobject@ParTable
        )
      }
      if (!.hasSlot(lavobject@Data, "ov.names.l")) {
        lavobject@Data@ov.names.l <- vector("list", 0L)
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "ov.names.l")) {
          lavobject@DataList[[j]]@ov.names.l <- vector("list", 0L)
        }
      }
    }

    # 0.6-1 (10 Mar 2017)
    if (!.hasSlot(lavobject@Model, "multilevel")) {
      lavobject@Model@multilevel <- FALSE
    }

    # 0.6-1 (19 Mar 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject, "loglik")) {
        lavobject@Model@ceq.simple.only <- FALSE
        lavobject@Model@cin.simple.only <- FALSE
        lavobject@loglik <- lav_model_loglik(
          lavdata = lavobject@Data,
          lavsamplestats = lavobject@SampleStats,
          lavimplied = lavobject@implied,
          lavmodel = lavobject@Model,
          lavoptions = lavobject@Options
        )
      }
    }

    # 0.6-1 (1 Oct 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "weights")) {
        lavobject@Data@weights <- vector("list", ngroups)
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "weights")) {
          lavobject@DataList[[j]]@weights <- vector("list", ngroups)
        }
      }
    }

    # 0.6-1 (3 Oct 2017)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject@Data, "sampling.weights")) {
        lavobject@Data@sampling.weights <- character(0L)
      }
    } else {
      for (j in seq_along(lavobject@DataList)) {
        if (!.hasSlot(lavobject@DataList[[j]], "sampling.weights")) {
          lavobject@DataList[[j]]@sampling.weights <- character(0L)
        }
      }
    }

    # 0.6-1 (2 May 2018)
    if (is_lavaan_object) {
      if (!.hasSlot(lavobject, "version")) lavobject@version <- "PRE 0.6"
    }
  } # no-version-flag (pre 0.6)

  ### from here on, we assume that the object is generated by lavaan 0.6-1 or
  ### higher

  # check if @test list is named
  if (is.null(names(lavobject@test))) {
    names(lavobject@test) <- sapply(lavobject@test, "[[", "test")
  }

  # 0.6-2 (12 Jun 2018)
  if (!.hasSlot(lavobject@Model, "x.free.var.idx")) {
    lavobject@Model@x.free.var.idx <- integer(0L)
  }

  # 0.6-3 (17 Sep 2018)
  if (!is_lavaan_object) {
    if (!.hasSlot(lavobject, "h1List")) {
      lavobject@h1List <- vector("list", 0L)
      lavobject@loglikList <- vector("list", 0L)
    }
  }

  # 0.6-4 (30 Mar 2019)
  if (!.hasSlot(lavobject@Model, "ov.efa.idx")) {
    lavobject@Model@ov.efa.idx <- vector("list", nblocks)
    lavobject@Model@lv.efa.idx <- vector("list", nblocks)
  }

  # 0.6-4 (11 Apr 2019)
  if (!.hasSlot(lavobject@Model, "nefa")) {
    lavobject@Model@nefa <- 0L
  }

  # 0.6-4 (24 Apr 2019)
  if (!.hasSlot(lavobject@Model, "H")) {
    lavobject@Model@H <- vector("list", 0L)
    lavobject@Model@lv.order <- vector("list", 0L)
  }

  # 0.6-4 (26 Apr 2019)
  if (!.hasSlot(lavobject@Model, "ceq.efa.JAC")) {
    lavobject@Model@ceq.efa.JAC <- matrix(0, nrow = 0L, ncol = 0L)
  }

  # 0.6-5 (7 Jul 2019)
  if (!is_lavaan_object) {
    if (!.hasSlot(lavobject, "baselineList")) {
      lavobject@baselineList <- vector("list", 0L)
    }
  }

  # 0.6-8 (29 Sep 2020)
  if (!.hasSlot(lavobject@Model, "rv.ov")) {
    lavobject@Model@rv.ov <- vector("list", 0L)
    lavobject@Model@rv.lv <- vector("list", 0L)
  }

  # 0.6-8 (18 Dec 2020)
  if (!.hasSlot(lavobject@Model, "estimator.args")) {
    lavobject@Model@estimator.args <- vector("list", 0L)
  }

  # 0.6-9 (15 Mar 2021)
  if (!.hasSlot(lavobject@Model, "modprop")) {
    lavobject@Model@modprop = lav_model_properties(
      glist = lavobject@Model@GLIST,
      lavpartable = lavobject@ParTable,
      nmat = lavobject@Model@nmat,
      m_free_idx = lavobject@Model@m.free.idx
    )
  }

  # 0.6-9 (22 Jun 2021)
  if (is_lavaan_object) {
    if (!.hasSlot(lavobject, "internal")) {
      lavobject@internal <- vector("list", 0L)
    }
  } else {
    if (!.hasSlot(lavobject, "internalList")) {
      lavobject@internalList <- vector("list", 0L)
    }
  }

  # 0.6-11 (28 Feb 2022)
  if (!.hasSlot(lavobject@Model, "nx.unco")) {
    # is not available, unco == free
    lavobject@Model@nx.unco <- lavobject@Model@nx.free
    lavobject@Model@x.unco.idx <- lavobject@Model@x.free.idx
    lavobject@Model@ceq.simple.only <- FALSE
    lavobject@Model@ceq.simple.K <- matrix(0, nrow = 0L, ncol = 0L)
  }

  # 0.6-13 (25 Jul 2022)
  if (!.hasSlot(lavobject@Model, "correlation")) {
    lavobject@Model@correlation <- FALSE
  }

  # 0.6-18 (25 Apr 2024)
  if (!is_lavaan_object) {
    if (!.hasSlot(lavobject, "version")) {
      lavobject@version <- "PRE 0.6.18"
    }
  }

  # 0.6-19 (27 Sep 2024)
  if (!.hasSlot(lavobject@Model, "cin.simple.only")) {
    lavobject@Model@cin.simple.only <- FALSE
  }

  # 0.6-20 (24 Jan 2025)
  if (!.hasSlot(lavobject@Model, "composites")) {
    lavobject@Model@composites <- any(lavobject@ParTable$op == "<~")
  }

  # check missing options
  object_options <- lavobject@Options
  all_options <- lavOptions()
  missing_idx <- which(!names(all_options) %in% names(object_options))
  new_options <- c(object_options, all_options[missing_idx])
  # fill in some "default" values
  if (new_options$estimator.orig == "default") {
    new_options$estimator.orig <- new_options$estimator
  }

  # 0.6-21
  new_options$gamma.vcov.mplus <- new_options$mimic == "Mplus"
  new_options$gamma.wls.mplus <- new_options$mimic == "Mplus"
  new_options$gls.v11.mplus <- new_options$mimic == "Mplus"
  new_options$rmsea.scaled.mplus <- new_options$mimic == "Mplus"
  new_options$cinformation.expected.mplus <- new_options$mimic == "Mplus"
  new_options$h1.information.meat <- "structured"
  new_options$mega.h1.information <- "unstructured"

  # 0.7-1
  if (is.null(new_options$rotation.args$mg_agreement)) {
    new_options$rotation.args$mg_agreement <- FALSE
    new_options$rotation.args$mg_agreement_weight <- 0.5
    new_options$rotation.args$mg_agreement_method <- "pairwise"
    new_options$rotation.args$mg_agreement_crit <- "procrustes"

  }
  if (is.null(new_options$rotation.args$mg_agreement_weight)) {
    new_options$rotation.args$mg_agreement_weight <- 0.5
  }

  lavobject@Options <- new_options

  lavobject
}

lav_step02_options <- function(slot_options = NULL,
                                      slot_data = NULL,
                                      flat_model = NULL,
                                      ordered = NULL,
                                      ordered_orig = NULL,
                                      sample_cov = NULL,
                                      sample_mean = NULL,
                                      sample_th = NULL,
                                      sample_nobs = NULL,
                                      ov_names_l = NULL,
                                      sampling_weights = NULL,
                                      constraints = NULL,
                                      group = NULL,
                                      ov_names_x = NULL,
                                      ov_names_y = NULL,
                                      dotdotdot = NULL,
                                      cluster = NULL,
                                      data = NULL) {
  # # # # # # # # # # # #
  # #  2. lavoptions  # #
  # # # # # # # # # # # #

  # if slot_options not NULL
  #   copy to lavoptions and modify categorical/clustered/multilevel
  #     inserting a "." in the first position
  #   if necessary, overwrite with values in dotdotdot and issue a warning
  #   check if all names in dotdotdot are possible options, if not *** error ***
  #   create complete option list (lav_options_default) and substitute values
  #     given in dotdotdot
  #   if data, slot_data and sample_cov NULL: opt$bounds = FALSE
  #   if slot_data$data.type != "full" or (slot_data and data = NULL):
  #     opt$missing = "listwise"
  #   set categorical mode ON if
  #     - an operator "|" (threshold) was used
  #     - data not NULL and one or more elements in ordered parameter
  #     - sample_th provided
  #     - at least one of the non-exogenous observed variables is "ordered"
  #       (ordered factor in R)
  #   if opt$estimator == "catml": set categorical mode OFF
  #       TODO: estimator = "CATML" isn't mentioned in lavOptions / estimator
  #             help text !?
  #   if cluster not NULL, set opt$.clustered TRUE and  *** error *** if
  #     categorical mode is ON
  #   opt$.multilevel = (length(ov.names.l) > 0L &&
  #                      length(ov.names.l[[1]]) > 1L)
  #   if sampling.weights not NULL en categorical mode OFF and opt$estimator
  #     in ("default", "ML", "PML")
  #     set opt$estimator to "MLR"
  #   if constraints present and estimator == "ML", set opt$information to
  #     c("observed", "observed")
  #   if there is an operator "~1" in flat_model and sample_mean not NULL,
  #     set opt$meanstructure TRUE
  #   if there are no exogenous variables but conditional.x explicitly
  #     requested: ** warning **
  #   if there are no exogenous variables set opt$conditional.x FALSE
  #   if there are no exogenous variables and fixed.x not explicitly requested,
  #     set opt$fixed.x to FALSE
  #   if allow.empty.cell and estimator not Bayes, issue a warning

  if (!is.null(slot_options)) {
    lavoptions <- slot_options

    # backwards compatibility
    if (!is.null(lavoptions$categorical)) {
      lavoptions$.categorical <- lavoptions$categorical
      lavoptions$categorical <- NULL
    }
    if (!is.null(lavoptions$clustered)) {
      lavoptions$.clustered <- lavoptions$clustered
      lavoptions$clustered <- NULL
    }
    if (!is.null(lavoptions$multilevel)) {
      lavoptions$.multilevel <- lavoptions$multilevel
      lavoptions$multilevel <- NULL
    }

    # but what if other 'options' are given anyway (eg 'start = ')?
    # give a warning!
    if (length(dotdotdot) > 0L) {
      dot_names <- names(dotdotdot)
      op_idx <- which(dot_names %in% names(slot_options))
      lav_msg_warn(gettext(
        "the following argument(s) override(s) the options in slot_options:"),
        paste(dot_names[op_idx], collapse = " ")
      )
      lavoptions[dot_names[op_idx]] <- dotdotdot[op_idx]
    }
  } else {
    if (lav_verbose()) {
      cat("lavoptions         ...")
    }

    # load default options
    opt <- lav_options_default()

    # catch unknown options
    ok_names <- names(opt)
    dot_names <- names(dotdotdot)
    wrong_idx <- which(!dot_names %in% ok_names)
    if (length(wrong_idx) > 0L) {
      # stop or warning?? stop for now (there could be more)
      lav_msg_stop(ngettext(length(wrong_idx),
        "unknown argument:", "unknown arguments:"),
        lav_msg_view(dot_names[wrong_idx], "none", FALSE)
      )
    }

    # modifyList
    if (is.list(dotdotdot$rotation.args)) {
      dotdotdot$rotation.args <- lav_snake_case(dotdotdot$rotation.args)
    }
    if (is.list(dotdotdot$em.args)) {
      dotdotdot$em.args <- lav_snake_case(dotdotdot$em.args)
    }
    if (is.list(dotdotdot$em.h1.args)) {
      dotdotdot$em.h1.args <- lav_snake_case(dotdotdot$em.h1.args)
    }
    opt <- modifyList(opt, dotdotdot)

    # extract estimator
    if (is.list(opt$estimator)) {
      estimator <- opt$estimator$estimator
    } else {
      estimator <- opt$estimator
    }

    # no data?
    if (is.null(slot_data) && is.null(data) && is.null(sample_cov)) {
      opt$bounds <- FALSE
    }

    # only sample moments?
    if (!is.null(slot_data) && !slot_data@data.type == "full") {
      opt$missing <- "listwise"
    } else if (is.null(slot_data) && is.null(data)) {
      opt$missing <- "listwise"
    }

    # categorical mode?
    opt$.categorical <- FALSE
    if (any(flat_model$op == "|")) {
      opt$.categorical <- TRUE
    } else if (!is.null(data) && length(ordered) > 0L) {
      opt$.categorical <- TRUE
    } else if (!is.null(sample_th)) {
      opt$.categorical <- TRUE
    } else if (is.data.frame(data)) {
      # first check if we can find ov.names.y in Data
      tmp_ov_names_y <- unique(unlist(ov_names_y))
      # remove possible interaction terms involving an y term
      int_idx <- grep(":", tmp_ov_names_y, fixed = TRUE)
      if (length(int_idx) > 0L) {
        tmp_ov_names_y <- tmp_ov_names_y[-int_idx]
      }
      idx_missing <- which(!(tmp_ov_names_y %in% names(data)))
      if (length(idx_missing)) {
        lav_msg_stop(
          gettext("missing observed variables in dataset:"),
          paste(tmp_ov_names_y[idx_missing], collapse = " ")
        )
      }
      if (any(sapply(data[, tmp_ov_names_y], inherits, "ordered"))) {
        opt$.categorical <- TRUE
      }
    }
    if (tolower(estimator) == "catml") {
      opt$.categorical <- FALSE
    }

    # IV estimator: default to two-stage missing-data handling when the data
    # contain missing values among the modeled variables and the user did not
    # request a 'missing' option. (opt$missing is still "default" here; it is
    # resolved to "listwise" later in lav_options_set().) This must happen
    # before lav_options_set() so the two.stage option cascade (meanstructure,
    # information, ...) and lavData (no listwise deletion) are set up correctly.
    if (toupper(estimator) == "IV" && identical(opt$missing, "default") &&
        !opt$.categorical && is.data.frame(data)) {
      tmp_ov <- unique(c(unlist(ov_names_y), unlist(ov_names_x)))
      tmp_ov <- tmp_ov[tmp_ov %in% names(data)]
      if (length(tmp_ov) > 0L && anyNA(data[, tmp_ov, drop = FALSE])) {
        opt$missing <- "two.stage"
      }
    }

    # check for WLSMV estimator with NULL ordered= argument (new in 0.7-1)
    # to avoid unintentional use of WLSMV
    # we do this here, because we need the value of opt$.categorical
    if (!opt$.categorical && is.null(ordered_orig) && !estimator == "default") {
      estimator <- toupper(estimator)
      if (estimator %in% c("WLSM", "WLSMV", "DWLS")) {
        lav_msg_stop(gettextf(
        "Estimator %s is typically used with categorical data only. To use it
         with continuous data, please explicitly set the ordered= argument
         to FALSE.",
        dQuote(lav_options_estimatorgroup(estimator))))
      }
    }

    # clustered?
    if (length(cluster) > 0L) {
      opt$.clustered <- TRUE
      # categorical + clustered (single-level): cluster-robust SEs/test via
      # the cluster-aggregated Gamma in muthen1984() (issue #254). Multilevel
      # (nlevels > 1) categorical is still unsupported.
      multilevel <- length(ov_names_l) > 0L && length(ov_names_l[[1]]) > 1L
      if (opt$.categorical && multilevel &&
        toupper(estimator) != "PML" &&
        !lav_options_estimatorgroup(tolower(estimator)) %in%
          c("WLS", "DWLS", "ULS")) {
        lav_msg_stop(gettext(
          "multilevel + categorical data is only supported for the
          (D)WLS/ULS estimators (e.g., WLSMV) (for now)."))
      }
    } else {
      opt$.clustered <- FALSE
    }

    # multilevel?
    if (length(ov_names_l) > 0L && length(ov_names_l[[1]]) > 1L) {
      opt$.multilevel <- TRUE
    } else {
      opt$.multilevel <- FALSE
    }

    # random slopes (rv() modifier)?
    opt$.rv <- !is.null(flat_model$rv) && any(nchar(flat_model$rv) > 0L)

    # sampling weights? (internal flag; used when resolving estimator
    # defaults, e.g. the se/test defaults for continuous ULS)
    opt$.sampling.weights <- !is.null(sampling_weights)

    # sampling weights? force MLR
    # HJ 18/10/23: Except for PML
    # YR 17/06/26: only force se/test if the user did not explicitly
    #              request a value for them; if the user insists on a
    #              different se=/test=, we honor that choice
    if (!is.null(sampling_weights) && !opt$.categorical &&
      toupper(estimator) %in% c("DEFAULT", "ML", "PML")) {
      if (opt$se != "none" && is.null(dotdotdot$se)) {
        opt$se <- "robust.huber.white"
      }
      if (opt$se != "none" && is.null(dotdotdot$test)) {
        opt$test <- "yuan.bentler.mplus"
      }
    }

    # constraints
    if (any(nchar(constraints) > 0L) && toupper(estimator) %in% c("ML")) {
      opt$information <- c("observed", "observed")
    }

    # meanstructure
    if (any(flat_model$op == "~1") || !is.null(sample_mean)) {
      opt$meanstructure <- TRUE
    }
    if (!is.null(group) && is.null(dotdotdot$meanstructure)) {
      opt$meanstructure <- TRUE
    }

    # conditional.x
    if ((is.list(ov_names_x) &&
      sum(sapply(ov_names_x, FUN = length)) == 0L) ||
      (is.character(ov_names_x) && length(ov_names_x) == 0L)) {
      # if explicitly set to TRUE, give warning
      if (is.logical(dotdotdot$conditional.x) && dotdotdot$conditional.x) {
        lav_msg_warn(gettext(
          "no exogenous covariates; conditional.x will be set to FALSE"))
      }
      opt$conditional.x <- FALSE
    }

    # fixed.x
    if ((is.list(ov_names_x) &&
      sum(sapply(ov_names_x, FUN = length)) == 0L) ||
      (is.character(ov_names_x) && length(ov_names_x) == 0L)) {
      # if explicitly set to TRUE, give warning
      if (is.logical(dotdotdot$fixed.x) && dotdotdot$fixed.x) {
        # ok, we respect this: keep fixed.x = TRUE
      } else {
        opt$fixed.x <- FALSE
      }
    }

    # allow.empty.cell
    if (opt$allow.empty.cell && opt$do.fit && toupper(estimator) != "BAYES") {
      lav_msg_warn(gettext(
        "allow.empty.cell is not intended to salvage estimation of this model,",
        " see ?lavOptions"))
    }

    # fill in remaining "default" values
    lavoptions <- lav_options_set(opt)

    # correlation = TRUE with observed product (interaction) terms: the
    # product of standardized variables does not have unit variance, so the
    # product terms must keep a free variance (github issue #427). We turn
    # the full correlation structure into a partial one that standardizes
    # all observed variables except the product terms. (If the user listed
    # the variables explicitly, we leave that choice alone.)
    if (isTRUE(lavoptions$correlation) &&
        length(lavoptions$.correlation.ov) == 0L) {
      ov_int <- unique(unlist(lav_pt_vnames(flat_model, "ov.interaction")))
      if (length(ov_int) > 0L) {
        ov_all <- unique(unlist(lav_pt_vnames(flat_model, "ov")))
        lavoptions$.correlation.ov <- setdiff(ov_all, ov_int)
      }
    }

    # store check.sigma.pd in lavaan_cache_env
    assign("opt_check_sigma_pd", opt$check.sigma.pd, lavaan_cache_env)

    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  lavoptions
}

# shared building blocks for the test-statistic machinery (lav_test*.R)
#
# YR 11 July 2026: initial version, collecting code that used to be
#                  duplicated across lav_test.R, lav_test_satorra_bentler.R,
#                  lav_test_yuan_bentler.R, lav_test_browne.R and
#                  lav_test_diff.R

# degrees of freedom of the user model, adding back the rank of the
# equality-constraint jacobian (inequality constraints, active or not,
# are ignored)
lav_test_df <- function(lavpartable = NULL, lavmodel = NULL) {
  df <- lav_pt_df(lavpartable)

  if (!lavmodel@cin.simple.only && nrow(lavmodel@con.jac) > 0L) {
    ceq_idx <- attr(lavmodel@con.jac, "ceq.idx")
    if (length(ceq_idx) > 0L) {
      neq <- qr(lavmodel@con.jac[ceq_idx, , drop = FALSE])$rank
      df <- df + neq
    }
  } else if (lavmodel@ceq.simple.only) {
    ndat <- lav_pt_ndat(lavpartable)
    npar <- max(lavpartable$free)
    df <- ndat - npar
  }

  df
}

# the meta-information attribute attached to the TEST list; used by
# lav_test_print() for independent printing
lav_test_info_attr <- function(lavdata = NULL, lavoptions = NULL) {
  list(
    ngroups = lavdata@ngroups, group.label = lavdata@group.label,
    information = lavoptions$information,
    h1.information = lavoptions$h1.information,
    observed.information = lavoptions$observed.information
  )
}

# do the test-specific information options (second element) match the ones
# used for the standard errors (first element)? if not, E.inv (and Delta,
# WLS.V) must be recomputed with the test-specific options; returns the
# lavoptions with the first elements overwritten by the second, and the
# recompute flag as an attribute
lav_test_change_options <- function(lavoptions = NULL) {
  if (length(lavoptions$information) == 1L &&
    length(lavoptions$h1.information) == 1L &&
    length(lavoptions$observed.information) == 1L) {
    recompute <- FALSE
  } else if (
    (lavoptions$information[1] == lavoptions$information[2]) &&
      (lavoptions$h1.information[1] == lavoptions$h1.information[2]) &&
      (lavoptions$information[2] == "expected" ||
        (lavoptions$observed.information[1] ==
          lavoptions$observed.information[2]))) {
    recompute <- FALSE
  } else {
    recompute <- TRUE
    # change information options
    lavoptions$information[1] <- lavoptions$information[2]
    lavoptions$h1.information[1] <- lavoptions$h1.information[2]
    lavoptions$observed.information[1] <- lavoptions$observed.information[2]
  }

  attr(lavoptions, "recompute") <- recompute
  lavoptions
}

# a 'failed' scaled-test entry: the standard test augmented with NA
# scaling quantities; used when the information matrix cannot be inverted
# (na_standard = TRUE) or when df <= 0 (na_standard = FALSE, no shift)
lav_test_scaled_na <- function(standard = NULL, test1 = NULL,
                               ngroups = 1L,
                               na_standard = TRUE, shift = TRUE) {
  if (na_standard) {
    standard$stat <- as.numeric(NA)
    standard$stat.group <- rep(as.numeric(NA), ngroups)
    standard$pvalue <- as.numeric(NA)
  }
  if (shift) {
    out <- c(standard,
      scaling.factor = as.numeric(NA),
      shift.parameter = as.numeric(NA),
      label = character(0)
    )
  } else {
    out <- c(standard,
      scaling.factor = as.numeric(NA),
      label = character(0)
    )
  }
  # to prevent lavTestLRT error when a robust test is detected for some
  # but not all models
  out$test <- test1

  list(standard = standard, failed = out)
}

# label for the scaled test statistics; the Mplus-flavored labels are used
# when information.expected.mplus is TRUE
lav_test_scaled_label <- function(test = "satorra.bentler",
                                  estimator = "ML",
                                  mplus_flag = FALSE) {
  if (test == "satorra.bentler") {
    plain <- "Satorra-Bentler correction"
    mplus <- c(
      ML = "Satorra-Bentler correction (Mplus variant)",
      DWLS = "Satorra-Bentler correction (WLSM)",
      ULS = "Satorra-Bentler correction (ULSM)"
    )
  } else if (test == "mean.var.adjusted") {
    plain <- "mean and variance adjusted correction"
    mplus <- c(
      ML = "mean and variance adjusted correction (MLMV)",
      DWLS = "mean and variance adjusted correction (WLSMV)",
      ULS = "mean and variance adjusted correction (ULSMV)"
    )
  } else if (test == "scaled.shifted") {
    plain <- "simple second-order correction"
    mplus <- c(
      ML = "simple second-order correction (MLMV)",
      DWLS = "simple second-order correction (WLSMV)",
      ULS = "simple second-order correction (ULSMV)"
    )
  } else {
    return(character(0L))
  }

  if (mplus_flag && estimator %in% names(mplus)) {
    unname(mplus[estimator])
  } else {
    plain
  }
}

# stack the per-group ingredients for the 'assembled' (explicit) UGamma
# computation: fg-weighted V blocks, Gamma_g / fg blocks (the
# Cov(sqrt(ntotal) s_g) stacking convention -- see the SCALING CONVENTIONS
# note in lav_samplestats_gamma.R), and the row-stacked Delta;
# a caller-supplied gamma_full (possibly with cross-group blocks) is used
# as-is instead of the block-diagonal assembly
lav_test_ug_stack <- function(m_gamma = NULL, delta = NULL, wls_v = NULL,
                              nobs = NULL, ntotal = NULL,
                              gamma_full = NULL) {
  ngroups <- length(wls_v)
  fg <- unlist(nobs) / ntotal

  v_g <- vector("list", ngroups)
  for (g in seq_len(ngroups)) {
    if (is.matrix(wls_v[[g]])) {
      v_g[[g]] <- fg[g] * wls_v[[g]]
    } else {
      v_g[[g]] <- fg[g] * diag(wls_v[[g]], nrow = length(wls_v[[g]]))
    }
  }

  if (is.null(gamma_full)) {
    gamma_f <- lav_gamma_rescale_ntotal(m_gamma,
      nobs = nobs, ntotal = ntotal
    )
    gamma_all <- lav_mat_bdiag(gamma_f)
  } else {
    gamma_all <- gamma_full
  }

  list(
    v_all = lav_mat_bdiag(v_g),
    gamma_all = gamma_all,
    delta_all = do.call("rbind", delta)
  )
}

# streamed computation of tr(U Gamma) and tr((U Gamma)^2), with
#
#   U = V - V Delta E.inv Delta' V
#
# assembled over groups in the Cov(sqrt(ntotal) s_g) stacking convention
# (fg-weighted V blocks, Gamma_g / fg blocks), WITHOUT ever forming a
# pstar x pstar product. With WD_g = V_g Delta_g, GWD_g = Gamma_g WD_g and
# B_g = WD_g' GWD_g (= Delta' V Gamma V Delta, npar x npar):
#
#   tr(U Gamma)     = sum_g [ tr(V_g Gamma_g) - fg_g tr(E.inv B_g) ]
#   tr((U Gamma)^2) = t1 - 2 t2 + t3, with
#      t1 = sum_g tr((V_g Gamma_g)^2)
#      t2 = tr(E.inv sum_g fg_g GWD_g' V_g GWD_g)
#      t3 = tr((E.inv C)^2),  C = sum_g fg_g B_g
#
# (the cross-group blocks of U Gamma enter tr((U Gamma)^2) only through
# E.inv, which is what the C accumulation captures)
#
# the V blocks may be given as vectors (diagonal weight matrices, e.g.
# DWLS/ULS) or full matrices; this matters in the categorical setting and
# whenever the number of variables grows, where pstar is large
#
# old_approach = TRUE reproduces the < 0.6-13 behavior where also
# tr((U Gamma)^2) was accumulated per group:
#   tr((U_g Gamma_g)^2) = t1_g - 2 fg_g tr(E.inv T2_g)
#                              + fg_g^2 tr((E.inv B_g)^2)
lav_test_ug_trace_stream <- function(m_gamma = NULL, delta = NULL,
                                     wls_v = NULL, e_inv = NULL,
                                     fg = NULL,
                                     satterthwaite = FALSE,
                                     old_approach = FALSE) {
  ngroups <- length(m_gamma)
  npar <- NCOL(delta[[1]])

  trace_group <- numeric(ngroups)
  trace2_group <- numeric(ngroups) # old_approach only
  t1 <- 0
  m_t2 <- m_c <- matrix(0, npar, npar)

  for (g in seq_len(ngroups)) {
    v <- wls_v[[g]]
    gamma_g <- m_gamma[[g]]
    delta_g <- delta[[g]]

    if (is.matrix(v)) {
      wd <- v %*% delta_g
      trv <- sum(v * gamma_g) # both symmetric
    } else {
      wd <- v * delta_g
      trv <- sum(v * diag(gamma_g))
    }
    gwd <- gamma_g %*% wd
    b_g <- crossprod(wd, gwd)
    trace_group[g] <- trv - fg[g] * sum(e_inv * b_g)

    if (satterthwaite) {
      if (is.matrix(v)) {
        vg <- v %*% gamma_g
        t1_g <- sum(vg * t(vg))
        t2_g <- crossprod(gwd, v %*% gwd)
      } else {
        t1_g <- drop(crossprod(v, (gamma_g * gamma_g) %*% v))
        t2_g <- crossprod(gwd, v * gwd)
      }
      if (old_approach) {
        eb <- e_inv %*% b_g
        trace2_group[g] <- t1_g - 2 * fg[g] * sum(e_inv * t2_g) +
          fg[g] * fg[g] * sum(eb * t(eb))
      } else {
        t1 <- t1 + t1_g
        m_t2 <- m_t2 + fg[g] * t2_g
        m_c <- m_c + fg[g] * b_g
      }
    }
  } # g

  trace_ugamma <- sum(trace_group)
  trace_ugamma2 <- as.numeric(NA)
  if (satterthwaite) {
    if (old_approach) {
      trace_ugamma2 <- sum(trace2_group)
    } else {
      m_ec <- e_inv %*% m_c
      trace_ugamma2 <- t1 - 2 * sum(e_inv * m_t2) + sum(m_ec * t(m_ec))
    }
  }

  list(
    trace.UGamma = trace_ugamma,
    trace.UGamma2 = trace_ugamma2,
    trace.UGamma.group = trace_group
  )
}

# which (non-scaled) test statistic shall we scale? (the scaled.test
# option; e.g. scaled.test = "browne.residual.nt.model" gives the
# RLS-based scaled tests)
lav_test_scaled_base <- function(test_1 = NULL, lavoptions = NULL,
                                 test = NULL) {
  unscaled_test <- test_1[[1]]
  if (lavoptions$scaled.test != "standard") {
    idx <- which(names(test_1) == lavoptions$scaled.test)
    if (length(idx) > 0L) {
      unscaled_test <- test_1[[idx[1]]]
    } else {
      lav_msg_warn(gettextf(
        "scaled.test [%1$s] not found among available (non scaled) tests:
        %2$s. Using standard test instead.",
        lavoptions$scaled.test, lav_msg_view(test)))
    }
  }
  unscaled_test
}

lav_step08_start <- function(slot_model = NULL,
                                    lavoptions = NULL,
                                    lavpartable = NULL,
                                    lavsamplestats = NULL,
                                    lavh1 = NULL) {
  # # # # # # # # # # #
  # #  8. lavstart # #
  # # # # # # # # # # #

  # if slot_model is NULL
  #   if lavpartable$est not NULL and lavoptions$start == "default"
  #     if there are free variances with est==0 or there are NA's in est
  #       compute start column in lavpartable via lav_start
  #     else
  #       set start column in lavpartable equal to est column
  #   else
  #     compute start column via lav_start and
  #       check via lav_start_check_cov if demanded (lavoptions$check.start)

  samplestats_flag <- TRUE
  if (!is.null(lavoptions$samplestats) &&
      !lavoptions$samplestats) {
    samplestats_flag <- FALSE
  }

  if (is.null(slot_model)) {
    # check if we have provided a full parameter table as model = input
    if (!is.null(lavpartable$est) && is.character(lavoptions$start) &&
      lavoptions$start == "default") {
      if (lav_verbose()) {
        cat("lavstart           ...")
      }
      # check if all 'est' values look ok
      # this is not the case, eg, if partables have been merged eg, as
      # in semTools' auxiliary() function

      # check for zero free variances and NA values
      if (is.null(lavpartable$lower)) {
        zero_idx <- which(lavpartable$free > 0L &
          lavpartable$op == "~~" &
          lavpartable$lhs == lavpartable$rhs &
          lavpartable$est == 0)
      } else {
        # ignore zero variances on the boundary; new in 0.6-21
        zero_idx <- which(lavpartable$free > 0L &
          lavpartable$op == "~~" &
          lavpartable$lhs == lavpartable$rhs &
          lavpartable$est == 0 &
          lavpartable$lower != 0)
      }

      if (length(zero_idx) > 0L || any(is.na(lavpartable$est))) {
        lavpartable$start <- lav_start(
          start_method = lavoptions$start,
          lavpartable = lavpartable,
          lavsamplestats = lavsamplestats,
          model_type = lavoptions$model.type,
          reflect = FALSE,
          samplestats_flag = samplestats_flag,
          # order.lv.by  = lavoptions$rotation.args$order_lv_by,
          order_lv_by = "none"
          )
      } else {
        lavpartable$start <- lavpartable$est
      }

      # check for exogenous parameters: if the dataset changed, we must
      # update them! (new in 0.6-16)
      # ... or not? (not compatible with how we bootstrap under fixed.x = T)
      # we really need to think about this more carefully...
      #
      # if (any(lavpartable$exo == 1L)) {
      #     # FIXME: there should be an easier way just to
      #     # (re)initialize the exogenous part of the model
      #     tmp <- lav_start(start.method   = "lavaan", # not "simple"
      #                                                 # if fixed.x = TRUE
      #                    lavpartable    = lavpartable,
      #                    lavsamplestats = lavsamplestats,
      #                    lavh1          = lavh1,
      #                    model.type     = lavoptions$model.type,
      #                    reflect      = FALSE,
      #                    #order.lv.by  = lavoptions$rotation.args$order_lv_by,
      #                    order.lv.by  = "none",
      #                    debug          = lav_debug())
      #     exo.idx <- which(lavpartable$exo == 1L)
      #     lavpartable$start[exo.idx] <- tmp[exo.idx]
      # }

      if (lav_verbose()) {
        cat(" done.\n")
      }
    } else {
      if (lav_verbose()) {
        cat("lavstart           ...")
      }
      start_values <- lav_start(
        start_method = lavoptions$start,
        lavpartable = lavpartable,
        lavsamplestats = lavsamplestats,
        lavh1 = lavh1,
        model_type = lavoptions$model.type,
        reflect = FALSE,
        samplestats_flag = samplestats_flag,
        # order.lv.by  = lavoptions$rotation.args$order_lv_by,
        order_lv_by = "none"
      )

      # sanity check
      if (!is.null(lavoptions$check.start) && lavoptions$check.start) {
        start_values <- lav_start_check_cov(
          lavpartable = lavpartable,
          start = start_values
        )
      }

      lavpartable$start <- start_values
      if (lav_verbose()) {
        cat(" done.\n")
      }
    }
  }

  lavpartable
}

# lav_gamma_recipe() -- single source of truth for "which Gamma (= NACOV,
# N times the asymptotic variance matrix of the sample statistics) does this
# analysis use, and with which settings?"
#
# Both the fit-time orchestrator (lav_samp_from_data) and the post-fit
# accessor (lav_object_gamma) consult this recipe, so that a Gamma that is
# recomputed after fitting cannot diverge from the one used at fit time.
#
# The recipe describes the DEFAULT policy implied by lavoptions + lavdata;
# caller-level overrides (a user-supplied NACOV= argument, or nacov =
# TRUE/FALSE) are handled by the callers themselves.

# YR 07 July 2026

lav_gamma_recipe <- function(lavoptions = NULL, lavdata = NULL,
                             nacov_compute = NULL) {
  stopifnot(!is.null(lavoptions), !is.null(lavdata))

  missing <- lavoptions$missing
  estimator <- lavoptions$estimator
  se <- lavoptions$se
  test <- lavoptions$test
  correlation <- isTRUE(lavoptions$correlation) # absent in old objects

  # data properties
  categorical <- any(lavdata@ov$type == "ordered")
  multilevel <- lavdata@nlevels > 1L
  clustered <- !multilevel && length(lavdata@cluster) > 0L
  wt <- length(lavdata@weights) > 0L && !is.null(lavdata@weights[[1]])

  # design vs frequency weighting of the (categorical/continuous) Gamma
  swt_type <- if (!is.null(lavoptions$sampling.weights.type)) {
    lavoptions$sampling.weights.type
  } else {
    "design"
  }

  # partial correlation structure: only these observed variables are scaled
  # to unit variance (empty -> all of them, the usual correlation structure)
  correlation_ov <- lavoptions$.correlation.ov
  if (is.null(correlation_ov)) {
    correlation_ov <- character(0L)
  }

  # 1. should NACOV be computed at fit time? (default policy)
  if (is.null(nacov_compute)) {
    nacov_compute <- FALSE
    if (se %in% c("robust.sem", "robust.sem.nt", "robust.cluster.sem") &&
        missing == "listwise") {
      nacov_compute <- TRUE
    }
    # note: test can be a vector...
    if (missing == "listwise" && any(test %in% c(
      "satorra.bentler",
      "mean.var.adjusted",
      "mean.var.adjusted.corrected",
      "scaled.shifted",
      "scaled.shifted.corrected"
    ))) {
      nacov_compute <- TRUE
    }
    if (missing == "listwise" &&
        any(vapply(test, lav_test_fmg_is_fmg, logical(1L)))) {
      nacov_compute <- TRUE
    }
    if (estimator == "IV" &&
        lavoptions$estimator.args$iv_vcov_stage1 == "gamma") {
      nacov_compute <- TRUE
    }
    if (any(missing == c("two.stage", "robust.two.stage")) &&
        estimator %in% c("ULS", "GLS", "WLS", "DLS")) {
      nacov_compute <- TRUE
    }
  }

  # 2. which flavor? (mirrors the branch order of the NACOV block in
  #    lav_samp_from_data, and the two-level builders in xxx_lavaan)
  #    - "adf"          ADF Gamma (lav_samp_gamma)
  #    - "nt"           normal-theory Gamma (lav_samp_gamma_nt); note: only
  #                     reachable for the (D)WLS-family estimators -- for
  #                     estimator = "ML", se = "robust.sem.nt" still stores
  #                     the ADF Gamma
  #    - "adf.cor"      (partial-)correlation ADF Gamma via the delta method
  #    - "missing.two.stage" / "missing.robust.two.stage"
  #                     NACOV of the saturated EM moments
  #    - "cat"          categorical: WLS.W (muthen1984), rescaled by N
  #    - "twolevel" / "twolevel.cat"
  #                     two-level (D)WLS builders (cluster sandwich)
  #    - "none"         no NACOV computed at fit time
  wls_family <- estimator %in% c("WLS", "DWLS", "ULS", "DLS", "IV", "catML")
  flavor <- "none"
  if (multilevel) {
    if (estimator %in% c("WLS", "DWLS", "ULS")) {
      flavor <- if (categorical) "twolevel.cat" else "twolevel"
    }
  } else if (nacov_compute && !categorical &&
             any(missing == c("two.stage", "robust.two.stage")) &&
             estimator %in% c("ULS", "GLS", "WLS", "DLS")) {
    flavor <- paste0("missing.", missing)
  } else if (estimator %in% c("ML", "GLS") && missing == "listwise" &&
             nacov_compute) {
    flavor <- if (correlation) "adf.cor" else "adf"
  } else if (wls_family) {
    if (categorical) {
      flavor <- "cat"
    } else if (correlation) {
      flavor <- "adf.cor"
    } else if (estimator %in% c("ULS", "DWLS") &&
               "robust.sem.nt" %in% se) {
      # only for ULS/DWLS (their continuous default se): for WLS/DLS
      # the (full) weight matrix is derived from the NACOV, which must
      # remain ADF
      flavor <- "nt"
    } else {
      flavor <- "adf"
    }
  }

  # 3. the knobs, as passed to the constructors
  #    note the asymmetry (as at fit time): gamma.wls.mplus only applies to
  #    the plain ADF Gamma of the (D)WLS-family estimators; the ML/GLS branch
  #    always uses mplus_wls = FALSE
  mplus_wls <- FALSE
  if (flavor == "adf" && wls_family && isTRUE(lavoptions$gamma.wls.mplus)) {
    mplus_wls <- TRUE
  }

  list(
    # decisions
    compute = nacov_compute,
    flavor = flavor,
    # data properties
    categorical = categorical,
    multilevel = multilevel,
    clustered = clustered,
    wt = wt,
    swt.type = swt_type,
    # model/options knobs
    meanstructure = lavoptions$meanstructure,
    fixed.x = lavoptions$fixed.x,
    conditional.x = lavoptions$conditional.x,
    correlation = correlation,
    correlation.ov = correlation_ov,
    n.minus.one = lavoptions$gamma.n.minus.one,
    unbiased = lavoptions$gamma.unbiased,
    mplus.wls = mplus_wls,
    group.w.free = lavoptions$group.w.free,
    missing = missing
  )
}

# which variables are scaled to unit variance under a (partial) correlation
# structure? (all of them, unless a subset was requested via
# correlation = c(...))
lav_gamma_recipe_cor_idx <- function(recipe, ov_names_g) {
  if (length(recipe$correlation.ov) > 0L) {
    which(ov_names_g %in% recipe$correlation.ov)
  } else {
    seq_along(ov_names_g)
  }
}

# create a thin forwarding alias for the function `target` (a name):
# the alias has the same formals as the target -- so documented usage,
# argument matching and missing() semantics are unaffected -- but its
# body merely re-dispatches the call to the target.
#
# rationale: a plain copy (`alias <- target`) stores the complete
# function a *second* time in the lazy-loading DB of the installed
# package; for the larger user-facing functions and their synonyms
# (parameterEstimates/parameterestimates, modificationIndices/..., ...)
# these copies added up to more than 100 Kb of installed size.
#
# note: the target must already exist when the alias is created, so the
# alias must come later in the collation order (00alias.R itself sorts
# before all lav_*.R files).
# the modindices() default `maximum_number = nrow(list_1)` refers to a
# variable that is only created inside the body; the thin forwarders
# below (which share the formals but not the body) would otherwise
# trigger a false-positive codetools NOTE
utils::globalVariables("list_1")

lav_alias <- function(target) {
  pkg_env <- parent.frame() # the package environment under construction
  fun <- eval(
    bquote(function() {
      sc <- sys.call()
      sc[[1L]] <- .(as.name(target)) # resolved in the lavaan namespace
      eval(sc, envir = parent.frame())
    }),
    envir = pkg_env
  )
  formals(fun) <- formals(get(target, envir = pkg_env, mode = "function"))
  fun
}

lav_model_group_mm_indices <- function(nmat) {
  nmat_cumsum <- cumsum(c(0L, nmat))
  lapply(seq_along(nmat), function(g) {
    seq_len(nmat[g]) + nmat_cumsum[g]
  })
}

lav_model_sigma_extra <- function(sigma_hat_g = NULL,
                                  check_sigma_pd = "chol") {
  # check if matrix is positive definite
  if (check_sigma_pd == "chol") {
    # fast path: try Cholesky directly
    c_s <- tryCatch(chol(sigma_hat_g), error = function(e) NULL)
    is_pd <- !is.null(c_s)
  } else {
    # slow path: eigenvalue-based PD check
    ev <- eigen(sigma_hat_g, symmetric = TRUE, only.values = TRUE)$values
    is_pd <- !(any(ev < sqrt(.Machine$double.eps)) || sum(ev) == 0)
  }

  if (!is_pd) {
    sigma_hat_inv <- MASS::ginv(sigma_hat_g)
    sigma_hat_log_det <- log(.Machine$double.eps)
    attr(sigma_hat_g, "po") <- FALSE
    attr(sigma_hat_g, "inv") <- sigma_hat_inv
    attr(sigma_hat_g, "log.det") <- sigma_hat_log_det
  } else {
    if (check_sigma_pd != "chol") {
      c_s <- chol(sigma_hat_g)
    }
    sigma_hat_inv <- chol2inv(c_s)
    d <- diag(c_s)
    sigma_hat_log_det <- 2 * sum(log(d))
    attr(sigma_hat_g, "po") <- TRUE
    attr(sigma_hat_g, "inv") <- sigma_hat_inv
    attr(sigma_hat_g, "log.det") <- sigma_hat_log_det
  }

  sigma_hat_g
}

lav_model_sigma <- function(lavmodel = NULL, glist = NULL, extra = FALSE,
                            delta = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  # check.sigma.pd -- new in 0.6-21
  check_sigma_pd <- get0("opt_check_sigma_pd", lavaan_cache_env,
                         ifnotfound = "chol")

  nmat <- lavmodel@nmat
  nblocks <- lavmodel@nblocks
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  sigma_hat <- vector("list", length = nblocks)

  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      sigma_hat[[g]] <- lav_lisrel_sigma(
        mlist = mlist,
        delta = delta
      )
    } else if (representation == "RAM") {
      sigma_hat[[g]] <- lav_ram_sigmahat(mlist = mlist, delta = delta)
    } else {
      lav_msg_stop(gettext(
        "only LISREL and RAM representation has been implemented for now"))
    }
    if (lav_debug()) print(sigma_hat[[g]])

    if (extra) {
      sigma_hat[[g]] <- lav_model_sigma_extra(
        sigma_hat_g = sigma_hat[[g]],
        check_sigma_pd = check_sigma_pd
      )
    }
  } # nblocks
  sigma_hat
}

## only if conditional.x = TRUE
## compute the (larger) unconditional 'joint' covariance matrix (y,x)
##
## Sigma (Joint ) = [ (S11, S12),
##                    (S21, S22) ] where
##     S11 = Sigma.res + PI %*% cov.x %*% t(PI)
##     S12 = PI %*% cov.x
##     S21 = cov.x %*% t(PI)
##     S22 = cov.x
lav_model_cond2joint_sigma <- function(lavmodel = NULL, glist = NULL,
                                       extra = FALSE, delta = TRUE) {
  stopifnot(lavmodel@conditional.x)

  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  # check.sigma.pd -- new in 0.6-21
  check_sigma_pd <- get0("opt_check_sigma_pd", lavaan_cache_env,
                         ifnotfound = "chol")

  nmat <- lavmodel@nmat
  nblocks <- lavmodel@nblocks
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  sigma_hat <- vector("list", length = nblocks)

  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      res_sigma <- lav_lisrel_sigma(mlist = mlist, delta = delta)
      # res.int <- lav_lisrel_mu(MLIST = MLIST)
      res_slopes <- lav_lisrel_pi(mlist = mlist)
      s_xx <- mlist$cov.x

      s_yy <- res_sigma + res_slopes %*% s_xx %*% t(res_slopes)
      s_yx <- res_slopes %*% s_xx
      s_xy <- s_xx %*% t(res_slopes)

      sigma_hat[[g]] <- rbind(cbind(s_yy, s_yx), cbind(s_xy, s_xx))
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }
    if (lav_debug()) print(sigma_hat[[g]])

    if (extra) {
      sigma_hat[[g]] <- lav_model_sigma_extra(
        sigma_hat_g = sigma_hat[[g]],
        check_sigma_pd = check_sigma_pd
      )
    }
  } # nblocks

  sigma_hat
}

lav_model_mu <- function(lavmodel = NULL, glist = NULL) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nmat <- lavmodel@nmat
  nblocks <- lavmodel@nblocks
  representation <- lavmodel@representation
  meanstructure <- lavmodel@meanstructure
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  mu_hat <- vector("list", length = nblocks)

  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (!meanstructure) {
      mu_hat[[g]] <- numeric(lavmodel@nvar[g])
    } else if (representation == "LISREL") {
      mu_hat[[g]] <- lav_lisrel_mu(mlist = mlist)
    } else if (representation == "RAM") {
      mu_hat[[g]] <- lav_ram_muhat(mlist = mlist)
    } else {
      lav_msg_stop(gettext(
        "only RAM and LISREL representation has been implemented for now"))
    }

    # new in 0.6-20: if a variable is ordinal, set its mean to zero
    # (even if NU is not all zero, as in a multiple group analysis with
    # group.equal = "thresholds")
    #
    # the logic is: Mu.hat is about 'y', not 'y-star'
    # the non-free intercepts (in TAU) are used when computing the
    # model-implied thresholds, but they do not say anything about the
    # 'observed' mean of 'y'
    if (lavmodel@categorical) {
      ord_idx <- unique(lavmodel@th.idx[[g]][lavmodel@th.idx[[g]] > 0L])
      mu_hat[[g]][ord_idx] <- 0
    }
  } # nblocks

  mu_hat
}

## only if conditional.x = TRUE
## compute the (larger) unconditional 'joint' mean vector (y,x)
##
## Mu (Joint ) = [ Mu.y, Mu.x ] where
##     Mu.y = res.int + PI %*% M.x
##     Mu.x = M.x
lav_model_cond2joint_mu <- function(lavmodel = NULL, glist = NULL) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nmat <- lavmodel@nmat
  nblocks <- lavmodel@nblocks
  representation <- lavmodel@representation
  meanstructure <- lavmodel@meanstructure
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  mu_hat <- vector("list", length = nblocks)

  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]

    if (!meanstructure) {
      mu_hat[[g]] <- numeric(lavmodel@nvar[g])
    } else if (representation == "LISREL") {
      mlist <- glist[mm_in_group]
      res_int <- lav_lisrel_mu(mlist = mlist)
      res_slopes <- lav_lisrel_pi(mlist = mlist)
      m_x <- mlist$mean.x

      mu_y <- res_int + res_slopes %*% m_x
      mu_x <- m_x
      mu_hat[[g]] <- c(mu_y, mu_x)
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }
  } # nblocks

  mu_hat
}

# TH.star = DELTA.star * (th.star - pi0.star)
# see Muthen 1984 eq 11
lav_model_th <- function(lavmodel = NULL, glist = NULL, delta = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  th_idx <- lavmodel@th.idx
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  th <- vector("list", length = nblocks)

  # compute TH for each group
  for (g in seq_len(nblocks)) {
    if (length(th_idx[[g]]) == 0) {
      th[[g]] <- numeric(0L)
      next
    }

    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]

    if (representation == "LISREL") {
      th[[g]] <- lav_lisrel_th(
        mlist = glist[mm_in_group],
        th_idx = th_idx[[g]], delta = delta
      )
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }
  }

  th
}

# PI = slope structure
# see Muthen 1984 eq 12
lav_model_pi <- function(lavmodel = NULL, glist = NULL, delta = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  conditional_x <- lavmodel@conditional.x
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  pi0 <- vector("list", length = nblocks)

  # compute TH for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (!conditional_x) {
      pi_g <- numeric(lavmodel@nvar[g])
    } else if (representation == "LISREL") {
      pi_g <- lav_lisrel_pi(mlist = mlist, delta = delta)
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    pi0[[g]] <- pi_g
  }

  pi0
}


# GW = group weight
lav_model_gw <- function(lavmodel = NULL, glist = NULL) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  group_w_free <- lavmodel@group.w.free
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  gw <- vector("list", length = nblocks)

  # compute GW for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (!group_w_free) {
      gw_g <- 0.0 # FIXME
    } else if (representation == "LISREL") {
      gw_g <- as.numeric(mlist$gw[1, 1])
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    gw[[g]] <- gw_g
  }

  # transform to proportions
  # gw <- unlist(GW)
  # gw <- exp(gw) / sum(exp(gw))
  # for(g in 1:nblocks) {
  #    GW[[g]] <- gw[g]
  # }

  gw
}

# *unconditional* variance/covariance matrix of Y
#  - same as Sigma.hat if all Y are continuous)
#  - if also Gamma, cov.x is used (only if categorical)
lav_model_vy <- function(lavmodel = NULL, glist = NULL, diagonal_only = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  vy <- vector("list", length = nblocks)

  # compute TH for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      vy_g <- lav_lisrel_vy(mlist = mlist)
    } else if (representation == "RAM") {
      # does not work for categorical setting yet
      stopifnot(!lavmodel@categorical)
      # does not work if conditional.x = TRUE
      stopifnot(!lavmodel@conditional.x)
      vy_g <- lav_ram_sigmahat(mlist = mlist)
    } else {
      lav_msg_stop(gettext(
        "only RAM and LISREL representation has been implemented for now"))
    }

    if (diagonal_only) {
      vy[[g]] <- diag(vy_g)
    } else {
      vy[[g]] <- vy_g
    }
  }

  vy
}

# V(ETA): latent variances variances/covariances
lav_model_veta <- function(lavmodel = NULL, glist = NULL,
                           remove_dummy_lv = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  veta <- vector("list", length = nblocks)

  # compute VETA for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      veta_g <- lav_lisrel_veta(mlist = mlist)

      if (remove_dummy_lv) {
        # remove all dummy latent variables
        lv_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[g]],
          lavmodel@ov.x.dummy.lv.idx[[g]]
        )
        if (!is.null(lv_idx)) {
          veta_g <- veta_g[-lv_idx, -lv_idx, drop = FALSE]
        }
      }
    } else if (representation == "RAM") {
      veta_g <- lav_ram_veta(mlist = mlist)
    } else {
      lav_msg_stop(gettext(
        "only LISREL and RAM representation has been implemented for now"))
    }

    veta[[g]] <- veta_g
  }

  veta
}

# V(ETA|x_i): latent variances variances/covariances, conditional on x_
# - this is always (I-B)^-1 PSI (I-B)^-T
# - by default the (observed) dummy latent variables are REMOVED, but if
#   remove_dummy_lv = FALSE, the full latent space is kept (needed when the
#   result must be conformable with the full LAMBDA, e.g. in lavPredict())
lav_model_vetax <- function(lavmodel = NULL, glist = NULL,
                            remove_dummy_lv = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  eta <- vector("list", length = nblocks)

  # compute ETA for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      if (remove_dummy_lv) {
        lv_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[g]],
          lavmodel@ov.x.dummy.lv.idx[[g]]
        )
      } else {
        lv_idx <- NULL
      }
      eta_g <- lav_lisrel_vetax(
        mlist = mlist,
        lv_dummy_idx = lv_idx
      )
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    eta[[g]] <- eta_g
  }

  eta
}

# COV: observed+latent variances variances/covariances
lav_model_cov_both <- function(lavmodel = NULL, glist = NULL,
                               remove_dummy_lv = FALSE, delta = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  cov_1 <- vector("list", length = nblocks)

  # compute COV for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      cov_g <- lav_lisrel_cov_both(mlist = mlist, delta = delta)

      if (remove_dummy_lv) {
        # remove all dummy latent variables
        lv_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[g]],
          lavmodel@ov.x.dummy.lv.idx[[g]]
        )
        if (!is.null(lv_idx)) {
          # offset for ov
          lambda_names <-
            lavmodel@dimNames[[which(names(glist) == "lambda")[g]]][[1L]]
          lv_idx <- lv_idx + length(lambda_names)
          cov_g <- cov_g[-lv_idx, -lv_idx, drop = FALSE]
        }
      }
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    cov_1[[g]] <- cov_g
  }

  cov_1
}


# E(ETA): expectation (means) of latent variables (return vector)
lav_model_eeta <- function(lavmodel = NULL, glist = NULL, lavsamplestats = NULL,
                           remove_dummy_lv = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  eeta <- vector("list", length = nblocks)

  # compute E(ETA) for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      eeta_g <- lav_lisrel_eeta(mlist,
        mean_x = lavsamplestats@mean.x[[g]],
        sample_mean = lavsamplestats@mean[[g]],
        ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
        ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
        ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
        ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]]
      )
      if (remove_dummy_lv) {
        # remove dummy
        lv_dummy_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[g]],
          lavmodel@ov.x.dummy.lv.idx[[g]]
        )
        if (length(lv_dummy_idx) > 0L) {
          eeta_g <- eeta_g[-lv_dummy_idx]
        }
      }
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    eeta[[g]] <- eeta_g
  }

  eeta
}

# E(ETA|x_i): conditional expectation (means) of latent variables
# for a given value of x_i (instead of E(x_i))
lav_model_eetax <- function(lavmodel = NULL, glist = NULL,
                            lavsamplestats = NULL, exo = NULL,
                            nobs = NULL, remove_dummy_lv = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  eetax <- vector("list", length = nblocks)

  # compute E(ETA) for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    exo_1 <- exo[[g]]
    if (is.null(exo_1)) {
      # create empty matrix
      exo_1 <- matrix(0, nobs[[g]], 0L)
    }

    if (representation == "LISREL") {
      eetax_g <- lav_lisrel_eetax(mlist,
        exo = exo_1, n = nobs[[g]],
        sample_mean = lavsamplestats@mean[[g]],
        ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
        ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
        ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
        ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]]
      )

      if (remove_dummy_lv) {
        # remove dummy
        lv_dummy_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[g]],
          lavmodel@ov.x.dummy.lv.idx[[g]]
        )
        if (length(lv_dummy_idx) > 0L) {
          eetax_g <- eetax_g[, -lv_dummy_idx, drop = FALSE]
        }
      }
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    eetax[[g]] <- eetax_g
  }

  eetax
}

# return 'regular' LAMBDA
lav_model_lambda <- function(lavmodel = NULL, glist = NULL,
                             handle_dummy_lv = TRUE,
                             remove_dummy_lv = FALSE,
                             use_wmat = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  mm_lambda <- vector("list", length = nblocks)

  # compute LAMBDA for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      if (handle_dummy_lv) {
        ov_y_dummy_ov_idx <- lavmodel@ov.y.dummy.ov.idx[[g]]
        ov_x_dummy_ov_idx <- lavmodel@ov.x.dummy.ov.idx[[g]]
        ov_y_dummy_lv_idx <- lavmodel@ov.y.dummy.lv.idx[[g]]
        ov_x_dummy_lv_idx <- lavmodel@ov.x.dummy.lv.idx[[g]]
      } else {
        ov_y_dummy_ov_idx <- NULL
        ov_x_dummy_ov_idx <- NULL
        ov_y_dummy_lv_idx <- NULL
        ov_x_dummy_lv_idx <- NULL
      }
      lambda_g <- lav_lisrel_lambda(
        mlist = mlist,
        ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
        ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
        ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
        ov_x_dummy_lv_idx = ov_x_dummy_lv_idx,
        remove_dummy_lv = remove_dummy_lv,
        use_wmat = use_wmat
      )
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    mm_lambda[[g]] <- lambda_g
  }

  mm_lambda
}

# THETA: observed (residual) variances
lav_model_theta <- function(lavmodel = NULL, glist = NULL, fix = TRUE,
                            use_wmat = FALSE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  mm_theta <- vector("list", length = nblocks)

  # compute THETA for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      if (fix) {
        theta_g <- lav_lisrel_theta(
          mlist = mlist,
          ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
          ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
          ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
          ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
          use_wmat = use_wmat
        )
      } else {
        theta_g <- lav_lisrel_theta(mlist = mlist)
      }
    } else if (representation == "RAM") {
      ov_idx <- as.integer(mlist$ov.idx[1, ])
      theta_g <- mlist$S[ov_idx, ov_idx, drop = FALSE]
    } else {
      lav_msg_stop(gettext(
        "only LISREL and RAM representation has been implemented for now"))
    }

    mm_theta[[g]] <- theta_g
  }

  mm_theta
}

# E(Y): expectation (mean) of observed variables
# returns vector 1 x nvar
lav_model_ey <- function(lavmodel = NULL, glist = NULL, lavsamplestats = NULL,
                         delta = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  nblocks <- lavmodel@nblocks
  nmat <- lavmodel@nmat
  representation <- lavmodel@representation
  mm_idx <- lav_model_group_mm_indices(nmat)

  # return a list
  ey <- vector("list", length = nblocks)

  # compute E(Y) for each group
  for (g in seq_len(nblocks)) {
    # which mm belong to group g?
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (representation == "LISREL") {
      ey_g <- lav_lisrel_ey(
        mlist = mlist,
        mean_x = lavsamplestats@mean.x[[g]],
        sample_mean = lavsamplestats@mean[[g]],
        ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
        ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
        ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
        ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
        delta = delta
      )
    } else {
      lav_msg_stop(gettext(
        "only representation LISREL has been implemented for now"))
    }

    ey[[g]] <- ey_g
  }

  ey
}


# E(Y | ETA, x_i): conditional expectation (means) of observed variables
# for a given value of x_i AND eta_i
lav_model_yhat <- function(lavmodel = NULL, glist = NULL, lavsamplestats = NULL,
                           exo = NULL, nobs = NULL, eta = NULL,
                           duplicate = FALSE, delta = TRUE) {
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST

  # ngroups, not nblocks!
  ngroups <- lavsamplestats@ngroups
  mm_idx <- lav_model_group_mm_indices(lavmodel@nmat)

  # return a list
  yhat <- vector("list", length = ngroups)

  # compute YHAT for each group
  for (g in seq_len(ngroups)) {
    # which mm belong to group g?
    # FIXME: what if more than g blocks???
    mm_in_group <- mm_idx[[g]]
    mlist <- glist[mm_in_group]

    if (is.null(exo[[g]]) && duplicate) {
      nobs_1 <- nobs[[g]]
    } else {
      nobs_1 <- 1L
    }

    if (lavmodel@representation == "LISREL") {
      if (lavmodel@conditional.x) {
        yhat[[g]] <- lav_lisrel_eyetax(
          mlist = mlist,
          exo = exo[[g]], eta = eta[[g]], n = nobs_1,
          sample_mean = lavsamplestats@mean[[g]],
          ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
          ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
          ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
          ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
          delta = delta
        )
      } else {
        # unconditional case
        yhat[[g]] <- lav_lisrel_eyetax3(
          mlist = mlist,
          eta = eta[[g]],
          sample_mean = lavsamplestats@mean[[g]],
          mean_x = lavsamplestats@mean.x[[g]],
          ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
          ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
          ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
          ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
          delta = delta
        )
        # impute back ov.y values that are NOT indicators
      }
    } else {
      lav_msg_stop(gettextf("representation %s not supported yet.",
                            lavmodel@representation))
    }
  }

  yhat
}

# overview NLMINB (default) versus CONSTR (=constrained optimization)

#                         | cin.simple | nonlinear | no cin
# ----------------------------------------------------------
# eq.constraints (linear) | CONSTR     | CONSTR    | NLMINB
# ceq.nonlinear           | CONSTR     | CONSTR    | CONSTR
# ceq.simple              | NLMINB     | CONSTR    | NLMINB
# no ceq                  | NLMINB     | CONSTR    | NLMINB



# model estimation
lav_model_est <- function(lavmodel = NULL,
                               lavpartable = NULL, # for parscale = "stand"
                               lavh1 = NULL, # for multilevel + parsc
                               lavsamplestats = NULL,
                               lavdata = NULL,
                               lavoptions = NULL,
                               lavcache = list(),
                               start = "model",
                               do_fit = TRUE) {
  lavpartable <- lav_pt_set_cache(lavpartable)
  estimator <- lavoptions$estimator
  verbose <- lav_verbose()
  debug <- lav_debug()
  ngroups <- lavsamplestats@ngroups

  if (lavsamplestats@missing.flag || estimator == "PML" ||
      lavdata@nlevels > 1L) {
    group_weight <- FALSE
  } else {
    group_weight <- TRUE
  }

  # backwards compatibility < 0.6-11
  if (is.null(lavoptions$optim.partrace)) {
    lavoptions$optim.partrace <- FALSE
  }

  if (lavoptions$optim.partrace) {
    # fx + parameter values
    penv <- new.env()
    penv$PARTRACE <- matrix(NA, nrow = 0, ncol = lavmodel@nx.free + 1L)
  }

  # starting values (ignoring equality constraints)
  x_unpack <- lav_model_get_parameters(lavmodel)

  # override? use simple instead? (new in 0.6-7)
  if (start == "simple") {
    start_1 <- numeric(length(lavpartable$lhs))
    # set loadings to 0.7
    loadings_idx <- which(lavpartable$free > 0L &
      lavpartable$op == "=~")
    if (length(loadings_idx) > 0L) {
      start_1[loadings_idx] <- 0.7
    }
    # set (only) variances to 1
    var_idx <- which(lavpartable$free > 0L &
      lavpartable$op == "~~" &
      lavpartable$lhs == lavpartable$rhs)
    if (length(var_idx) > 0L) {
      start_1[var_idx] <- 1
    }

    if (lavmodel@ceq.simple.only) {
      x_unpack <- start_1[lavpartable$free > 0L &
        !duplicated(lavpartable$free)]
    } else {
      x_unpack <- start_1[lavpartable$free > 0L]
    }

    # override? use random starting values instead? (new in 0.6-18)
  } else if (start == "random") {
    start_1 <- lav_pt_random(
      lavpartable = lavpartable,
      # needed if we still need to compute bounds:
      lavh1 = lavh1,
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      lavoptions = lavoptions
    )

    if (lavmodel@ceq.simple.only) {
      x_unpack <- start_1[lavpartable$free > 0L &
        !duplicated(lavpartable$free)]
    } else {
      x_unpack <- start_1[lavpartable$free > 0L]
    }
  }

  # new in 0.7-1: multilevel models where a whole block (level) is
  # saturated: the h1 estimates for this block are already available;
  # temporarily fix these parameters at their h1 values, so they do not
  # enter the optimization again as free parameters (this often helps
  # convergence); after optimization, they are free parameters again
  h1_sat_idx <- integer(0L)
  h1_sat_ok <- lavoptions$optim.fix.saturated
  if (is.null(h1_sat_ok)) { # backwards compatibility
    h1_sat_ok <- TRUE
  }
  if (h1_sat_ok && lavdata@nlevels > 1L && estimator == "ML" &&
      # only if the optimizer honours box constraints (NLMINB family)
      length(lavmodel@ceq.nonlinear.idx) == 0L &&
      (lavmodel@cin.simple.only ||
       (length(lavmodel@cin.linear.idx) == 0L &&
        length(lavmodel@cin.nonlinear.idx) == 0L)) &&
      (is.null(lavoptions$optim.method) ||
       toupper(lavoptions$optim.method) %in%
         c("NLMINB", "NLMINB0", "NLMINB1", "L.BFGS.B"))) {
    tmp <- lav_model_est_h1_saturated(
      lavmodel = lavmodel, lavpartable = lavpartable,
      lavdata = lavdata, lavh1 = lavh1
    )
    if (length(tmp$x.idx) > 0L) {
      h1_sat_idx <- tmp$x.idx
      x_unpack[h1_sat_idx] <- tmp$value
    }
  }

  # 1. parameter scaling (to handle data scaling, not parameter scaling)
  parscale <- rep(1.0, length(x_unpack))

  # for < 0.6 compatibility
  if (is.null(lavoptions$optim.parscale)) {
    lavoptions$optim.parscale <- "none"
  }

  if (lavoptions$optim.parscale == "none") {
    # do nothing, but still set SCALE, as before


    # 0.6-17:
    # only temporarily: 'keep' this mistake, and change it later:
    # (note the "standarized")
    # we only do this to avoid breaking a test in semlbci
    # } else if(lavoptions$optim.parscale %in% c("stand", "st", "standardize",
    #                                           "standarized", "stand.all")) {

    # this is what it should be:
  } else if (lavoptions$optim.parscale %in% c(
    "stand", "st", "standardize",
    "standardized", "stand.all"
  )) {
    # rescale parameters as if the data was standardized
    # new in 0.6-2
    #
    # FIXME: this works well, as long as the variances of the
    #        latent variables (which we do not know) are more or less
    #        equal to 1.0 (eg std.lv = TRUE)
    #
    #        Once we have better estimates of those variances, we could
    #        use them to set the scale
    #

    if (lavdata@nlevels > 1L) {
      if (length(lavh1) > 0L) {
        ov_var <- lapply(lavh1$implied$cov, diag)
      } else {
        ov_var <- lapply(
          do.call(c, lapply(lavdata@Lp, "[[", "ov.idx")),
          function(x) rep(1, length(x))
        )
      }
    } else {
      if (lavoptions$conditional.x) {
        ov_var <- lavsamplestats@res.var
      } else {
        ov_var <- lavsamplestats@var
      }
    }

    if (lavoptions$std.lv) {
      parscale <- lav_standardize_all(
        lavobject = NULL,
        est = rep(1, length(lavpartable$lhs)),
        est_std = rep(1, length(lavpartable$lhs)),
        cov_std = FALSE, ov_var = ov_var,
        lavmodel = lavmodel, lavpartable = lavpartable,
        cov_x = lavsamplestats@cov.x
      )
    } else {
      # needs good estimates for lv variances!
      # if there is a single 'marker' indicator, we could use
      # its observed variance as an upper bound

      # for the moment, set them to 1.0 (instead of 0.05)

      # TODO: USE Bentler's 1982 approach to get an estimate of
      # VETA; use those diagonal elements...
      # but only if we have 'marker' indicators for each LV
      lv_var <- vector("list", lavmodel@ngroups)
      for (g in seq_len(lavmodel@ngroups)) {
        mm_in_group <- 1:lavmodel@nmat[g] + cumsum(c(0, lavmodel@nmat))[g]
        mlist <- lavmodel@GLIST[mm_in_group]
        mm_lambda <- mlist$lambda
        n_lv <- ncol(mm_lambda)
        lv_var[[g]] <- rep(1.0, n_lv)
      }

      parscale <- lav_standardize_all(
        lavobject = NULL,
        est = rep(1, length(lavpartable$lhs)),
        # est.std = rep(1, length(lavpartable$lhs)),
        # here, we use whatever the starting values are
        # for the latent variances...
        cov_std = FALSE, ov_var = ov_var,
        lv_var = lv_var,
        lavmodel = lavmodel, lavpartable = lavpartable,
        cov_x = lavsamplestats@cov.x
      )
    }

    # in addition, take sqrt for variance parameters
    var_idx <- which(lavpartable$op == "~~" &
      lavpartable$lhs == lavpartable$rhs)
    if (length(var_idx) > 0L) {
      parscale[var_idx] <- sqrt(abs(parscale[var_idx]))
    }

    if (lavmodel@ceq.simple.only) {
      parscale <- parscale[lavpartable$free > 0 &
        !duplicated(lavpartable$free)]
    } else {
      parscale <- parscale[lavpartable$free > 0]
    }
  }
  # parscale should obey the equality constraints
  if (lavmodel@eq.constraints && lavoptions$optim.parscale != "none") {
    # pack
    p_pack <- as.numeric((parscale - lavmodel@eq.constraints.k0) %*%
      lavmodel@eq.constraints.K)
    # unpack
    parscale <- as.numeric(lavmodel@eq.constraints.K %*% p_pack) +
      lavmodel@eq.constraints.k0
  }
  if (debug) {
    cat("parscale = ", parscale, "\n")
  }
  z_unpack <- x_unpack * parscale

  # 2. pack (apply equality constraints)
  if (lavmodel@eq.constraints && ncol(lavmodel@eq.constraints.K) > 0L) {
    z_pack <- as.numeric((z_unpack - lavmodel@eq.constraints.k0) %*%
      lavmodel@eq.constraints.K)
  } else {
    z_pack <- z_unpack
  }

  # 3. transform (already constrained) variances to standard deviations?
  # TODO
  # if(lavoptions$optim.var.transform == "sqrt" &&
  #       length(lavmodel@x.free.var.idx) > 0L) {
  #    # transforming variances using atan (or another sigmoid function?)
  #    # FIXME: better approach?
  #    #start.x[lavmodel@x.free.var.idx] <-
  #    #    atan(start.x[lavmodel@x.free.var.idx])
  #    start.x[lavmodel@x.free.var.idx] <-
  #        sqrt(start.x[lavmodel@x.free.var.idx]) # assuming positive var
  # }

  # final starting values for optimizer
  start_x <- z_pack
  if (debug) {
    cat("start.x = ", start_x, "\n")
  }

  # user-specified bounds? (new in 0.6-2)
  if (is.null(lavpartable$lower)) {
    lower <- -Inf
  } else {
    if (lavmodel@ceq.simple.only) {
      free_idx <- which(lavpartable$free > 0L &
        !duplicated(lavpartable$free))
      lower <- lavpartable$lower[free_idx]
    } else if (lavmodel@eq.constraints) {
      # bounds have no effect any longer....
      # 0.6-19 -> we switch to constrained estimation
      #lav_msg_warn(gettext(
      #  "bounds have no effect in the presence of linear
      #          equality constraints"))
      lower <- -Inf
    } else {
      lower <- lavpartable$lower[lavpartable$free > 0L]
    }
  }
  if (is.null(lavpartable$upper)) {
    upper <- +Inf
  } else {
    if (lavmodel@ceq.simple.only) {
      free_idx <- which(lavpartable$free > 0L &
        !duplicated(lavpartable$free))
      upper <- lavpartable$upper[free_idx]
    } else if (lavmodel@eq.constraints) {
      # bounds have no effect any longer....
      if (is.null(lavpartable$lower)) {
        # bounds have no effect any longer....
        # 0.6-19 -> we switch to constrained estimation
        #lav_msg_warn(gettext(
        # "bounds have no effect in the presence of linear
        #          equality constraints"))
      }
      upper <- +Inf
    } else {
      upper <- lavpartable$upper[lavpartable$free > 0L]
    }
  }

  # check for inconsistent lower/upper bounds
  # this may happen if we have equality constraints; qr() may switch
  # the sign...
  bad_idx <- which(lower > upper)
  if (length(bad_idx) > 0L) {
    # switch
    # tmp <- lower[bad.idx]
    # lower[bad.idx] <- upper[bad.idx]
    # upper[bad.idx] <- tmp
    lower[bad_idx] <- -Inf
    upper[bad_idx] <- +Inf
  }

  # new in 0.7-1: fix the parameters of saturated blocks at their h1
  # values during optimization, by forcing lower == upper == start
  if (length(h1_sat_idx) > 0L) {
    lower <- rep(lower, length.out = length(start_x))
    upper <- rep(upper, length.out = length(start_x))
    lower[h1_sat_idx] <- upper[h1_sat_idx] <- start_x[h1_sat_idx]
  }

  # function to be minimized
  objective_function <- function(x, verbose = FALSE, inf_to_max = FALSE,
                                 debug = FALSE) {
    # 3. standard deviations to variances
    # WARNING: x is still packed here!
    # if(lavoptions$optim.var.transform == "sqrt" &&
    #   length(lavmodel@x.free.var.idx) > 0L) {
    #    #x[lavmodel@x.free.var.idx] <- tan(x[lavmodel@x.free.var.idx])
    #    x.var <- x[lavmodel@x.free.var.idx]
    #    x.var.sign <- sign(x.var)
    #    x[lavmodel@x.free.var.idx] <- x.var.sign * (x.var * x.var) # square!
    # }

    # 2. unpack
    if (lavmodel@eq.constraints) {
      x <- as.numeric(lavmodel@eq.constraints.K %*% x) +
        lavmodel@eq.constraints.k0
    }

    # 1. unscale
    x <- x / parscale

    # update GLIST (change `state') and make a COPY!
    glist <- lav_model_x2glist(lavmodel, x = x)

    fx <- lav_model_objective(
      lavmodel = lavmodel,
      glist = glist,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache
    )

    # only for PML: divide by N (to speed up convergence)
    if (estimator == "PML") {
      fx <- fx / lavsamplestats@ntotal
    }



    if (debug || verbose) {
      cat("  objective function  = ",
        sprintf("%18.16f", fx), "\n",
        sep = ""
      )
    }
    if (debug) {
      # cat("Current unconstrained parameter values =\n")
      # tmp.x <- lav_model_get_parameters(lavmodel, glist=GLIST, type="unco")
      # print(tmp.x); cat("\n")
      cat("Current free parameter values =\n")
      print(x)
      cat("\n")
    }

    if (lavoptions$optim.partrace) {
      penv$PARTRACE <- rbind(penv$PARTRACE, c(fx, x))
    }

    # for L-BFGS-B
    # if(infToMax && is.infinite(fx)) fx <- 1e20
    if (!is.finite(fx)) {
      fx_group <- attr(fx, "fx.group")
      fx <- 1e20
      attr(fx, "fx.group") <- fx_group # only for lav_model_fit()
    }

    fx
  }

  gradient_function <- function(x, verbose = FALSE, inf_to_max = FALSE,
                                debug = FALSE) {
    # transform variances back
    # if(lavoptions$optim.var.transform == "sqrt" &&
    #   length(lavmodel@x.free.var.idx) > 0L) {
    #    #x[lavmodel@x.free.var.idx] <- tan(x[lavmodel@x.free.var.idx])
    #    x.var <- x[lavmodel@x.free.var.idx]
    #    x.var.sign <- sign(x.var)
    #    x[lavmodel@x.free.var.idx] <- x.var.sign * (x.var * x.var) # square!
    # }

    # 2. unpack
    if (lavmodel@eq.constraints) {
      x <- as.numeric(lavmodel@eq.constraints.K %*% x) +
        lavmodel@eq.constraints.k0
    }

    # 1. unscale
    x <- x / parscale

    # update GLIST (change `state') and make a COPY!
    glist <- lav_model_x2glist(lavmodel, x = x)

    dx <- lav_model_grad(
      lavmodel = lavmodel,
      glist = glist,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache,
      type = "free",
      group_weight = group_weight, ### check me!!
      ceq_simple = lavmodel@ceq.simple.only
    )

    if (debug) {
      cat("Gradient function (analytical) =\n")
      print(dx)
      cat("\n")
    }

    # 1. scale (note: divide, not multiply!)
    dx <- dx / parscale

    # 2. pack
    if (lavmodel@eq.constraints) {
      dx <- as.numeric(dx %*% lavmodel@eq.constraints.K)
    }

    # 3. transform variances back
    # if(lavoptions$optim.var.transform == "sqrt" &&
    #   length(lavmodel@x.free.var.idx) > 0L) {
    #    x.var <- x[lavmodel@x.free.var.idx] # here in 'var' metric
    #    x.var.sign <- sign(x.var)
    #    x.var <- abs(x.var)
    #    x.sd <- sqrt(x.var)
    #    dx[lavmodel@x.free.var.idx] <-
    #        ( 2 * x.var.sign * dx[lavmodel@x.free.var.idx] * x.sd )
    # }

    # only for PML: divide by N (to speed up convergence)
    if (estimator == "PML") {
      dx <- dx / lavsamplestats@ntotal
    }

    if (debug) {
      cat("Gradient function (analytical, after eq.constraints.K) =\n")
      print(dx)
      cat("\n")
    }

    dx
  }

  gradient_function_numerical <- function(x, verbose = FALSE, debug = FALSE) {
    # NOTE: no need to 'transform' anything here (var/eq)
    # this is done anyway in objective_function

    # numerical approximation using the Richardson method
    npar <- length(x)
    h <- 10e-6
    dx <- numeric(npar)

    ## FIXME: call lav_model_objective directly!!
    for (i in 1:npar) {
      x_left <- x_left2 <- x_right <- x_right2 <- x
      x_left[i] <- x[i] - h
      x_left2[i] <- x[i] - 2 * h
      x_right[i] <- x[i] + h
      x_right2[i] <- x[i] + 2 * h
      fx_left <- objective_function(x_left, verbose = FALSE, debug = FALSE)
      fx_left2 <- objective_function(x_left2, verbose = FALSE, debug = FALSE)
      fx_right <- objective_function(x_right, verbose = FALSE, debug = FALSE)
      fx_right2 <- objective_function(x_right2, verbose = FALSE, debug = FALSE)
      dx[i] <- (fx_left2 - 8 * fx_left + 8 * fx_right - fx_right2) / (12 * h)
    }

    # dx <- lavGradientC(func=objective_function, x=x)
    # does not work if pnorm is involved... (eg PML)

    if (debug) {
      cat("Gradient function (numerical) =\n")
      print(dx)
      cat("\n")
    }

    dx
  }

  gradient_function_numerical_complex <- function(x, verbose = FALSE, debug = FALSE) { # nolint
    dx <- Re(lav_func_grad_complex(
      func = objective_function, x = x,
      h = sqrt(.Machine$double.eps)
    ))
    # does not work if pnorm is involved... (eg PML)

    if (debug) {
      cat("Gradient function (numerical complex) =\n")
      print(dx)
      cat("\n")
    }

    dx
  }


  # check if the initial values produce a positive definite Sigma
  # to begin with -- but only for estimator="ML"
  if (estimator %in% c("ML", "FML", "MML")) {
    sigma_hat <- lav_model_sigma(lavmodel, extra = TRUE)
    for (g in 1:ngroups) {
      if (!attr(sigma_hat[[g]], "po")) {
        group_txt <-
          if (ngroups > 1) gettextf(" in group %s.", g) else "."
        if (debug) {
          print(sigma_hat[[g]][, ])
        }
        lav_msg_warn(gettext(
          "initial model-implied matrix (Sigma) is not positive definite;
          check your model and/or starting parameters"), group_txt)
        x <- start_x
        fx <- as.numeric(NA)
        attr(fx, "fx.group") <- rep(as.numeric(NA), ngroups)
        attr(x, "converged") <- FALSE
        attr(x, "iterations") <- 0L
        attr(x, "control") <- lavoptions$control
        attr(x, "fx") <- fx
        return(x)
      }
    }
  }


  # parameter scaling
  # FIXME: what is the best way to set the scale??
  # current strategy: if startx > 1.0, we rescale by using
  # 1/startx
  scale_1 <- rep(1.0, length(start_x))
  if (lavoptions$optim.parscale == "none") {
    idx <- which(abs(start_x) > 1.0)
    if (length(idx) > 0L) {
      scale_1[idx] <- abs(1.0 / start_x[idx])
    }
  }
  if (debug) {
    cat("SCALE = ", scale_1, "\n")
  }


  # first try: check if starting values return a finite value
  fx <- objective_function(start_x, verbose = verbose, debug = debug)
  if (!is.finite(fx)) {
    # emergency change of start.x
    if (length(h1_sat_idx) > 0L) {
      # do not touch the (fixed) h1 values of saturated blocks
      start_x[-h1_sat_idx] <- start_x[-h1_sat_idx] / 10
    } else {
      start_x <- start_x / 10
    }
  }



  # first some nelder mead steps? (default = FALSE)
  init_nelder_mead <- lavoptions$optim.init_nelder_mead

  # gradient: analytic, numerical or NULL?
  if (is.character(lavoptions$optim.gradient)) {
    if (lavoptions$optim.gradient %in% c("analytic", "analytical")) {
      gradient <- gradient_function
    } else if (lavoptions$optim.gradient %in% c("numerical", "numeric")) {
      gradient <- gradient_function_numerical
    } else if (lavoptions$optim.gradient %in% c("numeric.complex", "complex")) {
      gradient <- gradient_function_numerical_complex
    } else if (lavoptions$optim.gradient %in% c("NULL", "null")) {
      gradient <- NULL
    } else {
      lav_msg_warn(gettext("gradient should be analytic, numerical or NULL"))
    }
  } else if (is.logical(lavoptions$optim.gradient)) {
    if (lavoptions$optim.gradient) {
      gradient <- gradient_function
    } else {
      gradient <- NULL
    }
  } else if (is.null(lavoptions$optim.gradient)) {
    gradient <- gradient_function
  }


  # default optimizer
  if (length(lavmodel@ceq.nonlinear.idx) == 0L &&
      (lavmodel@cin.simple.only || (length(lavmodel@cin.linear.idx)    == 0L &&
                                    length(lavmodel@cin.nonlinear.idx) == 0L))
     ) {
    if (is.null(lavoptions$optim.method)) {
      optimizer <- "NLMINB"
      # OPTIMIZER <- "BFGS"  # slightly slower, no bounds; better scaling!
      # OPTIMIZER <- "L-BFGS-B"  # trouble with Inf values for fx!
    } else {
      optimizer <- toupper(lavoptions$optim.method)
      stopifnot(optimizer %in% c(
        "NLMINB0", "NLMINB1", "NLMINB2",
        "NLMINB", "BFGS", "L.BFGS.B", "NONE"
      ))
      if (optimizer == "NLMINB1") {
        optimizer <- "NLMINB"
      }
    }
  } else {
    if (is.null(lavoptions$optim.method)) {
      optimizer <- "NLMINB.CONSTR"
    } else {
      optimizer <- toupper(lavoptions$optim.method)
      stopifnot(optimizer %in% c("NLMINB.CONSTR", "NLMINB", "NONE"))
    }
    if (optimizer == "NLMINB") {
      optimizer <- "NLMINB.CONSTR"
    }
  }

  if (init_nelder_mead) {
    if (verbose) cat("  initial Nelder-Mead step:\n")
    # trace <- 0L
    # if (verbose) trace <- 1L
    optim_out <- optim(
      par = start_x,
      fn = objective_function,
      method = "Nelder-Mead",
      # control=list(maxit=10L,
      #             parscale=SCALE,
      #             trace=trace),
      hessian = FALSE,
      verbose = verbose, debug = debug
    )
    cat("\n")
    start_x <- optim_out$par
  }



  if (optimizer == "NLMINB0") {
    if (verbose)
      cat("  quasi-Newton steps using NLMINB0 (no analytic gradient):\n")
    # if(debug) control$trace <- 1L;
    control_nlminb <- list(
      eval.max = 20000L,
      iter.max = 10000L,
      trace = 0L,
      # abs.tol=1e-20, ### important!! fx never negative
      abs.tol = (.Machine$double.eps * 10),
      rel.tol = 1e-10,
      # step.min=2.2e-14, # in =< 0.5-12
      step.min = 1.0, # 1.0 in < 0.5-21
      step.max = 1.0,
      x.tol = 1.5e-8,
      xf.tol = 2.2e-14
    )
    control_nlminb <- modifyList(control_nlminb, lavoptions$control)
    control <- control_nlminb[c(
      "eval.max", "iter.max", "trace",
      "step.min", "step.max",
      "abs.tol", "rel.tol", "x.tol", "xf.tol"
    )]
    # cat("DEBUG: control = "); print(str(control.nlminb)); cat("\n")
    optim_out <- nlminb(
      start = start_x,
      objective = objective_function,
      gradient = NULL,
      lower = lower,
      upper = upper,
      control = control,
      scale = scale_1,
      verbose = verbose, debug = debug
    )
    if (verbose) {
      cat("  convergence status (0=ok): ", optim_out$convergence, "\n")
      cat("  nlminb message says: ", optim_out$message, "\n")
      cat("  number of iterations: ", optim_out$iterations, "\n")
      cat(
        "  number of function evaluations [objective, gradient]: ",
        optim_out$evaluations, "\n"
      )
    }

    # try again
    if (optim_out$convergence != 0L) {
      optim_out <- nlminb(
        start = start_x,
        objective = objective_function,
        gradient = NULL,
        lower = lower,
        upper = upper,
        control = control,
        scale = scale_1,
        verbose = verbose, debug = debug
      )
    }

    iterations <- optim_out$iterations
    x <- optim_out$par
    if (optim_out$convergence == 0L) {
      converged <- TRUE
    } else {
      converged <- FALSE
    }
  } else if (optimizer == "NLMINB") {
    if (verbose) cat("  quasi-Newton steps using NLMINB:\n")
    # if(debug) control$trace <- 1L;
    control_nlminb <- list(
      eval.max = 20000L,
      iter.max = 10000L,
      trace = 0L,
      # abs.tol=1e-20, ### important!! fx never negative
      abs.tol = (.Machine$double.eps * 10),
      rel.tol = 1e-10,
      # step.min=2.2e-14, # in =< 0.5-12
      step.min = 1.0, # 1.0 in < 0.5-21
      step.max = 1.0,
      x.tol = 1.5e-8,
      xf.tol = 2.2e-14
    )
    control_nlminb <- modifyList(control_nlminb, lavoptions$control)
    control <- control_nlminb[c(
      "eval.max", "iter.max", "trace",
      "step.min", "step.max",
      "abs.tol", "rel.tol", "x.tol", "xf.tol"
    )]
    # cat("DEBUG: control = "); print(str(control.nlminb)); cat("\n")
    optim_out <- nlminb(
      start = start_x,
      objective = objective_function,
      gradient = gradient,
      lower = lower,
      upper = upper,
      control = control,
      scale = scale_1,
      verbose = verbose, debug = debug
    )
    if (verbose) {
      cat("  convergence status (0=ok): ", optim_out$convergence, "\n")
      cat("  nlminb message says: ", optim_out$message, "\n")
      cat("  number of iterations: ", optim_out$iterations, "\n")
      cat(
        "  number of function evaluations [objective, gradient]: ",
        optim_out$evaluations, "\n"
      )
    }

    iterations <- optim_out$iterations
    x <- optim_out$par
    if (optim_out$convergence == 0L) {
      converged <- TRUE
    } else {
      converged <- FALSE
    }
  } else if (optimizer == "BFGS") {
    # warning: Bollen example with estimator=GLS does NOT converge!
    # (but WLS works!)
    # - BB.ML works too

    control_bfgs <- list(
      trace = 0L, fnscale = 1,
      parscale = scale_1, ## or not?
      ndeps = 1e-3,
      maxit = 10000,
      abstol = 1e-20,
      reltol = 1e-10,
      REPORT = 1L
    )
    control_bfgs <- modifyList(control_bfgs, lavoptions$control)
    control <- control_bfgs[c(
      "trace", "fnscale", "parscale", "ndeps",
      "maxit", "abstol", "reltol", "REPORT"
    )]
    # trace <- 0L; if(verbose) trace <- 1L
    optim_out <- optim(
      par = start_x,
      fn = objective_function,
      gr = gradient,
      method = "BFGS",
      control = control,
      hessian = FALSE,
      verbose = verbose, debug = debug
    )
    if (verbose) {
      cat("  convergence status (0=ok): ", optim_out$convergence, "\n")
      cat("  optim BFGS message says: ", optim_out$message, "\n")
      # cat("number of iterations: ", optim.out$iterations, "\n")
      cat(
        "  number of function evaluations [objective, gradient]: ",
        optim_out$counts, "\n"
      )
    }

    # iterations <- optim.out$iterations
    iterations <- optim_out$counts[1]
    x <- optim_out$par
    if (optim_out$convergence == 0L) {
      converged <- TRUE
    } else {
      converged <- FALSE
    }
  } else if (optimizer == "L.BFGS.B") {
    # warning, does not cope with Inf values!!

    control_lbfgsb <- list(
      trace = 0L, fnscale = 1,
      parscale = scale_1, ## or not?
      ndeps = 1e-3,
      maxit = 10000,
      REPORT = 1L,
      lmm = 5L,
      factr = 1e7,
      pgtol = 0
    )
    control_lbfgsb <- modifyList(control_lbfgsb, lavoptions$control)
    control <- control_lbfgsb[c(
      "trace", "fnscale", "parscale",
      "ndeps", "maxit", "REPORT", "lmm",
      "factr", "pgtol"
    )]
    optim_out <- optim(
      par = start_x,
      fn = objective_function,
      gr = gradient,
      method = "L-BFGS-B",
      lower = lower,
      upper = upper,
      control = control,
      hessian = FALSE,
      verbose = verbose, debug = debug,
      inf_to_max = TRUE
    )
    if (verbose) {
      cat("  convergence status (0=ok): ", optim_out$convergence, "\n")
      cat("  optim L-BFGS-B message says: ", optim_out$message, "\n")
      # cat("number of iterations: ", optim.out$iterations, "\n")
      cat(
        "  number of function evaluations [objective, gradient]: ",
        optim_out$counts, "\n"
      )
    }

    # iterations <- optim.out$iterations
    iterations <- optim_out$counts[1]
    x <- optim_out$par
    if (optim_out$convergence == 0L) {
      converged <- TRUE
    } else {
      converged <- FALSE
    }
  } else if (optimizer == "NLMINB.CONSTR") {
    ocontrol <- list(verbose = verbose)
    if (!is.null(lavoptions$control$control.outer)) {
      ocontrol <- c(lavoptions$control$control.outer, verbose = verbose)
    }
    control_nlminb <- list(
      eval.max = 20000L,
      iter.max = 10000L,
      trace = 0L,
      # abs.tol=1e-20,
      abs.tol = (.Machine$double.eps * 10),
      rel.tol = 1e-9, # 1e-10 seems 'too strict'
      step.min = 1.0, # 1.0 in < 0.5-21
      step.max = 1.0,
      x.tol = 1.5e-8,
      xf.tol = 2.2e-14
    )
    control_nlminb <- modifyList(control_nlminb, lavoptions$control)
    control <- control_nlminb[c(
      "eval.max", "iter.max", "trace",
      "abs.tol", "rel.tol"
    )]
    cin <- cin_jac <- ceq <- ceq_jac <- NULL
    if (!is.null(body(lavmodel@cin.function))) cin <- lavmodel@cin.function
    if (!is.null(body(lavmodel@cin.jacobian))) cin_jac <- lavmodel@cin.jacobian
    if (!is.null(body(lavmodel@ceq.function))) ceq <- lavmodel@ceq.function
    if (!is.null(body(lavmodel@ceq.jacobian))) ceq_jac <- lavmodel@ceq.jacobian
    trace <- FALSE
    if (verbose) trace <- TRUE
    optim_out <- nlminb_constr(
      start = start_x,
      objective = objective_function,
      gradient = gradient,
      control = control,
      scale = scale_1,
      verbose = verbose, debug = debug,
      lower = lower,
      upper = upper,
      cin = cin, cin_jac = cin_jac,
      ceq = ceq, ceq_jac = ceq_jac,
      control_outer = ocontrol
    )
    if (verbose) {
      cat("  convergence status (0=ok): ", optim_out$convergence, "\n")
      cat("  nlminb_constr message says: ", optim_out$message, "\n")
      cat("  number of outer iterations: ", optim_out$outer.iterations, "\n")
      cat("  number of inner iterations: ", optim_out$iterations, "\n")
      cat(
        "  number of function evaluations [objective, gradient]: ",
        optim_out$evaluations, "\n"
      )
    }

    iterations <- optim_out$iterations
    x <- optim_out$par
    if (optim_out$convergence == 0) {
      converged <- TRUE
    } else {
      converged <- FALSE
    }
  } else if (optimizer == "NONE") {
    x <- start_x
    iterations <- 0L
    converged <- TRUE
    control <- list()

    # if inequality constraints, add con.jac/lambda
    # needed for df!
    if (length(lavmodel@ceq.nonlinear.idx) == 0L &&
        (lavmodel@cin.simple.only ||
         (length(lavmodel@cin.linear.idx) == 0L &&
          length(lavmodel@cin.nonlinear.idx) == 0L))) {
      optim_out <- list()
    } else {
      # if inequality constraints, add con.jac/lambda
      # needed for df!

      optim_out <- list()
      if (is.null(body(lavmodel@ceq.function))) {
        ceq <- function(x, ...) {
          numeric(0)
        }
      } else {
        ceq <- lavmodel@ceq.function
      }
      if (is.null(body(lavmodel@cin.function))) {
        cin <- function(x, ...) {
          numeric(0)
        }
      } else {
        cin <- lavmodel@cin.function
      }
      ceq0 <- ceq(start_x)
      cin0 <- cin(start_x)
      con0 <- c(ceq0, cin0)
      jac <- rbind(
        numDeriv::jacobian(ceq, x = start_x),
        numDeriv::jacobian(cin, x = start_x)
      )
      nceq <- length(ceq(start_x))
      ncin <- length(cin(start_x))
      ncon <- nceq + ncin
      ceq_idx <- cin_idx <- integer(0)
      if (nceq > 0L) ceq_idx <- 1:nceq
      if (ncin > 0L) cin_idx <- nceq + 1:ncin
      cin_flag <- rep(FALSE, ncon)
      if (ncin > 0L) cin_flag[cin_idx] <- TRUE

      inactive_idx <- integer(0L)
      cin_idx <- which(cin_flag)
      if (ncin > 0L) {
        # optimizer == "NONE": no multiplier information is available for
        # the externally supplied solution, so a constraint that touches
        # the boundary is conservatively treated as binding (this only
        # feeds the df bookkeeping); no strict-complementarity test here
        slack <- 1e-05
        inactive_idx <- which(cin_flag & con0 > slack)
      }
      attr(jac, "inactive.idx") <- inactive_idx
      attr(jac, "cin.idx") <- cin_idx
      attr(jac, "ceq.idx") <- ceq_idx

      optim_out$con.jac <- jac
      optim_out$lambda <- rep(0, ncon)
    }
  }

  # new in 0.6-19
  # if NLMINB() + cin.simple.only, add con.jac and lambda to optim.out
  if (optimizer %in% c("NLMINB", "NLMINB0", "L.BFGS.B") &&
      (lavmodel@cin.simple.only || lavmodel@ceq.simple.only)) {

    if (lavmodel@cin.simple.only && nrow(lavmodel@cin.JAC) > 0L) {
      # JAC
      cin_jac_1 <- lavmodel@cin.JAC

      # lambda (post-hoc): the bound rows involve one parameter each, so
      # the rows of cin.JAC are mutually orthogonal and the least-squares
      # multiplier reduces to the plain inner product with the gradient.
      # NOTE: for ceq.simple models both the parameter vector and the
      # gradient live in the reduced (packed) space, while cin.function
      # and cin.JAC operate on the full (unco) space -- unpack BOTH
      # (evaluating cin.function on the packed x returned garbage
      # constraint values, silently corrupting the old classification)
      dx <- gradient(x)
      if (lavmodel@ceq.simple.only) {
        unpack_idx <- lavpartable$free[lavpartable$free > 0]
        x_unpack <- x[unpack_idx]
        dx_unpack <- dx[unpack_idx]
      } else if (lavmodel@eq.constraints) {
        # unreachable via the standard pipeline: eq.constraints is a
        # packing flag that is only TRUE when equality constraints are the
        # ONLY constraints, which contradicts cin.simple.only; kept as a
        # safety net
        x_unpack <- as.numeric(lavmodel@eq.constraints.K %*% x) +
          lavmodel@eq.constraints.k0
        dx_unpack <- as.numeric(lavmodel@eq.constraints.K %*% dx)
      } else {
        x_unpack <- x
        dx_unpack <- dx
      }
      con0 <- lavmodel@cin.function(x_unpack)
      cin_lambda <- drop(cin_jac_1 %*% dx_unpack)

      # strict complementarity: a bound is only ACTIVE (binding) when the
      # solution sits at the bound AND the multiplier exerts force on the
      # gradient (see lav_con_cin_inactive_idx()); an interior solution
      # that merely grazes a bound keeps its ordinary standard error
      inactive_idx <- lav_con_cin_inactive_idx(
        con0 = con0, lambda = cin_lambda, jac = cin_jac_1,
        cin_flag = rep(TRUE, nrow(cin_jac_1))
      )
      cin_lambda[inactive_idx] <- 0

      # remove all inactive rows
      #if (length(inactive.idx) > 0L) {
      #  cin.JAC <- cin.JAC[-inactive.idx, , drop = FALSE]
      #  cin.lambda <- cin.lambda[-inactive.idx]
      #  inactive.idx <- integer(0L)
      #}
    } else {
      npar <- length(lavpartable$free[lavpartable$free > 0])
      cin_jac_1 <- matrix(0, nrow = 0L, ncol = npar)
      inactive_idx <- integer(0L)
      cin_lambda <- numeric(0L)
    }

    if (lavmodel@ceq.simple.only && nrow(lavmodel@ceq.simple.K) > 0L) {
      ceq_jac_1 <- t(lav_mat_ortho_complement(lavmodel@ceq.simple.K))
      ceq_lambda <- numeric(nrow(ceq_jac_1))
    } else {
      npar <- length(lavpartable$free[lavpartable$free > 0])
      ceq_jac_1 <- matrix(0, nrow = 0L, ncol = npar)
      ceq_lambda <- numeric(0L)
    }

    # combine
    jac <- rbind(cin_jac_1, ceq_jac_1)
    attr(jac, "inactive.idx") <- inactive_idx
    attr(jac, "cin.idx") <- seq_len(nrow(cin_jac_1))
    attr(jac, "ceq.idx") <- nrow(cin_jac_1) + seq_len(nrow(ceq_jac_1))
    lambda <- c(cin_lambda, ceq_lambda)

    optim_out$con.jac <- jac
    optim_out$lambda <- lambda
  }


  fx <- objective_function(x) # to get "fx.group" attribute

  # check convergence
  warn_txt <- ""
  if (converged) {
    # check.gradient
    if (!is.null(gradient) &&
      optimizer %in% c("NLMINB", "BFGS", "L.BFGS.B")) {
      # compute unscaled gradient
      dx <- gradient(x)

      # NOTE: unscaled gradient!!!
      if (converged && lavoptions$check.gradient &&
        any(abs(dx) > lavoptions$optim.dx.tol)) {
        # ok, identify the non-zero elements
        non_zero <- which(abs(dx) > lavoptions$optim.dx.tol)

        # which ones are 'boundary' points, defined by lower/upper?
        bound_idx <- integer(0L)
        if (!is.null(lavpartable$lower)) {
          bound_idx <- c(bound_idx, which(lower == x))
        }
        if (!is.null(lavpartable$upper)) {
          bound_idx <- c(bound_idx, which(upper == x))
        }
        # parameters fixed at their h1 values (saturated blocks) were
        # not part of the search
        if (length(h1_sat_idx) > 0L) {
          bound_idx <- c(bound_idx, h1_sat_idx)
        }
        if (length(bound_idx) > 0L) {
          non_zero <- non_zero[-which(non_zero %in% bound_idx)]
        }

        # this has many implications ... so should be careful to
        # avoid false alarm
        if (length(non_zero) > 0L) {
          converged <- FALSE
          warn_txt <- paste("the optimizer (", optimizer, ") ",
            "claimed the model converged,\n",
            "       but not all elements of the gradient are (near) zero;\n",
            "       the optimizer may not have found a local solution\n",
            "       use check.gradient = FALSE to skip this check.",
            sep = ""
          )
        }
      }
    } else {
      dx <- numeric(0L)
    }
  } else {
    dx <- numeric(0L)
    warn_txt <- "the optimizer warns that a solution has NOT been found!"
  }

  # transform back
  # 3.
  # if(lavoptions$optim.var.transform == "sqrt" &&
  #       length(lavmodel@x.free.var.idx) > 0L) {
  #    #x[lavmodel@x.free.var.idx] <- tan(x[lavmodel@x.free.var.idx])
  #    x.var <- x[lavmodel@x.free.var.idx]
  #    x.var.sign <- sign(x.var)
  #    x[lavmodel@x.free.var.idx] <- x.var.sign * (x.var * x.var) # square!
  # }

  # 2. unpack
  if (lavmodel@eq.constraints) {
    x <- as.numeric(lavmodel@eq.constraints.K %*% x) +
      lavmodel@eq.constraints.k0
  }

  # 1. unscale
  x <- x / parscale

  attr(x, "converged") <- converged
  attr(x, "start") <- start_x
  attr(x, "warn.txt") <- warn_txt
  attr(x, "iterations") <- iterations
  attr(x, "control") <- control
  attr(x, "fx") <- fx
  attr(x, "dx") <- dx
  attr(x, "parscale") <- parscale
  if (!is.null(optim_out$con.jac)) attr(x, "con.jac") <- optim_out$con.jac
  if (!is.null(optim_out$lambda)) attr(x, "con.lambda") <- optim_out$lambda
  if (lavoptions$optim.partrace) {
    attr(x, "partrace") <- penv$PARTRACE
  }

  x
}

# backwards compatibility
# estimateModel <- lav_model_estimate

# new in 0.7-1: identify free parameters that belong to a 'saturated'
# block of a multilevel model
#
# a block is considered to be saturated if it only contains (free,
# unconstrained) variances/covariances -- and means/intercepts -- of the
# observed variables of that block; in that case, the 'estimated' values of
# these parameters are already available in lavh1$implied, and they do not
# need to enter the optimization again as free parameters: they are
# (temporarily) fixed at their h1 values during optimization only; once
# model estimation is over, they are treated as free parameters again
# (eg to compute standard errors)
#
# returns a list with two elements:
#   - x.idx: positions of these parameters in the (packed) parameter vector
#   - value: the corresponding h1 estimates
lav_model_est_h1_saturated <- function(lavmodel = NULL,
                                       lavpartable = NULL,
                                       lavdata = NULL,
                                       lavh1 = NULL) {
  empty <- list(x.idx = integer(0L), value = numeric(0L))

  # multilevel only (for now), and we need the h1 estimates
  if (lavdata@nlevels == 1L || length(lavh1) == 0L ||
      is.null(lavh1$implied$cov)) {
    return(empty)
  }

  # general linear equality constraints: the optimizer operates in a
  # reduced (packed) parameter space; we cannot fix individual parameters
  if (lavmodel@eq.constraints) {
    return(empty)
  }

  nlevels <- lavdata@nlevels
  ngroups <- lavdata@ngroups

  # position of each free parameter in the (packed) parameter vector
  free_id <- lavpartable$free[lavpartable$free > 0L]
  uid <- unique(free_id) # ceq.simple: duplicated ids are packed out
  dup_id <- unique(free_id[duplicated(free_id)])

  # labels that show up in (in)equality constraints
  con_idx <- which(lavpartable$op %in% c("==", "<", ">"))
  con_ref <- unique(c(lavpartable$lhs[con_idx], lavpartable$rhs[con_idx]))

  x_idx <- integer(0L)
  value <- numeric(0L)

  for (g in seq_len(ngroups)) {
    for (l in seq_len(nlevels)) {
      b <- (g - 1L) * nlevels + l
      ovn <- lavdata@ov.names.l[[g]][[l]]
      p <- length(ovn)
      cov_h1 <- lavh1$implied$cov[[b]]
      mean_h1 <- lavh1$implied$mean[[b]]
      if (p == 0L || is.null(cov_h1) || nrow(cov_h1) != p ||
          length(mean_h1) != p) {
        next
      }

      row_idx <- which(lavpartable$block == b)
      if (length(row_idx) == 0L) {
        next
      }
      op <- lavpartable$op[row_idx]
      lhs <- lavpartable$lhs[row_idx]
      rhs <- lavpartable$rhs[row_idx]
      free <- lavpartable$free[row_idx]

      # only variances/covariances and intercepts of observed variables
      # (no latent variables, no regressions)
      if (any(!op %in% c("~~", "~1")) ||
          any(!lhs %in% ovn) ||
          any(op == "~~" & !rhs %in% ovn)) {
        next
      }

      # all p*(p+1)/2 variances/covariances must be present and free
      cov_flag <- op == "~~"
      if (any(free[cov_flag] == 0L)) {
        next
      }
      i1 <- match(lhs[cov_flag], ovn)
      i2 <- match(rhs[cov_flag], ovn)
      pair_id <- pmin(i1, i2) + p * pmax(i1, i2)
      if (length(unique(pair_id)) != p * (p + 1L) / 2L) {
        next
      }

      # intercepts/means: all present; either free, or fixed at their h1
      # value (eg the zero within-level means of shared variables)
      int_flag <- op == "~1"
      if (lavmodel@meanstructure) {
        if (sum(int_flag) != p || any(duplicated(lhs[int_flag]))) {
          next
        }
        int_fixed <- which(int_flag & free == 0L)
        if (length(int_fixed) > 0L) {
          fixed_val <- lavpartable$start[row_idx[int_fixed]]
          h1_val <- mean_h1[match(lhs[int_fixed], ovn)]
          if (any(abs(fixed_val - h1_val) > 1e-10)) {
            next
          }
        }
      } else if (any(int_flag)) {
        next
      }

      # none of the free parameters may be involved in equality or
      # inequality constraints
      block_id <- free[free > 0L]
      if (any(block_id %in% dup_id)) {
        next
      }
      lab <- c(lavpartable$label[row_idx], lavpartable$plabel[row_idx])
      lab <- lab[nzchar(lab)]
      if (length(con_ref) > 0L && any(lab %in% con_ref)) {
        next
      }

      # this block is saturated: collect positions + h1 values
      b_free <- which(free > 0L)
      b_value <- numeric(length(b_free))
      for (i in seq_along(b_free)) {
        r <- b_free[i]
        if (op[r] == "~~") {
          b_value[i] <- cov_h1[match(lhs[r], ovn), match(rhs[r], ovn)]
        } else {
          b_value[i] <- mean_h1[match(lhs[r], ovn)]
        }
      }
      x_idx <- c(x_idx, match(free[b_free], uid))
      value <- c(value, b_value)
    } # levels
  } # groups

  list(x.idx = x_idx, value = value)
}

# return 'attributes' of a lavaan partable -- generate a new set if necessary
lav_pt_attributes <- function(partable, pta = NULL) {
  if (is.null(pta)) {
    # attached to partable?
    pta <- attributes(partable)
    if (!is.null(pta$vnames) && !is.null(pta$nvar)) {
      # looks like a pta
      pta$ovda <- NULL
      return(pta)
    } else {
      pta <- list()
    }
  }

  # vnames
  pta$vnames <- lav_pt_vnames(partable, type = "*")

  # vidx
  tmp_ov <- pta$vnames$ov
  tmp_lv <- pta$vnames$lv
  nblocks <- length(pta$vnames$ov)
  pta$vidx <- lapply(names(pta$vnames), function(v) {
    lapply(seq_len(nblocks), function(b) {
      if (v == "lv.marker") {
        match(pta$vnames[[v]][[b]], tmp_ov[[b]])
      } else if (grepl("lv", v, fixed = TRUE)) {
        match(pta$vnames[[v]][[b]], tmp_lv[[b]])
      } else if (grepl("th", v, fixed = TRUE)) {
        # thresholds have '|t' pattern
        tmp_th <- sapply(strsplit(pta$vnames[[v]][[b]],
          "|t",
          fixed = TRUE
        ), "[[", 1L)
        match(tmp_th, tmp_ov[[b]])
      } else if (grepl("eqs", v, fixed = TRUE)) {
        # mixture of tmp.ov/tmp.lv
        integer(0L)
      } else {
        match(pta$vnames[[v]][[b]], tmp_ov[[b]])
      }
    })
  })
  names(pta$vidx) <- names(pta$vnames)

  # meanstructure
  pta$meanstructure <- any(partable$op == "~1")

  # nblocks
  pta$nblocks <- nblocks

  # ngroups
  pta$ngroups <- lav_pt_ngroups(partable)

  # nlevels
  pta$nlevels <- lav_pt_nlevels(partable)

  # nvar
  pta$nvar <- lapply(pta$vnames$ov, length)

  # nfac
  pta$nfac <- lapply(pta$vnames$lv, length)

  # nfac.nonnormal - for numerical integration
  pta$nfac.nonnormal <- lapply(pta$vnames$lv.nonnormal, length)

  # th.idx (new in 0.6-1)
  pta$th.idx <- lapply(seq_len(pta$nblocks), function(b) {
    out <- numeric(length(pta$vnames$th.mean[[b]]))
    idx <- (pta$vnames$th.mean[[b]] %in%
      pta$vnames$th[[b]])
    out[idx] <- pta$vidx$th[[b]]
    out
  })

  pta
}

lav_step10_cache <- function(slot_cache = NULL,
                                    lavdata = NULL,
                                    lavmodel = NULL,
                                    lavpartable = NULL,
                                    lavoptions = NULL,
                                    sampling_weights = NULL) {
  # # # # # # # # # # #
  # #  10. lavcache # #
  # # # # # # # # # # #

  # if slot_cache not NULL
  #   copy to lavcache
  # else
  #   lavcache = list of length lavdata@ngroups
  #   set tmp.ov.types = lavdata$ov$types
  #   if lavmodel@conditional.x and sum(lavmodel@nexo) > 0L remove elements
  #     lavpta$vidx$ov.x from tmp.ov.types
  #   if lavoptions$estimator == "PML" and all tmp.ov.types are "ordered"
  #     th = lav_model_th(lavmodel)
  #     bi = lav_tables_pairwise_freq_cells(lavdata)
  #     if lavoptions$missing is "available.cases" or "doubly.robust"
  #       uni = lav_tables_uni_freq_cell(lavdata)
  #       if lavoptions$missing is "doubly.robust"
  #         if lavoptions$control$pairwiseProbGivObs NULL: *** error ***
  #         if lavoptions$control$univariateProbGivObs NULL: *** error ***
  #     for all groups (1:lavdata@ngroups)
  #       set tmp.idx = 1:length(bi$ibs.freq)
  #       if bi$group not NULL and max(bi$group) > 1L set tmp.idx = indexes
  #         for this group in bi
  #       set bifreq = bi$obs.freq[tmp.idx]
  #       set binobs = bi$nobs[tmp.idx]
  #       set long = lav_pml_longvec_ind(no.x = ncol(lavdata@X[[g]]),
  #                             all.thres = th[[g]],
  #                             index.var.of.thres = lavmodel@th.idx[[g]])
  #       set lavcache[[g]] = list(bifreq = bifreq, nobs = binobs, long = long)
  #       if sampling.weights not NULL
  #         compute (for group g) lavcache[[g]]$sum_obs_weights_xixj_ab_vec (*)
  #       if lavoptions$missing is "available.cases" or "doubly.robust"
  #         set tmp.idx = 1:length(bi$ibs.freq)
  #         if bi$group not NULL and max(bi$group) > 1L set tmp.idx = indexes
  #           for this group in bi
  #         set lavcache[[g]]$unifreq = unifreq = uni$obs.freq[tmp.idx]
  #         set lavcache[[g]]$uninobs = uninobs = uni$nobs[tmp.idx]
  #         set lavcache[[g]]$uniweights.casewise = uniweights.casewise =
  #           rowSums(is.na(lavdata@X[[g]]))
  #         compute lavcache[[g]]$uniweights (*)
  #       if lavoptions$missing is "doubly.robust"
  #         lavcache[[g]]$pairwiseProbGivObs =
  #           lavoptions$control$pairwiseProbGivObs[[g]]
  #         lavcache[[g]]$univariateProbGivObs =
  #           lavoptions$control$univariateProbGivObs[[g]]
  #         compute members idx.y1, idx.gy2, idx.cat.y1, idx.cat.gy2 and
  #           id.uniPrGivObs from
  #           lavchache[[g]] (*)
  #   if lavdata$data.type is "full" and lavdata@Rp[[1L]] not NULL
  #     copy lavdata@Rp[[g]]$pat to lavcache[[g]]$pat for all groups g
  # if lavoptions$estimator is "MML"
  #   compute for all groups g lavcache[[g]]$GH via
  #     lav_integration_gauss_hermite
  #
  # (*) !!! computations too complicated to summarize here !!!

  if (!is.null(slot_cache)) {
    lavcache <- slot_cache
  } else {
    # prepare cache -- stuff needed for estimation, but also post-estimation
    lavcache <- vector("list", length = lavdata@ngroups)

    # ov.types? (for PML check)
    tmp_ov_types <- lavdata@ov$type
    if (lavmodel@conditional.x && sum(lavmodel@nexo) > 0L) {
      # remove ov.x
      tmp_ov_x_idx <- unlist(attr(lavpartable, "vidx")$ov.x)
      tmp_ov_types <- tmp_ov_types[-tmp_ov_x_idx]
    }

    if (lavoptions$estimator == "PML" && all(tmp_ov_types == "ordered")) {
      th <- lav_model_th(lavmodel)
      bi <- lav_tables_pairwise_freq_cell(lavdata)

      # handle option missing = "available.cases" or "doubly.robust"
      if (lavoptions$missing == "available.cases" ||
        lavoptions$missing == "doubly.robust") {
        uni <- lav_tables_uni_freq_cell(lavdata)
        # checks for missing = "double.robust"
        if (lavoptions$missing == "doubly.robust") {
          # check whether the probabilities pairwiseProbGivObs and
          # univariateProbGivObs are given by the user
          if (is.null(lavoptions$control$pairwiseProbGivObs)) {
            lav_msg_stop(gettext(
              "could not find `pairwiseProbGivObs' in control() list"))
          }
          if (is.null(lavoptions$control$univariateProbGivObs)) {
            lav_msg_stop(gettext(
              "could not find `univariateProbGivObs' in control() list"))
          }
        }
      }

      for (g in 1:lavdata@ngroups) {
        if (is.null(bi$group) || max(bi$group) == 1L) {
          bifreq <- bi$obs.freq
          binobs <- bi$nobs
        } else {
          idx <- which(bi$group == g)
          bifreq <- bi$obs.freq[idx]
          binobs <- bi$nobs[idx]
        }
        long <- lav_pml_longvec_ind(
          no_x = ncol(lavdata@X[[g]]),
          all_thres = th[[g]],
          index_var_of_thres = lavmodel@th.idx[[g]]
        )
        lavcache[[g]] <- list(
          bifreq = bifreq,
          nobs = binobs,
          long = long
        )

        # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

        # I need to add something that splits weights into g groups so
        # adjust what follows in the new code also compute the sum of
        # weights within a group, this will substitute n_g (group size)
        # of simple random sampling (SRS) and also compute the total the
        # total sum of weights over all observation over all groups,
        # this substitutes the total sample size of SRS.

        if (!is.null(sampling_weights)) {
          # Keep track of indices of the response categories (a,b) of a
          # pair of ordinal variables (xi,xj) appearing in the data as
          # well as the index of the pair.
          idx_ab_of_xixj_ab <- lapply(long[c(1:2, 5)], function(x) {
            x[(long$index.thres.var1.of.pair != 0) &
              (long$index.thres.var2.of.pair != 0)]
          })
          names(idx_ab_of_xixj_ab) <- c("idx_a", "idx_b", "idx_pairs")
          lavcache[[g]]$idx_ab_of_xixj_ab <- idx_ab_of_xixj_ab

          # Raw data for group g
          x_g <- lavdata@X[[g]]

          # I assume that X.g includes only the ordinal indicators nvar
          # gives the number of ordinal indicators
          nvar <- ncol(x_g)

          # pstar gives the number of pairs formed by the nvar ordinal
          # indicators
          pstar <- nvar * (nvar - 1) / 2

          # Keep track of the indices of variables forming each pair
          idx_vars_in_pair <- combn(nvar, 2)

          # The output of sapply below provides the sum of weights for
          # all bivariate response pattern for all pairs of indicators.

          # If all indicators have the same number of response
          # categories, the output of sapply function below is a matrix.
          # Each column refers to a different pair of indicators (i,j)
          # with j running faster than i, e.g. (1,2) (1,3) (2,3). Within
          # each column, each element (i.e. each row of the matrix)
          # refers to a different combination of response categories
          # (a,b) with a, the category index of indicator i, running
          # faster than b, the category index of indicator j, e.g.
          # (1,1), (2,1) (3,1) (1,2) (2,2) (3,2)

          # If the indicators have different number of response
          # categories, the output of sapply function below is a list.
          # Each element of the list refers to a different pair of
          # indicators (i,j) with j running faster than i and it is a
          # matrix with number of rows the number of response categories
          # of indicator i and ncol =  the number of response categories
          # of indicator j.

          sum_obs_weights_xixj_ab <- sapply(1:pstar, function(x) {
            tmp_idx_ab <- lapply(idx_ab_of_xixj_ab, function(y) {
              y[idx_ab_of_xixj_ab$idx_pairs == x]
            })
            tmp_idx_cols <- idx_vars_in_pair[, x]
            tmp_var1 <- factor(x_g[, tmp_idx_cols[1]],
              levels =
                as.character(unique(tmp_idx_ab$idx_a))
            )
            tmp_var2 <- factor(x_g[, tmp_idx_cols[2]],
              levels =
                as.character(unique(tmp_idx_ab$idx_b))
            )
            tapply(
              X = lavdata@weights[[g]],
              INDEX = list(tmp_var1, tmp_var2),
              FUN = sum
            )
          })

          # We need to transform the output of sapply into a vector
          # where the sum of weights (for all bivariate response
          # patterns for all pairs of indicators) are listed in the same
          # order as in pairwisePI vector, i.e. a runs the fastest,
          # followed by b, then by j and lastly by i.

          if (is.matrix(sum_obs_weights_xixj_ab)) {
            sum_obs_weights_xixj_ab_vec <- c(sum_obs_weights_xixj_ab)
          } else if (is.list(sum_obs_weights_xixj_ab)) {
            sum_obs_weights_xixj_ab_vec <-
              do.call(c, sum_obs_weights_xixj_ab)
          }

          # Note that sapply gives NA for these bivariate response
          # patterns which are not observed at all. Substitute NA with
          # 0.
          idx_na_sowxav <- is.na(sum_obs_weights_xixj_ab_vec)
          if (any(idx_na_sowxav)) {
            sum_obs_weights_xixj_ab_vec[idx_na_sowxav] <- 0
          }

          lavcache[[g]]$sum_obs_weights_xixj_ab_vec <-
            sum_obs_weights_xixj_ab_vec
        }

        # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

        # available cases
        if (lavoptions$missing == "available.cases" ||
          lavoptions$missing == "doubly.robust") {
          if (is.null(uni$group) || max(uni$group) == 1L) {
            unifreq <- uni$obs.freq
            uninobs <- uni$nobs
          } else {
            idx <- which(uni$group == g)
            unifreq <- uni$obs.freq[idx]
            uninobs <- uni$nobs[idx]
          }
          lavcache[[g]]$unifreq <- unifreq
          lavcache[[g]]$uninobs <- uninobs

          uniweights_casewise <- rowSums(is.na(lavdata@X[[g]]))
          lavcache[[g]]$uniweights.casewise <- uniweights_casewise

          # weights per response category per variable in the same
          # order as unifreq; i.e. w_ia, i = 1,...,p, (p variables),
          # a = 1,...,Ci, (Ci response categories for variable i),
          # a running faster than i
          tmp_uniweights <- apply(
            lavdata@X[[g]], 2,
            function(x) {
              tapply(uniweights_casewise, as.factor(x), sum,
                na.rm = TRUE
              )
            }
          )
          if (is.matrix(tmp_uniweights)) {
            lavcache[[g]]$uniweights <- c(tmp_uniweights)
          }
          if (is.list(tmp_uniweights)) {
            lavcache[[g]]$uniweights <- unlist(tmp_uniweights)
          }
        } # "available.cases" or "double.robust"

        # doubly.robust only
        if (lavoptions$missing == "doubly.robust") {
          # add the provided by the user probabilities
          # pairwiseProbGivObs and univariateProbGivObs in Cache
          lavcache[[g]]$pairwiseProbGivObs <-
            lavoptions$control$pairwiseProbGivObs[[g]]
          lavcache[[g]]$univariateProbGivObs <-
            lavoptions$control$univariateProbGivObs[[g]]
          # compute different indices vectors that will help to do
          # calculations
          ind_vec <- as.data.frame(long[1:5])
          ind_vec <-
            ind_vec[((ind_vec$index.thres.var1.of.pair != 0) &
              (ind_vec$index.thres.var2.of.pair != 0)), ]
          idx_cat_y1 <- ind_vec$index.thres.var1.of.pair
          idx_cat_y2 <- ind_vec$index.thres.var2.of.pair
          idx_pairs <- ind_vec$index.pairs.extended
          lavcache[[g]]$idx.pairs <- idx_pairs

          idx_cat_y1_split <- split(idx_cat_y1, idx_pairs)
          idx_cat_y2_split <- split(idx_cat_y2, idx_pairs)
          lavcache[[g]]$idx.cat.y1.split <- idx_cat_y1_split
          lavcache[[g]]$idx.cat.y2.split <- idx_cat_y2_split

          # generate the variables, categories indices vector which
          # keep track to which variables and categories the
          # elements of vector probY1Gy2 refer to
          nlev <- lavdata@ov$nlev
          nvar <- length(nlev)

          idx_var_matrix <- matrix(1:nvar, nrow = nvar, ncol = nvar)
          idx_diag <- diag(matrix(1:(nvar * nvar),
            nrow = nvar,
            ncol = nvar
          ))
          idx_y1gy2_matrix <- rbind(
            t(idx_var_matrix)[-idx_diag],
            idx_var_matrix[-idx_diag]
          )
          no_pairs_y1gy2 <- ncol(idx_y1gy2_matrix)
          idx_cat_y1 <- unlist(lapply(1:no_pairs_y1gy2, function(x) {
            rep(1:nlev[idx_y1gy2_matrix[1, x]],
              times = nlev[idx_y1gy2_matrix[2, x]]
            )
          }))
          idx_cat_gy2 <- unlist(lapply(1:no_pairs_y1gy2, function(x) {
            rep(1:nlev[idx_y1gy2_matrix[2, x]],
              each = nlev[idx_y1gy2_matrix[1, x]]
            )
          }))
          dim_pairs <- unlist(lapply(1:no_pairs_y1gy2, function(x) {
            nlev[idx_y1gy2_matrix[1, x]] *
              nlev[idx_y1gy2_matrix[2, x]]
          }))
          idx_y1 <- unlist(mapply(rep, idx_y1gy2_matrix[1, ],
            each = dim_pairs
          ))
          idx_gy2 <- unlist(mapply(rep, idx_y1gy2_matrix[2, ],
            each = dim_pairs
          ))

          lavcache[[g]]$idx.Y1 <- idx_y1
          lavcache[[g]]$idx.Gy2 <- idx_gy2
          lavcache[[g]]$idx.cat.Y1 <- idx_cat_y1
          lavcache[[g]]$idx.cat.Gy2 <- idx_cat_gy2

          # the vector below keeps track of the variable each column
          # of the matrix univariateProbGivObs refers to
          lavcache[[g]]$id.uniPrGivObs <-
            sort(c(
              unique(lavmodel@th.idx[[g]]),
              lavmodel@th.idx[[g]]
            ))
        } # doubly.robust
      } # g
    }
    # copy response patterns to cache -- FIXME!! (data not included
    # in Model only functions)
    if (lavdata@data.type == "full" && !is.null(lavdata@Rp[[1L]])) {
      for (g in 1:lavdata@ngroups) {
        lavcache[[g]]$pat <- lavdata@Rp[[g]]$pat
      }
    }
  }

  # random slopes (rv() modifier): store the index map and the
  # per-cluster crossproduct statistics (new in 0.7-1)
  if (isTRUE(lavoptions$.rv) ||
      length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L) {
    if (lavdata@data.type != "full") {
      lav_msg_stop(gettext(
        "random slopes (rv() modifier) require full (raw) data."))
    }
    for (g in 1:lavdata@ngroups) {
      if (is.null(lavcache[[g]]$rs)) {
        rs_info <- lav_mvn_cl_rs_info(
          lavmodel = lavmodel, lavpartable = lavpartable,
          lavdata = lavdata, lavoptions = lavoptions
        )
        rs_stats <- lav_mvn_cl_rs_stats(
          y1 = lavdata@X[[g]], lp = lavdata@Lp[[g]],
          rs_info = rs_info
        )
        lavcache[[g]]$rs <- list(info = rs_info, stats = rs_stats)
      }
    }
  }

  # If estimator = MML, store Gauss-Hermite nodes/weights
  if (lavoptions$estimator == "MML") {
    for (g in 1:lavdata@ngroups) {
      # count only the ones with non-normal indicators
      # nfac <- lavpta$nfac.nonnormal[[g]]
      nfac <- attr(lavpartable, "nfac")[[g]]
      lavcache[[g]]$GH <-
        lav_integration_gauss_hermite(
          n = lavoptions$integration.ngh,
          dnorm = TRUE,
          mean = 0, sd = 1,
          ndim = nfac
        )
      # lavcache[[g]]$DD <- lav_model_grad_dd(lavmodel, group = g)
    }
  }

  lavcache
}

# two-level ML estimation with random slopes (the rv() modifier)
#
# YR July 2026 (with help from Claude)
#
# Reference:
# - Rockwood, N. J. (2020). Maximum likelihood estimation of multilevel
#   structural equation models with random slopes for latent covariates.
#   Psychometrika, 85(2), 275-300.
#
# Two settings are implemented:
#
# 1) all random slopes involve *observed* level-1 covariates: all random
#    effects are 'linear' (Rockwood, 2020, section 3.2) and the marginal
#    (observed-data) loglikelihood is available in closed form (no
#    numerical integration);
#
# 2) one or more random slopes involve a *latent* level-1 covariate:
#    these slopes are 'nonlinear' random effects; conditional on their
#    values, the model reduces to setting (1), so the loglikelihood is
#    obtained by (Gauss-Hermite) quadrature over the nonlinear slopes
#    only, while all linear random effects are still integrated out
#    analytically (Rockwood, 2020, sections 3.3-3.6); see the
#    'quadrature' section below (lav_mvn_cl_rs_loglik_nl).
#
# NOTE: the likelihood is *conditional* on the covariates that carry random
# slopes (and on the between-level exogenous covariates): their joint
# distribution with the outcomes is no longer normal.
#
# Missing data (missing = "ml") is handled throughout: the per-case
# weight matrix becomes pattern-specific (see lav_mvn_cl_rs_loglik_m),
# the E-step additionally completes the missing y-values (see
# lav_mvn_cl_rs_estep_m), and lavPredict() conditions on the observed
# coordinates only.
#
# The model for cluster j (complete data, single group):
#
#   v_b = (y_b, z) ~ N(mu_v(w_j), Sigma_v)     [between; z = random slopes]
#   y_ij | v_b ~ N(mu_w + P x_ij + Q_ij v_b, Sigma_w),   independent over i
#
# where
#   - y_b     = between components of the 'both-level' variables (pb)
#   - z       = the random slopes, latent variables at level 2 (q)
#   - w_j     = between-level exogenous covariates (conditioned upon)
#   - x_ij    = level-1 covariates with random slopes (conditioned upon)
#   - Sigma_w = within-implied covariance of the level-1 y-vector (the
#               random-slope cells in GLIST are zero, so the standard
#               LISREL formula gives the *conditional* covariance)
#   - P       = reduced-form *fixed* slopes of y on x (from the within GLIST)
#   - Q_ij    = [ E_b , L_1 x_(1)ij , ..., L_q x_(q)ij ] with
#               L_r = (Lambda_w (I - B_w)^{-1})[, lhs(r)]
#
# Because Q_ij is linear in x_ij, all cluster-wise sums that appear in the
# loglikelihood (and in the EM E-step) reduce to per-cluster crossproducts
# of (y, x) -- computed once (lav_mvn_cl_rs_stats), so that each evaluation
# of the loglikelihood costs O(nclusters * (p + q)^3), without per-case
# loops. (cfr. the 'partial sufficient statistics' of Mplus.)


# collect and validate all static information needed for the random-slope
# kernels; returns an index map ('rs_info') that is computed only once
# (and cached in lavcache[[g]]$rs)
lav_mvn_cl_rs_info <- function(lavmodel = NULL, lavpartable = NULL,
                               lavdata = NULL, lavoptions = NULL) {
  # single group only (for now)
  if (lavdata@ngroups > 1L) {
    lav_msg_stop(gettext(
      "random slopes (rv() modifier) are not supported (yet) for
       multiple-group models."))
  }
  if (lavdata@nlevels != 2L) {
    lav_msg_stop(gettext(
      "random slopes (rv() modifier) require a two-level model."))
  }

  # number of quadrature nodes per dimension (latent covariates only)
  ngh <- 21L
  if (!is.null(lavoptions) && !is.null(lavoptions$integration.ngh)) {
    ngh <- as.integer(lavoptions$integration.ngh)
  }

  # path list: from the parameter table (with validation), or from
  # lavmodel@rv.ov/@rv.lv (no validation; used post-estimation)
  if (!is.null(lavpartable)) {
    pt <- lavpartable
    if (is.null(pt$level)) {
      lav_msg_stop(gettext("parameter table has no level column."))
    }

    # the random-slope paths at level 1
    # (the level column may hold user-chosen labels instead of 1L/2L)
    level_values <- lav_pt_level_values(pt)
    rs_idx <- which(nchar(pt$rv) > 0L & pt$op == "~" &
                    pt$level == level_values[1])
    # random slopes at level 2 are not allowed
    bad_idx <- which(nchar(pt$rv) > 0L & pt$op == "~" &
                     pt$level %in% level_values[-1L])
    if (length(bad_idx) > 0L) {
      lav_msg_stop(gettext(
        "random slopes (rv() modifier) can only be used in the level-1
         (within) part of the model."))
    }
    if (length(rs_idx) == 0L) {
      lav_msg_stop(gettext("no random-slope (rv) regressions found at
                            level 1."))
    }

    path_rv <- pt$rv[rs_idx]
    path_lhs <- pt$lhs[rs_idx]
    path_rhs <- pt$rhs[rs_idx]

    # covariate type per path: latent ("lv") or observed ("ov")
    # (this mirrors the rv.lv/rv.ov classification in lav_model();
    #  'split' both-level observed covariates count as latent: the
    #  random slope multiplies their latent within-cluster component)
    lv_names_1 <- lav_pt_vnames(pt, "lv", block = 1L)
    ov_names_pt_1 <- lav_pt_vnames(pt, "ov", block = 1L)
    ov_names_pt_2 <- lav_pt_vnames(pt, "ov", block = 2L)
    ov_12 <- ov_names_pt_2[ov_names_pt_2 %in% ov_names_pt_1]
    path_type <- ifelse(path_rhs %in% c(lv_names_1, ov_12),
                        "lv", "ov")
  } else {
    pt <- NULL
    rv_all <- c(lavmodel@rv.ov, lavmodel@rv.lv)
    path_rv <- names(rv_all)
    path_lhs <- sapply(rv_all, "[[", 1L)
    path_rhs <- sapply(rv_all, "[[", 2L)
    path_type <- rep(c("ov", "lv"),
                     c(length(lavmodel@rv.ov), length(lavmodel@rv.lv)))
  }

  z_names <- unique(path_rv) # canonical order of the random slopes
  q <- length(z_names)

  # slope type: a slope multiplying a latent covariate is a 'nonlinear'
  # random effect (numerical integration); a single label may not mix
  # covariate types
  z_type <- character(q)
  for (r in seq_len(q)) {
    tps <- unique(path_type[path_rv == z_names[r]])
    if (length(tps) > 1L) {
      lav_msg_stop(gettextf(
        "random-slope label %s is attached to both an observed and a
         latent covariate; this is not supported (yet).", z_names[r]))
    }
    z_type[r] <- tps
  }
  nl_z_idx <- which(z_type == "lv")
  nl_flag <- length(nl_z_idx) > 0L

  # variable names per level
  ov_names_1 <- lavdata@ov.names.l[[1]][[1]] # within
  ov_names_2 <- lavdata@ov.names.l[[1]][[2]] # between
  both_names <- ov_names_1[ov_names_1 %in% ov_names_2]
  within_only <- ov_names_1[!ov_names_1 %in% ov_names_2]
  between_only <- ov_names_2[!ov_names_2 %in% ov_names_1]

  # the observed covariates that are conditioned upon: the (observed)
  # rv covariates PLUS all other within-only exogenous (fixed.x)
  # covariates -- the loglikelihood is conditional on ALL of them
  # (cfr. Mplus TYPE = TWOLEVEL RANDOM, which conditions on all
  # observed x-covariates); the fixed regression paths of the
  # non-rv covariates enter through the reduced-form P matrix
  x_rv <- unique(path_rhs[path_type == "ov"])
  x_extra <- within_only[within_only %in% lavdata@ov.names.x[[1]] &
                         !within_only %in% x_rv]
  x_names <- c(x_rv, x_extra)

  # validation (only if we have a parameter table)
  if (!is.null(pt)) {
    # note on 'split' both-level covariates (rhs in ov_12): the random
    # slope multiplies the *latent* within-cluster component of the
    # covariate (latent decomposition/centering, cfr. Rockwood's
    # simulation model); the covariate is then jointly modeled: it is
    # part of the y-vector, its within dummy lv is the beta_w
    # injection target, and its between dummy lv is an ordinary linear
    # random effect inside v_b -- no special handling needed below

    # exogenous variables (fixed.x)
    ov_x_1 <- lav_pt_vnames(pt, "ov.x", block = 1L)
    ov_x_2 <- lav_pt_vnames(pt, "ov.x", block = 2L)

    # the observed rv covariates must be observed, within-only, and
    # exogenous
    bad <- x_rv[!x_rv %in% within_only]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "covariate(s) with a random slope must be observed within-only
         variables (they should not appear in the level-2 part of the
         model): %s.", paste(bad, collapse = " ")))
    }
    bad <- x_rv[!x_rv %in% ov_x_1]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "covariate(s) with a random slope must be exogenous (fixed.x)
         variables: %s.", paste(bad, collapse = " ")))
    }

    # the lhs of each random-slope path must be a level-1 dependent
    # variable
    bad <- path_lhs[!(path_lhs %in% lv_names_1 |
                      path_lhs %in% ov_names_1)]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "left-hand side of random-slope regression(s) not found among the
         level-1 (latent or observed) variables: %s.",
        paste(bad, collapse = " ")))
    }
    bad <- path_lhs[path_lhs %in% x_names]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "left-hand side of random-slope regression(s) cannot be an
         exogenous covariate: %s.", paste(bad, collapse = " ")))
    }

    # the random slopes must be 'pure' latent variables at level 2:
    # no (real) indicators
    bad_idx <- which(pt$op == "=~" & pt$lhs %in% z_names &
                     pt$rhs != pt$lhs)
    if (length(bad_idx) > 0L) {
      lav_msg_stop(gettextf(
        "random slope(s) cannot have indicators at level 2: %s.",
        paste(unique(pt$lhs[bad_idx]), collapse = " ")))
    }

  }

  # the within y-vector: all within variables, except the rv covariates
  y_names <- ov_names_1[!ov_names_1 %in% x_names]
  p1 <- length(y_names)

  # path table
  path_tab <- data.frame(
    rv = path_rv,
    lhs = path_lhs,
    rhs = path_rhs,
    type = path_type,
    z.idx = match(path_rv, z_names),
    x.idx = match(path_rhs, x_names), # NA for latent covariates
    stringsAsFactors = FALSE
  )

  # data column indices (in lavdata@X[[1]], columns = lavdata@ov.names[[1]])
  ov_names_data <- lavdata@ov.names[[1]]
  y_data_idx <- match(y_names, ov_names_data)
  x_data_idx <- match(x_names, ov_names_data)

  # between-only variables: exogenous covariates (w; conditioned
  # upon) vs *endogenous* z-outcomes (Rockwood's z-variables: observed
  # once per cluster, and modeled -- e.g., indicators of a
  # between-level factor)
  mm_b0 <- 1:lavmodel@nmat[2] + lavmodel@nmat[1]
  glist_names_b0 <- names(lavmodel@GLIST)[mm_b0]
  lambda_mm_b <- mm_b0[which(glist_names_b0 == "lambda")]
  lv_names_b_all <- lavmodel@dimNames[[lambda_mm_b]][[2]]
  if (!is.null(pt)) {
    exo_b_names <- between_only[between_only %in% ov_x_2]
  } else {
    # without a parameter table: the exogenous covariates appear as
    # dummy latent variables in the between block; the z-outcomes
    # (pure indicators) do not
    exo_b_names <- between_only[between_only %in% lv_names_b_all]
  }
  zb_names <- between_only[!between_only %in% exo_b_names]
  kz <- length(zb_names)
  exo_b_data_idx <- match(exo_b_names, ov_names_data)
  zb_data_idx <- match(zb_names, ov_names_data)

  # z-outcomes must be pure indicators: a z-variable that is itself
  # involved in a regression gets a dummy latent variable (and a
  # zero residual variance), which does not fit the 'extra
  # observation block' treatment below
  bad <- zb_names[zb_names %in% lv_names_b_all]
  if (length(bad) > 0L) {
    lav_msg_stop(gettextf(
      "between-only endogenous observed variables can only appear as
       indicators of a between-level latent variable (for now); not
       as outcome or predictor in a regression: %s.",
      paste(bad, collapse = " ")))
  }

  # the both-level variables
  yb_names <- both_names
  pb <- length(yb_names)
  # position of the y_b variables within the y-vector
  yb_y_idx <- match(yb_names, y_names)

  # the between random vector v_b = (eta_b, eps_b) lives in 'factor'
  # space: the between latent variables (excluding the dummy lv's of the
  # between-level exogenous covariates, which are conditioned out),
  # followed by the between residuals (eps_b) of the both-level
  # variables -- but only those with a free (or fixed to a nonzero
  # value) residual (co)variance. This keeps v_b minimal, and Sigma_v
  # positive definite even when (some) between residual variances are
  # fixed to zero (as in Mplus ex9.8).
  mm_b <- 1:lavmodel@nmat[2] + lavmodel@nmat[1]
  glist_names_b <- names(lavmodel@GLIST)[mm_b]
  theta_mm <- mm_b[which(glist_names_b == "theta")]
  theta_b <- lavmodel@GLIST[[theta_mm]]
  free_mat <- matrix(FALSE, nrow(theta_b), ncol(theta_b))
  free_mat[lavmodel@m.free.idx[[theta_mm]]] <- TRUE
  nonzero_mat <- (theta_b != 0) | free_mat
  ov_names_b_theta <- lavmodel@dimNames[[theta_mm]][[1]]
  yb_b_idx <- match(yb_names, ov_names_b_theta)
  eps_flag <- apply(nonzero_mat[yb_b_idx, , drop = FALSE], 1L, any)
  eps_names <- yb_names[eps_flag]

  # z-outcomes: the residual block Sigma_z must be positive definite
  # (all diagonal elements free or fixed to a nonzero value), and may
  # not covary with the between residuals of the both-level variables
  if (kz > 0L) {
    zb_b_idx <- match(zb_names, ov_names_b_theta)
    bad <- zb_names[!nonzero_mat[cbind(zb_b_idx, zb_b_idx)]]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "between-only endogenous observed variables must have a free
         (or fixed to a nonzero value) residual variance: %s.",
        paste(bad, collapse = " ")))
    }
    if (pb > 0L &&
        any(nonzero_mat[zb_b_idx, yb_b_idx, drop = FALSE])) {
      lav_msg_stop(gettext(
        "residual covariances between between-only endogenous
         observed variables and the both-level variables are not
         supported (yet)."))
    }
  }

  # structural fixed paths from the rv covariates? (a nonzero 'P'
  # matrix); the EM version cannot handle these (the per-case mean
  # P x_{ji} does not fit in the two-group complete-data model);
  # nlminb handles the general case
  mm_w <- 1:lavmodel@nmat[1]
  glist_names_w <- names(lavmodel@GLIST)[mm_w]
  beta_mm <- mm_w[which(glist_names_w == "beta")]
  p_flag <- FALSE
  if (length(beta_mm) > 0L) {
    beta_w <- lavmodel@GLIST[[beta_mm]]
    free_mat_w <- matrix(FALSE, nrow(beta_w), ncol(beta_w))
    free_mat_w[lavmodel@m.free.idx[[beta_mm]]] <- TRUE
    lv_names_w <- lavmodel@dimNames[[beta_mm]][[2]]
    x_lv_idx_w <- match(x_names, lv_names_w)
    x_lv_idx_w <- x_lv_idx_w[!is.na(x_lv_idx_w)]
    if (length(x_lv_idx_w) > 0L) {
      p_flag <- any(free_mat_w[, x_lv_idx_w]) ||
        any(beta_w[, x_lv_idx_w] != 0)
    }
  }

  # shared rv labels (the same slope attached to more than one path)?
  # the EM version requires a one-to-one path <-> slope mapping
  shared_flag <- (nrow(path_tab) != q) ||
    !identical(path_tab$z.idx, seq_len(q))

  # latent covariates: GLIST cell positions of the (fixed-to-zero)
  # within regression paths; the quadrature driver injects the node
  # values for the nonlinear slopes there
  nl_beta_row <- nl_beta_col <- integer(0L)
  nl_beta_mm <- integer(0L)
  if (nl_flag) {
    if (length(beta_mm) == 0L) {
      lav_msg_stop(gettext(
        "no beta matrix found in the within part of the model
         (random slopes for latent covariates)."))
    }
    lv_names_w_all <- lavmodel@dimNames[[beta_mm]][[2]]
    nl_path_idx <- which(path_type == "lv")
    nl_beta_row <- match(path_lhs[nl_path_idx], lv_names_w_all)
    nl_beta_col <- match(path_rhs[nl_path_idx], lv_names_w_all)
    if (anyNA(nl_beta_row) || anyNA(nl_beta_col)) {
      lav_msg_stop(gettext(
        "unable to locate the latent-covariate random-slope path(s) in
         the within beta matrix."))
    }
    nl_beta_mm <- beta_mm
  }

  list(
    z.names = z_names, q = q,
    x.names = x_names, nx = length(x_names),
    y.names = y_names, p1 = p1,
    yb.names = yb_names, pb = pb, yb.y.idx = yb_y_idx,
    eps.names = eps_names, neps = length(eps_names),
    exo.b.names = exo_b_names, nexo.b = length(exo_b_names),
    zb.names = zb_names, kz = kz, zb.data.idx = zb_data_idx,
    path.tab = path_tab,
    y.data.idx = y_data_idx,
    x.data.idx = x_data_idx,
    exo.b.data.idx = exo_b_data_idx,
    p.flag = p_flag, shared.flag = shared_flag,
    # latent covariates (nonlinear random slopes; quadrature)
    z.type = z_type, nl.flag = nl_flag,
    nl.z.idx = nl_z_idx, q.nl = length(nl_z_idx),
    nl.beta.mm = nl_beta_mm,
    nl.beta.row = nl_beta_row, nl.beta.col = nl_beta_col,
    ngh = ngh
  )
}

# per-cluster crossproduct statistics; computed once
#
# with missing data (missing = "ml"), the y-block may contain NAs; the
# crossproducts are then computed per (cluster, missing-pattern) cell
# and stored in the 'mp' component; the covariates (both the rv
# covariates and the between-level exogenous covariates) must be
# complete: cases with missing values on exogenous variables have
# already been removed by lav_data (fixed.x = TRUE)
lav_mvn_cl_rs_stats <- function(y1 = NULL, lp = NULL, rs_info = NULL) {
  cluster_idx <- lp$cluster.idx[[2]]
  nclusters <- lp$nclusters[[2]]
  cluster_size <- lp$cluster.size[[2]]

  p1 <- rs_info$p1
  nx <- rs_info$nx

  yy <- y1[, rs_info$y.data.idx, drop = FALSE]
  xx <- y1[, rs_info$x.data.idx, drop = FALSE]

  if (anyNA(xx)) {
    lav_msg_stop(gettext(
      "unexpected missing values in the covariate(s) with a random
       slope; such cases should have been removed (fixed.x = TRUE)."))
  }

  # between-level exogenous covariates: one row per cluster
  if (rs_info$nexo.b > 0L) {
    first_idx <- which(!duplicated(cluster_idx))
    # order by cluster index (1..G)
    first_idx <- first_idx[order(cluster_idx[first_idx])]
    exo_b <- y1[first_idx, rs_info$exo.b.data.idx, drop = FALSE]
    if (anyNA(exo_b)) {
      lav_msg_stop(gettext(
        "unexpected missing values in the between-level exogenous
         covariate(s); such cases should have been removed
         (fixed.x = TRUE)."))
    }
  } else {
    exo_b <- matrix(0, nclusters, 0L)
  }

  # between-only endogenous z-outcomes: one row per cluster
  if (rs_info$kz > 0L) {
    first_idx <- which(!duplicated(cluster_idx))
    first_idx <- first_idx[order(cluster_idx[first_idx])]
    zb <- y1[first_idx, rs_info$zb.data.idx, drop = FALSE]
    if (anyNA(zb)) {
      lav_msg_stop(gettext(
        "missing values in between-only endogenous observed variables
         are not supported (yet) in combination with random slopes."))
    }
  } else {
    zb <- matrix(0, nclusters, 0L)
  }

  # missing data in the y-block?
  mp <- NULL
  if (anyNA(yy)) {
    obs <- !is.na(yy)
    nobs_case <- rowSums(obs)

    # cases with no observed outcome values carry no information (the
    # likelihood is conditional on the covariates)
    empty_idx <- which(nobs_case == 0L)
    if (length(empty_idx) > 0L) {
      lav_msg_warn(gettextf(
        "%s case(s) have missing values on all within-level outcome
         variables; they do not contribute to the loglikelihood.",
        length(empty_idx)))
      keep_idx <- seq_len(nrow(yy))[-empty_idx]
    } else {
      keep_idx <- seq_len(nrow(yy))
    }

    # NA -> 0: the padded crossproduct entries are annihilated by the
    # (embedded) per-pattern weight matrices later on
    yy0 <- yy
    yy0[is.na(yy0)] <- 0

    # missing patterns (over the y-block only)
    case_id <- apply(1L * obs, 1L, paste, collapse = "")
    pat_ids <- unique(case_id[keep_idx])
    npatterns <- length(pat_ids)

    pattern <- vector("list", npatterns)
    for (p in seq_len(npatterns)) {
      case_p <- keep_idx[case_id[keep_idx] == pat_ids[p]]
      o <- which(obs[case_p[1L], ])
      clus_p <- cluster_idx[case_p]

      yp <- yy0[case_p, , drop = FALSE]
      xp <- xx[case_p, , drop = FALSE]

      # per-cluster sums for this pattern (rows ordered by cluster id)
      sy_p <- rowsum.default(yp, clus_p, reorder = TRUE)
      sx_p <- rowsum.default(xp, clus_p, reorder = TRUE)
      n_p <- as.numeric(rowsum.default(
        rep(1, length(case_p)), clus_p, reorder = TRUE))
      j1 <- sort(unique.default(clus_p))

      idx_split <- split.default(seq_along(case_p), clus_p)
      nclus_p <- length(j1)
      sxx_p <- vector("list", nclus_p)
      sxy_p <- vector("list", nclus_p)
      syy_p <- vector("list", nclus_p)
      for (jj in seq_len(nclus_p)) {
        ii <- idx_split[[jj]]
        yjj <- yp[ii, , drop = FALSE]
        xjj <- xp[ii, , drop = FALSE]
        syy_p[[jj]] <- crossprod(yjj)
        sxx_p[[jj]] <- crossprod(xjj)
        sxy_p[[jj]] <- crossprod(xjj, yjj)
      }

      pattern[[p]] <- list(
        o = o, j1 = j1, n = n_p, sy = sy_p, sx = sx_p,
        sxx = sxx_p, sxy = sxy_p, syy = syy_p
      )
    }

    mp <- list(
      npatterns = npatterns, pattern = pattern,
      nobs = sum(nobs_case), empty.idx = empty_idx
    )

    return(list(
      nclusters = nclusters, cluster.size = cluster_size,
      sy = NULL, sx = NULL, sxx = NULL, sxy = NULL, syy = NULL,
      exo.b = exo_b, zb = zb, ntotal = sum(cluster_size),
      ntotal.used = sum(cluster_size) - length(empty_idx), mp = mp
    ))
  }

  # per-cluster sums and crossproducts
  sy <- rowsum.default(yy, cluster_idx, reorder = TRUE) # G x p1
  sx <- rowsum.default(xx, cluster_idx, reorder = TRUE) # G x nx

  sxx <- vector("list", nclusters)
  sxy <- vector("list", nclusters)
  syy <- vector("list", nclusters)

  # split row indices per cluster (loop over clusters; matrices are small)
  row_idx <- split.default(seq_len(nrow(y1)), cluster_idx)
  for (j in seq_len(nclusters)) {
    idx <- row_idx[[j]]
    yj <- yy[idx, , drop = FALSE]
    xj <- xx[idx, , drop = FALSE]
    syy[[j]] <- crossprod(yj)
    sxx[[j]] <- crossprod(xj)
    sxy[[j]] <- crossprod(xj, yj)
  }

  list(
    nclusters = nclusters, cluster.size = cluster_size,
    sy = sy, sx = sx, sxx = sxx, sxy = sxy, syy = syy,
    exo.b = exo_b, zb = zb, ntotal = sum(cluster_size),
    ntotal.used = sum(cluster_size), mp = NULL
  )
}

# model-implied ingredients for the random-slope loglikelihood,
# computed from GLIST
#
# returns: sigma.w (p1 x p1), mu.w (p1), P (p1 x nx),
#          lmat (p1 x npaths), sigma.v (pv x pv), mu.v (pv),
#          cc (pv x nexo.b), mu.exo (nexo.b)
lav_mvn_cl_rs_implied <- function(lavmodel = NULL, glist = NULL,
                                  rs_info = NULL) {
  if (is.null(glist)) {
    glist <- lavmodel@GLIST
  }
  nmat <- lavmodel@nmat

  # within block
  mm_w <- 1:nmat[1]
  mlist_w <- glist[mm_w]
  dimnames_w <- lavmodel@dimNames[mm_w]
  names(dimnames_w) <- names(glist)[mm_w]

  ov_names_w <- dimnames_w[["lambda"]][[1]]
  lv_names_w <- dimnames_w[["lambda"]][[2]]

  y_idx <- match(rs_info$y.names, ov_names_w)
  x_lv_idx <- match(rs_info$x.names, lv_names_w)

  lambda_w <- mlist_w$lambda
  psi_w <- mlist_w$psi
  theta_w <- mlist_w$theta
  beta_w <- mlist_w$beta
  nu_w <- mlist_w$nu
  alpha_w <- mlist_w$alpha

  nlv_w <- ncol(lambda_w)
  if (is.null(beta_w)) {
    ib_inv_w <- diag(nlv_w)
  } else {
    ib_inv_w <- try(solve(diag(nlv_w) - beta_w), silent = TRUE)
    if (inherits(ib_inv_w, "try-error")) {
      return(NULL)
    }
  }

  # reduced form: ov on lv
  rf_w <- lambda_w %*% ib_inv_w

  # condition on the x-covariates: zero out their psi rows/cols and alpha
  psi_w0 <- psi_w
  alpha_w0 <- if (is.null(alpha_w)) {
    matrix(0, nlv_w, 1L)
  } else {
    alpha_w
  }
  if (length(x_lv_idx) > 0L) {
    psi_w0[x_lv_idx, ] <- 0
    psi_w0[, x_lv_idx] <- 0
    alpha_w0[x_lv_idx, 1L] <- 0
  }

  # Sigma_w: conditional (on x and z) covariance of the y-vector
  rf_y <- rf_w[y_idx, , drop = FALSE]
  sigma_w <- rf_y %*% psi_w0 %*% t(rf_y) +
    theta_w[y_idx, y_idx, drop = FALSE]

  # mu_w: within intercept part of the y-vector
  nu_w0 <- if (is.null(nu_w)) {
    matrix(0, nrow(lambda_w), 1L)
  } else {
    nu_w
  }
  mu_w <- as.numeric(nu_w0[y_idx, 1L] + rf_y %*% alpha_w0)

  # P: reduced-form *fixed* slopes of y on the x-covariates
  # (the random-slope cells in beta are zero, so only the fixed paths
  #  contribute)
  pmat <- rf_y[, x_lv_idx, drop = FALSE]

  # L_r columns, one per random-slope path
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)
  lmat <- matrix(0, rs_info$p1, npaths)
  for (i in seq_len(npaths)) {
    lhs_lv_idx <- match(path_tab$lhs[i], lv_names_w)
    if (is.na(lhs_lv_idx)) {
      # lhs is an observed variable without a dummy lv: direct row
      lhs_ov_idx <- match(path_tab$lhs[i], rs_info$y.names)
      lmat[lhs_ov_idx, i] <- 1
    } else {
      lmat[, i] <- rf_y[, lhs_lv_idx]
    }
  }

  # between block
  mm_b <- 1:nmat[2] + nmat[1]
  mlist_b <- glist[mm_b]
  dimnames_b <- lavmodel@dimNames[mm_b]
  names(dimnames_b) <- names(glist)[mm_b]

  ov_names_b <- dimnames_b[["lambda"]][[1]]
  lv_names_b <- dimnames_b[["lambda"]][[2]]

  yb_idx <- match(rs_info$yb.names, ov_names_b)
  z_lv_idx <- match(rs_info$z.names, lv_names_b)
  exo_ov_idx <- match(rs_info$exo.b.names, ov_names_b)

  lambda_b <- mlist_b$lambda
  psi_b <- mlist_b$psi
  theta_b <- mlist_b$theta
  beta_b <- mlist_b$beta
  nu_b <- mlist_b$nu
  alpha_b <- mlist_b$alpha

  nlv_b <- ncol(lambda_b)
  if (is.null(beta_b)) {
    ib_inv_b <- diag(nlv_b)
  } else {
    ib_inv_b <- try(solve(diag(nlv_b) - beta_b), silent = TRUE)
    if (inherits(ib_inv_b, "try-error")) {
      return(NULL)
    }
  }

  e_eta <- if (is.null(alpha_b)) {
    matrix(0, nlv_b, 1L)
  } else {
    ib_inv_b %*% alpha_b
  }
  v_eta <- ib_inv_b %*% psi_b %*% t(ib_inv_b)

  nu_b0 <- if (is.null(nu_b)) {
    matrix(0, nrow(lambda_b), 1L)
  } else {
    nu_b
  }

  # the between random vector v_b = (eta_b, eps_b) in 'factor' space:
  # - eta_b: all between latent variables (including the random slopes,
  #   and excluding the dummy lv's of the between-level exogenous
  #   covariates, which are conditioned out)
  # - eps_b: the between residuals of the both-level variables with a
  #   non-zero residual (co)variance (see rs_info$eps.names)
  pb <- rs_info$pb
  q <- rs_info$q
  nexo <- rs_info$nexo.b
  neps <- rs_info$neps

  exo_lv_idx <- match(rs_info$exo.b.names, lv_names_b)
  if (nexo > 0L) {
    eta_idx <- seq_len(nlv_b)[-exo_lv_idx]
  } else {
    eta_idx <- seq_len(nlv_b)
  }
  meta <- length(eta_idx)
  pv <- meta + neps

  # position of the random slopes within v_b
  z_v_idx <- match(z_lv_idx, eta_idx)

  # condition eta on the between-level exogenous covariates
  if (nexo > 0L) {
    v_ee <- v_eta[exo_lv_idx, exo_lv_idx, drop = FALSE]
    v_ee_inv <- lav_mat_sym_inverse(v_ee)
    cc_eta <- v_eta[eta_idx, exo_lv_idx, drop = FALSE] %*% v_ee_inv
    v_eta_c <- v_eta[eta_idx, eta_idx, drop = FALSE] -
      cc_eta %*% v_eta[exo_lv_idx, eta_idx, drop = FALSE]
    v_eta_c <- (v_eta_c + t(v_eta_c)) / 2
    mu_exo <- as.numeric(nu_b0[exo_ov_idx, 1L] + e_eta[exo_lv_idx, 1L])
  } else {
    cc_eta <- matrix(0, meta, 0L)
    v_eta_c <- v_eta
    mu_exo <- numeric(0L)
  }

  # Sigma_v and (unconditional part of) mu_v
  sigma_v <- matrix(0, pv, pv)
  sigma_v[seq_len(meta), seq_len(meta)] <- v_eta_c
  mu_v <- numeric(pv)
  mu_v[seq_len(meta)] <- e_eta[eta_idx, 1L]
  cc <- matrix(0, pv, nexo)
  cc[seq_len(meta), ] <- cc_eta
  if (neps > 0L) {
    eps_b_idx <- match(rs_info$eps.names, ov_names_b)
    eps_pos <- meta + seq_len(neps)
    sigma_v[eps_pos, eps_pos] <-
      theta_b[eps_b_idx, eps_b_idx, drop = FALSE]
  }

  # constant loading matrix of v_b into the y-vector
  q0 <- matrix(0, rs_info$p1, pv)
  yb_y_idx <- rs_info$yb.y.idx
  q0[yb_y_idx, seq_len(meta)] <- lambda_b[yb_idx, eta_idx, drop = FALSE]
  if (neps > 0L) {
    eps_y_idx <- match(rs_info$eps.names, rs_info$y.names)
    for (i in seq_len(neps)) {
      q0[eps_y_idx[i], meta + i] <- 1
    }
  }

  # constant part of the mean of the y-vector: within intercepts +
  # between (fixed) intercepts of the both-level variables
  mu_y_b <- numeric(rs_info$p1)
  mu_y_b[yb_y_idx] <- nu_b0[yb_idx, 1L]
  mu_y <- mu_w + mu_y_b

  # between-only endogenous z-outcomes: conditional on v_b,
  #   z_j ~ N(mu_z + G v_b, sigma_z)
  # with G the between loadings of the z-block on the eta part of v_b
  # (the z-outcomes are pure indicators: no loadings on the eps part,
  # no dummy latent variables)
  kz <- rs_info$kz
  gmat <- matrix(0, kz, pv)
  mu_z <- numeric(kz)
  sigma_z <- matrix(0, kz, kz)
  if (kz > 0L) {
    zb_idx <- match(rs_info$zb.names, ov_names_b)
    gmat[, seq_len(meta)] <- lambda_b[zb_idx, eta_idx, drop = FALSE]
    mu_z <- as.numeric(nu_b0[zb_idx, 1L])
    sigma_z <- theta_b[zb_idx, zb_idx, drop = FALSE]
  }

  list(
    sigma.w = sigma_w, mu.y = mu_y, mu.y.b = mu_y_b, P = pmat,
    lmat = lmat, q0 = q0, z.v.idx = z_v_idx, pv = pv,
    sigma.v = sigma_v, mu.v = mu_v, cc = cc, mu.exo = mu_exo,
    eta.names = lv_names_b[eta_idx], meta = meta,
    mu.z = mu_z, gmat = gmat, sigma.z = sigma_z
  )
}

# -2 * (observed-data, conditional-on-x) loglikelihood
# computed from the per-cluster crossproduct statistics
lav_mvn_cl_rs_loglik <- function(rs_stats = NULL, imp = NULL,
                                 rs_info = NULL,
                                 sinv_method = "eigen",
                                 log2pi = TRUE,
                                 minus_two = TRUE,
                                 per_cluster = FALSE) {
  # missing data? use the pattern-based version
  if (!is.null(rs_stats$mp)) {
    return(lav_mvn_cl_rs_loglik_m(
      rs_stats = rs_stats, imp = imp, rs_info = rs_info,
      sinv_method = sinv_method, log2pi = log2pi,
      minus_two = minus_two, per_cluster = per_cluster
    ))
  }

  # non-PD or failed implied?
  if (is.null(imp)) {
    return(+Inf)
  }

  p1 <- rs_info$p1
  nx <- rs_info$nx
  pv <- imp$pv
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)

  nclusters <- rs_stats$nclusters
  nj_all <- rs_stats$cluster.size

  # invert sigma.w
  sigma_w_inv <- lav_mat_sym_inverse(imp$sigma.w,
    logdet = TRUE, sinv_method = sinv_method
  )
  logdet_w <- attr(sigma_w_inv, "logdet")
  if (!is.finite(logdet_w)) {
    return(+Inf)
  }

  # NOTE: we never invert sigma.v; we use the identities
  #   logdet(Sigma_v) + logdet(Sigma_v^{-1} + A) = logdet(I + A Sigma_v)
  #   (Sigma_v^{-1} + A)^{-1} = Sigma_v (I + A Sigma_v)^{-1}
  # so that the loglikelihood remains well-defined when Sigma_v is
  # singular (eg between residual variances fixed to zero) or even
  # (mildly) indefinite (Heywood territory), as long as every
  # cluster-level marginal covariance matrix remains positive definite
  sigma_v <- imp$sigma.v
  if (anyNA(sigma_v)) {
    return(+Inf)
  }
  # is sigma.v PSD? (fast path: no per-cluster eigenvalue checks needed)
  ev_v <- eigen(sigma_v, symmetric = TRUE, only.values = TRUE)$values
  v_psd <- ev_v[length(ev_v)] > -1e-12

  w_mat <- sigma_w_inv
  lmat <- imp$lmat
  pmat <- imp$P
  mu_y <- imp$mu.y
  q0 <- imp$q0
  z_v_idx <- imp$z.v.idx

  # precompute
  wl <- w_mat %*% lmat # p1 x npaths
  ltwl <- crossprod(lmat, wl) # npaths x npaths
  wq0 <- w_mat %*% q0 # p1 x pv
  q0twq0 <- crossprod(q0, wq0) # pv x pv
  q0twl <- crossprod(q0, wl) # pv x npaths
  wp <- w_mat %*% pmat # p1 x nx
  ptwp <- crossprod(pmat, wp) # nx x nx
  mu_y_w <- as.numeric(w_mat %*% mu_y) # p1

  path_z <- path_tab$z.idx # 1..q
  path_x <- path_tab$x.idx # 1..nx
  path_zcol <- z_v_idx[path_z] # column in v_b, per path

  # between-only endogenous z-outcomes: one extra (per-cluster)
  # Gaussian observation block, z_j | v_b ~ N(mu_z + G v_b, Sigma_z)
  kz <- rs_info$kz
  if (kz > 0L) {
    wzb <- lav_mat_sym_inverse(imp$sigma.z,
      logdet = TRUE, sinv_method = sinv_method
    )
    logdet_zb <- attr(wzb, "logdet")
    if (!is.finite(logdet_zb)) {
      return(+Inf)
    }
    wzg <- wzb %*% imp$gmat # kz x pv
    a_zb <- crossprod(imp$gmat, wzg) # pv x pv
    zb_all <- rs_stats$zb
    mu_zb <- imp$mu.z
  }

  # more constants (hoisted out of the cluster loop)
  pmat_t <- t(pmat)
  wp_t <- t(wp)
  mu_y_w_p <- as.numeric(mu_y_w %*% pmat) # nx
  mu_y_quad <- sum(mu_y_w * mu_y)
  dpv <- diag(pv)
  sigma_v_local <- sigma_v
  mu_v <- imp$mu.v
  cc_mat <- imp$cc
  mu_exo <- imp$mu.exo
  nexo <- rs_info$nexo.b
  exo_b <- rs_stats$exo.b

  sy_all <- rs_stats$sy
  sx_all <- rs_stats$sx
  sxx_all <- rs_stats$sxx
  sxy_all <- rs_stats$sxy
  syy_all <- rs_stats$syy

  loglik_j <- numeric(nclusters)
  ok <- TRUE

  ok <- tryCatch({
  for (j in seq_len(nclusters)) {
    nj <- nj_all[j]
    sy_j <- sy_all[j, ]
    sx_j <- sx_all[j, ]
    sxx_j <- sxx_all[[j]]
    sxy_j <- sxy_all[[j]]
    syy_j <- syy_all[[j]]

    # centered sums: e~_i = y_i - mu_y - P x_i
    # sum_i e~_i
    se_j <- sy_j - nj * mu_y - as.numeric(pmat %*% sx_j)
    # sum_i x_ai e~_i' (nx x p1)
    sxe_j <- sxy_j - outer(sx_j, mu_y) - sxx_j %*% pmat_t

    # sum_i e~_i' W e~_i
    quad0 <- sum(w_mat * syy_j) -
      2 * sum(mu_y_w * sy_j) +
      nj * mu_y_quad -
      2 * sum(wp_t * sxy_j) +
      2 * sum(mu_y_w_p * sx_j) +
      sum(ptwp * sxx_j)

    # Q_ij = Q0 + sum_p L_p x_pij e_{z(p)}'
    # A_j = sum_i Q_ij' W Q_ij; p_j = sum_i Q_ij' W e~_i

    # constant part
    a_j <- nj * q0twq0
    p_j <- as.numeric(crossprod(wq0, se_j))

    # cross part: Q0' W (sum_i Lx_i) (only the z columns are involved)
    cz <- matrix(0, pv, pv)
    for (i in seq_len(npaths)) {
      zcol <- path_zcol[i]
      cz[, zcol] <- cz[, zcol] + q0twl[, i] * sx_j[path_x[i]]
    }
    a_j <- a_j + cz + t(cz)

    # z x z part: sum_i (Lx_i)' W (Lx_i)
    for (i in seq_len(npaths)) {
      for (k in seq_len(npaths)) {
        a_j[path_zcol[i], path_zcol[k]] <-
          a_j[path_zcol[i], path_zcol[k]] +
          ltwl[i, k] * sxx_j[path_x[i], path_x[k]]
      }
    }

    # p_j z part: sum_p L_p' W (sum_i x_pi e~_i)
    for (i in seq_len(npaths)) {
      zi <- path_zcol[i]
      p_j[zi] <- p_j[zi] + sum(wl[, i] * sxe_j[path_x[i], ])
    }

    # z-outcome contributions (the extra observation block)
    if (kz > 0L) {
      e_zb <- zb_all[j, ] - mu_zb
      a_j <- a_j + a_zb
      p_j <- p_j + as.numeric(crossprod(wzg, e_zb))
      quad0 <- quad0 + sum(e_zb * (wzb %*% e_zb))
    }

    # mean of v_b for this cluster (given the between covariates)
    if (nexo > 0L) {
      d_j <- mu_v + as.numeric(cc_mat %*% (exo_b[j, ] - mu_exo))
    } else {
      d_j <- mu_v
    }

    # center v_b at its mean
    p_tilde <- p_j - as.numeric(a_j %*% d_j)
    quad_j <- quad0 - 2 * sum(d_j * p_j) + sum(d_j * (a_j %*% d_j))

    # Z_j = I + A_j Sigma_v; logdet(Z_j) and Sigma_v Z_j^{-1} p~
    z_j <- dpv + a_j %*% sigma_v_local
    if (v_psd) {
      dt <- determinant(z_j, logarithm = TRUE)
      if (dt$sign <= 0 || !is.finite(dt$modulus)) {
        ok <- FALSE
        break
      }
      logdet_z <- as.numeric(dt$modulus)
    } else {
      # sigma.v indefinite: check that all eigenvalues of Z_j are
      # (real and) positive, so that the cluster-level marginal
      # covariance matrix is positive definite
      lambda <- Re(eigen(z_j, only.values = TRUE)$values)
      if (any(lambda < sqrt(.Machine$double.eps))) {
        ok <- FALSE
        break
      }
      logdet_z <- sum(log(lambda))
    }
    zp_j <- solve(z_j, p_tilde)

    # -2 loglik for this cluster
    loglik_j[j] <- nj * logdet_w + logdet_z +
      quad_j - sum(p_tilde * (sigma_v_local %*% zp_j))
    if (kz > 0L) {
      loglik_j[j] <- loglik_j[j] + logdet_zb
    }
    if (log2pi) {
      loglik_j[j] <- loglik_j[j] + (nj * p1 + kz) * log(2 * pi)
    }
  }
  ok
  }, error = function(e) FALSE) # tryCatch around the cluster loop

  if (!ok) {
    return(+Inf)
  }

  out <- sum(loglik_j)
  if (!minus_two) {
    out <- out / (-2)
    loglik_j <- loglik_j / (-2)
  }
  if (per_cluster) {
    attr(out, "loglik.cluster") <- loglik_j
  }

  out
}


# -2 * (observed-data, conditional-on-x) loglikelihood with missing
# data in the y-block (missing = "ml")
#
# the only change wrt the complete-data version: the per-case weight
# matrix is now pattern-specific, W_p = Sigma_w[o,o]^{-1} embedded in
# a p1 x p1 zero matrix; because the zero rows/columns of W_p
# annihilate the zero-padded entries of the per-(cluster, pattern)
# crossproducts, all accumulation formulas carry over unchanged --
# they are simply applied per pattern, and summed per cluster before
# the marginalization over v_b
lav_mvn_cl_rs_loglik_m <- function(rs_stats = NULL, imp = NULL,
                                   rs_info = NULL,
                                   sinv_method = "eigen",
                                   log2pi = TRUE,
                                   minus_two = TRUE,
                                   per_cluster = FALSE) {
  # non-PD or failed implied?
  if (is.null(imp)) {
    return(+Inf)
  }

  p1 <- rs_info$p1
  pv <- imp$pv
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)

  nclusters <- rs_stats$nclusters
  mp <- rs_stats$mp

  sigma_v <- imp$sigma.v
  if (anyNA(sigma_v)) {
    return(+Inf)
  }
  # is sigma.v PSD? (fast path: no per-cluster eigenvalue checks needed)
  ev_v <- eigen(sigma_v, symmetric = TRUE, only.values = TRUE)$values
  v_psd <- ev_v[length(ev_v)] > -1e-12

  lmat <- imp$lmat
  pmat <- imp$P
  mu_y <- imp$mu.y
  q0 <- imp$q0
  z_v_idx <- imp$z.v.idx

  path_z <- path_tab$z.idx # 1..q
  path_x <- path_tab$x.idx # 1..nx
  path_zcol <- z_v_idx[path_z] # column in v_b, per path

  pmat_t <- t(pmat)
  dpv <- diag(pv)
  mu_v <- imp$mu.v
  cc_mat <- imp$cc
  mu_exo <- imp$mu.exo
  nexo <- rs_info$nexo.b
  exo_b <- rs_stats$exo.b
  log2pi_c <- if (log2pi) log(2 * pi) else 0

  # between-only endogenous z-outcomes (always complete)
  kz <- rs_info$kz
  if (kz > 0L) {
    wzb <- lav_mat_sym_inverse(imp$sigma.z,
      logdet = TRUE, sinv_method = sinv_method
    )
    logdet_zb <- attr(wzb, "logdet")
    if (!is.finite(logdet_zb)) {
      return(+Inf)
    }
    wzg <- wzb %*% imp$gmat # kz x pv
    a_zb <- crossprod(imp$gmat, wzg) # pv x pv
    zb_all <- rs_stats$zb
    mu_zb <- imp$mu.z
  }

  # per-cluster accumulators
  ld_j <- numeric(nclusters) # sum_p n_jp (logdet_wp + p_o log(2 pi))
  quad_j <- numeric(nclusters)
  a_arr <- array(0, dim = c(pv, pv, nclusters))
  p_all <- matrix(0, nclusters, pv)
  loglik_j <- numeric(nclusters)

  ok <- tryCatch({
    # ---- accumulation: loop over patterns, then clusters within ----
    for (p in seq_len(mp$npatterns)) {
      cell <- mp$pattern[[p]]
      o <- cell$o
      po <- length(o)

      w_o <- lav_mat_sym_inverse(imp$sigma.w[o, o, drop = FALSE],
        logdet = TRUE, sinv_method = sinv_method
      )
      logdet_o <- attr(w_o, "logdet")
      if (!is.finite(logdet_o)) {
        # caught by the enclosing tryCatch() -> objective returns +Inf
        lav_msg_stop(gettext("sigma.w is not positive definite."))
      }
      # embed in the full p1 x p1 space
      w_mat <- matrix(0, p1, p1)
      w_mat[o, o] <- w_o

      # per-pattern constants
      wl <- w_mat %*% lmat # p1 x npaths
      ltwl <- crossprod(lmat, wl) # npaths x npaths
      wq0 <- w_mat %*% q0 # p1 x pv
      q0twq0 <- crossprod(q0, wq0) # pv x pv
      q0twl <- crossprod(q0, wl) # pv x npaths
      wp <- w_mat %*% pmat # p1 x nx
      ptwp <- crossprod(pmat, wp) # nx x nx
      wp_t <- t(wp)
      mu_y_w <- as.numeric(w_mat %*% mu_y) # p1
      mu_y_w_p <- as.numeric(mu_y_w %*% pmat) # nx
      mu_y_quad <- sum(mu_y_w * mu_y)
      ld_p <- logdet_o + po * log2pi_c

      j1 <- cell$j1
      for (jj in seq_along(j1)) {
        j <- j1[jj]
        n_jp <- cell$n[jj]
        sy_jp <- cell$sy[jj, ]
        sx_jp <- cell$sx[jj, ]
        sxx_jp <- cell$sxx[[jj]]
        sxy_jp <- cell$sxy[[jj]]
        syy_jp <- cell$syy[[jj]]

        ld_j[j] <- ld_j[j] + n_jp * ld_p

        # centered sums: e~_i = y_i - mu_y - P x_i
        se_jp <- sy_jp - n_jp * mu_y - as.numeric(pmat %*% sx_jp)
        sxe_jp <- sxy_jp - outer(sx_jp, mu_y) - sxx_jp %*% pmat_t

        # sum_i e~_i' W_p e~_i
        quad_j[j] <- quad_j[j] + sum(w_mat * syy_jp) -
          2 * sum(mu_y_w * sy_jp) +
          n_jp * mu_y_quad -
          2 * sum(wp_t * sxy_jp) +
          2 * sum(mu_y_w_p * sx_jp) +
          sum(ptwp * sxx_jp)

        # A contribution
        a_jp <- n_jp * q0twq0
        p_jp <- as.numeric(crossprod(wq0, se_jp))
        cz <- matrix(0, pv, pv)
        for (i in seq_len(npaths)) {
          zcol <- path_zcol[i]
          cz[, zcol] <- cz[, zcol] + q0twl[, i] * sx_jp[path_x[i]]
        }
        a_jp <- a_jp + cz + t(cz)
        for (i in seq_len(npaths)) {
          for (k in seq_len(npaths)) {
            a_jp[path_zcol[i], path_zcol[k]] <-
              a_jp[path_zcol[i], path_zcol[k]] +
              ltwl[i, k] * sxx_jp[path_x[i], path_x[k]]
          }
        }
        for (i in seq_len(npaths)) {
          zi <- path_zcol[i]
          p_jp[zi] <- p_jp[zi] + sum(wl[, i] * sxe_jp[path_x[i], ])
        }

        a_arr[, , j] <- a_arr[, , j] + a_jp
        p_all[j, ] <- p_all[j, ] + p_jp
      }
    }

    # ---- marginalization over v_b, per cluster ----
    for (j in seq_len(nclusters)) {
      a_j <- matrix(a_arr[, , j], pv, pv)
      p_j <- p_all[j, ]

      # z-outcome contributions (the extra observation block)
      if (kz > 0L) {
        e_zb <- zb_all[j, ] - mu_zb
        a_j <- a_j + a_zb
        p_j <- p_j + as.numeric(crossprod(wzg, e_zb))
        quad_j[j] <- quad_j[j] + sum(e_zb * (wzb %*% e_zb))
        ld_j[j] <- ld_j[j] + logdet_zb + kz * log2pi_c
      }

      # mean of v_b for this cluster (given the between covariates)
      if (nexo > 0L) {
        d_j <- mu_v + as.numeric(cc_mat %*% (exo_b[j, ] - mu_exo))
      } else {
        d_j <- mu_v
      }

      # center v_b at its mean
      p_tilde <- p_j - as.numeric(a_j %*% d_j)
      quad_full <- quad_j[j] - 2 * sum(d_j * p_j) +
        sum(d_j * (a_j %*% d_j))

      # Z_j = I + A_j Sigma_v; logdet(Z_j) and Sigma_v Z_j^{-1} p~
      z_j <- dpv + a_j %*% sigma_v
      if (v_psd) {
        dt <- determinant(z_j, logarithm = TRUE)
        if (dt$sign <= 0 || !is.finite(dt$modulus)) {
          # caught by the enclosing tryCatch() -> objective returns +Inf
          lav_msg_stop(gettext("Z matrix is not positive definite."))
        }
        logdet_z <- as.numeric(dt$modulus)
      } else {
        # sigma.v indefinite: check that all eigenvalues of Z_j are
        # (real and) positive
        lambda <- Re(eigen(z_j, only.values = TRUE)$values)
        if (any(lambda < sqrt(.Machine$double.eps))) {
          # caught by the enclosing tryCatch() -> objective returns +Inf
          lav_msg_stop(gettext("Z matrix is not positive definite."))
        }
        logdet_z <- sum(log(lambda))
      }
      zp_j <- solve(z_j, p_tilde)

      # -2 loglik for this cluster (log 2 pi already in ld_j)
      loglik_j[j] <- ld_j[j] + logdet_z +
        quad_full - sum(p_tilde * (sigma_v %*% zp_j))
    }
    TRUE
  }, error = function(e) FALSE) # tryCatch around both loops

  if (!ok) {
    return(+Inf)
  }

  out <- sum(loglik_j)
  if (!minus_two) {
    out <- out / (-2)
    loglik_j <- loglik_j / (-2)
  }
  if (per_cluster) {
    attr(out, "loglik.cluster") <- loglik_j
  }

  out
}


# ---------------------------------------------------------------------
# latent covariates: quadrature over the nonlinear random slopes
# ---------------------------------------------------------------------
#
# random slopes that multiply a *latent* level-1 covariate enter the
# within model nonlinearly (a product of two random variables); the
# marginal loglikelihood is no longer available in closed form.
# Following Rockwood (2020), we split the between random vector v_b
# into the nonlinear slopes z_nl and the remaining ('linear') part
# v_lin. Conditional on z_nl = b:
#
#   - the within model is the *linear* random-slope model with the
#     slope values b injected into the (fixed-to-zero) beta_w cells:
#     the SAME lav_mvn_cl_rs_implied() gives sigma_w(b), lmat(b), P(b)
#   - v_lin | z_nl = b is normal with moments given by standard
#     Gaussian conditioning (a common covariance matrix, and a mean
#     that is linear in b)
#   - the conditional (on b) loglikelihood is computed by the SAME
#     crossproduct-based lav_mvn_cl_rs_loglik() -- including the
#     missing-data version
#
# the marginal loglikelihood is then obtained by Gauss-Hermite
# quadrature over z_nl only (q.nl dimensions, typically 1):
#
#   loglik_j = logsumexp_q [ loglik_j(b_q) + log w_q ]
#
# with nodes b_q = mu_nl + chol(Sigma_nl)' t_q placed at the prior of
# the nonlinear slopes (Rockwood, 2020, eq. 36-37). Per-node cost is
# one closed-form evaluation (the crossproduct statistics are shared
# across the nodes).
#
# when between-level covariates are related to the nonlinear slopes
# ('Case B', e.g. a regression s ~ w; Rockwood's PISA model), the
# prior means mu_nl_j become cluster-specific. We then keep ONE
# shared grid -- centered at the average prior mean, with the grid
# covariance inflated by the between-cluster spread of the prior
# means (so the grid covers every cluster's prior) -- and repair the
# weights per cluster with the exact density ratio:
#
#   loglik_j = logsumexp_q [ loglik_j(b_q) + log w_q + log r_jq ],
#   log r_jq = log phi(b_q; mu_nl_j, Sigma_nl)
#              + log|A| - log phi(t_q; 0, I),
#
# with b_q = c + A' t_q the shared grid. This keeps the per-node cost
# structure of Case A (the within matrices are rebuilt per node, not
# per cluster x node); without between-level covariates (Case A) the
# corrections are skipped entirely (bit-identical to the plain
# scheme), and when covariates are present but unrelated to the
# slopes they vanish up to roundoff. Note that Sigma_nl (the
# conditional-on-w covariance) is always common across clusters:
# conditioning on the between covariates shifts means only.

# single entry point for the -2 loglikelihood: dispatches between the
# closed-form kernel (observed covariates only) and the quadrature
# version (one or more latent covariates)
lav_mvn_cl_rs_m2ll <- function(lavmodel = NULL, glist = NULL, rs = NULL,
                               sinv_method = "eigen", log2pi = TRUE,
                               minus_two = TRUE, per_cluster = FALSE) {
  if (isTRUE(rs$info$nl.flag)) {
    return(lav_mvn_cl_rs_loglik_nl(
      lavmodel = lavmodel, glist = glist,
      rs_stats = rs$stats, rs_info = rs$info,
      sinv_method = sinv_method, log2pi = log2pi,
      minus_two = minus_two, per_cluster = per_cluster
    ))
  }
  imp <- lav_mvn_cl_rs_implied(
    lavmodel = lavmodel, glist = glist, rs_info = rs$info
  )
  lav_mvn_cl_rs_loglik(
    rs_stats = rs$stats, imp = imp, rs_info = rs$info,
    sinv_method = sinv_method, log2pi = log2pi,
    minus_two = minus_two, per_cluster = per_cluster
  )
}

# build the conditional-model ingredients at all quadrature nodes:
# the reduced index map (rs_info_c), and one 'imp' list per node with
# the node values injected into beta_w and v_b conditioned on the
# nonlinear slopes; when the prior means of the nonlinear slopes are
# cluster-specific (Case B), exo_b (the per-cluster between-covariate
# rows) is needed, and the returned 'logr' matrix (nclusters x
# nnodes) holds the per-cluster log density-ratio corrections;
# returns NULL if the base implied matrices fail or Sigma_nl is not
# positive definite (individual nodes may still fail: those imp
# entries are NULL)
lav_mvn_cl_rs_cond <- function(lavmodel = NULL, glist = NULL,
                               rs_info = NULL, exo_b = NULL) {
  if (is.null(glist)) {
    glist <- lavmodel@GLIST
  }

  # base implied ingredients; only the between pieces (sigma.v, mu.v,
  # cc, z.v.idx) are used here -- they do not depend on the injected
  # beta_w cells
  imp0 <- lav_mvn_cl_rs_implied(
    lavmodel = lavmodel, glist = glist, rs_info = rs_info
  )
  if (is.null(imp0)) {
    return(NULL)
  }

  path_tab <- rs_info$path.tab
  nl_z <- rs_info$nl.z.idx # indices in z.names
  nl_v <- imp0$z.v.idx[nl_z] # positions in v_b
  lin_v <- setdiff(seq_len(imp0$pv), nl_v)

  # prior (between) moments of the nonlinear slopes; the quadrature
  # requires a positive-definite Sigma_nl
  s_nl <- imp0$sigma.v[nl_v, nl_v, drop = FALSE]
  mu_nl <- imp0$mu.v[nl_v]
  if (anyNA(s_nl)) {
    return(NULL)
  }
  ch <- try(chol(s_nl), silent = TRUE) # s_nl = t(ch) %*% ch
  if (inherits(ch, "try-error")) {
    return(NULL)
  }
  s_nl_inv <- chol2inv(ch)

  # cluster-specific shifts of the nonlinear prior means (Case B):
  # mu_nl_j = mu_nl + delta_j, delta_j = cc[nl, ] (w_j - mu_exo).
  # NOTE: the branch is taken whenever between-level covariates are
  # present (a *structural* decision), not only when the current
  # cc[nl, ] values are nonzero: a free slope-on-covariate parameter
  # that happens to be zero (e.g., at the starting values) still has
  # a nonzero *derivative* of the correction terms, which the
  # gradient must see. When cc[nl, ] = 0, the corrections vanish (up
  # to roundoff) and the general path reduces to the plain scheme.
  q_nl <- length(nl_v)
  delta <- NULL
  case_b <- rs_info$nexo.b > 0L
  if (case_b) {
    if (is.null(exo_b)) {
      lav_msg_stop(gettext(
        "internal error: exo_b is needed when the model contains both
         between-level covariates and nonlinear random slopes."))
    }
    wc <- exo_b - rep(imp0$mu.exo, each = nrow(exo_b))
    delta <- wc %*% t(imp0$cc[nl_v, , drop = FALSE]) # G x q_nl
  }

  # conditional distribution of the linear part given the slopes:
  # v_lin | z_nl = b ~ N(mu_lin + R (b - mu_nl) + cc_c (w_j - mu_exo),
  #                      sigma_v_c)
  # with cc_c the *conditional* regression on the between covariates
  # (equal to cc[lin, ] when the slopes do not depend on them)
  r_mat <- imp0$sigma.v[lin_v, nl_v, drop = FALSE] %*% s_nl_inv
  sigma_v_c <- imp0$sigma.v[lin_v, lin_v, drop = FALSE] -
    r_mat %*% imp0$sigma.v[nl_v, lin_v, drop = FALSE]
  sigma_v_c <- (sigma_v_c + t(sigma_v_c)) / 2
  cc_c <- imp0$cc[lin_v, , drop = FALSE]
  if (case_b) {
    cc_c <- cc_c - r_mat %*% imp0$cc[nl_v, , drop = FALSE]
  }

  # conditional index map: only the observed-covariate paths remain
  ov_rows <- which(path_tab$type == "ov")
  z_keep <- which(rs_info$z.type == "ov")
  path_tab_c <- path_tab[ov_rows, , drop = FALSE]
  path_tab_c$z.idx <- match(path_tab$rv[ov_rows],
                            rs_info$z.names[z_keep])
  rs_info_c <- rs_info
  rs_info_c$path.tab <- path_tab_c
  rs_info_c$z.names <- rs_info$z.names[z_keep]
  rs_info_c$q <- length(z_keep)
  rs_info_c$nl.flag <- FALSE
  # positions of the remaining (observed-type) slopes in reduced v_b
  z_v_idx_c <- match(imp0$z.v.idx[z_keep], lin_v)

  # Gauss-Hermite grid (dnorm kernel; tensor product over q.nl dims)
  ngh <- rs_info$ngh
  if (is.null(ngh)) {
    ngh <- 21L
  }
  gh <- lav_integration_gauss_hermite(n = ngh, dnorm = TRUE)
  if (q_nl == 1L) {
    tmat <- matrix(gh$x, nrow = 1L)
    logw <- log(gh$w)
  } else {
    tmat <- t(as.matrix(expand.grid(rep(list(gh$x), q_nl))))
    logw <- rowSums(as.matrix(
      expand.grid(rep(list(log(gh$w)), q_nl))
    ))
  }
  nnodes <- ncol(tmat)

  # grid placement: b_q = c + A' t_q
  # - Case A: c = mu_nl, A = chol(Sigma_nl) (no correction needed)
  # - Case B: c = average prior mean; grid covariance = Sigma_nl +
  #   between-cluster covariance of the prior means (so the grid
  #   covers every cluster's prior); exact per-cluster density-ratio
  #   corrections below
  if (case_b) {
    grid_c <- mu_nl + colMeans(delta)
    dc <- delta - rep(colMeans(delta), each = nrow(delta))
    s_grid <- s_nl + crossprod(dc) / nrow(delta)
    a_up <- try(chol(s_grid), silent = TRUE)
    if (inherits(a_up, "try-error")) {
      return(NULL)
    }
  } else {
    grid_c <- mu_nl
    a_up <- ch
  }
  bmat <- grid_c + t(a_up) %*% tmat # q_nl x nnodes

  # per-cluster log density-ratio corrections (Case B only):
  # log r_jq = log phi(b_q; mu_nl + delta_j, Sigma_nl)
  #            + log|A| - log phi(t_q; 0, I)
  logr <- NULL
  if (case_b) {
    mu_mat <- rep(mu_nl, each = nrow(delta)) + delta # G x q_nl
    mb <- s_nl_inv %*% bmat # q_nl x nnodes
    bmb <- colSums(bmat * mb) # nnodes
    mm <- rowSums((mu_mat %*% s_nl_inv) * mu_mat) # G
    cross <- mu_mat %*% mb # G x nnodes
    logdet_s <- 2 * sum(log(diag(ch)))
    logdet_a <- sum(log(diag(a_up)))
    tt <- colSums(tmat * tmat) # nnodes
    logr <- -0.5 * (outer(mm, bmb, "+") - 2 * cross) -
      0.5 * logdet_s + logdet_a +
      rep(0.5 * tt, each = nrow(delta))
  }

  # node-value injection map: nonlinear path -> beta_w cell
  nl_path_rows <- which(path_tab$type == "lv")
  path_nl_z <- match(path_tab$z.idx[nl_path_rows], nl_z)
  beta_mm <- rs_info$nl.beta.mm
  beta_cells <- cbind(rs_info$nl.beta.row, rs_info$nl.beta.col)

  imp_list <- vector("list", nnodes)
  glist_i <- glist
  for (i in seq_len(nnodes)) {
    b_i <- bmat[, i]
    glist_i[[beta_mm]][beta_cells] <- b_i[path_nl_z]
    imp_i <- lav_mvn_cl_rs_implied(
      lavmodel = lavmodel, glist = glist_i, rs_info = rs_info
    )
    if (is.null(imp_i)) {
      next
    }
    imp_list[[i]] <- list(
      sigma.w = imp_i$sigma.w, mu.y = imp_i$mu.y,
      mu.y.b = imp_i$mu.y.b, P = imp_i$P,
      lmat = imp_i$lmat[, ov_rows, drop = FALSE],
      q0 = imp_i$q0[, lin_v, drop = FALSE],
      z.v.idx = z_v_idx_c, pv = length(lin_v),
      sigma.v = sigma_v_c,
      mu.v = imp0$mu.v[lin_v] + as.numeric(r_mat %*% (b_i - mu_nl)),
      cc = cc_c,
      mu.exo = imp0$mu.exo,
      # z-outcomes: the slopes have no indicators, so the nl columns
      # of G are structurally zero and simply drop out
      mu.z = imp_i$mu.z,
      gmat = imp_i$gmat[, lin_v, drop = FALSE],
      sigma.z = imp_i$sigma.z
    )
  }

  list(rs_info_c = rs_info_c, imp = imp_list, logw = logw,
       nnodes = nnodes, logr = logr,
       # extras for the EAP (lavPredict) machinery
       bmat = bmat, lin.v = lin_v, nl.v = nl_v,
       eta.names = imp0$eta.names, meta = imp0$meta, pv = imp0$pv,
       z.v.idx = imp0$z.v.idx)
}

# -2 * loglikelihood by Gauss-Hermite quadrature over the nonlinear
# (latent-covariate) random slopes; all failure modes return +Inf
# (mirroring lav_mvn_cl_rs_loglik)
lav_mvn_cl_rs_loglik_nl <- function(lavmodel = NULL, glist = NULL,
                                    rs_stats = NULL, rs_info = NULL,
                                    sinv_method = "eigen",
                                    log2pi = TRUE,
                                    minus_two = TRUE,
                                    per_cluster = FALSE) {
  cond <- lav_mvn_cl_rs_cond(
    lavmodel = lavmodel, glist = glist, rs_info = rs_info,
    exo_b = rs_stats$exo.b
  )
  if (is.null(cond)) {
    return(+Inf)
  }

  nclusters <- rs_stats$nclusters
  nnodes <- cond$nnodes
  hmat <- matrix(-Inf, nclusters, nnodes)

  for (i in seq_len(nnodes)) {
    if (is.null(cond$imp[[i]])) {
      next
    }
    m2_i <- lav_mvn_cl_rs_loglik(
      rs_stats = rs_stats, imp = cond$imp[[i]],
      rs_info = cond$rs_info_c,
      sinv_method = sinv_method, log2pi = log2pi,
      minus_two = TRUE, per_cluster = TRUE
    )
    if (!is.finite(m2_i)) {
      next
    }
    hmat[, i] <- -0.5 * attr(m2_i, "loglik.cluster") + cond$logw[i]
  }

  # cluster-specific prior means (Case B): density-ratio corrections
  if (!is.null(cond$logr)) {
    hmat <- hmat + cond$logr
  }

  # per-cluster log-sum-exp over the nodes
  hmax <- apply(hmat, 1L, max)
  if (any(!is.finite(hmax))) {
    return(+Inf)
  }
  loglik_j <- hmax + log(rowSums(exp(hmat - hmax)))

  out <- sum(loglik_j)
  if (minus_two) {
    out <- -2 * out
    loglik_j <- -2 * loglik_j
  }
  if (per_cluster) {
    attr(out, "loglik.cluster") <- loglik_j
  }

  out
}

# analytic (adjoint-based) gradient of the quadrature loglikelihood:
#
#   d loglik_j / d theta = sum_i pi_ji d loglik_j(b_i) / d theta
#
# with pi_ji the posterior node weights (the derivative of the
# log-sum-exp); the conditional (per-node) derivatives reuse the
# closed-form adjoint pass (lav_mvn_cl_rs_dfphi) on the conditional
# model, chained through a small numerical Jacobian of the per-node
# ingredient map theta -> phi_c(theta; node i) -- which includes the
# dependence of the node values b_i(theta) = mu_nl + chol(Sigma_nl)'t_i
# on theta (no cluster loops); returns d(-2 loglik)/d theta, or the
# per-cluster score matrix (natural scale) when per_cluster = TRUE;
# NULL on failure (callers fall back to numerical differentiation)
lav_mvn_cl_rs_grad_nl <- function(lavmodel = NULL, rs = NULL,
                                  sinv_method = "eigen",
                                  per_cluster = FALSE,
                                  h = 1e-06) {
  rs_info <- rs$info
  rs_stats <- rs$stats

  cond0 <- lav_mvn_cl_rs_cond(lavmodel = lavmodel, rs_info = rs_info,
                              exo_b = rs_stats$exo.b)
  if (is.null(cond0)) {
    return(NULL)
  }
  nnodes <- cond0$nnodes
  nclusters <- rs_stats$nclusters
  case_b <- !is.null(cond0$logr)

  # per-node conditional logliks + adjoints at the current parameters
  hmat <- matrix(-Inf, nclusters, nnodes)
  gphi_list <- vector("list", nnodes)
  for (i in seq_len(nnodes)) {
    imp_i <- cond0$imp[[i]]
    if (is.null(imp_i)) {
      return(NULL)
    }
    m2_i <- lav_mvn_cl_rs_loglik(
      rs_stats = rs_stats, imp = imp_i, rs_info = cond0$rs_info_c,
      sinv_method = sinv_method, log2pi = FALSE,
      minus_two = TRUE, per_cluster = TRUE
    )
    if (!is.finite(m2_i)) {
      return(NULL)
    }
    hmat[, i] <- -0.5 * attr(m2_i, "loglik.cluster") + cond0$logw[i]
    gphi_list[[i]] <- lav_mvn_cl_rs_dfphi(
      rs_stats = rs_stats, imp = imp_i, rs_info = cond0$rs_info_c,
      sinv_method = sinv_method
    )
    if (is.null(gphi_list[[i]])) {
      return(NULL)
    }
  }
  if (case_b) {
    hmat <- hmat + cond0$logr
  }
  hmax <- apply(hmat, 1L, max)
  if (any(!is.finite(hmax))) {
    return(NULL)
  }
  loglik_j <- hmax + log(rowSums(exp(hmat - hmax)))
  pw <- exp(hmat - loglik_j) # posterior node weights (rows sum to 1)

  # numerical Jacobian of the per-node ingredient vectors phi_c
  # (central differences; no cluster loops); in Case B, the log
  # density-ratio corrections logr also depend on theta -- their
  # derivatives ride along in the same loop
  x0 <- lav_model_get_parameters(lavmodel)
  npar <- length(x0)
  cond_of <- function(x) {
    lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
    cond <- lav_mvn_cl_rs_cond(lavmodel = lavmodel2, rs_info = rs_info,
                               exo_b = rs_stats$exo.b)
    if (is.null(cond)) {
      return(NULL)
    }
    list(
      phi = lapply(cond$imp, function(m) {
        if (is.null(m)) NULL else lav_mvn_cl_rs_phi_vec(m)
      }),
      logr = cond$logr
    )
  }
  nphi <- length(lav_mvn_cl_rs_phi_vec(cond0$imp[[1]]))
  jac <- lapply(seq_len(nnodes), function(i) matrix(0, nphi, npar))
  dlr_sc <- if (case_b) matrix(0, nclusters, npar) else NULL
  for (k in seq_len(npar)) {
    h_k <- h * max(1, abs(x0[k]))
    x_right <- x_left <- x0
    x_right[k] <- x0[k] + h_k
    x_left[k] <- x0[k] - h_k
    cond_r <- cond_of(x_right)
    cond_l <- cond_of(x_left)
    if (is.null(cond_r) || is.null(cond_l)) {
      return(NULL)
    }
    for (i in seq_len(nnodes)) {
      if (is.null(cond_r$phi[[i]]) || is.null(cond_l$phi[[i]])) {
        return(NULL)
      }
      jac[[i]][, k] <- (cond_r$phi[[i]] - cond_l$phi[[i]]) / (2 * h_k)
    }
    if (case_b) {
      if (is.null(cond_r$logr) || is.null(cond_l$logr)) {
        return(NULL)
      }
      dlr_k <- (cond_r$logr - cond_l$logr) / (2 * h_k)
      dlr_sc[, k] <- rowSums(pw * dlr_k)
    }
  }

  if (per_cluster) {
    # per-cluster scores, natural loglikelihood scale
    sc <- matrix(0, nclusters, npar)
    for (i in seq_len(nnodes)) {
      sc <- sc + (pw[, i] * (-0.5) * gphi_list[[i]]) %*% jac[[i]]
    }
    if (case_b) {
      sc <- sc + dlr_sc
    }
    sc
  } else {
    # d(-2 loglik)/d theta
    dx <- numeric(npar)
    for (i in seq_len(nnodes)) {
      dx <- dx +
        as.numeric(colSums(pw[, i] * gphi_list[[i]]) %*% jac[[i]])
    }
    if (case_b) {
      dx <- dx - 2 * colSums(dlr_sc)
    }
    dx
  }
}


# ---------------------------------------------------------------------
# the EM algorithm
# ---------------------------------------------------------------------
#
# complete data = observed data + the between random vector v_b (in
# factor space, see above) + the 'within slope' variables
# z_w[r] = v_b[z_r] * x_r (one per random-slope path; completed with an
# arbitrary standard-normal marginal).
#
# - E-step: the posterior of v_b given the cluster data is
#       Sigma0_j = Sigma_v (I + A_j Sigma_v)^{-1},
#       mu0_j    = d_j + Sigma0_j (p_j - A_j d_j)
#   (the same A_j/p_j workhorses as the marginal loglikelihood); the
#   expected complete-data moments then collapse onto the same
#   per-cluster crossproducts (sy, sx, sxx, sxy, syy)
# - M-step: a standard TWO-GROUP SEM, fitted to the expected moments:
#   group 1 ('within', N obs) contains the y-vector plus the phantom
#   z_w variables (unit variance, zero mean, fixed loading 1 on the
#   lhs of their path); group 2 ('between', G obs) contains the
#   both-level variables, the random slopes (now 'observed') and the
#   between covariates
#
# limitations (fall back to nlminb): fixed paths from the rv
# covariates (P != 0), shared rv labels; between-level variances that
# are fixed to (near) zero are floored at 1e-4 in the M-step model
# (cfr. the Mplus 'minimum variance'); the nlminb hand-off (enabled by
# default) polishes the solution of the *exact* model afterwards

# check if the EM version can handle this model; returns character(0)
# if ok, otherwise the reason(s)
lav_mvn_cl_rs_em_ok <- function(rs = NULL) {
  reason <- character(0L)
  if (isTRUE(rs$info$nl.flag)) {
    reason <- c(reason, gettext(
      "the model contains random slopes for latent covariates"))
  }
  if (isTRUE(rs$info$kz > 0L)) {
    reason <- c(reason, gettext(
      "the model contains between-only endogenous observed variables"))
  }
  if (rs$info$p.flag) {
    reason <- c(reason, gettext(
      "the model contains fixed regression paths from the random-slope
       covariates"))
  }
  if (rs$info$shared.flag) {
    reason <- c(reason, gettext(
      "the same rv() label is attached to more than one regression
       path"))
  }
  reason
}

# E-step: expected complete-data moments
#
# returns:
#   r1 (p1 + q), t1 (p1 + q square): mean and covariance (about r1)
#     of the within complete-data vector (y_w, z_w) -- N observations
#   r2 (pb + q + nexo), t2 (square): mean and covariance of the
#     between complete-data vector (y_b, z, w) -- G observations
lav_mvn_cl_rs_estep <- function(rs_stats = NULL, imp = NULL,
                                rs_info = NULL,
                                sinv_method = "eigen") {
  # missing data? use the pattern-based version
  if (!is.null(rs_stats$mp)) {
    return(lav_mvn_cl_rs_estep_m(
      rs_stats = rs_stats, imp = imp, rs_info = rs_info,
      sinv_method = sinv_method
    ))
  }
  p1 <- rs_info$p1
  nx <- rs_info$nx
  q <- rs_info$q
  pb <- rs_info$pb
  nexo <- rs_info$nexo.b
  pv <- imp$pv
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)

  nclusters <- rs_stats$nclusters
  nj_all <- rs_stats$cluster.size
  ntotal <- rs_stats$ntotal

  sigma_w_inv <- lav_mat_sym_inverse(imp$sigma.w,
    logdet = TRUE, sinv_method = sinv_method
  )
  if (!is.finite(attr(sigma_w_inv, "logdet"))) {
    lav_msg_stop(gettext(
      "within covariance matrix is not positive definite in the
       E-step."))
  }
  w_mat <- sigma_w_inv
  sigma_v <- imp$sigma.v
  mu_y <- imp$mu.y
  nu_map <- imp$mu.y.b
  q0 <- imp$q0
  z_v_idx <- imp$z.v.idx
  path_zcol <- z_v_idx[path_tab$z.idx]
  path_x <- path_tab$x.idx

  # constants
  wq0 <- w_mat %*% q0
  q0twq0 <- crossprod(q0, wq0)
  wl <- w_mat %*% imp$lmat
  q0twl <- crossprod(q0, wl)
  ltwl <- crossprod(imp$lmat, wl)
  dpv <- diag(pv)

  # between mapping: (y_b, z) = c0 + G v
  gmat <- rbind(
    q0[rs_info$yb.y.idx, , drop = FALSE],
    dpv[z_v_idx, , drop = FALSE]
  )
  c0 <- c(nu_map[rs_info$yb.y.idx], rep(0, q))

  # accumulators
  sum1_w <- numeric(p1 + q) # within first moments
  sum2_w <- matrix(0, p1 + q, p1 + q) # within second moments
  sum1_b <- numeric(pb + q + nexo)
  sum2_b <- matrix(0, pb + q + nexo, pb + q + nexo)

  y_idx1 <- seq_len(p1)
  z_idx1 <- p1 + seq_len(q)
  b_idx2 <- seq_len(pb + q)
  w_idx2 <- pb + q + seq_len(nexo)

  for (j in seq_len(nclusters)) {
    nj <- nj_all[j]
    sy_j <- rs_stats$sy[j, ]
    sx_j <- rs_stats$sx[j, ]
    sxx_j <- rs_stats$sxx[[j]]
    sxy_j <- rs_stats$sxy[[j]]
    syy_j <- rs_stats$syy[[j]]

    # A_j and p_j (centered at mu.y; no P term: the EM requires P = 0)
    se_j <- sy_j - nj * mu_y
    a_j <- nj * q0twq0
    p_j <- as.numeric(crossprod(wq0, se_j))
    cz <- matrix(0, pv, pv)
    for (i in seq_len(npaths)) {
      zcol <- path_zcol[i]
      cz[, zcol] <- cz[, zcol] + q0twl[, i] * sx_j[path_x[i]]
    }
    a_j <- a_j + cz + t(cz)
    for (i in seq_len(npaths)) {
      for (k in seq_len(npaths)) {
        a_j[path_zcol[i], path_zcol[k]] <-
          a_j[path_zcol[i], path_zcol[k]] +
          ltwl[i, k] * sxx_j[path_x[i], path_x[k]]
      }
    }
    sxe_j <- sxy_j - outer(sx_j, mu_y)
    for (i in seq_len(npaths)) {
      zi <- path_zcol[i]
      p_j[zi] <- p_j[zi] + sum(wl[, i] * sxe_j[path_x[i], ])
    }

    # prior mean of v_b for this cluster
    if (nexo > 0L) {
      w_j <- rs_stats$exo.b[j, ]
      d_j <- imp$mu.v + as.numeric(imp$cc %*% (w_j - imp$mu.exo))
    } else {
      w_j <- numeric(0L)
      d_j <- imp$mu.v
    }

    # posterior of v_b
    z_j <- dpv + a_j %*% sigma_v
    sigma0_j <- sigma_v %*% solve(z_j)
    sigma0_j <- (sigma0_j + t(sigma0_j)) / 2
    p_tilde <- p_j - as.numeric(a_j %*% d_j)
    mu0_j <- d_j + as.numeric(sigma0_j %*% p_tilde)
    m0_j <- sigma0_j + tcrossprod(mu0_j) # E[v v']

    # ---------- within complete-data moments: (y_w, z_w) ----------
    # y_w = a - Q0 v, with a = y - nu_map (observed)
    sa_j <- sy_j - nj * nu_map # sum_i a_i
    saa_j <- syy_j - outer(sy_j, nu_map) - outer(nu_map, sy_j) +
      nj * tcrossprod(nu_map) # sum_i a_i a_i'
    q0mu <- as.numeric(q0 %*% mu0_j)
    q0m0q0 <- q0 %*% m0_j %*% t(q0)

    sum1_w[y_idx1] <- sum1_w[y_idx1] + sa_j - nj * q0mu
    sum2_w[y_idx1, y_idx1] <- sum2_w[y_idx1, y_idx1] +
      saa_j - outer(sa_j, q0mu) - outer(q0mu, sa_j) + nj * q0m0q0

    # z_w[r] = v[z_r] * (the covariate of path r)
    # (one-to-one path <-> slope mapping, so path r defines slope r;
    #  but note that two slopes may share the same covariate)
    sxa_j <- sxy_j - outer(sx_j, nu_map) # sum_i x_ai a_i'
    for (r in seq_len(q)) {
      zc_r <- z_v_idx[r]
      xr <- path_x[r] # covariate (column in the x-block) of path r
      sum1_w[z_idx1[r]] <- sum1_w[z_idx1[r]] + mu0_j[zc_r] * sx_j[xr]
      for (s in seq_len(q)) {
        zc_s <- z_v_idx[s]
        xs <- path_x[s]
        sum2_w[z_idx1[r], z_idx1[s]] <- sum2_w[z_idx1[r], z_idx1[s]] +
          m0_j[zc_r, zc_s] * sxx_j[xr, xs]
      }
      # cross y_w x z_w[r]:
      # sum_i x_(xr)i ( a_i mu0[z_r] - Q0 M0[, z_r] )
      cross_r <- sxa_j[xr, ] * mu0_j[zc_r] -
        sx_j[xr] * as.numeric(q0 %*% m0_j[, zc_r])
      sum2_w[y_idx1, z_idx1[r]] <- sum2_w[y_idx1, z_idx1[r]] + cross_r
      sum2_w[z_idx1[r], y_idx1] <- sum2_w[z_idx1[r], y_idx1] + cross_r
    }

    # ---------- between complete-data moments: (y_b, z, w) ----------
    b_j <- c0 + as.numeric(gmat %*% mu0_j)
    bb_j <- gmat %*% m0_j %*% t(gmat) +
      outer(c0, b_j) + outer(b_j, c0) - tcrossprod(c0)

    sum1_b[b_idx2] <- sum1_b[b_idx2] + b_j
    sum2_b[b_idx2, b_idx2] <- sum2_b[b_idx2, b_idx2] + bb_j
    if (nexo > 0L) {
      sum1_b[w_idx2] <- sum1_b[w_idx2] + w_j
      sum2_b[b_idx2, w_idx2] <- sum2_b[b_idx2, w_idx2] +
        outer(b_j, w_j)
      sum2_b[w_idx2, b_idx2] <- sum2_b[w_idx2, b_idx2] +
        outer(w_j, b_j)
      sum2_b[w_idx2, w_idx2] <- sum2_b[w_idx2, w_idx2] +
        tcrossprod(w_j)
    }
  } # j

  r1 <- sum1_w / ntotal
  t1 <- sum2_w / ntotal - tcrossprod(r1)
  r2 <- sum1_b / nclusters
  t2 <- sum2_b / nclusters - tcrossprod(r2)

  list(r1 = r1, t1 = t1, r2 = r2, t2 = t2)
}

# E-step with missing data in the y-block (missing = "ml")
#
# the complete data now include the missing y-values. Per pattern
# (observed set o, missing set m), define
#   B_mo = Sigma_w[m,o] Sigma_w[o,o]^{-1}
#   K_p  = (embedded m-selector) - (embedded B_mo)   [p1 x p1]
#   R_p  = I - K_p
#   C_p  = Schur complement Sigma_w[m,m] - B_mo Sigma_w[o,m], embedded
# then, conditional on (v_b, y_obs), the completed y-vector is
#   y_i = R_p y~_i + K_p (mu_y + Q_i v_b) + eps*,  Var(eps*) = C_p
# (y~_i = the zero-padded observed vector). Because K_p Q_i remains
# affine in x_i, all expected-moment sums again collapse onto the
# per-(cluster, pattern) crossproducts; for a complete pattern
# (K_p = 0, R_p = I, C_p = 0) every term reduces to the complete-data
# E-step. The M-step (the two-group SEM) is unchanged.
lav_mvn_cl_rs_estep_m <- function(rs_stats = NULL, imp = NULL,
                                  rs_info = NULL,
                                  sinv_method = "eigen") {
  p1 <- rs_info$p1
  nx <- rs_info$nx
  q <- rs_info$q
  pb <- rs_info$pb
  nexo <- rs_info$nexo.b
  pv <- imp$pv
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)

  nclusters <- rs_stats$nclusters
  mp <- rs_stats$mp

  sigma_w <- imp$sigma.w
  sigma_v <- imp$sigma.v
  mu_y <- imp$mu.y
  nu_map <- imp$mu.y.b
  q0 <- imp$q0
  lmat <- imp$lmat
  z_v_idx <- imp$z.v.idx
  path_zcol <- z_v_idx[path_tab$z.idx]
  path_x <- path_tab$x.idx
  dpv <- diag(pv)

  # between mapping: (y_b, z) = c0 + G v
  gmat <- rbind(
    q0[rs_info$yb.y.idx, , drop = FALSE],
    dpv[z_v_idx, , drop = FALSE]
  )
  c0 <- c(nu_map[rs_info$yb.y.idx], rep(0, q))

  # ---- per-pattern constants ----
  npat <- mp$npatterns
  pc <- vector("list", npat)
  for (p in seq_len(npat)) {
    o <- mp$pattern[[p]]$o
    m <- seq_len(p1)[-o]
    if (length(m) == 0L) {
      m <- integer(0L)
    }
    w_o <- lav_mat_sym_inverse(sigma_w[o, o, drop = FALSE],
      logdet = TRUE, sinv_method = sinv_method
    )
    if (!is.finite(attr(w_o, "logdet"))) {
      lav_msg_stop(gettext(
        "within covariance matrix is not positive definite in the
         E-step."))
    }
    w_mat <- matrix(0, p1, p1)
    w_mat[o, o] <- w_o

    k_mat <- matrix(0, p1, p1)
    ct_mat <- matrix(0, p1, p1)
    if (length(m) > 0L) {
      b_mo <- sigma_w[m, o, drop = FALSE] %*% w_o
      k_mat[cbind(m, m)] <- 1
      k_mat[m, o] <- -b_mo
      ct_mat[m, m] <- sigma_w[m, m, drop = FALSE] -
        b_mo %*% sigma_w[o, m, drop = FALSE]
    }
    r_mat <- diag(p1) - k_mat

    pc[[p]] <- list(
      w = w_mat,
      k = k_mat, r = r_mat, ct = ct_mat,
      g = as.numeric(k_mat %*% mu_y) - nu_map,
      h0 = -r_mat %*% q0, # = K Q0 - Q0 (p1 x pv)
      kl = k_mat %*% lmat, # p1 x npaths
      # constants for the A_j/p_j accumulation
      wl = w_mat %*% lmat,
      wq0 = w_mat %*% q0,
      q0twq0 = crossprod(q0, w_mat %*% q0),
      q0twl = crossprod(q0, w_mat %*% lmat),
      ltwl = crossprod(lmat, w_mat %*% lmat)
    )
  }

  # ---- phase 1: pattern-aware A_j and p_j (P = 0) ----
  a_arr <- array(0, dim = c(pv, pv, nclusters))
  p_all <- matrix(0, nclusters, pv)
  sx_tot <- matrix(0, nclusters, nx)
  sxx_tot <- array(0, dim = c(nx, nx, nclusters))
  ntot_used <- 0

  for (p in seq_len(npat)) {
    cell <- mp$pattern[[p]]
    cst <- pc[[p]]
    j1 <- cell$j1
    for (jj in seq_along(j1)) {
      j <- j1[jj]
      n_jp <- cell$n[jj]
      sy_jp <- cell$sy[jj, ]
      sx_jp <- cell$sx[jj, ]
      sxx_jp <- cell$sxx[[jj]]
      sxy_jp <- cell$sxy[[jj]]

      ntot_used <- ntot_used + n_jp
      sx_tot[j, ] <- sx_tot[j, ] + sx_jp
      sxx_tot[, , j] <- sxx_tot[, , j] + sxx_jp

      se_jp <- sy_jp - n_jp * mu_y
      a_jp <- n_jp * cst$q0twq0
      p_jp <- as.numeric(crossprod(cst$wq0, se_jp))
      cz <- matrix(0, pv, pv)
      for (i in seq_len(npaths)) {
        zcol <- path_zcol[i]
        cz[, zcol] <- cz[, zcol] + cst$q0twl[, i] * sx_jp[path_x[i]]
      }
      a_jp <- a_jp + cz + t(cz)
      for (i in seq_len(npaths)) {
        for (k in seq_len(npaths)) {
          a_jp[path_zcol[i], path_zcol[k]] <-
            a_jp[path_zcol[i], path_zcol[k]] +
            cst$ltwl[i, k] * sxx_jp[path_x[i], path_x[k]]
        }
      }
      sxe_jp <- sxy_jp - outer(sx_jp, mu_y)
      for (i in seq_len(npaths)) {
        zi <- path_zcol[i]
        p_jp[zi] <- p_jp[zi] + sum(cst$wl[, i] * sxe_jp[path_x[i], ])
      }

      a_arr[, , j] <- a_arr[, , j] + a_jp
      p_all[j, ] <- p_all[j, ] + p_jp
    }
  }

  # ---- posterior of v_b, per cluster ----
  mu0 <- matrix(0, nclusters, pv)
  m0_list <- vector("list", nclusters)
  for (j in seq_len(nclusters)) {
    a_j <- matrix(a_arr[, , j], pv, pv)
    if (nexo > 0L) {
      d_j <- imp$mu.v +
        as.numeric(imp$cc %*% (rs_stats$exo.b[j, ] - imp$mu.exo))
    } else {
      d_j <- imp$mu.v
    }
    z_j <- dpv + a_j %*% sigma_v
    sigma0_j <- sigma_v %*% solve(z_j)
    sigma0_j <- (sigma0_j + t(sigma0_j)) / 2
    p_tilde <- p_all[j, ] - as.numeric(a_j %*% d_j)
    mu0_j <- d_j + as.numeric(sigma0_j %*% p_tilde)
    mu0[j, ] <- mu0_j
    m0_list[[j]] <- sigma0_j + tcrossprod(mu0_j) # E[v v']
  }

  # ---- accumulators ----
  sum1_w <- numeric(p1 + q)
  sum2_w <- matrix(0, p1 + q, p1 + q)
  sum1_b <- numeric(pb + q + nexo)
  sum2_b <- matrix(0, pb + q + nexo, pb + q + nexo)

  y_idx1 <- seq_len(p1)
  z_idx1 <- p1 + seq_len(q)
  b_idx2 <- seq_len(pb + q)
  w_idx2 <- pb + q + seq_len(nexo)

  # ---- phase 2: within y-moments, per (pattern, cluster) cell ----
  # completed within part: y_w = c_i + H_i v + eps*, with
  #   c_i = R y~_i + g,  H_i = H0 + sum_r x_ir (K L_r) e_{z_r}'
  for (p in seq_len(npat)) {
    cell <- mp$pattern[[p]]
    cst <- pc[[p]]
    j1 <- cell$j1
    for (jj in seq_along(j1)) {
      j <- j1[jj]
      n_jp <- cell$n[jj]
      sy_jp <- cell$sy[jj, ]
      sx_jp <- cell$sx[jj, ]
      sxx_jp <- cell$sxx[[jj]]
      sxy_jp <- cell$sxy[[jj]]
      syy_jp <- cell$syy[[jj]]
      mu0_j <- mu0[j, ]
      m0_j <- m0_list[[j]]

      # sum_i c_i and sum_i c_i c_i'
      rsy <- as.numeric(cst$r %*% sy_jp)
      sc <- rsy + n_jp * cst$g
      cc_mat <- cst$r %*% syy_jp %*% t(cst$r) +
        outer(rsy, cst$g) + outer(cst$g, rsy) +
        n_jp * tcrossprod(cst$g)

      # sum_i H_i (p1 x pv)
      sh <- n_jp * cst$h0
      for (r in seq_len(q)) {
        zc <- path_zcol[r]
        sh[, zc] <- sh[, zc] + sx_jp[path_x[r]] * cst$kl[, r]
      }

      # sum_i x_ir c_i, per slope r (p1 vectors)
      cx <- vector("list", q)
      for (r in seq_len(q)) {
        xr <- path_x[r]
        cx[[r]] <- as.numeric(cst$r %*% sxy_jp[xr, ]) +
          sx_jp[xr] * cst$g
      }

      # sum_i c_i (H_i mu0)'
      h0mu <- as.numeric(cst$h0 %*% mu0_j)
      ch <- outer(sc, h0mu)
      for (r in seq_len(q)) {
        ch <- ch + mu0_j[path_zcol[r]] * outer(cx[[r]], cst$kl[, r])
      }

      # sum_i H_i M0 H_i'
      h0m0 <- cst$h0 %*% m0_j # p1 x pv
      hmh <- n_jp * h0m0 %*% t(cst$h0)
      for (r in seq_len(q)) {
        zc_r <- path_zcol[r]
        h0m_r <- h0m0[, zc_r]
        hmh <- hmh + sx_jp[path_x[r]] *
          (outer(h0m_r, cst$kl[, r]) + outer(cst$kl[, r], h0m_r))
        for (s in seq_len(q)) {
          hmh <- hmh + sxx_jp[path_x[r], path_x[s]] *
            m0_j[zc_r, path_zcol[s]] *
            outer(cst$kl[, r], cst$kl[, s])
        }
      }

      sum1_w[y_idx1] <- sum1_w[y_idx1] + sc + as.numeric(sh %*% mu0_j)
      sum2_w[y_idx1, y_idx1] <- sum2_w[y_idx1, y_idx1] +
        cc_mat + ch + t(ch) + hmh + n_jp * cst$ct

      # cross y_w x z_w[r]
      for (r in seq_len(q)) {
        zc_r <- path_zcol[r]
        xr <- path_x[r]
        cross_r <- mu0_j[zc_r] * cx[[r]] +
          sx_jp[xr] * h0m0[, zc_r]
        for (s in seq_len(q)) {
          cross_r <- cross_r + sxx_jp[xr, path_x[s]] *
            m0_j[path_zcol[s], zc_r] * cst$kl[, s]
        }
        sum2_w[y_idx1, z_idx1[r]] <- sum2_w[y_idx1, z_idx1[r]] + cross_r
        sum2_w[z_idx1[r], y_idx1] <- sum2_w[z_idx1[r], y_idx1] + cross_r
      }
    }
  }

  # ---- z_w moments and between moments, per cluster ----
  for (j in seq_len(nclusters)) {
    mu0_j <- mu0[j, ]
    m0_j <- m0_list[[j]]

    for (r in seq_len(q)) {
      zc_r <- z_v_idx[r]
      xr <- path_x[r]
      sum1_w[z_idx1[r]] <- sum1_w[z_idx1[r]] +
        mu0_j[zc_r] * sx_tot[j, xr]
      for (s in seq_len(q)) {
        sum2_w[z_idx1[r], z_idx1[s]] <- sum2_w[z_idx1[r], z_idx1[s]] +
          m0_j[zc_r, z_v_idx[s]] * sxx_tot[xr, path_x[s], j]
      }
    }

    # between complete-data moments: (y_b, z, w)
    b_j <- c0 + as.numeric(gmat %*% mu0_j)
    bb_j <- gmat %*% m0_j %*% t(gmat) +
      outer(c0, b_j) + outer(b_j, c0) - tcrossprod(c0)
    sum1_b[b_idx2] <- sum1_b[b_idx2] + b_j
    sum2_b[b_idx2, b_idx2] <- sum2_b[b_idx2, b_idx2] + bb_j
    if (nexo > 0L) {
      w_j <- rs_stats$exo.b[j, ]
      sum1_b[w_idx2] <- sum1_b[w_idx2] + w_j
      sum2_b[b_idx2, w_idx2] <- sum2_b[b_idx2, w_idx2] +
        outer(b_j, w_j)
      sum2_b[w_idx2, b_idx2] <- sum2_b[w_idx2, b_idx2] +
        outer(w_j, b_j)
      sum2_b[w_idx2, w_idx2] <- sum2_b[w_idx2, w_idx2] +
        tcrossprod(w_j)
    }
  }

  r1 <- sum1_w / ntot_used
  t1 <- sum2_w / ntot_used - tcrossprod(r1)
  r2 <- sum1_b / nclusters
  t2 <- sum2_b / nclusters - tcrossprod(r2)

  list(r1 = r1, t1 = t1, r2 = r2, t2 = t2, ntotal.used = ntot_used)
}

# construct the two-group M-step parameter table (computed once)
#
# returns: list(pt = the two-group partable (data.frame),
#               ov1/ov2 = the observed-variable order per group,
#               perm1/perm2 = permutations from the canonical E-step
#                             order to ov1/ov2,
#               free.idx = which(pt$free > 0),
#               fixed.x = any exo)
lav_mvn_cl_rs_mstep_pt <- function(lavpartable = NULL, rs_info = NULL,
                                   floor_value = 1e-4) {
  pt <- as.data.frame(lavpartable, stringsAsFactors = FALSE)
  # drop columns we do not need
  pt$est <- NULL
  pt$se <- NULL
  pt$start <- NULL
  pt$rv <- NULL
  pt$mod.idx <- NULL
  # two-group version: level -> group
  pt$group <- NULL
  names(pt)[names(pt) == "level"] <- "group"

  # remove all level-1 rows involving the rv covariates (their moments,
  # and the frozen random-slope paths)
  x_names <- rs_info$x.names
  rm_idx <- which(pt$group == 1L &
    (pt$lhs %in% x_names | pt$rhs %in% x_names))
  # remove the phantom marker rows (s =~ s) at the between level
  rm_idx <- c(rm_idx, which(pt$op == "=~" & pt$lhs == pt$rhs &
    pt$lhs %in% rs_info$z.names))
  if (length(rm_idx) > 0L) {
    pt <- pt[-rm_idx, , drop = FALSE]
  }

  # Mplus-style variance floor: between-level variances fixed to (near)
  # zero would make the complete-data model singular; floor them in the
  # M-step model only (the nlminb hand-off afterwards uses the exact
  # model)
  floor_idx <- which(pt$group == 2L & pt$op == "~~" &
    pt$lhs == pt$rhs & pt$free == 0L &
    !is.na(pt$ustart) & abs(pt$ustart) < floor_value)
  floored <- length(floor_idx) > 0L
  if (floored) {
    pt$ustart[floor_idx] <- floor_value
  }

  # phantom z_w variables at the within level; make sure the names do
  # not clash with existing variables
  znames <- paste0(rs_info$z.names, ".zw")
  taken <- unique(c(pt$lhs, pt$rhs))
  while (any(znames %in% taken)) {
    znames <- paste0(".", znames)
  }

  q <- rs_info$q
  path_tab <- rs_info$path.tab
  new_lhs <- character(0L)
  new_op <- character(0L)
  new_rhs <- character(0L)
  # the fixed unit loading of the path lhs on its z_w variable
  for (r in seq_len(q)) {
    new_lhs <- c(new_lhs, path_tab$lhs[r])
    new_op <- c(new_op, "~")
    new_rhs <- c(new_rhs, znames[r])
  }
  # unit variances, zero covariances, zero means (the standard-normal
  # completion)
  for (r in seq_len(q)) {
    for (s in r:q) {
      new_lhs <- c(new_lhs, znames[r])
      new_op <- c(new_op, "~~")
      new_rhs <- c(new_rhs, znames[s])
    }
  }
  new_lhs <- c(new_lhs, znames)
  new_op <- c(new_op, rep("~1", q))
  new_rhs <- c(new_rhs, rep("", q))
  nnew <- length(new_lhs)
  new_ustart <- numeric(nnew)
  new_ustart[seq_len(q)] <- 1 # the fixed loadings
  var_idx <- q + which(new_op[-seq_len(q)] == "~~" &
    new_lhs[-seq_len(q)] == new_rhs[-seq_len(q)])
  new_ustart[var_idx] <- 1 # the unit variances

  new_rows <- pt[rep(1L, nnew), , drop = FALSE]
  new_rows$id <- max(pt$id) + seq_len(nnew)
  new_rows$lhs <- new_lhs
  new_rows$op <- new_op
  new_rows$rhs <- new_rhs
  new_rows$user <- 2L
  new_rows$block <- 1L
  new_rows$group <- 1L
  new_rows$free <- 0L
  new_rows$ustart <- new_ustart
  new_rows$exo <- 0L
  if (!is.null(new_rows$label)) new_rows$label <- ""
  if (!is.null(new_rows$plabel)) new_rows$plabel <- ""
  if (!is.null(new_rows$efa)) new_rows$efa <- ""
  if (!is.null(new_rows$lower)) new_rows$lower <- -Inf
  if (!is.null(new_rows$upper)) new_rows$upper <- +Inf
  pt <- rbind(pt, new_rows)
  rownames(pt) <- NULL

  # observed-variable order per group
  ov1 <- lav_pt_vnames(pt, "ov", block = 1L)
  ov2 <- lav_pt_vnames(pt, "ov", block = 2L)

  # canonical E-step order: (y.names, znames) and (yb, z, exo)
  can1 <- c(rs_info$y.names, znames)
  can2 <- c(rs_info$yb.names, rs_info$z.names, rs_info$exo.b.names)
  if (length(ov1) != length(can1) || !all(ov1 %in% can1)) {
    lav_msg_fixme("within M-step variables do not match the E-step")
  }
  if (length(ov2) != length(can2) || !all(ov2 %in% can2)) {
    lav_msg_fixme("between M-step variables do not match the E-step")
  }

  list(
    pt = pt, ov1 = ov1, ov2 = ov2,
    perm1 = match(ov1, can1), perm2 = match(ov2, can2),
    free.idx = which(pt$free > 0L),
    fixed.x = any(pt$exo == 1L),
    znames = znames, floored = floored
  )
}

# EM estimation of the h0 model with random slopes
# (mirrors lav_mvn_cl_em_h0; single group)
lav_mvn_cl_rs_em_h0 <- function(lavsamplestats = NULL, lavdata = NULL,
                                lavpartable = NULL, lavmodel = NULL,
                                lavoptions = NULL, lavcache = NULL,
                                fx_tol = 1e-08, dx_tol = 1e-04,
                                max_iter = 5000L,
                                mstep_max_iter = 10000L,
                                mstep_rel_tol = 1e-10,
                                mstep_verbose = FALSE,
                                acceleration = "none") {
  stopifnot(lavdata@ngroups == 1L)

  # the rs cache
  rs <- lavcache[[1]]$rs
  if (is.null(rs)) {
    rs_info <- lav_mvn_cl_rs_info(lavmodel = lavmodel,
                                  lavdata = lavdata)
    rs_stats <- lav_mvn_cl_rs_stats(
      y1 = lavdata@X[[1]], lp = lavdata@Lp[[1]], rs_info = rs_info
    )
    rs <- list(info = rs_info, stats = rs_stats)
  }

  # feasibility
  reason <- lav_mvn_cl_rs_em_ok(rs)
  if (length(reason) > 0L) {
    lav_msg_stop(gettextf(
      "the EM algorithm is not available for this random-slope model
       (%s); use optim.method = \"nlminb\" instead.",
      paste(reason, collapse = "; ")))
  }

  # the two-group M-step model (built once)
  mstep <- lav_mvn_cl_rs_mstep_pt(
    lavpartable = lavpartable, rs_info = rs$info
  )
  ntot_used <- rs$stats$ntotal.used
  if (is.null(ntot_used)) {
    ntot_used <- rs$stats$ntotal
  }
  nobs_list <- list(ntot_used, rs$stats$nclusters)

  # M-step: fit the two-group model to the expected moments
  em_mstep <- function(estep, x_start) {
    local_pt <- mstep$pt
    local_pt$ustart[mstep$free.idx] <- x_start
    t1 <- estep$t1[mstep$perm1, mstep$perm1, drop = FALSE]
    r1 <- estep$r1[mstep$perm1]
    t2 <- estep$t2[mstep$perm2, mstep$perm2, drop = FALSE]
    r2 <- estep$r2[mstep$perm2]
    dimnames(t1) <- list(mstep$ov1, mstep$ov1)
    dimnames(t2) <- list(mstep$ov2, mstep$ov2)
    names(r1) <- mstep$ov1
    names(r2) <- mstep$ov2
    if (!mstep_verbose) {
      current_verbose <- lav_verbose()
      if (lav_verbose(FALSE)) {
        on.exit(lav_verbose(current_verbose), TRUE)
      }
    }
    lavaan(local_pt,
      sample_cov = list(within = t1, between = t2),
      sample_mean = list(within = r1, between = r2),
      sample_nobs = nobs_list,
      sample.cov.rescale = FALSE,
      control = list(
        iter.max = mstep_max_iter,
        rel.tol = mstep_rel_tol
      ),
      fixed.x = mstep$fixed.x,
      estimator = "ML",
      warn = FALSE,
      check.start = FALSE,
      check.post = FALSE,
      check.gradient = FALSE,
      check.vcov = FALSE,
      baseline = FALSE,
      fit.by.level = FALSE,
      h1 = FALSE,
      se = "none",
      test = "none"
    )
  }

  # one EM step as a map in the model parameter space: x -> x'
  em_step <- function(x, logl = FALSE) {
    lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
    imp <- lav_mvn_cl_rs_implied(lavmodel2, rs_info = rs$info)
    if (is.null(imp)) {
      lav_msg_stop(gettext(
        "model-implied matrices could not be computed in the E-step."))
    }
    estep <- lav_mvn_cl_rs_estep(
      rs_stats = rs$stats, imp = imp, rs_info = rs$info
    )
    local_fit <- em_mstep(estep, x_start = x)
    local_fit@optim$x
  }

  # observed-data loglikelihood at x
  em_logl <- function(x) {
    lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
    imp <- lav_mvn_cl_rs_implied(lavmodel2, rs_info = rs$info)
    out <- lav_mvn_cl_rs_loglik(
      rs_stats = rs$stats, imp = imp, rs_info = rs$info,
      log2pi = TRUE, minus_two = FALSE
    )
    as.numeric(out)
  }

  # initial values
  x_current <- lav_model_get_parameters(lavmodel)
  fx <- em_logl(x_current)

  if (lav_verbose()) {
    cat(
      "EM iter:", sprintf("%3d", 0),
      " fx =", sprintf("%17.10f", fx),
      "\n"
    )
  }

  if (acceleration %in% c("squarem", "qn")) {
    em_conv <- function(theta_old, theta, fx_old, fx) {
      if ((fx - fx_old) < fx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: fx.delta < ", fx_tol, "\n")
        }
        return(TRUE)
      }
      # the gradient is numerical (2 x npar objective evaluations):
      # only check it when the loglikelihood is no longer moving much
      if ((fx - fx_old) > 0.01) {
        return(FALSE)
      }
      lavmodel2 <- lav_model_set_parameters(lavmodel, x = theta)
      dx <- lav_model_grad(lavmodel2,
        lavdata = lavdata,
        lavsamplestats = lavsamplestats,
        lavcache = lavcache
      )
      if (max(abs(dx)) < dx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: max.dx < ", dx_tol, "\n")
        }
        return(TRUE)
      }
      FALSE
    }
    accel_fn <- if (acceleration == "qn") lav_em_qn else lav_em_squarem
    out_em <- accel_fn(
      theta = x_current, step_fn = em_step, logl_fn = em_logl,
      fx0 = fx, conv_fn = em_conv, max_iter = max_iter,
      logl_dec = 0 # inexact M-step: only accept monotone extrapolations
    )
    x <- out_em$theta
    fx <- out_em$fx
    em_converged <- out_em$converged || out_em$stalled
    em_iterations <- out_em$fpeval
  } else {
    # plain EM iterations
    fx_old <- fx
    x <- x_current
    for (i in 1:max_iter) {
      x <- em_step(x)
      fx <- em_logl(x)
      fx_delta <- fx - fx_old

      # the gradient is numerical (2 x npar objective evaluations):
      # only check it when the loglikelihood is no longer moving much
      if (fx_delta < 0.01) {
        lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
        dx <- lav_model_grad(lavmodel2,
          lavdata = lavdata,
          lavsamplestats = lavsamplestats,
          lavcache = lavcache
        )
        max_dx <- max(abs(dx))
      } else {
        max_dx <- as.numeric(NA)
      }

      if (lav_verbose()) {
        cat(
          "EM iter:", sprintf("%3d", i),
          " fx =", sprintf("%17.10f", fx),
          " fx.delta =", sprintf("%9.8f", fx_delta),
          " max.dx = ", sprintf("%9.8f", max_dx),
          "\n"
        )
      }

      if (fx_delta < fx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: fx.delta < ", fx_tol, "\n")
        }
        break
      } else {
        fx_old <- fx
      }
      if (is.finite(max_dx) && max_dx < dx_tol) {
        if (lav_verbose()) {
          cat("EM stopping rule reached: max.dx < ", dx_tol, "\n")
        }
        break
      }
    }
    em_converged <- i < max_iter
    em_iterations <- i
  }

  # add attributes
  if (em_converged) {
    attr(x, "converged") <- TRUE
    attr(x, "warn.txt") <- ""
  } else {
    attr(x, "converged") <- FALSE
    attr(x, "warn.txt") <- paste("maximum number of iterations (",
      max_iter, ") ",
      "was reached without convergence.\n",
      sep = ""
    )
  }
  attr(x, "iterations") <- em_iterations
  attr(x, "control") <- list(
    em.args = list(
      max_iter = max_iter,
      fx_tol = fx_tol,
      dx_tol = dx_tol,
      acceleration = acceleration
    )
  )
  attr(fx, "fx.group") <- fx # single group for now
  attr(x, "fx") <- fx
  # if the M-step model used the variance floor, the EM fixed point is
  # NOT the optimum of the exact model: the nlminb hand-off must always
  # run (even if the gradient looks small)
  attr(x, "rs.floored") <- mstep$floored

  x
}


# ---------------------------------------------------------------------
# analytic gradient (and per-cluster scores)
# ---------------------------------------------------------------------
#
# the gradient of the -2 loglikelihood is computed in two (chained)
# layers:
#
# 1) an ADJOINT pass (lav_mvn_cl_rs_dfphi): analytic derivatives of
#    F_j = sum_P n_jP ln|Sigma_w[o,o]| + ln|Z_j| + q~_j
#          - p~_j' Sigma_v Z_j^{-1} p~_j
#    with respect to the model-implied ingredients
#    phi = (Sigma_w, mu_y, P, lmat, q0, Sigma_v, mu_v, cc, mu_exo).
#    With u = Sigma_v Z^{-1} p~ (so mu0 = d + u, Sigma0 = Sigma_v
#    Z^{-1}), the derivatives wrt the per-cluster workhorses are
#    remarkably clean (Fisher's identity at work):
#       dF/dA      = Sigma0 + mu0 mu0' = M0
#       dF/dp      = -2 mu0
#       dF/dq      = 1
#       dF/dd      = -2 (p - A mu0)
#       dF/dSigma_v = sym(Z^{-1} A) - sym( (p~ - A u) (Z^{-1} p~)' )
#    (all inverse-free: valid for singular/indefinite Sigma_v);
#    chaining through A_j, p_j, q_j collapses -- once again -- onto
#    the per-(cluster, pattern) crossproducts: ONE pass over the data
#    cells, cost comparable to a single loglikelihood evaluation
#
# 2) the (small) Jacobian d phi / d theta of the implied map,
#    computed by central differences of lav_mvn_cl_rs_implied():
#    2 x npar evaluations of a few small matrix products -- no
#    cluster loops, negligible cost
#
# the same adjoint pass, kept per cluster, yields the per-cluster
# scores (for the robust/sandwich standard errors) at no extra cost

# flatten the theta-dependent implied ingredients into one vector
# (the adjoint pass uses the same layout)
lav_mvn_cl_rs_phi_vec <- function(imp) {
  c(
    as.numeric(imp$sigma.w), imp$mu.y, as.numeric(imp$P),
    as.numeric(imp$lmat), as.numeric(imp$q0), as.numeric(imp$sigma.v),
    imp$mu.v, as.numeric(imp$cc), imp$mu.exo,
    imp$mu.z, as.numeric(imp$gmat), as.numeric(imp$sigma.z)
  )
}

# adjoint pass: returns a (nclusters x nphi) matrix with the
# per-cluster derivatives of F_j wrt phi (NULL if something fails)
lav_mvn_cl_rs_dfphi <- function(rs_stats = NULL, imp = NULL,
                                rs_info = NULL,
                                sinv_method = "eigen") {
  p1 <- rs_info$p1
  nx <- rs_info$nx
  pv <- imp$pv
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)
  z_v_idx <- imp$z.v.idx
  path_zcol <- z_v_idx[path_tab$z.idx]
  path_x <- path_tab$x.idx
  nclusters <- rs_stats$nclusters
  nexo <- rs_info$nexo.b
  exo_b <- rs_stats$exo.b

  sigma_v <- imp$sigma.v
  mu_v <- imp$mu.v
  cc_mat <- imp$cc
  mu_exo <- imp$mu.exo
  mu_y <- imp$mu.y
  pmat <- imp$P
  pmat_t <- t(pmat)
  lmat <- imp$lmat
  q0 <- imp$q0
  dpv <- diag(pv)
  mu_y_op <- tcrossprod(mu_y)

  # flattened phi layout (must match lav_mvn_cl_rs_phi_vec)
  kz <- rs_info$kz
  sizes <- c(p1 * p1, p1, p1 * nx, p1 * npaths, p1 * pv,
             pv * pv, pv, pv * nexo, nexo,
             kz, kz * pv, kz * kz)
  off <- cumsum(c(0, sizes))
  i_sw <- off[1] + seq_len(sizes[1])
  i_muy <- off[2] + seq_len(sizes[2])
  i_p <- off[3] + seq_len(sizes[3])
  i_lm <- off[4] + seq_len(sizes[4])
  i_q0 <- off[5] + seq_len(sizes[5])
  i_sv <- off[6] + seq_len(sizes[6])
  i_mv <- off[7] + seq_len(sizes[7])
  i_cc <- off[8] + seq_len(sizes[8])
  i_me <- off[9] + seq_len(sizes[9])
  i_mz <- off[10] + seq_len(sizes[10])
  i_g <- off[11] + seq_len(sizes[11])
  i_sz <- off[12] + seq_len(sizes[12])
  nphi <- off[13]

  gphi <- matrix(0, nclusters, nphi)

  # z-outcome constants (the extra observation block)
  if (kz > 0L) {
    gmat_z <- imp$gmat
    mu_zb <- imp$mu.z
    zb_all <- rs_stats$zb
    wzb <- lav_mat_sym_inverse(imp$sigma.z,
      logdet = TRUE, sinv_method = sinv_method
    )
    if (!is.finite(attr(wzb, "logdet"))) {
      return(NULL)
    }
    wzg <- wzb %*% gmat_z # kz x pv
    a_zb <- crossprod(gmat_z, wzg) # pv x pv
  }

  # ---- z-block adjoints (mu_z, G, Sigma_z); the z-block is one
  #      extra 'cell' with constant loading G and weight Wz ----
  add_zcell <- function(j, g_a, g_p) {
    e_z <- zb_all[j, ] - mu_zb
    wze <- as.numeric(wzb %*% e_z)
    gga <- gmat_z %*% g_a # kz x pv
    ggp <- as.numeric(gmat_z %*% g_p) # kz
    # G adjoint (cfr. the q0 adjoint with n = 1, e~ = e_z)
    gg <- 2 * (wzb %*% gga) + outer(wze, g_p)
    # mu_z adjoint (cfr. the mu_y adjoint)
    gmz <- -as.numeric(wzb %*% (ggp + 2 * e_z))
    # Sigma_z adjoint (via Wz, plus the ln|Sigma_z| term)
    gw_z <- gga %*% t(gmat_z) + outer(ggp, e_z) + tcrossprod(e_z)
    gw_z <- (gw_z + t(gw_z)) / 2
    gsz <- wzb - wzb %*% gw_z %*% wzb
    gphi[j, i_mz] <<- gphi[j, i_mz] + gmz
    gphi[j, i_g] <<- gphi[j, i_g] + as.numeric(gg)
    gphi[j, i_sz] <<- gphi[j, i_sz] + as.numeric(gsz)
    invisible(NULL)
  }

  # ---- posterior pieces + (Sigma_v, mu_v, cc, mu_exo) adjoints ----
  # (returns the g_a/g_p adjoints needed for the cell collapse)
  post_adj <- function(j, a_j, p_j) {
    if (nexo > 0L) {
      d_j <- mu_v + as.numeric(cc_mat %*% (exo_b[j, ] - mu_exo))
    } else {
      d_j <- mu_v
    }
    p_t <- p_j - as.numeric(a_j %*% d_j)
    z_j <- dpv + a_j %*% sigma_v
    z_inv <- solve(z_j)
    zp <- as.numeric(z_inv %*% p_t)
    u <- as.numeric(sigma_v %*% zp)
    mu0 <- d_j + u

    s0 <- sigma_v %*% z_inv
    s0 <- (s0 + t(s0)) / 2
    g_a <- s0 + tcrossprod(mu0)
    g_p <- -2 * mu0
    g_d <- -2 * (p_j - as.numeric(a_j %*% mu0))
    za <- z_inv %*% a_j
    m1 <- outer(p_t - as.numeric(a_j %*% u), zp)
    g_sv <- (za + t(za)) / 2 - (m1 + t(m1)) / 2

    gphi[j, i_sv] <<- gphi[j, i_sv] + as.numeric(g_sv)
    gphi[j, i_mv] <<- gphi[j, i_mv] + g_d
    if (nexo > 0L) {
      gphi[j, i_cc] <<- gphi[j, i_cc] +
        as.numeric(outer(g_d, exo_b[j, ] - mu_exo))
      gphi[j, i_me] <<- gphi[j, i_me] -
        as.numeric(crossprod(cc_mat, g_d))
    }
    list(g_a = g_a, g_p = g_p)
  }

  # ---- collapse of one (cluster, pattern) cell onto the
  #      (Sigma_w, mu_y, P, lmat, q0) adjoints ----
  add_cell <- function(j, w_mat, n_c, sx_c, sxx_c, sy_c, sxy_c, syy_c,
                       se_c, sxe_c, g_a, g_p) {
    # SQ = sum_i Q_i
    sq <- n_c * q0
    for (r in seq_len(npaths)) {
      sq[, path_zcol[r]] <- sq[, path_zcol[r]] +
        sx_c[path_x[r]] * lmat[, r]
    }
    q0ga <- q0 %*% g_a

    # G_W (raw): <G_A, dA/dW> + <g_p, dp/dW> + dq/dW
    gw <- n_c * (q0ga %*% t(q0))
    for (r in seq_len(npaths)) {
      arv <- as.numeric(q0 %*% g_a[, path_zcol[r]])
      gw <- gw + sx_c[path_x[r]] *
        (outer(lmat[, r], arv) + outer(arv, lmat[, r]))
    }
    for (r in seq_len(npaths)) {
      for (s in seq_len(npaths)) {
        gw <- gw + sxx_c[path_x[r], path_x[s]] *
          g_a[path_zcol[r], path_zcol[s]] *
          outer(lmat[, r], lmat[, s])
      }
    }
    q0gp <- as.numeric(q0 %*% g_p)
    gw <- gw + outer(q0gp, se_c)
    for (r in seq_len(npaths)) {
      gw <- gw + g_p[path_zcol[r]] * outer(lmat[, r], sxe_c[path_x[r], ])
    }
    # sum_i e~_i e~_i'
    psx <- as.numeric(pmat %*% sx_c)
    see <- syy_c - outer(sy_c, mu_y) - outer(mu_y, sy_c) +
      n_c * mu_y_op -
      t(sxy_c) %*% pmat_t - pmat %*% sxy_c +
      outer(mu_y, psx) + outer(psx, mu_y) +
      pmat %*% sxx_c %*% pmat_t
    gw <- gw + see
    gw <- (gw + t(gw)) / 2

    # chain W = Sigma_w[o,o]^{-1} (embedded) -> Sigma_w,
    # including the n_c ln|Sigma_w[o,o]| term
    gsw <- n_c * w_mat - w_mat %*% gw %*% w_mat

    # q0 adjoint: sum_i (2 W Q_i G_A + W e~_i g_p')
    gq0 <- 2 * (w_mat %*% (sq %*% g_a)) +
      outer(as.numeric(w_mat %*% se_c), g_p)

    # lmat adjoint
    glm <- matrix(0, p1, npaths)
    for (r in seq_len(npaths)) {
      xq <- sx_c[path_x[r]] * q0
      for (s in seq_len(npaths)) {
        xq[, path_zcol[s]] <- xq[, path_zcol[s]] +
          sxx_c[path_x[r], path_x[s]] * lmat[, s]
      }
      glm[, r] <- 2 * as.numeric(w_mat %*% (xq %*% g_a[, path_zcol[r]])) +
        g_p[path_zcol[r]] * as.numeric(w_mat %*% sxe_c[path_x[r], ])
    }

    # mu_y adjoint: -sum_i (W Q_i g_p + 2 W e~_i)
    gmy <- -as.numeric(w_mat %*% (as.numeric(sq %*% g_p) + 2 * se_c))

    # P adjoint: -[ W sum_i (Q_i g_p) x_i' + 2 W sum_i e~_i x_i' ]
    mm <- outer(q0gp, sx_c)
    for (r in seq_len(npaths)) {
      mm <- mm + g_p[path_zcol[r]] * outer(lmat[, r], sxx_c[path_x[r], ])
    }
    gp_mat <- -(w_mat %*% (mm + 2 * t(sxe_c)))

    gphi[j, i_sw] <<- gphi[j, i_sw] + as.numeric(gsw)
    gphi[j, i_muy] <<- gphi[j, i_muy] + gmy
    if (nx > 0L) {
      gphi[j, i_p] <<- gphi[j, i_p] + as.numeric(gp_mat)
    }
    gphi[j, i_lm] <<- gphi[j, i_lm] + as.numeric(glm)
    gphi[j, i_q0] <<- gphi[j, i_q0] + as.numeric(gq0)
    invisible(NULL)
  }

  if (is.null(rs_stats$mp)) {
    # -------------------- complete data --------------------
    w_mat <- lav_mat_sym_inverse(imp$sigma.w,
      logdet = TRUE, sinv_method = sinv_method
    )
    if (!is.finite(attr(w_mat, "logdet"))) {
      return(NULL)
    }
    wl <- w_mat %*% lmat
    ltwl <- crossprod(lmat, wl)
    wq0 <- w_mat %*% q0
    q0twq0 <- crossprod(q0, wq0)
    q0twl <- crossprod(q0, wl)
    nj_all <- rs_stats$cluster.size

    for (j in seq_len(nclusters)) {
      nj <- nj_all[j]
      sy_j <- rs_stats$sy[j, ]
      sx_j <- rs_stats$sx[j, ]
      sxx_j <- rs_stats$sxx[[j]]
      sxy_j <- rs_stats$sxy[[j]]
      syy_j <- rs_stats$syy[[j]]

      se_j <- sy_j - nj * mu_y - as.numeric(pmat %*% sx_j)
      sxe_j <- sxy_j - outer(sx_j, mu_y) - sxx_j %*% pmat_t

      a_j <- nj * q0twq0
      p_j <- as.numeric(crossprod(wq0, se_j))
      cz <- matrix(0, pv, pv)
      for (i in seq_len(npaths)) {
        zcol <- path_zcol[i]
        cz[, zcol] <- cz[, zcol] + q0twl[, i] * sx_j[path_x[i]]
      }
      a_j <- a_j + cz + t(cz)
      for (i in seq_len(npaths)) {
        for (k in seq_len(npaths)) {
          a_j[path_zcol[i], path_zcol[k]] <-
            a_j[path_zcol[i], path_zcol[k]] +
            ltwl[i, k] * sxx_j[path_x[i], path_x[k]]
        }
      }
      for (i in seq_len(npaths)) {
        zi <- path_zcol[i]
        p_j[zi] <- p_j[zi] + sum(wl[, i] * sxe_j[path_x[i], ])
      }

      if (kz > 0L) {
        a_j <- a_j + a_zb
        p_j <- p_j + as.numeric(crossprod(wzg, zb_all[j, ] - mu_zb))
      }

      adj <- post_adj(j, a_j, p_j)
      if (kz > 0L) {
        add_zcell(j, adj$g_a, adj$g_p)
      }
      add_cell(
        j, w_mat, nj, sx_j, sxx_j, sy_j, sxy_j, syy_j,
        se_j, sxe_j, adj$g_a, adj$g_p
      )
    }
  } else {
    # -------------------- missing data --------------------
    mp <- rs_stats$mp
    npat <- mp$npatterns

    # per-pattern (embedded) weight matrices and constants
    w_list <- vector("list", npat)
    for (p in seq_len(npat)) {
      o <- mp$pattern[[p]]$o
      w_o <- lav_mat_sym_inverse(imp$sigma.w[o, o, drop = FALSE],
        logdet = TRUE, sinv_method = sinv_method
      )
      if (!is.finite(attr(w_o, "logdet"))) {
        return(NULL)
      }
      w_p <- matrix(0, p1, p1)
      w_p[o, o] <- w_o
      w_list[[p]] <- w_p
    }

    # pass 1: per-cluster A_j and p_j
    a_arr <- array(0, dim = c(pv, pv, nclusters))
    p_all <- matrix(0, nclusters, pv)
    for (p in seq_len(npat)) {
      cell <- mp$pattern[[p]]
      w_p <- w_list[[p]]
      wl <- w_p %*% lmat
      ltwl <- crossprod(lmat, wl)
      wq0 <- w_p %*% q0
      q0twq0 <- crossprod(q0, wq0)
      q0twl <- crossprod(q0, wl)
      for (jj in seq_along(cell$j1)) {
        j <- cell$j1[jj]
        n_jp <- cell$n[jj]
        sy_jp <- cell$sy[jj, ]
        sx_jp <- cell$sx[jj, ]
        sxx_jp <- cell$sxx[[jj]]
        sxy_jp <- cell$sxy[[jj]]

        se_jp <- sy_jp - n_jp * mu_y - as.numeric(pmat %*% sx_jp)
        sxe_jp <- sxy_jp - outer(sx_jp, mu_y) - sxx_jp %*% pmat_t

        a_jp <- n_jp * q0twq0
        p_jp <- as.numeric(crossprod(wq0, se_jp))
        cz <- matrix(0, pv, pv)
        for (i in seq_len(npaths)) {
          zcol <- path_zcol[i]
          cz[, zcol] <- cz[, zcol] + q0twl[, i] * sx_jp[path_x[i]]
        }
        a_jp <- a_jp + cz + t(cz)
        for (i in seq_len(npaths)) {
          for (k in seq_len(npaths)) {
            a_jp[path_zcol[i], path_zcol[k]] <-
              a_jp[path_zcol[i], path_zcol[k]] +
              ltwl[i, k] * sxx_jp[path_x[i], path_x[k]]
          }
        }
        for (i in seq_len(npaths)) {
          zi <- path_zcol[i]
          p_jp[zi] <- p_jp[zi] + sum(wl[, i] * sxe_jp[path_x[i], ])
        }
        a_arr[, , j] <- a_arr[, , j] + a_jp
        p_all[j, ] <- p_all[j, ] + p_jp
      }
    }

    # z-outcome contributions (once per cluster)
    if (kz > 0L) {
      for (j in seq_len(nclusters)) {
        a_arr[, , j] <- a_arr[, , j] + a_zb
        p_all[j, ] <- p_all[j, ] +
          as.numeric(crossprod(wzg, zb_all[j, ] - mu_zb))
      }
    }

    # posterior adjoints per cluster
    ga_list <- vector("list", nclusters)
    gp_list <- vector("list", nclusters)
    for (j in seq_len(nclusters)) {
      adj <- post_adj(j, matrix(a_arr[, , j], pv, pv), p_all[j, ])
      ga_list[[j]] <- adj$g_a
      gp_list[[j]] <- adj$g_p
      if (kz > 0L) {
        add_zcell(j, adj$g_a, adj$g_p)
      }
    }

    # pass 2: cell collapse
    for (p in seq_len(npat)) {
      cell <- mp$pattern[[p]]
      w_p <- w_list[[p]]
      for (jj in seq_along(cell$j1)) {
        j <- cell$j1[jj]
        n_jp <- cell$n[jj]
        sy_jp <- cell$sy[jj, ]
        sx_jp <- cell$sx[jj, ]
        sxx_jp <- cell$sxx[[jj]]
        sxy_jp <- cell$sxy[[jj]]
        syy_jp <- cell$syy[[jj]]
        se_jp <- sy_jp - n_jp * mu_y - as.numeric(pmat %*% sx_jp)
        sxe_jp <- sxy_jp - outer(sx_jp, mu_y) - sxx_jp %*% pmat_t
        add_cell(
          j, w_p, n_jp, sx_jp, sxx_jp, sy_jp, sxy_jp, syy_jp,
          se_jp, sxe_jp, ga_list[[j]], gp_list[[j]]
        )
      }
    }
  }

  gphi
}

# gradient of the -2 loglikelihood wrt the free parameters
# (per_cluster = TRUE: the per-cluster scores d loglik_j / d theta,
#  on the natural loglikelihood scale)
lav_mvn_cl_rs_grad <- function(lavmodel = NULL, rs = NULL,
                               sinv_method = "eigen",
                               per_cluster = FALSE,
                               h = 1e-06) {
  # latent covariates: posterior-weighted adjoint pass per quadrature
  # node (numerical fallback on failure, via the callers)
  if (isTRUE(rs$info$nl.flag)) {
    out <- try(lav_mvn_cl_rs_grad_nl(
      lavmodel = lavmodel, rs = rs, sinv_method = sinv_method,
      per_cluster = per_cluster, h = h
    ), silent = TRUE)
    if (inherits(out, "try-error")) {
      return(NULL)
    }
    return(out)
  }
  imp0 <- lav_mvn_cl_rs_implied(lavmodel, rs_info = rs$info)
  if (is.null(imp0)) {
    return(NULL)
  }
  gphi <- lav_mvn_cl_rs_dfphi(
    rs_stats = rs$stats, imp = imp0, rs_info = rs$info,
    sinv_method = sinv_method
  )
  if (is.null(gphi)) {
    return(NULL)
  }

  # the (small) Jacobian of the implied map: central differences,
  # no cluster loops
  x0 <- lav_model_get_parameters(lavmodel)
  npar <- length(x0)
  phi0 <- lav_mvn_cl_rs_phi_vec(imp0)
  jac <- matrix(0, length(phi0), npar)
  for (k in seq_len(npar)) {
    h_k <- h * max(1, abs(x0[k]))
    x_right <- x_left <- x0
    x_right[k] <- x0[k] + h_k
    x_left[k] <- x0[k] - h_k
    imp_r <- lav_mvn_cl_rs_implied(
      lav_model_set_parameters(lavmodel, x = x_right),
      rs_info = rs$info
    )
    imp_l <- lav_mvn_cl_rs_implied(
      lav_model_set_parameters(lavmodel, x = x_left),
      rs_info = rs$info
    )
    if (is.null(imp_r) || is.null(imp_l)) {
      return(NULL)
    }
    jac[, k] <- (lav_mvn_cl_rs_phi_vec(imp_r) -
      lav_mvn_cl_rs_phi_vec(imp_l)) / (2 * h_k)
  }

  if (per_cluster) {
    -0.5 * (gphi %*% jac)
  } else {
    as.numeric(colSums(gphi) %*% jac)
  }
}

# H0 scaling correction factor (cfr. the Mplus MLR output):
#
#   c_H0 = tr( H^{-1} J ) / npar
#
# with H the total *observed* information (minus the second derivative
# of the loglikelihood; via lav_model_hessian() with method =
# "central": central differences of the analytic gradient, relative
# step size) and J = crossprod of the per-cluster scores (the
# first-order information); this is the ingredient for scaled
# (loglikelihood) difference tests
lav_mvn_cl_rs_scaling_h0 <- function(lavobject = NULL, rs = NULL,
                                     h = 1e-04) {
  lavmodel <- lavobject@Model
  sc <- try(lav_mvn_cl_rs_scores(lavmodel = lavmodel, rs = rs),
            silent = TRUE)
  if (inherits(sc, "try-error")) {
    return(as.numeric(NA))
  }
  j_mat <- crossprod(sc)

  # hessian of the objective ( = -2 loglik / (2 N) ): the observed
  # information is hessian * ntotal
  h_obj <- lav_model_hessian(
    lavmodel = lavmodel,
    lavsamplestats = lavobject@SampleStats,
    lavdata = lavobject@Data,
    lavoptions = lavobject@Options,
    lavcache = lavobject@Cache,
    h = h, method = "central"
  )
  if (anyNA(h_obj)) {
    return(as.numeric(NA))
  }
  h_mat <- h_obj * lavobject@SampleStats@ntotal

  npar <- lavmodel@nx.free
  out <- try(sum(diag(solve(h_mat, j_mat))) / npar, silent = TRUE)
  if (inherits(out, "try-error") || !is.finite(out)) {
    return(as.numeric(NA))
  }
  out
}


# ---------------------------------------------------------------------
# per-cluster scores and the first-order (`meat') information
# ---------------------------------------------------------------------

# per-cluster scores: d loglik_j / d theta (natural loglikelihood
# scale); analytic (adjoint pass), with a numerical fallback
# (central differences of the per-cluster loglikelihood vector)
#
# returns a (nclusters x npar) matrix
lav_mvn_cl_rs_scores <- function(lavmodel = NULL, rs = NULL,
                                 h = 1e-05) {
  # analytic scores
  sc <- try(
    lav_mvn_cl_rs_grad(lavmodel = lavmodel, rs = rs,
                       per_cluster = TRUE),
    silent = TRUE
  )
  if (!inherits(sc, "try-error") && !is.null(sc)) {
    return(sc)
  }

  # numerical fallback
  x0 <- lav_model_get_parameters(lavmodel)
  npar <- length(x0)
  nclusters <- rs$stats$nclusters

  ll_cluster <- function(x) {
    lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
    out <- lav_mvn_cl_rs_m2ll(
      lavmodel = lavmodel2, rs = rs,
      log2pi = FALSE, minus_two = FALSE, per_cluster = TRUE
    )
    attr(out, "loglik.cluster")
  }

  sc <- matrix(0, nclusters, npar)
  for (k in seq_len(npar)) {
    h_k <- h * max(1, abs(x0[k]))
    x_right <- x_left <- x0
    x_right[k] <- x0[k] + h_k
    x_left[k] <- x0[k] - h_k
    ll_right <- ll_cluster(x_right)
    ll_left <- ll_cluster(x_left)
    if (is.null(ll_right) || is.null(ll_left)) {
      lav_msg_stop(gettext(
        "per-cluster scores could not be computed (non-finite
         loglikelihood near the parameter estimates)."))
    }
    sc[, k] <- (ll_right - ll_left) / (2 * h_k)
  }

  sc
}


# ---------------------------------------------------------------------
# empirical Bayes (posterior mean) predictions
# ---------------------------------------------------------------------
#
# - level 2: the posterior means mu0_j of the between latent variables
#   (the between factors AND the random slopes) -- the same posterior
#   as the EM E-step
# - level 1: the within factor scores, conditional on the posterior
#   means of the between random vector; because everything is linear
#   and Gaussian, plugging in mu0_j gives the *exact* posterior mean
#   E(eta_wij | all data of cluster j)
lav_mvn_cl_rs_eb <- function(lavmodel = NULL, lavdata = NULL,
                             lavcache = NULL, se = FALSE) {
  # the rs cache
  rs <- lavcache[[1]]$rs
  if (is.null(rs)) {
    rs_info <- lav_mvn_cl_rs_info(lavmodel = lavmodel,
                                  lavdata = lavdata)
    rs_stats <- lav_mvn_cl_rs_stats(
      y1 = lavdata@X[[1]], lp = lavdata@Lp[[1]], rs_info = rs_info
    )
    rs <- list(info = rs_info, stats = rs_stats)
  }
  rs_info <- rs$info
  rs_stats <- rs$stats

  # latent covariates: EAP predictions (posterior-weighted mixture
  # over the quadrature nodes)
  if (isTRUE(rs_info$nl.flag)) {
    return(lav_mvn_cl_rs_eb_nl(
      lavmodel = lavmodel, lavdata = lavdata, rs = rs, se = se
    ))
  }

  imp <- lav_mvn_cl_rs_implied(lavmodel, rs_info = rs_info)
  if (is.null(imp)) {
    lav_msg_stop(gettext(
      "model-implied matrices could not be computed."))
  }

  core <- lav_mvn_cl_rs_eb_core(
    lavmodel = lavmodel, glist = lavmodel@GLIST, lavdata = lavdata,
    rs_stats = rs_stats, rs_info = rs_info, imp = imp, se = se
  )

  # level-2 scores: the between latent variables (eta part of v_b)
  meta <- imp$meta
  l2 <- core$mu0[, seq_len(meta), drop = FALSE]
  colnames(l2) <- imp$eta.names
  out <- list(l1 = core$l1, l2 = l2, cluster.idx = core$cluster.idx)
  if (se) {
    se2 <- sqrt(pmax(core$var0[, seq_len(meta), drop = FALSE], 0))
    colnames(se2) <- imp$eta.names
    out$se2 <- se2
  }
  out
}

# the (linear-model) EB workhorse: per-cluster posterior of v_b
# (mu0, and the posterior variances var0 = diag(Sigma0)) plus the
# level-1 (within) factor scores, for the model described by
# (glist, imp, rs_info) -- which may be the *conditional* model of
# one quadrature node (latent covariates); the caller assembles
lav_mvn_cl_rs_eb_core <- function(lavmodel = NULL, glist = NULL,
                                  lavdata = NULL,
                                  rs_stats = NULL, rs_info = NULL,
                                  imp = NULL, se = FALSE) {
  p1 <- rs_info$p1
  nx <- rs_info$nx
  pv <- imp$pv
  meta <- imp$meta
  path_tab <- rs_info$path.tab
  npaths <- nrow(path_tab)
  z_v_idx <- imp$z.v.idx
  path_zcol <- z_v_idx[path_tab$z.idx]
  path_x <- path_tab$x.idx
  nclusters <- rs_stats$nclusters
  nj_all <- rs_stats$cluster.size

  sigma_w_inv <- lav_mat_sym_inverse(imp$sigma.w, logdet = TRUE)
  if (!is.finite(attr(sigma_w_inv, "logdet"))) {
    lav_msg_stop(gettext(
      "within covariance matrix is not positive definite."))
  }
  w_mat <- sigma_w_inv
  sigma_v <- imp$sigma.v
  mu_y <- imp$mu.y

  # constants (as in the loglikelihood/E-step)
  wl <- w_mat %*% imp$lmat
  wq0 <- w_mat %*% imp$q0
  q0twq0 <- crossprod(imp$q0, wq0)
  q0twl <- crossprod(imp$q0, wl)
  ltwl <- crossprod(imp$lmat, wl)
  dpv <- diag(pv)

  # between-only endogenous z-outcomes
  kz <- rs_info$kz
  if (kz > 0L) {
    wzb <- lav_mat_sym_inverse(imp$sigma.z)
    wzg <- wzb %*% imp$gmat
    a_zb <- crossprod(imp$gmat, wzg)
    zb_all <- rs_stats$zb
    mu_zb <- imp$mu.z
  }

  # ---- per-cluster posterior of v_b: mu0_j (and posterior sd) ----
  pmat <- imp$P
  pmat_t <- t(pmat)
  mu0 <- matrix(0, nclusters, pv)
  var0 <- if (se) matrix(0, nclusters, pv) else NULL

  if (!is.null(rs_stats$mp)) {
    # missing data: pattern-based accumulation of A_j and p_j
    # (cfr. lav_mvn_cl_rs_loglik_m)
    mp <- rs_stats$mp
    a_arr <- array(0, dim = c(pv, pv, nclusters))
    p_all <- matrix(0, nclusters, pv)
    for (p in seq_len(mp$npatterns)) {
      cell <- mp$pattern[[p]]
      o <- cell$o
      w_o <- lav_mat_sym_inverse(imp$sigma.w[o, o, drop = FALSE],
        logdet = TRUE)
      if (!is.finite(attr(w_o, "logdet"))) {
        lav_msg_stop(gettext(
          "within covariance matrix is not positive definite."))
      }
      w_p <- matrix(0, p1, p1)
      w_p[o, o] <- w_o
      wl_p <- w_p %*% imp$lmat
      wq0_p <- w_p %*% imp$q0
      q0twq0_p <- crossprod(imp$q0, wq0_p)
      q0twl_p <- crossprod(imp$q0, wl_p)
      ltwl_p <- crossprod(imp$lmat, wl_p)
      for (jj in seq_along(cell$j1)) {
        j <- cell$j1[jj]
        n_jp <- cell$n[jj]
        sy_jp <- cell$sy[jj, ]
        sx_jp <- cell$sx[jj, ]
        sxx_jp <- cell$sxx[[jj]]
        sxy_jp <- cell$sxy[[jj]]
        se_jp <- sy_jp - n_jp * mu_y - as.numeric(pmat %*% sx_jp)
        sxe_jp <- sxy_jp - outer(sx_jp, mu_y) - sxx_jp %*% pmat_t
        a_jp <- n_jp * q0twq0_p
        p_jp <- as.numeric(crossprod(wq0_p, se_jp))
        cz <- matrix(0, pv, pv)
        for (i in seq_len(npaths)) {
          zcol <- path_zcol[i]
          cz[, zcol] <- cz[, zcol] + q0twl_p[, i] * sx_jp[path_x[i]]
        }
        a_jp <- a_jp + cz + t(cz)
        for (i in seq_len(npaths)) {
          for (k in seq_len(npaths)) {
            a_jp[path_zcol[i], path_zcol[k]] <-
              a_jp[path_zcol[i], path_zcol[k]] +
              ltwl_p[i, k] * sxx_jp[path_x[i], path_x[k]]
          }
        }
        for (i in seq_len(npaths)) {
          zi <- path_zcol[i]
          p_jp[zi] <- p_jp[zi] + sum(wl_p[, i] * sxe_jp[path_x[i], ])
        }
        a_arr[, , j] <- a_arr[, , j] + a_jp
        p_all[j, ] <- p_all[j, ] + p_jp
      }
    }
    for (j in seq_len(nclusters)) {
      a_j <- matrix(a_arr[, , j], pv, pv)
      if (kz > 0L) {
        a_j <- a_j + a_zb
        p_all[j, ] <- p_all[j, ] +
          as.numeric(crossprod(wzg, zb_all[j, ] - mu_zb))
      }
      if (rs_info$nexo.b > 0L) {
        d_j <- imp$mu.v +
          as.numeric(imp$cc %*% (rs_stats$exo.b[j, ] - imp$mu.exo))
      } else {
        d_j <- imp$mu.v
      }
      z_j <- dpv + a_j %*% sigma_v
      sigma0_j <- sigma_v %*% solve(z_j)
      sigma0_j <- (sigma0_j + t(sigma0_j)) / 2
      p_tilde <- p_all[j, ] - as.numeric(a_j %*% d_j)
      mu0[j, ] <- d_j + as.numeric(sigma0_j %*% p_tilde)
      if (se) {
        var0[j, ] <- diag(sigma0_j)
      }
    }
  } else {
  for (j in seq_len(nclusters)) {
    nj <- nj_all[j]
    sy_j <- rs_stats$sy[j, ]
    sx_j <- rs_stats$sx[j, ]
    sxx_j <- rs_stats$sxx[[j]]
    sxy_j <- rs_stats$sxy[[j]]

    # centered sums: e~_i = y_i - mu_y - P x_i (P = the reduced-form
    # *fixed* slopes of y on the rv covariates)
    se_j <- sy_j - nj * mu_y - as.numeric(pmat %*% sx_j)
    a_j <- nj * q0twq0
    p_j <- as.numeric(crossprod(wq0, se_j))
    cz <- matrix(0, pv, pv)
    for (i in seq_len(npaths)) {
      zcol <- path_zcol[i]
      cz[, zcol] <- cz[, zcol] + q0twl[, i] * sx_j[path_x[i]]
    }
    a_j <- a_j + cz + t(cz)
    for (i in seq_len(npaths)) {
      for (k in seq_len(npaths)) {
        a_j[path_zcol[i], path_zcol[k]] <-
          a_j[path_zcol[i], path_zcol[k]] +
          ltwl[i, k] * sxx_j[path_x[i], path_x[k]]
      }
    }
    sxe_j <- sxy_j - outer(sx_j, mu_y) - sxx_j %*% pmat_t
    for (i in seq_len(npaths)) {
      zi <- path_zcol[i]
      p_j[zi] <- p_j[zi] + sum(wl[, i] * sxe_j[path_x[i], ])
    }

    if (kz > 0L) {
      a_j <- a_j + a_zb
      p_j <- p_j + as.numeric(crossprod(wzg, zb_all[j, ] - mu_zb))
    }

    if (rs_info$nexo.b > 0L) {
      d_j <- imp$mu.v +
        as.numeric(imp$cc %*% (rs_stats$exo.b[j, ] - imp$mu.exo))
    } else {
      d_j <- imp$mu.v
    }
    z_j <- dpv + a_j %*% sigma_v
    sigma0_j <- sigma_v %*% solve(z_j)
    sigma0_j <- (sigma0_j + t(sigma0_j)) / 2
    p_tilde <- p_j - as.numeric(a_j %*% d_j)
    mu0[j, ] <- d_j + as.numeric(sigma0_j %*% p_tilde)
    if (se) {
      var0[j, ] <- diag(sigma0_j)
    }
  }
  } # complete data

  # ---- level-1 scores: within factor scores given (mu0_j, x_ij) ----
  # within model pieces (same construction as lav_mvn_cl_rs_implied;
  # note: from 'glist', which may hold node-injected beta_w cells)
  nmat <- lavmodel@nmat
  mm_w <- 1:nmat[1]
  mlist_w <- glist[mm_w]
  dimnames_w <- lavmodel@dimNames[mm_w]
  names(dimnames_w) <- names(lavmodel@GLIST)[mm_w]
  ov_names_w <- dimnames_w[["lambda"]][[1]]
  lv_names_w <- dimnames_w[["lambda"]][[2]]
  y_idx <- match(rs_info$y.names, ov_names_w)
  x_lv_idx <- match(rs_info$x.names, lv_names_w)
  lambda_w <- mlist_w$lambda
  psi_w <- mlist_w$psi
  beta_w <- mlist_w$beta
  alpha_w <- mlist_w$alpha
  nlv_w <- ncol(lambda_w)
  if (is.null(beta_w)) {
    ib_inv_w <- diag(nlv_w)
  } else {
    ib_inv_w <- solve(diag(nlv_w) - beta_w)
  }
  psi_w0 <- psi_w
  alpha_w0 <- if (is.null(alpha_w)) {
    matrix(0, nlv_w, 1L)
  } else {
    alpha_w
  }
  if (length(x_lv_idx) > 0L) {
    psi_w0[x_lv_idx, ] <- 0
    psi_w0[, x_lv_idx] <- 0
    alpha_w0[x_lv_idx, 1L] <- 0
  }

  # 'real' latent variables: not an observed variable (dummy lv), and
  # not an rv covariate
  ov_names_data <- lavdata@ov.names[[1]]
  real_idx <- which(!lv_names_w %in% ov_names_data)

  y1 <- lavdata@X[[1]]
  yy <- y1[, rs_info$y.data.idx, drop = FALSE]
  xx <- y1[, rs_info$x.data.idx, drop = FALSE]
  cluster_idx <- lavdata@Lp[[1]]$cluster.idx[[2]]
  n1 <- nrow(y1)

  if (length(real_idx) > 0L) {
    # residuals: y_ij - E(y_ij | mu0_j, x_ij)
    vhat <- mu0[cluster_idx, , drop = FALSE]
    yhat <- matrix(mu_y, n1, p1, byrow = TRUE) +
      xx %*% t(imp$P) + vhat %*% t(imp$q0)
    # per-case slope inputs: z-hat * x
    zx <- matrix(0, n1, npaths)
    for (i in seq_len(npaths)) {
      zx[, i] <- vhat[, path_zcol[i]] * xx[, path_x[i]]
      yhat <- yhat + outer(zx[, i], imp$lmat[, i])
    }
    e_mat <- yy - yhat

    # conditional mean of the (full) within lv vector:
    # (I - B)^{-1} [ alpha0 + x (in the x-dummy slots)
    #                       + zhat*x (in the lhs equations) ]
    t_inp <- matrix(alpha_w0[, 1L], n1, nlv_w, byrow = TRUE)
    if (length(x_lv_idx) > 0L) {
      t_inp[, x_lv_idx] <- t_inp[, x_lv_idx, drop = FALSE] + xx
    }
    for (i in seq_len(npaths)) {
      lhs_col <- match(path_tab$lhs[i], lv_names_w)
      if (!is.na(lhs_col)) {
        t_inp[, lhs_col] <- t_inp[, lhs_col] + zx[, i]
      }
    }
    m_full <- t_inp %*% t(ib_inv_w)

    # regression update: + Cov(eta, y) Sigma_w^{-1} e
    veta0 <- ib_inv_w %*% psi_w0 %*% t(ib_inv_w)
    ceta_y <- (veta0 %*% t(lambda_w[y_idx, , drop = FALSE]))[real_idx, ,
      drop = FALSE
    ]
    if (is.null(rs_stats$mp)) {
      upd <- e_mat %*% (w_mat %*% t(ceta_y))
    } else {
      # missing data: the update uses the observed coordinates only
      # (cases with no observed outcomes get the plain conditional
      # mean, without a residual update)
      upd <- matrix(0, n1, length(real_idx))
      obs <- !is.na(yy)
      case_id <- apply(1L * obs, 1L, paste, collapse = "")
      for (pid in unique(case_id)) {
        ridx <- which(case_id == pid)
        o <- which(obs[ridx[1L], ])
        if (length(o) == 0L) {
          next
        }
        w_o <- lav_mat_sym_inverse(imp$sigma.w[o, o, drop = FALSE])
        upd[ridx, ] <- e_mat[ridx, o, drop = FALSE] %*%
          (w_o %*% t(ceta_y[, o, drop = FALSE]))
      }
    }
    l1 <- m_full[, real_idx, drop = FALSE] + upd
    colnames(l1) <- lv_names_w[real_idx]
  } else {
    l1 <- matrix(0, n1, 0L)
  }

  list(mu0 = mu0, var0 = var0, l1 = l1, cluster.idx = cluster_idx)
}

# EAP predictions for latent-covariate (quadrature) models: the
# posterior of the cluster-level random effects is a *mixture* over
# the quadrature nodes, with the posterior node weights
# pi_jq = exp(h_jq - loglik_j) (the same weights that drive the
# analytic gradient). Per node, the conditional model is
# linear-Gaussian, so the ordinary EB workhorse applies; the mixture
# is then assembled:
#
#   E[z_nl | y_j]   = sum_q pi_jq b_q
#   E[v_lin | y_j]  = sum_q pi_jq mu0_jq
#   Var[.. | y_j]   = sum_q pi_jq (Sigma0 + mu0 mu0') - mean mean'
#
# and analogously for the level-1 (within) factor scores, which are
# linear in the cluster-level effects per node
lav_mvn_cl_rs_eb_nl <- function(lavmodel = NULL, lavdata = NULL,
                                rs = NULL, se = FALSE) {
  rs_info <- rs$info
  rs_stats <- rs$stats

  cond <- lav_mvn_cl_rs_cond(
    lavmodel = lavmodel, rs_info = rs_info, exo_b = rs_stats$exo.b
  )
  if (is.null(cond)) {
    lav_msg_stop(gettext(
      "model-implied matrices could not be computed."))
  }
  nnodes <- cond$nnodes
  nclusters <- rs_stats$nclusters

  # posterior node weights (cfr. lav_mvn_cl_rs_grad_nl)
  hmat <- matrix(-Inf, nclusters, nnodes)
  for (i in seq_len(nnodes)) {
    if (is.null(cond$imp[[i]])) {
      next
    }
    m2_i <- lav_mvn_cl_rs_loglik(
      rs_stats = rs_stats, imp = cond$imp[[i]],
      rs_info = cond$rs_info_c,
      log2pi = FALSE, minus_two = TRUE, per_cluster = TRUE
    )
    if (!is.finite(m2_i)) {
      next
    }
    hmat[, i] <- -0.5 * attr(m2_i, "loglik.cluster") + cond$logw[i]
  }
  if (!is.null(cond$logr)) {
    hmat <- hmat + cond$logr
  }
  hmax <- apply(hmat, 1L, max)
  if (any(!is.finite(hmax))) {
    lav_msg_stop(gettext(
      "the (quadrature) loglikelihood could not be computed for one
       or more clusters."))
  }
  loglik_j <- hmax + log(rowSums(exp(hmat - hmax)))
  pw <- exp(hmat - loglik_j) # nclusters x nnodes; rows sum to 1

  # per-node conditional EB, mixed with the posterior node weights
  beta_mm <- rs_info$nl.beta.mm
  beta_cells <- cbind(rs_info$nl.beta.row, rs_info$nl.beta.col)
  path_tab <- rs_info$path.tab
  nl_path_rows <- which(path_tab$type == "lv")
  path_nl_z <- match(path_tab$z.idx[nl_path_rows], rs_info$nl.z.idx)

  pv_c <- length(cond$lin.v)
  m_lin <- matrix(0, nclusters, pv_c) # E[v_lin | y]
  v_lin <- if (se) matrix(0, nclusters, pv_c) else NULL # E[v^2] part
  l1_mix <- NULL
  glist_i <- lavmodel@GLIST
  for (i in seq_len(nnodes)) {
    if (is.null(cond$imp[[i]])) {
      # a failed node: its posterior weights are (numerically) zero
      next
    }
    b_i <- cond$bmat[, i]
    glist_i[[beta_mm]][beta_cells] <- b_i[path_nl_z]
    core <- lav_mvn_cl_rs_eb_core(
      lavmodel = lavmodel, glist = glist_i, lavdata = lavdata,
      rs_stats = rs_stats, rs_info = cond$rs_info_c,
      imp = cond$imp[[i]], se = se
    )
    m_lin <- m_lin + pw[, i] * core$mu0
    if (se) {
      v_lin <- v_lin + pw[, i] * (core$var0 + core$mu0^2)
    }
    if (is.null(l1_mix)) {
      l1_mix <- matrix(0, nrow(core$l1), ncol(core$l1))
      colnames(l1_mix) <- colnames(core$l1)
      cluster_idx <- core$cluster.idx
    }
    if (ncol(l1_mix) > 0L) {
      l1_mix <- l1_mix + pw[cluster_idx, i] * core$l1
    }
  }
  if (se) {
    v_lin <- v_lin - m_lin^2
  }

  # EAP of the nonlinear slopes: moments of the discrete posterior
  q_nl <- length(cond$nl.v)
  bt <- t(cond$bmat) # nnodes x q_nl
  m_nl <- pw %*% bt
  v_nl <- if (se) pw %*% bt^2 - m_nl^2 else NULL

  # assemble the level-2 output in the full eta ordering
  meta <- cond$meta
  l2 <- matrix(0, nclusters, meta)
  colnames(l2) <- cond$eta.names
  se2 <- if (se) l2 else NULL
  for (m in seq_len(meta)) {
    lin_pos <- match(m, cond$lin.v)
    if (!is.na(lin_pos)) {
      l2[, m] <- m_lin[, lin_pos]
      if (se) {
        se2[, m] <- sqrt(pmax(v_lin[, lin_pos], 0))
      }
    } else {
      nl_pos <- match(m, cond$nl.v)
      l2[, m] <- m_nl[, nl_pos]
      if (se) {
        se2[, m] <- sqrt(pmax(v_nl[, nl_pos], 0))
      }
    }
  }

  out <- list(l1 = l1_mix, l2 = l2, cluster.idx = cluster_idx)
  if (se) {
    out$se2 <- se2
  }
  out
}

# efa related functions
# YR - April 2019
#
# the lav_model_efa_rotate_x() function was based on a script originally
# written by Florian Scharf (Muenster University, Germany)

# apply a (block-level) rotation matrix 'hg' to all factor-related model
# matrices of a single block: lambda, psi, beta and alpha. This is the common
# transform used both when rotating the solution (lav_model_efa_rotate_x()) and
# when aligning the rotated factors of a bootstrap sample to the reference
# solution (lav_efa_bootstrap_align_coef()). nu and theta need no rotation.
#
# The general formulas below hold for any invertible 'hg'. For an orthonormal
# signed-permutation 'hg' (as used by the bootstrap alignment) they reduce to
# lambda %*% hg, t(hg) %*% beta %*% hg, etc., since solve(t(hg)) == hg.
lav_model_efa_apply_hg <- function(mlist = NULL, hg = NULL) {
  # 1. lambda
  mlist$lambda <- t(solve(hg, t(mlist$lambda)))

  # 2. psi (note: eq 22 Asp & Muthen, 2009: transpose reversed)
  mlist$psi <- t(hg) %*% mlist$psi %*% hg

  # 3. beta
  if (!is.null(mlist$beta)) {
    mlist$beta <- t(hg) %*% t(solve(hg, t(mlist$beta)))
  }

  # 4. alpha
  if (!is.null(mlist$alpha)) {
    mlist$alpha <- t(hg) %*% mlist$alpha
  }

  mlist
}

# reconstruct the full block-g LAMBDA from a block's model matrices, filling in
# the dummy latent variables used for observed-variable predictors/outcomes (and
# then removing them). Used by the rotation routines below.
lav_model_efa_block_lambda <- function(lavmodel = NULL, mlist = NULL, g = 1L) {
  lav_lisrel_lambda(
    mlist = mlist,
    ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
    ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
    ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
    ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
    remove_dummy_lv = TRUE
  )
}

# select the target (or target_mask) matrix for a single efa/esem set: a
# *named* list is keyed by the efa block labels (issue #516), so we return the
# element matching 'set_name'; an unnamed list is interpreted per group and is
# returned as-is (the group argument of lav_mat_rotate() selects the element
# downstream); a plain matrix (used for all sets) is also returned as-is.
lav_model_efa_target_set <- function(x = NULL, set_name = NULL,
                                     what = "target") {
  if (!is.list(x)) {
    return(x)
  }
  x_names <- names(x)
  if (is.null(x_names) || !any(nzchar(x_names))) {
    # unnamed list: per-group target/target_mask
    return(x)
  }
  if (length(set_name) == 0L || is.na(set_name)) {
    lav_msg_stop(gettextf(
      "rotation.args$%s is a named list, but the efa block labels could not be found in the model.",
      what))
  }
  idx <- match(set_name, x_names)
  if (is.na(idx)) {
    lav_msg_stop(gettextf(
      "rotation.args$%1$s is a named list, but contains no element for efa block \"%2$s\".",
      what, set_name))
  }
  x[[idx]]
}

# early validation of *named* per-set target/target_mask lists (issue #516):
# all names must be efa block labels, and every efa block with at least two
# factors needs an entry. Called by lav_model() so a mismatch is caught
# before the model is estimated.
lav_model_efa_check_target <- function(ropts = NULL, lavpartable = NULL,
                                       efa_values = NULL) {
  for (what in c("target", "target_mask")) {
    x <- ropts[[what]]
    if (!is.list(x)) {
      next
    }
    x_names <- names(x)
    if (is.null(x_names) || !any(nzchar(x_names))) {
      # unnamed list: per-group target/target_mask
      next
    }
    unknown <- setdiff(x_names, efa_values)
    if (length(unknown) > 0L) {
      lav_msg_stop(gettextf(
        "some names of the rotation.args$%1$s list (%2$s) do not match any efa block label in the model (%3$s).",
        what, lav_msg_view(unknown, log_sep = "none"),
        lav_msg_view(efa_values, log_sep = "none")))
    }
    for (set in efa_values) {
      nfac <- length(unique(lavpartable$lhs[lavpartable$op == "=~" &
        lavpartable$efa == set]))
      if (nfac > 1L && !set %in% x_names) {
        lav_msg_stop(gettextf(
          "the rotation.args$%1$s list contains no element for efa block \"%2$s\".",
          what, set))
      }
    }
  }
  invisible(NULL)
}

# reconstruct the full block-g THETA from a block's model matrices (filling in
# the dummy observed variables). Used by the rotation routines below.
lav_model_efa_block_theta <- function(lavmodel = NULL, mlist = NULL, g = 1L) {
  lav_lisrel_theta(
    mlist = mlist,
    ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
    ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
    ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
    ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]]
  )
}

# rotate solution
lav_model_efa_rotate <- function(lavmodel = NULL, lavoptions = NULL) {

  # sanity check
  if (lavmodel@nefa == 0L || lavoptions$rotation == "none") {
    return(lavmodel)
  }

  # save warn and verbose settings
  current_warn <- lav_warn()
  current_verbose <- lav_verbose()

  # extract unrotated parameters from lavmodel
  x_orig <- lav_model_get_parameters(lavmodel, type = "free", extra = FALSE)

  # rotate, extract information from 'extra' attribute
  tmp <- lav_model_efa_rotate_x(
    x = x_orig, lavmodel = lavmodel,
    lavoptions = lavoptions, extra = TRUE
  )
  extra <- attr(tmp, "extra")
  attr(tmp, "extra") <- NULL

  out <- list(GLIST = extra$GLIST, H = extra$H, lv.order = extra$lv.order)

  # restore warn/verbose settings
  lav_warn(current_warn)
  lav_verbose(current_verbose)

  out
}


# lower-level function, needed for numDeriv
lav_model_efa_rotate_x <- function(x, lavmodel = NULL, lavoptions = NULL,
                                   init_rot = NULL, extra = FALSE,
                                   type = "free") {
  # extract rotation options from lavoptions
  method <- lavoptions$rotation
  if (method == "none") {
    return(x)
  }
  ropts <- lavoptions$rotation.args

  # place parameters into model matrices
  lavmodel_orig <- lav_model_set_parameters(lavmodel, x = x)

  # GLIST
  glist <- lavmodel_orig@GLIST

  # H per block
  h <- vector("list", lavmodel@nblocks)
  order_1 <- vector("list", lavmodel@nblocks)

  # new in 0.7-1 -- three options:
  # 1. rotate per group (default)
  # 2. group.equal = "loadings": use first lambda matrix only
  # 3. rotate all groups together + agreement (eg De Roover & Vermunt, 2019)
  mg_group_equal_flag <- mg_agreement_flag <- FALSE
  if (lavmodel@ngroups > 1) {
    if ("loadings" %in% lavoptions$group.equal) {
      mg_group_equal_flag <- TRUE
    } else if (ropts$mg_agreement) {
      mg_agreement_flag <- TRUE
    }
  }

  # option 1 + 2 (single group or multipgroup + no agreement)
  if (!mg_agreement_flag) {
    # iterate over blocks, not groups: each block (eg the within and the
    # between level of a twolevel model) is rotated independently. The
    # efa-related slots (ov.efa.idx, lv.efa.idx, nmat, the dummy-ov indices
    # and H) are all indexed per block; for single-level models block == group,
    # so multigroup behaviour is unchanged.
    for (g in seq_len(lavmodel@nblocks)) {
      # select model matrices for this group
      mm_in_group <- seq_len(lavmodel@nmat[g]) + cumsum(c(0, lavmodel@nmat))[g]
      mlist <- glist[mm_in_group]

      # general rotation matrix (all latent variables)
      h[[g]] <- hg <- diag(ncol(mlist$lambda))
      lv_order <- seq_len(ncol(mlist$lambda))

      # reconstruct full mm_lambda + THETA (in case of dummy ov's)
      lambda_g <- lav_model_efa_block_lambda(lavmodel, mlist, g)
      theta_g <- lav_model_efa_block_theta(lavmodel, mlist, g)

      # fill in optimal rotation for each set
      for (set in seq_len(lavmodel@nefa)) {
        # which ov/lv's are involved in this set?
        ov_idx <- lavmodel@ov.efa.idx[[g]][[set]]
        lv_idx <- lavmodel@lv.efa.idx[[g]][[set]]

        # empty set?
        if (length(ov_idx) == 0L) {
          next
        }

        # just 1 factor?
        if (length(lv_idx) < 2L) {
          # new in 0.6-18: reflect if needed
          if (lavoptions$rotation.args$reflect) {
            tmp <- lambda_g[ov_idx, lv_idx, drop = TRUE]
            if (sum(tmp) < 0) {
              mlist$lambda[ov_idx, 1] <- -1 * mlist$lambda[ov_idx, 1]
            }
          }
          next
        }

        # unrotated 'A' for this set
        a <- lambda_g[ov_idx, lv_idx, drop = FALSE]

        # std.ov? we use diagonal of Sigma for this set of ov's only
        if (ropts$std_ov) {
          if (mg_group_equal_flag) {
            # compute mean THETA across groups, for this set of ov's only
            theta_list <- lapply(seq_len(lavmodel@ngroups), function(gg) {
              mm_in_gg <- (seq_len(lavmodel@nmat[gg]) +
                            cumsum(c(0, lavmodel@nmat))[gg])
              lav_model_efa_block_theta(lavmodel, glist[mm_in_gg], gg)
            })
            mm_theta <- (Reduce("+", theta_list) /
                          lavmodel@ngroups)[ov_idx, ov_idx, drop = FALSE]
          } else {
            mm_theta <- theta_g[ov_idx, ov_idx, drop = FALSE]
          }
          sigma_1 <- tcrossprod(a) + mm_theta
          this_ov_var <- diag(sigma_1)
        } else {
          this_ov_var <- NULL
        }

        # init.rot?
        if (!is.null(init_rot) && lavoptions$rotation.args$jac_init_rot) {
          init_rot_1 <- init_rot[[g]][lv_idx, lv_idx, drop = FALSE]
          rstarts <- 0
        } else {
          init_rot_1 <- NULL
          rstarts <- ropts$rstarts
        }
        # per-set target/target_mask? (named list, keyed by efa block label)
        set_name <- names(lavmodel@lv.efa.idx[[g]])[set]
        this_target <- lav_model_efa_target_set(
          ropts$target, set_name, what = "target")
        this_target_mask <- lav_model_efa_target_set(
          ropts$target_mask, set_name, what = "target_mask")
        # set warn and verbose to ropts-values
        current_warn <- lav_warn()
        current_verbose <- lav_verbose()
        if (lav_warn(ropts$warn))
          on.exit(lav_warn(current_warn), TRUE)
        if (lav_verbose(ropts$verbose))
          on.exit(lav_verbose(current_verbose), TRUE)
        # rotate this set
        res <- lav_mat_rotate(
          a = a,
          orthogonal = ropts$orthogonal,
          method = method,
          method_args = list(
            geomin_epsilon = ropts$geomin_epsilon,
            orthomax_gamma = ropts$orthomax_gamma,
            cf_gamma       = ropts$cf_gamma,
            oblimin_gamma  = ropts$oblimin_gamma,
            promax_kappa   = ropts$promax_kappa,
            target         = this_target,
            target_mask    = this_target_mask
          ),
          init_rot = init_rot_1,
          init_rot_check = FALSE,
          rstarts = rstarts,
          row_weights = ropts$row_weights,
          std_ov = ropts$std_ov,
          ov_var = this_ov_var,
          algorithm = ropts$algorithm,
          reflect = ropts$reflect,
          order_lv_by = ropts$order_lv_by,
          gpa_tol = ropts$gpa_tol,
          tol = ropts$tol,
          max_iter = ropts$max_iter,
          group = g
        )

        # extract rotation matrix (note, in Asp & Muthen, 2009; this is H')
        # note: as of 0.6-6, order.idx has already been applied to ROT,
        #       so no need to reorder rows/columns after rotation
        h_efa <- res$ROT

        # fill in optimal rotation for this set
        hg[lv_idx, lv_idx] <- h_efa

        # keep track of possible re-orderings
        lv_order[lv_idx] <- lv_idx[res$order.idx]
      } # set

      # rotate all the SEM parameters (lambda, psi, beta, alpha)
      mlist <- lav_model_efa_apply_hg(mlist, hg)

      # store rotated matrices in GLIST
      glist[mm_in_group] <- mlist

      # store rotation matrix + lv.order
      h[[g]] <- hg
      order_1[[g]] <- lv_order

      # check for group.equal
      if (mg_group_equal_flag) {
        break # only rotation once
      }
    } # group

    # if group.equal = "loadings", take care of the other groups
    if (mg_group_equal_flag) {
      mlist1 <- mlist # first group
      for (g in 2:lavmodel@ngroups) {
        mm_in_group <- (seq_len(lavmodel@nmat[g]) +
                         cumsum(c(0, lavmodel@nmat))[g])
        mlist <- glist[mm_in_group]
        # transform psi/beta/alpha; lambda is forced equal across groups
        mlist <- lav_model_efa_apply_hg(mlist, hg)
        mlist$lambda <- mlist1$lambda
        glist[mm_in_group] <- mlist
        h[[g]] <- hg
        order_1[[g]] <- lv_order
      }
    }

  # option 3: rotation + agreement
  } else {
    lambda_list <- vector("list", length = lavmodel@ngroups)
    for (g in seq_len(lavmodel@ngroups)) {
      # select model matrices for this group
      mm_in_group <- seq_len(lavmodel@nmat[g]) + cumsum(c(0, lavmodel@nmat))[g]
      mlist <- glist[mm_in_group]

      # reconstruct full mm_lambda (in case of dummy ov's)
      lambda_g <- lav_model_efa_block_lambda(lavmodel, mlist, g)

      lambda_list[[g]] <- lambda_g

      # prepare H and ORDER
      h[[g]] <- diag(ncol(lambda_g))
      order_1[[g]] <- seq_len(ncol(lambda_g))
    } # group

    # fill in optimal rotation for each set
    for (set in seq_len(lavmodel@nefa)) {
      # which ov/lv's are involved in this set?
      ov_idx <- lavmodel@ov.efa.idx[[1]][[set]] # assumed equal across groups!
      lv_idx <- lavmodel@lv.efa.idx[[1]][[set]] # assumed equal across groups!

      # empty set?
      if (length(ov_idx) == 0L) {
        next
      }

      # note: single-factor sets and std.ov are not (yet) supported in the
      # multigroup rotation + agreement path; this_ov_var is always NULL here

      # unrotated 'A' for this set
      alist_1 <- lapply(lambda_list,
        function(x) x[ov_idx, lv_idx, drop = FALSE])
      this_ov_var <- NULL

      # per-set target/target_mask? (named list, keyed by efa block label)
      set_name <- names(lavmodel@lv.efa.idx[[1]])[set]
      this_target <- lav_model_efa_target_set(
        ropts$target, set_name, what = "target")
      this_target_mask <- lav_model_efa_target_set(
        ropts$target_mask, set_name, what = "target_mask")

      # init.rot?
      if (!is.null(init_rot) && lavoptions$rotation.args$jac_init_rot) {
        init_rot_list <- vector("list", length = lavmodel@ngroups)
        init_rot_list[[g]] <- init_rot[[g]][lv_idx, lv_idx, drop = FALSE]
        rstarts <- 0
      } else {
        init_rot_list <- NULL
        rstarts <- ropts$rstarts
      }

      # set warn and verbose to ropts-values
      current_warn <- lav_warn()
      current_verbose <- lav_verbose()
      if (lav_warn(ropts$warn))
        on.exit(lav_warn(current_warn), TRUE)
      if (lav_verbose(ropts$verbose))
        on.exit(lav_verbose(current_verbose), TRUE)
      if (lav_verbose()) {
        cat("\n")
      }

      # rotate this set
      res <- lav_mat_rotate_mg(
        a_list = alist_1,
        orthogonal = ropts$orthogonal,
        method = method,
        method_args = list(
          geomin_epsilon = ropts$geomin_epsilon,
          orthomax_gamma = ropts$orthomax_gamma,
          cf_gamma       = ropts$cf_gamma,
          oblimin_gamma  = ropts$oblimin_gamma,
          promax_kappa   = ropts$promax_kappa,
          target         = this_target,
          target_mask    = this_target_mask
        ),
        init_rot_list = init_rot_list,
        init_rot_check = FALSE,
        rstarts = rstarts,
        #row_weights = ropts$row_weights,
        #std.ov = ropts$std_ov,
        #ov.var = this.ov.var,
        mg_algorithm = ropts$mg_agreement_algorithm,
        mg_agreement_method = ropts$mg_agreement_method,
        mg_agreement_weight = ropts$mg_agreement_weight,
        reflect = ropts$reflect,
        order_lv_by = ropts$order_lv_by,
        #gpa.tol = ropts$gpa_tol,
        tol = ropts$tol,
        max_iter = ropts$max_iter
      )
      for (g in seq_len(lavmodel@ngroups)) {
        # fill in optimal rotation for this set, for this group
        h[[g]][lv_idx, lv_idx] <- res$rotList[[g]]

        # keep track of possible re-orderings
        order_1[[g]][lv_idx] <- lv_idx[res$orderList[[g]]]
      }
    } # set

    # rotate all the SEM parameters
    for (g in seq_len(lavmodel@ngroups)) {
      mm_in_group <- seq_len(lavmodel@nmat[g]) + cumsum(c(0, lavmodel@nmat))[g]
      mlist <- glist[mm_in_group]
      hg <- h[[g]]

      # rotate all the SEM parameters (lambda, psi, beta, alpha)
      mlist <- lav_model_efa_apply_hg(mlist, hg)

      glist[mm_in_group] <- mlist
    }

    # rescale lambda again, so that sum-of-squares per columns is the same
    # for all groups
    lambda_list <- glist[names(glist) == "lambda"]
    psi_list    <- glist[names(glist) == "psi"]

    ss_list <- lapply(lambda_list, function(x) colSums(x^2))
    ss_ave <- Reduce("+", ss_list) / length(ss_list)
    scale_factor <- vector("list", length = ncol(lambda_list[[1]]))
    for (g in seq_len(lavmodel@ngroups)) {
      scale_factor[[g]] <- (sqrt(ss_list[[g]]) / sqrt(ss_ave))
      lambda_list[[g]] <- t(t(lambda_list[[g]]) / scale_factor[[g]])
    }

    # rescale psiList accordingly
    for (g in seq_len(lavmodel@ngroups)) {
      psi_list[[g]] <- (diag(scale_factor[[g]]) %*% psi_list[[g]] %*%
                        diag(scale_factor[[g]]))
    }

    # put back in GLIST
    glist[names(glist) == "lambda"] <- lambda_list
    glist[names(glist) == "psi"]    <- psi_list

  } # option 3

  # extract all rotated parameter estimates
  x_rot <- lav_model_get_parameters(lavmodel, glist = glist, type = type)

  # extra?
  if (extra) {
    attr(x_rot, "extra") <- list(GLIST = glist, H = h, lv.order = order_1)
  }

  # return rotated parameter estimates as a vector
  x_rot
}


# lower-level function, needed for numDeriv
lav_model_efa_rotate_border_x <- function(x, lavmodel = NULL,
                                          lavoptions = NULL,
                                          lavpartable = NULL) {
  # extract rotation options from lavoptions
  method <- lavoptions$rotation
  ropts <- lavoptions$rotation.args
  method_args <- list(
    geomin_epsilon = ropts$geomin_epsilon,
    orthomax_gamma = ropts$orthomax_gamma,
    cf_gamma = ropts$cf_gamma,
    oblimin_gamma = ropts$oblimin_gamma,
    promax_kappa = ropts$promax_kappa,
    target = ropts$target,
    target_mask = ropts$target_mask
  )

  # place parameters into model matrices
  lavmodel <- lav_model_set_parameters(lavmodel, x = x)

  # GLIST
  glist <- lavmodel@GLIST

  # res
  res <- numeric(0L)

  # per block (group x level): the rotation constraints must be generated for
  # every rotated block, otherwise the bordered information matrix is rank
  # deficient (and the rotated vcov is not positive definite). The efa-related
  # slots are indexed per block; for single-level models block == group.
  for (g in seq_len(lavmodel@nblocks)) {

    # group-specific method.args
  this_method_args <- method_args

  # set group-specific target/target_mask (if needed)
    # if target, check target matrix; an *unnamed* list is interpreted per
    # group; a *named* list is keyed by the efa block labels and is resolved
    # per set (in the set loop below)
    if (method == "target.strict" || method == "pst") {
      target <- method_args$target
      if (is.list(target) &&
          (is.null(names(target)) || !any(nzchar(names(target))))) {
        this_method_args$target <- target[[g]]
      }
    }
    if (method == "pst") {
      target_mask <- method_args$target_mask
      if (is.list(target_mask) &&
          (is.null(names(target_mask)) || !any(nzchar(names(target_mask))))) {
        this_method_args$target_mask <- target_mask[[g]]
      }
    }

    # select model matrices for this group
    mm_in_group <- seq_len(lavmodel@nmat[g]) + cumsum(c(0, lavmodel@nmat))[g]
    mlist <- glist[mm_in_group]

    # reconstruct full mm_lambda + THETA (in case of dummy ov's)
    lambda_g <- lav_model_efa_block_lambda(lavmodel, mlist, g)
    theta_g <- lav_model_efa_block_theta(lavmodel, mlist, g)

    # setnames
    set_names <- lav_pt_efa_values(lavpartable)

    # for each set
    for (set in seq_len(lavmodel@nefa)) {
      # check if we have any user=7 elements in this set
      # if not, skip constraints
      ind_idx <- which(lavpartable$op == "=~" &
        lavpartable$block == g &
        lavpartable$efa == set_names[set])
      if (!any(lavpartable$user[ind_idx] == 7L)) {
        next
      }

      # which ov/lv's are involved in this set?
      ov_idx <- lavmodel@ov.efa.idx[[g]][[set]]
      lv_idx <- lavmodel@lv.efa.idx[[g]][[set]]

      # empty set?
      if (length(ov_idx) == 0L) {
        next
      }

      # just 1 factor?
      if (length(lv_idx) < 2L) {
        next
      }

      a <- lambda_g[ov_idx, lv_idx, drop = FALSE]
      p <- nrow(a)
      m <- ncol(a)

      # per-set target/target_mask? (named list, keyed by efa block label)
      set_method_args <- this_method_args
      set_method_args$target <- lav_model_efa_target_set(
        this_method_args$target, set_names[set], what = "target")
      set_method_args$target_mask <- lav_model_efa_target_set(
        this_method_args$target_mask, set_names[set], what = "target_mask")

      # for oblique, we also need PSI
      if (!ropts$orthogonal) {
        mm_psi <- mlist$psi[lv_idx, lv_idx, drop = FALSE]
      }

      # std.ov? we use diagonal of Sigma for this set of ov's only
      if (ropts$std_ov) {
        mm_theta <- theta_g[ov_idx, ov_idx, drop = FALSE]
        sigma_1 <- tcrossprod(a) + mm_theta
        this_ov_var <- diag(sigma_1)
      } else {
        this_ov_var <- rep(1, p)
      }

      # choose method
      method <- tolower(method)
      if (method %in% c(
        "cf-quartimax", "cf-varimax", "cf-equamax",
        "cf-parsimax", "cf-facparsim"
      )) {
        method_fname <- "lav_mat_rotate_cf"
        set_method_args$cf_gamma <- switch(method,
          "cf-quartimax" = 0,
          "cf-varimax"   = 1 / p,
          "cf-equamax"   = m / (2 * p),
          "cf-parsimax"  = (m - 1) / (p + m - 2),
          "cf-facparsim" = 1
        )
      } else if (method == "target.strict") {
        method_fname <- "lav_mat_rotate_target"
      } else {
        method_fname <- paste("lav_mat_rotate_", method, sep = "")
      }

      # check if rotation method exists
      check <- try(get(method_fname), silent = TRUE)
      if (inherits(check, "try-error")) {
        lav_msg_stop(gettextf("unknown rotation method: %s", method_fname))
      }

      # 1. compute row weigths

      # 1.a cov -> cor?
      if (ropts$std_ov) {
        a <- a * 1 / sqrt(this_ov_var)
      }

      if (ropts$row_weights == "none") {
        weights <- rep(1.0, p)
      } else if (ropts$row_weights == "kaiser") {
        weights <- lav_mat_rotate_kaiser_weights(a)
      } else if (ropts$row_weights == "cureton-mulaik") {
        weights <- lav_mat_rotate_cm_weights(a)
      } else {
        lav_msg_stop(gettextf("row_weights can be",
          lav_msg_view(c("none", "kaiser", "cureton-mulaik"), "or")))
      }
      a <- a * weights

      # evaluate rotation criterion, extract GRAD
      q_1 <- do.call(
        method_fname,
        c(list(mm_lambda = a), set_method_args, list(grad = TRUE))
      )
      gq <- attr(q_1, "grad")
      attr(q_1, "grad") <- NULL

      # compute 'Z'
      z <- crossprod(a, gq)

      # compute constraints
      if (ropts$orthogonal) {
        # the constraint: Z == diagonal
        # or in other words, the non-diagonal elements of
        # Z - t(Z) are all zero
        tmp <- z - t(z)
        this_res <- lav_mat_vech(tmp, diagonal = FALSE)
      } else {
        psi_z <- mm_psi * diag(z) # rescale rows only
        tmp <- z - psi_z
        out1 <- lav_mat_vech(tmp, diagonal = FALSE)
        out2 <- lav_mat_vechu(tmp, diagonal = FALSE)
        this_res <- c(out1, out2)
      }

      res <- c(res, this_res)
    } # set
  } # block

  # return constraint vector
  res
}

# update lavdata object
# - new dataset (lav_data_update)
# - only subset of data (lav_data_update_subset) (for sam())

# YR - 18 Jan 2021 (so we don't need to export lav_data_*_patterns functions)
#    - 28 May 2023 lav_data_update_subset()

# update lavdata object with new dataset
# - assuming everything else stays the same
# - optionally, also provide boot_idx (per group) to adapt internal slots
#   (eXo, sampling weights); the row order of new_x must match boot_idx
# - optionally, also provide boot_clus (per group): a matrix of (relabeled)
#   cluster ids for the new data. This is used by the cluster bootstrap so
#   that a cluster that is resampled more than once is treated as several
#   distinct clusters. If NULL, the original cluster ids are reused.
lav_data_update <- function(lavdata = NULL,
                            new_x = NULL,
                            boot_idx = NULL,
                            boot_clus = NULL,
                            lavoptions = NULL,
                            ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  stopifnot(length(new_x) == lavdata@ngroups)
  stopifnot(!is.null(lavoptions))
  newdata <- lavdata

  # replace data 'X' slot for each group
  for (g in 1:lavdata@ngroups) {
    # replace raw data
    newdata@X[[g]] <- new_x[[g]]

    # if boot_idx is provided, also adapt eXo and sampling weights, using
    # the same (per-group) row ordering as new_x
    if (!is.null(boot_idx)) {
      boot_idx_g <- boot_idx[[g]]

      # eXo
      if (!is.null(lavdata@eXo[[g]])) {
        newdata@eXo[[g]] <- lavdata@eXo[[g]][boot_idx_g, , drop = FALSE]
      }

      # sampling weights
      if (!is.null(lavdata@weights[[g]])) {
        newdata@weights[[g]] <- lavdata@weights[[g]][boot_idx_g]
      }
    }

    # Mp + nobs
    if (lavoptions$missing != "listwise") {
      newdata@Mp[[g]] <- lav_data_mi_patterns(new_x[[g]],
        sort_freq = FALSE, coverage = TRUE
      )
      newdata@nobs[[g]] <-
        (nrow(newdata@X[[g]]) - length(newdata@Mp[[g]]$empty.idx))
    } else {
      # the number of rows may have changed (e.g., cluster bootstrap, where
      # the total N varies from sample to sample); keep nobs in sync
      newdata@nobs[[g]] <- nrow(newdata@X[[g]])
    }

    # Rp
    if (length(lavdata@ov.names.x[[g]]) == 0L &&
      all(lavdata@ov.names[[g]] %in%
        lavdata@ov$name[lavdata@ov$type == "ordered"])) {
      newdata@Rp[[g]] <- lav_data_resp_patterns(new_x[[g]])
    }

    # Lp (cluster structure): rebuild whenever there is a cluster variable.
    # This covers both two-level models (nlevels > 1) and single-level models
    # with cluster-robust standard errors (cluster = , nlevels == 1).
    if (length(lavdata@cluster) > 0L) {
      multilevel <- lavdata@nlevels > 1L
      n_clus_vars <- length(lavdata@cluster)
      # cluster id(s) for the new data
      if (!is.null(boot_clus)) {
        # cluster bootstrap: use the (relabeled) cluster ids, so that a
        # cluster that was resampled more than once is treated as several
        # distinct clusters
        clus <- boot_clus[[g]]
      } else {
        # reuse the original cluster ids (cluster.idx is stored from level 2)
        clus <- matrix(0L, nrow(new_x[[g]]), n_clus_vars)
        for (cc in seq_len(n_clus_vars)) {
          clus[, cc] <- lavdata@Lp[[g]]$cluster.idx[[cc + 1L]]
        }
      }
      # ALWAYS add ov.names.x at the end (as in lav_data_full())
      ov_names_2 <- unique(c(lavdata@ov.names[[g]], lavdata@ov.names.x[[g]]))
      newdata@Lp[[g]] <- lav_data_cl_patterns(
        y = new_x[[g]],
        clus = clus,
        cluster = lavdata@cluster,
        multilevel = multilevel,
        ov_names = ov_names_2,
        ov_names_x = lavdata@ov.names.x[[g]],
        ov_names_l = lavdata@ov.names.l[[g]]
      )
    }
  } # g

  # return updated data object
  newdata
}

# update lavdata, keeping only a subset of the observed variables
# (assuming everything else stays the same)
lav_data_update_subset <- function(lavdata = NULL, ov_names = NULL) {
  stopifnot(length(ov_names) == length(lavdata@ov.names))
  newdata <- lavdata

  # replace ov.names
  newdata@ov.names <- ov_names

  # ordered?
  if (length(lavdata@ordered) > 0L) {
    newdata@ordered <- lavdata@ordered[lavdata@ordered %in% ov_names]
  }

  # replace/update slots for each group
  for (g in 1:lavdata@ngroups) {
    # sanity check:
    if (all(lavdata@ov.names[[g]] %in% ov_names[[g]])) {
      # nothing to do
      next
    }

    # replace ov.names.x
    if (length(lavdata@ov.names.x[[g]]) > 0L) {
      newdata@ov.names.x[[g]] <-
        lavdata@ov.names.x[[g]][lavdata@ov.names.x[[g]] %in%
          ov_names[[g]]]
    }

    # replace ov.names.l
    if (newdata@nlevels > 1L) {
      for (l in 1:newdata@nlevels) {
        newdata@ov.names.l[[g]][[l]] <-
          lavdata@ov.names.l[[g]][[l]][lavdata@ov.names.l[[g]][[l]] %in%
            ov_names[[g]]]
      }
    }

    # ov table
    keep_idx <- which(lavdata@ov$name %in% unlist(ov_names))
    newdata@ov <- lapply(lavdata@ov, "[", keep_idx)

    # replace raw data
    newdata@X[[g]] <- lavdata@X[[g]][, lavdata@ov.names[[g]] %in%
      ov_names[[g]], drop = FALSE]

    # eXo
    if (length(newdata@ov.names.x[[g]]) == 0L) {
      newdata@eXo[g] <- list(NULL)
    } else {
      newdata@eXo[[g]] <- lavdata@eXo[[g]][, lavdata@ov.names.x[[g]] %in%
        ov_names[[g]], drop = FALSE]
    }

    # Mp + nobs
    if (lavdata@missing != "listwise") {
      newdata@Mp[[g]] <- lav_data_mi_patterns(newdata@X[[g]],
        sort_freq = FALSE, coverage = TRUE
      )
      newdata@nobs[[g]] <-
        (nrow(newdata@X[[g]]) - length(newdata@Mp[[g]]$empty.idx))
    }

    # Rp
    if (length(newdata@ordered) == 0L) {
      # nothing to do
    } else if (length(newdata@ov.names.x[[g]]) == 0L &&
      all(newdata@ov.names[[g]] %in%
        newdata@ov$name[newdata@ov$type == "ordered"])) {
      newdata@Rp[[g]] <- lav_data_resp_patterns(newdata@X[[g]])
    }

    # Lp: rebuild whenever there is a cluster variable; this covers both
    # two-level models (nlevels > 1) and single-level models with
    # cluster-robust standard errors (cluster = , nlevels == 1)
    if (length(newdata@cluster) > 0L) {
      # extract cluster variable(s), for this group
      # (cluster.idx is stored from level 2 onwards)
      n_clus_vars <- length(newdata@cluster)
      clus <- matrix(0L, nrow(newdata@X[[g]]), n_clus_vars)
      for (cc in seq_len(n_clus_vars)) {
        clus[, cc] <- lavdata@Lp[[g]]$cluster.idx[[cc + 1L]]
      }
      multilevel <- newdata@nlevels > 1L
      ov_names_1 <- unique(c(ov_names[[g]], newdata@ov.names.x[[g]]))
      newdata@Lp[[g]] <- lav_data_cl_patterns(
        y = newdata@X[[g]],
        clus = clus,
        cluster = newdata@cluster,
        multilevel = multilevel,
        ov_names = ov_names_1,
        ov_names_x = newdata@ov.names.x[[g]],
        ov_names_l = newdata@ov.names.l[[g]]
      )
    }
  } # g

  # return update data object
  newdata
}

# marginal ML
lav_model_lik_mml <- function(lavmodel = NULL,
                              mm_theta = NULL,
                              th = NULL,
                              glist = NULL,
                              group = 1L,
                              lavdata = NULL,
                              sample_mean = NULL,
                              sample_mean_x = NULL,
                              lavcache = NULL) {
  conditional_x <- lavmodel@conditional.x

  # data for this group
  x <- lavdata@X[[group]]
  nobs <- nrow(x)
  # nvar <- ncol(x)
  exo <- lavdata@eXo[[group]]

  # MLIST (for veta and yhat)
  mm_in_group <- 1:lavmodel@nmat[group] + cumsum(c(0, lavmodel@nmat))[group]
  mlist <- glist[mm_in_group]

  # quadrature points
  gh <- lavcache[[group]]$GH
  n_gh <- nrow(gh$x)
  # nfac <- ncol(gh$x)

  # compute VETAx (latent lv only)
  lv_dummy_idx <- c(
    lavmodel@ov.y.dummy.lv.idx[[group]],
    lavmodel@ov.x.dummy.lv.idx[[group]]
  )
  vetax <- lav_lisrel_vetax(
    mlist = mlist,
    lv_dummy_idx = lv_dummy_idx
  )
  # VETAx <- lav_lisrel_vetax(MLIST = MLIST)
  # check for negative values?
  if (any(diag(vetax) < 0)) {
    lav_msg_warn(gettext("--- VETAx contains negative values"))
    print(vetax)
    return(0)
  }

  # cholesky?
  # if(is.null(lavmodel@control$cholesky)) {
  cholesky <- TRUE
  # } else {
  #    CHOLESKY <- as.logical(lavmodel@control$cholesky)
  # if(nfac > 1L && !CHOLESKY) {
  #    warning("lavaan WARNING: CHOLESKY is OFF but nfac > 1L")
  # }
  # }

  if (!cholesky) {
    # we should still 'scale' the factors, if std.lv=FALSE
    eta_sd <- sqrt(diag(vetax))
  } else {
    # cholesky takes care of scaling
    tchol_veta <- try(chol(vetax), silent = TRUE)
    if (inherits(tchol_veta, "try-error")) {
      lav_msg_warn(gettext("--- VETAx not positive definite"))
      print(vetax)
      return(0)
    }
    if (!is.null(mlist$alpha) || !is.null(mlist$gamma)) {
      if (conditional_x) {
        eetax <- lav_lisrel_eetax(
          mlist = mlist, exo = exo, n = nobs,
          sample_mean = sample_mean,
          ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[group]],
          ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[group]],
          ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[group]],
          ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[group]]
        )
      } else {
        eeta <- lav_lisrel_eeta(
          mlist = mlist,
          mean_x = sample_mean_x,
          sample_mean = sample_mean,
          ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[group]],
          ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[group]],
          ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[group]],
          ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[group]]
        )
      }
      # if(length(lv.dummy.idx) > 0L) {
      #    EETAx <- EETAx[,-lv.dummy.idx,drop=FALSE]
      # }
    }
  }

  # compute (log)lik for each node, for each observation
  sum_log_fy <- matrix(0, nrow = n_gh, ncol = nobs)
  for (q in 1:n_gh) {
    # current value(s) for ETA
    # eta <- matrix(0, nrow = 1, ncol = ncol(MLIST$lambda))

    # non-dummy elements -> quadrature points
    # eta[1L, -lv.dummy.idx] <- GH$x[q,,drop=FALSE]
    xq <- gh$x[q, , drop = FALSE]

    # rescale/unwhiten
    if (cholesky) {
      # un-orthogonalize
      xq <- xq %*% tchol_veta
    } else {
      # no unit scale? (un-standardize)
      xq <- sweep(xq, MARGIN = 2, STATS = eta_sd, FUN = "*")
    }

    eta <- matrix(0, nrow = 1, ncol = ncol(mlist$lambda))
    if (length(lv_dummy_idx) > 0L) {
      eta[, -lv_dummy_idx] <- xq
    } else {
      eta <- xq
    }

    # eta_i = alpha + BETA eta_i + GAMMA eta_i + error
    #
    # - direct effect of BETA is already in VETAx, and hence tchol.VETA
    # - need to add alpha, and GAMMA eta_i
    if (!is.null(mlist$alpha) || !is.null(mlist$gamma)) {
      if (conditional_x) {
        eta <- sweep(eetax, MARGIN = 2, STATS = eta, FUN = "+")
      } else {
        eta <- eta + eeta
      }
    }

    # compute yhat for this node (eta)
    if (lavmodel@conditional.x) {
      yhat <- lav_lisrel_eyetax(
        mlist = mlist, exo = exo,
        eta = eta, sample_mean = sample_mean,
        ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[group]],
        ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[group]],
        ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[group]],
        ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[group]]
      )
    } else {
      yhat <- lav_lisrel_eyetax3(
        mlist = mlist,
        eta = eta, sample_mean = sample_mean,
        mean_x = sample_mean_x,
        ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[group]],
        ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[group]],
        ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[group]],
        ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[group]]
      )
    }

    # compute fy.var, for this node (eta): P(Y_i =  y_i | eta_i, x_i)
    log_fy_var <- lav_predict_fy_internal(
      x = x, yhat = yhat,
      th = th, mm_theta = mm_theta,
      num_idx = lavmodel@num.idx[[group]],
      th_idx = lavmodel@th.idx[[group]],
      link = lavmodel@link, log_1 = TRUE
    )

    # if log, fy is just the sum of log.fy.var
    log_fy <- apply(log_fy_var, 1L, sum)

    # store log likelihoods for this node
    sum_log_fy[q, ] <- log_fy
  }

  # integration
  lik <- as.numeric(t(gh$w) %*% exp(sum_log_fy))

  # avoid underflow
  idx <- which(lik < exp(-600))
  if (length(idx) > 0L) {
    lik[idx] <- exp(-600)
  }

  lik
}

lav_plot <- function(
  model = NULL,
  infile = NULL,
  varlv = FALSE,
  placenodes = NULL,
  edgelabelsbelow = NULL,
  group_covar_indicators = FALSE,
  common_opts = list(sloped_labels = TRUE,
                     mlovcolors = c("lightgreen", "lightblue"),
                     italic = TRUE,
                     lightness = 1,
                     auto_subscript = TRUE),
  rplot = list(outfile = "",
               addgrid = TRUE),
  tikz = list(outfile = "",
              cex = 1.3,
              standalone = FALSE),
  svg = list(outfile = "",
             stroke_width = 2L,
             font_size = 20L,
             idx_font_size = 15L,
             dy = 5L,
             font_family = "Latin Modern Math, arial, Arial, sans",
             standalone = FALSE)
) {
  tmp <- lav_model_plotinfo(model,
                            infile = infile,
                            varlv = varlv)
  tmp <- lav_plotinfo_positions(tmp,
                            placenodes = placenodes,
                            edgelabelsbelow = edgelabelsbelow,
                            group_covar_indicators = group_covar_indicators)
  mc <- match.call()
  create_rplot <- !is.null(mc$rplot) || (is.null(mc$tikz) && is.null(mc$svg))
  if (create_rplot)
    do.call(lav_plotinfo_rgraph, c(list(plotinfo = tmp),
                                   common_opts,
                                   rplot))
  if (!is.null(mc$tikz))
    do.call(lav_plotinfo_tikzcode, c(list(plotinfo = tmp),
                             common_opts,
                             tikz))
  if (!is.null(mc$svg))
    do.call(lav_plotinfo_svgcode, c(list(plotinfo = tmp),
                             common_opts,
                             svg))
}

# lav_start.R: provide starting values for model parameters
#
# YR 30/11/2010: initial version
# YR 08/06/2011: add fabin3 start values for factor loadings
# YR 14 Jan 2014: moved to lav_start.R

# fill in the 'ustart' column in a User data.frame with reasonable
# starting values, using the sample data

lav_start <- function(start_method = "default",
                      lavpartable = NULL,
                      lavsamplestats = NULL,
                      lavh1 = NULL, # fixme: only use lavh1?
                      model_type = "sem",
                      reflect = FALSE, # rotation only
                      samplestats_flag = TRUE,
                      order_lv_by = "none" # rotation only
                      ) {
  # check arguments
  stopifnot(is.list(lavpartable))

  # categorical?
  categorical <- any(lavpartable$op == "|")

  # correlation structure?
  correlation <- any(lavpartable$op == "~*~")

  # composites?
  composites <- any(lavpartable$op == "<~")

  # conditional.x?
  conditional_x <- any(lavpartable$exo == 1L &
    lavpartable$op %in% c("~", "<~"))
  # ord.names <- unique(lavpartable$lhs[ lavpartable$op == "|" ])

  # helper: the best available h1 (co)variance matrix for group g, to be
  # used for starting values; if conditional.x, the residual version
  # (note: lavh1$implied may not contain a cov/res.cov element at all,
  #  and NULL[[g]] would be an error, hence the length() checks)
  stats_h1_cov <- function(g) {
    if (conditional_x) {
      if (length(lavh1$implied$res.cov) >= g &&
          !is.null(lavh1$implied$res.cov[[g]])) {
        lavh1$implied$res.cov[[g]]
      } else if (length(lavsamplestats@missing.h1) >= g &&
                 !is.null(lavsamplestats@missing.h1[[g]]$res.cov)) {
        lavsamplestats@missing.h1[[g]]$res.cov
      } else {
        lavsamplestats@res.cov[[g]]
      }
    } else {
      if (length(lavh1$implied[["cov"]]) >= g &&
          !is.null(lavh1$implied[["cov"]][[g]])) {
        lavh1$implied[["cov"]][[g]]
      } else if (length(lavsamplestats@missing.h1) >= g &&
                 !is.null(lavsamplestats@missing.h1[[g]]$sigma)) {
        lavsamplestats@missing.h1[[g]]$sigma
      } else {
        lavsamplestats@cov[[g]]
      }
    }
  }

  # nlevels?
  nlevels <- lav_pt_nlevels(lavpartable)

  # reflect/order.lv.by
  if (is.null(reflect)) {
    reflect <- FALSE
  }
  if (is.null(order_lv_by)) {
    order_lv_by <- "index"
  }

  # check start.method
  #if (mimic == "lavaan") {
  start_initial <- "lavaan"
  #} else if (mimic == "Mplus") {
  #  start.initial <- "mplus"
  #} else {
  #  # FIXME: use LISREL/EQS/AMOS/.... schemes
  #  start.initial <- "lavaan"
  #}

  # start.method
  start_user <- NULL
  if (is.character(start_method)) {
    start_method_lc <- tolower(start_method)
  if (start_method_lc != "simple" && !samplestats_flag) {
      start_method_lc <- start_method <- "simple"
    }
    if (start_method_lc == "default") {
      # nothing to do
    } else if (start_method == "simple") {
      start <- numeric(length(lavpartable$ustart))
      # if(categorical || correlation) {
      start[which(lavpartable$op == "=~")] <- 0.7
      start[which(lavpartable$op == "<~")] <- 1
      # } else {
      #    start[ which(lavpartable$op == "=~") ] <- 1.0
      # }
      start[which(lavpartable$op == "~*~")] <- 1.0
      ov_names_ord <- lav_pt_vnames(lavpartable, "ov.ord")
      var_idx <- which(lavpartable$op == "~~" &
        lavpartable$lhs == lavpartable$rhs &
        !(lavpartable$lhs %in% ov_names_ord))
      start[var_idx] <- 1.0
      user_idx <- which(!is.na(lavpartable$ustart))
      start[user_idx] <- lavpartable$ustart[user_idx]
      return(start) # assuming fixed.x = FALSE!
    } else if (start_method == "est") {
      return(lavpartable$est)
    } else if (start_method_lc %in% c("simple", "lavaan")) {
      start_initial <- start_method_lc
    } else {
     lav_msg_stop(gettext("unknown value for start argument"))
    }
  } else if (is.list(start_method)) {
    start_user <- start_method
  } else if (is.numeric(start_method)) {
    nx_free <- sum(lavpartable$free > 0L)
    if (length(start_method) != nx_free) {
     lav_msg_stop(gettextf(
       "start argument contains %1$s elements; but parameter table
       expects %2$s free parameters.", length(start_method), nx_free))
    }
    lavpartable$ustart[lavpartable$free > 0L] <- start_method
  } else if (inherits(start_method, "lavaan")) {
    start_user <- parTable(start_method)
  }

  # check model list elements, if provided
  if (!is.null(start_user)) {
    if (is.null(start_user$lhs) ||
      is.null(start_user$op) ||
      is.null(start_user$rhs)) {
     lav_msg_stop(gettext(
       "problem with start argument: model list does not contain
       all elements: lhs/op/rhs"))
    }
    if (!is.null(start_user$est)) {
      # excellent, we got an est column; nothing to do
    } else if (!is.null(start_user$start)) {
      # no est column, but we use the start column
      start_user$est <- start_user$start
    } else if (!is.null(start_user$ustart)) {
      # not ideal, but better than nothing
      start_user$est <- start_user$ustart
    } else {
     lav_msg_stop(gettext(
       "problem with start argument: could not find est/start column
       in model list"))
    }
  }


  # global settings
  # 0. everything is zero
  start <- numeric(length(lavpartable$ustart))

  # 1. =~ factor loadings:
  if (categorical || correlation) {
    # if std.lv=TRUE, 0.8 is too large
    start[which(lavpartable$op == "=~")] <- 0.7
  } else {
    start[which(lavpartable$op == "=~")] <- 1.0
  }

  # 2. (residual) lv variances for latent variables
  lv_names <- lav_pt_vnames(lavpartable, "lv") # all groups
  lv_var_idx <- which(lavpartable$op == "~~" &
    lavpartable$lhs %in% lv_names &
    lavpartable$lhs == lavpartable$rhs)
  start[lv_var_idx] <- 0.05
  # start[lv.var.idx] <- 0.5 # new in 0.6-2? (for optim.parscale = "stand")

  # 3. latent response scales (if any)
  delta_idx <- which(lavpartable$op == "~*~")
  start[delta_idx] <- 1.0


  # group-specific settings
  ngroups <- lav_pt_ngroups(lavpartable)

  # for now, if no group column, add one (again), until we rewrite
  # this function to handle block/group hybrid settings
  if (is.null(lavpartable$group) && ngroups == 1L) {
    lavpartable$group <- rep(1L, length(lavpartable$lhs))
    lavpartable$group[lavpartable$block == 0L] <- 0L
  }

  # group values
  group_values <- lav_pt_group_values(lavpartable)

  for (g in 1:ngroups) {

    # info from user model for this group
    if (conditional_x) {
      ov_names <- lav_pt_vnames(lavpartable, "ov.nox",
                                                group = group_values[g])
    } else {
      ov_names <- lav_pt_vnames(lavpartable, "ov",
                                                group = group_values[g])
    }
    if (categorical) {
      ov_names_num <- lav_pt_vnames(lavpartable, "ov.num",
                                                group = group_values[g])
      ov_names_ord <- lav_pt_vnames(lavpartable, "ov.ord",
                                                group = group_values[g])
    } else {
      ov_names_num <- ov_names
    }
    lv_names <- lav_pt_vnames(lavpartable, "lv",
                                                 group = group_values[g])
    lv_names_efa <- lav_pt_vnames(lavpartable, "lv.efa",
                                                 group = group_values[g])
    ov_names_x <- lav_pt_vnames(lavpartable, "ov.x",
                                                 group = group_values[g])
    ov_ind_c <- lav_pt_vnames(lavpartable, "ov.cind",
                                                 group = group_values[g])
    lv_names_c <- lav_pt_vnames(lavpartable, "lv.composite",
                                                 group = group_values[g])

    # just for the nlevels >1 case
    ov_names <- unique(unlist(ov_names))
    ov_names_num <- unique(unlist(ov_names_num))
    lv_names <- unique(unlist(lv_names))
    lv_names_efa <- unique(unlist(lv_names_efa))
    ov_names_x <- unique(unlist(ov_names_x))
    ov_ind_c <- unique(unlist(ov_ind_c))
    lv_names_c <- unique(unlist(lv_names_c))
    lv_names_noc <- lv_names[!lv_names %in% lv_names_c]

    # residual ov variances (including exo/ind, to be overridden)
    ov_var_idx <- which(lavpartable$group == group_values[g] &
      lavpartable$op == "~~" &
      lavpartable$lhs %in% ov_names_num &
      lavpartable$lhs == lavpartable$rhs)
    sample_var_idx <- match(lavpartable$lhs[ov_var_idx], ov_names)
    if (model_type == "unrestricted") {
      h1_cov <- stats_h1_cov(g)
      start[ov_var_idx] <- diag(h1_cov)[sample_var_idx]
    } else {
      #if (start.initial == "mplus") {
      #  if (conditional.x && nlevels == 1L) {
      #    start[ov.var.idx] <-
      #      (1.0 - 0.50) * lavsamplestats@res.var[[1L]][sample.var.idx]
      #  } else {
      #    start[ov.var.idx] <-
      #      (1.0 - 0.50) * lavsamplestats@var[[1L]][sample.var.idx]
      #  }
      #} else {
        if (conditional_x && !is.null(lavsamplestats@res.cov[[g]])) {
          start[ov_var_idx] <-
            (1.0 - 0.50) * diag(lavsamplestats@res.cov[[g]])[sample_var_idx]
        } else {
          start[ov_var_idx] <-
            (1.0 - 0.50) * diag(lavsamplestats@cov[[g]])[sample_var_idx]
        }
      #}
      # composite indicators: fill in total variances
      if (composites) {
        start[ov_var_idx] <- diag(lavsamplestats@cov[[g]])[sample_var_idx]
      }
    }

    # 1-fac measurement models: loadings, psi, theta
    if (start_initial %in% c("lavaan") &&
      model_type %in% c("sem", "cfa")) {
      # fabin3 estimator (2sls) of Hagglund (1982) per factor
      for (f in lv_names_noc) {
        # not for efa factors
        if (f %in% lv_names_efa) {
          next
        }
        lambda_idx <- which(lavpartable$lhs == f &
          lavpartable$op == "=~" &
          lavpartable$group == group_values[g])
        # standardized?
        std_lv <- FALSE
        var_f_idx <- which(lavpartable$lhs == f &
          lavpartable$op == "~~" &
          lavpartable$group == group_values[g] &
          lavpartable$rhs == f)
        if (length(var_f_idx) > 0L &&
          all(lavpartable$free[var_f_idx] == 0) &&
          all(lavpartable$ustart[var_f_idx] == 1)) {
          std_lv <- TRUE
        }

        # no second order
        if (any(lavpartable$rhs[lambda_idx] %in% lv_names)) next

        # get observed indicators for this latent variable
        ov_idx <- match(lavpartable$rhs[lambda_idx], ov_names)
        if (length(ov_idx) > 0L && !any(is.na(ov_idx))) {
          if (lavsamplestats@missing.flag && nlevels == 1L) {
            h1_cov <- stats_h1_cov(g)
            cov_1 <- h1_cov[ov_idx, ov_idx, drop = FALSE]
          } else {
            if (conditional_x && !is.null(lavsamplestats@res.cov[[g]])) {
              cov_1 <- lavsamplestats@res.cov[[g]][ov_idx,
                ov_idx,
                drop = FALSE
              ]
            } else {
              cov_1 <- lavsamplestats@cov[[g]][ov_idx,
                ov_idx,
                drop = FALSE
              ]
            }
          }

          # fabin for 1-factor
          fabin <- lav_cfa_1fac_fabin(cov_1,
            std_lv = std_lv,
            lambda_only = TRUE,
            method = "fabin3"
          )

          # factor loadings
          tmp <- fabin$lambda
          tmp[!is.finite(tmp)] <- 1.0 # just in case (eg 0/0)

          # check for negative triad if nvar=3L (new in 0.6-8)
          if (!is.null(fabin$neg.triad) && fabin$neg.triad) {
            if (std_lv) {
              tmp <- rep(0.7, length(tmp))
            } else {
              tmp <- rep(1.0, length(tmp))
            }
          }

          # check for negative marker
          if (!std_lv && !is.na(lavpartable$ustart[lambda_idx[1]]) &&
              lavpartable$ustart[lambda_idx[1]] < 0) {
            tmp <- -1 * tmp
          }

          start[lambda_idx] <- tmp

          # factor variance
          # if(!std.lv) {
          #    start[var.f.idx] <- fabin$psi
          #    # if residual var, make smaller
          #    y.idx <- which(lavpartable$lhs == f &
          #                   lavpartable$group == group.values[g] &
          #                   lavpartable$op == "~")
          #    if(length(y.idx) > 0L) {
          #        # how much explained variance do we expect?
          #        # we take 0.50
          #        start[var.f.idx] <- 0.5 * start[var.f.idx]
          #    }
          #    # no negative variances (we get these if we have an
          #    # inconsistent triad (eg, covariance signs are +,+,-)
          #    if(start[var.f.idx] < 0) {
          #        start[var.f.idx] <- 0.05
          #    }
          # }

          # NOTE: fabin (sometimes) gives residual variances
          # that are larger than the original variances...

          # residual variances -- order?
          # res.idx <- which(lavpartable$lhs %in% ov.names[ov.idx] &
          #                 lavpartable$op == "~~" &
          #                 lavpartable$group == group.values[g] &
          #                 lavpartable$rhs == lavpartable$lhs)
          # start[res.idx] <- fabin$theta

          # negative variances?
          # neg.idx <- which(start[res.idx] < 0)
          # if(length(neg.idx) > 0L) {
          #    start[res.idx][neg.idx] <- 0.05
          # }
        }
      } # fabin3

      # efa?
      nefa <- lav_pt_nefa(lavpartable)
      if (nefa > 0L) {
        efa_values <- lav_pt_efa_values(lavpartable)
        # blocks belonging to this group (group-major, level-minor ordering);
        # the factor set is determined per block, as different blocks (eg the
        # within and between level, or different groups) may have a different
        # number of efa factors
        blocks_g <- (g - 1L) * nlevels + seq_len(nlevels)

        for (b in blocks_g) {
          for (set in seq_len(nefa)) {
            # determine ov idx for this set, within this block
            ov_efa <-
              unique(lavpartable$rhs[lavpartable$op == "=~" &
                lavpartable$block == b &
                lavpartable$efa == efa_values[set]])
            lv_efa <-
              unique(lavpartable$lhs[lavpartable$op == "=~" &
                lavpartable$block == b &
                lavpartable$efa == efa_values[set]])
            if (length(lv_efa) == 0L) {
              next
            }
            lambda_idx <- which(lavpartable$lhs %in% lv_efa &
              lavpartable$op == "=~" &
              lavpartable$block == b)

            theta_idx <- which(lavpartable$lhs %in% ov_efa &
              lavpartable$op == "~~" &
              lavpartable$lhs == lavpartable$rhs &
              lavpartable$block == b)

            # get observed indicators for these EFA lv variables
            ov_idx <- match(
              unique(lavpartable$rhs[lambda_idx]),
              ov_names
            )

            if (length(ov_idx) > 0L && !any(is.na(ov_idx))) {
              if (lavsamplestats@missing.flag && nlevels == 1L) {
                h1_cov <- stats_h1_cov(g)
                cov_1 <- h1_cov[ov_idx, ov_idx, drop = FALSE]
              } else {
                if (conditional_x) {
                  cov_1 <- lavsamplestats@res.cov[[g]][ov_idx,
                    ov_idx,
                    drop = FALSE
                  ]
                } else {
                  cov_1 <- lavsamplestats@cov[[g]][ov_idx,
                    ov_idx,
                    drop = FALSE
                  ]
                }
              }

              # EFA solution with zero upper-right corner
              efa_1 <- lav_efa_extraction(
                s = cov_1,
                nfactors = length(lv_efa),
                method = "ML",
                order_lv_by = order_lv_by,
                # order.lv.by = "none",
                # reflect = reflect,
                reflect = FALSE,
                corner = TRUE
              )

              # factor loadings
              tmp <- as.numeric(efa_1$LAMBDA)
              tmp[!is.finite(tmp)] <- 1.0 # just in case (eg 0/0)
              start[lambda_idx] <- tmp

              # residual variances
              tmp <- diag(efa_1$THETA)
              tmp[!is.finite(tmp)] <- 1.0 # just in case
              start[theta_idx] <- tmp
            }
          } # set
        } # block
      } # efa
    } # factor loadings

    if (model_type == "unrestricted") {
      # fill in 'covariances' from lavsamplestats
      cov_idx <- which(lavpartable$group == group_values[g] &
        lavpartable$op == "~~" &
        lavpartable$lhs != lavpartable$rhs)
      lhs_idx <- match(lavpartable$lhs[cov_idx], ov_names)
      rhs_idx <- match(lavpartable$rhs[cov_idx], ov_names)
      h1_cov <- stats_h1_cov(g)
      start[cov_idx] <- h1_cov[cbind(lhs_idx, rhs_idx)]
    }

    # composites
    if (composites) {
      std_lv <- FALSE
      var_f_idx <- which(lavpartable$lhs %in% lv_names_c &
        lavpartable$op == "~~" &
        lavpartable$group == group_values[g] &
        lavpartable$rhs %in% lv_names_c)
      if (length(var_f_idx) > 0L &&
        all(lavpartable$free[var_f_idx] == 0) &&
        !all(is.na(lavpartable$ustart[var_f_idx])) &&
        all(lavpartable$ustart[var_f_idx] == 1)) {
        std_lv <- TRUE
      }

      # weights
      cidx <- which(lavpartable$group == group_values[g] &
        lavpartable$op == "<~")
      if (std_lv) {
        start[cidx] <- 0.10
      } else {
        start[cidx] <- 1
      }

      # fill in 'covariances' from lavsamplestats
      cov_idx <- which(lavpartable$group == group_values[g] &
        lavpartable$op == "~~" &
        lavpartable$rhs %in% ov_ind_c &
        lavpartable$lhs != lavpartable$rhs)
      lhs_idx <- match(lavpartable$lhs[cov_idx], ov_names)
      rhs_idx <- match(lavpartable$rhs[cov_idx], ov_names)
      h1_cov <- stats_h1_cov(g)
      start[cov_idx] <- h1_cov[cbind(lhs_idx, rhs_idx)]
    }

    # variances of ordinal variables - set to 1.0
    if (categorical) {
      ov_var_ord_idx <- which(lavpartable$group == group_values[g] &
        lavpartable$op == "~~" &
        lavpartable$lhs %in% ov_names_ord &
        lavpartable$lhs == lavpartable$rhs)
      start[ov_var_ord_idx] <- 1.0
    }

    # 3g) intercepts/means
    ov_int_idx <- which(lavpartable$group == group_values[g] &
      lavpartable$op == "~1" &
      lavpartable$lhs %in% ov_names)
    sample_int_idx <- match(lavpartable$lhs[ov_int_idx], ov_names)
    if (lavsamplestats@missing.flag && nlevels == 1L) {
      # note: use exact list indexing below ([["mean"]], not $mean), as
      # the conditional.x implied list has mean.x/cov.x elements that $
      # would partially match
      if (conditional_x) {
        if (length(lavh1$implied$res.int) >= g &&
            !is.null(lavh1$implied$res.int[[g]])) {
          h1_mean <- lavh1$implied$res.int[[g]]
        } else if (length(lavsamplestats@missing.h1) >= g &&
                   !is.null(lavsamplestats@missing.h1[[g]]$res.int)) {
          h1_mean <- lavsamplestats@missing.h1[[g]]$res.int
        } else {
          h1_mean <- lavsamplestats@res.int[[g]]
        }
      } else if (!is.null(lavh1$implied[["mean"]][[g]])) {
        h1_mean <- lavh1$implied[["mean"]][[g]]
      } else if (!is.null(lavsamplestats@missing.h1[[g]])) {
        h1_mean <- lavsamplestats@missing.h1[[g]]$mu
      } else {
        h1_mean <- lavsamplestats@mean[[g]]
      }
      start[ov_int_idx] <- h1_mean[sample_int_idx]
    } else {
      if (conditional_x && nlevels == 1L) {
        start[ov_int_idx] <- lavsamplestats@res.int[[g]][sample_int_idx]
      } else {
        start[ov_int_idx] <- lavsamplestats@mean[[g]][sample_int_idx]
      }
    }

    # TODo: if marker.int.zero = TRUE, set lv means to marker means,
    #       and the non-marker means to
    #       lavsamplestats@mean[[g]] - LAMBDA %*% ALPHA
    #       where ALPHA = means of the markers

    # 4g) thresholds
    th_idx <- which(lavpartable$group == group_values[g] &
                      lavpartable$op == "|")
    if (length(th_idx) > 0L) {
      th_names_lavpartable <- paste(lavpartable$lhs[th_idx], "|",
        lavpartable$rhs[th_idx],
        sep = ""
      )
      th_names_sample <-
        lavsamplestats@th.names[[g]][lavsamplestats@th.idx[[g]] > 0L]
      # th.names.sample should identical to
      # lav_pt_vnames(lavpartable, "th", group = group.values[g])
      if (conditional_x && nlevels == 1L) {
        th_values <-
          lavsamplestats@res.th[[g]][lavsamplestats@th.idx[[g]] > 0L]
      } else {
        th_values <-
          lavsamplestats@th[[g]][lavsamplestats@th.idx[[g]] > 0L]
      }
      start[th_idx] <- th_values[match(
        th_names_lavpartable,
        th_names_sample
      )]
    }

    # 5g) exogenous `fixed.x' covariates
    if (length(ov_names_x) > 0) {
      exo_idx <- which(lavpartable$group == group_values[g] &
        lavpartable$op == "~~" &
        lavpartable$lhs %in% ov_names_x &
        lavpartable$rhs %in% ov_names_x)
      if (!conditional_x) {
        row_idx <- match(lavpartable$lhs[exo_idx], ov_names)
        col_idx <- match(lavpartable$rhs[exo_idx], ov_names)
        if (lavsamplestats@missing.flag && nlevels == 1L) {
          if (!is.null(lavh1$implied$cov[[g]])) {
            h1_cov <- lavh1$implied$cov[[g]]
          } else if (!is.null(lavsamplestats@missing.h1[[g]])) {
            h1_cov <- lavsamplestats@missing.h1[[g]]$sigma
          } else {
            h1_cov <- lavsamplestats@cov[[g]]
          }
          start[exo_idx] <- h1_cov[cbind(row_idx, col_idx)]
          # using slightly smaller starting values for free
          # variance/covariances (fixed.x = FALSE);
          # this somehow avoids false convergence in saturated models
          nobs <- lavsamplestats@nobs[[g]]
          this_idx <- which(seq_along(lavpartable$free) %in% exo_idx &
                              lavpartable$free > 0L)
          start[this_idx] <- start[this_idx] * (nobs - 1) / nobs
        } else {
          start[exo_idx] <- lavsamplestats@cov[[g]][cbind(row_idx, col_idx)]
        }
      } else {
        # cov.x
        row_idx <- match(lavpartable$lhs[exo_idx], ov_names_x)
        col_idx <- match(lavpartable$rhs[exo_idx], ov_names_x)
        start[exo_idx] <- lavsamplestats@cov.x[[g]][cbind(
          row_idx,
          col_idx
        )]
        # mean.x
        exo_int_idx <- which(lavpartable$group == group_values[g] &
          lavpartable$op == "~1" &
          lavpartable$lhs %in% ov_names_x)
        int_idx <- match(lavpartable$lhs[exo_int_idx], ov_names_x)
        start[exo_int_idx] <- lavsamplestats@mean.x[[g]][int_idx]
      }
    }

    # 6b. exogenous lv variances if single indicator -- new in 0.5-21
    lv_x <- lav_pt_vnames(lavpartable, "lv.x", group = group_values[g])
    # FIXME: also for multilevel?
    lv_x <- unique(unlist(lv_x))
    if (length(lv_x) > 0L) {
      for (ll in lv_x) {
        ind_idx <- which(lavpartable$op == "=~" &
          lavpartable$lhs == ll &
          lavpartable$group == group_values[g])
        if (length(ind_idx) == 1L) {
          single_ind <- lavpartable$rhs[ind_idx]
          single_fvar_idx <- which(lavpartable$op == "~~" &
            lavpartable$lhs == ll &
            lavpartable$rhs == ll &
            lavpartable$group == group_values[g])
          single_var_idx <- which(lavpartable$op == "~~" &
            lavpartable$lhs == single_ind &
            lavpartable$rhs == single_ind &
            lavpartable$group == group_values[g])
          # user-defined residual variance
          # fixme: we take the first, in case we have multiple matches
          # (eg nlevels)
          single_var <- lavpartable$ustart[single_var_idx[1]]
          if (is.na(single_var)) {
            single_var <- 1
          }
          ov_idx <- match(single_ind, ov_names)
          if (conditional_x && nlevels == 1L) {
            ov_var <- diag(lavsamplestats@res.cov[[g]])[ov_idx]
          } else {
            ov_var <- diag(lavsamplestats@cov[[g]])[ov_idx]
          }
          # take (1 - (rvar/ov.var) * ov.var
          tmp <- (1 - (single_var / ov_var)) * ov_var
          # just in case
          if (is.na(tmp) || tmp < 0.05) {
            tmp <- 0.05
          }
          start[single_fvar_idx] <- tmp
        }
      }
    }

    # 7g) regressions "~" # new in 0.6-10
    if (length(lv_names) == 0L && nlevels == 1L && !conditional_x) {
      # observed only
      reg_idx <- which(lavpartable$group == group_values[g] &
        lavpartable$op == "~")
      if (length(reg_idx) > 0L) {
        eqs_y <- unique(lavpartable$lhs[reg_idx])
        ny <- length(eqs_y)
        for (i in seq_len(ny)) {
          y_name <- eqs_y[i]
          start_idx <- which(lavpartable$group == group_values[g] &
            lavpartable$op == "~" &
            lavpartable$lhs == y_name)
          x_names <- lavpartable$rhs[start_idx]
          cov_1 <- lavsamplestats@cov[[g]]
          y_idx <- match(y_name, ov_names)
          x_idx <- match(x_names, ov_names)
          s_xx <- cov_1[x_idx, x_idx, drop = FALSE]
          s_xy <- cov_1[x_idx, y_idx, drop = FALSE]
          # regression coefficient(s)
          beta_i <- try(solve(s_xx, s_xy), silent = TRUE)
          if (inherits(beta_i, "try-error")) {
            start[start_idx] <- beta_i <- rep(0, length(start_idx))
          } else {
            start[start_idx] <- drop(beta_i)
          }
          # residual variance
          res_idx <- which(lavpartable$group == group_values[g] &
            lavpartable$op == "~~" &
            lavpartable$lhs == y_name &
            lavpartable$rhs == y_name)
          res_val <- cov_1[y_idx, y_idx] - drop(crossprod(beta_i, s_xy))
          if (res_val > 0.001 * cov_1[y_idx, y_idx] &&
            res_val < 0.999 * cov_1[y_idx, y_idx]) {
            start[res_idx] <- res_val
          } else {
            # do nothing (keep what we have)
          }
          # intercept
          int_idx <- which(lavpartable$group == group_values[g] &
            lavpartable$op == "~1" &
            lavpartable$lhs == y_name)
          if (length(int_idx) > 0L) {
            mean_1 <- lavsamplestats@mean[[g]]
            ybar <- mean_1[y_idx]
            xbar <- mean_1[x_idx]
            int_val <- ybar - drop(crossprod(beta_i, xbar))
            if (is.finite(int_val)) {
              start[int_idx] <- int_val
            }
          }
        }
      }
    }

    #  # 8 latent variances (new in 0.6-2)
    #  lv.names.y <- lav_pt_vnames(lavpartable,
    #                               "lv.y", group = group.values[g])
    #  lv.names.x <- lav_pt_vnames(lavpartable,
    #                               "lv.x", group = group.values[g])
    #  # multilevel? take first level only
    #  if(is.list(lv.names.y)) {
    #      lv.names.y <- unlist(lv.names.y) # for now
    #  }
    #  if(is.list(lv.names.x)) {
    #      lv.names.x <- unlist(lv.names.x) # for now
    #  }
    #  lv.names.xy <- unique(c(lv.names.x, lv.names.y))


    #  if(length(lv.names.xy) > 0L) {
    #      free.var.idx <- which(lavpartable$op == "~~" &
    #                            lavpartable$lhs %in% lv.names.xy &
    #                            lavpartable$rhs == lavpartable$lhs &
    #                            lavpartable$group == group.values[g])
    #      if(length(free.var.idx) > 0L) {
    #          this.lv.names <- lavpartable$lhs[free.var.idx]
    #          for(v in seq_len(length(free.var.idx))) {
    #              # single marker item?
    #              ind.idx <- which(lavpartable$op == "=~" &
    #                               lavpartable$lhs %in% this.lv.names[v] &
    #                               #lavpartable$rhs %in% ov.names.num &
    #                               lavpartable$free == 0L &
    #                               lavpartable$group == group.values[g])
    #              if(length(ind.idx) == 0) {
    #                  next
    #              } else if(length(ind.idx) > 1L) {
    #                  # FIXME! perhaps a random effect? do something clever
    #                  next
    #              } else if(length(ind.idx) == 1L) {
    #                  marker.ind <- lavpartable$rhs[ind.idx]
    #                  ov.idx <- match(marker.ind, ov.names)
    #                  if(conditional.x) {
    #                      ov.var <- diag(lavsamplestats@res.cov[[g]])[ov.idx]
    #                  } else {
    #                      ov.var <- diag(lavsamplestats@cov[[g]])[ov.idx]
    #                  }
    #
    #                  # exogenous? assume rel = 0.50
    #                  lambda <- lavpartable$ustart[ind.idx]
    #                  tmp <- (0.50 * ov.var)/lambda^2
    #                  if(this.lv.names[v] %in% lv.names.y) {
    #                      # endogenous, assume R2 = 0.2
    #                      tmp <- 0.8 * tmp
    #                  }
    #                  # within variance?
    #                  if(nlevels > 1L &&
    #                     lavpartable$level[ free.var.idx[v] ] == 1L) {
    #                      tmp <- tmp * 0.75
    #                  }
    #                  # between variance?
    #                  if(nlevels > 1L &&
    #                     lavpartable$level[ free.var.idx[v] ] > 1L) {
    #                      tmp <- tmp * 0.25
    #                  }
    #                  # just in case
    #                  if(is.na(tmp) || tmp < 0.05) {
    #                      tmp <- 0.05
    #                  }
    #                  start[ free.var.idx[v] ] <- tmp
    #              }
    #          } # v
    #      } # free.var.idx
    #  } # lv var

    # nlevels > 1L
    if (nlevels > 1L) {
      level_values <- lav_pt_level_values(lavpartable)
      # Note: ov.names.x contains all levels within a group!
      if (length(ov_names_x) > 0) {
        for (l in 1:nlevels) {
          # block number
          block <- (g - 1L) * nlevels + l

          this_block_x <- lav_pt_vnames(lavpartable, "ov.x",
            block = block
          )
          this_block_ov <- lav_pt_vnames(lavpartable, "ov",
            block = block
          )
          if (length(this_block_x) == 0L) {
            next
          }

          # var/cov
          exo_idx <- which(lavpartable$group == group_values[g] &
            lavpartable$level == level_values[l] &
            lavpartable$op == "~~" &
            lavpartable$lhs %in% this_block_x &
            lavpartable$rhs %in% this_block_x)

          if (!is.null(lavh1$implied$cov.x[[block]]) &&
              NROW(lavh1$implied$cov.x[[block]]) > 0L) {
            # the h1 cov blocks span the y variables only, and the x
            # moments are stored separately (two-level (D)WLS with
            # conditional.x)
            cov_1 <- lavh1$implied$cov.x[[block]]
            row_idx <- match(lavpartable$lhs[exo_idx], colnames(cov_1))
            col_idx <- match(lavpartable$rhs[exo_idx], colnames(cov_1))
          } else if (is.null(lavh1$implied$cov[[1]])) {
            row_idx <- match(lavpartable$lhs[exo_idx], ov_names)
            col_idx <- match(lavpartable$rhs[exo_idx], ov_names)
            if (l == 1L) {
              cov_1 <- lavsamplestats@YLp[[g]][[2]]$S.PW.start
            } else {
              cov_1 <- lavsamplestats@YLp[[g]][[l]]$Sigma.B
            }
          } else {
            row_idx <- match(lavpartable$lhs[exo_idx], this_block_ov)
            col_idx <- match(lavpartable$rhs[exo_idx], this_block_ov)
            cov_1 <- lavh1$implied$cov[[block]]
          }
          # make sure starting values for variances are positive
          neg_idx <- which(diag(cov_1) < 0.001)
          if (length(neg_idx) > 0L) {
            diag(cov_1)[neg_idx] <- 0.001
          }
          start[exo_idx] <- cov_1[cbind(row_idx, col_idx)]

          # intercepts
          ov_int_idx <- which(lavpartable$group == group_values[g] &
            lavpartable$level == level_values[l] &
            lavpartable$op == "~1" &
            lavpartable$lhs %in% this_block_x)

          if (!is.null(lavh1$implied$mean.x[[block]]) &&
              length(lavh1$implied$mean.x[[block]]) > 0L) {
            int <- lavh1$implied$mean.x[[block]]
            idx <- match(lavpartable$lhs[ov_int_idx], names(int))
          } else if (is.null(lavh1$implied$mean[[1]])) {
            idx <- match(lavpartable$lhs[ov_int_idx], ov_names)
            if (l == 1L) {
              int <- lavsamplestats@YLp[[g]][[2]]$Mu.W
            } else {
              int <- lavsamplestats@YLp[[g]][[l]]$Mu.B.start
            }
          } else {
            idx <- match(lavpartable$lhs[ov_int_idx], this_block_ov)
            int <- lavh1$implied$mean[[block]]
          }
          start[ov_int_idx] <- int[idx]

          # new in 0.6-12
          # very special case: conditional.x with a combination of
          # split-x and regular-x
          # here, we must:
          # 1) replace var/cov of split-x by *residual* varcov
          #    after regressing out regular-x
          # 2) replace means of split-x by intercepts
          # 3) fill split-x ~ regular-x regression coefficients
          if (conditional_x) {
            if (is.null(lavh1$implied$cov[[l]])) {
             lav_msg_stop(gettext(
               "lavh1 information is needed; please rerun with h1 = TRUE"))
            }
            blocks_within_group <- (g - 1L) * nlevels + seq_len(nlevels)
            other_block_names <- lav_pt_vnames(lavpartable, "ov",
                                    block = blocks_within_group[-block])
            ov_names_x_block <- this_block_x
            idx <- which(ov_names_x_block %in% other_block_names)
            if (length(idx) > 0L) {
              ov_names_x_block <- ov_names_x_block[-idx]
            }
            ov_names_x1 <- this_block_x[!this_block_x %in% ov_names_x_block]
            ov_names_x2 <- ov_names_x_block
            nx1 <- length(ov_names_x1) # split x
            nx2 <- length(ov_names_x2) # regular  x
            if (nx1 > 0L && nx2 > 0L) {
              # COV
              c1_idx <- match(ov_names_x1, this_block_ov)
              c2_idx <- match(ov_names_x2, this_block_ov)

              cov_y <- cov_1[c1_idx, c1_idx, drop = FALSE]
              cov_x <- cov_1[c2_idx, c2_idx, drop = FALSE]
              cov_yx <- cov_1[c1_idx, c2_idx, drop = FALSE]
              cov_xy <- cov_1[c2_idx, c1_idx, drop = FALSE]
              cov_xinv_yx <- solve(cov_x, cov_xy)
              res_cov <- cov_y - cov_yx %*% cov_xinv_yx

              res_cov_idx <- which(lavpartable$group == group_values[g] &
                lavpartable$level == level_values[l] &
                lavpartable$op == "~~" &
                lavpartable$lhs %in% ov_names_x1 &
                lavpartable$rhs %in% ov_names_x1)

              row_idx <- match(lavpartable$lhs[res_cov_idx], ov_names_x1)
              col_idx <- match(lavpartable$rhs[res_cov_idx], ov_names_x1)
              start[res_cov_idx] <- res_cov[cbind(row_idx, col_idx)]

              # INT
              int_y <- int[c1_idx]
              int_x <- int[c2_idx]
              res_int <- int_y - t(cov_xinv_yx) %*% int_x

              res_int_idx <- which(lavpartable$group == group_values[g] &
                lavpartable$level == level_values[l] &
                lavpartable$op == "~1" &
                lavpartable$lhs %in% ov_names_x1)
              idx <- match(lavpartable$lhs[res_int_idx], ov_names_x1)
              start[res_int_idx] <- res_int[idx]

              # REG
              reg_idx <- which(lavpartable$group == group_values[g] &
                lavpartable$level == level_values[l] &
                lavpartable$op == "~" &
                lavpartable$lhs %in% ov_names_x1 &
                lavpartable$rhs %in% ov_names_x2)
              row_idx <- match(lavpartable$lhs[reg_idx], ov_names_x1)
              col_idx <- match(lavpartable$rhs[reg_idx], ov_names_x2)
              start[reg_idx] <- t(cov_xinv_yx)[cbind(row_idx, col_idx)]
            } # special case
          } # conditional.x
        } # levels
      } # fixed.x
    } # nlevels > 1L
  } # groups




  # group weights
  group_idx <- which(lavpartable$lhs == "group" &
    lavpartable$op == "%")
  if (length(group_idx) > 0L) {
    ngroups <- length(group_idx)
    # prop <- rep(1/ngroups, ngroups)
    # use last group as reference
    # start[group.idx] <- log(prop/prop[ngroups])

    # poisson version
    start[group_idx] <- log(rep(lavsamplestats@ntotal / ngroups, ngroups))
  }

  # growth models:
  # - compute starting values for mean latent variables
  # - compute starting values for variance latent variables
  if (start_initial %in% c("lavaan") &&
    model_type == "growth") {
    ### DEBUG ONLY
    # lv.var.idx <- which(lavpartable$op == "~~"                &
    #                lavpartable$lhs %in% lv.names &
    #                lavpartable$lhs == lavpartable$rhs)

    ### DEBUG ONLY
    # lv.int.idx <- which(lavpartable$op == "~1"         &
    #                    lavpartable$lhs %in% lv.names)
  }

  # adjust if outside bounds -- new in 0.6-6
  if (!is.null(lavpartable$lower)) {
    bad_idx <- which(start < lavpartable$lower)
    if (length(bad_idx)) {
      start[bad_idx] <- lavpartable$lower[bad_idx]
    }
  }
  if (!is.null(lavpartable$upper)) {
    bad_idx <- which(start > lavpartable$upper)
    if (length(bad_idx)) {
      start[bad_idx] <- lavpartable$upper[bad_idx]
    }
  }


  # override if the model syntax contains explicit starting values (free only)
  # user.idx <- which(!is.na(lavpartable$ustart) &
  #                  lavpartable$user != 7L) # new in 0.6-7, if rotation and
  #                                          # and we change the order of lv's
  user_idx <- which(!is.na(lavpartable$ustart) & lavpartable$free > 0L)
  start[user_idx] <- lavpartable$ustart[user_idx]

  # override if a user list with starting values is provided
  # we only look at the 'est' column for now
  if (!is.null(start_user)) {
    if (is.null(lavpartable$group)) {
      lavpartable$group <- rep(1L, length(lavpartable$lhs))
    }
    if (is.null(start_user$group)) {
      start_user$group <- rep(1L, length(start_user$lhs))
    }

    # FIXME: avoid for loop!!!
    for (i in seq_along(lavpartable$lhs)) {
      # find corresponding parameters
      lhs <- lavpartable$lhs[i]
      op <- lavpartable$op[i]
      rhs <- lavpartable$rhs[i]
      grp <- lavpartable$group[i]

      start_user_idx <- which(start_user$lhs == lhs &
        start_user$op == op &
        start_user$rhs == rhs &
        start_user$group == grp)
      if (length(start_user_idx) == 1L &&
        is.finite(start_user$est[start_user_idx])) {
        start[i] <- start_user$est[start_user_idx]
      }
    }
  }

  # override fixed values with ustart values
  user_idx <- which(!is.na(lavpartable$ustart) & lavpartable$free == 0L)
  start[user_idx] <- lavpartable$ustart[user_idx]


  # new in 0.6-21
  # if any thresholds, make sure they are in increasing order
  # (could be an issue if the first and second were fixed to 0 and 1)
  # note: per block! (for two-level models, the level-1 thresholds are
  # all fixed to zero and must not trigger the repair for the (free)
  # level-2 thresholds of the same item)
  nblocks <- lav_pt_nblocks(lavpartable)
  block_values <- if (is.null(lavpartable$block)) {
    rep(1L, length(lavpartable$lhs))
  } else {
    lavpartable$block
  }
  for (b in seq_len(nblocks)) {
    th_idx <- which(block_values == b & lavpartable$op == "|")
    if (length(th_idx) > 0L) {
      # for every ov.ord, check if t1 < t2 < t3 < ...
      ov_ord <- unique(lavpartable$lhs[th_idx])
      for (oo in ov_ord) {
        this_idx <- which(block_values == b &
                          lavpartable$op == "|" & lavpartable$lhs == oo)
        this_free_idx <- which(block_values == b &
                               lavpartable$op == "|" & lavpartable$lhs == oo &
                               lavpartable$free > 0)
        item_th <- start[this_idx]
        if (length(item_th) > 1L && length(this_free_idx) > 0 &&
                                          !all(diff(item_th) > 0)) {
          start[this_free_idx] <-
            cumsum(abs(item_th))[match(this_free_idx, this_idx)]
        }
      }
    }
  }


  # two-level categorical: give the (free) between-level residual
  # variances of the ordinal variables a strictly positive starting value
  if (!is.null(lavpartable$level)) {
    level_values <- lav_pt_level_values(lavpartable)
    if (length(level_values) > 1L && any(lavpartable$op == "|")) {
      ord_names <- unique(lavpartable$lhs[lavpartable$op == "|"])
      ord_bvar_idx <- which(lavpartable$op == "~~" &
        lavpartable$lhs == lavpartable$rhs &
        lavpartable$lhs %in% ord_names &
        lavpartable$level %in% level_values[-1L] &
        lavpartable$free > 0L)
      if (length(ord_bvar_idx) > 0L) {
        small_idx <- ord_bvar_idx[start[ord_bvar_idx] < 0.01]
        start[small_idx] <- 0.05
      }
    }
  }

  # final check: no NaN or other non-finite values
  bad_idx <- which(!is.finite(start))
  if (length(bad_idx) > 0L) {
    cat("starting values:\n")
    print(start)
     lav_msg_warn(gettext(
       "some starting values are non-finite; replacing them with 0.5;
       please provide better starting values."))
    start[bad_idx] <- 0.5
  }

  if (lav_debug()) {
    cat("lavaan DEBUG: lavaanStart\n")
    print(start)
  }

  start
}

# backwards compatibility
# StartingValues <- lav_start

# sanity check: (user-specified) variances smaller than covariances
# but not for composites, as we have not 'set' their variances yet
lav_start_check_cov <- function(lavpartable = NULL, start = lavpartable$start) {
  nblocks <- lav_pt_nblocks(lavpartable)
  block_values <- lav_pt_block_values(lavpartable)

  for (g in 1:nblocks) {

    lv_names_c <- lav_pt_vnames(lavpartable, "lv.composite", block = g)

    # collect all non-zero covariances
    cov_idx <- which(lavpartable$op == "~~" &
      lavpartable$block == block_values[g] &
      !lavpartable$lhs %in% lv_names_c &
      lavpartable$lhs != lavpartable$rhs &
      !lavpartable$exo &
      start != 0)

    # for each covariance, use corresponding variances to standardize;
    # the end result should not exceed abs(1)
    for (cc in seq_along(cov_idx)) {
      this_cov_idx <- cov_idx[cc]

      # find corresponding variances
      var_lhs <- lavpartable$lhs[this_cov_idx]
      var_rhs <- lavpartable$rhs[this_cov_idx]

      var_lhs_idx <- which(lavpartable$op == "~~" &
        lavpartable$block == block_values[g] &
        lavpartable$lhs == var_lhs &
        lavpartable$lhs == lavpartable$rhs)

      var_rhs_idx <- which(lavpartable$op == "~~" &
        lavpartable$block == block_values[g] &
        lavpartable$lhs == var_rhs &
        lavpartable$lhs == lavpartable$rhs)

      var_lhs_value <- start[var_lhs_idx]
      var_rhs_value <- start[var_rhs_idx]

      block_txt <- ""
      if (nblocks > 1L) {
        block_txt <- paste(" [in block ", g, "]", sep = "")
      }

      # check for zero variances
      if (var_lhs_value == 0 || var_rhs_value == 0) {
        # this can only happen if it is user-specified
        # cov.idx free? set it to zero
        if (start[this_cov_idx] == 0) {
          # nothing to do
        } else if (lavpartable$free[this_cov_idx] > 0L) {
            lav_msg_warn(gettextf(
             "non-zero covariance element set to zero, due to fixed-to-zero
             variances variables involved are: %s", var_lhs), var_rhs,
             block_txt
            )
          start[this_cov_idx] <- 0
        } else {
          lav_msg_stop(gettextf(
            "please provide better fixed values for (co)variances;
            variables involved are: %s ", var_lhs), var_rhs, block_txt
          )
        }
        next
      }

      # which one is the smallest? abs() in case of negative variances
      if (abs(var_lhs_value) < abs(var_rhs_value)) {
        var_min_idx <- var_lhs_idx
        var_max_idx <- var_rhs_idx
      } else {
        var_min_idx <- var_rhs_idx
        var_max_idx <- var_lhs_idx
      }

      # check
      cor_1 <- abs(start[this_cov_idx] / sqrt(var_lhs_value * var_rhs_value))

      # NOTE: we treat this as an unconditional COR!

      if (!is.finite(cor_1)) {
        # force simple values
           lav_msg_warn(gettextf(
            "starting values imply NaN for a correlation value; variables
            involved are: %s", var_lhs), var_rhs, block_txt
          )
        start[var_lhs_idx] <- 1
        start[var_rhs_idx] <- 1
        start[this_cov_idx] <- 0
      } else if (cor_1 > 1) {
        txt <- gettextf(
          "starting values imply a correlation larger than 1; variables
          involved are: %1$s %2$s %3$s", var_lhs, var_rhs, block_txt)

        # three ways to fix it: rescale cov12, var1 or var2

        # we prefer a free parameter, and not user-specified
        if (lavpartable$free[this_cov_idx] > 0L &&
          is.na(lavpartable$ustart[this_cov_idx])) {
             lav_msg_warn(gettext(txt))
          start[this_cov_idx] <- start[this_cov_idx] / (cor_1 * 1.1)
        } else if (lavpartable$free[var_min_idx] > 0L &&
          is.na(lavpartable$ustart[var_min_idx])) {
             lav_msg_warn(gettext(txt))
          start[var_min_idx] <- start[var_min_idx] * (cor_1 * 1.1)^2
        } else if (lavpartable$free[var_max_idx] > 0L &&
          is.na(lavpartable$ustart[var_max_idx])) {
             lav_msg_warn(gettext(txt))
          start[var_max_idx] <- start[var_max_idx] * (cor_1 * 1.1)^2

          # not found? try just a free parameter
        } else if (lavpartable$free[this_cov_idx] > 0L) {
             lav_msg_warn(gettext(txt))
          start[this_cov_idx] <- start[this_cov_idx] / (cor_1 * 1.1)
        } else if (lavpartable$free[var_min_idx] > 0L) {
             lav_msg_warn(gettext(txt))
          start[var_min_idx] <- start[var_min_idx] * (cor_1 * 1.1)^2
        } else if (lavpartable$free[var_max_idx] > 0L) {
             lav_msg_warn(gettext(txt))
          start[var_max_idx] <- start[var_max_idx] * (cor_1 * 1.1)^2

          # nothing? abort or warn (and fail later...): warn
        } else {
             lav_msg_warn(gettext(txt))
          # lav_msg_stop(gettextf(
          # "please provide better fixed values for (co)variances;
          #  variables involved are: %s ", var.lhs), var.rhs, block.txt)
        }
      } # COR > 1
    } # cov.idx
  }
  start
}

# STEP 1 of the Muthen (1984) three-stage estimator: univariate models
#
# for each (endogenous) variable: means/thresholds/intercepts, slopes
# (if conditional.x), variances (numeric variables only), and the
# corresponding casewise scores
#
# YR/LDW 2026: refactored; empty-category compensation moved to
#              lav_samp_step1_empty_cells(); unobserved *middle* categories
#              now give a clear error message (they were never supported:
#              the old code stopped with an obscure R error)

lav_samp_step1 <- function(y,
                                  wt = NULL, # new in 0.6-6
                                  ov_names = NULL,
                                  ov_types = NULL,
                                  ov_levels = NULL,
                                  ov_names_x = character(0L),
                                  exo = NULL,
                                  scores_flag = TRUE, # scores?
                                  allow_empty_cell = TRUE,
                                           # allow empty categories?
                                  group = 1L) { # for error message


  # just in case Y is a vector
  y <- as.matrix(y)

  nvar <- NCOL(y)
  n <- NROW(y)
  n_th <- ov_levels - 1L
  n_th[n_th == -1L] <- 1L # numeric variables (nlev = 0): one 'mean' slot
  nth <- sum(n_th)
  th_end_idx <- cumsum(n_th)
  th_start_idx <- th_end_idx - (n_th - 1L)

  # variable types; default = numeric
  nexo <- length(ov_names_x)
  if (nexo > 0L) stopifnot(NCOL(exo) == nexo)

  # means/thresholds/intercepts, slopes, variances
  th <- vector("list", length = nvar)
  th_nox <- vector("list", length = nvar)
  th_names <- vector("list", length = nvar)
  th_idx_1 <- vector("list", length = nvar)
  # which declared threshold slots come from the fit (FALSE = pseudo value
  # for an empty category; see lav_samp_step1_empty_cells)
  th_keep <- vector("list", length = nvar)
  slopes <- matrix(as.numeric(NA), nrow = nvar, ncol = nexo) # if conditional.x
  var_1 <- numeric(length = nvar) # continuous variables only

  # SCORES
  sc_var <- matrix(0, n, nvar)
  sc_sl <- matrix(0, n, nvar * nexo)
  sc_th <- matrix(0, n, nth)

  # fitted objects
  fit_1 <- vector("list", length = nvar)

  # stage one - TH/SLOPES/VAR only
  for (i in 1:nvar) {
    th_idx <- th_start_idx[i]:th_end_idx[i]
    sl_idx <- seq(i, by = nvar, length.out = nexo)
    if (ov_types[i] == "numeric") {
      fit <- lav_uvreg_fit(y = y[, i], x = exo, wt = wt)
      if (any(is.na(fit$theta))) {
        lav_msg_stop(gettextf(
          "linear regression failed for %1$s;
          X may not be of full rank in group %2$s", ov_names[i], group))
      }
      fit_1[[i]] <- fit
      # compute mean and variance
      th[[i]] <- th_nox[[i]] <- fit$theta[1L]
      var_1[i] <- fit$theta[fit$var_idx]
      th_names[[i]] <- ov_names[i]
      th_idx_1[[i]] <- 0L
      th_keep[[i]] <- TRUE
      if (scores_flag) {
        scores <- lav_uvreg_sc(y = y[, i], x = exo, wt = wt)
        sc_th[, th_idx] <- scores[, 1L]
        sc_var[, i] <- scores[, fit$var_idx]
      }
      if (nexo > 0L) {
        slopes[i, ] <- fit$theta[-c(1L, fit$var_idx)]
        if (scores_flag) {
          sc_sl[, sl_idx] <- scores[, -c(1L, fit$var_idx), drop = FALSE]
        }
        th_nox[[i]] <- mean(y[, i], na.rm = TRUE)
      }
    } else if (ov_types[i] == "ordered") {
      # check if we have enough categories in this group
      # FIXME: should we be more tolerant here???
      y_freq <- tabulate(y[, i], nbins = ov_levels[i])
      if (any(y_freq == 0L) && !allow_empty_cell) {
        lav_msg_stop(gettextf(
          "some categories of variable `%1$s' are empty in group %2$s;
          frequencies are [%3$s]", ov_names[i], group,
          lav_msg_view(y_freq, "none")))
      }
      fit <- lav_uvord_fit(y = y[, i], x = exo, wt = wt)
      if (any(is.na(fit$theta))) {
        lav_msg_stop(gettextf(
          "probit regression failed for %1$s; X may not be of full rank
          in group %2$s", ov_names[i], group))
      }
      fit_1[[i]] <- fit
      th[[i]] <- fit$theta[fit$th_idx]
      fit_nox <- lav_uvord_th(y = y[, i], wt = wt)
      th_nox[[i]] <- fit_nox
      if (scores_flag) {
        scores <- lav_uvord_sc(y = y[, i], x = exo, wt = wt)
      }

      th_keep[[i]] <- rep(TRUE, n_th[i])
      if (allow_empty_cell) {
        if (any(y_freq == 0L)) {
          empty <- lav_samp_step1_empty_cells(
            y_freq = y_freq, nlev = ov_levels[i], fit = fit,
            fit_nox = fit_nox, ov_name = ov_names[i], group = group
          )
          th[[i]] <- empty$th
          th_nox[[i]] <- empty$th_nox
          th_keep[[i]] <- empty$keep_idx
          if (scores_flag) sc_th[, th_idx[empty$keep_idx]] <-
                                                scores[, fit$th_idx, drop = FALSE]
        } else {
          # no empty cells for this variable: fill the scores as usual
          # (up to 0.7-1, they were left at zero, giving a rank-deficient
          #  A11 and a broken WLS.W whenever allow.empty.cell = TRUE)
          if (scores_flag) sc_th[, th_idx] <- scores[, fit$th_idx,
                                                     drop = FALSE]
        }
        # note: local 'fit' only -- fit_1[[i]] (assigned above) keeps the
        # fitted (possibly shorter) th_idx, which step 2 relies on
        fit$th_idx <- 1:n_th[i]
      } else {
        if (scores_flag) sc_th[, th_idx] <- scores[, fit$th_idx, drop = FALSE]
      }
      slopes[i, ] <- fit$theta[fit$slope_idx]
      if (scores_flag) {
        sc_sl[, sl_idx] <- scores[, fit$slope_idx, drop = FALSE]
      }
      var_1[i] <- 1.0
      th_names[[i]] <- paste(ov_names[i], "|t", seq_along(th[[i]]),
        sep = ""
      )
      th_idx_1[[i]] <- rep(i, length(th[[i]]))
    } else {
      lav_msg_stop(gettext("unknown ov.types:"), ov_types[i])
    }
  }

  list(
    FIT = fit_1, VAR = var_1, SLOPES = slopes,
    TH = th, TH.NOX = th_nox, TH.IDX = th_idx_1, TH.NAMES = th_names,
    TH.KEEP = th_keep,
    SC.TH = sc_th, SC.VAR = sc_var, SC.SL = sc_sl,
    th.start.idx = th_start_idx, th.end.idx = th_end_idx
  )
}

# compensate for empty categories of an ordinal variable (allow_empty_cell)
#
# lav_uvord_fit() (implicitly) drops thresholds when *extreme* categories
# are unobserved: leading empty categories disappear in the 1..ncat
# recoding, and trailing empty categories shorten tabulate(y). The declared
# number of thresholds is nlev - 1, so we place the fitted thresholds in
# their declared slots (keep_idx) and fill the unidentified slots with
# pseudo-values: -4 (first), +4 (last), previous + .01 (in between).
#
# Unobserved *middle* categories are NOT supported: the fit then keeps a
# (degenerate) threshold for the empty category, and no consistent mapping
# to the declared slots exists.
lav_samp_step1_empty_cells <- function(y_freq = NULL, nlev = NULL,
                                       fit = NULL, fit_nox = NULL,
                                       ov_name = NULL, group = 1L) {
  keep_idx <- rep(TRUE, (nlev - 1))
  mis_idx <- !keep_idx
  zidx <- y_freq == 0L
  dz <- diff(zidx) == 1L
  if (y_freq[nlev] == 0L) {
    # trailing empty categories
    wdz <- which(dz)
    nhi <- nlev - wdz[length(wdz)]
    keep_idx[(nlev - nhi) : (nlev - 1)] <- FALSE
    mis_idx[(nlev - nhi) : (nlev - 1)] <- TRUE
  }
  if (any(dz[-length(dz)])) {
    # middle empty categories
    wdz <- which(dz[-length(dz)])
    keep_idx[wdz + 1] <- FALSE
    mis_idx[wdz + 1] <- TRUE
  }
  if (y_freq[1] == 0L) {
    # leading empty categories
    nlow <- which(diff(zidx) == -1)[1]
    keep_idx[1:nlow] <- FALSE
    mis_idx[1:nlow] <- TRUE
  }

  if (sum(keep_idx) != length(fit$th_idx)) {
    lav_msg_stop(gettextf(
      "cannot handle the empty categories of variable %1$s in group %2$s
      (frequencies are [%3$s]): only unobserved categories at the extremes
      are supported.", ov_name, group, lav_msg_view(y_freq, "none")))
  }

  th <- th_nox <- rep(0, nlev - 1)
  th[keep_idx] <- fit$theta[fit$th_idx]
  th_nox[keep_idx] <- fit_nox[keep_idx]
  for (k in which(mis_idx)) {
    if (k == 1) {
      th[k] <- -4
      th_nox[k] <- -4
    } else if (k == (nlev - 1)) {
      th[k] <- 4
      th_nox[k] <- 4
    } else {
      th[k] <- th[(k - 1)] + .01
      th_nox[k] <- th_nox[(k - 1)] + .01
    }
  }

  list(th = th, th_nox = th_nox, keep_idx = keep_idx)
}

# functions related to CFI and other 'incremental' fit indices

# lower-level functions:
# - lav_fit_cfi
# - lav_fit_rni (same as CFI, but without the max(0,))
# - lav_fit_tli/lav_fit_nnfi
# - lav_fit_rfi
# - lav_fit_nfi
# - lav_fit_pnfi
# - lav_fit_ifi

# higher-level functions:
# - lav_fit_cfi_lavobject

# Y.R. 20 July 2022

# CFI - comparative fit index (Bentler, 1990)
# robust version: Brosseau-Liard & Savalei MBR 2014, equation 15
# Brosseau-Liard, P. E., & Savalei, V. (2014). Adjusting incremental fit
# indices for nonnormality. Multivariate behavioral research, 49(5), 460-470.

# robust version MLMV (scaled.shifted)
# Savalei, V. (2018). On the computation of the RMSEA and CFI from the
# mean-and-variance corrected test statistic with nonnormal data in SEM.
# Multivariate behavioral research, 53(3), 419-429. eq 9

# note: robust MLM == robust MLMV

# categorical data:
# Savalei, V. (2021). Improving fit indices in structural equation modeling with
# categorical data. Multivariate Behavioral Research, 56(3), 390-407. doi:
# 10.1080/00273171.2020.1717922

# when missing = "fiml":
# Zhang, X., & Savalei, V. (2023). New computations for RMSEA and CFI following
# FIML and TS estimation with missing data. Psychological Methods, 28(2),
# 263-283. https://doi.org/10.1037/met0000445


lav_fit_cfi <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL,
                        c_hat = 1, c_hat_null = 1) {
  if (anyNA(c(x2, df, x2_null, df_null, c_hat, c_hat_null))) {
    return(as.numeric(NA))
  }

  # robust?
  if (df > 0 && !missing(c_hat) && !missing(c_hat_null) &&
    c_hat != 1 && c_hat_null != 1) {
    t1 <- max(c(x2 - (c_hat * df), 0))
    t2 <- max(c(x2 - (c_hat * df), x2_null - (c_hat_null * df_null), 0))
  } else {
    t1 <- max(c(x2 - df, 0))
    t2 <- max(c(x2 - df, x2_null - df_null, 0))
  }

  if (isTRUE(all.equal(t1, 0)) && isTRUE(all.equal(t2, 0))) {
    cfi <- 1
  } else {
    cfi <- 1 - t1 / t2
  }

  cfi
}

# RNI - relative noncentrality index (McDonald & Marsh, 1990)
# same as CFI, but without the max(0,)
lav_fit_rni <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL,
                        c_hat = 1, c_hat_null = 1) {
  if (anyNA(c(x2, df, x2_null, df_null, c_hat, c_hat_null))) {
    return(as.numeric(NA))
  }

  # robust?
  if (df > 0 && !missing(c_hat) && !missing(c_hat_null) &&
    c_hat != 1 && c_hat_null != 1) {
    t1 <- x2 - (c_hat * df)
    t2 <- x2_null - (c_hat_null * df_null)
  } else {
    t1 <- x2 - df
    t2 <- x2_null - df_null
  }

  if (isTRUE(all.equal(t2, 0))) {
    rni <- as.numeric(NA)
  } else if (!is.finite(t1) || !is.finite(t2)) {
    rni <- as.numeric(NA)
  } else {
    rni <- 1 - t1 / t2
  }

  rni
}

# TLI - Tucker-Lewis index (Tucker & Lewis, 1973)
# same as
# NNFI - nonnormed fit index (NNFI, Bentler & Bonett, 1980)
# note: formula in lavaan <= 0.5-20:
# t1 <- X2.null/df.null - X2/df
# t2 <- X2.null/df.null - 1
# if(t1 < 0 && t2 < 0) {
#    TLI <- 1
# } else {
#    TLI <- t1/t2
# }
# note: TLI original formula was in terms of fx/df, not X2/df
# then, t1 <- fx_0/df.null - fx/df
#       t2 <- fx_0/df.null - 1/N (or N-1 for wishart)

# note: in lavaan 0.5-21, we use the alternative formula:
# TLI <- 1 - ((X2 - df)/(X2.null - df.null) * df.null/df)
# - this one has the advantage that a 'robust' version
#   can be derived; this seems non-trivial for the original one
# - unlike cfi, we do not use 'max(0, )' for t1 and t2
#   therefore, t1 can go negative, and TLI can be > 1
lav_fit_tli <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL,
                        c_hat = 1, c_hat_null = 1) {
  if (anyNA(c(x2, df, x2_null, df_null, c_hat, c_hat_null))) {
    return(as.numeric(NA))
  }

  # robust?
  if (df > 0 && !missing(c_hat) && !missing(c_hat_null) &&
    c_hat != 1 && c_hat_null != 1) {
    t1 <- (x2 - c_hat * df) * df_null
    t2 <- (x2_null - c_hat_null * df_null) * df
  } else {
    t1 <- (x2 - df) * df_null
    t2 <- (x2_null - df_null) * df
  }

  if (df > 0 && abs(t2) > 0) {
    tli <- 1 - t1 / t2
  } else if (!is.finite(t1) || !is.finite(t2)) {
    tli <- as.numeric(NA)
  } else {
    tli <- 1
  }

  tli
}

# alias for nnfi
lav_fit_nnfi <- lav_alias("lav_fit_tli")

# RFI - relative fit index (Bollen, 1986; Joreskog & Sorbom 1993)
lav_fit_rfi <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL) {
  if (anyNA(c(x2, df, x2_null, df_null))) {
    return(as.numeric(NA))
  }

  if (df > df_null) {
    rli <- as.numeric(NA)
  } else if (df > 0 && df_null > 0) {
    t1 <- x2_null / df_null - x2 / df
    t2 <- x2_null / df_null
    if (!is.finite(t1) || !is.finite(t2)) {
      rli <- as.numeric(NA)
    } else if (t1 < 0 || t2 < 0) {
      rli <- 1
    } else {
      rli <- t1 / t2
    }
  } else {
    rli <- 1
  }

  rli
}

# NFI - normed fit index (Bentler & Bonett, 1980)
lav_fit_nfi <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL) {
  if (anyNA(c(x2, df, x2_null, df_null))) {
    return(as.numeric(NA))
  }

  if (df > df_null || isTRUE(all.equal(x2_null, 0))) {
    nfi <- as.numeric(NA)
  } else if (df > 0) {
    t1 <- x2_null - x2
    t2 <- x2_null
    nfi <- t1 / t2
  } else {
    nfi <- 1
  }

  nfi
}

# PNFI - Parsimony normed fit index (James, Mulaik & Brett, 1982)
lav_fit_pnfi <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL) {
  if (anyNA(c(x2, df, x2_null, df_null))) {
    return(as.numeric(NA))
  }

  if (df_null > 0 && x2_null > 0) {
    t1 <- x2_null - x2
    t2 <- x2_null
    pnfi <- (df / df_null) * (t1 / t2)
  } else {
    pnfi <- as.numeric(NA)
  }

  pnfi
}

# IFI - incremental fit index (Bollen, 1989; Joreskog & Sorbom, 1993)
lav_fit_ifi <- function(x2 = NULL, df = NULL, x2_null = NULL, df_null = NULL) {
  if (anyNA(c(x2, df, x2_null, df_null))) {
    return(as.numeric(NA))
  }

  t1 <- x2_null - x2
  t2 <- x2_null - df
  if (!is.finite(t1) || !is.finite(t2)) {
    ifi <- as.numeric(NA)
  } else if (t2 < 0) {
    ifi <- 1
  } else if (isTRUE(all.equal(t2, 0))) {
    ifi <- as.numeric(NA)
  } else {
    ifi <- t1 / t2
  }

  ifi
}

# higher-level function
lav_fit_cfi_lavobject <- function(lavobject = NULL, fit_measures = "cfi",
                                  baseline_model = NULL, h1_model = NULL,
                                  standard_test = "standard",
                                  scaled_test = "none",
                                  robust = TRUE,
                                  cat_nonpd = "na") {
  # check lavobject
  stopifnot(inherits(lavobject, "lavaan"))

  # check for categorical
  categorical_flag <- lavobject@Model@categorical

  # tests
  test <- lavobject@test
  test_names <- sapply(lavobject@test, "[[", "test")
  if (test_names[1] == "none" || standard_test == "none") {
    return(list())
  }
  test_idx <- which(test_names == standard_test)[1]
  if (length(test_idx) == 0L) {
    return(list())
  }

  scaled_flag <- FALSE
  if (!scaled_test %in% c("none", "standard", "default")) {
    scaled_idx <- which(test_names == scaled_test)
    if (length(scaled_idx) > 0L) {
      scaled_idx <- scaled_idx[1] # only the first one
      scaled_flag <- TRUE
    }
  }

  # robust?
  robust_flag <- FALSE
  if (robust && scaled_flag &&
    scaled_test %in% c(
      "satorra.bentler", "yuan.bentler.mplus",
      "yuan.bentler", "yuan.chan", "scaled.shifted",
      "scaled.shifted.corrected"
    )) {
    robust_flag <- TRUE
  }

  # FIML? (the FIML-corrected robust indices are only available for
  # single-level models; for multilevel data, we fall back to the regular
  # scaled/robust computation above, which handles missing data too)
  fiml_flag <- FALSE
  if (robust && lavobject@Options$missing %in% c("ml", "ml.x") &&
    lavobject@Data@nlevels == 1L) {
    fiml_flag <- robust_flag <- TRUE
    # check if we can compute corrected values
    if (scaled_flag) {
      version <- "V3"
    } else {
      version <- "V6"
    }
    fiml <- try(
      lav_fit_fiml_corrected(lavobject, baseline_model,
        version = version
      ),
      silent = TRUE
    )
    if (inherits(fiml, "try-error")) {
      lav_msg_warn(gettext("computation of robust CFI failed."))
      fiml <- list(
        XX3 = as.numeric(NA), df3 = as.numeric(NA),
        c.hat3 = as.numeric(NA), XX3.scaled = as.numeric(NA),
        XX3.null = as.numeric(NA), df3.null = as.numeric(NA),
        c.hat3.null = as.numeric(NA)
      )
    } else if (anyNA(c(
      fiml$XX3, fiml$df3, fiml$c.hat3, fiml$XX3.scaled,
      fiml$XX3.null, fiml$df3.null, fiml$c.hat3.null
    ))) {
      lav_msg_warn(gettext("computation of robust CFI resulted in NA values."))
    }
  }

  # supported fit measures in this function
  # baseline model
  fit_baseline <- c("baseline.chisq", "baseline.df", "baseline.pvalue")
  if (scaled_flag) {
    fit_baseline <- c(
      fit_baseline, "baseline.chisq.scaled",
      "baseline.df.scaled", "baseline.pvalue.scaled",
      "baseline.chisq.scaling.factor"
    )
  }

  fit_cfi_tli <- c("cfi", "tli")
  if (scaled_flag) {
    fit_cfi_tli <- c(fit_cfi_tli, "cfi.scaled", "tli.scaled")
  }
  if (robust_flag) {
    fit_cfi_tli <- c(fit_cfi_tli, "cfi.robust", "tli.robust")
  }

  # other incremental fit indices
  fit_cfi_other <- c("nnfi", "rfi", "nfi", "pnfi", "ifi", "rni")
  if (scaled_flag) {
    fit_cfi_other <- c(
      fit_cfi_other, "nnfi.scaled", "rfi.scaled",
      "nfi.scaled", "pnfi.scaled", "ifi.scaled", "rni.scaled"
    )
  }
  if (robust_flag) {
    fit_cfi_other <- c(fit_cfi_other, "nnfi.robust", "rni.robust")
  }

  # which one do we need?
  if (missing(fit_measures)) {
    # default set
    fit_measures <- c(fit_baseline, fit_cfi_tli)
  } else {
    # remove any not-CFI related index from fit.measures
    rm_idx <- which(!fit_measures %in%
      c(fit_baseline, fit_cfi_tli, fit_cfi_other))
    if (length(rm_idx) > 0L) {
      fit_measures <- fit_measures[-rm_idx]
    }
    if (length(fit_measures) == 0L) {
      return(list())
    }
  }


  # basic test statistics
  x2 <- test[[test_idx]]$stat
  df <- test[[test_idx]]$df
  # g <- lavobject@Data@ngroups # number of groups
  # n <- lav_inspect_ntotal(object = lavobject) # N vs N-1

  # scaled X2
  if (scaled_flag) {
    x2_scaled <- test[[scaled_idx]]$stat
    df_scaled <- test[[scaled_idx]]$df
  }
  if (robust_flag) {
    xx3 <- x2
    if (categorical_flag) {
      out <- try(lav_fit_catml_dwls(lavobject, nonpd = cat_nonpd),
        silent = TRUE
      )
      if (inherits(out, "try-error")) {
        xx3 <- df3 <- c_hat <- c_hat3 <- xx3_scaled <- as.numeric(NA)
      } else {
        xx3 <- out$XX3
        df3 <- out$df3
        c_hat3 <- c_hat <- out$c.hat3
        xx3_scaled <- out$XX3.scaled
      }
    } else if (fiml_flag) {
      xx3 <- fiml$XX3
      df3 <- fiml$df3
      c_hat3 <- c_hat <- fiml$c.hat3
      xx3_scaled <- fiml$XX3.scaled
    } else if (scaled_test %in% c(
      "scaled.shifted", "scaled.shifted.corrected"
    )) {
      # compute c.hat from a and b
      a <- test[[scaled_idx]]$scaling.factor
      b <- test[[scaled_idx]]$shift.parameter
      c_hat <- a * (df - b) / df
    } else {
      c_hat <- test[[scaled_idx]]$scaling.factor
    }
  }

  # output container
  indices <- list()

  # only do what is needed (per groups)
  cfi_baseline_flag <- cfi_tli_flag <- cfi_other_flag <- FALSE
  if (any(fit_baseline %in% fit_measures)) {
    cfi_baseline_flag <- TRUE
  }
  if (any(fit_cfi_tli %in% fit_measures)) {
    cfi_tli_flag <- TRUE
  }
  if (any(fit_cfi_other %in% fit_measures)) {
    cfi_other_flag <- TRUE
  }

  # 1. BASELINE model
  baseline_test <- NULL

  # is the (effective) baseline model the default independence model?
  # (needed below: the robust corrections for categorical data -- Savalei,
  #  2021, eq. 6b -- are only valid for the independence baseline)
  baseline_default_flag <- is.null(baseline_model) &&
    is.null(lavobject@external$baseline.model) &&
    !identical(lavobject@Options$baseline.type, "nested")

  # we use the following priority:
  # 1. user-provided baseline model
  # 2. baseline model in @external slot
  # 3. baseline model in @baseline slot
  # 4. nothing -> compute independence model

  # TDJ: Also check for user-supplied h1.model, using similar priority:
  #        1. user-provided h1 model
  #        2. h1 model in @external slot
  #        3. default h1 model (already in @h1 slot, no update necessary)

  # 1. user-provided h1 model
  if (!is.null(h1_model)) {
    stopifnot(inherits(h1_model, "lavaan"))

    # 2. h1 model in @external slot
  } else if (!is.null(lavobject@external$h1.model)) {
    stopifnot(inherits(lavobject@external$h1.model, "lavaan"))
    h1_model <- lavobject@external$h1.model
  } # else is.null

  # 1. user-provided baseline model
  if (!is.null(baseline_model)) {
    baseline_test <-
      lav_fit_check_baseline(
        fit_indep = baseline_model,
        object = lavobject,
        fit_h1 = h1_model # okay if NULL
      )
    # 2. baseline model in @external slot
  } else if (!is.null(lavobject@external$baseline.model)) {
    fit_indep <- lavobject@external$baseline.model
    baseline_test <-
      lav_fit_check_baseline(
        fit_indep = fit_indep,
        object = lavobject,
        fit_h1 = h1_model # okay if NULL
      )
    # 3. internal @baseline slot
  } else if (length(lavobject@baseline) > 0L &&
    !is.null(lavobject@baseline$test) &&
    ## if there is a custom h1.model, need  _check_baseline() to update @test
    is.null(h1_model)) {
    baseline_test <- lavobject@baseline$test
    # 4. (re)compute independence model
  } else {
    fit_indep <- try(lav_object_independence(lavobject), silent = TRUE)
    baseline_test <-
      lav_fit_check_baseline(
        fit_indep = fit_indep,
        object = lavobject,
        fit_h1 = h1_model # okay if NULL
      )
  }

  # baseline_test_idx
  baseline_test_idx <- which(names(baseline_test) == standard_test)[1]
  baseline_scaled_ok <- FALSE
  if (scaled_flag) {
    baseline_scaled_idx <- which(names(baseline_test) == scaled_test)[1]
    # the baseline model may lack the scaled test entry (eg a user-provided
    # baseline model that was fitted without the scaled test, or a baseline
    # that could not be refitted with the scaled test): degrade to NA for
    # the scaled/robust baseline quantities instead of failing
    baseline_scaled_ok <- !is.na(baseline_scaled_idx)
  }

  if (!is.null(baseline_test)) {
    x2_null <- baseline_test[[baseline_test_idx]]$stat
    df_null <- baseline_test[[baseline_test_idx]]$df
    if (scaled_flag) {
      if (baseline_scaled_ok) {
        x2_null_scaled <- baseline_test[[baseline_scaled_idx]]$stat
        df_null_scaled <- baseline_test[[baseline_scaled_idx]]$df
      } else {
        x2_null_scaled <- df_null_scaled <- as.numeric(NA)
      }
    }
    if (robust_flag) {
      xx3_null <- x2_null
      if (categorical_flag) {
        if (inherits(out, "try-error")) {
          xx3_null <- c_hat_null <- as.numeric(NA)
        } else if (!baseline_default_flag) {
          # the robust corrections for categorical data assume the
          # (default) independence baseline model; for any other baseline
          # model, mixing its test statistic/df with the independence-
          # based correction (tr Gamma) would be wrong -> return NA
          lav_msg_warn(gettext(
            "robust fit indices are currently not available for
             categorical data in combination with a non-default baseline
             model; returning NA."))
          xx3_null <- c_hat_null <- as.numeric(NA)
        } else {
          xx3_null <- out$XX3.null
          c_hat_null <- out$c.hat3.null
        }
      } else if (fiml_flag) {
        xx3_null <- fiml$XX3.null
        c_hat_null <- fiml$c.hat3.null
      } else if (!baseline_scaled_ok) {
        c_hat_null <- as.numeric(NA)
      } else if (scaled_test %in% c(
        "scaled.shifted", "scaled.shifted.corrected"
      )) {
        # compute c.hat from a and b
        a_null <-
          baseline_test[[baseline_scaled_idx]]$scaling.factor
        b_null <-
          baseline_test[[baseline_scaled_idx]]$shift.parameter
        c_hat_null <- a_null * (df_null - b_null) / df_null
      } else {
        c_hat_null <-
          baseline_test[[baseline_scaled_idx]]$scaling.factor
      }
    }
  } else {
    x2_null <- df_null <- as.numeric(NA)
    x2_null_scaled <- df_null_scaled <- as.numeric(NA)
    c_hat_null <- as.numeric(NA)
  }

  # check for NAs of nonfinite numbers
  if (!is.finite(x2) || !is.finite(df) ||
    !is.finite(x2_null) || !is.finite(df_null)) {
    indices[fit_measures] <- as.numeric(NA)
    return(indices)
  }

  # fill in baseline indices
  if (cfi_baseline_flag) {
    indices["baseline.chisq"] <- x2_null
    indices["baseline.df"] <- df_null
    indices["baseline.pvalue"] <- baseline_test[[baseline_test_idx]]$pvalue
    if (scaled_flag) {
      indices["baseline.chisq.scaled"] <- x2_null_scaled
      indices["baseline.df.scaled"] <- df_null_scaled
      if (baseline_scaled_ok) {
        indices["baseline.pvalue.scaled"] <-
          baseline_test[[baseline_scaled_idx]]$pvalue
        indices["baseline.chisq.scaling.factor"] <-
          baseline_test[[baseline_scaled_idx]]$scaling.factor
      } else {
        indices["baseline.pvalue.scaled"] <- as.numeric(NA)
        indices["baseline.chisq.scaling.factor"] <- as.numeric(NA)
      }
    }
  }

  # 2. CFI and TLI
  if (cfi_tli_flag) {
    indices["cfi"] <- lav_fit_cfi(
      x2 = x2, df = df,
      x2_null = x2_null, df_null = df_null
    )
    indices["tli"] <- lav_fit_tli(
      x2 = x2, df = df,
      x2_null = x2_null, df_null = df_null
    )
    if (scaled_flag) {
      indices["cfi.scaled"] <-
        lav_fit_cfi(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
      indices["tli.scaled"] <-
        lav_fit_tli(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
    }
    if (robust_flag) {
      indices["cfi.robust"] <-
        lav_fit_cfi(
          x2 = xx3, df = df,
          x2_null = xx3_null, df_null = df_null,
          c_hat = c_hat, c_hat_null = c_hat_null
        )
      indices["tli.robust"] <-
        lav_fit_tli(
          x2 = xx3, df = df,
          x2_null = xx3_null, df_null = df_null,
          c_hat = c_hat, c_hat_null = c_hat_null
        )
    }
  }

  # 3. other
  # c("nnfi", "rfi", "nfi", "pnfi", "ifi", "rni")
  if (cfi_other_flag) {
    indices["nnfi"] <-
      lav_fit_nnfi(x2 = x2, df = df, x2_null = x2_null, df_null = df_null)
    indices["rfi"] <-
      lav_fit_rfi(x2 = x2, df = df, x2_null = x2_null, df_null = df_null)
    indices["nfi"] <-
      lav_fit_nfi(x2 = x2, df = df, x2_null = x2_null, df_null = df_null)
    indices["pnfi"] <-
      lav_fit_pnfi(x2 = x2, df = df, x2_null = x2_null, df_null = df_null)
    indices["ifi"] <-
      lav_fit_ifi(x2 = x2, df = df, x2_null = x2_null, df_null = df_null)
    indices["rni"] <-
      lav_fit_rni(x2 = x2, df = df, x2_null = x2_null, df_null = df_null)

    if (scaled_flag) {
      indices["nnfi.scaled"] <-
        lav_fit_nnfi(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
      indices["rfi.scaled"] <-
        lav_fit_rfi(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
      indices["nfi.scaled"] <-
        lav_fit_nfi(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
      indices["pnfi.scaled"] <-
        lav_fit_pnfi(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
      indices["ifi.scaled"] <-
        lav_fit_ifi(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
      indices["rni.scaled"] <-
        lav_fit_rni(
          x2 = x2_scaled, df = df_scaled,
          x2_null = x2_null_scaled, df_null = df_null_scaled
        )
    }
    if (robust_flag) {
      indices["nnfi.robust"] <-
        lav_fit_nnfi(
          x2 = xx3, df = df,
          x2_null = xx3_null, df_null = df_null,
          c_hat = c_hat, c_hat_null = c_hat_null
        )
      indices["rni.robust"] <-
        lav_fit_rni(
          x2 = xx3, df = df,
          x2_null = xx3_null, df_null = df_null,
          c_hat = c_hat, c_hat_null = c_hat_null
        )
    }
  }

  # return only those that were requested
  indices[fit_measures]
}


# new in 0.6-5
# internal function to check the (external) baseline model, and
# return baseline 'test' list if everything checks out (and NULL otherwise)
lav_fit_check_baseline <- function(fit_indep = NULL, object = NULL,
                                            fit_h1 = NULL) {
  test <- NULL

  # check if everything is in order
  if (inherits(fit_indep, "try-error")) {
    lav_msg_warn(gettext("baseline model estimation failed"))
    return(NULL)
  } else if (!inherits(fit_indep, "lavaan")) {
    lav_msg_warn(gettext(
      "(user-provided) baseline model is not a fitted lavaan object"
    ))
    return(NULL)
  } else if (!fit_indep@optim$converged) {
    lav_msg_warn(gettext("baseline model did not converge"))
    return(NULL)
  } else {
    # evaluate if estimator/test matches original object
    # note: we do not need to check for 'se', as it may be 'none'
    same_test <- all(object@Options$test == fit_indep@Options$test)
    if (!same_test) {
      lav_msg_warn(gettextf(
        "Baseline model was using test(s) = %1$s, but original model was using
        test(s) = %2$s. Refitting baseline model!",
        lav_msg_view(fit_indep@Options$test, "none"),
        lav_msg_view(object@Options$test, "none")
      ))
    }
    same_estimator <- (object@Options$estimator ==
      fit_indep@Options$estimator)
    if (!same_estimator) {
      lav_msg_warn(gettextf(
        "Baseline model was using estimator = %1$s, but original model was
        using estimator = %2$s. Refitting baseline model!",
        dQuote(fit_indep@Options$estimator),
        dQuote(object@Options$estimator)
      ))
    }
    if (!same_test || !same_estimator) {
      lavoptions <- object@Options
      lavoptions$estimator <- object@Options$estimator
      lavoptions$se <- "none"
      lavoptions$baseline <- FALSE
      lavoptions$fit.by.level <- FALSE
      lavoptions$check.start <- FALSE
      lavoptions$check.post <- FALSE
      lavoptions$check.vcov <- FALSE
      lavoptions$test <- object@Options$test
      fit_indep <- try(
        lavaan(fit_indep,
          slot_options = lavoptions,
          slot_data = object@Data,
          slot_sample_stats = object@SampleStats,
          sloth1 = object@h1,
          slot_cache = object@Cache,
          verbose = FALSE
        ),
        silent = TRUE
      )
      # try again
      test <- lav_fit_check_baseline(
        fit_indep = fit_indep,
        object = object
      )
    } else {
      # extract what we need
      test <- fit_indep@test
    }
  } # converged lavaan object



  # TDJ: Check for user-supplied h1.model (here, the fit_h1 == argument)
  #      Similar to BASELINE model, use the following priority:
  #        1. user-provided h1 model
  #        2. h1 model in @external slot
  #        3. default h1 model (already in @h1 slot, no update necessary)
  # FIXME? user-supplied h1 model in object might be in fit.indep, too

  user_h1_exists <- FALSE
  # 1. user-provided h1 model
  if (!is.null(fit_h1)) {
    stopifnot(inherits(fit_h1, "lavaan"))
    user_h1_exists <- TRUE

    # 2. h1 model in @external slot
  } else if (!is.null(object@external$h1.model)) {
    stopifnot(inherits(object@external$h1.model, "lavaan"))
    fit_h1 <- object@external$h1.model
    user_h1_exists <- TRUE
  }

  if (user_h1_exists) {
    ## update @test slot
    test <- lav_update_test_custom_h1(
      lav_obj_h0 = fit_indep,
      lav_obj_h1 = fit_h1
    )@test
  }

  test
}

# data simulation engine behind simulateData() / lavSimulateData()
#
# This file contains the single, unified engine and its workers:
#  - lav_data_simulate()     : the public router (single-level vs multilevel)
#  - lav_data_simulate_sl()  : single-level worker (the historical engine);
#                              continuous, single-level output is byte-identical
#                              to the long-standing simulateData() behaviour, and
#                              it also handles ov_var / skewness / kurtosis /
#                              standardized / mass / categorical data
#  - lav_data_simulate_ml()  : multilevel worker; calls lavaan() directly to
#                              obtain the model-implied moments per block, and
#                              supports two levels (optionally multigroup)
#  - lav_data_simulate_old() : deprecated thin wrapper kept for backward
#                              compatibility
#  - helper functions (cluster_idx handling, Vale-Maurelli nonnormal data, ...)

# ---------------------------------------------------------------------------
# lav_data_simulate(): the single, unified data-simulation engine behind
# simulateData() / lavSimulateData().
#
# It routes to one of two internal workers, but presents a single interface:
#  - multilevel requests  -> lav_data_simulate_ml() (model-implied moments per
#    block, via a lavaan fit; supports two levels, optionally multigroup)
#  - single-level requests -> lav_data_simulate_sl() (the historical engine;
#    supports ov_var, skewness/kurtosis, standardized, mass, ...). The
#    continuous, single-level output is byte-identical to the historical
#    simulateData() output.
#
# The signature mirrors the historical simulateData() signature (so existing
# code keeps working), plus 'cluster_idx' (multilevel) and 'ordered_center'.
# ---------------------------------------------------------------------------
lav_data_simulate <- function(model = NULL,
                              model_type = "sem",
                              # model modifiers
                              meanstructure = FALSE,
                              int_ov_free = TRUE,
                              int_lv_free = FALSE,
                              marker_int_zero = FALSE,
                              conditional_x = FALSE,
                              composites = TRUE,
                              fixed_x = FALSE,
                              orthogonal = FALSE,
                              std_lv = TRUE,
                              auto_fix_first = FALSE,
                              auto_fix_single = FALSE,
                              auto_var = TRUE,
                              auto_cov_lv_x = TRUE,
                              auto_cov_y = TRUE,
                              ...,
                              # data properties
                              sample_nobs = 500L,
                              ov_var = NULL,
                              group_label = NULL,
                              skewness = NULL,
                              kurtosis = NULL,
                              cluster_idx = NULL,
                              # control
                              seed = NULL,
                              empirical = FALSE,
                              mass = TRUE,
                              ordered_center = TRUE,
                              return_type = "data.frame",
                              return_fit = FALSE,
                              debug = FALSE,
                              standardized = FALSE) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)

  multilevel <- !is.null(cluster_idx) || lav_simulate_is_multilevel(model)

  if (multilevel) {
    # ------- multilevel worker -------
    # the multilevel worker does not (yet) support these (single-level) features
    if (!is.null(ov_var) || !is.null(skewness) || !is.null(kurtosis) ||
        isTRUE(standardized) || isTRUE(mass)) {
      lav_msg_warn(gettext(
        "arguments 'ov_var', 'skewness', 'kurtosis', 'standardized' and 'mass'
        are not supported for multilevel data and will be ignored"))
    }
    return_value <- do.call(lav_data_simulate_ml, c(
      list(model = model, cmd_pop = model_type,
      # forward the model modifiers to the population-model fit; the ML worker
      # forwards these straight to cmd_pop (sem/cfa/lavaan), which expect the
      # dotted lavOptions names (not the snake_case local variable names)
      "int.ov.free" = int_ov_free, "int.lv.free" = int_lv_free,
      "marker.int.zero" = marker_int_zero, "conditional.x" = conditional_x,
      composites = composites, "fixed.x" = fixed_x, orthogonal = orthogonal,
      "std.lv" = std_lv, "auto.fix.first" = auto_fix_first,
      "auto.fix.single" = auto_fix_single, "auto.var" = auto_var,
      "auto.cov.lv.x" = auto_cov_lv_x, "auto.cov.y" = auto_cov_y),
      dotdotdot,
      list(sample_nobs = sample_nobs, cluster_idx = cluster_idx,
      seed = seed, empirical = empirical, ordered_center = ordered_center,
      return_fit = return_fit, output = return_type)))
    return(return_value)
  }

  # ------- single-level worker -------
  do.call(lav_data_simulate_sl, c(
    list(model = model, model_type = model_type,
    meanstructure = meanstructure, int_ov_free = int_ov_free,
    int_lv_free = int_lv_free, marker_int_zero = marker_int_zero,
    conditional_x = conditional_x, composites = composites,
    fixed_x = fixed_x, orthogonal = orthogonal, std_lv = std_lv,
    auto_fix_first = auto_fix_first, auto_fix_single = auto_fix_single,
    auto_var = auto_var, auto_cov_lv_x = auto_cov_lv_x,
    auto_cov_y = auto_cov_y), dotdotdot,
    # note: 'group_label' is deliberately NOT threaded here. it is a vestigial
    # argument (never read by the single-level worker, and never has been), so
    # forcing it would be pointless -- and harmful: it lets a caller-supplied
    # lazy default such as 'paste("G", 1:ngroups)' (an idiom copied from old
    # lavaan, where 'ngroups' only existed internally) stay unevaluated, as it
    # did before the simulateData() refactor.
    list(sample_nobs = sample_nobs, ov_var = ov_var,
    skewness = skewness, kurtosis = kurtosis, seed = seed,
    empirical = empirical, mass = mass, return_type = return_type,
    return_fit = return_fit, debug = debug, standardized = standardized,
    ordered_center = ordered_center)))
}

# exported synonym
lavSimulateData <- lav_alias("lav_data_simulate")

# detect whether a model (syntax string or parameter table) is multilevel
lav_simulate_is_multilevel <- function(model = NULL) {
  if (is.null(model)) {
    return(FALSE)
  }
  if (is.character(model)) {
    txt <- paste(model, collapse = "\n")
    return(grepl("(^|\n)[ \t]*level[ \t]*:", txt))
  }
  if (is.list(model)) {
    if (!is.null(model$level)) {
      lev <- model$level[!is.na(model$level)]
      return(length(unique(lev[lev > 0L])) > 1L)
    }
  }
  FALSE
}

# multilevel data-simulation worker behind lav_data_simulate()
lav_data_simulate_ml <- function(model = NULL,
                              cmd_pop = "sem",
                              ...,
                              # data properties
                              sample_nobs = 1000L,
                              cluster_idx = NULL,
                              # control
                              seed = NULL,
                              empirical = FALSE,
                              ordered_center = TRUE,
                              # output
                              add_labels = TRUE,
                              return_fit = FALSE,
                              output = "data.frame") {
  if (!is.null(seed)) set.seed(seed)

  # dotdotdot
  dotdotdot <- list(...)
  dotdotdot_orig <- dotdotdot

  # remove/override some options
  dotdotdot$verbose <- FALSE
  dotdotdot$debug <- FALSE
  dotdotdot$data <- NULL
  dotdotdot$sample.cov <- NULL
  dotdotdot$do.fit <- NULL

  # always use meanstructure = TRUE
  dotdotdot$meanstructure <- TRUE

  # helper to 'fit' the (no-data) population model; for multigroup models that
  # are only defined via c() modifiers, lavaan needs sample_nobs to know the
  # number of groups, so we retry with sample_nobs if the first attempt fails
  fit_pop_model <- function(dd) {
    out <- tryCatch(do.call(cmd_pop, args = c(list(model = model), dd)),
                    error = function(e) e)
    if (inherits(out, "error")) {
      out <- do.call(cmd_pop, args = c(list(model = model,
                                            sample_nobs = sample_nobs), dd))
    }
    out
  }

  # 'fit' population model: first pretend we generate continuous data only
  # (so the model-implied moments are the latent-response moments)
  dotdotdot_con <- dotdotdot
  dotdotdot_con$ordered <- NULL
  fit_con <- fit_pop_model(dotdotdot_con)

  # categorical? refit keeping the 'ordered' argument, to obtain thresholds
  if (!is.null(dotdotdot_orig$ordered) || fit_con@Model@categorical) {
    fit_pop <- fit_pop_model(dotdotdot)
  } else {
    fit_pop <- fit_con
  }

  # extract model implied statistics and data slot
  # NOTE: use delta = FALSE so that (for categorical data) we obtain the
  # *unscaled* latent-response covariances. The delta scaling standardizes
  # each block's variances to 1 for the WLS sample statistics, but for data
  # generation we need the actual within/between latent-response covariances,
  # so that y* = y*(within) + y*(between) has the correct variance
  # decomposition and the thresholds cut it correctly. For continuous data,
  # delta is the identity and this has no effect.
  lavimplied  <- lav_model_implied(fit_con@Model, delta = FALSE)
  lavdata     <- fit_pop@Data
  lavmodel    <- fit_pop@Model
  lavpartable <- fit_pop@ParTable
  lavoptions  <- fit_pop@Options

  # number of groups/levels/blocks
  ngroups <- lav_pt_ngroups(lavpartable)
  nblocks <- lav_pt_nblocks(lavpartable)
  nlevels <- lavdata@nlevels

  # check/expand the sample_nobs and cluster_idx arguments
  tmp <- lav_data_simulate_nobs(sample_nobs = sample_nobs,
                                cluster_idx = cluster_idx,
                                ngroups = ngroups, nblocks = nblocks,
                                nlevels = nlevels)
  sample_nobs <- tmp$sample_nobs
  cluster_idx <- tmp$cluster_idx

  # check if ov.names are the same for each group
  if (ngroups > 1L) {
    N1 <- lavdata@ov.names[[1]]
    if (!all(sapply(lavdata@ov.names, function(x) all(x %in% N1)))) {
      if (output == "data.frame") {
        output <- "matrix"
        lav_msg_warn(gettext(
          "groups do not contain the same set of variables; changing output=
          argument to \"matrix\""))
      }
    }
  }

  # prepare data containers
  X <- vector("list", length = nblocks)

  # generate data per BLOCK
  for (b in seq_len(nblocks)) {
    if (lavoptions$conditional.x) {
      lav_msg_stop(gettext("conditional.x = TRUE is not supported (yet) by the
                            multilevel data simulation"))
    } else {
      COV <- lavimplied$cov[[b]]
      MU  <- lavimplied$mean[[b]]
    }

    # if empirical = TRUE, rescale by N/(N-1), so that estimator = ML returns
    # exact results
    if (empirical) {
      if (sample_nobs[b] < NCOL(COV)) {
        lav_msg_stop(gettextf(
          "empirical = TRUE requires sample_nobs (= %1$s) to be larger than the
          number of variables (= %2$s) in block = %3$s",
          sample_nobs[b], NCOL(COV), b))
      }
      if (nlevels > 1L && (b %% nlevels == 1L)) {
        COV <- COV * sample_nobs[b] / (sample_nobs[b] - sample_nobs[b + 1])
      } else {
        COV <- COV * sample_nobs[b] / (sample_nobs[b] - 1)
      }
    }

    # generate normal data (using sign-invariant method for reproducibility)
    tmp <- try(lav_mvrnorm(
      n = sample_nobs[b],
      mu = MU, sigma_1 = COV, empirical = empirical),
      silent = TRUE)

    if (inherits(tmp, "try-error")) {
      # something went wrong; most likely: non-positive COV?
      ev <- eigen(COV, symmetric = TRUE, only.values = TRUE)$values
      if (any(ev < 0)) {
        lav_msg_stop(gettextf(
          "model-implied covariance matrix is not positive-definite in block =
          %1$s; smallest eigen value = %2$s; change the model parameters.",
          b, round(min(ev), 5)))
      } else {
        lav_msg_stop(gettextf("data generation failed for block = %s", b))
      }
    } else {
      X[[b]] <- unname(tmp)
    }
  } # block

  if (output == "block") {
    return(X)
  }

  # if multilevel, make a copy, and create X[[g]] per group
  if (nlevels > 1L) {
    x_block <- X
    X <- vector("list", length = ngroups)
  }

  # assemble data per group
  group_values <- lav_pt_group_values(lavpartable)
  for (g in 1:ngroups) {
    # multilevel?
    if (nlevels > 1L) {
      # which block?
      bb <- (g - 1) * nlevels + 1L

      Lp <- lavdata@Lp[[g]]
      p_tilde <- length(lavdata@ov.names[[g]])
      tmp1 <- matrix(0, nrow(x_block[[bb]]), p_tilde + 1L) # one extra column
      tmp2 <- matrix(0, nrow(x_block[[bb]]), p_tilde + 1L) # for the clus id

      # level 1
      tmp1[, Lp$ov.idx[[1]]] <- x_block[[bb]]

      # level 2 (expand cluster-level values to the level-1 units)
      tmp2[, Lp$ov.idx[[2]]] <- x_block[[bb + 1L]][cluster_idx[[g]], ,
                                                   drop = FALSE]
      # final
      X[[g]] <- tmp1 + tmp2

      # cluster id
      X[[g]][, p_tilde + 1L] <- cluster_idx[[g]]
    }

    # add variable names?
    if (add_labels) {
      if (nlevels > 1L) {
        colnames(X[[g]]) <- c(lavdata@ov.names[[g]], "cluster")
      } else {
        colnames(X[[g]]) <- lavdata@ov.names[[g]]
      }
    }

    # any categorical variables?
    ov_ord <- lav_object_vnames(fit_pop, "ov.ord", group = group_values[g])
    if (is.list(ov_ord)) {
      # multilevel -> use within level only
      ov_ord <- ov_ord[[1L]]
    }
    if (length(ov_ord) > 0L) {
      ov_names <- lavdata@ov.names[[g]]

      # which block?
      bb <- (g - 1) * nlevels + 1L

      # th/names
      # use delta = FALSE: the thresholds must be on the same (unscaled)
      # latent-response scale as the generated y* data (see the delta = FALSE
      # note where lavimplied is computed above)
      th__val <- as.numeric(lav_model_th(lavmodel = fit_pop@Model,
                                        delta = FALSE)[[bb]])
      if (length(lavmodel@num.idx[[bb]]) > 0L) {
        num_idx <- which(lavmodel@th_idx[[bb]] == 0)
        th__val <- th__val[-num_idx]
      }
      th_names <- fit_pop@pta$vnames$th[[bb]]
      th__names <- sapply(strsplit(th_names, split = "|", fixed = TRUE),
                         "[[", 1L)

      # use thresholds to cut
      for (o in ov_ord) {
        o_idx  <- which(o == ov_names)
        th_idx <- which(o == th__names)
        th_val <- c(-Inf, sort(th__val[th_idx]), +Inf)
        tmp <- X[[g]][, o_idx]
        if (ordered_center) {
          # center first (so the cut also works when the model-implied 'mean'
          # is nonzero); set ordered_center = FALSE for the 'old' behaviour
          tmp <- tmp - mean(tmp, na.rm = TRUE)
        }
        X[[g]][, o_idx] <- cut(tmp, th_val, labels = FALSE)
      }
    }
  }

  # output
  if (output == "matrix") {
    if (ngroups == 1L) {
      out <- X[[1L]]
    } else {
      out <- X
    }
  } else if (output == "data.frame") {
    if (ngroups == 1L) {
      out <- as.data.frame(X[[1L]], stringsAsFactors = FALSE)
    } else {
      # rbind + add group column
      out <- as.data.frame(do.call("rbind", X), stringsAsFactors = FALSE)
      out$group <- rep.int(1:ngroups, times = sapply(X, NROW))
    }
  } else if (output == "cov") {
    if (ngroups == 1L) {
      out <- cov(X[[1L]])
    } else {
      out <- lapply(X, cov)
    }
  } else {
    lav_msg_stop(gettextf("unknown option for argument output: %s", output))
  }

  if (return_fit) {
    attr(out, "fit") <- fit_pop
  }

  out
}

# check/expand the sample_nobs and cluster_idx arguments; for the multilevel
# case, return a per-block sample_nobs vector and a per-group cluster_idx list
lav_data_simulate_nobs <- function(sample_nobs = 1000L, cluster_idx = NULL,
                                   ngroups = 1L, nblocks = 1L, nlevels = 1L) {
  if (nlevels > 1L) {
    if (is.null(cluster_idx)) {
      # no cluster_idx given: build balanced clusters from sample_nobs
      # sample_nobs is interpreted (per group) as:
      #  - a single number: total number of level-1 units (clusters default
      #    to 1/10 of that, with a minimum of 2)
      #  - a vector c(n.level1, n.level2): #level-1 units and #clusters
      if (length(sample_nobs) == 1L) {
        n1 <- as.integer(sample_nobs)
        n2 <- max(2L, as.integer(round(n1 / 10)))
        nobs_l <- c(n1, n2)
      } else if (length(sample_nobs) == nlevels) {
        nobs_l <- as.integer(sample_nobs)
      } else {
        lav_msg_stop(gettext("for the multilevel case, sample_nobs should be a
                             single number, or a vector with one number per
                             level (eg c(1000, 100))"))
      }
      cluster_idx <- vector("list", ngroups)
      for (g in seq_len(ngroups)) {
        cluster_idx[[g]] <- lav_data_simulate_clusidx(nobs_l[1L], nobs_l[2L])
      }
    } else {
      # cluster_idx given
      if (!is.list(cluster_idx)) {
        cluster_idx <- rep(list(cluster_idx), ngroups)
      }
    }
    # cluster_idx values within each group should be 1, 2, ... nclus
    sample_nobs <- numeric(nblocks)
    for (g in seq_len(ngroups)) {
      cluster_idx[[g]] <- as.integer(as.factor(cluster_idx[[g]]))
      gg <- (g - 1L) * nlevels + 1L
      sample_nobs[gg]     <- length(cluster_idx[[g]])
      sample_nobs[gg + 1] <- length(unique(cluster_idx[[g]]))
    }
  } else {
    # single level
    if (length(sample_nobs) == ngroups) {
      # nothing to do
    } else if (ngroups > 1L && length(sample_nobs) == 1L) {
      sample_nobs <- rep.int(sample_nobs, ngroups)
    } else {
      lav_msg_stop(gettextf(
        "ngroups = %1$s but sample_nobs has length = %2$s",
        ngroups, length(sample_nobs)))
    }
  }

  list(sample_nobs = as.integer(sample_nobs), cluster_idx = cluster_idx)
}

# create a balanced cluster_idx vector with n1 level-1 units in n2 clusters
lav_data_simulate_clusidx <- function(n1, n2) {
  if (n2 < 1L) n2 <- 1L
  if (n2 > n1) n2 <- n1
  base <- n1 %/% n2
  rem  <- n1 %% n2
  sizes <- rep.int(base, n2)
  if (rem > 0L) sizes[seq_len(rem)] <- sizes[seq_len(rem)] + 1L
  rep.int(seq_len(n2), times = sizes)
}


# single-level data-simulation worker (the historical engine)
#
# initial version: YR 24 jan 2011
# revision for 0.4-11: YR 21 okt 2011
#
# This is the single-level worker behind lav_data_simulate(). It is kept as a
# separate function because the continuous, single-level output must remain
# byte-identical to the historical simulateData() output. The multilevel case is
# handled by lav_data_simulate_ml().
#
lav_data_simulate_sl <- function( # user-specified model    # nolint start
                         model = NULL,
                         model_type = "sem",
                         # model modifiers
                         meanstructure = FALSE,
                         int_ov_free = TRUE,
                         int_lv_free = FALSE,
                         marker_int_zero = FALSE,
                         conditional_x = FALSE,
                         composites = TRUE,
                         fixed_x = FALSE,
                         orthogonal = FALSE,
                         std_lv = TRUE,
                         auto_fix_first = FALSE,
                         auto_fix_single = FALSE,
                         auto_var = TRUE,
                         auto_cov_lv_x = TRUE,
                         auto_cov_y = TRUE,
                         ...,
                         # data properties
                         sample_nobs = 500L,
                         ov_var = NULL,
                         group_label = NULL, # vestigial; never used below
                         skewness = NULL,
                         kurtosis = NULL,
                         # control
                         seed = NULL,
                         empirical = FALSE,
                         mass = FALSE,
                         return_type = "data.frame",
                         return_fit = FALSE,
                         debug = FALSE,
                         standardized = FALSE,
                         ordered_center = FALSE) {         # nolint end
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  if (!is.null(seed)) set.seed(seed)
  # if(!exists(".Random.seed", envir = .GlobalEnv))
  #    runif(1)               # initialize the RNG if necessary
  # RNGstate <- .Random.seed

  # lav_model_pt
  if (is.list(model)) {
    # two possibilities: either model is already lavaanified
    # or it is something else...
    if (!is.null(model$lhs) && !is.null(model$op) &&
      !is.null(model$rhs) && !is.null(model$free)) {
      lav <- model

      # until 0.6-5, we only used the 'ustart' column
      # but what if 'lav' is a fitted lavaan object -> use 'est'
      if (!is.null(lav$est)) {
        lav$ustart <- lav$est
        lav$se <- NULL
        lav$est <- NULL
        lav$start <- NULL
      }
    } else if (is.character(model[[1]])) {
      lav_msg_stop(gettext("model is a list, but not a parameterTable?"))
    }
  } else {
    lav <- lav_model_pt(
      model = model,
      meanstructure = meanstructure,
      int_ov_free = int_ov_free,
      int_lv_free = int_lv_free,
      marker_int_zero = marker_int_zero,
      composites = composites,
      conditional_x = conditional_x,
      fixed_x = fixed_x,
      orthogonal = orthogonal,
      std_lv = std_lv,
      auto_fix_first = auto_fix_first,
      auto_fix_single = auto_fix_single,
      auto_var = auto_var,
      auto_cov_lv_x = auto_cov_lv_x,
      auto_cov_y = auto_cov_y,
      ngroups = length(sample_nobs)
    )
  }

  group_values <- lav_pt_group_values(lav)
  if (lav_debug()) {
    cat("initial lav\n")
    print(as.data.frame(lav))
  }

  # fill in any remaining NA values (needed for unstandardize)
  # 1 for variances and (unstandardized) factor loadings, 0 otherwise
  idx <- which(lav$op == "=~" & is.na(lav$ustart))
  if (length(idx) > 0L) {
    if (standardized) {
      lav$ustart[idx] <- 0.7
    } else {
      lav$ustart[idx] <- 1.0
    }
  }

  idx <- which(lav$op == "~~" & is.na(lav$ustart) & lav$lhs == lav$rhs)
  if (length(idx) > 0L) lav$ustart[idx] <- 1.0

  idx <- which(lav$op == "~" & is.na(lav$ustart))
  if (length(idx) > 0L) {
    lav_msg_warn(gettext(
      "some regression coefficients are unspecified and will be set to zero"))
  }

  idx <- which(is.na(lav$ustart))
  if (length(idx) > 0L) lav$ustart[idx] <- 0.0

  if (lav_debug()) {
    cat("lav + default values\n")
    print(as.data.frame(lav))
  }

  # set residual variances to enforce a standardized solution
  # but only if no *residual* variances have been specified in the syntax

  if (standardized) {
    # check if factor loadings are smaller than 1.0
    lambda_idx <- which(lav$op == "=~")
    if (any(lav$ustart[lambda_idx] >= 1.0)) {
      lav_msg_warn(gettext("standardized=TRUE but factor loadings are >= 1.0"))
    }

    # check if regression coefficients are smaller than 1.0
    reg_idx <- which(lav$op == "~")
    if (any(lav$ustart[reg_idx] >= 1.0)) {
      lav_msg_warn(gettext(
        "standardized=TRUE but regression coefficients are >= 1.0"))
    }

    # for ordered observed variables, we will get '0.0', but that is ok
    # so there is no need to make a distinction between numeric/ordered
    # here??
    ngroups <- lav_pt_ngroups(lav)
    ov_names <- lav_pt_vnames(lav, "ov")
    ov_nox <- lav_pt_vnames(lav, "ov.nox")
    # lv_names <- lav_pt_vnames(lav, "lv")
    # lv_y <- lav_pt_vnames(lav, "lv.y")
    lv_nox <- lav_pt_vnames(lav, "lv.nox")
    ov_var_idx <- which(lav$op == "~~" & lav$lhs %in% ov_nox &
      lav$rhs == lav$lhs)
    lv_var_idx <- which(lav$op == "~~" & lav$lhs %in% lv_nox &
      lav$rhs == lav$lhs)
    if (any(lav$user[c(ov_var_idx, lv_var_idx)] > 0L)) {
      lav_msg_warn(gettext(
        "if residual variances are specified, please use standardized=FALSE"))
    }

    # new in 0.6-20: - use lav_lisrel_residual_variances
    #                - use lav_lisrel_comp_set_intresvar
    dotdotdot <- list(...)
    dotdotdot$sample_nobs <- sample_nobs
    dotdotdot[["fixed.x"]] <- FALSE # for now (lavOptions name, not snake_case)
    dotdotdot$representation <- "LISREL"
    dotdotdot$composites <- composites
    dotdotdot$correlation <- TRUE # this is the trick
    tmp_fit <- do.call("lavaan", args = c(list(model = lav), dotdotdot))
    # set/get parameters to invoke lav_lisrel_residual_variances
    tmp_lav <- tmp_fit@ParTable
    tmp_x <- lav_model_get_parameters(tmp_fit@Model)
    tmp_model <- lav_model_set_parameters(tmp_fit@Model, x = tmp_x)
    tmp_lav$ustart <- lav_model_get_parameters(tmp_model, type = "user")

    # copy residual values to lav (without assuming parameter tables look the
    # same)
    res_idx <- c(ov_var_idx, lv_var_idx)
    for (i in seq_along(res_idx)) {
      # lookup this parameter in tmp.lav
      idx_in_lav <- res_idx[i]
      this_lhs <- lav$lhs[idx_in_lav]
      idx_in_tmp <- which(tmp_lav$op == "~~" & tmp_lav$lhs == this_lhs &
                          tmp_lav$rhs == tmp_lav$lhs)
      if (length(idx_in_tmp) == 0L) {
        # hm, not found? Give a warning?
      } else {
        vals <- tmp_lav$ustart[idx_in_tmp]
        # check if we have unfortunate values
        bad_idx <- which(!is.finite(vals) | vals < 0)
        vals[bad_idx] <- 1.0 # not pretty, but safe
        lav$ustart[idx_in_lav] <- vals
      }
    }

    # this is what we did <0.6-20
    # fit <- lavaan(model = lav, sample_nobs = sample_nobs, ...)
    # Sigma.hat <- lav_model_sigma(lavmodel = fit@Model)
    # ETA <- lav_model_veta(lavmodel = fit@Model)

    # if (lav_debug()) {
    #   cat("Sigma.hat:\n")
    #   print(Sigma.hat)
    #   cat("Eta:\n")
    #   print(ETA)
    # }

    # # stage 1: standardize LV
    # if (length(lv.nox) > 0L) {
    #   for (g in 1:ngroups) {
    #     var.group <- which(lav$op == "~~" & lav$lhs %in% lv.nox &
    #       lav$rhs == lav$lhs &
    #       lav$group == group_values[g])
    #     eta.idx <- match(lv.nox, lv.names)
    #     lav$ustart[var.group] <- 1 - diag(ETA[[g]])[eta.idx]
    #   }
    # }
    # # refit
    # fit <- lavaan(model = lav, sample_nobs = sample_nobs, ...)
    # Sigma.hat <- lav_model_sigma(lavmodel = fit@Model)

    # if (lav_debug()) {
    #   cat("after stage 1:\n")
    #   cat("Sigma.hat:\n")
    #   print(Sigma.hat)
    # }

    # # stage 2: standardize OV
    # for (g in 1:ngroups) {
    #   var.group <- which(lav$op == "~~" & lav$lhs %in% ov.nox &
    #     lav$rhs == lav$lhs &
    #     lav$group == group_values[g])
    #   ov.idx <- match(ov.nox, ov_names)
    #   lav$ustart[var.group] <- 1 - diag(Sigma.hat[[g]])[ov.idx]
    # }

    # if (lav_debug()) {
    #   cat("after standardisation lav\n")
    #   print(as.data.frame(lav))
    # }
  }


  # unstandardize
  if (!is.null(ov_var)) {
    # FIXME: if ov_var is named, check the order of the elements

    # 1. unstandardize observed variables
    lav$ustart <- lav_unstandardize_ov(partable = lav, ov_var = ov_var)

    # 2. unstandardized latent variables

    if (lav_debug()) {
      cat("after unstandardisation lav\n")
      print(as.data.frame(lav))
    }
  }

  # fit the model without data
  fit <- lavaan(model = lav, sample_nobs = sample_nobs, ...)

  # the model-implied moments for the population
  sigma_hat <- lav_model_sigma(lavmodel = fit@Model)
  mu_hat <- lav_model_mu(lavmodel = fit@Model)
  if (fit@Model@categorical) {
    th <- lav_model_th(lavmodel = fit@Model)
  }

  if (lav_debug()) {
    cat("\nModel-implied moments (before Vale-Maurelli):\n")
    print(sigma_hat)
    print(mu_hat)
    if (exists("TH")) print(th)
  }

  # ngroups
  ngroups <- length(sample_nobs)

  # prepare
  x <- vector("list", length = ngroups)
  # out <- vector("list", length = ngroups)

  for (g in 1:ngroups) {
    cov_1 <- sigma_hat[[g]]

    # if empirical = TRUE, rescale by N/(N-1), so that estimator=ML
    # returns exact results
    if (empirical) {
      cov_1 <- cov_1 * sample_nobs[g] / (sample_nobs[g] - 1)
    }

    # Using sign-invariant method for cross-machine reproducibility
    if (is.null(skewness) && is.null(kurtosis)) {
      if (mass) {
        x[[g]] <- MASS::mvrnorm(
          n = sample_nobs[g],
          mu = mu_hat[[g]],
          Sigma = cov_1,
          empirical = empirical
        )
      } else {
        x[[g]] <- lav_mvrnorm(
          n = sample_nobs[g],
          mu = mu_hat[[g]],
          sigma_1 = cov_1,
          empirical = empirical
        )
      }
    } else {
      # first generate Z
      z <- lav_data_valemaurelli1983(
        n = sample_nobs[g],
        cor_1 = cov2cor(cov_1),
        skewness = skewness, # FIXME: per group?
        kurtosis = kurtosis, mass = mass
      )
      # rescale
      # Note: 'scale()' will first center, and then scale
      # but we need to first scale, and then center...
      # this was reported by Jordan Brace (9 may 2014)
      # X[[g]] <- scale(Z, center = -Mu.hat[[g]],
      #                   scale  = 1/sqrt(diag(COV)))

      # first, we scale
      tmp <- scale(z,
        center = FALSE,
        scale = 1 / sqrt(diag(cov_1))
      )[, , drop = FALSE]

      # then, we center
      x[[g]] <- sweep(tmp, MARGIN = 2, STATS = mu_hat[[g]], FUN = "+")
    }

    # any categorical variables?
    ov_ord <- lav_pt_vnames(lav, type = "ov.ord", group = group_values[g])
    if (length(ov_ord) > 0L) {
      ov_names <- lav_pt_vnames(lav, type = "ov", group = group_values[g])
      # use thresholds to cut
      for (o in ov_ord) {
        o_idx <- which(o == ov_names)
        th_idx <- which(lav$op == "|" & lav$lhs == o &
          lav$group == group_values[g])
        th_val <- c(-Inf, sort(lav$ustart[th_idx]), +Inf)
        tmp_o <- x[[g]][, o_idx]
        if (ordered_center) {
          # center first (so the cut also works when the model-implied 'mean'
          # is nonzero); ordered_center = FALSE reproduces the historical cut
          tmp_o <- tmp_o - mean(tmp_o, na.rm = TRUE)
        }
        x[[g]][, o_idx] <- as.integer(cut(tmp_o, th_val))
      }
    }

    if (return_type == "data.frame") x[[g]] <- as.data.frame(x[[g]])
  }

  if (return_type == "matrix") {
    if (ngroups == 1L) {
      x[[1L]]
    } else {
      x
    }
  } else if (return_type == "data.frame") {
    data_1 <- x[[1L]]

    # if multiple groups, add group column
    if (ngroups > 1L) {
      for (g in 2:ngroups) {
        data_1 <- rbind(data_1, x[[g]])
      }
      data_1$group <- rep(1:ngroups, times = sample_nobs)
    }
    var_names <- lav_pt_vnames(fit@ParTable, type = "ov", group = 1L)
    if (ngroups > 1L) var_names <- c(var_names, "group")
    names(data_1) <- var_names
    if (return_fit) {
      attr(data_1, "fit") <- fit
    }
    data_1
  } else if (return_type == "cov") {
    if (ngroups == 1L) {
      cov(x[[1L]])
    } else {
      cov_list <- lapply(x, cov)
      cov_list
    }
  }
}
# deprecated: kept for backward compatibility (eg as the default data_function
# of lavSimulate()); routes through the unified engine lav_data_simulate(), which
# dispatches to lav_data_simulate_sl() (single level) or lav_data_simulate_ml()
# (multilevel). New code should call lavSimulateData() / lav_data_simulate().
lav_data_simulate_old <- function(..., ordered_center = FALSE) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  do.call(lav_data_simulate, c(dotdotdot, list(ordered_center = ordered_center)))
}



lav_data_valemaurelli1983 <- function(n = 100L, cor_1, skewness, kurtosis,
                                      mass = FALSE) {
  fleishman1978_abcd <- function(skewness, kurtosis) {
    system_function <- function(x, skewness, kurtosis) {
      b <- x[1L]
      c_1 <- x[2L]
      d <- x[3L]
      eq1 <- b * b + 6 * b * d + 2 * c_1 * c_1 + 15 * d * d - 1
      eq2 <- 2 * c_1 * (b * b + 24 * b * d + 105 * d * d + 2) - skewness
      eq3 <- 24 * (b * d + c_1 * c_1 * (1 + b * b + 28 * b * d) +
        d * d * (12 + 48 * b * d + 141 * c_1 * c_1 + 225 * d * d)) - kurtosis
      eq <- c(eq1, eq2, eq3)
      sum(eq * eq) ## SS
    }

    out <- nlminb(
      start = c(1, 0, 0), objective = system_function,
      scale = 10,
      control = list(trace = 0),
      skewness = skewness, kurtosis = kurtosis
    )
    if (out$convergence != 0 || out$objective > 1e-5) {
      lav_msg_warn(gettext("lav_data_valemaurelli1983 method did not converge,
                   or it did not find the roots"))
    }
    b <- out$par[1L]
    c_1 <- out$par[2L]
    d <- out$par[3L]
    a <- -c_1
    c(a, b, c_1, d)
  }

  get_icov <- function(b1, c1, d1, b2, c2, d2, r) {
    objective_function <- function(x, b1, c1, d1, b2, c2, d2, r) {
      rho <- x[1L]
      eq <- rho * (b1 * b2 + 3 * b1 * d2 + 3 * d1 * b2 + 9 * d1 * d2) +
        rho * rho * (2 * c1 * c2) + rho * rho * rho * (6 * d1 * d2) - r
      eq * eq
    }

    # gradientFunction <- function(x, bcd1, bcd2, R) {
    #
    # }

    out <- nlminb(
      start = r, objective = objective_function,
      scale = 10, control = list(trace = 0),
      b1 = b1, c1 = c1, d1 = d1, b2 = b2, c2 = c2, d2 = d2, r = r
    )
    if (out$convergence != 0 || out$objective > 1e-5)
      lav_msg_warn(gettext("no convergence"))
    rho <- out$par[1L]
    rho
  }

  # number of variables
  nvar <- ncol(cor_1)
  # check skewness
  if (is.null(skewness)) {
    sk <- rep(0, nvar)
  } else if (length(skewness) == nvar) {
    sk <- skewness
  } else if (length(skewness) == 1L) {
    sk <- rep(skewness, nvar)
  } else {
    lav_msg_stop(gettext("skewness has wrong length"))
  }

  if (is.null(kurtosis)) {
    ku <- rep(0, nvar)
  } else if (length(kurtosis) == nvar) {
    ku <- kurtosis
  } else if (length(kurtosis) == 1L) {
    ku <- rep(kurtosis, nvar)
  } else {
    lav_msg_stop(gettext("kurtosis has wrong length"))
  }

  # create Fleishman table
  ftable_1 <- matrix(0, nvar, 4L)
  for (i in 1:nvar) {
    ftable_1[i, ] <- fleishman1978_abcd(skewness = sk[i], kurtosis = ku[i])
  }

  # compute intermediate correlations between all pairs
  icor <- diag(nvar)
  for (j in 1:(nvar - 1L)) {
    for (i in (j + 1):nvar) {
      if (cor_1[i, j] == 0) next
      icor[i, j] <- icor[j, i] <-
        get_icov(ftable_1[i, 2], ftable_1[i, 3], ftable_1[i, 4],
          ftable_1[j, 2], ftable_1[j, 3], ftable_1[j, 4],
          r = cor_1[i, j]
        )
    }
  }

  if (lav_debug()) {
    cat("\nOriginal correlations (for Vale-Maurelli):\n")
    print(cor_1)
    cat("\nIntermediate correlations (for Vale-Maurelli):\n")
    print(icor)
    cat("\nEigen values ICOR:\n")
    print(eigen(icor)$values)
  }

  # generate Z (using sign-invariant method for cross-machine reproducibility)
  if (mass) {
    x <- z <- MASS::mvrnorm(n = n, mu = rep(0, nvar), Sigma = icor)
  } else {
    x <- z <- lav_mvrnorm(n = n, mu = rep(0, nvar), sigma_1 = icor)
  }

  # transform Z using Fleishman constants
  for (i in 1:nvar) {
    x[, i] <- ftable_1[i, 1L] + ftable_1[i, 2L] * z[, i] +
      ftable_1[i, 3L] * z[, i] * z[, i] +
      ftable_1[i, 4L] * z[, i] * z[, i] * z[, i]
  }

  x
}

# constructor for the 'lavSampleStats' class
#
# initial version: YR 25/03/2009
# major revision: YR 5/11/2011: separate data.obs and sample statistics
# YR 5/01/2016: add rescov, resvar, ... if conditional.x = TRUE

# YR 18 Jan 2021: use lavoptions

# helper: estimate the h1 (saturated) moments via the EM algorithm for a
# single group; if auxiliary variables are provided, they are included in the
# EM run to improve the moments under the missing-at-random (MAR) assumption,
# after which only the submatrix that corresponds to the model variables is
# kept (the auxiliary variables are dropped). When auxiliary variables are
# used, the saturated logl is recomputed over the model variables (at the
# aux-improved moments), so that it stays consistent with the reported moments.
lav_samp_mi_aux_moments <- function(y = NULL, aux = NULL, wt = NULL,
                                    mp = NULL, yp = NULL, nobs = NULL,
                                    max_iter = 500L, tol = 1e-05,
                                    non_pd_action = "warn",
                                    non_pd_tol = 1e-05,
                                    acceleration = "none") {
  has_aux <- !is.null(aux) && NCOL(aux) > 0L
  if (has_aux) {
    out <- lav_mvn_mi_h1_est_moments(
      y = cbind(y, aux), wt = wt,
      max_iter = max_iter, tol = tol,
      non_pd_action = non_pd_action, non_pd_tol = non_pd_tol,
      acceleration = acceleration
    )
    p_ov <- seq_len(NCOL(y))
    sigma <- out$Sigma[p_ov, p_ov, drop = FALSE]
    mu <- out$Mu[p_ov]
    n_h1 <- nobs
    if (length(mp$empty.idx) > 0L) {
      if (!is.null(wt)) {
        n_h1 <- n_h1 - sum(wt[mp$empty.idx])
      } else {
        n_h1 <- n_h1 - length(mp$empty.idx)
      }
    }
    h1 <- lav_mvn_mi_loglik_samp(
      yp = yp, mu = mu, sigma_1 = sigma,
      log2pi = FALSE, minus_two = TRUE
    ) / n_h1
    # keep the full augmented moments: needed for the two-stage standard
    # errors with auxiliary variables (Savalei & Bentler, 2009)
    list(sigma = sigma, mu = mu, h1 = h1,
         sigma.aug = out$Sigma, mu.aug = out$Mu)
  } else {
    out <- lav_mvn_mi_h1_est_moments(
      y = y, wt = wt, mp = mp, yp = yp,
      max_iter = max_iter, tol = tol,
      non_pd_action = non_pd_action, non_pd_tol = non_pd_tol,
      acceleration = acceleration
    )
    list(sigma = out$Sigma, mu = out$Mu, h1 = out$fx)
  }
}

lav_samp_from_data <- function(lavdata = NULL,        # nolint start
                                      lavoptions = NULL,
                                      wls_v = NULL,
                                      nacov = NULL) {        # nolint end
  # extra info from lavoptions
  stopifnot(!is.null(lavoptions))
  missing <- lavoptions$missing
  rescale <- lavoptions$sample.cov.rescale
  estimator <- lavoptions$estimator
  # mimic <- lavoptions$mimic
  meanstructure <- lavoptions$meanstructure
  correlation <- lavoptions$correlation
  # partial correlation structure: only these observed variables are scaled
  # to unit variance (empty -> all of them, the usual correlation structure)
  correlation_ov <- lavoptions$.correlation.ov
  if (is.null(correlation_ov)) {
    correlation_ov <- character(0L)
  }
  conditional_x <- lavoptions$conditional.x
  fixed_x <- lavoptions$fixed.x
  group_w_free <- lavoptions$group.w.free
  se <- lavoptions$se
  test <- lavoptions$test
  ridge <- lavoptions$ridge
  zero_add <- lavoptions$zero.add
  zero_keep_margins <- lavoptions$zero.keep.margins
  zero_cell_warn <- lavoptions$zero.cell.warn
  allow_empty_cell <- lavoptions$allow.empty.cell
  dls_a <- lavoptions$estimator.args$dls.a
  dls_gamma_nt <- lavoptions$estimator.args$dls.GammaNT

  # sample.icov (new in 0.6-9; ensure it exists, for older objects)
  sample_icov <- TRUE
  if (!is.null(lavoptions$sample.icov)) {
    sample_icov <- lavoptions$sample.icov
  }

  # ridge default
  if (ridge) {
    if (is.numeric(lavoptions$ridge.constant)) {
      ridge_eps <- lavoptions$ridge.constant
    } else {
      ridge_eps <- 1e-5
    }
  } else {
    ridge_eps <- 0.0
  }

  # check lavdata
  stopifnot(!is.null(lavdata))

  # the Gamma (NACOV) recipe: the single source of truth for which Gamma
  # this analysis uses, shared with the post-fit lav_object_gamma()
  gamma_recipe <- lav_gamma_recipe(lavoptions = lavoptions, lavdata = lavdata)
  # design vs frequency weighting of the (categorical/continuous) Gamma
  swt_type <- gamma_recipe$swt.type

  # lavdata slots (FIXME: keep lavdata@ names)
  x <- lavdata@X
  mp <- lavdata@Mp
  ngroups <- lavdata@ngroups
  nlevels <- lavdata@nlevels
  nobs <- lavdata@nobs
  ov_names <- lavdata@ov.names
  ov_names_x <- lavdata@ov.names.x
  data_ov <- lavdata@ov
  exo <- lavdata@eXo
  aux <- lavdata@aux
  wt <- lavdata@weights

  # new in 0.6-6
  # if sampling weights have been used, redefine nobs:
  # per group, we define nobs == sum(wt)
  for (g in seq_len(ngroups)) {
    if (!is.null(wt[[g]])) {
      nobs[[g]] <- sum(wt[[g]])
    }
  }

  # sample.cov.robust cannot be used if sampling weights are used
  if (lavoptions$sample.cov.robust) {
    if (!is.null(wt[[1]])) {
      lav_msg_stop(gettext(
        "sample.cov.robust = TRUE does not work (yet)
        if sampling weights are provided."))
    }
  }

  # sample statistics per group

  # joint (y,x)
  cov <- vector("list", length = ngroups)
  var <- vector("list", length = ngroups)
  mean <- vector("list", length = ngroups)
  th <- vector("list", length = ngroups)
  th_idx <- vector("list", length = ngroups)
  th_names <- vector("list", length = ngroups)

  # residual (y | x)
  res_cov <- vector("list", length = ngroups)
  res_var <- vector("list", length = ngroups)
  res_th <- vector("list", length = ngroups)
  res_th_nox <- vector("list", length = ngroups)
  res_slopes <- vector("list", length = ngroups)
  res_int <- vector("list", length = ngroups)

  # fixed.x
  mean_x <- vector("list", length = ngroups)
  cov_x <- vector("list", length = ngroups)

  # binary/ordinal
  bifreq <- vector("list", length = ngroups)

  # extra sample statistics per group
  icov <- vector("list", length = ngroups)
  cov_log_det <- vector("list", length = ngroups)
  res_icov <- vector("list", length = ngroups)
  res_cov_log_det <- vector("list", length = ngroups)
  wls_obs <- vector("list", length = ngroups)
  missing_1 <- vector("list", length = ngroups)
  missing_h1 <- vector("list", length = ngroups)
  missing_flag <- FALSE
  zero_cell_tables <- vector("list", length = ngroups)
  ylp <- vector("list", length = ngroups)

  # group weights
  group_w <- vector("list", length = ngroups)

  # convenience? # FIXME!
  x_idx <- vector("list", length = ngroups)


  wls_vd <- vector("list", length = ngroups)
  if (is.null(wls_v)) {
    wls_v <- vector("list", length = ngroups)
    wls_v_user <- FALSE
  } else {
    if (!is.list(wls_v)) {
      if (ngroups == 1L) {
        wls_v <- list(wls_v)
      } else {
        lav_msg_stop(gettextf(
          "wls_v argument should be a list of length %s", ngroups)
        )
      }
    } else {
      if (length(wls_v) != ngroups) {
        lav_msg_stop(gettextf(
          "wls_v assumes %1$s groups; data contains %2$s groups",
          length(wls_v), ngroups))
      }
    }

    # is wls_v full? check first
    if (is.null(dim(wls_v[[1]]))) {
      # we will assume it is the diagonal only
      wls_vd <- wls_v
      wls_v <- lapply(wls_vd, diag)
    } else {
      # create WLS.VD
      wls_vd <- lapply(wls_v, diag)
    }

    wls_v_user <- TRUE
    # FIXME: check dimension of wls_v!!
  }

  # a user-provided WLS.V is not used by the GLS estimation machinery
  # (since 0.6-10 the objective/gradient are direct functions of the
  # sample moments); ignore it altogether (and warn), rather than using
  # it inconsistently
  if (wls_v_user && estimator == "GLS") {
    lav_msg_warn(gettext(
      "WLS.V is ignored when estimator = \"GLS\"; use estimator = \"WLS\" to
      estimate with a user-provided weight matrix."))
    wls_v <- vector("list", length = ngroups)
    wls_vd <- vector("list", length = ngroups)
    wls_v_user <- FALSE
  }

  nacov_compute <- FALSE # since 0.6-6
  if (is.null(nacov)) {
    nacov <- vector("list", length = ngroups)
    nacov_user <- FALSE
    # default policy (se/test/missing/estimator triggers): see
    # lav_gamma_recipe() in lav_samplestats_gamma_recipe.R
    nacov_compute <- gamma_recipe$compute
  } else if (is.logical(nacov)) {
    if (!nacov) {
      nacov_compute <- FALSE
    } else {
      nacov_compute <- TRUE
    }
    nacov_user <- FALSE
    nacov <- vector("list", length = ngroups)
  } else {
    if (!is.list(nacov)) {
      if (ngroups == 1L) {
        nacov <- list(nacov)
      } else {
        lav_msg_stop(gettextf(
          "nacov= argument should be a list of length ", ngroups))
      }
    } else {
      if (length(nacov) != ngroups) {
        lav_msg_stop(gettextf(
          "nacov= assumes %1$s groups; data contains %2$s groups",
          length(nacov), ngroups))
      }
    }
    nacov_user <- TRUE
    # FIXME: check dimension of nacov!!
  }



  # compute some sample statistics per group
  for (g in 1:ngroups) {
    # switch off computing all sample statistics? (housekeeping only)
    if (!is.null(lavoptions$samplestats) && !lavoptions$samplestats) {
    next
  }
    # check nobs
    if (is.null(wt[[g]])) {
      if (nobs[[g]] < 2L) {
        if (nobs[[g]] == 0L) {
          if (ngroups > 1L) {
            lav_msg_stop(gettextf("data contains no observations in
                                   group %s", g))
          } else {
            lav_msg_stop(gettext("data contains no observations"))
          }
        } else if (!allow_empty_cell) {
          if (ngroups > 1L) {
            lav_msg_stop(gettextf("data contains only a single observation
                                   in group %s", g))
          } else {
            lav_msg_stop(gettext("data contains only a single observation"))
          }
        }
      }
    }

    # exogenous x?
    nexo <- length(ov_names_x[[g]])
    if (nexo) {
      stopifnot(nexo == NCOL(exo[[g]]))

      # two cases: ov.names contains 'x' variables, or not
      if (conditional_x) {
        # ov.names.x are NOT in ov.names
        x_idx[[g]] <- length(ov_names[[g]]) + seq_len(nexo)
      } else {
        if (fixed_x) {
          # ov.names.x are a subset of ov.names
          x_idx[[g]] <- match(ov_names_x[[g]], ov_names[[g]])
          stopifnot(!anyNA(x_idx[[g]]))
        } else {
          x_idx[[g]] <- integer(0L)
        }
      }
    } else {
      x_idx[[g]] <- integer(0L)
      conditional_x <- FALSE
      fixed_x <- FALSE
    }

    # group weight
    group_w[[g]] <- nobs[[g]] / sum(unlist(nobs))

    # check if we have categorical data in this group
    categorical <- FALSE
    ov_types <- data_ov$type[match(ov_names[[g]], data_ov$name)]
    ov_levels <- data_ov$nlev[match(ov_names[[g]], data_ov$name)]
    cat_1 <- list()
    if ("ordered" %in% ov_types) {
      categorical <- TRUE
      if (nlevels > 1L &&
        !lavoptions$estimator %in% c("WLS", "DWLS", "ULS")) {
        lav_msg_warn(gettext("multilevel + categorical not supported yet."))
      }
    }

    if (categorical) {
      # compute CAT

      if (estimator %in% c("ML", "REML", "PML", "FML", "MML", "none", "IV",
                           "ULS")) {
        wls_w <- FALSE
        if (estimator == "ULS" &&
            se %in% c("robust.sem", "robust.sem.nt", "robust.cluster.sem")) {
          wls_w <- TRUE
        } else if (estimator == "IV" &&
                  lavoptions$estimator.args$iv_vcov_stage1 == "gamma") {
          wls_w <- TRUE
        }
      } else {
        wls_w <- TRUE
      }
    # check cat.wls.w option (new in 0.6-18)
    if (!is.null(lavoptions$cat.wls.w) && !lavoptions$cat.wls.w) {
      wls_w <- FALSE # perhaps do.fit = FALSE? (eg sam())
    }
    # two-level categorical: the (single-level, flat) statistics computed
    # here only provide starting values; the real two-level statistics
    # (and their Gamma) come from the stage-wise estimation later
    if (nlevels > 1L) {
      wls_w <- FALSE
    }
      if (lav_verbose()) {
        cat("Estimating sample thresholds and correlations ... ")
      }

      # cluster-robust SEs/test for categorical data (issue #254): pass the
      # cluster index so muthen1984() can also return a cluster-robust WLS.W
      if (length(lavdata@cluster) > 0L && nlevels == 1L) {
        cluster_idx_cat <- lavdata@Lp[[g]]$cluster.idx[[2]]
      } else {
        cluster_idx_cat <- NULL
      }

      current_verbose <- lav_verbose()
      if (lav_verbose(lav_debug()))
        on.exit(lav_verbose(current_verbose), TRUE)
      # multilevel: X contains ALL ('tilde') variables (y + x); the flat
      # (single-level) muthen1984 run below -- starting values only --
      # needs the y columns only
      data_1_cat <- x[[g]]
      if (nlevels > 1L && conditional_x) {
        data_1_cat <- x[[g]][, seq_along(ov_names[[g]]), drop = FALSE]
      }
      if (conditional_x) {
        cat_1 <- muthen1984(
          data_1 = data_1_cat,
          wt = wt[[g]],
          sampling_weights_type = swt_type,
          ov_names = ov_names[[g]],
          ov_types = ov_types,
          ov_levels = ov_levels,
          ov_names_x = ov_names_x[[g]],
          exo = exo[[g]],
          group = g, # for error messages only
          wls_w = wls_w,
          zero_add = zero_add,
          zero_keep_margins = zero_keep_margins,
          zero_cell_warn = FALSE,
          zero_cell_tables = TRUE,
          allow_empty_cell = allow_empty_cell,
          cluster_idx = cluster_idx_cat
        )
      } else {
        cat_1 <- muthen1984(
          data_1 = x[[g]],
          wt = wt[[g]],
          sampling_weights_type = swt_type,
          ov_names = ov_names[[g]],
          ov_types = ov_types,
          ov_levels = ov_levels,
          ov_names_x = NULL,
          exo = NULL,
          group = g, # for error messages only
          wls_w = wls_w,
          zero_add = zero_add,
          zero_keep_margins = zero_keep_margins,
          zero_cell_warn = FALSE,
          zero_cell_tables = TRUE,
          allow_empty_cell = allow_empty_cell,
          cluster_idx = cluster_idx_cat
        )
      }
      lav_verbose(current_verbose)
      # empty cell tables
      zero_cell_tables[[g]] <- cat_1$zero.cell.tables
      if (lav_verbose()) cat("done\n")
    }

    if (categorical) {
      # convenience
      th_idx[[g]] <- unlist(cat_1$TH.IDX)
      th_names[[g]] <- unlist(cat_1$TH.NAMES)

      if (conditional_x) {
        # residual var/cov
        res_var[[g]] <- unlist(cat_1$VAR)
        res_cov[[g]] <- unname(cat_1$COV)
        if (ridge) {
          diag(res_cov[[g]]) <- diag(res_cov[[g]]) + ridge_eps
          res_var[[g]] <- diag(res_cov[[g]])
        }

        # th also contains the means of numeric variables
        res_th[[g]] <- unlist(cat_1$TH)
        res_th_nox[[g]] <- unlist(cat_1$TH.NOX)

        # for convenience, we store the intercept of numeric
        # variables in res.int
        nvar_1 <- NCOL(res_cov[[g]])
        mean[[g]] <- res_int[[g]] <- numeric(nvar_1)
        num_idx <- which(!seq_len(nvar_1) %in% th_idx[[g]])
        if (length(num_idx) > 0L) {
          num_idx_1 <- which(th_idx[[g]] == 0L)
          mean[[g]][num_idx] <- res_th_nox[[g]][num_idx_1]
          res_int[[g]][num_idx] <- res_th[[g]][num_idx_1]
        }

        # slopes
        res_slopes[[g]] <- cat_1$SLOPES
      } else {
        # var/cov
        var[[g]] <- unlist(cat_1$VAR)
        cov[[g]] <- unname(cat_1$COV)
        if (ridge) {
          diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          var[[g]] <- diag(cov[[g]])
        }

        # th also contains the means of numeric variables
        th[[g]] <- unlist(cat_1$TH)

        # mean (numeric only)
        nvar_1 <- NCOL(cov[[g]])
        mean[[g]] <- numeric(nvar_1)
        num_idx <- which(!seq_len(nvar_1) %in% th_idx[[g]])
        if (length(num_idx) > 0L) {
          num_idx_1 <- which(th_idx[[g]] == 0L)
          mean[[g]][num_idx] <- th[[g]][num_idx_1]
        }
      }

      # only for catML
      if (estimator == "catML") {
        cov_1 <- cov2cor(lav_mat_sym_force_pd(cov[[g]],
          tol = 1e-04
        ))
        # overwrite
        cov[[g]] <- cov_1
        out <- lav_samp_icov(
          cov_1 = cov_1,
          x_idx = x_idx[[g]],
          ngroups = ngroups, g = g
        )
        icov[[g]] <- out$icov
        cov_log_det[[g]] <- out$cov.log.det

        # the same for res.cov if conditional.x = TRUE
        if (conditional_x) {
          res_cov_1 <-
            cov2cor(lav_mat_sym_force_pd(res_cov[[g]],
              tol = 1e-04
            ))
          # overwrite
          res_cov[[g]] <- res_cov_1
          out <- lav_samp_icov(
            cov_1 = res_cov_1,
            ridge = 1e-05,
            x_idx = x_idx[[g]],
            ngroups = ngroups, g = g
          )
          res_icov[[g]] <- out$icov
          res_cov_log_det[[g]] <- out$cov.log.det
        }
      }
    # categorical

    # continuous -- multilevel
    } else if (nlevels > 1L) {
      # level-based sample statistics
      ylp[[g]] <- lav_samp_cl_patterns(
        y = x[[g]],
        lp = lavdata@Lp[[g]],
        conditional_x = lavoptions$conditional.x
      )


      if (conditional_x) {
        # for starting values only
        # no handling of missing data yet....
        if (missing %in% c(
          "ml", "ml.x",
          "two.stage", "robust.two.stage"
        )) {
          lav_msg_stop(gettextf(
            "missing = %s + conditional.x + two.level not supported yet",
            missing))
        }

        # residual covariances!
        y <- x[[g]] # contains eXo
        cov_1 <- unname(stats::cov(y, use = "pairwise.complete.obs"))
    # if we have missing values (missing by design?), replace them by 0
    cov_1[is.na(cov_1)] <- 0
        mean_1 <- unname(colMeans(y, na.rm = TRUE))
        var[[g]] <- diag(cov_1)
        # rescale cov by (N-1)/N? (only COV!)
        if (rescale) {
          # we 'transform' the sample cov (divided by n-1)
          # to a sample cov divided by 'n'
          cov_1 <- ((nobs[[g]] - 1) / nobs[[g]]) * cov_1
        }
        cov[[g]] <- cov_1
        if (ridge) {
          diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          var[[g]] <- diag(cov[[g]])
        }
        mean[[g]] <- mean_1

        a_1 <- cov_1[-x_idx[[g]], -x_idx[[g]], drop = FALSE]
        m_b <- cov_1[-x_idx[[g]], x_idx[[g]], drop = FALSE]
        c_1 <- cov_1[x_idx[[g]], x_idx[[g]], drop = FALSE]
        c_inv <- try(solve(c_1), silent = TRUE)
        if (inherits(c_inv, "try-error")) {
          lav_msg_warn(gettext(
            "Exogenous covariance matrix is singular; using generalized inverse."
          ))
          c_inv <- MASS::ginv(c_1)
        }
        res_cov[[g]] <- a_1 - m_b %*% c_inv %*% t(m_b)
        res_var[[g]] <- diag(cov[[g]])

        my <- mean_1[-x_idx[[g]]]
        mx <- mean_1[x_idx[[g]]]
        c3 <- rbind(
          c(1, mx),
          cbind(mx, c_1 + tcrossprod(mx))
        )
        b3 <- cbind(my, m_b + tcrossprod(my, mx))
        coef_1 <- try(unname(solve(c3, t(b3))), silent = TRUE)
        if (inherits(coef_1, "try-error")) {
          lav_msg_warn(gettext(
            "Augmented exogenous matrix is singular; using generalized inverse."
          ))
          coef_1 <- unname(MASS::ginv(c3) %*% t(b3))
        }

        res_int[[g]] <- coef_1[1, ] # intercepts
        res_slopes[[g]] <- t(coef_1[-1, , drop = FALSE]) # slopes
      } else {
        # FIXME: needed?
    cov_1 <- unname(stats::cov(x[[g]], use = "pairwise.complete.obs"))
        # if we have missing values (missing by design?), replace them by 0
        cov_1[is.na(cov_1)] <- 0
        cov[[g]] <- cov_1
        mean[[g]] <- unname(colMeans(x[[g]], na.rm = TRUE))
        var[[g]] <- diag(cov[[g]])

        # missing patterns
        if (missing %in% c("ml", "ml.x")) {
          missing_flag <- TRUE
          missing_1[[g]] <-
            lav_samp_mi_patterns(
              y = x[[g]],
              mp = mp[[g]],
              wt = wt[[g]],
              lp = lavdata@Lp[[g]]
            )
        }
      }
    } # multilevel

    # continuous -- single-level
    else {
      if (conditional_x) {
        # FIXME!
        # no correlation structures yet
        if (correlation) {
          lav_msg_stop(gettext(
            "conditional.x = TRUE is not supported (yet) for
            correlation structures."))
        }

        # missing values in the exogenous x variables cannot be handled
        # when we condition on x
        if (missing == "ml.x") {
          lav_msg_stop(gettext(
            "missing = \"ml.x\" is not supported when conditional.x = TRUE
            (missing values in the exogenous variables); use missing = \"ml\"
            (listwise deletion of cases with missing x values), or
            conditional.x = FALSE."))
        }

        # residual covariances!
        # (pairwise-complete sample statistics; if missing = \"ml\" or
        #  two.stage, these serve as starting values only, and the
        #  EM-based statistics are computed below)

        y <- cbind(x[[g]], exo[[g]])
    cov_1 <- unname(stats::cov(y, use = "pairwise.complete.obs"))
        # if we have missing values (missing by design?), replace them by 0
        cov_1[is.na(cov_1)] <- 0
        mean_1 <- unname(colMeans(y, na.rm = TRUE))
        # rescale cov by (N-1)/N? (only COV!)
        if (rescale) {
          # we 'transform' the sample cov (divided by n-1)
          # to a sample cov divided by 'n'
          cov_1 <- ((nobs[[g]] - 1) / nobs[[g]]) * cov_1
        }
        cov[[g]] <- cov_1
        var[[g]] <- diag(cov_1)
        if (ridge) {
          diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          var[[g]] <- diag(cov[[g]])
        }
        mean[[g]] <- mean_1

        a_1 <- cov_1[-x_idx[[g]], -x_idx[[g]], drop = FALSE]
        m_b <- cov_1[-x_idx[[g]], x_idx[[g]], drop = FALSE]
        c_1 <- cov_1[x_idx[[g]], x_idx[[g]], drop = FALSE]
        # FIXME: make robust against singular c_1!!!
        res_cov[[g]] <- a_1 - m_b %*% solve(c_1) %*% t(m_b)
        res_var[[g]] <- diag(cov[[g]])


        my <- mean_1[-x_idx[[g]]]
        mx <- mean_1[x_idx[[g]]]
        c3 <- rbind(
          c(1, mx),
          cbind(mx, c_1 + tcrossprod(mx))
        )
        b3 <- cbind(my, m_b + tcrossprod(my, mx))
        coef_1 <- unname(solve(c3, t(b3)))

        res_int[[g]] <- coef_1[1, ] # intercepts
        res_slopes[[g]] <- t(coef_1[-1, , drop = FALSE]) # slopes

        # missing data: FIML or two-stage
        if (missing %in% c("ml", "two.stage", "robust.two.stage")) {
          if (missing == "ml") {
            missing_flag <- TRUE
          } else {
            # two-stage: the sample statistics ARE the EM estimates
            missing_flag <- FALSE
            if (estimator %in% c("ULS", "GLS", "WLS", "DLS")) {
              lav_msg_stop(gettextf(
                "missing = %1$s + conditional.x is not supported (yet) for
                estimator %2$s; use estimator = \"ML\", or
                conditional.x = FALSE.", dQuote(missing), dQuote(estimator)))
            }
          }

          # extended pattern statistics: the usual per-pattern moments,
          # plus the crossproducts with the (complete) exogenous variables
          missing_1[[g]] <- lav_samp_mi_patterns_res(
            y = x[[g]], exo = exo[[g]], mp = mp[[g]], wt = wt[[g]]
          )

          # estimate the joint (y, x) moments of the unrestricted model
          # via EM; because x is complete, the joint normal likelihood
          # factorizes as f(y|x) * f(x) with variation-independent
          # parameter blocks, and the saturated moments of the conditional
          # model are obtained by partitioning the joint EM moments
          current_warn <- lav_warn()
          if (lav_warn(lavoptions$em.h1.args$warn))
            on.exit(lav_warn(current_warn), TRUE)
          # zero coverage?
          if (any(lav_mat_vech(mp[[g]]$coverage, diagonal = FALSE) == 0)) {
            # no EM estimates (h1 will be NULL)
          } else if (!allow_empty_cell) {
            # optionally augment the EM run with auxiliary variables to
            # improve the h1 moments under MAR (as in the unconditional
            # case; the auxiliary variables are dropped afterwards)
            aux_g <- if (length(aux) >= g) aux[[g]] else NULL
            has_aux <- !is.null(aux_g) && NCOL(aux_g) > 0L
            y2 <- cbind(x[[g]], exo[[g]])
            em_y <- if (has_aux) cbind(y2, aux_g) else y2
            em_out <- lav_mvn_mi_h1_est_moments(
              y = em_y, wt = wt[[g]],
              max_iter = lavoptions$em.h1.args$max_iter,
              tol = lavoptions$em.h1.args$tol,
              non_pd_action = lavoptions$em.h1.args$non_pd_action,
              non_pd_tol = lavoptions$em.h1.args$non_pd_tol,
              acceleration = lavoptions$em.h1.args$acceleration
            )
            p_ov <- seq_len(NCOL(y2))
            sigma_em <- em_out$Sigma[p_ov, p_ov, drop = FALSE]
            mu_em <- em_out$Mu[p_ov]

            # partition the joint moments into the residual moments of
            # the saturated conditional model (same algebra as above)
            a_em <- sigma_em[-x_idx[[g]], -x_idx[[g]], drop = FALSE]
            b_em <- sigma_em[-x_idx[[g]], x_idx[[g]], drop = FALSE]
            c_em <- sigma_em[x_idx[[g]], x_idx[[g]], drop = FALSE]
            res_cov_em <- a_em - b_em %*% solve(c_em) %*% t(b_em)
            my_em <- mu_em[-x_idx[[g]]]
            mx_em <- mu_em[x_idx[[g]]]
            c3_em <- rbind(
              c(1, mx_em),
              cbind(mx_em, c_em + tcrossprod(mx_em))
            )
            b3_em <- cbind(my_em, b_em + tcrossprod(my_em, mx_em))
            coef_em <- unname(solve(c3_em, t(b3_em)))
            res_int_em <- coef_em[1, ]
            res_slopes_em <- t(coef_em[-1, , drop = FALSE])

            # the saturated value of the conditional (-2/N) loglikelihood
            h1_em <- lav_mvreg_mi_loglik_samp(
              yp = missing_1[[g]],
              res_int = res_int_em, res_slopes = res_slopes_em,
              res_cov = res_cov_em,
              log2pi = FALSE, minus_two = TRUE
            ) / nobs[[g]]

            missing_h1[[g]] <- list(
              sigma = sigma_em, mu = mu_em, h1 = h1_em,
              res.cov = res_cov_em, res.int = res_int_em,
              res.slopes = res_slopes_em
            )
            if (has_aux) {
              # keep the full augmented moments (see the unconditional case)
              missing_h1[[g]]$sigma.aug <- em_out$Sigma
              missing_h1[[g]]$mu.aug <- em_out$Mu
            }

            # for two-stage (or FIML with sampling weights), replace the
            # sample statistics by the EM estimates
            if (!missing_flag || !is.null(wt[[g]])) {
              cov[[g]] <- sigma_em
              if (ridge) {
                diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
              }
              var[[g]] <- diag(cov[[g]])
              mean[[g]] <- mu_em
              res_cov[[g]] <- res_cov_em
              res_var[[g]] <- diag(cov[[g]])
              res_int[[g]] <- res_int_em
              res_slopes[[g]] <- res_slopes_em
            }
          }
          lav_warn(current_warn)
        }
      } else if (missing == "two.stage" ||
        missing == "robust.two.stage") {
        missing_flag <- FALSE # !!! just use sample statistics
        missing_1[[g]] <-
          lav_samp_mi_patterns(
            y = x[[g]],
            mp = mp[[g]],
            wt = wt[[g]]
          )
        current_warn <- lav_warn()
        if (lav_warn(lavoptions$em.h1.args$warn))
          on.exit(lav_warn(current_warn), TRUE)
        # optionally augment the EM run with auxiliary variables to improve
        # the moments under MAR (see lav_samp_mi_aux_moments); for two-stage
        # estimation, the sample statistics ARE the EM estimates, so the
        # auxiliary variables affect the model estimates as well
        aux_g <- if (length(aux) >= g) aux[[g]] else NULL
        missing_h1[[g]] <- lav_samp_mi_aux_moments(
          y = x[[g]], aux = aux_g, wt = wt[[g]],
          mp = mp[[g]], yp = missing_1[[g]], nobs = nobs[[g]],
          max_iter = lavoptions$em.h1.args$max_iter,
          tol = lavoptions$em.h1.args$tol,
          non_pd_action = lavoptions$em.h1.args$non_pd_action,
          non_pd_tol = lavoptions$em.h1.args$non_pd_tol,
          acceleration = lavoptions$em.h1.args$acceleration
        )
        lav_warn(current_warn)

        # here, sample statistics == EM estimates
        cov[[g]] <- missing_h1[[g]]$sigma
        if (ridge) {
          diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
        }
        var[[g]] <- diag(cov[[g]])
        mean[[g]] <- missing_h1[[g]]$mu
      } else if (missing %in% c("ml", "ml.x")) {
        missing_flag <- TRUE
        missing_1[[g]] <-
          lav_samp_mi_patterns(
            y = x[[g]],
            mp = mp[[g]],
            wt = wt[[g]]
          )

        if (nlevels == 1L) {
          # estimate moments unrestricted model
          current_warn <- lav_warn()
          if (lav_warn(lavoptions$em.h1.args$warn))
            on.exit(lav_warn(current_warn), TRUE)
          # zero coverage?
          if (any(lav_mat_vech(mp[[g]]$coverage, diagonal = FALSE) == 0)) {
            #out <- lav_mvn_mi_h1_est_moments_chol(
            #         lavdata = lavdata, lavoptions = lavoptions, group = g)
            #missing.h1.[[g]]$sigma <- out$Sigma
            #missing.h1.[[g]]$mu <- out$Mu
            #missing.h1.[[g]]$h1 <- out$fx
          } else if (!allow_empty_cell) {
            # optionally augment the EM run with auxiliary variables to
            # improve the h1 moments under MAR (see lav_samp_mi_aux_moments)
            aux_g <- if (length(aux) >= g) aux[[g]] else NULL
            missing_h1[[g]] <- lav_samp_mi_aux_moments(
              y = x[[g]], aux = aux_g, wt = wt[[g]],
              mp = mp[[g]], yp = missing_1[[g]], nobs = nobs[[g]],
              max_iter = lavoptions$em.h1.args$max_iter,
              tol = lavoptions$em.h1.args$tol,
              non_pd_action = lavoptions$em.h1.args$non_pd_action,
              non_pd_tol = lavoptions$em.h1.args$non_pd_tol,
              acceleration = lavoptions$em.h1.args$acceleration
            )
          }
          lav_warn(current_warn)
        }

        if (!is.null(wt[[g]])) {
          # here, sample statistics == EM estimates
          cov[[g]] <- missing_h1[[g]]$sigma
          if (ridge) {
            diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          }
          var[[g]] <- diag(cov[[g]])
          mean[[g]] <- missing_h1[[g]]$mu
        } else {
          # NEEDED? why not just EM-based?
          cov_1 <- unname(stats::cov(x[[g]], use = "pairwise.complete.obs"))
          # if we have missing values (missing by design?), replace them by 0
      cov_1[is.na(cov_1)] <- 0
          cov[[g]] <- cov_1
          # rescale cov by (N-1)/N? (only COV!)
          if (rescale) {
            # we 'transform' the sample cov (divided by n-1)
            # to a sample cov divided by 'n'
            cov[[g]] <- ((nobs[[g]] - 1) / nobs[[g]]) * cov[[g]]
          }
          if (ridge) {
            diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          }
          var[[g]] <- diag(cov[[g]])
          mean[[g]] <- colMeans(x[[g]], na.rm = TRUE)
        }
      } else {
        # LISTWISE
        if (!is.null(wt[[g]])) {
          out <- stats::cov.wt(x[[g]],
            wt = wt[[g]],
            method = "ML"
          )
      cov_1 <- out$cov
          # if we have missing values (missing by design?), replace them by 0
          cov_1[is.na(cov_1)] <- 0
          cov[[g]] <- cov_1
          # cov.wt(method = "ML") divides by sum(wt) (the 'N' version). If
          # rescale is FALSE (e.g. GLS/ULS/(D)WLS), the unweighted path uses
          # the unbiased 'N-1' version; mirror that here so that supplying
          # sampling weights does not change the covariance normalization
          # (nobs[[g]] == sum(wt[[g]]); see top of this function).
          if (!rescale) {
            cov[[g]] <- (nobs[[g]] / (nobs[[g]] - 1)) * cov[[g]]
          }
          if (ridge) {
            diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          }
          var[[g]] <- diag(cov[[g]])
          mean[[g]] <- out$center
        } else if (lavoptions$sample.cov.robust) {
          # fixme: allow prob/max.it to be options
          out <- lav_cov_huber(
            y = x[[g]], prob = 0.95,
            max_it = 200L, tol = 1e-07
          )
          cov[[g]] <- out$Sigma
          var[[g]] <- diag(cov[[g]])
          mean[[g]] <- out$Mu
        } else {
      cov_1 <- unname(stats::cov(x[[g]], use = "pairwise.complete.obs"))
          # if we have missing values (missing by design?), replace them by 0
          cov_1[is.na(cov_1)] <- 0
          cov[[g]] <- cov_1
          # rescale cov by (N-1)/N? (only COV!)
          if (rescale) {
            # we 'transform' the sample cov (divided by n-1)
            # to a sample cov divided by 'n'
            cov[[g]] <- ((nobs[[g]] - 1) / nobs[[g]]) * cov[[g]]
          }
          if (ridge) {
            diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
          }
          var[[g]] <- diag(cov[[g]])
          mean[[g]] <- colMeans(x[[g]], na.rm = TRUE)
        }
      }

      # correlation structure?
      if (correlation) {
        # which variables are scaled to unit variance? (all, or a subset
        # for a 'partial' correlation structure)
        if (length(correlation_ov) > 0L) {
          cor_idx <- which(ov_names[[g]] %in% correlation_ov)
        } else {
          cor_idx <- seq_len(ncol(cov[[g]]))
        }
        cov[[g]] <- lav_cov2cor_partial(cov[[g]], cor_idx)
        var[[g]] <- diag(cov[[g]])
        if (conditional_x) {
          res_cov[[g]] <- lav_cov2cor_partial(res_cov[[g]], cor_idx)
          res_var[[g]] <- diag(res_cov[[g]])
          cov_x[[g]] <- cov2cor(cov_x[[g]])
          # FIXME: slopes? more?
        }
      }

      # icov and cov.log.det (but not if missing)
      if (sample_icov && !missing %in% c("ml", "ml.x")) {
        out <- lav_samp_icov(
          cov_1 = cov[[g]], ridge = 1e-05,
          x_idx = x_idx[[g]],
          ngroups = ngroups, g = g
        )
        icov[[g]] <- out$icov
        cov_log_det[[g]] <- out$cov.log.det

        # the same for res.cov if conditional.x = TRUE
        if (conditional_x) {
          out <- lav_samp_icov(
            cov_1 = res_cov[[g]],
            ridge = 1e-05,
            x_idx = x_idx[[g]],
            ngroups = ngroups, g = g
          )
          res_icov[[g]] <- out$icov
          res_cov_log_det[[g]] <- out$cov.log.det
        }
      }
    } # continuous - single level


    # WLS.obs
    if (nlevels == 1L) {
      if (estimator == "catML") {
        # correlations only (for now)
        tmp_categorical <- FALSE
        tmp_meanstructure <- FALSE
      } else {
        tmp_categorical <- categorical
        tmp_meanstructure <- meanstructure
      }
      wls_obs[[g]] <- lav_samp_wls_obs(
        mean_g = mean[[g]],
        cov_g = cov[[g]], var_g = var[[g]], th_g = th[[g]],
        th_idx_g = th_idx[[g]], res_int_g = res_int[[g]],
        res_cov_g = res_cov[[g]], res_var_g = res_var[[g]],
        res_th_g = res_th[[g]], res_slopes_g = res_slopes[[g]],
        group_w_g = log(nobs[[g]]),
        categorical = tmp_categorical, conditional_x = conditional_x,
        meanstructure = tmp_meanstructure, correlation = correlation,
        num_idx = if (correlation && length(correlation_ov) > 0L) {
          which(!(ov_names[[g]] %in% correlation_ov))
        } else {
          integer(0L)
        },
        slopestructure = conditional_x,
        group_w_free = group_w_free
      )
    }

    # fill in the other slots
    if (!is.null(exo[[g]])) {
      if (!is.null(wt[[g]])) {
        if (missing != "listwise") {
          cov_x[[g]] <- missing_h1[[g]]$sigma[x_idx[[g]],
            x_idx[[g]],
            drop = FALSE
          ]
          mean_x[[g]] <- missing_h1[[g]]$mu[x_idx[[g]]]
        } else {
          out <- stats::cov.wt(exo[[g]],
            wt = wt[[g]],
            method = "ML"
          )
          cov_x[[g]] <- out$cov
          mean_x[[g]] <- out$center
        }
      } else {
        cov_x[[g]] <- cov(exo[[g]], use = "pairwise")
        if (rescale) {
          # we 'transform' the sample cov (divided by n-1)
          # to a sample cov divided by 'n'
          cov_x[[g]] <- ((nobs[[g]] - 1) / nobs[[g]]) * cov_x[[g]]
        }
        mean_x[[g]] <- colMeans(exo[[g]])
      }
    }

    # nacov (=GAMMA)
    if (!nacov_user && nlevels == 1L) {
      if (nacov_compute && !categorical &&
          any(missing == c("two.stage", "robust.two.stage")) &&
          estimator %in% c("ULS", "GLS", "WLS", "DLS")) {
        # two-stage missing data: the NACOV of the (saturated) EM moments,
        # in (mean, vech(cov)) order, to be used by the robust.sem sandwich
        # and the satorra.bentler test
        # - two.stage:        I_1^{-1}            (Savalei & Bentler, 2009)
        # - robust.two.stage: I_1^{-1} J_1 I_1^{-1} (Savalei & Falk, 2014)
        mu_g <- missing_h1[[g]]$mu
        sigma_g <- missing_h1[[g]]$sigma
        x_idx_g <- if (fixed_x) x_idx[[g]] else integer(0L)
        if (missing == "robust.two.stage") {
          nacov[[g]] <- lav_mvn_mi_h1_omega_sw(
            y = x[[g]], mp = mp[[g]],
            yp = missing_1[[g]], wt = wt[[g]],
            mu = mu_g, sigma_1 = sigma_g, x_idx = x_idx_g,
            information = "observed"
          )
        } else {
          i1 <- lav_mvnorm_missing_information_observed_samplestats(
            yp = missing_1[[g]],
            mu = mu_g, sigma_1 = sigma_g, x_idx = x_idx_g
          )
          nacov[[g]] <- lav_mat_sym_inverse(i1)
        }
      } else if (estimator %in% c("ML", "GLS") &&
                 !missing_flag && nacov_compute) {
        if (conditional_x) {
          y <- y
        } else {
          y <- x[[g]]
        }

        if (length(lavdata@cluster) > 0L) {
          cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
        } else {
          cluster_idx <- NULL
        }

        if (correlation) {
          # (partial) correlation ADF Gamma via the delta method (handles
          # fixed.x); cor_idx lists the unit-variance variables (all, unless
          # a subset was requested via correlation = c(...))
          cor_idx_g <- lav_gamma_recipe_cor_idx(gamma_recipe, ov_names[[g]])
          nacov[[g]] <- lav_samp_partial_cor_gamma(
            m_y = y, cor_idx = cor_idx_g,
            meanstructure = meanstructure,
            fixed_x = fixed_x, x_idx = x_idx[[g]]
          )
    } else {
          nacov[[g]] <-
            lav_samp_gamma(
              m_y = y,
              x_idx = x_idx[[g]],
              cluster_idx = cluster_idx,
              fixed_x = fixed_x,
              conditional_x = conditional_x,
              meanstructure = meanstructure,
              slopestructure = conditional_x,
              gamma_n_minus_one = gamma_recipe$n.minus.one,
              unbiased = gamma_recipe$unbiased,
              mplus_wls = FALSE, # never for ML/GLS (see recipe)
              wt = wt[[g]],
              sampling_weights_type = swt_type
            )
        }
      } else if (estimator %in% c("WLS", "DWLS", "ULS", "DLS", "IV", "catML")) {
        # ULS (continuous): the weight matrix is the identity, so Gamma
        # is not needed for estimation; skip the (expensive) ADF/NT Gamma
        # unless a consumer requests it: robust/scaled se or test (via
        # nacov_compute), or the browne.residual.adf test (the only
        # browne variant that reads the stored NACOV; the NT and
        # model-based variants use the fast kernel or recompute at test
        # time). Note lavInspect(fit, "gamma") will then return nothing,
        # just as for estimator ML with standard se/test.
        uls_gamma_skip <- (estimator == "ULS" && !categorical &&
          !nacov_compute &&
          !("browne.residual.adf" %in% lavoptions$test))
        if (uls_gamma_skip) {
          # no Gamma needed
        } else if (!categorical) {
          # sample size large enough?
          nvar <- ncol(x[[g]])
          # if(conditional.x && nexo > 0L) {
          #    nvar <- nvar - nexo
          # }
          pstar <- nvar * (nvar + 1) / 2
          if (meanstructure) pstar <- pstar + nvar
          if (conditional_x && nexo > 0L) {
            pstar <- pstar + (nvar * nexo)
          }
          # only when the (sample-based) ADF Gamma is computed below;
          # the normal-theory Gamma has no such sample-size requirement
          gamma_nt_flavor <- (!correlation &&
            estimator %in% c("ULS", "DWLS") &&
            "robust.sem.nt" %in% lavoptions$se)
          if (nrow(x[[g]]) < pstar && estimator != "IV" &&
              !gamma_nt_flavor) {
            if (ngroups > 1L) {
              lav_msg_warn(gettextf(
              "number of observations (%s) too small to compute Gamma",
              nrow(x[[g]]), " in group %s", g))
            } else {
              lav_msg_warn(gettextf(
              "number of observations (%s) too small to compute Gamma",
              nrow(x[[g]])))
            }
          }
          if (conditional_x) {
            y <- y
          } else {
            y <- x[[g]]
          }

          if (length(lavdata@cluster) > 0L) {
            cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
          } else {
            cluster_idx <- NULL
          }
      if (correlation) {
            # (partial) correlation ADF Gamma via the delta method (handles
            # fixed.x)
            cor_idx_g <- lav_gamma_recipe_cor_idx(gamma_recipe, ov_names[[g]])
            nacov[[g]] <- lav_samp_partial_cor_gamma(
              m_y = y, cor_idx = cor_idx_g,
              meanstructure = meanstructure,
              fixed_x = fixed_x, x_idx = x_idx[[g]]
            )
      } else {
            # the NT flavor may only replace the ADF Gamma for the
            # ULS/DWLS(-continuous) estimators when the robust.sem.nt
            # sandwich needs it (their default se since 0.6-21; for
            # DWLS the diagonal weights then also derive from the NT
            # Gamma, by design): for WLS/DLS the (full) weight matrix is
            # derived from this NACOV, and storing the NT Gamma instead
            # silently turned WLS into GLS (before 0.7-1, when
            # test = "browne.residual.nt" was combined with
            # estimator = "WLS"); note that the browne.residual.nt test
            # itself never reads this slot (it uses the fast kernel, or
            # recomputes via lav_object_gamma -- see lav_test_browne)
            if (estimator %in% c("ULS", "DWLS") &&
                "robust.sem.nt" %in% lavoptions$se) {
              nacov[[g]] <-
                lav_samp_gamma_nt(
                  m_y = y,
                  m_cov = cov[[g]],
                  m_mean = mean[[g]],
                  x_idx = x_idx[[g]],
                  # cluster.idx = cluster.idx, # not available
                  fixed_x = fixed_x,
                  conditional_x = conditional_x,
                  meanstructure = meanstructure,
                  slopestructure = conditional_x
                )
            } else {
              nacov[[g]] <-
                lav_samp_gamma(
                  m_y = y,
                  x_idx = x_idx[[g]],
                  cluster_idx = cluster_idx,
                  fixed_x = fixed_x,
                  conditional_x = conditional_x,
                  meanstructure = meanstructure,
                  slopestructure = conditional_x,
                  gamma_n_minus_one = gamma_recipe$n.minus.one,
                  unbiased = gamma_recipe$unbiased,
                  mplus_wls = gamma_recipe$mplus.wls,
                  wt = wt[[g]],
                  sampling_weights_type = swt_type
                )
            }
          }
        } else { # categorical case
          if (!is.null(cat_1$WLS.W)) {
            # issue #254: use the cluster-robust meat for Gamma/NACOV when
            # available (cluster= set). The WLS weight matrix (wls_v below)
            # still uses the standard cat_1$WLS.W, so point estimates are
            # unchanged relative to the non-clustered fit.
            if (!is.null(cat_1$WLS.W.cluster)) {
              nacov[[g]] <- cat_1$WLS.W.cluster * nobs[[g]]
              # finite-sample correction G/(G-1) on the cluster-aggregated
              # Gamma, where G is the number of clusters (matches Mplus
              # TYPE=COMPLEX + WLSMV/ULSMV).
              nc <- lavdata@Lp[[g]]$nclusters[[2]]
              nacov[[g]] <- nacov[[g]] * (nc / (nc - 1))
            } else {
              nacov[[g]] <- cat_1$WLS.W * nobs[[g]]
              if (gamma_recipe$n.minus.one) {
                nacov[[g]] <- nacov[[g]] * (nobs[[g]] / (nobs[[g]] - 1L))
              }
            }
          }
          if (estimator == "catML") {
            # remove all but the correlation part
            ntotal <- nrow(nacov[[g]])
            pstar <- nrow(cat_1$A22)
            nocor <- ntotal - pstar
            if (length(nocor) > 0L) {
              nacov[[g]] <- nacov[[g]][
                -seq_len(nocor),
                -seq_len(nocor)
              ]
            }
          }
        }
      } else if (estimator == "PML") {
        # no nacov ... for now
      }

      # group.w.free
      if (!is.null(nacov[[g]]) && group_w_free) {
        # unweight!!
        a <- group_w[[g]] * sum(unlist(nobs)) / nobs[[g]]
        # always 1!!!
        nacov[[g]] <- lav_mat_bdiag(matrix(a, 1, 1), nacov[[g]])
      }
    }

    # wls_v
    if (!wls_v_user && nlevels == 1L) {
      if (estimator == "DLS" && dls_gamma_nt == "sample" && dls_a < 1.0) {
        # compute GammaNT here
        if (correlation) {
          cor_idx_g <- lav_gamma_recipe_cor_idx(gamma_recipe, ov_names[[g]])
          gamma_nt <- lav_samp_partial_cor_gamma_nt(
            m_cov         = cov[[g]],
            cor_idx       = cor_idx_g,
            meanstructure = meanstructure,
            fixed_x       = fixed_x,
            x_idx         = x_idx[[g]]
          )
        } else {
          gamma_nt <- lav_samp_gamma_nt(
            m_cov          = cov[[g]],
            m_mean         = mean[[g]],
            rescale        = FALSE,
            x_idx          = x_idx[[g]],
            fixed_x        = fixed_x,
            conditional_x  = conditional_x,
            meanstructure  = meanstructure,
            slopestructure = conditional_x
          )
        }
      }

      # GLS: since 0.7-1 the estimation machinery no longer needs the
      # (large) WLS.V matrix (the objective uses the trace shortcut, the
      # gradient the omega approach, and the expected information is
      # streamed); when a consumer does need the full matrix (robust
      # SEs/tests, lavInspect), it is built on demand in
      # lav_model_h1_information. Only the variants below still
      # pre-compute it here.
      gls_v_lazy <- (estimator == "GLS" && !correlation &&
        !conditional_x && !group_w_free && missing == "listwise")
      if ((estimator == "GLS" && !gls_v_lazy) ||
        (estimator == "DLS" && dls_gamma_nt == "sample" &&
          dls_a == 1.0)) {
        # Note: we need the 'original' COV/MEAN/ICOV
        #        sample statistics; not the 'residual' version
        wls_v[[g]] <- lav_samp_wls_v_nt_g(
          m_cov         = cov[[g]],
          m_mean        = mean[[g]],
          m_icov        = icov[[g]],
          cor_idx       = lav_gamma_recipe_cor_idx(gamma_recipe,
                                                   ov_names[[g]]),
          correlation   = correlation,
          x_idx         = x_idx[[g]],
          fixed_x       = fixed_x,
          conditional_x = conditional_x,
          meanstructure = meanstructure
        )
        if (lavoptions$gls.v11.mplus && !conditional_x && meanstructure) {
          # bug in Mplus? V11 rescaled by nobs[[g]]/(nobs[[g]]-1)
          nvar <- NCOL(cov[[g]])
          wls_v[[g]][1:nvar, 1:nvar] <-
            wls_v[[g]][1:nvar, 1:nvar,
              drop = FALSE
            ] * (nobs[[g]] / (nobs[[g]] - 1))
        }
      } else if (estimator == "ML") {
        # no wls_v here, since function of model-implied moments
      } else if (estimator %in% c("WLS", "DWLS", "ULS", "DLS", "IV")) {
        if (!categorical) {
          if (estimator == "WLS" || estimator == "DLS") {
            if (!fixed_x) {
              if (estimator != "DLS") {
                # Gamma should be po before we invert
                ev <- eigen(nacov[[g]], # symmetric=FALSE,
                  only.values = TRUE
                )$values
                if (is.complex(ev)) {
                  lav_msg_stop(gettext(
                    "Gamma (NACOV) matrix is not positive-definite"))
                }
                if (any(Re(ev) < 0)) {
                  lav_msg_stop(gettext(
                    "Gamma (NACOV) matrix is not positive-definite"))
                }
              }
              if (estimator == "DLS" && dls_gamma_nt == "sample") {
                if (dls_a == 1.0) {
                  # nothing to do, use GLS version
                } else {
                  w_dls <-
                    (1 - dls_a) * nacov[[g]] + dls_a * gamma_nt
                  wls_v[[g]] <-
                    lav_mat_sym_inverse(w_dls)
                }
              } else { # WLS
                wls_v[[g]] <-
                  lav_mat_sym_inverse(nacov[[g]])
              }
            } else {
              # fixed.x: we have zero cols/rows
              # ginv does the trick, but perhaps this is overkill
              # just removing the zero rows/cols, invert, and
              # fill back in the zero rows/cols would do it
              # wls_v[[g]] <- MASS::ginv(nacov[[g]])
              if (estimator == "DLS" && dls_gamma_nt == "sample") {
                w_dls <- (1 - dls_a) * nacov[[g]] + dls_a * gamma_nt
                wls_v[[g]] <-
                  lav_mat_sym_inverse(w_dls)
              } else { # WLS
                wls_v[[g]] <-
                  lav_mat_sym_inverse(nacov[[g]])
              }
            }
          } else if (estimator == "DWLS") {
            dacov <- diag(nacov[[g]])
            if (!all(is.finite(dacov))) {
              lav_msg_stop(gettext(
                "diagonal of Gamma (NACOV) contains non finite values"))
            }
            if (fixed_x) {
              # structural zeroes!
              zero_idx <- which(dacov == 0.0)
              idacov <- 1 / dacov
              idacov[zero_idx] <- 0.0
            } else {
              idacov <- 1 / dacov
            }
            wls_v[[g]] <- diag(idacov,
              nrow = NROW(nacov[[g]]),
              ncol = NCOL(nacov[[g]])
            )
            wls_vd[[g]] <- idacov
          } else if (estimator == "ULS") {
            # wls_v[[g]] <- diag(length(WLS.obs[[g]]))
            wls_vd[[g]] <- rep(1, length(wls_obs[[g]]))
          }
        } else {
          if (estimator == "WLS") {
            c_s <- tryCatch(chol(cat_1$WLS.W * nobs[[g]]),
                           error = function(e) NULL)
            if (is.null(c_s)) {
              lav_msg_stop(gettext("could not invert CAT$WLS.W"))
            }
            wls_v[[g]] <- chol2inv(c_s)
          } else if (estimator == "DWLS") {
            dacov <- diag(cat_1$WLS.W * nobs[[g]])
            # wls_v[[g]] <- diag(1/dacov, nrow=NROW(CAT$WLS.W),
            #                            ncol=NCOL(CAT$WLS.W))
            wls_vd[[g]] <- 1 / dacov
          } else if (estimator == "ULS" || estimator == "IV") {
            # wls_v[[g]] <- diag(length(WLS.obs[[g]]))
            wls_vd[[g]] <- rep(1, length(wls_obs[[g]]))
          }
        }
      } else if (estimator == "PML" || estimator == "FML") {
        # no wls_v here
      }

      # group.w.free (only if categorical)
      # (continuous WLS gets its group-weight row via the bdiag'ed NACOV
      #  before inversion, see above; the GLS weight matrix is built
      #  directly from the sample moments, so add the row here)
      # FIXME: DLS + group.w.free: the (1 - a)*NACOV + a*GammaNT mix is
      #        non-conformable when group.w.free = TRUE
      if (group_w_free && (categorical || estimator == "GLS")) {
        if (!is.null(wls_v[[g]])) {
          # unweight!!
          a <- group_w[[g]] * sum(unlist(nobs)) / nobs[[g]]
          # always 1!!!
          # invert
          a <- 1 / a
          wls_v[[g]] <- lav_mat_bdiag(matrix(a, 1, 1), wls_v[[g]])
        }
        if (!is.null(wls_vd[[g]])) {
          # unweight!!
          a <- group_w[[g]] * sum(unlist(nobs)) / nobs[[g]]
          # always 1!!!
          # invert
          a <- 1 / a
          wls_vd[[g]] <- c(a, wls_vd[[g]])
        }
      }
    }
  } # ngroups

  # remove 'CAT', unless debug -- this is to save memory
  if (!lav_debug()) {
    cat_1 <- list()
  }

  # construct SampleStats object
  lav_sample_stats <- new("lavSampleStats",
    # sample moments
    th = th,
    th.idx = th_idx,
    th.names = th_names,
    mean = mean,
    cov = cov,
    var = var,

    # residual (y | x)
    res.cov = res_cov,
    res.var = res_var,
    res.th = res_th,
    res.th.nox = res_th_nox,
    res.slopes = res_slopes,
    res.int = res_int,
    mean.x = mean_x,
    cov.x = cov_x,
    bifreq = bifreq,
    group.w = group_w,

    # convenience
    nobs = nobs,
    ntotal = sum(unlist(nobs)),
    ngroups = ngroups,
    x.idx = x_idx,

    # extra sample statistics
    icov = icov,
    cov.log.det = cov_log_det,
    res.icov = res_icov,
    res.cov.log.det = res_cov_log_det,
    ridge = ridge_eps,
    WLS.obs = wls_obs,
    WLS.V = wls_v,
    WLS.VD = wls_vd,
    NACOV = nacov,
    NACOV.user = nacov_user,

    # cluster/levels
    YLp = ylp,

    # missingness
    missing.flag = missing_flag,
    missing = missing_1,
    missing.h1 = missing_h1,
    zero.cell.tables = zero_cell_tables
  )

  # just a SINGLE warning if we have empty cells
  if ((!is.null(lavoptions$samplestats) && lavoptions$samplestats) &&
      categorical && zero_cell_warn &&
      any(sapply(zero_cell_tables, nrow) > 0L)) {
    nempty <- sum(sapply(zero_cell_tables, nrow))
    lav_msg_warn(gettextf(
      "%s bivariate tables have empty cells; to see them, use:
      lavInspect(fit, \"zero.cell.tables\")", nempty)
    )
  }

  lav_sample_stats
}

# align a summary statistic that was provided as an attribute of sample.cov
# (res.slopes, cov.x, mean.x) with the model's internal variable order:
# sample.cov itself is name-matched against ov.names in
# lav_samp_from_moments(), but these attributes used to be consumed as-is --
# silently misaligning them whenever the provided order differs from the
# internal one (e.g., sam() with conditional.x = TRUE, where VETA is in
# measurement order, but the structural model is y-variables-first); if
# (dim)names are present, reorder by name; if absent, assume the statistic
# is already in the internal order (pre-0.7-1 behavior)
lav_samp_align_by_names <- function(stat = NULL, row_names = NULL,
                                    col_names = NULL, what = "res.slopes",
                                    g = 1L) {
  if (is.null(stat)) {
    return(NULL)
  }
  # a vector is expected (mean.x); accept a one-row/column matrix
  if (is.null(col_names)) {
    stat <- drop(stat)
    if (length(stat) != length(row_names)) {
      lav_msg_stop(gettextf(
        "%1$s attribute of sample.cov in group %2$s should contain %3$s
        elements.", what, g, length(row_names)))
    }
    if (!is.null(names(stat))) {
      idx <- match(row_names, names(stat))
      if (anyNA(idx)) {
        lav_msg_stop(gettextf(
          "names of the %1$s attribute of sample.cov (group %2$s) do not
          match the model. found: %3$s expected: %4$s", what, g,
          lav_msg_view(names(stat)), lav_msg_view(row_names)))
      }
      stat <- stat[idx]
    }
    return(unclass(unname(stat)))
  }
  # matrix
  if (!is.matrix(stat) || nrow(stat) != length(row_names) ||
      ncol(stat) != length(col_names)) {
    lav_msg_stop(gettextf(
      "%1$s attribute of sample.cov in group %2$s should be a %3$s by %4$s
      matrix.", what, g, length(row_names), length(col_names)))
  }
  if (!is.null(rownames(stat))) {
    row_idx <- match(row_names, rownames(stat))
    if (anyNA(row_idx)) {
      lav_msg_stop(gettextf(
        "rownames of the %1$s attribute of sample.cov (group %2$s) do not
        match the model. found: %3$s expected: %4$s", what, g,
        lav_msg_view(rownames(stat)), lav_msg_view(row_names)))
    }
    stat <- stat[row_idx, , drop = FALSE]
  }
  if (!is.null(colnames(stat))) {
    col_idx <- match(col_names, colnames(stat))
    if (anyNA(col_idx)) {
      lav_msg_stop(gettextf(
        "colnames of the %1$s attribute of sample.cov (group %2$s) do not
        match the model. found: %3$s expected: %4$s", what, g,
        lav_msg_view(colnames(stat)), lav_msg_view(col_names)))
    }
    stat <- stat[, col_idx, drop = FALSE]
  }
  unclass(unname(stat))
}


lav_samp_from_moments <- function(sample_cov = NULL,
                                         sample_mean = NULL,
                                         sample_th = NULL,
                                         sample_nobs = NULL,
                                         ov_names = NULL, # including x
                                         ov_names_x = NULL,
                                         wls_v = NULL,
                                         nacov = NULL,
                                         lavoptions = NULL) {
  # extract options
  estimator <- lavoptions$estimator
  # mimic <- lavoptions$mimic
  meanstructure <- lavoptions$meanstructure
  correlation <- lavoptions$correlation
  correlation_ov <- lavoptions$.correlation.ov
  if (is.null(correlation_ov)) {
    correlation_ov <- character(0L)
  }
  group_w_free <- lavoptions$group.w.free
  ridge <- lavoptions$ridge
  rescale <- lavoptions$sample.cov.rescale

  # no multilevel yet
  nlevels <- 1L

  # ridge default
  if (ridge) {
    if (is.numeric(lavoptions$ridge.constant)) {
      ridge_eps <- lavoptions$ridge.constant
    } else {
      ridge_eps <- 1e-5
    }
  } else {
    ridge_eps <- 0.0
  }

  # new in 0.6-3:
  # check if sample.cov has attributes if conditional.x = TRUE
  sample_res_slopes <- attr(sample_cov, "res.slopes")
  sample_cov_x <- attr(sample_cov, "cov.x")
  sample_mean_x <- attr(sample_cov, "mean.x")
  if (!is.null(sample_res_slopes)) {
    conditional_x <- TRUE
    # strip attributes
    attr(sample_cov, "res.slopes") <- NULL
    attr(sample_cov, "cov.x") <- NULL
    attr(sample_cov, "mean.x") <- NULL
    # make list
    if (!is.list(sample_res_slopes)) {
      sample_res_slopes <- list(sample_res_slopes)
    }
    if (!is.list(sample_cov_x)) {
      sample_cov_x <- list(sample_cov_x)
    }
    if (!is.list(sample_mean_x)) {
      sample_mean_x <- list(sample_mean_x)
    }
  } else if (!is.null(sample_cov_x)) {
    conditional_x <- FALSE
    fixed_x <- TRUE

    # strip attributes
    attr(sample_cov, "cov.x") <- NULL
    attr(sample_cov, "mean.x") <- NULL
    # make list
    if (!is.list(sample_cov_x)) {
      sample_cov_x <- list(sample_cov_x)
    }
    if (!is.list(sample_mean_x)) {
      sample_mean_x <- list(sample_mean_x)
    }
  } else if (is.null(sample_cov_x) && length(unlist(ov_names_x)) > 0L) {
    # fixed.x = TRUE, but only joint sample.cov is provided
    conditional_x <- FALSE
    fixed_x <- TRUE

    # create sample.cov.x and sample.mean.x later...
  } else {
    conditional_x <- FALSE
    fixed_x <- FALSE
  }

  # matrix -> list
  if (!is.list(sample_cov)) {
    sample_cov <- list(sample_cov)
  }

  # number of groups
  ngroups <- length(sample_cov)

  # ov.names
  if (!is.list(ov_names)) {
    ov_names <- rep(list(ov_names), ngroups)
  }
  if (!is.list(ov_names_x)) {
    ov_names_x <- rep(list(ov_names_x), ngroups)
  }

  if (!is.null(sample_mean)) {
    meanstructure <- TRUE
    if (!is.list(sample_mean)) {
      # check if sample.mean is string (between single quotes)
      if (is.character(sample_mean)) {
        sample_mean <- lav_char2num(sample_mean)
      }
      sample_mean <- list(unname(sample_mean))
    } else {
      sample_mean <- lapply(lapply(sample_mean, unname), unclass)
    }
  }

  if (!is.null(sample_th)) {
    th_idx <- attr(sample_th, "th.idx")
    attr(sample_th, "th.idx") <- NULL
    if (is.null(th_idx)) {
      lav_msg_stop(gettext("sample.th should have a th.idx attribute"))
    } else {
      if (is.list(th_idx)) {
        th_names <- lapply(th_idx, names)
        th_idx <- lapply(lapply(th_idx, unname), unclass)
      } else {
        th_names <- list(names(th_idx))
        th_idx <- list(unclass(unname(th_idx)))
      }
    }
    if (is.list(sample_th)) {
      # strip names and lavaan.vector class
      sample_th <- lapply(lapply(sample_th, unname), unclass)
    } else {
      # strip names and lavaan.vector class, make list
      sample_th <- list(unclass(unname(sample_th)))
    }
  } else {
    th_idx <- vector("list", length = ngroups)
    th_names <- vector("list", length = ngroups)
  }

  # sample statistics per group
  cov <- vector("list", length = ngroups)
  var <- vector("list", length = ngroups)
  mean <- vector("list", length = ngroups)
  th <- vector("list", length = ngroups)
  # th.idx      <- vector("list", length = ngroups)
  # th.names    <- vector("list", length = ngroups)

  # residual (y | x)
  res_cov <- vector("list", length = ngroups)
  res_var <- vector("list", length = ngroups)
  res_slopes <- vector("list", length = ngroups)
  res_int <- vector("list", length = ngroups)
  res_th <- vector("list", length = ngroups)
  res_th_nox <- vector("list", length = ngroups)

  # fixed.x / conditional.x
  mean_x <- vector("list", length = ngroups)
  cov_x <- vector("list", length = ngroups)

  bifreq <- vector("list", length = ngroups)

  # extra sample statistics per group
  icov <- vector("list", length = ngroups)
  cov_log_det <- vector("list", length = ngroups)
  res_icov <- vector("list", length = ngroups)
  res_cov_log_det <- vector("list", length = ngroups)
  wls_obs <- vector("list", length = ngroups)
  missing_1 <- vector("list", length = ngroups)
  missing_h1 <- vector("list", length = ngroups)
  missing_flag <- FALSE
  zero_cell_tables <- vector("list", length = ngroups)
  ylp <- vector("list", length = ngroups)

  # group weights
  group_w <- vector("list", length = ngroups)
  x_idx <- vector("list", length = ngroups)

  categorical <- FALSE
  if (!is.null(sample_th)) {
    categorical <- TRUE
  }

  wls_vd <- vector("list", length = ngroups)
  if (is.null(wls_v)) {
    wls_v <- vector("list", length = ngroups)
    wls_v_user <- FALSE
  } else {
    if (!is.list(wls_v)) {
      if (ngroups == 1L) {
        wls_v <- list(unclass(wls_v))
      } else {
        lav_msg_stop(gettextf("wls_v argument should be a list of length %s",
          ngroups)
        )
      }
    } else {
      if (length(wls_v) != ngroups) {
        lav_msg_stop(gettextf(
          "wls_v assumes %1$s groups; data contains %2$s groups",
          length(wls_v), ngroups))
      }
      wls_v <- lapply(wls_v, unclass)
    }

    # is wls_v full? check first
    if (is.null(dim(wls_v[[1]]))) {
      # we will assume it is the diagonal only
      wls_vd <- wls_v
      wls_v <- lapply(wls_vd, diag)
    } else {
      # create WLS.VD
      wls_vd <- lapply(wls_v, diag)
      # we could remove wls_v to save space...
    }

    wls_v_user <- TRUE
    # FIXME: check dimension of wls_v!!
  }

  # a user-provided WLS.V is not used by the GLS estimation machinery
  # (since 0.6-10 the objective/gradient are direct functions of the
  # sample moments); ignore it altogether (and warn), rather than using
  # it inconsistently
  if (wls_v_user && estimator == "GLS") {
    lav_msg_warn(gettext(
      "WLS.V is ignored when estimator = \"GLS\"; use estimator = \"WLS\" to
      estimate with a user-provided weight matrix."))
    wls_v <- vector("list", length = ngroups)
    wls_vd <- vector("list", length = ngroups)
    wls_v_user <- FALSE
  }

  if (is.null(nacov)) {
    nacov <- vector("list", length = ngroups)
    nacov_user <- FALSE
  } else {
    if (!is.list(nacov)) {
      if (ngroups == 1L) {
        nacov <- list(unclass(nacov))
      } else {
        lav_msg_stop(gettextf(
          "NACOV argument should be a list of length %s", ngroups))
      }
    } else {
      if (length(nacov) != ngroups) {
        lav_msg_stop(gettextf(
          "NACOV assumes %1$s groups; data contains %2$s groups",
          length(nacov), ngroups))
      }
      nacov <- lapply(nacov, unclass)
    }
    nacov_user <- TRUE
    # FIXME: check dimension of NACOV!!
  }

  nobs <- as.list(as.integer(sample_nobs))


  for (g in 1:ngroups) {
    # exogenous x?
    nexo <- length(ov_names_x[[g]])
    if (nexo) {
      # two cases: ov.names contains 'x' variables, or not
      if (conditional_x) {
        # ov.names.x are NOT in ov.names
        x_idx[[g]] <- which(ov_names[[g]] %in% ov_names_x[[g]])
      } else {
        if (fixed_x) {
          # ov.names.x are a subset of ov.names
          x_idx[[g]] <- match(ov_names_x[[g]], ov_names[[g]])
          stopifnot(!anyNA(x_idx[[g]]))
        } else {
          x_idx[[g]] <- integer(0L)
        }
      }
    } else {
      x_idx[[g]] <- integer(0L)
      conditional_x <- FALSE
      fixed_x <- FALSE
    }


    # group weight
    group_w[[g]] <- nobs[[g]] / sum(unlist(nobs))

    tmp_cov <- sample_cov[[g]]

    # make sure that the matrix is fully symmetric (NEEDED?)
    t_1 <- t(tmp_cov)
    tmp_cov[upper.tri(tmp_cov)] <- t_1[upper.tri(t_1)]

    # check dimnames
    if (!is.null(rownames(tmp_cov))) {
      cov_names <- rownames(tmp_cov)
    } else if (!is.null(colnames(tmp_cov))) {
      cov_names <- colnames(tmp_cov)
    } else {
      lav_msg_stop(gettext(
        "please provide row/col names for the covariance matrix!"))
    }

    # extract only the part we need (using ov.names)
    if (conditional_x) {
      idx <- match(ov_names[[g]][-x_idx[[g]]], cov_names)
    } else {
      idx <- match(ov_names[[g]], cov_names)
    }
    if (any(is.na(idx))) {
      cat("found: ", cov_names, "\n")
      cat("expected: ", ov_names[[g]], "\n")
      lav_msg_stop(gettextf(
        "rownames of covariance matrix do not match the model!
        found: %1$s expected: %2$s", lav_msg_view(cov_names),
        lav_msg_view(ov_names[[g]])))
    } else {
      tmp_cov <- tmp_cov[idx, idx, drop = FALSE]
    }

    # strip dimnames
    dimnames(tmp_cov) <- NULL

    if (is.null(sample_mean)) {
      # assume zero mean vector
      tmp_mean <- numeric(ncol(tmp_cov))
    } else {
      # extract only the part we need
      tmp_mean <- unclass(sample_mean[[g]][idx])
    }



    if (categorical) {
      # categorical + conditional.x = TRUE
      if (conditional_x) {
        th_g <- numeric(length(th_idx[[g]]))
        ord_idx <- which(th_idx[[g]] > 0)
        num_idx <- which(th_idx[[g]] == 0)
        if (length(ord_idx) > 0L) {
          th_g[ord_idx] <- sample_th[[g]][ord_idx]
        }
        if (length(num_idx) > 0L) {
          ord_var_idx <- unique(th_idx[[g]][th_idx[[g]] > 0])
          th_g[num_idx] <- -1 * sample_mean[[g]][-ord_var_idx]
        }
        res_th[[g]] <- th_g
        res_th_nox[[g]] <- sample_th[[g]]

        res_cov[[g]] <- tmp_cov
        if (ridge) {
          diag(res_cov[[g]]) <- diag(res_cov[[g]]) + ridge_eps
        }
        res_var[[g]] <- diag(tmp_cov)
        res_int[[g]] <- tmp_mean

        res_slopes[[g]] <- lav_samp_align_by_names(sample_res_slopes[[g]],
          row_names = ov_names[[g]][-x_idx[[g]]],
          col_names = ov_names_x[[g]], what = "res.slopes", g = g)
        cov_x[[g]] <- lav_samp_align_by_names(sample_cov_x[[g]],
          row_names = ov_names_x[[g]], col_names = ov_names_x[[g]],
          what = "cov.x", g = g)
        mean_x[[g]] <- lav_samp_align_by_names(sample_mean_x[[g]],
          row_names = ov_names_x[[g]], what = "mean.x", g = g)

        # th.idx and th.names are already ok

        # categorical + conditional.x = FALSE
      } else {
        th_g <- numeric(length(th_idx[[g]]))
        ord_idx <- which(th_idx[[g]] > 0)
        num_idx <- which(th_idx[[g]] == 0)
        if (length(ord_idx) > 0L) {
          th_g[ord_idx] <- sample_th[[g]][ord_idx]
        }
        if (length(num_idx) > 0L) {
          ord_var_idx <- unique(th_idx[[g]][th_idx[[g]] > 0])
          th_g[num_idx] <- -1 * sample_mean[[g]][-ord_var_idx]
        }
        th[[g]] <- th_g

        cov[[g]] <- tmp_cov
        if (ridge) {
          diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
        }
        var[[g]] <- diag(tmp_cov)
        mean[[g]] <- tmp_mean

        # fixed.x? (needed?)
        if (fixed_x) {
          cov_x[[g]] <- lav_samp_align_by_names(sample_cov_x[[g]],
            row_names = ov_names_x[[g]], col_names = ov_names_x[[g]],
            what = "cov.x", g = g)
          mean_x[[g]] <- lav_samp_align_by_names(sample_mean_x[[g]],
            row_names = ov_names_x[[g]], what = "mean.x", g = g)
        }

        # th, th.idx and th.names are already ok
      }

      # multilevel
    } else if (nlevels > 1L) {
      lav_msg_stop(gettext("multilevel + sample stats not ready yet"))


      # single level
    } else {
      # single-level + continuous + conditional.x = TRUE
      if (conditional_x) {
        res_cov[[g]] <- tmp_cov
        if (ridge) {
          diag(res_cov[[g]]) <- diag(res_cov[[g]]) + ridge_eps
        }
        res_var[[g]] <- diag(tmp_cov)
        res_int[[g]] <- tmp_mean
        res_slopes[[g]] <- lav_samp_align_by_names(sample_res_slopes[[g]],
          row_names = ov_names[[g]][-x_idx[[g]]],
          col_names = ov_names_x[[g]], what = "res.slopes", g = g)
        cov_x[[g]] <- lav_samp_align_by_names(sample_cov_x[[g]],
          row_names = ov_names_x[[g]], col_names = ov_names_x[[g]],
          what = "cov.x", g = g)
        mean_x[[g]] <- lav_samp_align_by_names(sample_mean_x[[g]],
          row_names = ov_names_x[[g]], what = "mean.x", g = g)

        # no rescale!

        # icov and cov.log.det
        # if(lavoptions$sample.icov) {
        out <- lav_samp_icov(
          cov_1 = res_cov[[g]],
          ridge = 1e-05,
          x_idx = x_idx[[g]],
          ngroups = ngroups, g = g
        )
        res_icov[[g]] <- out$icov
        res_cov_log_det[[g]] <- out$cov.log.det
        # }

        # continuous + conditional.x = FALSE
      } else {
        cov[[g]] <- tmp_cov
        mean[[g]] <- tmp_mean

        # rescale cov by (N-1)/N?
        if (rescale) {
          # we 'transform' the sample cov (divided by n-1)
          # to a sample cov divided by 'n'
          cov[[g]] <- ((nobs[[g]] - 1) / nobs[[g]]) * cov[[g]]
        }
        if (ridge) {
          diag(cov[[g]]) <- diag(cov[[g]]) + ridge_eps
        }
        var[[g]] <- diag(cov[[g]])

        # icov and cov.log.det
        # if(lavoptions$sample.icov) {
        out <- lav_samp_icov(
          cov_1 = cov[[g]],
          ridge = 1e-05,
          x_idx = x_idx[[g]],
          ngroups = ngroups,
          g = g
        )
        icov[[g]] <- out$icov
        cov_log_det[[g]] <- out$cov.log.det
        # }

        # fixed.x?
        if (fixed_x) {
          if (is.null(sample_cov_x)) {
            cov_x[[g]] <- cov[[g]][x_idx[[g]], x_idx[[g]],
              drop = FALSE
            ]
          } else {
            cov_x[[g]] <- lav_samp_align_by_names(sample_cov_x[[g]],
              row_names = ov_names_x[[g]], col_names = ov_names_x[[g]],
              what = "cov.x", g = g)
          }
          if (is.null(sample_mean_x)) {
            mean_x[[g]] <- mean[[g]][x_idx[[g]]]
          } else {
            mean_x[[g]] <- lav_samp_align_by_names(sample_mean_x[[g]],
              row_names = ov_names_x[[g]], what = "mean.x", g = g)
          }
        }
      }

      # correlation structure?
      if (correlation) {
        # which variables are scaled to unit variance? (all, or a subset
        # for a 'partial' correlation structure)
        if (length(correlation_ov) > 0L) {
          cor_idx <- which(ov_names[[g]] %in% correlation_ov)
        } else {
          cor_idx <- seq_len(ncol(cov[[g]]))
        }
        cov[[g]] <- lav_cov2cor_partial(cov[[g]], cor_idx)
        var[[g]] <- diag(cov[[g]])
        if (conditional_x) {
          res_cov[[g]] <- lav_cov2cor_partial(res_cov[[g]], cor_idx)
          res_var[[g]] <- diag(res_cov[[g]])
          cov_x[[g]] <- cov2cor(cov_x[[g]])
          # FIXME: slopes? more?
        }
      }
    }

    # WLS.obs
    wls_obs[[g]] <- lav_samp_wls_obs(
      mean_g = mean[[g]],
      cov_g = cov[[g]], var_g = var[[g]], th_g = th[[g]],
      th_idx_g = th_idx[[g]], res_int_g = res_int[[g]],
      res_cov_g = res_cov[[g]], res_var_g = res_var[[g]],
      res_th_g = res_th[[g]], res_slopes_g = res_slopes[[g]],
      group_w_g = log(nobs[[g]]),
      categorical = categorical, conditional_x = conditional_x,
      meanstructure = meanstructure, correlation = correlation,
      num_idx = if (correlation && length(correlation_ov) > 0L) {
        which(!(ov_names[[g]] %in% correlation_ov))
      } else {
        integer(0L)
      },
      slopestructure = conditional_x,
      group_w_free = group_w_free
    )

    # wls_v
    if (!wls_v_user) {
      # GLS: build WLS.V lazily (see lav_samp_from_data); note that this
      # moments-based branch never applied the gls.v11.mplus rescale, and
      # the on-demand build in lav_model_h1_information preserves that
      # (it checks lavdata@data.type)
      gls_v_lazy <- (estimator == "GLS" && !correlation &&
        !conditional_x && !group_w_free)
      if (estimator == "GLS" && !gls_v_lazy) {
        # FIXME: in <0.5-21, we had
        # V11 <- icov[[g]]
        #    if(mimic == "Mplus") { # is this a bug in Mplus?
        #        V11 <- V11 * nobs[[g]]/(nobs[[g]]-1)
        #    }
        wls_v[[g]] <- lav_samp_wls_v_nt_g(
          m_cov         = cov[[g]],
          m_mean        = mean[[g]],
          m_icov        = icov[[g]],
          cor_idx       = if (length(correlation_ov) > 0L) {
            which(ov_names[[g]] %in% correlation_ov)
          } else {
            seq_along(ov_names[[g]])
          },
          correlation   = correlation,
          x_idx         = x_idx[[g]],
          fixed_x       = fixed_x,
          conditional_x = conditional_x,
          meanstructure = meanstructure
        )
      } else if (estimator == "ULS") {
        # only the diagonal is stored (as in lav_samp_from_data); no
        # consumer needs the (pstar x pstar) identity matrix
        wls_vd[[g]] <- rep(1, length(wls_obs[[g]]))
      } else if (estimator == "WLS" || estimator == "DWLS") {
        if (is.null(wls_v[[g]])) {
          lav_msg_stop(gettext(
            "the (D)WLS estimator is only available with full data
            or with a user-provided wls_v"))
        }
      }

      # group.w.free
      if (!is.null(wls_v[[g]]) && group_w_free) {
        # FIXME!!!
        wls_v[[g]] <- lav_mat_bdiag(matrix(1, 1, 1), wls_v[[g]])
      }
    }

    # NACOV: nothing can be computed from moments alone, except the
    # normal-theory Gamma; compute it when the robust.sem.nt sandwich
    # asks for it (the default se for continuous ULS) -- before 0.7-1,
    # this failed with a cryptic error. (The browne.residual.nt test
    # does not read this slot: it uses the fast kernel, or recomputes
    # at test time.)
    if (!nacov_user && is.null(nacov[[g]]) &&
      !categorical && !correlation &&
      estimator %in% c("WLS", "DWLS", "ULS", "DLS") &&
      "robust.sem.nt" %in% lavoptions$se) {
      nacov[[g]] <- lav_samp_gamma_nt(
        m_cov          = cov[[g]],
        m_mean         = mean[[g]],
        x_idx          = x_idx[[g]],
        fixed_x        = fixed_x,
        conditional_x  = conditional_x,
        meanstructure  = meanstructure,
        slopestructure = conditional_x
      )
      # group.w.free: add the group-weight element (weight 1), as for
      # the Gamma in lav_samp_from_data
      if (group_w_free) {
        nacov[[g]] <- lav_mat_bdiag(matrix(1, 1, 1), nacov[[g]])
      }
    }
  } # ngroups

  # construct SampleStats object
  lav_sample_stats <- new("lavSampleStats",
    # sample moments
    th = th,
    th.idx = th_idx,
    th.names = th_names,
    mean = mean,
    cov = cov,
    var = var,

    # residual (y | x)
    res.cov = res_cov,
    res.var = res_var,
    res.th = res_th,
    res.th.nox = res_th_nox,
    res.slopes = res_slopes,
    res.int = res_int,

    # fixed.x
    mean.x = mean_x,
    cov.x = cov_x,

    # other
    bifreq = bifreq,
    group.w = group_w,

    # convenience
    nobs = nobs,
    ntotal = sum(unlist(nobs)),
    ngroups = ngroups,
    x.idx = x_idx,

    # extra sample statistics
    icov = icov,
    cov.log.det = cov_log_det,
    res.icov = res_icov,
    res.cov.log.det = res_cov_log_det,
    ridge = ridge_eps,
    WLS.obs = wls_obs,
    WLS.V = wls_v,
    WLS.VD = wls_vd,
    NACOV = nacov,
    NACOV.user = nacov_user,

    # cluster/level
    YLp = ylp,

    # missingness
    missing.flag = missing_flag,
    missing = missing_1,
    missing.h1 = missing_h1,
    zero.cell.tables = zero_cell_tables
  )

  lav_sample_stats
}

# compute sample statistics, per missing pattern
lav_samp_mi_patterns <- function(
  y = NULL, mp = NULL, wt = NULL,lp = NULL) {
  # coerce Y to matrix
  y <- as.matrix(y)

  # handle two-level data
  if (!is.null(lp)) {
    y_orig <- y
    # z <- NULL
    if (length(lp$between.idx[[2]]) > 0L) {
      y <- y[, -lp$between.idx[[2]], drop = FALSE]
      # z_idx <- which(!duplicated(lp$cluster.idx[[2]]))
      # z <- y_orig[z_idx, lp$between.idx[[2]], drop = FALSE]
    }
  }

  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y,
      sort_freq = FALSE, coverage = FALSE,
      lp = lp
    )
  }

  yp <- vector("list", length = mp$npatterns)

  # fill in pattern statistics
  for (p in seq_len(mp$npatterns)) {
    # extract raw data for these cases
    raw_1 <- y[mp$case.idx[[p]], mp$pat[p, ], drop = FALSE]

    # more than one case
    if (mp$freq[p] > 1L) {
      if (!is.null(wt)) {
        out <- stats::cov.wt(raw_1,
          wt = wt[mp$case.idx[[p]]],
          method = "ML"
        )
        sy <- out$cov
        my <- out$center
      } else {
        my <- base::.colMeans(raw_1, m = NROW(raw_1), n = NCOL(raw_1))
        # SY <- crossprod(RAW)/Mp$freq[p] - tcrossprod(MY)
        # bad practice, better like this:
        sy <- lav_mat_cov(raw_1)
      }
    # only a single observation (no need to weight!)
    } else {
      sy <- 0
      my <- as.numeric(raw_1)
    }

    if (!is.null(wt)) {
      freq <- sum(wt[mp$case.idx[[p]]])
    } else {
      freq <- mp$freq[p]
    }

    # store sample statistics, var.idx and freq
    yp[[p]] <- list(
      SY = sy, MY = my, var.idx = mp$pat[p, ],
      freq = freq
    )

    # if clustered data, add rowsum over all cases per cluster
    if (!is.null(lp)) {
      tmp <- rowsum.default(raw_1, group = mp$j.idx[[p]], reorder = FALSE)
      yp[[p]]$ROWSUM <- tmp
    }
  }

  # add Zp as an attribute
  # if(!is.null(Lp)) {
  #    Zp <- lav_samp_mi_patterns(Y = Z, Mp = Mp$Zp)
  #    for(p in Mp$Zp$npatterns) {
  #        this.z <- Z[Mp$Zp$case.idx[[p]], drop = FALSE]
  #        Zp[[p]]$ROWSUM <- t(this.z)
  #
  #    }
  #    attr(Yp, "Zp") <- Zp
  # }

  yp
}

# compute sample statistics, per cluster
lav_samp_cl_patterns <- function(y = NULL, lp = NULL,
                                             conditional_x = FALSE) {
  # coerce Y to matrix
  y1 <- as.matrix(y)
  n <- NROW(y1)
  p <- NCOL(y1)

  if (is.null(lp)) {
    lav_msg_stop(gettext("Lp is NULL"))
  }

  # how many levels?
  nlevels <- length(lp$cluster) + 1L

  # compute some sample statistics per level
  ylp <- vector("list", length = nlevels)
  for (l in 2:nlevels) {
    ncluster_sizes <- lp$ncluster.sizes[[l]]
    cluster_size <- lp$cluster.size[[l]]
    cluster_sizes <- lp$cluster.sizes[[l]]
    nclusters <- lp$nclusters[[l]]
    both_idx <- lp$both.idx[[l]]
    within_idx <- lp$within.idx[[l]]
    between_idx <- lp$between.idx[[l]]
    cluster_idx <- lp$cluster.idx[[l]]
    # cluster_size_ns <- lp$cluster.size.ns[[l]]

    # s <- (N^2 - sum(cluster.size^2)) / (N*(nclusters - 1L))
    # same as
    s <- (n - sum(cluster_size^2) / n) / (nclusters - 1)
    # NOTE: must be (nclusters - 1), otherwise, s is not average cluster
    # size even in the balanced case

    y1_means <- colMeans(y1, na.rm = TRUE)
    y1y1 <- lav_mat_crossprod(y1)
    both_idx <- all_idx <- seq_len(p)

    if (length(within_idx) > 0L ||
      length(between_idx) > 0L) {
      both_idx <- all_idx[-c(within_idx, between_idx)]
      # hm, this assumes the 'order' is the
      # same at both levels...
    }

    # cluster-means
    y2 <- rowsum.default(y1,
      group = cluster_idx, reorder = FALSE,
      na.rm = FALSE, # must be FALSE!
    ) / cluster_size
    y2c <- t(t(y2) - y1_means)

    # compute S.w
    # center within variables by grand mean instead of group mean?
    # (YR: apparently not for S.PW)

    y2a <- y2
    # if(length(within.idx) > 0L) {
    #    for(i in 1:length(within.idx)) {
    #        Y2a[, within.idx[i]] <- Y1.means[within.idx[i]]
    #    }
    # }
    y1a <- y1 - y2a[cluster_idx, , drop = FALSE]
    s_w <- lav_mat_crossprod(y1a) / (n - nclusters)

    # S.b
    # three parts: within/within, between/between, between/within
    # standard definition of the between variance matrix
    # divides by (nclusters - 1)
    s_b <- lav_mat_crossprod(y2c * cluster_size, y2c) / (nclusters - 1)

    # check for zero variances
    if (length(both_idx) > 0L) {
      zero_idx <- which(diag(s_b)[both_idx] < 0.0001)
      if (length(zero_idx) > 0L && !anyNA(y2)) {
        lav_msg_warn(gettext(
          "(near) zero variance at between level for split variable:"),
          paste(lp$both.names[[l]][zero_idx], collapse = " ")
        )
      }
    }

    s_1 <- cov(y1, use = "pairwise.complete.obs") * (n - 1L) / n
  # missing by design?
  s_1[is.na(s_1)] <- as.numeric(NA)

    # loglik.x
    # extract 'fixed' level-1 loglik from here
    wx_idx <- lp$ov.x.idx[[1]]
    if (length(wx_idx) > 0L) {
      loglik_x_w <- lav_mvn_h1_loglik_samp(
        sample_nobs = lp$nclusters[[1]],
        sample_cov  = s_1[wx_idx, wx_idx, drop = FALSE]
      )
    } else {
      loglik_x_w <- 0
    }
    # extract 'fixed' level-2 loglik
    bx_idx <- lp$ov.x.idx[[2]]
    if (length(bx_idx) > 0L) {
      covb <- cov(y2[, bx_idx, drop = FALSE]) * (nclusters - 1) / nclusters
      loglik_x_b <- lav_mvn_h1_loglik_samp(
        sample_nobs = lp$nclusters[[2]],
        sample_cov  = covb
      )
    } else {
      loglik_x_b <- 0
    }
    loglik_x <- loglik_x_w + loglik_x_b


    s_pw_start <- s_w
    if (length(within_idx) > 0L) {
      s_pw_start[within_idx, within_idx] <-
        s_1[within_idx, within_idx, drop = FALSE]
    }

    if (length(between_idx) > 0L) {
      s_w[between_idx, ] <- 0
      s_w[, between_idx] <- 0
      s_pw_start[between_idx, ] <- 0
      s_pw_start[, between_idx] <- 0
    }

    if (length(between_idx) > 0L) {
      # this is what is needed for MUML:
      s_b[, between_idx] <-
        (s * nclusters / n) * s_b[, between_idx, drop = FALSE]
      s_b[between_idx, ] <-
        (s * nclusters / n) * s_b[between_idx, , drop = FALSE]
      s_b[between_idx, between_idx] <-
        (s * lav_mat_crossprod(
          y2c[, between_idx, drop = FALSE],
          y2c[, between_idx, drop = FALSE]
        ) / nclusters)
    }

    sigma_b <- (s_b - s_w) / s
    sigma_b[within_idx, ] <- 0
    sigma_b[, within_idx] <- 0

    # what if we have negative variances in Sigma.B?
    # this may happen if 'split' a variable that has no between variance
    zero_idx <- which(diag(sigma_b) < 1e-10)
    if (length(zero_idx) > 0L) {
      sigma_b[zero_idx, ] <- 0
      sigma_b[, zero_idx] <- 0
    }


    mu_w <- numeric(p)
    mu_w[within_idx] <- y1_means[within_idx]

    mu_b <- y1_means
    mu_b[within_idx] <- 0
    if (length(between_idx) > 0L) {
      # replace between.idx by cov(Y2)[,] elements...
      mu_b[between_idx] <- colMeans(y2[, between_idx, drop = FALSE],
        na.rm = TRUE
      )

      s2 <- (cov(y2, use = "pairwise.complete.obs") *
        (nclusters - 1L) / nclusters)

      sigma_b[between_idx, between_idx] <-
        s2[between_idx, between_idx, drop = FALSE]
    }

    # FIXME: Mu.B not quite ok for (fixed.x) x variables if they
    # occur both at level 1 AND level 2
    mu_b_start <- mu_b
    # Mu.B.start[both.idx] <- Mu.B.start[both.idx] - colMeans(Y2c[,both.idx])

    # sample statistics PER CLUSTER-SIZE

    # summary statistics for complete data, conditional.x = FALSE
    # also needed for h1 (even if conditional.x = TRUE)
    cov_d <- vector("list", length = ncluster_sizes)
    mean_d <- vector("list", length = ncluster_sizes)
    for (clz in seq_len(ncluster_sizes)) {
      nj <- cluster_sizes[clz]
      # select clusters with this size
      d_idx <- which(cluster_size == nj)
      ns <- length(d_idx)
      # NOTE:!!!!
      # reorder columns
      # to match A.inv and m.k later on in objective!!!
      tmp2 <- y2[d_idx,
        c(between_idx, sort.int(c(both_idx, within_idx))),
        drop = FALSE
      ]
      mean_d[[clz]] <- colMeans(tmp2, na.rm = TRUE)
      bad_idx <- which(!is.finite(mean_d[[clz]])) # if nrow = 1 + NA
      if (length(bad_idx) > 0L) {
        mean_d[[clz]][bad_idx] <- 0 # ugly, only for starting values
      }
      if (length(d_idx) > 1L) {
        if (any(is.na(tmp2))) {
          # if full column has NA, this will fail...
          # not needed anyway
          # out <- lav_mvn_mi_h1_est_moments(Y = tmp2,
          #          max.iter = 10L)
          # cov.d[[clz]] <- out$Sigma
          cov_d[[clz]] <- 0
        } else {
          cov_d[[clz]] <- (cov(tmp2, use = "complete.obs") *
            (ns - 1) / ns)
        }
      } else {
        cov_d[[clz]] <- 0
      }
    } # clz

    # new in 0.6-12:
    # summary statistics for complete data, conditional.x = TRUE
    # ONLY for twolevel
    if (conditional_x) {
      within_x_idx <- lp$within.x.idx[[1]]
      between_y_idx <- lp$between.y.idx[[2]]
      between_x_idx <- lp$between.x.idx[[2]]
      y1_idx <- lp$ov.y.idx[[1]]
      x1_idx <- c(within_x_idx, between_x_idx) # in that order

      # data
      y1_wb <- y1[, y1_idx, drop = FALSE]
      y2_wb <- y2[, y1_idx, drop = FALSE]
      # if (length(between_y_idx) > 0L) {
      #   y2_z <- y2[, between_y_idx, drop = FALSE]
      # }
      if (length(x1_idx) > 0L) {
        exo_wb1 <- cbind(1, y1[, x1_idx, drop = FALSE])
        exo_wb2 <- cbind(1, y2[, x1_idx, drop = FALSE])
      } else {
        exo_wb1 <- matrix(1, nrow(y1), 1L)
        exo_wb2 <- matrix(1, nrow(y2), 1L)
      }

      # sample beta.wb (level 1)
      sample_wb <- solve(crossprod(exo_wb1), crossprod(exo_wb1, y1_wb))
      sample_yhat_wb1 <- exo_wb1 %*% sample_wb
      sample_yres_wb1 <- y1_wb - sample_yhat_wb1
      sample_yyres_wb1 <- crossprod(sample_yres_wb1)
      sample_xx_wb1 <- crossprod(exo_wb1)

      # sample beta.wb (level 2)
      xx_wb2 <- crossprod(exo_wb2)
      sample_wb2 <- try(solve(xx_wb2, crossprod(exo_wb2, y2_wb)),
        silent = TRUE
      )
      if (inherits(sample_wb2, "try-error")) {
        # this may happen if the covariate is cluster-centered
        # using the observed cluster means; then the 'means' will
        # be all (near) zero, and there is no variance
        sample_wb2 <- MASS::ginv(xx_wb2) %*% crossprod(exo_wb2, y2_wb)
      }
      sample_yhat_wb2 <- exo_wb2 %*% sample_wb2
      sample_yres_wb2 <- y2_wb - sample_yhat_wb2

      # weighted by cluster.size
      sample_yyres_wb2 <- crossprod(
        sample_yres_wb2,
        sample_yres_wb2 * cluster_size
      )
      sample_yres_x_wb2 <- crossprod(
        sample_yres_wb2,
        exo_wb2 * cluster_size
      )
      sample_xx_wb2 <- crossprod(
        exo_wb2,
        exo_wb2 * cluster_size
      )


      sample_clz_y2_res <- vector("list", ncluster_sizes)
      sample_clz_y2_xx <- vector("list", ncluster_sizes)
      sample_clz_y2_b <- vector("list", ncluster_sizes)

      if (length(between_y_idx) > 0L) {
        sample_clz_zz_res <- vector("list", ncluster_sizes)
        sample_clz_zz_xx <- vector("list", ncluster_sizes)
        sample_clz_zz_b <- vector("list", ncluster_sizes)

        sample_clz_yz_res <- vector("list", ncluster_sizes)
        sample_clz_yz_xx <- vector("list", ncluster_sizes)
        sample_clz_yres_xz <- vector("list", ncluster_sizes)
        sample_clz_xwzres <- vector("list", ncluster_sizes)
      }
      for (clz in seq_len(ncluster_sizes)) {
        # cluster size
        nj <- cluster_sizes[clz]
        nj_idx <- which(cluster_size == nj)

        # Y2
        y2_clz <- y2[nj_idx, y1_idx, drop = FALSE]
        if (length(x1_idx) > 0L) {
          exo2_clz <- cbind(1, y2[nj_idx, x1_idx, drop = FALSE])
        } else {
          exo2_clz <- matrix(1, nrow(y2_clz), 1L)
        }
        xx_clz <- crossprod(exo2_clz)
        clz_y2_b <- try(solve(xx_clz, crossprod(exo2_clz, y2_clz)),
          silent = TRUE
        )
        if (inherits(clz_y2_b, "try-error")) {
          clz_y2_b <-
            MASS::ginv(xx_clz) %*% crossprod(exo2_clz, y2_clz)
        }
        clz_y2_hat <- exo2_clz %*% clz_y2_b
        clz_y2_res <- y2_clz - clz_y2_hat
        sample_clz_y2_b[[clz]] <- clz_y2_b
        sample_clz_y2_res[[clz]] <- crossprod(clz_y2_res)
        sample_clz_y2_xx[[clz]] <- crossprod(exo2_clz)

        # Z
        if (length(between_y_idx) > 0L) {
          z_clz_z <- y2[nj_idx, between_y_idx, drop = FALSE]
          if (length(between_x_idx) > 0L) {
            exo_clz_z <- cbind(
              1,
              y2[nj_idx, between_x_idx, drop = FALSE]
            )
          } else {
            exo_clz_z <- matrix(1, nrow(z_clz_z), 1L)
          }
          zz_clz <- crossprod(exo_clz_z)
          clz_zz_b <- try(
            solve(
              zz_clz,
              crossprod(exo_clz_z, z_clz_z)
            ),
            silent = TRUE
          )
          if (inherits(clz_zz_b, "try-error")) {
            clz_zz_b <-
              MASS::ginv(zz_clz) %*% crossprod(exo_clz_z, z_clz_z)
          }
          clz_z_hat <- exo_clz_z %*% clz_zz_b
          clz_z_res <- z_clz_z - clz_z_hat
          sample_clz_zz_b[[clz]] <- clz_zz_b
          sample_clz_zz_res[[clz]] <- crossprod(clz_z_res)
          sample_clz_zz_xx[[clz]] <- crossprod(exo_clz_z)

          sample_clz_yz_res[[clz]] <- crossprod(clz_y2_res, clz_z_res)
          sample_clz_yz_xx[[clz]] <- crossprod(exo2_clz, exo_clz_z)
          sample_clz_yres_xz[[clz]] <- crossprod(clz_y2_res, exo_clz_z)
          sample_clz_xwzres[[clz]] <- crossprod(exo2_clz, clz_z_res)
        }
      } # clz
    } # conditional.x

    ylp[[l]] <- list(
      Y1Y1 = y1y1,
      Y2 = y2, s = s, S.b = s_b, S.PW.start = s_pw_start,
      Sigma.W = s_w, Mu.W = mu_w,
      Sigma.B = sigma_b, Mu.B = mu_b,
      Mu.B.start = mu_b_start, loglik.x = loglik_x,
      mean.d = mean_d, cov.d = cov_d
    )

    # if conditional, add more stuff
    if (conditional_x) {
      if (length(between_y_idx) > 0L) {
        extra <- list(
          sample.wb = sample_wb,
          sample.YYres.wb1 = sample_yyres_wb1,
          sample.XX.wb1 = sample_xx_wb1,
          sample.wb2 = sample_wb2,
          sample.YYres.wb2 = sample_yyres_wb2,
          sample.YresX.wb2 = sample_yres_x_wb2,
          sample.XX.wb2 = sample_xx_wb2,
          sample.clz.Y2.res = sample_clz_y2_res,
          sample.clz.Y2.XX = sample_clz_y2_xx,
          sample.clz.Y2.B = sample_clz_y2_b,
          sample.clz.ZZ.res = sample_clz_zz_res,
          sample.clz.ZZ.XX = sample_clz_zz_xx,
          sample.clz.ZZ.B = sample_clz_zz_b,
          sample.clz.YZ.res = sample_clz_yz_res,
          sample.clz.YZ.XX = sample_clz_yz_xx,
          sample.clz.YresXZ = sample_clz_yres_xz, # zero?
          sample.clz.XWZres = sample_clz_xwzres
        )
      } else {
        extra <- list(
          sample.wb = sample_wb,
          sample.YYres.wb1 = sample_yyres_wb1,
          sample.XX.wb1 = sample_xx_wb1,
          sample.wb2 = sample_wb2,
          sample.YYres.wb2 = sample_yyres_wb2,
          sample.YresX.wb2 = sample_yres_x_wb2,
          sample.XX.wb2 = sample_xx_wb2,
          sample.clz.Y2.res = sample_clz_y2_res,
          sample.clz.Y2.XX = sample_clz_y2_xx,
          sample.clz.Y2.B = sample_clz_y2_b
        )
      }
      ylp[[l]] <- c(ylp[[l]], extra)
    }
  } # l

  ylp
}

# check if the partable is complete/consistent
# for example, we may have added intercepts/variances (user = 0), fixed to zero
lav_pt_check <- function(partable, categorical = FALSE) {
  check <- TRUE

  # check for empty table - or should we WARN?
  if (length(partable$lhs) == 0) {
    return(check)
  }

  # get observed/latent variables
  ov_names <- lav_pt_vnames(partable, "ov.nox") # no need to specify exo??
  lv_names <- lav_pt_vnames(partable, "lv")
  lv_names_c <- lav_pt_vnames(partable, "lv.composite")
  lv_names_noc <- lv_names[!lv_names %in% lv_names_c]
  all_names <- c(ov_names, lv_names_noc)
  ov_names_ord <- lav_pt_vnames(partable, "ov.ord")

  nlevels <- lav_pt_nlevels(partable)

  # composites: within a single block, an indicator may not serve a composite
  # (<~) and a common factor (=~) at the same time (each indicator relates to
  # one construct only; see Schamberger et al.). Catch this here with a clear
  # message, otherwise it crashes later (non-positive-definite Sigma / NA in
  # loglik). NOTE: the check is per-block, because in a multilevel (or
  # multigroup) model the same observed variable may legitimately be a composite
  # indicator in one block and a common-factor indicator in another.
  if (length(lv_names_c) > 0L) {
    blk <- if (!is.null(partable$block)) partable$block else
      rep(1L, length(partable$lhs))
    for (b in unique(blk)) {
      cind <- unique(partable$rhs[partable$op == "<~" & blk == b])
      find <- unique(partable$rhs[partable$op == "=~" & blk == b])
      both <- cind[cind %in% find]
      if (length(both) > 0L) {
        lav_msg_stop(gettextf(
          "indicator(s) used by both a composite (<~) and a common factor (=~) %s: %s",
          "in the same block", lav_msg_view(both)))
      }
    }
  }

  # if categorical, we should have some ov.names.ord
  if (categorical && length(ov_names_ord) == 0L) {
    check <- FALSE
    lav_msg_warn(gettext("parameter table does not contain thresholds"))
  }

  # we should have a (residual) variance for *each* ov/lv
  # note: if lav_model_pt() has been used, this is always TRUE
  var_idx <- which(partable$op == "~~" &
    partable$lhs == partable$rhs & !partable$lhs %in% lv_names_c)
  missing_idx <- which(is.na(match(all_names, partable$lhs[var_idx])))
  if (length(missing_idx) > 0L) {
    check <- FALSE
    lav_msg_warn(gettextf(
      "parameter table does not contain (residual) variances for
      one or more variables: %s",
      lav_msg_view(all_names[missing_idx])))
  }

  # meanstructure?
  meanstructure <- any(partable$op == "~1")

  # if meanstructure, check for missing intercepts
  # note if lav_model_pt() has been used, this is always TRUE
  if (meanstructure) {
    # we should have an intercept for *each* ov/lv
    int_idx <- which(partable$op == "~1")
    missing_idx <- which(is.na(match(all_names, partable$lhs[int_idx])))
    if (length(missing_idx) > 0L) {
      check <- FALSE
      lav_msg_warn(gettextf(
        "parameter table does not contain intercepts
        for one or more variables: %s",
        lav_msg_view(all_names[missing_idx])))
    }
  }

  # ok, now the 'real' checks

  # do we have added (residual) variances (user = 0) that are fixed to zero?
  # this is not necessarily problematic!
  # eg. in latent change score models
  # therefore, we do NOT give a warning

  # var.fixed <- which(partable$op == "~~" &
  #                   partable$lhs == partable$rhs &
  #                   partable$user == 0 &
  #                   partable$free == 0)
  # if(length(var.fixed) > 0L) {
  #    check <- FALSE
  #     if(warn) {
  #        warning("lavaan WARNING: missing (residual) variances are set to",
  #        " zero: [", paste(partable$lhs[var.fixed],  collapse = " "), "]")
  #    }
  # }

  # do we have added intercepts (user = 0) that are fixed to zero?
  # this is not necessarily problematic; perhaps only for
  # exogenous variables?
  ov_ind <- unique(partable$rhs[partable$op == "=~"])
  lv_names <- unique(partable$lhs[partable$op == "=~"])
  int_fixed <- which(partable$op == "~1" &
    partable$user == 0L &
    partable$free == 0L &
    partable$ustart == 0L &
    # ignore block/group 1 -- typically within level exo
    !(partable$block %% nlevels == 1L) &
    # do not include factors
    !partable$lhs %in% lv_names &
    # do not include ordered variables
    !partable$lhs %in% ov_names_ord &
    # do not include indicators
    !partable$lhs %in% ov_ind &
    # do not include lv composites
    !partable$lhs %in% lv_names_c)

  if (length(int_fixed) > 0L) {
    check <- FALSE
    lav_msg_warn(gettext("automatically added intercepts are set to zero:"),
                 lav_msg_view(partable$lhs[int_fixed]))
  }

  # return check code
  check
}

# YR 11 feb 2017: initial version

# given a parameter table (PT), extract a part of the model:
# eg.:
# - only the measurement model (with saturated latent variables)
# - only the structural part
# - a single measurement block
# ...

# YR 25 June 2021: - add.exo.cov = TRUE for structural model
#                  - fixed.x = FALSE/TRUE -> exo flags


# FIXME:
# - but fixed-to-zero covariances may not be present in PT...
# - if indicators are regressed on exogenous covariates, should we
#   add them here? (no for now, unless add.ind.predictors = TRUE)
# - new in 0.6-20: - check for 2nd, 3rd order lv.names...
#                  - allow for conditional.x (global SAM)
lav_pt_subset_mm <- function(pt_1 = NULL,
                             lv_names = NULL,
                             add_lv_cov = TRUE,
                             add_ind_predictors = FALSE,
                             add_idx = FALSE,
                             idx_only = FALSE,
                             add_exo_reg = FALSE,
                             ov_names_x = NULL) {
  # PT
  pt_1 <- as.data.frame(pt_1, stringsAsFactors = FALSE)

  # check if we have free.unrotated column (rotation!)
  if (!is.null(pt_1$free.unrotated)) {
    pt_1$free.orig <- pt_1$free
    pt_1$free <- pt_1$free.unrotated
    pt_1$est.unrotated <- NULL
    pt_1$free.unrotated <- NULL
    pt_1$est.std <- NULL
  }

  # lavpta
  lavpta <- lav_pt_attributes(pt_1)

  # nblocks
  nblocks <- lavpta$nblocks
  block_values <- lav_pt_block_values(pt_1)

  # lv.names: list with element per block
  if (is.null(lv_names)) {
    lv_names <- lavpta$vnames$lv.regular
  } else if (!is.list(lv_names)) {
    lv_names <- rep(list(lv_names), nblocks)
  }

  # if lv.names contains a higher-order latent variable,
  # add its (latent) indicators
  for (g in 1:nblocks) {
    if (length(lavpta$vnames$lv.ind[[g]]) == 0L) {
      next
    }
    # ALL lv names
    lv_names_1 <- lavpta$vnames$lv[[g]]

    # lv.names in this block
    this_lv <- lv_names[[g]]

    all_ind_are_observed_flag <- FALSE
    new_lv <- character(0L)
    while (!all_ind_are_observed_flag) {
      # check indicators of all this.lv
      rhs <- pt_1$rhs[which(pt_1$op == "=~" & pt_1$block == g &
                          pt_1$lhs %in% this_lv)]
      lv_idx <- which(rhs %in% lv_names_1)
      if (length(lv_idx) > 0L) {
        # add these 'new' ones to new.lv
        new_lv <- c(new_lv, rhs[lv_idx])
        this_lv <- rhs[lv_idx]
      } else {
        all_ind_are_observed_flag <- TRUE
      }
    }

    # update lv.names for this block
    lv_names[[g]] <- unique(c(lv_names[[g]], new_lv))
  }

  # keep rows idx
  keep_idx <- integer(0L)

  # remove not-needed measurement models
  for (g in 1:nblocks) {
    # indicators for latent variables we keep
    ind_idx <- which(pt_1$op == "=~" &
      pt_1$lhs %in% lv_names[[g]] &
      pt_1$block == block_values[g])
    ind <- pt_1$rhs[ind_idx]
    # ind_plabel <- pt_1$plabel[ind_idx]

    # keep =~
    keep_idx <- c(keep_idx, ind_idx)

    # new in 0.6-17: indicators regressed on predictors
    if (add_ind_predictors) {
      pred_idx <- which(pt_1$op == "~" &
        pt_1$lhs %in% ind &
        pt_1$block == block_values[g])
      extra <- unique(pt_1$rhs[pred_idx])
      keep_idx <- c(keep_idx, pred_idx)
      # add them to IND, so we include their variances/intercepts
      ind <- c(ind, extra)
    }

    # keep ~~
    ov_var_idx <- which(pt_1$op == "~~" &
      pt_1$lhs %in% ind &
      pt_1$rhs %in% ind &
      pt_1$block == block_values[g])
    keep_idx <- c(keep_idx, ov_var_idx)

    lv_var_idx <- which(pt_1$op == "~~" &
      pt_1$lhs %in% lv_names[[g]] &
      pt_1$rhs %in% lv_names[[g]] &
      pt_1$block == block_values[g])
    keep_idx <- c(keep_idx, lv_var_idx)

    # intercepts indicators
    ov_int_idx <- which(pt_1$op == "~1" &
      pt_1$lhs %in% ind &
      pt_1$block == block_values[g])
    keep_idx <- c(keep_idx, ov_int_idx)

    # intercepts latent variables
    lv_int_idx <- which(pt_1$op == "~1" &
      pt_1$lhs %in% lv_names[[g]] &
      pt_1$block == block_values[g])
    keep_idx <- c(keep_idx, lv_int_idx)

    # thresholds
    th_idx <- which(pt_1$op == "|" &
      pt_1$lhs %in% ind &
      pt_1$block == block_values[g])
    keep_idx <- c(keep_idx, th_idx)

    # scaling factors
    sc_idx <- which(pt_1$op == "~*~" &
      pt_1$lhs %in% ind &
      pt_1$block == block_values[g])
    keep_idx <- c(keep_idx, sc_idx)

    # defined/constraints
    if (any(pt_1$op %in% c("==", "<", ">", ":="))) {
      # get the 'id' numbers and the labels involved in def/constraints
      pt2 <- pt_1
      pt2$free <- pt_1$id # use 'id' numbers instead of 'free' indices
      id <- lav_pt_con_label_id(pt2, def = TRUE)
      label <- names(id)

      # what are the row indices that we currently keep?
      free_id <- pt_1$id[keep_idx]
    }

    # defined parameters
    def_idx <- which(pt_1$op == ":=")
    if (length(def_idx) > 0L) {
      def_keep <- logical(length(def_idx))
      for (def in seq_along(def_idx)) {
        # rhs
        rhs_labels <- all.vars(as.formula(paste(
          "~",
          pt_1[def_idx[def], "rhs"]
        )))
        if (length(rhs_labels) > 0L) {
          # par id
          rhs_freeid <- id[match(rhs_labels, label)]

          # keep?
          if (all(rhs_freeid %in% free_id)) {
            def_keep[def] <- TRUE
          }
        } else { # only constants?
          def_keep[def] <- TRUE
        }
      }
      keep_idx <- c(keep_idx, def_idx[def_keep])
      # add 'id' numbers of := definitions that we keep
      free_id <- c(free_id, pt_1$id[def_idx[def_keep]])
    }

    # (in)equality constraints
    con_idx <- which(pt_1$op %in% c("==", "<", ">"))
    if (length(con_idx) > 0L) {
      con_keep <- logical(length(con_idx))
      for (con in seq_along(con_idx)) {
        lhs_keep <- FALSE
        rhs_keep <- FALSE

        # lhs
        lhs_labels <- all.vars(as.formula(paste(
          "~",
          pt_1[con_idx[con], "lhs"]
        )))
        if (length(lhs_labels) > 0L) {
          # par id
          lhs_freeid <- id[match(lhs_labels, label)]

          # keep?
          if (all(lhs_freeid %in% free_id)) {
            lhs_keep <- TRUE
          }
        } else {
          lhs_keep <- TRUE
        }

        # rhs
        rhs_labels <- all.vars(as.formula(paste(
          "~",
          pt_1[con_idx[con], "rhs"]
        )))
        if (length(rhs_labels) > 0L) {
          # par id
          rhs_freeid <- id[match(rhs_labels, label)]

          # keep?
          if (all(rhs_freeid %in% free_id)) {
            rhs_keep <- TRUE
          }
        } else {
          rhs_keep <- TRUE
        }

        if (lhs_keep && rhs_keep) {
          con_keep[con] <- TRUE
        }
      }

      keep_idx <- c(keep_idx, con_idx[con_keep])
    } # con
  } # block

  # remove any duplicated (only with higher-order factors)
  keep_idx <- keep_idx[!duplicated(keep_idx)]

  if (idx_only) {
    return(keep_idx)
  }

  pt_1 <- pt_1[keep_idx, , drop = FALSE]

  # check if we have enough indicators?
  # TODO

  # add covariances among latent variables?
  if (add_lv_cov) {
    pt_1 <- lav_pt_add_lv_cov(
      pt_1 = pt_1,
      lv_names = lv_names
    )
  }

  # add saturated regressions of the latent variables on the exogenous
  # covariates? (conditional.x = TRUE measurement blocks)
  if (add_exo_reg && length(unlist(ov_names_x)) > 0L) {
    pt_1 <- lav_pt_add_exo_reg(
      pt_1 = pt_1,
      lv_names = lv_names,
      ov_names_x = ov_names_x
    )
  }

  # clean up
  pt_1 <- lav_pt_complete(pt_1)

  if (add_idx) {
    attr(pt_1, "idx") <- keep_idx
  }

  pt_1
}

# add (free, saturated) regressions of the latent variables on the exogenous
# covariates -- used for the measurement blocks when conditional.x = TRUE
# (categorical data): the blocks must be estimated on the same conditional
# scale (Var(y*|x) = 1) as the joint model, so every block latent variable is
# regressed on ALL exogenous covariates (consistent with the h1 model, which
# conditions on all of them). For higher-order blocks only the first-order
# factors (those with observed indicators) are regressed on x: the top-factor
# slopes would not be separately identified from the first-order ones, and
# regressing the first-order factors is the saturated parameterization.
lav_pt_add_exo_reg <- function(pt_1, lv_names = NULL, ov_names_x = NULL) {
  # PT
  if (!is.data.frame(pt_1)) {
    pt_1 <- as.data.frame(pt_1, stringsAsFactors = FALSE)
  }

  # lavpta
  lavpta <- lav_pt_attributes(pt_1)

  # nblocks
  nblocks <- lavpta$nblocks
  block_values <- lav_pt_block_values(pt_1)

  # lv.names: list with element per block
  if (is.null(lv_names)) {
    lv_names <- lavpta$vnames$lv.regular
  } else if (!is.list(lv_names)) {
    lv_names <- rep(list(lv_names), nblocks)
  }
  if (!is.list(ov_names_x)) {
    ov_names_x <- rep(list(ov_names_x), nblocks)
  }

  for (b in seq_len(nblocks)) {
    if (length(ov_names_x[[b]]) == 0L) {
      next
    }
    for (lv in lv_names[[b]]) {
      if (!lv %in% lavpta$vnames$lv.regular[[b]]) {
        next
      }
      # skip higher-order factors (their indicators are latent variables)
      ind <- pt_1$rhs[pt_1$op == "=~" & pt_1$lhs == lv &
                      pt_1$block == block_values[b]]
      if (any(ind %in% lavpta$vnames$lv[[b]])) {
        next
      }
      for (x in ov_names_x[[b]]) {
        # already present?
        if (any(pt_1$op == "~" & pt_1$lhs == lv & pt_1$rhs == x &
                pt_1$block == block_values[b])) {
          next
        }
        add <- list(
          lhs = lv,
          op = "~",
          rhs = x,
          user = 3L,
          free = max(pt_1$free) + 1L,
          exo = 1L,
          block = b
        )
        # add group column
        if (!is.null(pt_1$group)) {
          add$group <- unique(pt_1$block[pt_1$block == b])
        }
        # add level column
        if (!is.null(pt_1$level)) {
          add$level <- unique(pt_1$level[pt_1$block == b])
        }
        # add lower column
        if (!is.null(pt_1$lower)) {
          add$lower <- as.numeric(-Inf)
        }
        # add upper column
        if (!is.null(pt_1$upper)) {
          add$upper <- as.numeric(+Inf)
        }
        pt_1 <- lav_pt_add(pt_1, add = add)
      }
    }
  }

  pt_1
}

# NOTE: only within same level
lav_pt_add_lv_cov <- function(pt_1, lv_names = NULL) {
  # PT
  if (!is.data.frame(pt_1)) {
    pt_1 <- as.data.frame(pt_1, stringsAsFactors = FALSE)
  }

  # lavpta
  lavpta <- lav_pt_attributes(pt_1)

  # nblocks
  nblocks <- lavpta$nblocks
  block_values <- lav_pt_block_values(pt_1)

  # lv.names: list with element per block
  if (is.null(lv_names)) {
    lv_names <- lavpta$vnames$lv.regular
  } else if (!is.list(lv_names)) {
    lv_names <- rep(list(lv_names), nblocks)
  }

  # check for higher-order models (new in 0.6-20)
  for (b in seq_len(nblocks)) {
    if (length(lavpta$vnames$lv.ind[[b]]) > 0L) {
      ind_idx <- which(lv_names[[b]] %in% lavpta$vnames$lv.ind[[b]])
      if (length(ind_idx) > 0L) {
        lv_names[[b]] <- lv_names[[b]][-ind_idx]
      }
    }
  }

  # remove lv.names if not present at same level/block
  if (nblocks > 1L) {
    for (b in seq_len(nblocks)) {
      rm_idx <- which(!lv_names[[b]] %in% lavpta$vnames$lv.regular[[b]])
      if (length(rm_idx) > 0L) {
        lv_names[[b]] <- lv_names[[b]][-rm_idx]
      }
    } # b
  }

  # add covariances among latent variables
  for (b in seq_len(nblocks)) {
    if (length(lv_names[[b]]) > 1L) {
      tmp <- utils::combn(lv_names[[b]], 2L)
      for (i in seq_len(ncol(tmp))) {
        # already present?
        cov1_idx <- which(pt_1$op == "~~" &
          pt_1$block == block_values[b] &
          pt_1$lhs == tmp[1, i] & pt_1$rhs == tmp[2, i])
        cov2_idx <- which(pt_1$op == "~~" &
          pt_1$block == block_values[b] &
          pt_1$lhs == tmp[2, i] & pt_1$rhs == tmp[1, i])

        # if not, add
        if (length(c(cov1_idx, cov2_idx)) == 0L) {
          add <- list(
            lhs = tmp[1, i],
            op = "~~",
            rhs = tmp[2, i],
            user = 3L,
            free = max(pt_1$free) + 1L,
            block = b
          )
          # add group column
          if (!is.null(pt_1$group)) {
            add$group <- unique(pt_1$block[pt_1$block == b])
          }
          # add level column
          if (!is.null(pt_1$level)) {
            add$level <- unique(pt_1$level[pt_1$block == b])
          }
          # add lower column
          if (!is.null(pt_1$lower)) {
            add$lower <- as.numeric(-Inf)
          }
          # add upper column
          if (!is.null(pt_1$upper)) {
            add$upper <- as.numeric(+Inf)
          }
          pt_1 <- lav_pt_add(pt_1, add = add)
        }
      }
    } # lv.names
  } # blocks

  pt_1
}

# this function takes a 'full' SEM (measurement models + structural part)
# and returns only the structural part
#
# - what to do if we have no regressions among the latent variables?
#   -> we return all covariances among the latent variables
#
lav_pt_subset_sm <- function(pt_1 = NULL,
                                                 add_idx = FALSE,
                                                 idx_only = FALSE,
                                                 add_exo_cov = FALSE,
                                                 fixed_x = FALSE,
                                                 conditional_x = FALSE,
                                                 free_fixed_var = FALSE,
                                                 meanstructure = FALSE) {
  # PT
  pt_1 <- as.data.frame(pt_1, stringsAsFactors = FALSE)

  # remove any EFA related information -- new in 0.6-18
  if (!is.null(pt_1$efa)) {
    pt_1$efa <- NULL
    pt_1$est.unrotated <- NULL
    seven_idx <- which(pt_1$user == 7L & pt_1$op == "~~")
    if (length(seven_idx) > 0L) {
      pt_1$user[seven_idx] <- 0L
      pt_1$free[seven_idx] <- 1L
      pt_1$ustart[seven_idx] <- as.numeric(NA)
      pt_1$est[seven_idx] <- pt_1$est.std[seven_idx]
    }
    pt_1$est.std <- NULL
  }

  # lavpta
  lavpta <- lav_pt_attributes(pt_1)

  # nblocks
  nblocks <- lavpta$nblocks
  block_values <- lav_pt_block_values(pt_1)

  # eqs.names
  # eqs_x_names <- lavpta$vnames$eqs.x
  # eqs_y_names <- lavpta$vnames$eqs.y
  lv_names <- lavpta$vnames$lv.regular

  # keep rows idx
  keep_idx <- integer(0L)

  # remove not-needed measurement models
  for (g in 1:nblocks) {

    # eqs.names
    eqs_names <- unique(c(
      lavpta$vnames$eqs.x[[g]],
      lavpta$vnames$eqs.y[[g]]
    ))
    if (length(eqs_names) == 0L) { # no structural model
      # only the latent variances/covariances are 'structural'; in the
      # higher-order setting, the lower-order factors (lv.ind) belong to
      # the measurement part
      eqs_names <- lv_names[[g]]
      lv_ind <- lavpta$vnames$lv.ind[[g]]
      if (length(lv_ind) > 0L) {
        eqs_names <- eqs_names[!eqs_names %in% lv_ind]
      }
    }
    # all.names <- unique(c(
    #   eqs.names,
    #   lavpta$vnames$lv.regular[[g]]
    # ))

    # regressions
    reg_idx <- which(pt_1$op == "~" & pt_1$block == block_values[g] &
      pt_1$lhs %in% eqs_names &
      pt_1$rhs %in% eqs_names)

    # the variances
    var_idx <- which(pt_1$op == "~~" & pt_1$block == block_values[g] &
      pt_1$lhs %in% eqs_names &
      pt_1$rhs %in% eqs_names &
      pt_1$lhs == pt_1$rhs)

    # optionally covariances (exo!)
    cov_idx <- which(pt_1$op == "~~" & pt_1$block == block_values[g] &
      pt_1$lhs %in% eqs_names &
      pt_1$rhs %in% eqs_names &
      pt_1$lhs != pt_1$rhs)

    # means/intercepts
    int_idx <- which(pt_1$op == "~1" & pt_1$block == block_values[g] &
      pt_1$lhs %in% eqs_names)

    keep_idx <- c(
      keep_idx, reg_idx, var_idx, cov_idx, int_idx
    )

    # defined/constraints
    if (any(pt_1$op %in% c("==", "<", ">", ":="))) {
      # get the 'id' numbers and the labels involved in def/constraints
      pt2 <- pt_1
      pt2$free <- pt_1$id # use 'id' numbers instead of 'free' indices
      id <- lav_pt_con_label_id(pt2, def = TRUE)
      label <- names(id)

      # what are the row indices that we currently keep?
      free_id <- pt_1$id[keep_idx]
    }

    # defined parameters
    def_idx <- which(pt_1$op == ":=")
    if (length(def_idx) > 0L) {
      def_keep <- logical(length(def_idx))
      for (def in seq_along(def_idx)) {
        # rhs
        rhs_labels <- all.vars(as.formula(paste(
          "~",
          pt_1[def_idx[def], "rhs"]
        )))
        if (length(rhs_labels) > 0L) {
          # par id
          rhs_freeid <- id[match(rhs_labels, label)]

          # keep?
          if (all(rhs_freeid %in% free_id)) {
            def_keep[def] <- TRUE
          }
        } else { # only constants?
          def_keep[def] <- TRUE
        }
      }
      keep_idx <- c(keep_idx, def_idx[def_keep])
      # add 'id' numbers of := definitions that we keep
      free_id <- c(free_id, pt_1$id[def_idx[def_keep]])
    }

    # (in)equality constraints
    con_idx <- which(pt_1$op %in% c("==", "<", ">"))
    if (length(con_idx) > 0L) {
      con_keep <- logical(length(con_idx))
      for (con in seq_along(con_idx)) {
        lhs_keep <- FALSE
        rhs_keep <- FALSE

        # lhs
        lhs_labels <- all.vars(as.formula(paste(
          "~",
          pt_1[con_idx[con], "lhs"]
        )))
        if (length(lhs_labels) > 0L) {
          # par id
          lhs_freeid <- id[match(lhs_labels, label)]

          # keep?
          if (all(lhs_freeid %in% free_id)) {
            lhs_keep <- TRUE
          }
        } else {
          lhs_keep <- TRUE
        }

        # rhs
        rhs_labels <- all.vars(as.formula(paste(
          "~",
          pt_1[con_idx[con], "rhs"]
        )))
        if (length(rhs_labels) > 0L) {
          # par id
          rhs_freeid <- id[match(rhs_labels, label)]

          # keep?
          if (all(rhs_freeid %in% free_id)) {
            rhs_keep <- TRUE
          }
        } else {
          rhs_keep <- TRUE
        }

        if (lhs_keep && rhs_keep) {
          con_keep[con] <- TRUE
        }
      }

      keep_idx <- c(keep_idx, con_idx[con_keep])
    } # con
  } # block

  if (idx_only) {
    return(keep_idx)
  }

  pt_1 <- pt_1[keep_idx, , drop = FALSE]

  # add any missing covariances among exogenous variables
  if (add_exo_cov) {
    if (conditional_x) {
      pt_1 <- lav_pt_add_exo_cov(pt_1, ov_names_x = lavpta$vnames$ov.x)
    } else {
      pt_1 <- lav_pt_add_exo_cov(pt_1)
    }
  }

  # if meanstructure, 'free' user=0 intercepts
  if (meanstructure) {
    int_idx <- which(pt_1$op == "~1" & pt_1$user == 0L & pt_1$free == 0L)
    if (conditional_x) {
      # the moments of the exogenous covariates are not modeled under
      # conditional.x: keep their intercepts fixed (exo)
      int_idx <- int_idx[!pt_1$lhs[int_idx] %in% unlist(lavpta$vnames$ov.x)]
    }
    if (length(int_idx) > 0L) {
      pt_1$free[int_idx] <- max(pt_1$free) + seq_along(int_idx)
      pt_1$ustart[int_idx] <- as.numeric(NA)
      pt_1$user[int_idx] <- 3L
    }
  }

  # if fixed.x = FALSE, remove all remaining (free) exo=1 elements
  if (!fixed_x) {
    exo_idx <- which(pt_1$exo != 0L)
    if (length(exo_idx) > 0L) {
      pt_1$exo[exo_idx] <- 0L
      pt_1$user[exo_idx] <- 3L
      pt_1$free[exo_idx] <- max(pt_1$free) + seq_along(exo_idx)
    }

  } else if (conditional_x) {
    # don't change exo column!
  } else {
    # first, wipe out exo column
    exo_idx <- which(pt_1$exo != 0L)
    if (length(exo_idx) > 0L) {
      pt_1$exo[exo_idx] <- 0L
    }

    # keep ov.x as in global model!
    for (g in 1:nblocks) {
      # ov.names.x <- lav_pt_vnames(PT,
      #   type = "ov.x",
      #   block = block.values[g]
      # )
      ov_names_x <- lavpta$vnames$ov.x[[g]]
      if (length(ov_names_x) == 0L) {
        next
      }

      # 1. variances/covariances
      exo_var_idx <- which(
        pt_1$op == "~~" &
          pt_1$block == block_values[g] &
          pt_1$rhs %in% ov_names_x &
          pt_1$lhs %in% ov_names_x &
          pt_1$user %in% c(0L, 3L)
      )
      if (length(exo_var_idx) > 0L) {
        pt_1$ustart[exo_var_idx] <- as.numeric(NA) # to be overridden
        pt_1$free[exo_var_idx] <- 0L
        pt_1$exo[exo_var_idx] <- 1L
        pt_1$user[exo_var_idx] <- 3L
      }

      # 2. intercepts
      exo_int_idx <- which(
        pt_1$op == "~1" &
          pt_1$block == block_values[g] &
          pt_1$lhs %in% ov_names_x &
          pt_1$user == 0L
      )
      if (length(exo_int_idx) > 0L) {
        pt_1$ustart[exo_int_idx] <- as.numeric(NA) # to be overridden
        pt_1$free[exo_int_idx] <- 0L
        pt_1$exo[exo_int_idx] <- 1L
        pt_1$user[exo_int_idx] <- 3L
      }
    } # blocks
  } # fixed.x

  # if conditional.x, check if we have 'additional' ov.x variables
  # that were not ov.x in the global model
  if (conditional_x) {
    for (b in 1:nblocks) {
      global_ov_x <- lavpta$vnames$ov.x[[b]]
      local_ov_x <- lav_pt_vnames(pt_1, type = "ov.x",
        block = block_values[b]
      )
      extra_idx <- which(!local_ov_x %in% global_ov_x)
      if (length(extra_idx) > 0L) {
        extra_ov_names <- local_ov_x[extra_idx]
        for (i in seq_along(extra_idx)) {
          add <- list(
            lhs = extra_ov_names[i],
            op = "~",
            rhs = global_ov_x[1],
            user = 3L,
            free = 0L,
            block = b,
            ustart = 0,
            exo = 1L
          )
          # add group column
          if (!is.null(pt_1$group)) {
            add$group <- unique(pt_1$block[pt_1$block == b])
          }
          # add level column
          if (!is.null(pt_1$level)) {
            add$level <- unique(pt_1$level[pt_1$block == b])
          }
          # add lower column
          if (!is.null(pt_1$lower)) {
            add$lower <- as.numeric(-Inf)
          }
          # add upper column
          if (!is.null(pt_1$upper)) {
            add$upper <- as.numeric(+Inf)
          }
          pt_1 <- lav_pt_add(pt_1, add = add)
        } # i
      } # extra.idx
    } # b
  } # conditional.x

  # if free.fixed.var, free up all 'fixed (to unity)' variances
  if (free_fixed_var) {
    fixed_var_idx <- which(pt_1$op == "~~" & pt_1$lhs == pt_1$rhs &
                       pt_1$free == 0 & pt_1$user == 0L & pt_1$ustart == 1)
    if (length(fixed_var_idx) > 0L) {
      pt_1$free[fixed_var_idx] <- max(pt_1$free) + seq_along(fixed_var_idx)
      pt_1$ustart[fixed_var_idx] <- as.numeric(NA)
      # reset any stale bounds: when the parent partable carries bounds
      # columns (explicit inequality constraints, bounds=), a FIXED
      # parameter has lower == upper == its fixed value; keeping that
      # lower bound after freeing pins the (residual) variance at 1
      if (!is.null(pt_1$lower)) {
        pt_1$lower[fixed_var_idx] <- -Inf
      }
      if (!is.null(pt_1$upper)) {
        pt_1$upper[fixed_var_idx] <- +Inf
      }
    }
  }

  # clean up
  pt_1 <- lav_pt_complete(pt_1)

  if (add_idx) {
    attr(pt_1, "idx") <- keep_idx
  }

  pt_1
}

# NOTE: only within same level
lav_pt_add_exo_cov <- function(pt_1, ov_names_x = NULL) {
  # PT
  pt_1 <- as.data.frame(pt_1, stringsAsFactors = FALSE)

  # lavpta
  lavpta <- lav_pt_attributes(pt_1)

  # nblocks
  nblocks <- lavpta$nblocks
  block_values <- lav_pt_block_values(pt_1)


  # ov.names.x: list with element per block
  if (is.null(ov_names_x)) {
    ov_names_x <- lavpta$vnames$ov.x
  } else if (!is.list(ov_names_x)) {
    ov_names_x <- rep(list(ov_names_x), nblocks)
  }

  # remove ov.names.x if not present at same level/block
  if (nblocks > 1L) {
    for (b in seq_len(nblocks)) {
      rm_idx <- which(!ov_names_x[[b]] %in% lavpta$vnames$ov.x[[b]])
      if (length(rm_idx) > 0L) {
        ov_names_x[[b]] <- ov_names_x[[b]][-rm_idx]
      }
    } # b
  }

  # add covariances among latent variables
  for (b in seq_len(nblocks)) {
    if (length(ov_names_x[[b]]) > 1L) {
      tmp <- utils::combn(ov_names_x[[b]], 2L)
      for (i in seq_len(ncol(tmp))) {
        # already present?
        cov1_idx <- which(pt_1$op == "~~" &
          pt_1$block == block_values[b] &
          pt_1$lhs == tmp[1, i] & pt_1$rhs == tmp[2, i])
        cov2_idx <- which(pt_1$op == "~~" &
          pt_1$block == block_values[b] &
          pt_1$lhs == tmp[2, i] & pt_1$rhs == tmp[1, i])

        # if not, add
        if (length(c(cov1_idx, cov2_idx)) == 0L) {
          add <- list(
            lhs = tmp[1, i],
            op = "~~",
            rhs = tmp[2, i],
            user = 3L,
            free = max(pt_1$free) + 1L,
            block = b,
            ustart = as.numeric(NA)
          )
          # add group column
          if (!is.null(pt_1$group)) {
            add$group <- unique(pt_1$block[pt_1$block == b])
          }
          # add level column
          if (!is.null(pt_1$level)) {
            add$level <- unique(pt_1$level[pt_1$block == b])
          }
          # add lower column
          if (!is.null(pt_1$lower)) {
            add$lower <- as.numeric(-Inf)
          }
          # add upper column
          if (!is.null(pt_1$upper)) {
            add$upper <- as.numeric(+Inf)
          }
          pt_1 <- lav_pt_add(pt_1, add = add)
        }
      }
    } # ov.names.x
  } # blocks

  pt_1
}

# rotation algorithms
#
# YR  3 April 2019 -- gradient projection algorithm
# YR 21 April 2019 -- pairwise rotation algorithm
# YR 11 May   2020 -- order.idx is done in rotation matrix
#                     (suggested by Florian Scharf)
# YR 02 June  2024 -- add group argument, so target and target_mask can
#                     be a list

# main function to rotate a single matrix 'A'
lav_mat_rotate <- function(a = NULL, # original matrix
                              orthogonal = FALSE, # default is oblique
                              method = "geomin", # default rot method
                              method_args = list(
                                geomin_epsilon = 0.01,
                                orthomax_gamma = 1,
                                cf_gamma = 0,
                                oblimin_gamma = 0,
                                promax_kappa = 4,
                                target = matrix(0, 0, 0),
                                target_mask = matrix(0, 0, 0)
                              ),
                              init_rot = NULL, # initial rotation matrix
                              init_rot_check = TRUE, # check if init ROT is ok
                              rstarts = 100L, # number of random starts
                              row_weights = "default", # row weighting
                              std_ov = FALSE, # rescale ov
                              ov_var = NULL, # ov variances
                              algorithm = "gpa", # rotation algorithm
                              reflect = TRUE, # reflect sign
                              order_lv_by = "index", # how to order the lv's
                              gpa_tol = 0.00001, # stopping tol gpa
                              tol = 1e-07, # stopping tol others
                              keep_rep = FALSE, # store replications
                              max_iter = 10000L, # max gpa iterations
                group = 1L) { # group number

  # check A
  if (!inherits(a, "matrix")) {
    lav_msg_stop(gettext("A does not seem to be a matrix"))
  }

  p <- nrow(a)
  m <- ncol(a)
  if (m < 2L) { # single dimension
    res <- list(
      mm_lambda = a, PHI = matrix(1, 1, 1), ROT = matrix(1, 1, 1),
      orthogonal = orthogonal, method = "none",
      method.args = list(), row_weights = "none",
      algorithm = "none", iter = 0L, converged = TRUE,
      method.value = 0
    )
    return(res)
  }

  # method
  method <- tolower(method)

  # if promax, skip everything, then call promax() later
  if (method == "promax") {
    # orig.algorithm <- algorithm
    # orig.rstarts <- rstarts

    algorithm <- "none"
    rstarts <- 0L
    init_rot <- NULL
    rot <- diag(m)
  }

  # check init.ROT
  if (!is.null(init_rot) && init_rot_check) {
    if (!inherits(init_rot, "matrix")) {
      lav_msg_stop(gettext("init.ROT does not seem to be a matrix"))
    }
    if (nrow(init_rot) != m) {
      lav_msg_stop(gettextf(
        "nrow(init.ROT) = %1$s does not equal ncol(A) = %2$s",
        nrow(init_rot), m))
    }
    if (nrow(init_rot) != ncol(init_rot)) {
      lav_msg_stop(gettextf(
        "nrow(init.ROT) = %1$s does not equal ncol(init.ROT) = %2$s",
        nrow(init_rot), ncol(init_rot)))
    }
    # rotation matrix?
    if (!lav_mat_rotate_check(init_rot, orthogonal = orthogonal)) {
      lav_msg_stop(gettext("init.ROT does not look like a rotation matrix"))
    }
  }

  # determine method function name + validate target/target_mask
  tmp_resolve <- lav_mat_rotate_resolve_method(
    method = method, method_args = method_args, p = p, m = m, group = group
  )
  method <- tmp_resolve$method
  method_fname <- tmp_resolve$method_fname
  method_args <- tmp_resolve$method_args

  # set orthogonal option
  if (missing(orthogonal)) {
    # the default is oblique, except for varimax, entropy and a few others
    if (method %in% c(
      "varimax", "entropy", "mccammon",
      "tandem1", "tandem2"
    )) {
      orthogonal <- TRUE
    } else {
      orthogonal <- FALSE
    }
  } else {
    if (!orthogonal && method %in% c(
      "varimax", "entropy", "mccammon",
      "tandem1", "tandem2"
    )) {
      lav_msg_warn(gettextf(
        "rotation method %s may not work with oblique rotation.",
        dQuote(method)
      ))
    }
  }

  # set row_weights
  row_weights <- tolower(row_weights)
  if (row_weights == "default") {
    # the default is "none", except for varimax
    if (method %in% c("varimax", "promax")) {
      row_weights <- "kaiser"
    } else {
      row_weights <- "none"
    }
  }

  # check algorithm
  algorithm <- tolower(algorithm)
  if (algorithm %in% c("gpa", "pairwise", "none")) {
    # nothing to do
  } else {
    lav_msg_stop(gettext("algorithm must be gpa or pairwise"))
  }



  # 1. compute row weigths

  # 1.a cov -> cor?
  if (std_ov) {
    a <- a * 1 / sqrt(ov_var)
  }

  if (row_weights == "none") {
    weights <- rep(1.0, p)
  } else if (row_weights == "kaiser") {
    weights <- lav_mat_rotate_kaiser_weights(a)
  } else if (row_weights == "cureton-mulaik") {
    weights <- lav_mat_rotate_cm_weights(a)
  } else {
    lav_msg_stop(gettext("row_weights can be none, kaiser or cureton-mulaik"))
  }
  a <- a * weights


  # 2. rotate

  # multiple random starts?
  if (rstarts > 0L) {
    rep_1 <- sapply(seq_len(rstarts), function(rep) {
      # random start (always orthogonal)
      init_rot <- lav_mat_rotate_gen(m = m, orthogonal = TRUE)
      # init.ROT <- lav_mat_rotate_gen(M = M, orthogonal = orthogonal)

      if (lav_verbose()) {
        cat("\n")
        cat("rstart = ", sprintf("%4d", rep), " start:\n")
      }


      # choose rotation algorithm
      if (algorithm == "gpa") {
        rot <- lav_mat_rotate_gpa(
          a = a, orthogonal = orthogonal,
          init_rot = init_rot,
          method_fname = method_fname,
          method_args = method_args,
          gpa_tol = gpa_tol,
          max_iter = max_iter
        )
        info <- attr(rot, "info")
        attr(rot, "info") <- NULL
        res <- c(info$method.value, lav_mat_vec(rot))
      } else if (algorithm == "pairwise") {
        rot <- lav_mat_rotate_pairwise(
          a = a,
          orthogonal = orthogonal,
          init_rot = init_rot,
          method_fname = method_fname,
          method_args = method_args,
          tol = tol,
          max_iter = max_iter
        )
        info <- attr(rot, "info")
        attr(rot, "info") <- NULL
        res <- c(info$method.value, lav_mat_vec(rot))
      }

      if (lav_verbose()) {
        cat(
          "rstart = ", sprintf("%4d", rep),
          " end; current crit = ", sprintf("%17.15f", res[1]), "\n"
        )
      }
      res
    })
    best_idx <- which.min(rep_1[1, ])
    rot <- matrix(rep_1[-1, best_idx], nrow = m, ncol = m)
    if (keep_rep) {
      info <- list(method.value = rep_1[1, best_idx], REP = rep_1)
    } else {
      info <- list(method.value = rep_1[1, best_idx])
    }
  } else if (algorithm != "none") {
    # initial rotation matrix
    if (is.null(init_rot)) {
      init_rot <- diag(m)
    }

    # Gradient Projection Algorithm
    if (algorithm == "gpa") {
      rot <- lav_mat_rotate_gpa(
        a = a, orthogonal = orthogonal,
        init_rot = init_rot,
        method_fname = method_fname,
        method_args = method_args,
        gpa_tol = gpa_tol,
        max_iter = max_iter
      )
    } else if (algorithm == "pairwise") {
      rot <- lav_mat_rotate_pairwise(
        a = a,
        orthogonal = orthogonal,
        init_rot = init_rot,
        method_fname = method_fname,
        method_args = method_args,
        tol = tol,
        max_iter = max_iter
      )
    }
    info <- attr(rot, "info")
    attr(rot, "info") <- NULL
  }

  # final rotation
  if (orthogonal) {
    # mm_lambda <- A %*% solve(t(ROT))
    # note: when ROT is orthogonal, solve(t(ROT)) == ROT
    mm_lambda <- a %*% rot
    phi <- diag(ncol(mm_lambda)) # correlation matrix == I
  } else {
    # mm_lambda <- A %*% solve(t(ROT))
    mm_lambda <- t(solve(rot, t(a)))
    phi <- crossprod(rot) # correlation matrix
  }

  # 3. undo row weighting
  mm_lambda <- mm_lambda / weights

  # here, after re-weighted, we run promax if needed
  if (method == "promax") {
    lambda_orig <- mm_lambda

    # first, run 'classic' varimax using varimax() from the stats package
    # we split varimax from promax, so we can control the normalize flag
    normalize_flag <- row_weights == "kaiser"
    xx <- stats::varimax(x = mm_lambda, normalize = normalize_flag)

    # promax
    kappa <- method_args$promax_kappa
    out <- lav_mat_rotate_promax(
      x = xx$loadings, m = kappa,
      varimax_rot = xx$rotmat
    )
    mm_lambda <- out$loadings
    phi <- solve(crossprod(out$rotmat))

    # compute 'ROT' to be compatible with GPa
    rott_inv <- solve(
      crossprod(lambda_orig),
      crossprod(lambda_orig, mm_lambda)
    )
    rot <- solve(t(rott_inv))

    info <- list(
      algorithm = "promax", iter = 0L, converged = TRUE,
      method.value = as.numeric(NA)
    )
  }

  # 3.b undo cov -> cor
  if (std_ov) {
    mm_lambda <- mm_lambda * sqrt(ov_var)
  }

  # 4.a reflect so that column sum is always positive
  if (reflect) {
    sum_1 <- colSums(mm_lambda)
    neg_idx <- which(sum_1 < 0)
    if (length(neg_idx) > 0L) {
      mm_lambda[, neg_idx] <- -1 * mm_lambda[, neg_idx, drop = FALSE]
      rot[, neg_idx] <- -1 * rot[, neg_idx, drop = FALSE]
      if (!orthogonal) {
        # recompute PHI
        phi <- crossprod(rot)
      }
    }
  }

  # 4.b reorder the columns
  order_idx <- lav_efa_order_idx(mm_lambda, order_lv_by)

  # do the same in PHI
  mm_lambda <- mm_lambda[, order_idx, drop = FALSE]
  phi <- phi[order_idx, order_idx, drop = FALSE]

  # new in 0.6-6, also do this in ROT, so we won't have to do this
  # again upstream
  rot <- rot[, order_idx, drop = FALSE]

  # 6. return results as a list
  res <- list(
    mm_lambda = mm_lambda, PHI = phi, ROT = rot, order.idx = order_idx,
    orthogonal = orthogonal, method = method,
    method_args = method_args, row_weights = row_weights
  )

  # add method info
  res <- c(res, info)

  res
}


# Gradient Projection Algorithm (Jennrich 2001, 2002)
#
# - this is a translation of the SAS PROC IML code presented in the Appendix
#   of Bernaards & Jennrich (2005)
# - as the orthogonal and oblique algorithm are so similar, they are
#   combined in a single function
# - the default is oblique rotation
#
lav_mat_rotate_gpa <- function(a = NULL, # original matrix
                                  orthogonal = FALSE, # default is oblique
                                  init_rot = NULL, # initial rotation
                                  method_fname = NULL, # criterion function
                                  method_args = list(), # optional method args
                                  gpa_tol = 0.00001,
                                  max_iter = 10000L) {
  # number of columns
  m <- ncol(a)

  # transpose of A (not needed for orthogonal)
  at <- t(a)

  # check init.ROT
  if (is.null(init_rot)) {
    rot <- diag(m)
  } else {
    rot <- init_rot
  }

  # set initial value of alpha to 1
  alpha <- 1

  # initial rotation
  if (orthogonal) {
    mm_lambda <- a %*% rot
  } else {
    mm_lambda <- t(solve(rot, at))
  }

  # using the current mm_lambda, evaluate the user-specified
  # rotation criterion; return Q (the criterion) and its gradient Gq
  q_1 <- do.call(
    method_fname,
    c(list(mm_lambda = mm_lambda), method_args, list(grad = TRUE))
  )
  gq <- attr(q_1, "grad")
  attr(q_1, "grad") <- NULL
  q_current <- q_1

  # compute gradient GRAD of f() at ROT from the gradient Gq of Q at mm_lambda
  # in a manner appropriate for orthogonal or oblique rotation
  if (orthogonal) {
    grad <- crossprod(a, gq)
  } else {
    grad <- -1 * solve(t(init_rot), crossprod(gq, mm_lambda))
  }

  # start iterations
  converged <- FALSE
  for (iter in seq_len(max_iter + 1L)) {
    # compute projection Gp of GRAD onto the linear manifold tangent at
    # ROT to the manifold of orthogonal or normal (for oblique) matrices
    #
    # this projection is zero if and only if ROT is a stationary point of
    # f() restricted to the orthogonal/normal matrices
    if (orthogonal) {
      mm <- crossprod(rot, grad)
      symm <- (mm + t(mm)) / 2
      gp <- grad - (rot %*% symm)
    } else {
      gp <- grad - t(t(rot) * colSums(rot * grad))
    }

    # check Frobenius norm of Gp
    frob <- sqrt(sum(gp * gp))

    # if verbose, print
    if (lav_verbose()) {
      cat(
        "  iter = ", sprintf("%4d", iter - 1),
        " Q = ", sprintf("%9.7f", q_current),
        " frob.log10 = ", sprintf("%10.7f", log10(frob)),
        " alpha = ", sprintf("%9.7f", alpha), "\n"
      )
    }

    if (frob < gpa_tol) {
      converged <- TRUE
      break
    }

    # update
    alpha <- 2 * alpha
    for (i in seq_len(1000)) { # make option?

      # step in the negative projected gradient direction
      # (note, the original algorithm in Jennrich 2001 used G, not Gp)
      x <- rot - alpha * gp

      if (orthogonal) {
        # use SVD to compute the projection ROTt of X onto the manifold
        # of orthogonal matrices
        svd_out <- svd(x)
        u <- svd_out$u
        v_1 <- svd_out$v
        rott <- u %*% t(v_1)
      } else {
        # compute the projection ROTt of X onto the manifold
        # of normal matrices
        v <- 1 / sqrt(apply(x^2, 2, sum))
        rott <- x %*% diag(v)
      }

      # rotate again
      if (orthogonal) {
        mm_lambda <- a %*% rott
      } else {
        mm_lambda <- t(solve(rott, at))
      }

      # evaluate criterion
      q_new <- do.call(method_fname, c(
        list(mm_lambda = mm_lambda),
        method_args, list(grad = TRUE)
      ))
      gq <- attr(q_new, "grad")
      attr(q_new, "grad") <- NULL

      # check stopping criterion
      if (q_new < q_current - 0.5 * frob * frob * alpha) {
        break
      } else {
        alpha <- alpha / 2
      }

      if (i == 1000) {
        lav_msg_warn(gettext("half-stepping failed in GPA"))
      }
    }

    # update
    rot <- rott
    q_current <- q_new

    if (orthogonal) {
      grad <- crossprod(a, gq)
    } else {
      grad <- -1 * solve(t(rot), crossprod(gq, mm_lambda))
    }
  } # iter

  # warn if no convergence
  if (!converged) {
    lav_msg_warn(gettextf(
      "GP rotation algorithm did not converge after %s iterations",
      max_iter
    ))
  }

  # algorithm information
  info <- list(
    algorithm = "gpa",
    iter = iter - 1L,
    converged = converged,
    method.value = q_current
  )

  attr(rot, "info") <- info

  rot
}


# pairwise rotation algorithm with direct line search
#
# based on Kaiser's (1959) algorithm and Jennrich and Sampson (1966) algorithm
# but to make it generic, a line search is used; inspired by Browne 2001
#
# - orthogonal: rotate one pair of columns (=plane) at a time
# - oblique: rotate 1 factor in one pair of columns (=plane) at a time
#            note: in the oblique case, (1,2) is not the same as (2,1)
# - BUT use optimize() to find the optimal angle (for each plane)
#   (see Browne, 2001, page 130)
# - repeat until the changes in the f() criterion are below tol
#

lav_mat_rotate_pairwise <- function(a = NULL, # original matrix
                                       orthogonal = FALSE,
                                       init_rot = NULL,
                                       method_fname = NULL, # crit function
                                       method_args = list(), # method args
                                       tol = 1e-8,
                                       max_iter = 1000L) {
  # number of columns
  m <- ncol(a)

  # initial mm_lambda + PHI
  if (is.null(init_rot)) {
    mm_lambda <- a
    if (!orthogonal) {
      phi <- diag(m)
    }
  } else {
    if (orthogonal) {
      mm_lambda <- a %*% init_rot
    } else {
      mm_lambda <- t(solve(init_rot, t(a)))
      phi <- crossprod(init_rot)
    }
  }

  # using the current mm_lambda, evaluate the user-specified
  # rotation criterion; return Q (the criterion) only
  q_current <- do.call(method_fname, c(
    list(mm_lambda = mm_lambda),
    method_args, list(grad = FALSE)
  ))

  # if verbose, print
  if (lav_verbose()) {
    cat(
      "  iter = ", sprintf("%4d", 0),
      " Q = ", sprintf("%13.11f", q_current), "\n"
    )
  }

  # plane combinations
  if (orthogonal) {
    plane <- utils::combn(m, 2)
  } else {
    tmp <- utils::combn(m, 2)
    plane <- cbind(tmp, tmp[c(2, 1), , drop = FALSE])
  }

  # define objective function -- orthogonal
  # note: 'a' is the *current* loading matrix for this plane sweep (passed in by
  # optimize() below), not the original unrotated matrix of the same name in the
  # enclosing function.
  objf_orth <- function(theta = 0, a = NULL, col1 = 0L, col2 = 0L) {
    # construct ROT
    rot <- lav_mat_givens_orth(m, col1, col2, theta)

    # rotate
    mm_lambda <- a %*% rot

    # evaluate criterion
    q_1 <- do.call(method_fname, c(
      list(mm_lambda = mm_lambda),
      method_args, list(grad = FALSE)
    ))

    q_1
  }

  # define objective function -- oblique
  objf_obliq <- function(delta = 0, a = NULL, col1 = 0L, col2 = 0L,
                         phi12 = 0) {
    # construct ROT
    rot <- lav_mat_givens_obliq(m, col1, col2, delta, phi12)

    # rotate
    mm_lambda <- a %*% rot

    # evaluate criterion
    q_1 <- do.call(method_fname, c(
      list(mm_lambda = mm_lambda),
      method_args, list(grad = FALSE)
    ))
    q_1
  }

  # start iterations
  converged <- FALSE
  q_old <- q_current
  for (iter in seq_len(max_iter)) {
    # rotate - one cycle
    for (pl in seq_len(ncol(plane))) {
      # choose plane
      col1 <- plane[1, pl]
      col2 <- plane[2, pl]

      # optimize
      if (orthogonal) {
        out <- optimize(
          f = objf_orth, interval = c(-pi / 4, +pi / 4),
          a = mm_lambda, col1 = col1, col2 = col2,
          maximum = FALSE, tol = .Machine$double.eps^0.25
        )
        # best rotation - for this plane
        theta <- out$minimum

        # construct ROT
        rot <- lav_mat_givens_orth(m, col1, col2, theta)
      } else {
        phi12 <- phi[col1, col2]
        out <- optimize(
          f = objf_obliq, interval = c(-1, +1),
          a = mm_lambda, col1 = col1, col2 = col2,
          phi12 = phi12,
          maximum = FALSE, tol = .Machine$double.eps^0.25
        )

        # best rotation - for this plane
        delta <- out$minimum

        # construct ROT (rot[col1, col1] == gamma, the PHI scaling factor)
        rot <- lav_mat_givens_obliq(m, col1, col2, delta, phi12)
        gamma <- rot[col1, col1]
      }

      # rotate
      mm_lambda <- mm_lambda %*% rot

      if (!orthogonal) {
        # rotate PHI
        phi[col1, ] <- (1 / gamma) * phi[col1, ] + (delta / gamma) * phi[col2, ]
        phi[, col1] <- phi[col1, ]
        phi[col1, col1] <- 1
      }
    } # all planes

    # check for convergence
    q_current <- do.call(method_fname, c(
      list(mm_lambda = mm_lambda),
      method_args, list(grad = FALSE)
    ))

    # absolute change in Q
    diff <- abs(q_old - q_current)

    # if verbose, print
    if (lav_verbose()) {
      cat(
        "  iter = ", sprintf("%4d", iter),
        " Q = ", sprintf("%13.11f", q_current),
        " change = ", sprintf("%13.11f", diff), "\n"
      )
    }

    if (diff < tol) {
      converged <- TRUE
      break
    } else {
      q_old <- q_current
    }
  } # iter

  # warn if no convergence
  if (!converged) {
    lav_msg_warn(gettextf(
      "pairwise rotation algorithm did not converge after %s iterations",
      max_iter
    ))
  }

  # compute final rotation matrix
  if (orthogonal) {
    rot <- solve(crossprod(a), crossprod(a, mm_lambda))
  } else {
    # to be compatible with GPa
    rott_inv <- solve(crossprod(a), crossprod(a, mm_lambda))
    rot <- solve(t(rott_inv))
  }

  # algorithm information
  info <- list(
    algorithm = "pairwise",
    iter = iter,
    converged = converged,
    method.value = q_current
  )

  attr(rot, "info") <- info

  rot
}

# the multivariate linear model using maximum likelihood

# 1) loglikelihood (from raw data, or sample statistics)
# 2) derivatives with respect to Beta, res.cov, vech(res.cov)
# 3) casewise scores with respect to Beta, vech(res.cov), Beta + vech(res.cov)
# 4) hessian Beta + vech(res.cov)
# 5) information h0 Beta + vech(res.cov)
#    5a: (unit)    expected information
#    5b: (unit)    observed information
#    5c: (unit) first.order information

# YR 24 Mar 2016: first version
# YR 20 Jan 2017: removed added 'N' in many equations, to be consistent with
#                 lav_mvnorm_*
# YR 18 Okt 2018: add 'information' functions, change arguments
#                 (X -> eXo, Sigma -> res.cov, Beta -> res.int + res.slopes)
# YR 03 Jul 2026: use shared kernels from lav_mvnorm_kernels.R

# 0. small helpers

# 0a. model-implied Beta (row 1 = intercepts, rows 2... = slopes)
lav_mvreg_beta1 <- function(beta_1 = NULL,
                            res_int = NULL,
                            res_slopes = NULL) {
  if (is.null(beta_1)) {
    beta_1 <- rbind(matrix(res_int, nrow = 1), t(res_slopes))
  }
  beta_1
}

# 0b. sample.xx = 1/N*crossprod(X1) (including intercept)
lav_mvreg_sample_xx <- function(sample_mean_x = NULL,
                                sample_cov_x = NULL) {
  rbind(
    cbind(1, matrix(sample_mean_x, nrow = 1, )),
    cbind(
      matrix(sample_mean_x, ncol = 1),
      sample_cov_x + tcrossprod(sample_mean_x)
    )
  )
}

# 1. loglikelihood

# 1a. input is raw data
lav_mvreg_loglik_data <- function(y = NULL,
                                  exo = NULL, # no intercept
                                  beta_1 = NULL,
                                  res_int = NULL,
                                  res_slopes = NULL,
                                  res_cov = NULL,
                                  casewise = FALSE,
                                  sinv_method = "eigen") {
  y <- unname(y)
  q_1 <- NCOL(y)
  n <- NROW(y)
  x <- cbind(1, unname(exo))

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  if (casewise) {
    log_2pi <- log(2 * pi)

    # invert res.cov
    if (sinv_method == "chol") {
      c_s <- chol(res_cov)
      ic_s <- backsolve(c_s, diag(q_1))
      logdet <- -2 * sum(log(diag(ic_s)))

      res <- y - x %*% beta_1
      dist_1 <- rowSums((res %*% ic_s)^2)
    } else {
      res_cov_inv <- lav_mat_sym_inverse(
        s = res_cov, logdet = TRUE,
        sinv_method = sinv_method
      )
      logdet <- attr(res_cov_inv, "logdet")

      res <- y - x %*% beta_1
      dist_1 <- rowSums(res %*% res_cov_inv * res)
    }

    loglik <- -(q_1 * log_2pi + logdet + dist_1) / 2
  } else {
    # invert res.cov
    res_cov_inv <- lav_mat_sym_inverse(
      s = res_cov, logdet = TRUE,
      sinv_method = sinv_method
    )
    logdet <- attr(res_cov_inv, "logdet")

    res <- y - x %*% beta_1
    # TOTAL <- TR( (Y - X%*%Beta) %*% res.cov.inv %*% t(Y - X%*%Beta) )
    total <- sum(rowSums(res %*% res_cov_inv * res))
    loglik <- -(n * q_1 / 2) * log(2 * pi) - (n / 2) * logdet - (1 / 2) * total
  }

  loglik
}


# 2b. input are sample statistics (res.int, res.slopes, res.cov, N) only
lav_mvreg_loglik_samp <- function(sample_res_int = NULL,
                                         sample_res_slopes = NULL,
                                         sample_res_cov = NULL,
                                         sample_mean_x = NULL,
                                         sample_cov_x = NULL,
                                         sample_nobs = NULL,
                                         beta_1 = NULL, # optional
                                         res_int = NULL,
                                         res_slopes = NULL,
                                         res_cov = NULL,
                                         sinv_method = "eigen",
                                         res_cov_inv = NULL) {
  q_1 <- NCOL(sample_res_cov)
  n <- sample_nobs
  # log_2pi <- log(2 * pi)

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # construct 'saturated' (sample-based) B
  sample_b <- rbind(matrix(sample_res_int, nrow = 1), t(sample_res_slopes))

  # construct sample.xx = 1/N*crossprod(X1) (including intercept)
  sample_xx <- lav_mvreg_sample_xx(sample_mean_x, sample_cov_x)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method, logdet = TRUE
  )
  logdet <- attr(res_cov_inv, "logdet")

  # tr(res.cov^{-1} %*% S)
  dist1 <- sum(res_cov_inv * sample_res_cov)

  # tr( res.cov^{-1} (B-beta)' X'X (B-beta)
  diff_1 <- sample_b - beta_1
  dist2 <- sum(res_cov_inv * crossprod(diff_1, sample_xx) %*% diff_1)

  loglik <- -(n / 2) * (q_1 * log(2 * pi) + logdet + dist1 + dist2)

  loglik
}




# 2. Derivatives

# 2a. derivative logl with respect to Beta (=intercepts and slopes)
lav_mvreg_dlogl_dbeta <- function(y = NULL,
                                  exo = NULL,
                                  beta_1 = NULL,
                                  res_int = NULL,
                                  res_slopes = NULL,
                                  res_cov = NULL,
                                  sinv_method = "eigen",
                                  res_cov_inv = NULL) {
  y <- unname(y)
  x <- cbind(1, unname(exo))

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # subtract 'X %*% Beta' from Y
  res <- y - x %*% beta_1

  # derivative
  dbeta <- as.numeric(t(x) %*% res %*% res_cov_inv)

  dbeta
}

# 2b: derivative logl with respect to res.cov (full matrix, ignoring symmetry)
lav_mvreg_dlogl_drescov <- function(y = NULL,
                                    exo = NULL,
                                    beta_1 = NULL,
                                    res_cov = NULL,
                                    res_int = NULL,
                                    res_slopes = NULL,
                                    sinv_method = "eigen",
                                    res_cov_inv = NULL) {
  y <- unname(y)
  n <- NROW(y)
  x <- cbind(1, unname(exo))

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # subtract 'X %*% Beta' from Y
  res <- y - x %*% beta_1

  # W.tilde
  w_tilde <- crossprod(res) / n

  # derivative
  dres_cov <- -(n / 2) * (res_cov_inv -
                         (res_cov_inv %*% w_tilde %*% res_cov_inv))

  dres_cov
}

# 2c: derivative logl with respect to vech(res.cov)
lav_mvreg_dlogl_dvechrescov <- function(y = NULL,
                                        exo = NULL,
                                        beta_1 = NULL,
                                        res_int = NULL,
                                        res_slopes = NULL,
                                        res_cov = NULL,
                                        sinv_method = "eigen",
                                        res_cov_inv = NULL) {
  dres_cov <- lav_mvreg_dlogl_drescov(
    y = y, exo = exo, beta_1 = beta_1,
    res_cov = res_cov, res_int = res_int, res_slopes = res_slopes,
    sinv_method = sinv_method, res_cov_inv = res_cov_inv
  )

  as.numeric(lav_mat_dup_pre(as.matrix(lav_mat_vec(dres_cov))))
}


# 3. Casewise scores

# 3a: casewise scores with respect to Beta (=intercepts and slopes)
#     column order: Y1_int, Y1_x1, Y1_x2, ...| Y2_int, Y2_x1, Y2_x2, ... |
lav_mvreg_sc_beta <- function(y = NULL,
                                  exo = NULL,
                                  beta_1 = NULL,
                                  res_int = NULL,
                                  res_slopes = NULL,
                                  res_cov = NULL,
                                  sinv_method = "eigen",
                                  res_cov_inv = NULL) {
  y <- unname(y)
  q_1 <- NCOL(y)
  x <- cbind(1, unname(exo))
  p <- NCOL(x)

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # subtract Mu
  res <- y - x %*% beta_1

  # post-multiply with res.cov.inv
  res <- res %*% res_cov_inv

  sc_beta <- x[, rep(1:p, times = q_1), drop = FALSE] *
    res[, rep(1:q_1, each = p), drop = FALSE]

  sc_beta
}


# 3b: casewise scores with respect to vech(res.cov)
lav_mvreg_sc_sigma <- function(y = NULL,
                                        exo = NULL,
                                        beta_1 = NULL,
                                        res_int = NULL,
                                        res_slopes = NULL,
                                        res_cov = NULL,
                                        sinv_method = "eigen",
                                        res_cov_inv = NULL) {
  y <- unname(y)
  x <- cbind(1, unname(exo))

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # subtract X %*% Beta
  res <- y - x %*% beta_1

  # score kernel
  lav_mvn_sc_kernel(yc = res, sigma_inv = res_cov_inv)$sigma
}


# 3c: casewise scores with respect to beta + vech(res.cov)
lav_mvreg_sc_beta_sigma <- function(y = NULL,
                                             exo = NULL,
                                             beta_1 = NULL,
                                             res_int = NULL,
                                             res_slopes = NULL,
                                             res_cov = NULL,
                                             sinv_method = "eigen",
                                             res_cov_inv = NULL) {
  y <- unname(y)
  q_1 <- NCOL(y)
  x <- cbind(1, unname(exo))
  p <- NCOL(x)

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # subtract X %*% Beta
  res <- y - x %*% beta_1

  # score kernel (res %*% res.cov.inv + scores wrt vech(res.cov))
  sc2 <- lav_mvn_sc_kernel(yc = res, sigma_inv = res_cov_inv)

  sc_beta <- x[, rep(1:p, times = q_1), drop = FALSE] *
    sc2$mu[, rep(1:q_1, each = p), drop = FALSE]

  cbind(sc_beta, sc2$sigma)
}

# 4. hessian of logl

# 4a. hessian logl Beta and vech(res.cov) from raw data
lav_mvreg_logl_hessian_data <- function(y = NULL,
                                        exo = NULL, # no int
                                        beta_1 = NULL, # int+slopes
                                        res_int = NULL,
                                        res_slopes = NULL,
                                        res_cov = NULL,
                                        res_cov_inv = NULL,
                                        sinv_method = "eigen") {
  # sample size
  n <- NROW(y)

  # observed information
  observed <- lav_mvreg_info_observed_data(
    y = y, exo = exo,
    beta_1 = beta_1, res_int = res_int, res_slopes = res_slopes,
    res_cov = res_cov, res_cov_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # hessian
  -n * observed
}

# 4b. hessian logl Beta and vech(res.cov) from samplestats
lav_mvreg_logl_hessian_samp <- function(
      sample_res_int = NULL,
      sample_res_slopes = NULL,
      sample_res_cov = NULL,
      sample_mean_x = NULL,
      sample_cov_x = NULL,
      sample_nobs = NULL,
      beta_1 = NULL, # int + slopes
      res_int = NULL, # intercepts only
      res_slopes = NULL, # slopes only (y x x)
      res_cov = NULL, # res.cov
      sinv_method = "eigen",
      res_cov_inv = NULL) {
  # sample size
  n <- sample_nobs

  # information
  observed <- lav_mvreg_information_observed_samplestats(
    sample_res_int = sample_res_int, sample_res_slopes = sample_res_slopes,
    sample_res_cov = sample_res_cov, sample_mean_x = sample_mean_x,
    sample_cov_x = sample_cov_x, beta_1 = beta_1, res_int = res_int,
    res_slopes = res_slopes, res_cov = res_cov, sinv_method = sinv_method,
    res_cov_inv = res_cov_inv
  )

  # hessian
  -n * observed
}


# Information h0

# 5a: unit expected information h0 Beta and vech(res.cov)
lav_mvreg_info_expected <- function(y = NULL, # not used
                                           exo = NULL, # not used
                                           sample_mean_x = NULL,
                                           sample_cov_x = NULL,
                                           sample_nobs = NULL,
                                           beta_1 = NULL, # not used
                                           res_int = NULL, # not used
                                           res_slopes = NULL, # not used
                                           res_cov = NULL,
                                           res_cov_inv = NULL,
                                           sinv_method = "eigen") {
  exo <- unname(exo)

  # res.cov.inv
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # N
  if (is.null(sample_nobs)) {
    sample_nobs <- nrow(exo) # hopefully not NULL either
  }

  # sample.mean.x + sample.cov.x
  if (is.null(sample_mean_x)) {
    sample_mean_x <- base::.colMeans(exo, m = NROW(exo), n = NCOL(exo))
  }
  if (is.null(sample_cov_x)) {
    sample_cov_x <- lav_mat_cov(exo)
  }

  # construct sample.xx = 1/N*crossprod(X1) (including intercept)
  sample_xx <- lav_mvreg_sample_xx(sample_mean_x, sample_cov_x)

  # expected information
  i11 <- res_cov_inv %x% sample_xx
  i22 <- lav_mvn_kron_dup_half(res_cov_inv)

  lav_mat_bdiag(i11, i22)
}

# 5b: unit observed information h0
lav_mvreg_info_observed_data <- function(y = NULL,
                                                exo = NULL, # no int
                                                beta_1 = NULL, # int+slopes
                                                res_int = NULL,
                                                res_slopes = NULL,
                                                res_cov = NULL,
                                                res_cov_inv = NULL,
                                                sinv_method = "eigen") {
  # create sample statistics
  y <- unname(y)
  x1 <- cbind(1, unname(exo))
  n <- NROW(y)

  # find 'B'
  qr_1 <- qr(x1)
  sample_b <- qr.coef(qr_1, y)

  sample_res_int <- as.numeric(sample_b[1, ])
  sample_res_slopes <- t(sample_b[-1, , drop = FALSE]) # transpose!
  sample_res_cov <- cov(qr.resid(qr_1, y)) * (n - 1) / n
  sample_mean_x <- base::.colMeans(exo, m = NROW(exo), n = NCOL(exo))
  sample_cov_x <- lav_mat_cov(exo)

  lav_mvreg_information_observed_samplestats(
    sample_res_int = sample_res_int,
    sample_res_slopes = sample_res_slopes, sample_res_cov = sample_res_cov,
    sample_mean_x = sample_mean_x, sample_cov_x = sample_cov_x,
    beta_1 = beta_1, res_int = res_int, res_slopes = res_slopes,
    res_cov = res_cov, sinv_method = sinv_method, res_cov_inv = res_cov_inv
  )
}


# 5b-bis: observed information h0 from sample statistics
lav_mvreg_information_observed_samplestats <-               # nolint
  function(sample_res_int = NULL,
           sample_res_slopes = NULL,
           sample_res_cov = NULL,
           sample_mean_x = NULL,
           sample_cov_x = NULL,
           beta_1 = NULL, # int + slopes
           res_int = NULL, # intercepts only
           res_slopes = NULL, # slopes only (y x x)
           res_cov = NULL, # res.cov
           sinv_method = "eigen",
           res_cov_inv = NULL) {
    # construct model-implied Beta
    beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

    # construct 'saturated' (sample-based) B
    sample_b <- rbind(matrix(sample_res_int, nrow = 1), t(sample_res_slopes))

    # construct sample.xx = 1/N*crossprod(X1) (including intercept)
    sample_xx <- lav_mvreg_sample_xx(sample_mean_x, sample_cov_x)

    # W.tilde = S + t(B - Beta) %*% (1/N)*X'X %*% (B - Beta)
    w_tilde <- (sample_res_cov +
      t(sample_b - beta_1) %*% sample_xx %*% (sample_b - beta_1))

    # res.cov.inv
    res_cov_inv <- lav_mvn_sigma_inv(
      sigma_1 = res_cov, sigma_inv = res_cov_inv,
      sinv_method = sinv_method
    )

    h11 <- res_cov_inv %x% sample_xx
    h21 <- lav_mat_dup_pre(res_cov_inv %x%
      (res_cov_inv %*% (crossprod(sample_b - beta_1, sample_xx))))
    h12 <- t(h21)

    aaa <- res_cov_inv %*% (2 * w_tilde - res_cov) %*% res_cov_inv
    h22 <- lav_mvn_kron_dup_half(res_cov_inv, aaa)

    out <- rbind(
      cbind(h11, h12),
      cbind(h21, h22)
    )
    out
  }


# 5c: unit first-order information h0
lav_mvreg_info_firstorder <- function(y = NULL,
                                             exo = NULL, # no int
                                             beta_1 = NULL, # int+slopes
                                             res_int = NULL,
                                             res_slopes = NULL,
                                             res_cov = NULL,
                                             res_cov_inv = NULL,
                                             sinv_method = "eigen") {
  n <- NROW(y)

  # scores
  sc <- lav_mvreg_sc_beta_sigma(
    y = y, exo = exo, beta_1 = beta_1,
    res_int = res_int, res_slopes = res_slopes, res_cov = res_cov,
    sinv_method = sinv_method, res_cov_inv = res_cov_inv
  )

  crossprod(sc) / n
}


# 6: inverted information h0

# 6a: inverted unit expected information h0 Beta and vech(res.cov)
#
# lav_mvreg_inverted_information_expected <- function(Y       = NULL, # unused!
# }

# FABIN = factor analysis by instrumental variables
# Hagglund 1982 (efa), 1986 (cfa)

lav_cfa_fabin2 <- function(s, marker_idx = NULL, lambda_nonzero_idx = NULL) {
  nvar <- ncol(s)
  nfac <- length(marker_idx)

  # overview of free/fixed
  mm_lambda <- matrix(0, nvar, nfac)
  mm_lambda[lambda_nonzero_idx] <- -1L

  lambda <- matrix(0, nvar, nfac)
  for (i in 1:nvar) {
    if (i %in% marker_idx) {
      lambda[i, marker_idx == i] <- 1.0
      next
    }
    free_idx <- mm_lambda[i, ] == -1L
    idx3 <- (1:nvar)[-c(i, marker_idx)]
    s23 <- s[i, idx3]
    fac_idx <- marker_idx[free_idx]

    if (length(fac_idx) == 1L) { # most common scenario in CFA
      s31 <- s13 <- s[idx3, fac_idx]
      lambda[i, free_idx] <- sum(s23 * s31) / sum(s13 * s13)
    } else {
      s31 <- s[idx3, fac_idx, drop = FALSE]
      s13 <- s[fac_idx, idx3, drop = FALSE]
      lambda[i, free_idx] <- solve(s13 %*% s31, drop(s23 %*% s31))
    }
  }

  lambda
}

lav_cfa_fabin3 <- function(s, marker_idx = NULL, lambda_nonzero_idx = NULL) {
  nvar <- ncol(s)
  nfac <- length(marker_idx)

  # overview of free/fixed
  mm_lambda <- matrix(0, nvar, nfac)
  mm_lambda[lambda_nonzero_idx] <- -1L

  s33_inv_1 <- try(solve(s[-marker_idx, -marker_idx, drop = FALSE]),
    silent = TRUE
  )
  if (inherits(s33_inv_1, "try-error")) {
    lav_msg_warn(gettext("fabin3 failed; switching to fabin2"))
    return(lav_cfa_fabin2(
      s = s, marker_idx = marker_idx,
      lambda_nonzero_idx = lambda_nonzero_idx
    ))
  }

  lambda <- matrix(0, nvar, nfac)
  rm3_idx <- 0L
  for (i in 1:nvar) {
    if (i %in% marker_idx) {
      lambda[i, marker_idx == i] <- 1.0
      next
    }
    free_idx <- mm_lambda[i, ] == -1L
    idx3 <- (1:nvar)[-c(i, marker_idx)]
    # s33 <- s[idx3, idx3, drop = FALSE]
    s23 <- s[i, idx3]
    fac_idx <- marker_idx[free_idx]
    rm3_idx <- rm3_idx + 1L
    # update inverse
    s33_inv <- lav_mat_sym_inverse_update(
      s_inv = s33_inv_1,
      rm_idx = rm3_idx
    )

    if (length(fac_idx) == 1L) { # most common scenario in CFA
      s31 <- s13 <- s[idx3, fac_idx]
      tmp <- s33_inv %*% s31 # or colSums(s33.inv * S31)
      lambda[i, free_idx] <- sum(s23 * tmp) / sum(s13 * tmp)
    } else {
      s31 <- s[idx3, fac_idx, drop = FALSE]
      s13 <- s[fac_idx, idx3, drop = FALSE]
      tmp <- s33_inv %*% s31
      # lambda[i, free.idx] <- ( s23 %*% solve(S33) %*% S31 %*%
      #                          solve(S13 %*% solve(S33) %*% S31) )
      lambda[i, free_idx] <- solve(s13 %*% tmp, drop(s23 %*% tmp))
    }
  }

  lambda
}

# internal function to be used inside lav_optim_noniter
# return 'x', the estimated vector of free parameters
lav_cfa_fabin_internal <- function(lavmodel = NULL, lavsamplestats = NULL,
                                   lavpartable = NULL,
                                   lavdata = NULL, lavoptions = NULL) {
  lavpta <- lav_pt_attributes(lavpartable)
  lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  # no structural part!
  if (any(lavpartable$op == "~")) {
    lav_msg_stop(gettext("FABIN estimator only available for CFA models"))
  }
  # no BETA matrix! (i.e., no higher-order factors)
  if (!is.null(lavmodel@GLIST$beta)) {
    lav_msg_stop(gettext(
    "FABIN estimator not available for models that require a BETA matrix"))
  }
  # no std.lv = TRUE for now
  if (lavoptions$std.lv) {
    lav_msg_stop(
      gettext("FABIN estimator not available if std.lv = TRUE"))
  }

  nblocks <- lav_pt_nblocks(lavpartable)
  stopifnot(nblocks == 1L) # for now
  b <- 1L
  sample_cov <- lavsamplestats@cov[[b]]
  nvar <- nrow(sample_cov)
  marker_idx <- lavpta$vidx$lv.marker[[b]]
  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")
  lambda_nonzero_idx <- lavmodel@m.free.idx[[lambda_idx]]
  # only diagonal THETA for now...
  # because if we have correlated residuals, we should remove the
  # corresponding variables as instruments before we estimate lambda...
  # (see MIIV)
  theta_idx <- which(names(lavmodel@GLIST) == "theta") # usually '2'
  m_theta <- lavmodel@m.free.idx[[theta_idx]]
  nondiag_idx <- m_theta[!m_theta %in% lav_mat_diag_idx(nvar)]
  if (length(nondiag_idx) > 0L) {
    lav_msg_warn(gettext(
      "this implementation of FABIN does not handle correlated residuals yet!"
      ))
  }

  # 1. estimate LAMBDA
  if (lavoptions$estimator == "FABIN2") {
    mm_lambda <- lav_cfa_fabin2(
      s = sample_cov, marker_idx = marker_idx,
      lambda_nonzero_idx = lambda_nonzero_idx
    )
  } else {
    mm_lambda <- lav_cfa_fabin3(
      s = sample_cov, marker_idx = marker_idx,
      lambda_nonzero_idx = lambda_nonzero_idx
    )
  }

  # 2. simple ULS method to get THETA and PSI (for now)
  gls_flag <- FALSE
  psi_mapping_ml_flag <- FALSE
  if (!is.null(lavoptions$estimator.args$thetapsi.method) &&
    lavoptions$estimator.args$thetapsi.method %in% c("GLS", "GLS.ML")) {
    gls_flag <- TRUE
  }
  if (!is.null(lavoptions$estimator.args$thetapsi.method) &&
    lavoptions$estimator.args$thetapsi.method %in% c("ULS.ML", "GLS.ML")) {
    psi_mapping_ml_flag <- TRUE
  }
  out <- lav_cfa_lambda2thetapsi(
    lambda = mm_lambda, s = sample_cov,
    s_inv = lavsamplestats@icov[[b]],
    gls = gls_flag,
    psi_mapping_ml = psi_mapping_ml_flag,
    nobs = lavsamplestats@ntotal
  )
  mm_theta <- diag(out$theta)
  mm_psi <- out$psi

  # 3. correlated residuals (if any) are just the difference between
  #    Sigma and S
  # if(length(nondiag.idx) > 0L) {
  #    Sigma <- LAMBDA %*% PSI %*% t(LAMBDA) + THETA
  #    THETA[nondiag.idx] <- (sample_cov - Sigma)[nondiag.idx]
  # }

  # store matrices in lavmodel@GLIST
  lavmodel@GLIST$lambda <- mm_lambda
  lavmodel@GLIST$theta <- mm_theta
  lavmodel@GLIST$psi <- mm_psi

  # extract free parameters only
  x <- lav_model_get_parameters(lavmodel)

  # apply bounds (if any)
  if (!is.null(lavpartable$lower)) {
    lower_x <- lavpartable$lower[lavpartable$free > 0]
    too_small_idx <- which(x < lower_x)
    if (length(too_small_idx) > 0L) {
      x[too_small_idx] <- lower_x[too_small_idx]
    }
  }
  if (!is.null(lavpartable$upper)) {
    upper_x <- lavpartable$upper[lavpartable$free > 0]
    too_large_idx <- which(x > upper_x)
    if (length(too_large_idx) > 0L) {
      x[too_large_idx] <- upper_x[too_large_idx]
    }
  }

  x
}

# lavSimulate: fit the *same* model, on simulated datasets
# YR - 4 July 2016: initial version
# YR - 15 Oct 2024: add iseed (similar to lavBootstrap)
# YR - 26 Oct 2024: rm pop.model, add est.true argument

lavSimulate <- function(model = NULL, # user model
                        data_function = lav_data_simulate_old,
                        data_function_args = list(),
                        est_true = NULL,
                        ndat = 1000L,
                        cmd = "sem",
                        ...,
                        store_slots = c("partable"),
                        fun = NULL,
                        show_progress = FALSE,
                        store_failed = FALSE,
                        parallel = c("no", "multicore", "snow"),
                        ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
                        cl = NULL,
                        iseed = NULL) {
  # dotdotdot
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)

  # dotdotdot for fit.pop
  dotdotdot_pop <- dotdotdot
  dotdotdot_pop$verbose <- FALSE
  dotdotdot_pop$debug <- FALSE
  dotdotdot_pop$data <- NULL
  dotdotdot_pop$sample_cov <- NULL

  # 'fit' model without data, check 'true' parameters
  fit_pop <- do.call(cmd,
    args = c(list(model = model), dotdotdot_pop)
  )

  # check est.true= argument
  stopifnot(!missing(est_true), is.numeric(est_true),
            length(fit_pop@ParTable$lhs) == length(est_true))
  fit_pop@ParTable$est <- est_true
  fit_pop@ParTable$start <- est_true

  # per default, use 'true' values as starting values
  if (is.null(dotdotdot$start)) {
    dotdotdot$start <- fit_pop
  }

  # honor fixed.x = TRUE: when the exogenous covariates are treated as fixed,
  # the analytic standard errors are *conditional* on these covariates (and
  # their sample (co)variances are fixed to the observed values). For the
  # simulation to match, the exogenous covariates must therefore be identical
  # in every replication. We hold the exogenous data fixed and regenerate the
  # endogenous variables conditionally: each replication is generated with the
  # user's data_function, but the endogenous part is re-expressed at the fixed
  # exogenous design, using the population regression of the (observed)
  # endogenous on the (observed) exogenous variables and the residuals of the
  # freshly generated dataset. This keeps the residual distribution intact
  # (e.g. non-normal) while fixing the exogenous (co)variances.
  fixed_x_flag <- lavInspect(fit_pop, "options")$fixed.x
  ov_x <- lavNames(fit_pop, "ov.x")
  if (isTRUE(fixed_x_flag) && length(ov_x) > 0L &&
      lavInspect(fit_pop, "ngroups") == 1L && is.function(data_function)) {
    ov_all <- lavNames(fit_pop)
    ov_y <- ov_all[!ov_all %in% ov_x]
    if (length(ov_y) > 0L) {
      # reproducible population/design generation
      if (!is.null(iseed)) {
        set.seed(iseed)
      }
      orig_data_function <- data_function
      orig_args <- data_function_args

      # estimate the (raw) population moments by stacking generated datasets
      d_one <- as.data.frame(do.call(orig_data_function, orig_args))
      if (!all(c(ov_x, ov_y) %in% names(d_one))) {
        lav_msg_stop(gettext(
          "fixed.x = TRUE: the data_function output must contain all observed
           variables of the model (as named columns)."))
      }
      n_one <- nrow(d_one)
      nrep_pop <- max(1L, min(200L, ceiling(1e5 / n_one)))
      if (nrep_pop > 1L) {
        d_more <- lapply(seq_len(nrep_pop - 1L), function(b) {
          as.data.frame(do.call(orig_data_function, orig_args))
        })
        d_big <- do.call("rbind", c(list(d_one), d_more))
      } else {
        d_big <- d_one
      }
      mu_pop <- colMeans(d_big[, ov_all, drop = FALSE])
      s_pop <- stats::cov(d_big[, ov_all, drop = FALSE])
      xi <- match(ov_x, ov_all)
      yi <- match(ov_y, ov_all)
      beta_yx <- s_pop[yi, xi, drop = FALSE] %*%
        solve(s_pop[xi, xi, drop = FALSE])
      mu_x <- mu_pop[xi]
      mu_y <- mu_pop[yi]

      # the fixed exogenous design (identical in every replication). Recolor
      # it so that its (co)variances match the population exogenous moments
      # exactly; otherwise this single realization would differ from the
      # population by O(1/sqrt(n)), and the conditional 'true' values would
      # not match est.true (which is defined at the population level).
      x_raw <- as.matrix(d_one[, ov_x, drop = FALSE])
      sx_pop <- s_pop[xi, xi, drop = FALSE]
      w_x <- solve(chol(stats::cov(x_raw))) %*% chol(sx_pop)
      x_fixed <- sweep(sweep(x_raw, 2L, colMeans(x_raw)) %*% w_x,
                       2L, mu_x, "+")
      colnames(x_fixed) <- ov_x

      data_function <- function(...) {
        d <- as.data.frame(do.call(orig_data_function, orig_args))
        if (nrow(d) != nrow(x_fixed)) {
          lav_msg_stop(gettext(
            "fixed.x = TRUE: the data_function must return a constant number
             of observations across replications."))
        }
        x_r <- as.matrix(d[, ov_x, drop = FALSE])
        y_r <- as.matrix(d[, ov_y, drop = FALSE])
        # residuals of the freshly generated endogenous data (under the
        # population regression)
        e_r <- sweep(y_r, 2L, mu_y) - sweep(x_r, 2L, mu_x) %*% t(beta_yx)
        # endogenous, conditional on the *fixed* exogenous design
        y_new <- sweep(sweep(x_fixed, 2L, mu_x) %*% t(beta_yx),
                       2L, mu_y, "+") + e_r
        d[, ov_y] <- y_new
        d[, ov_x] <- x_fixed
        d
      }
      data_function_args <- list() # original args are captured in the closure
    }
  }

  # no warnings during/after the simulations
  # add 'warn = FALSE' to args

  # generate simulations
  fit <- do.call("lavaanList", args = c(list(
    model = model,
    data_function = data_function,
    data_function_args = data_function_args,
    ndat = ndat, cmd = cmd,
    store_slots = store_slots, fun = fun,
    show_progress = show_progress,
    store_failed = store_failed,
    parallel = parallel, ncpus = ncpus,
    cl = cl, iseed = iseed
  ), dotdotdot))

  # flag this is a simulation
  fit@meta$lavSimulate <- TRUE

  # store 'true' parameters in meta$est.true
  fit@meta$est.true <- est_true

  fit
}

# level-specific fit measures for multilevel models, using the
# 'partially saturated model' approach (Ryu & West, 2009):
#
# - to assess the fit of the WITHIN part of the model, we fit an auxiliary
#   model where the between part is saturated (and vice versa); any misfit
#   of this partially saturated model can then be attributed to the
#   within part only
# - for the incremental fit indices (CFI/TLI), we also need level-specific
#   baseline models: the independence model at the target level, again
#   combined with a saturated model at the other level
#
# the (fitted) partially saturated models are stored in the @internal slot
# (internal$fit.by.level), and are used by fitMeasures(..., level = ) and
# by summary(..., fit.measures = TRUE)
#
# YR 11 July 2026 (initial version)

# generate 'saturated' rows for all blocks at the other level(s)
# (level != level.keep)
#
# important: any exogenous (fixed.x) covariates at the saturated level
# must REMAIN exogenous, because we re-use the sample statistics (and h1)
# slots of the original model: the saturated part is therefore saturated
# *conditionally* on the covariates (free y-(co)variances, free y~x
# slopes), just as if the user had written out a saturated level in the
# model syntax
lav_fit_by_level_sat_rows <- function(lavobject, level.keep = 1L) {
  lavdata <- lavobject@Data
  lavpta <- lavobject@pta
  lavh1 <- lavobject@h1
  nlevels <- lavdata@nlevels
  ngroups <- lavdata@ngroups
  fixed_x <- lavobject@Options$fixed.x

  lhs <- op <- rhs <- character(0L)
  block <- group <- level <- free <- exo <- integer(0L)
  ustart <- numeric(0L)

  add_rows <- function(lhs1, op1, rhs1, b, g, l, free1, exo1, ustart1) {
    n <- length(lhs1)
    lhs <<- c(lhs, lhs1)
    op <<- c(op, rep(op1, n))
    rhs <<- c(rhs, rhs1)
    block <<- c(block, rep(b, n))
    group <<- c(group, rep(g, n))
    level <<- c(level, rep(l, n))
    free <<- c(free, rep(free1, length.out = n))
    exo <<- c(exo, rep(exo1, length.out = n))
    ustart <<- c(ustart, rep(ustart1, length.out = n))
    invisible(NULL)
  }

  for (g in seq_len(ngroups)) {
    for (l in seq_len(nlevels)) {
      if (l == level.keep) {
        next
      }
      b <- (g - 1L) * nlevels + l

      ov_names <- lavpta$vnames$ov[[b]]
      if (fixed_x) {
        ov_x <- lavpta$vnames$ov.x[[b]]
      } else {
        ov_x <- character(0L)
      }
      ov_nox <- ov_names[!ov_names %in% ov_x]

      # h1 (unrestricted) moments of this block, for starting values
      sample_cov <- NULL
      sample_mean <- NULL
      if (length(lavh1) > 0L && !is.null(lavh1$implied$cov[[b]])) {
        sample_cov <- lavh1$implied$cov[[b]]
        sample_mean <- lavh1$implied$mean[[b]]
      }
      s_val <- function(v1, v2) {
        if (is.null(sample_cov)) {
          return(as.numeric(NA))
        }
        sample_cov[cbind(match(v1, ov_names), match(v2, ov_names))]
      }
      m_val <- function(v1) {
        if (is.null(sample_mean)) {
          return(as.numeric(NA))
        }
        sample_mean[match(v1, ov_names)]
      }

      # 1. variances of ALL observed variables first, in the same order
      # as lavpta$vnames$ov[[b]]: the first-appearance order of the ov's
      # in the parameter table must match the original model, because the
      # sample statistics (which are re-used) are stored in that order
      is_x <- ov_names %in% ov_x
      add_rows(ov_names, "~~", ov_names, b, g, l,
               free1 = ifelse(is_x, 0L, 1L),
               exo1 = ifelse(is_x, 1L, 0L),
               ustart1 = s_val(ov_names, ov_names))

      # 2. free covariances among the non-exogenous variables
      if (length(ov_nox) > 1L) {
        tmp <- utils::combn(ov_nox, 2L)
        add_rows(tmp[1L, ], "~~", tmp[2L, ], b, g, l,
                 free1 = 1L, exo1 = 0L,
                 ustart1 = s_val(tmp[1L, ], tmp[2L, ]))
      }

      # 3. free slopes of all y's on all (fixed) x's, plus the fixed
      # covariances among the exogenous variables
      if (length(ov_x) > 0L) {
        beta_start <- try({
          s_xx <- sample_cov[match(ov_x, ov_names),
                             match(ov_x, ov_names), drop = FALSE]
          s_xy <- sample_cov[match(ov_x, ov_names),
                             match(ov_nox, ov_names), drop = FALSE]
          solve(s_xx, s_xy) # nx x ny
        }, silent = TRUE)
        for (x in ov_x) {
          st <- rep(as.numeric(NA), length(ov_nox))
          if (!inherits(beta_start, "try-error") &&
              !is.null(sample_cov)) {
            st <- beta_start[match(x, ov_x), ]
          }
          add_rows(ov_nox, "~", rep(x, length(ov_nox)), b, g, l,
                   free1 = 1L, exo1 = 0L, ustart1 = st)
        }

        if (length(ov_x) > 1L) {
          tmp <- utils::combn(ov_x, 2L)
          add_rows(tmp[1L, ], "~~", tmp[2L, ], b, g, l,
                   free1 = 0L, exo1 = 1L,
                   ustart1 = s_val(tmp[1L, ], tmp[2L, ]))
        }
      }

      # 4. intercepts
      if (lavobject@Options$meanstructure) {
        if (l == 1L) {
          # within level: intercepts fixed to zero, except for
          # within-only variables (the same convention as lavaanify)
          b2 <- (g - 1L) * nlevels + 2L
          ov_l2 <- lavpta$vnames$ov[[b2]]
          within_only <- !ov_nox %in% ov_l2
          add_rows(ov_nox, "~1", rep("", length(ov_nox)), b, g, l,
                   free1 = ifelse(within_only, 1L, 0L), exo1 = 0L,
                   ustart1 = ifelse(within_only, m_val(ov_nox), 0))
        } else {
          add_rows(ov_nox, "~1", rep("", length(ov_nox)), b, g, l,
                   free1 = 1L, exo1 = 0L, ustart1 = m_val(ov_nox))
        }
        if (length(ov_x) > 0L) {
          add_rows(ov_x, "~1", rep("", length(ov_x)), b, g, l,
                   free1 = 0L, exo1 = 1L, ustart1 = m_val(ov_x))
        }
      }
    } # levels
  } # groups

  data.frame(
    lhs = lhs, op = op, rhs = rhs,
    user = rep(1L, length(lhs)),
    block = block, group = group, level = level,
    free = free, ustart = ustart, exo = exo,
    stringsAsFactors = FALSE
  )
}

# construct the parameter table for a partially saturated model:
# - the rows of the target level (level.keep) are taken from the user model
#   (or from the independence model, if independent = TRUE), while
# - the other level(s) are saturated
#
# returns NULL if the model contains constraints involving parameters from
# different levels (in that case, the misfit cannot be attributed to a
# single level)
lav_pt_partial_saturation <- function(lavobject, level.keep = 1L,
                                      independent = FALSE) {
  stopifnot(inherits(lavobject, "lavaan"))
  pt_user <- as.data.frame(lavobject@ParTable, stringsAsFactors = FALSE)

  # rows for the target level
  if (independent) {
    pt_keep <- as.data.frame(
      lav_pt_independence(lavobject = lavobject),
      stringsAsFactors = FALSE
    )
  } else {
    pt_keep <- pt_user
    # use the fitted values as starting values
    if (!is.null(pt_keep$est)) {
      pt_keep$ustart <- pt_keep$est
    }
  }

  # saturated rows for the other level(s); note: exogenous (fixed.x)
  # covariates remain exogenous (see lav_fit_by_level_sat_rows)
  pt_sat <- lav_fit_by_level_sat_rows(lavobject, level.keep = level.keep)

  # the columns we retain in the new parameter table
  core <- c(
    "lhs", "op", "rhs", "user", "block", "group", "level",
    "free", "ustart", "exo", "label", "plabel"
  )
  chr_core <- c("lhs", "op", "rhs", "label", "plabel")
  for (cc in core) {
    if (is.null(pt_keep[[cc]])) {
      if (cc == "group") {
        # single group: all (block > 0) rows belong to group 1
        pt_keep$group <- ifelse(pt_keep$block > 0L, 1L, 0L)
      } else {
        pt_keep[[cc]] <- if (cc %in% chr_core) "" else 0L
      }
    }
    if (is.null(pt_sat[[cc]])) {
      if (cc == "group") {
        pt_sat$group <- ifelse(pt_sat$block > 0L, 1L, 0L)
      } else {
        pt_sat[[cc]] <- if (cc %in% chr_core) "" else 0L
      }
    }
  }

  # the user parameter table may use labels in the group and level
  # columns (e.g., "g1", "within"), while the generated saturated and
  # independence partables use integer indices; map the indices to the
  # user values, so the combined partable is consistent
  group_values <- lav_pt_group_values(pt_user)
  level_values <- lav_pt_level_values(pt_user)
  level_keep_val <- level_values[level.keep]
  pt_sat$group <- group_values[pt_sat$group]
  pt_sat$level <- level_values[pt_sat$level]
  if (independent) {
    pt_keep$group <- group_values[pt_keep$group]
    pt_keep$level <- level_values[pt_keep$level]
  }

  con_ops <- c("==", "<", ">", ":=")
  keep_rows <- pt_keep[pt_keep$level %in% level_keep_val &
                       !pt_keep$op %in% con_ops, core, drop = FALSE]
  sat_rows <- pt_sat[!pt_sat$level %in% level_keep_val, core, drop = FALSE]

  # handle constraint/definition rows of the user model (they carry
  # level = 0): keep them if they only involve parameters of the target
  # level; silently drop them if they only involve parameters of the
  # other level(s); if they mix both, level-specific fit is not defined
  con_rows <- NULL
  if (!independent) {
    con_idx <- which(pt_user$op %in% con_ops)
    if (length(con_idx) > 0L) {
      mod_flag <- !pt_user$op %in% con_ops
      lab_all <- character(0L)
      if (!is.null(pt_user$plabel)) {
        lab_all <- pt_user$plabel
      } else {
        lab_all <- rep("", nrow(pt_user))
      }
      lab_user <- pt_user$label
      if (is.null(lab_user)) {
        lab_user <- rep("", nrow(pt_user))
      }
      keep_flag <- mod_flag & pt_user$level %in% level_keep_val
      drop_flag <- mod_flag & !pt_user$level %in% level_keep_val
      keep_syms <- c(lab_all[keep_flag], lab_user[keep_flag])
      drop_syms <- c(lab_all[drop_flag], lab_user[drop_flag])
      keep_syms <- keep_syms[nchar(keep_syms) > 0L]
      drop_syms <- drop_syms[nchar(drop_syms) > 0L]
      # a label may be attached to parameters at both levels (implying a
      # cross-level equality constraint)
      if (any(keep_syms %in% drop_syms)) {
        return(NULL)
      }

      keep_con <- logical(length(con_idx))
      for (ii in seq_along(con_idx)) {
        i <- con_idx[ii]
        refs <- character(0L)
        for (txt in c(pt_user$lhs[i], pt_user$rhs[i])) {
          if (nchar(txt) > 0L && is.na(suppressWarnings(as.numeric(txt)))) {
            refs <- c(refs, all.vars(parse(
              text = txt,
              keep.source = FALSE
            )))
          }
        }
        in_keep <- refs %in% keep_syms
        in_drop <- refs %in% drop_syms
        if (pt_user$op[i] == ":=") {
          if (length(refs) > 0L && all(in_keep)) {
            keep_con[ii] <- TRUE
            # the defined parameter becomes available for later rows
            keep_syms <- c(keep_syms, pt_user$lhs[i])
          } else {
            drop_syms <- c(drop_syms, pt_user$lhs[i])
          }
        } else {
          if (any(in_keep) && any(in_drop)) {
            # cross-level constraint
            return(NULL)
          }
          if (any(in_keep) && !any(in_drop)) {
            keep_con[ii] <- TRUE
          }
        }
      }
      if (any(keep_con)) {
        # pt_keep == pt_user here (independent = FALSE), but with all
        # core columns present
        con_rows <- pt_keep[con_idx[keep_con], core, drop = FALSE]
      }
    }
  }

  # combine, per block; constraint rows at the end
  pt_new <- rbind(keep_rows, sat_rows)
  pt_new <- pt_new[order(pt_new$block), , drop = FALSE]
  if (!is.null(con_rows)) {
    pt_new <- rbind(pt_new, con_rows)
  }

  # renumber the free parameters
  idx_free <- which(pt_new$free > 0L)
  pt_new$free <- 0L
  pt_new$free[idx_free] <- seq_along(idx_free)
  pt_new$id <- seq_len(nrow(pt_new))
  rownames(pt_new) <- NULL

  as.list(pt_new)
}

# fit the partially saturated models (one per level), together with their
# level-specific baseline models; returns a (possibly empty) named list
# with one fitted lavaan object per level; each object's @baseline slot
# contains the level-specific baseline, so that (for example) CFI/TLI are
# automatically computed with respect to the correct baseline model
lav_object_fit_by_level <- function(object) {
  out <- list()
  lavdata <- object@Data
  if (lavdata@nlevels < 2L) {
    return(out)
  }
  lavoptions <- object@Options

  # we need a converged fit and a test statistic
  if (!lavoptions$do.fit || !object@optim$converged ||
      object@optim$npar == 0L) {
    return(out)
  }
  if (length(lavoptions$test) == 1L && lavoptions$test == "none") {
    return(out)
  }
  if (length(object@h1) == 0L) {
    return(out)
  }
  # not (yet) for random slopes or efa blocks
  if (!is.null(object@ParTable$rv) &&
      any(nchar(object@ParTable$rv) > 0L)) {
    return(out)
  }
  if (!is.null(object@ParTable$efa) &&
      any(nchar(object@ParTable$efa) > 0L)) {
    return(out)
  }
  # not (yet) for categorical, conditional.x, correlation structures,
  # or free group weights
  if (object@Model@categorical || lavoptions$conditional.x ||
      object@Model@correlation || lavoptions$group.w.free) {
    return(out)
  }

  # options for the auxiliary fits
  lavoptions$se <- "none"
  lavoptions$h1 <- FALSE # already provided by sloth1
  lavoptions$baseline <- FALSE # we construct our own baselines
  lavoptions$fit.by.level <- FALSE # no recursion
  lavoptions$loglik <- TRUE
  lavoptions$implied <- TRUE
  lavoptions$check.start <- FALSE
  lavoptions$check.gradient <- FALSE
  lavoptions$check.post <- FALSE
  lavoptions$check.vcov <- FALSE
  lavoptions$bounds <- "none"
  lavoptions$optim.bounds <- list()
  lavoptions$start <- "default" # the ustart values are already good
  lavoptions$rstarts <- 0L
  lavoptions$do.fit <- TRUE
  if (lavoptions$optim.method != "noniter") {
    lavoptions$optim.method <- "nlminb"
  }

  level_names <- c("within", "between")

  for (l in seq_len(lavdata@nlevels)) {
    # partially saturated model for this level
    pt_ps <- lav_pt_partial_saturation(object,
      level.keep = l,
      independent = FALSE
    )
    if (is.null(pt_ps)) {
      lav_msg_warn(gettext(
        "level-specific fit measures are not available: the model contains
         (equality) constraints involving parameters from different
         levels."))
      return(list())
    }
    fit_ps <- try(
      lavaan(pt_ps,
        slot_options      = lavoptions,
        slot_sample_stats = object@SampleStats,
        slot_data         = lavdata,
        slot_cache        = object@Cache,
        sloth1            = object@h1
      ),
      silent = TRUE
    )
    if (inherits(fit_ps, "try-error") || !fit_ps@optim$converged) {
      lav_msg_warn(gettextf(
        "estimation of the partially saturated model for the %s level
         failed.", level_names[l]))
      next
    }

    # level-specific baseline model
    pt_bl <- lav_pt_partial_saturation(object,
      level.keep = l,
      independent = TRUE
    )
    fit_bl <- try(
      lavaan(pt_bl,
        slot_options      = lavoptions,
        slot_sample_stats = object@SampleStats,
        slot_data         = lavdata,
        slot_cache        = object@Cache,
        sloth1            = object@h1
      ),
      silent = TRUE
    )
    if (inherits(fit_bl, "try-error") || !fit_bl@optim$converged) {
      lav_msg_warn(gettextf(
        "estimation of the level-specific baseline model for the %s level
         failed.", level_names[l]))
    } else {
      fit_ps@baseline <- list(
        partable = fit_bl@ParTable,
        test = fit_bl@test
      )
    }

    # strip the slots that are shared with (and identical to) the
    # original object: the raw data can be huge, and storing two extra
    # copies would blow up the size of the fitted object; they are
    # re-injected by lav_fit_by_level_rehydrate() before use
    fit_ps@Data <- new("lavData")
    fit_ps@SampleStats <- new("lavSampleStats")
    fit_ps@Cache <- list()
    fit_ps@h1 <- list()

    out[[level_names[l]]] <- fit_ps
  }

  out
}

# re-inject the (stripped) slots that a stored partially saturated fit
# shares with the original object; the result is identical to the fit
# as it came out of lavaan(), because these slots were passed along
# unmodified (slot_data/slot_sample_stats/slot_cache/sloth1) when the
# partially saturated model was fitted
lav_fit_by_level_rehydrate <- function(fit_l, object) {
  fit_l@Data <- object@Data
  fit_l@SampleStats <- object@SampleStats
  fit_l@Cache <- object@Cache
  fit_l@h1 <- object@h1
  fit_l
}

# resolve a user-specified 'level' argument (1/2, "within"/"between", or
# the level label used in the model syntax) to "within" or "between"
lav_fit_by_level_target <- function(object, level = NULL) {
  if (object@Data@nlevels < 2L) {
    lav_msg_stop(gettext(
      "level-specific fit measures are only available for multilevel
       models."))
  }

  level_names <- c("within", "between")
  target <- NULL
  if (is.numeric(level) && length(level) == 1L &&
      level %in% seq_len(object@Data@nlevels)) {
    target <- level_names[as.integer(level)]
  } else if (is.character(level) && length(level) == 1L) {
    idx <- pmatch(tolower(level), level_names)
    if (is.na(idx)) {
      # perhaps the label of a level, as used in the model syntax
      idx <- pmatch(tolower(level), tolower(object@Data@level.label))
    }
    if (!is.na(idx)) {
      target <- level_names[idx]
    }
  }
  if (is.null(target)) {
    lav_msg_stop(gettextf(
      "level must be a single level number (1 or 2) or level name
       (%s).", lav_msg_view(c(level_names, object@Data@level.label),
                            log_sep = "or")))
  }

  target
}

# retrieve the partially saturated fit for a given level, computing it
# on the fly if it was not stored (e.g., fit.by.level = FALSE)
lav_fit_by_level_get <- function(object, level = NULL) {
  target <- lav_fit_by_level_target(object, level = level)

  fbl <- object@internal$fit.by.level
  if (is.null(fbl) || length(fbl) == 0L) {
    # not stored: compute on the fly
    fbl <- lav_object_fit_by_level(object)
  }
  fit_l <- NULL
  if (length(fbl) > 0L) {
    fit_l <- fbl[[target]]
  }
  if (is.null(fit_l)) {
    lav_msg_stop(gettextf(
      "no partially saturated model is available for the %s level.",
      target))
  }

  lav_fit_by_level_rehydrate(fit_l, object)
}

# build a small (measures x levels) matrix with level-specific fit
# measures, based on the STORED partially saturated fits (no on-the-fly
# computation); returns NULL if they are not available
# (used by summary(fit, fit.measures = TRUE))
lav_fit_by_level_fm <- function(object) {
  fbl <- object@internal$fit.by.level
  if (is.null(fbl) || length(fbl) == 0L) {
    return(NULL)
  }

  # scaled test statistics available?
  scaled <- any(!sapply(
    lapply(object@test, "[[", "scaling.factor"),
    is.null
  ))

  measures <- c("chisq", "df", "pvalue")
  if (scaled) {
    measures <- c(measures, "chisq.scaled", "df.scaled", "pvalue.scaled")
  }
  measures <- c(measures, "cfi", "tli", "rmsea")
  if (scaled) {
    measures <- c(measures, "cfi.robust", "tli.robust", "rmsea.robust")
  }

  level_names <- c("within", "between")
  srmr_names <- c("srmr_within", "srmr_between")
  res <- matrix(NA_real_,
    nrow = length(measures) + 1L,
    ncol = length(level_names)
  )
  rownames(res) <- c(measures, "srmr")
  colnames(res) <- level_names

  for (i in seq_along(level_names)) {
    fit_l <- fbl[[level_names[i]]]
    if (is.null(fit_l)) {
      next
    }
    fit_l <- lav_fit_by_level_rehydrate(fit_l, object)
    fm <- try(
      lav_fit(
        object = fit_l,
        fit_measures = list(fit.measures = c(measures, srmr_names[i])),
        output = "vector"
      ),
      silent = TRUE
    )
    if (inherits(fm, "try-error")) {
      next
    }
    idx <- match(c(measures, srmr_names[i]), names(fm))
    res[, i] <- unname(fm[idx])
  }

  # drop all-NA robust rows (e.g., when a baseline model failed)
  class(res) <- c("lavaan.matrix", "matrix", "array")
  res
}

# print the (measures x levels) matrix produced by lav_fit_by_level_fm()
# the layout mirrors lav_fitmeasures_print(): the Within column lines up
# with the regular (single/standard) column, and the Between column lines
# up with the second column that is used when a scaled test statistic is
# requested
lav_fit_by_level_print <- function(x, nd = 3L) {
  if (is.null(x) || !is.matrix(x) || nrow(x) == 0L) {
    return(invisible(x))
  }

  labels <- c(
    chisq          = gettext("Test statistic"),
    df             = gettext("Degrees of freedom"),
    pvalue         = gettext("P-value"),
    chisq.scaled   = gettext("Test statistic (scaled)"),
    df.scaled      = gettext("Degrees of freedom (scaled)"),
    pvalue.scaled  = gettext("P-value (scaled)"),
    cfi            = gettext("Comparative Fit Index (CFI)"),
    tli            = gettext("Tucker-Lewis Index (TLI)"),
    cfi.robust     = gettext("Robust Comparative Fit Index (CFI)"),
    tli.robust     = gettext("Robust Tucker-Lewis Index (TLI)"),
    rmsea          = gettext("RMSEA"),
    rmsea.robust   = gettext("Robust RMSEA"),
    srmr           = gettext("SRMR")
  )

  num_format <- paste("%", max(8L, nd + 5L), ".", nd, "f", sep = "")

  cat("\n")
  cat(gettext("Level-specific Fit Measures (other level saturated):"),
      "\n\n", sep = "")

  # container three columns (same layout as lav_fitmeasures_print)
  c1 <- ""
  c2 <- gettext("Within")
  c3 <- gettext("Between")

  fmt1 <- function(val, m) {
    if (is.na(val)) {
      return("NA")
    }
    if (m %in% c("df", "df.scaled") && val %% 1 == 0) {
      return(as.character(as.integer(val)))
    }
    sprintf(num_format, val)
  }

  for (r in seq_len(nrow(x))) {
    m <- rownames(x)[r]
    label <- labels[m]
    if (is.na(label)) {
      label <- m
    }
    c1 <- c(c1, label)
    c2 <- c(c2, fmt1(x[r, 1L], m))
    c3 <- c(c3, fmt1(x[r, 2L], m))
  }

  # format c1/c2/c3 (identical widths as lav_fitmeasures_print)
  c1 <- format(c1, width = 35L)
  c2 <- format(c2, width = 16L + max(0, (nd - 3L)) * 4L, justify = "right")
  c3 <- format(c3, width = 8L + nd, justify = "right")

  # create character matrix
  m <- cbind(c1, c2, c3, deparse.level = 0)
  colnames(m) <- rep("", ncol(m))
  rownames(m) <- rep(" ", nrow(m))

  # print
  write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)

  invisible(x)
}

# miscellaneous functions related to graphs
# often using an adjacency matrix as input
#
# collecting functions that were in lav_utils.R
# YR 15 Oct 2025

# check if adjacency matrix B represents an acyclic graph (DAG)
# B[i,j] != 0 means directed edge j -> i
# uses Kahn's topological sort: acyclic iff all nodes can be visited
lav_graph_is_acyclic <- function(m_b = NULL) {
  if (is.null(m_b) || !any(m_b != 0, na.rm = TRUE)) return(TRUE)
  nr <- nrow(m_b)
  indegree <- rowSums(m_b != 0)
  k <- 0L
  queue <- which(indegree == 0L)
  while (length(queue) > 0L) {
    v <- queue[1L]
    queue <- queue[-1L]
    k <- k + 1L
    for (u in which(m_b[, v] != 0)) {
      indegree[u] <- indegree[u] - 1L
      if (indegree[u] == 0L) queue <- c(queue, u)
    }
  }
  k == nr
}

# find index of 'ancestors' (predictors) for all nodes in a DAG
# given an adjacency matrix B (rows are y's, columns are x's)
#
# this (speedy!) version is written by Luc De Wilde
lav_graph_get_ancestors <- function(m_b = NULL) {
  m_b <- abs(m_b)
  nr <- nrow(m_b)
  out_env <- new.env(parent = emptyenv())

  # container to hold ancestor indices per node
  out_idx <- vector("list", length = nr)

  get_ancestors <- function(nr, callers) {
    if (any(callers == nr)) {
      lav_msg_warn(gettextf("Cycle detected for element nr %d !", nr))
      return(integer(0))
    }
    x <- get0(as.character(nr), envir = out_env, ifnotfound = NULL)
    if (!is.null(x)) return(x)
    retval <- integer(0L)
    x_direct_idx <- which(m_b[nr, ] != 0)
    for (j in seq_along(x_direct_idx)) {
      thisone <- x_direct_idx[j]
      retval <- c(retval, thisone)
      sub <- get_ancestors(thisone, c(callers, nr))
      if (all(sub != nr)) retval <- c(retval, sub)
    }
    retval <- sort.int(unique(retval))
    assign(as.character(nr), retval, envir = out_env)
    retval
  }

  # run over each node
  for (i in seq_len(nr)) {
    out_idx[[i]] <- get_ancestors(i, integer(0))
  } # all nodes
  out_idx
}


# cluster rows/cols that are linked/connected
# return list of connected nodes
# we assume A is the square/symmetric (binary) adjacency matrix of an
# undirected graph
#
# this version written by Luc De Wilde
lav_graph_get_connected_nodes <- function(a) {

  # make sure we have square symmetric matrix
  a <- as.matrix(a)

  # A must be square
  stopifnot(nrow(a) == ncol(a))

  # A must be symmetric
  stopifnot(isSymmetric(a))

  # set diagonal to zero (just in case)
  diag(a) <- 0L

  # make it logical
  a <- (a != 0)

  # number of cols/rows
  m <- ncol(a)

  # catch diagonal A
  if (all(lav_mat_vech(a, diagonal = FALSE) == 0L)) {
    return(seq_len(m))
  }

  visited <- rep(FALSE, m)         # track visited nodes
  membership <- integer(m)         # component id for each node
  component_id <- 0L               # current component id

  put_node_in_component <- function(node, componentid) {
    visited[node] <<- TRUE
    membership[node] <<- componentid
    toadd <- which(a[, node])
    for (n in toadd) {
      if (!visited[n]) put_node_in_component(n, componentid)
    }
  }

  for (node in seq_len(m)) {
    if (!visited[node]) {
      component_id <- component_id + 1L
      put_node_in_component(node, component_id)
    }
  }

  membership
}

# This routine tries to order the expressions such that all expressions that
# use a certain variable come after the expression that defines this variable.
# input adj.mat is an adjacency matrix
# when there is a 1 in element (r, c) this means that the variable
# defined in expression r is used in expression c.

# contributed by ldw (adapted from function defined by Kss2k, github issue #445)
lav_graph_order_adj_mat <- function(adj_mat, warn = TRUE) {
  adjmat <- adj_mat
  n <- nrow(adjmat)
  k <- 0
  testen <- 1:n
  while (TRUE) {
    found <- FALSE
    tests <- testen # which expressions are tested in the next loop
    for (i in tests) {
      if (all(adjmat[1:n, i] == 0)) { # var defined in expression i doesn't use
        # vars not already defined, select as next
        k <- k + 1            # increment order counter
        adjmat[i, ] <- 0      # remove usage indicator of defined variable
        adjmat[i, i] <- k     # keep order of expressions in diagonal element
        found <- TRUE
        testen <- testen[testen != i] # don't test this i in next loop
      }
    }
    if (k == n) return(order(diag(adjmat)))  # all done
    if (!found) {  # no definable var found
        if (warn) {
          lav_msg_warn(gettext("unable to sort `:=` parameters;",
                               "system of defined parameters contains a cycle"))
        }
        return(seq_len(n))    # cycle detected; return original order
    }
  }
}

  # This function performs the same task as lav_graph_order_adj_mat but
  # takes as input two vectors (character or integer) of the same length
  # defining the dependencies:
  # inputs are vectors defined and definedby of the same length, where
  # for each index j for these vectors
  # node definedby[j] is used in the definition of node defined[j].
  # This routine tries to order all nodes such that all node definitions that
  # use a certain node come after the definition of that node.

  # contributed by ldw as alternative for lav_graph_order_adj_mat
  # this function is faster, requires less memory and handles also
  # character vectors
lav_graph_topological_sort <- function(defined, definedby, warn = TRUE) {
  nodes <- unique(c(definedby, defined))
  n <- length(nodes)
  rv <- vector(mode(definedby), n)
  k <- 0L
  testnodes <- rep(TRUE, n)
  testedges <- rep(TRUE, length(defined))
  while (TRUE) {
    found <- FALSE
    tests <- which(testnodes)
    for (i in tests) {
      tocheck <- nodes[i]
      if (all(tocheck != defined[testedges])) {
        k <- k + 1L
        testnodes[i] <- FALSE
        rv[k] <- tocheck
        found <- TRUE
        testedges[definedby == tocheck] <- FALSE
      }
    }
    if (k == n) return(rv)  # all done
    if (!found) {  # no definable node found
      if (warn) {
        lav_msg_warn(gettext("unable to sort;",
                             "dependencies contain a cycle"))
      }
      return(nodes)    # cycle detected; return original order
    }
  }
}

# Topological grouping and placing of nodes in a matrix.
# Nodes which need to be placed at a border are given in argument bordernodes.
# This routine does a topological sort and returns a data.frame with the nodes,
# their position in a  matrix (rows, cols) and an indication of root
# (no dependencies, indic == "r") or
# leave (no other nodes depend on this one, indic == "l"), as follows :
# the first column contains all nodes without dependencies (*1) (*2)
# the second column contains nodes with only dependency
#                                       in the first column (*1)(*3)
# the third column contains nodes with only dependencies in the first and
#     second, with at least one in the second column and so on (*3)
# (*1) nodes with successors but no successors in the next column are promoted
#     to the column just before the least column of the successors, except
#     the nodes mentioned in bordernodes and in the first column.
# (*2) the rows in the first column are chosen so that they are in the
#     neighborhood of the mean of the rows from the nodes depending on them.(*4)
# (*3) the rows in the second to last column are chosen so that the sum of the
#     row-distances to connected nodes is minimized
# (*4) if there are nodes forced to the top border in columns 2 through
#     maxcol-1, the items in column1s 1 and maxcol cannot occupy the first row;
#     analogue for the bottom
lav_graph_topological_mat <- function(
  defined,
  definedby,
  bordernodes = character(0),
  warn = TRUE) {
  rv <- lav_graph_topological_sort(defined, definedby, warn)
  n <- length(rv)
  rvrow <- integer(n)
  rvcol <- integer(n)
  rvindic <- character(n)
  # position nodes in matrix, column by column, and mark root and leave nodes
  # rows are assigned but not yet adapted
  colmax <- 0L
  bordersincol <- 0L
  topborderfixed <- FALSE
  bottomborderfixed <- FALSE
  bordernodefound <- FALSE
  for (i in seq.int(n)) {
    predecessors <- definedby[rv[i] == defined]
    followers <- defined[rv[i] == definedby]
    isbordernode <- rv[i] %in% bordernodes
    for (predx in predecessors) {
      if (predx %in% bordernodes) bordernodefound <- TRUE
    }
    if (length(predecessors) == 0L) {
      rvcol[i] <- 1L
      colmax[1L] <- colmax[1L] + 1L
      rvrow[i] <- colmax[1L]
      rvindic[i] <- "r" # root
      if (isbordernode) bordernodefound <- TRUE
    } else if (!bordernodefound && isbordernode && length(followers) > 0L) {
      rvcol[i] <- 1L
      colmax[1L] <- colmax[1L] + 1L
      rvrow[i] <- colmax[1L]
      rvindic[i] <- "r" # root
    } else {
      if (length(followers) == 0L) {
        rvindic[i] <- "l"
      }
      predecessor_ind <- match(predecessors, rv)
      rvcol[i] <- max(rvcol[predecessor_ind]) + 1L
      if (length(colmax) < rvcol[i]) {
        colmax[rvcol[i]] <- 1L
        bordersincol <- 0L
      } else {
        colmax[rvcol[i]] <- colmax[rvcol[i]] + 1L
      }
      rvrow[i] <- colmax[rvcol[i]]
      if (rv[i] %in% bordernodes && rvindic[i] != "l") {
        bordersincol <- bordersincol + 1L
        topborderfixed <- TRUE
        if (bordersincol == 2L) bottomborderfixed <- TRUE
        if (bordersincol > 2L) {
          rvcol[i]  <- rvcol[i] + 1L
          colmax[rvcol[i]] <- 1L
          rvrow[i] <- colmax[rvcol[i]]
        }
      }
    }
  }
  # increment columns ? (*)
  incremented <- TRUE
  while (incremented) {
    incremented <- FALSE
    for (i in seq_along(rv)) {
      if (rvindic[i] != "l" && !(rv[i] %in% bordernodes && rvindic[i] == "r")) {
        followers <- defined[rv[i] == definedby]
        followers_ind <- match(followers, rv)
        mincol <- min(rvcol[followers_ind])
        if (rvcol[i] < mincol - 1L) {
          incremented <- TRUE
          curcol <- rvcol[i]
          rvcol[i] <- mincol - 1L
          # adapt rows in curcol
          rvrow[rvcol == curcol & rvrow > rvrow[i]] <-
            rvrow[rvcol == curcol & rvrow > rvrow[i]] - 1L
          colmax[curcol] <- colmax[curcol] - 1L
          # adapt rows in newcol
          colmax[mincol - 1L] <- colmax[mincol - 1L] + 1L
          rvrow[i] <- colmax[mincol - 1L]
        }
      }
    }
  }
  # place bordernodes in columns 2 through maxcol-1 in the first or last row
  for (col in seq.int(2L, length(colmax) - 1L)) {
    bordernodes_incol <- which(rvcol == col & rv %in% bordernodes)
    if (length(bordernodes_incol) > 0L) {
      totop <- bordernodes_incol[1L]
      if (rvrow[totop] != 1L) {
        swappie <- which(rvcol == col & rvrow == 1L)
        rvrow[swappie] <- rvrow[totop]
        rvrow[totop] <- 1L
      }
    }
    if (length(bordernodes_incol) > 1L) {
      tobottom <- bordernodes_incol[2L]
      if (rvrow[tobottom] != max(colmax)) {
        swappie <- which(rvcol == col & rvrow == colmax[col])
        rvrow[swappie] <- rvrow[tobottom]
        rvrow[tobottom] <- max(colmax)
      }
    }
  }
  # help function to order doubles as integers in a specified range
  order_doubles_interval <- function(inorder, outrange) {
    stopifnot(length(inorder) < outrange[2] - outrange[1] + 2)
    if (length(inorder) == 1) {
      if (inorder >= outrange[1L] && inorder <= outrange[2L]) {
        return(as.integer(inorder))
      }
      return(as.integer(sum(outrange) / 2))
    }
    in_order <- order(inorder)
    in_order[in_order] <- seq_along(inorder)
    if (length(inorder) == outrange[2] - outrange[1] + 1) {
      return(as.integer(outrange[1]) + in_order - 1L)
    }
    as.integer(outrange[1] + (in_order - 1L) *
      (outrange[2] - outrange[1] + 0.99) / (length(in_order) - 1L))
  }
  # arrange nodes in first column (***)
  rowmax <- max(colmax)
  addrows <- 0L
  if (topborderfixed) addrows <- 1L
  if (bottomborderfixed) addrows <- 2L
  rowmax <- max(colmax, colmax[1L] + addrows, colmax[length(colmax)] + addrows)
  if (addrows > 0L) { # adapt rows for nodes in maxrow and inner columns
    welke <- which(rvcol > 1L & rvcol < length(colmax) & rvrow == max(colmax))
    rvrow[welke] <- rowmax
  }
  nodescol1 <- which(rvcol == 1L)
  optimalrows <- sapply(nodescol1, function(ci) {
    nextnodes_ind <- which(rv[ci] == definedby)
    if (length(nextnodes_ind) == 0L) {
      colmax[1L] / 2
    } else {
      nextnodes <- defined[nextnodes_ind]
      mean(rvrow[match(nextnodes, rv)])
    }
  })
  interval1l <- c(
    if (topborderfixed) 2L else 1L,
    if (bottomborderfixed) rowmax - 1L else rowmax
  )
  rvrow[nodescol1] <- order_doubles_interval(optimalrows, interval1l)

  # adapt rows in columns to match as close as possible the rows of the
  # connected nodes in prior column(s), except for
  # borders for all but the last column: (**)
  if (length(colmax) > 1L) {
    for (c in seq.int(2L, length(colmax))) {
      if (c == length(colmax)) {
        cnodes_ind <- which(rvcol == c)
        interval <- interval1l
      } else {
        cnodes_ind <- which(rvcol == c & !(rv %in% bordernodes))
        if (length(cnodes_ind) == 0L) next
        interval <- range(rvrow[cnodes_ind])
        interval[2L] <- max(interval[2L], colmax - 1L)
      }
      optimalrows <- sapply(cnodes_ind, function(ci) {
        prevnodes_ind <- which(rv[ci] == defined)
        if (length(prevnodes_ind) == 0L) {
          rvrow[ci]
        } else {
          prevnodes <- definedby[prevnodes_ind]
          mean(rvrow[match(prevnodes, rv)])
        }
      })
      rvrow[cnodes_ind] <- order_doubles_interval(optimalrows, interval)
    }
  }
  # order on column, then row, and return
  neworder <- order(rvcol * 1000L + rvrow)
  stopifnot(length(neworder) == length(unique(neworder)))
  data.frame(
    nodes = rv[neworder],
    rows = rvrow[neworder],
    cols = rvcol[neworder],
    indic = rvindic[neworder]
  )
}

lav_label_code <- function(label = "", value = "", show = FALSE,
                           idx_font_size = 20L, dy = 7L,
                           italic = TRUE, auto_subscript = TRUE) {
  latexsymbols <- c("varepsilon",
    "Alpha", "Beta", "Gamma", "Delta", "Epsilon", "Zeta", "Eta", "Theta",
    "Iota", "Kappa", "Lambda", "Mu", "Nu", "Xi", "Omicron", "Pi",
    "Rho", "Sigma", "Tau", "Upsilon", "Phi", "Chi", "Psi", "Omega",
    "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta",
    "iota", "kappa", "lambda", "mu", "nu", "xi", "omicron", "pi",
    "rho", "sigma", "tau", "upsilon", "phi", "chi", "psi", "omega"
  )
  unicodes <- c("1013",
    "913", "914", "915", "916", "917", "918", "919", "920",
    "921", "922", "923", "924", "925", "926", "927", "928",
    "929", "931", "932", "933", "934", "935", "936", "937",
    "945", "946", "947", "948", "949", "950", "951", "952",
    "953", "954", "955", "956", "957", "958", "959", "960",
    "961", "963", "964", "965", "966", "967", "968", "969"
  )
  if (label == "" && value == "") return(list(svg = "", tikz = "", r = ""))
  if (auto_subscript) {
    label <- gsub("^([a-zA-Z]+)([0-9]+)", "\\1_\\2", label)
  }
  rexpression <- FALSE
  if (grepl("1van", label, fixed = TRUE)) label <- "1"
  if (value == "") {
    splitted <- strsplit(label, "=", fixed =  TRUE)[[1L]]
    if (length(splitted) > 1L) {
      label <- splitted[1L]
      value <- splitted[2L]
    }
  }
  if (label == "") {
    label <- value
    value <- ""
  }
  svgpart <- tikzpart <- rpart <- character(3)
  rpart[3L] <- tikzpart[3L] <- svgpart[3L] <- value
  if (label != "") {
    # separate label and value by equal sign
    if (svgpart[3L] != "") svgpart[3L] <- paste0("=", svgpart[3L])
    if (tikzpart[3L] != "") tikzpart[3L] <- paste0("=", tikzpart[3L])
    if (rpart[3L] != "") {
      rpart[3L] <- paste0(" == ", rpart[3L])
      rexpression <- TRUE
    }
    # subscript and Greek character set handling
    splitunderscore <- strsplit(label, "_", TRUE)[[1L]]
    svgpart[1] <- tikzpart[1] <- rpart[1] <- splitunderscore[1L]
    if (rpart[1] %in% latexsymbols) {
      rexpression <- TRUE
      svgpart[1] <- paste0("&#", unicodes[latexsymbols == svgpart[1]], ";")
      tikzpart[1] <- paste0("\\", tikzpart[1])
      if (rpart[1] == "varepsilon") rpart[1] <- "epsilon"
      rpart[1] <- str2expression(rpart[1])
    }
    if (length(splitunderscore) > 1L) {
      rexpression <- TRUE
      svgpart[2L] <- paste0("<tspan dy=\"", dy, "\" font-size=\"",
                     idx_font_size, "\">", splitunderscore[2L], "</tspan>")
      if (svgpart[3L] != "") {
        svgpart[3L] <- paste0("<tspan dy=\"-", dy, "\">",
                               svgpart[3L], "</tspan>")
      }
      tikzpart[2L] <- paste0("_{", splitunderscore[2L], "}")
      rpart[2L] <- paste0("[\"", splitunderscore[2L], "\"]")
    }
    if (!italic) tikzpart <- c("\\mathrm{", tikzpart, "}")
  }
  if (show) {
    plot(c(0, 3), c(0, 2), type = "n")
    if (rexpression) {
      text(1, 1, str2expression(paste(rpart, collapse = "")))
    } else {
      text(1, 1, paste(rpart, collapse = ""), font = ifelse(italic, 3L, 1L))
    }
  }
  list(svg = paste(svgpart, collapse = ""),
       tikz = paste(c("$", tikzpart, "$"), collapse = ""),
       r = ifelse(rexpression,
         str2expression(paste(rpart, collapse = "")),
         paste(rpart, collapse = ""))
  )
}

#' lav_export_estimation
#'
#' lavaan provides a range of optimization methods with the optim.method
#' argument (nlminb, BFGS, L-BFGS-B, GN, and nlminb_constr).
#' `lav_export_estimation` allows exporting objects and functions necessary
#' to pass a lavaan model into any optimizer that takes a combination of
#' (1) starting values, (2) fit-function, (3) gradient-function, and (4) upper
#' and lower bounds. This allows testing new optimization frameworks.
#'
#' @param lavaan_model a fitted lavaan model
#' @returns List with:
#' \itemize{
#'   \item get_coef - When working with equality constraints, lavaan internally
#'   uses some transformations. get_coef is a functions that recreates the coef
#'   function for the parameters.
#'   \item starting_values - starting_values to be used in the optimization
#'   \item objective_function - objective function, expecting the current
#'    parameter values and the lavaan model
#'   \item gradient_function - gradient function, expecting the current
#'    parameter values and the lavaan model
#'   \item lower - lower bounds for parameters
#'   \item upper - upper bound for parameters
#' }
#' @export
#' @examples
#' library(lavaan)
#' model <- "
#'   # latent variable definitions
#'      ind60 =~ x1 + x2 + x3
#'      dem60 =~ y1 + y2 + y3 + y4
#'      dem65 =~ y5 + a*y6 + y7 + y8
#'
#'   # regressions
#'     dem60 ~ ind60
#'     dem65 ~ ind60 + dem60
#' "
#'
#' fit <- sem(model,
#'   data = PoliticalDemocracy,
#'   do.fit = FALSE
#' )
#'
#' est <- lav_export_estimation(lavaan_model = fit)
#'
#' # The starting values are:
#' est$starting_values
#' # Note that these do not have labels (and may also differ from coef(fit)
#' # in case of equality constraints):
#' coef(fit)
#' # To get the same parameters, use:
#' est$get_coef(
#'   parameter_values = est$starting_values,
#'   lavaan_model = fit
#' )
#'
#' # The objective function can be used to compute the fit at the
#' # current estimates:
#' est$objective_function(
#'   parameter_values = est$starting_values,
#'   lavaan_model = fit
#' )
#'
#' # The gradient function can be used to compute the gradients at
#' # the current estimates:
#' est$gradient_function(
#'   parameter_values = est$starting_values,
#'   lavaan_model = fit
#' )
#'
#' # Together, these elements provide the means to estimate the parameters
#' # with a large range of optimizers. For simplicity, here is an example
#' # using optim:
#' est_fit <- optim(
#'   par = est$starting_values,
#'   fn = est$objective_function,
#'   gr = est$gradient_function,
#'   lavaan_model = fit,
#'   method = "BFGS"
#' )
#' est$get_coef(
#'   parameter_values = est_fit$par,
#'   lavaan_model = fit
#' )
#'
#' # This is identical to
#' coef(sem(model,
#'   data = PoliticalDemocracy
#' ))
#'
#' # Example using ridge regularization for parameter a
#' fn_ridge <- function(parameter_values, lavaan_model, est, lambda) {
#'   return(est$objective_function(
#'     parameter_values = parameter_values,
#'     lavaan_model = lavaan_model
#'   ) + lambda * parameter_values[6]^2)
#' }
#' ridge_fit <- optim(
#'   par = est$get_coef(est$starting_values,
#'     lavaan_model = fit
#'   ),
#'   fn = fn_ridge,
#'   lavaan_model = fit,
#'   est = est,
#'   lambda = 10
#' )
#' est$get_coef(
#'   parameter_values = ridge_fit$par,
#'   lavaan_model = fit
#' )
lav_export_estimation <- function(lavaan_model) {
  # define objective function
  objective_function <- function(parameter_values,
                                 lavaan_model) {
    if (lavaan_model@Model@eq.constraints) {
      parameter_values <-
        as.numeric(lavaan_model@Model@eq.constraints.K %*% parameter_values) +
        lavaan_model@Model@eq.constraints.k0
    }

    # create group list
    glist <- lav_model_x2glist(lavaan_model@Model, x = parameter_values)
    # get objective function **value**
    fx <- lav_model_objective(
      lavmodel = lavaan_model@Model,
      glist = glist,
      lavsamplestats = lavaan_model@SampleStats,
      lavdata = lavaan_model@Data,
      lavcache = list()
    )
    if (lavaan_model@Options$estimator == "PML") {
      # rescale objective function value
      fx <- fx / lavaan_model@SampleStats@ntotal
    }

    if (!is.finite(fx)) {
      fx_group <- attr(fx, "fx.group")
      fx <- 1e+20
      attr(fx, "fx.group") <- fx_group
    }
    fx
  }

  # define gradient function
  gradient_function <- function(parameter_values,
                                lavaan_model) {
    if (lavaan_model@Model@eq.constraints) {
      parameter_values <-
        as.numeric(lavaan_model@Model@eq.constraints.K %*% parameter_values) +
        lavaan_model@Model@eq.constraints.k0
    }

    glist <- lav_model_x2glist(lavaan_model@Model,
      x = parameter_values
    )
    current_verbose <- lav_verbose()
    if (lav_verbose(FALSE))
      on.exit(lav_verbose(current_verbose), TRUE)
    dx <- lav_model_grad(
      lavmodel = lavaan_model@Model,
      glist = glist,
      lavsamplestats = lavaan_model@SampleStats,
      lavdata = lavaan_model@Data,
      lavcache = list(),
      type = "free",
      group_weight = !(lavaan_model@SampleStats@missing.flag ||
                      lavaan_model@Options$estimator == "PML"),
      ceq_simple = lavaan_model@Model@ceq.simple.only
    )
    lav_verbose(current_verbose)

    if (lavaan_model@Model@eq.constraints) {
      dx <- as.numeric(dx %*% lavaan_model@Model@eq.constraints.K)
    }
    if (lavaan_model@Options$estimator == "PML") {
      dx <- dx / lavaan_model@SampleStats@ntotal
    }
    dx
  }

  # extract bounds
  lower <- lavaan_model@ParTable$lower[lavaan_model@ParTable$free > 0L]
  upper <- lavaan_model@ParTable$upper[lavaan_model@ParTable$free > 0L]

  # get starting values
  starting_values <- lav_model_get_parameters(lavaan_model@Model)
  if (lavaan_model@Model@eq.constraints) {
    starting_values <-
      as.numeric((starting_values - lavaan_model@Model@eq.constraints.k0) %*%
                  lavaan_model@Model@eq.constraints.K)
  }

  # lavaan internally uses transformations when there are equality constraints.
  # As a result, the parameters are not necessarily those one would expect when
  # fitting the model.
  # The parameters can be translated with the following function:
  get_coef <- function(parameter_values,
                       lavaan_model) {
    if (lavaan_model@Model@eq.constraints) {
      parameter_values <- as.numeric(lavaan_model@Model@eq.constraints.K %*%
                  parameter_values) + lavaan_model@Model@eq.constraints.k0
    }
    names(parameter_values) <- lav_pt_labels(lavaan_model@ParTable,
      type = "free"
    )
    parameter_values
  }

  # Now we just return everything so that the user can use their own optimizer
  list(
    get_coef = get_coef,
    starting_values = starting_values,
    objective_function = objective_function,
    gradient_function = gradient_function,
    lower = lower,
    upper = upper
  )
}

# the Muthen (1984) three-stage estimator for mixed continuous/ordinal data:
# univariate parameters (step 1), correlations (step 2), and the asymptotic
# covariance matrix WLS.W of all estimates (step 3, Muthen & Satorra 1995)
#
# This function was written in January 2012 -- Yves Rosseel
# First success: Friday 20 Jan 2012: the standard errors for
#                thresholds and polychoric correlations (in an
#                unrestricted/saturated model) are spot on!
# Second success: Saturday 9 June 2012: support for mixed (ordinal + metric)
#                 variables; thanks to the delta method to get the ACOV
#                 right (see H matrix)
# Third success: Monday 2 July 2012: support for fixed.x covariates
#
# Friday 13 July 2012: merge exo + non-exo code
# Monday 16 July 2012: fixed sign numeric in WLS.W; I think we got it right now
# YR 26 Nov 2015: move step1 + step2 to external functions
# YR/LDW 2026: refactored; step 3 decomposed into named helpers
#              (lav_m84_*); no change in behavior
#
# Step 3 computes the sandwich
#
#     WLS.W = B^{-1} INNER B^{-T}
#
# where the 'bread' B is the (block-triangular) matrix of derivatives of the
# stacked stage-wise estimating equations with respect to the parameters
#
#     B = | A11   0  |     A11: univariate (th/sl/var) information
#         | A21  A22 |     A21: d(cor scores)/d(univariate parameters)
#                          A22: diagonal correlation information
#
# and INNER is the crossproduct of the stacked casewise scores
# (with variants for sampling weights and cluster-robust aggregation).
# The H matrix afterwards converts the correlation metric to the
# covariance metric (delta rule) when numeric variables are present.

muthen1984 <- function(data_1 = NULL,
                       ov_names = NULL,
                       ov_types = NULL,
                       ov_levels = NULL,
                       ov_names_x = character(0L),
                       exo = NULL,
                       wt = NULL,
                       sampling_weights_type = "design",
                       wls_w = TRUE,
                       zero_add = c(0.5, 0.0),
                       zero_keep_margins = TRUE,
                       zero_cell_warn = FALSE,
                       zero_cell_tables = TRUE,
                       allow_empty_cell = TRUE,
                       cluster_idx = NULL,
                       group = 1L) { # group only for error messages

  # just in case Data is a vector
  data_1 <- as.matrix(data_1)

  nvar <- NCOL(data_1)
  n <- NROW(data_1)
  num_idx <- which(ov_types == "numeric")
  ord_idx <- which(ov_types == "ordered")
  nexo <- length(ov_names_x)
  if (nexo > 0L) stopifnot(NCOL(exo) == nexo)
  pstar <- nvar * (nvar - 1) / 2

  if (lav_verbose()) {
    cat("\nPreparing for WLS estimation -- STEP 1 + 2\n")
    cat("Number of endogenous variables: ", nvar, "\n")
    cat("Endogenous variable names:\n")
    print(ov_names)
    cat("\n")
    cat("Endogenous ov types:\n")
    print(ov_types)
    cat("\n")
    cat("Endogenous ov levels:\n ")
    print(ov_levels)
    cat("\n")
    cat("Number of exogenous variables: ", nexo, "\n")
    cat("Exogenous variable names:\n")
    print(ov_names_x)
    cat("\n")
  }

  step1 <- lav_samp_step1(
    y = data_1, wt = wt, ov_names = ov_names,
    ov_types = ov_types, ov_levels = ov_levels, ov_names_x = ov_names_x,
    exo = exo, scores_flag = wls_w, allow_empty_cell = allow_empty_cell,
    group = group
  )

  fit <- step1$FIT
  th <- step1$TH
  th_nox <- step1$TH.NOX
  th_idx_1 <- step1$TH.IDX
  th_names <- step1$TH.NAMES
  th_keep <- step1$TH.KEEP
  var_1 <- step1$VAR
  slopes <- step1$SLOPES
  sc_th <- step1$SC.TH
  sc_sl <- step1$SC.SL
  sc_var <- step1$SC.VAR
  th_start_idx <- step1$th.start.idx
  th_end_idx <- step1$th.end.idx

  # rm SC.VAR columns from ordinal variables
  if (wls_w && length(ord_idx) > 0L) {
    sc_var <- sc_var[, -ord_idx, drop = FALSE]
  }


  if (lav_verbose()) {
    cat("STEP 1: univariate statistics\n")
    cat("Threshold + means:\n")
    tthh <- unlist(th)
    names(tthh) <- unlist(th_names)
    print(tthh)
    cat("Slopes (if any):\n")
    colnames(slopes) <- ov_names_x
    rownames(slopes) <- ov_names
    print(slopes)
    cat("Variances:\n")
    names(var_1) <- ov_names
    print(unlist(var_1))
  }

  # stage two -- correlations

  if (lav_verbose()) cat("\n\nSTEP 2: covariances/correlations:\n")
  cor_1 <- lav_samp_step2(
    uni = fit, wt = wt, ov_names = ov_names,
    zero_add = zero_add,
    zero_keep_margins = zero_keep_margins,
    zero_cell_warn = zero_cell_warn,
    zero_cell_tables = zero_cell_tables
  )
  empty_cell_tables <- attr(cor_1, "zero.cell.tables")
  attr(cor_1, "zero.cell.tables") <- NULL

  if (lav_verbose()) {
    colnames(cor_1) <- rownames(cor_1) <- ov_names
    print(cor_1)
  }

  if (!wls_w) { # we do not need the asymptotic variance matrix
    if (any("numeric" %in% ov_types)) {
      cov_1 <- lav_cor2cov(r = cor_1, sds = sqrt(unlist(var_1)))
    } else {
      cov_1 <- cor_1
    }
    out <- list(
      TH = th, SLOPES = slopes, VAR = var_1, COR = cor_1, COV = cov_1,
      SC = NULL, TH.NOX = th_nox, TH.NAMES = th_names, TH.IDX = th_idx_1,
      INNER = NULL, A11 = NULL, A12 = NULL, A21 = NULL, A22 = NULL,
      WLS.W = NULL, H = NULL, zero.cell.tables = matrix("", 0, 2)
    )
    return(out)
  }


  # stage three -- WLS.W

  # index layout of the stacked parameter vector
  # (th/means | slopes (if any) | variances (numeric only) | correlations)
  layout <- lav_m84_layout(
    nvar = nvar, num_idx = num_idx,
    th_start_idx = th_start_idx, th_end_idx = th_end_idx,
    nth = NCOL(sc_th), nsl = NCOL(sc_sl), nexo = nexo
  )
  a11_size <- layout$a11_size

  # A21 (+ the H21/H22 blocks of the delta-rule matrix, and the
  #      correlation scores)
  tmp <- lav_m84_a21(
    fit = fit, ov_types = ov_types, cor_1 = cor_1, var_1 = var_1,
    wt = wt, n = n, layout = layout, th_keep = th_keep
  )
  sc_cor <- tmp$sc_cor
  a21 <- tmp$a21
  h21 <- tmp$h21
  h22 <- tmp$h22

  if (!is.null(wt)) {
    sc_cor <- sc_cor * wt # reweight
  }

  # stacked scores + meat
  sc <- cbind(sc_th, sc_sl, sc_var, sc_cor)
  inner <- lav_mat_crossprod(sc)

  # A11
  # new approach (2 June 2012): A11 is just a 'sparse' version of
  # (the left upper block of) INNER
  if (!is.null(wt)) {
    inner2 <- lav_mat_crossprod(sc / wt, sc)
  } else {
    inner2 <- inner
  }
  a11 <- lav_m84_a11(inner2 = inner2, layout = layout)

  # A22 (diagonal)
  a22 <- lav_m84_a22(sc_cor = sc_cor, wt = wt)

  # A12 (zero)
  a12 <- matrix(0, NROW(a11), NCOL(a22))

  # invert B as a block-triangular matrix (0.5-23)
  tmp <- lav_m84_b_inv(a11 = a11, a21 = a21, a22 = a22)
  b_inv <- tmp$b_inv
  # (warnings are emitted here, not in the helper, so that the message
  #  mentions muthen1984 as before)
  if (tmp$ginv_a11) {
    lav_msg_warn(gettext("trouble constructing W matrix;
                         used generalized inverse for A11 submatrix"))
  }
  if (tmp$ginv_a22) {
    lav_msg_warn(gettext("trouble constructing W matrix;
                         used generalized inverse for A22 submatrix"))
  }

  #  weight matrix (correlation metric)
  # The bread b_inv is built from sum(wt)-weighted quantities (a11/a22 use
  # 'inner2'). The choice of meat encodes how the sampling weights are
  # interpreted (sampling.weights.type):
  #  - "design"    : meat = 'inner' = crossprod(sc) where sc is wt-scaled,
  #                  i.e. weighted by sum(wt^2). b_inv %*% inner %*% b_inv is
  #                  then the design-based sandwich Gamma (matches Mplus).
  #  - "frequency" : meat = 'inner2' = crossprod(sc/wt, sc), weighted by
  #                  sum(wt); treats the weights as frequencies, so Gamma
  #                  (and the WLS weight matrix and robust SEs) matches a fit
  #                  on the row-replicated data.
  # Without sampling weights inner2 == inner, so the choice is a no-op.
  if (!is.null(wt) && identical(sampling_weights_type, "frequency")) {
    wls_w <- b_inv %*% inner2 %*% t(b_inv)
  } else {
    wls_w <- b_inv %*% inner %*% t(b_inv)
  }

  # cluster-robust meat (issue #254): for cluster-robust Gamma/NACOV we keep
  # the same 'bread' (b_inv), but aggregate the case-wise scores per cluster
  # before forming the outer product -- exactly as in the continuous case
  # (lav_samp_gamma: zc <- rowsum(zc, cluster_idx)). Only affects Gamma/NACOV
  # (hence the robust SEs and scaled test); the WLS weight matrix (and thus the
  # point estimates) is still built from the standard i.i.d. 'wls_w' above.
  wls_w_cl <- NULL
  if (!is.null(cluster_idx)) {
    sc_cl <- rowsum.default(sc, group = cluster_idx, reorder = FALSE)
    inner_cl <- lav_mat_crossprod(sc_cl)
    wls_w_cl <- b_inv %*% inner_cl %*% t(b_inv)
  }

  # COV matrix?
  if (any("numeric" %in% ov_types)) {
    cov_1 <- lav_cor2cov(r = cor_1, sds = sqrt(unlist(var_1)))

    # construct H matrix to apply delta rule (for the transformation
    # of rho_ij to cov_ij)
    h11 <- diag(NROW(a11))
    h12 <- matrix(0, NROW(a11), NCOL(a22))
    # H22 and H21 already filled in
    h <- rbind(
      cbind(h11, h12),
      cbind(h21, h22)
    )

    wls_w <- h %*% wls_w %*% t(h)
  } else {
    cov_1 <- cor_1
    h <- diag(NCOL(wls_w))
  }
  if (!is.null(wls_w_cl)) {
    wls_w_cl <- h %*% wls_w_cl %*% t(h)
  }

  # reverse sign numeric TH (because we provide -mu in WLS.obs)
  # (WOW, it took me a LOOONGGG time to realize this!)
  # YR 16 July 2012

  # NOTE: prior to 0.5-17, we used num.idx (instead of NUM.idx)
  # which is WRONG if we have more than one threshold per variable
  # (thanks to Sacha Epskamp for spotting this!)
  if (length(num_idx) > 0L) {
    num_idx_1 <- which(unlist(th_idx_1) == 0L)
    wls_w[num_idx_1, ] <- -wls_w[num_idx_1, ]
    wls_w[, num_idx_1] <- -wls_w[, num_idx_1]
    if (!is.null(wls_w_cl)) {
      wls_w_cl[num_idx_1, ] <- -wls_w_cl[num_idx_1, ]
      wls_w_cl[, num_idx_1] <- -wls_w_cl[, num_idx_1]
    }
  }


  out <- list(
    TH = th, SLOPES = slopes, VAR = var_1, COR = cor_1, COV = cov_1,
    SC = sc, TH.NOX = th_nox, TH.NAMES = th_names, TH.IDX = th_idx_1,
    INNER = inner, A11 = a11, A12 = a12, A21 = a21, A22 = a22,
    WLS.W = wls_w, WLS.W.cluster = wls_w_cl, H = h,
    zero.cell.tables = empty_cell_tables
  )
  out
}

# index layout of the stacked (univariate | correlation) parameter vector:
# for each variable, the positions of its thresholds/mean, slopes and
# (numeric only) variance within the A11 block; and the pstar mapping of
# the correlations
lav_m84_layout <- function(nvar = NULL, num_idx = NULL,
                           th_start_idx = NULL, th_end_idx = NULL,
                           nth = NULL, nsl = NULL, nexo = NULL) {
  pstar <- nvar * (nvar - 1) / 2
  pstar_1 <- matrix(0, nvar, nvar)
  pstar_1[lav_mat_vech_idx(nvar, diagonal = FALSE)] <- 1:pstar

  th_idx <- vector("list", length = nvar)
  sl_idx <- vector("list", length = nvar)
  var_idx <- vector("list", length = nvar)
  for (i in 1:nvar) {
    th_idx[[i]] <- th_start_idx[i]:th_end_idx[i]
    if (nexo > 0L) {
      sl_idx[[i]] <- nth + seq(i, by = nvar, length.out = nexo)
    } else {
      sl_idx[[i]] <- integer(0L)
    }
    # variance slot (numeric variables only; NA for ordinal)
    var_idx[[i]] <- nth + nsl + match(i, num_idx)
  }

  list(
    nvar = nvar, pstar = pstar, pstar_1 = pstar_1, nexo = nexo,
    th_idx = th_idx, sl_idx = sl_idx, var_idx = var_idx,
    a11_size = nth + nsl + length(num_idx)
  )
}

# the casewise scores of one correlation (pair i > j), together with the
# cross-derivatives with respect to the univariate parameters of both
# variables, mapped to a type-independent layout:
#   dx_rho, dx_th_i/dx_th_j (thresholds or mean), dx_sl_i/dx_sl_j (slopes),
#   dx_var_i/dx_var_j (residual variance; NULL for ordinal variables)
lav_m84_pair_sc <- function(fit = NULL, ov_types = NULL,
                            i = NULL, j = NULL, rho = NULL, wt = NULL) {
  ord_i <- ov_types[i] == "ordered"
  ord_j <- ov_types[j] == "ordered"
  if (!ord_i && !ord_j) {
    # pearson
    sc <- lav_bvreg_cor_sc(
      rho = rho, fit_y1 = fit[[i]], fit_y2 = fit[[j]], wt = wt
    )
    out <- list(
      dx_rho = sc$dx_rho,
      dx_th_i = sc$dx_mu_y1, dx_th_j = sc$dx_mu_y2,
      dx_sl_i = sc$dx_sl_y1, dx_sl_j = sc$dx_sl_y2,
      dx_var_i = sc$dx_var_y1, dx_var_j = sc$dx_var_y2
    )
  } else if (!ord_i && ord_j) {
    # polyserial (i = linear, j = ordinal)
    sc <- lav_bvmix_cor_sc(
      rho = rho, fit_y1 = fit[[i]], fit_y2 = fit[[j]], wt = wt
    )
    out <- list(
      dx_rho = sc$dx_rho,
      dx_th_i = sc$dx_mu_y1, dx_th_j = sc$dx_th_y2,
      dx_sl_i = sc$dx_sl_y1, dx_sl_j = sc$dx_sl_y2,
      dx_var_i = sc$dx_var_y1, dx_var_j = NULL
    )
  } else if (ord_i && !ord_j) {
    # polyserial (j = linear, i = ordinal)
    sc <- lav_bvmix_cor_sc(
      rho = rho, fit_y1 = fit[[j]], fit_y2 = fit[[i]], wt = wt
    )
    out <- list(
      dx_rho = sc$dx_rho,
      dx_th_i = sc$dx_th_y2, dx_th_j = sc$dx_mu_y1,
      dx_sl_i = sc$dx_sl_y2, dx_sl_j = sc$dx_sl_y1,
      dx_var_i = NULL, dx_var_j = sc$dx_var_y1
    )
  } else {
    # polychoric
    sc <- lav_bvord_cor_sc(
      rho = rho, fit_y1 = fit[[i]], fit_y2 = fit[[j]], wt = wt
    )
    out <- list(
      dx_rho = sc$dx_rho,
      dx_th_i = sc$dx_th_y1, dx_th_j = sc$dx_th_y2,
      dx_sl_i = sc$dx_sl_y1, dx_sl_j = sc$dx_sl_y2,
      dx_var_i = NULL, dx_var_j = NULL
    )
  }
  out
}

# the A21 block: for each correlation (row), the crossproduct of its scores
# with the cross-derivatives w.r.t. the univariate parameters (columns);
# also returns the correlation scores sc_cor (unweighted) and the H21/H22
# blocks of the delta-rule matrix (rho -> cov metric, numeric variables)
lav_m84_a21 <- function(fit = NULL, ov_types = NULL,
                        cor_1 = NULL, var_1 = NULL,
                        wt = NULL, n = NULL, layout = NULL,
                        th_keep = NULL) {
  nvar <- layout$nvar
  pstar <- layout$pstar

  sc_cor <- matrix(0, n, pstar)
  a21 <- matrix(0, pstar, layout$a11_size)
  h22 <- diag(pstar) # for the delta rule
  h21 <- matrix(0, pstar, layout$a11_size)

  # for this one, we need new scores: for each F_ij (cor), the
  # scores with respect to the TH, VAR, ...
  for (j in seq_len(nvar - 1L)) {
    for (i in (j + 1L):nvar) {
      pstar_idx <- layout$pstar_1[i, j]

      pair <- lav_m84_pair_sc(
        fit = fit, ov_types = ov_types, i = i, j = j,
        rho = cor_1[i, j], wt = wt
      )

      # RHO
      if (is.null(wt)) {
        sc_cor[, pstar_idx] <- pair$dx_rho
      } else {
        sc_cor[, pstar_idx] <- pair$dx_rho / wt # unweight
      }

      # TH
      # (th_keep: skip the pseudo-threshold slots of empty categories --
      #  the fit has no scores for them; their a21 entries stay zero)
      a21[pstar_idx, layout$th_idx[[i]][th_keep[[i]]]] <-
        lav_mat_crossprod(sc_cor[, pstar_idx], pair$dx_th_i)
      a21[pstar_idx, layout$th_idx[[j]][th_keep[[j]]]] <-
        lav_mat_crossprod(sc_cor[, pstar_idx], pair$dx_th_j)

      # SL
      if (layout$nexo > 0L) {
        a21[pstar_idx, layout$sl_idx[[i]]] <-
          lav_mat_crossprod(sc_cor[, pstar_idx], pair$dx_sl_i)
        a21[pstar_idx, layout$sl_idx[[j]]] <-
          lav_mat_crossprod(sc_cor[, pstar_idx], pair$dx_sl_j)
      }

      # VAR (numeric variables only) + H21/H22 delta-rule entries
      num_i <- !is.null(pair$dx_var_i)
      num_j <- !is.null(pair$dx_var_j)
      if (num_i) {
        a21[pstar_idx, layout$var_idx[[i]]] <-
          lav_mat_crossprod(sc_cor[, pstar_idx], pair$dx_var_i)
      }
      if (num_j) {
        a21[pstar_idx, layout$var_idx[[j]]] <-
          lav_mat_crossprod(sc_cor[, pstar_idx], pair$dx_var_j)
      }
      if (num_i && num_j) {
        h21[pstar_idx, layout$var_idx[[i]]] <-
          (sqrt(var_1[j]) * cor_1[i, j]) / (2 * sqrt(var_1[i]))
        h21[pstar_idx, layout$var_idx[[j]]] <-
          (sqrt(var_1[i]) * cor_1[i, j]) / (2 * sqrt(var_1[j]))
        h22[pstar_idx, pstar_idx] <- sqrt(var_1[i]) * sqrt(var_1[j])
      } else if (num_i) {
        h21[pstar_idx, layout$var_idx[[i]]] <-
          cor_1[i, j] / (2 * sqrt(var_1[i]))
        h22[pstar_idx, pstar_idx] <- sqrt(var_1[i])
      } else if (num_j) {
        h21[pstar_idx, layout$var_idx[[j]]] <-
          cor_1[i, j] / (2 * sqrt(var_1[j]))
        h22[pstar_idx, pstar_idx] <- sqrt(var_1[j])
      }
    }
  }

  list(sc_cor = sc_cor, a21 = a21, h21 = h21, h22 = h22)
}

# the A11 block: 'sparse' (per-variable block-diagonal) version of the
# left-upper block of INNER (crossprod of the univariate scores)
lav_m84_a11 <- function(inner2 = NULL, layout = NULL) {
  a11 <- matrix(0, layout$a11_size, layout$a11_size)
  for (i in seq_len(layout$nvar)) {
    a11_idx <- c(
      layout$th_idx[[i]], layout$sl_idx[[i]],
      layout$var_idx[[i]][!is.na(layout$var_idx[[i]])]
    )
    a11[a11_idx, a11_idx] <- inner2[a11_idx, a11_idx]
  }
  a11
}

# the A22 block: diagonal of the correlation-score crossproducts
lav_m84_a22 <- function(sc_cor = NULL, wt = NULL) {
  pstar <- NCOL(sc_cor)
  a22 <- matrix(0, pstar, pstar)
  for (i in seq_len(pstar)) {
    if (is.null(wt)) {
      a22[i, i] <- sum(sc_cor[, i] * sc_cor[, i], na.rm = TRUE)
    } else {
      a22[i, i] <- sum(sc_cor[, i] * sc_cor[, i] / wt, na.rm = TRUE)
    }
  }
  a22
}

# invert the block-triangular bread matrix
#
# B <- rbind( cbind(A11,A12),
#            cbind(A21,A22) )
#
# B.inv = A11^{-1}                   0
#         -A22^{-1} A21 A11^{-1}     A22^{-1}
#
lav_m84_b_inv <- function(a11 = NULL, a21 = NULL, a22 = NULL) {
  # invert A11
  ginv_a11 <- FALSE
  a11_inv <- try(solve(a11), silent = TRUE)
  if (inherits(a11_inv, "try-error")) {
    # brute force
    a11_inv <- MASS::ginv(a11)
    ginv_a11 <- TRUE
  }

  # invert A22 (diagonal)
  ginv_a22 <- FALSE
  da22 <- diag(a22)
  if (any(da22 == 0)) {
    ginv_a22 <- TRUE
    a22_inv <- MASS::ginv(a22)
  } else {
    a22_inv <- a22
    diag(a22_inv) <- 1 / da22
  }

  # lower-left block
  a21_inv <- -a22_inv %*% a21 %*% a11_inv

  # upper-right block remains zero
  a12_inv <- matrix(0, NROW(a11), NCOL(a22))

  # construct B.inv
  b_inv <- rbind(
    cbind(a11_inv, a12_inv),
    cbind(a21_inv, a22_inv)
  )

  list(b_inv = b_inv, ginv_a11 = ginv_a11, ginv_a22 = ginv_a22)
}

# the weighted bivariate linear regression model
# YR 14 March 2020 (replacing the old lav_pearson.R + lav_binorm.R routines)
#
# - bivariate standard normal (density/CDF primitives, also used by the
#   other bv modules)
# - pearson correlation
# - bivariate linear regression
# - using sampling weights wt
#
# YR/LDW 2026: refactored (see lav_uvbv_common.R for the shared cache/minfns
#              design); fixed two dormant bugs (see below); no change in
#              behavior otherwise
#
#  info concerning lintr package, Luc DW May 7, 2026
#  there is a known problem in codetools, transferred to lintr, that
#  variables assigned in a with() are marked as 'may not be used';
#  to this end the code in with() statements are excluded from linting!
#

# density of a bivariate __standard__ normal
lav_dbinorm <- function(u, v, rho, force_zero = FALSE) {
  # dirty hack to handle extreme large values for rho
  # note that u, v, and rho are vectorized!
  rho_limit <- 0.9999
  abs_rho <- abs(rho)
  idx <- which(abs_rho > rho_limit)
  if (length(idx) > 0L) {
    rho[idx] <- sign(rho[idx]) * rho_limit
  }

  r <- 1 - rho * rho
  out <- 1 / (2 * pi * sqrt(r)) *
       exp(-0.5 * (u * u - 2 * rho * u * v + v * v) / r)

  # if abs(u) or abs(v) are very large (say, >10), set result equal
  # to exactly zero
  idx <- which(abs(u) > 10 | abs(v) > 10)
  if (length(idx) > 0L && force_zero) {
    out[idx] <- 0
  }

  out
}

# the factor g(u, v; rho) in
#
#    d phi_2(u, v; rho) / d rho = phi_2(u, v; rho) * g(u, v; rho)
#
# i.e. the derivative of the log bivariate standard normal density with
# respect to rho; used for the rho-Hessians in lav_bvord.R
# note: Olsson 1979 A2 contains an error here!
lav_dbinorm_guv <- function(u, v, rho) {
  r <- (1 - rho * rho)
  (u * v * r -
    rho * ((u * u) - 2 * rho * u * v + (v * v)) + rho * r) / (r * r)
}

# switch between pbivnorm, mnormt, ...
pbinorm <- function(upper_x = NULL, upper_y = NULL, rho = 0.0,
                    lower_x = -Inf, lower_y = -Inf, check = FALSE) {
  pbinorm2(
    upper_x = upper_x, upper_y = upper_y, rho = rho,
    lower_x = lower_x, lower_y = lower_y, check = check
  )
}

# using vectorized version (a la pbivnorm)
pbinorm2 <- function(upper_x = NULL, upper_y = NULL, rho = 0.0,
                     lower_x = -Inf, lower_y = -Inf, check = FALSE) {
  n <- length(upper_x)
  stopifnot(length(upper_y) == n)
  if (n > 1L) {
    if (length(rho) == 1L) {
      rho <- rep(rho, n)
    }
    if (length(lower_x) == 1L) {
      lower_x <- rep(lower_x, n)
    }
    if (length(lower_y) == 1L) {
      lower_y <- rep(lower_y, n)
    }
  }

  upper_only <- all(lower_x == -Inf & lower_y == -Inf)
  if (upper_only) {
    upper_x[upper_x == +Inf] <- exp(10) # better pnorm?
    upper_y[upper_y == +Inf] <- exp(10)
    upper_x[upper_x == -Inf] <- -exp(10)
    upper_y[upper_y == -Inf] <- -exp(10)
    res <- pbivnorm(upper_x, upper_y, rho = rho)
  } else {
    # pbivnorm does not handle -Inf well...
    lower_x[lower_x == -Inf] <- -exp(10)
    lower_y[lower_y == -Inf] <- -exp(10)
    res <- pbivnorm(upper_x, upper_y, rho = rho) -
      pbivnorm(lower_x, upper_y, rho = rho) -
      pbivnorm(upper_x, lower_y, rho = rho) +
      pbivnorm(lower_x, lower_y, rho = rho)
  }

  res
}



# pearson correlation
# if no missing, solution is just cor(Y1,Y2) or cor(e1,e2)
# but if missing, two-step solution is NOT the same as cor(Y1,Y2) or cor(e1,e2)
lav_bvreg_cor_twostep_fit <- function(y1, y2, exo = NULL, wt = NULL,
                                      fit_y1 = NULL, fit_y2 = NULL,
                                      y1_name = NULL, y2_name = NULL,
                                      optim_method = "nlminb1",
                                      # optim.method = "none",
                                      optim_scale = 1,
                                      init_theta = NULL,
                                      control = list()) {
  # create cache environment
  cache <- lav_bvreg_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt,
    fit_y1 = fit_y1, fit_y2 = fit_y2
  )

  # the complete case is trivial
  if (!anyNA(cache$y1) && !anyNA(cache$y2)) {
    return(cache$theta[1L])
  }

  # optim.method
  minfns <- lav_uvbv_optim_fns(optim_method, lav_bvreg_min_fns())
  if (optim_method == "none") {
    return(cache$theta[1L])
  }

  # optimize
  if (is.null(control$trace)) {
    control$trace <- ifelse(lav_verbose(), 1, 0)
  }

  # init theta?
  if (!is.null(init_theta)) {
    start_x <- init_theta
  } else {
    start_x <- cache$theta
  }

  # try 1
  optim <- nlminb(
    start = start_x, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control,
    scale = optim_scale, lower = -0.999, upper = +0.999,
    cache = cache
  )

  # try 2 (scale = 10)
  if (optim$convergence != 0L) {
    optim <- nlminb(
      start = start_x, objective = minfns$objective,
      gradient = minfns$gradient, hessian = minfns$hessian,
      control = control,
      scale = 10, lower = -0.999, upper = +0.999,
      cache = cache
    )
  }

  # try 3 (start = 0, step.min = 0.1, gradient, no hessian)
  if (optim$convergence != 0L) {
    control$step.min <- 0.1
    # try again, with different starting value
    optim <- nlminb(
      start = 0, objective = minfns$objective,
      gradient = lav_bvreg_min_fns()$gradient, hessian = NULL,
      control = control,
      scale = optim_scale, lower = -0.999, upper = +0.999,
      cache = cache
    )
  }

  # check convergence
  if (optim$convergence != 0L) {
    if (!is.null(y1_name) && !is.null(y2_name)) {
      lav_msg_warn(gettextf(
        "estimation pearson correlation did not converge for variables %1$s
        and %2$s.", y1_name, y2_name))
    } else {
      lav_msg_warn(gettext(
        "estimation pearson correlation(s) did not always converge"))
    }

    # use init (as we always did in < 0.6-6; this is also what Mplus does)
    rho <- start_x
  } else {
    # store result
    rho <- optim$par
  }

  rho
}

# build the cache from user-level arguments: fit the univariate models if
# needed, apply parameter overrides, and initialize the cache environment
# (shared by lav_bvreg_cor_twostep_fit/_cor_sc/_logl/_lik)
lav_bvreg_cache_from_args <- function(y1, y2, exo = NULL, wt = NULL,
                                      rho = NULL,
                                      fit_y1 = NULL, fit_y2 = NULL,
                                      evar_y1 = NULL, beta_y1 = NULL,
                                      evar_y2 = NULL, beta_y2 = NULL,
                                      scores = FALSE) {
  if (is.null(fit_y1)) {
    fit_y1 <- lav_uvreg_fit(y = y1, x = exo, wt = wt)
  }
  if (is.null(fit_y2)) {
    fit_y2 <- lav_uvreg_fit(y = y2, x = exo, wt = wt)
  }

  # user specified parameters
  if (!is.null(evar_y1) || !is.null(beta_y1)) {
    fit_y1 <- lav_uvreg_update_fit(
      fit_y = fit_y1,
      evar_new = evar_y1, beta_new = beta_y1
    )
  }
  if (!is.null(evar_y2) || !is.null(beta_y2)) {
    fit_y2 <- lav_uvreg_update_fit(
      fit_y = fit_y2,
      evar_new = evar_y2, beta_new = beta_y2
    )
  }

  cache <- lav_bvreg_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, wt = wt,
    scores = scores
  )
  if (!is.null(rho)) {
    cache$theta <- rho
  }

  cache
}

# Y1 = linear
# Y2 = linear
lav_bvreg_init_cache <- function(fit_y1 = NULL,
                                 fit_y2 = NULL,
                                 wt = NULL,
                                 scores = FALSE,
                                 parent = parent.frame()) {
  # data
  y1 <- fit_y1$y
  y2 <- fit_y2$y
  exo <- fit_y1$x

  # Y1
  y1c <- y1 - fit_y1$yhat
  evar_y1 <- fit_y1$theta[fit_y1$var_idx]
  sd_y1 <- sqrt(evar_y1)
  eta_y1 <- fit_y1$yhat

  # Y2
  # (was fit_y2$theta[fit_y1$var_idx]: harmless as long as both fits share
  #  the same covariates, which all current callers do)
  y2c <- y2 - fit_y2$yhat
  evar_y2 <- fit_y2$theta[fit_y2$var_idx]
  sd_y2 <- sqrt(evar_y2)
  eta_y2 <- fit_y2$yhat


  # exo?
  if (is.null(exo)) {
    nexo <- 0L
  } else {
    nexo <- ncol(exo)
  }

  # nobs
  if (is.null(wt)) {
    n <- length(y1)
  } else {
    n <- sum(wt)
  }

  # starting value
  if (fit_y1$nexo > 0L) {
    e1 <- y1 - fit_y1$yhat
    e2 <- y2 - fit_y2$yhat
    if (is.null(wt)) {
      rho_init <- cor(e1, e2, use = "pairwise.complete.obs")
    } else {
      tmp <- na.omit(cbind(e1, e2, wt))
      rho_init <- cov.wt(tmp[, 1:2], wt = tmp[, 3], cor = TRUE)$cor[2, 1]
    }
  } else {
    if (is.null(wt)) {
      rho_init <- cor(y1, y2, use = "pairwise.complete.obs")
    } else {
      tmp <- na.omit(cbind(y1, y2, wt))
      rho_init <- cov.wt(tmp[, 1:2], wt = tmp[, 3], cor = TRUE)$cor[2, 1]
    }
  }

  # sanity check
  if (is.na(rho_init) || abs(rho_init) >= 1.0) {
    rho_init <- 0.0
  }

  # parameter vector
  theta <- rho_init # only

  # cache environment
  # (note: y1/y2 are only stored when scores = FALSE is not enough --
  #  they are needed by the twostep fit for the anyNA check)
  cache_list <- list(
    nexo = nexo, theta = theta, wt = wt, n = n,
    y1 = y1, y2 = y2, y1c = y1c, y2c = y2c,
    evar_y1 = evar_y1, sd_y1 = sd_y1, eta_y1 = eta_y1,
    evar_y2 = evar_y2, sd_y2 = sd_y2, eta_y2 = eta_y2
  )
  if (scores) {
    cache_list$exo <- exo
  }

  list2env(cache_list, parent = parent)
}


# casewise likelihoods, unweighted!
lav_bvreg_lik_cache <- function(cache = NULL) {
  with(cache, {   # nolint start
    rho <- theta[1L]

    cov_y12 <- rho * sqrt(evar_y1) * sqrt(evar_y2)
    sigma <- matrix(c(evar_y1, cov_y12, cov_y12, evar_y2), 2L, 2L)
    lik <- exp(lav_mvn_loglik_data(
      y = cbind(y1c, y2c), wt = NULL,
      mu = c(0, 0), sigma_1 = sigma,
      casewise = TRUE
    ))
    # only catch likelihoods that have underflowed to (essentially) zero;
    # we must NOT drop merely-small (but valid) likelihoods: 'lik' here is
    # a bivariate normal *density* that can be legitimately tiny for
    # outlying values of y1 and/or y2. The old threshold
    # sqrt(.Machine$double.eps) (~1.5e-8) discarded such cases, and because
    # *which* cases were dropped depended on rho, the objective became
    # non-smooth and the resulting correlation was biased (see the
    # analogous fix in lav_bvmix.R for the continuous/ordinal case).
    lik_toosmall_idx <- which(!(lik > 0))
    lik[lik_toosmall_idx] <- as.numeric(NA)

    return(lik)
  })             # nolint end
}

lav_bvreg_logl_cache <- function(cache = NULL) {
  with(cache, {       # nolint start
    lik <- lav_bvreg_lik_cache(cache) # unweighted!

    if (!is.null(wt)) {
      logl <- sum(wt * log(lik), na.rm = TRUE)
    } else {
      logl <- sum(log(lik), na.rm = TRUE)
    }

    return(logl)
  })                # nolint end
}

lav_bvreg_grad_cache <- function(cache = NULL) {
  with(cache, {     # nolint start
    rho <- theta[1L]
    r <- (1 - rho * rho)

    sd_y1_y2 <- sd_y1 * sd_y2
    t1 <- (y1c * y2c) / sd_y1_y2
    t2 <- (y1c * y1c) / evar_y1 - (2 * rho * t1) + (y2c * y2c) / evar_y2
    dx <- (rho + t1 - t2 * rho / r) / r

    # to be consistent with (log)lik_cache
    if (length(lik_toosmall_idx) > 0L) {
      dx[lik_toosmall_idx] <- as.numeric(NA)
    }

    if (is.null(wt)) {
      dx_rho <- sum(dx, na.rm = TRUE)
    } else {
      dx_rho <- sum(wt * dx, na.rm = TRUE)
    }

    return(dx_rho)
  })                  # nolint end
}

lav_bvreg_hessian_cache <- function(cache = NULL) {
  with(cache, {       # nolint start
    rho <- theta[1L]

    rho2 <- rho * rho
    r2 <- r * r
    r3 <- r * r * r

    h <- 1 / r - (2 * rho2 * t2) / r3 +
      2 * rho2 * (1 - t2 / r) / r2 + 4 * rho * t1 / r2 - t2 / r2

    # to be consistent with (log)lik_cache
    if (length(lik_toosmall_idx) > 0L) {
      h[lik_toosmall_idx] <- as.numeric(NA)
    }

    if (is.null(wt)) {
      h_1 <- sum(h, na.rm = TRUE)
    } else {
      h_1 <- sum(wt * h, na.rm = TRUE)
    }
    dim(h_1) <- c(1L, 1L) # for nlminb
    return(h_1)
  })               # nolint end
}

# nlminb objective/gradient/hessian (see lav_uvbv_common.R)
# note: this also fixes a dormant bug: the old lav_bvreg_min_hessian did
# not store 'x' in cache$theta before recomputing (never triggered: the
# default optim_method "nlminb1" does not use the hessian)
lav_bvreg_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_bvreg_logl_cache,
    grad_fun = lav_bvreg_grad_cache,
    hessian_fun = lav_bvreg_hessian_cache
  )
}

# casewise scores - cache
# FIXME: should we also set 'lik.toosmall.idx' cases to NA?
lav_bvreg_cor_sc_cache <- function(cache = NULL) {
  with(cache, {           # nolint start
    rho <- theta[1L]
    r <- (1 - rho * rho)

    # mu.y1
    dx_mu_y1 <- (2 * y1c / evar_y1 - 2 * rho * y2c / (sd_y1 * sd_y2)) / (2 * r)
    if (!is.null(wt)) {
      dx_mu_y1 <- wt * dx_mu_y1
    }

    # mu.y2
    dx_mu_y2 <- -(2 * rho * y1c / (sd_y1 * sd_y2) - 2 * y2c / evar_y2) / (2 * r)
    if (!is.null(wt)) {
      dx_mu_y2 <- wt * dx_mu_y2
    }

    # evar_y1
    dx_var_y1 <- -(0.5 / evar_y1 - ((y1c * y1c) / (evar_y1 * evar_y1) -
      rho * y1c * y2c / (evar_y1 * sd_y1 * sd_y2)) / (2 * r))
    if (!is.null(wt)) {
      dx_var_y1 <- wt * dx_var_y1
    }

    # var.y2
    dx_var_y2 <-
          -(0.5 / evar_y2 + (rho * y1c * y2c / (evar_y2 * sd_y1 * sd_y2) -
           (y2c * y2c) / (evar_y2 * evar_y2)) / (2 * r))
    if (!is.null(wt)) {
      dx_var_y2 <- wt * dx_var_y2
    }

    # sl.y1
    dx_sl_y1 <- NULL
    if (nexo > 0L) {
      dx_sl_y1 <- dx_mu_y1 * exo # weights already included in dx.mu.y1
    }

    # sl.y2
    dx_sl_y2 <- NULL
    if (nexo > 0L) {
      dx_sl_y2 <- dx_mu_y2 * exo # weights already included in dx.mu.y2
    }

    # rho
    z <- (y1c * y1c) / evar_y1 -
         2 * rho * y1c * y2c / (sd_y1 * sd_y2) + (y2c * y2c) / evar_y2
    dx_rho <- rho / r + (y1c * y2c / (sd_y1 * sd_y2 * r) - z * rho / (r * r))
    if (!is.null(wt)) {
      dx_rho <- wt * dx_rho
    }

    out <- list(
      dx_mu_y1 = dx_mu_y1, dx_var_y1 = dx_var_y1,
      dx_mu_y2 = dx_mu_y2, dx_var_y2 = dx_var_y2,
      dx_sl_y1 = dx_sl_y1, dx_sl_y2 = dx_sl_y2,
      dx_rho = dx_rho
    )
    return(out)
  })                       # nolint end
}

# casewise scores
#
# Y1 = linear
# Y2 = linear
lav_bvreg_cor_sc <- function(y1, y2, exo = NULL, wt = NULL,
                                 rho = NULL,
                                 fit_y1 = NULL, fit_y2 = NULL,
                                 evar_y1 = NULL, beta_y1 = NULL,
                                 evar_y2 = NULL, beta_y2 = NULL) {
  cache <- lav_bvreg_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    evar_y1 = evar_y1, beta_y1 = beta_y1,
    evar_y2 = evar_y2, beta_y2 = beta_y2,
    scores = TRUE
  )

  lav_bvreg_cor_sc_cache(cache = cache)
}

# logl - no cache
lav_bvreg_logl <- function(y1, y2, exo = NULL, wt = NULL,
                           rho = NULL,
                           fit_y1 = NULL, fit_y2 = NULL,
                           evar_y1 = NULL, beta_y1 = NULL,
                           evar_y2 = NULL, beta_y2 = NULL) {
  cache <- lav_bvreg_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    evar_y1 = evar_y1, beta_y1 = beta_y1,
    evar_y2 = evar_y2, beta_y2 = beta_y2,
    scores = TRUE
  )

  lav_bvreg_logl_cache(cache = cache)
}

# lik - no cache
lav_bvreg_lik <- function(y1, y2, exo = NULL, wt = NULL,
                          rho = NULL,
                          fit_y1 = NULL, fit_y2 = NULL,
                          evar_y1 = NULL, beta_y1 = NULL,
                          evar_y2 = NULL, beta_y2 = NULL,
                          .log = FALSE) {
  cache <- lav_bvreg_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    evar_y1 = evar_y1, beta_y1 = beta_y1,
    evar_y2 = evar_y2, beta_y2 = beta_y2,
    scores = TRUE
  )

  lik <- lav_bvreg_lik_cache(cache = cache)
  if (.log) {
    lik <- log(lik)
  }

  if (!is.null(wt)) {
    if (.log) {
      lik <- wt * lik
    } else {
      tmp <- wt * log(lik)
      lik <- exp(tmp)
    }
  }

  lik
}

# lav_options specific per estimator in separate functions LDW 06/04/2024

lav_options_est_ml <- function(opt) {
  # ML and friends: MLF, MLM, MLMV, MLMVS, MLR                     ####
  # se
  if (opt$se == "bootstrap" &&
      opt$estimator %in% c("mlf", "mlm", "mlmv", "mlmvs", "mlr")) {
    lav_msg_stop(gettext("use ML estimator for bootstrap"))
  } else if (opt$se == "default") {
    if (opt$estimator %in% c("ml", "mlf")) {
      opt$se <- "standard"
    } else if (opt$estimator %in% c("mlm", "mlmv", "mlmvs")) {
      opt$se <- "robust.sem"
    } else if (opt$estimator == "mlr") {
      opt$se <- "robust.huber.white"
    }
  } else if (opt$se == "robust") {
    if (opt$missing %in% c("ml", "ml.x")) {
      opt$se <- "robust.huber.white"
    } else if (opt$missing == "two.stage") { # needed?
      opt$se <- "two.stage"
    } else if (opt$missing == "robust.two.stage") { # needed?
      opt$se <- "robust.two.stage"
    } else {
      opt$se <- "robust.sem"
    }
  }
  # information
  if (opt$estimator == "mlf") {
    if (opt$information[1] == "default") {
      opt$information[1] <- "first.order"
    }
    if (length(opt$information) > 1L &&
        opt$information[2] == "default") {
      opt$information[2] <- "first.order"
    }
  }
  # test
  if (!opt$test[1] == "none") {
    if (opt$estimator %in% c("ml", "mlf")) {
      if (opt$test[1] == "default") {
        opt$test <- "standard"
      } # else {
      #    opt$test <- union("standard", opt$test)
      # }
    } else if (opt$estimator == "mlm") {
      if (opt$test[1] == "default") {
        opt$test <- "satorra.bentler"
      } else {
        opt$test <- union("satorra.bentler", opt$test)
      }
    } else if (opt$estimator == "mlmv") {
      if (opt$test[1] == "default") {
        opt$test <- "scaled.shifted"
      } else {
        opt$test <- union("scaled.shifted", opt$test)
      }
    } else if (opt$estimator == "mlmvs") {
      if (opt$test[1] == "default") {
        opt$test <- "mean.var.adjusted"
      } else {
        opt$test <- union("mean.var.adjusted", opt$test)
      }
    }
  }
  opt
}

lav_options_est_gls <- function(opt) {
  # GLS                                                            ####
  # FIXME: catch clustered, ...
  if (opt$.categorical) {
    lav_msg_stop(gettext(
      "ordered categorical data is not supported when estimator is GLS."))
  }
  if (opt$.multilevel) {
    # no two-level normal-theory weight matrix (yet); before 0.7-1 this
    # failed silently during estimation
    lav_msg_stop(gettext(
      "multilevel data is not supported when estimator is GLS."))
  }
  # two-stage missing data: estimation uses the (EM) saturated moments, while
  # the SEs use the robust.sem sandwich (with the two-stage NACOV) and the
  # scaled (satorra.bentler) test; se/test/missing are set in lav_options()
  two_stage <- any(opt$missing == c("two.stage", "robust.two.stage"))
  if (two_stage) {
    return(opt)
  }
  # se
  if (opt$se == "default") {
    # GLS uses a normal-theory weight matrix, so the 'standard' se/test are
    # only valid for normal, i.i.d. data. With sampling weights the moments
    # follow the sampling design, so default to the (ADF) sandwich se and the
    # ADF-based test instead (cf. ULS).
    if (isTRUE(opt$.sampling.weights)) {
      opt$se <- "robust.sem"
    } else {
      opt$se <- "standard"
    }
  } else if (opt$se == "robust") {
    opt$se <- "robust.sem"
  }
  # test
  if (opt$test[1] == "default") {
    if (isTRUE(opt$.sampling.weights)) {
      opt$test <- "browne.residual.adf"
    } else {
      opt$test <- "standard"
    }
  }
  bad_idx <- which(!opt$test %in% c(
    "standard", "none",
    "browne.residual.nt", # == standard
    "browne.residual.nt.model",
    "browne.residual.adf",
    "browne.residual.adf.model"
  ))
  if (length(bad_idx) > 0L) {
    lav_msg_stop(gettextf(
      "invalid value(s) in test= argument when estimator is GLS: %s.",
      paste(opt$test[bad_idx], collapse = " ")))
  }
  # missing
  opt$missing <- "listwise" # also pairwise?
  opt
}

lav_options_est_ntrls <- function(opt) {
  # NTRLS (experimental)                                           ####
  # optim.gradient
  opt$optim.gradient <- "numerical"
  # se
  if (opt$se == "default") {
    opt$se <- "standard"
  }
  # test
  if (opt$test[1] == "default") {
    opt$test <- "standard"
  }
  bad_idx <- which(!opt$test %in% c(
    "standard", "none",
    "browne.residual.nt",
    "browne.residual.nt.model",
    "browne.residual.adf",
    "browne.residual.adf.model"
  ))
  if (length(bad_idx) > 0L) {
    lav_msg_stop(gettextf(
      "invalid value(s) in test= argument when estimator is NTRLS: %s.",
      paste(opt$test[bad_idx], collapse = " ")))
  }
  # missing
  opt$missing <- "listwise"
  opt
}

lav_options_est_catml <- function(opt) {
  # catML (experimental)                                           ####
  # optim.gradient
  opt$optim.gradient <- "numerical" # for now
  # force correlation = TRUE, and categorical = FALSE
  opt$correlation <- TRUE
  opt$.categorical <- FALSE # we 'pretend' to have continuous data!
  # se
  if (opt$se == "default") {
    opt$se <- "robust.sem" # for now
  }
  # test
  if (opt$test[1] == "default") {
    opt$test <- "satorra.bentler"
  }
  # missing
  if (opt$missing %in% c("listwise", "pairwise")) {
    # nothing to do
  } else if (opt$missing == "default") {
    opt$missing <- "listwise"
  } else {
    lav_msg_stop(gettext(
      "missing argument should be listwise or pairwise if estimator is catML"))
  }
  opt
}

lav_options_est_wls <- function(opt) {
  # WLS                                                            ####
  # two-stage missing data: see lav_options_est_gls()
  if (any(opt$missing == c("two.stage", "robust.two.stage"))) {
    return(opt)
  }
  # se
  if (opt$se == "default") {
    opt$se <- "standard"
  } else if (opt$se == "robust.sem.nt") {
    # the WLS weight matrix is derived from the (single) NACOV slot,
    # which must then hold the ADF Gamma; a normal-theory sandwich is
    # not available
    lav_msg_stop(gettext(
      "se = \"robust.sem.nt\" is not supported when estimator is WLS;
      use se = \"robust.sem\" instead."))
  }
  # test
  if (opt$test[1] == "default") {
    opt$test <- "standard"
  }
  bad_idx <- which(!opt$test %in% c(
    "standard", "none",
    "browne.residual.nt",
    "browne.residual.nt.model",
    "browne.residual.adf", # == standard
    "browne.residual.adf.model"
  ))
  if (length(bad_idx) > 0L) {
    lav_msg_stop(gettextf(
      "invalid value(s) in test= argument when estimator is WLS: %s.",
      paste(opt$test[bad_idx], collapse = " ")))
  }
  # missing
  # opt$missing <- "listwise" (could be pairwise)
  opt
}

lav_options_est_dls <- function(opt) {
  # DLS                                                            ####
  # no categorical
  if (opt$.categorical) {
    lav_msg_stop(gettext(
      "ordered categorical data is not supported when estimator is DLS."))
  }
  # se
  if (opt$se == "default") {
    opt$se <- "robust.sem"
  } else if (opt$se == "robust.sem.nt") {
    # the DLS weight matrix is derived from the (single) NACOV slot,
    # which must then hold the ADF Gamma; a normal-theory sandwich is
    # not available
    lav_msg_stop(gettext(
      "se = \"robust.sem.nt\" is not supported when estimator is DLS;
      use se = \"robust.sem\" instead."))
  }
  # test
  if (opt$test[1] == "default") {
    opt$test <- "satorra.bentler"
  }
  bad_idx <- which(!opt$test %in% c(
    "standard", "none",
    "satorra.bentler",
    "browne.residual.nt", # == standard
    "browne.residual.nt.model",
    "browne.residual.adf",
    "browne.residual.adf.model"
  ))
  if (length(bad_idx) > 0L) {
    lav_msg_stop(gettextf(
      "invalid value(s) in test= argument when estimator is DLS: %s.",
      paste(opt$test[bad_idx], collapse = " ")))
  }
  # always include "satorra.bentler"
  if (opt$test[1] %in% c(
    "browne.residual.nt", "browne.residual.adf",
    "browne.residual.nt.model",
    "browne.residual.adf.model"
  )) {
    opt$test <- union("satorra.bentler", opt$test)
  }
  # missing (two-stage missing data keeps the EM moments; see
  # lav_options_est_gls())
  if (!any(opt$missing == c("two.stage", "robust.two.stage"))) {
    opt$missing <- "listwise"
  }
  # estimator.args
  if (is.null(opt$estimator.args)) {
    opt$estimator.args <- list(
      dls.a = 1.0, dls.GammaNT = "model",
      dls.FtimesNmin1 = FALSE
    )
  } else {
    if (is.null(opt$estimator.args$dls.a)) {
      opt$estimator.args$dls.a <- 1.0
    } else {
      stopifnot(is.numeric(opt$estimator.args$dls.a))
      if (opt$estimator.args$dls.a < 0.0 ||
          opt$estimator.args$dls.a > 1.0) {
        lav_msg_stop(gettext(
          "dls.a value in estimator.args must be between 0 and 1."))
      }
    }
    if (is.null(opt$estimator.args$dls.GammaNT)) {
      opt$estimator.args$dls.GammaNT <- "model"
    } else {
      stopifnot(is.character(opt$estimator.args$dls.GammaNT))
      opt$estimator.args$dls.GammaNT <-
        tolower(opt$estimator.args$dls.GammaNT)
      if (!opt$estimator.args$dls.GammaNT %in% c("sample", "model")) {
        lav_msg_stop(gettextf(
          "dls.GammaNT value in estimator.args must be either %s.",
          lav_msg_view(c("sample", "model"), log_sep = "or")))
      }
    }
    if (is.null(opt$estimator.args$dls.FtimesNminus1)) {
      opt$estimator.args$dls.FtimesNminus1 <- FALSE
    } else {
      stopifnot(is.logical(opt$estimator.args$dls.FtimesNminus1))
    }
  }
  if (opt$estimator.args$dls.GammaNT == "sample") {
    if (opt$optim.method %in% c("nlminb", "gn")) {
      # nothing to do
    } else if (opt$optim.method == "default") {
      opt$optim.method <- "gn"
    } else {
      lav_msg_stop(gettext(
        "optim.method must be either nlminb or gn if estimator is DLS."))
    }
  } else {
    if (opt$optim.method %in% c("gn")) {
      # nothing to do
    } else if (opt$optim.method == "default") {
      opt$optim.method <- "gn"
    } else if (opt$optim.method == "nlminb") {
      opt$optim.gradient <- "numerical"
    } else {
      lav_msg_stop(gettext(
        "optim.method must be either nlminb or gn if estimator is DLS."))
    }
  }
  opt
}

lav_options_est_dwls <- function(opt) {
  # DWLS, WLSM, WLSMV, WLSMVS                                      ####
  # new in 0.6-17: if !categorical, give a warning
  # changed in 0.7-1: catch this earlier
  #if (!opt$.categorical) {
  #  lav_msg_warn(gettextf(
  #    "estimator %s is not recommended for continuous data.
  #    Did you forget to set the ordered= argument?",
  #    dQuote(lav_options_estimatorgroup(opt$estimator))))
  #}
  # se
  if (opt$se == "bootstrap" &&
      opt$estimator %in% c("wlsm", "wlsmv", "wlsmvs")) {
    lav_msg_stop(gettext("use (D)WLS estimator for bootstrap"))
  } else if (opt$se == "default") {
    if (opt$estimator == "dwls" && !opt$.categorical) {
      # opt$se <- "standard"
      # with continuous data, DWLS follows the same defaults as ULS: with
      # sampling weights, use the ADF-based se (and, below, the ADF test)
      if (isTRUE(opt$.sampling.weights)) {
        opt$se <- "robust.sem"
      } else {
        opt$se <- "robust.sem.nt" # new in 0.6-21
      }
    } else {
      opt$se <- "robust.sem"
    }
  } else if (opt$se == "robust") {
    opt$se <- "robust.sem"
  } else if (opt$se == "robust.sem.nt" && opt$.categorical) {
    # no normal-theory Gamma in the categorical setting
    lav_msg_stop(gettext(
      "se = \"robust.sem.nt\" is not available for categorical data;
      use se = \"robust.sem\" instead."))
  }
  # test
  if (!opt$test[1] == "none") {
    if (opt$estimator == "dwls") {
      if (opt$test[1] == "default" && !opt$.categorical) {
        if (opt$se == "robust.sem") { # user-specified?
          opt$test <- "browne.residual.adf" # new in 0.7-1
          opt$standard.test <- "browne.residual.adf"
          opt$scaled.test <- "browne.residual.adf"
        } else {
          opt$test <- "browne.residual.nt" # new in 0.6-21
          opt$standard.test <- "browne.residual.nt"
          opt$scaled.test <- "browne.residual.nt"
        }
      } else if (opt$test[1] == "default" && opt$.categorical) {
        opt$test <- "standard" # bad choice!
      } else {
        opt$test <- union("standard", opt$test)
      }
    } else if (opt$estimator == "wlsm") {
      if (opt$test[1] == "default") {
        opt$test <- "satorra.bentler"
      } else {
        opt$test <- union("satorra.bentler", opt$test)
      }
    } else if (opt$estimator == "wlsmv") {
      if (opt$test[1] == "default") {
        opt$test <- "scaled.shifted"
      } else {
        opt$test <- union("scaled.shifted", opt$test)
      }
    } else if (opt$estimator == "wlsmvs") {
      if (opt$test[1] == "default") {
        opt$test <- "mean.var.adjusted"
      } else {
        opt$test <- union("mean.var.adjusted", opt$test)
      }
    }
  }
  opt
}

lav_options_est_uls <- function(opt) {
  # ULS, ULSM, ULSMV, ULSMVS                                       ####
  # two-stage missing data: see lav_options_est_gls()
  if (any(opt$missing == c("two.stage", "robust.two.stage"))) {
    return(opt)
  }
  # se
  if (opt$se == "bootstrap" &&
      opt$estimator %in% c("ulsm", "ulsmv", "ulsmvs")) {
    lav_msg_stop(gettext("use ULS estimator for bootstrap"))
  } else if (opt$se == "default") {
    if (opt$estimator == "uls" && !opt$.categorical) {
      #opt$se <- "standard"
      # with sampling weights, the (ADF) full Gamma is computed, so we can
      # default to the ADF-based robust SE (and, below, the ADF-based test)
      # instead of the normal-theory versions
      if (isTRUE(opt$.sampling.weights)) {
        opt$se <- "robust.sem"
      } else {
        opt$se <- "robust.sem.nt" # new in 0.6-21
      }
    } else {
      opt$se <- "robust.sem"
    }
  } else if (opt$se == "robust") {
    opt$se <- "robust.sem"
  } else if (opt$se == "robust.sem.nt" && opt$.categorical) {
    # no normal-theory Gamma in the categorical setting
    lav_msg_stop(gettext(
      "se = \"robust.sem.nt\" is not available for categorical data;
      use se = \"robust.sem\" instead."))
  }
  # test
  if (!opt$test[1] == "none") {
    if (opt$estimator == "uls") {
      if (opt$test[1] == "default" && !opt$.categorical) {
        if (opt$se == "robust.sem") { # user-specified?
          opt$test <- "browne.residual.adf" # new in 0.7-1
          opt$standard.test <- "browne.residual.adf"
          opt$scaled.test <- "browne.residual.adf"
        } else {
          opt$test <- "browne.residual.nt" # new in 0.6-21
          opt$standard.test <- "browne.residual.nt"
          opt$scaled.test <- "browne.residual.nt"
        }
      } else if (opt$test[1] == "default" && opt$.categorical) {
        opt$test <- "standard"
      } else {
        opt$test <- union("standard", opt$test)
      }
    } else if (opt$estimator == "ulsm") {
      if (opt$test[1] == "default") {
        opt$test <- "satorra.bentler"
      } else {
        opt$test <- union("satorra.bentler", opt$test)
      }
    } else if (opt$estimator == "ulsmv") {
      if (opt$test[1] == "default") {
        opt$test <- "scaled.shifted"
      } else {
        opt$test <- union("scaled.shifted", opt$test)
      }
    } else if (opt$estimator == "ulsmvs") {
      if (opt$test[1] == "default") {
        opt$test <- "mean.var.adjusted"
      } else {
        opt$test <- union("mean.var.adjusted", opt$test)
      }
    }
  }
  opt
}

lav_options_est_pml <- function(opt) {
  # PML                                                            ####
  # se
  if (opt$se == "default") {
    opt$se <- "robust.huber.white"
  }

  # information
  opt$information[1] <- "observed"
  if (length(opt$information) > 1L &&
      opt$information[2] == "default") {
    opt$information[2] <- "observed"
  }
  if (length(opt$observed.information) > 1L &&
      opt$observed.information[2] == "default") {
    opt$observed.information[2] <- "hessian"
  }

  # test
  if (length(opt$test) > 1L) {
    lav_msg_stop(gettext(
      "only a single test statistic is allowed when estimator is PML."))
  }
  if (!opt$test[1] == "none") {
    opt$test <- "mean.var.adjusted"
  }
  opt
}

lav_options_est_fml <- function(opt) {
  # FML - UMN                                                      ####
  # optim.gradient
  opt$optim.gradient <- "numerical"
  # se
  if (opt$se == "default") {
    opt$se <- "standard"
  }
  # information
  opt$information[1] <- "observed"
  if (length(opt$information) > 1L &&
      opt$information[2] == "default") {
    opt$information[2] <- "observed"
  }
  # test
  if (!opt$test[1] == "none") {
    opt$test <- "standard"
  }
  opt
}

lav_options_est_reml <- function(opt) {
  # REML                                                           ####
  # se
  if (opt$se == "default") {
    opt$se <- "standard"
  }
  # information
  opt$information[1] <- "observed"
  if (length(opt$information) > 1L &&
      opt$information[2] == "default") {
    opt$information[2] <- "observed"
  }
  # test
  if (!opt$test[1] == "none") {
    opt$test <- "standard"
  }
  # missing
  opt$missing <- "listwise"
  opt
}

lav_options_est_mml <- function(opt) {
  # MML                                                            ####
  # se
  if (opt$se == "default") {
    opt$se <- "standard"
  }
  # information
  opt$information[1] <- "observed"
  opt$meanstructure <- TRUE
  if (length(opt$information) > 1L &&
      opt$information[2] == "default") {
    opt$information[2] <- "observed"
  }
  # test
  opt$test <- "none"
  # link
  if (opt$link == "default") {
    # opt$link <- "logit"
    opt$link <- "probit"
  } else if (opt$link %in% c("logit", "probit")) {
    # nothing to do
  } else {
    lav_msg_stop(gettext("link must be `logit' or `probit'"))
  }
  # parameterization
  if (opt$parameterization == "default") {
    opt$parameterization <- "mml"
  } else {
    lav_msg_stop(gettext(
      "parameterization argument is ignored if estimator = MML"))
  }
  opt
}

lav_options_est_fabin <- function(opt) {
  # FABIN, MULTIPLE-GROUP-METHOD (MGM), BENTLER1982, ...           ####
  # experimental, for cfa or sam only
  # se
  if (opt$se == "default") {
    opt$se <- "none"
  }
  # bounds
  if (!is.null(opt$bounds) && opt$bounds == "default" &&
      length(opt$optim.bounds) == 0L) {
    opt$bounds <- "standard"
  }
  # test
  if (opt$test == "default") {
    opt$test <- "none" # for now
  }
  # missing
  opt$missing <- "listwise" # for now (until we have two-stage working)
  # options for fabin
  if (lav_options_estimatorgroup(opt$estimator) %in% c("FABIN2", "FABIN3")) {
    if (is.null(opt$estimator.args)) {
      opt$estimator.args <- list(thetapsi.method = "GLS")
    } else {
      if (is.null(opt$estimator.args$thetapsi.method)) {
        opt$estimator.args$thetapsi.method <- "GLS"
      } else {
        opt$estimator.args$thetapsi.method <-
          toupper(opt$estimator.args$thetapsi.method)
        if (opt$estimator.args$thetapsi.method %in% c(
          "ULS",
          "GLS", "WLS", "ULS.ML", "GLS.ML", "WLS.ML"
        )) {
          if (opt$estimator.args$thetapsi.method == "WLS") {
            opt$estimator.args$thetapsi.method <- "GLS"
          }
          if (opt$estimator.args$thetapsi.method == "WLS.ML") {
            opt$estimator.args$thetapsi.method <- "GLS.ML"
          }
        } else {
          lav_msg_stop(gettextf(
            "unknown value for estimator.args$thetapsi.method option: %s.",
            opt$estimator.args$thetapsi.method))
        }
      }
    }
  }
  # options for Bentler
  if (lav_options_estimatorgroup(opt$estimator) == "BENTLER1982") {
    if (is.null(opt$estimator.args)) {
      opt$estimator.args <- list(GLS = FALSE, quadprog = FALSE)
    } else {
      if (is.null(opt$estimator.args$GLS)) {
        opt$estimator.args$GLS <- FALSE
      }
      if (is.null(opt$estimator.args$quadprog)) {
        opt$estimator.args$quadprog <- FALSE
      }
    }
  }
  # options for guttman1952 multiple group method
  if (lav_options_estimatorgroup(opt$estimator) == "MGM") {
    if (is.null(opt$estimator.args)) {
      opt$estimator.args <- list(
        zero.after.efa = TRUE,
        psi.mapping = FALSE,
        quadprog = FALSE
      )
    } else {
      if (is.null(opt$estimator.args$zero.after.efa)) {
        opt$estimator.args$zero.after.efa <- TRUE
      }
      if (is.null(opt$estimator.args$psi.mapping)) {
        opt$estimator.args$psi.mapping <- FALSE
      }
      if (is.null(opt$estimator.args$quadprog)) {
        opt$estimator.args$quadprog <- FALSE
      }
    }
  }
  # brute-force override
  opt$optim.method <- "noniter"
  opt$start <- "simple"
  opt
}

lav_options_est_iv <- function(opt) {
  # (MI)IV-2SLS and friends                                          ####

  # the IV estimator.args use snake_case names
  if (is.list(opt$estimator.args) && length(opt$estimator.args) > 0L) {
    opt$estimator.args <- lav_snake_case(opt$estimator.args)
  }

  # brute-force override
  opt$optim.method <- "noniter"
  opt$marker.int.zero <- TRUE

  # ceq.simple lets the equation-by-equation estimator honor
  # them via a pooled (system) solve -- see lav_sem_miiv_pool_directed()
  opt$ceq.simple <- TRUE

  # missing data
  # - two.stage / robust.two.stage: use the (saturated) EM moments for point
  #   estimation, with standard errors corrected for the EM uncertainty (see
  #   lav_sem_miiv_vcov()); requires the sample-statistics path
  # - anything else falls back to listwise deletion
  two_stage <- any(opt$missing == c("two.stage", "robust.two.stage"))
  if (two_stage && isTRUE(opt$.categorical)) {
    lav_msg_warn(gettext(
      "[IV] two-stage missing-data handling is not available for categorical
       data; using listwise deletion instead."))
    two_stage <- FALSE
  }
  if (!two_stage) {
    opt$missing <- "listwise" # for now
  }

  # se: the general missing = two.stage handling (in lav_options()) sets se to
  # (robust.)two.stage, but the IV estimator uses its own SE machinery and
  # reads opt$missing to pick the two-stage moment covariance; reset to
  # "standard" here
  if (any(opt$se == c("default", "two.stage", "robust.two.stage"))) {
    opt$se <- "standard" # for now
  }
  # bounds
  if (!is.null(opt$bounds) && opt$bounds == "default" &&
      length(opt$optim.bounds) == 0L) {
    opt$bounds <- "standard"
  }
  # test (the general two.stage handling forces satorra.bentler; reset to the
  # IV default Browne residual test). With missing data the sample-based
  # version is not available, so use the model-based variant.
  if ((length(opt$test) == 1L && opt$test == "default") || two_stage) {
    if (opt$.categorical) {
      opt$test <- if (two_stage) {
        "browne.residual.adf.model"
      } else {
        "browne.residual.adf" # always sample-based
      }
    } else {
      opt$test <- if (two_stage) {
        "browne.residual.nt.model"
      } else {
        "browne.residual.nt" # sample-based (especially for baseline)
      }                       # model-based Sigma is here diagonal!
    }
  }
  opt$standard.test <- opt$test[1]

  # fixed.x
  opt$fixed.x <- FALSE # for now

  # sample.icov not needed
  opt$sample.icov <- FALSE

  # no loglik
  opt$loglik <- FALSE

  # no rescale of vcov
  opt$sample.cov.rescale <- FALSE

  # estimator options
  if (is.null(opt$estimator.args)) {
    # create default list
    opt$estimator.args <- list(iv_method = "2SLS",
                               iv_samplestats = TRUE,
                               iv_varcov_method = "RLS",
                               iv_sargan = TRUE,
                               iv_sargan_adjust = "none",
                               iv_weak = "warn",
                               iv_weak_threshold = 10,
                               iv_vcov_stage1 = "lm.vcov.dfres",
                               iv_vcov_stage2 = "h2",
                               iv_vcov_gamma_modelbased = TRUE,
                               iv_vcov_jack_numerical = FALSE,
                               iv_vcov_jaca_numerical = FALSE,
                               iv_vcov_jacb_numerical = FALSE,
                               iv_mimic_ml = FALSE)
  } else {
    if (is.null(opt$estimator.args$iv_method)) {
      opt$estimator.args$iv_method <- "2SLS"
    } else if (!opt$estimator.args$iv_method %in% "2SLS") {
      lav_msg_stop(gettext("iv_method should be 2SLS (for now)."))
    }
    if (is.null(opt$estimator.args$iv_samplestats)) {
      opt$estimator.args$iv_samplestats <- TRUE
    }
    # [[ ]] exact match: $ would partial-match iv_sargan to iv_sargan_adjust
    if (is.null(opt$estimator.args[["iv_sargan"]])) {
      opt$estimator.args$iv_sargan <- TRUE
    }
    # multiple-comparison adjustment for the per-equation Sargan p-values
    if (is.null(opt$estimator.args[["iv_sargan_adjust"]])) {
      opt$estimator.args$iv_sargan_adjust <- "none"
    } else if (!opt$estimator.args[["iv_sargan_adjust"]] %in%
               stats::p.adjust.methods) {
      lav_msg_stop(gettextf(
        "iv_sargan_adjust should be one of %s.",
        lav_msg_view(stats::p.adjust.methods, log_sep = "or")))
    }
    # NOTE: use [[ ]] (exact match); $ would partial-match iv_weak to
    # iv_weak_threshold
    if (is.null(opt$estimator.args[["iv_weak"]])) {
      opt$estimator.args$iv_weak <- "warn"
    } else if (!tolower(opt$estimator.args[["iv_weak"]]) %in%
               c("warn", "prune", "none")) {
      lav_msg_stop(gettext("iv_weak should be warn, prune, or none."))
    }
    if (is.null(opt$estimator.args[["iv_weak_threshold"]])) {
      opt$estimator.args$iv_weak_threshold <- 10
    } else if (!is.numeric(opt$estimator.args[["iv_weak_threshold"]]) ||
               opt$estimator.args[["iv_weak_threshold"]] < 0) {
      lav_msg_stop(gettext(
        "iv_weak_threshold should be a non-negative number."))
    }
    if (is.null(opt$estimator.args$iv_vcov_stage1)) {
      if (opt$.categorical) {
        opt$estimator.args$iv_vcov_stage1 <- "gamma"
      } else {
        opt$estimator.args$iv_vcov_stage1 <- "lm.vcov.dfres"
      }
      if (tolower(opt$se[1]) == "none") {
        opt$estimator.args$iv_vcov_stage1 <- "none"
      }
    } else if (!tolower(opt$estimator.args$iv_vcov_stage1) %in%
               c("lm.vcov", "lm.vcov.dfres", "gamma", "none")) {
      lav_msg_stop(gettext("iv_vcov_stage1 should be lm.vcov, lm.vcov.dfres,
                            gamma or none."))
    } else if (tolower(opt$estimator.args$iv_vcov_stage1) == "gamma" &&
               !opt$estimator.args$iv_samplestats) {
      lav_msg_stop(gettext("iv_vcov_stage1 cannot be gamma if
                            iv_samplestats is FALSE"))
    }
    if (is.null(opt$estimator.args$iv_vcov_stage2)) {
      if (tolower(opt$se[1]) == "none") {
        opt$estimator.args$iv_vcov_stage2 <- "none"
      } else {
        opt$estimator.args$iv_vcov_stage2 <- "h2"
      }
    } else if (!tolower(opt$estimator.args$iv_vcov_stage2) %in%
               c("h2", "delta", "none")) {
      lav_msg_stop(gettext("iv_vcov_stage2 should be h2, delta, or none."))
    } else if (tolower(opt$estimator.args$iv_vcov_stage2) == "h2" &&
               !opt$estimator.args$iv_samplestats) {
      lav_msg_stop(gettext("iv_vcov_stage2 should be delta (or none) if
                   iv_samplestats = FALSE."))
    }
    if (opt$.categorical) {
      opt$estimator.args$iv_samplestats <- TRUE
      opt$estimator.args$iv_varcov_method = "ULS"
    }
    if (is.null(opt$estimator.args$iv_varcov_method)) {
      opt$estimator.args$iv_varcov_method <- "RLS"
    } else if (!toupper(opt$estimator.args$iv_varcov_method) %in%
               c("ULS", "GLS", "2RLS", "RLS", "NONE")) {
      lav_msg_stop(gettext("iv_varcov_method should be ULS, GLS, 2RLS, RLS
                            or NONE."))
    }
    if (is.null(opt$estimator.args$iv_vcov_gamma_modelbased)) {
      opt$estimator.args$iv_vcov_gamma_modelbased <- TRUE
    }
    if (is.null(opt$estimator.args$iv_mean_structure)) {
      opt$estimator.args$iv_mean_structure <- "wls"
    } else if (!tolower(opt$estimator.args$iv_mean_structure) %in%
               c("moments", "wls")) {
      lav_msg_stop(gettext("iv_mean_structure should be moments or wls."))
    }
    if (is.null(opt$estimator.args$iv_vcov_jack_numerical)) {
      opt$estimator.args$iv_vcov_jack_numerical <- FALSE
    }
    if (is.null(opt$estimator.args$iv_vcov_jaca_numerical)) {
      opt$estimator.args$iv_vcov_jaca_numerical <- FALSE
    }
    if (is.null(opt$estimator.args$iv_vcov_jacb_numerical)) {
      opt$estimator.args$iv_vcov_jacb_numerical <- FALSE
    }
    if (is.null(opt$estimator.args$iv_mimic_ml)) {
      opt$estimator.args$iv_mimic_ml <- FALSE
    } else if (!is.logical(opt$estimator.args$iv_mimic_ml) ||
               length(opt$estimator.args$iv_mimic_ml) != 1L) {
      lav_msg_stop(gettext("iv_mimic_ml should be a single logical (TRUE/FALSE)."))
    }
  }

  # iv_mimic_ml: use the (maximum-likelihood) divisor N everywhere instead of
  # the default unbiased divisors
  if (isTRUE(opt$estimator.args$iv_mimic_ml)) {
    opt$sample.cov.rescale <- TRUE
  }

  if (two_stage) {
    opt$estimator.args$iv_samplestats <- TRUE
    if (tolower(opt$se) != "none" &&
        tolower(opt$estimator.args$iv_vcov_stage1) %in%
          c("lm.vcov", "lm.vcov.dfres")) {
      opt$estimator.args$iv_vcov_stage1 <- "gamma"
    }
  }

  # override test if iv_varcov_method is "none"
  if (tolower(opt$estimator.args$iv_varcov_method) == "none") {
    opt$test <- "none"
    opt$standard.test <- "none"
  }

  opt
}

# normalize the rbm.method estimator.arg (short forms allowed)
# ("none" is an undocumented value that falls back to plain ML; kept for
#  debugging / parity checks)
lav_options_est_rbm_method <- function(x) {
  if (length(x) == 0L) {
    return("implicit")
  }
  x <- tolower(as.character(x)[1L])
  if (x %in% c("none", "ml", "false")) {
    "none"
  } else if (startsWith(x, "e")) {
    "explicit" # explicit, erbm, e
  } else if (startsWith(x, "i")) {
    "implicit" # implicit, irbm, i
  } else {
    lav_msg_stop(gettextf(
      "estimator.args$rbm.method must be one of %s.",
      lav_msg_view(c("implicit", "explicit"), log_sep = "or")))
  }
}

lav_options_est_rbm <- function(opt) {
  # RBM reduced-bias M-estimation (continuous SEM, for now)  # FIXME
  #
  # penalized-ML point estimation; the discrepancy, implied moments, loglik,
  # test and information matrices are all the ML ones, so we keep the estimator
  # as "ml" internally. The dedicated fit function is triggered in step 11 by
  # the presence of estimator.args$rbm.method (the optimizer itself is nlminb).

  # estimator.args
  if (is.null(opt$estimator.args)) {
    opt$estimator.args <- list(rbm.method = "implicit")
  } else {
    opt$estimator.args$rbm.method <-
      lav_options_est_rbm_method(opt$estimator.args$rbm.method)
  }

  # information: the penalty/bias need the observed information
  opt$information[1] <- "observed"
  if (length(opt$information) > 1L &&
      opt$information[2] == "default") {
    opt$information[2] <- "observed"
  }

  # se: sandwich by default
  if (opt$se == "default") {
    opt$se <- "robust.huber.white"
  } else if (opt$se == "robust") {
    opt$se <- "robust.huber.white"
  }

  # test
  if (opt$test[1] == "default") {
    opt$test <- "standard"
  }

  # missing
  if (opt$missing == "default") {
    opt$missing <- "listwise" # for now
  }

  # the actual optimizer is nlminb (the rbm fit is triggered in step 11 by
  # estimator.args$rbm.method)
  opt$optim.method <- "nlminb"

  # internally, treat as ML for all downstream machinery
  opt$estimator <- "ml"

  opt
}

lav_options_est_none <- function(opt) {
  # NONE                                                           ####
  # se
  if (opt$se == "default") {
    opt$se <- "none"
  }
  # test
  if (opt$test[1] == "default") {
    opt$test <- "none"
  }
  opt
}

# tracing possibility in functions defined below, an example of use :
#
# in the function where you want to trace add a line
#     lav_trace(x)
# where x is a characterstring you want to show in the trace
#
# thereafter execute a script like this:
# library(lavaan)
# lavaan:::lav_trace_set(TRUE)
# model <- '
# # latent variable definitions
#    ind60 =~ x1 + x2 + x3
#    dem60 =~ y1 + a*y2 + b*y3 + c*y4
#    dem65 =~ y5 + a*y6 + b*y7 + c*y8
#
# # regressions
#   dem60 ~ ind60
#   dem65 ~ ind60 + dem60
#
# # residual correlations
#   y1 ~~ y5
#   y2 ~~ y4 + y6
#   y3 ~~ y7
#   y4 ~~ y8
#   y6 ~~ y8
# '
# fit <- sem(model, data = PoliticalDemocracy)
# summary(fit)
# lavaan:::lav_trace_set(FALSE)
# lavaan:::lav_trace_print("PolDem_trace.txt")
#

lav_trace <- function(content = "") {
  ignore_in_stack <- c(
    "eval", "try", "tryCatch", "tryCatchList", "tryCatchOne", "doTryCatch",
    "which", "unique", "as.list", "as.character", "unlist", "lav_trace",
    "source", "withVisible", "tryCatch.W.E", "withCallingHandlers", "do.call"
  )
  if (!exists("trace_on", lavaan_cache_env)) {
    return(invisible(NULL))
  }
  if (!exists("trace_nr", lavaan_cache_env))
              assign("trace_nr", 1L, lavaan_cache_env)
  tracenr <- get("trace_nr", lavaan_cache_env)
  x <- sub("[() ].*$", "", as.character(sys.calls()))
  if (length(x) == 0) {
    return(invisible(NULL))
  }
  a <- paste0("trc_", formatC(tracenr, format = "d", width = 5, flag = "0"))
  x <- x[!(x %in% ignore_in_stack)]
  if (length(x) > 0) {
    assign(a, list(stack = x, content = content, time = Sys.time()),
                        lavaan_cache_env)
    assign("trace_nr", tracenr + 1L, lavaan_cache_env)
  }

  invisible(NULL)
}

lav_trace_set <- function(state = NULL, silent = FALSE) {
  traceon <- exists("trace_on", lavaan_cache_env)
  msg <- ""
  if (is.null(state)) {
    rm(list = ls(lavaan_cache_env, pattern = "^trc_"), envir = lavaan_cache_env)
    if (exists("trace_nr", lavaan_cache_env))
        rm("trace_nr", envir = lavaan_cache_env)
    msg <- "Traces removed."
  } else if (state) {
    if (traceon) {
      msg <- "Trace already active!"
    } else {
      assign("trace_on", TRUE, lavaan_cache_env)
      msg <- "Trace on."
    }
  } else {
    if (traceon) {
      rm("trace_on", envir = lavaan_cache_env)
      msg <- "Trace off."
    } else {
      msg <- "Trace not active!"
    }
  }
  if (!silent) cat(msg, "\n", sep = "")

  invisible(NULL)
}

lav_trace_get <- function() {
  traceobjects <- ls(lavaan_cache_env, pattern = "^trc_")
  if (length(traceobjects) == 0) {
    return(list())
  }
  x <- mget(traceobjects, envir = lavaan_cache_env)
  x <- x[order(names(x))]

  x
}

lav_trace_print <- function(file = "", clean_after = (file != "")) {
  cat("Trace print on ", format(Sys.time(), format = "%F"), "\n\n", file = file)
  x <- lav_trace_get()
  for (x1 in x) {
    cat(format(x1$time, format = "%T"),
      paste(x1$stack, collapse = ">"), ":", x1$content,
      "\n",
      sep = " ", file = file, append = TRUE
    )
  }
  if (clean_after) lav_trace_set(NULL, TRUE)
}

lav_trace_summary <- function(file = "", clean_after = FALSE) {
  x <- lav_trace_get()
  temp <- new.env(parent = emptyenv())
  for (x1 in x) {
    nn <- length(x1$stack)
    mm <- paste(x1$stack[nn], paste(x1$stack[seq_len(nn - 1L)],
          collapse = ">"), sep = "\t")
    assign(mm, 1L + get0(mm, temp, ifnotfound = 0L), temp)
  }
  objects <- sort(ls(temp))
  for (i in seq_along(objects)) {
    cat(objects[i], get(objects[i], temp), "\n", file = file, append = TRUE)
  }
  if (clean_after) lav_trace_set(NULL, TRUE)
}

# `methods' for fitted lavaan objects
#
# standard (S4) methods:
# - show()
# - summary()
# - coef()
# - fitted.values() + fitted()
# - vcov()
# - logLik()
# - nobs()
# - update()
# - anova()

# lavaan-specific methods:
#
# - lavParameterEstimates()
# - standardizedSolution()
# - parameterTable()
# - varTable()


setMethod(
  "show", "lavaan",
  function(object) {
    # efa?
    efa_flag <- object@Options$model.type == "efa"

    # show only basic information
    res <- lav_object_summary(object,
      fit_measures = FALSE,
      estimates = FALSE,
      modindices = FALSE,
      efa = efa_flag
    )
    if (efa_flag) {
      # print (standardized) loadings only
      class(res) <- c("lavaan.efa", "list")
      print(res)
    } else {
      # print lavaan header
      print(res)
    }
    invisible(res)
  }
)

setMethod(
  "summary", "lavaan",
  function(object, header = TRUE,
           fit_measures = FALSE,
           residuals = FALSE,
           estimates = TRUE,
           ci = FALSE,
           fmi = FALSE,
           standardized = FALSE,
           std = standardized,
           std_nox = FALSE, # TODO: remove deprecated argument in early 2025
           remove_system_eq = TRUE,
           remove_eq = TRUE,
           remove_ineq = TRUE,
           remove_def = FALSE,
           remove_nonfree = FALSE,
           remove_step1 = TRUE,
           remove_unused = TRUE,
           plabel = FALSE,
           cov_std = TRUE,
           rsquare = FALSE,
           baseline_model = NULL,
           h1_model = NULL,
           fm_args = list(
             standard.test = "default",
             scaled.test = "default",
             rmsea.ci.level = 0.90,
             rmsea.h0.closefit = 0.05,
             rmsea.h0.notclosefit = 0.08,
             robust = TRUE,
             cat.nonpd = "na"
           ),
           modindices = FALSE,
           srmr_close_h0 = NULL,
           nd = 3L, cutoff = 0.3, dot_cutoff = 0.1, ...) {
    dotdotdot <- list(...)
    lav_adapt_func(environment(), dotdotdot, NULL)

    # efa?
    efa_flag <- object@Options$model.type == "efa"

    if (is.logical(fit_measures)) {
      fit_measures <- if (fit_measures) "default" else "none"
    }
    if (!is.list(fit_measures))
        fit_measures <- list(fit.measures = fit_measures)
    fit_measures <- lav_fit_measures_list_alias(fit_measures)
    if (is.logical(fit_measures$fit.measures)) {
      fit_measures$fit.measures <- if (fit_measures$fit.measures) "default" else "none"
    }
    if (!missing(fm_args)) {
      lav_deprecated_args("fit.measures", "fm_args")
      fit_measures <- c(fit_measures, fm_args)
    }
    res <- lav_object_summary(
      object = object, header = header,
      fit_measures = fit_measures, residuals = residuals,
      estimates = estimates,
      baseline_model = baseline_model,
      h1_model = h1_model,
      ci = ci, fmi = fmi, std = std, standardized = standardized,
      remove_system_eq = remove_system_eq,
      remove_eq = remove_eq, remove_ineq = remove_ineq,
      remove_def = remove_def, remove_nonfree = remove_nonfree,
      remove_step1 = remove_step1, remove_unused = remove_unused,
      plabel = plabel, cov_std = cov_std,
      rsquare = rsquare, efa = efa_flag,
      modindices = modindices,
      srmr_close_h0 = srmr_close_h0
    )
    # res has class c("lavaan.summary", "list")

    # what about nd? only used if we actually print; save as attribute
    attr(res, "nd") <- nd

    # if efa, add cutoff and dot.cutoff, and change class
    if (efa_flag) {
      # class(res) <- c("lavaan.summary.efa", "list")
      attr(res, "cutoff") <- cutoff
      attr(res, "dot.cutoff") <- dot_cutoff
    }

    res
  }
)


setMethod(
  "coef", "lavaan",
  function(object, type = "free", labels = TRUE, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
            sQuote("coef"))
        )
      }
    }
    # check object
    object <- lav_object_check_version(object)

    lav_inspect_coef(
      object = object, type = type,
      add_labels = labels, add_class = TRUE
    )
  }
)

# Warn when computing a *standardized* defined (:=) parameter is ambiguous.
#
# If a (user) label is attached to two or more model parameters -- ie they are
# constrained to be equal -- and that same label is referenced inside a defined
# parameter (:=), then the standardized value of the defined parameter is not
# well-defined: equal *unstandardized* coefficients need not have equal
# *standardized* coefficients (the latter also depend on the standard deviations
# of the variables involved, which typically differ across groups/equations).
# lavaan must still report a single number, and silently uses the value from the
# first occurrence of the label. This check issues an informative warning, as
# suggested by Keith Markus on the lavaan discussion list:
#   https://groups.google.com/g/lavaan/c/AWGORQnmKBg
#
# 'est.std' (optional) is the vector of standardized estimates aligned with the
# rows of 'partable'. When supplied, a label is only flagged if the standardized
# values of its (equal) parameters actually differ -- so the warning never fires
# when standardization happens to be invariant.
lav_partable_check_std_def <- function(partable, est.std = NULL,
                                       tol = 1e-08) {
  # we need labels, and at least one := definition
  if (is.null(partable$label)) {
    return(invisible(NULL))
  }
  def_idx <- which(partable$op == ":=")
  if (length(def_idx) == 0L) {
    return(invisible(NULL))
  }

  # labels attached to 2+ *model* parameters (these are the equality
  # constraints); exclude ==/</>/:= rows (the latter carries the def name)
  model_idx <- which(!partable$op %in% c("==", "<", ">", ":=") &
                       nzchar(partable$label))
  if (length(model_idx) == 0L) {
    return(invisible(NULL))
  }
  lab <- partable$label[model_idx]
  shared <- unique(lab[duplicated(lab)])
  if (length(shared) == 0L) {
    return(invisible(NULL))
  }

  def_rhs <- partable$rhs[def_idx]
  def_lhs <- partable$lhs[def_idx]
  flagged_lab <- character(0L)
  flagged_def <- character(0L)
  for (lb in shared) {
    # if standardized values are available, only flag when they truly differ
    if (!is.null(est.std)) {
      vals <- est.std[model_idx[lab == lb]]
      vals <- vals[is.finite(vals)]
      if (length(vals) < 2L ||
          diff(range(vals)) <= tol * max(1, max(abs(vals)))) {
        next
      }
    }
    # is the label referenced (as a whole token) inside any := definition?
    pat <- paste0("\\b", gsub("([][{}().|^$*+?\\\\])", "\\\\\\1", lb), "\\b")
    hit <- grepl(pat, def_rhs)
    if (any(hit)) {
      flagged_lab <- c(flagged_lab, lb)
      flagged_def <- unique(c(flagged_def, def_lhs[hit]))
    }
  }
  if (length(flagged_lab) == 0L) {
    return(invisible(NULL))
  }

  lav_msg_warn(gettextf(
    "the standardized value of the defined parameter(s) %s is ambiguous: it
     uses the label(s) %s, which are constrained equal across groups (or
     equations). Equal unstandardized coefficients can have different
     standardized coefficients, so lavaan reports the standardized defined
     parameter based on the first occurrence only. Interpret with caution.",
    paste(sQuote(flagged_def), collapse = ", "),
    paste(sQuote(flagged_lab), collapse = ", ")))

  invisible(list(def = flagged_def, label = flagged_lab))
}

standardizedSolution <- function(object,                     # nolint
                                 type = "std.all",
                                 se = TRUE,
                                 zstat = TRUE,
                                 pvalue = TRUE,
                                 ci = TRUE,
                                 level = 0.95,
                                 boot_ci_type = "perc",
                                 cov_std = TRUE,
                                 remove_eq = TRUE,
                                 remove_ineq = TRUE,
                                 remove_def = FALSE,
                                 remove_aux = TRUE,
                                 partable = NULL,
                                 glist = NULL,
                                 est = NULL,
                                 output = "data.frame",
                                 ...) {
    dotdotdot <- list(...)
    lav_adapt_func(environment(), dotdotdot, NULL)
    # check object
    object <- lav_object_check_version(object)

    # check type: either one of the std.* keywords, or a vector of (observed)
    # variable names, in which case only the parameters involving these
    # variables are standardized (a generalization of "std.nox")
    ov_std_user <- NULL
    std_types <- c("std.all", "std.lv", "std.nox")
    if (length(type) > 1L || !any(tolower(type) %in% std_types)) {
      ov_all <- lav_object_vnames(object, "ov")
      bad <- type[!(type %in% ov_all)]
      if (length(bad) > 0L) {
        lav_msg_warn(gettextf(
          "type= argument contains unknown observed variable(s): %s",
          paste(bad, collapse = ", ")))
      }
      ov_std_user <- type[type %in% ov_all]
      if (length(ov_std_user) == 0L) {
        lav_msg_stop(gettext(
          "type= must be one of std.lv, std.all, std.nox, or a vector of
           observed variable names."))
      }
      type <- "std.user"
    } else {
      type <- tolower(type)
      stopifnot(type %in% std_types)
      if (type == "std.nox") {
        # sanity checks (same as in parameterEstimates())
        if (length(lav_object_vnames(object, "ov.x")) == 0) {
          lav_msg_warn(gettext(
            "`std.nox' values are identical to `std.all' values in the
             absence of exogenous covariates."))
        } else if (!object@Options$fixed.x) {
          lav_msg_warn(gettext(
            "`std.nox' values are identical to `std.all' values when
             fixed.x = FALSE."))
        }
      }
    }

    # check output= argument
    output <- tolower(output)
    if (output %in% c("data.frame", "table")) {
      output <- "data.frame"
    } else if (output %in% c("text", "pretty")) {
      output <- "text"
    } else {
      lav_msg_stop(gettextf(
        "output must be %s or %s", sQuote("data.frame"), sQuote("text"))
      )
    }

    # no zstat + pvalue if estimator is Bayes
    if (object@Options$estimator == "Bayes") {
      zstat <- pvalue <- FALSE
    }

    # no se if class is not lavaan
    # using class() -- can't use inherits(), as this includes blavaan
    if (class(object)[1L] != "lavaan") {
      if (missing(se) || !se) {
        se <- FALSE
        zstat <- FALSE
        pvalue <- FALSE
      }
    }

    if (is.null(partable)) {
      tmp_partable <- lavInspect(object, "list")
    } else {
      tmp_partable <- partable
    }
    tmp_list <- tmp_partable[, c("lhs", "op", "rhs", "exo")]
    if (!is.null(tmp_partable$group)) {
      tmp_list$group <- tmp_partable$group
    }
    if (!is.null(tmp_partable$block)) {
      tmp_list$block <- tmp_partable$block
    }
    if (sum(nchar(tmp_partable$label)) != 0L) {
      tmp_list$label <- tmp_partable$label
    }

    # add std and std.all columns
    if (type == "std.lv") {
      tmp_list$est.std <- lav_standardize_lv(object,
        est = est, glist = glist,
        partable = partable, cov_std = cov_std
      )
    } else if (type == "std.all") {
      tmp_list$est.std <- lav_standardize_all(object,
        est = est, glist = glist,
        partable = partable, cov_std = cov_std
      )
    } else if (type == "std.nox") {
      tmp_list$est.std <- lav_standardize_all_nox(object,
        est = est, glist = glist,
        partable = partable, cov_std = cov_std
      )
    } else if (type == "std.user") {
      tmp_list$est.std <- lav_standardize_all(object,
        est = est, glist = glist,
        partable = partable, cov_std = cov_std, ov_std = ov_std_user
      )
    }

    # warn if a defined (:=) parameter standardizes a label that is constrained
    # equal across groups/equations (its standardized value is then ambiguous)
    lav_partable_check_std_def(tmp_partable, est.std = tmp_list$est.std)

    # bootstrap? If the model was fitted with se = "bootstrap" (and bootstrap
    # draws are available), we provide bootstrap standard errors and bootstrap
    # confidence intervals for the standardized parameters (just like
    # parameterEstimates() does for the unstandardized parameters). Every row
    # of the standardized solution is a function of the free parameters, so we
    # simply re-standardize each bootstrap draw.
    boot_std <- NULL
    boot_error_idx <- integer(0L)
    boot_bca_design <- NULL
    tmp_boot_free <- NULL
    tmp_fun_x <- NULL
    if (object@Options$se == "bootstrap" && se &&
        !is.null(object@boot$coef) && NROW(object@boot$coef) > 0L) {
      stopifnot(boot_ci_type %in%
        c("norm", "basic", "perc", "bca.simple", "bca"))
      # standardization as a function of the free parameter vector
      if (type == "std.lv") {
        tmp_fun_x <- lav_standardize_lv_x
      } else if (type == "std.all") {
        tmp_fun_x <- lav_standardize_all_x
      } else if (type == "std.nox") {
        tmp_fun_x <- lav_standardize_all_nox_x
      } else if (type == "std.user") {
        tmp_fun_x <- function(x, lavobject) {
          lav_standardize_all_x(x, lavobject = lavobject, ov_std = ov_std_user)
        }
      }
      tmp_boot_free <- lav_inspect_boot(object)
      boot_error_idx <- attr(tmp_boot_free, "error.idx")
      if (length(boot_error_idx) > 0L) {
        tmp_boot_free <- tmp_boot_free[-boot_error_idx, , drop = FALSE]
      }
      boot_std <- try(apply(tmp_boot_free, 1L, tmp_fun_x, lavobject = object),
                      silent = TRUE)
      if (inherits(boot_std, "try-error") || is.null(boot_std)) {
        boot_std <- NULL
      } else {
        if (is.matrix(boot_std)) {
          boot_std <- t(boot_std)
        } else {
          boot_std <- as.matrix(boot_std)
        }
        boot_std[!is.finite(boot_std)] <- as.numeric(NA)
        # label the columns like lavInspect(object, "coef.boot"); one column
        # per row of the (full) standardized solution, kept in sync with the
        # returned rows below (see the remove.* blocks) so users can retrieve
        # the standardized estimates in the bootstrap samples via
        # attr(., "boot_std")
        colnames(boot_std) <- lav_pt_labels(tmp_partable)
      }
    }

    if (object@Options$se != "none" && se && !is.null(boot_std)) {
      # bootstrap standard errors (MLE-normalised covariance, as elsewhere)
      nboot <- nrow(boot_std)
      tmp <- apply(boot_std, 2L, sd, na.rm = TRUE) * sqrt((nboot - 1) / nboot)
      # catch near-zero SEs
      zero_idx <- which(tmp < .Machine$double.eps^(1 / 4))
      if (length(zero_idx) > 0L) {
        tmp[zero_idx] <- 0.0
      }
      tmp_list$se <- tmp
      if (zstat) {
        tmp_se <- ifelse(tmp_list$se == 0.0, NA, tmp_list$se)
        tmp_list$z <- tmp_list$est.std / tmp_se
      }
      if (zstat && pvalue) {
        tmp_list$pvalue <- 2 * (1 - pnorm(abs(tmp_list$z)))
      }
    } else if (object@Options$se != "none" && se) {
      # add 'se' for standardized parameters
      tmp_vcov <- try(lav_inspect_vcov(object,
        standardized = TRUE,
        type = type, ov_std = ov_std_user, free_only = FALSE,
        add_labels = FALSE,
        add_class = FALSE
      ))
      if (inherits(tmp_vcov, "try-error") || is.null(tmp_vcov)) {
        tmp_list$se <- rep(NA, length(tmp_list$lhs))
        if (zstat) {
          tmp_list$z <- rep(NA, length(tmp_list$lhs))
        }
        if (pvalue) {
          tmp_list$pvalue <- rep(NA, length(tmp_list$lhs))
        }
      } else {
        tmp <- diag(tmp_vcov)
        # catch negative values
        min_idx <- which(tmp < 0)
        if (length(min_idx) > 0L) {
          tmp[min_idx] <- as.numeric(NA)
        }
        # now, we can safely take the square root
        tmp <- sqrt(tmp)

        # catch near-zero SEs
        zero_idx <- which(tmp < .Machine$double.eps^(1 / 4)) # was 1/2 < 0.6
        # was 1/3 < 0.6-9
        if (length(zero_idx) > 0L) {
          tmp[zero_idx] <- 0.0
        }
        tmp_list$se <- tmp

        # add 'z' column
        if (zstat) {
          tmp_se <- ifelse(tmp_list$se == 0.0, NA, tmp_list$se)
          tmp_list$z <- tmp_list$est.std / tmp_se
        }
        if (zstat && pvalue) {
          tmp_list$pvalue <- 2 * (1 - pnorm(abs(tmp_list$z)))
        }
      }
    }

    # simple symmetric confidence interval
    if (se && object@Options$se != "none" && ci) {
      # next three lines based on confint.lm
      a <- (1 - level) / 2
      a <- c(a, 1 - a)
      fac <- qnorm(a)
      ci <- tmp_list$est.std + tmp_list$se %o% fac

      if (!is.null(boot_std)) {
        # bootstrap confidence intervals (shared with parameterEstimates() and
        # lavEffects() via lav_bootstrap_ci(); see lav_bootstrap.R). Every row
        # is a derived quantity, so the CI is computed uniformly for all rows.
        t0 <- tmp_list$est.std
        if (boot_ci_type == "bca") {
          design <- lav_bootstrap_bca_design(object, boot_error_idx)
          stopifnot(nrow(design) == nrow(boot_std))
          # keep the empirical-influence design matrix, so users can retrieve
          # it via attr(., "design") (it is computed anyway for BCa)
          boot_bca_design <- design
          acc <- lav_bootstrap_acceleration(boot_std, design)
          ci <- lav_bootstrap_ci(boot_t = boot_std, t0 = t0,
            boot_ci_type = "bca", level = level, acc = acc)
        } else {
          # For "norm", the bias is evaluated at the mean of the bootstrapped
          # (free) parameters, exactly as in parameterEstimates().
          bias <- if (boot_ci_type == "norm") {
            as.numeric(tmp_fun_x(colMeans(tmp_boot_free, na.rm = TRUE),
                                 lavobject = object)) - t0
          } else {
            NULL
          }
          ci <- lav_bootstrap_ci(boot_t = boot_std, t0 = t0,
            boot_ci_type = boot_ci_type, level = level,
            se = tmp_list$se, bias = bias)
        }
      } else {
        # Monte Carlo: asymmetric percentile-based CI for standardized
        # defined parameters (Preacher & Selig 2012)
        mc_coef <- lav_inspect_mc(object)
        def_idx <- which(tmp_list$op == ":=")
        if (!is.null(mc_coef) && length(def_idx) > 0L) {
          if (type == "std.lv") {
            tmp_fun_mc <- lav_standardize_lv_x
          } else if (type == "std.all") {
            tmp_fun_mc <- lav_standardize_all_x
          } else if (type == "std.nox") {
            tmp_fun_mc <- lav_standardize_all_nox_x
          } else if (type == "std.user") {
            tmp_fun_mc <- function(x, lavobject) {
              lav_standardize_all_x(x, lavobject = lavobject,
                                    ov_std = ov_std_user)
            }
          }
          mc_std <- try(apply(mc_coef, 1L, tmp_fun_mc,
                              lavobject = object), silent = TRUE)
          if (!inherits(mc_std, "try-error")) {
            if (is.matrix(mc_std)) {
              mc_std <- t(mc_std)
            } else {
              mc_std <- as.matrix(mc_std)
            }
            alpha <- c((1 - level) / 2, 1 - (1 - level) / 2)
            mc_qq <- apply(mc_std[, def_idx, drop = FALSE], 2L,
                           quantile, probs = alpha, na.rm = TRUE, type = 7)
            ci[def_idx, ] <- t(mc_qq)
          }
        }
      }

      tmp_list$ci.lower <- ci[, 1]
      tmp_list$ci.upper <- ci[, 2]
    }


    # if single group, remove group column
    if (object@Data@ngroups == 1L) tmp_list$group <- NULL

    # remove == rows?
    if (remove_eq) {
      eq_idx <- which(tmp_list$op == "==")
      if (length(eq_idx) > 0L) {
        tmp_list <- tmp_list[-eq_idx, ]
        if (!is.null(boot_std)) {
          boot_std <- boot_std[, -eq_idx, drop = FALSE]
        }
      }
    }
    # remove <> rows?
    if (remove_ineq) {
      ineq_idx <- which(tmp_list$op %in% c("<", ">"))
      if (length(ineq_idx) > 0L) {
        tmp_list <- tmp_list[-ineq_idx, ]
        if (!is.null(boot_std)) {
          boot_std <- boot_std[, -ineq_idx, drop = FALSE]
        }
      }
    }
    # remove := rows?
    if (remove_def) {
      def_idx <- which(tmp_list$op == ":=")
      if (length(def_idx) > 0L) {
        tmp_list <- tmp_list[-def_idx, ]
        if (!is.null(boot_std)) {
          boot_std <- boot_std[, -def_idx, drop = FALSE]
        }
      }
    }
    # remove auxiliary (saturated-correlates) rows? (FIML aux= variables)
    if (remove_aux && length(object@Options$aux) > 0L) {
      aux_idx <- which(tmp_list$lhs %in% object@Options$aux |
        tmp_list$rhs %in% object@Options$aux)
      if (length(aux_idx) > 0L) {
        tmp_list <- tmp_list[-aux_idx, ]
        if (!is.null(boot_std)) {
          boot_std <- boot_std[, -aux_idx, drop = FALSE]
        }
      }
    }

    # remove random-slope phantom marker rows (eg s1 =~ s1, fixed to zero)
    if (!is.null(object@ParTable$rv) &&
        any(nchar(object@ParTable$rv) > 0L)) {
      rv_names <- unique(object@ParTable$rv[
        nchar(object@ParTable$rv) > 0L])
      rv_idx <- which(tmp_list$op == "=~" &
        tmp_list$lhs == tmp_list$rhs &
        tmp_list$lhs %in% rv_names)
      if (length(rv_idx) > 0L) {
        tmp_list <- tmp_list[-rv_idx, ]
        if (!is.null(boot_std)) {
          boot_std <- boot_std[, -rv_idx, drop = FALSE]
        }
      }
    }

    # remove attribute for data order
    attr(tmp_list, "ovda") <- NULL

    if (output == "text") {
      class(tmp_list) <- c(
        "lavaan.parameterEstimates", "lavaan.data.frame",
        "data.frame"
      )
      # tmp_list$exo is needed for printing, don't remove it
      attr(tmp_list, "group.label") <- object@Data@group.label
      attr(tmp_list, "level.label") <- object@Data@level.label
      # attr(tmp_list, "header") <- FALSE
    } else {
      tmp_list$exo <- NULL
      tmp_list$block <- NULL
      class(tmp_list) <- c("lavaan.data.frame", "data.frame")
    }

    # store the standardized estimates in the bootstrap samples (issue #597)
    # and, for BCa, the empirical-influence design matrix (issue #598), so
    # users can retrieve both without recomputing them
    if (!is.null(boot_std)) {
      attr(tmp_list, "boot_std") <- boot_std
    }
    if (!is.null(boot_bca_design)) {
      attr(tmp_list, "design") <- boot_bca_design
    }

    tmp_list
  }

lavParameterEstimates <- function(object,                      # nolint
                                  # select columns
                                  se = TRUE,
                                  zstat = TRUE,
                                  pvalue = TRUE,
                                  ci = TRUE,
                                  standardized = FALSE,
                                  fmi = FALSE,
                                  plabel = FALSE,
                                  # control
                                  level = 0.95,
                                  boot_ci_type = "perc",
                                  cov_std = TRUE,
                                  fmi_options = list(),
                                  # add rows
                                  rsquare = FALSE,
                                  # remove rows
                                  remove_system_eq = TRUE,
                                  remove_eq = TRUE,
                                  remove_ineq = TRUE,
                                  remove_def = FALSE,
                                  remove_nonfree = FALSE,
                                  remove_step1 = TRUE,
                                  remove_unused = FALSE,
                                  remove_aux = TRUE,
                                  # output
                                  add_attributes = FALSE,
                                  output = "data.frame",
                                  header = FALSE,
                                  ...) {

   dotdotdot <- list(...)
   lav_adapt_func(environment(), dotdotdot, NULL)

  # lavaan.fsr?
    if (inherits(object, "lavaan.fsr")) {
      return(object$PE)
    }

    # check object
    object <- lav_object_check_version(object)

    # deprecated add_attributes (for psycho/blavaan)
    if (add_attributes) {
      output <- "text"
    }


    # no se if class is not lavaan
    # can't use inherits(), as this would return TRUE if object is from blavaan
    if (class(object)[1L] != "lavaan") {
      if (missing(se) || !se) {
        se <- FALSE
        zstat <- FALSE
        pvalue <- FALSE
      }
    }

    # check output= argument
    output <- tolower(output)
    if (output %in% c("data.frame", "table")) {
      output <- "data.frame"
      header <- FALSE
    } else if (output %in% c("text", "pretty")) {
      output <- "text"
    } else {
      lav_msg_stop(gettextf(
        "output must be %s or %s", sQuote("data.frame"), sQuote("text"))
      )
    }

    # check fmi
    if (fmi) {
      if (inherits(object, "lavaanList")) {
        lav_msg_warn(gettext(
          "fmi not available for object of class \"lavaanList\""))
        fmi <- FALSE
      }
      if (object@Options$se != "standard") {
        lav_msg_warn(gettext(
          "fmi only available if se = \"standard\""))
        fmi <- FALSE
      }
      if (object@Options$estimator != "ML") {
        lav_msg_warn(gettext(
          "fmi only available if estimator = \"ML\""))
        fmi <- FALSE
      }
      if (!object@SampleStats@missing.flag) {
        lav_msg_warn(gettext(
          "fmi only available if missing = \"(fi)ml\""))
        fmi <- FALSE
      }
      if (!object@optim$converged) {
        lav_msg_warn(gettext(
          "fmi not available; model did not converge"))
        fmi <- FALSE
      }
    }

    # no zstat + pvalue if estimator is Bayes
    if (object@Options$estimator == "Bayes") {
      zstat <- pvalue <- FALSE
    }

    tmp_partable <- as.data.frame(object@ParTable, stringsAsFactors = FALSE)
    tmp_list <- tmp_partable[, c("lhs", "op", "rhs", "free")]
    if (!is.null(tmp_partable$user)) {
      tmp_list$user <- tmp_partable$user
    }
    if (!is.null(tmp_partable$block)) {
      tmp_list$block <- tmp_partable$block
    } else {
      tmp_list$block <- rep(1L, length(tmp_list$lhs))
    }
    if (!is.null(tmp_partable$level)) {
      tmp_list$level <- tmp_partable$level
    } else {
      tmp_list$level <- rep(1L, length(tmp_list$lhs))
    }
    if (!is.null(tmp_partable$group)) {
      tmp_list$group <- tmp_partable$group
    } else {
      tmp_list$group <- rep(1L, length(tmp_list$lhs))
    }
    if (!is.null(tmp_partable$step)) {
      tmp_list$step <- tmp_partable$step
    }
    if (!is.null(tmp_partable$efa)) {
      tmp_list$efa <- tmp_partable$efa
    }
    if (!is.null(tmp_partable$label)) {
      tmp_list$label <- tmp_partable$label
    } else {
      tmp_list$label <- rep("", length(tmp_list$lhs))
    }
    if (!is.null(tmp_partable$exo)) {
      tmp_list$exo <- tmp_partable$exo
    } else {
      tmp_list$exo <- rep(0L, length(tmp_list$lhs))
    }
    if (inherits(object, "lavaanList")) {
      # per default: nothing!
      # if("partable" %in% object@meta$store.slots) {
      #    COF <- sapply(object@ParTableList, "[[", "est")
      #    tmp_list$est <- rowMeans(COF)
      # }
      tmp_list$est <- NULL
    } else if (!is.null(tmp_partable$est)) {
      tmp_list$est <- tmp_partable$est
    } else {
      tmp_list$est <- lav_model_get_parameters(object@Model,
        type = "user",
        extra = TRUE
      )
    }
    if (!is.null(tmp_partable$lower)) {
      tmp_list$lower <- tmp_partable$lower
      tmp_list$lower[tmp_list$free > 0L] <-
        tmp_list$lower[tmp_list$free > 0L] + .Machine$double.eps
    }
    if (!is.null(tmp_partable$upper)) {
      tmp_list$upper <- tmp_partable$upper
      tmp_list$upper[tmp_list$free > 0L] <-
        tmp_list$upper[tmp_list$free > 0L] - .Machine$double.eps
    }


    # add se, zstat, pvalue
    if (se && object@Options$se != "none") {
      tmp_list$se <- lav_inspect_se(object)
      # handle tiny SEs
      tmp_list$se <- ifelse(tmp_list$se < sqrt(.Machine$double.eps),
        0, tmp_list$se
      )
      tmp_se <- ifelse(tmp_list$se < sqrt(.Machine$double.eps), NA, tmp_list$se)
      if (zstat) {
        tmp_list$z <- tmp_list$est / tmp_se
        if (pvalue) {
          tmp_list$pvalue <- 2 * (1 - pnorm(abs(tmp_list$z)))
          # remove p-value if bounds have been used
          if (!is.null(tmp_partable$lower)) {
            b_idx <- which(abs(tmp_partable$lower - tmp_partable$est) <
              sqrt(.Machine$double.eps) &
              tmp_partable$free > 0L)
            if (length(b_idx) > 0L) {
              tmp_list$pvalue[b_idx] <- as.numeric(NA)
            }
          }
          if (!is.null(tmp_partable$upper)) {
            b_idx <- which(abs(tmp_partable$upper - tmp_partable$est) <
              sqrt(.Machine$double.eps) &
              tmp_partable$free > 0L)
            if (length(b_idx) > 0L) {
              tmp_list$pvalue[b_idx] <- as.numeric(NA)
            }
          }
        }
      }
    }

    # extract bootstrap data (if any)
    if (object@Options$se == "bootstrap" ||
      "bootstrap" %in% object@Options$test ||
      "bollen.stine" %in% object@Options$test) {
      tmp_boot <- lav_inspect_boot(object)
      bootstrap_seed <- attr(tmp_boot, "seed") # for bca
      error_idx <- attr(tmp_boot, "error.idx")
      if (length(error_idx) > 0L) {
        tmp_boot <- tmp_boot[-error_idx, , drop = FALSE] # drops attributes
      }
    } else {
      tmp_boot <- NULL
    }

    bootstrap_successful <- NROW(tmp_boot) # should be zero if NULL
    boot_bca_design <- NULL # BCa empirical-influence design matrix (issue #598)

    # confidence interval
    if (se && object@Options$se != "none" && ci) {
      # next three lines based on confint.lm
      a <- (1 - level) / 2
      a <- c(a, 1 - a)
      if (object@Options$se != "bootstrap") {
        fac <- qnorm(a)
        ci <- tmp_list$est + tmp_list$se %o% fac
        # Monte Carlo: replace CI for defined parameters with
        # asymmetric percentile-based bounds (Preacher & Selig 2012)
        mc_coef <- lav_inspect_mc(object)
        def_idx <- which(object@ParTable$op == ":=")
        if (!is.null(mc_coef) && length(def_idx) > 0L) {
          mc_def <- apply(mc_coef, 1L, object@Model@def.function)
          if (length(def_idx) == 1L) {
            mc_def <- as.matrix(mc_def)
          } else {
            mc_def <- t(mc_def)
          }
          mc_def[!is.finite(mc_def)] <- as.numeric(NA)
          alpha <- c((1 - level) / 2, 1 - (1 - level) / 2)
          mc_qq <- apply(mc_def, 2L, quantile, probs = alpha,
                         na.rm = TRUE, type = 7)
          ci[def_idx, ] <- t(mc_qq)
        }
      } else if (object@Options$se == "bootstrap") {
        # bootstrap confidence interval; the order-statistic interpolation
        # (boot package 'norm.inter') is shared with lavEffects() via
        # lav_bootstrap_norm_inter() (see lav_bootstrap.R)
        stopifnot(!is.null(tmp_boot))
        stopifnot(boot_ci_type %in% c(
          "norm", "basic", "perc",
          "bca.simple", "bca"
        ))
        if (boot_ci_type == "norm") {
          fac <- qnorm(a)
          boot_x <- colMeans(tmp_boot, na.rm = TRUE)
          boot_est <-
            lav_model_get_parameters(object@Model,
              glist = lav_model_x2glist(object@Model, boot_x),
              type = "user", extra = TRUE
            )
          bias_est <- (boot_est - tmp_list$est)
          ci <- (tmp_list$est - bias_est) + tmp_list$se %o% fac
        } else if (boot_ci_type %in% c("basic", "perc", "bca.simple")) {
          # the per-quantity bound computation is shared with lavEffects()
          # via lav_bootstrap_ci() (see lav_bootstrap.R)
          ci <- cbind(tmp_list$est, tmp_list$est)

          # free parameters
          free_idx <- which(object@ParTable$free &
            !duplicated(object@ParTable$free))
          ci[free_idx, ] <- lav_bootstrap_ci(
            boot_t = tmp_boot, t0 = tmp_list$est[free_idx],
            boot_ci_type = boot_ci_type, level = level
          )

          # defined parameters
          def_idx <- which(object@ParTable$op == ":=")
          if (length(def_idx) > 0L) {
            boot_def <- apply(tmp_boot, 1L, object@Model@def.function)
            if (length(def_idx) == 1L) {
              boot_def <- as.matrix(boot_def)
            } else {
              boot_def <- t(boot_def)
            }
            ci[def_idx, ] <- lav_bootstrap_ci(
              boot_t = boot_def, t0 = tmp_list$est[def_idx],
              boot_ci_type = boot_ci_type, level = level
            )
          }

          # TODO: add cin/ceq?
        } else if (boot_ci_type == "bca") { # new in 0.6-12
          # we assume that the 'ordinary' (nonparametric) bootstrap was used.
          # The empirical-influence design matrix and the acceleration/CI
          # computations are shared with lavEffects() (see lav_bootstrap.R).
          design <- lav_bootstrap_bca_design(object, error_idx)
          stopifnot(nrow(design) == nrow(tmp_boot))
          # keep the design matrix, so users can retrieve it via
          # attr(., "design") (it is computed anyway for BCa)
          boot_bca_design <- design

          ci <- cbind(tmp_list$est, tmp_list$est)

          # free parameters
          free_idx <- which(object@ParTable$free &
            !duplicated(object@ParTable$free))
          stopifnot(length(free_idx) == ncol(tmp_boot))
          acc <- lav_bootstrap_acceleration(tmp_boot, design)
          ci[free_idx, ] <- lav_bootstrap_ci(
            boot_t = tmp_boot, t0 = tmp_list$est[free_idx],
            boot_ci_type = "bca", level = level, acc = acc
          )

          # defined parameters
          def_idx <- which(object@ParTable$op == ":=")
          if (length(def_idx) > 0L) {
            boot_def <- apply(tmp_boot, 1L, object@Model@def.function)
            if (length(def_idx) == 1L) {
              boot_def <- as.matrix(boot_def)
            } else {
              boot_def <- t(boot_def)
            }
            acc <- lav_bootstrap_acceleration(boot_def, design)
            ci[def_idx, ] <- lav_bootstrap_ci(
              boot_t = boot_def, t0 = tmp_list$est[def_idx],
              boot_ci_type = "bca", level = level, acc = acc
            )
          }

          # TODO: add cin/ceq?
        }

        # ceq.simple.only (e.g. estimator = "IV" with equality constraints):
        # parameters constrained equal share a single bootstrap column, so only
        # the 'leader' row (first occurrence of each free index) received a CI
        # above; copy it to the (duplicated) followers so they match instead of
        # collapsing to the point estimate
        if (object@Model@ceq.simple.only) {
          free_pt <- object@ParTable$free
          dup_idx <- which(free_pt > 0L & duplicated(free_pt))
          if (length(dup_idx) > 0L) {
            leader_idx <- match(free_pt[dup_idx], free_pt)
            ci[dup_idx, ] <- ci[leader_idx, , drop = FALSE]
          }
        }
      }

      tmp_list$ci.lower <- ci[, 1]
      tmp_list$ci.upper <- ci[, 2]
    }

    # standardized estimates?
    # 28 March 2024: TDJ adds option to select specific types
    # observed variables to standardize when standardized= is a vector of
    # (observed) variable names (a generalization of std.nox); see below
    ov_std_user <- NULL
    if (is.logical(standardized)) {
      if (standardized) {
        standardized <- c("std.lv", "std.all")
        if (length(lav_object_vnames(object, "ov.x")) &&
                                 object@Options$fixed.x) {
          standardized <- c(standardized, "std.nox")
        }
      } else {
        standardized <- character(0)
      } # corresponds to standardized=FALSE
    } else {
      # !is.logical(standardized)
      standardized <- as.character(standardized)
      std_types <- c("std.lv", "std.all", "std.nox")
      if (length(standardized) > 0L &&
          !any(tolower(standardized) %in% std_types)) {
        # interpret 'standardized' as a vector of observed variable names:
        # standardize ONLY the parameters involving these variables (a
        # generalization of "std.nox", where the exogenous 'x' are the ones
        # left unstandardized)
        ov_all <- lav_object_vnames(object, "ov")
        bad <- standardized[!(standardized %in% ov_all)]
        if (length(bad) > 0L) {
          lav_msg_warn(gettextf(
            "standardized= argument contains unknown observed variable(s): %s",
            paste(bad, collapse = ", ")))
        }
        ov_std_user <- standardized[standardized %in% ov_all]
        if (length(ov_std_user) == 0L) {
          standardized <- character(0)
        } else {
          standardized <- c("std.lv", "std.user")
        }
      } else {
        standardized <- tolower(standardized)
        if ("std.nox" %in% standardized) {
          # sanity checks
          if (length(lav_object_vnames(object, "ov.x")) == 0) {
            lav_msg_warn(gettext(
              "`std.nox' values are identical to `std.all' values in the
               absence of exogenous covariates."))
          } else if (!object@Options$fixed.x) {
            lav_msg_warn(gettext(
              "`std.nox' values are identical to `std.all' values when
               fixed.x = FALSE."))
          }
        }
      }
    }
    # Then add each requested type
    # (original source code, but now independently conditional)
    if ("std.lv" %in% standardized) {
      tmp_list$std.lv <- lav_standardize_lv(object, cov_std = cov_std)
    }
    if ("std.all" %in% standardized) {
      tmp_list$std.all <- lav_standardize_all(object,
        est_std = tmp_list$est.std,
        cov_std = cov_std
      )
    }
    if ("std.nox" %in% standardized) {
      tmp_list$std.nox <- lav_standardize_all_nox(object,
        est_std = tmp_list$est.std,
        cov_std = cov_std
      )
    }
    if ("std.user" %in% standardized) {
      tmp_list$std.user <- lav_standardize_all(object,
        est_std = tmp_list$est.std,
        cov_std = cov_std, ov_std = ov_std_user
      )
    }

    # warn if a defined (:=) parameter standardizes a label that is constrained
    # equal across groups/equations (its standardized value is then ambiguous);
    # base the numeric comparison on whichever standardized column is available
    if (length(standardized) > 0L) {
      tmp_std <- tmp_list$std.all
      if (is.null(tmp_std)) tmp_std <- tmp_list$std.nox
      if (is.null(tmp_std)) tmp_std <- tmp_list$std.user
      if (is.null(tmp_std)) tmp_std <- tmp_list$std.lv
      lav_partable_check_std_def(tmp_list, est.std = tmp_std)
    }

    # rsquare?
    if (rsquare) {
      r2 <- lavTech(object, "rsquare", add.labels = TRUE)
      tmp_names <- unlist(lapply(r2, names))
      nel <- length(tmp_names)
      if (nel == 0L) {
        lav_msg_warn(
          gettext("rsquare = TRUE, but there are no dependent variables"))
      } else {
        if (lav_pt_nlevels(tmp_list) == 1L) {
          block <- rep(seq_along(r2), sapply(r2, length))
          first_block_idx <- which(!duplicated(tmp_list$block) &
            tmp_list$block > 0L)
          gval <- tmp_list$group[first_block_idx]
          if (length(gval) > 0L) {
            group <- rep(gval, sapply(r2, length))
          } else {
            # single block, single group
            group <- rep(1L, length(block))
          }
          r2 <- data.frame(
            lhs = tmp_names, op = rep("r2", nel),
            rhs = tmp_names, block = block, group = group,
            est = unlist(r2), stringsAsFactors = FALSE
          )
        } else {
          # add level column
          block <- rep(seq_along(r2), sapply(r2, length))
          first_block_idx <- which(!duplicated(tmp_list$block) &
            tmp_list$block > 0L)
          # always at least two blocks
          gval <- tmp_list$group[first_block_idx]
          group <- rep(gval, sapply(r2, length))
          lval <- tmp_list$level[first_block_idx]
          level <- rep(lval, sapply(r2, length))
          r2 <- data.frame(
            lhs = tmp_names, op = rep("r2", nel),
            rhs = tmp_names,
            block = block, group = group,
            level = level,
            est = unlist(r2), stringsAsFactors = FALSE
          )
        }
        # add step column if needed
        if (!is.null(tmp_list$step)) {
          r2$step <- 2L # per default
          # simplification: we assume that only the
          # observed indicators of latent variables are step 1
          ov_ind <- unlist(object@pta$vnames$ov.ind)
          step1_idx <- which(r2$lhs %in% ov_ind)
          r2$step[step1_idx] <- 1L
        }
        tmp_list <- lav_pt_merge(pt1 = tmp_list, pt2 = r2, warn = FALSE)
      }
    }

    # fractional missing information (if estimator="fiml")
    if (fmi) {
      se_orig <- tmp_list$se

      # new in 0.6-6, use 'EM' based (unstructured) sample statistics
      # otherwise, it would be as if we use expected info, while the
      # original use observed, producing crazy results
      if (object@Data@ngroups > 1L) {
        em_cov <- lapply(lavInspect(object, "sampstat.h1"), "[[", "cov")
        em_mean <- lapply(lavInspect(object, "sampstat.h1"), "[[", "mean")
      } else {
        em_cov <- lavInspect(object, "sampstat.h1")$cov
        em_mean <- lavInspect(object, "sampstat.h1")$mean
      }

      tmp_pt <- parTable(object)
      tmp_pt$ustart <- tmp_pt$est
      tmp_pt$start <- tmp_pt$est <- NULL

      this_options <- object@Options
      if (!is.null(fmi_options) && is.list(fmi_options)) {
        # modify original options
        this_options <- modifyList(this_options, fmi_options)
      }
      # override
      this_options$optim.method <- "none"
      this_options$sample.cov.rescale <- FALSE
      this_options$check.gradient <- FALSE
      this_options$baseline <- FALSE
      this_options$h1 <- FALSE
      this_options$test <- FALSE

      fit_complete <- lavaan(
        model = tmp_pt,
        sample_cov = em_cov,
        sample_mean = em_mean,
        sample_nobs = lavInspect(object, "nobs"),
        slot_options = this_options
      )

      se_comp <- lavParameterEstimates(fit_complete,
        ci = FALSE, fmi = FALSE,
        zstat = FALSE, pvalue = FALSE, remove_system_eq = FALSE,
        remove_eq = FALSE, remove_ineq = FALSE,
        remove_def = FALSE, remove_nonfree = FALSE, remove_unused = FALSE,
        rsquare = rsquare, add_attributes = FALSE
      )$se

      se_comp <- ifelse(se_comp == 0.0, as.numeric(NA), se_comp)
      tmp_list$fmi <- 1 - (se_comp * se_comp) / (se_orig * se_orig)
    }

    # if single level, remove level column
    if (object@Data@nlevels == 1L) tmp_list$level <- NULL

    # if single group, remove group column
    if (object@Data@ngroups == 1L) tmp_list$group <- NULL

    # if single everything, remove block column
    if (object@Data@nlevels == 1L &&
      object@Data@ngroups == 1L) {
      tmp_list$block <- NULL
    }

    # if no user-defined labels, remove label column
    if (sum(nchar(object@ParTable$label)) == 0L) {
      tmp_list$label <- NULL
    }

    # if plabel = TRUE, add it (new in 0.6-20)
    if (plabel) {
      if (!is.null(tmp_partable$plabel)) {
        tmp_list$plabel <- tmp_partable$plabel
      } else {
        if (!is.null(tmp_partable$id)) {
          tmp_list$plabel <- paste(".p", tmp_list$id, ".", sep = "")
        } else {
          tmp_list$plabel <- paste(".p", seq_along(tmp_list$plabel),
                                   ".", sep = "")
        }
      }
    }

    # remove non-free parameters? (but keep ==, >, < and :=)
    if (remove_nonfree) {
      nonfree_idx <- which(tmp_list$free == 0L &
        !tmp_list$op %in% c("==", ">", "<", ":="))
      if (length(nonfree_idx) > 0L) {
        tmp_list <- tmp_list[-nonfree_idx, ]
      }
    }

    # remove 'unused' parameters
    # these are parameters that are automatically added (user == 0),
    # but with their final (est) values fixed to their default values
    # (typically 1 or 0).
    # currently only intercepts and scaling-factors (for now)
    # should we also remove fixed-to-1 variances? (parameterization = theta)?
    if (remove_unused) {
      # intercepts
      int_idx <- which(tmp_list$op == "~1" &
        tmp_list$user == 0L &
        tmp_list$free == 0L &
        tmp_list$est == 0)
      if (length(int_idx) > 0L) {
        tmp_list <- tmp_list[-int_idx, ]
      }

      # scaling factors
      scaling_idx <- which(tmp_list$op == "~*~" &
        tmp_list$user == 0L &
        tmp_list$free == 0L &
        tmp_list$est == 1)
      if (length(scaling_idx) > 0L) {
        tmp_list <- tmp_list[-scaling_idx, ]
      }
    }


    # remove 'free' column
    tmp_list$free <- NULL

    # remove == rows?
    if (remove_eq) {
      eq_idx <- which(tmp_list$op == "==" & tmp_list$user == 1L)
      if (length(eq_idx) > 0L) {
        tmp_list <- tmp_list[-eq_idx, ]
      }
    }
    if (remove_system_eq) {
      eq_idx <- which(tmp_list$op == "==" & tmp_list$user != 1L)
      if (length(eq_idx) > 0L) {
        tmp_list <- tmp_list[-eq_idx, ]
      }
    }
    # remove <> rows?
    if (remove_ineq) {
      ineq_idx <- which(tmp_list$op %in% c("<", ">"))
      if (length(ineq_idx) > 0L) {
        tmp_list <- tmp_list[-ineq_idx, ]
      }
    }
    # remove := rows?
    if (remove_def) {
      def_idx <- which(tmp_list$op == ":=")
      if (length(def_idx) > 0L) {
        tmp_list <- tmp_list[-def_idx, ]
      }
    }

    # remove step 1 rows?
    if (remove_step1 && !is.null(tmp_list$step)) {
      step1_idx <- which(tmp_list$step == 1L)
      if (length(step1_idx) > 0L) {
        tmp_list <- tmp_list[-step1_idx, ]
      }
      # remove step column
      tmp_list$step <- NULL
    }

    # remove auxiliary (saturated-correlates) rows? (FIML aux= variables)
    if (remove_aux && length(object@Options$aux) > 0L) {
      aux_idx <- which(tmp_list$lhs %in% object@Options$aux |
        tmp_list$rhs %in% object@Options$aux)
      if (length(aux_idx) > 0L) {
        tmp_list <- tmp_list[-aux_idx, ]
      }
    }

    # remove random-slope phantom marker rows (eg s1 =~ s1, fixed to zero)
    if (!is.null(object@ParTable$rv) &&
        any(nchar(object@ParTable$rv) > 0L)) {
      rv_names <- unique(object@ParTable$rv[
        nchar(object@ParTable$rv) > 0L])
      rv_idx <- which(tmp_list$op == "=~" &
        tmp_list$lhs == tmp_list$rhs &
        tmp_list$lhs %in% rv_names)
      if (length(rv_idx) > 0L) {
        tmp_list <- tmp_list[-rv_idx, ]
      }
    }

    # remove attribute for data order
    attr(tmp_list, "ovda") <- NULL

    # remove tmp_list$user
    tmp_list$user <- NULL

    if (output == "text") {
      class(tmp_list) <- c(
        "lavaan.parameterEstimates", "lavaan.data.frame",
        "data.frame"
      )
      if (header) {
        attr(tmp_list, "categorical") <- object@Model@categorical
        attr(tmp_list, "parameterization") <- object@Model@parameterization
        attr(tmp_list, "information") <- object@Options$information[1]
        attr(tmp_list, "information.meat") <- object@Options$information.meat
        attr(tmp_list, "se") <- object@Options$se
        attr(tmp_list, "group.label") <- object@Data@group.label
        attr(tmp_list, "level.label") <- object@Data@level.label
        attr(tmp_list, "bootstrap") <- object@Options$bootstrap
        attr(tmp_list, "bootstrap.successful") <- bootstrap_successful
        attr(tmp_list, "missing") <- object@Options$missing
        attr(tmp_list, "observed.information") <-
          object@Options$observed.information[1]
        attr(tmp_list, "h1.information") <- object@Options$h1.information[1]
        attr(tmp_list, "h1.information.meat") <-
          object@Options$h1.information.meat
        attr(tmp_list, "header") <- header
        # estimator = "IV": the standard errors are not based on an information
        # matrix, but on a two-stage (sandwich) procedure; report the stage-1
        # and stage-2 vcov methods and the type of moment ACOV (Gamma) instead
        if (identical(object@Options$estimator, "IV")) {
          attr(tmp_list, "estimator") <- "IV"
          attr(tmp_list, "iv.vcov.stage1") <-
            object@Options$estimator.args$iv_vcov_stage1
          attr(tmp_list, "iv.vcov.stage2") <-
            object@Options$estimator.args$iv_vcov_stage2
          # type of Gamma (moment ACOV) used in the stage-2 standard errors;
          # not applicable when no stage-2 covariance is computed
          if (!identical(tolower(object@Options$estimator.args$iv_vcov_stage2),
                         "none")) {
            attr(tmp_list, "iv.gamma") <- if (object@Model@categorical) {
              "ADF"
            } else if (object@Options$missing %in%
                       c("two.stage", "robust.two.stage")) {
              "TS"
            } else {
              "NT"
            }
          }
        }
        # FIXME: add more!!
      }
    } else {
      tmp_list$exo <- NULL
      tmp_list$lower <- tmp_list$upper <- NULL
      class(tmp_list) <- c("lavaan.data.frame", "data.frame")
    }

    # store the BCa empirical-influence design matrix (issue #598), so users
    # can retrieve it via attr(., "design") without recomputing it
    if (!is.null(boot_bca_design)) {
      attr(tmp_list, "design") <- boot_bca_design
    }

    tmp_list
}
parameterEstimates <- lav_alias("lavParameterEstimates") # synonym # nolint

parTable <- function(object) {
  # check object
  object <- lav_object_check_version(object)

  # convert to data.frame
  out <- as.data.frame(object@ParTable, stringsAsFactors = FALSE)

  class(out) <- c("lavaan.data.frame", "data.frame")
  out
}
parameterTable <- lav_alias("parTable") # synonym # nolint
parametertable <- lav_alias("parTable") # synonym # nolint
partable <- lav_alias("parTable") # synonym # nolint

varTable <- function(object, ov_names = names(object),  # nolint start
                     ov_names_x = NULL,
                     ordered = NULL, factor = NULL,
                     as_data_frame = TRUE,
                     ...) {           # nolint end
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  if (inherits(object, "lavaan")) {
    # check object
    object <- lav_object_check_version(object)
    tmp_var <- object@Data@ov
  } else if (inherits(object, "lavData")) {
    tmp_var <- object@ov
  } else if (inherits(object, "data.frame")) {
    tmp_var <- lav_dataframe_vartable(
      frame = object, ov_names = ov_names,
      ov_names_x = ov_names_x,
      ordered = ordered, factor = factor,
      as_data_frame = FALSE
    )
  } else {
    lav_msg_stop(gettext("object must of class lavaan or a data.frame"))
  }

  if (as_data_frame) {
    tmp_var <- as.data.frame(tmp_var,
      stringsAsFactors = FALSE,
      row.names = seq_along(tmp_var$name)
    )
    class(tmp_var) <- c("lavaan.data.frame", "data.frame")
  }

  tmp_var
}


setMethod(
  "fitted.values", "lavaan",
  function(object, type = "moments", labels = TRUE, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("fitted.values"))
        )
      }
    }
    # lowercase type
    type <- tolower(type)

    # check object
    object <- lav_object_check_version(object)

    # catch type="casewise"
    if (type %in% c("casewise", "case", "obs", "observations", "ov")) {
      return(lavPredict(object, type = "ov", label = labels))
    }

    lav_inspect_implied(object,
      add_labels = labels, add_class = TRUE,
      drop_list_single_group = TRUE
    )
  }
)


setMethod(
  "fitted", "lavaan",
  function(object, type = "moments", labels = TRUE, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("fitted"))
        )
      }
    }
    fitted.values(object, type = type, labels = labels)
  }
)

setMethod(
  "vcov", "lavaan",
  function(object, type = "free", labels = TRUE, remove.duplicated = FALSE, # nolint start
           standardized = NULL, free.only = TRUE, ...) {                    # nolint end
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("vcov"))
        )
      }
    }
    # check object
    object <- lav_object_check_version(object)

    # check for convergence first!
    if (object@optim$npar > 0L && !object@optim$converged) {
      lav_msg_stop(gettext("model did not converge"))
    }

    if (object@Options$se == "none") {
      lav_msg_stop(gettext("vcov not available if se=\"none\""))
    }

    ## verify there are any user-defined parameters
    #FIXME? smarter to check @ParTable for $op == ":="?
    if (is.null(formals(object@Model@def.function))) {
      type <- "free" # avoids error in lav_inspect_vcov_def()
    }

    if (!is.null(standardized)) {
      standardized <- tolower(standardized[1])
      stopifnot(standardized %in% c("std.lv", "std.all", "std.nox"))
    }

    if (type == "user" || type == "joint" || type == "all" || type == "full" ||
      type == "complete") {
      if (remove.duplicated) {
        lav_msg_stop(gettext(
          "argument \"remove.duplicated\" not supported if type = \"user\""
        ))
      }
      tmp_varcov <- lav_inspect_vcov_def(object,
        joint = TRUE,
        standardized = !is.null(standardized),
        type = ifelse(is.null(standardized), "std.all", standardized),
        add_labels = labels,
        add_class = TRUE
      )
    } else if (type == "free") {
      tmp_varcov <- lav_inspect_vcov(object,
        standardized = !is.null(standardized),
        type = ifelse(is.null(standardized), "std.all", standardized),
        free_only = free.only,
        add_labels = labels,
        add_class = TRUE,
        remove_duplicated = remove.duplicated
      )
    } else {
      lav_msg_stop(gettext("type argument should be \"user\" or \"free\""))
    }

    tmp_varcov
  }
)


# logLik (so that we can use the default AIC/BIC functions from stats4(
setMethod(
  "logLik", "lavaan",
  function(object, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("logLik"))
        )
      }
    }
    # check object
    object <- lav_object_check_version(object)

    if (object@Options$estimator != "ML") {
      lav_msg_warn(gettext("logLik only available if estimator is ML"))
    }
    if (object@optim$npar > 0L && !object@optim$converged) {
      lav_msg_warn(gettext("model did not converge"))
    }

    # new in 0.6-1: we use the @loglik slot (instead of fitMeasures)
    tmp_logl <- object@loglik
    if (is.null(tmp_logl$loglik)) {
      # empty @loglik slot (eg loglik = FALSE, or estimator != ML):
      # return NA instead of crashing (AIC/BIC then also return NA)
      logl <- as.numeric(NA)
      attr(logl, "df") <- object@optim$npar
      attr(logl, "nobs") <- object@SampleStats@ntotal
    } else {
      logl <- tmp_logl$loglik
      attr(logl, "df") <- tmp_logl$npar ### note: must be npar, not df!!
      attr(logl, "nobs") <- tmp_logl$ntotal
    }
    class(logl) <- "logLik"
    logl
  }
)

# nobs
if (!exists("nobs", envir = asNamespace("stats4"))) {
  setGeneric("nobs", function(object, ...) standardGeneric("nobs"))
}
setMethod(
  "nobs", signature(object = "lavaan"),
  function(object, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("nobs"))
        )
      }
    }
    object@SampleStats@ntotal
  }
)

# see: src/library/stats/R/update.R
setMethod(
  "update", signature(object = "lavaan"),
  function(object, model, add, ..., evaluate = TRUE) {
    # check object
    object <- lav_object_check_version(object)

    call <- object@call
    if (is.null(call)) {
      lav_msg_stop(gettext("need an object with call slot"))
    }
    # sam objects store the original sam() call in the call slot (>= 0.7-1);
    # in older versions, the call slot contained an internal lavaan() call
    # that cannot be re-evaluated
    if (!is.null(object@internal$sam.method) &&
        !any(as.character(call[[1]]) == "sam")) {
      lav_msg_stop(gettext(
        "update() does not work for a sam object created by an older
         version of lavaan; please rerun sam() first."))
    }
    current_call_names <- names(call)
    lavaan_formals <- names(formals(lavaan::lavaan))
    current_formals <- lav_snake_case(current_call_names) %in% lavaan_formals
    current_call_names[current_formals] <-
              lav_snake_case(current_call_names[current_formals])
    names(call) <- current_call_names

    extras <- match.call(expand.dots = FALSE)$...

    if (!missing(model)) {
      # call$formula <- update.formula(formula(object), formula.)
      call$model <- model
    } else if (exists(as.character(call$model))) {
      call$model <- eval(call$model, parent.frame())
    } else if (is.character(call$model)) {
      ## do nothing
      ## call$model <- call$model
    } else {
      call$model <- parTable(object)
      call$model$est <- NULL
      call$model$se <- NULL
    }
    if (!is.null(call$slot_par_table) && is.list(call$model)) {
      call$slot_par_table <- call$model
    }

    if (length(extras) > 0) {
      ## check for call$slot_options conflicts
      if (!is.null(call$slot_options)) {
        same_names <- intersect(names(lavOptions()), names(extras))
        for (i in same_names) {
          call$slot_options[[i]] <- extras[[i]]
          extras[i] <- NULL # not needed if they are in slotOptions
        }
      }
      existing <- !is.na(match(names(extras), names(call)))
      for (a in names(extras)[existing]) call[[a]] <- extras[[a]]
      if (any(!existing)) {
        call <- c(as.list(call), extras[!existing])
        call <- as.call(call)
      }
    }

    if (missing(add) && !evaluate) {
      return(call)
    }
    ## for any of the other 3 scenarios, we need the updated fit

    ## Check if "add" and "model" are both strings; combine them
    if (missing(add)) {
      add_allready_in_partable <- TRUE # because nothing to add
    } else {
      if (is.character(add) && is.character(call$model)) {
        call$model <- c(call$model, add)
        add_allready_in_partable <- TRUE
      } else {
        add_allready_in_partable <- FALSE
      }
    }
    newfit <- eval(call, parent.frame())
    if (add_allready_in_partable && evaluate) {
      return(newfit)
    }

    ## only remaining situations: "add" exists, but either "add" or "model"
    ## is a parameter table, so update the parameter table in the call
    if (!(mode(add) %in% c("list", "character"))) {
      lav_msg_stop(
        gettext("'add' argument must be model syntax or parameter table.
                See ?lav_model_pt help page.")
      )
    }
    tmp_pt <- lav_object_extended(newfit, add = add)@ParTable
    tmp_pt$user <- NULL # get rid of "10" category used in lavTestScore()
    ## group == 0L in new rows
    tmp_pt$group[tmp_pt$group == 0L] <- tmp_pt$block[tmp_pt$group == 0L]
    # tmp_pt$plabel == "" in new rows.  Consequences?
    tmp_pt$est <- NULL
    tmp_pt$se <- NULL
    call$model <- tmp_pt

    if (evaluate) {
      eval(call, parent.frame())
    } else {
      call
    }
  }
)

setMethod(
  "anova", signature(object = "lavaan"),
  function(object, ...) {
    # NOTE: if we add additional arguments, it is not the same generic
    # anova() function anymore, and match.call will be screwed up

    # NOTE: we need to extract the names of the models from match.call here,
    #       otherwise, we lose them in the call stack

    mcall <- match.call(expand.dots = TRUE)
    dots <- list(...)

    modp <- if (length(dots)) {
      sapply(dots, inherits, "lavaan")
    } else {
      logical(0)
    }
    tmp_names <- sapply(as.list(mcall)[c(FALSE, TRUE, modp)], deparse)

    lavTestLRT(object = object, ..., model_names = tmp_names)
  }
)

# synonyms (thin forwarders; see lav_alias in 00alias.R)
standardizedsolution <- lav_alias("standardizedSolution") # nolint
vartable <- lav_alias("varTable") # nolint

# utility functions
#
# initial version: YR 25/03/2009

# multivariate normal random number generation
# replacement for MASS::mvrnorm for better cross-machine reproducibility
# see: https://blog.djnavarro.net/posts/2025-05-18_multivariate-normal-sampling-floating-point/  # nolint
#
# the issue with MASS::mvrnorm is that it uses eigendecomposition with a
# transformation matrix (sqrt(Lambda) Q') that is NOT invariant to eigenvector
# sign flips. this leads to irreproducible results across machines even with
# the same random seed. this implementation follows mvtnorm::rmvnorm which uses:
# - eigen: Q sqrt(Lambda) Q' which IS invariant to sign flips (default)
# - svd: similar to eigen, also invariant
# - chol: unique factorization
# eigen is the default to match MASS::mvrnorm's approach while fixing the
# sign-flip issue with the invariant formula
lav_mvrnorm <- function(n = 1, mu, sigma_1, tol = 1e-06, empirical = FALSE,
                      method = "eigen", check_symmetry = TRUE, byrow = FALSE) {
  p <- length(mu)

  # check symmetry (like mvtnorm)
  if (check_symmetry && !isSymmetric(sigma_1,
    tol = sqrt(.Machine$double.eps),
    check.attributes = FALSE
  )) {
    lav_msg_stop(gettext("'Sigma' must be a symmetric matrix in lav_mvrnorm"))
  }

  # check dimensions (like mvtnorm)
  if (p != nrow(sigma_1)) {
    lav_msg_stop(gettext("incompatible arguments in lav_mvrnorm"))
  }

  # compute transformation matrix R based on method (following mvtnorm exactly)
  method <- match.arg(method, c("chol", "eigen", "svd"))

  r <- if (method == "eigen") {
    ev <- eigen(sigma_1, symmetric = TRUE)
    if (!all(ev$values >= -tol * abs(ev$values[1L]))) {
      lav_msg_warn(gettext(
        "sigma is numerically not positive semidefinite."))
    }
    # Q sqrt(Lambda) Q' - invariant to sign flips
    t(ev$vectors %*% (t(ev$vectors) * sqrt(pmax(ev$values, 0))))
  } else if (method == "svd") {
    s <- svd(sigma_1)
    if (!all(s$d >= -tol * abs(s$d[1L]))) {
      lav_msg_warn(gettext(
        "sigma is numerically not positive semidefinite."))
    }
    t(s$v %*% (t(s$u) * sqrt(pmax(s$d, 0))))
  } else if (method == "chol") {
    r <- chol(sigma_1, pivot = TRUE)
    r[, order(attr(r, "pivot"))]
  }

  # for names (fallback to dimnames of Sigma if mu has no names)
  nm <- names(mu)
  if (is.null(nm) && !is.null(dn <- dimnames(sigma_1))) {
    nm <- dn[[1L]]
  }

  if (empirical) {
    # generate standard normal, then apply empirical transformation
    x <- matrix(stats::rnorm(p * n), n, p, byrow = byrow)
    x <- scale(x, center = TRUE, scale = FALSE) # center
    x <- x %*% svd(x, nu = 0)$v # orthogonalize
    x <- scale(x, center = FALSE, scale = TRUE) # unit variance

    # transform by R
    x <- sweep(x %*% r, 2, mu, "+")
    colnames(x) <- nm

    if (n == 1) drop(x) else x
  } else {
    # generate samples (following mvtnorm exactly)
    x <- matrix(stats::rnorm(n * p), nrow = n, byrow = byrow) %*% r
    x <- sweep(x, 2, mu, "+")
    colnames(x) <- nm

    if (n == 1) drop(x) else x
  }
}

# sd with trimming
lav_sample_trimmed_sd <- function(x, na_rm = TRUE, trim = 0) {
  if (isTRUE(na_rm)) {
    x <- x[!is.na(x)]
  }
  n <- length(x)
  if (trim > 0 && n) {
    if (is.complex(x)) {
      lav_msg_stop(gettext("trimmed means are not defined for complex data"))
    }
    if (anyNA(x)) {
      return(NA_real_)
    }
    if (trim >= 0.5) {
      return(stats::median(x, na.rm = FALSE))
    }
    lo <- floor(n * trim) + 1
    hi <- n + 1 - lo
    x <- sort.int(x, partial = unique(c(lo, hi)))[lo:hi]
  }
  sd(x)
}

# convert correlation matrix + standard deviations to covariance matrix
# based on cov2cor in package:stats
lav_cor2cov <- function(r, sds, names = NULL, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  p <- (d <- dim(r))[1L]
  if (!is.numeric(r) || length(d) != 2L || p != d[2L]) {
    lav_msg_stop(gettext("'V' is not a square numeric matrix"))
  }

  if (any(!is.finite(sds))) {
    lav_msg_warn(gettext(
      "sds had 0 or NA entries; non-finite result is doubtful"
    ))
  }

  # if(sum(diag(r)) != p)
  #    stop("The diagonal of a correlation matrix should be all ones.")

  if (p != length(sds)) {
    lav_msg_stop(gettext("The standard deviation vector and correlation matrix
                         have a different number of variables"))
  }

  s <- r
  s[] <- sds * r * rep(sds, each = p)

  # optionally, add names
  if (!is.null(names)) {
    stopifnot(length(names) == p)
    rownames(s) <- colnames(s) <- names
  }

  s
}

# convert characters within single quotes to numeric vector
# eg. s <- '3 4.3 8e-3 2.0'
#     x <- lav_char2num(s)
lav_char2num <- function(s = "") {
  # first, strip all ',' or ';'
  s_1 <- chartr(",;", "  ", s)
  scan(text = s_1, quiet = TRUE)
}

# create full matrix based on lower.tri or upper.tri elements; add names
# always ROW-WISE!!
lav_getcov <- function(x, lower = TRUE, diagonal = TRUE, sds = NULL,
                       names = paste("V", 1:nvar, sep = "")) {
  # check x and sds
  if (is.character(x)) x <- lav_char2num(x)
  if (is.character(sds)) sds <- lav_char2num(sds)

  if (lower) {
    cov_1 <- lav_mat_lower2full(x, diagonal = diagonal)
  } else {
    cov_1 <- lav_mat_upper2full(x, diagonal = diagonal)
  }
  nvar <- ncol(cov_1)

  # if diagonal is false, assume unit diagonal
  if (!diagonal) diag(cov_1) <- 1

  # check if we have a sds argument
  if (!is.null(sds)) {
    stopifnot(length(sds) == nvar)
    cov_1 <- lav_cor2cov(cov_1, sds)
  }

  # names
  stopifnot(length(names) == nvar)
  rownames(cov_1) <- colnames(cov_1) <- names

  cov_1
}

lav_char2hash <- function(s = "") {
  stopifnot(is.character(s))
  nums <- utf8ToInt(paste(s, collapse = "\n"))
  rval <- 0x7EDCBA98L
  for (i in nums) {
    positions <- 1L + (bitwAnd(i, 15L))
    bits_r <- bitwShiftL(1L, positions) - 1L
    wrap1 <- bitwShiftL(bitwAnd(bits_r, rval), 31L - positions)
    wrap2 <- bitwShiftR(bitwAnd(bitwNot(bits_r), rval), positions)
    saw <- bitwOr(wrap1, wrap2)
    rval <- bitwXor(saw, i)
  }
  as.hexmode(rval)
}

# vectorize all (h0 or h1) sample statistics, in the same order
# as Gamma
lav_implied_to_vec <- function(implied = NULL, lavmodel = NULL,
                               drop_list = TRUE) {
  ngroups <- lavmodel@ngroups

  wls_obs <- vector("list", ngroups)
  for (g in seq_len(ngroups)) {

    var <- NULL
    res_var <- NULL
    if (lavmodel@conditional.x) {
      res_var <- diag(implied$res.cov[[g]])
      res_var <- res_var[res_var != 1]
    } else {
      var <- diag(implied$cov[[g]])
      var <- var[var != 1]
    }

    wls_obs[[g]] <- lav_samp_wls_obs(
      # plain
      mean_g = implied$mean[[g]],
      cov_g = implied$cov[[g]],
      var_g = var,
      th_g = implied$th[[g]],
      th_idx_g = lavmodel@th.idx[[g]],

      # conditional.x
      res_int_g = implied$res.int[[g]],
      res_cov_g = implied$res.cov[[g]],
      res_var_g = res_var,
      res_th_g = implied$res.th[[g]],
      res_slopes_g = implied$res.slopes[[g]],
      group_w_g = implied$group.w[[g]],

      # flags
      categorical = lavmodel@categorical,
      conditional_x = lavmodel@conditional.x,
      meanstructure = lavmodel@meanstructure,
      correlation = lavmodel@correlation,
      slopestructure = lavmodel@conditional.x,
      group_w_free = lavmodel@group.w.free
    )
  }

  if (drop_list) {
    out <- unlist(wls_obs)
  } else {
    out <- wls_obs
  }

  out
}

# given a vector of sample statistics, reconstruct implied
# entries (cov, mean, th, res.cov, ...)
lav_vec_to_implied <- function(x = NULL, lavmodel) {

  ngroups <- lavmodel@ngroups
  implied <- list()

  for (g in seq_len(ngroups)) {

    # number of variables
    nvar <- lavmodel@nvar[g]

    # if group.w.free, always comes first
    if (lavmodel@group.w.free) {
      idx <- 1L
      group_w <- x[idx]
      x <- x[-idx]
    } else {
      group_w <- 1
    }

    if (lavmodel@categorical) {
      if (lavmodel@conditional.x) {
        # TODO
        cat("\n NOT READY YET! \n")
      }

      # th
      nth <- length(lavmodel@th.idx[[g]])
      idx <- seq_len(nth)
      th_g <- x[idx]
      x <- x[-idx]

      # mean/var ov.num
      mean_g <- rep(0, nvar)
      var_g <- rep(1, nvar)
      if (any(lavmodel@th.idx[[g]] == 0)) {
        num_idx <- seq_len(nvar)
        num_idx <- num_idx[!num_idx %in% lavmodel@th.idx[[g]]]
        idx <- seq_along(num_idx)
        var_g[num_idx] <- x[idx]
        x <- x[-idx]
        # FIXME: change sign?
        mean_g[num_idx] <- th_g[lavmodel@th.idx[[g]] == 0]
      }

      # cov - lower only
      idx <- seq_len(nvar * (nvar - 1) / 2)
      cov_g <- lav_mat_vech_rev(x[idx], diagonal = FALSE)
      x <- x[-idx]
      diag(cov_g) <- var_g

      # fill in
      implied$cov[[g]] <- cov_g
      implied$mean[[g]] <- mean_g
      implied$th[[g]] <- th_g

    } else {
      # diag flag
      diag_flag <- TRUE
      if (lavmodel@correlation) {
        diag_flag <- FALSE
      }
      if (lavmodel@conditional.x) {
        # TODO
        cat("\n NOT READY YET! \n")
      }
      if (lavmodel@meanstructure) {
        idx <- seq_len(nvar)
        mean_g <- x[idx]
        x <- x[-idx]
      } else {
        mean_g <- numeric(nvar)
      }
      if (diag_flag) {
        idx <- seq_len(nvar * (nvar + 1) / 2)
      } else {
        idx <- seq_len(nvar * (nvar - 1) / 2)
      }
      cov_g <- lav_mat_vech_rev(x[idx], diagonal = diag_flag)
      x <- x[-idx]

      # fill in
      implied$cov[[g]] <- cov_g
      implied$mean[[g]] <- mean_g
      implied$th[g] <- list(NULL)
    }

    implied$group.w[[g]] <- group_w
  }

  implied
}


# find the solution of the linear equation:
#
# theta = solve(t(Delta) %*% W %*% Delta) %*% t(Delta) %*% W %*% svec
#
# where Delta is the Jacobian of Sigma wrt the free parameters,
# W is lav_mat_bdiag(S.inv, S22_inv)
# where S22_inv = 0.5 * lav_mat_dup_pre_post(S.inv %x% S.inv)
#  - S can be I (ULS) (then W = 0.5 * t(D) %*% D)
#  - S can be sample.cov (GLS)
#  - S can be Sigma (2RLS or RLS)
#
# svec is the vector with sample statistics (first means, then vech(S))
#
# strategy:
# - first we compute Q = W %*% Delta, using the fact that W only contains S.inv
#   elements (without explicitly constructing the kronecker product)
#   note that for the cov block:
#     W %*% Delta = w * vech(S.inv * unvech(v) * S.inv)
#   where w is 1/2 for diagonal elements, and 1 elsewhere
# - compute A = t(Delta) %*% Q, and b = Q %*% svec
# - compute solve(A,b) using cholesky back/forward, as A is symmetric and pd
#
# optionally, return H = solve(t(Delta) %*% W %*% Delta) %*% t(Delta) %*% W
# as an attribute
lav_utils_wls_linearization <- function(delta = NULL, s = NULL,
                                        meanstructure = FALSE,
                                        categorical = FALSE,
                                        #fixed.x = FALSE, # FIXME: needed?
                                        #x.idx = integer(0L),
                                        svec = NULL,
                                        return_h = FALSE,
                                        con_jac = NULL,
                                        con_rhs = NULL) {
  nvar <- nrow(s)
  if (categorical) {
    meanstructure <- FALSE
    # all ordinal (for now)
    pstar <- nvar * (nvar - 1L) / 2L
  } else {
    pstar <- nvar * (nvar + 1L) / 2L
  }

  jac_cov <- delta
  nc <- ncol(jac_cov)
  s_cov <- svec

  # vech
  w <- rep(1.0, pstar)
  if (!categorical) {
    # half-weight the diagonal elements of vech(): they are counted once,
    # while each off-diagonal element represents two (symmetric) entries.
    # (For categorical data, the cov part only contains off-diagonal
    # correlations, so no diagonal weighting is needed.)
    w[lav_mat_diagh_idx(nvar)] <- 0.5
  }

  # split Delta/svec
  if (categorical) {
    cov_idx  <- seq(nrow(delta) - pstar + 1L, nrow(delta))
    jac_cov  <- delta[cov_idx, , drop = FALSE]
    s_cov    <- svec[cov_idx]
  } else if (!categorical && meanstructure) {
    mean_idx <- seq_len(nvar)
    jac_mean <- delta[mean_idx, , drop = FALSE]
    jac_cov <- delta[-mean_idx, , drop = FALSE]
    s_mean <- svec[mean_idx]
    s_cov <- svec[-mean_idx]
  } else if (nrow(delta) != pstar) {
    lav_msg_stop(gettext("nrow(Delta) != pstar"))
  }

  # special case: S = I (ULS)
  uls_flag <- isTRUE(all.equal(s, diag(nvar), check.attributes = FALSE))
  if (uls_flag || categorical) {
    # cov part
    q_cov <- w * jac_cov
    a <- crossprod(jac_cov, q_cov)
    b <- drop(crossprod(q_cov, s_cov))

    # mean part
    if (!categorical && meanstructure) {
      a <- a + crossprod(jac_mean)
      b <- b + drop(crossprod(jac_mean, s_mean))
    }

    # tQ only needed if H is requested
    if (return_h) {
      t_q <- t(q_cov)
      if (meanstructure) t_q <- cbind(t(jac_mean), t_q)
    }
  } else {
    # S inverse
    c_s <- chol(s)
    si <- chol2inv(c_s)

    # cov part
    q_cov <- matrix(0, pstar, nc)
    for (j in seq_len(nc)) {
      vmat <- lav_mat_vech_rev(jac_cov[, j])
      m_w <- si %*% vmat %*% si
      q_cov[, j] <- w * lav_mat_vech(m_w)
    }

    a <- crossprod(jac_cov, q_cov)
    b <- drop(crossprod(q_cov, s_cov))

    # mean part
    if (meanstructure) {
      q_mean <- backsolve(c_s, forwardsolve(t(c_s), jac_mean))
      a <- a + crossprod(jac_mean, q_mean)
      b <- b + drop(crossprod(q_mean, s_mean))
    }

    if (return_h) {
      t_q <- t(q_cov)
      if (meanstructure) t_q <- cbind(t(q_mean), t_q)
    }
  }

  r <- chol(a)
  have_con <- !is.null(con_jac) && nrow(con_jac) > 0L
  if (!return_h && !have_con) {
    out <- backsolve(r, forwardsolve(t(r), b))
  } else {
    ainv  <- chol2inv(r)
    out <- drop(ainv %*% b)
    if (return_h) {
      h <- ainv %*% t_q
      attr(out, "H") <- h
    }
  }

  # impose linear equality constraints (con_jac %*% theta == con_rhs) on the
  # WLS estimate by projection onto the constraint set, using the WLS Hessian
  # 'a' as the metric: theta_c = theta - A^{-1} R' (R A^{-1} R')^{-1}(R theta-q)
  if (have_con) {
    air <- ainv %*% t(con_jac)
    rar <- con_jac %*% air
    lambda <- solve(rar, drop(con_jac %*% out) - con_rhs)
    h_attr <- attr(out, "H")
    out <- drop(out - air %*% lambda)
    if (!is.null(h_attr)) {
      attr(out, "H") <- h_attr
    }
  }

  out
}

# ----------------------- lav_snake_case -------------------------------- #
# function to transform names of variables to snake_case
# this function is used mainly to rename function arguments given in a list
#     where 'old' names are still accepted to avoid breaking other packages
# if input is a list, the names of the list are transformed.
lav_snake_case <- function(old) {
  curval <- c("B", "C", "D", "E", "K", "W", "PI",
             "TAU", "DELTA", "NU", "LAMBDA", "eXo",
              "WMAT", "THETA", "ALPHA", "BETA", "GAMMA", "PSI",
              "SminTheta")
  newval <- c("m_b", "m_c", "m_d", "m_e", "m_k", "m_w", "pi0",
             "mm_tau", "mm_delta", "mm_nu", "mm_lambda", "exo",
             "mm_wmat", "mm_theta", "mm_alpha", "mm_beta", "mm_gamma", "mm_psi",
             "s_min_theta")
  if (is.list(old)) old_names <- names(old) else old_names <- old
  # transform dot.case and CamelCase to snake_case
  varnames_new <- chartr(".A-Z", "_a-z",
                   gsub("([a-z])([A-Z])", "\\1_\\2", old_names))
  # apply standard modifications
  mtch <- match(old_names, curval)
  for (j in seq_along(old_names)) {
    if (!is.na(mtch[j])) varnames_new[j] <- newval[mtch[j]]
  }
  # remove trailing underscores in new names
  varnames_new <- gsub("_$", "", varnames_new)
  # check no doubles in new names
  tocheck <- varnames_new[varnames_new != ""]
  doubles <- anyDuplicated(tocheck)
  if (doubles) {
    lav_msg_stop(gettextf("At least one snake_cased name (%s) is duplicated!",
                   tocheck[doubles]))
  }
  if (is.list(old)) {
    names(old) <- varnames_new
    old
  } else {
    varnames_new
  }
}
# ----------------------- lav_option_names -------------------------------- #
# function to transform specified names of options to valid option names
# this function is used to accept uppercase and dots or underscores in
# option names where the real option name has lowercase, underscores or dots
# if input is a list, the names of the list are transformed.
lav_option_names <- function(old) {
  if (is.list(old)) old_names <- names(old) else old_names <- old
  # transform dot.case and CamelCase to snake_case
  varnames_new <- chartr("A-Z_", "a-z.", old_names)
  varnames_new[varnames_new == "optim.init.nelder.mead"] <- "optim.init_nelder_mead"
  if (is.list(old)) {
    names(old) <- varnames_new
    old
  } else {
    varnames_new
  }
}
# ------------------------- lav_adapt_func --------------------------------- #
# function to put arguments with old names (in ...) in the new named argument
# the function to adapt must have a ... argument and call this function in the
# beginning as follows :
#   dotdotdot <- list(...)
#   lav_adapt_func(environment(), dotdotdot, TRUE/FALSE/NULL)
# The argument options_ddd if set to TRUE transforms the names of dotdotdot to
# valid option names (if possible).
# If options_ddd is set to NULL an error "unused argument(s)" is generated when
# there are items left in dotdotdot. If dotdotdot is empty and options_ddd is
# NULL, dotdotdot is removed from the environment!
lav_adapt_func <- function(envir, dotdotdot, options_ddd) {
  lijst <- ls(pos = envir)
  if (length(dotdotdot) > 0L) {
    dddnames <- names(dotdotdot)
    dddnewnames <- lav_snake_case(dddnames)
    newargs <- dddnewnames %in% lijst
    for (j in which(newargs)) {
      assign(dddnewnames[j], dotdotdot[[dddnames[j]]], envir)
      dotdotdot[[dddnames[j]]] <- NULL
    }
    if (length(dotdotdot) > 0L) {
      if (is.null(options_ddd)) {
        lav_msg_stop(ngettext(length(dotdotdot),
                     "unused argument:", "unused arguments:"),
                     lav_msg_view(names(dotdotdot), "none", FALSE)
        )
      }
      if (options_ddd && length(dotdotdot) > 0L) {
        dotdotdot <- lav_option_names(dotdotdot)
      }
    }
    assign("dotdotdot", dotdotdot, envir)
  } else {
    if (is.null(options_ddd)) rm("dotdotdot", pos = envir)
  }
}

# for blavaan
# TDJ: add "..." to make the generic actually generic, for lavaan.mi objects

# S3 generic for S3 dispatch
fitMeasures <- function(object, fit_measures = "all",           # nolint
                        baseline_model = NULL, h1_model = NULL,
                        fm_args = list(
                          standard.test = "default",
                          scaled.test = "default",
                          rmsea.ci.level = 0.90,
                          rmsea.close.h0 = 0.05,
                          rmsea.notclose.h0 = 0.08,
                          robust = TRUE,
                          cat.nonpd = "na"
                        ),
                        output = "vector", level = NULL, ...) {
  UseMethod("fitMeasures", object)
}
fitmeasures <- function(object, fit_measures = "all",
                        baseline_model = NULL, h1_model = NULL,
                        fm_args = list(
                          standard.test = "default",
                          scaled.test = "default",
                          rmsea.ci.level = 0.90,
                          rmsea.close.h0 = 0.05,
                          rmsea.notclose.h0 = 0.08,
                          robust = TRUE,
                          cat.nonpd = "na"
                        ),
                        output = "vector", level = NULL, ...) {
  UseMethod("fitmeasures", object)
}


# S4 generic for S4 dispatch
setGeneric(
  "fitMeasures",
  function(object, fit_measures = "all",
           baseline_model = NULL, h1_model = NULL,
           fm_args = list(
             standard.test = "default",
             scaled.test = "default",
             rmsea.ci.level = 0.90,
             rmsea.close.h0 = 0.05,
             rmsea.notclose.h0 = 0.08,
             robust = TRUE,
             cat.nonpd = "na"
           ),
           output = "vector", level = NULL, ...) {
    standardGeneric("fitMeasures")
  }
)
setGeneric(
  "fitmeasures",
  function(object, fit_measures = "all",
           baseline_model = NULL, h1_model = NULL,
           fm_args = list(
             standard.test = "default",
             scaled.test = "default",
             rmsea.ci.level = 0.90,
             rmsea.close.h0 = 0.05,
             rmsea.notclose.h0 = 0.08,
             robust = TRUE,
             cat.nonpd = "na"
           ),
           output = "vector", level = NULL, ...) {
    standardGeneric("fitmeasures")
  }
)


# S3 generics
lavInspect <- function(object, what = "free",
                       add_labels = TRUE,
                       add_class = TRUE,
                       list_by_group = TRUE,
                       drop_list_single_group = TRUE,
                       ...) {
  UseMethod("lavInspect", object)
}

inspect <- function(object, what = "free", ...) {
    UseMethod("inspect", object)
}

lavTech <- function(object, what = "free",
                    add_labels = FALSE,
                    add_class = FALSE,
                    list_by_group = FALSE,
                    drop_list_single_group = FALSE,
                    ...) {
  UseMethod("lavTech", object)
}

lav_step15_baseline <- function(lavoptions = NULL,
                                       lavsamplestats = NULL,
                                       lavdata = NULL,
                                       lavcache = NULL,
                                       lavh1 = NULL,
                                       lavpartable = NULL) {
  # # # # # # # # # # #
  # #  15. baseline # #  (since 0.6-5)
  # # # # # # # # # # #

  # if options$do.fit and options$test not "none" and options$baseline = TRUE
  #   try fit.indep <- lav_object_independence(...)
  #   if not successful or not converged
  #     ** warning **
  #     lavbaseline < list()
  #   else
  #     lavbaseline <- list with partable and test of fit.indep
  lavbaseline <- list()
  if (lavoptions$do.fit &&
    !("none" %in% lavoptions$test) &&
    is.logical(lavoptions$baseline) && lavoptions$baseline) {
    if (lav_verbose()) {
      cat("lavbaseline ...")
    }
  current_verbose <- lav_verbose()
  lav_verbose(FALSE)
    fit_indep <- try(lav_object_independence(
      object = NULL,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache,
      lavoptions = lavoptions,
      lavpartable = lavpartable,
      lavh1 = lavh1
    ), silent = TRUE)
  lav_verbose(current_verbose)
    if (inherits(fit_indep, "try-error") || !fit_indep@optim$converged) {
      lav_msg_warn(gettext("estimation of the baseline model failed."))
      lavbaseline <- list()
      if (lav_verbose()) {
        cat(" FAILED.\n")
      }
    } else {
      # store relevant information
      lavbaseline <- list(
        partable = fit_indep@ParTable,
        test = fit_indep@test
      )
      if (lav_verbose()) {
        cat(" done.\n")
      }
    }
  }

  lavbaseline
}

# constrained optimization
# - references: * Nocedal & Wright (2006) Chapter 17
#               * Optimization with constraints by Madsen, Nielsen & Tingleff
#               * original papers: Powell, 1969 and Rockafeller, 1974
# - using 'nlminb' for the unconstrained subproblem
# - convergence scheme is based on the auglag function in the alabama package
nlminb_constr <- function(start, objective, gradient = NULL, hessian = NULL,
                          ..., scale = 1, control = list(),
                          lower = -Inf, upper = Inf,
                          ceq = NULL, ceq_jac = NULL,
                          cin = NULL, cin_jac = NULL,
                          control_outer = list()) {
  # we need a gradient
  stopifnot(!is.null(gradient))

  # if no 'ceq' or 'cin' function, we create a dummy one
  if (is.null(ceq)) {
    ceq <- function(x, ...) {
      numeric(0)
    }
  }
  if (is.null(cin)) {
    cin <- function(x, ...) {
      numeric(0)
    }
  }

  # if no user-supplied jacobian functions, create them
  if (is.null(ceq_jac)) {
    if (is.null(ceq)) {
      ceq_jac <- function(x, ...) {
        matrix(0, nrow = 0L, ncol = length(x))
      }
    } else {
      ceq_jac <- function(x, ...) {
        numDeriv::jacobian(func = ceq, x = x, ...)
      }
    }
  }
  if (is.null(cin_jac)) {
    if (is.null(cin)) {
      cin_jac <- function(x, ...) {
        matrix(0, nrow = 0L, ncol = length(x))
      }
    } else {
      cin_jac <- function(x, ...) {
        numDeriv::jacobian(func = cin, x = x, ...)
      }
    }
  }

  # how many ceq and cin constraints?
  nceq <- length(ceq(start))
  ncin <- length(cin(start))
  ncon <- nceq + ncin
  ceq_idx <- cin_idx <- integer(0)
  if (nceq > 0L) ceq_idx <- 1:nceq
  if (ncin > 0L) cin_idx <- nceq + 1:ncin
  cin_flag <- rep(FALSE, ncon)
  if (ncin > 0L) cin_flag[cin_idx] <- TRUE

  # control outer default values
  control_outer_default <- list(
    mu0 = 100,
    lambda0 = 10,
    tol = 1e-06, # changed this in 0.4-12
    itmax = 100L,
    verbose = FALSE
  )
  control_outer <- modifyList(control_outer_default, control_outer)


  # construct augmented lagrangian function
  auglag <- function(x, ...) {
    # apply constraints
    ceq0 <- ceq(x, ...)
    cin0 <- cin(x, ...)
    con0 <- c(ceq0, cin0)
    # 'release' inactive constraints
    if (ncin > 0L) {
      slack <- lambda / mu
      inactive_idx <- which(cin_flag & con0 > slack)
      con0[inactive_idx] <- slack[inactive_idx]
    }
    objective(x, ...) - sum(lambda * con0) + (mu / 2) * sum(con0 * con0)
  }

  fgrad <- function(x, ...) {
    # apply constraints
    ceq0 <- ceq(x, ...)
    cin0 <- cin(x, ...)
    con0 <- c(ceq0, cin0)
    # jacobian
    jac <- rbind(ceq_jac(x, ...), cin_jac(x, ...))
    lambda_jac <- lambda * jac

    # handle inactive constraints
    if (ncin > 0L) {
      slack <- lambda / mu
      inactive_idx <- which(cin_flag & con0 > slack)
      if (length(inactive_idx) > 0L) {
        jac <- jac[-inactive_idx, , drop = FALSE]
        lambda_jac <- lambda_jac[-inactive_idx, , drop = FALSE]
        con0 <- con0[-inactive_idx]
      }
    }

    if (nrow(jac) > 0L) {
      (gradient(x, ...) - colSums(lambda_jac) +
        mu * as.numeric(t(jac) %*% con0))
    } else {
      gradient(x, ...)
    }
  }


  # initialization
  ceq0 <- ceq(start, ...)
  cin0 <- cin(start, ...)
  con0 <- c(ceq0, cin0)
  lambda <- rep(control_outer$lambda0, length(con0))
  mu <- control_outer$mu0
  inactive_idx <- integer(0)
  if (ncin > 0L) {
    slack <- lambda / mu
    inactive_idx <- which(cin_flag & con0 > slack)
    con0[inactive_idx] <- slack[inactive_idx]
  }
  m_k <- max(abs(con0))
  if (control_outer$verbose) {
    cat("init cin0 values: ", cin0, "\n")
    cat("init ceq0 values: ", ceq0, "\n")
    cat("init slack values: ", lambda / mu, "\n")
    cat("init inactive idx: ", inactive_idx, "\n")
    cat("init con0 values: ", con0, "\n")
    cat("K = max con0: ", m_k, "\n")
  }

  r <- objective(start, ...) # r := obj
  feval <- 0L
  geval <- 0L
  niter <- 0L
  ilack <- 0L
  kprev <- m_k
  mu0 <- control_outer$mu0 / kprev
  if (is.infinite(mu0)) mu0 <- 1.0
  mu <- mu0

  m_k <- Inf
  x_par <- start
  for (i in 1:control_outer$itmax) {
    x_old <- x_par
    r_old <- r
    ############################################################
    if (control_outer$verbose) {
      cat("\nStarting inner optimization [", i, "]:\n")
      cat("lambda: ", lambda, "\n")
      cat("mu: ", mu, "\n")
    }
    optim_out <- nlminb(
      start = x_par, objective = auglag,
      gradient = fgrad, control = control,
      lower = lower, upper = upper,
      scale = scale, ...
    )
    ############################################################
    x_par <- optim_out$par
    r <- optim_out$objective
    feval <- feval + optim_out$evaluations[1]
    geval <- geval + optim_out$evaluations[2]
    niter <- niter + optim_out$iterations

    # check constraints
    ceq0 <- ceq(x_par, ...)
    cin0 <- cin(x_par, ...)
    con0 <- c(ceq0, cin0)
    if (ncin > 0L) {
      slack <- lambda / mu
      inactive_idx <- which(cin_flag & con0 > slack)
      con0[inactive_idx] <- slack[inactive_idx]
    }
    m_k <- max(abs(con0))
    if (control_outer$verbose) {
      cat("cin0 values: ", cin0, "\n")
      cat("ceq0 values: ", ceq0, "\n")
      cat("active threshold: ", lambda / mu, "\n")
      cat("inactive idx: ", inactive_idx, "\n")
      cat("con0 values: ", con0, "\n")
      cat("K = max con0: ", m_k, " Kprev = ", kprev, "\n")
    }

    # update K or mu (see Powell, 1969)
    if (m_k <= kprev / 4) {
      lambda <- lambda - (mu * con0)
      kprev <- m_k
    } else {
      mu <- 10 * mu
    }

    # check convergence
    pconv <- max(abs(x_par - x_old))
    if (pconv < control_outer$tol) {
      ilack <- ilack + 1L
    } else {
      ilack <- 0L
    }

    if ((is.finite(r) && is.finite(r_old) &&
      abs(r - r_old) < control_outer$tol && m_k < control_outer$tol) ||
      ilack >= 3) {
      break
    }
  }

  # output
  a <- list()

  if (i == control_outer$itmax) {
    a$convergence <- 10L
    a$message <- "nlminb_constr ran out of iterations and did not converge"
  } else if (m_k > control_outer$tol) {
    a$convergence <- 11L
    a$message <- "Convergence due to lack of progress in parameter updates"
  } else {
    a$convergence <- 0L
    a$message <- "converged"
  }
  a$par <- optim_out$par
  a$outer.iterations <- i
  a$lambda <- lambda
  a$mu <- mu
  # a$value <- objective(a$start, ...)
  # a$cin <- cin(a$start, ...)
  # a$ceq <- ceq(a$start, ...)
  a$evaluations <- c(feval, geval)
  a$iterations <- niter
  # a$kkt1 <- max(abs(a$fgrad)) <= 0.01 * (1 + abs(a$value))
  # a$kkt2 <- any(eigen(a$hessian)$value * control.optim$objectivescale> 0)

  # jacobian of ceq and 'active' cin
  ceq0 <- ceq(a$par, ...)
  cin0 <- cin(a$par, ...)
  con0 <- c(ceq0, cin0)
  jac <- rbind(ceq_jac(a$par, ...), cin_jac(a$par, ...))
  inactive_idx <- integer(0L)
  cin_idx <- which(cin_flag)
  # ceq.idx <- which(!cin.flag)
  if (ncin > 0L) {
    # strict complementarity: a row is only ACTIVE (binding) when it is
    # satisfied with near-equality AND its multiplier exerts force on the
    # gradient (see lav_con_cin_inactive_idx()); the multipliers of the
    # non-binding rows have been driven to zero by the augmented-Lagrangian
    # updates. (The old |con0| <= 1e-05 rule wrongly treated an interior
    # solution that merely grazes a bound as equality-constrained.)
    inactive_idx <- lav_con_cin_inactive_idx(
      con0 = con0, lambda = lambda, jac = jac, cin_flag = cin_flag
    )
  }
  attr(jac, "inactive.idx") <- inactive_idx
  attr(jac, "cin.idx") <- cin_idx
  attr(jac, "ceq.idx") <- ceq_idx
  a$con.jac <- jac

  a
}

# Magnus & Neudecker (1999) style matrix operations
# YR - 11 may 2011: initial version
# YR - 19 okt 2014: rename functions using lav_matrix_ prefix

# vec operator
#
# the vec operator (for 'vectorization') transforms a matrix into
# a vector by stacking the *columns* of the matrix one underneath the other
#
# M&N book: page 30
#
# note: we do not coerce to 'double/numeric' storage-mode (like as.numeric)
lav_mat_vec <- function(a) {
  as.vector(a)
}


# vecr operator
#
# the vecr operator transforms a matrix into
# a vector by stacking the *rows* of the matrix one underneath the other
lav_mat_vecr <- function(a) {
  # faster way??
  # nRow <- NROW(A); nCol <- NCOL(A)
  # idx <- (seq_len(nCol) - 1L) * nRow + rep(seq_len(nRow), each = nCol)
  # if (lav_use_lavaanC() && is.numeric(A)) {
  #   return(lavaanC::m_vecr(A))
  # }
  lav_mat_vec(t(a))
}


# vech
#
# the vech operator (for 'half vectorization') transforms a *symmetric* matrix
# into a vector by stacking the *columns* of the matrix one underneath the
# other, but eliminating all supradiagonal elements
#
# see Henderson & Searle, 1979
#
# M&N book: page 48-49
#
lav_mat_vech <- function(s, diagonal = TRUE) {
  # if (lav_use_lavaanC() && is.numeric(S)) {
  #   return(lavaanC::m_vech(S, diagonal))
  # }
  row_1 <- row(s)
  col_1 <- col(s)
  if (diagonal) s[row_1 >= col_1] else s[row_1 > col_1]
}


# vech of a symmetric gradient matrix, doubling the off-diagonal elements
# (chain rule when differentiating with respect to vech(Sigma):
#  off-diagonal elements appear twice in Sigma)
lav_mat_vech_dd <- function(s) {
  tmp <- s * 2
  diag(tmp) <- diag(s)
  lav_mat_vech(tmp)
}


# the vechr operator transforms a *symmetric* matrix
# into a vector by stacking the *rows* of the matrix one after the
# other, but eliminating all supradiagonal elements
lav_mat_vechr <- function(s, diagonal = TRUE) {
  # if (lav_use_lavaanC() && is.numeric(S)) {
  #   return(lavaanC::m_vechr(S, diagonal))
  # }
  s[lav_mat_vechr_idx(n = NCOL(s), diagonal = diagonal)]
}


# the vechu operator transforms a *symmetric* matrix
# into a vector by stacking the *columns* of the matrix one after the
# other, but eliminating all infradiagonal elements
lav_mat_vechu <- function(s, diagonal = TRUE) {
  # if (lav_use_lavaanC() && is.numeric(S)) {
  #   return(lavaanC::m_vechu(S, diagonal))
  # }
  s[lav_mat_vechu_idx(n = NCOL(s), diagonal = diagonal)]
}

# the vechru operator transforms a *symmetric* matrix
# into a vector by stacking the *rows* of the matrix one after the
# other, but eliminating all infradiagonal elements
#
# same as vech (but using upper-diagonal elements)
lav_mat_vechru <- function(s, diagonal = TRUE) {
  # if (lav_use_lavaanC() && is.numeric(S)) {
  #   return(lavaanC::m_vechru(S, diagonal))
  # }
  s[lav_mat_vechru_idx(n = NCOL(s), diagonal = diagonal)]
}

# return the *vector* indices of the lower triangular elements of a
# symmetric matrix of size 'n'
lav_mat_vech_idx <- function(n = 1L, diagonal = TRUE) {
  n <- as.integer(n)
  # if (lav_use_lavaanC() && n > 1L) {
  #   return(lavaanC::m_vech_idx(n, diagonal))
  # }
  if (n < 1L) {
    return(integer(0L))
  }
  if (diagonal) {
    sequence(n:1L) + rep((seq_len(n) - 1L) * (n + 1L), n:1L)
  } else {
    if (n == 1L) {
      return(integer(0L))
    }
    (sequence((n - 1L):1L) +
      rep((seq_len(n - 1L) - 1L) * (n + 1L) + 1L, (n - 1L):1L))
  }
}

# return the *row* indices of the lower triangular elements of a
# symmetric matrix of size 'n'
lav_mat_vech_row_idx <- function(n = 1L, diagonal = TRUE) {
  n <- as.integer(n)
  # if (lav_use_lavaanC() && n > 1L) {
  #   return(lavaanC::m_vech_row_idx(n, diagonal))
  # }
  if (diagonal) {
    sequence(n:1L) + rep(seq_len(n) - 1L, n:1L)
  } else {
    if (n <= 1L) {
      return(integer(0L))
    }
    sequence((n - 1L):1L) + rep(seq_len(n - 1L), (n - 1L):1L)
  }
}

# return the *col* indices of the lower triangular elements of a
# symmetric matrix of size 'n'
lav_mat_vech_col_idx <- function(n = 1L, diagonal = TRUE) {
  n <- as.integer(n)
  # if (lav_use_lavaanC() && n > 1L) {
  #   return(lavaanC::m_vech_col_idx(n, diagonal))
  # }
  if (!diagonal) {
    n <- n - 1L
  }
  rep.int(seq_len(n), times = rev(seq_len(n)))
}


# return the *vector* indices of the lower triangular elements of a
# symmetric matrix of size 'n' -- ROW-WISE
lav_mat_vechr_idx <- function(n = 1L, diagonal = TRUE) {
  n <- as.integer(n)
  # if (lav_use_lavaanC() && n > 1L) {
  #   return(lavaanC::m_vechr_idx(n, diagonal))
  # }
  if (n < 1L) {
    return(integer(0L))
  }
  if (diagonal) {
    row_idx <- rep(seq_len(n), seq_len(n))
    col_idx <- sequence(seq_len(n))
  } else {
    if (n == 1L) {
      return(integer(0L))
    }
    row_idx <- rep(seq_len(n)[-1L], seq_len(n - 1L))
    col_idx <- sequence(seq_len(n - 1L))
  }
  (col_idx - 1L) * n + row_idx
}

# return the *vector* indices of the upper triangular elements of a
# symmetric matrix of size 'n' -- COLUMN-WISE
lav_mat_vechu_idx <- function(n = 1L, diagonal = TRUE) {
  n <- as.integer(n)
  # if (lav_use_lavaanC() && n > 1L) {
  #   return(lavaanC::m_vechu_idx(n, diagonal))
  # }
  if (n < 1L) {
    return(integer(0L))
  }
  if (diagonal) {
    sequence(seq_len(n)) + rep((seq_len(n) - 1L) * n, seq_len(n))
  } else {
    if (n == 1L) {
      return(integer(0L))
    }
    sequence(seq_len(n - 1L)) + rep(seq_len(n - 1L) * n, seq_len(n - 1L))
  }
}

# return the *vector* indices of the upper triangular elements of a
# symmetric matrix of size 'n' -- ROW-WISE
lav_mat_vechru_idx <- function(n = 1L, diagonal = TRUE) {
  n <- as.integer(n)
  # if (lav_use_lavaanC() && n > 1L) {
  #   return(lavaanC::m_vechru_idx(n, diagonal))
  # }
  if (n < 1L) {
    return(integer(0L))
  }
  if (diagonal) {
    row_idx <- rep(seq_len(n), n:1L)
    col_idx <- sequence(n:1L) + rep(seq_len(n) - 1L, n:1L)
  } else {
    if (n == 1L) {
      return(integer(0L))
    }
    row_idx <- rep(seq_len(n - 1L), (n - 1L):1L)
    col_idx <- sequence((n - 1L):1L) + rep(seq_len(n - 1L), (n - 1L):1L)
  }
  (col_idx - 1L) * n + row_idx
}


# vech.reverse and vechru.reverse (aka `upper2full')
#
# given the output of vech(S) --or vechru(S) which is identical--
# reconstruct S
lav_mat_vech_rev <- function(x, diagonal = TRUE) {
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_vech_reverse(x, diagonal))
  # }
  # guess dimensions
  if (diagonal) {
    p <- (sqrt(1 + 8 * length(x)) - 1) / 2
  } else {
    p <- (sqrt(1 + 8 * length(x)) + 1) / 2
  }
  stopifnot(p == round(p, 0))

  s <- numeric(p * p)
  s[lav_mat_vech_idx(p, diagonal = diagonal)] <- x
  s[lav_mat_vechru_idx(p, diagonal = diagonal)] <- x

  attr(s, "dim") <- c(p, p)
  s
}
lav_mat_vechru_rev <- lav_alias("lav_mat_vech_rev")
lav_mat_upper2full <- lav_alias("lav_mat_vech_rev")


# vechr.reverse vechu.reverse (aka `lower2full')
#
# given the output of vechr(S) --or vechu(S) which is identical--
# reconstruct S
lav_mat_vechr_rev <- function(x, diagonal = TRUE) {
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_vechr_reverse(x, diagonal))
  # }
  # guess dimensions
  if (diagonal) {
    p <- (sqrt(1 + 8 * length(x)) - 1) / 2
  } else {
    p <- (sqrt(1 + 8 * length(x)) + 1) / 2
  }
  stopifnot(p == round(p, 0))

  s <- numeric(p * p)
  s[lav_mat_vechr_idx(p, diagonal = diagonal)] <- x
  s[lav_mat_vechu_idx(p, diagonal = diagonal)] <- x

  attr(s, "dim") <- c(p, p)
  s
}
lav_mat_vechu_rev <- lav_alias("lav_mat_vechr_rev")
lav_mat_lower2full <- lav_alias("lav_mat_vechr_rev")


# return the *vector* indices of the diagonal elements of a symmetric
# matrix of size 'n'
lav_mat_diag_idx <- function(n = 1L) {
  # if(n < 1L) return(integer(0L))
  n <- as.integer(n)
  if (n < 1L) {
    return(integer(0L))
  }
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_diag_idx(n))
  # }
  1L + (seq_len(n) - 1L) * (n + 1L)
}


# return the *vector* indices of the diagonal elements of the LOWER part
# of a symmetric matrix of size 'n'
lav_mat_diagh_idx <- function(n = 1L) {
  n <- as.integer(n)
  if (n < 1L) {
    return(integer(0L))
  }
  if (n == 1L) {
    return(1L)
  }
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_diagh_idx(n))
  # }
  c(1L, cumsum(n:2L) + 1L)
}


# return the *vector* indices of the ANTI diagonal elements of a symmetric
# matrix of size 'n'
lav_mat_antidiag_idx <- function(n = 1L) {
  if (n < 1L) {
    return(integer(0L))
  }
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_antidiag_idx(n))
  # }
  1L + seq_len(n) * (n - 1L)
}

# return the *vector* indices of 'idx' elements in a vech() matrix
#
# eg if n = 4 and type == "and" and idx = c(2,4)
#    we create matrix A =
#       [,1]  [,2]  [,3]  [,4]
# [1,] FALSE FALSE FALSE FALSE
# [2,] FALSE  TRUE FALSE  TRUE
# [3,] FALSE FALSE FALSE FALSE
# [4,] FALSE  TRUE FALSE  TRUE
#
# and the result is c(5,7,10)
#
# eg if n = 4 and type == "or" and idx = c(2,4)
#    we create matrix A =
#       [,1] [,2]  [,3] [,4]
# [1,] FALSE TRUE FALSE TRUE
# [2,]  TRUE TRUE  TRUE TRUE
# [3,] FALSE TRUE FALSE TRUE
# [4,]  TRUE TRUE  TRUE TRUE
#
# and the result is c(2, 4, 5, 6, 7, 9, 10)
#
lav_mat_vech_which_idx <- function(n = 1L, diagonal = TRUE,
                                      idx = integer(0L), type = "and",
                                      add_idx_at_start = FALSE) {
  if (length(idx) == 0L) {
    return(integer(0L))
  }
  n <- as.integer(n)
  idx <- as.integer(idx)
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_vech_which_idx(n, diagonal,
  #                         idx, type, add.idx.at.start))
  # }
  a <- matrix(FALSE, n, n)
  if (type == "and") {
    a[idx, idx] <- TRUE
  } else if (type == "or") {
    a[idx, ] <- TRUE
    a[, idx] <- TRUE
  }
  pstar_idx <- which(lav_mat_vech(a, diagonal = diagonal))

  if (add_idx_at_start) {
    pstar_idx <- c(idx, pstar_idx + n)
  }

  pstar_idx
}

# similar to lav_mat_vech_which_idx(), but
# - only 'type = and'
# - order of idx matters!
lav_mat_vech_match_idx <- function(n = 1L, diagonal = TRUE,
                                      idx = integer(0L)) {
  if (length(idx) == 0L) {
    return(integer(0L))
  }
  n <- as.integer(n)
  idx <- as.integer(idx)
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_vech_match_idx(n, diagonal, idx))
  # }
  pstar <- n * (n + 1) / 2
  a <- lav_mat_vech_rev(seq_len(pstar))
  m_b <- a[idx, idx, drop = FALSE]
  lav_mat_vech(m_b, diagonal = diagonal)
}

# check if square matrix is diagonal
lav_mat_is_diagonal <- function(a = NULL, tol = .Machine$double.eps) {
  a <- as.matrix.default(a)
  stopifnot(nrow(a) == ncol(a))
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_is_diagonal(A))
  # }
  diag(a) <- 0
  all(abs(a) <= tol)
}


# create the duplication matrix (D_n): it 'duplicates' the elements
# in vech(S) to create vec(S) (where S is symmetric)
#
# D %*% vech(S) == vec(S)
#
# M&N book: pages 48-50
#

# dup3: using col idx only
# D7 <- dup(7L); x<- apply(D7, 1, function(x) which(x > 0)); matrix(x,7,7)
lav_mat_dup <- function(n = 1L) {
  n <- as.integer(n)
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_duplication(n))
  # }
  if ((n < 1L) || (round(n) != n)) {
    lav_msg_stop(gettext("n must be a positive integer"))
  }

  if (n > 255L) {
    lav_msg_stop(gettext("n is too large"))
  }

  nstar <- n * (n + 1) / 2
  n2 <- n * n
  # THIS is the real bottleneck: allocating an ocean of zeroes...
  x <- numeric(n2 * nstar)

  tmp <- matrix(0L, n, n)
  tmp[lav_mat_vech_idx(n)] <- seq_len(nstar)
  tmp[lav_mat_vechru_idx(n)] <- seq_len(nstar)

  idx <- seq_len(n2) + (lav_mat_vec(tmp) - 1L) * n2

  x[idx] <- 1.0

  attr(x, "dim") <- c(n2, nstar)
  x
}

# compute t(D) %*% A (without explicitly computing D)
# sqrt(nrow(A)) is an integer
# A is not symmetric, and not even square, only n^2 ROWS
lav_mat_dup_pre <- function(a = matrix(0, 0, 0)) {
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_duplication_pre(A))
  # }
  # number of rows
  n2 <- NROW(a)

  # square nrow(A) only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)

  # dup idx
  idx1 <- lav_mat_vech_idx(n)
  idx2 <- lav_mat_vechru_idx(n)

  out <- a[idx1, , drop = FALSE] + a[idx2, , drop = FALSE]
  u <- lav_mat_diagh_idx(n)
  out[u, ] <- out[u, ] / 2.0

  out
}

# compute A %*% D (without explicitly computing D)
# sqrt(ncol(A)) must be an integer
# A is not symmetric, and not even square, only n^2 COLUMNS
lav_mat_dup_post <- function(a = matrix(0, 0, 0)) {
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_duplication_post(A))
  # }
  # number of columns
  n2 <- NCOL(a)

  # square a only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)

  # dup idx
  idx1 <- lav_mat_vech_idx(n)
  idx2 <- lav_mat_vechru_idx(n)

  out <- a[, idx1, drop = FALSE] + a[, idx2, drop = FALSE]
  u <- lav_mat_diagh_idx(n)
  out[, u] <- out[, u] / 2.0

  out
}


# compute t(D) %*% A %*% D (without explicitly computing D)
# A must be a square matrix and sqrt(ncol) an integer
lav_mat_dup_pre_post <- function(a = matrix(0, 0, 0)) {
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_duplication_pre_post(A))
  # }
  # number of columns
  n2 <- NCOL(a)

  # square A only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(NROW(a) == n2, n * n == n2)

  # dup idx
  idx1 <- lav_mat_vech_idx(n)
  idx2 <- lav_mat_vechru_idx(n)

  out <- a[idx1, , drop = FALSE] + a[idx2, , drop = FALSE]
  u <- lav_mat_diagh_idx(n)
  out[u, ] <- out[u, ] / 2.0
  out <- out[, idx1, drop = FALSE] + out[, idx2, drop = FALSE]
  out[, u] <- out[, u] / 2.0

  out
}

# 'partial' correlation transform of a covariance matrix: scale only the
# variables in 'cor_idx' to unit variance, leaving the others in their
# original metric. If 'cor_idx' covers all variables, this is cov2cor().
lav_cov2cor_partial <- function(s = matrix(0, 0, 0), cor_idx = NULL) {
  if (is.null(cor_idx) || length(cor_idx) == 0L) {
    return(stats::cov2cor(s))
  }
  scale <- rep(1.0, nrow(s))
  scale[cor_idx] <- sqrt(diag(s)[cor_idx])
  s / tcrossprod(scale)
}

# compute t(D) %*% A %*% D (without explicitly computing D)
# A must be a square matrix and sqrt(ncol) an integer
# correlation version: ignoring diagonal elements
lav_mat_dup_cor_pre_post <- function(a = matrix(0, 0, 0)) {
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_duplication_cor_pre_post(A))
  # }
  # number of columns
  n2 <- NCOL(a)

  # square A only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(NROW(a) == n2, n * n == n2)

  # dup idx
  idx1 <- lav_mat_vech_idx(n, diagonal = FALSE)
  idx2 <- lav_mat_vechru_idx(n, diagonal = FALSE)

  out <- a[idx1, , drop = FALSE] + a[idx2, , drop = FALSE]
  out <- out[, idx1, drop = FALSE] + out[, idx2, drop = FALSE]

  out
}


# create the generalized inverse of the duplication matrix (D^+_n):
# it removes the duplicated elements in vec(S) to create vech(S)
# if S is symmetric
#
# D^+ %*% vec(S) == vech(S)
#
# M&N book: page 49
#
# D^+ == solve(t(D_n %*% D_n) %*% t(D_n)

# create DUP.ginv without transpose
lav_mat_dup_ginv <- function(n = 1L) {
  # if (lav_use_lavaanC()) {
  #   n <- as.integer(n)
  #   return(lavaanC::m_duplication_ginv(n))
  # }
  if ((n < 1L) || (round(n) != n)) {
    lav_msg_stop(gettext("n must be a positive integer"))
  }

  if (n > 255L) {
    lav_msg_stop(gettext("n is too large"))
  }

  nstar <- n * (n + 1) / 2
  n2 <- n * n
  # THIS is the real bottleneck: allocating an ocean of zeroes...
  x <- numeric(nstar * n2)

  x[(lav_mat_vech_idx(n) - 1L) * nstar + seq_len(nstar)] <- 0.5
  x[(lav_mat_vechru_idx(n) - 1L) * nstar + seq_len(nstar)] <- 0.5
  x[(lav_mat_diag_idx(n) - 1L) * nstar + lav_mat_diagh_idx(n)] <- 1.0

  attr(x, "dim") <- c(nstar, n2)
  x
}

# pre-multiply with D^+
# number of rows in A must be 'square' (n*n)
lav_mat_dup_ginv_pre <- function(a = matrix(0, 0, 0)) {
  m_a <- as.matrix.default(a)

  # number of rows
  n2 <- NROW(m_a)

  # square nrow(A) only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)
  nstar <- n * (n + 1) / 2

  idx1 <- lav_mat_vech_idx(n)
  idx2 <- lav_mat_vechru_idx(n)
  out <- (m_a[idx1, , drop = FALSE] + m_a[idx2, , drop = FALSE]) / 2
  out
}

# post-multiply with t(D^+)
# number of columns in A must be 'square' (n*n)
lav_mat_dup_ginv_post <- function(a = matrix(0, 0, 0)) { #  nolint
  m_a <- as.matrix.default(a)

  # number of columns
  n2 <- NCOL(m_a)

  # square A only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)

  idx1 <- lav_mat_vech_idx(n)
  idx2 <- lav_mat_vechru_idx(n)
  out <- (m_a[, idx1, drop = FALSE] + m_a[, idx2, drop = FALSE]) / 2
  out
}

# pre AND post-multiply with D^+: D^+ %*% A %*% t(D^+)
# for square matrices only, with ncol = nrow = n^2
lav_mat_dup_ginv_pre_post <- function(a = matrix(0, 0, 0)) {
  m_a <- as.matrix.default(a)
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_duplication_ginv_pre_post(A))
  # }
  # number of columns
  n2 <- NCOL(m_a)

  # square A only, n2 = n^2
  n <- as.integer(round(sqrt(n2)))
  stopifnot(NROW(m_a) == n2, n * n == n2)

  idx1 <- lav_mat_vech_idx(n)
  idx2 <- lav_mat_vechru_idx(n)
  out <- (m_a[idx1, , drop = FALSE] + m_a[idx2, , drop = FALSE]) / 2
  out <- (out[, idx1, drop = FALSE] + out[, idx2, drop = FALSE]) / 2
  out
}


# create the commutation matrix (K_mn)
# the mn x mx commutation matrix is a permutation matrix which
# transforms vec(A) into vec(A')
#
# K_mn %*% vec(A) == vec(A')
#
# (in Henderson & Searle 1979, it is called the vec-permutation matrix)
# M&N book: pages 46-48
#
# note: K_mn is a permutation matrix, so it is orthogonal: t(K_mn) = K_mn^-1
#       K_nm %*% K_mn == I_mn
#
# it is called the 'commutation' matrix because it enables us to interchange
# ('commute') the two matrices of a Kronecker product, eg
#   K_pm (A %x% B) K_nq == (B %x% A)
#
# important property: it allows us to transform a vec of a Kronecker product
# into the Kronecker product of the vecs (if A is m x n and B is p x q):
#   vec(A %x% B) == (I_n %x% K_qm %x% I_p)(vec A %x% vec B)

# first attempt
lav_mat_com <- function(m = 1L, n = 1L) {
  # if (lav_use_lavaanC()) {
  #   m <- as.integer(m)
  #   n <- as.integer(n)
  #   return(lavaanC::m_commutation(m, n))
  # }

  if ((m < 1L) || (round(m) != m)) {
    lav_msg_stop(gettext("m must be a positive integer"))
  }

  if ((n < 1L) || (round(n) != n)) {
    lav_msg_stop(gettext("n must be a positive integer"))
  }

  p <- m * n
  x <- numeric(p * p)

  pattern <- rep(c(rep((m + 1L) * n, (m - 1L)), n + 1L), n)
  idx <- c(1L, 1L + cumsum(pattern)[-p])

  x[idx] <- 1.0
  attr(x, "dim") <- c(p, p)

  x
}

# compute K_n %*% A without explicitly computing K
# K_n = K_nn, so sqrt(nrow(A)) must be an integer!
# = permuting the rows of A
lav_mat_com_pre <- function(a = matrix(0, 0, 0)) {
  a <- as.matrix(a)

  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_commutation_pre(A))
  # }

  # number of rows of A
  n2 <- nrow(a)

  # K_nn only (n2 = m * n)
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)

  # compute row indices
  # row.idx <- as.integer(t(matrix(1:n2, n, n)))
  row_idx <- rep(seq_len(n), each = n) + (seq_len(n) - 1L) * n

  out <- a[row_idx, , drop = FALSE]
  out
}

# compute A %*% K_n without explicitly computing K
# K_n = K_nn, so sqrt(ncol(A)) must be an integer!
# = permuting the columns of A
lav_mat_com_post <- function(a = matrix(0, 0, 0)) {
  m_a <- as.matrix(a)

  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_commutation_post(a))
  # }

  # number of columns of a
  n2 <- ncol(m_a)

  # K_nn only (n2 = m * n)
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)

  # compute col indices
  # row.idx <- as.integer(t(matrix(1:n2, n, n)))
  col_idx <- rep(seq_len(n), each = n) + (seq_len(n) - 1L) * n

  out <- m_a[, col_idx, drop = FALSE]
  out
}

# compute K_n %*% A %*% K_n without explicitly computing K
# K_n = K_nn, so sqrt(ncol(A)) must be an integer!
# = permuting both the rows AND columns of A
lav_mat_com_pre_post <- function(a = matrix(0, 0, 0)) {
  m_a <- as.matrix(a)
  # if (lav_use_lavaanC()) {
  #   return(lavaanC::m_commutation_pre_post(A))
  # }
  # number of columns of A
  n2 <- NCOL(m_a)

  # K_nn only (n2 = m * n)
  n <- as.integer(round(sqrt(n2)))
  stopifnot(n * n == n2)

  # compute col indices
  row_idx <- rep(seq_len(n), each = n) + (seq_len(n) - 1L) * n
  col_idx <- row_idx

  out <- m_a[row_idx, col_idx, drop = FALSE]
  out
}


# compute K_mn %*% A without explicitly computing K
# = permuting the rows of A
lav_mat_com_mn_pre <- function(a, m = 1L, n = 1L) {
  # number of rows of A
  mn <- NROW(a)
  stopifnot(mn == m * n)

  # compute row indices
  # row.idx <- as.integer(t(matrix(1:mn, m, n)))
  row_idx <- rep(seq_len(m), each = n) + (seq_len(n) - 1L) * m

  out <- a[row_idx, , drop = FALSE]
  out
}

# shortcut for the idiom 't(D) %*% (S %x% S) %*% D'
# where S is symmetric, and D is the duplication matrix
# lav_matrix_tD_SxS_D <- function(S) {

#TODO: continue coding here

# }

# square root of a positive definite symmetric matrix
lav_mat_sym_sqrt <- function(s = matrix(0, 0, 0)) {
  n <- NROW(s)

  # eigen decomposition, assume symmetric matrix
  s_eigen <- eigen(s, symmetric = TRUE)
  v <- s_eigen$vectors
  d <- s_eigen$values

  # 'fix' slightly negative tiny numbers
  d[d < 0] <- 0.0

  # sqrt the eigenvalues and reconstruct (scale columns of V,
  #                                 avoid n*n diagonal alloc)
  s_sqrt <- tcrossprod(v * rep(sqrt(d), each = n), v)

  s_sqrt
}

# orthogonal complement of a matrix A
# see Satorra (1992). Sociological Methodology, 22, 249-278, footnote 3:
#
# To compute such an orthogonal matrix, consider the p* x p* matrix P = I -
# A(A'A)^-1A', which is idempotent of rank p* - q. Consider the singular value
# decomposition P = HVH', where H is a p* x (p* - q) matrix of full column rank,
# and V is a (p* - q) x (p* - q) diagonal matrix. It is obvious that H'A = 0;
# hence, H is the desired orthogonal complement. This method of constructing an
# orthogonal complement was proposed by Heinz Neudecker (1990, pers. comm.).
#
# update YR 21 okt 2014:
# - note that A %*% solve(t(A) %*% A) %*% t(A) == tcrossprod(qr.Q(qr(A)))
# - if we are using qr, we can as well use qr.Q to get the complement
#
lav_mat_ortho_complement <- function(a = matrix(0, 0, 0)) {
  qr_1 <- qr(a)
  ran_k <- qr_1$rank

  # following Heinz Neudecker:
  # n <- nrow(A)
  # P <- diag(n) - tcrossprod(qr.Q(QR))
  # OUT <- svd(P)$u[, seq_len(n - ranK), drop = FALSE]

  q_1 <- qr.Q(qr_1, complete = TRUE)
  # get rid of the first ranK columns
  out <- q_1[, -seq_len(ran_k), drop = FALSE]

  out
}

# construct block diagonal matrix from a list of matrices
# ... can contain multiple arguments, which will be coerced to a list
#     or it can be a single list (of matrices)
lav_mat_bdiag <- function(...) {
  if (nargs() == 0L) {
    return(matrix(0, 0, 0))
  }
  dots <- list(...)

  # create list of matrices
  if (is.list(dots[[1]])) {
    mlist <- dots[[1]]
  } else {
    mlist <- dots
  }
  if (length(mlist) == 1L) {
    return(mlist[[1]])
  }

  # more than 1 matrix
  nmat <- length(mlist)
  nrows <- sapply(mlist, NROW)
  crows <- cumsum(nrows)
  ncols <- sapply(mlist, NCOL)
  ccols <- cumsum(ncols)
  trows <- sum(nrows)
  tcols <- sum(ncols)
  # empty names will shorten the c/r-names vector
  cnames <- unlist(lapply(mlist, colnames))
  rnames <- unlist(lapply(mlist, rownames))

  x <- numeric(trows * tcols)

  for (m in seq_len(nmat)) {
    if (m > 1L) {
      rcoffset <- trows * ccols[m - 1] + crows[m - 1]
    } else {
      rcoffset <- 0L
    }
    m_idx <- (rep((seq_len(ncols[m]) - 1L) * trows, each = nrows[m]) +
      rep(seq_len(nrows[m]), ncols[m]) + rcoffset)
    x[m_idx] <- mlist[[m]]
  }

  attr(x, "dim") <- c(trows, tcols)
  # col/row names (only if complete)
  if (nrow(x) == length(rnames) && ncol(x) == length(cnames)) {
    attr(x, "dimnames") <- list(rnames, cnames)
  }

  x
}

# trace of a single square matrix, or the trace of a product of (compatible)
# matrices resulting in a single square matrix
lav_mat_trace <- function(..., check = TRUE) {
  if (nargs() == 0L) {
    return(as.numeric(NA))
  }
  dots <- list(...)

  # create list of matrices
  if (is.list(dots[[1]])) {
    mlist <- dots[[1]]
  } else {
    mlist <- dots
  }

  # number of matrices
  n_mat <- length(mlist)

  # single matrix
  if (n_mat == 1L) {
    s <- mlist[[1]]
    if (check) {
      # check if square
      stopifnot(NROW(s) == NCOL(s))
    }
    out <- sum(s[lav_mat_diag_idx(n = NROW(s))])
  } else if (n_mat == 2L) {
    # dimension check is done by '*'
    out <- sum(mlist[[1]] * t(mlist[[2]]))
  } else if (n_mat == 3L) {
    a <- mlist[[1]]
    m_b <- mlist[[2]]
    m_c <- mlist[[3]]

    # A, B, m_c
    # below is the logic; to be coded inline
    # DIAG <- numeric( NROW(A) )
    # for(i in seq_len(NROW(A))) {
    #     DIAG[i] <- sum( rep(A[i,], times = NCOL(B)) *
    #                  as.vector(B) *
    #                  rep(m_c[,i], each=NROW(B)) )
    # }
    # out <- sum(DIAG)

    # FIXME:

    # dimension check is automatic
    b2 <- m_b %*% m_c
    out <- sum(a * t(b2))
  } else {
    # nRows <- sapply(mlist, NROW)
    # nCols <- sapply(mlist, NCOL)

    # check if product is ok
    # stopifnot(all(nCols[seq_len(nMat-1L)] == nRows[2:nMat]))

    # check if product is square
    # stopifnot(nRows[1] == nCols[nMat])

    m1 <- mlist[[1]]
    m2 <- mlist[[2]]
    for (m in 3L:n_mat) {
      m2 <- m2 %*% mlist[[m]]
    }
    out <- sum(m1 * t(m2))
  }

  out
}

# crossproduct, but handling NAs pairwise, if needed
# otherwise, just call base::crossprod
lav_mat_crossprod <- function(a, m_b) {
  # single argument?
  if (missing(m_b)) {
    if (!anyNA(a)) {
      return(base::crossprod(a))
    }
    m_b <- a
    # no missings?
  } else if (!anyNA(a) && !anyNA(m_b)) {
    return(base::crossprod(a, m_b))
  }

  # A and B must be matrices
  if (!inherits(a, "matrix")) {
    a <- matrix(a)
  }
  if (!inherits(m_b, "matrix")) {
    m_b <- matrix(m_b)
  }

  out <- apply(m_b, 2L, function(x) colSums(a * x, na.rm = TRUE))

  # only when A is a vector, and B is a matrix, we get back a vector
  # while the result should be a matrix with 1-row
  if (!is.matrix(out)) {
    out <- t(matrix(out))
  }

  out
}


# reduced row echelon form of A
lav_mat_rref <- function(a, tol = sqrt(.Machine$double.eps)) {
  # MATLAB documentation says rref uses: tol = (max(size(A))*eps *norm(A,inf)
  if (missing(tol)) {
    a_norm <- max(rowSums(abs(a)))
    tol <- max(dim(a)) * a_norm * .Machine$double.eps
  }

  # check if A is a matrix
  stopifnot(is.matrix(a))

  # dimensions
  n_row <- NROW(a)
  n_col <- NCOL(a)
  pivot <- integer(min(n_row, n_col))
  pivot_n <- 0L

  # catch empty matrix
  if (n_row == 0 && n_col == 0) {
    return(matrix(0, 0, 0))
  }

  row_index <- col_index <- 1
  while (row_index <= n_row &&
    col_index <= n_col) {
    # look for largest (in absolute value) element in this column:
    i_below <- which.max(abs(a[row_index:n_row, col_index]))
    i <- i_below + row_index - 1L
    p <- a[i, col_index]

    # check if column is empty
    if (abs(p) <= tol) {
      a[row_index:n_row, col_index] <- 0L # clean up
      col_index <- col_index + 1
    } else {
      # store pivot column
      pivot_n <- pivot_n + 1L
      pivot[pivot_n] <- col_index

      # do we need to swap column?
      if (row_index != i) {
        a[c(row_index, i), col_index:n_col] <-
          a[c(i, row_index), col_index:n_col]
      }

      # scale pivot to be 1.0
      a[row_index, col_index:n_col] <- a[row_index, col_index:n_col] / p

      # create zeroes below and above pivot
      other <- seq_len(n_row)[-row_index]
      a[other, col_index:n_col] <-
        a[other, col_index:n_col] - tcrossprod(
          a[other, col_index],
          a[row_index, col_index:n_col]
        )
      # next row/col
      row_index <- row_index + 1
      col_index <- col_index + 1
    }
  }

  # rounding?

  list(R = a, pivot = pivot[seq_len(pivot_n)])
}


# inverse of a non-singular (not necessarily positive-definite) symmetric matrix
# FIXME: error handling?
lav_mat_sym_inverse <- function(s, logdet = FALSE,
                                         sinv_method = "eigen",
                                         zero_warn = FALSE) {
  # catch zero cols/rows (S is symmetric, so zero row iff zero col)
  zero_idx <- which(rowSums(abs(s)) == 0)
  s_orig <- s
  if (length(zero_idx) > 0L) {
    if (zero_warn) {
      lav_msg_warn(gettext("matrix to be inverted contains zero cols/rows"))
    }
    s <- s[-zero_idx, -zero_idx, drop = FALSE]
  }

  p <- NCOL(s)

  if (p == 0L) {
    s_inv <- matrix(0, 0, 0)
    if (logdet) {
      attr(s_inv, "logdet") <- 0
    }
    return(s_inv)
  } else if (p == 1L) {
    tmp <- s[1, 1]
    s_inv <- matrix(1 / tmp, 1, 1)
    if (logdet) {
      if (tmp > 0) {
        attr(s_inv, "logdet") <- log(tmp)
      } else {
        attr(s_inv, "logdet") <- -Inf
      }
    }
  } else if (p == 2L) {
    a11 <- s[1, 1]
    a12 <- s[1, 2]
    a21 <- s[2, 1]
    a22 <- s[2, 2]
    tmp <- a11 * a22 - a12 * a21
    if (tmp == 0) {
      s_inv <- matrix(c(Inf, Inf, Inf, Inf), 2, 2)
      if (logdet) {
        attr(s_inv, "logdet") <- -Inf
      }
    } else {
      s_inv <- matrix(c(a22 / tmp, -a21 / tmp, -a12 / tmp, a11 / tmp), 2, 2)
      if (logdet) {
        if (tmp > 0) {
          attr(s_inv, "logdet") <- log(tmp)
        } else {
          attr(s_inv, "logdet") <- -Inf
        }
      }
    }
  } else if (sinv_method == "eigen") {
    ev_1 <- eigen(s, symmetric = TRUE)
    # V %*% diag(1/d) %*% V^{-1}, where V^{-1} = V^T
    s_inv <-
      tcrossprod(
        ev_1$vectors / rep(ev_1$values, each = length(ev_1$values)),
        ev_1$vectors
      )

    # 0.5 version
    # S.inv <- tcrossprod(sweep(EV$vectors, 2L,
    #                          STATS = (1/EV$values), FUN="*"), EV$vectors)

    if (logdet) {
      if (all(ev_1$values >= 0)) {
        attr(s_inv, "logdet") <- sum(log(ev_1$values))
      } else {
        attr(s_inv, "logdet") <- as.numeric(NA)
      }
    }
  } else if (sinv_method == "solve") {
    s_inv <- solve.default(s)
    if (logdet) {
      ev <- eigen(s, symmetric = TRUE, only.values = TRUE)
      if (all(ev$values >= 0)) {
        attr(s_inv, "logdet") <- sum(log(ev$values))
      } else {
        attr(s_inv, "logdet") <- as.numeric(NA)
      }
    }
  } else if (sinv_method == "chol") {
    # this will break if S is not positive definite
    c_s <- chol.default(s)
    s_inv <- chol2inv(c_s)
    if (logdet) {
      diag_c_s <- diag(c_s)
      attr(s_inv, "logdet") <- sum(log(diag_c_s * diag_c_s))
    }
  } else {
    lav_msg_stop(gettext("method must be either `eigen', `solve' or `chol'"))
  }

  if (length(zero_idx) > 0L) {
    logdet <- attr(s_inv, "logdet")
    tmp <- s_orig
    tmp[-zero_idx, -zero_idx] <- s_inv
    s_inv <- tmp
    attr(s_inv, "logdet") <- logdet
    attr(s_inv, "zero.idx") <- zero_idx
  }

  s_inv
}

# solve(A) %*% B, where A is symmetric, and B is vector of matrix
lav_mat_sym_solve_spd <- function(a, m_b, tol = 1e-10) {
  chol_a <- try(chol(a), silent = TRUE)
  if (!inherits(chol_a, "try-error")) {
    # Cholesky solve
    backsolve(chol_a, forwardsolve(t(chol_a), m_b))
  } else {
    # Eigen fallback (pseudo-inverse)
    eig <- eigen(a, symmetric = TRUE)
    keep <- eig$values > tol * max(eig$values)
    ainv <- eig$vectors[, keep, drop = FALSE] %*%
      diag(1 / eig$values[keep], length(eig$values[keep])) %*%
      t(eig$vectors[, keep, drop = FALSE])
    ainv %*% m_b
  }
}

# update inverse of S, after removing 1 or more rows (and corresponding
# colums) from S, a symmetric matrix
#
# - only removal for now!
#
lav_mat_sym_inverse_update <- function(s_inv, rm_idx = integer(0L),
                                                logdet = FALSE,
                                                s_logdet = NULL) {
  ndel <- length(rm_idx)

  if (ndel == 0L) {
    out <- s_inv
    if (logdet) {
      attr(out, "logdet") <- s_logdet
    }

  # rank-1 update
  } else if (ndel == 1L) {
    h <- s_inv[rm_idx, rm_idx]
    if (h > 0) {
      a <- s_inv[-rm_idx, rm_idx, drop = FALSE] / sqrt(h)
      out <- s_inv[-rm_idx, -rm_idx, drop = FALSE] - tcrossprod(a)
      if (logdet) {
        attr(out, "logdet") <- s_logdet + log(h)
      }
    } else {
      out <- s_inv[-rm_idx, -rm_idx, drop = FALSE]
      if (logdet) {
        attr(out, "logdet") <- as.numeric(NA)
      }
    }
  }

  # rank-n update
  else if (ndel < NCOL(s_inv)) {
    a_1 <- s_inv[rm_idx, -rm_idx, drop = FALSE]
    h_1 <- s_inv[rm_idx, rm_idx, drop = FALSE]
    out <- (s_inv[-rm_idx, -rm_idx, drop = FALSE] -
      crossprod(a_1, solve.default(h_1, a_1)))
    if (logdet) {
      c_h <- try(chol.default(h_1), silent = TRUE)
      if (!inherits(c_h, "try-error")) {
        diag_c_h <- diag(c_h)
        h_logdet <- sum(log(diag_c_h * diag_c_h))
      } else {
        ev <- eigen(h_1, symmetric = TRUE, only.values = TRUE)$values
        h_logdet <- sum(log(abs(ev)))
      }
      attr(out, "logdet") <- s_logdet + h_logdet
    }

    # erase all col/rows...
  } else if (ndel == NCOL(s_inv)) {
    out <- matrix(0, 0, 0)
  } else {
    lav_msg_stop(gettext("column indices exceed number of columns in S.inv"))
  }

  out
}

# compute `lambda': the smallest root of the determinantal equation
# |M - lambda*P| = 0 (see Fuller 1987, p.125 or p.172
#
# the function allows for zero rows/columns in P, by regressing them out
# this approach was suggested to me by Wayne A. Fuller, personal communication,
# 12 Nov 2020
#
# YR 12 June 2026: P no longer needs to be strictly positive definite: any
# rank deficiency of P that is not visible as zero rows/columns (a zero
# eigenvalue with a nonzero diagonal) is now handled in the same way, by
# regressing out the null space of P in its eigenbasis: writing
# P = V diag(d1, 0) t(V) and Mt = t(V) M V (partitioned accordingly),
#   |M - lambda*P| = |Mt22| * |Mt11 - Mt12 solve(Mt22) Mt21 - lambda*diag(d1)|
# so the finite roots are the eigenvalues of the Schur complement pencil;
# the null space directions of P produce no (finite) roots
#
# if P is (numerically) zero, there are no finite roots, and we return +Inf
# (the callers test 'lambda < cutoff', and with P = 0 the correction term
# vanishes anyway)
lav_mat_sym_diff_smallest_root <- function(m = NULL, p = NULL) {
  # check input (we will 'assume' they are square and symmetric)
  stopifnot(is.matrix(m), is.matrix(p))

  # check if P is diagonal or not
  pdiag_flag <- FALSE
  tmp <- p
  diag(tmp) <- 0
  if (all(abs(tmp) < sqrt(.Machine$double.eps))) {
    pdiag_flag <- TRUE
  }

  # diagonal elements of P
  n_p <- nrow(p)
  diag_p <- p[lav_mat_diag_idx(n_p)]

  # force diagonal elements of P to be nonnegative (warn?)
  neg_idx <- which(diag_p < 0)
  if (length(neg_idx) > 0L) {
    lav_msg_warn(gettext(
      "some diagonal elements of the matrix that is subtracted are negative (and set to zero)"
    ))
    diag(p)[neg_idx] <- diag_p[neg_idx] <- 0
  }

  # check for (near)zero diagonal elements
  zero_idx <- which(abs(diag_p) < sqrt(.Machine$double.eps))

  # P is (numerically) zero: no finite roots
  if (length(zero_idx) == n_p) {
    return(+Inf)
  }

  # regress out the M-block corresponding to zero diagonal elements in P
  # (for a positive semidefinite P, a zero diagonal element implies a zero
  # row/column, i.e. a null space direction of P)
  if (length(zero_idx) > 0L) {
    m_pp <- m[-zero_idx, -zero_idx, drop = FALSE]
    m_pn <- m[-zero_idx, zero_idx, drop = FALSE]
    m_np <- m[zero_idx, -zero_idx, drop = FALSE]
    m_nn <- m[zero_idx, zero_idx, drop = FALSE]

    m <- m_pp - m_pn %*% lav_mat_sym_solve_spd(m_nn, m_np)
    p <- p[-zero_idx, -zero_idx, drop = FALSE]
  }

  # compute the smallest root of |M - lambda*P| = 0
  if (pdiag_flag) {
    # diagonal P with strictly positive diagonal elements
    ldiag <- 1 / sqrt(diag(p))
    lml <- t(ldiag * m) * ldiag
    lambda <- eigen(lml, symmetric = TRUE, only.values = TRUE)$values[nrow(p)]
  } else {
    eig_p <- eigen(p, symmetric = TRUE)
    ev <- eig_p$values

    # (numerical) rank of P; eigenvalues below the tolerance (including
    # any negative ones) span the null space of P
    ev_tol <- sqrt(.Machine$double.eps) * max(ev[1], 0)
    rank_p <- sum(ev > ev_tol)
    if (rank_p == 0L) {
      # no positive eigenvalues left: no finite roots
      return(+Inf)
    }
    if (any(ev < -1000 * ev_tol)) {
      lav_msg_warn(gettext(
        "matrix to subtract is not positive semidefinite; using its positive part only"))
    }

    vt_mv <- crossprod(eig_p$vectors, m %*% eig_p$vectors)
    if (rank_p < nrow(p)) {
      # hidden rank deficiency: regress out the null space of P
      pos_idx <- seq_len(rank_p)
      m_pp <- vt_mv[pos_idx, pos_idx, drop = FALSE]
      m_pn <- vt_mv[pos_idx, -pos_idx, drop = FALSE]
      m_np <- vt_mv[-pos_idx, pos_idx, drop = FALSE]
      m_nn <- vt_mv[-pos_idx, -pos_idx, drop = FALSE]
      vt_mv <- m_pp - m_pn %*% lav_mat_sym_solve_spd(m_nn, m_np)
    }

    d_p <- ev[seq_len(rank_p)]
    lml <- t(vt_mv / sqrt(d_p)) / sqrt(d_p)
    lambda <- eigen(lml, symmetric = TRUE, only.values = TRUE)$values[rank_p]
  }

  lambda
}

# force a symmetric matrix to be positive definite
# simple textbook version (see Matrix::nearPD for a more sophisticated version)
#
lav_mat_sym_force_pd <- function(s, tol = 1e-06) {
  if (ncol(s) == 1L) {
    return(matrix(max(s[1, 1], tol), 1L, 1L))
  }

  # eigen decomposition
  s_eigen <- eigen(s, symmetric = TRUE)

  # eigen values
  ev <- s_eigen$values

  # replace small/negative eigen values
  ev[ev / abs(ev[1]) < tol] <- tol * abs(ev[1])

  # reconstruct (scale columns of V, avoid n*n diagonal alloc)
  n <- ncol(s_eigen$vectors)
  out <- tcrossprod(s_eigen$vectors * rep(ev, each = n), s_eigen$vectors)

  out
}

# compute sample covariance matrix, divided by 'N' (not N-1, as in cov)
#
# mu is not supposed to be ybar, but close
# if provided, we compute S as 1/N*crossprod(y - mu) instead of
#                              1/N*crossprod(y - ybar)
lav_mat_cov <- function(y, mu = NULL) {
  n <- NROW(y)
  s1 <- stats::cov(y) # uses a corrected two-pass algorithm
  s <- s1 * (n - 1) / n

  # mu?
  if (!is.null(mu)) {
    p <- NCOL(y)
    ybar <- base::.colMeans(y, m = n, n = p)
    s <- s + tcrossprod(ybar - mu)
  }

  s
}

# transform a matrix to match a given target mean/covariance
lav_mat_transform_mean_cov <- function(y,
                                          target_mean = numeric(NCOL(y)),
                                          target_cov = diag(NCOL(y))) {
  # coerce to matrix
  y <- as.matrix.default(y)

  # convert to vector
  target_mean <- as.vector(target_mean)

  s <- lav_mat_cov(y)
  s_inv <- lav_mat_sym_inverse(s)
  s_inv_sqrt <- lav_mat_sym_sqrt(s_inv)
  target_cov_sqrt <- lav_mat_sym_sqrt(target_cov)

  # transform cov
  x <- y %*% s_inv_sqrt %*% target_cov_sqrt

  # shift mean
  xbar <- colMeans(x)
  x <- t(t(x) - xbar + target_mean)

  x
}

# weighted column variances
#
# for each column in y: var = sum(wt * (y - w.mean(y))^2) / N
#
# where N = sum(wt) - 1 (method = "unbiased") assuming wt are frequency weights
#   or  N = sum(wt)     (method = "ML")
#
# Note: another approach (when the weights are 'reliability weights' is to
#       use N = sum(wt) - sum(wt^2)/sum(wt) (not implemented here)
#
# if we have missing values, we use only the observations and weights
# that are NOT missing
#
lav_mat_var_wt <- function(y, wt = NULL, method = c("unbiased", "ML")) {
  y <- unname(as.matrix.default(y))
  dim_1 <- dim(y)

  if (is.null(wt)) {
    wt <- rep(1, nrow(y))
  }

  if (anyNA(y)) {
    wt_1 <- wt * !is.na(y)
    w_n <- .colSums(wt_1, m = dim_1[1], n = dim_1[2])
    w_mean <- .colSums(wt * y, m = dim_1[1], n = dim_1[2], na.rm = TRUE) / w_n
    ytc <- t(t(y) - w_mean)
    tmp <- .colSums(wt * ytc * ytc, m = dim_1[1], n = dim_1[2], na.rm = TRUE)
    out <- switch(match.arg(method),
      unbiased = tmp / (w_n - 1),
      ML = tmp / w_n
    )
  } else {
    w_mean <- .colSums(wt * y, m = dim_1[1], n = dim_1[2]) / sum(wt)
    ytc <- t(t(y) - w_mean)
    tmp <- .colSums(wt * ytc * ytc, m = dim_1[1], n = dim_1[2])
    out <- switch(match.arg(method),
      unbiased = tmp / (sum(wt) - 1),
      ML = tmp / sum(wt)
    )
  }

  out
}

# compute (I-A)^{-1} where A is square
# using a (truncated) Neumann series:  (I-A)^{-1} = \sum_k=0^{\infty} A^k
#                                                 = I + A + A^2 + A^3 + ...
#
# note: this only works if the largest eigenvalue for A is < 1; but if A
#       represents regressions, the diagonal will be zero, and all eigenvalues
#       are zero
#
# as A is typically sparse, we can stop if all elements in A^k are zero for,
# say, k<=6
lav_mat_inverse_iminus <- function(a = NULL) {
  nr <- nrow(a)
  nc <- ncol(a)
  stopifnot(nr == nc)

  # create I + A
  ia <- a
  diag_idx <- lav_mat_diag_idx(nr)
  ia[diag_idx] <- ia[diag_idx] + 1

  # initial approximation
  ia_inv <- ia

  # first order
  a2 <- a %*% a
  if (all(a2 == 0)) {
    # we are done
    return(ia_inv)
  } else {
    ia_inv <- ia_inv + a2
  }

  # second order
  a3 <- a2 %*% a
  if (all(a3 == 0)) {
    # we are done
    return(ia_inv)
  } else {
    ia_inv <- ia_inv + a3
  }

  # third order
  a4 <- a3 %*% a
  if (all(a4 == 0)) {
    # we are done
    return(ia_inv)
  } else {
    ia_inv <- ia_inv + a4
  }

  # fourth order
  a5 <- a4 %*% a
  if (all(a5 == 0)) {
    # we are done
    return(ia_inv)
  } else {
    ia_inv <- ia_inv + a5
  }

  # fifth order
  a6 <- a5 %*% a
  if (all(a6 == 0)) {
    # we are done
    ia_inv
  } else {
    # naive version (for now)
    tmp <- -a
    tmp[diag_idx] <- tmp[diag_idx] + 1
    ia_inv <- try(solve(tmp), silent = TRUE)
    if (inherits(ia_inv, "try-error")) {
      lav_msg_warn(gettext("(I - A) is singular; using generalized inverse."))
      ia_inv <- MASS::ginv(tmp)
    }
    ia_inv
  }
}

# let Gamma_NT = 2 * D.plus %*% (S %x% S) %*% t(D.plus), where
# D.plus is the elimination matrix: D.plus = ginv(D)
# we wish to compute R = K %*% Gamma_NT %*% t(K)
# but without computing the kronecker product explicitly
# - we allow for meanstructure TRUE/FALSE
# - we allow for fixed.x = TRUE, if x.idx is not empty
# - but NOT for conditional.x = TRUE (for now)
lav_mat_k_gammant_kt <- function(m_k, s, meanstructure = FALSE,
                                    x_idx = integer(0L)) {
  # dimensions of S
  nvar <- nrow(s)
  pstar <- nvar * (nvar + 1L) / 2L

  # dimension output
  q <- nrow(m_k)

  # sanity checks
  expected_q <- if (meanstructure) nvar + pstar else pstar
  if (ncol(m_k) != expected_q) {
    lav_msg_stop(gettextf(
      "ncol(K) must be %d when meanstructure = %s",
      expected_q, toupper(meanstructure)
    ))
  }
  x_idx <- as.integer(x_idx)
  if (length(x_idx) > 0L && any(x_idx < 1L | x_idx > nvar)) {
    lav_msg_stop(gettext("x.idx contains indices outside 1:nvar."))
  }

  y_idx <- setdiff(seq_len(nvar), x_idx)
  nvar_y <- length(y_idx)
  nvar_x <- length(x_idx)
  if (nvar_y == 0L) {
    lav_msg_stop(gettextf("No random variables remain after removing x.idx."))
  }

  # Cholesky of S (no check: we assume S is pd)
  l_s <- t(chol(s))

  # fixed.x: conditional covariance YbarX and low-rank factor Q for R = Q Q'
  if (nvar_x > 0L) {
    a_yy <- s[y_idx, y_idx, drop = FALSE]
    b_yx <- s[y_idx, x_idx, drop = FALSE]
    c_xx <- s[x_idx, x_idx, drop = FALSE]
    l_x <- t(chol(c_xx))

    bl <- forwardsolve(l_x, t(b_yx))
    ybar_x <- a_yy - t(bl) %*% bl

    f_full <- matrix(0, nvar, nvar_x)
    f_full[y_idx, ] <- b_yx
    f_full[x_idx, ] <- c_xx
    q_1 <- t(forwardsolve(l_x, t(f_full)))
  } else {
    ybar_x <- s
    q_1 <- matrix(0, nvar, 0L)
  }
  l_ybar_x <- t(chol(ybar_x))

  # build W (d^2 x p) with col j = vec(L' N_j L)
  # B2 is q x pstar (rows are vech vectors); L is nvar x d
  build_w <- function(b2, l) {
    d <- ncol(l)
    if (d == 0L) {
      return(matrix(0, 0L, q))
    }

    # expand vech row j of B2 into nvar x nvar_q block matrix V_mat
    v_mat <- matrix(0, nvar, nvar * q)
    for (j in seq_len(q)) {
      m <- lav_mat_vech_rev(b2[j, ])
      diag(m) <- diag(m) * 2
      v_mat[, (j - 1L) * nvar + seq_len(nvar)] <- m / 2
    }
    # batch compute t(L) N_j L via two (d x nvar_q) multiplies + aperm trick
    lt_v <- t(l) %*% v_mat
    lt_v3 <- array(lt_v, c(d, nvar, q))
    # aperm transposes each block: since M_i symmetric, t(L' M_j) = M_j L
    lt_vt <- matrix(aperm(lt_v3, c(2L, 1L, 3L)), nvar, d * q)
    matrix(t(l) %*% lt_vt, d^2, q)
  }

  # assemble result
  k2 <- if (meanstructure) {
    m_k[, nvar + seq_len(pstar), drop = FALSE]
  } else {
    m_k
  }
  w_s <- build_w(k2, l_s)
  w_r <- build_w(k2, q_1)
  out <- 2 * (crossprod(w_s) - crossprod(w_r))

  if (meanstructure) {
    k1 <- m_k[, seq_len(nvar), drop = FALSE]
    z <- k1[, y_idx, drop = FALSE] %*% l_ybar_x
    t1 <- tcrossprod(z)
    out <- out + t1
  }

  out
}

# LDW 11/4/24 : overwrite defaults depending on mimic in separate function
#
lav_options_mimic <- function(opt) {
  mlr_test <- "yuan.bentler.mplus" # for now
  if (opt$gls.v11.mplus == "default") {
    opt$gls.v11.mplus <- (opt$mimic == "Mplus")
  }
  if (opt$gamma.vcov.mplus == "default") {
    opt$gamma.vcov.mplus <- (opt$mimic == "Mplus")
  }
  if (opt$gamma.wls.mplus == "default") {
    opt$gamma.wls.mplus <- (opt$mimic == "Mplus")
  }
  if (opt$information.expected.mplus == "default") {
    opt$information.expected.mplus <- (opt$mimic == "Mplus")
  }
  if (opt$rmsea.scaled.mplus == "default") {
    opt$rmsea.scaled.mplus <- (opt$mimic == "Mplus")
  }
  if (opt$mimic == "lavaan") {
    if (is.character(opt$conditional.x)) { # = "default"
      if (lav_options_estimatorgroup(opt$estimator) == "ML") {
        opt$conditional.x <- FALSE
      }
    }
    if (opt$fixed.x == "default") {
      if (any(lav_options_estimatorgroup(opt$estimator) ==
                                                     c("MML", "ML", "IV")) &&
          is.character(opt$start) && opt$start != "simple") { # new in 0.6-12
        opt$fixed.x <- TRUE
      }
    }
    if (is.character(opt$zero.keep.margins)) { # = "default"
      opt$zero.keep.margins <- TRUE
    }
  } else if (opt$mimic == "Mplus") {
    if (length(opt$group.equal) == 0L || all(nchar(opt$group.equal) == 0L)) {
      if (opt$.categorical) {
        opt$group.equal <- c("loadings", "thresholds")
      } else {
        if (is.logical(opt$meanstructure) && !opt$meanstructure) {
          opt$group.equal <- "loadings"
        } else {
          opt$group.equal <- c("loadings", "intercepts")
        }
      }
    }
    if (opt$missing == "default") {
      if (!opt$.categorical && any(opt$estimator == c("ml", "mlr"))) {
        # since version 5?
        opt$missing <- "ml"
        # check later if this is ok
      }
    }
    if (opt$estimator != "pml") {
      if (opt$meanstructure == "default") opt$meanstructure <- TRUE
    }
    if (opt$estimator == "mlr") mlr_test <- "yuan.bentler.mplus"
    if (any(lav_options_estimatorgroup(opt$estimator) ==
            c("ML", "REML", "NTRLS", "catML"))) {
      }
    if (is.character(opt$conditional.x)) { # = "default"
      if (lav_options_estimatorgroup(opt$estimator) == "ML") {
        opt$conditional.x <- FALSE
      }
    }
    if (opt$fixed.x == "default") {
      if (any(lav_options_estimatorgroup(opt$estimator) == c("MML", "ML")) &&
          is.character(opt$start) && opt$start != "simple") { # new in 0.6-12
        opt$fixed.x <- TRUE
      }
    }
    if (is.character(opt$zero.keep.margins)) { # = "default"
      opt$zero.keep.margins <- TRUE
    }
    opt$baseline.conditional.x.free.slopes <- FALSE
  } else if (opt$mimic == "EQS") {
    opt$baseline.fixed.x.free.cov <- FALSE
    if (opt$estimator == "mlr") mlr_test <- "yuan.bentler"
    if (any(lav_options_estimatorgroup(opt$estimator) ==
            c("ML", "REML", "NTRLS", "catML"))) {
      if (opt$likelihood == "default") opt$likelihood <- "wishart"
    }
  } else if (opt$mimic == "LISREL") {
    opt$baseline.fixed.x.free.cov <- FALSE
    if (any(lav_options_estimatorgroup(opt$estimator) ==
            c("ML", "REML", "NTRLS", "catML"))) {
      if (opt$likelihood == "default") opt$likelihood <- "wishart"
    }
  }
  if (opt$estimator == "mlr") {
    if (opt$test[1] == "default") {
      opt$test <- mlr_test
    } else if (!any(opt$test == "none")) {
      # respect test = "none" (as the MLM/MLMV/ULSM/... blocks do); adding
      # the MLR-implied scaled test would produce the invalid c("<scaled>",
      # "none") combination (eg sam() fits its internal dummy model with
      # test = "none")
      opt$test <- union(mlr_test, opt$test)
    }
  }
  opt
}

# in separate file LDW 06/04/2024

# rename names of se values, and check for invalid values
lav_options_check_se <- function(opt = NULL) {
  # unlike test=, se= should be a single character string
  if (length(opt$se) > 1L) opt$se <- opt$se[1]

  # backwards compatibility (0.4 -> 0.5)
  if (opt$se == "first.order") {
    opt$se <- "standard"
    opt$information[1] <- "first.order"
    if (length(opt$information) > 1L &&
        opt$information[2] == "default") {
      opt$information[2] <- "first.order"
    }
  } else if (opt$se == "observed") {
    opt$se <- "standard"
    opt$information[1] <- "observed"
    if (length(opt$information) > 1L &&
        opt$information[2] == "default") {
      opt$information[2] <- "observed"
    }
  } else if (opt$se == "expected") {
    opt$se <- "standard"
    opt$information[1] <- "expected"
    if (length(opt$information) > 1L &&
        opt$information[2] == "default") {
      opt$information[2] <- "expected"
    }
  }

  # handle generic 'robust' (except clustered/multilevel)
  # else if(opt$se == "robust" && !opt$.clustered && !opt$.multilevel) {
  #    if(opt$missing %in% c("ml", "ml.x")) {
  #        opt$se <- "robust.huber.white"
  #    } else if(opt$missing == "two.stage") {
  #        opt$se <- "two.stage"
  #    } else if(opt$missing == "robust.two.stage") {
  #        opt$se <- "robust.two.stage"
  #    } else {
  #        # depends on estimator!
  #        opt$se <- "robust.sem"
  #    }
  # }

  # GLS, NTRLS, FML, UMN
  ok_flag <- TRUE
  if (any(opt$estimator == c("gls", "ntrls", "fml"))) {
    ok_flag <- any(opt$se == c(
      "default", "none", "standard",
      "bootstrap", "external",
      "robust", "robust.sem"
    ))
  # WLS, DLS, DWLS, WLSM, WLSMV, WLSMVS, ULS, ULSM, ULSMV, ULSMVS
  } else if (any(opt$estimator == c(
    "wls", "dls",
    "dwls", "wlsm", "wlsmv", "wlsmvs",
    "uls", "ulsm", "ulsmv", "ulsmvs"
  ))) {
    ok_flag <- any(opt$se == c(
      "default", "none", "standard",
      "bootstrap", "external",
      "robust", "robust.sem", "robust.sem.nt"
    ))
  # PML
  } else if (opt$estimator == "pml") {
    ok_flag <- any(opt$se == c(
      "default", "none", "standard",
      "bootstrap", "external",
      "robust.huber.white"
    ))
  # FABIN, GUTTMAN1952, BENTLER1982, ...
  } else if (any(opt$estimator == c(
    "fabin2", "fabin3", "mgm"))) {
    ok_flag <- any(opt$se == c("default", "none", "bootstrap", "external"))
  # OTHERS
  } else if (any(opt$estimator == c("fml", "mml", "reml"))) {
    ok_flag <- any(opt$se == c("default", "none", "standard", "external"))
  }

  if (!ok_flag) {
    lav_msg_stop(gettextf(
      "invalid value (%1$s) in se= argument for estimator %2$s.",
      opt$se, toupper(opt$estimator)))
  }

  opt
}

lav_step00_parameters <- function(matchcall = NULL,
                                  syscall = NULL,
                                  dotdotdot = NULL) {
  # 1. resolve a problem where parameter 'cl' is matched to 'cluster'
  #    and shouldn't
  # 2. apply default options for cfa/sem/growth functions
  # 3. if dotdotdot$control present, copy to dotdotdot$... for
  #   optim.method,
  #   optim.force.converged,
  #   gradient -> optim.gradient !!! overwritten by dotdotdot$gradient
  #     if present
  #   init_nelder_mead -> optim.init_nelder_mead

  mc <- matchcall
  sc <- syscall
  ddd <- dotdotdot

  # catch partial matching of 'cl' (expanded to cluster)
  if (!is.null(sc[["cl"]]) &&
    is.null(sc[["cluster"]]) &&
    !is.null(mc[["cluster"]])) {
    mc[["cl"]] <- mc[["cluster"]]
    mc[["cluster"]] <- NULL
    ddd$cl <- sc[["cl"]]
  }
  if (!is.null(mc$cluster)) mc$cluster <- eval(mc$cluster, parent.frame(2))

  # default options
  if (any(ddd$model.type == c("sem", "cfa", "growth"))) {
    # default options for sem/cfa or growth
    defaults <- list(
      int.ov.free     = ddd$model.type != "growth",
      int.lv.free     = ddd$model.type == "growth",
      auto.fix.first  = TRUE, # (re)set in lav_options_set
      auto.fix.single = TRUE,
      auto.var        = TRUE,
      auto.cov.lv.x   = TRUE,
      auto.cov.y      = TRUE,
      auto.th         = TRUE,
      auto.delta      = TRUE,
      auto.efa        = TRUE
    )
    for (dflt_i in seq_along(defaults)) {
      argname <- names(defaults)[dflt_i]
      if (is.null(mc[[argname]])) {
        mc[[argname]] <- defaults[[dflt_i]]
        ddd[[argname]] <- defaults[[dflt_i]]
      }
    }
  }

  # backwards compatibility, control= argument (<0.5-23)
  if (!is.null(ddd$control)) {
    # optim.method
    if (!is.null(ddd$control$optim.method)) {
      ddd$optim.method <- ddd$control$optim.method
    }
    # cor.optim.method
    if (!is.null(ddd$control$cor.optim.method)) {
      # ignore it silently
    }
    # control$optim.force.converged
    if (!is.null(ddd$control$optim.force.converged)) {
      ddd$optim.force.converged <- ddd$control$optim.force.converged
    }
    # gradient
    if (!is.null(ddd$control$gradient)) {
      ddd$optim.gradient <- ddd$control$gradient
    }
    if (!is.null(ddd$gradient)) {
      ddd$optim.gradient <- ddd$gradient
    }
    # init_nelder_mead
    if (!is.null(ddd$control$init_nelder_mead)) {
      ddd$optim.init_nelder_mead <- ddd$control$init_nelder_mead
    }
  }

  list(mc = mc, dotdotdot = ddd)
}

lav_step00_checkdata <- function(data = NULL,
                                        dotdotdot = NULL,
                                        sample_cov = NULL,
                                        sample_nobs = NULL,
                                        sample_mean = NULL,
                                        sample_th = NULL,
                                        nacov = NULL,
                                        wls_v = NULL,
                                        ov_order = NULL) {
  # if data not NULL:
  #   if it is an 'enriched' data.frame (e.g. a tibble), simplify to an
  #   ordinary data.frame
  #   if class is 'lavMoments':
  #         check if it contains sample.cov and sample.nobs (***error*** if not)
  #         if sample.mean, sample.th, NACOV and/or WLS.V present,
  #                   copy to corresponding arguments of the function
  #         if lavOptions present in data, copy those that are not provided
  #         in the function call to dotdotdot$...
  #             set data to NULL
  #   if it is a function --> ***error***
  #   TODO: other tests are present in lav_lavdata(), should we copy them here?
  # if nacov or wls_v not NULL, set ov_order to "data"

  if (!is.null(data)) {
    if (inherits(data, "data.frame")) {
      # just in case it is not a traditional data.frame
      data <- as.data.frame(data)
    } else if (inherits(data, "lavMoments")) {
      # This object must contain summary statistics
      # e.g., created by lavaan.mi::poolSat

      # set required-data arguments
      if ("sample.cov" %in% names(data)) {
        sample_cov <- data$sample.cov
      } else {
        lav_msg_stop(gettext(
          "When data= is of class lavMoments, it must contain sample.cov"))
      }

      if ("sample.nobs" %in% names(data)) {
        sample_nobs <- data$sample.nobs
      } else {
        lav_msg_stop(gettext(
          "When data= is of class lavMoments, it must contain sample.nobs"))
      }

      # check for optional-data arguments
      if ("sample.mean" %in% names(data)) sample_mean <- data$sample.mean
      if ("sample.th" %in% names(data)) sample_th <- data$sample.th
      if ("NACOV" %in% names(data)) nacov <- data$NACOV
      if ("WLS.V" %in% names(data)) wls_v <- data$WLS.V

      # set other args not included in dotdotdot
      if (length(data$lavOptions)) {
        newdots <- setdiff(names(data$lavOptions), names(dotdotdot))
        if (length(newdots)) {
          for (dd in newdots) dotdotdot[[dd]] <- data$lavOptions[[dd]]
        }
      }

      # FIXME: Should WLS.V be an I(dentity) matrix when ULS is requested?
      #       Unused for point estimates, but still used to scale/shift test
      # if (!is.null(dotdotdot$estimator)) {
      #   if (grepl(pattern = "ULS", x = toupper(dotdotdot$estimator[1L])) &&
      #       !is.null(WLS.V)) {
      #     #  set to diagonal
      #     if (is.list(WLS.V)) {
      #       WLS.V <- lapply(WLS.V, function(w) {diag(w) <- 1 ; return(w) })
      #     } else diag(WLS.V) <- 1
      #   }
      # }

      # get rid of data= argument
      data <- NULL
    }

    if (is.function(data)) {
      lav_msg_stop(gettext("data is a function; it should be a data.frame"))
    }
  }
  # new in 0.6-14: if NACOV and/or WLS.V are provided, we force
  # ov_order="data" for now
  # until we have reliable code to re-arrange/select col/rows for
  # of NACOV/WLS.V based on the model-based ov.names
  if (!is.null(nacov) || !is.null(wls_v)) {
    if (ov_order != "force.model") { # used by sam()
      ov_order <- "data"
    } else {
      ov_order <- "model"
    }
  }

  list(
    data = data, dotdotdot = dotdotdot, sample_cov = sample_cov,
    sample_nobs = sample_nobs, sample_mean = sample_mean,
    sample_th = sample_th, nacov = nacov, wls_v = wls_v, ov_order = ov_order
  )
}

# shared computational kernels for the lav_mvnorm_* / lav_mvreg_* family
#
# these helpers centralize small blocks of code that used to be copy-pasted
# across lav_mvnorm.R, lav_mvnorm_h1.R, lav_mvnorm_missing.R,
# lav_mvnorm_missing_h1.R, lav_mvreg.R (and, for some, the cluster files)
#
# YR/CC 03 July 2026: first version (refactoring; no change in behavior)

# 1) ensure we have Sigma^{-1} (optionally with a "logdet" attribute)
#
#    - if sigma_inv is NULL, invert sigma_1 using lav_mat_sym_inverse()
#    - if sigma_inv is given but the "logdet" attribute is missing (and
#      logdet = TRUE), compute -log|Sigma^{-1}| from the eigenvalues
lav_mvn_sigma_inv <- function(sigma_1 = NULL,
                              sigma_inv = NULL,
                              sinv_method = "eigen",
                              logdet = FALSE) {
  if (is.null(sigma_inv)) {
    sigma_inv <- lav_mat_sym_inverse(
      s = sigma_1, logdet = logdet,
      sinv_method = sinv_method
    )
  } else if (logdet && is.null(attr(sigma_inv, "logdet"))) {
    # compute - ln|Sigma.inv|
    ev <- eigen(sigma_inv, symmetric = TRUE, only.values = TRUE)
    attr(sigma_inv, "logdet") <- -1 * sum(log(ev$values))
  }

  sigma_inv
}

# 2) total number of observations: sum(wt), or the number of rows of y
lav_mvn_nobs <- function(y = NULL, wt = NULL) {
  if (!is.null(wt)) {
    sum(wt)
  } else {
    NROW(y)
  }
}

# 2b) ML sample statistics (mean + cov) from raw data, weighted or not
lav_mvn_samp_stats <- function(y = NULL, wt = NULL) {
  if (!is.null(wt)) {
    out <- stats::cov.wt(y, wt = wt, method = "ML")
    list(mean = out$center, cov = out$cov)
  } else {
    list(
      mean = base::.colMeans(y, m = NROW(y), n = NCOL(y)),
      cov = lav_mat_cov(y)
    )
  }
}

# 3) W.tilde = S + (ybar - mu)(ybar - mu)' from raw data
lav_mvn_w_tilde <- function(y = NULL, wt = NULL, mu = NULL) {
  if (!is.null(wt)) {
    out <- stats::cov.wt(y, wt = wt, method = "ML")
    out$cov + tcrossprod(out$center - mu)
  } else {
    lav_mat_cov(y, mu = mu)
  }
}

# 3b) W.tilde from sample statistics
lav_mvn_w_tilde_samp <- function(sample_mean = NULL,
                                 sample_cov = NULL,
                                 mu = NULL) {
  sample_cov + tcrossprod(as.numeric(sample_mean) - as.numeric(mu))
}

# 4) casewise score kernel
#
#    input:  yc = centered data (y - mu), sigma_inv
#    output: list with
#            $mu    = casewise scores with respect to mu          (= yc Sinv)
#            $sigma = casewise scores with respect to vech(Sigma)
lav_mvn_sc_kernel <- function(yc = NULL, sigma_inv = NULL) {
  p <- NCOL(sigma_inv)

  # vech(Sigma.inv)
  isigma <- lav_mat_vech(sigma_inv)

  # postmultiply with Sigma.inv
  ysi <- yc %*% sigma_inv

  # tcrossprod
  idx1 <- lav_mat_vech_col_idx(p)
  idx2 <- lav_mat_vech_row_idx(p)
  z <- ysi[, idx1, drop = FALSE] * ysi[, idx2, drop = FALSE]

  # subtract isigma from each row
  sc <- t(t(z) - isigma)

  # adjust for lav_mat_dup_pre (not vech!)
  sc[, lav_mat_diagh_idx(p)] <- sc[, lav_mat_diagh_idx(p)] / 2

  list(mu = ysi, sigma = sc)
}

# 5) observed-information kernel (the i11/i21/i22 algebra)
#
#    given Sigma^{-1}, Sigma, W.tilde and (ybar - mu), return the blocks of
#    the unit observed information for mu + vech(Sigma):
#      i11 = Sigma^{-1}
#      i21 = D' ((Sigma^{-1} (ybar - mu)) %x% Sigma^{-1})
#      i22 = 1/2 D' (Sigma^{-1} %x% AAA) D,
#            AAA = Sigma^{-1} (2 W.tilde - Sigma) Sigma^{-1}
#    if mean_diff is NULL, only i22 is computed
lav_mvn_info_obs_kernel <- function(sigma_inv = NULL,
                                    sigma_1 = NULL,
                                    w_tilde = NULL,
                                    mean_diff = NULL) {
  aaa <- sigma_inv %*% (2 * w_tilde - sigma_1) %*% sigma_inv
  i22 <- lav_mvn_kron_dup_half(sigma_inv, aaa)

  if (is.null(mean_diff)) {
    list(i22 = i22)
  } else {
    i21 <- lav_mat_dup_pre((sigma_inv %*% mean_diff) %x% sigma_inv)
    list(i11 = sigma_inv, i21 = i21, i22 = i22)
  }
}

# 6) 1/2 * D' (x %x% y) D  (y defaults to x)
#
#    this is the ONE home for a (future) lavaanC fast path:
#    lavaanC::m_kronecker_dup_pre_post(x, y, 0.5) or
#    lavaanC::m_kronecker_dup_cor_pre_post(x, multiplicator = 0.5)
lav_mvn_kron_dup_half <- function(x, y = NULL, correlation = FALSE) {
  if (is.null(y)) {
    y <- x
  }
  if (correlation) {
    0.5 * lav_mat_dup_cor_pre_post(x %x% y)
  } else {
    0.5 * lav_mat_dup_pre_post(x %x% y)
  }
}

# 6b) 2 * D^+ (x %x% y) D^{+T}  (y defaults to x)
#
#     (single home for a future lavaanC fast path:
#      lavaanC::m_kronecker_dup_ginv_pre_post(x, y, multiplicator = 2.0))
lav_mvn_kron_dup_ginv2 <- function(x, y = NULL) {
  if (is.null(y)) {
    y <- x
  }
  2 * lav_mat_dup_ginv_pre_post(x %x% y)
}

# 7) zero out the rows/cols of an information matrix (for mu + vech(Sigma),
#    or vech(Sigma) only) that correspond to fixed exogenous variables
lav_mvn_zero_x_idx <- function(out,
                               n = NULL,
                               x_idx = integer(0L),
                               meanstructure = TRUE) {
  if (length(x_idx) == 0L) {
    return(out)
  }
  not_x <- lav_mat_vech_which_idx(
    n = n, idx = x_idx,
    add_idx_at_start = meanstructure
  )
  out[not_x, ] <- 0
  out[, not_x] <- 0

  out
}


# 8) missing-data (FIML) helpers

# 8a) effective sample size: total weight/frequency of the non-empty cases
lav_mvn_mi_nobs <- function(mp = NULL, wt = NULL) {
  if (!is.null(wt)) {
    if (length(mp$empty.idx) > 0L) {
      sum(wt) - sum(wt[mp$empty.idx])
    } else {
      sum(wt)
    }
  } else {
    sum(mp$freq) # removed empty cases!
  }
}

# 8b) (weighted) frequency of missing pattern p
lav_mvn_mi_pattern_freq <- function(mp = NULL, p = NULL, wt = NULL) {
  if (!is.null(wt)) {
    sum(wt[mp$case.idx[[p]]])
  } else {
    mp$freq[p]
  }
}

# 8c) Sigma^{-1} for the observed part of a missing pattern, computed by
#     downdating the complete-data inverse; complete patterns return
#     sigma_inv as is (including its "logdet" attribute, if any)
lav_mvn_mi_pattern_inv <- function(sigma_inv = NULL,
                                   na_idx = integer(0L),
                                   logdet = FALSE,
                                   sigma_logdet = NULL) {
  if (length(na_idx) > 0L) {
    lav_mat_sym_inverse_update(
      s_inv = sigma_inv, rm_idx = na_idx,
      logdet = logdet, s_logdet = sigma_logdet
    )
  } else {
    sigma_inv
  }
}

# 8d) observed-information blocks (i11/i21/i22) for a single missing
#     pattern, zero-padded to the full p_1 x p_1 layout
#
#     my/sy are the pattern's sample mean/cov (observed variables only),
#     sigma_inv_1 the pattern's inverse (see lav_mvn_mi_pattern_inv)
lav_mvn_mi_info_obs_pattern <- function(sigma_inv_1 = NULL,
                                        var_idx = NULL,
                                        p_1 = NULL,
                                        my = NULL,
                                        sy = NULL,
                                        mu = NULL,
                                        sigma_1 = NULL) {
  s_inv <- matrix(0, p_1, p_1)
  s_inv[var_idx, var_idx] <- sigma_inv_1

  tmp21 <- matrix(0, p_1, 1)
  tmp21[var_idx, 1] <- sigma_inv_1 %*% (my - mu[var_idx])

  w_tilde <- sy + tcrossprod(my - mu[var_idx])
  aaa <- (sigma_inv_1 %*%
    (2 * w_tilde - sigma_1[var_idx, var_idx, drop = FALSE]) %*%
    sigma_inv_1)
  tmp22 <- matrix(0, p_1, p_1)
  tmp22[var_idx, var_idx] <- aaa

  list(
    i11 = s_inv,
    i21 = lav_mat_dup_pre(tmp21 %x% s_inv),
    i22 = lav_mvn_kron_dup_half(s_inv, tmp22)
  )
}

# 8e) casewise score engine for missing data
#
#     computes, pattern by pattern, the casewise scores with respect to
#     mu ($dmu) and vech(Sigma) ($sc) in the full p_1-column layout;
#     rows of empty cases remain NA; no weighting/fixed.x handling here
lav_mvn_mi_sc_engine <- function(y = NULL,
                                 mp = NULL,
                                 mu = NULL,
                                 sigma_inv = NULL) {
  p_1 <- NCOL(y)
  ny <- NROW(y)

  # for the tcrossprod
  idx1 <- lav_mat_vech_col_idx(p_1)
  idx2 <- lav_mat_vech_row_idx(p_1)

  # vech(Sigma.inv)
  i_sigma <- lav_mat_vech(sigma_inv)

  # subtract Mu
  yc <- t(t(y) - mu)

  # dmu / SC per case
  dmu <- matrix(as.numeric(NA), ny, p_1)
  sc <- matrix(as.numeric(NA), nrow = ny, ncol = length(i_sigma))

  # for each pattern, compute Yc %*% sigma.inv
  for (p in seq_len(mp$npatterns)) {
    # observed values for this pattern
    var_idx <- mp$pat[p, ]

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # cases with this pattern
    case_idx <- mp$case.idx[[p]]

    # invert Sigma for this pattern
    if (length(na_idx) > 0L) {
      sigma_inv_1 <- lav_mvn_mi_pattern_inv(
        sigma_inv = sigma_inv, na_idx = na_idx
      )
      tmp <- matrix(0, p_1, p_1)
      tmp[var_idx, var_idx] <- sigma_inv_1
      isigma <- lav_mat_vech(tmp)
    } else {
      sigma_inv_1 <- sigma_inv
      isigma <- i_sigma
    }

    # compute dMu for all observations of this pattern
    dmu[case_idx, var_idx] <-
      yc[case_idx, var_idx, drop = FALSE] %*% sigma_inv_1

    # postmultiply these cases with sigma.inv
    yc[case_idx, var_idx] <- dmu[case_idx, var_idx]

    # tcrossprod
    sc[case_idx, ] <- yc[case_idx, idx1] * yc[case_idx, idx2]

    # subtract isigma from each row
    sc[case_idx, ] <- t(t(sc[case_idx, , drop = FALSE]) - isigma)
  }

  # adjust for lav_mat_dup_pre (not vech!)
  sc[, lav_mat_diagh_idx(p_1)] <- sc[, lav_mat_diagh_idx(p_1)] / 2

  list(dmu = dmu, sc = sc)
}

# observed information for the two-level model with missing data,
# using Louis's (1982) method
#
# YR + Claude, July 2026
#
# Louis's identity:
#   -H_obs(theta) = E[-H_complete | v.obs] - Cov[S_complete | v.obs]
# where C is the complete-data loglikelihood: treating the cluster random
# effects, the missing y-values and the missing z-values as 'complete',
# C is simply the sum of two independent Gaussian samples (a two-group
# model):
#   d_ji = y_ji - nu_j ~ N(Mu.W, Sigma.W)   (one term per level-1 unit)
#   u_j  = (nu_j[both], z_j) ~ N(Mu.B, Sigma.B) (one term per cluster)
#
# the first term only requires the expected complete-data moments (one
# E-step); the second term is computed in closed form below, using the
# posterior distribution of the cluster-level unknowns (the same
# quantities as the E-step of lav_mvn_cl_mi_em_engine)
#
# this replaces the numerical Hessian (4 x npar full-data gradient
# passes) by a single full-data pass

# ---------------------------------------------------------------------
# Cov[ S_complete | v.obs ] in the *moment* space, where S_complete is
# the complete-data score of (-2 * logl) with respect to
# (Mu.W, vech(Sigma.W), Mu.B, vech(Sigma.B)) -- the same packing (and
# scale) as lav_mvn_cl_mi_dlogl_2l_samp()
#
# notation, per cluster j (centered): w = (beta, ztilde[mis]) with
# posterior N(m1, C1); per unit i with missing pattern p:
#   e_ji = d_ji - Mu.W = a_i + F_p beta + eta_i,  eta_i ~ N(0, V_p)
# where a_i = G_p yc_i[obs], G_p = [I; B_p] (observed rows identity,
# missing rows the sweep regression), F_p = -G_p S_ob (S_ob selects the
# observed both-columns), V_p = C_p embedded in the missing block;
# and t_j = u_j - Mu.B = k_j + K_q w (k_j holds the observed ztilde)
#
# the law of total covariance then gives two parts:
# (1) E_w[ Cov(S | w) ]: only the within scores are random given w;
#     unit-wise Magnus-Neudecker moments, aggregated per pattern
# (2) Cov_w( E[S | w] ): E[S|w] is a quadratic polynomial in w; for
#     centered w ~ N(0, C1): Cov(b'w + w'Rw, ...) = b1'C b2 +
#     2 tr(R1 C R2 C)
lav_mvn_cl_mi_scov_louis <- function(y1 = NULL,
                                     y2 = NULL,
                                     lp = NULL,
                                     mp = NULL,
                                     mu_w = NULL, # implied Mu.W
                                     sigma_w = NULL, # implied Sigma.W
                                     mu_b = NULL, # implied Mu.B
                                     sigma_b = NULL, # implied Sigma.B
                                     extra = FALSE,
                                     out = NULL) { # pre-converted 2l list
  # map implied to 2l matrices (unless the caller already did)
  if (is.null(out)) {
    out <- lav_mvn_cl_implied22l(
      lp = lp, mu_w = mu_w, mu_b = mu_b,
      sigma_w = sigma_w, sigma_b = sigma_b
    )
  }
  mu_y <- out$mu.y
  mu_z <- out$mu.z
  sw <- out$sigma.w # == sigma_w (within/y1w space)
  sb <- out$sigma.b[lp$both.idx[[2]], lp$both.idx[[2]], drop = FALSE]
  sigma_zz <- out$sigma.zz
  syz <- out$sigma.yz[lp$both.idx[[2]], , drop = FALSE]

  # dimensions and indices
  nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]
  between_idx <- lp$between.idx[[2]]
  both_idx <- lp$both.idx[[2]]
  nz <- length(between_idx)
  nb <- length(both_idx)
  zp <- mp$Zp
  if (nz > 0L) {
    y1w <- y1[, -between_idx, drop = FALSE]
    z <- y2[, between_idx, drop = FALSE]
    zc <- t(t(z) - mu_z)
  } else {
    y1w <- y1
    z <- NULL
  }
  ny <- ncol(y1w)
  p2 <- nb + nz
  y1w_c <- t(t(y1w) - mu_y)

  # score-space dimensions and offsets:
  # [Mu.W (ny), vech Sigma.W, Mu.B (p2), vech Sigma.B]
  nvw <- ny * (ny + 1L) / 2L
  nvb <- p2 * (p2 + 1L) / 2L
  ntot <- ny + nvw + p2 + nvb
  i_muw <- seq_len(ny)
  i_sw <- ny + seq_len(nvw)
  i_mub <- ny + nvw + seq_len(p2)
  i_sb <- ny + nvw + p2 + seq_len(nvb)

  # constants
  ow <- solve.default(sw)
  ob <- solve.default(sigma_b) # implied Sigma.B == [[sb, syz],[szy, szz]]
  gow <- lav_mat_dup_pre(ow %x% ow) # Gdd_w (Omega x Omega): nvw x ny^2
  gob <- lav_mat_dup_pre(ob %x% ob) # nvb x p2^2
  kny <- lav_mat_com(ny, ny)

  # ---- posterior machinery (same as the E-step) --------------------
  # (shared with lav_mvn_cl_mi_estep_ranef)
  post <- lav_mvn_cl_mi_posterior(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    mu_w = mu_w, sigma_w = sigma_w, mu_b = mu_b, sigma_b = sigma_b,
    out = out
  )
  breg_p <- post$breg
  ccond_p <- post$ccond
  m1_list <- post$m1
  c1_list <- post$c1
  zmis_q <- post$zmis_q
  zpat2j <- post$zpat2j
  cluster_idx <- lp$cluster.idx[[2]]

  # ---- per-pattern constants + data aggregates ---------------------
  # G_p, F_p, V_p; per (pattern, cluster): n_pj and sum of yc[obs]
  np <- mp$npatterns
  g_list <- f_list <- v_list <- vector("list", np)
  ycsum_list <- jmap_list <- nvec_list <- vector("list", np)
  for (p in seq_len(np)) {
    o_idx <- which(mp$pat[p, ])
    na_idx <- which(!mp$pat[p, ])
    g_p <- matrix(0, ny, length(o_idx))
    g_p[o_idx, ] <- diag(length(o_idx))
    if (length(na_idx) > 0L) {
      g_p[na_idx, ] <- breg_p[[p]]
    }
    g_list[[p]] <- g_p
    # F_p = -G_p S_ob: columns for both-vars; zero if unobserved
    f_p <- matrix(0, ny, nb)
    obs_b <- match(both_idx, o_idx) # NA if this both-var is missing
    seen <- which(!is.na(obs_b))
    if (length(seen) > 0L) {
      f_p[, seen] <- -g_p[, obs_b[seen], drop = FALSE]
    }
    f_list[[p]] <- f_p
    v_p <- matrix(0, ny, ny)
    if (length(na_idx) > 0L) {
      v_p[na_idx, na_idx] <- ccond_p[[p]]
    }
    v_list[[p]] <- v_p
    # per-cluster aggregates for this pattern
    case_idx <- mp$case.idx[[p]]
    j_idx <- mp$j.idx[[p]]
    ycs <- rowsum.default(y1w_c[case_idx, o_idx, drop = FALSE],
      group = j_idx, reorder = FALSE
    )
    jmap_list[[p]] <- as.integer(rownames(ycs))
    ycsum_list[[p]] <- ycs
    nvec <- rowsum.default(rep(1, length(case_idx)),
      group = j_idx, reorder = FALSE
    )
    nvec_list[[p]] <- drop(nvec)
  }

  # ---- part 1: E_w[ Cov(S | w) ] -- per-pattern totals -------------
  # per pattern we need (summing over all units in the pattern):
  #   chat_tot = sum_i E[c_i]  and  cc_tot = sum_i E[c_i c_i']
  # with c_i = a_i + F_p beta, beta | v.obs ~ N(m1[1:nb], C1[bb])
  scov <- matrix(0, ntot, ntot)
  escore <- numeric(ntot) # E[S | v.obs] (Fisher identity check)
  se_tot <- matrix(0, ny, ny) # sum_i E[e e'] (for escore)
  chat_all <- numeric(ny) # sum_i E[c_i] over everything
  for (p in seq_len(np)) {
    g_p <- g_list[[p]]
    f_p <- f_list[[p]]
    v_p <- v_list[[p]]
    jm <- jmap_list[[p]]
    nvec <- nvec_list[[p]]
    ycs <- ycsum_list[[p]]
    o_idx <- which(mp$pat[p, ])
    freq <- mp$freq[p]

    # pattern totals
    m1b <- do.call(rbind, lapply(m1_list[jm], function(m) m[seq_len(nb)]))
    if (nb == 0L) {
      m1b <- matrix(0, length(jm), 0L)
    }
    # chat_tot = G_p sum_j ycsum_pj + F_p sum_j n_pj m1b_j
    chat_tot <- drop(g_p %*% colSums(ycs)) +
      drop(f_p %*% colSums(nvec * m1b))
    # cc_tot = G (sum yc yc') G' + cross terms + n-weighted quadratic
    yy <- crossprod(y1w_c[mp$case.idx[[p]], o_idx, drop = FALSE])
    m1s <- crossprod(ycs, m1b) # sum_j ycsum_pj m1b_j'
    w2 <- matrix(0, nb, nb) # sum_j n_pj (m1b m1b' + C1[bb])
    for (k in seq_along(jm)) {
      c1bb <- c1_list[[jm[k]]][seq_len(nb), seq_len(nb), drop = FALSE]
      w2 <- w2 + nvec[k] * (tcrossprod(m1b[k, ]) + c1bb)
    }
    cross <- g_p %*% m1s %*% t(f_p)
    cc_tot <- g_p %*% yy %*% t(g_p) + cross + t(cross) +
      f_p %*% w2 %*% t(f_p)

    # (mu.w, mu.w): 4 freq * Ow V Ow
    scov[i_muw, i_muw] <- scov[i_muw, i_muw] +
      4 * freq * (ow %*% v_p %*% ow)
    # (mu.w, sigma.w): 2 Ow [V x chat' + chat' x V] t(gow)
    m_x <- (v_p %x% t(chat_tot)) + (t(chat_tot) %x% v_p)
    scov[i_muw, i_sw] <- scov[i_muw, i_sw] + 2 * (ow %*% m_x %*% t(gow))
    # (sigma.w, sigma.w): gow (I+K)[freq VxV + V x cc + cc x V] t(gow)
    m_q <- freq * (v_p %x% v_p) + (v_p %x% cc_tot) + (cc_tot %x% v_p)
    m_q <- m_q + kny %*% m_q
    scov[i_sw, i_sw] <- scov[i_sw, i_sw] + gow %*% m_q %*% t(gow)

    # bookkeeping for the expected score
    se_tot <- se_tot + cc_tot + freq * v_p
    chat_all <- chat_all + chat_tot
  }
  # empty units: e ~ N(0, Sigma.W)
  n_empty <- length(mp$empty.idx)
  if (n_empty > 0L) {
    scov[i_muw, i_muw] <- scov[i_muw, i_muw] + 4 * n_empty * ow
    m_q <- sw %x% sw
    m_q <- m_q + kny %*% m_q
    scov[i_sw, i_sw] <- scov[i_sw, i_sw] + n_empty * (gow %*% m_q %*% t(gow))
    se_tot <- se_tot + n_empty * sw
  }

  # ---- part 2: Cov_w( E[S | w] ) -- per-cluster --------------------
  # E[S|w] = const + b'wtilde + wtilde' R wtilde, wtilde ~ N(0, C1)
  et_tot <- matrix(0, p2, p2) # sum_j E[t t'] (for escore)
  that_all <- numeric(p2)
  for (j in seq_len(nclusters)) {
    m1 <- m1_list[[j]]
    c1 <- c1_list[[j]]
    nw <- length(m1)
    nzm <- nw - nb
    # K_q: w -> t (p2 x nw)
    k_q <- matrix(0, p2, nw)
    if (nb > 0L) {
      k_q[seq_len(nb), seq_len(nb)] <- diag(nb)
    }
    if (nzm > 0L) {
      zm_idx <- zmis_q[[zpat2j[j]]]
      k_q[nb + zm_idx, nb + seq_len(nzm)] <- diag(nzm)
    }
    # that = k_j + K m1
    that <- numeric(p2)
    if (nz > 0L) {
      zrow <- zc[j, ]
      zrow[is.na(zrow)] <- 0
      that[nb + seq_len(nz)] <- zrow
    }
    that <- that + drop(k_q %*% m1)

    # B matrix (ntot x nw) and R stack (ntot x nw^2)
    bmat <- matrix(0, ntot, nw)
    rstk <- matrix(0, ntot, nw * nw)
    # within rows: sums over the patterns present in this cluster
    if (nb > 0L) {
      fsum <- matrix(0, ny, nb) # sum_p n_pj F_p
      usum <- matrix(0, ny * ny, nb) # sum_p (F x chat_pj + chat_pj x F)
      ffsum <- matrix(0, ny * ny, nb * nb) # sum_p n_pj (F x F)
      m1b <- m1[seq_len(nb)]
      for (p in seq_len(np)) {
        k <- match(j, jmap_list[[p]])
        if (is.na(k)) {
          next
        }
        n_pj <- nvec_list[[p]][k]
        f_p <- f_list[[p]]
        chat_pj <- drop(g_list[[p]] %*% ycsum_list[[p]][k, ]) +
          n_pj * drop(f_p %*% m1b)
        fsum <- fsum + n_pj * f_p
        usum <- usum + (f_p %x% chat_pj) + (chat_pj %x% f_p)
        ffsum <- ffsum + n_pj * (f_p %x% f_p)
      }
      bmat[i_muw, seq_len(nb)] <- -2 * (ow %*% fsum)
      bmat[i_sw, seq_len(nb)] <- -(gow %*% usum)
      # quadratic rows: R = reshape of -(gow ffsum), beta-block only;
      # embed the nb x nb block into the nw x nw vec-space
      rsw <- -(gow %*% ffsum) # nvw x nb^2
      bpos <- as.vector(outer(seq_len(nb), seq_len(nb),
        function(r, s) (s - 1L) * nw + r
      ))
      rstk[i_sw, bpos] <- rsw
    }
    # between rows
    bmat[i_mub, ] <- -2 * (ob %*% k_q)
    bmat[i_sb, ] <- -(gob %*% ((k_q %x% that) + (that %x% k_q)))
    rstk[i_sb, ] <- -(gob %*% (k_q %x% k_q))

    # symmetrize the quadratic part: R <- (R + R')/2
    com_idx <- as.vector(outer(seq_len(nw), seq_len(nw),
      function(r, s) (r - 1L) * nw + s
    ))
    rstk <- (rstk + rstk[, com_idx, drop = FALSE]) / 2

    # Cov contribution: B C1 B' + 2 R (C1 x C1) R'
    scov <- scov + bmat %*% c1 %*% t(bmat) +
      2 * (rstk %*% (c1 %x% c1) %*% t(rstk))

    # bookkeeping for the expected score
    et_tot <- et_tot + tcrossprod(that) + k_q %*% c1 %*% t(k_q)
    that_all <- that_all + that
  }
  # part 1 has no (mu.w x sigma.w) transpose yet
  scov[i_sw, i_muw] <- t(scov[i_muw, i_sw])

  # ---- expected score (Fisher identity: == observed-data score) ----
  escore[i_muw] <- -2 * drop(ow %*% chat_all)
  escore[i_sw] <- drop(nobs * lav_mat_vech_dd(ow) -
    gow %*% as.vector(se_tot))
  escore[i_mub] <- -2 * drop(ob %*% that_all)
  escore[i_sb] <- drop(nclusters * lav_mat_vech_dd(ob) -
    gob %*% as.vector(et_tot))

  res <- list(scov = scov, escore = escore)
  if (extra) {
    res$post <- list(
      m1 = m1_list, c1 = c1_list,
      breg = breg_p, ccond = ccond_p,
      mu_y = mu_y, mu_z = mu_z, both_idx = both_idx,
      between_idx = between_idx,
      zmis_q = if (nz > 0L) zmis_q else NULL,
      zpat2j = if (nz > 0L) zpat2j else NULL
    )
  }
  res
}

# ---------------------------------------------------------------------
# observed information (on the scale of lav_model_hessian(), i.e. the
# Hessian of the objective function -logl/N) for a two-level model with
# missing data, using Louis's method; single group only
lav_mvn_cl_mi_h_louis <- function(lavmodel = NULL,
                                  lavsamplestats = NULL,
                                  lavdata = NULL,
                                  ceq_simple = FALSE,
                                  h = 1e-05) {
  stopifnot(lavdata@ngroups == 1L, lavdata@nlevels == 2L)
  lp <- lavdata@Lp[[1]]
  mp <- lavdata@Mp[[1]]
  y1 <- lavdata@X[[1]]
  y2 <- lavsamplestats@YLp[[1]][[2]]$Y2
  ntotal <- lavsamplestats@ntotal
  nclusters <- lp$nclusters[[2]]
  # note: the complete-data model has one 'within' term per level-1 unit,
  # *including* units with all level-1 variables missing; nunits may
  # therefore exceed ntotal (which excludes the empty units)
  nunits <- lp$nclusters[[1]]

  # current parameter values and implied moments
  implied <- lav_model_implied(lavmodel)
  mu_w <- implied$mean[[1]]
  sigma_w <- implied$cov[[1]]
  mu_b <- implied$mean[[2]]
  sigma_b <- implied$cov[[2]]

  # 1. one E-step: expected complete-data moments (per unit/cluster)
  engine <- lav_mvn_cl_mi_em_engine(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    min_variance = .Machine$double.eps # no clamping here
  )
  out2 <- lav_mvn_cl_implied22l(
    lp = lp, mu_w = mu_w, mu_b = mu_b,
    sigma_w = sigma_w, sigma_b = sigma_b
  )
  theta <- engine$pack(
    out2$mu.w, out2$mu.b, out2$sigma.w, out2$sigma.b,
    out2$mu.z, out2$sigma.zz, out2$sigma.yz
  )
  estat <- engine$implied(engine$step(theta))

  # convention shift: the E-step machinery imputes nu centered at the
  # *total* mean (the '2l' convention: mu.b.2l = Mu.B + Mu.W[both]),
  # while the complete-data score below is written in the 'implied'
  # convention (d ~ N(Mu.W, Sigma.W), u ~ N(Mu.B, Sigma.B), with Mu.W
  # possibly nonzero for both-variables, e.g. with within-level
  # covariates); the score *vector* is identical under both conventions
  # (the residuals coincide), but the expected means differ by the fixed
  # shift s0 = Mu.W[both] (at the current parameter values):
  #   E[d.impl] = E[d.2l] + s0,   E[u.impl] = E[u.2l] - s0
  # (the central second moments are shift-invariant)
  both_idx <- lp$both.idx[[2]]
  nb <- length(both_idx)
  s0 <- drop(mu_w)[both_idx]
  estat_mu_w <- drop(estat$Mu.W)
  estat_mu_w[both_idx] <- estat_mu_w[both_idx] + s0
  estat_mu_b <- drop(estat$Mu.B)
  estat_mu_b[seq_len(nb)] <- estat_mu_b[seq_len(nb)] - s0

  # 2. Cov[ S_complete | v.obs ] in moment space (-2 logl scale)
  #    (out2 avoids re-converting the same implied moments)
  scov <- lav_mvn_cl_mi_scov_louis(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    mu_w = mu_w, sigma_w = sigma_w, mu_b = mu_b, sigma_b = sigma_b,
    out = out2
  )$scov

  # 3. term A: jacobian of the complete-data gradient (in theta space),
  #    with the expected sufficient statistics held fixed; this is the
  #    gradient of the two-group model fitted to the expected moments,
  #    on the same scale as lav_model_grad() (objective = -logl/N);
  #    with simple equality constraints, we (like lav_model_delta) work
  #    in the unconstrained space, and reduce at the very end
  if (lavmodel@ceq.simple.only) {
    npar <- lavmodel@nx.unco
    type_glist <- "unco"
  } else {
    npar <- lavmodel@nx.free
    type_glist <- "free"
  }
  cgrad <- function(x) {
    glist2 <- lav_model_x2glist(lavmodel, type = type_glist, x = x)
    implied_x <- lav_model_implied(lavmodel, glist = glist2)
    delta_x <- lav_model_delta(lavmodel, glist = glist2)[[1]]
    # complete-data score (-2 logl scale) at the expected moments
    ow_x <- solve.default(implied_x$cov[[1]])
    resw <- estat_mu_w - drop(implied_x$mean[[1]])
    m_w <- nunits * (estat$Sigma.W + tcrossprod(resw))
    s_muw <- -2 * nunits * drop(ow_x %*% resw)
    s_sw <- lav_mat_vech_dd(nunits * ow_x - ow_x %*% m_w %*% ow_x)
    ob_x <- solve.default(implied_x$cov[[2]])
    resb <- estat_mu_b - drop(implied_x$mean[[2]])
    m_b <- nclusters * (estat$Sigma.B + tcrossprod(resb))
    s_mub <- -2 * nclusters * drop(ob_x %*% resb)
    s_sb <- lav_mat_vech_dd(nclusters * ob_x - ob_x %*% m_b %*% ob_x)
    drop(c(s_muw, s_sw, s_mub, s_sb) %*% delta_x) / (2 * ntotal)
  }
  x <- lav_model_get_parameters(lavmodel = lavmodel)
  if (lavmodel@ceq.simple.only) {
    # unpack to the unconstrained space
    x <- drop(x %*% t(lavmodel@ceq.simple.K))
  }
  term_a <- matrix(0, npar, npar)
  for (jj in seq_len(npar)) {
    x_r <- x_l <- x
    x_r[jj] <- x[jj] + h
    x_l[jj] <- x[jj] - h
    term_a[, jj] <- (cgrad(x_r) - cgrad(x_l)) / (2 * h)
  }
  term_a <- (term_a + t(term_a)) / 2

  # 4. term B: Delta' Cov Delta, on the objective scale;
  #    obj = (-2 logl)/(2N) and -2 H_O = E[-2 H_C] - Cov(S2)/2, so
  #    H_obj = term_a - Delta' Cov Delta / (4N)
  #    (lav_model_delta() columns are in the unconstrained space when
  #    ceq.simple.only, matching term_a)
  delta <- lav_model_delta(lavmodel = lavmodel)[[1]]
  term_b <- crossprod(delta, scov %*% delta) / (4 * ntotal)

  hessian <- term_a - term_b
  if (ceq_simple && lavmodel@ceq.simple.only) {
    # reduce to the free parameters
    k_mat <- lavmodel@ceq.simple.K
    hessian <- crossprod(k_mat, hessian %*% k_mat)
  }
  hessian
}

# what are the block values (not necessarily integers)
lav_pt_block_values <- function(partable) {
  if (is.null(partable$block)) {
    block_values <- 1L
  } else {
    # always integers
    tmp <- partable$block[partable$block > 0L & # non-zero only
      !partable$op %in% c("==", "<", ">", ":=")]
    block_values <- unique(na.omit(tmp)) # could be, eg, '2' only
  }

  block_values
}

# guess number of blocks from a partable
lav_pt_nblocks <- function(partable) {
  length(lav_pt_block_values(partable))
}

# what are the group values (not necessarily integers)
lav_pt_group_values <- function(partable) {
  # FLAT?
  if (any(partable$op == ":")) {
    colon_idx <- which(partable$op == ":" &
      tolower(partable$lhs) == "group")
    if (length(colon_idx) > 0L) {
      group_values <- unique(partable$rhs[colon_idx])
    }
    # regular partable
  } else if (is.null(partable$group)) {
    group_values <- 1L
  } else if (is.numeric(partable$group)) {
    tmp <- partable$group[partable$group > 0L &
      !partable$op %in% c("==", "<", ">", ":=")]
    group_values <- unique(na.omit(tmp))
  } else { # character
    tmp <- partable$group[nchar(partable$group) > 0L &
      !partable$op %in% c("==", "<", ">", ":=")]
    group_values <- unique(na.omit(tmp))
  }

  group_values
}

# guess number of groups from a partable
lav_pt_ngroups <- function(partable) {
  length(lav_pt_group_values(partable))
}

# what are the level values (not necessarily integers)
lav_pt_level_values <- function(partable) {
  # FLAT?
  if (any(partable$op == ":")) {
    colon_idx <- which(partable$op == ":" &
      tolower(partable$lhs) == "level")
    level_values <- integer(0L)
    if (length(colon_idx) > 0L) {
      level_values <- unique(partable$rhs[colon_idx])
    }
    # regular partable
  } else if (is.null(partable$level)) {
    level_values <- 1L
  } else if (is.numeric(partable$level)) {
    tmp <- partable$level[partable$level > 0L &
      !partable$op %in% c("==", "<", ">", ":=")]
    level_values <- unique(na.omit(tmp))
  } else { # character
    tmp <- partable$level[nchar(partable$level) > 0L &
      !partable$op %in% c("==", "<", ">", ":=")]
    level_values <- unique(na.omit(tmp))
  }

  level_values
}

# guess number of levels from a partable
lav_pt_nlevels <- function(partable) {
  length(lav_pt_level_values(partable))
}

# efa sets values
lav_pt_efa_values <- function(partable) {
  if (is.null(partable$efa)) {
    efa_values <- character(0L)
  } else { # should be character
    tmp_efa <- as.character(partable$efa)
    tmp <- tmp_efa[nchar(tmp_efa) > 0L &
      !partable$op %in% c("==", "<", ">", ":=")]
    efa_values <- unique(na.omit(tmp))
  }

  efa_values
}

# number of efa sets from a partable
lav_pt_nefa <- function(partable) {
  length(lav_pt_efa_values(partable))
}




# number of sample statistics per block
lav_pt_ndat <- function(partable) {
  # global
  meanstructure <- any(partable$op == "~1")
  fixed_x <- any(partable$exo > 0L & partable$free == 0L)
  conditional_x <- any(partable$exo > 0L & partable$op == "~")
  categorical <- any(partable$op == "|")
  composites <- any(partable$op == "<~")
  correlation <- any(partable$op == "~*~")
  if (categorical) {
    meanstructure <- TRUE
  }

  # blocks
  nblocks <- lav_pt_nblocks(partable)
  nlevels <- lav_pt_nlevels(partable)
  ndat <- integer(nblocks)

  for (b in seq_len(nblocks)) {
    # how many observed variables in this block?
    if (conditional_x) {
      ov_names <- lav_pt_vnames(partable, "ov.nox", block = b)
    } else {
      ov_names <- lav_pt_vnames(partable, "ov", block = b)
    }
    nvar <- length(ov_names)

    # pstar
    pstar <- nvar * (nvar + 1) / 2
    if (meanstructure) {
      pstar <- pstar + nvar

      # no meanstructure if within level, except ov.x which is not
      # decomposed
      if (nlevels > 1L && (b %% nlevels) == 1L) {
        # all zero
        pstar <- pstar - nvar

        # except within-only 'y'
        ov_names_y <- lav_pt_vnames(partable, "ov.nox", block = b)
        ov_names_y2 <- unlist(lav_pt_vnames(partable, "ov",
          block = seq_len(nblocks)[-b]
        ))
        ov_names_y <- ov_names_y[!ov_names_y %in% ov_names_y2]
        if (length(ov_names_y) > 0L) {
          pstar <- pstar + length(ov_names_y)
        }

        # except within-only 'x' (unless fixed.x)
        ov_names_x <- lav_pt_vnames(partable, "ov.x", block = b)
        ov_names_x2 <- unlist(lav_pt_vnames(partable, "ov",
          block = seq_len(nblocks)[-b]
        ))
        ov_names_x <- ov_names_x[!ov_names_x %in% ov_names_x2]
        if (!fixed_x && length(ov_names_x) > 0L) {
          pstar <- pstar + length(ov_names_x)
        }
      }
    }

    ndat[b] <- pstar

    # correction for fixed.x?
    if (!conditional_x && fixed_x) {
      ov_names_x <- lav_pt_vnames(partable, "ov.x", block = b)
      nvar_x <- length(ov_names_x)
      pstar_x <- nvar_x * (nvar_x + 1) / 2
      if (meanstructure) {
        if (nlevels > 1L && (b %% nlevels) == 1L) {
          # do nothing, they are already removed
        } else {
          pstar_x <- pstar_x + nvar_x
        }
      }
      ndat[b] <- ndat[b] - pstar_x
    }

    # composites?
    if (composites) {
      ov_cind <- lav_pt_vnames(partable, "ov.cind", block = b)
      covar_idx <- which(partable$op == "~~" &
                         partable$lhs %in% ov_cind &
                         partable$rhs %in% ov_cind &
                         partable$free == 0L)
      ndat[b] <- ndat[b] - length(covar_idx)
    }

    # correction for ordinal data?
    if (categorical) {
      ov_names_x <- lav_pt_vnames(partable, "ov.x", block = b)
      nexo <- length(ov_names_x)
      ov_ord <- lav_pt_vnames(partable, "ov.ord", block = b)
      nvar_ord <- length(ov_ord)
      th <- lav_pt_vnames(partable, "th", block = b)
      nth <- length(th)
      if (nlevels > 1L) {
        # two-level categorical (theta parameterization):
        if ((b %% nlevels) == 1L) {
          # within level: no variances for the ordinal variables (fixed
          # to 1); the means were never counted; the thresholds are
          # between-level statistics
          ndat[b] <- ndat[b] - nvar_ord
        } else {
          # between level: no means for the ordinal variables (fixed to
          # 0), but additional thresholds; the between variances of the
          # ordinal variables ARE statistics
          ndat[b] <- ndat[b] - nvar_ord
          ndat[b] <- ndat[b] + nth
        }
        # add slopes (conditional.x): per block, every 'y' variable is
        # regressed on the block's exogenous covariates
        if (conditional_x) {
          ndat[b] <- ndat[b] + (nvar * nexo)
        }
      } else {
        # no variances
        ndat[b] <- ndat[b] - nvar_ord
        # no means
        ndat[b] <- ndat[b] - nvar_ord
        # but additional thresholds
        ndat[b] <- ndat[b] + nth
        # add slopes
        if (conditional_x) {
          ov_names_x <- lav_pt_vnames(partable, "ov.x", block = b)
          nexo <- length(ov_names_x)
          ndat[b] <- ndat[b] + (nvar * nexo)
        }
      }
    }

    # correction for correlation not categorical: every variable that is
    # standardized to unit variance (i.e. has a '~*~' scaling row) removes
    # one sample statistic. For a 'partial' correlation structure this is a
    # subset of the variables, so count the '~*~' rows rather than 'nvar'.
    if (correlation && !categorical) {
      n_cor <- sum(partable$op == "~*~" & partable$block == b)
      ndat[b] <- ndat[b] - n_cor
    }

    # correction for conditional.x not categorical
    if (conditional_x && !categorical) {
      ov_names_x <- lav_pt_vnames(partable, "ov.x", block = b)
      nexo <- length(ov_names_x)
      # add slopes
      ndat[b] <- ndat[b] + (nvar * nexo)
    }

    # correction for group proportions?
    group_idx <- which(partable$lhs == "group" &
      partable$op == "%" &
      partable$block == b)
    if (length(group_idx) > 0L) {
      # ndat <- ndat + (length(group.idx) - 1L) # G - 1 (sum to one)
      ndat[b] <- ndat[b] + 1L # poisson: each cell a parameter
    }
  } # blocks

  # sum over all blocks
  sum(ndat)
}

# total number of free parameters (ignoring equality constraints)
lav_pt_npar <- function(partable) {
  # we only assume non-zero values
  npar <- length(which(partable$free > 0L))
  npar
}

# global degrees of freedom: ndat - npar
# ignoring constraints! (not very useful)
#
# we need to find the rank of con.jac to find the exact amount
# of non-redundant equality constraints (this is done in lav_test.R)
lav_pt_df <- function(partable) {
  npar <- lav_pt_npar(partable)
  ndat <- lav_pt_ndat(partable)

  # degrees of freedom
  df <- ndat - npar

  as.integer(df)
}

# check order of covariances: we only fill the upper.tri
# therefore, we 'switch' lhs & rhs if they appear in the wrong order
lav_pt_covariance_reorder <- function(partable,
                                            ov_names = NULL,
                                            lv_names = NULL) {
  # shortcut
  cov_idx <- which(partable$op == "~~" & partable$lhs != partable$rhs)
  if (length(cov_idx) == 0L) {
    # nothing to do
    return(partable)
  }

  # get names
  if (is.null(ov_names)) {
    ov_names <- lav_pt_vnames(partable, "ov")
  } else {
    ov_names <- unlist(ov_names)
  }
  if (is.null(lv_names)) {
    lv_names <- lav_pt_vnames(partable, "lv")
    # add random slopes (if any)
    if (!is.null(partable$rv) && any(nchar(partable$rv) > 0L)) {
      rv_names <- unique(partable$rv[nchar(partable$rv) > 0L])
      lv_names <- c(lv_names, rv_names)
    }
  } else {
    lv_names <- unlist(lv_names)
  }
  lv_ov_names <- c(lv_names, ov_names)

  # identify wrong ordering
  lhs_idx <- match(partable$lhs[cov_idx], lv_ov_names)
  rhs_idx <- match(partable$rhs[cov_idx], lv_ov_names)
  swap_idx <- cov_idx[lhs_idx > rhs_idx]

  if (length(swap_idx) == 0L) {
    # nothing to do
    return(partable)
  }

  # swap!
  tmp <- partable$lhs[swap_idx]
  partable$lhs[swap_idx] <- partable$rhs[swap_idx]
  partable$rhs[swap_idx] <- tmp

  partable
}

# add a single parameter to an existing parameter table
lav_pt_add <- function(partable = NULL, add = list()) {
  # treat partable as list, not as a data.frame
  partable <- as.list(partable)

  # number of elements
  nel <- length(partable$lhs)

  # add copy of last row
  for (c in seq_along(partable)) {
    if (is.integer(partable[[c]][[1]])) {
      if (partable[[c]][nel] == 0L) {
        partable[[c]][nel + 1] <- 0L
      } else if (partable[[c]][nel] == 1L) {
        partable[[c]][nel + 1] <- 1L
      } else {
        partable[[c]][nel + 1] <- partable[[c]][nel] + 1L
      }
    } else if (is.character(partable[[c]][[1]])) {
      partable[[c]][nel + 1] <- ""
    } else if (is.numeric(partable[[c]][[1]])) {
      partable[[c]][nel + 1] <- 0
    } else {
      partable[[c]][nel + 1] <- partable[[c]][nel]
    }

    # replace
    if (names(partable)[c] %in% names(add)) {
      partable[[c]][nel + 1] <- add[[names(partable)[c]]]
    }
  }

  partable
}


# look for p2-row-idx of p1 elements
# p1 is usually a subset of p2
# return NA if not found
lav_pt_map_id_p1_in_p2 <- function(p1, p2, stopifnotfound = TRUE,
                                         exclude_nonpar = TRUE) {
  # check if we have a 'block' column (in both p1 and p2)
  if (is.null(p1$block)) {
    if (is.null(p1$group)) {
      p1$block <- rep.int(1L, length(p1$lhs))
    } else {
      p1$block <- p1$group
    }
  }
  if (is.null(p2$block)) {
    if (is.null(p2$group)) {
      p2$block <- rep.int(1L, length(p2$lhs))
    } else {
      p2$block <- p2$group
    }
  }

  # ALL rows from p1, or only 'parameters'?
  if (exclude_nonpar) {
    # get all parameters that have a '.p*' plabel
    # (they exclude "==", "<", ">", ":=")
    p1_idx <- grep(".p", p1$plabel, fixed = TRUE)
  } else {
    # all of it
    # note: block should be '0' in both p1 and p2
    p1_idx <- seq_along(p1$lhs)
  }
  np1 <- length(p1_idx)

  # return p2.id
  p2_id <- integer(np1)

  # check every parameter in p1
  for (i in seq_len(np1)) {
    # identify parameter in p1
    lhs <- p1$lhs[i]
    op <- p1$op[i]
    rhs <- p1$rhs[i]
    block <- p1$block[i]

    # search for corresponding parameter in p2
    p2_idx <- which(p2$lhs == lhs & p2$op == op & p2$rhs == rhs &
      p2$block == block)

    # found?
    if (length(p2_idx) == 0L) {
      if (stopifnotfound) {
        lav_msg_stop(gettext("parameter in p1 not found in p2:"),
          paste(lhs, op, rhs, "(block = ", block, ")", sep = " ")
        )
      } else {
        p2_id[i] <- as.integer(NA)
      }
    } else {
      p2_id[i] <- p2_idx
    }
  }

  p2_id
}

# the weighted bivariate ordinal/linear model
# YR 08 March 2020 (replacing the old lav_polyserial.R routines)
#
# - polyserial (and biserial) correlations
# - bivariate ordinal/linear regression
# - using sampling weights wt
#
# YR/LDW 2026: refactored (see lav_uvbv_common.R for the shared cache/minfns
#              design); no change in behavior
#
#  info concerning lintr package, Luc DW May 7, 2026
#  there is a known problem in codetools, transferred to lintr, that
#  variables assigned in a with() are marked as 'may not be used';
#  to this end the code in with() statements are excluded from linting!
#

# polyserial correlation
#
# Y1 = linear
# Y2 = ordinal
#
lav_bvmix_cor_twostep_fit <- function(y1, y2, exo = NULL, wt = NULL,
                                      fit_y1 = NULL, fit_y2 = NULL,
                                      y1_name = NULL, y2_name = NULL,
                                      optim_method = "nlminb1", # 0.6-7
                                      optim_scale = 1.0,
                                      init_theta = NULL,
                                      control = list()) {
  # create cache environment
  cache <- lav_bvmix_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt,
    fit_y1 = fit_y1, fit_y2 = fit_y2
  )

  # optim.method
  minfns <- lav_uvbv_optim_fns(optim_method, lav_bvmix_min_fns())

  # optimize
  if (is.null(control$trace)) {
    control$trace <- ifelse(lav_verbose(), 1, 0)
  }

  # init theta?
  if (!is.null(init_theta)) {
    start_x <- init_theta
  } else {
    start_x <- cache$theta
  }

  # try 1
  optim <- nlminb(
    start = start_x, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control,
    scale = optim_scale, lower = -0.995, upper = +0.995,
    cache = cache
  )

  # try 2
  if (optim$convergence != 0L) {
    optim <- nlminb(
      start = start_x, objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control,
      scale = optim_scale, lower = -0.995, upper = +0.995,
      cache = cache
    )
  }

  # try 3
  if (optim$convergence != 0L) {
    optim <- nlminb(
      start = 0, objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control,
      scale = 10, lower = -0.995, upper = +0.995,
      cache = cache
    )
  }

  # try 4 -- new in 0.6-8
  if (optim$convergence != 0L) {
    optim <- optimize(
      f = minfns$objective, interval = c(-0.995, +0.995),
      cache = cache, tol = .Machine$double.eps
    )
    if (is.finite(optim$minimum)) {
      optim$convergence <- 0L
      optim$par <- optim$minimum
    }
  }

  # check convergence
  if (optim$convergence != 0L) {
    if (!is.null(y1_name) && !is.null(y2_name)) {
      lav_msg_warn(gettextf(
        "estimation polyserial correlation did not converge
        for variables %1$s and %2$s", y1_name, y2_name))
    } else {
      lav_msg_warn(gettext(
        "estimation polyserial correlation(s) did not always converge"))
    }
    rho <- cache$theta # starting value
  } else {
    rho <- optim$par
  }

  rho
}

# build the cache from user-level arguments: fit the univariate models if
# needed, apply parameter overrides, and initialize the cache environment
# (shared by lav_bvmix_cor_twostep_fit/_cor_sc/_logl/_lik)
#
# Y1 = linear   (evar_y1/beta_y1 overrides)
# Y2 = ordinal  (th_y2/sl_y2 overrides)
lav_bvmix_cache_from_args <- function(y1, y2, exo = NULL, wt = NULL,
                                      rho = NULL,
                                      fit_y1 = NULL, fit_y2 = NULL,
                                      evar_y1 = NULL, beta_y1 = NULL,
                                      th_y2 = NULL, sl_y2 = NULL,
                                      scores = FALSE) {
  if (is.null(fit_y1)) {
    fit_y1 <- lav_uvreg_fit(y = y1, x = exo, wt = wt)
  }
  if (is.null(fit_y2)) {
    fit_y2 <- lav_uvord_fit(y = y2, x = exo, wt = wt)
  }

  # update parameters/z1/z2 if needed (used in lav_pml_dploglik_dimplied()
  #                                    in lav_model_gradient_pml.R)
  fit_y1 <- lav_uvreg_update_fit(
    fit_y = fit_y1, evar_new = evar_y1,
    beta_new = beta_y1
  )
  fit_y2 <- lav_uvord_update_fit(
    fit_y = fit_y2, th_new = th_y2,
    sl_new = sl_y2
  )

  cache <- lav_bvmix_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, wt = wt,
    scores = scores
  )
  if (!is.null(rho)) {
    cache$theta <- rho
  }

  cache
}

# Y1 = linear
# Y2 = ordinal
lav_bvmix_init_cache <- function(fit_y1 = NULL,
                                 fit_y2 = NULL,
                                 wt = NULL,
                                 scores = FALSE,
                                 parent = parent.frame()) {
  # data
  y1 <- fit_y1$y
  y2 <- fit_y2$y
  exo <- fit_y1$x

  # extract parameters

  # Y1
  y1_var <- fit_y1$theta[fit_y1$var_idx]
  y1_sd <- sqrt(y1_var)
  y1_eta <- fit_y1$yhat
  z <- (y1 - y1_eta) / y1_sd

  # Y2
  th_y2 <- fit_y2$theta[fit_y2$th_idx]

  # exo?
  if (is.null(exo)) {
    nexo <- 0L
  } else {
    nexo <- ncol(exo)
  }

  # nobs
  if (is.null(wt)) {
    n <- length(y1)
  } else {
    n <- sum(wt)
  }

  # starting value -- Olsson 1982 eq 38
  # (with covariates, based on the standardized residual z instead of y1)
  v1 <- if (nexo > 0L) z else y1
  if (is.null(wt)) {
    cor_1 <- cor(v1, y2, use = "pairwise.complete.obs")
    sd_1 <- sd(y2, na.rm = TRUE) * sqrt((n - 1) / n)
  } else {
    tmp <- na.omit(cbind(v1, y2, wt))
    cor_1 <- cov.wt(x = tmp[, 1:2], wt = tmp[, 3], cor = TRUE)$cor[2, 1]
    sd_1 <- sqrt(lav_mat_var_wt(tmp[, 2], wt = tmp[, 3]))
  }
  rho_init <- (cor_1 * sd_1 / sum(dnorm(th_y2)))

  # sanity check
  if (is.na(rho_init)) {
    rho_init <- 0.0
  } else if (abs(rho_init) > 0.9) {
    rho_init <- rho_init / 2
  }

  # parameter vector
  theta <- rho_init # only

  # cache environment
  cache_list <- list(
    nexo = nexo, theta = theta, wt = wt, n = n,
    y1 = y1, y1_var = y1_var, y1_sd = y1_sd, y1_eta = y1_eta, z = z,
    fit_y2_z1 = fit_y2$z1, fit_y2_z2 = fit_y2$z2
  )
  if (scores) {
    cache_list <- c(cache_list, list(
      exo = exo,
      y2_y1 = fit_y2$y1, y2_y2 = fit_y2$y2
    ))
  }

  list2env(cache_list, parent = parent)
}


# casewise likelihoods, unweighted!
lav_bvmix_lik_cache <- function(cache = NULL) {
  with(cache, {                # nolint start
    rho <- theta[1L]
    r <- sqrt(1 - rho * rho)

    # p(Y2|Y1)
    tauj_star <- (fit_y2_z1 - rho * z) / r
    tauj1_star <- (fit_y2_z2 - rho * z) / r
    py2y1 <- pnorm(tauj_star) - pnorm(tauj1_star)
    # TODO, check when to use 1 - pnorm()
    py2y1[py2y1 < .Machine$double.eps] <- .Machine$double.eps

    # p(Y1)
    py1 <- dnorm(y1, mean = y1_eta, sd = y1_sd)

    # lik
    lik <- py1 * py2y1

    # only catch likelihoods that have underflowed to (essentially) zero;
    # we must NOT drop merely-small (but valid) likelihoods: unlike the
    # ordinal/ordinal case, p(y1) here is a *density* that can be
    # legitimately tiny for outlying values of the continuous variable y1.
    # The old threshold sqrt(.Machine$double.eps) (~1.5e-8) discarded such
    # cases, and because *which* cases were dropped depended on rho, the
    # objective became non-smooth and the polyserial correlation was biased
    # (badly so for skewed y1), preventing convergence to the optimal
    # solution. This restores the pre-0.6-6 (lav_polyserial.R) behavior.
    lik_toosmall_idx <- which(!(lik > 0))
    lik[lik_toosmall_idx] <- as.numeric(NA)

    return(lik)
  })                                 # nolint end
}

lav_bvmix_logl_cache <- function(cache = NULL) {
  with(cache, {                      # nolint start
    lik <- lav_bvmix_lik_cache(cache) # unweighted!

    if (!is.null(wt)) {
      logl <- sum(wt * log(lik), na.rm = TRUE)
    } else {
      logl <- sum(log(lik), na.rm = TRUE)
    }

    return(logl)
  })                                # nolint end
}

lav_bvmix_grad_cache <- function(cache = NULL) {
  with(cache, {                     # nolint start
    rho <- theta[1L]

    y_z1 <- dnorm(tauj_star)
    y_z2 <- dnorm(tauj1_star)
    pyx_inv_r3 <- 1 / (py2y1 * r * r * r)

    # rho
    d1 <- fit_y2_z1 * rho - z
    d2 <- fit_y2_z2 * rho - z
    dx <- pyx_inv_r3 * (y_z1 * d1 - y_z2 * d2)

    # to be consistent with (log)lik_cache
    if (length(lik_toosmall_idx) > 0L) {
      dx[lik_toosmall_idx] <- as.numeric(NA)
    }

    if (is.null(wt)) {
      dx_rho <- sum(dx, na.rm = TRUE)
    } else {
      dx_rho <- sum(wt * dx, na.rm = TRUE)
    }

    return(dx_rho)
  })                                  # nolint end
}

# YR 29 March 2020
# obtained by using 'Deriv' (from package Deriv) on the
# gradient function, and cleaning up
# correct, but not good enough
lav_bvmix_hessian_cache <- function(cache = NULL) {
  with(cache, {                       # nolint start
    rho <- theta[1L]
    r2 <- r * r

    t1 <- z - rho * tauj_star / r
    t2 <- z - rho * tauj1_star / r

    tmp <- (y_z1 * (d1 * ((3 * rho / r2) + tauj_star * t1 / r)
      + fit_y2_z1 + dx * r2 * t1)

    - y_z2 * (d2 * ((3 * rho / r2) + tauj1_star * t2 / r)
        + fit_y2_z2 + dx * r2 * t2)
    )

    # to be consistent with (log)lik_cache
    if (length(lik_toosmall_idx) > 0L) {
      tmp[lik_toosmall_idx] <- as.numeric(NA)
    }

    if (is.null(wt)) {
      h <- sum(tmp * pyx_inv_r3, na.rm = TRUE)
    } else {
      h <- sum(wt * (tmp * pyx_inv_r3), na.rm = TRUE)
    }
    dim(h) <- c(1L, 1L) # for nlminb

    return(h)
  })                                    # nolint end
}

# nlminb objective/gradient/hessian (see lav_uvbv_common.R)
# note: this also fixes a dormant bug: the old lav_bvmix_min_hessian did
# not store 'x' in cache$theta before recomputing (never triggered: the
# default optim_method "nlminb1" does not use the hessian)
lav_bvmix_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_bvmix_logl_cache,
    grad_fun = lav_bvmix_grad_cache,
    hessian_fun = lav_bvmix_hessian_cache
  )
}


lav_bvmix_cor_sc_cache <- function(cache = NULL,
                                       sigma_correction = FALSE,
                                       na_zero = FALSE) {
  # make the arguments visible inside with(cache, ...): symbols there
  # resolve through the cache environment (and its parents), NOT through
  # this function's frame
  cache$sigma_correction <- sigma_correction
  cache$na_zero <- na_zero
  with(cache, {                           # nolint start
    rho <- theta[1L]
    r <- sqrt(1 - rho * rho)

    tauj_star <- (fit_y2_z1 - rho * z) / r
    tauj1_star <- (fit_y2_z2 - rho * z) / r
    y_z1 <- dnorm(tauj_star)
    y_z2 <- dnorm(tauj1_star)

    # p(Y2|Y1)
    py2y1 <- pnorm(tauj_star) - pnorm(tauj1_star)
    py2y1[py2y1 < .Machine$double.eps] <- .Machine$double.eps
    pyx_inv <- 1 / py2y1

    # mu.y1
    y_z1_y_z2 <- y_z1 - y_z2
    dx_mu_y1 <- 1 / y1_sd * (z + (pyx_inv * (rho / r) * y_z1_y_z2))
    if (!is.null(wt)) {
      dx_mu_y1 <- wt * dx_mu_y1
    }

    # var.y1
    dx_var_y1 <- 1 / (2 * y1_var) * (((z * z) - 1) +
      (pyx_inv * rho * z / r) * y_z1_y_z2)
    if (!is.null(wt)) {
      dx_var_y1 <- wt * dx_var_y1
    }

    # th.y2
    dx_th_y2 <- (y2_y1 * y_z1 - y2_y2 * y_z2) * 1 / r * pyx_inv
    if (!is.null(wt)) {
      dx_th_y2 <- wt * dx_th_y2
    }

    # sl.y1
    dx_sl_y1 <- NULL
    if (nexo > 0L) {
      dx_sl_y1 <- dx_mu_y1 * exo
      # if(!is.null(wt)) {
      # dx.mu.y1 had already been weighted
      # }
    }

    # sl.y2
    dx_sl_y2 <- NULL
    if (nexo > 0L) {
      dx_sl_y2 <- (y_z2 - y_z1) * exo * 1 / r * pyx_inv
      if (!is.null(wt)) {
        dx_sl_y2 <- wt * dx_sl_y2
      }
    }

    # rho
    tauj <- y_z1 * (fit_y2_z1 * rho - z)
    tauj1 <- y_z2 * (fit_y2_z2 * rho - z)
    dx_rho <- pyx_inv * 1 / (r * r * r) * (tauj - tauj1)
    if (!is.null(wt)) {
      dx_rho <- wt * dx_rho
    }

    # FIXME: only tested for non_exo!
    # used by lav_pml_dploglik_dimplied()
    if (sigma_correction) {
      dx_rho_orig <- dx_rho
      dx_var_y1_orig <- dx_var_y1

      # sigma
      dx_rho <- dx_rho_orig / y1_sd

      # var
      cov_1 <- rho * y1_sd
      dx_var_y1 <- (dx_var_y1_orig -
        1 / 2 * cov_1 / y1_var * 1 / y1_sd * dx_rho_orig)
    }

    out <- list(
      dx_mu_y1 = dx_mu_y1, dx_var_y1 = dx_var_y1,
      dx_th_y2 = dx_th_y2,
      dx_sl_y1 = dx_sl_y1, dx_sl_y2 = dx_sl_y2,
      dx_rho = dx_rho
    )

    return(out)
  })                                        # nolint end
}


# casewise scores
#
# Y1 = linear
# Y2 = ordinal
lav_bvmix_cor_sc <- function(y1, y2, exo = NULL, wt = NULL,
                                 rho = NULL,
                                 fit_y1 = NULL, fit_y2 = NULL,
                                 evar_y1 = NULL, beta_y1 = NULL,
                                 th_y2 = NULL, sl_y2 = NULL,
                                 sigma_correction = FALSE,
                                 na_zero = FALSE) {
  cache <- lav_bvmix_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    evar_y1 = evar_y1, beta_y1 = beta_y1,
    th_y2 = th_y2, sl_y2 = sl_y2,
    scores = TRUE
  )

  lav_bvmix_cor_sc_cache(
    cache = cache,
    sigma_correction = sigma_correction,
    na_zero = na_zero
  )
}

# logl - no cache
lav_bvmix_logl <- function(y1, y2, exo = NULL, wt = NULL,
                           rho = NULL,
                           fit_y1 = NULL, fit_y2 = NULL,
                           evar_y1 = NULL, beta_y1 = NULL,
                           th_y2 = NULL, sl_y2 = NULL) {
  cache <- lav_bvmix_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    evar_y1 = evar_y1, beta_y1 = beta_y1,
    th_y2 = th_y2, sl_y2 = sl_y2,
    scores = TRUE
  )

  lav_bvmix_logl_cache(cache = cache)
}

# lik - no cache
lav_bvmix_lik <- function(y1, y2, exo = NULL, wt = NULL,
                          rho = NULL,
                          fit_y1 = NULL, fit_y2 = NULL,
                          evar_y1 = NULL, beta_y1 = NULL,
                          th_y2 = NULL, sl_y2 = NULL,
                          take_log = FALSE) {
  cache <- lav_bvmix_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    evar_y1 = evar_y1, beta_y1 = beta_y1,
    th_y2 = th_y2, sl_y2 = sl_y2,
    scores = TRUE
  )

  lik <- lav_bvmix_lik_cache(cache = cache) # unweighted
  if (take_log) {
    lik <- log(lik)
  }

  if (!is.null(wt)) {
    if (take_log) {
      lik <- wt * lik
    } else {
      tmp <- wt * log(lik)
      lik <- exp(tmp)
    }
  }

  lik
}

# bootstrap based NVCOV
lav_model_nvcov_bootstrap <- function(lavmodel = NULL,
                                      lavsamplestats = NULL,
                                      lavoptions = NULL,
                                      lavimplied = NULL,
                                      lavh1 = NULL,
                                      lavdata = NULL,
                                      lavcache = NULL,
                                      lavpartable = NULL) {
  # number of bootstrap draws
  if (!is.null(lavoptions$bootstrap) &&
      !is.null(lavoptions$bootstrap$R)) {
    r <- as.integer(lavoptions$bootstrap$R)
    stopifnot(r > 0L)
  } else {
    r <- 1000L
  }

  boot_type <- "ordinary"
  if ("bollen.stine" %in% lavoptions$test) {
    boot_type <- "bollen.stine"
  }

  test <- NULL
  coef_1 <- lav_bootstrap_internal(
    object = NULL,
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavpartable = lavpartable,
    lavoptions = lavoptions,
    lavdata = lavdata,
    r = r,
    check_post = lavoptions$check.post,
    type = boot_type,
    fun = ifelse(boot_type == "bollen.stine",
      "coeftest", "coef"
    )
  )
  # warn            = -1L)
  coef_orig <- coef_1

  # new in 0.6-12: always warn for failed and nonadmissible
  error_idx <- attr(coef_1, "error.idx")
  nfailed <- length(error_idx) # zero if NULL
  if (nfailed > 0L) {
    lav_msg_warn(gettextf(
      "%s bootstrap runs failed or did not converge.", nfailed))
  }

  notok <- length(attr(coef_1, "nonadmissible")) # zero if NULL
  if (notok > 0L) {
    lav_msg_warn(gettextf(
      "%s bootstrap runs resulted in nonadmissible solutions.", notok))
  }

  if (length(error_idx) > 0L) {
    # new in 0.6-13: we must still remove them!
    coef_1 <- coef_1[-error_idx, , drop = FALSE]
    # this also drops the attributes
  }

  if (boot_type == "bollen.stine") {
    nc <- ncol(coef_1)
    test <- coef_1[, nc]
    coef_1 <- coef_1[, -nc, drop = FALSE]
    # the last column is the (bollen.stine) test statistic, not a
    # coefficient; remove it from BOOT.COEF too (it is stored separately
    # in BOOT.TEST), but preserve the attributes (error.idx, nonadmissible,
    # seed, ...) that subsetting would otherwise drop
    keep_attr <- attributes(coef_orig)
    keep_attr[c("dim", "dimnames")] <- NULL
    coef_orig <- coef_orig[, -nc, drop = FALSE]
    attributes(coef_orig) <- c(attributes(coef_orig), keep_attr)
  }

  # new in 0.6-20: check for outliers, ie big difference between sd() and mad()
  # see github issue 347
  sd_mad_ratio <- (apply(coef_1, 2,  sd, na.rm = TRUE) /
                   apply(coef_1, 2, mad, na.rm = TRUE))
  crit_ratio <- 5
  if (any(sd_mad_ratio > crit_ratio)) {
    names_1 <- lav_pt_labels(lavpartable, type = "free")
    params_w_outliers <- paste(names_1[sd_mad_ratio > crit_ratio],
                               collapse = " ")
    lav_msg_warn(gettextf(
      "The following bootstrapped free parameters have a high (>5)
      ratio of standard deviation to median absolute deviation: %s.
      P-values and confidence intervals may not match.", params_w_outliers))
  }

  # FIXME: cov rescale? Yes for now
  nboot <- nrow(coef_1)
  # The bootstrap directly estimates Cov(coef), which IS the target VCOV.
  # lav_model_vcov() will afterwards divide this NVCOV by 'n' to obtain the
  # VCOV, so we scale by the same 'n' here, making the two cancel. For
  # multilevel models, lavaan divides by the (total) number of clusters --
  # not by ntotal -- so we must use the same denominator here (otherwise the
  # bootstrap standard errors would be off by a factor sqrt(ntotal/nclusters)).
  # This mirrors the 'n' computation in lav_model_vcov().
  likelihood <- lavoptions$likelihood
  if (lavdata@nlevels > 1L &&
      lavmodel@estimator %in% c("ML", "PML", "FML") &&
      !is.null(likelihood) && likelihood == "normal") {
    n <- 0
    for (g in seq_len(lavsamplestats@ngroups)) {
      n <- n + lavdata@Lp[[g]]$nclusters[[2]]
    }
  } else {
    n <- lavsamplestats@ntotal
  }
  nvar_cov <- n * (cov(coef_1) * (nboot - 1) / nboot)

  # ceq.simple.only models (e.g. estimator = "IV" with equality constraints):
  # the bootstrap draws are in the compact (free) space (one column per
  # non-collapsed free parameter), but the rest of the vcov machinery
  # (lav_model_vcov_se) expects the 'unco' space (one row/col per parameter,
  # with shared rows/cols for parameters constrained equal). Expand via
  # ceq.simple.K so that, e.g., loadings constrained equal get identical
  # standard errors. The BOOT.COEF stored below stays in the compact space
  # (the bootstrap CI code in lavParameterEstimates() expects that).
  if (lavmodel@ceq.simple.only && nrow(lavmodel@ceq.simple.K) > 0L &&
      ncol(nvar_cov) == ncol(lavmodel@ceq.simple.K)) {
    tmp_k <- lavmodel@ceq.simple.K
    nvar_cov <- tmp_k %*% nvar_cov %*% t(tmp_k)
  }

  # save COEF and TEST (if any)
  attr(nvar_cov, "BOOT.COEF") <- coef_orig # including attributes
  attr(nvar_cov, "BOOT.TEST") <- test

  nvar_cov
}


# robust `sem' NVCOV (see Browne, 1984,  bentler & dijkstra 1985)
lav_model_nvcov_robust_sem <- function(lavmodel = NULL,
                                       lavsamplestats = NULL,
                                       lavdata = NULL,
                                       lavcache = NULL,
                                       lavimplied = NULL,
                                       lavh1 = NULL,
                                       lavoptions = NULL,
                                       use_ginv = FALSE,
                                       attr_delta = TRUE,
                                       attr_t_dvgvd = FALSE,
                                       attr_e_inv = FALSE,
                                       attr_wls_v = FALSE) {
  # compute inverse of the expected(!) information matrix
  if (lavmodel@estimator == "ML" && lavoptions$information.expected.mplus) {
    # YR - 11 aug 2010 - what Mplus seems to do is (see Muthen apx 4 eq102)
    # - A1 is not based on Sigma.hat and Mu.hat,
    # but on lavsamplestats@cov and lavsamplestats@mean... ('unstructured')
    # - gamma is not identical to what is used for WLS; closer to EQS
    # - N/N-1 bug in G11 for NVarCov (but not test statistic)
    # - we divide by N-1! (just like EQS)
    e_inv <- lav_model_info_expected_mlm(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      extra = TRUE,
      augmented = TRUE,
      inverted = TRUE,
      use_ginv = use_ginv
    )
  } else {
    e_inv <- lav_model_info(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavoptions = lavoptions,
      extra = TRUE,
      augmented = TRUE,
      inverted = TRUE,
      use_ginv = use_ginv
    )
  }

  # check if E.inv is ok
  if (inherits(e_inv, "try-error")) {
    return(e_inv)
  }

  delta <- attr(e_inv, "Delta")
  wls_v <- attr(e_inv, "WLS.V")
  attr(e_inv, "Delta") <- NULL
  attr(e_inv, "WLS.V") <- NULL
  # gamma
  gamma <- lavsamplestats@NACOV
  if (lavmodel@estimator == "ML" &&
    lavoptions$gamma.vcov.mplus && !lavsamplestats@NACOV.user) {
    # 'fix' G11 part of gamma (NOTE: this is NOT needed for SB test
    # statistic
    for (g in 1:lavsamplestats@ngroups) {
      gg1 <- (lavsamplestats@nobs[[g]] - 1) / lavsamplestats@nobs[[g]]
      if (lavmodel@conditional.x) {
        nvar <- NCOL(lavsamplestats@res.cov[[g]])
      } else {
        nvar <- NCOL(lavsamplestats@cov[[g]])
      }
      g11 <- gamma[[g]][1:nvar, 1:nvar, drop = FALSE]
      gamma[[g]][1:nvar, 1:nvar] <- g11 * gg1
    } # g
  }


  t_dvgvd <- matrix(0, ncol = ncol(e_inv), nrow = nrow(e_inv))
  for (g in 1:lavsamplestats@ngroups) {
    fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
    if (lavoptions$gamma.vcov.mplus) {
      fg1 <- (lavsamplestats@nobs[[g]] - 1) / lavsamplestats@ntotal
    } else {
      # from 0.6 onwards, we use fg1 == fg, to be more consistent with
      # lav_test()
      fg1 <- fg
    }
    # fg twice for WLS.V, 1/fg1 once for GaMMA
    # if fg==fg1, there would be only one fg, as in Satorra 1999 p.8
    # t(Delta) * WLS.V %*% gamma %*% WLS.V %*% Delta
    if (lavmodel@estimator == "DWLS" || lavmodel@estimator == "ULS") {
      # diagonal weight matrix
      wd <- wls_v[[g]] * delta[[g]]
    } else {
      # full weight matrix
      wd <- wls_v[[g]] %*% delta[[g]]
    }
    t_dvgvd <- t_dvgvd + fg * fg / fg1 * crossprod(wd, gamma[[g]] %*% wd)
  } # g
  nvar_cov <- (e_inv %*% t_dvgvd %*% e_inv)

  # to be reused by lav_test()
  if (attr_delta) {
    attr(nvar_cov, "Delta") <- delta
  }
  # for twostep.robust in sam()
  if (attr_t_dvgvd) {
    attr(nvar_cov, "tDVGVD") <- t_dvgvd
  }

  if ((lavoptions$information[1] == lavoptions$information[2]) &&
    (lavoptions$h1.information[1] == lavoptions$h1.information[2]) &&
    (lavoptions$information[2] == "expected" ||
      lavoptions$observed.information[1] ==
        lavoptions$observed.information[2])) {
    # only when same type of information is used # new in 0.6-6
    attr(nvar_cov, "E.inv") <- e_inv
    attr(nvar_cov, "WLS.V") <- wls_v
  }

  # user override
  if (attr_e_inv && is.null(attr(nvar_cov, "E.inv"))) {
    attr(nvar_cov, "E.inv") <- e_inv
  }
  if (attr_wls_v && is.null(attr(nvar_cov, "WLS.V"))) {
    attr(nvar_cov, "WLS.V") <- wls_v
  }

  nvar_cov
}

lav_model_nvcov_robust_sandwich <- function(lavmodel = NULL,    # nolint
                                            lavsamplestats = NULL,
                                            lavdata = NULL,
                                            lavoptions = NULL,
                                            lavimplied = NULL,
                                            lavh1 = NULL,
                                            lavcache = NULL,
                                            use_ginv = FALSE) {
  # sandwich estimator: A.inv %*% B %*% t(A.inv)
  # where A.inv == E.inv
  #       B == outer product of case-wise scores

  # inverse observed/expected information matrix
  e_inv <- lav_model_info(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    lavcache = lavcache,
    lavimplied = lavimplied,
    lavh1 = lavh1,
    lavoptions = lavoptions,
    extra = FALSE,
    augmented = TRUE,
    inverted = TRUE,
    use_ginv = use_ginv
  )

  # check if E.inv is ok
  if (inherits(e_inv, "try-error")) {
    return(e_inv)
  }

  # new in 0.6-6, check for h1.information.meat
  lavoptions2 <- lavoptions
  if (!is.null(lavoptions$information.meat)) {
    lavoptions2$information <- lavoptions$information.meat
  }
  if (!is.null(lavoptions$h1.information.meat)) {
    lavoptions2$h1.information <- lavoptions$h1.information.meat
  }

  # outer product of case-wise scores
  b0 <-
    lav_model_info_firstorder(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavoptions = lavoptions2,
      extra = TRUE,
      check_pd = FALSE,
      augmented = FALSE,
      inverted = FALSE,
      use_ginv = use_ginv
    )

  # compute sandwich estimator
  nvar_cov <- e_inv %*% b0 %*% e_inv

  attr(nvar_cov, "B0.group") <- attr(b0, "B0.group")

  if ((lavoptions$information[1] == lavoptions$information[2]) &&
    (lavoptions$h1.information[1] == lavoptions$h1.information[2]) &&
    (lavoptions$information[2] == "expected" ||
      lavoptions$observed.information[1] ==
        lavoptions$observed.information[2])) {
    # only when same type of information is used # new in 0.6-6
    attr(nvar_cov, "E.inv") <- e_inv
  }

  nvar_cov
}

# two stage
# - two.stage: gamma = I_1^{-1}
# - robust.two.stage: gamma = incomplete gamma (I_1^{-1} J_1 I_1^{-1})
# where I_1 and J_1 are based on the (saturated) model h1
# (either unstructured, or structured)
#
# references:
#
# - Savalei \& Bentler (2009) eq (6) for se = "two.stage"
# - Savalei \& Falk (2014) eq  (3)   for se = "robust.two.stage"
# - Yuan \& Bentler (2000)
lav_model_nvcov_two_stage <- function(lavmodel = NULL,
                                      lavsamplestats = NULL,
                                      lavoptions = NULL,
                                      lavimplied = NULL,
                                      lavh1 = NULL,
                                      lavdata = NULL,
                                      use_ginv = FALSE) {
  # expected OR observed, depending on lavoptions$information
  if (is.null(lavoptions) && is.null(lavoptions$information[1])) {
    lavoptions <- list(
      information = "observed",
      observed.information = "h1",
      h1.information = "structured"
    )
  }

  # restrictions:
  # only works if:
  # - information is expected,
  # - or information is observed but with observed.information == "h1"
  if (lavoptions$information[1] == "observed" &&
    lavoptions$observed.information[1] != "h1") {
    lav_msg_stop(
      gettext("two.stage + observed information currently only works
              with observed.information = 'h1'"))
  }
  # no weights (yet)
  if (!is.null(lavdata@weights[[1]])) {
    lav_msg_stop(gettext("two.stage + sampling.weights is not supported yet"))
  }
  # no fixed.x (yet)
  # if(!is.null(lavsamplestats@x.idx) &&
  #   length(lavsamplestats@x.idx[[1]]) > 0L) {
  #    lav_msg_stop(gettext("two.stage + fixed.x = TRUE is not supported yet"))
  # }


  # information matrix
  e_inv <- lav_model_info(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    lavoptions = lavoptions,
    lavimplied = lavimplied,
    lavh1 = lavh1,
    extra = TRUE,
    augmented = TRUE,
    inverted = TRUE,
    use_ginv = use_ginv
  )
  delta <- attr(e_inv, "Delta")
  wls_v <- attr(e_inv, "WLS.V") # this is 'H' or 'A1' in the literature
  attr(e_inv, "Delta") <- NULL
  attr(e_inv, "WLS.V") <- NULL

  # check if E.inv is ok
  if (inherits(e_inv, "try-error")) {
    return(e_inv)
  }

  # check WLS.V = A1
  if (is.null(wls_v)) {
    lav_msg_stop(gettext("WLS.V/H/A1 is NULL, observed.information = hessian?"))
  }

  # gamma
  gamma <- vector("list", length = lavsamplestats@ngroups)

  # handle multiple groups
  t_dvgvd <- matrix(0, ncol = ncol(e_inv), nrow = nrow(e_inv))
  for (g in 1:lavsamplestats@ngroups) {
    fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
    # fg1 <- (lavsamplestats@nobs[[g]]-1)/lavsamplestats@ntotal
    fg1 <- fg
    # fg twice for WLS.V, 1/fg1 once for GaMMA
    # if fg==fg1, there would be only one fg, as in Satorra 1999 p.8
    # t(Delta) * WLS.V %*% gamma %*% WLS.V %*% Delta
    if (lavmodel@estimator == "DWLS" || lavmodel@estimator == "ULS") {
      # diagonal weight matrix (stored as a vector)
      wd <- wls_v[[g]] * delta[[g]]
    } else {
      # full weight matrix
      wd <- wls_v[[g]] %*% delta[[g]]
    }

    # conditional.x: the stage-1 'saturated' parameters live in the
    # conditional metric (vec(Beta), vech(res.cov)); compute Omega with
    # the lav_mvreg_mi_* kernels (x is complete). Note: with complete x,
    # the joint information is block-diagonal between the conditional
    # parameters and the x moments, so the conditional Omega is the
    # correct stage-1 ACOV.
    if (lavmodel@conditional.x) {
      aux_g2 <- if (length(lavdata@aux) >= g) lavdata@aux[[g]] else NULL
      if (!is.null(aux_g2) && NCOL(aux_g2) > 0L) {
        lav_msg_stop(gettext(
          "auxiliary variables (aux =) are not supported (yet) for
          two-stage standard errors when conditional.x = TRUE."))
      }
      if (lavoptions$h1.information[1] == "unstructured") {
        res_int_g <- lavh1$implied$res.int[[g]]
        res_slopes_g <- lavh1$implied$res.slopes[[g]]
        res_cov_g <- lavh1$implied$res.cov[[g]]
      } else {
        res_int_g <- lavimplied$res.int[[g]]
        res_slopes_g <- lavimplied$res.slopes[[g]]
        res_cov_g <- lavimplied$res.cov[[g]]
      }

      if (lavoptions$se == "two.stage") {
        # Savalei & Bentler (2009), conditional metric
        if (lavoptions$information[1] == "expected") {
          info <- lav_mvreg_mi_info_expected(
            yp = lavsamplestats@missing[[g]],
            res_cov = res_cov_g
          )
        } else {
          info <- lav_mvreg_mi_information_observed_samplestats(
            yp = lavsamplestats@missing[[g]],
            res_int = res_int_g, res_slopes = res_slopes_g,
            res_cov = res_cov_g
          )
        }
        omega_g <- lav_mat_sym_inverse(info)
      } else { # robust.two.stage
        # Savalei & Falk (2014), conditional metric
        if (length(lavdata@cluster) > 0L) {
          cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
        } else {
          cluster_idx <- NULL
        }
        omega_g <- lav_mvreg_mi_h1_omega_sw(
          y = lavdata@X[[g]],
          exo = lavdata@eXo[[g]],
          mp = lavdata@Mp[[g]],
          yp = lavsamplestats@missing[[g]],
          wt = lavdata@weights[[g]],
          cluster_idx = cluster_idx,
          res_int = res_int_g, res_slopes = res_slopes_g,
          res_cov = res_cov_g,
          information = lavoptions$information[1]
        )
      }
      gamma[[g]] <- omega_g

      # compute
      t_dvgvd <- t_dvgvd + fg * fg / fg1 * crossprod(wd, gamma[[g]] %*% wd)
      next
    }

    # to compute (incomplete) GAMMA, should we use
    # structured or unstructured mean/sigma?
    #
    # we use the same setting as to compute 'H' (the h1 information matrix)
    # so that at Omega = H if data is complete
    if (lavoptions$h1.information[1] == "unstructured") {
      #MU <- lavsamplestats@missing.h1[[g]]$mu
      #SIGMA <- lavsamplestats@missing.h1[[g]]$sigma
      mu <- lavh1$implied$mean[[g]]
      sigma_1 <- lavh1$implied$cov[[g]]
    } else {
      mu <- lavimplied$mean[[g]]
      sigma_1 <- lavimplied$cov[[g]]
    }

    # auxiliary variables (aux=): with two-stage estimation, the stage-1
    # saturated moments are estimated over the augmented set [model vars, aux],
    # so the stage-1 ACOV (Omega) must be computed over that augmented set and
    # then reduced to the model-variable moments (Savalei & Bentler, 2009).
    # We only do this when there are no fixed-x covariates (the common case).
    aux_g <- if (length(lavdata@aux) >= g) lavdata@aux[[g]] else NULL
    use_aux <- !is.null(aux_g) && NCOL(aux_g) > 0L &&
      length(lavsamplestats@x.idx[[g]]) == 0L
    if (use_aux) {
      y_g <- cbind(lavdata@X[[g]], aux_g)
      mp_g <- lav_data_mi_patterns(y_g)
      mu_g <- lavsamplestats@missing.h1[[g]]$mu.aug
      sigma_g <- lavsamplestats@missing.h1[[g]]$sigma.aug
      if (is.null(mu_g) || is.null(sigma_g)) {
        em_aug <- lav_mvn_mi_h1_est_moments(y_g,
          mp = mp_g, max_iter = lavoptions$em.h1.args$max_iter,
          tol = lavoptions$em.h1.args$tol,
          non_pd_action = lavoptions$em.h1.args$non_pd_action,
          non_pd_tol = lavoptions$em.h1.args$non_pd_tol,
          acceleration = lavoptions$em.h1.args$acceleration
        )
        mu_g <- em_aug$Mu
        sigma_g <- em_aug$Sigma
      }
      midx <- lav_aux_moment_idx(
        p_model = NCOL(lavdata@X[[g]]), p_aug = NCOL(y_g)
      )
    } else {
      y_g <- lavdata@X[[g]]
      mp_g <- lavdata@Mp[[g]]
      mu_g <- mu
      sigma_g <- sigma_1
      midx <- NULL
    }

    # compute 'gamma' (or Omega.beta)
    if (lavoptions$se == "two.stage") {
      # this is Savalei & Bentler (2009)
      if (lavoptions$information[1] == "expected") {
        info <- lav_mvn_mi_info_expected(
          y = y_g, mp = mp_g,
          wt = lavdata@weights[[g]],
          mu = mu_g, sigma_1 = sigma_g,
          x_idx = lavsamplestats@x.idx[[g]]
        )
      } else {
        info <- lav_mvnorm_missing_information_observed_samplestats(
          yp = if (use_aux) {
            lav_samp_mi_patterns(y = y_g, mp = mp_g)
          } else {
            lavsamplestats@missing[[g]]
          },
          # wt not needed
          mu = mu_g, sigma_1 = sigma_g,
          x_idx = lavsamplestats@x.idx[[g]]
        )
      }
      omega_g <- lav_mat_sym_inverse(info)
    } else { # we assume "robust.two.stage"
      # NACOV is here incomplete gamma
      # Savalei & Falk (2014)
      #
      if (length(lavdata@cluster) > 0L) {
        cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
      } else {
        cluster_idx <- NULL
      }
      omega_g <- lav_mvn_mi_h1_omega_sw(
        y = y_g,
        mp = mp_g,
        yp = if (use_aux) {
          lav_samp_mi_patterns(y = y_g, mp = mp_g)
        } else {
          lavsamplestats@missing[[g]]
        },
        wt = lavdata@weights[[g]],
        cluster_idx = cluster_idx,
        mu = mu_g, sigma_1 = sigma_g,
        x_idx = lavsamplestats@x.idx[[g]],
        information = lavoptions$information[1]
      )
    }

    # auxiliary variables: reduce the augmented Omega to the model-variable
    # moments (sub-matrix of the inverse information)
    if (use_aux) {
      gamma[[g]] <- omega_g[midx, midx, drop = FALSE]
    } else {
      gamma[[g]] <- omega_g
    }

    # compute
    t_dvgvd <- t_dvgvd + fg * fg / fg1 * crossprod(wd, gamma[[g]] %*% wd)
  } # g

  nvar_cov <- (e_inv %*% t_dvgvd %*% e_inv)

  # to be reused by lavaanTest
  attr(nvar_cov, "Delta") <- delta
  attr(nvar_cov, "Gamma") <- gamma
  if ((lavoptions$information[1] == lavoptions$information[2]) &&
    (lavoptions$h1.information[1] == lavoptions$h1.information[2]) &&
    (lavoptions$information[2] == "expected" ||
      lavoptions$observed.information[1] ==
        lavoptions$observed.information[2])) {
    # only when same type of information is used # new in 0.6-6
    attr(nvar_cov, "E.inv") <- e_inv
    attr(nvar_cov, "WLS.V") <- wls_v
  }

  nvar_cov
}



lav_model_vcov <- function(lavmodel = NULL,
                           lavsamplestats = NULL,
                           lavoptions = NULL,
                           lavdata = NULL,
                           lavpartable = NULL,
                           lavcache = NULL,
                           lavimplied = NULL,
                           lavh1 = NULL,
                           use_ginv = FALSE) {
  likelihood <- lavoptions$likelihood
#  information <- lavoptions$information[1] # first one is for vcov
  se <- lavoptions$se
#  mimic <- lavoptions$mimic

  # special cases
  if (se == "none" || se == "external" || se == "twostep") {
    return(matrix(0, 0, 0))
  }

  if (se == "standard") {
    nvar_cov <- lav_model_info(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavoptions = lavoptions,
      extra = FALSE,
      augmented = TRUE,
      inverted = TRUE,
      use_ginv = use_ginv
    )
  } else if (se == "first.order") {
    nvar_cov <-
      lav_model_info_firstorder(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        lavcache = lavcache,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        lavoptions = lavoptions,
        extra = TRUE,
        check_pd = FALSE,
        augmented = TRUE,
        inverted = TRUE,
        use_ginv = use_ginv
      )
  } else if (se %in% c("robust.sem", "robust.sem.nt", "robust.cluster.sem")) {
    nvar_cov <-
      lav_model_nvcov_robust_sem(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavcache = lavcache,
        lavdata = lavdata,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        lavoptions = lavoptions,
        use_ginv = use_ginv
      )
  } else if (se == "robust.huber.white" || se == "robust.cluster") {
    nvar_cov <-
      lav_model_nvcov_robust_sandwich(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        lavcache = lavcache,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        lavoptions = lavoptions,
        use_ginv = use_ginv
      )
  } else if (se %in% c("two.stage", "robust.two.stage")) {
    nvar_cov <-
      lav_model_nvcov_two_stage(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavoptions = lavoptions,
        lavdata = lavdata,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        use_ginv = use_ginv
      )
  } else if (se == "bootstrap") {
    nvar_cov <- try(
      lav_model_nvcov_bootstrap(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavoptions = lavoptions,
        lavdata = lavdata,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        lavcache = lavcache,
        lavpartable = lavpartable
      ),
      silent = TRUE
    )
  } else {
    lav_msg_warn(gettextf("unknown se type: %s", se))
  }

  if (!inherits(nvar_cov, "try-error")) {
    # denominator!
    if (lavmodel@estimator %in% c("ML", "PML", "FML") &&
      likelihood == "normal") {
      if (lavdata@nlevels == 1L) {
        n <- lavsamplestats@ntotal
        # new in 0.6-9 (to mimic method="lm" in effectLite)
        # special case: univariate regression in each group
        if (lavoptions$mimic == "lm" &&
          all(lavmodel@modprop$uvreg)) {
          n <- sum(unlist(lavsamplestats@nobs) -
            (unlist(lavmodel@modprop$nexo) + 1L))
          # always adding the intercept (for now)
        }
      } else {
        # total number of clusters (over groups)
        n <- 0
        for (g in 1:lavsamplestats@ngroups) {
          n <- n + lavdata@Lp[[g]]$nclusters[[2]]
        }
      }
    } else {
      n <- lavsamplestats@ntotal - lavsamplestats@ngroups
    }

    var_cov <- 1 / n * nvar_cov

    # check if VarCov is pd -- new in 0.6-2
    # mostly important if we have (in)equality constraints (MASS::ginv!)
    if (lavmodel@ceq.simple.only) {
      # do nothing
    } else if (!is.null(lavoptions$check.vcov) && lavoptions$check.vcov) {
      eigvals <- eigen(var_cov,
        symmetric = TRUE,
        only.values = TRUE
      )$values
      # correct for (in)equality constraints: the effective number of
      # imposed constraints is the RANK of the stacked equality + active
      # inequality rows (a plain row count over-corrects when rows are
      # redundant, within or across the two sets)
      neiq <- 0L
      if (nrow(lavmodel@con.jac) > 0L) {
        ceq_idx <- attr(lavmodel@con.jac, "ceq.idx")
        cin_idx <- attr(lavmodel@con.jac, "cin.idx")
        ina_idx <- attr(lavmodel@con.jac, "inactive.idx")
        act_idx <- c(ceq_idx, setdiff(cin_idx, ina_idx)) # only active cin
        if (length(act_idx) > 0L) {
          neiq <- qr(lavmodel@con.jac[act_idx, , drop = FALSE])$rank
        }
        if (neiq > 0L) {
          eigvals <- rev(eigvals)[-seq_len(neiq)]
        }
      }
      min_val <- min(eigvals)
      # if(any(eigvals < -1 * sqrt(.Machine$double.eps)) &&
      if (min_val < .Machine$double.eps^(3 / 4)) {
        # VarCov.chol <- suppressWarnings(try(chol(VarCov,
        #                   pivot = TRUE), silent = TRUE))
        # VarCov.rank <- attr(VarCov.chol, "rank")
        # VarCov.pivot <- attr(VarCov.chol, "pivot")
        # VarCov.badidx <- VarCov.pivot[ VarCov.rank + 1L ]
        # pt.idx <- which(lavpartable$free == VarCov.badidx)
        # par.string <- paste(lavpartable$lhs[pt.idx],
        #                    lavpartable$op[ pt.idx],
        #                    lavpartable$rhs[pt.idx])
        # if(lavdata@ngroups > 1L) {
        #    par.string <- paste0(par.string, " in group ",
        #                         lavpartable$group[pt.idx])
        # }
        # if(lavdata@nlevels > 1L) {
        #    par.string <- paste0(par.string, " in level ",
        #                         lavpartable$level[pt.idx])
        # }

        if (min_val > 0) {
          lav_msg_warn(
            gettextf("The variance-covariance matrix of the estimated
                    parameters (vcov) does not appear to be positive
                    definite! The smallest eigenvalue (= %e) is close
                     to zero. This may be a symptom that the model is
                     not identified.", min(min_val)))
        } else {
          lav_msg_warn(
            gettextf("The variance-covariance matrix of the estimated parameters
                  (vcov) does not appear to be positive definite! The smallest
                  eigenvalue (= %e) is smaller than zero. This may be a
                  symptom that the model is not identified.",
                  min(min_val)))
        }
      }
    }
  } else {
    lav_msg_warn(
     gettext("Could not compute standard errors! The information matrix
             could not be inverted. This may be a symptom that the model
             is not identified.")
    )
    var_cov <- NULL
  } # could not invert

  var_cov
}

# Generate Monte Carlo draws for the free parameters from a multivariate
# normal distribution with mean = coef_hat (the point estimate) and
# covariance = VCOV. Used for Preacher & Selig (2012)-style Monte Carlo
# confidence intervals for defined parameters.
#
# Returns an R x n_free matrix. The number of draws and (optional) seed
# are taken from lavoptions$monte.carlo.
lav_model_vcov_mc <- function(lavmodel = NULL, vcov = NULL,
                              lavoptions = NULL) {
  if (is.null(vcov) || inherits(vcov, "try-error")) {
    return(NULL)
  }
  r <- 20000L
  seed <- NULL
  if (!is.null(lavoptions) && !is.null(lavoptions$monte.carlo)) {
    if (!is.null(lavoptions$monte.carlo$R)) {
      r <- as.integer(lavoptions$monte.carlo$R)
    }
    if (!is.null(lavoptions$monte.carlo$seed)) {
      seed <- lavoptions$monte.carlo$seed
    }
  }
  stopifnot(r > 0L)

  coef_hat <- lav_model_get_parameters(lavmodel = lavmodel, type = "free")
  # ensure we sample in the dimension of vcov: typically n_free, but
  # with simple equality constraints (ceq.simple.only) vcov is in unco
  # space; in that case lift coef_hat via t(K)
  if (lavmodel@ceq.simple.only && nrow(vcov) == nrow(lavmodel@ceq.simple.K)) {
    coef_hat <- drop(lavmodel@ceq.simple.K %*% coef_hat)
  }

  # save and restore RNG state so mc sampling does not perturb the
  # user's global seed
  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    saved_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    on.exit(assign(".Random.seed", saved_seed, envir = .GlobalEnv), add = TRUE)
  } else {
    on.exit({
      if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
        rm(".Random.seed", envir = .GlobalEnv)
      }
    }, add = TRUE)
  }
  if (!is.null(seed)) {
    set.seed(seed)
  }

  # symmetrize vcov for safety
  vcov_sym <- 0.5 * (vcov + t(vcov))
  mc_coef <- lav_mvrnorm(n = r, mu = coef_hat, sigma_1 = vcov_sym,
                         check_symmetry = FALSE)
  if (!is.matrix(mc_coef)) {
    mc_coef <- matrix(mc_coef, nrow = r)
  }
  mc_coef
}

# Is the Monte Carlo method requested for defined parameters?
lav_model_vcov_se_mc_active <- function(lavoptions) {
  isTRUE(!is.null(lavoptions) &&
    identical(lavoptions$se.def, "monte.carlo"))
}

# Reduce a vcov matrix from the 'unco' space (one row/column per
# non-collapsed free parameter, nx.unco) to the 'compact' space (one
# row/column per free parameter, nx.free) for ceq.simple.only models. This is
# needed whenever the vcov must be combined with a jacobian/hessian that is
# expressed in the compact (nx.free) space (e.g. defined parameters, Wald
# tests). vcov is returned unchanged if it is not in the unco space.
lav_model_vcov_unco_to_free <- function(lavmodel, vcov) {
  if (lavmodel@ceq.simple.only && nrow(lavmodel@ceq.simple.K) > 0L &&
      nrow(vcov) == nrow(lavmodel@ceq.simple.K)) {
    rep_idx <- apply(lavmodel@ceq.simple.K, 2L,
                     function(col) which(col != 0)[1L])
    vcov <- vcov[rep_idx, rep_idx, drop = FALSE]
  }
  vcov
}

lav_model_vcov_se <- function(lavmodel, lavpartable, vcov = NULL,
                              boot = NULL, mc = NULL,
                              lavoptions = NULL, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # 0. special case
  if (is.null(vcov)) {
    se <- rep(NA_real_, lavmodel@nx.user)
    se[lavpartable$free == 0L] <- 0.0
    return(se)
  }

  # 1. free parameters only
  x_var <- diag(vcov)
  # check for negative values (what to do: NA or 0.0?)
  x_var[x_var < 0] <- as.numeric(NA)
  x_se <- sqrt(x_var)
  if (lavmodel@ceq.simple.only) {
    glist <- lav_model_x2glist(
      lavmodel = lavmodel, x = x_se,
      type = "unco"
    )
  } else {
    glist <- lav_model_x2glist(
      lavmodel = lavmodel, x = x_se,
      type = "free"
    )
  }

  # se for full parameter table, but with 0.0 entries for def/ceq/cin
  # elements
  se <- lav_model_get_parameters(
    lavmodel = lavmodel, glist = glist,
    type = "user", extra = FALSE
  )


  # 2. fixed parameters -> se = 0.0
  se[which(lavpartable$free == 0L)] <- 0.0


  # 3. defined parameters:
  def_idx <- which(lavpartable$op == ":=")
  if (length(def_idx) > 0L) {
    # if Monte Carlo samples are not supplied but requested, draw them
    if (is.null(mc) && is.null(boot) &&
        lav_model_vcov_se_mc_active(lavoptions)) {
      mc <- lav_model_vcov_mc(lavmodel = lavmodel, vcov = vcov,
                              lavoptions = lavoptions)
    }
    if (!is.null(mc)) {
      mc_def <- apply(mc, 1L, lavmodel@def.function)
      if (length(def_idx) == 1L) {
        mc_def <- as.matrix(mc_def)
      } else {
        mc_def <- t(mc_def)
      }
      # def.function maps invalid evaluations (NaN) to +Inf;
      # treat those draws as missing so that cov() ignores them
      mc_def[!is.finite(mc_def)] <- as.numeric(NA)
      def_cov <- cov(mc_def, use = "pairwise.complete.obs")
      diag_def_cov <- diag(def_cov)
      diag_def_cov[diag_def_cov < 0] <- as.numeric(NA)
      se[def_idx] <- sqrt(diag_def_cov)
      return(se)
    }
    if (!is.null(boot)) {
      # we must remove the NA rows (and hope we have something left)
      error_idx <- attr(boot, "error.idx")
      boot <- boot
      if (length(error_idx) > 0L) {
        boot <- boot[-error_idx, , drop = FALSE] # drops attributes
      }

      boot_def <- apply(boot, 1L, lavmodel@def.function)
      if (length(def_idx) == 1L) {
        boot_def <- as.matrix(boot_def)
      } else {
        boot_def <- t(boot_def)
      }
      # new in 0.6-20: check for outliers, big difference between sd() and mad()
      # see github issue 347
      sd_mad_ratio <- (apply(boot_def, 2,  sd, na.rm = TRUE) /
                       apply(boot_def, 2, mad, na.rm = TRUE))
      crit_ratio <- 5
      if (any(sd_mad_ratio > crit_ratio)) {
        names_1 <- colnames(boot_def)
        def_w_outliers <- paste(names_1[sd_mad_ratio > crit_ratio],
                                collapse = " ")
        lav_msg_warn(gettextf(
          "The following bootstrapped defined parameters have a high (>5)
          ratio of standard deviation to median absolute deviation: %s.
          P-values and confidence intervals may not match.", def_w_outliers))
      }
      nboot <- nrow(boot_def)
      def_cov <- cov(boot_def) * (nboot - 1) / nboot
    } else {
      # regular delta method
      x <- lav_model_get_parameters(lavmodel = lavmodel, type = "free")
      jac <- try(lav_func_jacobian_complex(func = lavmodel@def.function, x = x),
        silent = TRUE
      )
      use_complex <- !inherits(jac, "try-error")
      if (!use_complex) { # eg. pnorm()
        jac <- lav_func_jacobian_simple(func = lavmodel@def.function, x = x)
      }

      # second-order delta method: compute Hessian for each defined param
      hess_list <- NULL
      if (!is.null(lavoptions) &&
        isTRUE(lavoptions$se.delta.second.order)) {
        if (use_complex) {
          hess_list <- try(lav_func_hessian_complex(
            func = lavmodel@def.function, x = x), silent = TRUE)
          if (inherits(hess_list, "try-error")) {
            hess_list <- try(lav_func_hessian_simple(
              func = lavmodel@def.function, x = x), silent = TRUE)
          }
        } else {
          hess_list <- try(lav_func_hessian_simple(
            func = lavmodel@def.function, x = x), silent = TRUE)
        }
        if (inherits(hess_list, "try-error")) {
          lav_msg_warn(gettext(
            "Could not compute the Hessian of the defined parameters;
             falling back to the first-order delta method."))
          hess_list <- NULL
        }
      }

      # jac (and hess_list) are in the compact (nx.free) space; when
      # ceq.simple.only, vcov is in the larger 'unco' space, so reduce vcov to
      # the compact space. NB: previously jac was *expanded* to the unco space
      # via ceq.simple.K, which multiply-counts the shared parameters and
      # inflates the SE of defined parameters that involve them.
      vcov_def <- lav_model_vcov_unco_to_free(lavmodel, vcov)
      def_cov <- jac %*% vcov_def %*% t(jac)

      # add second-order delta method correction:
      # Cov(g_i, g_j) ~ grad_i' V grad_j + 0.5 tr(H_i V H_j V)
      if (!is.null(hess_list)) {
        ndef_1 <- length(hess_list)
        correction <- matrix(0, ndef_1, ndef_1)
        hv_list <- lapply(hess_list, function(h) h %*% vcov_def)
        for (i in seq_len(ndef_1)) {
          for (j in i:ndef_1) {
            # tr(A B) where A = H_i V, B = H_j V
            #   = sum_{a,b} A_{a,b} B_{b,a}
            #   = sum element-wise of A and t(B)
            correction[i, j] <- 0.5 * sum(hv_list[[i]] * t(hv_list[[j]]))
            if (i != j) {
              correction[j, i] <- correction[i, j]
            }
          }
        }
        def_cov <- def_cov + correction
      }
    }
    # check for negative se's
    diag_def_cov <- diag(def_cov)
    diag_def_cov[diag_def_cov < 0] <- as.numeric(NA)
    se[def_idx] <- sqrt(diag_def_cov)
  }

  se
}

# compute 'Omega' == A1^{-1} B1 A1^{-1}
# where A1 is the expected/observed information matrix of the unrestricted (h1)
# model, and B1 is the first-order information matrix of the unrestricted (h1)
# model
#
# but the exact result will depend on the options:
# for 'A':
# - omega.information ("expected" or "observed")
# - omega.h1.information ("structured" or "unstructured")
# for 'B':
# - omega.information.meat ("first-order")
# - omega.h1.information.meat ("structured" or "unstructured")
#
# special case: if data is complete, A is expected/unstructured, and B is
# unstructured, we get (sample-based) 'Gamma'
#
# YR 28 Oct 2020

lav_model_h1_omega <- function(lavobject = NULL,
                               lavmodel = NULL,
                               lavsamplestats = NULL,
                               lavdata = NULL,
                               lavimplied = NULL,
                               lavh1 = NULL,
                               lavcache = NULL,
                               lavoptions = NULL) {
  lav_model_h1_unpack(lavobject, need_h1 = TRUE, need_implied = TRUE)

  # set options for A
  a1_options <- lavoptions
  a1_options$information <- lavoptions$omega.information
  a1_options$h1.information <- lavoptions$omega.h1.information

  b1_options <- lavoptions
  b1_options$information <- lavoptions$omega.information.meat # unused
  b1_options$h1.information <- lavoptions$omega.h1.information.meat

  # compute A1 (per group)
  a1_1 <- lav_model_h1_info(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats, lavdata = lavdata,
    lavimplied = lavimplied, lavh1 = lavh1,
    lavcache = lavcache, lavoptions = a1_options
  )

  # compute B1 (per group)
  b1 <- lav_model_h1_info_firstorder(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats, lavdata = lavdata,
    lavimplied = lavimplied, lavh1 = lavh1,
    lavcache = lavcache, lavoptions = b1_options
  )

  # return Omega per group
  omega <- vector("list", length = lavdata@ngroups)
  trace_h1 <- numeric(lavdata@ngroups)
  h1_ndat <- numeric(lavdata@ngroups)
  for (g in seq_len(lavdata@ngroups)) {
    a1_g <- a1_1[[g]]
    b1_g <- b1[[g]]

    # mask independent 'fixed-x' variables
    zero_idx <- which(diag(a1_g) == 0)
    if (length(zero_idx) > 0L) {
      a1_inv_1 <- matrix(0, nrow(a1_g), ncol(a1_g))
      a1 <- a1_g[-zero_idx, -zero_idx, drop = FALSE]
      a1_inv <- solve(a1)
      a1_inv_1[-zero_idx, -zero_idx] <- a1_inv
    } else {
      a1_inv_1 <- solve(a1_g)
    }
    trace_h1[g] <- sum(b1_g * t(a1_inv_1))
    h1_ndat[g] <- ncol(a1_g) - length(zero_idx)

    omega[[g]] <- a1_inv_1 %*% b1_g %*% a1_inv_1
  }

  # store trace.h1 as an attribute (to be used in yuan-bentler)
  attr(omega, "trace.h1") <- trace_h1
  attr(omega, "h1.ndat") <- h1_ndat
  attr(omega, "A.information") <- paste(a1_options$information,
    a1_options$h1.information,
    sep = "."
  )
  attr(omega, "B.information") <- paste(b1_options$information,
    b1_options$h1.information,
    sep = "."
  )

  omega
}

lav_step06_h1 <- function(sloth1 = NULL,
                                 lavoptions = NULL,
                                 lavsamplestats = NULL,
                                 lavdata = NULL,
                                 lavpartable = NULL) {
  # # # # # # # # #
  # #  6. lavh1 # #
  # # # # # # # # #

  # if sloth1 not NULL
  #   copy to lavh1
  # else
  #   if lavoptions$h1 TRUE
  #     if length(lavsamplestats$ntotal) > 0
  #       compute lavh1 via lav_h1_implied_logl
  #   else
  #     check lavoptions$h1 is logical, if not *** error ***

  if (!is.null(sloth1)) {
    lavh1 <- sloth1
  } else {
    lavh1 <- list()
    if (is.logical(lavoptions$h1) && lavoptions$h1) {
      if (length(lavsamplestats@ntotal) > 0L ||
	      (!is.null(lavoptions$samplestats) && !lavoptions$samplestats)) {
		  # lavsamplestats filled in
        if (lav_verbose()) {
          cat("lavh1              ... start:\n")
        }

        # implied h1 statistics and logl (if available)
        lavh1 <- lav_h1_implied_logl(
          lavdata = lavdata,
          lavsamplestats = lavsamplestats,
          lavpartable = lavpartable,
          lavoptions = lavoptions
        )
        if (lav_debug()) {
          print(lavh1)
        }
        if (lav_verbose()) {
          cat("lavh1              ... done.\n")
        }
      } else {
        # do nothing for now
      }
    } else {
      if (!is.logical(lavoptions$h1)) {
        lav_msg_stop(gettext("argument `h1' must be logical (for now)"))
      }
      # TODO: allow h1 to be either a model syntax, a parameter table,
      # or a fitted lavaan object
    }
  }

  lavh1
}

# compute two-step standard errors for SAM models
#
# several possibilities:
# 1) se = "twostep": classic (but global) two-step corrected SEs
#     - create 'global' model, only to get the 'joint' information matrix
#     - partition information matrix (step 1, step 2)
#     - apply two-step correction for second step
#     - 'insert' these corrected SEs (and vcov) in JOINT
# 2) se = "standard": using I.22.inv (but without correction term)
# 3) se = "naive": grab (naive) VCOV from FIT.PA
# 4) se = "local": grab (robust) VCOV from FIT.PA

# se = "bootstrap": bootstrap the whole two-step procedure, and compute
# the vcov of the (combined step 1 + step 2) coefficients
lav_sam_step2_se_bootstrap <- function(sam_object = NULL, bootstrap = list()) {
  default_args <- list(R = 1000L, type = "ordinary",
                       show.progress = FALSE,
                       check.post = TRUE, keep.idx = FALSE)
  this_args <- modifyList(default_args, bootstrap)
  coef_1 <- lav_bootstrap_internal(object = sam_object,
    r = this_args$R, show_progress = this_args$show.progress,
    type = this_args$type, fun = "coef",
    check_post = this_args$check.post, keep_idx = this_args$keep.idx)
  coef_orig <- coef_1
  error_idx <- attr(coef_1, "error.idx")
  nfailed <- length(error_idx) # zero if NULL
  if (nfailed > 0L) {
    lav_msg_warn(gettextf(
      "%s bootstrap runs failed or did not converge.", nfailed))
  }
  notok <- length(attr(coef_1, "nonadmissible")) # zero if NULL
  if (notok > 0L) {
    lav_msg_warn(gettextf(
      "%s bootstrap runs resulted in nonadmissible solutions.", notok))
  }
  if (length(error_idx) > 0L) {
    # new in 0.6-13: we must still remove them!
    coef_1 <- coef_1[-error_idx, , drop = FALSE]
    # this also drops the attributes
  }
  nboot <- nrow(coef_1)
  var_cov <- cov(coef_1) * (nboot - 1) / nboot

  list(VCOV = var_cov, boot.coef = coef_orig, R = this_args$R)
}

# WLS.V_g %*% Delta_g (the 'WD' building block of the robust SEM sandwich),
# handling the DWLS/ULS case where lavaan stores the diagonal weight matrix as
# a plain vector. Optionally restrict Delta to a subset of columns.
lav_sam_wls_delta <- function(wls_v_g, delta_g, cols = NULL) {
  if (!is.null(cols)) {
    delta_g <- delta_g[, cols, drop = FALSE]
  }
  if (is.matrix(wls_v_g)) {
    wls_v_g %*% delta_g
  } else {
    # DWLS/ULS: WLS.V holds the diagonal of the weight matrix
    wls_v_g * delta_g
  }
}

# Case B local SEs: across-group (equality) constraints in the measurement
# model couple vech(VETA_g) across groups, so the NACOV of the stacked
# structural statistics (step1$Gamma.eta.full) has nonzero cross-group blocks
# that lavaan's multigroup robust.sem (a per-group NACOV list) cannot use.
# We therefore compute the robust sandwich ourselves, reusing the structural
# model's Delta, WLS.V (weight) and the augmented inverse information E.inv:
#
#   VCOV = (1/N^2) E.inv [ WD' (Dn Cov.full Dn) WD ] E.inv
#
# with WD = rbind_g( WLS.V_g %*% Delta_g ), Dn = diag(n_g) (repeated over the
# statistics of group g), Cov.full = the full covariance of the stacked
# structural statistics. When Cov.full is block-diagonal (Case A) this reduces
# exactly to lavaan's E.inv %*% (sum_g (n_g/N) WD_g' Gamma_g WD_g) %*% E.inv / N.
lav_sam_step2_se_local_caseB <- function(fit_pa, step1) {
  ngroups <- fit_pa@Data@ngroups
  ntotal  <- fit_pa@SampleStats@ntotal
  nobs_g  <- unlist(fit_pa@SampleStats@nobs)

  # structural model Delta, WLS.V (weight) and augmented E.inv
  tmp <- lav_model_nvcov_robust_sem(
    lavmodel = fit_pa@Model, lavsamplestats = fit_pa@SampleStats,
    lavcache = fit_pa@cache, lavdata = fit_pa@Data,
    lavimplied = fit_pa@implied, lavh1 = fit_pa@h1,
    lavoptions = fit_pa@Options, use_ginv = FALSE,
    attr_delta = TRUE, attr_e_inv = TRUE, attr_wls_v = TRUE)
  e_inv <- attr(tmp, "E.inv")
  delta <- attr(tmp, "Delta")
  wls_v <- attr(tmp, "WLS.V")

  # WD = rbind_g( W_g Delta_g ); dn = n_g repeated over group g's statistics
  wd_list <- vector("list", ngroups)
  dn_list <- vector("list", ngroups)
  for (g in seq_len(ngroups)) {
    wd <- lav_sam_wls_delta(wls_v[[g]], delta[[g]])
    wd_list[[g]] <- wd
    dn_list[[g]] <- rep.int(nobs_g[g], nrow(wd))
  }
  wd_all <- do.call(rbind, wd_list)
  dn <- unlist(dn_list)

  cov_full <- step1$Gamma.eta.full
  if (nrow(cov_full) != length(dn)) {
    lav_msg_stop(gettext(
      "internal error: dimension mismatch while assembling the multigroup
       (Case B) local sandwich."))
  }

  # M = WD' (Dn Cov.full Dn) WD  ;  (Dn Cov.full Dn)[i,j] = n_i cov[i,j] n_j
  dcd <- cov_full * outer(dn, dn)
  m_cov <- crossprod(wd_all, dcd %*% wd_all)
  vcov_1 <- (1 / (ntotal * ntotal)) * (e_inv %*% m_cov %*% e_inv)
  vcov_1
}

# Casewise (huber.white) building blocks for the two-step sandwich
# (se = "twostep.huber.white"): the Yuan & Chan V22/V21/V11 pieces, but with
# casewise-SCORE crossproducts instead of Gamma-based expressions:
#
#   g2_i = casewise score of the JOINT model w.r.t. theta2, at (theta2.hat,
#          theta1.hat)
#   h_i  = casewise influence on theta1.hat: per measurement block b,
#          h_i^(b) = (N/n_b) I_b^{-1} g1_i^(b)   (unit block information I_b,
#          n_b = cases used by block b; zero rows for cases without data in b)
#
#   V22 = (1/N) sum g2_i g2_i',  V21 = (1/N) sum g2_i h_i',
#   V11 = (1/N) sum h_i h_i'
#
# Fed into the same A/B assembly as the twostep.robust branch, this yields the
# stacked M-estimation (Murphy-Topel) sandwich with first-order meat -- the
# two-step analogue of se = "robust.huber.white". Because the meat is built
# from casewise scores, it needs no Gamma: missing = "ml" (FIML scores per
# missing pattern) and multiple groups (zero scores outside a case's group)
# are covered by the same expressions.
#
# The scores are centered before the crossproducts: an exact no-op for
# sam.method = "global" (the joint theta2 score is solved: sum_i g2_i = 0, and
# sum_i g1_i = 0 per block), and the correct removal of the O(1) offset for
# the local methods (whose step-2 estimator does not solve the joint score
# exactly -- the same asymptotic-equivalence argument that underlies the
# twostep.robust linearization).
lav_sam_step2_se_hw_v <- function(joint = NULL, step1 = NULL,
                                  step2_free_idx = NULL, ntot = NULL) {
  npar <- max(joint@ParTable$free)
  # for the local sam methods, the joint object is assembled without fitting
  # and carries Options$estimator == "none"; the casewise scores are those of
  # the (ML) joint likelihood in Model@estimator
  if (!joint@Options$estimator %in% c("ML", "WLS", "GLS", "ULS")) {
    joint@Options$estimator <- joint@Model@estimator
  }
  # ignore_constraints: the sam models may carry synthesized INEQUALITY
  # constraints (the measurement blocks are fitted with bounds =
  # "wide.zerovar", which lavaan represents as linear cin rows in con.jac);
  # these are inactive at an interior solution, but lav_sc()'s constraint
  # projection would still project the scores onto them, corrupting the
  # casewise influences. Equality constraints are excluded from this se
  # flavour upstream, so skipping the projection is exact here.
  sc <- lav_sc(joint, remove_empty_cases = FALSE,
               ignore_constraints = TRUE)
  sc[is.na(sc)] <- 0 # empty cases contribute a zero score
  if (ncol(sc) != npar) {
    lav_msg_stop(gettext(
      "internal error: the joint casewise scores do not match the free
       parameters (twostep.huber.white)."))
  }
  g2 <- sc[, step2_free_idx, drop = FALSE]

  # step-1 casewise influence, in the joint free-parameter numbering
  h_full <- matrix(0, nrow(sc), npar)
  pt_free <- step1$PT.free
  for (mm in seq_along(step1$MM.FIT)) {
    fb <- step1$MM.FIT[[mm]]
    scb <- lav_sc(fb, remove_empty_cases = FALSE,
                  ignore_constraints = TRUE) # see note above
    scb[is.na(scb)] <- 0 # cases with no data in this block: zero score
    if (nrow(scb) != nrow(sc)) {
      lav_msg_stop(gettext(
        "internal error: case alignment failure between a measurement block
         and the joint model (twostep.huber.white)."))
    }
    ib <- lavTech(fb, "information") # unit information (honors information=)
    nb <- fb@SampleStats@ntotal
    hb <- (ntot / nb) * scb %*% solve(ib)
    # map the block's free parameters into the joint free numbering (the same
    # mapping used for the Sigma.11 assembly in lav_sam_step1())
    mm_idx <- step1$block.mm.idx[[mm]]
    ptm_idx <- step1$block.ptm.idx[[mm]]
    par_idx <- pt_free[mm_idx[ptm_idx]]
    keep_idx <- fb@ParTable$free[ptm_idx]
    ok <- par_idx > 0L & keep_idx > 0L # drop := etc.
    h_full[, par_idx[ok]] <- hb[, keep_idx[ok], drop = FALSE]
  }

  # center (see note above) and build the V pieces
  g2c <- t(t(g2) - colMeans(g2))
  h1c <- t(t(h_full) - colMeans(h_full))
  list(
    v22 = crossprod(g2c) / ntot,
    v21 = crossprod(g2c, h1c) / ntot, # cols = FULL joint numbering
    v11 = crossprod(h1c) / ntot # rows/cols = FULL joint numbering
  )
}

# grab the 'naive' vcov from FIT.PA (computing it if absent), removing
# any rows/cols in step2_rm_idx
lav_sam_step2_se_vcov_pa <- function(fit_pa, step2_rm_idx = integer(0L)) {
  if (is.null(fit_pa@vcov$vcov)) {
    fit_pa@Options$se <- "standard"
    vcov_pa <- lavTech(fit_pa, "vcov")
  } else {
    vcov_pa <- fit_pa@vcov$vcov
  }
  if (length(step2_rm_idx) > 0L) {
    vcov_pa <- vcov_pa[-step2_rm_idx, -step2_rm_idx, drop = FALSE]
  }
  vcov_pa
}

# the robust (sandwich) vcov of FIT.PA, using the NACOV (= Gamma.eta) it was
# fitted with -- FIT.PA itself was fitted with se = "standard" (so that the
# naive vcov is available for the alpha blending), so recompute here
lav_sam_step2_se_vcov_pa_robust <- function(fit_pa,
                                            step2_rm_idx = integer(0L)) {
  fit_pa@Options$se <- "robust.sem"
  fit_pa@vcov$vcov <- NULL # force recompute
  vcov_pa <- lavTech(fit_pa, "vcov")
  if (length(step2_rm_idx) > 0L) {
    vcov_pa <- vcov_pa[-step2_rm_idx, -step2_rm_idx, drop = FALSE]
  }
  vcov_pa
}

# Standard errors for latent (residual) variances that are 'fixed' (free == 0)
# in the JOINT model but are in fact data-derived. Under std.lv = TRUE, SAM
# step 1 fixes the *total* variance of each factor to 1; for an *endogenous*
# factor the residual variance is then a derived quantity
#   r_j = total_j - explained_j   with   VETA = (I - B)^-1 Psi (I - B)^-T
# (total_j fixed to 1). Although r_j is fixed-by-label in the joint model (and
# therefore excluded from the joint vcov, giving se = 0), it is a function of
# the free structural parameters and has genuine sampling variability. We
# recover its SE by the delta method, using a numerical jacobian of r_j
# through the model-implied VETA and the (se-method-specific) joint vcov.
# Exogenous factor variances have no incoming paths, so their jacobian is zero
# and they keep se = 0 automatically.
#
# Returns a numeric vector aligned with the rows of joint@ParTable (0 for every
# row that is not such a derived latent variance).
lav_sam_step2_se_lv_var <- function(joint = NULL) {
  lavmodel    <- joint@Model
  lavpartable <- joint@ParTable
  VCOV        <- joint@vcov$vcov
  se_out      <- numeric(length(lavpartable$lhs))

  if (is.null(VCOV) || lavmodel@representation != "LISREL" || anyNA(VCOV)) {
    return(se_out)
  }

  # 'psi' matrix index in GLIST per block, plus its latent-variable names
  nmat      <- lavmodel@nmat
  mm_idx    <- lav_model_group_mm_indices(nmat)
  psi_glist <- which(names(lavmodel@GLIST) == "psi")
  if (length(psi_glist) == 0L) {
    return(se_out)
  }
  lv_block <- lapply(seq_len(lavmodel@nblocks), function(g) {
    psi_mm <- intersect(mm_idx[[g]], psi_glist)
    lavmodel@dimNames[[psi_mm[1]]][[1]]
  })

  # target rows: latent (residual) variances, fixed (free == 0), and not
  # explicitly fixed by the user (user != 1)
  target <- which(
    lavpartable$op == "~~" &
      lavpartable$lhs == lavpartable$rhs &
      lavpartable$free == 0L &
      lavpartable$user != 1L
  )
  target <- target[vapply(target, function(i) {
    lavpartable$lhs[i] %in% lv_block[[lavpartable$block[i]]]
  }, logical(1L))]
  if (length(target) == 0L) {
    return(se_out)
  }

  # parameter vector in the space matching VCOV (unco if simple eq constraints)
  ceq_simple <- lavmodel@ceq.simple.only &&
    nrow(lavmodel@ceq.simple.K) == nrow(VCOV)
  vtype <- if (ceq_simple) "unco" else "free"
  x0 <- lav_model_get_parameters(lavmodel) # nx.free
  if (ceq_simple) {
    x0 <- drop(lavmodel@ceq.simple.K %*% x0) # nx.unco
  }
  npar <- length(x0)
  if (npar != ncol(VCOV)) { # spaces do not line up: stay safe
    return(se_out)
  }

  # explained_j: model-implied total variance of factor j in block g, with its
  # own residual psi[j, j] set to zero
  explained <- function(x, g, j) {
    gl <- lav_model_x2glist(lavmodel, x = x, type = vtype)
    psi_mm <- intersect(mm_idx[[g]], psi_glist)[1]
    gl[[psi_mm]][j, j] <- 0
    lav_model_veta(lavmodel = lavmodel, glist = gl)[[g]][j, j]
  }

  eps <- 1e-6
  for (row in target) {
    g <- lavpartable$block[row]
    j <- match(lavpartable$lhs[row], lv_block[[g]])
    if (is.na(j)) next
    v <- try(
      {
        e0 <- explained(x0, g, j)
        jac <- numeric(npar)
        for (k in seq_len(npar)) {
          xk <- x0
          xk[k] <- xk[k] + eps
          jac[k] <- -(explained(xk, g, j) - e0) / eps
        }
        as.numeric(crossprod(jac, VCOV %*% jac))
      },
      silent = TRUE
    )
    if (!inherits(v, "try-error") && is.finite(v) && v >= 0) {
      se_out[row] <- sqrt(v)
    }
  }

  se_out
}

lav_sam_step2_se <- function(fit = NULL, joint = NULL,
                             step1 = NULL, step2 = NULL,
                             local_options = list()) {
  out <- list()
  sigma_11 <- step1$Sigma.11
  step1_free_idx <- step1$step1.free.idx
  step2_free_idx <- step2$step2.free.idx
  lavoptions <- fit@Options
  nlevels <- fit@pta$nlevels
  fit_pa <- step2$FIT.PA
  extra_id <- step2$extra.id

  # catch empty step2.free.idx
  if (length(step2_free_idx) == 0L) {
    # no (free) structural parameters at all!
    out <- list(
      V1 = matrix(0, 0, 0), V2 = matrix(0, 0, 0),
      VCOV = matrix(0, 0, 0)
    )
    return(out)
  }

  if (!lavoptions$se %in%
    c("none", "standard", "naive", "twostep", "twostep.robust",
      "twostep.huber.white", "local", "local.nt")) {
    lav_msg_warn(gettextf(
      "unknown se= argument: %s. Switching to twostep.",
      lavoptions$se
    ))
  }

  if (lavoptions$se == "none") {
    return(out)
  }

  if (lav_verbose()) {
    cat("Computing ", lavoptions$se, " standard errors ... ", sep = "")
  }

  # conditional.x + local sam method: the joint-score (Yuan & Chan)
  # linearization of the twostep.robust branch below describes the
  # joint-model-given-theta1 estimator, which under conditional.x (with
  # non-centered exogenous covariates) is NOT asymptotically equivalent to
  # the local step-2 estimator: the structural fit is saturated in the
  # latent-on-x slopes and cannot pool information across statistics the way
  # the joint estimator does, so the joint-score vcov underestimates the
  # sampling variability (badly so for binary covariates). Compute the
  # twostep.robust SEs as the structural-space sandwich instead (FIT.PA with
  # NACOV = Gamma.eta), which linearizes the actual two-step estimator and
  # makes them identical to se = "local" in this setting. (For sam.method =
  # "global" step 2 IS the joint estimator, and the branch below applies.)
  tsrobust_condx_flag <- lavoptions$se == "twostep.robust" &&
    fit@Model@conditional.x &&
    isTRUE(step1$sam.method %in% c("local", "fsr", "cfsr")) &&
    !is.null(step1$Gamma.eta) && !is.null(step1$Gamma.eta[[1]])

  # PML + twostep.huber.white: the assembled joint carries
  # Options$estimator == "none" and no PML cache (bifreq/long/uniweights),
  # while both the joint information (the numerical hessian of the PML
  # gradient) and the casewise scores need them. Patch a local copy of the
  # joint (R copy semantics: the returned object is not affected).
  if (lavoptions$se == "twostep.huber.white" &&
      joint@Model@estimator == "PML") {
    joint@Options$estimator <- "PML"
    if (is.null(joint@Cache[[1]]$long)) {
      joint@Cache <- lav_step10_cache(
        slot_cache = NULL,
        lavdata = joint@Data,
        lavmodel = joint@Model,
        lavpartable = joint@ParTable,
        lavoptions = joint@Options,
        sampling_weights = if (length(joint@Data@sampling.weights) > 0L) {
          joint@Data@sampling.weights
        } else {
          NULL
        }
      )
    }
  }

  if (lavoptions$se %in% c("naive", "twostep", "twostep.robust",
                           "twostep.huber.white")) {
    info <- lavInspect(joint, "information")
    i_12 <- info[step1_free_idx, step2_free_idx, drop = FALSE]
    i_22 <- info[step2_free_idx, step2_free_idx, drop = FALSE]
    i_21 <- info[step2_free_idx, step1_free_idx, drop = FALSE]
  }

  # V2
  if (nlevels > 1L) {
    # FIXME: not ok for multigroup multilevel
    n <- fit@Data@Lp[[1]]$nclusters[[2]] # first group only
  } else {
    n <- nobs(fit)
  }

  # total number of free parameters STRUC
  if (fit_pa@Model@ceq.simple.only) {
    npar <- fit_pa@Model@nx.unco
    pts_free <- fit_pa@ParTable$free
    pts_free[pts_free > 0] <- seq_len(npar)
  } else {
    npar <- fit_pa@Model@nx.free
    pts_free <- fit_pa@ParTable$free
  }

  # do we have 'extra' free parameters in FIT.PA that are not free in JOINT?
  step2_rm_idx <- integer(0L)
  if (length(extra_id) > 0L) {
    id_idx <- which(fit_pa@ParTable$id %in% extra_id &
      fit_pa@ParTable$free > 0L)
    step2_rm_idx <- pts_free[id_idx]
  }

  # invert augmented information, for I.22 block only
  # new in 0.6-16 (otherwise, eq constraints in struc part are ignored)
  if (lavoptions$se %in% c("standard", "twostep", "twostep.robust",
                           "twostep.huber.white")) {
    # Fix for EFA/ESEM: when rotation is used, FIT.PA@Model@con.jac includes
    # columns for rotation identification constraints that are not part of
    # step2.free.idx. These extra columns cause a dimension mismatch in
    # lav_model_info_augment_invert(). Remove them via rm.idx. NOTE: this
    # adjustment is only relevant for the augmented-inverse path below; it must
    # NOT be added to step2_rm_idx itself, because the naive/local path reads
    # the FIT.PA vcov directly and would then drop legitimate parameters (this
    # used to break multigroup models with across-group equality constraints,
    # where con.jac legitimately has more columns than step2.free.idx).
    aug_rm_idx <- step2_rm_idx
    if (nrow(fit_pa@Model@con.jac) > 0L) {
      n_jac_cols <- ncol(fit_pa@Model@con.jac)
      n_step2 <- length(step2_free_idx)
      # 'extra' FIT.PA parameters that are not free in the JOINT model are
      # already in step2_rm_idx; only columns BEYOND n_step2 + those extras
      # are rotation-identification columns that must be removed in
      # addition. (Comparing n_jac_cols against n_step2 alone wrongly
      # trimmed legitimate constraint columns whenever FIT.PA had extra
      # parameters, eg multigroup group.equal = "regressions".)
      n_expected <- n_step2 + length(step2_rm_idx)
      if (n_jac_cols > n_expected) {
        aug_rm_idx <- union(step2_rm_idx, (n_expected + 1):n_jac_cols)
      }
    }
    i_22_inv <-
      lav_model_info_augment_invert(
        lavmodel = fit_pa@Model,
        information = i_22,
        inverted = TRUE,
        use_ginv = FALSE, # if interaction, SEs end up smaller than naive...
        rm_idx = aug_rm_idx
      )
    if (inherits(i_22_inv, "try-error")) {
      # hm, not good
      if (lavoptions$se != "naive") {
        lav_msg_warn(gettext(
          "problem inverting information matrix (I.22); -> switching
            to naive standard errors!"
        ))
        lavoptions$se <- "naive"
      }
    }
  } # se needs I.22.inv

  # note: computing I.22.inv as a Schur complement of JOINT@vcov$vcov
  # would allow a 'robust' vcov for the joint model, but does not work if
  # we have equality constraints in the MM (the (1,1) block is singular)

  # se = "standard"
  if (lavoptions$se == "standard") {
    vcov_1 <- 1 / n * i_22_inv
    out$VCOV <- vcov_1

  # se = "naive" or "local": grab VCOV directly from FIT.PA
  # (also twostep.robust under conditional.x -- see tsrobust_condx_flag above)
  } else if (lavoptions$se %in% c("naive", "local", "local.nt") ||
             tsrobust_condx_flag) {
    if (isTRUE(step1$caseB) && lavoptions$se %in% c("local", "local.nt")) {
      # across-group constraints in the measurement model: build the
      # cross-group sandwich (the per-group FIT.PA vcov is incomplete)
      vcov_1 <- lav_sam_step2_se_local_caseB(fit_pa = fit_pa, step1 = step1)
      if (length(step2_rm_idx) > 0L) {
        vcov_1 <- vcov_1[-step2_rm_idx, -step2_rm_idx, drop = FALSE]
      }
    } else if (tsrobust_condx_flag) {
      vcov_1 <- lav_sam_step2_se_vcov_pa_robust(fit_pa, step2_rm_idx)
    } else {
      vcov_1 <- lav_sam_step2_se_vcov_pa(fit_pa, step2_rm_idx)
    }
    # order rows/cols of VCOV, so that they correspond with the (step 2)
    # parameters of the JOINT model, but remove := parameters first
    pt_idx <- step2$pt.idx
    nonpar_id <- fit@ParTable$id[fit@ParTable$op %in% c(":=", "==", "<", ">")]
    rm_idx <- which(step2$pt.idx %in% nonpar_id)
    if (length(rm_idx) > 0L) {
      pt_idx <- step2$pt.idx[-rm_idx]
    }
    idx <- sort.int(pt_idx, index.return = TRUE)$ix
    vcov_1 <- vcov_1[idx, idx, drop = FALSE]

    # drop parameters that are free in FIT.PA, but fixed in the JOINT
    # model (eg std.lv = TRUE: the lv (residual) variances are freed in
    # the structural model, but fixed (to 1) in the joint model); there is
    # no room for their sampling variability in the joint vcov, and
    # step2.free.idx does not include them
    pt_2 <- step2$PT
    keep_idx <- which(pt_2$free[sort.int(pt_idx)] > 0L)
    if (length(keep_idx) < length(pt_idx)) {
      vcov_1 <- vcov_1[keep_idx, keep_idx, drop = FALSE]
    }

    out$VCOV <- vcov_1

  # se = "twostep", "twostep.robust" or "twostep.huber.white"
  } else if (lavoptions$se %in% c("twostep", "twostep.robust",
                                  "twostep.huber.white")) {

    robust <- (lavoptions$se == "twostep.robust")
    huber_white <- (lavoptions$se == "twostep.huber.white")
    hw_v <- NULL
    if (huber_white) {
      # casewise-score V pieces; on failure fall back to plain twostep
      hw_v <- tryCatch(
        lav_sam_step2_se_hw_v(
          joint = joint, step1 = step1,
          step2_free_idx = step2_free_idx, ntot = n
        ),
        error = function(e) e
      )
      if (inherits(hw_v, "error")) {
        lav_msg_warn(gettextf(
          "huber.white standard errors (se = \"twostep.huber.white\") could
           not be computed (%s); twostep standard errors are reported
           instead.", conditionMessage(hw_v)))
        huber_white <- FALSE
        lavoptions$se <- "twostep"
      }
    }
    if (robust && fit@Model@conditional.x &&
        isTRUE(step1$sam.method %in% c("local", "fsr", "cfsr"))) {
      # only reached when Gamma.eta is unavailable (tsrobust_condx_flag was
      # FALSE): the joint-score correction below is not valid for the local
      # step-2 estimator under conditional.x -> plain twostep
      lav_msg_warn(gettext(
        "robust standard errors (se = \"twostep.robust\") need Gamma.eta
         under conditional.x = TRUE, which could not be computed; twostep
         standard errors are reported instead."))
      robust <- FALSE
      lavoptions$se <- "twostep"
    }
    if (robust) {
      # get P (the influence of the step 1 / measurement parameters on the
      # sample statistics) and align its rows with step1.free.idx (the rows of
      # P are in ascending free-parameter order). Single group: one matrix;
      # multigroup: the columns are the stacked statistics of all groups, with
      # attr 'stat.offset' giving the per-group column boundaries.
      p <- lav_sam_step1_local_jac(step1 = step1, fit = fit, p_only = TRUE)
      p_row_idx <- match(step1_free_idx, attr(p, "free.idx"))
      if (anyNA(p_row_idx)) {
        # P cannot be aligned with the step 1 free parameters: the robust
        # correction is not available -> fall back to (non-robust) twostep
        # SEs. (Should not happen: within-block and across-group equality
        # constraints are handled; this is a defensive guard.)
        lav_msg_warn(gettext(
          "robust standard errors (se = \"twostep.robust\") are not available
           for this model; twostep standard errors are reported instead."))
        robust <- FALSE
        lavoptions$se <- "twostep"
      }
    }
    if (huber_white) {
      # stacked M-estimation sandwich with casewise-score meat: same A/B
      # assembly as the twostep.robust branch below, with the V pieces
      # replaced by their casewise counterparts (see lav_sam_step2_se_hw_v)
      m_b <- -1 * info[step2_free_idx, step1_free_idx, drop = FALSE]
      v22 <- hw_v$v22
      v21 <- hw_v$v21[, step1_free_idx, drop = FALSE]
      v11 <- hw_v$v11[step1_free_idx, step1_free_idx, drop = FALSE]
      v12 <- t(v21)

      a_inv <- -1 * i_22_inv
      v2 <- 1 / n * (a_inv %*% v22 %*% a_inv)
      v1 <- 1 / n * (a_inv %*%
        (m_b %*% v12 + v21 %*% t(m_b) + m_b %*% v11 %*% t(m_b)) %*% a_inv)
    } else if (!robust) {
      v2 <- 1 / n * i_22_inv # not the same as FIT.PA@vcov$vcov!!
      v1 <- i_22_inv %*% i_21 %*% sigma_11 %*% i_12 %*% i_22_inv
    } else {
      # following Yuan & Chan 2002, eqs 4, 10, 11, 12, 13 and 14
      # but for V11, V12, V21, V22: we use index '1' for step1, and '2'
      # for step 2!!
      m_b <- -1 * info[step2_free_idx, step1_free_idx, drop = FALSE]
      stat_offset <- attr(p, "stat.offset") # NULL if single group
      p <- p[p_row_idx, , drop = FALSE]

      # get V22 (= the robust covariance of the step 2 estimating function,
      # already aggregated over groups by lav_model_nvcov_robust_sem())
      if (is.null(joint@SampleStats@NACOV[[1]])) {
        joint@SampleStats@NACOV <- lavTech(joint, "gamma")
      }
      tmp <- lav_model_nvcov_robust_sem(
        lavmodel = joint@Model, lavsamplestats = joint@SampleStats,
        lavcache = joint@cache, lavdata = joint@Data,
        lavimplied = joint@implied, lavh1 = joint@h1,
        lavoptions = joint@Options, use_ginv = FALSE,
        attr_delta = TRUE, attr_t_dvgvd = TRUE, attr_e_inv = TRUE,
        attr_wls_v = TRUE)
      delta <- attr(tmp, "Delta")
      wls_v <- attr(tmp, "WLS.V")
      t_dvgvd <- attr(tmp, "tDVGVD")

      v22 <- t_dvgvd[step2_free_idx, step2_free_idx, drop = FALSE] # ok

      # V11 = Cov(sqrt(N) * step1 params) and V21 = Cov(sqrt(N) * step2 score,
      # sqrt(N) * step1 params). Because the groups are independent samples,
      # Gamma is block-diagonal across groups, so both aggregate as a sum over
      # groups. The group weights differ per block, however, because the step 2
      # estimating function G is the sample-size weighted average of the
      # per-group scores (G = sum_g f_g G_g), while the step 1 estimates gamma_g
      # are per-group. With Gamma_g = Cov(sqrt(n_g) s_g) and f_g = n_g/N:
      #   V22 = Cov(sqrt(N) G)         -> sum_g f_g   WD_g' Gamma_g WD_g  (= t_dvgvd)
      #   V21 = Cov(sqrt(N) G, gamma)  -> sum_g  1    WD_g' Gamma_g P_g'
      #   V11 = Cov(sqrt(N) gamma)     -> sum_g 1/f_g P_g  Gamma_g P_g'
      # (each replacement of a 'score' index by a 'parameter' index trades one
      # factor f_g for 1/f_g). This makes the three terms of V1 below scale
      # consistently, so for independent groups (Case A) the per-group result
      # equals a separate single-group fit. It also covers across-group
      # measurement constraints (Case B): the cross-group coupling is encoded
      # in the per-group columns of P. For a single group f_g = 1 and the loop
      # reduces to the original (unweighted) single-group expressions.
      gamma <- lavTech(joint, "gamma")
      ngroups <- joint@Data@ngroups
      ntot <- joint@SampleStats@ntotal
      v11 <- matrix(0, nrow(p), nrow(p))
      v21 <- matrix(0, length(step2_free_idx), nrow(p))
      for (g in seq_len(ngroups)) {
        fg <- joint@SampleStats@nobs[[g]] / ntot
        gamma_g <- gamma[[g]]
        if (is.null(stat_offset)) {
          p_g <- p
        } else {
          p_g <- p[, (stat_offset[g] + 1L):stat_offset[g + 1L], drop = FALSE]
        }
        wd_g <- lav_sam_wls_delta(wls_v[[g]], delta[[g]], step2_free_idx)
        v11 <- v11 + (1 / fg) * (p_g %*% gamma_g %*% t(p_g))
        v21 <- v21 + crossprod(wd_g, gamma_g %*% t(p_g))
      }
      v12 <- t(v21)

      a_inv <- -1 * i_22_inv
      v2 <- 1 / n * (a_inv %*% v22 %*% a_inv)
      v1 <- 1 / n * (a_inv %*%
        (m_b %*% v12 + v21 %*% t(m_b) + m_b %*% v11 %*% t(m_b)) %*% a_inv)
    }

    # V for second step
    if (!is.null(local_options$alpha.correction) &&
      local_options$alpha.correction > 0) {
      alpha_n1 <- lav_sam_alpha_n1(local_options$alpha.correction, n)
      vcov_naive <- lav_sam_step2_se_vcov_pa(fit_pa, step2_rm_idx)
      vcov_corrected <- v2 + v1
      vcov_1 <- alpha_n1 * vcov_naive + (1 - alpha_n1) * vcov_corrected
    } else {
      # no alpha correction
      vcov_1 <- v2 + v1
    }

    # store in out
    out$V2 <- v2
    out$V1 <- v1
    out$VCOV <- vcov_1
  } # twostep

  # store se
  out$se <- lavoptions$se # in case it changed

  if (lav_verbose()) {
    cat("done.\n")
  }

  out
}

# here, we generate new models based on the original model in lavobject
# 1. the independence model
# 2. the unrestricted model
# 3. model + extra parameters (for modindices/lavTestScore)
# 4. catML fit based on DWLS fit (for robust RMSEA/CFI)


# 1. fit an 'independence/baseline' model
#    note that for ML (and ULS and DWLS), the 'estimates' of the
#    independence model are simply the observed variances
#    but for GLS and WLS, this is not the case!!
#
#  - YR 01 Feb 2026: allow for baseline.type = "nested"
lav_object_baseline <- function(object = NULL,
                                # or
                                lavsamplestats = NULL,
                                lavdata = NULL,
                                lavcache = NULL,
                                lavoptions = NULL,
                                lavpartable = NULL,
                                lavh1 = NULL,
                                # local options
                                se = FALSE) {
  # object or slots?
  if (!is.null(object)) {
    stopifnot(inherits(object, "lavaan"))
    object <- lav_object_check_version(object)

    # extract needed slots
    lavsamplestats <- object@SampleStats
    lavdata <- object@Data
    lavcache <- object@Cache
    lavoptions <- object@Options
    lavpartable <- object@ParTable
    lavpta <- object@pta
    lavh1 <- object@h1
    if (is.null(lavoptions$estimator.args)) {
      lavoptions$estimator.args <- list()
    }
  } else {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }

  # if two-level, force conditional.x = FALSE (for now)
  # exception: two-level (D)WLS, where the baseline model supports (and
  # needs) conditional.x
  if (lavdata@nlevels > 1L && lavoptions$conditional.x &&
      !lavoptions$estimator %in% c("WLS", "DWLS", "ULS")) {
    lavoptions$conditional.x <- FALSE
  }

  # construct parameter table for independence model
  if (!is.null(lavoptions$baseline.type) &&
    lavoptions$baseline.type == "nested") {
    lavpartable <- lav_pt_baseline(
      lavobject = NULL,
      lavpartable = lavpartable, lavh1 = lavh1
    )
  } else {
    lavpartable <- lav_pt_indep_or_unrestricted(
      lavobject = NULL,
      lavdata = lavdata, lavpta = lavpta, lavoptions = lavoptions,
      lavsamplestats = lavsamplestats, lavh1 = lavh1, independent = TRUE
    )
    # auxiliary variables (FIML saturated correlates): free the auxiliary
    # covariances in the baseline, so the auxiliary variables stay saturated
    # (the model observed variables remain uncorrelated); see lav_aux.R
    if (length(lavoptions$aux) > 0L) {
      lavpartable <- lav_aux_satcor_baseline(
        indep_partable = lavpartable,
        aux = lavoptions$aux,
        ov_names = lavdata@ov.names,
        ngroups = lavdata@ngroups
      )
    }
  }

  # new in 0.6-6: add lower bounds for ov.var
  if (!is.null(lavoptions$optim.bounds)) {
    lavoptions$bounds <- "doe.maar"
    lavoptions$effect.coding <- "" # to avoid warning
    lavoptions$optim.bounds <- list(lower = "ov.var")
    lavpartable <- lav_pt_add_bounds(
      partable = lavpartable,
      lavh1 = lavh1, lavdata = lavdata,
      lavsamplestats = lavsamplestats, lavoptions = lavoptions
    )
  }

  # new in 0.6-8: if DLS, change to sample-based
  if (lavoptions$estimator == "DLS") {
    if (lavoptions$estimator.args$dls.GammaNT == "sample") {
      # nothing to do
    } else {
      lavoptions$estimator.args$dls.GammaNT <- "sample"
      for (g in 1:lavsamplestats@ngroups) {
        # overwrite
        lavsamplestats@WLS.V[[g]] <- lav_dls_wls_v_g(
          m_cov         = lavsamplestats@cov[[g]],
          m_mean        = lavsamplestats@mean[[g]],
          nacov_g       = lavsamplestats@NACOV[[g]],
          dls_a         = lavoptions$estimator.args$dls.a,
          x_idx         = lavsamplestats@x.idx[[g]],
          fixed_x       = lavoptions$fixed.x,
          conditional_x = lavoptions$conditional.x,
          meanstructure = lavoptions$meanstructure
        )
      }
    }
  }

  # se
  if (se) {
    if (lavoptions$se == "none") {
      lavoptions$se <- "standard"
    }
  } else {
    # the baseline standard errors are never used (only the baseline
    # test statistics feed CFI/TLI and friends), so skip the vcov step
    # (since 0.7-1); the scaled/robust tests recompute E.inv, Delta and
    # WLS.V themselves when no vcov is available -- as has always been
    # the case for se = "bootstrap", where the baseline was already
    # fitted with se = "none" since 0.6-20
    # exception: two.stage/robust.two.stage, where the scaled test
    # reuses the two-stage vcov ingredients; keep se as-is there
    if (!any(lavoptions$missing == c("two.stage", "robust.two.stage"))) {
      lavoptions$se <- "none"
    }
  }

  # change options
  lavoptions$h1 <- FALSE # already provided by lavh1
  lavoptions$baseline <- FALSE # of course
  lavoptions$fit.by.level <- FALSE
  lavoptions$loglik <- TRUE # eg for multilevel
  lavoptions$implied <- TRUE # needed for loglik (multilevel)
  lavoptions$check.start <- FALSE
  lavoptions$check.gradient <- FALSE
  lavoptions$check.post <- FALSE
  lavoptions$check.vcov <- FALSE
  lavoptions$optim.bounds <- list() # we already have the bounds
  lavoptions$start <- "default" # don't re-use user-specified starting values
  lavoptions$rstarts <- 0L # no random starts

  # ALWAYS do.fit and set optim.method = "nlminb" (if npar > 0)
  npar <- lav_pt_npar(lavpartable)
  if (npar > 0L) {
    lavoptions$do.fit <- TRUE
    if (lavoptions$optim.method != "noniter") {
      lavoptions$optim.method <- "nlminb"
    }
  } else {
    # perhaps a correlation structure?
    lavoptions$optim.method <- "none"
    lavoptions$optim.force.converged <- TRUE
  }

  # needed?
  if (any(lavpartable$op == "~1")) lavoptions$meanstructure <- TRUE

  # FIXME: it is crucial that the order of the ov's, as returned by
  # lav_object_vnames() remains the same
  # so lav_object_vnames(object) should equal lav_object_vnames(lavpartable)
  # otherwise, we will use the wrong sample statistics!!!
  #
  # this seems ok now, because we first generate the covariances in
  # lavpartable, and they should be in the right order (unlike the
  # intercepts)

  fit <- lavaan(lavpartable,
    slot_options     = lavoptions,
    slot_sample_stats = lavsamplestats,
    slot_data        = lavdata,
    slot_cache       = lavcache,
    sloth1          = lavh1
  )

  fit
}


# 3. extended model (used for modification indices)
lav_object_extended <- function(object, add = NULL,
                                remove_duplicated = TRUE,
                                all_free = FALSE,
                                do_fit = FALSE) {
  object <- lav_object_check_version(object)
  # partable original model
  partable <- object@ParTable[c(
    "lhs", "op", "rhs", "free", "exo", "label",
    "plabel"
  )]

  # new in 0.6-3: check for non-parameters
  nonpar_idx <- which(partable$op %in% c("==", ":=", "<", ">"))

  # always add block/group/level
  if (!is.null(object@ParTable$group)) {
    partable$group <- object@ParTable$group
  } else {
    partable$group <- rep(1L, length(partable$lhs))
    if (length(nonpar_idx) > 0L) {
      partable$group[nonpar_idx] <- 0L
    }
  }
  if (!is.null(object@ParTable$level)) {
    partable$level <- object@ParTable$level
  } else {
    partable$level <- rep(1L, length(partable$lhs))
    if (length(nonpar_idx) > 0L) {
      partable$level[nonpar_idx] <- 0L
    }
  }
  if (!is.null(object@ParTable$block)) {
    partable$block <- object@ParTable$block
  } else {
    partable$block <- rep(1L, length(partable$lhs))
    if (length(nonpar_idx) > 0L) {
      partable$block[nonpar_idx] <- 0L
    }
  }

  # TDJ: Added to prevent error when lav_pt_merge() is called below.
  #      Problematic if object@ParTable is missing one of the requested slots,
  #      which returns a NULL slot with a missing <NA> name.  For example:
  #        example(cfa)
  #        lav_pt_independence(lavdata = fit@Data, lavpta = fit@pta,
  #                                  lavoptions = lavInspect(fit, "options"))
  #     Has no "label" or "plabel" elements.
  empties <- which(sapply(partable, is.null))
  if (length(empties)) {
    partable[empties] <- NULL
  }

  if (all_free) {
    partable$user <- rep(1L, length(partable$lhs))
    non_free_idx <- which(partable$free == 0L & partable$op != "==" &
      partable$op != ":=" & partable$op != "<" &
      partable$op != ">")
    partable$free[non_free_idx] <- 1L
    partable$user[non_free_idx] <- 10L
  }

  # replace 'start' column, since lav_model will fill these in in GLIST
  partable$start <- lavParameterEstimates(object,
    remove_system_eq = FALSE,
    remove_def = FALSE,
    remove_eq = FALSE,
    remove_ineq = FALSE,
    remove_nonfree = FALSE,
    remove_unused = FALSE
  )$est

  # add new parameters, extend model
  if (is.list(add)) {
    stopifnot(
      !is.null(add$lhs),
      !is.null(add$op),
      !is.null(add$rhs)
    )
    add_1 <- add
  } else if (is.character(add)) {
    ngroups <- lav_pt_ngroups(partable)
    add_orig <- lav_model_pt(add, ngroups = ngroups)
    add_1 <- add_orig[, c("lhs", "op", "rhs", "user", "label")] # minimum

    # always add block/group/level
    if (!is.null(add_orig$group)) {
      add_1$group <- add_orig$group
    } else {
      add_1$group <- rep(1L, length(add_1$lhs))
    }
    if (!is.null(add_orig$level)) {
      add_1$level <- add_orig$level
    } else {
      add_1$level <- rep(1L, length(add_1$lhs))
    }
    if (!is.null(add_orig$block)) {
      add_1$block <- add_orig$block
    } else {
      add_1$block <- rep(1L, length(add_1$lhs))
    }

    remove_idx <- which(add_1$user == 0)
    if (length(remove_idx) > 0L) {
      add_1 <- add_1[-remove_idx, ]
    }
    add_1$start <- rep(0, nrow(add_1))
    add_1$free <- rep(1, nrow(add_1))
    add_1$user <- rep(10, nrow(add_1))
  }

  # merge
  list_1 <- lav_pt_merge(partable, add_1,
    remove_duplicated = remove_duplicated,
    warn = FALSE
  )

  # remove nonpar?
  # if(remove.nonpar) {
  #    nonpar.idx <- which(LIST$op %in% c("==", ":=", "<", ">"))
  #    if(length(nonpar.idx) > 0L) {
  #        LIST <- LIST[-nonpar.idx,]
  #    }
  # }

  # redo 'free'
  free_idx <- which(list_1$free > 0)
  list_1$free[free_idx] <- seq_along(free_idx)

  # adapt options
  lavoptions <- object@Options

  # do.fit?
  lavoptions$do.fit <- do_fit

  # needed?
  if (any(list_1$op == "~1")) lavoptions$meanstructure <- TRUE

  # switch off 'consider switching to parameterization = theta' warning
  # (modindices!)
  lavoptions$check.delta.cat.mediator <- FALSE

  # no need for the partially saturated models (multilevel)
  lavoptions$fit.by.level <- FALSE

  fit <- lavaan(list_1,
    slot_options     = lavoptions,
    slot_sample_stats = object@SampleStats,
    slot_data        = object@Data,
    slot_cache       = object@Cache,
    sloth1          = object@h1
  )

  fit
}

# 4. catml model
# allow_refit: if an input correlation matrix is not positive definite,
# it is always smoothed; but the model parameters are only re-estimated
# if allow_refit = TRUE (if FALSE, we keep the original -- typically
# DWLS -- estimates, and merely re-evaluate; as in lavaan <= 0.6-13)
lav_object_catml <- function(lavobject = NULL, allow_refit = TRUE) {
  stopifnot(inherits(lavobject, "lavaan"))
  stopifnot(lavobject@Model@categorical)

  # extract slots
  lavdata <- lavobject@Data
  lavsamplestats <- lavobject@SampleStats
  lavoptions <- lavobject@Options
  lavpta <- lavobject@pta

  # if only categorical variables: remove thresholds and intercepts
  refit <- FALSE
  partable_catml <- parTable(lavobject)
  rm_idx <- which(partable_catml$op %in% c("|", "~1"))
  # remove also any constraints that refer to the labels that belong
  # to these intercepts/thresholds (new in 0.6-20)
  th_int_labels <- unique(
    partable_catml$label[rm_idx],
    partable_catml$plabel[rm_idx]
  )
  if (any(nchar(th_int_labels) == 0L)) {
    th_int_labels <- th_int_labels[-which(nchar(th_int_labels) == 0L)]
  }
  con_idx <- which(partable_catml$op %in% c("==", "<", ">", ":=") &
    (partable_catml$lhs %in% th_int_labels |
      partable_catml$rhs %in% th_int_labels))
  partable_catml <- partable_catml[-c(rm_idx, con_idx), ]
  # never rotate (new in 0.6-19), as we only need fit measures
  if (!is.null(partable_catml$efa)) {
    partable_catml$efa <- NULL
    if (!is.null(partable_catml$free.unrotated)) { # e.g., rotation = "none"
      partable_catml$free <- partable_catml$free.unrotated
    }
  }

  if (!all(lavdata@ov$type == "ordered")) {
    refit <- TRUE
    partable_catml$start <- partable_catml$est
    partable_catml$se <- NULL
    partable_catml$ustart <- partable_catml$est
    for (b in seq_len(lavpta$nblocks)) {
      ov_names_num <- lavpta$vnames$ov.num[[b]]
      ov_var_idx <- which(partable_catml$op == "~~" &
        partable_catml$lhs %in% ov_names_num &
        partable_catml$lhs == partable_catml$rhs)
      partable_catml$free[ov_var_idx] <- 0L
    }
  }
  # re-attach the data-based ov order (ov_order = "data"). parTable() has
  # stripped the "ovda" attribute and the row-subsetting above drops
  # attributes, so without this the model would be rebuilt in model order
  # while the reused sample statistics / data remain in data order, leading
  # to a mismatched (inflated) catML chi-square. Harmless no-op when the
  # data order already equals the model order.
  attr(partable_catml, "ovda") <- lavdata@ov.names[[1]]

  partable_catml <- lav_pt_complete(partable_catml)

  # adapt lavsamplestats
  for (g in seq_len(lavdata@ngroups)) {
    lavsamplestats@WLS.V[[g]] <- NULL
    lavsamplestats@WLS.VD[[g]] <- NULL
    cor_1 <- lavsamplestats@cov[[g]]
    # check if COV is pd or not
    ev <- eigen(cor_1, symmetric = TRUE, only.values = TRUE)$values
    if (any(ev < .Machine$double.eps^(1 / 2))) {
      # not PD!
      cov_1 <- cov2cor(lav_mat_sym_force_pd(cor_1, tol = 1e-04))
      lavsamplestats@cov[[g]] <- cov_1
      lavsamplestats@var[[g]] <- diag(cov_1)
      if (allow_refit) {
        refit <- TRUE
      }
    } else {
      cov_1 <- cor_1
    }

    current_warn <- lav_warn()
    if (lav_warn(FALSE)) {
      on.exit(lav_warn(current_warn), TRUE)
    }
    out <- lav_samp_icov(
      cov_1 = cov_1, ridge = 1e-05,
      x_idx = lavsamplestats@x.idx[[g]],
      ngroups = lavdata@ngroups, g = g
    )
    lav_warn(current_warn)
    lavsamplestats@icov[[g]] <- out$icov
    lavsamplestats@cov.log.det[[g]] <- out$cov.log.det

    # NACOV <- lavsamplestats@NACOV[[g]]
    # nvar   <- nrow(COV)
    # ntotal <- nrow(NACOV)
    # pstar <- nvar*(nvar-1)/2
    # nocor <- ntotal - pstar
    # if(length(nocor) > 0L) {
    #    lavsamplestats@NACOV[[g]] <- NACOV[-seq_len(nocor),
    #                                       -seq_len(nocor)]
    # }
  }

  # adapt lavoptions
  lavoptions$estimator <- "catML"
  lavoptions$.categorical <- FALSE
  lavoptions$categorical <- FALSE
  lavoptions$correlation <- TRUE
  lavoptions$meanstructure <- FALSE
  lavoptions$conditional.x <- FALSE # fixme
  lavoptions$information <- c("expected", "expected")
  lavoptions$h1.information <- c("structured", "structured") # unlike DWLS
  lavoptions$se <- "none"
  lavoptions$test <- "standard" # always for now
  lavoptions$rotation <- "none" # new in 0.6-19
  lavoptions$baseline <- TRUE # also for RMSEA?
  # the robust corrections assume the (default) independence baseline
  # (Savalei, 2021, eq. 6b); never use a 'nested' baseline here
  lavoptions$baseline.type <- "independence"
  lavoptions$fit.by.level <- FALSE
  if (!refit) {
    lavoptions$optim.method <- "none"
    lavoptions$optim.force.converged <- TRUE
  } else {
    lavoptions$optim.gradient <- "numerical"
  }

  # dummy fit
  fit <- lavaan(
    slot_par_table = partable_catml,
    slot_sample_stats = lavsamplestats,
    slot_data = lavdata,
    slot_options = lavoptions
  )

  fit
}

# synonym (thin forwarder; see lav_alias in 00alias.R)
lav_object_independence <- lav_alias("lav_object_baseline")

# collection of functions that deal with rotation matrices
# YR  3 April 2019 -- initial version
# YR  6 Jan 2023: add promax
# YR 30 Jan 2024: orthogonal rotation matrix now reaches the full space

# generate random orthogonal rotation matrix
#
# reference for the orthogonal case:
#
# Stewart, G. W. (1980). The Efficient Generation of Random Orthogonal Matrices
# with an Application to Condition Estimators. SIAM Journal on Numerical
# Analysis, 17(3), 403-409. http://www.jstor.org/stable/2156882
#
lav_mat_rotate_gen <- function(m = 10L, orthogonal = TRUE) {
  # catch M=1
  if (m == 1L) {
    return(matrix(1, 1, 1))
  }

  # create random normal matrix
  tmp <- matrix(rnorm(m * m), nrow = m, ncol = m)

  if (orthogonal) {
    # use QR decomposition
    qr_out <- qr(tmp)
    q_1 <- qr.Q(qr_out)
    r <- qr.R(qr_out)
    # ... "normalized so that the diagonal elements of R are positive"
    sign_diag_r <- sign(diag(r))
    out <- q_1 * rep(sign_diag_r, each = m)
  } else {
    # just normalize *columns* of tmp -> crossprod(out) has 1 on diagonal
    out <- t(t(tmp) / sqrt(diag(crossprod(tmp))))
  }

  out
}

# check if ROT is an orthogonal matrix if orthogonal = TRUE, or normal if
# orthogonal = FALSE
lav_mat_rotate_check <- function(rot = NULL, orthogonal = TRUE,
                                    tolerance = sqrt(.Machine$double.eps)) {
  # we assume ROT is a matrix
  m <- nrow(rot)

  # crossprod
  rr <- crossprod(rot)

  # target
  if (orthogonal) {
    # ROT^T %*% ROT = I
    target <- diag(m)
  } else {
    # diagonal should be 1
    target <- rr
    diag(target) <- 1
  }

  # compare for near-equality
  res <- all.equal(target = target, current = rr, tolerance = tolerance)

  # return TRUE or FALSE
  if (is.logical(res) && res) {
    out <- TRUE
  } else {
    out <- FALSE
  }

  out
}

# get weights vector needed to weight the rows using Kaiser normalization
lav_mat_rotate_kaiser_weights <- function(a = NULL) {
  normalize <- 1 / sqrt(rowSums(a * a))
  idx_zero <- which(normalize == 0)
  # catch rows with all zero (thanks to Coen Bernaards for suggesting this)
  normalize[idx_zero] <- normalize[idx_zero] + .Machine$double.eps
  normalize
}

# get weights vector needed to weight the rows using Cureton & Mulaik (1975)
# standardization
# see also Browne (2001) page 128-129
#
# Note: the 'final' weights are multiplied by the Kaiser weights (see CEFA)
#
lav_mat_rotate_cm_weights <- function(a = NULL) {
  p <- nrow(a)
  m <- ncol(a)

  # first principal component of AA'
  a_eigen <- eigen(tcrossprod(a), symmetric = TRUE)
  a_1 <- a_eigen$vectors[, 1] * sqrt(a_eigen$values[1])

  kaiser_weights <- 1 / sqrt(rowSums(a * a))
  a_star <- abs(a_1 * kaiser_weights) # always between 0 and 1

  m_sqrt_inv <- 1 / sqrt(m)
  acos_m_sqrt_inv <- acos(m_sqrt_inv)

  delta <- numeric(p)
  delta[a_star < m_sqrt_inv] <- pi / 2

  tmp <- (acos_m_sqrt_inv - acos(a_star)) / (acos_m_sqrt_inv - delta) * (pi / 2)

  # add constant (see Cureton & Mulaik, 1975, page 187)
  cm <- cos(tmp) * cos(tmp) + 0.001

  # final weights = weighted by Kaiser weights
  cm * kaiser_weights
}

# build an orthogonal Givens (plane) rotation matrix: an m x m identity with a
# 2x2 rotation by 'theta' in the (col1, col2) plane.
lav_mat_givens_orth <- function(m = 2L, col1 = 1L, col2 = 2L, theta = 0) {
  rot <- diag(m)
  rot[col1, col1] <- base::cos(theta)
  rot[col1, col2] <- base::sin(theta)
  rot[col2, col1] <- -1 * base::sin(theta)
  rot[col2, col2] <- base::cos(theta)
  rot
}

# build an oblique (plane) rotation matrix for the (col1, col2) plane, given the
# shear 'delta' and the current factor correlation phi12. rot[col1, col1] equals
# the scaling factor 'gamma' (= sqrt(abs(gamma2))) used to update PHI.
lav_mat_givens_obliq <- function(m = 2L, col1 = 1L, col2 = 2L, delta = 0,
                                 phi12 = 0) {
  rot <- diag(m)
  gamma2 <- 1 + (2 * delta * phi12) + (delta * delta)
  rot[col1, col1] <- sqrt(abs(gamma2))
  rot[col1, col2] <- -1 * delta
  rot[col2, col1] <- 0
  rot[col2, col2] <- 1
  rot
}

# resolve the rotation criterion shared by lav_mat_rotate() and
# lav_mat_rotate_mg(): map the (already lower-cased) method name to its gradient
# function ('method_fname'), fill in the method-specific cf_gamma for the
# Crawford-Ferguson family, validate target/target_mask matrices (dimensions
# p x m), and promote a 'target.strict' with NA entries to 'pst'. 'group'
# selects the per-group matrix when target/target_mask are given as a list.
# Returns the (possibly updated) method, method_fname and method_args.
lav_mat_rotate_resolve_method <- function(method = NULL, method_args = list(),
                                          p = 0L, m = 0L, group = 1L) {
  # determine method function name
  if (method %in% c(
    "cf-quartimax", "cf-varimax", "cf-equamax",
    "cf-parsimax", "cf-facparsim"
  )) {
    method_fname <- "lav_mat_rotate_cf"
    method_args$cf_gamma <- switch(method,
      "cf-quartimax" = 0,
      "cf-varimax"   = 1 / p,
      "cf-equamax"   = m / (2 * p),
      "cf-parsimax"  = (m - 1) / (p + m - 2),
      "cf-facparsim" = 1
    )
  } else if (method %in% c("bi-quartimin", "biquartimin")) {
    method_fname <- "lav_mat_rotate_biquartimin"
  } else if (method %in% c("bi-geomin", "bigeomin")) {
    method_fname <- "lav_mat_rotate_bigeomin"
  } else if (method == "target.strict") {
    method_fname <- "lav_mat_rotate_target"
  } else {
    method_fname <- paste("lav_mat_rotate_", method, sep = "")
  }

  # check if rotation method exists
  check <- try(get(method_fname), silent = TRUE)
  if (inherits(check, "try-error")) {
    lav_msg_stop(gettext("unknown rotation method:"), method_fname)
  }

  # if target, check target matrix
  if (method == "target.strict" || method == "pst") {
    target <- method_args$target
    if (is.list(target)) {
      method_args$target <- target <- target[[group]]
    }
    # check dimension of target/A
    if (nrow(target) != p) {
      lav_msg_stop(gettext("nrow(target) != nrow(A)"))
    }
    if (ncol(target) != m) {
      lav_msg_stop(gettext("ncol(target) != ncol(A)"))
    }
  }
  if (method == "pst") {
    target_mask <- method_args$target_mask
    if (is.list(target_mask)) {
      method_args$target_mask <- target_mask <- target_mask[[group]]
    }
    # check dimension of target_mask/A
    if (nrow(target_mask) != p) {
      lav_msg_stop(gettext("nrow(target_mask) != nrow(A)"))
    }
    if (ncol(target_mask) != m) {
      lav_msg_stop(gettext("col(target_mask) != ncol(A)"))
    }
  }
  # we keep this here, so lav_mat_rotate() can be used independently
  if (method == "target.strict" && anyNA(target)) {
    method <- "pst"
    method_fname <- "lav_mat_rotate_pst"
    target_mask <- matrix(1, nrow = nrow(target), ncol = ncol(target))
    target_mask[is.na(target)] <- 0
    method_args$target_mask <- target_mask
  }

  list(method = method, method_fname = method_fname,
       method_args = method_args)
}

# taken from the stats package, but skipping varimax (already done):
lav_mat_rotate_promax <- function(x, m = 4, varimax_rot = NULL) {
  # this is based on promax() from factanal.R in /src/library/stats/R

  # 1. create 'ideal' pattern matrix
  q_1 <- x * abs(x) ^ (m - 1)

  # 2. regress x on Q to obtain 'rotation matrix' (same as 'procrustes')
  u <- lm.fit(x, q_1)$coefficients

  # 3. rescale so that solve(crossprod(U)) has 1 on the diagonal

  d <- diag(solve(t(u) %*% u))
  u <- u %*% diag(sqrt(d))
  dimnames(u) <- NULL

  # 4. create rotated factor matrix
  z <- x %*% u

  # 5. update rotation matrix
  u <- varimax_rot %*% u # here we plugin the rotation matrix from varimax

  list(loadings = z, rotmat = u)
}

# Gamma: N times the asymptotic variance matrix of the sample statistics
# (means/thresholds/slopes first, then vech of the (co)variances).
#
# Two computational routes exist, tied together by the identity
#
#     Gamma_ADF = A1^{-1} %*% J1 %*% A1^{-1}
#
# with A1 the expected h1 information (A1^{-1} = Gamma_NT) and J1 the
# first-order h1 information (crossprod of the casewise h1 scores / n),
# both at the saturated sample moments:
#
# - the DIRECT route (lav_samp_gamma below): crossprod of the casewise
#   moment contributions. Since sc_i = A1 %*% zc_i, this equals the
#   sandwich exactly (machine precision) and is 1.5-3x faster (the
#   sandwich adds two pstar x pstar multiplications); it is the canonical
#   engine for complete single-level data.
# - the SANDWICH route: needed whenever the casewise contributions are not
#   iid draws of a simple moment vector -- sampling weights (the wt branch
#   of lav_samp_gamma below), missing data (lav_mvn_mi_h1_omega_sw), and
#   two-level data (lav_samp_gamma_2l_g), where it is the only route.
#
# (evidence + benchmarks: aja_identity.R in the gamma-refactor test dir)
#
# SCALING CONVENTIONS
# - @SampleStats@NACOV[[g]] stores Gamma_g = n_g * acov(s_g)
#   = Cov(sqrt(n_g) * s_g), per group. (For two-level data n_g is the
#   number of OBSERVATIONS, even though the clusters are the independent
#   units: Gamma_g = n_g * acov always holds.)
# - the h1 information matrices (lav_model_h1_information.R) are UNIT
#   (per-observation single-level, per-cluster two-level); acov versions
#   (lav_model_h1_acov) carry the 1/n_g; the categorical WLS.W is stored
#   as Gamma/n and rescaled at consumption.
# - multigroup consumers stack per-group blocks into one 'global' Gamma.
#   Two equivalent conventions are in use:
#     (a) Gamma_g / fg  (= Cov(sqrt(ntotal) * s_g)) with fg-weighted
#         information/weight blocks appearing TWICE in U
#         (satorra.bentler, yuan.bentler, catml_dwls fit indices);
#     (b) fg * Gamma_g with UNweighted U blocks (lav_test_diff).
#   These give identical tr(U Gamma) and tr((U Gamma)^2): every (g,h)
#   cross-term picks up exactly one fg and one fh under either placement
#   (verified numerically in fg_convention_check.R, gamma-refactor dir).

# rescale a per-group Gamma list from the Cov(sqrt(n_g) s_g) convention
# (as stored in @NACOV) to the Cov(sqrt(ntotal) s_g) convention
# (Gamma_g / fg); convention (a) above, used when stacking the per-group
# blocks into one 'global' Gamma
lav_gamma_rescale_ntotal <- function(gamma, nobs, ntotal = NULL) {
  nobs <- unlist(nobs)
  if (is.null(ntotal)) {
    ntotal <- sum(nobs)
  }
  fg <- nobs / ntotal
  for (g in seq_along(gamma)) {
    gamma[[g]] <- gamma[[g]] / fg[g]
  }
  gamma
}

# 'the' Gamma (NACOV) of a fitted object: the fit-time stored NACOV when
# available, otherwise recomputed with the fit-time settings (the recipe).
# The single entry point for consumers that need "the Gamma that was (or
# would have been) used by this object" (satorra.bentler/fmg tests, sam,
# lavInspect "gamma", ...).
lav_gamma_used <- function(lavobject = NULL,
                           # or individual slots
                           lavdata = NULL,
                           lavoptions = NULL,
                           lavsamplestats = NULL,
                           lavh1 = NULL,
                           lavimplied = NULL) {
  if (!is.null(lavobject)) {
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
    lavsamplestats <- lavobject@SampleStats
    lavh1 <- lavobject@h1
    lavimplied <- lavobject@implied
  }

  # fit-time NACOV?
  if (length(lavsamplestats@NACOV) > 0L &&
      !is.null(lavsamplestats@NACOV[[1]])) {
    return(lavsamplestats@NACOV)
  }

  # recompute with the fit-time settings
  lav_object_gamma(
    lavdata = lavdata,
    lavoptions = lavoptions,
    lavsamplestats = lavsamplestats,
    lavh1 = lavh1,
    lavimplied = lavimplied,
    adf = TRUE,
    model_based = FALSE
  )
}

# the ADF 'meat': crossproduct of the (centered) casewise moment
# contributions zc, divided by n; when cluster_idx is given, the
# contributions are aggregated within clusters first (the G/(G-1)
# finite-sample correction is applied by the caller). Shared by the three
# (plain / fixed.x / conditional.x) branches of lav_samp_gamma().
lav_samp_gamma_meat <- function(zc, cluster_idx = NULL, n = nrow(zc)) {
  if (length(cluster_idx) > 0L) {
    zc <- rowsum(zc, cluster_idx)
  }
  if (anyNA(zc)) {
    lav_mat_crossprod(zc) / n
  } else {
    base::crossprod(zc) / n
  }
}

# weighted crossproduct of casewise scores (the 'meat' of a sandwich) under
# sampling weights; "design" weights the crossproduct by wt^2, "frequency"
# by wt (as if each case were physically replicated wt times). Shared by
# the ADF Gamma and the (PML) first-order h1 information.
lav_samp_wt_crossprod <- function(sc, wt, sampling_weights_type = "design") {
  if (identical(sampling_weights_type, "frequency")) {
    crossprod(sqrt(wt) * sc)
  } else {
    crossprod(wt * sc)
  }
}

# DLS with a model-based GammaNT: the (inverted) weight matrix for one
# group, W_DLS^{-1}, where
#     W_DLS = (1 - a) * Gamma_ADF (= NACOV) + a * Gamma_NT(m_cov, m_mean)
# shared by the objective, gradient, h1-information and object-generate code
lav_dls_wls_v_g <- function(m_cov = NULL, m_mean = NULL,
                            nacov_g = NULL, dls_a = 1.0,
                            x_idx = integer(0L), fixed_x = FALSE,
                            conditional_x = FALSE, meanstructure = FALSE) {
  gamma_nt <- lav_samp_gamma_nt(
    m_cov          = m_cov,
    m_mean         = m_mean,
    x_idx          = x_idx,
    fixed_x        = fixed_x,
    conditional_x  = conditional_x,
    meanstructure  = meanstructure,
    slopestructure = conditional_x
  )
  w_dls <- (1 - dls_a) * nacov_g + dls_a * gamma_nt
  lav_mat_sym_inverse(w_dls)
}

# for internal use -- lavobject or internal slots
lav_object_gamma <- function(lavobject = NULL,
                             # or individual slots
                             lavdata = NULL,
                             lavoptions = NULL,
                             lavsamplestats = NULL,
                             lavh1 = NULL,
                             lavimplied = NULL,
                             # other options
                             adf = TRUE, model_based = FALSE,
                             mplus_wls = NULL) {
  # extract slots
  if (!is.null(lavobject)) {
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
    lavsamplestats <- lavobject@SampleStats
    lavh1 <- lavobject@h1
    lavimplied <- lavobject@implied
  }

  # the Gamma recipe: the same settings as used at fit time
  recipe <- lav_gamma_recipe(lavoptions = lavoptions, lavdata = lavdata,
                             nacov_compute = TRUE)

  missing <- recipe$missing
  fixed_x <- recipe$fixed.x
  conditional_x <- recipe$conditional.x
  meanstructure <- recipe$meanstructure
  gamma_n_minus_one <- recipe$n.minus.one
  gamma_unbiased <- recipe$unbiased
  if (is.null(mplus_wls)) {
    # gamma.wls.mplus, but only where fit time applies it (plain ADF Gamma
    # for the (D)WLS-family estimators)
    mplus_wls <- recipe$mplus.wls
  }

  # output container
  out <- vector("list", length = lavdata@ngroups)

  # ---- two-level data: cluster sandwich at the h1 estimates --------------
  if (recipe$multilevel) {
    if (recipe$categorical) {
      lav_msg_stop(gettext(
        "Gamma can not be (re)computed for two-level categorical data; the
         (stage-wise) Gamma used at fit time (if any) is stored in the
         NACOV slot."))
    }
    if (model_based || !adf) {
      lav_msg_stop(gettext(
        "only the sandwich-type Gamma (at the saturated h1 estimates) is
         available for two-level data."))
    }
    if (is.null(lavh1) || length(lavh1$implied) == 0L) {
      lav_msg_stop(gettext(
        "the h1 (saturated) estimates are needed to compute the two-level
         Gamma, but they are not available in this object."))
    }
    for (g in seq_len(lavdata@ngroups)) {
      gamma_g <- lav_samp_gamma_2l_g(
        lavsamplestats = lavsamplestats,
        lavh1 = lavh1,
        lavdata = lavdata,
        g = g
      )
      attr(gamma_g, "keep") <- NULL
      out[[g]] <- gamma_g
    }
    return(out)
  }

  # ---- categorical data (single-level): muthen1984 three-stage -----------
  if (recipe$categorical) {
    if (!adf || model_based) {
      lav_msg_stop(gettext(
        "only the sample-based (ADF-type) Gamma is available for
         categorical data."))
    }
    data_ov <- lavdata@ov
    for (g in seq_len(lavdata@ngroups)) {
      ov_names_g <- lavdata@ov.names[[g]]
      ov_types_g <- data_ov$type[match(ov_names_g, data_ov$name)]
      ov_levels_g <- data_ov$nlev[match(ov_names_g, data_ov$name)]
      if (length(lavdata@cluster) > 0L) {
        cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
      } else {
        cluster_idx <- NULL
      }
      cat_1 <- muthen1984(
        data_1 = lavdata@X[[g]],
        wt = lavdata@weights[[g]],
        sampling_weights_type = recipe$swt.type,
        ov_names = ov_names_g,
        ov_types = ov_types_g,
        ov_levels = ov_levels_g,
        ov_names_x = if (conditional_x) lavdata@ov.names.x[[g]] else NULL,
        exo = if (conditional_x) lavdata@eXo[[g]] else NULL,
        group = g, # for error messages only
        wls_w = TRUE,
        zero_add = lavoptions$zero.add,
        zero_keep_margins = lavoptions$zero.keep.margins,
        zero_cell_warn = FALSE,
        zero_cell_tables = TRUE,
        allow_empty_cell = lavoptions$allow.empty.cell,
        cluster_idx = cluster_idx
      )
      nobs_g <- lavsamplestats@nobs[[g]]
      if (!is.null(cat_1$WLS.W.cluster)) {
        # cluster-robust meat + G/(G-1) correction, as at fit time
        nc <- lavdata@Lp[[g]]$nclusters[[2]]
        gamma_g <- cat_1$WLS.W.cluster * nobs_g * (nc / (nc - 1))
      } else {
        gamma_g <- cat_1$WLS.W * nobs_g
        if (gamma_n_minus_one) {
          gamma_g <- gamma_g * (nobs_g / (nobs_g - 1L))
        }
      }
      if (lavoptions$estimator == "catML") {
        # remove all but the correlation part, as at fit time
        ntotal <- nrow(gamma_g)
        pstar <- nrow(cat_1$A22)
        nocor <- ntotal - pstar
        if (length(nocor) > 0L) {
          gamma_g <- gamma_g[-seq_len(nocor), -seq_len(nocor)]
        }
      }
      if (recipe$group.w.free) {
        gamma_g <- lav_mat_bdiag(matrix(1, 1, 1), gamma_g)
      }
      out[[g]] <- gamma_g
    }
    return(out)
  }

  # ---- incomplete data (ml / two.stage / robust.two.stage) ---------------
  # Gamma = N x acov of the saturated estimates under missingness, from the
  # missing-data h1 information at the EM estimates (or at the model-implied
  # moments if model_based): the sandwich I^{-1} J I^{-1} (adf) or the
  # inverted observed information I^{-1} (not adf). This mirrors both the
  # fit-time NACOV of the two.stage least-squares estimators and the
  # two-stage robust vcov machinery.
  if (!missing %in% c("listwise", "pairwise")) {
    sandwich <- adf
    if (missing == "two.stage") {
      sandwich <- FALSE
    } else if (missing == "robust.two.stage") {
      sandwich <- TRUE
    }
    for (g in seq_len(lavdata@ngroups)) {
      if (conditional_x) {
        # conditional metric: Gamma is the N x acov of the saturated
        # (vec(Beta), vech(res.cov)) estimates (x is complete)
        if (model_based) {
          res_int_g <- lavimplied$res.int[[g]]
          res_slopes_g <- lavimplied$res.slopes[[g]]
          res_cov_g <- lavimplied$res.cov[[g]]
        } else {
          res_int_g <- lavsamplestats@missing.h1[[g]]$res.int
          res_slopes_g <- lavsamplestats@missing.h1[[g]]$res.slopes
          res_cov_g <- lavsamplestats@missing.h1[[g]]$res.cov
        }
        if (is.null(res_cov_g)) {
          lav_msg_stop(gettext(
            "Gamma is not available: no EM estimates of the saturated
            conditional moments (zero coverage?)."))
        }
        if (sandwich) {
          gamma_g <- lav_mvreg_mi_h1_omega_sw(
            y = lavdata@X[[g]], exo = lavdata@eXo[[g]],
            mp = lavdata@Mp[[g]],
            yp = lavsamplestats@missing[[g]], wt = lavdata@weights[[g]],
            res_int = res_int_g, res_slopes = res_slopes_g,
            res_cov = res_cov_g,
            information = "observed"
          )
        } else {
          i1 <- lav_mvreg_mi_information_observed_samplestats(
            yp = lavsamplestats@missing[[g]],
            res_int = res_int_g, res_slopes = res_slopes_g,
            res_cov = res_cov_g
          )
          gamma_g <- lav_mat_sym_inverse(i1)
        }
        if (recipe$group.w.free) {
          gamma_g <- lav_mat_bdiag(matrix(1, 1, 1), gamma_g)
        }
        out[[g]] <- gamma_g
        next
      }
      if (model_based) {
        mu_g <- lavimplied$mean[[g]]
        sigma_g <- lavimplied$cov[[g]]
      } else {
        mu_g <- lavsamplestats@missing.h1[[g]]$mu
        sigma_g <- lavsamplestats@missing.h1[[g]]$sigma
      }
      x_idx_g <- if (fixed_x) lavsamplestats@x.idx[[g]] else integer(0L)
      if (sandwich) {
        gamma_g <- lav_mvn_mi_h1_omega_sw(
          y = lavdata@X[[g]], mp = lavdata@Mp[[g]],
          yp = lavsamplestats@missing[[g]], wt = lavdata@weights[[g]],
          mu = mu_g, sigma_1 = sigma_g, x_idx = x_idx_g,
          information = "observed"
        )
      } else {
        i1 <- lav_mvnorm_missing_information_observed_samplestats(
          yp = lavsamplestats@missing[[g]],
          mu = mu_g, sigma_1 = sigma_g, x_idx = x_idx_g
        )
        gamma_g <- lav_mat_sym_inverse(i1)
      }
      if (recipe$group.w.free) {
        gamma_g <- lav_mat_bdiag(matrix(1, 1, 1), gamma_g)
      }
      out[[g]] <- gamma_g
    }
    return(out)
  }

  # ---- continuous, complete-data, single-level ----------------------------
  if (adf && lavdata@data.type != "full") {
    lav_msg_stop(gettext(
      "the (ADF) Gamma matrix can not be computed without full data; please
       provide the NACOV= argument."))
  }
  if (adf && model_based && conditional_x) {
    lav_msg_stop(gettext(
      "ADF + model.based + conditional.x is not supported yet."))
  }
  if (adf && model_based && recipe$correlation) {
    lav_msg_stop(gettext(
      "ADF + model.based is not supported yet for correlation structures."))
  }

  # compute Gamma matrix for each group
  for (g in seq_len(lavdata@ngroups)) {
    x_idx <- lavsamplestats@x.idx[[g]]
    cov_1 <- mean_1 <- NULL
    if (!adf || model_based) {
      implied <- lavh1$implied # saturated/unstructured
      if (model_based) {
        implied <- lavimplied # model-based/structured
      }
      if (conditional_x) {
        # convert to joint COV/MEAN
        res_s <- implied$res.cov[[g]]
        res_slopes <- implied$res.slopes[[g]]
        res_int <- implied$res.int[[g]]
        s_xx <- implied$cov.x[[g]]
        m_x <- implied$mean.x[[g]]

        s_yy <- res_s + res_slopes %*% s_xx %*% t(res_slopes)
        s_yx <- res_slopes %*% s_xx
        s_xy <- s_xx %*% t(res_slopes)
        m_y <- res_int + res_slopes %*% m_x

        cov_1 <- rbind(cbind(s_yy, s_yx), cbind(s_xy, s_xx))
        mean_1 <- c(m_y, m_x)
      } else {
        # not conditional.x
        cov_1 <- implied$cov[[g]]
        mean_1 <- implied$mean[[g]]
      }
    } # COV/MEAN

    if (adf) {
      if (conditional_x) {
        y <- cbind(lavdata@X[[g]], lavdata@eXo[[g]])
      } else {
        y <- lavdata@X[[g]]
      }
      if (length(lavdata@cluster) > 0L) {
        cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
      } else {
        cluster_idx <- NULL
      }
      if (recipe$correlation) {
        # (partial) correlation ADF Gamma via the delta method, as at fit
        # time (no cluster/weights support in the leaf function)
        cor_idx_g <- lav_gamma_recipe_cor_idx(recipe, lavdata@ov.names[[g]])
        out[[g]] <- lav_samp_partial_cor_gamma(
          m_y = y, cor_idx = cor_idx_g,
          meanstructure = meanstructure,
          fixed_x = fixed_x, x_idx = x_idx
        )
      } else {
        out[[g]] <- lav_samp_gamma(
          m_y = y,
          m_mu = mean_1,
          m_sigma = cov_1,
          x_idx = x_idx,
          cluster_idx = cluster_idx,
          fixed_x = fixed_x,
          conditional_x = conditional_x,
          meanstructure = meanstructure,
          slopestructure = conditional_x,
          gamma_n_minus_one = gamma_n_minus_one,
          unbiased = gamma_unbiased,
          mplus_wls = mplus_wls,
          wt = lavdata@weights[[g]],
          sampling_weights_type = recipe$swt.type
        )
      }
    } else {
      if (recipe$correlation) {
        cor_idx_g <- lav_gamma_recipe_cor_idx(recipe, lavdata@ov.names[[g]])
        out[[g]] <- lav_samp_partial_cor_gamma_nt(
          m_cov = cov_1, cor_idx = cor_idx_g,
          meanstructure = meanstructure,
          fixed_x = fixed_x, x_idx = x_idx
        )
      } else {
        out[[g]] <- lav_samp_gamma_nt(
          m_cov = cov_1, # joint!
          m_mean = mean_1, # joint!
          x_idx = x_idx,
          fixed_x = fixed_x,
          conditional_x = conditional_x,
          meanstructure = meanstructure,
          slopestructure = conditional_x
        )
      }
    }

    # group.w.free: prepend the group-weight element; the fit-time weight
    # a = group_w * ntotal / nobs is identically 1, so this matches the
    # NACOV stored at fit time
    if (recipe$group.w.free) {
      out[[g]] <- lav_mat_bdiag(matrix(1, 1, 1), out[[g]])
    }

  } # g

  out
}





# NOTE:
#  - three types:
#       1) plain         (conditional.x = FALSE, fixed.x = FALSE)
#       2) fixed.x       (conditional.x = FALSE, fixed.x = TRUE)
#       3) conditional.x (conditional.x = TRUE)
#  - if conditional.x = TRUE, we ignore fixed.x (can be TRUE or FALSE)

# the E[(1,x)(1,x)'] moment matrix of the (intercept, exo) regressors; the
# building block of the mean/slope block of the conditional.x NT Gamma
# (lav_samp_gamma_nt) and of its direct inverse (lav_samp_gamma_inverse_nt)
lav_samp_gamma_c3 <- function(m_mx, m_c) {
  rbind(
    c(1, m_mx),
    cbind(m_mx, m_c + tcrossprod(m_mx))
  )
}

# NORMAL-THEORY
lav_samp_gamma_nt <- function(m_y = NULL, # should include
                                     # eXo if
                                     # conditional.x=TRUE
                                     wt = NULL,
                                     m_cov = NULL, # joint!
                                     m_mean = NULL, # joint!
                                     rescale = FALSE,
                                     x_idx = integer(0L),
                                     fixed_x = FALSE,
                                     conditional_x = FALSE,
                                     meanstructure = FALSE,
                                     slopestructure = FALSE) {
  # check arguments
  if (length(x_idx) == 0L) {
    conditional_x <- FALSE
    fixed_x <- FALSE
  }

  # compute m_cov from m_y
  if (is.null(m_cov)) {
    stopifnot(!is.null(m_y))

    # coerce to matrix
    m_y <- unname(as.matrix(m_y))
    n <- nrow(m_y)
    if (is.null(wt)) {
      m_cov <- cov(m_y)
      if (rescale) {
        m_cov <- m_cov * (n - 1) / n # (normal) ML version
      }
    } else {
      out <- stats::cov.wt(m_y, wt = wt, method = "ML")
      m_cov <- out$cov
    }
  } else {
    if (!missing(rescale)) {
      lav_msg_warn(gettext("rescale= argument has no effect if m_cov is given"))
    }
    if (!missing(wt)) {
      lav_msg_warn(gettext("wt= argument has no effect if m_cov is given"))
    }
  }

  # if needed, compute m_mean from m_y
  if (conditional_x && length(x_idx) > 0L && is.null(m_mean) &&
    (meanstructure || slopestructure)) {
    stopifnot(!is.null(m_y))
    if (is.null(wt)) {
      m_mean <- colMeans(m_y, na.rm = TRUE)
    } else {
      m_mean <- out$center
    }
  }

  # rename
  m_s <- m_cov
  m_m <- m_mean

  # unconditional
  if (!conditional_x) {
    # unconditional - stochastic x
    if (!fixed_x) {
      # if (lav_use_lavaanC()) {
      #   m_gamma <- lavaanC::m_kronecker_dup_ginv_pre_post(m_s,
      #                                              multiplicator = 2.0)
      # } else {
        m_gamma <- 2 * lav_mat_dup_ginv_pre_post(m_s %x% m_s)
      # }
      if (meanstructure) {
        m_gamma <- lav_mat_bdiag(m_s, m_gamma)
      }

      # unconditional - fixed x
    } else {
      # handle fixed_x = TRUE

      # cov(m_y|X) = m_a - m_b m_c^{-1} m_b'
      # where m_a = cov(m_y), m_b = cov(m_y,X), m_c = cov(X)
      m_a <- m_s[-x_idx, -x_idx, drop = FALSE]
      m_b <- m_s[-x_idx, x_idx, drop = FALSE]
      m_c <- m_s[x_idx, x_idx, drop = FALSE]
      ybarx <- m_a - m_b %*% solve(m_c, t(m_b))

      # reinsert ybarx in m_y+X (residual) covariance matrix
      ybarx_aug <- matrix(0, nrow = NROW(m_s), ncol = NCOL(m_s))
      ybarx_aug[-x_idx, -x_idx] <- ybarx

      # take difference
      m_r <- m_s - ybarx_aug

      # if (lav_use_lavaanC()) {
      #   gamma_s <- lavaanC::m_kronecker_dup_ginv_pre_post(m_s,
      #                                               multiplicator = 2.0)
      #   gamma_r <- lavaanC::m_kronecker_dup_ginv_pre_post(m_r,
      #                                               multiplicator = 2.0)
      # } else {
        gamma_s <- 2 * lav_mat_dup_ginv_pre_post(m_s %x% m_s)
        gamma_r <- 2 * lav_mat_dup_ginv_pre_post(m_r %x% m_r)
      # }
      m_gamma <- gamma_s - gamma_r

      if (meanstructure) {
        m_gamma <- lav_mat_bdiag(ybarx_aug, m_gamma)
      }
    }
  } else {
    # conditional_x

    # 4 possibilities:
    # - no meanstructure, no slopes
    # -    meanstructure, no slopes
    # - no meanstructure, slopes
    # -    meanstructure, slopes

    # regress m_y on X, and compute covariance of residuals 'm_r'
    m_a <- m_s[-x_idx, -x_idx, drop = FALSE]
    m_b <- m_s[-x_idx, x_idx, drop = FALSE]
    m_c <- m_s[x_idx, x_idx, drop = FALSE]
    cov_ybarx <- m_a - m_b %*% solve(m_c) %*% t(m_b)
    # if (lav_use_lavaanC()) {
    #   m_gamma <- lavaanC::m_kronecker_dup_ginv_pre_post(cov_ybarx,
    #                                                  multiplicator = 2.0)
    # } else {
      m_gamma <- 2 *
               lav_mat_dup_ginv_pre_post(cov_ybarx %x% cov_ybarx)
    # }

    if (meanstructure || slopestructure) {
      m_mx <- m_m[x_idx]
      c3 <- lav_samp_gamma_c3(m_mx, m_c)
      # B3 <- cbind(m_my, m_b + tcrossprod(m_my,m_mx))
    }

    if (meanstructure) {
      if (slopestructure) {
        # a11 <- solve(c3) %x% cov_ybarx
        a11 <- cov_ybarx %x% solve(c3)
      } else {
        # a11 <- solve(c3)[1, 1, drop=FALSE] %x% cov_ybarx
        a11 <- cov_ybarx %x% solve(c3)[1, 1, drop = FALSE]
      }
    } else {
      if (slopestructure) {
        # a11 <- solve(c3)[-1, -1, drop=FALSE] %x% cov_ybarx
        a11 <- cov_ybarx %x% solve(c3)[-1, -1, drop = FALSE]
      } else {
        a11 <- matrix(0, 0, 0)
      }
    }

    m_gamma <- lav_mat_bdiag(a11, m_gamma)
  }

  m_gamma
}

# NOTE:
#  - three types:
#       1) plain         (conditional_x = FALSE, fixed_x = FALSE)
#       2) fixed_x       (conditional_x = FALSE, fixed_x = TRUE)
#       3) conditional_x (conditional_x = TRUE)
#  - if conditional_x = TRUE, we ignore fixed_x (can be TRUE or FALSE)
#
#  - new in 0.6-1: if Mu/Sigma is provided, compute 'model-based' Gamma
#                  (only if conditional_x = FALSE, for now)
#  - new in 0.6-2: if cluster.idx is not NULL, correct for clustering
#  - new in 0.6-13: add unbiased = TRUE (for the 'plain' setting only)

# ADF THEORY
lav_samp_gamma <- function(m_y, # Y+X if cond!
                                  m_mu = NULL,
                                  m_sigma = NULL,
                                  x_idx = integer(0L),
                                  cluster_idx = NULL,
                                  fixed_x = FALSE,
                                  conditional_x = FALSE,
                                  meanstructure = FALSE,
                                  slopestructure = FALSE,
                                  gamma_n_minus_one = FALSE,
                                  unbiased = FALSE,
                                  mplus_wls = FALSE,
                                  wt = NULL,
                                  sampling_weights_type = "design") {
  # coerce to matrix
  m_y <- unname(as.matrix(m_y))
  n <- nrow(m_y)
  p <- ncol(m_y)

  # sampling weights?
  # Gamma is the asymptotic (co)variance of the sample moments, which for
  # the saturated (h1) model equals the sandwich I^{-1} J I^{-1}, where I is
  # the expected and J the first-order h1 information matrix. We reuse the
  # (weight-aware) expected information I, and construct the meat J according
  # to how the sampling weights are interpreted (sampling.weights.type):
  #  - "design"    : J = crossprod(wt * sc) / sum(wt), i.e. sum(wt^2)-weighted,
  #                  the design-based sandwich; the parameter vcov is then
  #                  invariant to the overall scale of the weights, matching
  #                  the continuous ML robust (huber.white) SEs.
  #  - "frequency" : J = crossprod(sqrt(wt) * sc) / sum(wt), i.e. sum(wt)-
  #                  weighted; treats the weights as frequencies, so the result
  #                  equals the Gamma from physically replicating each case wt
  #                  times. (NB: this also feeds the WLS/DWLS/DLS weight matrix,
  #                  so it shifts those point estimates relative to "design".)
  # Under normality J = I and this reduces to the normal-theory Gamma.
  if (!is.null(wt)) {
    if (conditional_x) {
      lav_msg_stop(gettext(
        "Gamma with sampling weights is not available (yet) if
         conditional.x = TRUE."))
    }
    if (length(cluster_idx) > 0L) {
      lav_msg_stop(gettext(
        "Gamma with sampling weights is not available (yet) for
         clustered data."))
    }
    out <- stats::cov.wt(m_y, wt = wt, method = "ML")
    i_mat <- lav_mvn_h1_info_expected(
      y = m_y, wt = wt, sample_cov = out$cov, x_idx = x_idx,
      meanstructure = meanstructure
    )
    # unit (unweighted) casewise scores of the saturated model, evaluated
    # at the weighted sample moments
    sc <- lav_mvn_sc_mu_sigma(
      y = m_y, mu = out$center, sigma_1 = out$cov, x_idx = x_idx
    )
    if (!meanstructure) {
      sc <- sc[, -seq_len(p), drop = FALSE]
    }
    j_mat <- lav_samp_wt_crossprod(sc, wt, sampling_weights_type) / sum(wt)
    # fixed.x zeroes the x-blocks of I, so use a pseudo-inverse there
    if (length(x_idx) > 0L) {
      i_inv <- MASS::ginv(i_mat)
    } else {
      i_inv <- solve(i_mat)
    }
    return(i_inv %*% j_mat %*% i_inv)
  }

  # unbiased?
  if (unbiased) {
    if (conditional_x || fixed_x || !is.null(m_sigma) ||
      !is.null(cluster_idx)) {
      lav_msg_stop(
        gettext("unbiased Gamma only available for the simple
        (not conditional_x or fixed_x or model-based or clustered) setting."))
    } else {
    # data really should be complete
      m_cov <- stats::cov(m_y, use = "pairwise.complete.obs")
      m_cov <- m_cov * (n - 1) / n
      cov_vech <- lav_mat_vech(m_cov)
    }
  }

  # model-based?
  if (!is.null(m_sigma)) {
    stopifnot(!conditional_x)
    model_based <- TRUE
    if (meanstructure) {
      stopifnot(!is.null(m_mu))
      sigma <- c(as.numeric(m_mu), lav_mat_vech(m_sigma))
    } else {
      m_mu <- colMeans(m_y, na.rm = TRUE) # for centering!
      sigma <- lav_mat_vech(m_sigma)
    }
  } else {
    model_based <- FALSE
  }

  # denominator
  if (gamma_n_minus_one) {
    n <- n - 1
  }

  # check arguments
  if (length(x_idx) == 0L) {
    conditional_x <- FALSE
    fixed_x <- FALSE
  }
  if (mplus_wls) {
    stopifnot(!conditional_x, !fixed_x)
  }

  if (!conditional_x && !fixed_x) {
    # center, so we can use crossprod instead of cov
    if (model_based) {
      m_yc <- t(t(m_y) - as.numeric(m_mu))
    } else {
      m_yc <- t(t(m_y) - colMeans(m_y, na.rm = TRUE))
    }

    # create z where the rows_i contain the following elements:
    #  - Y_i (if meanstructure is TRUE)
    #  - vech(Yc_i' %*% Yc_i) where Yc_i are the residuals
    idx1 <- lav_mat_vech_col_idx(p)
    idx2 <- lav_mat_vech_row_idx(p)
    if (meanstructure) {
      z <- cbind(m_y, m_yc[, idx1, drop = FALSE] *
        m_yc[, idx2, drop = FALSE])
    } else {
      z <- (m_yc[, idx1, drop = FALSE] *
        m_yc[, idx2, drop = FALSE])
    }

    if (model_based) {
      if (meanstructure) {
        stopifnot(!is.null(m_mu))
        sigma <- c(as.numeric(m_mu), lav_mat_vech(m_sigma))
      } else {
        sigma <- lav_mat_vech(m_sigma)
      }
      zc <- t(t(z) - sigma)
    } else {
      zc <- t(t(z) - colMeans(z, na.rm = TRUE))
    }

    # meat (aggregated within clusters first, if clustered)
    m_gamma <- lav_samp_gamma_meat(zc, cluster_idx = cluster_idx, n = n)
  } else if (!conditional_x && fixed_x) {
    if (model_based) {
      m_yc <- t(t(m_y) - as.numeric(m_mu))
      y_bar <- colMeans(m_y, na.rm = TRUE)
      res_cov <- (m_sigma[-x_idx, -x_idx, drop = FALSE] -
        m_sigma[-x_idx, x_idx, drop = FALSE] %*%
        solve(m_sigma[x_idx, x_idx, drop = FALSE]) %*%
        m_sigma[x_idx, -x_idx, drop = FALSE])
      res_slopes <- (solve(m_sigma[x_idx, x_idx, drop = FALSE]) %*%
        m_sigma[x_idx, -x_idx, drop = FALSE])
      res_int <- (y_bar[-x_idx] -
        as.numeric(colMeans(m_y[, x_idx, drop = FALSE],
          na.rm = TRUE
        ) %*% res_slopes))
      x_bar <- y_bar[x_idx]
      yhat__bar <- as.numeric(res_int + as.numeric(x_bar) %*% res_slopes)
      yhat_bar <- numeric(p)
      yhat_bar[-x_idx] <- yhat__bar
      yhat_bar[x_idx] <- x_bar
      yhat_cov <- m_sigma
      yhat_cov[-x_idx, -x_idx] <- m_sigma[-x_idx, -x_idx] - res_cov


      yhat <- cbind(1, m_y[, x_idx]) %*% rbind(res_int, res_slopes)
      y_hat <- m_y
      y_hat[, -x_idx] <- yhat
      # y_hat <- cbind(yhat, m_y[,x_idx])
      y_hatc <- t(t(y_hat) - yhat_bar)
      idx1 <- lav_mat_vech_col_idx(p)
      idx2 <- lav_mat_vech_row_idx(p)
      if (meanstructure) {
        z <- (cbind(m_y, m_yc[, idx1, drop = FALSE] *
          m_yc[, idx2, drop = FALSE]) -
          cbind(y_hat, y_hatc[, idx1, drop = FALSE] *
            y_hatc[, idx2, drop = FALSE]))
        sigma1 <- c(m_mu, lav_mat_vech(m_sigma))
        sigma2 <- c(yhat_bar, lav_mat_vech(yhat_cov))
      } else {
        z <- (m_yc[, idx1, drop = FALSE] *
          m_yc[, idx2, drop = FALSE] -
          y_hatc[, idx1, drop = FALSE] *
            y_hatc[, idx2, drop = FALSE])
        sigma1 <- lav_mat_vech(m_sigma)
        sigma2 <- lav_mat_vech(yhat_cov)
      }
      zc <- t(t(z) - (sigma1 - sigma2))
    } else {
      m_qr <- qr(cbind(1, m_y[, x_idx, drop = FALSE]))
      yhat <- qr.fitted(m_qr, m_y[, -x_idx, drop = FALSE])
      # y_hat <- cbind(yhat, m_y[,x_idx])
      y_hat <- m_y
      y_hat[, -x_idx] <- yhat

      m_yc <- t(t(m_y) - colMeans(m_y, na.rm = TRUE))
      y_hatc <- t(t(y_hat) - colMeans(y_hat, na.rm = TRUE))
      idx1 <- lav_mat_vech_col_idx(p)
      idx2 <- lav_mat_vech_row_idx(p)
      if (meanstructure) {
        z <- (cbind(m_y, m_yc[, idx1, drop = FALSE] *
          m_yc[, idx2, drop = FALSE]) -
          cbind(y_hat, y_hatc[, idx1, drop = FALSE] *
            y_hatc[, idx2, drop = FALSE]))
      } else {
        z <- (m_yc[, idx1, drop = FALSE] *
          m_yc[, idx2, drop = FALSE] -
          y_hatc[, idx1, drop = FALSE] *
            y_hatc[, idx2, drop = FALSE])
      }
      zc <- t(t(z) - colMeans(z, na.rm = TRUE))
    }

    # meat (aggregated within clusters first, if clustered)
    m_gamma <- lav_samp_gamma_meat(zc, cluster_idx = cluster_idx, n = n)
  } else {
    # conditional_x

    # 4 possibilities:
    # - no meanstructure, no slopes
    # -    meanstructure, no slopes
    # - no meanstructure, slopes
    # -    meanstructure, slopes

    # regress m_y on m_x, and compute residuals
    m_x <- cbind(1, m_y[, x_idx, drop = FALSE])
    m_qr <- qr(m_x)
    m_res <- qr.resid(m_qr, m_y[, -x_idx, drop = FALSE])
    p <- ncol(m_res)

    idx1 <- lav_mat_vech_col_idx(p)
    idx2 <- lav_mat_vech_row_idx(p)

    if (meanstructure || slopestructure) {
      xtx_inv <- unname(solve(crossprod(m_x)))
      xi <- (m_x %*% xtx_inv) * n ## FIXME, shorter way?
      nc_x <- NCOL(m_x)
      nc_y <- NCOL(m_res)
    }

    if (meanstructure) {
      if (slopestructure) {
        # xi_idx <- rep(seq_len(nc_x), each  = nc_y)
        # res_idx <- rep(seq_len(nc_y), times = nc_x)
        xi_idx <- rep(seq_len(nc_x), times = nc_y)
        res_idx <- rep(seq_len(nc_y), each = nc_x)
        z <- cbind(
          xi[, xi_idx, drop = FALSE] *
            m_res[, res_idx, drop = FALSE],
          m_res[, idx1, drop = FALSE] *
            m_res[, idx2, drop = FALSE]
        )
      } else {
        xi_idx <- rep(1L, each = nc_y)
        z <- cbind(
          xi[, xi_idx, drop = FALSE] *
            m_res,
          m_res[, idx1, drop = FALSE] *
            m_res[, idx2, drop = FALSE]
        )
      }
    } else {
      if (slopestructure) {
        # xi_idx <- rep(seq_len(nc_x), each  = nc_y)
        # xi_idx <- xi_idx[ -seq_len(nc_y) ]
        xi_idx <- rep(seq(2, nc_x), times = nc_y)
        # res_idx <- rep(seq_len(nc_y), times = (nc_x - 1L))
        res_idx <- rep(seq_len(nc_y), each = (nc_x - 1L))
        z <- cbind(
          xi[, xi_idx, drop = FALSE] *
            m_res[, res_idx, drop = FALSE],
          m_res[, idx1, drop = FALSE] *
            m_res[, idx2, drop = FALSE]
        )
      } else {
        z <- m_res[, idx1, drop = FALSE] * m_res[, idx2, drop = FALSE]
      }
    }

    if (model_based) {
      zc <- t(t(z) - sigma)
    } else {
      zc <- t(t(z) - colMeans(z, na.rm = TRUE))
    }

    # meat (aggregated within clusters first, if clustered)
    m_gamma <- lav_samp_gamma_meat(zc, cluster_idx = cluster_idx, n = n)
  }


  # only to mimic Mplus when estimator = "WLS"
  if (mplus_wls && !fixed_x && !conditional_x) {
    # adjust G_22 (the varcov part)
    m_s <- cov(m_y, use = "pairwise")
    w <- lav_mat_vech(m_s)
    w_biased <- (n - 1) / n * w
    diff <- outer(w, w) - outer(w_biased, w_biased)
    if (meanstructure) {
      m_gamma[-seq_len(p), -seq_len(p)] <-
        m_gamma[-seq_len(p), -seq_len(p), drop = FALSE] - diff
    } else {
      m_gamma <- m_gamma - diff
    }

    if (meanstructure) {
      # adjust G_12/G_21 (third-order)
      # strange rescaling?
      n1 <- (n - 1) / n
      m_gamma[seq_len(p), -seq_len(p)] <- m_gamma[seq_len(p), -seq_len(p)] * n1
      m_gamma[-seq_len(p), seq_len(p)] <- m_gamma[-seq_len(p), seq_len(p)] * n1
    }
  }

  # clustered? G/(G-1) finite-sample correction, G = number of clusters
  if (length(cluster_idx) > 0L) {
    nc <- length(unique(cluster_idx))
    m_gamma <- m_gamma * nc / (nc - 1)
  }

  # unbiased?
  if (unbiased) {
    # normal-theory m_gamma (cov only)
    # if (lav_use_lavaanC()) {
    #   gammant_cov <- lavaanC::m_kronecker_dup_ginv_pre_post(m_cov,
    #                                               multiplicator = 2.0)
    # } else {
      gammant_cov <- 2 * lav_mat_dup_ginv_pre_post(m_cov %x% m_cov)
    # }

    if (meanstructure) {
      gamma_cov <- m_gamma[-(1:p), -(1:p), drop = FALSE]
      gamma_mean_cov <- m_gamma[1:p, -(1:p), drop = FALSE]
    } else {
      gamma_cov <- m_gamma
    }

    # Browne's unbiased DF estimator (m_cov part)
    gamma_u <- (n * (n - 1) / (n - 2) / (n - 3) * gamma_cov -
      n / (n - 2) / (n - 3) * (gammant_cov -
        2 / (n - 1) * tcrossprod(cov_vech)))
    if (meanstructure) {
      m_gamma <- lav_mat_bdiag(m_cov, gamma_u)

      # 3-rd order:
      m_gamma[1:p, (p + 1):ncol(m_gamma)] <- gamma_mean_cov * n / (n - 2)
      m_gamma[(p + 1):ncol(m_gamma), 1:p] <- t(gamma_mean_cov * n / (n - 2))
    } else {
      m_gamma <- gamma_u
    }
  } # unbiased

  m_gamma
}

# ADF Gamma for correlations
#
# 30 May 2024: basic version: fixed_x=FALSE, conditional_x=FALSE, ...
# Jacobian of the (partial) correlation moment vector
#   [ mean?, var(num_idx), off-diagonal (co)variances ]
# with respect to the raw moments [ mean?, vech(S) ], evaluated at 'm_s'.
# 'cor_idx' lists the variables that are standardized to unit variance;
# num_idx (= complement) keep their original metric. This is the building
# block of the delta-method correlation Gamma (both NT and ADF), and works
# for full (cor_idx = all) as well as partial correlation structures.
lav_partial_cor_jacobian <- function(m_s, cor_idx = NULL,
                                      meanstructure = FALSE) {
  p <- nrow(m_s)
  is_cor <- logical(p)
  is_cor[cor_idx] <- TRUE
  num_idx <- which(!is_cor)
  # scaling: sd for correlation variables, 1 for the others
  d <- rep(1.0, p)
  d[is_cor] <- sqrt(diag(m_s)[is_cor])

  # vech bookkeeping (full, with diagonal)
  vr <- lav_mat_vech_row_idx(p)
  vc <- lav_mat_vech_col_idx(p)
  pstar <- length(vr)
  diag_pos <- which(vr == vc) # variances, in variable order
  off_pos <- which(vr != vc) # covariances
  var_pos <- diag_pos[num_idx] # variances we keep (non-correlation vars)

  # Jacobian (covariance part): rows = [var(num_idx), offdiag], cols = vech(S)
  nkeep <- length(var_pos) + length(off_pos)
  j_cov <- matrix(0, nkeep, pstar)
  if (length(var_pos) > 0L) {
    j_cov[cbind(seq_along(var_pos), var_pos)] <- 1
  }
  off0 <- length(var_pos)
  for (i in seq_along(off_pos)) {
    pos <- off_pos[i]
    k <- vr[pos]
    l <- vc[pos]
    m_kl <- m_s[k, l] / (d[k] * d[l])
    row <- off0 + i
    j_cov[row, pos] <- 1 / (d[k] * d[l])
    if (is_cor[k]) {
      j_cov[row, diag_pos[k]] <- j_cov[row, diag_pos[k]] -
        0.5 * m_kl / m_s[k, k]
    }
    if (is_cor[l]) {
      j_cov[row, diag_pos[l]] <- j_cov[row, diag_pos[l]] -
        0.5 * m_kl / m_s[l, l]
    }
  }

  if (meanstructure) {
    # partial moments: [mean(p), var(num_idx), offdiag]; the mean of a
    # standardized variable is scaled by 1/sd (first order), others by 1
    j_mat <- matrix(0, p + nkeep, p + pstar)
    j_mat[cbind(seq_len(p), seq_len(p))] <- 1 / d
    j_mat[(p + 1L):(p + nkeep), (p + 1L):(p + pstar)] <- j_cov
  } else {
    j_mat <- j_cov
  }

  j_mat
}

# ADF (sandwich) Gamma for a (partial) correlation structure, via the delta
# method: Gamma_cor = J %*% Gamma_full %*% t(J), where Gamma_full is the raw
# ADF Gamma (lav_samp_gamma, which also handles fixed.x via x_idx) and J is
# the Jacobian above.
lav_samp_partial_cor_gamma <- function(m_y, cor_idx = NULL,
                                       meanstructure = FALSE,
                                       fixed_x = FALSE,
                                       x_idx = integer(0L)) {
  m_y <- unname(as.matrix(m_y))
  n <- nrow(m_y)
  m_s <- cov(m_y) * (n - 1) / n

  j_mat <- lav_partial_cor_jacobian(m_s, cor_idx = cor_idx,
                                    meanstructure = meanstructure)
  gamma_full <- lav_samp_gamma(m_y, meanstructure = meanstructure,
                               x_idx = x_idx, fixed_x = fixed_x)

  j_mat %*% gamma_full %*% t(j_mat)
}

# Normal-theory Gamma for a (partial) correlation structure, via the delta
# method: Gamma_cor = J %*% Gamma_full_NT %*% t(J), where Gamma_full_NT is the
# raw NT Gamma (lav_samp_gamma_nt, which also handles fixed.x).
lav_samp_partial_cor_gamma_nt <- function(m_cov, cor_idx = NULL,
                                          meanstructure = FALSE,
                                          fixed_x = FALSE,
                                          x_idx = integer(0L)) {
  j_mat <- lav_partial_cor_jacobian(m_cov, cor_idx = cor_idx,
                                    meanstructure = meanstructure)
  gamma_full <- lav_samp_gamma_nt(m_cov = m_cov, meanstructure = meanstructure,
                                  x_idx = x_idx, fixed_x = fixed_x)

  j_mat %*% gamma_full %*% t(j_mat)
}

# functions related to AIC and other information criteria

# lower-level functions:
# - lav_fit_aic
# - lav_fit_bic
# - lav_fit_sabic

# higher-level functions:
# - lav_fit_aic_lavobject

# Y.R. 21 July 2022

lav_fit_aic <- function(logl = NULL, npar = NULL) {
  aic <- (-2 * logl) + (2 * npar)
  aic
}

lav_fit_bic <- function(logl = NULL, npar = NULL, n = NULL) {
  bic <- (-2 * logl) + (npar * log(n))
  bic
}

lav_fit_sabic <- function(logl = NULL, npar = NULL, n = NULL) {
  n_star <- (n + 2) / 24
  sabic <- (-2 * logl) + (npar * log(n_star))
  sabic
}

lav_fit_aic_lavobject <- function(lavobject = NULL, fit_measures = "aic",
                                  standard_test = "standard",
                                  scaled_test = "none",
                                  estimator = "ML") {
  # check lavobject
  stopifnot(inherits(lavobject, "lavaan"))
  # check object
  lavobject <- lav_object_check_version(lavobject)

  # tests
  test <- lavobject@test
  test_names <- sapply(lavobject@test, "[[", "test")
  if (test_names[1] == "none" || standard_test == "none") {
    return(list())
  }
  test_idx <- which(test_names == standard_test)[1]
  if (length(test_idx) == 0L) {
    return(list())
  }

  scaled_flag <- FALSE
  if (!scaled_test %in% c("none", "standard", "default")) {
    scaled_idx <- which(test_names == scaled_test)
    if (length(scaled_idx) > 0L) {
      scaled_idx <- scaled_idx[1] # only the first one
      scaled_flag <- TRUE
    }
  }

  # estimator?
  if (missing(estimator)) {
    estimator <- lavobject@Options$estimator
  }

  # supported fit measures in this function
  if (estimator == "MML") {
    fit_logl <- c("logl", "aic", "bic", "ntotal", "bic2")
  } else {
    fit_logl <- c(
      "logl", "unrestricted.logl", "aic", "bic",
      "ntotal", "bic2"
    )
  }
  if (scaled_flag &&
    scaled_test %in% c("yuan.bentler", "yuan.bentler.mplus")) {
    fit_logl <- c(fit_logl, "scaling.factor.h1", "scaling.factor.h0")
  }

  # which one do we need?
  if (missing(fit_measures)) {
    # default set
    fit_measures <- fit_logl
  } else {
    # remove any not-CFI related index from fit.measures
    rm_idx <- which(!fit_measures %in% fit_logl)
    if (length(rm_idx) > 0L) {
      fit_measures <- fit_measures[-rm_idx]
    }
    if (length(fit_measures) == 0L) {
      return(list())
    }
  }

  # output container
  indices <- list()

  # non-ML values
  indices["logl"] <- as.numeric(NA)
  indices["unrestricted.logl"] <- as.numeric(NA)
  indices["aic"] <- as.numeric(NA)
  indices["bic"] <- as.numeric(NA)
  indices["ntotal"] <- lavobject@SampleStats@ntotal
  indices["bic2"] <- as.numeric(NA)

  if (estimator %in% c("ML", "MML")) {
    # do we have a @h1 slot?
    if (length(lavobject@h1) > 0L) {
      indices["unrestricted.logl"] <- lavobject@h1$logl$loglik
    } else {
      lavh1 <- lav_h1_implied_logl(
        lavdata = lavobject@Data,
        lavsamplestats = lavobject@SampleStats,
        lavoptions = lavobject@Options
      )
      indices["unrestricted.logl"] <- lavh1$logl$loglik
    }

    # logl H0
    loglik <- lavobject@loglik
    indices["logl"] <- loglik$loglik
    indices["aic"] <- loglik$AIC
    indices["bic"] <- loglik$BIC
    indices["ntotal"] <- loglik$ntotal
    indices["bic2"] <- loglik$BIC2

    # scaling factor for MLR
    if (scaled_test %in% c("yuan.bentler", "yuan.bentler.mplus")) {
      indices["scaling.factor.h1"] <- test[[scaled_idx]]$scaling.factor.h1
      indices["scaling.factor.h0"] <- test[[scaled_idx]]$scaling.factor.h0
    }
  } # ML

  # return only those that were requested
  indices[fit_measures]
}

# the univariate two-level ordinal probit (random-intercept) model
#
#     y*_ij = xw_ij' pw + xb_j' pb + y_w,ij + b_j,
#     y_w,ij ~ N(0, 1),  b_j ~ N(0, vb)
#     y_ij = k  <=>  tau_k < y*_ij <= tau_{k+1}
#
# i.e. conditional on the cluster effect b_j:
#
#     P(y_ij = k | b_j) = Phi(tau_k+1 - eta_ij - b_j)
#                         - Phi(tau_k - eta_ij - b_j)
#
# with eta_ij = xw_ij' pw + xb_j' pb (cfr. the single-level convention in
# lav_uvord.R: the linear predictor enters with a minus sign relative to
# the thresholds). The within-level latent residual variance is fixed to
# 1 (theta parameterization); there is no mean parameter (absorbed by the
# thresholds). This is stage 1 of the stage-wise unrestricted two-level
# model for a categorical variable.
#
# YR 2026 (two-level WLS, phase 2; covariates added in phase 7)
#
# parameter vector: theta = c(tau_1, ..., tau_nth, vb[, pw, pb])
# the slope parameters are APPENDED, so that the positions of the
# thresholds and vb do not depend on the presence of covariates
#
# The cluster likelihood integrates over b_j with (standardized)
# Gauss-Hermite quadrature; the per-node conditional kernel reuses
# lav_uvord_z12() with offset = eta + b_q:
#
#     L_j = sum_q w_q prod_i P(y_ij | b_q),   b_q = sqrt(vb) x_q
#
# computed per cluster on the log scale (log-sum-exp). The analytic
# gradient is posterior-weighted: with pw_jq = w_q exp(ll_jq) / L_j,
#
#   d logL_j / d tau = sum_q pw_jq * sum_{i in j} (y1 p1 - y2 p2)/pi |_q
#   d logL_j / d vb  = sum_q pw_jq * [sum_{i in j} (p2 - p1)/pi |_q]
#                                  * b_q / (2 vb)
#   d logL_j / d pw  = sum_q pw_jq * sum_{i in j} x_ij (p2 - p1)/pi |_q
#
# (the standardized nodes x_q and weights w_q do not depend on vb).
#
# WITHOUT covariates the conditional kernel only depends on the data
# through the response CATEGORY, so all per-node quantities are computed
# per category and aggregated per cluster via the (nclusters x ncat)
# count matrix -- the cost is (nearly) independent of the number of
# observations. With covariates every observation has its own linear
# predictor, so the kernels are computed per observation.
#
# fit object (list) contract -- see lav_uvord_2l_fit():
#   theta, nth, th_idx, bvar_idx, wslope_idx, bslope_idx,
#   nexo_w, nexo_b, nexo, y, x_w, x_b, z, cluster_idx, ngh,
#   logl, converged

lav_uvord_2l_fit <- function(y = NULL,
                             x_w = NULL, # within-level covariates
                             x_b = NULL, # between-level covariates
                             cluster_idx = NULL,
                             ngh = 21L,
                             optim_method = "nlminb1",
                             control = list(),
                             output = "list") {
  # create cache environment
  cache <- lav_uvord_2l_init_cache(
    y = y, x_w = x_w, x_b = x_b, cluster_idx = cluster_idx, ngh = ngh
  )
  nexo <- cache$nexo

  # optim
  minfns <- lav_uvbv_optim_fns(
    optim_method, lav_uvord_2l_min_fns(covariates = (nexo > 0L))
  )

  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  lower <- c(rep(-Inf, cache$nth), 1e-6, rep(-Inf, nexo))

  # try 1
  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb, lower = lower,
    cache = cache
  )

  # try 2: different starting value for vb
  if (optim$convergence != 0L) {
    start_2 <- cache$theta_start
    start_2[cache$nth + 1L] <- 0.5
    optim <- nlminb(
      start = start_2, objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control_nlminb, lower = lower,
      cache = cache
    )
  }

  if (optim$convergence != 0L) {
    lav_msg_warn(gettext(
      "estimation of the two-level ordinal probit model did not converge"))
  }

  if (output == "cache") {
    return(cache)
  }

  list(
    theta = optim$par,
    nth = cache$nth,
    th_idx = seq_len(cache$nth),
    bvar_idx = cache$nth + 1L,
    wslope_idx = if (cache$nexo_w > 0L) cache$nth + 1L +
                   seq_len(cache$nexo_w)
                 else integer(0L),
    bslope_idx = if (cache$nexo_b > 0L) cache$nth + 1L + cache$nexo_w +
                   seq_len(cache$nexo_b)
                 else integer(0L),
    nexo_w = cache$nexo_w, nexo_b = cache$nexo_b, nexo = nexo,
    y = cache$y,
    x_w = cache$x_w, x_b = cache$x_b, z = cache$z,
    cluster_idx = cache$cluster_idx,
    ngh = ngh,
    logl = {
      cache$theta <- optim$par
      if (nexo > 0L) {
        lav_uvord_2l_x_loglik_cache(cache = cache)
      } else {
        lav_uvord_2l_loglik_cache(cache = cache)
      }
    },
    converged = (optim$convergence == 0L)
  )
}

# prepare cache environment
lav_uvord_2l_init_cache <- function(y = NULL,
                                    x_w = NULL,
                                    x_b = NULL,
                                    cluster_idx = NULL,
                                    ngh = 21L,
                                    parent = parent.frame()) {
  # y
  if (!is.integer(y)) {
    y <- as.integer(y)
  }
  if (anyNA(y)) {
    lav_msg_stop(gettext(
      "missing values are not supported (yet) in lav_uvord_2l_fit()."))
  }
  if (!min(y) == 1L) {
    y <- as.integer(ordered(y))
  }

  nobs <- length(y)
  y_freq <- tabulate(y)
  ncat <- length(y_freq)
  nth <- ncat - 1L
  y_prop <- y_freq / sum(y_freq)

  nexo_w <- if (is.null(x_w)) 0L else NCOL(x_w)
  nexo_b <- if (is.null(x_b)) 0L else NCOL(x_b)
  nexo <- nexo_w + nexo_b

  # cluster x category count matrix
  cl_sorted <- sort(unique(cluster_idx))
  cl_pos <- match(cluster_idx, cl_sorted)
  nclusters <- length(cl_sorted)
  cmat <- unclass(table(
    factor(cl_pos, levels = seq_len(nclusters)),
    factor(y, levels = seq_len(ncat))
  ))
  dimnames(cmat) <- NULL
  storage.mode(cmat) <- "double"

  # category-level offsets and indicator matrices
  ycat <- seq_len(ncat)
  o12 <- lav_uvord_o12(y = ycat, nth = nth)
  y1 <- matrix(seq_len(nth), ncat, nth, byrow = TRUE) == ycat
  y2 <- matrix(seq_len(nth), ncat, nth, byrow = TRUE) == (ycat - 1L)

  # standardized GH nodes
  gh <- lav_2l_gh_xw(ngh = ngh)

  # starting values: marginal thresholds (and slopes) live on the
  # sd = sqrt(1 + vb) scale; take vb0 = 0.2
  vb0 <- 0.2
  s0 <- sqrt(1 + vb0)
  if (nexo == 0L) {
    th_start <- qnorm(cumsum(y_prop[-length(y_prop)])) * s0
    theta <- c(th_start, vb0)
    z <- NULL
    o1_i <- o2_i <- NULL
  } else {
    z <- cbind(x_w, x_b)
    storage.mode(z) <- "double"
    if (anyNA(z)) {
      lav_msg_stop(gettext(
        "missing values are not supported (yet) in lav_uvord_2l_fit()."))
    }
    # single-level probit fit for the starting values
    fit0 <- lav_uvord_fit(y = y, x = z)
    theta <- c(fit0$theta[fit0$th_idx] * s0, vb0,
               fit0$theta[fit0$slope_idx] * s0)
    # observation-level offsets (for the per-observation kernels)
    o12_i <- lav_uvord_o12(y = y, nth = nth)
    o1_i <- o12_i$o1
    o2_i <- o12_i$o2
  }

  list2env(
    list(
      y = y, x_w = x_w, x_b = x_b, z = z,
      cluster_idx = cluster_idx, cmat = cmat,
      ycat = ycat, ncat = ncat,
      o1 = o12$o1, o2 = o12$o2, y1 = y1, y2 = y2,
      o1_i = o1_i, o2_i = o2_i,
      nexo_w = nexo_w, nexo_b = nexo_b, nexo = nexo,
      nobs = nobs, n = nobs, nth = nth, nclusters = nclusters,
      gh = gh, ngh = ngh,
      theta = theta, theta_start = theta
    ),
    parent = parent
  )
}

# per-node conditional cell probabilities (with the same numerical guards
# as the single-level lav_uvord_loglik_cache)
lav_uvord_2l_pi_q <- function(y = NULL, th = NULL, o1 = NULL, o2 = NULL,
                              b_q = NULL) {
  z12 <- lav_uvord_z12(y = y, th = th, o1 = o1, o2 = o2, offset = b_q)
  z1 <- z12$z1
  z2 <- z12$z2
  pi_i <- pnorm(z1) - pnorm(z2)
  large_idx <- which(z2 > 1)
  if (length(large_idx) > 0L) {
    pi_i[large_idx] <- (pnorm(z2[large_idx], lower.tail = FALSE) -
      pnorm(z1[large_idx], lower.tail = FALSE))
  }
  # guard against zero cells
  pi_i[pi_i < .Machine$double.eps] <- .Machine$double.eps
  list(pi_i = pi_i, z1 = z1, z2 = z2)
}

lav_uvord_2l_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    th <- theta[1:nth]
    vb <- theta[nth + 1L]
    s_b <- sqrt(vb)

    # category x node conditional log-probabilities
    logp <- matrix(0, ncat, ngh)
    for (q in seq_len(ngh)) {
      b_q <- s_b * gh$x[q]
      pq <- lav_uvord_2l_pi_q(y = ycat, th = th, o1 = o1, o2 = o2,
                              b_q = b_q)
      logp[, q] <- log(pq$pi_i)
    }

    # cluster x node log-likelihood contributions
    ll <- cmat %*% logp

    # log-sum-exp per cluster
    lwll <- sweep(ll, 2L, gh$logw, "+")
    mmax <- apply(lwll, 1L, max)
    logl_j <- mmax + log(rowSums(exp(lwll - mmax)))

    # posterior node weights (for the gradient)
    pw <- exp(lwll - logl_j)

    loglik <- sum(logl_j)
    return(loglik)
  })                     # nolint end
}

lav_uvord_2l_grad_cache <- function(cache = NULL) {
  colSums(lav_uvord_2l_sc_cache(cache = cache))
}

# cluster-wise scores (of the loglik); assumes the loglik has been
# evaluated at cache$theta (pw available)
lav_uvord_2l_sc_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    th <- theta[1:nth]
    vb <- theta[nth + 1L]
    s_b <- sqrt(vb)

    sc <- matrix(0, nclusters, nth + 1L)
    for (q in seq_len(ngh)) {
      b_q <- s_b * gh$x[q]
      pq <- lav_uvord_2l_pi_q(y = ycat, th = th, o1 = o1, o2 = o2,
                              b_q = b_q)
      p1 <- dnorm(pq$z1)
      p2 <- dnorm(pq$z2)
      inv_pi <- 1 / pq$pi_i

      # category-level score pieces -> cluster level via the count matrix
      # thresholds: (y1 p1 - y2 p2)/pi
      sc_th_q <- cmat %*% ((y1 * p1 - y2 * p2) * inv_pi)
      # cluster intercept: (p2 - p1)/pi
      g_b_q <- cmat %*% ((p2 - p1) * inv_pi)

      sc[, 1:nth] <- sc[, 1:nth] + pw[, q] * sc_th_q
      sc[, nth + 1L] <- sc[, nth + 1L] +
        pw[, q] * g_b_q * (b_q / (2 * vb))
    }

    return(sc)
  })                     # nolint end
}

# --- with covariates (per-observation kernels) ---

lav_uvord_2l_x_loglik_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    th <- theta[1:nth]
    vb <- theta[nth + 1L]
    s_b <- sqrt(vb)
    beta <- theta[nth + 1L + seq_len(nexo)]
    eta <- drop(z %*% beta)

    # cluster x node log-likelihood contributions
    ll <- matrix(0, nclusters, ngh)
    for (q in seq_len(ngh)) {
      pq <- lav_uvord_2l_pi_q(y = y, th = th, o1 = o1_i, o2 = o2_i,
                              b_q = eta + s_b * gh$x[q])
      ll[, q] <- drop(rowsum.default(log(pq$pi_i), group = cluster_idx,
                                     reorder = TRUE))
    }

    # log-sum-exp per cluster
    lwll <- sweep(ll, 2L, gh$logw, "+")
    mmax <- apply(lwll, 1L, max)
    logl_j <- mmax + log(rowSums(exp(lwll - mmax)))

    # posterior node weights (for the gradient)
    pw <- exp(lwll - logl_j)

    loglik <- sum(logl_j)
    return(loglik)
  })                     # nolint end
}

lav_uvord_2l_x_grad_cache <- function(cache = NULL) {
  colSums(lav_uvord_2l_x_sc_cache(cache = cache))
}

# cluster-wise scores, covariate case; assumes the loglik has been
# evaluated at cache$theta (pw available);
# columns: (tau_1..nth, vb, slopes)
lav_uvord_2l_x_sc_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    th <- theta[1:nth]
    vb <- theta[nth + 1L]
    s_b <- sqrt(vb)
    beta <- theta[nth + 1L + seq_len(nexo)]
    eta <- drop(z %*% beta)

    # z1 involves tau_y (y <= nth), z2 involves tau_{y-1} (y >= 2)
    idx1 <- which(y <= nth)
    idx2 <- which(y >= 2L)

    sc <- matrix(0, nclusters, nth + 1L + nexo)
    for (q in seq_len(ngh)) {
      pq <- lav_uvord_2l_pi_q(y = y, th = th, o1 = o1_i, o2 = o2_i,
                              b_q = eta + s_b * gh$x[q])
      p1 <- dnorm(pq$z1)
      p2 <- dnorm(pq$z2)
      inv_pi <- 1 / pq$pi_i

      # observation-level score pieces
      # thresholds: d logp_i / d tau_k = (p1 1{y=k} - p2 1{y=k+1}) / pi
      kern_th <- matrix(0, nobs, nth)
      kern_th[cbind(idx1, y[idx1])] <- (p1 * inv_pi)[idx1]
      kern_th[cbind(idx2, y[idx2] - 1L)] <-
        kern_th[cbind(idx2, y[idx2] - 1L)] - (p2 * inv_pi)[idx2]
      # offset (cluster intercept/slopes): (p2 - p1)/pi
      g_i <- (p2 - p1) * inv_pi

      cl <- rowsum.default(cbind(kern_th, g_i, g_i * z),
                           group = cluster_idx, reorder = TRUE)
      sc[, 1:nth] <- sc[, 1:nth] + pw[, q] * cl[, 1:nth, drop = FALSE]
      sc[, nth + 1L] <- sc[, nth + 1L] +
        pw[, q] * cl[, nth + 1L] * (s_b * gh$x[q] / (2 * vb))
      sc[, nth + 1L + seq_len(nexo)] <- sc[, nth + 1L + seq_len(nexo)] +
        pw[, q] * cl[, nth + 1L + seq_len(nexo), drop = FALSE]
    }

    return(sc)
  })                     # nolint end
}

# cluster-wise scores at given (or fitted) parameter values - no cache
lav_uvord_2l_sc <- function(fit_y = NULL, theta = NULL) {
  cache <- lav_uvord_2l_init_cache(
    y = fit_y$y, x_w = fit_y$x_w, x_b = fit_y$x_b,
    cluster_idx = fit_y$cluster_idx, ngh = fit_y$ngh
  )
  if (is.null(theta)) {
    theta <- fit_y$theta
  }
  cache$theta <- theta
  if (cache$nexo > 0L) {
    tmp <- lav_uvord_2l_x_loglik_cache(cache = cache) # populates pw
    sc <- lav_uvord_2l_x_sc_cache(cache = cache)
    colnames(sc) <- c(
      paste0("th", seq_len(cache$nth)), "bvar",
      lav_uvreg_2l_slope_names(cache$nexo_w, cache$nexo_b)
    )
  } else {
    tmp <- lav_uvord_2l_loglik_cache(cache = cache) # populates pw
    sc <- lav_uvord_2l_sc_cache(cache = cache)
    colnames(sc) <- c(paste0("th", seq_len(cache$nth)), "bvar")
  }
  sc
}

# total loglik - no cache
lav_uvord_2l_logl <- function(fit_y = NULL, theta = NULL) {
  cache <- lav_uvord_2l_init_cache(
    y = fit_y$y, x_w = fit_y$x_w, x_b = fit_y$x_b,
    cluster_idx = fit_y$cluster_idx, ngh = fit_y$ngh
  )
  if (is.null(theta)) {
    theta <- fit_y$theta
  }
  cache$theta <- theta
  if (cache$nexo > 0L) {
    lav_uvord_2l_x_loglik_cache(cache = cache)
  } else {
    lav_uvord_2l_loglik_cache(cache = cache)
  }
}

# reject descending thresholds (cfr. lav_uvord_pre_objective)
lav_uvord_2l_pre_objective <- function(x, cache = NULL) {
  if (cache$nth > 1L && x[1] > x[2]) {
    return(+Inf)
  }
  if (cache$nth > 2L && x[2] > x[3]) {
    return(+Inf)
  }
  if (cache$nth > 3L && x[3] > x[4]) {
    return(+Inf)
  }
  NULL
}

# nlminb objective/gradient (see lav_uvbv_common.R)
lav_uvord_2l_min_fns <- function(covariates = FALSE) {
  if (covariates) {
    lav_uvbv_min_fns(
      logl_fun = lav_uvord_2l_x_loglik_cache,
      grad_fun = lav_uvord_2l_x_grad_cache,
      pre_objective = lav_uvord_2l_pre_objective
    )
  } else {
    lav_uvbv_min_fns(
      logl_fun = lav_uvord_2l_loglik_cache,
      grad_fun = lav_uvord_2l_grad_cache,
      pre_objective = lav_uvord_2l_pre_objective
    )
  }
}

# acceleration of EM-type fixed-point iterations
#
# YR + Claude, July 2026
#
# SQUAREM-1 (Varadhan & Roland, 2008, Scand. J. Statist. 35, 335-353):
# each cycle takes two EM steps, extrapolates along the 'squared' step,
# and then takes one stabilizing EM step from the extrapolated point;
# a loglikelihood safeguard falls back to the plain (double) EM step
# whenever the extrapolation fails or decreases the loglikelihood too much

# theta     = packed parameter vector
# step_fn   = the EM map: theta -> theta' (one E-step + M-step); the map
#             must return a valid parameter vector for any theta it accepts
# logl_fn   = loglikelihood at theta (to be maximized); used to safeguard
#             the extrapolation and to monitor convergence
# fx0       = loglikelihood at the initial theta (recomputed if NULL)
# conv      = convergence criterion (per cycle): "logl" = |change in logl|
#             < tol (as in the plain multilevel EM iterations); "param" =
#             largest absolute change in the parameter values < tol (as in
#             the plain single-level EM iterations)
# conv_fn   = optional user-supplied convergence check, overriding conv;
#             a function(theta_old, theta, fx_old, fx) returning TRUE
#             (converged) or FALSE (continue); used by the h0 EM optimizer
#             to combine the logl criterion with a gradient check
# tol       = convergence tolerance
# max_iter  = maximum number of EM map evaluations (not cycles), so that
#             the em.h1.args$max_iter option keeps the same meaning as for
#             the non-accelerated EM iterations
# step_max0 = initial upper bound for the extrapolation step length
# mstep_factor = growth/shrink factor for the step length bound
# logl_dec  = maximum allowed *decrease* in the loglikelihood for an
#             extrapolated point to be accepted (mild non-monotonicity
#             speeds up convergence; see Varadhan & Roland); if even the
#             plain (fallback) EM step decreases the loglikelihood by more
#             than logl_dec, the iterations stop at the previous point
#             (returned with stalled = TRUE)
lav_em_squarem <- function(theta = NULL,
                           step_fn = NULL,
                           logl_fn = NULL,
                           fx0 = NULL,
                           conv = "logl",
                           conv_fn = NULL,
                           tol = 1e-04,
                           max_iter = 5000L,
                           step_max0 = 1,
                           mstep_factor = 4,
                           logl_dec = 1) {
  if (is.null(fx0)) {
    fx0 <- logl_fn(theta)
  }
  fx <- fx_old <- fx0
  fpeval <- 0L
  cycle <- 0L
  converged <- FALSE
  stalled <- FALSE
  step_max <- step_max0

  while (fpeval < max_iter) {
    cycle <- cycle + 1L

    # two plain EM steps
    p0 <- theta
    p1 <- step_fn(p0)
    fpeval <- fpeval + 1L
    q1 <- p1 - p0
    p2 <- step_fn(p1)
    fpeval <- fpeval + 1L
    q2 <- p2 - p1
    v <- q2 - q1
    sv2 <- sum(v * v)

    # extrapolate: theta' = p0 + 2*alpha*q1 + alpha^2*(q2 - q1)
    # with steplength alpha = sqrt(|q1|^2 / |q2 - q1|^2) ("SqS3"),
    # bounded by step_max; note that alpha = 1 gives theta' = p2, the
    # plain (double) EM step
    alpha <- 1
    if (sv2 > .Machine$double.eps) {
      alpha <- max(1, min(step_max, sqrt(sum(q1 * q1) / sv2)))
    }
    accepted <- FALSE
    if (alpha > 1.01) {
      p_new <- p0 + 2 * alpha * q1 + alpha * alpha * v
      # stabilize: one extra EM step from the extrapolated point; this
      # also repairs (mildly) infeasible extrapolations, as the output
      # of an M-step is always a valid parameter vector
      p_try <- try(step_fn(p_new), silent = TRUE)
      fpeval <- fpeval + 1L
      if (!inherits(p_try, "try-error") &&
          all(is.finite(p_try))) {
        fx_try <- try(logl_fn(p_try), silent = TRUE)
        if (!inherits(fx_try, "try-error") && is.finite(fx_try) &&
            fx_try >= fx_old - logl_dec) {
          theta <- p_try
          fx <- fx_try
          accepted <- TRUE
        }
      }
      if (!accepted) {
        # rejected: shrink the steplength bound (if we hit it)
        if (alpha >= step_max) {
          step_max <- max(step_max0, step_max / mstep_factor)
        }
        alpha <- 1
      }
    }
    if (!accepted) {
      # plain (double) EM step
      theta <- p2
      fx <- logl_fn(theta)
      if (!is.finite(fx)) {
        lav_msg_stop(gettext(
          "The EM iterations failed; some matrices may be singular;
           please check your data for (near-)perfect correlations."))
      }
      if (fx < fx_old - logl_dec) {
        # even the plain (double) EM step decreased the loglikelihood
        # substantially: the map is not behaving like a monotone EM map
        # at this point (this can happen when the M-step is only solved
        # approximately, as in the h0 EM optimizer); revert and stop
        theta <- p0
        fx <- fx_old
        stalled <- TRUE
        break
      }
      accepted <- TRUE
    }
    # if we hit the bound, allow longer steps in the next cycle
    if (accepted && alpha >= step_max) {
      step_max <- mstep_factor * step_max
    }

    fx_delta <- fx - fx_old
    if (lav_verbose()) {
      cat(
        "SQUAREM:", sprintf("%3d", cycle),
        " fx =", sprintf("%17.10f", fx),
        " fx.delta =", sprintf("%9.8f", fx_delta),
        " alpha =", sprintf("%6.2f", alpha),
        "\n"
      )
    }

    # convergence check (per cycle)
    if (!is.null(conv_fn)) {
      converged <- isTRUE(conv_fn(p0, theta, fx_old, fx))
    } else if (conv == "param") {
      converged <- max(abs(theta - p0)) < tol
    } else {
      converged <- abs(fx_delta) < tol
    }
    if (converged) {
      break
    }
    fx_old <- fx
  }

  list(theta = theta, fx = fx, converged = converged, stalled = stalled,
       fpeval = fpeval, cycles = cycle)
}

# QN: quasi-Newton acceleration of EM-type fixed-point iterations
# (Zhou, Alexander & Lange, 2011, Stat. Comput. 21, 261-273; the 'qn'
# method of the turboEM package). A rank-q secant approximation of the
# Jacobian of the EM map F yields a quasi-Newton step towards the fixed
# point x = F(x), without any analytic gradient or Hessian. Like SQUAREM,
# a loglikelihood safeguard falls back to a plain (double) EM step whenever
# the quasi-Newton step is infeasible or does not improve on it.
#
# the arguments have the same meaning as in lav_em_squarem() above;
# qn_order = the number of secant pairs (the 'order' q of the scheme; the
# turboEM default is 2). Note: each iteration takes two EM map evaluations
# (versus up to three for a SQUAREM cycle), plus up to two loglikelihood
# evaluations for the safeguard.
lav_em_qn <- function(theta = NULL,
                      step_fn = NULL,
                      logl_fn = NULL,
                      fx0 = NULL,
                      conv = "logl",
                      conv_fn = NULL,
                      tol = 1e-04,
                      max_iter = 5000L,
                      qn_order = 2L,
                      logl_dec = 1) {
  npar <- length(theta)
  qn_order <- max(1L, min(as.integer(qn_order), npar))

  # safeguarded loglikelihood: -Inf if the point is infeasible (so an
  # infeasible extrapolation is always rejected in favor of the plain step)
  safe_logl <- function(x) {
    if (!all(is.finite(x))) {
      return(-Inf)
    }
    fx <- try(logl_fn(x), silent = TRUE)
    if (inherits(fx, "try-error") || !is.finite(fx)) {
      return(-Inf)
    }
    fx
  }

  fpeval <- 0L
  iter <- 0L
  converged <- FALSE
  stalled <- FALSE

  # helper: a single guarded EM map evaluation
  emstep <- function(x) {
    out <- try(step_fn(x), silent = TRUE)
    fpeval <<- fpeval + 1L
    if (inherits(out, "try-error") || !all(is.finite(out))) {
      return(NULL)
    }
    out
  }

  # ---- initialize the secant window (U, V) ----
  # U_i = F(x_i) - x_i ; V_i = F(F(x_i)) - F(x_i) (as in turboEM)
  U <- NULL
  ok <- TRUE
  for (i in seq_len(qn_order)) {
    pnew <- emstep(theta)
    if (is.null(pnew)) {
      ok <- FALSE
      break
    }
    U <- cbind(U, pnew - theta)
    theta <- pnew
  }
  if (ok) {
    pnew <- emstep(theta)
    if (is.null(pnew)) {
      ok <- FALSE
    }
  }
  if (ok) {
    U <- cbind(U[, -1, drop = FALSE], pnew - theta)
    pnew2 <- emstep(pnew)
    if (is.null(pnew2)) {
      ok <- FALSE
    }
  }
  if (!ok) {
    # could not build the secant window; hand back the best plain-EM
    # point we reached, and let the caller treat it as a (non-)stall
    fx <- safe_logl(theta)
    return(list(theta = theta, fx = fx, converged = FALSE,
                stalled = TRUE, fpeval = fpeval, cycles = 0L))
  }
  V <- cbind(U[, -1, drop = FALSE], pnew2 - pnew)
  # 'theta' is the current base point; fx_old is its loglikelihood
  fx_old <- safe_logl(theta)
  if (!is.finite(fx_old)) {
    fx_old <- if (!is.null(fx0)) fx0 else -Inf
  }
  fx <- fx_old

  # ---- main loop ----
  while (fpeval < max_iter) {
    iter <- iter + 1L
    theta_old <- theta

    # quasi-Newton step: solve the (small) q x q system; on a singular
    # system, fall back to the plain (double) EM step pnew2
    tmp <- try(solve(crossprod(U, U - V)), silent = TRUE)
    took_qn <- !inherits(tmp, "try-error")
    if (took_qn) {
      p_next <- as.numeric(pnew -
        (V %*% tmp) %*% crossprod(U, theta - pnew))
      # safeguard: keep the QN step only if feasible and at least as good
      # as the plain (double) EM step
      fx_next <- safe_logl(p_next)
      fx_pnew2 <- safe_logl(pnew2)
      if (is.finite(fx_next) && fx_next >= fx_pnew2) {
        theta <- p_next
        fx <- fx_next
      } else {
        theta <- pnew2
        fx <- fx_pnew2
      }
    } else {
      theta <- pnew2
      fx <- safe_logl(pnew2)
    }

    # stall: even the plain (double) EM step no longer improves the
    # loglikelihood (can happen with an inexact M-step, as in the h0 EM);
    # revert to the previous point and stop
    if (!is.finite(fx) || fx < fx_old - logl_dec) {
      theta <- theta_old
      fx <- fx_old
      stalled <- TRUE
      break
    }

    fx_delta <- fx - fx_old
    if (lav_verbose()) {
      cat(
        "QN:", sprintf("%3d", iter),
        " fx =", sprintf("%17.10f", fx),
        " fx.delta =", sprintf("%9.8f", fx_delta),
        "\n"
      )
    }

    # convergence check (per iteration)
    if (!is.null(conv_fn)) {
      converged <- isTRUE(conv_fn(theta_old, theta, fx_old, fx))
    } else if (conv == "param") {
      converged <- max(abs(theta - theta_old)) < tol
    } else {
      converged <- abs(fx_delta) < tol
    }
    if (converged) {
      break
    }

    # advance the secant window from the accepted point
    pnew <- emstep(theta)
    if (is.null(pnew)) {
      stalled <- TRUE
      break
    }
    U <- cbind(U[, -1, drop = FALSE], pnew - theta)
    pnew2 <- emstep(pnew)
    if (is.null(pnew2)) {
      stalled <- TRUE
      break
    }
    V <- cbind(V[, -1, drop = FALSE], pnew2 - pnew)
    fx_old <- fx
  }

  list(theta = theta, fx = fx, converged = converged, stalled = stalled,
       fpeval = fpeval, cycles = iter)
}

# fitting function for standard ML
lav_model_objective_ml <- function(sigma_hat = NULL, mu_hat = NULL,
                         data_cov = NULL, data_mean = NULL,
                         data_cov_log_det = NULL,
                         meanstructure = FALSE) {
  # FIXME: WHAT IS THE BEST THING TO DO HERE??
  # CURRENTLY: return Inf  (at least for nlminb, this works well)
  if (!attr(sigma_hat, "po")) {
    return(Inf)
  }


  sigma_hat_inv <- attr(sigma_hat, "inv")
  sigma_hat_log_det <- attr(sigma_hat, "log.det")
  nvar <- ncol(sigma_hat)

  if (!meanstructure) {
    fx <- (sigma_hat_log_det + sum(data_cov * sigma_hat_inv) -
      data_cov_log_det - nvar)
  } else {
    w_tilde <- data_cov + tcrossprod(data_mean - mu_hat)
    fx <- (sigma_hat_log_det + sum(w_tilde * sigma_hat_inv) -
      data_cov_log_det - nvar)
  }

  # no negative values
  if (is.finite(fx) && fx < 0.0) fx <- 0.0

  fx
}

# fitting function for standard ML
lav_model_objective_ml_res <- function(sigma_hat = NULL, mu_hat = NULL,
                                       pi0 = NULL, res_cov = NULL,
                                       res_int = NULL, res_slopes = NULL,
                                       res_cov_log_det = NULL,
                                       cov_x = NULL, mean_x = NULL) {
  if (!attr(sigma_hat, "po")) {
    return(Inf)
  }

  # augmented mean.x + cov.x matrix
  c3 <- rbind(
    c(1, mean_x),
    cbind(mean_x, cov_x + tcrossprod(mean_x))
  )

  sigma_hat_inv <- attr(sigma_hat, "inv")
  sigma_hat_log_det <- attr(sigma_hat, "log.det")
  nvar <- ncol(sigma_hat)

  # sigma
  objective_sigma <- (sigma_hat_log_det + sum(res_cov * sigma_hat_inv) -
    res_cov_log_det - nvar)
  # beta
  obs <- t(cbind(res_int, res_slopes))
  est <- t(cbind(mu_hat, pi0))
  diff_1 <- obs - est
  objective_beta <- sum(sigma_hat_inv * crossprod(diff_1, c3) %*% diff_1)

  fx <- objective_sigma + objective_beta

  # no negative values
  if (is.finite(fx) && fx < 0.0) fx <- 0.0

  fx
}


# fitting function for restricted ML
lav_model_objective_reml <- function(sigma_hat = NULL, mu_hat = NULL,
                           data_cov = NULL, data_mean = NULL,
                           data_cov_log_det = NULL,
                           meanstructure = FALSE,
                           group = 1L, lavmodel = NULL,
                           lavsamplestats = NULL, lavdata = NULL) {
  if (!attr(sigma_hat, "po")) {
    return(Inf)
  }

  sigma_hat_inv <- attr(sigma_hat, "inv")
  sigma_hat_log_det <- attr(sigma_hat, "log.det")
  nvar <- ncol(sigma_hat)

  if (!meanstructure) {
    fx <- (sigma_hat_log_det + sum(data_cov * sigma_hat_inv) -
      data_cov_log_det - nvar)
  } else {
    w_tilde <- data_cov + tcrossprod(data_mean - mu_hat)
    fx <- (sigma_hat_log_det + sum(w_tilde * sigma_hat_inv) -
      data_cov_log_det - nvar)
  }

  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")
  mm_lambda <- lavmodel@GLIST[[lambda_idx[group]]]
  data_cov_inv <- lavsamplestats@icov[[group]]
  reml_h0 <- log(det(t(mm_lambda) %*% sigma_hat_inv %*% mm_lambda))
  reml_h1 <- log(det(t(mm_lambda) %*% data_cov_inv %*% mm_lambda))
  nobs <- lavsamplestats@nobs[[group]]

  # fx <- (Sigma.hat.log.det + tmp - data.cov.log.det - nvar) +
  #                        1/Ng * (reml.h0  - reml.h1)
  fx <- fx + (1 / nobs * (reml_h0 - reml_h1))

  # no negative values
  if (is.finite(fx) && fx < 0.0) fx <- 0.0

  fx
}

# 'classic' fitting function for GLS
# used again since 0.6-10 (we used the much slower lav_model_objective_wls before)
lav_model_objective_gls <- function(sigma_hat = NULL,
                                    mu_hat = NULL,
                                    data_cov = NULL,
                                    data_cov_inv = NULL,
                                    data_mean = NULL,
                                    meanstructure = FALSE,
                                    correlation = FALSE) {
  tmp <- data_cov_inv %*% (data_cov - sigma_hat)
  # tmp is not perfectly symmetric, so we use t(tmp) on the next line
  # to obtain the same value as lav_model_objective_wls
  fx <- 0.5 * sum(tmp * t(tmp))

  if (correlation) {
    # Bentler & Savalei (2010) eq 1.31
    dd <- as.matrix(diag(tmp))
    tt <- diag(nrow(data_cov)) + data_cov * data_cov_inv
    fx <- fx - drop(t(dd) %*% solve(tt) %*% dd)
  }

  if (meanstructure) {
    tmp2 <- sum(data_cov_inv * tcrossprod(data_mean - mu_hat))
    fx <- fx + tmp2
  }

  # no negative values
  if (is.finite(fx) && fx < 0.0) fx <- 0.0

  fx
}

# general WLS estimator (Muthen, Appendix 4, eq 99 single group)
# full weight (WLS.V) matrix
lav_model_objective_wls <- function(wls_est = NULL,
                                 wls_obs = NULL, wls_v = NULL) {
  # diff <- as.matrix(WLS.obs - WLS.est)
  # fx <- as.numeric( t(diff) %*% WLS.V %*% diff )

  # since 0.5-17, we use crossprod twice
  diff <- wls_obs - wls_est
  fx <- as.numeric(crossprod(crossprod(wls_v, diff), diff))
  # todo alternative: using chol(WLS.V)

  # no negative values
  if (is.finite(fx) && fx < 0.0) fx <- 0.0

  fx
}

# diagonally weighted LS (DWLS)
lav_model_objective_dwls <- function(wls_est = NULL,
                            wls_obs = NULL, wls_vd = NULL) {
  diff <- wls_obs - wls_est
  fx <- sum(diff * diff * wls_vd)

  # no negative values
  if (is.finite(fx) && fx < 0.0) fx <- 0.0

  fx
}

# Full Information ML estimator (FIML) handling the missing values
lav_model_objective_fiml <- function(sigma_hat = NULL, mu_hat = NULL, yp = NULL,
                           h1 = NULL, n = NULL) {
  if (is.null(n)) {
    n <- sum(sapply(yp, "[[", "freq"))
  }

  # Note: we ignore x.idx (if any)
  fx <- lav_mvn_mi_loglik_samp(
    yp = yp,
    mu = mu_hat, sigma_1 = sigma_hat,
    log2pi = FALSE,
    minus_two = TRUE
  ) / n

  # ajust for h1
  if (!is.null(h1)) {
    fx <- fx - h1

    # no negative values
    if (is.finite(fx) && fx < 0.0) fx <- 0.0
  }

  fx
}

# FIML for the conditional.x case: the (conditional) residual-moment
# counterpart of lav_model_objective_fiml (mu_hat = res.int,
# pi0 = res.slopes, sigma_hat = res.cov)
lav_model_objective_fiml_res <- function(sigma_hat = NULL, mu_hat = NULL,
                                         pi0 = NULL, yp = NULL,
                                         h1 = NULL, n = NULL) {
  if (is.null(n)) {
    n <- sum(sapply(yp, "[[", "freq"))
  }

  fx <- lav_mvreg_mi_loglik_samp(
    yp = yp,
    res_int = mu_hat, res_slopes = pi0, res_cov = sigma_hat,
    log2pi = FALSE,
    minus_two = TRUE
  ) / n

  # ajust for h1
  if (!is.null(h1)) {
    fx <- fx - h1

    # no negative values
    if (is.finite(fx) && fx < 0.0) fx <- 0.0
  }

  fx
}

# pairwise maximum likelihood
# this is adapted from code written by Myrsini Katsikatsou
#
# some changes:
# - no distinction between x/y (ksi/eta)
# - 29/03/2016: adapt for exogenous covariates
# - 21/09/2016: added code for missing = doubly.robust (contributed by
#   Myrsini Katsikatsou)
# - HJ 18/10/2023: For sampling weights the lavcache$bifreq are weighted
lav_model_objective_pml <- function(sigma_hat = NULL, # model-based var/cov/cor
                          mu_hat = NULL, # model-based means
                          th = NULL, # model-based thresholds + means
                          pi0 = NULL, # slopes
                          th_idx = NULL, # threshold idx per variable
                          num_idx = NULL, # which variables are numeric
                          x = NULL, # raw data
                          exo = NULL, # eXo data
                          wt = NULL, # case weights
                          lavcache = NULL, # housekeeping stuff
                          missing = NULL) { # how to deal with missings?

  # YR 3 okt 2012
  # - the idea is to compute for each pair of variables, the model-based
  #   probability (or likelihood in mixed case) (that we observe the data
  #   for this pair under the model)
  # - if we have exogenous variables + conditional.x, do this for each case
  # - after taking logs, the sum over the cases gives the
  #   log probablity/likelihood for this pair
  # - the sum over all pairs gives the final PL based logl

  # first of all: check if all correlations are within [-1,1]
  # if not, return Inf; (at least with nlminb, this works well)

  # diagonal of Sigma.hat is not necessarily 1, even for categorical vars
  sigma_hat2 <- sigma_hat
  if (length(num_idx) > 0L) {
    diag(sigma_hat2)[-num_idx] <- 1
  } else {
    diag(sigma_hat2) <- 1
  }
  # all positive variances? (for continuous variables)
  if (any(diag(sigma_hat2) < 0)) {
    out <- +Inf
    attr(out, "logl") <- as.numeric(NA)
    return(out)
  }
  cor_hat <- cov2cor(sigma_hat2) # to get correlations (rho!)
  cors <- lav_mat_vech(cor_hat, diagonal = FALSE)

  if (length(cors) > 0L && (any(abs(cors) > 1) ||
    any(is.na(cors)))) {
    # question: what is the best approach here??
    out <- +Inf
    attr(out, "logl") <- as.numeric(NA)
    return(out)
  }

  nvar <- nrow(sigma_hat)
  if (is.null(exo)) {
    nexo <- 0L
  } else {
    nexo <- NCOL(exo)
  }
  pstar <- nvar * (nvar - 1) / 2
  ov_types <- rep("ordered", nvar)
  if (length(num_idx) > 0L) {
    ov_types[num_idx] <- "numeric"
  }


  ##### Three cases:
  ##### 1) all ordered, no exogenous (fast!)
  ##### 2) mixed ordered + continuous, no exogenous
  ##### 3) mixed ordered + continuous, exogenous (conditional.x = TRUE)




  ##### Case 1:
  #####  all ordered
  #####  no exogenous covariates
  #####
  if (all(ov_types == "ordered") && nexo == 0L) {
    # prepare for Myrsini's vectorization scheme
    long2 <- lav_pml_longvec_th_rho(
      no_x = nvar,
      all_thres = th,
      index_var_of_thres = th_idx,
      rho_xixj = cors
    )
    # get expected probability per table, per pair
    pairwise_pi <- lav_pml_expprob_vec(
      ind_vec = lavcache$long,
      th_rho_vec = long2
    )
    pairwise_pi_orig <- pairwise_pi # for doubly.robust

    # get frequency per table, per pair
    logl <- sum(lavcache$bifreq * log(pairwise_pi))

    # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    # FYI the bifreq are already weighted so this will work. Alternatively:
    if (!is.null(wt)) {
      logl <- sum(lavcache$sum_obs_weights_xixj_ab_vec * log(pairwise_pi))
    }

    # more convenient fit function
    prop <- lavcache$bifreq / lavcache$nobs
    freq <- lavcache$bifreq
    if (!is.null(wt)) {
      prop <- lavcache$sum_obs_weights_xixj_ab_vec / sum(wt)
      freq <- lavcache$sum_obs_weights_xixj_ab_vec
    }

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    # remove zero props # FIXME!!! or add 0.5???
    # zero.idx <- which(prop == 0.0)
    zero_idx <- which((prop == 0.0) | !is.finite(prop))
    if (length(zero_idx) > 0L) {
      freq <- freq[-zero_idx]
      prop <- prop[-zero_idx]
      pairwise_pi <- pairwise_pi[-zero_idx]
    }
    ## Fmin <- sum( prop*log(prop/pairwisePI) )
    fmin <- sum(freq * log(prop / pairwise_pi)) # to avoid 'N'

    if (missing == "available.cases" || missing == "doubly.robust") {
      uni_pi <- lav_pml_th_uni_prob(th = th, th_idx = th_idx)

      # shortcuts
      unifreq <- lavcache$unifreq
      uninobs <- lavcache$uninobs
      uniweights <- lavcache$uniweights

      logl <- logl + sum(uniweights * log(uni_pi))

      uniprop <- unifreq / uninobs

      # remove zero props
      # uni.zero.idx <- which(uniprop == 0.0)
      uni_zero_idx <- which((uniprop == 0.0) | !is.finite(uniprop))
      if (length(uni_zero_idx) > 0L) {
        uniprop <- uniprop[-uni_zero_idx]
        uni_pi <- uni_pi[-uni_zero_idx]
        uniweights <- uniweights[-uni_zero_idx]
      }

      fmin <- fmin + sum(uniweights * log(uniprop / uni_pi))
    }

    if (missing == "doubly.robust") {
      # COMPUTE THE SUM OF THE EXPECTED BIVARIATE CONDITIONAL LIKELIHOODS
      # SUM_{i,j} [ E_{Yi,Yj|y^o}(lnf(Yi,Yj))) ]

      # First compute the terms of the summand. Since the cells of
      # pairwiseProbGivObs are zero for the pairs of variables that at least
      # one of the variables is observed (hence not contributing to the summand)
      # there is no need to construct an index vector for summing appropriately
      # within each individual.
      log_pairwise_pi_orig <- log(pairwise_pi_orig)
      pairwise_prob_giv_obs <- lavcache$pairwiseProbGivObs
      tmp_prod <- t(t(pairwise_prob_giv_obs) * log_pairwise_pi_orig)

      sum_elnfij_casewise <- apply(tmp_prod, 1, sum)
      sum_elnfij <- sum(sum_elnfij_casewise)
      logl <- logl + sum_elnfij
      fmin <- fmin - sum_elnfij

      # COMPUTE THE THE SUM OF THE EXPECTED UNIVARIATE CONDITIONAL LIKELIHOODS
      # SUM_{i,j} [ E_{Yj|y^o}(lnf(Yj|yi))) ]

      # First compute the model-implied conditional univariate probabilities
      # p(y_i=a|y_j=b). Let ModProbY1Gy2 be the vector of these
      # probabilities. The order the probabilities
      # are listed in the vector ModProbY1Gy2 is as follows:
      # y1|y2, y1|y3, ..., y1|yp, y2|y1, y2|y3, ..., y2|yp,
      # ..., yp|y1, yp|y2, ..., yp|y(p-1). Within each pair of variables the
      # index "a" which represents the response category of variable yi runs
      # faster than "b" which represents the response category of the given
      # variable yj.
      # The computation of these probabilities are based on the model-implied
      # bivariate probabilities p(y_i=a,y_j=b). To do the appropriate summations
      # and divisions we need some index vectors to keep track of the index i,j,
      # a, and b, as well as the pair index. These index vectors should be
      # computed once and stored in lavcache. About where in the lavaan code
      # we will add the computations and how they will be done please see the
      # file "new objects in lavcache for DR-PL.r"

      idx_pairs <- lavcache$idx.pairs
      idx_cat_y2_split <- lavcache$idx.cat.y2.split
      idx_cat_y1_split <- lavcache$idx.cat.y1.split
      idx_y1 <- lavcache$idx.Y1
      idx_gy2 <- lavcache$idx.Gy2
      # idx_cat_y1 <- lavcache$idx.cat.Y1
      idx_cat_gy2 <- lavcache$idx.cat.Gy2
      id_uni_pr_giv_obs <- lavcache$id.uniPrGivObs
      # the latter keeps track which variable each column of the matrix
      # univariateProbGivObs refers to

      # For the function lav_pml_bivprob_unicondprob see the .r file
      # with the same name.
      mod_prob_y1gy2 <- lav_pml_bivprob_unicondprob(
        biv_prob = pairwise_pi_orig,
        nvar = nvar,
        idx_pairs = idx_pairs,
        idx_y1 = idx_y1,
        idx_gy2 = idx_gy2,
        idx_cat_y1_split = idx_cat_y1_split,
        idx_cat_y2_split = idx_cat_y2_split
      )

      log_mod_prob_y1gy2 <- log(mod_prob_y1gy2)

      # Let univariateProbGivObs be the matrix of the conditional univariate
      # probabilities Pr(y_i=a|y^o) that has been computed in advance and are
      # fed to the DR-PL function. The rows represent different individuals,
      # i.e. nrow=nobs, and the columns different probabilities. The columns
      # are listed as follows: a runs faster than i.

      # Note that the number of columns of univariateProbGivObs is not the
      # same with the length(log_ModProbY1Gy2), actually
      # ncol(univariateProbGivObs) < length(log_ModProbY1Gy2).
      # For this we use the following commands in order to multiply correctly.

      # Compute for each case the product  Pr(y_i=a|y^o) * log[ p(y_i=a|y_j=b) ]
      # i.e. univariateProbGivObs * log_ModProbY1Gy2
      univariate_prob_giv_obs <- lavcache$univariateProbGivObs
      nobs <- nrow(x)
      uniweights_casewise <- lavcache$uniweights.casewise
      id_cases_with_missing <- which(uniweights_casewise > 0)
      no_cases_with_missing <- length(id_cases_with_missing)
      no_obs_casewise <- nvar - uniweights_casewise
      idx_missing_var <- apply(x, 1, function(x) {
        which(is.na(x))
      })
      idx_observed_var <- lapply(idx_missing_var, function(x) {
        c(1:nvar)[-x]
      })
      idx_cat_observed_var <- sapply(1:nobs, function(i) {
        x[i, idx_observed_var[[i]]]
      })
      elnyi_givyjb_casewise <- sapply(1:no_cases_with_missing, function(i) {
        tmp_id_case <- id_cases_with_missing[i]
        tmp_no_mis <- uniweights_casewise[tmp_id_case]
        tmp_idx_mis <- idx_missing_var[[tmp_id_case]]
        tmp_idx_obs <- idx_observed_var[[tmp_id_case]]
        tmp_no_obs <- no_obs_casewise[tmp_id_case]
        tmp_idx_cat_obs <- idx_cat_observed_var[[tmp_id_case]]
        tmp_uni_prob_giv_obs_i <- univariate_prob_giv_obs[tmp_id_case, ]
        sapply(1:tmp_no_mis, function(k) {
          tmp_idx_mis_var <- tmp_idx_mis[k]
          tmp_uni_prob_giv_obs_ik <-
            tmp_uni_prob_giv_obs_i[id_uni_pr_giv_obs == tmp_idx_mis_var]
          tmp_log_mod_prob_y1gy2 <- sapply(1:tmp_no_obs, function(z) {
            log_mod_prob_y1gy2[idx_y1 == tmp_idx_mis_var &
              idx_gy2 == tmp_idx_obs[z] &
              idx_cat_gy2 == tmp_idx_cat_obs[z]]
          })
          sum(tmp_log_mod_prob_y1gy2 * tmp_uni_prob_giv_obs_ik)
        })
      })
      elnyi_givyjb <- sum(unlist(elnyi_givyjb_casewise))
      logl <- logl + elnyi_givyjb
      # for the Fmin function
      fmin <- fmin - elnyi_givyjb
    } # end of if (missing =="doubly.robust")


    ##### Case 2:
    #####  mixed ordered + numeric
    #####  no exogenous covariates
    #####
  } else if (nexo == 0L) {
    # mixed ordered/numeric variables, but no exogenous covariates
    # - no need to compute 'casewise' (log)likelihoods

    pstar_1 <- matrix(0, nvar, nvar) # utility matrix, to get indices
    pstar_1[lav_mat_vech_idx(nvar, diagonal = FALSE)] <- 1:pstar
    # n <- NROW(x)

    log_lik_pair <- numeric(pstar) # logl per pair (summed over cases)
    for (j in seq_len(nvar - 1L)) {
      for (i in (j + 1L):nvar) {
        pstar_idx <- pstar_1[i, j]
        if (ov_types[i] == "numeric" &&
          ov_types[j] == "numeric") {
          log_lik <- lav_mvn_loglik_data(
            y = x[, c(i, j)], wt = wt, mu = mu_hat[c(i, j)],
            sigma_1 = sigma_hat[c(i, j), c(i, j)], casewise = TRUE
          )
          log_lik_pair[pstar_idx] <- sum(log_lik, na.rm = TRUE)
        } else if (ov_types[i] == "numeric" &&
          ov_types[j] == "ordered") {
          # polyserial correlation
          log_lik <- lav_bvmix_lik(
            y1 = x[, i], y2 = x[, j],
            wt = wt,
            evar_y1 = sigma_hat[i, i],
            beta_y1 = mu_hat[i],
            th_y2 = th[th_idx == j],
            rho = cor_hat[i, j], take_log = TRUE
          )
          log_lik_pair[pstar_idx] <- sum(log_lik, na.rm = TRUE)
        } else if (ov_types[j] == "numeric" &&
          ov_types[i] == "ordered") {
          # polyserial correlation
          log_lik <- lav_bvmix_lik(
            y1 = x[, j], y2 = x[, i],
            wt = wt,
            evar_y1 = sigma_hat[j, j],
            beta_y1 = mu_hat[j],
            th_y2 = th[th_idx == i],
            rho = cor_hat[i, j], take_log = TRUE
          )
          log_lik_pair[pstar_idx] <- sum(log_lik, na.rm = TRUE)
        } else if (ov_types[i] == "ordered" &&
          ov_types[j] == "ordered") {
          # polychoric correlation
          pairwise_pi <- lav_bvord_noexo_pi(
            rho = cor_hat[i, j],
            th_y1 = th[th_idx == i],
            th_y2 = th[th_idx == j]
          )
          # avoid zeroes
          pairwise_pi[pairwise_pi < .Machine$double.eps] <-
            .Machine$double.eps
          # note: missing values are just not counted
          freq_1 <- lav_bvord_freq(x[, i], x[, j], wt = wt)
          log_lik_pair[pstar_idx] <- sum(freq_1 * log(pairwise_pi))
        }
      }
    } # all pairs

    na_idx <- which(is.na(log_lik_pair))
    if (length(na_idx) > 0L) {
      lav_msg_warn(gettext("some pairs produce NA values for logl:"),
        lav_msg_view(round(log_lik_pair, 3), "none")
      )
    }

    # sum over pairs
    logl <- sum(log_lik_pair)

    # Fmin
    fmin <- (-1) * logl


    ##### Case 3:
    #####  mixed ordered + numeric
    #####  exogenous covariates
    #####  (conditional.x = TRUE)
  } else {
    lik <- matrix(0, nrow(x), pstar) # likelihood per case, per pair
    pstar_1 <- matrix(0, nvar, nvar) # utility matrix, to get indices
    pstar_1[lav_mat_vech_idx(nvar, diagonal = FALSE)] <- 1:pstar
    # n <- NROW(x)

    for (j in seq_len(nvar - 1L)) {
      for (i in (j + 1L):nvar) {
        pstar_idx <- pstar_1[i, j]
        # cat("pstar.idx =", pstar.idx, "i = ", i, " j = ", j, "\n")
        if (ov_types[i] == "numeric" &&
            ov_types[j] == "numeric") {
          # ordinary pearson correlation
          lik[, pstar_idx] <-
            lav_bvreg_lik(
              y1 = x[, i], y2 = x[, j], exo = exo,
              wt = wt,
              evar_y1 = sigma_hat[i, i],
              beta_y1 = c(mu_hat[i], pi0[i, ]),
              evar_y2 = sigma_hat[j, j],
              beta_y2 = c(mu_hat[j], pi0[j, ]),
              rho = cor_hat[i, j]
            )
        } else if (ov_types[i] == "numeric" &&
                   ov_types[j] == "ordered") {
          # polyserial correlation
          ### FIXME: th.y2 should go into ps_lik!!!
          lik[, pstar_idx] <-
            lav_bvmix_lik(
              y1 = x[, i], y2 = x[, j], exo = exo,
              wt = wt,
              evar_y1 = sigma_hat[i, i],
              beta_y1 = c(mu_hat[i], pi0[i, ]),
              th_y2 = th[th_idx == j],
              sl_y2 = pi0[j, ],
              rho = cor_hat[i, j]
            )
        } else if (ov_types[j] == "numeric" &&
                   ov_types[i] == "ordered") {
          # polyserial correlation
          ### FIXME: th.y1 should go into ps_lik!!!
          lik[, pstar_idx] <-
            lav_bvmix_lik(
              y1 = x[, j], y2 = x[, i], exo = exo,
              wt = wt,
              evar_y1 = sigma_hat[j, j],
              beta_y1 = c(mu_hat[j], pi0[j, ]),
              th_y2 = th[th_idx == i],
              sl_y2 = pi0[i, ],
              rho = cor_hat[i, j]
            )
        } else if (ov_types[i] == "ordered" &&
                   ov_types[j] == "ordered") {
          lik[, pstar_idx] <-
            lav_pml_bi_lik_x(
              y1 = x[, i],
              y2 = x[, j],
              rho = sigma_hat[i, j],
              th_y1 = th[th_idx == i],
              th_y2 = th[th_idx == j],
              exo = exo,
              pi_y1 = pi0[i, ],
              pi_y2 = pi0[j, ],
              missing_ind = missing
            )
        }
      }
    } # all pairs

    # check for zero likelihoods/probabilities
    # FIXME: or should we replace them with a tiny number?
    if (any(lik == 0.0, na.rm = TRUE)) {
      out <- +Inf
      attr(out, "logl") <- as.numeric(NA)
      return(out)
    }

    # loglikelihood
    log_lik_cases <- log(lik)

    # sum over cases
    log_lik_pairs <- colSums(log_lik_cases, na.rm = TRUE)

    # sum over pairs
    logl <- logl_pairs <- sum(log_lik_pairs)

    if (missing == "available.cases" && all(ov_types == "ordered") &&
      nexo != 0L) {
      uni_lik <- matrix(0, nrow(x), ncol(x))
      for (i in seq_len(nvar)) {
        uni_lik[, i] <- lav_pml_uni_lik(
          y1 = x[, i],
          th_y1 = th[th_idx == i],
          exo = exo,
          pi_y1 = pi0[i, ]
        )
      }

      if (any(uni_lik == 0.0, na.rm = TRUE)) {
        out <- +Inf
        attr(out, "logl") <- as.numeric(NA)
        return(out)
      }

      uni_log_lik_cases <- log(uni_lik) * lavcache$uniweights.casewise

      # sum over cases
      uni_log_lik_varwise <- colSums(uni_log_lik_cases)

      # sum over variables
      uni_log_lik <- sum(uni_log_lik_varwise)

      # add with the pairwise part of LogLik
      logl <- logl_pairs + uni_log_lik
    }

    # we minimise
    fmin <- (-1) * logl
  }


  # here, we should have two quantities: logl and Fmin

  # function value as returned to the minimizer
  fx <- fmin

  # attach 'loglikelihood'
  attr(fx, "logl") <- logl

  fx
}

# full information maximum likelihood
# underlying multivariate normal approach (see Joreskog & Moustaki, 2001)
#
lav_model_objective_fml <- function(sigma_hat = NULL, # model-based var/cov/cor
                          th = NULL, # model-based thresholds + means
                          th_idx = NULL, # threshold idx per variable
                          num_idx = NULL, # which variables are numeric
                          x = NULL, # raw data
                          lavcache = NULL) { # patterns

  # YR 27 aug 2013
  # just for fun, and to compare with PML for small models

  # first of all: check if all correlations are within [-1,1]
  # if not, return Inf; (at least with nlminb, this works well)
  cors <- sigma_hat[lower.tri(sigma_hat)]

  if (any(abs(cors) > 1)) {
    return(+Inf)
  }

  nvar <- nrow(sigma_hat)
  # pstar <- nvar * (nvar - 1) / 2
  ov_types <- rep("ordered", nvar)
  if (length(num_idx) > 0L) ov_types[num_idx] <- "numeric"
  mean_1 <- rep(0, nvar)

  # shortcut for all ordered - per pattern
  if (all(ov_types == "ordered")) {
    pat <- lavcache$pat
    npatterns <- nrow(pat)
    freq <- as.numeric(rownames(pat))
    pi0 <- numeric(npatterns)
    th_var <- lapply(1:nvar, function(x) c(-Inf, th[th_idx == x], +Inf))
    # FIXME!!! ok to set diagonal to 1.0?
    diag(sigma_hat) <- 1.0
    for (r in 1:npatterns) {
      # compute probability for each pattern
      lower <- sapply(1:nvar, function(x) th_var[[x]][pat[r, x]])
      upper <- sapply(1:nvar, function(x) th_var[[x]][pat[r, x] + 1L])


      # how accurate must we be here???
      pi0[r] <- sadmvn(lower, upper,
        mean = mean_1, varcov = sigma_hat,
        maxpts = 10000 * nvar, abseps = 1e-07
      )
    }
    # sum (log)likelihood over all patterns
    # LogLik <- sum(log(PI) * freq)

    # more convenient fit function
    prop <- freq / sum(freq)
    # remove zero props # FIXME!!! or add 0.5???
    zero_idx <- which(prop == 0.0)
    if (length(zero_idx) > 0L) {
      prop <- prop[-zero_idx]
      pi0 <- pi0[-zero_idx]
    }
    fmin <- sum(prop * log(prop / pi0))
  } else { # case-wise
    pi0 <- numeric(nobs)
    for (i in 1:nobs) {
      # compute probability for each case
      pi0[i] <- lav_msg_stop(gettext("not implemented"))
    }
    # sum (log)likelihood over all observations
    # log_lik <- sum(log(pi0))
    lav_msg_stop(gettext("not implemented"))
  }

  # function value as returned to the minimizer
  # fx <- -1 * LogLik
  fx <- fmin

  fx
}

lav_model_objective_mml <- function(lavmodel = NULL,
                          mm_theta = NULL,
                          th = NULL,
                          glist = NULL,
                          group = 1L,
                          lavdata = NULL,
                          sample_mean = NULL,
                          sample_mean_x = NULL,
                          lavcache = NULL) {
  # compute case-wise likelihoods
  lik <- lav_model_lik_mml(
    lavmodel = lavmodel, mm_theta = mm_theta, th = th,
    glist = glist, group = group, lavdata = lavdata,
    sample_mean = sample_mean, sample_mean_x = sample_mean_x,
    lavcache = lavcache
  )

  # log + sum over observations
  logl <- sum(log(lik))

  # function value as returned to the minimizer
  fx <- -logl

  fx
}

lav_model_objective_2l <- function(lavmodel = NULL,
                         glist = NULL,
                         y1 = NULL, # only for missing
                         lp = NULL,
                         mp = NULL,
                         lavsamplestats = NULL,
                         lavcache_group = NULL, # only for random slopes
                         group = 1L) {
  # random slopes? (rv() modifier) -> use the random-slope kernel
  if (length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L) {
    rs <- lavcache_group$rs
    if (is.null(rs)) {
      lav_msg_fixme(
        "no rs element found in lavcache; this should not happen")
    }
    loglik <- lav_mvn_cl_rs_m2ll(
      lavmodel = lavmodel, glist = glist, rs = rs,
      log2pi = FALSE, minus_two = TRUE
    )
    objective <- as.numeric(loglik) / (lavsamplestats@ntotal * 2)
    return(objective)
  }

  # compute model-implied statistics for all blocks
  implied <- lav_model_implied(lavmodel, glist = glist)

  # here, we assume only 2!!! levels, at [[1]] and [[2]]
  if (lavmodel@conditional.x) {
    res_sigma_w <- implied$res.cov[[(group - 1) * 2 + 1]]
    res_int_w   <- implied$res.int[[(group - 1) * 2 + 1]]
    res_pi_w    <- implied$res.slopes[[(group - 1) * 2 + 1]]

    res_sigma_b <- implied$res.cov[[(group - 1) * 2 + 2]]
    res_int_b   <- implied$res.int[[(group - 1) * 2 + 2]]
    res_pi_b    <- implied$res.slopes[[(group - 1) * 2 + 2]]
  } else {
    sigma_w <- implied$cov[[(group - 1) * 2 + 1]]
    mu_w    <- implied$mean[[(group - 1) * 2 + 1]]
    sigma_b <- implied$cov[[(group - 1) * 2 + 2]]
    mu_b    <- implied$mean[[(group - 1) * 2 + 2]]
  }

  if (lavsamplestats@missing.flag) {
    if (lavmodel@conditional.x) {
      lav_msg_stop(gettext("multilevel + conditional.x is not ready yet for
                           fiml; rerun with conditional.x = FALSE"))
    }
    y2 <- lavsamplestats@YLp[[group]][[2]]$Y2
    # yp <- lavsamplestats@missing[[group]]
    loglik <- lav_mvn_cl_mi_loglik_samp_2l(
      y1 = y1,
      y2 = y2, lp = lp, mp = mp,
      mu_w = mu_w, sigma_w = sigma_w,
      mu_b = mu_b, sigma_b = sigma_b,
      log2pi = FALSE, minus_two = TRUE
    )
  } else {
    ylp <- lavsamplestats@YLp[[group]]
    if (lavmodel@conditional.x) {
      loglik <- lav_mvreg_cl_loglik_samp_2l(
        ylp = ylp, lp = lp,
        res_sigma_w = res_sigma_w,
        res_int_w = res_int_w, res_pi_w = res_pi_w,
        res_sigma_b = res_sigma_b,
        res_int_b = res_int_b, res_pi_b = res_pi_b,
        log2pi = FALSE, minus_two = TRUE
      )
    } else {
      loglik <- lav_mvn_cl_loglik_samp_2l(
        ylp = ylp, lp = lp,
        mu_w = mu_w, sigma_w = sigma_w,
        mu_b = mu_b, sigma_b = sigma_b,
        log2pi = FALSE, minus_two = TRUE
      )
    }
  }

  # minimize (we already did -2*)
  objective <- 1 * loglik

  # divide by (N*2)
  objective <- objective / (lavsamplestats@ntotal * 2)

  # should be strictly positive
  # if(objective < 0) {
  #   objective <- +Inf
  # }

  objective
}

# find ALL (eligible) *model-implied* instrumental variables (no pruning yet)
#
# three algorithms:
#   1) using 'total effects' of errors/disturbances, based on Bollen & Bauer
#      (2004) [algorithm = "bb2004"]
#   2) treating errors/disturbances as latent variables (as in the
#      comprehensive RAM model); similar to miivs() in MIIVsem package
#      [algorithm = "miivsem"]
#   3) using an explicit expression (due to Albert Maydeu-Olivares) for
#      cov(u,y) [algorithm = "covuy"] (the default)
#
# YR 29 Dec 2025 - first version (bb2004 + miivsem)
# YR 24 Jan 2025 - covuy algorithm (including higher-order factors)

lav_model_find_iv <- function(lavobject = NULL, lavmodel = NULL,
                              lavpartable = NULL, algorithm = "covuy",
                              output = "list", drop_list_single_group = FALSE) {
  # check output
  output <- tolower(output)
  stopifnot(output %in% c("list", "table"))

  # lavobject or components?
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))
    lavpartable <- lavobject@ParTable
    lavpta <- lavobject@pta
    lavmodel <- lavobject@Model
  } else {
    # get lavpta
    lavpta <- lav_pt_attributes(lavpartable)
  }

  # sanity checks
  nblocks <- lavpta$nblocks
  for (b in seq_len(nblocks)) {
    lv_idx <- lavpta$vidx$lv.regular[[b]]
    lv_marker <- lavpta$vnames$lv.marker[[b]]
    if (length(lv_idx) > 0L) {
      # do have 'clear' marker/scaling indicators?
      empty_idx <- which(nchar(lv_marker) == 0L)
      if (length(empty_idx) > 0L) {
        tmp_string <- paste(names(lv_marker)[empty_idx], collapse = " ")
        lav_msg_stop(gettextf("no clear marker/scaling indicator found for
                               factor(s): %s", tmp_string))
      }
      # marker/scaling indicator cannot be a dependent variable in an ~ equation
      badmarker_idx <- which(lv_marker %in% lavpta$vnames$eqs.y[[b]])
      if (length(badmarker_idx) > 0L) {
        tmp_string <- paste(lv_marker[badmarker_idx], collapse = " ")
        lav_msg_stop(gettextf("marker/scaling indicator cannot be a dependent
                               variable in a regression: %s", tmp_string))
      }
    }
  }

  # first find all model-implied instruments
  algorithm <- tolower(algorithm)
  if (algorithm == "bb2004") {
    iv_list <- lav_model_find_iv_bb2004(lavmodel = lavmodel, lavpta = lavpta)
  } else if (algorithm == "miivsem") {
    iv_list <- lav_model_find_iv_miivsem(lavmodel = lavmodel, lavpta = lavpta)
  } else {
    iv_list <- lav_model_find_iv_covuy(lavmodel = lavmodel, lavpta = lavpta)
  }

  # check for user-specified instruments
  if (any(lavpartable$op == "|~")) {
    # restrict the |~ rows to the current block, so that in a multiple-group
    # analysis the per-group instruments are not pooled (which would duplicate
    # them, e.g. iv = c("z", "z"), and make the instrument covariance singular)
    pt_block <- if (!is.null(lavpartable$block)) {
      lavpartable$block
    } else {
      rep(1L, length(lavpartable$op))
    }
    for (b in seq_len(nblocks)) {
      iv_list[[b]] <- lapply(iv_list[[b]], function(eq) {
        # the user specifies the dependent variable of the equation on the
        # lhs of |~; for a latent variable, this is its scaling indicator,
        # which is stored in eq$lhs_new (eq$lhs holds the latent name). Match
        # on either, so that |~ works for both observed and latent dependents.
        lhs <- eq$lhs[1] # we assume a single lhs
        lhs_new <- eq$lhs_new[1]
        iv_idx <- which(lavpartable$op == "|~" & pt_block == b &
          (lavpartable$lhs == lhs | lavpartable$lhs == lhs_new))
        if (length(iv_idx)) {
          # override iv
          eq$iv <- lavpartable$rhs[iv_idx]
          eq$iv_type <- "user"
        }
        eq
      })
    } # blocks
  }

  if (output == "table") {
    table <- vector("list", length = nblocks)
    for (b in seq_len(nblocks)) {
      eqs <- iv_list[[b]]
      lhs <- sapply(eqs, "[[", "lhs")
      rhs <- sapply(lapply(eqs, "[[", "rhs"), paste, collapse = " + ")
      lhs_new <- sapply(eqs, "[[", "lhs_new")
      rhs_new <- sapply(lapply(eqs, "[[", "rhs_new"), paste, collapse = " + ")
      iv <- sapply(lapply(eqs, "[[", "iv"), paste, collapse = ", ")
      type <- sapply(eqs, "[[", "iv_type")
      table[[b]] <- data.frame(
        lhs = lhs, rhs = rhs,
        lhs_new = lhs_new, rhs_new = rhs_new, type = type, iv = iv
      )
      class(table[[b]]) <- c("lavaan.data.frame", "data.frame")
    }
    out <- table
  } else {
    out <- iv_list
  }

  if (nblocks == 1L && drop_list_single_group) {
    out <- out[[1]]
  }

  out
}


# algorithm 1:
# loosely based on Bollen & Bauer (2004), but with support for higher-order
# factors
lav_model_find_iv_bb2004 <- function(lavmodel = NULL, lavpta = NULL) {
  # check representation
  if (lavmodel@representation != "LISREL") {
    lav_msg_stop(gettext(
      "this function only works with LISREL representation",
      " (for now)"
    ))
  }

  # create model matrices with 'user/partable' entries
  glist <- lavmodel@GLIST
  for (mm in seq_along(glist)) {
    dimnames(glist[[mm]]) <- lavmodel@dimNames[[mm]]
    m_idx <- lavmodel@m.user.idx[[mm]]
    x_idx <- lavmodel@x.user.idx[[mm]]
    glist[[mm]][, ] <- 0.0
    glist[[mm]][m_idx] <- x_idx
  }

  # number of blocks
  nblocks <- lavpta$nblocks
  lambda_idx <- which(names(glist) == "lambda")
  beta_idx <- which(names(glist) == "beta")
  psi_idx <- which(names(glist) == "psi")
  theta_idx <- which(names(glist) == "theta")
  if (lavmodel@meanstructure) {
    nu_idx <- which(names(glist) == "nu")
    alpha_idx <- which(names(glist) == "alpha")
  }

  # repeat for every block
  iv_list <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    # extract information to create Bollen & Bauer (2004) matrices
    ov_names <- lavpta$vnames$ov[[b]]
    nvar <- length(ov_names)
    lv_names <- lavpta$vnames$lv.regular[[b]]
    lv_marker <- lavpta$vnames$lv.marker[[b]]
    lv_marker_orig <- lavpta$vnames$lv.marker[[b]]
    lv_ho <- lavpta$vnames$lv.ho[[b]]

    lv_idx <- lavpta$vidx$lv.regular[[b]]
    lv_x_idx <- lavpta$vidx$lv.x[[b]]
    lv_marker_idx <- lavpta$vidx$lv.marker[[b]]

    # dummy y?
    if (length(lavmodel@ov.y.dummy.ov.idx[[b]]) > 0L) {
      lv_idx <- c(lv_idx, lavmodel@ov.y.dummy.lv.idx[[b]])
      lv_marker <- c(lv_marker_orig, ov_names[lavmodel@ov.y.dummy.ov.idx[[b]]])
      names(lv_marker) <- c(
        names(lv_marker_orig),
        ov_names[lavmodel@ov.y.dummy.ov.idx[[b]]]
      )
      lv_marker_idx <- c(lv_marker_idx, lavmodel@ov.y.dummy.ov.idx[[b]])
    }

    # keep track of higher order factors
    lv_ho_idx <- which(is.na(lv_marker_idx)) # same as lavpta$vidx$lv.ho
    if (length(lv_ho_idx) > 0L) {
      lv_in_marker <- lv_marker[lv_marker %in% lv_names]
      while (length(lv_in_marker) > 0L) {
        new_markers <- unname(lv_marker[lv_in_marker])
        lv_marker[match(lv_in_marker, lv_marker)] <- new_markers
        lv_in_marker <- lv_marker[lv_marker %in% lv_names]
      }
      lv_marker_idx <- match(lv_marker, ov_names)
    }

    # model matrices for this block
    lambda <- glist[[lambda_idx[b]]]
    beta <- glist[[beta_idx[b]]]
    theta <- glist[[theta_idx[b]]]
    psi <- glist[[psi_idx[b]]]
    if (lavmodel@meanstructure) {
      nu <- glist[[nu_idx[b]]]
      alpha <- glist[[alpha_idx[b]]]
    }
    if (is.null(beta)) {
      beta <- matrix(0, ncol(lambda), ncol(lambda))
      colnames(beta) <- rownames(beta) <- colnames(lambda)
    }

    # binary model matrices: nonzero = 1, zero = 0
    if (length(lv_idx) == 0L) {
      lambda_bin <- diag(nrow = ncol(lambda))
    } else {
      lambda_bin <- (lambda != 0) * 1L
    }
    beta_bin <- (beta != 0) * 1L
    ibinv <- solve(diag(nrow(beta_bin)) - beta_bin)
    ibinv_bin <- (ibinv != 0) * 1L

    # construct pred
    if (length(lv_idx) > 0L) {
      reg_bin <- rbind(lambda_bin, beta_bin)
      pred_orig <- rbind(lambda, beta)
      if (lavmodel@meanstructure) {
        int_orig <- rbind(nu, alpha)
      }
    } else {
      reg_bin <- beta_bin
      pred_orig <- beta
      if (lavmodel@meanstructure) {
        int_orig <- alpha
      }
    }
    pred_idx <- numeric(ncol(reg_bin))
    if (length(lv_marker_idx) > 0L) {
      pred_idx[lv_idx] <- lv_marker_idx
    }
    if (length(lavmodel@ov.x.dummy.ov.idx[[b]]) > 0L) {
      pred_idx[lavmodel@ov.x.dummy.lv.idx[[b]]] <-
        lavmodel@ov.x.dummy.ov.idx[[b]]
    }
    if (length(lavmodel@ov.y.dummy.ov.idx[[b]]) > 0L) {
      pred_idx[lavmodel@ov.y.dummy.lv.idx[[b]]] <-
        lavmodel@ov.y.dummy.ov.idx[[b]]
    }
    tmp <- t(t(reg_bin) * pred_idx)
    # remove scaling '1' for markers (if any)
    if (length(lv_marker_idx) > 0L) {
      row_idx <- match(lv_marker_orig, rownames(tmp))
      col_idx <- match(names(lv_marker_orig), colnames(tmp))
      tmp[cbind(row_idx, col_idx)] <- 0L
    }
    # keep only the 'ov' part
    pred <- tmp[seq_along(ov_names), , drop = FALSE]
    # replace marker rows, with beta entries of their corresponding lv's
    # dummy y + regular lv's
    if (length(lv_idx) > 0L && !is.null(beta)) {
      tmp_lv <- tmp_lv2 <- tmp[length(ov_names) + lv_idx, , drop = FALSE]
      orig_lv <- pred_orig[length(ov_names) + lv_idx, , drop = FALSE]
      if (lavmodel@meanstructure) {
        int_lv <- int_orig[length(ov_names) + lv_idx, , drop = FALSE]
        tmp_lv2 <- cbind(int_lv, tmp_lv)
      }

      # tmp[lv.marker.idx, ] <- tmp[length(ov.names) + lv.idx, ]
      # careful, lv.marker.idx may contain the same element multiple times!
      zerolv_idx <- which(apply(tmp_lv2, 1, function(x) all(x == 0)))
      if (length(zerolv_idx) > 0L) {
        tmp_lv <- tmp_lv[-zerolv_idx, , drop = FALSE]
        orig_lv <- orig_lv[-zerolv_idx, , drop = FALSE]
        if (lavmodel@meanstructure) {
          int_lv <- int_lv[-zerolv_idx, , drop = FALSE]
        }
      }
      r_idx <- match(lv_marker[rownames(tmp_lv)], rownames(pred))
      pred[r_idx, ] <- tmp_lv
      pred_orig[r_idx, ] <- orig_lv
      rownames(pred_orig)[r_idx] <- rownames(tmp_lv)
      pred_orig <- pred_orig[1:nvar, , drop = FALSE]
      if (lavmodel@meanstructure) {
        int_orig[r_idx, ] <- int_lv
        # also exogenous ov's!
        lv_x_dummy_idx <- lavmodel@ov.x.dummy.lv.idx[[b]]
        if (length(lv_x_dummy_idx) > 0L) {
          int_orig[lavmodel@ov.x.dummy.ov.idx[[b]], 1] <-
                                                alpha[lv_x_dummy_idx, 1]
        }
        rownames(int_orig)[r_idx] <- rownames(int_lv)
        int_orig <- int_orig[1:nvar, , drop = FALSE]
      }
    }
    # non-zero entries of 'pred' contain the marker idx of the latent
    # predictors (and perhaps the dummy ov indices)

    # construct comp (composite error term)
    # - for each dependent: the corresponding element in Theta/Psi (y_res)
    # - for each (latent) predictor: the corresponding *marker* element in
    #   Theta (x_res)
    # - for the markers: the disturbance (psi) of the corresponding
    #   lv's (included in x_res)
    # - for higher-order factors: the disturbance (psi) of their 'marker'
    #   factor
    y_res <- diag(theta)
    if (length(lv_marker_idx) > 0L) {
      x_res <- pred
      # x_res[pred != 0] <- y_res[as.vector(pred[pred != 0])]
      x_res[pred != 0] <- y_res[pred]
      x_res[cbind(lv_marker_idx, seq_along(lv_idx))] <- diag(psi)[lv_idx]
      if (length(lv_ho_idx) > 0L) {
        # higher-order factors: what are their indicators?
        # add diag(psi) value in this row
        x_res <- t(apply(x_res, 1, function(x) {
          res <- x
          hof <- x[lv_ho_idx]
          hof_nonzero <- hof[hof != 0]
          if (length(hof_nonzero) > 0L) {
            # keep going until we only have first-order factors (or ovs)
            target <- unname(lv_marker_orig[names(hof_nonzero)])
            ho_in_target <- target[target %in% lv_ho]
            while (length(ho_in_target) > 0L) {
              new <- unname(lv_marker_orig[ho_in_target])
              target <- c(target, new)
              ho_in_target <- new[new %in% lv_ho]
            }

            if (length(target)) {
              res[target] <- diag(psi)[target]
            }
          }
          res
        }))
      }
      # comp <- cbind(y_res, x_res)[1:nvar, ]
      comp <- cbind(y_res, x_res)
    } else {
      # no latent variables
      comp <- cbind(y_res)
    }

    # construct total: total effects of errors/disturbances on
    #                  EACH observed variable in the model
    # - direct effect of 'epsilon': diag(theta)
    # - total effect of 'disturbance': Lambda %*% (ii - Beta)^{-1}
    lv_res <- diag(psi)
    lv_res[lv_x_idx] <- 0 # remove 'total/exo' variances from psi
    y_res <- diag(theta)
    t_res <- t(t(lambda_bin %*% ibinv_bin) * lv_res)
    total <- cbind(y_res, t_res)

    # construct piv (potential ivs):
    # initial set of potential instruments for each equation, by selecting
    # observed variables that are unaffected by the disturbances or
    # uniquenesses in that equation
    piv <- matrix(1:nvar, nrow = nvar, ncol = nvar, byrow = TRUE)
    colnames(piv) <- rownames(piv) <- ov_names
    for (i in 1:nvar) {
      res_id <- unique(comp[i, comp[i, ] != 0])
      tmp <- apply(total, 2, function(x) !x %in% res_id)
      rm_idx <- which(!apply(tmp, 1, all))
      piv[i, rm_idx] <- 0
    }
    # diagonal must be zero (not needed, but just to be safe)
    diag(piv) <- 0
    # ov.y -> zero column
    if (length(lavpta$vidx$ov.y[[b]]) > 0L) {
      piv[, lavpta$vidx$ov.y[[b]]] <- 0
    }
    # ov.x -> zero row
    if (length(lavpta$vidx$ov.x[[b]]) > 0L) {
      piv[lavpta$vidx$ov.x[[b]], ] <- 0
    }


    # construct iv
    # Starting from piv, remove any potential ivs that are affected by a
    # disturbance that correlates with any disturbance in the composite
    # disturbance term
    iv <- piv
    for (i in 1:nvar) {
      # check if we need instruments at all
      rhs <- pred[i, pred[i, ] != 0, drop = FALSE]
      if (length(rhs) > 0L && all(comp[i, colnames(rhs)] == 0)) {
        # no instruments needed
        iv[i, -rhs] <- 0
      } else {
        for (p in 1:nvar) {
          # do check for this iv, if false, set to zero
          comp_idx <- which(y_res %in% comp[i, ])
          total_idx <- which(y_res %in% total[piv[i, p], ])
          cov_values <- theta[as.matrix(expand.grid(comp_idx, total_idx))]
          if (any(cov_values != 0)) {
            iv[i, p] <- 0
          }
        }
      } # instruments needed
    }

    # add intercepts here
    if (lavmodel@meanstructure) {
      colnames(int_orig) <- "1"
      pred <- cbind(int_orig, pred)
      pred_orig <- cbind(int_orig, pred_orig)

      idx <- match(names(lv_marker), rownames(pred))
      rownames(pred)[idx] <- lv_marker
    }

    # remove 'empty' rows
    empty_idx <- which(apply(pred, 1L, function(x) all(x == 0)) &
      rownames(pred) %in% ov_names)
    if (length(empty_idx) > 0L) {
      iv <- iv[-empty_idx, , drop = FALSE]
      comp <- comp[-empty_idx, , drop = FALSE]
      pred <- pred[-empty_idx, , drop = FALSE]
      pred_orig <- pred_orig[-empty_idx, , drop = FALSE]
      if (lavmodel@meanstructure) {
        int_orig <- int_orig[-empty_idx, , drop = FALSE]
      }
    }
    # replace row/colnames of pred by marker names
    idx <- match(names(lv_marker), colnames(pred))
    colnames(pred)[idx] <- lv_marker

    # prepare list
    eqs <- vector("list", length = nrow(iv))
    for (j in seq_along(eqs)) {
      cet <- colnames(comp)[which(comp[j, ] != 0)]
      x_idx <- which(pred[j, ] != 0)
      ptint <- integer(0L)
      if (lavmodel@meanstructure) {
        x_idx <- x_idx[-1]
        ptint <- unname(pred_orig[j, 1])
      }
      if (lavmodel@meanstructure && length(x_idx) == 0L) {
        rhs_new <- "1"
        rhs <- "1"
        miiv <- "1"
      } else {
        rhs_new <- colnames(pred[j, x_idx, drop = FALSE])
        rhs <- colnames(pred_orig[j, x_idx, drop = FALSE])
        miiv <- colnames(iv[j, iv[j, ] != 0, drop = FALSE])
      }
      iv_type <- "miiv"
      if (identical(rhs_new, miiv)) {
        iv_type <- "ols"
      }
      eqs[[j]] <- list(
        lhs_new = rownames(pred)[j],
        rhs_new = rhs_new,
        lhs = rownames(pred_orig)[j],
        rhs = rhs,
        pt = unname(pred_orig[j, x_idx]),
        ptint = ptint,
        cet = cet,
        markers = unique(lv_marker[cet[-1]]),
        iv_type = iv_type,
        iv = miiv,
        miiv = miiv
      )
    }

    # reorder so that ov lhs come first, then lv lhs
    lhs <- sapply(eqs, "[[", "lhs")
    ov_idx <- match(ov_names[ov_names %in% lhs], lhs)
    lv_idx <- seq_along(lhs)[-ov_idx]
    eqs <- eqs[c(ov_idx, lv_idx)]

    iv_list[[b]] <- eqs
  } # blocks

  iv_list
}

# algorithm 2: create a model where error/disturbance terms are latent
#              variables (a bit like the comprehensive RAM representation)
#              and compute 'big_sigma':
#              - to identify ov's that are uncorrelated with the errors in
#                the composite error term
#              - to identify ov's that are correlated with at least one
#                marker
#              the intersection give the (full list of) miivs
#
# the implementation below is loosely based on (but not identical to) the
# 'miivs()' function from the MIIVsem package
#
lav_model_find_iv_miivsem <- function(lavmodel = NULL, lavpta = NULL) {
  # check representation
  if (lavmodel@representation != "LISREL") {
    lav_msg_stop(gettext(
      "this function only works with LISREL representation",
      " (for now)"
    ))
  }

  # create model matrices with 'user/partable' entries
  glist <- lavmodel@GLIST
  for (mm in seq_along(glist)) {
    dimnames(glist[[mm]]) <- lavmodel@dimNames[[mm]]
    m_idx <- lavmodel@m.user.idx[[mm]]
    x_idx <- lavmodel@x.user.idx[[mm]]
    glist[[mm]][, ] <- 0.0
    glist[[mm]][m_idx] <- x_idx
  }
  lambda_idx <- which(names(glist) == "lambda")
  beta_idx <- which(names(glist) == "beta")
  psi_idx <- which(names(glist) == "psi")
  theta_idx <- which(names(glist) == "theta")
  if (lavmodel@meanstructure) {
    nu_idx <- which(names(glist) == "nu")
    alpha_idx <- which(names(glist) == "alpha")
  }

  # nblocks
  nblocks <- lavmodel@nblocks

  # repeat for every block
  iv_list <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    lv_names <- lavpta$vnames$lv.regular[[b]]
    ov_names <- lavpta$vnames$ov[[b]]
    lv_marker <- lavpta$vnames$lv.marker[[b]]

    # model matrices for this block
    lambda <- glist[[lambda_idx[b]]]
    beta <- glist[[beta_idx[b]]]
    theta <- glist[[theta_idx[b]]]
    psi <- glist[[psi_idx[b]]]
    if (lavmodel@meanstructure) {
      nu <- glist[[nu_idx[b]]]
      alpha <- glist[[alpha_idx[b]]]
    }


    # 'exogenous' variables (in beta/psi) (including dummy ov's)
    if (is.null(beta)) {
      lv_x_idx <- lavpta$vidx$lv.x[[b]]
    } else {
      lv_x_idx <- which(apply(beta, 1L, function(x) all(x == 0)))
    }

    # lambda+beta
    if (is.null(beta)) {
      beta <- matrix(0, ncol(lambda), ncol(lambda))
    }
    lambda_beta <- rbind(lambda, beta)
    pred_orig <- lambda_beta
    lambda_beta[lambda_beta != 0] <- as.numeric(NA)
    # add markers
    row_idx <- match(lavpta$vnames$lv.marker[[b]], rownames(lambda_beta))
    col_idx <- match(names(lavpta$vnames$lv.marker[[b]]), colnames(lambda_beta))
    lambda_beta[cbind(row_idx, col_idx)] <- 1
    pred <- lambda_beta
    pred_orig_full <- pred_orig
    empty_idx <- which(apply(lambda_beta, 1L, function(x) all(x == 0)))
    if (length(empty_idx) > 0) {
      lambda_beta <- lambda_beta[-empty_idx, , drop = FALSE]
      pred_orig <- pred_orig[-empty_idx, , drop = FALSE]
    }

    # error + zero matrix, to create gamma and beta
    error_matrix <- diag(nrow = nrow(lambda_beta))
    rownames(error_matrix) <- rownames(lambda_beta)
    colnames(error_matrix) <- paste("e.", rownames(lambda_beta), sep = "")
    zero_matrix <- error_matrix * 0
    colnames(zero_matrix) <- rownames(lambda_beta)

    # construct gamma
    gamma <- cbind(
      lambda_beta[, lv_x_idx, drop = FALSE],
      error_matrix
    )
    # gamma_orig <- cbind(
    #   pred_orig[, lv_x_idx, drop = FALSE],
    #   error_matrix
    # )

    # construct beta
    beta <- beta_orig <- zero_matrix
    tmp <- lambda_beta[, -lv_x_idx, drop = FALSE]
    beta[, colnames(tmp)] <- tmp
    tmp <- pred_orig[, -lv_x_idx, drop = FALSE]
    beta_orig[, colnames(tmp)] <- tmp

    # construct Phi
    phi <- lav_mat_bdiag(
      psi[lv_x_idx, lv_x_idx, drop = FALSE],
      theta,
      psi[-lv_x_idx, -lv_x_idx, drop = FALSE]
    )
    empty_idx <- which(apply(phi, 1L, function(x) all(x == 0)))
    if (length(empty_idx) > 0) {
      phi <- phi[-empty_idx, -empty_idx, drop = FALSE]
    }
    colnames(phi) <- rownames(phi) <- colnames(gamma)
    phi[phi != 0] <- as.numeric(NA)

    tmp <- crossprod(gamma)
    tmp[, ] <- 0
    beta_1 <- lav_mat_bdiag(beta, tmp)
    ii <- diag(nrow(beta_1))
    diag(tmp) <- 1
    gamma <- rbind(gamma, tmp)

    # to compute big_sigma, we can replace NA by 1
    beta_1[is.na(beta_1)] <- 1
    gamma[is.na(gamma)] <- 1
    phi[is.na(phi)] <- 1
    ib_inv <- solve(ii - beta_1)
    big_sigma <- ib_inv %*% gamma %*% phi %*% t(gamma) %*% t(ib_inv)
    sigma_ov <- big_sigma[ov_names, , drop = FALSE] # only rows with ov.names
    # e_names <- colnames(gamma)[grep("e\\.", colnames(gamma))]


    # matrix of all regressions (nothing has been removed yet, here)
    gamma_beta <- pred
    gamma_beta_orig <- pred_orig_full

    # add intercepts here
    if (lavmodel@meanstructure) {
      int <- rbind(nu, alpha)
      colnames(int) <- "1"
      int_orig <- int
      int[int != 0] <- as.numeric(NA)
      idx <- match(lv_marker, rownames(int))
      int[idx, 1] <- 1
      gamma_beta <- cbind(int, gamma_beta)
      gamma_beta_orig <- cbind(int_orig, gamma_beta_orig)
    }

    # select only those rows of gamma_beta that have at least one free (NA)
    # element
    free_idx <- which(apply(is.na(gamma_beta), 1, any))
    eqs_y_free <- unique(rownames(gamma_beta)[free_idx])
    eqs <- vector("list", length = length(eqs_y_free))
    for (j in seq_along(eqs)) {
      # lhs
      eq_lhs <- eq_lhs_orig <- eqs_y_free[j]

      # index in gamma_beta
      jj <- free_idx[j]

      # lhs + rhs
      zero_idx <- which(gamma_beta[jj, ] != 0) # fixed or error term
      na_idx <- which(is.na(gamma_beta[jj, ])) # free/NA
      eq_rhs <- eq_rhs_orig <- colnames(gamma_beta)[c(zero_idx, na_idx)]
      pt <- unname(gamma_beta_orig[jj, na_idx])
      ptint <- integer(0L)
      if (lavmodel@meanstructure) {
        ptint <- unname(int_orig[jj, 1L])
        eq_rhs <- eq_rhs[-1]
        eq_rhs_orig <- eq_rhs_orig[-1]
      }
      eq_all <- c(eq_lhs, eq_rhs)

      # markers + composite error term (cet)
      cet <- paste("e.", eq_lhs, sep = "")

      # replace all latent variables by their markers
      markers <- character(0L)
      lv_in_eq <- eq_all[eq_all %in% lv_names]
      if (length(lv_in_eq) > 0L) {
        markers <- unname(lv_marker[lv_in_eq])
        # replace in eq_all
        eq_all[match(lv_in_eq, eq_all)] <- markers
        # add marker errors to composite error term
        cet <- c(cet, paste("e.", markers, sep = ""))
      }

      # any markers that are still latent? (eg higher-order factors)
      if (any(eq_all %in% lv_names)) {
        lv_in_eq <- eq_all[eq_all %in% lv_names]
        while (length(lv_in_eq) > 0L) {
          new_markers <- unname(lv_marker[lv_in_eq])
          eq_all[match(lv_in_eq, eq_all)] <- new_markers
          markers[match(lv_in_eq, markers)] <- new_markers
          cet <- c(cet, paste("e.", new_markers, sep = ""))
          lv_in_eq <- eq_all[eq_all %in% lv_names]
        }
      }
      markers <- unique(markers)
      eq_lhs <- eq_all[1]
      eq_rhs <- eq_all[-1]

      if (length(eq_rhs) > 0L) {
        # observed variables that are uncorrelated with all the components of
        # the composite error term
        ov_uncorrelated_with_cet <-
          ov_names[apply(sigma_ov[, cet, drop = FALSE] == 0, 1L, all)]

        # observed variables that have at least one non-zero implied correlation
        # with any of the marker indicators (vector i).
        ov_correlated_with_markers <-
          ov_names[apply(sigma_ov[, markers, drop = FALSE] != 0, 1, all)]

        # valid instruments
        miivs <- intersect(ov_uncorrelated_with_cet, ov_correlated_with_markers)
      } else {
        eq_rhs <- "1"
        eq_rhs_orig <- "1"
        miivs <- "1"
      }

      iv_type <- "miiv"
      if (identical(eq_rhs, miivs)) {
        iv_type <- "ols"
      }

      eqs[[j]] <- list(
        lhs_new = eq_lhs,
        rhs_new = eq_rhs,
        lhs = eq_lhs_orig,
        rhs = eq_rhs_orig,
        pt = pt,
        ptint = ptint,
        cet = cet,
        markers = markers,
        iv_type = iv_type,
        iv = miivs,
        miiv = miivs
      )
    }

    # reorder so that ov lhs come first, then lv lhs
    lhs <- sapply(eqs, "[[", "lhs")
    ov_idx <- match(ov_names[ov_names %in% lhs], lhs)
    lv_idx <- seq_along(lhs)[-ov_idx]
    eqs <- eqs[c(ov_idx, lv_idx)]

    iv_list[[b]] <- eqs
  } # nblocks


  iv_list
}

# algorithm 3:
# - use random data to create sigma (used to check if predictors are correlated
#   with instruments)
# - use an explicit expression to compute cov(u,y), where u_j is the disturbance
#   term of an equation (used to verify that instruments are not correlated
#   with this disturbance term)
lav_model_find_iv_covuy <- function(lavmodel = NULL, lavpta = NULL) {
  # check representation
  if (lavmodel@representation != "LISREL") {
    lav_msg_stop(gettext(
      "this function only works with LISREL representation",
      " (for now)"
    ))
  }

  # create model matrices with 1) 'user/partable' entries, 2) random entries
  glist <- glistr <- lavmodel@GLIST
  glist_names <- names(glist)
  for (mm in seq_along(glist)) {
    dimnames(glist[[mm]]) <- lavmodel@dimNames[[mm]]
    m_idx <- lavmodel@m.user.idx[[mm]]
    x_idx <- lavmodel@x.user.idx[[mm]]
    # partable entries
    glist[[mm]][, ] <- 0.0
    glist[[mm]][m_idx] <- x_idx
    # random entries (overriding user specified!)
    f_idx <- lavmodel@m.free.idx[[mm]]
    glistr[[mm]][f_idx] <- (runif(length(f_idx), min = 0.1, max = 0.5) *
      sign(runif(length(f_idx), min = -1, max = 1)))
    if (glist_names[mm] %in% c("theta", "psi")) {
      tmp <- glistr[[mm]]
      # make sure diagonal is large enough (but only for nonzero values)
      zerodiag_idx <- which(diag(tmp) == 0)
      diag(tmp) <- diag(tmp) + 1.5
      diag(tmp)[zerodiag_idx] <- 0
      # make symmetric
      glistr[[mm]] <- (tmp + t(tmp)) / 2
    }
  }

  # generate random sigma per block
  sigma <- lav_model_sigma(lavmodel, glist = glistr, extra = FALSE)
  sigma_aug <- lav_model_cov_both(lavmodel, glist = glistr)

  # number of blocks
  nblocks <- lavpta$nblocks
  lambda_idx <- which(glist_names == "lambda")
  beta_idx <- which(glist_names == "beta")
  if (lavmodel@meanstructure) {
    nu_idx <- which(names(glist) == "nu")
    alpha_idx <- which(names(glist) == "alpha")
  }


  # repeat for every block
  iv_list <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    # extract information to create Bollen & Bauer (2004) matrices
    ov_names <- lavmodel@dimNames[[lambda_idx[b]]][[1]]
    lv_names <- lavmodel@dimNames[[lambda_idx[b]]][[2]]
    both_names <- c(ov_names, lv_names)
    nvar <- length(ov_names)
    nfac <- length(lv_names)

    # model matrices for this block
    lambda <- glist[[lambda_idx[b]]]
    beta <- glist[[beta_idx[b]]]
    if (is.null(beta)) {
      beta <- matrix(0, nfac, nfac)
    }
    if (lavmodel@meanstructure) {
      nu <- glist[[nu_idx[b]]]
      alpha <- glist[[alpha_idx[b]]]
    }

    # sigma + sigma_aug
    this_sigma <- sigma[[b]]
    rownames(this_sigma) <- colnames(this_sigma) <- ov_names
    this_sigma_aug <- sigma_aug[[b]]
    rownames(this_sigma_aug) <- colnames(this_sigma_aug) <- both_names

    lambdar <- glistr[[lambda_idx[b]]]
    if (is.null(glistr[[beta_idx[b]]])) {
      betar <- diag(1, nrow = nfac)
    } else {
      betar <- glistr[[beta_idx[b]]]
    }

    # all markers (including higher-order)
    lv_marker <- lv_names
    names(lv_marker) <- lv_names
    lv_marker[match(names(lavpta$vnames$lv.marker[[b]]), lv_names)] <-
      lavpta$vnames$lv.marker[[b]]
    lv_marker_idx <- match(lv_marker, both_names)

    # construct cov_u_y
    pi_mat <- rbind(lambdar, betar)
    rownames(pi_mat) <- both_names
    colnames(pi_mat) <- lv_names
    pi_mat[lv_marker_idx, ] <- betar
    pi_zero_mat <- matrix(0, nrow = nrow(pi_mat), ncol = nrow(pi_mat))
    pi_zero_mat[, lv_marker_idx] <- pi_mat
    cov_u_y_aug <- this_sigma_aug - pi_zero_mat %*% this_sigma_aug
    # zap small elements to be exactly zero
    cov_u_y_aug[abs(cov_u_y_aug) < 1e-07] <- 0.0

    if (length(lavpta$vnames$lv.ho[[b]]) > 0L) {
      # 'add' lv marker rows to ov marker rows
      ov_marker <- lv_marker
      lv_names_nox <- lv_names[!lv_names %in% ov_names]
      lv_in_marker <- lv_beta <- ov_marker[ov_marker %in% lv_names_nox]
      while (length(lv_in_marker) > 0L) {
        new_markers <- unname(ov_marker[lv_in_marker])
        ov_marker[match(lv_in_marker, ov_marker)] <- new_markers
        lv_in_marker <- ov_marker[ov_marker %in% lv_names_nox]
      }
      ho_idx <- match(names(lv_beta), both_names)
      target_idx <- match(ov_marker[names(lv_beta)], both_names)
      cov_u_y_aug[target_idx, ] <-
        cov_u_y_aug[target_idx, ] + cov_u_y_aug[ho_idx, ]
    } else {
      ov_marker <- lv_marker
    }
    cov_u_y <- cov_u_y_aug[seq_len(nvar), seq_len(nvar)]

    # construct pred
    pred <- rbind(lambda, beta)

    # if meanstructure, add intercepts
    if (lavmodel@meanstructure) {
      int <- rbind(nu, alpha)
      colnames(int) <- "1" # shorter than "intercept"
      pred <- cbind(int, pred)
    }

    pred_orig <- pred
    # replace col/rownames of pred by marker names
    idx <- match(names(ov_marker), colnames(pred))
    idx <- idx[!is.na(idx)]
    colnames(pred)[idx] <- ov_marker
    idx <- match(names(ov_marker), rownames(pred))
    idx <- idx[!is.na(idx)]
    rownames(pred)[idx] <- ov_marker

    # remove scaling '1' for markers (if any)
    if (length(lv_marker_idx) > 0L) {
      pred[lv_marker_idx, ] <- 0
    }

    # remove 'empty' rows
    empty_idx <- which(apply(pred, 1L, function(x) all(x == 0)))
    if (length(empty_idx) > 0L) {
      pred <- pred[-empty_idx, , drop = FALSE]
      pred_orig <- pred_orig[-empty_idx, , drop = FALSE]
    }

    # prepare list
    eqs <- vector("list", length = nrow(pred))
    for (j in seq_along(eqs)) {
      lhs <- rownames(pred)[j]
      x_idx <- which(pred[j, ] != 0)
      rhs <- colnames(pred)[x_idx]
      ptint <- integer(0L)
      if (lavmodel@meanstructure && pred[j, 1] != 0) {
        ptint <- unname(pred_orig[j, 1])
        rhs <- rhs[-1]
        x_idx <- x_idx[-1]
      }

      # instruments needed?
      iv_flag <- TRUE
      if (lavmodel@meanstructure && length(x_idx) == 0L) {
        iv_flag <- FALSE
        miiv <- "1"
      } else if (all(cov_u_y[lhs, rhs] == 0)) {
        iv_flag <- FALSE
        miiv <- rhs
      } else {
        # correlated with at least one predictor
        correlated_with_pred <- apply(
          this_sigma[, rhs, drop = FALSE], 1L,
          function(x) any(x != 0)
        )
        # instruments should be uncorrelated with the  term of the eq (u_j)
        uncorrelated_with_u <- cov_u_y[lhs, ] == 0
        miiv <- ov_names[correlated_with_pred & uncorrelated_with_u]
      }

      # prepare rhs, rhs_new, pt
      if (length(x_idx) == 0) {
        rhs_new <- "1"
        rhs <- "1"
        pt <- integer(0L)
      } else {
        rhs_new <- colnames(pred[j, x_idx, drop = FALSE])
        rhs <- colnames(pred_orig[j, x_idx, drop = FALSE])
        pt <- as.integer(unname(pred_orig[j, x_idx]))
      }

      iv_type <- "miiv"
      if (identical(rhs_new, miiv)) {
        iv_type <- "ols"
      }

      eqs[[j]] <- list(
        lhs_new = rownames(pred)[j],
        rhs_new = rhs_new,
        lhs = rownames(pred_orig)[j],
        rhs = rhs,
        pt = pt,
        ptint = ptint,
        iv_flag = iv_flag,
        iv_type = iv_type,
        iv = miiv,
        miiv = miiv
      )
    }

    # reorder so that ov lhs come first, then lv lhs
    lhs <- sapply(eqs, "[[", "lhs")
    ov_idx <- match(ov_names[ov_names %in% lhs], lhs)
    lv_idx <- seq_along(lhs)[-ov_idx]
    eqs <- eqs[c(ov_idx, lv_idx)]

    iv_list[[b]] <- eqs
  }

  iv_list
}

# and matrix-representation specific functions:
# - lav_model_sigma
# - lav_model_mu
# - derivative.F

# initial version: YR 2011-01-21: LISREL stuff
# updates:          YR 2011-12-01: group specific extraction
#                   YR 2012-05-17: thresholds
#                   YR 2021-10-04: rename representation.LISREL -> lav_lisrel

lav_lisrel <- function(lavpartable = NULL,
                       target = NULL,
                       extra = FALSE,
                       allow_composites = TRUE,
                       remove_nonexisting = TRUE) {
  # prepare target list
  if (is.null(target)) target <- lavpartable

  stopifnot(!is.null(target$block))

  # prepare output
  n <- length(target$lhs)
  tmp_mat <- character(n)
  tmp_row <- integer(n)
  tmp_col <- integer(n)

  # global settings
  meanstructure <- any(lavpartable$op == "~1")
  # categorical <- any(lavpartable$op == "|")
  composites <- any(lavpartable$op == "<~") && allow_composites
  # group_w_free <- any(lavpartable$lhs == "group" & lavpartable$op == "%")

  # gamma? only if conditional.x
  if (any(lavpartable$op %in% c("~", "<~") & lavpartable$exo == 1L) &&
    !composites) {
    gamma <- TRUE
  } else {
    gamma <- FALSE
  }

  # number of blocks
  nblocks <- lav_pt_nblocks(lavpartable)

  # multilevel?
  nlevels <- lav_pt_nlevels(lavpartable)
  ngroups <- lav_pt_ngroups(lavpartable)

  ov_dummy_names_nox <- vector("list", nblocks)
  ov_dummy_names_x <- vector("list", nblocks)
  if (extra) {
    rep_mm_names <- vector("list", nblocks)
    rep_mm_number <- vector("list", nblocks)
    rep_mm_rows <- vector("list", nblocks)
    rep_mm_cols <- vector("list", nblocks)
    rep_mm_dim_names <- vector("list", nblocks)
    rep_mm_symmetric <- vector("list", nblocks)
  }

  for (g in 1:nblocks) {
    # info from user model per block
    if (gamma) {
      ov_names <- lav_pt_vnames(lavpartable, "ov.nox", block = g)
    } else {
      ov_names <- lav_pt_vnames(lavpartable, "ov", block = g)
    }
    nvar <- length(ov_names)
    lv_names <- lav_pt_vnames(lavpartable, "lv", block = g)
    nfac <- length(lv_names)
    ov_th <- lav_pt_vnames(lavpartable, "th", block = g)
    nth <- length(ov_th)
    ov_names_x <- lav_pt_vnames(lavpartable, "ov.x", block = g)
    nexo <- length(ov_names_x)
    ov_names_nox <- lav_pt_vnames(lavpartable, "ov.nox", block = g)

    # in this representation, we need to create 'phantom/dummy' latent
    # variables for all `x' and `y' variables not in lv.names
    # (only y if conditional.x = TRUE)

    # regression dummys
    if (gamma) {
      tmp_names <-
        unique(lavpartable$lhs[(lavpartable$op == "~" |
          lavpartable$op == "<~") &
          lavpartable$block == g])
      # new in 0.6-12: fix for multilevel + conditional.x: split ov.x
      # are removed from ov.x
      if (nlevels > 1L) {
        if (ngroups == 1L) {
          other_block_names <- lav_pt_vnames(lavpartable, "ov",
            block = seq_len(nblocks)[-g]
          )
        } else {
          # TEST ME
          this_group <- ceiling(g / nlevels)
          blocks_within_group <- (this_group - 1L) * nlevels + seq_len(nlevels)
          other_block_names <- lav_pt_vnames(lavpartable,
            "ov",
            block = blocks_within_group[-g]
          )
        }
        if (length(ov_names_x) > 0L) {
          idx <- which(ov_names_x %in% other_block_names)
          if (length(idx) > 0L) {
            tmp_names <- unique(c(tmp_names, ov_names_x[idx]))
            ov_names_nox <- unique(c(ov_names_nox, ov_names_x[idx]))
            ov_names_x <- ov_names_x[-idx]
            nexo <- length(ov_names_x)
            ov_names <- ov_names_nox
            nvar <- length(ov_names)
          }
        }
      }
    } else {
      if (composites) {
        tmp_names <-
          unique(c(
            lavpartable$lhs[(lavpartable$op == "~") &
              lavpartable$block == g],
            lavpartable$rhs[(lavpartable$op == "~") &
              lavpartable$block == g]
          ))
      } else {
        # old behavior < 0.6-20
        tmp_names <-
          unique(c(
            lavpartable$lhs[(lavpartable$op == "~" |
              lavpartable$op == "<~") &
              lavpartable$block == g],
            lavpartable$rhs[(lavpartable$op == "~" |
              lavpartable$op == "<~") &
              lavpartable$block == g]
          ))
      }
    }
    dummy_names1 <- tmp_names[!tmp_names %in% lv_names]
    # covariances involving dummys
    dummy_cov_idx <- which(lavpartable$op == "~~" & lavpartable$block == g &
      (lavpartable$lhs %in% dummy_names1 |
        lavpartable$rhs %in% dummy_names1))
    # new in 0.5-21: also include covariances involving these covariances...
    dummy_cov_idx1 <- which(lavpartable$op == "~~" & lavpartable$block == g &
      (lavpartable$lhs %in% lavpartable$lhs[dummy_cov_idx] |
        lavpartable$rhs %in% lavpartable$rhs[dummy_cov_idx]))
    dummy_cov_idx <- unique(c(dummy_cov_idx, dummy_cov_idx1))

    dummy_names2 <- unique(c(
      lavpartable$lhs[dummy_cov_idx],
      lavpartable$rhs[dummy_cov_idx]
    ))


    # new in 0.6-7: ~~ between latent and observed
    dummy_cov_ov_lv_idx1 <- which(lavpartable$op == "~~" &
      lavpartable$block == g &
      lavpartable$lhs %in% ov_names &
      lavpartable$rhs %in% lv_names)
    dummy_cov_ov_lv_idx2 <- which(lavpartable$op == "~~" &
      lavpartable$block == g &
      lavpartable$lhs %in% lv_names &
      lavpartable$rhs %in% ov_names)
    dummy_names3 <- unique(c(
      lavpartable$lhs[dummy_cov_ov_lv_idx1],
      lavpartable$rhs[dummy_cov_ov_lv_idx2]
    ))

    # new in 0.6-10: ~~ between observed and observed, but not in ~
    # dummy_orphan_idx <- which(lavpartable$op == "~~" &
    #   lavpartable$block == g &
    #   lavpartable$lhs %in% ov_names &
    #   lavpartable$rhs %in% ov_names &
    #   (!lavpartable$lhs %in% c(
    #     dummy_names1,
    #     dummy_names2
    #   ) |
    #     !lavpartable$rhs %in% c(
    #       dummy_names1,
    #       dummy_names2
    #     )))

    # collect all dummy variables
    dummy_names <- unique(c(dummy_names1, dummy_names2, dummy_names3))


    if (length(dummy_names)) {
      # make sure order is the same as ov.names
      ov_dummy_names_nox[[g]] <-
        ov_names_nox[ov_names_nox %in% dummy_names]
      ov_dummy_names_x[[g]] <-
        ov_names_x[ov_names_x %in% dummy_names]

      # combine them, make sure order is identical to ov.names
      tmp <- ov_names[ov_names %in% dummy_names]

      # same for ov.names.x (if they are not in ov.names) (conditional.x)
      if (length(ov_names_x) > 0L) {
        tmp_x <- ov_names_x[ov_names_x %in% dummy_names]
        tmp <- unique(c(tmp, tmp_x))
      }

      # extend lv.names
      lv_names <- c(lv_names, tmp)
      nfac <- length(lv_names)

      # add 'dummy' =~ entries
      # dummy_mat <- rep("lambda", length(dummy_names))
    } else {
      ov_dummy_names_nox[[g]] <- character(0)
      ov_dummy_names_x[[g]] <- character(0)
    }

    # 1a. "=~" regular indicators
    idx <- which(target$block == g &
      target$op == "=~" & !(target$rhs %in% lv_names))
    tmp_mat[idx] <- "lambda"
    tmp_row[idx] <- match(target$rhs[idx], ov_names)
    tmp_col[idx] <- match(target$lhs[idx], lv_names)

    # 1b. "=~" regular higher-order lv indicators
    idx <- which(target$block == g &
      target$op == "=~" & !(target$rhs %in% ov_names))
    tmp_mat[idx] <- "beta"
    tmp_row[idx] <- match(target$rhs[idx], lv_names)
    tmp_col[idx] <- match(target$lhs[idx], lv_names)

    # 1c. "=~" indicators that are both in ov and lv
    idx <- which(target$block == g &
      target$op == "=~" & target$rhs %in% ov_names &
      target$rhs %in% lv_names)
    tmp_mat[idx] <- "beta"
    tmp_row[idx] <- match(target$rhs[idx], lv_names)
    tmp_col[idx] <- match(target$lhs[idx], lv_names)

    # 1d. "<~" indicators
    if (composites) {
      idx <- which(target$block == g &
        target$op == "<~" & !(target$rhs %in% lv_names))
      tmp_mat[idx] <- "wmat"
      tmp_row[idx] <- match(target$rhs[idx], ov_names)
      tmp_col[idx] <- match(target$lhs[idx], lv_names)
    }

    # 2. "~" regressions
    if (gamma) {
      # gamma
      if (composites) {
        idx <- which(target$rhs %in% ov_names_x &
          target$block == g & target$op == "~")
      } else {
        idx <- which(target$rhs %in% ov_names_x &
          target$block == g & (target$op == "~" |
          target$op == "<~"))
      }
      tmp_mat[idx] <- "gamma"
      tmp_row[idx] <- match(target$lhs[idx], lv_names)
      tmp_col[idx] <- match(target$rhs[idx], ov_names_x)

      # beta
      if (composites) {
        idx <- which(!target$rhs %in% ov_names_x &
          target$block == g & target$op == "~")
      } else {
        idx <- which(!target$rhs %in% ov_names_x &
          target$block == g & (target$op == "~" |
          target$op == "<~"))
      }
      tmp_mat[idx] <- "beta"
      tmp_row[idx] <- match(target$lhs[idx], lv_names)
      tmp_col[idx] <- match(target$rhs[idx], lv_names)
    } else {
      if (composites) {
        idx <- which(target$block == g & target$op == "~")
      } else {
        idx <- which(target$block == g & (target$op == "~" |
          target$op == "<~"))
      }
      tmp_mat[idx] <- "beta"
      tmp_row[idx] <- match(target$lhs[idx], lv_names)
      tmp_col[idx] <- match(target$rhs[idx], lv_names)
    }

    # 3a. "~~" ov
    idx <- which(target$block == g &
      target$op == "~~" & !(target$lhs %in% lv_names))
    tmp_mat[idx] <- "theta"
    tmp_row[idx] <- match(target$lhs[idx], ov_names)
    tmp_col[idx] <- match(target$rhs[idx], ov_names)

    # 3aa. "~~" ov.x
    if (gamma) {
      idx <- which(target$block == g &
        target$op == "~~" & (target$lhs %in% ov_names_x))
      tmp_mat[idx] <- "cov.x"
      tmp_row[idx] <- match(target$lhs[idx], ov_names_x)
      tmp_col[idx] <- match(target$rhs[idx], ov_names_x)
    }

    # 3b. "~~" lv
    idx <- which(target$block == g &
      target$op == "~~" & target$rhs %in% lv_names)
    tmp_mat[idx] <- "psi"
    tmp_row[idx] <- match(target$lhs[idx], lv_names)
    tmp_col[idx] <- match(target$rhs[idx], lv_names)

    # 4a. "~1" ov
    idx <- which(target$block == g &
      target$op == "~1" & !(target$lhs %in% lv_names))
    tmp_mat[idx] <- "nu"
    tmp_row[idx] <- match(target$lhs[idx], ov_names)
    tmp_col[idx] <- 1L

    # 4aa, "~1" ov.x
    if (gamma) {
      idx <- which(target$block == g &
        target$op == "~1" & (target$lhs %in% ov_names_x))
      tmp_mat[idx] <- "mean.x"
      tmp_row[idx] <- match(target$lhs[idx], ov_names_x)
      tmp_col[idx] <- 1L
    }

    # 4b. "~1" lv
    idx <- which(target$block == g &
      target$op == "~1" & target$lhs %in% lv_names)
    tmp_mat[idx] <- "alpha"
    tmp_row[idx] <- match(target$lhs[idx], lv_names)
    tmp_col[idx] <- 1L

    # 5. "|" th
    label <- paste(target$lhs, target$op, target$rhs, sep = "")
    idx <- which(target$block == g &
      target$op == "|" & label %in% ov_th)
    th <- paste(target$lhs[idx], "|", target$rhs[idx], sep = "")
    tmp_mat[idx] <- "tau"
    tmp_row[idx] <- match(th, ov_th)
    tmp_col[idx] <- 1L

    # 6. "~*~" scales
    idx <- which(target$block == g &
      target$op == "~*~")
    tmp_mat[idx] <- "delta"
    tmp_row[idx] <- match(target$lhs[idx], ov_names)
    tmp_col[idx] <- 1L

    # new 0.5-12: catch lower-elements in theta/psi
    idx_lower <- which(tmp_mat %in% c("theta", "psi") & tmp_row > tmp_col)
    if (length(idx_lower) > 0L) {
      tmp <- tmp_row[idx_lower]
      tmp_row[idx_lower] <- tmp_col[idx_lower]
      tmp_col[idx_lower] <- tmp
    }

    # new 0.5-16: group weights
    idx <- which(target$block == g & target$lhs == "group" &
      target$op == "%")
    tmp_mat[idx] <- "gw"
    tmp_row[idx] <- 1L
    tmp_col[idx] <- 1L

    # new in 0.7-1: instruments
    idx <- which(target$block == g & target$lhs == "group" &
      target$op == "|~")
    tmp_mat[idx] <- "miiv"
    tmp_row[idx] <- 0L
    tmp_col[idx] <- 0L

    if (extra) {
      # mRows
      mm_rows <- list(
        tau = nth,
        delta = nvar,
        nu = nvar,
        lambda = nvar,
        wmat = nvar,
        theta = nvar,
        alpha = nfac,
        beta = nfac,
        gamma = nfac,
        cov.x = nexo,
        mean.x = nexo,
        gw = 1L,
        psi = nfac
      )

      # mCols
      mm_cols <- list(
        tau = 1L,
        delta = 1L,
        nu = 1L,
        lambda = nfac,
        wmat = nfac,
        theta = nvar,
        alpha = 1L,
        beta = nfac,
        gamma = nexo,
        cov.x = nexo,
        mean.x = 1L,
        gw = 1L,
        psi = nfac
      )

      # dimNames for LISREL model matrices
      mm_dim_names <- list(
        tau = list(ov_th, "threshold"),
        delta = list(ov_names, "scales"),
        nu = list(ov_names, "intercept"),
        lambda = list(ov_names, lv_names),
        wmat = list(ov_names, lv_names),
        theta = list(ov_names, ov_names),
        alpha = list(lv_names, "intercept"),
        beta = list(lv_names, lv_names),
        gamma = list(lv_names, ov_names_x),
        cov.x = list(ov_names_x, ov_names_x),
        mean.x = list(ov_names_x, "intercepts"),
        gw = list("group", "weight"),
        psi = list(lv_names, lv_names)
      )

      # isSymmetric
      mm_symmetric <- list(
        tau = FALSE,
        delta = FALSE,
        nu = FALSE,
        lambda = FALSE,
        wmat = FALSE,
        theta = TRUE,
        alpha = FALSE,
        beta = FALSE,
        gamma = FALSE,
        cov.x = TRUE,
        mean.x = FALSE,
        gw = FALSE,
        psi = TRUE
      )

      # which mm's do we need? (always include lambda, theta and psi)
      # new: 0.6 this block only!!
      idx_1 <- which(target$block == g)
      if ("wmat" %in% tmp_mat[idx_1]) {
        mm_names <- c("lambda", "wmat", "theta", "psi")
      } else {
        mm_names <- c("lambda", "theta", "psi")
      }

      if ("beta" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "beta")
      }
      if (meanstructure) {
        mm_names <- c(mm_names, "nu", "alpha")
      }
      if ("tau" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "tau")
      }
      if ("delta" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "delta")
      }
      if ("gamma" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "gamma")
      }
      if ("gw" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "gw")
      }
      if ("cov.x" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "cov.x")
      }
      if ("mean.x" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "mean.x")
      }

      rep_mm_names[[g]] <- mm_names
      rep_mm_number[[g]] <- length(mm_names)
      rep_mm_rows[[g]] <- unlist(mm_rows[mm_names])
      rep_mm_cols[[g]] <- unlist(mm_cols[mm_names])
      rep_mm_dim_names[[g]] <- mm_dim_names[mm_names]
      rep_mm_symmetric[[g]] <- unlist(mm_symmetric[mm_names])
    } # extra
  } # nblocks

  rep_1 <- list(
    mat = tmp_mat,
    row = tmp_row,
    col = tmp_col
  )

  # remove non-existing (NAs)?
  # here we remove `non-existing' parameters; this depends on the matrix
  # representation (eg in LISREL rep, there is no ~~ between lv and ov)
  # if(remove.nonexisting) {
  #    idx <- which( nchar(REP$mat) > 0L &
  #                 !is.na(REP$row) & REP$row > 0L &
  #                 !is.na(REP$col) & REP$col > 0L )
  #   # but keep ==, :=, etc.
  #   idx <- c(idx, which(lavpartable$op %in% c("==", ":=", "<", ">")))
  #   REP$mat <- REP$mat[idx]
  #   REP$row <- REP$row[idx]
  #   REP$col <- REP$col[idx]
  #

  # always add 'ov.dummy.*.names' attributes
  attr(rep_1, "ov.dummy.names.nox") <- ov_dummy_names_nox
  attr(rep_1, "ov.dummy.names.x") <- ov_dummy_names_x

  if (extra) {
    attr(rep_1, "mmNames") <- rep_mm_names
    attr(rep_1, "mmNumber") <- rep_mm_number
    attr(rep_1, "mmRows") <- rep_mm_rows
    attr(rep_1, "mmCols") <- rep_mm_cols
    attr(rep_1, "mmDimNames") <- rep_mm_dim_names
    attr(rep_1, "mmSymmetric") <- rep_mm_symmetric
  }

  rep_1
}


# ETA:
# 1) EETA
# 2) EETAx
# 3) VETA
# 4) VETAx

# 1) EETA
# compute E(ETA): expected value of latent variables (marginal over x)
# - if no eXo (and GAMMA):
#     E(ETA) = (I-B)^-1 ALPHA
# - if eXo and GAMMA:
#     E(ETA) = (I-B)^-1 ALPHA + (I-B)^-1 GAMMA mean.x
lav_lisrel_eeta <- function(mlist = NULL, mean_x = NULL,
                            sample_mean = NULL,
                            ov_y_dummy_ov_idx = NULL,
                            ov_x_dummy_ov_idx = NULL,
                            ov_y_dummy_lv_idx = NULL,
                            ov_x_dummy_lv_idx = NULL) {
  mm_beta <- mlist$beta
  mm_gamma <- mlist$gamma

  # ALPHA? (reconstruct, but no 'fix')
  mm_alpha <- lav_lisrel_alpha0(
    mlist = mlist, sample_mean = sample_mean,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # BETA?
  if (!is.null(mm_beta)) {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    # GAMMA?
    if (!is.null(mm_gamma)) {
      eeta <- as.vector(ib_inv %*% mm_alpha + ib_inv %*% mm_gamma %*% mean_x)
    } else {
      eeta <- as.vector(ib_inv %*% mm_alpha)
    }
  } else {
    # GAMMA?
    if (!is.null(mm_gamma)) {
      eeta <- as.vector(mm_alpha + mm_gamma %*% mean_x)
    } else {
      eeta <- as.vector(mm_alpha)
    }
  }

  eeta
}

# 2) EETAx
# compute E(ETA|x_i): conditional expected value of latent variable,
#                     given specific value of x_i
# - if no eXo (and GAMMA):
#     E(ETA) = (I-B)^-1 ALPHA
#     we return a matrix of size [nobs x nfac] replicating E(ETA)
# - if eXo and GAMMA:
#     E(ETA|x_i) = (I-B)^-1 ALPHA + (I-B)^-1 GAMMA x_i
#     we return  a matrix of size [nobs x nfac]
#
lav_lisrel_eetax <- function(mlist = NULL, exo = NULL, n = nrow(exo),
                             sample_mean = NULL,
                             ov_y_dummy_ov_idx = NULL,
                             ov_x_dummy_ov_idx = NULL,
                             ov_y_dummy_lv_idx = NULL,
                             ov_x_dummy_lv_idx = NULL) {
  mm_lambda <- mlist$lambda
  mm_beta <- mlist$beta
  mm_gamma <- mlist$gamma
  nfac <- ncol(mm_lambda)
  # if eXo, N must be nrow(eXo)
  if (!is.null(exo)) {
    n <- nrow(exo)
  }

  # ALPHA?
  mm_alpha <- lav_lisrel_alpha0(
    mlist = mlist, sample_mean = sample_mean,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # construct [nobs x nfac] matrix (repeating ALPHA)
  eeta <- matrix(mm_alpha, n, nfac, byrow = TRUE)

  # put back eXo values if dummy
  if (length(ov_x_dummy_lv_idx) > 0L) {
    eeta[, ov_x_dummy_lv_idx] <- exo
  }

  # BETA?
  if (!is.null(mm_beta)) {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    eeta <- eeta %*% t(ib_inv)
  }

  # GAMMA?
  if (!is.null(mm_gamma)) {
    if (!is.null(mm_beta)) {
      eeta <- eeta + exo %*% t(ib_inv %*% mm_gamma)
    } else {
      eeta <- eeta + exo %*% t(mm_gamma)
    }
  }

  eeta
}

# 3) VETA
# compute V(ETA): variances/covariances of latent variables
# - if no eXo (and GAMMA)
#     V(ETA) = (I-B)^-1 PSI (I-B)^-T
# - if eXo and GAMMA: (cfr lisrel submodel 3a with ksi=x)
#     V(ETA) = (I-B)^-1 [ GAMMA  cov.x t(GAMMA) + PSI] (I-B)^-T
lav_lisrel_veta <- function(mlist = NULL) {
  # mm_lambda <- mlist$lambda
  # nvar <- nrow(mm_lambda)
  mm_psi <- mlist$psi
  # mm_theta <- mlist$theta
  mm_beta <- mlist$beta
  mm_gamma <- mlist$gamma

  if (!is.null(mm_gamma)) {
    cov_x <- mlist$cov.x
    # we treat 'x' as 'ksi' in the LISREL model; cov.x is PHI
    mm_psi <- tcrossprod(mm_gamma %*% cov_x, mm_gamma) + mm_psi
  }

  # beta?
  if (is.null(mm_beta)) {
    veta <- mm_psi
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    veta <- tcrossprod(ib_inv %*% mm_psi, ib_inv)
  }

  veta
}

# 4) VETAx
# compute V(ETA|x_i): variances/covariances of latent variables
#     V(ETA) = (I-B)^-1 PSI (I-B)^-T  + remove dummies
lav_lisrel_vetax <- function(mlist = NULL, lv_dummy_idx = NULL) {
  mm_psi <- mlist$psi
  mm_beta <- mlist$beta

  # beta?
  if (is.null(mm_beta)) {
    veta <- mm_psi
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    veta <- tcrossprod(ib_inv %*% mm_psi, ib_inv)
  }

  # remove dummy lv?
  if (!is.null(lv_dummy_idx)) {
    veta <- veta[-lv_dummy_idx, -lv_dummy_idx, drop = FALSE]
  }

  veta
}


# Y
# 1) EY
# 2) EYx
# 3) EYetax
# 4) VY
# 5) VYx
# 6) VYetax

# 1) EY
# compute E(Y): expected value of observed
# E(Y) = NU + LAMBDA %*% E(eta)
#      = NU + LAMBDA %*% (IB.inv %*% ALPHA) # no exo, no GAMMA
#      = NU + LAMBDA %*% (IB.inv %*% ALPHA + IB.inv %*% GAMMA %*% mean.x) # eXo
# if DELTA -> E(Y) = delta * E(Y)
#
# this is similar to lav_model_mu but:
# - we ALWAYS compute NU+ALPHA, even if meanstructure=FALSE
# - never used if GAMMA, since we then have categorical variables, and the
#   'part 1' structure contains the (thresholds +) intercepts, not
#   the means
lav_lisrel_ey <- function(mlist = NULL, mean_x = NULL, sample_mean = NULL,
                          ov_y_dummy_ov_idx = NULL,
                          ov_x_dummy_ov_idx = NULL,
                          ov_y_dummy_lv_idx = NULL,
                          ov_x_dummy_lv_idx = NULL, delta = TRUE) {
  mm_lambda <- mlist$lambda

  # get NU, but do not 'fix'
  mm_nu <- lav_lisrel_nu0(
    mlist = mlist, sample_mean = sample_mean,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # compute E(ETA)
  eeta <- lav_lisrel_eeta(
    mlist = mlist, sample_mean = sample_mean,
    mean_x = mean_x,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # EY
  ey <- as.vector(mm_nu) + as.vector(mm_lambda %*% eeta)

  # if delta, scale
  if (delta && !is.null(mlist$delta)) {
    ey <- ey * as.vector(mlist$delta)
  }

  ey
}

# 3) EYetax
# compute E(Y|eta_i,x_i): conditional expected value of observed variable
#                         given specific value of eta_i AND x_i
#
# E(y*_i|eta_i, x_i) = NU + LAMBDA eta_i + KAPPA x_i
#
# where eta_i = predict(fit) = factor scores OR specific values for eta_i
# (as in GH integration)
#
# if nexo = 0, and eta_i is single row, YHAT is the same for each observation
# in this case, we return a single row, unless Nobs > 1L, in which case
# we return Nobs identical rows
#
# NOTE: we assume that any effect of x_i on eta_i has already been taken
#       care off

# categorical version
lav_lisrel_eyetax <- function(mlist = NULL,
                              exo = NULL,
                              eta = NULL,
                              n = nrow(exo),
                              sample_mean = NULL,
                              ov_y_dummy_ov_idx = NULL,
                              ov_x_dummy_ov_idx = NULL,
                              ov_y_dummy_lv_idx = NULL,
                              ov_x_dummy_lv_idx = NULL,
                              delta = TRUE) {
  mm_lambda <- mlist$lambda
  mm_beta <- mlist$beta
  if (!is.null(exo)) {
    n <- nrow(exo)
  } else if (!is.null(n)) {
    # nothing to do
  } else {
    n <- 1L
  }

  # create ETA matrix
  if (nrow(eta) == 1L) {
    eta <- matrix(eta, n, ncol(eta), byrow = TRUE)
  }

  # always augment ETA with 'dummy values' (0 for ov.y, eXo for ov.x)
  # ndummy <- length(c(ov.y.dummy.lv.idx, ov.x.dummy.lv.idx))
  # if(ndummy > 0L) {
  #    ETA2 <- cbind(ETA, matrix(0, N, ndummy))
  # } else {
  eta2 <- eta
  # }

  # only if we have dummy ov.y, we need to compute the 'yhat' values
  # beforehand
  if (length(ov_y_dummy_lv_idx) > 0L) {
    # insert eXo values
    if (length(ov_x_dummy_lv_idx) > 0L) {
      eta2[, ov_x_dummy_lv_idx] <- exo
    }
    # zero ov.y values
    if (length(ov_y_dummy_lv_idx) > 0L) {
      eta2[, ov_y_dummy_lv_idx] <- 0
    }

    # ALPHA? (reconstruct, but no 'fix')
    mm_alpha <- lav_lisrel_alpha0(
      mlist = mlist, sample_mean = sample_mean,
      ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
      ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
      ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
      ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
    )
    # BETA?
    # (add mm_alpha to each row; faster than sweep(., 2L, ., "+"))
    if (!is.null(mm_beta)) {
      eta2 <- tcrossprod(eta2, mm_beta)
    }
    eta2 <- eta2 + rep(mm_alpha, each = nrow(eta2))

    # put back eXo values
    if (length(ov_x_dummy_lv_idx) > 0L) {
      eta2[, ov_x_dummy_lv_idx] <- exo
    }

    # put back ETA values for the 'real' latent variables
    dummy_idx <- c(ov_x_dummy_lv_idx, ov_y_dummy_lv_idx)
    if (length(dummy_idx) > 0L) {
      lv_regular_idx <- seq_len(min(dummy_idx) - 1L)
      eta2[, lv_regular_idx] <- eta[, lv_regular_idx, drop = FALSE]
    }
  }

  # get NU, but do not 'fix'
  mm_nu <- lav_lisrel_nu0(
    mlist = mlist,
    sample_mean = sample_mean,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # EYetax (add mm_nu to each row; faster than sweep(., 2L, ., "+"))
  eyetax <- tcrossprod(eta2, mm_lambda)
  eyetax <- eyetax + rep(mm_nu, each = nrow(eyetax))

  # if delta, scale each row by the delta diagonal
  if (delta && !is.null(mlist$delta)) {
    eyetax <- eyetax * rep(mlist$delta, each = nrow(eyetax))
  }

  eyetax
}

# unconditional version
lav_lisrel_eyetax3 <- function(mlist = NULL,
                               eta = NULL,
                               sample_mean = NULL,
                               mean_x = NULL,
                               ov_y_dummy_ov_idx = NULL,
                               ov_x_dummy_ov_idx = NULL,
                               ov_y_dummy_lv_idx = NULL,
                               ov_x_dummy_lv_idx = NULL,
                               delta = TRUE) {
  mm_lambda <- mlist$lambda

  # special case: empty lambda
  if (ncol(mm_lambda) == 0L) {
    return(matrix(sample_mean,
      nrow(eta), length(sample_mean),
      byrow = TRUE
    ))
  }

  # lv idx
  dummy_idx <- c(ov_y_dummy_lv_idx, ov_x_dummy_lv_idx)
  if (length(dummy_idx) > 0L) {
    nondummy_idx <- seq_len(min(dummy_idx) - 1L)
  } else {
    nondummy_idx <- seq_len(ncol(mlist$lambda))
  }

  # beta?
  if (is.null(mlist$beta) || length(ov_y_dummy_lv_idx) == 0L ||
    length(nondummy_idx) == 0L) {
    lambda__ib_inv <- mm_lambda
  } else {
    # only keep those columns of BETA that correspond to the
    # the `regular' latent variables
    # (ie. ignore the structural part altogether)
    mlist2 <- mlist
    mlist2$beta[, dummy_idx] <- 0
    ib_inv <- lav_lisrel_ibinv(mlist = mlist2)
    lambda__ib_inv <- mm_lambda %*% ib_inv
  }

  # compute model-implied means
  ey <- lav_lisrel_ey(
    mlist = mlist, mean_x = mean_x,
    sample_mean = sample_mean,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  eeta <- lav_lisrel_eeta(
    mlist = mlist, sample_mean = sample_mean,
    mean_x = mean_x,
    ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
    ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
    ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
    ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
  )

  # center regular lv only
  eta[, nondummy_idx] <- sweep(eta[, nondummy_idx, drop = FALSE], 2L,
    STATS = eeta[nondummy_idx], FUN = "-"
  )

  # project from lv to ov, if we have any lv
  if (length(nondummy_idx) > 0) {
    eyetax <- sweep(
      tcrossprod(
        eta[, nondummy_idx, drop = FALSE],
        lambda__ib_inv[, nondummy_idx, drop = FALSE]
      ),
      2L,
      STATS = ey, FUN = "+"
    )
  } else {
    eyetax <- eta
  }

  # put back eXo variables
  if (length(ov_x_dummy_lv_idx) > 0L) {
    eyetax[, ov_x_dummy_ov_idx] <- eta[, ov_x_dummy_lv_idx, drop = FALSE]
  }

  # if delta, scale
  if (delta && !is.null(mlist$delta)) {
    eyetax <- sweep(eyetax, 2L, STATS = mlist$delta, FUN = "*")
  }

  eyetax
}

# 4) VY
# compute the *un*conditional variance/covariance of y: V(Y) or V(Y*)
# 'unconditional' model-implied (co)variances
#  - same as Sigma.hat if all Y are continuous
#  - diagonal is 1.0 (or delta^2) if categorical
#  - if also Gamma, cov.x is used (only if conditional.x)
#    only in THIS case, VY is different from diag(VYx)
#
# V(Y) = LAMBDA V(ETA) t(LAMBDA) + THETA
lav_lisrel_vy <- function(mlist = NULL) {
  mm_lambda <- mlist$lambda
  mm_theta <- mlist$theta

  veta <- lav_lisrel_veta(mlist = mlist)
  vy <- tcrossprod(mm_lambda %*% veta, mm_lambda) + mm_theta
  vy
}

# 5) VYx
# compute V(Y*|x_i) == model-implied covariance matrix
# this equals V(Y*) if no (explicit) eXo no GAMMA
#
# in >0.6-20: special treatment for composites
#
lav_lisrel_sigma <- function(mlist = NULL, delta = TRUE) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  mm_psi <- mlist$psi
  mm_theta <- mlist$theta
  mm_beta <- mlist$beta
  mm_wmat <- mlist$wmat

  # standard: no composites
  if (is.null(mm_wmat)) {
    # beta?
    if (is.null(mm_beta)) {
      lambda__ib_inv <- mm_lambda
    } else {
      ib_inv <- lav_lisrel_ibinv(mlist = mlist)
      lambda__ib_inv <- mm_lambda %*% ib_inv
    }
    # compute V(Y*|x_i)
    vyx <- tcrossprod(lambda__ib_inv %*% mm_psi, lambda__ib_inv) + mm_theta

    # composites, or mix of composites and latent variables
  } else {
    # - first join LAMBDA and WMAT
    # - create 'T' matrix: - identity for regular lv's,
    #                      - THETA block-diagonal for composites
    # - create C_0: VETA, but zero diagonal elements for composites
    # composite indicators (rows) and composite latent variables (columns) are
    # identified from WMAT (which holds the composite weights), NOT from empty
    # rows/columns of LAMBDA: a higher-order common factor whose only indicators
    # are latent variables (e.g. a growth factor measured by composites) also
    # has an empty LAMBDA column and must not be treated as a composite.
    cov_idx <- which(rowSums(mm_wmat != 0) > 0L)
    clv_idx <- which(colSums(mm_wmat != 0) > 0L)
    # regular latent variables
    # rlv_idx <- seq_len(ncol(mm_lambda))[-clv_idx]

    # combine LAMBDA and WMAT
    lw <- mm_lambda + mm_wmat

    tmat <- diag(nrow(mm_lambda))
    tmat[cov_idx, cov_idx] <- mm_theta[cov_idx, cov_idx]
    wtw <- t(lw[, clv_idx, drop = FALSE]) %*% tmat %*%
                               lw[, clv_idx, drop = FALSE]
    wtw_inv <- solve(wtw)
    wtw_inv_1 <- diag(ncol(mm_lambda))
    wtw_inv_1[clv_idx, clv_idx] <- wtw_inv

    if (is.null(mm_beta)) {
      ib_inv <- diag(nrow(mm_psi))
    } else {
      ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    }
    veta <- ib_inv %*% mm_psi %*% t(ib_inv)
    c0 <- veta
    diag(c0)[clv_idx] <- 0

    vyx <- tmat %*% lw %*% wtw_inv_1 %*% c0 %*%
                       t(wtw_inv_1) %*% t(lw) %*% tmat + mm_theta
  }

  # if delta, scale
  if (delta && !is.null(mlist$delta)) {
    mm_delta <- diag(mlist$delta[, 1L], nrow = nvar, ncol = nvar)
    vyx <- mm_delta %*% vyx %*% mm_delta
  }

  vyx
}


### compute model-implied sample statistics
#
# 1) MuHat (similar to EY, but continuous only)
# 2) TH
# 3) PI
# 4) SigmaHat == VYx

# compute MuHat for a single block/group; only for the continuous case (no eXo)
#
# this is a special case of E(Y) where
# - we have no (explicit) eXogenous variables
# - only continuous
lav_lisrel_mu <- function(mlist = NULL) {
  mm_nu <- mlist$nu
  mm_alpha <- mlist$alpha
  mm_lambda <- mlist$lambda
  mm_beta <- mlist$beta

  # shortcut
  if (is.null(mm_alpha) || is.null(mm_nu)) {
    return(matrix(0, nrow(mm_lambda), 1L))
  }

  # beta?
  if (is.null(mm_beta)) {
    lambda__ib_inv <- mm_lambda
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    lambda__ib_inv <- mm_lambda %*% ib_inv
  }

  # compute Mu Hat
  mu_hat <- mm_nu + lambda__ib_inv %*% mm_alpha

  mu_hat
}

# compute TH for a single block/group
lav_lisrel_th <- function(mlist = NULL, th_idx = NULL, delta = TRUE) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)
  mm_beta <- mlist$beta
  mm_tau <- mlist$tau
  nth <- nrow(mm_tau)

  # missing alpha
  if (is.null(mlist$alpha)) {
    mm_alpha <- matrix(0, nfac, 1L)
  } else {
    mm_alpha <- mlist$alpha
  }

  # missing nu
  if (is.null(mlist$nu)) {
    mm_nu <- matrix(0, nvar, 1L)
  } else {
    mm_nu <- mlist$nu
  }

  if (is.null(th_idx)) {
    th_idx <- seq_len(nth)
    nlev <- rep(1L, nvar)
    k_nu <- diag(nvar)
  } else {
    nlev <- tabulate(th_idx, nbins = nvar)
    nlev[nlev == 0L] <- 1L
    k_nu <- matrix(0, sum(nlev), nvar)
    k_nu[cbind(seq_len(sum(nlev)), rep(seq_len(nvar), times = nlev))] <- 1.0
  }

  # shortcut
  if (is.null(mm_tau)) {
    return(matrix(0, length(th_idx), 1L))
  }

  # beta?
  if (is.null(mm_beta)) {
    lambda__ib_inv <- mm_lambda
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    lambda__ib_inv <- mm_lambda %*% ib_inv
  }

  # compute pi0
  pi0 <- mm_nu + lambda__ib_inv %*% mm_alpha

  # interleave th's with zeros where we have numeric variables
  th <- numeric(length(th_idx))
  th[th_idx > 0L] <- mm_tau[, 1L]

  # compute TH
  th_1 <- th - (k_nu %*% pi0)

  # if delta, scale
  if (delta && !is.null(mlist$delta)) {
    delta_diag <- mlist$delta[, 1L]
    delta_star_diag <- rep(delta_diag, times = nlev)
    th_1 <- th_1 * delta_star_diag
  }

  as.vector(th_1)
}

# compute PI for a single block/group
lav_lisrel_pi <- function(mlist = NULL, delta = TRUE) {
  mm_lambda <- mlist$lambda
  mm_beta <- mlist$beta
  mm_gamma <- mlist$gamma

  # shortcut
  if (is.null(mm_gamma)) {
    return(matrix(0, nrow(mm_lambda), 0L))
  }

  # beta?
  if (is.null(mm_beta)) {
    lambda__ib_inv <- mm_lambda
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    lambda__ib_inv <- mm_lambda %*% ib_inv
  }

  # compute PI
  pi0 <- lambda__ib_inv %*% mm_gamma

  # if delta, scale
  if (delta && !is.null(mlist$delta)) {
    delta_diag <- mlist$delta[, 1L]
    pi0 <- pi0 * delta_diag
  }

  pi0
}

lav_lisrel_lambda <- function(mlist = NULL,
                              ov_y_dummy_ov_idx = NULL,
                              ov_x_dummy_ov_idx = NULL,
                              ov_y_dummy_lv_idx = NULL,
                              ov_x_dummy_lv_idx = NULL,
                              remove_dummy_lv = FALSE,
                              use_wmat = FALSE) {
  lv_dummy_idx <- c(ov_y_dummy_lv_idx, ov_x_dummy_lv_idx)

  # get lambda
  mm_lambda <- mlist$lambda

  # should we transform lambda for composites?
  if (use_wmat && !is.null(mlist$wmat) && !is.null(mlist$theta)) {
    mm_wmat <- mlist$wmat
    mm_theta <- mlist$theta

    mm_lambda <- (
      mm_lambda + mm_theta %*% mm_wmat %*%
      MASS::ginv(t(mm_wmat) %*% mm_theta %*% mm_wmat)
    )
  }

  # dummy ovs?
  if (length(ov_y_dummy_ov_idx) > 0L && !is.null(mlist$beta)) {
    mm_lambda[ov_y_dummy_ov_idx, ] <- mlist$beta[ov_y_dummy_lv_idx, ]
  }

  # remove dummy lv?
  if (remove_dummy_lv && length(lv_dummy_idx) > 0L) {
    mm_lambda <- mm_lambda[, -lv_dummy_idx, drop = FALSE]
  }

  mm_lambda
}

lav_lisrel_theta <- function(mlist = NULL,
                             ov_y_dummy_ov_idx = NULL,
                             ov_x_dummy_ov_idx = NULL,
                             ov_y_dummy_lv_idx = NULL,
                             ov_x_dummy_lv_idx = NULL,
                             use_wmat = FALSE) {
  ov_dummy_idx <- c(ov_y_dummy_ov_idx, ov_x_dummy_ov_idx)
  lv_dummy_idx <- c(ov_y_dummy_lv_idx, ov_x_dummy_lv_idx)

  # get theta
  mm_theta <- mlist$theta

  # should we transform theta for composites?
  if (use_wmat && !is.null(mlist$wmat) && !is.null(mlist$lambda)) {
    mm_wmat <- mlist$wmat
    mm_lambda <- mlist$lambda

    mm_lambda <- (
      mm_lambda + mm_theta %*% mm_wmat %*%
      MASS::ginv(t(mm_wmat) %*% mm_theta %*% mm_wmat)
    )

    mm_theta <- (
      mm_theta -
      mm_lambda %*% t(mm_wmat) %*% mm_theta %*% mm_wmat %*% t(mm_lambda)
    )
  }

  # dummy ovs?
  if (length(ov_dummy_idx) > 0L) {
    mm_theta[ov_dummy_idx, ov_dummy_idx] <-
      mlist$psi[lv_dummy_idx, lv_dummy_idx]
  }

  mm_theta
}

# compute (I - BETA)^{-1}
# new in 0.7-1: check structure of BETA
#  1. BETA absent / all-zero        -> identity
#  1b.BETA is complex               -> general solve()
#  2. BETA strictly lower triangular-> forwardsolve
#  3. BETA strictly upper triangular-> backsolve
#  4. BETA is a DAG in any order    -> Kahn topological sort, permute to
#                                      lower tri, forwardsolve, unpermute
#  5. BETA has directed cycles      -> general solve()
lav_lisrel_ibinv <- function(mlist = NULL) {
  mm_beta <- mlist$beta
  nr   <- nrow(mlist$psi)

  # case 1: no BETA
  if (is.null(mm_beta)) {
    return(diag(nr))
  }

  # the diagonal of BETA is structurally zero in the (I - BETA)^-1 formula;
  # some representations (e.g. composites with composites = FALSE) leave a
  # non-zero value on the diagonal. We force it to zero here, so that the
  # diagonal of (I - BETA) is always one. This matches the historical
  # solve()-based implementation (tmp <- -BETA; diag(tmp) <- 1; solve(tmp))
  # and keeps the triangular shortcuts below non-singular.
  diag(mm_beta) <- 0

  # forwardsolve/backsolve do not support complex values; solve() does.
  # We also route to the generic solve() path when BETA contains missing
  # values (e.g. NA while a model is being filled in during post-estimation,
  # as blavaan does): solve() propagates NA/NaN silently, matching the
  # historical behavior, whereas the logical shortcuts below would trip with
  # "missing value where TRUE/FALSE needed".
  if (is.complex(mm_beta) || anyNA(mm_beta)) {
    tmp <- -mm_beta
    diag(tmp) <- 1
    return(solve(tmp))
  }

  # BETA is identically zero
  if (all(mm_beta == 0)) {
    return(diag(nr))
  }

  # case 2: strictly lower triangular
  if (all(mm_beta[upper.tri(mm_beta)] == 0)) {
    return(forwardsolve(diag(nr) - mm_beta, diag(nr)))
  }

  # case 3: strictly upper triangular
  if (all(mm_beta[lower.tri(mm_beta)] == 0)) {
    return(backsolve(diag(nr) - mm_beta, diag(nr)))
  }

  # case 4: recursive/DAG
  # BETA[i,j] != 0  <=>  directed edge j -> i
  indegree <- rowSums(mm_beta != 0)
  result <- integer(nr)
  k <- 0L
  queue <- which(indegree == 0)
  while (length(queue) > 0L) {
    v <- queue[1L]
    queue <- queue[-1L]
    k <- k + 1L
    result[k] <- v
    for (u in which(mm_beta[, v] != 0)) {
      indegree[u] <- indegree[u] - 1L
      if (indegree[u] == 0L) queue <- c(queue, u)
    }
  }
  if (k == nr) {
    # recursive model: permute B to strictly lower triangular, solve, unpermute
    ib_inv <- forwardsolve(diag(nr) - mm_beta[result, result], diag(nr))
    inv_order <- order(result)
    return(ib_inv[inv_order, inv_order, drop = FALSE])
  }

  # case 5: non-recursive model
  tmp <- -mm_beta
  diag(tmp) <- 1
  solve(tmp)
}

# only if ALPHA=NULL but we need it anyway
# we 'reconstruct' ALPHA here (including dummy entries), no fixing
#
# without any dummy variables, this is just the zero vector
# but if we have dummy variables, we need to fill in their values
#
#
lav_lisrel_alpha0 <- function(mlist = NULL, sample_mean = NULL,
                              ov_y_dummy_ov_idx = NULL,
                              ov_x_dummy_ov_idx = NULL,
                              ov_y_dummy_lv_idx = NULL,
                              ov_x_dummy_lv_idx = NULL) {
  if (!is.null(mlist$alpha)) {
    return(mlist$alpha)
  }

  mm_lambda <- mlist$lambda
  nfac <- ncol(mm_lambda)

  ov_dummy_idx <- c(ov_y_dummy_ov_idx, ov_x_dummy_ov_idx)
  lv_dummy_idx <- c(ov_y_dummy_lv_idx, ov_x_dummy_lv_idx)

  if (length(ov_dummy_idx) > 0L) {
    mm_alpha <- matrix(0, nfac, 1L)
    # Note: instead of sample_mean, we need 'intercepts'
    # sample_mean = NU + LAMBDA..IB.inv %*% ALPHA
    # so,
    # solve(LAMBDA..IB.inv) %*% (sample_mean - NU) = ALPHA
    # where
    # - LAMBDA..IB.inv only contains 'dummy' variables, and is square
    # - NU elements are not needed (since not in ov.dummy.idx)
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    lambda__ib_inv <- mm_lambda %*% ib_inv
    lambda__ib_inv_dummy <- lambda__ib_inv[ov_dummy_idx, lv_dummy_idx, drop = FALSE]
    mm_alpha[lv_dummy_idx] <-
      solve(lambda__ib_inv_dummy, sample_mean[ov_dummy_idx])
  } else {
    mm_alpha <- matrix(0, nfac, 1L)
  }

  mm_alpha
}

# only if NU=NULL but we need it anyway
#
# since we have no meanstructure, we can assume NU is unrestricted
# and contains either:
#     1) the sample means (if not eXo)
#     2) the intercepts, if we have exogenous covariates
#        since sample_mean = NU + LAMBDA %*% E(eta)
#        we have NU = sample_mean - LAMBDA %*% E(eta)
lav_lisrel_nu0 <- function(mlist = NULL, sample_mean = NULL,
                           ov_y_dummy_ov_idx = NULL,
                           ov_x_dummy_ov_idx = NULL,
                           ov_y_dummy_lv_idx = NULL,
                           ov_x_dummy_lv_idx = NULL) {
  if (!is.null(mlist$nu)) {
    return(mlist$nu)
  }

  # if nexo > 0, subtract lambda %*% EETA
  if (length(ov_x_dummy_ov_idx) > 0L) {
    eeta <- lav_lisrel_eeta(mlist,
      mean_x = NULL,
      sample_mean = sample_mean,
      ov_y_dummy_ov_idx = ov_y_dummy_ov_idx,
      ov_x_dummy_ov_idx = ov_x_dummy_ov_idx,
      ov_y_dummy_lv_idx = ov_y_dummy_lv_idx,
      ov_x_dummy_lv_idx = ov_x_dummy_lv_idx
    )

    # 'regress' NU on X
    mm_nu <- sample_mean - mlist$lambda %*% eeta

    # just to make sure we have exact zeroes for all dummies
    mm_nu[c(ov_y_dummy_ov_idx, ov_x_dummy_ov_idx)] <- 0
  } else {
    # unrestricted mean
    mm_nu <- sample_mean
  }

  mm_nu
}

# set (total/residual) variances of composites
# and while we at it, also set intercepts of composites
lav_lisrel_comp_set_intresvar <- function(mlist = NULL,
                                          tol = .Machine$double.eps,
                                          debug = FALSE) {
  mm_lambda <- mlist$lambda
  mm_beta <- mlist$beta
  mm_psi <- mlist$psi
  mm_wmat <- mlist$wmat
  # mm_theta <- mlist$theta

  # nothing to do if this block has no composites (e.g. in a multilevel or
  # multigroup model where composites appear in some blocks only)
  if (is.null(mm_wmat) || all(mm_wmat == 0)) {
    return(mlist)
  }

  # std.lv or not?
  marker_idx <- lav_utils_get_marker(mlist$wmat)
  std_lv <- FALSE
  if (all(is.na(marker_idx))) {
    std_lv <- TRUE
  }

  # housekeeping (identify composites from WMAT, not from empty LAMBDA rows/cols;
  # see lav_lisrel_sigma() for why a higher-order factor would otherwise be
  # mistaken for a composite)
  ovc_idx <- which(rowSums(mm_wmat != 0) > 0L)
  lvc_idx <- which(colSums(mm_wmat != 0) > 0L)
  lvc_flag <- logical(nrow(mm_psi))
  lvc_flag[lvc_idx] <- TRUE
  tmat <- diag(nrow(mm_lambda))
  tmat[ovc_idx, ovc_idx] <- mlist$theta[ovc_idx, ovc_idx]

  if (std_lv) {
    target_psi <- rep(1, ncol(mm_wmat))
  } else {
    # total variances composites
    target_psi <- diag(t(mm_wmat) %*% tmat %*% mm_wmat)
  }
  # fill in PSI element for non-composites
  target_psi[!lvc_flag] <- diag(mm_psi)[!lvc_flag]

  # initial values (including exogenous variances)
  diag(mm_psi) <- target_psi

  # no regressions
  if (is.null(mm_beta)) {
    # store PSI
    mlist$psi <- mm_psi

    # fix intercept (if needed)
    if (!is.null(mlist$alpha)) {
      tmp <- t(mm_wmat) %*% mlist$nu
      mlist$alpha[lvc_idx, 1L] <- tmp[lvc_idx, 1L]
    }

    return(mlist)
  }

  # set (residual) variances in psi
  abs_beta <- abs(mlist$beta)
  # x_idx <- which(apply(abs_beta, 1L, sum) == 0 & lvc_flag)
  y_idx <- which(apply(abs_beta, 1L, sum) != 0 & lvc_flag)

  # compute IB.inv
  ib_inv <- lav_lisrel_ibinv(mlist)

  # check if BETA is acyclic
  if (!lav_graph_is_acyclic(mm_beta)) {
    # damn, we have a cyclic model; use nlminb()
    mm_psi <- lav_mlist_target_psi(
      ib_inv = ib_inv, mm_psi = mm_psi,
      target_psi = target_psi, y_idx = y_idx
    )
  } else {
    # for an acyclic model, we should be able to find the
    # residual analytically; simply by computing the model-based
    # total variances of the RHS of each regression, and set the
    # residual of the y variable so that it is exactly equal to unity
    # (yes, this will result in negative residual variances if needed)

    # ideally, we first sort the variables 'in topological order'; then
    # we need only one run; but here we are somewhat lazy, and we
    # use a few runs, each time setting more variables right

    # get ancestors list for each node/variable
    ancestors <- lav_graph_get_ancestors(mm_beta)

    # for each y variable, compute IB.inv %*% psi %*% t(IB.inv), without y
    ny <- length(y_idx)
    max_rep <- ny * 4
    for (rep in seq_len(max_rep)) {
      if (debug) {
        cat("rep = ", rep, "\n")
      }
      # check current diagonal
      current_diag <- diag(ib_inv %*% mm_psi %*% t(ib_inv))
      if (debug) {
        cat("target.psi   = ", target_psi[y_idx], "\n")
        cat("current.diag = ", current_diag[y_idx], "\n")
      }
      if (all(abs(current_diag[y_idx] - target_psi[y_idx]) < tol)) {
        # we are done, bail out
        break
      }
      for (i in seq_len(ny)) {
        this_y_idx <- y_idx[i]
        this_x_idx <- ancestors[[this_y_idx]]
        ib_inv_y <- ib_inv[this_y_idx, this_x_idx, drop = FALSE]
        psi_x <- mm_psi[this_x_idx, this_x_idx, drop = FALSE]
        var_y <- drop(ib_inv_y %*% psi_x %*% t(ib_inv_y))
        mm_psi[this_y_idx, this_y_idx] <- target_psi[this_y_idx] - var_y
      }
    }

    # final check?
    current_diag <- diag(ib_inv %*% mm_psi %*% t(ib_inv))
    if (debug) {
      cat("final current.diag = ", current_diag[y_idx], "\n")
    }
    # don't be too strict here
    if (any(abs(current_diag[y_idx] - target_psi[y_idx]) > sqrt(tol))) {
      # as a last resort, use optimization
      mm_psi <- lav_mlist_target_psi(
        ib_inv = ib_inv, mm_psi = mm_psi,
        target_psi = target_psi, y_idx = y_idx
      )
    }
  } # acyclic

  # store PSI
  mlist$psi <- mm_psi

  # fix composite mean (if needed)
  if (!is.null(mlist$alpha)) {
    tmp <- t(mm_wmat) %*% mlist$nu # total means
    # step 1: set exogenous alpha values
    mlist$alpha[lvc_idx, 1L] <- tmp[lvc_idx, 1L]
    # step 2: set endogenous alpha values
    exo_only <- mlist$alpha
    exo_only[y_idx] <- 0
    # compensate alpha so that total mean stays the same
    tmp2 <- ib_inv %*% exo_only
    mlist$alpha[y_idx, 1L] <- mlist$alpha[y_idx, 1L] - tmp2[y_idx, 1L]
  }

  mlist
}

# if DELTA parameterization, compute residual elements (in theta, or psi)
# - typically for (endogenous) observed *categorical* variables only
# - but could also be all (endogenous) observed variables, if correlation = TRUE
# - or all (endogenous) latent and observed variables, if ov.only = FALSE
#
# new version YR 29 Oct 2024: try harder for psi elements (but this only works
#                             for acyclic models)
#             YR 01 Nov 2024: for non-acyclic models: use optimization
lav_lisrel_residual_variances <- function(mlist = NULL,
                                          num_idx = NULL,
                                          ov_y_dummy_ov_idx = NULL,
                                          ov_y_dummy_lv_idx = NULL,
                                          ov_only = TRUE,
                                          tol = .Machine$double.eps,
                                          debug = FALSE) {
  mm_beta <- mlist$beta
  mm_psi <- mlist$psi
  if (is.null(mlist$delta)) {
    delta <- rep(1, nrow(mlist$lambda))
  } else {
    delta <- mlist$delta
  }

  # remove num.idx from ov.y.dummy.*
  if (length(num_idx) > 0L && length(ov_y_dummy_ov_idx) > 0L) {
    n_idx <- which(ov_y_dummy_ov_idx %in% num_idx)
    if (length(n_idx) > 0L) {
      ov_y_dummy_ov_idx <- ov_y_dummy_ov_idx[-n_idx]
      ov_y_dummy_lv_idx <- ov_y_dummy_lv_idx[-n_idx]
    }
  }

  # if delta, the target may not be unity, but DELTA^(-2)
  target_all <- 1 / (delta * delta) # often the unit vector
  target_psi <- rep(1, nrow(mm_psi))
  if (length(ov_y_dummy_ov_idx) > 0L) {
    target_psi[ov_y_dummy_lv_idx] <- target_all[ov_y_dummy_ov_idx]
  }

  # phase 1: set (residual) variances in psi
  if (!is.null(mm_beta) && length(ov_y_dummy_ov_idx) > 0L) {
    abs_beta <- abs(mlist$beta)
    x_idx <- which(apply(abs_beta, 1L, sum) == 0)
    if (ov_only) {
      y_idx <- ov_y_dummy_lv_idx
      # remove x.idx elements (fixed.x = FALSE)
      if (any(y_idx %in% x_idx)) {
        y_idx <- y_idx[-which(y_idx %in% x_idx)]
      }
    } else {
      y_idx <- which(apply(abs_beta, 1L, sum) != 0)
    }

    nr <- nrow(mm_beta)
    ib <- -mm_beta
    ib[lav_mat_diag_idx(nr)] <- 1
    ib_inv <- solve(ib)


    # check if BETA is acyclic
    if (!lav_graph_is_acyclic(mm_beta)) {
      # damn, we have a cyclic model; use nlminb()
      mm_psi <- lav_mlist_target_psi(
        ib_inv = ib_inv, mm_psi = mm_psi,
        target_psi = target_psi, y_idx = y_idx
      )
    } else {
      # for an acyclic model, we should be able to find the
      # residual analytically; simply by computing the model-based
      # total variances of the RHS of each regression, and set the
      # residual of the y variable so that it is exactly equal to unity
      # (yes, this will result in negative residual variances if needed)

      # ideally, we first sort the variables 'in topological order'; then
      # we need only one run; but here we are somewhat lazy, and we
      # use a few runs, each time setting more variables right

      # get ancestors list for each node/variable
      ancestors <- lav_graph_get_ancestors(mm_beta)

      # for each y variable, compute IB.inv %*% psi %*% t(IB.inv), without y
      ny <- length(y_idx)
      max_rep <- ny * 4
      for (rep in seq_len(max_rep)) {
        if (debug) {
          cat("rep = ", rep, "\n")
        }
        # check current diagonal
        current_diag <- diag(ib_inv %*% mm_psi %*% t(ib_inv))
        if (debug) {
          cat("target.psi   = ", target_psi[y_idx], "\n")
          cat("current.diag = ", current_diag[y_idx], "\n")
        }
        if (all(abs(current_diag[y_idx] - target_psi[y_idx]) < tol)) {
          # we are done, bail out
          break
        }
        for (i in seq_len(ny)) {
          this_y_idx <- y_idx[i]
          this_x_idx <- ancestors[[this_y_idx]]
          ib_inv_y <- ib_inv[this_y_idx, this_x_idx, drop = FALSE]
          psi_x <- mm_psi[this_x_idx, this_x_idx, drop = FALSE]
          var_y <- drop(ib_inv_y %*% psi_x %*% t(ib_inv_y))
          mm_psi[this_y_idx, this_y_idx] <- target_psi[this_y_idx] - var_y
        }
      }

      # final check?
      current_diag <- diag(ib_inv %*% mm_psi %*% t(ib_inv))
      if (debug) {
        cat("final current.diag = ", current_diag[y_idx], "\n")
      }
      # don't be too strict here
      if (any(abs(current_diag[y_idx] - target_psi[y_idx]) > sqrt(tol))) {
        # as a last resort, use optimization
        mm_psi <- lav_mlist_target_psi(
          ib_inv = ib_inv, mm_psi = mm_psi,
          target_psi = target_psi, y_idx = y_idx
        )
      }
    } # acyclic
  } # phase 1

  # store PSI
  mlist$psi <- mm_psi

  # phase 2: set residual variances in theta

  # force non-numeric theta elements to be zero
  if (length(num_idx) > 0L) {
    diag(mlist$theta)[-num_idx] <- 0.0
  } else {
    diag(mlist$theta) <- 0.0
  }

  sigma_hat <- lav_lisrel_sigma(mlist = mlist, delta = FALSE)
  diag_sigma <- diag(sigma_hat)
  # theta = DELTA^(-2) - diag( LAMBDA (I-B)^-1 PSI (I-B)^-T t(LAMBDA) )
  theta_diag <- target_all - diag_sigma
  not_idx <- unique(c(num_idx, ov_y_dummy_ov_idx))
  if (length(not_idx) > 0L) {
    diag(mlist$theta)[-not_idx] <- theta_diag[-not_idx]
  } else {
    diag(mlist$theta) <- theta_diag
  }

  mlist
}

# if THETA parameterization, compute delta elements
# of observed categorical variables, as a function of other model parameters
lav_lisrel_delta <- function(mlist = NULL, num_idx = NULL) {
  sigma_hat <- lav_lisrel_sigma(mlist = mlist, delta = FALSE)
  diag_sigma <- diag(sigma_hat)

  # (1/delta^2) = diag( LAMBDA (I-B)^-1 PSI (I-B)^-T t(LAMBDA) ) + THETA
  # tmp <- diag.Sigma + THETA
  tmp <- diag_sigma
  tmp[tmp < 0] <- as.numeric(NA)
  mlist$delta[, 1L] <- sqrt(1 / tmp)

  # numeric delta's stay 1.0
  if (length(num_idx) > 0L) {
    mlist$delta[num_idx] <- 1.0
  }

  mlist
}

# compute Sigma/ETA: variances/covariances of BOTH observed and latent variables
lav_lisrel_cov_both <- function(mlist = NULL, delta = TRUE) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  mm_psi <- mlist$psi
  nlat <- nrow(mm_psi)
  mm_theta <- mlist$theta
  mm_beta <- mlist$beta

  # 'extend' matrices
  lambda2 <- rbind(mm_lambda, diag(nlat))
  theta2 <- lav_mat_bdiag(mm_theta, matrix(0, nlat, nlat))


  # beta?
  if (is.null(mm_beta)) {
    lambda__ib_inv <- lambda2
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    lambda__ib_inv <- lambda2 %*% ib_inv
  }

  # compute augment COV matrix
  cov_1 <- tcrossprod(lambda__ib_inv %*% mm_psi, lambda__ib_inv) + theta2

  # if delta, scale
  if (delta && !is.null(mlist$delta)) {
    mm_delta <- diag(mlist$delta[, 1L], nrow = nvar, ncol = nvar)
    cov_1[seq_len(nvar), seq_len(nvar)] <-
      mm_delta %*% cov_1[seq_len(nvar), seq_len(nvar)] %*% mm_delta
  }


  # if GAMMA, also x part
  mm_gamma <- mlist$gamma
  if (!is.null(mm_gamma)) {
    cov_x <- mlist$cov.x
    if (is.null(mm_beta)) {
      sx <- tcrossprod(mm_gamma %*% cov_x, mm_gamma)
    } else {
      ib_inv__gamma <- ib_inv %*% mm_gamma
      sx <- tcrossprod(ib_inv__gamma %*% cov_x, ib_inv__gamma)
    }
    cov_1[(nvar + 1):(nvar + nlat), (nvar + 1):(nvar + nlat)] <-
      cov_1[(nvar + 1):(nvar + nlat), (nvar + 1):(nvar + nlat)] + sx
  }

  cov_1
}


# derivative of the objective function
lav_lisrel_df_dmlist <- function(mlist = NULL, omega = NULL, omega_mu = NULL) {
  mm_lambda <- mlist$lambda
  mm_psi <- mlist$psi
  mm_beta <- mlist$beta
  mm_alpha <- mlist$alpha
  # mm_wmat <- mlist$wmat

  lambda_deriv <- NULL
  beta_deriv <- NULL
  theta_deriv <- NULL
  psi_deriv <- NULL
  nu_deriv <- NULL
  alpha_deriv <- NULL
  group_w_deriv <- NULL
  wmat_deriv <- NULL

  # beta?
  if (is.null(mm_beta)) {
    lambda__ib_inv <- mm_lambda
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    lambda__ib_inv <- mm_lambda %*% ib_inv
  }

  # meanstructure?
  meanstructure <- FALSE
  if (!is.null(omega_mu)) meanstructure <- TRUE

  # group weight?
  group_w_free <- FALSE
  if (!is.null(mlist$gw)) group_w_free <- TRUE

  # pre-compute some values
  t_lambda__ib_inv <- t(lambda__ib_inv)
  if (!is.null(mm_beta)) {
    omega__lambda__ib_inv__psi__t_ib_inv <-                 # nolint
      (omega %*% lambda__ib_inv %*% mm_psi %*% t(ib_inv))
  } else {
    omega__lambda <- omega %*% mm_lambda
  }

  # 1. LAMBDA
  if (!is.null(mm_beta)) {
    if (meanstructure) {
      lambda_deriv <- -1.0 * (omega_mu %*% t(mm_alpha) %*% t(ib_inv) +
        omega__lambda__ib_inv__psi__t_ib_inv)
    } else {
      lambda_deriv <- -1.0 * omega__lambda__ib_inv__psi__t_ib_inv
    }
  } else {
    # no BETA
    if (meanstructure) {
      lambda_deriv <- -1.0 * (omega_mu %*% t(mm_alpha) +
        omega__lambda %*% mm_psi)
    } else {
      lambda_deriv <- -1.0 * (omega__lambda %*% mm_psi)
    }
  }

  # 2. BETA
  if (!is.null(mm_beta)) {
    if (meanstructure) {
      beta_deriv <- -1.0 * ((t(ib_inv) %*%
        (t(mm_lambda) %*% omega_mu %*% t(mm_alpha)) %*%
        t(ib_inv)) +
        (t_lambda__ib_inv %*%
          omega__lambda__ib_inv__psi__t_ib_inv))
    } else {
      beta_deriv <- -1.0 * (t_lambda__ib_inv %*%
        omega__lambda__ib_inv__psi__t_ib_inv)
    }
  }

  # 3. PSI
  psi_deriv <- -1.0 * (t_lambda__ib_inv %*% omega %*% lambda__ib_inv)
  diag(psi_deriv) <- 0.5 * diag(psi_deriv)

  # 4. THETA
  theta_deriv <- -1.0 * omega
  diag(theta_deriv) <- 0.5 * diag(theta_deriv)

  if (meanstructure) {
    # 5. NU
    nu_deriv <- -1.0 * omega_mu

    # 6. ALPHA
    alpha_deriv <- -1.0 * t(t(omega_mu) %*% lambda__ib_inv)
  }

  if (group_w_free) {
    group_w_deriv <- 0.0
  }

  list(
    lambda = lambda_deriv,
    wmat = wmat_deriv,
    beta = beta_deriv,
    theta = theta_deriv,
    psi = psi_deriv,
    nu = nu_deriv,
    alpha = alpha_deriv,
    gw = group_w_deriv
  )
}

# dSigma/dx -- per model matrix
lav_lisrel_dsigma_dx <- function(mlist = NULL,
                                 m = "lambda",
                                 # all model matrix elements, or only a few?
                                 # NOTE: for symmetric matrices,
                                 # we assume that the have full size
                                 # (nvar*nvar) (but already correct for
                                 # symmetry)
                                 idx = seq_along(mlist[[m]]),
                                 vech = TRUE,
                                 delta = TRUE) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)
  mm_psi <- mlist$psi
  mm_wmat <- mlist$wmat

  # for composites (vec version)
  compute_sigma <- function(x, mm = "wmat", mlist = NULL) {
    mlist_1 <- mlist
    if (mm %in% c("psi", "theta")) {
      mlist_1[[mm]] <- lav_mat_vech_rev(x)
    } else {
      mlist_1[[mm]][, ] <- x
    }
    lav_mat_vec(lav_lisrel_sigma(mlist_1))
  }

  composites <- FALSE
  if (!is.null(mm_wmat)) {
    composites <- TRUE
  }

  # only lower.tri part of sigma (not same order as elimination matrix?)
  v_idx <- lav_mat_vech_idx(nvar)
  pstar <- nvar * (nvar + 1) / 2

  # shortcut for gamma, nu, alpha, tau,.... : empty matrix
  if (m == "nu" || m == "alpha" || m == "tau" || m == "gamma" ||
    m == "gw" || m == "cov.x" || m == "mean.x") {
    return(matrix(0.0, nrow = pstar, ncol = length(idx)))
  }

  # Delta?
  delta_flag <- FALSE
  if (delta && !is.null(mlist$delta)) {
    mm_delta <- mlist$delta
    delta_flag <- TRUE
  } else if (m == "delta") { # modindices?
    return(matrix(0.0, nrow = pstar, ncol = length(idx)))
  }

  # beta?
  if (!is.null(mlist$ibeta.inv)) {
    ib_inv <- mlist$ibeta.inv
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
  }

  # pre
  # if(m == "lambda" || m == "beta")
  #    IK <- diag(nvar*nvar) + lav_mat_com(nvar, nvar)
  if (m == "lambda" || m == "beta") {
    l1 <- mm_lambda %*% ib_inv %*% mm_psi %*% t(ib_inv)
  }
  if (m == "beta" || m == "psi") {
    lambda__ib_inv <- mm_lambda %*% ib_inv
  }

  # here we go:
  if (m == "lambda") {
    kol_idx <- matrix(1:(nvar * nfac), nvar, nfac, byrow = TRUE)[idx]
    dx <- (l1 %x% diag(nvar))[, idx, drop = FALSE] +
      (diag(nvar) %x% l1)[, kol_idx, drop = FALSE]
  } else if (m == "beta") {
    if (composites) {
      dx <- lav_func_jacobian_complex(
        func = compute_sigma,
        x = lav_mat_vec(mlist$beta),
        mm = "beta", mlist = mlist
      )
      dx <- dx[, idx, drop = FALSE]
    } else {
      kol_idx <- matrix(1:(nfac * nfac), nfac, nfac, byrow = TRUE)[idx]
      dx <- (l1 %x% lambda__ib_inv)[, idx, drop = FALSE] +
        (lambda__ib_inv %x% l1)[, kol_idx, drop = FALSE]
      # this is not really needed (because we select idx=m.el.idx)
      # but just in case we need all elements of beta...
      dx[, which(idx %in% lav_mat_diag_idx(nfac))] <- 0.0
    }
  } else if (m == "psi") {
    if (composites) {
      tmp <- lav_func_jacobian_complex(
        func = compute_sigma,
        x = lav_mat_vech(mlist$psi),
        mm = "psi", mlist = mlist
      )
      dx <- matrix(0, nrow = nrow(tmp), ncol = length(mm_psi))
      dx[, lav_mat_vech_idx(nrow(mm_psi))] <- tmp
      dx[, lav_mat_vechu_idx(nrow(mm_psi), diagonal = FALSE)] <-
        dx[, lav_mat_vech_idx(nrow(mm_psi), diagonal = FALSE), drop = FALSE]
      dx <- dx[, idx, drop = FALSE]
    } else {
      dx <- (lambda__ib_inv %x% lambda__ib_inv)
      # symmetry correction, but keeping all duplicated elements
      # since we depend on idx=m.el.idx
      lower_idx <- lav_mat_vech_idx(nfac, diagonal = FALSE)
      upper_idx <- lav_mat_vechru_idx(nfac, diagonal = FALSE)
      offdiag_sum <- dx[, lower_idx] + dx[, upper_idx]
      dx[, c(lower_idx, upper_idx)] <- cbind(offdiag_sum, offdiag_sum)
      dx <- dx[, idx, drop = FALSE]
    }
  } else if (m == "theta") {
    # DX <- diag(nvar*nvar) # very sparse...
    dx <- matrix(0, nvar * nvar, length(idx))
    dx[cbind(idx, seq_along(idx))] <- 1
    # symmetry correction not needed, since all off-diagonal elements
    # are zero?
  } else if (m == "delta") {
    omega <- lav_lisrel_sigma(mlist, delta = FALSE)
    dd <- diag(mm_delta[, 1], nvar, nvar)
    dd_omega <- (dd %*% omega)
    a <- dd_omega %x% diag(nvar)
    m_b <- diag(nvar) %x% dd_omega
    dx <- a[, lav_mat_diag_idx(nvar), drop = FALSE] +
      m_b[, lav_mat_diag_idx(nvar), drop = FALSE]
    dx <- dx[, idx, drop = FALSE]
  } else if (m == "wmat") {
    # just a dummy to get us going
    dx <- lav_func_jacobian_complex(
      func = compute_sigma,
      x = lav_mat_vec(mm_wmat),
      mm = "wmat", mlist = mlist
    )
    dx <- dx[, idx, drop = FALSE]

    # KOL.idx <- matrix(1:(nvar * nfac), nvar, nfac, byrow = TRUE)[idx]
    # VETA <- IB.inv %*% PSI %*% t(IB.inv)
    # C0 <- VETA; diag(C0) <- 0
    # cov.idx <- which(apply(LAMBDA, 1L,
    #                        function(x) sum(x == 0) == ncol(LAMBDA)))
    # clv.idx <- which(apply(LAMBDA, 2L,
    #                        function(x) sum(x == 0) == nrow(LAMBDA)))
    # Tmat <- diag(nrow(LAMBDA))
    # Tmat[cov.idx, cov.idx] <- MLIST$theta[cov.idx, cov.idx]
    # L1 <- Tmat %*% WMAT %*% C0
    # DX <- (L1 %x% diag(nvar))[, idx, drop = FALSE] +
    #   (diag(nvar) %x% L1)[, KOL.idx, drop = FALSE]
    # DX <- DX * nfac
  } else {
    lav_msg_stop(gettext("wrong model matrix name:"), m)
  }

  if (delta_flag && !m == "delta") {
    dx <- dx * as.vector(mm_delta %x% mm_delta)
  }

  # vech?
  if (vech) {
    dx <- dx[v_idx, , drop = FALSE]
  }

  dx
}

# dMu/dx -- per model matrix
lav_lisrel_dmu_dx <- function(mlist = NULL,
                              m = "alpha",
                              # all model matrix elements, or only a few?
                              idx = seq_along(mlist[[m]])) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)
  # mm_wmat <- mlist$wmat

  # shortcut for empty matrices
  if (m == "gamma" || m == "psi" || m == "theta" || m == "tau" ||
    m == "delta" || m == "gw" || m == "cov.x" || m == "mean.x") {
    return(matrix(0.0, nrow = nvar, ncol = length(idx)))
  }

  # missing alpha
  if (is.null(mlist$alpha)) {
    mm_alpha <- matrix(0, nfac, 1L)
  } else {
    mm_alpha <- mlist$alpha
  }


  # beta?
  if (!is.null(mlist$ibeta.inv)) {
    ib_inv <- mlist$ibeta.inv
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
  }

  if (m == "nu") {
    dx <- diag(nvar)
  } else if (m == "lambda") {
    dx <- t(ib_inv %*% mm_alpha) %x% diag(nvar)
  } else if (m == "wmat") {
    # dummy, just to get us going
    dx <- t(ib_inv %*% mm_alpha) %x% diag(nvar)
  } else if (m == "beta") {
    dx <- t(ib_inv %*% mm_alpha) %x% (mm_lambda %*% ib_inv)
    # this is not really needed (because we select idx=m.el.idx)
    dx[, lav_mat_diag_idx(nfac)] <- 0.0
  } else if (m == "alpha") {
    dx <- mm_lambda %*% ib_inv
  } else {
    lav_msg_stop(gettext("wrong model matrix names:"), m)
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dTh/dx -- per model matrix
lav_lisrel_dth_dx <- function(mlist = NULL,
                              m = "tau",
                              # all model matrix elements, or only a few?
                              idx = seq_along(mlist[[m]]),
                              th_idx = NULL,
                              delta = TRUE) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)
  mm_tau <- mlist$tau
  nth <- nrow(mm_tau)

  # missing alpha
  if (is.null(mlist$alpha)) {
    mm_alpha <- matrix(0, nfac, 1L)
  } else {
    mm_alpha <- mlist$alpha
  }

  # missing nu
  if (is.null(mlist$nu)) {
    mm_nu <- matrix(0, nvar, 1L)
  } else {
    mm_nu <- mlist$nu
  }

  # Delta?
  delta_flag <- FALSE
  if (delta && !is.null(mlist$delta)) {
    mm_delta <- mlist$delta
    delta_flag <- TRUE
  }

  if (is.null(th_idx)) {
    th_idx <- seq_len(nth)
    nlev <- rep(1L, nvar)
    k_nu <- diag(nvar)
  } else {
    nlev <- tabulate(th_idx, nbins = nvar)
    nlev[nlev == 0L] <- 1L
    k_nu <- matrix(0, sum(nlev), nvar)
    k_nu[cbind(seq_len(sum(nlev)), rep(seq_len(nvar), times = nlev))] <- 1.0
  }

  # shortcut for empty matrices
  if (m == "gamma" || m == "psi" || m == "theta" || m == "gw" ||
    m == "cov.x" || m == "mean.x") {
    return(matrix(0.0, nrow = length(th_idx), ncol = length(idx)))
  }

  # beta?
  if (!is.null(mlist$ibeta.inv)) {
    ib_inv <- mlist$ibeta.inv
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
  }

  if (m == "tau") {
    dx <- matrix(0, nrow = length(th_idx), ncol = nth)
    dx[th_idx > 0L, ] <- diag(nth)
    if (delta_flag) {
      dx <- dx * as.vector(k_nu %*% mm_delta)
    }
  } else if (m == "nu") {
    dx <- (-1) * k_nu
    if (delta_flag) {
      dx <- dx * as.vector(k_nu %*% mm_delta)
    }
  } else if (m == "lambda") {
    dx <- (-1) * t(ib_inv %*% mm_alpha) %x% diag(nvar)
    dx <- k_nu %*% dx
    if (delta_flag) {
      dx <- dx * as.vector(k_nu %*% mm_delta)
    }
  } else if (m == "beta") {
    dx <- (-1) * t(ib_inv %*% mm_alpha) %x% (mm_lambda %*% ib_inv)
    # this is not really needed (because we select idx=m.el.idx)
    dx[, lav_mat_diag_idx(nfac)] <- 0.0
    dx <- k_nu %*% dx
    if (delta_flag) {
      dx <- dx * as.vector(k_nu %*% mm_delta)
    }
  } else if (m == "alpha") {
    dx <- (-1) * mm_lambda %*% ib_inv
    dx <- k_nu %*% dx
    if (delta_flag) {
      dx <- dx * as.vector(k_nu %*% mm_delta)
    }
  } else if (m == "delta") {
    dx1 <- matrix(0, nrow = length(th_idx), ncol = 1)
    dx1[th_idx > 0L, ] <- mm_tau
    dx2 <- mm_nu + mm_lambda %*% ib_inv %*% mm_alpha
    dx2 <- k_nu %*% dx2
    dx <- k_nu * as.vector(dx1 - dx2)
  } else {
    lav_msg_stop(gettext("wrong model matrix names:"), m)
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dPi/dx -- per model matrix
lav_lisrel_dpi_dx <- function(mlist = NULL,
                              m = "lambda",
                              # all model matrix elements, or only a few?
                              idx = seq_along(mlist[[m]])) {
  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)
  mm_gamma <- mlist$gamma
  nexo <- ncol(mm_gamma)

  # Delta?
  delta_flag <- FALSE
  if (!is.null(mlist$delta)) {
    delta_diag <- mlist$delta[, 1L]
    delta_flag <- TRUE
  }

  # shortcut for empty matrices
  if (m == "tau" || m == "nu" || m == "alpha" || m == "psi" ||
    m == "theta" || m == "gw" || m == "cov.x" || m == "mean.x") {
    return(matrix(0.0, nrow = nvar * nexo, ncol = length(idx)))
  }

  # beta?
  if (!is.null(mlist$ibeta.inv)) {
    ib_inv <- mlist$ibeta.inv
  } else {
    ib_inv <- lav_lisrel_ibinv(mlist = mlist)
  }

  if (m == "lambda") {
    dx <- t(ib_inv %*% mm_gamma) %x% diag(nvar)
    if (delta_flag) {
      dx <- dx * delta_diag
    }
  } else if (m == "beta") {
    dx <- t(ib_inv %*% mm_gamma) %x% (mm_lambda %*% ib_inv)
    # this is not really needed (because we select idx=m.el.idx)
    dx[, lav_mat_diag_idx(nfac)] <- 0.0
    if (delta_flag) {
      dx <- dx * delta_diag
    }
  } else if (m == "gamma") {
    dx <- diag(nexo) %x% (mm_lambda %*% ib_inv)
    if (delta_flag) {
      dx <- dx * delta_diag
    }
  } else if (m == "delta") {
    pre <- rep(1, nexo) %x% diag(nvar)
    dx <- pre * as.vector(mm_lambda %*% ib_inv %*% mm_gamma)
  } else {
    lav_msg_stop(gettext("wrong model matrix names:"), m)
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dGW/dx -- per model matrix
lav_lisrel_dgw_dx <- function(mlist = NULL,
                              m = "gw",
                              # all model matrix elements, or only a few?
                              idx = seq_along(mlist[[m]])) {
  # shortcut for empty matrices
  if (m != "gw") {
    return(matrix(0.0, nrow = 1L, ncol = length(idx)))
  } else {
    # m == "gw"
    dx <- matrix(1.0, 1, 1)
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dlambda/dx -- per model matrix
lav_lisrel_dlambda_dx <- function(mlist = NULL,
                                  m = "lambda",
                                  # all model matrix elements, or only a few?
                                  idx = seq_along(mlist[[m]])) {
  mm_lambda <- mlist$lambda

  # shortcut for empty matrices
  if (m != "lambda") {
    return(matrix(0.0, nrow = length(mm_lambda), ncol = length(idx)))
  } else {
    # m == "lambda"
    dx <- diag(1, nrow = length(mm_lambda), ncol = length(mm_lambda))
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dpsi/dx -- per model matrix - FIXME!!!!!
lav_lisrel_dpsi_dx <- function(mlist = NULL,
                               m = "psi",
                               # all model matrix elements, or only a few?
                               idx = seq_along(mlist[[m]])) {
  mm_psi <- mlist$psi
  nfac <- nrow(mm_psi)
  v_idx <- lav_mat_vech_idx(nfac)

  # shortcut for empty matrices
  if (m != "psi") {
    dx <- matrix(0.0, nrow = length(mm_psi), ncol = length(idx))
    return(dx[v_idx, , drop = FALSE])
  } else {
    # m == "psi"
    dx <- diag(1, nrow = length(mm_psi), ncol = length(mm_psi))
  }

  dx <- dx[v_idx, idx, drop = FALSE]
  dx
}

# dtheta/dx -- per model matrix
lav_lisrel_dtheta_dx <- function(mlist = NULL,
                                 m = "theta",
                                 # all model matrix elements, or only a few?
                                 idx = seq_along(mlist[[m]])) {
  mm_theta <- mlist$theta
  nvar <- nrow(mm_theta)
  v_idx <- lav_mat_vech_idx(nvar)

  # shortcut for empty matrices
  if (m != "theta") {
    dx <- matrix(0.0, nrow = length(mm_theta), ncol = length(idx))
    return(dx[v_idx, , drop = FALSE])
  } else {
    # m == "theta"
    dx <- diag(1, nrow = length(mm_theta), ncol = length(mm_theta))
  }

  dx <- dx[v_idx, idx, drop = FALSE]
  dx
}


# dbeta/dx -- per model matrix
lav_lisrel_dbeta_dx <- function(mlist = NULL,
                                m = "beta",
                                # all model matrix elements, or only a few?
                                idx = seq_along(mlist[[m]])) {
  mm_beta <- mlist$beta

  # shortcut for empty matrices
  if (m != "beta") {
    return(matrix(0.0, nrow = length(mm_beta), ncol = length(idx)))
  } else {
    # m == "beta"
    dx <- diag(1, nrow = length(mm_beta), ncol = length(mm_beta))
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dgamma/dx -- per model matrix
lav_lisrel_dgamma_dx <- function(mlist = NULL,
                                 m = "gamma",
                                 # all model matrix elements, or only a few?
                                 idx = seq_along(mlist[[m]])) {
  mm_gamma <- mlist$gamma

  # shortcut for empty matrices
  if (m != "gamma") {
    return(matrix(0.0, nrow = length(mm_gamma), ncol = length(idx)))
  } else {
    # m == "gamma"
    dx <- diag(1, nrow = length(mm_gamma), ncol = length(mm_gamma))
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dnu/dx -- per model matrix
lav_lisrel_dnu_dx <- function(mlist = NULL,
                              m = "nu",
                              # all model matrix elements, or only a few?
                              idx = seq_along(mlist[[m]])) {
  mm_nu <- mlist$nu

  # shortcut for empty matrices
  if (m != "nu") {
    return(matrix(0.0, nrow = length(mm_nu), ncol = length(idx)))
  } else {
    # m == "nu"
    dx <- diag(1, nrow = length(mm_nu), ncol = length(mm_nu))
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# dtau/dx -- per model matrix
lav_lisrel_dtau_dx <- function(mlist = NULL,
                               m = "tau",
                               # all model matrix elements, or only a few?
                               idx = seq_along(mlist[[m]])) {
  mm_tau <- mlist$tau

  # shortcut for empty matrices
  if (m != "tau") {
    return(matrix(0.0, nrow = length(mm_tau), ncol = length(idx)))
  } else {
    # m == "tau"
    dx <- diag(1, nrow = length(mm_tau), ncol = length(mm_tau))
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}


# dalpha/dx -- per model matrix
lav_lisrel_dalpha_dx <- function(mlist = NULL,
                                 m = "alpha",
                                 # all model matrix elements, or only a few?
                                 idx = seq_along(mlist[[m]])) {
  mm_alpha <- mlist$alpha

  # shortcut for empty matrices
  if (m != "alpha") {
    return(matrix(0.0, nrow = length(mm_alpha), ncol = length(idx)))
  } else {
    # m == "alpha"
    dx <- diag(1, nrow = length(mm_alpha), ncol = length(mm_alpha))
  }

  dx <- dx[, idx, drop = FALSE]
  dx
}

# MLIST = NULL; meanstructure=TRUE; th=TRUE; delta=TRUE; pi=TRUE; gw=FALSE
# lav_mat_vech_idx <- lavaan:::lav_mat_vech_idx;
# lav_mat_vechru_idx <- lavaan:::lav_mat_vechru_idx
# vec <- lavaan:::vec;
# lav_func_jacobian_complex <- lavaan:::lav_func_jacobian_complex
# lav_lisrel_sigma <- lavaan:::lav_lisrel_sigma
# lav_lisrel_delta <- lavaan:::lav_lisrel_delta
lav_lisrel_test_derivatives <- function(mlist = NULL,
                                        nvar = NULL, nfac = NULL, nexo = NULL,
                                        th_idx = NULL, num_idx = NULL,
                                        meanstructure = TRUE,
                                        th = TRUE, delta = TRUE, pi = TRUE,
                                        gw = FALSE, theta = FALSE) {
  if (is.null(mlist)) {
    # create artificial matrices, compare 'numerical' vs 'analytical'
    # derivatives
    # nvar <- 12; nfac <- 3; nexo <- 4 # this combination is special?
    if (is.null(nvar)) {
      nvar <- 20
    }
    if (is.null(nfac)) {
      nfac <- 6
    }
    if (is.null(nexo)) {
      nexo <- 5
    }
    if (is.null(num_idx)) {
      num_idx <- sort(sample(seq_len(nvar), ceiling(nvar / 2)))
    }
    if (is.null(th_idx)) {
      th_idx <- integer(0L)
      for (i in seq_len(nvar)) {
        if (i %in% num_idx) {
          th_idx <- c(th_idx, 0)
        } else {
          th_idx <- c(th_idx, rep(i, sample(c(1, 1, 2, 6), 1L)))
        }
      }
    }
    nth <- sum(th_idx > 0L)

    mlist <- list()
    mlist$lambda <- matrix(0, nvar, nfac)
    mlist$beta <- matrix(0, nfac, nfac)
    mlist$theta <- matrix(0, nvar, nvar)
    mlist$psi <- matrix(0, nfac, nfac)
    if (meanstructure) {
      mlist$alpha <- matrix(0, nfac, 1L)
      mlist$nu <- matrix(0, nvar, 1L)
    }
    if (th) mlist$tau <- matrix(0, nth, 1L)
    if (delta) mlist$delta <- matrix(0, nvar, 1L)
    mlist$gamma <- matrix(0, nfac, nexo)
    if (gw) mlist$gw <- matrix(0, 1L, 1L)

    # feed random numbers
    mlist <- lapply(mlist, function(x) {
      x[, ] <- rnorm(length(x))
      x
    })
    # fix
    diag(mlist$beta) <- 0.0
    diag(mlist$theta) <- diag(mlist$theta) * diag(mlist$theta) * 10
    diag(mlist$psi) <- diag(mlist$psi) * diag(mlist$psi) * 10
    mlist$psi[lav_mat_vechru_idx(nfac)] <-
      mlist$psi[lav_mat_vech_idx(nfac)]
    mlist$theta[lav_mat_vechru_idx(nvar)] <-
      mlist$theta[lav_mat_vech_idx(nvar)]
    if (delta) mlist$delta[, ] <- abs(mlist$delta) * 10
  } else {
    nvar <- nrow(mlist$lambda)
  }

  compute_sigma <- function(x, mm = "lambda", mlist = NULL) {
    mlist_1 <- mlist
    if (mm %in% c("psi", "theta")) {
      mlist_1[[mm]] <- lav_mat_vech_rev(x)
    } else {
      mlist_1[[mm]][, ] <- x
    }
    if (theta) {
      mlist_1 <- lav_lisrel_delta(mlist = mlist_1, num_idx = num_idx)
    }
    lav_mat_vech(lav_lisrel_sigma(mlist_1))
  }

  compute_mu <- function(x, mm = "lambda", mlist = NULL) {
    mlist_1 <- mlist
    if (mm %in% c("psi", "theta")) {
      mlist_1[[mm]] <- lav_mat_vech_rev(x)
    } else {
      mlist_1[[mm]][, ] <- x
    }
    if (theta) {
      mlist_1 <- lav_lisrel_delta(mlist = mlist_1, num_idx = num_idx)
    }
    lav_lisrel_mu(mlist_1)
  }

  compute_th2 <- function(x, mm = "tau", mlist = NULL, th_idx) {
    mlist_1 <- mlist
    if (mm %in% c("psi", "theta")) {
      mlist_1[[mm]] <- lav_mat_vech_rev(x)
    } else {
      mlist_1[[mm]][, ] <- x
    }
    if (theta) {
      mlist_1 <- lav_lisrel_delta(mlist = mlist_1, num_idx = num_idx)
    }
    lav_lisrel_th(mlist_1, th_idx = th_idx)
  }

  compute_pi <- function(x, mm = "lambda", mlist = NULL) {
    mlist_1 <- mlist
    if (mm %in% c("psi", "theta")) {
      mlist_1[[mm]] <- lav_mat_vech_rev(x)
    } else {
      mlist_1[[mm]][, ] <- x
    }
    if (theta) {
      mlist_1 <- lav_lisrel_delta(mlist = mlist_1, num_idx = num_idx)
    }
    lav_lisrel_pi(mlist_1)
  }

  compute_gw <- function(x, mm = "gw", mlist = NULL) {
    mlist_1 <- mlist
    if (mm %in% c("psi", "theta")) {
      mlist_1[[mm]] <- lav_mat_vech_rev(x)
    } else {
      mlist_1[[mm]][, ] <- x
    }
    if (theta) {
      mlist_1 <- lav_lisrel_delta(mlist = mlist_1, num_idx = num_idx)
    }
    mlist_1$gw[1, 1]
  }

  # if theta, set MLIST$delta
  if (theta) {
    mlist <- lav_lisrel_delta(mlist = mlist, num_idx = num_idx)
  }

  for (mm in names(mlist)) {
    if (mm %in% c("psi", "theta")) {
      x <- lav_mat_vech(mlist[[mm]])
    } else {
      x <- lav_mat_vec(mlist[[mm]])
    }
    if (mm == "delta" && theta) next
    if (lav_debug()) {
      cat("### mm = ", mm, "\n")
    }

    # 1. sigma
    dx1 <- lav_func_jacobian_complex(func = compute_sigma,
                                      x = x, mm = mm, mlist = mlist)
    dx2 <- lav_lisrel_dsigma_dx(
      mlist = mlist, m = mm, idx = seq_along(mlist[[mm]]),
      delta = !theta
    )
    if (mm %in% c("psi", "theta")) {
      # remove duplicated columns of symmetric matrices
      idx <- lav_mat_vechru_idx(sqrt(ncol(dx2)), diagonal = FALSE)
      if (length(idx) > 0L) dx2 <- dx2[, -idx]
    }
    if (theta) {
      sigma_hat <- lav_lisrel_sigma(mlist = mlist, delta = FALSE)
      r <- lav_deriv_cov2cor(sigma_hat, num_idx = num_idx)

      dx3 <- dx2
      dx2 <- r %*% dx2
    }
    if (lav_debug()) {
      cat("[SIGMA] mm = ", sprintf("%-8s:", mm), "DX1 (numerical):\n")
      print(zapsmall(dx1))
      cat("\n")
      cat("[SIGMA] mm = ", sprintf("%-8s:", mm), "DX2 (analytical):\n")
      print(dx2)
      cat("\n")
      if (theta) {
        cat("[SIGMA] mm = ", sprintf("%-8s:", mm), "DX3 (analytical):\n")
        print(dx3)
        cat("\n")
      }
    }
    cat(
      "[SIGMA] mm = ", sprintf("%-8s:", mm), "sum delta = ",
      sprintf("%12.9f", sum(dx1 - dx2)), "  max delta = ",
      sprintf("%12.9f", max(dx1 - dx2)), "\n"
    )

    # 2. mu
    dx1 <- lav_func_jacobian_complex(func = compute_mu,
                                    x = x, mm = mm, mlist = mlist)
    dx2 <- lav_lisrel_dmu_dx(
      mlist = mlist,
      m = mm, idx = seq_along(mlist[[mm]])
    )
    if (mm %in% c("psi", "theta")) {
      # remove duplicated columns of symmetric matrices
      idx <- lav_mat_vechru_idx(sqrt(ncol(dx2)), diagonal = FALSE)
      if (length(idx) > 0L) dx2 <- dx2[, -idx]
    }
    cat(
      "[MU   ] mm = ", sprintf("%-8s:", mm), "sum delta = ",
      sprintf("%12.9f", sum(dx1 - dx2)), "  max delta = ",
      sprintf("%12.9f", max(dx1 - dx2)), "\n"
    )
    if (lav_debug()) {
      cat("[MU   ] mm = ", sprintf("%-8s:", mm), "DX1 (numerical):\n")
      print(zapsmall(dx1))
      cat("\n")
      cat("[MU   ] mm = ", sprintf("%-8s:", mm), "DX2 (analytical):\n")
      print(dx2)
      cat("\n")
    }

    # 3. th
    if (th) {
      dx1 <- lav_func_jacobian_complex(
        func = compute_th2, x = x, mm = mm, mlist = mlist, th_idx = th_idx
      )
      dx2 <- lav_lisrel_dth_dx(
        mlist = mlist, m = mm, idx = seq_along(mlist[[mm]]),
        th_idx = th_idx,
        delta = TRUE
      )
      if (theta) {
        # 1. compute dDelta.dx
        dx_sigma <-
          lav_lisrel_dsigma_dx(
            m = mm, idx = seq_along(mlist[[mm]]),
            mlist = mlist, delta = !theta
          )
        var_idx <- which(!lav_mat_vech_idx(nvar) %in%
          lav_mat_vech_idx(nvar, diagonal = FALSE))
        sigma_hat <- lav_lisrel_sigma(mlist = mlist, delta = FALSE)
        dsigma <- diag(sigma_hat)
        # dy/ddsigma = -0.5/(ddsigma*sqrt(ddsigma))
        d_delta_dx <- dx_sigma[var_idx, , drop = FALSE] * -0.5 /
          (dsigma * sqrt(dsigma))

        # 2. compute dth.dDelta
        dth_d_delta <-
          lav_lisrel_dth_dx(
            mlist = mlist,
            m = "delta",
            idx = seq_along(mlist[["delta"]]),
            th_idx = th_idx
          )

        # 3. add dth.dDelta %*% dDelta.dx
        no_num_idx <- which(th_idx > 0)
        dx2[no_num_idx, ] <- dx2[no_num_idx, , drop = FALSE] +
          (dth_d_delta %*% d_delta_dx)[no_num_idx, , drop = FALSE]
        # DX2 <- DX2 + dth.dDelta %*% dDelta.dx
      }
      if (mm %in% c("psi", "theta")) {
        # remove duplicated columns of symmetric matrices
        idx <- lav_mat_vechru_idx(sqrt(ncol(dx2)), diagonal = FALSE)
        if (length(idx) > 0L) dx2 <- dx2[, -idx]
      }
      cat(
        "[TH   ] mm = ", sprintf("%-8s:", mm), "sum delta = ",
        sprintf("%12.9f", sum(dx1 - dx2)), "  max delta = ",
        sprintf("%12.9f", max(dx1 - dx2)), "\n"
      )
      if (lav_debug()) {
        cat("[TH   ] mm = ", sprintf("%-8s:", mm), "DX1 (numerical):\n")
        print(zapsmall(dx1))
        cat("\n")
        cat("[TH   ] mm = ", sprintf("%-8s:", mm), "DX2 (analytical):\n")
        print(dx2)
        cat("\n")
      }
    }

    # 4. pi
    if (pi) {
      dx1 <- lav_func_jacobian_complex(func = compute_pi,
                                         x = x, mm = mm, mlist = mlist)
      dx2 <- lav_lisrel_dpi_dx(
        mlist = mlist,
        m = mm, idx = seq_along(mlist[[mm]])
      )
      if (mm %in% c("psi", "theta")) {
        # remove duplicated columns of symmetric matrices
        idx <- lav_mat_vechru_idx(sqrt(ncol(dx2)), diagonal = FALSE)
        if (length(idx) > 0L) dx2 <- dx2[, -idx]
      }
      if (theta) {
        # 1. compute dDelta.dx
        dx_sigma <-
          lav_lisrel_dsigma_dx(
            mlist = mlist, m = mm, idx = seq_along(mlist[[mm]]),
            delta = !theta
          )
        if (mm %in% c("psi", "theta")) {
          # remove duplicated columns of symmetric matrices
          idx <- lav_mat_vechru_idx(sqrt(ncol(dx_sigma)), diagonal = FALSE)
          if (length(idx) > 0L) dx_sigma <- dx_sigma[, -idx]
        }
        var_idx <- which(!lav_mat_vech_idx(nvar) %in%
          lav_mat_vech_idx(nvar, diagonal = FALSE))
        sigma_hat <- lav_lisrel_sigma(mlist = mlist, delta = FALSE)
        dsigma <- diag(sigma_hat)
        # dy/ddsigma = -0.5/(ddsigma*sqrt(ddsigma))
        d_delta_dx <- dx_sigma[var_idx, , drop = FALSE] * -0.5 /
          (dsigma * sqrt(dsigma))

        # 2. compute dpi.dDelta
        dpi_d_delta <-
          lav_lisrel_dpi_dx(
            mlist = mlist,
            m = "delta",
            idx = seq_along(mlist[["delta"]])
          )

        # 3. add dpi.dDelta %*% dDelta.dx
        no_num_idx <- which(!seq.int(1L, nvar) %in% num_idx)
        no_num_idx <- rep(seq.int(0, nexo - 1) * nvar,
          each = length(no_num_idx)
        ) + no_num_idx
        dx2[no_num_idx, ] <- dx2[no_num_idx, , drop = FALSE] +
          (dpi_d_delta %*% d_delta_dx)[no_num_idx, , drop = FALSE]
      }
      cat(
        "[PI   ] mm = ", sprintf("%-8s:", mm), "sum delta = ",
        sprintf("%12.9f", sum(dx1 - dx2)), "  max delta = ",
        sprintf("%12.9f", max(dx1 - dx2)), "\n"
      )
      if (lav_debug()) {
        cat("[PI   ] mm = ", sprintf("%-8s:", mm), "DX1 (numerical):\n")
        print(zapsmall(dx1))
        cat("\n")
        cat("[PI   ] mm = ", sprintf("%-8s:", mm), "DX2 (analytical):\n")
        print(dx2)
        cat("\n")
      }
    }

    # 5. gw
    if (gw) {
      dx1 <- lav_func_jacobian_complex(func = compute_gw,
                                          x = x, mm = mm, mlist = mlist)
      dx2 <- lav_lisrel_dgw_dx(
        mlist = mlist,
        m = mm, idx = seq_along(mlist[[mm]])
      )
      if (mm %in% c("psi", "theta")) {
        # remove duplicated columns of symmetric matrices
        idx <- lav_mat_vechru_idx(sqrt(ncol(dx2)), diagonal = FALSE)
        if (length(idx) > 0L) dx2 <- dx2[, -idx]
      }
      cat(
        "[GW   ] mm = ", sprintf("%-8s:", mm), "sum delta = ",
        sprintf("%12.9f", sum(dx1 - dx2)), "  max delta = ",
        sprintf("%12.9f", max(dx1 - dx2)), "\n"
      )
      if (lav_debug()) {
        cat("[GW   ] mm = ", sprintf("%-8s:", mm), "DX1 (numerical):\n")
        print(dx1)
        cat("\n\n")
        cat("[GW   ] mm = ", sprintf("%-8s:", mm), "DX2 (analytical):\n")
        print(dx2)
        cat("\n\n")
      }
    }
  }

  mlist$th.idx <- th_idx
  mlist$num.idx <- num_idx

  mlist
}

# check for marker indicators:
#   - if std.lv = FALSE: a single '1' per factor, everything else zero
#   - if std.lv = TRUE: a single non-zero value per factor, everything else zero
lav_utils_get_marker <- function(mm_lambda = NULL, std_lv = FALSE) {
  mm_lambda <- as.matrix(mm_lambda)
  # nvar <- nrow(mm_lambda)
  nfac <- ncol(mm_lambda)

  # round values
  mm_lambda <- round(mm_lambda, 3L)

  marker_idx <- numeric(nfac)
  for (f in seq_len(nfac)) {
    if (std_lv) {
      marker_idx[f] <- which(rowSums(cbind(
        mm_lambda[, f] != 0,
        mm_lambda[, -f] == 0
      )) == nfac)[1]
    } else {
      marker_idx[f] <- which(rowSums(cbind(
        mm_lambda[, f] == 1,
        mm_lambda[, -f] == 0
      )) == nfac)[1]
    }
  }

  marker_idx
}

# find the residual variances (diagonal element of PSI) in such a way
# so that the diagonal elements of IB.inv %*% PSI %*% t(IB.inv) are
# equal to the elements of target.psi (usually the 1 vector)
#
# YR 01 Nov 2024: initial version; no bounds for now... (so we may end up
#                 with negative variances)
lav_mlist_target_psi <- function(mm_beta = NULL, ib_inv = NULL, mm_psi = NULL,
                                 target_psi = NULL, y_idx = NULL) {
  nr <- nrow(mm_psi)

  # IB.inv (if not given)
  if (is.null(ib_inv)) {
    ib <- -mm_beta
    ib[lav_mat_diag_idx(nr)] <- 1
    ib_inv <- solve(ib)
  }

  # target.psi
  if (is.null(target_psi)) {
    target_psi <- rep(1, nr)
  }

  # y.idx
  if (is.null(y_idx)) {
    y_idx <- seq_len(nr)
  }

  # nothing to solve for (e.g. a cyclic model where all composites are
  # exogenous); bail out before we construct a 0x0 linear system
  if (length(y_idx) == 0L) {
    return(mm_psi)
  }

  # The unknown (residual) variances diag(PSI)[y_idx] enter
  #   diag(IB.inv %*% PSI %*% t(IB.inv))[y_idx]
  # *linearly*: writing v_i = sum_{j in y} IB.inv[i,j]^2 * PSI[j,j] + known_i,
  # the requirement v_i = target_i (i in y_idx) is the square linear system
  #   A %*% PSI_yy = target[y] - known[y],   A[i,j] = IB.inv[i,j]^2.
  # Solving it directly (instead of via nlminb) is exact and fast, and -- unlike
  # nlminb -- it propagates a complex perturbation, so complex-step
  # differentiation through lav_lisrel_comp_set_intresvar() works for cyclic
  # (non-recursive) models too (avoiding 'imaginary parts discarded' warnings
  # and incorrect derivatives). This holds for any BETA, acyclic or not.
  psi0 <- mm_psi
  diag(psi0)[y_idx] <- 0
  known <- diag(ib_inv %*% psi0 %*% t(ib_inv))[y_idx]
  amat <- (ib_inv[y_idx, y_idx, drop = FALSE])^2
  bvec <- target_psi[y_idx] - known
  psi_yy <- tryCatch(solve(amat, bvec), error = function(e) {
    # singular system (e.g. an under-identified cyclic model): fall back to a
    # least-squares solution via the normal equations (real-valued only).
    drop(solve(crossprod(amat) +
               diag(sqrt(.Machine$double.eps), length(bvec)),
               crossprod(amat, bvec)))
  })

  # return updated PSI matrix
  diag(mm_psi)[y_idx] <- psi_yy
  mm_psi
}

# second-order derivatives of Sigma wrt elements of Lambda/Beta/Psi/Theta
# (elementwise)
#
# work in progress
lav_lisrel_d2sigma_lambda_psi <- function(mlist = NULL,
                                          i = 1L, j = 1L, # lambda
                                          k = 1L, l = 1L) { # psi
  # helper functions
  # elementary matrix: p x q matrix with 1 in position (i,j), 0 elsewhere
  elem_mat <- function(i, j, p, q) {
    m_e <- matrix(0, p, q)
    m_e[i, j] <- 1
    m_e
  }

  # elementary symmetric matrix for vech index (k,l) of a q x q symmetric matrix
  elem_mat_sym <- function(k, l, q) {
    m_e <- matrix(0, q, q)
    m_e[k, l] <- 1
    if (k != l) m_e[l, k] <- 1
    m_e
  }

  mm_lambda <- mlist$lambda
  nvar <- nrow(mm_lambda) # p
  nfac <- ncol(mm_lambda) # q
  ib_inv <- mlist$IB.inv
  if (is.null(ib_inv)) {
    if (is.null(mlist$beta)) {
      ib_inv <- diag(1, nrow = nfac)
    } else {
      ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    }
  }

  lib_inv <- mm_lambda %*% ib_inv

  d_lambda <- elem_mat(i, j, nvar, nfac) # p x q,  elementary matrix for Lambda
  d_psi <- elem_mat_sym(k, l, nfac) # q x q,  elementary symm matrix for Psi

  x <- d_lambda %*% ib_inv %*% d_psi %*% t(lib_inv) # p x p
  x + t(x) # p x p symmetric matrix
}

lav_lisrel_d2sigma_beta_psi <- function(mlist = NULL,
                                        i = 1L, j = 1L, # beta
                                        k = 1L, l = 1L) { # psi
  # helper functions
  # elementary matrix: p x q matrix with 1 in position (i,j), 0 elsewhere
  elem_mat <- function(i, j, p, q) {
    m_e_e <- matrix(0, p, q)
    m_e_e[i, j] <- 1
    m_e_e
  }

  # elementary symmetric matrix for vech index (k,l) of a q x q symmetric matrix
  elem_mat_sym <- function(k, l, q) {
    m_e_e <- matrix(0, q, q)
    m_e_e[k, l] <- 1
    if (k != l) m_e_e[l, k] <- 1
    m_e_e
  }

  mm_lambda <- mlist$lambda
  # nvar <- nrow(mm_lambda) # p
  nfac <- ncol(mm_lambda) # q
  ib_inv <- mlist$IB.inv
  if (is.null(ib_inv)) {
    if (is.null(mlist$beta)) {
      ib_inv <- diag(1, nrow = nfac)
    } else {
      ib_inv <- lav_lisrel_ibinv(mlist = mlist)
    }
  }
  lib_inv <- mm_lambda %*% ib_inv

  d_beta <- elem_mat(i, j, nfac, nfac) # q x q, elementary matrix for Beta
  d_psi <- elem_mat_sym(k, l, nfac) # q x q, elementary symm matrix for Psi

  x <- lib_inv %*% d_beta %*% ib_inv %*% d_psi %*% t(lib_inv) # p x p
  x + t(x) # p x p symmetric matrix
}


# Jacobian of the model-implied moments wrt the free parameters
# (single block, LISREL representation).
#
# Inputs:
#   MLIST         : named list of model matrices for the block (lambda, psi,
#                   theta, [beta], [delta], [wmat], [nu], [alpha], [tau])
#   m.free.idx    : list of length(MLIST); m.free.idx[[mm]] holds the
#                   matrix-element indices of the free elements of MLIST[[mm]]
#   x.free.idx    : list of length(MLIST); x.free.idx[[mm]] holds the
#                   parameter-vector positions of those free elements
#   nx.free       : total number of free parameters (column dimension of out)
#   meanstructure : logical
#   categorical   : logical
#   correlation   : logical (continuous correlation structure)
#   num.idx       : integer vector of numeric-variable indices (block)
#   th.idx        : integer vector of threshold indices (block)
#   group.w.free  : logical; if TRUE, prepend a row for the free group weight
#
# Output: a matrix with nx.free columns, rows are the model-implied moments
# in the order used elsewhere in lavaan:
# [group weight (if group.w.free) | thresholds | means] then [vech(Sigma)].
lav_lisrel_dimplied_dx <- function(mlist           = NULL,
                                   m_free_idx      = NULL,
                                   x_free_idx      = NULL,
                                   nx_free         = NULL,
                                   meanstructure   = FALSE,
                                   categorical     = FALSE,
                                   correlation     = FALSE,
                                   conditional_x   = FALSE,
                                   num_idx         = integer(0L),
                                   th_idx          = integer(0L),
                                   group_w_free    = FALSE,
                                   parameterization = "delta") {

  # model matrices in this block
  mnames <- names(mlist)

  delta_flag <- beta_flag <- FALSE
  if (!is.null(mlist$delta)) {
    mm_delta_idx0 <- which(mnames == "delta")
    delta_is_free <- length(m_free_idx[[mm_delta_idx0]]) > 0L
    if (any(mlist$delta[, 1] != 1) || delta_is_free) {
      delta_flag <- TRUE
    }
  }
  if (!is.null(mlist$beta)) {
    beta_flag <- TRUE
  }

  mm_lambda_idx <- which(mnames == "lambda")
  x_lambda_idx <- x_free_idx[[mm_lambda_idx]]
  m_lambda_idx <- m_free_idx[[mm_lambda_idx]]
  n_lam <- length(m_lambda_idx)

  mm_psi_idx <- which(mnames == "psi")
  tmp <- x_free_idx[[mm_psi_idx]]
  x_psi_idx <- x_psi_idx <- tmp[!duplicated(tmp)]
  m_psi_idx <- m_free_idx[[mm_psi_idx]][!duplicated(tmp)]
  n_psi <- length(m_psi_idx)


  mm_theta_idx <- which(mnames == "theta")
  tmp <- x_free_idx[[mm_theta_idx]]
  x_theta_idx <- tmp[!duplicated(tmp)]
  m_theta_idx <- m_free_idx[[mm_theta_idx]][!duplicated(tmp)]
  n_the <- length(m_theta_idx)

  n_bet <- 0L
  x_beta_idx <- integer(0L)
  if (beta_flag) {
    mm_beta_idx <- which(mnames == "beta")
    x_beta_idx <- x_free_idx[[mm_beta_idx]]
    m_beta_idx <- m_free_idx[[mm_beta_idx]]
    n_bet <- length(m_beta_idx)
  }

  n_del <- 0L
  x_delta_idx <- integer(0L)
  if (delta_flag) {
    mm_delta_idx <- which(mnames == "delta")
    x_delta_idx <- x_free_idx[[mm_delta_idx]]
    m_delta_idx <- m_free_idx[[mm_delta_idx]]
    n_del <- length(m_delta_idx)
  }

  # composite: wmat
  n_wmat <- 0L
  x_wmat_idx <- integer(0L)
  m_wmat_idx <- integer(0L)
  wmat_flag <- !is.null(mlist$wmat)
  if (wmat_flag) {
    mm_wmat_idx <- which(mnames == "wmat")
    if (length(mm_wmat_idx) > 0L) {
      x_wmat_idx <- x_free_idx[[mm_wmat_idx]]
      m_wmat_idx <- m_free_idx[[mm_wmat_idx]]
      n_wmat <- length(m_wmat_idx)
    }
  }

  # in case !meanstructure or !categorical
  n_nu  <- 0L
  m_nu_idx    <- integer(0L)
  x_nu_idx    <- integer(0L)
  n_alp <- 0L
  m_alpha_idx <- integer(0L)
  x_alpha_idx <- integer(0L)
  n_gam <- 0L
  m_gamma_idx <- integer(0L)
  x_gamma_idx <- integer(0L)


  if (meanstructure) {
    mm_nu_idx <- which(mnames == "nu")
    x_nu_idx <- x_free_idx[[mm_nu_idx]]
    m_nu_idx <- m_free_idx[[mm_nu_idx]]
    n_nu <- length(m_nu_idx)

    mm_alpha_idx <- which(mnames == "alpha")
    x_alpha_idx <- x_free_idx[[mm_alpha_idx]]
    m_alpha_idx <- m_free_idx[[mm_alpha_idx]]
    n_alp <- length(m_alpha_idx)
  }

  if (conditional_x) {
    mm_gamma_idx <- which(mnames == "gamma")
    if (length(mm_gamma_idx) > 0L) {
      x_gamma_idx <- x_free_idx[[mm_gamma_idx]]
      m_gamma_idx <- m_free_idx[[mm_gamma_idx]]
      n_gam <- length(m_gamma_idx)
    }
  }

  if (categorical) {
    mm_th_idx <- which(mnames == "tau")
    x_th_idx <- x_free_idx[[mm_th_idx]]
    m_th_idx <- m_free_idx[[mm_th_idx]]
    n_th <- length(m_th_idx)
  }

  nvar <- nrow(mlist$lambda)
  nfac <- ncol(mlist$lambda)
  pstar <- nvar * (nvar + 1L) / 2L

  # precompute
  if (beta_flag) {
    a_1 <- lav_lisrel_ibinv(mlist)
    m <- mlist$lambda %*% a_1
    g <- a_1 %*% mlist$psi %*% t(a_1)
    lg <- mlist$lambda %*% g
  } else {
    m <- mlist$lambda
    g <- mlist$psi
    lg <- mlist$lambda %*% mlist$psi
  }

  if (delta_flag) {
    delta  <- as.vector(mlist$delta)
  }


  # vech structure for Sigma
  r_s <- lav_mat_vech_row_idx(nvar)
  c_s <- lav_mat_vech_col_idx(nvar)
  sigma_lut <- lav_mat_vech_rev(seq_len(pstar))

  if (delta_flag) {
    # scaling weights: delta[r] * delta[s] for each vech position
    delta_weight <- delta[r_s] * delta[c_s]
  }

  # helper: vec index -> (row, col)
  vec2rc <- function(idx, nr) {
    cbind(row = (idx - 1L) %% nr + 1L,
          col = (idx - 1L) %/% nr + 1L)
  }

  # prepare output matrix for sigma
  n_free <- n_lam + n_wmat + n_bet + n_psi + n_the + n_del
  jac_sigma <- matrix(0, pstar, n_free)

  col <- 1L


  if (!wmat_flag) {
    # Lambda [i,j]: dS0[r,s] = LG[s,j]*I(r==i) + LG[r,j]*I(s==i)
    if (n_lam > 0L) {
      rc  <- vec2rc(m_lambda_idx, nvar)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]

      t1 <- lg[c_s, j_v, drop = FALSE] * outer(r_s, i_v, `==`)
      t2 <- lg[r_s, j_v, drop = FALSE] * outer(c_s, i_v, `==`)

      dx <- t1 + t2
      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_lam - 1L)] <- dx
      col <- col + n_lam
    }

    # Beta [i,j]: dS0[r,s] = M[r,i]*LG[s,j] + LG[r,j]*M[s,i]
    if (n_bet > 0L) {
      rc  <- vec2rc(m_beta_idx, nfac)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]

      t1 <- m[r_s,  i_v, drop = FALSE] * lg[c_s, j_v, drop = FALSE]
      t2 <- lg[r_s, j_v, drop = FALSE] * m[c_s,  i_v, drop = FALSE]

      dx <- t1 + t2
      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_bet - 1L)] <- dx
      col <- col + n_bet
    }
   # Psi [k,l] symmetric: dS0[r,s] = M[r,k]*M[s,l] + M[r,l]*M[s,k]
    # diagonal (k==l): halve
    if (n_psi > 0L) {
      rc  <- vec2rc(m_psi_idx, nfac)
      k_v <- pmax(rc[, 1L], rc[, 2L])
      l_v <- pmin(rc[, 1L], rc[, 2L])

      t1 <- m[r_s, k_v, drop = FALSE] * m[c_s, l_v, drop = FALSE]
      t2 <- m[r_s, l_v, drop = FALSE] * m[c_s, k_v, drop = FALSE]
      dx <- t1 + t2

      diag_mask <- (k_v == l_v)
      if (any(diag_mask)) {
        dx[, diag_mask] <- dx[, diag_mask] * 0.5
      }

      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_psi - 1L)] <- dx
      col <- col + n_psi
    }

    # Theta [k,l] symmetric: dS = Delta %*% dTheta %*% Delta
    #  -> unit vector at vech pos (k,l)
    if (n_the > 0L) {
      rc  <- vec2rc(m_theta_idx, nvar)
      k_v <- pmax(rc[, 1L], rc[, 2L])
      l_v <- pmin(rc[, 1L], rc[, 2L])

      jt <- matrix(0, pstar, n_the)
      vech_pos <- sigma_lut[cbind(k_v, l_v)]
      if (delta_flag) {
        jt[cbind(vech_pos, seq_len(n_the))] <- delta_weight[vech_pos]
      } else {
        jt[cbind(vech_pos, seq_len(n_the))] <- 1
      }

      jac_sigma[, col:(col + n_the - 1L)] <- jt
      col <- col + n_the
    }

    # Delta[k] (diagonal only):
    # dS[r,s] = I(r==k)*Sigma0[k,s]*delta[s] + delta[r]*Sigma0[r,k]*I(s==k)
    if (n_del > 0L) {
      sigma0 <- m %*% mlist$psi %*% t(m) + mlist$theta
      k_v <- m_delta_idx
      t1 <- sigma0[c_s, k_v, drop = FALSE] * delta[c_s] * outer(r_s, k_v, `==`)
      t2 <- delta[r_s] * sigma0[r_s, k_v, drop = FALSE] * outer(c_s, k_v, `==`)
      jac_sigma[, col:(col + n_del - 1L)] <- t1 + t2
    }

  } else {
    # composite ov (cov) and composite lv (clv): identify from WMAT, not from
    # empty LAMBDA rows/cols (a higher-order factor measured only by latent
    # variables also has an empty LAMBDA column; see lav_lisrel_sigma()).
    cov_idx <- which(rowSums(mlist$wmat != 0) > 0L)
    clv_idx <- which(colSums(mlist$wmat != 0) > 0L)

    lw  <- mlist$lambda + mlist$wmat
    tmat <- diag(nvar)
    tmat[cov_idx, cov_idx] <- mlist$theta[cov_idx, cov_idx]
    wtw <- t(lw[, clv_idx, drop = FALSE]) %*% tmat %*%
             lw[, clv_idx, drop = FALSE]
    wtw_inv <- solve(wtw)
    wtw_inv_1 <- diag(nfac)
    wtw_inv_1[clv_idx, clv_idx] <- wtw_inv

    if (!beta_flag) {
      a_1 <- diag(nfac)
    }

    # C0: V(eta) with composite diagonal entries zeroed
    c0 <- g
    diag(c0)[clv_idx] <- 0

    # precompute
    lambdafull <- tmat %*% lw %*% wtw_inv_1
    lg_c <- lambdafull %*% c0 %*% wtw_inv_1
    h_c <- lambdafull %*% a_1
    f_c <- lambdafull %*% g
    sig0s <- lambdafull %*% c0 %*% t(lambdafull)
    lamc <- tmat %*% lw[, clv_idx, drop = FALSE] %*% wtw_inv
    # dLambdafull/dW[i,j] for i in ovc, j in clv expands to
    #   wtw.inv[j_loc,c_loc]*(Tmat - LAMc %*% wtw %*% t(LAMc))[r,i]
    #   - LAMc[i,c_loc]*LAMc[r,j_loc].
    # The "Pmat1" and "Pmat2" multipliers below feed into the wmat-block
    # T1+T2-T3-T4 expansion. Earlier versions used Tmat - LAMc %*% t(LAMc)
    # and LAMc %*% wtw.inv respectively, which are correct only when
    # wtw == I; the generally correct forms use the wtw factor explicitly:
    pmat1 <- tmat - lamc %*% wtw %*% t(lamc)
    pmat2 <- lamc

    # Beta/Psi correction matrices
    xrs_xss <- lambdafull[r_s, clv_idx, drop = FALSE] *
              lambdafull[c_s, clv_idx, drop = FALSE]
    a_sub  <- a_1[clv_idx, , drop = FALSE]
    g_sub  <- g[clv_idx, , drop = FALSE]
    # lambda
    if (n_lam > 0L) {
      rc  <- vec2rc(m_lambda_idx, nvar)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]

      t1 <- lg_c[c_s, j_v, drop = FALSE] * outer(r_s, i_v, `==`)
      t2 <- lg_c[r_s, j_v, drop = FALSE] * outer(c_s, i_v, `==`)

      dx <- t1 + t2
      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_lam - 1L)] <- dx
      col <- col + n_lam
    }

    # wmat
    if (n_wmat > 0L) {
      rc <- vec2rc(m_wmat_idx, nvar)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]
      j_loc <- match(j_v, clv_idx)

      # path 1: xi_i terms
      t1 <- pmat1[r_s, i_v, drop = FALSE] * lg_c[c_s, j_v, drop = FALSE]
      t2 <- pmat1[c_s, i_v, drop = FALSE] * lg_c[r_s, j_v, drop = FALSE]

      # path 2: phi_j * sigma_i correction
      t3 <- pmat2[r_s, j_loc, drop = FALSE] * sig0s[c_s, i_v, drop = FALSE]
      t4 <- pmat2[c_s, j_loc, drop = FALSE] * sig0s[r_s, i_v, drop = FALSE]

      dx <- t1 + t2 - t3 - t4
      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_wmat - 1L)] <- dx
      col <- col + n_wmat
    }

    # beta
    if (n_bet > 0L) {
      rc  <- vec2rc(m_beta_idx, nfac)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]

      t1 <- h_c[r_s, i_v, drop = FALSE] * f_c[c_s, j_v, drop = FALSE]
      t2 <- f_c[r_s, j_v, drop = FALSE] * h_c[c_s, i_v, drop = FALSE]

      ag     <- a_sub[, i_v, drop = FALSE] * g_sub[, j_v, drop = FALSE]
      r_corr <- 2 * xrs_xss %*% ag

      dx <- t1 + t2 - r_corr
      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_bet - 1L)] <- dx
      col <- col + n_bet
    }

    # psi
    if (n_psi > 0L) {
      rc <- vec2rc(m_psi_idx, nfac)
      k_v <- pmax(rc[, 1L], rc[, 2L])
      l_v <- pmin(rc[, 1L], rc[, 2L])

      t1 <- h_c[r_s, k_v, drop = FALSE] * h_c[c_s, l_v, drop = FALSE]
      t2 <- h_c[r_s, l_v, drop = FALSE] * h_c[c_s, k_v, drop = FALSE]

      a_kl <- a_sub[, k_v, drop = FALSE] * a_sub[, l_v, drop = FALSE]
      r_corr <- 2 * xrs_xss %*% a_kl

      dx <- t1 + t2 - r_corr

      diag_mask <- (k_v == l_v)
      if (any(diag_mask)) {
        dx[, diag_mask] <- dx[, diag_mask] * 0.5
      }

      if (delta_flag) {
        dx <- dx * delta_weight
      }
      jac_sigma[, col:(col + n_psi - 1L)] <- dx
      col <- col + n_psi
    }

    # theta
    if (n_the > 0L) {
      rc  <- vec2rc(m_theta_idx, nvar)
      k_v <- pmax(rc[, 1L], rc[, 2L])
      l_v <- pmin(rc[, 1L], rc[, 2L])

      jt <- matrix(0, pstar, n_the)
      vech_pos <- sigma_lut[cbind(k_v, l_v)]
      if (delta_flag) {
        jt[cbind(vech_pos, seq_len(n_the))] <- delta_weight[vech_pos]
      } else {
        jt[cbind(vech_pos, seq_len(n_the))] <- 1
      }

      # free composite-indicator (co)variances (a free 'T' matrix).
      # For an ordinary residual, dSigma[k,l]/dtheta[k,l] = 1 (the unit entry
      # set above). But an entry of the composite 'T' matrix also enters Sigma
      # through Lambda_c = T W (W'TW)^{-1} and Theta_c = T - Lambda_c W'TW
      # Lambda_c' (see lav_lisrel_sigma()). Replace the unit column with the
      # exact *direct* derivative dSigma/dT (holding psi fixed), obtained by
      # complex-step differentiation of lav_lisrel_sigma(). The remaining,
      # indirect path theta -> psi[clv,clv] = W'TW -> Sigma is added later by
      # the composite chain-rule correction (which differentiates psi[clv,clv]
      # w.r.t. all free parameters, T included).
      is_t <- rc[, 1L] %in% cov_idx & rc[, 2L] %in% cov_idx
      if (any(is_t)) {
        t_rc_cap <- rc[is_t, , drop = FALSE]
        mlist_cap <- mlist
        rs_cap <- r_s
        cs_cap <- c_s
        compute_sig_t <- function(tt) {
          ml <- mlist_cap
          th <- ml$theta
          for (i in seq_along(tt)) {
            th[t_rc_cap[i, 1L], t_rc_cap[i, 2L]] <- tt[i]
            th[t_rc_cap[i, 2L], t_rc_cap[i, 1L]] <- tt[i]
          }
          ml$theta <- th
          lav_lisrel_sigma(ml, delta = FALSE)[cbind(rs_cap, cs_cap)]
        }
        tt0 <- mlist$theta[t_rc_cap]
        j_t <- lav_func_jacobian_complex(func = compute_sig_t, x = tt0)
        if (delta_flag) {
          j_t <- j_t * delta_weight
        }
        jt[, which(is_t)] <- j_t
      }

      jac_sigma[, col:(col + n_the - 1L)] <- jt
      col <- col + n_the
    }

    # delta
    if (n_del > 0L) {
      sigma0_c <- sig0s + mlist$theta
      k_v      <- m_delta_idx
      t1 <- sigma0_c[c_s, k_v, drop = FALSE] * delta[c_s] *
                                              outer(r_s, k_v, `==`)
      t2 <- delta[r_s] * sigma0_c[r_s, k_v, drop = FALSE] *
                                              outer(c_s, k_v, `==`)
      jac_sigma[, col:(col + n_del - 1L)] <- t1 + t2
    }

  } # wmat.flag

  # theta parameterization (categorical / correlation only).
  #
  # In the "theta" approach (Muthen & Asparouhov, Mplus Web Note 4), the
  # diagonal of THETA is a free parameter and the diagonal scaling factor
  # Delta is *not* a free parameter; it is determined implicitly by
  #     Delta_j^{-2} = Sigma*_jj   (j ordinal),     Delta_j = 1   (j numeric).
  # The implied moments are
  #     Sigma_obs[r,s] = Sigma*[r,s] / (s_r * s_s),    s_j = sqrt(Sigma*_jj),
  #     TH_obs[t]     = (tau - nu - lambda * alpha) / s_{v(t)}.
  #
  # Up to this point, jac_sigma has been built using the prepopulated
  # MLIST$delta (= 1/s_j) values, i.e. it equals
  #     d(Delta * Sigma* * Delta) / dphi
  # treating Delta as a *constant*. The implicit dependence of Delta on
  # the other free parameters is added directly here, by writing
  #     d Sigma_obs / dphi
  #         =  (1/(s_r s_s)) * dSigma*_rs/dphi
  #          - 0.5 * Sigma_obs[r,s]/Sigma*_rr * dSigma*_rr/dphi   (r ordinal)
  #          - 0.5 * Sigma_obs[r,s]/Sigma*_ss * dSigma*_ss/dphi   (s ordinal)
  # The first term is already in jac_sigma (the Delta=const value). The two
  # correction terms are added below using jac_sigma diagonal rows
  # (which equal (1/Sigma*_jj) * dSigma*_jj/dphi for ordinal j and
  #  dSigma*_jj/dphi for numeric j -- both consistent with the formula
  #  above when combined with sigma_obs_v as the multiplier).
  jac_diag_theta <- NULL
  if (parameterization == "theta" && (categorical || correlation)) {
    sigma_star <- lav_lisrel_sigma(mlist = mlist, delta = FALSE)
    s_diag <- sqrt(diag(sigma_star))
    if (length(num_idx) > 0L) {
      s_diag[num_idx] <- 1.0
    }
    sigma_obs <- sigma_star / tcrossprod(s_diag)
    sigma_obs_v <- lav_mat_vech(sigma_obs)

    diag_pos <- lav_mat_diagh_idx(nvar)
    jac_diag_theta <- jac_sigma[diag_pos, , drop = FALSE]

    # If Delta itself is in the (compact) column block (e.g. the modindices
    # "extended" model where every model-matrix element is free), Sigma*
    # does not depend on Delta -- so dSigma*_jj/d(Delta_k) = 0. The chain
    # rule below uses jac_sigma's diagonal rows as a proxy for
    # (1/Sigma*_jj) * dSigma*_jj/dphi (correct for lambda/beta/psi/theta
    # columns), but those rows pick up the delta-block contribution of
    # jac_sigma[(r,r),:], which would spuriously feed back into the chain.
    # Zero those columns so the chain term at the delta columns is zero.
    if (n_del > 0L) {
      jac_diag_theta[,
        (ncol(jac_diag_theta) - n_del + 1L):ncol(jac_diag_theta)] <- 0
    }

    is_ord <- rep(TRUE, nvar)
    if (length(num_idx) > 0L) {
      is_ord[num_idx] <- FALSE
    }

    coef_r <- -0.5 * sigma_obs_v * is_ord[r_s]
    coef_s <- -0.5 * sigma_obs_v * is_ord[c_s]

    jac_sigma <- jac_sigma +
                 coef_r * jac_diag_theta[r_s, , drop = FALSE] +
                 coef_s * jac_diag_theta[c_s, , drop = FALSE]
  }

  # categorical: reorder
  if (categorical) {
    # reorder: first variances (of numeric), then covariances
    cov_idx <- lav_mat_vech_idx(nvar)
    covd_idx <- lav_mat_vech_idx(nvar, diagonal = FALSE)

    var_idx <- which(is.na(match(
      cov_idx,
      covd_idx
    )))[num_idx]
    cor_idx <- match(covd_idx, cov_idx)

    jac_sigma <- rbind(
      jac_sigma[var_idx, , drop = FALSE],
      jac_sigma[cor_idx, , drop = FALSE]
    )
  }

  # correlation structure
  if (!categorical && correlation) {
    if (length(num_idx) > 0L) {
      # partial correlation structure: keep the (free) variances of the
      # non-correlation variables (num_idx), drop only the correlation
      # variables' diagonal -- mirror the categorical layout (variances of
      # num_idx first, then the off-diagonal covariances)
      cov_idx <- lav_mat_vech_idx(nvar)
      covd_idx <- lav_mat_vech_idx(nvar, diagonal = FALSE)
      var_idx <- which(is.na(match(cov_idx, covd_idx)))[num_idx]
      offd_idx <- match(covd_idx, cov_idx)
      jac_sigma <- rbind(
        jac_sigma[var_idx, , drop = FALSE],
        jac_sigma[offd_idx, , drop = FALSE]
      )
    } else {
      rm_idx <- lav_mat_diagh_idx(nvar)
      jac_sigma <- jac_sigma[-rm_idx, , drop = FALSE]
    }
  }

  jac_th   <- NULL
  jac_beta <- NULL
  nth_full <- 0L
  nexo_g   <- 0L
  if (conditional_x && !is.null(mlist$gamma)) {
    nexo_g <- ncol(mlist$gamma)
  }
  if (!categorical) {
    if (conditional_x) {
      # conditional.x continuous case: implied moments are
      #   mu_i  = nu_i + (LAMBDA(I-B)^{-1} alpha)_i
      #   pi_{i,k} = (LAMBDA(I-B)^{-1} GAMMA)_{i,k}
      #   Sigma = LAMBDA(I-B)^{-1} PSI (I-B)^{-T} LAMBDA' + THETA
      # The Beta-style row layout is interleaved column-major of a
      # (nexo+1) x nvar matrix (intercept first, then nexo slopes),
      # matching what lav_mvreg_scores_* expects.
      if (beta_flag) {
        a <- drop(a_1 %*% mlist$alpha)
      } else {
        a <- if (!is.null(mlist$alpha)) as.numeric(mlist$alpha) else
             rep(0, nfac)
      }
      gamma_ib <- if (nexo_g > 0L) {
        if (beta_flag) a_1 %*% mlist$gamma else mlist$gamma
      } else {
        matrix(0, nfac, 0L)
      }

      n_beta_rows <- (nexo_g + 1L) * nvar
      mu_slots    <- (seq_len(nvar) - 1L) * (nexo_g + 1L) + 1L

      n_free_beta <- n_nu + n_lam + n_bet + n_alp + n_gam
      jac_beta <- matrix(0, n_beta_rows, n_free_beta)

      col <- 1L
      # nu[i]: only affects mu_i (intercept slot for variable i).
      if (n_nu > 0L) {
        jac_beta[cbind(mu_slots[m_nu_idx], col + seq_len(n_nu) - 1L)] <- 1
        col <- col + n_nu
      }

      # lambda[i,j]: dmu_i = a[j] * I(r==i);
      #                                dpi_{i,k} = gamma_IB[j,k] * I(r==i).
      if (n_lam > 0L) {
        rc <- vec2rc(m_lambda_idx, nvar)
        i_v <- rc[, 1L]
        j_v <- rc[, 2L]
        # mu-row: a[j_v] at row mu_slots[i_v]
        jac_beta[cbind(mu_slots[i_v], col + seq_len(n_lam) - 1L)] <- a[j_v]
        # pi-rows (k = 1..nexo)
        if (nexo_g > 0L) {
          for (k in seq_len(nexo_g)) {
            jac_beta[cbind(mu_slots[i_v] + k,
                           col + seq_len(n_lam) - 1L)] <- gamma_ib[j_v, k]
          }
        }
        col <- col + n_lam
      }

      # beta[i,j]: dmu_r = M[r,i]*a[j]; dpi_{r,k} = M[r,i]*gamma_IB[j,k].
      if (n_bet > 0L) {
        rc <- vec2rc(m_beta_idx, nfac)
        i_v <- rc[, 1L]
        j_v <- rc[, 2L]
        # mu rows
        jac_beta[mu_slots, col:(col + n_bet - 1L)] <-
          m[, i_v, drop = FALSE] * rep(a[j_v], each = nvar)
        # pi rows
        if (nexo_g > 0L) {
          for (k in seq_len(nexo_g)) {
            jac_beta[mu_slots + k, col:(col + n_bet - 1L)] <-
              m[, i_v, drop = FALSE] *
                rep(gamma_ib[j_v, k], each = nvar)
          }
        }
        col <- col + n_bet
      }

      # alpha[k]: only affects mu (dmu = M[,k]).
      if (n_alp > 0L) {
        jac_beta[mu_slots, col:(col + n_alp - 1L)] <-
          m[, m_alpha_idx, drop = FALSE]
        col <- col + n_alp
      }

      # gamma[i,k]: only affects pi_{r,k} -> M[r,i].
      if (n_gam > 0L) {
        rc <- vec2rc(m_gamma_idx, nfac)
        i_v <- rc[, 1L]
        k_v <- rc[, 2L]
        for (c2 in seq_len(n_gam)) {
          jac_beta[mu_slots + k_v[c2], col + c2 - 1L] <- m[, i_v[c2]]
        }
      }
    } else if (meanstructure) {
      # precompute: IB.inv * alpha
      if (beta_flag) {
        a <- drop(a_1 %*% mlist$alpha)
      } else {
        a <- mlist$alpha
      }

      n_free <- n_nu + n_lam + n_bet + n_alp
      jac_mean <- matrix(0, nvar, n_free)

      col <- 1L
      # nu[i]:  dmu = e_i
      if (n_nu > 0L) {
        jn <- matrix(0, nvar, n_nu)
        jn[cbind(m_nu_idx, seq_len(n_nu))] <- 1
        jac_mean[, col:(col + n_nu - 1L)] <- jn
        col <- col + n_nu
      }

      # Lambda[i,j]:  dmu[r] = a[j] * I(r == i)
      if (n_lam > 0L) {
        rc  <- vec2rc(m_lambda_idx, nvar)
        i_v <- rc[, 1L]
        j_v <- rc[, 2L]

        jl <- matrix(0, nvar, n_lam)
        jl[cbind(i_v, seq_len(n_lam))] <- a[j_v]
        jac_mean[, col:(col + n_lam - 1L)] <- jl
        col <- col + n_lam
      }

      # Beta[i,j]:  dmu[r] = M[r,i] * a[j]
      if (n_bet > 0L) {
        rc  <- vec2rc(m_beta_idx, nfac)
        i_v <- rc[, 1L]
        j_v <- rc[, 2L]

        jac_mean[, col:(col + n_bet - 1L)] <-
          m[, i_v, drop = FALSE] * rep(a[j_v], each = nvar)
        col <- col + n_bet
      }

      # alpha[k]:  dmu = M[,k]
      if (n_alp > 0L) {
        jac_mean[, col:(col + n_alp - 1L)] <- m[, m_alpha_idx, drop = FALSE]
      }
    } # meanstructure


  # categorical
  } else if (categorical) {

    th_idx_g <- th_idx
    nth_full  <- length(th_idx_g)

    # v_slot[t] = observed-variable index for TH element t
    nlev_v <- tabulate(th_idx_g, nbins = nvar)
    nlev_v[nlev_v == 0L] <- 1L
    v_slot <- rep(seq_len(nvar), times = nlev_v)

    # per-slot delta scale
    if (delta_flag) {
      delta_star <- delta[v_slot]
    } else {
      delta_star <- rep(1, nth_full)
    }

    # ord_slots[k] = position in TH vector that corresponds to tau row k
    ord_slots <- which(th_idx_g > 0L)

    # a_vec = IB.inv * alpha  (for lambda/beta derivatives)
    if (!is.null(mlist$alpha)) {
      a_vec <- if (beta_flag) {
        as.vector(a_1 %*% mlist$alpha)
      } else {
        as.vector(mlist$alpha)
      }
    } else {
      a_vec <- rep(0, nfac)
    }

    # nu_vec (intercepts; zero for purely ordinal models)
    nu_vec <- if (!is.null(mlist$nu)) as.vector(mlist$nu) else rep(0, nvar)

    # pi0 = nu + Lambda * IB.inv * alpha = nu + M * alpha
    if (!is.null(mlist$alpha)) {
      pi0 <- nu_vec + as.vector(m %*% as.vector(mlist$alpha))
    } else {
      pi0 <- nu_vec
    }

    # tau_full: tau values at ordinal TH slots, 0 at numeric TH slots
    tau_full <- numeric(nth_full)
    if (!is.null(mlist$tau) && nrow(mlist$tau) > 0L) {
      tau_full[ord_slots] <- as.vector(mlist$tau)
    }

    # unscaled TH: TH0[t] = tau_full[t] - pi0[v(t)]
    th0 <- tau_full - pi0[v_slot]

    # build jac_th  (nth_full rows, n_free_th columns)
    # column order: tau | delta | nu | lambda | beta | alpha
    n_free_th <- n_th + n_del + n_nu + n_lam + n_bet + n_alp
    jac_th <- matrix(0, nth_full, n_free_th)
    col_th <- 1L

    # tau[k]: delta_star[t] * I(t == ord_slots[k])
    if (n_th > 0L) {
      t_slots <- ord_slots[m_th_idx]
      jt <- matrix(0, nth_full, n_th)
      jt[cbind(t_slots, seq_len(n_th))] <- delta_star[t_slots]
      jac_th[, col_th:(col_th + n_th - 1L)] <- jt
      col_th <- col_th + n_th
    }

    # delta[k]: I(v_slot[t]==k) * TH0[t]
    if (n_del > 0L) {
      jac_th[, col_th:(col_th + n_del - 1L)] <-
        outer(v_slot, m_delta_idx, `==`) * th0
      col_th <- col_th + n_del
    }

    # nu[i]: -delta_star[t] * I(v_slot[t]==i)
    if (n_nu > 0L) {
      jac_th[, col_th:(col_th + n_nu - 1L)] <-
        -outer(v_slot, m_nu_idx, `==`) * delta_star
      col_th <- col_th + n_nu
    }

    # lambda[i,j]: -delta_star[t]*I(v_slot[t]==i)*a_vec[j]
    if (n_lam > 0L) {
      rc  <- vec2rc(m_lambda_idx, nvar)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]
      jac_th[, col_th:(col_th + n_lam - 1L)] <-
        -(outer(v_slot, i_v, `==`) * delta_star) *
         matrix(a_vec[j_v], nth_full, n_lam, byrow = TRUE)
      col_th <- col_th + n_lam
    }

    # beta[i,j]: -delta_star[t]*M[v_slot[t],i]*a_vec[j]
    if (n_bet > 0L) {
      rc  <- vec2rc(m_beta_idx, nfac)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]
      jac_th[, col_th:(col_th + n_bet - 1L)] <-
        -delta_star * m[v_slot, i_v, drop = FALSE] *
         matrix(a_vec[j_v], nth_full, n_bet, byrow = TRUE)
      col_th <- col_th + n_bet
    }

    # alpha[k]: -delta_star[t] * M[v_slot[t], k]
    if (n_alp > 0L) {
     jac_th[, col_th:(col_th + n_alp - 1L)] <-
       -delta_star * m[v_slot, m_alpha_idx, drop = FALSE]
    }
  }

  # categorical + conditional.x: build jac_pi (slope rows).
  #   pi_obs[r, k] = Delta[r] * (LAMBDA(I-B)^{-1} GAMMA)[r, k]
  # vec(PI) row layout is column-major: index (r, k) -> (k-1)*nvar + r.
  # Free params (in this column order): lambda | beta | gamma | delta.
  jac_pi <- NULL
  n_pi_rows <- 0L
  if (categorical && conditional_x && nexo_g > 0L) {
    n_pi_rows <- nvar * nexo_g

    gamma_ib <- if (beta_flag) a_1 %*% mlist$gamma else mlist$gamma
    delta_var <- if (delta_flag) delta else rep(1, nvar)

    n_free_pi <- n_lam + n_bet + n_gam + n_del
    jac_pi <- matrix(0, n_pi_rows, n_free_pi)
    col_pi <- 1L

    # lambda[i,j]: dpi[r,k] = Delta[r] * I(r==i) * gamma_IB[j, k]
    if (n_lam > 0L) {
      rc <- vec2rc(m_lambda_idx, nvar)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]
      for (k in seq_len(nexo_g)) {
        jac_pi[cbind((k - 1L) * nvar + i_v,
                     col_pi + seq_len(n_lam) - 1L)] <-
          delta_var[i_v] * gamma_ib[j_v, k]
      }
      col_pi <- col_pi + n_lam
    }

    # beta[i,j]: dpi[r,k] = Delta[r] * M[r,i] * gamma_IB[j, k]
    if (n_bet > 0L) {
      rc <- vec2rc(m_beta_idx, nfac)
      i_v <- rc[, 1L]
      j_v <- rc[, 2L]
      for (k in seq_len(nexo_g)) {
        rows_k <- ((k - 1L) * nvar + 1L):(k * nvar)
        jac_pi[rows_k, col_pi:(col_pi + n_bet - 1L)] <-
          delta_var * m[, i_v, drop = FALSE] *
            rep(gamma_ib[j_v, k], each = nvar)
      }
      col_pi <- col_pi + n_bet
    }

    # gamma[i,k']: dpi[r,k] = Delta[r] * M[r,i] * I(k==k')
    if (n_gam > 0L) {
      rc <- vec2rc(m_gamma_idx, nfac)
      i_v <- rc[, 1L]
      k_v <- rc[, 2L]
      for (c2 in seq_len(n_gam)) {
        rows_k <- ((k_v[c2] - 1L) * nvar + 1L):(k_v[c2] * nvar)
        jac_pi[rows_k, col_pi + c2 - 1L] <- delta_var * m[, i_v[c2]]
      }
      col_pi <- col_pi + n_gam
    }

    # delta[r] (delta-parameterization free):
    #                               dpi[r,k] = (M GAMMA)[r, k] (unscaled).
    if (n_del > 0L) {
      m_gamma <- m %*% mlist$gamma  # nvar x nexo
      for (c2 in seq_len(n_del)) {
        kd <- m_delta_idx[c2]
        for (k in seq_len(nexo_g)) {
          jac_pi[(k - 1L) * nvar + kd, col_pi + c2 - 1L] <- m_gamma[kd, k]
        }
      }
    }
  }

  # sigma
  out <- matrix(0, nrow = nrow(jac_sigma), ncol = nx_free)
  el_idx_sigma <- if (!wmat_flag) {
    c(x_lambda_idx, x_beta_idx, x_psi_idx, x_theta_idx, x_delta_idx)
  } else {
    c(x_lambda_idx, x_wmat_idx, x_beta_idx, x_psi_idx, x_theta_idx, x_delta_idx)
  }
  out[, el_idx_sigma] <- jac_sigma

  # meanstructure / conditional.x
  if (!categorical && conditional_x && !is.null(jac_beta)) {
    el_idx_beta <- c(x_nu_idx, x_lambda_idx, x_beta_idx, x_alpha_idx,
                     x_gamma_idx)
    outm <- matrix(0, nrow = nrow(jac_beta), ncol = nx_free)
    outm[, el_idx_beta] <- jac_beta
    out <- rbind(outm, out)
  } else if (!categorical && meanstructure) {
    # right order
    el_idx <- c(x_nu_idx, x_lambda_idx, x_beta_idx, x_alpha_idx)
    outm <- matrix(0, nrow = nrow(jac_mean), ncol = nx_free)
    outm[, el_idx] <- jac_mean
    out <- rbind(outm, out)
  } else if (categorical && !is.null(jac_th)) {
    el_idx_th <- c(
      if (n_th  > 0L) x_th_idx     else integer(0L),
      if (n_del > 0L) x_delta_idx  else integer(0L),
      if (n_nu  > 0L) x_nu_idx     else integer(0L),
      if (n_lam > 0L) x_lambda_idx else integer(0L),
      if (n_bet > 0L) x_beta_idx   else integer(0L),
      if (n_alp > 0L) x_alpha_idx  else integer(0L)
    )
    out_th <- matrix(0, nth_full, nx_free)
    out_th[, el_idx_th] <- jac_th

    # jac_diag_full only needed under theta parameterization (cached for
    # tau- and pi-side chain corrections).
    jac_diag_full <- NULL
    is_ord_v <- rep(TRUE, nvar)
    if (length(num_idx) > 0L) {
      is_ord_v[num_idx] <- FALSE
    }
    if (parameterization == "theta" && !is.null(jac_diag_theta)) {
      jac_diag_full <- matrix(0, nvar, nx_free)
      jac_diag_full[, el_idx_sigma] <- jac_diag_theta
    }

    # theta parameterization: chain rule for tau-side moments.
    #   TH_obs[t] = (tau - nu - lambda*alpha) / s_{v(t)},  v(t) ordinal,
    # was filled treating Delta as a constant. The implicit dependence
    # of Delta on the other free parameters adds
    #   d TH_obs[t] / dphi += -0.5 * TH_obs[t] / Sigma*_v(t),v(t)
    #                              * dSigma*_v(t),v(t) / dphi
    # which, in nx.free column space, is
    #   chain[t, :] = -0.5 * TH_obs[t] * jac_diag_full[v_slot[t], :]
    # Numeric phantom slots get no correction (Delta_num = 1 by construction,
    # so it does not depend on the free parameters).
    if (!is.null(jac_diag_full) && nth_full > 0L) {
      th_obs <- delta_star * th0 * is_ord_v[v_slot]
      out_th <- out_th +
                (-0.5 * th_obs) * jac_diag_full[v_slot, , drop = FALSE]
    }

    out <- rbind(out_th, out)

    # categorical + conditional.x: insert pi rows between th and sigma.
    if (!is.null(jac_pi)) {
      el_idx_pi <- c(
        if (n_lam > 0L) x_lambda_idx else integer(0L),
        if (n_bet > 0L) x_beta_idx   else integer(0L),
        if (n_gam > 0L) x_gamma_idx  else integer(0L),
        if (n_del > 0L) x_delta_idx  else integer(0L)
      )
      out_pi <- matrix(0, n_pi_rows, nx_free)
      out_pi[, el_idx_pi] <- jac_pi

      # theta parameterization: chain rule for pi-side moments.
      #   pi_obs[r, k] = Delta_r * unscaled_pi[r, k], with Delta_r = 1/s_r
      #   d pi_obs[r, k] / dphi += -0.5 * pi_obs[r, k] / Sigma*_rr * dSigma*_rr
      # only for ordinal r. Numeric variables have Delta_r = 1 (constant).
      if (!is.null(jac_diag_full)) {
        # construct vec(PI_obs) following the column-major row layout
        unscaled_pi <- m %*% mlist$gamma
        pi_obs_mat  <- if (delta_flag) unscaled_pi * delta else unscaled_pi
        pi_obs_v <- as.numeric(pi_obs_mat)              # nvar*nexo, vec
        r_pi     <- rep.int(seq_len(nvar), nexo_g)      # variable per row
        out_pi <- out_pi +
                  (-0.5 * pi_obs_v * is_ord_v[r_pi]) *
                    jac_diag_full[r_pi, , drop = FALSE]
      }

      # row order: out_th already on top; pi between th and sigma.
      # 'out' currently is rbind(out_th, sigma); reassemble:
      n_th_rows <- nrow(out_th)
      n_sig_rows <- nrow(out) - n_th_rows
      out <- rbind(out[seq_len(n_th_rows), , drop = FALSE],
                   out_pi,
                   out[n_th_rows + seq_len(n_sig_rows), , drop = FALSE])
    }
  }

  # composite chain-rule correction.
  # lav_lisrel_comp_set_intresvar() implicitly reparameterizes psi[clv,clv]
  # diagonal (and alpha[clv] when meanstructure is present) as a function of
  # the other free parameters (wmat / nu / theta / beta / lambda). The
  # analytical jac_sigma / jac_mean above were computed treating psi*/alpha*
  # as inputs; we add the chain term:
  #   chain[, j] = sum_{k in clv} dSigma/dpsi*[k,k] * dpsi*[k,k]/dphi_j
  #              + sum_{k in clv} dMu/dalpha*[k]   * dalpha*[k]/dphi_j
  # The closed-form dSigma/dpsi*[k,k] is the same expression used for free
  # psi[k,k] (T1+T2-R_corr halved), reused with k ranging over clv.idx;
  # dMu/dalpha*[k] = M[r,k] (zero for composite-OV rows by construction).
  # dpsi*/dphi and dalpha*/dphi are obtained by complex-step differentiation
  # through lav_lisrel_comp_set_intresvar() itself.
  if (wmat_flag && length(clv_idx) > 0L) {
    n_lvc <- length(clv_idx)

    # current free parameter vector
    x0_chain <- numeric(nx_free)
    for (mm in seq_along(mlist)) {
      if (length(m_free_idx[[mm]])) {
        x0_chain[x_free_idx[[mm]]] <- mlist[[mm]][m_free_idx[[mm]]]
      }
    }

    ml_tmpl    <- mlist
    m_free_cap <- m_free_idx
    x_free_cap <- x_free_idx
    clv_cap    <- clv_idx
    meanstr_cap <- meanstructure
    compute_lvc <- function(xx) {
      ml <- ml_tmpl
      for (mm in seq_along(ml)) {
        if (length(m_free_cap[[mm]])) {
          ml[[mm]][m_free_cap[[mm]]] <- xx[x_free_cap[[mm]]]
        }
      }
      ml <- lav_lisrel_comp_set_intresvar(ml)
      o <- ml$psi[cbind(clv_cap, clv_cap)]
      if (meanstr_cap && !is.null(ml$alpha)) {
        o <- c(o, ml$alpha[clv_cap, 1L])
      }
      o
    }
    j_lvc <- lav_func_jacobian_complex(func = compute_lvc, x = x0_chain)
    j_psi <- j_lvc[seq_len(n_lvc), , drop = FALSE]

    # dSigma/dpsi*[k,k] for k in clv.idx (closed form):
    #   H_c[r,k]*H_c[s,k] -
    #             sum_{k' in clv} Lambdafull[r,k']*Lambdafull[s,k']*A[k',k]^2
    h_c_lvc <- h_c[, clv_idx, drop = FALSE]
    m_psi   <- h_c_lvc[r_s, , drop = FALSE] *
                 h_c_lvc[c_s, , drop = FALSE]
    a_lvc_lvc <- a_1[clv_idx, clv_idx, drop = FALSE]
    m_psi   <- m_psi - xrs_xss %*% (a_lvc_lvc^2)

    sigma_chain <- m_psi %*% j_psi

    if (!categorical && meanstructure) {
      sigma_row_offset <- nvar
    } else if (categorical && !is.null(jac_th)) {
      sigma_row_offset <- nth_full
    } else {
      sigma_row_offset <- 0L
    }
    sig_rows <- sigma_row_offset + seq_len(nrow(jac_sigma))
    out[sig_rows, ] <- out[sig_rows, , drop = FALSE] + sigma_chain

    # dMu/dalpha*[k] for k in clv.idx: column M[, k]
    if (meanstructure && !categorical) {
      j_alpha  <- j_lvc[n_lvc + seq_len(n_lvc), , drop = FALSE]
      mu_chain <- m[, clv_idx, drop = FALSE] %*% j_alpha
      out[seq_len(nvar), ] <- out[seq_len(nvar), , drop = FALSE] + mu_chain
    }
  }

  # group weight: prepend a row with 1.0 at the gw column
  if (group_w_free) {
    mm_gw_idx <- which(mnames == "gw")
    if (length(mm_gw_idx) > 0L) {
      x_gw_idx <- x_free_idx[[mm_gw_idx]]
      out_gw <- matrix(0, nrow = 1L, ncol = nx_free)
      out_gw[1L, x_gw_idx] <- 1.0
      out <- rbind(out_gw, out)
    }
  }

  out
}

# compare two nested models, by default using the chi-square
# difference test

# - in 0.5-16, SB.classic = TRUE is the default again (for now)
# - in 0.5-18, SB.classic is replaced by 'method', with the following
#   options:
#     method = "default" (we choose a default method, based on the estimator)
#     method = "standard" (option to explicitly avoid robust adjustment)
#     method = "Satorra.2000"
#     method = "Satorra.Bentler.2001"
#     method = "Satorra.Bentler.2010"
#     method = "mean.var.adjusted.PLRT"
#
# - 0.6-13: RMSEA.D (also known as 'RDR') is added to the table (also if scaled,
#           since 0.6-20)
# - 0.6-13: fix multiple-group UG^2 bug in Satorra.2000 (reported by
#           Gronneberg, Foldnes and Moss)
#
# - 0.6-18:
#     New option method = "standard" (to explicitly avoid robust adjustment)
#     New test= argument to select scaled stat when
#                           method="satorra.bentler.2001/2010"


lavTestLRT <- function(object, ..., method = "default", test = "default",   # nolint
                       a_method = "delta", scaled_shifted = TRUE, # only when method="Satorra.2000"
                       type = "Chisq", model_names = NULL) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)
  type <- tolower(type[1])
  test <- tolower(test[1])
  method_orig <- method[1]
  method <- tolower(gsub("[-_\\.]", "", method[1]))
  if (type %in% c("browne", "browne.residual.adf", "browne.residual.nt")) {
    if (type == "browne") {
      type <- "browne.residual.adf"
    }
    if (!method %in% c("default", "standard")) {
      lav_msg_stop(gettext(
        "method cannot be used if type is browne.residual.adf or
        browne.residual.nt"))
    }
    method <- "default"
  }

  # NOTE: if we add additional arguments, it is not the same generic
  # anova() function anymore, and match.call will be screwed up

  mcall <- match.call(expand.dots = TRUE)
  modp <- if (length(dotdotdot)) {
    sapply(dotdotdot, inherits, "lavaan")
  } else {
    logical(0L)
  }
  # check object
  object <- lav_object_check_version(object)
  # check models in dotdotdot
  dotdotdot[modp] <- lapply(dotdotdot[modp], lav_object_check_version)

  # sam objects (issue #517): when comparing multiple models that were all
  # fitted with sam() and a local sam.method, delegate to the stored step-2
  # structural fits: the LRT then compares the structural parts, assuming
  # identical (fixed) step-1 measurement models. (A single local sam object
  # needs no special treatment: its @test slot already holds the structural
  # test statistics.)
  if (any(modp)) {
    all_mods <- c(list(object), dotdotdot[modp])
    local_flags <- vapply(all_mods, lav_sam_local_flag, logical(1L))
    if (all(local_flags)) {
      struc_list <- lapply(all_mods, lav_sam_struc_object)
      if (any(vapply(struc_list, is.null, logical(1L)))) {
        lav_msg_stop(gettext(
          "no stored structural fit found for at least one model: sam
           object(s) created by an older version of lavaan; please rerun
           sam()."))
      }
      # the comparison only makes sense if the measurement part is the same
      mm_1 <- all_mods[[1]]@internal$sam.mm.table
      same_mm <- vapply(all_mods[-1], function(x) {
        isTRUE(all.equal(x@internal$sam.mm.table, mm_1))
      }, logical(1L))
      if (!all(same_mm)) {
        lav_msg_warn(gettext(
          "the models do not share the same measurement blocks; the
           chi-square difference test of the structural parts may not be
           meaningful"))
      }
      lav_msg_note(gettext(
        "the chi-square difference test compares the structural parts only,
         conditional on identical (fixed) step-1 measurement models."))
      object <- struc_list[[1]]
      dotdotdot[modp] <- struc_list[-1]
    } else if (any(local_flags)) {
      lav_msg_stop(gettext(
        "lavTestLRT() cannot compare a model fitted with sam() (using a
         local sam.method) with other models."))
    }
  }

  # some general properties (taken from the first model)
  estimator <- object@Options$estimator
  ngroups <- object@Data@ngroups
  ntotal <- object@SampleStats@ntotal

  # TDJ: check for user-supplied h1 model
  user_h1_exists <- FALSE
  if (!is.null(object@external$h1.model)) {
    if (inherits(object@external$h1.model, "lavaan")) {
      user_h1_exists <- TRUE
    }
  }

  # shortcut for single argument (just plain LRT)
  if (!any(modp) && !user_h1_exists) {
    if (lav_test_fmg_is_preset(test)) {
      test <- lav_test_fmg_resolve_preset(
        test, nested = FALSE, ngroups = ngroups
      )
    }
    if (type == "cf") {
      lav_msg_warn(gettext("`type' argument is ignored for a single model"))
    }
    return(lav_test_lrt_single_model(object, method = method,
       test = test, type = type))
  }

  # list of models
  mods <- c(list(object), dotdotdot[modp])
  if (!is.null(model_names)) {
    names(mods) <- model_names
  } else {
    names(mods) <- sapply(
      as.list(mcall)[which(c(FALSE, TRUE, modp))],
      function(x) deparse(x)
    )
  }
  # TDJ: Add user-supplied h1 model, if it exists
  if (user_h1_exists) mods$user_h1 <- object@external$h1.model

  # ensure model names are unique (they are used as row.names later)
  names(mods) <- make.unique(names(mods))

  # put them in order (using degrees of freedom)
  ndf <- sapply(mods, function(x) x@test[[1]]$df)
  order_idx <- order(ndf)
  mods <- mods[order_idx]
  ndf <- ndf[order_idx]

  if (lav_test_fmg_is_preset(test)) {
    test <- lav_test_fmg_resolve_preset(
      test, nested = TRUE, ngroups = ngroups
    )
  }

  # here come the checks -- eventually, an option may skip this
  if (TRUE) {
    # 1. same set of observed variables?
    ov_names <- lapply(mods, function(x) {
      sort(lav_object_vnames(x))
    })
    ov <- ov_names[[1L]] # the observed variable names of the first model
    if (!all(sapply(ov_names, function(x) identical(x, ov)))) {
      lav_msg_warn(gettext(
        "some models are based on a different set of observed variables"))
    }
    ## wow FIXME: we may need to reorder the rows/columns first!!
    # COVS <- lapply(mods, function(x) slot(slot(x, "Sample"), "cov")[[1]])
    # if(!all(sapply(COVS, all.equal, COVS[[1]]))) {
    #    stop("lavaan ERROR: models must be fit to the same data")
    # }
    # 2. nested models? *different* npars?

    # TODO!

    # 3. all meanstructure?
    mean_structure <- sapply(mods, lavInspect, "meanstructure")
    if (sum(mean_structure) > 0L &&
      sum(mean_structure) < length(mean_structure)) {
      lav_msg_warn(gettext("not all models have a meanstructure"))
    }

    # 4. all converged?
    if (!all(sapply(mods, lavInspect, "converged"))) {
      lav_msg_warn(gettext("not all models converged"))
    }
  }

  if (type == "chisq" && lav_test_fmg_is_fmg(test)) {
    return(lav_test_lrt_fmg(
      mods = mods,
      test = test,
      method = method,
      method_orig = method_orig,
      estimator = estimator,
      ntotal = ntotal,
      ngroups = ngroups,
      missing = object@Options$missing
    ))
  }

  mods_scaled <- unlist(lapply(mods, function(x) {
    any(c(
      "satorra.bentler", "yuan.bentler", "yuan.bentler.mplus",
      "mean.var.adjusted", "scaled.shifted",
      "mean.var.adjusted.corrected", "scaled.shifted.corrected"
    ) %in%
      unlist(sapply(slot(x, "test"), "[[", "test")))
  }))

  if (all(mods_scaled | ndf == 0) && any(mods_scaled)) {
    # Note: if df=0, test is not really robust, hence the above condition
    scaled <- TRUE
    # which test to choose by default?
    # i.e., not determined by method=
    scaled_list <- sapply(mods[[which(ndf > 0)[1]]]@test,
                         # first mod with df>0
                         #FIXME: ? If no mods have df > 0,
                         #         this still yields error
                         function(x) !is.null(x$scaled.test.stat))
    if (!any(scaled_list)) {
      # no test entry carries a scaled.test.stat field (eg PML objects
      # created before 0.7-1): fall back to matching the scaled-test names
      scaled_list <- sapply(mods[[which(ndf > 0)[1]]]@test,
        function(x) {
          x$test[1] %in% c(
            "satorra.bentler", "yuan.bentler", "yuan.bentler.mplus",
            "mean.var.adjusted", "scaled.shifted",
            "mean.var.adjusted.corrected", "scaled.shifted.corrected"
          )
        })
    }
    scaled_idx <- which(scaled_list)[[1]]
    default_test <- object@test[[scaled_idx]]$test
    if (test == "default") {
      test_1 <- default_test
    } else if (!test %in% c("satorra.bentler", "yuan.bentler",
     "yuan.bentler.mplus", "mean.var.adjusted", "scaled.shifted",
     "mean.var.adjusted.corrected", "scaled.shifted.corrected")) {
      lav_msg_stop(gettextf(
        "test = %s not found in object. See available tests in
        lavInspect(object, \"options\")$test.", dQuote(test)))
    } else {
      test_1 <- test
    }

    ## is the test available from all models?
    check_scaled <- unlist(lapply(mods, function(x) {
      test_1 %in% unlist(sapply(slot(x, "test"), "[[", "test"))
    }))

    if (any(!check_scaled)) {
      lav_msg_stop(gettextf(
        "test = %1$s not found in model(s): %2$s. Find available tests per model
        using lavInspect(fit, \"options\")$test.", dQuote(test),
        lav_msg_view(names(mods)[which(!check_scaled)], "none")))
    }

  } else if (!any(mods_scaled)) { # thanks to R.M. Bee to fix this
    scaled <- FALSE
    # if the user explicitly requested a robust difference test, but none of
    # the models were fitted with a robust (scaled) test statistic, the
    # method= argument cannot be honored: warn that it is ignored and a
    # standard (regular) chi-squared difference test is computed instead
    # (see GitHub issue #383)
    if (method %in% c("satorra2000", "satorrabentler2001",
                      "satorrabentler2010", "meanvaradjustedplrt")) {
      lav_msg_warn(gettextf(
        "method = %s requires robust test statistics, but none of the models
        were fitted with a robust test; the method= argument is ignored, and a
        standard (regular) chi-squared difference test is computed instead.",
        dQuote(method_orig)))
    }
    test_1 <- "standard"
    method <- "standard"
  } else {
    lav_msg_stop(gettext(
      "some models (but not all) have scaled test statistics"))
  }

  if (type %in% c("browne.residual.adf", "browne.residual.nt")) {
    scaled <- FALSE
    method <- "standard"
  }

  if (method == "standard") {
    scaled <- FALSE
  }

  # select method
  if (method == "default") {
    if (estimator == "PML") {
      method <- "mean.var.adjusted.PLRT"
    } else if (scaled) {
      if (test_1 %in% c(
        "satorra.bentler", "yuan.bentler",
        "yuan.bentler.mplus"
      )) {
        method <- "satorra.bentler.2001"
      } else {
        method <- "satorra.2000"
        if (test_1 %in% c(
          "mean.var.adjusted.corrected", "scaled.shifted.corrected"
        )) {
          # the corrected (Hayakawa 2018) trace estimator is not (yet)
          # available for difference tests
          lav_msg_warn(gettext(
            "the difference test uses the standard (uncorrected) trace
            estimates, not the corrected (Hayakawa 2018) versions."))
        }
      }
    } else {
      # nothing to do
    }
  } else if (method == "meanvaradjustedplrt" ||
    method == "mean.var.adjusted.PLRT") {
    method <- "mean.var.adjusted.PLRT"
    stopifnot(estimator == "PML")
  } else if (method == "satorra2000") {
    method <- "satorra.2000"
  } else if (method == "satorrabentler2001") {
    method <- "satorra.bentler.2001"
  } else if (method == "satorrabentler2010") {
    method <- "satorra.bentler.2010"

    ## only option left:
  } else if (method != "standard") {
    lav_msg_stop(
      gettextf("Unknown method for scaled difference test: %s.", method))
  }

  ## in case users specify method= or test= (but still type="chisq"),
  ## make sure the arguments are consistent for scaled tests
  if (method %in% c("satorra.bentler.2001", "satorra.bentler.2010") && scaled &&
    (!test_1 %in% c("satorra.bentler", "yuan.bentler", "yuan.bentler.mplus"))) {
    lav_msg_stop(gettextf(
      "method = %s only available when models are fitted with test =
      \"satorra.bentler\", \"yuan.bentler\", or \"yuan.bentler.mplus\".",
      dQuote(method)))
  } else {
    ## method="satorra.2000" still available when test_1 != scaled_shifted
    ## Or !scaled, so nothing to do.
  }

  # NOTE: when scaled = FALSE, method has already been reset to "standard"
  # above (either because none of the models used a robust test statistic --
  # in which case a warning was issued -- or because the user explicitly
  # requested method = "standard" or a Browne residual type). So there is no
  # need to check for a robust method here anymore.

  # which models have used a MEANSTRUCTURE?
  mods_meanstructure <- sapply(mods, function(x) {
    unlist(slot(
      slot(x, "Model"),
      "meanstructure"
    ))
  })
  if (!all(mods_meanstructure) && sum(mods_meanstructure) != 0) {
    lav_msg_stop(gettext("some models (but not all) have a meanstructure"))
  }

  # collect statistics for each model
  if (type == "chisq") {
    df_1 <- sapply(mods, function(x) slot(x, "test")[[1]]$df)
    stat_1 <- sapply(mods, function(x) slot(x, "test")[[1]]$stat)
  } else if (type == "browne.residual.nt") {
    testlist <- lapply(
      mods,
      function(x) lavTest(x, test = "browne.residual.nt",
                          drop_list_single = FALSE)
    )
    df_1 <- sapply(testlist, function(x) x[[type]]$df)
    stat_1 <- sapply(testlist, function(x) x[[type]]$stat)
  } else if (type == "browne.residual.adf") {
    testlist <- lapply(
      mods,
      function(x) lavTest(x, test = "browne.residual.adf",
                          drop_list_single = FALSE)
    )
    df_1 <- sapply(testlist, function(x) x[[type]]$df)
    stat_1 <- sapply(testlist, function(x) x[[type]]$stat)
  } else if (type == "cf") {
    tmp <- lapply(mods, lavTablesFitCf)
    stat_1 <- unlist(tmp)
    df_1 <- unlist(lapply(tmp, attr, "DF"))
  } else {
    lav_msg_stop(gettextf("test type unknown: %s", type))
  }

  # difference statistics
  stat_delta <- stat_delta_orig <- c(NA, diff(stat_1))
  df_delta <- df_delta_orig <- c(NA, diff(df_1))

  # check for negative values in stat_delta
  # but with a tolerance (0.6-12)!
  if (any(stat_delta[-1] < -1 * .Machine$double.eps^(1 / 3))) {
    lav_msg_warn(gettextf(
      "Some restricted models fit better than less restricted models; either
      these models are not nested, or the less restricted model failed to reach
      a global optimum. Smallest difference = %s.", min(stat_delta[-1])))
  }

  # prepare for scaling versions
  if (method == "satorra.2000" && scaled_shifted) {
    a_delta <- b_delta <- rep(as.numeric(NA), length(stat_1))
    c_delta <- NULL
  } else if (method %in% c("satorra.bentler.2001",
           "satorra.bentler.2010", "satorra.2000")) {
    c_delta <- rep(as.numeric(NA), length(stat_1))
  } else {
    # eg method = "mean.var.adjusted.PLRT" or "standard": no scaling
    # attributes (c_delta is consulted below, so it must exist)
    c_delta <- NULL
  }

  # correction for scaled test statistics
  if (type == "chisq" && scaled) {
    if (method == "satorra.bentler.2001") {
      # use formula from Satorra & Bentler 2001
      for (m in seq_len(length(mods) - 1L)) {
        out <- lav_test_diff_sb2001(mods[[m]], mods[[m + 1]],
                                                # in case not @test[[2]]:
                                                test = test_1)
        stat_delta[m + 1] <- out$T.delta
        df_delta[m + 1] <- out$df.delta
        c_delta[m + 1] <- out$scaling.factor
      }
    } else if (method == "mean.var.adjusted.PLRT") {
      for (m in seq_len(length(mods) - 1L)) {
        out <- lav_pml_test_plrt(mods[[m]], mods[[m + 1]])
        stat_delta[m + 1] <- out$FSMA.PLRT
        df_delta[m + 1] <- out$adj.df
      }
    } else if (method == "satorra.bentler.2010") {
      for (m in seq_len(length(mods) - 1L)) {
        out <- lav_test_diff_sb2010(mods[[m]], mods[[m + 1]],
          test = test_1, # in case not @test[[2]]
          h1 = FALSE
        ) # must be F

        stat_delta[m + 1] <- out$T.delta
        df_delta[m + 1] <- out$df.delta
        c_delta[m + 1] <- out$scaling.factor
      }
    } else if (method == "satorra.2000") {
      for (m in seq_len(length(mods) - 1L)) {
        if (test_1 %in% c(
          "satorra.bentler", "yuan.bentler",
          "yuan.bentler.mplus"
        )) {
          satterthwaite <- FALSE
        } else {
          satterthwaite <- TRUE
         }
        out <- lav_test_diff_satorra2000(mods[[m]], mods[[m + 1]],
          h1 = TRUE,
          satterthwaite = satterthwaite,
          scaled_shifted = scaled_shifted,
          a_method = a_method
        )
        stat_delta[m + 1] <- out$t_delta
        df_delta[m + 1] <- out$df_delta
        if (scaled_shifted) {
          a_delta[m + 1] <- out$a
          b_delta[m + 1] <- out$b
        } else {
          c_delta[m + 1] <- out$scaling.factor
        }
      }
    }
  }

  # check if scaled diff failed somehow
  if (scaled &&
     ((method %in% c("satorra.bentler.2001", "satorra.bentler.2010") &&
          is.na(out$scaling.factor)) ||
         (method == "satorra.2000" && scaled_shifted && is.na(out$a)) ||
         (method == "satorra.2000" && !scaled_shifted &&
          is.na(out$scaling.factor)))
     ) {
    scaled <- FALSE
  }


  # unname
  stat_delta <- unname(stat_delta)
  # zap small values (for near-zero values) (anova class does not use rounding)
  stat_delta <- round(stat_delta, 10)
  df_delta <- unname(df_delta)
  stat_delta_orig <- unname(stat_delta_orig)
  df_delta_orig <- unname(df_delta_orig)
  if (scaled && !is.null(c_delta)) {
    c_delta <- unname(c_delta)
  }

  # Pvalue
  pvalue_delta <- pchisq(stat_delta, df_delta, lower.tail = FALSE)

  # new in 0.6-13: RMSEA (RMSEA.D or RDR)
  if (object@Options$missing == "listwise") {
    if (scaled && !is.null(c_delta)) {
      c_hat <- c_delta[-1]
    } else {
      c_hat <- rep(1, length(stat_delta_orig) - 1L)
    }
    rmsea_delta <- c(NA, lav_fit_rmsea(
      x2 = stat_delta_orig[-1],
      df = df_delta_orig[-1],
      n = ntotal,
      g = ngroups,
      c_hat = c_hat
    ))
  }

  # AIC/BIC
  aic <- bic <- rep(NA, length(mods))
  if (estimator == "ML") {
    aic <- sapply(mods, FUN = AIC)
    bic <- sapply(mods, FUN = BIC)
  } else if (estimator == "PML") {
    out_1 <- lapply(mods, lav_pml_object_aic_bic)
    aic <- sapply(out_1, "[[", "PL_AIC")
    bic <- sapply(out_1, "[[", "PL_BIC")
  }

  if (estimator == "PML") {
    val <- data.frame(
      Df = df_1,
      PL_AIC = aic,
      PL_BIC = bic,
      Chisq = stat_1,
      "Chisq diff" = stat_delta,
      "Df diff" = df_delta,
      "Pr(>Chisq)" = pvalue_delta,
      row.names = names(mods),
      check.names = FALSE
    )
  } else if (object@Options$missing == "listwise") {
    val <- data.frame(
      Df = df_1,
      AIC = aic,
      BIC = bic,
      Chisq = stat_1,
      "Chisq diff" = stat_delta,
      "RMSEA" = rmsea_delta,
      "Df diff" = df_delta,
      "Pr(>Chisq)" = pvalue_delta,
      row.names = names(mods),
      check.names = FALSE
    )
  } else {
    val <- data.frame(
      Df = df_1,
      AIC = aic,
      BIC = bic,
      Chisq = stat_1,
      "Chisq diff" = stat_delta,
      #"RMSEA" = rmsea_delta, # if missing, not yet...
      "Df diff" = df_delta,
      "Pr(>Chisq)" = pvalue_delta,
      row.names = names(mods),
      check.names = FALSE
    )
  }

  # catch df_delta == 0 cases (reported by Florian Zsok in Zurich)
  # but only if there are no inequality constraints! (0.6-1)
  idx <- which(val[, "Df diff"] == 0)
  if (length(idx) > 0L) {
    # remove models with inequality constraints
    ineq_idx <- which(sapply(lapply(mods, function(x) {
      slot(slot(x, "Model"), "x.cin.idx")
    }), length) > 0L)
    rm_idx <- which(idx %in% ineq_idx)
    if (length(rm_idx) > 0L) {
      idx <- idx[-rm_idx]
    }
  }
  if (length(idx) > 0L) {
    val[idx, "Pr(>Chisq)"] <- as.numeric(NA)
    lav_msg_warn(gettext("some models have the same degrees of freedom"))
  }

  if (type == "chisq") {
    if (scaled) {
      txt <- paste("The ", dQuote("Chisq"), " column contains standard ",
        "test statistics, not the robust test that should be ",
        "reported per model. A robust difference test is a ",
        "function of two standard (not robust) statistics.",
        sep = ""
      )
      # describe the difference test that is actually computed, so the title
      # is not misleading (see GitHub issue #523). For method="satorra.2000"
      # the exact statistic depends on both scaled_shifted= and the test that
      # was used to fit the models (test_1); see lav_test_diff_satorra2000().
      if (method == "satorra.2000" && scaled_shifted) {
        stat_label <- "Scaled and Shifted"
      } else if (method == "satorra.2000" &&
        !test_1 %in% c("satorra.bentler", "yuan.bentler",
          "yuan.bentler.mplus")) {
        # satterthwaite-style (mean and variance adjusted) difference test
        stat_label <- "Mean and Variance Adjusted"
      } else if (method == "mean.var.adjusted.PLRT") {
        stat_label <- "Mean and Variance Adjusted"
      } else {
        # satorra.bentler.2001 / satorra.bentler.2010, or satorra.2000 with a
        # satorra.bentler/yuan.bentler test: a (Satorra-Bentler) scaled test
        stat_label <- "Scaled"
      }
      attr(val, "heading") <-
        paste("\n", stat_label, " Chi-Squared Difference Test (method = ",
          dQuote(method), ")\n\n",
          lav_msg(paste("lavaan NOTE:", txt), showheader = TRUE),
          "\n",
          sep = ""
        )
      if (method == "satorra.2000" && scaled_shifted) {
        attr(val, "scale") <- a_delta
        attr(val, "shift") <- b_delta
      } else if (method %in% c("satorra.bentler.2001",
                "satorra.bentler.2010", "satorra.2000")) {
        attr(val, "scale") <- c_delta
      }
    } else {
      attr(val, "heading") <- "\nChi-Squared Difference Test\n"
    }
  } else if (type == "browne.residual.adf") {
    attr(val, "heading") <-
      "\nChi-Squared Difference Test based on Browne's residual (ADF) Test\n"
  } else if (type == "browne.residual.nt") {
    attr(val, "heading") <-
      "\nChi-Squared Difference Test based on Browne's residual (NT) Test\n"
  } else if (type == "cf") {
    colnames(val)[c(3, 4)] <- c("Cf", "Cf diff")
    attr(val, "heading") <- "\nCf Difference Test\n"
  }
  class(val) <- c("anova", class(val))

  val
}


# anova table for a single model
lav_test_lrt_single_model <- function(object, method = "default",
                                      test = "default", type = "Chisq") {
  estimator <- object@Options$estimator

  aic <- bic <- c(NA, NA)
  if (estimator == "ML") {
    aic <- c(NA, AIC(object))
    bic <- c(NA, BIC(object))
  }

  ## determine which @test element
  tn <- names(object@test)
  if (is.null(tn)) {
    tn <- "standard" # for lavaan <0.6 objects
  }
  if (length(tn) == 1L) {
    test_1 <- 1L # only choice

    ## More than 1.  Cycle through possible user specifications:
  } else if (method[1] == "standard") {
    test_1 <- 1L
  } else if (grepl("browne", type, fixed = TRUE) && type %in% tn) {
    test_1 <- type
  } else if (test %in% tn) {
    test_1 <- test
  } else {
    ## Nothing explicitly (or validly) requested.
    ## But there is > 1 test, so take the second element (old default)
    test_1 <- 2L
  }

  ## anova table
  val <- data.frame(
    Df = c(0, object@test[[test_1]]$df),
    AIC = aic,
    BIC = bic,
    Chisq = c(0, object@test[[test_1]]$stat),
    "Chisq diff" = c(NA, object@test[[test_1]]$stat),
    "Df diff" = c(NA, object@test[[test_1]]$df),
    "Pr(>Chisq)" = c(NA, object@test[[test_1]]$pvalue),
    row.names = c("Saturated", "Model"),
    check.names = FALSE
  )
  ## scale/shift attributes
  if (!is.null(object@test[[test_1]]$scaling.factor)) {
    attr(val, "scale") <- c(NA, object@test[[test_1]]$scaling.factor)
  }
  if (!is.null(object@test[[test_1]]$shift.parameter)) {
    attr(val, "shift") <- c(NA, object@test[[test_1]]$shift.parameter)
  }

  ## heading
  if (grepl(pattern = "browne", x = test_1, fixed = TRUE)) {
    attr(val, "heading") <- object@test[[test_1]]$label

  } else if (test_1 == 1L) {
    attr(val, "heading") <- "Chi-Squared Test Statistic (unscaled)\n"

  } else {
    label <- object@test[[test_1]]$label
    attr(val, "heading") <- paste0("Chi-Squared Test Statistic (scaled",
                                   ifelse(test_1 == "scaled.shifted",
                                          yes = " and shifted)", no = ")"),
                                   ifelse(is.null(label),
                                          yes = "\n", no = paste("\n ", label)),
                                   "\n")
  }

  class(val) <- c("anova", class(val))

  val
}

lav_test_lrt_fmg <- function(mods, test = "pall_ug_ml", method = "default",
                             method_orig = method,
                             estimator = "ML", ntotal = NULL,
                             ngroups = NULL, missing = "listwise") {
  if (!method %in% c("default", "standard", "satorra2000")) {
    lav_msg_stop(gettextf(
      "FMG nested tests require method= %1$s, %2$s, or %3$s; found %4$s.",
      dQuote("default"), dQuote("standard"), dQuote("satorra.2000"),
      dQuote(method_orig)
    ))
  }
  invisible(lapply(
    mods,
    function(x) {
      lav_test_fmg_check_estimator(x@Options, context = "FMG nested tests")
      lav_test_fmg_check_mi(x@Options, context = "FMG nested tests")
    }
  ))
  estimators <- vapply(mods, function(x) x@Options$estimator.orig,
                       character(1L))
  if (length(unique(estimators)) > 1L) {
    lav_msg_stop(gettextf(
      "FMG nested tests require all models to use the same estimator; found %1$s.",
      lav_msg_view(unique(estimators), "and")
    ))
  }

  parsed <- lav_test_fmg_parse(test)
  chisq <- lav_test_fmg_resolve_chisq(parsed, lavoptions = mods[[1]]@Options)
  unbiased <- lav_test_fmg_resolve_unbiased(
    parsed, lavoptions = mods[[1]]@Options
  )
  label <- lav_test_fmg_label(parsed, unbiased = unbiased, chisq = chisq)
  heading_label <- sub(" p-value test \\(", "; ", label)
  heading_label <- sub("\\)$", "", heading_label)

  if (chisq == "rls") {
    TESTlist <- lapply(
      mods,
      function(x) lav_test_fmg_browne_nt_model(lavobject = x)
    )
    Df <- sapply(TESTlist, function(x) x$df)
    STAT <- sapply(TESTlist, function(x) x$stat)
  } else {
    Df <- sapply(mods, function(x) slot(x, "test")[[1]]$df)
    STAT <- sapply(mods, function(x) slot(x, "test")[[1]]$stat)
  }

  stat_delta <- stat_delta_orig <- c(NA, diff(STAT))
  df_delta <- df_delta_orig <- c(NA, diff(Df))
  pvalue_delta <- rep(as.numeric(NA), length(mods))

  if (length(mods) > 1L) {
    for (m in seq_len(length(mods) - 1L)) {
      out <- lav_test_fmg_nested(
        m0 = mods[[m + 1L]],
        m1 = mods[[m]],
        test = test
      )
      stat_delta[m + 1L] <- out$stat
      df_delta[m + 1L] <- out$df
      pvalue_delta[m + 1L] <- out$pvalue
    }
  }

  stat_delta <- round(unname(stat_delta), 10)
  df_delta <- unname(df_delta)
  stat_delta_orig <- unname(stat_delta_orig)
  df_delta_orig <- unname(df_delta_orig)

  aic <- bic <- rep(NA, length(mods))
  if (estimator == "ML") {
    aic <- sapply(mods, FUN = AIC)
    bic <- sapply(mods, FUN = BIC)
  } else if (estimator == "PML") {
    OUT <- lapply(mods, lav_pml_object_aic_bic)
    aic <- sapply(OUT, "[[", "PL_AIC")
    bic <- sapply(OUT, "[[", "PL_BIC")
  }

  if (missing == "listwise") {
    rmsea_delta <- c(NA, lav_fit_rmsea(
      x2 = stat_delta_orig[-1],
      df = df_delta_orig[-1],
      n = ntotal,
      g = ngroups,
      c_hat = rep(1, length(stat_delta_orig) - 1L)
    ))

    val <- data.frame(
      Df = Df,
      AIC = aic,
      BIC = bic,
      Chisq = STAT,
      "Chisq diff" = stat_delta,
      "RMSEA" = rmsea_delta,
      "Df diff" = df_delta,
      "Pr(>Chisq)" = pvalue_delta,
      row.names = names(mods),
      check.names = FALSE
    )
  } else {
    val <- data.frame(
      Df = Df,
      AIC = aic,
      BIC = bic,
      Chisq = STAT,
      "Chisq diff" = stat_delta,
      "Df diff" = df_delta,
      "Pr(>Chisq)" = pvalue_delta,
      row.names = names(mods),
      check.names = FALSE
    )
  }

  idx <- which(val[, "Df diff"] == 0)
  if (length(idx) > 0L) {
    val[idx, "Pr(>Chisq)"] <- as.numeric(NA)
    lav_msg_warn(gettext("some models have the same degrees of freedom"))
  }

  txt <- if (chisq == "rls") {
    paste("The ", dQuote("Chisq"), " column contains source RLS ",
      "test statistics, not the FMG test that should be reported per ",
      "model. An FMG difference test is a function of two source ",
      "statistics.",
      sep = ""
    )
  } else {
    paste("The ", dQuote("Chisq"), " column contains standard ",
      "test statistics, not the FMG test that should be reported per ",
      "model. An FMG difference test is a function of two standard ",
      "(not FMG) statistics.",
      sep = ""
    )
  }
  note <- paste(strwrap(
    paste("lavaan NOTE:", txt),
    width = getOption("width", 80L) - 3L
  ), collapse = "\n")
  note <- paste0(
    "lavaan->lavTestLRT():  \n   ",
    gsub("\n", "\n   ", note, fixed = TRUE)
  )
  attr(val, "heading") <- paste0(
    "\nFMG Chi-Squared Difference Test (", heading_label, ")\n\n",
    note,
    "\n"
  )
  class(val) <- c("anova", class(val))

  val
}

# James-Stein estimator
#
# Burghgraeve, E., De Neve, J., & Rosseel, Y. (2021). Estimating structural
# equation models using James-Stein type shrinkage estimators. Psychometrika,
# 86(1), 96-130.
#
# YR 08 Feb 2023: - first version in lavaan, cfa only (for now)

lav_cfa_jamesstein <- function(s,
                               y = NULL, # raw data
                               marker_idx = NULL,
                               lambda_nonzero_idx = NULL,
                               theta = NULL, # vector!
                               theta_bounds = TRUE,
                               aggregated = FALSE) { # aggregated?
  # dimensions
  nvar <- ncol(s)
  nfac <- length(marker_idx)
  stopifnot(length(theta) == nvar)
  stopifnot(ncol(y) == nvar)

  # overview of lambda structure
  m_b <- mm_lambda <- b_nomarker <- matrix(0, nvar, nfac)
  lambda_marker_idx <- (seq_len(nfac) - 1L) * nvar + marker_idx
  m_b[lambda_marker_idx] <- mm_lambda[lambda_marker_idx] <- 1L
  m_b[lambda_nonzero_idx] <- b_nomarker[lambda_nonzero_idx] <- 1L

  # Nu
  mm_nu <- numeric(nvar)

  # do we first 'clip' the theta values so they are within standard bounds?
  # (Question: do we need the 0.01 and 0.99 multipliers?)
  diag_s <- diag(s)
  if (theta_bounds) {
    # lower bound
    lower_bound <- diag_s * 0 # * 0.01
    too_small_idx <- which(theta < lower_bound)
    if (length(too_small_idx) > 0L) {
      theta[too_small_idx] <- lower_bound[too_small_idx]
    }

    # upper bound
    upper_bound <- diag_s * 1 # * 0.99
    too_large_idx <- which(theta > upper_bound)
    if (length(too_large_idx) > 0L) {
      theta[too_large_idx] <- upper_bound[too_large_idx]
    }
  }

  # compute conditional expectation conditional on the scaling indicator
  e_js1 <- lav_cfa_jamesstein_ce(
    y = y, marker_idx = marker_idx,
    resvars_markers = theta[marker_idx]
  )

  # compute LAMBDA
  for (f in seq_len(nfac)) {
    nomarker_idx <- which(b_nomarker[, f] == 1)
    y_nomarker_f <- y[, nomarker_idx, drop = FALSE]

    # regress no.marker.idx data on E(\eta|Y)
    fit <- lm(y_nomarker_f ~ e_js1[, f, drop = FALSE])

    # extract 'lambda' values
    mm_lambda[nomarker_idx, f] <- drop(coef(fit)[-1, ])

    # (optional) extract means
    # NU[nomarker.idx] <- drop(coef(fit)[1,])

    if (aggregated) {
      # local copy of 'scaling' LAMBDA
      lambda_scaling <- mm_lambda

      j_1 <- length(nomarker_idx)
      for (j in seq_len(j_1)) {
        # data without this indicator
        j_idx <- nomarker_idx[j]
        no_j_idx <- c(marker_idx[f], nomarker_idx[-j])
        y_agg <- y[, no_j_idx, drop = FALSE]
        y_j <- y[, j_idx, drop = FALSE]

        # retrieve estimated values scaling JS
        lambda_js_scaling <- lambda_scaling[no_j_idx, f, drop = FALSE]

        # optimize the weights
        starting_weights <- rep(1 / j_1, times = j_1)
        w <- optim(
          par = starting_weights,
          fn = lav_cfa_jamesstein_rel,
          data = y_agg,
          resvars = theta[no_j_idx]
        )$par

        # make sure the weights sum up to 1
        w_optim <- w / sum(w)

        # compute aggregated indicator using the optimal weights
        y_agg_1 <- t(t(w_optim) %*% t(y_agg))

        # compute error variance of the aggregated indicator
        var_eps_agg <- drop(t(w_optim) %*%
          diag(theta[no_j_idx], nrow = length(no_j_idx)) %*% w_optim)

        # compute conditional expectation using aggregated indicator
        tmp <- lav_cfa_jamesstein_ce(
          y = y_agg_1, marker_idx = 1L,
          resvars_markers = var_eps_agg
        )
        ce_agg <- tmp / drop(w_optim %*% lambda_js_scaling)

        # compute factor loading
        fit <- lm(y_j ~ ce_agg)
        mm_lambda[j_idx, f] <- drop(coef(fit)[-1])

        # (optional) extract means
        # NU[j.idx] <- drop(coef(fit)[1,])
      } # j
    } # aggregate
  } # f

  list(lambda = mm_lambda, nu = mm_nu)
}

# internal function to be used inside lav_optim_noniter
# return 'x', the estimated vector of free parameters
lav_cfa_jamesstein_internal <- function(lavobject = NULL, # convenience
                                        # internal slot
                                        lavmodel = NULL,
                                        lavsamplestats = NULL,
                                        lavpartable = NULL,
                                        lavdata = NULL,
                                        lavoptions = NULL,
                                        theta_bounds = TRUE) {
  lavpta <- NULL
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))

    # extract slots
    lavmodel <- lavobject@Model
    lavsamplestats <- lavobject@SampleStats
    lavpartable <- lav_pt_set_cache(lavobject@ParTable, lavobject@pta)
    lavpta <- lavobject@pta
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
  }
  if (is.null(lavpta)) {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }

  # no structural part!
  if (any(lavpartable$op == "~")) {
    lav_msg_stop(gettext(
      "JS(A) estimator only available for CFA models (for now)"))
  }
  # no BETA matrix! (i.e., no higher-order factors)
  if (!is.null(lavmodel@GLIST$beta)) {
    lav_msg_stop(gettext(
      "JS(A) estimator not available for models that require a BETA matrix"))
  }
  # no std.lv = TRUE for now
  if (lavoptions$std.lv) {
    lav_msg_stop(gettext("JS(A) estimator not available if std.lv = TRUE"))
  }

  nblocks <- lav_pt_nblocks(lavpartable)
  stopifnot(nblocks == 1L) # for now
  b <- 1L
  sample_cov <- lavsamplestats@cov[[b]]
  nvar <- nrow(sample_cov)
  lv_names <- lavpta$vnames$lv.regular[[b]]
  nfac <- length(lv_names)
  marker_idx <- lavpta$vidx$lv.marker[[b]]
  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")
  lambda_nonzero_idx <- lavmodel@m.free.idx[[lambda_idx]]
  # only diagonal THETA for now...
  theta_idx <- which(names(lavmodel@GLIST) == "theta") # usually '2'
  m_theta <- lavmodel@m.free.idx[[theta_idx]]
  nondiag_idx <- m_theta[!m_theta %in% lav_mat_diag_idx(nvar)]
  if (length(nondiag_idx) > 0L) {
    lav_msg_warn(gettext(
      "this implementation of JS/JSA does not handle correlated residuals yet!"
      ))
  }


  # 1. obtain estimate for (diagonal elements of) THETA
  #    for now we use Spearman per factor
  m_b <- matrix(0, nvar, nfac)
  lambda_marker_idx <- (seq_len(nfac) - 1L) * nvar + marker_idx
  m_b[lambda_marker_idx] <- 1L
  m_b[lambda_nonzero_idx] <- 1L
  theta <- numeric(nvar)
  for (f in seq_len(nfac)) {
    ov_idx <- which(m_b[, f] == 1L)
    s_fac <- sample_cov[ov_idx, ov_idx, drop = FALSE]
    theta[ov_idx] <- lav_cfa_theta_spearman(s_fac, bounds = "wide")
  }
  mm_theta <- diag(theta, nrow = nvar)

  # 2. run James-Stein algorithm
  y <- lavdata@X[[1]] # raw data
  aggregated <- FALSE
  if (lavoptions$estimator == "JSA") {
    aggregated <- TRUE
  }
  out <- lav_cfa_jamesstein(
    s = sample_cov, y = y, marker_idx = marker_idx,
    lambda_nonzero_idx = lambda_nonzero_idx,
    theta = theta,
    # experimental
    theta_bounds = theta_bounds,
    #
    aggregated = aggregated
  )
  mm_lambda <- out$lambda

  # 3. PSI
  mm_psi <- lav_cfa_lambdatheta2psi(
    lambda = mm_lambda, theta = theta,
    s = sample_cov, mapping = "ML"
  )

  # store matrices in lavmodel@GLIST
  lavmodel@GLIST$lambda <- mm_lambda
  lavmodel@GLIST$theta <- mm_theta
  lavmodel@GLIST$psi <- mm_psi

  # extract free parameters only
  x <- lav_model_get_parameters(lavmodel)

  # apply bounds (if any)
  if (!is.null(lavpartable$lower)) {
    lower_x <- lavpartable$lower[lavpartable$free > 0]
    too_small_idx <- which(x < lower_x)
    if (length(too_small_idx) > 0L) {
      x[too_small_idx] <- lower_x[too_small_idx]
    }
  }
  if (!is.null(lavpartable$upper)) {
    upper_x <- lavpartable$upper[lavpartable$free > 0]
    too_large_idx <- which(x > upper_x)
    if (length(too_large_idx) > 0L) {
      x[too_large_idx] <- upper_x[too_large_idx]
    }
  }

  x
}


# Conditional expectation (Section 2.1, eq. 10)
lav_cfa_jamesstein_ce <- function(y = NULL,
                                  marker_idx = NULL,
                                  resvars_markers = NULL) {
  y <- as.matrix(y)

  # sample size
  n <- nrow(y)
  n1 <- n - 1
  n3 <- n - 3

  # markers only
  y_marker <- y[, marker_idx, drop = FALSE]

  # means and variances
  mean_1 <- colMeans(y_marker, na.rm = TRUE)
  var_1 <- apply(y_marker, 2, var, na.rm = TRUE)

  # 1 - R per marker
  onemin_r <- n3 * resvars_markers / (n1 * var_1)

  # R per marker
  r <- 1 - onemin_r

  # create E(\eta | Y)
  e_eta_cond_y <- t(t(y_marker) * r + onemin_r * mean_1)

  e_eta_cond_y
}

# Reliability function used to obtain the weights (Section 4, Aggregation)
lav_cfa_jamesstein_rel <- function(w = NULL, data = NULL, resvars = NULL) {
  # construct weight vector
  w <- matrix(w, ncol = 1)

  # construct aggregated indicator: y_agg = t(w) %*% y_i
  y_agg <- t(t(w) %*% t(data))

  # calculate variance of aggregated indicator
  var_y_agg <- var(y_agg)

  # calculate error variance of the aggregated indicator
  var_eps_agg <- t(w) %*% diag(resvars) %*% w

  # reliability function to be maximized
  rel <- (var_y_agg - var_eps_agg) %*% solve(var_y_agg)

  # return value
  -rel
}

# residual diagnostics

# two types:
# 1) residuals for summary statistics
# 2) case-wise residuals

# this (new) version written around Aug/Sept 2018 for 0.6-3
# - based on obsList (inspect_sampstat) and estList (inspect_implied)
# - pre-scaling for type = "cor.bollen" and type = "cor.bentler"
# - summary statistics: rmr, srmr, crmr, urmr, usrmr, ucrmr; standard errors,
#       confidence intervals (for u(cs)rmr),
#       z-statistics (exact test, close test), p-values
# - type = "normalized" is based on lav_model_h1_acov(), and should now work
#   for all estimators
# - type = "standardized" now uses the correct formula, and should work for
#   for all estimators
# - type = "standardized.mplus" uses the simplified Mplus/LISREL version,
#   often resulting in NAs due to negative var(resid) estimates
#   (this was "standardized" in lavaan < 0.6.3

# conditional.x = TRUE is supported (both categorical and continuous): the
# moment vector then includes a regression-slopes block. Slopes are not
# rescaled in the cor metric (they are not rescaled in the residual point
# estimate either).
#
# Multilevel (twolevel): residual SEs/z-statistics and the (u)rmr/(u)srmr/
# (u)crmr summaries are supported for a single group (raw, cor.bentler and
# cor.bollen). The moment vector stacks the level-specific blocks, each ordered
# [mean, vech(cov)]; the within-level means are fixed at zero (degenerate
# moments) and get a zero SE, and the cor standardization is a block-diagonal
# jacobian over the levels. Only multigroup + multilevel is not ready yet.

# - change 0.6-6: we enforce observed.information = "h1" to ensure 'Q' is a
#                 projection matrix (see lav_residuals_acov)

# - change 0.6-13: fixed.x = TRUE is ignored (to conform with 'tradition')

# - change 0.7-1: two multigroup ACOV fixes: (1) the stored @NACOV is scaled
#                  by 1/n_g (not 1/ntotal) to obtain ACOV(s_g), and (2) the
#                  standardized-residual ACOV is computed from the full
#                  stacked Q over all groups, so that under across-group
#                  equality constraints the Var(sigma-hat) contributions of
#                  the other groups (and, for the combined summary table, the
#                  cross-group residual covariance) are no longer dropped.

# - change 0.7-1: refactor (issue #291): ACOV.obs (= the expensive ACOV of
#                  the observed h1 sample statistics) is computed at most once
#                  per call and reused for both the residual SEs and the
#                  summary statistics SEs (see lav_residuals_acov_obs())
# - change 0.7-1: the (undocumented, never-functional) custom.rmr argument
#                  has been removed
# - change 0.7-1: residual SEs/z-statistics and the (u)rmr/(u)srmr/(u)crmr
#                  summaries are now available for the categorical case,
#                  including mixed (continuous + ordinal), and for both
#                  categorical and continuous conditional.x (with a
#                  regression-slopes block). The cor.bentler/cor.bollen
#                  standardization uses a cov->cor jacobian built for the
#                  categorical moment ordering (see
#                  lav_residuals_cor_jacobian_cat()).
# - change 0.7-1: residual SEs and the (u)rmr/(u)srmr/(u)crmr summaries are now
#                  available for (single-group) multilevel models, with per-
#                  block (within/between) summaries; the within-level means are
#                  degenerate (fixed at zero) and lav_model_h1_acov() now drops
#                  such zero-information moments before inverting (reinserting
#                  zeros) instead of returning NaN; the cor standardization is a
#                  block-diagonal jacobian (lav_residuals_cor_jacobian_ml()).
# - change 0.7-1: new h1= argument (lavResiduals) to supply a user-provided
#                  saturated model (a fitted lavaan object, a lavh1 list, or a
#                  moments list); it is swapped into object@h1 and used as the
#                  observed summary statistics (see lav_residuals_resolve_h1())
# - change 0.7-1: conditional.x (both continuous and categorical): the
#                  regression slopes are now standardized (b * s.x / s.y; for
#                  categorical endogenous variables s.y = 1) in the
#                  cor.bentler/cor.bollen metric. The continuous $summary table
#                  reports res.cov / res.int / res.slopes / total (instead of
#                  cov / mean / total); the categorical $summary table reports
#                  res.cov / res.th / res.slopes / total (instead of cor /
#                  thresholds / total). The 'total' (and hence the
#                  srmr/rmr/crmr fitMeasures totals) now includes the slopes;
#                  fitMeasures reads the res.cov column.
# - change 0.7-1: output = "table" now lists *all* residual elements (not just
#                  the off-diagonal covariances): (co)variances, means/inter-
#                  cepts, thresholds and -- for conditional.x -- the regression
#                  slopes, with optional se/z columns (see
#                  lav_residuals_table_block()); structurally fixed moments
#                  (zero/NA se) are filtered out.
# - change 0.7-1: new output = "text" (alias "pretty"): prints (a) a table with
#                  the n_largest largest residuals per element group/block (in
#                  the parameterEstimates(output="text") style, with a Residual
#                  column and optional Std.Err/Z-value columns; the header
#                  reflects the residual type) and (b) the residual summary
#                  table(s) in the fit-measures style (see
#                  lav_residuals_summary_print()).
# - change 0.7-1: new usrmr_ci_level (default 0.90) and usrmr_close_h0 (default
#                  0.05) arguments to set the confidence level of the reported
#                  CI and the H0 value of the close-fit test for the unbiased
#                  urmr/usrmr/ucrmr (cf. the rmsea.* arguments in fitMeasures).
# - change 0.7-1: new elementwise= argument (default TRUE). If FALSE, the
#                  element-wise residuals are omitted (for both output = "list"
#                  and output = "text") and only the summary information is
#                  returned/shown. summary = FALSE now also suppresses the
#                  summary in the text output (it is no longer force-enabled).
# - change 0.7-1: new combine= argument. When TRUE (and there
#                  are multiple groups or levels), the per-block summary tables
#                  are replaced by a single overall table that pools the
#                  residual elements across all blocks. The per-block and the
#                  combined tables share one code path (column "specs"); the
#                  combined ACOV is the joint (stacked) ACOV across all blocks:
#                  across the (correlated) levels of a multilevel model, and
#                  across groups -- where the residuals are correlated whenever
#                  across-group equality constraints couple theta-hat.

setMethod(
  "residuals", "lavaan",
  function(object, type = "raw", labels = TRUE, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("residuals"))
        )
      }
    }
    # lowercase type
    type <- tolower(type)

    # type = "casewise"
    if (type %in% c("casewise", "case", "obs", "observations", "ov")) {
      return(lav_residuals_casewise(object, labels = labels))
    } else {
      out <- lav_residuals(
        object = object, type = type, h1 = TRUE,
        add_type = TRUE,
        rename_cov_cor = FALSE, # should become FALSE!
        # after packages (eg jmv)
        # have adapted 0.6-3 style
        add_labels = labels, add_class = TRUE,
        drop_list_single_group = TRUE
      )
    }

    out
  }
)

setMethod(
  "resid", "lavaan",
  function(object, type = "raw", ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("resid"))
        )
      }
    }
    residuals(object, type = type, labels = TRUE)
  }
)


# turn one block's residual list (as returned by lav_residuals()) into a long
# data.frame, with one row per residual element-cell: covariances/variances
# (v1~~v2), intercepts/means (v1~1), thresholds (v1|t1) and -- for conditional.x
# -- the regression slopes (endo~exo). The exogenous moments (cov.x, mean.x) are
# conditioned on and therefore omitted. Each row also carries a 'group' label
# (Covariances / Variances / Intercepts / Thresholds / Regressions) and its
# lhs/op/rhs components, used by the output = "text" largest-residuals tables.
# Structurally fixed moments (zero or unavailable se) are dropped. If the block
# list carries '.se'/'.z' companions (se = TRUE / zstat = TRUE), the 'se'/'z'
# columns are filled; otherwise they are NA.
lav_residuals_table_rows <- function(block_list, cov_diagonal = TRUE) {
  sym_names   <- c("cov", "res.cov")     # symmetric (co)variance matrices
  int_names   <- c("mean", "res.int")    # means / intercepts
  th_names    <- c("th", "res.th")       # thresholds
  slope_names <- c("res.slopes")         # regression slopes (endo x exo)
  skip_names  <- c("type", "summary", "cov.x", "mean.x")

  base_names <- setdiff(names(block_list), skip_names)
  base_names <- base_names[!grepl("\\.(se|z)$", base_names)]

  parts <- list()
  for (nm in base_names) {
    res_el <- block_list[[nm]]
    se_el  <- block_list[[paste0(nm, ".se")]]
    z_el   <- block_list[[paste0(nm, ".z")]]
    if (is.null(res_el)) next

    if (nm %in% sym_names) {
      p <- nrow(res_el)
      vidx <- lav_mat_vech_idx(p, diagonal = cov_diagonal)
      rc <- arrayInd(vidx, c(p, p))
      rn <- rownames(res_el)
      if (is.null(rn)) rn <- as.character(seq_len(p))
      lhs <- rn[rc[, 1]]
      rhs <- rn[rc[, 2]]
      op <- "~~"
      grp <- ifelse(rc[, 1] == rc[, 2], "Variances", "Covariances")
      name <- paste0(lhs, "~~", rhs)
      sel <- function(x) x[vidx]
    } else if (nm %in% slope_names) {
      rn <- rownames(res_el)
      cn <- colnames(res_el)
      if (is.null(rn)) rn <- as.character(seq_len(nrow(res_el)))
      if (is.null(cn)) cn <- as.character(seq_len(ncol(res_el)))
      grid <- expand.grid(r = seq_len(nrow(res_el)), c = seq_len(ncol(res_el)))
      lhs <- rn[grid$r]
      rhs <- cn[grid$c]
      op <- "~"
      grp <- "Regressions"
      name <- paste0(lhs, "~", rhs)
      sel <- function(x) as.vector(x)
    } else if (nm %in% int_names) {
      vn <- names(res_el)
      if (is.null(vn)) vn <- as.character(seq_along(res_el))
      lhs <- vn
      rhs <- ""
      op <- "~1"
      grp <- "Intercepts"
      name <- paste0(vn, "~1")
      sel <- function(x) as.vector(x)
    } else if (nm %in% th_names) {
      vn <- names(res_el)
      if (is.null(vn)) vn <- as.character(seq_along(res_el))
      lhs <- vn # already in "y1|t1" form
      rhs <- ""
      op <- ""
      grp <- "Thresholds"
      name <- vn
      sel <- function(x) as.vector(x)
    } else {
      next # unknown element type: skip
    }

    n_el <- length(name)
    parts[[nm]] <- data.frame(
      group = grp, lhs = lhs, op = op, rhs = rhs, name = name,
      res = sel(res_el),
      se = if (!is.null(se_el)) sel(se_el) else rep(NA_real_, n_el),
      z = if (!is.null(z_el)) sel(z_el) else rep(NA_real_, n_el),
      stringsAsFactors = FALSE
    )
  }

  if (length(parts) == 0L) {
    rows <- data.frame(
      group = character(0), lhs = character(0), op = character(0),
      rhs = character(0), name = character(0), res = numeric(0),
      se = numeric(0), z = numeric(0), stringsAsFactors = FALSE
    )
  } else {
    rows <- do.call(rbind, parts)
  }

  # drop rows for structurally fixed moments (a saturated mean or just-
  # identified threshold, the intercept of an ordinal variable, or an exactly
  # reproduced variance): these are not estimable residuals. The reliable
  # signal is a zero or unavailable standard error -- the point estimate itself
  # is only numerically (not exactly) zero. When no standard errors are
  # available (e.g. the gated multigroup + multilevel case), fall back to the
  # residual magnitude.
  if (any(is.finite(rows$se))) {
    keep <- is.finite(rows$se) & abs(rows$se) > 1e-08
  } else {
    keep <- is.finite(rows$res) & abs(rows$res) > 1e-06
  }
  rows <- rows[keep, , drop = FALSE]
  rownames(rows) <- NULL

  rows
}

# the flat (name / res / [se] / [z]) data.frame used by output = "table"
lav_residuals_table_block <- function(block_list, se = FALSE, zstat = FALSE,
                                      cov_diagonal = TRUE) {
  rows <- lav_residuals_table_rows(block_list, cov_diagonal = cov_diagonal)
  tab <- rows[, c("name", "res", "se", "z"), drop = FALSE]
  rownames(tab) <- NULL

  # drop the se/z columns if not requested
  if (!se) tab$se <- NULL
  if (!zstat) tab$z <- NULL

  tab
}

# for one block: the 'n_largest' residuals (in absolute value) per element
# group (Covariances, Variances, Intercepts, Thresholds, Regressions). Returns
# a data.frame with the group order preserved (used by output = "text").
lav_residuals_largest_block <- function(block_list, cov_diagonal = TRUE,
                                        n_largest = 5L) {
  rows <- lav_residuals_table_rows(block_list, cov_diagonal = cov_diagonal)
  if (nrow(rows) == 0L) {
    return(rows)
  }
  group_order <- c(
    "Covariances", "Variances", "Intercepts", "Thresholds", "Regressions"
  )
  rows$group <- factor(rows$group, levels = group_order)
  parts <- lapply(split(rows, rows$group, drop = TRUE), function(g) {
    g <- g[order(abs(g$res), decreasing = TRUE), , drop = FALSE]
    if (n_largest > 0L && n_largest < nrow(g)) {
      g <- g[seq_len(n_largest), , drop = FALSE]
    }
    g
  })
  out <- do.call(rbind, parts)
  out$group <- as.character(out$group)
  rownames(out) <- NULL
  out
}

# section title for the largest-residuals tables, reflecting the residual type
lav_residuals_largest_title <- function(type) {
  t <- tolower(type)[1]
  if (t %in% c("raw", "rmr")) {
    return("Largest Raw Residuals")
  }
  if (t == "normalized") {
    return("Largest Normalized Residuals")
  }
  if (t == "standardized") {
    return("Largest Standardized Residuals")
  }
  if (t == "standardized.mplus") {
    return("Largest Standardized Residuals (Mplus)")
  }
  pretty <- switch(t,
    cor = "Cor.Bollen", cor.bentler = "Cor.Bentler", cor.eqs = "Cor.Bentler",
    cor.bollen = "Cor.Bollen", srmr = "Cor.Bentler", crmr = "Cor.Bollen",
    t
  )
  paste0("Largest Standardized Residuals (", pretty, ")")
}


# user-visible function
lavResiduals <- function(object, type = "cor.bentler", h1 = NULL,         # nolint
                         se = FALSE, zstat = TRUE,
                         summary = TRUE, elementwise = TRUE,
                         combine = FALSE,
                         usrmr_ci_level = 0.90, usrmr_close_h0 = 0.05,
                         h1_acov = "unstructured",
                         add_type = TRUE, add_labels = TRUE, add_class = TRUE,
                         drop_list_single_group = TRUE,
                         maximum_number = 0L, n_largest = 5L,
                         output = "list",
                        ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)

  # sam objects (issue #517)
  if (!is.null(object@internal$sam.method)) {
    if (lav_sam_local_flag(object)) {
      # delegate to the stored step-2 structural fit: the residuals of a
      # local SAM model are the residuals of the structural part (the
      # estimated latent moments vs the structural model)
      fit_pa <- lav_sam_struc_object(object)
      if (is.null(fit_pa)) {
        lav_msg_stop(gettext(
          "no stored structural fit found: this sam object was created by an
           older version of lavaan; please rerun sam()."))
      }
      lav_msg_note(gettext(
        "the residuals are computed for the structural part only (the
         estimated latent moments), conditional on the (fixed) measurement
         model of step 1."))
      object <- fit_pa
    } else {
      lav_msg_stop(gettext(
        "lavResiduals() is not available if sam.method = \"global\": the
         standard errors of the residuals require an h1 ACOV that a two-step
         (sam) procedure does not provide; use residuals() to inspect the
         raw residuals."))
    }
  }

  # normalize output argument ("pretty" is an alias for "text")
  output <- tolower(output)[1]
  if (output == "pretty") {
    output <- "text"
  }
  if (!output %in% c("list", "table", "text")) {
    lav_msg_stop(gettextf(
      "output= should be one of %s.",
      lav_msg_view(c("list", "table", "text"), "or")
    ))
  }

  # for the table and text outputs we always need the standard errors
  # internally: they tell apart estimable residuals (finite, non-zero se) from
  # structurally fixed moments (zero/NA se), which are filtered out. The summary
  # statistics are not used by the table, so they are skipped there. The
  # 'summary' and 'elementwise' arguments are honored for both list and text
  # output (summary = FALSE -> no summary; elementwise = FALSE -> only summary).
  table_output <- (output == "table")
  text_output <- (output == "text")
  out <- lav_residuals(
    object = object, type = type, h1 = h1,
    se = se || table_output || (text_output && elementwise), zstat = zstat,
    summary = summary && !table_output,
    summary_options = list(
      se = TRUE, zstat = TRUE, pvalue = TRUE,
      unbiased = TRUE, unbiased_se = TRUE, unbiased_ci = TRUE,
      unbiased_ci_level = usrmr_ci_level, unbiased_zstat = TRUE,
      unbiased_test_val = usrmr_close_h0, unbiased_pvalue = TRUE
    ),
    h1_acov = h1_acov, add_type = add_type,
    add_labels = add_labels, add_class = add_class,
    drop_list_single_group = drop_list_single_group,
    summary_blocks_combined = combine
  )

  if (output == "table") {
    # variance (diagonal) residuals are informative for the raw-style metrics
    # but structurally zero in the correlation metrics, so we only list them
    # for the former
    cov_diagonal <- tolower(type)[1] %in%
      c("raw", "rmr", "normalized", "standardized", "standardized.mplus")

    # one table per block (single block -> the list is not nested)
    nblocks <- lav_pt_nblocks(object@ParTable)
    out_list <- vector("list", length = nblocks)
    for (block in seq_len(nblocks)) {
      block_list <- if (nblocks == 1L) out else out[[block]]
      tab <- lav_residuals_table_block(block_list,
        se = se, zstat = zstat, cov_diagonal = cov_diagonal
      )

      # round the numeric columns
      num_cols <- intersect(c("res", "se", "z"), names(tab))
      tab[num_cols] <- lapply(tab[num_cols], round, digits = 3)

      # sort by absolute residual (largest first)
      tab <- tab[order(abs(tab$res), decreasing = TRUE), , drop = FALSE]
      rownames(tab) <- NULL

      # show first rows only (maximum_number == 0 -> all)
      maximum_number <- maximum_number
      if (maximum_number > 0L && maximum_number < nrow(tab)) {
        tab <- tab[seq_len(maximum_number), , drop = FALSE]
      }

      if (add_class) {
        class(tab) <- c("lavaan.data.frame", "data.frame")
      }
      out_list[[block]] <- tab
    }
    if (nblocks == 1L) {
      out <- out_list[[1]]
    } else {
      out <- out_list
      names(out) <- object@Data@block.label
    }
  } else if (output == "text") {
    out <- lav_residuals_list_to_text(out,
      type = type, nblocks = lav_pt_nblocks(object@ParTable),
      drop_single = drop_list_single_group, n_largest = n_largest,
      show_se = isTRUE(se), show_z = isTRUE(zstat), ci_level = usrmr_ci_level,
      combine = isTRUE(combine), do_summary = isTRUE(summary),
      do_largest = isTRUE(elementwise)
    )
  } else if (!elementwise) {
    # output = "list" but only the summary information is requested: return the
    # summary table(s), dropping the (element-wise) residual matrices
    nblocks <- lav_pt_nblocks(object@ParTable)
    if (nblocks == 1L) {
      blk <- if (drop_list_single_group) out else out[[1]]
      out <- blk[["summary"]]
    } else if (!is.null(out[["summary"]])) {
      out <- out[["summary"]] # combine = TRUE: a single (pooled) summary
    } else {
      block_names <- setdiff(names(out), "summary")
      sm <- lapply(out[block_names], function(b) b[["summary"]])
      names(sm) <- block_names
      out <- sm
    }
  } else {
    # output = "list", elementwise = TRUE -> nothing to do
  }

  out
}

# turn a lavResiduals(output = "list") result into the 'lavaan.residuals.summary'
# object that is printed in the lavaan style (largest residuals + summary
# tables). Factored out so both lavResiduals(output = "text") and the residuals
# section of summary() (lav_object_summary / lav_summary_print) can use it.
lav_residuals_list_to_text <- function(out, type, nblocks, drop_single = TRUE,
                                       n_largest = 5L, show_se = FALSE,
                                       show_z = TRUE, ci_level = 0.90,
                                       combine = FALSE, do_summary = TRUE,
                                       do_largest = TRUE) {
  # per-block residual lists
  if (nblocks == 1L) {
    blk <- if (drop_single) out else out[[1]]
    block_lists <- list(blk)
    block_labels <- ""
  } else {
    block_names <- setdiff(names(out), "summary")
    block_lists <- out[block_names]
    block_labels <- block_names
  }

  # summary table(s) -- some estimators (e.g. IV) do not produce residual
  # summary statistics, so the per-block 'summary' element may be missing;
  # keep summary_tables NULL in that case so nothing is printed
  summary_tables <- NULL
  if (do_summary) {
    if (nblocks == 1L) {
      summary_tables <- list(block_lists[[1]][["summary"]])
      names(summary_tables) <- ""
    } else if (!is.null(out[["summary"]])) {
      summary_tables <- list(out[["summary"]]) # combine = TRUE: one summary
      names(summary_tables) <- ""
    } else {
      summary_tables <- lapply(block_lists, function(b) b[["summary"]])
      names(summary_tables) <- block_labels
    }
    # drop blocks without a summary table; if none remain, store NULL
    keep <- !vapply(summary_tables, is.null, logical(1))
    summary_tables <- if (any(keep)) summary_tables[keep] else NULL
  }

  # largest residuals per block (variances only for raw-style metrics)
  largest <- NULL
  if (do_largest) {
    cov_diagonal <- tolower(type)[1] %in%
      c("raw", "rmr", "normalized", "standardized", "standardized.mplus")
    largest <- lapply(block_lists, function(b) {
      lav_residuals_largest_block(b, cov_diagonal = cov_diagonal,
        n_largest = n_largest
      )
    })
    names(largest) <- block_labels
  }

  structure(
    list(
      type = type, largest = largest, summary_tables = summary_tables,
      show_se = isTRUE(show_se), show_z = isTRUE(show_z), ci_level = ci_level,
      combine = isTRUE(combine) && nblocks > 1L
    ),
    class = "lavaan.residuals.summary"
  )
}

# pretty-print a single residual summary table (one block), in the lavaan
# fit-measures style: a left-justified label column followed by one right-
# justified value column per moment block (e.g. cov / mean / total). The
# (biased) point estimate and its exact-fit test are separated from the
# (unbiased) estimate, its confidence interval and its close-fit test by a
# blank line, mirroring the RMSEA / Robust RMSEA layout.
lav_residuals_summary_print_one <- function(tab, metric, nd = 3L,
                                            ci_level = 0.90) {
  num_format <- paste0("%", max(8L, nd + 5L), ".", nd, "f")
  big_m <- toupper(metric)
  rn <- rownames(tab)
  cn <- colnames(tab)
  ci_pct <- format(round(ci_level * 100, 1))

  # capitalized column headers (text output only), following the style of
  # parameterEstimates(., output = "text")
  col_disp_map <- c(
    cov = "Cov", mean = "Mean", total = "Total", cor = "Cor",
    thresholds = "Th", res.cov = "Res.cov", res.int = "Res.int",
    res.th = "Res.th", res.slopes = "Res.slps"
  )
  cn_disp <- col_disp_map[cn]
  cn_disp[is.na(cn_disp)] <- cn[is.na(cn_disp)]
  cn_disp <- unname(cn_disp)

  # nice row labels
  label_map <- c(
    "Standard error",
    "Test statistic (exact fit, H0: value = 0)",
    "P-value (one-sided)",
    "Standard error",
    paste0(ci_pct, " Percent confidence interval - lower"),
    paste0(ci_pct, " Percent confidence interval - upper"),
    "H0 value (close fit)",
    "Test statistic (close fit)",
    "P-value (one-sided)"
  )
  names(label_map) <- c(
    paste0(metric, ".se"),
    paste0(metric, ".exactfit.z"),
    paste0(metric, ".exactfit.pvalue"),
    paste0("u", metric, ".se"),
    paste0("u", metric, ".ci.lower"),
    paste0("u", metric, ".ci.upper"),
    paste0("u", metric, ".closefit.h0.value"),
    paste0("u", metric, ".closefit.z"),
    paste0("u", metric, ".closefit.pvalue")
  )
  nice <- function(r) {
    if (r == metric) {
      return(big_m)
    }
    if (r == paste0("u", metric)) {
      return(paste0("Unbiased ", big_m, " (U", big_m, ")"))
    }
    if (!is.na(label_map[r])) {
      return(unname(label_map[r]))
    }
    r
  }
  labels <- vapply(rn, nice, character(1))

  # formatted value matrix (NA -> blank); the column width must also fit the
  # widest (display) column name. The extra +2 (rather than +1) shifts the
  # value column one space further right, so the digits line up with the
  # fit-measure values when printed by summary().
  val_w <- max(8L, nd + 5L, max(nchar(cn_disp))) + 2L
  vm <- matrix("", nrow(tab), length(cn))
  for (j in seq_along(cn)) {
    v <- tab[[cn[j]]]
    vm[, j] <- ifelse(is.na(v), "", sprintf(num_format, v))
  }

  lab_w <- 44L
  fmt_lab <- function(s) formatC(paste0("  ", s), width = lab_w, flag = "-")
  fmt_val <- function(s) formatC(s, width = val_w)

  # header (column names), then the rows, with a blank line before the unbiased
  # block
  ub_start <- match(paste0("u", metric), rn)
  lines <- paste0(
    fmt_lab(""), paste(vapply(cn_disp, fmt_val, character(1)), collapse = "")
  )
  for (i in seq_len(nrow(tab))) {
    if (!is.na(ub_start) && i == ub_start) {
      lines <- c(lines, "")
    }
    lines <- c(lines, paste0(
      fmt_lab(labels[i]),
      paste(vapply(seq_along(cn), function(j) fmt_val(vm[i, j]), character(1)),
        collapse = ""
      )
    ))
  }
  # terminate each line; the next section adds its own single leading blank line
  cat(paste0(lines, "\n"), sep = "")
}

# pretty-print a single block's largest-residuals table, in the style of
# parameterEstimates(., output = "text"): residual elements grouped into
# sections (Covariances, Variances, Intercepts, Thresholds, Regressions), with
# a left-justified "lhs op rhs" label and right-justified Residual / Std.Err /
# Z-value columns.
lav_residuals_largest_print_one <- function(df, show_se = FALSE,
                                            show_z = TRUE, nd = 3L) {
  if (nrow(df) == 0L) {
    cat("\n  (no non-zero residuals)\n")
    return(invisible())
  }
  num_format <- paste0("%", max(8L, nd + 5L), ".", nd, "f")
  val_w <- max(8L, nd + 5L) + 1L
  fmt_val <- function(s) formatC(s, width = val_w)
  fmt_num <- function(v) ifelse(is.na(v), "", sprintf(num_format, v))

  # "lhs op rhs" label (thresholds already carry their full "y1|t1" name)
  lab <- ifelse(df$op == "", df$lhs,
    ifelse(nzchar(df$rhs), paste(df$lhs, df$op, df$rhs), paste(df$lhs, df$op))
  )
  lab_w <- max(nchar(lab), 8L) + 4L
  fmt_lab <- function(s) formatC(paste0("  ", s), width = lab_w, flag = "-")

  col_hdr <- c("Residual", if (show_se) "Std.Err", if (show_z) "Z-value")
  lines <- paste0(
    fmt_lab(""), paste(vapply(col_hdr, fmt_val, character(1)), collapse = "")
  )
  group_order <- c(
    "Covariances", "Variances", "Intercepts", "Thresholds", "Regressions"
  )
  groups <- intersect(group_order, unique(df$group))
  for (g_idx in seq_along(groups)) {
    grp <- groups[g_idx]
    if (g_idx > 1L) {
      lines <- c(lines, "") # blank line between sections
    }
    lines <- c(lines, paste0(grp, ":"))
    for (i in which(df$group == grp)) {
      vals <- c(
        fmt_num(df$res[i]),
        if (show_se) fmt_num(df$se[i]),
        if (show_z) fmt_num(df$z[i])
      )
      lines <- c(lines, paste0(
        fmt_lab(lab[i]),
        paste(vapply(vals, fmt_val, character(1)), collapse = "")
      ))
    }
  }
  # shift the whole table two spaces to the right (blank lines stay empty)
  lines <- ifelse(nzchar(lines), paste0("  ", lines), lines)
  # terminate each line; the next section adds its own single leading blank line
  cat(paste0(lines, "\n"), sep = "")
}

# print method for the output = "text" result of lavResiduals(): first the
# largest residuals per block, then the residual summary table(s)
lav_residuals_summary_print <- function(x, ..., nd = 3L) {
  # ---- largest residuals (per block) ----
  if (!is.null(x$largest)) {
    cat("\n", lav_residuals_largest_title(x$type), ":\n", sep = "")
    multi <- length(x$largest) > 1L
    for (b in seq_along(x$largest)) {
      lab <- names(x$largest)[b]
      if (multi && !is.null(lab) && nzchar(lab)) {
        cat("\n  [ ", lab, " ]\n", sep = "")
      }
      cat("\n")
      lav_residuals_largest_print_one(x$largest[[b]],
        show_se = isTRUE(x$show_se), show_z = isTRUE(x$show_z), nd = nd
      )
    }
  }

  # ---- residual summary table(s) ----
  tables <- x$summary_tables
  if (length(tables) > 0L && !is.null(rownames(tables[[1]]))) {
    metric <- rownames(tables[[1]])[1] # "rmr" / "srmr" / "crmr"
    full_name <- switch(metric,
      rmr = "Root Mean Square Residual (RMR)",
      srmr = "Standardized Root Mean Square Residual (SRMR)",
      crmr = "Correlation Root Mean Square Residual (CRMR)",
      "Residual-based summary"
    )
    header <- full_name
    if (isTRUE(x$combine)) {
      header <- paste0(full_name, " (pooled across blocks)")
    }
    cat("\n", header, ":\n", sep = "")

    for (b in seq_along(tables)) {
      lab <- names(tables)[b]
      if (!is.null(lab) && nzchar(lab)) {
        cat("\n  [ ", lab, " ]\n", sep = "")
      }
      cat("\n")
      lav_residuals_summary_print_one(tables[[b]], metric, nd = nd,
        ci_level = if (is.null(x$ci_level)) 0.90 else x$ci_level
      )
    }
  }

  invisible(x)
}

# summary_options with everything switched off; used for the settings where
# we only report point estimates (no SE, no z, no unbiased statistics)
lav_residuals_summary_options_off <- function() {
  list(
    se = FALSE, zstat = FALSE,
    pvalue = FALSE, unbiased = FALSE,
    unbiased_se = FALSE,
    unbiased_ci = FALSE, unbiased_ci_level = 0.90,
    unbiased_zstat = FALSE, unbiased_test_val = 0.05,
    unbiased_pvalue = FALSE
  )
}

# resolve a user-provided h1= argument to a 'lavh1' list (list with an
# $implied element), so it can be swapped into object@h1 and used as the
# observed (saturated) sample statistics. Accepts:
#  - a fitted lavaan object       -> its @h1 slot
#  - a lavh1 list (has $implied)  -> as-is
#  - an implied-moments list      -> wrapped as list(implied = .)
# Returns NULL when no user override is requested (h1 is NULL or logical).
lav_residuals_resolve_h1 <- function(object, h1 = NULL) {
  if (is.null(h1) || is.logical(h1)) {
    return(NULL)
  }

  if (inherits(h1, "lavaan")) {
    lavh1 <- h1@h1
    if (length(lavh1) == 0L || is.null(lavh1$implied)) {
      lav_msg_stop(gettext(
        "the h1= lavaan object has no (non-empty) h1 slot; refit it with the
         default option h1 = TRUE"))
    }
  } else if (is.list(h1)) {
    if (!is.null(h1$implied)) {
      lavh1 <- h1 # already a lavh1 list
    } else if (!is.null(h1$cov) || !is.null(h1$res.cov)) {
      lavh1 <- list(implied = h1) # an 'implied' moments list
    } else {
      lav_msg_stop(gettext(
        "the h1= list must be a lavh1 list (with an 'implied' element) or an
         implied-moments list (with a 'cov' or 'res.cov' element)"))
    }
  } else {
    lav_msg_stop(gettext(
      "h1= must be a fitted lavaan object, a lavh1 list, or a moments list"))
  }

  # check that the saturated moments are compatible with the fitted object
  cov_nm <- if (object@Model@conditional.x) "res.cov" else "cov"
  usr_cov <- lavh1$implied[[cov_nm]]
  if (is.null(usr_cov) || !is.list(usr_cov)) {
    lav_msg_stop(gettextf(
      "the h1= implied moments must contain a (per-block) list element %s",
      dQuote(cov_nm)))
  }
  ref_cov <- object@h1$implied[[cov_nm]]
  if (!is.null(ref_cov)) {
    if (length(usr_cov) != length(ref_cov)) {
      lav_msg_stop(gettext(
        "the h1= saturated model has a different number of blocks/groups than
         the fitted object"))
    }
    for (b in seq_along(ref_cov)) {
      if (!all(dim(as.matrix(usr_cov[[b]])) == dim(as.matrix(ref_cov[[b]])))) {
        lav_msg_stop(gettextf(
          "the h1= saturated model and the fitted object have incompatible
           dimensions in block %d", b))
      }
    }
  }

  lavh1
}

# main function
lav_residuals <- function(object, type = "raw", h1 = TRUE,
                          se = FALSE, zstat = FALSE, summary = FALSE,
                          summary_options = list(
                            se = TRUE, zstat = TRUE,
                            pvalue = TRUE, unbiased = TRUE, unbiased_se = TRUE,
                            unbiased_ci = TRUE, unbiased_ci_level = 0.90,
                            unbiased_zstat = FALSE, unbiased_test_val = 0.05,
                            unbiased_pvalue = FALSE
                          ),
                          h1_acov = "unstructured", add_type = FALSE,
                          rename_cov_cor = FALSE,
                          add_labels = FALSE, add_class = FALSE,
                          drop_list_single_group = FALSE,
                          summary_blocks_combined = FALSE) {
  # check object
  object <- lav_object_check_version(object)

  # user-provided saturated model (h1=)? if so, swap it into object@h1 so the
  # observed (saturated) sample statistics -- and their ACOV -- are taken from
  # it; the model-implied moments still come from 'object'
  if (is.null(h1)) {
    h1 <- TRUE
  } else if (is.logical(h1)) {
    # logical TRUE/FALSE: use object's own saturated model (keep as-is)
  } else if (inherits(h1, "lavaan") || is.list(h1)) {
    object@h1 <- lav_residuals_resolve_h1(object, h1)
    h1 <- TRUE
  } else {
    lav_msg_stop(gettext(
      "h1= must be NULL, logical, a fitted lavaan object, a lavh1 list, or a
       moments list"))
  }

  # working version summary options
  summary_options_1 <- summary_options

  # type
  type <- tolower(type)[1]

  # check type
  if (!type %in% c(
    "raw", "cor", "cor.bollen", "cor.bentler", "cor.eqs",
    "rmr", "srmr", "crmr",
    "normalized", "standardized", "standardized.mplus"
  )) {
    lav_msg_stop(gettext("unknown argument for type:"), dQuote(type))
  }

  # if cor, choose 'default'
  if (type == "cor") {
    if (object@Options$mimic == "EQS") {
      type <- "cor.bentler"
    } else {
      type <- "cor.bollen"
    }
  }
  if (type == "cor.eqs") {
    type <- "cor.bentler"
  }
  if (type == "rmr") {
    type <- "raw"
  }
  if (type == "srmr") {
    type <- "cor.bentler"
  }
  if (type == "crmr") {
    type <- "cor.bollen"
  }

  # slots
  lavdata <- object@Data
  lavmodel <- object@Model

  # change options if multilevel (for now)
  if (lavdata@nlevels > 1L) {
    # per-element residual SEs are supported for a single group (all types);
    # multigroup + multilevel is not ready yet
    if (lavdata@ngroups > 1L) {
      zstat <- se <- FALSE
      summary <- FALSE
      summary_options_1 <- lav_residuals_summary_options_off()
    }
  }

  # residual SEs/summaries are supported for the unconditional case (pure
  # categorical, continuous, and mixed continuous + ordinal) as well as the
  # conditional.x case (the moment vector then includes a regression-slopes
  # block, both for categorical and continuous models); nothing to disable.

  # change options if estimator = "IV"
  # The IV (MIIV-2SLS) estimator does not minimize an ML/GLS discrepancy
  # function, so the standardized residual ACOV (Ogasawara, 2001; Bentler &
  # Dijkstra, 1985) -- which the per-element SEs/z-statistics and the SRMR
  # exact-fit/close-fit tests rely on -- is not valid here (the projection
  # matrix Q = I - Delta A0^-1 Delta' A1 assumes the estimator minimizes the
  # corresponding discrepancy). The residual point estimates (raw residuals,
  # RMR/SRMR/CRMR) are plain functions of the residuals and remain valid, so
  # we keep the summary but report point estimates only (no SE/z/tests).
  if (lavmodel@estimator == "IV") {
    zstat <- se <- FALSE
    summary_options_1 <- lav_residuals_summary_options_off()
  }

  # observed and fitted sample statistics
  obs_list <- lav_inspect_sampstat(object,
    h1 = h1,
    add_labels = add_labels, add_class = add_class,
    drop_list_single_group = FALSE
  )
  est_list <- lav_inspect_implied(object,
    add_labels = add_labels, add_class = add_class,
    drop_list_single_group = FALSE
  )
  # blocks
  nblocks <- length(obs_list)

  # pre-scale?
  if (type %in% c("cor.bentler", "cor.bollen")) {
    for (b in seq_len(nblocks)) {
      var_obs <- if (lavmodel@conditional.x) {
        diag(obs_list[[b]][["res.cov"]])
      } else {
        diag(obs_list[[b]][["cov"]])
      }
      var_est <- if (lavmodel@conditional.x) {
        diag(est_list[[b]][["res.cov"]])
      } else {
        diag(est_list[[b]][["cov"]])
      }

      # rescale obsList
      obs_list[[b]] <-
        lav_residuals_rescale(x = obs_list[[b]], diag_cov = var_obs)
      # rescale estList
      if (type == "cor.bentler") { # use obsList
        est_list[[b]] <-
          lav_residuals_rescale(x = est_list[[b]], diag_cov = var_obs)
      } else if (type == "cor.bollen") { # use estList for COV only
        est_list[[b]] <- lav_residuals_rescale(
          x = est_list[[b]],
          diag_cov = var_est, diag_cov2 = var_obs
        )
      }

      # standardize the regression slopes (conditional.x, both continuous and
      # categorical): b -> b * s.x / s.y, using the observed SDs for both obs
      # and est (like the intercepts), so the slope residual is in the
      # standardized metric. For categorical endogenous variables s.y = 1
      # (unit-variance latent response), so the factor reduces to s.x.
      if (lavmodel@conditional.x) {
        s_x <- sqrt(diag(obs_list[[b]][["cov.x"]]))
        scale_sl <- outer(1 / sqrt(var_obs), s_x) # [j, k] = s.x[k] / s.y[j]
        obs_list[[b]][["res.slopes"]] <-
          obs_list[[b]][["res.slopes"]] * scale_sl
        est_list[[b]][["res.slopes"]] <-
          est_list[[b]][["res.slopes"]] * scale_sl
      }
    }
  }

  # compute residuals: (observed - implied)
  res_list <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    res_list[[b]] <- lapply(seq_along(obs_list[[b]]),
      FUN = function(el) {
        obs_list[[b]][[el]] - est_list[[b]][[el]]
      }
    )
    # always name the elements, even if add_labels = FALSE
    names_1 <- names(obs_list[[b]])
    names(res_list[[b]]) <- names_1
  }

  # do we need seList?
  #
  # ACOV.obs (the ACOV of the observed h1 sample statistics) is the expensive
  # building block underlying both the residual SEs and the summary statistics
  # SEs. We compute it (at most) once here and pass it down to avoid computing
  # it twice (see issue #291).
  acov_obs <- NULL
  if (se || zstat) {
    acov_obs <- lav_residuals_acov_obs(object, h1_acov = h1_acov)
    se_list <- lav_residuals_se(object,
      type = type, z_type = "standardized",
      h1_acov = h1_acov,
      add_class = add_class, add_labels = add_labels,
      acov_obs = acov_obs
    )
  } else if (type %in% c("normalized", "standardized", "standardized.mplus")) {
    acov_obs <- lav_residuals_acov_obs(object, h1_acov = h1_acov)
    se_list <- lav_residuals_se(object,
      type = "raw", z_type = type,
      h1_acov = h1_acov,
      add_class = add_class, add_labels = add_labels,
      acov_obs = acov_obs
    )
  } else {
    se_list <- NULL
  }

  # normalize/standardize?
  if (type %in% c("normalized", "standardized", "standardized.mplus")) {
    for (b in seq_len(nblocks)) {
      if (add_labels) {
        names_1 <- names(res_list[[b]])
      }
      res_list[[b]] <- lapply(seq_along(res_list[[b]]),
        FUN = function(el) {
          a_1 <- res_list[[b]][[el]]
          b_1 <- se_list[[b]][[el]]
          near_zero_idx <- which(abs(a_1) < 1e-05)
          if (length(near_zero_idx) > 0L) {
            b_1[near_zero_idx] <- 1
          }
          a_1 / b_1
        }
      )
      if (add_labels) {
        names(res_list[[b]]) <- names_1
      }
    }
  }

  # add se
  res_list_orig <- res_list
  if (se) {
    for (b in seq_len(nblocks)) {
      names_res <- names(res_list[[b]])
      names_se <- paste0(names_res, ".se")
      res_list[[b]] <- c(res_list[[b]], se_list[[b]])
      names(res_list[[b]]) <- c(names_res, names_se)
    }
  }


  # add zstat
  if (zstat) {
    for (b in seq_len(nblocks)) {
      names_res <- names(res_list[[b]])
      names_z <- paste0(names(res_list_orig[[b]]), ".z")
      tmp <- lapply(seq_along(res_list_orig[[b]]),
        FUN = function(el) {
          a_1 <- res_list_orig[[b]][[el]]
          b_1 <- se_list[[b]][[el]]
          # NOTE: which threshold should we use?
          # used to be 1e-05
          # changed to 1e-04 in 0.6-4
          near_zero_idx <- which(abs(a_1) < 1e-04)
          if (length(near_zero_idx) > 0L) {
            # B[near_zero_idx] <- as.numeric(NA)
            b_1[near_zero_idx] <- 1.0
          }
          a_1 / b_1
        }
      )
      res_list[[b]] <- c(res_list[[b]], tmp)
      names(res_list[[b]]) <- c(names_res, names_z)
    }
  }

  # add summary statistics (rms, mabs)
  if (summary) {
    summary_options_1 <- lav_snake_case(summary_options_1)
    # lav_residuals_summary() computes its ACOV.obs using h1_acov =
    # "unstructured"; reuse the one we already have if it matches (issue #291)
    summary_acov_obs <- if (identical(h1_acov, "unstructured")) {
      acov_obs
    } else {
      NULL
    }
    args <- c(
      list(
        object = object, type = type, h1_acov = h1_acov,
        add_class = add_class,
        summary_blocks_combined = summary_blocks_combined,
        acov_obs = summary_acov_obs
      ),
      summary_options_1
    )
    sum_stat <- do.call("lav_residuals_summary", args)
    if (isTRUE(attr(sum_stat, "combined"))) {
      # a single overall summary table (across all blocks); attached at the top
      # level of the output (see below), not per block
      combined_summary <- sum_stat[[1]][[1]]
    } else {
      for (b in seq_len(nblocks)) {
        names_1 <- names(res_list[[b]])
        res_list[[b]] <- c(res_list[[b]], list(sum_stat[[b]][[1]])) # only 1
        names_1 <- c(names_1, "summary")
        names(res_list[[b]]) <- names_1
      }
    }
  }

  # last: add type
  if (add_type) {
    for (b in seq_len(nblocks)) {
      names_1 <- names(res_list[[b]])
      res_list[[b]] <- c(type, res_list[[b]])
      names_1 <- c("type", names_1)
      names(res_list[[b]]) <- names_1
    }
  }

  # optional: rename 'cov' to 'cor' (if type = "cor")
  if (rename_cov_cor && type %in% c("cor.bentler", "cor.bollen")) {
    for (b in seq_len(nblocks)) {
      names_1 <- names(res_list[[b]])
      names_1 <- gsub("cov", "cor", names_1, fixed = TRUE)
      names(res_list[[b]]) <- names_1
    }
  }


  # output
  out <- res_list
  if (nblocks == 1L && drop_list_single_group) {
    out <- out[[1]]
  } else {
    if (lavdata@nlevels == 1L &&
      length(lavdata@group.label) > 0L) {
      names(out) <- unlist(lavdata@group.label)
    } else if (lavdata@nlevels > 1L &&
      length(lavdata@group.label) == 0L) {
      names(out) <- lavdata@level.label
    }
  }

  # combined summary (combine = TRUE): a single overall table,
  # attached at the top level so it is reachable as out$summary (consistent with
  # the single-block case), rather than repeated/omitted per block
  if (summary && exists("combined_summary", inherits = FALSE)) {
    out$summary <- combined_summary
  }

  out
}

# compute ACOV for the observed h1 sample statistics (ACOV == Gamma/n_g)
#
# This is the (often expensive) building block underlying all residual
# standard errors. It only depends on 'object' and 'h1_acov' (NOT on 'type'
# or 'z_type'), so it can be computed once and reused across the various
# lav_residuals_acov() calls (see issue #291).
lav_residuals_acov_obs <- function(object, h1_acov = "unstructured") {
  lavsamplestats <- object@SampleStats

  if (!is.null(lavsamplestats@NACOV[[1]])) {
    nacov_obs <- lavsamplestats@NACOV
                # if this changes, tag @TDJorgensen in commit message
    # @NACOV[[g]] stores Gamma_g = Cov(sqrt(n_g) s_g) (see the SCALING
    # CONVENTIONS note in lav_samplestats_gamma.R), so ACOV(s_g) =
    # NACOV[[g]] / n_g. (before 0.7-1, this divided by ntotal instead,
    # shrinking all multigroup residual SEs by the factor sqrt(n_g/ntotal);
    # the h1_acov branch below always used 1/n_g.)
    acov_obs <- lapply(seq_along(nacov_obs), function(g) {
      nacov_obs[[g]] / lavsamplestats@nobs[[g]]
    })
  } else {
    acov_obs <- lav_model_h1_acov(
      lavobject = object,
      h1_information = h1_acov
    )
  }

  acov_obs
}

# return ACOV as list per group
#
# 'acov_obs' (the ACOV of the observed h1 sample statistics) may be passed in
# by the caller to avoid recomputing it; if NULL, it is computed here.
lav_residuals_acov <- function(object, type = "raw", z_type = "standardized",
                               h1_acov = "unstructured", acov_obs = NULL) {
  # check type
  if (z_type %in% c("normalized", "standardized.mplus") && type != "raw") {
    lav_msg_stop(gettextf(
      "z.type = %1$s can only be used with type = %2$s",
      dQuote(z_type), dQuote("raw")))
  }

  # slots
  lavdata <- object@Data
  lavmodel <- object@Model

  # return list per group
  acov_res <- vector("list", length = lavdata@ngroups)

  # compute ACOV for observed h1 sample statistics (reuse if provided)
  if (is.null(acov_obs)) {
    acov_obs <- lav_residuals_acov_obs(object, h1_acov = h1_acov)
  }

  # shortcut for normalized
  if (z_type == "normalized") {
    acov_res <- acov_obs
    return(acov_res)
  } else {
    if (z_type == "standardized") {
      a1 <- lav_model_h1_info(object)
      if (lavmodel@estimator == "DWLS" || lavmodel@estimator == "ULS") {
        # A1 is diagonal matrix
        a1 <- lapply(a1, diag)
      }
      if (type %in% c("cor.bentler", "cor.bollen")) {
        sampstat <- lavTech(object, "sampstat")
      }
    } else if (z_type == "standardized.mplus") {
      vcov_1 <- lavTech(object, "vcov")
    }
    mm_delta <- lavTech(object, "delta")
  }

  if (z_type == "standardized") {
    # see Ogasawara (2001) using Bentler & Dijkstra (1985) eq 1.7.4

    # NVarCov, but always 'not' robust
    #
    # new in 0.6-6: to ensure Q is a projection matrix, we
    #               force observed.information = "h1"
    #               (only needed if information is observed)
    this_options <- object@Options
    this_options$observed.information[1] <- "h1"
    a0_inv <- lav_model_info(
      lavmodel = lavmodel,
      lavsamplestats = object@SampleStats,
      lavdata = lavdata,
      lavcache = object@Cache,
      lavimplied = object@implied,
      lavh1 = object@h1,
      lavoptions = this_options,
      extra = FALSE,
      augmented = TRUE,
      inverted = TRUE,
      use_ginv = TRUE
    )

    # ACOV of the residual vector STACKED over all groups: linearizing,
    #   r_g = (s_g - sigma_g)
    #         - Delta_g a0_inv sum_h gw_h Delta_h' A1_h (s_h - sigma_h)
    # i.e. r = Q (s - sigma) with
    #   Q = I - Delta a0_inv Delta' W,   W = bdiag(gw_g A1_g)
    #   ACOV(r) = Q bdiag(ACOV(s_g)) Q'
    # Under across-group equality constraints a0_inv (the constrained,
    # augmented inverse) is NOT block-diagonal across the groups, so the
    # diagonal blocks of ACOV(r) pick up Var(sigma-hat_g) contributions
    # from every group's sampling variability. (before 0.7-1, this was
    # computed per group as Q_gg ACOV(s_g) Q_gg', which drops those
    # cross-group terms and understates the residual SEs whenever equality
    # constraints couple the groups; without such constraints the two
    # coincide.)
    gw_all <- unlist(object@SampleStats@nobs) / object@SampleStats@ntotal
                 # if this changes, tag @TDJorgensen in commit message
    delta_all <- do.call(rbind, mm_delta)
    w_all <- lav_mat_bdiag(lapply(
      seq_len(lavdata@ngroups),
      function(g) gw_all[g] * a1[[g]]
    ))
    q_all <- diag(nrow = nrow(delta_all)) -
      delta_all %*% a0_inv %*% t(delta_all) %*% w_all
    acov_res_all <- q_all %*% lav_mat_bdiag(acov_obs) %*% t(q_all)

    # start/end of each group's block in the stacked matrix
    block_n <- vapply(acov_obs, nrow, integer(1L))
    block_end <- cumsum(block_n)
    block_start <- block_end - block_n + 1L

    # per-group cov->cor correction matrices (NULL if type == "raw"),
    # collected to also correct the full stacked matrix afterwards
    cor_mat <- vector("list", lavdata@ngroups)
  }

  # for each group, compute ACOV
  for (g in seq_len(lavdata@ngroups)) {
    if (z_type == "standardized.mplus") { # simplified formula
      # also used by LISREL?
      # see https://www.statmodel.com/download/StandardizedResiduals.pdf

      acov_est_g <- mm_delta[[g]] %*% vcov_1 %*% t(mm_delta[[g]])
      acov_res[[g]] <- acov_obs[[g]] - acov_est_g
    } else if (z_type == "standardized") {
      # this group's diagonal block of the stacked ACOV(r)
      idx_g <- block_start[g]:block_end[g]
      acov_res[[g]] <- acov_res_all[idx_g, idx_g, drop = FALSE]

      # correct ACOV.res for type = "cor.bentler" or type = "cor.bollen"
      if (type == "cor.bentler") {
        if (lavmodel@categorical) {
          mixed_or_condx <- lavmodel@conditional.x ||
            length(unlist(lavmodel@num.idx)) > 0L
          if (mixed_or_condx) {
            # mixed continuous + ordinal and/or conditional.x: transform to
            # correlation metric using the (residual) covariance matrix
            cov_1 <- if (lavmodel@conditional.x) {
              sampstat[[g]][["res.cov"]]
            } else {
              sampstat[[g]][["cov"]]
            }
            nexo <- if (lavmodel@conditional.x) lavmodel@nexo[g] else 0L
            s_x <- if (lavmodel@conditional.x) {
              sqrt(diag(sampstat[[g]][["cov.x"]]))
            } else {
              NULL
            }
            jac <- lav_residuals_cor_jacobian_cat(
              lavmodel, g, cov_1, "cor.bentler", nexo = nexo, s_x = s_x
            )
            cor_mat[[g]] <- jac
          } else {
            # pure categorical: already in correlation metric, nothing to do
          }
        } else if (lavdata@nlevels > 1L) {
          # multilevel: block-diagonal jacobian over the level-specific blocks
          jac <- lav_residuals_cor_jacobian_ml(object, sampstat, "cor.bentler")
          cor_mat[[g]] <- jac
        } else {
          # Ogasawara (2001), eq (13), or
          # Maydeu-Olivares (2017), eq (16)
          cov_1 <- if (lavmodel@conditional.x) {
            sampstat[[g]][["res.cov"]]
          } else {
            sampstat[[g]][["cov"]]
          }
          ss <- 1 / sqrt(diag(cov_1))
          tmp <- lav_mat_vech(tcrossprod(ss))
          g_inv_sqrt <- diag(tmp, nrow = length(tmp))
          if (lavmodel@conditional.x) {
            # moment order [vecr(cbind(int, slopes)), vech(cov)]: intercepts
            # are rescaled by 1/s.y, regression slopes by s.x/s.y
            s_x <- sqrt(diag(sampstat[[g]][["cov.x"]]))
            d_intsl <- lav_residuals_condx_intsl_scale(
              ss, s_x, lavmodel@meanstructure
            )
            gg <- lav_mat_bdiag(
              diag(d_intsl, nrow = length(d_intsl)), g_inv_sqrt
            )
          } else if (lavmodel@meanstructure) {
            gg <- lav_mat_bdiag(
              diag(ss, nrow = length(ss)),
              g_inv_sqrt
            )
          } else {
            gg <- g_inv_sqrt
          }
          cor_mat[[g]] <- gg
        } # continuous
      } else if (type == "cor.bollen") {
        if (lavmodel@categorical) {
          mixed_or_condx <- lavmodel@conditional.x ||
            length(unlist(lavmodel@num.idx)) > 0L
          if (mixed_or_condx) {
            # mixed continuous + ordinal and/or conditional.x: transform to
            # correlation metric using the (residual) covariance matrix
            cov_1 <- if (lavmodel@conditional.x) {
              sampstat[[g]][["res.cov"]]
            } else {
              sampstat[[g]][["cov"]]
            }
            nexo <- if (lavmodel@conditional.x) lavmodel@nexo[g] else 0L
            s_x <- if (lavmodel@conditional.x) {
              sqrt(diag(sampstat[[g]][["cov.x"]]))
            } else {
              NULL
            }
            jac <- lav_residuals_cor_jacobian_cat(
              lavmodel, g, cov_1, "cor.bollen", nexo = nexo, s_x = s_x
            )
            cor_mat[[g]] <- jac
          } else {
            # pure categorical: already in correlation metric, nothing to do
          }
        } else if (lavdata@nlevels > 1L) {
          # multilevel: block-diagonal jacobian over the level-specific blocks
          jac <- lav_residuals_cor_jacobian_ml(object, sampstat, "cor.bollen")
          cor_mat[[g]] <- jac
        } else {
          # here we use the Maydeu-Olivares (2017) approach, see eq 17
          cov_1 <- if (lavmodel@conditional.x) {
            sampstat[[g]][["res.cov"]]
          } else {
            sampstat[[g]][["cov"]]
          }
          f1 <- lav_deriv_cov2cor_b(cov_1)
          if (lavmodel@conditional.x) {
            ss <- 1 / sqrt(diag(cov_1))
            s_x <- sqrt(diag(sampstat[[g]][["cov.x"]]))
            d_intsl <- lav_residuals_condx_intsl_scale(
              ss, s_x, lavmodel@meanstructure
            )
            ff <- lav_mat_bdiag(diag(d_intsl, nrow = length(d_intsl)), f1)
          } else if (lavmodel@meanstructure) {
            ss <- 1 / sqrt(diag(cov_1))
            ff <- lav_mat_bdiag(diag(ss, nrow = length(ss)), f1)
          } else {
            ff <- f1
          }
          cor_mat[[g]] <- ff
        } # continuous
      } # cor.bollen

      # apply the cov->cor correction to this group's block
      if (!is.null(cor_mat[[g]])) {
        acov_res[[g]] <- cor_mat[[g]] %*% acov_res[[g]] %*% t(cor_mat[[g]])
      }
    } # z.type = "standardized"
  } # g

  # attach the full stacked ACOV (all groups; correlation-metric corrected
  # like the per-group blocks): used by the combined residual summaries,
  # where the cross-group covariance between the residuals (nonzero under
  # across-group equality constraints) must be retained
  if (z_type == "standardized") {
    if (any(!vapply(cor_mat, is.null, logical(1L)))) {
      cor_all <- lav_mat_bdiag(lapply(
        seq_len(lavdata@ngroups),
        function(g) {
          if (is.null(cor_mat[[g]])) {
            diag(nrow = block_n[g])
          } else {
            cor_mat[[g]]
          }
        }
      ))
      acov_res_all <- cor_all %*% acov_res_all %*% t(cor_all)
    }
    attr(acov_res, "full") <- acov_res_all
  }

  acov_res
}

# index map of the (categorical) h1 sample-statistics vector, in the order
# produced by lav_samp_wls_obs():
#   [ th-block (thresholds + continuous means/intercepts, by variable),
#     regression slopes (vec, only if conditional.x = TRUE; nvar * nexo),
#     continuous variances (num.idx),
#     covariances (vech, no diagonal) ]
#
# Returns, as positions into that vector: $th (thresholds), $mean (continuous
# means/intercepts), $slopes (regression slopes), $var (continuous variances),
# $cov (off-diagonal (co)variances), plus $num_idx (continuous variable
# indices), $nvar and $nexo. For conditional.x = FALSE, $slopes is empty and
# nexo = 0.
lav_residuals_cat_idx <- function(lavmodel, g = 1L, nvar = NULL, nexo = 0L) {
  th_idx <- lavmodel@th.idx[[g]]
  nth_block <- length(th_idx)
  num_idx <- lavmodel@num.idx[[g]]
  n_num <- length(num_idx)
  if (is.null(nvar)) {
    nvar <- length(unique(th_idx[th_idx != 0])) + n_num
  }
  n_slopes <- nvar * nexo
  n_offdiag <- nvar * (nvar - 1) / 2

  list(
    th = which(th_idx != 0L), # threshold positions in the th-block
    mean = which(th_idx == 0L), # continuous mean/intercept positions
    slopes = if (n_slopes > 0L) nth_block + seq_len(n_slopes) else integer(0),
    var = nth_block + n_slopes + seq_len(n_num), # continuous variances
    cov = nth_block + n_slopes + n_num + seq_len(n_offdiag), # off-diag covs
    num_idx = num_idx,
    nvar = nvar,
    nexo = nexo
  )
}

# Build the cov->cor standardization jacobian for the *categorical* moment
# vector ordering [th-block, (slopes), continuous variances, vech(cov, no
# diagonal)], used to transform the raw-residual ACOV into the correlation
# metric for mixed (continuous + ordinal) models. 'cov_1' is the observed (h1)
# (residual) covariance matrix (ordinal variables have a unit diagonal, so
# 1/sqrt(diag) is 1 for them and 1/SD for the continuous ones). 'type' is
# either "cor.bentler" (observed SDs treated as fixed -> diagonal scaling) or
# "cor.bollen" (full cov2cor jacobian). 'nexo' > 0 for conditional.x (a slopes
# block is present). Returns J such that ACOV.cor = J %*% ACOV.raw %*% t(J).
#
# Thresholds are left unchanged. The regression slopes are standardized as
# b -> b * s.x / s.y (matching the point estimate, cf. the prescaling loop in
# lav_residuals_acov()); for ordinal endogenous variables s.y = 1, so the
# factor reduces to s.x. 's_x' holds the SDs of the exogenous covariates (only
# needed when a slopes block is present). For a pure-categorical model (no
# continuous variables and no slopes) this reduces to the identity; we
# therefore only use it in the mixed / conditional.x case.
lav_residuals_cor_jacobian_cat <- function(lavmodel, g, cov_1, type,
                                           nexo = 0L, s_x = NULL) {
  idx <- lav_residuals_cat_idx(lavmodel, g, nvar = nrow(cov_1), nexo = nexo)
  nvar <- idx$nvar
  num_idx <- idx$num_idx
  p <- length(idx$th) + length(idx$mean) + length(idx$slopes) +
    length(idx$var) + length(idx$cov)

  ss <- 1 / sqrt(diag(cov_1)) # 1 for ordinal (unit diagonal), 1/SD otherwise

  jac <- matrix(0, nrow = p, ncol = p)
  # thresholds: unchanged
  if (length(idx$th) > 0L) {
    jac[cbind(idx$th, idx$th)] <- 1
  }
  # regression slopes: scaled by s.x / s.y (vec / column-major order, matching
  # idx$slopes); leaving residual and SE scaled by the same factor keeps the
  # per-element z-statistic invariant
  if (length(idx$slopes) > 0L) {
    slope_scale <- if (!is.null(s_x)) as.vector(outer(ss, s_x)) else 1
    jac[cbind(idx$slopes, idx$slopes)] <- slope_scale
  }
  # continuous means/intercepts: scaled by 1/SD (observed SD treated as fixed)
  if (length(idx$mean) > 0L) {
    jac[cbind(idx$mean, idx$mean)] <- ss[num_idx]
  }

  if (type == "cor.bentler") {
    # diagonal scaling; observed SDs treated as fixed constants
    if (length(idx$var) > 0L) {
      jac[cbind(idx$var, idx$var)] <- ss[num_idx]^2
    }
    if (length(idx$cov) > 0L) {
      jac[cbind(idx$cov, idx$cov)] <-
        lav_mat_vech(tcrossprod(ss), diagonal = FALSE)
    }
  } else {
    # cor.bollen: full cov2cor jacobian, restricted to the moments that are
    # actually free (continuous variances + off-diagonal covariances); the
    # ordinal variances are fixed at 1 and are dropped
    f1 <- lav_deriv_cov2cor_b(cov_1)
    diagh <- lav_mat_diagh_idx(nvar)
    offdiag <- setdiff(seq_len(nvar * (nvar + 1) / 2), diagh)
    full_idx <- c(diagh[num_idx], offdiag)
    catcov <- c(idx$var, idx$cov) # variances first, then off-diagonals
    jac[catcov, catcov] <- f1[full_idx, full_idx]
  }

  jac
}

# scaling factors for the [vecr(cbind(intercepts, slopes))] block of the
# *continuous* conditional.x moment vector: intercepts are rescaled by 1/SD
# (ss = 1/s.y), regression slopes by s.x/s.y (b -> b * s.x / s.y). 'ss' is
# 1/s.y per endogenous variable; 's_x' is the SD of each exogenous covariate.
# Returns the diagonal in vecr (row-major) order. If meanstructure = FALSE the
# block contains only the slopes.
lav_residuals_condx_intsl_scale <- function(ss, s_x, meanstructure) {
  b <- if (meanstructure) {
    # row j: [ss_j (intercept), ss_j * s.x_1, ...] = ss_j * c(1, s_x)
    outer(ss, c(1, s_x))
  } else {
    outer(ss, s_x)
  }
  as.vector(t(b))
}

# block-diagonal cov->cor standardization jacobian for a multilevel model.
# The moment vector stacks the level-specific blocks, each ordered
# [mean, vech(cov)] and independent of the others, so the jacobian is the
# direct sum of the per-block continuous jacobians (means rescaled by 1/SD;
# covariances by diagonal scaling for cor.bentler, by lav_deriv_cov2cor_b for
# cor.bollen). 'sampstat' is the per-block lavTech(object, "sampstat").
lav_residuals_cor_jacobian_ml <- function(object, sampstat, type) {
  nblocks <- object@Model@nblocks
  block_jac <- vector("list", nblocks)
  for (b in seq_len(nblocks)) {
    cov_b <- sampstat[[b]][["cov"]]
    ss <- 1 / sqrt(diag(cov_b))
    cov_jac <- if (type == "cor.bentler") {
      diag(lav_mat_vech(tcrossprod(ss)), nrow = length(ss) * (length(ss) + 1) / 2)
    } else {
      lav_deriv_cov2cor_b(cov_b)
    }
    # block order: [mean, vech(cov)]
    block_jac[[b]] <- lav_mat_bdiag(diag(ss, nrow = length(ss)), cov_jac)
  }
  do.call(lav_mat_bdiag, block_jac)
}

# return resList with 'se' values for each residual
lav_residuals_se <- function(object, type = "raw", z_type = "standardized",
                             h1_acov = "unstructured",
                             add_class = FALSE, add_labels = FALSE,
                             acov_obs = NULL) {
  # slots
  lavdata <- object@Data
  lavmodel <- object@Model
  lavpta <- object@pta

  # return list per block (= per group for single-level data; for multilevel
  # data there is one block per level within each group)
  se_list <- vector("list", length = lavmodel@nblocks)

  # get ACOV per group (reuse acov_obs if provided)
  acov_res <- lav_residuals_acov(
    object = object, type = type,
    z_type = z_type, h1_acov = h1_acov, acov_obs = acov_obs
  )

  # labels
  if (add_labels) {
    ov_names <- object@pta$vnames$ov
    # ov_names_res <- object@pta$vnames$ov.nox
    # ov_names_x <- object@pta$vnames$ov.x
  }

  # for each group, compute 'se' values, and fill list
  for (g in seq_len(lavdata@ngroups)) {
    nvar <- object@pta$nvar[[g]] # block or group-based?
    diag_acov <- diag(acov_res[[g]])

    # take care of negative, or non-finite diag.ACOV elements
    diag_acov[!is.finite(diag_acov)] <- NA
    diag_acov[diag_acov < 0] <- NA

    # categorical
    if (lavmodel@categorical) {
      conditional_x <- lavmodel@conditional.x
      nexo <- if (conditional_x) lavmodel@nexo[g] else 0L
      # endogenous variables only (res.cov is over the endogenous ov);
      # for conditional.x the exogenous covariates are NOT in the moment vector
      nvar_endo <- if (conditional_x) {
        length(lavpta$vnames$ov.nox[[g]])
      } else {
        nvar
      }

      # index map of the categorical moment vector (handles the mixed
      # continuous + ordinal case, and the conditional.x slopes block)
      idx <- lav_residuals_cat_idx(lavmodel, g, nvar = nvar_endo, nexo = nexo)
      num_idx <- idx$num_idx

      # COV (off-diagonal); continuous variances go on the diagonal, the
      # ordinal-variable diagonal stays zero (variance is fixed at 1)
      cov_se <- lav_mat_vech_rev(sqrt(diag_acov[idx$cov]), diagonal = FALSE)
      if (length(num_idx) > 0L) {
        diag(cov_se)[num_idx] <- sqrt(diag_acov[idx$var])
      }

      # MEAN/INTERCEPT (continuous variables only; ordinal entries are NA)
      mean_se <- rep(as.numeric(NA), nvar_endo)
      if (length(idx$mean) > 0L) {
        mean_se[num_idx] <- sqrt(diag_acov[idx$mean])
      }

      # TH (thresholds only)
      th_se <- sqrt(diag_acov[idx$th])

      if (conditional_x) {
        # SLOPES (nvar_endo x nexo, vec ordering)
        slopes_se <- matrix(sqrt(diag_acov[idx$slopes]),
          nrow = nvar_endo, ncol = nexo
        )
        # exogenous moments are conditioned on (not in the vector): no SE
        covx_se <- matrix(as.numeric(NA), nexo, nexo)
        meanx_se <- rep(as.numeric(NA), nexo)
        if (add_class) {
          class(cov_se) <- c("lavaan.matrix.symmetric", "matrix")
          class(covx_se) <- c("lavaan.matrix.symmetric", "matrix")
          class(slopes_se) <- c("lavaan.matrix", "matrix")
          class(mean_se) <- class(meanx_se) <- class(th_se) <-
            c("lavaan.vector", "numeric")
        }
        if (add_labels) {
          rownames(cov_se) <- colnames(cov_se) <- lavpta$vnames$ov.nox[[g]]
          names(mean_se) <- lavpta$vnames$ov.nox[[g]]
          names(th_se) <- lavpta$vnames$th[[g]]
          rownames(slopes_se) <- lavpta$vnames$ov.nox[[g]]
          colnames(slopes_se) <- lavpta$vnames$ov.x[[g]]
          rownames(covx_se) <- colnames(covx_se) <- lavpta$vnames$ov.x[[g]]
          names(meanx_se) <- lavpta$vnames$ov.x[[g]]
        }
        # order must match lav_inspect_sampstat(): res.cov, res.int, res.th,
        # res.slopes, cov.x, mean.x
        se_list[[g]] <- list(
          res.cov.se = cov_se, res.int.se = mean_se, res.th.se = th_se,
          res.slopes.se = slopes_se, cov.x.se = covx_se, mean.x.se = meanx_se
        )
      } else {
        if (add_class) {
          class(cov_se) <- c("lavaan.matrix.symmetric", "matrix")
          class(mean_se) <- c("lavaan.vector", "numeric")
          class(th_se) <- c("lavaan.vector", "numeric")
        }
        if (add_labels) {
          rownames(cov_se) <- colnames(cov_se) <- ov_names[[g]]
          names(mean_se) <- ov_names[[g]]
          # thresholds only (no continuous means); equals th.mean for pure
          # categorical models, excludes continuous means in mixed ones
          names(th_se) <- lavpta$vnames$th[[g]]
        }
        se_list[[g]] <- list(
          cov.se = cov_se, mean.se = mean_se,
          th.se = th_se
        )
      }

      # continuous -- single level
    } else if (lavdata@nlevels == 1L) {
      if (lavmodel@conditional.x) {
        # moment order: [vecr(cbind(res.int, res.slopes)), vech(res.cov)]
        # (intercepts only if meanstructure); res.slopes is nvar.endo x nexo
        nexo <- lavmodel@nexo[g]
        nvar_endo <- length(lavpta$vnames$ov.nox[[g]])
        ncol_b <- if (lavmodel@meanstructure) 1L + nexo else nexo
        n_intsl <- nvar_endo * ncol_b
        # row-major (vecr) reshape of the intercept/slopes block
        b_se <- matrix(sqrt(diag_acov[seq_len(n_intsl)]),
          nrow = nvar_endo, ncol = ncol_b, byrow = TRUE
        )
        cov_se <- lav_mat_vech_rev(sqrt(diag_acov[-seq_len(n_intsl)]),
          diagonal = TRUE
        )
        covx_se <- matrix(as.numeric(NA), nexo, nexo)
        meanx_se <- rep(as.numeric(NA), nexo)
        if (lavmodel@meanstructure) {
          int_se <- b_se[, 1]
          slopes_se <- b_se[, -1, drop = FALSE]
        } else {
          int_se <- NULL
          slopes_se <- b_se
        }
        if (add_class) {
          class(cov_se) <- c("lavaan.matrix.symmetric", "matrix")
          class(covx_se) <- c("lavaan.matrix.symmetric", "matrix")
          class(slopes_se) <- c("lavaan.matrix", "matrix")
          class(meanx_se) <- c("lavaan.vector", "numeric")
          if (!is.null(int_se)) class(int_se) <- c("lavaan.vector", "numeric")
        }
        if (add_labels) {
          rownames(cov_se) <- colnames(cov_se) <- lavpta$vnames$ov.nox[[g]]
          rownames(slopes_se) <- lavpta$vnames$ov.nox[[g]]
          colnames(slopes_se) <- lavpta$vnames$ov.x[[g]]
          rownames(covx_se) <- colnames(covx_se) <- lavpta$vnames$ov.x[[g]]
          names(meanx_se) <- lavpta$vnames$ov.x[[g]]
          if (!is.null(int_se)) names(int_se) <- lavpta$vnames$ov.nox[[g]]
        }
        # order must match lav_inspect_sampstat(): res.cov, res.int (if
        # meanstructure), res.slopes, cov.x, mean.x
        if (lavmodel@meanstructure) {
          se_list[[g]] <- list(
            res.cov.se = cov_se, res.int.se = int_se,
            res.slopes.se = slopes_se, cov.x.se = covx_se, mean.x.se = meanx_se
          )
        } else {
          se_list[[g]] <- list(
            res.cov.se = cov_se, res.slopes.se = slopes_se,
            cov.x.se = covx_se, mean.x.se = meanx_se
          )
        }
      } else {
        if (lavmodel@meanstructure) {
          tmp <- sqrt(diag_acov[-(1:nvar)])
          cov_se <- lav_mat_vech_rev(tmp)
          mean_se <- sqrt(diag_acov[1:nvar])
          if (add_class) {
            class(cov_se) <- c("lavaan.matrix.symmetric", "matrix")
            class(mean_se) <- c("lavaan.vector", "numeric")
          }
          if (add_labels) {
            rownames(cov_se) <- colnames(cov_se) <- ov_names[[g]]
            names(mean_se) <- ov_names[[g]]
          }
          se_list[[g]] <- list(cov.se = cov_se, mean.se = mean_se)
        } else {
          cov_se <- lav_mat_vech_rev(sqrt(diag_acov))
          if (add_class) {
            class(cov_se) <- c("lavaan.matrix.symmetric", "matrix")
          }
          if (add_labels) {
            rownames(cov_se) <- colnames(cov_se) <- ov_names[[g]]
          }
          se_list[[g]] <- list(cov.se = cov_se)
        }
      }

      # continuous -- multilevel
    } else if (lavdata@nlevels > 1L) {
      if (lavdata@ngroups > 1L) {
        # multigroup + multilevel: not ready yet
        lav_msg_stop(gettext("not ready yet"))
      }
      # the moment vector stacks the level-specific blocks, each ordered
      # [mean, vech(cov)]; the within-level means are fixed at zero and thus
      # have a zero standard error
      offset <- 0L
      for (b in seq_len(lavmodel@nblocks)) {
        nvb <- object@pta$nvar[[b]]
        pstar_b <- nvb * (nvb + 1) / 2
        mean_se <- sqrt(diag_acov[offset + seq_len(nvb)])
        cov_se <- lav_mat_vech_rev(
          sqrt(diag_acov[offset + nvb + seq_len(pstar_b)])
        )
        offset <- offset + nvb + pstar_b
        if (add_class) {
          class(cov_se) <- c("lavaan.matrix.symmetric", "matrix")
          class(mean_se) <- c("lavaan.vector", "numeric")
        }
        if (add_labels) {
          rownames(cov_se) <- colnames(cov_se) <- ov_names[[b]]
          names(mean_se) <- ov_names[[b]]
        }
        se_list[[b]] <- list(cov.se = cov_se, mean.se = mean_se)
      }
    }
  } # g

  se_list
}

# return summary statistics (rmr/srmr/crmr) as a list per group
#
# For each requested 'type' and each group, we summarize the residuals into
# one or more RMS columns (e.g. cov/mean/total or cor/thresholds/total). The
# heavy lifting (point estimate, SE, z-statistic, unbiased version, CI, close-
# fit test) is done by lav_residuals_summary_rms(); this function only decides
# *which* residuals enter *which* column, and slices the corresponding ACOV.
lav_residuals_summary <- function(object, type = c("rmr", "srmr", "crmr"),
                                  h1_acov = "unstructured",
                                  se = FALSE, zstat = FALSE, pvalue = FALSE,
                                  unbiased = FALSE, unbiased_se = FALSE,
                                  unbiased_ci = FALSE, unbiased_ci_level = 0.90,
                                  unbiased_zstat = FALSE,
                                  unbiased_test_val = 0.05,
                                  unbiased_pvalue = FALSE,
                                  summary_blocks_combined = FALSE,
                                  add_class = FALSE, acov_obs = NULL) {
  # imply flags
  if (pvalue) {
    zstat <- TRUE
  }
  if (zstat) {
    se <- TRUE
  }
  if (unbiased_pvalue) {
    unbiased_zstat <- TRUE
  }
  if (unbiased_zstat) {
    unbiased_se <- TRUE
  }

  if (!all(type %in% c(
    "rmr", "srmr", "crmr",
    "raw", "cor.bentler", "cor.bollen"
  ))) {
    lav_msg_stop(gettext("unknown type:"), dQuote(type))
  }

  # change type name to the canonical rmr/srmr/crmr
  type[type == "raw"] <- "rmr"
  type[type == "cor.bentler"] <- "srmr"
  type[type == "cor.bollen"] <- "crmr"

  # slots
  lavdata <- object@Data
  lavmodel <- object@Model
  conditional_x <- lavmodel@conditional.x

  # residual point estimates (and, if needed, their ACOV) per requested type;
  # acov_obs is reused across types to avoid recomputing it (issue #291)
  if ("rmr" %in% type) {
    rmr_list <- lav_residuals(object = object, type = "raw")
    if (se || unbiased) {
      rmr_list_se <- lav_residuals_acov(
        object = object, type = "raw", z_type = "standardized",
        h1_acov = "unstructured", acov_obs = acov_obs
      )
    }
  }
  if ("srmr" %in% type) {
    srmr_list <- lav_residuals(object = object, type = "cor.bentler")
    if (se || unbiased) {
      srmr_list_se <- lav_residuals_acov(
        object = object, type = "cor.bentler", z_type = "standardized",
        h1_acov = "unstructured", acov_obs = acov_obs
      )
    }
  }
  if ("crmr" %in% type) {
    crmr_list <- lav_residuals(object = object, type = "cor.bollen")
    if (se || unbiased) {
      crmr_list_se <- lav_residuals_acov(
        object = object, type = "cor.bollen", z_type = "standardized",
        h1_acov = "unstructured", acov_obs = acov_obs
      )
    }
  }

  # select the residual (and ACOV) lists for a given type + group
  get_rms_lists <- function(type_name, g) {
    est <- switch(type_name,
      rmr = rmr_list[[g]], srmr = srmr_list[[g]], crmr = crmr_list[[g]]
    )
    se_g <- NULL
    if (se || unbiased) {
      se_g <- switch(type_name,
        rmr = rmr_list_se[[g]], srmr = srmr_list_se[[g]],
        crmr = crmr_list_se[[g]]
      )
    }
    list(est = est, se = se_g)
  }

  # compute one RMS summary column (captures the various reporting flags)
  do_rms <- function(stats, acov, pstar, type_name) {
    lav_residuals_summary_rms(
      stats = stats, acov = acov, se = se, zstat = zstat, pvalue = pvalue,
      unbiased = unbiased, unbiased_se = unbiased_se,
      unbiased_ci = unbiased_ci, unbiased_ci_level = unbiased_ci_level,
      unbiased_zstat = unbiased_zstat, unbiased_test_val = unbiased_test_val,
      unbiased_pvalue = unbiased_pvalue, pstar = pstar, type = type_name
    )
  }

  # The per-block summary is built from a list of "column specs". A column spec
  # is list(stats = , lidx = , pstar = , offset = ): 'stats' are the residual
  # elements entering that column, 'lidx' their (local) positions in the block's
  # ACOV, 'pstar' the divisor of the RMS, and 'offset' the block's start row in
  # the per-type global ACOV (only used for combining multilevel blocks, where
  # the ACOV is one joint matrix). The same specs feed both the per-block tables
  # (finalize_block_table) and the single combined table (combine_specs_table),
  # so there is exactly one place that decides which residuals enter which
  # column.

  # continuous cov/mean/total column specs for one type
  continuous_specs <- function(rms_list_g, nvar_b, mean_idx, cov_idx,
                               cov_nm, mean_nm, ty, has_mean, offset = 0L) {
    specs <- list()
    # COV
    stats <- lav_mat_vech(rms_list_g[[cov_nm]])
    pstar <- length(stats)
    if (ty == "crmr") {
      pstar <- pstar - nvar_b # CRMR excludes the (co)variance diagonal
    }
    specs[["cov"]] <- list(
      stats = stats, lidx = cov_idx, pstar = pstar, offset = offset
    )
    if (has_mean) {
      # MEAN / INTERCEPT
      stats <- rms_list_g[[mean_nm]]
      specs[["mean"]] <- list(
        stats = stats, lidx = mean_idx, pstar = length(stats), offset = offset
      )
      # TOTAL (means/intercepts + covariances)
      stats <- c(rms_list_g[[mean_nm]], lav_mat_vech(rms_list_g[[cov_nm]]))
      pstar <- length(stats)
      if (ty == "crmr") {
        pstar <- pstar - nvar_b
      }
      specs[["total"]] <- list(
        stats = stats, lidx = c(mean_idx, cov_idx), pstar = pstar,
        offset = offset
      )
    }
    specs
  }

  # continuous conditional.x column specs: res.cov / res.int / res.slopes /
  # total. The moment vector is [vecr(cbind(res.int, res.slopes)), vech(res.cov)]
  condx_specs <- function(rms_list_g, nvar_b, nexo, has_mean, ty, offset = 0L) {
    ncol_b <- if (has_mean) 1L + nexo else nexo
    n_intsl <- nvar_b * ncol_b
    pstar_cov <- nvar_b * (nvar_b + 1) / 2
    int_idx <- if (has_mean) {
      seq.int(1L, by = ncol_b, length.out = nvar_b)
    } else {
      integer(0)
    }
    slope_idx <- setdiff(seq_len(n_intsl), int_idx) # vecr (row-major) order
    cov_idx <- n_intsl + seq_len(pstar_cov)

    specs <- list()
    # res.cov
    stats <- lav_mat_vech(rms_list_g[["res.cov"]])
    pstar <- length(stats)
    if (ty == "crmr") pstar <- pstar - nvar_b
    specs[["res.cov"]] <- list(
      stats = stats, lidx = cov_idx, pstar = pstar, offset = offset
    )
    # res.int (kept before res.slopes to match the column order)
    if (has_mean) {
      stats <- rms_list_g[["res.int"]]
      specs[["res.int"]] <- list(
        stats = stats, lidx = int_idx, pstar = length(stats), offset = offset
      )
    }
    # res.slopes
    stats <- lav_mat_vecr(rms_list_g[["res.slopes"]])
    specs[["res.slopes"]] <- list(
      stats = stats, lidx = slope_idx, pstar = length(stats), offset = offset
    )
    # total (all moments, in moment-vector order)
    int_sl <- if (has_mean) {
      lav_mat_vecr(cbind(rms_list_g[["res.int"]], rms_list_g[["res.slopes"]]))
    } else {
      lav_mat_vecr(rms_list_g[["res.slopes"]])
    }
    stats <- c(int_sl, lav_mat_vech(rms_list_g[["res.cov"]]))
    pstar <- length(stats)
    if (ty == "crmr") pstar <- pstar - nvar_b
    specs[["total"]] <- list(
      stats = stats, lidx = seq_len(n_intsl + pstar_cov), pstar = pstar,
      offset = offset
    )
    specs
  }

  # categorical column specs: cor / thresholds / total (unconditional) or
  # res.cov / res.th / res.slopes / total (conditional.x). 'idx' is the
  # moment-vector index map (lav_residuals_cat_idx); it is NULL when no ACOV is
  # needed. The canonical lavaan WLS ordering for the total is [thresholds,
  # (slopes), covariances].
  categorical_specs <- function(rms_list_g, idx, nvar_endo, conditional_x,
                                cov_nm, th_nm, ty, offset = 0L) {
    specs <- list()
    # COR (off-diagonal (co)variances); pstar = p*(p+1)/2 for SRMR, p*(p-1)/2
    # for CRMR
    stats <- lav_mat_vech(rms_list_g[[cov_nm]], diagonal = FALSE)
    pstar <- if (ty == "crmr") length(stats) else length(stats) + nvar_endo
    specs[[if (conditional_x) "res.cov" else "cor"]] <- list(
      stats = stats, lidx = idx$cov, pstar = pstar, offset = offset
    )
    # THRESHOLDS
    stats <- rms_list_g[[th_nm]]
    specs[[if (conditional_x) "res.th" else "thresholds"]] <- list(
      stats = stats, lidx = idx$th, pstar = length(stats), offset = offset
    )
    # SLOPES (conditional.x only); vec / column-major order, matching idx$slopes
    if (conditional_x) {
      stats <- as.vector(rms_list_g[["res.slopes"]])
      specs[["res.slopes"]] <- list(
        stats = stats, lidx = idx$slopes, pstar = length(stats), offset = offset
      )
    }
    # TOTAL: [thresholds, (slopes), covariances]
    stats <- c(
      rms_list_g[[th_nm]],
      if (conditional_x) as.vector(rms_list_g[["res.slopes"]]),
      lav_mat_vech(rms_list_g[[cov_nm]], diagonal = FALSE)
    )
    pstar <- if (ty == "crmr") length(stats) else length(stats) + nvar_endo
    specs[["total"]] <- list(
      stats = stats, lidx = c(idx$th, idx$slopes, idx$cov), pstar = pstar,
      offset = offset
    )
    specs
  }

  # build one summary table from a list of column specs, slicing the block's
  # ACOV (block_acov) with the (local) indices stored in each spec
  finalize_block_table <- function(specs, block_acov, ty) {
    cols <- lapply(specs, function(sp) {
      acov <- if (!is.null(block_acov)) {
        block_acov[sp$lidx, sp$lidx, drop = FALSE]
      } else {
        NULL
      }
      do_rms(sp$stats, acov, sp$pstar, ty)
    })
    table_1 <- as.data.frame(do.call(cbind, cols))
    colnames(table_1) <- names(specs)
    if (add_class) {
      class(table_1) <- c("lavaan.data.frame", "data.frame")
    }
    table_1
  }

  # build the single combined table by pooling, per column, the residual
  # elements across all blocks. The combined ACOV is sliced from the one
  # joint ACOV: across levels the blocks are correlated by construction,
  # and across (independent) groups the residuals are still correlated
  # whenever across-group equality constraints couple theta-hat (before
  # 0.7-1, the group blocks were assumed independent and bdiag'ed, dropping
  # that cross-group covariance). The bdiag path remains as a fallback only.
  combine_specs_table <- function(block_specs_ty, block_acov_ty, full_acov,
                                  ty) {
    col_names <- names(block_specs_ty[[1]])
    cols <- lapply(col_names, function(cn) {
      stats <- unlist(lapply(block_specs_ty, function(bs) bs[[cn]]$stats))
      pstar <- sum(vapply(
        block_specs_ty, function(bs) bs[[cn]]$pstar, numeric(1)
      ))
      acov <- NULL
      if (se || unbiased) {
        if (!is.null(full_acov)) {
          # one joint ACOV: gather the (global) indices across the blocks
          gidx <- unlist(lapply(
            block_specs_ty, function(bs) bs[[cn]]$offset + bs[[cn]]$lidx
          ))
          acov <- full_acov[gidx, gidx, drop = FALSE]
        } else {
          # fallback: block-diagonal of the per-block column slices
          blocks <- lapply(seq_along(block_specs_ty), function(b) {
            li <- block_specs_ty[[b]][[cn]]$lidx
            block_acov_ty[[b]][li, li, drop = FALSE]
          })
          acov <- if (length(blocks) == 1L) {
            blocks[[1]]
          } else {
            do.call(lav_mat_bdiag, blocks)
          }
        }
      }
      do_rms(stats, acov, pstar, ty)
    })
    table_1 <- as.data.frame(do.call(cbind, cols))
    colnames(table_1) <- col_names
    if (add_class) {
      class(table_1) <- c("lavaan.data.frame", "data.frame")
    }
    table_1
  }

  # ACOV list for a given type (one element per group for single-level data;
  # for multilevel data a single joint ACOV across the levels)
  get_se_list <- function(ty) {
    if (!(se || unbiased)) {
      return(NULL)
    }
    switch(ty,
      rmr = rmr_list_se, srmr = srmr_list_se, crmr = crmr_list_se
    )
  }

  multilevel <- (lavdata@nlevels > 1L)
  if (multilevel && lavdata@ngroups > 1L) {
    lav_msg_stop(gettext("not ready yet")) # multigroup + multilevel
  }

  # Gather, per type and per block, the column specs and the block's ACOV.
  # block_specs[[ty]][[b]] is a named list of column specs; block_acov[[ty]][[b]]
  # is the block's ACOV (or NULL). For multilevel data, all blocks share one
  # joint ACOV; each block's ACOV is the corresponding diagonal sub-block, and
  # the spec 'offset' records where the block starts in the joint matrix.
  block_specs <- vector("list", length(type))
  block_acov <- vector("list", length(type))
  names(block_specs) <- names(block_acov) <- type

  if (multilevel) {
    # multilevel: one block per level; the moment vector of each block is
    # ordered [mean, vech(cov)]
    for (ty in type) {
      block_specs[[ty]] <- vector("list", lavmodel@nblocks)
      block_acov[[ty]] <- vector("list", lavmodel@nblocks)
    }
    offset <- 0L
    for (b in seq_len(lavmodel@nblocks)) {
      nvb <- object@pta$nvar[[b]]
      pstar_b <- nvb * (nvb + 1) / 2
      blk_size <- nvb + pstar_b
      blk <- offset + seq_len(blk_size) # [mean, cov] block range
      for (ty in type) {
        est_b <- switch(ty,
          rmr = rmr_list[[b]], srmr = srmr_list[[b]], crmr = crmr_list[[b]]
        )
        block_specs[[ty]][[b]] <- continuous_specs(
          est_b, nvb,
          mean_idx = seq_len(nvb), cov_idx = nvb + seq_len(pstar_b),
          cov_nm = "cov", mean_nm = "mean", ty = ty,
          has_mean = lavmodel@meanstructure, offset = offset
        )
        se_list <- get_se_list(ty)
        if (!is.null(se_list)) {
          block_acov[[ty]][[b]] <- se_list[[1]][blk, blk, drop = FALSE]
        }
      }
      offset <- offset + blk_size
    }
  } else {
    for (ty in type) {
      block_specs[[ty]] <- vector("list", lavdata@ngroups)
      block_acov[[ty]] <- vector("list", lavdata@ngroups)
    }
    offset_g <- 0L # start row of this group's block in the joint ACOV
    for (g in seq_len(lavdata@ngroups)) {
      nvar <- object@pta$nvar[[g]]
      if (lavmodel@categorical) {
        # categorical (incl. mixed continuous + ordinal and conditional.x);
        # endogenous variables only
        nexo <- if (conditional_x) lavmodel@nexo[g] else 0L
        nvar_endo <- if (conditional_x) {
          length(object@pta$vnames$ov.nox[[g]])
        } else {
          nvar
        }
        cov_nm <- if (conditional_x) "res.cov" else "cov"
        th_nm <- if (conditional_x) "res.th" else "th"
        idx <- if (se || unbiased) {
          lav_residuals_cat_idx(lavmodel, g, nvar = nvar_endo, nexo = nexo)
        } else {
          NULL
        }
        for (ty in type) {
          sel <- get_rms_lists(ty, g)
          block_specs[[ty]][[g]] <- categorical_specs(
            sel$est, idx, nvar_endo, conditional_x, cov_nm, th_nm, ty,
            offset = offset_g
          )
          block_acov[[ty]][g] <- list(sel$se) # single [ keeps NULL slots
        }
      } else if (conditional_x) {
        # continuous conditional.x: res.cov / res.int / res.slopes / total
        nvar_endo <- length(object@pta$vnames$ov.nox[[g]])
        nexo <- lavmodel@nexo[g]
        for (ty in type) {
          sel <- get_rms_lists(ty, g)
          block_specs[[ty]][[g]] <- condx_specs(
            sel$est, nvar_endo, nexo,
            has_mean = lavmodel@meanstructure, ty = ty, offset = offset_g
          )
          block_acov[[ty]][g] <- list(sel$se) # single [ keeps NULL slots
        }
      } else {
        # continuous unconditional: cov / mean / total, order [mean, vech(cov)]
        nvar_endo <- nvar
        mean_acov_idx <- if (lavmodel@meanstructure) {
          seq_len(nvar_endo)
        } else {
          integer(0)
        }
        cov_acov_idx <- length(mean_acov_idx) +
          seq_len(nvar_endo * (nvar_endo + 1) / 2)
        for (ty in type) {
          sel <- get_rms_lists(ty, g)
          block_specs[[ty]][[g]] <- continuous_specs(
            sel$est, nvar_endo,
            mean_idx = mean_acov_idx, cov_idx = cov_acov_idx,
            cov_nm = "cov", mean_nm = "mean", ty = ty,
            has_mean = lavmodel@meanstructure, offset = offset_g
          )
          block_acov[[ty]][g] <- list(sel$se) # single [ keeps NULL slots
        }
      }
      # advance to the next group's block in the joint ACOV
      if (se || unbiased) {
        offset_g <- offset_g + nrow(get_se_list(type[1])[[g]])
      }
    } # g
  }

  # combine the block-specific summaries into a single overall table?
  if (summary_blocks_combined && lavmodel@nblocks > 1L) {
    out <- vector("list", length(type))
    names(out) <- type
    for (ty in type) {
      full_acov <- NULL
      if (se || unbiased) {
        if (multilevel) {
          # one joint ACOV across the levels
          full_acov <- get_se_list(ty)[[1]]
        } else {
          # single-level multigroup: the joint (stacked) ACOV attached by
          # lav_residuals_acov(), which retains the cross-group covariance
          # of the residuals under across-group equality constraints
          full_acov <- attr(get_se_list(ty), "full")
        }
      }
      out[[ty]] <- combine_specs_table(
        block_specs[[ty]], block_acov[[ty]], full_acov, ty
      )
    }
    sum_stat <- list(out)
    attr(sum_stat, "combined") <- TRUE
    return(sum_stat)
  }

  # otherwise: one summary table per block
  sum_stat <- vector("list", length = lavmodel@nblocks)
  for (b in seq_len(lavmodel@nblocks)) {
    out <- vector("list", length(type))
    names(out) <- type
    for (ty in type) {
      out[[ty]] <- finalize_block_table(
        block_specs[[ty]][[b]], block_acov[[ty]][[b]], ty
      )
    }
    sum_stat[[b]] <- out
  }

  sum_stat
}



lav_residuals_summary_rms <- function(stats = NULL, acov = NULL,
                                      se = FALSE,
                                      level = 0.90,
                                      zstat = FALSE, pvalue = FALSE,
                                      unbiased = FALSE,
                                      unbiased_se = FALSE,
                                      unbiased_ci = FALSE,
                                      unbiased_ci_level = 0.90,
                                      unbiased_zstat = FALSE,
                                      unbiased_test_val = 0.05,
                                      unbiased_pvalue = FALSE,
                                      pstar = 0, type = "rms") {
  out <- vector("list", length = 0L)


  # covariance matrix
  if (length(stats) > 0L) {
    rms <- sqrt(sum(stats * stats) / pstar)
  } else {
    rms <- 0
    se <- unbiased <- zstat <- FALSE
  }

  # default is NULL
  rms_se <- rms_z <- rms_pvalue <- NULL
  urms <- urms_se <- urms_z <- urms_pvalue <- NULL
  urms_ci_lower <- urms_ci_upper <- NULL
  if (!unbiased_zstat) {
    unbiased_test_val <- NULL
  }

  if (se || unbiased) {
    tr2 <- sum(diag(acov %*% acov))
    tr1 <- sum(diag(acov))
    if (se) {
      rms_avar <- tr2 / (tr1 * 2 * pstar)
      if (!is.finite(rms_avar) || rms_avar < .Machine$double.eps) {
        rms_se <- as.numeric(NA)
      } else {
        rms_se <- sqrt(rms_avar)
      }
    }
  }

  if (zstat) {
    e_rms <- (sqrt(tr1 / pstar) * (4 * tr1 * tr1 - tr2) / (4 * tr1 * tr1))
    rms_z <- max((rms - e_rms), 0) / rms_se
    if (pvalue) {
      rms_pvalue <- 1 - pnorm(rms_z)
    }
  }

  if (unbiased) {
    t_cov <- as.numeric(crossprod(stats))
    e_ve <- as.numeric(t(stats) %*% acov %*% stats)
    k_cov <- 1 - (tr2 + 2 * e_ve) / (4 * t_cov * t_cov)
    urms <- (1 / k_cov * sqrt(max((t_cov - tr1), 0) / pstar))
    if (unbiased_se) {
      urms_avar <- (1 / (k_cov * k_cov) * (tr2 + 2 * e_ve) /
                                               (2 * pstar * t_cov))
      if (!is.finite(urms_avar) || urms_avar < .Machine$double.eps) {
        urms_se <- as.numeric(NA)
      } else {
        urms_se <- sqrt(urms_avar)
      }
      if (unbiased_ci) {
        a <- (1 - unbiased_ci_level) / 2
        a <- c(a, 1 - a)
        fac <- stats::qnorm(a)
        urms_ci_lower <- urms + urms_se * fac[1]
        urms_ci_upper <- urms + urms_se * fac[2]
      }
      if (unbiased_zstat) {
        urms_z <- (urms - unbiased_test_val) / urms_se
        if (unbiased_pvalue) {
          urms_pvalue <- 1 - pnorm(urms_z)
        }
      }
    }
  }

  # labels
  if (type == "rmr") {
    out <- list(
      rmr = rms, rmr.se = rms_se,
      rmr.exactfit.z = rms_z, rmr.exactfit.pvalue = rms_pvalue,
      urmr = urms, urmr.se = urms_se,
      urmr.ci.lower = urms_ci_lower,
      urmr.ci.upper = urms_ci_upper,
      urmr.closefit.h0.value = unbiased_test_val,
      urmr.closefit.z = urms_z,
      urmr.closefit.pvalue = urms_pvalue
    )
  } else if (type == "srmr") {
    out <- list(
      srmr = rms, srmr.se = rms_se,
      srmr.exactfit.z = rms_z, srmr.exactfit.pvalue = rms_pvalue,
      usrmr = urms, usrmr.se = urms_se,
      usrmr.ci.lower = urms_ci_lower,
      usrmr.ci.upper = urms_ci_upper,
      usrmr.closefit.h0.value = unbiased_test_val,
      usrmr.closefit.z = urms_z,
      usrmr.closefit.pvalue = urms_pvalue
    )
  } else if (type == "crmr") {
    out <- list(
      crmr = rms, crmr.se = rms_se,
      crmr.exactfit.z = rms_z, crmr.exactfit.pvalue = rms_pvalue,
      ucrmr = urms, ucrmr.se = urms_se,
      ucrmr.ci.lower = urms_ci_lower,
      ucrmr.ci.upper = urms_ci_upper,
      ucrmr.closefit.h0.value = unbiased_test_val,
      ucrmr.closefit.z = urms_z,
      ucrmr.closefit.pvalue = urms_pvalue
    )
  }

  unlist(out)
}


# x is a list with sample statistics (eg output of lavInspect(fit, "sampstat")
# y is another (possibly the same) list with sample statistics
#
# to avoid many 'NAs', we set the scale-factor to 1
# if the to-be-scaled value is < 1e-05 (in absolute value)
lav_residuals_rescale <- function(x, diag_cov = NULL, diag_cov2 = NULL) {
  if (is.null(diag_cov2)) {
    diag_cov2 <- diag_cov
  }

  # make sure we can take the sqrt and invert
  diag_cov[!is.finite(diag_cov)] <- NA
  diag_cov[diag_cov < .Machine$double.eps] <- NA
  scale_cov <- tcrossprod(1 / sqrt(diag_cov))

  # for the mean, we use diag.cov2
  diag_cov2[!is.finite(diag_cov2)] <- NA
  diag_cov2[diag_cov2 < .Machine$double.eps] <- NA
  scale_mean <- 1 / sqrt(diag_cov2)

  # rescale cov
  if (!is.null(x[["cov"]])) {
    # catch (near) zero elements in x$cov
    near_zero_idx <- which(abs(x[["cov"]]) < 1e-05)
    scale_cov[near_zero_idx] <- 1
    x[["cov"]][] <- x[["cov"]] * scale_cov
  }
  if (!is.null(x[["res.cov"]])) {
    # catch (near) zero elements in x$res.cov
    near_zero_idx <- which(abs(x[["res.cov"]]) < 1e-05)
    scale_cov[near_zero_idx] <- 1
    x[["res.cov"]][] <- x[["res.cov"]] * scale_cov
  }

  # rescale int/mean
  if (!is.null(x[["res.int"]])) {
    # catch (near) zero elements in x$res.int
    near_zero_idx <- which(abs(x[["res.int"]]) < 1e-05)
    scale_mean[near_zero_idx] <- 1
    x[["res.int"]] <- x[["res.int"]] * scale_mean
  }

  if (!is.null(x[["mean"]])) {
    # catch (near) zero elements in x$mean
    near_zero_idx <- which(abs(x[["mean"]]) < 1e-05)
    scale_mean[near_zero_idx] <- 1
    x[["mean"]] <- x[["mean"]] * scale_mean
  }

  # FIXME: do something sensible for th, slopes, ...

  x
}

# Gauss-Newton optimization: Fisher scoring with Levenberg-Marquardt
# globalization
#
# The search direction solves the damped normal equations
#
#     (H + mu * D) d = -g
#
# where g is the (scoring) gradient of the discrepancy function,
# H = t(Delta) %*% A1 %*% Delta is the expected (Gauss-Newton)
# information, and D = diag(H) is the Marquardt scaling. The damping
# parameter mu is adapted from the gain ratio (actual versus predicted
# reduction): mu -> 0 recovers the undamped Gauss-Newton step (fast
# local convergence), while a large mu turns the step into a short
# (scaled) steepest-descent step (globally safe). A step is only taken
# when it decreases the objective (or, for estimators whose scoring
# gradient is not the exact gradient of the monitored objective, when it
# decreases the scoring residual; see below).
#
# constraints:
# - linear equality constraints are eliminated exactly by optimizing in
#   the packed space x = K %*% z + k0 (the same route as nlminb); as in
#   lav_model_est, bounds are ignored in this case
# - simple equality constraints (ceq.simple) are handled by reducing the
#   gradient/information via ceq.simple.K
# - simple bounds (the lower/upper columns of the parameter table) are
#   honored with projected steps plus an active set: parameters pinned at
#   a bound with an outward-pointing gradient are frozen, the damped
#   system is solved for the remaining parameters, and the trial point is
#   projected onto the box
#
# estimator = "DLS" with dls.GammaNT = "model": the weight matrix is a
# function of the model-implied moments, and the (IRLS) scoring gradient
# omits the dV/dtheta terms of the monitored objective. The estimator is
# defined by the scoring fixed point g(theta) = 0, so near convergence
# the objective value may (slightly) increase along the scoring
# direction; such steps are accepted whenever they shrink the scoring
# residual ||g||.

# YR - 19 Jan 2021: initial version, needed for DLS
# YR - 10 Jul 2026: rewritten: Levenberg-Marquardt damping, packed
#      equality constraints, active-set bounds, Delta/A1 computed once
#      per iteration (shared by gradient and information), streamed
#      ML/GLS kernels via lav_model_grad/lav_model_info_expected,
#      gradient-based convergence test

# evaluate the objective at x; when objective_only = FALSE, also return
# the scoring gradient g and the expected (Gauss-Newton) information H,
# both in the free-parameter space (nx.free)
lav_optim_gn_eval <- function(x, lavmodel = NULL, lavsamplestats = NULL,
                              lavdata = NULL, lavoptions = NULL,
                              objective_only = TRUE,
                              fast_ls = FALSE, group_w = NULL,
                              group_weight = TRUE) {
  lavmodel2 <- lav_model_set_parameters(lavmodel = lavmodel, x = x)
  fx <- lav_model_objective(
    lavmodel = lavmodel2, lavdata = lavdata,
    lavsamplestats = lavsamplestats
  )
  fx <- as.numeric(fx)[1]

  if (objective_only) {
    return(list(fx = fx, g = NULL, H = NULL))
  }

  if (fast_ls) {
    # single-level moment-based estimators (WLS/DWLS/ULS/DLS): compute
    # Delta and the weight matrix A1 once, and share them between the
    # gradient and the information; for DLS with dls.GammaNT = "model"
    # this builds the (expensive) theta-dependent weight matrix a single
    # time per iteration
    lavimplied <- lav_model_implied(lavmodel = lavmodel2)
    delta <- lav_model_delta(lavmodel = lavmodel2)
    a1 <- lav_model_h1_info_expected(
      lavobject = NULL,
      lavmodel = lavmodel2,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavoptions = lavoptions,
      lavimplied = lavimplied,
      lavh1 = NULL,
      lavcache = NULL
    )
    wls_est <- lav_model_wls_est(
      lavmodel = lavmodel2,
      lavimplied = lavimplied
    )
    diagonal_v <- lavmodel@estimator %in% c("DWLS", "ULS")

    g <- H <- NULL
    for (gr in seq_len(lavsamplestats@ngroups)) {
      diff_g <- lavsamplestats@WLS.obs[[gr]] - wls_est[[gr]]
      if (diagonal_v) {
        # t(Delta) %*% diag(a1)
        pre_g <- t(a1[[gr]] * delta[[gr]])
      } else {
        pre_g <- crossprod(delta[[gr]], a1[[gr]])
      }
      g_gr <- group_w[gr] * drop(pre_g %*% diff_g)
      h_gr <- group_w[gr] * (pre_g %*% delta[[gr]])
      if (gr == 1L) {
        g <- -1 * g_gr
        H <- h_gr
      } else {
        g <- g - g_gr
        H <- H + h_gr
      }
    }

    # simple equality constraints: reduce from nx.unco to nx.free
    if (lavmodel@ceq.simple.only &&
      length(g) == lavmodel@nx.unco &&
      lavmodel@nx.unco > lavmodel@nx.free) {
      k_s <- lavmodel@ceq.simple.K
      g <- drop(crossprod(k_s, g))
      H <- crossprod(k_s, H %*% k_s)
    }
  } else {
    # generic route: the same analytic gradient as the nlminb path, and
    # the expected information; both use the streamed ML/GLS kernels
    # when available (the pstar x pstar weight matrix is never formed)
    g <- lav_model_grad(
      lavmodel = lavmodel2,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      type = "free",
      group_weight = group_weight,
      ceq_simple = TRUE
    )
    g <- as.numeric(g)
    H <- lav_model_info_expected(
      lavmodel = lavmodel2,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavoptions = lavoptions
    )
    # simple equality constraints: reduce from nx.unco to nx.free
    # (the gradient is already reduced via ceq_simple = TRUE)
    if (lavmodel@ceq.simple.only &&
      NCOL(H) == lavmodel@nx.unco &&
      lavmodel@nx.unco > lavmodel@nx.free) {
      k_s <- lavmodel@ceq.simple.K
      H <- crossprod(k_s, H %*% k_s)
    }
  }

  # enforce symmetry (guards against round-off in the group sums)
  H <- (H + t(H)) / 2

  list(fx = fx, g = g, H = H)
}

lav_optim_gn <- function(lavmodel = NULL, lavsamplestats = NULL,
                         lavpartable = NULL,
                         lavdata = NULL, lavoptions = NULL) {
  # no support (yet) for nonlinear constraints
  nonlinear_idx <- c(
    lavmodel@ceq.nonlinear.idx,
    lavmodel@cin.nonlinear.idx
  )
  if (length(nonlinear_idx) > 0L) {
    lav_msg_stop(gettext(
      "nonlinear constraints not supported (yet) with optim.method = \"GN\"."))
  }

  # no support (yet) for general inequality constraints. Simple ones are
  # fine: when cin.simple.only is TRUE, every inequality constraint (and
  # every bounds= bound, which is encoded as a cin.function too) has been
  # lowered into the lower/upper columns of the parameter table, which the
  # projected Gauss-Newton iterations below honor natively.
  if (!is.null(body(lavmodel@cin.function)) && !lavmodel@cin.simple.only) {
    lav_msg_stop(gettext(
      "inequality constraints not supported (yet) with optim.method = \"GN\"."))
  }

  # only for estimators with an analytic (scoring) gradient and an
  # expected information in moment space
  if (!lavmodel@estimator %in% c(
    "ML", "GLS", "WLS", "DWLS", "ULS", "DLS", "NTRLS", "catML"
  )) {
    lav_msg_stop(gettextf(
      "optim.method = \"GN\" is not available for estimator %s.",
      dQuote(lavmodel@estimator)))
  }

  # options (gn.args sublist)
  max_iter <- as.integer(lavoptions$gn.args$max_iter)
  tol_x <- lavoptions$gn.args$tol_x
  tol_g <- lavoptions$gn.args$tol_g

  # current set of free parameters
  x <- lav_model_get_parameters(lavmodel)
  npar <- length(x)

  # nothing to do?
  if (npar == 0L) {
    fx <- lav_model_objective(
      lavmodel = lavmodel, lavdata = lavdata,
      lavsamplestats = lavsamplestats
    )
    attr(x, "converged") <- TRUE
    attr(x, "warn.txt") <- ""
    attr(x, "iterations") <- 0L
    attr(x, "control") <- list(iter.max = max_iter, tol.x = tol_x,
                               tol.g = tol_g)
    attr(x, "fx") <- fx
    return(x)
  }

  # linear (non-simple) equality constraints: optimize in the packed
  # space x = K %*% z + k0 (the constraints then hold exactly along the
  # whole path); this mirrors the nlminb route in lav_model_est
  eq_flag <- (lavmodel@eq.constraints &&
    ncol(lavmodel@eq.constraints.K) > 0L)
  if (eq_flag) {
    eq_k <- lavmodel@eq.constraints.K
    eq_k0 <- lavmodel@eq.constraints.k0
  }
  unpack <- function(z) {
    if (eq_flag) {
      as.numeric(eq_k %*% z) + eq_k0
    } else {
      z
    }
  }

  # simple bounds; as in lav_model_est, bounds are dropped in the
  # presence of (non-simple) linear equality constraints (the packed
  # coordinates are linear combinations of the original parameters)
  lb <- rep(-Inf, npar)
  ub <- rep(+Inf, npar)
  if (!eq_flag && !is.null(lavpartable) &&
    (!is.null(lavpartable$lower) || !is.null(lavpartable$upper))) {
    if (lavmodel@ceq.simple.only) {
      free_idx <- which(lavpartable$free > 0L &
        !duplicated(lavpartable$free))
    } else {
      free_idx <- which(lavpartable$free > 0L)
    }
    if (!is.null(lavpartable$lower)) {
      lb <- lavpartable$lower[free_idx]
    }
    if (!is.null(lavpartable$upper)) {
      ub <- lavpartable$upper[free_idx]
    }
    stopifnot(length(lb) == npar, length(ub) == npar)
    # inconsistent bounds may occur with equality constraints (qr() may
    # switch the sign); disable them
    bad_idx <- which(lb > ub)
    if (length(bad_idx) > 0L) {
      lb[bad_idx] <- -Inf
      ub[bad_idx] <- +Inf
    }
    # make the starting values feasible
    x <- pmin(pmax(x, lb), ub)
  }
  bounds_flag <- any(is.finite(lb)) || any(is.finite(ub))

  # group weights for the fast *LS route, matching the analytic gradient
  # (lav_model_grad) so that g is the exact gradient of the objective
  ngroups <- lavsamplestats@ngroups
  group_w <- numeric(ngroups)
  for (g in seq_len(ngroups)) {
    if (lavmodel@estimator == "DLS" &&
      !isTRUE(lavmodel@estimator.args$dls.FtimesNminus1)) {
      group_w[g] <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
    } else {
      group_w[g] <- (lavsamplestats@nobs[[g]] - 1) / lavsamplestats@ntotal
    }
  }

  # single-level moment-based estimators: share Delta/A1 between the
  # gradient and the information (see lav_optim_gn_eval)
  fast_ls <- (lavdata@nlevels == 1L &&
    lavmodel@estimator %in% c("WLS", "DWLS", "ULS", "DLS") &&
    !lavmodel@group.w.free &&
    !lavsamplestats@missing.flag)

  # generic route: same group-weight switch as lav_model_est
  group_weight <- !(lavsamplestats@missing.flag || lavdata@nlevels > 1L)

  # is the scoring gradient the exact gradient of the objective? not for
  # DLS with a model-based weight matrix (see the top of this file)
  scoring_flag <- (lavmodel@estimator == "DLS" &&
    lavmodel@estimator.args$dls.GammaNT == "model")

  eval_point <- function(x_1, objective_only = TRUE) {
    lav_optim_gn_eval(
      x = x_1, lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavoptions = lavoptions, objective_only = objective_only,
      fast_ls = fast_ls, group_w = group_w, group_weight = group_weight
    )
  }

  # move the gradient/information to the packed space (if needed)
  pack_point <- function(fit) {
    if (eq_flag && !is.null(fit$g)) {
      fit$g <- as.numeric(fit$g %*% eq_k)
      fit$H <- crossprod(eq_k, fit$H %*% eq_k)
    }
    fit
  }

  # working coordinates
  if (eq_flag) {
    z <- as.numeric((x - eq_k0) %*% eq_k)
    lb_z <- rep(-Inf, length(z))
    ub_z <- rep(+Inf, length(z))
  } else {
    z <- x
    lb_z <- lb
    ub_z <- ub
  }
  nz <- length(z)

  # initial evaluation
  cur <- pack_point(eval_point(unpack(z), objective_only = FALSE))
  warn_txt <- ""
  converged <- FALSE
  iter <- 0L

  if (!is.finite(cur$fx) || anyNA(cur$g) || anyNA(cur$H)) {
    x <- unpack(z)
    fx <- lav_model_objective(
      lavmodel = lav_model_set_parameters(lavmodel = lavmodel, x = x),
      lavdata = lavdata, lavsamplestats = lavsamplestats
    )
    attr(x, "converged") <- FALSE
    attr(x, "warn.txt") <- gettext(
      "objective or gradient could not be evaluated at the starting values.")
    attr(x, "iterations") <- 0L
    attr(x, "control") <- list(iter.max = max_iter, tol.x = tol_x,
                               tol.g = tol_g)
    attr(x, "fx") <- fx
    return(x)
  }

  if (lav_verbose()) {
    cat("iter = ", sprintf("%3d", 0L),
      ": obj = ", sprintf("%13.10f", cur$fx), "\n",
      sep = ""
    )
  }

  # scaled norm of the projected gradient (Dennis & Schnabel style
  # relative gradient); the projection removes the components that point
  # outward at an active bound
  relgrad <- function(z_1, fit) {
    pg <- fit$g
    if (bounds_flag) {
      pg[z_1 <= lb_z & pg > 0] <- 0
      pg[z_1 >= ub_z & pg < 0] <- 0
    }
    max(abs(pg) * pmax(abs(z_1), 1)) / max(abs(fit$fx), 1)
  }

  # Levenberg-Marquardt damping state
  mu <- 1e-04
  nu <- 2
  mu_max <- 1e+12
  reject_max <- 25L
  bound_eps <- 1e-12

  while (iter < max_iter) {
    # gradient-based convergence test
    if (relgrad(z, cur) < tol_g) {
      converged <- TRUE
      if (lav_verbose()) {
        cat("Gauss-Newton: relative gradient is below tolerance (",
          sprintf("%g", tol_g), ")\n",
          sep = ""
        )
      }
      break
    }

    iter <- iter + 1L

    # active set: parameters at a bound with an outward-pointing gradient
    # stay put this iteration
    if (bounds_flag) {
      active <- (z <= lb_z + bound_eps & cur$g > 0) |
        (z >= ub_z - bound_eps & cur$g < 0)
    } else {
      active <- logical(nz)
    }
    free_set <- which(!active)
    if (length(free_set) == 0L) {
      # all parameters pinned at a bound; nothing can improve
      converged <- TRUE
      break
    }

    h_ff <- cur$H[free_set, free_set, drop = FALSE]
    g_f <- cur$g[free_set]
    # Marquardt scaling, with a floor for (near-)zero curvature
    d_diag <- diag(h_ff)
    d_floor <- max(abs(d_diag), 1) * 1e-08
    d_diag <- pmax(abs(d_diag), d_floor)

    # inner loop: increase the damping until an acceptable step is found
    accepted <- FALSE
    new_full <- NULL
    n_reject <- 0L
    rho <- as.numeric(NA)
    while (!accepted && n_reject <= reject_max && mu <= mu_max) {
      m_damp <- h_ff
      diag(m_damp) <- diag(m_damp) + mu * d_diag
      r_chol <- try(chol(m_damp), silent = TRUE)
      if (inherits(r_chol, "try-error")) {
        # not positive definite at this damping level: increase and retry
        mu <- mu * nu
        nu <- 2 * nu
        n_reject <- n_reject + 1L
        next
      }
      d_step <- numeric(nz)
      d_step[free_set] <- -1 * backsolve(
        r_chol,
        backsolve(r_chol, g_f, transpose = TRUE)
      )

      # trial point (projected onto the box)
      z_new <- pmin(pmax(z + d_step, lb_z), ub_z)
      d_eff <- z_new - z
      # predicted reduction of the quadratic model along the actual
      # (projected) step
      pred <- -1 * (sum(cur$g * d_eff) +
        0.5 * sum(d_eff * drop(cur$H %*% d_eff)))

      trial <- eval_point(unpack(z_new), objective_only = TRUE)
      ared <- cur$fx - trial$fx

      if (is.finite(trial$fx) && ared > 0) {
        # objective decreased: accept, and relax the damping according
        # to the gain ratio (Nielsen)
        accepted <- TRUE
        if (is.finite(pred) && pred > 0) {
          rho <- ared / pred
        } else {
          rho <- 1
        }
        mu <- max(mu * max(1 / 3, 1 - (2 * rho - 1)^3), 1e-12)
        nu <- 2
      } else if (scoring_flag && is.finite(trial$fx)) {
        # scoring fallback: the (IRLS) fixed point is not the exact
        # minimizer of the monitored objective; accept the step if it
        # shrinks the scoring residual
        new_full <- pack_point(eval_point(unpack(z_new),
          objective_only = FALSE
        ))
        if (!anyNA(new_full$g) && !anyNA(new_full$H) &&
          sqrt(sum(new_full$g^2)) <
            (1 - 1e-04) * sqrt(sum(cur$g^2))) {
          accepted <- TRUE
          new_full$fx <- trial$fx
          rho <- as.numeric(NA)
          nu <- 2
        } else {
          new_full <- NULL
        }
      }

      if (!accepted) {
        mu <- mu * nu
        nu <- 2 * nu
        n_reject <- n_reject + 1L
      }
    } # inner

    if (!accepted) {
      # no acceptable step, even with (very) heavy damping: we are either
      # at a (numerically) stationary point, or stuck
      if (relgrad(z, cur) < sqrt(tol_g)) {
        converged <- TRUE
      } else {
        warn_txt <- gettext(
          "the optimizer could no longer improve the objective
          (Gauss-Newton step rejected at maximum damping).")
      }
      break
    }

    # accepted step: update the state
    x_old <- unpack(z)
    z <- z_new
    x_new <- unpack(z)
    rms_x <- sqrt(mean((x_new - x_old)^2))

    if (is.null(new_full)) {
      new_full <- pack_point(eval_point(x_new, objective_only = FALSE))
      new_full$fx <- trial$fx
    }
    cur <- new_full

    if (lav_verbose()) {
      # (kept within 80 columns)
      cat("iter = ", sprintf("%3d", iter),
        ": obj = ", sprintf("%13.10f", cur$fx),
        " mu = ", sprintf("%7.1e", mu),
        if (is.finite(rho)) {
          paste0(" rho = ", sprintf("%6.3f", rho))
        } else {
          " (scoring)"
        },
        " rms.x = ", sprintf("%12.10f", rms_x), "\n",
        sep = ""
      )
    }

    if (anyNA(cur$g) || anyNA(cur$H)) {
      warn_txt <- gettext(
        "the gradient could not be evaluated at the updated parameter
        values.")
      break
    }

    # step-based convergence test (backward compatible with < 0.7)
    if (rms_x < tol_x) {
      converged <- TRUE
      if (lav_verbose()) {
        cat("Gauss-Newton algorithm converged: rms.x = ",
          sprintf("%12.12f", rms_x), " < ",
          sprintf("%12.12f", tol_x), "\n",
          sep = ""
        )
      }
      break
    }
  } # outer

  x <- unpack(z)

  # one last evaluation, to get the fx.group attribute
  lavmodel <- lav_model_set_parameters(lavmodel = lavmodel, x = x)
  fx <- lav_model_objective(
    lavmodel = lavmodel, lavdata = lavdata,
    lavsamplestats = lavsamplestats
  )

  # add attributes
  if (!converged && iter >= max_iter && !nzchar(warn_txt)) {
    warn_txt <- paste("maximum number of iterations (",
      max_iter, ") ",
      "was reached without convergence.\n",
      sep = ""
    )
  }
  attr(x, "converged") <- converged
  attr(x, "warn.txt") <- warn_txt
  attr(x, "iterations") <- iter
  attr(x, "control") <- list(
    iter.max = max_iter,
    tol.x = tol_x,
    tol.g = tol_g
  )
  attr(x, "fx") <- fx

  x
}

# shared infrastructure for the univariate/bivariate TWO-LEVEL modules:
#
#   lav_uvreg_2l.R  univariate two-level Gaussian (random intercept)
#   lav_uvord_2l.R  univariate two-level ordinal probit (random intercept)
#   lav_bvreg_2l.R  bivariate two-level Gaussian (two-step)
#
# YR 2026 (two-level WLS, phase 2)
#
# These are the stage-wise ('univariate + bivariate') fitters for the
# unrestricted two-level model. The Gaussian kernel below works with the
# per-cluster sufficient statistics
#
#     nj      cluster size
#     ybar_j  cluster mean vector (d)
#     W_j     within-cluster scatter sum (y_ij - ybar_j)(y_ij - ybar_j)'
#
# and evaluates, for Sigma_j = I_nj (x) Sigma_w + J_nj (x) Sigma_b,
#
#  -2 logl_j = nj*d*log(2pi) + (nj-1) log|Sw| + log|Vj|
#              + tr(Sw^{-1} W_j) + nj r_j' Vj^{-1} r_j
#
# with Vj = Sw + nj Sb and r_j = ybar_j - mu. The derivative engine takes
# the pattern matrices (dmu, dSw, dSb) of any parameter and returns the
# per-cluster derivative of -2 logl_j:
#
#  d(-2 logl_j) = (nj-1) tr(Sw^{-1} dSw) - tr(Sw^{-1} dSw Sw^{-1} W_j)
#               + tr(Vj^{-1} dVj) - nj r_j' Vj^{-1} dVj Vj^{-1} r_j
#               - 2 nj r_j' Vj^{-1} dmu
#
# with dVj = dSw + nj dSb.

# per-cluster sufficient statistics
# y: numeric matrix (n x d); cluster_idx: integer vector (1..J)
lav_2l_cluster_stats <- function(y = NULL, cluster_idx = NULL) {
  y <- as.matrix(y)
  d <- NCOL(y)
  njs <- tabulate(cluster_idx)
  nclusters <- length(njs)

  ysum <- rowsum.default(y, group = cluster_idx, reorder = TRUE)
  ybar <- ysum / njs

  # within scatter, stored as vech per cluster (J x d*(d+1)/2)
  yc <- y - ybar[cluster_idx, , drop = FALSE]
  cp_idx <- which(lower.tri(diag(d), diag = TRUE), arr.ind = TRUE)
  wmat <- matrix(0, NROW(y), NROW(cp_idx))
  for (k in seq_len(NROW(cp_idx))) {
    wmat[, k] <- yc[, cp_idx[k, 1]] * yc[, cp_idx[k, 2]]
  }
  wvech <- rowsum.default(wmat, group = cluster_idx, reorder = TRUE)

  list(
    d = d, njs = njs, nclusters = nclusters, nobs = NROW(y),
    ybar = ybar, wvech = wvech,
    wtot = colSums(wmat) # total within scatter (vech)
  )
}

# make a d x d symmetric matrix from its vech
lav_2l_vech2sym <- function(x, d) {
  s <- matrix(0, d, d)
  s[lower.tri(s, diag = TRUE)] <- x
  s[upper.tri(s)] <- t(s)[upper.tri(s)]
  s
}

# per-cluster -2 loglik of the two-level Gaussian model
# cs: output of lav_2l_cluster_stats(); mu (d), sigma_w, sigma_b (d x d)
# returns a vector of length nclusters (or +Inf if not admissible)
lav_2l_gauss_m2ll <- function(cs = NULL, mu = NULL,
                              sigma_w = NULL, sigma_b = NULL) {
  d <- cs$d
  njs <- cs$njs
  nsizes <- sort(unique(njs))

  sw_inv <- try(solve(sigma_w), silent = TRUE)
  det_sw <- det(sigma_w)
  if (inherits(sw_inv, "try-error") || det_sw <= 0) {
    return(rep(+Inf, cs$nclusters))
  }

  # within part: (nj-1) log|Sw| + tr(Sw^{-1} W_j)
  # tr(Sw^{-1} W_j) from the vech: diagonal entries once, off-diag twice
  swi_vech <- sw_inv[lower.tri(sw_inv, diag = TRUE)]
  mult <- ifelse(
    which(lower.tri(diag(d), diag = TRUE)) %in%
      which(lower.tri(diag(d), diag = FALSE)), 2, 1
  )
  tr_w <- drop(cs$wvech %*% (swi_vech * mult))
  out <- njs * d * log(2 * pi) + (njs - 1) * log(det_sw) + tr_w

  # between part per unique cluster size
  r <- cs$ybar - matrix(mu, cs$nclusters, d, byrow = TRUE)
  for (nj in nsizes) {
    idx <- which(njs == nj)
    vj <- sigma_w + nj * sigma_b
    vj_inv <- try(solve(vj), silent = TRUE)
    det_vj <- det(vj)
    if (inherits(vj_inv, "try-error") || det_vj <= 0) {
      return(rep(+Inf, cs$nclusters))
    }
    rj <- r[idx, , drop = FALSE]
    quad <- rowSums((rj %*% vj_inv) * rj)
    out[idx] <- out[idx] + log(det_vj) + nj * quad
  }

  out
}

# per-cluster derivative of -2 logl_j for a list of parameters
# params: list of list(dmu = (d), dsw = (d x d), dsb = (d x d)); any of the
#         three may be NULL (treated as zero)
# returns a (nclusters x length(params)) matrix
lav_2l_gauss_dm2ll <- function(cs = NULL, mu = NULL,
                               sigma_w = NULL, sigma_b = NULL,
                               params = NULL) {
  d <- cs$d
  njs <- cs$njs
  nsizes <- sort(unique(njs))
  npar <- length(params)

  sw_inv <- solve(sigma_w)
  r <- cs$ybar - matrix(mu, cs$nclusters, d, byrow = TRUE)

  out <- matrix(0, cs$nclusters, npar)

  # precompute Sw^{-1} dSw and Sw^{-1} dSw Sw^{-1} per parameter
  for (p in seq_len(npar)) {
    dmu <- params[[p]]$dmu
    dsw <- params[[p]]$dsw
    dsb <- params[[p]]$dsb

    # within part
    if (!is.null(dsw)) {
      a <- sw_inv %*% dsw
      tr_a <- sum(diag(a))
      b <- a %*% sw_inv # Sw^{-1} dSw Sw^{-1}
      b_vech <- b[lower.tri(b, diag = TRUE)]
      mult <- ifelse(
        which(lower.tri(diag(d), diag = TRUE)) %in%
          which(lower.tri(diag(d), diag = FALSE)), 2, 1
      )
      # note: tr(Sw^{-1} dSw Sw^{-1} W_j); B is symmetric iff dSw is
      tr_bw <- drop(cs$wvech %*% (b_vech * mult))
      out[, p] <- out[, p] + (njs - 1) * tr_a - tr_bw
    }

    # between part per cluster size
    for (nj in nsizes) {
      idx <- which(njs == nj)
      vj <- sigma_w + nj * sigma_b
      vj_inv <- solve(vj)
      dvj <- matrix(0, d, d)
      if (!is.null(dsw)) dvj <- dvj + dsw
      if (!is.null(dsb)) dvj <- dvj + nj * dsb
      rj <- r[idx, , drop = FALSE]
      rv <- rj %*% vj_inv
      if (!is.null(dsw) || !is.null(dsb)) {
        out[idx, p] <- out[idx, p] + sum(diag(vj_inv %*% dvj)) -
          njs[idx] * rowSums((rv %*% dvj) * rv)
      }
      if (!is.null(dmu)) {
        out[idx, p] <- out[idx, p] - 2 * njs[idx] * drop(rv %*% dmu)
      }
    }
  }

  out
}

# per-observation GLS-weighted residuals m = Sigma_j^{-1} e (n x d):
#
#     m_i = Sw^{-1} e_i - Sw^{-1} (nj Sb) Vj^{-1} ebar_j
#
# the score of the loglik with respect to a mean parameter of variable a
# with observation-level design column x is sum_i x_i m_ia (cluster-wise:
# rowsum(x * m[, a])); used for the slope parameters (covariates)
lav_2l_gauss_mresid <- function(e = NULL, cluster_idx = NULL,
                                sigma_w = NULL, sigma_b = NULL) {
  e <- as.matrix(e)
  d <- NCOL(e)
  njs <- tabulate(cluster_idx)
  ebar <- rowsum.default(e, group = cluster_idx, reorder = TRUE) / njs
  sw_inv <- solve(sigma_w)
  m <- e %*% sw_inv
  adj <- matrix(0, length(njs), d)
  for (nj in sort(unique(njs))) {
    idx <- which(njs == nj)
    vj_inv <- solve(sigma_w + nj * sigma_b)
    hmat <- sw_inv %*% (nj * sigma_b) %*% vj_inv
    adj[idx, ] <- ebar[idx, , drop = FALSE] %*% t(hmat)
  }
  m - adj[cluster_idx, , drop = FALSE]
}

# standardized Gauss-Hermite nodes for integrating over N(0, 1):
#   integral f(u) phi(u) du ~= sum_q w_q f(x_q)
# (the caller rescales: b_q = sqrt(vb) * x_q)
lav_2l_gh_xw <- function(ngh = 21L) {
  gh <- lav_integration_gauss_hermite(n = ngh, dnorm = TRUE,
                                      mean = 0, sd = 1)
  list(x = gh$x, w = gh$w, logw = log(gh$w))
}

# keep 'old' names for some function names that have been used
# (or are still being used) by external packages
# rsem ==> tools::testinstalledpackages successful without next line
# computeExpectedInformation <- lav_model_information_expected
# only for simsem ==> tools::testinstalledpackages successful without next line

# standardize function names in lav_utils.R / 31 Oct 2025
getCov <- function(x, lower = TRUE, diagonal = TRUE, sds = NULL,  # nolint start
                   names = paste("V", 1:nvar, sep = "")) {
  lav_deprecated("lav_getcov")
  if (diagonal) {
    nvar <- (sqrt(1 + 8 * length(x)) - 1)/2
  }
  else {
    nvar <- (sqrt(1 + 8 * length(x)) + 1)/2
  }
  sc <- sys.call()
  sc[[1L]] <- quote(lavaan::lav_getcov)
  eval(sc, parent.frame())
}
char2num <- function(s = "") {
    lav_deprecated("lav_char2num")
    sc <- sys.call()
    sc[[1L]] <- quote(lavaan::lav_char2num)
    eval(sc, parent.frame())
}
cor2cov <- function(r, sds, names = NULL, ...) {
    lav_deprecated("lav_cor2cov")
    sc <- sys.call()
    sc[[1L]] <- quote(lavaan::lav_cor2cov)
    eval(sc, parent.frame())
}

# standardize function names in lav_mplus_lavaan / 7 Nov 2025
mplus2lavaan.modelSyntax <- function(syntax) {
    lav_deprecated("lav_mplus_syntax_model")
    sc <- sys.call()
    sc[[1L]] <- quote(lavaan::lav_mplus_syntax_model)
    eval(sc, parent.frame())
}
mplus2lavaan  <- function(inpfile, run = TRUE)  {
    lav_deprecated("lav_mplus_lavaan")
    sc <- sys.call()
    sc[[1L]] <- quote(lavaan::lav_mplus_lavaan)
    eval(sc, parent.frame())
}

# standardize function names in lav_partable_vnames.R / 9 December 2025
lavaanNames <- function(object, type = "ov", ...) {
    lav_deprecated("lavNames", times = 0L) #--> for now no warning
    sc <- sys.call()
    sc[[1L]] <- quote(lavaan::lavNames)
    eval(sc, parent.frame())
}

# standardize function names in lav_data_simulate.R / 9 December 2025
simulateData <- function(
                         model = NULL,
                         model_type = "sem",
                         meanstructure = FALSE,
                         int_ov_free = TRUE,
                         int_lv_free = FALSE,
                         marker_int_zero = FALSE,
                         conditional_x = FALSE,
                         composites = TRUE,
                         fixed_x = FALSE,
                         orthogonal = FALSE,
                         std_lv = TRUE,
                         auto_fix_first = FALSE,
                         auto_fix_single = FALSE,
                         auto_var = TRUE,
                         auto_cov_lv_x = TRUE,
                         auto_cov_y = TRUE,
                         ...,
                         sample_nobs = 500L,
                         ov_var = NULL,
                         group_label = NULL,
                         skewness = NULL,
                         kurtosis = NULL,
                         cluster_idx = NULL,
                         seed = NULL,
                         empirical = FALSE,
                         mass = FALSE,
                         ordered_center = TRUE,
                         return_type = "data.frame",
                         return_fit = FALSE,
                         debug = FALSE,
                         standardized = FALSE) {
  lav_deprecated("lavSimulateData", times = 0L)   #--> for now no warning
  sc <- sys.call()
  # rename any old dotted argument names to their snake_case equivalents
  # (e.g. 'group.label' -> 'group_label') at the *name* level, without
  # evaluating the argument expressions. this routes such arguments to the
  # matching formal of the new function instead of into '...', where
  # 'list(...)' would eagerly force them. it preserves backward compatibility
  # for callers that copied lavaan's historical (and unused) default idiom
  # 'group.label = paste("G", 1:ngroups, sep = "")', where 'ngroups' only
  # ever existed inside lavaan: as long as such an argument is never used, it
  # stays a lazy promise and is never evaluated (as it was before 0.6-19).
  nms <- names(sc)
  if (!is.null(nms)) {
    named <- which(nzchar(nms))
    if (length(named) > 0L) {
      names(sc)[named] <- lav_snake_case(nms[named])
    }
  }
  # call new function
  sc[[1L]] <- quote(lavaan::lavSimulateData)
  eval(sc, parent.frame())
}
# standardize function names in lav_bootstrap.R / 9 December 2025
bootstrapLavaan <- function(object,
                            r = 1000L,
                            type = "ordinary",
                            verbose = FALSE,
                            fun = "coef",
                            keep_idx = FALSE,
                            parallel = c("no", "multicore", "snow"),
                            ncpus = min(2L, max(1L, parallel::detectCores() - 2L, na.rm = TRUE)),
                            cl = NULL,
                            iseed = NULL,
                            h0_rmsea = NULL,
                            ...) {
    lav_deprecated("lavBootstrap", times = 0L) #--> for now no warning
    sc <- sys.call()
    # call new function
    sc[[1L]] <- quote(lavaan::lavBootstrap)
    eval(sc, parent.frame())
}

# standardize exported functions  / 13 December 2025
inspectSampleCov <- function(model, data, ...) {
    lav_deprecated("lavInspectSampleCov") #--> for now no warning
    sc <- sys.call()
    sc[[1L]] <- quote(lavaan::lavInspectSampleCov)
    eval(sc, parent.frame())
}                                                                  # nolint end

# parameterestimates <- function(object,
#                                se = TRUE, zstat = TRUE, pvalue = TRUE,
#                                ci = TRUE, standardized = FALSE,
#                                fmi = FALSE, plabel = FALSE,
#                                level = 0.95, boot_ci_type = "perc",
#                                cov_std = TRUE, fmi_options = list(),
#                                rsquare = FALSE,
#                                remove_system_eq = TRUE, remove_eq = TRUE,
#                                remove_ineq = TRUE, remove_def = FALSE,
#                                remove_nonfree = FALSE, remove_step1 = TRUE,
#                                remove_unused = FALSE, add_attributes = FALSE,
#                                output = "data.frame", header = FALSE) {
#   lav_deprecated("lavParameterEstimates", times = 1L) #--> for now no warning
#   sc <- sys.call()
#   sc[[1L]] <- quote(lavaan::lavParameterEstimates)
#   eval(sc, parent.frame())
# }
parameterestimates <- lav_alias("lavParameterEstimates") # alias

# for tidySEM in 0.6-21 only
# vnames <- lav_partable_vnames

## STEP 1b: compute Var(eta) and E(eta) per block
##          only needed for local/fsr approach!
lav_sam_step1_local <- function(step1 = NULL, fit = NULL, y = NULL,
                                sam_method = "local",
                                local_options = list(
                                  M.method = "ML",
                                  lambda.correction = TRUE,
                                  alpha.correction = 0L,
                                  twolevel.method = "h1"
                                ),
                                return_cov_iveta2 = TRUE,
                                return_fs = FALSE) {
  # local.M.method
  local_m_method <- toupper(local_options[["M.method"]])
  if (!local_m_method %in% c("GLS", "ML", "ULS")) {
    lav_msg_stop(gettext(
      "local option M.method should be one of ML, GLS or ULS."))
  }

  lavoptions <- fit@Options
  lavpta <- fit@pta
  nblocks <- lavpta$nblocks

  # flags
  lv_interaction_flag <- FALSE
  lv_higherorder_flag <- FALSE
  if (length(unlist(lavpta$vnames$lv.interaction)) > 0L) {
    lv_interaction_flag <- TRUE
  }
  if (length(unlist(lavpta$vnames$lv.ind)) > 0L) {
    lv_higherorder_flag <- TRUE
  }

  if (lav_verbose()) {
    cat("Constructing the mapping matrix using the ",
      local_m_method, " method ... ",
      sep = ""
    )
  }

  # all the measurement parameters are already stored in PT
  pt_1 <- step1$PT
  if (fit@Model@ceq.simple.only) {
    x_free <- pt_1$est[pt_1$free > 0 & !duplicated(pt_1$free)]
  } else {
    x_free <- pt_1$est[pt_1$free > 0]
  }
  # check for NA values (eg in BETA); set them to zero
  x_free[!is.finite(x_free)] <- 0

  lavmodel_tmp <- lav_model_set_parameters(fit@Model, x = x_free)
  mm_lambda <- mm_theta <- mm_beta <- mm_psi <- mm_nu <- mm_delta <- NULL

  # create LAMBDA
  lambda_idx <- which(names(fit@Model@GLIST) == "lambda")
  mm_lambda <- lavmodel_tmp@GLIST[lambda_idx]

  # create THETA
  theta_idx <- which(names(fit@Model@GLIST) == "theta")
  mm_theta <- lavmodel_tmp@GLIST[theta_idx]

  # NU
  if (fit@Model@meanstructure) {
    nu_idx <- which(names(fit@Model@GLIST) == "nu")
    mm_nu <- lavmodel_tmp@GLIST[nu_idx]
  }

  # DELTA
  if (fit@Model@categorical || fit@Model@correlation) {
    delta_idx <- which(names(fit@Model@GLIST) == "delta")
    mm_delta <- lavmodel_tmp@GLIST[delta_idx]
  }

  # BETA/PSI
  if (lv_higherorder_flag) {
    beta_idx <- which(names(fit@Model@GLIST) == "beta")
    mm_beta <- lavmodel_tmp@GLIST[beta_idx]
    psi_idx <- which(names(fit@Model@GLIST) == "psi")
    mm_psi <- lavmodel_tmp@GLIST[psi_idx]
  }

  # handle dummy's + higher-order + rank-deficient
  lsam_analytic_flag <- rep(TRUE, nblocks)
  l_veta <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    # new in 0.6-10: check if any indicators are also involved
    # in the structural part; if so, set THETA row/col to zero
    # and make sure LAMBDA element is correctly set
    # (we also need to adjust M)
    dummy_ov_idx <- fit@Model@ov.y.dummy.ov.idx[[b]]
    dummy_lv_idx <- fit@Model@ov.y.dummy.lv.idx[[b]]
    if (length(dummy_ov_idx)) {
      mm_theta[[b]][dummy_ov_idx, ] <- 0
      mm_theta[[b]][, dummy_ov_idx] <- 0
      mm_lambda[[b]][dummy_ov_idx, ] <- 0
      mm_lambda[[b]][cbind(dummy_ov_idx, dummy_lv_idx)] <- 1
    }
    if (fit@Model@meanstructure) {
      if (length(dummy_ov_idx)) {
        mm_nu[[b]][dummy_ov_idx, 1] <- 0
      }
    }

    # get ALL lv names (including dummy ov.x/ov.y)
    lv_names <- fit@Model@dimNames[[lambda_idx[b]]][[2L]]

    # handle higher-order factors here
    if (length(lavpta$vidx$lv.ind[[b]]) > 0L) {
      lv_ind_names <- lavpta$vnames$lv.ind[[b]]
      lv_target <- lv_names[!lv_names %in% lv_ind_names]

      target_idx <- match(lv_target, lv_names)
      other_idx <- seq_along(lv_names)[-target_idx]

      this_beta <- mm_beta[[b]]
      this_beta[is.na(this_beta)] <- 0

      ib <- diag(nrow(this_beta)) - this_beta
      ib_inv <- solve(ib)
      lb_inv <- mm_lambda[[b]] %*% ib_inv

      # replace LAMBDA
      mm_lambda[[b]] <- lb_inv[, target_idx, drop = FALSE]

      psi_other <- mm_psi[[b]][other_idx, other_idx, drop = FALSE]
      lb_inv2 <- lb_inv[, other_idx, drop = FALSE]

      # replace THETA
      mm_theta[[b]] <- lb_inv2 %*% psi_other %*% t(lb_inv2) + mm_theta[[b]]
    }

    # check if LAMBDA has full column rank
    this_lambda <- mm_lambda[[b]]
    if (length(lavpta$vidx$lv.interaction[[b]]) > 0L) {
      if (length(lavpta$vidx$lv.ind[[b]]) > 0L) {
        rm_idx <- c(match(lavpta$vnames$lv.ind[[b]], lv_names),
                    match(lavpta$vnames$lv.interaction[[b]], lv_names))
        this_lambda <- this_lambda[, -rm_idx, drop = FALSE]
      } else {
        rm_idx <- match(lavpta$vnames$lv.interaction[[b]], lv_names)
        this_lambda <- this_lambda[, -rm_idx, drop = FALSE]
      }
    }
    if (qr(this_lambda)$rank < ncol(this_lambda)) {
      if (sam_method == "local" && !lv_interaction_flag) {
        lsam_analytic_flag[b] <- FALSE
        # we will try an iterative solution
      } else {
        # eg cfsr or lv interactions: no idea what to do here (yet)
        print(this_lambda)
        lav_msg_stop(gettext(
          "LAMBDA has no full column rank. Please use sam.method = global"))
      }
    }

    # if lambda has full rank, check if cov.lv is unrestricted
    if (!lsam_analytic_flag[b]) {
      veta_symbolic_1 <- lav_sam_veta_pt(fit, block = b)
      # this is the tricky thing: which rows/cols should we remove?
      # none for now
      if (fit@Options$std.lv) {
        veta_symbolic <- lav_mat_vech(veta_symbolic_1, diagonal = FALSE)
      } else {
        veta_symbolic <- lav_mat_vech(veta_symbolic_1, diagonal = TRUE)
      }
      if (any(veta_symbolic == 0)) {
        lsam_analytic_flag[b] <- FALSE
        nfac <- ncol(veta_symbolic_1)
        m_free <- which(veta_symbolic_1 != 0)
        tmp <- lav_mat_vech_rev(lav_mat_vech_idx(nfac))
        x_free <- tmp[which(veta_symbolic_1 != 0)]
        unique_idx <- unique(x_free)
        row_idx <- match(x_free, unique_idx)
        l_psi <- matrix(0L, nrow = nfac * nfac, ncol = length(unique_idx))
        idx <- cbind(m_free, row_idx)
        l_psi[idx] <- 1L
        l_veta[[b]] <- l_psi
      }
    }
  } # b

  # store LAMBDA/THETA/NU per block
  step1$LAMBDA <- mm_lambda
  step1$THETA <- mm_theta
  if (fit@Model@meanstructure) {
    step1$NU <- mm_nu
  }
  if (fit@Model@categorical || fit@Model@correlation) {
    step1$DELTA <- mm_delta
  }

  veta       <- vector("list", nblocks)
  msm_list   <- vector("list", nblocks)
  mtm_list   <- vector("list", nblocks)
  fs_mean    <- vector("list", nblocks)
  fs         <- vector("list", nblocks)
  cov_iveta2 <- vector("list", nblocks)
  rel        <- vector("list", nblocks)
  alpha      <- vector("list", nblocks)
  lambda     <- vector("list", nblocks)
  lambda1    <- vector("list", nblocks)
  if (lavoptions$meanstructure) {
    eeta <- vector("list", nblocks)
  } else {
    eeta <- NULL
  }
  m <- vector("list", nblocks)
  lv_names_1 <- vector("list", nblocks)
  d_std_all <- vector("list", nblocks)

  # compute VETA/EETA per block
  for (b in seq_len(nblocks)) {

    # which group is this?
    this_group <- floor(b / fit@Data@nlevels + 0.5)

    # lv.names, including dummy-lv covariates
    psi_idx <- which(names(fit@Model@GLIST) == "psi")[b]
    lv_names_b <- fit@Model@dimNames[[psi_idx]][[1L]] # including dummy/inter.
    rm_idx <- integer(0L)

    # higher-order? remove lower-order factors
    if (lv_higherorder_flag && length(lavpta$vnames$lv.ind[[b]]) > 0L) {
      rm_idx <- c(rm_idx, match(lavpta$vnames$lv.ind[[b]], lv_names_b))
    }

    # interaction terms? remove them for VETA
    if (lv_interaction_flag && length(lavpta$vnames$lv.interaction[[b]]) > 0L) {
      rm_idx <- c(rm_idx, match(lavpta$vnames$lv.interaction[[b]], lv_names_b))
      lv_int_names <- lavpta$vnames$lv.interaction[[b]]
    }

    # final names for EETA/VETA (not including interaction terms!)
    lv_names1 <- lv_names_b
    if (length(rm_idx) > 0L) {
      lv_names1 <- lv_names_b[-rm_idx]
    }
    lv_names_1[[b]] <- lv_names1

    # get sample statistics for this block
    cov_1 <- step1$COV[[b]]
    ybar <- drop(step1$YBAR[[b]])

    # rescale COV?
    if (fit@Data@nlevels == 1L &&
        (fit@Model@categorical || fit@Model@correlation)) {
      scale_vector <- 1 / (drop(mm_delta[[b]]))
      cov_1 <- scale_vector * cov_1 * rep(scale_vector, each = ncol(cov_1))
      ybar <- scale_vector * ybar # Checkme!
    }

    # do we need ICOV?
    if (local_m_method == "GLS") {
      if (fit@Options$sample.cov.rescale) {
        # get unbiased S
        n <- fit@SampleStats@nobs[[this_group]]
        cov_unbiased <- cov_1 * n / (n - 1)
        icov <- solve(cov_unbiased)
      } else {
        icov <- solve(cov_1)
      }
    }

    # compute mapping matrix 'M'
    this_lambda <- mm_lambda[[b]]
    if (length(lavpta$vidx$lv.interaction[[b]]) > 0L) {
      this_lambda <- this_lambda[, -lavpta$vidx$lv.interaction[[b]]]
    }
    if (lsam_analytic_flag[b]) {
      mb <- lav_sam_mapping_mat(
        mm_lambda = this_lambda,
        mm_theta = mm_theta[[b]],
        s = cov_1, s_inv = icov,
        method = local_m_method
      )
    } else {
      mb <- matrix(as.numeric(NA), ncol(this_lambda), nrow(this_lambda))
    }
    if (length(lavpta$vidx$lv.interaction[[b]]) > 0L) {
      tmp <- mb
      mb <- matrix(0, nrow = ncol(mm_lambda[[b]]), ncol = nrow(mm_lambda[[b]]))
      mb[-lavpta$vidx$lv.interaction[[b]], ] <- tmp
    }

    # handle observed-only variables (needed?)
    dummy_ov_idx <- c(
      fit@Model@ov.x.dummy.ov.idx[[b]],
      fit@Model@ov.y.dummy.ov.idx[[b]]
    )
    dummy_lv_idx <- c(
      fit@Model@ov.x.dummy.lv.idx[[b]],
      fit@Model@ov.y.dummy.lv.idx[[b]]
    )

    # fix dummy.lv.idx if we have higher-order factors!
    if (lv_higherorder_flag) {
      dummy_lv_idx <- match(lv_names_b[dummy_lv_idx], lv_names1)
    }

    if (length(dummy_ov_idx)) {
      mb[dummy_lv_idx, ] <- 0
      mb[cbind(dummy_lv_idx, dummy_ov_idx)] <- 1
    }

    # here, we remove the lv.interaction row(s) from Mb
    # FIXME: if we have higher order factors!
    if (length(lavpta$vidx$lv.interaction[[b]]) > 0L) {
      mb <- mb[-lavpta$vidx$lv.interaction[[b]], ]
    }

    # compute EETA
    if (lavoptions$meanstructure) {
      if (lsam_analytic_flag[b]) {
        eeta[[b]] <- lav_sam_eeta(m = mb, ybar = ybar, mm_nu = mm_nu[[b]])
      } else {
        # EETA is constrained somehow
        lav_msg_stop(gettext("not ready yet"))
      }
      fs_mean[[b]] <- eeta[[b]] # ok if no interaction
    }

    # compute VETA
    if (sam_method == "local") {
      if (lsam_analytic_flag[b]) {
        tmp <- lav_sam_veta(
          m = mb, s = cov_1, mm_theta = mm_theta[[b]],
          alpha_correction = local_options[["alpha.correction"]],
          lambda_correction = local_options[["lambda.correction"]],
          n = fit@SampleStats@nobs[[this_group]],
          dummy_lv_idx = dummy_lv_idx,
          extra = TRUE
        )
        veta[[b]] <- tmp[, , drop = FALSE] # drop attributes
        alpha[[b]]  <- attr(tmp, "alpha")
        lambda[[b]] <- attr(tmp, "lambda.star")
        msm_list[[b]] <- attr(tmp, "MSM")
        mtm_list[[b]] <- attr(tmp, "MTM")
        # effective first-order coefficient on the (pd-forced) MTM matrix,
        # so that VETA = MSM - lambda1 * MTM; this absorbs both the
        # lambda.star correction (1 if the correction was not needed) and
        # the alpha correction (lav_sam_veta() rescales MTM by
        # (1 - alpha/(n-1)) *before* the lambda step); when lv interactions
        # are present, lambda[[b]] will be overwritten by the second-order
        # lambda.star, but the casewise decomposition of EETA2/VETA2 (see
        # lav_sam_veta2() and lav_sam_gamma_add()) needs this first-order
        # value
        nobs_b <- fit@SampleStats@nobs[[this_group]]
        alpha_n1 <- lav_sam_alpha_n1(local_options[["alpha.correction"]],
                                     nobs_b)
        l1 <- if (is.finite(lambda[[b]])) lambda[[b]] else 1
        lambda1[[b]] <- (1 - alpha_n1) * l1
      } else {
        # VETA is constrained somehow
        veta[[b]] <- lav_sam_veta_con(s = cov_1, mm_lambda = mm_lambda[[b]],
                                      mm_theta = mm_theta[[b]],
                                      l_veta = l_veta[[b]],
                                      local_m_method = local_m_method)
        alpha[[b]]   <- as.numeric(NA)
        lambda[[b]]  <- as.numeric(NA)
        lambda1[[b]] <- as.numeric(NA)
        msm_list[[b]] <- matrix(0, 0, 0)
        mtm_list[[b]] <- matrix(0, 0, 0)
      }
    } else if (sam_method == "cfsr") {
      # first, we need the 'true' VETA (to get Sigma)
      tmp <- lav_sam_veta(
        m = mb, s = cov_1, mm_theta = mm_theta[[b]],
        alpha_correction = 0L,
        lambda_correction = local_options[["lambda.correction"]],
        n = fit@SampleStats@nobs[[this_group]],
        dummy_lv_idx = dummy_lv_idx,
        extra = FALSE
      )
      veta[[b]] <- tmp[, , drop = FALSE]
      # compute 'Sigma'
      sigma_1 <- this_lambda %*% veta[[b]] %*% t(this_lambda) + mm_theta[[b]]
      tmat <- lav_predict_tmat_det_internal(sigma_1 = sigma_1, veta = veta[[b]],
                                            lambda = this_lambda)
      a <- tmat %*% mb
      veta[[b]] <- a %*% cov_1 %*% t(a)
    } else {
      # FSR -- no correction
      veta[[b]] <- mb %*% cov_1 %*% t(mb)
    }

    # standardize? not really needed, but we may have 1.0000001
    # as variances, and this may lead to false convergence
    # d_std holds the rescaling factors (1 = no rescaling); when lv
    # interactions are present, we will need them to express the factor
    # scores (and related matrices) in the same standardized metric
    d_std <- rep(1, ncol(veta[[b]]))
    if (fit@Options$std.lv) {
      # we should only standardize the LVs, not the observed variables
      # (dummy lvs); but note that the cross-blocks between the dummy lvs
      # and the latent variables must be rescaled by the factor-side
      # scaling factors as well (new in June 2026; before, these
      # cross-blocks were left unrescaled, and -- when interaction terms
      # were present -- the dummy lvs were accidentally standardized too,
      # because dummy_lv_idx refers to the full set of lv names, while
      # veta[[b]] only contains lv_names1)
      dummy_lv_idx1 <- dummy_lv_idx
      if (!lv_higherorder_flag) { # if higher-order, already remapped
        dummy_lv_idx1 <- match(lv_names_b[dummy_lv_idx], lv_names1)
      }
      if (length(dummy_lv_idx1) == 0L) {
        d_std <- sqrt(diag(veta[[b]]))
        veta[[b]] <- stats::cov2cor(veta[[b]])
      } else {
        d_std[-dummy_lv_idx1] <- sqrt(diag(veta[[b]]))[-dummy_lv_idx1]
        veta[[b]] <- t(veta[[b]] / d_std) / d_std
      }
    }
    d_std_all[[b]] <- d_std
    colnames(veta[[b]]) <- rownames(veta[[b]]) <- lv_names1

    # compute model-based RELiability
    if (lsam_analytic_flag[b]) {
      rel[[b]] <- diag(veta[[b]]) / diag(msm_list[[b]])
    } else {
      rel[[b]] <- as.numeric(NA)
    }

    # check for lv.interactions
    if (lv_interaction_flag && length(lv_int_names) > 0L) {
      if (fit@Model@categorical || fit@Model@correlation) {
        lav_msg_stop(gettext("SAM + lv interactions do not work (yet) if
                             correlation structures are used."))
      }

      # compute Bartlett factor scores here
      if (is.null(y)) {
        yb <- fit@Data@X[[b]]
      } else {
        yb <- y[[b]]
      }
      # center
      yb_c <- t(t(yb) - drop(mm_nu[[b]]))
      fs_b <- yb_c %*% t(mb)
      colnames(fs_b) <- lv_names1
      # FIXME: what about observed covariates?

      # if std.lv = TRUE, express the factor scores, EETA and the mapping
      # matrix in the same standardized metric as the cov2cor() rescaled
      # VETA above, so that the augmented statistics EETA2/VETA2 (and
      # their casewise decomposition in lav_sam_veta2()) are
      # metric-coherent
      eeta1 <- eeta[[b]]
      mb_v2 <- mb
      if (fit@Options$std.lv) {
        fs_b <- t(t(fs_b) / d_std)
        eeta1 <- eeta1 / d_std
        mb_v2 <- mb / d_std
      }

      # EETA2
      eeta[[b]] <- lav_sam_eeta2(
        eeta = eeta1, veta = veta[[b]],
        lv_names = lv_names1,
        lv_int_names = lv_int_names
      )

      # VETA2
      if (sam_method == "local") {
        tmp <- lav_sam_veta2(
          fs = fs_b, m = mb_v2,
          veta = veta[[b]], eeta = eeta1,
          mm_theta = mm_theta[[b]],
          lv_names = lv_names1,
          lv_int_names = lv_int_names,
          dummy_lv_names = lv_names_b[dummy_lv_idx],
          alpha_correction = local_options[["alpha.correction"]],
          lambda_correction = local_options[["lambda.correction"]],
          lambda1 = lambda1[[b]],
          return_fs = return_fs,
          return_cov_iveta2 = return_cov_iveta2,
          extra = TRUE
        )
        veta[[b]] <- tmp[, , drop = FALSE] # drop attributes
        alpha[[b]] <- attr(tmp, "alpha")
        lambda[[b]] <- attr(tmp, "lambda.star")
        msm_list[[b]] <- attr(tmp, "MSM")
        mtm_list[[b]] <- attr(tmp, "MTM")
        fs_mean[[b]] <- attr(tmp, "FS.mean")
        if (return_fs) {
          fs[[b]] <- attr(tmp, "FS")
        }
        if (return_cov_iveta2) {
          cov_iveta2[[b]] <- attr(tmp, "cov.iveta2")
        }
      } else {
        lav_msg_fixme("not ready yet!")
        # FSR -- no correction
        veta[[b]] <- lav_sam_fs2(
          fs = fs_b,
          lv_names = lv_names1, lv_int_names = lv_int_names
        )
      }
    }

    # store Mapping matrix for this block
    m[[b]] <- mb
  } # blocks

  # label blocks
  if (nblocks > 1L) {
    if (!is.null(eeta)) {
      names(eeta)     <- fit@Data@block.label
    }
    names(veta)       <- fit@Data@block.label
    names(rel)        <- fit@Data@block.label
    names(msm_list)   <- fit@Data@block.label
    names(mtm_list)   <- fit@Data@block.label
    names(fs_mean)    <- fit@Data@block.label
    names(fs)         <- fit@Data@block.label
    names(cov_iveta2) <- fit@Data@block.label
  }

  # handle conditional.x: add res.slopes, cov.x and mean.x
  if (fit@Model@conditional.x) {
    if (lv_interaction_flag) {
      lav_msg_stop(gettext(
        "sam() does not support latent interactions in combination with
        conditional.x = TRUE (yet)."))
    }
    res_slopes <- vector("list", length = nblocks)
    cov_x  <- fit@h1$implied$cov.x
    mean_x <- fit@h1$implied$mean.x
    for (b in seq_len(nblocks)) {
      x_names <- fit@Data@ov.names.x[[b]]
      # y-level slopes: read from step1$SLOPES (set by lav_sam_get_cov_ybar,
      # and perturbed by the categorical numeric jacobian); fall back to the
      # h1 implied slopes
      slopes_b <- step1$SLOPES[[b]]
      if (is.null(slopes_b)) {
        slopes_b <- fit@h1$implied$res.slopes[[b]]
      }
      # categorical/correlation: rescale to the model's latent-response
      # metric, exactly like COV/YBAR above
      if (fit@Data@nlevels == 1L &&
          (fit@Model@categorical || fit@Model@correlation)) {
        slopes_b <- (1 / drop(mm_delta[[b]])) * slopes_b
      }
      # slopes of the latent variables on the exogenous covariates; if
      # std.lv = TRUE, put them (and EETA, the latent intercepts) on the same
      # standardized metric as the cov2cor() rescaled VETA above, so that the
      # full conditional moment vector (intercepts, slopes, residual
      # covariances) is metric-coherent
      res_slopes[[b]] <- (m[[b]] %*% slopes_b) / d_std_all[[b]]
      if (fit@Options$std.lv && !is.null(eeta)) {
        eeta[[b]] <- eeta[[b]] / d_std_all[[b]]
      }
      # (dim)names, so that the step-2 summary-statistics call can align
      # these statistics with its own internal variable order
      rownames(res_slopes[[b]]) <- colnames(veta[[b]])
      colnames(res_slopes[[b]]) <- x_names
      dimnames(cov_x[[b]]) <- list(x_names, x_names)
      mean_x[[b]] <- drop(mean_x[[b]])
      names(mean_x[[b]]) <- x_names
    }
    attr(veta, "res.slopes") <- res_slopes
    attr(veta, "cov.x") <- cov_x
    attr(veta, "mean.x") <- mean_x
    # needed by the (analytic) conditional.x JACb in the Gamma.eta machinery
    step1$D.STD <- d_std_all
  }

  # store EETA/VETA/M/alpha/lambda
  step1$VETA     <- veta
  step1$EETA     <- eeta
  step1$REL      <- rel
  step1$M        <- m
  step1$lambda   <- lambda
  step1$lambda1  <- lambda1
  step1$alpha    <- alpha
  step1$MSM      <- msm_list
  step1$MTM      <- mtm_list
  step1$FS.mean  <- fs_mean
  step1$FS       <- fs
  step1$COV.IVETA2 <- cov_iveta2
  step1$LV.NAMES <- lv_names_1
  # store also sam.method and local.options
  step1$sam.method <- sam_method
  step1$local.options <- local_options

  if (lav_verbose()) {
    cat("done.\n")
  }

  step1
}


# Memory-lean Gamma.eta for group g: JAC %*% Gamma %*% t(JAC) without ever
# forming the p* x p* observed Gamma. Returns NULL when the (continuous, simple)
# setting does not apply, so the caller falls back to the explicit Gamma.
#
# The ADF Gamma of the WLS.obs statistics is the casewise crossprod
#   Gamma = crossprod(z) / n.eff,   z_i = [ y_i (meanstructure),
#                                           vech((y_i - ybar)(y_i - ybar)') ]
# so JAC %*% Gamma %*% t(JAC) = crossprod(z %*% t(JAC)) / n.eff (the U below).
#
# With the unbiased (Browne) Gamma -- the sam() default -- the cov block also
# needs the normal-theory Gamma and a rank-1 term:
#   Gamma_u_cov = c1*Gamma_cov - c2*(Gamma_NT - 2/(n-1)*tcrossprod(vech(S)))
# with c1 = n(n-1)/((n-2)(n-3)), c2 = n/((n-2)(n-3)). The NT sandwich is built
# without the p* x p* Gamma_NT via [JAC Gamma_NT JAC']_rc = 2*tr(G_r S G_c S),
# where G_r is the symmetric matrix with vech G_r = (cov rows of) JAC[r, ]
# (off-diagonals halved). The mean-mean block is unchanged and the mean-cov
# block is scaled by n/(n-2).
lav_sam_gamma_eta_g <- function(fit = NULL, jac_g = NULL, g = 1L) {
  lavoptions <- fit@Options
  lavmodel <- fit@Model
  unbiased <- isTRUE(lavoptions$gamma.unbiased)
  if (lavmodel@categorical || lavmodel@correlation || lavmodel@conditional.x ||
      (isTRUE(lavoptions$fixed.x) && length(fit@Data@ov.names.x[[g]]) > 0L) ||
      fit@Data@nlevels > 1L || length(fit@Data@cluster) > 0L ||
      (unbiased && isTRUE(lavoptions$gamma.n.minus.one))) {
    return(NULL)
  }
  m_y <- unname(as.matrix(fit@Data@X[[g]]))
  if (anyNA(m_y)) {
    return(NULL) # missing data: leave it to the (pairwise) NACOV path
  }
  n <- nrow(m_y)
  p <- ncol(m_y)
  meanstr <- lavmodel@meanstructure
  m_yc <- t(t(m_y) - colMeans(m_y))
  idx1 <- lav_mat_vech_col_idx(p)
  idx2 <- lav_mat_vech_row_idx(p)
  zc_cov <- m_yc[, idx1, drop = FALSE] * m_yc[, idx2, drop = FALSE]
  zc_cov <- t(t(zc_cov) - colMeans(zc_cov))

  if (meanstr) {
    jac_m <- jac_g[, seq_len(p), drop = FALSE]
    jac_c <- jac_g[, -seq_len(p), drop = FALSE]
    u_m <- m_yc %*% t(jac_m)
  } else {
    jac_c <- jac_g
  }
  u_c <- zc_cov %*% t(jac_c)

  # biased ADF Gamma: simple casewise crossprod
  if (!unbiased) {
    n_eff <- if (isTRUE(lavoptions$gamma.n.minus.one)) n - 1L else n
    u <- if (meanstr) u_m + u_c else u_c
    return(crossprod(u) / n_eff)
  }

  # unbiased (Browne): cov block needs the NT Gamma + rank-1 correction
  s_cov <- crossprod(m_yc) / n
  cov_vech <- lav_mat_vech(s_cov)
  mstar <- nrow(jac_g)
  offu <- which(idx1 != idx2)
  # G_mat: rows = vec(G_r), the symmetric matrix with halved off-diagonals
  g_mat <- matrix(0, mstar, p * p)
  lin_11 <- (idx1 - 1L) * p + idx2
  lin_22 <- (idx2 - 1L) * p + idx1
  g_mat[, lin_11] <- jac_c
  g_mat[, lin_22] <- jac_c
  g_mat[, lin_11[offu]] <- jac_c[, offu, drop = FALSE] / 2
  g_mat[, lin_22[offu]] <- jac_c[, offu, drop = FALSE] / 2
  # M_mat: rows = vec(S G_r S)
  m_mat <- t(apply(g_mat, 1L, function(gv) {
    as.vector(s_cov %*% matrix(gv, p, p) %*% s_cov)
  }))
  gamma_nt <- 2 * tcrossprod(g_mat, m_mat) # JAC Gamma_NT JAC'  (mstar x mstar)
  d_vec <- drop(jac_c %*% cov_vech)
  c1 <- n * (n - 1) / ((n - 2) * (n - 3))
  c2 <- n / ((n - 2) * (n - 3))
  ge_cov <- c1 * (crossprod(u_c) / n) -
    c2 * (gamma_nt - 2 / (n - 1) * tcrossprod(d_vec))
  if (!meanstr) {
    return(ge_cov)
  }
  ge_mm <- crossprod(u_m) / n
  ge_mc <- (n / (n - 2)) * (crossprod(u_m, u_c) / n)
  ge_mm + ge_cov + ge_mc + t(ge_mc)
}

# Shared helpers for lav_sam_step1_local_jac() (single group) and
# lav_sam_step1_local_jac_mg() (multigroup) -- they build the same jacobian
# (JAC = JACc %*% JACa + JACb [+ psi channel]); the multigroup version simply
# stacks the per-group statistics and structural moments. Factoring the common
# pieces out keeps the two entry points thin and avoids drift between them.

# Indices (in step1$PT) of the free 'measurement' parameters whose influence
# d(theta.mm)/d(stats) forms the rows of JACa: factor loadings (=~), residual
# (co)variances of observed indicators (~~), intercepts (~1, if meanstructure),
# scaling factors (~*~) and thresholds (|, categorical), plus the higher-order
# loadings (=~ on an lv indicator) and latent (co)variances (~~ between lv
# indicators). One row per free parameter: by default de-duplicated on the
# (constrained) pt_1$free numbering; pass dedup_key = step1$PT.free to de-dup on
# the *unconstrained* numbering instead (needed for the p_only 'P' matrix under
# ceq.simple, where across-group equality-constrained parameters share a
# constrained free number but are distinct rows of P / step1.free.idx).
lav_sam_meas_keep_idx <- function(pt_1, ov_names, lavmodel, lv_ind,
                                  meanstructure, dedup_key = NULL) {
  if (is.null(dedup_key)) {
    dedup_key <- pt_1$free
  }
  dup <- duplicated(dedup_key)
  lambda_idx <- which(pt_1$op == "=~" & pt_1$free > 0L & !dup)
  theta_idx  <- which(pt_1$op == "~~" & pt_1$free > 0L & !dup &
                      pt_1$lhs %in% ov_names & pt_1$rhs %in% ov_names)
  nu_idx <- integer(0L)
  if (meanstructure) {
    nu_idx <- which(pt_1$op == "~1" & pt_1$free > 0L & !dup &
                    pt_1$lhs %in% ov_names)
  }
  delta_idx <- integer(0L)
  if (lavmodel@categorical || lavmodel@correlation) {
    delta_idx <- which(pt_1$op == "~*~" & pt_1$free > 0L & !dup)
  }
  th_idx <- integer(0L)
  if (lavmodel@categorical) {
    th_idx <- which(pt_1$op == "|" & pt_1$free > 0L & !dup)
  }
  beta_idx <- psi_idx <- integer(0L)
  if (length(lv_ind) > 0L) {
    beta_idx <- which(pt_1$op == "=~" & pt_1$free > 0L & !dup &
                      pt_1$rhs %in% lv_ind)
    psi_idx <- which(pt_1$op == "~~" & pt_1$free > 0L & !dup &
                     pt_1$rhs %in% lv_ind & pt_1$lhs %in% lv_ind)
  }
  sort(c(lambda_idx, theta_idx, nu_idx, delta_idx, th_idx, beta_idx, psi_idx))
}

# Stack the structural moments c(EETA_g, vech(VETA_g)) over the requested groups
# (a single g for a per-group JACb, or all groups for JACc/psi). This is the
# function whose jacobian (w.r.t. the sample statistics / measurement params)
# the numeric-difference helpers below compute.
#
# conditional.x: the structural moments follow the conditional WLS.obs layout
# c( vecr(cbind(EETA, RS_eta)), vech(VETA) ), where RS_eta = the res.slopes
# attribute of VETA (slopes of the latent variables on the exogenous
# covariates) and vecr interleaves per variable: (int_1, b_11..b_1q, int_2,
# ...). The variables are in VETA (measurement) order here; the permutation to
# the structural model's internal order happens once, on Gamma.eta (see
# lav_sam_condx_perm_idx()).
lav_sam_veta_vec <- function(step1_obj, groups, meanstructure,
                             conditional_x = FALSE) {
  unlist(lapply(groups, function(g) {
    if (conditional_x) {
      rs_g <- attr(step1_obj$VETA, "res.slopes")[[g]]
      c(lav_mat_vecr(cbind(step1_obj$EETA[[g]], rs_g)),
        lav_mat_vech(step1_obj$VETA[[g]]))
    } else if (meanstructure) {
      c(step1_obj$EETA[[g]], lav_mat_vech(step1_obj$VETA[[g]]))
    } else {
      lav_mat_vech(step1_obj$VETA[[g]])
    }
  }))
}

# Numeric jacobian of the stacked structural moments w.r.t. a set of
# free-parameter rows of step1$PT$est (rows_idx). Used for JACc (rows_idx =
# keep_idx) and the psi channel (rows_idx = the joint factor (co)variance rows).
lav_sam_jac_est_rows <- function(rows_idx, step1, fit, groups, meanstructure) {
  conditional_x <- fit@Model@conditional.x
  ff <- function(x) {
    step1_1 <- step1
    step1_1$PT$est[rows_idx] <- x
    step1_1 <- lav_sam_step1_local(step1 = step1_1, fit = fit,
      sam_method = step1$sam.method, local_options = step1$local.options)
    lav_sam_veta_vec(step1_1, groups, meanstructure, conditional_x)
  }
  verbose_flag <- lav_verbose()
  lav_verbose(FALSE)
  on.exit(lav_verbose(verbose_flag))
  numDeriv::jacobian(func = ff, x = step1$PT$est[rows_idx])
}

# Categorical JACb for one group g: numeric jacobian of vech(VETA_g) w.r.t. the
# off-diagonal (poly)choric correlations and the free continuous variances of
# step1$COV[[g]] (keeping theta.mm fixed), with the columns mapped (by label)
# into group g's joint WLS.obs vector. Thresholds/means stay zero. Under
# conditional.x the y-on-x slopes (step1$SLOPES[[g]], raw probit metric) are
# perturbed as well: the structural moments include RS_eta = M %*% slopes.
lav_sam_jacb_cat_g <- function(g, step1, fit, joint_labels_g, meanstructure) {
  conditional_x <- fit@Model@conditional.x
  cov_g <- step1$COV[[g]]
  low_idx <- lower.tri(cov_g) # strict lower triangle (column-major)
  n_low <- sum(low_idx)
  nvar_g <- ncol(cov_g)
  # continuous variables: their variances (the diagonal) are free statistics
  num_idx <- which(!seq_len(nvar_g) %in% fit@SampleStats@th.idx[[g]])
  x0 <- c(cov_g[low_idx], diag(cov_g)[num_idx])
  n_covvar <- length(x0)
  if (conditional_x) {
    slopes_g <- step1$SLOPES[[g]]
    x0 <- c(x0, as.vector(slopes_g)) # column-major, like lav_mat_vec()
  }
  ffb_cat <- function(xx) {
    step1_1 <- step1
    this_cov <- step1$COV[[g]]
    this_cov[low_idx] <- xx[seq_len(n_low)]
    this_cov[upper.tri(this_cov)] <- t(this_cov)[upper.tri(this_cov)]
    if (length(num_idx) > 0L) {
      dd <- diag(this_cov)
      dd[num_idx] <- xx[n_low + seq_along(num_idx)]
      diag(this_cov) <- dd
    }
    step1_1$COV[[g]] <- this_cov
    if (conditional_x) {
      step1_1$SLOPES[[g]] <- matrix(xx[-seq_len(n_covvar)],
                                    nrow = nrow(slopes_g),
                                    ncol = ncol(slopes_g))
    }
    step1_1 <- lav_sam_step1_local(step1 = step1_1, fit = fit,
      sam_method = step1$sam.method, local_options = step1$local.options)
    lav_sam_veta_vec(step1_1, g, meanstructure, conditional_x)
  }
  verbose_flag <- lav_verbose()
  lav_verbose(FALSE)
  jacb_corr <- numDeriv::jacobian(func = ffb_cat, x = x0)
  lav_verbose(verbose_flag)

  # map the perturbed-statistic columns into group g's joint WLS.obs vector (by
  # label, so any difference in variable ordering is handled)
  ov_g <- colnames(cov_g)
  if (is.null(ov_g)) {
    ov_g <- fit@Data@ov.names[[g]]
  }
  r_idx <- row(cov_g)[low_idx]
  c_idx <- col(cov_g)[low_idx]
  var_labels <- if (length(num_idx) > 0L) {
    paste0(ov_g[num_idx], "|var")
  } else {
    character(0L)
  }
  my_labels <- c(vapply(seq_along(r_idx), function(k) {
    paste(sort(c(ov_g[r_idx[k]], ov_g[c_idx[k]])), collapse = "~~")
  }, character(1L)), var_labels)
  if (conditional_x) {
    x_names_g <- fit@Data@ov.names.x[[g]]
    my_labels <- c(my_labels,
      paste0(rep(ov_g, times = length(x_names_g)), "~",
             rep(x_names_g, each = nvar_g)))
  }
  col_idx <- match(my_labels, joint_labels_g)
  if (anyNA(col_idx)) {
    lav_msg_stop(gettext(
      "internal error: unable to map the polychoric correlations into the
       statistics vector of the joint model (categorical JACb)"))
  }
  jb <- matrix(0, nrow = nrow(jacb_corr),
               ncol = length(fit@SampleStats@WLS.obs[[g]]))
  jb[, col_idx] <- jacb_corr
  jb
}

# Continuous JACb for one group g: vech(VETA_g) = M_g S_g M_g', so
# d vech(VETA_g) / d vech(S_g) is the (row-selected) duplication-post product of
# M_g %x% M_g, with a leading M_g block for the means (if meanstructure).
lav_sam_jacb_cont_g <- function(mb, meanstructure) {
  mbx_mb <- mb %x% mb
  row_idx <- lav_mat_vech_idx(nrow(mb))
  jb <- lav_mat_dup_post(mbx_mb)[row_idx, , drop = FALSE]
  if (meanstructure) {
    jb <- lav_mat_bdiag(mb, jb)
  }
  jb
}

# Continuous JACb for one group g under conditional.x. Rows follow the
# conditional structural-moment layout c(vecr(cbind(EETA, RS_eta)),
# vech(VETA)) of lav_sam_veta_vec(); columns follow the joint conditional
# WLS.obs layout c(vecr(cbind(res.int, res.slopes)), vech(res.cov)).
# With C_eta = cbind(EETA, RS_eta) = M_g %*% C_y - [M_g nu, 0] and
# vecr(A) = vec(A'):
#   d vecr(C_eta) / d vecr(C_y)     = M_g %x% I_(q+1)
#   d vech(VETA)  / d vech(res.cov) = as in the unconditional case
# (theta.mm -- and hence M_g -- is held fixed here; its influence goes
# through the JACc channel). If std.lv = TRUE, the intercept/slope rows are
# on the same standardized metric as VETA (divide the M_g rows by d_std,
# matching the res.slopes attribute); the vech(VETA) block keeps the raw
# M_g %x% M_g map, as in the unconditional case.
lav_sam_jacb_cont_condx_g <- function(mb, nexo, d_std = NULL) {
  mb_m <- mb
  if (!is.null(d_std)) {
    mb_m <- mb / d_std
  }
  jb_mean <- mb_m %x% diag(nexo + 1L)
  row_idx <- lav_mat_vech_idx(nrow(mb))
  jb_cov <- lav_mat_dup_post(mb %x% mb)[row_idx, , drop = FALSE]
  lav_mat_bdiag(jb_mean, jb_cov)
}

# conditional.x, continuous measurement blocks: the blocks themselves are fit
# unconditionally (their unconditional sample moments are exact functions of
# the joint conditional statistics, given the -- fixed -- moments of the
# exogenous covariates):
#   ybar_B = res.int_B + res.slopes_B %*% mean.x
#   S_BB   = res.cov_BB + res.slopes_B %*% cov.x %*% res.slopes_B'
# This helper returns the jacobian T of the block statistics c(ybar_B,
# vech(S_BB)) w.r.t. the joint conditional WLS.obs vector, so that the
# block influence (columns = block statistics) can be mapped into the joint
# conditional statistics space as mm_jac %*% T.
lav_sam_condx_t_g <- function(fit, g, ov_block, meanstructure) {
  ov_nox <- fit@pta$vnames$ov.nox[[g]]
  ov_x <- fit@Data@ov.names.x[[g]]
  ny <- length(ov_nox)
  q <- length(ov_x)
  bs <- fit@h1$implied$res.slopes[[g]] # ny x q
  bsx <- bs %*% fit@h1$implied$cov.x[[g]] # ny x q
  xbar <- drop(fit@h1$implied$mean.x[[g]])

  y_idx <- match(ov_block, ov_nox)
  if (anyNA(y_idx)) {
    lav_msg_stop(gettext(
      "internal error: unable to map the measurement-block variables into
      the joint conditional statistics vector (conditional.x JACa)"))
  }
  nb <- length(y_idx)
  ncol_t <- ny * (1L + q) + ny * (ny + 1L) / 2L
  cov_off <- ny * (1L + q)
  # column index of the intercept / slopes of joint y-variable j
  slope_cols <- function(j) (j - 1L) * (q + 1L) + 1L + seq_len(q)
  # vech position of the joint pair (j, l)
  pos <- matrix(0L, ny, ny)
  pos[lower.tri(pos, diag = TRUE)] <- seq_len(ny * (ny + 1L) / 2L)
  pos[upper.tri(pos)] <- t(pos)[upper.tri(pos)]

  t_mean <- NULL
  if (meanstructure) {
    t_mean <- matrix(0, nb, ncol_t)
    for (i in seq_len(nb)) {
      j <- y_idx[i]
      t_mean[i, (j - 1L) * (q + 1L) + 1L] <- 1
      if (q > 0L) {
        t_mean[i, slope_cols(j)] <- xbar
      }
    }
  }
  r_i <- lav_mat_vech_row_idx(nb)
  c_i <- lav_mat_vech_col_idx(nb)
  t_cov <- matrix(0, nb * (nb + 1L) / 2L, ncol_t)
  for (k in seq_len(nrow(t_cov))) {
    j <- y_idx[r_i[k]]
    l <- y_idx[c_i[k]]
    t_cov[k, cov_off + pos[j, l]] <- 1
    if (q > 0L) {
      t_cov[k, slope_cols(j)] <- t_cov[k, slope_cols(j)] + bsx[l, ]
      t_cov[k, slope_cols(l)] <- t_cov[k, slope_cols(l)] + bsx[j, ]
    }
  }
  rbind(t_mean, t_cov)
}

lav_sam_step1_local_jac <- function(step1 = NULL, fit = NULL, p_only = FALSE,
                                    return_jac = FALSE) {

  lavdata <- fit@Data
  # lavsamplestats <- fit@SampleStats
  lavmodel <- fit@Model
  lavpta <- fit@pta
  nblocks <- lavpta$nblocks

  # local_options <- step1$local.options
  # sam_method <- step1$sam.method

  ngroups <- lavdata@ngroups
  # categorical local SEs: supported for the single-group setting. The
  # categorical d(VETA)/d(stats) jacobian needs an extra channel that the
  # continuous case does not: in the delta (correlation) parameterization the
  # residual variances theta are not free but are pinned by the unit-diagonal
  # constraint theta_jj = 1 - (Lambda Psi Lambda')_jj. Hence VETA also depends
  # on the within-block factor (co)variances Psi (through theta, and so through
  # the mapping matrix M). Psi is not a measurement parameter in 'keep_idx', so
  # this 'psi channel' is otherwise dropped; without it the latent (co)variance
  # SEs come out ~0.6x of their true value. It is added to JAC below (the
  # continuous case is unaffected, because there theta is free and
  # d(VETA)/d(Psi) = 0). Block-wise mixed models (some measurement blocks fully
  # categorical, others fully continuous) are supported: the JACa loop picks the
  # influence path per block, and JACb also perturbs the free continuous
  # variances. Higher-order categorical models are supported too (the structural
  # part reduces to the second-order factor (co)variances, as in the continuous
  # case). Under conditional.x the measurement blocks are themselves fitted
  # with conditional.x = TRUE (same latent-response scale as the joint model;
  # see lav_sam_step1()), the joint statistics gain the y-on-x slope entries
  # (label "y~x"), and JACb also perturbs step1$SLOPES. Block-wise mixed +
  # conditional.x is stopped in lav_sam_step1().

  # multiple groups: dispatch to the (single-level, no latent interaction)
  # multigroup implementation, which builds one stacked jacobian over all
  # groups and returns a list of per-group jacobians. Across-group (equality)
  # constraints in the measurement model (Case B) are detected there (via the
  # cross-group blocks of the stacked jacobian) and handled with the full
  # cross-group Gamma in step 2 (SEs and the robust test statistic).
  if (ngroups > 1L) {
    if (lavdata@nlevels > 1L) {
      lav_msg_stop(gettext(
        "local SEs: not available for multigroup multilevel models (yet)!"))
    }
    if (lavmodel@conditional.x) {
      lav_msg_stop(gettext(
        "local SEs: not available with multiple groups in combination with
         conditional.x = TRUE (yet)!"))
    }
    if (length(unlist(lavpta$vnames$lv.interaction)) > 0L) {
      lav_msg_stop(gettext(
        "local SEs: not available for latent interactions with multiple
         groups (yet)!"))
    }
    return(lav_sam_step1_local_jac_mg(step1 = step1, fit = fit,
                                      p_only = p_only,
                                      return_jac = return_jac))
  }

  g <- 1L
  n_mmblocks <- length(step1$MM.FIT)

  # categorical: we map the statistics of each measurement block into the
  # WLS statistics vector of the joint model by label
  if (lavmodel@categorical) {
    joint_labels <- lav_sam_wls_obs_labels(
      ov_names = if (lavmodel@conditional.x) {
        lavpta$vnames$ov.nox[[g]]
      } else {
        lavdata@ov.names[[g]]
      },
      th_idx = fit@SampleStats@th.idx[[g]],
      ov_names_x = if (lavmodel@conditional.x) {
        lavdata@ov.names.x[[g]]
      } else {
        NULL
      }
    )
  }

  # JAC = (JACc %*% JACa) + JACb
  # - rows are the elements of vech(VETA)
  # - cols are the elements of vech(S)

  # JACa: mm.theta x vech(S)
  # JACc: vech(VETA) x mm.theta (keeping S fixed)
  # JACb: vech(VETA) x vech(S)  (keeping mm.theta fixed)

  # JACa: jacobian of theta.mm = f(vech(S))
  jaca <- matrix(0, nrow = length(fit@ParTable$lhs), # we select later
                    ncol = length(fit@SampleStats@WLS.obs[[g]]))
  # 'psi channel' accumulators (categorical only): the influence d(Psi)/d(stats)
  # for the within-block factor (co)variances, and their joint partable rows
  psi_infl_list <- list()
  psi_jrow_list <- integer(0L)
  for (mm in seq_len(n_mmblocks)) {
    fit_mm_block <- step1$MM.FIT[[mm]]

    # influence d(theta.mm)/d(block stats). When the joint model is categorical
    # we use the (D)WLS sandwich ingredients for EVERY block, including a fully
    # continuous one (block-wise mixed): lavTech(., "h1.information.expected")
    # is not available for a continuous measurement block that is embedded in a
    # categorical SAM, whereas lav_model_nvcov_robust_sem() returns a valid
    # Delta/E.inv/WLS.V for it.
    if (lavmodel@categorical) {
      # influence of the (D)WLS estimator: (Delta' W Delta)^-1 Delta' W
      # (we reuse the same ingredients that produced the robust vcov of
      # this measurement block)
      tmp <- lav_model_nvcov_robust_sem(
        lavmodel = fit_mm_block@Model,
        lavsamplestats = fit_mm_block@SampleStats,
        lavcache = fit_mm_block@cache, lavdata = fit_mm_block@Data,
        lavimplied = fit_mm_block@implied, lavh1 = fit_mm_block@h1,
        lavoptions = fit_mm_block@Options, use_ginv = FALSE,
        attr_delta = TRUE, attr_e_inv = TRUE, attr_wls_v = TRUE)
      mm_e_inv <- attr(tmp, "E.inv")
      mm_wls_v <- attr(tmp, "WLS.V")[[g]]
      mm_delta_1 <- attr(tmp, "Delta")[[g]]
      if (is.matrix(mm_wls_v)) {
        mm_jac <- mm_e_inv %*% t(mm_wls_v %*% mm_delta_1)
      } else {
        # DWLS/ULS: WLS.V holds the diagonal of the weight matrix
        mm_jac <- mm_e_inv %*% t(mm_wls_v * mm_delta_1)
      }
    } else {
      mm_h1_expected <- lavTech(fit_mm_block, "h1.information.expected")
      mm_delta       <- lavTech(fit_mm_block, "Delta")
      # JACa = d(theta.mm)/d(stats) = invI . Delta' . A1 (A1 = saturated/h1
      # weight). Which information to invert depends on the SE method, and the
      # two choices are NOT interchangeable in principle (they coincide only for
      # a just-identified measurement block, where s = sigma):
      #   - se = "twostep.robust" (p_only): the Yuan & Chan (2002, eq. 12a)
      #     ASYMPTOTIC sandwich, whose bread is the population A-matrix -> the
      #     EXPECTED information (Delta'A1Delta)^-1, when information = "expected"
      #     (the lavaan default; honor the option otherwise).
      #   - se = "local"/"ij": the infinitesimal jackknife = an EMPIRICAL,
      #     finite-sample sandwich. Its influence d(theta.mm_hat)/d(stats) is the
      #     exact realized jacobian (implicit function theorem at the solution),
      #     i.e. the OBSERVED information; this also keeps JACa coherent with the
      #     observed/ADF Gamma it is later sandwiched with in Gamma.eta =
      #     JAC . Gamma . JAC'. So local MUST use observed (do not "unify" this
      #     with the twostep.robust choice).
      if (p_only && fit@Options$information[1] == "expected") {
        mm_inv_observed <-
          lavTech(fit_mm_block, "inverted.information.expected")
      } else {
        mm_inv_observed <-
          lavTech(fit_mm_block, "inverted.information.observed")
      }
      mm_jac <- t(mm_h1_expected[[g]] %*% mm_delta[[g]] %*% mm_inv_observed)
    }
    # full (unfiltered) influence: needed for the psi channel below
    mm_jac_full <- mm_jac

    # keep only rows that are also in FIT@ParTable. The rows of mm_jac are in
    # the *unconstrained* parameter order (Delta and the augmented inverted
    # information keep one column/row per unconstrained parameter), so under
    # within-block (ceq.simple) equality constraints the block's constrained
    # 'free' numbering must be remapped to the unconstrained numbering first
    # (the constrained members then each select their own -- identical --
    # influence row); indexing by the constrained numbering would shift every
    # row after the first constrained pair by one. (Same remap as the
    # Sigma.11 assembly in lav_sam_step1().)
    mm_ptm_free <- fit_mm_block@ParTable$free
    if (fit_mm_block@Model@ceq.simple.only) {
      mm_ptm_free[mm_ptm_free > 0L] <- seq_len(fit_mm_block@Model@nx.unco)
    }
    mm_keep_idx <- mm_ptm_free[step1$block.ptm.idx[[mm]]]
    mm_jac <- mm_jac[mm_keep_idx, , drop = FALSE]

    # map the columns of mm_jac (the sample statistics of this
    # measurement block) into the statistics vector of the joint model
    mm_row_idx <- step1$block.mm.idx[[mm]][step1$block.ptm.idx[[mm]]]
    if (lavmodel@categorical) {
      # label-based mapping (handles any variable ordering and a block-wise
      # mix of categorical and continuous measurement blocks)
      mm_col_idx <- lav_sam_mg_cat_col_idx(fit_mm_block, g, joint_labels,
                                           ncol(mm_jac_full))
      # sign flips (continuous conditional.x block: block intercept vs the
      # joint NEGATIVE intercept |t1 entry)
      mm_sign <- attr(mm_col_idx, "sign")
      if (!is.null(mm_sign)) {
        mm_jac <- t(t(mm_jac) * mm_sign)
      }
      jaca[mm_row_idx, mm_col_idx] <- mm_jac
    } else if (lavmodel@conditional.x) {
      # the (unconditional) block statistics are exact functions of the joint
      # conditional statistics: chain through T (see lav_sam_condx_t_g())
      t_block <- lav_sam_condx_t_g(fit, g,
        ov_block = fit_mm_block@Data@ov.names[[g]],
        meanstructure = fit_mm_block@Model@meanstructure)
      jaca[mm_row_idx, ] <- mm_jac %*% t_block
    } else {
      mm_ov_idx <- match(step1$MM.FIT[[mm]]@Data@ov.names[[g]],
                         lavdata@ov.names[[g]])
      mm_nvar <- length(lavdata@ov.names[[g]])
      mm_col_idx <- lav_mat_vech_which_idx(mm_nvar, idx = mm_ov_idx,
        add_idx_at_start = lavmodel@meanstructure)
      jaca[mm_row_idx, mm_col_idx] <- mm_jac
    }

    # psi channel: capture the influence d(Psi)/d(stats) for the within-block
    # factor (co)variances (categorical blocks only), mapped to the joint stats
    if (lavmodel@categorical && fit_mm_block@Model@categorical) {
      blv <- unlist(fit_mm_block@pta$vnames$lv)
      bpt <- fit_mm_block@ParTable
      bpsi <- which(bpt$op == "~~" & bpt$lhs %in% blv & bpt$rhs %in% blv &
                    bpt$free > 0L & !duplicated(bpt$free))
      for (k in bpsi) {
        ja_row <- numeric(ncol(jaca))
        ja_row[mm_col_idx] <- mm_jac_full[mm_ptm_free[k], ]
        # joint partable row for this factor (co)variance (canonical pair)
        jrow <- which(step1$PT$op == "~~" &
          ((step1$PT$lhs == bpt$lhs[k] & step1$PT$rhs == bpt$rhs[k]) |
           (step1$PT$lhs == bpt$rhs[k] & step1$PT$rhs == bpt$lhs[k])))[1]
        psi_infl_list[[length(psi_infl_list) + 1L]] <- ja_row
        psi_jrow_list <- c(psi_jrow_list, jrow)
      }
    }
  }

  # keep only the free 'measurement' parameters
  pt_1 <- step1$PT
  # only ov.names that are actually used in the measurement models
  ov_names <- unique(unlist(lapply(step1$MM.FIT, lav_object_vnames, "ov")))
  lv_ind <- unlist(lavpta$vnames$lv.ind)
  keep_idx <- lav_sam_meas_keep_idx(pt_1, ov_names, lavmodel, lv_ind,
                                    lavmodel@meanstructure)
  if (p_only) {
    # the rows of P are in ascending free-parameter order; return the
    # free indices, so the caller can align them with step1.free.idx
    # (which is ordered per measurement block). Under (within-block)
    # ceq.simple equality constraints, step1.free.idx is in the
    # *unconstrained* numbering (one entry per constrained member), so
    # de-duplicate on step1$PT.free instead of pt_1$free: the constrained
    # members then each keep their own (identical) influence row, exactly
    # as in the multigroup (_mg) path.
    p_keep_idx <- lav_sam_meas_keep_idx(pt_1, ov_names, lavmodel, lv_ind,
                                        lavmodel@meanstructure,
                                        dedup_key = step1$PT.free)
    jaca <- jaca[p_keep_idx, , drop = FALSE]
    attr(jaca, "free.idx") <- step1$PT.free[p_keep_idx]
    return(jaca)
  }
  jaca <- jaca[keep_idx, , drop = FALSE]

  # JACb: jacobian of the function vech(VETA) = f(vech(S), theta.mm)
  #       (treating theta.mm as fixed)
  if (length(unlist(lavpta$vnames$lv.interaction)) > 0L) {
    ffb <- function(x) {
      if (lavmodel@meanstructure) {
        nvar <- nrow(fit@h1$implied$cov[[g]])
        this_ybar <- x[seq_len(nvar)]
        this_cov <- lav_mat_vech_rev(x[-seq_len(nvar)])
      } else {
        this_ybar <- fit@h1$implied$mean[[g]]
        this_cov <- lav_mat_vech_rev(x)
      }

      # change COV/YBAR
      step1_1 <- step1
      step1_1$COV[[1]] <- this_cov
      if (lavmodel@meanstructure) {
        step1_1$YBAR[[1]] <- this_ybar
      }

      # transform data to comply with the new COV/YBAR
      y <- fit@Data@X[[1]]
      ytrans <- vector("list", nblocks)
      ytrans[[1]] <- lav_mat_transform_mean_cov(y, target_mean = this_ybar,
                                                   target_cov = this_cov)
      colnames(ytrans[[1]]) <- fit@pta$vnames$ov[[1]]

      step1_1 <- lav_sam_step1_local(step1 = step1_1, fit = fit, y = ytrans,
           sam_method = step1$sam.method, local_options = step1$local.options)
      if (lavmodel@meanstructure) {
        out <- c(step1_1$EETA[[1]], lav_mat_vech(step1_1$VETA[[1]]))
      } else {
        out <- lav_mat_vech(step1_1$VETA[[1]])
      }
      out
    }
    # shut off verbose
    verbose_flag <- lav_verbose()
    lav_verbose(FALSE)
    this_x <- lav_mat_vech(fit@h1$implied$cov[[g]])
    if (lavmodel@meanstructure) {
      this_x <- c(fit@h1$implied$mean[[g]], this_x)
    }
    jacb <- numDeriv::jacobian(func = ffb, x = this_x)
    lav_verbose(verbose_flag)
  } else if (lavmodel@categorical) {
    # categorical (and block-wise mixed): JACb depends on the (poly)choric
    # correlations and the free continuous variances of step1$COV (not the
    # thresholds/means), and through the DELTA rescaling / Croon correction is
    # not a simple M %x% M map; obtain it numerically (see lav_sam_jacb_cat_g()).
    jacb <- lav_sam_jacb_cat_g(g, step1, fit, joint_labels,
                               lavmodel@meanstructure)
  } else if (lavmodel@conditional.x) {
    jacb <- lav_sam_jacb_cont_condx_g(step1$M[[g]],
      nexo = length(fit@Data@ov.names.x[[g]]),
      d_std = step1$D.STD[[g]])
  } else { # no latent interactions
    jacb <- lav_sam_jacb_cont_g(step1$M[[g]], lavmodel@meanstructure)
  }

  # JACc: jacobian of vech(VETA) = f(theta.mm), keeping vech(S) fixed
  jacc <- lav_sam_jac_est_rows(keep_idx, step1, fit, g, lavmodel@meanstructure)

  # assemble JAC
  jac <- (jacc %*% jaca) + jacb

  # psi channel (categorical): add (d VETA / d Psi) %*% (d Psi / d stats), the
  # contribution of the within-block factor (co)variances that enter VETA
  # through the delta constraint theta_jj = 1 - (Lambda Psi Lambda')_jj. Without
  # it the categorical latent (co)variance SEs are ~0.6x too small. For
  # continuous indicators d VETA / d Psi = 0, so this leaves the continuous
  # (and mixed-within-block) results that do not depend on Psi unchanged.
  jacpsi <- NULL
  # drop any factor (co)variance already propagated as a measurement parameter
  # in keep_idx (e.g. the lv.ind (co)variances of a higher-order model), to
  # avoid double counting
  if (length(psi_jrow_list) > 0L) {
    dup_idx <- which(psi_jrow_list %in% keep_idx)
    if (length(dup_idx) > 0L) {
      psi_jrow_list <- psi_jrow_list[-dup_idx]
      psi_infl_list <- psi_infl_list[-dup_idx]
    }
  }
  if (lavmodel@categorical && length(psi_jrow_list) > 0L) {
    jaca_psi <- do.call(rbind, psi_infl_list)
    jacpsi <- lav_sam_jac_est_rows(psi_jrow_list, step1, fit, g,
                                   lavmodel@meanstructure)
    jac <- jac + (jacpsi %*% jaca_psi)
  }

  if (return_jac) {
    attr(jac, "JACa") <- jaca
    attr(jac, "JACb") <- jacb
    attr(jac, "JACc") <- jacc
    attr(jac, "JACpsi") <- jacpsi
  }

  # eventually, this will be a list per group/block
  list(jac)
}

# PML structural-space Gamma.eta (se = "local" for estimator = "PML").
#
# For PML the step-1 estimator is NOT a function of the saturated statistics
# s = (thresholds, polychoric correlations): the pairwise likelihood loads on
# the full bivariate tables. The influence route Gamma.eta = JAC Gamma JAC'
# (which writes theta1.hat as a function of s) therefore does not apply.
# Instead we build the CASEWISE joint influences of the two ingredients of
# the structural statistics c(EETA, vech(VETA)) = f(s, theta1.hat):
#
#   sqrt(N) (stats - target) ~ (1/sqrt(N)) sum_i u_i,
#   u_i = JACb zeta_i + JACc h_i
#
# with
#   zeta_i = casewise influence of the saturated statistics: the muthen1984
#            two-stage estimator's influence function, zeta_i = N B^-1 sc_i
#            (SC/A11/A21/A22 from muthen1984()); their crossprod/N
#            reproduces the categorical Gamma (NACOV) exactly;
#   h_i    = casewise influence of the step-1 (PML) block estimates: per
#            block, I_b^-1 g1_i with g1_i the casewise pairwise scores
#            (lav_sc()) and I_b the unit block information;
#   JACb   = d(stats)/d s'       (theta1 fixed; lav_sam_jacb_cat_g());
#   JACc   = d(stats)/d theta1'  (s fixed; lav_sam_jac_est_rows() over ALL
#            free block parameters, which automatically includes the
#            within-block factor (co)variances -- the 'psi channel' of the
#            (D)WLS categorical path: in the delta parameterization VETA
#            depends on Psi through theta_jj = 1 - (Lambda Psi Lambda')_jj).
#
# Gamma.eta = (1/N) sum_i (u_i - ubar)(u_i - ubar)': the joint casewise
# (infinitesimal-jackknife) covariance, including the cross-covariance
# between s and theta1.hat (same cases). Downstream, step 2 consumes it
# exactly like the (D)WLS Gamma.eta: robust (local) SEs + the corrected
# Satorra-Bentler structural test.
#
# Scope (guarded): single group, single level, complete data, all-ordinal,
# no exogenous covariates, no sampling weights, no equality constraints.
lav_sam_gamma_eta_pml <- function(step1 = NULL, fit = NULL) {
  lavdata <- fit@Data
  lavmodel <- fit@Model
  g <- 1L

  # guards
  if (lavdata@ngroups > 1L) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: not available with multiple groups (yet)!"))
  }
  if (lavdata@nlevels > 1L) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: not available for multilevel models (yet)!"))
  }
  if (lavmodel@conditional.x) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: not available with exogenous covariates
       (conditional.x = TRUE) (yet)!"))
  }
  if (anyNA(lavdata@X[[g]])) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: not available for missing data (yet)!"))
  }
  if (length(lavdata@sampling.weights) > 0L) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: not available with sampling weights (yet)!"))
  }
  if (lavmodel@eq.constraints || lavmodel@ceq.simple.only) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: not available with equality constraints (yet)!"))
  }
  ov_idx <- match(lavdata@ov.names[[g]], lavdata@ov$name)
  ov_types <- lavdata@ov$type[ov_idx]
  if (!all(ov_types == "ordered")) {
    lav_msg_stop(gettext(
      "PML Gamma.eta: all observed variables must be ordered (yet)!"))
  }

  y <- lavdata@X[[g]]
  n <- nrow(y)
  meanstructure <- lavmodel@meanstructure

  # --- 1. zeta_i: casewise influence of the saturated (th, rho) statistics
  cat_1 <- muthen1984(
    data_1 = y,
    ov_names = lavdata@ov.names[[g]],
    ov_types = ov_types,
    ov_levels = lavdata@ov$nlev[ov_idx],
    wls_w = TRUE
  )
  b_inv <- lav_m84_b_inv(
    a11 = cat_1$A11, a21 = cat_1$A21, a22 = cat_1$A22
  )$b_inv
  zeta <- n * (cat_1$SC %*% t(b_inv)) # N x nstats, WLS.obs layout

  # --- 2. JACb: d(stats)/d s' (theta1 fixed)
  joint_labels <- lav_sam_wls_obs_labels(
    ov_names = lavdata@ov.names[[g]],
    th_idx = fit@SampleStats@th.idx[[g]],
    ov_names_x = NULL
  )
  jacb <- lav_sam_jacb_cat_g(g, step1, fit, joint_labels, meanstructure)

  # --- 3. rows: ALL free block parameters, mapped to step1$PT rows
  pt_1 <- step1$PT
  n_mmblocks <- length(step1$MM.FIT)
  rows_list <- cols_list <- vector("list", n_mmblocks)
  for (mm in seq_len(n_mmblocks)) {
    fb <- step1$MM.FIT[[mm]]
    ptm <- fb@ParTable
    ptm_idx <- step1$block.ptm.idx[[mm]]
    ok <- ptm$free[ptm_idx] > 0L
    rows_list[[mm]] <- step1$block.mm.idx[[mm]][ptm_idx][ok]
    cols_list[[mm]] <- ptm$free[ptm_idx][ok]
  }
  rows_idx <- sort(unique(unlist(rows_list)))

  # JACc: d(stats)/d theta1' (s fixed), columns in rows_idx order
  jacc <- lav_sam_jac_est_rows(rows_idx, step1, fit, g, meanstructure)

  # --- 4. h_i: casewise influence of the block (PML) estimates,
  #        columns in rows_idx order
  h <- matrix(0, n, length(rows_idx))
  for (mm in seq_len(n_mmblocks)) {
    fb <- step1$MM.FIT[[mm]]
    # ignore_constraints: the blocks are fitted with bounds =
    # "wide.zerovar", which lavaan represents as (inactive) linear
    # INEQUALITY constraints; lav_sc()'s constraint projection would
    # project the scores onto them and corrupt the casewise influences.
    # Equality constraints are excluded from this route (guard above).
    scb <- lav_sc(fb, remove_empty_cases = FALSE,
                  ignore_constraints = TRUE)
    scb[is.na(scb)] <- 0
    if (nrow(scb) != n) {
      lav_msg_stop(gettext(
        "internal error: case alignment failure between a measurement
         block and the data (PML Gamma.eta)."))
    }
    ib <- lavTech(fb, "information") # unit information
    nb <- fb@SampleStats@ntotal
    hb <- (n / nb) * scb %*% solve(ib)
    h[, match(rows_list[[mm]], rows_idx)] <-
      hb[, cols_list[[mm]], drop = FALSE]
  }

  # --- 5. assemble the joint casewise influences + Gamma.eta
  u <- zeta %*% t(jacb) + h %*% t(jacc)
  u <- t(t(u) - colMeans(u))
  list(crossprod(u) / n)
}

# Two-level (single group) Gamma.eta: the NACOV of the structural statistics
# c(EETA_b, vech(VETA_b)) per level, plus their full (cross-level) covariance.
#
# The 'sample' statistics of the two-level local approach are the h1
# (saturated) estimates s = c(Mu.W, vech(Sigma.W), Mu.B, vech(Sigma.B)), whose
# sampling variability is the cluster-sandwich Gamma of the two-level (D)WLS
# machinery (see lav_samplestats_wls_2l.R):
#   Gamma.s = nobs * (I1^-1 J1 I1^-1) / nclusters
# with I1/J1 the expected/first-order per-cluster h1 informations at the h1
# estimates. The jacobian of the structural statistics w.r.t. s uses the same
# decomposition as the single-level lav_sam_step1_local_jac():
#   JAC_b = (JACc_b %*% JACa) + JACb_b
# where JACa (the influence of s on the step-1 measurement parameters) comes
# from the two-level measurement-block fits -- their Delta rows and h1
# information are laid out exactly like s, so the single-level influence
# formula invI . Delta' . A1 carries over, with the block statistics mapped
# into the joint s per level; JACb_b is the direct map of level b's segment
# of s (VETA_b = M_b S_b M_b'); and JACc_b is the numeric jacobian through
# the measurement estimates, reusing lav_sam_jac_est_rows() (which re-runs
# the two-level lav_sam_step1_local()).
#
# The structural model is refit as a two-'group' model (one group per level,
# nobs = c(N, nclusters)), so the per-level Gamma.eta list must follow
# lavaan's per-group NACOV convention Gamma_g = Cov(sqrt(n_g) stats_g):
#   Gamma.eta[[1]] =         JAC_1 Gamma.s JAC_1'    (n_1 = N; Gamma.s is
#                                                     N-scaled already)
#   Gamma.eta[[2]] = (G/N) * JAC_2 Gamma.s JAC_2'    (n_2 = G = nclusters)
# The two levels' statistics are correlated (they share s and the step 1
# chain), which the per-group list cannot express; we therefore also return
# the plain covariance of the STACKED statistics (Gamma.eta.full), and step 2
# computes the exact cross-level sandwich via the Case B path.
lav_sam_gamma_eta_2l <- function(step1 = NULL, fit = NULL) {
  lavdata <- fit@Data
  lavmodel <- fit@Model
  g <- 1L

  # supported setting: two levels, single group, continuous, complete data,
  # h1 statistics
  if (lavdata@ngroups > 1L) {
    lav_msg_stop(gettext(
      "two-level Gamma.eta: not available for multigroup multilevel
       models (yet)!"))
  }
  if (lavmodel@categorical) {
    lav_msg_stop(gettext(
      "two-level Gamma.eta: not available for categorical data (yet)!"))
  }
  if (length(unlist(fit@pta$vnames$lv.interaction)) > 0L) {
    lav_msg_stop(gettext(
      "two-level Gamma.eta: not available with latent interactions (yet)!"))
  }
  if (!is.null(step1$local.options$twolevel.method) &&
      step1$local.options$twolevel.method != "h1") {
    lav_msg_stop(gettext(
      "two-level Gamma.eta: only available for twolevel.method = \"h1\"."))
  }
  if (anyNA(lavdata@X[[g]])) {
    lav_msg_stop(gettext(
      "two-level Gamma.eta: not available for missing data (yet)!"))
  }

  lp <- lavdata@Lp[[g]]
  meanstructure <- lavmodel@meanstructure # always TRUE for two-level

  # level variable sets, in the order used by the h1 estimates and the
  # Delta rows (the Lp$ov.idx order; note ov.idx[[2]] may be unsorted)
  ov_l <- lapply(1:2, function(l) lavdata@ov.names[[g]][lp$ov.idx[[l]]])
  p_l <- lengths(ov_l)
  seg_len <- p_l + p_l * (p_l + 1L) / 2L
  seg_offset <- c(0L, seg_len[1]) # start of each level's (mean, vech) segment
  n_s <- sum(seg_len)

  # h1 (saturated) estimates: step1$COV/YBAR hold them per level (this is
  # the twolevel.method = "h1" path of lav_sam_get_cov_ybar())
  mu_w <- drop(step1$YBAR[[1]])
  sigma_w <- step1$COV[[1]]
  mu_b <- drop(step1$YBAR[[2]])
  sigma_b <- step1$COV[[2]]
  if (ncol(sigma_w) != p_l[1] || ncol(sigma_b) != p_l[2]) {
    lav_msg_stop(gettext(
      "internal error: two-level Gamma.eta: step1$COV dimensions do not
       match the level variable sets."))
  }

  # Gamma.s: cluster-sandwich NACOV of the h1 estimates (the same
  # computation as lav_samp_wls_2l(), N-scaled convention)
  nobs <- fit@SampleStats@nobs[[g]]
  nclusters <- lp$nclusters[[2]]
  i1 <- lav_mvn_cl_info_expected(
    lp = lp, mu_w = mu_w, sigma_w = sigma_w, mu_b = mu_b, sigma_b = sigma_b,
    x_idx = fit@SampleStats@x.idx[[g]]
  )
  j1 <- lav_mvn_cl_info_firstorder(
    y1 = lavdata@X[[g]], ylp = fit@SampleStats@YLp[[g]], lp = lp,
    mu_w = mu_w, sigma_w = sigma_w, mu_b = mu_b, sigma_b = sigma_b,
    x_idx = fit@SampleStats@x.idx[[g]], divide_by_two = TRUE
  )
  diag_i1 <- diag(i1)
  keep <- which(abs(diag_i1) > sqrt(.Machine$double.eps) * max(abs(diag_i1)))
  i1_inv <- matrix(0, nrow(i1), ncol(i1))
  inv_keep <- try(lav_mat_sym_inverse(i1[keep, keep, drop = FALSE]),
                  silent = TRUE)
  if (inherits(inv_keep, "try-error")) {
    lav_msg_stop(gettext(
      "two-level Gamma.eta: could not invert the h1 information matrix."))
  }
  i1_inv[keep, keep] <- inv_keep
  gamma_s <- nobs * (i1_inv %*% j1 %*% i1_inv) / nclusters
  gamma_s <- (gamma_s + t(gamma_s)) / 2

  # JACa: influence d(theta.mm)/d(s) of the two-level measurement blocks
  jaca <- matrix(0, nrow = length(fit@ParTable$lhs), ncol = n_s)
  for (mm in seq_len(length(step1$MM.FIT))) {
    fit_mm_block <- step1$MM.FIT[[mm]]
    if (fit_mm_block@Data@nlevels != 2L) {
      lav_msg_stop(gettext(
        "two-level Gamma.eta: all measurement blocks must be two-level
         models."))
    }
    # invI . Delta' . A1, exactly as in the single-level case (se = "local"
    # uses the observed information; see lav_sam_step1_local_jac())
    mm_h1_expected <- lavTech(fit_mm_block, "h1.information.expected")
    mm_delta <- lavTech(fit_mm_block, "Delta")
    mm_inv_observed <- lavTech(fit_mm_block, "inverted.information.observed")
    mm_jac <- t(mm_h1_expected[[g]] %*% mm_delta[[g]] %*% mm_inv_observed)

    # keep only rows that are also in FIT@ParTable
    mm_keep_idx <- fit_mm_block@ParTable$free[step1$block.ptm.idx[[mm]]]
    mm_jac <- mm_jac[mm_keep_idx, , drop = FALSE]

    # map the block's statistics (mu_w, vech(Sigma.W), mu_b, vech(Sigma.B)
    # over the block variables) into the joint s, per level
    mm_lp <- fit_mm_block@Data@Lp[[g]]
    mm_col_idx <- unlist(lapply(1:2, function(l) {
      mm_ov_l <- fit_mm_block@Data@ov.names[[g]][mm_lp$ov.idx[[l]]]
      idx_l <- match(mm_ov_l, ov_l[[l]])
      if (anyNA(idx_l)) {
        lav_msg_stop(gettext(
          "internal error: two-level Gamma.eta: unable to map the
           measurement-block variables into the joint statistics vector."))
      }
      seg_offset[l] + lav_mat_vech_which_idx(p_l[l], idx = idx_l,
                                             add_idx_at_start = TRUE)
    }))
    mm_row_idx <- step1$block.mm.idx[[mm]][step1$block.ptm.idx[[mm]]]
    jaca[mm_row_idx, mm_col_idx] <- mm_jac
  }

  # keep only the free 'measurement' parameters
  pt_1 <- step1$PT
  ov_names <- unique(unlist(lapply(step1$MM.FIT, lav_object_vnames, "ov")))
  lv_ind <- unlist(fit@pta$vnames$lv.ind)
  keep_idx <- lav_sam_meas_keep_idx(pt_1, ov_names, lavmodel, lv_ind,
                                    meanstructure)
  jaca <- jaca[keep_idx, , drop = FALSE]

  # per level: JAC_b = (JACc_b %*% JACa) + JACb_b
  jac <- vector("list", 2L)
  for (b in 1:2) {
    # direct part: (EETA_b, vech(VETA_b)) = f(mean_b, vech(S_b)), theta fixed
    jacb_core <- lav_sam_jacb_cont_g(step1$M[[b]], meanstructure)
    jacb <- matrix(0, nrow = nrow(jacb_core), ncol = n_s)
    jacb[, seg_offset[b] + seq_len(seg_len[b])] <- jacb_core
    # indirect part: through the measurement estimates
    jacc <- lav_sam_jac_est_rows(keep_idx, step1, fit, b, meanstructure)
    jac[[b]] <- (jacc %*% jaca) + jacb
  }

  # per-level NACOV (Gamma_g = Cov(sqrt(n_g) stats_g), n = c(N, nclusters))
  # + the full covariance of the stacked statistics (cross-level blocks)
  gamma_eta <- vector("list", 2L)
  gamma_eta[[1]] <- jac[[1]] %*% gamma_s %*% t(jac[[1]])
  gamma_eta[[2]] <- (nclusters / nobs) * (jac[[2]] %*% gamma_s %*% t(jac[[2]]))
  jac_full <- rbind(jac[[1]], jac[[2]])
  gamma_eta_full <- jac_full %*% (gamma_s / nobs) %*% t(jac_full)

  list(Gamma.eta = gamma_eta, Gamma.eta.full = gamma_eta_full, JAC = jac)
}

# multigroup version of lav_sam_step1_local_jac() (single level, no latent
# interactions, continuous data).
#
# Strategy: build ONE 'stacked' jacobian
#   JAC.full = JACc %*% JACa + JACb
# where
#   - the columns run over the stacked sample statistics vech(S_g) (per group),
#   - the rows run over the stacked c(EETA_g, vech(VETA_g)) (per group).
# The sample statistics are independent across groups, so the (full) Gamma of
# the stacked vech(S) is block-diagonal. When the measurement model has NO
# across-group (equality) constraints, the measurement parameters of the
# different groups are estimated independently, hence JAC.full is itself
# block-diagonal across groups, and we can return a simple list of per-group
# jacobians jac[[g]] (the caller then forms Gamma.eta[[g]] = jac[[g]] %*%
# Gamma[[g]] %*% t(jac[[g]]), and feeds it as a per-group NACOV to the
# multigroup structural model).
#
# If the measurement model DOES contain across-group constraints, JAC.full has
# nonzero off-diagonal (cross-group) blocks: vech(VETA_g) then depends on the
# sample statistics of the OTHER groups too. A per-group NACOV list can no
# longer capture this (the structural step would need the full cross-group
# Gamma.eta); we detect this situation and stop with an informative message.
# map the statistics of a (categorical or continuous) measurement block of
# group g into the labels of the joint WLS statistics vector of group g, and
# return the matching column indices (within group g's statistics).
lav_sam_mg_cat_col_idx <- function(fit_mm_block, g, joint_labels, ncol_full) {
  if (fit_mm_block@Model@categorical) {
    mm_labels <- lav_sam_wls_obs_labels(
      ov_names = if (fit_mm_block@Model@conditional.x) {
        fit_mm_block@pta$vnames$ov.nox[[g]]
      } else {
        fit_mm_block@Data@ov.names[[g]]
      },
      th_idx = fit_mm_block@SampleStats@th.idx[[g]],
      ov_names_x = if (fit_mm_block@Model@conditional.x) {
        fit_mm_block@Data@ov.names.x[[g]]
      } else {
        NULL
      })
  } else if (fit_mm_block@Model@conditional.x) {
    # continuous block, fitted with conditional.x = TRUE, within a
    # categorical + conditional.x joint model: the influence columns follow
    # the continuous conditional layout c(vecr(cbind(int, slopes)),
    # vech(res.cov, diag = TRUE)). The joint represents each numeric
    # variable by its NEGATIVE intercept (|t1), its slopes (~x), its
    # residual variance (|var) and residual covariances (~~); the |t1 sign
    # flip is returned in the "sign" attribute.
    block_ov <- fit_mm_block@Data@ov.names[[g]]
    block_x <- fit_mm_block@Data@ov.names.x[[g]]
    nvar_b <- length(block_ov)
    q_b <- length(block_x)
    mean_labels <- unlist(lapply(block_ov, function(y) {
      c(paste0(y, "|t1"), paste0(y, "~", block_x))
    }))
    mean_signs <- rep(c(-1, rep(1, q_b)), times = nvar_b)
    v_lin <- lav_mat_vech_idx(nvar_b)
    zero_b <- matrix(0L, nvar_b, nvar_b)
    v_r <- row(zero_b)[v_lin]
    v_c <- col(zero_b)[v_lin]
    cov_labels <- ifelse(v_r == v_c,
      paste0(block_ov[v_r], "|var"),
      vapply(seq_along(v_r), function(k)
        paste(sort(c(block_ov[v_r[k]], block_ov[v_c[k]])), collapse = "~~"),
        character(1L)))
    mm_labels <- c(mean_labels, cov_labels)
    mm_signs <- c(mean_signs, rep(1, length(cov_labels)))
    col_idx <- match(mm_labels, joint_labels)
    if (anyNA(col_idx)) {
      lav_msg_stop(gettext(
        "internal error: unable to map the statistics of a measurement block
         into the statistics vector of the joint model"))
    }
    attr(col_idx, "sign") <- mm_signs
    return(col_idx)
  } else {
    # continuous block within a categorical joint model: the influence columns
    # are ordered (means, vech(cov)); the joint represents each continuous
    # variable by its mean (|t1), variance (|var) and covariances (~~)
    block_ov <- fit_mm_block@Data@ov.names[[g]]
    nvar_b <- length(block_ov)
    v_lin <- lav_mat_vech_idx(nvar_b)
    zero_b <- matrix(0L, nvar_b, nvar_b)
    v_r <- row(zero_b)[v_lin]
    v_c <- col(zero_b)[v_lin]
    cov_labels <- ifelse(v_r == v_c,
      paste0(block_ov[v_r], "|var"),
      vapply(seq_along(v_r), function(k)
        paste(sort(c(block_ov[v_r[k]], block_ov[v_c[k]])), collapse = "~~"),
        character(1L)))
    if (ncol_full == nvar_b + length(v_lin)) {
      mm_labels <- c(paste0(block_ov, "|t1"), cov_labels) # with means
    } else {
      mm_labels <- cov_labels
    }
  }
  col_idx <- match(mm_labels, joint_labels)
  if (anyNA(col_idx)) {
    lav_msg_stop(gettext(
      "internal error: unable to map the statistics of a measurement block
       into the statistics vector of the joint model"))
  }
  col_idx
}

lav_sam_step1_local_jac_mg <- function(step1 = NULL, fit = NULL,
                                       p_only = FALSE,
                                       return_jac = FALSE) {
  lavdata  <- fit@Data
  lavmodel <- fit@Model
  lavpta   <- fit@pta
  ngroups  <- lavdata@ngroups
  n_mmblocks <- length(step1$MM.FIT)
  meanstructure <- lavmodel@meanstructure

  # per-group size of the (joint) sample-statistics vector vech(S_g)[+ means]
  stat_n <- vapply(seq_len(ngroups),
    function(g) length(fit@SampleStats@WLS.obs[[g]]), integer(1L))
  stat_offset <- c(0L, cumsum(stat_n)) # length ngroups + 1
  stat_tot <- sum(stat_n)

  # NOTE: veta_n / veta_offset (the per-group size of c(EETA_g, vech(VETA_g)))
  # are only needed for the JACb/JACc stacking below; they rely on step1$VETA,
  # which is only computed for the local/fsr/cfsr methods. They are therefore
  # deferred until after the p_only early return (twostep.robust, also used by
  # sam.method = "global", does not need them and does not have step1$VETA).

  # categorical: per-group labels of the joint WLS statistics vector, used to
  # map each measurement block's statistics into the right joint columns
  if (lavmodel@categorical) {
    joint_labels <- lapply(seq_len(ngroups), function(g)
      lav_sam_wls_obs_labels(ov_names = lavdata@ov.names[[g]],
                             th_idx = fit@SampleStats@th.idx[[g]]))
  }

  # JACa: jacobian of theta.mm = f(vech(S_all))
  # rows: all parameters in fit@ParTable (we select 'measurement' rows later)
  # cols: stacked vech(S_g) over the groups
  jaca <- matrix(0, nrow = length(fit@ParTable$lhs), ncol = stat_tot)
  # psi-channel accumulators (categorical only): the influence d(Psi)/d(stats)
  # of the within-block factor (co)variances, and their joint partable rows
  psi_infl_list <- list()
  psi_jrow_list <- integer(0L)
  for (mm in seq_len(n_mmblocks)) {
    fit_mm_block   <- step1$MM.FIT[[mm]]
    if (lavmodel@categorical) {
      # influence of the (D)WLS estimator: (Delta' W Delta)^-1 Delta' W, reusing
      # the robust-vcov ingredients. We use this for EVERY block (incl. a fully
      # continuous one in a block-wise mixed model), because
      # lavTech(., "h1.information.expected") is not available for a continuous
      # block embedded in a categorical SAM.
      tmp <- lav_model_nvcov_robust_sem(
        lavmodel = fit_mm_block@Model,
        lavsamplestats = fit_mm_block@SampleStats,
        lavcache = fit_mm_block@cache, lavdata = fit_mm_block@Data,
        lavimplied = fit_mm_block@implied, lavh1 = fit_mm_block@h1,
        lavoptions = fit_mm_block@Options, use_ginv = FALSE,
        attr_delta = TRUE, attr_e_inv = TRUE, attr_wls_v = TRUE)
      mm_e_inv  <- attr(tmp, "E.inv")
      mm_delta  <- attr(tmp, "Delta")
      mm_wls_v  <- attr(tmp, "WLS.V")
    } else {
      mm_h1_expected <- lavTech(fit_mm_block, "h1.information.expected")
      mm_delta       <- lavTech(fit_mm_block, "Delta")
      # same expected-vs-observed information choice as the single-group path
      # (see lav_sam_step1_local_jac()): EXPECTED for se = "twostep.robust"
      # (p_only; Yuan & Chan asymptotic sandwich, honoring the information
      # option), OBSERVED for se = "local"/"ij" (infinitesimal jackknife =
      # empirical sandwich, coherent with the observed/ADF Gamma).
      if (p_only && fit@Options$information[1] == "expected") {
        mm_inv_observed <-
          lavTech(fit_mm_block, "inverted.information.expected")
      } else {
        mm_inv_observed <-
          lavTech(fit_mm_block, "inverted.information.observed")
      }
    }

    # rows of mm_jac that we keep (the free params present in fit@ParTable),
    # together with the corresponding row numbers in fit@ParTable.
    # NOTE: Delta and (inverted) information have one row/col per *unconstrained*
    # free parameter (nx.unco); with simple equality constraints (ceq.simple,
    # e.g. across-group equal loadings) the ParTable$free numbering (nx.free)
    # is smaller and would mis-index the rows of mm_jac. Remap to the
    # unconstrained position (cf. lav_sam_step1()).
    mm_free <- fit_mm_block@ParTable$free
    if (fit_mm_block@Model@ceq.simple.only) {
      mm_free[mm_free > 0L] <- seq_len(fit_mm_block@Model@nx.unco)
    }
    mm_keep_idx <- mm_free[step1$block.ptm.idx[[mm]]]
    mm_row_idx  <- step1$block.mm.idx[[mm]][step1$block.ptm.idx[[mm]]]

    # psi channel: identify this block's free factor (co)variances (ALL groups)
    # ONCE, with their (unconstrained) block-row index and their joint partable
    # row. We accumulate d(Psi)/d(stats) over the groups below, so that with
    # across-group measurement constraints (Case B) a group's factor variance
    # also picks up its dependence on the OTHER groups' statistics.
    psi_fp <- integer(0L); psi_jr <- integer(0L)
    if (lavmodel@categorical && fit_mm_block@Model@categorical) {
      blv <- unlist(fit_mm_block@pta$vnames$lv)
      bpt <- fit_mm_block@ParTable
      bpsi <- which(bpt$op == "~~" & bpt$lhs %in% blv & bpt$rhs %in% blv &
                    bpt$free > 0L & !duplicated(bpt$free))
      for (k in bpsi) {
        psi_fp <- c(psi_fp, mm_free[k]) # unconstrained block-row index
        jrow <- which(step1$PT$op == "~~" & step1$PT$group == bpt$group[k] &
          ((step1$PT$lhs == bpt$lhs[k] & step1$PT$rhs == bpt$rhs[k]) |
           (step1$PT$lhs == bpt$rhs[k] & step1$PT$rhs == bpt$lhs[k])))[1]
        psi_jr <- c(psi_jr, jrow)
      }
    }
    psi_acc <- matrix(0, nrow = length(psi_fp), ncol = stat_tot)

    # lavaan stores a single 'unit' (per-observation averaged) information
    # matrix for the whole (multigroup) block; the group-g diagonal block of
    # its inverse is therefore (N/n_g) times the inverse one would obtain from
    # group g alone, while Delta[[g]] and h1.information.expected[[g]] are
    # unweighted. To recover the per-group influence d(theta.mm_g)/d(s_g) we
    # undo this normalization with the group proportion w_g = n_g / N.
    mm_ntotal <- fit_mm_block@SampleStats@ntotal

    for (g in seq_len(ngroups)) {
      w_g <- fit_mm_block@SampleStats@nobs[[g]] / mm_ntotal
      # influence of group g's sample statistics on ALL block parameters
      # (in the no-across-group-constraint case, only group g's parameters
      #  have nonzero rows; with across-group constraints, other groups'
      #  parameter rows are nonzero too -> cross-group coupling)
      if (lavmodel@categorical) {
        wv <- mm_wls_v[[g]]
        if (is.matrix(wv)) {
          mm_jac <- w_g * (mm_e_inv %*% t(wv %*% mm_delta[[g]]))
        } else {
          # DWLS/ULS: WLS.V holds the diagonal of the weight matrix
          mm_jac <- w_g * (mm_e_inv %*% t(wv * mm_delta[[g]]))
        }
      } else {
        mm_jac <- w_g *
          t(mm_h1_expected[[g]] %*% mm_delta[[g]] %*% mm_inv_observed)
      }
      # full (unfiltered) influence: needed for the psi channel below
      mm_jac_full <- mm_jac
      mm_jac <- mm_jac[mm_keep_idx, , drop = FALSE]

      # map the columns (block statistics of group g) into the joint
      # statistics vector of group g
      if (lavmodel@categorical) {
        mm_col_idx <- stat_offset[g] +
          lav_sam_mg_cat_col_idx(fit_mm_block, g, joint_labels[[g]],
                                 ncol(mm_jac_full))
      } else {
        mm_ov_idx <- match(fit_mm_block@Data@ov.names[[g]],
                           lavdata@ov.names[[g]])
        mm_nvar   <- length(lavdata@ov.names[[g]])
        mm_col_idx <- stat_offset[g] +
          lav_mat_vech_which_idx(mm_nvar, idx = mm_ov_idx,
            add_idx_at_start = meanstructure)
      }
      jaca[mm_row_idx, mm_col_idx] <- mm_jac

      # psi channel: accumulate d(Psi)/d(group-g stats) for this block's factor
      # (co)variances at group g's columns. Summed over the group loop, this
      # gives the full (cross-group) influence; for Case A only the own-group
      # columns are nonzero.
      if (length(psi_fp) > 0L) {
        psi_acc[, mm_col_idx] <- psi_acc[, mm_col_idx, drop = FALSE] +
          mm_jac_full[psi_fp, , drop = FALSE]
      }
    }

    # store the accumulated psi-channel influence rows for this block
    for (i in seq_along(psi_fp)) {
      psi_infl_list[[length(psi_infl_list) + 1L]] <- psi_acc[i, ]
      psi_jrow_list <- c(psi_jrow_list, psi_jr[i])
    }
  }

  # keep only the free 'measurement' parameters
  pt_1 <- step1$PT
  ov_names <- unique(unlist(lapply(step1$MM.FIT, lav_object_vnames, "ov")))
  lv_ind <- unlist(lavpta$vnames$lv.ind)
  keep_idx <- lav_sam_meas_keep_idx(pt_1, ov_names, lavmodel, lv_ind,
                                    meanstructure)

  if (p_only) {
    # twostep.robust: we only need the 'P' matrix = influence of the
    # measurement (step 1) parameters on the (stacked) sample statistics.
    # The rows are in ascending free-parameter order; the caller aligns them
    # with step1.free.idx and splits the columns per group using stat.offset.
    # (No need for the costly JACb/JACc numeric jacobians below; this branch
    # is therefore also usable for sam.method = "global", which does not
    # populate step1$VETA / step1$M / step1$COV.)
    #
    # keep_idx de-duplicates on pt_1$free (the *constrained* free number). With
    # across-group equality constraints in the measurement model (Case B)
    # several partable rows -- one per group -- share that number but are
    # distinct in step1.free.idx (the *unconstrained* numbering). For the P
    # matrix we need one row per unconstrained step 1 parameter, so de-duplicate
    # on step1$PT.free instead. (When there are no simple equality constraints
    # the two numberings coincide, so p_keep_idx == keep_idx.)
    p_keep_idx <- if (lavmodel@ceq.simple.only) {
      lav_sam_meas_keep_idx(pt_1, ov_names, lavmodel, lv_ind, meanstructure,
                            dedup_key = step1$PT.free)
    } else {
      keep_idx
    }
    pjac <- jaca[p_keep_idx, , drop = FALSE]
    attr(pjac, "free.idx") <- step1$PT.free[p_keep_idx]
    attr(pjac, "stat.offset") <- stat_offset
    return(pjac)
  }

  jaca <- jaca[keep_idx, , drop = FALSE]

  # per-group size of c(EETA_g, vech(VETA_g)) (needed for the JACb/JACc
  # stacking below; relies on step1$VETA, only computed for local/fsr/cfsr)
  veta_n <- vapply(seq_len(ngroups), function(g) {
    p <- ncol(step1$VETA[[g]])
    # NOTE: %/% binds tighter than *, so parenthesize the vech length
    # (p*(p+1)/2); without the parens p*(p+1L)%/%2L is wrong for even p
    (if (meanstructure) p else 0L) + (p * (p + 1L)) %/% 2L
  }, integer(1L))
  veta_offset <- c(0L, cumsum(veta_n))
  veta_tot <- sum(veta_n)

  # JACb: jacobian of vech(VETA) = f(vech(S)), keeping theta.mm fixed
  # (block-diagonal across groups: VETA_g = M_g S_g M_g')
  jacb_list <- vector("list", ngroups)
  if (lavmodel@categorical) {
    for (g in seq_len(ngroups)) {
      jacb_list[[g]] <- lav_sam_jacb_cat_g(g, step1, fit, joint_labels[[g]],
                                           meanstructure)
    }
  } else {
    for (g in seq_len(ngroups)) {
      jacb_list[[g]] <- lav_sam_jacb_cont_g(step1$M[[g]], meanstructure)
    }
  }
  jacb <- lav_mat_bdiag(jacb_list)

  # JACc: jacobian of vech(VETA) = f(theta.mm), keeping vech(S) fixed
  # (over the stacked c(EETA_g, vech(VETA_g)) of all groups)
  jacc <- lav_sam_jac_est_rows(keep_idx, step1, fit, seq_len(ngroups),
                               meanstructure)

  # assemble the stacked JAC
  jac_full <- (jacc %*% jaca) + jacb

  # psi channel (categorical): add (d VETA / d Psi) %*% (d Psi / d stats), the
  # contribution of the within-block factor (co)variances that enter VETA
  # through the delta constraint theta_jj = 1 - (Lambda Psi Lambda')_jj. Without
  # it the categorical latent (co)variance SEs are ~0.6x too small. (For
  # continuous indicators d VETA / d Psi = 0.)
  if (length(psi_jrow_list) > 0L) {
    # drop any factor (co)variance already propagated as a measurement
    # parameter in keep_idx (e.g. lv.ind (co)variances), to avoid double counting
    dup_idx <- which(psi_jrow_list %in% keep_idx)
    if (length(dup_idx) > 0L) {
      psi_jrow_list <- psi_jrow_list[-dup_idx]
      psi_infl_list <- psi_infl_list[-dup_idx]
    }
  }
  if (lavmodel@categorical && length(psi_jrow_list) > 0L) {
    jaca_psi <- do.call(rbind, psi_infl_list)
    jacpsi <- lav_sam_jac_est_rows(psi_jrow_list, step1, fit,
                                   seq_len(ngroups), meanstructure)
    jac_full <- jac_full + (jacpsi %*% jaca_psi)
  }

  # split into per-group jacobians, and detect cross-group (off-diagonal)
  # coupling. Negligible off-diagonal blocks => the measurement model has no
  # across-group constraints (Case A): the per-group jac[[g]] are sufficient,
  # and the caller forms a per-group NACOV list. Non-negligible off-diagonal
  # blocks => across-group measurement constraints (Case B): vech(VETA_g) then
  # depends on the OTHER groups' sample statistics, so the full (stacked)
  # jacobian must be kept and a cross-group sandwich computed in step 2.
  jac <- vector("list", ngroups)
  caseb_flag <- FALSE
  for (g in seq_len(ngroups)) {
    r_idx <- (veta_offset[g] + 1L):veta_offset[g + 1L]
    c_idx <- (stat_offset[g] + 1L):stat_offset[g + 1L]
    jac[[g]] <- jac_full[r_idx, c_idx, drop = FALSE]

    other_c <- if (stat_tot > length(c_idx)) {
      seq_len(stat_tot)[-c_idx]
    } else {
      integer(0L)
    }
    if (length(other_c) > 0L) {
      off_block <- max(abs(jac_full[r_idx, other_c, drop = FALSE]))
      this_block <- max(abs(jac[[g]]), 1)
      # numeric (finite-difference) noise is ~1e-7; genuine across-group
      # coupling is O(1), so a relative tolerance of 1e-4 separates them
      if (off_block > 1e-4 * this_block) {
        caseb_flag <- TRUE
      }
    }
  }

  # always expose the stacked jacobian + block layout, so step 2 can build the
  # cross-group sandwich when needed (Case B)
  attr(jac, "jac.full") <- jac_full
  attr(jac, "caseB") <- caseb_flag
  attr(jac, "stat.offset") <- stat_offset
  attr(jac, "veta.offset") <- veta_offset

  if (return_jac) {
    attr(jac, "JACa") <- jaca
    attr(jac, "JACb") <- jacb
    attr(jac, "JACc") <- jacc
  }

  jac
}

# semi-analytic version
# YR 4 June 2025: works, but still needs cleanup + avoid redundant calculations
# (eg when multiplied with a matrix with many zero rows/cols)
# YR 12 June 2026: rewritten to avoid the loop over the observations:
# because differentiation is linear, the average (over the observations) of
# the casewise jacobians equals the jacobian of the *average* casewise
# contribution, and the latter can be computed with a few (small) matrix
# operations; the centering terms (involving the mean of the second-order
# factor scores) can be ignored, as they cancel out when summing over the
# observations
lav_sam_gamma_add <- function(step1 = NULL, fit = NULL, group = 1L,
                              use_analytic = TRUE) {

  lavdata <- fit@Data
  lavmodel <- fit@Model
  lavpta <- fit@pta

  ngroups <- lavdata@ngroups
  if (ngroups > 1L) {
    lav_msg_stop(gettext("IJ local SEs: not available with multiple groups!\n"))
  }
  g <- group
  y <- fit@Data@X[[g]]
  n <- nrow(y)

  # NAMES + lv.keep
  lv_names <- c("..int..", step1$LV.NAMES[[1]])
  tmp <- lav_sam_lvnames2(lv_names)
  lv_keep <- colnames(step1$VETA[[1]])
  keep_idx <- match(lv_keep, tmp$names)

  # row/column pairs of the kept elements of (eta* %x% eta*)
  i1k <- tmp$idx1[keep_idx]
  i2k <- tmp$idx2[keep_idx]

  # the second-order lambda.star, and the effective first-order lambda1,
  # are treated as fixed constants (not differentiated)
  lambda_star <- step1$lambda[[1]]
  lambda1 <- step1$lambda1[[1]]
  if (is.null(lambda1) || !is.finite(lambda1)) {
    lambda1 <- 1
  }
  # effective second-order multiplier of the unscaled var.error term,
  # absorbing the alpha correction (see lav_sam_veta2())
  alpha_n1 <- lav_sam_alpha_n1(step1$alpha[[1]], n)
  lambda2_eff <- lambda_star * (1 - alpha_n1)

  # dummy lvs (observed variables in the structural part): mirror the
  # patches of lav_sam_step1_local() inside lbar() below; a dummy lv's
  # factor score is the observed variable itself (no measurement error),
  # and the patched rows/columns are *constants* (zero derivative); for
  # M.method = "ML" the patches are usually redundant (the model matrices
  # already encode lambda = 1/theta = 0, and the mapping matrix reproduces
  # the unit rows), but for M.method = "GLS" the unpatched mapping matrix
  # rows of the dummy lvs are *not* unit vectors
  lambda_gidx <- which(names(lavmodel@GLIST) == "lambda")[1]
  lv_names_full <- lavmodel@dimNames[[lambda_gidx]][[2L]]
  rm_idx <- lavpta$vidx$lv.interaction[[1]]
  lv_names_noint <- lv_names_full
  if (length(rm_idx) > 0L) {
    lv_names_noint <- lv_names_full[-rm_idx]
  }
  dummy_ov_y_idx <- lavmodel@ov.y.dummy.ov.idx[[1]]
  dummy_lv_y_idx <- match(lv_names_full[lavmodel@ov.y.dummy.lv.idx[[1]]],
                          lv_names_noint)
  dummy_ov_idx <- c(lavmodel@ov.x.dummy.ov.idx[[1]], dummy_ov_y_idx)
  dummy_lv_idx <- match(lv_names_full[c(lavmodel@ov.x.dummy.lv.idx[[1]],
                                        lavmodel@ov.y.dummy.lv.idx[[1]])],
                        lv_names_noint)

  # std.lv = TRUE? if so, lav_sam_step1_local() has rescaled VETA
  # (cov2cor) and expressed the factor scores in the same standardized
  # metric; we must mirror this inside lbar() below (the rescaling factors
  # depend on the step 1 parameters, so they are differentiated through)
  std_lv_flag <- fit@Options$std.lv

  # step 1 free parameters
  step1_idx <- which(step1$PT$free %in% step1$step1.free.idx)
  x_step1 <- step1$PT$est[step1_idx]
  pt_1 <- step1$PT

  # the average casewise contribution to the augmented summary statistics
  #   Lbar = [ E(eta* %x% eta*)[keep], vech(Var(eta* %x% eta*)[keep, keep]) ]
  # as a function of the step 1 free parameters (holding the data, and
  # lambda.star/lambda1, fixed); see the casewise contributions (iveta2_1)
  # in lav_sam_veta2(); the averages equal the EETA2/VETA2 statistics that
  # are actually used in step 2:
  # with f_i = (1, M (y_i - nu)), B = bdiag(0, M THETA t(M)),
  # C2 = (1/N) \sum_i f_i t(f_i), E = C2 - lambda1 * B, and writing
  # (r1, s1) and (r2, s2) for the row/column pairs of elements 'a' and 'b'
  # of (eta* %x% eta*), the averages of the casewise contributions are:
  # - first-order part:
  #     Lbar1[a] = (1/N) \sum_i (f_i %x% f_i)[a] - lambda1 * vec(B)[a]
  # - second-order part (before taking vech):
  #     Lbar2[a, b] = Var(fs2)[a, b] - lambda2.eff * tmpbar[a, b]
  #   with lambda2.eff = lambda.star * (1 - alpha/(n-1))
  #   where fs2_i = f_i %x% f_i, and tmpbar is the average of
  #     tmp_i = ((F_i - lambda1 * B) %x% B) + (B %x% (F_i - lambda1 * B)) +
  #             ((F_i - lambda1 * B) %x% B) %*% K +
  #             K %*% ((F_i - lambda1 * B) %x% B) +
  #             (I + K) %*% (B %x% B)              (with F_i = f_i t(f_i))
  #   which depends on the f_i's only through C2; elementwise:
  #     tmpbar[a, b] = E[r1, r2] B[s1, s2] + B[r1, r2] E[s1, s2] +
  #                    E[r1, s2] B[s1, r2] + E[s1, r2] B[r1, s2] +
  #                    B[r1, r2] B[s1, s2] + B[s1, r2] B[r1, s2]
  lbar <- function(x, .return_all = FALSE) {
    pt_x <- pt_1
    pt_x$est[step1_idx] <- x
    this_lavmodel <- lav_model_set_parameters(lavmodel,
                       x = pt_x$est[pt_x$free > 0 & !duplicated(pt_x$free)])
    this_nu     <- drop(this_lavmodel@GLIST$nu)
    # no interaction columns!
    if (length(rm_idx) > 0L) {
      this_lambda <- this_lavmodel@GLIST$lambda[, -rm_idx, drop = FALSE]
    } else {
      this_lambda <- this_lavmodel@GLIST$lambda
    }
    this_theta  <- this_lavmodel@GLIST$theta
    # dummy lv patches (cfr. lav_sam_step1_local())
    if (length(dummy_ov_y_idx) > 0L) {
      this_theta[dummy_ov_y_idx, ] <- 0
      this_theta[, dummy_ov_y_idx] <- 0
      this_lambda[dummy_ov_y_idx, ] <- 0
      this_lambda[cbind(dummy_ov_y_idx, dummy_lv_y_idx)] <- 1
      this_nu[dummy_ov_y_idx] <- 0
    }
    this_m <- lav_sam_mapping_mat(mm_lambda = this_lambda,
                                     mm_theta = this_theta,
                                     s = step1$COV[[1]],
                                     method = step1$local.options$M.method)
    if (length(dummy_ov_idx) > 0L) {
      this_m[dummy_lv_idx, ] <- 0
      this_m[cbind(dummy_lv_idx, dummy_ov_idx)] <- 1
    }
    m_pre_std <- this_m
    d_std <- NULL
    if (std_lv_flag) {
      # mirror the cov2cor() rescaling of the first-order VETA: rescale
      # the rows of the mapping matrix so that the (lambda1-corrected)
      # factor-score covariance matrix has unit diagonal; dummy lvs are
      # never rescaled
      msm_diag <- rowSums((this_m %*% step1$COV[[1]]) * this_m)
      mtm_diag <- rowSums((this_m %*% this_theta) * this_m)
      d_std <- sqrt(msm_diag - lambda1 * mtm_diag)
      d_std[dummy_lv_idx] <- 1
      this_m <- this_m / d_std
    }
    b_aug <- lav_mat_bdiag(0, this_m %*% this_theta %*% t(this_m))

    # (Bartlett) factor scores, augmented with an intercept
    fs <- cbind(1, t(t(y) - this_nu) %*% t(this_m))

    # C2 = (1/N) \sum_i f_i t(f_i)
    c2 <- crossprod(fs) / n
    e_mat <- c2 - lambda1 * b_aug

    # first-order part
    out1 <- e_mat[cbind(i2k, i1k)]

    # second-order part: only the kept columns of fs2 are needed
    fs2k <- fs[, i1k, drop = FALSE] * fs[, i2k, drop = FALSE]
    fs2kc <- t(t(fs2k) - colMeans(fs2k))
    var_fs2k <- crossprod(fs2kc) / n

    tmpbar <- (e_mat[i1k, i1k] * b_aug[i2k, i2k] +
               b_aug[i1k, i1k] * e_mat[i2k, i2k] +
               e_mat[i1k, i2k] * b_aug[i2k, i1k] +
               e_mat[i2k, i1k] * b_aug[i1k, i2k] +
               b_aug[i1k, i1k] * b_aug[i2k, i2k] +
               b_aug[i2k, i1k] * b_aug[i1k, i2k])

    out_vec <- c(out1, lav_mat_vech(var_fs2k - lambda2_eff * tmpbar))
    if (!.return_all) {
      return(out_vec)
    }
    # intermediates needed by the analytic jacobian below
    list(vec = out_vec, nu = this_nu, lambda = this_lambda,
         theta = this_theta, m_pre_std = m_pre_std, m = this_m,
         b = b_aug, e = e_mat, fs = fs, fs2kc = fs2kc, d_std = d_std)
  }

  # jacobian of Lbar w.r.t. the step 1 parameters: cveta[a, k] = d Lbar[a]/d x_k
  # By default we use a closed-form (analytic) jacobian, which avoids the
  # repeated mapping-matrix solves and full factor-score recomputations of the
  # numerical (numDeriv) jacobian. The analytic path applies the chain rule
  # through the model-implied factor scores, with the key building block being
  # the differential of the mapping matrix M = (Lambda' W Lambda)^-1 Lambda' W
  # (W = I, S^-1, Theta^-1 for ULS/GLS/ML):
  #   dM = A dLambda' W Pp  -  M dLambda M  [ -  M dTheta Theta^-1 Pp (ML) ]
  # with A = (Lambda' W Lambda)^-1 and Pp = I - Lambda M. It is only taken for
  # the 'clean' continuous setting (standard mapping matrix, no dummy latent
  # variables, no empty Lambda columns, invertible Theta for ML); otherwise we
  # fall back to numDeriv. See lav_sam_gamma_add_jac().
  cveta <- NULL
  if (use_analytic) {
    cveta <- tryCatch(
      lav_sam_gamma_add_jac(
        base = lbar(x_step1, .return_all = TRUE),
        x_step1 = x_step1, lavmodel = lavmodel, pt_1 = pt_1,
        step1_idx = step1_idx, rm_idx = rm_idx,
        dummy_ov_idx = dummy_ov_idx, s_cov = step1$COV[[1]],
        method = step1$local.options$M.method, std_lv_flag = std_lv_flag,
        lambda1 = lambda1, lambda2_eff = lambda2_eff,
        i1k = i1k, i2k = i2k, y = y, n = n),
      error = function(e) NULL)
  }
  if (is.null(cveta)) {
    cveta <- numDeriv::jacobian(func = lbar, x = x_step1)
  }

  gamma_addition <- n * (cveta %*% step1$Sigma.11 %*% t(cveta))
  gamma_addition
}

# Closed-form jacobian of the augmented summary statistics Lbar w.r.t. the
# step 1 parameters (the analytic counterpart of numDeriv::jacobian(lbar)).
# Returns the d_L x n_par jacobian matrix, or NULL if the 'clean' continuous
# setting does not apply (the caller then falls back to numDeriv).
lav_sam_gamma_add_jac <- function(base = NULL, x_step1 = NULL, lavmodel = NULL,
                                  pt_1 = NULL, step1_idx = NULL, rm_idx = NULL,
                                  dummy_ov_idx = NULL, s_cov = NULL,
                                  method = "ML", std_lv_flag = FALSE,
                                  lambda1 = 1, lambda2_eff = 1,
                                  i1k = NULL, i2k = NULL, y = NULL, n = NULL) {
  Lam0 <- base$lambda; Th0 <- base$theta; nu0 <- base$nu
  M0 <- base$m_pre_std; Mf <- base$m; fs0 <- base$fs
  e0 <- base$e; b0 <- base$b; fs2kc0 <- base$fs2kc; d_std0 <- base$d_std
  S <- s_cov
  p <- nrow(Lam0); mlv <- ncol(Lam0)
  method <- toupper(method)

  # eligibility: no dummy lvs, no empty Lambda columns, standard mapping matrix
  if (length(dummy_ov_idx) > 0L) {
    return(NULL)
  }
  if (any(apply(Lam0, 2L, function(z) all(z == 0)))) {
    return(NULL)
  }

  # weight matrix W and A = (Lambda' W Lambda)^-1
  if (method == "ULS") {
    W <- diag(p)
  } else if (method == "GLS") {
    W <- solve(S)
  } else if (method == "ML") {
    if (any(abs(diag(Th0)) < 1e-4)) {
      return(NULL)
    }
    W <- solve(Th0)
  } else {
    return(NULL)
  }
  LtW <- crossprod(Lam0, W) # mlv x p
  A <- solve(LtW %*% Lam0)
  # is the mapping matrix the standard weighted pseudo-inverse? (if lbar used
  # a fallback -- tmat/ginv/ULS -- the closed form below does not apply)
  if (max(abs(A %*% LtW - M0)) > 1e-7) {
    return(NULL)
  }
  Pp <- diag(p) - Lam0 %*% M0 # residual projector (p x p)
  WPp <- W %*% Pp
  ThiPp <- if (method == "ML") W %*% Pp else NULL # Theta^-1 Pp

  # cheap GLIST extractor (set_parameters only; no mapping/scores). The
  # Lambda/Theta/nu entries are linear in the free parameters, so the forward
  # difference below is exact (up to round-off).
  get_lln <- function(x) {
    pt_x <- pt_1
    pt_x$est[step1_idx] <- x
    ml <- lav_model_set_parameters(lavmodel,
            x = pt_x$est[pt_x$free > 0 & !duplicated(pt_x$free)])
    if (length(rm_idx) > 0L) {
      lam <- ml@GLIST$lambda[, -rm_idx, drop = FALSE]
    } else {
      lam <- ml@GLIST$lambda
    }
    list(nu = drop(ml@GLIST$nu), lambda = lam, theta = ml@GLIST$theta)
  }

  Ymc <- t(t(y) - nu0) # n x p  = Y - 1 nu'
  np <- length(x_step1)
  cv <- matrix(0, length(base$vec), np)
  eps_g <- 1e-6
  for (k in seq_len(np)) {
    xk <- x_step1
    xk[k] <- xk[k] + eps_g
    gk <- get_lln(xk)
    dLam <- (gk$lambda - Lam0) / eps_g
    dTh  <- (gk$theta  - Th0)  / eps_g
    dnu  <- (gk$nu     - nu0)  / eps_g

    # raw mapping-matrix derivative (pre-std rescaling)
    dM0 <- A %*% crossprod(dLam, WPp) - M0 %*% (dLam %*% M0)
    if (!is.null(ThiPp)) {
      dM0 <- dM0 - M0 %*% (dTh %*% ThiPp)
    }

    # std.lv rescaling chain: Mf = M0 / d_std, d_std_j = sqrt(msm_j - l1 mtm_j)
    if (std_lv_flag) {
      d_msm <- 2 * rowSums((dM0 %*% S) * M0)
      d_mtm <- 2 * rowSums((dM0 %*% Th0) * M0) + rowSums((M0 %*% dTh) * M0)
      d_dstd <- (d_msm - lambda1 * d_mtm) / (2 * d_std0)
      dMf <- dM0 / d_std0 - M0 * (d_dstd / d_std0^2)
    } else {
      dMf <- dM0
    }

    # b_aug = bdiag(0, M Theta M'), augmented factor scores fs, C2, E
    dMtMt <- dMf %*% Th0 %*% t(Mf)
    d_b <- lav_mat_bdiag(0, dMtMt + Mf %*% dTh %*% t(Mf) + t(dMtMt))
    d_score <- Ymc %*% t(dMf) - matrix(drop(Mf %*% dnu), n, mlv, byrow = TRUE)
    d_fs <- cbind(0, d_score)
    d_e <- (crossprod(d_fs, fs0) + crossprod(fs0, d_fs)) / n - lambda1 * d_b

    # first-order part
    d_out1 <- d_e[cbind(i2k, i1k)]

    # second-order part
    d_fs2k <- d_fs[, i1k, drop = FALSE] * fs0[, i2k, drop = FALSE] +
              fs0[, i1k, drop = FALSE] * d_fs[, i2k, drop = FALSE]
    d_fs2kc <- t(t(d_fs2k) - colMeans(d_fs2k))
    d_var <- (crossprod(d_fs2kc, fs2kc0) + crossprod(fs2kc0, d_fs2kc)) / n
    d_tmp <- (d_e[i1k, i1k] * b0[i2k, i2k] + e0[i1k, i1k] * d_b[i2k, i2k] +
              d_b[i1k, i1k] * e0[i2k, i2k] + b0[i1k, i1k] * d_e[i2k, i2k] +
              d_e[i1k, i2k] * b0[i2k, i1k] + e0[i1k, i2k] * d_b[i2k, i1k] +
              d_e[i2k, i1k] * b0[i1k, i2k] + e0[i2k, i1k] * d_b[i1k, i2k] +
              d_b[i1k, i1k] * b0[i2k, i2k] + b0[i1k, i1k] * d_b[i2k, i2k] +
              d_b[i2k, i1k] * b0[i1k, i2k] + b0[i2k, i1k] * d_b[i1k, i2k])

    cv[, k] <- c(d_out1, lav_mat_vech(d_var - lambda2_eff * d_tmp))
  }

  cv
}

# generate labels for each parameter
lav_pt_labels <- function(partable,
                                blocks = c("group", "level"),
                                group_equal = "", group_partial = "",
                                type = "user") {
  # catch empty partable
  if (length(partable$lhs) == 0L) {
    return(character(0L))
  }

  # default labels
  label <- paste(partable$lhs, partable$op, partable$rhs, sep = "")

  # handle multiple groups
  if ("group" %in% blocks) {
    if (is.character(partable$group)) {
      group_label <- unique(partable$group)
      group_label <- group_label[nchar(group_label) > 0L]
      ngroups <- length(group_label)
    } else {
      ngroups <- lav_pt_ngroups(partable)
      group_label <- 1:ngroups
    }
    if (ngroups > 1L) {
      for (g in 2:ngroups) {
        label[partable$group == group_label[g]] <-
          paste(label[partable$group == group_label[g]],
            ".g", g,
            sep = ""
          )
      }
    }
  } else {
    ngroups <- 1L
  }

  # cat("DEBUG: label start:\n"); print(label); cat("\n")
  # cat("group_equal = ", group_equal, "\n")
  # cat("group_partial = ", group_partial, "\n")

  # use group_equal so that equal sets of parameters get the same label
  if (ngroups > 1L && length(group_equal) > 0L) {
    if ("intercepts" %in% group_equal ||
      "residuals" %in% group_equal ||
      "residual.covariances" %in% group_equal) {
      ov_names_nox <- vector("list", length = ngroups)
      for (g in 1:ngroups) {
        ov_names_nox[[g]] <- unique(unlist(
          lav_pt_vnames(partable, "ov.nox", group = g)))
      }
    }
    if ("thresholds" %in% group_equal) {
      ov_names_ord <- vector("list", length = ngroups)
      for (g in 1:ngroups) {
        ov_names_ord[[g]] <- unique(unlist(
          lav_pt_vnames(partable, "ov.ord", group = g)))
      }
    }
    if ("means" %in% group_equal ||
      "lv.variances" %in% group_equal ||
      "lv.covariances" %in% group_equal) {
      lv_names <- vector("list", length = ngroups)
      for (g in 1:ngroups) {
        lv_names[[g]] <- unique(unlist(
          lav_pt_vnames(partable, "lv", group = g)))
      }
    }

    # eq_flag: TRUE if the parameter is a candidate for a (cross-group)
    # equality constraint. Note that the models need NOT be the same across
    # groups (eg when the 'group:' syntax is used): a parameter qualifies in
    # a given group if it belongs to the requested set *in that group*.
    eq_flag <- logical(length(partable$lhs))

    # helper: TRUE for a row if its 'lhs' is in 'names_list[[group]]' of the
    # group that the row belongs to
    in_group_names <- function(names_list) {
      out <- logical(length(partable$lhs))
      for (g in seq_len(ngroups)) {
        out[partable$group == g &
          partable$lhs %in% names_list[[g]]] <- TRUE
      }
      out
    }

    # LOADINGS
    if ("loadings" %in% group_equal) {
      eq_flag[partable$op == "=~"] <- TRUE
    }
    # COMPOSITE LOADINGS (new in 0.6-4)
    if ("composite.loadings" %in% group_equal) {
      # new setting (0.6-20): <~
      if (any(partable$op == "<~")) {
        lav_msg_warn(gettext("composite.loadings are in fact composite weights;
                              better use composite.weights"))
        eq_flag[partable$op == "<~"] <- TRUE
      } else {
        # old school: composites are phantom constructs with zero residual...
        lv_f_names <- unique(unlist(
          lav_pt_vnames(partable, "lv.formative")))
        eq_flag[partable$op == "~" &
                partable$lhs %in% lv_f_names] <- TRUE
      }
    }
    # COMPOSITE WEIGHTS (new in 0.6-20) # same as 'loadings'...
    if ("composite.weights" %in% group_equal) {
      eq_flag[partable$op == "<~"] <- TRUE
    }
    # INTERCEPTS (OV)
    if ("intercepts" %in% group_equal) {
      eq_flag[partable$op == "~1" & in_group_names(ov_names_nox)] <- TRUE
    }
    # THRESHOLDS (OV-ORD)
    if ("thresholds" %in% group_equal) {
      eq_flag[partable$op == "|" & in_group_names(ov_names_ord)] <- TRUE
    }
    # MEANS (LV)
    if ("means" %in% group_equal) {
      eq_flag[partable$op == "~1" & in_group_names(lv_names)] <- TRUE
    }
    # REGRESSIONS
    if ("regressions" %in% group_equal) {
      eq_flag[partable$op == "~"] <- TRUE
    }
    # RESIDUAL variances (FIXME: OV ONLY!)
    if ("residuals" %in% group_equal) {
      eq_flag[partable$op == "~~" & in_group_names(ov_names_nox) &
        partable$lhs == partable$rhs] <- TRUE
    }
    # RESIDUAL covariances (FIXME: OV ONLY!)
    if ("residual.covariances" %in% group_equal) {
      eq_flag[partable$op == "~~" & in_group_names(ov_names_nox) &
        partable$lhs != partable$rhs] <- TRUE
    }
    # LV VARIANCES
    if ("lv.variances" %in% group_equal) {
      eq_flag[partable$op == "~~" & in_group_names(lv_names) &
        partable$lhs == partable$rhs] <- TRUE
    }
    # LV COVARIANCES
    if ("lv.covariances" %in% group_equal) {
      eq_flag[partable$op == "~~" & in_group_names(lv_names) &
        partable$lhs != partable$rhs] <- TRUE
    }

    # if group_partial, do NOT constrain these parameters (in any group)
    if (length(group_partial) > 0L) {
      base_label <- paste(partable$lhs, partable$op, partable$rhs, sep = "")
      eq_flag[base_label %in% group_partial] <- FALSE
    }

    # Give all candidate parameters that share the same lhs/op/rhs the same
    # label, so they become equality-constrained -- across *every* group that
    # contains them, independently of group 1. Parameters that appear in a
    # single group only keep their unique label (and stay free). This way the
    # constraints are symmetric: a parameter present in groups 2 and 3 (but
    # not group 1) is constrained just as one present in groups 1 and 2.
    eq_idx <- which(eq_flag)
    if (length(eq_idx) > 0L) {
      eq_key <- paste(partable$lhs, partable$op, partable$rhs,
        sep = "")[eq_idx]
      for (key in unique(eq_key)) {
        same_idx <- eq_idx[eq_key == key]
        if (length(same_idx) > 1L) {
          label[same_idx] <- label[same_idx[1L]]
        }
      }
    }
  }

  # cat("DEBUG: g1.idx = ", g1.idx, "\n")
  # cat("DEBUG: label after group_equal:\n"); print(label); cat("\n")

  # handle other block identifier (not 'group')
  for (block in blocks) {
    if (block == "group") {
      next
    } else if (block == "level" && !is.null(partable[[block]])) {
      # all but first level
      lev_vals <- lav_pt_level_values(partable)
      idx <- which(partable[[block]] != lev_vals[1])
      label[idx] <- paste(label[idx], ".", "l",
        partable[[block]][idx],
        sep = ""
      )
    } else if (!is.null(partable[[block]])) {
      label <- paste(label, ".", block, partable[[block]], sep = "")
    }
  }

  # user-specified labels -- override everything!!
  user_idx <- which(nchar(partable$label) > 0L)
  label[user_idx] <- partable$label[user_idx]

  # cat("DEBUG: user.idx = ", user.idx, "\n")
  # cat("DEBUG: label after user.idx:\n"); print(label); cat("\n")

  # which labels do we need?
  if (type == "user") {
    idx <- seq_along(label)
  } else if (type == "free") {
    # idx <- which(partable$free > 0L & !duplicated(partable$free))
    idx <- which(partable$free > 0L)
    # } else if(type == "unco") {
    #    idx <- which(partable$unco > 0L & !duplicated(partable$unco))
  } else {
    lav_msg_stop(gettext("argument `type' must be one of free or user"))
  }

  label[idx]
}

# create random starting values starting from a parameter table
# - using the lower/upper bounds and runif() for factor loadings
#   and variances
# - using runif(,-0.5,+0.5) for correlations; rescale using variances
# - check if Sigma.hat is PD; if not, try again
#
# YR 26 Feb 2024

lav_pt_random <- function(lavpartable = NULL,
                                # needed if we still need to compute bounds:
                                lavh1 = NULL, lavdata = NULL,
                                lavsamplestats = NULL, lavoptions = NULL) {

  lavpta <- lav_pt_attributes(lavpartable)

  # ALWAYS (recompute) bounds, as user may have provide other
  # bounds (eg "pos.var") (0.6-20)
  lavoptions2 <- lavoptions
  lavoptions2$bounds <- "standard"
  lavoptions2$optim.bounds <-
    list(
      lower = c(
        "ov.var", "lv.var", "loadings",
        "covariances"
      ),
      upper = c(
        "ov.var", "lv.var", "loadings",
        "covariances"
      ),
      lower.factor = c(1.0, 1.0, 1.0, 0.999),
      upper.factor = c(1.0, 1.0, 1.0, 0.999),
      min.reliability.marker = 0.1,
      min.var.lv.endo = 0.005
    )
  lavpartable <- lav_pt_add_bounds(
    partable = lavpartable,
    lavh1 = lavh1, lavdata = lavdata,
    lavsamplestats = lavsamplestats, lavoptions = lavoptions2
  )

  # replace -Inf/Inf by -1/1 * .Machine$double.eps (for runif)
  inf_idx <- which(lavpartable$lower < -1e+16)
  if (length(inf_idx) > 0L) {
    lavpartable$lower[inf_idx] <- -1e+16
  }
  inf_idx <- which(lavpartable$upper > 1e+16)
  if (length(inf_idx) > 0L) {
    lavpartable$upper[inf_idx] <- 1e+16
  }

  # empty lavpartable$start?
  if (is.null(lavpartable$start)) {
    start_1 <- numeric(length(lavpartable$lhs))
    # set loadings to 0.7
    loadings_idx <- which(lavpartable$free > 0L &
      lavpartable$op == "=~")
    if (length(loadings_idx) > 0L) {
      start_1[loadings_idx] <- 0.7
    }
    # set (only) variances to 1
    var_idx <- which(lavpartable$free > 0L &
      lavpartable$op == "~~" &
      lavpartable$lhs == lavpartable$rhs)
    if (length(var_idx) > 0L) {
      start_1[var_idx] <- 1
    }

    lavpartable$start <- start_1
  }

  # initial values
  start_1 <- lavpartable$start

  nblocks <- lav_pt_nblocks(lavpartable)
  block_values <- lav_pt_block_values(lavpartable)
  for (b in 1:nblocks) {
    ov_names <- lavpta$vnames$ov[[b]]
    lv_names <- lavpta$vnames$lv[[b]]
    ov_ind_names <- lavpta$vnames$ov.ind[[b]]

    # start with the lv (residual) variances
    lv_var_idx <- which(lavpartable$block == block_values[b] &
      lavpartable$op == "~~" &
      lavpartable$lhs %in% lv_names &
      lavpartable$rhs %in% lv_names &
      lavpartable$lhs == lavpartable$rhs)

    if (length(lv_var_idx) > 0L) {
      for (i in lv_var_idx) {
        if (lavpartable$free[i] > 0L &&
          (lavpartable$lower[i] < lavpartable$upper[i])) {
          start_1[i] <- runif(
            n = 1L, min = lavpartable$lower[i],
            max = lavpartable$upper[i]
          )
        }
      }
    }

    # first, we generate lv correlations, and then rescale to covariances
    lv_cov_idx <- which(lavpartable$block == block_values[b] &
      lavpartable$op == "~~" &
      lavpartable$lhs %in% lv_names &
      lavpartable$rhs %in% lv_names &
      lavpartable$lhs != lavpartable$rhs)

    if (length(lv_cov_idx) > 0L) {
      for (i in lv_cov_idx) {
        if (lavpartable$free[i] > 0L &&
          (lavpartable$lower[i] < lavpartable$upper[i])) {
          cor_val <- runif(n = 1L, -0.5, +0.5)
          var1_idx <- which(lavpartable$block == block_values[b] &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$lhs[i] &
            lavpartable$lhs == lavpartable$rhs)
          var2_idx <- which(lavpartable$block == block_values[b] &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$rhs[i] &
            lavpartable$lhs == lavpartable$rhs)
          start_1[i] <- cor_val * sqrt(start_1[var1_idx]) *
                                  sqrt(start_1[var2_idx])
        }
      }
    }

    # next, (residual) ov variances
    ov_var_idx <- which(lavpartable$block == block_values[b] &
      lavpartable$op == "~~" &
      lavpartable$lhs %in% ov_names &
      lavpartable$rhs %in% ov_names &
      lavpartable$lhs == lavpartable$rhs)

    if (length(ov_var_idx) > 0L) {
      for (i in ov_var_idx) {
        if (lavpartable$free[i] > 0L &&
          (lavpartable$lower[i] < lavpartable$upper[i])) {
          start_1[i] <- runif(
            n = 1L, min = lavpartable$lower[i],
            max = lavpartable$upper[i]
          )
        }
      }
    }

    # (residual) ov covariances (if any)
    ov_cov_idx <- which(lavpartable$block == block_values[b] &
      lavpartable$op == "~~" &
      lavpartable$lhs %in% ov_names &
      lavpartable$rhs %in% ov_names &
      lavpartable$lhs != lavpartable$rhs)

    if (length(ov_cov_idx) > 0L) {
      for (i in ov_cov_idx) {
        if (lavpartable$free[i] > 0L &&
          (lavpartable$lower[i] < lavpartable$upper[i])) {
          cor_val <- runif(n = 1L, -0.5, +0.5)
          var1_idx <- which(lavpartable$block == block_values[b] &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$lhs[i] &
            lavpartable$lhs == lavpartable$rhs)
          var2_idx <- which(lavpartable$block == block_values[b] &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$rhs[i] &
            lavpartable$lhs == lavpartable$rhs)
          start_1[i] <- cor_val * sqrt(start_1[var1_idx]) *
                                  sqrt(start_1[var2_idx])
        }
      }
    }

    # finally, the lambda values, keeping in mind that
    # lambda_p^(u) = sqrt( upper(res.var.indicators_p) /
    #                      lower(var.factor) )
    lambda_idx <- which(lavpartable$block == block_values[b] &
      lavpartable$op == "=~" &
      lavpartable$lhs %in% lv_names &
      lavpartable$rhs %in% ov_ind_names)

    if (length(lambda_idx)) {
      for (i in lambda_idx) {
        if (lavpartable$free[i] > 0L &&
          (lavpartable$lower[i] < lavpartable$upper[i])) {
          varov_idx <- which(lavpartable$block == block_values[b] &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$rhs[i] &
            lavpartable$lhs == lavpartable$rhs)
          varlv_idx <- which(lavpartable$block == block_values[b] &
            lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$lhs[i] &
            lavpartable$lhs == lavpartable$rhs)
          lambda_u <- sqrt(start_1[varov_idx] / start_1[varlv_idx])
          start_1[i] <- runif(n = 1, -lambda_u, lambda_u)
        }
      }
    }
  }

  # sanity check; needed?
  current_warn <- lav_warn()
  if (lav_warn(TRUE))
      on.exit(lav_warn(current_warn), TRUE)
  start_1 <- lav_start_check_cov(
    lavpartable = lavpartable, start = start_1
  )

  start_1
}

# compute WLS.est (as a list per group)
#
# multilevel: the per-block (within/between) vectors are stacked per group:
# c(within, between), matching WLS.obs (lav_samp_wls_2l) and the per-group
# Delta matrix (lav_model_delta)
lav_model_wls_est <- function(lavmodel = NULL, glist = NULL,
                              lavimplied = NULL) {
  # two-level + categorical: the standardized stage-wise layout
  # (see lav_model_wls_2l_cat.R); already stacked per group
  if (lavmodel@multilevel && lavmodel@categorical) {
    return(lav_m07_wls_est(lavmodel = lavmodel, glist = glist))
  }
  nblocks <- lavmodel@nblocks
  meanstructure <- lavmodel@meanstructure
  correlation <- lavmodel@correlation
  categorical <- lavmodel@categorical
  group_w_free <- lavmodel@group.w.free
  num_idx <- lavmodel@num.idx

  # model-implied statistics
  if (is.null(lavimplied)) {
    lavimplied <- lav_model_implied(lavmodel, glist = glist)
  }

  wls_est_1 <- vector("list", length = nblocks)
  for (g in 1:nblocks) {
    if (categorical) {
      # order of elements is important here:
      # 1. thresholds + means (interleaved)
      # 2. slopes (if any, columnwise per exo)
      # 3. variances (if any)
      # 4. correlations (no diagonal!)
      if (lavmodel@conditional.x) {
        wls_est <- c(
          lavimplied$res.th[[g]],
          lav_mat_vec(lavimplied$res.slopes[[g]]),
          diag(lavimplied$res.cov[[g]])[num_idx[[g]]],
          lav_mat_vech(lavimplied$res.cov[[g]],
            diagonal = FALSE
          )
        )
      } else {
        wls_est <- c(
          lavimplied$th[[g]],
          diag(lavimplied$cov[[g]])[num_idx[[g]]],
          lav_mat_vech(lavimplied$cov[[g]],
            diagonal = FALSE
          )
        )
      }
    } else {
      # CONTINUOUS
      diag_1 <- TRUE
      if (correlation) {
        diag_1 <- FALSE
      }

      if (lavmodel@conditional.x && lavmodel@nexo[g] > 0L) {
        # order = vec(Beta), where first row are intercepts
        # cbind(res.int, res.slopes) is t(Beta)
        # so we need vecr
        if (meanstructure) {
          wls_est <- c(
            lav_mat_vecr(
              cbind(
                lavimplied$res.int[[g]],
                lavimplied$res.slopes[[g]]
              )
            ),
            lav_mat_vech(lavimplied$res.cov[[g]],
              diagonal = diag_1
            )
          )
        } else {
          wls_est <- c(
            lav_mat_vecr(lavimplied$res.slopes[[g]]),
            lav_mat_vech(lavimplied$res.cov[[g]],
              diagonal = diag_1
            )
          )
        }
      } else {
        if (correlation && length(num_idx[[g]]) > 0L) {
          # partial correlation structure: keep the variances of the
          # non-correlation variables (num.idx), drop only the correlation
          # variables' diagonal (mirrors the categorical layout above)
          v_est <- c(
            diag(lavimplied$cov[[g]])[num_idx[[g]]],
            lav_mat_vech(lavimplied$cov[[g]], diagonal = FALSE)
          )
        } else {
          v_est <- lav_mat_vech(lavimplied$cov[[g]], diagonal = diag_1)
        }
        if (meanstructure) {
          wls_est <- c(lavimplied$mean[[g]], v_est)
        } else {
          wls_est <- v_est
        }
      } # conditional.x = FALSE
    } # categorical = FALSE

    if (group_w_free) {
      wls_est <- c(lavimplied$group.w[[g]], wls_est)
    }

    wls_est_1[[g]] <- wls_est
  }

  # multilevel: stack the blocks of each group (within on top)
  if (lavmodel@multilevel) {
    nlevels <- 2L
    ngroups <- nblocks %/% nlevels
    wls_est_2 <- vector("list", length = ngroups)
    for (g in seq_len(ngroups)) {
      wls_est_2[[g]] <- c(
        wls_est_1[[(g - 1) * nlevels + 1L]],
        wls_est_1[[(g - 1) * nlevels + 2L]]
      )
    }
    wls_est_1 <- wls_est_2
  }

  wls_est_1
}

# Note: lav_model_wls_v() is replaced by lav_model_h1_info() in 0.6-1

# various fit measures

# - lav_fit_cn
# - lav_fit_wrmr
# - lav_fit_mfi
# - lav_fit_ecvi

# Y.R. 21 July 2022

# Hoelter Critical N (CN)
lav_fit_cn <- function(x2 = NULL, df = NULL, n = NULL, alpha = 0.05) {
  # catch df=0, X2=0
  if (df == 0 && x2 < sqrt(.Machine$double.eps)) { # added sqrt() in 0.6-21
    cn <- as.numeric(NA)
  } else {
    cn <- qchisq(p = (1 - alpha), df = df) / (x2 / n) + 1
  }

  cn
}

# WRMR
# we use the definition: wrmr = sqrt ( 2*N*F / nel )
# Note: when multiple groups, 'nel' only seems to correspond to the
# first group???
lav_fit_wrmr <- function(x2 = NULL, nel = NULL) {
  if (nel > 0) {
    wrmr <- sqrt(x2 / nel)
  } else {
    wrmr <- as.numeric(NA)
  }

  wrmr
}

# MFI - McDonald Fit Index (McDonald, 1989)
lav_fit_mfi <- function(x2 = NULL, df = NULL, n = NULL) {
  mfi <- exp(-0.5 * (x2 - df) / n)
  mfi
}

# ECVI - cross-validation index (Brown & Cudeck, 1989, eq 5)
# "In the special case where F = F_ML, Equation 5 [=ECVI] is the
# rescaled AIC employed by Cudeck and Browne (1983, Equation 5.1). This
# result is concordant with a finding of Stone (1977). He showed under general
# conditions that if the "leaving one out at a time" method of cross-validation
# (Stone, 1974; Geisser, 1975) is employed, a log-likelihood measure of
# predictive validity is asymptotically equivalent to the AIC." (p. 448)

# not defined for multiple groups and/or models with meanstructures
# TDJ: According to Dudgeon (2004, p. 317), "ECVI requires no adjustment
#      when a model is fitted simultaneously in multiple samples."
#      And I think the lack of mean structure in Brown & Cudeck (1989)
#      was a matter of habitual simplification back then, not necessity.
# YR: - why does Dudgeon eq 22 use (df + 2*npar) instead of (2*npar)??
lav_fit_ecvi <- function(x2 = NULL, npar = npar, n = n) {
  ecvi <- x2 / n + (2 * npar) / n
  ecvi
}

# collect information about the model that we can use
# (eg. is theta diagonal or not, is the structural model recursive or not,
#  is the model just a regression model, etc)
#
# initial version: YR 15 March 2021
# - YR 05 Oct 2021: use det(I - B) to check if B is acyclic
# - YR 11 Nov 2021: if no latents, and conditional.x = TRUE, we may have no
#                   beta matrix


# note: there is no 'lavmodel' yet, because we call this in lav_model.R
lav_model_properties <- function(glist, lavpartable = NULL,
                                 nmat = NULL, m_free_idx = NULL) {
  lavpta <- lav_pt_attributes(lavpartable)
  nblocks <- lavpta$nblocks

  # is the model a univariate/multivariate linear multiple regression
  # model (per block)?
  uvreg <- logical(nblocks)
  uvord <- logical(nblocks)
  mvreg <- logical(nblocks)
  acyclic <- rep(as.logical(NA), nblocks)
  bowfree <- rep(as.logical(NA), nblocks)
  nexo <- integer(nblocks)

  for (g in seq_len(nblocks)) {
    # at least 1 regression
    if (length(lavpta$vnames$eqs.y[[g]]) == 0L) {
      next
    }

    # find beta index for this block
    mm_in_block <- 1:nmat[g] + cumsum(c(0L, nmat))[g]
    mlist <- glist[mm_in_block]
    beta_idx <- which(names(mlist) == "beta") + cumsum(c(0L, nmat))[g]
    psi_idx <- which(names(mlist) == "psi") + cumsum(c(0L, nmat))[g]

    if (length(beta_idx) > 0L) {
      # 1. acyclic?
      m_b <- glist[[beta_idx]]
      # keep fixed values (if any); fill in 1 in all 'free' positions
      m_b[m_free_idx[[beta_idx]]] <- 1
      acyclic[g] <- lav_graph_is_acyclic(m_b)

      # 2. bow-free?
      b_one <- as.integer(m_b != 0)
      psi <- glist[[psi_idx]]
      # keep fixed values (if any); fill in 1 in all 'free' positions
      psi[m_free_idx[[psi_idx]]] <- 1
      psi_one <- as.integer(psi != 0)
      both_one <- b_one + psi_one
      if (any(both_one > 1)) {
        bowfree[g] <- FALSE
      } else {
        bowfree[g] <- TRUE
      }
    } else {
      # perhaps conditional.x = TRUE?
      # if there is no BETA, then we only have Gamma, and the
      # system must be acyclic
      acyclic[g] <- TRUE
      # and also bowfree
      bowfree[g] <- TRUE
    }


    # no latent variables, at least 1 dependent variable
    if (lavpta$nfac[[g]] > 0L) {
      next
    }

    # no mediators
    if (length(lavpta$vnames$eqs.y[[g]]) !=
      length(lavpta$vnames$ov.y[[g]])) {
      next
    }

    # categorical y?
    if (length(lavpta$vnames$ov.ord[[g]]) > 0L) {
      # we only flag the univariate version
      if (length(lavpta$vnames$ov.ord[[g]]) == 1L &&
        length(lavpta$vnames$ov.y[[g]]) == 1L &&
        lavpta$vnames$ov.ord[[g]][1] == lavpta$vnames$ov.y[[g]][1]) {
        uvord[g] <- TRUE
      }

      # mvreg?
    } else {
      if (length(lavpta$vnames$ov.y[[g]]) > 1L) {
        mvreg[g] <- TRUE
      } else {
        uvreg[g] <- TRUE
      }
    }

    nexo[g] <- length(lavpta$vnames$eqs.x[[g]])
  } # g

  modprop <- list(
    uvreg = uvreg, uvord = uvord, mvreg = mvreg,
    nexo = nexo, acyclic = acyclic, bowfree = bowfree
  )

  modprop
}

# special functions for the one-factor model
# YR 24 June 2018

# 1-factor model with (only) three indicators:
# no iterations needed; can be solved analytically

# denote s11, s22, s33 the diagonal elements, and
# s21, s31, s32 the off-diagonal elements

# under the 1-factor model; typically, either psi == 1, or l1 == 1
# - s11 == l1^2*psi + theta1
# - s22 == l2^2*psi + theta2
# - s33 == l3^2*psi + theta3
# - s21 == l2*l1*psi
# - s31 == l3*l1*psi
# - s32 == l3*l2*psi
# 6 unknowns, 6 knowns

# note: if the triad of covariances is negative, there is no
# `valid' solution, for example:
#
# > S
#      [,1] [,2] [,3]
# [1,]  1.0  0.6  0.3
# [2,]  0.6  1.0 -0.1
# [3,]  0.3 -0.1  1.0
#
# (note: all eigenvalues are positive)

lav_cfa_1fac_3ind <- function(sample_cov, std_lv = FALSE,
                              warn_neg_triad = TRUE, bounds = TRUE) {
  # check sample cov
  stopifnot(is.matrix(sample_cov))
  n_row <- NROW(sample_cov)
  n_col <- NCOL(sample_cov)
  stopifnot(n_row == n_col, n_row < 4L, n_col < 4L)
  nvar <- n_row

  # we expect a 3x3 sample covariance matrix
  # however, if we get a 2x2 (or 1x1 covariance matrix), do something
  # useful anyways...
  if (nvar == 1L) {
    # lambda = 1, theta = 0, psi = sample_cov[1,1]
    # lambda = 1, theta = 0, psi = 1 (for now, until NlsyLinks is fixed)
    sample_cov <- matrix(1, 3L, 3L) * 1.0
  } else if (nvar == 2L) {
    # hm, we could force both lambda's to be 1, but if the second
    # one is negative, this will surely lead to non-convergence issues
    #
    # just like lavaan < 0.6.2, we will use the regression of y=marker
    # on x=item2
    mean_2var <- mean(diag(sample_cov))
    max_var <- max(diag(sample_cov))
    extra <- c(mean_2var, sample_cov[2, 1])
    sample_cov <- rbind(cbind(sample_cov, extra, deparse.level = 0),
      c(extra, max_var),
      deparse.level = 0
    )
  }

  s11 <- sample_cov[1, 1]
  s22 <- sample_cov[2, 2]
  s33 <- sample_cov[3, 3]
  stopifnot(s11 > 0, s22 > 0, s33 > 0)

  s21 <- sample_cov[2, 1]
  s31 <- sample_cov[3, 1]
  s32 <- sample_cov[3, 2]
  # note: s21*s31*s32 should be positive!
  neg_triad <- FALSE
  if (s21 * s31 * s32 < 0) {
    neg_triad <- TRUE
    if (warn_neg_triad) {
      lav_msg_warn(gettext("product of the three covariances is negative!"))
    }
  }

  # first, we assume l1 = 1
  psi <- (s21 * s31) / s32 # note that we assume |s32|>0
  l1 <- 1
  l2 <- s32 / s31 # l2 <- s21/psi
  l3 <- s32 / s21 # l3 <- s31/psi
  theta1 <- s11 - psi
  theta2 <- s22 - l2 * l2 * psi
  theta3 <- s33 - l3 * l3 * psi

  # sanity check (new in 0.6-11): apply standard bounds
  if (bounds) {
    lower_psi <- s11 - (1 - 0.1) * s11 # we assume REL(y1) >= 0.1
    psi <- min(max(psi, lower_psi), s11)

    l2_bound <- sqrt(s22 / lower_psi)
    l2 <- min(max(-l2_bound, l2), l2_bound)
    l3_bound <- sqrt(s33 / lower_psi)
    l3 <- min(max(-l3_bound, l3), l3_bound)

    theta1 <- min(max(theta1, 0), s11)
    theta2 <- min(max(theta2, 0), s22)
    theta3 <- min(max(theta3, 0), s33)
  }

  lambda <- c(l1, l2, l3)
  theta <- c(theta1, theta2, theta3)


  # std.lv?
  if (std_lv) {
    # we allow for negative psi (if bounds = FALSE)
    lambda <- lambda * sign(psi) * sqrt(abs(psi))
    psi <- 1
  }

  # special cases
  if (nvar == 1L) {
    lambda <- lambda[1]
    theta <- theta[1]
  } else if (nvar == 2L) {
    lambda <- lambda[1:2]
    theta <- theta[1:2]
    psi <- psi / 2 # smaller works better?
  }

  list(lambda = lambda, theta = theta, psi = psi, neg.triad = neg_triad)
}

# FABIN (Hagglund, 1982)
# 1-factor only
lav_cfa_1fac_fabin <- function(s, lambda_only = FALSE, method = "fabin3",
                               std_lv = FALSE, bounds = TRUE) {
  # check arguments
  if (std_lv) {
    lambda_only <- FALSE # we need psi
  }

  nvar <- NCOL(s)

  # catch nvar < 4
  if (nvar < 4L) {
    out <- lav_cfa_1fac_3ind(
      sample_cov = s, std_lv = std_lv,
      warn_neg_triad = FALSE
    )
    return(out)
  }

  # 1. lambda
  lambda <- numeric(nvar)
  lambda[1L] <- 1.0
  for (i in 2:nvar) {
    idx3 <- (1:nvar)[-c(i, 1L)]
    s23 <- s[i, idx3]
    s31 <- s13 <- s[idx3, 1L]
    if (method == "fabin3") {
      s33 <- s[idx3, idx3]
      tmp <- try(solve(s33, s31), silent = TRUE) # GaussJordanPivot is
      # slightly more efficient
      if (inherits(tmp, "try-error")) {
        lambda[i] <- sum(s23 * s31) / sum(s13^2)
      } else {
        lambda[i] <- sum(s23 * tmp) / sum(s13 * tmp)
      }
    } else {
      lambda[i] <- sum(s23 * s31) / sum(s13^2)
    }
  }

  # bounds? (new in 0.6-11)
  if (bounds) {
    s11 <- s[1, 1]
    lower_psi <- s11 - (1 - 0.1) * s11 # we assume REL(y1) >= 0.1
    for (i in 2:nvar) {
      l_bound <- sqrt(s[i, i] / lower_psi)
      lambda[i] <- min(max(-l_bound, lambda[i]), l_bound)
    }
  }

  if (lambda_only) {
    return(list(
      lambda = lambda, psi = as.numeric(NA),
      theta = rep(as.numeric(NA), nvar)
    ))
  }

  # 2. theta

  # GLS version
  # W <- solve(S)
  # LAMBDA <- as.matrix(lambda)
  # A1 <- solve(t(LAMBDA) %*% W %*% LAMBDA) %*% t(LAMBDA) %*% W
  # A2 <- W %*% LAMBDA %*% A1

  # tmp1 <- W*W - A2*A2
  # tmp2 <- diag( W %*% S %*% W - A2 %*% S %*% A2 )
  # theta.diag <- solve(tmp1, tmp2)

  # 'least squares' version, assuming W = I
  d <- tcrossprod(lambda) / sum(lambda^2)
  theta <- solve(diag(nvar) - d * d, diag(s - (d %*% s %*% d)))

  # 3. psi (W=I)
  s1 <- s - diag(theta)
  l2 <- sum(lambda^2)
  psi <- sum(colSums(as.numeric(lambda) * s1) * lambda) / (l2 * l2)

  # std.lv?
  if (std_lv) {
    # we allow for negative psi
    lambda <- lambda * sign(psi) * sqrt(abs(psi))
    psi <- 1
  }

  list(lambda = lambda, theta = theta, psi = psi)
}

lav_pml_plrt2 <- function(lavobject = NULL, lavmodel = NULL, lavdata = NULL,
                          lavsamplestats = NULL, lavpartable = NULL,
                          lavoptions = NULL, x = NULL, vcov_1 = NULL,
                          lavcache = NULL) {
  lavpta <- NULL
  if (!is.null(lavobject)) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
    lavsamplestats <- lavobject@SampleStats
    lavcache <- lavobject@Cache
    lavpartable <- lav_pt_set_cache(lavobject@ParTable, lavobject@pta)
    lavpta <- lavobject@pta
  }
  if (is.null(lavpta)) {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }

  if (is.null(x)) {
    # compute 'fx' = objective function value
    # (NOTE: since 0.5-18, NOT divided by N!!)
    fx <- lav_model_objective(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache
    )
    h0_fx <- as.numeric(fx)
    h0_fx_group <- attr(fx, "fx.group")
  } else {
    h0_fx <- attr(attr(x, "fx"), "fx.pml")
    h0_fx_group <- attr(attr(x, "fx"), "fx.group")
  }

  # fit a saturated model 'fittedSat'
  model_sat <- lav_pt_unrestricted(
    lavobject = NULL,
    lavdata = lavdata,
    lavoptions = lavoptions,
    lavpta = lavpta,
    lavsamplestats = lavsamplestats
  )

  # FIXME: se="none", test="none"??
  options_1 <- lavoptions
  options_1$se <- "none"
  options_1$test <- "none"
  options_1$baseline <- FALSE
  options_1$h1 <- FALSE
  fitted_sat <- lavaan(model_sat,
    slot_options = options_1, verbose = FALSE,
    slot_sample_stats = lavsamplestats,
    slot_data = lavdata, slot_cache = lavcache
  )
  fx <- lav_model_objective(
    lavmodel = fitted_sat@Model,
    lavsamplestats = fitted_sat@SampleStats,
    lavdata = fitted_sat@Data,
    lavcache = fitted_sat@Cache
  )
  sat_fx <- as.numeric(fx)
  sat_fx_group <- attr(fx, "fx.group")

  # we also need a `saturated model', but where the moments are based
  # on the model-implied sample statistics under H0
  model_sat2 <-
    lav_pt_unrestricted(
      lavobject = NULL,
      lavdata = lavdata,
      lavoptions = lavoptions,
      lavpta = lavpta,
      lavsamplestats = NULL,
      sample_cov = lav_model_sigma(lavmodel),
      sample_mean = lav_model_mu(lavmodel),
      sample_th = lav_model_th(lavmodel),
      sample_th_idx = lavsamplestats@th.idx
    )

  options2 <- options_1
  options2$optim.method <- "none"
  options2$optim.force.converged <- TRUE
  fitted_sat2 <- lavaan(model_sat2,
    slot_options = options2, verbose = FALSE,
    slot_sample_stats = lavsamplestats,
    slot_data = lavdata, slot_cache = lavcache
  )

  # the code below was contributed by Myrsini Katsikatsou (Jan 2015)

  # for now, only a single group is supported:
  # g = 1L


  ########################### The code for PLRT for overall goodness of fit

  ##### Section 1. Compute the asymptotic mean and variance
  #####            of the first quadratic quantity
  # if(is.null(VCOV)) {
  #    VCOV <- lav_model_vcov(lavmodel       = lavmodel,
  #                           lavsamplestats = lavsamplestats,
  #                           lavoptions     = lavoptions,
  #                           lavdata        = lavdata,
  #                           lavpartable    = lavpartable,
  #                           lavcache       = lavcache)
  # }
  # G.inv
  # InvG_attheta0 <- lavsamplestats@ntotal * VCOV[,]
  # Hessian
  # H_attheta0 <- solve(attr(VCOV, "E.inv"))

  # inverted observed information ('H.inv')
  if (!is.null(vcov_1)) {
    h0_inv <- attr(vcov_1, "E.inv")
  }
  if (is.null(vcov_1) || is.null(h0_inv)) {
    # lav_model_vcov() only attaches "E.inv" when the vcov and test
    # information options coincide (they do not, e.g., with cluster =);
    # (re)compute it here
    h0_inv <- lav_model_info_observed(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavoptions = lavoptions,
      lavcache = lavcache, augmented = TRUE, inverted = TRUE
    )
  }

  # first order information ('J')
  if (is.null(vcov_1)) {
    j0 <- lav_model_info_firstorder(
      lavmodel = lavmodel, lavoptions = lavoptions,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavcache = lavcache
    )[, ]
  } else {
    # we do not get J, but J.group, FIXME?
    j0 <- lav_model_info_firstorder(
      lavmodel = lavmodel, lavoptions = lavoptions,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavcache = lavcache
    )[, ]
  }

  # inverted Godambe information
  g0_inv <- h0_inv %*% j0 %*% h0_inv

  h0tmp_prod1 <- h0_inv %*% j0
  # H0tmp_prod1 <- InvG_attheta0 %*% H_attheta0
  h0tmp_prod2 <- h0tmp_prod1 %*% h0tmp_prod1
  e_tww <- sum(diag(h0tmp_prod1))
  var_tww <- 2 * sum(diag(h0tmp_prod2))

  ##### Section 2: Compute the asymptotic mean and variance
  #####            of the second quadratic quantity.
  tmp_options <- fitted_sat2@Options
  tmp_options$se <- "robust.huber.white"
  vcov_sat2 <- lav_model_vcov(
    lavmodel = fitted_sat2@Model,
    lavsamplestats = fitted_sat2@SampleStats,
    lavoptions = tmp_options,
    lavdata = fitted_sat2@Data,
    lavpartable = fitted_sat2@ParTable,
    lavcache = fitted_sat2@Cache
  )
  # G.inv at vartheta_0
  inv_g_at_vartheta0 <- lavsamplestats@ntotal * vcov_sat2[, ]
  # Hessian at vartheta_0
  h_at_vartheta0 <- solve(attr(vcov_sat2, "E.inv")) # should always work
  # H1.inv <- lavTech(fittedSat2, "inverted.information.observed")
  # J1     <- lavTech(fittedSat2, "information.first.order")
  # H1tmp_prod1 <- H1.inv %*% J1
  h1tmp_prod1 <- inv_g_at_vartheta0 %*% h_at_vartheta0
  h1tmp_prod2 <- h1tmp_prod1 %*% h1tmp_prod1
  e_tzz <- sum(diag(h1tmp_prod1))
  var_tzz <- 2 * sum(diag(h1tmp_prod2))


  ##### Section 3: Compute the asymptotic covariance
  #####            of the two quadratic quantities

  drhodpsi_mat_1 <- vector("list", length = lavsamplestats@ngroups)
  group_values <- lav_pt_group_values(fitted_sat2@ParTable)
  for (g in 1:lavsamplestats@ngroups) {
    delta_g <- lav_model_delta(lavmodel)[[g]]
    # order of the rows: first the thresholds, then the correlations
    # we need to map the rows of delta.g to the rows/cols of H_at_vartheta0
    # of H1

    pt_1 <- fitted_sat2@ParTable
    pt_1$label <- lav_pt_labels(pt_1)
    free_idx <- which(pt_1$free > 0 & pt_1$group == group_values[g])
    parlabel <- pt_1$label[free_idx]

    # for now, we can assume that lav_model_delta will always return
    # the thresholds first, then the correlations
    #
    # later, we should add a (working) add.labels = TRUE option to
    # lav_model_delta
    th_names <- lavpta$vnames$th[[g]]
    ov_names <- lavpta$vnames$ov[[g]]
    tmp <- utils::combn(ov_names, 2)
    cor_names <- paste(tmp[1, ], "~~", tmp[2, ], sep = "")
    names_1 <- c(th_names, cor_names)
    if (g > 1L) {
      names_1 <- paste(names_1, ".g", g, sep = "")
    }

    par_idx <- match(parlabel, names_1)
    drhodpsi_mat_1[[g]] <- delta_g[par_idx, , drop = FALSE]
  }
  drhodpsi_mat <- do.call(rbind, drhodpsi_mat_1)

  # tmp_prod <- ( t(drhodpsi_mat) %*% H_at_vartheta0 %*%
  #                 drhodpsi_mat %*% InvG_attheta0 %*%
  #                 H_attheta0 %*% InvG_attheta0 )
  tmp_prod <- (t(drhodpsi_mat) %*% h_at_vartheta0 %*%
    drhodpsi_mat %*% h0_inv %*% j0 %*% g0_inv)
  cov_tzztww <- 2 * sum(diag(tmp_prod))

  ##### Section 4: compute the adjusted PLRT and its p-value
  plrth0sat <- 2 * (h0_fx - sat_fx)
  plrth0sat_group <- 2 * (h0_fx_group - sat_fx_group)
  asym_mean_plrth0sat <- e_tzz - e_tww
  asym_var_plrth0sat <- var_tzz + var_tww - 2 * cov_tzztww
  scaling_factor <- (asym_mean_plrth0sat / (asym_var_plrth0sat / 2))
  fsa_plrt_sem <- (asym_mean_plrth0sat / (asym_var_plrth0sat / 2)) * plrth0sat
  adjusted_df <- (asym_mean_plrth0sat * asym_mean_plrth0sat) /
    (asym_var_plrth0sat / 2)
  # In some very few cases (simulations show very few cases
  #                         in small sample sizes)
  # the adjusted_df is a negative number, we should then
  # print a warning like: "The adjusted df is computed to be a negative number
  # and for this the first and second moment adjusted PLRT is not computed." .
  pvalue <- 1 - pchisq(fsa_plrt_sem, df = adjusted_df)

  list(
    PLRTH0Sat = plrth0sat, PLRTH0Sat.group = plrth0sat_group,
    stat = fsa_plrt_sem, df = adjusted_df, p.value = pvalue,
    scaling.factor = scaling_factor
  )
}
############################################################################

# add parameter bounds to the parameter table
# lavoptions$optim.bounds
lav_pt_add_bounds <- function(partable = NULL,
                                    lavh1 = NULL,
                                    lavdata = NULL,
                                    lavsamplestats = NULL,
                                    lavoptions = NULL) {
  # no support (yet) for multilevel
  if (lav_pt_nlevels(partable) > 1L) {
    return(partable)
  }

  # check optim.bounds
  if (is.null(lavoptions$optim.bounds)) {
    # <0.6-6 version
    return(partable)
  } else if (!is.null(lavoptions$samplestats) && !lavoptions$samplestats) {
    # no sample statistics
    return(partable)
  } else {
    if (!is.null(lavoptions$bounds) && lavoptions$bounds == "none") {
      # no bounds needed
      return(partable)
    }

    # no support from effect.coding (for now)
    if (!is.null(lavoptions$effect.coding) &&
      nchar(lavoptions$effect.coding[1L]) > 0L) {
      lav_msg_warn(gettext(
        "automatic bounds not available (yet) if effect.coding is used"
        ))
      return(partable)
    }

    optim_bounds <- lavoptions$optim.bounds

    # check the elements
    if (is.null(optim_bounds$lower)) {
      optim_bounds$lower <- character(0L)
    } else {
      optim_bounds$lower <- as.character(optim_bounds$lower)
    }
    if (is.null(optim_bounds$upper)) {
      optim_bounds$upper <- character(0L)
    } else {
      optim_bounds$upper <- as.character(optim_bounds$upper)
    }

    if (is.null(optim_bounds$min.reliability.marker)) {
      optim_bounds$min.reliability.marker <- 0.0
    } else {
      if (optim_bounds$min.reliability.marker < 0 ||
        optim_bounds$min.reliability.marker > 1.0) {
        lav_msg_stop(gettextf(
          "optim.bounds$min.reliability.marker is out of range: %s",
          optim_bounds$min.reliability.marker
        ))
      }
    }

    if (is.null(optim_bounds$min.var.ov)) {
      optim_bounds$min.var.ov <- -Inf
    }

    if (is.null(optim_bounds$min.var.lv.exo)) {
      optim_bounds$min.var.lv.exo <- 0.0
    }

    if (is.null(optim_bounds$min.var.lv.endo)) {
      optim_bounds$min.var.lv.endo <- 0.0
    }

    if (is.null(optim_bounds$max.r2.lv.endo)) {
      optim_bounds$max.r2.lv.endo <- 1.0
    }

    if (is.null(optim_bounds$lower.factor)) {
      optim_bounds$lower.factor <- rep(1.0, length(optim_bounds$lower))
    } else {
      if (length(optim_bounds$lower.factor) == 1L &&
        is.numeric(optim_bounds$lower.factor)) {
        optim_bounds$lower.factor <- rep(
          optim_bounds$lower.factor,
          length(optim_bounds$lower)
        )
      } else if (length(optim_bounds$lower.factor) !=
        length(optim_bounds$lower)) {
        lav_msg_stop(
          gettext("length(optim.bounds$lower.factor) is not equal to
                  length(optim.bounds$lower)")
        )
      }
    }
    lower_factor <- optim_bounds$lower.factor

    if (is.null(optim_bounds$upper.factor)) {
      optim_bounds$upper.factor <- rep(1.0, length(optim_bounds$upper))
    } else {
      if (length(optim_bounds$upper.factor) == 1L &&
        is.numeric(optim_bounds$upper.factor)) {
        optim_bounds$upper.factor <- rep(
          optim_bounds$upper.factor,
          length(optim_bounds$upper)
        )
      } else if (length(optim_bounds$upper.factor) !=
        length(optim_bounds$upper)) {
        lav_msg_stop(
          gettext("length(optim.bounds$lower.factor) is not equal to
                  length(optim.bounds$upper)")
        )
      }
    }
    upper_factor <- optim_bounds$upper.factor
  }

  # new in 0.6-17: check if we have theta parameterization
  theta_parameterization_flag <- FALSE
  if (any(partable$op == "~*~") && lavoptions$parameterization == "theta") {
    # some fixed-to-1 theta elements?
    ov_scaled <- partable$lhs[partable$op == "~*~"]
    ov_var_idx <- which(partable$op == "~~" &
      partable$lhs %in% ov_scaled &
      partable$free == 0L &
      partable$ustart == 1)
    if (length(ov_var_idx) > 0L) {
      theta_parameterization_flag <- TRUE
      theta_parameterization_names <- partable$lhs[ov_var_idx]
    }
  }

  # shortcut
  rel <- optim_bounds$min.reliability.marker

  # nothing to do
  if (length(optim_bounds$lower) == 0L &&
    length(optim_bounds$upper) == 0L) {
    return(partable)
  } else {
    # we compute ALL bounds, then we select what we need
    # (otherwise, we can not use the 'factor')

    if (!is.null(partable$lower)) {
      lower_user <- partable$lower
    } else {
      partable$lower <- lower_user <- rep(-Inf, length(partable$lhs))
    }
    if (!is.null(partable$upper)) {
      upper_user <- partable$upper
    } else {
      partable$upper <- upper_user <- rep(+Inf, length(partable$lhs))
    }

    # the 'automatic' bounds
    lower_auto <- rep(-Inf, length(partable$lhs))
    upper_auto <- rep(+Inf, length(partable$lhs))
  }

  lavpta <- lav_pt_attributes(partable)

  # check blocks
  if (is.null(partable$block)) {
    partable$block <- rep(1L, length(partable$lhs))
  }
  # block_values <- lav_pt_block_values(partable)

  # check groups
  if (is.null(partable$group)) {
    partable$group <- rep(1L, length(partable$lhs))
  }
  group_values <- lav_pt_group_values(partable)
  ngroups <- length(group_values)

  # compute bounds per group ### TODO: add levels/classes/...
  b <- 0L
  for (g in seq_len(ngroups)) {
    # next block
    b <- b + 1L

    # for this block
    ov_names <- lavpta$vnames$ov[[b]]
    lv_names <- lavpta$vnames$lv[[b]]
    lv_names_x <- lavpta$vnames$lv.x[[b]]
    if (length(lv_names_x) > 0L) {
      lv_names_endo <- lv_names[!lv_names %in% lv_names_x]
    } else {
      lv_names_endo <- lv_names
    }
    lv_marker <- lavpta$vnames$lv.marker[[b]]

    # OV.VAR for this group
    if (lavsamplestats@missing.flag && lavdata@nlevels == 1L) {
      if (!is.null(lavh1$implied$cov[[g]])) {
        ov_var <- diag(lavh1$implied$cov[[g]])
      } else {
        ov_var <- diag(lavsamplestats@missing.h1[[g]]$sigma)
      }
    } else {
      if (lavoptions$conditional.x) {
        ov_var <- diag(lavsamplestats@res.cov[[g]])
      } else {
        ov_var <- diag(lavsamplestats@cov[[g]])
      }
    }

    # new in 0.6-17: increase observed variances for 'scaled' parameters
    # if theta parameterization
    if (theta_parameterization_flag) {
      sc_idx <- match(theta_parameterization_names, ov_names)
      ov_var[sc_idx] <- ov_var[sc_idx] / rel
    }


    # we 'process' the parameters per 'type', so we can choose
    # to apply (or not) upper/lower bounds for each type separately

    ################################
    ## 1. (residual) ov variances ##
    ################################
    par_idx <- which(partable$group == group_values[g] &
      partable$op == "~~" &
      partable$lhs %in% ov_names &
      partable$lhs == partable$rhs)

    if (length(par_idx) > 0L) {
      # lower == 0
      lower_auto[par_idx] <- 0

      # upper == var(ov)
      var_idx <- match(partable$lhs[par_idx], ov_names)
      upper_auto[par_idx] <- ov_var[var_idx]

      # if reliability > 0, adapt marker indicators only
      if (rel > 0) {
        marker_idx <- which(partable$group == group_values[g] &
          partable$op == "~~" &
          partable$lhs %in% lv_marker &
          partable$lhs == partable$rhs)
        marker_var_idx <- match(partable$lhs[marker_idx], ov_names)

        # upper = (1-REL)*OVAR
        upper_auto[marker_idx] <- (1 - rel) * ov_var[marker_var_idx]
      }

      # range
      bound_range <- upper_auto[par_idx] - pmax(lower_auto[par_idx], 0)

      # enlarge lower?
      if ("ov.var" %in% optim_bounds$lower) {
        factor <- lower_factor[which(optim_bounds$lower == "ov.var")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          diff <- abs(new_range - bound_range)
          lower_auto[par_idx] <- lower_auto[par_idx] - diff
        }
      }

      # enlarge upper?
      if ("ov.var" %in% optim_bounds$upper) {
        factor <- upper_factor[which(optim_bounds$upper == "ov.var")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          diff <- abs(new_range - bound_range)
          upper_auto[par_idx] <- upper_auto[par_idx] + diff
        } else if (is.finite(factor) && factor == 1.0) {
          # new in 0.6-20
          # enlarge anyway, but only with 0.5%
          # this is in particular useful for exogenous variances, otherwise,
          # they will always end up on the boundary
          new_range <- bound_range * 1.005
          diff <- abs(new_range - bound_range)
          upper_auto[par_idx] <- upper_auto[par_idx] + diff
        }
      }

      # min.var.ov?
      min_idx <- which(lower_auto[par_idx] < optim_bounds$min.var.ov)
      if (length(min_idx) > 0L) {
        lower_auto[par_idx[min_idx]] <- optim_bounds$min.var.ov
      }

      # requested?
      if ("ov.var" %in% optim_bounds$lower) {
        partable$lower[par_idx] <- lower_auto[par_idx]
      }
      if ("ov.var" %in% optim_bounds$upper) {
        partable$upper[par_idx] <- upper_auto[par_idx]
      }
    } # (res) ov variances

    ################################
    ## 2. (residual) lv variances ##
    ################################

    # first collect lower/upper bounds for TOTAL variances in lv.names
    lv_var_lb <- numeric(length(lv_names))
    lv_var_ub <- numeric(length(lv_names))

    if (lavoptions$std.lv) {
      lv_var_lb <- rep(1.0, length(lv_names))
      lv_var_ub <- rep(1.0, length(lv_names))
    } else {
      for (i in seq_along(lv_names)) {
        # this_lv_name <- lv_names[i]
        this_lv_marker <- lv_marker[i]

        if (nchar(this_lv_marker) > 0L && this_lv_marker %in% ov_names) {
          marker_var <- ov_var[match(this_lv_marker, ov_names)]
          lower <- marker_var - (1 - rel) * marker_var
          lv_var_lb[i] <- max(lower, optim_bounds$min.var.lv.exo)
          # LV.VAR.UB[i] <- marker.var - REL*marker.var
          lv_var_ub[i] <- marker_var

          # new in 0.6-17
          if (theta_parameterization_flag) {
            lv_var_lb[i] <- rel
          }
        } else {
          lv_var_lb[i] <- optim_bounds$min.var.lv.exo
          lv_var_ub[i] <- max(ov_var)
        }
      }
    }

    # use these bounds for the free parameters
    par_idx <- which(partable$group == group_values[g] &
      partable$op == "~~" &
      partable$lhs %in% lv_names &
      partable$lhs == partable$rhs)

    if (length(par_idx) > 0L) {
      # adjust for endogenous lv
      lv_var_lb2 <- lv_var_lb
      endo_idx <- which(lv_names %in% lv_names_endo)
      if (length(endo_idx) > 0L) {
        lv_var_lb2[endo_idx] <- optim_bounds$min.var.lv.endo
        if (optim_bounds$max.r2.lv.endo != 1) {
          lv_var_lb2[endo_idx] <-
            (1 - optim_bounds$max.r2.lv.endo) * lv_var_ub[endo_idx]
        }
      }
      exo_idx <- which(!lv_names %in% lv_names_endo)
      if (length(exo_idx) > 0L && optim_bounds$min.var.lv.exo != 0) {
        lv_var_lb2[exo_idx] <- optim_bounds$min.var.lv.exo
      }

      lower_auto[par_idx] <- lv_var_lb2[match(
        partable$lhs[par_idx],
        lv_names
      )]
      upper_auto[par_idx] <- lv_var_ub[match(
        partable$lhs[par_idx],
        lv_names
      )]

      # new in 0.6-20: if std.lv = TRUE, the latent variances are normally
      # fixed to 1.0; but in some settings (eg multiple groups +
      # group.equal = "loadings") the latent variances in groups 2+ are
      # free parameters. For those free parameters, the [1, 1] bounds are
      # wrong (they would fix the variance to 1.0). Use data-driven bounds
      # instead (std.lv has no marker indicator, so no marker-based bounds).
      if (lavoptions$std.lv) {
        std_free_idx <- which(partable$free[par_idx] > 0L)
        if (length(std_free_idx) > 0L) {
          fidx <- par_idx[std_free_idx]
          lower_auto[fidx] <- optim_bounds$min.var.lv.exo
          upper_auto[fidx] <- max(ov_var)
          # endogenous lv: respect min.var.lv.endo / max.r2.lv.endo
          endo_f <- which(partable$lhs[fidx] %in% lv_names_endo)
          if (length(endo_f) > 0L) {
            lower_auto[fidx[endo_f]] <- optim_bounds$min.var.lv.endo
            if (optim_bounds$max.r2.lv.endo != 1) {
              lower_auto[fidx[endo_f]] <-
                (1 - optim_bounds$max.r2.lv.endo) * upper_auto[fidx[endo_f]]
            }
          }
        }
      }

      # range
      bound_range <- upper_auto[par_idx] - pmax(lower_auto[par_idx], 0)

      # enlarge lower?
      if ("lv.var" %in% optim_bounds$lower) {
        factor <- lower_factor[which(optim_bounds$lower == "lv.var")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          diff <- abs(new_range - bound_range)
          lower_auto[par_idx] <- lower_auto[par_idx] - diff
        }
      }

      # enlarge upper?
      if ("lv.var" %in% optim_bounds$upper) {
        factor <- upper_factor[which(optim_bounds$upper == "lv.var")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          diff <- abs(new_range - bound_range)
          upper_auto[par_idx] <- upper_auto[par_idx] + diff
        }
      }

      # requested?
      if ("lv.var" %in% optim_bounds$lower) {
        partable$lower[par_idx] <- lower_auto[par_idx]
      }
      if ("lv.var" %in% optim_bounds$upper) {
        partable$upper[par_idx] <- upper_auto[par_idx]
      }
    } # lv variances


    #############################################
    ## 3. factor loadings (ov indicators only) ##
    #############################################

    # lambda_p^(u) = sqrt( upper(res.var.indicators_p) /
    #                      lower(var.factor) )

    ov_ind_names <- lavpta$vnames$ov.ind[[b]]
    par_idx <- which(partable$group == group_values[g] &
      partable$op == "=~" &
      partable$lhs %in% lv_names &
      partable$rhs %in% ov_ind_names)

    if (length(par_idx) > 0L) {
      # if negative LV variances are allowed (due to factor > 1)
      # make them equal to zero
      lv_var_lb[lv_var_lb < 0] <- 0.0

      var_all <- ov_var[match(partable$rhs[par_idx], ov_names)]
      tmp <- lv_var_lb[match(partable$lhs[par_idx], lv_names)]
      tmp[is.na(tmp)] <- 0 # just in case...
      lower_auto[par_idx] <- -1 * sqrt(var_all / tmp) # -Inf if tmp==0
      upper_auto[par_idx] <- +1 * sqrt(var_all / tmp) # +Inf if tmp==0

      # if std.lv = TRUE, force 'first' loading to be positive?
      # if(lavoptions$std.lv) {
      #    # get index 'first' indicators
      #    first.idx <- which(!duplicated(partable$lhs[par.idx]))
      #    lower.auto[par.idx][first.idx] <- 0
      # }

      # range
      bound_range <- upper_auto[par_idx] - lower_auto[par_idx]

      # enlarge lower?
      if ("loadings" %in% optim_bounds$lower) {
        factor <- lower_factor[which(optim_bounds$lower == "loadings")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          ok_idx <- is.finite(new_range)
          if (length(ok_idx) > 0L) {
            diff <- abs(new_range[ok_idx] - bound_range[ok_idx])
            lower_auto[par_idx][ok_idx] <-
              lower_auto[par_idx][ok_idx] - diff
          }
        }
      }

      # enlarge upper?
      if ("loadings" %in% optim_bounds$upper) {
        factor <- upper_factor[which(optim_bounds$upper == "loadings")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          ok_idx <- is.finite(new_range)
          if (length(ok_idx) > 0L) {
            diff <- abs(new_range[ok_idx] - bound_range[ok_idx])
            upper_auto[par_idx][ok_idx] <-
              upper_auto[par_idx][ok_idx] + diff
          }
        }
      }



      # requested?
      if ("loadings" %in% optim_bounds$lower) {
        partable$lower[par_idx] <- lower_auto[par_idx]
      }
      if ("loadings" %in% optim_bounds$upper) {
        partable$upper[par_idx] <- upper_auto[par_idx]
      }
    } # lambda


    ####################
    ## 4. covariances ##
    ####################

    # | sqrt(var(x)) sqrt(var(y)) | <= cov(x,y)

    par_idx <- which(partable$group == group_values[g] &
      partable$op == "~~" &
      partable$lhs != partable$rhs)

    if (length(par_idx) > 0L) {
      for (i in seq_along(par_idx)) {
        # this lhs/rhs
        this_lhs <- partable$lhs[par_idx[i]]
        this_rhs <- partable$rhs[par_idx[i]]

        # 2 possibilities:
        # - variances are free parameters
        # - variances are fixed (eg std.lv = TRUE)

        # var idx
        lhs_var_idx <- which(partable$group == group_values[g] &
          partable$op == "~~" &
          partable$lhs == this_lhs &
          partable$lhs == partable$rhs)
        rhs_var_idx <- which(partable$group == group_values[g] &
          partable$op == "~~" &
          partable$lhs == this_rhs &
          partable$lhs == partable$rhs)
        # upper bounds
        lhs_upper <- upper_auto[lhs_var_idx]
        rhs_upper <- upper_auto[rhs_var_idx]

        # compute upper bounds for this cov (assuming >0 vars)
        if (is.finite(lhs_upper) && is.finite(rhs_upper)) {
          upper_cov <- sqrt(lhs_upper) * sqrt(rhs_upper)
          upper_auto[par_idx[i]] <- +1 * upper_cov
          lower_auto[par_idx[i]] <- -1 * upper_cov
        }
      }

      # range
      bound_range <- upper_auto[par_idx] - lower_auto[par_idx]

      # enlarge lower?
      if ("covariances" %in% optim_bounds$lower) {
        factor <-
          lower_factor[which(optim_bounds$lower == "covariances")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          ok_idx <- is.finite(new_range)
          if (length(ok_idx) > 0L) {
            diff <- new_range[ok_idx] - bound_range[ok_idx]
            lower_auto[par_idx][ok_idx] <-
              lower_auto[par_idx][ok_idx] - diff
          }
        }
      }

      # enlarge upper?
      if ("covariances" %in% optim_bounds$upper) {
        factor <-
          upper_factor[which(optim_bounds$upper == "covariances")]
        if (is.finite(factor) && factor != 1.0) {
          new_range <- bound_range * factor
          ok_idx <- is.finite(new_range)
          if (length(ok_idx) > 0L) {
            diff <- new_range[ok_idx] - bound_range[ok_idx]
            upper_auto[par_idx][ok_idx] <-
              upper_auto[par_idx][ok_idx] + diff
          }
        }
      }

      # requested?
      if ("covariances" %in% optim_bounds$lower) {
        partable$lower[par_idx] <- lower_auto[par_idx]
      }
      if ("covariances" %in% optim_bounds$upper) {
        partable$upper[par_idx] <- upper_auto[par_idx]
      }
    } # covariances
  } # g

  # overwrite with lower.user (except -Inf)
  not_inf_idx <- which(lower_user > -Inf)
  if (length(not_inf_idx) > 0L) {
    partable$lower[not_inf_idx] <- lower_user[not_inf_idx]
  }

  # overwrite with upper.user (except +Inf)
  not_inf_idx <- which(upper_user < +Inf)
  if (length(not_inf_idx) > 0L) {
    partable$upper[not_inf_idx] <- upper_user[not_inf_idx]
  }

  # non-free
  non_free_idx <- which(partable$free == 0L)
  if (length(non_free_idx) > 0L && !is.null(partable$ustart)) {
    partable$lower[non_free_idx] <- partable$ustart[non_free_idx]
    partable$upper[non_free_idx] <- partable$ustart[non_free_idx]
  }

  partable
}

# two-level WLS/DWLS/ULS with categorical data: sample 'statistics' and
# their asymptotic covariance matrix, via the stage-wise (univariate +
# bivariate) estimation of the unrestricted two-level model
# (lav_muthen2007)
#
# YR 2026 (two-level WLS, phase 5)
#
# The per-group statistic vector follows the lavaan two-level WLS layout
# (the within and between blocks stacked per group, matching the rows of
# the per-group Delta matrix):
#
#   within block (categorical layout, theta parameterization):
#     1. th: thresholds (ordinal) and -mu entries (continuous, all zero
#        here: the means of shared variables live at the between level)
#     2. variances of the continuous variables (vw)
#     3. within 'correlations' (vech, no diagonal): cw in the theta
#        metric (ordinal latent responses have within variance 1)
#   between block (continuous layout):
#     4. between means: mu (continuous), 0 (ordinal, fixed)
#     5. vech(Sigma_b, diagonal = TRUE): vb / cb
#
# note: the exact layout is aligned with lav_model_wls_est() /
# lav_model_delta() for the two-level categorical model.

lav_samp_wls_2l_cat <- function(lavsamplestats = NULL,
                                lavdata = NULL,
                                lavoptions = NULL) {
  stopifnot(lavdata@nlevels == 2L)
  estimator <- lavoptions$estimator
  ngroups <- lavsamplestats@ngroups
  nlevels <- 2L

  wls_obs <- vector("list", length = ngroups)
  wls_v <- vector("list", length = ngroups)
  wls_vd <- vector("list", length = ngroups)
  nacov <- vector("list", length = ngroups)

  # 'h1' (unrestricted model) implied statistics, in the standard
  # multilevel layout (within/between block per group); consumed by
  # fit measures (srmr), the baseline model (starting values), ...
  if (isTRUE(lavoptions$conditional.x)) {
    # note: we ALSO fill in the (residual, y-only) cov/mean entries:
    # several consumers (starting values, ...) look for
    # lavh1$implied$cov[[b]] regardless of conditional.x
    h1_implied <- list(
      cov = vector("list", ngroups * nlevels),
      mean = vector("list", ngroups * nlevels),
      res.cov = vector("list", ngroups * nlevels),
      res.int = vector("list", ngroups * nlevels),
      res.slopes = vector("list", ngroups * nlevels),
      cov.x = vector("list", ngroups * nlevels),
      mean.x = vector("list", ngroups * nlevels)
    )
  } else {
    h1_implied <- list(
      cov = vector("list", ngroups * nlevels),
      mean = vector("list", ngroups * nlevels)
    )
  }

  conditional_x <- isTRUE(lavoptions$conditional.x)

  for (g in seq_len(ngroups)) {
    lp <- lavdata@Lp[[g]]
    nobs <- lavsamplestats@nobs[[g]]

    if (!conditional_x && length(lavsamplestats@x.idx[[g]]) > 0L) {
      lav_msg_stop(gettext(
        "fixed.x = TRUE with conditional.x = FALSE is not supported (yet)
        for two-level (D)WLS estimation."))
    }

    # y variables per level (excluding exogenous covariates), in the
    # block variable order; the covariates (if any) come from the Lp
    # within.x/between.x index sets ('split' covariates -- appearing at
    # both levels -- have already been moved to the y set).
    # NOTE: all Lp index sets refer to the 'tilde' universe (lp$ov.names
    # = the y variables followed by the x variables), which is also the
    # column order of lavdata@X[[g]] for multilevel data
    ov_names_all <- lp$ov.names
    x_w <- x_b <- NULL
    xw_names <- xb_names <- NULL
    if (conditional_x) {
      y_idx1 <- lp$ov.y.idx[[1]]
      y_idx2 <- lp$ov.y.idx[[2]]
      xw_idx <- lp$within.x.idx[[1]]
      xb_idx <- lp$between.x.idx[[2]]
      if (length(xw_idx) > 0L) {
        x_w <- lavdata@X[[g]][, xw_idx, drop = FALSE]
        xw_names <- ov_names_all[xw_idx]
      }
      if (length(xb_idx) > 0L) {
        x_b <- lavdata@X[[g]][, xb_idx, drop = FALSE]
        xb_names <- ov_names_all[xb_idx]
      }
      # 'split' covariates (exogenous, but appearing at both levels) are
      # not supported (yet): every covariate must be within-only or
      # between-only
      split_x <- setdiff(lavdata@ov.names.x[[g]], c(xw_names, xb_names))
      if (length(split_x) > 0L) {
        lav_msg_stop(gettextf(
          "two-level (D)WLS estimation does not support exogenous
          covariates that appear at both levels (yet): %s. Each covariate
          should be used at a single level only (cfr. within/between
          covariates); alternatively, use fixed.x = FALSE.",
          lav_msg_view(split_x, "none")))
      }
    } else {
      y_idx1 <- lp$ov.idx[[1]]
      y_idx2 <- lp$ov.idx[[2]]
    }
    ov_names_1 <- ov_names_all[y_idx1]
    ov_names_2 <- ov_names_all[y_idx2]
    if (length(setdiff(ov_names_2, ov_names_1)) > 0L) {
      lav_msg_stop(gettext(
        "two-level (D)WLS estimation with categorical data does not
        support between-only variables (yet)."))
    }
    if (!identical(ov_names_2,
                   ov_names_1[ov_names_1 %in% ov_names_2])) {
      lav_msg_stop(gettext(
        "two-level (D)WLS estimation with categorical data requires the
        between-level variables to appear in the same relative order as
        at the within level."))
    }

    # variable info (in the within-block variable order)
    ov_types <- lavdata@ov$type[match(ov_names_1, lavdata@ov$name)]
    between_flag <- ov_names_1 %in% ov_names_2
    if (any(ov_types == "ordered" & !between_flag)) {
      lav_msg_stop(gettext(
        "two-level (D)WLS estimation does not support within-only ORDINAL
        variables (yet)."))
    }

    # stage-wise estimation of the unrestricted two-level model
    m07 <- lav_muthen2007(
      y = lavdata@X[[g]][, y_idx1, drop = FALSE],
      ov_names = ov_names_1,
      ov_types = ov_types,
      cluster_idx = lp$cluster.idx[[2]],
      between_flag = between_flag,
      x_w = x_w, x_b = x_b,
      xw_names = xw_names, xb_names = xb_names,
      ngh = lavoptions$integration.ngh,
      ngh2 = 13L,
      wls_w = TRUE
    )

    # map the stage-wise statistic vector + ACOV to the lavaan layout
    out <- lav_m07_repackage(m07 = m07, ov_types = ov_types,
                             between_flag = between_flag)

    wls_obs[[g]] <- out$wls_obs
    nacov[[g]] <- nobs * out$acov

    w <- lav_samp_wls_v_2l(
      gamma_1 = nacov[[g]], keep = out$free_idx,
      estimator = estimator, nclusters = lp$nclusters[[2]], group = g
    )
    wls_v[[g]] <- w$wls_v
    if (!is.null(w$wls_vd)) {
      wls_vd[[g]] <- w$wls_vd
    }

    # store the stage-wise 'sample' statistics for inspection/start values
    lavsamplestats@th[[g]] <- out$th
    lavsamplestats@th.idx[[g]] <- out$th_idx

    # unrestricted implied statistics (theta metric: the diagonal of
    # SIGMA.W is 1 for the ordinal variables); the between block spans
    # the between-level variables only, and the means of within-only
    # variables live at the within level
    b_idx <- which(between_flag)
    mu_w1 <- numeric(NROW(m07$SIGMA.W))
    mu_w1[!between_flag] <- m07$MU[!between_flag]
    w_pos <- (g - 1) * nlevels + 1L
    b_pos <- (g - 1) * nlevels + 2L
    if (conditional_x) {
      h1_implied$cov[[w_pos]] <- m07$SIGMA.W
      h1_implied$cov[[b_pos]] <-
        m07$SIGMA.B[b_idx, b_idx, drop = FALSE]
      h1_implied$mean[[w_pos]] <- mu_w1
      h1_implied$mean[[b_pos]] <- m07$MU[b_idx]
      h1_implied$res.cov[[w_pos]] <- m07$SIGMA.W
      h1_implied$res.cov[[b_pos]] <-
        m07$SIGMA.B[b_idx, b_idx, drop = FALSE]
      h1_implied$res.int[[w_pos]] <- mu_w1
      h1_implied$res.int[[b_pos]] <- m07$MU[b_idx]
      h1_implied$res.slopes[[w_pos]] <- m07$SLOPES.W
      h1_implied$res.slopes[[b_pos]] <-
        m07$SLOPES.B[b_idx, , drop = FALSE]
      # observed covariate moments (within: observation level; between:
      # cluster level), with the covariate names as dimnames
      if (!is.null(x_w)) {
        n1 <- NROW(x_w)
        cx1 <- cov(x_w) * (n1 - 1) / n1
        dimnames(cx1) <- list(xw_names, xw_names)
        h1_implied$cov.x[[w_pos]] <- cx1
        h1_implied$mean.x[[w_pos]] <- setNames(colMeans(x_w), xw_names)
      } else {
        h1_implied$cov.x[[w_pos]] <- matrix(0, 0L, 0L)
        h1_implied$mean.x[[w_pos]] <- numeric(0L)
      }
      if (!is.null(x_b)) {
        xbj <- x_b[!duplicated(lp$cluster.idx[[2]]), , drop = FALSE]
        nj <- NROW(xbj)
        cx2 <- cov(xbj) * (nj - 1) / nj
        dimnames(cx2) <- list(xb_names, xb_names)
        h1_implied$cov.x[[b_pos]] <- cx2
        h1_implied$mean.x[[b_pos]] <- setNames(colMeans(xbj), xb_names)
      } else {
        h1_implied$cov.x[[b_pos]] <- matrix(0, 0L, 0L)
        h1_implied$mean.x[[b_pos]] <- numeric(0L)
      }
    } else {
      h1_implied$cov[[w_pos]] <- m07$SIGMA.W
      h1_implied$cov[[b_pos]] <-
        m07$SIGMA.B[b_idx, b_idx, drop = FALSE]
      h1_implied$mean[[w_pos]] <- mu_w1
      h1_implied$mean[[b_pos]] <- m07$MU[b_idx]
    }
  }

  lavsamplestats@WLS.obs <- wls_obs
  lavsamplestats@WLS.V <- wls_v
  lavsamplestats@WLS.VD <- wls_vd
  if (!lavsamplestats@NACOV.user) {
    lavsamplestats@NACOV <- nacov
  }

  list(
    lavsamplestats = lavsamplestats,
    lavh1 = list(
      implied = h1_implied,
      logl = list(loglik = as.numeric(NA), loglik.group = as.numeric(NA))
    )
  )
}

# repackage the stage-wise parameter vector (per variable, then per pair)
# and its ACOV into the lavaan two-level WLS layout:
#
#   1. th (per within variable: thresholds for ordinal, fixed 0 for
#      continuous)
#   2. within means/intercepts of the within-only (continuous) variables
#   3. vec(Pi_w): within slopes, column-major over (variable, covariate)
#   4. within variances of the continuous variables
#   5. within covariances (vech, no diagonal)
#   6. between means/intercepts (continuous: mu; ordinal: fixed 0) --
#      between variables only
#   7. vec(Pi_b): between slopes, column-major -- between variables only
#   8. vech(Sigma_b, diagonal = TRUE) -- between variables only
lav_m07_repackage <- function(m07 = NULL, ov_types = NULL,
                              between_flag = NULL) {
  nvar <- length(ov_types)
  ord <- (ov_types == "ordered")
  if (is.null(between_flag)) {
    between_flag <- rep(TRUE, nvar)
  }
  npar_uni <- m07$npar.uni
  uni_end <- cumsum(npar_uni)
  uni_start <- uni_end - npar_uni + 1L
  ntheta <- sum(npar_uni)
  pstar <- nvar * (nvar - 1L) / 2L
  pair_neta <- m07$pair.neta
  if (is.null(pair_neta)) {
    pair_neta <- rep(2L, pstar)
  }
  eta_end <- cumsum(pair_neta)
  eta_start <- eta_end - pair_neta + 1L
  nexo_w <- if (is.null(m07$nexo.w)) 0L else m07$nexo.w
  nexo_b <- if (is.null(m07$nexo.b)) 0L else m07$nexo.b

  # per-variable offsets (within the univariate block) of the 'base'
  # parameters and the slopes; the univariate block layout is
  #   ordinal:              (tau_1..K, vb, wsl..., bsl...)
  #   continuous (between): (mu, vw, vb, wsl..., bsl...)
  #   continuous (within):  (mu, vw, wsl...)
  nth_p <- integer(nvar) # thresholds per variable (0 = continuous)
  vb_off <- integer(nvar) # offset of vb (NA-like 0 if absent)
  wsl_off <- integer(nvar) # offset of the last param before wsl
  for (p in seq_len(nvar)) {
    if (ord[p]) {
      nth_p[p] <- npar_uni[p] - 1L - nexo_w -
        (if (between_flag[p]) nexo_b else 0L)
      vb_off[p] <- nth_p[p] + 1L
      wsl_off[p] <- nth_p[p] + 1L
    } else if (between_flag[p]) {
      vb_off[p] <- 3L
      wsl_off[p] <- 3L
    } else {
      vb_off[p] <- 0L
      wsl_off[p] <- 2L
    }
  }

  # source positions (in the stage-wise vector) of each target entry;
  # 0 = structurally-fixed entry (value stored in fixed_val)
  src <- integer(0L)
  fixed_val <- numeric(0L)
  th_idx <- integer(0L)

  # --- within block ---
  # 1. th: thresholds (ordinal); fixed 0 slots (continuous)
  for (p in seq_len(nvar)) {
    if (ord[p]) {
      src <- c(src, uni_start[p] - 1L + seq_len(nth_p[p]))
      fixed_val <- c(fixed_val, rep(NA_real_, nth_p[p]))
      th_idx <- c(th_idx, rep(p, nth_p[p]))
    } else {
      src <- c(src, 0L)
      fixed_val <- c(fixed_val, 0)
      th_idx <- c(th_idx, 0L)
    }
  }
  # 2. within means of the within-only (continuous) variables
  for (p in seq_len(nvar)) {
    if (!between_flag[p]) {
      src <- c(src, uni_start[p]) # mu
      fixed_val <- c(fixed_val, NA_real_)
    }
  }
  # 3. vec(Pi_w): all variables for covariate 1, then covariate 2, ...
  for (q in seq_len(nexo_w)) {
    for (p in seq_len(nvar)) {
      src <- c(src, uni_start[p] - 1L + wsl_off[p] + q)
      fixed_val <- c(fixed_val, NA_real_)
    }
  }
  # 4. variances of the continuous variables (vw)
  for (p in seq_len(nvar)) {
    if (!ord[p]) {
      src <- c(src, uni_start[p] + 1L) # (mu, vw, ...)
      fixed_val <- c(fixed_val, NA_real_)
    }
  }
  # 5. within covariances/'correlations' (vech, no diagonal): cw
  for (pp in seq_len(pstar)) {
    src <- c(src, ntheta + eta_start[pp]) # cw is the first free param
    fixed_val <- c(fixed_val, NA_real_)
  }

  # --- between block (between variables only) ---
  b_idx <- which(between_flag)
  # 6. between means: mu (continuous); 0 (ordinal, fixed)
  for (p in b_idx) {
    if (ord[p]) {
      src <- c(src, 0L)
      fixed_val <- c(fixed_val, 0)
    } else {
      src <- c(src, uni_start[p]) # mu
      fixed_val <- c(fixed_val, NA_real_)
    }
  }
  # 7. vec(Pi_b) over the between variables
  for (q in seq_len(nexo_b)) {
    for (p in b_idx) {
      src <- c(src, uni_start[p] - 1L + wsl_off[p] + nexo_w + q)
      fixed_val <- c(fixed_val, NA_real_)
    }
  }
  # 8. vech(Sigma_b, diagonal = TRUE): vb (diag) and cb (off-diagonal)
  # in lav_mat_vech (column-major) order over the between variables
  pair_pos <- matrix(0L, nvar, nvar)
  pp <- 0L
  for (j in seq_len(nvar - 1L)) {
    for (i in (j + 1L):nvar) {
      pp <- pp + 1L
      pair_pos[i, j] <- pp
    }
  }
  nb <- length(b_idx)
  for (jj in seq_len(nb)) {
    for (ii in jj:nb) {
      i <- b_idx[ii]; j <- b_idx[jj]
      if (i == j) {
        src <- c(src, uni_start[j] - 1L + vb_off[j])
        fixed_val <- c(fixed_val, NA_real_)
      } else {
        pp <- pair_pos[i, j]
        stopifnot(pair_neta[pp] == 2L)
        src <- c(src, ntheta + eta_start[pp] + 1L) # cb
        fixed_val <- c(fixed_val, NA_real_)
      }
    }
  }

  # assemble
  nstat <- length(src)
  wls_obs <- numeric(nstat)
  free <- (src > 0L)
  wls_obs[free] <- m07$theta[src[free]]
  wls_obs[!free] <- fixed_val[!free]

  acov <- matrix(0, nstat, nstat)
  acov[free, free] <- m07$WLS.W[src[free], src[free]]

  # th vector (thresholds + zero means, as in the single-level layout)
  nth_all <- length(th_idx)
  th <- wls_obs[seq_len(nth_all)]

  list(
    wls_obs = wls_obs, acov = acov,
    free_idx = which(free),
    th = th, th_idx = th_idx
  )
}

# This file will (eventually) contain functions that can be used to
# 'predict' the values of outcome variables (y), given the values of
# input variables (x).

# first version YR 2 Nov 2022

# method = "conditional.mean" is based on the following article:

# Mark de Rooij, Julian D. Karch, Marjolein Fokkema, Zsuzsa Bakk, Bunga Citra
# Pratiwi & Henk Kelderman (2022) SEM-Based Out-of-Sample Predictions,
# StructuralEquation Modeling: A Multidisciplinary Journal
# DOI:10.1080/10705511.2022.2061494

# YR 31 Jan 2023: we always 'force' meanstructure = TRUE (for now)

# internal function: prepare the data and the y/x indices that are needed
# by both lavPredictY() and lavResidualsY()
lav_predict_y_prepare <- function(object,
                                  newdata = NULL,
                                  ynames = lav_object_vnames(object, "ov.y"),
                                  xnames = lav_object_vnames(object, "ov.x")) {

  lavmodel <- object@Model
  lavdata <- object@Data
  lavimplied <- object@implied

  # check meanstructure
  if (!lavmodel@meanstructure) {
    lavimplied$mean <- lapply(object@SampleStats@mean, as.matrix)
  }

  new_data <- NULL
  newdata_bygroup <- FALSE
  group_idx <- seq_len(lavdata@ngroups)

  # need full data set
  if (is.null(newdata)) {
    # use internal copy:
    if (lavdata@data.type != "full") {
      lav_msg_stop(gettext(
        "sample statistics were used for fitting and newdata is empty"
      ))
    } else if (is.null(lavdata@X[[1]])) {
      lav_msg_stop(gettext("no local copy of data; FIXME!"))
    } else {
      data_obs <- lavdata@X
      ov_names <- lavdata@ov.names
    }
    # eXo <- lavdata@eXo
  } else if (is.list(newdata) && !is.data.frame(newdata)) {
    # newdata is provided per group, as a list with one element per group.
    # Elements may be NULL to skip a group. This is useful when the groups
    # have a different set of variables, so that ynames= and xnames= can be
    # specified per group (see GitHub issue #369). Each element is processed
    # as a single-group dataset, using the ov.names of the corresponding
    # group in the fitted model.
    newdata_bygroup <- TRUE
    if (length(newdata) != lavdata@ngroups) {
      lav_msg_stop(gettextf(
        "newdata is a list of length %1$d, but the model has %2$d group(s);
        please provide one data.frame per group (use NULL to skip a group).",
        length(newdata), lavdata@ngroups
      ))
    }
    ov <- lavdata@ov
    ordered_names <- ov$name[ov$type == "ordered"]
    data_obs <- ov_names <- vector("list", lavdata@ngroups)
    for (g in seq_len(lavdata@ngroups)) {
      if (is.null(newdata[[g]])) {
        next # skip this group; data_obs[[g]] stays NULL
      }
      new_data_g <- lav_lavdata(
        data = newdata[[g]],
        group = NULL, # treat as a single group
        ov_names = lavdata@ov.names[[g]],
        ov_names_x = lavdata@ov.names.x[[g]],
        ordered = ordered_names,
        lavoptions = list(
          std.ov = lavdata@std.ov,
          group.label = character(0L),
          missing = "ml.x", # always!
          warn = TRUE
        ),
        allow_single_case = TRUE
      )
      data_obs[[g]] <- new_data_g@X[[1L]]
      ov_names[[g]] <- new_data_g@ov.names[[1L]]
    }
    group_idx <- which(!vapply(data_obs, is.null, logical(1)))
    if (length(group_idx) == 0L) {
      lav_msg_stop(gettext(
        "newdata is a list, but all its elements are NULL; please provide the
        data for at least one group."
      ))
    }
  } else {
    # newdata is given (a single data.frame containing all groups)!

    # create lavData object
    ov <- lavdata@ov
    new_data <- lav_lavdata(
      data = newdata,
      group = lavdata@group,
      ov_names = lavdata@ov.names,
      ov_names_x = lavdata@ov.names.x,
      ordered = ov$name[ov$type == "ordered"],
      lavoptions = list(
        std.ov = lavdata@std.ov,
        group.label = lavdata@group.label,
        missing = "ml.x", # always!
        warn = TRUE
      ),
      allow_single_case = TRUE
    )
    # if ordered, check if number of levels is still the same (new in 0.6-7)
    if (lavmodel@categorical) {
      orig_ordered_idx <- which(lavdata@ov$type == "ordered")
      orig_ordered_lev <- lavdata@ov$nlev[orig_ordered_idx]
      match_new_idx <- match(
        lavdata@ov$name[orig_ordered_idx],
        new_data@ov$name
      )
      new_ordered_lev <- new_data@ov$nlev[match_new_idx]
      if (any(orig_ordered_lev - new_ordered_lev != 0)) {
        lav_msg_stop(gettext(
          "mismatch number of categories for some ordered variables in
          newdata compared to original data."
        ))
      }
    }

    data_obs <- new_data@X
    # eXo <- newData@eXo
    ov_names <- new_data@ov.names
  } # newdata

  # check ynames
  if (length(ynames) == 0L) {
    lav_msg_stop(gettext(
      "please specify the y-variables in the ynames= argument"
    ))
  }

  if (anyDuplicated((ynames))) {
    lav_msg_stop(gettext(
      "ynames contains duplicate variable names"
    ))
  }

  if (!is.list(ynames)) {
    ynames <- rep(list(ynames), lavdata@ngroups)
  }

  # check xnames
  if (length(xnames) == 0L) {
    lav_msg_stop(gettext(
      "please specify the x-variables in the xnames= argument"
    ))
  }

  if (anyDuplicated((xnames))) {
    lav_msg_stop(gettext(
      "ynames contains duplicate variable names"
    ))
  }

  if (!is.list(xnames)) {
    xnames <- rep(list(xnames), lavdata@ngroups)
  }

  # create y.idx and x.idx (only for the groups we will predict for)
  y_idx <- x_idx <- vector("list", lavdata@ngroups)
  for (g in group_idx) {
    # ynames in ov.names for this group?
    missing_idx <- which(!ynames[[g]] %in% ov_names[[g]])
    if (length(missing_idx) > 0L) {
      lav_msg_stop(
        gettext(
          "some variable names in ynames do not appear in the dataset:"
        ),
        lav_msg_view(ynames[[g]][missing_idx], "none")
      )
    } else {
      y_idx[[g]] <- match(ynames[[g]], ov_names[[g]])
    }

    # xnames in ov.names for this group?
    missing_idx <- which(!xnames[[g]] %in% ov_names[[g]])
    if (length(missing_idx) > 0L) {
      lav_msg_stop(
        gettext(
          "some variable names in xnames do not appear in the dataset:"
        ),
        lav_msg_view(xnames[[g]][missing_idx], "none")
      )
    } else {
      x_idx[[g]] <- match(xnames[[g]], ov_names[[g]])
    }
  }

  list(
    lavmodel = lavmodel, lavdata = lavdata, lavimplied = lavimplied,
    new_data = new_data, data_obs = data_obs, ov_names = ov_names,
    ynames = ynames, xnames = xnames, y_idx = y_idx, x_idx = x_idx,
    newdata_bygroup = newdata_bygroup, group_idx = group_idx
  )
}


# internal function: label + (optionally) assemble the per-group output
# matrices (predictions or residuals) into the user-facing result. This is
# shared by lavPredictY() and lavResidualsY().
lav_predict_y_assemble <- function(out, ynames, lavdata,
                                   newdata = NULL, new_data = NULL,
                                   label = TRUE, assemble = TRUE) {
  # label?
  if (label) {
    # column names
    for (g in seq_len(lavdata@ngroups)) {
      colnames(out[[g]]) <- ynames[[g]]
    }

    # group.labels
    if (lavdata@ngroups > 1L) {
      names(out) <- lavdata@group.label
    }
  }

  # lavaan.matrix
  out <- lapply(out, "class<-", c("lavaan.matrix", "matrix"))

  if (lavdata@ngroups == 1L) {
    res <- out[[1L]]
  } else {
    res <- out
  }

  # assemble multiple groups into a single data.frame?
  if (lavdata@ngroups > 1L && assemble) {
    if (!is.null(newdata)) {
      lavdata <- new_data
    }
    data_1 <- matrix(as.numeric(NA),
      nrow = sum(unlist(lavdata@norig)),
      ncol = ncol(out[[1L]])
    ) # assume == per g
    colnames(data_1) <- colnames(out[[1L]])
    for (g in seq_len(lavdata@ngroups)) {
      data_1[lavdata@case.idx[[g]], ] <- out[[g]]
    }
    data_1 <- as.data.frame(data_1, stringsAsFactors = FALSE)

    if (!is.null(newdata)) {
      data_1[, lavdata@group] <- newdata[, lavdata@group]
    } else {
      # add group
      data_1[, lavdata@group] <- rep(as.character(NA), nrow(data_1))
      if (lavdata@missing == "listwise") {
        # we will lose the group label of omitted variables!
        data_1[unlist(lavdata@case.idx), lavdata@group] <-
          rep(lavdata@group.label, unlist(lavdata@nobs))
      } else {
        data_1[unlist(lavdata@case.idx), lavdata@group] <-
          rep(lavdata@group.label, unlist(lavdata@norig))
      }
    }

    res <- data_1
  }

  res
}


# internal function: label the per-group output matrices when newdata was
# provided as a per-group list (see lav_predict_y_prepare()). Only the
# groups in group_idx are returned; there is no single-data.frame layout to
# reassemble into, so a (labeled) list is returned -- or a single matrix if
# only one group was requested.
lav_predict_y_assemble_bygroup <- function(out, ynames, lavdata, group_idx,
                                           label = TRUE) {
  active <- out[group_idx]
  if (label) {
    for (i in seq_along(group_idx)) {
      colnames(active[[i]]) <- ynames[[group_idx[i]]]
    }
    if (lavdata@ngroups > 1L) {
      names(active) <- lavdata@group.label[group_idx]
    }
  }
  active <- lapply(active, "class<-", c("lavaan.matrix", "matrix"))
  if (length(active) == 1L) {
    active[[1L]]
  } else {
    active
  }
}


# main function
lavPredictY <- function(object,                                    # nolint
                        newdata = NULL,
                        ynames = lav_object_vnames(object, "ov.y"),
                        xnames = lav_object_vnames(object, "ov.x"),
                        method = "conditional.mean",
                        label = TRUE,
                        assemble = TRUE,
                        force_zero_mean = FALSE,
                        lambda = 0,
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  stopifnot(inherits(object, "lavaan"))
  # check object
  object <- lav_object_check_version(object)

  # prepare data + y/x indices
  prep <- lav_predict_y_prepare(
    object = object, newdata = newdata,
    ynames = ynames, xnames = xnames
  )

  # prediction method
  method <- tolower(method)
  if (method == "conditional.mean") {
    out <- lav_predict_y_conditional_mean(
      lavobject = NULL,
      lavmodel = prep$lavmodel, lavdata = prep$lavdata,
      lavimplied = prep$lavimplied,
      data_obs = prep$data_obs, y_idx = prep$y_idx, x_idx = prep$x_idx,
      force_zero_mean = force_zero_mean,
      lambda = lambda
    )
  } else {
    lav_msg_stop(gettext("method must be \"conditional.mean\" (for now)."))
  }

  # label + assemble
  if (prep$newdata_bygroup) {
    lav_predict_y_assemble_bygroup(
      out = out, ynames = prep$ynames, lavdata = prep$lavdata,
      group_idx = prep$group_idx, label = label
    )
  } else {
    lav_predict_y_assemble(
      out = out, ynames = prep$ynames, lavdata = prep$lavdata,
      newdata = newdata, new_data = prep$new_data,
      label = label, assemble = assemble
    )
  }
}


# compute residuals (observed - predicted) for the y-variables, given the
# x-variables. See lavPredictY() for the prediction part. First version:
# YR 2 July 2026 (feature request in GitHub issue #269).
lavResidualsY <- function(object,                                  # nolint
                          newdata = NULL,
                          ynames = lav_object_vnames(object, "ov.y"),
                          xnames = lav_object_vnames(object, "ov.x"),
                          method = "conditional.mean",
                          label = TRUE,
                          assemble = TRUE,
                          force_zero_mean = FALSE,
                          lambda = 0,
                        ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  stopifnot(inherits(object, "lavaan"))
  # check object
  object <- lav_object_check_version(object)

  # prepare data + y/x indices (identical to lavPredictY)
  prep <- lav_predict_y_prepare(
    object = object, newdata = newdata,
    ynames = ynames, xnames = xnames
  )

  # prediction method
  method <- tolower(method)
  if (method == "conditional.mean") {
    ypred <- lav_predict_y_conditional_mean(
      lavobject = NULL,
      lavmodel = prep$lavmodel, lavdata = prep$lavdata,
      lavimplied = prep$lavimplied,
      data_obs = prep$data_obs, y_idx = prep$y_idx, x_idx = prep$x_idx,
      force_zero_mean = force_zero_mean,
      lambda = lambda
    )
  } else {
    lav_msg_stop(gettext("method must be \"conditional.mean\" (for now)."))
  }

  # residuals = observed - predicted (only for the groups we predicted for)
  out <- vector("list", length = prep$lavdata@ngroups)
  for (g in prep$group_idx) {
    yobs_g <- prep$data_obs[[g]][, prep$y_idx[[g]], drop = FALSE]
    out[[g]] <- yobs_g - ypred[[g]]
  }

  # label + assemble
  if (prep$newdata_bygroup) {
    lav_predict_y_assemble_bygroup(
      out = out, ynames = prep$ynames, lavdata = prep$lavdata,
      group_idx = prep$group_idx, label = label
    )
  } else {
    lav_predict_y_assemble(
      out = out, ynames = prep$ynames, lavdata = prep$lavdata,
      newdata = newdata, new_data = prep$new_data,
      label = label, assemble = assemble
    )
  }
}


# method = "conditional.mean"
lav_predict_y_conditional_mean <- function(
    lavobject = NULL, # for convenience
    # object ingredients
    lavmodel = NULL,
    lavdata = NULL,
    lavimplied = NULL,
    # new data
    data_obs = NULL,
    # y and x
    y_idx = NULL,
    x_idx = NULL,
    # options
    force_zero_mean = FALSE,
    lambda = lambda,
    level = 1L) { # not used for now

  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    # lavsamplestats <- lavobject@SampleStats
    lavimplied <- lavobject@implied
  } else {
    stopifnot(
      !is.null(lavmodel), !is.null(lavdata),
      # !is.null(lavsamplestats),
      !is.null(lavimplied)
    )
  }

  # data.obs?
  if (is.null(data_obs)) {
    data_obs <- lavdata@X
  }

  # checks
  if (lavmodel@categorical) {
    lav_msg_stop(gettext("no support for categorical data (yet)."))
  }
  if (lavdata@nlevels > 1L) {
    lav_msg_stop(gettext("no support for multilevel data (yet)."))
  }

  # conditional.x?
  if (lavmodel@conditional.x) {
    sigma_hat <- lav_model_cond2joint_sigma(lavmodel)
    if (lavmodel@meanstructure) {
      mu_hat <- lav_model_cond2joint_mu(lavmodel)
    }
  } else {
    sigma_hat <- lavimplied$cov
    mu_hat <- lavimplied$mean
  }

  # output container
  ypred <- vector("list", length = lavdata@ngroups)

  # run over all groups
  for (g in 1:lavdata@ngroups) {
    # multiple levels?
    if (lavdata@nlevels > 1L) {
      # TODO!
      lav_msg_stop(gettext("no support for multilevel data (yet)!"))
    } else {
      data_obs_g <- data_obs[[g]]

      # no (new)data for this group? (e.g., per-group newdata list with a
      # NULL entry); leave ypred[[g]] as NULL and move on
      if (is.null(data_obs_g)) {
        next
      }

      # model-implied variance-covariance matrix for this group
      cov_g <- sigma_hat[[g]]

      # model-implied mean vector for this group
      if (force_zero_mean) {
        mean_g <- rep(0, ncol(data_obs_g))
      } else {
        mean_g <- as.numeric(mu_hat[[g]])
      }

      # indices (in ov.names)
      y_idx_g <- y_idx[[g]]
      x_idx_g <- x_idx[[g]]

      # partition y/x
      sxx <- cov_g[x_idx_g, x_idx_g, drop = FALSE]
      sxy <- cov_g[x_idx_g, y_idx_g, drop = FALSE]

      # x-data only
      xtest <- data_obs_g[, x_idx_g, drop = FALSE]

      # mx/my
      mx <- mean_g[x_idx_g]
      my <- mean_g[y_idx_g]

      # center using mx
      xtest <- t(t(xtest) - mx)

      # Apply regularization
      sxx <- sxx + lambda * diag(nrow(sxx))

      # prediction rule
      tmp <- xtest %*% solve(sxx, sxy)
      ypred[[g]] <- t(t(tmp) + my)
    } # single level
  } # g

  ypred
}


# Takes a sequence of lambdas and performs k-fold cross-validation to determine
# the best lambda
lavPredictY_cv <- function(                    # nolint
    object,
    data = NULL,
    xnames = lav_object_vnames(object, "ov.x"),
    ynames = lav_object_vnames(object, "ov.y"),
    n_folds = 10L,
    lambda_seq = seq(0, 1, 0.1),
  ...) {

  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # object should be (or inherit from) a lavaan object
  stopifnot(inherits(object, "lavaan"))
  # check object
  object <- lav_object_check_version(object)

  # results container
  results <- matrix(as.numeric(NA),
    nrow = length(lambda_seq) * n_folds,
    ncol = 2L
  )
  colnames(results) <- c("mse", "lambda")

  # shuffle folds
  folds <- sample(rep(1:n_folds, length.out = nrow(data)))

  # extract Y-data
  y <- as.matrix(data[, ynames, drop = FALSE])

  j <- 0L
  for (i in 1:n_folds) {
    indis <- which(folds == i)
    fold_fit <- try(update(object,
      data = data[-indis, , drop = FALSE],
      warn = FALSE
    ), silent = TRUE)
    if (inherits(fold_fit, "try-error")) {
      lav_msg_warn(gettextf("failed fit in fold %s", i))
      next
    }
    for (l in lambda_seq) {
      j <- j + 1L
      yhat <- lavPredictY(
        fold_fit,
        newdata = data[indis, , drop = FALSE],
        xnames = xnames,
        ynames = ynames,
        lambda = l
      )
      y_error <- y[indis, , drop = FALSE] - yhat
      mse <- mean(y_error * y_error)
      results[j, ] <- c(mse, l)
    }
  }

  # Group by lambda and determine average MSE per group
  avg <- aggregate(results[, "mse"],
    by = list(results[, "lambda"]),
    FUN = mean, na.rm = TRUE
  )
  avg <- avg[order(avg[, 2]), ]
  names(avg) <- c("lambda", "mse")
  lambda_min <- avg[1L, "lambda"]

  list(results = avg, lambda.min = lambda_min)
}

lav_step09_model <- function(slot_model = NULL,
                                    lavoptions = NULL,
                                    lavpartable = NULL,
                                    lavsamplestats = NULL,
                                    lavdata = NULL) {
  # # # # # # # # # # #
  # #  9. lavmodel # #
  # # # # # # # # # # #

  # if slot_model not NULL
  #   copy to lavmodel
  # else
  #   compute lavmodel via lav_model
  #   if lavdata@data.type == "none" and categorical mode
  #     set parameters in lavmodel via lav_model_set_parameters and
  #       re-adjust start column in lavpartable
  #     if differences between start and ustart column (in lavpartable)
  #       if lavmodel$parameterization == "delta"
  #         if user specified delta values : ** warning **
  #       if lavmodel$parameterization == "theta"
  #         if user specified theta values : ** warning **

  if (!is.null(slot_model)) {
    lavmodel <- slot_model
  } else {
    if (lav_verbose()) {
      cat("lavmodel           ...")
    }
    lavmodel <- lav_model(
      lavpartable = lavpartable,
      lavoptions = lavoptions,
      th_idx = lavsamplestats@th.idx
    )
    # no longer needed: x values are in start
    # cov.x            = lavsamplestats@cov.x,
    # mean.x           = lavsamplestats@mean.x)

    # if no data, call lav_model_set_parameters once (for categorical case)
    if (lavdata@data.type == "none" && lavmodel@categorical) {
      lavmodel <- lav_model_set_parameters(
        lavmodel = lavmodel,
        x = lav_model_get_parameters(lavmodel)
      )
      # re-adjust parameter table
      lavpartable$start <- lav_model_get_parameters(lavmodel, type = "user")

      # check/warn if theta/delta values make sense
      if (!all(lavpartable$start == lavpartable$ustart)) {
        if (lavmodel@parameterization == "delta") {
          # did the user specify theta values?
          user_var_idx <- which(lavpartable$op == "~~" &
            lavpartable$lhs == lavpartable$rhs &
            lavpartable$lhs %in% unlist(attr(lavpartable, "vnames")$ov_ord) &
            lavpartable$user == 1L)
          if (length(user_var_idx)) {
            lav_msg_warn(
              gettextf("variance (theta) values for categorical variables
                       are ignored if parameterization = %s!",
                       "'delta'")
            )
          }
        } else if (lavmodel@parameterization == "theta") {
          # did the user specify theta values?
          user_delta_idx <- which(lavpartable$op == "~*~" &
            lavpartable$lhs == lavpartable$rhs &
            lavpartable$lhs %in% unlist(attr(lavpartable, "vnames")$ov_ord) &
            lavpartable$user == 1L)
          if (length(user_delta_idx)) {
            lav_msg_warn(
              gettextf("scaling (~*~) values for categorical variables
                       are ignored if parameterization = %s!",
                       "'theta'")
            )
          }
        }
      }

    # same for composites: call lav_model_set_parameters once to set
    # total/residual variances of composites in PSI
    } else if (lavmodel@composites) {
      lavmodel <- lav_model_set_parameters(
        lavmodel = lavmodel,
        x = lav_model_get_parameters(lavmodel)
      )
      # re-adjust parameter table
      lavpartable$start <- lav_model_get_parameters(lavmodel, type = "user")
    }
    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  # if parameterization = "delta" and categorical/correlation: check if
  # we have an observed mediator (new in 0.6-19)
  # but fixed (for recursive models) in 0.6-20
#   check.flag <- FALSE
#   if (!is.null(lavoptions$check.delta.cat.mediator) && # lavaan >= 0.6.19
#       lavoptions$check.delta.cat.mediator) {
#     check.flag <- TRUE
#   }
#   if (check.flag &&
#       (lavmodel@categorical || lavmodel@correlation) &&
#       lavmodel@representation == "LISREL" &&
#       lavmodel@parameterization == "delta") {
#     # get idx BETA matrices
#   beta.idx   <- which(names(lavmodel@GLIST) == "beta")
#   # for every block
#   for (i in seq_len(length(beta.idx))) {
#     this.beta <- abs(lavmodel@GLIST[[beta.idx[i]]])
#       this.beta[lavmodel@m.free.idx[[beta.idx[i]]]] <- 1.0
#     # exogenous variables:       have zero rows
#     # endogenous variables:      have non-zero rows
#     # endogenous-only variables: have non-zero rows and zero columns
#     # mediators:                 have non-zero rows and non-zero columns
#     m.idx <- which(apply(this.beta, 1, sum) != 0 &
#                      apply(this.beta, 2, sum) != 0)
#     if (length(m.idx) > 0L) {
#       # we have (at least) one mediator
#       # is one of them an observed variable? -> warn
#         m.names <- lavmodel@dimNames[[beta.idx[i]]][[1]][m.idx]
#     ov.names   <- unlist(lavdata@ov.names)
#         ov.ordered <- unlist(lavdata@ordered)
#         # correlation model
#     if (lavmodel@correlation && any(m.names %in% ov.names)) {
#       bad.names <- m.names[m.names %in% ov.names]
#       if (length(beta.idx) == 1L) {
#             lav_msg_warn(gettextf("model contains at least one observed
#                 mediator: [%s]; consider switching to parameterization =
#                 \"theta\"",
#             paste(bad.names, collapse = " ")))
#       } else {
#             lav_msg_warn(gettextf("model contains at least one observed
#             mediator in block %i: [%s]; consider switching to parameterization
#             = \"theta\"", i,
#             paste(bad.names, collapse = " ")))
#       }
#         # categorical mode
#     } else if (lavmodel@categorical && any(m.names %in% ov.ordered)) {
#           bad.names <- m.names[m.names %in% ov.ordered]
#           if (length(beta.idx) == 1L) {
#             lav_msg_warn(gettextf("model contains at least one observed
#             categorical mediator: [%s]; consider switching to parameterization
#             = \"theta\"",
#             paste(bad.names, collapse = " ")))
#           } else {
#             lav_msg_warn(gettextf("model contains at least one observed
#                     categorical mediator in block %i: [%s]; consider
#                     switching to parameterization = \"theta\"", i,
#             paste(bad.names, collapse = " ")))
#           }
#         }
#     } # mediators
#   } # block
#   } # delta parameterization

  list(
    lavpartable = lavpartable,
    lavmodel    = lavmodel
  )
}

# functions related to the RMSEA index of approximate fit

# lower-level functions: no checking of input: just compute the number(s):
# - lav_fit_rmsea
# - lav_fit_rmsea_ci
# - lav_fit_rmsea_closefit
# - lav_fit_rmsea_notclosefit (TODO)

# higher-level functions:
# - lav_fit_rmsea_lavobject

# Y.R. 19 July 2022

# we assume X2 = N * F.val
# lambda = (X2 - df) is the non-centrality parameter

# RMSEA: sqrt( (X2 - df)/(N * df) )
#      = sqrt( lambda/(N * df) )
#      = sqrt( ((N*F.val) - df)/(N * df) )
#      = sqrt( (N.F.val)/(N * df) - df/(N * df) )
#      = sqrt( F.val/df - 1/N )
#      = sqrt( (X2/N)/df - 1/N )


# 'scaled' RMSEA: X2 is replaced by X2-SB (or any other 'scaled' statistic)

# robust RMSEA: sqrt( (X2/N)/df - c.hat/N )
# note:
# - robust RMSEA == scaled RMSEA * sqrt(c.hat)
# - robust RMSEA CI == scaled RMSEA CI * sqrt(c.hat)

# robust RMSEA for MLMV (ie. scaled.shifted):
# - c == a * (df - b) / df
# - robust RMSEA MLM == robust RMSEA MLMV
# - robust RMSEA MLMV == scaled RMSEA MLMV * sqrt(a)
# - robust RMSEA CI MLMV == scaled RMSEA CI MLMV * sqrt(a)

# References:
# Steiger, J. H., & Lind, J. C. (1980, May). Statistically based tests for the
# number of common factors.  Paper presented at the annual meeting of the
# Psychometric Society, Iowa City, IA.

# confidence interval:
# Browne, M. W., & Cudeck, R. (1993). Alternative ways of assessing model fit.
# In K. A. Bollen & J. S. Long (Eds.), Testing structural equation models (pp.
# 136-162). Newbury Park, CA: Sage.

# problems with low df
# Kenny, D. A., Kaniskan, B., & McCoach, D. B. (2015).  The performance of RMSEA
# in models with small degrees of freedom.  Sociological Methods & Research, 44,
# 486-507.

# robust version MLM
# Patricia E. Brosseau-Liard , Victoria Savalei & Libo Li (2012) An
# Investigation of the Sample Performance of Two Nonnormality Corrections for
# RMSEA, Multivariate Behavioral Research, 47:6, 904-930, DOI:
# 10.1080/00273171.2012.715252

# robust version MLMV (scaled.shifted)
# Savalei, V. (2018). On the computation of the RMSEA and CFI from the
# mean-and-variance corrected test statistic with nonnormal data in SEM.
# Multivariate behavioral research, 53(3), 419-429.

# categorical data:
# Savalei, V. (2021). Improving fit indices in structural equation modeling with
# categorical data. Multivariate Behavioral Research, 56(3), 390-407. doi:
# 10.1080/00273171.2020.1717922

# missing = "fiml":
# Zhang, X., & Savalei, V. (2022). New computations for RMSEA and CFI following
# FIML and TS estimation with missing data. Psychological Methods.


# always using N (if a user needs N-1, just replace N by N-1)
# vectorized!
lav_fit_rmsea <- function(x2 = NULL, df = NULL, n = NULL,
                          f_val = NULL, g = 1L, c_hat = 1.0) {
  # did we get a sample size?
  if (missing(n) && !missing(f_val)) {
    # population version
    rmsea <- sqrt(f_val / df)
  } else {
    nel <- length(x2)
    if (nel == 0) {
      return(as.numeric(NA))
    }
    rmsea <- ifelse(df > 0,
      # 'standard' way to compute RMSEA
      rmsea <- sqrt(pmax((x2 / n) / df - c_hat / n, rep(0, nel))) * sqrt(g),
      0
    ) # if df == 0
  }

  rmsea
}


# note: for 'robust' version, X2 should be SB-X2
lav_fit_rmsea_ci <- function(x2 = NULL, df = NULL, n = NULL,
                             g = 1L, c_hat = 1, level = 0.90) {
  if (missing(n) || missing(x2) || missing(df) ||
    !is.finite(x2) || !is.finite(df) || !is.finite(n)) {
    return(list(
      rmsea.ci.lower = as.numeric(NA),
      rmsea.ci.upper = as.numeric(NA)
    ))
  }

  if (!is.finite(level) || level < 0 || level > 1.0) {
    lav_msg_warn(gettextf(
      "invalid level value [%s] set to default 0.90.", level
    ))
    level <- 0.90
  }

  upper_perc <- (1 - (1 - level) / 2)
  lower_perc <- (1 - level) / 2

  # internal function
  lower_lambda <- function(lambda) {
    (pchisq(x2, df = df, ncp = lambda) - upper_perc)
  }

  upper_lambda <- function(lambda) {
    (pchisq(x2, df = df, ncp = lambda) - lower_perc)
  }


  # lower bound
  if (df < 1 || lower_lambda(0) < 0.0) {
    rmsea_ci_lower <- 0
  } else {
    lambda_l <- try(uniroot(f = lower_lambda, lower = 0, upper = x2)$root,
      silent = TRUE
    )
    if (inherits(lambda_l, "try-error")) {
      lambda_l <- as.numeric(NA)
    }
    # lower bound
    rmsea_ci_lower <- sqrt((c_hat * lambda_l) / (n * df))

    # multiple groups? -> correction
    if (g > 1L) {
      rmsea_ci_lower <- rmsea_ci_lower * sqrt(g)
    }
  }

  # upper bound
  n_rmsea <- max(n, x2 * 4)
  if (df < 1 || upper_lambda(n_rmsea) > 0 || upper_lambda(0) < 0) {
    rmsea_ci_upper <- 0
  } else {
    lambda_u <- try(
      uniroot(
        f = upper_lambda, lower = 0,
        upper = n_rmsea
      )$root,
      silent = TRUE
    )
    if (inherits(lambda_u, "try-error")) {
      lambda_u <- NA
    }
    # upper bound
    rmsea_ci_upper <- sqrt((c_hat * lambda_u) / (n * df))

    # multiple groups? -> correction
    if (g > 1L) {
      rmsea_ci_upper <- rmsea_ci_upper * sqrt(g)
    }
  }

  list(
    rmsea.ci.lower = rmsea_ci_lower,
    rmsea.ci.upper = rmsea_ci_upper
  )
}

# H_0: RMSEA <= rmsea.h0
lav_fit_rmsea_closefit <- function(x2 = NULL, df = NULL, n = NULL,
                                   g = 1L, c_hat = 1, rmsea_h0 = 0.05) {
  if (missing(n) || missing(x2) || missing(df) ||
    !is.finite(x2) || !is.finite(df) || !is.finite(n)) {
    return(as.numeric(NA))
  }

  rmsea_pvalue <- as.numeric(NA)
  if (df > 0) {
    # see Dudgeon 2004, eq 16 for the 'G' correction
    ncp <- (n * df * 1 / c_hat * rmsea_h0^2) / g
    rmsea_pvalue <- 1 - pchisq(x2, df = df, ncp = ncp)
  }

  rmsea_pvalue
}

# MacCallum, R. C., Browne, M. W., & Sugawara, H. M. (1996).
# H_0: RMSEA >= rmsea.h0
lav_fit_rmsea_notclosefit <- function(x2 = NULL, df = NULL, n = NULL,
                                      g = 1L, c_hat = 1, rmsea_h0 = 0.05) {
  if (missing(n) || missing(x2) || missing(df) ||
    !is.finite(x2) || !is.finite(df) || !is.finite(n)) {
    return(as.numeric(NA))
  }

  rmsea_pvalue <- as.numeric(NA)
  if (df > 0) {
    # see Dudgeon 2004, eq 16 for the 'G' correction
    ncp <- (n * df * 1 / c_hat * rmsea_h0^2) / g
    rmsea_pvalue <- pchisq(x2, df = df, ncp = ncp)
  }

  rmsea_pvalue
}


lav_fit_rmsea_lavobject <- function(lavobject = NULL, fit_measures = "rmsea",
                                    standard_test = "standard",
                                    scaled_test = "none",
                                    ci_level = 0.90,
                                    close_h0 = 0.05, notclose_h0 = 0.08,
                                    robust = TRUE,
                                    cat_nonpd = "na") {
  # check lavobject
  stopifnot(inherits(lavobject, "lavaan"))

  # check for categorical
  categorical_flag <- lavobject@Model@categorical

  # tests
  test <- lavobject@test
  test_names <- sapply(lavobject@test, "[[", "test")
  if (test_names[1] == "none" || standard_test == "none") {
    return(list())
  }
  test_idx <- which(test_names == standard_test)[1]
  if (length(test_idx) == 0L) {
    return(list())
  }

  scaled_flag <- FALSE
  if (!scaled_test %in% c("none", "standard", "default")) {
    scaled_idx <- which(test_names == scaled_test)
    if (length(scaled_idx) > 0L) {
      scaled_idx <- scaled_idx[1] # only the first one
      scaled_flag <- TRUE
    }
  }

  # robust?
  robust_flag <- FALSE
  if (robust && scaled_flag &&
    scaled_test %in% c(
      "satorra.bentler", "yuan.bentler.mplus",
      "yuan.bentler", "yuan.chan", "scaled.shifted",
      "scaled.shifted.corrected"
    )) {
    robust_flag <- TRUE
  }

  # FIML? (the FIML-corrected robust indices are only available for
  # single-level models; for multilevel data, we fall back to the regular
  # scaled/robust computation above, which handles missing data too)
  fiml_flag <- FALSE
  if (robust && lavobject@Options$missing %in% c("ml", "ml.x") &&
    lavobject@Data@nlevels == 1L) {
    fiml_flag <- robust_flag <- TRUE
    # check if we can compute corrected values
    if (scaled_flag) {
      version <- "V3"
    } else {
      version <- "V6"
    }
    fiml <- try(
      lav_fit_fiml_corrected(lavobject,
        baseline_model = NULL,
        version = version
      ),
      silent = TRUE
    )
    if (inherits(fiml, "try-error")) {
      lav_msg_warn(gettext("computation of robust RMSEA failed."))
      fiml <- list(
        XX3 = as.numeric(NA), df3 = as.numeric(NA),
        c.hat3 = as.numeric(NA), XX3.scaled = as.numeric(NA)
      )
    } else if (anyNA(c(fiml$XX3, fiml$df3, fiml$c.hat3, fiml$XX3.scaled))) {
      lav_msg_warn(gettext(
        "computation of robust RMSEA resulted in NA values."
      ))
    }
  }

  # supported fit measures in this function
  fit_rmsea <- c(
    "rmsea",
    "rmsea.ci.lower", "rmsea.ci.upper", "rmsea.ci.level",
    "rmsea.pvalue", "rmsea.close.h0",
    "rmsea.notclose.pvalue", "rmsea.notclose.h0"
  )
  if (scaled_flag) {
    fit_rmsea <- c(
      fit_rmsea, "rmsea.scaled", "rmsea.ci.lower.scaled",
      "rmsea.ci.upper.scaled", "rmsea.pvalue.scaled",
      "rmsea.notclose.pvalue.scaled"
    )
  }
  if (robust_flag) {
    fit_rmsea <- c(
      fit_rmsea, "rmsea.robust", "rmsea.ci.lower.robust",
      "rmsea.ci.upper.robust", "rmsea.pvalue.robust",
      "rmsea.notclose.pvalue.robust"
    )
  }

  # which one do we need?
  if (missing(fit_measures)) {
    # default set
    fit_measures <- fit_rmsea
  } else {
    # remove any not-RMSEA related index from fit.measures
    rm_idx <- which(!fit_measures %in% fit_rmsea)
    if (length(rm_idx) > 0L) {
      fit_measures <- fit_measures[-rm_idx]
    }
    if (length(fit_measures) == 0L) {
      return(list())
    }
  }


  # basic test statistics
  x2 <- test[[test_idx]]$stat
  df <- test[[test_idx]]$df
  g <- lavobject@Data@ngroups # number of groups
  n <- lav_inspect_ntotal(object = lavobject) # N vs N-1

  # scaled X2/df values
  #
  # Two reference distributions are in use here, and they differ:
  #
  #  - lavaan (default): the UNSCALED statistic is referred to a chi-square
  #    with tr[U Gamma] degrees of freedom. X2 is asymptotically a mixture
  #    sum(lambda_i * chisq_1), and this matches its MEAN (E[X2] = tr[U Gamma]).
  #
  #  - Mplus: the SCALED statistic is referred to a chi-square with the scaled
  #    test's own df. This matches the mean on the other scale
  #    (E[X2/c] = df, since c = tr[U Gamma]/df).
  #
  # Both are mean-matching approximations; they differ in variance, so they
  # give different confidence intervals and p-values. lavaan already uses the
  # Mplus convention for the scaled.shifted family (there is no meaningful
  # tr[U Gamma] route once the statistic has been shifted), which is why
  # WLSMV/MLMV agree with Mplus while WLSM/MLM/MLMVS/MLR do not.
  #
  # rmsea.scaled.mplus = TRUE (set by mimic = "Mplus") applies the Mplus
  # convention throughout. See Mplus techappen.pdf, Appendix 5, eq (125) with
  # (106): "with categorical outcomes, Mplus replaces d in (125) by tr[U Gamma]"
  # -- which, since tr[U Gamma] = c * d, is the same as referring the scaled
  # statistic to d.
  if (scaled_flag) {
    if (scaled_test %in% c("scaled.shifted", "scaled.shifted.corrected") ||
      isTRUE(lavobject@Options$rmsea.scaled.mplus)) {
      xx2 <- test[[scaled_idx]]$stat
      # the scaled test carries its own df (equal to df for satorra.bentler
      # and the shifted family; a Satterthwaite value for mean.var.adjusted)
      df2 <- test[[scaled_idx]]$df
      if (is.null(df2) || !is.finite(df2) || df2 == 0) {
        df2 <- df
      }
    } else {
      xx2 <- x2
      df2 <- sum(test[[scaled_idx]]$trace.UGamma)
      if (!is.finite(df2) || df2 == 0) {
        df2 <- as.numeric(NA)
      }
    }
  }

  # robust ingredients
  n3 <- n
  if (robust_flag) {
    if (categorical_flag) {
      out <- try(lav_fit_catml_dwls(lavobject, nonpd = cat_nonpd),
        silent = TRUE
      )
      if (inherits(out, "try-error")) {
        xx3 <- df3 <- c_hat3 <- xx3_scaled <- as.numeric(NA)
      } else {
        xx3 <- out$XX3
        df3 <- out$df3
        c_hat3 <- c_hat <- out$c.hat3
        xx3_scaled <- out$XX3.scaled
      }
      # new in 0.7-1: the robust quantities (xx3, c.hat3, ...) are all
      # ML-based (chi-square = N * F_ML), so we use N -- not N-1 -- in
      # the denominator; this also matches what a direct
      # estimator = "catML" run would use
      n3 <- lavobject@SampleStats@ntotal
    } else if (fiml_flag) {
      xx3 <- fiml$XX3
      df3 <- fiml$df3
      c_hat3 <- c_hat <- fiml$c.hat3
      xx3_scaled <- fiml$XX3.scaled
    } else {
      xx3 <- x2
      df3 <- df
      c_hat <- test[[scaled_idx]]$scaling.factor
      if (scaled_test %in% c(
        "scaled.shifted", "scaled.shifted.corrected"
      )) {
        # compute c.hat from a and b
        a <- test[[scaled_idx]]$scaling.factor
        b <- test[[scaled_idx]]$shift.parameter
        c_hat3 <- a * (df - b) / df
      } else {
        c_hat3 <- c_hat
      }
      xx3_scaled <- test[[scaled_idx]]$stat
    }
  }

  # output container
  indices <- list()

  # what do we need?
  rmsea_val_flag <- rmsea_ci_flag <- rmsea_pvalue_flag <- FALSE
  if (any(c("rmsea", "rmsea.scaled", "rmsea.robust") %in% fit_measures)) {
    rmsea_val_flag <- TRUE
  }
  if (any(c(
    "rmsea.ci.lower", "rmsea.ci.upper", "rmsea.ci.level",
    "rmsea.ci.lower.scaled", "rmsea.ci.upper.scaled",
    "rmsea.ci.lower.robust", "rmsea.ci.upper.robust"
  )
  %in% fit_measures)) {
    rmsea_ci_flag <- TRUE
  }
  if (any(c(
    "rmsea.pvalue", "rmsea.pvalue.scaled", "rmsea.pvalue.robust",
    "rmsea.notclose.pvalue", "rmsea.notclose.pvalue.scaled",
    "rmsea.notclose.pvalue.robust",
    "rmsea.close.h0", "rmsea.notclose.h0"
  )
  %in% fit_measures)) {
    rmsea_pvalue_flag <- TRUE
  }


  # 1. RMSEA
  if (rmsea_val_flag) {
    indices["rmsea"] <- lav_fit_rmsea(x2 = x2, df = df, n = n, g = g)
    if (scaled_flag) {
      indices["rmsea.scaled"] <- lav_fit_rmsea(
        x2 = xx2, df = df2,
        n = n, g = g
      )
    }
    if (robust_flag) {
      indices["rmsea.robust"] <-
        lav_fit_rmsea(x2 = xx3, df = df3, n = n3, c_hat = c_hat3, g = g)
    }
  }

  # 2. RMSEA CI
  if (rmsea_ci_flag) {
    indices["rmsea.ci.level"] <- ci_level
    ci <- lav_fit_rmsea_ci(x2 = x2, df = df, n = n, g = g, level = ci_level)
    indices["rmsea.ci.lower"] <- ci$rmsea.ci.lower
    indices["rmsea.ci.upper"] <- ci$rmsea.ci.upper

    if (scaled_flag) {
      ci_scaled <- lav_fit_rmsea_ci(
        x2 = xx2, df = df2, n = n, g = g,
        level = ci_level
      )
      indices["rmsea.ci.lower.scaled"] <- ci_scaled$rmsea.ci.lower
      indices["rmsea.ci.upper.scaled"] <- ci_scaled$rmsea.ci.upper
    }
    if (robust_flag) {
      # note: input is scaled test statistic!
      ci_robust <- lav_fit_rmsea_ci(
        x2 = xx3_scaled,
        df = df3, n = n3, g = g, c_hat = c_hat,
        level = ci_level
      )
      indices["rmsea.ci.lower.robust"] <- ci_robust$rmsea.ci.lower
      indices["rmsea.ci.upper.robust"] <- ci_robust$rmsea.ci.upper
    }
  }

  # 3. RMSEA pvalue
  if (rmsea_pvalue_flag) {
    indices["rmsea.close.h0"] <- close_h0
    indices["rmsea.notclose.h0"] <- notclose_h0
    indices["rmsea.pvalue"] <-
      lav_fit_rmsea_closefit(
        x2 = x2, df = df, n = n, g = g,
        rmsea_h0 = close_h0
      )
    indices["rmsea.notclose.pvalue"] <-
      lav_fit_rmsea_notclosefit(
        x2 = x2, df = df, n = n, g = g,
        rmsea_h0 = notclose_h0
      )
    if (scaled_flag) {
      indices["rmsea.pvalue.scaled"] <-
        lav_fit_rmsea_closefit(
          x2 = xx2, df = df2, n = n, g = g,
          rmsea_h0 = close_h0
        )
      indices["rmsea.notclose.pvalue.scaled"] <-
        lav_fit_rmsea_notclosefit(
          x2 = xx2, df = df2, n = n, g = g,
          rmsea_h0 = notclose_h0
        )
    }
    if (robust_flag) {
      indices["rmsea.pvalue.robust"] <- # new in 0.6-13
        lav_fit_rmsea_closefit(
          x2 = xx3_scaled,
          df = df3,
          n = n3, g = g, c_hat = c_hat,
          rmsea_h0 = close_h0
        )
      indices["rmsea.notclose.pvalue.robust"] <- # new in 0.6-13
        lav_fit_rmsea_notclosefit(
          x2 = xx3_scaled,
          df = df3,
          n = n3, g = g, c_hat = c_hat,
          rmsea_h0 = notclose_h0
        )
    }
  }

  # return only those that were requested
  indices[fit_measures]
}

lav_points_beziers <- function(x,
  y = NULL,
  col = par("col"),
  lwd = par("lwd")
) {
  if (is.null(y)) {
    if (dim(x)[1L] == 2) {
      p_x <- x[1L, ]
      p_y <- x[2L, ]
    } else {
      p_x <- x[, 1L]
      p_y <- x[, 2L]
    }
  } else {
    p_x <- x
    p_y <- y
  }
  stopifnot(length(p_x) == length(p_y))
  t <- seq(0, 1, length.out = 50)
  if (length(p_x) == 3L) {
    punten_x <- (1 - t)^2 * p_x[1] + 2 * (1 - t) * t * p_x[2] + t^2 * p_x[3]
    punten_y <- (1 - t)^2 * p_y[1] + 2 * (1 - t) * t * p_y[2] + t^2 * p_y[3]
  } else {
    punten_x <- (1 - t)^3 * p_x[1] + 3 * t * (1 - t)^2 * p_x[2] +
      3 * t^2 * (1 - t) * p_x[3] + t^3 * p_x[4]
    punten_y <- (1 - t)^3 * p_y[1] + 3 * t * (1 - t)^2 * p_y[2] +
      3 * t^2 * (1 - t) * p_y[3] + t^3 * p_y[4]
  }
  lines(punten_x, punten_y, col = col, lwd = lwd)
}

lav_plotinfo_rgraph <- function(plotinfo,
                                sloped_labels = TRUE,
                                outfile = "",
                                addgrid = TRUE,
                                mlovcolors = c("lightgreen", "lightblue"),
                                lightness = 1,
                                italic = TRUE,
                                auto_subscript = TRUE) {
  font <- ifelse(italic, 3L, 1L)
  node_elements <- function(nodetiepe, noderadius) {
    # define form, color and anchors for a node type
    thetas <- switch(nodetiepe,
      lv = ,
      varlv = seq(0, 2 * pi, length.out = 50L),
      ov = seq(pi / 4, 2 * pi, by = pi / 2),
      wov = ,
      bov = c(
        seq(pi / 4 - pi / 10, pi / 4 + pi / 10, by = pi / 60),
        seq(3 * pi / 4 - pi / 10, 3 * pi / 4 + pi / 10, by = pi / 60),
        seq(5 * pi / 4 - pi / 10, 5 * pi / 4 + pi / 10, by = pi / 60),
        seq(7 * pi / 4 - pi / 10, 7 * pi / 4 + pi / 10, by = pi / 60)
      ),
      cv = seq(0, 2 * pi, by = pi / 3),
      const = seq(pi / 2, 2 * pi, by = 2 * pi / 3)
    )
    localradius <- noderadius
    if (nodetiepe == "varlv") localradius <- noderadius * .8
    drawx <- localradius * cos(thetas)
    drawy <- localradius * sin(thetas)
    wovbovflat <- max(drawx)
    boxcol <- switch(nodetiepe,
      lv = NA_integer_,
      varlv = NA_integer_,
      ov = NA_integer_,
      wov = mlovcolors[1L],
      bov = mlovcolors[2L],
      cv = NA_integer_,
      const = NA_integer_
    )
    n <- c(0, switch(nodetiepe,
      lv = ,
      varlv = ,
      const = localradius,
      ov = localradius * sqrt(2) / 2,
      wov = ,
      bov = wovbovflat,
      cv = localradius * sqrt(3) / 2
    ))
    s <- c(0, switch(nodetiepe,
      lv = ,
      varlv = -localradius,
      ov = -localradius * sqrt(2) / 2,
      wov = ,
      bov = -wovbovflat,
      cv = -localradius * sqrt(3) / 2,
      const = -localradius * 0.5
    ))
    e <- switch(nodetiepe,
      lv = ,
      varlv = ,
      cv = c(localradius, 0),
      ov = c(localradius * sqrt(2) / 2, 0),
      wov = ,
      bov = c(wovbovflat, 0),
      const = c(localradius * sqrt(3) / 2, -localradius * 0.5)
    )
    w <- c(-e[1L], e[2L])
    ne <- switch(nodetiepe,
      lv = ,
      varlv = ,
      ov = ,
      wov = ,
      bov = localradius * sqrt(0.5) * c(1, 1),
      cv = localradius * c(0.5, sqrt(3) / 2),
      const = e
    )
    nw <- c(-ne[1L], ne[2L])
    se <- switch(nodetiepe,
      lv = ,
      varlv = ,
      ov = ,
      wov = ,
      bov = localradius * sqrt(0.5) * c(1, -1),
      cv = localradius * c(0.5, -sqrt(3) / 2),
      const = e
    )
    sw <- c(-se[1L], se[2L])
    list(
      drawx = drawx, drawy = drawy, boxcol = boxcol, n = n, ne = ne, e = e,
      se = se, s = s, sw = sw, w = w, nw = nw
    )
  }

  vecrotate <- function(vec, angle) {
    c(
      cos(angle) * vec[1] + sin(angle) * vec[2],
      -sin(angle) * vec[1] + cos(angle) * vec[2]
    )
  }
  plot_arrow <- function(tip, dirvec) {
    unitvec <- dirvec / sqrt(sum(dirvec^2))
    arrowangle <- pi * 25 / 180
    arrowinset <- 0.4
    args <- rbind(
      tip,
      tip + vecrotate(-unitvec * arrowlength, arrowangle),
      tip - unitvec * arrowlength * (1 - arrowinset),
      tip + vecrotate(-unitvec * arrowlength, -arrowangle)
    )
    polygon(args, col = "black", border = NA)
  }
  plot_edge <- function(van, naar, label = "", dubbel = FALSE,
                        control = NA_real_, below = FALSE, txtcex = 0.9) {
    labele <- lav_label_code(label, auto_subscript = auto_subscript)$r
    dirvec <- naar - van
    theta <- atan2(naar[2] - van[2], naar[1] - van[1])
    srt <- ifelse(sloped_labels, 180 * theta / pi, 0)
    if (srt > 90) srt <- srt - 180
    if (srt < -90) srt <- srt + 180
    if (is.na(control[1L])) {
      args <- rbind(van, naar)
      lines(args, lwd = 2)
      plot_arrow(naar, dirvec)
      if (dubbel) plot_arrow(van, -dirvec)
      midden <- (van + naar) * 0.5
    } else {
      # gebogen lijn (quadratic lav_points_beziers)
      lav_points_beziers(rbind(van, control, naar), lwd = 2)
      midden <- (van + naar) / 4 + control / 2
      plot_arrow(naar, naar - control)
      if (dubbel) plot_arrow(van, van - control)
    }
    if (label != "") {
      if (below) {
        if (theta >= 0 && theta < 90) {
          text(midden[1L], midden[2L], labele,
            adj = c(0, 1),
            srt = srt, cex = txtcex, font = font
          )
        } else if (theta >= 90) {
          text(midden[1L], midden[2L], labele,
            adj = c(1, 1),
            srt = srt, cex = txtcex, font = font
          )
        } else if (theta < -90) {
          text(midden[1L], midden[2L], labele,
            adj = c(0, 1),
            srt = srt, cex = txtcex, font = font
          )
        } else {
          text(midden[1L], midden[2L], labele,
            adj = c(1, 1),
            srt = srt, cex = txtcex, font = font
          )
        }
      } else {
        if (theta >= 0 && theta < 90) {
          text(midden[1L], midden[2L], labele,
            adj = c(1, 0),
            srt = srt, cex = txtcex, font = font
          )
        } else if (theta >= 90) {
          text(midden[1L], midden[2L], labele,
            adj = c(0, 0),
            srt = srt, cex = txtcex, font = font
          )
        } else if (theta < -90) {
          text(midden[1L], midden[2L], labele,
            adj = c(1, 0),
            srt = srt, cex = txtcex, font = font
          )
        } else {
          text(midden[1L], midden[2L], labele,
            adj = c(0, 0),
            srt = srt, cex = txtcex, font = font
          )
        }
      }
    }
  }
  plot_var <- function(waar, noderadius, label = "", side = "n", txtcex = 0.9) {
    labele <- lav_label_code(label, auto_subscript = auto_subscript)$r
    thetarange <- c(pi / 6, 11 * pi / 6)
    if (side == "s") thetarange <- thetarange + pi / 2
    if (side == "e") thetarange <- thetarange + pi
    if (side == "n") thetarange <- thetarange + 3 * pi / 2
    localradius <- noderadius * 0.8
    middelpt <- switch(side,
      n = c(0, localradius),
      w = c(-localradius, 0),
      s = c(0, -localradius),
      e = c(localradius, 0)
    )
    middelpt <- middelpt + waar
    # cirkelsegment
    thetas <- seq(thetarange[1L], thetarange[2L], length.out = 40)
    straal <- localradius
    xs <- middelpt[1] + cos(thetas) * straal
    ys <- middelpt[2] + sin(thetas) * straal
    lines(xs, ys)
    # pijlen
    plot_arrow(c(xs[1], ys[1]), c(sin(thetarange[1]), -cos(thetarange[1])))
    plot_arrow(c(xs[40], ys[40]), c(-sin(thetarange[2]), cos(thetarange[2])))
    # label
    if (label != "") {
      text(middelpt[1L], middelpt[2L], labele,
        adj = 0.5, cex = txtcex * 0.8,
        font = font
      )
    }
  }
  plot_node <- function(waar, tiepe, label = "", txtcex = 0.9) {
    labele <- lav_label_code(label, auto_subscript = auto_subscript)$r
    elems <- node_elements(tiepe, noderadius)
    x <- waar[1] + elems$drawx
    y <- waar[2] + elems$drawy
    polygon(x, y, col = elems$boxcol, lwd = 1)
    text(waar[1L], waar[2L], labele, adj = 0.5, cex = txtcex, font = font)
  }
  mlrij <- plotinfo$mlrij
  if (is.null(mlrij)) {
    lav_msg_stop(gettext(
      "plotinfo hasn't been processed by lav_plotinfo_positions!"
    ))
  }
  nodes <- plotinfo$nodes
  edges <- plotinfo$edges
  noderadius <- 0.3
  arrowlength <- noderadius / 3
  rijen <- max(nodes$rij)
  kolommen <- max(nodes$kolom)
  if (outfile != "") png(outfile, 960, 960, "px")
  opar <- par(mar = c(1L, 1L, 1L, 1L) + 0.1)
  plot.default(
    x = c(0, lightness * kolommen + 1),
    c(0, lightness * rijen + 1), type = "n",
    xlab = "", ylab = "", axes = FALSE, asp = 1
  )
  if (addgrid) {
    abline(
      v = seq.int(kolommen) * lightness,
      h = seq.int(rijen) * lightness,
      lwd = 1,
      lty = "dotted", col = "lightgray"
    )
    text(seq.int(kolommen) * lightness, 0.3,
      labels = seq.int(kolommen), adj = 1, cex = 0.7
    )
    text(0.3, seq.int(rijen) * lightness,
      labels = seq.int(rijen, 1), adj = 1, cex = 0.7
    )
  }
  if (mlrij > 0L) abline(h = rijen - mlrij + 1, lwd = 2)
  for (j in seq.int(nrow(edges))) {
    if (edges$naar[j] != edges$van[j]) {
      van <- which(nodes$id == edges$van[j])
      naar <- which(nodes$id == edges$naar[j])
      adrvan <- lightness *
        c(nodes$kolom[van], rijen - nodes$rij[van] + 1)
      elems <- node_elements(nodes$tiepe[van], noderadius)
      adrvan <- adrvan + elems[[edges$vananker[j]]]
      adrnaar <- lightness *
        c(nodes$kolom[naar], rijen - nodes$rij[naar] + 1)
      elems <- node_elements(nodes$tiepe[naar], noderadius)
      adrnaar <- adrnaar + elems[[edges$naaranker[j]]]
      if (is.na(edges$controlpt.rij[j])) {
        plot_edge(adrvan, adrnaar, edges$label[j],
          dubbel = (edges$tiepe[j] == "~~"),
          below = edges$labelbelow[j]
        )
      } else {
        controlpt <- lightness * c(
          edges$controlpt.kol[j],
          rijen - edges$controlpt.rij[j] + 1
        )
        plot_edge(adrvan, adrnaar, edges$label[j],
          dubbel = (edges$tiepe[j] == "~~"),
          below = edges$labelbelow[j],
          control = controlpt
        )
      }
    } else {
      van <- which(nodes$id == edges$van[j])
      adrvan <- lightness * c(nodes$kolom[van], rijen - nodes$rij[van] + 1)
      elems <- node_elements(nodes$tiepe[van], noderadius)
      adrvan <- adrvan + elems[[edges$vananker[j]]]
      plot_var(adrvan, noderadius, edges$label[j], edges$vananker[j])
    }
  }
  for (j in seq.int(nrow(nodes))) {
    plot_node(
      lightness *
        c(nodes$kolom[j], rijen - nodes$rij[j] + 1),
      nodes$tiepe[j],
      nodes$naam[j]
    )
  }
  par(opar)
  if (outfile != "") dev.off()
  invisible(NULL)
}

# the multivariate linear model + missing (y) values
#  ('FIML' for the conditional.x case)
#
# the model is y_i | x_i ~ N(Beta' X1_i, res.cov), where X1_i = c(1, x_i);
# the x (exogenous) values are assumed to be complete; the y values may
# contain missing values, organized in missing patterns; each case
# contributes the loglikelihood of its observed y values only:
#   log f(y_obs,i | x_i) with mean (Beta' X1_i)[obs] and
#   covariance res.cov[obs, obs]

# 0) extended pattern statistics (sufficient statistics per pattern)
# 1) loglikelihood (from raw data, or pattern statistics)
# 2) derivatives with respect to Beta, res.cov, vech(res.cov)
# 3) casewise scores with respect to Beta + vech(res.cov)
# 4) hessian of Beta + vech(res.cov)
# 5) (unit) information of Beta + vech(res.cov)
#    5a: (unit)    expected information
#    5b: (unit)    observed information
#    5c: (unit) first.order information

# the parameter order is always: vec(Beta) | vech(res.cov), where
# vec(Beta) is the column-major vec of the (1+nexo) x nvar matrix Beta,
# i.e. per y variable: intercept, slope_1, slope_2, ...; this matches
# both lav_mvreg_sc_beta() and the conditional.x DELTA rows

# YR/CC 13 Jul 2026: first version

# 0. extended per-pattern sufficient statistics
#
#    because the conditional mean is linear in x, the crossproducts
#    WXX = sum_i X1_i X1_i', WXY = sum_i X1_i y_obs,i' and
#    WYY = sum_i y_obs,i y_obs,i' (sums over the cases within a pattern,
#    weighted if wt is given) are sufficient for the loglikelihood, the
#    gradient, the hessian and the expected information; only the
#    casewise scores need the raw data
#
#    the result extends the lav_samp_mi_patterns() output: each element
#    also contains the usual SY, MY, var.idx and freq fields
lav_samp_mi_patterns_res <- function(y = NULL,
                                     exo = NULL, # no intercept
                                     mp = NULL,
                                     wt = NULL) {
  y <- as.matrix(unname(y))
  x1 <- cbind(1, as.matrix(unname(exo)))

  # missing patterns (y side only; x is complete)
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y, sort_freq = FALSE, coverage = FALSE)
  }

  # the usual pattern statistics (SY, MY, var.idx, freq)
  yp <- lav_samp_mi_patterns(y = y, mp = mp, wt = wt)

  # add the crossproduct statistics
  for (p in seq_len(mp$npatterns)) {
    case_idx <- mp$case.idx[[p]]
    var_idx <- mp$pat[p, ]

    x1_p <- x1[case_idx, , drop = FALSE]
    y_p <- y[case_idx, var_idx, drop = FALSE]

    if (!is.null(wt)) {
      wx1_p <- x1_p * wt[case_idx]
      yp[[p]]$WXX <- crossprod(wx1_p, x1_p)
      yp[[p]]$WXY <- crossprod(wx1_p, y_p)
      yp[[p]]$WYY <- crossprod(y_p * wt[case_idx], y_p)
    } else {
      yp[[p]]$WXX <- crossprod(x1_p)
      yp[[p]]$WXY <- crossprod(x1_p, y_p)
      yp[[p]]$WYY <- crossprod(y_p)
    }
  }

  yp
}

# small helper: per-pattern beta kernel and total residual crossprod
#
#   C.p      = WXY - WXX %*% Beta[, obs]         (the beta score kernel)
#   W.tilde  = sum_i r_i r_i'  (total, NOT divided by freq), where
#              r_i = y_obs,i - Beta[, obs]' X1_i
lav_mvreg_mi_pat_kernel <- function(yp_1 = NULL, beta_1 = NULL) {
  b_o <- beta_1[, yp_1$var.idx, drop = FALSE]
  xyb <- yp_1$WXY - yp_1$WXX %*% b_o
  w_tilde <- (yp_1$WYY - crossprod(yp_1$WXY, b_o) - crossprod(b_o, yp_1$WXY)
              + crossprod(b_o, yp_1$WXX %*% b_o))
  list(c_p = xyb, w_tilde = w_tilde)
}

# 1. loglikelihood

# 1a: input is raw data (casewise = TRUE returns casewise loglikelihoods;
#     rows of empty cases remain NA)
lav_mvreg_mi_loglik_data <- function(y = NULL,
                                     exo = NULL, # no intercept
                                     mp = NULL,
                                     wt = NULL,
                                     beta_1 = NULL,
                                     res_int = NULL,
                                     res_slopes = NULL,
                                     res_cov = NULL,
                                     casewise = FALSE,
                                     sinv_method = "eigen",
                                     res_cov_inv = NULL,
                                     log2pi = TRUE,
                                     minus_two = FALSE) {
  y <- as.matrix(unname(y))
  x1 <- cbind(1, as.matrix(unname(exo)))
  ny <- NROW(y)
  log_2pi <- log(2 * pi)

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y, sort_freq = FALSE, coverage = FALSE)
  }

  # global inverse + logdet
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method, logdet = TRUE
  )
  res_cov_logdet <- attr(res_cov_inv, "logdet")

  # residuals (NAs propagate)
  res <- y - x1 %*% beta_1

  # DIST/logdet per case
  dist_1 <- logdet <- p_log_2pi <- rep(as.numeric(NA), ny)

  for (p in seq_len(mp$npatterns)) {
    # observed values for this pattern
    var_idx <- mp$pat[p, ]

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # cases with this pattern
    case_idx <- mp$case.idx[[p]]

    # constant
    p_log_2pi[case_idx] <- sum(var_idx) * log_2pi

    # invert res.cov for this pattern
    sigma_inv <- lav_mvn_mi_pattern_inv(
      sigma_inv = res_cov_inv, na_idx = na_idx,
      logdet = TRUE, sigma_logdet = res_cov_logdet
    )
    logdet[case_idx] <- attr(sigma_inv, "logdet")

    if (mp$freq[p] == 1L) {
      dist_1[case_idx] <- sum(sigma_inv *
        crossprod(res[case_idx, var_idx, drop = FALSE]))
    } else {
      dist_1[case_idx] <-
        rowSums(res[case_idx, var_idx, drop = FALSE] %*% sigma_inv *
          res[case_idx, var_idx, drop = FALSE])
    }
  }

  # compute casewise loglikelihoods
  if (log2pi) {
    llik <- -(p_log_2pi + logdet + dist_1) / 2
  } else {
    llik <- -(logdet + dist_1) / 2
  }

  # minus.two
  if (minus_two) {
    llik <- -2 * llik
  }

  # weights?
  if (!is.null(wt)) {
    llik <- llik * wt
  }

  if (casewise) {
    loglik <- llik
  } else {
    loglik <- sum(llik, na.rm = TRUE)
  }

  loglik
}

# 1b: input are the extended pattern statistics (see
#     lav_samp_mi_patterns_res)
lav_mvreg_mi_loglik_samp <- function(yp = NULL,
                                     beta_1 = NULL,
                                     res_int = NULL,
                                     res_slopes = NULL,
                                     res_cov = NULL,
                                     sinv_method = "eigen",
                                     res_cov_inv = NULL,
                                     log2pi = TRUE,
                                     minus_two = FALSE) {
  log_2pi <- log(2 * pi)
  pat_n <- length(yp)

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # global inverse + logdet
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method, logdet = TRUE
  )
  res_cov_logdet <- attr(res_cov_inv, "logdet")

  # DIST/logdet per pattern
  dist_1 <- logdet <- p_log_2pi <- numeric(pat_n)

  for (p in seq_len(pat_n)) {
    # observed variables for this pattern
    var_idx <- yp[[p]]$var.idx

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # constant
    p_log_2pi[p] <- sum(var_idx) * log_2pi * yp[[p]]$freq

    # invert res.cov for this pattern
    sigma_inv <- lav_mvn_mi_pattern_inv(
      sigma_inv = res_cov_inv, na_idx = na_idx,
      logdet = TRUE, sigma_logdet = res_cov_logdet
    )
    logdet[p] <- attr(sigma_inv, "logdet") * yp[[p]]$freq

    # total residual crossprod for this pattern
    kern <- lav_mvreg_mi_pat_kernel(yp_1 = yp[[p]], beta_1 = beta_1)
    dist_1[p] <- sum(sigma_inv * kern$w_tilde)
  }

  # loglikelihood all data
  if (log2pi) {
    loglik <- sum(-(p_log_2pi + logdet + dist_1) / 2)
  } else {
    loglik <- sum(-(logdet + dist_1) / 2)
  }

  if (minus_two) {
    loglik <- -2 * loglik
  }

  loglik
}


# 2. Derivatives

# 2a: derivative logl with respect to Beta (=intercepts and slopes),
#     returned as vec(Beta) (column-major of the (1+nexo) x nvar matrix)
lav_mvreg_mi_dlogl_dbeta_samp <- function(yp = NULL,
                                          beta_1 = NULL,
                                          res_int = NULL,
                                          res_slopes = NULL,
                                          res_cov = NULL,
                                          sinv_method = "eigen",
                                          res_cov_inv = NULL) {
  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)
  p_1 <- NCOL(beta_1)

  # global inverse
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  dbeta <- matrix(0, NROW(beta_1), p_1)

  for (p in seq_len(length(yp))) {
    var_idx <- yp[[p]]$var.idx

    # invert res.cov for this pattern
    sigma_inv <- lav_mvn_mi_pattern_inv(
      sigma_inv = res_cov_inv, na_idx = which(!var_idx)
    )

    kern <- lav_mvreg_mi_pat_kernel(yp_1 = yp[[p]], beta_1 = beta_1)
    dbeta[, var_idx] <- dbeta[, var_idx] + kern$c_p %*% sigma_inv
  }

  as.numeric(dbeta)
}

# 2b: derivative logl with respect to res.cov (full matrix, ignoring
#     symmetry)
lav_mvreg_mi_dlogl_drescov_samp <- function(yp = NULL,
                                            beta_1 = NULL,
                                            res_int = NULL,
                                            res_slopes = NULL,
                                            res_cov = NULL,
                                            sinv_method = "eigen",
                                            res_cov_inv = NULL) {
  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)
  p_1 <- NCOL(beta_1)

  # global inverse
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  dres_cov <- matrix(0, p_1, p_1)

  for (p in seq_len(length(yp))) {
    var_idx <- yp[[p]]$var.idx
    pat_freq <- yp[[p]]$freq

    # invert res.cov for this pattern
    sigma_inv <- lav_mvn_mi_pattern_inv(
      sigma_inv = res_cov_inv, na_idx = which(!var_idx)
    )

    kern <- lav_mvreg_mi_pat_kernel(yp_1 = yp[[p]], beta_1 = beta_1)
    dres_cov[var_idx, var_idx] <- (dres_cov[var_idx, var_idx]
      - (pat_freq / 2) * (sigma_inv -
        (sigma_inv %*% (kern$w_tilde / pat_freq) %*% sigma_inv)))
  }

  dres_cov
}

# 2c: derivative logl with respect to vech(res.cov)
lav_mvreg_mi_dlogl_dvechrescov_samp <- function(yp = NULL,
                                                beta_1 = NULL,
                                                res_int = NULL,
                                                res_slopes = NULL,
                                                res_cov = NULL,
                                                sinv_method = "eigen",
                                                res_cov_inv = NULL) {
  dres_cov <- lav_mvreg_mi_dlogl_drescov_samp(
    yp = yp, beta_1 = beta_1,
    res_int = res_int, res_slopes = res_slopes, res_cov = res_cov,
    sinv_method = sinv_method, res_cov_inv = res_cov_inv
  )

  as.numeric(lav_mat_dup_pre(as.matrix(lav_mat_vec(dres_cov))))
}


# 3. Casewise scores

# 3a: casewise scores with respect to Beta + vech(res.cov)
#     column order: Y1_int, Y1_x1, ... | Y2_int, Y2_x1, ... | vech(res.cov)
#     NAs are left in place (empty cases, and the vech(res.cov) columns
#     that involve an unobserved variable); callers should zero them out
lav_mvreg_mi_sc_beta_sigma <- function(y = NULL,
                                       exo = NULL, # no intercept
                                       mp = NULL,
                                       wt = NULL,
                                       beta_1 = NULL,
                                       res_int = NULL,
                                       res_slopes = NULL,
                                       res_cov = NULL,
                                       sinv_method = "eigen",
                                       res_cov_inv = NULL) {
  y <- as.matrix(unname(y))
  q_1 <- NCOL(y)
  x1 <- cbind(1, as.matrix(unname(exo)))
  p <- NCOL(x1)

  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y, sort_freq = FALSE, coverage = FALSE)
  }

  # global inverse
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # residuals (NAs propagate)
  res <- y - x1 %*% beta_1

  # score engine (res %*% res.cov.inv per pattern + vech(res.cov) scores)
  eng <- lav_mvn_mi_sc_engine(
    y = res, mp = mp, mu = numeric(q_1),
    sigma_inv = res_cov_inv
  )

  sc_beta <- x1[, rep(1:p, times = q_1), drop = FALSE] *
    eng$dmu[, rep(1:q_1, each = p), drop = FALSE]

  out <- cbind(sc_beta, eng$sc)

  # weights
  if (!is.null(wt)) {
    out <- out * wt
  }

  out
}


# 4. hessian of logl

# 4a. hessian logl Beta and vech(res.cov) from the extended pattern
#     statistics
lav_mvreg_mi_logl_hessian_samp <- function(yp = NULL,
                                           beta_1 = NULL,
                                           res_int = NULL,
                                           res_slopes = NULL,
                                           res_cov = NULL,
                                           sinv_method = "eigen",
                                           res_cov_inv = NULL) {
  # construct model-implied Beta
  beta_1 <- lav_mvreg_beta1(beta_1, res_int, res_slopes)
  p_1 <- NCOL(beta_1)
  nx1 <- NROW(beta_1) # 1 + nexo
  pstar <- p_1 * (p_1 + 1) / 2

  # global inverse
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  h11 <- matrix(0, nx1 * p_1, nx1 * p_1)
  h21 <- matrix(0, pstar, nx1 * p_1)
  h22 <- matrix(0, pstar, pstar)

  for (p in seq_len(length(yp))) {
    var_idx <- yp[[p]]$var.idx
    pat_freq <- yp[[p]]$freq

    # invert res.cov for this pattern
    sigma_inv_1 <- lav_mvn_mi_pattern_inv(
      sigma_inv = res_cov_inv, na_idx = which(!var_idx)
    )

    # zero-padded pattern inverse
    s_inv <- matrix(0, p_1, p_1)
    s_inv[var_idx, var_idx] <- sigma_inv_1

    kern <- lav_mvreg_mi_pat_kernel(yp_1 = yp[[p]], beta_1 = beta_1)

    # beta-beta block: vec(Beta) positions of the observed y variables
    obs_idx <- which(var_idx)
    idx_o <- as.vector(outer(seq_len(nx1), (obs_idx - 1L) * nx1, "+"))
    h11[idx_o, idx_o] <- (h11[idx_o, idx_o]
      + (sigma_inv_1 %x% yp[[p]]$WXX))

    # sigma-beta block: zero-padded C.p (per-pattern beta kernel)
    c_bar <- matrix(0, nx1, p_1)
    c_bar[, var_idx] <- kern$c_p
    h21 <- h21 + lav_mat_dup_pre(s_inv %x% (s_inv %*% t(c_bar)))

    # sigma-sigma block
    aaa <- (sigma_inv_1 %*%
      (2 * (kern$w_tilde / pat_freq) -
        res_cov[var_idx, var_idx, drop = FALSE]) %*%
      sigma_inv_1)
    tmp22 <- matrix(0, p_1, p_1)
    tmp22[var_idx, var_idx] <- aaa
    h22 <- h22 + pat_freq * lav_mvn_kron_dup_half(s_inv, tmp22)
  }

  h12 <- t(h21)

  -1 * rbind(
    cbind(h11, h12),
    cbind(h21, h22)
  )
}

# 4b. hessian logl Beta and vech(res.cov) from raw data
lav_mvreg_mi_logl_hessian_data <- function(y = NULL,
                                           exo = NULL, # no intercept
                                           mp = NULL,
                                           wt = NULL,
                                           beta_1 = NULL,
                                           res_int = NULL,
                                           res_slopes = NULL,
                                           res_cov = NULL,
                                           sinv_method = "eigen",
                                           res_cov_inv = NULL) {
  yp <- lav_samp_mi_patterns_res(y = y, exo = exo, mp = mp, wt = wt)

  lav_mvreg_mi_logl_hessian_samp(
    yp = yp, beta_1 = beta_1,
    res_int = res_int, res_slopes = res_slopes, res_cov = res_cov,
    sinv_method = sinv_method, res_cov_inv = res_cov_inv
  )
}


# 5. Information

# 5a: expected unit information Beta and vech(res.cov)
#     (expected given the observed missing patterns; as in the
#     unconditional case, this is only strictly valid under MCAR)
lav_mvreg_mi_info_expected <- function(yp = NULL,
                                       res_cov = NULL,
                                       res_cov_inv = NULL,
                                       sinv_method = "eigen") {
  p_1 <- NCOL(res_cov)
  nx1 <- NROW(yp[[1]]$WXX) # 1 + nexo
  pstar <- p_1 * (p_1 + 1) / 2

  # global inverse
  res_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = res_cov, sigma_inv = res_cov_inv,
    sinv_method = sinv_method
  )

  # N
  n <- sum(sapply(yp, "[[", "freq")) # implicitly: removed empty cases!

  i11 <- matrix(0, nx1 * p_1, nx1 * p_1)
  i22 <- matrix(0, pstar, pstar)

  for (p in seq_len(length(yp))) {
    var_idx <- yp[[p]]$var.idx
    pat_freq <- yp[[p]]$freq

    # invert res.cov for this pattern
    sigma_inv_1 <- lav_mvn_mi_pattern_inv(
      sigma_inv = res_cov_inv, na_idx = which(!var_idx)
    )

    # beta block
    obs_idx <- which(var_idx)
    idx_o <- as.vector(outer(seq_len(nx1), (obs_idx - 1L) * nx1, "+"))
    i11[idx_o, idx_o] <- (i11[idx_o, idx_o]
      + (sigma_inv_1 %x% yp[[p]]$WXX))

    # sigma block
    s_inv <- matrix(0, p_1, p_1)
    s_inv[var_idx, var_idx] <- sigma_inv_1
    i22 <- i22 + pat_freq * lav_mvn_kron_dup_half(s_inv)
  }

  lav_mat_bdiag(i11, i22) / n
}

# 5b: unit observed information Beta and vech(res.cov) from the extended
#     pattern statistics
lav_mvreg_mi_information_observed_samplestats <- # nolint
  function(yp = NULL,
           beta_1 = NULL,
           res_int = NULL,
           res_slopes = NULL,
           res_cov = NULL,
           sinv_method = "eigen",
           res_cov_inv = NULL) {
    n <- sum(sapply(yp, "[[", "freq")) # implicitly: removed empty cases!

    # observed information
    observed <- lav_mvreg_mi_logl_hessian_samp(
      yp = yp, beta_1 = beta_1,
      res_int = res_int, res_slopes = res_slopes, res_cov = res_cov,
      sinv_method = sinv_method, res_cov_inv = res_cov_inv
    )

    -observed / n
  }

# 5c: unit first-order information Beta and vech(res.cov) from raw data
lav_mvreg_mi_info_firstorder <- function(y = NULL,
                                         exo = NULL, # no intercept
                                         mp = NULL,
                                         wt = NULL,
                                         cluster_idx = NULL,
                                         beta_1 = NULL,
                                         res_int = NULL,
                                         res_slopes = NULL,
                                         res_cov = NULL,
                                         sinv_method = "eigen",
                                         res_cov_inv = NULL) {
  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y, sort_freq = FALSE, coverage = FALSE)
  }

  # N
  n <- lav_mvn_mi_nobs(mp = mp, wt = wt)

  sc <- lav_mvreg_mi_sc_beta_sigma(
    y = y, exo = exo, mp = mp, wt = wt,
    beta_1 = beta_1, res_int = res_int, res_slopes = res_slopes,
    res_cov = res_cov, sinv_method = sinv_method,
    res_cov_inv = res_cov_inv
  )

  # handle clustering
  if (!is.null(cluster_idx)) {
    # take the sum within each cluster
    sc <- rowsum(sc, group = cluster_idx, reorder = FALSE, na.rm = TRUE)

    # lower bias if number of clusters is not very high
    n_c <- nrow(sc)
    correction_factor <- n_c / (n_c - 1)
    sc <- sc * sqrt(correction_factor)
  } else {
    sc[is.na(sc)] <- 0
  }

  lav_mat_crossprod(sc) / n
}


# 6. sandwich ACOV of the (saturated) Beta + vech(res.cov) estimates:
#    A^{-1} B A^{-1} (Savalei & Falk, 2014, in the conditional metric),
#    used for se = "robust.two.stage" + conditional.x
lav_mvreg_mi_h1_omega_sw <- function(y = NULL,
                                     exo = NULL, # no intercept
                                     mp = NULL,
                                     wt = NULL,
                                     cluster_idx = NULL,
                                     yp = NULL,
                                     sinv_method = "eigen",
                                     res_int = NULL,
                                     res_slopes = NULL,
                                     res_cov = NULL,
                                     res_cov_inv = NULL,
                                     information = "observed") {
  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # extended pattern statistics
  if (is.null(yp)) {
    yp <- lav_samp_mi_patterns_res(y = y, exo = exo, mp = mp, wt = wt)
  }

  # bread
  if (information == "expected") {
    a <- lav_mvreg_mi_info_expected(
      yp = yp, res_cov = res_cov,
      res_cov_inv = res_cov_inv, sinv_method = sinv_method
    )
  } else {
    a <- lav_mvreg_mi_information_observed_samplestats(
      yp = yp, res_int = res_int, res_slopes = res_slopes,
      res_cov = res_cov, res_cov_inv = res_cov_inv,
      sinv_method = sinv_method
    )
  }
  a_inv <- lav_mat_sym_inverse(
    s = a, logdet = FALSE,
    sinv_method = sinv_method
  )

  # meat
  m_b <- lav_mvreg_mi_info_firstorder(
    y = y, exo = exo, mp = mp, wt = wt,
    cluster_idx = cluster_idx,
    res_int = res_int, res_slopes = res_slopes, res_cov = res_cov,
    res_cov_inv = res_cov_inv, sinv_method = sinv_method
  )

  # sandwich
  a_inv %*% m_b %*% a_inv
}

# Corrected (mean and variance) adjusted test statistics
#
# Hayakawa, K. (2018). Corrected goodness-of-fit test in covariance
# structure analysis. Psychological Methods, 24(3), 371-389.
#
# The mean-and-variance adjusted test statistic (test = "mean.var.adjusted")
# divides the test statistic by tr(UGamma^2)/tr(UGamma), with degrees of
# freedom tr(UGamma)^2/tr(UGamma^2). But the naive (plug-in) estimator of
# tr(UGamma^2) is severely (upward) biased unless p* = p(p+1)/2 is small
# relative to N; as a result, the adjusted test underrejects (often towards
# zero) when the number of variables grows. Hayakawa (2018) replaces the
# naive estimator by the unbiased estimator of a trace of a squared
# covariance matrix of Srivastava (2005) and Himeno & Yamada (2014), which
# remains accurate in high dimensions:
#
#   a2c = [ (N-2)(N-1) tr(H^2) - N(N-1) tr(D^2) + (tr H)^2 ] /
#         [ N(N-1)(N-2)(N-3) ]
#
# where H = sum_i y_i y_i', D = diag(y_1'y_1, ..., y_N'y_N),
# y_i = U^{1/2} (s_i - sbar), and s_i is the casewise (saturated) moment
# vector of observation i.
#
# Note that U^{1/2} is never needed: with Y = [y_1, ..., y_N] and
# Zc the N x p~ matrix with rows (s_i - sbar)', the N x N Gram matrix
# M = Zc U Zc' = Y'Y satisfies tr(H) = tr(M), tr(H^2) = tr(M^2), and
# y_i'y_i = diag(M), so all three ingredients come from M directly.
#
# The corrected tests are exposed as test = "mean.var.adjusted.corrected"
# and (a natural extension, not in the paper) "scaled.shifted.corrected";
# they combine with scaled.test = "browne.residual.nt.model" to give the
# RLS-based versions recommended by Hayakawa (2018).

# eligibility: the corrected trace estimator assumes iid casewise moment
# vectors from a single sample, computed from complete continuous data
lav_test_hayakawa_check <- function(lavoptions = NULL, lavdata = NULL,
                                    lavmodel = NULL) {
  context <- gettext("corrected adjusted tests")

  if (!is.null(lavdata)) {
    if (lavdata@ngroups > 1L) {
      lav_msg_stop(gettextf(
        "%s are only available for single-group models.", context))
    }
    if (lavdata@nlevels > 1L) {
      lav_msg_stop(gettextf(
        "%s are only available for single-level models.", context))
    }
    if (length(lavdata@X) == 0L || is.null(lavdata@X[[1]])) {
      lav_msg_stop(gettextf(
        "%s require the raw data.", context))
    }
  }

  if (!is.null(lavmodel) && lavmodel@categorical) {
    lav_msg_stop(gettextf(
      "%s are only available for continuous data.", context))
  }

  if (!is.null(lavoptions)) {
    if (!isTRUE(lavoptions$estimator == "ML")) {
      lav_msg_stop(gettextf(
        "%1$s require estimator = %2$s; found %3$s.",
        context, dQuote("ML"), dQuote(lavoptions$estimator)))
    }
    if (!isTRUE(lavoptions$missing[1] == "listwise")) {
      lav_msg_stop(gettextf(
        "%1$s require missing = %2$s; found %3$s.",
        context, dQuote("listwise"), dQuote(lavoptions$missing[1])))
    }
    if (isTRUE(lavoptions$conditional.x)) {
      lav_msg_stop(gettextf(
        "%s are not available if conditional.x = TRUE.", context))
    }
  }

  invisible(TRUE)
}

# unbiased estimator of tr((U Omega)^2), Omega = Var(s_i)
# (Srivastava 2005; Himeno & Yamada 2014; Hayakawa 2018, eq. 19)
#
# u:      the U matrix (as in UfromUGamma), single group
# m_y:    raw data matrix (N x p), complete cases
# meanstructure: if TRUE, s_i = c(y_i, vech((y_i - ybar)(y_i - ybar)'));
#                if FALSE, only the vech part
#
# returns a list with both traces computed from the same casewise Gram
# matrix, exactly as in Hayakawa (2018):
#  - trace.UGamma  = tr(U Omega.hat) with the (unbiased) N-1 divisor
#  - trace.UGamma2 = the corrected (unbiased) estimator of tr((U Omega)^2)
lav_test_hayakawa_trace2 <- function(u = NULL, m_y = NULL,
                                     meanstructure = FALSE) {
  m_y <- unname(as.matrix(m_y))
  n <- nrow(m_y)
  p <- ncol(m_y)

  if (n < 4L) {
    lav_msg_stop(gettext(
      "the corrected trace estimator requires at least 4 observations."))
  }

  # casewise (saturated) moment vectors s_i
  m_yc <- t(t(m_y) - colMeans(m_y))
  idx1 <- lav_mat_vech_col_idx(p)
  idx2 <- lav_mat_vech_row_idx(p)
  z <- m_yc[, idx1, drop = FALSE] * m_yc[, idx2, drop = FALSE]
  if (meanstructure) {
    z <- cbind(m_y, z)
  }

  # sanity: U must match the moment vector
  if (ncol(z) != NROW(u)) {
    lav_msg_stop(gettextf(
      "dimension of U (%1$d) does not match the casewise moment vector
       (%2$d).", NROW(u), ncol(z)))
  }

  # center (rows become (s_i - sbar)')
  zc <- t(t(z) - colMeans(z))

  # N x N Gram matrix M = Zc U Zc' (= Y'Y with y_i = U^{1/2}(s_i - sbar);
  # no matrix square root needed)
  m_gram <- tcrossprod(zc %*% u, zc)

  tr_h <- sum(diag(m_gram))
  tr_h2 <- sum(m_gram * m_gram) # symmetric: tr(M^2)
  tr_d2 <- sum(diag(m_gram)^2)

  trace_ugamma2_c <-
    ((n - 2) * (n - 1) * tr_h2 - n * (n - 1) * tr_d2 + tr_h * tr_h) /
    (n * (n - 1) * (n - 2) * (n - 3))

  list(
    trace.UGamma = tr_h / (n - 1),
    trace.UGamma2 = trace_ugamma2_c
  )
}

# the weighted bivariate ordinal model
# YR 19 Feb 2020 (replacing the old lav_polychor.R routines)
#
# - polychoric (and tetrachoric) correlations
# - bivariate ordinal regression
# - using sampling weights wt
#
# YR/LDW 2026: refactored (see lav_uvbv_common.R for the shared cache/minfns
#              design); no change in behavior
#
#  info concerning lintr package, Luc DW May 7, 2026
#  there is a known problem in codetools, transferred to lintr, that
#  variables assigned in a with() are marked as 'may not be used';
#  to this end the code in with() statements are excluded from linting!
#

# two-way frequency table
# only works if Y = 1,2,3,...
lav_bvord_freq <- function(y1, y2, wt = NULL) {
  max_y1 <- max(y1, na.rm = TRUE)
  max_y2 <- max(y2, na.rm = TRUE)

  bin <- y1 - 1L
  bin <- bin + max_y1 * (y2 - 1L)
  bin <- bin + 1L

  if (is.null(wt)) {
    bin <- bin[!is.na(bin)]
    out <- array(tabulate(bin, nbins = max_y1 * max_y2),
      dim = c(max_y1, max_y2)
    )
  } else {
    if (anyNA(y1) || anyNA(y2)) {
      wt[is.na(y1) | is.na(y2)] <- 0
      bin[is.na(bin)] <- 0
    }
    y_ncat <- max_y1 * max_y2
    y_freq <- numeric(y_ncat)
    for (cat in seq_len(y_ncat)) {
      y_freq[cat] <- sum(wt[bin == cat])
    }
    out <- array(y_freq, dim = c(max_y1, max_y2))
  }

  out
}

# polychoric correlation
#
# zero.add is a vector: first element is for 2x2 tables only, second element
#                       for general tables
# zero.keep.margins is only used for 2x2 tables
#
lav_bvord_cor_twostep_fit <- function(y1, y2, exo = NULL, wt = NULL,
                                      fit_y1 = NULL, fit_y2 = NULL,
                                      freq = NULL,
                                      zero_add = c(0.5, 0.0),
                                      zero_keep_margins = TRUE,
                                      zero_cell_warn = FALSE,
                                      zero_cell_flag = FALSE,
                                      optim_method = "nlminb2",
                                      optim_scale = 1.0,
                                      init_theta = NULL,
                                      control = list(step.min = 0.1), # 0.6-7
                                      y1_name = NULL, y2_name = NULL) {
  # create cache environment
  cache <- lav_bvord_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt,
    fit_y1 = fit_y1, fit_y2 = fit_y2
  )

  # empty cells or not
  empty_cells <- FALSE

  # check for zero cells (if not exo), and catch some special cases
  if (cache$nexo == 0L) {
    freq <- cache$freq
    nr <- nrow(freq)
    nc <- ncol(freq)

    # check for empty cells
    if (any(freq == 0L)) {
      empty_cells <- TRUE
      if (zero_cell_warn) {
        if (!is.null(y1_name) && !is.null(y2_name)) {
          lav_msg_warn(gettextf(
            "empty cell(s) in bivariate table of %1$s x %2$s",
            y1_name, y2_name))
        } else {
          lav_msg_warn(gettext("empty cell(s) in bivariate table"))
        }
      }
    }

    # treat 2x2 tables
    if (nr == 2L && nc == 2L) {
      idx <- which(freq == 0L)
      # catch 2 empty cells: perfect correlation!
      if (length(idx) == 2L) {
        lav_msg_warn(gettext("two empty cells in 2x2 table"))
        if (freq[1, 1] > 0L) {
          rho <- 1.0
          if (zero_cell_flag) {
            attr(rho, "zero.cell.flag") <- empty_cells
          }
          return(rho)
        } else {
          rho <- -1.0
          if (zero_cell_flag) {
            attr(rho, "zero.cell.flag") <- empty_cells
          }
          return(rho)
        }
      } else if (length(idx) == 1L && zero_add[1] > 0.0) {
        if (zero_keep_margins) {
          # add + compensate to preserve margins
          if (idx == 1L || idx == 4L) { # main diagonal
            freq[1, 1] <- freq[1, 1] + zero_add[1]
            freq[2, 2] <- freq[2, 2] + zero_add[1]
            freq[2, 1] <- freq[2, 1] - zero_add[1]
            freq[1, 2] <- freq[1, 2] - zero_add[1]
          } else {
            freq[1, 1] <- freq[1, 1] - zero_add[1]
            freq[2, 2] <- freq[2, 2] - zero_add[1]
            freq[2, 1] <- freq[2, 1] + zero_add[1]
            freq[1, 2] <- freq[1, 2] + zero_add[1]
          }
        } else {
          freq[idx] <- freq[idx] + zero_add[1]
        }
      }
      # general table
    } else {
      if (any(freq == 0L) && zero_add[2] > 0.0) {
        # general table: just add zero.add to the empty cell(s)
        freq[freq == 0] <- zero_add[2]
      }
    }

    # update (possibly change) freq table
    cache$freq <- freq

    # catch special cases for 2x2 tables
    if (nr == 2L && nc == 2L) {
      # 1. a*d == c*d
      storage.mode(freq) <- "numeric" # to avoid integer overflow
      if (freq[1, 1] * freq[2, 2] == freq[1, 2] * freq[2, 1]) {
        rho <- 0.0
        if (zero_cell_flag) {
          attr(rho, "zero.cell.flag") <- empty_cells
        }
        return(rho)
      }
      # 2. equal margins (th1 = th2 = 0)
      if (cache$th_y1[1] == 0 && cache$th_y2[1] == 0) {
        # see eg Brown & Benedetti 1977 eq 2
        rho <- -cos(2 * pi * freq[1, 1] / sum(freq))
        if (zero_cell_flag) {
          attr(rho, "zero.cell.flag") <- empty_cells
        }
        return(rho)
      }
    }
  } # non-exo

  # optim.method
  minfns <- lav_uvbv_optim_fns(optim_method, lav_bvord_min_fns())

  # optimize
  if (is.null(control$trace)) {
    control$trace <- ifelse(lav_verbose(), 1, 0)
  }

  # init theta?
  if (!is.null(init_theta)) {
    start_x <- init_theta
  } else {
    start_x <- cache$theta
  }

  # try 1
  optim <- nlminb(
    start = start_x, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control,
    scale = optim_scale, lower = -0.999, upper = +0.999,
    cache = cache
  )

  # try 2
  if (optim$convergence != 0L) {
    # try again, with different starting value
    optim <- nlminb(
      start = 0, objective = minfns$objective,
      gradient = NULL, hessian = NULL,
      control = control,
      scale = optim_scale, lower = -0.995, upper = +0.995,
      cache = cache
    )
  }

  # check convergence
  if (optim$convergence != 0L) {
    if (!is.null(y1_name) && !is.null(y2_name)) {
      lav_msg_warn(gettextf(
        "estimation polychoric correlation did not converge for
                    variables %1$s and %2$s", y1_name, y2_name))
    } else {
      lav_msg_warn(gettext(
        "estimation polychoric correlation(s) did not always converge"))
    }
    rho <- start_x
  } else {
    rho <- optim$par
  }

  # zero.cell.flag
  if (zero_cell_flag) {
    attr(rho, "zero.cell.flag") <- empty_cells
  }

  rho
}

# build the cache from user-level arguments: fit the univariate models if
# needed, apply parameter overrides, and initialize the cache environment
# (shared by lav_bvord_cor_twostep_fit/_cor_sc/_logl/_lik)
lav_bvord_cache_from_args <- function(y1, y2, exo = NULL, wt = NULL,
                                      rho = NULL,
                                      fit_y1 = NULL, fit_y2 = NULL,
                                      th_y1 = NULL, th_y2 = NULL,
                                      sl_y1 = NULL, sl_y2 = NULL,
                                      scores = FALSE) {
  if (is.null(fit_y1)) {
    fit_y1 <- lav_uvord_fit(y = y1, x = exo, wt = wt)
  }
  if (is.null(fit_y2)) {
    fit_y2 <- lav_uvord_fit(y = y2, x = exo, wt = wt)
  }

  # update z1/z2 if needed (used in lav_pml_dploglik_dimplied() in
  #                                     lav_model_gradient_pml.R)
  fit_y1 <- lav_uvord_update_fit(
    fit_y = fit_y1,
    th_new = th_y1, sl_new = sl_y1
  )
  fit_y2 <- lav_uvord_update_fit(
    fit_y = fit_y2,
    th_new = th_y2, sl_new = sl_y2
  )

  cache <- lav_bvord_init_cache(
    fit_y1 = fit_y1, fit_y2 = fit_y2, wt = wt,
    scores = scores
  )
  if (!is.null(rho)) {
    cache$theta <- rho
  }

  cache
}

# prepare cache environment
lav_bvord_init_cache <- function(fit_y1 = NULL,
                                 fit_y2 = NULL,
                                 wt = NULL,
                                 scores = FALSE,
                                 parent = parent.frame()) {
  # data
  y1 <- fit_y1$y
  y2 <- fit_y2$y
  exo <- fit_y1$x

  # exo?
  if (is.null(exo)) {
    nexo <- 0L
    freq <- lav_bvord_freq(y1 = y1, y2 = y2, wt = wt)
    th_y1 <- fit_y1$theta[fit_y1$th_idx]
    th_y2 <- fit_y2$theta[fit_y2$th_idx]
    nth_y1 <- length(th_y1)
    nth_y2 <- length(th_y2)
    pth_y1 <- pnorm(th_y1)
    pth_y2 <- pnorm(th_y2)
  } else {
    nexo <- ncol(exo)
    freq <- NULL
    fit_y1_z1 <- fit_y1$z1
    fit_y2_z1 <- fit_y2$z1
    fit_y1_z2 <- fit_y1$z2
    fit_y2_z2 <- fit_y2$z2

    # take care of missing values
    if (length(fit_y1$missing_idx) > 0L || length(fit_y2$missing_idx) > 0L) {
      missing_idx <- unique(c(fit_y1$missing_idx, fit_y2$missing_idx))
      fit_y1_z1[missing_idx] <- 0
      fit_y2_z1[missing_idx] <- 0
      fit_y1_z2[missing_idx] <- 0
      fit_y2_z2[missing_idx] <- 0
    } else {
      missing_idx <- integer(0L)
    }
  }

  # nobs
  if (is.null(wt)) {
    n <- length(y1)
  } else {
    n <- sum(wt)
  }

  # starting value (for both exo and not-exo)
  # if (is.null(wt)) {
  if (sd(y1, na.rm = TRUE) == 0 || sd(y2, na.rm = TRUE) == 0) {
    rho_init <- 0.0
  } else {
    rho_init <- cor(y1, y2, use = "pairwise.complete.obs")
  }
  # }
  # cov.wt does not handle missing values...
  # rho.init <- cov.wt(cbind(Y1, Y2), wt = wt, cor = TRUE)$cor[2,1]
  if (is.na(rho_init) || abs(rho_init) >= 1.0) {
    rho_init <- 0.0
  }

  # parameter vector
  theta <- rho_init # only, for now

  # cache environment: common part + exo/noexo specific parts
  # (+ the case x th indicator matrices when scores are needed)
  cache_list <- list(
    nexo = nexo, theta = theta, wt = wt, n = n,
    y1 = y1, y2 = y2, freq = freq
  )
  if (nexo == 0L) {
    cache_list <- c(cache_list, list(
      th_y1 = th_y1, th_y2 = th_y2,
      nth_y1 = nth_y1, nth_y2 = nth_y2,
      pth_y1 = pth_y1, pth_y2 = pth_y2
    ))
    if (scores) {
      cache_list <- c(cache_list, list(
        fit_y1_z1 = fit_y1$z1, fit_y1_z2 = fit_y1$z2,
        fit_y2_z1 = fit_y2$z1, fit_y2_z2 = fit_y2$z2
      ))
    }
  } else {
    cache_list <- c(cache_list, list(
      fit_y1_z1 = fit_y1_z1, fit_y1_z2 = fit_y1_z2,
      fit_y2_z1 = fit_y2_z1, fit_y2_z2 = fit_y2_z2,
      missing_idx = missing_idx
    ))
    if (scores) {
      cache_list <- c(cache_list, list(exo = exo))
    }
  }
  if (scores) {
    cache_list <- c(cache_list, list(
      y1_y1 = fit_y1$y1, y1_y2 = fit_y1$y2,
      y2_y1 = fit_y2$y1, y2_y2 = fit_y2$y2
    ))
  }

  list2env(cache_list, parent = parent)
}

# probabilities for each cell, given rho, th_y1 and th_y2
lav_bvord_noexo_pi_cache <- function(cache = NULL) {
  with(cache, {                     # nolint start
    pi0 <- lav_bvord_noexo_pi(
      rho = theta[1L], th_y1 = th_y1, th_y2 = th_y2
    )
    return(pi0)
  })                                         # nolint end
}

# the four-corner accumulation: given the (nth_y1 x nth_y2) matrix of values
# at all threshold crossings, compute the (nth_y1+1 x nth_y2+1) per-cell
# rectangle combination x[i,j] - x[i-1,j] - x[i,j-1] + x[i-1,j-1]
lav_bvord_corner_sum <- function(x = NULL, nth_y1 = NULL, nth_y2 = NULL) {
  p1 <- p2 <- p3 <- p4 <- matrix(0, nth_y1 + 1L, nth_y2 + 1L)
  t1_idx <- seq_len(nth_y1)
  t2_idx <- seq_len(nth_y2)

  # p1 is left-upper corner
  p1[t1_idx, t2_idx] <- x
  # p2 is left-lower corner
  p2[t1_idx + 1L, t2_idx] <- x
  # p3 is right-upper corner
  p3[t1_idx, t2_idx + 1L] <- x
  # p4 is right-lower corner
  p4[t1_idx + 1L, t2_idx + 1L] <- x

  p1 - p2 - p3 + p4
}

# partial derivative of CDF(th_y1, th_y2, rho) with respect to rho
lav_bvord_noexo_phi_cache <- function(cache = NULL) {
  with(cache, {                  #  nolint start
    rho <- theta[1L]

    # compute lav_dbinorm for all possible combinations
    t1 <- rep(th_y1, times = nth_y2)
    t2 <- rep(th_y2, each = nth_y1)
    dbi_norm <- matrix(lav_dbinorm(t1, t2, rho),
      nrow = nth_y1, ncol = nth_y2
    )

    phi <- lav_bvord_corner_sum(dbi_norm, nth_y1, nth_y2)
    return(phi)
  })                                          # nolint end
}

# derivative of phi_2(y1,y2;rho) wrt rho equals
# phi_2(y1,y2;rho) * guv(y1,y2;rho) -- see lav_dbinorm_guv() in lav_bvreg.R
lav_bvord_noexo_gnorm_cache <- function(cache = NULL) {
  with(cache, {                         # nolint start
    rho <- theta[1L]

    # compute gnorm for all possible combinations
    gnorm_1 <- dbi_norm * matrix(lav_dbinorm_guv(t1, t2, rho), nth_y1, nth_y2)

    gnorm <- lav_bvord_corner_sum(gnorm_1, nth_y1, nth_y2)
    return(gnorm)
  })                                          # nolint end
}


# casewise likelihoods, unweighted!
lav_bvord_lik_cache <- function(cache = NULL) {
  with(cache, {                     # nolint start
    rho <- theta[1L]

    # no exo
    if (nexo == 0L) {
      pi0 <- lav_bvord_noexo_pi_cache(cache)
      lik <- pi0[cbind(y1, y2)]

      # exo
    } else {
      lik <- pbinorm(
        upper_x = fit_y1_z1, upper_y = fit_y2_z1,
        lower_x = fit_y1_z2, lower_y = fit_y2_z2, rho = rho
      )
      if (length(missing_idx) > 0L) {
        lik[missing_idx] <- NA
      }
      # catch very small values
      lik_toosmall_idx <- which(lik < .Machine$double.eps ^ (2 / 3))
      lik[lik_toosmall_idx] <- as.numeric(NA)
    }

    return(lik)
  })                                     # nolint end
}

lav_bvord_logl_cache <- function(cache = NULL) {
  with(cache, {                  # nolint start
    rho <- theta[1L]

    # no exo
    if (nexo == 0L) {
      pi0 <- lav_bvord_noexo_pi_cache(cache)
      logl <- sum(freq * log(pi0), na.rm = TRUE)

      # exo
    } else {
      lik <- lav_bvord_lik_cache(cache) # unweighted!
      if (!is.null(wt)) {
        logl <- sum(wt * log(lik), na.rm = TRUE)
      } else {
        logl <- sum(log(lik), na.rm = TRUE)
      }
    }

    return(logl)
  })                                     # nolint end
}

lav_bvord_grad_cache <- function(cache = NULL) {
  with(cache, {              # nolint start
    rho <- theta[1L]

    # no exo
    if (nexo == 0L) {
      phi <- lav_bvord_noexo_phi_cache(cache)
      bad_idx <- which(pi0 < .Machine$double.eps^(2/3))
      if (length(bad_idx) > 0L) {
        pi0[bad_idx] <- as.numeric(NA)
      }
      dx_rho <- sum((freq * phi) / pi0, na.rm = TRUE)

      # exo
    } else {
      d1 <- lav_dbinorm(fit_y1_z1, fit_y2_z1, rho)
      d2 <- lav_dbinorm(fit_y1_z2, fit_y2_z1, rho)
      d3 <- lav_dbinorm(fit_y1_z1, fit_y2_z2, rho)
      d4 <- lav_dbinorm(fit_y1_z2, fit_y2_z2, rho)
      phi <- (d1 - d2 - d3 + d4)

      # avoid dividing by very tiny numbers (new in 0.6-6)
      # -> done automatically: lik == NA in this case

      dx2 <- phi / lik

      if (is.null(wt)) {
        dx_rho <- sum(dx2, na.rm = TRUE)
      } else {
        dx_rho <- sum(wt * dx2, na.rm = TRUE)
      }
    }

    return(dx_rho)
  })                                     # nolint end
}

lav_bvord_hessian_cache <- function(cache = NULL) {
  with(cache, {                    # nolint start
    rho <- theta[1L]

    # no exo
    if (nexo == 0L) {
      bad_idx <- which(pi0 < .Machine$double.eps^(2/3))
      if (length(bad_idx) > 0L) {
        pi0[bad_idx] <- as.numeric(NA)
      }
      gnorm <- lav_bvord_noexo_gnorm_cache(cache)
      # H <- sum( freq * (gnorm/PI - (phi*phi)/(PI*PI)), na.rm = TRUE)
      h <- (sum((freq * gnorm) / pi0, na.rm = TRUE) -
        sum((freq * phi * phi) / (pi0 * pi0), na.rm = TRUE))
      dim(h) <- c(1L, 1L)

      # exo
    } else {
      gnorm <- ((d1 * lav_dbinorm_guv(fit_y1_z1, fit_y2_z1, rho)) -
        (d2 * lav_dbinorm_guv(fit_y1_z2, fit_y2_z1, rho)) -
        (d3 * lav_dbinorm_guv(fit_y1_z1, fit_y2_z2, rho)) +
        (d4 * lav_dbinorm_guv(fit_y1_z2, fit_y2_z2, rho)))

      if (is.null(wt)) {
        h <- sum(gnorm / lik - (phi * phi) / (lik * lik), na.rm = TRUE)
      } else {
        h <- sum(wt * (gnorm / lik - (phi * phi) / (lik * lik)), na.rm = TRUE)
      }

      dim(h) <- c(1L, 1L)
    }

    return(h)
  })                                              # nolint end
}

# nlminb objective/gradient/hessian (see lav_uvbv_common.R)
lav_bvord_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_bvord_logl_cache,
    grad_fun = lav_bvord_grad_cache,
    hessian_fun = lav_bvord_hessian_cache
  )
}


# casewise scores
lav_bvord_cor_sc_cache <- function(cache = NULL, na_zero = FALSE,
                                       use_weights = TRUE) {
  # make the arguments visible inside with(cache, ...): symbols there
  # resolve through the cache environment (and its parents), NOT through
  # this function's frame
  cache$na_zero <- na_zero
  cache$use_weights <- use_weights
  with(cache, {                 # nolint start
    rho <- theta[1L]
    r <- sqrt(1 - rho * rho)

    # lik
    lik <- lav_bvord_lik_cache(cache = cache)
    bad_idx <- which(lik < .Machine$double.eps^(2/3))
    if (length(bad_idx) > 0L) {
      lik[bad_idx] <- as.numeric(NA)
    }

    d_y1_z1 <- dnorm(fit_y1_z1)
    d_y1_z2 <- dnorm(fit_y1_z2)
    d_y2_z1 <- dnorm(fit_y2_z1)
    d_y2_z2 <- dnorm(fit_y2_z2)

    # th_y1
    if (identical(r, 0.0)) {
      y1_z1 <- d_y1_z1 * 0.5
      y1_z2 <- d_y1_z2 * 0.5
    } else {
      y1_z1 <- (d_y1_z1 * pnorm((fit_y2_z1 - rho * fit_y1_z1) / r) -
        d_y1_z1 * pnorm((fit_y2_z2 - rho * fit_y1_z1) / r))
      y1_z2 <- (d_y1_z2 * pnorm((fit_y2_z1 - rho * fit_y1_z2) / r) -
        d_y1_z2 * pnorm((fit_y2_z2 - rho * fit_y1_z2) / r))
    }
    dx_th_y1 <- (y1_y1 * y1_z1 - y1_y2 * y1_z2) / lik
    if (na_zero) {
      dx_th_y1[is.na(dx_th_y1)] <- 0
    }

    # th_y2
    if (identical(r, 0.0)) {
      y2_z1 <- d_y2_z1 * 0.5
      y2_z2 <- d_y2_z2 * 0.5
    } else {
      y2_z1 <- (d_y2_z1 * pnorm((fit_y1_z1 - rho * fit_y2_z1) / r) -
        d_y2_z1 * pnorm((fit_y1_z2 - rho * fit_y2_z1) / r))
      y2_z2 <- (d_y2_z2 * pnorm((fit_y1_z1 - rho * fit_y2_z2) / r) -
        d_y2_z2 * pnorm((fit_y1_z2 - rho * fit_y2_z2) / r))
    }
    dx_th_y2 <- (y2_y1 * y2_z1 - y2_y2 * y2_z2) / lik
    if (na_zero) {
      dx_th_y2[is.na(dx_th_y2)] <- 0
    }

    # slopes
    dx_sl_y1 <- dx_sl_y2 <- NULL
    if (nexo > 0L) {
      # sl.y1
      dx_sl_y1 <- (y1_z2 - y1_z1) * exo / lik
      if (na_zero) {
        dx_sl_y1[is.na(dx_sl_y1)] <- 0
      }

      # sl.y2
      dx_sl_y2 <- (y2_z2 - y2_z1) * exo / lik
      if (na_zero) {
        dx_sl_y2[is.na(dx_sl_y2)] <- 0
      }
    }

    # rho
    if (nexo == 0L) {
      phi <- lav_bvord_noexo_phi_cache(cache)
      dx <- phi[cbind(y1, y2)]
    } else {
      dx <- (lav_dbinorm(fit_y1_z1, fit_y2_z1, rho) -
        lav_dbinorm(fit_y1_z2, fit_y2_z1, rho) -
        lav_dbinorm(fit_y1_z1, fit_y2_z2, rho) +
        lav_dbinorm(fit_y1_z2, fit_y2_z2, rho))
    }
    dx_rho <- dx / lik
    if (na_zero) {
      dx_rho[is.na(dx_rho)] <- 0
    }

    if (!is.null(wt) && use_weights) {
      dx_th_y1 <- dx_th_y1 * wt
      dx_th_y2 <- dx_th_y2 * wt
      if (nexo > 0L) {
        dx_sl_y1 <- dx_sl_y1 * wt
        dx_sl_y2 <- dx_sl_y2 * wt
      }
      dx_rho <- dx_rho * wt
    }

    out <- list(
      dx_th_y1 = dx_th_y1, dx_th_y2 = dx_th_y2,
      dx_sl_y1 = dx_sl_y1, dx_sl_y2 = dx_sl_y2, dx_rho = dx_rho
    )
    return(out)
  })                                                   # nolint end
}


# casewise scores - no cache
lav_bvord_cor_sc <- function(y1, y2, exo = NULL, wt = NULL,
                                 rho = NULL,
                                 fit_y1 = NULL, fit_y2 = NULL,
                                 th_y1 = NULL, th_y2 = NULL,
                                 sl_y1 = NULL, sl_y2 = NULL,
                                 na_zero = FALSE, use_weights = TRUE) {
  cache <- lav_bvord_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    th_y1 = th_y1, th_y2 = th_y2,
    sl_y1 = sl_y1, sl_y2 = sl_y2,
    scores = TRUE
  )

  lav_bvord_cor_sc_cache(
    cache = cache, na_zero = na_zero,
    use_weights = use_weights
  )
}

# logl - no cache
lav_bvord_logl <- function(y1, y2, exo = NULL, wt = NULL,
                           rho = NULL,
                           fit_y1 = NULL, fit_y2 = NULL,
                           th_y1 = NULL, th_y2 = NULL,
                           sl_y1 = NULL, sl_y2 = NULL) {
  cache <- lav_bvord_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    th_y1 = th_y1, th_y2 = th_y2,
    sl_y1 = sl_y1, sl_y2 = sl_y2
  )

  lav_bvord_logl_cache(cache = cache)
}

# lik - no cache
lav_bvord_lik <- function(y1, y2, exo = NULL, wt = NULL,
                          rho = NULL,
                          fit_y1 = NULL, fit_y2 = NULL,
                          th_y1 = NULL, th_y2 = NULL,
                          sl_y1 = NULL, sl_y2 = NULL,
                          .log = FALSE) {
  cache <- lav_bvord_cache_from_args(
    y1 = y1, y2 = y2, exo = exo, wt = wt, rho = rho,
    fit_y1 = fit_y1, fit_y2 = fit_y2,
    th_y1 = th_y1, th_y2 = th_y2,
    sl_y1 = sl_y1, sl_y2 = sl_y2
  )

  lik <- lav_bvord_lik_cache(cache = cache) # unweighted
  if (.log) {
    lik <- log(lik)
  }

  if (!is.null(wt)) {
    if (.log) {
      lik <- wt * lik
    } else {
      tmp <- wt * log(lik)
      lik <- exp(tmp)
    }
  }

  lik
}

# probabilities for each cell, given rho, th_y1 and th_y2
# (also used directly by lav_tables.R and the PML objective)
lav_bvord_noexo_pi <- function(rho = NULL, th_y1 = NULL, th_y2 = NULL) {
  nth_y1 <- length(th_y1)
  nth_y2 <- length(th_y2)
  pth_y1 <- pnorm(th_y1)
  pth_y2 <- pnorm(th_y2)

  # catch special case: rho = 0.0
  if (rho == 0.0) {
    row_pi <- base::diff(c(0, pth_y1, 1))
    col_pi <- base::diff(c(0, pth_y2, 1))
    pi_ij <- base::outer(row_pi, col_pi)
    return(pi_ij)
  }

  # prepare for a single call to pbivnorm
  upper_y <- rep(th_y2, times = rep.int(nth_y1, nth_y2))
  upper_x <- rep(th_y1, times = ceiling(length(upper_y)) / nth_y1)

  bi <- pbivnorm::pbivnorm(x = upper_x, y = upper_y, rho = rho)
  dim(bi) <- c(nth_y1, nth_y2)
  bi <- rbind(0, bi, pth_y2, deparse.level = 0L)
  bi <- cbind(0, bi, c(0, pth_y1, 1), deparse.level = 0L)

  # get probabilities
  nr <- nrow(bi)
  nc <- ncol(bi)
  pi0 <- bi[-1L, -1L] - bi[-1L, -nc] - bi[-nr, -1L] + bi[-nr, -nc]

  # all elements should be strictly positive
  pi0[pi0 < .Machine$double.eps ^ (2 / 3)] <-
                   .Machine$double.eps ^ (2 / 3)
  pi0
}

# parse user-supplied constraint/definition text, re-raising any parse error
# as a clean lavaan error (instead of a raw R parser error)
lav_pt_con_parse <- function(text, what = gettext("constraint(s)")) {
  res <- tryCatch(parse(file = "", text = text),
                  error = function(e) conditionMessage(e))
  if (!is.expression(res)) {
    lav_msg_stop(gettextf(
      "could not parse %1$s: %2$s", what, res))
  }
  res
}

# return the function names used in one or more expression strings that
# can NOT be resolved (as a function) from environment 'env'
# (used to give a clean error for unknown functions in := / constraints)
lav_pt_con_undefined_fun <- function(txt, env) {
  txt <- txt[nchar(txt) > 0L]
  if (length(txt) == 0L) {
    return(character(0L))
  }
  expr <- tryCatch(parse(file = "", text = txt), error = function(e) NULL)
  if (is.null(expr)) {
    return(character(0L))
  }
  # function names = all names that are not plain variables
  fun_names <- setdiff(all.names(expr), all.vars(expr))
  if (length(fun_names) == 0L) {
    return(character(0L))
  }
  ok <- vapply(fun_names, function(fn) {
    exists(fn, envir = env, mode = "function", inherits = TRUE)
  }, logical(1L))
  fun_names[!ok]
}

# build def function from partable
lav_pt_con_def <- function(partable, con = NULL, debug = FALSE, # nolint start
                                         txt_only = FALSE, warn = TRUE) {      # nolint end
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  # empty function
  def_function <- function() NULL

  # if 'con', merge partable + con
  if (!is.null(con)) {
    partable$lhs <- c(partable$lhs, con$lhs)
    partable$op <- c(partable$op, con$op)
    partable$rhs <- c(partable$rhs, con$rhs)
  }

  # get := definitions
  def_idx <- which(partable$op == ":=")

  # catch empty def
  if (length(def_idx) == 0L) {
    if (txt_only) {
      return(character(0L))
    } else {
      return(def_function)
    }
  }

  # sort order of def.idx by dependencies
  deps <- lapply(partable$rhs[def_idx], FUN = function(x) {
    all.vars(lav_pt_con_parse(x, gettext("parameter definition(s)"))) })
  lab_unsorted <- partable$lhs[def_idx]
  adj_mat <- matrix(0L, nrow = length(def_idx), ncol = length(def_idx))
  for (i in seq_along(lab_unsorted)) {
    adj_mat[lab_unsorted %in% deps[[i]], i] <- 1L
  }
  # note: our validation below gives a clean error for any cyclic system,
  # so we suppress the (less informative) cycle warning from the sorter
  def_idx <- def_idx[lav_graph_order_adj_mat(adj_mat, warn = FALSE)]

  # validate that, in sorted order, each ':=' definition only refers to model
  # parameter labels or to ':=' parameters that have been defined EARLIER;
  # this catches forward/self/circular references (which would otherwise
  # produce cryptic 'object not found'/'non-numeric argument' errors)
  def_lhs_sorted <- as.character(partable$lhs[def_idx])
  # available names = model parameter labels, EXCLUDING the ':=' parameters
  # themselves (in a full partable, ':=' rows carry a $label equal to their
  # name, so we drop the def names from the label set)
  model_labels <- partable$label[nchar(partable$label) > 0L]
  available <- setdiff(unique(model_labels), def_lhs_sorted)
  for (i in seq_along(def_idx)) {
    rhs_vars <- all.vars(parse(file = "", text = partable$rhs[def_idx[i]]))
    # only flag references to (other/self) ':=' names not yet available;
    # genuinely unknown labels are reported by the check further below
    bad <- rhs_vars[rhs_vars %in% def_lhs_sorted & !rhs_vars %in% available]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "defined parameter `%1$s' in `%2$s' refers to parameter(s) that are not (yet) defined: %3$s",
        def_lhs_sorted[i],
        paste(def_lhs_sorted[i], ":=", partable$rhs[def_idx[i]]),
        lav_msg_view(bad, "none")))
    }
    available <- c(available, def_lhs_sorted[i])
  }

  # create function
  formals(def_function) <- alist(.x. = , ... = )
  if (txt_only) {
    body_txt <- ""
  } else {
    body_txt <- paste("{\n# parameter definitions\n\n")
  }

  lhs_names <- partable$lhs[def_idx]
  def_labels <- all.vars(parse(file = "", text = partable$rhs[def_idx]))
  # remove the ones in lhs.names
  idx <- which(def_labels %in% lhs_names)
  if (length(idx) > 0L) def_labels <- def_labels[-idx]

  # get corresponding 'x' indices
  def_x_idx <- partable$free[match(def_labels, partable$label)]
  if (any(is.na(def_x_idx))) {
    lav_msg_stop(gettext(
      "unknown label(s) in variable definition(s):"),
      lav_msg_view(def_labels[which(is.na(def_x_idx))], "none")
    )
  }
  if (any(def_x_idx == 0)) {
    lav_msg_stop(gettext(
      "non-free parameter(s) in variable definition(s):"),
      lav_msg_view(def_labels[which(def_x_idx == 0)], "none")
    )
  }
  def_x_lab <- paste(".x.[", def_x_idx, "]", sep = "")
  # put both the labels the function BODY
  if (length(def_x_idx) > 0L) {
    body_txt <- paste(body_txt, "# parameter labels\n",
      paste(def_labels, " <- ", def_x_lab, collapse = "\n"),
      "\n",
      sep = ""
    )
  }

  # write the definitions literally
  body_txt <- paste(body_txt, "\n# parameter definitions\n", sep = "")
  for (i in seq_along(def_idx)) {
    body_txt <- paste(body_txt,
      lhs_names[i], " <- ", partable$rhs[def_idx[i]], "\n",
      sep = ""
    )
  }

  if (txt_only) {
    return(body_txt)
  }

  # put the results in 'out'
  body_txt <- paste(body_txt, "\nout <- ",
    paste("c(", paste(lhs_names, collapse = ","), ")\n", sep = ""),
    sep = ""
  )
  # what to do with NA values? -> return +Inf???
  body_txt <- paste(body_txt, "out[is.na(out)] <- Inf\n", sep = "")
  body_txt <- paste(body_txt, "names(out) <- ",
    paste("c(\"", paste(lhs_names, collapse = "\",\""), "\")\n", sep = ""),
    sep = ""
  )
  body_txt <- paste(body_txt, "return(out)\n}\n", sep = "")

  body(def_function) <- parse(file = "", text = body_txt)

  # check for unknown functions in the definitions (clean error at build time)
  bad_fun <- lav_pt_con_undefined_fun(partable$rhs[def_idx],
                                      environment(def_function))
  if (length(bad_fun) > 0L) {
    lav_msg_stop(gettext(
      "could not find function(s) used in variable definition(s):"),
      lav_msg_view(bad_fun, "none")
    )
  }

  if (lav_debug()) {
    cat("def.function = \n")
    print(def_function)
    cat("\n")
  }

  def_function
}

# build ceq function from partable
#     non-trivial equality constraints (linear or nonlinear)
#     convert to 'ceq(x)' function where 'x' is the (free) parameter vector
#     and ceq(x) returns the evaluated equality constraints
#
#     eg. if b1 + b2 == 2 (and b1 correspond to, say,  x[10] and x[17])
#         ceq <- function(x) {
#             out <- rep(NA, 1)
#             b1 = x[10]; b2 = x[17]
#             out[1] <- b1 + b2 - 2
#         }
lav_pt_con_ceq <- function(partable, con = NULL, debug = FALSE, # nolint start
                                         txt_only = FALSE) {                   # nolint end
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  # empty function
  ceq_function <- function() NULL

  # if 'con', merge partable + con
  if (!is.null(con)) {
    partable$lhs <- c(partable$lhs, con$lhs)
    partable$op <- c(partable$op, con$op)
    partable$rhs <- c(partable$rhs, con$rhs)
  }

  # get equality constraints
  eq_idx <- which(partable$op == "==")

  # catch empty ceq
  if (length(eq_idx) == 0L) {
    if (txt_only) {
      return(character(0L))
    } else {
      return(ceq_function)
    }
  }

  # create function
  formals(ceq_function) <- alist(.x. = , ... = )
  if (txt_only) {
    body_txt <- ""
  } else {
    body_txt <- paste("{\nout <- rep(NA, ", length(eq_idx), ")\n", sep = "")
  }

  # first come the variable definitions
  def_txt <- lav_pt_con_def(partable, txt_only = TRUE,
                                          warn = FALSE)
  def_idx <- which(partable$op == ":=")
  body_txt <- paste(body_txt, def_txt, "\n", sep = "")


  # extract labels
  eq_what <- gettext("equality constraint(s)")
  lhs_labels <- all.vars(lav_pt_con_parse(partable$lhs[eq_idx], eq_what))
  rhs_labels <- all.vars(lav_pt_con_parse(partable$rhs[eq_idx], eq_what))
  eq_labels <- unique(c(lhs_labels, rhs_labels))
  # remove def.names from eq.labels
  if (length(def_idx) > 0L) {
    def_names <- as.character(partable$lhs[def_idx])
    d_idx <- which(eq_labels %in% def_names)
    if (length(d_idx) > 0) eq_labels <- eq_labels[-d_idx]
  }
  eq_x_idx <- rep(as.integer(NA), length(eq_labels))
  # get user-labels ids
  ulab_idx <- which(eq_labels %in% partable$label)
  if (length(ulab_idx) > 0L) {
    eq_x_idx[ulab_idx] <- partable$free[match(
      eq_labels[ulab_idx],
      partable$label
    )]
  }
  # get plabels ids
  plab_idx <- which(eq_labels %in% partable$plabel)
  if (length(plab_idx) > 0L) {
    eq_x_idx[plab_idx] <- partable$free[match(
      eq_labels[plab_idx],
      partable$plabel
    )]
  }

  # check if we have found the label
  if (any(is.na(eq_x_idx))) {
    lav_msg_stop(gettext("unknown label(s) in equality constraint(s):"),
      lav_msg_view(eq_labels[which(is.na(eq_x_idx))], "none")
    )
  }
  # check if they are all 'free'
  if (any(eq_x_idx == 0)) {
    fixed_eq_idx <- which(eq_x_idx == 0)
    # FIXME: what should we do here? we used to stop with an error
    # from 0.5.18, we give a warning, and replace the non-free label
    # with its fixed value in ustart
    # warning(
    #    "lavaan WARNING: non-free parameter(s) in equality constraint(s): ",
    #    paste(eq.labels[fixed.eq.idx], collapse=" "))

    fixed_lab_lhs <- eq_labels[fixed_eq_idx]
    fixed_lab_rhs <- numeric(length(fixed_lab_lhs))

    for (i in seq_along(fixed_lab_lhs)) {
      # first try label
      idx <- match(fixed_lab_lhs[i], partable$label)
      # then try plabel
      if (is.na(idx)) {
        idx <- match(fixed_lab_lhs[i], partable$plabel)
      }
      if (is.na(idx)) {
        # hm, not found? fill in zero, or NA?
      } else {
        fixed_lab_rhs[i] <- partable$ustart[idx]
      }
    }

    body_txt <- paste(body_txt, "# non-free parameter labels\n",
      paste(fixed_lab_lhs, "<-", fixed_lab_rhs, collapse = "\n"),
      "\n",
      sep = ""
    )

    eq_x_idx <- eq_x_idx[-fixed_eq_idx]
    eq_labels <- eq_labels[-fixed_eq_idx]
  }

  # put the labels the function BODY
  eq_x_lab <- paste(".x.[", eq_x_idx, "]", sep = "")
  if (length(eq_x_idx) > 0L) {
    body_txt <- paste(body_txt, "# parameter labels\n",
      paste(eq_labels, "<-", eq_x_lab, collapse = "\n"),
      "\n",
      sep = ""
    )
  }

  # write the equality constraints literally
  body_txt <- paste(body_txt, "\n# equality constraints\n", sep = "")
  for (i in seq_along(eq_idx)) {
    lhs <- partable$lhs[eq_idx[i]]
    rhs <- partable$rhs[eq_idx[i]]
    if (rhs == "0") {
      eq_string <- lhs
    } else {
      eq_string <- paste(lhs, " - (", rhs, ")", sep = "")
    }
    body_txt <- paste(body_txt, "out[", i, "] <- ", eq_string, "\n", sep = "")
  }

  if (txt_only) {
    return(body_txt)
  }

  # put the results in 'out'
  # BODY.txt <- paste(BODY.txt, "\nout <- ",
  #    paste("c(", paste(lhs.names, collapse=","),")\n", sep=""), sep="")

  # what to do with NA values? -> return +Inf???
  body_txt <- paste(body_txt, "\n", "out[is.na(out)] <- Inf\n", sep = "")
  body_txt <- paste(body_txt, "return(out)\n}\n", sep = "")
  body(ceq_function) <- parse(file = "", text = body_txt)

  # check for unknown functions in the constraints (clean error at build time)
  bad_fun <- lav_pt_con_undefined_fun(
    c(partable$lhs[eq_idx], partable$rhs[eq_idx]),
    environment(ceq_function)
  )
  if (length(bad_fun) > 0L) {
    lav_msg_stop(gettext(
      "could not find function(s) used in equality constraint(s):"),
      lav_msg_view(bad_fun, "none")
    )
  }

  if (lav_debug()) {
    cat("ceq.function = \n")
    print(ceq_function)
    cat("\n")
  }

  ceq_function
}


# build ciq function from partable
#     non-trivial inequality constraints (linear or nonlinear)
#     convert to 'cin(x)' function where 'x' is the (free) parameter vector
#     and cin(x) returns the evaluated inequality constraints
#
#     eg. if b1 + b2 > 2 (and b1 correspond to, say,  x[10] and x[17])
#         cin <- function(x) {
#             out <- rep(NA, 1)
#             b1 = x[10]; b2 = x[17]
#             out[1] <- b1 + b2 - 2
#         }
#
#     new in 0.6-19: we also add the lower/upper bounds
#
# NOTE: very similar, but not identitical to ceq, because we need to take
#       care of the difference between '<' and '>'
lav_pt_con_ciq <- function(partable, con = NULL, debug = FALSE, # nolint start
                                         txt_only = FALSE) {                   # nolint end
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  # empty function
  cin_function <- function() NULL

  # if 'con', merge partable + con
  if (!is.null(con)) {
    partable$lhs <- c(partable$lhs, con$lhs)
    partable$op <- c(partable$op, con$op)
    partable$rhs <- c(partable$rhs, con$rhs)
  }

  # get explicit inequality constraints
  ineq_idx <- which(partable$op == ">" | partable$op == "<")

  # get lower/upper bounds
  upper_idx <- integer(0L)
  lower_idx <- integer(0L)
  if (!is.null(partable$upper)) {
    upper_idx <- which(partable$free > 0L  & is.finite(partable$upper))
  }
  if (!is.null(partable$lower)) {
    lower_idx <- which(partable$free > 0L & is.finite(partable$lower))
  }

  # add them to ineq.idx
  ineq_only_idx <- ineq_idx
  ineq_idx <- c(upper_idx, lower_idx, ineq_idx)

  # catch empty ciq
  if (length(ineq_idx) == 0L) {
    if (txt_only) {
      return(character(0L))
    } else {
      return(cin_function)
    }
  }

  # create function
  formals(cin_function) <- alist(.x. = , ... = )
  if (txt_only) {
    body_txt <- ""
  } else {
    body_txt <- paste("{\nout <- rep(NA, ", length(ineq_idx), ")\n", sep = "")
  }

  # first come the variable definitions
  def_txt <- lav_pt_con_def(partable, txt_only = TRUE,
                                          warn = FALSE)
  def_idx <- which(partable$op == ":=")
  body_txt <- paste(body_txt, def_txt, "\n", sep = "")

  # extract labels
  ineq_what <- gettext("inequality constraint(s)")
  lhs_labels <- all.vars(lav_pt_con_parse(partable$lhs[ineq_only_idx], ineq_what))
  rhs_labels <- all.vars(lav_pt_con_parse(partable$rhs[ineq_only_idx], ineq_what))
  ineq_labels <- unique(c(lhs_labels, rhs_labels))
  # remove def.names from ineq.labels
  if (length(def_idx) > 0L) {
    def_names <- as.character(partable$lhs[def_idx])
    d_idx <- which(ineq_labels %in% def_names)
    if (length(d_idx) > 0) ineq_labels <- ineq_labels[-d_idx]
  }
  ineq_x_idx <- rep(as.integer(NA), length(ineq_labels))
  # get user-labels ids
  ulab_idx <- which(ineq_labels %in% partable$label)
  if (length(ulab_idx) > 0L) {
    ineq_x_idx[ulab_idx] <- partable$free[match(
      ineq_labels[ulab_idx],
      partable$label
    )]
  }
  # get plabels ids
  plab_idx <- which(ineq_labels %in% partable$plabel)
  if (length(plab_idx) > 0L) {
    ineq_x_idx[plab_idx] <- partable$free[match(
      ineq_labels[plab_idx],
      partable$plabel
    )]
  }

  # check if we have found the label
  if (length(ineq_x_idx) > 0L && any(is.na(ineq_x_idx))) {
    lav_msg_stop(gettext("unknown label(s) in inequality constraint(s):"),
      lav_msg_view(ineq_labels[which(is.na(ineq_x_idx))], "none")
    )
  }

  # check if they are all 'free'
  if (length(ineq_x_idx) > 0L && any(ineq_x_idx == 0)) {
    fixed_ineq_idx <- which(ineq_x_idx == 0)
    # FIXME: what should we do here? we used to stop with an error
    # from 0.5.18, we give a warning, and replace the non-free label
    # with its fixed value in ustart
    lav_msg_warn(gettext("non-free parameter(s) in inequality constraint(s):"),
      lav_msg_view(ineq_labels[fixed_ineq_idx], "none")
    )

    fixed_lab_lhs <- ineq_labels[fixed_ineq_idx]
    fixed_lab_rhs <- partable$ustart[match(fixed_lab_lhs, partable$label)]
    body_txt <- paste(body_txt, "# non-free parameter labels\n",
      paste(fixed_lab_lhs, "<-", fixed_lab_rhs, collapse = "\n"),
      "\n",
      sep = ""
    )

    ineq_x_idx <- ineq_x_idx[-fixed_ineq_idx]
    ineq_labels <- ineq_labels[-fixed_ineq_idx]
  }

  # put the labels the function BODY
  if (length(ineq_x_idx) > 0L) {
    ineq_x_lab <- paste(".x.[", ineq_x_idx, "]", sep = "")
    if (length(ineq_x_idx) > 0L) {
      body_txt <- paste(body_txt, "# parameter labels\n",
        paste(ineq_labels, "<-", ineq_x_lab, collapse = "\n"),
        "\n",
        sep = ""
      )
    }
  }

  # write the constraints literally
  body_txt <- paste(body_txt, "\n# inequality constraints\n", sep = "")
  free <- partable$free
  free[free > 0] <- seq_along(free[free > 0])
  for (i in seq_along(ineq_idx)) {
    lhs <- partable$lhs[ineq_idx[i]]
    op <- partable$op[ineq_idx[i]]
    rhs <- partable$rhs[ineq_idx[i]]

    if (ineq_idx[i] %in% ineq_only_idx) {
      # EXPLICIT inequality constraints
      # note,this is different from ==, because we have < AND >
      if (rhs == "0" && op == ">") {
        ineq_string <- lhs
      } else if (rhs == "0" && op == "<") {
        ineq_string <- paste(rhs, " - (", lhs, ")", sep = "")
      } else if (rhs != "0" && op == ">") {
        ineq_string <- paste(lhs, " - (", rhs, ")", sep = "")
      } else if (rhs != "0" && op == "<") {
        ineq_string <- paste(rhs, " - (", lhs, ")", sep = "")
      }
    } else if (ineq_idx[i] %in% upper_idx) {
      # simple upper bound
      val <- partable$upper[ineq_idx[i]]
      xlab <- paste(".x.[", free[ineq_idx[i]], "]", sep = "")
      ineq_string <- paste(val, " - (", xlab, ")", sep = "")
    } else if (ineq_idx[i] %in% lower_idx) {
      # simple lower bound
      val <- partable$lower[ineq_idx[i]]
      xlab <- paste(".x.[", free[ineq_idx[i]], "]", sep = "")
      ineq_string <- paste(xlab, " - (", val, ")", sep = "")
    }

    body_txt <- paste(body_txt, "out[", i, "] <- ", ineq_string, "\n", sep = "")
  }

  if (txt_only) {
    return(body_txt)
  }

  # put the results in 'out'
  body_txt <- paste(body_txt, "\n", "out[is.na(out)] <- Inf\n", sep = "")
  if (length(upper_idx) > 0L || length(lower_idx) > 0L) {
    bound_idx <- which(ineq_idx %in% c(upper_idx, lower_idx))
    bound_txt <- paste("c(", paste(bound_idx, collapse = ", "), ")", sep = "")
    body_txt <- paste(body_txt, "attr(out, \"bound.idx\") <- ", bound_txt,
                      "\n", sep = "")
  }
  body_txt <- paste(body_txt, "return(out)\n}\n", sep = "")
  body(cin_function) <- parse(file = "", text = body_txt)

  # check for unknown functions in the constraints (clean error at build time)
  if (length(ineq_only_idx) > 0L) {
    bad_fun <- lav_pt_con_undefined_fun(
      c(partable$lhs[ineq_only_idx], partable$rhs[ineq_only_idx]),
      environment(cin_function)
    )
    if (length(bad_fun) > 0L) {
      lav_msg_stop(gettext(
        "could not find function(s) used in inequality constraint(s):"),
        lav_msg_view(bad_fun, "none")
      )
    }
  }

  if (lav_debug()) {
    cat("cin.function = \n")
    print(cin_function)
    cat("\n")
  }

  cin_function
}

# return a named vector of the 'free' indices, for the labels that
# are used in a constrained (or optionally a definition)
# (always 0 for definitions)
lav_pt_con_label_id <- function(partable, con = NULL,
                                              def = TRUE) {
  # if 'con', merge partable + con
  if (!is.null(con)) {
    partable$lhs <- c(partable$lhs, con$lhs)
    partable$op <- c(partable$op, con$op)
    partable$rhs <- c(partable$rhs, con$rhs)
  }

  # get constraints
  if (def) {
    con_idx <- which(partable$op %in% c("==", "<", ">", ":="))
  } else {
    con_idx <- which(partable$op %in% c("==", "<", ">"))
  }

  # catch empty con
  if (length(con_idx) == 0L) {
    return(integer(0L))
  }

  def_idx <- which(partable$op == ":=")

  # extract labels
  lhs_labels <- all.vars(lav_pt_con_parse(partable$lhs[con_idx]))
  rhs_labels <- all.vars(lav_pt_con_parse(partable$rhs[con_idx]))
  con_labels <- unique(c(lhs_labels, rhs_labels))

  # remove def.names from con.labels (unless def = TRUE)
  if (!def && length(def_idx) > 0L) {
    def_names <- as.character(partable$lhs[def_idx])
    d_idx <- which(con_labels %in% def_names)
    if (length(d_idx) > 0) {
      con_labels <- con_labels[-d_idx]
    }
  }
  con_x_idx <- rep(as.integer(NA), length(con_labels))

  # get user-labels ids
  ulab_idx <- which(con_labels %in% partable$label)
  if (length(ulab_idx) > 0L) {
    con_x_idx[ulab_idx] <- partable$free[match(
      con_labels[ulab_idx],
      partable$label
    )]
  }
  # get plabels ids
  plab_idx <- which(con_labels %in% partable$plabel)
  if (length(plab_idx) > 0L) {
    con_x_idx[plab_idx] <- partable$free[match(
      con_labels[plab_idx],
      partable$plabel
    )]
  }

  # check if we have found the label
  if (any(is.na(con_x_idx))) {
    lav_msg_warn(gettext("unknown label(s) in equality constraint(s):"),
      lav_msg_view(con_labels[which(is.na(con_x_idx))], "none")
    )
  }

  # return named integer vector
  names(con_x_idx) <- con_labels

  con_x_idx
}

# New version of parser, written by Luc De Wilde in september/october 2023
# Adapt to style guide by LDW on April 4, 2026

# ----------------------- lav_create_enum ------------------------------------ #
# function to create an Enumerable like structure in R
#  usage example  mycolors <- lav_create_enum(c("black", "white",
#                                     "orange", "green", "red", "blue"))
#                 xyz <- mycolors$red
# values are default 1L, ..., number of names, but can be user specified
# ---------------------------------------------------------------------------- #
lav_create_enum <- function(names, values = seq_along(names)) {
  stopifnot(identical(unique(names), names), is.character(names))
  stopifnot(length(names) == length(values))
  res <- as.list(setNames(values, names))
  res$enum_names <- names
  res$enum_values <- values
  res$enum_size <- length(values)
  res <- as.environment(res)
  lockEnvironment(res, bindings = TRUE)
  res
}

# ------------------------ lav_parse_sublist --------------------------------- #
# function to create a list with only some indexes for all members
# ---------------------------------------------------------------------------- #
lav_parse_sublist <- function(inlist, indexes) {
  for (j in seq_along(inlist)) {
    inlist[[j]] <- inlist[[j]][indexes]
  }
  inlist
}

# ------------------------ lav_parse_eval_r_expression ------------------- #
# help function to evaluate the value of an r expression formed by the elements
# with index 'from' to 'to' of a formula 'formul1'
# returns "_error_" if evaluation failed
# used only in lav_parse_modifier
# ---------------------------------------------------------------------------- #
lav_parse_eval_r_expression <- function(formul1, from, to, types) {
  strings <- vapply(seq.int(from, to), function(x) {
    if (formul1$elem_type[x] == types$stringliteral) {
      paste0('"', formul1$elem_text[x], '"')
    } else {
      formul1$elem_text[x]
    }
  }, "")
  txt <- paste(strings, collapse = "")
  result <- try(eval(parse(text = txt),
    envir = NULL,
    enclos = baseenv()
  ), silent = TRUE)
  if (inherits(result, "try-error")) {
    return("_error_")
  }
  result
}
# ------------------------ lav_parse_txtloc  --------------------------------- #
# function which translates a position in the model source string to a
# user friendly locator (=[1L]) and the line with position (=[2L])
# ---------------------------------------------------------------------------- #
lav_parse_txtloc <- function(modelsrc, position) {
  txt <- c("", "")
  if (nchar(modelsrc) >= position && position > 0) {
    newlines <- gregexpr("\n", paste0(modelsrc, "\n"), fixed = TRUE)[[1]]
    lijn <- which(newlines >= position)[1]
    if (lijn == 1L) {
      pos <- position
      lijnchar <- substr(modelsrc, 1L, newlines[1])
    } else {
      pos <- position - newlines[lijn - 1L]
      lijnchar <- substr(modelsrc, newlines[lijn - 1L] + 1L, newlines[lijn])
    }
    if (nchar(lijnchar) == 1L) {
      lijnchar <- ""
    } else {
      lijnchar <- substr(lijnchar, 1L, nchar(lijnchar) - 1)
    }
    # adapt line number when first line blank :
    if (grepl("^[ \t]*\n", modelsrc)) lijn <- lijn - 1L
    txt <- c(
      gettextf(" at line %1$s, pos %2$s", lijn, pos),
      paste(lijnchar, "\n", strrep(" ", pos - 1L), "^\n", sep = "")
    )
  }
  txt
}

# ------------------------ lav_parse_text_tokens ----------------------------- #
# function to split the model source in tokens.
# Returns a list with tokens with their attributes
#   elem_pos  : position in source
#   elem_type : type of token (cf. definition of types
#               in lav_parse_model_string)
#   elem_text : the text of the token
#   elem_formule.number : sequence number of the 'logical'
#                         formula where the token occurs
# the function returns the stored tokens in a list
# ---------------------------------------------------------------------------- #
lav_parse_text_tokens <- function(modelsrc, types) {
  nmax <- nchar(modelsrc)
  elem_pos <- vector("integer", nmax)
  elem_type <- elem_pos
  elem_text <- vector("character", nmax)
  elem_i <- 1L
  modelsrcw <- paste0(modelsrc, "\n") # working model, must end
  # with a newline for tests via regexpr
  stringliterals <- gregexpr("\"[^\"]*?[\"\n]", modelsrcw)[[1L]]
  if (stringliterals[1L] > -1L) {
    stringliteral_lengths <- attr(stringliterals, "match.length")
    for (i in seq_along(stringliterals)) {
      pfpos <- stringliterals[i]
      pflen <- stringliteral_lengths[i]
      substr(modelsrcw, pfpos + 1L, pfpos + pflen - 2L) <-
        strrep(" ", pflen - 2L)
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos + 1L, pfpos + pflen - 2L)
      elem_type[elem_i] <- types$stringliteral
      elem_i <- elem_i + 1L
    }
  }
  comments <- gregexpr("[#!].*?\n", modelsrcw)[[1L]]
  if (comments[1] > -1L) {
    comment_lengths <- attr(comments, "match.length")
    for (i in seq_along(comments)) {
      substr(modelsrcw, comments[i], comments[i] + comment_lengths[i] - 1L) <-
        strrep(" ", comment_lengths[i] - 1L)
      # check for stringliterals in comment
      str_in_comment <- (elem_pos > comments[i] &
        elem_pos < comments[i] + comment_lengths[i])
      if (any(str_in_comment)) {
        elem_type[str_in_comment] <- 0
      }
    }
  }
  modelsrcw <- gsub("\t", " ", modelsrcw)
  newlines <- gregexpr("[;\n]", modelsrcw)[[1L]]
  if (newlines[1L] > -1L) {
    for (i in seq_along(newlines)) {
      pfpos <- newlines[i]
      substr(modelsrcw, pfpos, pfpos) <- "\n"
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- "\n"
      elem_type[elem_i] <- types$newline
      elem_i <- elem_i + 1L
    }
  }
  # --------------------- handling spaces in operators ----------------------- #
  if (grepl("= +~", modelsrcw)) {
    waar <- regexpr("= +~", modelsrcw)[1]
    modelsrcw <- gsub("=( +)~", "=~\\1", modelsrcw)
    tl <- lav_parse_txtloc(modelsrc, waar)
    lav_msg_warn(gettext("splitting of '=~' deprecated"),
      tl[1L],
      footer = tl[2L]
    )
  }
  if (grepl("[^=~]~ +~", modelsrcw)) {
    waar <- regexpr("[^=~]~ +~", modelsrcw)[1]
    modelsrcw <- gsub("([^=~])~( +)~", "\\1~~\\2", modelsrcw)
    tl <- lav_parse_txtloc(modelsrc, waar + 1L)
    lav_msg_warn(gettext("splitting of '~~' deprecated"),
      tl[1L],
      footer = tl[2L]
    )
  }
  # -------------------------------------------------------------------------- #
  lavops <- gregexpr("=~|<~|~\\*~|~~|~|\\|~|==|<|>|:=|:|\\||%", modelsrcw)[[1]]
  if (lavops[1L] > -1L) {
    lavop_lengths <- attr(lavops, "match.length")
    for (i in seq_along(lavops)) {
      pfpos <- lavops[i]
      pflen <- lavop_lengths[i]
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrcw, pfpos, pfpos + pflen - 1L)
      elem_type[elem_i] <- types$lavaanoperator
      substr(modelsrcw, pfpos, pfpos + pflen - 1L) <- strrep(" ", pflen)
      elem_i <- elem_i + 1L
    }
  }
  symbols <- gregexpr("[,()/*?^']", modelsrcw)[[1L]] # f1=~x2 + 0.5 ? x3
  symbols1 <- gregexpr("[-+][^.0-9]", modelsrcw)[[1L]] # f1=~x2+x3
  symbols2 <- gregexpr("[._0-9a-df-zA-DF-Z)] *[-+][.0-9]", modelsrcw)[[1L]]
  # f1=~x2+2*x3, len-2 !
  symbols3 <- gregexpr("[^.0-9][eE] *[-+][.0-9]", modelsrcw)[[1L]]
  # f1=~xe+2*x3, len-2 !
  if (symbols1[1L] > -1L) {
    if (symbols[1L] == -1L) {
      symbols <- symbols1
    } else {
      symbols <- c(symbols, symbols1)
    }
  }
  if (symbols2[1L] > -1L) {
    symbols2_lengths <- attr(symbols2, "match.length")
    symbols2 <- symbols2 + symbols2_lengths - 2L
    if (symbols[1L] == -1L) {
      symbols <- symbols2
    } else {
      symbols <- c(symbols, symbols2)
    }
  }
  if (symbols3[1L] > -1L) {
    symbols3_lengths <- attr(symbols3, "match.length")
    symbols3 <- symbols3 + symbols3_lengths - 2L
    if (symbols[1L] == -1L) {
      symbols <- symbols3
    } else {
      symbols <- c(symbols, symbols3)
    }
  }
  if (symbols[1L] > -1L) {
    for (i in seq_along(symbols)) {
      pfpos <- symbols[i]
      substr(modelsrcw, pfpos, pfpos) <- " "
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos, pfpos)
      elem_type[elem_i] <- types$symbol
      elem_i <- elem_i + 1L
    }
  }

  numliterals <- gregexpr(
    "([ \n][-+][.0-9]|[ \n]\\.[0-9]|[ \n][0-9])[-+\\.0-9eE]*",
    paste0(" ", modelsrcw)
  )[[1]]
  if (numliterals[1L] > -1L) {
    numliteral_lengths <- attr(numliterals, "match.length") - 1L
    for (i in seq_along(numliterals)) {
      pfpos <- numliterals[i]
      pflen <- numliteral_lengths[i]
      substr(modelsrcw, pfpos, pfpos + pflen - 1L) <- strrep(" ", pflen)
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos, pfpos + pflen - 1L)
      elem_type[elem_i] <- types$numliteral
      elem_i <- elem_i + 1L
    }
  }
  identifiers <- gregexpr("[ \n][_.[:alpha:]][_.[:alnum:]]*",
                 paste0(" ", modelsrcw)
  )[[1]]
  if (identifiers[1L] > -1L) {
    identifier_lengths <- attr(identifiers, "match.length") - 1L
    for (i in seq_along(identifiers)) {
      pfpos <- identifiers[i]
      pflen <- identifier_lengths[i]
      substr(modelsrcw, pfpos, pfpos + pflen - 1L) <- strrep(" ", pflen)
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos, pfpos + pflen - 1L)
      elem_type[elem_i] <- types$identifier
      elem_i <- elem_i + 1L
    }
  }
  # check for uninterpreted chars
  wrong <- regexpr("[^\"\n ]", modelsrcw)
  if (wrong != -1L) {
    tl <- lav_parse_txtloc(modelsrc, wrong)
    lav_msg_stop(gettext("unexpected character"),
      tl[1L],
      footer = tl[2L]
    )
  }
  # remove unused elements from vectors
  elements <- which(elem_type > 0L)
  elem_pos <- elem_pos[elements]
  elem_type <- elem_type[elements]
  elem_text <- elem_text[elements]
  # order tokens
  token_order <- order(elem_pos)
  elem_pos <- elem_pos[token_order]
  elem_type <- elem_type[token_order]
  elem_text <- elem_text[token_order]

  # concatenate identifiers with only spaces in between - LDW 22/4/2024
  elem_i <- length(elem_pos)
  concatenated <- FALSE
  while (elem_i > 1L) {
    if (any(elem_type[elem_i] == c(types$identifier, types$numliteral)) &&
      elem_type[elem_i - 1L] == types$identifier) {
      spaces_between <- elem_pos[elem_i] - elem_pos[elem_i - 1L] -
        length(elem_text[elem_i - 1L])
      elem_text[elem_i - 1L] <- paste0(
        elem_text[elem_i - 1L],
        strrep(" ", spaces_between),
        elem_text[elem_i]
      )
      elem_type[elem_i] <- 0L
      concatenated <- TRUE
    }
    elem_i <- elem_i - 1L
  }
  if (concatenated) { # remove items with type 0
    elements <- which(elem_type > 0L)
    elem_pos <- elem_pos[elements]
    elem_type <- elem_type[elements]
    elem_text <- elem_text[elements]
  }

  # to set formula number
  elem_formula_number <- rep(0L, length(elem_type))
  frm_number <- 1L
  frm_hasefa <- FALSE
  frm_lastplus <- FALSE
  frm_incremented <- FALSE
  for (i in seq_along(elem_type)) {
    elem_formula_number[i] <- frm_number
    if (elem_type[i] == types$identifier && elem_text[i] == "efa") {
      frm_hasefa <- TRUE
    }
    if (any(elem_text[i] ==
      c("+", "*", "=~", "-", "<~", "~*~", "~~", "~", "|~", "|", "%"))) {
      if (frm_incremented) {
        frm_number <- frm_number - 1L
        elem_formula_number[i] <- frm_number
        frm_incremented <- FALSE
      }
      frm_lastplus <- TRUE
    } else {
      if (any(elem_type[i] == c(
        types$stringliteral, types$identifier, types$numliteral,
        types$stringliteral, types$symbol
      ))) {
        frm_lastplus <- FALSE
      }
      if (i > 1 && elem_type[i] != types$newline &&
        elem_type[i - 1L] == types$lavaanoperator) {
        frm_hasefa <- FALSE
      }
    }
    if (elem_type[i] == types$newline) {
      if (i > 1 && elem_type[i - 1L] != types$newline) { # ignore multiple nl's
        if (!frm_hasefa && !frm_lastplus) {
          frm_number <- frm_number + 1L
          frm_incremented <- TRUE
        } else {
          frm_hasefa <- FALSE
        }
      }
    } else {
      frm_incremented <- FALSE
    }
  }
  list(
    elem_pos = elem_pos, elem_type = elem_type,
    elem_text = elem_text, elem_formula_number = elem_formula_number
  )
}

# ------------------------ lav_parse_tokens_formulas ------------------------- #
# function to group the modellist tokens in 'mono' formulas.
# mono means that the terms (for formulas other than blocks and constraints)
#   are split in separate formulas, e.g.
#   a1 + a2 =~ b1 + b2  becomes
#     /  a1 =~ b1
#     |  a1 =~ b2
#     |  a2 =~ b1
#     \  a2 =~ b2
# newlines are removed
# the function returns a list of formulas
# ---------------------------------------------------------------------------- #
lav_parse_tokens_formulas <- function(modellist, modelsrc, types) {
  real_operators <- c("=~", "<~", "~*~", "~~", "~", "|~", "|", "%")
  welke <- modellist$elem_type != types$newline
  formula_numbers <- unique(modellist$elem_formula_number[welke])
  formulas <- lapply(formula_numbers, function(s) {
    welkenu <- modellist$elem_formula_number == s & welke
    list(
      elem_pos = modellist$elem_pos[welkenu],
      elem_type = modellist$elem_type[welkenu],
      elem_text = modellist$elem_text[welkenu]
    )
  })
  maxnum <- length(formula_numbers) + sum(modellist$elem_text == "+")
  outval <- vector(mode = "list", length = maxnum)
  realnum <- 0L
  for (i in seq_along(formulas)) {
    formul1 <- formulas[[i]]
    opi <- which(formul1$elem_type == types$lavaanoperator)
    nelem <- length(formul1$elem_type)
    if (length(opi) == 0L) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1])
      lav_msg_stop(gettext("formula without valid operator"),
        tl[1L],
        footer = tl[2L]
      )
    }
    if (length(opi) > 1L) opi <- opi[1] # only first operator taken
    if (any(formul1$elem_text[opi] == real_operators) &&
      sum(formul1$elem_text == "+") > 0) {
      # check + symbols outside parentheses in left and right hand side
      lhplusjes <- integer(0)
      openparentheses <- 0L
      for (jj in seq.int(1L, opi - 1L)) {
        if (formul1$elem_text[jj] == "(") {
          openparentheses <- openparentheses + 1L
          next
        }
        if (formul1$elem_text[jj] == ")") {
          openparentheses <- openparentheses - 1L
          next
        }
        if (formul1$elem_text[jj] == "+" && openparentheses == 0L) {
          lhplusjes <- c(lhplusjes, jj)
        }
      }
      lhplusjes <- c(lhplusjes, opi)
      plusjes <- integer(0)
      openparentheses <- 0L
      for (jj in seq.int(opi + 1L, nelem)) {
        if (formul1$elem_text[jj] == "(") {
          openparentheses <- openparentheses + 1L
          next
        }
        if (formul1$elem_text[jj] == ")") {
          openparentheses <- openparentheses - 1L
          next
        }
        if (formul1$elem_text[jj] == "+" && openparentheses == 0L) {
          plusjes <- c(plusjes, jj)
        }
      }
      plusjes <- c(plusjes, nelem + 1)
      # splitting lhs and rhs on '+' signs
      for (j in seq_along(lhplusjes)) {
        j0 <- 1L
        if (j > 1L) j0 <- lhplusjes[j - 1L] + 1L
        j1 <- lhplusjes[j] - 1L
        if (j1 < j0) next # skip empty parts
        for (k in seq_along(plusjes)) {
          k0 <- opi + 1L
          k1 <- plusjes[k] - 1L
          if (k > 1L) k0 <- plusjes[k - 1L] + 1L
          if (k1 < k0) next # skip empty parts
          welke <- c(seq.int(j0, j1), opi, seq.int(k0, k1))
          realnum <- realnum + 1L
          outval[[realnum]] <- lav_parse_sublist(formul1, welke)
        }
      }
    } else {
      realnum <- realnum + 1L
      outval[[realnum]] <- formul1
    }
  }
  outval[seq_len(realnum)]
}
# ------------------------ lav_parse_check_name ------------------------ #
# checks if an element of the elem_text member in a list is a valid r-name
# ---------------------------------------------------------------------------- #
lav_parse_check_name <- function(formul1, ind, modelsrc) {
  # allow spaces, LDW 22/4/2024
  testitem <- gsub(" ", "_", formul1$elem_text[ind], fixed = TRUE)

  if (make.names(testitem) != testitem) {
    tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[ind])
    lav_msg_stop(
      gettext("identifier is either a reserved word (in R) or
              contains an illegal character"),
      tl[1L],
      footer = tl[2L]
    )
  }
  invisible(NULL)
}
# ------------------------ lav_parse_modifier ---------------------------- #
# The function takes a list with tokens belonging to a single 'mono' lavaan
# formula as input. The other arguments are:
#       lhs : check for lhs or rhs modifier
#       opi : index of the lavaan operator in the list-items
#  modelsrc : the model source string (for error messages and warnings)
#     types : the types of tokens
#       rme : index of last element of modifier in formula (*)
#   rmeprev : index of first element of modifier in formula - 1L (*)
# The function return the modifier detected as element of a list
#  with name the modifier type (efa, fixed, start, label, lower, upper, prior or
#   rv) and value an array of values (length > 1 if vector via c(...)) for the
#   modifier value.
# (*) if rme > remprev the rhs is limited to the elements with index
#     rmeprev+1:rme, this is to support multiple modifiers for the same element.
# An error message is produced when no modifier can be determined.
# ---------------------------------------------------------------------------- #
lav_parse_modifier <- function(formul1, lhs, opi, modelsrc, types,
                                   rme = 0L, rmeprev = 0L, modenv) {
  if (rme > rmeprev) {
    welke <- c(seq.int(1L, opi), seq.int(rmeprev + 1L, rme),
               length(formul1$elem_type))
    formul1 <- lav_parse_sublist(formul1, welke)
  }
  nelem <- length(formul1$elem_type)
  # remove unnecessary parentheses (one element between parentheses, previous
  # no identifier)
  check_more <- TRUE
  while (check_more && nelem > 4L) {
    check_more <- FALSE
    for (par.i in seq.int(3L, nelem - 1L)) {
      if (formul1$elem_text[par.i - 1L] == "(" &&
        formul1$elem_text[par.i + 1L] == ")" &&
        formul1$elem_type[par.i - 2L] != types$identifier) {
        formul1$elem_type[par.i - 1L] <- 0L
        formul1$elem_type[par.i + 1L] <- 0L
        check_more <- TRUE
      }
    }
    if (check_more) {
      formul1 <- lav_parse_sublist(formul1, which(formul1$elem_type > 0))
      nelem <- length(formul1$elem_type)
    }
  }
  if (lhs) {
    # modifier on left hand side
    # only 1 possibility : efa ( expression-resulting-in-char ) *
    #                                        identifier operator ... (rhs) ...
    if (formul1$elem_text[1L] == "efa" &&
      formul1$elem_text[2L] == "(" &&
      formul1$elem_text[opi - 3L] == ")" &&
      formul1$elem_text[opi - 2L] == "*") {
      temp <- lav_parse_eval_r_expression(formul1, 3L, opi - 4L, types)
      if (is.character(temp) && temp[1] != "_error_") {
        return(list(efa = temp))
      }
    }
    tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1L])
    lav_msg_stop(gettext("invalid left hand side modifier"),
      tl[1L],
      footer = tl[2L]
    )
  } else {
    getmodifier <- function(s) {
      if (s %in% c("c", "t")) {
        tmp <- s
        attr(tmp, "tiepe") <- "label"
        return(tmp)
      }
      v <- all.vars(str2expression(s))
      for (v1 in v) {
        assign(v1, v1, modenv)
      }
      tmp <- eval(str2expression(s), modenv)
      rm(list = v, pos = modenv)
      if (is.null(attr(tmp, "tiepe"))) {
        if (is.numeric(tmp) || is.logical(tmp)) {
          attr(tmp, "tiepe") <- "fixed"
        } else {
          attr(tmp, "tiepe") <- "label"
        }
      }
      tmp
    }
    strings <- vapply(seq.int(opi + 1L, nelem - 2L), function(x) {
      if (formul1$elem_type[x] == types$stringliteral) {
        paste0('"', formul1$elem_text[x], '"')
      } else {
        formul1$elem_text[x]
      }
    }, "")
    txt <- paste(strings, collapse = "")
    if (formul1$elem_text[nelem - 1L] == "?") {
      formul1$elem_text[nelem - 1L] <- "*"
      txt <- paste0("start(", txt, ")")
    }
    if (formul1$elem_text[nelem - 1L] != "*") {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[nelem - 1L])
      lav_msg_stop(gettext("invalid modifier symbol (should be '*' or '?')"),
        tl[1L],
        footer = tl[2L]
      )
    }
    modifier <- try(getmodifier(txt), silent = TRUE)
    if (inherits(modifier, "try-error")) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi + 1L])
      lav_msg_stop(gettext("invalid modifier specification"),
        tl[1L],
        footer = tl[2L]
      )
    }
    if (attr(modifier, "tiepe") %in% c("label", "rv")) {
      if (!is.character(modifier)) {
        tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi + 1L])
        lav_msg_stop(gettext("invalid label modifier (should be character)"),
          tl[1L],
          footer = tl[2L]
        )
      }
    } else {
      if (!is.numeric(modifier) && !all(is.na(modifier))) {
        tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi + 1L])
        lav_msg_stop(gettext("invalid numeric modifier"),
          tl[1L],
          footer = tl[2L]
        )
      }
    }
    modifierlist <- list(as.vector(modifier))
    names(modifierlist) <- attr(modifier, "tiepe")
    modifierlist
  }
}

# -------------------- main parsing function --------------------------------- #
lav_parse_model_string <- function(model_syntax = "", as_data_frame = FALSE) {
  # replace 'strange' tildes (in some locales) (new in 0.6-6)
  modelsrc <- gsub(
    pattern = "\u02dc",
    replacement = "~",
    paste(unlist(model_syntax), "", collapse = "\n"),
    fixed = TRUE
  )
  # check for empty, whitespace-only or comment-only model syntax
  if (length(model_syntax) == 0L ||
      !grepl("[^[:space:];]", gsub("[#!][^\n]*", "", modelsrc))) {
    lav_msg_stop(gettext("Model syntax is empty."))
  }
  hashstring <- paste0("mdl_", lav_char2hash(paste0(modelsrc, as_data_frame)))
  if (exists(hashstring, envir = lavaan_cache_env)) {
    return(get(hashstring, envir = lavaan_cache_env))
  }
  modenv <- new.env()
assign("label", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "label"
  x
}, modenv)
assign("start", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "start"
  x
}, modenv)
assign("fixed", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "fixed"
  x
}, modenv)
assign("upper", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "upper"
  x
}, modenv)
assign("lower", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "lower"
  x
}, modenv)
assign("rv", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "rv"
  x
}, modenv)
assign("prior", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "prior"
  x
}, modenv)
assign("equal", function(...) {
  x <- do.call("c", args = list(...))
  attr(x, "tiepe") <- "label"
  x
}, modenv)
  types <- lav_create_enum(c(
    "identifier", "numliteral", "stringliteral",
    "symbol", "lavaanoperator", "newline"
  ))
  modellist <- lav_parse_text_tokens(modelsrc, types)
  if (lav_debug()) {
    print(data.frame(
      pos = modellist$elem_pos,
      type = types$enum_names[modellist$elem_type],
      text = modellist$elem_text,
      formula = modellist$elem_formula_number
    ))
  }
  formulalist <- lav_parse_tokens_formulas(modellist, modelsrc, types)
  #---- analyse syntax formulas and put in flat_-----
  max_mono_formulas <- length(formulalist)
  flat_lhs <- character(max_mono_formulas)
  flat_op <- character(max_mono_formulas)
  flat_rhs <- character(max_mono_formulas)
  flat_rhs_mod_idx <- integer(max_mono_formulas)
  flat_block <- integer(max_mono_formulas) # keep track of groups using ":" opr
  flat_fixed <- character(max_mono_formulas) # only for display purposes!
  flat_start <- character(max_mono_formulas) # only for display purposes!
  flat_lower <- character(max_mono_formulas) # only for display purposes!
  flat_upper <- character(max_mono_formulas) # only for display purposes!
  flat_label <- character(max_mono_formulas) # only for display purposes!
  flat_prior <- character(max_mono_formulas)
  flat_efa <- character(max_mono_formulas)
  flat_rv <- character(max_mono_formulas)
  flat_idx <- 0L
  mod_idx <- 0L
  constraints <- list()
  mod <- list()
  block <- 1L
  block_op <- FALSE
  if (lav_debug()) {
    cat("formula to analyse:\n")
  }
  #  operators <- c("=~", "<~", "~*~", "~~", "~", "|~", "==", "<", ">", ":=",
  #                 ":", "|", "%")
  constraint_operators <- c("==", "<", ">", ":=")
  for (s in seq_along(formulalist)) {
    formul1 <- formulalist[[s]]
    if (lav_debug()) {
      cat(vapply(seq_along(formul1$elem_type), function(j) {
        if (formul1$elem_type[j] == types$stringliteral) {
          return(dQuote(formul1$elem_text[j], FALSE))
        }
        formul1$elem_text[j]
      }, ""), "\n")
    }
    nelem <- length(formul1$elem_type)
    # where is the operator
    opi <- which(formul1$elem_type == types$lavaanoperator)
    if (length(opi) > 1L) { # if more than 1 operator skip operators ':'
      opii <- 1L
      while (formul1$elem_text[opi[opii]] == ":" && opii < length(opi)) {
        opii <- opii + 1L
      }
      opi <- opi[opii]
    }
    op <- formul1$elem_text[opi]
    if (any(op == constraint_operators)) { # ----- constraints -------
      lhs <- paste(formul1$elem_text[seq.int(1L, opi - 1L)], collapse = "")
      rhs <- paste(formul1$elem_text[seq.int(opi + 1L, nelem)], collapse = "")
      constraints <- c(
        constraints,
        list(list(
          op = op,
          lhs = lhs,
          rhs = rhs,
          user = 1L
        ))
      )
      next
    }
    if (op == ":") { # ------------------------- block start ----------------- #
      if (opi == 1L) {
        tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1])
        lav_msg_stop(
          gettext(
          "Missing block identifier. The correct syntax is: \"LHS: RHS\",
          where LHS is a block identifier (eg group or level), and RHS is
          the group/level/block number or label."
          ), tl[1L], footer = tl[2L]
        )
      }
      if (opi > 2L || all(tolower(formul1$elem_text[1]) !=
        c("group", "level", "block", "class"))) {
        tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1])
        lav_msg_stop(
          gettext(
          "Invalid block identifier. The correct syntax is: \"LHS: RHS\",
          where LHS is a block identifier (eg group or level), and RHS is
          the group/level/block number or label."
          ),
          tl[1L], footer = tl[2L]
        )
      }
      if (nelem != 3 || all(formul1$elem_type[3] !=
        c(types$stringliteral, types$identifier, types$numliteral))) {
        tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1])
        lav_msg_stop(
          gettext("syntax contains block identifier \"group\" with missing or
                invalid number/label.The correct syntax is: \"LHS: RHS\", where
                LHS is a block identifier (eg group or level), and RHS is the
                group/level/block number or label."),
          tl[1L],
          footer = tl[2L]
        )
      }
      flat_idx <- flat_idx + 1L
      flat_lhs[flat_idx] <- formul1$elem_text[1]
      flat_op[flat_idx] <- op
      flat_rhs[flat_idx] <- formul1$elem_text[3]
      flat_rhs_mod_idx[flat_idx] <- 0L
      if (block_op) {
        block <- block + 1L
      } else {
        if (flat_idx != 1) {
          tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1])
          lav_msg_warn(
            gettext("First block defined after other formulas"),
            tl[1L],
            footer = tl[2L]
          )
        }
      }
      flat_block[flat_idx] <- block
      block_op <- TRUE
      next
    }
    # ------------------ relational operators -------------------------------- #
    # warn if some identifiers contain spaces
    contsp <- which(formul1$elem_type == types$identifier &
      grepl(" ", formul1$elem_text, fixed = TRUE))
    if (length(contsp) > 0L) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[contsp[1L]])
      lav_msg_warn(
        gettextf(
          "having identifiers with spaces ('%s') is deprecated",
          formul1$elem_text[contsp[1]]
        ),
        tl[1L],
        footer = tl[2L]
      )
    }
    # checks for valid names in lhs and rhs
    lav_parse_check_name(formul1, opi - 1L, modelsrc) # valid name lhs
    for (j in seq.int(opi + 1L, nelem)) { # valid names rhs
      if (formul1$elem_type[j] == types$identifier &&
        formul1$elem_text[j] != "NA") {
        lav_parse_check_name(formul1, j, modelsrc)
      }
    }
    if (formul1$elem_type[nelem] != types$identifier &&
        (formul1$elem_type[nelem] != types$numliteral ||
         all(op != c("~", "|~", "=~")))) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[nelem])
      lav_msg_stop(
        gettext("Last element of rhs part expected to be an identifier or,
                for operator ~, |~ or =~, a numeric literal!"),
        tl[1L],
        footer = tl[2L]
      )
    }
    # intercept fixed on 0
    # replace 'lhs ~ 0' => 'lhs ~ 0 * 1' - intercept fixed on zero
    if (formul1$elem_text[nelem] == "0" && op == "~" && opi == nelem - 1L) {
      formul1$elem_type <- c(formul1$elem_type, types$symbol, types$numliteral)
      formul1$elem_text <- c(formul1$elem_text, "*", "1")
      formul1$elem_pos <- c(formul1$elem_pos, rep(formul1$elem_pos[nelem], 2))
      nelem <- length(formul1$elem_type)
    }
    # phantom latent variable
    # replace 'lhs =~ 0' => 'lhs =~ fixed(0)*lhs', 0 can be other numliteral
    #          also, lhs is last element before '=~'
    if (formul1$elem_type[nelem] == types$numliteral && op == "=~") {
      formul1$elem_type <- c(
        formul1$elem_type[seq.int(1L, nelem - 1L)], types$identifier,
        types$symbol, types$numliteral, types$symbol, types$symbol,
        types$identifier
      )
      formul1$elem_text <- c(
        formul1$elem_text[seq.int(1L, nelem - 1L)], "fixed", "(",
        formul1$elem_text[nelem], ")", "*", formul1$elem_text[opi - 1L]
      )
      formul1$elem_pos <- c(
        formul1$elem_pos[seq.int(1L, nelem - 1L)],
        rep(formul1$elem_pos[nelem], 6)
      )
      nelem <- length(formul1$elem_type)
    }
    # handling interaction variable types
    colons <- which(formul1$elem_text[seq.int(1L, nelem - 1L)] == ":" &
      formul1$elem_type[seq.int(2L, nelem)] == types$identifier)
    # check at most 1 colon
    if (length(colons) > 2L ||
        (length(colons) == 2L && (colons[1L] > opi || colons[2L] < opi))) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[colons[2]])
      lav_msg_stop(
        gettext(
        "Three-way or higher-order interaction terms (using multiple
         colons) are not supported in the lavaan syntax; please manually
         construct the product terms yourself in the data.frame, give
         them an appropriate name, and then you can use these interaction
         variables as any other (observed) variable in the model syntax."
        ), tl[1L], footer = tl[2L]
      )
    }
    if (length(colons) > 0L) {
      # collapse items around colon "a" ":" "b" => "a:b"
      formul1$elem_text[colons[1L] - 1L] <-
        paste(formul1$elem_text[seq.int(colons[1L] - 1L, colons[1L] + 1L)],
          collapse = ""
        )
      formul1 <- lav_parse_sublist(formul1,
          setdiff(seq.int(1L, nelem), seq.int(colons[1L], colons[1L] + 1L)))
      nelem <- length(formul1$elem_type)
      if (colons[1L] < opi) {
        opi <- opi - 2L # is in LHS
        if (length(colons) == 2L) colons[2L] <- colons[2L] - 2L
      }
    }
    if (length(colons) == 2L) {
      # collapse items around colon "a" ":" "b" => "a:b"
      formul1$elem_text[colons[2L] - 1L] <-
        paste(formul1$elem_text[seq.int(colons[2L] - 1L, colons[2L] + 1L)],
              collapse = ""
        )
      formul1 <- lav_parse_sublist(formul1,
          setdiff(seq.int(1L, nelem), seq.int(colons[2L], colons[2L] + 1L)))
      nelem <- length(formul1$elem_type)
    }
    # modifiers
    rhsmodelems <- which(seq_along(formul1$elem_type) > opi &
      formul1$elem_type == types$symbol &
      (formul1$elem_text == "*" | formul1$elem_text == "?"))
    for (j in seq_along(rhsmodelems)) {
      if (sum(formul1$elem_text[seq.int(opi, rhsmodelems[j])] == "(") !=
          sum(formul1$elem_text[seq.int(opi, rhsmodelems[j])] == ")"))
        rhsmodelems[j] <- 0L
    }
    rhsmodelems <- rhsmodelems[rhsmodelems != 0L]
    if (length(rhsmodelems) == 0L) rhsmodelems <- opi
    lhs <- formul1$elem_text[opi - 1L]
    rhs <- formul1$elem_text[nelem]
    for (rmei in seq_along(rhsmodelems)) {
      rme <- rhsmodelems[rmei]
      rmeprev <- if (rmei == 1L) opi else rhsmodelems[rmei - 1L]
      already <- which(flat_lhs == lhs & flat_op == op & flat_block == block &
        (flat_rhs == rhs | (flat_rhs == "" & (op == "~" | op == "|~") &
          formul1$elem_type[nelem] == types$numliteral)))
      if (length(already) == 1L) {
        idx <- already
      } else {
        flat_idx <- flat_idx + 1L
        idx <- flat_idx
        flat_lhs[idx] <- lhs
        flat_op[idx] <- op
        flat_rhs[idx] <- rhs
        flat_block[idx] <- block
        if (formul1$elem_type[nelem] == types$numliteral) {
          if (op == "~" || op == "|~") flat_rhs[idx] <- ""
        }
      }
      lhsmod <- list()
      if (opi > 2 && rmei == 1L) {
        lhsmod <- lav_parse_modifier(
          formul1,
          TRUE, opi, modelsrc, types, modenv = modenv
        )
      }
      rhsmod <- list()
      if (nelem - opi > 1) {
        rhsmod <- lav_parse_modifier(
          formul1,
          FALSE, opi, modelsrc, types, rme, rmeprev, modenv = modenv
        )
      }
      flat_fixed[idx] <- if (is.null(rhsmod$fixed)) {
        flat_fixed[idx]
      } else {
        paste(rhsmod$fixed, collapse = ";")
      }
      flat_start[idx] <- if (is.null(rhsmod$start)) {
        flat_start[idx]
      } else {
        paste(rhsmod$start, collapse = ";")
      }
      flat_label[idx] <- if (is.null(rhsmod$label)) {
        flat_label[idx]
      } else {
        paste(rhsmod$label, collapse = ";")
      }
      flat_lower[idx] <- if (is.null(rhsmod$lower)) {
        flat_lower[idx]
      } else {
        paste(rhsmod$lower, collapse = ";")
      }
      flat_upper[idx] <- if (is.null(rhsmod$upper)) {
        flat_upper[idx]
      } else {
        paste(rhsmod$upper, collapse = ";")
      }
      flat_prior[idx] <- if (is.null(rhsmod$prior)) {
        flat_prior[idx]
      } else {
        paste(rhsmod$prior, collapse = ";")
      }
      flat_efa[idx] <- if (is.null(lhsmod$efa)) {
        flat_efa[idx]
      } else {
        paste(lhsmod$efa, collapse = ";")
      }
      flat_rv[idx] <- if (is.null(rhsmod$rv)) {
        flat_rv[idx]
      } else {
        paste(rhsmod$rv, collapse = ";")
      }
      modnu <- c(lhsmod, rhsmod)
      if (length(modnu) > 0L) { # there is a modifier here
        if (length(already) == 0) { # unknown element
          mod_idx <- mod_idx + 1L
          cur_mod_idx <- mod_idx
          mod[[cur_mod_idx]] <- modnu
          flat_rhs_mod_idx[idx] <- cur_mod_idx
        } else { # known element
          if (flat_rhs_mod_idx[idx] == 0) { # not yet modifier
            mod_idx <- mod_idx + 1L
            cur_mod_idx <- mod_idx
            mod[[cur_mod_idx]] <- modnu
            flat_rhs_mod_idx[idx] <- cur_mod_idx
          } else { # use existing modifier index
            cur_mod_idx <- flat_rhs_mod_idx[idx]
            overwrite <- names(modnu)[names(modnu) %in%
              names(mod[[cur_mod_idx]])]
            if (length(overwrite) > 0) {
              tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[rmeprev + 1L])
              lav_msg_warn(
                gettextf(
                  "modifier %s specified multiple times, overwritten",
                  overwrite[1L]
                ), tl[1L],
                footer = tl[2L]
              )
            }
            mod[[cur_mod_idx]] <- modifyList(mod[[cur_mod_idx]], modnu)
          }
        }
      }
    }
    # check for variable regressed on itself
    if (formul1$elem_text[opi] == "~" &&
        formul1$elem_text[opi - 1L] == formul1$elem_text[nelem]) {
      if (!grepl("^0\\.?0*$", flat_fixed[idx])) {
        tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi])
        lav_msg_stop(
          gettext("a variable cannot be regressed on itself"),
          tl[1L],
          footer = tl[2L]
        )
      }
    }
  }
  # create flat (omit items without operator)
  filled_ones <- which(flat_op != "")
  flat <- list(
    lhs = flat_lhs[filled_ones],
    op = flat_op[filled_ones],
    rhs = flat_rhs[filled_ones],
    mod.idx = flat_rhs_mod_idx[filled_ones],
    block = flat_block[filled_ones],
    fixed = flat_fixed[filled_ones],
    start = flat_start[filled_ones],
    lower = flat_lower[filled_ones],
    upper = flat_upper[filled_ones],
    label = flat_label[filled_ones],
    prior = flat_prior[filled_ones],
    efa = flat_efa[filled_ones],
    rv = flat_rv[filled_ones]
  )
  # change op for intercepts (for convenience only)
  int_idx <- which(flat_op == "~" & flat_rhs == "")
  if (length(int_idx) > 0L) {
    flat$op[int_idx] <- "~1"
  }
  # change op for instruments (for convenience only)
  int_idx <- which(flat_op == "|~" & flat_rhs == "")
  if (length(int_idx) > 0L) {
    flat$op[int_idx] <- "|~1"
  }
  # if there are constraints that are simple lower or upper limits, put
  # them in these members, add a modifier and remove constraint
  aantal <- length(constraints)
  if (aantal > 0) {
    for (j in aantal:1) {
      if (any(flat$label == constraints[[j]]$lhs) &&
          any(constraints[[j]]$op == c("<", ">"))) {
          rhslang <- str2lang(constraints[[j]]$rhs)
          numbound <- NA_real_
          if (mode(rhslang) == "numeric") {
            numbound <- as.numeric(constraints[[j]]$rhs)
          } else {
            if (mode(rhslang) == "call") {
              if (is.numeric(tryCatch(eval(rhslang),
                                      error = function(e) "error"))) {
                numbound <- eval(rhslang)
              }
            }
          }
          if (!is.na(numbound)) {
            nrs <- which(flat$label == constraints[[j]]$lhs)
            for (nr in nrs) {
              nrm <- length(mod) + 1L
              if (flat$mod.idx[nr] > 0L) {
                nrm <- flat$mod.idx[nr]
              } else {
                flat$mod.idx[nr] <- nrm
                mod <- c(mod, list(label = constraints[[j]]$lhs))
              }
              if (constraints[[j]]$op == "<") {
                flat$upper[nr] <- as.character(numbound)
                mod[[nrm]]$upper <- numbound
              } else {
                flat$lower[nr] <- as.character(numbound)
                mod[[nrm]]$lower <- numbound
              }
            }
            constraints <- constraints[-j]
          }
        }
    }
  }
  # new in 0.6, reorder covariances here!
  flat <- lav_pt_covariance_reorder(flat)
  if (as_data_frame) {
    flat <- as.data.frame(flat, stringsAsFactors = FALSE)
  }
  # new in 0.6-4: check for 'group' within 'level'
  if (any(flat_op == ":")) {
    op_idx <- which(flat_op == ":")
    if (length(op_idx) < 2L) {
      # only 1 block identifier? this is weird -> give warning
      lav_msg_warn(gettext("syntax contains only a single block identifier!"))
    } else {
      first_block <- flat_lhs[op_idx[1L]]
      second_block <- flat_lhs[op_idx[2L]]
      if (first_block == "level" && second_block == "group") {
        lav_msg_stop(gettext("groups can not be nested within levels!"))
      }
    }
  }
  # create output
  attr(flat, "modifiers") <- mod
  attr(flat, "constraints") <- constraints
  assign(hashstring, flat, envir = lavaan_cache_env)
  flat
}

# the 'multiple group' method as described in Guttman, 1952
#
# Guttman, L. (1952). Multiple group methods for common-factor analysis,
# their basis, computation, and interpretation. Psychometrika, 17(2) 209--222
#
# YR 02 Feb 2023: - first version in lavaan, using quadprog (no std.lv yet)

lav_cfa_guttman1952 <- function(s,
                                marker_idx = NULL,
                                lambda_nonzero_idx = NULL,
                                theta = NULL, # vector!
                                theta_bounds = FALSE,
                                force_pd = FALSE,
                                zero_after_efa = FALSE,
                                quadprog = FALSE,
                                psi_mapping = FALSE,
                                nobs = 20L) { # for cutoff
  # dimensions
  nvar <- ncol(s)
  nfac <- length(marker_idx)
  stopifnot(length(theta) == nvar)

  # overview of lambda structure
  m_b <- matrix(0, nvar, nfac)
  lambda_marker_idx <- (seq_len(nfac) - 1L) * nvar + marker_idx
  m_b[lambda_marker_idx] <- 1L
  m_b[lambda_nonzero_idx] <- 1L

  # this method does not support crossloadings!
  if (any(rowSums(m_b) > 1)) {
    lav_msg_stop(gettext("the guttman1952 procedure does not support ",
                       "crossloadings; consider fabin or bentler1982 instead"))
  }

  # if we wish to keep SminTheta PD, we must keep theta within bounds
  if (force_pd) {
    theta_bounds <- TRUE
  }
  if (psi_mapping) {
    theta_bounds <- TRUE
    force_pd <- TRUE
  }

  # do we first 'clip' the theta values so they are within standard bounds?
  # (Question: do we need the 0.01 and 0.99 multipliers?)
  diag_s <- diag(s)
  if (theta_bounds) {
    # lower bound
    lower_bound <- diag_s * 0 # * 0.01
    too_small_idx <- which(theta < lower_bound)
    if (length(too_small_idx) > 0L) {
      theta[too_small_idx] <- lower_bound[too_small_idx]
    }

    # upper bound
    upper_bound <- diag_s * 1 # * 0.99
    too_large_idx <- which(theta > upper_bound)
    if (length(too_large_idx) > 0L) {
      theta[too_large_idx] <- upper_bound[too_large_idx]
    }
  }

  # compute SminTheta: S where we replace diagonal with 'communalities'
  diag_theta <- diag(theta, nvar)
  s_min_theta <- s - diag_theta
  if (force_pd) {
    lambda <- try(lav_mat_sym_diff_smallest_root(s, diag_theta),
      silent = TRUE
    )
    if (inherits(lambda, "try-error")) {
      lav_msg_warn(gettext("failed to compute lambda"))
      s_min_theta <- s - diag_theta # and hope for the best
    } else {
      cutoff <- 1 + 1 / (nobs - 1)
      if (lambda < cutoff) {
        lambda_star <- lambda - 1 / (nobs - 1)
        s_min_theta <- s - lambda_star * diag_theta
      } else {
        s_min_theta <- s - diag_theta
      }
    }
  } else {
    # at least we force the diagonal elements of SminTheta to be nonnegative
    lower_bound <- diag_s * 0.001
    too_small_idx <- which(diag(s_min_theta) < lower_bound)
    if (length(too_small_idx) > 0L) {
      diag(s_min_theta)[too_small_idx] <- lower_bound[too_small_idx]
    }
  }

  # compute covariances among 1) (corrected) variables, and
  #                           2) (corrected) sum-scores
  ys_cov <- s_min_theta %*% m_b

  # compute covariance matrix of corrected sum-scores
  # SS.COV <- t(B) %*% SminTheta %*% B
  ss_cov <- crossprod(m_b, ys_cov)

  # scaling factors
  # D.inv.sqrt <- diag(1/sqrt(diag(SS.COV)))
  d_inv_sqrt <- 1 / sqrt(diag(ss_cov))

  # factor correlation matrix
  # PHI <- D.inv.sqrt %*% SS.COV %*% D.inv.sqrt
  phi <- t(ss_cov * d_inv_sqrt) * d_inv_sqrt

  # factor *structure* matrix
  # (covariances corrected Y & corrected normalized sum-scores)
  # YS.COR <- YS.COV %*% D.inv.sqrt
  ys_cor <- t(ys_cov) * d_inv_sqrt # transposed!

  if (zero_after_efa) {
    # we initially assume a saturated LAMBDA (like EFA)
    # then, we just fix the zero-elements to zero

    mm_lambda <- t(solve(phi, ys_cor)) # = unconstrained EFA version
    # force zeroes
    mm_lambda <- mm_lambda * m_b
  } else if (quadprog) {
    # constrained version using quadprog
    # only useful if (in)equality constraints are needed (TODo)

    # PHI MUST be positive-definite
    phi <- cov2cor(lav_mat_sym_force_pd(phi,
      tol = 1e-04
    )) # option?
    dmat <- lav_mat_bdiag(rep(list(phi), nvar))
    dvec <- as.vector(ys_cor)
    eq_idx <- which(t(m_b) != 1) # these must be zero (row-wise!)
    rmat <- diag(nrow(dmat))[eq_idx, , drop = FALSE]
    bvec <- rep(0, length(eq_idx)) # optional, 0=default
    out <- try(quadprog::solve.QP(
      Dmat = dmat, dvec = dvec, Amat = t(rmat),
      meq = length(eq_idx), bvec = bvec
    ), silent = TRUE)
    if (inherits(out, "try-error")) {
      lav_msg_warn(gettext("solve.QP failed to find a solution"))
      lambda_1 <- m_b
      lambda_1[lambda_nonzero_idx] <- as.numeric(NA)
      theta_1 <- diag(rep(as.numeric(NA), nvar), nvar)
      psi <- matrix(as.numeric(NA), nfac, nfac)
      return(list(lambda = lambda_1, theta = theta_1, psi = psi))
    } else {
      mm_lambda <- matrix(out$solution, nrow = nvar, ncol = nfac, byrow = TRUE)
      # zap almost zero elements
      mm_lambda[abs(mm_lambda) < sqrt(.Machine$double.eps)] <- 0
    }
  } else {
    # default, if no (in)equality constraints
    mm_lambda <- t(solve(phi, ys_cor)) # = unconstrained EFA version
    #YS.COR0 <- YS.COR
    #YS.COR0[t(B) != 1] <- 0
    #LAMBDA <- t(YS.COR0)
  }

  # rescale LAMBDA, so that 'marker' indicator == 1
  marker_lambda <- mm_lambda[lambda_marker_idx]
  lambda_1 <- t(t(mm_lambda) * (1 / marker_lambda))

  # rescale PHI, covariance metric
  psi <- t(phi * marker_lambda) * marker_lambda

  # redo psi using ML mapping function?
  if (psi_mapping) {
    ti <- 1 / theta
    zero_theta_idx <- which(abs(theta) < 0.01) # be conservative
    if (length(zero_theta_idx) > 0L) {
      ti[zero_theta_idx] <- 1
    }

    # ML mapping function
    m <- solve(t(lambda_1) %*% diag(ti, nvar) %*% lambda_1) %*%
         t(lambda_1) %*% diag(ti, nvar)
    psi <- m %*% s_min_theta %*% t(m)
  }

  list(lambda = lambda_1, theta = theta, psi = psi)
}

# internal function to be used inside lav_optim_noniter
# return 'x', the estimated vector of free parameters
lav_cfa_guttman1952_internal <- function(lavobject = NULL, # convenience
                                         # internal slot
                                         lavmodel = NULL,
                                         lavsamplestats = NULL,
                                         lavpartable = NULL,
                                         lavdata = NULL,
                                         lavoptions = NULL,
                                         theta_bounds = TRUE,
                                         force_pd = TRUE,
                                         zero_after_efa = FALSE,
                                         quadprog = FALSE,
                                         psi_mapping = TRUE) {
  lavpta <- NULL
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))

    # extract slots
    lavmodel <- lavobject@Model
    lavsamplestats <- lavobject@SampleStats
    lavpartable <- lav_pt_set_cache(lavobject@ParTable, lavobject@pta)
    lavpta <- lavobject@pta
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
  }
  if (is.null(lavpta)) {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }

  if (missing(zero_after_efa) &&
    !is.null(lavoptions$estimator.args$zero.after.efa)) {
    zero_after_efa <- lavoptions$estimator.args$zero.after.efa
  }

  if (missing(psi_mapping) &&
    !is.null(lavoptions$estimator.args$psi.mapping)) {
    psi_mapping <- lavoptions$estimator.args$psi.mapping
  }

  if (missing(quadprog) &&
    !is.null(lavoptions$estimator.args$quadprog)) {
    quadprog <- lavoptions$estimator.args$quadprog
  }

  # no structural part!
  if (any(lavpartable$op == "~")) {
    lav_msg_stop(gettext("GUTTMAN1952 estimator only available for CFA models"))
  }
  # no BETA matrix! (i.e., no higher-order factors)
  if (!is.null(lavmodel@GLIST$beta)) {
    lav_msg_stop(gettext(
   "GUTTMAN1952 estimator not available for models that require a BETA matrix"))
  }
  # no std.lv = TRUE for now
  if (lavoptions$std.lv) {
    lav_msg_stop(gettext(
      "GUTTMAN1952 estimator not available if std.lv = TRUE"))
  }

  nblocks <- lav_pt_nblocks(lavpartable)
  stopifnot(nblocks == 1L) # for now
  b <- 1L
  sample_cov <- lavsamplestats@cov[[b]]
  nvar <- nrow(sample_cov)
  lv_names <- lavpta$vnames$lv.regular[[b]]
  nfac <- length(lv_names)
  marker_idx <- lavpta$vidx$lv.marker[[b]]
  lambda_idx <- which(names(lavmodel@GLIST) == "lambda")
  lambda_nonzero_idx <- lavmodel@m.free.idx[[lambda_idx]]
  # only diagonal THETA for now...
  # because if we have correlated residuals, we should remove the
  # corresponding variables as instruments before we estimate lambda...
  # (see MIIV)
  theta_idx <- which(names(lavmodel@GLIST) == "theta") # usually '2'
  m_theta <- lavmodel@m.free.idx[[theta_idx]]
  nondiag_idx <- m_theta[!m_theta %in% lav_mat_diag_idx(nvar)]
  if (length(nondiag_idx) > 0L) {
    lav_msg_warn(gettext(
      "this implementation of FABIN does not handle correlated residuals yet!"))
  }

  # 1. obtain estimate for (diagonal elements of) THETA
  #    for now we use Spearman per factor
  m_b <- matrix(0, nvar, nfac)
  lambda_marker_idx <- (seq_len(nfac) - 1L) * nvar + marker_idx
  m_b[lambda_marker_idx] <- 1L
  m_b[lambda_nonzero_idx] <- 1L
  theta <- numeric(nvar)
  for (f in seq_len(nfac)) {
    ov_idx <- which(m_b[, f] == 1L)
    s_fac <- sample_cov[ov_idx, ov_idx, drop = FALSE]
    theta[ov_idx] <- lav_cfa_theta_spearman(s_fac, bounds = "wide")
  }

  # 2. run Guttman1952 'Multiple Groups' algorithm
  out <- lav_cfa_guttman1952(
    s = sample_cov, marker_idx = marker_idx,
    lambda_nonzero_idx = lambda_nonzero_idx,
    theta = theta,
    # experimental
    theta_bounds = theta_bounds,
    force_pd = force_pd,
    zero_after_efa = zero_after_efa,
    quadprog = quadprog,
    psi_mapping = psi_mapping,
    #
    nobs = lavsamplestats@ntotal
  )
  mm_lambda <- out$lambda
  mm_theta <- diag(out$theta, nvar)
  mm_psi <- out$psi

  # store matrices in lavmodel@GLIST
  lavmodel@GLIST$lambda <- mm_lambda
  lavmodel@GLIST$theta <- mm_theta
  lavmodel@GLIST$psi <- mm_psi

  # extract free parameters only
  x <- lav_model_get_parameters(lavmodel)

  # apply bounds (if any)
  if (!is.null(lavpartable$lower)) {
    lower_x <- lavpartable$lower[lavpartable$free > 0]
    too_small_idx <- which(x < lower_x)
    if (length(too_small_idx) > 0L) {
      x[too_small_idx] <- lower_x[too_small_idx]
    }
  }
  if (!is.null(lavpartable$upper)) {
    upper_x <- lavpartable$upper[lavpartable$free > 0]
    too_large_idx <- which(x > upper_x)
    if (length(too_large_idx) > 0L) {
      x[too_large_idx] <- upper_x[too_large_idx]
    }
  }

  x
}

# contributed by Ed Merkle (17 Jan 2013)
# WLS version contributed by Franz Classe (March 2024)
# (adapted for inclusion in lavaan by YR)


# YR 12 Feb 2013: small changes to match the results of lav_model_gradient
#                 in the multiple group case
# YR 30 May 2014: handle 1-variable case (fixing apply in lines 56, 62, 108)
# YR 05 Nov 2015: add remove.duplicated = TRUE, to cope with strucchange in
#                 case of simple equality constraints
# YR 19 Nov 2015: if constraints have been used, compute case-wise Lagrange
#                 multipliers, and define the scores as: SC + (t(R) lambda)
# YR 05 Feb 2016: catch conditional.x = TRUE: no support (for now), until
#                 we can use the generic 0.6 infrastructure for scores,
#                 including the missing-values case
# YR 16 Feb 2016: adapt to changed @Mp slot elements; add remove.empty.cases=
#                 argument
# YR 12 Mar 2024: make lintr (more) happy; include WLS code from Franz Classe
#                 move ML-specific code to lav_sc_ml() function
# YR 26 Apr 2025: add lav_scores_gls()

lav_sc <- function(object, scaling = FALSE,
                   ignore_constraints = FALSE,
                   remove_duplicated = TRUE,
                   remove_empty_cases = TRUE) {
  stopifnot(inherits(object, "lavaan"))

  # check object
  object <- lav_object_check_version(object)

  # what if estimator is not ML or WLS?
  # avoid hard error (using stop); throw a warning, and return an empty matrix
  if (!object@Options$estimator %in% c("ML", "WLS", "GLS", "ULS", "PML")) {
    lav_msg_warn(gettext("scores only available if estimator is ML"))
    return(matrix(0, 0, 0))
  }

  # check if conditional.x = TRUE
  # (exceptions: PML supports exogenous covariates through the pairwise
  # likelihood machinery itself; for ML (single-level, continuous) the
  # lav_mvreg_* score kernels are used)
  if (object@Model@conditional.x &&
      !(object@Options$estimator == "PML" ||
        (object@Options$estimator == "ML" &&
         !object@Model@categorical &&
         object@Data@nlevels == 1L))) {
    lav_msg_stop(gettext("scores not available (yet) if conditional.x = TRUE"))
  }

  # shortcuts
  lavdata <- object@Data
  lavmodel <- object@Model
  lavsamplestats <- object@SampleStats
  lavoptions <- object@Options

  ## number variables/sample size
  # ntab <- unlist(lavsamplestats@nobs)
  ## change in 0.5-17: we keep the 'empty cases'
  ##                   and 'fill' in the scores at their 'case.idx'
  ##                   later, we remove the 'empty rows'
  # ntot <- max( object@Data@case.idx[[ object@Data@ngroups ]] )
  ntab <- unlist(lavdata@norig)
  ntot <- sum(ntab)
  npar <- lav_inspect_npar(object, ceq = FALSE)
  # with simple equality constraints (ceq.simple), the casewise scores are first
  # computed in the larger 'unco' space (one column per non-collapsed parameter,
  # matching Delta); they are collapsed to the free parameters further below
  if (lavmodel@ceq.simple.only) {
    npar <- lavmodel@nx.unco
  }

  if (object@Options$estimator == "ML") {
    moments <- fitted(object)
    score_matrix <- lav_sc_ml(
      ntab = ntab, ntot = ntot, npar = npar,
      moments = moments, lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavmodel = lavmodel, lavoptions = lavoptions, scaling = scaling
    )
  } else if (object@Options$estimator == "WLS" && lavmodel@categorical) {
    # check if ALL observed variables are ordered
    ov_names <- unlist(lavdata@ov.names)
    ov_idx <- which(lavdata@ov$name %in% ov_names)
    if (!all(lavdata@ov$type[ov_idx] == "ordered")) {
      lav_msg_stop(gettext(
        "WLS scores only available if all observed variables are ordered."))
    }

    # compute WLS scores
    score_matrix <- lav_sc_wls(
      ntab = ntab, ntot = ntot, npar = npar,
      lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavmodel = lavmodel, lavoptions = lavoptions
    )
  } else if (!lavmodel@categorical &&
             object@Options$estimator %in% c("GLS", "ULS", "WLS")) {
    # compute WLS/GLS/ULS `scores'
    score_matrix <- lav_sc_ls(
      ntab = ntab, ntot = ntot, npar = npar,
      lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavmodel = lavmodel, lavoptions = lavoptions
    )
  } else if (object@Options$estimator == "PML") {
    # casewise scores of the pairwise log-likelihood
    score_matrix <- lav_sc_pml(
      ntot = ntot, npar = npar,
      lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavmodel = lavmodel, lavimplied = object@implied,
      lavcache = object@Cache
    )
  } else {
    # should not happen
    lav_msg_fixme("this should not happen")
  }

  # sampling weights? the casewise score contributions become wt_i * s_i,
  # so that the column sums equal the (weighted) gradient (= 0 at the MLE),
  # and crossprod() yields the design-based information that is also used
  # for the (robust) standard errors. This is a no-op without weights.
  if (length(lavdata@sampling.weights) > 0L) {
    wt_full <- numeric(ntot)
    for (g in seq_len(lavdata@ngroups)) {
      wt_full[lavdata@case.idx[[g]]] <- lavdata@weights[[g]]
    }
    score_matrix <- score_matrix * wt_full
  }

  # handle empty rows
  if (remove_empty_cases) {
    # empty.idx <- which( apply(score_matrix, 1L,
    #                        function(x) sum(is.na(x))) == ncol(score_matrix) )
    empty_idx <- unlist(lapply(lavdata@Mp, "[[", "empty.idx"))
    if (length(empty_idx) > 0L) {
      score_matrix <- score_matrix[-empty_idx, , drop = FALSE]
    }
  }

  # provide column names
  colnames(score_matrix) <- names(lav_inspect_coef(object,
    type = "free", add_labels = TRUE
  ))

  # handle general EQUALITY constraints, so that the sum of the columns
  # equals zero
  # note: for ceq.simple.only models there are no explicit ceq/cin indices, but
  # the (synthesized) con.jac encodes the simple equalities in the 'unco' space;
  # projecting here makes the full (remove.duplicated = FALSE) scores respect the
  # constraints, exactly as for explicit equality constraints. This projection
  # does NOT affect the collapsed scores below, since con.jac %*% ceq.simple.K = 0
  #
  # INEQUALITY constraints are excluded from the projection (fixed July
  # 2026): at an interior solution their Lagrange multipliers are zero and
  # the estimator's scores have genuine sampling variability in those
  # directions; projecting on them (as the old code did, for ALL con.jac
  # rows) forced those score components to sum to zero and corrupted the
  # scores -- eg for any fit with bounds = "wide.zerovar", whose bounds are
  # synthesized as linear cin rows.
  if (!ignore_constraints && nrow(lavmodel@con.jac) > 0L) {
    r_full <- lavmodel@con.jac[, , drop = FALSE]
    # equality rows of con.jac (the "ceq.idx" attribute); for the
    # synthesized ceq.simple con.jac (no explicit constraints) all rows are
    # equalities. If the attribute is absent (very old objects), fall back
    # to the historical all-rows behaviour.
    ceq_rows <- attr(lavmodel@con.jac, "ceq.idx")
    if (is.null(ceq_rows)) {
      if (lavmodel@ceq.simple.only ||
          sum(lavmodel@ceq.linear.idx, lavmodel@ceq.nonlinear.idx) > 0) {
        ceq_rows <- seq_len(nrow(r_full))
      } else {
        ceq_rows <- integer(0L)
      }
    }
    if (length(ceq_rows) > 0L) {
      if (length(ceq_rows) < nrow(r_full)) {
        # drop the INEQUALITY rows BEFORE computing the multipliers (the
        # 'pre' factor is built from the full constraint jacobian and is
        # not row-separable)
        object@Model@con.jac <- r_full[ceq_rows, , drop = FALSE]
      }
      r_matrix <- object@Model@con.jac[, , drop = FALSE]
      pre <- lav_con_lambda_pre(object)
      # LAMBDA <- -1 * t(pre %*% t(score_matrix))
      # RLAMBDA <- t(t(r_matrix) %*% t(LAMBDA))
      score_matrix <- score_matrix -
        t(t(r_matrix) %*% pre %*% t(score_matrix))
    }
  }

  # handle simple equality constraints
  if (remove_duplicated && lavmodel@eq.constraints) {
    simple_flag <- lav_con_check_simple(lavmodel)
    if (simple_flag) {
      k_matrix <- lav_con_r2k(lavmodel)
      score_matrix <- score_matrix %*% k_matrix
    } else {
      lav_msg_warn(gettext(
        "remove.duplicated is TRUE, but equality constraints do not appear
        to be simple; returning full scores"))
    }
  } else if (remove_duplicated && lavmodel@ceq.simple.only) {
    # collapse the 'unco'-space scores to the free parameters: the score for a
    # free parameter is the sum of the scores of the parameters that share it
    # (ceq.simple.K is the nx.unco x nx.free duplication matrix)
    score_matrix <- score_matrix %*% lavmodel@ceq.simple.K
  }

  score_matrix
}

lavScores <- function(                                # nolint
                      object,
                      scaling = FALSE,
                      ignore_constraints = FALSE,
                      remove_duplicated = TRUE,
                      remove_empty_cases = TRUE,
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  lav_sc(object, scaling = scaling,
        ignore_constraints = ignore_constraints,
        remove_duplicated = remove_duplicated,
        remove_empty_cases = remove_empty_cases)
}

# YR 18 Jun 2026: the casewise H1 (saturated-model) scores are now computed by
#                 the (better tested) lav_mvn_sc_mu_sigma() (complete data) and
#                 lav_mvn_mi_sc_mu_sigma() (incomplete data) functions from the
#                 lav_mvnorm*.R files; this function only assembles them into the
#                 model-parameter scores via the Delta matrix. The mvn casewise
#                 score has the opposite sign of the old 'scores_h1' (so the old
#                 '-scores_h1 %*% Delta' becomes 'sc %*% Delta'), and produces
#                 vech(Sigma) with the diagonal already halved -- matching Delta.
lav_sc_ml <- function(ntab = 0L,
                      ntot = 0L,
                      npar = 0L,
                      moments = NULL,
                      lavdata = NULL,
                      lavsamplestats = NULL,
                      lavmodel = NULL,
                      lavoptions = NULL,
                      scaling = FALSE) {
  score_matrix <- matrix(NA, ntot, npar)

  # Delta matrix
  delta <- lav_model_delta(lavmodel = lavmodel)

  # rename moments
  moments_groups <- moments

  # group weights (only used if scaling = TRUE)
  group_w <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal

  for (g in 1:lavsamplestats@ngroups) {
    if (lavsamplestats@ngroups > 1) {
      moments <- moments_groups[[g]]
    }

    # conditional.x: casewise scores of the conditional (mvreg) likelihood,
    # with respect to (vec(Beta), vech(res.cov)) -- the same stat order as
    # the conditional.x Delta rows (meanstructure is always TRUE here)
    if (lavmodel@conditional.x) {
      if (!lavsamplestats@missing.flag) { # complete data
        sc <- lav_mvreg_sc_beta_sigma(
          y = lavdata@X[[g]], exo = lavdata@eXo[[g]],
          res_int = moments$res.int, res_slopes = moments$res.slopes,
          res_cov = moments$res.cov
        )
      } else { # incomplete data
        sc <- lav_mvreg_mi_sc_beta_sigma(
          y = lavdata@X[[g]], exo = lavdata@eXo[[g]],
          mp = lavdata@Mp[[g]],
          res_int = moments$res.int, res_slopes = moments$res.slopes,
          res_cov = moments$res.cov
        )
        # the (saturated) casewise score entries that refer to the
        # unobserved cells of a case are zero (see lav_sc_ml below)
        sc[is.na(sc)] <- 0
      }

      if (scaling && lavsamplestats@missing.flag) {
        sc <- group_w[g] * sc
      }

      wi <- lavdata@case.idx[[g]]
      score_matrix[wi, ] <- sc %*% delta[[g]]
      if (scaling) {
        score_matrix[wi, ] <- (-1 / ntot) * score_matrix[wi, ]
      }
      next
    }

    sigma_hat <- moments$cov
    nvar <- ncol(lavsamplestats@cov[[g]])

    # H1 (saturated) mean: model-implied if meanstructure, else sample mean
    if (lavmodel@meanstructure) {
      mu_hat <- moments$mean
    } else {
      mu_hat <- lavsamplestats@mean[[g]]
    }

    if (!lavsamplestats@missing.flag) { # complete data
      sc <- lav_mvn_sc_mu_sigma(
        y = lavdata@X[[g]], mu = mu_hat, sigma_1 = sigma_hat
      )

      # wishart likelihood: the second-order (vech(Sigma)) block uses the
      # unbiased cross-product, i.e. tcrossprod(y - mu) is scaled by n/(n-1).
      # This rescales the vech(Sigma) columns of the casewise score, and adds a
      # (constant) shift proportional to vech(Sigma.inv).
      if (lavoptions$likelihood == "wishart") {
        nobs1 <- lavsamplestats@nobs[[g]] / (lavsamplestats@nobs[[g]] - 1)
        j2 <- matrix(1, nvar, nvar)
        diag(j2) <- 0.5
        d_row <- lav_mat_vech(-j2 * chol2inv(chol(sigma_hat)))
        sig_idx <- seq.int(nvar + 1L, ncol(sc))
        sc[, sig_idx] <- nobs1 * sc[, sig_idx, drop = FALSE] -
          matrix((nobs1 - 1) * d_row, nrow(sc), length(d_row), byrow = TRUE)
      }
    } else { # incomplete data
      sc <- lav_mvn_mi_sc_mu_sigma(
        y = lavdata@X[[g]], mp = lavdata@Mp[[g]],
        mu = mu_hat, sigma_1 = sigma_hat
      )
      # lav_mvn_mi_sc_mu_sigma() leaves NA in the score entries that refer to
      # the *unobserved* cells of a case; the (saturated) casewise score for a
      # missing variable is zero, so we zero them here before mapping through
      # Delta (the old implementation initialised these entries to 0).
      sc[is.na(sc)] <- 0
    } # missing

    # no mean structure: drop the (leading) mu columns
    if (!lavmodel@meanstructure) {
      sc <- sc[, -seq_len(nvar), drop = FALSE]
    }

    # scaling? (matches the historical behaviour: group weights are applied to
    # the incomplete-data scores only, then everything is scaled by -1/ntot)
    if (scaling && lavsamplestats@missing.flag) {
      sc <- group_w[g] * sc
    }

    wi <- lavdata@case.idx[[g]]
    score_matrix[wi, ] <- sc %*% delta[[g]]
    if (scaling) {
      score_matrix[wi, ] <- (-1 / ntot) * score_matrix[wi, ]
    }
  } # g

  score_matrix
}

# casewise scores of the PAIRWISE log-likelihood (estimator = "PML"),
# YR July 2026. The chain is identical to the first-order information for
# PML (see lav_model_h1_info_firstorder()): lav_pml_dploglik_dimplied()
# provides the casewise derivatives with respect to the implied statistics
# (thresholds/means, slopes, variances, correlations), and the Delta matrix
# maps them to the model parameters:
#   sc_i(theta) = dlogPL_i/dtheta' = dlogPL_i/dimplied' %*% Delta
# The scores are in +logPL convention (they sum to ~zero at the PML
# solution), one row per (original) case, placed at case.idx -- exactly as
# lav_sc_ml(). Consequently crossprod(sc)/N reproduces the first-order
# (unit) information, and the huber.white sandwich built from these scores
# reproduces the default robust PML standard errors.
# Supported: complete data ("listwise") and missing = "available.cases"
# (the univariate part is included in the casewise scores); exogenous
# covariates go through the pairwise machinery itself (conditional.x).
# missing = "doubly.robust" is NOT covered (its correction terms are not
# part of lav_pml_dploglik_dimplied()), matching the first-order
# information code.
lav_sc_pml <- function(ntot = 0L,
                       npar = 0L,
                       lavdata = NULL,
                       lavsamplestats = NULL,
                       lavmodel = NULL,
                       lavimplied = NULL,
                       lavcache = NULL) {
  if (lavdata@missing == "doubly.robust") {
    lav_msg_stop(gettext(
      "casewise PML scores are not available (yet) for
       missing = \"doubly.robust\""))
  }
  score_matrix <- matrix(NA, ntot, npar)

  # Delta matrix
  delta <- lav_model_delta(lavmodel = lavmodel)

  # model-implied statistics
  if (is.null(lavimplied) || length(lavimplied) == 0L) {
    lavimplied <- lav_model_implied(lavmodel)
  }

  for (g in 1:lavsamplestats@ngroups) {
    if (lavmodel@conditional.x) {
      sigma_1 <- lavimplied$res.cov[[g]]
      mu <- lavimplied$res.mean[[g]]
      if (is.null(mu)) {
        mu <- lavimplied$res.int[[g]]
      }
      th <- lavimplied$res.th[[g]]
      pi0 <- lavimplied$res.slopes[[g]]
      exo <- lavdata@eXo[[g]]
    } else {
      sigma_1 <- lavimplied$cov[[g]]
      mu <- lavimplied$mean[[g]]
      th <- lavimplied$th[[g]]
      pi0 <- NULL
      exo <- NULL
    }

    sc <- lav_pml_dploglik_dimplied(
      sigma_hat = sigma_1,
      mu_hat = mu,
      th = th,
      th_idx = lavmodel@th.idx[[g]],
      num_idx = lavmodel@num.idx[[g]],
      x = lavdata@X[[g]],
      exo = exo,
      wt = NULL,
      pi0 = pi0,
      lavcache = lavcache[[g]],
      missing = lavdata@missing,
      scores = TRUE,
      negative = FALSE
    )

    wi <- lavdata@case.idx[[g]]
    score_matrix[wi, ] <- sc %*% delta[[g]]
  } # g

  score_matrix
}

# this function is based on code originally written by Franz Classe (Munich)
# for categorical data only!
lav_sc_wls <- function(ntab = 0L,
                       ntot = 0L,
                       npar = 0L,
                       lavdata = NULL,
                       lavsamplestats = NULL,
                       lavmodel = NULL,
                       lavoptions = NULL) {
  # internal function
  do_dummy_single_var <- function(x_1, lv, ntot, num) {
    xd <- matrix(NA, nrow = ntot, ncol = lv[num] - 1)
    x <- x_1[, num]
    minx <- min(x)
    categ <- minx - 1
    v <- 1
    while (categ < lv[num] - 1) {
      categ <- categ + 1
      xd[, v] <- ifelse(x > categ, 1, 0)
      v <- v + 1
    }

    xd
  }

  # container for scores
  score_matrix <- matrix(NA, ntot, npar)

  # Delta matrix
  delta <- lav_model_delta(lavmodel = lavmodel)

  # shortcuts
  lv <- lavdata@ov[["nlev"]]

  for (g in 1:lavsamplestats@ngroups) {
    nvar <- ncol(lavsamplestats@cov[[g]])
    x_1 <- lavdata@X[[g]]

    # convert categorical data to dummy variables
    # FIXME: skip continuous variables
    xd <- do.call(
      cbind,
      lapply(
        1:nvar,
        function(i) do_dummy_single_var(x_1, lv, ntot, i)
      )
    )

    # e1
    musd <- colMeans(xd)
    e1 <- t(t(xd) - musd)

    # e2
    mus <- colMeans(x_1)
    y_minus_mu <- t(apply(x_1, 1L, function(x) x - mus))
    s_vech <- t(apply(y_minus_mu, 1L, function(i) {
      lavaan::lav_mat_vech(tcrossprod(i), diagonal = FALSE)
    })) # s=c( (y1-mu1)(y2-mu2)....
    sigma <- colMeans(s_vech)
    e2 <- t(apply(s_vech, 1L, function(x) x - sigma))

    # e
    e <- cbind(e1, e2)

    # weight matrix
    m_w <- lavsamplestats@WLS.V[[g]]

    # combine matrices
    wi <- lavdata@case.idx[[g]]
    score_matrix[wi, ] <- t(t(delta[[g]]) %*% m_w %*% t(e))
  } # g

  score_matrix
}


# wls/gls/uls (continuous data only)
lav_sc_ls <- function(ntab = 0L,
                      ntot = 0L,
                      npar = 0L,
                      lavdata = NULL,
                      lavsamplestats = NULL,
                      lavmodel = NULL,
                      lavoptions = NULL) {

  # container for scores
  score_matrix <- matrix(NA, ntot, npar)

  # estimator
  estimator <- lavoptions$estimator

  # Delta matrix
  delta <- lav_model_delta(lavmodel = lavmodel)

  # implied stats
  implied <- lav_model_implied(lavmodel)

  for (g in 1:lavsamplestats@ngroups) {
    nvar <- ncol(lavsamplestats@cov[[g]])
    nobs <- lavsamplestats@nobs[[g]]
    y <- lavdata@X[[g]]

    # center (not using model-implied!)
    yc <- t(t(y) - colMeans(y, na.rm = TRUE))

    # create Z where the rows_i contain the following elements:
    #  - Y_i (if meanstructure is TRUE)
    #  - vech(Yc_i' %*% Yc_i) where Yc_i are the residuals
    idx1 <- lav_mat_vech_col_idx(nvar)
    idx2 <- lav_mat_vech_row_idx(nvar)
    if (lavmodel@meanstructure) {
      z <- cbind(y, yc[, idx1, drop = FALSE] * yc[, idx2, drop = FALSE])
    } else {
      z <- (yc[, idx1, drop = FALSE] * yc[, idx2, drop = FALSE])
    }

    # model-based sample statistics
    if (lavmodel@meanstructure) {
      sigma <- c(as.numeric(implied$mean[[g]]),
                 lav_mat_vech(implied$cov[[g]]))
    } else {
      sigma <- lav_mat_vech(implied$cov[[g]])
    }

    # adjust sigma for N-1, so that colMeans(scores) == gradient
    # (not for the means)
    if (lavmodel@meanstructure) {
      z[, -seq_len(nvar)] <-
                z[, -seq_len(nvar), drop = FALSE] * nobs / (nobs - 1)
    } else {
      z <- z * nobs / (nobs - 1)
    }

    # compute Zc
    zc <- t(t(z) - sigma)

    # combine matrices
    wi <- lavdata@case.idx[[g]]
    if (estimator == "ULS") {
      # identity weight matrix
      score_matrix[wi, ] <- zc %*% delta[[g]]
    } else if (estimator == "GLS" &&
      is.null(lavsamplestats@WLS.V[[g]])) {
      # GLS: WLS.V is no longer pre-computed (0.7-1); stream
      # WLS.V %*% Delta without forming the matrix
      v11_scale <- 1.0
      if (isTRUE(lavmodel@estimator.args$gls.v11.mplus) &&
        lavmodel@meanstructure && lavdata@data.type == "full") {
        v11_scale <- lavsamplestats@nobs[[g]] /
          (lavsamplestats@nobs[[g]] - 1)
      }
      wd <- lav_samp_wls_v_nt_prod(
        m_icov = lavsamplestats@icov[[g]],
        x = delta[[g]],
        meanstructure = lavmodel@meanstructure,
        fixed_x = lavmodel@fixed.x,
        x_idx = lavsamplestats@x.idx[[g]],
        v11_scale = v11_scale
      )
      score_matrix[wi, ] <- zc %*% wd
    } else {
      m_w <- lavsamplestats@WLS.V[[g]]
      score_matrix[wi, ] <- zc %*% m_w %*% delta[[g]]
    }
  } # g

  score_matrix
}

# estfun method for the sandwich package (thin forwarder; see lav_alias
# in 00alias.R)
estfun.lavaan <- lav_alias("lavScores")        # nolint

lav_pt_flat <- function(flat = NULL,
                              blocks = "group",
                              block_id = NULL,
                              meanstructure = FALSE,
                              int_ov_free = FALSE,
                              int_lv_free = FALSE,
                              orthogonal = FALSE,
                              orthogonal_y = FALSE,
                              orthogonal_x = FALSE,
                              orthogonal_efa = FALSE,
                              std_lv = FALSE,
                              correlation = FALSE,
                              composites = TRUE,
                              composites_cov_free = FALSE,
                              conditional_x = FALSE,
                              fixed_x = TRUE,
                              parameterization = "delta",
                              auto_fix_first = FALSE,
                              marker = NULL,
                              auto_fix_single = FALSE,
                              auto_var = FALSE,
                              auto_cov_lv_x = FALSE,
                              auto_cov_x = FALSE,
                              auto_cov_y = FALSE,
                              auto_th = FALSE,
                              auto_delta = FALSE,
                              auto_efa = FALSE,
                              var_table = NULL,
                              group_equal = NULL,
                              group_w_free = FALSE,
                              ngroups = 1L,
                              nthresholds = NULL,
                              ov_names_x_block = NULL) {
  categorical <- FALSE

  # auto.cov.x implies auto.cov.lv.x
  if (auto_cov_x) {
    auto_cov_lv_x <- TRUE
  }

  ### tmp.default elements: parameters that are typically not specified by
  ###                   users, but should typically be considered,
  ###                   either free or fixed

  # extract `names' of various types of variables:
  lv_names <- lav_pt_vnames(flat, type = "lv") # latent variables
  # lv.names.r   <- lav_pt_vnames(FLAT, type="lv.regular")
  # regular latent variables
  if (composites) {
    lv_names_f <- character(0L)
    lv_names_c <- lav_pt_vnames(flat, type = "lv.composite")
    ov_ind_c <- lav_pt_vnames(flat, type = "ov.cind")
    lv_names_noc <- lv_names[!lv_names %in% lv_names_c]
  } else {
    lv_names_c <- character(0L)
    ov_ind_c <- character(0L)
    lv_names_f <- lav_pt_vnames(flat, type = "lv.formative")
    lv_names_noc <- lv_names
  }

  # formative latent variables
  ov_names <- lav_pt_vnames(flat, type = "ov")
  # observed variables
  ov_names_x <- lav_pt_vnames(flat, type = "ov.x")
  # exogenous x covariates
  lv_names_int <- lav_pt_vnames(flat, type = "lv.interaction")
  # lv interactions

  if (is.null(ov_names_x_block)) {
    ov_names_x_block <- ov_names_x
  }
  ov_names_nox <- lav_pt_vnames(flat, type = "ov.nox")
  lv_names_x <- lav_pt_vnames(flat, type = "lv.x") # exogenous lv
  ov_names_y <- lav_pt_vnames(flat, type = "ov.y") # dependent ov
  lv_names_y <- lav_pt_vnames(flat, type = "lv.y") # dependent lv
  lv_names_efa <- lav_pt_vnames(flat, type = "lv.efa")
  # lvov.names.y <- c(ov.names.y, lv.names.y)
  lvov_names_y <- c(lv_names_y, ov_names_y)

  # get 'ordered' variables, either from FLAT or var_table
  ov_names_ord1 <- lav_pt_vnames(flat, type = "ov.ord")
  # check if we have "|" for exogenous variables
  if (length(ov_names_ord1) > 0L) {
    idx <- which(ov_names_ord1 %in% ov_names_x)
    if (length(idx) > 0L) {
      lav_msg_warn(gettext("thresholds are defined for exogenous variables:"),
                   lav_msg_view(ov_names_ord1[idx], "none"))
    }
  }

  # check data
  if (!is.null(var_table)) {
    ov_names_ord2 <-
      as.character(var_table$name[var_table$type == "ordered"])
    # remove fixed.x variables
    idx <- which(ov_names_ord2 %in% ov_names_x)
    if (length(idx) > 0L) {
      ov_names_ord2 <- ov_names_ord2[-idx]
    }

    # remove those that do appear in the model syntax
    idx <- which(!ov_names_ord2 %in% ov_names)
    if (length(idx) > 0L) {
      ov_names_ord2 <- ov_names_ord2[-idx]
    }
  } else {
    ov_names_ord2 <- character(0L)
  }

  # check nthresholds, if it is a named vector
  ov_names_ord3 <- character(0L)
  if (!is.null(nthresholds)) {
    if (!is.null(var_table)) {
      lav_msg_stop(gettext(
        "the var_table and nthresholds arguments should not be used together."))
    }
    if (!is.numeric(nthresholds)) {
      lav_msg_stop(gettext("nthresholds should be a named vector of integers."))
    }
    nth_names <- names(nthresholds)
    if (!is.null(nth_names)) {
      ov_names_ord3 <- nth_names
    } else {
      # if nthresholds is just a number, all is good; otherwise it
      # should be a names vector
      if (length(nthresholds) > 1L) {
        lav_msg_warn(gettext("nthresholds must be a named vector of integers."))
      }
      # just a single number -> assume ALL y variables are ordered
      ov_names_ord3 <- ov_names_nox
    }
  }

  # final ov.names.ord
  tmp <- unique(c(ov_names_ord1, ov_names_ord2, ov_names_ord3))
  ov_names_ord <- ov_names[ov_names %in% tmp]

  # if we have the "|" in the model syntax, check the number of thresholds
  # if(!is.null(var_table) && length(ov.names.ord1) > 0L) {
  #    for(o in ov.names.ord1) {
  #        nth <- var_table$nlev[ var_table$name == o ] - 1L
  #        nth.in.partable <- sum(FLAT$op == "|" & FLAT$lhs == o)
  #        if(nth != nth.in.partable) {
  #            stop("lavaan ERROR: expected ", max(0,nth),
  #                 " threshold(s) for variable ",
  #                 sQuote(o), "; syntax contains ", nth.in.partable, "\n")
  #        }
  #    }
  # }

  if (length(ov_names_ord) > 0L) {
    categorical <- TRUE
  }

  # do we have any EFA lv's? they need special treatment if auto.efa = TRUE
  if (!is.null(flat$efa) && auto_efa) {
    lv_names_efa <- unique(flat$lhs[flat$op == "=~" &
      nchar(flat$efa) > 0L])
    # remove them from lv.names.x
    # if(length(lv.names.x) > 0L) {
    #    both.idx <- which(lv.names.x %in% lv.names.efa)
    #    if(length(both.idx) > 0L) {
    #        lv.names.x <- lv.names.x[ -both.idx ]
    #    }
    # }

    # remove them from lvov.names.y
    # if(length(lvov.names.y) > 0L) {
    #    both.idx <- which(lvov.names.y %in% lv.names.efa)
    #    if(length(both.idx) > 0L) {
    #        lvov.names.y <- lvov.names.y[ -both.idx ]
    #    }
    # }
  } else {
    lv_names_efa <- character(0)
  }

  lhs <- rhs <- character(0)

  # 1. THRESHOLDS (based on var_table)
  #    NOTE: - new in 0.5-18: ALWAYS include threshold parameters in partable,
  #            but only free them if auto.th = TRUE
  #          - [only ov.names.ord2, because ov.names.ord1 are already
  #            in tmp.user and we only need to add 'default' parameters here]
  #            (not any longer: we create them for ALL ordered var (0.6-12)
  nth <- 0L
  # if(auto.th && length(ov.names.ord2) > 0L) {
  # if(length(ov.names.ord2) > 0L) {
  if (length(ov_names_ord) > 0L) {
    # for(o in ov.names.ord2) {
    for (o in ov_names_ord) {
      if (!is.null(var_table)) {
        nth <- var_table$nlev[var_table$name == o] - 1L
      } else if (!is.null(nthresholds)) {
        if (length(nthresholds) == 1L && is.null(nth_names)) {
          nth <- nthresholds
        } else {
          # we can assume nthresholds is a named vector
          nth <- unname(nthresholds[o])
          if (is.na(nth)) {
            lav_msg_stop(gettextf("ordered variable %s not found in the
                                  named vector nthresholds.", o))
          }
        }
      }
      if (nth < 1L) next
      lhs <- c(lhs, rep(o, nth))
      rhs <- c(rhs, paste("t", seq_len(nth), sep = ""))
    }
    nth <- length(lhs)
  }

  # 2. default (residual) variances and covariances

  # a) (residual) VARIANCES (all ov's except exo, and all lv's)
  # NOTE: change since 0.5-17: we ALWAYS include the vars in the
  #       parameter table; but only if auto.var = TRUE, we set them free
  # if(auto.var) {
  ov_var <- ov_names_nox
  # auto-remove ordinal variables
  # idx <- match(ov.names.ord, ov.var)
  # if(length(idx)) ov.var <- ov.var[-idx]
  lhs <- c(lhs, ov_var, lv_names)
  rhs <- c(rhs, ov_var, lv_names)
  # }

  # b) `independent` latent variable COVARIANCES (lv.names.x)
  if (auto_cov_lv_x && length(lv_names_x) > 1L) {
    tmp <- utils::combn(lv_names_x, 2)
    lhs <- c(lhs, tmp[1, ]) # to fill upper.tri
    rhs <- c(rhs, tmp[2, ])
  }

  # b2) COVARIANCES between `independent` latent variables (lv.names.x)
  #     and observed exogenous covariates (ov.names.x) -- new in 0.7-1
  #     (not if conditional.x: the covariates are conditioned out, so these
  #      covariances cannot be represented)
  if (auto_cov_x && !conditional_x &&
      length(lv_names_x) > 0L && length(ov_names_x_block) > 0L) {
    lhs <- c(lhs, rep(lv_names_x, each = length(ov_names_x_block)))
    rhs <- c(rhs, rep(ov_names_x_block, times = length(lv_names_x)))
  }

  # c) `dependent` latent variables COVARIANCES (lv.y.idx + ov.y.lv.idx)
  if (auto_cov_y && length(lvov_names_y) > 1L) {
    tmp <- utils::combn(lvov_names_y, 2L)
    lhs <- c(lhs, tmp[1, ]) # to fill upper.tri
    rhs <- c(rhs, tmp[2, ])
  }

  # d) exogenous x covariates: VARIANCES + COVARIANCES
  if ((nx <- length(ov_names_x)) > 0L) {
    if (conditional_x) {
      # new in 0.6-12: we make a distinction between ov.names.x and
      # ov.names.x.block: we treat them 'separately' (with no covariances
      # among them)
      # but we add 'regressions' instead (see below)
      ov_names_x1 <- ov_names_x[!ov_names_x %in% ov_names_x_block]
      ov_names_x2 <- ov_names_x_block
      nx1 <- length(ov_names_x1) # split x
      nx2 <- length(ov_names_x2) # regular  x
      if (nx1 > 0L) {
        idx <- lower.tri(matrix(0, nx1, nx1), diag = TRUE)
        lhs <- c(lhs, rep(ov_names_x1, each = nx1)[idx]) # fill upper.tri
        rhs <- c(rhs, rep(ov_names_x1, times = nx1)[idx])
      }
      if (nx2 > 0L) {
        idx <- lower.tri(matrix(0, nx2, nx2), diag = TRUE)
        lhs <- c(lhs, rep(ov_names_x2, each = nx2)[idx]) # fill upper.tri
        rhs <- c(rhs, rep(ov_names_x2, times = nx2)[idx])
      }
    } else {
      idx <- lower.tri(matrix(0, nx, nx), diag = TRUE)
      lhs <- c(lhs, rep(ov_names_x, each = nx)[idx]) # fill upper.tri
      rhs <- c(rhs, rep(ov_names_x, times = nx)[idx])
    }
  }

  # e) indicators of composites: COVARIANCES
  #    but only within/intra blocks
  if ((ncx <- length(ov_ind_c)) > 0L) {
    # create W1
    w1 <- matrix(0, length(ov_ind_c), length(lv_names_c))
    c_idx <- which(flat$op == "<~")
    w1[cbind(match(flat$rhs[c_idx], ov_ind_c),
             match(flat$lhs[c_idx], lv_names_c))] <- 1
    w1w1 <- tcrossprod(w1)
    w1w1[upper.tri(w1w1, diag = TRUE)] <- 0 # keep lower.tri only
    if (ncx > 1L) {
      lhs <- c(lhs, ov_ind_c[col(w1w1)[as.logical(w1w1)]])
      rhs <- c(rhs, ov_ind_c[row(w1w1)[as.logical(w1w1)]])
    }
  }

  # f) efa latent variables COVARIANCES; only needed for 'mediators'
  #    (not in lv.names.x, not in lv.names.y) -- added in 0.6-18
  if (auto_efa && length(lv_names_efa) > 1L) {
    efa_values <- lav_pt_efa_values(flat)
    for (set in efa_values) {
      # correlated factors within each set
      this_set_lv <- unique(flat$lhs[flat$op == "=~" &
        !flat$lhs %in% lv_names_x &
        !flat$lhs %in% lv_names_y &
        flat$efa == set])
      if (length(this_set_lv) > 0L) {
        tmp <- utils::combn(this_set_lv, 2)
        lhs <- c(lhs, tmp[1, ]) # to fill upper.tri
        rhs <- c(rhs, tmp[2, ])
      }
    }
  }

  # create 'op' (thresholds come first, then variances)
  op <- rep("~~", length(lhs))
  op[seq_len(nth)] <- "|"

  # LATENT RESPONSE SCALES (DELTA)
  #    NOTE: - new in 0.5-19: ALWAYS include scaling parameters in partable,
  #            but only free them if auto.delta = TRUE (and parameterization
  #            is "delta"
  # if(auto.delta && auto.th && length(ov.names.ord) > 0L &&
  #   # length(lv.names) > 0L &&
  #   (ngroups > 1L || any(FLAT$op == "~*~") || parameterization == "theta")) {
  if (length(ov_names_ord) > 0L) {
    lhs <- c(lhs, ov_names_ord)
    rhs <- c(rhs, ov_names_ord)
    op <- c(op, rep("~*~", length(ov_names_ord)))
  }

  # same for correlation structures, but now for ALL variables (or a
  # subset, if 'correlation' is a character vector of variable names)
  if (!categorical && (isTRUE(correlation) || is.character(correlation))) {
    cor_ov <- if (is.character(correlation)) {
      ov_names[ov_names %in% correlation]
    } else {
      ov_names
    }
    lhs <- c(lhs, cor_ov)
    rhs <- c(rhs, cor_ov)
    op <- c(op, rep("~*~", length(cor_ov)))
  }

  # 3. INTERCEPTS
  if (meanstructure) {
    # if(conditional.x) {
    #    ov.int <- ov.names.nox
    # } else {
    ov_int <- ov_names
    # }
    # auto-remove ordinal variables
    # idx <- which(ov.int %in% ov.names.ord)
    # if(length(idx)) ov.int <- ov.int[-idx]

    int_lhs <- c(ov_int, lv_names)
    lhs <- c(lhs, int_lhs)
    rhs <- c(rhs, rep("", length(int_lhs)))
    op <- c(op, rep("~1", length(int_lhs)))
  }

  # 4. REGRESSIONS
  if (conditional_x) {
    # new in 0.6-12: we make a distinction between ov.names.x and
    # ov.names.x.block: we treat them 'separately' (with no covariances
    # among them)
    # but we add 'regressions' instead!
    ov_names_x1 <- ov_names_x[!ov_names_x %in% ov_names_x_block]
    ov_names_x2 <- ov_names_x_block
    nx1 <- length(ov_names_x1) # split x
    nx2 <- length(ov_names_x2) # regular  x
    if (nx1 > 0L && nx2 > 0L) {
      # add regressions for split-x ~ regular-x
      lhs <- c(lhs, rep(ov_names_x1, times = nx2))
      op <- c(op, rep("~", nx2 * nx1))
      rhs <- c(rhs, rep(ov_names_x2, each = nx1))
    }
  }

  # free group weights
  if (group_w_free) {
    lhs <- c(lhs, "group")
    rhs <- c(rhs, "w")
    op <- c(op, "%")
  }

  tmp_default <- data.frame(
    lhs = lhs, op = op, rhs = rhs,
    mod.idx = rep(0L, length(lhs)),
    stringsAsFactors = FALSE
  )


  # 4. USER: user-specified elements
  lhs <- flat$lhs
  op <- flat$op
  rhs <- flat$rhs
  mod_idx <- flat$mod.idx

  lv_names <- lav_pt_vnames(flat, type = "lv") # latent variables
  ov_names <- lav_pt_vnames(flat, type = "ov") # observed variables
  tmp_user <- data.frame(
    lhs = lhs, op = op, rhs = rhs, mod.idx = mod_idx,
    stringsAsFactors = FALSE
  )

  # check for duplicated elements in tmp.user
  tmp_tmp <- tmp_user[, 1:3]
  idx <- which(duplicated(tmp_tmp))
  if (length(idx) > 0L) {
    txt <- sapply(seq_along(idx), function(i) {
      paste(
        "    ", tmp_tmp[idx[i], "lhs"],
        tmp_tmp[idx[i], "op"],
        tmp_tmp[idx[i], "rhs"]
      )
    })
    lav_msg_warn(gettext(
      "duplicated elements in model syntax have been ignored:"),
      lav_msg_view(txt, "none"))
    tmp_user <- tmp_user[-idx, ]
  }

  # check for duplicated elements in tmp.default
  # - FIXME: can we not avoid this somehow??
  # - for example, if the user model includes 'x1 ~~ x1'
  #   or 'x1 ~ 1'
  # - remove them from tmp.default
  tmp_tmp <- rbind(tmp_default[, 1:3], tmp_user[, 1:3])
  idx <- which(duplicated(tmp_tmp, fromLast = TRUE))
  # idx should be in tmp.default
  if (length(idx)) {
    for (i in idx) {
      flat_idx <- which(tmp_user$lhs == tmp_default$lhs[i] &
        tmp_user$op == tmp_default$op[i] &
        tmp_user$rhs == tmp_default$rhs[i])
      if (length(flat_idx) != 1L) {
        cat("[lavaan DEBUG] idx in tmp.tmp: i = ", i, "\n")
        print(tmp_tmp[i, ])
        cat("[lavaan DEBUG] idx in tmp.default: i = ", i, "\n")
        print(tmp_default[i, ])
        cat("[lavaan DEBUG] flat.idx:")
        print(flat_idx)
      }
    }
    tmp_default <- tmp_default[-idx, ]
  }

  # now that we have removed all duplicated elements, we can construct
  # the tmp.list for a single group/block
  lhs <- c(tmp_user$lhs, tmp_default$lhs)
  op <- c(tmp_user$op, tmp_default$op)
  rhs <- c(tmp_user$rhs, tmp_default$rhs)
  user <- c(
    rep(1L, length(tmp_user$lhs)),
    rep(0L, length(tmp_default$lhs))
  )
  mod_idx <- c(tmp_user$mod.idx, tmp_default$mod.idx)

  # by default: everything is free!
  free <- rep(1L, length(lhs))
  ustart <- rep(as.numeric(NA), length(lhs))
  # label   <- paste(lhs, op, rhs, sep="")
  label <- rep(character(1), length(lhs))
  exo <- rep(0L, length(lhs))

  # 0a. if auto.th = FALSE, set fix the thresholds
  if (!auto_th) {
    th_idx <- which(op == "|" & user == 0L)
    free[th_idx] <- 0L
  }

  # 0b. if auto.var = FALSE, set the unspecified variances to zero
  if (!auto_var) {
    var_idx <- which(op == "~~" &
      lhs == rhs &
      !lhs %in% ov_ind_c &
      user == 0L)
    ustart[var_idx] <- 0.0
    free[var_idx] <- 0L
  } else if (length(lv_names_f) > 0L) {
    # 'formative' (residual) variances are set to zero by default
    var_idx <- which(op == "~~" &
      lhs == rhs &
      lhs %in% lv_names_f &
      user == 0L)
    ustart[var_idx] <- 0.0
    free[var_idx] <- 0L
  }

  # 0c. for the ~~ for composite indicators (the 'T' matrix):
  #     by default these are fixed to their sample (co)variances; if
  #     composites_cov_free = TRUE, they are estimated as free parameters
  #     instead (needed for multilevel models, see lav_options). Either way
  #     ustart is set to NA, so good starting values (the sample (co)variances)
  #     are filled in later by lav_start().
  if (length(ov_ind_c) > 0) {
    if (composites_cov_free) {
      # free only the (co)variances *among* composite indicators (the T block)
      var_idx <- which(op == "~~" &
                       lhs %in% ov_ind_c & rhs %in% ov_ind_c)
      ustart[var_idx] <- as.numeric(NA)
      free[var_idx] <- 1L
    } else {
      var_idx <- which(op == "~~" & lhs %in% ov_ind_c)
      ustart[var_idx] <- as.numeric(NA)
      free[var_idx] <- 0L
    }
  }

  # 0d. variances for composites: ALWAYS fixed (should be set later
  #     by lav_lisrel_comp_set_intresvar
  if (length(lv_names_c) > 0) {
    var_idx <- which(op == "~~" & lhs %in% lv_names_c & lhs == rhs)
    ustart[var_idx] <- as.numeric(NA)
    free[var_idx] <- 0L
  }

  # 0e. instruments: ALWAYS nonfree
  iv_idx <- which(op == "|~")
  ustart[iv_idx] <- 0.0
  free[iv_idx] <- 0L

  # 1. fix metric of regular latent variables
  if (std_lv) {
    # fix metric by fixing the variance of the latent variable
    # (composites are excluded: their variance is determined by W'TW and
    #  cannot be fixed here; they are scaled via their first weight below)
    lv_var_idx <- which(op == "~~" &
      lhs %in% lv_names_noc & lhs == rhs)
    ustart[lv_var_idx] <- 1.0
    free[lv_var_idx] <- 0L
  }
  if (auto_efa && length(lv_names_efa) > 0L) {
    # fix lv variances of efa blocks to unity
    lv_var_idx <- which(op == "~~" &
      lhs %in% lv_names_efa & lhs == rhs)
    ustart[lv_var_idx] <- 1.0
    free[lv_var_idx] <- 0L
  }
  if (auto_fix_first) {
    # fix metric by fixing the loading of the first indicator
    # (but not for efa factors)
    #
    # if 'marker' is provided (a named vector lv -> indicator), fix the
    # loading of that indicator instead of the first one; this is used by
    # the bad.marker.crit mechanism to switch to another marker if the first
    # indicator turns out to be a poor item (see lav_pt_marker_adapt())
    mm_idx <- which(op == "=~" & !(lhs %in% lv_names_efa))
    if (is.null(marker)) {
      first_idx <- mm_idx[which(!duplicated(lhs[mm_idx]))]
      ustart[first_idx] <- 1.0
      free[first_idx] <- 0L
    } else {
      for (lv in unique(lhs[mm_idx])) {
        lv_rows <- mm_idx[lhs[mm_idx] == lv]
        marker_row <- lv_rows[1L] # default: first indicator
        if (lv %in% names(marker) && !is.na(marker[[lv]])) {
          tmp_row <- lv_rows[rhs[lv_rows] == marker[[lv]]]
          if (length(tmp_row) > 0L) {
            marker_row <- tmp_row[1L]
          }
        }
        ustart[marker_row] <- 1.0
        free[marker_row] <- 0L
      }
    }
  }
  # composites always need a scale. Unlike common factors, a composite's
  # variance is determined within the model (= W'TW) and cannot be fixed via
  # the std.lv mechanism above. Therefore composites default to marker scaling
  # (first weight = 1) under BOTH auto.fix.first and std.lv. Users who want a
  # different marker can free the first weight explicitly (e.g. NA*x1 + 1*x2).
  #
  # if 'marker' is provided (a named vector lv -> indicator), fix the weight
  # of that indicator instead of the first one; this is used by the
  # bad.marker.crit mechanism to switch to another marker if the implied
  # weight of the first indicator is (near) zero (see lav_pt_marker_adapt())
  if (composites && length(lv_names_c) > 0L && (auto_fix_first || std_lv)) {
    mm_idx <- which(op == "<~")
    if (is.null(marker)) {
      first_idx <- mm_idx[which(!duplicated(lhs[mm_idx]))]
      ustart[first_idx] <- 1.0
      free[first_idx] <- 0L
    } else {
      for (lv in unique(lhs[mm_idx])) {
        lv_rows <- mm_idx[lhs[mm_idx] == lv]
        marker_row <- lv_rows[1L] # default: first indicator
        if (lv %in% names(marker) && !is.na(marker[[lv]])) {
          tmp_row <- lv_rows[rhs[lv_rows] == marker[[lv]]]
          if (length(tmp_row) > 0L) {
            marker_row <- tmp_row[1L]
          }
        }
        ustart[marker_row] <- 1.0
        free[marker_row] <- 0L
      }
    }
  }

  # 2. fix residual variance of single indicators to zero
  if (auto_var && auto_fix_single) {
    mm_idx <- which(op == "=~")
    tmp_t <- table(lhs[mm_idx])
    if (any(tmp_t == 1L)) {
      # ok, we have a LV with only a single indicator
      lv_names_single <- names(tmp_t)[tmp_t == 1L]
      # get corresponding indicator if unique
      lhs_mm <- lhs[mm_idx]
      rhs_mm <- rhs[mm_idx]
      single_ind <- rhs_mm[which(lhs_mm %in% lv_names_single &
        lhs_mm != rhs_mm & # exclude phantom
        !(duplicated(rhs_mm) |
          duplicated(rhs_mm, fromLast = TRUE)))]
      # is the indicator unique?
      if (length(single_ind) > 0L) {
        var_idx <- which(op == "~~" & lhs %in% single_ind &
          rhs %in% single_ind &
          lhs == rhs &
          user == 0L)
        ustart[var_idx] <- 0.0
        free[var_idx] <- 0L
      }
    }
  }

  # 3. orthogonal = TRUE?
  if (orthogonal) {
    lv_cov_idx <- which(op == "~~" &
      lhs %in% lv_names &
      rhs %in% lv_names &
      lhs != rhs &
      user == 0L)
    ustart[lv_cov_idx] <- 0.0
    free[lv_cov_idx] <- 0L
  }
  # 3b. orthogonal.y = TRUE?
  if (orthogonal_y) {
    lv_cov_idx <- which(op == "~~" &
      lhs %in% lv_names_y &
      rhs %in% lv_names_y &
      lhs != rhs &
      user == 0L)
    ustart[lv_cov_idx] <- 0.0
    free[lv_cov_idx] <- 0L
  }
  # 3c. orthogonal.x = TRUE?
  if (orthogonal_x) {
    # note: the rhs may also be an observed exogenous covariate
    # (only if auto.cov.x = TRUE)
    lv_cov_idx <- which(op == "~~" &
      lhs %in% lv_names_x &
      rhs %in% c(lv_names_x, ov_names_x_block) &
      lhs != rhs &
      user == 0L)
    ustart[lv_cov_idx] <- 0.0
    free[lv_cov_idx] <- 0L
  }
  # 3d. orthogonal.efa = TRUE?
  if (orthogonal_efa) {
    lv_cov_idx <- which(op == "~~" &
      lhs %in% lv_names_efa &
      rhs %in% lv_names_efa &
      lhs != rhs &
      user == 0L)
    ustart[lv_cov_idx] <- 0.0
    free[lv_cov_idx] <- 0L
  }

  # 4. intercepts
  if (meanstructure) {
    if (categorical) {
      # zero intercepts/means ordinal variables
      ov_int_idx <- which(op == "~1" &
        lhs %in% ov_names_ord &
        user == 0L)
      ustart[ov_int_idx] <- 0.0
      free[ov_int_idx] <- 0L
    }
    if (int_ov_free == FALSE) {
      # zero intercepts/means observed variables
      ov_int_idx <- which(op == "~1" &
        lhs %in% ov_names &
        user == 0L)
      ustart[ov_int_idx] <- 0.0
      free[ov_int_idx] <- 0L
    }
    if (int_lv_free == FALSE) {
      # zero intercepts/means latent variables
      lv_int_idx <- which(op == "~1" &
        lhs %in% lv_names &
        user == 0L)
      ustart[lv_int_idx] <- 0.0
      free[lv_int_idx] <- 0L
    }
    # 4b. fixed effect (only if we have random slopes)
    if (!is.null(flat$rv) && any(nchar(flat$rv) > 0L)) {
      lv_names_rv <- lav_pt_vnames(flat, "lv.rv")
      lv_rv_idx <- which(op == "~1" &
        lhs %in% lv_names_rv &
        user == 0L)
      ustart[lv_rv_idx] <- as.numeric(NA)
      free[lv_rv_idx] <- 1L
    }
    if (length(lv_names_int) > 0L) {
      lv_int_idx <- which(op == "~1" &
        lhs %in% lv_names_int &
        user == 0L)
      ustart[lv_int_idx] <- as.numeric(NA)
      free[lv_int_idx] <- 1L
    }
    # composites: the composite mean is determined (= W'nu), so the intercept
    # is always non-free, with ustart = NA; the value is filled in later as a
    # function of the other parameters (see lav_lisrel_comp_set_intresvar).
    # This holds even if the user writes 'composite ~ 1' explicitly (just like
    # composite variances above): otherwise it adds a redundant free parameter
    # that does not affect the fit and makes the information matrix singular.
    if (length(lv_names_c) > 0L) {
      c_int_idx <- which(op == "~1" & lhs %in% lv_names_c)
      ustart[c_int_idx] <- as.numeric(NA)
      free[c_int_idx] <- 0L
    }
  }

  # 4b. fixed effect (only if we have random slopes)
  # if(!is.null(FLAT$rv)) {
  #    }

  # 5. handle exogenous `x' covariates
  # usually, ov.names.x.block == ov.names.x
  # except if multilevel, where 'split' ov.x are treated as endogenous

  # 5a conditional.x = FALSE
  if (!conditional_x && fixed_x && length(ov_names_x_block) > 0) {
    # 1. variances/covariances
    exo_var_idx <- which(op == "~~" &
      rhs %in% ov_names_x_block &
      lhs %in% ov_names_x_block &
      user == 0L)
    ustart[exo_var_idx] <- as.numeric(NA) # should be overridden later!
    free[exo_var_idx] <- 0L
    exo[exo_var_idx] <- 1L

    # 2. intercepts
    exo_int_idx <- which(op == "~1" &
      lhs %in% ov_names_x_block &
      user == 0L)
    ustart[exo_int_idx] <- as.numeric(NA) # should be overridden later!
    free[exo_int_idx] <- 0L
    exo[exo_int_idx] <- 1L
  }

  # 5a-bis. conditional.x = TRUE
  if (conditional_x && length(ov_names_x) > 0L) {
    # 1. variances/covariances
    exo_var_idx <- which(op == "~~" &
      rhs %in% ov_names_x &
      lhs %in% ov_names_x &
      user == 0L)
    if (fixed_x) {
      ustart[exo_var_idx] <- as.numeric(NA) # should be overridden later!
      free[exo_var_idx] <- 0L
    }
    exo[exo_var_idx] <- 1L

    # 2. intercepts
    exo_int_idx <- which(op == "~1" &
      lhs %in% ov_names_x &
      user == 0L)
    if (fixed_x) {
      ustart[exo_int_idx] <- as.numeric(NA) # should be overridden later!
      free[exo_int_idx] <- 0L
    }
    exo[exo_int_idx] <- 1L

    # 3. regressions ov + lv
    exo_reg_idx <- which(op %in% c("~", "<~") &
      lhs %in% c(lv_names, ov_names_nox) &
      rhs %in% ov_names_x)
    exo[exo_reg_idx] <- 1L

    # 3b regression split.x ~ regular.x
    exo_reg2_idx <- which(op %in% c("~", "<~") &
      lhs %in% ov_names_x &
      rhs %in% ov_names_x)
    if (fixed_x) {
      ustart[exo_reg2_idx] <- as.numeric(NA) # should be overridden later!
      free[exo_reg2_idx] <- 0L
    }
    exo[exo_reg2_idx] <- 1L
  }

  # 5b. residual variances of ordinal variables?
  if (length(ov_names_ord) > 0L) {
    ord_idx <- which(lhs %in% ov_names_ord &
      op == "~~" &
      user == 0L & ## New in 0.6-1
      lhs == rhs)
    ustart[ord_idx] <- 1L ## FIXME!! or 0?? (0 breaks ex3.12)
    free[ord_idx] <- 0L
  }

  # 5c latent response scales of ordinal variables?
  #    by default, all fixed to 1.0
  if (length(ov_names_ord) > 0L) {
    delta_idx <- which(op == "~*~" &
      user == 0L) ## New in 0.6-1
    ustart[delta_idx] <- 1.0
    free[delta_idx] <- 0L
  }

  # correlation structure (new in 0.6-13)
  # 'correlation' may be TRUE (all ov) or a character vector (a subset);
  # only the selected variables get a fixed unit variance + scaling factor
  if (isTRUE(correlation) || is.character(correlation)) {
    cor_ov <- if (is.character(correlation)) {
      ov_names[ov_names %in% correlation]
    } else {
      ov_names
    }
    var_idx <- which(lhs %in% cor_ov &
      op == "~~" &
      user == 0L &
      lhs == rhs)
    ustart[var_idx] <- 1L
    free[var_idx] <- 0L

    delta_idx <- which(op == "~*~" &
      user == 0L)
    ustart[delta_idx] <- 1.0
    free[delta_idx] <- 0L
  }

  # group proportions (group 1L)
  if (group_w_free) {
    group_idx <- which(lhs == "group" & op == "%")
    # if(ngroups > 1L) {
    free[group_idx] <- 1L
    ustart[group_idx] <- as.numeric(NA)
    # } else {
    #      free[ group.idx ] <- 0L
    #    ustart[ group.idx ] <- 0.0 # last group
    # }
  }

  # 6. multiple groups?
  group <- rep(1L, length(lhs))
  if (ngroups > 1) {

    # only if "loadings" in group.equal and !std.lv:
    # construct temporary tmp.list to obtain lv.marker
    if (!std_lv && "loadings" %in% group_equal) {
      tmp_list <- list(
        id          = seq_along(lhs),
        lhs         = lhs,
        op          = op,
        rhs         = rhs,
        free        = free,
        ustart      = ustart,
        block       = rep(1, length(rhs)))
      lv_marker <- lav_pt_vnames(tmp_list, "lv.marker")
    }


    group <- rep(1:ngroups, each = length(lhs))
    user <- rep(user, times = ngroups)
    lhs <- rep(lhs, times = ngroups)
    op <- rep(op, times = ngroups)
    rhs <- rep(rhs, times = ngroups)
    free <- rep(free, times = ngroups)
    ustart <- rep(ustart, times = ngroups)
    mod_idx <- rep(mod_idx, times = ngroups)
    label <- rep(label, times = ngroups)
    exo <- rep(exo, times = ngroups)

    # specific changes per group
    for (g in 2:ngroups) {
      # free/fix intercepts latent variables
      if (meanstructure) {
        int_idx <- which(op == "~1" &
          lhs %in% lv_names_noc &
          user == 0L &
          group == g)
        if (int_lv_free == FALSE && g > 1 &&
          ("intercepts" %in% group_equal) &&
          !("means" %in% group_equal)) {
          free[int_idx] <- 1L
          ustart[int_idx] <- as.numeric(NA)
        }
      }

      # free intercept indicators if equal thresholds (new in 0.6-20)
      if (meanstructure && length(ov_names_ord) > 0L) {
        ord_idx <- which(op == "~1" &
          lhs %in% ov_names_ord &
          user == 0L &
          group == g)
        if (int_lv_free == FALSE && g > 1 &&
           "thresholds" %in% group_equal) {
          free[ord_idx] <- 1L
          ustart[ord_idx] <- as.numeric(NA)
        }
      }

      # latent variances if std.lv = TRUE (new in 0.6-4)
      if (std_lv && "loadings" %in% group_equal &&
        !"lv.variances" %in% group_equal) {
        lv_var_idx <- which(op == "~~" &
          lhs %in% lv_names &
          !lhs %in% lv_names_efa &
          lhs == rhs &
          user == 0L &
          group == g)
        if (length(lv_var_idx) > 0L) {
          free[lv_var_idx] <- 1L
          ustart[lv_var_idx] <- as.numeric(NA)
        }
      # marker indicator if std.lv = FALSE (new in 0.6-20)
      } else if (!std_lv && "loadings" %in% group_equal) {
        marker_idx <- which(op == "=~" &
          rhs %in% lv_marker &
          free == 0L &
          ustart == 1L &
          group == g)
        if (length(marker_idx) > 0L) {
          free[marker_idx] <- 1L
          ustart[marker_idx] <- as.numeric(NA)
        }
      }

      # latent variances if efa = TRUE (new in 0.6-5)
      if (length(lv_names_efa) > 0L &&
          auto_efa && "loadings" %in% group_equal &&
        !"lv.variances" %in% group_equal) {
        lv_var_idx <- which(op == "~~" &
          lhs %in% lv_names_efa &
          lhs == rhs &
          user == 0L &
          group == g)
        if (length(lv_var_idx) > 0L) {
          free[lv_var_idx] <- 1L
          ustart[lv_var_idx] <- as.numeric(NA)
        }
      }

      # latent response scaling -- categorical only
      # - if thresholds are equal -> free scalings/residual variances
      # - but not for binary indicators!
      if (length(ov_names_ord) > 0L) {
        nth <- sapply(ov_names_ord,
          function(x) sum(lhs == x & op == "|" & group == 1L))
        ov_names_ord_notbinary <- ov_names_ord[nth > 1L]
        if (auto_delta && parameterization == "delta") {
          if (any(op == "~*~" & group == g) &&
            ("thresholds" %in% group_equal)) {
            delta_idx <- which(op == "~*~" & group == g &
              lhs %in% ov_names_ord_notbinary)
            free[delta_idx] <- 1L
            ustart[delta_idx] <- as.numeric(NA)
          }
        } else if (parameterization == "theta") {
          if (any(op == "~*~" & group == g) &&
            ("thresholds" %in% group_equal)) {
            var_ord_idx <- which(op == "~~" & group == g &
              lhs %in% ov_names_ord_notbinary & lhs == rhs)
            free[var_ord_idx] <- 1L
            ustart[var_ord_idx] <- as.numeric(NA)
          }
        }
      }

      # group proportions
      if (group_w_free) {
        group_idx <- which(lhs == "group" & op == "%" & group == g)
        # if(g == ngroups) {
        #      free[ group.idx ] <- 0L
        #    ustart[ group.idx ] <- 0.0 # last group
        # } else {
        free[group_idx] <- 1L
        ustart[group_idx] <- as.numeric(NA)
        # }
      }
    } # g
  } # ngroups

  # construct tmp.list
  tmp_list <- list(
    id          = seq_along(lhs),
    lhs         = lhs,
    op          = op,
    rhs         = rhs,
    user        = user
  )

  # add block column (before group/level columns)
  if (!is.null(block_id)) {
    # only one block
    tmp_list$block <- rep(block_id, length(lhs))
  } else {
    # block is a combination of at least group, level, ...
    # for now, only group
    tmp_list$block <- group
  }

  # block columns (typically only group)
  for (block in blocks) {
    if (block == "group") {
      tmp_list[[block]] <- group
    } else {
      tmp_list[[block]] <- rep(0L, length(lhs))
    }
  }

  # other columns
  tmp_list2 <- list(
    mod.idx     = mod_idx,
    free        = free,
    ustart      = ustart,
    exo         = exo,
    label       = label
  )

  tmp_list <- c(tmp_list, tmp_list2)
}

# initial version: YR 03/05/2017

# major change: YR 14/06/2022 for 0.6-12
# - summary() is now silent if not printed
# - here, we only collect the necessary ingredients, and store them in a
#   a list
# - the result is a S3 class lavaan.summary
# - the actual printing is done by lav_summary_print (see lav_print.R)

# YR 26 July 2022: add fm.args= argument to change the way (some) fit measures
#                  are computed
# YR 24 Sept 2022: add efa= argument
# YR 19 Nov  2023: add remove_unused= argument
# TDJ 28 March 2024: deprecate std.nox= argument ("std.nox" can be %in%
#                                                          standardized=)

# Is at least one defined (:=) parameter a NONLINEAR function of the free
# parameters? The (first-order) delta method is exact for linear definitions
# (sums/differences), so the se.def note in lav_summary_print() is only
# relevant when a nonlinear definition is present. Detected numerically: a
# function is linear iff its Jacobian does not depend on the parameter values.
lav_object_def_nonlinear <- function(object) {
  if (sum(object@ParTable$op == ":=") == 0L) {
    return(FALSE)
  }
  lavmodel <- object@Model
  if (!is.function(lavmodel@def.function)) {
    return(FALSE)
  }
  x <- try(lav_model_get_parameters(lavmodel, type = "free"), silent = TRUE)
  if (inherits(x, "try-error") || length(x) < 1L) {
    return(FALSE)
  }
  jac <- function(p) {
    out <- try(lav_func_jacobian_complex(lavmodel@def.function, p),
               silent = TRUE)
    if (inherits(out, "try-error")) {
      out <- try(lav_func_jacobian_simple(lavmodel@def.function, p),
                 silent = TRUE)
    }
    out
  }
  j1 <- jac(x)
  j2 <- jac(x + 1e-3 * (abs(x) + 1)) # deterministic perturbation
  if (inherits(j1, "try-error") || inherits(j2, "try-error")) {
    return(FALSE)
  }
  ok <- is.finite(j1) & is.finite(j2)
  isTRUE(any(abs(j1[ok] - j2[ok]) > 1e-8))
}

# create summary of a lavaan object
lav_object_summary <- function(object, header = TRUE,
                               fit_measures = FALSE,
                               baseline_model = NULL,
                               h1_model = NULL,
                               fm_args =
                                 list(
                                   standard.test = "default",
                                   scaled.test = "default",
                                   rmsea.ci.level = 0.90,
                                   rmsea.h0.closefit = 0.05,
                                   rmsea.h0.notclosefit = 0.08
                                 ),
                               residuals = FALSE,
                               estimates = TRUE,
                               ci = FALSE,
                               fmi = FALSE,
                               standardized = FALSE,
                               std = standardized,
                               remove_system_eq = TRUE,
                               remove_eq = TRUE,
                               remove_ineq = TRUE,
                               remove_def = FALSE,
                               remove_nonfree = FALSE,
                               remove_step1 = TRUE,
                               remove_unused = TRUE,
                               plabel = FALSE,
                               cov_std = TRUE,
                               rsquare = FALSE,
                               efa = FALSE,
                               efa_args =
                                 list(
                                   lambda = TRUE,
                                   theta = TRUE,
                                   psi = TRUE,
                                   eigenvalues = TRUE,
                                   sumsq_table = TRUE,
                                   lambda_structure = FALSE,
                                   fs_determinacy = FALSE,
                                   se = FALSE,
                                   zstat = FALSE,
                                   pvalue = FALSE
                                 ),
                               modindices = FALSE,
                               srmr_close_h0 = NULL,
                              ...) {
   dotdotdot <- list(...)
   lav_adapt_func(environment(), dotdotdot, NULL)
  # check object
  object <- lav_object_check_version(object)

  # default fm.args
  default_fm_args <- list(
    standard.test = "default",
    scaled.test = "default",
    rmsea.ci.level = 0.90,
    rmsea.close.h0 = 0.05,
    rmsea.notclose.h0 = 0.08,
    robust = TRUE,
    cat.nonpd = "na"
  )
  if (is.logical(fit_measures)) {
    if (fit_measures) {
      fit_measures <- "default"
    } else {
      fit_measures <- "none"
    }
  }
  if (!missing(fm_args)) {
    lav_deprecated_args("fit.measures", "fm.args")
    fm_args <- modifyList(default_fm_args, fm_args)
  } else {
    fm_args <- default_fm_args
  }
  if (is.list(fit_measures)) {
    fit_measures <- lav_fit_measures_list_alias(fit_measures)
    if (is.null(names(fit_measures)) ||
        is.null(fit_measures$fit.measures)) {
      lav_msg_stop(gettextf(
        "If %s is a list, it must contain a named element %s.",
        "fit.measures", "fit.measures (or fit_measures)"
      ))
    }
    temp <- fit_measures$fit.measures
    fit_measures$fit.measures <- NULL
    fm_args <- modifyList(default_fm_args, fit_measures)
    fit_measures <- temp
  }


  # return a list with the main ingredients
  res <- list()

  # this is to avoid partial matching of 'std' with std.nox
  if (is.logical(std) && is.logical(standardized)) {
    standardized <- std || standardized
  } else {
    # At least 1 is not logical.
    tmp <- union(
      if (is.logical(std)) character(0L) else as.character(std),
      if (is.logical(standardized)) character(0L) else as.character(standardized)
    )
    std_types <- c("std.lv", "std.all", "std.nox")
    if (length(tmp) > 0L && !any(tolower(tmp) %in% std_types)) {
      # a vector of observed variable names: standardize only those (a
      # generalization of std.nox); pass through unchanged
      standardized <- tmp
    } else {
      # Retain only valid standardization options.
      standardized <- intersect(tolower(tmp), std_types)
    }
  }

  # create the 'short' summary
  if (header) {
    # 1. collect header information
    version <- object@version

    res$header <- list(
      lavaan.version = version,
      sam.approach = !is.null(object@internal$sam.method),
      optim.method = object@Options$optim.method,
      optim.iterations = object@optim$iterations,
      optim.converged = object@optim$converged
    )

    # sam or sem?
    if (!is.null(object@internal$sam.method)) {
      # SAM version

      # 2. sam header
      res$sam.header <-
        list(
          sam.method = object@internal$sam.method,
          sam.local.options = object@internal$sam.local.options,
          sam.mm.list = object@internal$sam.mm.list,
          sam.mm.estimator = object@internal$sam.mm.estimator,
          sam.struc.estimator = object@internal$sam.struc.estimator
        )

      # 3. no EFA (for now)?

      # 4. summarize lavdata
      res$data <- lav_data_summary_short(object@Data)

      # 5a. sam local test statistics
      res$sam <- list(
        sam.method = object@internal$sam.method,
        sam.mm.table = object@internal$sam.mm.table,
        sam.mm.rel = object@internal$sam.mm.rel,
        sam.struc.fit = object@internal$sam.struc.fit,
        ngroups = object@Data@ngroups,
        group.label = object@Data@group.label,
        nlevels = object@Data@nlevels,
        level.label = object@Data@level.label,
        block.label = object@Data@block.label
      )

      # 5b. global test statistics (for global only)
      if (object@internal$sam.method == "global") {
        test <- object@test
        # attach the meta info that lav_test_print() needs (the SAM branch
        # does not go through the SEM path that normally sets this)
        if (is.null(attr(test, "info"))) {
          lavdata <- object@Data
          lavoptions <- object@Options
          attr(test, "info") <-
            list(
              ngroups = lavdata@ngroups,
              group.label = lavdata@group.label,
              information = lavoptions$information,
              h1.information = lavoptions$h1.information,
              observed.information = lavoptions$observed.information
            )
        }
        res$test <- test
      } else if (!identical(fit_measures, "none") &&
                 !is.null(lav_sam_struc_object(object))) {
        # 5c. local family + fit.measures: print the (structural) chi-square
        # test statistic(s) as usual, instead of the one-line
        # 'Summary Information Structural part:' (issue #517)
        fit_pa <- lav_sam_struc_object(object)
        test <- fit_pa@test
        if (is.null(attr(test, "info"))) {
          attr(test, "info") <-
            list(
              ngroups = fit_pa@Data@ngroups,
              group.label = fit_pa@Data@group.label,
              information = fit_pa@Options$information,
              h1.information = fit_pa@Options$h1.information,
              observed.information = fit_pa@Options$observed.information
            )
        }
        # the header is printed before the test statistics, making clear
        # that the model test (and the fit measures below it) concern the
        # structural part only
        attr(test, "header") <- gettext(
          "Fit measures of the structural part (given the measurement model):")
        res$test <- test
        res$sam$sam.struc.fit <- NULL
      }
    } else {
      # SEM version

      # 2. summarize optim info (including estimator)
      nrow_ceq_jac <- nrow(object@Model@ceq.JAC)
      #if (object@Model@ceq.simple.only) {
        # not needed, as nrow.ceq.jac is already zero
      #  nrow.ceq.jac <- 0L
      #}
      cin_simple_only <- FALSE
      ceq_simple_only <- FALSE
      cin_simple_only <- object@Model@cin.simple.only
      ceq_simple_only <- object@Model@ceq.simple.only
      nrow_cin_jac <- nrow(object@Model@cin.JAC)
      if (cin_simple_only) {
        nrow_cin_jac <- 0L
      }
      if (ceq_simple_only && cin_simple_only) {
        nrow_con_jac <- 0L
        con_jac_rank <- 0L
      } else {
        nrow_con_jac <- nrow(object@Model@con.jac)
        con_jac_rank <- qr(object@Model@con.jac)$rank
      }

      res$optim <- list(
        estimator = object@Options$estimator,
        estimator.args = object@Options$estimator.args,
        optim.method = object@Options$optim.method,
        npar = object@Model@nx.free,
        eq.constraints = object@Model@eq.constraints,
        nrow.ceq.jac = nrow_ceq_jac,
        nrow.cin.jac = nrow_cin_jac,
        nrow.con.jac = nrow_con_jac,
        con.jac.rank = con_jac_rank
      )


      # 3. if EFA/ESEM, summarize rotation info
      if (object@Model@nefa > 0L) {
        res$rotation <-
          list(
            rotation = object@Options$rotation,
            rotation.args = object@Options$rotation.args
          )
      }

      # 4. summarize lavdata
      res$data <- lav_data_summary_short(object@Data)

      # 5. test statistics
      test <- object@test
      # TDJ: check for user-supplied h1 model
      if (!is.null(object@external$h1.model)) {
        stopifnot(inherits(object@external$h1.model, "lavaan"))
        ## update @test slot
        test <- lav_update_test_custom_h1(
          lav_obj_h0 = object, lav_obj_h1 = object@external$h1.model)@test
      }
      # double check if we have attr(TEST, "info") (perhaps old object?)
      if (is.null(attr(test, "info"))) {
        lavdata <- object@Data
        lavoptions <- object@Options
        attr(test, "info") <-
          list(
            ngroups = lavdata@ngroups,
            group.label = lavdata@group.label,
            information = lavoptions$information,
            h1.information = lavoptions$h1.information,
            observed.information = lavoptions$observed.information
          )
      }
      res$test <- test
    } # regular sem
  } # header

  # efa-related info
  if (efa) {
    res$efa <- lav_efa_summary(object, efa_args = efa_args)
  } # efa

  # only if requested, add the additional fit measures
  # (fit_measures may be a vector of measure names; only the single
  #  "none" turns this section off)
  if (!identical(fit_measures, "none")) {
    # some early warnings (to avoid a hard stop)
    if (object@Data@data.type == "none") {
      lav_msg_warn(gettext(
        "fit measures not available if there is no data"))
    } else if (lav_sam_local_flag(object) &&
      is.null(lav_sam_struc_object(object))) {
      lav_msg_warn(gettext(
        "fit measures not available: this sam object was created by an
         older version of lavaan; please rerun sam()"))
    } else if (length(object@Options$test) == 1L &&
      object@Options$test == "none" &&
      length(object@Model@rv.ov) == 0L &&
      length(object@Model@rv.lv) == 0L) {
      # (random-slope models do provide the loglikelihood-based
      #  measures, despite test = "none")
      lav_msg_warn(gettext(
        "fit measures not available if test = \"none\""))
    } else if (object@optim$npar > 0L && !object@optim$converged) {
      lav_msg_warn(gettext(
        "fit measures not available if model did not converge"))
    } else {
      fit <- lav_fit(object,
       fit_measures = c(list(fit.measures = fit_measures), fm_args),
       baseline_model = baseline_model,
       h1_model = h1_model)
      # local sam objects: the 'structural part' header is already printed
      # once, before the test statistics (see res$test above)
      if (!is.null(attr(res$test, "header"))) {
        attr(fit, "header") <- NULL
      }
      res$fit <- fit

      # multilevel: add level-specific fit measures (based on the stored
      # partially saturated fits; see fit.by.level option)
      if (object@Data@nlevels > 1L) {
        res$fit.by.level <- lav_fit_by_level_fm(object)
      }
    }
  }

  # only if requested, add the residuals (and their summary). 'residuals' may be
  # TRUE/FALSE, or a (named) list of arguments passed on to lavResiduals().
  if ((is.list(residuals) || isTRUE(residuals)) &&
      object@optim$npar > 0L && !object@optim$converged) {
    lav_msg_warn(gettext("residuals not available if model did not converge"))
    residuals <- FALSE
  }
  # sam objects: warn-and-skip where lavResiduals() would stop (see the
  # sam branch in lavResiduals())
  if ((is.list(residuals) || isTRUE(residuals)) &&
      !is.null(object@internal$sam.method)) {
    if (lav_sam_local_flag(object) &&
        is.null(lav_sam_struc_object(object))) {
      lav_msg_warn(gettext(
        "residuals not available: this sam object was created by an older
         version of lavaan; please rerun sam()"))
      residuals <- FALSE
    } else if (!lav_sam_local_flag(object)) {
      lav_msg_warn(gettext(
        "residual summary statistics are not available if sam.method =
        \"global\"; use residuals() to inspect the raw residuals"))
      residuals <- FALSE
    }
  }
  if (is.list(residuals) || isTRUE(residuals)) {
    res_args <- if (is.list(residuals)) residuals else list()
    # defaults (overridable via the list form), matching lavResiduals()
    res_defaults <- list(
      type = "cor.bentler", se = FALSE, zstat = TRUE, n.largest = 5L,
      combine = FALSE, usrmr.ci.level = 0.90, usrmr.close.h0 = 0.05
    )
    res_args <- modifyList(res_defaults, res_args)
    # the summary needs both the element-wise residuals and the summary tables
    res_args <- c(
      res_args,
      list(
        object = object, summary = TRUE, elementwise = TRUE, output = "list"
      )
    )
    res_list <- do.call(lavResiduals, res_args)
    res$residuals <- res_list

    # for local sam objects, lavResiduals() delegated to the structural part
    res_pt <- object@ParTable
    if (lav_sam_local_flag(object)) {
      res_pt <- lav_sam_struc_object(object)@ParTable
    }

    # pre-build the text representation used by lav_summary_print()
    res_text <- lav_residuals_list_to_text(
      res_list, type = res_args$type,
      nblocks = lav_pt_nblocks(res_pt), drop_single = TRUE,
      n_largest = res_args$n.largest, show_se = isTRUE(res_args$se),
      show_z = isTRUE(res_args$zstat), ci_level = res_args$usrmr.ci.level,
      combine = isTRUE(res_args$combine), do_summary = TRUE, do_largest = TRUE
    )
    attr(res$residuals, "text") <- res_text

    # if the fit measures are also shown AND the residual summary is the SRMR,
    # drop the SRMR from the fit measures to avoid printing it twice
    if (!is.null(res$fit) && !is.null(res_text$summary_tables)) {
      r_metric <- rownames(res_text$summary_tables[[1]])[1]
      drop_srmr <- c("srmr", "srmr_nomean", "srmr_within", "srmr_between")
      if (identical(r_metric, "srmr") &&
          any(names(res$fit) %in% drop_srmr)) {
        fit_attr <- attributes(res$fit)
        res$fit <- res$fit[setdiff(names(res$fit), drop_srmr)]
        # restore the class (and any other attributes) lost by subsetting
        for (a in setdiff(names(fit_attr), "names")) {
          attr(res$fit, a) <- fit_attr[[a]]
        }
      }
    }
  }

  # main ingredient: the parameter table
  if (estimates) {
    pe <- lavParameterEstimates(object,
      ci = ci, standardized = standardized,
      rsquare = rsquare, fmi = fmi,
      cov_std = cov_std,
      remove_eq = remove_eq, remove_system_eq = remove_system_eq,
      remove_ineq = remove_ineq, remove_def = remove_def,
      remove_nonfree = remove_nonfree,
      remove_step1 = remove_step1,
      remove_unused = remove_unused,
      plabel = plabel,
      output = "text",
      header = TRUE
    )
    res$pe <- as.data.frame(pe)

    # flag: nonlinear defined (:=) parameters whose SE/CI rely on the
    # (first-order) delta method -> suggest se.def="mc" or bootstrap
    # (printed by lav_summary_print(); silent otherwise). Stored as an
    # attribute of the parameter table (res$pe) instead of a separate
    # list element, to avoid disturbing code that relies on partial
    # matching of names(summary(object)) (e.g. $p -> $pe).
    attr(res$pe, "delta.note") <-
      identical(object@Options$se.def, "default") &&
      !identical(object@Options$se, "bootstrap") &&
      !identical(object@Options$se, "none") &&
      lav_object_def_nonlinear(object)
  }

  # modification indices?
  if (is.logical(modindices)) {
    if (modindices) modindices <- list(standardized = TRUE, cov_std = cov_std)
  }
  if (is.list(modindices)) {
    mi <- do.call("modificationIndices", c(list(object = object), modindices))
    res$mi <- mi
  }

  # create lavaan.summary S3 class
  class(res) <- c("lavaan.summary", "list")

  res
}

# export a fitted lavaan object to Mplus input (+ a data file Mplus can read)
#
# The generated input, when run in Mplus, reproduces the lavaan model:
#  - a single data file is written, with extra columns for the grouping
#    variable (GROUPING) and/or the cluster variable (CLUSTER) when needed
#  - equality constraints are imposed through shared (clean) parameter labels;
#    other constraints / defined parameters go to a MODEL CONSTRAINT block
#  - multigroup models use group-specific MODEL blocks; multilevel models use
#    %WITHIN% / %BETWEEN% sections (nested within group blocks if both apply)

# main entry point: returns the complete Mplus input file as a string
# (data_file gives the name(s) of the data file(s) referenced in the DATA block)
lav_export_mplus <- function(object, data_file = "data.raw") {
  if (inherits(object, "lavaan")) {
    pt      <- as.data.frame(object@ParTable, stringsAsFactors = FALSE)
    ngroups <- object@Data@ngroups
    nlevels <- object@Data@nlevels
    opt     <- object@Options
  } else {
    pt      <- as.data.frame(lav_export_check(object), stringsAsFactors = FALSE)
    ngroups <- if (is.null(pt$group)) 1L else lav_pt_ngroups(pt)
    nlevels <- if (is.null(pt$level)) 1L else max(pt$level)
    opt     <- list(estimator = "ML", se = "standard", test = "standard",
                    information = "expected", meanstructure = FALSE)
  }
  nn <- length(pt$lhs)
  if (is.null(pt$free))   pt$free   <- rep(0L, nn)
  if (is.null(pt$ustart)) pt$ustart <- if (!is.null(pt$est)) pt$est else rep(NA, nn)
  if (is.null(pt$label))  pt$label  <- rep("", nn)
  if (is.null(pt$plabel)) pt$plabel <- rep("", nn)
  if (is.null(pt$exo))    pt$exo    <- rep(0L, nn)
  if (is.null(pt$block))  pt$block  <- rep(1L, nn)
  if (is.null(pt$group))  pt$group  <- ifelse(pt$block == 0L, 0L, 1L)
  if (is.null(pt$level))  pt$level  <- ifelse(pt$block == 0L, 0L, 1L)
  pt$label[is.na(pt$label)] <- ""

  # NOTE: variable names are sanitized (dots -> underscores) only when emitting
  # model statements (in lav_export_mplus_block), *after* label/constraint
  # resolution, so that plabel references keep matching.

  is_con <- pt$op %in% c("==", "<", ">", ":=")
  display_label <- lav_export_lavaan_labels(pt, is_con)
  mlabel <- vapply(display_label, lav_export_mplus_label, character(1L),
                   USE.NAMES = FALSE)

  ov_ord <- lav_export_mplus_var(lav_pt_vnames(pt, type = "ov.ord"))

  # Mplus is case-insensitive, so a latent variable name (eg factor 'C1') may
  # collide with an observed variable name that differs only in case (eg
  # indicator 'c1'). Mplus rejects such input; rename offending latent
  # variables to a collision-free name (see lav_export_mplus_lv_rename).
  lv_rename <- lav_export_mplus_lv_rename(pt)

  # Mplus group labels (used both in the GROUPING statement and the
  # group-specific MODEL headers, so they must agree)
  group_labels <- lav_export_mplus_group_labels(object, ngroups)

  # ---- MODEL block(s) ----
  body <- character(0L)
  for (g in seq_len(ngroups)) {
    header <- if (g == 1L) "MODEL:" else paste0("MODEL ", group_labels[g], ":")
    body <- c(body, header)
    if (nlevels > 1L) {
      for (l in seq_len(nlevels)) {
        sec <- if (l == 1L) "%WITHIN%" else "%BETWEEN%"
        body <- c(body, paste0("  ", sec))
        rows <- which(!is_con & pt$group == g & pt$level == l)
        # in Mplus, means/intercepts exist only on the between level
        body <- c(body, lav_export_mplus_block(pt, rows, mlabel, ov_ord,
                                               indent = "    ",
                                               drop_means = (l == 1L),
                                               lv_rename = lv_rename))
      }
    } else {
      rows <- which(!is_con & pt$group == g)
      body <- c(body, lav_export_mplus_block(pt, rows, mlabel, ov_ord,
                                            indent = "  ",
                                            lv_rename = lv_rename))
    }
  }

  # ---- MODEL CONSTRAINT block ----
  con_df <- lav_export_constraints_df(pt, is_con, display_label)
  con_block <- ""
  if (nrow(con_df)) {
    lines <- character(0L)
    new_par <- con_df$lhs[con_df$op == ":="]
    if (length(new_par)) {
      lines <- c(lines, paste0("  NEW(", paste(new_par, collapse = " "), ");"))
    }
    for (i in seq_len(nrow(con_df))) {
      op <- con_df$op[i]
      mop <- switch(op, "==" = "=", ":=" = "=", op)
      lhs <- lav_export_mplus_expr(con_df$lhs[i])
      rhs <- lav_export_mplus_expr(con_df$rhs[i])
      lines <- c(lines, paste0("  ", lhs, " ", mop, " ", rhs, ";"))
    }
    con_block <- paste0("MODEL CONSTRAINT:\n", paste(lines, collapse = "\n"),
                        "\n")
  }

  # ---- header (TITLE / DATA / VARIABLE / ANALYSIS) ----
  head <- lav_export_mplus_header(object, pt, data_file = data_file,
                                  ngroups = ngroups, nlevels = nlevels,
                                  ov_ord = ov_ord, opt = opt,
                                  group_labels = group_labels)

  out <- paste0(head,
                paste(body, collapse = "\n"), "\n",
                con_block,
                "OUTPUT:\n  sampstat standardized tech1;\n")
  class(out) <- c("lavaan.character", "character")
  out
}

# emit Mplus MODEL statements for a set of parameter rows
lav_export_mplus_block <- function(pt, rows, mlabel, ov_ord, indent = "  ",
                                   drop_means = FALSE,
                                   lv_rename = character(0L)) {
  if (length(rows) == 0L) {
    return(character(0L))
  }
  out <- character(0L)
  for (i in rows) {
    op <- pt$op[i]
    lhs <- lav_export_mplus_var(pt$lhs[i])
    rhs <- lav_export_mplus_var(pt$rhs[i])

    # rename latent variables that would clash (case-insensitively) with an
    # observed variable name in (case-insensitive) Mplus
    if (length(lv_rename)) {
      if (lhs %in% names(lv_rename)) lhs <- unname(lv_rename[lhs])
      if (rhs %in% names(lv_rename)) rhs <- unname(lv_rename[rhs])
    }

    # skip parameters that Mplus handles implicitly
    if (op %in% c("~~", "~1") && pt$exo[i] == 1L) next            # fixed.x exo
    if (op == "~~" && lhs == rhs && lhs %in% ov_ord) next         # ordinal var
    if (op == "~1" && lhs %in% ov_ord) next                       # ordinal int
    if (op == "~1" && drop_means) next             # within level: no means

    lab <- if (nzchar(mlabel[i])) paste0(" (", mlabel[i], ")") else ""
    mod <- lav_export_mplus_mod(pt$free[i], pt$ustart[i])

    stmt <- switch(op,
      "=~" = paste0(lhs, " BY ", rhs, mod, lab, ";"),
      "<~" = paste0(lhs, " BY ", rhs, mod, lab, ";"),
      "~"  = paste0(lhs, " ON ", rhs, mod, lab, ";"),
      "~~" = if (lhs == rhs) {
               paste0(lhs, mod, lab, ";")
             } else {
               paste0(lhs, " WITH ", rhs, mod, lab, ";")
             },
      "~1" = paste0("[", lhs, mod, "]", lab, ";"),
      "|"  = paste0("[", lhs, "$", sub("t", "", rhs), mod, "]", lab, ";"),
      "~*~" = paste0("{", lhs, mod, lab, "};"),
      NULL
    )
    if (!is.null(stmt)) out <- c(out, paste0(indent, stmt))
  }
  out
}

# Mplus value modifier: @value for fixed, * for free
lav_export_mplus_mod <- function(free, ustart) {
  if (free == 0L) {
    if (is.na(ustart)) {
      return("")
    }
    return(paste0("@", lav_export_num(ustart)))
  }
  "*"
}

# sanitize a variable name for Mplus (no dots)
lav_export_mplus_var <- function(x) {
  gsub(".", "_", x, fixed = TRUE)
}

# Mplus is not case-sensitive: a latent variable (lhs of =~) may never share a
# name with an observed variable that differs only in case (eg factor 'C1' vs
# indicator 'c1'). Build a named map (original -> new) renaming each offending
# latent variable to a collision-free name by appending 'F'. Names are compared
# after the usual dot->underscore sanitization, so the map keys/values match the
# tokens emitted in the MODEL block.
lav_export_mplus_lv_rename <- function(pt) {
  lv_names <- lav_export_mplus_var(lav_pt_vnames(pt, type = "lv"))
  ov_names <- lav_export_mplus_var(lav_pt_vnames(pt, type = "ov"))
  rename <- character(0L)
  if (length(lv_names) == 0L) {
    return(rename)
  }
  ov_lower <- tolower(ov_names)
  # names already in use (case-insensitive); grow it as we assign new names so
  # that two renamed factors cannot end up clashing with each other
  taken <- tolower(c(ov_names, lv_names))
  for (lv in lv_names) {
    if (!(tolower(lv) %in% ov_lower)) next        # no collision
    new <- paste0(lv, "F")
    while (tolower(new) %in% taken) {
      new <- paste0(new, "F")
    }
    rename[lv] <- new
    taken <- c(taken, tolower(new))
  }
  rename
}

# sanitize a parameter label for Mplus (alnum/underscore, must start w/ letter)
lav_export_mplus_label <- function(x) {
  if (!nzchar(x)) {
    return("")
  }
  x <- gsub("[^A-Za-z0-9_]", "", x)
  if (!grepl("^[A-Za-z]", x)) x <- paste0("p", x)
  x
}

# rewrite a constraint expression: sanitize labels embedded in it and convert
# the few operators that differ between lavaan and Mplus (eg ^ -> **)
lav_export_mplus_expr <- function(expr) {
  if (is.na(expr) || !nzchar(expr)) {
    return(expr)
  }
  expr <- gsub("^", "**", expr, fixed = TRUE)
  # drop any leftover dots inside plabel-style tokens
  gsub(".", "", expr, fixed = TRUE)
}

# Mplus group labels for a multigroup model. We use the actual lavaan group
# labels (object@Data@group.label) when they can be represented as valid Mplus
# names (letters/digits/underscore, beginning with a letter, max 8 chars, and
# unique case-insensitively, since Mplus is case-insensitive). Labels starting
# with a digit are prefixed with 'g' (e.g. "1" -> "g1"). If the labels cannot
# be represented safely (empty, or collapse to duplicates after sanitizing),
# we fall back to generic g1..gN.
lav_export_mplus_group_labels <- function(object, ngroups) {
  fallback <- paste0("g", seq_len(ngroups))
  if (!inherits(object, "lavaan")) {
    return(fallback)
  }
  labs <- object@Data@group.label
  if (length(labs) != ngroups || ngroups < 1L) {
    return(fallback)
  }
  safe <- vapply(as.character(labs), function(x) {
    y <- gsub("[^A-Za-z0-9_]", "", x)
    if (!nzchar(y) || !grepl("^[A-Za-z]", y)) {
      y <- paste0("g", y)
    }
    substr(y, 1L, 8L)
  }, character(1L), USE.NAMES = FALSE)
  if (any(!nzchar(safe)) || anyDuplicated(tolower(safe))) {
    return(fallback)
  }
  safe
}

# build the TITLE / DATA / VARIABLE / ANALYSIS header
lav_export_mplus_header <- function(object, pt, data_file, ngroups, nlevels,
                                    ov_ord, opt, group_labels = NULL) {
  if (is.null(group_labels)) {
    group_labels <- lav_export_mplus_group_labels(object, ngroups)
  }
  estimator <- if (inherits(object, "lavaan")) {
    lav_mplus_estimator(object)
  } else {
    "ML"
  }
  meanstructure <- isTRUE(opt$meanstructure)
  information <- opt$information
  if (is.null(information)) information <- "expected"

  ov_names <- lav_export_mplus_var(lav_pt_vnames(pt, type = "ov"))
  weight_name <- if (inherits(object, "lavaan")) {
    object@Data@sampling.weights
  } else {
    character(0L)
  }
  weight_name <- lav_export_mplus_var(weight_name)

  # single-level clustered data (cluster= but no level: structure) -> Mplus
  # TYPE=COMPLEX with a CLUSTER variable (cluster-robust SEs/test, issue #254)
  clustered <- nlevels == 1L && inherits(object, "lavaan") &&
    length(object@Data@cluster) > 0L

  # all columns present in the data file (must match the written data)
  data_cols <- ov_names
  if (length(weight_name)) data_cols <- c(data_cols, weight_name)
  if (ngroups > 1L) data_cols <- c(data_cols, "GRP")
  if (nlevels > 1L || clustered) data_cols <- c(data_cols, "CLUS")

  data_type <- if (inherits(object, "lavaan")) {
    object@Data@data.type
  } else {
    "full"
  }
  listwise <- isTRUE(opt$missing == "listwise")

  # TITLE
  c_title <- "TITLE:\n  this Mplus input file is autogenerated by lavExport()\n"

  # DATA
  c_data <- "DATA:\n"
  if (ngroups > 1L && length(data_file) > 1L) {
    for (g in seq_len(ngroups)) {
      c_data <- paste0(c_data, "  file (g", g, ") is ", data_file[g], ";\n")
    }
  } else {
    c_data <- paste0(c_data, "  file is ", data_file[1L], ";\n")
  }
  if (data_type == "full") {
    c_data <- paste0(c_data, "  type is individual;\n")
    if (listwise) c_data <- paste0(c_data, "  listwise = on;\n")
  } else if (data_type == "moment") {
    c_data <- paste0(c_data, "  type is fullcov;\n")
    c_data <- paste0(c_data, "  nobservations are ",
                     object@Data@nobs[[1L]], ";\n")
  }

  # VARIABLE
  c_var <- paste0("VARIABLE:\n  names are\n",
                  lav_export_mplus_wrap(data_cols), ";\n")
  c_var <- paste0(c_var, "  usevariables are\n",
                  lav_export_mplus_wrap(ov_names), ";\n")
  if (data_type == "full") {
    c_var <- paste0(c_var, "  missing are all (-999999);\n")
  }
  if (length(ov_ord)) {
    c_var <- paste0(c_var, "  categorical are\n",
                    lav_export_mplus_wrap(ov_ord), ";\n")
  }
  if (ngroups > 1L && length(data_file) <= 1L) {
    glab <- paste0(seq_len(ngroups), "=", group_labels)
    c_var <- paste0(c_var, "  grouping is GRP (", paste(glab, collapse = " "),
                    ");\n")
  }
  if (clustered || nlevels > 1L) {
    c_var <- paste0(c_var, "  cluster is CLUS;\n")
  }
  if (nlevels > 1L) {
    # declare variables that are modeled only at the within or between level
    if (inherits(object, "lavaan")) {
      ovl <- object@Data@ov.names.l[[1L]]
      if (length(ovl) >= 2L) {
        within_only  <- lav_export_mplus_var(setdiff(ovl[[1L]], ovl[[2L]]))
        between_only <- lav_export_mplus_var(setdiff(ovl[[2L]], ovl[[1L]]))
        if (length(within_only)) {
          c_var <- paste0(c_var, "  within = ",
                          paste(within_only, collapse = " "), ";\n")
        }
        if (length(between_only)) {
          c_var <- paste0(c_var, "  between = ",
                          paste(between_only, collapse = " "), ";\n")
        }
      }
    }
  }
  if (length(weight_name)) {
    c_var <- paste0(c_var, "  weight is ", weight_name, ";\n")
  }

  # ANALYSIS
  atype <- if (nlevels > 1L) "twolevel" else if (clustered) "complex" else "general"
  c_ana <- paste0("ANALYSIS:\n  type = ", atype, ";\n")
  c_ana <- paste0(c_ana, "  estimator = ", toupper(estimator), ";\n")
  if (toupper(estimator) %in% c("ML", "MLR", "MLF") && nlevels == 1L) {
    c_ana <- paste0(c_ana, "  information = ", information[1L], ";\n")
  }
  if (!meanstructure && nlevels == 1L) {
    c_ana <- paste0(c_ana, "  model = nomeanstructure;\n")
  }

  paste0(c_title, c_data, c_var, c_ana)
}

# wrap a vector of names over multiple indented lines (6 per line)
lav_export_mplus_wrap <- function(x, per = 6L, indent = "    ") {
  if (length(x) == 0L) {
    return("")
  }
  grp <- (seq_along(x) - 1L) %/% per
  lines <- tapply(x, grp, function(z) paste0(indent, paste(z, collapse = " ")))
  paste(lines, collapse = "\n")
}


# helper functions ----------------------------------------------------------

lav_mplus_estimator <- function(object) {
  estimator <- object@Options$estimator
  if (estimator == "DWLS") {
    estimator <- "WLS"
  }

  # only 1 argument for 'test' is allowed: drop the (default) "standard" test
  # and keep the scaled/robust one if present
  test <- object@Options$test
  test <- test[test != "standard"]
  if (length(test) == 0L) {
    test <- "standard"
  } else if (length(test) > 1L) {
    test <- test[1L]
  }
  object@Options$test <- test

  if (estimator == "ML") {
    if (object@Options$test %in% c("yuan.bentler", "yuan.bentler.mplus")) {
      estimator <- "MLR"
    } else if (object@Options$test == "satorra.bentler") {
      estimator <- "MLM"
    } else if (object@Options$test == "scaled.shifted") {
      estimator <- "MLMV"
    } else if (object@Options$se == "first.order") {
      estimator <- "MLF"
    }
  } else if (estimator %in% c("ULS", "WLS")) {
    if (object@Options$test == "satorra.bentler") {
      estimator <- paste(estimator, "M", sep = "")
    } else if (object@Options$test == "scaled.shifted") {
      estimator <- paste(estimator, "MV", sep = "")
    }
  } else if (estimator == "MML") {
    estimator <- "ML"
  }

  estimator
}

# - 0.6-13: fix multiple-group UG^2 bug (reported by Gronneberg, Foldnes and
#           Moss) when Satterthwaite = TRUE, ngroups > 1, and eq constraints.
#
#           Note however that Satterthwaite = FALSE always (for now), so
#           the fix has no (visible) effect
# - 0.7-1:  the traces are computed by the streamed worker
#           lav_test_ug_trace_stream() (lav_test_utils.R): no more
#           pstar x pstar products (A1 %*% ... %*% A1)

lav_test_yb <- function(lavobject = NULL,
                                  lavsamplestats = NULL,
                                  lavmodel = NULL,
                                  lavimplied = NULL,
                                  lavh1 = NULL,
                                  lavoptions = NULL,
                                  lavdata = NULL,
                                  test_unscaled = NULL,
                                  e_inv = NULL,
                                  b0_group = NULL,
                                  test = "yuan.bentler",
                                  mimic = "lavaan",
                                  # method          = "default",
                                  ug2_old_approach = FALSE,
                                  return_ugamma = FALSE) {
  test_1 <- list()

  if (!is.null(lavobject)) {
    lavsamplestats <- lavobject@SampleStats
    lavmodel <- lavobject@Model
    lavoptions <- lavobject@Options
    # lavpartable <- lavobject@ParTable
    lavimplied <- lavobject@implied
    lavh1 <- lavobject@h1
    lavdata <- lavobject@Data
    test_1$standard <- lavobject@test[[1]]
  } else {
    test_1$standard <- test_unscaled
  }

  # ug2.old.approach
  if (missing(ug2_old_approach)) {
    if (!is.null(lavoptions$ug2.old.approach)) {
      ug2_old_approach <- lavoptions$ug2.old.approach
    } else {
      ug2_old_approach <- FALSE
    }
  }

  # E.inv ok?
  lavoptions <- lav_test_change_options(lavoptions)
  e_inv_recompute <- attr(lavoptions, "recompute")
  attr(lavoptions, "recompute") <- NULL
  if (!is.null(e_inv)) {
    e_inv_recompute <- FALSE # user-provided
  }

  # check test
  if (!all(test %in% c(
    "yuan.bentler",
    "yuan.bentler.mplus"
  ))) {
    lav_msg_warn(gettext("test must be one of `yuan.bentler', or
                         `yuan.bentler.mplus'; will use `yuan.bentler' only"))
    test <- "yuan.bentler"
  }

  # do we have E.inv?
  if (is.null(e_inv) || e_inv_recompute) {
    e_inv <- try(
      lav_model_info(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        lavimplied = lavimplied,
        lavoptions = lavoptions,
        extra = FALSE,
        augmented = TRUE,
        inverted = TRUE
      ),
      silent = TRUE
    )
    if (inherits(e_inv, "try-error")) {
      if (return_ugamma) {
        lav_msg_warn(gettext(
          "could not invert information matrix needed for UGamma"))
        return(NULL)
      } else {
        na_out <- lav_test_scaled_na(test_1$standard,
          test1 = test[1],
          ngroups = lavdata@ngroups
        )
        test_1$standard <- na_out$standard
        test_1[[test[1]]] <- na_out$failed
        lav_msg_warn(gettext("could not invert information matrix needed for
                             robust test statistic"))
        return(test_1)
      }
    }
  }

  # catch df == 0
  if (test_1$standard$df == 0L || test_1$standard$df < 0) {
    na_out <- lav_test_scaled_na(test_1$standard,
      test1 = test[1],
      na_standard = FALSE, shift = FALSE
    )
    test_1[[test[1]]] <- na_out$failed
    return(test_1)
  }

  # FIXME: should we not always use 'unstructured' here?
  # if the model is, say, the independence model, the
  # 'structured' information (A1) will be so far away from B1
  # that we will end up with 'NA'
  h1_options <- lavoptions
  if (test == "yuan.bentler.mplus") {
    # always 'unstructured' H1 information
    h1_options$h1.information <- "unstructured"
  }

  # A1 is usually expected or observed
  a1_group <- lav_model_h1_info(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    lavimplied = lavimplied,
    lavh1 = lavh1,
    lavoptions = h1_options
  )
  # B1 is always first.order
  b1_group <- lav_model_h1_info_firstorder(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    lavimplied = lavimplied,
    lavh1 = lavh1,
    lavoptions = h1_options
  )

  if (test == "yuan.bentler.mplus") {
    if (is.null(b0_group)) {
      b0 <- lav_model_info_firstorder(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        lavh1 = lavh1,
        lavoptions = lavoptions,
        extra = TRUE,
        check_pd = FALSE,
        augmented = FALSE,
        inverted = FALSE
      )
      b0_group <- attr(b0, "B0.group")
    }
    trace_ugamma <-
      lav_test_yb_mplus_trace(
        lavsamplestats = lavsamplestats,
        a1_group = a1_group,
        b1_group = b1_group,
        b0_group = b0_group,
        e_inv = e_inv,
        meanstructure = lavmodel@meanstructure
      )
  } else if (test == "yuan.bentler") {
    # compute Delta
    delta <- lav_model_delta(lavmodel = lavmodel)

    # compute Omega/Gamma
    omega <- lav_model_h1_omega(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavoptions = lavoptions
    )

    # compute trace 'U %*% Gamma' (or 'U %*% Omega')
    trace_ugamma <- lav_test_yb_trace(
      lavsamplestats   = lavsamplestats,
      meanstructure    = lavmodel@meanstructure,
      a1_group         = a1_group,
      b1_group         = b1_group,
      delta            = delta,
      omega            = omega,
      e_inv            = e_inv,
      ug2_old_approach = ug2_old_approach,
      satterthwaite    = FALSE
    ) # for now
  }

  # unscaled test
  df <- test_1$standard$df

  scaling_factor <- trace_ugamma / df
  if (scaling_factor < 0) scaling_factor <- as.numeric(NA)
  chisq_scaled <- test_1$standard$stat / scaling_factor
  pvalue_scaled <- 1 - pchisq(chisq_scaled, df)

  ndat <- sum(attr(trace_ugamma, "h1.ndat"))
  npar <- lavmodel@nx.free

  scaling_factor_h1 <- sum(attr(trace_ugamma, "h1")) / ndat
  scaling_factor_h0 <- sum(attr(trace_ugamma, "h0")) / npar
  trace_ugamma2 <- attr(trace_ugamma, "trace.UGamma2")
  attributes(trace_ugamma) <- NULL

  if ("yuan.bentler" %in% test) {
    test_1$yuan.bentler <-
      list(
        test = test,
        stat = chisq_scaled,
        stat.group = (test_1$standard$stat.group /
          scaling_factor),
        df = df,
        pvalue = pvalue_scaled,
        scaling.factor = scaling_factor,
        scaling.factor.h1 = scaling_factor_h1,
        scaling.factor.h0 = scaling_factor_h0,
        label = "Yuan-Bentler correction",
        trace.UGamma = trace_ugamma,
        trace.UGamma2 = trace_ugamma2,
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test
      )
  } else if ("yuan.bentler.mplus" %in% test) {
    test_1$yuan.bentler.mplus <-
      list(
        test = test,
        stat = chisq_scaled,
        stat.group = (test_1$standard$stat.group /
          scaling_factor),
        df = df,
        pvalue = pvalue_scaled,
        scaling.factor = scaling_factor,
        scaling.factor.h1 = scaling_factor_h1,
        scaling.factor.h0 = scaling_factor_h0,
        label =
          "Yuan-Bentler correction (Mplus variant)",
        trace.UGamma = trace_ugamma,
        trace.UGamma2 = as.numeric(NA),
        scaled.test.stat = test_1$standard$stat,
        scaled.test = test_1$standard$test
      )
  }

  test_1
}

# trace of U %*% Gamma (U %*% Omega), with U = A1 - A1 Delta E.inv Delta' A1,
# via the streamed worker (same structure as the Satorra-Bentler U, with
# V := A1); trace_h0 = tr(B1 Delta E.inv Delta') is accumulated per group
# as tr(E.inv Delta' B1 Delta)
lav_test_yb_trace <- function(lavsamplestats = lavsamplestats,
                                        meanstructure = TRUE,
                                        a1_group = NULL,
                                        b1_group = NULL,
                                        delta = NULL,
                                        omega = NULL,
                                        e_inv = NULL,
                                        ug2_old_approach = FALSE,
                                        satterthwaite = FALSE) {
  ngroups <- lavsamplestats@ngroups
  fg <- unlist(lavsamplestats@nobs) / lavsamplestats@ntotal

  trace_h1 <- attr(omega, "trace.h1")
  h1_ndat <- attr(omega, "h1.ndat")

  # tr(U Gamma) (and tr((U Gamma)^2) if satterthwaite)
  out <- lav_test_ug_trace_stream(
    m_gamma = omega, delta = delta, wls_v = a1_group, e_inv = e_inv,
    fg = fg, satterthwaite = satterthwaite,
    old_approach = ug2_old_approach
  )

  # trace.h0 per group; fg cancels against the Gamma_g / fg convention:
  # sum( (fg * B1_g) * (Delta E.inv Delta') ) = fg * tr(E.inv Delta' B1 Delta)
  trace_h0 <- numeric(ngroups)
  for (g in seq_len(ngroups)) {
    trace_h0[g] <- fg[g] *
      sum(e_inv * crossprod(delta[[g]], b1_group[[g]] %*% delta[[g]]))
  }

  trace_ugamma <- out$trace.UGamma
  attr(trace_ugamma, "h1") <- trace_h1
  attr(trace_ugamma, "h0") <- trace_h0
  attr(trace_ugamma, "h1.ndat") <- h1_ndat
  if (satterthwaite) {
    attr(trace_ugamma, "trace.UGamma2") <- out$trace.UGamma2
  }

  trace_ugamma
}

lav_test_yb_mplus_trace <- function(lavsamplestats = NULL,
                                              a1_group = NULL,
                                              b1_group = NULL,
                                              b0_group = NULL,
                                              e_inv = NULL,
                                              meanstructure = TRUE) {
  # typical for Mplus:
  # - do NOT use the YB formula, but use an approximation
  #   relying  on A0 ~= Delta' A1 Delta and the same for B0
  #
  # NOTE: if A0 is based on the hessian, then A0 only approximates
  #       Delta' A1 Delta
  #
  # - always use h1.information = "unstructured"!!!

  # ngroups <- lavsamplestats@ngroups

  trace_ugamma <- numeric(lavsamplestats@ngroups)
  trace_h1 <- numeric(lavsamplestats@ngroups)
  trace_h0 <- numeric(lavsamplestats@ngroups)
  h1_ndat <- numeric(lavsamplestats@ngroups)

  for (g in 1:lavsamplestats@ngroups) {
    # group weight
    fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal

    a1_1 <- a1_group[[g]]
    b1 <- b1_group[[g]]

    # mask independent 'fixed-x' variables
    zero_idx <- which(diag(a1_1) == 0)
    if (length(zero_idx) > 0L) {
      a1_inv_1 <- matrix(0, nrow(a1_1), ncol(a1_1))
      a1 <- a1_1[-zero_idx, -zero_idx]
      a1_inv <- solve(a1)
      a1_inv_1[-zero_idx, -zero_idx] <- a1_inv
    } else {
      a1_inv_1 <- solve(a1_1)
    }
    h1_ndat[g] <- ncol(a1_1) - length(zero_idx)

    # if data is complete, why not just A1 %*% Gamma?
    trace_h1[g] <- sum(b1 * t(a1_inv_1))
    trace_h0[g] <- fg * sum(b0_group[[g]] * t(e_inv))
    trace_ugamma[g] <- (trace_h1[g] - trace_h0[g])
  }

  # we take the sum here
  trace_ugamma <- sum(trace_ugamma)

  attr(trace_ugamma, "h1") <- trace_h1
  attr(trace_ugamma, "h0") <- trace_h0
  attr(trace_ugamma, "h1.ndat") <- h1_ndat

  trace_ugamma
}

# This function takes as input a plotinfo structure with two data.frames:
#   nodes: id, naam, tiepe, voorkeur, blok
#   edges: id, label, van, naar, tiepe
# , parameters to control the placement of nodes and edge labels, a
# switch to indicate whether indicators for which a covariance is explicit in
# the model should be grouped at the same side of the graph and a debug
# switch.
# It returns a plotinfo strucure with modified nodes and edges data.frames and
# an integer mlrij giving the position at which a line should be drawn for
# multilevel models.
# The data.frames have new columns defined as follows:
# nodes:
#   rij: index of the row where the node will be placed.
#   kolom: index of the column where the node will be placed.
# edges
#   vananker: character, anchor point for starting node.
#   naaranker: character, anchor point for destination node.
#   controlpt.kol: real, column position of control point if the
#     edge has to be drawn as a quadratic Beziers curve.
#   controlpt.row: real, row position of control point if the
#     edge has to be drawn as a quadratic Beziers curve.
#   labelbelow: logical, TRUE if label has to be positioned under
#     the line, initialized FALSE.
#
lav_plotinfo_positions <- function(
  plotinfo,
  placenodes = NULL,
  edgelabelsbelow = NULL,
  group_covar_indicators = FALSE,
  debug = FALSE
) {
  # add new columns to nodes and edges data.frames
  nodes <- plotinfo$nodes
  nodes$rij <- NA_integer_
  nodes$kolom <- NA_integer_
  edges <- plotinfo$edges
  edges$vananker <- NA_character_
  edges$naaranker <- NA_character_
  edges$controlpt.kol <- NA_real_
  edges$controlpt.rij <- NA_real_
  edges$labelbelow <- FALSE
  # if only one node, place it at (1, 1) and return
  if (length(nodes$rij) == 1L) {
    # Only 1 node, e.g. model = 'x ~~ x'
    nodes$rij[1L] <- 1L
    nodes$kolom[1L] <- 1L
    return(list(
      nodes = nodes,
      edges = edges,
      mlrij = 0L
    ))
  }
  # if there are multiple levels, split the plotinfo in two and call function
  # lav_plotinfo_positions_one separately for the two levels and combine the
  # results
  if (any(nodes$blok > 0L)) {
    # Multilevel, only level:1 and level:2 accepted
    nodes1 <- nodes[nodes$blok >= 2L, ]
    edges1 <- edges[edges$van %in% nodes1$id, ]
    edges1$van <- match(edges1$van, nodes1$id)
    edges1$naar <- match(edges1$naar, nodes1$id)
    nodes1$id <- seq_along(nodes1$tiepe)
    nodes1$blok <- 0L
    nodes2 <- nodes[nodes$blok == 1L, ]
    edges2 <- edges[edges$van %in% nodes2$id, ]
    edges2$van <- match(edges2$van, nodes2$id)
    edges2$naar <- match(edges2$naar, nodes2$id)
    nodes2$id <- seq_along(nodes2$tiepe)
    nodes2$blok <- 0L
    result1 <- lav_plotinfo_positions_one(
      list(nodes = nodes1, edges = edges1),
      placenodes,
      edgelabelsbelow,
      group_covar_indicators,
      debug
    )
    result2 <- lav_plotinfo_positions_one(
      list(nodes = nodes2, edges = edges2),
      placenodes,
      edgelabelsbelow,
      group_covar_indicators,
      debug
    )
    rijen1 <- max(result1$nodes$rij)
    result2$nodes$rij <- result2$nodes$rij + rijen1 + 1L
    result2$edges$controlpt.rij <- result2$edges$controlpt.rij + rijen1 + 1L
    result1$nodes$blok <- 2L
    result2$nodes$blok <- 1L
    result2$nodes$id <- result2$nodes$id + length(result1$nodes)
    result2$edges$van <- result2$edges$van + length(result1$nodes)
    result2$edges$naar <- result2$edges$naar + length(result1$nodes)
    nodes <- rbind(result1$nodes, result2$nodes)
    edges <- rbind(result1$edges, result2$edges)
    return(list(
      nodes = nodes,
      edges = edges,
      mlrij = rijen1 + 1L
    ))
  }
  # if there is only one level, call function lav_plotinfo_positions_one
  plotinfo <- list(nodes = nodes, edges = edges)
  lav_plotinfo_positions_one(
    plotinfo,
    placenodes,
    edgelabelsbelow,
    group_covar_indicators,
    debug
  )
}

# This function computes the positions for the nodes and the anchors and
# control points for the edges in the diagram for a single level.
lav_plotinfo_positions_one <- function(
  plotinfo,
  placenodes,
  edgelabelsbelow,
  group_covar_indicators,
  debug
) {
  nods <- plotinfo$nodes
  edgs <- plotinfo$edges
  # assign groupnumbers to the nodes, i.e. partition the nodes in groups
  # which belong together, e.g. indicators and the corresponding latent
  # variable
  nods <- lav_plotinfo_nodes_groups(plotinfo, group_covar_indicators)
  plotinfo <- list(nodes = nods, edges = edgs)
  # compute the information for the groups
  groups <- lav_plotinfo_groups(plotinfo)
  # order the groups in a matrix via topological sorting
  groups <- lav_groups_order(groups, plotinfo)
  # debug <- TRUE
  if (debug) {
    cat("debug start\nnodes\n")
    nods1 <- within(nods, {
      rm(blok, rij, kolom)
    })
    print(nods1)
    cat("edges\n")
    edgs1 <- within(edgs, {
      rm(vananker, naaranker, controlpt.kol, controlpt.rij)
    })
    print(edgs1)
    rm(edgs1, nods1)
    cat("matrix with groups after ordering\n")
    print(lav_groups_mat(groups))
    cat("debug end\n")
  }
  for (g in seq_along(groups)) {
    group <- groups[[g]]
    group <- lav_group_order(group, plotinfo)
    if (debug && group$nb.nodes > 1L) {
      plot(group$offsets.lin, -group$offsets.out,
        pch = 16, xlim = c(0, 1 + max(group$offsets.lin)),
        axes = FALSE, main = paste("Group", g))
      arrows(0.5, 0, 0.5, -max(group$offsets.out), col = "blue")
      text(0.6, -0.6, "out", col = "blue")
      text(group$offsets.lin  + 0.2, -group$offsets.out,
           nods$naam[group$nodes.id])
    }
    groups[[g]] <- group
  }
  width <- sapply(groups, function(g) g$width.height[1L])
  height <- sapply(groups, function(g) g$width.height[2L])
  columns <- sapply(groups, function(g) g$matrixrowcol[2L])
  rows <- sapply(groups, function(g) g$matrixrowcol[1L])
  colnums <- sort(unique(columns))
  rownums <- sort(unique(rows))
  colwidths <- sapply(colnums, function(i) max(width[columns == i]))
  rowheights <- sapply(rownums, function(i) max(height[rows == i]))
  for (g in groups) {
    thisrow <- g$matrixrowcol[1L]
    thiscol <- g$matrixrowcol[2L]
    gtop <- 1L
    if (thisrow > 1L) gtop <- gtop + cumsum(rowheights)[thisrow - 1L]
    gleft <- 1L
    if (thiscol > 1L) gleft <- gleft + cumsum(colwidths)[thiscol - 1L]
    for (n in seq_along(g$nodes.id)) {
      k <- which(nods$id == g$nodes.id[n])
      if (g$loc == "l" || g$loc == "?") {
        nods$rij[k] <- gtop + g$offsets.lin[n]
        nods$kolom[k] <- gleft + colwidths[g$matrixrowcol[2L]] -
                         g$offsets.out[n] - 1L
      } else if (g$loc == "r") {
        nods$rij[k] <- gtop + g$offsets.lin[n]
        nods$kolom[k] <- gleft + g$offsets.out[n]
      } else if (g$loc == "t") {
        nods$rij[k] <- gtop + rowheights[1] - g$offsets.out[n] - 1L
        nods$kolom[k] <- gleft + g$offsets.lin[n]
      } else if (g$loc == "b") {
        nods$rij[k] <- gtop + g$offsets.out[n]
        nods$kolom[k] <- gleft + g$offsets.lin[n]
      }
    }
  }
  # compress graph by joining adjacent rows without common column-elements and
  # adjacent columns without common row-elements
  maxrow <- max(nods$rij)
  maxcol <- max(nods$kolom)
  for (j in seq(maxrow - 1L, 1L, -1L)) {
    col1 <- nods$kolom[nods$rij == j]
    col2 <- nods$kolom[nods$rij == j + 1L]
    if (all(is.na(match(col1, col2)))) {
      dimins <- which(nods$rij > j)
      nods$rij[dimins] <- nods$rij[dimins] - 1L
    }
  }
  for (j in seq(maxcol - 1L, 1L, -1L)) {
    col1 <- nods$rij[nods$kolom == j]
    col2 <- nods$rij[nods$kolom == j + 1L]
    if (all(is.na(match(col1, col2)))) {
      dimins <- which(nods$kolom > j)
      nods$kolom[dimins] <- nods$kolom[dimins] - 1L
    }
  }
  # check for debug
  if (debug) {
      plot(nods$kolom, -nods$rij,
        pch = 16, xlim = c(0, 1 + max(nods$kolom)), axes = FALSE,
        main = "Absolute positions nodes")
      text(nods$kolom + 0.3, -nods$rij, nods$naam)
  }
  #### place nodes demanded by user ? ####
  if (!is.null(placenodes)) {
    for (nn in names(placenodes)) {
      w <- which(nods$naam == nn)
      if (length(w) == 0) {
        lav_msg_warn(gettextf("placenodes: node name %s not found!", nn))
      }
      nods$rij[w] <- placenodes[[nn]][1L]
      nods$kolom[w] <- placenodes[[nn]][2L]
    }
  }
  #### place anchors ####
  groupmaxcol <- max(sapply(groups, function(g) g$matrixrowcol[2L]))
  groupmaxrow <- max(sapply(groups, function(g) g$matrixrowcol[1L]))
  for (j in seq_along(edgs$id)) {
    thisedge <- edgs[j, ]
    van <- nods[which(nods$id == thisedge$van), ]
    naar <- nods[which(nods$id == thisedge$naar), ]
    vangroup <- groups[[van$group]]
    if (thisedge$tiepe == "=~") { # define latent variable
      if (vangroup$matrixrowcol[[2L]] == 1L) {
        thisedge$vananker <- "w"
        thisedge$naaranker <- "e"
      } else if (vangroup$matrixrowcol[[2L]] == groupmaxcol) {
        thisedge$vananker <- "e"
        thisedge$naaranker <- "w"
      } else if (vangroup$matrixrowcol[[1L]] == 1L) {
        thisedge$vananker <- "n"
        thisedge$naaranker <- "s"
      } else if (vangroup$matrixrowcol[[1L]] == groupmaxrow) {
        thisedge$vananker <- "s"
        thisedge$naaranker <- "n"
      } else {
        lav_msg_fixme(gettext("measurement group in unexpected position"))
      }
    } else if (thisedge$tiepe == "<~") { # define composite variable
      if (vangroup$matrixrowcol[[2L]] == 1L) {
        thisedge$vananker <- "e"
        thisedge$naaranker <- "w"
      } else if (vangroup$matrixrowcol[[2L]] == groupmaxcol) {
        thisedge$vananker <- "w"
        thisedge$naaranker <- "e"
      } else if (vangroup$matrixrowcol[[1L]] == 1L) {
        thisedge$vananker <- "s"
        thisedge$naaranker <- "n"
      } else if (vangroup$matrixrowcol[[1L]] == groupmaxrow) {
        thisedge$vananker <- "n"
        thisedge$naaranker <- "s"
      } else {
        lav_msg_fixme(gettext("measurement group in unexpected position"))
      }
    } else if (thisedge$tiepe == "~." || thisedge$tiepe == "~") {
      # regression/varlv
      if (van$rij == naar$rij) {
        thisedge$vananker <- if (van$kolom < naar$kolom) "e" else "w"
        thisedge$naaranker <- if (van$kolom < naar$kolom) "w" else "e"
      } else if (van$kolom == naar$kolom) {
        thisedge$vananker <- if (van$rij < naar$rij) "s" else "n"
        thisedge$naaranker <- if (van$rij < naar$rij) "n" else "s"
      } else {
        thisedge <- lav_plotinfo_anchors(thisedge, van, naar,
                                         max(nods$kolom), max(nods$rij))
      }
    } else if (thisedge$tiepe == "~~~") { # (remaining) variance
        thisedge <- lav_plotinfo_anchors(thisedge, van, naar,
                                         max(nods$kolom), max(nods$rij))
    } else { # covariance
        thisedge <- lav_plotinfo_anchors(thisedge, van, naar,
                                         max(nods$kolom), max(nods$rij))
    }
    edgs[j, ] <- thisedge
  }
  #### labelsbelow demanded by user ? ####
  if (!is.null(edgelabelsbelow)) {
    for (i in seq_along(edgelabelsbelow)) {
      n1 <- which(nods$naam == edgelabelsbelow[[i]][1L])
      if (length(n1) == 0) {
        lav_msg_warn(gettextf(
          "edgelabelsbelow: node name %s not found!",
          edgelabelsbelow[[i]][1L]
        ))
      }
      n2 <- which(nods$naam == edgelabelsbelow[[i]][2L])
      if (length(n2) == 0) {
        lav_msg_warn(gettextf(
          "edgelabelsbelow: node name %s not found!",
          edgelabelsbelow[[i]][2L]
        ))
      }
      ed <- which(
        edgs$van == nods$id[n1] &
          edgs$naar == nods$id[n2]
      )
      if (length(ed) == 0L) {
        ed <- which(edgs$naar == nods$id[n1] & edgs$van == nods$id[n2])
      }
      if (length(ed) == 0L) {
        lav_msg_warn(
          gettextf(
            "edgelabelsbelow: edge %s -- %s not found!",
            nods$naam[n1],
            nods$naam[n2]
          )
        )
      }
      edgs$labelbelow[ed] <- TRUE
    }
  }
  #### RETURN ####
  list(nodes = nods, edges = edgs, mlrij = 0L)
}

lav_plotinfo_nodes_groups <- function(plotinfo, group_covar_indicators) {
  nodes <- plotinfo$nodes
  edges <- plotinfo$edges
  nodes$group <- seq.int(nrow(nodes)) # each node in its own group
  merge_groups <- function(group, i1, i2) { # merge group of i1 into group of i2
    group[group == group[i1]] <- group[i2]
    group
  }
  indicator_ids <- lav_plotinfo_indicator_ids(nodes, edges)
  # process edges to form groups
  for (j in seq.int(nrow(edges))) {
    van_id <- which(nodes$id == edges$van[j])
    naar_id <- which(nodes$id == edges$naar[j])
    merge_them <- FALSE
    if (edges$tiepe[j] == "=~") {
      if (nodes$tiepe[naar_id] != "lv" && nodes$tiepe[naar_id] != "cv") {
        merge_them <- TRUE
      }
    } else if (edges$tiepe[j] == "<~") {
      if (nodes$tiepe[van_id] != "lv" && nodes$tiepe[van_id] != "cv") {
        merge_them <- TRUE
      }
    } else if (edges$tiepe[j] == "~.") {
      merge_them <- TRUE
    } else if (edges$tiepe[j] == "~~") {
      if (group_covar_indicators &&
        any(edges$van[j] == indicator_ids) &&
        any(edges$naar[j] == indicator_ids)
      ) {
        merge_them <- TRUE
      }
    } else if (edges$tiepe[j] == "~") {
      if (nodes$tiepe[van_id] == "varlv") {
        merge_them <- TRUE
      }
    }
    if (merge_them) {
      nodes$group <- merge_groups(nodes$group, naar_id, van_id)
    }
  }
  nodes$group <- match(nodes$group, unique(nodes$group)) # renumber 1:n
  nodes
}
lav_plotinfo_indicator_ids <- function(nodes, edges) {
  # compute ids of observed indicators
  indicator_ids <- unique(c(
    edges$naar[edges$tiepe == "=~"],
    edges$van[edges$tiepe == "<~"]
  ))
  for (i in seq_along(indicator_ids)) {
    indicator_id <- which(nodes$id == indicator_ids[i])
    if (
      nodes$tiepe[indicator_id] == "lv" ||
        nodes$tiepe[indicator_id] == "cv"
    ) {
      indicator_ids[i] <- 0L
    }
  }
  indicator_ids <- indicator_ids[indicator_ids != 0L]
  indicator_ids
}
lav_plotinfo_groups <- function(plotinfo) {
  nodes <- plotinfo$nodes
  edges <- plotinfo$edges
  nb_of_groups <- max(nodes$group)
  indicator_ids <- lav_plotinfo_indicator_ids(nodes, edges)
  groups <- lapply(seq_len(nb_of_groups), function(g) {
    group <- list(
      id = g,
      nodes.id = integer(),
      edges.id = integer(),
      offsets.out = integer(),
      offsets.lin = integer(),
      nb.nodes = integer(),
      nb.indicators = 0L,
      measurement = FALSE,
      loc = "?",
      matrixrowcol = c(0L, 0L),
      width.height = c(1L, 1L),
      indic = ""
    )
    group$nodes.id <- which(nodes$group == g)
    group$nodes.id <- nodes$id[group$nodes.id]
    group$edges.id <- which(
      edges$van %in% group$nodes.id & edges$naar %in% group$nodes.id
    )
    group$edges.id <- edges$id[group$edges.id]
    group$nb.nodes <- length(group$nodes.id)
    group$nb.indicators <-
      as.integer(sum(nodes$group[nodes$id %in% indicator_ids] == g))
    group$measurement <- (group$nb.indicators > 0L)
    group
  })
  groups
}
lav_groups_mat <- function(groups) {
  maxrow <- 1L
  maxcol <- 1L
  for (group in groups) {
    rowcol <- group$matrixrowcol
    if (rowcol[1L] > maxrow) {
      maxrow <- rowcol[1L]
    }
    if (rowcol[2L] > maxcol) {
      maxcol <- rowcol[2L]
    }
  }
  m <- matrix(0L, nrow = maxrow, ncol = maxcol)
  for (group in groups) {
    rowcol <- group$matrixrowcol
    m[rowcol[1L], rowcol[2L]] <- group$id
  }
  m
}

 # internal ordering of a group
 lav_group_order <- function(group_object, plotinfo) {
  edges <- plotinfo$edges
  group <- group_object
  group$offsets.out <- rep(0L, group$nb.nodes)
  group$offsets.lin <- rep(0L, group$nb.nodes)
  if (group$nb.nodes > 1L) {
    defined <- integer()
    definedby <- integer()
    bordernodes <- integer()
    for (j in group$edges.id) {
      thisedge <- edges[which(edges$id == j), ]
      if (thisedge$tiepe == "~.") {
        definedby <- c(definedby, thisedge$van)
        defined <- c(defined, thisedge$naar)
      } else {
        if (thisedge$tiepe == "=~") {
          definedby <- c(definedby, thisedge$naar)
          defined <- c(defined, thisedge$van)
          bordernodes <- c(bordernodes, thisedge$van)
        } else if (thisedge$tiepe == "<~") {
          definedby <- c(definedby, thisedge$van)
          defined <- c(defined, thisedge$naar)
          bordernodes <- c(bordernodes, thisedge$naar)
        } else { # include nodes not yet present indefined or definedby
          if (!(thisedge$van %in% definedby || thisedge$van %in% defined)) {
            definedby <- c(definedby, thisedge$van)
            defined <- c(defined, 999L)
          }
          if (!(thisedge$naar %in% definedby || thisedge$naar %in% defined)) {
            definedby <- c(definedby, thisedge$naar)
            defined <- c(defined, 999L)
          }
        }
      }
    }
    internaldf <- lav_graph_topological_mat(defined, definedby,
                                               bordernodes = bordernodes)
    for (j in seq_along(group$nodes.id)) {
      group$offsets.lin[j] <- internaldf$rows[internaldf$nodes ==
                                              group$nodes.id[j]] - 1L
      group$offsets.out[j] <- max(internaldf$cols) -
          internaldf$cols[internaldf$nodes == group$nodes.id[j]]
    }
  }
  if (group$loc == "l" || group$loc == "r" || group$loc == "?") {
    group$width.height <- c(
      1L + max(group$offsets.out),
      1L + max(group$offsets.lin)
    )
  } else {
    group$width.height <- c(
      1L + max(group$offsets.lin),
      1L + max(group$offsets.out)
    )
  }
  group
 }

lav_groups_order <- function(groups, plotinfo) {
  nodes <- plotinfo$nodes
  edges <- plotinfo$edges
  dependencies <- data.frame(defined = integer(0), definedby = integer(0))
  add_dependency <- function(depends, def, defby) {
    if (any(depends$defined == def & depends$definedby == defby)) {
      return(depends)
    }
    rbind(depends, data.frame(defined = def, definedby = defby))
  }
  for (j in seq_along(edges$id)) {
    van_id <- which(nodes$id == edges$van[j])
    naar_id <- which(nodes$id == edges$naar[j])
    if (nodes$group[van_id] == nodes$group[naar_id]) {
      next
    }
    if (edges$tiepe[j] == "=~") {
      if (nodes$tiepe[naar_id] == "lv" || nodes$tiepe[naar_id] == "cv") {
        dependencies <- add_dependency(
          dependencies,
          nodes$group[van_id],
          nodes$group[naar_id]
        )
      }
    } else if (edges$tiepe[j] == "<~") {
      if (nodes$tiepe[van_id] == "lv" || nodes$tiepe[van_id] == "cv") {
        dependencies <- add_dependency(
          dependencies,
          nodes$group[naar_id],
          nodes$group[van_id]
        )
      }
    } else if (edges$tiepe[j] == "~") {
      dependencies <- add_dependency(
        dependencies,
        nodes$group[naar_id],
        nodes$group[van_id]
      )
    }
  }
  bordernodes <- integer(0)
  for (g in groups) {
    if (g$measurement) bordernodes <- c(bordernodes, g$id)
  }
  if (nrow(dependencies) == 0L) {
    for (g in groups) dependencies <- add_dependency(dependencies, 999, g$id)
  }
  groupmatrixdf <- lav_graph_topological_mat(
    dependencies$defined,
    dependencies$definedby,
    bordernodes = bordernodes,
    warn = TRUE
  )
  for (g in seq_along(groups)) {
    group <- groups[[g]]
    gmcol <- which(groupmatrixdf$nodes == group$id)
    group$matrixrowcol <- as.integer(c(
      groupmatrixdf$rows[gmcol],
      groupmatrixdf$cols[gmcol]
    ))
    group$indic <- groupmatrixdf$indic[gmcol]
    groups[[g]] <- group
  }
  rm(groupmatrixdf)
  # Set loc = "l" for all measurement groups in first column.
  # Set loc = "r" for all measurement groups in last column.
  # Set loc = "t" for measurement groups in first row and
  #                               not first or last column.
  # Set loc = "b" for measurement groups in another row and
  #                               not first or last column.
  group_matrix <- lav_groups_mat(groups)
  gmcols <- ncol(group_matrix)
  for (g in seq_along(groups)) {
    group <- groups[[g]]
    if (group$measurement) {
      if (group$matrixrowcol[2L] == 1L) {
        group$loc <- "l"
      } else if (group$matrixrowcol[2L] == gmcols) {
        group$loc <- "r"
      } else {
        if (group$matrixrowcol[1L] == 1L) {
          group$loc <- "t"
        } else {
          group$loc <- "b"
        }
      }
      groups[[g]] <- group
    }
  }
  groups
}

lav_points_normalform <- function(p1, p2) {
  # compute normal form of a line
  # the constant in the normal form will always be <= 0
  xy1 <- matrix(
    c(
      p1, 1, p2, 1
    ),
    byrow = TRUE,
    ncol = 3
  )
  a <- det(xy1[, 2:3])
  b <- -det(xy1[, c(
    1L, 3L
  )])
  c <- det(xy1[, 1:2])
  fac <- 1 / sqrt(a * a + b * b)
  if (c > 0) {
    fac <- -fac
  }
  fac *
    c(
      a, b, c
    )
}
lav_pointslope_normalform <- function(p, slope) {
  # compute normal form of a line defined by intercept a and slope b
  # a vertical line is defined by x = a and b = Inf or -Inf
  if (is.infinite(slope)) {
    return(c(
      1, 0, -p[1]
    ))
  }
  coeffs <- c(
    -slope, 1, slope * p[1] - p[2]
  ) /
    sqrt(1 + slope^2)
  if (coeffs[3] > 0) {
    coeffs <- -coeffs
  }
  coeffs
}
lav_lines_intersection <- function(line1, line2) {
  # get the intersection point of two lines for which coefficients of
  # equations are given
  m <- matrix(
    c(
      line1[1:2], line2[1:2]
    ),
    byrow = TRUE,
    ncol = 2L
  )
  b <- matrix(
    c(
      -line1[3], -line2[3]
    ),
    ncol = 1L
  )
  if (abs(det(m)) < 1e-12) {
    return(NA_real_)
  }
  as.vector(solve(m) %*% b)
}
lav_edge_bezierscontrolpoint <- function(van, naar, maxrij, maxcol) {
  middelpunt <- c(
    maxrij + 1, maxcol + 1
  ) /
    2
  delta <- sqrt(sum((van - naar)^2)) /
    sqrt(sum(
      c(
        maxrij - 1, maxcol - 1
      )^2
    ))
  lijn <- lav_points_normalform(van, naar)
  middenlijn <- (van + naar) / 2
  lijnmidden <- sum(
    lijn *
      c(
        middelpunt, 1
      )
  )
  if (lijnmidden > 0) {
    lijn[3] <- lijn[3] + 0.5 + delta
  } else {
    lijn[3] <- lijn[3] - 0.5 - delta
  }
  orthoslope <- lijn[2] / lijn[1]
  loodlijn <- lav_pointslope_normalform(middenlijn, orthoslope)
  lav_lines_intersection(lijn, loodlijn)
}
lav_edge_bezierscp_corner <- function(van, naar, wvannaar, maxrij, maxcol) {
  dif <- (abs(van[1L] - naar[1L]) + abs(van[2L] - naar[2L]) - 1) /
    (maxrij + maxcol - 2)
  p <- switch(wvannaar,
    ne = ,
    en = c(
      1, maxcol
    ) +
      c(
        -dif, dif
      ),
    nw = ,
    wn = c(
      1, 1
    ) +
      c(
        -dif, -dif
      ),
    se = ,
    es = c(
      maxrij, maxcol
    ) +
      c(
        dif, dif
      ),
    sw = ,
    ws = c(
      maxrij, 1
    ) +
      c(
        dif, -dif
      )
  )
  2 * (p - 0.25 * (van + naar))
}

lav_plotinfo_anchors <- function(edge, nodevan, nodenaar, maxkol, maxrij) {
  if (edge$tiepe == "~~~") {
    if (nodevan$kolom == 1L) {
      edge$vananker <- "w"
      edge$naaranker <- "w"
    } else if (nodevan$kolom == maxkol) {
      edge$vananker <- "e"
      edge$naaranker <- "e"
    } else if (nodevan$rij == 1L) {
      edge$vananker <- "n"
      edge$naaranker <- "n"
    } else if (nodevan$rij == maxrij) {
      edge$vananker <- "s"
      edge$naaranker <- "s"
    } else {
      edge$vananker <- "n"
      edge$naaranker <- "n"
    }
  } else if (edge$tiepe == "~~") {
    thrucorner <- FALSE
    if (edge$tiepe == "~~") {
      if (nodevan$kolom == nodenaar$kolom) {
        if (nodevan$kolom < maxkol / 2) {
          edge$vananker <- edge$naaranker <- "w"
        } else {
          edge$vananker <- edge$naaranker <- "e"
        }
      } else if (nodevan$rij == nodenaar$rij) {
        if (nodevan$rij < maxrij / 2) {
          edge$vananker <- edge$naaranker <- "n"
        } else {
          edge$vananker <- edge$naaranker <- "s"
        }
      } else {
        if (nodevan$rij < nodenaar$rij) {
          if (nodevan$kolom < nodenaar$kolom) {
            if (nodevan$kolom == 1L && nodenaar$rij == maxrij) {
              edge$vananker <- "w"
              edge$naaranker <- "s"
              thrucorner <- TRUE
            } else {
              edge$vananker <- "s"
              edge$naaranker <- "w"
            }
          } else {
            if (nodevan$kolom == maxkol && nodenaar$rij == maxrij) {
              edge$vananker <- "e"
              edge$naaranker <- "s"
              thrucorner = TRUE
            } else {
              edge$vananker <- "s"
              edge$naaranker <- "e"
            }
            }
        } else {
          if (nodevan$kolom < nodenaar$kolom) {
            if (nodevan$kolom == 1L && nodenaar$rij == 1L) {
              edge$vananker <- "w"
              edge$naaranker <- "n"
              thrucorner <- TRUE
            } else {
              edge$vananker <- "n"
              edge$naaranker <- "w"
            }
          } else {
            if (nodevan$kolom == maxkol && nodenaar$rij == 1L) {
              edge$vananker <- "e"
              edge$naaranker <- "n"
              thrucorner <- TRUE
            } else {
              edge$vananker <- "n"
              edge$naaranker <- "e"
            }
          }
        }
      }
      if (thrucorner) {
        bc <- lav_edge_bezierscp_corner(
                  c(
          nodevan$rij, nodevan$kolom
        ),
        c(
          nodenaar$rij, nodenaar$kolom
        ),
        paste0(edge$vananker, edge$naaranker),
        maxrij,
        maxkol
        )
      } else {
      bc <- lav_edge_bezierscontrolpoint(
        c(
          nodevan$rij, nodevan$kolom
        ),
        c(
          nodenaar$rij, nodenaar$kolom
        ),
        maxrij,
        maxkol
      )
      }
      edge$controlpt.rij <- bc[1L]
      edge$controlpt.kol <- bc[2L]
    }
  } else {
    breaks <- c(
    -pi - 0.01,
    -7 * pi / 8,
    -5 * pi / 8,
    -3 * pi / 8,
    -pi / 8,
    pi / 8,
    3 * pi / 8,
    5 * pi / 8,
    7 * pi / 8,
    pi + 0.01
    )
    winds <- c("w", "sw", "s", "se", "e", "ne", "n", "nw", "w")
    hoek <- atan2(nodevan$rij - nodenaar$rij, nodenaar$kolom - nodevan$kolom)
    wind <- cut(hoek, breaks, winds)
    edge$vananker <- as.character(wind)
    if (hoek > 0) {
      hoek <- hoek - pi
    } else {
      hoek <- hoek + pi
    }
    wind <- cut(hoek, breaks, winds)
    edge$naaranker <- as.character(wind)
  }
  edge
}

# get missing patterns
lav_data_mi_patterns <- function(y, sort_freq = FALSE, coverage = FALSE,
                                      lp = NULL) {
  # handle two-level data
  if (!is.null(lp)) {
    y_orig <- y
    z <- NULL
    if (length(lp$between.idx[[2]]) > 0L) {
      y <- y[, -lp$between.idx[[2]], drop = FALSE]
      z_idx <- which(!duplicated(lp$cluster.idx[[2]]))
      z <- y_orig[z_idx, lp$between.idx[[2]], drop = FALSE]
    }
  }

  # construct TRUE/FALSE matrix: TRUE if value is observed
  obs <- !is.na(y)

  # empty cases
  empty_idx <- which(rowSums(obs) == 0L)

  # pattern of observed values per observation
  case_id <- apply(1L * obs, 1L, paste, collapse = "")

  # remove empty patterns
  if (length(empty_idx)) {
    case_id_nonempty <- case_id[-empty_idx]
  } else {
    case_id_nonempty <- case_id
  }

  # sort non-empty patterns (from high occurrence to low occurrence)
  if (sort_freq) {
    table_1 <- sort(table(case_id_nonempty), decreasing = TRUE)
  } else {
    table_1 <- table(case_id_nonempty)
  }

  # unique pattern ids
  pat_id <- names(table_1)

  # number of patterns
  pat_npatterns <- length(pat_id)

  # case idx per pattern
  pat_case_idx <- lapply(
    seq_len(pat_npatterns),
    function(p) which(case_id == pat_id[p])
  )

  # unique pattern frequencies
  pat_freq <- as.integer(table_1)

  # first occurrence of each pattern
  pat_first <- match(pat_id, case_id)

  # TRUE/FALSE for each pattern
  pat_obs <- obs[pat_first, , drop = FALSE] # observed per pattern

  mp <- list(
    npatterns = pat_npatterns, id = pat_id, freq = pat_freq,
    case.idx = pat_case_idx, pat = pat_obs, empty.idx = empty_idx,
    nel = sum(obs)
  )

  if (coverage) {
    # FIXME: if we have empty cases, include them in N?
    # no for now
    mp$coverage <- crossprod(obs) / sum(pat_freq)
    # Mp$coverage <- crossprod(OBS) / NROW(Y)
  }

  # additional info if we have two-level data
  if (!is.null(lp)) {
    mp$j.idx <- lapply(
      seq_len(pat_npatterns),
      function(p) lp$cluster.idx[[2]][mp$case.idx[[p]]]
    )
    mp$j1.idx <- lapply(
      seq_len(pat_npatterns),
      function(p) sort(unique.default(mp$j.idx[[p]])) # new in 0.6-19: sort!!
                                                    # because table/tabulate
                            # also sorts...
    )
    mp$j.freq <- lapply(
      seq_len(pat_npatterns),
      function(p) as.integer(unname(table(sort(mp$j.idx[[p]])))) # sort needed?
    )

    # between-level patterns
    if (!is.null(z)) {
      mp$Zp <- lav_data_mi_patterns(z,
        sort_freq = FALSE,
        coverage = FALSE, lp = NULL
      )
    }
  }

  mp
}

# get response patterns (ignore empty cases!)
lav_data_resp_patterns <- function(y) {
  # construct TRUE/FALSE matrix: TRUE if value is observed
  obs <- !is.na(y)

  # empty cases
  empty_idx <- which(rowSums(obs) == 0L)

  # remove empty cases
  if (length(empty_idx) > 0L) {
    y <- y[-empty_idx, , drop = FALSE]
  }

  # ntotal <- nrow(y)
  # nvar <- ncol(y)

  # identify, label and sort response patterns
  id <- apply(y, MARGIN = 1, paste, collapse = "")

  # sort patterns (from high occurrence to low occurrence)
  table_1 <- sort(table(id), decreasing = TRUE)
  order <- names(table_1)
  npatterns <- length(table_1)
  pat <- y[match(order, id), , drop = FALSE]
  row.names(pat) <- as.character(table_1)

  # handle NA?
  y[is.na(y)] <- -9
  total_patterns <- prod(apply(y, 2, function(x) length(unique(x))))
  empty_patterns <- total_patterns - npatterns
  # return a list
  # out <- list(nobs=ntotal, nvar=nvar,
  #            id=id, npatterns=npatterns,
  #            order=order, pat=pat)

  # only return pat
  out <- list(
    npatterns = npatterns, pat = pat, total.patterns = total_patterns,
    empty.patterns = empty_patterns
  )

  out
}

# get cluster information
# - cluster can be a vector!
# - clus can contain multiple columns!
lav_data_cl_patterns <- function(y = NULL,
                                      clus = NULL, # the cluster ids
                                      cluster = NULL, # the cluster 'names'
                                      multilevel = FALSE,
                                      ov_names = NULL,
                                      ov_names_x = NULL,
                                      ov_names_l = NULL) {
  # how many levels?
  nlevels <- length(cluster) + 1L

  # did we get any data (or is this just for lav_data_simulate_old)
  if (!is.null(y) && !is.null(clus)) {
    have_data <- TRUE
  } else {
    have_data <- FALSE
  }

  # check clus
  if (have_data) {
    stopifnot(ncol(clus) == (nlevels - 1L), nrow(y) == nrow(clus))
  }

  cluster_size    <- vector("list", length = nlevels)
  cluster_id      <- vector("list", length = nlevels)
  cluster_idx     <- vector("list", length = nlevels)
  nclusters       <- vector("list", length = nlevels)
  cluster_sizes   <- vector("list", length = nlevels)
  ncluster_sizes  <- vector("list", length = nlevels)
  cluster_size_ns <- vector("list", length = nlevels)

  ov_idx          <- vector("list", length = nlevels)
  ov_x_idx        <- vector("list", length = nlevels)
  ov_y_idx        <- vector("list", length = nlevels)
  both_idx        <- vector("list", length = nlevels)
  within_idx      <- vector("list", length = nlevels)
  within_x_idx    <- vector("list", length = nlevels)
  within_y_idx    <- vector("list", length = nlevels)
  between_idx     <- vector("list", length = nlevels)
  between_x_idx   <- vector("list", length = nlevels)
  between_y_idx   <- vector("list", length = nlevels)
  both_names      <- vector("list", length = nlevels)
  within_names    <- vector("list", length = nlevels)
  within_x_names  <- vector("list", length = nlevels)
  within_y_names  <- vector("list", length = nlevels)
  between_names   <- vector("list", length = nlevels)
  between_x_names <- vector("list", length = nlevels)
  between_y_names <- vector("list", length = nlevels)

  # level-1 is special
  if (have_data) {
    nclusters[[1]] <- NROW(y)
  }

  # higher levels:
  for (l in 2:nlevels) {
    if (have_data) {
      clus_1 <- clus[, (l - 1L)]

    # cluster.id: original cluster identifier
      cluster_id[[l]] <- unique(clus_1)

    # cluster.idx: internal cluster idx (always an integer from 1 to J
    # for every observation;
    # the first cluster id we observe is '1', the second '2', etc...)
      cluster_idx[[l]] <- match(clus_1, cluster_id[[l]])

      cluster_size[[l]] <- tabulate(cluster_idx[[l]])
      nclusters[[l]] <- length(cluster_size[[l]])
      # check if we have more observations than clusters
      if (nclusters[[1]] == nclusters[[l]]) {
        lav_msg_stop(gettext("every cluster contains only one observation."))
      }
      mean_cluster_size <- mean(cluster_size[[l]])
      if (mean_cluster_size < 1.5) {
        lav_msg_warn(gettextf(
          "mean cluster size is %s. This means that many clusters only
          contain a single observation.", mean_cluster_size))
        }
      cluster_sizes[[l]] <- unique(cluster_size[[l]])
      ncluster_sizes[[l]] <- length(cluster_sizes[[l]])
      cluster_size_ns[[l]] <- as.integer(table(factor(cluster_size[[l]],
        levels = as.character(cluster_sizes[[l]])
      )))
    } else {
      cluster_id[[l]] <- integer(0L)
      cluster_idx[[l]] <- integer(0L)
      cluster_size[[l]] <- integer(0L)
      nclusters[[l]] <- integer(0L)
      cluster_sizes[[l]] <- integer(0L)
      ncluster_sizes[[l]] <- integer(0L)
      cluster_size_ns[[l]] <- integer(0L)
    }
  }

  # for all levels:
  if (multilevel) {
    for (l in 1:nlevels) {
      # index of ov.names for this level
      ov_idx[[l]] <- match(ov_names_l[[l]], ov_names)

      # new in 0.6-12: always preserve the order of ov.idx[[l]]
      idx <- which(ov_names %in% ov_names_l[[1]] &
        ov_names %in% ov_names_l[[2]])
      both_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]

      idx <- which(ov_names %in% ov_names_l[[1]] &
        !ov_names %in% ov_names_l[[2]])
      within_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]
      # backwards compatibility: also store in within.idx[[2]]
      if (l == 2) {
        within_idx[[l]] <- within_idx[[1]]
      }

      idx <- which(!ov_names %in% ov_names_l[[1]] &
        ov_names %in% ov_names_l[[2]])
      between_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]

      # names
      # both.names[[l]]     <- ov.names[ ov.names %in% ov.names.l[[1]] &
      #                                 ov.names %in% ov.names.l[[2]] ]
      # within.names[[l]]   <- ov.names[ ov.names %in% ov.names.l[[1]] &
      #                                !ov.names %in% ov.names.l[[2]] ]
      # between.names[[l]]  <- ov.names[!ov.names %in% ov.names.l[[1]] &
      #                                 ov.names %in% ov.names.l[[2]] ]
      both_names[[l]] <- ov_names[both_idx[[l]]]
      within_names[[l]] <- ov_names[within_idx[[l]]]
      between_names[[l]] <- ov_names[between_idx[[l]]]
    }
  }

  # fixed.x wrt variable index
  if (multilevel && length(ov_names_x) > 0L) {
    for (l in 1:nlevels) {
      # some ov.names.x could be 'split', and end up in both.names
      # they should NOT be part ov.x.idx (as they become latent variables)
      idx <- which(ov_names %in% ov_names_x &
        ov_names %in% ov_names_l[[l]] &
        !ov_names %in% unlist(both_names))
      ov_x_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]

      # not any longer, we split them, but still treat them as 'fixed'
      # ov.x.idx[[l]] <- which( ov.names %in% ov.names.x &
      #                        ov.names %in% ov.names.l[[l]] )

      # if some ov.names.x have been 'split', and end up in both.names,
      # they should become part of ov.y.idx (despite being exogenous)
      # as they are now latent variables
      idx <- which(ov_names %in% ov_names_l[[l]] &
        !ov_names %in% ov_names_x[!ov_names_x %in% unlist(both_names)])
      ov_y_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]

      # not any longer, ov.x stays ov.x (even if we split)
      # ov.y.idx[[l]] <- which( ov.names %in% ov.names.l[[l]] &
      #                       !ov.names %in% ov.names.x )


      # if(l == 1L) {
      #    next
      # }
      # below, we only fill in the [[2]] element (and higher)

      idx <- which(ov_names %in% ov_names_l[[1]] &
        !ov_names %in% ov_names_l[[2]] &
        ov_names %in% ov_names_x)
      within_x_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]
      # backwards compatibility: also store in within.x.idx[[2]]
      if (l == 2) {
        within_x_idx[[l]] <- within_x_idx[[1]]
      }

      idx <- which(ov_names %in% ov_names_l[[1]] &
        !ov_names %in% ov_names_l[[2]] &
        !ov_names %in% ov_names_x)
      within_y_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]
      # backwards compatibility: also store in within.y.idx[[2]]
      if (l == 2) {
        within_y_idx[[l]] <- within_y_idx[[1]]
      }

      idx <- which(!ov_names %in% ov_names_l[[1]] &
        ov_names %in% ov_names_l[[2]] &
        ov_names %in% ov_names_x)
      between_x_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]

      idx <- which(!ov_names %in% ov_names_l[[1]] &
        ov_names %in% ov_names_l[[2]] &
        !ov_names %in% ov_names_x)
      between_y_idx[[l]] <- ov_idx[[l]][ov_idx[[l]] %in% idx]

      # within.x.names[[l]]   <- ov.names[ ov.names %in% ov.names.l[[1]] &
      #                                   ov.names %in% ov.names.x &
      #                                  !ov.names %in% ov.names.l[[2]] ]
      # within.y.names[[l]]   <- ov.names[ ov.names %in% ov.names.l[[1]] &
      #                                  !ov.names %in% ov.names.x &
      #                                  !ov.names %in% ov.names.l[[2]] ]
      # between.x.names[[l]]  <- ov.names[!ov.names %in% ov.names.l[[1]] &
      #                                   ov.names %in% ov.names.x &
      #                                   ov.names %in% ov.names.l[[2]] ]
      # between.y.names[[l]]  <- ov.names[!ov.names %in% ov.names.l[[1]] &
      #                                  !ov.names %in% ov.names.x &
      #                                   ov.names %in% ov.names.l[[2]] ]
      within_x_names[[l]] <- ov_names[within_x_idx[[l]]]
      within_y_names[[l]] <- ov_names[within_y_idx[[l]]]
      between_x_names[[l]] <- ov_names[between_x_idx[[l]]]
      between_y_names[[l]] <- ov_names[between_y_idx[[l]]]
    }
  } else {
    ov_y_idx <- ov_idx
  }

  out <- list(
    ov.names = ov_names, ov.names.x = ov_names_x, # for this group
    cluster = cluster, # clus = clus,
    # per level
    nclusters = nclusters,
    cluster.size = cluster_size, cluster.id = cluster_id,
    cluster.idx = cluster_idx, cluster.sizes = cluster_sizes,
    ncluster.sizes = ncluster_sizes,
    cluster.size.ns = cluster_size_ns,
    ov.idx = ov_idx, ov.x.idx = ov_x_idx, ov.y.idx = ov_y_idx,
    both.idx = both_idx, within.idx = within_idx,
    within.x.idx = within_x_idx, within.y.idx = within_y_idx,
    between.idx = between_idx,
    between.x.idx = between_x_idx, between.y.idx = between_y_idx,
    both.names = both_names, within.names = within_names,
    within.x.names = within_x_names,
    within.y.names = within_y_names,
    between.names = between_names,
    between.x.names = between_x_names,
    between.y.names = between_y_names
  )

  out
}

# handle ov_order = "data" by adding attribute "ovda" to FLAT
lav_pt_ov_from_data <- function(flat = NULL,
                                      data = NULL,
                                      sample_cov = NULL,
                                      slot_data = NULL) {
  # current model-based ov.names
  ov_names <- lav_pt_vnames(flat, type = "ov")

  # get data-based ov.names
  data_names <- NULL
  if (!is.null(data)) {
    data_names <- names(data)
  } else if (!is.null(sample_cov)) {
    # multiple group/blocks?
    if (is.list(sample_cov)) {
      data_names <- unique(unlist(lapply(sample_cov, colnames)))
      if (is.null(data_names)) {
        # try again with rows
        data_names <- unique(unlist(lapply(sample_cov, rownames)))
      }
    } else {
      data_names <- colnames(sample_cov)
      if (is.null(data_names)) {
        # try again with rows
        data_names <- rownames(sample_cov)
      }
    }
  } else if (!is.null(slot_data)) {
    data_names <- unique(unlist(slot_data@ov.names))
  }

  if (is.null(data_names) || length(data_names) == 0L) {
    lav_msg_stop(gettext("could not find variable names in data/sample.cov"))
  }

  # extract needed ov.names in the same order as the data
  ov_names_data <- data_names[data_names %in% ov_names]

  # check if we have all of them
  if (length(ov_names_data) != length(ov_names)) {
    idx_missing <- which(!(ov_names %in% ov_names_data))
    lav_msg_stop(gettextf(
      "some (observed) variables specified in the model are not found
      in the data: %s",
      lav_msg_view(ov_names[idx_missing], "none")))
  }

  # check if the order is the same
  if (!identical(ov_names, ov_names_data)) {
    attr(flat, "ovda") <- ov_names_data
  }
  flat
}

# How does lavaan work?

library(lavaan)

# PART 1: from syntax to matrices

model <- 'f =~ x1 + a*x2 + 3*x3'

# parsing the syntax
lavParseModelString(model, as_data_frame = TRUE)
# creates a 'FLAT' initial parameter table
FLAT <- lavParseModelString(model)
lavNames(FLAT)
lavNames(FLAT, "lv")

# lavParTable()
# - first creates FLAT
# - needs information from the data (eg number of groups)
# - builds the initial parameter table
# - the parameter table is the core representation of the model in lavaan
lavParTable(model)

# if sem/cfa is used, more parameters are set free
lavParTable(model, auto = TRUE)

# example: equality constraints using labels
model <- 'f =~ x1 + a*x2 + a*x3'
lavParTable(model, auto = TRUE)

# alternative for 'simple' equality constraints
# (will become the default soon)
lavParTable(model, auto = TRUE, ceq_simple = TRUE)

# explicit equality constraints
model <- 'f =~ x1 + a*x2 + b*x3; a == b'
lavParTable(model, auto = TRUE, ceq_simple = TRUE)

# multiple groups/blocks
model <- 'f =~ x1 + c(a1,a2)*x2 + c(b1, b2)*x3'
lavParTable(model, auto = TRUE, ngroups = 2)

# matrix representation: LISREL-all-y
model <- 'f =~ x1 + x2 + x3'
PT <- lavParTable(model, auto = TRUE, as_data_frame = TRUE)
PT
# map every parameter to a matrix element
MAT <- as.data.frame(lavaan:::lav_lisrel(PT))
cbind(PT, MAT)

# alternative matrix representation: RAM
MAT <- as.data.frame(lavaan:::lav_ram(PT))
cbind(PT, MAT)

# first create default lavoptions
fit <- sem(model = 'f =~x1 + x2 + x3', data = HolzingerSwineford1939)
lavoptions <- fit@Options

# create 'Model' object (S4 class)
# (often called lavmodel in internal functions)
Model <- lavaan:::lav_model(PT, lavoptions = lavoptions)
# another representation of the model, more suitable for computations
Model@nmat
Model@isSymmetric
Model@dimNames

# the matrix representation is in GLIST (group list)
# GLIST may contain the matrices of multiple groups, levels, blocks, ...
MLIST <- Model@GLIST

# MLIST is the matrix representation of a single block
# MLIST is used to compute model-based statistics, e.g.,
# the model-implied variance-covariance matrix
lavaan:::lav_lisrel_sigma(MLIST)




# PART 2: lavaan() workflow:
# 18 steps in total (many of them can be skipped)

# 1. lavParseModelString + ov.names + ngroups + nlevels ...

# 2. set/check options (eg meanstructure = TRUE)
#   lav_options_set() # lav_options.R

# 3. check the (raw) data, or the sample statistics
#   lav_lavdata()
lavdata <- lavaan:::lav_lavdata(data = HolzingerSwineford1939,
                            ov_names = c("x1", "x2", "x3"))
slotNames(lavdata)
lavdata@ov.names
lavdata@ngroups
lavdata@X[[1]] # the raw data in group/block 1

# 4. lavpartable: create the parameter table
# needs to know: how many groups, how many categories, ...
lavpartable <- lavParTable(model = FLAT, auto = TRUE)

# 4b. compute and store partable attributes (ov.names, ov.names.x, ...)
lavpta <- lav_pt_attributes(lavpartable)
lavpta$vnames$ov

# 5. lavsamplestats
# compute sample statistics (cov, mean, Gamma, ...)
lavsamplestats <- lavaan:::lav_samp_from_data(lavdata,
                                                     lavoptions = lavoptions)
slotNames(lavsamplestats)
lavsamplestats@cov[[1]] # observed covariance matrix first group/block
lavsamplestats@icov[[1]] # inverse observed covariance matrix
lavsamplestats@cov.log.det[[1]] # log determinant covariance matrix

# 6. lavh1
# summary statistics of the 'unrestricted' (h1) model
# with complete data, this is trivial (cov, mean)
# when data is missing, we need to estimate cov/mean using EM
lavh1 <- lavaan:::lav_h1_implied_logl(lavdata = lavdata,
                             lavsamplestats   = lavsamplestats,
                             lavoptions       = lavoptions)
lavh1$implied

# 7. parameter bounds (needs lavh1$implied)
# only used if bounded estimation is requested
lavoptions$bounds <- "standard"
lavoptions$optim.bounds <- list(lower = c("ov.var", "loadings"),
                                upper = c("ov.var", "loadings"),
                                min.reliability.marker = 0.1)
lavpartable <- lavaan:::lav_pt_add_bounds(partable = lavpartable,
                lavh1 = lavh1, lavdata = lavdata, lavsamplestats = lavsamplestats,
                lavoptions = lavoptions)
# remove bounds again to save space
lavpartable$lower <- NULL
lavpartable$upper <- NULL

# 8. compute starting values
lavpartable$start <- lavaan:::lav_start(start_method = lavoptions$start,
                               lavpartable     = lavpartable,
                               lavsamplestats  = lavsamplestats)
lavpartable

# 9. lavmodel: create internal model representation (with GLIST)
lavmodel <- lavaan:::lav_model(lavpartable      = lavpartable,
                               lavoptions       = lavoptions)
lavmodel@GLIST

# 10. lavcache: compute some additional summary statistis
# only used when estimator = "PML", and "MML" (for now)
lavcache <- list()

# 11. estimation
# - default: lav_model_est() + nlminb() (quasi-Newton optimization)
# - lav_optim_gn(): Gauss-Newton optimization
# - lav_optim_noniter(): non-iterative procedures
# - lav_mvn_cl_em_h0: EM for multilevel models
lavaan:::lav_verbose(TRUE) # be verbose
x <- try(lavaan:::lav_model_est(lavmodel        = lavmodel,
                            lavpartable     = lavpartable,
                            lavsamplestats  = lavsamplestats,
                            lavdata         = lavdata,
                            lavoptions      = lavoptions,
                            lavcache        = lavcache))
lavaan:::lav_verbose(FALSE) # switch off verbose

# store parameters in lavmodel
lavmodel <- lav_model_set_parameters(lavmodel, x = as.numeric(x))

# store parameters in @ParTable$est
lavpartable$est <- lav_model_get_parameters(lavmodel = lavmodel,
                                            type = "user", extra = TRUE)
lavpartable

# 12. lavimplied + lavloglik
# store model implied statistics in @implied
# if likelihood-based method, store also loglik in @loglik
lavimplied <- lav_model_implied(lavmodel)
lavloglik <- lavaan:::lav_model_loglik(lavdata        = lavdata,
                              lavsamplestats = lavsamplestats,
                              lavimplied     = lavimplied,
                              lavmodel       = lavmodel,
                              lavoptions     = lavoptions)

# 13. compute 'vcov': the variance matrix of the (free) parameters
# this is needed to compute standard errors
VCOV <- lavaan:::lav_model_vcov(lavmodel        = lavmodel,
                       lavsamplestats  = lavsamplestats,
                       lavoptions      = lavoptions,
                       lavdata         = lavdata,
                       lavpartable     = lavpartable,
                       lavcache        = lavcache,
                       lavimplied      = lavimplied,
                       lavh1           = lavh1)
VCOV

# prepare lavvcov slot
lavvcov <- list(se = lavoptions$se, information = lavoptions$information,
                vcov = VCOV)

# store standard errors in parameter table
lavpartable$se <- lavaan::lav_model_vcov_se(lavmodel = lavmodel,
                                    lavpartable = lavpartable,
                                    vcov = VCOV)

# 14. compute global test statistic (chi-square)
# trivial for standard test (=N * F_ML)
# more work for 'robust' test statistics (eg test = "satorra-bentler")
TEST <- lavaan:::lav_model_test(lavmodel            = lavmodel,
                       lavpartable         = lavpartable,
                       lavsamplestats      = lavsamplestats,
                       lavimplied          = lavimplied,
                       lavh1               = lavh1,
                       lavoptions          = lavoptions,
                       x                   = x,
                       vcov_1              = VCOV,
                       lavdata             = lavdata,
                       lavcache            = lavcache,
                       lavloglik           = lavloglik)
lavtest <- TEST

# 14bis. lavfit
# store 'fit'information
# no longer used, but if I remove it, a dozen (old) packages break...

# 15. fit baseline model
fit.indep <- try(lavaan:::lav_object_independence(object = NULL,
                                         lavsamplestats = lavsamplestats,
                                         lavdata        = lavdata,
                                         lavcache       = lavcache,
                                         lavoptions     = lavoptions,
                                         lavh1          = lavh1), silent = TRUE)

# 16. rotation
# only needed if efa() blocks are defined
# lavmodel <- lav_model_efa_rotate(lavmodel   = lavmodel,
#                                  lavoptions = lavoptions)

# 17. create lavaan object

# don't run, as some pieces have not been created...
# lavaan <- new("lavaan",
#               version      = as.character(packageVersion("lavaan")),
#               call         = mc,                  # match.call
#               timing       = timing,              # list
#               Options      = lavoptions,          # list
#               ParTable     = lavpartable,         # list
#               pta          = lavpta,              # list
#               Data         = lavdata,             # S4 class
#               SampleStats  = lavsamplestats,      # S4 class
#               Model        = lavmodel,            # S4 class
#               Cache        = lavcache,            # list
#               Fit          = lavfit,              # S4 class
#               boot         = lavboot,             # list
#               optim        = lavoptim,            # list
#               implied      = lavimplied,          # list
#               loglik       = lavloglik,           # list
#               vcov         = lavvcov,             # list
#               test         = lavtest,             # list
#               h1           = lavh1,               # list
#               baseline     = lavbaseline,         # list
#               internal     = list(),              # empty list
#               external     = list()               # empty list
# )

# 18. post-fitting check of parameters
# lavInspect(lavaan, "post.check")

# the sem/cfa/growth function just set some
# options to user-friendly settings:

# default options for sem/cfa call
# mc$int.ov.free     = TRUE
# mc$int.lv.free     = FALSE
# #mc$auto.fix.first  = !std.lv
# mc$auto.fix.first  = TRUE # (re)set in lav_options_set
# mc$auto.fix.single = TRUE
# mc$auto.var        = TRUE
# mc$auto.cov.lv.x   = TRUE
# mc$auto.cov.y      = TRUE
# mc$auto.th         = TRUE
# mc$auto.delta      = TRUE
# mc$auto.efa        = TRUE


# PART 3: extractor functions
fit <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939)

lavParameterEstimates(fit)
# = subset of parTable(fit), but with additional columsn (z, pvalues, ...)

lavParameterEstimates(fit, output = "text")
# this is a big part of the summary() output

# summary()
# first creates summary output as a list:
out <- summary(fit)
names(out)
class(out)
# with a specific print function
out

# fit indices
fitMeasures(fit)
fitMeasures(fit, output = "matrix")
fitMeasures(fit, output = "text")

# lavInspect/lavTech
lavInspect(fit, "est")
lavTech(fit, "est")


# PART 4: lavaan slots

fit <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939)
class(fit)
slotNames(fit)

# 1. lavaan version used to create this object
fit@version

# 2. user-specified call
fit@call

# 3. timings of several substeps
# unlist(fit@timing)

# 4. options used for this object
unlist(fit@Options)

# 5. the parameter table
fit@ParTable  # list
parTable(fit) # return as data.frame

# 6. the parameter table attributes
names(fit@pta)
fit@pta$vnames$ov
fit@pta$vidx$ov
fit@pta$nfac
fit@pta$nblocks

# 7. Data slot (S4)
fit@Data # has its own print function
slotNames(fit@Data)
as.data.frame(fit@Data@ov)
str(fit@Data)

# 8. SampleStats (S4)
fit@SampleStats
slotNames(fit@SampleStats)
fit@SampleStats@cov[[1]] # list with element per group

# 9. Model (S4)
slotNames(fit@Model)
fit@Model@x.free.idx # parameter index in parameter table
fit@Model@m.free.idx # parameter index in model matrix

# 10. Cache (list)
# cached information, only used for estimator PML and MML (for now)

# 11. Fit
# deprecated, only kept to avoid breaking some (old) packages (eg rsem)

# 12. boot (list)
# only used if bootstrapping was used
fitb <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939,
            se = "bootstrap", bootstrap = 100L, iseed = 1234, verbose = TRUE)
head(fitb@boot$coef)

# 13. optim -- info about (iterative) optimization process
str(fit@optim)

# 14. loglik -- loglikelihood information (ML estimation only)
unlist(fit@loglik)

# 15. implied -- implied sample statistics (per group)
fit@implied

# 16. vcov -- variance-covariance matrix of the (free) model parameters
fit@vcov

# 17. test -- global test statistics
fit2 <- sem(model = 'f =~ x1 + x2 + x3 + x4', data = HolzingerSwineford1939,
            test = "satorra-bentler")
names(fit2@test)
fit2@test$satorra.bentler

# 18. h1 -- sample statistics + logl unrestricted/saturated model
# often similar to info in @SampleStats, but not if missing data,
# multilevel data, ...
fit@h1$implied # this is what is used for lavInspect(fit, "implied")
fit@h1$logl

# 19. baseline -- information about the baseline model (needed for CFI/TLI)
names(fit@baseline)
as.data.frame(fit@baseline$partable)
fit@baseline$test$standard

# 20. internal -- list to store specials flags/info, for internal use only

# 21. external -- list to store specials flags/info if you are an external
# developer


# PART 5: source code structure

# 5 main functions:

# 1) xxx_lavaan.R
# 2) xxx_lavaanList.R
# 3) xxx_efa.R (new in 0.6-13)
# 4) xxx_sam.R
# 5) xxx_fsr.R (only for repro reasons; access via lavaan:::fsr())

# files that start with ctr_  contain contributed code
# written by others (although we are slowly converting them to lav_
# files)

# 00class.R contains S4 class definitions
# 00generic.R defines 'generic' functions (that can be specialized)
# fitMeasures(), lavInspect(), lavTech()

# zzz_OLDNAMES.R contains aliases for (very old) function names
# that are still used by some packages
# eg computeExpectedInformation <- lav_model_information_expected

# zzz.R traditionally contains the (in)famous startup message

# most files start with the lav_ prefix
# the second term often refers the type of object for which the file
# contains functions, for example
# lav_matrix.R
# lav_partable_subset.R
# lav_model_estimate.R
# lav_object_post_check.R

# but sometimes, it refers to what is created, or what is done
# lav_test.R  # creates @test slot
# lav_print.R # prints various objects

# for ML, an important set of functions are:
# lav_mvnorm.R
# lav_mvnorm_h1.R
# lav_mvnorm_missing.R
# lav_mvnorm_missing_h1.R
# lav_mvnorm_cluster.R
# lav_mvnorm_cluster_missing.R

# the standard ML discrepancy function is found at the top of
# lav_objective.R

# sometimes, lavaan needs to do trivial things (like regression), but
# needs a bit more info than what we get from standard functions (like lm):
# lav_uvreg.R          # univariate regression
# lav_uvord.R          # univariate probit regression
# lav_mvreg.R          # multivariate regression
# lav_mvreg_cluster.R  # multivariate twolevel regression

# during iterative estimation, we need to compute the value of the
# objective function (i.e., the discrepancy function) many times

# lav_model_estimate.R # runs the iterations
# and defines the following objective function:

   # function to be minimized
    objective_function <- function(x, verbose = FALSE, infToMax = FALSE,
                                   debug = FALSE) {
        # 2. unpack
        if(lavmodel@eq.constraints) {
            x <- as.numeric(lavmodel@eq.constraints.K %*% x) +
                            lavmodel@eq.constraints.k0
        }

        # 1. unscale
        x <- x / parscale

        # update GLIST (change `state') and make a COPY!
        GLIST <- lav_model_x2glist(lavmodel, x = x)

        fx <- lav_model_objective(lavmodel       = lavmodel,
                                  glist          = GLIST,
                                  lavsamplestats = lavsamplestats,
                                  lavdata        = lavdata,
                                  lavcache       = lavcache,
                                  verbose        = verbose)

        # only for PML: divide by N (to speed up convergence)
        if(estimator == "PML") {
            fx <- fx / lavsamplestats@ntotal
        }

        if(debug || verbose) {
            cat("  objective function  = ",
                sprintf("%18.16f", fx), "\n", sep="")
        }
        if(debug) {
            cat("Current free parameter values =\n"); print(x); cat("\n")
        }

        if(lavoptions$optim.partrace) {
            PENV$PARTRACE <- rbind(PENV$PARTRACE, c(fx, x))
        }

        # for L-BFGS-B
        #if(infToMax && is.infinite(fx)) fx <- 1e20
        if(!is.finite(fx)) {
            fx.group <- attr(fx, "fx.group")
            fx <- 1e20
            attr(fx, "fx.group") <- fx.group # only for lav_model_fit()
        }

        fx
    }

# lav_model_objective() can be found in lav_model_objective.R
# 1) compute model implied summary statistics (for each group)
#    using eg lav_model_sigma()
# 2) compute value discrepancy function
#    eg lav_model_objective_gls() or lav_model_objective_ml() # see lav_objective.R
# 3) if multiple groups, combine the values using group weights
# 4) return value (fx)

# YR 18 Dec 2015
# - functions to (directly) compute the inverse of 'Gamma' (the asymptotic
#   variance matrix of the sample statistics)
# - often used as 'WLS.V' (the weight matrix in WLS estimation)
#   and when computing the expected information matrix

# NOTE:
#  - three types:
#       1) plain         (conditional_x = FALSE, fixed_x = FALSE)
#       2) fixed_x       (conditional_x = FALSE, fixed_x = TRUE)
#       3) conditional_x (conditional_x = TRUE)
#  - if conditional_x = TRUE, we ignore fixed_x (can be TRUE or FALSE)

# GLS (and DLS with dls.a = 1): the normal-theory weight matrix for one
# group -- the inverse of the NT Gamma, in the (partial) correlation metric
# when correlation = TRUE. Shared by lav_samp_from_data() and
# lav_samp_from_moments().
lav_samp_wls_v_nt_g <- function(m_cov = NULL, m_mean = NULL, m_icov = NULL,
                                cor_idx = NULL, correlation = FALSE,
                                x_idx = integer(0L), fixed_x = FALSE,
                                conditional_x = FALSE,
                                meanstructure = FALSE) {
  if (correlation) {
    gamma_nt <- lav_samp_partial_cor_gamma_nt(
      m_cov         = m_cov,
      cor_idx       = cor_idx,
      meanstructure = meanstructure,
      fixed_x       = fixed_x,
      x_idx         = x_idx
    )
    lav_mat_sym_inverse(gamma_nt)
  } else {
    lav_samp_gamma_inverse_nt(
      m_icov         = m_icov,
      m_cov          = m_cov,
      m_mean         = m_mean,
      rescale        = FALSE,
      x_idx          = x_idx,
      fixed_x        = fixed_x,
      conditional_x  = conditional_x,
      meanstructure  = meanstructure,
      slopestructure = conditional_x
    )
  }
}

# GLS: compute WLS.V %*% x -- where WLS.V is the (unconditional)
# normal-theory weight matrix lav_samp_wls_v_nt_g() would return --
# WITHOUT constructing WLS.V. The cov block of WLS.V is
#   0.5 t(D) (S.inv %x% S.inv) D
# so for each column a of x (in vech metric) we have
#   0.5 t(D) vec(S.inv A S.inv)   with A = vech.reverse(a)
# and the (optional) mean block is S.inv itself (rescaled by v11_scale
# for gls.v11.mplus). Cost is O(nvar^3) per column instead of O(nvar^4).
# note: correlation = TRUE and conditional.x are NOT supported here;
# their weight matrices are not this simple Kronecker sandwich
lav_samp_wls_v_nt_prod <- function(m_icov = NULL, x = NULL,
                                   meanstructure = FALSE,
                                   fixed_x = FALSE,
                                   x_idx = integer(0L),
                                   v11_scale = 1.0) {
  x <- as.matrix(x)
  out <- matrix(0, nrow = NROW(x), ncol = NCOL(x))
  nvar <- NROW(m_icov)
  pstar <- (nvar * (nvar + 1L)) %/% 2L

  # mean block: V11 = S.inv (zero rows/cols for the x-variables if fixed.x)
  if (meanstructure) {
    m_v11 <- m_icov * v11_scale
    if (fixed_x && length(x_idx) > 0L) {
      m_v11[x_idx, ] <- 0
      m_v11[, x_idx] <- 0
    }
    out[seq_len(nvar), ] <- m_v11 %*% x[seq_len(nvar), , drop = FALSE]
    cov_idx <- nvar + seq_len(pstar)
  } else {
    cov_idx <- seq_len(pstar)
  }

  # fixed.x: WLS.V has zero rows/cols for the x/x combinations
  # (see lav_samp_gamma_inverse_nt); emulate by zeroing the corresponding
  # elements of x (the zero columns) and of the result (the zero rows)
  zero_idx <- integer(0L)
  if (fixed_x && length(x_idx) > 0L) {
    m_tmp <- matrix(0L, nvar, nvar)
    m_tmp[lav_mat_vech_idx(nvar)] <- seq_len(pstar)
    zero_idx <- lav_mat_vech(m_tmp[x_idx, x_idx, drop = FALSE])
  }

  x_cov <- x[cov_idx, , drop = FALSE]
  if (length(zero_idx) > 0L) {
    x_cov[zero_idx, ] <- 0
  }
  for (j in seq_len(NCOL(x))) {
    m_a <- lav_mat_vech_rev(x_cov[, j])
    out[cov_idx, j] <- 0.5 * lav_mat_dup_pre(
      matrix(m_icov %*% m_a %*% m_icov, ncol = 1L))
  }
  if (length(zero_idx) > 0L) {
    out[cov_idx[zero_idx], ] <- 0
  }

  out
}

# NORMAL-THEORY
lav_samp_gamma_inverse_nt <- function(m_y = NULL,
                                             m_cov = NULL,
                                             m_icov = NULL,
                                             m_mean = NULL,
                                             rescale = TRUE,
                                             x_idx = integer(0L),
                                             fixed_x = FALSE,
                                             conditional_x = FALSE,
                                             meanstructure = FALSE,
                                             slopestructure = FALSE) {
  # check arguments
  if (length(x_idx) == 0L) {
    conditional_x <- FALSE
    fixed_x <- FALSE
  }

  if (is.null(m_icov)) {
    if (is.null(m_cov)) {
      stopifnot(!is.null(m_y))

      # coerce to matrix
      m_y <- unname(as.matrix(m_y))
      n <- nrow(m_y)
      m_cov <- cov(m_y)
      if (rescale) {
        m_cov <- m_cov * (n - 1) / n # ML version
      }
    }

    m_icov <- solve(m_cov)
  }

  # if conditional_x, we may also need m_cov and m_mean
  if (conditional_x && length(x_idx) > 0L &&
    (meanstructure || slopestructure)) {
    if (is.null(m_cov)) {
      stopifnot(!is.null(m_y))

      # coerce to matrix
      m_y <- unname(as.matrix(m_y))
      n <- nrow(m_y)
      m_cov <- cov(m_y)

      if (rescale) {
        m_cov <- m_cov * (n - 1) / n # ML version
      }
    }

    if (is.null(m_mean)) {
      stopifnot(!is.null(m_y))
      m_mean <- unname(colMeans(m_y))
    }
  }

  # rename
  s_inv <- m_icov
  m_s <- m_cov
  m_m <- m_mean

  # unconditional
  if (!conditional_x) {
    # unconditional - stochastic x
    if (!fixed_x) {
      # if (lav_use_lavaanC()) {
      #   gamma_inv <- lavaanC::m_kronecker_dup_pre_post(s_inv,
      #                                                multiplicator = 0.5)
      # } else {
        gamma_inv <- 0.5 * lav_mat_dup_pre_post(s_inv %x% s_inv)
      # }
      if (meanstructure) {
        gamma_inv <- lav_mat_bdiag(s_inv, gamma_inv)
      }

      # unconditional - fixed x
    } else {
      # handle fixed_x = TRUE
      # if (lav_use_lavaanC()) {
      #   gamma_inv <- lavaanC::m_kronecker_dup_pre_post(s_inv,
      #                                                multiplicator = 0.5)
      # } else {
        gamma_inv <- 0.5 * lav_mat_dup_pre_post(s_inv %x% s_inv)
      # }

      # zero rows/cols corresponding with x/x combinations
      nvar <- NROW(m_icov)
      pstar <- nvar * (nvar + 1) / 2
      m_m <- matrix(0, nvar, nvar)
      m_m[lav_mat_vech_idx(nvar)] <- seq_len(pstar)
      zero_idx <- lav_mat_vech(m_m[x_idx, x_idx, drop = FALSE])
      gamma_inv[zero_idx, ] <- 0
      gamma_inv[, zero_idx] <- 0

      if (meanstructure) {
        s_inv_nox <- s_inv
        s_inv_nox[x_idx, ] <- 0
        s_inv_nox[, x_idx] <- 0
        gamma_inv <- lav_mat_bdiag(s_inv_nox, gamma_inv)
      }
    }
  } else {
    # conditional_x

    # 4 possibilities:
    # - no meanstructure, no slopes
    # -    meanstructure, no slopes
    # - no meanstructure, slopes
    # -    meanstructure, slopes

    s11 <- s_inv[-x_idx, -x_idx, drop = FALSE]

    # if (lav_use_lavaanC()) {
    #   gamma_inv <- lavaanC::m_kronecker_dup_pre_post(s11, multiplicator = 0.5)
    # } else {
      gamma_inv <- 0.5 * lav_mat_dup_pre_post(s11 %x% s11)
    # }

    if (meanstructure || slopestructure) {
      m_c <- m_s[x_idx, x_idx, drop = FALSE]
      m_mx <- m_m[x_idx]
      c3 <- lav_samp_gamma_c3(m_mx, m_c)
    }

    # the [int|slopes] block of Gamma is cov_ybarx %x% solve(c3) (see
    # lav_samp_gamma_nt): the statistics are in vecr(cbind(res.int,
    # res.slopes)) order (all coefficients of y1, all coefficients of
    # y2, ...), so the fast-running index is the coefficient index;
    # the inverse is then s11 %x% c3 (idem for the sub-cases)
    if (meanstructure) {
      if (slopestructure) {
        a11 <- s11 %x% c3
      } else {
        c11 <- 1 / solve(c3)[1, 1, drop = FALSE]
        a11 <- s11 %x% c11
      }
    } else {
      if (slopestructure) {
        # (solve(c3)[-1, -1])^{-1} == m_c (Schur complement)
        a11 <- s11 %x% m_c
      } else {
        a11 <- matrix(0, 0, 0)
      }
    }

    if (meanstructure || slopestructure) {
      gamma_inv <- lav_mat_bdiag(a11, gamma_inv)
    }
  }

  gamma_inv
}

# loglikelihood clustered/twolevel data in the presence of missing data

# YR:
# - objective function: first version around March 2021 (see Psych paper)
# - analytic gradient: first version around May 2021

# NOTE: see the top of lav_mvnorm_cluster.R for the 'implied' vs '2l'
# representations and the invariants of the conversion functions


# Mu.W, Mu.B, Sigma.W, Sigma.B are the model-implied statistics; a
# pre-converted 2l list (as returned by lav_mvn_cl_implied22l) may be
# passed as out= instead, to avoid re-converting
lav_mvn_cl_mi_loglik_samp_2l <- function(y1 = NULL,
                                         y2 = NULL,
                                         lp = NULL,
                                         mp = NULL,
                                         mu_w = NULL,
                                         sigma_w = NULL,
                                         mu_b = NULL,
                                         sigma_b = NULL,
                                         sinv_method = "eigen",
                                         log2pi = FALSE,
                                         loglik_x = 0,
                                         minus_two = TRUE,
                                         out = NULL) {
  # map implied to 2l matrices (unless the caller already did)
  if (is.null(out)) {
    out <- lav_mvn_cl_implied22l(
      lp = lp, mu_w = mu_w, mu_b = mu_b,
      sigma_w = sigma_w, sigma_b = sigma_b
    )
  }
  mu_y <- out$mu.y
  mu_z <- out$mu.z
  sigma_w_1 <- out$sigma.w
  sigma_b_1 <- out$sigma.b
  sigma_zz <- out$sigma.zz
  sigma_yz <- out$sigma.yz

  # Lp
  nclusters <- lp$nclusters[[2]]
  between_idx <- lp$between.idx[[2]]
  both_idx <- lp$both.idx[[2]]
  cluster_idx <- lp$cluster.idx[[2]]

  # sanity checks
  if (any(diag(sigma_w_1) < 0) || any(diag(sigma_b_1) < 0)) {
    return(+Inf)
  }

  # check is both.idx part of sigma.b is 'too' negative; if so, return +Inf
  ev <- eigen(sigma_b_1[both_idx, both_idx, drop = FALSE],
    symmetric = TRUE,
    only.values = TRUE
  )$values
  if (any(ev < -0.05)) {
    return(+Inf)
  }

  # cat("sigma.w = \n"); print(sigma.w)
  # cat("sigma.b = \n"); print(sigma.b)
  # cat("mu.y = \n"); print(mu.y)

  # global
  sigma_w_inv <- solve.default(sigma_w_1)
  sigma_w_logdet <- log(det(sigma_w_1))
  sigma_b_1 <- sigma_b_1[both_idx, both_idx, drop = FALSE] # only both part

  # y
  ny <- ncol(sigma_w_1)
  if (length(between_idx) > 0L) {
    y1w <- y1[, -between_idx, drop = FALSE]
  } else {
    y1w <- y1
  }
  y1w_c <- t(t(y1w) - mu_y)
  pij <- matrix(0, nrow(y1w_c), ny)

  # z
  nz <- length(between_idx)
  if (nz > 0L) {
    # check is sigma_zz is PD; if not, return +Inf
    ev <- eigen(sigma_zz, symmetric = TRUE, only.values = TRUE)$values
    if (any(ev < sqrt(.Machine$double.eps))) {
      return(+Inf)
    }

    z <- y2[, between_idx, drop = FALSE]
    z_c <- t(t(z) - mu_z)

    sigma_yz <- sigma_yz[both_idx, , drop = FALSE] # only both part
    sigma_zy <- t(sigma_yz)
    sigma_zz_inv <- solve.default(sigma_zz)
    sigma_zz_logdet <- log(det(sigma_zz))
    sigma_zi_zy <- sigma_zz_inv %*% sigma_zy
    sigma_b_z <- sigma_b_1 - sigma_yz %*% sigma_zi_zy
    gz <- z_c %*% sigma_zz_inv # for complete cases only
  }

  # containters per cluster
  q_yy_b <- q_zy <- q_zz_b <- numeric(nclusters)
  ibza_j_logdet <- numeric(nclusters)
  alist_1 <- rep(list(matrix(
    0, length(both_idx),
    length(both_idx)
  )), nclusters)

  # Z per missing pattern
  if (nz > 0L) {
    zp_1 <- mp$Zp
    zpat2j <- integer(nclusters) # which sigma.b.z per cluster
    sigma_b_z_1 <- vector("list", length = zp_1$npatterns + 1L)

    sigma_j_zz_logdet <- q_zz_a <- 0
    for (p in seq_len(zp_1$npatterns)) {
      freq <- zp_1$freq[p]
      z_na_idx <- which(!zp_1$pat[p, ])
      j_idx <- zp_1$case.idx[[p]] # cluster indices with this pattern
      zpat2j[j_idx] <- p

      if (length(z_na_idx) > 0L) {
        # zp <- sigma_zz[-z_na_idx, -z_na_idx, drop = FALSE]
        zp_inv <- lav_mat_sym_inverse_update(
          s_inv = sigma_zz_inv, rm_idx = z_na_idx,
          logdet = TRUE, s_logdet = sigma_zz_logdet
        )
        zp_logdet <- attr(zp_inv, "logdet")
        sigma_j_zz_logdet <- sigma_j_zz_logdet + (zp_logdet * freq)

        gz[j_idx, -z_na_idx] <- z_c[j_idx, -z_na_idx] %*% zp_inv

        yziy <- (sigma_yz[, -z_na_idx, drop = FALSE] %*% zp_inv %*%
          sigma_zy[-z_na_idx, , drop = FALSE])
        sigma_b_z_1[[p]] <- (sigma_b_1 - yziy)
      } else {
        # complete case
        sigma_j_zz_logdet <-
          sigma_j_zz_logdet + (sigma_zz_logdet * freq)
        sigma_b_z_1[[p]] <- sigma_b_z
      }
    } # p

    # add empty patterns (if any)
    if (length(zp_1$empty.idx) > 0L) {
      zpat2j[zp_1$empty.idx] <- p + 1L
      sigma_b_z_1[[p + 1L]] <- sigma_b_1
    }

    q_zz_a <- sum(gz * z_c, na.rm = TRUE)
    gz0 <- gz
    gz0[is.na(gz0)] <- 0
    gj <- gz0 %*% sigma_zy # only both part
  }

  # Y per missing pattern
  w_logdet <- 0
  for (p in seq_len(mp$npatterns)) {
    freq <- mp$freq[p]
    na_idx <- which(!mp$pat[p, ])
    j_idx <- mp$j.idx[[p]]
    j1_idx <- mp$j1.idx[[p]]
    tab <- integer(nclusters)
    tab[j1_idx] <- mp$j.freq[[p]]

    # compute sigma.w.inv for this pattern
    if (length(na_idx) > 0L) {
      wp_inv <- lav_mat_sym_inverse_update(
        s_inv = sigma_w_inv, rm_idx = na_idx,
        logdet = TRUE, s_logdet = sigma_w_logdet
      )
      wp_logdet <- attr(wp_inv, "logdet")
      w_logdet <- w_logdet + (wp_logdet * freq)

      pij[mp$case.idx[[p]], -na_idx] <-
        y1w_c[mp$case.idx[[p]], -na_idx] %*% wp_inv

      a_j <- matrix(0, ny, ny)
      a_j[-na_idx, -na_idx] <- wp_inv
      for (j in j1_idx) {
        alist_1[[j]] <- alist_1[[j]] + (a_j[both_idx, both_idx] * tab[j])
      }
      # WIP[[p]][-na.idx, -na.idx] <- wp.inv
    } else {
      # complete case
      w_logdet <- w_logdet + (sigma_w_logdet * freq)
      pij[mp$case.idx[[p]], ] <-
        y1w_c[mp$case.idx[[p]], ] %*% sigma_w_inv
      for (j in j1_idx) {
        alist_1[[j]] <-
          alist_1[[j]] + (sigma_w_inv[both_idx, both_idx] * tab[j])
      }
    }
  } # p
  q_yy_a <- sum(pij * y1w_c, na.rm = TRUE)
  pj <- rowsum.default(pij[, both_idx], cluster_idx,
    reorder = FALSE,
    na.rm = TRUE
  ) # only both part is needed

  # per cluster
  both_diag_idx <- lav_mat_diag_idx(length(both_idx))
  for (j in seq_len(nclusters)) {
    # we only need the 'both.idx' part of A.j, sigma.b.z, p.j, g.j ,...
    a_j <- alist_1[[j]]
    p_j <- pj[j, ]
    if (nz > 0L) {
      sigma_b_z <- sigma_b_z_1[[zpat2j[j]]]
    } else {
      sigma_b_z <- sigma_b_1
    }
    ibza_j <- sigma_b_z %*% a_j
    ibza_j[both_diag_idx] <- ibza_j[both_diag_idx] + 1
    # logdet IBZA.j
    tmp <- determinant.matrix(ibza_j, logarithm = TRUE)
    ibza_j_logdet[j] <- tmp$modulus * tmp$sign
    # IBZA.j.inv.BZ.p
    ibza_j_inv_bz_p <- solve.default(ibza_j, drop(sigma_b_z %*% p_j))
    q_yy_b[j] <- sum(p_j * ibza_j_inv_bz_p)

    if (nz > 0L) {
      g_j <- gj[j, ]
      ibza_j_inv_g <- solve.default(ibza_j, g_j)
      a_ibza_j_inv_g <- a_j %*% ibza_j_inv_g

      q_zz_b[j] <- sum(g_j * a_ibza_j_inv_g)
      q_zy[j] <- -sum(p_j * ibza_j_inv_g)
    }
  }


  if (nz > 0L) {
    p_1 <- mp$nel + zp_1$nel
    dist_1 <- (q_yy_a - sum(q_yy_b)) + 2 * sum(q_zy) + (q_zz_a + sum(q_zz_b))
    logdet <- w_logdet + sum(ibza_j_logdet) + sigma_j_zz_logdet
  } else {
    p_1 <- mp$nel
    dist_1 <- (q_yy_a - sum(q_yy_b))
    logdet <- w_logdet + sum(ibza_j_logdet)
  }

  # loglik?
  if (log2pi && !minus_two) {
    log_2pi <- log(2 * pi)
    loglik <- -(p_1 * log_2pi + logdet + dist_1) / 2
  } else {
    loglik <- dist_1 + logdet
  }

  # loglik.x (only if loglik is requested)
  if (length(unlist(lp$ov.x.idx)) > 0L && log2pi && !minus_two) {
    loglik <- loglik - loglik_x
  }

  loglik
}

# estimate ML estimates of Mu.W, Mu.B, Sigma.W, Sigma.B in the presence
# of missing data, using the EM algorithm (new in 0.7-1)
#
# this extends lav_mvn_cl_em_sat() to the missing data setting: in the
# E-step, we integrate over both the (latent) cluster random effects and
# the missing values; this mimics the approach used by Mplus (H1
# estimation with TYPE = TWOLEVEL and missing data)
#
# notation (same '2l' parameterization as the loglik function above):
# - y_ij = mu.y + u_j + e_ij, where mu.y = mu.w + mu.b,
#   u_j ~ N(0, sigma.b) (only the 'both' part has nonzero variance), and
#   e_ij ~ N(0, sigma.w)
# - z_j are the between-only variables: z_j = mu.z + (z-part), with
#   Var(z) = sigma.zz and Cov(u, z) = sigma.yz (only the 'both' rows)
#
# E-step, per cluster j (writing beta = u_j['both'] and centering all
# observed data):
# - the observed units contribute the precision A.j = sum_i T_i' W_i^{-1} T_i
#   and the information vector p.j = sum_i T_i' W_i^{-1} (y_i.obs - mu.y),
#   where W_i is the observed block of sigma.w for the pattern of unit i
#   (these are the same quantities used in the loglik function above)
# - the prior of w = (beta, z.mis) given z.obs is N(m0, C0) (standard
#   conditioning of the joint normal); the posterior is then
#       m1 = (I + C0 %*% Atilde)^{-1} (m0 + C0 %*% ptilde)
#       C1 = (I + C0 %*% Atilde)^{-1} C0
#   with Atilde = blkdiag(A.j, 0) and ptilde = c(p.j, 0); this form remains
#   valid even if C0 is singular
# - the missing y-values of unit i, given (beta, y_i.obs), follow the usual
#   within-pattern regression: mean B_p (y.obs - mu.y[obs] - beta[obs]),
#   covariance C_p
#
# the M-step is the standard saturated update based on the expected
# complete-data sufficient statistics
#
# the E-step machinery is also used by the h0 EM optimizer
# (lav_mvn_cl_em_h0, when missing = "ml"); it is therefore exposed as an
# 'engine': a list of closures over the data and the missing patterns:
# - theta0:  packed (available-case) starting values
# - pack():  pack the internal '2l' matrices into a parameter vector
# - step():  one EM step (E-step + saturated M-step update)
# - logl():  observed-data loglikelihood at theta
# - implied(): the implied-space matrices (Sigma.W, Sigma.B, Mu.W, Mu.B)
lav_mvn_cl_mi_em_engine <- function(y1 = NULL,
                                    y2 = NULL,
                                    lp = NULL,
                                    mp = NULL,
                                    loglik_x = 0,
                                    min_variance = 1e-05) {
  if (is.null(loglik_x)) {
    loglik_x <- 0
  }

  # dimensions and indices
  nobs <- lp$nclusters[[1]]
  nclusters <- lp$nclusters[[2]]
  cluster_idx <- lp$cluster.idx[[2]]
  between_idx <- lp$between.idx[[2]]
  both_idx <- lp$both.idx[[2]] # between-only variables always come last,
                               # so these indices are also valid in y-space
  nz <- length(between_idx)
  nb <- length(both_idx)
  zp <- mp$Zp

  # y data (level-1 variables only) and z data (one row per cluster)
  if (nz > 0L) {
    y1w <- y1[, -between_idx, drop = FALSE]
    z <- y2[, between_idx, drop = FALSE]
  } else {
    y1w <- y1
    z <- NULL
  }
  ny <- ncol(y1w)

  # starting values (available-case)
  mu_y0 <- colMeans(y1w, na.rm = TRUE)
  mu_y0[!is.finite(mu_y0)] <- 0
  var_y0 <- apply(y1w, 2L, var, na.rm = TRUE)
  var_y0[!is.finite(var_y0) | var_y0 < min_variance] <- 1
  mu_w <- mu_y0
  mu_w[both_idx] <- 0
  mu_b <- numeric(ny)
  mu_b[both_idx] <- mu_y0[both_idx]
  sigma_w <- diag(var_y0, nrow = ny)
  sigma_w[both_idx, both_idx] <- diag(var_y0[both_idx] / 2,
                                      nrow = nb)
  sigma_b <- matrix(0, ny, ny)
  sigma_b[both_idx, both_idx] <- diag(var_y0[both_idx] / 2, nrow = nb)
  if (nz > 0L) {
    mu_z <- colMeans(z, na.rm = TRUE)
    mu_z[!is.finite(mu_z)] <- 0
    var_z0 <- apply(z, 2L, var, na.rm = TRUE)
    var_z0[!is.finite(var_z0) | var_z0 < min_variance] <- 1
    sigma_zz <- diag(var_z0, nrow = nz)
    sigma_yz <- matrix(0, ny, nz)
  } else {
    mu_z <- numeric(0L)
    sigma_zz <- matrix(0, 0L, 0L)
    sigma_yz <- matrix(0, ny, 0L)
  }

  # z-values per cluster: use the between-level missing patterns (Zp)
  # to decide which entries are observed
  if (nz > 0L) {
    zpat2j <- integer(nclusters)
    for (q in seq_len(zp$npatterns)) {
      zpat2j[zp$case.idx[[q]]] <- q
    }
    if (length(zp$empty.idx) > 0L) {
      zpat2j[zp$empty.idx] <- zp$npatterns + 1L
    }
  }

  # constants for the fused loglikelihood (a byproduct of the E-step;
  # see em_step below)
  p_1 <- mp$nel
  if (nz > 0L) {
    p_1 <- p_1 + zp$nel
  }
  has_x <- length(unlist(lp$ov.x.idx)) > 0L
  log_2pi <- log(2 * pi)

  # pack/unpack the internal parameter vector (structural zeros stay zero
  # under linear extrapolation, so we simply pack the full matrices)
  em_pack <- function(mu_w, mu_b, sigma_w, sigma_b, mu_z, sigma_zz,
                      sigma_yz) {
    c(mu_w, mu_b, lav_mat_vech(sigma_w), lav_mat_vech(sigma_b), mu_z,
      lav_mat_vech(sigma_zz), as.vector(sigma_yz))
  }
  em_unpack <- function(theta) {
    i <- 0L
    mu_w <- theta[i + seq_len(ny)]; i <- i + ny
    mu_b <- theta[i + seq_len(ny)]; i <- i + ny
    sigma_w <- lav_mat_vech_rev(theta[i + seq_len(ny * (ny + 1) / 2)])
    i <- i + ny * (ny + 1) / 2
    sigma_b <- lav_mat_vech_rev(theta[i + seq_len(ny * (ny + 1) / 2)])
    i <- i + ny * (ny + 1) / 2
    mu_z <- theta[i + seq_len(nz)]; i <- i + nz
    sigma_zz <- if (nz > 0L) {
      lav_mat_vech_rev(theta[i + seq_len(nz * (nz + 1) / 2)])
    } else {
      matrix(0, 0L, 0L)
    }
    i <- i + nz * (nz + 1) / 2
    sigma_yz <- matrix(theta[i + seq_len(ny * nz)], ny, nz)
    list(mu_w = mu_w, mu_b = mu_b, sigma_w = sigma_w, sigma_b = sigma_b,
         mu_z = mu_z, sigma_zz = sigma_zz, sigma_yz = sigma_yz)
  }

  # one EM step: theta -> theta'; if logl = TRUE, the observed-data
  # loglikelihood at the *input* theta -- a cheap byproduct of the E-step
  # quantities -- is attached to the result as the "logl" attribute (it
  # matches the value of lav_mvn_cl_mi_loglik_samp_2l() exactly)
  em_step <- function(theta, logl = FALSE) {
    th <- em_unpack(theta)
    mu_w <- th$mu_w
    mu_b <- th$mu_b
    sigma_w <- th$sigma_w
    sigma_b <- th$sigma_b
    mu_z <- th$mu_z
    sigma_zz <- th$sigma_zz
    sigma_yz <- th$sigma_yz

    # current 'both' blocks
    sb <- sigma_b[both_idx, both_idx, drop = FALSE]
    if (nz > 0L) {
      syz <- sigma_yz[both_idx, , drop = FALSE]
    }
    mu_y <- mu_w + mu_b
    y1w_c <- t(t(y1w) - mu_y)

    sigma_w_inv <- try(solve.default(sigma_w), silent = TRUE)
    if (inherits(sigma_w_inv, "try-error")) {
      lav_msg_stop(gettext(
        "EM steps of the saturated (H1) model failed; some matrices may
         be singular; please check your data for (near-)perfect
         correlations."))
    }
    sigma_w_logdet <- 0
    if (logl) {
      sigma_w_logdet <- c(determinant.matrix(sigma_w,
                                             logarithm = TRUE)$modulus)
      # accumulators for the fused loglikelihood
      w_logdet <- zz_logdet <- ibza_logdet <- 0
      q_yy_a <- q_zz_a <- q_corr <- 0
    }

    # 1. per-pattern quantities + A.j and p.j per cluster
    winv_p <- breg_p <- ccond_p <- vector("list", mp$npatterns)
    alist_1 <- rep(list(matrix(0, nb, nb)), nclusters)
    pij <- matrix(0, nrow(y1w), ny)
    for (p in seq_len(mp$npatterns)) {
      o_idx <- which(mp$pat[p, ])
      na_idx <- which(!mp$pat[p, ])
      case_idx <- mp$case.idx[[p]]

      if (length(na_idx) > 0L) {
        wp_inv <- lav_mat_sym_inverse_update(
          s_inv = sigma_w_inv, rm_idx = na_idx,
          logdet = logl, s_logdet = sigma_w_logdet
        )
        if (logl) {
          w_logdet <- w_logdet + (attr(wp_inv, "logdet") * mp$freq[p])
        }
        breg_p[[p]] <- sigma_w[na_idx, o_idx, drop = FALSE] %*% wp_inv
        ccond_p[[p]] <- sigma_w[na_idx, na_idx, drop = FALSE] -
          breg_p[[p]] %*% sigma_w[o_idx, na_idx, drop = FALSE]
      } else {
        wp_inv <- sigma_w_inv
        if (logl) {
          w_logdet <- w_logdet + (sigma_w_logdet * mp$freq[p])
        }
      }
      winv_p[[p]] <- wp_inv

      pij[case_idx, o_idx] <-
        y1w_c[case_idx, o_idx, drop = FALSE] %*% wp_inv

      # contribution to A.j ('both' block only)
      a_full <- matrix(0, ny, ny)
      a_full[o_idx, o_idx] <- wp_inv
      a_bb <- a_full[both_idx, both_idx, drop = FALSE]
      j1_idx <- mp$j1.idx[[p]]
      j_freq <- mp$j.freq[[p]]
      for (k in seq_along(j1_idx)) {
        alist_1[[j1_idx[k]]] <- alist_1[[j1_idx[k]]] + (a_bb * j_freq[k])
      }
    } # patterns
    pj <- rowsum.default(pij[, both_idx, drop = FALSE], cluster_idx,
      reorder = FALSE, na.rm = TRUE
    )
    if (logl) {
      q_yy_a <- sum(pij * y1w_c, na.rm = TRUE)
    }

    # 2. prior of (beta, z.mis) given z.obs, per z-pattern
    if (nz > 0L) {
      zc <- t(t(z) - mu_z)
      nqz <- zp$npatterns + (length(zp$empty.idx) > 0L)
      c0_q <- m0coef_q <- zmis_q <- vector("list", nqz)
      for (q in seq_len(nqz)) {
        if (q > zp$npatterns) { # empty pattern: no z observed
          zo_idx <- integer(0L)
          zm_idx <- seq_len(nz)
        } else {
          zo_idx <- which(zp$pat[q, ])
          zm_idx <- which(!zp$pat[q, ])
        }
        zmis_q[[q]] <- zm_idx
        k_q <- rbind(
          cbind(sb, syz[, zm_idx, drop = FALSE]),
          cbind(
            t(syz[, zm_idx, drop = FALSE]),
            sigma_zz[zm_idx, zm_idx, drop = FALSE]
          )
        )
        if (length(zo_idx) > 0L) {
          l_q <- rbind(
            syz[, zo_idx, drop = FALSE],
            sigma_zz[zm_idx, zo_idx, drop = FALSE]
          )
          szz_o <- sigma_zz[zo_idx, zo_idx, drop = FALSE]
          soo_inv <- solve.default(szz_o)
          m0coef_q[[q]] <- l_q %*% soo_inv
          c0_q[[q]] <- k_q - m0coef_q[[q]] %*% t(l_q)

          if (logl) {
            # fused loglikelihood: marginal of the observed z-values
            zz_logdet <- zz_logdet +
              (c(determinant.matrix(szz_o, logarithm = TRUE)$modulus) *
               zp$freq[q])
            zc_o <- zc[zp$case.idx[[q]], zo_idx, drop = FALSE]
            q_zz_a <- q_zz_a + sum((zc_o %*% soo_inv) * zc_o)
          }
        } else {
          m0coef_q[[q]] <- matrix(0, nb + length(zm_idx), 0L)
          c0_q[[q]] <- k_q
        }
      }
    }

    # 3. per-cluster posterior of (beta, z.mis); accumulate the
    #    between-level (and z) expected sufficient statistics
    eb <- matrix(0, nclusters, nb) # E(beta | data)
    vb_list <- vector("list", nclusters) # Cov(beta | data)
    tb1 <- numeric(nb)
    tb2 <- matrix(0, nb, nb)
    if (nz > 0L) {
      tz1 <- numeric(nz)
      tz2 <- matrix(0, nz, nz)
      tzv <- matrix(0, nz, nb)
    }
    for (j in seq_len(nclusters)) {
      a_j <- alist_1[[j]]
      p_j <- pj[j, ]
      if (nz > 0L) {
        q <- zpat2j[j]
        c0 <- c0_q[[q]]
        zm_idx <- zmis_q[[q]]
        nzm <- length(zm_idx)
        if (q > zp$npatterns) {
          m0 <- numeric(nb + nzm)
        } else {
          zo_idx <- which(zp$pat[q, ])
          m0 <- drop(m0coef_q[[q]] %*% zc[j, zo_idx])
        }
      } else {
        c0 <- sb
        nzm <- 0L
        m0 <- numeric(nb)
      }

      # posterior of w = (beta, z.mis)
      nw <- nb + nzm
      m_j <- c0[, seq_len(nb), drop = FALSE] %*% a_j # C0 %*% Atilde
      m_j <- cbind(m_j, matrix(0, nw, nzm))
      m_j[lav_mat_diag_idx(nw)] <- m_j[lav_mat_diag_idx(nw)] + 1
      rhs <- cbind(m0 + drop(c0[, seq_len(nb), drop = FALSE] %*% p_j), c0)
      sol <- try(solve.default(m_j, rhs), silent = TRUE)
      if (inherits(sol, "try-error")) {
        lav_msg_stop(gettext(
          "EM steps of the saturated (H1) model failed; some matrices may
           be singular; please check your data for (near-)perfect
           correlations."))
      }
      m1 <- sol[, 1L]
      c1 <- sol[, -1L, drop = FALSE]
      c1 <- (c1 + t(c1)) / 2

      ebeta <- m1[seq_len(nb)]
      vb <- c1[seq_len(nb), seq_len(nb), drop = FALSE]
      eb[j, ] <- ebeta
      vb_list[[j]] <- vb

      if (logl) {
        # fused loglikelihood: m_j is block lower-triangular, so
        # det(m_j) = det(I + C0[beta, beta] %*% A.j); the second term is
        # the correction from integrating out w = (beta, z.mis)
        ldm <- determinant.matrix(m_j, logarithm = TRUE)
        ibza_logdet <- ibza_logdet + (c(ldm$modulus) * ldm$sign)
        m0_b <- m0[seq_len(nb)]
        q_corr <- q_corr + sum(p_j * (m0_b + ebeta)) -
          sum(m0_b * (a_j %*% ebeta))
      }

      # E(u_j | data), uncentered ('both' part)
      ev_b <- mu_b[both_idx] + ebeta
      tb1 <- tb1 + ev_b
      tb2 <- tb2 + vb + tcrossprod(ev_b)

      if (nz > 0L) {
        ez <- z[j, ]
        if (nzm > 0L) {
          ez[zm_idx] <- mu_z[zm_idx] + m1[nb + seq_len(nzm)]
        }
        tz1 <- tz1 + ez
        ezz <- tcrossprod(ez)
        ezv <- tcrossprod(ez, ev_b)
        if (nzm > 0L) {
          ezz[zm_idx, zm_idx] <- ezz[zm_idx, zm_idx, drop = FALSE] +
            c1[nb + seq_len(nzm), nb + seq_len(nzm), drop = FALSE]
          ezv[zm_idx, ] <- ezv[zm_idx, , drop = FALSE] +
            c1[nb + seq_len(nzm), seq_len(nb), drop = FALSE]
        }
        tz2 <- tz2 + ezz
        tzv <- tzv + ezv
      }
    } # clusters

    # 4. within-level expected sufficient statistics
    #    d_i = y_i - u_j; E(d_i) and E(d_i d_i') per pattern
    tw1 <- numeric(ny)
    tw2 <- matrix(0, ny, ny)
    eb_full <- matrix(0, nclusters, ny)
    eb_full[, both_idx] <- eb
    for (p in seq_len(mp$npatterns)) {
      o_idx <- which(mp$pat[p, ])
      na_idx <- which(!mp$pat[p, ])
      case_idx <- mp$case.idx[[p]]
      j_idx <- mp$j.idx[[p]]

      # E(d_i): observed part = y - mu.b - E(beta); missing part via B_p
      f_o <- y1w_c[case_idx, o_idx, drop = FALSE] -
        eb_full[j_idx, o_idx, drop = FALSE]
      e_full <- matrix(0, length(case_idx), ny)
      e_full[, o_idx] <- t(t(f_o) + mu_w[o_idx])
      if (length(na_idx) > 0L) {
        e_full[, na_idx] <- t(t(f_o %*% t(breg_p[[p]])) + mu_w[na_idx])
      }
      tw1 <- tw1 + colSums(e_full)
      tw2 <- tw2 + crossprod(e_full)

      # covariance contributions, per cluster in this pattern
      ob <- which(o_idx %in% both_idx) # both-var positions within o_idx
      b_pos <- match(o_idx[ob], both_idx)
      j1_idx <- mp$j1.idx[[p]]
      j_freq <- mp$j.freq[[p]]
      b_p <- breg_p[[p]]
      for (k in seq_along(j1_idx)) {
        v_oo <- matrix(0, length(o_idx), length(o_idx))
        if (length(ob) > 0L) {
          v_oo[ob, ob] <- vb_list[[j1_idx[k]]][b_pos, b_pos, drop = FALSE]
        }
        m_cov <- matrix(0, ny, ny)
        m_cov[o_idx, o_idx] <- v_oo
        if (length(na_idx) > 0L) {
          bv <- b_p %*% v_oo
          m_cov[na_idx, o_idx] <- bv
          m_cov[o_idx, na_idx] <- t(bv)
          m_cov[na_idx, na_idx] <- tcrossprod(bv, b_p) + ccond_p[[p]]
        }
        tw2 <- tw2 + (j_freq[k] * m_cov)
      }
    } # patterns

    # empty units (all level-1 variables missing): E(d) = mu.w,
    # Cov(d) = sigma.w -- no data contribution
    n_empty <- length(mp$empty.idx)
    if (n_empty > 0L) {
      tw1 <- tw1 + (n_empty * mu_w)
      tw2 <- tw2 + n_empty * (sigma_w + tcrossprod(mu_w))
    }

    # 5. M-step (saturated updates)
    mu_w <- tw1 / nobs
    sigma_w <- tw2 / nobs - tcrossprod(mu_w)
    sigma_w <- (sigma_w + t(sigma_w)) / 2
    m_b <- tb1 / nclusters
    mu_b <- numeric(ny)
    mu_b[both_idx] <- m_b
    sigma_b <- matrix(0, ny, ny)
    sigma_b[both_idx, both_idx] <- tb2 / nclusters - tcrossprod(m_b)
    if (nz > 0L) {
      mu_z <- tz1 / nclusters
      sigma_zz <- tz2 / nclusters - tcrossprod(mu_z)
      sigma_zz <- (sigma_zz + t(sigma_zz)) / 2
      sigma_yz <- matrix(0, ny, nz)
      sigma_yz[both_idx, ] <- t(tzv / nclusters - tcrossprod(mu_z, m_b))
    }

    # check for (near-zero) variances at the within level, and set
    # them to min.variance
    zero_var <- which(diag(sigma_w) < min_variance)
    if (length(zero_var) > 0L) {
      sigma_w[, zero_var] <- 0
      sigma_w[zero_var, ] <- 0
      diag(sigma_w)[zero_var] <- min_variance
    }

    theta_new <- em_pack(mu_w, mu_b, sigma_w, sigma_b, mu_z, sigma_zz,
                         sigma_yz)
    if (logl) {
      # assemble the loglikelihood at the input theta (note: all the
      # accumulators were computed *before* the M-step overwrote the
      # parameter matrices)
      fx_in <- -(p_1 * log_2pi + (w_logdet + ibza_logdet + zz_logdet) +
                 (q_yy_a + q_zz_a - q_corr)) / 2
      if (has_x) {
        fx_in <- fx_in - loglik_x
      }
      attr(theta_new, "logl") <- fx_in
    }
    theta_new
  } # em_step

  # loglikelihood at theta -- the state is already in 2l form, so we can
  # skip the 2l -> implied -> 2l round trip (mu.y = mu.w + mu.b, with
  # mu.b[within-only] structurally zero; see the invariants at the top of
  # lav_mvnorm_cluster.R)
  em_logl <- function(theta) {
    th <- em_unpack(theta)
    out <- list(
      sigma.w = th$sigma_w, sigma.b = th$sigma_b,
      sigma.zz = th$sigma_zz, sigma.yz = th$sigma_yz,
      mu.z = th$mu_z, mu.y = th$mu_w + th$mu_b,
      mu.w = th$mu_w, mu.b = th$mu_b
    )
    lav_mvn_cl_mi_loglik_samp_2l(
      y1 = y1, y2 = y2, lp = lp, mp = mp,
      loglik_x = loglik_x, log2pi = TRUE, minus_two = FALSE,
      out = out
    )
  }

  # implied-space matrices at theta
  em_implied <- function(theta) {
    th <- em_unpack(theta)
    lav_mvn_cl_2l2implied(
      lp = lp, sigma_w = th$sigma_w, sigma_b = th$sigma_b,
      sigma_zz = th$sigma_zz, sigma_yz = th$sigma_yz, mu_z = th$mu_z,
      mu_y = NULL, mu_w = th$mu_w, mu_b = th$mu_b
    )
  }

  list(
    theta0 = em_pack(mu_w, mu_b, sigma_w, sigma_b, mu_z, sigma_zz,
                     sigma_yz),
    pack = em_pack,
    step = em_step,
    step_logl = function(theta) em_step(theta, logl = TRUE),
    logl = em_logl,
    implied = em_implied
  )
}

lav_mvn_cl_mi_em_sat <- function(y1 = NULL,
                                 y2 = NULL,
                                 lp = NULL,
                                 mp = NULL,
                                 loglik_x = 0,
                                 tol = 1e-04, # = em.h1.args$tol
                                 max_iter = 5000L, # = em.h1.args$max_iter
                                 min_variance = 1e-05,
                                 acceleration = "none",
                                 fused = TRUE) { # = em.h1.args$fused
  if (is.null(fused)) {
    fused <- TRUE
  }
  engine <- lav_mvn_cl_mi_em_engine(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    loglik_x = loglik_x, min_variance = min_variance
  )

  theta <- engine$theta0

  # EM iterations
  if (acceleration %in% c("squarem", "qn")) {
    # report initial fx
    fx <- engine$logl(theta)
    if (lav_verbose()) {
      cat(
        "EM iter:", sprintf("%3d", 0),
        " fx =", sprintf("%17.10f", fx),
        "\n"
      )
    }
    accel_fn <- if (acceleration == "qn") lav_em_qn else lav_em_squarem
    out_em <- accel_fn(
      theta = theta, step_fn = engine$step, logl_fn = engine$logl,
      fx0 = fx, tol = tol, max_iter = max_iter
    )
    theta <- out_em$theta
    fx <- out_em$fx
    converged <- out_em$converged
  } else {
    # plain EM iterations; if fused = TRUE, the loglikelihood at theta is
    # obtained as a cheap byproduct of engine$step_logl(theta) (returned
    # as the "logl" attribute), so no separate loglikelihood evaluations
    # are needed; we merely look one EM step ahead, and at the end
    # (theta, fx) still form an exact pair; if fused = FALSE, the
    # stand-alone loglikelihood function is used instead (same iterates,
    # same stopping rule, same result)
    if (fused) {
      theta_next <- engine$step_logl(theta)
      fx <- attr(theta_next, "logl") # logl at theta0
    } else {
      fx <- engine$logl(theta)
    }
    if (lav_verbose()) {
      cat(
        "EM iter:", sprintf("%3d", 0),
        " fx =", sprintf("%17.10f", fx),
        "\n"
      )
    }
    fx_old <- fx
    converged <- FALSE
    for (i in seq_len(max_iter)) {
      if (fused) {
        theta <- theta_next
        theta_next <- engine$step_logl(theta)
        fx <- attr(theta_next, "logl") # logl at the current theta
      } else {
        theta <- engine$step(theta)
        fx <- engine$logl(theta)
      }

      # fx.delta
      fx_delta <- fx - fx_old

      # check if fx.delta is finite
      if (!is.finite(fx_delta)) {
        implied2 <- engine$implied(theta)
        cat("\n")
        cat("FATAL problem: dumping Sigma.W and Sigma.B matrices:\n\n")
        cat("Sigma.W:\n")
        print(implied2$Sigma.W)
        cat("\n")
        cat("Sigma.B:\n")
        print(implied2$Sigma.B)
        cat("\n")
        lav_msg_stop(gettext(
          "EM steps of the saturated (H1) model failed; some matrices may
           be singular; please check your data for (near-)perfect
           correlations."))
      }

      # what if fx.delta is negative?
      if (fx_delta < -1e-06) {
        lav_msg_warn(gettext(
          "logl decreased during EM steps of the saturated (H1) model"))
      }

      if (lav_verbose()) {
        cat(
          "EM iter:", sprintf("%3d", i),
          " fx =", sprintf("%17.10f", fx),
          " fx.delta =", sprintf("%9.8f", fx_delta),
          "\n"
        )
      }

      # convergence check
      if (abs(fx_delta) < tol) {
        converged <- TRUE
        break
      } else {
        fx_old <- fx
      }
    }
  } # EM iterations

  # warning?
  if (!converged) {
    lav_msg_warn(gettext(
      "Maximum number of iterations reached when computing the sample
       moments of the saturated (H1) model using EM; increase the max_iter
       element of the em.h1.args= argument to increase the number of
       iterations"))
  }

  # final implied matrices
  implied2 <- engine$implied(theta)

  list(
    Sigma.W = implied2$Sigma.W, Sigma.B = implied2$Sigma.B,
    Mu.W = implied2$Mu.W, Mu.B = implied2$Mu.B, logl = fx
  )
}

# the posterior distribution of the cluster-level unknowns, given the
# observed data and the model-implied statistics (Mu.W, Sigma.W, Mu.B,
# Sigma.B): per cluster j, w_j = (beta_j, ztilde_j[mis]) has posterior
# N(m1_j, C1_j) (see the E-step of lav_mvn_cl_mi_em_engine); also
# returns the per-pattern sweep regressions (breg, ccond) needed to
# impute the missing level-1 values, and various intermediate
# quantities (used by lav_mvn_cl_mi_scov_louis and
# lav_mvn_cl_mi_estep_ranef)
lav_mvn_cl_mi_posterior <- function(y1 = NULL,
                                    y2 = NULL,
                                    lp = NULL,
                                    mp = NULL,
                                    mu_w = NULL, # implied Mu.W
                                    sigma_w = NULL, # implied Sigma.W
                                    mu_b = NULL, # implied Mu.B
                                    sigma_b = NULL, # implied Sigma.B
                                    out = NULL) { # pre-converted 2l list
  # map implied to 2l matrices (unless the caller already did)
  if (is.null(out)) {
    out <- lav_mvn_cl_implied22l(
      lp = lp, mu_w = mu_w, mu_b = mu_b,
      sigma_w = sigma_w, sigma_b = sigma_b
    )
  }
  mu_y <- out$mu.y
  mu_z <- out$mu.z
  sw <- out$sigma.w
  sb <- out$sigma.b[lp$both.idx[[2]], lp$both.idx[[2]], drop = FALSE]
  sigma_zz <- out$sigma.zz
  syz <- out$sigma.yz[lp$both.idx[[2]], , drop = FALSE]

  # dimensions and indices
  nclusters <- lp$nclusters[[2]]
  cluster_idx <- lp$cluster.idx[[2]]
  between_idx <- lp$between.idx[[2]]
  both_idx <- lp$both.idx[[2]]
  nz <- length(between_idx)
  nb <- length(both_idx)
  zp <- mp$Zp
  if (nz > 0L) {
    y1w <- y1[, -between_idx, drop = FALSE]
    z <- y2[, between_idx, drop = FALSE]
    zc <- t(t(z) - mu_z)
  } else {
    y1w <- y1
    z <- NULL
    zc <- NULL
  }
  ny <- ncol(y1w)
  y1w_c <- t(t(y1w) - mu_y)

  # per-pattern quantities + A.j and p.j per cluster
  sigma_w_inv <- solve.default(sw)
  winv_p <- breg_p <- ccond_p <- vector("list", mp$npatterns)
  alist_1 <- rep(list(matrix(0, nb, nb)), nclusters)
  pij <- matrix(0, nrow(y1w), ny)
  for (p in seq_len(mp$npatterns)) {
    o_idx <- which(mp$pat[p, ])
    na_idx <- which(!mp$pat[p, ])
    case_idx <- mp$case.idx[[p]]
    if (length(na_idx) > 0L) {
      wp_inv <- lav_mat_sym_inverse_update(
        s_inv = sigma_w_inv, rm_idx = na_idx
      )
      breg_p[[p]] <- sw[na_idx, o_idx, drop = FALSE] %*% wp_inv
      ccond_p[[p]] <- sw[na_idx, na_idx, drop = FALSE] -
        breg_p[[p]] %*% sw[o_idx, na_idx, drop = FALSE]
    } else {
      wp_inv <- sigma_w_inv
    }
    winv_p[[p]] <- wp_inv
    pij[case_idx, o_idx] <- y1w_c[case_idx, o_idx, drop = FALSE] %*% wp_inv
    a_full <- matrix(0, ny, ny)
    a_full[o_idx, o_idx] <- wp_inv
    a_bb <- a_full[both_idx, both_idx, drop = FALSE]
    j1_idx <- mp$j1.idx[[p]]
    j_freq <- mp$j.freq[[p]]
    for (k in seq_along(j1_idx)) {
      alist_1[[j1_idx[k]]] <- alist_1[[j1_idx[k]]] + (a_bb * j_freq[k])
    }
  }
  pj <- rowsum.default(pij[, both_idx, drop = FALSE], cluster_idx,
    reorder = FALSE, na.rm = TRUE
  )

  # z-patterns: prior of (beta, z.mis) given z.obs
  zmis_q <- zpat2j <- NULL
  if (nz > 0L) {
    zpat2j <- integer(nclusters)
    for (q in seq_len(zp$npatterns)) {
      zpat2j[zp$case.idx[[q]]] <- q
    }
    if (length(zp$empty.idx) > 0L) {
      zpat2j[zp$empty.idx] <- zp$npatterns + 1L
    }
    nqz <- zp$npatterns + (length(zp$empty.idx) > 0L)
    c0_q <- m0coef_q <- zmis_q <- vector("list", nqz)
    for (q in seq_len(nqz)) {
      if (q > zp$npatterns) {
        zo_idx <- integer(0L)
        zm_idx <- seq_len(nz)
      } else {
        zo_idx <- which(zp$pat[q, ])
        zm_idx <- which(!zp$pat[q, ])
      }
      zmis_q[[q]] <- zm_idx
      k_q <- rbind(
        cbind(sb, syz[, zm_idx, drop = FALSE]),
        cbind(
          t(syz[, zm_idx, drop = FALSE]),
          sigma_zz[zm_idx, zm_idx, drop = FALSE]
        )
      )
      if (length(zo_idx) > 0L) {
        l_q <- rbind(
          syz[, zo_idx, drop = FALSE],
          sigma_zz[zm_idx, zo_idx, drop = FALSE]
        )
        soo_inv <- solve.default(sigma_zz[zo_idx, zo_idx, drop = FALSE])
        m0coef_q[[q]] <- l_q %*% soo_inv
        c0_q[[q]] <- k_q - m0coef_q[[q]] %*% t(l_q)
      } else {
        m0coef_q[[q]] <- matrix(0, nb + length(zm_idx), 0L)
        c0_q[[q]] <- k_q
      }
    }
  }

  # per-cluster posterior of w = (beta, z.mis): N(m1, C1)
  m1_list <- c1_list <- vector("list", nclusters)
  for (j in seq_len(nclusters)) {
    a_j <- alist_1[[j]]
    p_j <- pj[j, ]
    if (nz > 0L) {
      q <- zpat2j[j]
      c0 <- c0_q[[q]]
      zm_idx <- zmis_q[[q]]
      nzm <- length(zm_idx)
      if (q > zp$npatterns) {
        m0 <- numeric(nb + nzm)
      } else {
        zo_idx <- which(zp$pat[q, ])
        m0 <- drop(m0coef_q[[q]] %*% zc[j, zo_idx])
      }
    } else {
      c0 <- sb
      nzm <- 0L
      m0 <- numeric(nb)
    }
    nw <- nb + nzm
    m_j <- c0[, seq_len(nb), drop = FALSE] %*% a_j
    m_j <- cbind(m_j, matrix(0, nw, nzm))
    m_j[lav_mat_diag_idx(nw)] <- m_j[lav_mat_diag_idx(nw)] + 1
    rhs <- cbind(m0 + drop(c0[, seq_len(nb), drop = FALSE] %*% p_j), c0)
    sol <- solve.default(m_j, rhs)
    m1_list[[j]] <- sol[, 1L]
    c1 <- sol[, -1L, drop = FALSE]
    c1_list[[j]] <- (c1 + t(c1)) / 2
  }

  list(
    m1 = m1_list, c1 = c1_list,
    breg = breg_p, ccond = ccond_p, winv = winv_p,
    alist = alist_1, pj = pj,
    y1w = y1w, y1w_c = y1w_c, z = z, zc = zc,
    mu_y = mu_y, mu_z = mu_z,
    sb = sb, syz = syz, sigma_zz = sigma_zz, sigma_w = sw,
    mu_b_2l = out$mu.b, mu_w_2l = out$mu.w,
    both_idx = both_idx, between_idx = between_idx,
    zmis_q = zmis_q, zpat2j = zpat2j
  )
}

# the E-step 'random effects': the posterior means of the cluster-level
# components nu_j = (mu_b + beta_j), given ALL the observed data; this
# is the missing-data counterpart of lav_mvn_cl_em_estep_ranef()
#
# returns a nclusters x ny matrix (ny = number of level-1 variables) in
# the same convention as the complete-data version: for both-level
# variables, the (uncentered, total-mean scale) posterior mean of nu_j;
# zero for within-only variables
#
# if se = TRUE, the posterior standard deviations are returned in the
# "se" attribute
#
# if impute = TRUE, two extra attributes are returned:
# - "y1w.imputed": the level-1 data (level-1 variables only), with the
#   missing values replaced by their posterior means
#   E[y.mis | y.obs, E(beta)] (exact, as the conditional mean is linear
#   in beta)
# - "z.imputed": the cluster-level data (between-only variables, one
#   row per cluster), with the missing values replaced by their
#   posterior means
lav_mvn_cl_mi_estep_ranef <- function(y1 = NULL,
                                      y2 = NULL,
                                      lp = NULL,
                                      mp = NULL,
                                      mu_w = NULL, # implied Mu.W
                                      sigma_w = NULL, # implied Sigma.W
                                      mu_b = NULL, # implied Mu.B
                                      sigma_b = NULL, # implied Sigma.B
                                      se = FALSE,
                                      impute = FALSE) {
  post <- lav_mvn_cl_mi_posterior(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    mu_w = mu_w, sigma_w = sigma_w, mu_b = mu_b, sigma_b = sigma_b
  )
  nclusters <- lp$nclusters[[2]]
  both_idx <- post$both_idx
  nb <- length(both_idx)
  nz <- length(post$between_idx)
  ny <- ncol(post$y1w)

  # E(nu_j | all observed data), in the y1w space (cf. the complete-data
  # version: mu_b_2l[both] is the total mean of the both-level variables)
  eb <- matrix(0, nclusters, nb)
  for (j in seq_len(nclusters)) {
    eb[j, ] <- post$m1[[j]][seq_len(nb)]
  }
  mb_j <- matrix(0, nrow = nclusters, ncol = ny)
  mb_j[, both_idx] <- rep(post$mu_b_2l[both_idx], each = nclusters) + eb

  # posterior standard deviations
  if (se) {
    se_j <- matrix(0, nrow = nclusters, ncol = ny)
    for (j in seq_len(nclusters)) {
      c1_diag <- diag(post$c1[[j]])[seq_len(nb)]
      c1_diag[c1_diag < 0] <- 0
      se_j[j, both_idx] <- sqrt(c1_diag)
    }
    attr(mb_j, "se") <- se_j
  }

  # impute the missing values by their posterior means
  if (impute) {
    # level-1 variables: E[y.mis | y.obs, beta] is linear in beta, so
    # plugging in E[beta] gives the exact posterior mean
    y1w_i <- post$y1w
    eb_full <- matrix(0, nclusters, ny)
    eb_full[, both_idx] <- eb
    for (p in seq_len(mp$npatterns)) {
      na_idx <- which(!mp$pat[p, ])
      if (length(na_idx) == 0L) {
        next
      }
      o_idx <- which(mp$pat[p, ])
      case_idx <- mp$case.idx[[p]]
      j_idx <- mp$j.idx[[p]]
      f_o <- post$y1w_c[case_idx, o_idx, drop = FALSE] -
        eb_full[j_idx, o_idx, drop = FALSE]
      y1w_i[case_idx, na_idx] <-
        t(t(f_o %*% t(post$breg[[p]])) + post$mu_y[na_idx]) +
        eb_full[j_idx, na_idx, drop = FALSE]
    }
    # empty units (all level-1 variables missing)
    if (length(mp$empty.idx) > 0L) {
      cl_empty <- lp$cluster.idx[[2]][mp$empty.idx]
      y1w_i[mp$empty.idx, ] <-
        t(t(eb_full[cl_empty, , drop = FALSE]) + post$mu_y)
    }
    attr(mb_j, "y1w.imputed") <- y1w_i

    # between-only variables (one row per cluster)
    if (nz > 0L) {
      z_i <- post$z
      for (j in seq_len(nclusters)) {
        zm_idx <- post$zmis_q[[post$zpat2j[j]]]
        if (length(zm_idx) > 0L) {
          z_i[j, zm_idx] <- post$mu_z[zm_idx] +
            post$m1[[j]][nb + seq_along(zm_idx)]
        }
      }
      attr(mb_j, "z.imputed") <- z_i
    }
  }

  mb_j
}

# Mu.W, Mu.B, Sigma.W, Sigma.B are the model-implied statistics
lav_mvn_cl_mi_dlogl_2l_samp <- function(
                                        y1 = NULL,
                                        y2 = NULL,
                                        lp = NULL,
                                        mp = NULL,
                                        mu_w = NULL,
                                        sigma_w = NULL,
                                        mu_b = NULL,
                                        sigma_b = NULL,
                                        sinv_method = "eigen",
                                        return_list = FALSE) {
  lav_mvn_cl_mi_grad_engine(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    mu_w = mu_w, sigma_w = sigma_w,
    mu_b = mu_b, sigma_b = sigma_b,
    score_mode = FALSE, return_list = return_list,
    sinv_method = sinv_method
  )
}

# cluster-wise scores -2*logl wrt Mu.W, Mu.B, Sigma.W, Sigma.B
lav_mvn_cl_mi_sc_2l <- function(
                                y1 = NULL,
                                y2 = NULL,
                                lp = NULL,
                                mp = NULL,
                                mu_w = NULL,
                                sigma_w = NULL,
                                mu_b = NULL,
                                sigma_b = NULL,
                                sinv_method = "eigen") {
  lav_mvn_cl_mi_grad_engine(
    y1 = y1, y2 = y2, lp = lp, mp = mp,
    mu_w = mu_w, sigma_w = sigma_w,
    mu_b = mu_b, sigma_b = sigma_b,
    score_mode = TRUE,
    sinv_method = sinv_method
  )
}

# first-order information: outer crossprod of scores per cluster
lav_mvn_cl_mi_info_firstorder <- function(
                                          y1 = NULL,
                                          y2 = NULL,
                                          lp = NULL,
                                          mp = NULL,
                                          mu_w = NULL,
                                          sigma_w = NULL,
                                          mu_b = NULL,
                                          sigma_b = NULL,
                                          x_idx = NULL,
                                          divide_by_two = FALSE,
                                          sinv_method = "eigen") {

  scores <- lav_mvn_cl_mi_sc_2l(
    y1 = y1,
    y2 = y2,
    lp = lp,
    mp = mp,
    mu_w = mu_w,
    sigma_w = sigma_w,
    mu_b = mu_b,
    sigma_b = sigma_b,
    sinv_method = sinv_method
  )

  # divide by 2 (if we want scores wrt objective function)
  if (divide_by_two) {
    scores <- scores / 2
  }

  # unit information
  information <- crossprod(scores) / lp$nclusters[[2]]

  # if x.idx, set rows/cols to zero
  information <- lav_mvn_cl_zero_x_idx(information,
    lp = lp, mu_w = mu_w, mu_b = mu_b, x_idx = x_idx
  )

  information
}

# observed information
# order: mu.w within, vech(sigma.w) within, mu.b between, vech(sigma.b) between
# mu.w rows/cols that are splitted within/between are forced to zero
#
# numerical approximation (for now)
lav_mvn_cl_mi_info_observed <- function(
                                        y1 = NULL,
                                        y2 = NULL,
                                        lp = NULL,
                                        mp = NULL,
                                        ylp = NULL,
                                        mu_w = NULL,
                                        sigma_w = NULL,
                                        mu_b = NULL,
                                        sigma_b = NULL,
                                        x_idx = integer(0L),
                                        sinv_method = "eigen") {
  lav_mvn_cl_info_obs_engine(
    lp = lp,
    mu_w = mu_w, sigma_w = sigma_w,
    mu_b = mu_b, sigma_b = sigma_b,
    x_idx = x_idx, sinv_method = sinv_method,
    dlogl_fn = lav_mvn_cl_mi_dlogl_2l_samp,
    dlogl_args = list(y1 = y1, y2 = y2, mp = mp)
  )
}

# lavaanList: fit the *same* model, on different datasets
# YR - 29 Jun 2016
# YR - 27 Jan 2017: change lavoptions; add dotdotdot to each call
# TDJ - 23 Aug 2018: change wrappers to preserve arguments from match.call()
# YR - 15 Oct 2024: add iseed (as in lavBootstrap)

lavaanList <- function(model = NULL, # model                    # nolint
                       data_list = NULL, # list of datasets
                       data_function = NULL, # generating function
                       data_function_args = list(), # optional arguments
                       ndat = length(data_list), # how many datasets?
                       cmd = "lavaan",
                       ...,
                       store_slots = c("partable"), # default is partable
                       fun = NULL, # arbitrary fun
                       show_progress = FALSE,
                       store_failed = FALSE,
                       parallel = c("no", "multicore", "snow"),
                       ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
                       cl = NULL,
                       iseed = NULL) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, TRUE)

  # store_slots call
  mc <- match.call()

  # store current random seed (if any) and method
  rngkind_old <- RNGkind()
  #cat("RNGkind = ", paste(RNGkind_old, collapse = " "), "\n")
  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    init_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  } else {
    init_seed <- NULL
  }

  # initial seed (for the precomputations)
  if (!is.null(iseed)) {
    set.seed(min(1, iseed - 1L))
  }

  # check store_slots
  store_slots <- tolower(store_slots)
  if (length(store_slots) == 1L && store_slots == "all") {
    store_slots <- c(
      "timing", "partable", "data", "samplestats",
      "cache", "loglik", "h1", "baseline", "external",
      "vcov", "test", "optim", "implied"
    )
  }

  # data_list or function?
  if (is.function(data_function)) {
    if (ndat == 0L) {
      lav_msg_stop(gettext(
        "please specify number of requested datasets (ndat)"))
    }
    # here we already use the random generator
    first_data <- do.call(data_function, args = data_function_args)
    # data_list  <- vector("list", length = ndat)
  } else {
    first_data <- data_list[[1]]
  }

  # check data
  if (is.matrix(first_data)) {
    # check if we have column names?
    names_1 <- colnames(first_data)
    if (is.null(names_1)) {
      lav_msg_stop(gettext("data is a matrix without column names"))
    }
  } else if (inherits(first_data, "data.frame")) {
    # check?
  } else {
    lav_msg_stop(gettext("(generated) data is not a data.frame (or a matrix)"))
  }

  # if 'model' is a lavaan object (perhaps from lavSimulate), no need to
  # call `cmd'
  if (inherits(model, "lavaan")) {
    fit <- model
  } else {
    # adapt for FIT
    # dotdotdotFIT <- dotdotdot
    # dotdotdotFIT$do.fit  <- TRUE    # to get starting values
    # dotdotdotFIT$se      <- "none"
    # dotdotdotFIT$test    <- "none"

    # initial model fit, using first dataset
    fit <- do.call(cmd,
      args = c(list(
        model = model,
        data = first_data
      ), dotdotdot)
    )
  }

  lavoptions <- fit@Options
  lavoptions$fit.by.level <- FALSE # no partially saturated models
  lavmodel <- fit@Model
  lavpartable <- fit@ParTable
  lavpta <- fit@pta

  # remove any options in lavoptions from dotdotdot (sem/lavaan only)
  # because we can use the lavoptions slot
  if (length(dotdotdot) > 0L && cmd %in% c("lavaan", "sem", "cfa", "growth")) {
    rm_idx <- which(names(dotdotdot) %in% names(lavoptions))
    if (length(rm_idx) > 0L) {
      dotdotdot <- dotdotdot[-rm_idx]
    }
  }

  # remove start/est/se columns from lavpartable
  lavpartable$start <- lavpartable$est <- lavpartable$se <- NULL

  # empty slots
  timing_list <- par_table_list <- local_data_list <- sample_stats_list <-
    cache_list <- vcov_list <- test_list <- optim_list <-
    h1list <- loglik_list <- baseline_list <-
    implied_list <- fun_list <- list()

  # prepare store_slots slots
  if ("timing" %in% store_slots) {
    timing_list <- vector("list", length = ndat)
  }
  if ("partable" %in% store_slots) {
    par_table_list <- vector("list", length = ndat)
  }
  if ("data" %in% store_slots) {
    local_data_list <- vector("list", length = ndat)
  }
  if ("samplestats" %in% store_slots) {
    sample_stats_list <- vector("list", length = ndat)
  }
  if ("cache" %in% store_slots) {
    cache_list <- vector("list", length = ndat)
  }
  if ("vcov" %in% store_slots) {
    vcov_list <- vector("list", length = ndat)
  }
  if ("test" %in% store_slots) {
    test_list <- vector("list", length = ndat)
  }
  if ("optim" %in% store_slots) {
    optim_list <- vector("list", length = ndat)
  }
  if ("implied" %in% store_slots) {
    implied_list <- vector("list", length = ndat)
  }
  if ("loglik" %in% store_slots) {
    loglik_list <- vector("list", length = ndat)
  }
  if ("h1" %in% store_slots) {
    h1list <- vector("list", length = ndat)
  }
  if ("baseline" %in% store_slots) {
    baseline_list <- vector("list", length = ndat)
  }

  if (!is.null(fun)) {
    fun_list <- vector("list", length = ndat)
  }

  # single run
  fn <- function(i) {
    if (show_progress) {
      cat("   ... data set number:", sprintf("%4d", i))
    }

    # get new dataset
    if (i == 1L) {
      data_1 <- first_data
    } else if (is.function(data_function)) {
      data_1 <- do.call(data_function, args = data_function_args)
    } else if (is.list(data_list)) {
      data_1 <- data_list[[i]]
    }

    # if categorical, check if we have enough response categories
    # for each ordered variables in DATA
    data_ok_flag <- TRUE
    if (fit@Model@categorical) {
      # expected nlev
      ord_idx <- unique(unlist(fit@pta$vidx$ov.ord))
      nlev_exp <- fit@Data@ov$nlev[ord_idx]
      # observed nlev
      nlev_obs <- sapply(
        data_1[, unique(unlist(fit@pta$vnames$ov.ord)),
          drop = FALSE
        ],
        function(x) length(unique(na.omit(x)))
      )
      wrong_idx <- which(nlev_exp - nlev_obs != 0)
      if (length(wrong_idx) > 0L) {
        data_ok_flag <- FALSE
      }
    }

    # adapt lavmodel for this new dataset
    # - starting values will be different
    # - ov.x variances/covariances
    # FIXME: can we not make the changes internally?
    # if(lavmodel@fixed.x &&
    #     length(lav_pt_vnames(lavpartable, "ov.x")) > 0L) {
    # for(g in 1:FIT@Data@ngroups) {
    #
    # }
    lavmodel <- NULL
    # }

    # fit model with this (new) dataset
    if (data_ok_flag) {
      if (cmd %in% c("lavaan", "sem", "cfa", "growth")) {
        # lavoptions$start <- FIT # FIXME: needed?
        lavobject <- try(
          do.call("lavaan",
            args = c(
              list(
                slot_options = lavoptions,
                slot_par_table = lavpartable,
                slot_model = lavmodel,
                # start        = FIT,
                data = data_1
              ),
              dotdotdot
            )
          ),
          silent = TRUE
        )
      } else if (cmd == "fsr") {
        # the old (never exported) fsr() function has been removed;
        # sam() is its successor
        lav_msg_stop(gettext(
          "cmd = \"fsr\" is no longer supported; use cmd = \"sam\" instead."))
      } else if (cmd == "sam") {
        lavobject <- try(
          do.call("sam",
            args = c(
              list(
                model = model,
                data = data_1
              ),
              dotdotdot
            )
          ),
          silent = TRUE
        )
      } else {
        lavobject <- try(do.call(cmd,
          args = c(
              list(
                model = model,
                data = data_1
              ),
              dotdotdot
            )
          ),
          silent = TRUE
        )
        #lav_msg_stop(gettext("unknown cmd:"), cmd)
      }
    } # data.ok.flag

    res <- list(
      ok = FALSE, timing = NULL, ParTable = NULL,
      Data = NULL, SampleStats = NULL, vcov = NULL,
      test = NULL, optim = NULL, implied = NULL,
      baseline = NULL, baseline.ok = FALSE, fun = NULL
    )

    if (data_ok_flag && inherits(lavobject, "lavaan") &&
      lavInspect(lavobject, "converged")) {
      res$ok <- TRUE

      if (show_progress) {
        cat(
          "   OK -- niter = ",
          sprintf("%3d", lavInspect(lavobject, "iterations")),
          "\n"
        )
      }

      # extract slots from fit
      if ("timing" %in% store_slots) {
        res$timing <- lavobject@timing
      }
      if ("partable" %in% store_slots) {
        res$ParTable <- lavobject@ParTable
        # if the (possibly asymmetric) confidence intervals are based on
        # the bootstrap or the Monte Carlo method, the internal ParTable
        # does not contain them; compute and store ci.lower/ci.upper here,
        # so summary() (eg of a lavSimulate() object) can evaluate the
        # coverage of these intervals. The intervals are computed at the
        # 95% level (the lavaan default).
        if (identical(lavobject@Options$se, "bootstrap") ||
            identical(lavobject@Options$se.def, "monte.carlo")) {
          pe_ci <- try(lavParameterEstimates(lavobject,
            se = TRUE, zstat = FALSE, pvalue = FALSE, ci = TRUE,
            level = 0.95, standardized = FALSE,
            remove_system_eq = FALSE, remove_eq = FALSE,
            remove_ineq = FALSE, remove_def = FALSE,
            remove_nonfree = FALSE, remove_step1 = FALSE,
            remove_unused = FALSE, output = "data.frame"),
            silent = TRUE)
          if (!inherits(pe_ci, "try-error") && !is.null(pe_ci$ci.lower) &&
              length(pe_ci$ci.lower) == length(res$ParTable$lhs)) {
            res$ParTable$ci.lower <- pe_ci$ci.lower
            res$ParTable$ci.upper <- pe_ci$ci.upper
          } else {
            # res$ParTable is a list (not a data.frame), so we must use a
            # full-length vector here, not a scalar
            res$ParTable$ci.lower <- rep(as.numeric(NA),
                                         length(res$ParTable$lhs))
            res$ParTable$ci.upper <- rep(as.numeric(NA),
                                         length(res$ParTable$lhs))
          }
        }
      }
      if ("data" %in% store_slots) {
        res$Data <- lavobject@Data
      }
      if ("samplestats" %in% store_slots) {
        res$SampleStats <- lavobject@SampleStats
      }
      if ("cache" %in% store_slots) {
        res$Cache <- lavobject@Cache
      }
      if ("vcov" %in% store_slots) {
        res$vcov <- lavobject@vcov
      }
      if ("test" %in% store_slots) {
        res$test <- lavobject@test
      }
      if ("optim" %in% store_slots) {
        res$optim <- lavobject@optim
      }
      if ("implied" %in% store_slots) {
        res$implied <- lavobject@implied
      }
      if ("loglik" %in% store_slots) {
        res$loglik <- lavobject@loglik
      }
      if ("h1" %in% store_slots) {
        res$h1 <- lavobject@h1
      }
      if ("baseline" %in% store_slots) {
        res$baseline <- lavobject@baseline
        if (length(lavobject@baseline) > 0L) {
          res$baseline.ok <- TRUE
        }
      }

      # custom fun
      if (!is.null(fun)) {
        res$fun <- fun(lavobject)
      }
    } else { # failed!
      if (show_progress) {
        if (data_ok_flag) {
          if (inherits(lavobject, "lavaan")) {
            cat("   FAILED: no convergence\n")
          } else {
            cat("   FAILED: could not construct lavobject\n")
            print(lavobject)
          }
        } else {
          cat("   FAILED: nlev too low for some vars\n")
        }
      }
      if ("partable" %in% store_slots) {
        res$ParTable <- lavpartable
        res$ParTable$est <- res$ParTable$start
        res$ParTable$est[res$ParTable$free > 0] <- as.numeric(NA)
        res$ParTable$se <- numeric(length(lavpartable$lhs))
        res$ParTable$se[res$ParTable$free > 0] <- as.numeric(NA)
        # keep the ci.lower/ci.upper columns in sync with the converged
        # runs (see above), so all stored ParTables have the same columns.
        # res$ParTable is a list, so use full-length vectors, not scalars.
        if (identical(lavoptions$se, "bootstrap") ||
            identical(lavoptions$se.def, "monte.carlo")) {
          res$ParTable$ci.lower <- rep(as.numeric(NA),
                                       length(lavpartable$lhs))
          res$ParTable$ci.upper <- rep(as.numeric(NA),
                                       length(lavpartable$lhs))
        }
      }
      if (store_failed) {
        tmpfile <- tempfile(pattern = "lavaanListData")
        datfile <- paste0(tmpfile, ".csv")
        write.csv(data_1, file = datfile, row.names = FALSE)
        if (data_ok_flag) {
          # or only if lavobject is of class lavaan?
          objfile <- paste0(tmpfile, ".RData")
          save(lavobject, file = objfile)
        }
      }
    }

    res
  }


  # the next 8 lines are borrowed from the boot package
  have_mc <- have_snow <- FALSE
  if (missing(parallel)) {
    parallel <- "no"
  }
  parallel <- match.arg(parallel)
  if (parallel != "no" && ncpus > 1L) {
    if (parallel == "multicore") {
      have_mc <- .Platform$OS.type != "windows"
    } else if (parallel == "snow") {
      have_snow <- TRUE
    }
    if (!have_mc && !have_snow) ncpus <- 1L
    loadNamespace("parallel") # before recording seed!
  }

  # iseed:
  # this follows a proposal of Shu Fai Cheung (see github issue #240)
  # - iseed is used for both serial and parallel
  # - if iseed is not set, iseed is generated + .Random.seed created/updated
  #     -> tmp.seed <- NA
  # - if iseed is set: don't touch .Random.seed (if it exists)
  #     -> tmp.seed <- .Random.seed (if it exists)
  #     -> tmp.seed <- NULL (if it does not exist)
  if (is.null(iseed)) {
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      runif(1)
    }
    # identical(temp.seed, NA): Will not change .Random.seed in GlobalEnv
    temp_seed <- NA
    iseed <- runif(1, 0, 999999999)
  } else {
    if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
      temp_seed <-
        get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
    } else {
      # is.null(temp.seed): Will remove .Random.seed in GlobalEnv
      #                     if serial.
      #                     If parallel, .Random.seed will not be touched.
      temp_seed <- NULL
    }
  }
  if (!(ncpus > 1L && (have_mc || have_snow))) { # Only for serial
    set.seed(iseed)
  }

  # this is adapted from the boot function in package boot
  res <- if (ncpus > 1L && (have_mc || have_snow)) {
    if (have_mc) {
      #RNGkind_old <- RNGkind() # store current kind
      RNGkind("L'Ecuyer-CMRG") # to allow for reproducible results
      set.seed(iseed)
      parallel::mclapply(seq_len(ndat), fn, mc.cores = ncpus)
    } else if (have_snow) {
      list(...) # evaluate any promises
      if (is.null(cl)) {
        cl <- parallel::makePSOCKcluster(rep("localhost", ncpus))
        # # No need for
        # if(RNGkind()[1L] == "L'Ecuyer-CMRG")
        # clusterSetRNGStream() always calls `RNGkind("L'Ecuyer-CMRG")`
        parallel::clusterSetRNGStream(cl, iseed = iseed)
        res <- parallel::parLapply(cl, seq_len(ndat), fn)
        parallel::stopCluster(cl)
        res
      } else {
        parallel::parLapply(cl, seq_len(ndat), fn)
      }
    }
  } else {
    lapply(seq_len(ndat), fn)
  }


  # restructure
  if ("baseline" %in% store_slots) {
    meta <- list(
      ndat = ndat, ok = sapply(res, "[[", "ok"),
      baseline.ok = sapply(res, "[[", "baseline.ok"),
      store.slots = store_slots
    )
  } else {
    meta <- list(
      ndat = ndat, ok = sapply(res, "[[", "ok"),
      store.slots = store_slots
    )
  }

  # extract store_slots slots
  if ("timing" %in% store_slots) {
    timing_list <- lapply(res, "[[", "timing")
  }
  if ("partable" %in% store_slots) {
    par_table_list <- lapply(res, "[[", "ParTable")
  }
  if ("data" %in% store_slots) {
    local_data_list <- lapply(res, "[[", "Data")
  }
  if ("samplestats" %in% store_slots) {
    sample_stats_list <- lapply(res, "[[", "SampleStats")
  }
  if ("cache" %in% store_slots) {
    cache_list <- lapply(res, "[[", "Cache")
  }
  if ("vcov" %in% store_slots) {
    vcov_list <- lapply(res, "[[", "vcov")
  }
  if ("test" %in% store_slots) {
    test_list <- lapply(res, "[[", "test")
  }
  if ("optim" %in% store_slots) {
    optim_list <- lapply(res, "[[", "optim")
  }
  if ("implied" %in% store_slots) {
    implied_list <- lapply(res, "[[", "implied")
  }
  if ("h1" %in% store_slots) {
    h1list <- lapply(res, "[[", "h1")
  }
  if ("loglik" %in% store_slots) {
    loglik_list <- lapply(res, "[[", "loglik")
  }
  if ("baseline" %in% store_slots) {
    baseline_list <- lapply(res, "[[", "baseline")
  }
  if (!is.null(fun)) {
    fun_list <- lapply(res, "[[", "fun")
  }

  # create lavaanList object
  lavaan_list <- new("lavaanList",
    version = packageDescription("lavaan", fields = "Version"),
    call = mc,
    Options = lavoptions,
    ParTable = lavpartable,
    pta = lavpta,
    Model = lavmodel,
    Data = fit@Data,

    # meta
    meta = meta,

    # per dataset
    timingList = timing_list,
    ParTableList = par_table_list,
    DataList = local_data_list,
    SampleStatsList = sample_stats_list,
    CacheList = cache_list,
    vcovList = vcov_list,
    testList = test_list,
    optimList = optim_list,
    impliedList = implied_list,
    h1List = h1list,
    loglikList = loglik_list,
    baselineList = baseline_list,
    funList = fun_list,
    external = list()
  )

  # restore random seed
  if (!is.null(init_seed)) {
    assign(".Random.seed", init_seed, envir = .GlobalEnv)    # nolint
  } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    # initially there was no .Random.seed, but we created one along the way
    # clean up
    remove(".Random.seed", envir = .GlobalEnv)
  }
  # restore random generator method (if changed)
  RNGkind(rngkind_old[1], rngkind_old[2], rngkind_old[3])

  lavaan_list
}

semList <- function(model = NULL,                            # nolint
                    data_list = NULL,
                    data_function = NULL,
                    data_function_args = list(),
                    ndat = length(data_list),
                    ...,
                    store_slots = c("partable"),
                    fun = NULL,
                    show_progress = FALSE,
                    store_failed = FALSE,
                    parallel = c("no", "multicore", "snow"),
                    ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
                    cl = NULL,
                    iseed = NULL) {
  mc <- match.call(expand.dots = TRUE)
  mc$cmd <- "sem"
  mc[[1L]] <- quote(lavaan::lavaanList)
  eval(mc, parent.frame())
}

cfaList <- function(model = NULL,                         # nolint
                    data_list = NULL,
                    data_function = NULL,
                    data_function_args = list(),
                    ndat = length(data_list),
                    ...,
                    store_slots = c("partable"),
                    fun = NULL,
                    show_progress = FALSE,
                    store_failed = FALSE,
                    parallel = c("no", "multicore", "snow"),
                    ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
                    cl = NULL,
                    iseed = NULL) {
  mc <- match.call(expand.dots = TRUE)
  mc$cmd <- "cfa"
  mc[[1L]] <- quote(lavaan::lavaanList)
  eval(mc, parent.frame())
}

# construct 1D, 2D or pattern-based frequency tables
# YR. 10 April 2013
# Notes:
# - we do NOT make a distinction here between unordered and ordered categorical
#   variables
# - object can be a matrix (most likely with integers), a full data frame,
#   a fitted lavaan object, or a lavData object
# - 11 May 2013: added collapse=TRUE, min.std.resid options (suggested
#   by Myrsini Katsikatsou
# - 11 June 2013: added dimension, to get one-way and two-way (three-way?)
#   tables
# - 20 Sept 2013: - allow for sample-based or model-based cell probabilities
#                   re-organize/re-name to provide a more consistent interface
#                   rows in the output can be either: cells, tables or patterns
#                 - dimension=0 equals type="pattern
#                 - collapse=TRUE is replaced by type="table"
#                 - changed names of statistics: std.resid is now GR.average
#                 - added many more statistics; some based on the model, some
#                   on the unrestricted model
# - 8 Nov 2013:   - skip empty cells for G2, instead of adding 0.5 to obs
# - 7 Feb 2016:   - take care of conditional.x = TRUE

lavTables <- function(object,                         # nolint
                      # what type of table?
                      dimension = 2L,
                      type = "cells",
                      # if raw data, additional attributes
                      categorical = NULL,
                      group = NULL,
                      # which statistics / fit indices?
                      statistic = "default",
                      g2_min = 3.0, # needed for G2.{p/n}large
                      x2_min = 3.0, # needed for X2.{p/n}large
                      # pvalues for statistics?
                      p_value = FALSE,
                      # Bonferroni
                      # alpha.adj    = FALSE,
                      # output format
                      output = "data.frame",
                      pattern_as_string = TRUE,
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # check object
  object <- lav_object_check_version(object)

  # check input
  if (!(dimension == 0L || dimension == 1L || dimension == 2L)) {
    lav_msg_stop(gettext(
      "dimension must be 0, 1 or 2 for pattern, one-way or two-way tables"))
  }
  stopifnot(type %in% c("cells", "table", "pattern"))
  if (type == "pattern") {
    dimension <- 0L
  }

  # extract or create lavdata
  lavdata <- lav_lavdata(object, ordered = categorical, group = group)

  # is 'object' a lavaan object?
  lavobject <- NULL
  if (inherits(object, "lavaan")) {
    lavobject <- object
  }

  # case 1: response patterns
  if (dimension == 0L) {
    out <- lav_tables_pattern(
      lavobject = lavobject, lavdata = lavdata,
      statistic = statistic,
      pattern_as_string = pattern_as_string
    )
    # output format
    if (output == "data.frame") {
      class(out) <- c("lavaan.data.frame", "data.frame")
    } else {
      lav_msg_warn(gettextf("output option `%s' is not available; ignored.",
                            output))
    }

    # case 2: one-way/univariate
  } else if (dimension == 1L) {
    out <- lav_tables_oneway(
      lavobject = lavobject, lavdata = lavdata,
      statistic = statistic
    )

    # output format
    if (output == "data.frame") {
      class(out) <- c("lavaan.data.frame", "data.frame")
    } else {
      lav_msg_warn(gettextf("output option `%s' is not available; ignored.",
                            output))
    }

    # case 3a: two-way/pairwise/bivariate + cells
  } else if (dimension == 2L && type == "cells") {
    out <- lav_tables_pairwise_cells(
      lavobject = lavobject,
      lavdata = lavdata,
      statistic = statistic
    )
    # output format
    if (output == "data.frame") {
      class(out) <- c("lavaan.data.frame", "data.frame")
    } else if (output == "table") {
      out <- lav_tables_cells_format(out, lavdata = lavdata)
    } else {
      lav_msg_warn(gettextf("output option `%s' is not available; ignored.",
                            output))
    }

    #  case 3b: two-way/pairwise/bivariate + collapsed table
  } else if (dimension == 2L && (type == "table" || type == "tables")) {
    out <- lav_tables_pairwise_table(
      lavobject = lavobject,
      lavdata = lavdata,
      statistic = statistic,
      g2_min = g2_min,
      x2_min = x2_min,
      p_value = p_value
    )
    # output format
    if (output == "data.frame") {
      class(out) <- c("lavaan.data.frame", "data.frame")
    } else if (output == "table") {
      out <- lav_tables_table_format(out,
        lavdata = lavdata,
        lavobject = lavobject
      )
    } else {
      lav_msg_warn(gettext("output option `%s' is not available; ignored.",
                           output))
    }
  }

  if ((is.data.frame(out) && nrow(out) == 0L) ||
    (is.list(out) && length(out) == 0L)) {
    # empty table (perhaps, no categorical variables)
    return(invisible(out))
  }

  out
}

# shortcut, always dim=2, type="cells"
# lavTablesFit <- function(object,
#                         # if raw data, additional attributes
#                         categorical  = NULL,
#                         group        = NULL,
#                         # which statistics / fit indices?
#                         statistic    = "default",
#                         G2.min       = 3.0,
#                         X2.min       = 3.0,
#                         # pvalues for statistics?
#                         p_value      = FALSE,
#                         # output format
#                         output       = "data.frame") {
#
#    lavTables(object = object, dimension = 2L, type = "table",
#              categorical = categorical, group = group,
#              statistic = statistic,
#              g2_min = g2_min, x2_min = x2_min, p_value = p_value,
#              output = output, pattern_as_string = FALSE)
# }

# lavTables1D <- function(object,
#                        # if raw data, additional attributes
#                        categorical  = NULL,
#                       group        = NULL,
#                       # which statistics / fit indices?
#                        statistic    = "default",
#                        # output format
#                        output = "data.frame") {
#
#   lavTables(object = object, dimension = 1L,
#              categorical = categorical, group = group,
#              statistic = statistic, p_value = FALSE,
#              output = output, pattern_as_string = FALSE)
# }


lav_tables_pattern <- function(lavobject = NULL, lavdata = NULL,
                               statistic = NULL, pattern_as_string = TRUE) {
  # this only works if we have 'categorical' variables
  cat_idx <- which(lavdata@ov$type %in% c("ordered", "factor"))
  if (length(cat_idx) == 0L) {
    lav_msg_warn(gettext("no categorical variables are found"))
    return(data.frame(
      pattern = character(0L), nobs = integer(0L),
      obs.freq = integer(0L), obs.prop = numeric(0L)
    ))
  }
  # no support yet for mixture of endogenous ordered + numeric variables
  if (!is.null(lavobject) &&
    length(lav_object_vnames(lavobject, "ov.nox")) > length(cat_idx)) {
    lav_msg_warn(gettext("some endogenous variables are not categorical"))
    return(data.frame(
      pattern = character(0L), nobs = integer(0L),
      obs.freq = integer(0L), obs.prop = numeric(0L)
    ))
  }

  # default statistics
  if (!is.null(lavobject)) {
    if (length(statistic) == 1L && statistic == "default") {
      statistic <- c("G2", "X2")
    } else {
      stopifnot(statistic %in% c("G2.un", "X2.un", "G2", "X2"))
    }
  } else {
    # only data
    if (length(statistic) == 1L && statistic == "default") {
      # if data, none by default
      statistic <- character(0L)
    } else {
      stopifnot(statistic %in% c("G2.un", "X2.un"))
    }
  }

  # first, create basic table with response patterns
  for (g in 1:lavdata@ngroups) {
    pat <- lav_data_resp_patterns(lavdata@X[[g]])$pat
    obs_freq <- as.integer(rownames(pat))
    if (pattern_as_string) {
      pat <- data.frame(
        pattern = apply(pat, 1, paste, collapse = ""),
        stringsAsFactors = FALSE
      )
    } else {
      pat <- as.data.frame(pat, stringsAsFactors = FALSE)
      names(pat) <- lavdata@ov.names[[g]]
    }
    # pat$id <- 1:nrow(pat)
    if (lavdata@ngroups > 1L) {
      pat$group <- rep(g, nrow(pat))
    }
    nobs_1 <- sum(obs_freq)
    pat$nobs <- rep(nobs_1, nrow(pat))
    pat$obs.freq <- obs_freq
    rownames(pat) <- NULL
    if (g == 1L) {
      out <- pat
    } else {
      out <- rbind(out, pat)
    }
  }

  out$obs.prop <- out$obs.freq / out$nobs

  if (any(c("X2.un", "G2.un") %in% statistic)) {
    # not a good statistic... we only have uni+bivariate information
    lav_msg_warn(gettext(
      "limited information used for thresholds and correlations;
      but X2/G2 assumes full information"))
    pi0 <- lav_tables_resp_pi(
      lavobject = lavobject, lavdata = lavdata,
      est = "h1"
    )

    out$est.prop.un <- unlist(pi0)
    if ("G2.un" %in% statistic) {
      out$G2.un <- lav_tables_stat_g2(
        out$obs.prop, out$est.prop.un,
        out$nobs
      )
    }
    if ("X2.un" %in% statistic) {
      out$X2.un <- lav_tables_stat_x2(
        out$obs.prop, out$est.prop.un,
        out$nobs
      )
    }
  }
  if (any(c("X2", "G2") %in% statistic)) {
    if (lavobject@Options$estimator %in% c("FML")) {
      # ok, nothing to say
    } else if (lavobject@Options$estimator %in%
      c("WLS", "DWLS", "PML", "ULS")) {
      lav_msg_warn(gettextf(
        "estimator %s is not using full information while est.prop is
        using full information", lavobject@Options$estimator))
    } else {
      lav_msg_stop(gettextf(
        "estimator %s is not supported.", lavobject@Options$estimator))
    }

    pi0 <- lav_tables_resp_pi(
      lavobject = lavobject, lavdata = lavdata,
      est = "h0"
    )

    out$est.prop <- unlist(pi0)
    if ("G2" %in% statistic) {
      out$G2 <- lav_tables_stat_g2(
        out$obs.prop, out$est.prop,
        out$nobs
      )
    }
    if ("X2" %in% statistic) {
      out$X2 <- lav_tables_stat_x2(
        out$obs.prop, out$est.prop,
        out$nobs
      )
    }
  }

  # remove nobs?
  # out$nobs <- NULL

  out
}

# pairwise tables, rows = table cells
lav_tables_pairwise_cells <- function(lavobject = NULL, lavdata = NULL,
                                      statistic = character(0L)) {
  # this only works if we have at least two 'categorical' variables
  cat_idx <- which(lavdata@ov$type %in% c("ordered", "factor"))
  if (length(cat_idx) == 0L) {
    lav_msg_warn(gettext("no categorical variables are found"))
    return(data.frame(
      id = integer(0L), lhs = character(0L), rhs = character(0L),
      nobs = integer(0L), row = integer(0L), col = integer(0L),
      obs.freq = integer(0L), obs.prop = numeric(0L)
    ))
  }
  if (length(cat_idx) == 1L) {
    lav_msg_warn(gettext("at least two categorical variables are needed"))
    return(data.frame(
      id = integer(0L), lhs = character(0L), rhs = character(0L),
      nobs = integer(0L), row = integer(0L), col = integer(0L),
      obs.freq = integer(0L), obs.prop = numeric(0L)
    ))
  }

  # default statistics
  if (!is.null(lavobject)) {
    if (length(statistic) == 1L && statistic == "default") {
      statistic <- c("X2")
    } else {
      stopifnot(statistic %in% c(
        "cor", "th", "X2", "G2",
        "cor.un", "th.un", "X2.un", "G2.un"
      ))
    }
  } else {
    if (length(statistic) == 1L && statistic == "default") {
      # if data, none by default
      statistic <- character(0L)
    } else {
      stopifnot(statistic %in% c("cor.un", "th.un", "X2.un", "G2.un"))
    }
  }

  # initial table, observed cell frequencies
  out <- lav_tables_pairwise_freq_cell(
    lavdata = lavdata,
    as_data_frame = TRUE
  )
  out$obs.prop <- out$obs.freq / out$nobs

  if (any(c("cor.un", "th.un", "X2.un", "G2.un") %in% statistic)) {
    pi0 <- lav_tables_pairwise_sample_pi(
      lavobject = lavobject,
      lavdata = lavdata
    )
    out$est.prop.un <- unlist(pi0)
    if ("G2.un" %in% statistic) {
      out$G2.un <- lav_tables_stat_g2(
        out$obs.prop, out$est.prop.un,
        out$nobs
      )
    }
    if ("X2.un" %in% statistic) {
      out$X2.un <- lav_tables_stat_x2(
        out$obs.prop, out$est.prop.un,
        out$nobs
      )
    }

    if ("cor.un" %in% statistic) {
      cor_1 <- attr(pi0, "COR")
      cor_all <- unlist(lapply(cor_1, function(x) {
        x[lower.tri(x, diag = FALSE)]
      }))
      out$cor.un <- cor_all[out$id]
    }
  }
  if (any(c("cor", "th", "X2", "G2") %in% statistic)) {
    pi0 <- lav_tables_pairwise_model_pi(lavobject = lavobject)
    out$est.prop <- unlist(pi0)
    if ("G2" %in% statistic) {
      out$G2 <- lav_tables_stat_g2(
        out$obs.prop, out$est.prop,
        out$nobs
      )
    }
    if ("X2" %in% statistic) {
      out$X2 <- lav_tables_stat_x2(
        out$obs.prop, out$est.prop,
        out$nobs
      )
    }
    if ("cor" %in% statistic) {
      cor_1 <- attr(pi0, "COR")
      cor_all <- unlist(lapply(cor_1, function(x) {
        x[lower.tri(x, diag = FALSE)]
      }))
      out$cor <- cor_all[out$id]
    }
  }

  out
}

# G2 statistic
lav_tables_stat_g2 <- function(obs_prop = NULL, est_prop = NULL, nobs = NULL) {
  # not defined if out$obs_prop is (close to) zero
  zero_idx <- which(obs_prop < .Machine$double.eps)
  if (length(zero_idx)) {
    obs_prop[zero_idx] <- as.numeric(NA)
  }
  # the usual G2 formula
  g2 <- 2 * nobs * (obs_prop * log(obs_prop / est_prop))
  g2
}

# X2 (aka X2) statistic
lav_tables_stat_x2 <- function(obs_prop = NULL, est_prop = NULL, nobs = NULL) {
  res_prop <- obs_prop - est_prop
  x2 <- nobs * (res_prop * res_prop) / est_prop
  x2
}

# pairwise tables, rows = tables
lav_tables_pairwise_table <- function(lavobject = NULL, lavdata = NULL,
                                      statistic = character(0L),
                                      g2_min = 3.0,
                                      x2_min = 3.0,
                                      p_value = FALSE) {
  # default statistics
  if (!is.null(lavobject)) {
    if (length(statistic) == 1L && statistic == "default") {
      statistic <- c("X2", "X2.average")
    } else {
      stopifnot(statistic %in% c(
        "X2", "G2", "X2.un", "G2.un",
        "cor", "cor.un",
        "RMSEA.un", "RMSEA",
        "G2.average",
        "G2.nlarge",
        "G2.plarge",
        "X2.average",
        "X2.nlarge",
        "X2.plarge"
      ))
    }
  } else {
    if (length(statistic) == 1L && statistic == "default") {
      # if data, none by default
      statistic <- character(0L)
    } else {
      stopifnot(statistic %in% c(
        "cor.un", "X2.un", "G2.un",
        "RMSEA.un"
      ))
    }
  }

  # identify 'categorical' variables
  # cat.idx <- which(lavdata@ov$type %in% c("ordered","factor"))

  # pairwise tables
  # pairwise.tables <- utils::combn(vartable$name[cat.idx], m=2L)
  # pairwise.tables <- rbind(seq_len(ncol(pairwise.tables)),
  #                         pairwise.tables)
  # ntables <- ncol(pairwise.tables)

  # initial table, observed cell frequencies
  # out <- as.data.frame(t(pairwise.tables))
  # names(out) <- c("id", "lhs", "rhs")

  # collapse approach
  stat_cell <- character(0)
  if (any(c("G2", "G2.average", "G2.plarge", "G2.nlarge") %in% statistic)) {
    stat_cell <- c(stat_cell, "G2")
  }
  if (any(c("X2", "X2.average", "X2.plarge", "X2.nlarge") %in% statistic)) {
    stat_cell <- c(stat_cell, "X2")
  }
  if ("G2" %in% statistic || "RMSEA" %in% statistic) {
    stat_cell <- c(stat_cell, "G2")
  }
  if ("X2.un" %in% statistic) {
    stat_cell <- c(stat_cell, "X2.un")
  }
  if ("G2.un" %in% statistic || "RMSEA.un" %in% statistic) {
    stat_cell <- c(stat_cell, "G2.un")
  }
  if ("cor.un" %in% statistic) {
    stat_cell <- c(stat_cell, "cor.un")
  }
  if ("cor" %in% statistic) {
    stat_cell <- c(stat_cell, "cor")
  }

  # get table with table cells
  out_cell <- lav_tables_pairwise_cells(
    lavobject = lavobject,
    lavdata = lavdata,
    statistic = stat_cell
  )
  # only 1 row per table
  row_idx <- which(!duplicated(out_cell$id))
  if (is.null(out_cell$group)) {
    out <- out_cell[row_idx, c("lhs", "rhs", "nobs"), drop = FALSE]
  } else {
    out <- out_cell[row_idx, c("lhs", "rhs", "group", "nobs"), drop = FALSE]
  }

  # df
  if (length(statistic) > 0L) {
    nrow <- tapply(out_cell$row, INDEX = out_cell$id, FUN = max)
    ncol <- tapply(out_cell$col, INDEX = out_cell$id, FUN = max)
    out$df <- nrow * ncol - nrow - ncol
  }

  # cor
  if ("cor" %in% statistic) {
    out$cor <- out_cell[row_idx, "cor"]
  }

  # cor.un
  if ("cor.un" %in% statistic) {
    out$cor.un <- out_cell[row_idx, "cor.un"]
  }

  # X2
  if ("X2" %in% statistic) {
    out$X2 <- tapply(out_cell$X2,
      INDEX = out_cell$id, FUN = sum,
      na.rm = TRUE
    )
    if (p_value) {
      out$X2.pval <- pchisq(out$X2, df = out$df, lower.tail = FALSE)
    }
  }
  if ("X2.un" %in% statistic) {
    out$X2.un <- tapply(out_cell$X2.un,
      INDEX = out_cell$id, FUN = sum,
      na.rm = TRUE
    )
    if (p_value) {
      out$X2.un.pval <- pchisq(out$X2.un, df = out$df, lower.tail = FALSE)
    }
  }

  # G2
  if ("G2" %in% statistic) {
    out$G2 <- tapply(out_cell$G2,
      INDEX = out_cell$id, FUN = sum,
      na.rm = TRUE
    )
    if (p_value) {
      out$G2.pval <- pchisq(out$G2, df = out$df, lower.tail = FALSE)
    }
  }
  if ("G2.un" %in% statistic) {
    out$G2.un <- tapply(out_cell$G2.un,
      INDEX = out_cell$id, FUN = sum,
      na.rm = TRUE
    )
    if (p_value) {
      out$G2.un.pval <- pchisq(out$G2.un, df = out$df, lower.tail = FALSE)
    }
  }

  if ("RMSEA" %in% statistic) {
    g2 <- tapply(out_cell$G2, INDEX = out_cell$id, FUN = sum, na.rm = TRUE)
    # note: there seems to be a mistake in Appendix 1 eqs 43/44 of Joreskog
    # SSI paper (2005) 'SEM with ordinal variables using LISREL'
    # 2*N*d should N*d
    out$RMSEA <- sqrt(pmax(0, (g2 - out$df) / (out$nobs * out$df)))
    if (p_value) {
      # note: MUST use 1 - pchisq (instead of lower.tail = FALSE)
      # because for ncp > 80, routine only computes lower tail
      out$RMSEA.pval <- 1.0 - pchisq(g2,
        ncp = 0.1 * 0.1 * out$nobs * out$df,
        df = out$df, lower.tail = TRUE
      )
    }
  }
  if ("RMSEA.un" %in% statistic) {
    g2 <- tapply(out_cell$G2.un,
      INDEX = out_cell$id, FUN = sum,
      na.rm = TRUE
    )
    # note: there seems to be a mistake in Appendix 1 eqs 43/44 of Joreskog
    # SSI paper (2005) 'SEM with ordinal variables using LISREL'
    # 2*N*d should N*d
    out$RMSEA.un <- sqrt(pmax(0, (g2 - out$df) / (out$nobs * out$df)))
    if (p_value) {
      # note: MUST use 1 - pchisq (instead of lower.tail = FALSE)
      # because for ncp > 80, routine only computes lower tail
      out$RMSEA.un.pval <- 1.0 - pchisq(g2,
        ncp = 0.1 * 0.1 * out$nobs * out$df,
        df = out$df, lower.tail = TRUE
      )
    }
  }

  if ("G2.average" %in% statistic) {
    out$G2.average <- tapply(out_cell$G2,
      INDEX = out_cell$id, FUN = mean,
      na.rm = TRUE
    )
  }

  if ("G2.nlarge" %in% statistic) {
    out$G2.min <- rep(g2_min, length(out$lhs))
    out$G2.nlarge <- tapply(out_cell$G2,
      INDEX = out_cell$id,
      FUN = function(x) sum(x > g2_min, na.rm = TRUE)
    )
  }

  if ("G2.plarge" %in% statistic) {
    out$G2.min <- rep(g2_min, length(out$lhs))
    out$G2.plarge <- tapply(out_cell$G2,
      INDEX = out_cell$id,
      FUN = function(x) sum(x > g2_min, na.rm = TRUE) / length(x)
    )
  }

  if ("X2.average" %in% statistic) {
    out$X2.average <- tapply(out_cell$X2,
      INDEX = out_cell$id, FUN = mean,
      na.rm = TRUE
    )
  }

  if ("X2.nlarge" %in% statistic) {
    out$X2.min <- rep(x2_min, length(out$lhs))
    out$X2.nlarge <- tapply(out_cell$X2,
      INDEX = out_cell$id,
      FUN = function(x) sum(x > x2_min, na.rm = TRUE)
    )
  }

  if ("X2.plarge" %in% statistic) {
    out$X2.min <- rep(x2_min, length(out$lhs))
    out$X2.plarge <- tapply(out_cell$X2,
      INDEX = out_cell$id,
      FUN = function(x) sum(x > x2_min, na.rm = TRUE) / length(x)
    )
  }

  out
}


lav_tables_oneway <- function(lavobject = NULL, lavdata = NULL,
                              statistic = NULL) {
  # shortcuts
  vartable <- lavdata@ov
  xx <- lavdata@X

  # identify 'categorical' variables
  cat_idx <- which(vartable$type %in% c("ordered", "factor"))
  ncat <- length(cat_idx)

  # do we have any categorical variables?
  if (length(cat_idx) == 0L) {
    lav_msg_warn(gettext("no categorical variables are found"))
    return(data.frame(
      id = integer(0L), lhs = character(0L), rhs = character(0L),
      nobs = integer(0L),
      obs.freq = integer(0L), obs.prop = numeric(0L),
      est.prop = numeric(0L), X2 = numeric(0L)
    ))
  } else {
    labels <- strsplit(vartable$lnam[cat_idx], "\\|")
  }

  # ok, we have an overview of all categorical variables in the data
  ngroups <- length(xx)

  # for each group, for each categorical variable, collect information
  tables <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    tables[[g]] <- lapply(seq_len(ncat),
      FUN = function(x) {
        idx <- cat_idx[x]
        nrow <- vartable$nlev[idx]
        ncell <- nrow
        nvar <- length(lavdata@ov.names[[g]])
        id <- (g - 1) * nvar + x

        # compute observed frequencies
        freq <- tabulate(xx[[g]][, idx], nbins = ncell)

        list(
          id = rep.int(id, ncell),
          lhs = rep.int(vartable$name[idx], ncell),
          # op = rep.int("freq", ncell),
          rhs = labels[[x]],
          group = rep.int(g, ncell),
          nobs = rep.int(sum(freq), ncell),
          obs.freq = freq,
          obs.prop = freq / sum(freq)
        )
      }
    )
  }

  for (g in 1:ngroups) {
    table_1 <- tables[[g]]
    table_1 <- lapply(table_1, as.data.frame, stringsAsFactors = FALSE)
    if (g == 1L) {
      out <- do.call(rbind, table_1)
    } else {
      out <- rbind(out, do.call(rbind, table_1))
    }
  }
  if (g == 1) {
    # remove group column
    out$group <- NULL
  }

  # default statistics
  if (!is.null(lavobject)) {
    if (length(statistic) == 1L && statistic == "default") {
      statistic <- c("X2")
    } else {
      stopifnot(statistic %in% c(
        "th.un",
        "th", "G2", "X2"
      ))
    }

    # sample based
    # note, there is no G2.un or X2.un: always saturated!
    if ("th.un" %in% statistic) {
      # sample based
      th_un <- unlist(lapply(1:lavdata@ngroups, function(x) {
        if (lavobject@Model@conditional.x) {
          th <- lavobject@SampleStats@res.th[[x]][
            lavobject@SampleStats@th.idx[[x]] > 0
          ]
        } else {
          th <- lavobject@SampleStats@th[[x]][
            lavobject@SampleStats@th.idx[[x]] > 0
          ]
        }
        th_idx <- lavobject@SampleStats@th.idx[[x]][
          lavobject@SampleStats@th.idx[[x]] > 0
        ]
        unname(unlist(tapply(th,
          INDEX = th_idx,
          function(y) c(y, Inf)
        )))
      }))
      # overwrite obs.prop
      # NOTE: if we have exogenous variables, obs.prop will NOT
      #       correspond with qnorm(th_un)
      out$obs.prop <- unname(unlist(tapply(th_un,
        INDEX = out$id,
        FUN = function(x) {
          (pnorm(c(x, Inf)) -
            pnorm(c(-Inf, x)))[-(length(x) + 1)]
        }
      )))

      out$th.un <- th_un
    }

    # model based
    if (any(c("th", "G2", "X2") %in% statistic)) {
      # model based
      th_h0 <- unlist(lapply(1:lavdata@ngroups, function(x) {
        if (lavobject@Model@conditional.x) {
          th <- lavobject@implied$res.th[[x]][
            lavobject@SampleStats@th.idx[[x]] > 0
          ]
        } else {
          th <- lavobject@implied$th[[x]][
            lavobject@SampleStats@th.idx[[x]] > 0
          ]
        }
        th_idx <- lavobject@SampleStats@th.idx[[x]][
          lavobject@SampleStats@th.idx[[x]] > 0
        ]
        unname(unlist(tapply(th,
          INDEX = th_idx,
          function(x) c(x, Inf)
        )))
      }))

      est_prop <- unname(unlist(tapply(th_h0,
        INDEX = out$id,
        FUN = function(x) {
          (pnorm(c(x, Inf)) -
            pnorm(c(-Inf, x)))[-(length(x) + 1)]
        }
      )))
      out$est.prop <- est_prop

      if ("th" %in% statistic) {
        out$th <- th_h0
      }
      if ("G2" %in% statistic) {
        out$G2 <- lav_tables_stat_g2(
          out$obs.prop, out$est.prop,
          out$nobs
        )
      }
      if ("X2" %in% statistic) {
        out$X2 <- lav_tables_stat_x2(
          out$obs.prop, out$est.prop,
          out$nobs
        )
      }
    }
  } else {
    if (length(statistic) == 1L && statistic == "default") {
      # if data, none by default
      statistic <- character(0L)
    } else {
      stopifnot(statistic %in% c("th.un"))
    }

    if ("th.un" %in% statistic) {
      out$th.un <- unlist(tapply(out$obs.prop,
        INDEX = out$id,
        FUN = function(x) qnorm(cumsum(x))
      ))
    }
  }

  out
}

# HJ 15/1/2023 MODIFIED to add sampling weights
# compute pairwise (two-way) frequency tables
lav_tables_pairwise_freq_cell <- function(lavdata = NULL,
                                          as_data_frame = TRUE) {
  # shortcuts
  vartable <- as.data.frame(lavdata@ov, stringsAsFactors = FALSE)
  xx <- lavdata@X
  # ov_names <- lavdata@ov.names
  ngroups <- lavdata@ngroups
  # wt <- lavdata@weights

  # identify 'categorical' variables
  cat_idx <- which(vartable$type %in% c("ordered", "factor"))

  # do we have any categorical variables?
  if (length(cat_idx) == 0L) {
    lav_msg_stop(gettext("no categorical variables are found"))
  } else if (length(cat_idx) == 1L) {
    lav_msg_stop(gettext("at least two categorical variables are needed"))
  }

  # pairwise tables
  pairwise_tables <- utils::combn(vartable$name[cat_idx], m = 2L)
  pairwise_tables <- rbind(pairwise_tables, seq_len(ncol(pairwise_tables)))
  ntables <- ncol(pairwise_tables)

  # for each group, for each pairwise table, collect information
  tables <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    tables[[g]] <- apply(pairwise_tables,
      MARGIN = 2,
      FUN = function(x) {
        idx1 <- which(vartable$name == x[1])
        idx2 <- which(vartable$name == x[2])
        id <- (g - 1) * ntables + as.numeric(x[3])
        nrow <- vartable$nlev[idx1]
        ncol <- vartable$nlev[idx2]
        ncell <- nrow * ncol

        # compute two-way observed frequencies
        y1 <- xx[[g]][, idx1]
        y2 <- xx[[g]][, idx2]
        # freq <- table(y1, y2) # we lose missings; useNA is ugly
        freq <- lav_bvord_freq(y1, y2)

        # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

        # If we want to use weighted frequencies we can use the code
        # below. However, it will probably make sense only when the
        # weights are normalised. If they're not, we may get quite ugly
        # and nonsensical numbers here. So for now, just keep the
        # lavtables as is (using non-weighted frequencies).
        #
        # freq <- lav_bvord_freq(y1, Y2, wt[[g]])

        # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


        list(
          id = rep.int(id, ncell),
          lhs = rep.int(x[1], ncell),
          # op = rep.int("table", ncell),
          rhs = rep.int(x[2], ncell),
          group = rep.int(g, ncell),
          nobs = rep.int(sum(freq), ncell),
          row = rep.int(seq_len(nrow), times = ncol),
          col = rep(seq_len(ncol), each = nrow),
          obs.freq = lav_mat_vec(freq) # col by col!
        )
      }
    )
  }

  if (as_data_frame) {
    for (g in 1:ngroups) {
      table_1 <- tables[[g]]
      table_1 <- lapply(table_1, as.data.frame, stringsAsFactors = FALSE)
      if (g == 1) {
        out <- do.call(rbind, table_1)
      } else {
        out <- rbind(out, do.call(rbind, table_1))
      }
    }
    if (g == 1) {
      # remove group column
      out$group <- NULL
    }
  } else {
    if (ngroups == 1L) {
      out <- tables[[1]]
    } else {
      out <- tables
    }
  }

  out
}


# low-level function to compute expected proportions per cell
# object
lav_tables_pairwise_model_pi <- function(lavobject = NULL) {
  stopifnot(lavobject@Model@categorical)

  # shortcuts
  lavmodel <- lavobject@Model
  implied <- lavobject@implied
  ngroups <- lavobject@Data@ngroups
  ov_types <- lavobject@Data@ov$type
  th_idx <- lavobject@Model@th.idx
  sigma_hat <- if (lavmodel@conditional.x) implied$res.cov else implied$cov
  th <- if (lavmodel@conditional.x) implied$res.th else implied$th

  pi0 <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    sigmahat <- sigma_hat[[g]]
    cors <- sigmahat[lower.tri(sigmahat)]
    if (any(abs(cors) > 1)) {
      lav_msg_warn(gettext(
        "some model-implied correlations are larger than 1.0"))
    }
    nvar <- nrow(sigmahat)

    # shortcut for all ordered - tablewise
    if (all(ov_types == "ordered") && !is.null(lavobject@Cache[[g]]$long)) {
      # FREQ.OBS <- c(FREQ.OBS, lavobject@Cache[[g]]$bifreq)
      long2 <- lav_pml_longvec_th_rho(
        no_x = nvar,
        all_thres = th[[g]],
        index_var_of_thres = th_idx[[g]],
        rho_xixj = cors
      )
      # get expected probability per table, per pair
      pi0[[g]] <- lav_pml_expprob_vec(
        ind_vec = lavobject@Cache[[g]]$long,
        th_rho_vec = long2
      )
    } else {
      pi_group <- integer(0)
      # order! first i, then j, lav_mat_vec(table)!
      for (i in seq_len(nvar - 1L)) {
        for (j in (i + 1L):nvar) {
          if (ov_types[i] == "ordered" && ov_types[j] == "ordered") {
            pi_table <- lav_bvord_noexo_pi(
              rho = sigmahat[i, j],
              th_y1 = th[[g]][th_idx[[g]] == i],
              th_y2 = th[[g]][th_idx[[g]] == j]
            )
            pi_group <- c(pi_group, lav_mat_vec(pi_table))
          }
        }
      }
      pi0[[g]] <- pi_group
    }
  } # g

  # add  COR/th/TH.IDX
  attr(pi0, "COR") <- sigma_hat
  attr(pi0, "TH") <- th
  attr(pi0, "TH.IDX") <- th_idx

  pi0
}

# low-level function to compute expected proportions per cell
# using sample-based correlations + thresholds
#
# object can be either lavData or lavaan class
lav_tables_pairwise_sample_pi <- function(lavobject = NULL, lavdata = NULL) {
  # get COR, TH and th.idx
  if (!is.null(lavobject)) {
    if (lavobject@Model@conditional.x) {
      cor_1 <- lavobject@SampleStats@res.cov
      th <- lavobject@SampleStats@res.th
    } else {
      cor_1 <- lavobject@SampleStats@cov
      th <- lavobject@SampleStats@th
    }
    th_idx <- lavobject@SampleStats@th.idx
  } else if (!is.null(lavdata)) {
    fit_un <- lav_object_cor(object = lavdata, se = "none", output = "fit")
    if (fit_un@Model@conditional.x) {
      cor_1 <- fit_un@SampleStats@res.cov
      th <- fit_un@SampleStats@res.th
    } else {
      cor_1 <- fit_un@SampleStats@cov
      th <- fit_un@SampleStats@th
    }
    th_idx <- fit_un@SampleStats@th.idx
  } else {
    lav_msg_stop(gettext("both lavobject and lavdata are NULL"))
  }

  lav_tables_pairwise_sample_pi_cor(
    cor_1 = cor_1, th = th,
    th_idx = th_idx
  )
}

# low-level function to compute expected proportions per cell
lav_tables_pairwise_sample_pi_cor <- function(cor_1 = NULL, th = NULL, # nolint
                                              th_idx = NULL) {
  ngroups <- length(cor_1)

  pi0 <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    sigmahat <- cor_1[[g]]
    cors <- sigmahat[lower.tri(sigmahat)]
    if (any(abs(cors) > 1)) {
      lav_msg_warn(gettext(
        "some model-implied correlations are larger than 1.0"))
    }
    nvar <- nrow(sigmahat)
    th_idx_1 <- th_idx[[g]]

    # reconstruct ov.types
    ov_types <- rep("numeric", nvar)
    ord_idx <- unique(th_idx_1[th_idx_1 > 0])
    ov_types[ord_idx] <- "ordered"

    pi_group <- integer(0)
    # order! first i, then j, lav_mat_vec(table)!
    for (i in seq_len(nvar - 1L)) {
      for (j in (i + 1L):nvar) {
        if (ov_types[i] == "ordered" && ov_types[j] == "ordered") {
          pi_table <- lav_bvord_noexo_pi(
            rho = sigmahat[i, j],
            th_y1 = th[[g]][th_idx_1 == i],
            th_y2 = th[[g]][th_idx_1 == j]
          )
          pi_group <- c(pi_group, lav_mat_vec(pi_table))
        }
      }
    }
    pi0[[g]] <- pi_group
  } # g

  # add  COR/TH/TH.IDX
  attr(pi0, "COR") <- cor_1
  attr(pi0, "TH") <- th
  attr(pi0, "TH.IDX") <- th_idx

  pi0
}

# low-level function to compute expected proportions per PATTERN
# using sample-based correlations + thresholds
#
# object can be either lavData or lavaan class
#
# only valid if estimator = FML, POM or NOR
#
lav_tables_resp_pi <- function(lavobject = NULL, lavdata = NULL,
                               est = "h0") {
  # shortcuts
  if (!is.null(lavobject)) {
    lavmodel <- lavobject@Model
    implied <- lavobject@implied
  }
  ngroups <- lavdata@ngroups

  # h0 or unrestricted?
  if (est == "h0") {
    sigma_hat <- if (lavmodel@conditional.x) implied$res.cov else implied$cov
    th <- if (lavmodel@conditional.x) implied$res.th else implied$th
    th_idx_1 <- lavobject@SampleStats@th.idx
  } else {
    if (is.null(lavobject)) {
      fit_un <- lav_object_cor(object = lavdata, se = "none", output = "fit")
      sigma_hat <-
        if (fit_un@Model@conditional.x) fit_un@implied$res.cov else fit_un@implied$cov
      th <-
        if (fit_un@Model@conditional.x) fit_un@implied$res.th else fit_un@implied$th
      th_idx_1 <- fit_un@SampleStats@th.idx
    } else {
      if (lavobject@Model@conditional.x) {
        sigma_hat <- lavobject@SampleStats@res.cov
        th <- lavobject@SampleStats@res.th
      } else {
        sigma_hat <- lavobject@SampleStats@cov
        th <- lavobject@SampleStats@th
      }
      th_idx_1 <- lavobject@SampleStats@th.idx
    }
  }

  pi0 <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    sigmahat <- sigma_hat[[g]]
    cors <- sigmahat[lower.tri(sigmahat)]
    if (any(abs(cors) > 1)) {
      lav_msg_warn(gettext(
        "some model-implied correlations are larger than 1.0"))
    }
    nvar <- nrow(sigmahat)
    th_idx <- th_idx_1[[g]]
    mean_1 <- rep(0, nvar)

    # reconstruct ov.types
    ov_types <- rep("numeric", nvar)
    ord_idx <- unique(th_idx[th_idx > 0])
    ov_types[ord_idx] <- "ordered"

    if (all(ov_types == "ordered")) {
      # get patterns ## FIXME GET it
      if (!is.null(lavdata@Rp[[g]]$pat)) {
        pat_1 <- lavdata@Rp[[g]]$pat
      } else {
        pat_1 <- lav_data_resp_patterns(lavdata@X[[g]])$pat
      }
      npatterns <- nrow(pat_1)
      freq <- as.numeric(rownames(pat_1))
      pi_group <- numeric(npatterns)
      th_var <- lapply(
        1:nvar,
        function(x) c(-Inf, th[[g]][th_idx == x], +Inf)
      )
      # FIXME!!! ok to set diagonal to 1.0?
      diag(sigmahat) <- 1.0
      for (r in 1:npatterns) {
        # compute probability for each pattern
        lower <- sapply(1:nvar, function(x) th_var[[x]][pat_1[r, x]])
        upper <- sapply(1:nvar, function(x) th_var[[x]][pat_1[r, x] + 1L])
        # handle missing values
        na_idx <- which(is.na(pat_1[r, ]))
        if (length(na_idx) > 0L) {
          lower <- lower[-na_idx]
          upper <- upper[-na_idx]
          mean_r <- mean_1[-na_idx]
          sigmahat_r <- sigmahat[-na_idx, -na_idx, drop = FALSE]
        } else {
          mean_r <- mean_1
          sigmahat_r <- sigmahat
        }
        pi_group[r] <- sadmvn(lower, upper,
          mean = mean_r,
          varcov = sigmahat_r
        )
      }
    } else { # case-wise
      pi_group <- rep(as.numeric(NA), lavdata@nobs[[g]])
      lav_msg_warn(gettext("casewise PI not implemented"))
    }

    pi0[[g]] <- pi_group
  } # g

  pi0
}

lav_tables_table_format <- function(out, lavdata = lavdata,
                                    lavobject = lavobject) {
  # determine column we need
  names_1 <- names(out)
  stat_idx <- which(names_1 %in% c(
    "cor", "cor.un",
    "G2", "G2.un",
    "X2", "X2.un",
    "RMSEA", "RMSEA.un",
    "G2.average", "G2.plarge", "G2.nlarge",
    "X2.average", "X2.plarge", "X2.nlarge"
  ))
  if (length(stat_idx) == 0) {
    if (!is.null(out$obs.freq)) {
      stat_idx <- which(names_1 == "obs.freq")
    } else if (!is.null(out$nobs)) {
      stat_idx <- which(names_1 == "nobs")
    }
    uni <- NULL
  } else if (length(stat_idx) > 1) {
    lav_msg_stop(gettext(
      "more than one statistic for table output:"),
      paste(names_1[stat_idx], collapse = " ")
    )
  } else {
    # univariate version of same statistic
    if (names_1[stat_idx] == "G2.average") {
      uni <- lavTables(lavobject, dimension = 1L, statistic = "G2")
    } else if (names_1[stat_idx] == "X2.average") {
      uni <- lavTables(lavobject, dimension = 1L, statistic = "X2")
    } else {
      uni <- NULL
    }
  }

  out_1 <- vector("list", length = lavdata@ngroups)
  for (g in 1:lavdata@ngroups) {
    if (lavdata@ngroups == 1L) { # no group column
      stat <- out[[stat_idx]]
    } else {
      stat <- out[[stat_idx]][out$group == g]
    }
    rn <- lavdata@ov.names[[g]]
    out_1[[g]] <- lav_getcov(stat, diagonal = FALSE, lower = FALSE, names = rn)
    # change diagonal elements: replace by univariate stat
    # if possible
    diag(out_1[[g]]) <- as.numeric(NA)
    if (!is.null(uni)) {
      if (!is.null(uni$group)) {
        idx <- which(uni$group == g)
      } else {
        idx <- seq_along(uni$lhs)
      }
      if (names_1[stat_idx] == "G2.average") {
        diag(out_1[[g]]) <- tapply(uni$G2[idx],
          INDEX = uni$id[idx],
          FUN = mean
        )
      } else if (names_1[stat_idx] == "X2.average") {
        diag(out_1[[g]]) <- tapply(uni$X2[idx],
          INDEX = uni$id[idx],
          FUN = mean
        )
      }
    } else if (names_1[stat_idx] %in% c("cor", "cor.un")) {
      diag(out_1[[g]]) <- 1
    }
    class(out_1[[g]]) <- c("lavaan.matrix.symmetric", "matrix")
  }
  if (lavdata@ngroups > 1L) {
    names(out_1) <- lavdata@group.label
    out <- out_1
  } else {
    out <- out_1[[1]]
  }

  out
}

lav_tables_cells_format <- function(out, lavdata = lavdata,
                                    drop_list_single_group = FALSE) {
  out_1 <- vector("list", length = lavdata@ngroups)
  if (is.null(out$group)) {
    out$group <- rep(1L, length(out$lhs))
  }
  # do we have a statistic?
  # determine column we need
  names_1 <- names(out)
  stat_idx <- which(names_1 %in% c(
    "cor", "cor.un",
    "G2", "G2.un",
    "X2", "X2.un",
    "RMSEA", "RMSEA.un",
    "G2.average", "G2.plarge", "G2.nlarge",
    "X2.average", "X2.plarge", "X2.nlarge"
  ))
  if (length(stat_idx) == 0) {
    statistic <- "obs.freq"
  } else if (length(stat_idx) > 1) {
    lav_msg_stop(gettext(
      "more than one statistic for table output:"),
      paste(names_1[stat_idx], collapse = " ")
    )
  } else {
    statistic <- names_1[stat_idx]
  }

  for (g in 1:lavdata@ngroups) {
    case_idx <- which(out$group == g)
    id_group <- unique(out$id[out$group == g])
    tmp <- lapply(id_group, function(x) {
      tx <- out[out$id == x, ]
      m <- matrix(
        tx[, statistic],
        max(tx$row), max(tx$col)
      )
      rownames(m) <- unique(tx$row)
      colnames(m) <- unique(tx$col)
      class(m) <- c("lavaan.matrix", "matrix")
      m
    })
    names(tmp) <- unique(paste(out$lhs[case_idx], out$rhs[case_idx],
      sep = "_"
    ))
    out_1[[g]] <- tmp
  }

  if (lavdata@ngroups == 1L && drop_list_single_group) {
    out_1 <- out_1[[1]]
  } else {
    if (length(lavdata@group.label) > 0L) {
      names(out_1) <- unlist(lavdata@group.label)
    }
  }

  out_1
}



# The function lav_tables_uni_freq_cell computes the univariate (one-way)
# frequency tables.
# The function closely follows the "logic" of the lavaan function
# lav_tables_pairwise_freq_cell.
# The output is either a list or a data.frame depending on the value the logical
# input argument as.data.frame. Either way, the same information is contained
# which is:
# a) the observed (univariate) frequencies f_ia, i=1,...,p (variables),
#    a=1,...,ci (response categories), with a index running faster than i index.
# b) an index vector with the name varb which indicates which variable each
#    frequency refers to.
# c) an index vector with the name group which indicates which group each
#    frequency refers to when multi-group analysis.
# d) an index vector with the name level which indicates which level within
#    each ordinal variable each frequency refers to.
# e) a vector nobs which gives how many cases were considered to compute the
#    corresponding frequency. Since we use the available data for each variable
#    when missing=="available_cases" we expect these numbers to differ when
#    missing values are present.
# f) an index vector with the name id indexing each univariate table,
#    1 goes to first variable in the first group, 2 to 2nd variable in the
#    second group and so on. The last table has the index equal to
#                                   (no of groups) * (no of variables).

lav_tables_uni_freq_cell <- function(lavdata = NULL,
                                            as_data_frame = TRUE) {
  # shortcuts
  vartable <- as.data.frame(lavdata@ov, stringsAsFactors = FALSE)
  xx <- lavdata@X
  # ov_names <- lavdata@ov.names
  ngroups <- lavdata@ngroups

  # identify 'categorical' variables
  cat_idx <- which(vartable$type %in% c("ordered", "factor"))

  # do we have any categorical variables?
  if (length(cat_idx) == 0L) {
    lav_msg_stop(gettext("no categorical variables are found"))
  }

  # univariate tables
  univariate_tables <- vartable$name[cat_idx]
  univariate_tables <- rbind(univariate_tables,
    seq_along(univariate_tables),
    deparse.level = 0
  )
  ntables <- ncol(univariate_tables)

  # for each group, for each pairwise table, collect information
  uni_tables <- vector("list", length = ngroups)
  for (g in 1:ngroups) {
    uni_tables[[g]] <- apply(univariate_tables,
      MARGIN = 2,
      FUN = function(x) {
        idx1 <- which(vartable$name == x[1])
        id <- (g - 1) * ntables + as.numeric(x[2])
        ncell <- vartable$nlev[idx1]

        # compute one-way observed frequencies
        y1 <- xx[[g]][, idx1]
        uni_freq <- tabulate(y1, nbins = max(y1, na.rm = TRUE))

        list(
          id = rep.int(id, ncell),
          varb = rep.int(x[1], ncell),
          group = rep.int(g, ncell),
          nobs = rep.int(sum(uni_freq), ncell),
          level = seq_len(ncell),
          obs.freq = uni_freq
        )
      }
    )
  }

  if (as_data_frame) {
    for (g in 1:ngroups) {
      uni_table <- uni_tables[[g]]
      uni_table <- lapply(uni_table, as.data.frame,
        stringsAsFactors = FALSE
      )
      if (g == 1) {
        out <- do.call(rbind, uni_table)
      } else {
        out <- rbind(out, do.call(rbind, uni_table))
      }
    }
    if (g == 1) {
      # remove group column
      out$group <- NULL
    }
  } else {
    if (ngroups == 1L) {
      out <- uni_tables[[1]]
    } else {
      out <- uni_tables
    }
  }

  out
}

# YR - 26 Nov 2013: generate partable for the unrestricted model
# YR - 19 Mar 2017: handle twolevel model
# YR - 27 May 2021: added lav_pt_unrestricted_chol so we can use
#                   a cholesky parameterization: S = LAMBDA %*% t(LAMBDA)

lav_pt_unrestricted <- function(lavobject = NULL,          # nolint start
                                      # if no object is available,
                                      lavdata = NULL,
                                      lavpta = NULL,
                                      lavoptions = NULL,
                                      lavsamplestats = NULL,
                                      lavh1 = NULL,
                                      # optional user-provided sample stats
                                      sample_cov = NULL,
                                      sample_mean = NULL,
                                      sample_slopes = NULL,
                                      sample_th = NULL,
                                      sample_th_idx = NULL,
                                      sample_cov_x = NULL,
                                      sample_mean_x = NULL) {   # nolint end
  lav_pt_indep_or_unrestricted(
    lavobject = lavobject,
    lavdata = lavdata, lavpta = lavpta, lavoptions = lavoptions,
    lavsamplestats = lavsamplestats, lavh1 = lavh1,
    sample_cov = sample_cov, sample_mean = sample_mean,
    sample_slopes = sample_slopes,
    sample_th = sample_th, sample_th_idx = sample_th_idx,
    independent = FALSE
  )
}

# generate parameter table for an independence model
# YR - 12 Sep 2017: special case of lav_pt_unrestricted()
lav_pt_independence <- function(lavobject = NULL,         # nolint start
                                      # if no object is available,
                                      lavdata = NULL,
                                      lavpta = NULL,
                                      lavoptions = NULL,
                                      lavsamplestats = NULL,
                                      lavh1 = NULL,
                                      # optional user-provided sample stats
                                      sample_cov = NULL,
                                      sample_mean = NULL,
                                      sample_slopes = NULL,
                                      sample_th = NULL,
                                      sample_th_idx = NULL,
                                      sample_cov_x = NULL,
                                      sample_mean_x = NULL) {   # nolint end
  lav_pt_indep_or_unrestricted(
    lavobject = lavobject,
    lavdata = lavdata, lavpta = lavpta, lavoptions = lavoptions,
    lavsamplestats = lavsamplestats, lavh1 = lavh1,
    sample_cov = sample_cov, sample_mean = sample_mean,
    sample_slopes = sample_slopes,
    sample_th = sample_th, sample_th_idx = sample_th_idx,
    independent = TRUE
  )
}

lav_pt_indep_or_unrestricted <- function(lavobject = NULL,
                                               # if no object is available,
                                               lavdata = NULL,
                                               lavpta = NULL,
                                               lavoptions = NULL,
                                               lavsamplestats = NULL,
                                               lavh1 = NULL,
                                           # optional user-provided sample stats
                                               sample_cov = NULL,
                                               sample_mean = NULL,
                                               sample_slopes = NULL,
                                               sample_th = NULL,
                                               sample_th_idx = NULL,
                                               sample_cov_x = NULL,
                                               sample_mean_x = NULL,
                                               independent = FALSE) {
  # grab everything from lavaan lavobject
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))

    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
    lavsamplestats <- lavobject@SampleStats
    lavpta <- lavobject@pta
    lavh1 <- lavobject@h1
  }

  if (lavdata@data.type == "none") {
    lavsamplestats <- NULL
  }

  # conditional.x ? check res.cov[[1]] slot
  conditional_x <- FALSE
  if (!is.null(lavsamplestats) && !is.null(lavsamplestats@res.cov[[1]])) {
    conditional_x <- TRUE
  } else if (!is.null(lavoptions) && lavoptions$conditional.x) {
    conditional_x <- TRUE
  }

  # group.w.free?
  group_w_free <- FALSE
  if (!is.null(lavoptions) && lavoptions$group.w.free) {
    group_w_free <- TRUE
  }
  # we use CAPS below for the list version, so we can use 'small caps'
  # within the for() loop

  # get sample statistics, all groups
  sample_cov_1 <- sample_cov
  if (is.null(sample_cov_1) && !is.null(lavsamplestats)) {
    if (conditional_x) {
      sample_cov_1 <- lavsamplestats@res.cov
    } else {
      sample_cov_1 <- lavsamplestats@cov
    }
  }

  sample_mean_1 <- sample_mean
  if (is.null(sample_mean_1) && !is.null(lavsamplestats)) {
    if (conditional_x) {
      sample_mean_1 <- lavsamplestats@res.int
    } else {
      sample_mean_1 <- lavsamplestats@mean
    }
  }

  sample_slopes_1 <- sample_slopes
  if (conditional_x && is.null(sample_slopes_1) && !is.null(lavsamplestats)) {
    sample_slopes_1 <- lavsamplestats@res.slopes
  }

  sample_th_1 <- sample_th
  if (is.null(sample_th_1) && !is.null(lavsamplestats)) {
    if (conditional_x) {
      sample_th_1 <- lavsamplestats@res.th
    } else {
      sample_th_1 <- lavsamplestats@th
    }
  }

  sample_th_idx_1 <- sample_th_idx
  if (is.null(sample_th_idx_1) && !is.null(lavsamplestats)) {
    sample_th_idx_1 <- lavsamplestats@th.idx
  }

  sample_cov_x_1 <- sample_cov_x
  if (is.null(sample_cov_x_1) && !is.null(lavsamplestats)) {
    sample_cov_x_1 <- lavsamplestats@cov.x
  }

  sample_mean_x_1 <- sample_mean_x
  if (is.null(sample_mean_x_1) && !is.null(lavsamplestats)) {
    sample_mean_x_1 <- lavsamplestats@mean.x
  }



  ov <- lavdata@ov
  meanstructure <- lavoptions$meanstructure
  categorical <- any(ov$type == "ordered")
  ngroups <- lavdata@ngroups
  nlevels <- lavdata@nlevels
  if (lavoptions$estimator == "catML") {
    categorical <- FALSE
  }
  correlation <- FALSE
  if (!is.null(lavoptions$correlation)) {
    correlation <- lavoptions$correlation
  }

  # what with fixed.x?
  # - does not really matter; fit will be saturated anyway
  # - fixed.x = TRUE may avoid convergence issues with non-numeric
  #             x-covariates
  fixed_x <- lavoptions$fixed.x

  # if multilevel
  if (nlevels > 1L) {
    # fixed.x       <- FALSE # for now
    if (!lavoptions$estimator %in% c("WLS", "DWLS", "ULS")) {
      categorical <- FALSE # for now
      conditional_x <- FALSE # for now
    }
    # else: two-level (D)WLS supports conditional.x (slopes per block)
  }

  lhs <- rhs <- op <- character(0)
  group <- block <- level <- free <- exo <- integer(0)
  ustart <- numeric(0)

  # block number
  b <- 0L
  for (g in 1:ngroups) {
    # only for multilevel
    if (nlevels > 1L) {
      # ylp <- lavsamplestats@YLp[[g]]
      lp <- lavdata@Lp[[g]]
    }

    # local copy
    sample_cov <- sample_cov_1[[g]]
    sample_mean <- sample_mean_1[[g]]
    sample_slopes <- sample_slopes_1[[g]]
    sample_th <- sample_th_1[[g]]
    sample_th_idx <- sample_th_idx_1[[g]]
    sample_cov_x <- sample_cov_x_1[[g]]
    sample_mean_x <- sample_mean_x_1[[g]]

    # force local sample.cov to be pd -- just for starting values anyway
    if (!is.null(sample_cov) && !anyNA(sample_cov)) {
      sample_cov <- lav_mat_sym_force_pd(sample_cov)
    }

    for (l in 1:nlevels) {
      # block
      b <- b + 1L

      # ov.names for this block
      if (is.null(lavpta)) { # only data was used
      if (nlevels > 1L) {
      ov_names <- lavdata@ov.names.l[[g]][[(ngroups - 1) * g + b]]
    } else {
          ov_names <- lavdata@ov.names[[g]]
    }
        ov_names_x <- lavdata@ov.names.x[[g]]
        ov_names_nox <- ov_names[!ov_names %in% ov_names_x]
      } else {
        if (conditional_x) {
          ov_names <- lavpta$vnames$ov.nox[[b]]
        } else {
          ov_names <- lavpta$vnames$ov[[b]]
        }
        ov_names_x <- lavpta$vnames$ov.x[[b]]
        ov_names_nox <- lavpta$vnames$ov.nox[[b]]
      }

      # only for multilevel, overwrite sample.cov and sample_mean
      if (nlevels > 1L) {
        if (independent) {
          ov_names_x <- lavpta$vnames$ov.x[[b]]
          ov_names_nox <- lavpta$vnames$ov.nox[[b]]
          if (!is.null(lavh1$implied$cov.x) &&
              !is.null(lavh1$implied$cov.x[[b]]) &&
              NROW(lavh1$implied$cov.x[[b]]) > 0L) {
            # conditional.x with y-only h1 cov blocks and separately
            # stored covariate moments (two-level (D)WLS)
            sample_cov_x <- lavh1$implied$cov.x[[b]]
            sample_mean_x <- lavh1$implied$mean.x[[b]]
          } else if (length(ov_names_x) > 0L) {
            # better use lavdata@Lp[[g]]$ov.x.idx??
            # in case we have x/y mismatch across levels?
            ov_x_idx <- lavpta$vidx$ov.x[[b]]
            sample_cov_x <- lavh1$implied$cov[[b]][ov_x_idx,
              ov_x_idx,
              drop = FALSE
            ]
            sample_mean_x <- lavh1$implied$mean[[b]][ov_x_idx]
          } else {
            sample_cov_x <- NULL
            sample_mean_x <- NULL
          }
        } else {
          ov_names_x <- character(0L)
          ov_names_nox <- ov_names
        }

        if (length(lavh1) > 0L) {
          sample_cov <- lavh1$implied$cov[[b]]
          sample_mean <- lavh1$implied$mean[[b]]
          if (conditional_x && !is.null(lavh1$implied$res.slopes) &&
              !is.null(lavh1$implied$res.slopes[[b]])) {
            sample_slopes <- lavh1$implied$res.slopes[[b]]
          }
        } else {
          sample_cov <- diag(length(ov_names))
          sample_mean <- numeric(length(ov_names))
        }

        # if(l == 1L) {
        #    sample.cov  <- YLp[[2]]$Sigma.W[block.idx, block.idx,
        #                                    drop = FALSE]
        #    sample_mean <- YLp[[2]]$Mu.W[block.idx]
        # } else {
        #    sample.cov  <- YLp[[2]]$Sigma.B[block.idx, block.idx,
        #                                    drop = FALSE]
        #    sample_mean <- YLp[[2]]$Mu.B[block.idx]
        # }

        # force local sample.cov to be strictly pd (and exaggerate)
        # just for starting values anyway, but at least the first
        # evaluation will be feasible
        sample_cov <- lav_mat_sym_force_pd(sample_cov,
          tol = 1e-03
        )
      }


      # a) VARIANCES (all ov's, if !conditional.x, also exo's)
      nvar <- length(ov_names)

      lhs <- c(lhs, ov_names)
      op <- c(op, rep("~~", nvar))
      rhs <- c(rhs, ov_names)
      block <- c(block, rep(b, nvar))
      group <- c(group, rep(g, nvar))
      level <- c(level, rep(l, nvar))
      if (correlation) {
        free <- c(free, rep(0L, nvar))
      } else {
        free <- c(free, rep(1L, nvar))
      }
      exo <- c(exo, rep(0L, nvar))

      # starting values -- variances
      if (correlation) {
        ustart <- c(ustart, rep(1, nvar))
      } else if (!is.null(sample_cov)) {
        ustart <- c(ustart, diag(sample_cov))
      } else {
        ustart <- c(ustart, rep(as.numeric(NA), nvar))
      }

      # COVARIANCES!
      if (!independent) {
        pstar <- nvar * (nvar - 1) / 2
        if (pstar > 0L) { # only if more than 1 variable
          tmp <- utils::combn(ov_names, 2)
          lhs <- c(lhs, tmp[1, ]) # to fill upper.tri
          op <- c(op, rep("~~", pstar))
          rhs <- c(rhs, tmp[2, ])
          block <- c(block, rep(b, pstar))
          group <- c(group, rep(g, pstar))
          level <- c(level, rep(l, pstar))
          free <- c(free, rep(1L, pstar))
          exo <- c(exo, rep(0L, pstar))
        }

        # starting values -- covariances
        if (!is.null(sample_cov)) {
          sample_cov_vech <- lav_mat_vech(sample_cov, diagonal = FALSE)
          ustart <- c(ustart, sample_cov_vech)
          # check for 'missing by design' cells: here, the sample.cov
          # element is *exactly* zero (new in 0.6-18)
          zero_cov <- which(sample_cov_vech == 0)
          if (length(zero_cov) > 0L && !is.null(lavh1)) {
            n_tmp <- length(free)
            ones_and_zeroes <- rep(1L, pstar)
            ones_and_zeroes[zero_cov] <- 0L
            free[(n_tmp - pstar + 1):n_tmp] <- ones_and_zeroes
          }
        } else {
          ustart <- c(ustart, rep(as.numeric(NA), pstar))
        }
      }

      # ordered? fix variances, add thresholds
      ord_names <- character(0L)
      if (categorical) {
        ord_names <- ov$name[ov$type == "ordered"]
        # only for this group
        ord_names <- ov_names[which(ov_names %in% ord_names)]

        if (length(ord_names) > 0L) {
          # fix variances to 1.0
          # (two-level, theta parameterization: at the WITHIN level only;
          #  the between-level variances of the ordinal variables are free)
          if (nlevels > 1L) {
            if (l == 1L) {
              idx <- which(lhs %in% ord_names & op == "~~" & lhs == rhs &
                           block == b)
              ustart[idx] <- 1.0
              free[idx] <- 0L
            }
          } else {
            idx <- which(lhs %in% ord_names & op == "~~" & lhs == rhs)
            ustart[idx] <- 1.0
            free[idx] <- 0L
          }

          # add thresholds
          # (two-level: FREE at the between level; fixed-to-zero
          #  placeholder rows at the within level -- the same convention
          #  as lavaanify() for the user model)
          lhs_th <- character(0)
          rhs_th <- character(0)
          for (o in ord_names) {
            nth <- ov$nlev[ov$name == o] - 1L
            if (nth < 1L) next
            lhs_th <- c(lhs_th, rep(o, nth))
            rhs_th <- c(rhs_th, paste("t", seq_len(nth), sep = ""))
          }
          nel <- length(lhs_th)
          lhs <- c(lhs, lhs_th)
          rhs <- c(rhs, rhs_th)
          op <- c(op, rep("|", nel))
          block <- c(block, rep(b, nel))
          group <- c(group, rep(g, nel))
          level <- c(level, rep(l, nel))
          exo <- c(exo, rep(0L, nel))
          if (nlevels == 1L || l == nlevels) {
            free <- c(free, rep(1L, nel))
            # starting values
            if (!is.null(sample_th) && !is.null(sample_th_idx)) {
              th_start <- sample_th[sample_th_idx > 0L]
              ustart <- c(ustart, th_start)
            } else {
              ustart <- c(ustart, rep(as.numeric(NA), nel))
            }
          } else {
            # within level: fixed-to-zero placeholders
            free <- c(free, rep(0L, nel))
            ustart <- c(ustart, rep(0, nel))
          } # thresholds

          # fixed-to-zero intercepts (since 0.5.17)
          ov_int <- ord_names
          nel <- length(ov_int)
          lhs <- c(lhs, ov_int)
          op <- c(op, rep("~1", nel))
          rhs <- c(rhs, rep("", nel))
          block <- c(block, rep(b, nel))
          group <- c(group, rep(g, nel))
          level <- c(level, rep(l, nel))
          free <- c(free, rep(0L, nel))
          exo <- c(exo, rep(0L, nel))
          ustart <- c(ustart, rep(0, nel))

          # ~*~ (since 0.6-1)
          nel <- length(ov_int)
          lhs <- c(lhs, ov_int)
          op <- c(op, rep("~*~", nel))
          rhs <- c(rhs, ov_int)
          block <- c(block, rep(b, nel))
          group <- c(group, rep(g, nel))
          level <- c(level, rep(l, nel))
          free <- c(free, rep(0L, nel))
          exo <- c(exo, rep(0L, nel))
          ustart <- c(ustart, rep(1, nel))
        }
      } # categorical

      # correlation structure?
      if (!categorical && correlation) {
        nel <- nvar
        lhs <- c(lhs, ov_names)
        op <- c(op, rep("~*~", nel))
        rhs <- c(rhs, ov_names)
        block <- c(block, rep(b, nel))
        group <- c(group, rep(g, nel))
        level <- c(level, rep(l, nel))
        free <- c(free, rep(0L, nel))
        exo <- c(exo, rep(0L, nel))
        ustart <- c(ustart, rep(1, nel))
      }

      # meanstructure?
      if (meanstructure) {
        # auto-remove ordinal variables
        ov_int <- ov_names
        idx <- which(ov_int %in% ord_names)
        if (length(idx)) ov_int <- ov_int[-idx]

        nel <- length(ov_int)
        lhs <- c(lhs, ov_int)
        op <- c(op, rep("~1", nel))
        rhs <- c(rhs, rep("", nel))
        block <- c(block, rep(b, nel))
        group <- c(group, rep(g, nel))
        level <- c(level, rep(l, nel))
        # if multilevel, level=1 has fixed zeroes
        if (nlevels > 1L && l == 1L) {
          within_1 <- rep(0L, nel)
          # within-only variables have a free mean at level 1; match by
          # NAME: ov_int may exclude ordinal (and exogenous) variables,
          # while within.idx refers to the full ('tilde') variable set
          # FIXME: assuming 1 group
          within_names <- lp$ov.names[lp$within.idx[[2]]]
          within_1[ov_int %in% within_names] <- 1L
          free <- c(free, within_1)
        } else {
          free <- c(free, rep(1L, nel))
        }
        exo <- c(exo, rep(0L, nel))

        # starting values
        if (!is.null(sample_mean)) {
          sample_int_idx <- match(ov_int, ov_names)
          ustart <- c(ustart, sample_mean[sample_int_idx])
        } else {
          ustart <- c(ustart, rep(as.numeric(NA), length(ov_int)))
        }
      }


      # fixed.x exogenous variables?
      if (!conditional_x && (nx <- length(ov_names_x)) > 0L) {
        if (independent && lavoptions$baseline.fixed.x.free.cov) {
          # add covariances for eXo
          pstar <- nx * (nx - 1) / 2
          if (pstar > 0L) { # only if more than 1 variable
            tmp <- utils::combn(ov_names_x, 2)
            lhs <- c(lhs, tmp[1, ]) # to fill upper.tri
            op <- c(op, rep("~~", pstar))
            rhs <- c(rhs, tmp[2, ])
            block <- c(block, rep(b, pstar))
            group <- c(group, rep(g, pstar))
            level <- c(level, rep(l, pstar))
            free <- c(free, rep(1L, pstar))
            exo <- c(exo, rep(0L, pstar))

            # starting values
            if (!is.null(sample_cov_x)) {
              rhs_idx <- match(tmp[1, ], ov_names_x)
              lhs_idx <- match(tmp[2, ], ov_names_x)
              ustart <- c(
                ustart,
                sample_cov_x[cbind(rhs_idx, lhs_idx)]
              )
            } else {
              ustart <- c(ustart, rep(as.numeric(NA), pstar))
            }
          }
        }

        if (fixed_x) {
          # fix variances/covariances
          exo_idx <- which(rhs %in% ov_names_x &
            lhs %in% ov_names_x &
            op == "~~" & group == g) # ok
          exo[exo_idx] <- 1L
          free[exo_idx] <- 0L

          # fix means
          exo_idx <- which(lhs %in% ov_names_x &
            op == "~1" & group == g) # ok
          exo[exo_idx] <- 1L
          free[exo_idx] <- 0L
        }
      }

      # conditional.x?
      if (conditional_x && (nx <- length(ov_names_x)) > 0L) {
        # eXo variances
        nel <- length(ov_names_x)
        lhs <- c(lhs, ov_names_x)
        op <- c(op, rep("~~", nel))
        rhs <- c(rhs, ov_names_x)
        block <- c(block, rep(b, nel))
        group <- c(group, rep(g, nel))
        level <- c(level, rep(l, nel))
        if (fixed_x) {
          free <- c(free, rep(0L, nel))
          exo <- c(exo, rep(1L, nel))
        } else {
          free <- c(free, rep(1L, nel))
          exo <- c(exo, rep(0L, nel))
        }

        # starting values
        if (!is.null(sample_cov_x)) {
          ustart <- c(ustart, diag(sample_cov_x))
        } else {
          ustart <- c(ustart, rep(as.numeric(NA), nel))
        }


        # eXo covariances
        pstar <- nx * (nx - 1) / 2
        if (pstar > 0L) { # only if more than 1 variable
          tmp <- utils::combn(ov_names_x, 2)
          lhs <- c(lhs, tmp[1, ]) # to fill upper.tri
          op <- c(op, rep("~~", pstar))
          rhs <- c(rhs, tmp[2, ])
          block <- c(block, rep(b, pstar))
          group <- c(group, rep(g, pstar))
          level <- c(level, rep(l, pstar))
          if (fixed_x) {
            free <- c(free, rep(0L, pstar))
            exo <- c(exo, rep(1L, pstar))
          } else {
            free <- c(free, rep(1L, pstar))
            exo <- c(exo, rep(0L, pstar))
          }

          # starting values
          if (!is.null(sample_cov_x)) {
            rhs_idx <- match(tmp[1, ], ov_names_x)
            lhs_idx <- match(tmp[2, ], ov_names_x)
            ustart <- c(
              ustart,
              sample_cov_x[cbind(rhs_idx, lhs_idx)]
            )
          } else {
            ustart <- c(ustart, rep(as.numeric(NA), pstar))
          }
        }

        # eXo means
        if (meanstructure) {
          ov_int <- ov_names_x

          nel <- length(ov_int)
          lhs <- c(lhs, ov_int)
          op <- c(op, rep("~1", nel))
          rhs <- c(rhs, rep("", nel))
          group <- c(group, rep(g, nel))
          block <- c(block, rep(b, nel))
          level <- c(level, rep(l, nel))
          if (fixed_x) {
            free <- c(free, rep(0L, nel))
            exo <- c(exo, rep(1L, nel))
          } else {
            free <- c(free, rep(1L, nel))
            exo <- c(exo, rep(0L, nel))
          }

          # starting values
          if (!is.null(sample_mean_x)) {
            sample_int_idx <- match(ov_int, ov_names_x)
            ustart <- c(ustart, sample_mean_x[sample_int_idx])
          } else {
            ustart <- c(ustart, rep(as.numeric(NA), length(ov_int)))
          }
        }

        # slopes
        nnox <- length(ov_names_nox)
        nel <- nnox * nx

        lhs <- c(lhs, rep(ov_names_nox, times = nx))
        op <- c(op, rep("~", nel))
        rhs <- c(rhs, rep(ov_names_x, each = nnox))
        block <- c(block, rep(b, nel))
        group <- c(group, rep(g, nel))
        level <- c(level, rep(l, nel))
        if (independent) {
          if (lavoptions$baseline.conditional.x.free.slopes) {
            free <- c(free, rep(1L, nel))
          } else {
            free <- c(free, rep(0L, nel))
          }
        } else {
          free <- c(free, rep(1L, nel))
        }
        exo <- c(exo, rep(1L, nel))

        # starting values -- slopes
        if (independent) {
          # FIXME: zero slope-structure provides a fit that
          # is equal to the conditional.x = FALSE version;
          # in principle, we could just fix the slope-structure
          # to the sample-based slopes

          # to get the old behaviour:
          if (!lavoptions$baseline.conditional.x.free.slopes) {
            ustart <- c(ustart, rep(0, nel))
          } else {
            # but we probably should do:
            ustart <- c(ustart, lav_mat_vec(sample_slopes))
          }
        } else if (!is.null(sample_slopes)) {
          ustart <- c(ustart, lav_mat_vec(sample_slopes))
        } else {
          ustart <- c(ustart, rep(as.numeric(NA), nel))
        }
      } # conditional.x

      # group.w.free (new in 0.6-8)
      if (group_w_free) {
        lhs <- c(lhs, "group")
        op <- c(op, "%")
        rhs <- c(rhs, "w")
        block <- c(block, b)
        group <- c(group, g)
        level <- c(level, l)
        free <- c(free, 1L)
        exo <- c(exo, 0L)
        ustart <- c(ustart, lavsamplestats@WLS.obs[[g]][1])
      }
    } # levels
  } # ngroups

  # free counter
  idx_free <- which(free > 0)
  free[idx_free] <- seq_along(idx_free)

  list_1 <- list(
    id = seq_along(lhs),
    lhs = lhs,
    op = op,
    rhs = rhs,
    user = rep(1L, length(lhs)),
    block = block,
    group = group,
    level = level,
    free = free,
    ustart = ustart,
    exo = exo # ,
    # label       = rep("",  length(lhs))
    # eq.id       = rep(0L,  length(lhs)),
    # unco        = free
  )


  # keep level column if no levels? (no for now)
  if (nlevels < 2L) {
    list_1$level <- NULL
  }

  list_1
}

# - currently only used for continuous twolevel data
# - conditional.x not supported (yet)
lav_pt_unrestricted_chol <- function(lavobject = NULL,
                                           # if no object is available,
                                           lavdata = NULL,
                                           lavpta = NULL, # optional
                                           lavoptions = NULL,
                                           group = NULL) {
  # grab everything from lavaan lavobject
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))

    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
    # lavsamplestats <- lavobject@SampleStats
    lavpta <- lavobject@pta
    # lavh1 <- lavobject@h1
  }

  ov <- lavdata@ov
  meanstructure <- lavoptions$meanstructure
  categorical <- any(ov$type == "ordered")
  if (categorical) {
    lav_msg_stop(gettext("categorical data not supported in this function"))
  }
  ngroups <- lavdata@ngroups
  nlevels <- lavdata@nlevels

  # select groups
  if (is.null(group)) {
    group_select <- seq_len(ngroups)
  } else {
    stopifnot(is.numeric(group), all(group <= ngroups))
    group_select <- group
    if (length(group_select) == 0L) {
      lav_msg_stop(gettext("no groups selected"))
    }
  }

  # what with fixed.x?
  # - does not really matter; fit will be saturated anyway
  # - fixed.x = TRUE may avoid convergence issues with non-numeric
  #             x-covariates
  # fixed_x <- lavoptions$fixed.x

  # if multilevel
  if (nlevels > 1L) {
    # fixed.x       <- FALSE # for now
    # conditional_x <- FALSE # for now
    categorical <- FALSE # for now
  }

  lhs <- rhs <- op <- character(0)
  group <- block <- level <- free <- exo <- integer(0)
  ustart <- lower <- numeric(0)

  # block number
  b <- 0L
  for (g in 1:ngroups) {

    # select group?
    if (! g %in% group_select) {
      next
    }

    # only for multilevel
    if (nlevels > 1L) {
      lp <- lavdata@Lp[[g]]
    }

    for (l in 1:nlevels) {
      # block
      b <- b + 1L

      if (is.null(lavpta)) {
        ov_names <- lavdata@ov.names[[b]]
        # ov_names_x <- lavdata@ov.names.x[[b]]
        # ov_names_nox <- ov_names[!ov_names %in% ov_names_x]
      } else {
        ov_names <- lavpta$vnames$ov[[b]]
        # ov_names_x <- lavpta$vnames$ov.x[[b]]
        # ov_names_nox <- lavpta$vnames$ov.nox[[b]]
      }

      # only for multilevel, overwrite sample.cov and sample_mean
      if (nlevels > 1L) {
        # ov_names_x <- character(0L)
        # ov_names_nox <- ov_names
      }

      # create lv.names == ov.names
      lv_names <- paste("f", ov_names, sep = "")

      # a) OV VARIANCES -> fixed to zero
      nvar <- length(ov_names)
      lhs <- c(lhs, ov_names)
      op <- c(op, rep("~~", nvar))
      rhs <- c(rhs, ov_names)
      block <- c(block, rep(b, nvar))
      group <- c(group, rep(g, nvar))
      level <- c(level, rep(l, nvar))
      ustart <- c(ustart, rep(0.0001, nvar)) ### Force PD!! (option?)
      free <- c(free, rep(0L, nvar))
      exo <- c(exo, rep(0L, nvar))
      lower <- c(lower, rep(0.0, nvar))

      # b) LV VARIANCES -> fixed to 1.0
      nvar <- length(lv_names)
      lhs <- c(lhs, lv_names)
      op <- c(op, rep("~~", nvar))
      rhs <- c(rhs, lv_names)
      block <- c(block, rep(b, nvar))
      group <- c(group, rep(g, nvar))
      level <- c(level, rep(l, nvar))
      ustart <- c(ustart, rep(1.0, nvar))
      free <- c(free, rep(0L, nvar))
      exo <- c(exo, rep(0L, nvar))
      lower <- c(lower, rep(1.0, nvar))

      # c) LOADINGS self
      nvar <- length(ov_names)
      lhs <- c(lhs, lv_names)
      op <- c(op, rep("=~", nvar))
      rhs <- c(rhs, ov_names)
      block <- c(block, rep(b, nvar))
      group <- c(group, rep(g, nvar))
      level <- c(level, rep(l, nvar))
      ustart <- c(ustart, rep(as.numeric(NA), nvar))
      free <- c(free, rep(1L, nvar))
      exo <- c(exo, rep(0L, nvar))
      lower <- c(lower, rep(0.0, nvar)) # lower bound!

      # d) LOADINGS other
      if (length(ov_names) > 1L) {
        tmp <- utils::combn(ov_names, 2)
        pstar <- ncol(tmp)
        lhs <- c(lhs, paste("f", tmp[1, ], sep = ""))
        op <- c(op, rep("=~", pstar))
        rhs <- c(rhs, tmp[2, ])
        block <- c(block, rep(b, pstar))
        group <- c(group, rep(g, pstar))
        level <- c(level, rep(l, pstar))
        free <- c(free, rep(1L, pstar))
        exo <- c(exo, rep(0L, pstar))
        lower <- c(lower, rep(-Inf, pstar))
        ustart <- c(ustart, rep(as.numeric(NA), pstar))
      }

      # check for zero coverage at level 1 (new in 0.6-18)
      if (lavdata@missing %in% c("ml", "ml.x") && l == 1 &&
          !is.null(lavdata@Mp[[g]])) {
        coverage <- lavdata@Mp[[g]]$coverage
        sample_cov_vech <- lav_mat_vech(coverage, diagonal = FALSE)
        zero_cov <- which(sample_cov_vech == 0)
        if (length(zero_cov) > 0L) {
          n_tmp <- length(free)
          ones_and_zeroes <- rep(1L, pstar)
          ones_and_zeroes[zero_cov] <- 0L
      inf_and_zeroes <- rep(-Inf, pstar)
      inf_and_zeroes[zero_cov] <- 0
      na_and_zeroes <- rep(as.numeric(NA), pstar)
      na_and_zeroes[zero_cov] <- 0
          free[(n_tmp - pstar + 1):n_tmp] <- ones_and_zeroes
      ustart[(n_tmp - pstar + 1):n_tmp] <- na_and_zeroes
      lower[(n_tmp - pstar + 1):n_tmp] <- inf_and_zeroes
        }
      }

      # meanstructure?
      if (meanstructure) {
        # OV
        ov_int <- ov_names

        nel <- length(ov_int)
        lhs <- c(lhs, ov_int)
        op <- c(op, rep("~1", nel))
        rhs <- c(rhs, rep("", nel))
        block <- c(block, rep(b, nel))
        group <- c(group, rep(g, nel))
        level <- c(level, rep(l, nel))
        # if multilevel, level=1 has fixed zeroes
        if (nlevels > 1L && l == 1L) {
          within_1 <- rep(0L, nel)
          within_idx <- match(lp$within.idx[[2]], lp$ov.idx[[1]])
          within_1[within_idx] <- 1L
          free <- c(free, within_1)
        } else {
          free <- c(free, rep(1L, nel))
        }
        exo <- c(exo, rep(0L, nel))
        lower <- c(lower, rep(-Inf, nel))
        ustart <- c(ustart, rep(as.numeric(NA), nel))

        # LV
        ov_int <- lv_names

        nel <- length(ov_int)
        lhs <- c(lhs, ov_int)
        op <- c(op, rep("~1", nel))
        rhs <- c(rhs, rep("", nel))
        block <- c(block, rep(b, nel))
        group <- c(group, rep(g, nel))
        level <- c(level, rep(l, nel))
        free <- c(free, rep(0L, nel))
        exo <- c(exo, rep(0L, nel))
        ustart <- c(ustart, rep(0.0, nel))
        lower <- c(lower, rep(-Inf, nel))
      }
    } # levels
  } # ngroups

  # free counter
  idx_free <- which(free > 0)
  free[idx_free] <- seq_along(idx_free)

  list_1 <- list(
    id = seq_along(lhs),
    lhs = lhs,
    op = op,
    rhs = rhs,
    user = rep(1L, length(lhs)),
    block = block,
    group = group,
    level = level,
    free = free,
    ustart = ustart,
    exo = exo,
    lower = lower # ,
    # label       = rep("",  length(lhs))
    # eq.id       = rep(0L,  length(lhs)),
    # unco        = free
  )


  # keep level column if no levels? (no for now)
  if (nlevels < 2L) {
    list_1$level <- NULL
  }

  list_1
}

# create a 'baseline' model from an existing object/partable
# - fix all (free) directed effects (factor loadings, regressions) to zero
# - fix all (free) covariances to zero
# - neutralize the (measured) latent variables by fixing their variances
#   to zero
# - keep all other (relevant) constraints
# - this *should* result in a baseline model that is always nested within
#   the original model
lav_pt_baseline <- function(lavobject = NULL,
                                  # if no object is available,
                                  lavpartable = NULL,
                                  lavh1 = NULL) {

  # grab everything from lavaan lavobject
  if (!is.null(lavobject)) {
    stopifnot(inherits(lavobject, "lavaan"))
    lavpartable <- lavobject@ParTable
    lavh1 <- lavobject@h1
  }

  # number of blocks
  nblocks <- lav_pt_nblocks(lavpartable)

  # conditional.x ? check res.cov[[1]] slot
  conditional_x <- FALSE
  if (!is.null(lavh1) && !is.null(lavh1$implied$res.cov[[1]])) {
    conditional_x <- TRUE
  }

  # get sample statistics, all groups
  if (conditional_x) {
    sample_cov <- lavh1$implied$res.cov
    # sample_mean <- lavh1$implied$res.int
  } else {
    sample_cov <- lavh1$implied$cov
    # sample_mean <- lavh1$implied$mean
  }

  # shortcut of lavpartable
  pt_1 <- lavpartable

  # keep exo=1 starting values
  exo_idx <- which(pt_1$exo == 1L)
  if (!is.null(pt_1$est)) {
    pt_1$ustart[exo_idx] <- pt_1$est[exo_idx]
  } else if (!is.null(pt_1$start)) {
    pt_1$iustart[exo_idx] <- pt_1$start[exo_idx]
  }

  # remove est/se columns, if present, but keep start column
  pt_1$est <- pt_1$se <- NULL

  # lv.names
  lv_names <- lav_pt_vnames(pt_1, "lv")

  # zero-out all directed effects and covariances
  directed_idx <- which(pt_1$op %in% c("=~", "~") & pt_1$free > 0L &
                                                            pt_1$exo != 1L)
  cov_idx <- which(pt_1$op == "~~" & pt_1$lhs != pt_1$rhs & pt_1$free > 0L)

  # neutralize latent variables
  lv_var_idx <- which(pt_1$op == "~~" & pt_1$lhs %in% lv_names &
                                        pt_1$lhs == pt_1$rhs & pt_1$free > 0L)
  #lv.int.idx <- which(PT$op == "~1" & PT$lhs %in% lv.names & PT$free > 0L)
  #zero.idx <- c(directed.idx, cov.idx, lv.var.idx, lv.int.idx)
  zero_idx <- c(directed_idx, cov_idx, lv_var_idx)

  pt_1$free[zero_idx] <- 0L
  pt_1$start[zero_idx] <- 0
  pt_1$ustart[zero_idx] <- 0

  # Question: fill in more elements in PT$start? (only ov variances for now)
  for (b in seq_len(nblocks)) {
    ov_names_num <- lav_pt_vnames(lavpartable, "ov.num", block = b)
    if (conditional_x) {
      ov_names_x <- lav_pt_vnames(lavpartable, "ov.x", block = b)
      ov_names_num <- ov_names_num[!ov_names_num %in% ov_names_x]
    }
    ovar_idx <- which(lavpartable$block == b &
      lavpartable$op == "~~" &
      lavpartable$lhs %in% ov_names_num &
      lavpartable$lhs == lavpartable$rhs)
    sample_var_idx <- match(lavpartable$lhs[ovar_idx], ov_names_num)
    pt_1$ustart[ovar_idx] <- diag(sample_cov[[b]])[sample_var_idx]
  }

  pt_1 <- lav_pt_complete(pt_1)
  pt_1
}

lav_step13_vcov_boot <- function(lavoptions = NULL,
                                        lavmodel = NULL,
                                        lavsamplestats = NULL,
                                        lavdata = NULL,
                                        lavpartable = NULL,
                                        lavcache = NULL,
                                        lavimplied = NULL,
                                        lavh1 = NULL,
                                        x = NULL) {
  # # # # # # # # # # # # # # # #
  # #  13. lavvcov + lavboot # #
  # # # # # # # # # # # # # # # #

  # note: for models with (rotated) exploratory factors (EFA/ESEM) and
  # se = "bootstrap", the bootstrap is deferred to step 16 (lav_step16_rotation)
  # where the rotated parameter table exists, so that each bootstrap sample can
  # be refit, rotated, and aligned to the original solution (see github #376).
  # The condition below therefore skips the (analytic) VCOV here for that case.

  # set VCOV to NULL
  # if lavoptions$se not "none", "external", "twostep" and
  #   lavmodel@nx.free > 0L and x converged or optim.method == "none"
  #   compute VCOV via lav_model_vcov
  # if attribute BOOT.COEFF of VCOV not NULL, store it in lavboot$coef
  # lavvcov <- list(se = lavoptions$se, information = lavoptions$information,
  #   vcov = VCOV1)
  #   where VCOV1 = VCOV without attributes (except dim) or ...
  #                 NULL if lavoptions$store.vcov FALSE or
  #                 store.vcov=="default" and rotation="none"
  # if lavoptions$se == "external"
  #   if lavpartable$se NULL
  #     lavpartable$se <- lav_model_vcov_se(..., vcov=NULL, boot=NULL)
  #       + ** warning **
  # if lavpartable not "external" or "none" or "twostep"
  #     lavpartable$se <- lav_model_vcov_se(...)

  vcov_1 <- NULL
  if (lavoptions$se != "none" && lavoptions$se != "external" &&
    lavoptions$se != "twostep" &&
    (lavmodel@nefa == 0L ||
     (lavmodel@nefa > 0L && lavoptions$rotation == "none") ||
     (lavmodel@nefa > 0L && lavoptions$rotation.se == "delta")
    ) &&
    lavmodel@nx.free > 0L && (attr(x, "converged") ||
    lavoptions$optim.method == "none")) {
    if (lav_verbose()) {
      cat("computing VCOV for      se =", lavoptions$se, "...")
    }
    # special case: estimator = "IV" (analytic two-stage standard errors);
    # se = "bootstrap" falls through to the generic bootstrap machinery below
    if (lavoptions$estimator %in% "IV" && lavoptions$se != "bootstrap" &&
        !is.null(attr(x, "eqs"))) {
      vcov_1 <- lav_sem_miiv_vcov(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavoptions = lavoptions,
        lavpartable = lavpartable,
        lavimplied = lavimplied,
        lavh1 = lavh1,
        lavdata = lavdata,
        eqs = attr(x, "eqs")
      )
    } else {
      # everything else:
      vcov_1 <- lav_model_vcov(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavoptions = lavoptions,
        lavdata = lavdata,
        lavpartable = lavpartable,
        lavcache = lavcache,
        lavimplied = lavimplied,
        lavh1 = lavh1
      )
    }
    if (lav_verbose()) {
      cat(" done.\n")
    }
  } # VCOV

  # extract bootstrap results (if any)
  if (!is.null(attr(vcov_1, "BOOT.COEF"))) {
    lavboot <- list()
    lavboot$coef <- attr(vcov_1, "BOOT.COEF")
  } else {
    lavboot <- list()
  }

  # draw Monte Carlo samples (if requested) for defined parameters
  # only when bootstrap is not used (bootstrap covers def CIs already)
  lav_monte_carlo <- list()
  def_idx_count <- sum(lavpartable$op == ":=")
  if (def_idx_count > 0L && lavoptions$se != "bootstrap" &&
    lav_model_vcov_se_mc_active(lavoptions) &&
    !is.null(vcov_1) && !inherits(vcov_1, "try-error")) {
    if (lav_verbose()) {
      cat("drawing", lavoptions$monte.carlo$R,
          "Monte Carlo samples for defined parameters ...")
    }
    mc_coef <- lav_model_vcov_mc(lavmodel = lavmodel, vcov = vcov_1,
                                 lavoptions = lavoptions)
    lav_monte_carlo$coef <- mc_coef
    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  # store VCOV in vcov
  # strip all attributes but 'dim'
  tmp_attr <- attributes(vcov_1)
  vcov1 <- vcov_1
  attributes(vcov1) <- tmp_attr["dim"]
  # store vcov? new in 0.6-6
  if (!is.null(lavoptions$store.vcov) && !is.null(vcov1)) {
    if (is.logical(lavoptions$store.vcov) && !lavoptions$store.vcov) {
      vcov1 <- NULL
    }
    if (is.character(lavoptions$store.vcov) &&
      lavoptions$rotation == "none" &&
      lavoptions$store.vcov == "default" &&
      ncol(vcov1) > 200L) {
      vcov1 <- NULL
    }
  }
  lavvcov <- list(
    se = lavoptions$se,
    information = lavoptions$information[1],
    vcov = vcov1
  )

  # store se in partable
  if (lavoptions$se == "external") {
    if (is.null(lavpartable$se)) {
      lavpartable$se <- lav_model_vcov_se(
        lavmodel = lavmodel,
        lavpartable = lavpartable,
        vcov = NULL, boot = NULL,
        lavoptions = lavoptions
      )
      lav_msg_warn(gettext(
        "se = \"external\" but parameter table does not contain a `se' column"))
    }
  } else if (lavoptions$se %in% c("none", "twostep")) {
    # do nothing
  } else {
    lavpartable$se <- lav_model_vcov_se(
      lavmodel = lavmodel,
      lavpartable = lavpartable,
      vcov = vcov_1,
      boot = lavboot$coef,
      mc = lav_monte_carlo$coef,
      lavoptions = lavoptions
    )
  }

  list(
    lavpartable      = lavpartable,
    lavvcov          = lavvcov,
    VCOV             = vcov_1,
    lavmodel         = lavmodel,
    lavboot          = lavboot,
    lav_monte_carlo  = lav_monte_carlo
  )
}

# utility function related to EFA

# generate 'efa' syntax for a single block of factors
# generate EFA model syntax
#
# 'nfactors' is a per-block vector of (integer) factor counts; the blocks are
# ordered group-major, level-minor, ie block = (g - 1L) * nlevels + level.
# When there is a single block (ngroups == 1 and nlevels == 1), the output is
# just the '=~' lines (no 'group:'/'level:' header), identical to the original
# single-block syntax. With multiple levels and/or groups, the blocks are
# wrapped in 'level:' and/or 'group:' sections.
lav_syntax_efa <- function(ov_names = NULL, nfactors = 1L,
                           nlevels = 1L, ngroups = 1L) {
  # 'twolevel = TRUE' was the old (pre per-block) calling convention
  if (isTRUE(nlevels)) {
    nlevels <- 2L
  }
  nblocks <- ngroups * nlevels

  # recycle a single factor count to all blocks
  if (length(nfactors) == 1L) {
    nfactors <- rep.int(nfactors, nblocks)
  }
  stopifnot(length(nfactors) == nblocks)

  # the '=~' lines for a single block with 'k' factors
  block_lines <- function(k) {
    vapply(seq_len(k), function(f) {
      paste0('efa("efa")*f', f, " =~ ", paste(ov_names, collapse = " + "))
    }, character(1L))
  }

  # single block: no header (backward compatible)
  if (nblocks == 1L) {
    return(block_lines(nfactors[1L]))
  }

  model <- character(0L)
  for (g in seq_len(ngroups)) {
    if (ngroups > 1L) {
      model <- c(model, paste("group:", g))
    }
    for (l in seq_len(nlevels)) {
      block <- (g - 1L) * nlevels + l
      if (nlevels > 1L) {
        model <- c(model, paste("level:", l))
      }
      model <- c(model, block_lines(nfactors[block]))
    }
  }

  model
}

# extract *standardized* loadings from efaList
lav_efa_get_loadings <- function(object, ...) {
  # kill object$loadings if present
  object[["loadings"]] <- NULL

  out <- lapply(object, function(x) {
    std <- lavTech(x, "std",
      add.class = TRUE, add.labels = TRUE,
      list.by.group = FALSE
    )
    lambda_idx <- which(names(std) == "lambda")
    mm_lambda <- std[lambda_idx]
    names(mm_lambda) <- NULL
    # if only single block, drop list
    if (length(mm_lambda) == 1L) {
      mm_lambda <- mm_lambda[[1]]
    } else {
      names(mm_lambda) <- x@Data@block.label
    }
    mm_lambda
  })

  # drop list if only a single model
  if (length(out) == 1L) {
    out <- out[[1]]
  }

  out
}

# determine the column order for a (rotated or unrotated) loading matrix,
# according to 'order_lv_by': "sumofsquares", "index" (Asparouhov & Muthen
# 2009, Appendix D) or "none". Returns an integer permutation vector.
lav_efa_order_idx <- function(mm_lambda = NULL, order_lv_by = "index") {
  if (order_lv_by == "sumofsquares") {
    l2 <- mm_lambda * mm_lambda
    order_idx <- base::order(colSums(l2), decreasing = TRUE)
  } else if (order_lv_by == "index") {
    # reorder using Asparouhov & Muthen 2009 criterion (see Appendix D)
    max_loading <- apply(abs(mm_lambda), 2, max)
    # 1: per factor, number of the loadings that are at least 0.8 of the
    #    highest loading of the factor
    # 2: mean of the index numbers
    average_index <- sapply(seq_len(ncol(mm_lambda)), function(i) {
      mean(which(abs(mm_lambda[, i]) >= 0.8 * max_loading[i]))
    })
    # order of the factors
    order_idx <- base::order(average_index)
  } else if (order_lv_by == "none") {
    order_idx <- seq_len(ncol(mm_lambda))
  } else {
    lav_msg_stop(gettext("order must be index, sumofsquares or none"))
  }
  order_idx
}

# reflect the columns of a loading matrix so that each column sum is
# non-negative; returns the (possibly sign-flipped) matrix.
lav_efa_reflect_cols <- function(mm_lambda = NULL) {
  neg_idx <- which(colSums(mm_lambda) < 0)
  if (length(neg_idx) > 0L) {
    mm_lambda[, neg_idx] <- -1 * mm_lambda[, neg_idx, drop = FALSE]
  }
  mm_lambda
}

# optionally reflect, then reorder the columns of a loading matrix. This is the
# common post-processing applied by the EFA extraction/PACE routines (the
# rotation engine applies the same ordering, but also needs to permute ROT/PHI,
# so it calls lav_efa_order_idx() directly).
lav_efa_reflect_and_reorder <- function(mm_lambda = NULL, reflect = TRUE,
                                        order_lv_by = "index") {
  if (reflect) {
    mm_lambda <- lav_efa_reflect_cols(mm_lambda)
  }
  order_idx <- lav_efa_order_idx(mm_lambda, order_lv_by)
  mm_lambda[, order_idx, drop = FALSE]
}

# find the best ordering of the columns in lambda_target, to minimize
# the difference with the reference lambda matrix (lambda_ref)
#
# enumerate all m! permutations of seq_len(m), one per row. The row order is the
# one produced by expand.grid(); both callers below rely on it for tie-breaking
# (they keep the first optimum), so do not change it.
lav_efa_permutations <- function(m = 1L) {
  tmp <- unname(as.matrix(expand.grid(rep(list(seq_len(m)), m))))
  # keep only rows where all numbers appear (i.e. actual permutations)
  tmp[apply(tmp, 1L, function(x) length(unique(x)) == m), , drop = FALSE]
}

# return the optimal order of the column for lambda_target
lav_efa_find_best_order <- function(lambda_ref = NULL, lambda_target = NULL,
                                    crit = "rmse") {
  m <- ncol(lambda_ref)
  stopifnot(ncol(lambda_target) == m)

  # all possible permutations
  perm <- lav_efa_permutations(m)

  rmse_perm <- numeric(nrow(perm))
  for (p in seq_len(nrow(perm))) {
    diff <- lambda_ref - lambda_target[, perm[p, ], drop = FALSE]
    diff2 <- diff * diff
    mse <- mean(diff2)
    rmse_perm[p] <- sqrt(mse)
  }
  best_idx <- which.min(rmse_perm)

  # return 'best' permutation
  perm[best_idx, ]
}

# find the signed permutation matrix T (m x m) such that 'lambda %*% T' best
# matches the reference loading matrix 'lambda_ref' (in a least-squares sense).
#
# this resolves the rotational indeterminacy of an EFA solution relative to a
# reference solution: factors may be re-ordered (column permutation) and/or
# reflected (sign change). It is used to align the rotated loadings of each
# bootstrap sample to the original (full-sample) rotated solution, so that the
# bootstrap distribution of the rotated parameters is meaningful.
#
# T is an orthonormal signed permutation matrix (T^-1 == t(T)), with a single
# non-zero (+1 or -1) entry per row/column. Applying it as an extra 'rotation'
# transforms all factor-related model matrices consistently (see
# lav_efa_bootstrap_align_coef()).
lav_efa_align_signed_perm <- function(lambda = NULL, lambda_ref = NULL) {
  m <- ncol(lambda)
  stopifnot(ncol(lambda_ref) == m)

  # single factor: only a possible sign flip
  if (m < 2L) {
    s <- if (sum(lambda * lambda_ref) < 0) -1 else 1
    return(matrix(s, 1L, 1L))
  }

  # all possible column permutations
  perm <- lav_efa_permutations(m)

  best_val <- Inf
  best_t <- diag(m)
  for (p in seq_len(nrow(perm))) {
    this_perm <- perm[p, ]
    lambda_p <- lambda[, this_perm, drop = FALSE]
    # optimal per-column sign: align each column with its reference column
    s <- sign(colSums(lambda_p * lambda_ref))
    s[s == 0] <- 1
    lambda_ps <- sweep(lambda_p, 2L, s, "*")
    val <- sum((lambda_ps - lambda_ref)^2)
    if (val < best_val) {
      best_val <- val
      tt <- matrix(0, m, m)
      for (j in seq_len(m)) {
        tt[this_perm[j], j] <- s[j]
      }
      best_t <- tt
    }
  }

  best_t
}

# stage-wise estimation of the unrestricted TWO-LEVEL model for mixed
# continuous/ordinal data, and the asymptotic covariance matrix of the
# estimates (the two-level analogue of muthen1984())
#
# YR 2026 (two-level WLS, phase 4; covariates added in phase 7)
#
# stage 1: univariate two-level models per variable
#            continuous: (mu, vw, vb[, slopes]) -- lav_uvreg_2l_fit()
#            ordinal:    (tau_1..K, vb[, slopes]) -- lav_uvord_2l_fit()
#          (ordinal: within latent variance fixed to 1, no mean)
#          exogenous covariates (conditional.x): within-level covariates
#          x_w enter every regression; between-level covariates x_b only
#          the regressions of variables with a between part
# stage 2: bivariate two-level models per pair, univariate parameters
#          (incl. the slopes) fixed, estimating theta = (cw, cb) -- the
#          within and between RESIDUAL covariance -- jointly:
#            cont/cont: lav_bvreg_2l_cov_twostep_fit()
#            cont/ord:  lav_bvmix_2l_cov_twostep_fit()
#            ord/ord:   lav_bvord_2l_cor_twostep_fit()
# stage 3: sandwich ACOV of all estimates, with the CLUSTERS as the
#          independent units:
#
#              WLS.W = B^{-1} INNER B^{-T}
#
#          B = | A11   0  |  A11: block-diagonal univariate information
#              | A21  A22 |  A21: d(pair scores)/d(univariate params)
#                            A22: block-diagonal (2 x 2) pair information
#
#          all blocks estimated from the cluster-wise scores via the
#          information identities (cfr. lav_m84_* for the single-level
#          case); INNER = crossprod of the stacked cluster-wise scores.
#
# The parameter (statistic) vector s is ordered per variable first
# (mu/vw/vb or tau_1..K/vb, followed by the within + between slopes),
# then per pair (cw, cb), pairs in column-major lower-triangular order
# ((2,1), (3,1), ..., (3,2), ...). The mapping to the lavaan WLS.obs
# layout is done by the caller.

lav_muthen2007 <- function(y = NULL,
                           ov_names = NULL,
                           ov_types = NULL,
                           cluster_idx = NULL,
                           between_flag = NULL, # per variable: does it
                                                # have a between part?
                           x_w = NULL, # within-level covariates
                           x_b = NULL, # between-level covariates
                           xw_names = NULL,
                           xb_names = NULL,
                           ngh = 21L, # stage 1 + mixed pairs
                           ngh2 = 13L, # ordinal pairs (per dimension)
                           wls_w = TRUE) {
  y <- as.matrix(y)
  nvar <- NCOL(y)
  if (is.null(ov_names)) {
    ov_names <- colnames(y)
    if (is.null(ov_names)) {
      ov_names <- paste0("y", seq_len(nvar))
    }
  }
  ord_flag <- (ov_types == "ordered")
  if (is.null(between_flag)) {
    between_flag <- rep(TRUE, nvar)
  }
  if (any(ord_flag & !between_flag)) {
    lav_msg_stop(gettext(
      "within-only ORDINAL variables are not supported (yet) in the
      two-level stage-wise estimation."))
  }

  # covariates
  nexo_w <- if (is.null(x_w)) 0L else NCOL(x_w)
  nexo_b <- if (is.null(x_b)) 0L else NCOL(x_b)
  if (nexo_w > 0L && is.null(xw_names)) {
    xw_names <- colnames(x_w)
    if (is.null(xw_names)) {
      xw_names <- paste0("xw", seq_len(nexo_w))
    }
  }
  if (nexo_b > 0L && is.null(xb_names)) {
    xb_names <- colnames(x_b)
    if (is.null(xb_names)) {
      xb_names <- paste0("xb", seq_len(nexo_b))
    }
  }

  # ---- stage 1: univariate fits ----
  fit <- vector("list", nvar)
  for (p in seq_len(nvar)) {
    x_w_p <- if (nexo_w > 0L) x_w else NULL
    x_b_p <- if (nexo_b > 0L && between_flag[p]) x_b else NULL
    if (ord_flag[p]) {
      fit[[p]] <- lav_uvord_2l_fit(
        y = y[, p], x_w = x_w_p, x_b = x_b_p,
        cluster_idx = cluster_idx, ngh = ngh
      )
    } else {
      fit[[p]] <- lav_uvreg_2l_fit(
        y = y[, p], x_w = x_w_p, x_b = x_b_p,
        cluster_idx = cluster_idx,
        between = between_flag[p]
      )
    }
  }

  # univariate parameter blocks
  # free_idx_uni: positions of the FREE univariate parameters within
  # fit$theta (excludes the pinned vb of within-only variables);
  # sc_names_uni: the matching cluster-score column names
  npar_uni <- integer(nvar)
  free_idx_uni <- vector("list", nvar)
  sc_names_uni <- vector("list", nvar)
  th <- vector("list", nvar)
  th_idx <- vector("list", nvar)
  mu <- numeric(nvar)
  vw <- numeric(nvar)
  vb <- numeric(nvar)
  slopes_w <- matrix(0, nvar, nexo_w)
  slopes_b <- matrix(0, nvar, nexo_b)
  for (p in seq_len(nvar)) {
    f <- fit[[p]]
    sln <- lav_uvreg_2l_slope_names(f$nexo_w, f$nexo_b)
    if (ord_flag[p]) {
      th[[p]] <- f$theta[f$th_idx]
      th_idx[[p]] <- rep(p, length(th[[p]]))
      mu[p] <- 0
      vw[p] <- 1.0
      vb[p] <- f$theta[f$bvar_idx]
      free_idx_uni[[p]] <- c(f$th_idx, f$bvar_idx,
                             f$wslope_idx, f$bslope_idx)
      sc_names_uni[[p]] <- c(paste0("th", seq_along(f$th_idx)),
                             "bvar", sln)
    } else {
      th[[p]] <- f$theta[f$mu_idx]
      th_idx[[p]] <- 0L
      mu[p] <- f$theta[f$mu_idx]
      vw[p] <- f$theta[f$wvar_idx]
      vb[p] <- f$theta[f$bvar_idx]
      if (between_flag[p]) {
        free_idx_uni[[p]] <- c(1L, 2L, 3L, f$wslope_idx, f$bslope_idx)
        sc_names_uni[[p]] <- c("mu", "wvar", "bvar", sln)
      } else {
        # within-only variables: vb is fixed at 0 (not a parameter)
        free_idx_uni[[p]] <- c(1L, 2L, f$wslope_idx)
        sc_names_uni[[p]] <- c("mu", "wvar", sln)
      }
    }
    if (f$nexo_w > 0L) {
      slopes_w[p, ] <- f$theta[f$wslope_idx]
    }
    if (f$nexo_b > 0L) {
      slopes_b[p, ] <- f$theta[f$bslope_idx]
    }
    npar_uni[p] <- length(free_idx_uni[[p]])
  }

  # ---- stage 2: pairwise fits ----
  pstar <- nvar * (nvar - 1L) / 2L
  sigma_w <- diag(vw, nrow = nvar)
  sigma_b <- diag(vb, nrow = nvar)
  pair_theta <- matrix(0, pstar, 2L) # (cw, cb) per pair
  pair_vars <- matrix(0L, pstar, 2L) # (j, i) with j < i
  pair_type <- character(pstar)
  # pairs involving a within-only variable have cb fixed at 0: only cw
  # is a free parameter
  pair_neta <- integer(pstar)
  pp <- 0L
  for (j in seq_len(nvar - 1L)) {
    for (i in (j + 1L):nvar) {
      pp <- pp + 1L
      pair_vars[pp, ] <- c(j, i)
      if (!ord_flag[j] && !ord_flag[i]) {
        pair_type[pp] <- "bvreg"
        est <- lav_bvreg_2l_cov_twostep_fit(
          fit_y1 = fit[[j]], fit_y2 = fit[[i]],
          y1_name = ov_names[j], y2_name = ov_names[i]
        )
      } else if (ord_flag[j] && ord_flag[i]) {
        pair_type[pp] <- "bvord"
        est <- lav_bvord_2l_cor_twostep_fit(
          fit_y1 = fit[[j]], fit_y2 = fit[[i]], ngh = ngh2,
          y1_name = ov_names[j], y2_name = ov_names[i]
        )
      } else if (!ord_flag[j] && ord_flag[i]) {
        pair_type[pp] <- "bvmix"
        est <- lav_bvmix_2l_cov_twostep_fit(
          fit_y1 = fit[[j]], fit_y2 = fit[[i]], ngh = ngh,
          y1_name = ov_names[j], y2_name = ov_names[i]
        )
      } else {
        # ordinal (j) / continuous (i): continuous variable first
        pair_type[pp] <- "bvmix.swap"
        est <- lav_bvmix_2l_cov_twostep_fit(
          fit_y1 = fit[[i]], fit_y2 = fit[[j]], ngh = ngh,
          y1_name = ov_names[i], y2_name = ov_names[j]
        )
      }
      pair_theta[pp, ] <- est
      pair_neta[pp] <- if (between_flag[j] && between_flag[i]) 2L else 1L
      sigma_w[i, j] <- sigma_w[j, i] <- est[1]
      sigma_b[i, j] <- sigma_b[j, i] <- est[2]
    }
  }

  # the stage-wise parameter (statistic) vector
  s_uni <- unlist(lapply(seq_len(nvar), function(p) {
    fit[[p]]$theta[free_idx_uni[[p]]]
  }))
  s_pair <- unlist(lapply(seq_len(pstar), function(pp) {
    pair_theta[pp, seq_len(pair_neta[pp])]
  }))
  s_all <- c(s_uni, s_pair)
  par_names <- c(
    unlist(lapply(seq_len(nvar), function(p) {
      f <- fit[[p]]
      sl_nm <- character(0L)
      if (f$nexo_w > 0L) {
        sl_nm <- c(sl_nm, paste0(ov_names[p], "~", xw_names))
      }
      if (f$nexo_b > 0L) {
        sl_nm <- c(sl_nm, paste0(ov_names[p], "~", xb_names))
      }
      if (ord_flag[p]) {
        c(paste0(ov_names[p], "|t", seq_along(th[[p]])),
          paste0(ov_names[p], "~b~", ov_names[p]), sl_nm)
      } else if (between_flag[p]) {
        c(paste0(ov_names[p], "~1"),
          paste0(ov_names[p], "~w~", ov_names[p]),
          paste0(ov_names[p], "~b~", ov_names[p]), sl_nm)
      } else {
        c(paste0(ov_names[p], "~1"),
          paste0(ov_names[p], "~w~", ov_names[p]), sl_nm)
      }
    })),
    unlist(lapply(seq_len(pstar), function(pp) {
      j <- pair_vars[pp, 1]; i <- pair_vars[pp, 2]
      nm <- paste0(ov_names[i], "~w~", ov_names[j])
      if (pair_neta[pp] == 2L) {
        nm <- c(nm, paste0(ov_names[i], "~b~", ov_names[j]))
      }
      nm
    }))
  )
  names(s_all) <- par_names

  out <- list(
    FIT = fit,
    TH = th, TH.IDX = th_idx,
    MU = mu, SIGMA.W = sigma_w, SIGMA.B = sigma_b,
    SLOPES.W = slopes_w, SLOPES.B = slopes_b,
    pair.theta = pair_theta, pair.vars = pair_vars,
    pair.type = pair_type, pair.neta = pair_neta,
    between.flag = between_flag,
    nexo.w = nexo_w, nexo.b = nexo_b,
    theta = s_all, npar.uni = npar_uni,
    ord.flag = ord_flag,
    WLS.W = NULL, SC = NULL
  )

  if (!wls_w) {
    return(out)
  }

  # ---- stage 3: sandwich ----
  # global column layout of the univariate blocks
  uni_end <- cumsum(npar_uni)
  uni_start <- uni_end - npar_uni + 1L
  ntheta <- sum(npar_uni)
  eta_end <- cumsum(pair_neta)
  eta_start <- eta_end - pair_neta + 1L
  neta <- sum(pair_neta)
  npar_all <- ntheta + neta
  nclusters <- length(unique(cluster_idx))

  # univariate cluster-wise scores
  sc_all <- matrix(0, nclusters, npar_all)
  for (p in seq_len(nvar)) {
    if (ord_flag[p]) {
      sc_p <- lav_uvord_2l_sc(fit[[p]])
    } else {
      sc_p <- lav_uvreg_2l_sc(fit[[p]])
    }
    sc_p <- sc_p[, sc_names_uni[[p]], drop = FALSE]
    sc_all[, uni_start[p]:uni_end[p]] <- sc_p
  }

  # A11: block-diagonal univariate information (information identity)
  a11 <- matrix(0, ntheta, ntheta)
  a11_inv <- matrix(0, ntheta, ntheta)
  for (p in seq_len(nvar)) {
    idx <- uni_start[p]:uni_end[p]
    a11_p <- crossprod(sc_all[, idx, drop = FALSE])
    a11[idx, idx] <- a11_p
    a11_inv[idx, idx] <- lav_mat_sym_inverse(a11_p)
  }

  # pair-score column names for a member variable, in the same order as
  # its univariate block (sc_names_uni): base names + slope names
  m07_pair_var_names <- function(f, base_names, prefix_sl) {
    sln <- lav_uvreg_2l_slope_names(f$nexo_w, f$nexo_b)
    if (length(sln) == 0L) {
      return(base_names)
    }
    c(base_names, paste0(prefix_sl, sln))
  }

  # pairwise scores; A21 and A22
  a21 <- matrix(0, neta, ntheta)
  a22_inv <- matrix(0, neta, neta)
  for (pp in seq_len(pstar)) {
    j <- pair_vars[pp, 1]; i <- pair_vars[pp, 2]
    cw <- pair_theta[pp, 1]; cb <- pair_theta[pp, 2]
    type <- pair_type[pp]

    if (type == "bvreg") {
      sc_pair <- lav_bvreg_2l_sc(
        fit_y1 = fit[[j]], fit_y2 = fit[[i]], cw = cw, cb = cb
      )
      free_names <- c("wcov", "bcov")
      var1_names <- m07_pair_var_names(
        fit[[j]], c("mu_y1", "wvar_y1",
                    if (between_flag[j]) "bvar_y1"), "y1_")
      var2_names <- m07_pair_var_names(
        fit[[i]], c("mu_y2", "wvar_y2",
                    if (between_flag[i]) "bvar_y2"), "y2_")
    } else if (type == "bvord") {
      sc_pair <- lav_bvord_2l_sc(
        fit_y1 = fit[[j]], fit_y2 = fit[[i]], rho_w = cw, cb = cb,
        ngh = ngh2
      )
      free_names <- c("rho_w", "cb")
      var1_names <- m07_pair_var_names(
        fit[[j]], c(paste0("tau_a", seq_along(th[[j]])), "vb_a"), "a_")
      var2_names <- m07_pair_var_names(
        fit[[i]], c(paste0("tau_c", seq_along(th[[i]])), "vb_c"), "c_")
    } else if (type == "bvmix") {
      # y1 = continuous (j), y2 = ordinal (i)
      sc_pair <- lav_bvmix_2l_sc(
        fit_y1 = fit[[j]], fit_y2 = fit[[i]], cw = cw, cb = cb, ngh = ngh
      )
      free_names <- c("cw", "cb")
      var1_names <- m07_pair_var_names(
        fit[[j]], c("mu_a", "vw_a",
                    if (between_flag[j]) "vb_a"), "a_")
      var2_names <- m07_pair_var_names(
        fit[[i]], c(paste0("tau_c", seq_along(th[[i]])), "vb_c"), "c_")
    } else { # bvmix.swap: y1 = continuous (i), y2 = ordinal (j)
      sc_pair <- lav_bvmix_2l_sc(
        fit_y1 = fit[[i]], fit_y2 = fit[[j]], cw = cw, cb = cb, ngh = ngh
      )
      free_names <- c("cw", "cb")
      # variable j is the ordinal one here (the 'c' side of the module)
      var1_names <- m07_pair_var_names(
        fit[[j]], c(paste0("tau_c", seq_along(th[[j]])), "vb_c"), "c_")
      var2_names <- m07_pair_var_names(
        fit[[i]], c("mu_a", "vw_a",
                    if (between_flag[i]) "vb_a"), "a_")
    }

    free_idx <- match(free_names, colnames(sc_pair))
    var1_idx <- match(var1_names, colnames(sc_pair))
    var2_idx <- match(var2_names, colnames(sc_pair))
    if (anyNA(c(free_idx, var1_idx, var2_idx))) {
      wanted <- c(free_names, var1_names, var2_names)
      missing_names <- wanted[!wanted %in% colnames(sc_pair)]
      lav_msg_stop(gettextf(
        "internal error: pair score columns do not match the univariate
        parameter blocks (pair %1$s/%2$s, missing: %3$s; available: %4$s).",
        ov_names[j], ov_names[i],
        paste(missing_names, collapse = ", "),
        paste(colnames(sc_pair), collapse = ", ")))
    }

    # pairs with a within-only member: cb is fixed at 0, cw only
    free_idx <- free_idx[seq_len(pair_neta[pp])]

    eidx <- eta_start[pp]:eta_end[pp] # position within the eta block
    sc_all[, ntheta + eidx] <- sc_pair[, free_idx, drop = FALSE]

    # A22 block and its inverse
    a22_pp <- crossprod(sc_pair[, free_idx, drop = FALSE])
    a22_inv[eidx, eidx] <- solve(a22_pp)

    # A21 blocks: -E[d v / d theta] = E[v u_pair'] (information identity)
    g1 <- uni_start[j]:uni_end[j]
    g2 <- uni_start[i]:uni_end[i]
    a21[eidx, g1] <- crossprod(
      sc_pair[, free_idx, drop = FALSE],
      sc_pair[, var1_idx, drop = FALSE]
    )
    a21[eidx, g2] <- crossprod(
      sc_pair[, free_idx, drop = FALSE],
      sc_pair[, var2_idx, drop = FALSE]
    )
  }

  # block-triangular inverse of the bread
  a21_inv <- -a22_inv %*% a21 %*% a11_inv
  b_inv <- rbind(
    cbind(a11_inv, matrix(0, ntheta, neta)),
    cbind(a21_inv, a22_inv)
  )

  # meat: crossprod of the stacked cluster-wise scores
  inner <- crossprod(sc_all)

  # sandwich = ACOV of the stage-wise estimates
  wls_w_1 <- b_inv %*% inner %*% t(b_inv)
  wls_w_1 <- (wls_w_1 + t(wls_w_1)) / 2
  rownames(wls_w_1) <- colnames(wls_w_1) <- par_names

  out$WLS.W <- wls_w_1
  out$SC <- sc_all
  out$A11 <- a11
  out$A21 <- a21
  out
}

# numeric approximation of the Hessian
# using an analytic gradient
#
# method = "richardson" (the default): 4-point Richardson
#   extrapolation (4 x npar gradient evaluations), with a fixed
#   (absolute) step size h
# method = "central": plain central differences (2 x npar gradient
#   evaluations), with a *relative* step size h * max(1, |x_j|);
#   cheaper, and the appropriate choice when the gradient itself
#   contains a numerical layer (eg random-slope models)
#
# `NA' semantics: if the gradient cannot be computed (or is not
# finite) at one of the perturbed parameter values, an all-NA matrix
# is returned (callers can check with anyNA()), rather than a hard
# stop or a silently wrong Hessian
lav_model_hessian <- function(lavmodel = NULL,
                              lavsamplestats = NULL,
                              lavdata = NULL,
                              lavoptions = NULL,
                              lavcache = NULL,
                              group_weight = TRUE,
                              ceq_simple = FALSE,
                              h = 1e-06,
                              method = "richardson") {
  # estimator <- lavmodel@estimator

  # catch numerical gradient
  if (lavoptions$optim.gradient == "numerical") {
    obj_f <- function(x) {
      lavmodel2 <- lav_model_set_parameters(lavmodel, x = x)
      lav_model_objective(
        lavmodel = lavmodel2,
        lavsamplestats = lavsamplestats, lavdata = lavdata,
        lavcache = lavcache
      )[1]
    }
    x <- lav_model_get_parameters(lavmodel = lavmodel)
    hessian <- try(numDeriv::hessian(func = obj_f, x = x),
                   silent = TRUE)
    if (inherits(hessian, "try-error")) {
      hessian <- matrix(as.numeric(NA), length(x), length(x))
    }
    return(hessian)
  }

  # two-level + missing data: use Louis's (1982) method (one analytic
  # pass) instead of numerically differentiating the gradient (which
  # needs 4 x npar full-data gradient passes); new in 0.7-1
  if (!is.null(lavdata) && lavdata@nlevels > 1L &&
      lavdata@ngroups == 1L &&
      lavsamplestats@missing.flag &&
      lavmodel@estimator == "ML" &&
      !lavmodel@conditional.x &&
      length(lavmodel@rv.ov) == 0L && length(lavmodel@rv.lv) == 0L &&
      is.null(lavdata@weights[[1]])) {
    hessian <- try(
      lav_mvn_cl_mi_h_louis(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        ceq_simple = ceq_simple
      ),
      silent = TRUE
    )
    if (!inherits(hessian, "try-error") && all(is.finite(hessian))) {
      return(hessian)
    }
    # if the Louis computation fails, fall through to the numerical
    # approximation below
  }

  # computing the Richardson extrapolation (or central differences)
  if (!ceq_simple && lavmodel@ceq.simple.only) {
    npar <- lavmodel@nx.unco
    type_glist <- "unco"
  } else {
    npar <- lavmodel@nx.free
    type_glist <- "free"
  }
  hessian <- matrix(0, npar, npar)
  x <- lav_model_get_parameters(lavmodel = lavmodel)
  if (!ceq_simple && lavmodel@ceq.simple.only) {
    # unpack
    x <- drop(x %*% t(lavmodel@ceq.simple.K))
  }

  # the gradient at a perturbed x; NULL if it cannot be computed
  grad_at <- function(x_p) {
    out <- try(
      lav_model_grad(
        lavmodel = lavmodel,
        glist = lav_model_x2glist(
          lavmodel = lavmodel, type = type_glist, x = x_p
        ),
        lavsamplestats = lavsamplestats,
        lavdata = lavdata,
        lavcache = lavcache,
        type = "free",
        group_weight = group_weight,
        ceq_simple = ceq_simple
      ),
      silent = TRUE
    )
    if (inherits(out, "try-error") || anyNA(out) ||
        any(!is.finite(out))) {
      return(NULL)
    }
    out
  }
  hessian_na <- matrix(as.numeric(NA), npar, npar)

  for (j in seq_len(npar)) {
    if (method == "central") {
      # central differences, relative step size
      h_j <- h * max(1, abs(x[j]))
      x_left <- x_right <- x
      x_left[j] <- x[j] - h_j
      x_right[j] <- x[j] + h_j

      g_left <- grad_at(x_left)
      g_right <- grad_at(x_right)
      if (is.null(g_left) || is.null(g_right)) {
        return(hessian_na)
      }

      hessian[, j] <- (g_right - g_left) / (2 * h_j)
    } else {
      # Richardson extrapolation
      # FIXME: the number below should vary as a function of 'x[j]'
      h_j <- h
      x_left <- x_left2 <- x_right <- x_right2 <- x
      x_left[j] <- x[j] - h_j
      x_left2[j] <- x[j] - 2 * h_j
      x_right[j] <- x[j] + h_j
      x_right2[j] <- x[j] + 2 * h_j

      g_left <- grad_at(x_left)
      g_left2 <- grad_at(x_left2)
      g_right <- grad_at(x_right)
      g_right2 <- grad_at(x_right2)
      if (is.null(g_left) || is.null(g_left2) ||
          is.null(g_right) || is.null(g_right2)) {
        return(hessian_na)
      }

      hessian[, j] <- (g_left2 - 8 * g_left + 8 * g_right - g_right2) /
        (12 * h_j)
    }
  }

  # check if Hessian is (almost) symmetric, as it should be
  max_diff <- max(abs(hessian - t(hessian)))
  if (max_diff > 1e-05 * max(diag(hessian))) {
    # hm, Hessian is not symmetric -> WARNING!
    lav_msg_warn(gettextf(
    "Hessian is not fully symmetric. Max diff = %1$s (Max diag Hessian = %2$s)",
    max_diff, max(diag(hessian))))
    # FIXME: use numDeriv::hessian instead?
  }
  hessian <- (hessian + t(hessian)) / 2.0

  hessian
}

# Albert (1944a/b) & Ihara & Kano 1986 method to estimate residual variances
# of indicators using a PArtitioned Covariance matrix Estimator (PACE)
#
# The implementation is based on Cudeck 1991:

# Cudeck, R. (1991). Noniterative factor analysis estimators, with algorithms
# for subset and instrumental variable selection. Journal of Educational
# Statistics, 16(1), 35-52.

# YR -- 14 FEB 2020

# - 'fast' version; only (2*nfactors + 1) iterations are needed
# - scale-invariant (by default)
# - always assuming unit variances for the factors
lav_efa_pace <- function(s, nfactors = 1L, p_idx = seq_len(ncol(s)),
                         reflect = TRUE, order_lv_by = "none",
                         use_r = TRUE, theta_only = TRUE) {
  s <- unname(s)
  nvar <- ncol(s)
  theta <- numeric(nvar)
  stopifnot(nfactors < nvar / 2)

  # because subset selection is not scale-invariant, we transform
  # S to R, compute theta based on R, and then rescale again
  if (use_r) {
    s_var <- diag(s)
    r <- stats::cov2cor(s)
  } else {
    r <- s
  }

  # find principal variables ('largest' sub-block)
  a <- r

  # row indices
  v_r <- integer(0L)
  # column indices
  v_c <- integer(0L)

  for (h in seq_len(nfactors)) {
    # mask
    mask_idx <- c(v_r, v_c)
    tmp <- abs(a)
    if (length(mask_idx) > 0L) {
      tmp[mask_idx, ] <- 0
      tmp[, mask_idx] <- 0
    }
    diag(tmp) <- 0

    # find maximum off-diagonal element
    idx <- which(tmp == max(tmp), arr.ind = TRUE, useNames = FALSE)[1, ]
    k <- idx[1]
    l <- idx[2]
    v_r <- c(v_r, k)
    v_c <- c(v_c, l)

    # non-symmetric sweep operator
    a_kl <- a[k, l]
    if (abs(a_kl) < sqrt(.Machine$double.eps)) {
      out <- a
      out[k, ] <- 0
      out[, l] <- 0
    } else {
      out <- a - tcrossprod(a[, l], a[k, ]) / a_kl
      out[k, ] <- a[k, ] / a_kl
      out[, l] <- -a[, l] / a_kl
      out[k, l] <- 1 / a_kl
    }
    a <- out
  }
  # diagonal elements are estimates of theta
  # for all variables not in (v.r, v.c)
  all_idx <- seq_len(nvar)
  # v_r_init <- v_r
  # v_c_init <- v_c
  other_idx <- all_idx[-c(v_r, v_c)]
  theta[other_idx] <- diag(a)[other_idx]


  # now fill in theta for the 2*m remaining variables in c(v.r.init, v.c.init)
  for (i in p_idx) {
    if (i %in% other_idx) {
      next
    }

    # row indices
    v_r <- integer(0L)
    # column indices
    v_c <- integer(0L)

    a <- r
    for (h in seq_len(nfactors)) {
      # mask
      mask_idx <- c(i, v_r, v_c)
      tmp <- abs(a)
      tmp[mask_idx, ] <- 0
      tmp[, mask_idx] <- 0
      diag(tmp) <- 0

      # find maximum off-diagonal element
      idx <- which(tmp == max(tmp), arr.ind = TRUE, useNames = FALSE)[1, ]
      k <- idx[1]
      l <- idx[2]
      v_r <- c(v_r, k)
      v_c <- c(v_c, l)

      # non-symmetric sweep operator
      a_kl <- a[k, l]
      if (abs(a_kl) < sqrt(.Machine$double.eps)) {
        out <- a
        out[k, ] <- 0
        out[, l] <- 0
      } else {
        out <- a - tcrossprod(a[, l], a[k, ]) / a_kl
        out[k, ] <- a[k, ] / a_kl
        out[, l] <- -a[, l] / a_kl
        out[k, l] <- 1 / a_kl
      }
      a <- out
    }

    # diagonal element is estimate of theta
    theta[i] <- a[i, i]
  }

  # return theta elements only
  if (theta_only) {
    # rescale back to S metric
    if (use_r) {
      theta <- theta * s_var
    }
    return(theta[p_idx])
  }

  # compute LAMBDA using the 'eigenvalue' method
  ev <- eigen(r, symmetric = TRUE)
  s_sqrt <- ev$vectors %*% sqrt(diag(ev$values)) %*% t(ev$vectors)
  s_inv_sqrt <- ev$vectors %*% sqrt(diag(1 / ev$values)) %*% t(ev$vectors)
  rtr <- s_inv_sqrt %*% diag(theta) %*% s_inv_sqrt

  ev <- eigen(rtr, symmetric = TRUE)
  omega_m <- ev$vectors[, 1L + nvar - seq_len(nfactors), drop = FALSE]
  gamma_m <- ev$values[1L + nvar - seq_len(nfactors)]
  gamma_m_1 <- diag(gamma_m, nrow = nfactors, ncol = nfactors)

  # Cudeck 1991 page 37 bottom of the page:
  lambda_dot <- s_sqrt %*% omega_m %*% sqrt(diag(nfactors) - gamma_m_1)

  if (use_r) {
    # IF (and only if) the input is a correlation matrix,
    # we must rescale so that the diag(R.implied) == 1

    # R.unscaled <- tcrossprod(LAMBDA.dot) + diag(theta)
    # r.var.inv <- 1/diag(R.unscaled)

    # LAMBDA/THETA in correlation metric
    # LAMBDA.R <- sqrt(r.var.inv) * LAMBDA.dot
    # THETA.R <- diag(r.var.inv * theta)

    # convert to 'S' metric
    mm_lambda <- sqrt(s_var) * lambda_dot
    mm_theta <- diag(s_var * theta)
  } else {
    mm_lambda <- lambda_dot
    mm_theta <- diag(theta)
  }

  # reflect (optional) + reorder the columns
  mm_lambda <- lav_efa_reflect_and_reorder(mm_lambda, reflect, order_lv_by)

  list(LAMBDA = mm_lambda, THETA = mm_theta)
}

# build a bare-bones parameter table from a fitted lm object
#
# YR: this function was broken since Mar 3, 2017, but nobody noticed this!
#     fixed again Apr 29, 2025.

lav_pt_from_lm <- function(object, est = FALSE, label = FALSE,
                                 as_data_frame = FALSE) {
  # sanity check
  if (!inherits(object, "lm")) {
    lav_msg_stop(gettext("object must be of class lm"))
  }

  object_terms <- terms(object)

  response_index <- attr(object_terms, "response")
  var_names <- as.character(attr(object_terms, "variables"))[-1]
  response_name <- var_names[response_index]

  pred_coef <- coef(object)
  pred_names <- names(pred_coef)

  lhs <- rep(response_name, length(pred_names))
  op <- rep("~", length(pred_names))
  rhs <- pred_names

  # intercept?
  if (attr(object_terms, "intercept")) {
    int_idx <- which(rhs == "(Intercept)")
    op[int_idx] <- "~1"
    rhs[int_idx] <- ""
  }

  # always add residual variance?
  # lhs <- c(lhs, responseName)
  # op <- c(op, "~~")
  # rhs <- c(rhs, responseName)

  # construct minimal partable
  partable <- list(lhs = lhs, op = op, rhs = rhs)

  # include 'est' column?
  if (est) {
    # partable$est <- c(as.numeric(predCoef),
    #                  sum(resid(object)^2) / object$df.residual)
    partable$est <- as.numeric(pred_coef)
  }

  # include 'label' column?
  if (label) {
    # partable$label <- c(predNames, responseName)
    partable$label <- pred_names

    # convert all ':' to '.'
    partable$label <- gsub("[:()]", ".", partable$label)
  }

  # convert to data.frame?
  if (as_data_frame) {
    partable <- as.data.frame(partable, stringsAsFactors = FALSE)
  }

  partable
}

# the information matrix of the unrestricted (H1) model
# taking into account:
#   - the estimator (ML or (D)WLS/ULS)
#   - missing or not
#   - fixed.x = TRUE or FALSE
#   - conditional.x = TRUE or FALSE
#   - h1.information is "structured" or "unstructured"
#
# Note: this replaces the (old) lav_model_wls_v() function
#
# - YR 22 Okt 2017: initial version
# - YR 03 Dec 2017: add lavh1, implied is either lavimplied or lavh1
#                   add support for clustered data: first.order
# - YR 03 Jan 2018: add support for clustered data: expected
# - YR 23 Aug 2018: lav_model_h1_acov (0.6-3)
# - YR 08 Jul 2026: merge the (almost fully parallel) expected/observed
#                   scaffolding into a single worker (lav_model_h1_info_ed)


## For the lavaan.mi package, TDJ provides pooled versions of all the
## sample moments called in these functions.  If any updates to these functions
## require NEW information (from @SampleStats or @implied or @h1),
## PLEASE ADD A TAG     @TDJorgensen
## at the end of the commit message on GitHub, so TDJ can check whether
## lavaan.mi::lavResiduals.mi() needs to be updated accordingly.


# all lav_model_h1_* functions accept either a fitted lavaan object
# (lavobject) or the individual slots; this helper unpacks the object into
# the slot variables of the CALLER's frame, and (optionally) fills in
# lavh1/lavimplied when they were not supplied
lav_model_h1_unpack <- function(lavobject = NULL, env = parent.frame(),
                                need_h1 = FALSE, need_implied = FALSE) {
  if (!is.null(lavobject) && inherits(lavobject, "lavaan")) {
    env$lavmodel <- lavobject@Model
    env$lavsamplestats <- lavobject@SampleStats
    env$lavdata <- lavobject@Data
    env$lavimplied <- lavobject@implied
    env$lavh1 <- lavobject@h1
    env$lavcache <- lavobject@Cache
    env$lavoptions <- lavobject@Options
  }

  # sanity check
  if (need_h1 && length(env$lavh1) == 0L) {
    env$lavh1 <- lav_h1_implied_logl(
      lavdata = env$lavdata,
      lavsamplestats = env$lavsamplestats,
      lavoptions = env$lavoptions
    )
  }
  if (need_implied && length(env$lavimplied) == 0L) {
    env$lavimplied <- lav_model_implied(lavmodel = env$lavmodel)
  }

  invisible(NULL)
}


lav_model_h1_info <- function(lavobject = NULL,
                                     lavmodel = NULL,
                                     lavsamplestats = NULL,
                                     lavdata = NULL,
                                     lavimplied = NULL,
                                     lavh1 = NULL,
                                     lavcache = NULL,
                                     lavoptions = NULL) {
  lav_model_h1_unpack(lavobject, need_h1 = TRUE, need_implied = TRUE)

  # information
  information <- lavoptions$information[1] # ALWAYS take the first one
  # the caller must control it


  # compute information matrix
  if (information == "observed") {
    i1 <- lav_model_h1_info_observed(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, lavoptions = lavoptions
    )
  } else if (information == "expected") {
    i1 <- lav_model_h1_info_expected(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, lavoptions = lavoptions
    )
  } else if (information == "first.order") {
    i1 <- lav_model_h1_info_firstorder(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, lavoptions = lavoptions
    )
  }

  # I1 information, as a list per group
  i1
}

# shared worker for the expected and observed h1 information: the two are
# structurally identical (same estimator dispatch, same structured /
# unstructured moment selection, same group.w.free handling) -- only the
# innermost kernel calls differ. Returns NULL for unsupported estimators
# (the named wrappers below raise the error, so the message carries their
# name).
lav_model_h1_info_ed <- function(what = "expected",
                                 lavmodel = NULL,
                                 lavsamplestats = NULL,
                                 lavdata = NULL,
                                 lavoptions = NULL,
                                 lavimplied = NULL,
                                 lavh1 = NULL) {
  observed <- (what == "observed")

  # sanity check
  # (not needed for the estimators that read the precomputed WLS.V/WLS.VD
  #  slots -- and for two-level categorical WLS there is no h1 object)
  if (length(lavh1) == 0L &&
    !lavmodel@estimator %in% c("GLS", "WLS", "DLS", "DWLS", "ULS")) {
    lavh1 <- lav_h1_implied_logl(
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      lavoptions = lavoptions
    )
  }
  if (length(lavimplied) == 0L) {
    lavimplied <- lav_model_implied(lavmodel = lavmodel)
  }

  # structured or unstructured? (since 0.5-23)
  if (!is.null(lavoptions) &&
    !is.null(lavoptions$h1.information[1]) &&
    lavoptions$h1.information[1] == "unstructured") {
    structured <- FALSE
  } else {
    structured <- TRUE
  }

  # which single-level ML-family estimators are handled by branch 3a below?
  # the observed h1 information is only available for "ML" itself
  # (estimator "DLS" is caught by the dedicated branch 1b)
  if (observed) {
    ml_family <- "ML"
  } else {
    ml_family <- c("ML", "NTRLS", "catML", "IV")
  }

  # 1. WLS.V (=A1) for GLS/WLS
  if (lavmodel@estimator == "GLS" || lavmodel@estimator == "WLS") {
    a1 <- lavsamplestats@WLS.V
    # GLS: since 0.7-1 the NT weight matrix is no longer pre-computed at
    # the samplestats stage (estimation does not need it); build it here
    # on demand. Only reached for the 'plain' single-level variant: the
    # correlation/conditional.x/group.w.free/two.stage variants still
    # store WLS.V eagerly (see lav_samp_from_data).
    if (lavmodel@estimator == "GLS" && is.null(a1[[1]])) {
      for (g in seq_len(lavsamplestats@ngroups)) {
        a1[[g]] <- lav_samp_wls_v_nt_g(
          m_cov         = lavsamplestats@cov[[g]],
          m_mean        = lavsamplestats@mean[[g]],
          m_icov        = lavsamplestats@icov[[g]],
          correlation   = FALSE,
          x_idx         = lavsamplestats@x.idx[[g]],
          fixed_x       = lavmodel@fixed.x,
          conditional_x = FALSE,
          meanstructure = lavmodel@meanstructure
        )
        # mimic Mplus: V11 rescale (only for full-data input; the
        # moments-based lav_samp_from_moments never applied it)
        if (isTRUE(lavmodel@estimator.args$gls.v11.mplus) &&
          lavmodel@meanstructure &&
          !is.null(lavdata) && lavdata@data.type == "full") {
          nvar <- NCOL(lavsamplestats@cov[[g]])
          a1[[g]][1:nvar, 1:nvar] <- a1[[g]][1:nvar, 1:nvar, drop = FALSE] *
            (lavsamplestats@nobs[[g]] / (lavsamplestats@nobs[[g]] - 1))
        }
      }
    }

  # 1b. DLS: for dls.GammaNT = "model", the weight matrix is a function of
  #     the model-implied moments and must be recomputed here (the @WLS.V
  #     slot only holds the plain inv(NACOV) fallback)
  } else if (lavmodel@estimator == "DLS") {
    if (lavmodel@estimator.args$dls.GammaNT == "sample") {
      a1 <- lavsamplestats@WLS.V
    } else {
      a1 <- vector("list", length = lavsamplestats@ngroups)
      for (g in seq_len(lavsamplestats@ngroups)) {
        a1[[g]] <- lav_dls_wls_v_g(
          m_cov         = lavimplied$cov[[g]],
          m_mean        = lavimplied$mean[[g]],
          nacov_g       = lavsamplestats@NACOV[[g]],
          dls_a         = lavmodel@estimator.args$dls.a,
          x_idx         = lavsamplestats@x.idx[[g]],
          fixed_x       = lavmodel@fixed.x,
          conditional_x = lavmodel@conditional.x,
          meanstructure = lavmodel@meanstructure
        )
      }
    }

  # 2. DWLS/ULS diagonal @WLS.VD slot
  } else if (lavmodel@estimator == "DWLS" || lavmodel@estimator == "ULS") {
    # diagonal only!!
    a1 <- lavsamplestats@WLS.VD

  # 3a. ML single level
  } else if (lavmodel@estimator %in% ml_family && lavdata@nlevels == 1L) {
    a1 <- vector("list", length = lavsamplestats@ngroups)

    for (g in 1:lavsamplestats@ngroups) {
      if (lavsamplestats@missing.flag && lavmodel@conditional.x) {
        # mvreg, missing data
        if (lavmodel@meanstructure && structured) {
          res_int <- lavimplied$res.int[[g]]
          res_slopes <- lavimplied$res.slopes[[g]]
        } else {
          res_int <- lavh1$implied$res.int[[g]]
          res_slopes <- lavh1$implied$res.slopes[[g]]
        }
        if (structured) {
          res_cov <- lavimplied$res.cov[[g]]
        } else {
          res_cov <- lavh1$implied$res.cov[[g]]
        }

        if (observed) {
          a1[[g]] <- lav_mvreg_mi_information_observed_samplestats(
            yp = lavsamplestats@missing[[g]],
            res_int = res_int,
            res_slopes = res_slopes,
            res_cov = res_cov
          )
        } else {
          a1[[g]] <- lav_mvreg_mi_info_expected(
            yp = lavsamplestats@missing[[g]],
            res_cov = res_cov
          )
        }
      } else if (lavsamplestats@missing.flag) {
        # mvnorm, missing data
        # FIXME: allow for meanstructure = FALSE
        if (lavmodel@meanstructure && structured) {
          mean_1 <- lavimplied$mean[[g]]
        } else {
          mean_1 <- lavh1$implied$mean[[g]]
        }
        if (structured) {
          sigma_1 <- lavimplied$cov[[g]]
        } else {
          sigma_1 <- lavh1$implied$cov[[g]]
        }

        if (observed) {
          a1[[g]] <-
            lav_mvnorm_missing_information_observed_samplestats(
              yp = lavsamplestats@missing[[g]],
              # wt not needed
              mu = mean_1,
              sigma_1 = sigma_1,
              x_idx = lavsamplestats@x.idx[[g]]
            )
        } else {
          a1[[g]] <-
            lav_mvn_mi_info_expected(
              y = lavdata@X[[g]],
              mp = lavdata@Mp[[g]],
              wt = lavdata@weights[[g]],
              mu = mean_1,
              sigma_1 = sigma_1,
              x_idx = lavsamplestats@x.idx[[g]]
            )
        }
      } else if (lavmodel@conditional.x) {
        # mvreg
        if (lavmodel@meanstructure && structured) {
          res_int <- lavimplied$res.int[[g]]
          res_slopes <- lavimplied$res.slopes[[g]]
        } else {
          res_int <- lavsamplestats@res.int[[g]]
          res_slopes <- lavsamplestats@res.slopes[[g]]
        }
        if (structured) {
          res_cov <- lavimplied$res.cov[[g]]
        } else {
          res_cov <- lavsamplestats@res.cov[[g]]
        }

        if (observed) {
          a1[[g]] <- lav_mvreg_information_observed_samplestats(
            sample_res_int    = lavsamplestats@res.int[[g]],
            sample_res_slopes = lavsamplestats@res.slopes[[g]],
            sample_res_cov    = lavsamplestats@res.cov[[g]],
            sample_mean_x     = lavsamplestats@mean.x[[g]],
            sample_cov_x      = lavsamplestats@cov.x[[g]],
            res_int           = res_int,
            res_slopes        = res_slopes,
            res_cov           = res_cov
          )
        } else {
          a1[[g]] <- lav_mvreg_info_expected(
            sample_mean_x     = lavsamplestats@mean.x[[g]],
            sample_cov_x      = lavsamplestats@cov.x[[g]],
            sample_nobs       = lavsamplestats@nobs[[g]],
            res_int           = res_int,
            res_slopes        = res_slopes,
            res_cov           = res_cov
          )
        }
      } else {
        # complete data, conditional.x = FALSE
        # mvnorm
        if (observed) {
          if (lavmodel@meanstructure && structured) {
            mean_1 <- lavimplied$mean[[g]]
          } else {
            mean_1 <- lavsamplestats@mean[[g]]
          }

          if (structured) {
            a1[[g]] <- lav_mvn_info_observed_samp(
              sample_mean   = lavsamplestats@mean[[g]],
              sample_cov    = lavsamplestats@cov[[g]],
              mu            = mean_1,
              sigma_1       = lavimplied$cov[[g]],
              x_idx         = lavsamplestats@x.idx[[g]],
              meanstructure = lavmodel@meanstructure
            )
          } else {
            a1[[g]] <- lav_mvn_h1_info_observed_samp(
              sample_mean    = lavsamplestats@mean[[g]],
              sample_cov     = lavsamplestats@cov[[g]],
              sample_cov_inv = lavsamplestats@icov[[g]],
              x_idx          = lavsamplestats@x.idx[[g]],
              meanstructure  = lavmodel@meanstructure
            )
          }
        } else {
          if (structured) {
            a1[[g]] <- lav_mvn_info_expected(
              sigma_1       = lavimplied$cov[[g]],
              x_idx         = lavsamplestats@x.idx[[g]],
              meanstructure = lavmodel@meanstructure,
              correlation   = lavmodel@correlation
            )
          } else {
            a1[[g]] <- lav_mvn_h1_info_expected(
              sample_cov_inv = lavsamplestats@icov[[g]],
              x_idx          = lavsamplestats@x.idx[[g]],
              meanstructure  = lavmodel@meanstructure,
              correlation    = lavmodel@correlation
            )
          }
        }
      } # complete/missing, conditional.x

      # stochastic group weight
      if (lavmodel@group.w.free) {
        # unweight!! (as otherwise, we would 'weight' again)
        a <- exp(lavimplied$group.w[[g]]) / lavsamplestats@nobs[[g]]
        a1[[g]] <- lav_mat_bdiag(matrix(a, 1L, 1L), a1[[g]])
      }
    } # g

  # 3b. ML + multilevel
  } else if (lavmodel@estimator == "ML" && lavdata@nlevels > 1L) {
    a1 <- vector("list", length = lavsamplestats@ngroups)

    # structured? lavimplied vs lavh1
    if (structured) {
      implied <- lavimplied
    } else {
      implied <- lavh1$implied
    }

    for (g in 1:lavsamplestats@ngroups) {
      mu_w <- implied$mean[[(g - 1) * lavdata@nlevels + 1L]]
      mu_b <- implied$mean[[(g - 1) * lavdata@nlevels + 2L]]
      sigma_w <- implied$cov[[(g - 1) * lavdata@nlevels + 1L]]
      sigma_b <- implied$cov[[(g - 1) * lavdata@nlevels + 2L]]

      if (observed && lavdata@missing %in% c("ml", "ml.x")) {
        a1[[g]] <- lav_mvn_cl_mi_info_observed(
          y1            = lavdata@X[[g]],
          y2            = lavsamplestats@YLp[[g]][[2]]$Y2,
          lp            = lavdata@Lp[[g]],
          mp            = lavdata@Mp[[g]],
          mu_w          = mu_w,
          sigma_w       = sigma_w,
          mu_b          = mu_b,
          sigma_b       = sigma_b,
          x_idx         = lavsamplestats@x.idx[[g]]
        )
      } else if (observed) {
        a1[[g]] <- lav_mvn_cl_info_observed(
          lp           = lavdata@Lp[[g]],
          ylp          = lavsamplestats@YLp[[g]],
          mu_w         = mu_w,
          sigma_w      = sigma_w,
          mu_b         = mu_b,
          sigma_b      = sigma_b,
          x_idx        = lavsamplestats@x.idx[[g]]
        )
      } else {
        a1[[g]] <- lav_mvn_cl_info_expected(
          lp           = lavdata@Lp[[g]],
          mu_w         = mu_w,
          sigma_w      = sigma_w,
          mu_b         = mu_b,
          sigma_b      = sigma_b,
          x_idx        = lavsamplestats@x.idx[[g]]
        )
      }
    } # g
  } else {
    # no branch matched (e.g. PML): the wrappers raise the error
    return(NULL)
  }

  a1
}

# fisher/expected information of H1
lav_model_h1_info_expected <- function(lavobject = NULL,
                                              lavmodel = NULL,
                                              lavsamplestats = NULL,
                                              lavdata = NULL,
                                              lavoptions = NULL,
                                              lavimplied = NULL,
                                              lavh1 = NULL,
                                              lavcache = NULL) {
  lav_model_h1_unpack(lavobject)

  a1 <- lav_model_h1_info_ed(
    what = "expected",
    lavmodel = lavmodel, lavsamplestats = lavsamplestats,
    lavdata = lavdata, lavoptions = lavoptions,
    lavimplied = lavimplied, lavh1 = lavh1
  )
  if (is.null(a1)) {
    # unsupported estimator (e.g. PML): fail with a clear message instead
    # of an obscure "object 'a1' not found"
    lav_msg_stop(gettextf(
      "the (expected) h1 information matrix is not available for
       estimator %s.", lavmodel@estimator))
  }

  a1
}

lav_model_h1_info_observed <- function(lavobject = NULL,
                                              lavmodel = NULL,
                                              lavsamplestats = NULL,
                                              lavdata = NULL,
                                              lavimplied = NULL,
                                              lavh1 = NULL,
                                              lavcache = NULL,
                                              lavoptions = NULL) {
  lav_model_h1_unpack(lavobject)

  a1 <- lav_model_h1_info_ed(
    what = "observed",
    lavmodel = lavmodel, lavsamplestats = lavsamplestats,
    lavdata = lavdata, lavoptions = lavoptions,
    lavimplied = lavimplied, lavh1 = lavh1
  )
  if (is.null(a1)) {
    # unsupported estimator (e.g. PML): fail with a clear message instead
    # of an obscure "object 'a1' not found"
    lav_msg_stop(gettextf(
      "the (observed) h1 information matrix is not available for
       estimator %s.", lavmodel@estimator))
  }

  a1
}

# outer product of the case-wise scores (gradients)
# HJ 18/10/2023: Adjust J matrix correctly using weights. Note: H matrix is
# based on lav_model_hessian so no changes required.
lav_model_h1_info_firstorder <- function(lavobject = NULL,
                                                lavmodel = NULL,
                                                lavsamplestats = NULL,
                                                lavdata = NULL,
                                                lavimplied = NULL,
                                                lavh1 = NULL,
                                                lavcache = NULL,
                                                lavoptions = NULL) {
  lav_model_h1_unpack(lavobject, need_h1 = TRUE, need_implied = TRUE)

  estimator <- lavmodel@estimator
  if (!estimator %in% c("ML", "PML")) {
    lav_msg_stop(gettext(
      "information = \"first.order\" not available for estimator"),
      sQuote(estimator))
  }

  # structured?
  if (!is.null(lavoptions) &&
    !is.null(lavoptions$h1.information[1]) &&
    lavoptions$h1.information[1] == "unstructured") {
    structured <- FALSE
  } else {
    structured <- TRUE
  }

  # clustered?
  if (!is.null(lavoptions) &&
    !is.null(lavoptions$.clustered) &&
    lavoptions$.clustered) {
    clustered <- TRUE
    if (is.null(lavdata@Lp[[1]])) {
      lav_msg_stop(gettext("lavdata@Lp is empty, while clustered = TRUE"))
    }
    # if (estimator == "PML") {
    #   lav_msg_stop(gettext(
    #     "clustered information is not (yet) available when estimator = 'PML'"))
    # }
    # if(lavsamplestats@missing.flag) {
    #    stop("lavaan ERROR: clustered information is not (yet) available
    #                               when missing = \"ML\"")
    # }
    # if(lavmodel@conditional.x) {
    #    stop("lavaan ERROR: clustered information is not (yet) available
    #                               when conditional.x = TRUE")
    # }
    # if(!structured) {
    #    stop("lavaan ERROR: clustered information is not (yet) available
    #                               when h1.information = \"unstructured\"")
    # }
  } else {
    clustered <- FALSE
  }

  # structured? compute model-implied statistics
  if (estimator == "PML" || structured) {
    if (length(lavimplied) == 0L) {
      lavimplied <- lav_model_implied(lavmodel)
    }
  }

  # structured? lavimplied vs lavh1
  if (structured) {
    implied <- lavimplied
  } else {
    implied <- lavh1$implied
  }

  b1 <- vector("list", length = lavsamplestats@ngroups)
  for (g in 1:lavdata@ngroups) {
    wt <- lavdata@weights[[g]]

    if (estimator == "PML") {
      # slow approach: compute outer product of case-wise scores

      if (lavmodel@conditional.x) {
        sigma_1 <- implied$res.cov[[g]]
        mu <- implied$res.mean[[g]]
        th <- implied$res.th[[g]]
        pi0 <- implied$res.slopes[[g]]
        exo <- lavdata@eXo[[g]]
      } else {
        sigma_1 <- implied$cov[[g]]
        mu <- implied$mean[[g]]
        th <- implied$th[[g]]
        pi0 <- NULL
        exo <- NULL
      }
      sc <- lav_pml_dploglik_dimplied(
        sigma_hat = sigma_1,
        mu_hat = mu,
        th = th,
        th_idx = lavmodel@th.idx[[g]],
        num_idx = lavmodel@num.idx[[g]],
        x = lavdata@X[[g]],
        exo = exo,
        wt = NULL,
        pi0 = pi0,
        lavcache = lavcache[[g]],
        missing = lavdata@missing,
        scores = TRUE,
        negative = FALSE
      )

      # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


      # information H1

      if (isTRUE(clustered)) {
        # For clustered data, need to compute (centred) crossprod within each
        # cluster and sum them all up.
        clusters     <- lavdata@Lp[[g]]$cluster.id[[2]]  # why list of 2?
        clusters_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
        nclust       <- length(clusters)
        zb <- list()

        if (is.null(wt)) wt <- rep(1, length(clusters_idx))

        for (b in seq_along(clusters)) {
          # drop = FALSE: a single-observation cluster must stay a matrix
          sc_b <- sc[clusters_idx == b, , drop = FALSE]
          wt_b <- wt[clusters_idx == b]
          zb[[b]] <- colSums(sc_b * wt_b)
        }
        zbar <- apply(do.call(cbind, zb), 1, mean)
        b1c <- Reduce(f = `+`, lapply(zb, function(z) tcrossprod(z - zbar)))
        b1[[g]] <- nclust / (nclust - 1) * b1c

      } else {
        if (is.null(wt)) {
          b1[[g]] <- lav_mat_crossprod(sc)
        } else {
          # sc holds the unit (unweighted) casewise scores. How the sampling
          # weights enter the first-order information (the meat of the robust
          # sandwich / the scaled test) is controlled by sampling.weights.type:
          #  - "design"    : crossprod(wt * sc), weighted by sum(wt^2), the
          #                  design-based version (matches Mplus).
          #  - "frequency" : crossprod(sqrt(wt) * sc), weighted by sum(wt),
          #                  i.e. treats the weights as frequencies, so the
          #                  result matches a fit on the row-replicated data.
          swt_type <- if (!is.null(lavoptions$sampling.weights.type)) {
            lavoptions$sampling.weights.type
          } else {
            "design"
          }
          b1[[g]] <- lav_samp_wt_crossprod(sc, wt, swt_type)
        }
      }


    # ML
    } else if (estimator == "ML" && lavdata@nlevels > 1L) {
      # if not-structured, we use lavh1, and that is always
      # 'unconditional' (for now)
      if (lavmodel@conditional.x && structured) {
      if (lavdata@missing %in% c("ml", "ml.x")) {
      lav_msg_stop(gettext("firstorder information matrix not available
                                (yet) if conditional.x + fiml"))
    }
        res_sigma_w <- implied$res.cov[[(g - 1) * lavdata@nlevels + 1L]]
        res_int_w <- implied$res.int[[(g - 1) * lavdata@nlevels + 1L]]
        res_pi_w <- implied$res.slopes[[(g - 1) * lavdata@nlevels + 1L]]

        res_sigma_b <- implied$res.cov[[(g - 1) * lavdata@nlevels + 2L]]
        res_int_b <- implied$res.int[[(g - 1) * lavdata@nlevels + 2L]]
        res_pi_b <- implied$res.slopes[[(g - 1) * lavdata@nlevels + 2L]]
        b1[[g]] <- lav_mvreg_cl_info_firstorder(
          y1            = lavdata@X[[g]],
          ylp           = lavsamplestats@YLp[[g]],
          lp            = lavdata@Lp[[g]],
          res_sigma_w   = res_sigma_w,
          res_int_w     = res_int_w,
          res_pi_w      = res_pi_w,
          res_sigma_b   = res_sigma_b,
          res_int_b     = res_int_b,
          res_pi_b      = res_pi_b,
          divide_by_two = TRUE
        )
        # reorder from the kernel statistic order to the Delta row order
        perm <- lav_mvreg_cl_stat_perm(
          res_int_w = res_int_w, res_pi_w = res_pi_w,
          res_int_b = res_int_b, res_pi_b = res_pi_b
        )
        b1[[g]] <- b1[[g]][perm, perm, drop = FALSE]
      } else {
        # note: if conditional.x, the (unconditional) unstructured b1
        # may only be combined with matching unconditional-space factors
        # (as in the yuan.bentler.mplus trace), never with the
        # conditional.x Delta; user-requested h1.information =
        # "unstructured" is blocked in lav_options_set()
        mu_w <- implied$mean[[(g - 1) * lavdata@nlevels + 1L]]
        mu_b <- implied$mean[[(g - 1) * lavdata@nlevels + 2L]]
        sigma_w <- implied$cov[[(g - 1) * lavdata@nlevels + 1L]]
        sigma_b <- implied$cov[[(g - 1) * lavdata@nlevels + 2L]]
    if (lavdata@missing %in% c("ml", "ml.x")) {
          b1[[g]] <- lav_mvn_cl_mi_info_firstorder(
            y1            = lavdata@X[[g]],
            y2            = lavsamplestats@YLp[[g]][[2]]$Y2,
            lp            = lavdata@Lp[[g]],
            mp            = lavdata@Mp[[g]],
            mu_w          = mu_w,
            sigma_w       = sigma_w,
            mu_b          = mu_b,
            sigma_b       = sigma_b,
            x_idx         = lavsamplestats@x.idx[[g]],
            divide_by_two = TRUE
          )
    } else {
      # no missing values
          b1[[g]] <- lav_mvn_cl_info_firstorder(
            y1            = lavdata@X[[g]],
            ylp           = lavsamplestats@YLp[[g]],
            lp            = lavdata@Lp[[g]],
            mu_w          = mu_w,
            sigma_w       = sigma_w,
            mu_b          = mu_b,
            sigma_b       = sigma_b,
            x_idx         = lavsamplestats@x.idx[[g]],
            divide_by_two = TRUE
          )
    }
      }
    } else if (estimator == "ML" && lavdata@nlevels == 1L) {
      if (length(lavdata@cluster) > 0L) {
        cluster_idx <- lavdata@Lp[[g]]$cluster.idx[[2]]
      } else {
        cluster_idx <- NULL
      }

      if (lavsamplestats@missing.flag && lavmodel@conditional.x) {
        # mvreg, missing data
        if (lavmodel@meanstructure && structured) {
          res_int <- lavimplied$res.int[[g]]
          res_slopes <- lavimplied$res.slopes[[g]]
        } else {
          res_int <- lavh1$implied$res.int[[g]]
          res_slopes <- lavh1$implied$res.slopes[[g]]
        }

        b1[[g]] <- lav_mvreg_mi_info_firstorder(
          y = lavdata@X[[g]],
          exo = lavdata@eXo[[g]],
          mp = lavdata@Mp[[g]], wt = wt,
          cluster_idx = cluster_idx,
          res_int = res_int,
          res_slopes = res_slopes,
          res_cov = implied$res.cov[[g]]
        )
      } else if (lavsamplestats@missing.flag) {
        # mvnorm
        # FIXME: allow for meanstructure = FALSE
        if (lavmodel@meanstructure && structured) {
          mean_1 <- lavimplied$mean[[g]]
        } else {
          #MEAN <- lavsamplestats@missing.h1[[g]]$mu
          mean_1 <- lavh1$implied$mean[[g]]
        }

        b1[[g]] <- lav_mvn_mi_info_firstorder(
          y = lavdata@X[[g]],
          mp = lavdata@Mp[[g]], wt = wt,
          cluster_idx = cluster_idx,
          mu = mean_1,
          # meanstructure = lavmodel@meanstructure,
          sigma_1 = implied$cov[[g]],
          x_idx = lavsamplestats@x.idx[[g]]
        )
      } else {
        if (lavmodel@conditional.x) {
          # mvreg
          if (lavmodel@meanstructure && structured) {
            res_int <- lavimplied$res.int[[g]]
            res_slopes <- lavimplied$res.slopes[[g]]
          } else {
            res_int <- lavsamplestats@res.int[[g]]
            res_slopes <- lavsamplestats@res.slopes[[g]]
          }

          b1[[g]] <- lav_mvreg_info_firstorder(
            y              = lavdata@X[[g]],
            exo            = lavdata@eXo[[g]],
            res_int        = res_int,
            res_slopes     = res_slopes,
            # wt            = WT,
            # meanstructure = lavmodel@meanstructure,
            res_cov        = implied$res.cov[[g]]
          )
        } else {
          # conditional.x = FALSE
          # mvnorm
          if (lavmodel@meanstructure && structured) {
            mean_1 <- lavimplied$mean[[g]]
          } else {
            # NOTE: the information matrix will be the same (minus
            # the meanstructure block), but once INVERTED, the
            # standard errors will be (slightly) smaller!!!
            # This is only visible when estimator = "MLF"
            # (or information = "first.order")
            mean_1 <- lavsamplestats@mean[[g]] # saturated
          }

          if (structured) {
            b1[[g]] <- lav_mvn_info_firstorder(
              y = lavdata@X[[g]],
              mu = mean_1, sigma_1 = lavimplied$cov[[g]],
              wt = wt,
              cluster_idx = cluster_idx,
              x_idx = lavsamplestats@x.idx[[g]],
              meanstructure = lavmodel@meanstructure
            )
          } else {
            b1[[g]] <- lav_mvn_h1_info_firstorder(
              y = lavdata@X[[g]],
              sample_cov_inv = lavsamplestats@icov[[g]],
              gamma_1 = lavsamplestats@NACOV[[g]],
              wt = wt,
              cluster_idx = cluster_idx, # only if wt
              x_idx = lavsamplestats@x.idx[[g]],
              meanstructure = lavmodel@meanstructure
            )
          }
        } # mvnorm
      } # missing
    } # ML

    # stochastic group weight
    if (lavmodel@group.w.free) {
      # unweight!!
      a <- exp(lavimplied$group.w[[g]]) / lavsamplestats@nobs[[g]]
      b1[[g]] <- lav_mat_bdiag(matrix(a, 1, 1), b1[[g]])
    }
  } # g

  b1
}


# asymptotic variance matrix (=Gamma/N) of the unrestricted (H1)
# sample statistics
#
# FIXME: make this work for categorical/GLS/WLS/...
#
lav_model_h1_acov <- function(lavobject = NULL,
                              lavmodel = NULL,
                              lavsamplestats = NULL,
                              lavdata = NULL,
                              lavoptions = NULL,
                              lavimplied = NULL,
                              lavh1 = NULL,
                              lavcache = NULL,
                              meanstructure = NULL, # if specified, use it
                              h1_information = NULL, # if specified, use it
                              se = NULL) { # if specified, use it
  lav_model_h1_unpack(lavobject, need_h1 = TRUE, need_implied = TRUE)

  # override
  if (!is.null(meanstructure)) {
    lavoptions$meanstructure <- meanstructure
  }
  if (!is.null(h1_information)) {
    lavoptions$h1.information[1] <- h1_information
  }
  if (!is.null(se)) {
    lavoptions$se <- se
  }


  # compute information matrix (dispatch on lavoptions$information[1])
  i1 <- lav_model_h1_info(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats, lavdata = lavdata,
    lavimplied = lavimplied, lavh1 = lavh1,
    lavcache = lavcache, lavoptions = lavoptions
  )

  if (lavoptions$se %in% c("robust.huber.white", "robust.sem",
                           "robust.cluster", "robust.cluster.sem")) {
    # for clustered data (robust.cluster[.sem]), the first-order kernels
    # aggregate the casewise scores within clusters (they receive
    # cluster_idx whenever lavdata@cluster is set), so the same sandwich
    # below yields the cluster-robust ACOV
    j1 <- lav_model_h1_info_firstorder(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, lavoptions = lavoptions
    )
  }

  # compute ACOV per group
  acov <- vector("list", length = lavdata@ngroups)
  for (g in 1:lavdata@ngroups) {
    # denominator
    if (lavdata@nlevels == 1L) {
      ng <- lavsamplestats@nobs[[g]]
    } else {
      ng <- lavdata@Lp[[g]]$nclusters[[2]]
    }

    # invert information
    #
    # some h1 sample statistics can be fixed/degenerate, with zero information
    # (e.g. the within-level means in a multilevel model, which are fixed at
    # zero). Those rows/columns make the full information matrix singular, so
    # we drop them before inverting and reinsert zeros afterwards (a zero
    # asymptotic variance for fixed statistics). For single-level models there
    # are no such columns and this reduces to the plain inverse.
    diag_i1 <- diag(i1[[g]])
    keep <- which(abs(diag_i1) > sqrt(.Machine$double.eps) * max(abs(diag_i1)))
    if (length(keep) < nrow(i1[[g]])) {
      i1_g_inv <- matrix(0, nrow(i1[[g]]), ncol(i1[[g]]))
      inv_keep <- try(
        lav_mat_sym_inverse(i1[[g]][keep, keep, drop = FALSE]),
        silent = TRUE
      )
      if (inherits(inv_keep, "try-error")) {
        lav_msg_stop(gettext(
          "could not invert h1 information matrix in group"), g)
      }
      i1_g_inv[keep, keep] <- inv_keep
    } else {
      i1_g_inv <- try(lav_mat_sym_inverse(i1[[g]]), silent = TRUE)
      if (inherits(i1_g_inv, "try-error")) {
        lav_msg_stop(gettext(
          "could not invert h1 information matrix in group"), g)
      }
    }

    # which type of se?
    if (lavoptions$se %in% c("standard", "none")) {
      acov[[g]] <- 1 / ng * i1_g_inv
    } else if (lavoptions$se %in% c("robust.huber.white", "robust.sem",
                                    "robust.cluster", "robust.cluster.sem")) {
      acov[[g]] <- 1 / ng * (i1_g_inv %*% j1[[g]] %*% i1_g_inv)
    } else {
      lav_msg_stop(gettextf(
        "h1 ACOV not available for se = %s.", dQuote(lavoptions$se)))
    }
  }

  acov
}

# parse lavaan syntax
# YR 14 Jan 2014: move to lav_syntax.R
# YR 17 Oct 2023: add ldw parser
# YR 23 Oct 2024: switch to "c.r"
# LDW 9 Jan 2026: remove c.r option

lavParseModelString <- function(model_syntax = "", as_data_frame = FALSE,   # nolint
                                parser = "open", warn = TRUE, debug = FALSE, ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  if (!missing(debug)) {
    current_debug <- lav_debug()
    if (lav_debug(debug))
      on.exit(lav_debug(current_debug), TRUE)
  }
  if (!missing(warn)) {
    current_warn <- lav_warn()
    if (lav_warn(warn))
      on.exit(lav_warn(current_warn), TRUE)
  }
  parser <- tolower(parser)
  parser_choices <- c("old", "new", "open")
  if (!parser %in% parser_choices) {
    lav_msg_stop(gettextf("parser= argument should be %s",
     lav_msg_view(parser_choices, "or")))
  }

  switch(parser,
  old = # original/classic parser
    lav_parse_model_string_orig(model_syntax = model_syntax,
       as_data_frame = as_data_frame),
  new = # new parser
    lav_parse_model_string(model_syntax = model_syntax,
      as_data_frame = as_data_frame),
  open = # open parser
    lav_parse_model_string_open(model_syntax = model_syntax,
      as_data_frame = as_data_frame)
  )
}

# the 'original' parser (up to 0.6-17)
lav_parse_model_string_orig <- function(model_syntax = "",
                                        as_data_frame = FALSE) {
  # check for empty syntax
  if (length(model_syntax) == 0) {
    lav_msg_stop(gettextf("lavaan ERROR: empty model syntax"))
  }

  # remove comments prior to split:
  # match from comment character to newline, but don't eliminate newline
  model_syntax <- gsub("[#!].*(?=\n)", "", model_syntax, perl = TRUE)

  # replace semicolons with newlines prior to split
  model_syntax <- gsub(";", "\n", model_syntax, fixed = TRUE)

  # remove all whitespace prior to split
  model_syntax <- gsub("[ \t]+", "", model_syntax, perl = TRUE)
  # remove any occurrence of >= 2 consecutive newlines to eliminate
  # blank statements; this retains a blank newline at the beginning,
  # if such exists, but parser will not choke because of start.idx
  model_syntax <- gsub("\n{2,}", "\n", model_syntax, perl = TRUE)

  # replace 'strange' tildes (in some locales) (new in 0.6-6)
  model_syntax <- gsub(pattern = "\u02dc", replacement = "~", model_syntax, fixed = TRUE)

  # break up in lines
  model <- unlist(strsplit(model_syntax, "\n"))

  # check for multi-line formulas: they contain no operator symbol
  # but before we do that, we remove all strings between double quotes
  # to avoid confusion with for example equal("f1=~x1") statements
  # model.simple <- gsub("\\(.*\\)\\*", "MODIFIER*", model)
  model_simple <- gsub("\\\".[^\\\"]*\\\"", "LABEL", model)

  # start.idx <- grep("[~=<>:|%]", model.simple)
  operators <- c(
    "=~", "<~", "~*~", "~~", "~", "==", "<", ">", ":=",
    ":", "\\|", "%"
  )
  lhs_modifiers <- c("efa")
  operators_extra <- c(operators, lhs_modifiers)
  start_idx <- grep(paste(operators_extra, collapse = "|"), model_simple)

  # check for empty start.idx: no operator found (new in 0.6-1)
  if (length(start_idx) == 0L) {
    lav_msg_stop(gettext("lavaan ERROR: model does not contain lavaan
                          syntax (no operators found)"))
  }

  # check for lonely lhs modifiers (only efa() for now):
  # if found, remove the following start.idx
  efa_idx <- grep("efa\\(", model_simple)
  op_idx <- grep(paste(operators, collapse = "|"), model_simple)
  both_idx <- which(efa_idx %in% op_idx)
  if (length(both_idx) > 0L) {
    efa_idx <- efa_idx[-which(efa_idx %in% op_idx)]
  }
  if (length(efa_idx) > 0L) {
    start_idx <- start_idx[-(match(efa_idx, start_idx) + 1L)]
  }

  # check for non-empty string, without an operator in the first lines
  # (new in 0.6-1)
  if (start_idx[1] > 1L) {
    # two possibilities:
    # - we have an empty line (ok)
    # - the element contains no operator (warn!)
    for (el in 1:(start_idx[1] - 1L)) {
      # not empty?
      if (nchar(model_simple[el]) > 0L) {
        lav_msg_warn(gettextf("lavaan WARNING: no operator found in this
                               syntax line: ", model_simple[el], "\n",
                              "     This syntax line will be ignored!"))
      }
    }
  }

  end_idx <- c(start_idx[-1] - 1, length(model))
  model_orig <- model
  model <- character(length(start_idx))
  for (i in seq_along(start_idx)) {
    model[i] <- paste(model_orig[start_idx[i]:end_idx[i]], collapse = "")
  }

  # ok, in all remaining lines, we should have an operator outside the ""
  model_simple <- gsub("\\\".[^\\\"]*\\\"", "LABEL", model)
  idx_wrong <- which(!grepl(
    paste(operators, collapse = "|"),
    model_simple
  ))
  # idx.wrong <- which(!grepl("[~=<>:|%]", model.simple))
  if (length(idx_wrong) > 0) {
    cat("lavaan: missing operator in formula(s):\n")
    print(model[idx_wrong])
    lav_msg_stop(gettext("lavaan ERROR: syntax error in lavaan model syntax"))
  }

  # but perhaps we have a '+' as the first character?
  idx_wrong <- grep("^\\+", model)
  if (length(idx_wrong) > 0) {
    cat("lavaan: some formula(s) start with a plus (+) sign:\n")
    print(model[idx_wrong])
    lav_msg_stop(gettext("lavaan ERROR: syntax error in lavaan model syntax"))
  }


  # main operation: flatten formulas into single bivariate pieces
  # with a left-hand-side (lhs), an operator (eg "=~"), and a
  # right-hand-side (rhs)
  # both lhs and rhs can have a modifier
  flat_lhs <- character(0)
  flat_op <- character(0)
  flat_rhs <- character(0)
  flat_rhs_mod_idx <- integer(0)
  flat_block <- integer(0) # keep track of groups using ":" operator

  flat_fixed <- character(0) # only for display purposes!
  flat_start <- character(0) # only for display purposes!
  flat_lower <- character(0) # only for display purposes!
  flat_upper <- character(0) # only for display purposes!
  flat_label <- character(0) # only for display purposes!
  flat_prior <- character(0)
  flat_efa <- character(0)
  flat_rv <- character(0)
  flat_idx <- 0L
  mod_idx_1 <- 0L
  con_idx <- 0L
  mod_1 <- vector("list", length = 0L)
  con <- vector("list", length = 0L)
  block <- 1L
  block_op <- FALSE
  for (i in seq_along(model)) {
    x <- model[i]
    if (lav_debug()) {
      cat("formula to parse:\n")
      print(x)
      cat("\n")
    }

    # 1. which operator is used?
    line_simple <- gsub("\\\".[^\\\"]*\\\"", "LABEL", x)
    # "=~" operator?
    if (grepl("=~", line_simple, fixed = TRUE)) {
      op <- "=~"
      # "<~" operator?
    } else if (grepl("<~", line_simple, fixed = TRUE)) {
      op <- "<~"
    } else if (grepl("~*~", line_simple, fixed = TRUE)) {
      op <- "~*~"
      # "~~" operator?
    } else if (grepl("~~", line_simple, fixed = TRUE)) {
      op <- "~~"
      # "~" operator?
    } else if (grepl("~", line_simple, fixed = TRUE)) {
      op <- "~"
      # "==" operator?
    } else if (grepl("==", line_simple, fixed = TRUE)) {
      op <- "=="
      # "<" operator?
    } else if (grepl("<", line_simple, fixed = TRUE)) {
      op <- "<"
      # ">" operator?
    } else if (grepl(">", line_simple, fixed = TRUE)) {
      op <- ">"
      # ":=" operator?
    } else if (grepl(":=", line_simple, fixed = TRUE)) {
      op <- ":="
      # ":" operator?
    } else if (grepl(":", line_simple, fixed = TRUE)) {
      op <- ":"
      # "|" operator?
    } else if (grepl("|", line_simple, fixed = TRUE)) {
      op <- "|"
      # "%" operator?
    } else if (grepl("%", line_simple, fixed = TRUE)) {
      op <- "%"
    } else {
      lav_msg_stop(gettext("unknown operator in ", model[i]))
    }

    # 2. split by operator (only the *first* occurrence!)
    # check first if equal/label modifier has been used on the LEFT!
    if (substr(x, 1, 6) == "label(") {
      lav_msg_stop(gettext("label modifier can not be used on the",
            "left-hand side of the operator"))
    }
    if (op == "|") {
      op_idx <- regexpr("\\|", x)
    } else if (op == "~*~") {
      op_idx <- regexpr("~\\*~", x)
    } else {
      op_idx <- regexpr(op, x)
    }
    lhs <- substr(x, 1L, op_idx - 1L)

    # right-hand side string
    rhs <- substr(x, op_idx + attr(op_idx, "match.length"), nchar(x))

    # check if first character of rhs is '+'; if so, remove silently
    # (for those who copied multiline R input from a website/pdf)
    if (substr(rhs, 1, 1) == "+") {
      rhs <- substr(rhs, 2, nchar(rhs))
    }

    # 2b. if operator is "==" or "<" or ">" or ":=", put it in CON
    if (op == "==" || op == "<" || op == ">" || op == ":=") {
      # remove quotes, if any
      lhs <- gsub("\\\"", "", lhs)
      rhs <- gsub("\\\"", "", rhs)
      con_idx <- con_idx + 1L
      con[[con_idx]] <- list(op = op, lhs = lhs, rhs = rhs, user = 1L)
      next
    }

    # 2c if operator is ":", put it in BLOCK
    if (op == ":") {
      # check if rhs is empty (new in 0.6-4)
      if (nchar(rhs) == 0L) {
        lav_msg_stop(gettextf(
          "syntax contains block identifier %s with missing number/label. The
          correct syntax is: \"LHS: RHS\", where LHS is a block identifier (eg
          group or level), and RHS is the group/level/block number or label.",
          dQuote(lhs))
        )
      }

      # check lhs (new in 0.6-4) - note: class is for nlsem
      lhs_orig <- lhs
      lhs <- tolower(lhs)
      if (!lhs %in% c("group", "level", "block", "class")) {
        lav_msg_stop(gettextf(
          "unknown block identifier: %s. Block identifier should be
           group, level or block.", dQuote(lhs_orig))
        )
      }

      flat_idx <- flat_idx + 1L
      flat_lhs[flat_idx] <- lhs
      flat_op[flat_idx] <- op
      flat_rhs[flat_idx] <- rhs
      flat_fixed[flat_idx] <- ""
      flat_start[flat_idx] <- ""
      flat_lower[flat_idx] <- ""
      flat_upper[flat_idx] <- ""
      flat_label[flat_idx] <- ""
      flat_prior[flat_idx] <- ""
      flat_efa[flat_idx] <- ""
      flat_rv[flat_idx] <- ""
      flat_rhs_mod_idx[flat_idx] <- 0L
      if (block_op) {
        block <- block + 1L
      }
      flat_block[flat_idx] <- block
      block_op <- TRUE
      next
    }

    # 3. parse left hand

    # new in 0.6-3
    # first check if all lhs names are valid (in R); see ?make.names
    # and ?reserved
    # for example, 'NA' is a reserved keyword, and should not be used
    # this usually only happens for latent variable names
    #
    # check should not come earlier, as we do not need it for :,==,<,>,:=
    lhs_1 <- strsplit(lhs, split = "+", fixed = TRUE)[[1]]
    # remove modifiers
    lhs_1 <- gsub("^\\S*\\*", "", lhs_1)
    if (!all(make.names(lhs_1) == lhs_1)) {
      lav_msg_stop(gettextf(
        "lavaan ERROR: left hand side (lhs) of this formula: %1$s %2$s %3$s
        contains either a reserved word (in R) or an illegal character: %4$s.
        See ?reserved for a list of reserved words in R. Please use a variable
        name that is not a reserved word in R and use only characters, digits,
        or the dot symbol.",
        lhs, op, rhs, dQuote(lhs_1[!make.names(lhs_1) == lhs_1])
      ))
    }

    lhs_formula <- as.formula(paste("~", lhs))
    lhs_out <- lav_syntax_parse_rhs(rhs = lhs_formula[[2L]], op = op)
    lhs_names <- names(lhs_out)


    # new in 0.6-4
    # handle LHS modifiers (if any)
    # if(sum(sapply(lhs.out, length)) > 0L) {
    # warning("lavaan WARNING: left-hand side of formula ",
    #         "below contains modifier:\n", x,"\n")
    # }

    # 4. lav_syntax_parse_rhs (as rhs of a single-sided formula)

    # new 0.5-12: before we do this, replace '0.2?' by 'start(0.2)*'
    # requested by the simsem folks
    rhs <- gsub("\\(?([-]?[0-9]*\\.?[0-9]*)\\)?\\?", "start(\\1)\\*", rhs)




    # new in 0.6-6, check for rhs NAMES that are reserved names
    # like in foo =~ in + out
    rhs_1 <- strsplit(rhs, split = "+", fixed = TRUE)[[1]]
    rhs_names <- gsub("^\\S*\\*", "", rhs_1)
    bad <- c("if", "else", "repeat", "while", "function", "for", "in")
    if (any(rhs_names %in% c(bad, "NA"))) { # "NA" added in 0.6-8
      lav_msg_stop(gettextf(
        "Right hand side (rhs) of this formula:\n    ",
        lhs, " ", op, " ", rhs,
        "\n    contains either a reserved word (in R) ",
        "or an illegal character: ",
        dQuote(rhs_names[which(rhs_names %in% bad)[1]]),
        "\n    See ?reserved for a list of reserved words in R",
        "\n    Please use a variable name that is not a reserved word in R",
        "\n    and use only characters, digits, or the dot symbol."
      ))
    }
    # new in 0.6-6, check for rhs LABELS that are reserved names
    # like in foo =~ in*bar
    rhs_1 <- strsplit(rhs, split = "+", fixed = TRUE)[[1]]
    rhs_labels <- gsub("\\*\\S*$", "", rhs_1)
    if (any(rhs_labels %in% bad)) {
      lav_msg_stop(gettextf(
        "Right hand side (rhs) of this formula:\n    ",
        lhs, " ", op, " ", rhs,
        "\n    contains either a reserved word (in R) ",
        "or an illegal character: ",
        dQuote(rhs_names[which(rhs_labels %in% bad)[1]]),
        "\n    See ?reserved for a list of reserved words in R",
        "\n    Please use a variable name that is not a reserved word in R",
        "\n    and use only characters, digits, or the dot symbol."
      ))
    }
    # new in 0.6-12: check for three-way interaction terms (which we do
    # NOT support)
    if (any(grepl(":", rhs_names))) {
      ncolon <- sapply(gregexpr(":", rhs_names), length)
      if (any(ncolon > 1L)) {
        idx <- which(ncolon > 1L)
        lav_msg_stop(gettext(
        "Three-way or higher-order interaction terms (using
multiple colons) are not supported in the lavaan syntax; please manually
construct the product terms yourself in the data.frame, give them an
appropriate name, and then you can use these interaction variables as any
other (observed) variable in the model syntax. Problematic term is: "),
          rhs_names[idx[1]]
        )
      }
    }


    rhs_formula <- as.formula(paste("~", rhs))
    out <- lav_syntax_parse_rhs(rhs = rhs_formula[[2L]], op = op)

    if (lav_debug()) print(out)

    # for each lhs element
    for (l in seq_along(lhs_names)) {
      # for each rhs element
      for (j in seq_along(out)) {
        # catch intercepts
        if (names(out)[j] == "intercept") {
          if (op == "~") {
            rhs_name <- ""
          } else {
            # either number (1), or reserved name?
            lav_msg_stop(gettextf("Right-hand side of
             formula contains an invalid variable name:\n    %s", x))
          }
        } else if (names(out)[j] == "..zero.." && op == "~") {
          rhs_name <- ""
        } else if (names(out)[j] == "..constant.." && op == "~") {
          rhs_name <- ""
        } else {
          rhs_name <- names(out)[j]
        }

        # move this 'check' to post-parse
        # if(op == "|") {
        #    th.name <- paste("t", j, sep="")
        #    if(names(out)[j] != th.name) {
        #        stop("lavaan ERROR: threshold ", j, " of variable ",
        #             sQuote(lhs.names[1]), " should be named ",
        #             sQuote(th.name), "; found ",
        #             sQuote(names(out)[j]), "\n")
        #    }
        # }

        # catch lhs = rhs and op = "=~"
        if (op == "=~" && lhs_names[l] == names(out)[j]) {
          lav_msg_stop(gettextf(
            "Latent variable `%s` can not be measured by itself",
              lhs_names[l]))
        }

        # check if we not already have this combination (in this group)
        # 1. asymmetric (=~, ~, ~1)
        if (op != "~~") {
          idx <- which(flat_lhs == lhs_names[l] &
            flat_op == op &
            flat_block == block &
            flat_rhs == rhs_name)
          if (length(idx) > 0L) {
            lav_msg_stop(gettextf(
              "Duplicate model element in: %s.", model[i]))
          }
        } else {
          # 2. symmetric (~~)
          idx <- which(flat_lhs == rhs_name &
            flat_op == "~~" &
            flat_block == block &
            flat_rhs == lhs_names[l])
          if (length(idx) > 0L) {
            lav_msg_stop(gettextf(
              "Duplicate model element in: %s.", model[i]))
          }
        }

        # check if we have a self-loop (y ~ y)
        if (op %in% c("~", "<~") && rhs_name == lhs_names[l]) {
          # stop("lavaan ERROR: lhs and rhs are the same in: ",
          #     model[i])
          # this breaks pompom package, example uSEM
          lav_msg_warn(gettextf(
            "Lhs and rhs are the same in: %s.",
            model[i]
          ))
        }


        flat_idx <- flat_idx + 1L
        flat_lhs[flat_idx] <- lhs_names[l]
        flat_op[flat_idx] <- op
        flat_rhs[flat_idx] <- rhs_name
        flat_block[flat_idx] <- block
        flat_fixed[flat_idx] <- ""
        flat_start[flat_idx] <- ""
        flat_label[flat_idx] <- ""
        flat_lower[flat_idx] <- ""
        flat_upper[flat_idx] <- ""
        flat_prior[flat_idx] <- ""
        flat_efa[flat_idx] <- ""
        flat_rv[flat_idx] <- ""

        mod <- list()
        rhs_mod <- 0L
        if (length(lhs_out[[l]]$efa) > 0L) {
          mod$efa <- lhs_out[[l]]$efa
          flat_efa[flat_idx] <- paste(mod$efa, collapse = ";")
          rhs_mod <- 1L # despite being a LHS modifier
        }
        if (length(out[[j]]$fixed) > 0L) {
          mod$fixed <- out[[j]]$fixed
          flat_fixed[flat_idx] <- paste(mod$fixed, collapse = ";")
          rhs_mod <- 1L
        }
        if (length(out[[j]]$start) > 0L) {
          mod$start <- out[[j]]$start
          flat_start[flat_idx] <- paste(mod$start, collapse = ";")
          rhs_mod <- 1L
        }
        if (length(out[[j]]$lower) > 0L) {
          mod$lower <- out[[j]]$lower
          flat_lower[flat_idx] <- paste(mod$lower, collapse = ";")
          rhs_mod <- 1L
        }
        if (length(out[[j]]$upper) > 0L) {
          mod$upper <- out[[j]]$upper
          flat_upper[flat_idx] <- paste(mod$upper, collapse = ";")
          rhs_mod <- 1L
        }
        if (length(out[[j]]$label) > 0L) {
          mod$label <- out[[j]]$label
          flat_label[flat_idx] <- paste(mod$label, collapse = ";")
          rhs_mod <- 1L
        }
        if (length(out[[j]]$rv) > 0L) {
          mod$rv <- out[[j]]$rv
          flat_rv[flat_idx] <- paste(mod$rv, collapse = ";")
          rhs_mod <- 1L
        }
        if (length(out[[j]]$prior) > 0L) {
          mod$prior <- out[[j]]$prior
          flat_prior[flat_idx] <- paste(mod$prior, collapse = ";")
          rhs_mod <- 1L
        }
        # if(op == "~1" && rhs == "0") {
        #    mod$fixed <- 0
        #    FLAT.fixed[FLAT.idx] <- paste(mod$fixed, collapse=";")
        #    rhs.mod <- 1L
        # }
        if (op == "=~" && rhs == "0") {
          mod$fixed <- 0
          flat_rhs[flat_idx] <- flat_lhs[flat_idx]
          flat_fixed[flat_idx] <- paste(mod$fixed, collapse = ";")
          rhs_mod <- 1L
        }

        flat_rhs_mod_idx[flat_idx] <- rhs_mod

        if (rhs_mod > 0L) {
          mod_idx_1 <- mod_idx_1 + 1L
          mod_1[[mod_idx_1]] <- mod
        }
      } # rhs elements
    } # lhs elements
  } # model elements

  # enumerate modifier indices
  mod_idx <- which(flat_rhs_mod_idx > 0L)
  flat_rhs_mod_idx[mod_idx] <- seq_along(mod_idx)

  flat <- list(
    lhs = flat_lhs, op = flat_op, rhs = flat_rhs,
    mod.idx = flat_rhs_mod_idx, block = flat_block,
    fixed = flat_fixed, start = flat_start,
    lower = flat_lower, upper = flat_upper,
    label = flat_label, prior = flat_prior,
    efa = flat_efa, rv = flat_rv
  )

  # change op for intercepts (for convenience only)
  int_idx <- which(flat$op == "~" & flat$rhs == "")
  if (length(int_idx) > 0L) {
    flat$op[int_idx] <- "~1"
  }

  # new in 0.6, reorder covariances here!
  flat <- lav_pt_covariance_reorder(flat)

  if (as_data_frame) {
    flat <- as.data.frame(flat, stringsAsFactors = FALSE)
  }

  # new in 0.6-4: check for 'group' within 'level'
  if (any(flat$op == ":")) {
    op_idx <- which(flat$op == ":")
    if (length(op_idx) < 2L) {
      # only 1 block identifier? this is weird -> give warning
      lav_msg_warn(gettextf(
        "Syntax contains only a single block identifier: %s",
         flat$lhs[op_idx]))
    } else {
      first_block <- flat$lhs[op_idx[1L]]
      second_block <- flat$lhs[op_idx[2L]]
      if (first_block == "level" &&
        second_block == "group") {
        lav_msg_stop(gettext(
          "Groups can not be nested within levels."))
      }
    }
  }

  attr(flat, "modifiers") <- mod_1
  attr(flat, "constraints") <- con

  flat
}

lav_syntax_parse_rhs <- function(rhs, op = "") {
  # new version YR 15 dec 2011!
  # - no 'equal' field anymore (only labels!)
  # - every modifier is evaluated
  # - unquoted labels are allowed (eg. x1 + x2 + c(v1,v2,v3)*x3)

  # fill in rhs list
  out <- list()
  repeat {
    if (length(rhs) == 1L) { # last one and only a single element
      out <- c(vector("list", 1L), out)
      name <- all.vars(rhs)
      if (length(name) > 0L) {
        names(out)[1L] <- name
      } else { # intercept or zero?
        if (as.character(rhs) == "1") {
          names(out)[1L] <- "intercept"
        } else if (as.character(rhs) == "0") {
          names(out)[1L] <- "..zero.."
          out[[1L]]$fixed <- 0
        } else {
          names(out)[1L] <- "..constant.."
          out[[1L]]$fixed <- 0
        }
      }
      break
    } else if (rhs[[1L]] == "*") { # last one, but with modifier
      out <- c(vector("list", 1L), out)
      name <- all.vars(rhs[[3L]])

      if (length(name) > 0L) { # not an intercept
        # catch interaction term
        rhs3_names <- all.names(rhs[[3L]])
        if (rhs3_names[1L] == ":") {
          if (length(name) == 1) {
            name <- paste(name[1L], ":", name[1L], sep = "")
          } else {
            name <- paste(name[1L], ":", name[2L], sep = "")
          }
        }
        names(out)[1L] <- name
      } else { # intercept
        names(out)[1L] <- "intercept"
      }
      i_var <- all.vars(rhs[[2L]], unique = FALSE)
      if (length(i_var) > 0L) {
        # modifier are unquoted labels
        out[[1L]]$label <- i_var
      } else {
        # modifier is something else
        out[[1L]] <- lav_syntax_get_modifier(rhs[[2L]])
      }
      break
    } else if (rhs[[1L]] == ":") { # last one, but interaction term
      out <- c(vector("list", 1L), out)
      name <- all.vars(rhs)
      if (length(name) == 1) {
        name <- paste(name[1L], ":", name[1L], sep = "")
      } else {
        name <- paste(name[1L], ":", name[2L], sep = "")
      }
      names(out)[1L] <- name
      break
    } else if (rhs[[1L]] == "+") { # not last one!

      # three possibilities:
      # 1. length(rhs[[3]] == 3), and rhs[[3L]][[1]] == "*" -> modifier
      # 2. length(rhs[[3]] == 3), and rhs[[3L]][[1]] == ":" -> interaction
      # 3. length(rhs[[3]] == 1) -> single element

      out <- c(vector("list", 1L), out)

      # modifier or not?
      if (length(rhs[[3L]]) == 3L && rhs[[3L]][[1]] == "*") {
        # modifier!!
        name <- all.vars(rhs[[3L]][[3]])

        if (length(name) > 0L) { # not an intercept
          # catch interaction term
          rhs3_names <- all.names(rhs[[3L]][[3]])
          if (rhs3_names[1L] == ":") {
            if (length(name) == 1) {
              name <- paste(name[1L], ":", name[1L], sep = "")
            } else {
              name <- paste(name[1L], ":", name[2L], sep = "")
            }
          }
          names(out)[1L] <- name
        } else { # intercept
          names(out)[1L] <- "intercept"
        }
        i_var <- all.vars(rhs[[3]][[2L]], unique = FALSE)
        if (length(i_var) > 0L) {
          # modifier are unquoted labels
          out[[1L]]$label <- i_var
        } else {
          # modifier is something else
          out[[1L]] <- lav_syntax_get_modifier(rhs[[3]][[2L]])
        }

        # interaction term?
      } else if (length(rhs[[3L]]) == 3L && rhs[[3L]][[1]] == ":") {
        # interaction term, without modifier
        name <- all.vars(rhs[[3L]])
        if (length(name) == 1) {
          name <- paste(name[1L], ":", name[1L], sep = "")
        } else {
          name <- paste(name[1L], ":", name[2L], sep = "")
        }
        names(out)[1L] <- name
      } else { # no modifier!!
        name <- all.vars(rhs[[3]])
        if (length(name) > 0L) {
          names(out)[1L] <- name
        } else { # intercept or zero?
          if (as.character(rhs[[3]]) == "1") {
            names(out)[1L] <- "intercept"
          } else if (as.character(rhs[[3]]) == "0") {
            names(out)[1L] <- "..zero.."
            out[[1L]]$fixed <- 0
          } else {
            names(out)[1L] <- "..constant.."
            out[[1L]]$fixed <- 0
          }
        }
      }


      # next element
      rhs <- rhs[[2L]]
    } else {
      lav_msg_stop(gettextf(
        "lavaan ERROR: I'm confused parsing this line: ", rhs, "\n"))
    }
  }

  # if multiple elements, check for duplicated elements and merge if found
  if (length(out) > 1L) {
    rhs_names <- names(out)
    while (!is.na(idx <- which(duplicated(rhs_names))[1L])) {
      dup_name <- rhs_names[idx]
      orig_idx <- match(dup_name, rhs_names)
      merged <- c(out[[orig_idx]], out[[idx]])
      if (!is.null(merged)) { # be careful, NULL will delete element
        out[[orig_idx]] <- merged
      }
      out <- out[-idx]
      rhs_names <- names(out)
    }
  }

  # if thresholds, check order and reorder if necessary
  # if(op == "|") {
  #    t.names <- names(out)
  #    idx <- match(sort(t.names), t.names)
  #    out <- out[idx]
  # }

  out
}


lav_syntax_get_modifier <- function(mod) {
  if (length(mod) == 1L) {
    # three possibilities: 1) numeric, 2) NA, or 3) quoted character
    if (is.numeric(mod)) {
      return(list(fixed = mod))
    }
    if (is.na(mod)) {
      return(list(fixed = as.numeric(NA)))
    }
    if (is.character(mod)) {
      list(label = mod)
    }
  } else if (mod[[1L]] == "start") {
    cof <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    list(start = cof)
  } else if (mod[[1L]] == "lower") {
    cof <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    list(lower = cof)
  } else if (mod[[1L]] == "upper") {
    cof <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    list(upper = cof)
  } else if (mod[[1L]] == "equal") {
    label <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    list(label = label)
  } else if (mod[[1L]] == "label") {
    label <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    label[is.na(label)] <- "" # catch 'NA' elements in a label
    list(label = label)
  } else if (mod[[1L]] == "rv") {
    rv <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    if (anyNA(rv)) {
      lav_msg_stop(gettextf("lavaan ERROR: some rv() labels are NA"))
    }
    list(rv = rv)
  } else if (mod[[1L]] == "prior") {
    prior <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    list(prior = prior)
  } else if (mod[[1L]] == "efa") {
    efa <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    list(efa = efa)
  } else if (mod[[1L]] == "c") {
    # vector: we allow numeric and character only!
    cof <- unlist(lapply(as.list(mod)[-1],
      eval,
      envir = NULL, enclos = NULL
    ))
    if (all(is.na(cof))) {
      list(fixed = rep(as.numeric(NA), length(cof)))
    } else if (is.numeric(cof)) {
      list(fixed = cof)
    } else if (is.character(cof)) {
      cof[is.na(cof)] <- "" # catch 'NA' elements in a label
      list(label = cof)
    } else {
      lav_msg_stop(gettextf("lavaan ERROR: can not parse modifier:", mod, "\n"))
    }
  } else {
    # unknown expression
    # as a final attempt, we will evaluate it and coerce it
    # to either a numeric or character (vector)
    cof <- try(eval(mod, envir = NULL, enclos = NULL), silent = TRUE)
    if (inherits(cof, "try-error")) {
      lav_msg_stop(gettextf(
        "lavaan ERROR: evaluating modifier failed: ",
        paste(as.character(mod)[[1]], "()*", sep = ""), "\n"
      ))
    } else if (is.numeric(cof)) {
      list(fixed = cof)
    } else if (is.character(cof)) {
      list(label = cof)
    } else {
      lav_msg_stop(gettextf(
        "lavaan ERROR: can not parse modifier: ",
        paste(as.character(mod)[[1]], "()*", sep = ""), "\n"
      ))
    }
  }
}

# print object from lavData class
#

setMethod(
  "show", "lavData",
  function(object) {
    # print 'lavData' object
    res <- lav_data_summary_short(object)
    lav_data_print_short(res, nd = 3L)
  }
)

# create summary information for @lavdata slot
lav_data_summary_short <- function(object) {
  # which object?
  if (inherits(object, "lavaan")) {
    # check object
    object <- lav_object_check_version(object)
    lavdata <- object@Data
  } else if (inherits(object, "lavData")) {
    lavdata <- object
  } else {
    lav_msg_stop(gettext("object must be lavaan or lavData object"))
  }

  # two or three columns (depends on nobs/norig)
  threecolumn <- FALSE
  for (g in 1:lavdata@ngroups) {
    if (lavdata@nobs[[g]] != lavdata@norig[[g]]) {
      threecolumn <- TRUE
      break
    }
  }

  # clustered data?
  clustered <- length(lavdata@cluster) > 0L

  # multilevel data?
  multilevel <- lavdata@nlevels > 1L

  # extract summary information
  datasummary <- list(
    ngroups = lavdata@ngroups,
    nobs = unlist(lavdata@nobs)
  )

  # norig?
  if (threecolumn) {
    datasummary$norig <- unlist(lavdata@norig)
  }

  # multiple groups?
  if (lavdata@ngroups > 1L) {
    datasummary$group.label <- lavdata@group.label
  }

  # sampling weights?
  if (!is.null(lavdata@weights[[1L]])) {
    datasummary$sampling.weights <- lavdata@sampling.weights
  }

  # clustered/multilevel data?
  if (clustered) {
    if (multilevel) {
      datasummary$nlevels <- lavdata@nlevels
    }
    datasummary$cluster <- lavdata@cluster

    if (lavdata@ngroups == 1L) {
      datasummary$nclusters <- unlist(lavdata@Lp[[1]]$nclusters)
    } else {
      tmp <- vector("list", length = lavdata@ngroups)
      for (g in seq_len(lavdata@ngroups)) {
        tmp[[g]] <- unlist(lavdata@Lp[[g]]$nclusters)
      }
      datasummary$nclusters <- tmp
    }
  }

  # missing data?
  if (!is.null(lavdata@Mp[[1L]])) {
    datasummary$npatterns <- sapply(lavdata@Mp, "[[", "npatterns")
    if (multilevel && !is.null(lavdata@Mp[[1L]]$Zp)) {
      datasummary$npatterns2 <- sapply(lapply(
        lavdata@Mp,
        "[[", "Zp"
      ), "[[", "npatterns")
    }
  }

  datasummary
}

lav_data_print_short <- function(object, nd = 3L) {
  # object should data summary
  if (inherits(object, "lavaan")) {
    object <- lav_data_summary_short(object)
  }
  datasummary <- object

  # num_format <- paste("%", max(8L, nd + 5L), ".", nd, "f", sep = "")

  # threecolumn
  threecolumn <- !is.null(datasummary$norig)

  # multilevel?
  multilevel <- !is.null(datasummary$nlevels)

  # clustered?
  clustered <- !is.null(datasummary$cluster) && is.null(datasummary$nlevels)

  # header? no, for historical reasons only
  # cat("Data information:\n\n")

  c1 <- c2 <- c3 <- character(0L)

  # number of observations
  if (datasummary$ngroups == 1L) {
    if (threecolumn) {
      c1 <- c(c1, "")
      c2 <- c(c2, "Used")
      c3 <- c(c3, "Total")
    }
    c1 <- c(c1, "Number of observations")
    c2 <- c(c2, datasummary$nobs)
    c3 <- c(c3, ifelse(threecolumn, datasummary$norig, ""))
  } else {
    c1 <- c(c1, "Number of observations per group:")
    if (threecolumn) {
      c2 <- c(c2, "Used")
      c3 <- c(c3, "Total")
    } else {
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }
    for (g in 1:datasummary$ngroups) {
      c1 <- c(c1, sprintf("  %-40s", datasummary$group.label[g]))
      c2 <- c(c2, datasummary$nobs[g])
      c3 <- c(c3, ifelse(threecolumn, datasummary$norig[g], ""))
    } # g
  }

  # number of clusters
  if (datasummary$ngroups == 1L) {
    if (multilevel) {
      for (l in 2:datasummary$nlevels) {
        c1 <- c(
          c1,
          paste("Number of clusters [",
            datasummary$cluster[l - 1], "]",
            sep = ""
          )
        )
        c2 <- c(c2, datasummary$nclusters[l])
        c3 <- c(c3, "")
      }
    } else if (clustered) {
      c1 <- c(c1, paste("Number of clusters [", datasummary$cluster, "]",
        sep = ""
      ))
      c2 <- c(c2, datasummary$nclusters[2])
      c3 <- c(c3, "")
    }
  } else {
    if (multilevel) {
      for (l in 2:datasummary$nlevels) {
        c1 <- c(
          c1,
          paste("Number of clusters [", datasummary$cluster[l - 1], "]:",
            sep = ""
          )
        )
        c2 <- c(c2, "")
        c3 <- c(c3, "")
        for (g in 1:datasummary$ngroups) {
          c1 <- c(c1, sprintf("  %-40s", datasummary$group.label[g]))
          c2 <- c(c2, datasummary$nclusters[[g]][l])
          c3 <- c(c3, "")
        }
      }
    } else if (clustered) {
      c1 <- c(
        c1,
        paste("Number of clusters [", datasummary$cluster, "]:", sep = "")
      )
      c2 <- c(c2, "")
      c3 <- c(c3, "")
      for (g in 1:datasummary$ngroups) {
        c1 <- c(c1, sprintf("  %-40s", datasummary$group.label[g]))
        c2 <- c(c2, datasummary$nclusters[[g]][2])
        c3 <- c(c3, "")
      }
    }
  }

  # missing patterns?
  if (!is.null(datasummary$npatterns)) {
    if (datasummary$ngroups == 1L) {
      if (multilevel) {
        c1 <- c(c1, "Number of missing patterns -- level 1")
        c2 <- c(c2, datasummary$npatterns)
        c3 <- c(c3, "")
        if (!is.null(datasummary$npatterns2)) {
          c1 <- c(c1, "Number of missing patterns -- level 2")
          c2 <- c(c2, datasummary$npatterns2)
          c3 <- c(c3, "")
        }
      } else {
        c1 <- c(c1, "Number of missing patterns")
        c2 <- c(c2, datasummary$npatterns)
        c3 <- c(c3, "")
      }
    } else {
      if (multilevel) {
        c1 <- c(c1, "Number of missing patterns per group:")
        c2 <- c(c2, "")
        c3 <- c(c3, "")
        for (g in 1:datasummary$ngroups) {
          c1 <- c(
            c1,
            paste(sprintf(
              "  %-40s",
              datasummary$group.label[g]
            ), "-- level 1")
          )
          c2 <- c(c2, datasummary$npatterns[g])
          c3 <- c(c3, "")
          if (!is.null(datasummary$npatterns2)) {
            c1 <- c(
              c1,
              paste(sprintf(
                "  %-40s",
                datasummary$group.label[g]
              ), "-- level 2")
            )
            c2 <- c(c2, datasummary$npatterns2[g])
            c3 <- c(c3, "")
          }
        }
      } else {
        c1 <- c(c1, "Number of missing patterns per group:")
        c2 <- c(c2, "")
        c3 <- c(c3, "")
        for (g in 1:datasummary$ngroups) {
          c1 <- c(c1, sprintf("  %-40s", datasummary$group.label[g]))
          c2 <- c(c2, datasummary$npatterns[g])
          c3 <- c(c3, "")
        }
      }
    }
  }

  # sampling weights?
  if (!is.null(datasummary$sampling.weights)) {
    c1 <- c(c1, "Sampling weights variable")
    c2 <- c(c2, datasummary$sampling.weights)
    c3 <- c(c3, "")
  }

  # format c1/c2
  c1 <- format(c1, width = 43L)
  c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
  c3 <- format(c3, width = 8L + nd, justify = "right")

  # create character matrix
  if (threecolumn) {
    m <- cbind(c1, c2, c3, deparse.level = 0)
  } else {
    m <- cbind(c1, c2, deparse.level = 0)
  }
  colnames(m) <- rep("", ncol(m))
  rownames(m) <- rep(" ", nrow(m))

  # print
  write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)

  invisible(m)
}

# Browne's residual test statistic
# see Browne (1984) eq 2.20a

# T_B = (N-1) * t(RES) %*% Delta.c %*%
#                          solve(t(Delta.c) %*% Gamma %*% Delta.c) %*%
#                          t(Delta.c) %*% RES
#
#     = (N-1) * t(RES) %*% (Gamma.inv -
#                           Gamma.inv %*% Delta %*%
#                                  solve(t(Delta) %*% Gamma.inv %*% Delta) %*%
#                           t(Delta) %*% Gamma.inv) %*% RES

# Note: if Gamma == solve(Weight matrix), then:
# t(Delta) %*% solve(Gamma) %*% RES == 0-vector!
#
# Therefore:
# - if estimator = "WLS", X2 == Browne's residual ADF statistic
# - if estimator = "GLS", X2 == Browne's residual NT statistic
#
# - if estimator = "NTRLS", X2 == Browne's residual NT statistic (model-based)
#   also known as the RLS test statistic
#   ... except in multigroup + equality constraints, where
#   t(Delta) %*% solve(Gamma) %*% RES not zero everywhere!?

# YR 26 July 2022: add alternative slots, if lavobject = NULL
# YR 22 Jan  2023: allow for model-based 'structured' Sigma

# TODo: - allow for non-linear equality constraints
#         (see Browne, 1982, eq 1.7.19; although we may face singular matrices)

lav_test_browne <- function(lavobject = NULL,
                            # or
                            lavdata = NULL,
                            lavsamplestats = NULL, # WLS.obs, NACOV
                            lavmodel = NULL,
                            lavpartable = NULL, # DF
                            lavoptions = NULL,
                            lavh1 = NULL,
                            lavimplied = NULL,
                            # further options:
                            n_minus_one = "default",
                            adf = TRUE,
                            model_based = FALSE) {
  if (!is.null(lavobject)) {
    # check input
    if (!inherits(lavobject, "lavaan")) {
      lav_msg_stop(gettext("object is not a lavaan object."))
    }

    # slots
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
    lavmodel <- lavobject@Model
    lavpartable <- lavobject@ParTable
    lavoptions <- lavobject@Options
    lavh1 <- lavobject@h1
    lavimplied <- lavobject@implied
  }

  if (lavmodel@categorical) {
    if (!adf) {
      lav_msg_warn(gettext("normal theory version of Browne's residual test
        not available in the categorical setting; switching to ADF."))
      adf <- TRUE
    }
    if (model_based) {
      lav_msg_stop(gettext("model-based version of Browne's residual test not
        available in the categorical setting; switching to sample-based."))
      model_based <- FALSE
    }
  }
  if (lavdata@missing != "listwise" && !model_based) {
    lav_msg_stop(gettext("sample-based version of Browne's test is not
      available when data is missing; switching to model-based."))
    model_based <- TRUE
  }
  if (lavdata@nlevels > 1L) {
    lav_msg_stop(gettext("Browne's test is not available when data is
                         multilevel."))
  }
  if (length(lavmodel@ceq.nonlinear.idx) > 0L) {
    lav_msg_stop(gettext("Browne's test is not available (yet) when nonlinear
                         equality constraints are involved."))
  }

  if (!is.logical(n_minus_one)) {
    if (lavoptions$estimator == "ML" &&
      lavoptions$likelihood == "normal") {
      n_minus_one <- FALSE
    } else {
      n_minus_one <- TRUE
    }
  }

  # linear equality constraints? NOTE: do not test the packing flags
  # (@eq.constraints / @ceq.simple.only) here -- they are both FALSE when
  # equality constraints coexist with inequality constraints or bounds,
  # and the constraints would silently be dropped from the projection
  eq_basis <- lav_con_eq_basis(lavmodel)
  lineq_flag <- !is.null(eq_basis)

  # can we use the fast version?
  fast_flag <- FALSE
  if (!adf && !lineq_flag && !lavmodel@conditional.x &&
      !lavmodel@group.w.free) {
    fast_flag <- TRUE
    if (model_based) {
      implied <- lavimplied
    } else {
      implied <- lavh1$implied
    }
  }

  # ingredients
  delta <- lav_model_delta(lavmodel)
  if (adf) {
    # ADF version
    # note: the stored NACOV can only be used for the sample-based
    # version; the model-based version must recompute Gamma at the
    # model-implied moments (before 0.7-1, browne.residual.adf.model
    # silently returned the sample-based statistic whenever a NACOV
    # was stored, e.g. for the (D)WLS-family estimators)
    if (!model_based &&
        !is.null(lavsamplestats@NACOV[[1]]) &&
        !lavoptions$se == "robust.sem.nt") { # eg, estimator = "ULS"/"DWLS"
      gamma_1 <- lavsamplestats@NACOV
    } else {
      if (!is.null(lavobject)) {
        if (lavobject@Data@data.type != "full") {
          lav_msg_stop(gettext("ADF version not available without full data or
                               user-provided Gamma/NACOV matrix"))
        }
        gamma_1 <- lav_object_gamma(lavobject,
          adf = TRUE,
          model_based = model_based
        )
      } else {
        if (lavdata@data.type != "full") {
          lav_msg_stop(gettext("ADF version not available without full data or
                               user-provided Gamma/NACOV matrix"))
        }
        gamma_1 <- lav_object_gamma(
          lavobject = NULL,
          lavdata = lavdata,
          lavoptions = lavoptions,
          lavsamplestats = lavsamplestats,
          lavh1 = lavh1,
          lavimplied = lavimplied,
          adf = TRUE,
          model_based = model_based
        )
      }
    }
  } else if (!fast_flag) {
    # NT version
    if (!is.null(lavobject)) {
      gamma_1 <- lav_object_gamma(lavobject,
        adf = FALSE,
        model_based = model_based
      )
    } else {
      gamma_1 <- lav_object_gamma(
        lavobject = NULL,
        lavdata = lavdata,
        lavoptions = lavoptions,
        lavsamplestats = lavsamplestats,
        lavh1 = lavh1,
        lavimplied = lavimplied,
        adf = FALSE,
        model_based = model_based
      )
    }
  }
  wls_obs <- lavsamplestats@WLS.obs
  wls_est <- lav_model_wls_est(lavmodel)
  nobs <- lavsamplestats@nobs
  ntotal <- lavsamplestats@ntotal

  # compute T.B per group
  ngroups <- length(wls_obs)
  stat_group <- numeric(ngroups)

  # 1. standard setting: no equality constraints
  if (!lineq_flag) {
    for (g in seq_len(ngroups)) {
      res <- wls_obs[[g]] - wls_est[[g]]
      delta_g <- delta[[g]]
      if (n_minus_one) {
        ng <- nobs[[g]] - 1L
      } else {
        ng <- nobs[[g]]
      }
      if (fast_flag) {
        stat_group[g] <- lav_test_browne_nt_fast(
          res = res, delta = delta_g,
          sample_cov = implied$cov[[g]], sample_nobs = ng,
          meanstructure = lavmodel@meanstructure,
          x_idx = lavsamplestats@x.idx[[g]]
        )
      } else {
        # naive formula base computation (slow!)
        delta_c <- lav_mat_ortho_complement(delta_g)
        t_dgd <- crossprod(delta_c, gamma_1[[g]]) %*% delta_c
        # if fixed.x = TRUE, gamma_1[[g]] may contain zero col/rows
        t_dgd_inv <- lav_mat_sym_inverse(t_dgd)
        t_res_delta_c <- crossprod(res, delta_c)
        stat_group[g] <-
          ng * drop(t_res_delta_c %*% t_dgd_inv %*% t(t_res_delta_c))
      }
    }
    stat <- sum(stat_group)

    # 2. linear equality constraint
  } else if (lineq_flag) {
    res_all <- do.call("c", wls_obs) - do.call("c", wls_est)
    delta_all <- do.call("rbind", delta)
    delta_g <- delta_all %*% eq_basis
    gamma_inv_weighted <- vector("list", ngroups)
    for (g in seq_len(ngroups)) {
      if (n_minus_one) {
        ng <- nobs[[g]] - 1L
      } else {
        ng <- nobs[[g]]
      }
      gamma_inv_temp <- try(solve(gamma_1[[g]]), silent = TRUE)
      if (inherits(gamma_inv_temp, "try-error")) {
        # TDJ: This will happen whenever an (otherwise) unrestricted
        #      covariance matrix has a structure to it, such as equal
        #      variances (and certain covariances) for 2 members of an
        #      indistinguishable dyad (represented as 2 columns).  In
        #      such cases, their (N)ACOV elements are also identical.
        gamma_inv_temp <- MASS::ginv(gamma_1[[g]])
      }
      gamma_inv_weighted[[g]] <- gamma_inv_temp * ng / ntotal
    }
    gi <- lav_mat_bdiag(gamma_inv_weighted)
    t_dgi_d <- t(delta_g) %*% gi %*% delta_g
    t_dgi_d_inv <- MASS::ginv(t_dgi_d) # GI may be rank-deficient
    q1 <- drop(t(res_all) %*% gi %*% res_all)
    q2 <- drop(t(res_all) %*%
      gi %*% delta_g %*% t_dgi_d_inv %*% t(delta_g) %*% gi %*%
      res_all)
    stat <- ntotal * (q1 - q2)
    stat_group <- stat * unlist(nobs) / ntotal # proxy only
  }
  # (nonlinear equality constraints were caught at the top)


  # DF
  if (!is.null(lavobject)) {
    df_1 <- lavobject@test[[1]]$df
  } else {
    # same approach as in lav_test.R
    df_1 <- lav_test_df(lavpartable = lavpartable, lavmodel = lavmodel)
  }

  if (adf) {
    if (model_based) {
      # using model-based Gamma
      name <- "browne.residual.adf.model"
      label <- "Browne's residual (ADF model-based) test"
    } else {
      # regular one
      name <- "browne.residual.adf"
      label <- "Browne's residual-based (ADF) test"
    }
  } else {
    if (model_based) {
      # using model-implied Sigma (instead of S)
      # also called the 'reweighted least-squares (RLS)' version
      name <- "browne.residual.nt.model"
      label <- "Browne's residual (NT model-based) test"
    } else {
      # regular one
      name <- "browne.residual.nt"
      label <- "Browne's residual-based (NT) test"
    }
  }
  out <- list(
    test = name,
    stat = stat,
    stat.group = stat_group,
    df = df_1,
    refdistr = "chisq",
    pvalue = 1 - pchisq(stat, df_1),
    label = label
  )
  out
}


# faster version for the NT setting with no constraints
lav_test_browne_nt_fast <- function(res = NULL, delta = NULL,
                                    sample_cov = NULL, sample_nobs = NULL,
                                    meanstructure = FALSE,
                                    x_idx = integer(0L)) {
  nvar <- nrow(sample_cov)
  pstar <- nvar * (nvar + 1L) / 2L
  q <- ncol(delta)
  nrow_expected <- if (meanstructure) nvar + pstar else pstar
  stopifnot(length(res) == nrow_expected, nrow(delta) == nrow_expected)

  # only once
  s_inv <- solve(sample_cov)

  # handle fixed.x = TRUE
  if (length(x_idx) > 0L) {
    s_inv[x_idx, ] <- 0.0
    s_inv[, x_idx] <- 0.0
  }

  # vech
  diag_idx <- lav_mat_diagh_idx(nvar)

  # applies gamma_cov_inv = 0.5 * t(D) %*% (S.inv %x% S.inv) %*% D
  # to a pstar-vector 'x'
  apply_gamma_inv <- function(x) {
    x_cov <- x
    if (meanstructure) {
      x_mean <- x[seq_len(nvar)]
      x_cov <- x[nvar + seq_len(pstar)]
      out_mean <- drop(s_inv %*% x_mean)
    }

    m_w <- lav_mat_vech_rev(x_cov)
    z <- s_inv %*% m_w %*% s_inv
    out_cov <- lav_mat_vech(z)
    out_cov[diag_idx] <- out_cov[diag_idx] / 2.0
    out <- out_cov
    if (meanstructure) {
      out <- c(out_mean, out_cov)
    }
    out
  }

  u <- apply_gamma_inv(res)
  gamma_inv_delta <- matrix(0.0, nrow_expected, q)
  for (j in seq_len(q)) {
    gamma_inv_delta[, j] <- apply_gamma_inv(delta[, j])
  }
  a <- crossprod(delta, gamma_inv_delta)
  b <- crossprod(delta, u)
  ab_inv_b <- tryCatch(
    {
      r_chol <- chol(a)
      backsolve(r_chol, forwardsolve(t(r_chol), b))
    },
    error = function(e) {
      MASS::ginv(a) %*% b
    }
  )

  term1 <- as.numeric(crossprod(res, u)) # t(res) Ginv res
  term2 <- as.numeric(crossprod(b, ab_inv_b)) # t(b) A^{-1} b

  stat <- sample_nobs * (term1 - term2)
  stat
}

lav_pml_plrt <- function(lavobject = NULL, lavmodel = NULL, lavdata = NULL,
                         lavsamplestats = NULL, lavpartable = NULL,
                         lavoptions = NULL, x = NULL, vcov_1 = NULL,
                         lavcache = NULL) {
  lavpta <- NULL
  if (!is.null(lavobject)) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavoptions <- lavobject@Options
    lavsamplestats <- lavobject@SampleStats
    lavcache <- lavobject@Cache
    lavpartable <- lav_pt_set_cache(lavobject@ParTable, lavobject@pta)
    lavpta <- lavobject@pta
  }
  if (is.null(lavpta)) {
    lavpta <- lav_pt_attributes(lavpartable)
    lavpartable <- lav_pt_set_cache(lavpartable, lavpta)
  }

  if (is.null(x)) {
    # compute 'fx' = objective function value
    # (NOTE: since 0.5-18, NOT divided by N!!)
    fx <- lav_model_objective(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache
    )
    h0_fx <- as.numeric(fx)
    h0_fx_group <- attr(fx, "fx.group")
  } else {
    h0_fx <- attr(attr(x, "fx"), "fx.pml")
    h0_fx_group <- attr(attr(x, "fx"), "fx.group")
  }

  # fit a saturated model 'fittedSat'
  model_sat <- lav_pt_unrestricted(
    lavobject = NULL,
    lavdata = lavdata,
    lavoptions = lavoptions,
    lavpta = lavpta,
    lavsamplestats = lavsamplestats
  )

  # FIXME: se="none", test="none"??
  options_1 <- lavoptions
  options_1$se <- "none"
  options_1$test <- "none"
  options_1$baseline <- FALSE
  options_1$h1 <- FALSE
  fitted_sat <- lavaan(model_sat,
    slot_options = options_1, verbose = FALSE,
    slot_sample_stats = lavsamplestats,
    slot_data = lavdata, slot_cache = lavcache
  )
  fx <- lav_model_objective(
    lavmodel = fitted_sat@Model,
    lavsamplestats = fitted_sat@SampleStats,
    lavdata = fitted_sat@Data,
    lavcache = fitted_sat@Cache
  )
  sat_fx <- as.numeric(fx)
  sat_fx_group <- attr(fx, "fx.group")

  # we also need a `saturated model', but where the moments are based
  # on the model-implied sample statistics under H0
  model_sat2 <-
    lav_pt_unrestricted(
      lavobject = NULL,
      lavdata = lavdata,
      lavoptions = lavoptions,
      lavsamplestats = NULL,
      sample_cov = lav_model_sigma(lavmodel),
      sample_mean = lav_model_mu(lavmodel),
      sample_th = lav_model_th(lavmodel),
      sample_th_idx = lavsamplestats@th.idx
    )

  options2 <- options_1
  options2$optim.method <- "none"
  options2$optim.force.converged <- TRUE
  options2$check.start <- FALSE
  options2$check.gradient <- FALSE
  options2$check.post <- FALSE
  options2$check.vcov <- FALSE
  fitted_sat2 <- lavaan(model_sat2,
    slot_options = options2, verbose = FALSE,
    slot_sample_stats = lavsamplestats,
    slot_data = lavdata, slot_cache = lavcache
  )

  # the code below was contributed by Myrsini Katsikatsou (Jan 2015)

  # for now, only a single group is supported:
  # g = 1L


  ########################### The code for PLRT for overall goodness of fit

  # First define the number of non-redundant elements of the (fitted)
  # covariance/correlation matrix of the underlying variables.
  # nvar <- lavmodel@nvar[[g]]
  # dSat <- nvar*(nvar-1)/2
  # if(length(lavmodel@num.idx[[g]]) > 0L) {
  #    dSat <- dSat + length(lavmodel@num.idx[[g]])
  # }

  # select `free' parameters (excluding thresholds) from fittedSat2 model
  pt_sat2 <- fitted_sat2@ParTable
  d_sat_idx <- pt_sat2$free[pt_sat2$free > 0L & pt_sat2$op != "|"]
                                                        # remove thresholds

  # Secondly, we need to specify the indices of the rows/columns of vcov(),
  # hessian, and variability matrix that refer to all SEM parameters
  # except thresholds.
  pt_1 <- lavpartable
  index_par <- pt_1$free[pt_1$free > 0L & pt_1$op != "|"]

  # Thirdly, specify the sample size.
  # nsize <- lavdata@nobs[[g]]
  nsize <- lavsamplestats@ntotal #TODO: check if var nsize needed ??

  # Now we can proceed to the computation of the quantities needed for PLRT.
  # Briefly, to say that PLRT is equal to the difference of two quadratic forms.
  # To compute the first and second moment adjusted PLRT we should compute
  # the asymptotic mean and variance of each quadratic quantity as well as
  # their asymptotic covariance.

  ##### Section 1. Compute the asymptotic mean and variance
  #####            of the first quadratic quantity
  # Below I assume that lavobject is the output of lavaan function. I guess
  # vcov(lavobject) can be substituted by VCOV object inside lavaan function
  # defined at lines 703 -708. But what is the object inside lavaan function
  # for getHessian(lavobject)?
  if (is.null(vcov_1)) {
    lavoptions$se <- "robust.huber.white"
    vcov_1 <- lav_model_vcov(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavoptions = lavoptions,
      lavdata = lavdata,
      lavpartable = lavpartable,
      lavcache = lavcache
    )
  }
  inv_g_to_psipsi_attheta0 <- (lavsamplestats@ntotal * vcov_1)[index_par,
                             index_par, drop = FALSE] # G^psipsi(theta0)
  # below the lavaan function getHessian is used
  # Hattheta0 <- (-1) * H0.Hessian
  # Hattheta0 <- H0.Hessian
  # InvHattheta0 <- solve(Hattheta0)
  inv_hattheta0 <- attr(vcov_1, "E.inv")
  if (is.null(inv_hattheta0)) {
    # lav_model_vcov() only attaches "E.inv" when the vcov and test
    # information options coincide (they do not, e.g., with cluster =
    # where information = c("observed", "expected")); recompute it here.
    # The PLRT 'H' matrix is the observed (Hessian-based) information --
    # the flavor that E.inv holds in the non-clustered case
    h0_options <- lavoptions
    h0_options$information <- c("observed", "observed")
    inv_hattheta0 <- lav_model_info(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavcache = lavcache,
      lavoptions = h0_options,
      extra = FALSE,
      augmented = TRUE,
      inverted = TRUE,
      use_ginv = TRUE
    )
  }
  inv_h_to_psipsi_attheta0 <- inv_hattheta0[index_par, index_par, drop = FALSE]
                                                           # H^psipsi(theta0)
  # with equality constraints, inv_hattheta0 is the *constrained*
  # (augmented) inverse, which is singular along the constrained
  # directions, so its submatrix must be ginv-inverted (the ginv(M) %*% M
  # projection then restricts the trace to the constrained tangent space).
  # NOTE: do not gate this on @eq.constraints alone -- that packing flag is
  # FALSE when equality constraints coexist with inequality constraints or
  # bounds, and solve() below would then fail on a singular matrix
  has_ceq <- lavmodel@eq.constraints || lavmodel@ceq.simple.only ||
    nrow(lavmodel@ceq.JAC) > 0L
  if (length(index_par) == 0L) {
    # eg independence model (YR 26 June 2018)
    inv_of_inv_h_to_psipsi_attheta0 <- matrix(0, 0, 0)     # nolint
  } else if (has_ceq) {
    inv_of_inv_h_to_psipsi_attheta0 <-                     # nolint
      MASS::ginv(inv_h_to_psipsi_attheta0)
  } else {
    inv_of_inv_h_to_psipsi_attheta0 <-                     # nolint
      solve(inv_h_to_psipsi_attheta0) # [H^psipsi(theta0)]^(-1)
  }

  h0tmp_prod1 <- inv_of_inv_h_to_psipsi_attheta0 %*% inv_g_to_psipsi_attheta0
  h0tmp_prod2 <- h0tmp_prod1 %*% h0tmp_prod1
  e_tww <- sum(diag(h0tmp_prod1)) # expected mean of first quadratic quantity
  var_tww <- 2 * sum(diag(h0tmp_prod2)) # variance of first quadratic quantity

  ##### Section 2: Compute the asymptotic mean and variance
  #####            of the second quadratic quantity.
  # Now we need to evaluate the fitted (polychoric) correlation/ covariance
  # matrix using the estimates of SEM parameters derived under the fitted model
  # which is the model of the null hypothesis. We also need to compute the
  # vcov matrix of these estimates (estimates of polychoric correlations)
  # as well as the related hessian and variability matrix.
  tmp_options <- fitted_sat2@Options
  tmp_options$se <- lavoptions$se
  vcov_sat2 <- lav_model_vcov(
    lavmodel = fitted_sat2@Model,
    lavsamplestats = fitted_sat2@SampleStats,
    lavoptions = tmp_options,
    lavdata = fitted_sat2@Data,
    lavpartable = fitted_sat2@ParTable,
    lavcache = fitted_sat2@Cache,
    use_ginv = TRUE
  )
  inv_g_to_sigmasigma_attheta0 <- lavsamplestats@ntotal * vcov_sat2[d_sat_idx,
                                d_sat_idx, drop = FALSE] # G^sigmasigma(theta0)
  # Hattheta0 <- (-1)* getHessian(fittedSat2)
  # Hattheta0 <- getHessian(fittedSat2)
  # InvHattheta0 <- solve(Hattheta0)
  inv_hattheta0 <- attr(vcov_sat2, "E.inv")
  if (is.null(inv_hattheta0)) {
    # not attached when the vcov/test information options differ (e.g.,
    # with cluster =); recompute the observed-information version
    sat2_options <- tmp_options
    sat2_options$information <- c("observed", "observed")
    inv_hattheta0 <- lav_model_info(
      lavmodel = fitted_sat2@Model,
      lavsamplestats = fitted_sat2@SampleStats,
      lavdata = fitted_sat2@Data,
      lavcache = fitted_sat2@Cache,
      lavoptions = sat2_options,
      extra = FALSE,
      augmented = TRUE,
      inverted = TRUE,
      use_ginv = TRUE
    )
  }
  inv_h_to_sigmasigma_attheta0 <-
              inv_hattheta0[d_sat_idx, d_sat_idx, drop = FALSE]
                                                    # H^sigmasigma(theta0)
  # Inv_of_InvH_to_sigmasigma_attheta0 <- solve(InvH_to_sigmasigma_attheta0)
  #                                                 #[H^sigmasigma(theta0)]^(-1)
  inv_of_inv_h_to_sigmasigma_attheta0 <- MASS::ginv(       # nolint
    inv_h_to_sigmasigma_attheta0,
    tol = .Machine$double.eps ^ (3 / 4)
  )
  h1tmp_prod1 <- inv_of_inv_h_to_sigmasigma_attheta0 %*%
                            inv_g_to_sigmasigma_attheta0
  h1tmp_prod2 <- h1tmp_prod1 %*% h1tmp_prod1
  e_tzz <- sum(diag(h1tmp_prod1)) # expected mean of the second
                                  # quadratic quantity
  var_tzz <- 2 * sum(diag(h1tmp_prod2)) # variance of the second
                                        # quadratic quantity

  ##### Section 3: Compute the asymptotic covariance of
  #####            the two quadratic quantities

  drhodpsi_mat_1 <- vector("list", length = lavsamplestats@ngroups)
  group_values <- lav_pt_group_values(fitted_sat2@ParTable)
  for (g in 1:lavsamplestats@ngroups) {
    # delta.g <- lav_model_delta(lavmodel)[[g]] # [[1]] to be substituted by g?
    # The above gives the derivatives of thresholds and polychoric correlations
    # with respect to SEM param (including thresholds) evaluated under H0.
    # From deltamat we need to exclude the rows and columns referring
    # to thresholds.
    # For this:

    # order of the rows: first the thresholds, then the correlations
    # we need to map the rows of delta.g to the rows/cols of H_at_vartheta0
    # of H1

    pt_1 <- fitted_sat2@ParTable
    pt_1$label <- lav_pt_labels(pt_1)
    free_idx <- which(pt_1$free > 0 & pt_1$op != "|" &
                       pt_1$group == group_values[g])
    parlabel <- pt_1$label[free_idx]

    # for now, we can assume that lav_model_delta will always return
    # the thresholds first, then the correlations
    #
    # later, we should add a (working) add.labels = TRUE option to
    # lav_model_delta
    # th.names <- lavobject@pta$vnames$th[[g]]
    # ov.names <- lavobject@pta$vnames$ov[[g]]
    # th.names <- lav_object_vnames(lavpartable, "th")
    # ov.names <- lav_object_vnames(lavpartable, "ov.nox")
    # ov.names.x <- lav_object_vnames(lavpartable, "ov.x")
    # tmp <- utils::combn(ov.names, 2)
    # cor.names <- paste(tmp[1,], "~~", tmp[2,], sep = "")

    # added by YR - 22 Okt 2017 #####################################
    # ov.names.x <- lav_object_vnames(lavpartable, "ov.x")
    # if(length(ov.names.x)) {
    #    slope.names <- apply(expand.grid(ov.names, ov.names.x), 1L,
    #                             paste, collapse = "~")
    # } else {
    #    slope.names <- character(0L)
    # }
    #################################################################

    # NAMES <- c(th.names, slope.names, cor.names)

    # added by YR - 26 April 2018, for 0.6-1
    # we now can get 'labelled' delta rownames
    delta_g <- lav_inspect_delta_internal(
      lavmodel = lavmodel,
      lavdata = lavdata, lavpartable = lavpartable,
      add_labels = TRUE, add_class = FALSE,
      drop_list_single_group = FALSE
    )[[g]]
    names_1 <- rownames(delta_g)
    if (g > 1L) {
      names_1 <- paste(names_1, ".g", g, sep = "")
    }

    par_idx <- match(parlabel, names_1)
    if (any(is.na(par_idx))) {
      lav_msg_warn(gettextf(
        "mismatch between DELTA labels and PAR labels!
        PARLABEL: %1$s, DELTA LABELS: %2$s", lav_msg_view(parlabel),
         lav_msg_view(names_1)))
    }

    drhodpsi_mat_1[[g]] <- delta_g[par_idx, index_par, drop = FALSE]
  }
  drhodpsi_mat <- do.call(rbind, drhodpsi_mat_1)

  tmp_prod <- t(drhodpsi_mat) %*% inv_of_inv_h_to_sigmasigma_attheta0 %*%
    drhodpsi_mat %*% inv_g_to_psipsi_attheta0 %*% h0tmp_prod1
  cov_tzztww <- 2 * sum(diag(tmp_prod))

  ##### Section 4: compute the adjusted PLRT and its p-value
  # PLRTH0Sat <- 2*nsize*(lavfit@fx - fittedSat@Fit@fx)
  plrth0sat <- 2 * (h0_fx - sat_fx)
  plrth0sat_group <- 2 * (h0_fx_group - sat_fx_group)
  asym_mean_plrth0sat <- e_tzz - e_tww
  # catch zero value for asym_mean_PLRTH0Sat
  if (asym_mean_plrth0sat == 0) {
    asym_var_plrth0sat <- 0
    scaling_factor <- as.numeric(NA)
    fsa_plrt_sem <- as.numeric(NA)
    adjusted_df <- as.integer(NA)
    pvalue <- as.numeric(NA)
  } else if (any(is.na(c(var_tzz, var_tww, cov_tzztww)))) {
    asym_var_plrth0sat <- as.numeric(NA)
    scaling_factor <- as.numeric(NA)
    fsa_plrt_sem <- as.numeric(NA)
    adjusted_df <- as.integer(NA)
    pvalue <- as.numeric(NA)
  } else {
    asym_var_plrth0sat <- var_tzz + var_tww - 2 * cov_tzztww
    scaling_factor <- (asym_mean_plrth0sat / (asym_var_plrth0sat / 2))
    fsa_plrt_sem <- (asym_mean_plrth0sat / (asym_var_plrth0sat / 2)) * plrth0sat
    adjusted_df <- (asym_mean_plrth0sat * asym_mean_plrth0sat) /
      (asym_var_plrth0sat / 2)
    # In some very few cases (simulations show very few cases in small
    # sample sizes) the adjusted_df is a negative number, we should then
    # print a warning like: "The adjusted df is computed to be a negative number
    # and for this the first and second moment adjusted PLRT is not computed."
    if (scaling_factor > 0) {
      pvalue <- 1 - pchisq(fsa_plrt_sem, df = adjusted_df)
    } else {
      pvalue <- as.numeric(NA)
    }
  }

  list(
    PLRTH0Sat = plrth0sat, PLRTH0Sat.group = plrth0sat_group,
    stat = fsa_plrt_sem, df = adjusted_df, p.value = pvalue,
    scaling.factor = scaling_factor
  )
}
############################################################################


lav_pml_object_aic_bic <- function(lavobject) {
  ########################## The code for PL version fo AIC and BIC
  # The following should be done because it is not the pl log-likelihood
  # that is maximized but a fit function that should be minimized. So, we
  # should find the value of log-PL at the estimated parameters through the
  # value of the fitted function.
  # The following may need to be updated if we change the fit function
  # so that it is correct for the case of missing values as well.

  log_pl <- lavobject@optim$logl
  nsize <- lavobject@SampleStats@ntotal

  # inverted observed unit information
  h_inv <- lavTech(lavobject, "inverted.information.observed")

  # first order unit information
  j <- lavTech(lavobject, "information.first.order")

  # trace (J %*% H.inv) = sum (J * t(H.inv))
  dim_theta <- sum(j * h_inv)


  # computations of PL versions of AIC and BIC
  pl_aic <- (-2) * log_pl + 2 * dim_theta
  pl_bic <- (-2) * log_pl + dim_theta * log(nsize)

  list(logPL = log_pl, PL_AIC = pl_aic, PL_BIC = pl_bic)
}

# the univariate ordinal (probit/logit) regression model
#
# - probit regression / ordinal probit regression
# - logit regression / ordinal logit regression (logistic = TRUE)
# - using case weights wt
#
# YR 25 Nov 2019 (replacing the old lav_probit.R routines)
# YR/LDW 2026: refactored (see lav_uvbv_common.R for the shared cache/minfns
#              design); no change in behavior
#
# Note: the idea of using 'o1' and 'o2' when computing z1/z2 comes from
#       the dissertation of Christensen, 2012 (see also his `ordinal' package)
#
# The model for case i with response category y_i = k (k = 1, .., nth + 1):
#
#     P(y_i = k) = pfun(z1_i) - pfun(z2_i)
#
# with the latent-response 'rectangle' bounds
#
#     z1_i = th[k]     - eta_i  (+100 if k == nth + 1, i.e. th[nth+1] = +Inf)
#     z2_i = th[k - 1] - eta_i  (-100 if k == 1,       i.e. th[0]     = -Inf)
#
# and linear predictor eta_i = X_i %*% beta (0 if no covariates).
# The parameter vector is theta = c(th, beta).
#
# fit object (list) contract -- see lav_uvord_fit():
#   theta, nexo, nth, th_idx, slope_idx, missing_idx,
#   y (integer, recoded to 1..ncat), wt, y1, y2 (case x th indicator
#   matrices), z1, z2 (bounds at theta), x
# The z1/z2 values are those of the *last* evaluation inside nlminb;
# lav_uvord_update_fit() recomputes them for new parameter values.

lav_uvord_fit <- function(y = NULL,
                          x = NULL,
                          wt = rep(1, length(y)),
                          lower = -Inf,
                          upper = +Inf,
                          optim_method = "nlminb",
                          logistic = FALSE, # probit is the default
                          control = list(),
                          output = "list") {
  # y
  if (!is.integer(y)) {
    # brute force, no checking! (this is a lower-level function)
    y <- as.integer(y)
  }
  if (!min(y, na.rm = TRUE) == 1L) {
    y <- as.integer(ordered(y))
  }

  # check weights
  wt <- lav_uvbv_wt_check(y = y, wt = wt)

  # optim.method
  minfns <- lav_uvbv_optim_fns(optim_method, lav_uvord_min_fns())
  y_ncat <- length(tabulate(y)) # number of response categories
  if (y_ncat <= 1L) {
    minfns$gradient <- minfns$hessian <- NULL
  }

  # create cache environment
  cache <- lav_uvord_init_cache(y = y, x = x, wt = wt, logistic = logistic)

  # optimize -- only changes from defaults
  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L,
    trace = 0L, abs.tol = (.Machine$double.eps * 10)
  )
  control_nlminb <- modifyList(control_nlminb, control)

  optim <- nlminb(
    start = cache$theta, objective = minfns$objective,
    gradient = minfns$gradient, hessian = minfns$hessian,
    control = control_nlminb, lower = lower, upper = upper,
    cache = cache
  )

  if (output == "cache") {
    return(cache)
  }

  # return results as a list (to be compatible with lav_bvord.R)
  list(
    theta = optim$par,
    nexo = cache$nexo,
    nth = cache$nth,
    th_idx = seq_len(cache$nth),
    slope_idx = seq_along(optim$par)[-seq_len(cache$nth)],
    missing_idx = cache$missing_idx,
    y = cache$y,
    wt = cache$wt,
    y1 = cache$y1,
    y2 = cache$y2,
    z1 = cache$z1,
    z2 = cache$z2,
    x = cache$x
  )
}

# shortcut to get (possibly weighted) thresholds only, if no eXo
lav_uvord_th <- function(y = NULL, wt = NULL) {
  y_freq <- tabulate(y) # unweighted
  y_ncat <- length(y_freq) # number of response categories
  if (is.null(wt)) {
    y_prop <- y_freq / sum(y_freq)
  } else {
    y_freq <- numeric(y_ncat) # numeric! weights...
    for (cat in seq_len(y_ncat)) {
      y_freq[cat] <- sum(wt[y == cat], na.rm = TRUE)
    }
    y_prop <- y_freq / sum(y_freq)
  }

  qnorm(cumsum(y_prop[-length(y_prop)]))
}

# the -Inf/+Inf 'offsets': add +100 (-100) to z1 (z2) for the extreme
# categories, so that pfun() evaluates to 1 (0) there
lav_uvord_o12 <- function(y = NULL, nth = NULL) {
  list(
    o1 = ifelse(y == nth + 1, 100, 0),
    o2 = ifelse(y == 1, -100, 0)
  )
}

# the latent-response rectangle bounds z1/z2, given thresholds th, (optional)
# slopes beta for covariates x, and an (optional) additional per-case offset
# (the offset is not used by the single-level code, but allows a cluster
#  random effect to shift eta in a two-level extension)
lav_uvord_z12 <- function(y = NULL, th = NULL, o1 = NULL, o2 = NULL,
                          x = NULL, beta = NULL, offset = NULL) {
  th_1 <- c(0, th, 0) # th[0] and th[nth+1]; o1/o2 take over there
  eta <- NULL
  if (!is.null(x) && length(beta) > 0L) {
    eta <- drop(x %*% beta)
  }
  if (!is.null(offset)) {
    eta <- if (is.null(eta)) offset else eta + offset
  }
  if (!is.null(eta)) {
    z1 <- th_1[y + 1L] - eta + o1
    z2 <- th_1[y] - eta + o2
  } else {
    z1 <- th_1[y + 1L] + o1
    z2 <- th_1[y] + o2
  }
  list(z1 = z1, z2 = z2)
}

# prepare cache environment
lav_uvord_init_cache <- function(y = NULL,
                                 x = NULL,
                                 wt = rep(1, length(y)),
                                 logistic = FALSE,
                                 parent = parent.frame()) {
  nobs <- length(y)

  # number of response categories
  y_ncat <- length(tabulate(y)) # unweighted
  # number of thresholds
  nth <- y_ncat - 1L

  # X
  if (is.null(x)) {
    nexo <- 0L
  } else {
    x <- unname(x)
    nexo <- ncol(x)
    # new in 0.6-17: check if X is full rank
    if (!anyNA(x)) {
      if (qr(x)$rank < ncol(x)) {
        lav_msg_stop(gettext(
          "matrix of exogenous covariates is rank deficient! (i.e., some x
          variables contain redundant information)"))
      }
    }
  }

  # nobs
  if (is.null(wt)) {
    n <- nobs
  } else {
    n <- sum(wt)
  }

  # frequencies (possibly weighted by wt)
  y_freq <- numeric(y_ncat) # numeric! weights...
  for (cat in seq_len(y_ncat)) {
    y_freq[cat] <- sum(wt[y == cat], na.rm = TRUE)
  }
  y_prop <- y_freq / sum(y_freq)

  # missing values
  missing_idx <- which(is.na(y))

  # missing values
  if (any(is.na(y)) || (!is.null(x) && any(is.na(x)))) {
    lav_crossprod <- lav_mat_crossprod
  } else {
    lav_crossprod <- base::crossprod
  }

  # distribution
  if (logistic) {
    pfun <- plogis
    dfun <- dlogis
    gfun <- function(x) {
      # FIXMe: is it worth making this work for abs(x) > 200?
      out <- numeric(length(x))
      out[is.na(x)] <- NA
      x_ok <- which(abs(x) < 200)
      e <- exp(-x[x_ok])
      e1 <- 1 + e
      e2 <- e1 * e1
      e4 <- e2 * e2
      out[x_ok] <- -e / e2 + e * (2 * (e * e1)) / e4
      out
    }
  } else {
    pfun <- pnorm
    dfun <- dnorm
    gfun <- function(x) {
      -x * dnorm(x)
    }
  }

  # offsets -Inf/+Inf
  o12 <- lav_uvord_o12(y = y, nth = nth)
  o1 <- o12$o1
  o2 <- o12$o2

  # TH matrices (Matrix logical?)
  if (nth > 0L) {
    y1 <- matrix(1:nth, nobs, nth, byrow = TRUE) == y
    y2 <- matrix(1:nth, nobs, nth, byrow = TRUE) == (y - 1L)
  } else {
    y1 <- y2 <- matrix(nrow = nobs, ncol = 0)
  }

  # starting values
  if (nexo == 0L && nth > 0L) {
    if (logistic) {
      th_start <- qlogis(cumsum(y_prop[-length(y_prop)]))
    } else {
      th_start <- qnorm(cumsum(y_prop[-length(y_prop)]))
    }
  } else if ((nth == 1L && nexo > 0L) || nth == 0L) {
    th_start <- 0
  } else {
    if (logistic) {
      th_start <- qlogis((1:nth) / (nth + 1))
    } else {
      th_start <- qnorm((1:nth) / (nth + 1))
    }
  }
  beta_start <- rep(0, nexo)
  theta <- c(th_start, beta_start)

  list2env(
    list(
      y = y, x = x, wt = wt, o1 = o1, o2 = o2,
      missing_idx = missing_idx, n = n,
      pfun = pfun, dfun = dfun, gfun = gfun,
      lav_crossprod = lav_crossprod,
      nth = nth, nobs = nobs, y.ncat = y_ncat, nexo = nexo,
      y1 = y1, y2 = y2,
      theta = theta
    ),
    parent = parent
  )
}

# compute total (log)likelihood
lav_uvord_loglik <- function(y = NULL,
                             x = NULL,
                             wt = rep(1, length(y)),
                             logistic = FALSE,
                             cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvord_fit(
      y = y, x = x, wt = wt,
      logistic = logistic, output = "cache"
    )
  }
  lav_uvord_loglik_cache(cache = cache)
}

lav_uvord_loglik_cache <- function(cache = NULL) {
  with(cache, {      # nolint start
    # Note: we could treat the binary case separately,
    # avoiding calling pfun() twice

    # free parameters
    th <- theta[1:nth]
    beta <- theta[-c(1:nth)]
    z12 <- lav_uvord_z12(y = y, th = th, o1 = o1, o2 = o2, x = x, beta = beta)
    z1 <- z12$z1
    z2 <- z12$z2
    pi_i <- pfun(z1) - pfun(z2)

    # avoid numerical degradation if z2 (and therefore z1) are both 'large'
    # and the pfuns are close to 1.0
    large_idx <- which(z2 > 1)
    if (length(large_idx) > 0L) {
      pi_i[large_idx] <- (pfun(z2[large_idx], lower.tail = FALSE) -
        pfun(z1[large_idx], lower.tail = FALSE))
    }

    loglik <- sum(wt * log(pi_i), na.rm = TRUE)

    return(loglik)
  })                      # nolint end
}

# casewise scores
lav_uvord_sc <- function(y = NULL,
                             x = NULL,
                             wt = rep(1, length(y)),
                             use_weights = TRUE,
                             logistic = FALSE,
                             cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvord_fit(
      y = y, x = x, wt = wt,
      logistic = logistic, output = "cache"
    )
  }
  sc <- lav_uvord_sc_cache(cache = cache)

  if (!is.null(wt) && use_weights) {
    sc <- sc * wt
  }

  sc
}

lav_uvord_sc_cache <- function(cache = NULL) {
  with(cache, {
    # d logl / d pi
    dldpi <- 1 / pi_i # unweighted!

    # we assume z1/z2 are available
    p1 <- dfun(z1)
    p2 <- dfun(z2)

    # th
    scores_th <- dldpi * (y1 * p1 - y2 * p2)

    # beta
    if (nexo > 0L) {
      scores_beta <- dldpi * (-x) * (p1 - p2)
      return(cbind(scores_th, scores_beta, deparse.level = 0))
    } else {
      return(scores_th)
    }
  })
}

lav_uvord_grad_cache <- function(cache = NULL) {
  with(cache, {
    # d logl / d pi
    wtp <- wt / pi_i
    p1 <- dfun(z1)
    p2 <- dfun(z2)

    # th
    dxa <- y1 * p1 - y2 * p2
    scores_th <- wtp * dxa

    # beta
    if (nexo > 0L) {
      dxb <- x * (p1 - p2) # == x*p1 - x*p2
      scores_beta <- wtp * (-dxb)
      return(colSums(cbind(scores_th, scores_beta, deparse.level = 0),
        na.rm = TRUE
      ))
    } else {
      return(colSums(scores_th, na.rm = TRUE))
    }
  })
}


# compute total Hessian
lav_uvord_hessian <- function(y = NULL,
                              x = NULL,
                              wt = rep(1, length(y)),
                              logistic = FALSE,
                              cache = NULL) {
  if (is.null(cache)) {
    cache <- lav_uvord_fit(
      y = y, x = x, wt = wt,
      logistic = logistic, output = "cache"
    )
  }
  tmp <- lav_uvord_loglik_cache(cache = cache)
  tmp <- lav_uvord_grad_cache(cache = cache)
  lav_uvord_hessian_cache(cache = cache)
}

lav_uvord_hessian_cache <- function(cache = NULL) {
  with(cache, {          # nolint start
    wtp2 <- wt / (pi_i * pi_i)

    g1w <- gfun(z1) * wtp
    g2w <- gfun(z2) * wtp

    y1gw <- y1 * g1w
    y2gw <- y2 * g2w

    dx2_tau <- (lav_crossprod(y1gw, y1) - lav_crossprod(y2gw, y2) -
      lav_crossprod(dxa, dxa * wtp2))

    if (nexo == 0L) {
      return(dx2_tau)
    }

    dxb2 <- dxb * wtp2
    dx2_beta <- (lav_crossprod(x * g1w, x) - lav_crossprod(x * g2w, x) -
      lav_crossprod(dxb, dxb2))

    dx_taubeta <- (-lav_crossprod(y1gw, x) + lav_crossprod(y2gw, x) +
      lav_crossprod(dxa, dxb2))

    hessian <- rbind(cbind(dx2_tau, dx_taubeta, deparse.level = 0),
      cbind(t(dx_taubeta), dx2_beta, deparse.level = 0),
      deparse.level = 0
    )

    return(hessian)
  })                             # nolint end
}

# reject descending values for the first few thresholds (new in 0.6-8)
lav_uvord_pre_objective <- function(x, cache = NULL) {
  if (cache$nth > 1L && x[1] > x[2]) {
    return(+Inf)
  }
  if (cache$nth > 2L && x[2] > x[3]) {
    return(+Inf)
  }
  if (cache$nth > 3L && x[3] > x[4]) {
    return(+Inf)
  }
  NULL
}

# nlminb objective/gradient/hessian (see lav_uvbv_common.R)
lav_uvord_min_fns <- function() {
  lav_uvbv_min_fns(
    logl_fun = lav_uvord_loglik_cache,
    grad_fun = lav_uvord_grad_cache,
    hessian_fun = lav_uvord_hessian_cache,
    pre_objective = lav_uvord_pre_objective
  )
}

# get 'z1' and 'z2' values, given (new) values for the parameters
# only needed for lav_bvord_cor_sc(), which is called from
# lav_pml_dploglik_dimplied() in lav_model_gradient_pml.R
lav_uvord_update_fit <- function(fit_y = NULL, th_new = NULL, sl_new = NULL) {
  # return fit.y with 'update' z1/z2 values
  if (is.null(th_new) && is.null(sl_new)) {
    return(fit_y)
  }

  if (!is.null(th_new)) {
    fit_y$theta[fit_y$th_idx] <- th_new
  }
  if (!is.null(sl_new)) {
    fit_y$theta[fit_y$slope_idx] <- sl_new
  }

  nth <- length(fit_y$th_idx)
  o12 <- lav_uvord_o12(y = fit_y$y, nth = nth)

  theta <- fit_y$theta
  th <- theta[1:nth]
  beta <- theta[-c(1:nth)]
  z12 <- lav_uvord_z12(
    y = fit_y$y, th = th, o1 = o12$o1, o2 = o12$o2,
    x = fit_y$x, beta = beta
  )
  fit_y$z1 <- z12$z1
  fit_y$z2 <- z12$z2

  fit_y
}

lav_step05_samp <- function(slot_sample_stats = NULL,
                                          lavdata = NULL,
                                          lavoptions = NULL,
                                          wls_v = NULL,
                                          nacov = NULL,
                                          sample_cov = NULL,
                                          sample_mean = NULL,
                                          sample_th = NULL,
                                          sample_nobs = NULL,
                                          ov_names = NULL,
                                          ov_names_x = NULL,
                                          lavpartable = NULL) {
  # # # # # # # # # # # # # #
  # #  5. lavsamplestats  # #
  # # # # # # # # # # # # # #

  # if slot_sample_stats not NULL
  #   copy to lavsamplestats
  # else
  #   if lavdata@data.type == "full"
  #     compute lavsamplestats via lav_samplestats_from_data
  #   else
  #     if lavdata@data.type == "moment"
  #       if lavoptions$meanstructure TRUE but sample_mean is NULL:
  #         ** warning **
  #       compute lavsamplestats via lav_samplestats_from_moments
  #     else
  #       create lavsamplestats object (type lavSampleStats) with data from
  #         lavdata and lavpta

  if (!is.null(slot_sample_stats)) {
    lavsamplestats <- slot_sample_stats
  } else if (lavdata@data.type == "full") {
    if (lav_verbose()) {
      cat("lavsamplestats     ...")
    }
    lavsamplestats <- lav_samp_from_data(
      lavdata       = lavdata,
      lavoptions    = lavoptions,
      wls_v         = wls_v,
      nacov         = nacov
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }
  } else if (lavdata@data.type == "moment") {
    if (lav_verbose()) {
      cat("lavsamplestats ...")
    }
    # check if we have sample_mean and meanstructure = TRUE
    if (lavoptions$meanstructure && is.null(sample_mean)) {
      lav_msg_warn(
        gettext("sample_mean= argument is missing, but model contains
                mean/intercept parameters."))
    }
    lavsamplestats <- lav_samp_from_moments(
      sample_cov    = sample_cov,
      sample_mean   = sample_mean,
      sample_th     = sample_th,
      sample_nobs   = sample_nobs,
      ov_names      = ov_names,
      ov_names_x    = ov_names_x,
      wls_v         = wls_v,
      nacov         = nacov,
      lavoptions    = lavoptions
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }
  } else {
    # no data
    lavsamplestats <- new("lavSampleStats",
      ngroups = lavdata@ngroups,
      nobs = as.list(rep(0L, lavdata@ngroups)),
      cov.x = vector("list", length = lavdata@ngroups),
      mean.x = vector("list", length = lavdata@ngroups),
      th.idx = attr(lavpartable, "th.idx"),
      missing.flag = FALSE
    )
  }

  if (lav_debug()) {
    print(str(lavsamplestats))
  }

  lavsamplestats
}

# YR 19 September 2022
#
# Entry function to handle noniterative estimators


lav_optim_noniter <- function(lavmodel = NULL, lavsamplestats = NULL,
                              lavpartable = NULL, lavh1 = NULL,
                              lavdata = NULL, lavoptions = NULL) {
  lavpta <- lav_pt_attributes(lavpartable)
  lavpartable <- lav_pt_set_cache(lavpartable, lavpta)

  # no support for many things:
  # (the IV estimator supports multiple groups, including cross-group equality
  # constraints for measurement invariance)
  if (lavmodel@ngroups > 1L && !identical(lavoptions$estimator, "IV")) {
    lav_msg_stop(gettext(
      "multiple groups not supported (yet) with optim.method = 'NONITER'."))
  }

  if (lavdata@nlevels > 1L) {
    lav_msg_stop(gettext(
      "multilevel not supported (yet) with optim.method = 'NONITER'."))
  }

  # no support (yet) for nonlinear constraints
  nonlinear_idx <- c(
    lavmodel@ceq.nonlinear.idx,
    lavmodel@cin.nonlinear.idx
  )
  if (length(nonlinear_idx) > 0L) {
    lav_msg_stop(gettext(
      "nonlinear constraints not supported (yet) with optim.method = 'NONITER'."
      ))
  }

  # no support (yet) for inequality constraints
  if (!lavmodel@cin.simple.only && !is.null(body(lavmodel@cin.function))) {
    lav_msg_stop(gettext(
    "inequality constraints not supported (yet) with optim.method = 'NONITER'."
    ))
  }

  # linear equality constraints are supported by the IV estimator (handled by
  # a constrained system solve); other noniterative estimators do not (yet)
  # support them
  if (length(lavmodel@ceq.linear.idx) > 0L &&
      !identical(lavoptions$estimator, "IV")) {
    lav_msg_stop(gettext(
      "equality constraints not supported (yet) with optim.method = 'NONITER'."
      ))
  }

  # extract current set of free parameters
  x_old <- lav_model_get_parameters(lavmodel)
  # npar <- length(x_old)


  # fabin?
  ok_flag <- FALSE
  if (lavoptions$estimator %in% c("FABIN2", "FABIN3")) {
    x <- try(lav_cfa_fabin_internal(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavpartable = lavpartable,
      lavoptions = lavoptions
    ), silent = TRUE)
  } else if (lavoptions$estimator == "MGM") {
    x <- try(lav_cfa_guttman1952_internal(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavpartable = lavpartable,
      lavoptions = lavoptions
    ), silent = TRUE)
  } else if (lavoptions$estimator == "BENTLER1982") {
    x <- try(lav_cfa_bentler1982_internal(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavpartable = lavpartable,
      lavoptions = lavoptions
    ), silent = TRUE)
  } else if (lavoptions$estimator %in% c("JS", "JSA")) {
    x <- try(lav_cfa_jamesstein_internal(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavpartable = lavpartable,
      lavdata = lavdata,
      lavoptions = lavoptions
    ), silent = TRUE)
  } else if (lavoptions$estimator == "BENTLER1982") {
    x <- try(lav_cfa_bentler1982_internal(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavpartable = lavpartable,
      lavoptions = lavoptions
    ), silent = TRUE)
  } else if (lavoptions$estimator == "IV") {
    x <- try(lav_sem_miiv_internal(
      lavmodel = lavmodel, lavdata = lavdata, lavh1 = lavh1,
      lavsamplestats = lavsamplestats, lavpartable = lavpartable,
      lavoptions = lavoptions
    ), silent = TRUE)
  } else {
    lav_msg_warn(
      gettextf("unknown (noniterative) estimator: %s
               (returning starting values)", lavoptions$estimator)
    )
  }
  if (inherits(x, "try-error")) {
    x <- x_old
  } else {
    ok_flag <- TRUE
  }

  # closing
  fx <- 0
  attr(fx, "fx.group") <- rep(0, lavmodel@ngroups)
  if (ok_flag) {
    attr(x, "converged") <- TRUE
    attr(x, "warn.txt") <- ""
  } else {
    attr(x, "converged") <- FALSE
    attr(x, "warn.txt") <- "noniterative estimation failed"
  }
  attr(x, "iterations") <- 1L
  attr(x, "control") <- list()
  attr(x, "fx") <- fx

  x
}

# deprecated: only kept in order to avoid some older packages
lav_model_fit <- function(lavpartable = NULL,
                          lavmodel = NULL,
                          lavimplied = NULL,
                          x = NULL,
                          vcov_1 = NULL,
                          test = NULL) {
  stopifnot(is.list(lavpartable), inherits(lavmodel, "lavModel"))

  # extract information from 'x'
  iterations <- attr(x, "iterations")
  converged <- attr(x, "converged")
  fx <- attr(x, "fx")
  fx_group <- attr(fx, "fx.group")
  if (!is.null(attr(fx, "logl.group"))) {
    logl_group <- attr(fx, "logl.group")
    logl <- sum(logl_group)
  } else {
    logl_group <- as.numeric(NA)
    logl <- as.numeric(NA)
  }
  # print(fx.group)
  control <- attr(x, "control")
  attributes(fx) <- NULL
  x_copy <- x # we are going to change it (remove attributes)
  attributes(x_copy) <- NULL
  est <- lav_model_get_parameters(lavmodel = lavmodel, type = "user")

  # did we compute standard errors?
  if (is.null(lavpartable$se)) {
    if (is.null(vcov_1)) {
      se <- rep(as.numeric(NA), lavmodel@nx.user)
      se[lavpartable$free == 0L] <- 0
    } else {
      se <- lav_model_vcov_se(
        lavmodel = lavmodel,
        lavpartable = lavpartable,
        vcov = vcov_1,
        boot = attr(vcov_1, "BOOT.COEF")
      )
    }
  } else {
    se <- as.numeric(lavpartable$se) # could be logical NA
  }

  # did we compute test statistics
  if (is.null(test)) {
    test_1 <- list()
  } else {
    test_1 <- test
  }

  # for convenience: compute lavmodel-implied Sigma and Mu
  if (is.null(lavimplied) || length(lavimplied) == 0L) {
    implied <- lav_model_implied(lavmodel)
  } else {
    implied <- lavimplied
  }

  # if bootstrapped parameters, add attr to 'est'
  if (!is.null(attr(vcov_1, "BOOT.COEF"))) {
    attr(est, "BOOT.COEF") <- attr(vcov_1, "BOOT.COEF")
  }

  # partrace?
  if (!is.null(attr(x, "partrace"))) {
    partrace <- attr(x, "partrace")
  } else {
    partrace <- matrix(0, 0L, 0L)
  }

  new("Fit",
    npar       = max(lavpartable$free),
    x          = x_copy,
    partrace   = partrace,
    start      = lavpartable$start, # needed? (package stremo!)
    est        = est, # at least 5 packages!!
    se         = se,
    fx         = fx,
    fx.group   = fx_group,
    logl       = logl,
    logl.group = logl_group,
    iterations = iterations,
    converged  = converged,
    control    = control,
    Sigma.hat  = if (lavmodel@conditional.x) implied$res.cov else implied$cov,
    Mu.hat     = if (lavmodel@conditional.x) implied$res.int else implied$mean,
    TH         = if (lavmodel@conditional.x) implied$res.th else implied$th,
    test       = test_1
  )
}

# user-visible routine to
# compute polychoric/polyserial/... correlations
#
# YR 17 Sept 2013
#
# - YR 26 Nov 2013: big change - make it a wrapper around lavaan()
#                   estimator = "none" means two.step (starting values)

lav_object_cor <- function(object,
                   # lav.data options
                   ordered = NULL,
                   group = NULL,
                   missing = "listwise",
                   ov_names_x = NULL,
                   sampling_weights = NULL,
                   # lavaan options
                   se = "none",
                   test = "none",
                   estimator = "two.step",
                   baseline = FALSE,
                   # other options (for lavaan)
                   ...,
                   cor_smooth = FALSE,
                   cor_smooth_tol = 1e-04, # was 1e-06 in <0.6-14
                   output = "cor") {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, TRUE)
  # shortcut if object = lavaan object
  if (inherits(object, "lavaan")) {
    # check object
    object <- lav_object_check_version(object)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          "function lav_object_cor for lavaan-object")
        )
      }
    }
    out <- lav_object_cor_output(object, output = output)
    return(out)
  }

  # check estimator
  estimator <- tolower(estimator)
  if (estimator %in% c("two.step", "two.stage")) {
    estimator <- "none"
  }

  # se?
  se <- tolower(se)
  output <- tolower(output)
  if (se != "none") {
    if (output %in% c("cor", "cov", "sampstat", "th", "thresholds")) {
      lav_msg_warn(gettext("argument `se' is ignored since standard errors
                           are not needed for the requested `output'"))
      se <- "none"
    }
  }

  # extract sampling.weights.normalization from dots (for lav_lavdata() call)
  dots <- dotdotdot
  sampling_weights_normalization <- "group"
  if (!is.null(dots$sampling.weights.normalization)) {
    sampling_weights_normalization <- dots$sampling.weights.normalization
  }


  # check object class
  if (inherits(object, "lavData")) {
    lav_data <- object
  } else if (inherits(object, "data.frame") ||
    inherits(object, "matrix")) {
    object <- as.data.frame(object)
    names_1 <- names(object)
    if (!is.null(group)) {
      names_1 <- names_1[-match(group, names_1)]
    }
    if (!is.null(sampling_weights)) {
      names_1 <- names_1[-match(sampling_weights, names_1)]
    }
    if (is.logical(ordered)) {
      ordered_flag <- ordered
      if (ordered_flag) {
        ordered <- names_1
        if (length(ov_names_x) > 0L) {
          ordered <- ordered[-which(ordered %in% ov_names_x)]
        }
      } else {
        ordered <- character(0L)
      }
    } else if (is.null(ordered)) {
      ordered <- character(0L)
    } else if (!is.character(ordered)) {
      lav_msg_stop(gettext("ordered argument must be a character vector"))
    } else if (length(ordered) == 1L && nchar(ordered) == 0L) {
      ordered <- character(0L)
    } else {
      # check if all names in "ordered" occur in the dataset?
      missing_idx <- which(!ordered %in% names_1)
      if (length(missing_idx) > 0L) {
        lav_msg_warn(gettextf(
          "ordered variable(s): %s could not be found
          in the data and will be ignored",
          lav_msg_view(ordered[missing_idx])))
      }
    }
    lav_data <- lav_lavdata(
      data = object, group = group,
      ov_names = names_1, ordered = ordered,
      sampling_weights = sampling_weights,
      ov_names_x = ov_names_x,
      lavoptions = list(
        missing = missing,
        sampling.weights.normalization = sampling_weights_normalization
      )
    )
  } else {
    lav_msg_stop(gettext("lav_object_cor can not handle objects of class"),
      paste(class(object), collapse = " ")
    )
  }

  # set default estimator if se != "none"
  categorical <- any(lav_data@ov$type == "ordered")
  if (se != "none" && estimator == "none") {
    if (categorical) {
      estimator <- "WLSMV"
    } else {
      estimator <- "ML"
    }
  }

  # extract more partable options from dots
  meanstructure <- FALSE
  fixed_x <- FALSE
  mimic <- "lavaan"
  conditional_x <- FALSE
  if (!is.null(dots$meanstructure)) {
    meanstructure <- dots$meanstructure
  }
  if (lav_data@ngroups > 1L || categorical ||
      tolower(missing) %in% c("ml", "fiml", "direct")) {
    meanstructure <- TRUE
  }
  if (!is.null(dots$fixed.x)) {
    fixed_x <- dots$fixed.x
  }
  if (!is.null(dots$mimic)) {
    mimic <- dots$mimic
  }
  if (!is.null(dots$conditional.x)) {
    conditional_x <- dots$conditional.x
  }

  # override, only for backwards compatibility (eg moments() in JWileymisc)
  # if(missing %in% c("ml", "fiml")) {
  #    meanstructure = TRUE
  # }

  # generate partable for unrestricted model
  pt_un <-
    lav_pt_unrestricted(
      lavobject = NULL,
      lavdata = lav_data,
      lavoptions = list(
        meanstructure = meanstructure,
        fixed.x = fixed_x,
        conditional.x = conditional_x,
        # sampling.weights.normalization = sampling.weights.normalization,
        group.w.free = FALSE,
        missing = missing,
        estimator = estimator,
        mimic = mimic
      ),
      sample_cov = NULL,
      sample_mean = NULL,
      sample_th = NULL
    )


  fit <- do.call(lavaan, c(list(
    slot_par_table = pt_un, slot_data = lav_data,
    model.type = "unrestricted",
    missing = missing,
    baseline = baseline, h1 = TRUE, # must be TRUE!
    se = se, test = test, estimator = estimator), dotdotdot)
  )

  out <- lav_object_cor_output(fit, output = output)

  # smooth correlation matrix? (only if output = "cor")
  if (output == "cor" && cor_smooth) {
    tmp_attr <- attributes(out)
    out <- cov2cor(lav_mat_sym_force_pd(out, tol = cor_smooth_tol))
    # we lost most of the attributes
    attributes(out) <- tmp_attr
  }

  out
}
lavCor <- lav_alias("lav_object_cor") # synonym #nolint

lav_object_cor_output <- function(object, output = "cor") {
  # check output
  if (output %in% c("cor", "cov")) {
    out <- lavInspect(object, "sampstat")
    if (object@Data@ngroups == 1L) {
      if (object@Model@conditional.x) {
        out <- out$res.cov
      } else {
        out <- out$cov
      }
      if (output == "cor") {
        out <- cov2cor(out)
      }
    } else {
      if (object@Model@conditional.x) {
        out <- lapply(out, "[[", "res.cov")
      } else {
        out <- lapply(out, "[[", "cov")
      }
      if (output == "cor") {
        out <- lapply(out, cov2cor)
      }
    }
  } else if (output %in% c("th", "thresholds")) {
    out <- lavInspect(object, "sampstat")
    if (object@Data@ngroups == 1L) {
      if (object@Model@conditional.x) {
        out <- out$res.th
      } else {
        out <- out$th
      }
    } else {
      if (object@Model@conditional.x) {
        out <- lapply(out, "[[", "res.th")
      } else {
        out <- lapply(out, "[[", "th")
      }
    }
  } else if (output %in% c("sampstat")) {
    out <- lavInspect(object, "sampstat")
  } else if (output %in% c(
    "parameterEstimates", "pe",
    "parameterestimates", "est"
  )) {
    out <- standardizedSolution(object)
  } else {
    out <- object
  }

  out
}

# # # # # # # # # # # # # # #
# # lavaan main function  # #
# # # # # # # # # # # # # # #
#
# main user-visible cfa/sem/growth functions
#
# initial version: YR 25/03/2009
# added lavoptions YR 02/08/2010
# major revision: YR 9/12/2010: - new workflow (since 0.4-5)
#                               - merge cfa/sem/growth functions
# YR 25/02/2012: changed data slot (from list() to S4); data@X contains data

# YR 26 Jan 2017: use '...' to capture the never-ending list of options
# YR 07 Feb 2023: add ov.order= argument
# HJ 18 Oct 2023: extend PML to allow sampling weights
# LDW 26 Feb 2024: split lavaan in smaller steps
# LDW June 2026: argument names in snake_case

lavaan <- function(
    # user specified model: can be syntax, parameter Table, ...
    model = NULL,
    # data (second argument, most used)
    data = NULL,
    # variable information
    ordered = NULL,
    # auxiliary observed variables (not part of the model)
    aux = NULL,
    # sampling weights
    sampling_weights = NULL,
    # summary data
    sample_cov = NULL,
    sample_mean = NULL,
    sample_th = NULL,
    sample_nobs = NULL,
    # multiple groups?
    group = NULL,
    # multiple levels?
    cluster = NULL,
    # constraints
    constraints = "",
    # user-specified variance matrices
    wls_v = NULL,
    nacov = NULL,
    # internal order of ov.names
    ov_order = "model",
    # full slots from previous fits
    slot_options = NULL,
    slot_par_table = NULL,
    slot_sample_stats = NULL,
    slot_data = NULL,
    slot_model = NULL,
    slot_cache = NULL,
    sloth1 = NULL,
    # options (dotdotdot)
    ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, TRUE)
  # start timer
  start_time0 <- proc.time()[3]
  timing <- list()
  timing$start_time <- start_time0

  # ------------- adapt parameters -----------------
  mc <- match.call(expand.dots = TRUE)
  temp <- lav_step00_parameters(
    matchcall = mc,
    syscall   = sys.call(), # to get main arguments without partial matching
    dotdotdot = dotdotdot
  )
  lavmc <- temp$mc
  dotdotdot <- temp$dotdotdot
  cluster <- lavmc$cluster
  rm(mc)

  # store current random seed (if any)
  if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    temp_seed <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  } else {
    temp_seed <- NULL
  }

  # ------------- handling of warn/debug/verbose switches ----------
  if (!is.null(dotdotdot$debug)) {
    current_debug <- lav_debug()
    if (lav_debug(dotdotdot$debug))
      on.exit(lav_debug(current_debug), TRUE)
    dotdotdot$debug <- NULL
    if (lav_debug()) {
      dotdotdot$warn <- TRUE       # force warnings if debug
      dotdotdot$verbose <- TRUE    # force verbose if debug
    }
  }
  if (!is.null(dotdotdot$warn)) {
    current_warn <- lav_warn()
    if (lav_warn(dotdotdot$warn))
      on.exit(lav_warn(current_warn), TRUE)
    dotdotdot$warn <- NULL
  }
  if (!is.null(dotdotdot$verbose)) {
    current_verbose <- lav_verbose()
    if (lav_verbose(dotdotdot$verbose))
      on.exit(lav_verbose(current_verbose), TRUE)
    dotdotdot$verbose <- NULL
  }

  # ------------ check data ------------------------
  temp <- lav_step00_checkdata(
    data        = data,
    dotdotdot   = dotdotdot,
    sample_cov  = sample_cov,
    sample_nobs = sample_nobs,
    sample_mean = sample_mean,
    sample_th   = sample_th,
    nacov       = nacov,
    wls_v       = wls_v,
    ov_order    = ov_order
  )
  data <- temp$data
  dotdotdot <- temp$dotdotdot
  sample_cov <- temp$sample_cov
  sample_nobs <- temp$sample_nobs
  sample_mean <- temp$sample_mean
  sample_th <- temp$sample_th
  nacov <- temp$nacov
  wls_v <- temp$wls_v
  ov_order <- temp$ov_order

  timing <- lav_add_timing(timing, "init")

  # ------------ ov.names 1 -----  initial flat model --------------------
  # if parser not specified, take default one
  if (is.null(dotdotdot$parser)) {
    opt_default <- lav_options_default()
    useparser <- opt_default$parser
  } else {
    useparser <- dotdotdot$parser
  }

  flat_model <- lav_step01_ovnames_initflat(
    slot_par_table   = slot_par_table,
    model            = model,
    dotdotdot_parser = useparser
  )

  # ------------ ov.names 1b ----- handle 'old way' for composites -------
  if (!is.null(dotdotdot[["composites", exact = TRUE]]) &&
     !dotdotdot[["composites", exact = TRUE]] &&
     any(flat_model$op == "<~")) {
    flat_model <- lav_step01_ovnames_composites(flat_model)
  }

  # ------------ ov.names 2 ------ handle ov_order -----------------------
  flat_model <- lav_step01_ovnames_ovorder(
    flat_model = flat_model,
    ov_order   = ov_order,
    data       = data,
    sample_cov = sample_cov,
    slot_data  = slot_data
  )

  # ------------ ov.names 3 ------- group blocks ------------------
  ngroups <- 1L # default value
  temp <- lav_step01_ovnames_group(
    flat_model = flat_model,
    ngroups    = ngroups
  )
  flat_model <- temp$flat_model
  ov_names <- temp$ov_names
  ov_names_x <- temp$ov_names_x
  ov_names_y <- temp$ov_names_y
  lv_names <- temp$lv_names
  group_values <- temp$group_values
  ngroups <- temp$ngroups

  # ------------ ov.names 4 ------ sanity checks ------------------
  lav_step01_ovnames_checklv(
    lv_names     = lv_names,
    ov_names     = ov_names,
    data         = data,
    sample_cov   = sample_cov,
    dotdotdot    = dotdotdot,
    slot_options = slot_options
  )

  # ------------ ov.names 5 ------ handle ov.names.l --------------
  temp <- lav_step01_ovnames_namesl(
    data         = data,
    cluster      = cluster,
    flat_model   = flat_model,
    group_values = group_values,
    ngroups      = ngroups
  )
  flat_model <- temp$flat_model
  ov_names_l <- temp$ov_names_l

  # ------------ ov.names 6 ------ sanity check ordered --------------
  ordered_orig <- ordered
  ordered <- lav_step01_ovnames_ordered(
    ordered    = ordered,
    flat_model = flat_model,
    data       = data
  )
  timing <- lav_add_timing(timing, "ov.names")

  # ------------ auxiliary variables (FIML saturated correlates) ----------
  # if aux= is supplied and the user requests full-information ML, add the
  # auxiliary variables to the model as a block of 'saturated correlates'
  # (Graham, 2003): they become ordinary (endogenous) observed variables that
  # are freely correlated with each other and with every model variable, and
  # have a free mean. The (enlarged) model is then estimated with casewise
  # FIML, so the auxiliary variables affect the estimates and standard errors.
  # The auxiliary parameters are hidden from the default output. The two-stage
  # missing-data options use a different mechanism (see lav_data_aux_check()).
  aux_satcor         <- FALSE
  aux_fiml_attempted <- FALSE
  aux_names          <- character(0L)
  if (!is.null(aux)) {
    user_missing <- tolower(
      if (!is.null(dotdotdot$missing)) dotdotdot$missing else "")
    user_estimator <- toupper(
      if (!is.null(dotdotdot$estimator)) dotdotdot$estimator else "ML")
    is_fiml <- user_missing %in% c("ml", "fiml", "direct", "direct.ml")
    is_ml_estimator <- user_estimator %in%
      c("", "DEFAULT", "ML", "MLR", "MLM", "MLMV", "MLMVS", "MLF")
    if (is_fiml && is_ml_estimator) {
      aux_fiml_attempted <- TRUE
      aux_names <- lav_aux_clean_names(
        aux = aux, data = data, ov_names = ov_names, ordered = ordered
      )
      if (length(aux_names) > 0L) {
        aux_satcor <- TRUE
        # add the saturated-correlates block to the (flat) model
        flat_model <- lav_aux_satcor_flat(flat_model, aux_names, ov_names)
        # add the auxiliary variables to the observed variable names, so they
        # are read from the data and enter the FIML likelihood (endogenous)
        if (is.list(ov_names)) {
          ov_names   <- lapply(ov_names,   function(x) unique(c(x, aux_names)))
          ov_names_y <- lapply(ov_names_y, function(x) unique(c(x, aux_names)))
        } else {
          ov_names   <- unique(c(ov_names,   aux_names))
          ov_names_y <- unique(c(ov_names_y, aux_names))
        }
        # auxiliary means require a mean structure; auxiliary covariances with
        # exogenous covariates require fixed.x = FALSE
        dotdotdot$meanstructure <- TRUE
        dotdotdot$fixed.x       <- FALSE
        lav_msg_note(gettextf(
          "auxiliary (aux) variable(s) added to the model as saturated
           correlates (estimated with FIML and hidden from the output): %s",
          paste(aux_names, collapse = " ")))
      }
    }
  }

  # ------------ external instruments (estimator = "IV") ----------------
  # instruments supplied with the |~ operator that are NOT otherwise part of
  # the model are 'external' instruments. Like auxiliary variables, they must
  # be read from the data (so that their covariances with the model variables
  # are available to the 2SLS estimator), but they must NOT enter the
  # model-implied summary statistics. They are routed through the generic
  # extra-observed-variable channel (ov_names_aux -> lavData@aux); the IV
  # estimator then augments the covariance matrix with these columns (see
  # lav_sem_miiv.R).
  ext_iv_names <- character(0L)
  if (!aux_satcor) {
    user_estimator2 <- toupper(
      if (!is.null(dotdotdot$estimator)) {
        as.character(dotdotdot$estimator)[1L]
      } else "")
    if (identical(user_estimator2, "IV")) {
      ext_iv_names <- lav_iv_external_names(
        flat_model = flat_model, ov_names = ov_names, data = data)
    }
  }

  # ------------ lavoptions --------------------
  lavoptions <- lav_step02_options(
    slot_options      = slot_options,
    slot_data         = slot_data,
    flat_model       = flat_model,
    ordered          = ordered,
    ordered_orig     = ordered_orig,
    sample_cov       = sample_cov,
    sample_mean      = sample_mean,
    sample_th        = sample_th,
    sample_nobs      = sample_nobs,
    ov_names_l       = ov_names_l,
    sampling_weights = sampling_weights,
    constraints      = constraints,
    group            = group,
    ov_names_x       = ov_names_x,
    ov_names_y       = ov_names_y,
    dotdotdot        = dotdotdot,
    cluster          = cluster,
    data             = data
  )
  # fixed.x = FALSE? set ov_names_x = character(0L)
  # new in 0.6-1
  if (!lavoptions$fixed.x) {
    ov_names_x <- character(0L)
  }
  timing <- lav_add_timing(timing, "Options")

  # ------------ auxiliary variables --------------
  # if the auxiliary variables were already added as saturated correlates
  # (FIML, see above), they are part of the model now; otherwise, validate
  # them for the EM-based (two-stage) channel (see lav_data_aux_check())
  if (aux_satcor) {
    ov_names_aux       <- character(0L)
    lavoptions$aux     <- aux_names
  } else if (aux_fiml_attempted) {
    # FIML saturated-correlates path was attempted, but no usable auxiliary
    # variables remained (already validated/warned above); nothing to do
    ov_names_aux <- character(0L)
  } else if (length(ext_iv_names) > 0L) {
    # external instruments (estimator = "IV"): route through the extra-ov
    # channel so they are read from the data, but keep them out of the model
    ov_names_aux <- ext_iv_names
  } else {
    ov_names_aux <- lav_data_aux_check(
      aux        = aux,
      data       = data,
      ov_names   = ov_names,
      ordered    = ordered,
      lavoptions = lavoptions
    )
  }

  # ------------ lavdata ------------------------
  temp <- lav_step03_data(
    slot_data         = slot_data,
    lavoptions       = lavoptions,
    ov_names         = ov_names,
    ov_names_y       = ov_names_y,
    group            = group,
    data             = data,
    cluster          = cluster,
    ov_names_x       = ov_names_x,
    ov_names_l       = ov_names_l,
    ov_names_aux     = ov_names_aux,
    ordered          = ordered,
    sampling_weights = sampling_weights,
    sample_cov       = sample_cov,
    sample_mean      = sample_mean,
    sample_th        = sample_th,
    sample_nobs      = sample_nobs,
    slot_par_table   = slot_par_table,
    ngroups          = ngroups,
    dotdotdot        = dotdotdot,
    flat_model       = flat_model,
    model            = model, # in case model is a lavaan object
    nacov            = nacov,
    wls_v            = wls_v
  )
  lavdata <- temp$lavdata
  lavoptions <- temp$lavoptions

  timing <- lav_add_timing(timing, "Data")

  # ------------ lavpartable -------------------
  temp <- lav_step04_pt(
    slot_par_table = slot_par_table,
    model          = model,
    flat_model     = flat_model,
    lavoptions     = lavoptions,
    lavdata        = lavdata,
    constraints    = constraints
  )
  lavoptions <- temp$lavoptions
  lavpartable <- temp$lavpartable
  timing <- lav_add_timing(timing, "ParTable")

  # ------------ lavpta ------------------------
  # lavpta <- lav_lavaan_step04_pta(
  #   lavpartable = lavpartable,
  #   lavoptions  = lavoptions
  # )
  # timing <- lav_add_timing(timing, "lavpta")

  # ------------ lavsamplestats ---------------
  lavsamplestats <- lav_step05_samp(
    slot_sample_stats = slot_sample_stats,
    lavdata           = lavdata,
    lavoptions        = lavoptions,
    wls_v             = wls_v,
    nacov             = nacov,
    sample_cov        = sample_cov,
    sample_mean       = sample_mean,
    sample_th         = sample_th,
    sample_nobs       = sample_nobs,
    ov_names          = ov_names,
    ov_names_x        = ov_names_x,
    lavpartable       = lavpartable
  )
  timing <- lav_add_timing(timing, "SampleStats")

  # ------------ lavh1 ------------------------
  lavh1 <- lav_step06_h1(
    sloth1         = sloth1,
    lavoptions     = lavoptions,
    lavsamplestats = lavsamplestats,
    lavdata        = lavdata,
    lavpartable    = lavpartable
  )
  timing <- lav_add_timing(timing, "h1")

  # ------------ two-level WLS statistics -----
  # for two-level (D)WLS/ULS, the sample 'statistics' are the h1 estimates,
  # and their acov (Gamma) is the cluster-based sandwich; both can only be
  # computed once the h1 model has been estimated
  if (lavdata@nlevels > 1L &&
    lavoptions$estimator %in% c("WLS", "DWLS", "ULS") &&
    lavdata@data.type != "none") {
    if (lavoptions$.categorical) {
      # stage-wise (univariate + bivariate) estimation; this also
      # provides the 'h1' (unrestricted) implied statistics
      tmp <- lav_samp_wls_2l_cat(
        lavsamplestats = lavsamplestats,
        lavdata        = lavdata,
        lavoptions     = lavoptions
      )
      lavsamplestats <- tmp$lavsamplestats
      lavh1 <- tmp$lavh1
    } else {
      lavsamplestats <- lav_samp_wls_2l(
        lavsamplestats = lavsamplestats,
        lavh1          = lavh1,
        lavdata        = lavdata,
        lavoptions     = lavoptions
      )
    }
    timing <- lav_add_timing(timing, "wls2l")
  }

  # ------------ adapt marker ------------------
  # if the first indicator (the default marker) of a latent variable turns
  # out to be a poor item, switch to a better marker (and warn); this avoids
  # convergence problems caused by a (very) poor marker item. We keep the
  # original parameter table around: if the switched model fails to converge,
  # we retry with the original marker(s) further below, so that the
  # bad.marker.crit mechanism can never make convergence worse.
  lavpartable_orig <- lavpartable
  marker_switched  <- FALSE
  # composites use marker scaling (first weight fixed to 1) under BOTH
  # auto.fix.first and std.lv (see lav_partable_flat()), so the check must
  # also run when std.lv = TRUE and the model contains composites; the
  # 'clean default' guards in lav_pt_marker_adapt() keep reflective factors
  # (whose loadings are all free under std.lv) out of the check
  if ((isTRUE(lavoptions$auto.fix.first) ||
       (isTRUE(lavoptions$std.lv) && any(lavpartable$op == "<~"))) &&
      lavoptions$bad.marker.crit > 0) {
    adapt <- lav_pt_marker_adapt(
      lavpartable = lavpartable,
      lavh1       = lavh1,
      lavoptions  = lavoptions,
      lavdata     = lavdata,
      threshold   = lavoptions$bad.marker.crit
    )
    if (!is.null(adapt)) {
      info_fac <- adapt$info[adapt$info$type == "factor", , drop = FALSE]
      if (nrow(info_fac) > 0L) {
        lav_msg_warn(gettextf(
          "the first indicator of the following latent variable(s) is a poor
           item; switching to another marker item (to set the metric) to avoid
           convergence problems; use bad.marker.crit = 0 to switch off
           this behavior: %s",
          paste0(info_fac$lv, " (", info_fac$old, " -> ",
                 info_fac$new, ")", collapse = ", ")))
      }
      info_comp <- adapt$info[adapt$info$type == "composite", , drop = FALSE]
      if (nrow(info_comp) > 0L) {
        lav_msg_warn(gettextf(
          "the first indicator of the following composite(s) has a (near)
           zero implied weight; switching to another marker indicator (to set
           the metric) to avoid convergence problems; use bad.marker.crit = 0
           to switch off this behavior: %s",
          paste0(info_comp$lv, " (", info_comp$old, " -> ",
                 info_comp$new, ")", collapse = ", ")))
      }
      # rebuild the parameter table using the new marker(s)
      temp <- lav_step04_pt(
        slot_par_table = slot_par_table,
        model          = model,
        flat_model     = flat_model,
        lavoptions     = lavoptions,
        lavdata        = lavdata,
        constraints    = constraints,
        marker         = adapt$marker
      )
      lavpartable <- temp$lavpartable
      marker_switched <- TRUE
    }
  }

  # ------------ bounds + start + model + cache + optim ------------
  # steps 07-11 are wrapped in a local function so they can be re-run with the
  # original marker(s) as a fail-safe (see the marker switch above and the
  # retry just below)
  lav_bounds_to_optim <- function(lavpartable, timing) {
    # ------------ bounds ------------------------
    lavpartable <- lav_step07_bounds(
      lavoptions     = lavoptions,
      lavh1          = lavh1,
      lavdata        = lavdata,
      lavsamplestats = lavsamplestats,
      lavpartable    = lavpartable
    )
    timing <- lav_add_timing(timing, "bounds")

    # ------------ lavstart ----------------------
    lavpartable <- lav_step08_start(
      slot_model      = slot_model,
      lavoptions      = lavoptions,
      lavpartable     = lavpartable,
      lavsamplestats  = lavsamplestats,
      lavh1           = lavh1
    )
    timing <- lav_add_timing(timing, "start")

    # ------------ model -------------------------
    temp <- lav_step09_model(
      slot_model     = slot_model,
      lavoptions     = lavoptions,
      lavpartable    = lavpartable,
      lavsamplestats = lavsamplestats,
      lavdata        = lavdata
    )
    lavpartable <- temp$lavpartable
    lavmodel <- temp$lavmodel
    timing <- lav_add_timing(timing, "Model")

    # -------- lavcache ----------------------------------
    lavcache <- lav_step10_cache(
      slot_cache       = slot_cache,
      lavdata          = lavdata,
      lavmodel         = lavmodel,
      lavpartable      = lavpartable,
      lavoptions       = lavoptions,
      sampling_weights = sampling_weights
    )
    timing <- lav_add_timing(timing, "cache")

    # -------- est + lavoptim ----------------------------
    temp <- lav_step11_estoptim(
      lavdata        = lavdata,
      lavmodel       = lavmodel,
      lavcache       = lavcache,
      lavsamplestats = lavsamplestats,
      lavh1          = lavh1,
      lavoptions     = lavoptions,
      lavpartable    = lavpartable
    )
    timing <- lav_add_timing(timing, "optim")

    list(lavpartable = temp$lavpartable,
         lavmodel    = temp$lavmodel,
         lavcache    = lavcache,
         lavoptim    = temp$lavoptim,
         x           = temp$x,
         lavoptions  = temp$lavoptions,
         timing      = timing)
  }

  temp <- lav_bounds_to_optim(lavpartable, timing)

  # fail-safe: if we switched marker(s) but the model did not converge, retry
  # with the original marker(s). bad.marker.crit should never make convergence
  # worse: if the original (un-switched) model converges we keep that; if it
  # does not converge either, we still revert (so the result matches what
  # bad.marker.crit = 0 would have produced).
  if (marker_switched && !isTRUE(attr(temp$x, "converged"))) {
    temp_orig <- lav_bounds_to_optim(lavpartable_orig, temp$timing)
    if (isTRUE(attr(temp_orig$x, "converged"))) {
      lav_msg_note(gettext(
        "the model with the switched marker item(s) did not converge;
         reverting to the original marker item(s) (which did converge)."))
    } else {
      lav_msg_note(gettext(
        "the model with the switched marker item(s) did not converge;
         reverting to the original marker item(s)."))
    }
    temp <- temp_orig
  }

  lavpartable <- temp$lavpartable
  lavmodel    <- temp$lavmodel
  lavcache    <- temp$lavcache
  lavoptim    <- temp$lavoptim
  x           <- temp$x
  timing      <- temp$timing
  if (!is.null(temp$lavoptions)) {
    # e.g., the (quiet) em -> nlminb fallback: record the optimizer
    # that was actually used
    lavoptions <- temp$lavoptions
  }

  # store eqs if present in x
  laveqs <- list()
  if (!is.null(attr(x, "eqs"))) {
    laveqs <- attr(x, "eqs")
  }

  # -------- lavimplied + lavloglik --------------------
  lavimplied <- lav_step12_implied(
    lavoptions = lavoptions,
    lavmodel   = lavmodel
  )
  timing <- lav_add_timing(timing, "implied")

  lavloglik <- lav_step12_loglik(
    lavcache = lavcache,
    lavoptions     = lavoptions,
    lavdata        = lavdata,
    lavsamplestats = lavsamplestats,
    lavh1          = lavh1,
    lavimplied     = lavimplied,
    lavmodel       = lavmodel
  )
  timing <- lav_add_timing(timing, "loglik")

  # ----------- lavvcov + lavboot -------------------
  temp <- lav_step13_vcov_boot(
    lavoptions     = lavoptions,
    lavmodel       = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata        = lavdata,
    lavpartable    = lavpartable,
    lavcache       = lavcache,
    lavimplied     = lavimplied,
    lavh1          = lavh1,
    x              = x
  )
  lavpartable <- temp$lavpartable
  lavvcov <- temp$lavvcov
  vcov <- temp$VCOV
  lavmodel <- temp$lavmodel
  lavboot <- temp$lavboot
  lav_monte_carlo <- temp$lav_monte_carlo

  timing <- lav_add_timing(timing, "vcov")

  # ----------- lavtest ----------
  lavtest <- lav_step14_test(
    lavoptions     = lavoptions,
    lavmodel       = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata        = lavdata,
    lavpartable    = lavpartable,
    lavcache       = lavcache,
    lavimplied     = lavimplied,
    lavh1          = lavh1,
    x              = x,
    vcov           = vcov,
    lavloglik      = lavloglik
  )
  timing <- lav_add_timing(timing, "test")

  # ----------- lavfit ----------
  lavfit <- lav_step14_fit(
    lavpartable = lavpartable,
    lavmodel    = lavmodel,
    lavimplied  = lavimplied,
    x           = x,
    vcov        = vcov,
    lavtest     = lavtest
  )
  timing <- lav_add_timing(timing, "Fit")

  # ----------- baseline ----------------------------
  lavbaseline <- lav_step15_baseline(
    lavoptions     = lavoptions,
    lavsamplestats = lavsamplestats,
    lavdata        = lavdata,
    lavcache       = lavcache,
    lavh1          = lavh1,
    lavpartable    = lavpartable
  )
  timing <- lav_add_timing(timing, "baseline")

  # ----------- rotation ---------------------------
  temp <- lav_step16_rotation(
    lavoptions     = lavoptions,
    lavmodel       = lavmodel,
    lavpartable    = lavpartable,
    lavh1          = lavh1,
    lavdata        = lavdata,
    x              = x,
    lavvcov        = lavvcov,
    vcov           = vcov,
    lavcache       = lavcache,
    lavimplied     = lavimplied,
    lavsamplestats = lavsamplestats
  )
  lavpartable <- temp$lavpartable
  lavmodel <- temp$lavmodel
  lavvcov <- temp$lavvcov
  # rotated bootstrap (se = "bootstrap" + rotation) is computed in step 16,
  # after the rotated parameter table exists; keep step 13's lavboot otherwise
  if (!is.null(temp$lavboot)) {
    lavboot <- temp$lavboot
  }

  timing <- lav_add_timing(timing, "rotation")

  # ------ lavaan result  ----------------
  out <- lav_step17_lavaan(
    lavmc          = lavmc,
    timing         = timing,
    lavoptions     = lavoptions,
    lavpartable    = lavpartable,
    lavdata        = lavdata,
    lavsamplestats = lavsamplestats,
    lavmodel       = lavmodel,
    lavcache       = lavcache,
    lavfit         = lavfit,
    lavboot        = lavboot,
    lavoptim       = lavoptim,
    lavimplied     = lavimplied,
    lavloglik      = lavloglik,
    lavvcov        = lavvcov,
    lavtest        = lavtest,
    lavh1          = lavh1,
    lavbaseline    = lavbaseline,
    laveqs         = laveqs,
    start_time0    = start_time0,
    lav_monte_carlo = lav_monte_carlo
  )

  # ----------- fit.by.level (multilevel only) ----------
  # fit the partially saturated models (one per level), providing
  # level-specific fit measures; see fitMeasures(..., level = )
  if (lavdata@nlevels > 1L && isTRUE(lavoptions$fit.by.level)) {
    fbl <- lav_object_fit_by_level(out)
    if (length(fbl) > 0L) {
      out@internal$fit.by.level <- fbl
    }
  }

  # restore random seed
  if (!is.null(temp_seed)) {
    assign(".Random.seed", temp_seed, envir = .GlobalEnv)             # nolint
  } else if (exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) {
    # initially there was no .Random.seed, but we created one along the way
    # clean up
    remove(".Random.seed", envir = .GlobalEnv)
  }

  out
}
# # # # # #
# # cfa # #
# # # # # #
cfa <- function(
  model = NULL,
  data = NULL,
  ordered = NULL,
  aux = NULL,
  sampling_weights = NULL,
  sample_cov = NULL,
  sample_mean = NULL,
  sample_th = NULL,
  sample_nobs = NULL,
  group = NULL,
  cluster = NULL,
  constraints = "",
  wls_v = NULL,
  nacov = NULL,
  ov_order = "model",
  ...) {
  sc <- sys.call()
  sc[["model.type"]] <- quote("cfa")
  sc[["cmd"]] <- quote("cfa")
  # call mother function
  sc[[1L]] <- quote(lavaan::lavaan)
  eval(sc, parent.frame())
}
# # # # # #
# # sem # #
# # # # # #
sem <- function(
    model = NULL,
    data = NULL,
    ordered = NULL,
    aux = NULL,
    sampling_weights = NULL,
    sample_cov = NULL,
    sample_mean = NULL,
    sample_th = NULL,
    sample_nobs = NULL,
    group = NULL,
    cluster = NULL,
    constraints = "",
    wls_v = NULL,
    nacov = NULL,
    ov_order = "model",
    ...) {
  sc <- sys.call()
  sc[["model.type"]] <- quote("sem")
  sc[["cmd"]] <- quote("sem")
  # call mother function
  sc[[1L]] <- quote(lavaan::lavaan)
  eval(sc, parent.frame())
}
# # # # # # # #
# # growth  # #
# # # # # # # #
growth <- function(
    model = NULL,
    data = NULL,
    ordered = NULL,
    aux = NULL,
    sampling_weights = NULL,
    sample_cov = NULL,
    sample_mean = NULL,
    sample_th = NULL,
    sample_nobs = NULL,
    group = NULL,
    cluster = NULL,
    constraints = "",
    wls_v = NULL,
    nacov = NULL,
    ov_order = "model",
    ...) {
  sc <- sys.call()
  sc[["model.type"]] <- quote("growth")
  sc[["cmd"]] <- quote("growth")
  # call mother function
  sc[[1L]] <- quote(lavaan::lavaan)
  eval(sc, parent.frame())
}

# # # # # # # # # # # # # # # # # # #
# # help function lav_add_timing  # #
# # # # # # # # # # # # # # # # # # #
lav_add_timing <- function(timing, part) {
  # timing is a list with element start_time
  # this function adds an element with name as specified in parameter part
  # and the duration of the interval from start_time up to now
  # thereafter the element start_time is set to now (prepare for next call)
  # the adapted list is returned
  timenow <- proc.time()[3]
  timing[[part]] <- (timenow - timing$start_time)
  timing$start_time <- timenow

  timing
}

lav_beziersq_beziersc <- function(x) {
  # x contains the beziers points P1, P, P2 (P = control) for quadratic beziers
  # x is a matrix with 2 rows and 3 columns
  # returns rtval which contains P1, C1, C2, P2 so that the cubic beziers
  # with control points C1 and C2 is as 'high' as the quadratic one
  # rtval is a matrix with 2 rows and 4 columns
  matrix(c(x[, 1L], x[, 1L] / 3 + 2 * x[, 2L] / 3,
         x[, 3L] / 3 + 2 * x[, 2L] / 3, x [, 3L]), nrow = 2)
}
lav_node_coordinates <- function(nodeid, anker, nodes, maxrij) {
  nodenr <- which(nodes$id == nodeid)
  middelpunt <- c(nodes$kolom[nodenr], maxrij - nodes$rij[nodenr])
  delta <- switch(anker, n = c(0, 0.3), ne = c(0.3, 0.3), e = c(0.3, 0),
                  se = c(0.3, -0.3), s = c(0, -0.3), sw = c(-0.3, -0.3),
                  w = c(-0.3, 0), nw = c(-0.3, 0.3))
  middelpunt + delta
}
lav_plotinfo_tikzcode <- function(plotinfo,
                          outfile = "",
                          cex = 1.3,
                          sloped_labels = TRUE,
                          standalone = FALSE,
                          mlovcolors = c("lightgreen", "lightblue"),
                          lightness = 1,
                          italic = TRUE,
                          auto_subscript = TRUE
                          ) {
  tmpcol <- col2rgb(mlovcolors)
  wovcol <- paste(round(tmpcol[, 1L] / 255, 2), collapse = ",")
  bovcol <- paste(round(tmpcol[, 2L] / 255, 2), collapse = ",")
  nodenaam <- function(nm, blk) {
    if (blk > 0L) return(gsub("_", "", paste0("B", blk, nm), fixed = TRUE))
    gsub("_", "", nm, fixed = TRUE)
  }
  mlrij <- plotinfo$mlrij
  if (is.null(mlrij))
    lav_msg_stop(gettext(
      "plotinfo hasn't been processed by lav_plotinfo_positions!"))
  nodes <- plotinfo$nodes
  edges <- plotinfo$edges
  if (lightness != 1) {
    mlrij <- lightness * mlrij
    nodes$kolom <- lightness * nodes$kolom
    nodes$rij <- lightness * nodes$rij
    edges$controlpt.kol <- lightness * edges$controlpt.kol
    edges$controlpt.rij <- lightness * edges$controlpt.rij
  }
  if (is.character(outfile)) {
    zz <- file(outfile, open = "w")
    closezz <- TRUE
  } else {
    zz <- outfile
    closezz <- FALSE
  }
  if (standalone) writeLines(c(
    "\\documentclass{article}",
    "\\usepackage{amsmath, amssymb}",
    "\\usepackage{amsfonts}",
    "\\usepackage[utf8]{inputenc}",
    "\\usepackage[english]{babel}",
    "\\usepackage{color}",
    "\\usepackage{tikz}"), zz)
  commstyle <- paste0("draw, minimum size=", round(6 * cex), "mm")
  writeLines(c(
    "\\usetikzlibrary {shapes.geometric}",
    paste0("\\definecolor{wovcol}{rgb}{", wovcol, "}"),
    paste0("\\definecolor{bovcol}{rgb}{", bovcol, "}"),
    "\\tikzset{",
    ">=stealth,",
    paste0("x={(", cex, "cm,0cm)}, y={(0cm,", cex, "cm)},"),
    paste0("lv/.style={circle, ", commstyle, ", thick},"),
    paste0("varlv/.style={circle, draw, minimum size=",
     round(4 * cex), "mm, semithick},"),
    paste0("cv/.style={regular polygon, regular polygon sides=6, ",
     commstyle, ", thick},"),
    paste0("ov/.style={rectangle, ", commstyle, ", thick},"),
    paste0("wov/.style={rectangle, rounded corners, fill=wovcol, ",
     commstyle, ", thick},"),
    paste0("bov/.style={rectangle, rounded corners, fill=bovcol, ",
     commstyle, ", thick},"),
    paste0("const/.style={regular polygon, regular polygon sides=3, ",
     commstyle, ", thick}"),
    "}"), zz)
  if (standalone) writeLines("\\begin{document}", zz)
  writeLines("\\begin{tikzpicture}", zz)
  maxrij <- max(nodes$rij)
  maxcol <- max(nodes$kolom)
  if (mlrij > 0L) {
    writeLines(paste("\\draw (0, ", maxrij - mlrij, ") -- (",
     maxcol, ",", maxrij - mlrij, ");", sep = ""), zz)
  }
  for (j in seq.int(nrow(nodes))) {
    xpos <- nodes$kolom[j]
    ypos <- maxrij - nodes$rij[j]
    writeLines(paste(
      "\\node[", nodes$tiepe[j], "] (",
      nodenaam(nodes$naam[j], nodes$blok[j]),
      ") at (", xpos, ",", ypos, ") {",
      lav_label_code(nodes$naam[j], italic = italic,
                    auto_subscript = auto_subscript)$tikz,
                    "};", sep = ""), zz)
  }
  for (j in seq.int(nrow(edges))) {
    van <- which(nodes$id == edges$van[j])
    vannaam <- nodenaam(nodes$naam[van], nodes$blok[van])
    naar <- which(nodes$id == edges$naar[j])
    naarnaam <- nodenaam(nodes$naam[naar], nodes$blok[naar])
    nodelabel <- lav_label_code(edges$label[j], italic = italic,
                                  auto_subscript = auto_subscript)$tikz
    if (van == naar) { # self
      if (nodes$kolom[van] == 1L) {
        writeLines(paste("\\path[<->] (", vannaam,
                         ") edge [in=160, out=-160, looseness=8] node[right] {",
                         nodelabel, "} (",
                         vannaam, ");",
                         sep = ""), zz)
      } else if (nodes$rij[van] == maxrij) {
        writeLines(paste("\\path[<->] (", vannaam,
                         ") edge [in=-110, out=-70, looseness=8] node[above] {",
                         nodelabel, "} (",
                         vannaam, ");",
                         sep = ""), zz)
      } else if (nodes$kolom[van] == maxcol) {
        writeLines(paste("\\path[<->] (", vannaam,
                         ") edge [in=20, out=-20, looseness=8] node[left] {",
                         nodelabel, "} (",
                         vannaam, ");",
                         sep = ""), zz)
      } else {
        writeLines(paste("\\path[<->] (", vannaam,
                         ") edge [in=110, out=70, looseness=8] node[below] {",
                         nodelabel, "} (",
                         vannaam, ");",
                         sep = ""), zz)
      }
    } else {
      anchorv <- switch(edges$vananker[j],
                        n = ".north", e = ".east", s = ".south", w = ".west")
      anchorn <- switch(edges$naaranker[j],
                        n = ".north", e = ".east", s = ".south", w = ".west")
      if (is.na(edges$controlpt.kol[j])) {
        pathtype <- " -- "
      } else {
        vanadr <- lav_node_coordinates(edges$van[j], edges$vananker[j],
          nodes, maxrij)
        naaradr <- lav_node_coordinates(edges$naar[j], edges$naaranker[j],
          nodes, maxrij)
        controlq <- c(edges$controlpt.kol[j], maxrij - edges$controlpt.rij[j])
        beziersc <- lav_beziersq_beziersc(
          matrix(c(vanadr, controlq, naaradr), nrow = 2L)
        )
        pathtype <- paste0(" .. controls (", beziersc[1L, 2L], ",",
                           beziersc[2L, 2L], ") and (", beziersc[1L, 3L], ",",
                           beziersc[2L, 3L], ") .. ")
      }
      thelabel <- lav_label_code(edges$label[j], italic = italic,
                                   auto_subscript = auto_subscript)$tikz
      if (thelabel != "") {
        thelabel <- paste0("node[pos=0.5,",
                           ifelse(edges$labelbelow[j], "below", "above"),
                           ifelse(sloped_labels, ",sloped", ""),
                           "] {", thelabel, "} ")
      }
      pijl <- ifelse(edges$tiepe[j] %in% c("~~", "~~~"), "<->", "->")
      writeLines(paste0("\\draw[", pijl, "] (", vannaam, anchorv, ")",
                pathtype, "(", naarnaam, anchorn, ") ",
                thelabel, ";", sep = ""), zz)
    }
  }
  writeLines("\\end{tikzpicture}", zz)
  if (standalone) writeLines("\\end{document}", zz)
  if (closezz) close(zz)
  invisible(NULL)
}

lav_step17_lavaan <- function(lavmc = NULL,
                              timing = NULL,
                              lavoptions = NULL,
                              lavpartable = NULL,
                              lavdata = NULL,
                              lavsamplestats = NULL,
                              lavmodel = NULL,
                              lavcache = NULL,
                              lavfit = NULL,
                              lavboot = NULL,
                              lavoptim = NULL,
                              lavimplied = NULL,
                              lavloglik = NULL,
                              lavvcov = NULL,
                              lavtest = NULL,
                              lavh1 = NULL,
                              lavbaseline = NULL,
                              laveqs = NULL,
                              start_time0 = NULL,
                              lav_monte_carlo = NULL) {
  # # # # # # # # # #
  # #  17. lavaan # #
  # # # # # # # # # #

  # stop timer
  # create lavaan object
  # if lavmodel@nefa > 0
  #   compute standardizedSolution and store in ParTable slot in lavaan object
  # if post-checking demanded and converged, execute
  #   lavInspect(lavaan, "post.check")
  #
  timing$total <- (proc.time()[3] - start_time0)
  timing$start.time <- NULL
  lavpta <- lav_pt_attributes(lavpartable)
  lavpartable <- lav_pt_remove_cache(lavpartable)
  lavaan <- new("lavaan", # type_of_slot - where created or modified ?
    # ------------   ------------------------- -
    version = packageDescription("lavaan", fields = "Version"),
    call = lavmc, # match.call - lav_step00_parameters
    timing = timing, # list - ldw_add_timing
    Options = lavoptions, # list - options (2) / data (3) / partable (4)
    ParTable = lavpartable,
    # list - partable/bounds/start/model/estoptim/vcovboot/rotation
    pta = lavpta, # list - lav_partable_attributes
    Data = lavdata, # S4 class - data (3)
    SampleStats = lavsamplestats, # S4 class - samplestats (5)
    Model = lavmodel, # S4 class - model (9) / estoptim (11) / vcovboot (13)
    Cache = lavcache, # list - cache (10)
    Fit = lavfit, # S4 class - lav_model_fit (14bis)
    boot = lavboot, # list - vcovboot (13)
    optim = lavoptim, # list - estoptim (11)
    implied = lavimplied, # list - lav_model_implied (12)
    loglik = lavloglik, # list - lav_model_loglik (12)
    vcov = lavvcov, # list - vcovboot (13)
    test = lavtest, # list - test (14)
    h1 = lavh1, # list - h1 (6)
    baseline = lavbaseline, # list - baseline (15)
    internal = if (is.null(lav_monte_carlo) || length(lav_monte_carlo) == 0L) {
      list()
    } else {
      list(monte.carlo = lav_monte_carlo)
    },
    external = list() # empty list
  )

  # if model.type = "efa", add standardized solution to partable
  if (lavmodel@nefa > 0L) {
    if (lav_verbose()) {
      cat("computing standardized solution ... ")
    }
    std <- standardizedSolution(lavaan,
      remove_eq = FALSE,
      remove_ineq = FALSE, remove_def = FALSE
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }
    lavaan@ParTable$est.std <- std$est.std
  if (!is.null(std$se)) {
      lavaan@ParTable$se.std <- std$se
    }
  }

  # eqs?
  if (length(laveqs) > 0L) {
    lavaan@internal$eqs <- laveqs
  }

  # post-fitting check of parameters
  if (!is.null(lavoptions$check.post) && lavoptions$check.post &&
    lavTech(lavaan, "converged")) {
    if (lav_verbose()) {
      cat("post check  ...")
    }
    lavInspect(lavaan, "post.check")
    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  lavaan
}

# lavPredict() contains a collection of `predict' methods
# the unifying theme is that they all rely on the (unknown, to be estimated)
# or (known, apriori specified) values for the latent variables
#
# lv: latent variables (aka `factor scores')
# ov: predict linear part of y_i
#
# - YR 11 June 2013: first version, in order to get factor scores for the
#                    categorical case
# - YR 12 Jan 2014: refactoring + lav_predict_fy (to be used by estimator MML)
#
# - YR 25 Mar 2025: transformed factor scores (correlation-preserving)

# overload standard R function `predict'
setMethod(
  "predict", "lavaan",
  function(object, newdata = NULL, ...) {
    dotdotdot <- list(...)
    if (length(dotdotdot) > 0L) {
      for (j in seq_along(dotdotdot)) {
        lav_msg_warn(gettextf(
          "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
          sQuote("predict"))
        )
      }
    }
    lavPredict(
      object = object, newdata = newdata, type = "lv", method = "EBM",
      fsm = FALSE, rel = FALSE, optim_method = "bfgs"
    )
  }
)

# efaList version
lav_efalist_predict <- function(object, ...) {
  # kill object$loadings if present
  object[["loadings"]] <- NULL

  if (length(object) == 1L) {
    # unlist
    object <- object[[1]]
  } else {
    # use the 'last' one per default
    object <- object[[length(object)]]
  }

  predict(object, ...)
}

# public function
lavPredict <- function(object, newdata = NULL, # keep order of predict(), 0.6-7 # nolint
                       type = "lv", method = "EBM", transform = FALSE,
                       se = "none", acov = "none", label = TRUE, fsm = FALSE,
                       mdist = FALSE, rel = FALSE,
                       append_data = FALSE, assemble = FALSE, # or TRUE?
                       level = 1L, optim_method = "bfgs", eta = NULL,
                       parallel = c("auto", "no", "multicore", "snow"),
                       ncpus = NULL, cl = NULL,
                       drop_list_single_group = TRUE,
                       mdist_draws = 2000L,
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # check object
  object <- lav_object_check_version(object)

  parallel <- match.arg(parallel)

  # catch efaList objects
  if (inherits(object, "efaList")) {
    # kill object$loadings if present
    object[["loadings"]] <- NULL
    if (length(object) == 1L) {
      # unlist
      object <- object[[1]]
    } else {
      # use the 'last' one per default
      object <- object[[length(object)]]
    }
  }

  stopifnot(inherits(object, "lavaan"))
  lavmodel <- object@Model
  lavdata <- object@Data
  lavsamplestats <- object@SampleStats
  lavh1 <- object@h1
  lavimplied <- object@implied
  lavpartable <- object@ParTable

  # random slopes? use the dedicated empirical Bayes route:
  # - level = 2: posterior means of the between latent variables
  #   (including the random slopes); se = "standard" gives the
  #   posterior standard deviations
  # - level = 1: within factor scores, conditional on the posterior
  #   means of the cluster-level random effects (exact posterior means)
  if (length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L) {
    if (tolower(type) != "lv") {
      lav_msg_stop(gettext(
        "for models with random slopes, only type = \"lv\" is
         supported (for now)."))
    }
    if (!tolower(method) %in% c("ebm", "regression")) {
      lav_msg_stop(gettext(
        "for models with random slopes, only method = \"EBM\" is
         supported (the posterior means)."))
    }
    if (!is.null(newdata)) {
      lav_msg_stop(gettext(
        "newdata is not supported (yet) for models with random
         slopes."))
    }
    if (fsm || mdist || rel || transform || append_data) {
      lav_msg_stop(gettext(
        "the fsm/mdist/rel/transform/append_data options are not
         supported (yet) for models with random slopes."))
    }
    se_flag <- !is.null(se) && se != "none"
    if (se_flag && level == 1L) {
      lav_msg_stop(gettext(
        "standard errors for the level-1 factor scores are not
         available (yet) for models with random slopes."))
    }
    eb <- lav_mvn_cl_rs_eb(
      lavmodel = lavmodel, lavdata = lavdata,
      lavcache = object@Cache, se = se_flag
    )
    if (level == 2L) {
      out1 <- eb$l2
      se1 <- eb$se2
    } else {
      out1 <- eb$l1
      se1 <- NULL
    }
    if (!label) {
      colnames(out1) <- NULL
    }
    class(out1) <- c("lavaan.matrix", "matrix")
    if (!drop_list_single_group) {
      out1 <- list(out1)
      if (!is.null(se1)) {
        se1 <- list(se1)
      }
    }
    if (se_flag) {
      attr(out1, "se") <- se1
    }
    return(out1)
  }

  # warn if the model does not contain any 'regular' latent variables
  if (length(lav_object_vnames(lavpartable, "lv.regular")) == 0L) {
    lav_msg_warn(gettextf("fitted model does not contain regular
    (i.e., measured) latent variables; the matrix of factor scores may
    contain no columns"))
  }

  res <- lav_predict_internal(
    lavmodel = lavmodel, lavdata = lavdata,
    lavsamplestats = lavsamplestats, lavimplied = lavimplied, lavh1 = lavh1,
    lavpartable = lavpartable, newdata = newdata, type = type, method = method,
    transform = transform, se = se, acov = acov, label = label,
    fsm = fsm, rel = rel,
    mdist = mdist, append_data = append_data, assemble = assemble,
    level = level, optim_method = optim_method, eta = eta,
    parallel = parallel, ncpus = ncpus, cl = cl,
    drop_list_single_group = drop_list_single_group,
    mdist_draws = mdist_draws
  )

  res
}

# internal version, to be used if lavobject does not exist yet
lav_predict_internal <- function(lavmodel = NULL,
                               lavdata = NULL,
                               lavsamplestats = NULL,
                               lavh1 = NULL,
                               lavimplied = NULL,
                               lavpartable = NULL,
                               # standard options
                               newdata = NULL, # keep order of predict(), 0.6-7
                               type = "lv", method = "EBM", transform = FALSE,
                               se = "none", acov = "none", label = TRUE,
                               fsm = FALSE, rel = FALSE,
                               mdist = FALSE,
                               append_data = FALSE, assemble = FALSE, # or TRUE?
                               level = 1L, optim_method = "bfgs", eta = NULL,
                               parallel = "no", ncpus = NULL, cl = NULL,
                               drop_list_single_group = TRUE,
                               mdist_draws = 2000L) {
  # type
  type <- tolower(type)
  lavpta <- lav_pt_attributes(lavpartable)
  if (type %in% c("latent", "lv", "factor", "factor.score", "factorscore")) {
    type <- "lv"
  } else if (type %in% c("ov", "yhat")) {
    type <- "yhat"
  } else if (type %in% c("residuals", "resid", "error")) {
    type <- "resid"
  }

  # if resid, not for categorical, unless mdist = TRUE: in that case the
  # expected residuals of the latent responses are computed by Monte Carlo
  # integration (Mansolf & Reise, 2017; see lav_predict_mdist_cat)
  if (type == "resid" && lavmodel@categorical && !mdist) {
    lav_msg_stop(gettext(
      "casewise residuals not available if data is categorical
      (unless mdist = TRUE)"))
  }

  # append_data? check level
  if (append_data && level > 1L) {
    lav_msg_warn(gettext("append_data not available if level > 1L"))
    append_data <- FALSE
  }

  # mdist? -> fsm = TRUE (continuous data only; for categorical data the
  # distances are computed by Monte Carlo integration, and no factor score
  # matrix is involved)
  mdist_cat_flag <- mdist && lavmodel@categorical
  if (mdist && !mdist_cat_flag) {
    fsm <- TRUE
  }

  # se?
  if (acov != "none") {
    se <- acov # acov implies se
  }
  if (se != "none") {
    if (is.logical(se) && se) {
      se <- "standard"
      if (acov != "none") {
        acov <- se # reverse-imply upstream
      }
    }
    if (type != "lv") {
      lav_msg_stop(gettext("standard errors only available if type = \"lv\""))
    }
    # NOTE: for categorical data, the standard errors are computed in
    # lav_predict_eta_ebm_ml() from the curvature of the (EBM/ML) objective at
    # the optimum, and so differ from one response pattern to the next.
    # if(lavdata@missing %in% c("ml", "ml.x")) {
    #    se <- acov <- "none"
    #    warning("lavaan WARNING: standard errors not
    #                 available (yet) for missing data + fiml")
    # }
  }

  # need full data set supplied
  if (is.null(newdata)) {
    # use internal copy:
    if (lavdata@data.type != "full") {
      lav_msg_stop(gettext(
        "sample statistics were used for fitting and newdata is empty"))
    } else if (is.null(lavdata@X[[1]])) {
      lav_msg_stop(gettext("no local copy of data; FIXME!"))
    } else {
      data_obs <- lavdata@X
      ov_names <- lavdata@ov.names
    }
    exo <- lavdata@eXo
  } else {
    ov <- lavdata@ov
    # for multilevel models, the cluster variable must be present in
    # newdata, so that a fresh cluster structure (Lp) can be built
    if (lavdata@nlevels > 1L &&
        !all(lavdata@cluster %in% names(newdata))) {
      lav_msg_stop(gettextf(
        "newdata does not contain the cluster variable(s): %s",
        lav_msg_view(lavdata@cluster, "none")))
    }
    new_data <- lav_lavdata(
      data = newdata,
      group = lavdata@group,
      cluster = lavdata@cluster,
      ov_names = lavdata@ov.names,
      ov_names_x = lavdata@ov.names.x,
      ov_names_l = lavdata@ov.names.l,
      ordered = ov$name[ov$type == "ordered"],
      lavoptions = list(
        std.ov = lavdata@std.ov,
        group.label = lavdata@group.label,
        level.label = lavdata@level.label,
        missing = lavdata@missing
      ), # was FALSE before?
      allow_single_case = TRUE
    )
    # if ordered, check if number of levels is still the same (new in 0.6-7)
    if (lavmodel@categorical) {
      orig_ordered_idx <- which(lavdata@ov$type == "ordered")
      orig_ordered_lev <- lavdata@ov$nlev[orig_ordered_idx]
      match_new_idx <- match(
        lavdata@ov$name[orig_ordered_idx],
        new_data@ov$name
      )
      new_ordered_lev <- new_data@ov$nlev[match_new_idx]
      if (any(orig_ordered_lev - new_ordered_lev != 0)) {
        lav_msg_stop(
          gettext("mismatch number of categories for some ordered variables
                  in newdata compared to original data.")
        )
      }
    }
    data_obs <- new_data@X
    exo <- new_data@eXo
    ov_names <- new_data@ov.names

    # for multilevel models, replace lavdata by the freshly-built
    # newdata object: the downstream code needs the newdata cluster
    # structure (Lp), missing patterns (Mp) and dimensions (new in
    # 0.7-1); lavsamplestats is kept as-is (only used for the
    # model-implied quantities)
    if (lavdata@nlevels > 1L) {
      lavdata <- new_data
    }
  }

  # categorical data + mdist: expected (squared) M-distances of the latent
  # responses, via Monte Carlo integration (Mansolf & Reise, 2017)
  if (mdist_cat_flag) {
    if (!type %in% c("lv", "resid", "yhat")) {
      lav_msg_stop(gettext(
        "mdist is only available if type is one of: lv resid yhat"))
    }
    if (level > 1L) {
      lav_msg_stop(gettext(
        "mdist with categorical data is not available if level > 1L"))
    }
    mdist_cat <- lav_predict_mdist_cat(
      lavmodel = lavmodel, lavdata = lavdata, lavimplied = lavimplied,
      data_obs = data_obs, exo = exo,
      type = type, method = method, ndraws = mdist_draws
    )
    # (squared metric, as in the continuous case)
    mdist_1 <- lapply(mdist_cat, "[[", "d2")
  }

  if (type == "lv") {
    if (!is.null(eta)) {
      lav_msg_warn(gettext("lvs will be predicted here;
                           supplying eta= has no effect"))
    }

    # post fit check (lv pd?)
    # ok <- lav_object_post_check(object)
    # if(!ok) {
    #    stop("lavaan ERROR: lavInspect(,\"post.check\") is not TRUE;
    #       factor scores can not be computed. See the WARNING message.")
    # }

    out <- lav_predict_eta(
      lavobject = NULL, lavmodel = lavmodel,
      lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavimplied = lavimplied, lavpta = lavpta,
      se = se, acov = acov, level = level,
      data_obs = data_obs, exo = exo, method = method,
      fsm = fsm, rel = rel, transform = transform, optim_method = optim_method,
      parallel = parallel, ncpus = ncpus, cl = cl
    )

    # extract fsm here
    if (fsm) {
      fsm_1 <- attr(out, "fsm")
    }

  # extract rel here
  if (rel) {
    rel_1 <- attr(out, "rel")
  }

    # extract se here
    if (se != "none") {
      se_1 <- attr(out, "se")
      if (acov != "none") {
        acov_1 <- attr(out, "acov")
      }
    }

    # remove dummy lv? (removes attr!)
    out <- lapply(seq_len(lavdata@ngroups), function(g) {
      # determine block
      b <- lav_predict_block_idx(lavdata, g, level)
      lv_idx <- c(
        lavmodel@ov.y.dummy.lv.idx[[b]],
        lavmodel@ov.x.dummy.lv.idx[[b]]
      )
      ret <- out[[g]]
      if (length(lv_idx) > 0L) {
        ret <- out[[g]][, -lv_idx, drop = FALSE]
      }
      ret
    })

    # we need to remove the dummy's before we transform
    # (update 0.6-20: no longer needed... as transform happens internally)
    if (fsm) {
      fsm_1 <- lapply(seq_len(lavdata@ngroups), function(g) {
        # determine block
        b <- lav_predict_block_idx(lavdata, g, level)
        lv_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[b]],
          lavmodel@ov.x.dummy.lv.idx[[b]]
        )
        # ov.idx <- lavmodel@ov.x.dummy.ov.idx[[b]]
        # or should we use pta$vidx$ov.ind?
        ov_ind <- lavpta$vidx$ov.ind[[b]]
        ret <- fsm_1[[g]]
        if (length(lv_idx) > 0L) {
          if (is.matrix(fsm_1[[g]])) {
            ret <- fsm_1[[g]][-lv_idx, ov_ind, drop = FALSE]
          } else if (is.list(fsm_1[[g]])) {
            fsm_1[[g]] <- lapply(fsm_1[[g]], function(x) {
              ret <- x[-lv_idx, ov_ind, drop = FALSE]
              ret
            })
          }
        }
        ret
      })
    }

#   # new in 0.6-16
#   # we assume the dummy lv's have already been removed
#     if (transform) {
#       # VETA <- lav_model_veta(lavmodel = lavmodel, remove.dummy.lv = TRUE)
#       EETA <- lav_model_eeta(
#         lavmodel = lavmodel,
#         lavsamplestats = lavsamplestats, remove.dummy.lv = TRUE
#       )
#       # compute transformation matrix
#       if (tolower(method) %in% c("ebm", "regression")) {
#         tmat <- lav_predict_tmat_green(lavmodel = lavmodel,
#                                        lavimplied = lavimplied)
#       } else {
#         tmat <- lav_predict_tmat_det(lavmodel = lavmodel,
#                                      lavimplied = lavimplied)
#       }
#
#       # update fsm_1
#       if (fsm) {
#         fsm_1 <- lapply(seq_len(lavdata@ngroups), function(g) {
#           # determine block
#           if (lavdata@nlevels == 1L) {
#             b <- g
#           } else {
#             b <- (g - 1) * lavdata@nlevels + level
#           }
#           ret <- tmat[[b]] %*% fsm_1[[g]]
#           ret
#         })
#       }
#
#       out <- lapply(seq_len(lavdata@ngroups), function(g) {
#         # determine block
#         if (lavdata@nlevels == 1L) {
#           b <- g
#         } else {
#           b <- (g - 1) * lavdata@nlevels + level
#         }
#
#         FS.centered <- scale(out[[g]], center = TRUE, scale = FALSE)
#         #FS.cov <- crossprod(FS.centered) / nrow(FS.centered)
#         #FS.cov.inv <- try(solve(FS.cov), silent = TRUE)
#         #if (inherits(FS.cov.inv, "try-error")) {
#         #  lav_msg_warn(
#         #    gettext("could not invert (co)variance matrix of factor scores;
#         #            returning original factor scores."))
#         #  return(out[[g]])
#         #}
#         #fs.inv.sqrt <- lav_mat_sym_sqrt(FS.cov.inv)
#         #veta.sqrt <- lav_mat_sym_sqrt(VETA[[g]])
#         #tmp <- FS.centered %*% fs.inv.sqrt %*% veta.sqrt
#         tmp <- FS.centered %*% t(tmat[[b]])
#         ret <- t(t(tmp) + drop(EETA[[g]]))
#
#         ret
#       })
#     }

    # new in 0.6-17
    if (mdist && !mdist_cat_flag) {
      veta <- lav_model_veta(lavmodel = lavmodel, remove_dummy_lv = TRUE)
      eeta <- lav_model_eeta(
        lavmodel = lavmodel,
        lavsamplestats = lavsamplestats, remove_dummy_lv = TRUE
      )
      # under conditional.x, lavimplied has no joint 'cov'; the relevant Sigma
      # is Cov(y|x) = res.cov (partial matching would wrongly pick cov.x)
      implied_cov <- if (lavmodel@conditional.x) {
        lavimplied$res.cov
      } else {
        lavimplied$cov
      }
      mdist_1 <- lapply(seq_len(lavdata@ngroups), function(g) {
        a <- fsm_1[[g]]
        sigma <- implied_cov[[g]]
        if (transform) {
          fs_cov <- veta[[g]]
        } else {
          fs_cov <- a %*% sigma %*% t(a)
        }
        fs_cov_inv <- solve(fs_cov)
        # Mahalanobis distance
        fs_c <- t(t(out[[g]]) - eeta[[g]]) # center
        df_squared <- rowSums((fs_c %*% fs_cov_inv) * fs_c)
        ret <- df_squared # squared!
        ret
      })
    }

    # append original/new data? (also remove attr)
    if (append_data && level == 1L) {
      out <- lapply(seq_len(lavdata@ngroups), function(g) {
        ret <- cbind(out[[g]], data_obs[[g]])
        ret
      })
    }

    if (se != "none") {
      se_1 <- lapply(seq_len(lavdata@ngroups), function(g) {
        # determine block
        b <- lav_predict_block_idx(lavdata, g, level)
        lv_idx <- c(
          lavmodel@ov.y.dummy.lv.idx[[b]],
          lavmodel@ov.x.dummy.lv.idx[[b]]
        )
        ret <- se_1[[g]]
        if (length(lv_idx) > 0L) {
          ret <- se_1[[g]][, -lv_idx, drop = FALSE]
        }
        ret
      })
      if (acov != "none") {
        acov_1 <- lapply(seq_len(lavdata@ngroups), function(g) {
          # determine block
          b <- lav_predict_block_idx(lavdata, g, level)
          lv_idx <- c(
            lavmodel@ov.y.dummy.lv.idx[[b]],
            lavmodel@ov.x.dummy.lv.idx[[b]]
          )
          ret <- acov_1[[g]]
          if (length(lv_idx) > 0L) {
            if (is.matrix(acov_1[[g]])) {
              ret <- acov_1[[g]][-lv_idx, -lv_idx, drop = FALSE]
            } else if (is.list(acov_1[[g]])) {
              ret <- lapply(acov_1[[g]], function(x) {
                ret <- x[-lv_idx, -lv_idx, drop = FALSE]
                ret
              })
            }
          }
          ret
        })
      } # acov
    } # se

    # label?
    if (label) {
      for (g in seq_len(lavdata@ngroups)) {
        gg <- lav_predict_block_idx(lavdata, g, level)

        if (append_data) {
          colnames(out[[g]]) <- c(
            lavpta$vnames$lv[[gg]],
            ov_names[[g]]
          ) # !not gg
        } else {
          colnames(out[[g]]) <- lavpta$vnames$lv[[gg]]
        }

        if (fsm) {
          # column names = manifest indicators. For models with composites the
          # factor-score matrix also has columns for the composite indicators,
          # so use the full set of observed indicators (ov = ov.ind + ov.cind)
          # rather than ov.ind alone.
          fsm_ov <- if (lavmodel@composites) {
            lavpta$vnames$ov[[gg]]
          } else {
            lavpta$vnames$ov.ind[[gg]]
          }
          if (is.null(fsm_1[[g]])) {
            # skip
          } else if (is.matrix(fsm_1[[g]])) {
            if (ncol(fsm_1[[g]]) == length(fsm_ov)) {
              dimnames(fsm_1[[g]]) <- list(lavpta$vnames$lv[[gg]], fsm_ov)
            }
          } else if (is.list(fsm_1[[g]])) {
            fsm_1[[g]] <- lapply(fsm_1[[g]], function(x) {
              if (ncol(x) == length(fsm_ov)) {
                dimnames(x) <- list(lavpta$vnames$lv[[gg]], fsm_ov)
              }
              x
            })
          }
        }

        if (se != "none") {
          if (!is.null(se_1[[g]])) {
            colnames(se_1[[g]]) <- lavpta$vnames$lv[[gg]]
          }
        }

        if (rel) {
          if (!is.null(rel_1[[g]])) {
            names(rel_1[[g]]) <- lavpta$vnames$lv[[gg]]
          }
        }

        if (acov != "none") {
          if (is.null(acov_1[[g]])) {
            # skip
          } else if (is.matrix(acov_1[[g]])) {
            dimnames(acov_1[[g]]) <- list(
              lavpta$vnames$lv[[gg]],
              lavpta$vnames$lv[[gg]]
            )
          } else if (is.list(acov_1[[g]])) {
            acov_1[[g]] <- lapply(acov_1[[g]], function(x) {
              dimnames(x) <- list(
                lavpta$vnames$lv[[gg]],
                lavpta$vnames$lv[[gg]]
              )
              x
            })
          }
        }
      } # g

      # group.labels
      if (lavdata@ngroups > 1L) {
        names(out) <- lavdata@group.label
        if (se != "none") {
          names(se_1) <- lavdata@group.label
        }
        if (acov != "none") {
          names(acov_1) <- lavdata@group.label
        }
      }
    } # label

    # yhat: estimated value for the observed indicators, given (estimated)
    # factor scores
    # resid: y - yhat
  } else if (type %in% c("yhat", "resid")) {
    resid_flag <- type == "resid"
    if (resid_flag && mdist_cat_flag) {
      # categorical data: the expected casewise residuals of the latent
      # responses (a by-product of the Monte Carlo integration)
      out <- lapply(mdist_cat, "[[", "resid")
    } else {
      out <- lav_predict_yhat(
        lavobject = NULL, lavmodel = lavmodel,
        lavdata = lavdata, lavsamplestats = lavsamplestats,
        lavimplied = lavimplied,
        data_obs = data_obs, exo = exo,
        eta = eta, method = method, optim_method = optim_method,
        fsm = fsm,
        resid_flag = resid_flag
      )
      if (fsm) {
        fsm_1 <- attr(out, "fsm")
      }
    }

    # label?
    if (label) {
      for (g in seq_len(lavdata@ngroups)) {
        if (ncol(out[[g]]) == length(lavpta$vnames$ov[[g]])) {
          colnames(out[[g]]) <- lavpta$vnames$ov[[g]]
        } else {
          # conditional.x: the columns only cover the non-exogenous ov's
          colnames(out[[g]]) <- lavpta$vnames$ov.nox[[g]]
        }
      }
    }

    # mdist
    if (mdist && !mdist_cat_flag) {
      mm_lambda <- lav_model_lambda(
        lavmodel = lavmodel,
        remove_dummy_lv = FALSE,
        use_wmat = TRUE
      )
      # under conditional.x, lavimplied has no joint 'cov'; the relevant Sigma
      # is Cov(y|x) = res.cov (partial matching would wrongly pick cov.x)
      implied_cov <- if (lavmodel@conditional.x) {
        lavimplied$res.cov
      } else {
        lavimplied$cov
      }
      mdist_1 <- lapply(seq_len(lavdata@ngroups), function(g) {
        sigma <- implied_cov[[g]]
        la <- mm_lambda[[g]]
        if (type == "resid") {
          ila <- diag(ncol(sigma)) - la %*% fsm_1[[g]]
          omega_e <- ila %*% sigma %*% t(ila)
          eig <- eigen(omega_e, symmetric = TRUE)
          a <- eig$vectors[, seq_len(nrow(la) - ncol(la)),
            drop = FALSE
          ]
        } else if (type == "yhat") {
          laa <- la %*% fsm_1[[g]]
          omega_e <- laa %*% sigma %*% t(laa)
          eig <- eigen(omega_e, symmetric = TRUE)
          a <- eig$vectors[, seq_len(ncol(la)), drop = FALSE]
        }
        out_a <- apply(out[[g]], 1L, function(x) {
          colSums(a * x, na.rm = TRUE)
        })
        if (is.matrix(out_a)) {
          out_a <- t(out_a)
        } else {
          out_a <- as.matrix(out_a)
        }
        # if(lavmodel@meanstructure) {
        #     est.mean <- drop(t(lavimplied$mean[[g]]) %*% a)
        #     if(type == "resid") {
        #         obs.mean <- drop(lavh1$implied$mean[[g]] %*% a)
        #         est.mean <- drop(t(lavimplied$mean[[g]]) %*% a)
        #         outA.mean <- obs.mean - est.mean
        #     } else if(type == "yhat") {
        #         outA.mean <- est.mean
        #     }
        # } else {
        #     outA.mean <- colMeans(out_a)
        # }
        out_a_cov <- t(a) %*% omega_e %*% a
        out_a_cov_inv <- solve(out_a_cov)
        # Mahalanobis distance
        # out_a_c <- t( t(out_a) - outA.mean ) # center
        out_a_c <- out_a
        df_squared <- rowSums((out_a_c %*% out_a_cov_inv) * out_a_c)
        ret <- df_squared # squared!
        ret
      })
    }


    # density for each observed item, given (estimated) factor scores
  } else if (type == "fy") {
    out <- lav_predict_fy(
      lavobject = NULL, lavmodel = lavmodel,
      lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavimplied = lavimplied,
      data_obs = data_obs, exo = exo,
      eta = eta, method = method, optim_method = optim_method
    )

    # label?
    if (label) {
      for (g in seq_len(lavdata@ngroups)) {
        colnames(out[[g]]) <- lavpta$vnames$ov[[g]]
      }
    }
  } else {
    lav_msg_stop(gettext("type must be one of: lv yhat fy"))
  }

  # lavaan.matrix
  out <- lapply(out, "class<-", c("lavaan.matrix", "matrix"))

  if (lavdata@ngroups == 1L && drop_list_single_group) {
    res <- out[[1L]]
  } else {
    res <- out
  }

  # assemble multiple groups into a single data.frame? (new in 0.6-4)
  if (lavdata@ngroups > 1L && assemble) {
    if (!is.null(newdata)) {
      lavdata <- new_data
    }
    data_1 <- matrix(as.numeric(NA),
      nrow = sum(unlist(lavdata@norig)),
      ncol = ncol(out[[1L]])
    ) # assume == per g
    colnames(data_1) <- colnames(out[[1L]])
    for (g in seq_len(lavdata@ngroups)) {
      data_1[lavdata@case.idx[[g]], ] <- out[[g]]
    }
    data_1 <- as.data.frame(data_1, stringsAsFactors = FALSE)

    if (!is.null(newdata)) {
      data_1[, lavdata@group] <- newdata[, lavdata@group]
    } else {
      # add group
      data_1[, lavdata@group] <- rep(as.character(NA), nrow(data_1))
      if (lavdata@missing == "listwise") {
        # we will lose the group label of omitted variables!
        data_1[unlist(lavdata@case.idx), lavdata@group] <-
          rep(lavdata@group.label, unlist(lavdata@nobs))
      } else {
        data_1[unlist(lavdata@case.idx), lavdata@group] <-
          rep(lavdata@group.label, unlist(lavdata@norig))
      }
    }

    res <- data_1
  }

  if (fsm && type == "lv") {
    attr(res, "fsm") <- fsm_1
  }

  if (rel && type == "lv") {
    attr(res, "rel") <- rel_1
  }

  if (mdist) {
    attr(res, "mdist") <- mdist_1
  }

  if (se != "none") {
    attr(res, "se") <- se_1
    # return full sampling covariance matrix?
    if (acov == "standard") {
      attr(res, "acov") <- acov_1
    }
  }

  res
}

# map a group index 'g' (and a level) to the corresponding block index 'b'
# in the per-block model matrices (GLIST, veta, eeta, ...). For single-level
# models the block is just the group; for multilevel models the blocks are
# interleaved per group as (g - 1) * nlevels + level.
lav_predict_block_idx <- function(lavdata, g, level = 1L) {
  if (lavdata@nlevels == 1L) {
    g
  } else {
    (g - 1L) * lavdata@nlevels + level
  }
}

# internal function
lav_predict_eta <- function(lavobject = NULL, # for convenience
                            # sub objects
                            lavmodel = NULL, lavdata = NULL,
                            lavsamplestats = NULL,
                            lavimplied = NULL, lavpta = NULL,
                            # new data
                            data_obs = NULL, exo = NULL,
                            # options
                            method = "EBM",
                            fsm = FALSE,
              rel = FALSE,
                            transform = FALSE,
                            se = "none", acov = "none",
                            level = 1L,
                            optim_method = "bfgs",
                            parallel = "no", ncpus = NULL, cl = NULL) {
  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavdata <- lavobject@Data
  } else {
    stopifnot(!is.null(lavdata))
  }

  # method
  method <- tolower(method)

  # alias
  if (method == "regression") {
    method <- "ebm"
  } else if (method == "bartlett" || method == "bartlet") {
    method <- "ml"
  }

  # at this point, method is either "ebm" (= regression/EB) or "ml" (= Bartlett)
  if (!method %in% c("ebm", "ml")) {
    lav_msg_stop(gettextf("unknown method: %s.", method))
  }

  if (all(lavdata@ov$type == "numeric")) {
    # normal case: closed-form regression (EBM) or Bartlett (ML) scores
    out <- lav_predict_eta_normal(
      lavobject = lavobject,
      lavmodel = lavmodel, lavdata = lavdata,
      lavimplied = lavimplied, lavpta = lavpta, se = se, acov = acov,
      level = level, lavsamplestats = lavsamplestats,
      data_obs = data_obs, exo = exo, fsm = fsm, rel = rel,
      transform = transform, method = method
    )
  } else {
    # non-normal case: numerical optimisation of the latent scores
    out <- lav_predict_eta_ebm_ml(
      lavobject = lavobject,
      lavmodel = lavmodel, lavdata = lavdata,
      lavsamplestats = lavsamplestats, se = se, acov = acov,
      level = level, data_obs = data_obs, exo = exo,
      ml = (method == "ml"), optim_method = optim_method,
      parallel = parallel, ncpus = ncpus, cl = cl
    )
  }

  out
}


# (pseudo-)inverse used in the factor-score machinery.
# For a converged model the model-implied Sigma is positive definite, and
# (lambda' Sigma.inv lambda) is invertible in all but a few "by construction"
# singular settings (e.g. a single-indicator construct with a fixed zero
# residual variance). In particular, for models with composites both matrices
# are well-conditioned, so we use an exact (and fast) solve() here; we only fall
# back to the Moore-Penrose pseudo-inverse for the genuinely singular cases that
# originally motivated MASS::ginv(). This keeps the common path exact and avoids
# relying on ginv() for composites.
lav_predict_solve <- function(x) {
  out <- try(solve(x), silent = TRUE)
  if (inherits(out, "try-error")) {
    out <- MASS::ginv(x)
  }
  # drop dimnames so the result behaves exactly like MASS::ginv() downstream
  dimnames(out) <- NULL
  out
}

# factor scores - normal case
# NOTE: this is the classic 'regression' method; for the linear/continuous
#       case, this is equivalent to both EB and EBM
# factor scores - normal (continuous) case - closed-form
#
# Two methods are supported, differing only in the factor-score coefficient
# matrix 'C' (mapping centered data to latent scores) and in the corresponding
# conditional covariance used for standard errors:
#
#  - method = "regression" (aka EB/EBM/Thurstone): C = Psi Lambda' Sigma.inv
#       For the linear/continuous case this is equivalent to both EB and EBM.
#
#  - method = "bartlett" (aka ML): C = (Lambda' Sigma.inv Lambda).inv Lambda'
#                                      Sigma.inv
#       The usual formula uses Theta.inv, but to deal with a singular THETA
#       (zeroes on the diagonal) we use the 'GLS' version with Sigma.inv
#       instead (Bentler & Yuan, 1997, 'Optimal Conditionally Unbiased
#       Equivariant Factor Score Estimators', in Berkane (Ed) 'Latent variable
#       modeling with applications to causality', Springer-Verlag).
#
# NOTE: instead of solve(), we use lav_predict_solve() (a pseudo-inverse
#       fallback) for the special settings where -by construction-
#       (Lambda' Sigma.inv Lambda) is singular. For Bartlett scores this
#       destroys the conditionally-unbiased property.
lav_predict_eta_normal <- function(lavobject = NULL, # for convenience
                                   # sub objects
                                   lavmodel = NULL, lavdata = NULL,
                                   lavsamplestats = NULL,
                                   lavimplied = NULL, lavpta = NULL,
                                   # optional new data
                                   data_obs = NULL, exo = NULL,
                                   se = "none", acov = "none", level = 1L,
                                   fsm = FALSE, rel = FALSE,
                                   transform = FALSE,
                                   method = "regression") {
  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
    lavimplied <- lavobject@implied
    lavpta <- lavobject@pta
  } else {
    stopifnot(
      !is.null(lavmodel), !is.null(lavdata),
      !is.null(lavsamplestats), !is.null(lavimplied)
    )
  }

  # Bartlett (ML) or regression (EB/EBM) scores?
  bartlett <- tolower(method) %in% c("ml", "bartlett", "bartlet")

  # method-specific factor-score coefficient matrix 'C' (maps centered data to
  # latent scores), as a function of Lambda, Sigma.inv and Psi (= veta_g)
  #
  # higher-order factors (Bartlett only): a higher-order factor has no
  # observed indicators, so its column in Lambda is empty, and the
  # (pseudo-inverse based) Bartlett mapping returns all-zero scores for this
  # factor; if lambda_star (the 'collapsed' loadings, see below) is provided,
  # the rows for the higher-order factors are taken from the Bartlett mapping
  # based on lambda_star instead (the other rows are left untouched)
  compute_fsc <- function(lambda, sigma_inv, veta_g, lambda_star = NULL,
                          fixes = NULL) {
    if (bartlett) {
      fsc <- lav_predict_solve(t(lambda) %*% sigma_inv %*% lambda) %*%
        t(lambda) %*% sigma_inv
      if (!is.null(lambda_star)) {
        for (fx in fixes) {
          ls_t <- lambda_star[, fx$tcols, drop = FALSE]
          fsc_star <- lav_predict_solve(
            t(ls_t) %*% sigma_inv %*% ls_t) %*% t(ls_t) %*% sigma_inv
          fsc[fx$cols, ] <- fsc_star[match(fx$cols, fx$tcols), ,
                                     drop = FALSE]
        }
      }
      fsc
    } else {
      veta_g %*% t(lambda) %*% sigma_inv
    }
  }

  # method-specific conditional (co)variance of the factor scores, used for the
  # standard errors and the full sampling covariance matrix (acov)
  #
  # if 'fsc' (and 'sigma') are provided (higher-order factors, Bartlett only),
  # the general expression Var(fs - eta) = C Sigma C' - C Lambda Psi -
  # Psi Lambda' C' + Psi is used instead: the standard expression below
  # assumes C Lambda = I, which no longer holds for the higher-order rows
  compute_fscov <- function(lambda, sigma_inv, veta_g,
                            fsc = NULL, sigma = NULL) {
    if (bartlett) {
      # traditional formula uses solve(Lambda' Theta.inv Lambda); we use the
      # Sigma.inv version (minus Psi) to handle negative/zero variances, with
      # a pseudo-inverse fallback (lav_predict_solve)
      if (is.null(fsc)) {
        lav_predict_solve(t(lambda) %*% sigma_inv %*% lambda) - veta_g
      } else {
        cl_psi <- fsc %*% lambda %*% veta_g
        fsc %*% sigma %*% t(fsc) - cl_psi - t(cl_psi) + veta_g
      }
    } else {
      veta_g - veta_g %*% t(lambda) %*% sigma_inv %*% lambda %*% veta_g
    }
  }

  if (is.null(data_obs)) {
    data_obs <- lavdata@X
    newdata_flag <- FALSE
  } else {
    newdata_flag <- TRUE
  }
  # eXo not needed

  # missings? and missing = "ml"?
  if (lavdata@missing %in% c("ml", "ml.x")) {
    if (newdata_flag) {
      mp_1 <- vector("list", lavdata@ngroups)
      for (g in seq_len(lavdata@ngroups)) {
        mp_1[[g]] <- lav_data_mi_patterns(data_obs[[g]])
      }
    } else {
      mp_1 <- lavdata@Mp
    }
  }

  mm_lambda <- lav_model_lambda(lavmodel = lavmodel, remove_dummy_lv = FALSE,
                                use_wmat = TRUE)

  # conditional.x: the model-implied Sigma/means refer to the residual y|x
  # system, and the exogenous x-part is handled separately (per case). We
  # therefore use the *conditional* (given-x) model-implied quantities:
  #  - Sigma      = Cov(y|x) = res.cov (NOT the joint 'cov', which under
  #                 conditional.x does not exist in lavimplied)
  #  - V(ETA)     = Cov(eta|x) = (I-B)^-1 Psi (I-B)^-T (no Gamma cov.x Gamma'
  #                 term), keeping the dummy latent variables so that it stays
  #                 conformable with the full LAMBDA
  #  - E(ETA|x_i) and E(Y|x_i) are per-case quantities (computed in the loop)
  cond_x <- lavmodel@conditional.x
  if (cond_x) {
    # exo may be NULL when called via the 'lavobject' convenience path
    if (is.null(exo)) {
      exo <- lavdata@eXo
    }
    sigma_hat <- lavimplied$res.cov
    veta <- lav_model_vetax(lavmodel = lavmodel, remove_dummy_lv = FALSE)
    eetax <- lav_model_eetax(
      lavmodel = lavmodel, lavsamplestats = lavsamplestats,
      exo = exo, nobs = lapply(data_obs, NROW)
    )
    res_int <- lavimplied$res.int
    res_slopes <- lavimplied$res.slopes
  } else {
    sigma_hat <- lavimplied$cov
    veta <- lav_model_veta(lavmodel = lavmodel)
    eeta <- lav_model_eeta(lavmodel = lavmodel,
                           lavsamplestats = lavsamplestats)
    ey <- lav_model_ey(lavmodel = lavmodel, lavsamplestats = lavsamplestats)
  }
  sigma_inv <- lapply(sigma_hat, lav_predict_solve)

  # Bartlett + higher-order factors (new in 0.7-1): collapse the measurement
  # chain, just like in sam(): LAMBDA.star = LAMBDA %*% solve(I - B), where B
  # only contains the factor loadings of the higher-order factors (the rows
  # of the latent indicators); for each higher-order factor, the columns of
  # LAMBDA.star for its 'layer' (the factor itself + all factors that are
  # neither an ancestor nor a descendant in the measurement chain, keeping
  # only the topmost ones) define a collapsed measurement model
  # y = LAMBDA.star ETA2 + epsilon.star, and the (Sigma.inv form of the)
  # Bartlett mapping based on these columns provides the mapping-matrix row
  # for that factor; factors of the same layer share the same mapping
  mm_lambda_star <- vector("list", length = length(mm_lambda))
  ho_fixes <- vector("list", length = length(mm_lambda))
  if (bartlett && !is.null(lavpta) && !is.null(lavpta$vnames$lv.ind)) {
    lambda_mm_idx <- which(names(lavmodel@GLIST) == "lambda")
    mm_idx_block <- lav_model_group_mm_indices(lavmodel@nmat)
    for (b in seq_along(mm_lambda)) {
      lv_ind_names <- lavpta$vnames$lv.ind[[b]]
      if (length(lv_ind_names) == 0L) next
      empty_idx <- which(apply(mm_lambda[[b]], 2L,
                               function(x) all(x == 0)))
      if (length(empty_idx) == 0L) next
      mlist <- lavmodel@GLIST[mm_idx_block[[b]]]
      if (is.null(mlist$beta)) next
      lv_names <- lavmodel@dimNames[[lambda_mm_idx[b]]][[2L]]
      # keep only the rows of the latent indicators; other (structural)
      # regressions must not enter the collapsed loadings
      this_beta <- mlist$beta
      this_beta[is.na(this_beta)] <- 0
      keep_row_idx <- match(lv_ind_names, lv_names)
      bstar <- matrix(0, nrow = nrow(this_beta), ncol = ncol(this_beta))
      bstar[keep_row_idx, ] <- this_beta[keep_row_idx, , drop = FALSE]
      ib_inv <- solve(diag(nrow(bstar)) - bstar)
      lambda_star_b <- mm_lambda[[b]] %*% ib_inv
      # reach[r, j] = TRUE: factor j influences factor r through the
      # measurement chain (r is a latent indicator of j, possibly indirectly)
      reach <- abs(ib_inv) > 1e-12
      diag(reach) <- FALSE
      # the factors we can fix: empty column in LAMBDA, but non-empty
      # column in LAMBDA.star
      fix_idx <- empty_idx[apply(lambda_star_b[, empty_idx, drop = FALSE],
                                 2L, function(x) any(x != 0))]
      if (length(fix_idx) == 0L) next
      fixes <- list()
      for (j in fix_idx) {
        # this factor's layer: neither ancestors nor descendants of j, ...
        allowed <- which(!reach[j, ] & !reach[, j])
        # ... keeping only the topmost factors (no ancestor within the set)
        tcols <- allowed[!apply(reach[allowed, allowed, drop = FALSE],
                                1L, any)]
        key <- paste(tcols, collapse = "-")
        if (is.null(fixes[[key]])) {
          fixes[[key]] <- list(tcols = tcols, cols = j)
        } else {
          fixes[[key]]$cols <- c(fixes[[key]]$cols, j)
        }
      }
      mm_lambda_star[[b]] <- lambda_star_b
      ho_fixes[[b]] <- fixes
    }
  }

  fs <- vector("list", length = lavdata@ngroups)
  if (fsm) {
    fsm_1 <- vector("list", length = lavdata@ngroups)
  }
  if (rel) {
    rel_1 <- vector("list", length = lavdata@ngroups)
  }
  if (transform) {
    # correlation-preserving transformation: Green (regression) or
    # Krijnen/McDonald (Bartlett) determinacy matrix
    if (bartlett) {
      tmat <- lav_predict_tmat_det(lavmodel = lavmodel, lavimplied = lavimplied)
    } else {
      tmat <- lav_predict_tmat_green(lavmodel = lavmodel,
                                     lavimplied = lavimplied)
    }
  }

  if (acov != "none") {
    se <- acov # acov implies SE
  }
  if (se != "none") {
    se_1 <- vector("list", length = lavdata@ngroups)
    # return full sampling covariance matrix?
    if (acov != "none") {
      acov_1 <- vector("list", length = lavdata@ngroups)
    }
  }

  for (g in 1:lavdata@ngroups) {
    # determine block
    b <- lav_predict_block_idx(lavdata, g, level)

    veta_g <- veta[[b]]
    lambda_g <- mm_lambda[[b]]
    sigma_inv_g <- sigma_inv[[b]]
    if (!cond_x) {
      eeta_g <- eeta[[b]]
      ey_g <- ey[[b]]
    }
    # higher-order factors (Bartlett only; NULL otherwise)
    lambda_star_g <- mm_lambda_star[[b]]
    ho_fixes_g <- ho_fixes[[b]]

    if (lavdata@nlevels > 1L) {
      lp <- lavdata@Lp[[g]]

      # cluster means (Y2); recompute from data_obs (rather than the
      # sample statistics) so that the same code path handles newdata,
      # where the cluster structure comes from lavdata (= the newdata
      # object; see lav_predict_internal)
      y2_g <- rowsum.default(data_obs[[g]],
        group = lp$cluster.idx[[2]], reorder = FALSE, na.rm = FALSE
      ) / lp$cluster.size[[2]]

      # implied for this group
      group_idx <- (g - 1) * lavdata@nlevels + seq_len(lavdata@nlevels)
      implied_group <- lapply(lavimplied, function(x) x[group_idx])

      # random effects (=random intercepts or cluster means)
      if (lavdata@missing %in% c("ml", "ml.x")) {
        # missing data: use the posterior means from the E-step,
        # conditioning on ALL the observed data (new in 0.7-1); the
        # missing values are replaced by their posterior means, so
        # that the (linear) factor score computation below yields the
        # exact posterior means E(eta | all observed data)
        mb_j <- lav_mvn_cl_mi_estep_ranef(
          y1 = data_obs[[g]], y2 = y2_g,
          lp = lp, mp = lavdata@Mp[[g]],
          mu_w = implied_group$mean[[1]],
          sigma_w = implied_group$cov[[1]],
          mu_b = implied_group$mean[[2]],
          sigma_b = implied_group$cov[[2]],
          se = FALSE, impute = TRUE
        )
      } else {
        out <- lav_mvn_cl_implied22l(
          lp = lp,
          implied = implied_group
        )
        mb_j <- lav_mvn_cl_em_estep_ranef(
          ylp = list(list(), list(Y2 = y2_g)), lp = lp,
          sigma_w = out$sigma.w, sigma_b = out$sigma.b,
          sigma_zz = out$sigma.zz, sigma_yz = out$sigma.yz,
          mu_z = out$mu.z, mu_w = out$mu.w, mu_b = out$mu.b,
          se = FALSE
        )
      }

      ov_idx <- lp$ov.idx

      if (level == 1L) {
        data_w <- data_obs[[g]][, ov_idx[[1]]]
        # with missing = "ml": use the (posterior-mean) completed data
        if (!is.null(attr(mb_j, "y1w.imputed"))) {
          data_w <- attr(mb_j, "y1w.imputed")
        }
        data_b <- mb_j[lp$cluster.idx[[2]], , drop = FALSE]

        # center
        data_obs_g <- data_w - data_b
      } else if (level == 2L) {
        data_b_1 <- matrix(0,
          nrow = nrow(mb_j),
          ncol = ncol(data_obs[[g]])
        )
        data_b_1[, ov_idx[[1]]] <- mb_j
        between_idx <- lp$between.idx[[2 * g]]
        if (length(between_idx) > 0L) {
          data_b_1[, between_idx] <- data_obs[[g]][
            !duplicated(lp$cluster.idx[[2]]),
            between_idx
          ]
          # with missing = "ml": use the (posterior-mean) completed
          # between-only variables
          if (!is.null(attr(mb_j, "z.imputed"))) {
            data_b_1[, between_idx] <- attr(mb_j, "z.imputed")
          }
        }
        data_obs_g <- data_b_1[, ov_idx[[2]]]
      } else {
        lav_msg_stop(gettext("only 2 levels are supported"))
      }
    } else {
      data_obs_g <- data_obs[[b]]
    }

    nfac <- ncol(veta_g)
    if (nfac == 0L) {
      fs[[g]] <- matrix(0, lavdata@nobs[[g]], nfac)
      next
    }

    # center data and set up the per-case latent means (eeta_case_g); under
    # conditional.x both the observed centering means E(Y|x_i) and the latent
    # means E(ETA|x_i) vary per case (as a function of x_i)
    if (cond_x) {
      exo_g <- exo[[g]]
      # E(Y|x_i) = res.int + res.slopes %*% x_i
      eyx_g <- matrix(res_int[[b]], nrow = nrow(data_obs_g),
                      ncol = length(res_int[[b]]), byrow = TRUE)
      if (!is.null(exo_g) && ncol(exo_g) > 0L) {
        eyx_g <- eyx_g + exo_g %*% t(res_slopes[[b]])
      }
      yc <- data_obs_g - eyx_g
      eeta_case_g <- eetax[[b]] # [nobs x nfac]
    } else {
      yc <- t(t(data_obs_g) - ey_g)
      eeta_case_g <- matrix(eeta_g, nrow = nrow(data_obs_g),
                            ncol = length(eeta_g), byrow = TRUE)
    }

    # NOTE: do NOT scale yc by the sampling weights here. Factor scores are
    # per-case predictions; a case's score must not depend on its own
    # sampling weight. Weights already enter via the (weighted) parameter
    # estimates and the weighted mean used for centering above.

    # global factor score coefficient matrix 'C'
    fsc <- compute_fsc(lambda_g, sigma_inv_g, veta_g,
      lambda_star = lambda_star_g, fixes = ho_fixes_g
    )

    # transform?
    if (transform) {
      fsc <- tmat[[b]] %*% fsc
    }

    # store fsm?
    if (fsm) {
      fsm_g <- fsc
    }

    # reliability?
    if (rel) {
      var_f <- fsc %*% sigma_hat[[g]] %*% t(fsc)  # or S?
      cov_f_eta <- fsc %*% lambda_g %*% veta_g
      var_eta <- veta_g
      # FS.determinacy <- diag( diag(1/sqrt(diag(Var.f))) %*%
      #                         Cov.f.eta %*%
      #                         diag(1/sqrt(diag(Var.eta)))
      #                       )
      fs_determinacy <- (diag(cov_f_eta) /
                          (sqrt(diag(var_f)) * sqrt(diag(var_eta))))
      rel_g <- fs_determinacy * fs_determinacy
    }

    # compute factor scores
    # (for two-level models with missing = "ml", the data have been
    # completed by their posterior means above, so the complete-data
    # expressions apply)
    if (lavdata@missing %in% c("ml", "ml.x") && lavdata@nlevels == 1L) {
      # missing patterns for this group
      mp <- mp_1[[g]]

      # factor scores container
      fs_g <- matrix(as.numeric(NA), nrow(yc), ncol = nfac)

      if (se == "standard") {
        se_g <- matrix(as.numeric(NA), nrow(yc), ncol = nfac)
      }

      if (acov == "standard") {
        acov_g <- vector("list", length = mp$npatterns)
      }

      # compute FSC per pattern
      for (p in seq_len(mp$npatterns)) {
        var_idx <- mp$pat[p, ] # observed
        na_idx <- which(!var_idx) # missing

        # extract observed data for these (centered) cases
        oc <- yc[mp$case.idx[[p]], mp$pat[p, ], drop = FALSE]

        # invert sigma (Sigma_22, observed part only) for this pattern
        sigma_22_inv <- try(lav_mat_sym_inverse_update(
          s_inv = sigma_inv_g, rm_idx = na_idx, logdet = FALSE
        ), silent = TRUE)
        if (inherits(sigma_22_inv, "try-error")) {
          lav_msg_stop(gettext("Sigma_22.inv cannot be inverted"))
        }

        lambda <- lambda_g[var_idx, , drop = FALSE]
        lambda_star_p <- NULL
        if (!is.null(lambda_star_g)) {
          lambda_star_p <- lambda_star_g[var_idx, , drop = FALSE]
        }
        fsc <- compute_fsc(lambda, sigma_22_inv, veta_g,
          lambda_star = lambda_star_p, fixes = ho_fixes_g
        )

        # if FSC contains rows that are all-zero, replace by NA
        #
        # this happens eg if all the indicators of a single factor
        # are missing; then this column in lambda only contains zeroes
        # and therefore the corresponding row in FSC contains only
        # zeroes, leading to factor score 0
        #
        # showing 'NA' is better than getting 0
        #
        # (Note that this is not needed for the 'regression' method,
        #  only for Bartlett)
        if (bartlett) {
          fsc_nona <- fsc # keep a NA-free copy for the SEs below
          zero_idx <- which(apply(fsc, 1L, function(x) all(x == 0)))
          if (length(zero_idx) > 0L) {
            fsc[zero_idx, ] <- NA
          }
        }

        # factor scores for this pattern
        fs_g[mp$case.idx[[p]], ] <- t(fsc %*% t(oc)) +
          eeta_case_g[mp$case.idx[[p]], , drop = FALSE]

        # SE?
        if (se == "standard") {
          if (is.null(lambda_star_p)) {
            tmp <- compute_fscov(lambda, sigma_22_inv, veta_g)
          } else {
            tmp <- compute_fscov(lambda, sigma_22_inv, veta_g,
              fsc = fsc_nona, sigma = sigma_hat[[b]][var_idx, var_idx,
                                                     drop = FALSE]
            )
          }
          tmp_d <- diag(tmp)
          tmp_d[tmp_d < 1e-05] <- as.numeric(NA)

          # all cases in this pattern get the same SEs
          se_g[mp$case.idx[[p]], ] <- matrix(sqrt(tmp_d),
            nrow = length(mp$case.idx[[p]]),
            ncol = ncol(se_g), byrow = TRUE
          )
        }

        # acov?
        if (acov == "standard") {
          acov_g[[p]] <- tmp # for this pattern
        }
      } # p
    } else {
      # compute factor scores
      fs_g <- t(fsc %*% t(yc)) + eeta_case_g
    }

    # replace values in dummy lv's by their observed counterpart
    if (length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L && level == 1L) {
      fs_g[, lavmodel@ov.y.dummy.lv.idx[[g]]] <-
        data_obs_g[, lavmodel@ov.y.dummy.ov.idx[[g]], drop = FALSE]
    }
    if (length(lavmodel@ov.x.dummy.lv.idx[[g]]) > 0L && level == 1L) {
      fs_g[, lavmodel@ov.x.dummy.lv.idx[[g]]] <-
        data_obs_g[, lavmodel@ov.x.dummy.ov.idx[[g]], drop = FALSE]
    }

    fs[[g]] <- fs_g

    # FSM
    if (fsm) {
      fsm_1[[g]] <- fsm_g
    }

    # REL
    if (rel) {
      rel_1[[g]] <- rel_g
    }

    # standard error
    if (se == "standard" && !transform) { # for now
      if (lavdata@missing %in% c("ml", "ml.x") && lavdata@nlevels == 1L) {
        se_1[[g]] <- se_g
        if (acov == "standard") {
          acov_1[[g]] <- acov_g
        }
      } else { # complete data
        if (is.null(lambda_star_g)) {
          tmp <- compute_fscov(lambda_g, sigma_inv_g, veta_g)
        } else {
          tmp <- compute_fscov(lambda_g, sigma_inv_g, veta_g,
            fsc = fsc, sigma = sigma_hat[[b]]
          )
        }
        tmp_d <- diag(tmp)
        tmp_d[tmp_d < 1e-05] <- as.numeric(NA)
        se_1[[g]] <- matrix(sqrt(tmp_d), nrow = 1L)

        # return full sampling covariance matrix?
        if (acov == "standard") {
          acov_1[[g]] <- tmp
        }
      }
    } # se = "standard"
  } # g

  if (fsm) {
    attr(fs, "fsm") <- fsm_1
  }
  if (rel) {
    attr(fs, "rel") <- rel_1
  }
  if (se != "none") {
    attr(fs, "se") <- se_1
    # return full sampling covariance matrix?
    if (acov == "standard") {
      attr(fs, "acov") <- acov_1
    }
  }

  fs
}

# decide whether (and how) to parallelise the per-case factor-score
# optimisation for a given number of independent tasks 'n_tasks'.
#  - "no"        : always serial
#  - "multicore" : fork-based (unix only), as requested by the user
#  - "snow"      : PSOCK cluster, as requested by the user
#  - "auto"      : fork-based, but only when it is clearly worthwhile, i.e.
#                  enough independent optimisations to amortise the overhead
#                  (the dedup step often leaves only a handful of unique cases)
# returns one of "no", "multicore", "snow".
lav_predict_parallel_mode <- function(parallel = "no", ncpus = NULL,
                                      n_tasks = 1L) {
  # number of unique optimisations above which "auto" switches on forking
  auto_threshold <- 1000L

  if (is.null(ncpus) || ncpus < 2L || n_tasks < 2L || parallel == "no") {
    return("no")
  }
  unix <- .Platform$OS.type != "windows"
  if (parallel == "multicore") {
    if (unix) "multicore" else "no"
  } else if (parallel == "snow") {
    "snow"
  } else if (parallel == "auto") {
    # auto only uses fork, and only when there is enough work to gain from it
    if (unix && n_tasks >= auto_threshold) "multicore" else "no"
  } else {
    "no"
  }
}

# factor scores - EBM or ML
lav_predict_eta_ebm_ml <- function(lavobject = NULL, # for convenience
                                   # sub objects
                                   lavmodel = NULL, lavdata = NULL,
                                   lavsamplestats = NULL,
                                   # optional new data
                                   data_obs = NULL, exo = NULL,
                                   se = "none", acov = "none", level = 1L,
                                   ml = FALSE,
                                   optim_method = "bfgs",
                                   parallel = "no", ncpus = NULL, cl = NULL) {
  optim_method <- tolower(optim_method)

  stopifnot(optim_method %in% c("nlminb", "bfgs"))

  # parallelisation: the per-case optimisation below is embarrassingly parallel
  # and fully deterministic (no RNG handling needed). Resolve the number of
  # cores now; the actual serial/parallel decision is made per group, once the
  # number of unique cases to optimise is known (see lav_predict_parallel_mode).
  parallel <- parallel[1L]
  if (parallel != "no" && is.null(ncpus)) {
    # CRAN policy: never use more than 2 cores by default; users may request
    # more explicitly via ncpus=.
    ncpus <- min(2L, max(1L, parallel::detectCores() - 2L, na.rm = TRUE))
  }
  if (parallel != "no") {
    loadNamespace("parallel")
  }

  ### FIXME: if all indicators of a factor are normal, can we not
  ###        just use the `classic' regression method??
  ###        (perhaps after whitening, to get uncorrelated factors...)

  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
  } else {
    stopifnot(
      !is.null(lavmodel), !is.null(lavdata),
      !is.null(lavsamplestats)
    )
  }

  # new data?
  if (is.null(data_obs)) {
    data_obs <- lavdata@X
  }
  if (is.null(exo)) {
    exo <- lavdata@eXo
  }

  # se?
  if (acov != "none") {
    se <- acov # acov implies SE
  }
  # For categorical data the factor scores are posterior-mode (EBM) or ML
  # estimates obtained by numerical optimisation. Their standard errors are
  # based on the curvature of the objective at the optimum: the (co)variance
  # matrix is the inverse of the Hessian of the negative log-posterior (EBM) or
  # negative log-likelihood (ML), i.e. a Laplace/observed-information
  # approximation. Unlike the linear-Gaussian case, this curvature depends on
  # the response pattern, so the standard errors differ from one (pattern of)
  # observation(s) to the next.
  se_flag <- (se != "none")
  acov_flag <- (acov != "none")
  if (se_flag) {
    se_1 <- vector("list", length = lavdata@ngroups)
    if (acov_flag) {
      acov_1 <- vector("list", length = lavdata@ngroups)
    }
  }

  vetax <- lav_model_vetax(lavmodel = lavmodel)
  vetax_inv <- vetax
  for (g in seq_len(lavdata@ngroups)) {
    if (nrow(vetax[[g]]) > 0L) {
      vetax_inv[[g]] <- solve(vetax[[g]])
    }
  }
  eetax <- lav_model_eetax(
    lavmodel = lavmodel, lavsamplestats = lavsamplestats,
    exo = exo, nobs = lapply(data_obs, NROW),
    remove_dummy_lv = TRUE
  ) ## FIXME?
  th_1 <- lav_model_th(lavmodel = lavmodel, delta = FALSE)
  mm_theta <- lav_model_theta(lavmodel = lavmodel)

  # check for zero entries in THETA (new in 0.6-4)
  for (g in seq_len(lavdata@ngroups)) {
    if (any(diag(mm_theta[[g]]) == 0)) {
      lav_msg_stop(gettext(
        "(residual) variance matrix THETA contains zero elements
        on the diagonal."))
    }
  }

  # Model matrices per group, used by the conditional-density evaluation. They
  # are loop-invariant within a group, so we extract them once here instead of
  # rebuilding them on every objective evaluation (the objective is called many
  # times per case by the optimiser).
  fy_mlist_g <- lapply(seq_len(lavdata@ngroups), function(gg) {
    mm_idx <- 1:lavmodel@nmat[gg] + cumsum(c(0, lavmodel@nmat))[gg]
    lavmodel@GLIST[mm_idx]
  })

  # local objective function: x = lv values. The per-case dummy ov.y values are
  # passed in (rather than read via a case index 'i') so the objective does not
  # depend on the enclosing loop variable -- this lets us evaluate cases via
  # lapply()/mclapply() instead of a for() loop.
  f_eta_i <- function(x, y_i, x_i, mu_i, dummy_y) {
    # add 'dummy' values (if any) for ov.y
    if (length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L) {
      x2 <- c(x - mu_i, dummy_y)
    } else {
      x2 <- x - mu_i
    }

    # conditional density of y, given eta.i(=x)
    log_fy <- lav_predict_fy_eta_i(
      lavmodel = lavmodel,
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      y_i = y_i,
      x_i = x_i,
      eta_i = matrix(x2, nrow = 1L), # <---- eta!
      theta_sd = theta_sd,
      th = th,
      th_idx = th_idx,
      g = g,
      log = TRUE,
      mlist = fy_mlist_g[[g]]
    )

    if (ml) {
      # NOTE: 'true' ML is simply  -1*sum(log.fy)
      #  - but there is no upper/lower bound for the extrema:
      #    a pattern of all (in)correct drives the 'theta' parameter
      #    towards +/- Inf
      # - therefore, we add a vague prior, just to stabilize
      #
      diff <- t(x) - mu_i
      v <- diag(length(x)) * 1e-05
      tmp <- as.numeric(0.5 * diff %*% v %*% t(diff))
      out <- 1 + tmp - sum(log_fy, na.rm = TRUE)
    } else {
      diff <- t(x) - mu_i
      v <- vetax_inv[[g]]
      tmp <- as.numeric(0.5 * diff %*% v %*% t(diff))
      out <- tmp - sum(log_fy, na.rm = TRUE)
    }
    out
  }

  fs <- vector("list", length = lavdata@ngroups)
  for (g in seq_len(lavdata@ngroups)) {
    nfac <- ncol(vetax[[g]])
    nfac2 <- nfac
    if (length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L) {
      nfac2 <- nfac2 + length(lavmodel@ov.y.dummy.lv.idx[[g]])
    }
    fs[[g]] <- matrix(as.numeric(NA), nrow(data_obs[[g]]), nfac2)

    # special case: no regular lv's
    if (nfac == 0) {
      # impute dummy ov.y (if any)
      fs[[g]][, lavmodel@ov.y.dummy.ov.idx[[g]]] <-
        data_obs[[g]][, lavmodel@ov.y.dummy.ov.idx[[g]], drop = FALSE]
      next
    }

    ## FIXME: factor scores not identical (but close) to Mplus
    #         if delta elements not equal to 1??
    # mm_in_group <- 1:lavmodel@nmat[g] + cumsum(c(0, lavmodel@nmat))[g]
    # mlist <- lavmodel@GLIST[mm_in_group]

    # check for negative values
    neg_var_idx <- which(diag(mm_theta[[g]]) < 0)
    if (length(neg_var_idx) > 0) {
      lav_msg_warn(
        gettext("factor scores could not be computed due to at least
                one negative (residual) variance"))
      next
    }

    # common values
    theta_sd <- sqrt(diag(mm_theta[[g]]))
    th <- th_1[[g]]
    th_idx <- lavmodel@th.idx[[g]]

    # Casewise optimisation. Two cases that share the same inputs -- the data
    # row y_i, the exogenous covariates x_i, and the conditional mean mu_i --
    # define exactly the same objective function and starting value; since the
    # optimiser is deterministic, they yield identical factor scores. We
    # therefore optimise once per unique combination and copy the result to all
    # matching cases. For categorical data many cases share the same response
    # pattern, so this can reduce the number of optimisations by an order of
    # magnitude.
    n <- nrow(data_obs[[g]])

    # key each case by everything the objective depends on, using the exact
    # (round-tripping) hexadecimal representation so that only bitwise-identical
    # rows are ever merged
    key_cols <- list(data_obs[[g]], eetax[[g]])
    if (!is.null(exo[[g]])) {
      key_cols <- c(key_cols, list(exo[[g]]))
    }
    key_mat <- do.call(cbind, key_cols)
    case_key <- do.call(paste, c(
      lapply(seq_len(ncol(key_mat)), function(j) sprintf("%a", key_mat[, j])),
      sep = "\r"
    ))
    rep_idx <- which(!duplicated(case_key)) # one representative case / pattern
    case_map <- rep_idx[match(case_key, case_key[rep_idx])] # case -> rep row

    # optimise the latent scores for a single (representative) case 'i'.
    # Returns a list with the score 'eta' (incl. any dummy ov.y values) and,
    # when requested, the standard errors 'se' and sampling covariance 'acov'
    # of the (regular) latent scores.
    solve_i <- function(i) {
      # eXo?
      if (!is.null(exo[[g]])) {
        x_i <- exo[[g]][i, , drop = FALSE]
      } else {
        x_i <- NULL
      }
      mu_i <- eetax[[g]][i, , drop = FALSE]
      y_i <- data_obs[[g]][i, , drop = FALSE]

      # dummy ov.y values for this case (passed to the objective and appended
      # to the score below)
      if (length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L) {
        dummy_y <- data_obs[[g]][i, lavmodel@ov.y.dummy.ov.idx[[g]],
          drop = FALSE
        ]
      } else {
        dummy_y <- NULL
      }

      start_1 <- numeric(nfac) # initial values for eta
      se_i <- rep(as.numeric(NA), nfac)
      acov_i <- matrix(as.numeric(NA), nfac, nfac)

      if (!all(is.na(y_i))) {
        # find best values for eta.i
        if (optim_method == "nlminb") {
          out <- nlminb(
            start = start_1, objective = f_eta_i,
            gradient = NULL, # for now
            control = list(rel.tol = 1e-8),
            y_i = y_i, x_i = x_i, mu_i = mu_i, dummy_y = dummy_y
          )
        } else if (optim_method == "bfgs") {
          out <- optim(
            par = start_1, fn = f_eta_i,
            gr = NULL,
            control = list(reltol = 1e-8, fnscale = 1.1),
            method = "BFGS",
            y_i = y_i, x_i = x_i, mu_i = mu_i, dummy_y = dummy_y
          )
        }
        if (out$convergence == 0L) {
          eta_reg <- out$par

          # standard errors from the curvature of the objective at the
          # optimum: f_eta_i is the negative log-posterior (EBM) or negative
          # log-likelihood (ML), so its Hessian at the mode is the inverse of
          # the (Laplace) sampling covariance of the latent scores.
          if (se_flag) {
            hess <- try(stats::optimHess(eta_reg, f_eta_i,
              y_i = y_i, x_i = x_i, mu_i = mu_i, dummy_y = dummy_y
            ), silent = TRUE)
            if (!inherits(hess, "try-error")) {
              tmp_acov <- try(solve(hess), silent = TRUE)
              if (!inherits(tmp_acov, "try-error")) {
                acov_i <- tmp_acov
                tmp_d <- diag(tmp_acov)
                tmp_d[tmp_d < 0] <- as.numeric(NA) # catch non-pd Hessians
                se_i <- sqrt(tmp_d)
              }
            }
          }
        } else {
          eta_reg <- rep(as.numeric(NA), nfac)
        }
      } else {
        eta_reg <- rep(as.numeric(NA), nfac)
      }

      # add dummy ov.y lv values
      eta_i <- eta_reg
      if (length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L) {
        eta_i <- c(eta_i, dummy_y)
      }

      list(eta = eta_i, se = se_i, acov = acov_i)
    }

    # optimise once per unique case, serially or in parallel
    run_mode <- lav_predict_parallel_mode(parallel, ncpus, length(rep_idx))
    if (run_mode == "multicore") {
      res_list <- parallel::mclapply(rep_idx, solve_i,
        mc.cores = min(ncpus, length(rep_idx))
      )
    } else if (run_mode == "snow") {
      if (is.null(cl)) {
        cl0 <- parallel::makePSOCKcluster(
          rep("localhost", min(ncpus, length(rep_idx)))
        )
        on.exit(parallel::stopCluster(cl0), add = TRUE)
        res_list <- parallel::parLapply(cl0, rep_idx, solve_i)
      } else {
        res_list <- parallel::parLapply(cl, rep_idx, solve_i)
      }
    } else {
      res_list <- lapply(rep_idx, solve_i)
    }
    for (u in seq_along(rep_idx)) {
      fs[[g]][rep_idx[u], ] <- res_list[[u]]$eta
    }

    # copy each representative's factor scores to all cases in its pattern
    fs[[g]] <- fs[[g]][case_map, , drop = FALSE]

    # expand the (per-pattern) standard errors / sampling covariances to all
    # cases. The se/acov refer to the nfac regular factors; they are placed in
    # the leading columns of an nfac2-wide structure (dummy ov.y columns left
    # as NA) so that the dummy-lv handling in lav_predict_internal() applies
    # identically to the scores and to their standard errors.
    if (se_flag) {
      u_of_case <- match(case_key, case_key[rep_idx]) # case -> unique index
      se_rep <- do.call(rbind, lapply(res_list, "[[", "se")) # n_unique x nfac
      se_g <- matrix(as.numeric(NA), n, nfac2)
      se_g[, seq_len(nfac)] <- se_rep[u_of_case, , drop = FALSE]
      se_1[[g]] <- se_g
      if (acov_flag) {
        acov_rep <- lapply(res_list, "[[", "acov") # list: nfac x nfac
        acov_1[[g]] <- lapply(u_of_case, function(u) {
          m <- matrix(as.numeric(NA), nfac2, nfac2)
          m[seq_len(nfac), seq_len(nfac)] <- acov_rep[[u]]
          m
        })
      }
    }
  }

  if (se_flag) {
    attr(fs, "se") <- se_1
    if (acov_flag) {
      attr(fs, "acov") <- acov_1
    }
  }

  fs
}

# predicted value for response y*_i, conditional on the predicted latent
# variable scores
# `measurement part':
#     y*_i = nu + lambda eta_i + K x_i + epsilon_i
#
#    where eta_i = latent variable value for i (either given or from predict)
#
# Two types: 1) nrow(eta) = nrow(X) (factor scores)
#            2) nrow(eta) = 1L (given values)
#
# in both cases, we return [nobs x nvar] matrix per group
lav_predict_yhat <- function(lavobject = NULL, # for convience
                             # sub objects
                             lavmodel = NULL, lavdata = NULL,
                             lavsamplestats = NULL,
                             lavimplied = NULL,
                             # new data
                             data_obs = NULL, exo = NULL,
                             # eta values
                             eta = NULL,
                             # options
                             method = "EBM",
                             duplicate = FALSE,
                             optim_method = "bfgs",
                             fsm = FALSE,
                             resid_flag = FALSE) {
  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
    lavimplied <- lavobject@implied
  } else {
    stopifnot(
      !is.null(lavmodel), !is.null(lavdata),
      !is.null(lavsamplestats), !is.null(lavimplied)
    )
  }

  # new data?
  if (is.null(data_obs)) {
    data_obs <- lavdata@X
  }
  if (is.null(exo)) {
    exo <- lavdata@eXo
  }

  # do we get values for eta? If not, use `predict' to get plausible values
  if (is.null(eta)) {
    eta <- lav_predict_eta(
      lavobject = NULL, lavmodel = lavmodel,
      lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavimplied = lavimplied,
      data_obs = data_obs, exo = exo, method = method,
      optim_method = optim_method, fsm = fsm
    )
    fsm_1 <- attr(eta, "fsm")
  } else {
    # matrix
    if (is.matrix(eta)) { # user-specified?
      if (nrow(eta) == 1L) {
        tmp <- matrix(eta, lavsamplestats@ntotal, length(eta),
          byrow = TRUE
        )
      } else if (nrow(eta) != lavsamplestats@ntotal) {
        lav_msg_stop(gettext("nrow(eta) != lavsamplestats@ntotal"))
      } else {
        tmp <- eta
      }
      eta <- lapply(1:lavdata@ngroups, function(i) tmp[lavdata@case.idx[[i]], ])
      # vector: just 1 row of factor-scores
    } else if (is.numeric(eta)) {
      # convert to matrix
      tmp <- matrix(eta, lavsamplestats@ntotal, length(eta), byrow = TRUE)
      eta <- lapply(1:lavdata@ngroups, function(i) tmp[lavdata@case.idx[[i]], ])
    } else if (is.list(eta)) {
      stopifnot(lavdata@ngroups == length(eta))
    }
  }

  yhat <- lav_model_yhat(
    lavmodel = lavmodel, glist = NULL,
    lavsamplestats = lavsamplestats, exo = exo,
    nobs = lapply(data_obs, NROW),
    eta = eta, duplicate = duplicate
  )

  # if conditional.x, paste exo
  if (lavmodel@categorical && !is.null(exo)) {
    yhat <- lapply(seq_len(lavdata@ngroups), function(g) {
      ret <- cbind(yhat[[g]], exo[[g]])
      ret
    })
  }

  # residuals? compute y - yhat
  if (resid_flag) {
    res <- lapply(seq_len(lavdata@ngroups), function(g) {
      ret <- data_obs[[g]] - yhat[[g]]
      ret
    })
  } else {
    res <- yhat
  }

  # fsm?
  if (fsm) {
    attr(res, "fsm") <- fsm_1
  }

  res
}

# conditional density y -- assuming independence!!
# f(y_i | eta_i, x_i) for EACH item
#
lav_predict_fy <- function(lavobject = NULL, # for convience
                           # sub objects
                           lavmodel = NULL, lavdata = NULL,
                           lavsamplestats = NULL,
                           lavimplied = NULL,
                           # new data
                           data_obs = NULL, exo = NULL,
                           # eta values
                           eta = NULL,
                           # options
                           method = "EBM",
                           log_1 = FALSE,
                           optim_method = "bfgs") {
  # full object?
  if (inherits(lavobject, "lavaan")) {
    lavmodel <- lavobject@Model
    lavdata <- lavobject@Data
    lavsamplestats <- lavobject@SampleStats
    lavimplied <- lavobject@implied
  } else {
    stopifnot(
      !is.null(lavmodel), !is.null(lavdata),
      !is.null(lavsamplestats), !is.null(lavimplied)
    )
  }

  # new data?
  if (is.null(data_obs)) {
    data_obs <- lavdata@X
  }
  if (is.null(exo)) {
    exo <- lavdata@eXo
  }

  # we need the YHATs (per group)
  yhat <- lav_predict_yhat(
    lavobject = NULL, lavmodel = lavmodel,
    lavdata = lavdata, lavsamplestats = lavsamplestats,
    lavimplied = lavimplied,
    data_obs = data_obs, exo = exo, eta = eta, method = method,
    duplicate = FALSE, optim_method = optim_method
  )

  mm_theta <- lav_model_theta(lavmodel = lavmodel)
  th <- lav_model_th(lavmodel = lavmodel, delta = FALSE)

  fy <- vector("list", length = lavdata@ngroups)
  for (g in seq_len(lavdata@ngroups)) {
    fy[[g]] <- lav_predict_fy_internal(
      x = data_obs[[g]], yhat = yhat[[g]],
      th = th[[g]], mm_theta = mm_theta[[g]],
      num_idx = lavmodel@num.idx[[g]],
      th_idx = lavmodel@th.idx[[g]],
      link = lavmodel@link, log_1 = log_1
    )
  }

  fy
}


# single group, internal function
lav_predict_fy_internal <- function(x = NULL, yhat = NULL,
                                    th = NULL, mm_theta = NULL,
                                    num_idx = NULL, th_idx = NULL,
                                    link = NULL, log_1 = FALSE) {
  # shortcuts
  theta_var <- diag(mm_theta)

  # check size yhat (either 1L or Nobs rows)
  if (!(nrow(yhat) == 1L || nrow(yhat) == nrow(x))) {
    lav_msg_stop(gettext("nrow(yhat[[g]]) not 1L and not nrow(X))"))
  }

  fy_group <- matrix(0, nrow(x), ncol(x))
  # if(NORMAL) {
  #    if(nrow(yhat) == nrow(X)) {
  #        tmp <- (X - yhat)^2
  #    } else {
  #        tmp <- sweep(X, MARGIN=2, STATS=yhat, FUN="-")^2
  #    }
  #    tmp1 <- sweep(tmp, MARGIN=2, theta.var, "/")
  #    tmp2 <- exp( -0.5 * tmp1 )
  #    tmp3 <- sweep(tmp2, MARGIN=2, sqrt(2*pi*theta.var), "/")
  #    if(log.) {
  #        FY.group <- log(tmp3)
  #    } else {
  #        FY.group <- tmp3
  #    }
  # } else {
  # mixed items

  ord_idx <- unique(th_idx[th_idx > 0L])

  # first, NUMERIC variables
  if (length(num_idx) > 0L) {
    for (v in num_idx) {
      fy_group[, v] <- dnorm(x[, v],
        # yhat may change or not per case
        mean = yhat[, v],
        sd   = sqrt(theta_var[v]),
        log  = log_1
      )
    }
  }

  # second, ORDERED variables
  for (v in ord_idx) {
    th_y <- th[th_idx == v]
    th_y_1 <- c(-Inf, th_y, Inf)
    ncat <- length(th_y) + 1L
    fy <- numeric(ncat)
    theta_v <- sqrt(theta_var[v])
    yhat_v <- yhat[, v]

    # two cases: yhat.v is a scalar, or has length = nobs
    fy <- matrix(0, nrow = length(yhat_v), ncol = ncat)

    # for each category
    for (k in seq_len(ncat)) {
      if (link == "probit") {
        fy[, k] <- pnorm((th_y_1[k + 1] - yhat_v) / theta_v) -
          pnorm((th_y_1[k] - yhat_v) / theta_v)
      } else if (link == "logit") {
        fy[, k] <- plogis((th_y_1[k + 1] - yhat_v) / theta_v) -
          plogis((th_y_1[k] - yhat_v) / theta_v)
      } else {
        lav_msg_stop(gettext("link must be probit or logit"))
      }
    }

    # underflow
    idx <- which(fy < .Machine$double.eps)
    if (length(idx) > 0L) {
      fy[idx] <- .Machine$double.eps
    }

    # log?
    if (log_1) {
      fy <- log(fy)
    }

    # case-wise expansion/selection
    if (length(yhat_v) == 1L) {
      # expand category probabilities for all observations
      fy_group[, v] <- fy[1L, x[, v]]
    } else {
      # select correct category probability per observation
      fy_group[, v] <- fy[cbind(seq_len(nrow(fy)), x[, v])]
    }
  } # ord

  fy_group
}



# conditional density y -- assuming independence!!
# f(y_i | eta_i, x_i)
#
# but for a SINGLE observation y_i (and x_i), for given values of eta_i
#
lav_predict_fy_eta_i <- function(lavmodel = NULL, lavdata = NULL,
                                 lavsamplestats = NULL,
                                 y_i = NULL, x_i = NULL,
                                 eta_i = NULL, theta_sd = NULL, g = 1L,
                                 th = NULL, th_idx = NULL, log = TRUE,
                                 mlist = NULL) {
  # model matrices for this group; loop-invariant, so callers in a tight loop
  # may pass a precomputed 'mlist' to avoid rebuilding it on every call
  if (is.null(mlist)) {
    mm_in_group <- 1:lavmodel@nmat[g] + cumsum(c(0, lavmodel@nmat))[g]
    mlist <- lavmodel@GLIST[mm_in_group]
  }

  # linear predictor for all items
  yhat <-
    lav_lisrel_eyetax(
      mlist = mlist,
      exo = x_i,
      eta = eta_i,
      sample_mean = lavsamplestats@mean[[g]],
      ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
      ov_x_dummy_ov_idx = lavmodel@ov.x.dummy.ov.idx[[g]],
      ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]],
      ov_x_dummy_lv_idx = lavmodel@ov.x.dummy.lv.idx[[g]],
      delta = FALSE
    )

  # P(y_i | eta_i, x_i) for all items
  if (all(lavdata@ov$type == "numeric")) {
    # NORMAL case
    fy <- dnorm(y_i, mean = yhat, sd = theta_sd, log = log)
  } else {
    nvar_g <- lavmodel@nvar[g]
    var_type <- lavdata@ov$type[seq_len(nvar_g)]
    fy <- numeric(nvar_g)

    # numeric variables: vectorised normal density
    num_idx <- which(var_type == "numeric")
    if (length(num_idx) > 0L) {
      fy[num_idx] <- dnorm(y_i[num_idx],
        mean = yhat[num_idx], sd = theta_sd[num_idx], log = log
      )
    }

    # ordered variables: P(y = k) = Phi((tau_k - yhat)/sd) -
    # Phi((tau_{k-1} - yhat)/sd), with tau_0 = -Inf and tau_{ncat} = +Inf.
    # Computed for all ordered items at once (two vectorised pnorm calls
    # instead of a per-item loop with two scalar pnorm calls each).
    ord_idx <- which(var_type == "ordered")
    if (length(ord_idx) > 0L) {
      # missing responses -> NA density
      na_mask <- is.na(y_i[ord_idx])
      if (any(na_mask)) {
        fy[ord_idx[na_mask]] <- as.numeric(NA)
      }
      obs <- ord_idx[!na_mask]
      if (length(obs) > 0L) {
        k <- y_i[obs] # observed category (1-based)
        # number of categories per variable, and the 0-based offset of each
        # variable's thresholds within 'th' (match() locates the first
        # threshold, which is robust to placeholder entries for the
        # numeric variables, i.e. th.idx == 0)
        ncat <- tabulate(th_idx, nbins = nvar_g)[obs] + 1L
        th_off <- match(obs, th_idx) - 1L

        # upper threshold tau_k (Inf for the top category)
        upper <- th[th_off + k]
        upper[k == ncat] <- Inf
        # lower threshold tau_{k-1} (-Inf for the first category)
        lower <- rep(-Inf, length(obs))
        nz <- k > 1L
        lower[nz] <- th[th_off[nz] + k[nz] - 1L]

        sd_obs <- theta_sd[obs]
        yhat_obs <- yhat[obs]
        prob <- pnorm((upper - yhat_obs) / sd_obs) -
          pnorm((lower - yhat_obs) / sd_obs)
        prob[prob < .Machine$double.eps] <- .Machine$double.eps
        fy[obs] <- if (log) log(prob) else prob
      }
    }

    # guard against unexpected variable types
    bad_idx <- which(!var_type %in% c("numeric", "ordered"))
    if (length(bad_idx) > 0L) {
      v <- bad_idx[1L]
      lav_msg_stop(gettextf("unknown type: `%1$s' for variable: %2$s",
        lavdata@ov$type[v], lavdata@ov$name[v])
      )
    }
  }

  fy
}

# gather the (unconditional) per-block ingredients shared by both
# correlation-preserving transformations: model-implied Sigma, latent
# covariance Psi (veta) and factor loadings Lambda
lav_predict_tmat_blocks <- function(lavobject = NULL,
                                    lavmodel = NULL, lavimplied = NULL) {
  if (!is.null(lavobject)) {
    lavmodel <- lavobject@Model
    lavimplied <- lavobject@implied
  }

  if (is.null(lavimplied) || length(lavimplied) == 0L) {
    lavimplied <- lav_model_implied(lavmodel)
  }
  if (lavmodel@conditional.x) {
    lavimplied <- lav_model_implied_cond2uncond(lavimplied)
  }

  list(
    sigma = lavimplied$cov,
    veta = lav_model_veta(lavmodel = lavmodel, remove_dummy_lv = FALSE),
    lambda = lav_model_lambda(lavmodel, remove_dummy_lv = FALSE,
                              use_wmat = TRUE),
    nblocks = lavmodel@nblocks
  )
}

# compute `transformation' matrix to convert regression factor scores
# to (Green's) correlation-preserving factor scores
lav_predict_tmat_green <- function(lavobject = NULL,
                                   lavmodel = NULL, lavimplied = NULL) {
  mm <- lav_predict_tmat_blocks(lavobject, lavmodel, lavimplied)

  lapply(seq_len(mm$nblocks), function(b) {
    lav_predict_tmat_green_internal(mm$sigma[[b]], mm$veta[[b]], mm$lambda[[b]])
  })
}

# compute `transformation' matrix to convert Bartlett factor scores
# to (Krijnen/McDonald) correlation-preserving factor scores
lav_predict_tmat_det <- function(lavobject = NULL,
                                 lavmodel = NULL, lavimplied = NULL) {
  mm <- lav_predict_tmat_blocks(lavobject, lavmodel, lavimplied)

  lapply(seq_len(mm$nblocks), function(b) {
    lav_predict_tmat_det_internal(mm$sigma[[b]], mm$veta[[b]], mm$lambda[[b]])
  })
}

# single block only (Green/regression)
lav_predict_tmat_green_internal <- function(sigma_1 = NULL, veta = NULL,
                                            lambda = NULL) {
  sigma_inv <- solve(sigma_1)
  veta_sqrt <- lav_mat_sym_sqrt(veta)
  veta32 <- veta %*% veta_sqrt
  tmp <- veta32 %*% t(lambda) %*% sigma_inv %*% lambda %*% veta32
  tmp_inv_sqrt <- lav_mat_sym_sqrt(solve(tmp))
  veta_sqrt %*% tmp_inv_sqrt %*% veta_sqrt
}

# single block only (Bartlett/determinacy), for internal use in
# lav_sam_step1_local()
lav_predict_tmat_det_internal <- function(sigma_1 = NULL, veta = NULL,
                                          lambda = NULL) {
    sigma_inv <- solve(sigma_1)
    veta_sqrt <- lav_mat_sym_sqrt(veta)
    veta_inv_sqrt <- lav_mat_sym_sqrt(solve(veta))
    tmp <- veta_sqrt %*% t(lambda) %*% sigma_inv %*% lambda %*% veta_sqrt
    tmp_sqrt <- lav_mat_sym_sqrt(tmp)
    tmat <- veta_sqrt %*% tmp_sqrt %*% veta_inv_sqrt
    tmat
}

# inspect a lavaanList object

lav_lavaanlist_inspect <- function(object, what = "free", ...) {
  dotdotdot <- list(...)
  if (length(dotdotdot) > 0L) {
    for (j in seq_along(dotdotdot)) {
      lav_msg_warn(gettextf(
        "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
        sQuote("inspect"))
      )
    }
  }
  lavListInspect(
    object = object,
    what = what,
    add_labels = TRUE,
    add_class = TRUE,
    drop_list_single_group = TRUE
  )
}

# the `tech' version: no labels, full matrices, ... for further processing
lav_lavaanlist_lavtech <- function(object,
                               what = "free",
                               add_labels = FALSE,
                               add_class = FALSE,
                               list_by_group = FALSE,
                               drop_list_single_group = FALSE,
                              ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)

  lavListInspect(
    object = object, what = what,
    add_labels = add_labels, add_class = add_class,
    list_by_group = list_by_group,
    drop_list_single_group = drop_list_single_group
  )
}

lavListTech <- lav_alias("lav_lavaanlist_lavtech") # synonym # nolint

# just in case someone uses lavInspect on a lavaanList object
lav_lavaanlist_lavinspect <- function(object,
                                  what = "free",
                                  add_labels = TRUE,
                                  add_class = TRUE,
                                  list_by_group = TRUE,
                                  drop_list_single_group = TRUE,
                                  ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)

  lavListInspect(
    object = object, what = what,
    add_labels = add_labels, add_class = add_class,
    list_by_group = list_by_group,
    drop_list_single_group = drop_list_single_group
  )
}

lavListInspect <- function(object,
                           what = "free",
                           add_labels = TRUE,
                           add_class = TRUE,
                           list_by_group = TRUE,
                           drop_list_single_group = TRUE,
                           ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # object must inherit from class lavaanList
  stopifnot(inherits(object, "lavaanList"))

  # only a single argument
  if (length(what) > 1) {
    lav_msg_stop(gettext(
      "`what' argument contains multiple arguments; only one is allowed"))
  }

  # be case insensitive
  what <- tolower(what)


  #### model matrices, with different contents ####
  if (what == "free") {
    lav_lavaanlist_inspect_mms(object,
      what = "free",
      type = "free", add_labels = add_labels, add_class = add_class,
      list_by_group = list_by_group,
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "partable" || what == "user") {
    lav_lavaanlist_inspect_mms(object,
      what = "free",
      type = "partable", add_labels = add_labels, add_class = add_class,
      list_by_group = list_by_group,
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "start" || what == "starting.values") {
    lav_lavaanlist_inspect_mms(object,
      what = "start",
      add_labels = add_labels, add_class = add_class,
      list_by_group = list_by_group,
      drop_list_single_group = drop_list_single_group
    )


    #### parameter table ####
  } else if (what == "list") {
    parTable(object)

    #### data + missingness ####
  } else if (what == "ngroups") {
    object@Data@ngroups
  } else if (what == "group") {
    object@Data@group
  } else if (what == "cluster") {
    object@Data@cluster
  } else if (what == "nlevels") {
    object@Data@nlevels
  } else if (what == "nclusters") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "nclusters",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "ncluster.size") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "ncluster.size",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "cluster.size") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "cluster.size",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "cluster.id") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "cluster.id",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "cluster.idx") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "cluster.idx",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "cluster.label") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "cluster.label",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "cluster.sizes") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "cluster.sizes",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "average.cluster.size") {
    lav_inspect_cl_info(object,
      level = 2L,
      what = "average.cluster.size",
      drop_list_single_group = drop_list_single_group
    )
  } else if (what == "ordered") {
    object@Data@ordered
  } else if (what == "group.label") {
    object@Data@group.label
  } else if (what == "level.label") {
    object@Data@level.label
  } else if (what == "nobs") { # only for original!
    unlist(object@Data@nobs)
  } else if (what == "norig") { # only for original!
    unlist(object@Data@norig)
  } else if (what == "ntotal") { # only for original!
    sum(unlist(object@Data@nobs))

    #### from the model object (but stable) over datasets? ####
  } else if (what == "th.idx") {
    lav_lavaanlist_inspect_th_idx(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group
    )


    #### meanstructure, categorical ####
  } else if (what == "meanstructure") {
    object@Model@meanstructure
  } else if (what == "categorical") {
    object@Model@categorical
  } else if (what == "fixed.x") {
    object@Model@fixed.x
  } else if (what == "parameterization") {
    object@Model@parameterization

    # options
  } else if (what == "options" || what == "lavoptions") {
    object@Options

    # call
  } else if (what == "call") {
    as.list(object@call)

    #### not found ####
  } else {
    lav_msg_stop(gettextf(
      "unknown `what' argument in inspect function: `%s'", what))
  }
}


lav_lavaanlist_inspect_start <- function(object) {
  # from 0.5-19, they are in the partable
  if (!is.null(object@ParTable$start)) {
    out <- object@ParTable$start
  } else {
    # in < 0.5-19, we should look in @Fit@start
    out <- object@Fit@start
  }

  out
}

lav_lavaanlist_inspect_mms <- function(
    object, what = "free",
    type = "free", add_labels = FALSE, add_class = FALSE,
    list_by_group = FALSE,
    drop_list_single_group = FALSE) {
  glist <- object@Model@GLIST

  for (mm in seq_along(glist)) {
    if (add_labels) {
      dimnames(glist[[mm]]) <- object@Model@dimNames[[mm]]
    }

    if (what == "free") {
      # fill in free parameter counts
      if (type == "free") {
        m_el_idx <- object@Model@m.free.idx[[mm]]
        x_el_idx <- object@Model@x.free.idx[[mm]]
        # } else if(type == "unco") {
        #    m.el.idx <- object@Model@m.unco.idx[[mm]]
        #    x.el.idx <- object@Model@x.unco.idx[[mm]]
      } else if (type == "partable") {
        m_el_idx <- object@Model@m.user.idx[[mm]]
        x_el_idx <- object@Model@x.user.idx[[mm]]
      } else {
        lav_msg_stop(gettextf("unknown type argument: %s", type))
      }
      # erase everything
      glist[[mm]][, ] <- 0.0
      glist[[mm]][m_el_idx] <- x_el_idx
    } else if (what == "start") {
      # fill in starting values
      m_user_idx <- object@Model@m.user.idx[[mm]]
      x_user_idx <- object@Model@x.user.idx[[mm]]
      start_1 <- lav_lavaanlist_inspect_start(object)
      glist[[mm]][m_user_idx] <- start_1[x_user_idx]
    }

    # class
    if (add_class) {
      if (object@Model@isSymmetric[mm]) {
        class(glist[[mm]]) <- c("lavaan.matrix.symmetric", "matrix")
      } else {
        class(glist[[mm]]) <- c("lavaan.matrix", "matrix")
      }
    }
  }

  # try to reflect `equality constraints'
  con_flag <- FALSE
  if (what == "free" && object@Model@eq.constraints) {
    # extract constraints from parameter table
    pt_1 <- parTable(object)
    con_1 <- pt_1[pt_1$op %in% c("==", "<", ">"), c("lhs", "op", "rhs")]
    rownames(con_1) <- NULL

    # replace 'labels' by parameter numbers
    id <- lav_pt_con_label_id(pt_1)
    label <- names(id)
    for (con in seq_len(nrow(con_1))) {
      # lhs
      lhs_labels <- all.vars(as.formula(paste("~", con_1[con, "lhs"])))

      if (length(lhs_labels) > 0L) {
        # par id
        lhs_freeid <- id[match(lhs_labels, label)]

        # substitute
        tmp <- con_1[con, "lhs"]
        for (pat in seq_along(lhs_labels)) {
          tmp <- sub(lhs_labels[pat], lhs_freeid[pat], tmp)
        }
        con_1[con, "lhs"] <- tmp
      }

      # rhs
      rhs_labels <- all.vars(as.formula(paste("~", con_1[con, "rhs"])))

      if (length(rhs_labels) > 0L) {
        # par id
        rhs_freeid <- id[match(rhs_labels, label)]
        # substitute
        tmp <- con_1[con, "rhs"]
        for (pat in seq_along(rhs_labels)) {
          tmp <- sub(rhs_labels[pat], rhs_freeid[pat], tmp)
        }
        con_1[con, "rhs"] <- tmp
      }
    } # con

    # add this info at the top
    # GLIST <- c(constraints = list(CON), GLIST)
    # no, not a good idea, it does not work with list_by_group

    # add it as a 'header' attribute?
    attr(con_1, "header") <- "Note: model contains equality constraints:"
    con_flag <- TRUE
  }

  # should we group them per group?
  if (list_by_group) {
    lavmodel <- object@Model
    nmat <- lavmodel@nmat

    out <- vector("list", length = object@Data@ngroups)
    for (g in 1:object@Data@ngroups) {
      # which mm belong to group g?
      mm_in_group <- 1:nmat[g] + cumsum(c(0, nmat))[g]

      out[[g]] <- glist[mm_in_group]
    }

    if (object@Data@ngroups == 1L && drop_list_single_group) {
      out <- out[[1]]
    } else {
      if (length(object@Data@group.label) > 0L) {
        names(out) <- unlist(object@Data@group.label)
      }
    }
  } else {
    out <- glist
  }

  # header
  if (con_flag) {
    attr(out, "header") <- con_1
  }

  # lavaan.list
  if (add_class) {
    class(out) <- c("lavaan.list", "list")
  }

  out
}

lav_lavaanlist_inspect_th_idx <- function(
    object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # thresholds idx -- usually, we get it from SampleStats
  # but fortunately, there is a copy in Model, but no names...
  out <- object@Model@th.idx

  # nblocks
  nblocks <- length(out)

  # labels + class
  for (b in seq_len(nblocks)) {
    # if(add_labels && length(OUT[[b]]) > 0L) {
    #    names(OUT[[b]]) <- object@SampleStats@th.names[[b]]
    # }
    if (add_class && !is.null(out[[b]])) {
      class(out[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    out <- out[[1]]
  } else {
    if (object@Data@nlevels == 1L &&
      length(object@Data@group.label) > 0L) {
      names(out) <- unlist(object@Data@group.label)
    } else if (object@Data@nlevels > 1L &&
      length(object@Data@group.label) == 0L) {
      names(out) <- object@Data@level.label
    }
  }

  out
}

# univariate modification indices
#

modindices <- function(object,
                       standardized = TRUE,
                       cov_std = TRUE,
                       information = "expected",
                       # power statistics?
                       power = FALSE,
                       delta = 0.1,
                       alpha = 0.05,
                       high_power = 0.75,
                       # customize output
                       sort = FALSE,
                       minimum_value = 0.0,
                       maximum_number = nrow(list_1),
                       free_remove = TRUE,
                       na_remove = TRUE,
                       op = NULL,
                      ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # rename modified argument
  high_power <- high_power

  # check object
  object <- lav_object_check_version(object)

  # sam objects (issue #517)
  if (!is.null(object@internal$sam.method)) {
    if (lav_sam_local_flag(object)) {
      # delegate to the stored step-2 structural fit: modification indices
      # for a local SAM model concern the structural part only
      fit_pa <- lav_sam_struc_object(object)
      if (is.null(fit_pa)) {
        lav_msg_stop(gettext(
          "no stored structural fit found: this sam object was created by an
           older version of lavaan; please rerun sam()."))
      }
      lav_msg_note(gettext(
        "the modification indices are computed for the structural part only,
         conditional on the (fixed) measurement model of step 1."))
      object <- fit_pa
    } else {
      lav_msg_stop(gettext(
        "modification indices are not available if sam.method = \"global\"."))
    }
  }

  # check if model has converged
  if (object@optim$npar > 0L && !object@optim$converged) {
    lav_msg_warn(gettext("model did not converge"))
  }

  # not ready for estimator = "PML"
  if (object@Options$estimator == "PML") {
    lav_msg_stop(gettext(
      "modification indices for estimator PML are not implemented yet."))
  }

  # not (yet) available for models with composites: the saturated model used to
  # generate candidate parameters (lav_pt_full) is not composite-aware (it would
  # generate '=~' loadings and indicator covariances for composites that the
  # LISREL representation cannot map), and freeing the composite weights/indicator
  # (co)variances breaks composite identification. Use lavTestScore() to assess
  # specific constraints instead.
  if (.hasSlot(object@Model, "composites") && object@Model@composites) {
    lav_msg_stop(gettext(
      "modification indices are not available for models with composites (<~);
      use lavTestScore() to assess specific (constrained) parameters."))
  }

  # new in 0.6-17: check if the model contains equality constraints
  if (object@Model@eq.constraints) {
    lav_msg_warn(
      gettext("the modindices() function ignores equality constraints;
              use lavTestScore() to assess the impact of releasing one or
              multiple constraints.")
    )
  }

  # sanity check
  if (power) {
    standardized <- TRUE
  }

  # extended list (fixed-to-zero parameters)
  strict_exo <- FALSE
  if (object@Model@conditional.x) {
    strict_exo <- TRUE
  }
  full <- lav_pt_full(
    partable = lav_pt_set_cache(object@ParTable, object@pta),
    free = TRUE, start = TRUE,
    strict_exo = strict_exo
  )
  full$free <- rep(1L, nrow(full))
  full$user <- rep(10L, nrow(full))

  fit <- lav_object_extended(object, add = full, all_free = TRUE)
  list_1 <- fit@ParTable


  # compute information matrix 'extended model'
  # ALWAYS use *expected* information (for now)
  information_1 <- lavTech(fit, paste("information", information, sep = "."))

  # compute gradient 'extended model'
  score <- lavTech(fit, "gradient.logl")

  # Saris, Satorra & Sorbom 1987
  # partition Q into Q_11, Q_22 and Q_12/Q_21
  # which elements of Q correspond with 'free' and 'nonfree' parameters?
  model_idx <- list_1$free[list_1$free > 0L & list_1$user != 10L]
  extra_idx <- list_1$free[list_1$free > 0L & list_1$user == 10L]

  # catch empty extra.idx (no modification indices!)
  if (length(extra_idx) == 0L) {
    # 2 possibilities: either model is saturated, or we have constraints
    if (object@test[[1]]$df == 0) {
      lav_msg_warn(gettext(
        "list with extra parameters is empty; model is saturated"))
    } else {
      lav_msg_warn(gettext(
        "list with extra parameters is empty; to release equality
        constraints, use lavTestScore()"))
    }
    list_1 <- data.frame(
      lhs = character(0), op = character(0),
      rhs = character(0), group = integer(0),
      mi = numeric(0), epc = numeric(0),
      sepc.lv = numeric(0), sepc.all = numeric(0),
      sepc.nox = numeric(0)
    )
    return(list_1)
  }

  # partition
  i11 <- information_1[extra_idx, extra_idx, drop = FALSE]
  i12 <- information_1[extra_idx, model_idx, drop = FALSE]
  i21 <- information_1[model_idx, extra_idx, drop = FALSE]
  i22 <- information_1[model_idx, model_idx, drop = FALSE]

  # ALWAYS use *expected* information (for now)
  i22_inv <- try(
    lavTech(object, paste("inverted.information",
      information,
      sep = "."
    )),
    silent = TRUE
  )
  # just in case...
  if (inherits(i22_inv, "try-error")) {
    lav_msg_stop(gettext(
      "could not compute modification indices; information matrix is singular"))
  }

  v <- i11 - i12 %*% i22_inv %*% i21
  v_diag <- diag(v)
  # dirty hack: catch very small or negative values in diag(V)
  # this is needed eg when parameters are not identified if freed-up;
  idx <- which(v_diag < .Machine$double.eps^(1 / 3)) # was 1/2 <0.6-14
  if (length(idx) > 0L) {
    v_diag[idx] <- as.numeric(NA)
  }

  # create and fill in mi
  if (object@Data@nlevels == 1L) {
    n <- object@SampleStats@ntotal
    #if (object@Model@estimator %in% ("ML")) {
      score <- -1 * score # due to gradient.logl
    #}
  } else {
    # total number of clusters (over groups)
    n <- 0
    for (g in 1:object@SampleStats@ngroups) {
      n <- n + object@Data@Lp[[g]]$nclusters[[2]]
    }
    # score <- score * (2 * object@SampleStats@ntotal) / N
    score <- score / 2 # -2 * LRT
  }
  mi <- numeric(length(score))
  mi[extra_idx] <- n * (score[extra_idx] * score[extra_idx]) / v_diag
  if (length(model_idx) > 0L) {
    mi[model_idx] <- n * (score[model_idx] * score[model_idx]) / diag(i22)
  }

  list_1$mi <- rep(as.numeric(NA), length(list_1$lhs))
  list_1$mi[list_1$free > 0] <- mi

  # handle equality constraints (if any)
  # eq.idx <- which(LIST$op == "==")
  # if(length(eq.idx) > 0L) {
  #    OUT <- lavTestScore(object, warn = FALSE)
  #    LIST$mi[ eq.idx ] <- OUT$uni$X2
  # }

  # scaled?
  # if(length(object@test) > 1L) {
  #    LIST$mi.scaled <- LIST$mi / object@test[[2]]$scaling.factor
  # }

  # EPC
  d <- (-1 * n) * score
  # needed? probably not; just in case
  d[which(abs(d) < 1e-15)] <- 1.0
  list_1$epc[list_1$free > 0] <- mi / d


  # standardize?
  if (standardized) {
    epc <- list_1$epc

    if (cov_std) {
      # replace epc values for variances by est values
      var_idx <- which(list_1$op == "~~" & list_1$lhs == list_1$rhs &
        list_1$exo == 0L)
      epc[var_idx] <- list_1$est[var_idx]
    }

    # two problems:
    #   - EPC of variances can be negative, and that is
    #     perfectly legal
    #   - EPC (of variances) can be tiny (near-zero), and we should
    #     not divide by tiny variables
    small_idx <- which(list_1$op == "~~" &
      list_1$lhs == list_1$rhs &
      abs(epc) < sqrt(.Machine$double.eps))
    if (length(small_idx) > 0L) {
      epc[small_idx] <- as.numeric(NA)
    }

    # get the sign
    epc_sign <- sign(list_1$epc)

    list_1$sepc.lv <- epc_sign * lav_standardize_lv(object,
      partable = list_1,
      est = abs(epc),
      cov_std = cov_std
    )
    if (length(small_idx) > 0L) {
      list_1$sepc.lv[small_idx] <- 0
    }
    list_1$sepc.all <- epc_sign * lav_standardize_all(object,
      partable = list_1,
      est = abs(epc),
      cov_std = cov_std
    )
    if (length(small_idx) > 0L) {
      list_1$sepc.all[small_idx] <- 0
    }
    list_1$sepc.nox <- epc_sign * lav_standardize_all_nox(object,
      partable = list_1,
      est = abs(epc),
      cov_std = cov_std
    )
    if (length(small_idx) > 0L) {
      list_1$sepc.nox[small_idx] <- 0
    }
  }

  # power?
  if (power) {
    list_1$delta <- delta
    # FIXME: this is using epc in unstandardized metric
    #        this would be much more useful in standardized metric
    #        we need a lav_standardize_all.reverse function...
    list_1$ncp <- (list_1$mi / (list_1$epc * list_1$epc)) * (delta * delta)
    list_1$power <- 1 - pchisq(qchisq((1.0 - alpha), df = 1),
      df = 1, ncp = list_1$ncp
    )
    list_1$decision <- character(length(list_1$power))

    # five possibilities (Table 6 in Saris, Satorra, van der Veld, 2009)
    mi_significant <- ifelse(1 - pchisq(list_1$mi, df = 1) < alpha,
      TRUE, FALSE
    )
    high_power <- list_1$power > high_power
    # FIXME: sepc.all or epc??
    # epc.high <- abs(LIST$sepc.all) > LIST$delta
    epc_high <- abs(list_1$epc) > list_1$delta

    list_1$decision[which(!mi_significant & !high_power)] <- "(i)"
    list_1$decision[which(mi_significant & !high_power)] <- "**(m)**"
    list_1$decision[which(!mi_significant & high_power)] <- "(nm)"
    list_1$decision[which(mi_significant & high_power &
      !epc_high)] <- "epc:nm"
    list_1$decision[which(mi_significant & high_power &
      epc_high)] <- "*epc:m*"

    # LIST$decision[ which(mi.significant &  high_power) ] <- "epc"
    # LIST$decision[ which(mi.significant & !high_power) ] <- "***"
    # LIST$decision[ which(!mi.significant & !high_power) ] <- "(i)"
  }

  # remove rows corresponding to 'fixed.x' exogenous parameters
  # exo.idx <- which(LIST$exo == 1L & nchar(LIST$plabel) > 0L)
  # if(length(exo.idx) > 0L) {
  #    LIST <- LIST[-exo.idx,]
  # }

  # remove some columns
  list_1$id <- list_1$ustart <- list_1$exo <- list_1$label <-
                                              list_1$plabel <- NULL
  list_1$start <- list_1$free <- list_1$est <- list_1$se <-
                                               list_1$prior <- NULL
  list_1$upper <- list_1$lower <- NULL

  if (power) {
    list_1$sepc.lv <- list_1$sepc.nox <- NULL
  }

  # create data.frame
  list_1 <- as.data.frame(list_1, stringsAsFactors = FALSE)
  class(list_1) <- c("lavaan.data.frame", "data.frame")

  # remove rows corresponding to 'old' free parameters
  if (free_remove) {
    old_idx <- which(list_1$user != 10L)
    if (length(old_idx) > 0L) {
      list_1 <- list_1[-old_idx, ]
    }
  }

  # remove rows corresponding to 'equality' constraints
  eq_idx <- which(list_1$op == "==")
  if (length(eq_idx) > 0L) {
    list_1 <- list_1[-eq_idx, ]
  }

  # remove even more columns
  list_1$user <- NULL

  # remove block/group/level is only single block
  if (lav_pt_nblocks(list_1) == 1L) {
    list_1$block <- NULL
    list_1$group <- NULL
    list_1$level <- NULL
  }

  # sort?
  if (sort) {
    list_1 <- list_1[order(list_1$mi, decreasing = TRUE), ]
  }
  if (minimum_value > 0.0) {
    list_1 <- list_1[!is.na(list_1$mi) & list_1$mi > minimum_value, ]
  }
  if (maximum_number < nrow(list_1)) {
    list_1 <- list_1[seq_len(maximum_number), ]
  }
  if (na_remove) {
    idx <- which(is.na(list_1$mi))
    if (length(idx) > 0) {
      list_1 <- list_1[-idx, ]
    }
  }
  if (!is.null(op)) {
    idx <- list_1$op %in% op
    if (length(idx) > 0) {
      list_1 <- list_1[idx, ]
    }
  }

  # add header
  # TODO: small explanation of the columns in the header?
  #    attr(LIST, "header") <-
  # c("modification indices for newly added parameters only; to\n",
  #   "see the effects of releasing equality constraints, use the\n",
  #   "lavTestScore() function")

  list_1
}

# aliases
modificationIndices <- lav_alias("modindices") # nolint
modificationindices <- lav_alias("modindices") # nolint

# here, we compute various versions of the `information' matrix
# NOTE:
# 1) we ALWAYS compute the UNIT information (not the total information)
#
# 2) by default, we ignore the constraints (we deal with this when we
#    take the inverse later on)

lav_model_info <- function(lavmodel = NULL,
                                  lavsamplestats = NULL,
                                  lavdata = NULL,
                                  lavimplied = NULL,
                                  lavh1 = NULL,
                                  delta = NULL,
                                  lavcache = NULL,
                                  lavoptions = NULL,
                                  extra = FALSE,
                                  augmented = FALSE,
                                  inverted = FALSE,
                                  use_ginv = FALSE) {

  information <- lavoptions$information[1] # ALWAYS used the first one
  # caller can control it

  # rotation?
  # if(!is.null(lavoptions$rotation) && lavoptions$rotation != "none") {
  #    use.ginv <- TRUE
  # }

  if (is.null(lavh1)) {
    lavh1 <- lav_h1_implied_logl(
      lavdata = lavdata,
      lavsamplestats = lavsamplestats,
      lavoptions = lavoptions
    )
  }

  # compute information matrix
  if (information == "observed") {
    if (lavsamplestats@missing.flag || lavdata@nlevels > 1L) {
      group_weight <- FALSE
    } else {
      group_weight <- TRUE
    }
    m_e <- lav_model_info_observed(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, group_weight = group_weight,
      lavoptions = lavoptions, extra = extra,
      augmented = augmented, inverted = inverted, use_ginv = use_ginv
    )
  } else if (information == "expected") {
    m_e <- lav_model_info_expected(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, lavoptions = lavoptions, extra = extra,
      augmented = augmented, inverted = inverted, use_ginv = use_ginv
    )
  } else if (information == "first.order") {
    m_e <- lav_model_info_firstorder(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavimplied = lavimplied, lavh1 = lavh1,
      lavcache = lavcache, lavoptions = lavoptions, # extra = extra,
      check_pd = FALSE,
      augmented = augmented, inverted = inverted, use_ginv = use_ginv
    )
  }

  # information, augmented information, or inverted information
  m_e
}

# sum a list of per-group information blocks
lav_model_info_group_sum <- function(info_group) {
  information <- info_group[[1]]
  if (length(info_group) > 1L) {
    for (g in 2:length(info_group)) {
      information <- information + info_group[[g]]
    }
  }
  information
}

# assemble the parameter-space information from the h1 information:
#
#     Info = sum_g fg * t(Delta_g) %*% A1_g %*% Delta_g
#
# with a diagonal shortcut when A1_g is stored as the diagonal vector of
# the weight matrix (the DWLS/ULS @WLS.VD slot); note that
# lav_model_info_expected_mlm() passes diagonal = FALSE, as its A1 is
# always a full matrix, whatever the estimator
lav_model_info_assemble <- function(lavmodel = NULL,
                                    lavsamplestats = NULL,
                                    delta = NULL,
                                    a1 = NULL,
                                    diagonal =
                                      lavmodel@estimator %in%
                                        c("DWLS", "ULS")) {
  info_group <- vector("list", length = lavsamplestats@ngroups)
  for (g in 1:lavsamplestats@ngroups) {
    # note LISREL documentation suggests (Ng - 1) instead of Ng...
    fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
    # compute information for this group
    if (diagonal) {
      # diagonal weight matrix
      delta2 <- sqrt(a1[[g]]) * delta[[g]]
      info_group[[g]] <- fg * crossprod(delta2)
    } else {
      # full weight matrix
      info_group[[g]] <- fg * (crossprod(delta[[g]], a1[[g]]) %*% delta[[g]])
    }
  }

  # assemble over groups
  lav_model_info_group_sum(info_group)
}

# fisher/expected information
#
# information = Delta' I1 Delta, where I1 is the unit information of
# the saturated model (evaluated either at the structured or unstructured
# estimates)
lav_model_info_expected <- function(lavmodel = NULL,
                                           lavsamplestats = NULL,
                                           lavdata = NULL,
                                           lavoptions = NULL,
                                           lavimplied = NULL,
                                           lavh1 = NULL,
                                           delta = NULL,
                                           lavcache = NULL,
                                           extra = FALSE,
                                           augmented = FALSE,
                                           inverted = FALSE,
                                           use_ginv = FALSE) {
  if (inverted) {
    augmented <- TRUE
  }

  # 1. Delta
  if (is.null(delta)) {
    delta <- lav_model_delta(lavmodel = lavmodel)
  }


  # two-level least-squares? (WLS.V/WLS.VD hold the stacked two-level
  # weight matrix; the standard Delta' A1 Delta path applies per group)
  wls_2l <- (lavdata@nlevels > 1L &&
    lavmodel@estimator %in% c("WLS", "DWLS", "ULS"))

  # GLS/ML: A1 is bdiag(W, 0.5 t(D) (W %x% W) D) with W the inverse of
  # the sample (GLS, ML-unstructured) or model-implied (ML-structured)
  # covariance matrix; when the caller does not need the full matrix
  # (extra = FALSE), stream A1 %*% Delta column-wise instead of forming
  # the (large) A1 -- O(nvar^3) per free parameter instead of O(nvar^4)
  gls_stream <- (lavmodel@estimator == "GLS" && !extra &&
    lavdata@nlevels == 1L &&
    is.null(lavsamplestats@WLS.V[[1]]) &&
    !lavmodel@correlation &&
    !lavmodel@conditional.x &&
    !lavmodel@group.w.free)
  ml_stream <- (lavmodel@estimator == "ML" && !extra &&
    lavdata@nlevels == 1L &&
    !lavsamplestats@missing.flag &&
    !lavmodel@correlation &&
    !lavmodel@conditional.x &&
    !lavmodel@group.w.free)

  # 2. H1 information (single level, or two-level least-squares)
  # for two-level ML there is no pstar-space A1 here (the information is
  # computed directly in parameter space below); a1 stays NULL and the
  # extra = TRUE attribute "WLS.V" is simply absent
  a1 <- NULL
  if (gls_stream || ml_stream) {
    # no A1 needed
  } else if (lavdata@nlevels == 1L || wls_2l) {
    a1 <- lav_model_h1_info_expected(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavoptions = lavoptions,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavcache = lavcache
    )
  } else {
    # force conditional.x = FALSE
    lavimplied <- lav_model_implied_cond2uncond(lavimplied)
  }

  # 3. compute Information per group + assemble over groups
  if (gls_stream || ml_stream) {
    # ML: structured (default) evaluates W at the model-implied moments,
    # unstructured at the sample moments (mirroring the corresponding
    # branch in lav_model_h1_information); GLS is always sample-based
    ml_structured <- TRUE
    if (!is.null(lavoptions) &&
      !is.null(lavoptions$h1.information[1]) &&
      lavoptions$h1.information[1] == "unstructured") {
      ml_structured <- FALSE
    }
    if (ml_stream && ml_structured && length(lavimplied) == 0L) {
      lavimplied <- lav_model_implied(lavmodel = lavmodel)
    }
    info_group <- vector("list", length = lavsamplestats@ngroups)
    for (g in 1:lavsamplestats@ngroups) {
      fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
      v11_scale <- 1.0
      if (gls_stream) {
        w_inv <- lavsamplestats@icov[[g]]
        # mimic Mplus: V11 rescale (full-data input only, see
        # lav_model_h1_information)
        if (isTRUE(lavmodel@estimator.args$gls.v11.mplus) &&
          lavmodel@meanstructure && lavdata@data.type == "full") {
          v11_scale <- lavsamplestats@nobs[[g]] /
            (lavsamplestats@nobs[[g]] - 1)
        }
      } else if (ml_structured) {
        # same inversion route as lav_mvn_info_expected
        w_inv <- lav_mvn_sigma_inv(
          sigma_1 = lavimplied$cov[[g]],
          sinv_method = "eigen"
        )
      } else {
        w_inv <- lavsamplestats@icov[[g]]
      }
      wd <- lav_samp_wls_v_nt_prod(
        m_icov = w_inv,
        x = delta[[g]],
        meanstructure = lavmodel@meanstructure,
        fixed_x = lavmodel@fixed.x,
        x_idx = lavsamplestats@x.idx[[g]],
        v11_scale = v11_scale
      )
      info_group[[g]] <- fg * crossprod(delta[[g]], wd)
    }
    information <- lav_model_info_group_sum(info_group)
  } else if (lavdata@nlevels > 1L && !wls_2l) {
    # multilevel (ML): the information is computed directly in parameter
    # space (no pstar-space A1)
    info_group <- vector("list", length = lavsamplestats@ngroups)
    for (g in 1:lavsamplestats@ngroups) {
      fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal

      # here, we assume only 2 levels, at [[1]] and [[2]]
      if (lavoptions$h1.information[1] == "structured") {
        implied <- lavimplied
      } else {
        implied <- lavh1$implied
      }
      sigma_w <- implied$cov[[(g - 1) * 2 + 1]]
      mu_w <- implied$mean[[(g - 1) * 2 + 1]]
      sigma_b <- implied$cov[[(g - 1) * 2 + 2]]
      mu_b <- implied$mean[[(g - 1) * 2 + 2]]

      info_g <-
        lav_mvn_cl_info_expected_delta(
          lp = lavdata@Lp[[g]],
          delta = delta[[g]],
          mu_w = mu_w,
          sigma_w = sigma_w,
          mu_b = mu_b,
          sigma_b = sigma_b,
          sinv_method = "eigen"
        )
      info_group[[g]] <- fg * info_g
    } # g
    information <- lav_model_info_group_sum(info_group)
  } else {
    information <- lav_model_info_assemble(
      lavmodel = lavmodel, lavsamplestats = lavsamplestats,
      delta = delta, a1 = a1
    )
  }

  # 5. augmented information?
  if (augmented) {
    information <-
      lav_model_info_augment_invert(
        lavmodel = lavmodel,
        information = information,
        inverted = inverted,
        use_ginv = use_ginv
      )
  }

  if (extra) {
    attr(information, "Delta") <- delta
    attr(information, "WLS.V") <- a1 # unweighted
  }

  # possibly augmented/inverted
  information
}

# only for Mplus MLM
lav_model_info_expected_mlm <- function(lavmodel = NULL,
                                               lavsamplestats = NULL,
                                               m_delta = NULL,
                                               extra = FALSE,
                                               augmented = FALSE,
                                               inverted = FALSE,
                                               use_ginv = FALSE) {
  if (inverted) {
    augmented <- TRUE
  }

  if (is.null(m_delta)) {
    m_delta <- lav_model_delta(lavmodel = lavmodel)
  }

  # compute A1
  a1 <- vector("list", length = lavsamplestats@ngroups)
  if (lavmodel@group.w.free) {
    gw <- unlist(lav_model_gw(lavmodel = lavmodel))
  }
  for (g in 1:lavsamplestats@ngroups) {
    a1[[g]] <- lav_mvn_h1_info_expected(
      sample_cov     = lavsamplestats@cov[[g]],
      sample_cov_inv = lavsamplestats@icov[[g]],
      x_idx          = lavsamplestats@x.idx[[g]]
    )
    # the same as GLS... (except for the N/N-1 scaling)
    if (lavmodel@group.w.free) {
      # unweight!!
      a <- exp(gw[g]) / lavsamplestats@nobs[[g]]
      # a <- exp(gw[g]) * lavsamplestats@ntotal / lavsamplestats@nobs[[g]]
      a1[[g]] <- lav_mat_bdiag(matrix(a, 1, 1), a1[[g]])
    }
  }

  # compute Information per group + assemble over groups
  information <- lav_model_info_assemble(
    lavmodel = lavmodel, lavsamplestats = lavsamplestats,
    delta = m_delta, a1 = a1, diagonal = FALSE
  )

  # augmented information?
  if (augmented) {
    information <-
      lav_model_info_augment_invert(
        lavmodel = lavmodel,
        information = information,
        inverted = inverted,
        use_ginv = use_ginv
      )
  }

  if (extra) {
    attr(information, "Delta") <- m_delta
    attr(information, "WLS.V") <- a1 # unweighted
  }

  information
}


lav_model_info_observed <- function(lavmodel = NULL,
                                           lavsamplestats = NULL,
                                           lavdata = NULL,
                                           lavimplied = NULL,
                                           lavh1 = NULL,
                                           lavcache = NULL,
                                           lavoptions = NULL,
                                           extra = FALSE,
                                           group_weight = TRUE,
                                           augmented = FALSE,
                                           inverted = FALSE,
                                           use_ginv = FALSE) {
  if (inverted) {
    augmented <- TRUE
  }

  # observed.information:
  #     - "hessian": second derivative of objective function
  #     - "h1": observed information matrix of saturated (h1) model,
  #             pre- and post-multiplied by the jacobian of the model
  #             parameters (Delta), usually evaluated at the structured
  #             sample statistics (but this depends on the h1.information
  #             option)
  if (!is.null(lavoptions) &&
    !is.null(lavoptions$observed.information[1]) &&
    lavoptions$observed.information[1] == "h1") {
    observed_information <- "h1"
  } else {
    observed_information <- "hessian"
  }

  # HESSIAN based
  if (observed_information == "hessian") {
    hessian <- lav_model_hessian(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavoptions = lavoptions,
      lavcache = lavcache,
      group_weight = group_weight,
      ceq_simple = FALSE
    )

    # NOTE! What is the relationship between the Hessian of the objective
    # function, and the `information' matrix (unit or total)

    # 1. in lavaan, we ALWAYS minimize, so the Hessian is already pos def
    # 2. currently, all estimators give unit information, except MML and PML
    #    so, no need to divide by N
    information <- hessian

    # divide by 'N' for MML and PML
    if (lavmodel@estimator == "PML" || lavmodel@estimator == "MML") {
      information <- information / lavsamplestats@ntotal
      # HJ: Does this need to be divided by sum of weights instead?
    }

    # if multilevel, we should divide by 'J', the number of clusters
    if (lavdata@nlevels > 1L) {
      nc <- 0
      for (g in 1:lavsamplestats@ngroups) {
        nc <- nc + lavdata@Lp[[g]]$nclusters[[2]]
      }
      information <- information * lavsamplestats@ntotal / nc
    }
  }

  # using 'observed h1 information'
  # we need DELTA and 'WLS.V' (=A1)

  if (observed_information == "h1" || extra) {
    # 1. Delta
    delta <- lav_model_delta(lavmodel = lavmodel)

    # 2. H1 information

    a1 <- lav_model_h1_info_observed(
      lavmodel = lavmodel,
      lavsamplestats = lavsamplestats,
      lavdata = lavdata,
      lavoptions = lavoptions,
      lavimplied = lavimplied,
      lavh1 = lavh1,
      lavcache = lavcache
    )
  }

  if (observed_information == "h1") {
    # compute Information per group + assemble over groups
    information <- lav_model_info_assemble(
      lavmodel = lavmodel, lavsamplestats = lavsamplestats,
      delta = delta, a1 = a1
    )
  }

  # augmented information?
  if (augmented) {
    information <-
      lav_model_info_augment_invert(
        lavmodel = lavmodel,
        information = information,
        inverted = inverted,
        use_ginv = use_ginv
      )
  }

  if (extra) {
    attr(information, "Delta") <- delta
    attr(information, "WLS.V") <- a1
  }

  information
}

# outer product of the case-wise scores (gradients)
# HJ 18/10/23: Need to divide sum of crossproduct of individual log-likelihoods
# by sum of weights rather than sample size.
lav_model_info_firstorder <- function(lavmodel = NULL,
                                             lavsamplestats = NULL,
                                             lavdata = NULL,
                                             lavimplied = NULL,
                                             lavh1 = NULL,
                                             lavcache = NULL,
                                             lavoptions = NULL,
                                             check_pd = FALSE,
                                             extra = FALSE,
                                             augmented = FALSE,
                                             inverted = FALSE,
                                             use_ginv = FALSE) {
  if (!lavmodel@estimator %in% c("ML", "PML")) {
    lav_msg_stop(gettext(
      "information = \"first.order\" not available for estimator"),
      sQuote(lavmodel@estimator))
  }

  if (inverted) {
    augmented <- TRUE
  }

  # random slopes (rv() modifier): the h1 route is not available;
  # compute the first-order information directly from (numerically
  # obtained) per-cluster scores in model-parameter space; the 'unit'
  # convention for two-level models is per *cluster*: B0 =
  # crossprod(scores) / nclusters (matching the hessian-based observed
  # information, which is also expressed per cluster)
  if (length(lavmodel@rv.ov) > 0L || length(lavmodel@rv.lv) > 0L) {
    g <- 1L # single group only (enforced elsewhere)
    rs <- lavcache[[g]]$rs
    if (is.null(rs)) {
      rs_info <- lav_mvn_cl_rs_info(lavmodel = lavmodel,
                                    lavdata = lavdata)
      rs_stats <- lav_mvn_cl_rs_stats(
        y1 = lavdata@X[[g]], lp = lavdata@Lp[[g]], rs_info = rs_info
      )
      rs <- list(info = rs_info, stats = rs_stats)
    }
    sc <- lav_mvn_cl_rs_scores(lavmodel = lavmodel, rs = rs)
    nclusters <- lavdata@Lp[[g]]$nclusters[[2]]
    information <- crossprod(sc) / nclusters
    b0_group <- list(information)

    if (augmented) {
      information <-
        lav_model_info_augment_invert(
          lavmodel = lavmodel,
          information = information,
          check_pd = check_pd,
          inverted = inverted,
          use_ginv = use_ginv
        )
    }
    if (extra) {
      attr(information, "B0.group") <- b0_group
    }
    return(information)
  }

  b0_group <- vector("list", lavsamplestats@ngroups)

  # 1. Delta
  delta <- lav_model_delta(lavmodel = lavmodel)

  # 2. H1 information
  b1 <- lav_model_h1_info_firstorder(
    lavmodel = lavmodel,
    lavsamplestats = lavsamplestats,
    lavdata = lavdata,
    lavoptions = lavoptions,
    lavimplied = lavimplied,
    lavh1 = lavh1,
    lavcache = lavcache
  )

  # 3. compute Information per group
  info_group <- vector("list", length = lavsamplestats@ngroups)
  for (g in 1:lavsamplestats@ngroups) {
    # unweighted (needed in lav_test?)
    b0_group[[g]] <- t(delta[[g]]) %*% b1[[g]] %*% delta[[g]]

    # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    wt <- lavdata@weights[[g]]
    if (is.null(wt)) {
      fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
    } else {
      totalwt <- sum(unlist(lavdata@weights))
      fg <- sum(wt) / totalwt
    }

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    # compute information for this group
    if (lavmodel@estimator == "PML") {
      # (fixed July 2026, resolving the "WHAT IS THE ROLE OF fg?" note):
      # for PML, b0_group is a TOTAL (summed-over-cases) information --
      # the crossprod of the casewise pairwise scores -- so the correct
      # aggregate is the plain sum, divided by N further below. The extra
      # fg weight double-counted the group sizes and, combined with the
      # matching error in the hessian-based information, inflated all
      # multigroup PML standard errors by sqrt(N/n_g). Single-group fits
      # are unaffected (fg = 1).
      info_group[[g]] <- b0_group[[g]]
    } else {
      info_group[[g]] <- fg * b0_group[[g]]
    }
  }

  # 4. assemble over groups
  information <- lav_model_info_group_sum(info_group)

  # NOTE: for MML and PML, we get 'total' information (instead of unit) divide
  # by 'N' for MML and PML. For weighted sample, use the sum of weights
  # instead of sample size

  # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

  if (lavmodel@estimator == "PML" || lavmodel@estimator == "MML") {
    if (length(lavdata@sampling.weights) == 0) {
      the_n <- lavsamplestats@ntotal
    } else {
      the_n <- sum(unlist(lavdata@weights))
    }
    information <- information / the_n
    for (g in 1:lavsamplestats@ngroups) {
      b0_group[[g]] <- b0_group[[g]] / the_n
    }
  }
  # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

  # augmented information?
  if (augmented) {
    information <-
      lav_model_info_augment_invert(
        lavmodel = lavmodel,
        information = information,
        check_pd = check_pd,
        inverted = inverted,
        use_ginv = use_ginv
      )
  }

  if (extra) {
    attr(information, "B0.group") <- b0_group
    attr(information, "Delta") <- delta
    attr(information, "WLS.V") <- b1
  }

  information
}


# create augmented information matrix (if needed), and take the inverse
# (if inverted = TRUE), returning only the [1:npar, 1:npar] elements
#
# rm.idx is used by lav_sam_step2_se; it used when the structural model
# contains more parameters than the joint model; therefore, information
# will be 'too small', and we need to remove some columns in H
#
# the constrained inverse is computed via the null-space identity: the
# [1:npar, 1:npar] block of the inverse of the bordered (KKT) matrix
# equals Z (Z' I Z)^-1 Z', with Z an orthonormal basis of the null space
# of the active constraint Jacobian. This needs one solve of order
# npar - rank(h) instead of the SVD of the (npar + 2m) x (npar + 2m)
# bordered matrix, and avoids the MASS::ginv() tolerance altogether. When
# Z'IZ is singular (the model is not identified even on the constrained
# tangent space), we fall back to the bordered matrix + Moore-Penrose
# route, preserving the legacy behavior in that regime. The full bordered
# matrix is still built when inverted = FALSE (lavInspect
# "augmented.information").
lav_model_info_augment_invert <- function(lavmodel = NULL,
                                                 information = NULL,
                                                 inverted = FALSE,
                                                 check_pd = FALSE,
                                                 use_ginv = FALSE,
                                                 rm_idx = integer(0L)) {
  npar <- nrow(information)

  # collect the ACTIVE constraint rows (row-normalized), if any
  h <- NULL
  lambda <- numeric(0L)
  border_type <- "none"
  if (nrow(lavmodel@con.jac) > 0L) {
    border_type <- "general"
    h <- lavmodel@con.jac
    # read the attribute BEFORE any subsetting: `[` drops attributes, so
    # reading it after h[, -rm_idx] silently kept the inactive inequality
    # rows (eg synthesized bounds) and imposed them as active equalities
    # whenever rm_idx was non-empty (the SAM step-2 path)
    inactive_idx <- attr(h, "inactive.idx")
    lambda <- lavmodel@con.lambda # lagrangean coefs
    if (length(inactive_idx) > 0L) {
      h <- h[-inactive_idx, , drop = FALSE]
      lambda <- lambda[-inactive_idx]
    }
    if (length(rm_idx) > 0L) {
      h <- h[, -rm_idx, drop = FALSE]
    }
    # normalize the rows of h: the null space of h (and therefore the
    # constrained inverse) is invariant to row scaling, but crossprod(h)
    # puts the SQUARED row norms into the spectrum of the bordered
    # matrix, and MASS::ginv() uses a tolerance RELATIVE to the largest
    # singular value: a badly scaled constraint (eg 1000*a == 1000*b)
    # raised the cutoff enough to zero genuine directions of the
    # information, silently collapsing the standard errors. Rows that
    # lost all their entries to rm_idx (constraints involving only
    # removed parameters) are dropped.
    row_norm <- sqrt(rowSums(h * h))
    nonzero_idx <- which(row_norm > 0)
    if (length(nonzero_idx) < nrow(h)) {
      h <- h[nonzero_idx, , drop = FALSE]
      lambda <- lambda[nonzero_idx]
      row_norm <- row_norm[nonzero_idx]
    }
    if (nrow(h) > 0L) {
      h <- h / row_norm
      # the multiplier of a rescaled constraint rescales inversely
      lambda <- lambda * row_norm
    }
  } else if (lavmodel@ceq.simple.only) {
    border_type <- "simple"
    h <- t(lav_mat_ortho_complement(lavmodel@ceq.simple.K))
    if (length(rm_idx) > 0L) {
      h <- h[, -rm_idx, drop = FALSE]
      # rows may have lost (some of) their entries; drop empty rows and
      # re-normalize the rest (see the note above)
      row_norm <- sqrt(rowSums(h * h))
      nonzero_idx <- which(row_norm > 0)
      h <- h[nonzero_idx, , drop = FALSE] / row_norm[nonzero_idx]
    }
    lambda <- numeric(nrow(h))
  }
  is_augmented <- !is.null(h) && nrow(h) > 0L

  # build the full bordered (KKT) matrix; only needed when the caller
  # asks for the un-inverted augmented matrix, or as the Moore-Penrose
  # fallback of the inverted path below
  build_border <- function() {
    h0 <- matrix(0, nrow(h), nrow(h))
    # adding crossprod(h) leaves the [1:npar, 1:npar] block of the
    # inverse unchanged (h x = 0 implies (info + h'h) x = info x), but
    # makes the bordered matrix nonsingular when the information is
    # singular only along the constraint-normal directions
    info <- information + crossprod(h)
    if (border_type == "general") {
      # note: the middle 2*diag(lambda) block is fully decoupled (zero
      # off-diagonal blocks on both sides), so it cannot affect the
      # [1:npar, 1:npar] slice of the inverse; it is kept for the shape
      # of the full augmented matrix (lavInspect "augmented.information")
      h10 <- matrix(0, npar, nrow(h))
      dl <- 2 * diag(lambda, nrow(h), nrow(h))
      rbind(
        cbind(info, h10, t(h)),
        cbind(t(h10), dl, h0),
        cbind(h, h0, h0)
      )
    } else {
      rbind(
        cbind(info, t(h)),
        cbind(h, h0)
      )
    }
  }

  if (check_pd) {
    # for a constrained model, the relevant identification check is the
    # information restricted to the constrained tangent space (a bordered
    # matrix itself is a saddle matrix and always indefinite)
    if (is_augmented) {
      z <- lav_mat_ortho_complement(t(h))
      m_check <- crossprod(z, information %*% z)
    } else {
      m_check <- information
    }
    eigvals <- eigen(m_check,
      symmetric = TRUE,
      only.values = TRUE
    )$values
    if (any(eigvals < -1 * .Machine$double.eps^(3 / 4))) {
      lav_msg_warn(gettext(
        "information matrix is not positive definite;
        the model may not be identified"))
    }
  }

  if (!inverted) {
    if (is_augmented) {
      return(build_border())
    }
    return(information)
  }

  # inverted
  if (!is_augmented) {
    if (use_ginv) {
      # note: default tol in MASS::ginv is sqrt(.Machine$double.eps)
      #       which seems a bit too conservative
      #       from 0.5-20, we changed this to .Machine$double.eps^(3/4)
      out <- try(
        MASS::ginv(information,
          tol = .Machine$double.eps^(3 / 4)
        ),
        silent = TRUE
      )
    } else {
      out <- try(solve(information), silent = TRUE)
    }
    return(out)
  }

  # constrained inverse: null-space route
  out <- try(
    {
      z <- lav_mat_ortho_complement(t(h))
      ziz <- crossprod(z, information %*% z)
      v <- z %*% solve(ziz, t(z))
      # enforce exact symmetry
      (v + t(v)) / 2
    },
    silent = TRUE
  )
  if (!inherits(out, "try-error")) {
    return(out)
  }

  # fallback: bordered matrix + Moore-Penrose
  try(
    MASS::ginv(build_border(),
      tol = .Machine$double.eps^(3 / 4)
    )[1:npar,
      1:npar,
      drop = FALSE
    ],
    silent = TRUE
  )
}

# IV/MIIV estimation

# internal function to be used inside lav_optim_noniter
# return 'x', the estimated vector of free parameters
lav_sem_miiv_internal <- function(lavmodel = NULL, lavh1 = NULL,
                                  lavsamplestats = NULL,
                                  lavpartable = NULL,
                                  lavdata = NULL, lavoptions = NULL) {
  # IV options
  iv_method <- toupper(lavoptions$estimator.args$iv_method)
  stopifnot(iv_method %in% "2SLS")
  iv_varcov_method <- toupper(lavoptions$estimator.args$iv_varcov_method)
  iv_samplestats <- lavoptions$estimator.args$iv_samplestats
  if (lavdata@data.type == "moment") {
    # force iv_samplestats = TRUE
    iv_samplestats <- TRUE
  }
  iv_vcov_stage1 <- tolower(lavoptions$estimator.args$iv_vcov_stage1)
  iv_vcov_stage2 <- tolower(lavoptions$estimator.args$iv_vcov_stage2)
  # [[ ]] exact match: $ would partial-match iv_sargan to iv_sargan_adjust
  iv_sargan <- lavoptions$estimator.args[["iv_sargan"]]
  iv_mean_structure <- tolower(lavoptions$estimator.args$iv_mean_structure)
  if (length(iv_mean_structure) == 0L) {
    iv_mean_structure <- "moments"
  }
  # ML divisor (N) instead of the unbiased N-P divisor for the residual
  # variances (see lav_options_est_iv())
  iv_mimic_ml <- isTRUE(lavoptions$estimator.args$iv_mimic_ml)
  # just in case
  if (lavmodel@categorical) {
    iv_samplestats <- TRUE
  }
  stopifnot(iv_varcov_method %in% c("ULS", "GLS", "2RLS", "RLS", "NONE"))

  # get lavpta
  # lavpta <- lav_pt_attributes(lavpartable)

  # we assume the blocks are independent groups for now
  stopifnot(lavdata@nlevels == 1L)

  # directed versus undirected (free) parameters
  undirected_idx <- which(lavpartable$free > 0L &
    !duplicated(lavpartable$free) & # if ceq.simple
    lavpartable$op == "~~")
  directed_idx <- which(lavpartable$free > 0L &
    !duplicated(lavpartable$free) & # if ceq.simple
    lavpartable$op %in% c("=~", "~", "~1"))
  free_directed_idx <- unique(lavpartable$free[directed_idx])
  free_undirected_idx <- unique(lavpartable$free[undirected_idx])

  # external instruments with categorical data: equality constraints among the
  # directed coefficients (eg measurement invariance) are not supported, because
  # the constraint-aware standard errors would require a numerical moment
  # Jacobian, which is not available for categorical data (the analytic
  # per-equation Jacobian cannot represent the pooling). Fail cleanly rather
  # than report unreliable standard errors.
  if (lavmodel@categorical && length(unlist(lavdata@ov.names.aux)) > 0L) {
    dir_slope_free <- lavpartable$free[lavpartable$free > 0L &
      lavpartable$op %in% c("=~", "~")]
    if (anyDuplicated(dir_slope_free) > 0L ||
        length(lavmodel@ceq.linear.idx) > 0L) {
      lav_msg_stop(gettext(
        "external instruments with categorical data do not support equality
         constraints among the directed coefficients (eg measurement
         invariance) yet."))
    }
  }

  # general linear equality constraints are handled stage-wise: among the
  # directed slopes (stage 1) and among the undirected parameters (stage 2).
  # Warn once about constraints we cannot honor this way (constraints on
  # intercepts, or spanning both stages).
  if (length(lavmodel@ceq.linear.idx) > 0L) {
    noint_idx <- which(lavpartable$free > 0L & !duplicated(lavpartable$free) &
      lavpartable$op %in% c("=~", "~"))
    free_noint_idx <- unique(lavpartable$free[noint_idx])
    con_dir <- lav_sem_miiv_linear_con(lavmodel, free_noint_idx)
    con_und <- lav_sem_miiv_linear_con(lavmodel, free_undirected_idx)
    n_handled <- (if (is.null(con_dir)) 0L else nrow(con_dir$jac)) +
      (if (is.null(con_und)) 0L else nrow(con_und$jac))
    if (length(lavmodel@ceq.linear.idx) > n_handled) {
      lav_msg_warn(gettext(
        "[IV] some linear equality constraints involve intercepts or span both
         the directed and undirected parameters; these are not honored by the
         (sequential) IV estimator and are ignored."))
    }
  }

  # find ALL model-implied instrumental variables (miivs) per equation
  eqs <- lav_model_find_iv(lavmodel = lavmodel, lavpartable = lavpartable)

  # weak-instrument diagnostics (warn) or pruning, based on the first-stage
  # F-statistic of each overidentified equation
  # use [[ ]] (exact match): $ would partial-match iv_weak to iv_weak_threshold
  iv_weak <- tolower(lavoptions$estimator.args[["iv_weak"]])
  if (length(iv_weak) > 0L && iv_weak != "none") {
    eqs <- lav_sem_miiv_weak(
      eqs = eqs, lavpta = lav_pt_attributes(lavpartable),
      lavh1 = lavh1, lavsamplestats = lavsamplestats, lavdata = lavdata,
      threshold = lavoptions$estimator.args[["iv_weak_threshold"]],
      action = iv_weak
    )
  }

  # initial parameter vector
  x <- lav_model_get_parameters(lavmodel)

  ########################################
  # first stage: directed parameter only #
  ########################################
  theta1 <- numeric(0L)
  if (length(free_directed_idx) > 0L) {
    if (iv_samplestats) {
      theta1 <- lav_sem_miiv_2sls_samp(
        x = NULL, samplestats = FALSE, eqs = eqs,
        lavmodel = lavmodel, lavpartable = lavpartable,
        lavdata = lavdata, lavsamplestats = lavsamplestats,
        lavh1 = lavh1, free_directed_idx = free_directed_idx,
        iv_vcov_stage1 = iv_vcov_stage1, iv_sargan = iv_sargan,
        iv_vcov_stage2 = iv_vcov_stage2, iv_mimic_ml = iv_mimic_ml
      )
    } else {
      theta1 <- lav_sem_miiv_2sls(
        eqs = eqs,
        lavmodel = lavmodel, lavpartable = lavpartable,
        lavdata = lavdata, free_directed_idx = free_directed_idx,
        iv_vcov_stage1 = iv_vcov_stage1, iv_sargan = iv_sargan,
        iv_vcov_stage2 = iv_vcov_stage2, iv_mimic_ml = iv_mimic_ml
      )
    }
    # update equations
    eqs <- attr(theta1, "eqs")
    theta1 <- as.numeric(theta1) # drop attributes
    # store theta1 elements in x
    x[free_directed_idx] <- theta1
  }

  #######################################
  # second stage: undirected parameters #
  #######################################

  # compute theta2 using ULS/GLS/RLS/2RLS
  if (length(free_undirected_idx) > 0L && iv_varcov_method != "NONE") {
    theta2 <- lav_sem_miiv_varcov(
      x = theta1,
      lavmodel = lavmodel, lavpartable = lavpartable,
      lavsamplestats = lavsamplestats,
      lavh1 = lavh1, free_directed_idx = free_directed_idx,
      free_undirected_idx = free_undirected_idx,
      iv_varcov_method = iv_varcov_method
    )
  } else {
    theta2 <- rep(as.numeric(NA), length(free_undirected_idx))
  }
  # store theta2 elements in x
  x[free_undirected_idx] <- theta2

  # mean structure: re-estimate the free mean parameters (observed intercepts
  # and latent means) jointly by GLS (iv_mean_structure = "wls"), which matches
  # the ML mean estimate and pools shared intercepts/latent means across groups
  # (e.g. for scalar invariance). The default "moments" path keeps the
  # per-equation, nobs-pooled estimates from stage 1.
  if (lavmodel@meanstructure && identical(iv_mean_structure, "wls")) {
    free_mean_idx <- unique(lavpartable$free[lavpartable$op == "~1" &
      lavpartable$free > 0L & !duplicated(lavpartable$free)])
    if (length(free_mean_idx) > 0L) {
      x[free_mean_idx] <- lav_sem_miiv_mean_wls(
        lavmodel = lavmodel, lavsamplestats = lavsamplestats, x = x,
        free_mean_idx = free_mean_idx)
    }
  }

  attr(x, "eqs") <- eqs
  x
}

# joint mean-structure GLS solve. Given the estimated loadings/regressions and
# variances, re-estimate all free mean parameters (observed intercepts nu and
# latent means alpha) by a single GLS solve that fits the model-implied means
# to the sample means across all groups, weighting by nobs * Sigma^{-1}. The
# model-implied means are linear in the mean parameters, so this is a linear
# solve; shared mean parameters (e.g. equal intercepts for scalar invariance)
# are pooled jointly, reproducing the ML mean estimate (given Sigma). The "vcov"
# attribute holds the GLS information inverse A^{-1} (= the SE covariance for
# the continuous, complete-data case).
lav_sem_miiv_mean_wls <- function(lavmodel, lavsamplestats, x, free_mean_idx) {
  implied <- lav_model_implied(lav_model_set_parameters(lavmodel, x = x))
  delta_list <- lav_sem_miiv_delta(lav_model_set_parameters(lavmodel, x = x))
  # fixed mean contribution (free mean parameters set to 0)
  x0 <- x
  x0[free_mean_idx] <- 0
  implied0 <- lav_model_implied(lav_model_set_parameters(lavmodel, x = x0))

  nfree <- length(free_mean_idx)
  amat <- matrix(0, nfree, nfree)
  bvec <- numeric(nfree)
  for (g in seq_len(lavmodel@nblocks)) {
    nvar <- lavmodel@nvar[[g]]
    m_g <- delta_list[[g]][seq_len(nvar), free_mean_idx, drop = FALSE]
    resid <- lavsamplestats@mean[[g]] - implied0$mean[[g]]
    w_g <- try(lav_mat_sym_inverse(implied$cov[[g]]), silent = TRUE)
    if (inherits(w_g, "try-error")) {
      w_g <- diag(nvar)
    }
    w_g <- w_g * lavsamplestats@nobs[[g]]
    amat <- amat + crossprod(m_g, w_g %*% m_g)
    bvec <- bvec + crossprod(m_g, w_g %*% resid)
  }
  ainv <- lav_mat_sym_inverse(amat)
  out <- drop(ainv %*% bvec)
  attr(out, "vcov") <- ainv
  out
}


# external instruments (estimator = "IV"): build the saturated moments over
# [model variables, external instruments] for block 'b' by augmenting the
# model (saturated) covariance/mean with the extra observed variables
# (lavData@aux). The model-variable block is kept exactly equal to the supplied
# model moments; the instrument blocks (and means) come from the raw data.
# Model variables come first, instruments last (so the model sub-block can be
# recovered later). Returns NULL when there are no external instruments for
# this block (so callers keep their model-only moments unchanged).
lav_sem_miiv_aug_moments <- function(lavdata = NULL, b = 1L, ov_names = NULL,
                                     sample_cov = NULL, sample_mean = NULL) {
  aux_names_b <- if (length(lavdata@ov.names.aux) >= b) {
    lavdata@ov.names.aux[[b]]
  } else {
    character(0L)
  }
  if (length(aux_names_b) == 0L || length(lavdata@X) < b ||
    is.null(lavdata@X[[b]]) || is.null(lavdata@aux[[b]])) {
    return(NULL)
  }
  p_model <- length(ov_names)
  full_b <- cbind(lavdata@X[[b]], lavdata@aux[[b]])
  aug_cov <- stats::cov(full_b, use = "pairwise.complete.obs")
  aug_cov[seq_len(p_model), seq_len(p_model)] <- sample_cov
  aug_mean <- NULL
  if (!is.null(sample_mean)) {
    aug_mean <- colMeans(full_b, na.rm = TRUE)
    aug_mean[seq_len(p_model)] <- sample_mean
  }
  list(cov = aug_cov, mean = aug_mean, ov.names = c(ov_names, aux_names_b))
}


# external instruments, CATEGORICAL data: build the augmented sample statistics
# over [model variables, instruments] for block 'b'. The instruments must be
# ordinal (the IV/PIV estimator works on the all-ordinal threshold/correlation
# structure). Runs muthen1984() over the augmented data to obtain the augmented
# polychoric correlation matrix R, the thresholds, and the asymptotic
# covariance of the sample statistics (NACOV = WLS.W * nobs). The model
# correlation block (and the model thresholds) are overwritten with the model's
# own saturated values, so that the model-variable estimates are unchanged.
# Returns NULL when there are no external instruments for this block.
lav_sem_miiv_aug_cat_moments <- function(lavdata = NULL, b = 1L,
                                         model_cor = NULL, model_th = NULL,
                                         nobs = NULL) {
  aux_names_b <- if (length(lavdata@ov.names.aux) >= b) {
    lavdata@ov.names.aux[[b]]
  } else {
    character(0L)
  }
  if (length(aux_names_b) == 0L || length(lavdata@X) < b ||
    is.null(lavdata@X[[b]]) || is.null(lavdata@aux[[b]])) {
    return(NULL)
  }
  model_ov <- lavdata@ov.names[[b]]
  p_model <- length(model_ov)

  # variable types/levels for the augmented set
  m_type <- lavdata@ov$type[match(model_ov, lavdata@ov$name)]
  m_lev <- lavdata@ov$nlev[match(model_ov, lavdata@ov$name)]
  a_type <- lavdata@ov.aux$type[match(aux_names_b, lavdata@ov.aux$name)]
  a_lev <- lavdata@ov.aux$nlev[match(aux_names_b, lavdata@ov.aux$name)]

  # ordinal instruments only (for now)
  if (any(a_type != "ordered")) {
    lav_msg_stop(gettextf(
      "external instrument(s) must be ordered (categorical) when the model is
       categorical; continuous instrument(s): %s",
      paste(aux_names_b[a_type != "ordered"], collapse = " ")))
  }

  full_b <- cbind(lavdata@X[[b]], lavdata@aux[[b]])
  cat1 <- muthen1984(
    data_1 = full_b, ov_names = c(model_ov, aux_names_b),
    ov_types = c(m_type, a_type), ov_levels = c(m_lev, a_lev),
    wls_w = TRUE, allow_empty_cell = FALSE
  )

  R <- unname(cat1$COR)
  th <- unlist(cat1$TH)
  th_idx <- unlist(cat1$TH.IDX)
  # keep the model block exactly equal to the model's saturated statistics
  if (!is.null(model_cor)) {
    R[seq_len(p_model), seq_len(p_model)] <- model_cor
  }
  if (!is.null(model_th)) {
    th[th_idx <= p_model] <- model_th
  }
  nacov <- if (!is.null(cat1$WLS.W) && !is.null(nobs)) {
    cat1$WLS.W * nobs
  } else {
    NULL
  }

  list(R = R, th = th, th.idx = th_idx, nacov = nacov,
       ov.names = c(model_ov, aux_names_b), p.model = p_model)
}


# external instruments, CATEGORICAL data: the column indices, within the
# augmented categorical moment vector [thresholds, off-diagonal correlations],
# that correspond to the model moments (model thresholds + correlations among
# model variables). 'th.idx' is the per-threshold variable index of the
# augmented threshold block.
lav_sem_miiv_aug_cat_keep_idx <- function(p_model = 0L, p_aug = 0L,
                                          th_idx = integer(0L)) {
  nth <- length(th_idx)
  th_keep <- which(th_idx <= p_model)
  ri <- lav_mat_vech_row_idx(p_aug)
  ci <- lav_mat_vech_col_idx(p_aug)
  offdiag <- ri > ci
  ri_off <- ri[offdiag]
  ci_off <- ci[offdiag]
  cor_keep <- which(ri_off <= p_model & ci_off <= p_model)
  c(th_keep, nth + cor_keep)
}


# external instruments: assemble the full (augmented-moment) directed Jacobian
# jac_k for block 'b' from the per-equation k_mat_full matrices stored by
# lav_sem_miiv_2sls_samp(). This mirrors lav_sem_miiv_utils_jack_eqs(), but the
# columns span the AUGMENTED moments [model variables, instruments] (so the
# directed coefficients' dependence on the instrument moments is retained).
# Returns NULL when no augmented Jacobian is available for this block.
lav_sem_miiv_jack_eqs_aug <- function(eqs = NULL, block = 1L, lavmodel = NULL,
                                      lavpartable = NULL,
                                      free_directed_idx = integer(0L)) {
  b <- block
  ntheta1 <- length(free_directed_idx)
  # augmented moment dimension, taken from the first available k_mat_full
  m_aug <- NULL
  for (j in seq_along(eqs[[b]])) {
    if (!is.null(eqs[[b]][[j]]$k_mat_full)) {
      m_aug <- ncol(eqs[[b]][[j]]$k_mat_full)
      break
    }
  }
  if (is.null(m_aug)) {
    return(NULL)
  }
  k_mat <- matrix(0.0, nrow = ntheta1, ncol = m_aug)
  for (j in seq_along(eqs[[b]])) {
    eq <- eqs[[b]][[j]]
    if (is.null(eq$k_mat_full)) {
      next
    }
    tmp <- lavpartable$free[eq$pt]
    tmp <- tmp[tmp != 0]
    free_idx <- match(tmp, free_directed_idx)
    nx <- nrow(eq$k_mat_full)
    if (length(free_idx) > 0L && nx > 0L) {
      if (all(free_idx > 0L)) {
        k_mat[free_idx, ] <- eq$k_mat_full
      } else {
        zero_idx <- which(free_idx == 0L)
        free_idx <- free_idx[-zero_idx]
        if (length(free_idx) > 0L) {
          k_mat[free_idx, ] <- eq$k_mat_full[-zero_idx, , drop = FALSE]
        }
      }
    }
    if (lavmodel@meanstructure) {
      tmp <- lavpartable$free[eq$ptint]
      tmp <- tmp[tmp != 0]
      free_int_idx <- match(tmp, free_directed_idx)
      if (length(free_int_idx) > 0L && free_int_idx > 0L) {
        k_mat[free_int_idx, ] <- eq$k_mat_int_full
      }
    }
  }
  k_mat
}


# external instruments: map a model-moment quantity into the augmented moment
# layout [model variables, instruments]. Returns the column indices, within the
# augmented (mean, vech(cov)) vector, that correspond to the model moments.
lav_sem_miiv_aug_keep_idx <- function(p_model = 0L, p_aug = 0L,
                                      meanstructure = FALSE) {
  if (meanstructure) {
    lav_aux_moment_idx(p_model, p_aug)
  } else {
    ri <- lav_mat_vech_row_idx(p_aug)
    ci <- lav_mat_vech_col_idx(p_aug)
    which(ri <= p_model & ci <= p_model)
  }
}


# external instruments: build the augmented saturated moment vector at the
# estimate, concatenated over blocks (each block: [mean, vech(cov)] of the
# augmented [model variables, instruments] covariance). Returns the vector and
# the per-block augmented variable count (p_aug), as consumed by the aug_vec
# mode of lav_sem_miiv_2sls_samp().
lav_sem_miiv_aug_vec0 <- function(lavdata = NULL, lavmodel = NULL,
                                  lavh1 = NULL) {
  ms <- lavmodel@meanstructure
  nblocks <- lavmodel@nblocks
  vec <- numeric(0L)
  pdim <- integer(nblocks)
  for (b in seq_len(nblocks)) {
    augb <- lav_sem_miiv_aug_moments(lavdata = lavdata, b = b,
      ov_names = lavdata@ov.names[[b]],
      sample_cov = lavh1$implied$cov[[b]],
      sample_mean = if (ms) lavh1$implied$mean[[b]] else NULL)
    pdim[b] <- ncol(augb$cov)
    seg <- if (ms) {
      c(augb$mean, lav_matrix_vech(augb$cov))
    } else {
      lav_matrix_vech(augb$cov)
    }
    vec <- c(vec, seg)
  }
  list(vec = vec, pdim = pdim)
}


# external instruments: the full (augmented-moment) directed Jacobian computed
# NUMERICALLY, by differentiating the (constraint-aware) 2SLS directed solve
# with respect to the augmented moment vector. This is used when there are
# equality constraints among the directed coefficients (which the per-equation
# analytic Jacobian, lav_sem_miiv_jack_eqs_aug(), cannot represent because it
# ignores the cross-equation pooling). Returns an n_directed x ntot_aug matrix,
# with the columns in the same per-block [mean, vech(cov)] order used elsewhere.
lav_sem_miiv_jac_full_numeric <- function(eqs = NULL, lavmodel = NULL,
                                          lavpartable = NULL, lavdata = NULL,
                                          lavsamplestats = NULL, lavh1 = NULL,
                                          free_directed_idx = NULL) {
  av <- lav_sem_miiv_aug_vec0(lavdata = lavdata, lavmodel = lavmodel,
                              lavh1 = lavh1)
  numDeriv::jacobian(func = function(v) {
    lav_sem_miiv_2sls_samp(x = NULL, samplestats = TRUE, eqs = eqs,
      lavmodel = lavmodel, lavpartable = lavpartable, lavdata = lavdata,
      lavsamplestats = lavsamplestats, lavh1 = lavh1,
      free_directed_idx = free_directed_idx,
      aug_vec = v, aug_pdim = av$pdim)
  }, x = av$vec)
}


# assess instrument strength for each IV equation via the first-stage
# F-statistic (Staiger & Stock, 1997): for each endogenous regressor in an
# equation, regress it on the equation's instruments and compute the F for the
# instruments. A minimum F below 'threshold' (rule of thumb: 10) flags weak
# instruments. 'action' controls the response:
#  - "warn":  warn about the affected equations (default); the user can supply
#             stronger instruments manually via the |~ operator
#  - "prune": greedily drop the weakest excess instrument from weak,
#             overidentified equations (never below just-identified)
#  - "none":  do nothing (caller skips this function)
# Returns the (possibly pruned) eqs list.
lav_sem_miiv_weak <- function(eqs = NULL, lavpta = NULL, lavh1 = NULL,
                              lavsamplestats = NULL, lavdata = NULL,
                              threshold = 10,
                              action = "warn") {
  # first-stage R^2 of regressor 'xi' on instruments 'zi' (indices)
  r2_fun <- function(s, xi, zi) {
    if (length(zi) == 0L) {
      return(0)
    }
    beta <- try(solve(s[zi, zi, drop = FALSE], s[zi, xi]), silent = TRUE)
    if (inherits(beta, "try-error")) {
      return(NA_real_)
    }
    min(max(drop(s[xi, zi, drop = FALSE] %*% beta) / s[xi, xi], 0), 1 - 1e-12)
  }
  # first-stage F from an R^2 with nz instruments and N observations
  f_fun <- function(r2, nz, n) {
    df_res <- n - nz - 1L
    if (is.na(r2) || df_res <= 0L || nz <= 0L) {
      return(NA_real_)
    }
    (r2 / nz) / ((1 - r2) / df_res)
  }

  weak_msgs <- character(0L)
  pruned_msgs <- character(0L)

  for (b in seq_along(eqs)) {
    ov_names <- lavpta$vnames$ov[[b]]
    s <- lavh1$implied$cov[[b]]
    n <- lavsamplestats@nobs[[b]]
    # external instruments: use the covariance augmented with the extra
    # observed variables (lavData@aux), so instruments not in the model can be
    # assessed for strength
    if (!is.null(lavdata)) {
      aug <- lav_sem_miiv_aug_moments(
        lavdata = lavdata, b = b, ov_names = ov_names, sample_cov = s
      )
      if (!is.null(aug)) {
        s <- aug$cov
        ov_names <- aug$ov.names
      }
    }
    for (j in seq_along(eqs[[b]])) {
      eq <- eqs[[b]][[j]]
      iv_flag <- if (!is.null(eq$iv_type)) {
        eq$iv_type != "ols"
      } else {
        !identical(eq$rhs_new, eq$miiv)
      }
      if (!iv_flag || identical(eq$rhs_new, "1")) {
        next
      }
      endog <- setdiff(eq$rhs_new, eq$iv) # regressors needing instruments
      if (length(endog) == 0L) {
        next
      }
      iv <- eq$iv
      min_f <- function(iv_set) {
        zi <- match(iv_set, ov_names)
        min(vapply(endog, function(xv) {
          f_fun(r2_fun(s, match(xv, ov_names), zi), length(iv_set), n)
        }, numeric(1)), na.rm = TRUE)
      }
      cur_f <- min_f(iv)

      # prune the weakest excess instruments
      if (action == "prune" && is.finite(cur_f) && cur_f < threshold) {
        repeat {
          if (length(iv) <= length(eq$rhs_new) || min_f(iv) >= threshold) {
            break
          }
          # binding (weakest) regressor; drop the excluded instrument whose
          # removal best preserves that regressor's R^2
          f_each <- vapply(endog, function(xv) {
            f_fun(r2_fun(s, match(xv, ov_names), match(iv, ov_names)),
                  length(iv), n)
          }, numeric(1))
          xi <- match(endog[which.min(f_each)], ov_names)
          cand <- setdiff(iv, eq$rhs_new)
          if (length(cand) == 0L) {
            break
          }
          keep_r2 <- vapply(cand, function(z) {
            r2_fun(s, xi, match(setdiff(iv, z), ov_names))
          }, numeric(1))
          iv <- setdiff(iv, cand[which.max(keep_r2)])
        }
        if (!identical(iv, eq$iv)) {
          new_f <- min_f(iv)
          pruned_msgs <- c(pruned_msgs, sprintf(
            "%s ~ %s: dropped %s (first-stage F %.1f -> %.1f)",
            eq$lhs_new, paste(eq$rhs_new, collapse = " + "),
            paste(setdiff(eq$iv, iv), collapse = ", "), cur_f, new_f))
          eqs[[b]][[j]]$iv <- iv
          eqs[[b]][[j]]$iv_type <- "user" # fixed set downstream
          cur_f <- new_f
        }
      }

      if (is.finite(cur_f) && cur_f < threshold) {
        weak_msgs <- c(weak_msgs, sprintf(
          "%s ~ %s (first-stage F = %.1f)",
          eq$lhs_new, paste(eq$rhs_new, collapse = " + "), cur_f))
      }
    }
  }

  if (length(pruned_msgs) > 0L) {
    lav_msg_warn(gettextf(
      "[IV] weak instruments pruned (iv_weak = \"prune\"):\n%s",
      paste(" -", pruned_msgs, collapse = "\n")))
  }
  if (length(weak_msgs) > 0L) {
    lav_msg_warn(gettextf(
      "[IV] weak instruments (first-stage F < %g) for:\n%s\nConsider supplying
       stronger instruments via the |~ operator, or set
       estimator.args = list(..., iv_weak = \"prune\") to drop weak ones.",
      threshold, paste(" -", weak_msgs, collapse = "\n")))
  }

  eqs
}


# pool the directed (slope) coefficients across equations subject to simple
# equality constraints (ie parameters that share the same 'free' column, such
# as equal factor loadings in different equations).
#
# each equation j contributes a quadratic block to a stacked least-squares
# problem: amat_j = Xhat_j' Xhat_j (centered) and bvec_j = Xhat_j' y_j. Without
# constraints, the stacked system is block-diagonal and the joint solution
# reproduces the per-equation 2SLS estimates exactly. Simple equality
# constraints become linear equality restrictions across the blocks, which we
# enforce with quadprog::solve.QP() (meq = number of equality rows).
#
# 'blocks' is a list with, per equation, $amat (nx x nx), $bvec (nx) and
# $gcol (length nx: the global 'free' column for each slope). The per-equation
# blocks are accumulated into a single system D (over the distinct free slope
# columns) and d; simple equality constraints (parameters that share a free
# column) are handled automatically by this accumulation. General linear
# equality constraints are passed in 'con_jac' (n_con x n, columns aligned to
# sort(unique(gcol))) and 'con_rhs' (length n_con) and enforced with
# quadprog::solve.QP(). Returns a named numeric vector: the pooled estimate per
# distinct free slope column.
lav_sem_miiv_pool_directed <- function(blocks = NULL, con_jac = NULL,
                                       con_rhs = NULL) {
  # drop equations without slopes
  blocks <- blocks[vapply(blocks, function(b) length(b$gcol) > 0L, logical(1))]
  if (length(blocks) == 0L) {
    return(numeric(0L))
  }

  # distinct free slope columns
  free_slope_idx <- sort(unique(unlist(lapply(blocks, `[[`, "gcol"))))
  n <- length(free_slope_idx)

  # accumulate the per-equation blocks into one system (shared columns sum)
  dmat <- matrix(0, n, n)
  dvec <- numeric(n)
  for (bl in blocks) {
    p <- match(bl$gcol, free_slope_idx)
    dmat[p, p] <- dmat[p, p] + bl$amat
    dvec[p] <- dvec[p] + bl$bvec
  }

  if (!is.null(con_jac) && nrow(con_jac) > 0L) {
    sol <- quadprog::solve.QP(
      Dmat = dmat, dvec = dvec, Amat = t(con_jac),
      bvec = con_rhs, meq = nrow(con_jac)
    )$solution
  } else {
    sol <- solve(dmat, dvec)
  }

  names(sol) <- as.character(free_slope_idx)
  sol
}


# extract the linear equality constraints (rows of ceq.JAC) that involve ONLY
# the free columns in 'target_idx'. Returns NULL if there are none, else a list
# with $jac (n_con x length(target_idx), columns aligned to 'target_idx' in the
# given order) and $rhs (length n_con). Constraints whose support is not fully
# inside 'target_idx' (eg cross-stage or intercept constraints) are skipped.
lav_sem_miiv_linear_con <- function(lavmodel = NULL, target_idx = NULL) {
  lin_idx <- lavmodel@ceq.linear.idx
  if (length(lin_idx) == 0L || is.null(lavmodel@ceq.JAC) ||
      nrow(lavmodel@ceq.JAC) == 0L || length(target_idx) == 0L) {
    return(NULL)
  }
  jac <- lavmodel@ceq.JAC[lin_idx, , drop = FALSE]
  rhs <- lavmodel@ceq.rhs[lin_idx]
  keep <- apply(jac, 1L, function(r) {
    nz <- which(abs(r) > 0)
    length(nz) > 0L && all(nz %in% target_idx)
  })
  if (!any(keep)) {
    return(NULL)
  }
  list(jac = jac[keep, target_idx, drop = FALSE], rhs = rhs[keep])
}


# apply the pooled directed solve to the parameter vector 'x'. This is engaged
# when a free (directed slope) column is shared by more than one equation (a
# simple equality constraint, handled via column accumulation) OR when there is
# a general linear equality constraint among the directed slopes (handled via
# quadprog::solve.QP). When neither applies, the per-equation fills are already
# correct and 'x' is returned unchanged. Intercepts (if free) are recomputed
# from the pooled slopes.
lav_sem_miiv_apply_directed_pool <- function(eqs_b = NULL, x = NULL,
                                             lavmodel = NULL,
                                             lavpartable = NULL) {
  slope_blocks <- lapply(eqs_b, `[[`, "slope_block")
  slope_blocks <- slope_blocks[!vapply(slope_blocks, is.null, logical(1))]
  if (length(slope_blocks) == 0L) {
    return(x)
  }
  all_gcol <- unlist(lapply(slope_blocks, `[[`, "gcol"))
  free_slope_idx <- sort(unique(all_gcol))

  # general linear equality constraints among the directed slopes (if any)
  con <- lav_sem_miiv_linear_con(lavmodel, free_slope_idx)

  # intercepts that are shared across equations/groups (e.g. equal intercepts
  # for scalar measurement invariance) must also be pooled, not left as
  # last-write-wins
  all_int <- unlist(lapply(eqs_b, function(e) {
    if (!is.null(e$slope_block) && !is.null(e$ptint) &&
        length(e$ptint) == 1L) {
      lavpartable$free[e$ptint]
    } else {
      integer(0L)
    }
  }))
  shared_int <- anyDuplicated(all_int[all_int > 0L]) > 0L

  # nothing to pool?
  if (anyDuplicated(all_gcol) == 0L && is.null(con) && !shared_int) {
    return(x)
  }

  theta_pool <- lav_sem_miiv_pool_directed(
    slope_blocks,
    con_jac = if (is.null(con)) NULL else con$jac,
    con_rhs = if (is.null(con)) NULL else con$rhs
  )
  # 1. write the pooled slopes
  for (j in seq_along(eqs_b)) {
    sb <- eqs_b[[j]]$slope_block
    if (is.null(sb)) {
      next
    }
    x[sb$gcol] <- theta_pool[as.character(sb$gcol)]
  }
  # 2. recompute the intercepts from the pooled slopes. When an intercept free
  #    column is shared across equations/groups, pool the per-equation
  #    intercepts by an nobs-weighted average (consistent with the nobs-scaled
  #    cross-products used for the slopes); a non-shared intercept reduces to
  #    its single per-equation value.
  if (lavmodel@meanstructure) {
    int_idx <- integer(0L)
    int_val <- numeric(0L)
    int_wgt <- numeric(0L)
    for (j in seq_along(eqs_b)) {
      sb <- eqs_b[[j]]$slope_block
      if (is.null(sb)) {
        next
      }
      eq <- eqs_b[[j]]
      free_int_idx <- lavpartable$free[eq$ptint]
      if (length(free_int_idx) != 1L || free_int_idx <= 0L) {
        next
      }
      # full slope vector in rhs order (free pooled values + fixed ustart)
      eq_free_idx <- lavpartable$free[eq$pt]
      b_full <- lavpartable$ustart[eq$pt]
      b_full[is.na(b_full)] <- 0
      free_pos <- which(eq_free_idx > 0L)
      b_full[free_pos] <- x[eq_free_idx[free_pos]]
      beta0 <- as.vector(sb$y_bar - sb$x_bar %*% b_full)
      int_idx <- c(int_idx, free_int_idx)
      int_val <- c(int_val, beta0)
      int_wgt <- c(int_wgt, if (!is.null(eq$nobs)) eq$nobs else 1)
    }
    if (length(int_idx) > 0L) {
      num <- tapply(int_wgt * int_val, int_idx, sum)
      den <- tapply(int_wgt, int_idx, sum)
      pooled_int <- num / den
      x[as.integer(names(pooled_int))] <- pooled_int
    }
  }
  x
}


# stage 1: use 2SLS to find the regression coefficients of all
#          directed effects in the model -- continuous/raw-data version
lav_sem_miiv_2sls <- function(eqs = NULL, lavmodel = NULL, lavpartable = NULL,
                              lavdata = NULL, free_directed_idx = NULL,
                              iv_vcov_stage1 = "lm.vcov.dfres",
                              iv_vcov_stage2 = "delta",
                              iv_sargan = TRUE, iv_mimic_ml = FALSE) {
  # this function is for continuous/raw-data only
  stopifnot(lavdata@data.type == "full")
  stopifnot(!lavmodel@categorical)
  iv_vcov_stage1 <- tolower(iv_vcov_stage1)
  stopifnot(iv_vcov_stage1 %in% c("lm.vcov.dfres", "lm.vcov", "none"))

  # get lavpta
  lavpta <- lav_pt_attributes(lavpartable)

  if (is.null(eqs)) {
    # find ALL model-implied instrumental variables (miivs) per equation
    eqs <- lav_model_find_iv(lavmodel = lavmodel, lavpartable = lavpartable)
  }

  # number of blocks
  nblocks <- lavmodel@nblocks

  # parameter vector
  x <- lav_model_get_parameters(lavmodel)

  # 2SLS per equation
  # for now: - no equality constraints (yet)
  #          - only OLS (not robust, no lasso, ...)
  for (b in seq_len(nblocks)) {
    # ov.names for this block
    ov_names <- lavpta$vnames$ov[[b]]

    # raw data for this block
    xy <- lavdata@X[[b]]

    # external instruments (estimator = "IV"): append the extra observed
    # variables (lavData@aux) as additional columns so that instruments which
    # are not part of the model can be used by the per-equation 2SLS below. The
    # model variables come first, the instruments last.
    aux_names_b <- if (length(lavdata@ov.names.aux) >= b) {
      lavdata@ov.names.aux[[b]]
    } else {
      character(0L)
    }
    if (length(aux_names_b) > 0L && !is.null(lavdata@aux[[b]])) {
      xy <- cbind(xy, lavdata@aux[[b]])
      ov_names <- c(ov_names, aux_names_b)
    }

    # ingredients for the robust (Browne residual-based) overidentification
    # test: the distribution-free (ADF) ACOV of the sample (co)variances and
    # the sample covariance matrix of all variables in this block. This test
    # is valid under non-normality, where the classical Sargan test is not.
    # The moment block of the ADF Gamma is vech(S) (with the diagonal), so the
    # index map runs over the full vech (including variances).
    gamma_block <- NULL
    cov_block <- NULL
    moment_idx_mat <- NULL
    weights_b <- lavdata@weights[[b]]
    if (iv_sargan && nrow(xy) > ncol(xy)) {
      gamma_block <- try(lav_samp_gamma(
        m_y = xy, meanstructure = FALSE, wt = weights_b
      ), silent = TRUE)
      if (inherits(gamma_block, "try-error")) {
        gamma_block <- NULL
      } else {
        nvar_b <- ncol(xy)
        # ML sample covariance (divisor N), matching the ADF Gamma scale
        if (is.null(weights_b)) {
          cov_block <- crossprod(scale(xy, center = TRUE, scale = FALSE)) /
            nrow(xy)
        } else {
          cov_block <- stats::cov.wt(xy, wt = weights_b, method = "ML")$cov
        }
        # vech index map (column-major lower triangle, WITH the diagonal)
        moment_idx_mat <- matrix(0L, nvar_b, nvar_b)
        moment_idx_mat[lower.tri(moment_idx_mat, diag = TRUE)] <-
          seq_len(nvar_b * (nvar_b + 1L) / 2L)
        moment_idx_mat[upper.tri(moment_idx_mat)] <-
          t(moment_idx_mat)[upper.tri(moment_idx_mat)]
      }
    }

    # estimation per equation
    for (j in seq_along(eqs[[b]])) {
      # this equation
      eq <- eqs[[b]][[j]]

      # iv_flag?
      iv_flag <- TRUE
      if (!is.null(eq$iv_type)) {
        iv_flag <- eq$iv_type != "ols"
      } else {
        # if rhs_new matches miiv, iv_flag is FALSE
        if (identical(eq$rhs_new, eq$miiv)) {
          iv_flag <- FALSE
        }
      }

      # Y: there is always an y variable
      y_idx <- match(eq$lhs_new, ov_names)
      yvec <- xy[, y_idx, drop = TRUE] # y-variable (always scalar)

      # X: usually, there are x variables (apart from the "1")
      if (identical(eq$rhs_new, "1")) {
        x_idx <- integer(0L)
        xmat <- matrix(0, nrow = nrow(xy), ncol = 0L)
      } else {
        x_idx <- match(eq$rhs_new, ov_names)
        xmat <- xy[, x_idx, drop = FALSE]
      }

      # Z: instruments
      if (iv_flag) {
        i_idx <- match(eq$iv, ov_names)
        imat <- cbind(1, xy[, i_idx, drop = FALSE]) # instruments
      }

      # weights
      weights <- lavdata@weights[[b]]

      # sargan vector (classic test) and browne vector (robust test)
      sargan <- rep(as.numeric(NA), 3L)
      names(sargan) <- c("stat", "df", "pvalue")
      browne <- sargan

      # 0. check
      if (iv_flag && length(eq$iv) < length(eq$rhs)) {
        # what to do? skip, or proceed anyway?
        eqs[[b]][[j]]$coef <- numeric(0L)
        eqs[[b]][[j]]$nobs <- nrow(xmat)
        eqs[[b]][[j]]$df_res <- as.integer(NA)
        eqs[[b]][[j]]$resvar <- as.numeric(NA)
        eqs[[b]][[j]]$resvar_df_res <- as.numeric(NA)
        eqs[[b]][[j]]$XX <- crossprod(cbind(1, xmat))
        eqs[[b]][[j]]$vcov <- matrix(0, 0L, 0L)
        eqs[[b]][[j]]$sargan <- sargan
        eqs[[b]][[j]]$browne <- browne
        next
      }

      # 1. regress x on instruments
      if (iv_flag) {
        if (is.null(weights)) {
          fit_x_on_z <- lm.fit(x = imat, y = xmat)
        } else {
          fit_x_on_z <- lm.wfit(x = imat, y = xmat, weights = weights)
        }
        # check for NA's
        if (anyNA(fit_x_on_z$coefficients)) {
          lav_msg_warn(gettextf(
            "regression %s on %s failed (NAs);
                                 redundant instruments?",
            paste(ov_names[x_idx], collapse = " + "),
            paste(ov_names[i_idx], collapse = " + ")
          ))
        }
        xhat <- as.matrix(fit_x_on_z$fitted.values)
      } else {
        # just plain regression: x == xhat
        xhat <- xmat
      }

      # 2. regress y on xhat
      if (is.null(weights)) {
        fit_y_on_xhat <- lm.fit(x = cbind(1, xhat), y = yvec)
      } else {
        fit_y_on_xhat <- lm.wfit(
          x = cbind(1, xhat), y = yvec,
          weights = weights
        )
      }

      # store the (centered) slope system for the pooled solve that handles
      # simple equality constraints across equations (see below)
      if (length(x_idx) > 0L) {
        if (is.null(weights)) {
          x_bar <- colMeans(xhat)
          y_bar <- mean(yvec)
          xc <- sweep(xhat, 2L, x_bar)
          yc <- yvec - y_bar
          amat <- crossprod(xc)
          bvec <- crossprod(xc, yc)
        } else {
          sw <- sum(weights)
          x_bar <- colSums(weights * xhat) / sw
          y_bar <- sum(weights * yvec) / sw
          xc <- sweep(xhat, 2L, x_bar)
          yc <- yvec - y_bar
          amat <- crossprod(xc * weights, xc)
          bvec <- crossprod(xc * weights, yc)
        }
        eq_free_idx <- lavpartable$free[eq$pt]
        fix_idx <- which(eq_free_idx == 0L)
        if (length(fix_idx) > 0L) {
          b_fix <- lavpartable$ustart[eq$pt][fix_idx]
          b_fix[is.na(b_fix)] <- 0
          amat_free <- amat[-fix_idx, -fix_idx, drop = FALSE]
          bvec_free <- bvec[-fix_idx, , drop = FALSE] -
            amat[-fix_idx, fix_idx, drop = FALSE] %*% b_fix
          gcol_free <- eq_free_idx[-fix_idx]
        } else {
          amat_free <- amat
          bvec_free <- bvec
          gcol_free <- eq_free_idx
        }
        if (length(gcol_free) > 0L) {
          eqs[[b]][[j]]$slope_block <- list(
            amat = amat_free, bvec = drop(bvec_free), gcol = gcol_free,
            x_idx = x_idx, x_bar = x_bar, y_bar = y_bar
          )
        }
      }

      # 3. fill estimates in x
      # - eq$pt contains partable/user rows
      # - lavpartable$free[eq$pt] should give the free idx
      free_idx <- lavpartable$free[eq$pt]
      if (length(x_idx) > 0L) {
        if (all(free_idx > 0L)) {
          x[free_idx] <- fit_y_on_xhat$coefficients[-1]
        } else {
          # remove non-free elements
          zero_idx <- which(free_idx == 0L)
          free_idx <- free_idx[-zero_idx]
          if (length(free_idx) > 0) {
            x[free_idx] <- fit_y_on_xhat$coefficients[-1][-zero_idx]
          }
        }
      }
      if (lavmodel@meanstructure) {
        free_int_idx <- lavpartable$free[eq$ptint]
        if (free_int_idx > 0L) {
          x[free_int_idx] <- fit_y_on_xhat$coefficients[1]
        }
      }

      # 4. resvar
      resvar_df_res <- resvar <- df_res <- NULL
      if (iv_vcov_stage1 %in% c("lm.vcov.dfres", "lm.vcov")) {
        df_res <- fit_y_on_xhat$df.residual
        notna_idx <- unname(which(!is.na(fit_y_on_xhat$coefficients)))
        ycoef <- fit_y_on_xhat$coefficients[notna_idx]
        res <- yvec - drop(cbind(1, xmat)[, notna_idx, drop = FALSE] %*% ycoef)
        if (is.null(weights)) {
          sse <- sum(res * res)
        } else {
          sse <- sum(weights * res * res)
        }
        resvar_df_res <- sse / df_res # what we should do...
        resvar <- sse / length(res)
        # iv_mimic_ml: use the ML divisor (N) for the residual variance
        # used in the standard errors / Sargan test as well
        if (iv_mimic_ml) {
          resvar_df_res <- resvar
        }
      }

      # 5. naive cov (for standard errors) (see summary.lm in base R)
      vcov <- NULL
      if (iv_vcov_stage1 %in% c("lm.vcov.dfres", "lm.vcov")) {
        p1 <- 1L:fit_y_on_xhat$rank
        r <- chol2inv(fit_y_on_xhat$qr$qr[p1, p1, drop = FALSE])
        if (iv_vcov_stage1 == "lm.vcov") {
          vcov <- r * resvar
        } else {
          vcov <- r * resvar_df_res
        }
      }

      # 6. Sargan test (see summary.ivreg.R 363--371)
      if (iv_flag && iv_sargan && iv_vcov_stage1 != "none") {
        sargan["df"] <- length(i_idx) - length(x_idx)
        if (sargan["df"] > 0L) {
          if (is.null(weights)) {
            fit_yres_on_z <- lm.fit(x = imat, y = res)
            rssr <- sum((res - mean(res))^2)
            sse2 <- sum(fit_yres_on_z$residuals * fit_yres_on_z$residuals)
          } else {
            fit_yres_on_z <- lm.wfit(x = imat, y = res, weights = weights)
            rssr <- sum(weights * (res - weighted.mean(res, weights))^2)
            sse2 <- sum(weights * fit_yres_on_z$residuals *
              fit_yres_on_z$residuals)
          }
          sargan["stat"] <- length(res) * (1 - sse2 / rssr)
          sargan["pvalue"] <- pchisq(sargan["stat"], sargan["df"],
            lower.tail = FALSE
          )
        }
      }

      # 6b. robust residual-based (Browne) overidentification test, using the
      # distribution-free (ADF) ACOV. The residual g = cov(z, y - x beta) is
      # linear in the sample covariances; the helper extracts the relevant
      # sub-block of the ADF Gamma and forms the test. Skipped when an
      # instrument is redundant (NA slope) or the ADF Gamma is unavailable.
      if (iv_flag && iv_sargan && !is.null(gamma_block) &&
        length(i_idx) - length(x_idx) > 0L) {
        beta_x <- fit_y_on_xhat$coefficients[-1]
        if (!anyNA(beta_x)) {
          browne <- lav_sem_miiv_browne_test(
            y_idx = y_idx, x_idx = x_idx, i_idx = i_idx,
            beta = beta_x, sample_cov = cov_block,
            nacov = gamma_block, nth = 0L,
            moment_idx_mat = moment_idx_mat, nobs = nrow(xy)
          )
        }
      }

      # add info to eqs list
      eqs[[b]][[j]]$coef <- unname(fit_y_on_xhat$coefficients)
      eqs[[b]][[j]]$nobs <- nrow(xmat)
      eqs[[b]][[j]]$df_res <- df_res
      eqs[[b]][[j]]$resvar <- resvar
      eqs[[b]][[j]]$resvar_df_res <- resvar_df_res
      # eqs[[b]][[j]]$XX <- crossprod(cbind(1, xmat)) # needed? or only vcov?
      eqs[[b]][[j]]$vcov <- vcov
      eqs[[b]][[j]]$sargan <- sargan
      eqs[[b]][[j]]$browne <- browne
    } # eqs
  } # nblocks

  # pooled (system) solve for simple equality constraints among the directed
  # slopes, across ALL blocks/groups so that cross-group constraints (e.g.
  # equal loadings for measurement invariance) are pooled jointly rather than
  # left as last-write-wins (no-op when nothing is shared)
  x <- lav_sem_miiv_apply_directed_pool(
    eqs_b = do.call(c, eqs), x = x, lavmodel = lavmodel,
    lavpartable = lavpartable
  )

  # return theta1
  theta1 <- x[free_directed_idx]

  # add equations as an attribute
  attr(theta1, "eqs") <- eqs

  theta1
}

# stage 1: use 2SLS to find the regression coefficients of all
#          directed effects in the model -- samplestats version
# if samplestats = TRUE, x should contain the sample statistics
# (taken from lavh1$implied)
lav_sem_miiv_2sls_samp <- function(x = NULL, samplestats = FALSE,
                                          eqs = NULL, lavmodel = NULL,
                                          lavpartable = NULL, lavdata = NULL,
                                          lavsamplestats = NULL, lavh1 = NULL,
                                          iv_vcov_stage1 = "lm.vcov.dfres",
                                          iv_vcov_stage2 = "delta",
                                          iv_sargan = TRUE,
                                          free_directed_idx = NULL,
                                          aug_vec = NULL, aug_pdim = NULL,
                                          iv_mimic_ml = FALSE) {
  # no conditional.x for now!
  stopifnot(!lavmodel@conditional.x)
  iv_vcov_stage1 <- tolower(iv_vcov_stage1)
  stopifnot(iv_vcov_stage1 %in%
    c("lm.vcov.dfres", "lm.vcov", "gamma", "none"))

  # get lavpta
  lavpta <- lav_pt_attributes(lavpartable)

  if (is.null(eqs)) {
    # find ALL model-implied instrumental variables (miivs) per equation
    eqs <- lav_model_find_iv(lavmodel = lavmodel, lavpartable = lavpartable)
  }

  # external instruments: 'aug_vec' mode. The directed coefficients are
  # computed from an AUGMENTED moment vector [model variables, instruments]
  # supplied directly (one segment per block, concatenated), instead of from
  # the model moments plus the raw-data instrument blocks. This lets a
  # numerical Jacobian differentiate the (constraint-aware) directed solve with
  # respect to the instrument moments as well (see lav_sem_miiv_vcov()).
  use_aug_vec <- !is.null(aug_vec)
  if (use_aug_vec) {
    aug_ms <- lavmodel@meanstructure
    aug_md <- if (aug_ms) {
      aug_pdim + aug_pdim * (aug_pdim + 1L) / 2L
    } else {
      aug_pdim * (aug_pdim + 1L) / 2L
    }
    aug_off <- cumsum(c(0L, aug_md))
  } else if (samplestats) {
    implied <- lav_vec_to_implied(x, lavmodel = lavmodel)
  } else {
    implied <- lavh1$implied
  }

  # number of blocks
  nblocks <- lavmodel@nblocks

  # parameter vector -> all NA (for the free parameters)
  x <- lav_model_get_parameters(lavmodel) * as.numeric(NA)

  # 2SLS per equation
  # for now: - no equality constraints (yet)
  #          - only OLS (not robust, no lasso, ...)
  for (b in seq_len(nblocks)) {
    # ov.names for this block
    ov_names <- lavpta$vnames$ov[[b]]

    # nobs for this 'block'
    nobs <- lavsamplestats@nobs[[b]]

    p_model <- length(ov_names)
    aug_nth <- NULL      # external instruments, categorical: augmented #th
    aug_th_idx <- NULL   # external instruments, categorical: augmented th.idx
    if (use_aug_vec) {
      # reconstruct the augmented covariance/mean directly from aug_vec
      seg <- aug_vec[(aug_off[b] + 1L):aug_off[b + 1L]]
      p_aug <- aug_pdim[b]
      aux_names_b <- if (length(lavdata@ov.names.aux) >= b) {
        lavdata@ov.names.aux[[b]]
      } else {
        character(0L)
      }
      ov_names <- c(ov_names, aux_names_b)
      has_ext_iv <- length(aux_names_b) > 0L
      if (aug_ms) {
        sample_mean <- seg[seq_len(p_aug)]
        sample_cov <- lav_matrix_vech_reverse(seg[-seq_len(p_aug)])
      } else {
        sample_cov <- lav_matrix_vech_reverse(seg)
        # no mean structure: the directed slopes do not depend on the means,
        # but the engine still needs a (held-fixed) mean vector for centering
        mm <- lavh1$implied$mean[[b]]
        if (is.null(mm)) {
          mm <- numeric(p_model)
        }
        instr_mean <- if (has_ext_iv && !is.null(lavdata@aux[[b]])) {
          colMeans(lavdata@aux[[b]], na.rm = TRUE)
        } else {
          numeric(0L)
        }
        sample_mean <- c(mm, instr_mean)
      }
    } else {
      # sample statistics for this block
      sample_cov <- implied$cov[[b]]
      sample_mean <- implied$mean[[b]]

      # external instruments (estimator = "IV"): augment the saturated sample
      # statistics with the extra observed variables (lavData@aux) so the 2SLS
      # estimator can use instruments that are not part of the model. The
      # model-variable block is kept exactly equal to the saturated statistics;
      # the instrument blocks come from the data. Model variables come first
      # and instruments last, so that the model-moment sub-block can be
      # recovered for the (model-dimensioned) Jacobians (k_mat) below.
      if (lavmodel@categorical) {
        # categorical: augmented polychoric correlation matrix (the instruments
        # must be ordinal); the slopes use the correlations only
        augc <- lav_sem_miiv_aug_cat_moments(
          lavdata = lavdata, b = b, model_cor = sample_cov, nobs = nobs
        )
        has_ext_iv <- !is.null(augc)
        if (has_ext_iv) {
          sample_cov <- augc$R
          ov_names <- augc$ov.names
          sample_mean <- c(sample_mean,
            rep(0, length(ov_names) - length(sample_mean)))
          aug_nth <- length(augc$th)
          aug_th_idx <- augc$th.idx
        }
      } else {
        aug <- lav_sem_miiv_aug_moments(
          lavdata = lavdata, b = b, ov_names = ov_names,
          sample_cov = sample_cov, sample_mean = sample_mean
        )
        has_ext_iv <- !is.null(aug)
        if (has_ext_iv) {
          sample_cov <- aug$cov
          sample_mean <- aug$mean
          ov_names <- aug$ov.names
        }
      }
    }

    # ingredients for the robust (Browne residual-based) overidentification
    # test (see lav_sem_miiv_browne_test()). It needs the ACOV (Gamma) of the
    # sample statistics plus a map from a variable pair to its position in the
    # moment block of that ACOV.
    #  - categorical: the moments are polychoric correlations and the classic
    #    Sargan test is not valid; the ACOV is the polychoric ACOV over
    #    [thresholds, correlations] (off-diagonal vech, eq_nth thresholds).
    #  - continuous: the robust test uses the distribution-free (ADF) ACOV of
    #    the sample covariances (vech WITH diagonal, eq_nth = 0), computed from
    #    the (complete) raw data; this is valid under non-normality.
    # 'eq_cov' is the sample covariance/correlation matrix on the SAME scale as
    # the ACOV, used to form the residual moments. Not computed in the 'aug_vec'
    # (numerical-differentiation) calls, where the results are discarded anyway.
    eq_nacov <- NULL
    eq_nth <- 0L
    moment_idx_mat <- NULL
    eq_cov <- sample_cov
    if (!use_aug_vec && iv_sargan) {
      nvar_b <- nrow(sample_cov)
      if (lavmodel@categorical) {
        moment_idx_mat <- matrix(0L, nvar_b, nvar_b)
        moment_idx_mat[lower.tri(moment_idx_mat)] <-
          seq_len(nvar_b * (nvar_b - 1L) / 2L)
        moment_idx_mat <- moment_idx_mat + t(moment_idx_mat)
        if (has_ext_iv) {
          eq_nacov <- augc$nacov
          eq_nth <- aug_nth
        } else {
          eq_nacov <- lavsamplestats@NACOV[[b]]
          eq_nth <- nrow(eq_nacov) - nvar_b * (nvar_b - 1L) / 2L
        }
      } else {
        # continuous: ADF Gamma from the (complete) raw data, ordered to match
        # ov_names (model variables first, external instruments appended)
        xy_b <- if (length(lavdata@X) >= b) lavdata@X[[b]] else NULL
        if (has_ext_iv && !is.null(lavdata@aux[[b]])) {
          xy_b <- cbind(xy_b, lavdata@aux[[b]])
        }
        if (!is.null(xy_b) && !anyNA(xy_b) && nrow(xy_b) > ncol(xy_b)) {
          wt_b <- lavdata@weights[[b]]
          g_try <- try(lav_samp_gamma(
            m_y = xy_b, meanstructure = FALSE, wt = wt_b
          ), silent = TRUE)
          if (!inherits(g_try, "try-error")) {
            eq_nacov <- g_try
            eq_nth <- 0L
            pvar <- ncol(xy_b)
            moment_idx_mat <- matrix(0L, pvar, pvar)
            moment_idx_mat[lower.tri(moment_idx_mat, diag = TRUE)] <-
              seq_len(pvar * (pvar + 1L) / 2L)
            moment_idx_mat[upper.tri(moment_idx_mat)] <-
              t(moment_idx_mat)[upper.tri(moment_idx_mat)]
            # ML sample covariance (divisor N), matching the ADF Gamma scale
            eq_cov <- if (is.null(wt_b)) {
              crossprod(scale(xy_b, center = TRUE, scale = FALSE)) / nrow(xy_b)
            } else {
              stats::cov.wt(xy_b, wt = wt_b, method = "ML")$cov
            }
          }
        }
      }
    }

    # estimation per equation
    for (j in seq_along(eqs[[b]])) {
      # this equation
      eq <- eqs[[b]][[j]]

      # iv_flag?
      iv_flag <- TRUE
      if (!is.null(eq$iv_type)) {
        iv_flag <- eq$iv_type != "ols"
      } else {
        # if rhs_new matches miiv, iv_flag is FALSE
        if (identical(eq$rhs_new, eq$miiv)) {
          iv_flag <- FALSE
        }
      }

      # Y: there is always an y variable
      y_idx <- match(eq$lhs_new, ov_names)
      y_bar <- sample_mean[y_idx]

      # X: usually, there are x variables (apart from the "1")
      if (identical(eq$rhs_new, "1")) {
        x_idx <- integer(0L)
      } else {
        x_idx <- match(eq$rhs_new, ov_names)
        x_bar <- sample_mean[x_idx]
      }
      nx <- length(x_idx)

      # Z: instruments
      s_xy <- sample_cov[x_idx, y_idx, drop = FALSE]
      s_xx <- sample_cov[x_idx, x_idx, drop = FALSE]
      s_yy <- sample_cov[y_idx, y_idx, drop = FALSE]
      nz <- 0L
      if (iv_flag) {
        i_idx <- match(eq$iv, ov_names)
        nz <- length(i_idx)
        s_xz <- sample_cov[x_idx, i_idx, drop = FALSE]
        s_zx <- sample_cov[i_idx, x_idx, drop = FALSE]
        s_zz <- sample_cov[i_idx, i_idx, drop = FALSE]
        s_zy <- sample_cov[i_idx, y_idx, drop = FALSE]
      }

      # sargan vector (classic test) and browne vector (robust test)
      sargan <- rep(as.numeric(NA), 3L)
      names(sargan) <- c("stat", "df", "pvalue")
      browne <- sargan

      # 0. check
      if (iv_flag && length(eq$iv) < length(eq$rhs)) {
        # what to do? skip, or proceed anyway?
        eqs[[b]][[j]]$coef <- numeric(0L)
        eqs[[b]][[j]]$nobs <- nobs
        eqs[[b]][[j]]$df_res <- as.integer(NA)
        eqs[[b]][[j]]$resvar <- as.numeric(NA)
        eqs[[b]][[j]]$resvar_df_res <- as.numeric(NA)
        eqs[[b]][[j]]$XX <- matrix(0, 0L, 0L)
        eqs[[b]][[j]]$vcov <- matrix(0, 0L, 0L)
        eqs[[b]][[j]]$sargan <- sargan
        eqs[[b]][[j]]$browne <- browne
        next
      }

      # 1. + 2. OLS + 2SLS
      fit_y_on_xhat <- list()
      if (iv_flag) {
        # Step 1: compute S_ZZ^{-1} S_ZX and S_ZZ^{-1} S_Zy
        m_w <- lav_mat_sym_solve_spd(s_zz, s_zx)
        v <- lav_mat_sym_solve_spd(s_zz, s_zy)
        # Step 2: build reduced system
        amat <- s_xz %*% m_w
        bvec <- s_xz %*% v
        beta_slopes <- drop(lav_mat_sym_solve_spd(amat, bvec))
        beta0 <- as.vector(y_bar - t(x_bar) %*% beta_slopes)
      } else {
        if (nx > 0L) {
          amat <- s_xx
          bvec <- s_xy
          beta_slopes <- drop(lav_mat_sym_solve_spd(amat, bvec))
          beta0 <- as.vector(y_bar - t(x_bar) %*% beta_slopes)
        } else {
          beta_slopes <- numeric(0L)
          beta0 <- y_bar
        }
      }
      fit_y_on_xhat$coefficients <- c(beta0, beta_slopes)

      # store the (centered) slope system for the pooled solve that handles
      # simple equality constraints across equations (see below). fixed slopes
      # (free == 0) are moved to the right-hand side.
      if (nx > 0L) {
        eq_free_idx <- lavpartable$free[eq$pt]
        fix_idx <- which(eq_free_idx == 0L)
        if (length(fix_idx) > 0L) {
          b_fix <- lavpartable$ustart[eq$pt][fix_idx]
          b_fix[is.na(b_fix)] <- 0
          amat_free <- amat[-fix_idx, -fix_idx, drop = FALSE]
          bvec_free <- bvec[-fix_idx, , drop = FALSE] -
            amat[-fix_idx, fix_idx, drop = FALSE] %*% b_fix
          gcol_free <- eq_free_idx[-fix_idx]
        } else {
          amat_free <- amat
          bvec_free <- bvec
          gcol_free <- eq_free_idx
        }
        if (length(gcol_free) > 0L) {
          eqs[[b]][[j]]$slope_block <- list(
            amat = amat_free, bvec = drop(bvec_free), gcol = gcol_free,
            x_idx = x_idx, x_bar = x_bar, y_bar = y_bar
          )
        }
      }

      # 2b. jac_eqs_k
      k_mat <- NULL
      k_mat_int <- NULL
      # full (augmented-moment) directed Jacobians, kept only when external
      # instruments are present (see the reduction below and the stage-2
      # standard errors in lav_sem_miiv_vcov())
      k_mat_full <- NULL
      k_mat_int_full <- NULL
      if (iv_vcov_stage1 == "gamma" || iv_vcov_stage2 == "h2") {
        nvar <- nrow(sample_cov)
        pstar <- nvar * (nvar + 1) / 2
        # Ex[k, j] = 1 iff x.idx[k] == j
        ex <- matrix(0.0, nx, nvar)
        if (nx > 0L) {
          ex[cbind(seq_len(nx), x_idx)] <- 1.0
        }
        # ---- Jacobian of beta_slopes w.r.t. vech(S) --------------
        a_vec <- lav_mat_vech_row_idx(nvar)
        b_vec <- lav_mat_vech_col_idx(nvar)
        offdiag <- (a_vec > b_vec) # logical: strictly off-diagonal elements

        if (!iv_flag && nx == 0L) {
          j_slopes_s <- matrix(0.0, nrow = 0L, ncol = pstar)
        } else if (!iv_flag) {
          g <- -as.vector(crossprod(ex, beta_slopes))
          g[y_idx] <- g[y_idx] + 1.0
          # build M (p x m) column-by-column in vectorised form
          m <- sweep(ex[, b_vec, drop = FALSE], 2L, g[a_vec], `*`) +
            sweep(ex[, a_vec, drop = FALSE], 2L, g[b_vec] * offdiag, `*`)
          j_slopes_s <- lav_mat_sym_solve_spd(s_xx, m)
        } else {
          r_z <- drop(s_zy) - s_zx %*% beta_slopes
          u <- drop(lav_mat_sym_solve_spd(s_zz, r_z))
          ez <- matrix(0.0, nz, nvar)
          ez[cbind(seq_len(nz), i_idx)] <- 1.0
          ew <- crossprod(m_w, ez)
          cu <- as.vector(crossprod(ez, u))
          cb_x <- as.vector(crossprod(ex, beta_slopes))
          h <- -cu
          h[y_idx] <- h[y_idx] + 1.0
          # build M (p x m) vectorised
          m <-
            sweep(ex[, b_vec, drop = FALSE], 2L, cu[a_vec], `*`) +
            sweep(ew[, a_vec, drop = FALSE], 2L, h[b_vec], `*`) -
            sweep(ew[, b_vec, drop = FALSE], 2L, cb_x[a_vec], `*`) +
            sweep(ex[, a_vec, drop = FALSE], 2L, cu[b_vec] * offdiag, `*`) +
            sweep(ew[, b_vec, drop = FALSE], 2L, h[a_vec] * offdiag, `*`) -
            sweep(ew[, a_vec, drop = FALSE], 2L, cb_x[b_vec] * offdiag, `*`)
          j_slopes_s <- solve(amat, m)
        }

        # fill in k_mat_int and k_mat
        k_mat_int <- numeric(0L)
        if (!lavmodel@categorical && lavmodel@meanstructure) {
          k_mat_int <- numeric(nvar + pstar)
          k_mat_int[y_idx] <- 1.0
          if (nx > 0L) {
            k_mat_int[x_idx] <- -beta_slopes
            k_mat_int[nvar + seq_len(pstar)] <- -x_bar %*% j_slopes_s
          }
          k_mat <- matrix(0.0, nx, nvar + pstar)
          if (nx > 0L) {
            k_mat[, nvar + seq_len(pstar)] <- j_slopes_s
          }
        } else if (lavmodel@categorical) {
          # split: var_num, off
          # FIXME: all ordinal for now (ie no variances)
          tmp <- j_slopes_s[, -lav_mat_diagh_idx(nvar), drop = FALSE]
          # number of thresholds: for external instruments the moments span the
          # augmented [thresholds, correlations] set (aug_nth thresholds), else
          # the model NACOV size minus the correlation block
          nth <- if (has_ext_iv) {
            aug_nth
          } else {
            nrow(lavsamplestats@NACOV[[b]]) - ncol(tmp)
          }
          k_mat_int <- numeric(nth + ncol(tmp))
          k_mat <- matrix(0.0, nx, nth + ncol(tmp))
          if (nx > 0L) {
            k_mat[, nth + seq_len(ncol(tmp))] <- tmp
          }
        } else {
          k_mat <- matrix(0.0, nrow = nx, ncol = pstar)
          if (nx > 0L) {
            k_mat <- j_slopes_s
          }
        }

        # external instruments: the Jacobians here are over the AUGMENTED
        # moments [model variables, instruments]. Keep the full (augmented)
        # versions for the stage-2 standard errors (so the instrument sampling
        # variance is folded into the variance/covariance parameter standard
        # errors, see lav_sem_miiv_vcov()), and also produce model-moment-only
        # versions (the model variables come first) for the model-dimensioned
        # consumers that do not augment (e.g. lav_sem_miiv_utils_jack_eqs()).
        if (has_ext_iv && lavmodel@categorical) {
          k_mat_full <- k_mat
          k_mat_int_full <- k_mat_int
          keep <- lav_sem_miiv_aug_cat_keep_idx(p_model, nvar, aug_th_idx)
          k_mat <- k_mat[, keep, drop = FALSE]
          k_mat_int <- k_mat_int[keep]
        } else if (has_ext_iv && !lavmodel@categorical) {
          k_mat_full <- k_mat
          k_mat_int_full <- k_mat_int
          if (lavmodel@meanstructure) {
            keep <- lav_aux_moment_idx(p_model, nvar)
            k_mat <- k_mat[, keep, drop = FALSE]
            k_mat_int <- k_mat_int[keep]
          } else {
            ri <- lav_mat_vech_row_idx(nvar)
            ci <- lav_mat_vech_col_idx(nvar)
            vech_keep <- which(ri <= p_model & ci <= p_model)
            k_mat <- k_mat[, vech_keep, drop = FALSE]
          }
        }
        k_mat
      }

      # 3. fill estimates in x
      # - eq$pt contains partable/user rows
      # - lavpartable$free[eq$pt] should give the free idx
      free_idx <- lavpartable$free[eq$pt]
      if (nx > 0L) {
        if (all(free_idx > 0L)) {
          x[free_idx] <- fit_y_on_xhat$coefficients[-1]
        } else {
          # remove non-free elements
          zero_idx <- which(free_idx == 0L)
          free_idx <- free_idx[-zero_idx]
          if (length(free_idx) > 0) {
            x[free_idx] <- fit_y_on_xhat$coefficients[-1][-zero_idx]
          }
        }
      }
      if (lavmodel@meanstructure) {
        free_int_idx <- lavpartable$free[eq$ptint]
        if (free_int_idx > 0L) {
          x[free_int_idx] <- fit_y_on_xhat$coefficients[1]
        }
      }

      # 4. resvar
      resvar_df_res <- resvar <- df_res <- NULL
      if (iv_vcov_stage1 %in% c("lm.vcov.dfres", "lm.vcov") || iv_sargan) {
        if (nx > 0L) {
          tmp <- s_xx %*% beta_slopes
          resvar <- as.numeric(s_yy - 2 * t(beta_slopes) %*% s_xy +
            t(beta_slopes) %*% tmp)
        } else {
          resvar <- drop(s_yy)
        }
        df_res <- nobs - (nx + 1L)
        resvar_df_res <- resvar * nobs / df_res
        # iv_mimic_ml: use the ML divisor (N) instead of N-P, so the residual
        # variance used in the standard errors / Sargan test stays on the same
        # (ML) scale as the rescaled sample covariance (resvar = RSS / N)
        if (iv_mimic_ml) {
          resvar_df_res <- resvar
        }
      }

      # 5. naive cov (for standard errors) (see summary.lm in base R)
      vcov <- NULL
      if (iv_vcov_stage1 %in% c("lm.vcov.dfres", "lm.vcov")) {
        this_resvar <- resvar_df_res
        if (iv_vcov_stage1 == "lm.vcov") {
          this_resvar <- resvar
        }
        if (nx > 0L) {
          ainv <-
            lav_mat_sym_solve_spd(amat, diag(x = 1, nrow = nrow(amat)))
          vcov_slopes <- (this_resvar / nobs) * ainv
          vcov_beta0 <-
            (this_resvar / nobs) + t(x_bar) %*% vcov_slopes %*% x_bar
          cov_beta0_slopes <- -vcov_slopes %*% x_bar
          vcov <- matrix(0, nx + 1L, nx + 1L)
          vcov[1, 1] <- vcov_beta0
          vcov[1, -1] <- t(cov_beta0_slopes)
          vcov[-1, 1] <- cov_beta0_slopes
          vcov[-1, -1] <- vcov_slopes
        } else {
          # only intercept
          vcov <- matrix(this_resvar / nobs, 1L, 1L)
        }
      }

      # 6. overidentification test(s)
      if (iv_flag && iv_sargan && iv_vcov_stage1 != "none") {
        # classic Sargan test (see summary.ivreg.R 363--371); for categorical
        # data the p-value is left NA because the classical Sargan test is not
        # valid for polychoric correlations (use the Browne test below instead)
        sargan["df"] <- nz - nx
        if (sargan["df"] > 0L) {
          g <- s_zy - s_zx %*% beta_slopes
          w <- lav_mat_sym_solve_spd(s_zz, g)
          sargan["stat"] <- as.numeric(nobs * t(g) %*% w / resvar_df_res)
          if (!lavmodel@categorical) {
            sargan["pvalue"] <- pchisq(sargan["stat"], sargan["df"],
              lower.tail = FALSE
            )
          }
        }
        # robust residual-based (Browne) test using the ACOV of the sample
        # statistics (polychoric ACOV for categorical data, distribution-free
        # ADF ACOV for continuous data); only when that ACOV is available (it is
        # not in the 'aug_vec' numerical-differentiation calls, nor for missing
        # data, where eq_nacov stays NULL)
        if (!use_aug_vec && !is.null(eq_nacov)) {
          browne <- lav_sem_miiv_browne_test(
            y_idx = y_idx, x_idx = x_idx, i_idx = i_idx,
            beta = beta_slopes, sample_cov = eq_cov,
            nacov = eq_nacov, nth = eq_nth,
            moment_idx_mat = moment_idx_mat, nobs = nobs
          )
        }
      }

      # XX (needed?)
      # XX <- NULL
      #       if (!lavmodel@categorical) {
      #         if (nx > 0L) {
      #           XX <- tcrossprod(c(1, x.bar)) * nobs
      #           XX[-1, -1] <- (S_XX + tcrossprod(x.bar)) * nobs
      #         } else {
      #           XX <- matrix(nobs, 1L, 1L)
      #         }
      #       }

      # add info to eqs list
      eqs[[b]][[j]]$coef <- unname(fit_y_on_xhat$coefficients)
      eqs[[b]][[j]]$nobs <- nobs
      eqs[[b]][[j]]$df_res <- df_res
      eqs[[b]][[j]]$resvar <- resvar
      eqs[[b]][[j]]$resvar_df_res <- resvar_df_res
      # eqs[[b]][[j]]$XX <- XX
      eqs[[b]][[j]]$vcov <- vcov
      eqs[[b]][[j]]$sargan <- sargan
      eqs[[b]][[j]]$browne <- browne
      eqs[[b]][[j]]$k_mat_int <- k_mat_int
      eqs[[b]][[j]]$k_mat <- k_mat
      eqs[[b]][[j]]$k_mat_full <- k_mat_full
      eqs[[b]][[j]]$k_mat_int_full <- k_mat_int_full
    } # eqs
  } # nblocks

  # pooled (system) solve for simple equality constraints among the directed
  # slopes: if a free column is shared by >1 equation (within a group, or
  # across groups for measurement invariance), the per-equation fills above are
  # last-write-wins and must be replaced by a joint solution over ALL
  # blocks/groups (no-op when nothing is shared).
  x <- lav_sem_miiv_apply_directed_pool(
    eqs_b = do.call(c, eqs), x = x, lavmodel = lavmodel,
    lavpartable = lavpartable
  )

  # return theta1
  theta1 <- x[free_directed_idx]

  # add equations as an attribute
  if (!samplestats) {
    attr(theta1, "eqs") <- eqs
  }

  theta1
}


# second stage: compute theta_2 as follows:
# theta_2 = solve(t(Delta2) %*% W2 %*% Delta2) %*% t(Delta2) %*% W2 %*% svec
#
# this is the fast but single-block, no-equality-constraints version
# (for now)
#
# see lav_sem_miiv_varcov_old() for an alternative
# model derivative matrix Delta (per block), with the ceq.simple shared
# columns collapsed so that the columns are indexed by the *free* parameter
# number (matching free_directed_idx / free_undirected_idx). Under
# ceq.simple.only, lav_model_delta() returns nx.unco columns (one per
# unconstrained parameter); indexing those by free numbers would select the
# wrong columns and corrupt the second-stage (co)variance estimates (e.g. the
# residual-variance blow-up seen with equal factor loadings).
lav_sem_miiv_delta <- function(lavmodel, glist = NULL) {
  delta <- lav_model_delta(lavmodel = lavmodel, glist = glist)
  if (lavmodel@ceq.simple.only) {
    delta <- lapply(delta, function(d) d %*% lavmodel@ceq.simple.K)
  }
  delta
}

# solve the undirected (variance/covariance) parameters for a single block
# 'b', given that block's collapsed Delta and the free-parameter indices 'fu'
# of its undirected parameters. ULS/GLS use a single WLS linearization; 2RLS
# and RLS reweight with the model-implied covariance of block 'b'. Returns
# theta2 for 'fu' (with the H attribute when return_h = TRUE).
lav_sem_miiv_varcov_block <- function(b, fu, delta_b, implied, lavmodel, x,
                                      iv_varcov_method, return_h) {
  delta2 <- delta_b[, fu, drop = FALSE]

  # general linear equality constraints among the undirected parameters (if
  # any); aligned to the 'fu' (ie delta2 column) order
  con_und <- lav_sem_miiv_linear_con(lavmodel, fu)
  con_jac <- if (is.null(con_und)) NULL else con_und$jac
  con_rhs <- if (is.null(con_und)) NULL else con_und$rhs

  # svec
  sample_cov <- implied$cov[[b]]
  sample_mean <- implied$mean[[b]]
  if (lavmodel@categorical) {
    svec <- c(implied$th[[b]], lav_mat_vech(sample_cov, diagonal = FALSE))
  } else if (lavmodel@meanstructure) {
    svec <- c(sample_mean, lav_mat_vech(sample_cov))
  } else {
    svec <- lav_mat_vech(sample_cov)
  }

  # initial estimate for theta.2
  if (iv_varcov_method == "GLS") {
    s <- sample_cov
  } else { # ULS, or starting point for 2RLS/RLS
    s <- diag(1, nrow = nrow(sample_cov))
  }
  theta2 <- lav_utils_wls_linearization(delta = delta2, s = s,
                                        meanstructure = lavmodel@meanstructure,
                                        categorical = lavmodel@categorical,
                                        svec = svec, return_h = return_h,
                                        con_jac = con_jac, con_rhs = con_rhs)

  # 2RLS
  if (iv_varcov_method == "2RLS") {
    x[fu] <- theta2
    lavmodel_tmp <- lav_model_set_parameters(lavmodel = lavmodel, x = x)
    sigma <- lav_model_sigma(lavmodel = lavmodel_tmp, extra = FALSE)[[b]]
    # fall back to the ULS estimate if the model-implied Sigma (from the ULS
    # step) is not positive definite (eg a Heywood case)
    new_theta2 <- try(
      lav_utils_wls_linearization(delta = delta2, s = sigma,
                                  meanstructure = lavmodel@meanstructure,
                                  svec = svec, return_h = return_h,
                                  con_jac = con_jac, con_rhs = con_rhs),
      silent = TRUE
    )
    if (inherits(new_theta2, "try-error")) {
      lav_msg_warn(gettext("2RLS for variances/covariances fell back to the
                            ULS estimate: the model-implied covariance matrix
                            is not positive definite."))
    } else {
      theta2 <- new_theta2
    }
  # RLS
  } else if (iv_varcov_method == "RLS") {
    # typically, we need 5-15 iterations, so max = 200 should be enough
    for (i in seq_len(200L)) {
      old_x <- theta2
      x[fu] <- theta2
      lavmodel_tmp <- lav_model_set_parameters(lavmodel = lavmodel, x = x)
      sigma <- lav_model_sigma(lavmodel = lavmodel_tmp, extra = FALSE)[[b]]
      # the RLS weight needs a positive-definite model-implied Sigma; if the
      # current iterate yields a non-PD Sigma (eg a Heywood case), keep the
      # last valid estimate rather than aborting the whole estimator
      new_theta2 <- try(
        lav_utils_wls_linearization(delta = delta2, s = sigma,
                                    meanstructure = lavmodel@meanstructure,
                                    svec = svec, return_h = return_h,
                                    con_jac = con_jac, con_rhs = con_rhs),
        silent = TRUE
      )
      if (inherits(new_theta2, "try-error")) {
        lav_msg_warn(gettext("RLS for variances/covariances stopped early:
                              the model-implied covariance matrix is not
                              positive definite; returning the last valid
                              estimate."))
        theta2 <- old_x
        break
      }
      theta2 <- new_theta2
      sse <- Re(sum((old_x - theta2)^2))
      if (sse < 1e-12 * Re(1 + sum(theta2^2))) {
        break
      } else if (i == 200L) {
        lav_msg_warn(gettext("RLS for variances/covariances did not converge
                              after 200 iterations."))
      }
    }
  }

  theta2
}

lav_sem_miiv_varcov <- function(x = NULL, samplestats = FALSE,
                                lavmodel = NULL, lavpartable = NULL,
                                lavsamplestats = NULL,
                                lavh1 = NULL, free_directed_idx = NULL,
                                free_undirected_idx = NULL,
                                add_h2 = FALSE,
                                iv_varcov_method = "RLS") {

  # x = NULL?
  if (is.null(x)) {
    # calling from vcov?
    x <- lav_model_get_parameters(lavmodel)
    lavmodel_tmp <- lavmodel
    implied <- lavh1$implied
  } else if (samplestats) {
    # only used to compute jac_b numerically
    implied <- lav_vec_to_implied(x, lavmodel = lavmodel)
    x <- lav_model_get_parameters(lavmodel)
    lavmodel_tmp <- lavmodel
  } else {
    # first time, or to compute jac_a numerically
    theta1 <- x
    x <- lav_model_get_parameters(lavmodel)
    x[free_directed_idx] <- theta1
    lavmodel_tmp <- lav_model_set_parameters(lavmodel = lavmodel, x = x)
    implied <- lavh1$implied
  }

  # number of blocks
  nblocks <- lavmodel@nblocks
  delta_list <- lav_sem_miiv_delta(lavmodel_tmp)

  # single group: solve directly (also keeps the H attribute for the SE path)
  if (nblocks == 1L) {
    return(lav_sem_miiv_varcov_block(
      b = 1L, fu = free_undirected_idx, delta_b = delta_list[[1L]],
      implied = implied, lavmodel = lavmodel, x = x,
      iv_varcov_method = iv_varcov_method, return_h = add_h2))
  }

  # multiple groups: each group's moments contribute its own block; solve the
  # undirected parameters per block and assemble in free_undirected_idx order.
  # (Cross-group equality constraints among loadings/intercepts are imposed in
  # stage 1; cross-group constraints among variances, if any, are handled per
  # block here.)
  theta2 <- numeric(length(free_undirected_idx))
  for (b in seq_len(nblocks)) {
    fu_b <- unique(lavpartable$free[lavpartable$op == "~~" &
      lavpartable$free > 0L & lavpartable$block == b])
    fu_b <- fu_b[fu_b %in% free_undirected_idx]
    if (length(fu_b) == 0L) {
      next
    }
    tb <- lav_sem_miiv_varcov_block(
      b = b, fu = fu_b, delta_b = delta_list[[b]], implied = implied,
      lavmodel = lavmodel, x = x, iv_varcov_method = iv_varcov_method,
      return_h = FALSE)
    theta2[match(fu_b, free_undirected_idx)] <- as.numeric(tb)
  }

  theta2
}


# restricted-2SLS covariance for the directed parameters under (simple)
# equality constraints. Each equation contributes its per-equation 2SLS
# coefficient information (the inverse of eq$vcov) to the shared free-parameter
# positions; summing the information across equations that share a free
# parameter yields the variance-weighted restricted-2SLS covariance used by
# MIIVsem. This is algebraically the restricted-GLS covariance K (K' A K)^{-1}
# K' (with A the block-diagonal per-equation information and K the sharing
# map), and reproduces the top-left block of MIIVsem's KKT system
# [[A, R'], [R, 0]]^{-1}. Returns the covariance block aligned to
# free_directed_idx (in the compact / nx.free space).
lav_sem_miiv_directed_vcov_restricted <- function(eqs, lavpartable,
                                                  free_directed_idx, nfree) {
  info <- matrix(0, nfree, nfree)
  for (b in seq_along(eqs)) {
    for (eq in eqs[[b]]) {
      if (is.null(eq$vcov) || nrow(eq$vcov) == 0L) {
        next
      }
      # coefficient -> free parameter index; eq$vcov is ordered as
      # (intercept, slopes), matching c(eq$ptint, eq$pt)
      int_free <- if (!is.null(eq$ptint) && length(eq$ptint) > 0L) {
        lavpartable$free[eq$ptint]
      } else {
        0L
      }
      slope_free <- if (length(eq$rhs) > 0L) {
        lavpartable$free[eq$pt]
      } else {
        integer(0L)
      }
      coef_free <- c(int_free, slope_free)
      keep <- which(coef_free > 0L)
      if (length(keep) == 0L) {
        next
      }
      a_j <- solve(eq$vcov[keep, keep, drop = FALSE])
      fidx <- coef_free[keep]
      info[fidx, fidx] <- info[fidx, fidx] + a_j
    }
  }
  solve(info[free_directed_idx, free_directed_idx, drop = FALSE])
}


# VCOV for free parameters
lav_sem_miiv_vcov <- function(lavmodel = NULL, lavsamplestats = NULL,
                              lavoptions = NULL, lavpartable = NULL,
                              lavimplied = NULL,
                              lavh1 = NULL, lavdata = NULL, eqs = NULL) {
  # iv options
  iv_vcov_stage1 <- tolower(lavoptions$estimator.args$iv_vcov_stage1)
  iv_vcov_stage2 <- tolower(lavoptions$estimator.args$iv_vcov_stage2)
  # iv_samplestats <- lavoptions$estimator.args$iv_samplestats
  iv_vcov_gamma_modelbased <-
    lavoptions$estimator.args$iv_vcov_gamma_modelbased
  iv_varcov_method <- toupper(lavoptions$estimator.args$iv_varcov_method)
  iv_vcov_jack_numerical <- lavoptions$estimator.args$iv_vcov_jack_numerical
  iv_vcov_jaca_numerical <- lavoptions$estimator.args$iv_vcov_jaca_numerical
  # iv_vcov_jacb_numerical <- lavoptions$estimator.args$iv_vcov_jacb_numerical

  # empty vcov
  vcov <- matrix(0, lavmodel@nx.free, lavmodel@nx.free)

  # nblocks
  nblocks <- lavmodel@nblocks

  # eqs?
  if (is.null(eqs)) {
    eqs <- lav_model_find_iv(lavmodel = lavmodel, lavpartable = lavpartable)
  }

  # the raw-data engine (iv_samplestats = FALSE) produces the per-equation
  # covariances (eq$vcov) used for the directed standard errors, but not the
  # analytic moment Jacobian (eq$k_mat / eq$k_mat_full). When those are
  # missing, the directed Jacobian needed for the stage-2 (and gamma) standard
  # errors is computed numerically instead (from the sample statistics, which
  # does not require k_mat).
  eqs_have_kmat <- FALSE
  eqs_have_kmat_full <- FALSE
  for (eqs_b in eqs) {
    for (e in eqs_b) {
      if (!is.null(e$k_mat)) {
        eqs_have_kmat <- TRUE
      }
      if (!is.null(e$k_mat_full)) {
        eqs_have_kmat_full <- TRUE
      }
    }
  }
  if (!eqs_have_kmat) {
    iv_vcov_jack_numerical <- TRUE
  }

  # directed versus undirected (free) parameters
  undirected_idx <- which(lavpartable$free > 0L &
    !duplicated(lavpartable$free) & # if ceq.simple
    lavpartable$op == "~~")
  directed_idx <- which(lavpartable$free > 0L &
    !duplicated(lavpartable$free) & # if ceq.simple
    lavpartable$op %in% c("=~", "~", "~1")) # WITH intercepts!!
  directed_noint_idx <- which(lavpartable$free > 0L &
    !duplicated(lavpartable$free) & # if ceq.simple
    lavpartable$op %in% c("=~", "~")) # NO intercepts!!
  if (lavmodel@categorical) {
    th_idx <- which(lavpartable$free > 0L &
    !duplicated(lavpartable$free) & # if ceq.simple
    lavpartable$op == "|")
    if (length(unlist(lavmodel@num.idx)) > 0L) {
      lav_msg_warn(gettext(
        "[IV] no standard errors (yet) when variables are both ordinal
         and continuous"
      ))
      iv_vcov_stage1 <- "none"
      iv_vcov_stage2 <- "none"
    }
  }

  free_directed_idx <- unique(lavpartable$free[directed_idx])
  free_directed_noint_idx <- unique(lavpartable$free[directed_noint_idx])
  free_undirected_idx <- unique(lavpartable$free[undirected_idx])
  if (lavmodel@categorical) {
    free_th_idx <- unique(lavpartable$free[th_idx])
  }

  # directed free parameters only
  x <- lav_model_get_parameters(lavmodel, type = "free")
  # theta1 <- x[free_directed_idx]
  theta1_noint <- x[free_directed_noint_idx]
  theta2 <- x[free_undirected_idx] # final estimate

  # equality constraints? Two flavors:
  #  - simple (parameters sharing a free column, ceq.simple.only path)
  #  - general linear (rows of ceq.JAC)
  # The analytic Jacobians assume independent equations / no constraints, so
  # when a constraint touches the directed (undirected) parameters we fall back
  # to the numerical Jacobian of the constrained point-estimation function,
  # which differentiates the actual pooled / projected solve.
  dir_slope_free <- lavpartable$free[lavpartable$free > 0L &
    lavpartable$op %in% c("=~", "~")]
  has_shared_dir <- anyDuplicated(dir_slope_free) > 0L
  con_dir <- lav_sem_miiv_linear_con(lavmodel, free_directed_noint_idx)
  con_und <- lav_sem_miiv_linear_con(lavmodel, free_undirected_idx)
  has_dir_con <- has_shared_dir || !is.null(con_dir)
  has_und_con <- !is.null(con_und)

  # how to handle equality constraints among the *directed* parameters:
  #  - lm.vcov(.dfres) (default): variance-weighted restricted-2SLS covariance
  #    for the directed block (matches MIIVsem) -- see
  #    lav_sem_miiv_directed_vcov_restricted()
  #  - gamma: the original delta-method (moment-Jacobian) standard errors, with
  #    a numerical Jacobian of the constrained pooled solve
  # Either way stage 2 still needs the constraint-aware (numerical) directed
  # Jacobian to propagate the directed uncertainty into the undirected block.
  use_restricted_directed <- FALSE
  if (has_dir_con && length(free_directed_idx) > 0L &&
      iv_vcov_stage1 != "none") {
    if (iv_vcov_stage1 %in% c("lm.vcov", "lm.vcov.dfres")) {
      if (isTRUE(lavoptions$estimator.args$iv_samplestats)) {
        # the restricted-2SLS covariance only handles *simple* equality
        # constraints (parameters that share a free column, ceq.simple.only);
        # for general linear directed constraints (rows of ceq.JAC) fall back
        # to the delta method
        if (lavmodel@ceq.simple.only && has_shared_dir && is.null(con_dir)) {
          use_restricted_directed <- TRUE
        } else {
          iv_vcov_stage1 <- "gamma"
        }
      } else {
        lav_msg_warn(gettext(
          "[IV] equality constraints are ignored by the raw-data standard
           errors; use iv_samplestats = TRUE for constraint-aware
           standard errors."))
      }
    }
    iv_vcov_jack_numerical <- TRUE
  }

  # general linear constraints among the undirected parameters: the analytic
  # stage-2 Jacobians (and the h2 weight) ignore the constraint, so switch to
  # the delta method with numerical jac_a / jac_b (which differentiate the
  # constrained projection in lav_sem_miiv_varcov())
  # (iv_vcov_jaca_numerical gates both the numerical jac_a and jac_b below)
  if (has_und_con && iv_vcov_stage2 != "none") {
    iv_vcov_stage2 <- "delta"
    iv_vcov_jaca_numerical <- TRUE
  }

  # switch off iv_vcov_stage2?
  if (length(free_directed_idx) > 0L && iv_vcov_stage1 == "none") {
    iv_vcov_stage2 <- "none"
  }

  # ---- multiple groups (continuous data) ----------------------------------
  # The sample moments are independent across groups, so the moment covariance
  # is block-diagonal. We build that block-diagonal 'gamma_big' (each block its
  # NT moment covariance divided by its nobs), a block-diagonal stage-2 map h2,
  # the (already multiblock) numerical directed Jacobian jac_k, and the
  # restricted-2SLS directed covariance (which loops over all blocks/equations
  # and so handles cross-group constraints). The ceq.simple expansion at the
  # end maps the compact vcov to the parameter table.
  if (nblocks > 1L && iv_vcov_stage1 != "none") {
    two_stage <- any(lavoptions$missing == c("two.stage", "robust.two.stage"))
    # two-stage missing data and categorical data need the explicit moment
    # covariance for the directed block; complete continuous data uses the
    # (exact) restricted-2SLS directed covariance
    use_explicit_gamma <- two_stage || lavmodel@categorical
    vec_list <- lav_implied_to_vec(implied = lavh1$implied,
                                   lavmodel = lavmodel, drop_list = FALSE)
    md <- vapply(vec_list, length, integer(1L)) # moment dim per block
    moff <- cumsum(c(0L, md))                    # column offsets
    ntot_moment <- sum(md)

    # block-diagonal moment covariance gamma_big = bdiag(Gamma_b / nobs_b),
    # where Gamma_b is the (per-observation) asymptotic covariance of the
    # sample moments of group b: the categorical ADF gamma (NACOV) for ordered
    # data, the two-stage gamma (Savalei & Bentler / Savalei & Falk) under
    # two-stage missing data, or the normal-theory gamma otherwise
    gamma_g <- vector("list", nblocks)
    for (b in seq_len(nblocks)) {
      x_idx_b <-
        if (lavmodel@fixed.x) lavsamplestats@x.idx[[b]] else integer(0L)
      if (lavmodel@categorical) {
        gamma_g[[b]] <- lavsamplestats@NACOV[[b]] / lavsamplestats@nobs[[b]]
      } else if (two_stage) {
        mu_b <- lavh1$implied$mean[[b]]
        sig_b <- lavh1$implied$cov[[b]]
        if (identical(tolower(lavoptions$missing), "robust.two.stage")) {
          gg <- lav_mvn_mi_h1_omega_sw(y = lavdata@X[[b]], mp = lavdata@Mp[[b]],
            yp = lavsamplestats@missing[[b]], mu = mu_b, sigma_1 = sig_b,
            x_idx = x_idx_b, information = "observed")
        } else {
          i1 <- lav_mvnorm_missing_information_observed_samplestats(
            yp = lavsamplestats@missing[[b]], mu = mu_b, sigma_1 = sig_b,
            x_idx = x_idx_b)
          gg <- lav_mat_sym_inverse(i1)
        }
        gamma_g[[b]] <- gg / lavsamplestats@nobs[[b]]
      } else {
        cov_b <- if (iv_vcov_gamma_modelbased) lavimplied$cov[[b]] else
          lavh1$implied$cov[[b]]
        if (inherits(try(chol(cov_b), silent = TRUE), "try-error")) {
          cov_b <- lavh1$implied$cov[[b]]
        }
        gamma_g[[b]] <- lav_mat_k_gammant_kt(m_k = diag(md[b]), s = cov_b,
          meanstructure = lavmodel@meanstructure, x_idx = x_idx_b) /
          lavsamplestats@nobs[[b]]
      }
    }
    gamma_big <- lav_mat_bdiag(gamma_g)

    # external instruments with CATEGORICAL data (multiple groups): replace the
    # per-block model gamma (NACOV) and the directed Jacobian with their
    # augmented [thresholds, correlations] versions, so the instrument sampling
    # variability is folded into all standard errors. The augmented directed
    # Jacobian is assembled analytically from k_mat_full (the numerical route is
    # not valid for categorical data).
    ext_cat_mg <- lavmodel@categorical && length(free_directed_idx) > 0L &&
      length(unlist(lavdata@ov.names.aux)) > 0L && eqs_have_kmat_full
    md_aug <- NULL
    moff_aug <- NULL
    cat_keep_b <- NULL
    if (ext_cat_mg) {
      md_aug <- integer(nblocks)
      cat_keep_b <- vector("list", nblocks)
      gamma_aug_b <- vector("list", nblocks)
      for (b in seq_len(nblocks)) {
        augb <- lav_sem_miiv_aug_cat_moments(lavdata = lavdata, b = b,
          model_cor = lavh1$implied$cov[[b]], model_th = lavsamplestats@th[[b]],
          nobs = lavsamplestats@nobs[[b]])
        gamma_aug_b[[b]] <- augb$nacov / lavsamplestats@nobs[[b]]
        cat_keep_b[[b]] <- lav_sem_miiv_aug_cat_keep_idx(
          lavmodel@nvar[[b]], ncol(augb$R), augb$th.idx)
        md_aug[b] <- nrow(augb$nacov)
      }
      moff_aug <- cumsum(c(0L, md_aug))
      gamma_big <- lav_mat_bdiag(gamma_aug_b)
    }

    # numerical directed Jacobian over the stacked moments (jac_k differentiates
    # the per-group 2SLS solve, including the cross-group pooled constraints)
    jac_k <- NULL
    if (length(free_directed_idx) > 0L && ext_cat_mg) {
      # categorical external instruments: analytic augmented Jacobian
      ntot_aug <- sum(md_aug)
      jac_k <- matrix(0, length(free_directed_idx), ntot_aug)
      for (b in seq_len(nblocks)) {
        jb <- lav_sem_miiv_jack_eqs_aug(eqs = eqs, block = b,
          lavmodel = lavmodel, lavpartable = lavpartable,
          free_directed_idx = free_directed_idx)
        jac_k[, moff_aug[b] + seq_len(md_aug[b])] <- jb
      }
    } else if (length(free_directed_idx) > 0L) {
      vec <- unlist(vec_list)
      jac_k <- numDeriv::jacobian(lav_sem_miiv_2sls_samp, x = vec,
        samplestats = TRUE, eqs = eqs, lavmodel = lavmodel,
        lavpartable = lavpartable, lavsamplestats = lavsamplestats,
        lavh1 = lavh1, lavdata = lavdata,
        free_directed_idx = free_directed_idx)
    }

    # directed covariance. Complete continuous data: restricted-2SLS
    # (block-diagonal for unconstrained groups, pooled across groups for
    # cross-group constraints). Two-stage missing data: the explicit
    # jac_k %*% gamma %*% t(jac_k), since the per-equation covariances do not
    # carry the EM-uncertainty correction.
    if (length(free_directed_idx) > 0L) {
      v_dir <- NULL
      if (!use_explicit_gamma) {
        v_dir <- try(lav_sem_miiv_directed_vcov_restricted(eqs = eqs,
          lavpartable = lavpartable, free_directed_idx = free_directed_idx,
          nfree = lavmodel@nx.free), silent = TRUE)
      }
      if (!is.null(v_dir) && !inherits(v_dir, "try-error")) {
        vcov[free_directed_idx, free_directed_idx] <- v_dir
      } else {
        vcov[free_directed_idx, free_directed_idx] <-
          jac_k %*% gamma_big %*% t(jac_k)
      }
    }

    # undirected covariance: block-diagonal stage-2 map, propagating the
    # directed uncertainty via jac_k
    if (iv_vcov_stage2 != "none" && length(free_undirected_idx) > 0L &&
        iv_varcov_method != "NONE") {
      delta_list <- lav_sem_miiv_delta(lavmodel)
      h2_full <- matrix(0, length(free_undirected_idx), ntot_moment)
      delta1_full <- matrix(0, ntot_moment, length(free_directed_idx))
      for (b in seq_len(nblocks)) {
        cols <- moff[b] + seq_len(md[b])
        if (length(free_directed_idx) > 0L) {
          delta1_full[cols, ] <-
            delta_list[[b]][, free_directed_idx, drop = FALSE]
        }
        fu_b <- unique(lavpartable$free[lavpartable$op == "~~" &
          lavpartable$free > 0L & lavpartable$block == b])
        fu_b <- fu_b[fu_b %in% free_undirected_idx]
        if (length(fu_b) == 0L) {
          next
        }
        delta2_b <- delta_list[[b]][, fu_b, drop = FALSE]
        if (lavmodel@categorical || iv_varcov_method == "ULS") {
          s_mat <- diag(lavmodel@nvar[[b]])
        } else if (iv_varcov_method == "GLS") {
          s_mat <- lavh1$implied$cov[[b]]
        } else {
          s_mat <- lavimplied$cov[[b]]
          if (inherits(try(chol(s_mat), silent = TRUE), "try-error")) {
            s_mat <- lavh1$implied$cov[[b]]
          }
        }
        tmp_b <- lav_utils_wls_linearization(delta = delta2_b, s = s_mat,
          meanstructure = lavmodel@meanstructure,
          categorical = lavmodel@categorical,
          svec = vec_list[[b]], return_h = TRUE)
        h2_b <- attr(tmp_b, "H")
        # the stage-2 map covers the (co)variance/correlation moments; for
        # categorical data the threshold moments come first in each block and
        # are left at zero (the (co)variance parameters do not depend on them)
        nth_b <- md[b] - ncol(h2_b)
        h2_cols <- moff[b] + nth_b + seq_len(ncol(h2_b))
        h2_full[match(fu_b, free_undirected_idx), h2_cols] <- h2_b
      }
      # external instruments (continuous data): fold the instrument sampling
      # variance into the variance/covariance parameter standard errors, by
      # rebuilding the stage-2 map over the per-block AUGMENTED moments
      # [model variables, instruments] and sandwiching with the block-diagonal
      # normal-theory gamma of the augmented covariances (see the single-group
      # path for the rationale). The directed coefficients' dependence on the
      # instrument moments enters through the full (augmented) directed
      # Jacobian assembled analytically from k_mat_full.
      ext_aux_mg <- !use_explicit_gamma && length(free_directed_idx) > 0L &&
        length(unlist(lavdata@ov.names.aux)) > 0L
      if (ext_aux_mg) {
        ms <- lavmodel@meanstructure
        md_aug <- integer(nblocks)
        cov_aug_b <- vector("list", nblocks)
        keep_b <- vector("list", nblocks)
        gamma_aug_b <- vector("list", nblocks)
        for (b in seq_len(nblocks)) {
          x_idx_b <-
            if (lavmodel@fixed.x) lavsamplestats@x.idx[[b]] else integer(0L)
          augb <- lav_sem_miiv_aug_moments(lavdata = lavdata, b = b,
            ov_names = lavdata@ov.names[[b]],
            sample_cov = lavh1$implied$cov[[b]],
            sample_mean = if (ms) lavh1$implied$mean[[b]] else NULL)
          cov_aug_b[[b]] <- augb$cov
          p_model_b <- lavmodel@nvar[[b]]
          p_aug_b <- ncol(augb$cov)
          keep_b[[b]] <- lav_sem_miiv_aug_keep_idx(p_model_b, p_aug_b, ms)
          md_aug[b] <- if (ms) {
            p_aug_b + p_aug_b * (p_aug_b + 1L) / 2L
          } else {
            p_aug_b * (p_aug_b + 1L) / 2L
          }
          gamma_aug_b[[b]] <- lav_mat_k_gammant_kt(m_k = diag(md_aug[b]),
            s = cov_aug_b[[b]], meanstructure = ms, x_idx = x_idx_b) /
            lavsamplestats@nobs[[b]]
        }
        moff_aug <- cumsum(c(0L, md_aug))
        ntot_aug <- sum(md_aug)
        gamma_big_aug <- lav_mat_bdiag(gamma_aug_b)
        jac_k_aug <- matrix(0, length(free_directed_idx), ntot_aug)
        delta1_aug <- matrix(0, ntot_aug, length(free_directed_idx))
        h2_aug <- matrix(0, length(free_undirected_idx), ntot_aug)
        for (b in seq_len(nblocks)) {
          cols_b <- moff_aug[b] + seq_len(md_aug[b])
          model_cols_b <- moff_aug[b] + keep_b[[b]]
          if (!has_dir_con && eqs_have_kmat_full) {
            jb <- lav_sem_miiv_jack_eqs_aug(eqs = eqs, block = b,
              lavmodel = lavmodel, lavpartable = lavpartable,
              free_directed_idx = free_directed_idx)
            jac_k_aug[, cols_b] <- jb
          }
          delta1_aug[model_cols_b, ] <-
            delta_list[[b]][, free_directed_idx, drop = FALSE]
          fu_b <- unique(lavpartable$free[lavpartable$op == "~~" &
            lavpartable$free > 0L & lavpartable$block == b])
          fu_b <- fu_b[fu_b %in% free_undirected_idx]
          if (length(fu_b) == 0L) {
            next
          }
          delta2_b <- delta_list[[b]][, fu_b, drop = FALSE]
          if (iv_varcov_method == "ULS") {
            s_mat <- diag(lavmodel@nvar[[b]])
          } else if (iv_varcov_method == "GLS") {
            s_mat <- lavh1$implied$cov[[b]]
          } else {
            s_mat <- lavimplied$cov[[b]]
            if (inherits(try(chol(s_mat), silent = TRUE), "try-error")) {
              s_mat <- lavh1$implied$cov[[b]]
            }
          }
          tmp_b <- lav_utils_wls_linearization(delta = delta2_b, s = s_mat,
            meanstructure = ms, categorical = FALSE,
            svec = vec_list[[b]], return_h = TRUE)
          h2_b <- attr(tmp_b, "H")
          nth_b <- md[b] - ncol(h2_b) # mean offset (continuous)
          h2_b <- cbind(matrix(0, nrow(h2_b), nth_b), h2_b)
          h2_aug[match(fu_b, free_undirected_idx), model_cols_b] <- h2_b
        }
        if (has_dir_con || !eqs_have_kmat_full) {
          # use the numerical augmented Jacobian when the analytic k_mat_full
          # is unavailable (raw-data engine) or when there are equality
          # constraints among the directed coefficients (the per-equation
          # analytic Jacobian cannot represent the cross-equation/cross-group
          # pooling)
          jac_k_aug <- lav_sem_miiv_jac_full_numeric(
            eqs = eqs, lavmodel = lavmodel, lavpartable = lavpartable,
            lavdata = lavdata, lavsamplestats = lavsamplestats,
            lavh1 = lavh1, free_directed_idx = free_directed_idx)
        }
        tmp <- h2_aug %*% (diag(ntot_aug) - delta1_aug %*% jac_k_aug)
        vcov[free_undirected_idx, free_undirected_idx] <-
          tmp %*% gamma_big_aug %*% t(tmp)
      } else if (ext_cat_mg) {
        # categorical external instruments: build the stage-2 map and directed
        # derivative over the per-block augmented [thresholds, correlations]
        # layout; gamma_big and jac_k are already augmented (see above)
        ntot_aug <- sum(md_aug)
        delta1_aug <- matrix(0, ntot_aug, length(free_directed_idx))
        h2_aug <- matrix(0, length(free_undirected_idx), ntot_aug)
        for (b in seq_len(nblocks)) {
          model_cols_b <- moff_aug[b] + cat_keep_b[[b]]
          delta1_aug[model_cols_b, ] <-
            delta_list[[b]][, free_directed_idx, drop = FALSE]
          fu_b <- unique(lavpartable$free[lavpartable$op == "~~" &
            lavpartable$free > 0L & lavpartable$block == b])
          fu_b <- fu_b[fu_b %in% free_undirected_idx]
          if (length(fu_b) == 0L) {
            next
          }
          delta2_b <- delta_list[[b]][, fu_b, drop = FALSE]
          s_mat <- diag(lavmodel@nvar[[b]])
          tmp_b <- lav_utils_wls_linearization(delta = delta2_b, s = s_mat,
            meanstructure = lavmodel@meanstructure, categorical = TRUE,
            svec = vec_list[[b]], return_h = TRUE)
          h2_b <- attr(tmp_b, "H")
          nth_b <- md[b] - ncol(h2_b)
          h2_b <- cbind(matrix(0, nrow(h2_b), nth_b), h2_b)
          h2_aug[match(fu_b, free_undirected_idx), model_cols_b] <- h2_b
        }
        tmp <- h2_aug %*% (diag(ntot_aug) - delta1_aug %*% jac_k)
        vcov[free_undirected_idx, free_undirected_idx] <-
          tmp %*% gamma_big %*% t(tmp)
      } else {
        if (length(free_directed_idx) > 0L) {
          tmp <- h2_full %*% (diag(ntot_moment) - delta1_full %*% jac_k)
        } else {
          tmp <- h2_full
        }
        vcov[free_undirected_idx, free_undirected_idx] <-
          tmp %*% gamma_big %*% t(tmp)
      }
    }

    # thresholds (categorical): the threshold parameters equal the sample
    # thresholds, so their covariance is the corresponding block of the moment
    # covariance, delta_th' gamma delta_th
    if (lavmodel@categorical && length(free_th_idx) > 0L) {
      delta_list_th <- lav_sem_miiv_delta(lavmodel)
      delta_full <- do.call("rbind", delta_list_th)
      delta_th <- delta_full[, free_th_idx, drop = FALSE]
      gamma_th <- gamma_big
      if (ext_cat_mg) {
        # map the per-block model threshold derivative into the augmented layout
        delta_th <- matrix(0, sum(md_aug), length(free_th_idx))
        for (b in seq_len(nblocks)) {
          cols <- moff[b] + seq_len(md[b])
          delta_th[moff_aug[b] + cat_keep_b[[b]], ] <-
            delta_list_th[[b]][, free_th_idx, drop = FALSE]
        }
        gamma_th <- gamma_big
      }
      zero_idx <- which(rowSums(abs(delta_th)) == 0)
      if (length(zero_idx) > 0L) {
        delta_th <- delta_th[-zero_idx, , drop = FALSE]
        gamma_th <- gamma_th[-zero_idx, -zero_idx, drop = FALSE]
      }
      vcov[free_th_idx, free_th_idx] <-
        t(delta_th) %*% gamma_th %*% delta_th
    }

    # ceq.simple: expand the compact vcov to the 'unco' space
    if (lavmodel@ceq.simple.only && nrow(lavmodel@ceq.simple.K) > 0L) {
      vcov <- lavmodel@ceq.simple.K %*% vcov %*% t(lavmodel@ceq.simple.K)
    }
    return(vcov)
  }

  # missing data: two-stage standard errors. Under missing = "two.stage" /
  # "robust.two.stage" the sample moments are the (saturated) EM estimates;
  # their asymptotic covariance ('gamma') is no longer the complete-data NT
  # gamma but the inverse saturated information (two.stage) or its sandwich
  # (robust.two.stage). We build that 'gamma_big' explicitly (in the same
  # (mean, vech(cov)) ordering as jac_k) and route the SEs through the same
  # explicit jac %*% gamma %*% t(jac) path used for categorical data.
  two_stage <- any(lavoptions$missing == c("two.stage", "robust.two.stage"))
  use_explicit_gamma <- lavmodel@categorical || two_stage

  # do we need gamma_big?
  if (iv_vcov_stage1 == "gamma" ||
     (iv_vcov_stage2 != "none" && length(free_undirected_idx) > 0L)) {
      gamma_g <- vector("list", lavmodel@ngroups)
    for (g in seq_len(lavmodel@ngroups)) {
      fg <- lavsamplestats@nobs[[g]] / lavsamplestats@ntotal
      # (model-implied or h1) covariance, used by the continuous NT-gamma
      # path; needed whether or not NACOV is available
      if (iv_vcov_gamma_modelbased) {
        cov_g <- lavimplied$cov[[g]]
        # the model-implied covariance may be non-PD (eg a Heywood case); the
        # continuous NT-gamma path needs a PD matrix, so fall back to the
        # (PD) sample covariance in that case
        if (!lavmodel@categorical &&
            inherits(try(chol(cov_g), silent = TRUE), "try-error")) {
          cov_g <- lavh1$implied$cov[[g]]
        }
      } else {
        cov_g <- lavh1$implied$cov[[g]]
      }
      if (two_stage) {
        # two-stage moment covariance, in (mean, vech(cov)) order
        mu_g <- lavh1$implied$mean[[g]]
        sig_g <- lavh1$implied$cov[[g]]
        x_idx_g <-
          if (lavmodel@fixed.x) lavsamplestats@x.idx[[g]] else integer(0L)
        if (identical(tolower(lavoptions$missing), "robust.two.stage")) {
          # sandwich I_1^{-1} J_1 I_1^{-1} (Savalei & Falk, 2014)
          gg <- lav_mvn_mi_h1_omega_sw(
            y = lavdata@X[[g]], mp = lavdata@Mp[[g]],
            yp = lavsamplestats@missing[[g]],
            mu = mu_g, sigma_1 = sig_g, x_idx = x_idx_g,
            information = "observed"
          )
        } else {
          # I_1^{-1} (Savalei & Bentler, 2009)
          i1 <- lav_mvnorm_missing_information_observed_samplestats(
            yp = lavsamplestats@missing[[g]],
            mu = mu_g, sigma_1 = sig_g, x_idx = x_idx_g
          )
          gg <- lav_mat_sym_inverse(i1)
        }
        gamma_g[[g]] <- fg * gg
      } else if (!is.null(lavsamplestats@NACOV[[g]])) {
        gamma_g[[g]] <- fg * lavsamplestats@NACOV[[g]]
      }
    }
    if (two_stage || !is.null(lavsamplestats@NACOV[[g]])) {
      gamma_big <- lav_mat_bdiag(gamma_g)
    }
  }

  # compute jac_k if needed
  if (length(free_directed_idx) > 0L && iv_vcov_stage1 != "none"
      && (iv_vcov_stage1 == "gamma" || iv_vcov_stage2 == "h2")) {
    if (iv_vcov_jack_numerical) {
      # compute K matrix
      vec <- lav_implied_to_vec(
        implied = lavh1$implied, lavmodel = lavmodel,
        drop_list = TRUE
      )
      jac_k <- numDeriv::jacobian(
        lav_sem_miiv_2sls_samp,
        x = vec, samplestats = TRUE, eqs = eqs,
        lavmodel = lavmodel, lavpartable = lavpartable,
        lavsamplestats = lavsamplestats,
        lavh1 = lavh1, lavdata = lavdata,
        free_directed_idx = free_directed_idx
      )
    } else {
      # analytic version -- # FIXME: multiple blocks!!!
      stopifnot(lavmodel@nblocks == 1L)
      jac_k <- lav_sem_miiv_utils_jack_eqs(
        eqs = eqs, block = 1L,
        lavmodel = lavmodel, lavpartable = lavpartable,
        free_directed_idx = free_directed_idx
      )
    }
  }

  # external instruments with CATEGORICAL data (single group): replace the
  # model gamma (NACOV) and the directed Jacobian with their augmented
  # [thresholds, correlations] versions, so the instrument sampling variability
  # is folded into all standard errors (directed coefficients and the stage-2
  # parameters). The augmented directed Jacobian is assembled analytically from
  # k_mat_full (the numerical route reconstructs a Pearson covariance and is
  # not valid for categorical data; equality constraints among directed
  # coefficients are therefore not supported here).
  ext_cat <- lavmodel@categorical && lavmodel@nblocks == 1L &&
    length(unlist(lavdata@ov.names.aux)) > 0L &&
    length(free_directed_idx) > 0L && eqs_have_kmat_full
  cat_keep <- NULL
  if (ext_cat) {
    augc <- lav_sem_miiv_aug_cat_moments(
      lavdata = lavdata, b = 1L, model_cor = lavh1$implied$cov[[1]],
      model_th = lavsamplestats@th[[1]], nobs = lavsamplestats@nobs[[1]])
    # single-group explicit-gamma path uses fg * NACOV (fg = nobs/ntotal = 1),
    # i.e. the augmented NACOV on the same scale as lavsamplestats@NACOV
    gamma_big <- (lavsamplestats@nobs[[1]] / lavsamplestats@ntotal) * augc$nacov
    jac_k <- lav_sem_miiv_jack_eqs_aug(
      eqs = eqs, block = 1L, lavmodel = lavmodel,
      lavpartable = lavpartable, free_directed_idx = free_directed_idx)
    cat_keep <- lav_sem_miiv_aug_cat_keep_idx(
      lavmodel@nvar[[1]], ncol(augc$R), augc$th.idx)
  }

  # stage 1: directed effects
  if (length(free_directed_idx) > 0L && iv_vcov_stage1 != "none") {
    if (use_restricted_directed) {
      # variance-weighted restricted-2SLS covariance (matches MIIVsem)
      v_dir <- try(lav_sem_miiv_directed_vcov_restricted(
        eqs = eqs, lavpartable = lavpartable,
        free_directed_idx = free_directed_idx, nfree = lavmodel@nx.free),
        silent = TRUE)
      if (!inherits(v_dir, "try-error")) {
        vcov[free_directed_idx, free_directed_idx] <- v_dir
      }
    } else if (iv_vcov_stage1 == "gamma") {
      if (use_explicit_gamma) {
        # explicit moment covariance: categorical (ADF) or two-stage missing
        k_gamma_adf_kt <- jac_k %*% gamma_big %*% t(jac_k)
        vcov[free_directed_idx, free_directed_idx] <-
          k_gamma_adf_kt / lavsamplestats@ntotal
      } else {
        # vcov = K %*% Gamma_NT %*% t(K)
        # k_gammant_kt <- jac_k %*% gamma_big %*% t(jac_k)
        x_idx <-
          if (lavmodel@fixed.x) lavsamplestats@x.idx[[1]] else integer(0L)
        if (length(unlist(lavdata@ov.names.aux)) > 0L) {
          # external instruments: the 2SLS slopes depend on the instrument
          # moments (and not on the model moments), so the model-moment
          # Jacobian would be ~0. Use the full (augmented) directed Jacobian
          # and the augmented covariance instead.
          jac_full <- lav_sem_miiv_jac_full_numeric(
            eqs = eqs, lavmodel = lavmodel, lavpartable = lavpartable,
            lavdata = lavdata, lavsamplestats = lavsamplestats,
            lavh1 = lavh1, free_directed_idx = free_directed_idx)
          cov_aug <- lav_sem_miiv_aug_moments(
            lavdata = lavdata, b = 1L, ov_names = lavdata@ov.names[[1]],
            sample_cov = lavh1$implied$cov[[1]],
            sample_mean = if (lavmodel@meanstructure) {
              lavh1$implied$mean[[1]]
            } else {
              NULL
            })$cov
          k_gammant_kt <- lav_mat_k_gammant_kt(m_k = jac_full, s = cov_aug,
            meanstructure = lavmodel@meanstructure, x_idx = x_idx)
        } else {
          k_gammant_kt <- lav_mat_k_gammant_kt(m_k = jac_k, s = cov_g,
            meanstructure = lavmodel@meanstructure, x_idx = x_idx)
        }
        vcov[free_directed_idx, free_directed_idx] <-
          k_gammant_kt / lavsamplestats@ntotal
      }

    # iv_vcov_stage1 = lm.vcov or lm.vcov.dfres
    } else {
      for (b in seq_len(nblocks)) {
        neqs <- length(eqs[[b]])
        for (j in seq_len(neqs)) {
          eq <- eqs[[b]][[j]]

          eq_vcov <- NULL
          if (!is.null(eq$vcov) && nrow(eq$vcov) > 0L) {
            eq_vcov <- eq$vcov # always includes intercept
          } else if (!is.null(eq$XX) && !is.null(eq$resvar)) {
            # reconstruct using XX and resvar (less stable!)
            eq_vcov <- solve(crossprod(eq$XX)) * eq$resvar # or resvar_df_res
          }
          if (is.null(eq_vcov)) {
            next
          }

          # only keep free parameters
          free_idx <- lavpartable$free[eq$pt]
          if (length(eq$rhs) > 0L) {
            if (all(free_idx > 0L)) {
              vcov[free_idx, free_idx] <- eq_vcov[-1, -1]
            } else {
              # remove non-free elements
              zero_idx_1 <- which(free_idx == 0L)
              free_idx <- free_idx[-zero_idx_1]
              if (length(free_idx) > 0L) {
                vcov[free_idx, free_idx] <-
                  eq_vcov[-1, -1][-zero_idx_1, -zero_idx_1, drop = FALSE]
              }
            }
          }
          if (lavmodel@meanstructure) {
            free_int_idx <- lavpartable$free[eq$ptint]
            if (free_int_idx > 0L) {
              vcov[free_int_idx, free_int_idx] <- eq_vcov[1, 1]
            }
          }
        } # neqs
      } # nblocks
    } # continuous/vcov version
  } # stage 1

  # stage 2: undirected effects (note: intercepts have no effect!)

  if (iv_vcov_stage2 != "none" && length(free_undirected_idx) > 0L &&
    iv_varcov_method != "NONE") {
    if (iv_vcov_stage2 == "h2") {
      # delta_list <- lav_model_delta(lavmodel) # per block/group

      # # block-wise
      # delta1_block <- vector("list", length = nblocks)
      # delta2_block <- vector("list", length = nblocks)
      # w2_block <- vector("list", length = nblocks)
      # for (b in seq_len(nblocks)) {
      #   nvar <- lavmodel@nvar[b]
      #   # delta1 + delta2
      #   delta1_block[[b]] <- delta_list[[b]][, free_directed_idx,
      #                                                  drop = FALSE]
      #   delta2_block[[b]] <-
      #     delta_list[[b]][, free_undirected_idx, drop = FALSE]
      #   # w2
      #   if (lavmodel@categorical || iv_varcov_method == "ULS") {
      #     s.inv <- diag(1, nrow = nvar)
      #   } else if (iv_varcov_method == "GLS") {
      #     s.inv <- lavh1$implied$cov[[b]]
      #   } else {
      #     s.inv <- lavimplied$cov[[b]]
      #   }
      #   w2_22 <- 0.5 * lav_mat_dup_pre_post(s.inv %x% s.inv)
      #   if (lavmodel@meanstructure) {
      #     w2_block[[b]] <- lav_mat_bdiag(s.inv, w2_22)
      #   } else {
      #     w2_block[[b]] <- w2_22
      #   }
      # } # blocks
      # delta1 <- do.call("rbind", delta1_block)
      # delta2 <- do.call("rbind", delta2_block)
      # w2 <- lav_mat_bdiag(w2_block)
      # h2 <- solve(t(delta2) %*% w2 %*% delta2, t(delta2) %*% w2)

      # tmp <- lav_sem_miiv_varcov(
      #   x = NULL, lavmodel = lavmodel,
      #   lavpartable = lavpartable, lavsamplestats = lavsamplestats,
      #   lavh1 = lavh1, free_directed_idx = free_directed_idx,
      #   free_undirected_idx = free_undirected_idx, add_h2 = TRUE,
      #   iv_varcov_method = iv_varcov_method
      # )
      # h2 <- attr(tmp, "H")
      stopifnot(lavmodel@nblocks == 1L)
      delta <- lav_sem_miiv_delta(lavmodel)[[1L]]
      delta2 <- delta[, free_undirected_idx, drop = FALSE]
      svec <- lav_implied_to_vec(
          implied = lavh1$implied, lavmodel = lavmodel,
          drop_list = TRUE
      )
      if (lavmodel@categorical || iv_varcov_method == "ULS") {
        s_mat <- diag(lavmodel@nvar[[1]])
      } else if (iv_varcov_method == "GLS") {
        s_mat <- lavh1$implied$cov[[1]]
      } else {
        # 2RLS/RLS use the model-implied covariance, which may be non-PD (eg a
        # Heywood case); fall back to the (PD) sample covariance if so
        s_mat <- lavimplied$cov[[1]]
        if (!lavmodel@categorical &&
            inherits(try(chol(s_mat), silent = TRUE), "try-error")) {
          s_mat <- lavh1$implied$cov[[1]]
        }
      }
      tmp <- lav_utils_wls_linearization(delta = delta2, s = s_mat,
        meanstructure = lavmodel@meanstructure,
        categorical = lavmodel@categorical, svec = svec, return_h = TRUE)
      h2 <- attr(tmp, "H")
      if (length(free_directed_idx) > 0L) {
        # assuming a SINGLE block for now
        delta1 <- delta[, free_directed_idx, drop = FALSE]
        nth <- nrow(delta1) - ncol(h2)
        h2 <- cbind(matrix(0, nrow(h2), ncol = nth), h2)
        if (ext_cat) {
          # place the (model-moment) stage-2 map and directed derivative into
          # the augmented [thresholds, correlations] layout, then propagate the
          # directed uncertainty through the augmented Jacobian
          m_aug <- ncol(jac_k)
          h2_aug <- matrix(0, nrow(h2), m_aug)
          h2_aug[, cat_keep] <- h2
          d1_aug <- matrix(0, m_aug, ncol(delta1))
          d1_aug[cat_keep, ] <- delta1
          tmp <- h2_aug %*% (diag(m_aug) - d1_aug %*% jac_k)
        } else {
          idiag <- diag(1, nrow = nrow(delta1))
          tmp <- h2 %*% (idiag - delta1 %*% jac_k)
        }
      } else {
        tmp <- h2
      }
      if (use_explicit_gamma) {
        tmp_gamma_tmpt <- tmp %*% gamma_big %*% t(tmp)
        vcov[free_undirected_idx, free_undirected_idx] <-
          tmp_gamma_tmpt / lavsamplestats@ntotal
        # add thresholds (categorical only)
        nth <- if (lavmodel@categorical) length(free_th_idx) else 0L
        if (length(nth) > 0L && nth > 0L) {
          delta_th <- delta[, free_th_idx, drop = FALSE]
          if (ext_cat) {
            # map the model threshold derivative into the augmented layout
            dta <- matrix(0, ncol(jac_k), ncol(delta_th))
            dta[cat_keep, ] <- delta_th
            delta_th <- dta
            gamma_th <- gamma_big
          } else {
            gamma_th <- gamma_big
          }
          # FIXME: remove zero rows from delta
          zero_idx <- which(rowSums(abs(delta_th)) == 0)
          if (length(zero_idx) > 0L) {
            delta_th <- delta_th[-zero_idx, , drop = FALSE]
            gamma_th <- gamma_th[-zero_idx, -zero_idx, drop = FALSE]
          }
          th_gamma_tht <- t(delta_th) %*% gamma_th %*% delta_th
          vcov[free_th_idx, free_th_idx] <-
            th_gamma_tht / lavsamplestats@ntotal
        }
      } else {
        x_idx <-
          if (lavmodel@fixed.x) lavsamplestats@x.idx[[1]] else integer(0L)
        # external instruments: fold the instrument sampling variance into the
        # variance/covariance parameter standard errors. The directed
        # coefficients depend on the instrument moments, so we propagate the
        # uncertainty through the FULL (augmented-moment) directed Jacobian and
        # sandwich with the normal-theory gamma of the augmented covariance,
        # rather than treating the instrument moments as fixed.
        jac_full <- NULL
        if (length(unlist(lavdata@ov.names.aux)) > 0L &&
            length(free_directed_idx) > 0L) {
          if (has_dir_con || !eqs_have_kmat_full) {
            # use the numerical augmented Jacobian when the analytic k_mat_full
            # is unavailable (raw-data engine) or when there are equality
            # constraints among the directed coefficients (the per-equation
            # analytic Jacobian cannot represent the cross-equation pooling)
            jac_full <- lav_sem_miiv_jac_full_numeric(
              eqs = eqs, lavmodel = lavmodel, lavpartable = lavpartable,
              lavdata = lavdata, lavsamplestats = lavsamplestats,
              lavh1 = lavh1, free_directed_idx = free_directed_idx)
          } else {
            jac_full <- lav_sem_miiv_jack_eqs_aug(
              eqs = eqs, block = 1L, lavmodel = lavmodel,
              lavpartable = lavpartable, free_directed_idx = free_directed_idx)
          }
        }
        if (!is.null(jac_full)) {
          # the instrument moments have no model-implied counterpart, so for a
          # coherent augmented covariance (needed by the normal-theory gamma)
          # we use the saturated (H1) sample covariance for the whole matrix
          aug <- lav_sem_miiv_aug_moments(
            lavdata = lavdata, b = 1L, ov_names = lavdata@ov.names[[1]],
            sample_cov = lavh1$implied$cov[[1]],
            sample_mean = if (lavmodel@meanstructure) {
              lavh1$implied$mean[[1]]
            } else {
              NULL
            })
          cov_aug <- aug$cov
          p_model <- lavmodel@nvar[[1]]
          p_aug <- ncol(cov_aug)
          keep <- lav_sem_miiv_aug_keep_idx(p_model, p_aug,
            lavmodel@meanstructure)
          m_aug <- ncol(jac_full)
          h2_aug <- matrix(0, nrow(h2), m_aug)
          h2_aug[, keep] <- h2
          d1_aug <- matrix(0, m_aug, ncol(delta1))
          d1_aug[keep, ] <- delta1
          tmp_aug <- h2_aug %*% (diag(m_aug) - d1_aug %*% jac_full)
          tmp_gammant_tmpt <- lav_mat_k_gammant_kt(m_k = tmp_aug, s = cov_aug,
            meanstructure = lavmodel@meanstructure, x_idx = x_idx)
        } else {
          tmp_gammant_tmpt <- lav_mat_k_gammant_kt(m_k = tmp, s = cov_g,
            meanstructure = lavmodel@meanstructure, x_idx = x_idx)
        }
        #tmp_gammant_tmpt_bis <- tmp %*% gamma_big %*% t(tmp)
        vcov[free_undirected_idx, free_undirected_idx] <-
          tmp_gammant_tmpt / lavsamplestats@ntotal
      }

      # iv_vcov_stage2 == "delta"
    } else {
      # part a: effect of theta1
      if (!iv_vcov_jaca_numerical) {
        # use analytic expressions
        if (iv_varcov_method %in% c("ULS", "GLS")) {
          jac_a <- lav_sem_miiv_utils_jaca_uls_gls(
            lavmodel = lavmodel,
            lavpartable = lavpartable, lavh1 = lavh1,
            free_directed_idx = free_directed_noint_idx,
            free_undirected_idx = free_undirected_idx,
            iv_varcov_method = iv_varcov_method
          )
        } else if (iv_varcov_method == "2RLS") {
          jac_a <- lav_sem_miiv_utils_jaca_2rls(
            lavmodel = lavmodel,
            lavpartable = lavpartable, lavh1 = lavh1,
            free_directed_idx = free_directed_noint_idx,
            free_undirected_idx = free_undirected_idx
          )
        } else if (iv_varcov_method == "RLS") {
          jac_a <- lav_sem_miiv_utils_jaca_rls(
            lavmodel = lavmodel,
            lavpartable = lavpartable, lavh1 = lavh1,
            free_directed_idx = free_directed_noint_idx,
            free_undirected_idx = free_undirected_idx
          )
        }
        # for checking only: numerical approximation
      } else {
        jac_a <- lav_func_jacobian_complex(
          func = lav_sem_miiv_varcov,
          x = theta1_noint, samplestats = FALSE, lavmodel = lavmodel,
          lavpartable = lavpartable, lavsamplestats = lavsamplestats,
          lavh1 = lavh1, free_directed_idx = free_directed_noint_idx,
          free_undirected_idx = free_undirected_idx,
          iv_varcov_method = iv_varcov_method
        )

        # jac_a <- lav_func_jacobian_complex(
        #   func = lav_sem_miiv_varcov_old,
        #   x = theta1_noint, samplestats = FALSE, lavmodel = lavmodel,
        #   lavpartable = lavpartable, lavsamplestats = lavsamplestats,
        #   lavh1 = lavh1, free_directed_idx = free_directed_noint_idx,
        #   free_undirected_idx = free_undirected_idx,
        #   iv_varcov_method = iv_varcov_method
        # )
      }

      # compute vcov_a (due to theta1)
      vcov_directed <-
        vcov[free_directed_noint_idx, free_directed_noint_idx, drop = FALSE]
      vcov_a <- jac_a %*% vcov_directed %*% t(jac_a)

      # part b: effect of sample statistics
      # - get jacobian lav_sem_miiv_varcov wrt samplestats_vec
      # - get Gamma (NT or ADF)
      # - vcov_undirected_b <- jac_b %*% (1/N * gamma_mat) %*% t(jac_b)

      if (!iv_vcov_jaca_numerical) {
        delta_block <- lav_sem_miiv_delta(lavmodel = lavmodel)
        jac_b_block <- vector("list", length = nblocks)
        for (b in seq_len(nblocks)) {
          sample_cov <- lavh1$implied$cov[[b]]
          this_delta2 <- delta_block[[b]][, free_undirected_idx, drop = FALSE]
          if (iv_varcov_method == "ULS") {
            jac_b_block[[b]] <-
              lav_sem_miiv_utils_jacb_uls(sample_cov, delta2 = this_delta2)
          } else if (iv_varcov_method == "GLS") {
            jac_b_block[[b]] <-
              lav_sem_miiv_utils_jacb_gls(sample_cov, delta2 = this_delta2,
                                          theta2 = theta2)
          } else if (iv_varcov_method == "2RLS") {
            jac_b_block[[b]] <-
              lav_sem_miiv_utils_jacb_2rls(sample_cov, delta2 = this_delta2,
                                           theta2 = theta2)
          } else if (iv_varcov_method == "RLS") {
            jac_b_block[[b]] <-
              lav_sem_miiv_utils_jacb_rls(sample_cov, delta2 = this_delta2,
                                          theta2 = theta2)
          }
        } # blocks
        jac_b <- lav_mat_bdiag(jac_b_block)

      # for checking only: numerical approximation (very slow!)
      } else {
        vec <- lav_implied_to_vec(
          implied = lavh1$implied, lavmodel = lavmodel,
          drop_list = TRUE
        )

        jac_b <- lav_func_jacobian_complex(
          func = lav_sem_miiv_varcov,
          x = vec, samplestats = TRUE, lavmodel = lavmodel,
          lavpartable = lavpartable, lavsamplestats = lavsamplestats,
          lavh1 = lavh1, free_directed_idx = free_directed_noint_idx,
          free_undirected_idx = free_undirected_idx,
          iv_varcov_method = iv_varcov_method
        )

        # jac_b <- lav_func_jacobian_complex(
        #   func = lav_sem_miiv_varcov_old,
        #   x = vec, samplestats = TRUE, lavmodel = lavmodel,
        #   lavpartable = lavpartable, lavsamplestats = lavsamplestats,
        #   lavh1 = lavh1, free_directed_idx = free_directed_noint_idx,
        #   free_undirected_idx = free_undirected_idx,
        #   iv_varcov_method = iv_varcov_method
        # )
      }
      if (use_explicit_gamma) {
        jacb_gamma_jacbt <- jac_b %*% gamma_big %*% t(jac_b)
      } else {
        x_idx <-
          if (lavmodel@fixed.x) lavsamplestats@x.idx[[1]] else integer(0L)
        jacb_gamma_jacbt <- lav_mat_k_gammant_kt(m_k = jac_b, s = cov_g,
          meanstructure = lavmodel@meanstructure, x_idx = x_idx)
      }
      vcov_b <- jacb_gamma_jacbt / lavsamplestats@ntotal
      vcov_ab <- vcov_a + vcov_b
      vcov[free_undirected_idx, free_undirected_idx] <- vcov_ab
    } # continuous
  } # iv_vcov_stage2 != "none"

  # the rest of lavaan expects the vcov of a ceq.simple.only model to be in the
  # 'unco' space (one row/column per non-collapsed parameter, nx.unco), eg in
  # lav_model_vcov_se(). The vcov above is built in the compact (nx.free)
  # space, so expand it via the ceq.simple.K mapping: constrained parameters
  # then share their (co)variance, and the SEs map to the parameter table
  # correctly (without this, the constrained parameters get inconsistent SEs
  # and the trailing parameters get NA).
  if (lavmodel@ceq.simple.only && nrow(lavmodel@ceq.simple.K) > 0L) {
    vcov <- lavmodel@ceq.simple.K %*% vcov %*% t(lavmodel@ceq.simple.K)
  }

  vcov
}

# small functions to do something useful with the common
# plot commands

# suggested by JEB
lav_lavaan_pairs <- function(x, group = 1L, ...) {
  m_x <- x@Data@X[[group]]
  colnames(m_x) <- x@Data@ov.names[[group]]
  pairs(m_x, ...)
}

# inspect a fitted lavaan object

# backward compatibility -- wrapper around lavInspect
lav_lavaan_inspect <- function(object, what = "free", ...) {
  dotdotdot <- list(...)
  if (length(dotdotdot) > 0L) {
    for (j in seq_along(dotdotdot)) {
      lav_msg_warn(gettextf(
        "Unknown argument %s for %s", sQuote(names(dotdotdot)[j]),
        sQuote("inspect"))
      )
    }
  }
  lav_lavaan_lavinspect(object              = object,
    what                   = what,
    add_labels             = TRUE,
    add_class              = TRUE,
    drop_list_single_group = TRUE)
}

# the `tech' version: no labels, full matrices, ... for further processing
lav_lavaan_lavtech <- function(object,
                           what                   = "free",
                           add_labels             = FALSE,
                           add_class              = FALSE,
                           list_by_group          = FALSE,
                           drop_list_single_group = FALSE,
                           ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  lav_lavaan_lavinspect(object, what = what,
    add_labels = add_labels, add_class = add_class,
    list_by_group =  list_by_group,
    drop_list_single_group =  drop_list_single_group)
}

# the `user' version: with defaults for display only
lav_lavaan_lavinspect <- function(object,
                              what                   = "free",
                              add_labels             = TRUE,
                              add_class              = TRUE,
                              list_by_group          = TRUE,
                              drop_list_single_group = TRUE,
                              ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # object must inherit from class lavaan
  stopifnot(inherits(object, "lavaan"))

  # check object
  object <- lav_object_check_version(object)

  # store partable with pta in object to use cache in called functions
  object@ParTable <- lav_pt_set_cache(object@ParTable, object@pta)

  # only a single argument
  if (length(what) > 1) {
    lav_msg_stop(gettextf("argument %s cannot have more than one element",
                          "what"))
  }

  # be case insensitive
  what <- tolower(what)

  #### model matrices, with different contents ####
  if (what == "free") {
    lav_inspect_modelmatrices(object, what = "free",
      type = "free", add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "impute" ||
    what == "imputed") { # just to ease the transition for semTools!
    object@imputed
  } else if (what == "partable" || what == "user") {
    lav_inspect_modelmatrices(object, what = "free",
      type = "partable", add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "se" ||
    what == "std.err" ||
    what == "standard.errors") {
    lav_inspect_modelmatrices(object, what = "se",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "se.std" ||
    what == "std.se") {
    lav_inspect_modelmatrices(object, what = "std.se",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "start" || what == "starting.values") {
    lav_inspect_modelmatrices(object, what = "start",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "est"  || what == "estimates" ||
    what == "x") {
    lav_inspect_modelmatrices(object, what = "est",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "est.unrotated") {
    lav_inspect_modelmatrices(object, what = "est.unrotated",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mm.lambda" || what == "lambda") {
    lav_inspect_mm(object, mm_name = "lambda",
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mm.delta") {
    lav_inspect_mm(object, mm_name = "delta",
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "dx.free") {
    lav_inspect_modelmatrices(object, what = "dx.free",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "dx.all") {
    lav_inspect_modelmatrices(object, what = "dx.all",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "std" || what == "std.all" ||
    what == "est.std" || what == "std.est" ||
    what == "standardized") {
    lav_inspect_modelmatrices(object, what = "std.all",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "std.lv") {
    lav_inspect_modelmatrices(object, what = "std.lv",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "std.nox") {
    lav_inspect_modelmatrices(object, what = "std.nox",
      add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)


    #### parameter table ####
  } else if (what == "list") {
    parTable(object)

    #### bootstrap coef ####
  } else if (what %in% c("boot", "bootstrap", "boot.coef", "coef.boot")) {
    lav_inspect_boot(object, add_labels = add_labels,
      add_class = add_class)

    #### Monte Carlo coef draws ####
  } else if (what %in% c("monte.carlo", "mc", "mc.coef", "coef.mc")) {
    lav_inspect_mc(object, add_labels = add_labels,
      add_class = add_class)

    #### fit indices ####
  } else if (what == "fit" ||
    what == "fitmeasures" ||
    what == "fit.measures" ||
    what == "fit.indices") {
    fitMeasures(object)

    #### baseline model ####
  } else if (what == "baseline.partable") {
    out <- as.data.frame(object@baseline$partable, stringsAsFactors = FALSE)
    if (add_class) {
      class(out) <- c("lavaan.data.frame", "data.frame")
    }
    out
  } else if (what == "baseline.test") {
    object@baseline$test

    #### modification indices ####
  } else if (what == "mi" ||
    what == "modindices" ||
    what == "modification.indices") {
    modificationIndices(object)


    #### sample statistics #####
  } else if (what == "obs" ||
    what == "observed" ||
    what == "sampstat" ||
    what == "sampstats" ||
    what == "samplestats" ||
    what == "samp" ||
    what == "sample" ||
    what == "samplestatistics") {
    # new in 0.6-3: always use h1 = TRUE!!!
    lav_inspect_sampstat(object, h1 = TRUE, std = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "obs.std" ||
    what == "observed.std" ||
    what == "sampstat.std" ||
    what == "sampstats.std" ||
    what == "samplestats.std" ||
    what == "samp.std" ||
    what == "sample.std" ||
    what == "samplestatistics.std") {
    lav_inspect_sampstat(object, h1 = TRUE, std = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "h1" || what == "missing.h1" || what == "sampstat.h1") {
    lav_inspect_sampstat(object, h1 = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    #### wls.est - wls.obs - wls.v ####
  } else if (what == "wls.est") {
    lav_inspect_wls_est(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "wls.obs") {
    lav_inspect_wls_obs(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "wls.v") {
    lav_inspect_wls_v(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)



    #### data + missingness ####
  } else if (what == "data") {
    lav_inspect_data(object, add_labels = add_labels,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "case.idx") {
    lav_inspect_case_idx(object,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "ngroups") {
    object@Data@ngroups
  } else if (what == "group") {
    object@Data@group
  } else if (what == "cluster") {
    object@Data@cluster
  } else if (what == "nlevels") {
    object@Data@nlevels
  } else if (what == "nclusters") {
    lav_inspect_cl_info(object, level = 2L,
      what = "nclusters",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "ncluster.size") {
    lav_inspect_cl_info(object, level = 2L,
      what = "ncluster.size",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cluster.size") {
    lav_inspect_cl_info(object, level = 2L,
      what = "cluster.size",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cluster.id") {
    lav_inspect_cl_info(object, level = 2L,
      what = "cluster.id",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cluster.idx") {
    lav_inspect_cl_info(object, level = 2L,
      what = "cluster.idx",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cluster.label") {
    lav_inspect_cl_info(object, level = 2L,
      what = "cluster.label",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cluster.sizes") {
    lav_inspect_cl_info(object, level = 2L,
      what = "cluster.sizes",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "average.cluster.size") {
    lav_inspect_cl_info(object, level = 2L,
      what = "average.cluster.size",
      drop_list_single_group = drop_list_single_group)
  } else if (what == "ordered") {
    object@Data@ordered
  } else if (what == "group.label") {
    object@Data@group.label
  } else if (what == "level.label") {
    object@Data@level.label
  } else if (what == "nobs") {
    # use @SampleStats@nobs (not @Data@nobs): if sampling weights are
    # used, the effective sample size is sum(wt), and this is what the
    # nobs() generic and all fit computations rely on (see also "ntotal")
    unlist(object@SampleStats@nobs)
  } else if (what == "norig") {
    unlist(object@Data@norig)
  } else if (what == "ntotal") {
    # effective total sample size; equals sum(wt) when sampling weights
    # are used (consistent with the nobs() method)
    object@SampleStats@ntotal
  } else if (what == "coverage") {
    lav_inspect_mi_coverage(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("patterns", "pattern")) {
    lav_inspect_mi_patterns(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "empty.idx") {
    lav_inspect_empty_idx(object,
      drop_list_single_group = drop_list_single_group)


    #### rsquare ####
  } else if (what == "rsquare" || what == "r-square" || what == "r2") {
    lav_inspect_rsquare(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)


    #### model-implied sample statistics ####
  } else if (what == "implied" || what == "fitted" ||
    what == "expected" || what == "exp") {
    lav_inspect_implied(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "resid" || what == "res" || what == "residual" ||
    what == "residuals") {
    lav_inspect_residuals(object, h1 = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cov.lv" || what == "veta") {
    lav_inspect_cov_lv(object,
      correlation_metric = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cor.lv") {
    lav_inspect_cov_lv(object,
      correlation_metric = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mean.lv" || what == "eeta") {
    lav_inspect_mean_lv(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cov.all") {
    lav_inspect_cov_all(object,
      correlation_metric = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cor.all") {
    lav_inspect_cov_all(object,
      correlation_metric = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cov.ov" || what == "sigma" || what == "sigma.hat") {
    lav_inspect_cov_ov(object,
      correlation_metric = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "cor.ov") {
    lav_inspect_cov_ov(object,
      correlation_metric = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mean.ov" || what == "mu" || what == "mu.hat") {
    lav_inspect_mean_ov(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "th" || what == "thresholds") {
    lav_inspect_th(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "th.idx") {
    lav_inspect_th_idx(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "vy") {
    lav_inspect_vy(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("fs.reliability", "fs.rel", "fs.reliabilities")) {
    lav_inspect_fs_determinacy(object, squared = TRUE,
      fs_method = "regression",
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("fs_determinacy", "fs.det", "fs.determin",
                         "fs.determinacies")) {
    lav_inspect_fs_determinacy(object, squared = FALSE,
      fs_method = "regression",
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("fs_reliability_bartlett",
                         "fs_reliability_Bartlett",
                         "fs.reliability.Bartlett",
                         "fs.reliability.bartlett",
                         "fs.rel.bartlett", "fs.rel.Bartlett",
                         "fs.reliabilities.bartlett",
                         "fs.reliabilities.Bartlett")) {
    lav_inspect_fs_determinacy(object, squared = TRUE,
      fs_method = "Bartlett",
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("fs_determinacy_bartlett",
                         "fs_determinacy_Bartlett",
                         "fs.determinacy.Bartlett",
                         "fs.determinacy.bartlett",
                         "fs.det.bartlett", "fs.det.Bartlett",
                         "fs.determin.bartlett", "fs.determin.Bartlett",
                         "fs.determinacies.bartlett",
                         "fs.determinacies.Bartlett")) {
    lav_inspect_fs_determinacy(object, squared = FALSE,
      fs_method = "Bartlett",
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)



    #### specific model matrices? ####
  } else if (what == "mm.theta" || what == "theta" || what == "theta.cov") {
    lav_inspect_theta(object,  correlation_metric = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "theta.cor") {
    lav_inspect_theta(object,  correlation_metric = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    #### (squared) Mahalanobis distances ####
  } else if (what == "mdist2.fs") {
    lav_inspect_mdist2(object, type = "lv", squared = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mdist2.resid") {
    lav_inspect_mdist2(object, type = "resid", squared = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mdist.fs") {
    lav_inspect_mdist2(object, type = "lv", squared = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "mdist.resid") {
    lav_inspect_mdist2(object, type = "resid", squared = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    #### convergence, meanstructure, categorical ####
  } else if (what == "converged") {
    object@optim$converged
  } else if (what == "iterations" ||
    what == "iter" ||
    what == "niter") {
    object@optim$iterations
  } else if (what == "meanstructure") {
    object@Model@meanstructure
  } else if (what == "categorical") {
    object@Model@categorical
  } else if (what == "fixed.x") {
    object@Model@fixed.x
  } else if (what == "parameterization") {
    object@Model@parameterization
  } else if (what == "npar") {
    lav_inspect_npar(object, ceq = FALSE) # ignore equality constraints
  } else if (what == "coef") {
    # this breaks simsem and semTools -- 0.6-1
    # lav_inspect_coef(object, type = "free",
    #                        add_labels = add_labels, add_class = add_class)
    lav_inspect_modelmatrices(object, what = "est",
      type = "free", add_labels = add_labels, add_class = add_class,
      list_by_group =  list_by_group,
      drop_list_single_group = drop_list_single_group)


    #### NACOV samplestats ####
  } else if (what == "gamma" || what == "sampstat.nacov") {
    lav_inspect_sampstat_gamma(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)


    #### gradient, Hessian, information, first.order, vcov ####
  } else if (what == "gradient") {
    lav_inspect_grad(object,
      add_labels = add_labels, add_class = add_class, logl = FALSE)
  } else if (what == "gradient.logl") {
    lav_inspect_grad(object,
      add_labels = add_labels, add_class = add_class, logl = TRUE)
  } else if (what == "optim.gradient") {
    lav_inspect_grad(object,
      add_labels = add_labels, add_class = add_class, optim = TRUE)
  } else if (what == "hessian") {
    lav_inspect_hessian(object,
      add_labels = add_labels, add_class = add_class)

  } else if (what == "information") {
    lav_inspect_info(object, information = "default",
      augmented = FALSE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "information.expected") {
    lav_inspect_info(object, information = "expected",
      augmented = FALSE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "information.observed") {
    lav_inspect_info(object, information = "observed",
      augmented = FALSE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "information.first.order" ||
    what == "information.firstorder"  ||
    what == "first.order") {
    lav_inspect_info(object, information = "first.order",
      augmented = FALSE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)

  } else if (what == "augmented.information") {
    lav_inspect_info(object, information = "default",
      augmented = TRUE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "augmented.information.expected") {
    lav_inspect_info(object, information = "expected",
      augmented = TRUE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "augmented.information.observed") {
    lav_inspect_info(object, information = "observed",
      augmented = TRUE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "augmented.information.first.order" ||
    what == "augmented.first.order") {
    lav_inspect_info(object, information = "first.order",
      augmented = TRUE, inverted = FALSE,
      add_labels = add_labels, add_class = add_class)

  } else if (what == "inverted.information") {
    lav_inspect_info(object, information = "default",
      augmented = TRUE, inverted = TRUE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "inverted.information.expected") {
    lav_inspect_info(object, information = "expected",
      augmented = TRUE, inverted = TRUE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "inverted.information.observed") {
    lav_inspect_info(object, information = "observed",
      augmented = TRUE, inverted = TRUE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "inverted.information.first.order" ||
    what == "inverted.first.order") {
    lav_inspect_info(object, information = "first.order",
      augmented = TRUE, inverted = TRUE,
      add_labels = add_labels, add_class = add_class)

  } else if (what == "h1.information") {
    lav_inspect_h1_info(object, information = "default",
      h1_information = "default", inverted = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "h1.information.expected") {
    lav_inspect_h1_info(object, information = "expected",
      h1_information = "default", inverted = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "h1.information.observed") {
    lav_inspect_h1_info(object, information = "observed",
      h1_information = "default", inverted = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "h1.information.first.order" ||
    what == "h1.information.firstorder"  ||
    what == "h1.first.order") {
    lav_inspect_h1_info(object,
      information = "first.order", h1_information = "default",
      inverted = FALSE, add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

  } else if (what == "vcov") {
    lav_inspect_vcov(object,
      standardized = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.std.all" || what == "vcov.standardized" ||
    what == "vcov.std") {
    lav_inspect_vcov(object,
      standardized = TRUE, type = "std.all",
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.std.lv") {
    lav_inspect_vcov(object,
      standardized = TRUE, type = "std.lv",
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.std.nox") {
    lav_inspect_vcov(object,
      standardized = TRUE, type = "std.nox",
      add_labels = add_labels, add_class = add_class)

  } else if (what == "vcov.def") {
    lav_inspect_vcov_def(object, joint = FALSE,
      standardized = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.def.std.all" || what == "vcov.def.standardized" ||
    what == "vcov.def.std") {
    lav_inspect_vcov_def(object, joint = FALSE,
      standardized = TRUE, type = "std.all",
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.def.std.lv") {
    lav_inspect_vcov_def(object, joint = FALSE,
      standardized = TRUE, type = "std.lv",
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.def.std.nox") {
    lav_inspect_vcov_def(object, joint = FALSE,
      standardized = TRUE, type = "std.nox",
      add_labels = add_labels, add_class = add_class)

  } else if (what == "vcov.def.joint") {
    lav_inspect_vcov_def(object, joint = TRUE,
      standardized = FALSE,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.def.joint.std.all" ||
    what == "vcov.def.joint.standardized" ||
    what == "vcov.def.joint.std") {
    lav_inspect_vcov_def(object, joint = TRUE,
      standardized = TRUE, type = "std.all",
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.def.joint.std.lv") {
    lav_inspect_vcov_def(object, joint = TRUE,
      standardized = TRUE, type = "std.lv",
      add_labels = add_labels, add_class = add_class)
  } else if (what == "vcov.def.joint.std.nox") {
    lav_inspect_vcov_def(object, joint = TRUE,
      standardized = TRUE, type = "std.nox",
      add_labels = add_labels, add_class = add_class)

  } else if (what == "ugamma" || what == "ug" || what == "u.gamma") {
    lav_inspect_ugamma(object,
      add_labels = add_labels, add_class = add_class)
  } else if (what == "ufromugamma" || what == "u") {
    lav_inspect_u_from_ugamma(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    ### jacobians ####
  } else if (what == "delta") {
    lav_inspect_delta(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "delta.rownames") {
    lav_inspect_delta_rownames(object,
      drop_list_single_group = drop_list_single_group)

    ### casewise loglikelihoods ###
  } else if (what == "loglik.casewise") {
    lav_inspect_loglik_casewise(object, log_1 = TRUE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "lik.casewise") {
    lav_inspect_loglik_casewise(object, log_1 = FALSE,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    # multilevel #
  } else if (what == "icc") {
    lav_inspect_icc(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)
  } else if (what == "ranef") {
    lav_inspect_ranef(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    # instrumental variables
  } else if (what %in% c("iv", "ivs", "miiv", "miivs", "instr", "instruments")) {
    lav_inspect_iv(object,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("eqs")) {
    lav_inspect_eqs(object,
      drop_list_single_group = drop_list_single_group)
  } else if (what %in% c("sargan", "hansen")) {
    lav_inspect_sargan(object,
      drop_list_single_group = drop_list_single_group)

    # sam: the stored step-2 structural fit (local sam.method only)
  } else if (what %in% c("sam.struc.fit.object", "sam.struc")) {
    if (is.null(object@internal$sam.method)) {
      lav_msg_stop(gettext("this model was not fitted with sam()."))
    }
    fit_pa <- lav_sam_struc_object(object)
    if (is.null(fit_pa)) {
      lav_msg_stop(gettext(
        "no stored structural fit found: either sam.method = \"global\" was
         used, or this sam object was created by an older version of lavaan."))
    }
    fit_pa

    # post-checking
  } else if (what == "post.check" || what == "post") {
    lav_object_post_check(object)

    # options
  } else if (what == "options" || what == "lavoptions") {
    object@Options

    # version
  } else if (what == "version") {
    object@version

    # call
  } else if (what == "call") {
    as.list(object@call)

    # timing
  } else if (what == "timing") {
    object@timing

    # optim
  } else if (what == "optim") {
    object@optim

    # test
  } else if (what == "test") {
    object@test

    # zero cell tables
  } else if (what == "zero.cell.tables") {
    lav_inspect_zero_cell_tables(object,
      add_labels = add_labels, add_class = add_class,
      drop_list_single_group = drop_list_single_group)

    #### not found ####
  } else {
    lav_msg_stop(gettextf(
      "%1$s argument unknown: %2$s",
      "what", lav_msg_view(what)
    ))
  }

}


# helper functions (mostly to deal with older 'object' that may have
# been saved somewhere)
lav_inspect_est <- function(object, unrotated = FALSE) {

  if (inherits(object, "lavaan")) {
    # from 0.5-19, they are in the partable
    if (!is.null(object@ParTable$est)) {
      if (unrotated) {
        return_value <- object@ParTable$est.unrotated
      } else {
        return_value <- object@ParTable$est
                  # if this changes, tag @TDJorgensen in commit message
      }
    } else {
      partable <- parTable(object)
      return_value <- rep(as.numeric(NA), length(partable$lhs))
    }
  } else {
    # try generic coef()
    return_value <- coef(object, type = "user")
    if (is.matrix(return_value)) {
      # lavaanList?
      return_value <- rowMeans(return_value)
    }
  }

  return_value
}

lav_inspect_se <- function(object) {
  # from 0.5-19, they are in the partable
  if (!is.null(object@ParTable$se)) {
    return_value <- object@ParTable$se
  } else {
    partable <- parTable(object)
    return_value <- rep(as.numeric(NA), length(partable$lhs))
  }

  return_value
}

lav_inspect_std_se <- function(object) {

  if (!is.null(object@ParTable$se.std)) {
    return_value <- object@ParTable$se.std
  } else {
    tmp_std <- standardizedSolution(object)
    return_value <- tmp_std$se
  }

  return_value
}

lav_inspect_start <- function(object) {
  # from 0.5-19, they are in the partable
  if (!is.null(object@ParTable$start)) {
    return_value <- object@ParTable$start
  } else {
    # in < 0.5-19, we should look in @Fit@start
    return_value <- object@Fit@start
  }

  return_value
}

lav_inspect_mc <- function(object, add_labels = FALSE,
                                  add_class = FALSE) {
  # retrieve the Monte Carlo coefficient draws used for defined parameter
  # confidence intervals (Preacher & Selig 2012). Returns NULL if MC was
  # not requested.
  mc_coef <- NULL
  int <- try(slot(object, "internal"), silent = TRUE)
  if (!inherits(int, "try-error") && !is.null(int$monte.carlo)) {
    mc_coef <- int$monte.carlo$coef
  }
  if (is.null(mc_coef)) {
    return(NULL)
  }
  if (add_labels) {
    colnames(mc_coef) <- names(coef(object))
  }
  if (add_class) {
    class(mc_coef) <- c("lavaan.matrix", "matrix")
  }
  mc_coef
}

lav_inspect_boot <- function(object, add_labels = FALSE,
                                    add_class = FALSE) {

  if (object@Options$se   != "bootstrap" &&
    !any(c("bootstrap", "bollen.stine") %in% object@Options$test)) {
    lav_msg_stop(gettext("bootstrap was not used."))
  }

  # from 0.5-19. they are in a separate slot
  tmp <- try(slot(object, "boot"), silent = TRUE)
  if (inherits(tmp, "try-error")) {
    # older version of object?
    est <- lav_inspect_est(object)
    tmp_boot <- attr(est, "tmp.boot.COEF")
  } else {
    # 0.5-19 way
    tmp_boot <- object@boot$coef
  }

  # add coef names
  if (add_labels) {
    colnames(tmp_boot) <- names(coef(object))
  }

  # add class
  if (add_class) {
    class(tmp_boot) <- c("lavaan.matrix", "matrix")
  }

  tmp_boot
}


lav_inspect_modelmatrices <- function(object, what = "free",
    type = "free", add_labels = FALSE, add_class = FALSE,
    list_by_group = FALSE,
    drop_list_single_group = FALSE) {

  glist <- object@Model@GLIST

  current_verbose <- lav_verbose()
  if (what == "dx.free") {
    if (lav_verbose(FALSE)) on.exit(lav_verbose(current_verbose), TRUE)
    tmp_dx <- lav_model_grad(
      lavmodel       = object@Model,
      glist          = NULL,
      lavsamplestats = object@SampleStats,
      lavdata        = object@Data,
      lavcache       = object@Cache,
      type           = "free",
      group_weight   = TRUE,
      ceq_simple     = TRUE,
      delta          = NULL)
  } else if (what == "dx.all") {
    # the gradient for ALL model matrix elements (type = "allofthem")
    # has not been implemented for these model classes
    if (object@Model@multilevel || object@Model@categorical ||
        object@Model@conditional.x) {
      lav_msg_stop(gettext("dx.all is not available for multilevel,
        categorical or conditional.x models"))
    }
    if (lav_verbose(FALSE)) on.exit(lav_verbose(current_verbose), TRUE)
    glist <- lav_model_grad(
      lavmodel   = object@Model,
      glist          = NULL,
      lavsamplestats = object@SampleStats,
      lavdata        = object@Data,
      lavcache       = object@Cache,
      type           = "allofthem",
      group_weight   = TRUE,
      ceq_simple     = FALSE,
      delta          = NULL)
    names(glist) <- names(object@Model@GLIST)
  } else if (what == "std.all") {
    tmp_std <- lav_standardize_all(object)
  } else if (what == "std.lv") {
    tmp_std <- lav_standardize_lv(object)
  } else if (what == "std.nox") {
    tmp_std <- lav_standardize_all_nox(object)
  } else if (what == "se") {
    tmp_se <- lav_inspect_se(object)
  } else if (what == "std.se") {
    tmp_se <- lav_inspect_std_se(object)
  } else if (what == "start") {
    tmp_start <- lav_inspect_start(object)
  } else if (what == "est") {
    tmp_est <- lav_inspect_est(object)
  } else if (what == "est.unrotated") {
    # only an unrotated solution exists if the model contains EFA blocks
    # and a rotation (other than "none") was actually applied
    if (object@Model@nefa == 0L ||
      is.null(object@Options$rotation) ||
      object@Options$rotation == "none") {
      tmp_est <- lav_inspect_est(object, unrotated = FALSE)
    } else {
      tmp_est <- lav_inspect_est(object, unrotated = TRUE)
    }
  }

  for (mm in seq_along(glist)) {

    if (add_labels) {
      dimnames(glist[[mm]]) <- object@Model@dimNames[[mm]]
    }

    if (what == "free") {
      # fill in free parameter counts
      if (type == "free") {
        m_el_idx <- object@Model@m.free.idx[[mm]]
        x_el_idx <- object@Model@x.free.idx[[mm]]
        # } else if(type == "unco") {
        #    m.el.idx <- object@Model@m.unco.idx[[mm]]
        #    x.el.idx <- object@Model@x.unco.idx[[mm]]
      } else if (type == "partable") {
        m_el_idx <- object@Model@m.user.idx[[mm]]
        x_el_idx <- object@Model@x.user.idx[[mm]]
      } else {
        lav_msg_stop(gettextf(
          "%1$s argument unknown: %2$s",
          "type", lav_msg_view(type)
        ))
      }
      # erase everything
      glist[[mm]][, ] <- 0.0
      glist[[mm]][m_el_idx] <- x_el_idx
    } else if (what == "se" || what == "std.se") {
      # fill in standard errors
      m_user_idx <- object@Model@m.user.idx[[mm]]
      x_user_idx <- object@Model@x.user.idx[[mm]]
      # erase everything
      glist[[mm]][, ] <- 0.0
      glist[[mm]][m_user_idx] <- tmp_se[x_user_idx]
    } else if (what == "start") {
      # fill in starting values
      m_user_idx <- object@Model@m.user.idx[[mm]]
      x_user_idx <- object@Model@x.user.idx[[mm]]
      glist[[mm]][m_user_idx] <- tmp_start[x_user_idx]
    } else if (what %in% c("est", "est.unrotated")) {
      # fill in estimated parameter values
      m_user_idx <- object@Model@m.user.idx[[mm]]
      x_user_idx <- object@Model@x.user.idx[[mm]]
      glist[[mm]][m_user_idx] <- tmp_est[x_user_idx]
    } else if (what == "dx.free") {
      # fill in derivatives free parameters
      m_el_idx <- object@Model@m.free.idx[[mm]]
      x_el_idx <- object@Model@x.free.idx[[mm]]
      # erase everything
      glist[[mm]][, ] <- 0.0
      glist[[mm]][m_el_idx] <- tmp_dx[x_el_idx]
    } else if (what %in% c("std.all", "std.lv", "std.nox")) {
      m_user_idx <- object@Model@m.user.idx[[mm]]
      x_user_idx <- object@Model@x.user.idx[[mm]]
      glist[[mm]][m_user_idx] <- tmp_std[x_user_idx]
    }

    # class
    if (add_class) {
      if (object@Model@isSymmetric[mm]) {
        class(glist[[mm]]) <- c("lavaan.matrix.symmetric", "matrix")
      } else {
        class(glist[[mm]]) <- c("lavaan.matrix", "matrix")
      }
    }
  }

  # try to reflect `equality constraints'
  con_flag <- FALSE
  if (what == "free" && object@Model@eq.constraints) {
    # extract constraints from parameter table
    partable <- parTable(object)
    tmp_con <-  partable[partable$op %in% c("==", "<", ">"),
                         c("lhs", "op", "rhs")]
    rownames(tmp_con) <- NULL

    # replace 'labels' by parameter numbers
    tmp_id <- lav_pt_con_label_id(partable)
    tmp_label <- names(tmp_id)
    for (con in seq_len(nrow(tmp_con))) {
      # lhs
      lhs_labels <- all.vars(as.formula(paste("~", tmp_con[con, "lhs"])))

      if (length(lhs_labels) > 0L) {
        # par id
        lhs_freeid <- tmp_id[match(lhs_labels, tmp_label)]

        # substitute
        tmp <- tmp_con[con, "lhs"]
        for (pat in seq_along(lhs_labels)) {
          tmp <- sub(lhs_labels[pat], lhs_freeid[pat], tmp)
        }
        tmp_con[con, "lhs"] <- tmp
      }

      # rhs
      rhs_labels <- all.vars(as.formula(paste("~", tmp_con[con, "rhs"])))

      if (length(rhs_labels) > 0L) {
        # par id
        rhs_freeid <- tmp_id[match(rhs_labels, tmp_label)]
        # substitute
        tmp <- tmp_con[con, "rhs"]
        for (pat in seq_along(rhs_labels)) {
          tmp <- sub(rhs_labels[pat], rhs_freeid[pat], tmp)
        }
        tmp_con[con, "rhs"] <- tmp
      }
    } # con

    # add this info at the top
    # glist <- c(constraints = list(tmp.con), glist)
    # no, not a good idea, it does not work with list_by_group

    # add it as a 'header' attribute?
    attr(tmp_con, "header") <- "Note: model contains equality constraints:"
    con_flag <- TRUE
  }

  # should we group them per block?
  if (list_by_group) {
    lavmodel       <- object@Model
    nmat           <- lavmodel@nmat

    return_value <- vector("list", length = lavmodel@nblocks)
    for (b in seq_len(lavmodel@nblocks)) {
      # which mm belong to this block?
      mm_in_group <- 1:nmat[b] + cumsum(c(0, nmat))[b]

      return_value[[b]] <- glist[mm_in_group]
    }

    if (lavmodel@nblocks == 1L && drop_list_single_group) {
      return_value <- return_value[[1]]
    } else if (lavmodel@nblocks > 1L) {
      names(return_value) <- object@Data@block.label
    }
  } else {
    return_value <- glist
  }

  # header
  if (con_flag) {
    attr(return_value, "header") <- tmp_con
  }

  # lavaan.list
  if (add_class) {
    class(return_value) <- c("lavaan.list", "list")
  }

  return_value
}




# - fixme, should we export this function?
# - since 0.5-21, conditional.x = TRUE returns residual sample statistics
#    for ML, we have both joint and residual cov/var/...; but for
#    categorical = TRUE, we only have residual cov/var...; so, we
#    only return residual in both cases, whenever residual
# - since 0.6-3, we always extract the values from the @h1 slot (if present)
#   if meanstructure = FALSE, do NOT include $mean elements any longer
lav_inspect_sampstat <- function(object, h1 = TRUE,
    std = FALSE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  nblocks <- object@Model@nblocks
  ov_names <- object@pta$vnames$ov
  ov_names_res <- object@pta$vnames$ov.nox
  ov_names_x   <- object@pta$vnames$ov.x

  # slots
  lavsamplestats <- object@SampleStats
  lavmodel <- object@Model

  # if nlevels, override h1 to be TRUE, and set conditional.x = FALSE
  if (object@Data@nlevels > 1L) {
    h1 <- TRUE
    conditional_x <- FALSE # for now (0.6-12)
  } else {
    conditional_x <- lavmodel@conditional.x
  }

  # check if we have a non-empty @h1 slot
  if (length(object@h1) == 0L) {
    h1 <- FALSE
  } else {
    h1_implied <- object@h1$implied
  }

  # if h1 = FALSE and nlevels > 1L, nothing to show...
  if (!h1 && object@Data@nlevels > 1L) {
    lav_msg_stop(gettext(
      "sample statistics not available; refit with option h1 = TRUE"))
  }

  return_value <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {

    if (!conditional_x) {
      # covariance matrix
      if (h1) {
        return_value[[b]]$cov  <- h1_implied$cov[[b]]
      } else {
        return_value[[b]]$cov  <- lavsamplestats@cov[[b]]
      }
      if (std) {
        diag_orig <- diag(return_value[[b]]$cov)
        return_value[[b]]$cov <- cov2cor(return_value[[b]]$cov)
      }
      if (add_labels && !is.null(return_value[[b]]$cov)) {
        rownames(return_value[[b]]$cov) <- colnames(return_value[[b]]$cov) <-
          ov_names[[b]]
      }
      if (add_class) {
        class(return_value[[b]]$cov) <- c("lavaan.matrix.symmetric", "matrix")
      }

      # mean vector
      if (lavmodel@meanstructure) {
        if (h1) {
          return_value[[b]]$mean <- as.numeric(h1_implied$mean[[b]])
        } else {
          return_value[[b]]$mean <- as.numeric(lavsamplestats@mean[[b]])
        }
        if (std) {
          diag_orig[diag_orig < .Machine$double.eps] <- NA
          return_value[[b]]$mean <- return_value[[b]]$mean / sqrt(diag_orig)
        }
        if (add_labels) {
          names(return_value[[b]]$mean) <- ov_names[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$mean) <- c("lavaan.vector", "numeric")
        }
      }

      # thresholds
      if (lavmodel@categorical) {
        if (h1) {
          return_value[[b]]$th <- as.numeric(h1_implied$th[[b]])
        } else {
          return_value[[b]]$th <- as.numeric(lavsamplestats@th[[b]])
        }
        if (length(lavmodel@num.idx[[b]]) > 0L) {
          num_idx <- which(lavmodel@th.idx[[b]] == 0)
          return_value[[b]]$th <- return_value[[b]]$th[-num_idx]
        }
        # FIXME: what to do if std = TRUE (depends on delta/theta)
        if (add_labels) {
          names(return_value[[b]]$th) <- object@pta$vnames$th[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$th) <- c("lavaan.vector", "numeric")
        }
      }
      # !conditional.x
    } else {
      # if conditional.x = TRUE

      # residual covariance matrix
      if (h1) {
        return_value[[b]]$res.cov  <- h1_implied$res.cov[[b]]
      } else {
        return_value[[b]]$res.cov  <- lavsamplestats@res.cov[[b]]
      }
      if (std) {
        diag_orig <- diag(return_value[[b]]$res.cov)
        return_value[[b]]$res.cov <- cov2cor(return_value[[b]]$res.cov)
      }
      if (add_labels) {
        rownames(return_value[[b]]$res.cov) <-
          colnames(return_value[[b]]$res.cov) <-
          ov_names_res[[b]]
      }
      if (add_class) {
        class(return_value[[b]]$res.cov) <-
          c("lavaan.matrix.symmetric", "matrix")
      }

      # intercepts
      if (lavmodel@meanstructure) {
        if (h1) {
          return_value[[b]]$res.int <- as.numeric(h1_implied$res.int[[b]])
        } else {
          return_value[[b]]$res.int <- as.numeric(lavsamplestats@res.int[[b]])
        }
        if (std) {
          diag_orig[diag_orig < .Machine$double.eps] <- NA
          return_value[[b]]$res.int <- return_value[[b]]$res.int /
            sqrt(diag_orig)
        }
        if (add_labels) {
          names(return_value[[b]]$res.int) <- ov_names_res[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$res.int) <- c("lavaan.vector", "numeric")
        }
      }

      # thresholds
      if (lavmodel@categorical) {
        if (h1) {
          return_value[[b]]$res.th <- as.numeric(h1_implied$res.th[[b]])
        } else {
          return_value[[b]]$res.th <- as.numeric(lavsamplestats@res.th[[b]])
        }
        if (length(lavmodel@num.idx[[b]]) > 0L) {
          num_idx <- which(lavmodel@th.idx[[b]] == 0)
          return_value[[b]]$res.th <- return_value[[b]]$res.th[-num_idx]
        }
        # FIXME: if std: what to do?
        if (add_labels) {
          names(return_value[[b]]$res.th) <- object@pta$vnames$th[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$res.th) <- c("lavaan.vector", "numeric")
        }
      }

      # slopes
      if (lavmodel@nexo[b] > 0L) {
        if (h1) {
          return_value[[b]]$res.slopes  <- h1_implied$res.slopes[[b]]
        } else {
          return_value[[b]]$res.slopes  <- lavsamplestats@res.slopes[[b]]
        }
        # FIXME: if std: what to do? (here: b.z = b * s.x /s.y)
        if (std) {
          tmp_y <- matrix(sqrt(diag_orig),
            nrow(return_value[[b]]$res.slopes),
            ncol(return_value[[b]]$res.slopes))
          tmp_x <- matrix(sqrt(diag(lavsamplestats@cov.x[[b]])),
            nrow(return_value[[b]]$res.slopes),
            ncol(return_value[[b]]$res.slopes), byrow = TRUE)
          return_value[[b]]$res.slopes <- return_value[[b]]$res.slopes /
                                          tmp_y * tmp_x
        }
        if (add_labels) {
          rownames(return_value[[b]]$res.slopes) <- ov_names_res[[b]]
          colnames(return_value[[b]]$res.slopes) <- ov_names_x[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$res.slopes) <- c("lavaan.matrix", "matrix")
        }
      }

      # cov.x
      if (lavmodel@nexo[b] > 0L) {
        return_value[[b]]$cov.x  <- lavsamplestats@cov.x[[b]]
        if (std) {
          diag_orig <- diag(return_value[[b]]$cov.x)
          return_value[[b]]$cov.x <- cov2cor(return_value[[b]]$cov.x)
        }
        if (add_labels) {
          rownames(return_value[[b]]$cov.x) <- ov_names_x[[b]]
          colnames(return_value[[b]]$cov.x) <- ov_names_x[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$cov.x) <-
            c("lavaan.matrix.symmetric", "matrix")
        }
      }

      # mean.x
      if (lavmodel@nexo[b] > 0L) {
        return_value[[b]]$mean.x <- as.numeric(object@SampleStats@mean.x[[b]])
        if (std) {
          diag_orig[diag_orig < .Machine$double.eps] <- NA
          return_value[[b]]$mean.x <- return_value[[b]]$mean.x / sqrt(diag_orig)
        }
        if (add_labels) {
          names(return_value[[b]]$mean.x) <- ov_names_x[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$mean.x) <- c("lavaan.vector", "numeric")
        }
      }

    } # conditional.x

    # stochastic weights
    if (lavmodel@group.w.free) {
      # to be consistent with the 'implied' values,
      # transform so group.w is the 'log(group.freq)'
      return_value[[b]]$group.w <-
        log(lavsamplestats@group.w[[b]] * lavsamplestats@ntotal)
      if (add_labels) {
        names(return_value[[b]]$group.w) <- "w"
      }
      if (add_class) {
        class(return_value[[b]]$group.w) <- c("lavaan.vector", "numeric")
      }
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}


lav_inspect_data <- function(object, add_labels = FALSE,
                                    drop_list_single_group = FALSE) {

  n_g <- object@Data@ngroups
  if (object@Model@conditional.x) {
    return_value <- vector("list", length = n_g)
    for (g in 1:n_g) {
      return_value[[g]] <- cbind(object@Data@X[[g]],
        object@Data@eXo[[g]])
    }
  } else {
    return_value <- object@Data@X
  }

  if (add_labels) {
    for (g in 1:n_g) {
      if (object@Model@conditional.x) {
        colnames(return_value[[g]]) <- c(object@Data@ov.names[[g]],
          object@Data@ov.names.x[[g]])
      } else {
        colnames(return_value[[g]]) <- object@Data@ov.names[[g]]
      }
    }
  }

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

lav_inspect_case_idx <- function(object,
                                        drop_list_single_group = FALSE) {
  n_g <- object@Data@ngroups

  return_value <- object@Data@case.idx

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

# lav_inspect_case_idx <- function(object, level = 1L,
#                                        drop_list_single_group = FALSE) {
#    #FIXME: if lavaan ever allows 3-level or cross-classifed models,
#    #      "level=" should be a character indicating the clustering variable
#
#    n.g <- object@Data@ngroups
#    nlevels <- object@Data@nlevels
#    if (nlevels == 1L) level <- 1L
#                         # if what="cluster.idx" for single-level model
#
#    if (level == 2L) {
#        # level-2 (cluster) IDs
#        return.value <- lapply(object@Data@Lp, function(gg)
#                           gg$cluster.id[[2]][ gg$cluster.idx[[2]] ])
#        #FIXME: update if lavaan ever accepts 3-level or
#                                           cross-classified models
#
#    } else return.value <- object@Data@case.idx # level-1 (casewise) IDs
#
#    if(n.g == 1L && drop_list_single_group) {
#        return.value <- return.value[[1]]
#    } else {
#        if(length(object@Data@group.label) > 0L) {
#            names(return.value) <- unlist(object@Data@group.label)
#        }
#    }
#    return.value
# }
#

# cluster info
lav_inspect_cl_info <- function(
    object,
    what = "cluster.size",
    level = 2L,
    drop_list_single_group = FALSE) {

  n_g <- object@Data@ngroups
  nlevels <- object@Data@nlevels

  # just in case we have no clusters
  if (nlevels == 1L) {
    if (what %in% c("nclusters", "ncluster.size", "cluster.id")) {
      return_value <- as.list(rep(1L, n_g))
    } else if (what %in% c("cluster.size", "cluster.sizes")) {
      return_value <- object@Data@nobs
    } else if (what %in% c("cluster.idx", "cluster.label")) {
      # everybody belongs to cluster 1
      return_value <- lapply(seq_len(n_g),
        function(gg) rep(1L, object@Data@nobs[[gg]]))
    } else if (what == "average.cluster.size") {
      lav_msg_stop(gettext(
        "average cluster size only available for clustered data"))
    }
  }

  # if we do have clusters
  if (nlevels > 1L) {
    return_value <- vector("list", length = n_g)

    for (g in seq_len(n_g)) {
      tmp_lp <- object@Data@Lp[[g]]
      if (what == "nclusters") {
        return_value[[g]] <- tmp_lp$nclusters[[level]]
      } else if (what == "ncluster.size") {
        return_value[[g]] <- tmp_lp$ncluster.size[[level]]
      } else if (what == "cluster.size") {
        return_value[[g]] <- tmp_lp$cluster.size[[level]]
      } else if (what == "cluster.id") {
        return_value[[g]] <- tmp_lp$cluster.id[[level]]
      } else if (what == "cluster.idx") {
        return_value[[g]] <- tmp_lp$cluster.idx[[level]]
      } else if (what == "cluster.label") {
        return_value[[g]] <-
          tmp_lp$cluster.id[[level]][tmp_lp$cluster.idx[[level]]]
      } else if (what == "cluster.sizes") {
        return_value[[g]] <- tmp_lp$cluster.sizes[[level]]
      } else if (what == "average.cluster.size") {
        nn_g <- object@Data@nobs[[g]]
        cluster_size <- tmp_lp$cluster.size[[level]]
        nclusters <- tmp_lp$nclusters[[level]]
        return_value[[g]] <- (nn_g^2 - sum(cluster_size^2)) /
          (nn_g * (nclusters - 1L))
      }
    } # g
  } # nlevels > 1L

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}


# count the number of clusters, or obtain tmp.n within each cluster
# lav_inspect_ncluster <- function(object, sizes = FALSE, #level = 2L,
#                                        drop_list_single_group = FALSE) {
#    n.g <- object@Data@ngroups
#    nlevels <- object@Data@nlevels
#
#    if (nlevels == 1L) {
#      # single-level model, return sample size(s) or count 1 cluster per group
#      return.value <- if (sizes) unlist(object@Data@nobs) else rep(1L, n.g)
#
#    } else if (sizes) {
#      # for each group, a vector of cluster sizes
#      return.value <- lapply(object@Data@Lp, function(gg) gg$cluster.size[[2]])
#      #FIXME: update if lavaan ever accepts 3-level or cross-classified models
#
#      if (n.g == 1L && drop_list_single_group)
#                       return.value <- return.value[[1]]
#
#    } else {
#      # number of clusters in each group
#      return.value <- sapply(object@Data@Lp, function(gg) gg$nclusters[[2]])
#      #FIXME: update if lavaan ever accepts 3-level or cross-classified models
#    }
#
#    # assign group names, if applicable
#    if (n.g > 1L) names(return.value) <- unlist(object@Data@group.label)
#    return.value
# }

lav_inspect_rsquare <- function(object, est_std_all = NULL,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  nblocks <- object@Model@nblocks
  return_value <- vector("list", length = nblocks)

  if (is.null(est_std_all)) {
    est_std_all <- lav_standardize_all(object)
  }

  partable <- object@ParTable
  partable$rsquare <- 1.0 - est_std_all
  # no values > 1.0
  partable$rsquare[partable$rsquare > 1.0] <- as.numeric(NA)

  for (b in seq_len(nblocks)) {
    ind_names <- partable$rhs[which(partable$op == "=~" &
      partable$block == b)]
    eqs_y_names <- partable$lhs[which(partable$op == "~"  &
      partable$block == b)]
    y_names <- unique(c(ind_names, eqs_y_names))

    idx <- which(partable$op == "~~" & partable$lhs %in% y_names &
      partable$rhs == partable$lhs & partable$block == b)
    tmp <- partable$rsquare[idx]

    if (add_labels && length(tmp) > 0L) {
      names(tmp) <- partable$lhs[idx]
    }
    if (add_class) {
      class(tmp) <- c("lavaan.vector", "numeric")
    }

    return_value[[b]] <- tmp
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

# model implied sample stats
lav_inspect_implied <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  nblocks <- object@Model@nblocks
  ov_names <- object@pta$vnames$ov
  ov_names_res <- object@pta$vnames$ov.nox
  ov_names_x   <- object@pta$vnames$ov.x

  # slots
  lavimplied <- object@implied
  lavmodel   <- object@Model

  # if nlevels, always set conditional.x = FALSE
  if (object@Data@nlevels > 1L) {
    lavimplied <- lav_model_implied_cond2uncond(lavimplied)
    conditional_x <- FALSE # for now (0.6-12)
  } else {
    conditional_x <- lavmodel@conditional.x
  }

  return_value <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {

    if (!conditional_x) {
      # covariance matrix
      return_value[[b]]$cov  <- lavimplied$cov[[b]]
      if (add_labels && !is.null(return_value[[b]]$cov)) {
        rownames(return_value[[b]]$cov) <- colnames(return_value[[b]]$cov) <-
          ov_names[[b]]
      }
      if (add_class) {
        class(return_value[[b]]$cov) <- c("lavaan.matrix.symmetric", "matrix")
      }

      # mean vector
      if (lavmodel@meanstructure) {
        return_value[[b]]$mean <- as.numeric(lavimplied$mean[[b]])
        if (add_labels) {
          names(return_value[[b]]$mean) <- ov_names[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$mean) <- c("lavaan.vector", "numeric")
        }
      }

      # thresholds
      if (lavmodel@categorical) {
        return_value[[b]]$th <- as.numeric(lavimplied$th[[b]])
        if (length(object@Model@num.idx[[b]]) > 0L) {
          num_idx <- which(object@Model@th.idx[[b]] == 0)
          return_value[[b]]$th <- return_value[[b]]$th[-num_idx]
        }
        if (add_labels) {
          names(return_value[[b]]$th) <- object@pta$vnames$th[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$th) <- c("lavaan.vector", "numeric")
        }
      }
      # !conditional.x
    } else {
      # if conditional.x = TRUE
      # residual covariance matrix
      return_value[[b]]$res.cov  <- lavimplied$res.cov[[b]]
      if (add_labels) {
        rownames(return_value[[b]]$res.cov) <-
          colnames(return_value[[b]]$res.cov) <-
          ov_names_res[[b]]
      }
      if (add_class) {
        class(return_value[[b]]$res.cov) <-
          c("lavaan.matrix.symmetric", "matrix")
      }

      # intercepts
      if (lavmodel@meanstructure) {
        return_value[[b]]$res.int <- as.numeric(lavimplied$res.int[[b]])
        if (add_labels) {
          names(return_value[[b]]$res.int) <- ov_names_res[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$res.int) <- c("lavaan.vector", "numeric")
        }
      }

      # thresholds
      if (lavmodel@categorical) {
        return_value[[b]]$res.th <- as.numeric(lavimplied$res.th[[b]])
        if (length(object@Model@num.idx[[b]]) > 0L) {
          num_idx <- which(object@Model@th.idx[[b]] == 0)
          return_value[[b]]$res.th <- return_value[[b]]$res.th[-num_idx]
        }
        if (add_labels) {
          names(return_value[[b]]$res.th) <- object@pta$vnames$th[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$res.th) <- c("lavaan.vector", "numeric")
        }
      }

      # slopes
      if (lavmodel@nexo[b] > 0L) {
        return_value[[b]]$res.slopes  <- lavimplied$res.slopes[[b]]
        if (add_labels) {
          rownames(return_value[[b]]$res.slopes) <- ov_names_res[[b]]
          colnames(return_value[[b]]$res.slopes) <- ov_names_x[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$res.slopes) <- c("lavaan.matrix", "matrix")
        }
      }

      # cov.x
      if (lavmodel@nexo[b] > 0L) {
        return_value[[b]]$cov.x  <- lavimplied$cov.x[[b]]
        if (add_labels) {
          rownames(return_value[[b]]$cov.x) <- ov_names_x[[b]]
          colnames(return_value[[b]]$cov.x) <- ov_names_x[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$cov.x) <-
            c("lavaan.matrix.symmetric", "matrix")
        }
      }

      # mean.x
      if (lavmodel@nexo[b] > 0L) {
        return_value[[b]]$mean.x  <- as.numeric(lavimplied$mean.x[[b]])
        if (add_labels) {
          names(return_value[[b]]$mean.x) <- ov_names_x[[b]]
        }
        if (add_class) {
          class(return_value[[b]]$mean.x) <- c("lavaan.vector", "numeric")
        }
      }


    } # conditional.x

    # stochastic weights
    if (lavmodel@group.w.free) {
      return_value[[b]]$group.w <- lavimplied$group.w[[b]]
      if (add_labels) {
        names(return_value[[b]]$group.w) <- "w" # somewhat redundant
      }
      if (add_class) {
        class(return_value[[b]]$group.w) <- c("lavaan.vector", "numeric")
      }
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}


# residuals: _inspect_sampstat - _inspect_implied
lav_inspect_residuals <- function(object, h1 = TRUE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  lav_residuals(object, type = "raw", h1 = h1,
    add_labels = add_labels, add_class = add_class,
    drop_list_single_group = drop_list_single_group)
}


lav_inspect_cov_lv <- function(object, correlation_metric = FALSE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # compute lv covar
  return_value <- lav_model_veta(lavmodel = object@Model,
                                 remove_dummy_lv = TRUE)

  # nblocks
  nblocks <- length(return_value)

  # cor + labels + class
  for (b in seq_len(nblocks)) {

    if (correlation_metric && nrow(return_value[[b]]) > 1L) {
      # note: cov2cor fails if matrix is empty!
      return_value[[b]] <- cov2cor(return_value[[b]])
    }

    if (add_labels) {
      colnames(return_value[[b]]) <- rownames(return_value[[b]]) <-
        object@pta$vnames$lv[[b]]
    }

    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.matrix.symmetric", "matrix")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_mean_lv <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # compute lv means
  return_value <- lav_model_eeta(lavmodel = object@Model,
    lavsamplestats = object@SampleStats,
    remove_dummy_lv = TRUE)

  # nblocks
  nblocks <- length(return_value)

  # ensure numeric
  return_value <- lapply(return_value, as.numeric)

  # labels + class
  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L) {
      names(return_value[[b]]) <- object@pta$vnames$lv[[b]]
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_cov_all <- function(object, correlation_metric = FALSE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # compute extended model implied covariance matrix (both ov and lv)
  return_value <- lav_model_cov_both(lavmodel = object@Model,
    remove_dummy_lv = TRUE)

  # nblocks
  nblocks <- length(return_value)

  # cor + labels + class
  for (b in seq_len(nblocks)) {

    if (correlation_metric && nrow(return_value[[b]]) > 1L) {
      # note: cov2cor fails if matrix is empty!
      return_value[[b]] <- cov2cor(return_value[[b]])
    }

    if (add_labels) {
      tmp_names <- c(object@pta$vnames$ov.model[[b]],
        object@pta$vnames$lv[[b]])
      colnames(return_value[[b]]) <- rownames(return_value[[b]]) <- tmp_names
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.matrix.symmetric", "matrix")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}


lav_inspect_cov_ov <- function(object, correlation_metric = FALSE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # get model-implied covariance matrix observed
  if (object@Model@conditional.x) {
    return_value <- object@implied$res.cov
  } else {
    return_value <- object@implied$cov
  }

  # nblocks
  nblocks <- length(return_value)

  # cor + labels + class
  for (b in seq_len(nblocks)) {

    if (correlation_metric && nrow(return_value[[b]]) > 1L) {
      # note: cov2cor fails if matrix is empty!
      return_value[[b]] <- cov2cor(return_value[[b]])
    }

    if (add_labels) {
      colnames(return_value[[b]]) <- rownames(return_value[[b]]) <-
        object@pta$vnames$ov.model[[b]]
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.matrix.symmetric", "matrix")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_mean_ov <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # compute ov means
  if (object@Model@conditional.x) {
    return_value <- object@implied$res.int
  } else {
    return_value <- object@implied$mean
  }

  # nblocks
  nblocks <- length(return_value)

  # make numeric
  return_value <- lapply(return_value, as.numeric)

  # labels + class
  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L) {
      names(return_value[[b]]) <- object@pta$vnames$ov.model[[b]]
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_th <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # thresholds
  if (object@Model@conditional.x) {
    return_value <- object@implied$res.th
  } else {
    return_value <- object@implied$th
  }

  # nblocks
  nblocks <- length(return_value)

  # make numeric
  return_value <- lapply(return_value, as.numeric)

  # labels + class
  for (b in seq_len(nblocks)) {
    # note: for (continuous) multilevel models, @Model@th.idx may contain
    # fewer elements than nblocks; but then there is nothing to remove
    if (length(object@Model@num.idx[[b]]) > 0L &&
        b <= length(object@Model@th.idx) &&
        length(object@Model@th.idx[[b]]) > 0L) {
      num_idx <- which(object@Model@th.idx[[b]] == 0)
      return_value[[b]] <- return_value[[b]][-num_idx]
    }
    if (add_labels && length(return_value[[b]]) > 0L) {
      names(return_value[[b]]) <- object@pta$vnames$th[[b]]
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_th_idx <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # thresholds idx
  return_value <- object@SampleStats@th.idx

  # nblocks
  nblocks <- length(return_value)

  # labels + class
  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L) {
      names(return_value[[b]]) <- object@SampleStats@th.names[[b]]
    }
    if (add_class && !is.null(return_value[[b]])) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_vy <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # 'unconditional' model-implied variances
  #  - same as diag(Sigma.hat) if all Y are continuous)
  #  - 1.0 (or delta^2) if categorical
  #  - if also Gamma, cov.x is used (only if categorical)

  return_value <- lav_model_vy(lavmodel = object@Model, glist = NULL,
    diagonal_only = TRUE)

  # nblocks
  nblocks <- length(return_value)

  # labels + class
  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L) {
      if (object@Model@categorical || object@Model@conditional.x) {
        # y-side variables only (no exogenous x covariates)
        names(return_value[[b]]) <- object@pta$vnames$ov.nox[[b]]
      } else {
        names(return_value[[b]]) <- object@pta$vnames$ov[[b]]
      }
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_fs_determinacy <- function(object,
  squared = TRUE, fs_method = "regression",
  add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  fs <- lavPredict(object, type = "lv", method = fs_method, rel = TRUE)
  return_value <- attr(fs, "rel")

  # determinacies or reliabilities?
  if (!squared) {
    return_value <- lapply(return_value, sqrt)
  }

  # nblocks
  nblocks <- length(return_value)

  # ensure numeric
  return_value <- lapply(return_value, as.numeric)

  # labels + class
  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L) {
      names(return_value[[b]]) <- object@pta$vnames$lv[[b]]
    }
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}


lav_inspect_theta <- function(object, correlation_metric = FALSE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {
  # get residual covariances
  return_value <- lav_model_theta(lavmodel = object@Model)

  # nblocks
  nblocks <- length(return_value)

  # labels + class
  for (b in seq_len(nblocks)) {

    if (correlation_metric && nrow(return_value[[b]]) > 0L) {
      if (all(return_value[[b]] == 0)) {
        return_value[[b]] <- return_value[[b]]
      } else {
        return_value[[b]] <- cov2cor(return_value[[b]])
      }
    }

    if (add_labels && length(return_value[[b]]) > 0L) {
      colnames(return_value[[b]]) <- rownames(return_value[[b]]) <-
        object@pta$vnames$ov.model[[b]]
    }

    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.matrix.symmetric", "matrix")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}


lav_inspect_mi_coverage <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  n_g <- object@Data@ngroups
  return_value <- vector("list", n_g)

  for (g in 1:n_g) {
    if (!is.null(object@Data@Mp[[g]])) {
      return_value[[g]] <- object@Data@Mp[[g]]$coverage
    } else {
      nvar <- length(object@Data@ov.names[[g]])
      return_value[[g]] <- matrix(1.0, nvar, nvar)
    }

    if (add_labels && length(return_value[[g]]) > 0L) {
      colnames(return_value[[g]]) <- rownames(return_value[[g]]) <-
        object@pta$vnames$ov.model[[g]]
    }

    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.matrix.symmetric", "matrix")
    }
  }

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

lav_inspect_mi_patterns <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  n_g <- object@Data@ngroups
  return_value <- vector("list", n_g)

  for (g in 1:n_g) {
    if (!is.null(object@Data@Mp[[g]])) {
      return_value[[g]] <- object@Data@Mp[[g]]$pat
    } else {
      nvar <- length(object@Data@ov.names[[g]])
      return_value[[g]] <- matrix(TRUE, 1L, nvar)
      rownames(return_value[[g]]) <- object@Data@nobs[[g]]
    }

    if (add_labels && length(return_value[[g]]) > 0L) {
      colnames(return_value[[g]]) <- object@pta$vnames$ov.model[[g]]
    }

    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.matrix", "matrix")
    }
  }

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

lav_inspect_empty_idx <- function(object,
                                         drop_list_single_group = FALSE) {

  n_g <- object@Data@ngroups

  # get empty idx
  return_value <- vector("list", n_g)

  for (g in 1:n_g) {
    if (!is.null(object@Data@Mp[[g]])) {
      return_value[[g]] <- object@Data@Mp[[g]]$empty.idx
    } else {
      return_value[[g]] <- integer(0L)
    }
  }

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}


lav_inspect_wls_est <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  return_value <- lav_model_wls_est(object@Model)

  if (add_labels) {
    tmp_names <- lav_inspect_delta_rownames(object,
      drop_list_single_group = FALSE)
  }

  # nblocks
  nblocks <- length(return_value)

  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L &&
        object@Data@nlevels == 1L) {
      names(return_value[[b]]) <- tmp_names[[b]]
    }

    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_wls_obs <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  return_value <- object@SampleStats@WLS.obs ### FIXME: should be in @h1??

  # two-level models (e.g., estimator = "ML"): the WLS.obs slot may not
  # be filled in; construct the observed statistic vector from the h1
  # (saturated) model instead, using the same layout as wls.est
  if (object@Data@nlevels > 1L && any(sapply(return_value, is.null))) {
    if (length(object@h1) == 0L) {
      lav_msg_stop(gettext("h1 slot is not available; refit with h1 = TRUE"))
    }
    return_value <- lav_model_wls_est(lavmodel = object@Model,
      lavimplied = object@h1$implied)
  }

  if (add_labels) {
    tmp_names <- lav_inspect_delta_rownames(object,
      drop_list_single_group = FALSE)
  }

  # nblocks
  nblocks <- length(return_value)

  for (b in seq_len(nblocks)) {
    if (add_labels && length(return_value[[b]]) > 0L &&
        object@Data@nlevels == 1L) {
      names(return_value[[b]]) <- tmp_names[[b]]
    }

    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.vector", "numeric")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}

lav_inspect_wls_v <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  # return.value <- lav_model_wls_v(lavmodel       = object@Model,
  #                       lavsamplestats = object@SampleStats,
  #                       structured     = TRUE,
  #                       lavdata        = object@Data)
  # WLS.V == (traditionally) h1 expected information
  return_value <- lav_model_h1_info_expected(lavobject = object)
  # this affects fit measures gfi, agfi, pgfi


  # nblocks
  nblocks <- length(return_value)

  # if estimator == "DWLS" or "ULS", we only stored the diagonal
  # hence, we create a full matrix here
  if (object@Options$estimator %in% c("DWLS", "ULS")) {
    return_value <- lapply(return_value,
      function(x) {
        nr <- NROW(x)
        diag(x, nrow = nr, ncol = nr)
      })
  }

  if (add_labels) {
    tmp_names <- lav_inspect_delta_rownames(object,
      drop_list_single_group = FALSE)
  }

  # label + class
  for (b in seq_len(nblocks)) {
    # labels
    if (add_labels && nrow(return_value[[b]]) > 0L &&
        object@Data@nlevels == 1L) {
      colnames(return_value[[b]]) <-
        rownames(return_value[[b]]) <- tmp_names[[b]]
    }

    # class
    if (add_class) {
      class(return_value[[b]]) <- c("lavaan.matrix", "matrix")
    }
  }

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (nblocks > 1L) {
    names(return_value) <- object@Data@block.label
  }

  return_value
}


lav_inspect_sampstat_gamma <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  return_value <- lav_gamma_used(object)

  if (add_labels) {
    tmp_names <- lav_inspect_delta_rownames(object,
      drop_list_single_group = FALSE)
  }

  # nblocks
  nblocks <- length(return_value)

  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]

    # labels
    if (add_labels) {
      colnames(return_value) <- rownames(return_value) <- tmp_names[[1]]
    }

    # class
    if (add_class) {
      class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
    }
  } else {
    if (object@Data@nlevels == 1L && length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)

      # labels
      if (add_labels) {
        for (g in seq_len(object@Data@ngroups)) {
          colnames(return_value[[g]]) <-
            rownames(return_value[[g]]) <- tmp_names[[g]]
        }
      }

      # class
      if (add_class) {
        for (g in seq_len(object@Data@ngroups)) {
          class(return_value[[g]]) <- c("lavaan.matrix.symmetric", "matrix")
        }
      }
    } else if (object@Data@nlevels > 1L &&
      length(object@Data@group.label) == 0L) {
      # two-level: NACOV/Gamma is stored per group (one matrix covering both
      # levels), not per level; only add level labels if the lengths match
      if (length(return_value) == length(object@Data@level.label)) {
        names(return_value) <- object@Data@level.label
      }
    }
  }

  return_value
}


lav_inspect_grad <- function(object,
    add_labels = FALSE, add_class = FALSE, logl = FALSE,
    optim = FALSE) {

  lavmodel       <- object@Model
  lavdata        <- object@Data
  lavsamplestats <- object@SampleStats

  if (optim) {
    logl <- FALSE
  }

  if (lavsamplestats@missing.flag ||
    object@Options$estimator == "PML") {
    group_weight <- FALSE
  } else {
    group_weight <- TRUE
  }
  current_verbose <- lav_verbose()
  if (lav_verbose(FALSE)) on.exit(lav_verbose(current_verbose), TRUE)
  dx <- lav_model_grad(
    lavmodel       = lavmodel,
    glist          = NULL,
    lavsamplestats = lavsamplestats,
    lavdata        = object@Data,
    lavcache       = object@Cache,
    type           = "free",
    group_weight   = group_weight)
  lav_verbose(current_verbose)
  # if logl, rescale to get gradient wrt the loglikelihood
  if (logl) {
    if (lavmodel@estimator %in% c("ML")) {
      if (lavdata@nlevels == 1L) {
        # currently, this is just a sign switch
        dx <- -1 * dx
      } else {
        lavpartable <- object@ParTable
        # gradient.log = gradient.obj * (2 * tmp.n) / nclusters

        if (lavdata@ngroups == 1L) {
          tmp_n <- lavdata@Lp[[1]]$nclusters[[1]]
          nclusters <- lavdata@Lp[[1]]$nclusters[[2]]
          dx <- dx * (2 * tmp_n) / nclusters
        } else {
          group_values <- lav_pt_group_values(lavpartable)
          for (g in seq_len(lavdata@ngroups)) {
            tmp_n <- lavdata@Lp[[g]]$nclusters[[1]]
            nclusters <- lavdata@Lp[[g]]$nclusters[[2]]
            g_idx <-
              which((lavpartable$group ==
                group_values[g])[lavpartable$free > 0L])
            dx[g_idx] <- dx[g_idx] * (2 * tmp_n) / nclusters
          }
        }
      }
    } else {
      # FIXME:
      # non-likelihood: what to do? just switch the sign for now.
      # Note: this is used in lavTestScore()
      dx <- -1 * dx
    }
  }

  # optim?
  if (optim) {
    # 1. scale (note: divide, not multiply!)
    if (!is.null(object@optim$parscale)) {
      dx <- dx / object@optim$parscale
    }

    # 2. pack
    if (lavmodel@eq.constraints) {
      dx <- as.numeric(dx %*% lavmodel@eq.constraints.K)
    }
    # only for PML: divide by tmp.n (to speed up convergence)
    if (lavmodel@estimator == "PML") {
      dx <- dx / lavsamplestats@ntotal
    }
  }

  # labels
  if (add_labels) {
    if (optim && lavmodel@eq.constraints) {
      tmp_names_all <- lav_pt_labels(object@ParTable, type = "free")
      tmp_seq <- seq_along(tmp_names_all)
      pack_seq <- as.numeric((tmp_seq - lavmodel@eq.constraints.k0) %*%
        +lavmodel@eq.constraints.K)
      ok_idx <- which(pack_seq %in% tmp_seq)
      tmp_names <- rep("(eq.con)", length(pack_seq))
      tmp_names[ok_idx] <- tmp_names_all[pack_seq[ok_idx]]
      names(dx) <- tmp_names
    } else {
      names(dx) <- lav_pt_labels(object@ParTable, type = "free")
    }
  }

  # class
  if (add_class) {
    class(dx) <- c("lavaan.vector", "numeric")
  }

  dx
}

lav_inspect_hessian <- function(object,
    add_labels = FALSE, add_class = FALSE) {

  return_value <- lav_model_hessian(
    lavmodel       = object@Model,
    lavsamplestats = object@SampleStats,
    lavdata        = object@Data,
    lavcache       = object@Cache,
    lavoptions     = object@Options,
    group_weight   = TRUE)

  # labels
  if (add_labels) {
    colnames(return_value) <- rownames(return_value) <-
      lav_pt_labels(object@ParTable, type = "free")
  }

  # class
  if (add_class) {
    class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
  }

  return_value
}

lav_inspect_info <- function(object,
    information = "default", augmented = FALSE, inverted = FALSE,
    add_labels = FALSE, add_class = FALSE) {

  if (information != "default") {
    # override option
    object@Options$information <- information
  }

  return_value <- lav_model_info(lavmodel =  object@Model,
    lavsamplestats = object@SampleStats,
    lavdata        = object@Data,
    lavcache       = object@Cache,
    lavimplied     = object@implied,
    lavh1          = object@h1,
    lavoptions     = object@Options,
    extra          = FALSE,
    augmented      = augmented,
    inverted       = inverted)

  # labels
  if (add_labels) {
    tmp_names <- lav_pt_labels(object@ParTable, type = "free")
    if (augmented) {
      n_extra <- nrow(return_value) - length(tmp_names)
      if (n_extra > 0L) {
        tmp_names <- c(tmp_names, paste("aug", 1:n_extra, sep = ""))
      }
    }
    colnames(return_value) <- rownames(return_value) <- tmp_names
  }

  # class
  if (add_class) {
    class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
  }

  return_value
}

lav_inspect_h1_info <- function(object,
    information = "default", h1_information = "default", inverted = FALSE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  if (information != "default") {
    # override option
    object@Options$information <- information
  }
  if (h1_information != "default") {
    # override option
    object@Options$h1.information <- h1_information
  }

  lavmodel <- object@Model
  lavdata  <- object@Data

  # list!
  return_value <- lav_model_h1_info(lavmodel = lavmodel,
    lavsamplestats = object@SampleStats,
    lavdata        = lavdata,
    lavcache       = object@Cache,
    lavimplied     = object@implied,
    lavh1          = object@h1,
    lavoptions     = object@Options)

  # inverted? (NOT USED)
  # if(inverted) {
  #    return.value <- lapply(return.value, solve) # FIXME: handle errors...
  # }

  if (add_labels) {
    tmp_names <- lav_inspect_delta_rownames(object,
      drop_list_single_group = FALSE)
  }

  # labels/class per group
  for (g in seq_len(lavmodel@ngroups)) {
    # labels
    if (add_labels) {
      colnames(return_value[[g]]) <-
        rownames(return_value[[g]]) <- tmp_names[[g]]
    }

    # class
    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.matrix.symmetric", "matrix")
    }
  }

  # drop list?
  if (lavmodel@ngroups == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (!is.null(lavdata)) {
    if (length(lavdata@group.label) > 0L) {
      names(return_value) <- unlist(lavdata@group.label)
    }
  }

  return_value
}

lav_inspect_vcov <- function(object, standardized = FALSE,
    type = "std.all", ov_std = NULL, free_only = TRUE,
    add_labels = FALSE, add_class = FALSE, remove_duplicated = FALSE) {

  lavmodel <- object@Model
  lavoptions <- object@Options

  # store partable with pta in object to use cache in called functions
  if (is.null(attr(object, "vnames"))) {
    object@ParTable <- lav_pt_set_cache(object@ParTable, object@pta)
  }

  if (object@optim$npar == 0) {
    return_value <- matrix(0, 0, 0)
  } else {
    # check if we already have it
    # tmp <- try(slot(object, "vcov"), silent = TRUE)
    # if( !inherits(tmp, "try-error") && !is.null(object@vcov$vcov)
    #   && !(rotation && standardized)) {
    if (!is.null(object@vcov$vcov) &&
        !(standardized && lavmodel@ceq.simple.only)) {
      return_value <- object@vcov$vcov
                 # if this changes, tag @TDJorgensen in commit message
    } else if (standardized && lavmodel@ceq.simple.only &&
               !is.null(object@vcov$vcov) &&
               nrow(object@vcov$vcov) == nrow(lavmodel@ceq.simple.K)) {
      # special case: ceq.simple.only + standardized
      # tmp.jac (below) is in the 'compact' (nx.free) space, but the stored
      # vcov lives in the larger 'unco' space (it duplicates the simple
      # equality parameters). Reduce the stored vcov to the compact space.
      # NB: do NOT recompute with ceq.simple.only dropped, as that corrupts
      # the (expected) information matrix for the shared parameters.
      return_value <- lav_model_vcov_unco_to_free(lavmodel, object@vcov$vcov)
    } else {
      # compute it again
      # if(rotation && standardized) {
      #    lavmodel <- lav_model_set_parameters(lavmodel,
      #                    x = object@optim$x)
      #    lavoptions <- object@Options
      #    lavoptions$rotation.se <- "delta"
      # }

      return_value <- lav_model_vcov(
        lavmodel       = lavmodel,
        lavsamplestats = object@SampleStats,
        lavoptions     = lavoptions,
        lavdata        = object@Data,
        lavcache       = object@Cache,
        lavimplied     = object@implied,
        lavh1          = object@h1
      )
      # if(rotation && !standardized) {
      #    # fixme: compute tmp.vcov.rot manually...
      #    stop("lavaan ERROR: rerun with store.vcov = TRUE")
      # }

      if (is.null(return_value)) {
        return(return_value)
      }

      # special case: ceq.simple.only + standardized
      # tmp.jac (below) is in the 'compact' (nx.free) space, but the recomputed
      # vcov may live in the larger 'unco' space (it duplicates the simple
      # equality parameters). Reduce it to the compact space if needed.
      if (standardized && lavmodel@ceq.simple.only) {
        return_value <- lav_model_vcov_unco_to_free(lavmodel, return_value)
      }
    }
  }

  # strip attributes
  attr(return_value, "E.inv") <- NULL
  attr(return_value, "B0") <- NULL
  attr(return_value, "B0.group") <- NULL
  attr(return_value, "Delta") <- NULL
  attr(return_value, "WLS.V") <- NULL
  attr(return_value, "tmp.boot.COEF") <- NULL
  attr(return_value, "tmp.boot.TEST") <- NULL

  # standardized?
  if (standardized) {
    if (type == "std.lv") {
      tmp_fun <- lav_standardize_lv_x
    } else if (type == "std.all") {
      tmp_fun <- lav_standardize_all_x
    } else if (type == "std.nox") {
      tmp_fun <- lav_standardize_all_nox_x
    } else if (type == "std.user") {
      tmp_fun <- function(x, lavobject, ...) {
        lav_standardize_all_x(x, lavobject = lavobject, ov_std = ov_std, ...)
      }
    }

    x_vec <- lav_model_get_parameters(lavmodel)
    tmp_jac <- try(lav_func_jacobian_complex(func = tmp_fun, x = x_vec,
      lavobject = object),
    silent = TRUE)
    use_complex <- !inherits(tmp_jac, "try-error")
    if (!use_complex) { # eg. pnorm()
      tmp_jac <- lav_func_jacobian_simple(func = tmp_fun, x = x_vec,
        lavobject = object)
    }
    # }

    # second-order delta method: compute Hessians for each row of tmp_jac
    tmp_hess <- NULL
    if (isTRUE(lavoptions$se.delta.second.order)) {
      if (use_complex) {
        tmp_hess <- try(lav_func_hessian_complex(func = tmp_fun, x = x_vec,
          lavobject = object), silent = TRUE)
        if (inherits(tmp_hess, "try-error")) {
          tmp_hess <- try(lav_func_hessian_simple(func = tmp_fun, x = x_vec,
            lavobject = object), silent = TRUE)
        }
      } else {
        tmp_hess <- try(lav_func_hessian_simple(func = tmp_fun, x = x_vec,
          lavobject = object), silent = TRUE)
      }
      if (inherits(tmp_hess, "try-error")) {
        lav_msg_warn(gettext(
          "Could not compute the Hessian for the standardized solution;
           falling back to the first-order delta method."))
        tmp_hess <- NULL
      }
    }

    # handle ceq.simple.only
    #if (object@Model@ceq.simple.only) {
    #  tmp.jac <- tmp.jac %*% t(object@Model@ceq.simple.K)
    #}

    # rows of tmp.jac contain *all* the parameters in the parameter table
    if (free_only) {
      if (object@Model@ceq.simple.only) {
        pt_free_idx <- which(object@ParTable$free > 0L &
          !duplicated(object@ParTable$free))
      } else {
        pt_free_idx <- which(object@ParTable$free > 0L)
      }
      tmp_jac <- tmp_jac[pt_free_idx, , drop = FALSE]
      if (!is.null(tmp_hess)) {
        tmp_hess <- tmp_hess[pt_free_idx]
      }
    }

    # apply Delta method
    param_vcov <- return_value
    return_value <- tmp_jac %*% return_value %*% t(tmp_jac)

    # add second-order delta method correction:
    # Cov(g_i, g_j) ~ grad_i' V grad_j + 0.5 tr(H_i V H_j V)
    if (!is.null(tmp_hess)) {
      ndef_1 <- length(tmp_hess)
      correction <- matrix(0, ndef_1, ndef_1)
      hv_list <- lapply(tmp_hess, function(h) h %*% param_vcov)
      for (i in seq_len(ndef_1)) {
        for (j in i:ndef_1) {
          correction[i, j] <- 0.5 * sum(hv_list[[i]] * t(hv_list[[j]]))
          if (i != j) {
            correction[j, i] <- correction[i, j]
          }
        }
      }
      return_value <- return_value + correction
    }

    # force return.value to be symmetric and pd
    return_value <- (return_value + t(return_value)) / 2
    # return.value <- lav_mat_sym_force_pd(return.value,
    #                                     tol = 1e-09) # was 1e-06 < 0.6-9
  }

  # labels
  if (add_labels) {
    # if(rotation && !free.only) {
    #    # todo
    # } else {
    colnames(return_value) <- rownames(return_value) <-
      lav_pt_labels(object@ParTable,
                          ## add "user" labels?
                          type = ifelse(standardized && !free_only,
                                        "user", "free"))
    # }
  }

  # alias?
  if (remove_duplicated && lavmodel@eq.constraints) {
    simple_flag <- lav_con_check_simple(lavmodel)
    if (simple_flag) {
      tmp_lab <- lav_pt_labels(object@ParTable, type = "free")
      dup_flag <- duplicated(tmp_lab)
      return_value <- return_value[!dup_flag, !dup_flag, drop = FALSE]
    } else {
      lav_msg_warn(
        gettext("alias is TRUE, but equality constraints do not appear
                to be simple; returning full vcov"))
    }
  }

  # class
  if (add_class) {
    class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
  }

  return_value
}

lav_inspect_vcov_def <- function(object, joint = FALSE,
    standardized = FALSE, type = "std.all",
    add_labels = FALSE, add_class = FALSE) {

  lavmodel    <- object@Model
  lavpartable <- object@ParTable
  free_idx <- which(lavpartable$free > 0L)
  def_idx <- which(lavpartable$op == ":=")
  joint_idx <- c(free_idx, def_idx)

  if (!joint && length(def_idx) == 0L) {
    return(matrix(0, 0, 0))
  } else if (joint && length(joint_idx) == 0L) {
    return(matrix(0, 0, 0))
  }

  if (standardized) {
    # Monte Carlo path: empirical cov of the standardized def parameters
    mc_coef_std <- lav_inspect_mc(object)
    if (!is.null(mc_coef_std) && length(def_idx) > 0L) {
      if (type == "std.lv") {
        tmp_fun_std <- lav_standardize_lv_x
      } else if (type == "std.all") {
        tmp_fun_std <- lav_standardize_all_x
      } else if (type == "std.nox") {
        tmp_fun_std <- lav_standardize_all_nox_x
      }
      mc_std <- try(apply(mc_coef_std, 1L, tmp_fun_std,
                          lavobject = object), silent = TRUE)
      if (!inherits(mc_std, "try-error")) {
        if (is.matrix(mc_std)) {
          mc_std <- t(mc_std)
        } else {
          mc_std <- as.matrix(mc_std)
        }
        mc_std[!is.finite(mc_std)] <- as.numeric(NA)
        if (joint) {
          mc_std_sub <- mc_std[, joint_idx, drop = FALSE]
        } else {
          mc_std_sub <- mc_std[, def_idx, drop = FALSE]
        }
        return_value <- cov(mc_std_sub, use = "pairwise.complete.obs")
      } else {
        mc_coef_std <- NULL
      }
    }
    # delta-based standardized cov (default path)
    if (is.null(mc_coef_std) || length(def_idx) == 0L) {
      tmp_vcov <- lav_inspect_vcov(object,
        standardized = TRUE,
        type = type, free_only = FALSE,
        add_labels = FALSE, add_class = FALSE)
      if (joint) {
        return_value <- tmp_vcov[joint_idx, joint_idx, drop = FALSE]
      } else {
        return_value <- tmp_vcov[def_idx, def_idx, drop = FALSE]
      }
    }
  } else {
    # get free parameters
    x <- lav_model_get_parameters(lavmodel, type = "free")

    # bootstrap, Monte Carlo or delta?
    mc_coef <- lav_inspect_mc(object)
    if (!is.null(object@boot$coef)) {
      tmp_boot <- object@boot$coef
      # remove NA rows
      error_idx <- attr(tmp_boot, "error.idx")
      if (length(error_idx) > 0L) {
        tmp_boot <- tmp_boot[-error_idx, , drop = FALSE] # drops attributes
      }
      tmp_boot_def <- apply(tmp_boot, 1L, lavmodel@def.function)
      if (length(def_idx) == 1L) {
        tmp_boot_def <- as.matrix(tmp_boot_def)
      } else {
        tmp_boot_def <- t(tmp_boot_def)
      }
      if (joint) {
        tmp_boot_def <- cbind(tmp_boot, tmp_boot_def)
      }
      return_value <- cov(tmp_boot_def)
    } else if (!is.null(mc_coef)) {
      mc_def <- apply(mc_coef, 1L, lavmodel@def.function)
      if (length(def_idx) == 1L) {
        mc_def <- as.matrix(mc_def)
      } else {
        mc_def <- t(mc_def)
      }
      if (joint) {
        mc_def <- cbind(mc_coef, mc_def)
      }
      nmc <- nrow(mc_def)
      return_value <- cov(mc_def) * (nmc - 1) / nmc
    } else {
      # tmp.vcov
      tmp_vcov <- lav_inspect_vcov(object,
        standardized = FALSE,
        type = type, free_only = TRUE,
        add_labels = FALSE,
        add_class = FALSE)
      # tmp_jac (below) is in the compact (nx.free) space; when ceq.simple.only,
      # tmp_vcov is in the larger 'unco' space, so reduce it to the compact space
      tmp_vcov <- lav_model_vcov_unco_to_free(lavmodel, tmp_vcov)

      # regular delta method
      tmp_jac <- try(lav_func_jacobian_complex(func = lavmodel@def.function,
        x = x), silent = TRUE)
      use_complex <- !inherits(tmp_jac, "try-error")
      if (!use_complex) { # eg. pnorm()
        tmp_jac <- lav_func_jacobian_simple(func = lavmodel@def.function,
          x = x)
      }

      # second-order delta method
      tmp_hess <- NULL
      if (isTRUE(object@Options$se.delta.second.order)) {
        if (use_complex) {
          tmp_hess <- try(lav_func_hessian_complex(
            func = lavmodel@def.function, x = x), silent = TRUE)
          if (inherits(tmp_hess, "try-error")) {
            tmp_hess <- try(lav_func_hessian_simple(
              func = lavmodel@def.function, x = x), silent = TRUE)
          }
        } else {
          tmp_hess <- try(lav_func_hessian_simple(
            func = lavmodel@def.function, x = x), silent = TRUE)
        }
        if (inherits(tmp_hess, "try-error")) {
          lav_msg_warn(gettext(
            "Could not compute the Hessian of the defined parameters;
             falling back to the first-order delta method."))
          tmp_hess <- NULL
        }
      }

      if (joint) {
        # the free block maps the (compact) free parameters to the 'user' free
        # parameters: identity normally, but ceq.simple.K when ceq.simple.only
        # (so the duplicated simple-equality parameters are expanded back)
        if (lavmodel@ceq.simple.only &&
            ncol(lavmodel@ceq.simple.K) == ncol(tmp_jac)) {
          free_block <- lavmodel@ceq.simple.K
        } else {
          free_block <- diag(nrow = ncol(tmp_jac))
        }
        tmp_jac2 <- rbind(free_block, tmp_jac)
        return_value <- tmp_jac2 %*% tmp_vcov %*% t(tmp_jac2)
        if (!is.null(tmp_hess)) {
          # second-order correction: only the def-def block is non-zero
          # (the free part is linear, so its Hessian is zero)
          nfree <- nrow(free_block)
          ndef_1 <- length(tmp_hess)
          correction <- matrix(0, nfree + ndef_1, nfree + ndef_1)
          hv_list <- lapply(tmp_hess, function(h) h %*% tmp_vcov)
          for (i in seq_len(ndef_1)) {
            for (j in i:ndef_1) {
              correction[nfree + i, nfree + j] <-
                0.5 * sum(hv_list[[i]] * t(hv_list[[j]]))
              if (i != j) {
                correction[nfree + j, nfree + i] <-
                  correction[nfree + i, nfree + j]
              }
            }
          }
          return_value <- return_value + correction
        }
      } else {
        return_value <- tmp_jac %*% tmp_vcov %*% t(tmp_jac)
        if (!is.null(tmp_hess)) {
          ndef_1 <- length(tmp_hess)
          correction <- matrix(0, ndef_1, ndef_1)
          hv_list <- lapply(tmp_hess, function(h) h %*% tmp_vcov)
          for (i in seq_len(ndef_1)) {
            for (j in i:ndef_1) {
              correction[i, j] <-
                0.5 * sum(hv_list[[i]] * t(hv_list[[j]]))
              if (i != j) {
                correction[j, i] <- correction[i, j]
              }
            }
          }
          return_value <- return_value + correction
        }
      }
    }
  }

  # labels: the same parameter names as coef()/vcov() (user labels override
  # the lhs-op-rhs default; := rows keep their name) (github issue #504)
  if (add_labels) {
    tmp_names <- lav_pt_labels(lavpartable, type = "user")
    if (joint) {
      tmp_names <- tmp_names[joint_idx]
    } else {
      tmp_names <- tmp_names[def_idx]
    }
    colnames(return_value) <- rownames(return_value) <- tmp_names
  }

  # class
  if (add_class) {
    class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
  }

  return_value
}

lav_inspect_ugamma <- function(object,
    add_labels = FALSE, add_class = FALSE) {

  out <- lav_test_sb(lavobject = object,
    method = "original",
    return_ugamma = TRUE)
  return_value <- out$UGamma

  # labels
  # if (add_labels) {
  # colnames(return.value) <- rownames(return.value) <-
  # }

  # class
  if (add_class) {
    class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
  }

  return_value
}

lav_inspect_u_from_ugamma <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  out <- lav_test_sb(lavobject     = object,
    method        = "original",
    return_u      = TRUE)
  return_value <- out$UfromUGamma

  # labels
  # if (add_labels) {
  # colnames(return.value) <- rownames(return.value) <-
  # }

  # class
  if (add_class) {
    class(return_value) <- c("lavaan.matrix.symmetric", "matrix")
  }

  return_value
}


# Extract a single (estimated) model matrix (by name) from the @GLIST slot,
# per block. Used by the "mm.*" options (e.g. "mm.lambda", "mm.delta"). Note
# that for "delta" this is the scaling matrix, NOT the Delta jacobian (see
# lav_inspect_delta below).
lav_inspect_mm <- function(object, mm_name = "lambda", add_labels = FALSE,
    add_class = FALSE, drop_list_single_group = FALSE) {

  # estimated model matrices, per block
  MM <- lav_inspect_modelmatrices(object, what = "est",
    add_labels = add_labels, add_class = add_class,
    list_by_group = TRUE, drop_list_single_group = FALSE)

  nblocks <- length(MM)
  return_value <- vector("list", length = nblocks)
  names(return_value) <- names(MM)
  for (b in seq_len(nblocks)) {
    if (is.null(MM[[b]][[mm_name]])) {
      lav_msg_stop(gettextf(
        "model matrix \"%s\" is not available for this model", mm_name))
    }
    return_value[[b]] <- MM[[b]][[mm_name]]
  }

  # drop list?
  if (nblocks == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  }

  return_value
}

# Delta (jacobian: d samplestats / d free_parameters)
lav_inspect_delta <- function(object,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  lavmodel    <- object@Model
  lavdata     <- object@Data
  lavpartable <- object@ParTable
  # lavpta      <- object@pta

  return_value <- lav_inspect_delta_internal(lavmodel = lavmodel,
    lavdata = lavdata, lavpartable = lavpartable,
    add_labels = add_labels, add_class = add_class,
    drop_list_single_group = drop_list_single_group)

  return_value
}

lav_inspect_delta_rownames <- function(
    object,
    lavmodel = NULL,
    lavpartable = NULL,
    drop_list_single_group = FALSE) {
  if (!is.null(object)) {
    lavmodel    <- object@Model
    lavpartable <- object@ParTable
    lavpta      <- object@pta
    lavdata     <- object@Data
  } else {
    lavdata     <- NULL
    lavpta <- lav_pt_attributes(lavpartable)
  }

  categorical    <- lavmodel@categorical
  correlation    <- lavmodel@correlation
  conditional_x  <- lavmodel@conditional.x
  group_w_free   <- lavmodel@group.w.free
  nvar           <- lavmodel@nvar
  num_idx        <- lavmodel@num.idx
  th_idx         <- lavmodel@th.idx
  nblocks        <- lavmodel@nblocks

  # store names per block, rbind later
  tmp_names <- vector("list", length = nblocks)

  # output is per group
  return_value <- vector("list", lavmodel@ngroups)

  for (g in 1:nblocks) {

    if (conditional_x) {
      ov_names <- lavpta$vnames$ov.nox[[g]]
    } else {
      ov_names <- lavpta$vnames$ov[[g]]
    }
    ov_names_x <- lavpta$vnames$ov.x[[g]]
    nvar <- length(ov_names)


    names_cov <- names_cor <- names_var <- character(0L)
    names_mu <- names_pi <- names_th <- character(0L)
    names_gw <- character(0L)

    # Sigma
    # - if continuous: vech(Sigma)
    # - if categorical: first numeric variances, then
    # tmp <- apply(expand.grid(ov.names, ov.names), 1L,
    #             paste, collapse = "~~")

    # if(categorical) {
    #    names.cor <- tmp[lav_mat_vech_idx(nvar, diagonal = FALSE)]
    #    names.var <- tmp[lav_mat_diag_idx(nvar)[num.idx[[g]]]]
    # } else {
    #    names.cov <- tmp[lav_mat_vech_idx(nvar, diagonal = TRUE)]
    # }

    # NOTE: in 0.6-1, we use the same order, but 'label' in row-wise
    # format (eg x1 ~~ x2 instead of x2 ~~ x1)
    tmp <- matrix(apply(expand.grid(ov_names, ov_names), 1L,
      paste, collapse = "~~"), nrow = nvar)
    if (categorical) {
      names_cor <- lav_mat_vechru(tmp, diagonal = FALSE)
      names_var <- diag(tmp)[num_idx[[g]]]
    } else if (correlation) {
      names_cor <- lav_mat_vechru(tmp, diagonal = FALSE)
    } else {
      names_cov <- lav_mat_vechru(tmp, diagonal = TRUE)
    }


    # Mu
    if (!categorical && lavmodel@meanstructure) {
      names_mu <- paste(ov_names, "~1", sep = "")
    }

    # Pi
    if (conditional_x && lavmodel@nexo[g] > 0L) {
      names_pi <- apply(expand.grid(ov_names, ov_names_x), 1L,
        paste, collapse = "~")
    }

    # th
    if (categorical) {
      names_th <- lavpta$vnames$th[[g]]
      # interweave numeric intercepts, if any
      if (length(num_idx[[g]]) > 0L) {
        tmp <- character(length(th_idx[[g]]))
        tmp[th_idx[[g]] > 0] <- names_th
        tmp[th_idx[[g]] == 0] <- paste(ov_names[num_idx[[g]]],
          "~1", sep = "")
        names_th <- tmp
      }
    }

    # gw
    if (group_w_free) {
      names_gw <- "w"
    }

    tmp_names[[g]] <- c(names_gw,
      names_th, names_mu,
      names_pi,
      names_cov, names_var, names_cor)

  } # blocks

  # multilevel?
  if (lavmodel@multilevel) {
    for (g in 1:lavmodel@ngroups) {
      return_value[[g]] <- c(tmp_names[[(g - 1) * 2 + 1]],
        tmp_names[[(g - 1) * 2 + 2]])
    }
  } else {
    return_value <- tmp_names
  }

  # drop list?
  if (lavmodel@ngroups == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (!is.null(lavdata)) {
    if (length(lavdata@group.label) > 0L) {
      names(return_value) <- unlist(lavdata@group.label)
    }
  }

  return_value
}

lav_inspect_delta_internal <- function(
    lavmodel = NULL, lavdata  = NULL,
    lavpartable = NULL,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  return_value <- lav_model_delta(lavmodel)

  if (add_labels) {
    tmp_pnames <- lav_pt_labels(lavpartable, type = "free")
    tmp_rownames  <- lav_inspect_delta_rownames(object = NULL,
      lavmodel = lavmodel, lavpartable = lavpartable,
      drop_list_single_group = FALSE)
  }

  for (g in seq_len(lavmodel@ngroups)) {
    # add labels
    if (add_labels) {
      colnames(return_value[[g]]) <- tmp_pnames
      rownames(return_value[[g]]) <- tmp_rownames[[g]]
    }

    # add class
    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.matrix", "matrix")
    }

  } # ngroups

  # drop list?
  if (lavmodel@ngroups == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(lavdata@group.label) > 0L) {
      names(return_value) <- unlist(lavdata@group.label)
    }
  }

  return_value
}

lav_inspect_zero_cell_tables <-                          # nolint
  function(object, add_labels = FALSE, add_class = FALSE,
            drop_list_single_group = FALSE) {
  # categorical?
  if (!object@Model@categorical) {
    lav_msg_warn(gettext("no categorical variables in fitted model"))
    return(invisible(list()))
  }

  lavdata <- object@Data

  # create 2-way tables
  tmp_table <- lavTables(object, dimension = 2L, output = "data.frame",
    statistic = NULL)

  # select tables with empty cells
  empty_id <- tmp_table$id[which(tmp_table$obs.freq == 0)]


  if (length(empty_id) == 0L) {
    # only when lavInspect() is used, give message
    if (add_class) {
      cat("(There are no tables with empty cells for this fitted model)\n")
    }
    return(invisible(list()))
  }
  lav_tables_cells_format(
      tmp_table[tmp_table$id %in% empty_id, ],
      lavdata = lavdata,
      drop_list_single_group = drop_list_single_group)
}

lav_inspect_coef <- function(object, type = "free",
                                    add_labels = FALSE, add_class = FALSE) {

  if (type == "user" || type == "all") {
    type <- "user"
    idx <- seq_along(object@ParTable$lhs)
  } else if (type == "free") {
    #idx <- which(object@ParTable$free > 0L & !duplicated(object@ParTable$free))
    idx <- which(object@ParTable$free > 0L)
  } else {
    lav_msg_stop(gettextf(
      "%1$s argument must be either %2$s or %3$s",
      "type", "free", "user"))
  }
  tmp_est <- lav_inspect_est(object)
  cof <- tmp_est[idx]

  # labels?
  if (add_labels) {
    names(cof) <- lav_pt_labels(object@ParTable, type = type)
  }

  # class
  if (add_class) {
    class(cof) <- c("lavaan.vector", "numeric")
  }

  cof
}

lav_inspect_npar <- function(object, ceq = FALSE) {

  # free parameters (going to the optimizer)
  npar <- object@Model@nx.free

  # account for equality constraints?
  # note: when ceq.simple.only is TRUE, nx.free already excludes the simple
  # equality constraints (they are absorbed via ceq.simple.K), and the
  # (synthesized) con.jac lives in the larger 'unco' space; subtracting its
  # rank here would double-count the constraints (see lav_model_loglik.R)
  if (ceq && !object@Model@ceq.simple.only &&
      nrow(object@Model@con.jac) > 0L) {
    ceq_idx <- attr(object@Model@con.jac, "ceq.idx")
    if (length(ceq_idx) > 0L) {
      neq <- qr(object@Model@con.jac[ceq_idx, , drop = FALSE])$rank
      npar <- npar - neq
    }
  }

  npar
}

lav_inspect_icc <- function(object, add_labels = FALSE,
                                   add_class = FALSE,
                                   drop_list_single_group = FALSE) {

  lavdata <- object@Data
  n_g <- lavdata@ngroups
  return_value <- vector("list", n_g)

  # clustered data?
  if (length(lavdata@cluster) == 0L) {
    lav_msg_stop(gettext(
      "intraclass correlation only available for clustered data"))
  } else if (lavdata@nlevels == 1L) {
    lav_msg_stop(gettext(
      "intraclass correlation only available if the model syntax
       contains levels"))
  }

  if (length(object@h1) == 0L) {
    lav_msg_stop(gettext("h1 slot is not available; refit with h1 = TRUE"))
  }

  # implied statistics
  implied <- object@h1$implied

  for (g in 1:n_g) {
    sigma_w <- implied$cov[[(g - 1) * lavdata@nlevels + 1]]
    sigma_b <- implied$cov[[(g - 1) * lavdata@nlevels + 2]]

    w_diag <- diag(sigma_w)
    b_diag <- diag(sigma_b)

    return_value[[g]] <- numeric(length(w_diag))

    ov_names_l <- lavdata@ov.names.l[[g]]
    w_idx <- which(ov_names_l[[1]] %in% ov_names_l[[2]])
    w_names <- ov_names_l[[1]][w_idx]
    b_idx <- match(w_names, ov_names_l[[2]])

    return_value[[g]][w_idx] <- b_diag[b_idx] / (w_diag[w_idx] + b_diag[b_idx])

    # label
    if (add_labels) {
      names(return_value[[g]]) <- ov_names_l[[1]]
    }

    # class
    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.vector", "numeric")
    }
  } # g

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

lav_inspect_ranef <- function(object, add_labels = FALSE,
                                     add_class = FALSE,
                                     drop_list_single_group = FALSE) {

  lavdata <- object@Data
  lavsamplestats <- object@SampleStats

  n_g <- lavdata@ngroups
  return_value <- vector("list", n_g)

  # multilevel?
  if (lavdata@nlevels == 1L) {
    lav_msg_stop(gettext(
      "random effects only available for clustered data (in the long format)"))
  }

  # implied statistics
  lavimplied <- object@implied

  for (g in 1:n_g) {

    tmp_lp <- lavdata@Lp[[g]]
    tmp_ylp <- lavsamplestats@YLp[[g]]

    # implied for this group
    group_idx <- (g - 1) * lavdata@nlevels + seq_len(lavdata@nlevels)
    implied_group <- lapply(lavimplied, function(x) x[group_idx])

    # random effects (=random intercepts or cluster means)
    if (lavdata@missing %in% c("ml", "ml.x")) {
      # missing data: posterior means from the E-step, conditioning on
      # ALL the observed data (new in 0.7-1)
      mb_j <- lav_mvn_cl_mi_estep_ranef(
        y1 = lavdata@X[[g]], y2 = tmp_ylp[[2]]$Y2,
        lp = tmp_lp, mp = lavdata@Mp[[g]],
        mu_w = implied_group$mean[[1]],
        sigma_w = implied_group$cov[[1]],
        mu_b = implied_group$mean[[2]],
        sigma_b = implied_group$cov[[2]],
        se = FALSE)
    } else {
      out <- lav_mvn_cl_implied22l(lp = tmp_lp, implied = implied_group)
      mb_j <- lav_mvn_cl_em_estep_ranef(ylp = tmp_ylp, lp = tmp_lp,
        sigma_w = out$sigma.w, sigma_b = out$sigma.b,
        sigma_zz = out$sigma.zz, sigma_yz = out$sigma.yz,
        mu_z = out$mu.z, mu_w = out$mu.w, mu_b = out$mu.b,
        se = FALSE)
    }
    return_value[[g]] <- mb_j

    ov_names_l <- lavdata@ov.names.l[[g]]

    # label
    if (add_labels) {
      colnames(return_value[[g]]) <- ov_names_l[[1]]
    }

    # class
    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.matrix", "matrix")

    }
  } # g

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

# casewise loglikelihood contributions
lav_inspect_loglik_casewise <- function(object, log_1 = TRUE,
    add_labels = FALSE, add_class = FALSE, drop_list_single_group = FALSE) {

  lavdata <- object@Data
  lavsamplestats <- object@SampleStats
  lavimplied <- object@implied
  lavoptions <- object@Options

  n_g <- lavdata@ngroups
  return_value <- vector("list", n_g)

  # multilevel?
  if (lavdata@nlevels > 1L) {
    lav_msg_stop(gettext("casewise (log)likelihoods contributions
     not yet available for clustered data"))
  }

  # estimator ML?
  if (object@Options$estimator != "ML") {
    lav_msg_stop(gettextf("casewise (log)likelihoods contributions
     only available for estimator = %s",
    dQuote("ML")))
  }

  for (g in 1:n_g) {

    if (lavsamplestats@missing.flag) {
      return_value[[g]] <-
        lav_mvn_mi_llik_casewise(y = lavdata@X[[g]],
          wt = lavdata@weights[[g]],
          mu = lavimplied$mean[[g]],
          sigma_1 = lavimplied$cov[[g]],
          x_idx  = lavsamplestats@x.idx[[g]])
    } else { # single-level, complete data
      if (lavoptions$conditional.x) {
        if (!is.null(lavdata@weights[[g]])) {
          lav_msg_stop(gettext("no support (yet) if weights are used."))
        }
        return_value[[g]] <- lav_mvreg_loglik_data(
          y          = lavdata@X[[g]],
          exo        = lavdata@eXo[[g]],
          res_int    = lavimplied$res.int[[g]],
          res_slopes = lavimplied$res.slopes[[g]],
          res_cov    = lavimplied$res.cov[[g]],
          casewise   = TRUE)

      } else {

        if (object@Model@meanstructure) {
          tmp_mean <- lavimplied$mean[[g]]
        } else {
          tmp_mean <- lavsamplestats@mean[[g]]
        }
        return_value[[g]] <-
          lav_mvn_loglik_data(y = lavdata@X[[g]],
            wt = lavdata@weights[[g]],
            mu = tmp_mean,
            sigma_1 = lavimplied$cov[[g]],
            x_idx = lavsamplestats@x.idx[[g]],
            casewise = TRUE)
      }
    } # single-level, complete data

    # log. = FALSE?
    if (!log_1) {
      return_value[[g]] <- exp(return_value[[g]])
    }

    # label
    # if(add_labels) {
    # }

    # class
    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.vector", "numeric")

    }

  } # g

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

# Mahalanobis distances for factor scores or casewise residuals
# type = "lv" -> factor scores
# type = "resid" -> casewise residuals
#
# we always use Bartlett factor scores (see Yuan & Hayashi 2010)
# (this has no impact on the m-distances for the factor scores,
#  and only a very slight impact on the m-distances for the casewise
#  residuals; but asymptotically, only when we use Bartlett factor
#  scores are the 'true scores' (=LAMBDA %*% FS) orthogonal to the
#  casewise residuals)
lav_inspect_mdist2 <- function(object, type = "resid", squared = TRUE,
                                      add_labels = FALSE, add_class = FALSE,
                                      drop_list_single_group = FALSE) {

  lavdata <- object@Data
  n_g <- lavdata@ngroups

  if (object@Model@categorical) {
    # ordered categorical (or mixed) data: the expected M-distances of the
    # latent responses, obtained by Monte Carlo integration (Mansolf &
    # Reise, 2017); note that here the expected distance E[d] is not the
    # square root of the expected squared distance E[d^2], so both are
    # computed directly
    out <- lav_predict_mdist_cat(lavobject = object, type = type,
      method = "ML") # = Bartlett
    if (squared) {
      return_value <- lapply(out, "[[", "d2")
    } else {
      return_value <- lapply(out, "[[", "d")
    }
  } else {
    # lavPredict()
    out <- lavPredict(object, type = type, method = "ML", # = Bartlett
      label = FALSE, fsm = TRUE, mdist = TRUE,
      se = "none", acov = "none")
    return_value <- attr(out, "mdist")

    # squared?
    if (!squared) {
      return_value <- lapply(return_value, sqrt)
    }
  }

  for (g in seq_len(n_g)) {
    # labels?
    # if(add_labels) {
    # }

    # class
    if (add_class) {
      class(return_value[[g]]) <- c("lavaan.vector", "numeric")
    }
  } # g

  if (n_g == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else {
    if (length(object@Data@group.label) > 0L) {
      names(return_value) <- unlist(object@Data@group.label)
    }
  }

  return_value
}

# N versus N-1 (or N versus N-G in the multiple group setting)
# Changed 0.5-15: suggestion by Mark Seeto
lav_inspect_ntotal <- function(object) {
  if (object@Options$estimator %in% c("ML", "PML", "FML", "catML") &&
    object@Options$likelihood %in% c("default", "normal")) {
    n <- object@SampleStats@ntotal
  } else {
    n <- object@SampleStats@ntotal - object@SampleStats@ngroups
  }

  n
}

lav_inspect_iv <- function(object, drop_list_single_group = FALSE) {

  if (is.null(object@internal$eqs)) {
    lav_msg_stop(gettext("no equations/ivs found"))
  }
  lavmodel <- object@Model
  lavdata <- object@Data

  # grab equations
  iv_list <- object@internal$eqs

  # nblocks
  nblocks <- object@pta$nblocks

  table <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    eqs <- iv_list[[b]]
    lhs <- sapply(eqs, "[[", "lhs")
    rhs <- sapply(lapply(eqs, "[[", "rhs"), paste, collapse = " + ")
    lhs_new <- sapply(eqs, "[[", "lhs_new")
    rhs_new <- sapply(lapply(eqs, "[[", "rhs_new"), paste, collapse = " + ")
    iv <- sapply(lapply(eqs, "[[", "iv"), paste, collapse = ", ")
    type <- sapply(eqs, "[[", "iv_type")
    table[[b]] <- data.frame(
      lhs = lhs, rhs = rhs,
      lhs.new = lhs_new, rhs.new = rhs_new, type = type, instruments = iv
    )
    class(table[[b]]) <- c("lavaan.data.frame", "data.frame")
  }

  # return value
  return_value <- table

  # drop list?
  if (lavmodel@ngroups == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (!is.null(lavdata)) {
    if (length(lavdata@group.label) > 0L) {
      names(return_value) <- unlist(lavdata@group.label)
    }
  }

  return_value
}

lav_inspect_eqs <- function(object, drop_list_single_group = FALSE) {

  if (is.null(object@internal$eqs)) {
    lav_msg_stop(gettext("no equations/ivs found"))
  }
  lavmodel <- object@Model
  lavdata <- object@Data

  # grab equations
  eqs <- object@internal$eqs

  # return value
  return_value <- eqs

  # drop list?
  if (lavmodel@ngroups == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (!is.null(lavdata)) {
    if (length(lavdata@group.label) > 0L) {
      names(return_value) <- unlist(lavdata@group.label)
    }
  }

  return_value
}

lav_inspect_sargan <- function(object, drop_list_single_group = FALSE) {

  if (is.null(object@internal$eqs)) {
    lav_msg_stop(gettext("no equations/ivs found"))
  }
  lavmodel <- object@Model
  lavdata <- object@Data

  # grab equations
  iv_list <- object@internal$eqs

  # decide which overidentification test to report:
  #  - categorical (polychoric) data: the classical Sargan test is not valid
  #    for polychoric correlations, so report only the robust Browne
  #    residual-based (RLS) test (computed with the polychoric ACOV)
  #  - continuous data with the default test (browne.residual.nt): report only
  #    the classical Sargan test
  #  - continuous data where the user switched to test = browne.residual.adf:
  #    report only the (distribution-free) Browne residual-based test
  # see lav_sem_miiv_browne_test()
  test <- object@Options$test
  show_browne <- lavmodel@categorical ||
    any(grepl("browne.residual.adf", test, fixed = TRUE))

  # nblocks
  nblocks <- object@pta$nblocks

  # multiple-comparison adjustment for the per-equation Sargan p-values
  # (applied within each block, over the overidentified equations)
  adjust <- object@Options$estimator.args[["iv_sargan_adjust"]]
  if (is.null(adjust)) {
    adjust <- "none"
  }

  # helper to pull a named element from a per-equation test vector (the
  # 'browne' slot may be absent in objects fitted by older versions)
  pull <- function(eqs, slot, what) {
    sapply(seq_along(eqs), function(x) {
      v <- eqs[[x]][[slot]]
      if (is.null(v)) as.numeric(NA) else v[what]
    })
  }

  table <- vector("list", length = nblocks)
  for (b in seq_len(nblocks)) {
    eqs <- iv_list[[b]]
    lhs <- sapply(eqs, "[[", "lhs")
    rhs <- sapply(lapply(eqs, "[[", "rhs"), paste, collapse = " + ")
    sargan_stat <- pull(eqs, "sargan", "stat")
    sargan_df <- pull(eqs, "sargan", "df")
    sargan_pvalue <- pull(eqs, "sargan", "pvalue")
    browne_stat <- pull(eqs, "browne", "stat")
    browne_pvalue <- pull(eqs, "browne", "pvalue")
    if (show_browne) {
      table[[b]] <- data.frame(
        lhs = lhs, rhs = rhs,
        df = sargan_df,
        browne.stat = browne_stat, browne.pval = browne_pvalue
      )
    } else {
      table[[b]] <- data.frame(
        lhs = lhs, rhs = rhs,
        df = sargan_df,
        sargan.stat = sargan_stat, sargan.pval = sargan_pvalue
      )
    }

    # remove rows for which the test was not computed (just-identified
    # equations, where the statistic is NA)
    na_idx <- which(is.na(sargan_stat) & is.na(browne_stat))
    if (length(na_idx) > 0L) {
      table[[b]] <- table[[b]][-na_idx, , drop = FALSE]
    }

    # adjusted p-values over the (overidentified) equations in this block,
    # applied to the p-values of the reported test
    if (adjust != "none" && nrow(table[[b]]) > 0L) {
      if (show_browne) {
        table[[b]]$browne.pval.adj <- stats::p.adjust(
          table[[b]]$browne.pval, method = adjust)
      } else {
        table[[b]]$sargan.pval.adj <- stats::p.adjust(
          table[[b]]$sargan.pval, method = adjust)
      }
    }

    class(table[[b]]) <- c("lavaan.data.frame", "data.frame")
  }

  # return value
  return_value <- table

  # drop list?
  if (lavmodel@ngroups == 1L && drop_list_single_group) {
    return_value <- return_value[[1]]
  } else if (!is.null(lavdata)) {
    if (length(lavdata@group.label) > 0L) {
      names(return_value) <- unlist(lavdata@group.label)
    }
  }

  return_value
}

# the multivariate normal distribution + missing values
#  (so-called 'FIML')

# 1) loglikelihood (from raw data, or sample statistics)
# 2) derivatives with respect to mu, Sigma, vech(Sigma)
# 3) casewise scores with respect to mu, vech(Sigma), mu + vech(Sigma)
# 4) hessian of mu + vech(Sigma)
# 5) (unit) information of mu + vech(Sigma)
#    5a: (unit)    expected information
#    5b: (unit)    observed information
#    5c: (unit) first.order information
#    5d: lav_mvn_mi_info_both (both observed + first.order)

# 6) inverted information h0 mu + vech(Sigma)
#    6a: /
#    6b: /
#    6c: /
# 7) ACOV h0 mu + vech(Sigma)
#    7a: 1/N * inverted expected    information
#    7b: 1/N * inverted observed    information
#    7c: 1/N * inverted first-order information
#    7d: sandwich acov

# 10) additional functions
#      - lav_mvnorm_missing_impute_pattern
#      - lav_mvnorm_missing_estep


# YR 09 Feb 2016: first version
# YR 19 Mar 2017: 10)
# YR 01 Jul 2018: first_order functions gain cluster.idx= argument
# YR 03 Okt 2018: a few functions gain a wt= argument
# YR 03 Jul 2026: use shared kernels from lav_mvnorm_kernels.R


# 1) likelihood

# 1a: input is raw data
#  - two strategies: 1) using missing patterns (pattern = TRUE)
#                    2) truly case per case    (pattern = FALSE)
#    depending on the sample size, missing patterns, etc... one can be
#    (much) faster than the other
lav_mvn_mi_loglik_data <- function(y = NULL,
                                           mu = NULL,
                                           wt = NULL,
                                           sigma_1 = NULL,
                                           x_idx = integer(0L),
                                           casewise = FALSE,
                                           pattern = TRUE,
                                           sinv_method = "eigen",
                                           log2pi = TRUE,
                                           minus_two = FALSE) {
  if (pattern) {
    llik <- lav_mvn_mi_llik_pattern(
      y = y, wt = wt, mu = mu,
      sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
      log2pi = log2pi, minus_two = minus_two
    )
  } else {
    llik <- lav_mvn_mi_llik_casewise(
      y = y, wt = wt, mu = mu,
      sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
      log2pi = log2pi, minus_two = minus_two
    )
  }

  if (casewise) {
    loglik <- llik
  } else {
    loglik <- sum(llik, na.rm = TRUE)
  }

  loglik
}

# 1b: input are sample statistics (mean, cov, N) per pattern
lav_mvn_mi_loglik_samp <- function(yp = NULL,
                                                  mu = NULL,
                                                  sigma_1 = NULL,
                                                  x_idx = integer(0L),
                                                  x_mean = NULL,
                                                  x_cov = NULL,
                                                  sinv_method = "eigen",
                                                  log2pi = TRUE,
                                                  minus_two = FALSE) {
  log_2pi <- log(2 * pi)
  pat_n <- length(yp)
  p_1 <- length(yp[[1]]$var.idx)

  # global inverse + logdet
  sigma_inv_1 <- lav_mat_sym_inverse(
    s = sigma_1, logdet = TRUE,
    sinv_method = sinv_method
  )
  sigma_logdet <- attr(sigma_inv_1, "logdet")

  # DIST/logdet per pattern
  dist_1 <- logdet <- p_log_2pi <- numeric(pat_n)

  # for each pattern, compute sigma.inv/logdet; compute DIST for all
  # observations of this pattern
  for (p in seq_len(pat_n)) {
    # observed variables for this pattern
    var_idx <- yp[[p]]$var.idx

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # constant
    p_log_2pi[p] <- sum(var_idx) * log_2pi * yp[[p]]$freq

    # invert Sigma for this pattern
    sigma_inv <- lav_mvn_mi_pattern_inv(
      sigma_inv = sigma_inv_1, na_idx = na_idx,
      logdet = TRUE, sigma_logdet = sigma_logdet
    )
    logdet[p] <- attr(sigma_inv, "logdet") * yp[[p]]$freq

    tt <- yp[[p]]$SY + tcrossprod(yp[[p]]$MY - mu[var_idx])
    dist_1[p] <- sum(sigma_inv * tt) * yp[[p]]$freq
  }

  # loglikelihood all data
  if (log2pi) {
    loglik <- sum(-(p_log_2pi + logdet + dist_1) / 2)
  } else {
    loglik <- sum(-(logdet + dist_1) / 2)
  }

  if (minus_two) {
    loglik <- -2 * loglik
  }

  # x.idx
  if (length(x_idx) > 0L) {
    # conditional loglikelihood (fixed.x = TRUE): subtract the marginal
    # loglikelihood of the x (exogenous) variables. This marginal must be
    # computed using FIML over the missing-data patterns of the x variables
    # themselves; using the complete-data formula (as we did before) is wrong
    # whenever the x variables contain missing values (eg missing = "fiml.x"),
    # as it ignores their missingness (see GitHub issue #226).
    # Note: mu[x_idx]/Sigma[x_idx, x_idx] equal the (model-implied) moments of
    # the x variables, so we do not need x_mean/x_cov.
    xp <- vector("list", length = length(yp))
    for (p in seq_len(length(yp))) {
      var_idx_full <- yp[[p]]$var.idx
      # which x variables are observed in this pattern?
      x_obs <- var_idx_full[x_idx]
      # skip patterns where no x variable is observed (no contribution)
      if (!any(x_obs)) {
        next
      }
      # positions of the observed x variables within the observed block
      obs_full <- which(var_idx_full)
      pos_in_obs <- match(x_idx[x_obs], obs_full)
      if (yp[[p]]$freq > 1L) {
        sy_x <- yp[[p]]$SY[pos_in_obs, pos_in_obs, drop = FALSE]
      } else {
        sy_x <- 0
      }
      xp[[p]] <- list(
        SY = sy_x, MY = yp[[p]]$MY[pos_in_obs],
        var.idx = x_obs, freq = yp[[p]]$freq
      )
    }
    # drop the (skipped) empty patterns
    xp <- xp[!vapply(xp, is.null, logical(1L))]

    loglik_x <- lav_mvn_mi_loglik_samp(
      yp = xp,
      mu = mu[x_idx],
      sigma_1 = sigma_1[x_idx, x_idx, drop = FALSE],
      x_idx = integer(0L),
      sinv_method = sinv_method,
      log2pi = log2pi,
      minus_two = minus_two
    )

    loglik <- loglik - loglik_x
  }

  loglik
}

## casewise loglikelihoods

# casewise Sinv.method
lav_mvn_mi_llik_casewise <- function(y = NULL,
                                             wt = NULL,
                                             mu = NULL,
                                             sigma_1 = NULL,
                                             x_idx = integer(0L),
                                             sinv_method = "eigen",
                                             log2pi = TRUE,
                                             minus_two = FALSE) {
  p <- NCOL(y)
  log_2pi <- log(2 * pi)
  mu <- as.numeric(mu)
  ny <- NROW(y)

  # global inverse + logdet
  sigma_inv_1 <- lav_mat_sym_inverse(
    s = sigma_1, logdet = TRUE,
    sinv_method = sinv_method
  )
  sigma_logdet <- attr(sigma_inv_1, "logdet")

  # subtract Mu
  yc <- t(t(y) - mu)

  # DIST/logdet per case
  dist_1 <- logdet <- p_log_2pi <- rep(as.numeric(NA), ny)

  # missing pattern per case
  obs <- !is.na(y)
  p_i <- rowSums(obs)

  # constant
  p_log_2pi <- p_i * log_2pi

  # complete cases first (only an advantage if we have mostly complete
  # observations)
  other_idx <- seq_len(ny)
  complete_idx <- which(p_i == p)
  if (length(complete_idx) > 0L) {
    other_idx <- other_idx[-complete_idx]
    dist_1[complete_idx] <-
      rowSums(yc[complete_idx, , drop = FALSE] %*% sigma_inv_1 *
        yc[complete_idx, , drop = FALSE])
    logdet[complete_idx] <- sigma_logdet
  }

  # non-complete cases
  for (i in other_idx) {
    na_idx <- which(!obs[i, ])

    # catch empty cases
    if (length(na_idx) == p) next

    # invert Sigma for this pattern
    sigma_inv <- lav_mat_sym_inverse_update(
      s_inv = sigma_inv_1,
      rm_idx = na_idx, logdet = TRUE, s_logdet = sigma_logdet
    )
    logdet[i] <- attr(sigma_inv, "logdet")

    # distance for this case
    dist_1[i] <- sum(sigma_inv * crossprod(yc[i, obs[i, ], drop = FALSE]))
  }

  # compute casewise loglikelihoods
  if (log2pi) {
    llik <- -(p_log_2pi + logdet + dist_1) / 2
  } else {
    llik <- -(logdet + dist_1) / 2
  }

  # minus.two
  if (minus_two) {
    llik <- -2 * llik
  }

  # weights?
  if (!is.null(wt)) {
    llik <- llik * wt
  }

  # x.idx
  if (length(x_idx) > 0L) {
    llik_x <- lav_mvn_mi_llik_casewise(
      y = y[, x_idx, drop = FALSE],
      wt = wt, mu = mu[x_idx],
      sigma_1 = sigma_1[x_idx, x_idx, drop = FALSE],
      x_idx = integer(0L), sinv_method = sinv_method,
      log2pi = log2pi, minus_two = minus_two
    )
    llik <- llik - llik_x
  }

  llik
}

# pattern-based, but casewise loglikelihoods
lav_mvn_mi_llik_pattern <- function(y = NULL,
                                            mp = NULL,
                                            wt = NULL,
                                            mu = NULL,
                                            sigma_1 = NULL,
                                            x_idx = integer(0L),
                                            sinv_method = "eigen",
                                            log2pi = TRUE,
                                            minus_two = FALSE) {
  p_1 <- NCOL(y)
  log_2pi <- log(2 * pi)
  mu <- as.numeric(mu)
  ny <- NROW(y)

  # global inverse + logdet
  sigma_inv_1 <- lav_mat_sym_inverse(
    s = sigma_1, logdet = TRUE,
    sinv_method = sinv_method
  )
  sigma_logdet <- attr(sigma_inv_1, "logdet")

  # subtract Mu
  yc <- t(t(y) - mu)

  # DIST/logdet per case
  dist_1 <- logdet <- p_log_2pi <- rep(as.numeric(NA), ny)

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # for each pattern, compute sigma.inv/logdet; compute DIST for all
  # observations of this pattern
  for (p in seq_len(mp$npatterns)) {
    # observed values for this pattern
    var_idx <- mp$pat[p, ]

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # identify cases with this pattern
    case_idx <- mp$case.idx[[p]]

    # constant
    p_log_2pi[case_idx] <- sum(var_idx) * log_2pi

    # invert Sigma for this pattern
    sigma_inv <- lav_mvn_mi_pattern_inv(
      sigma_inv = sigma_inv_1, na_idx = na_idx,
      logdet = TRUE, sigma_logdet = sigma_logdet
    )
    logdet[case_idx] <- attr(sigma_inv, "logdet")

    if (mp$freq[p] == 1L) {
      dist_1[case_idx] <- sum(sigma_inv *
        crossprod(yc[case_idx, var_idx, drop = FALSE]))
    } else {
      dist_1[case_idx] <-
        rowSums(yc[case_idx, var_idx, drop = FALSE] %*% sigma_inv *
          yc[case_idx, var_idx, drop = FALSE])
    }
  }

  # compute casewise loglikelihoods
  if (log2pi) {
    llik <- -(p_log_2pi + logdet + dist_1) / 2
  } else {
    llik <- -(logdet + dist_1) / 2
  }

  # minus.two
  if (minus_two) {
    llik <- -2 * llik
  }

  # weights?
  if (!is.null(wt)) {
    llik <- llik * wt
  }

  # x.idx -- using casewise (as patterns for Y may not be the same as
  #                          patterns for Y[,-x.idx])
  if (length(x_idx) > 0L) {
    llik_x <- lav_mvn_mi_llik_casewise(
      y = y[, x_idx, drop = FALSE],
      wt = wt, mu = mu[x_idx],
      sigma_1 = sigma_1[x_idx, x_idx, drop = FALSE],
      x_idx = integer(0L), sinv_method = sinv_method,
      log2pi = log2pi, minus_two = minus_two
    )
    llik <- llik - llik_x
  }

  llik
}




# 2. Derivatives

# 2a: derivative logl with respect to mu
lav_mvn_mi_dlogl_dmu <- function(y = NULL,
                                         wt = NULL,
                                         mu = NULL,
                                         sigma_1 = NULL,
                                         x_idx = integer(0L),
                                         sigma_inv = NULL,
                                         sinv_method = "eigen") {
  sc <- lav_mvn_mi_sc_mu(
    y = y, wt = wt, mu = mu, sigma_1 = sigma_1,
    x_idx = x_idx, sigma_inv = sigma_inv, sinv_method = sinv_method
  )

  colSums(sc, na.rm = TRUE)
}

# 2abis: using samplestats
lav_mvn_mi_dlogl_dmu_samp <- function(yp = NULL,
                                                     mu = NULL,
                                                     sigma_1 = NULL,
                                                     x_idx = integer(0L),
                                                     sigma_inv = NULL,
                                                     sinv_method = "eigen") {
  pat_n <- length(yp)
  p_1 <- length(yp[[1]]$var.idx)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # dmu
  dmu <- numeric(p_1)

  # for each pattern, compute sigma.inv
  for (p in seq_len(pat_n)) {
    # observed variables for this pattern
    var_idx <- yp[[p]]$var.idx

    # invert Sigma for this pattern
    sigma_inv_1 <- lav_mvn_mi_pattern_inv(
      sigma_inv = sigma_inv, na_idx = which(!var_idx)
    )

    # dmu for this pattern
    dmu_pattern <- as.numeric(sigma_inv_1 %*% (yp[[p]]$MY - mu[var_idx]))

    # update mu
    dmu[var_idx] <- dmu[var_idx] + (dmu_pattern * yp[[p]]$freq)
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    dmu[x_idx] <- 0
  }

  dmu
}



# 2b: derivative logl with respect to Sigma (full matrix, ignoring symmetry)
lav_mvn_mi_dlogl_dsigma <- function(m_y = NULL,
                                            l_mp = NULL,
                                            wt = NULL,
                                            m_mu = NULL,
                                            m_sigma = NULL,
                                            x_idx = integer(0L),
                                            sigma_inv = NULL,
                                            sinv_method = "eigen") {
  # missing patterns
  if (is.null(l_mp)) {
    l_mp <- lav_data_mi_patterns(m_y)
  }

  # per-pattern sample statistics
  yp <- lav_samp_mi_patterns(y = m_y, mp = l_mp, wt = wt)

  lav_mvn_mi_dlogl_dsigma_samp(
    yp = yp, mu = as.numeric(m_mu), sigma_1 = m_sigma,
    x_idx = x_idx, sigma_inv = sigma_inv, sinv_method = sinv_method
  )
}

# 2bbis: using samplestats
lav_mvn_mi_dlogl_dsigma_samp <- function(yp = NULL,
                                                        mu = NULL,
                                                        sigma_1 = NULL,
                                                        x_idx = integer(0L),
                                                        sigma_inv = NULL,
                                                        sinv_method = "eigen") {
  pat_n <- length(yp)
  p_1 <- length(yp[[1]]$var.idx)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # dvechSigma
  d_sigma <- matrix(0, p_1, p_1)

  # for each pattern
  for (p in seq_len(pat_n)) {
    # observed variables for this pattern
    var_idx <- yp[[p]]$var.idx

    # invert Sigma for this pattern
    sigma_inv_1 <- lav_mvn_mi_pattern_inv(
      sigma_inv = sigma_inv, na_idx = which(!var_idx)
    )

    w_tilde <- yp[[p]]$SY + tcrossprod(yp[[p]]$MY - mu[var_idx])

    # dSigma for this pattern
    d_sigma_pattern <- matrix(0, p_1, p_1)
    d_sigma_pattern[var_idx, var_idx] <- -(1 / 2) * (sigma_inv_1 -
      (sigma_inv_1 %*% w_tilde %*% sigma_inv_1))

    # update dSigma
    d_sigma <- d_sigma + (d_sigma_pattern * yp[[p]]$freq)
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    d_sigma[x_idx, x_idx] <- 0
  }

  d_sigma
}


# 2c: derivative logl with respect to vech(Sigma)
lav_mvn_mi_dlogl_dvechsigma <- function(y = NULL,
                                                wt = NULL,
                                                mu = NULL,
                                                x_idx = integer(0L),
                                                sigma_1 = NULL,
                                                sigma_inv = NULL,
                                                sinv_method = "eigen") {
  d_sigma <- lav_mvn_mi_dlogl_dsigma(
    m_y = y, wt = wt, m_mu = mu,
    m_sigma = sigma_1, x_idx = x_idx, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  dvech_sigma <- as.numeric(lav_mat_dup_pre(
    as.matrix(lav_mat_vec(d_sigma))
  ))

  dvech_sigma
}

# 2cbis: using samplestats
lav_mvnorm_missing_dlogl_dvechsigma_samplestats <- # nolint
  function(yp = NULL,
           mu = NULL,
           sigma_1 = NULL,
           x_idx = integer(0L),
           sigma_inv = NULL,
           sinv_method = "eigen") {
    pat_n <- length(yp)
    p_1 <- length(yp[[1]]$var.idx)

    sigma_inv <- lav_mvn_sigma_inv(
      sigma_1 = sigma_1, sigma_inv = sigma_inv,
      sinv_method = sinv_method
    )

    # dvechSigma
    dvech_sigma <- numeric(p_1 * (p_1 + 1) / 2)

    # for each pattern
    for (p in seq_len(pat_n)) {
      # observed variables for this pattern
      var_idx <- yp[[p]]$var.idx

      # invert Sigma for this pattern
      sigma_inv_1 <- lav_mvn_mi_pattern_inv(
        sigma_inv = sigma_inv, na_idx = which(!var_idx)
      )

      w_tilde <- yp[[p]]$SY + tcrossprod(yp[[p]]$MY - mu[var_idx])

      # dSigma for this pattern
      d_sigma_pattern <- matrix(0, p_1, p_1)
      d_sigma_pattern[var_idx, var_idx] <- -(1 / 2) * (sigma_inv_1 -
        (sigma_inv_1 %*% w_tilde %*% sigma_inv_1))

      # fixed.x?
      if (length(x_idx) > 0L) {
        d_sigma_pattern[x_idx, x_idx] <- 0
      }

      # convert to vechSigma
      dvech_sigma_pattern <- as.numeric(lav_mat_dup_pre(
        as.matrix(lav_mat_vec(d_sigma_pattern))
      ))

      # update dvechSigma
      dvech_sigma <- dvech_sigma + (dvech_sigma_pattern * yp[[p]]$freq)
    }

    dvech_sigma
  }




# 3. Casewise scores

# 3a: casewise scores with respect to mu
lav_mvn_mi_sc_mu <- function(y = NULL,
                                         wt = NULL,
                                         mp = NULL,
                                         mu = NULL,
                                         sigma_1 = NULL,
                                         x_idx = integer(0L),
                                         sigma_inv = NULL,
                                         sinv_method = "eigen") {
  p_1 <- NCOL(y)
  mu <- as.numeric(mu)
  ny <- NROW(y)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # subtract Mu
  yc <- t(t(y) - mu)

  # dmu per case
  dmu <- matrix(as.numeric(NA), ny, p_1)

  # for each pattern, compute sigma.inv
  for (p in seq_len(mp$npatterns)) {
    # observed values for this pattern
    var_idx <- mp$pat[p, ]

    case_idx <- mp$case.idx[[p]]

    # invert Sigma for this pattern
    sigma_inv_1 <- lav_mvn_mi_pattern_inv(
      sigma_inv = sigma_inv, na_idx = which(!var_idx)
    )

    # compute dMu for all observations of this pattern
    dmu[case_idx, var_idx] <-
      yc[case_idx, var_idx, drop = FALSE] %*% sigma_inv_1
  }

  # weights
  if (!is.null(wt)) {
    dmu <- dmu * wt
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    dmu[, x_idx] <- 0
  }

  dmu
}

# 3b: casewise scores with respect to vech(Sigma)
lav_mvn_mi_sc_sigma <- function(y = NULL,
                                                 wt = NULL,
                                                 mp = NULL,
                                                 mu = NULL,
                                                 sigma_1 = NULL,
                                                 x_idx = integer(0L),
                                                 sigma_inv = NULL,
                                                 sinv_method = "eigen") {
  p_1 <- NCOL(y)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # score engine
  sc <- lav_mvn_mi_sc_engine(
    y = y, mp = mp, mu = mu,
    sigma_inv = sigma_inv
  )$sc

  # weights
  if (!is.null(wt)) {
    sc <- sc * wt
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    sc[, lav_mat_vech_which_idx(n = p_1, idx = x_idx)] <- 0
  }

  sc
}

# 3c: casewise scores with respect to mu + vech(Sigma)
lav_mvn_mi_sc_mu_sigma <- function(y = NULL,
                                                    mp = NULL,
                                                    wt = NULL,
                                                    mu = NULL,
                                                    sigma_1 = NULL,
                                                    x_idx = integer(0L),
                                                    sigma_inv = NULL,
                                                    sinv_method = "eigen") {
  p_1 <- NCOL(y)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # score engine
  eng <- lav_mvn_mi_sc_engine(
    y = y, mp = mp, mu = mu,
    sigma_inv = sigma_inv
  )

  out <- cbind(eng$dmu, eng$sc)

  # weights
  if (!is.null(wt)) {
    out <- out * wt
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    not_x <- lav_mat_vech_which_idx(n = p_1, idx = x_idx,
                                       #diagonal = !correlation,
                                       add_idx_at_start = TRUE)
    out[, not_x] <- 0
  }

  out
}


# 4) Hessian of logl
lav_mvn_mi_logl_hessian_data <- function(y = NULL,
                                                 mp = NULL,
                                                 wt = NULL,
                                                 mu = NULL,
                                                 sigma_1 = NULL,
                                                 x_idx = integer(0L),
                                                 sinv_method = "eigen",
                                                 sigma_inv = NULL) {
  # missing patterns
  yp <- lav_samp_mi_patterns(y = y, mp = mp, wt = wt)

  lav_mvnorm_missing_logl_hessian_samplestats(
    yp = yp, mu = mu,
    sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
    sigma_inv = sigma_inv
  )
}

lav_mvnorm_missing_logl_hessian_samplestats <- # nolint
  function(yp = NULL,
           # wt not needed
           mu = NULL,
           sigma_1 = NULL,
           x_idx = integer(0L),
           sinv_method = "eigen",
           sigma_inv = NULL) {
    pat_n <- length(yp)
    p_1 <- length(yp[[1]]$var.idx)

    sigma_inv <- lav_mvn_sigma_inv(
      sigma_1 = sigma_1, sigma_inv = sigma_inv,
      sinv_method = sinv_method
    )

    h11 <- matrix(0, p_1, p_1)
    h21 <- matrix(0, p_1 * (p_1 + 1) / 2, p_1)
    h22 <- matrix(0, p_1 * (p_1 + 1) / 2, p_1 * (p_1 + 1) / 2)

    # for each pattern, compute sigma.inv
    for (p in seq_len(pat_n)) {
      # observed variables
      var_idx <- yp[[p]]$var.idx
      pat_freq <- yp[[p]]$freq

      # invert Sigma for this pattern
      sigma_inv_1 <- lav_mvn_mi_pattern_inv(
        sigma_inv = sigma_inv, na_idx = which(!var_idx)
      )

      # observed-information blocks for this pattern
      info <- lav_mvn_mi_info_obs_pattern(
        sigma_inv_1 = sigma_inv_1, var_idx = var_idx, p_1 = p_1,
        my = yp[[p]]$MY, sy = yp[[p]]$SY, mu = mu, sigma_1 = sigma_1
      )

      h11 <- h11 + pat_freq * info$i11
      h21 <- h21 + pat_freq * info$i21
      h22 <- h22 + pat_freq * info$i22
    }

    h12 <- t(h21)

    out <- -1 * rbind(
      cbind(h11, h12),
      cbind(h21, h22)
    )

    # fixed.x?
    out <- lav_mvn_zero_x_idx(out,
      n = p_1, x_idx = x_idx,
      meanstructure = TRUE
    )

    out
  }




# 5) Information

# 5a: expected unit information Mu and vech(Sigma)
#     (only useful under MCAR)
# (old term: Abeta, expected)
lav_mvn_mi_info_expected <- function(y = NULL,
                                                    mp = NULL,
                                                    wt = NULL,
                                                    mu = NULL, # unused
                                                    sigma_1 = NULL,
                                                    x_idx = integer(0L),
                                                    sigma_inv = NULL,
                                                    sinv_method = "eigen") {
  p_1 <- NCOL(y)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # N
  n <- lav_mvn_mi_nobs(mp = mp, wt = wt)

  i11 <- matrix(0, p_1, p_1)
  i22 <- matrix(0, p_1 * (p_1 + 1) / 2, p_1 * (p_1 + 1) / 2)

  # for each pattern, compute sigma.inv
  for (p in seq_len(mp$npatterns)) {
    # observed variables
    var_idx <- mp$pat[p, ]

    # invert Sigma for this pattern
    sigma_inv_1 <- lav_mvn_mi_pattern_inv(
      sigma_inv = sigma_inv, na_idx = which(!var_idx)
    )

    s_inv <- matrix(0, p_1, p_1)
    s_inv[var_idx, var_idx] <- sigma_inv_1

    s2_inv <- lav_mvn_kron_dup_half(s_inv)

    freq <- lav_mvn_mi_pattern_freq(mp = mp, p = p, wt = wt)

    i11 <- i11 + freq * s_inv
    i22 <- i22 + freq * s2_inv
  }

  out <- lav_mat_bdiag(i11, i22) / n

  # fixed.x?
  out <- lav_mvn_zero_x_idx(out,
    n = p_1, x_idx = x_idx,
    meanstructure = TRUE
  )

  out
}

# 5b: unit observed information Mu and vech(Sigma) from raw data
# (old term: Abeta, observed)
lav_mvn_mi_info_observed_data <- function(y = NULL,
                                                         mp = NULL,
                                                         wt = NULL,
                                                         mu = NULL,
                                                         sigma_1 = NULL,
                                                         x_idx = integer(0L),
                                                         sinv_method = "eigen",
                                                         sigma_inv = NULL) {
  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # N
  n <- lav_mvn_mi_nobs(mp = mp, wt = wt)

  # observed information
  observed <- lav_mvn_mi_logl_hessian_data(
    y = y, mp = mp, wt = wt,
    mu = mu, sigma_1 = sigma_1, x_idx = x_idx,
    sinv_method = sinv_method, sigma_inv = sigma_inv
  )

  -observed / n
}

# 5b-bis: unit observed information Mu and vech(Sigma) from samplestats
lav_mvnorm_missing_information_observed_samplestats <- # nolint
  function(yp = NULL,
           # wt not needed
           mu = NULL,
           sigma_1 = NULL,
           x_idx = integer(0L),
           sinv_method = "eigen",
           sigma_inv = NULL) {
    n <- sum(sapply(yp, "[[", "freq")) # implicitly: removed empty cases!

    # observed information
    observed <- lav_mvnorm_missing_logl_hessian_samplestats(
      yp = yp, mu = mu,
      sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
      sigma_inv = sigma_inv
    )

    -observed / n
  }

# 5c: unit first-order information Mu and vech(Sigma) from raw data
# (old term: Bbeta)
lav_mvn_mi_info_firstorder <- function(y = NULL,
                                                      mp = NULL,
                                                      wt = NULL,
                                                      cluster_idx = NULL,
                                                      mu = NULL,
                                                      sigma_1 = NULL,
                                                      x_idx = integer(0L),
                                                      sinv_method = "eigen",
                                                      sigma_inv = NULL) {
  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # N
  n <- lav_mvn_mi_nobs(mp = mp, wt = wt)

  sc <- lav_mvn_mi_sc_mu_sigma(
    y = y, mp = mp, wt = wt,
    mu = mu, sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
    sigma_inv = sigma_inv
  )

  # handle clustering
  if (!is.null(cluster_idx)) {
    # take the sum within each cluster
    sc <- rowsum(sc, group = cluster_idx, reorder = FALSE, na.rm = TRUE)

    # lower bias is number of clusters is not very high
    n_c <- nrow(sc)
    correction_factor <- n_c / (n_c - 1)
    sc <- sc * sqrt(correction_factor)
  }

  lav_mat_crossprod(sc) / n
}

# 5d: both unit first-order information and expected/observed information
#     from raw data, in one go
lav_mvn_mi_info_both <- function(y = NULL,
                                                mp = NULL,
                                                wt = NULL,
                                                cluster_idx = NULL,
                                                mu = NULL,
                                                sigma_1 = NULL,
                                                x_idx = integer(0L),
                                                sinv_method = "eigen",
                                                sigma_inv = NULL,
                                                information = "observed") {
  p_1 <- NCOL(y)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  # N
  n <- lav_mvn_mi_nobs(mp = mp, wt = wt)

  # Abeta: expected or observed unit information
  if (information == "expected") {
    abeta <- lav_mvn_mi_info_expected(
      y = y, mp = mp, wt = wt,
      sigma_1 = sigma_1, x_idx = x_idx, sigma_inv = sigma_inv,
      sinv_method = sinv_method
    )
  } else {
    yp <- lav_samp_mi_patterns(y = y, mp = mp, wt = wt)
    abeta <- lav_mvnorm_missing_information_observed_samplestats(
      yp = yp, mu = mu,
      sigma_1 = sigma_1, x_idx = x_idx, sinv_method = sinv_method,
      sigma_inv = sigma_inv
    )
  }

  # Bbeta: first-order unit information (from the casewise scores)
  eng <- lav_mvn_mi_sc_engine(
    y = y, mp = mp, mu = mu,
    sigma_inv = sigma_inv
  )
  sc <- cbind(eng$dmu, eng$sc)

  # weights
  if (!is.null(wt)) {
    sc <- sc * wt
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    not_x <- lav_mat_vech_which_idx(n = p_1, idx = x_idx,
                                       #diagonal = !correlation,
                                       add_idx_at_start = TRUE)
    sc[, not_x] <- 0
  }

  # handle clustering
  if (!is.null(cluster_idx)) {
    # take the sum within each cluster
    sc <- rowsum(sc, group = cluster_idx, reorder = FALSE, na.rm = TRUE)

    # lower bias is number of clusters is not very high
    n_c <- nrow(sc)
    correction_factor <- n_c / (n_c - 1)
    sc <- sc * sqrt(correction_factor)
  }

  # first order information
  bbeta <- lav_mat_crossprod(sc) / n

  list(Abeta = abeta, Bbeta = bbeta)
}


# 6) inverted information h0 mu + vech(Sigma)

#    6a: (unit) inverted expected information
#    NOT USED: is not equal to solve(expected)
#    (although it does converge to the same solution eventually)

#    6b: /

#    6c: /


# 7) ACOV h0 mu + vech(Sigma)

#    7a: 1/N * inverted expected    information

#    7b: 1/N * inverted observed    information

#    7c: 1/N * inverted first-order information

#    7d: sandwich acov



# 10) other stuff

# single imputation missing cells, under the normal model, pattern-based
# FIXME: add wt
lav_mvn_mi_impute_pattern <- function(y = NULL,
                                              mp = NULL,
                                              mu = NULL,
                                              sigma_1 = NULL,
                                              sigma_inv = NULL,
                                              sinv_method = "eigen") {
  mu <- as.numeric(mu)

  # complete data
  y_complete <- y

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # subtract Mu
  yc <- t(t(y) - mu)

  # fill in data per pattern
  for (p in seq_len(mp$npatterns)) {
    # observed values for this pattern
    var_idx <- mp$pat[p, ]

    # if complete, nothing to do
    if (all(var_idx)) {
      next
    }

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # extract observed data for these (centered) cases
    oc <- yc[mp$case.idx[[p]], mp$pat[p, ], drop = FALSE]

    # invert Sigma (Sigma_22, observed part only) for this pattern
    sigma_22_inv <- try(
      lav_mat_sym_inverse_update(
        s_inv = sigma_inv, rm_idx = na_idx, logdet = FALSE
      ),
      silent = TRUE
    )
    if (inherits(sigma_22_inv, "try-error")) {
      lav_msg_stop(gettext("Sigma_22.inv cannot be inverted"))
    }

    # estimate missing values in this pattern
    sigma_12 <- sigma_1[!var_idx, var_idx, drop = FALSE]
    y_missing <- t(sigma_12 %*% sigma_22_inv %*% t(oc) + mu[!var_idx])

    # complete data for this pattern
    y_complete[mp$case.idx[[p]], !var_idx] <- y_missing
  }

  y_complete
}


# E-step: expectations of sum, sum of squares, sum of crossproducts
# plus correction
lav_mvn_mi_estep <- function(y = NULL,
                                     mp = NULL,
                                     wt = NULL,
                                     mu = NULL,
                                     sigma_1 = NULL,
                                     sigma_inv = NULL,
                                     sinv_method = "eigen") {
  p_1 <- NCOL(y)
  mu <- as.numeric(mu)

  # missing patterns
  if (is.null(mp)) {
    mp <- lav_data_mi_patterns(y)
  }

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # T1, T2
  t1 <- numeric(p_1)
  t2 <- matrix(0, p_1, p_1)

  # update T1 and T2 per pattern
  for (p in seq_len(mp$npatterns)) {
    # observed values for this pattern
    var_idx <- mp$pat[p, ]

    # extract observed data
    o <- y[mp$case.idx[[p]], mp$pat[p, ], drop = FALSE]

    # if complete, just compute first and second moments
    if (all(var_idx)) {
      if (!is.null(wt)) {
        wt_1 <- wt[mp$case.idx[[p]]]
        t1 <- t1 + colSums(wt_1 * o)
        t2 <- t2 + crossprod(sqrt(wt_1) * o)
      } else {
        # complete pattern
        t1 <- t1 + colSums(o)
        t2 <- t2 + crossprod(o)
      }
      next
    }

    # missing values for this pattern
    na_idx <- which(!var_idx)

    # partition Sigma (1=missing, 2=complete)
    sigma_11 <- sigma_1[!var_idx, !var_idx, drop = FALSE]
    sigma_12 <- sigma_1[!var_idx, var_idx, drop = FALSE]
    sigma_21 <- sigma_1[var_idx, !var_idx, drop = FALSE]

    # invert Sigma (Sigma_22, observed part only) for this pattern
    sigma_22_inv <- try(
      lav_mat_sym_inverse_update(
        s_inv = sigma_inv, rm_idx = na_idx, logdet = FALSE
      ),
      silent = TRUE
    )
    if (inherits(sigma_22_inv, "try-error")) {
      lav_msg_stop(gettext("Sigma_22.inv cannot be inverted"))
    }

    # estimate missing values in this pattern
    oc <- t(t(o) - mu[var_idx])
    y_missing <- t(sigma_12 %*% sigma_22_inv %*% t(oc) + mu[!var_idx])

    # complete data for this pattern
    y_complete <- matrix(0, mp$freq[[p]], p_1)
    y_complete[, var_idx] <- o
    y_complete[, !var_idx] <- y_missing

    if (!is.null(wt)) {
      wt_1 <- wt[mp$case.idx[[p]]]
      t1_pat <- colSums(wt_1 * y_complete)
      t2_pat <- crossprod(sqrt(wt_1) * y_complete)
    } else {
      # 1. SUM `completed' pattern
      t1_pat <- colSums(y_complete)

      # 2. CROSSPROD `completed' pattern
      t2_pat <- crossprod(y_complete)
    }

    # correction for missing cells: conditional covariances
    t2_p11 <- sigma_11 - (sigma_12 %*% sigma_22_inv %*% sigma_21)
    if (!is.null(wt)) {
      t2_pat[!var_idx, !var_idx] <-
        t2_pat[!var_idx, !var_idx] + (t2_p11 * sum(wt_1))
    } else {
      t2_pat[!var_idx, !var_idx] <-
        t2_pat[!var_idx, !var_idx] + (t2_p11 * mp$freq[[p]])
    }

    # accumulate
    t1 <- t1 + t1_pat
    t2 <- t2 + t2_pat
  }

  list(T1 = t1, T2 = t2)
}

# functions related to GFI and other 'absolute' fit indices

# lower-level functions:
# - lav_fit_gfi_lisrel    (the 'old' LISREL GFI)
# - lav_fit_agfi_lisrel
# - lav_fit_pgfi
# - lav_fit_gfi           (the 'new' GFI: Maydeu-Olivares et al. 2024)
# - lav_fit_gfi_ci

# higher-level functions:
# - lav_fit_gfi_lavobject

# Y.R. 21 July 2022
# Y.R. 24 June 2026: rename old gfi/agfi to gfi_lisrel/agfi_lisrel and add the
#                    'new' GFI (with confidence interval) of Maydeu-Olivares et
#                    al. (2024)

# ----------------------------------------------------------------------------
# the 'old' (LISREL) GFI and friends
# ----------------------------------------------------------------------------

# original formulas were given in Joreskog and Sorbom (1984) user's guide
# for LISREL VI (one for ML, and another for ULS)

# here we use the more 'general' formulas
# (generalized to allow for meanstructures etc)

# References:

# Mulaik, S. A., James, L. R., Van Alstine, J., Bennett, N., Lind, S., &
# Stilwell, C. D. (1989). Evaluation of goodness-of-fit indices for structural
# equation models. Psychological bulletin, 105(3), 430.

# Tanaka, J. S., & Huba, G. J. (1985). A fit index for covariance structure
# models under arbitrary GLS estimation. British Journal of Mathematical and
# Statistical Psychology, 38,197-201.

lav_fit_gfi_lisrel <- function(wls_obs = NULL, wls_est = NULL, wls_v = NULL,
                               nobs_1 = NULL) {
  # number of groups
  g_1 <- length(wls_obs)

  # compute gfi per group
  gfi_group <- numeric(g_1)
  for (g in 1:g_1) {
    wls_obs_1 <- wls_obs[[g]]
    wls_est_1 <- wls_est[[g]]
    wls_v_1 <- wls_v[[g]]

    if (is.null(wls_v_1)) {
      gfi_group[g] <- as.numeric(NA)
    } else {
      wls_diff <- wls_obs_1 - wls_est_1
      if (is.matrix(wls_v_1)) {
        # full weight matrix
        t1 <- crossprod(wls_diff, wls_v_1) %*% wls_diff
        t2 <- crossprod(wls_obs_1, wls_v_1) %*% wls_obs_1
      } else {
        # diagonal weight matrix
        t1 <- as.numeric(crossprod(wls_diff^2, wls_v_1))
        t2 <- as.numeric(crossprod(wls_obs_1^2, wls_v_1))
      }
      gfi_group[g] <- 1 - t1 / t2
    }
  }

  if (g_1 > 1) {
    ## CHECKME: get the scaling right
    nobs_1 <- unlist(nobs_1)
    gfi <- as.numeric((nobs_1 %*% gfi_group) / sum(nobs_1))
  } else {
    gfi <- gfi_group[1L]
  }

  gfi
}

# 'adjusted' GFI (adjusted for degrees of freedom)
lav_fit_agfi_lisrel <- function(gfi = NULL, nel = NULL, df = NULL) {
  if (!is.finite(gfi) || !is.finite(nel) || !is.finite(df)) {
    agfi <- as.numeric(NA)
  } else if (df > 0) {
    agfi <- 1 - (nel / df) * (1 - gfi)
  } else {
    agfi <- 1
  }

  agfi
}

# PGFI: parsimony goodness-of-fit index

# Mulaik, S. A., James, L. R., Van Alstine, J., Bennett, N., Lind, S., &
# Stilwell, C. D. (1989). Evaluation of goodness-of-fit indices for structural
# equation models. Psychological bulletin, 105(3), 430.

# LISREL formula (Simplis book 2002, p. 126)
lav_fit_pgfi <- function(gfi = NULL, nel = NULL, df = NULL) {
  if (!is.finite(gfi) || !is.finite(nel) || !is.finite(df)) {
    pgfi <- as.numeric(NA)
  } else if (nel == 0) {
    pgfi <- as.numeric(NA)
  } else {
    pgfi <- (df / nel) * gfi
  }

  pgfi
}

# ----------------------------------------------------------------------------
# the 'new' GFI (with confidence interval)
# ----------------------------------------------------------------------------

# Maydeu-Olivares, A., Ximenez, C., & Revuelta, J. (2024). Percentage of
# variance accounted for in structural equation models: The rediscovery of the
# goodness of fit index. Psychological Methods.

# The GFI is interpreted as the percentage of variance (of the sum of squared
# covariances) accounted for by the model; it is the SEM analogue of R^2 in
# regression (without an intercept), using the saturated model as baseline.

# point estimate (Maydeu-Olivares et al., 2024, Eq. 20):
#
#   GFI = min( N*p / (N*p + 2*c_hat*T - 2*c_hat*df), 1 )
#
# where T is the test statistic, p the number of observed variables, N the
# (total) sample size, df the degrees of freedom and c_hat a scaling constant.
#
# Maydeu-Olivares et al. (2024) use the reweighted least squares (RLS)
# statistic T_RLS (== 'browne.residual.nt.model' in lavaan), arguing it offers
# a more reliable assessment of exact fit than the likelihood ratio (LR)
# statistic. We compute the GFI from both the LR statistic (gfi_lrt) and the
# RLS statistic (gfi_rls); the (printed) 'gfi' is the RLS version.
#
# Following the implementation of the (robust) RMSEA in lavaan, we use the
# (unscaled) statistic X2 as 'c_hat*T' (since c_hat*T_M equals the unadjusted
# RLS/LR statistic) and apply c_hat only to the degrees of freedom:
#
#   GFI = min( N*p / (N*p + 2*X2 - 2*c_hat*df), 1 )
#
# - 'standard' (normal-theory, unbiased) version: c_hat = 1
# - 'robust' (to nonnormality) version: c_hat = scaling factor
#
# Note: the scaling constant c (and the shift b for scaled.shifted) is the same
# for T_LR and T_RLS (Maydeu-Olivares et al., 2024, Eq. 19), so the robust RLS
# version reuses the scaling factor(s) obtained for the LR statistic.
#
# Note: this matches the relationship GFI = p / (p + 2*df*RMSEA^2) (their
# Eq. 26), with the corresponding (standard or robust) RMSEA value.
#
# always using N (if a user needs N-1, just replace N by N-1)
lav_fit_gfi <- function(x2 = NULL, df = NULL, n = NULL, p = NULL,
                        c_hat = 1.0) {
  if (length(x2) == 0L || !is.finite(x2) || !is.finite(df) ||
    !is.finite(n) || !is.finite(p)) {
    return(as.numeric(NA))
  }

  np <- n * p
  denom <- np + 2 * x2 - 2 * c_hat * df
  if (!is.finite(denom) || denom <= 0) {
    return(as.numeric(NA))
  }

  gfi <- np / denom

  # see Eq. 20: GFI can not be larger than 1 (eg when the model fits 'too well')
  if (gfi > 1) {
    gfi <- 1
  }

  gfi
}

# confidence interval for the 'new' GFI (Maydeu-Olivares et al., 2024, Eq. 23)
#
# the confidence interval is obtained from the noncentrality parameter of the
# (noncentral) chi-square distribution (exactly as for the RMSEA), and then
# transformed to the GFI metric:
#
#   GFI_bound = min( N*p / (N*p + 2*c_hat*lambda), 1 )
#
# note: the *lower* noncentrality (lambda_l) yields the *upper* GFI bound, and
# the *upper* noncentrality (lambda_u) yields the *lower* GFI bound.
#
# note: for the 'robust' version, x2 should be the scaled test statistic
lav_fit_gfi_ci <- function(x2 = NULL, df = NULL, n = NULL, p = NULL,
                           c_hat = 1, level = 0.90) {
  if (length(x2) == 0L || !is.finite(x2) || !is.finite(df) ||
    !is.finite(n) || !is.finite(p)) {
    return(list(
      gfi.ci.lower = as.numeric(NA),
      gfi.ci.upper = as.numeric(NA)
    ))
  }

  if (!is.finite(level) || level < 0 || level > 1.0) {
    lav_msg_warn(gettextf(
      "invalid level value [%s] set to default 0.90.", level
    ))
    level <- 0.90
  }

  np <- n * p

  upper_perc <- (1 - (1 - level) / 2)
  lower_perc <- (1 - level) / 2

  # internal functions
  lower_lambda <- function(lambda) {
    (pchisq(x2, df = df, ncp = lambda) - upper_perc)
  }
  upper_lambda <- function(lambda) {
    (pchisq(x2, df = df, ncp = lambda) - lower_perc)
  }

  # transform a noncentrality parameter to the GFI metric
  gfi_from_lambda <- function(lambda) {
    if (!is.finite(lambda)) {
      return(as.numeric(NA))
    }
    val <- np / (np + 2 * c_hat * lambda)
    if (val > 1) {
      val <- 1
    }
    val
  }

  # lower noncentrality -> upper GFI bound
  if (df < 1 || lower_lambda(0) < 0.0) {
    gfi_ci_upper <- 1
  } else {
    lambda_l <- try(uniroot(f = lower_lambda, lower = 0, upper = x2)$root,
      silent = TRUE
    )
    if (inherits(lambda_l, "try-error")) {
      lambda_l <- as.numeric(NA)
    }
    gfi_ci_upper <- gfi_from_lambda(lambda_l)
  }

  # upper noncentrality -> lower GFI bound
  n_gfi <- max(n, x2 * 4)
  if (df < 1 || upper_lambda(n_gfi) > 0 || upper_lambda(0) < 0) {
    gfi_ci_lower <- 1
  } else {
    lambda_u <- try(
      uniroot(
        f = upper_lambda, lower = 0,
        upper = n_gfi
      )$root,
      silent = TRUE
    )
    if (inherits(lambda_u, "try-error")) {
      lambda_u <- as.numeric(NA)
    }
    gfi_ci_lower <- gfi_from_lambda(lambda_u)
  }

  list(
    gfi.ci.lower = gfi_ci_lower,
    gfi.ci.upper = gfi_ci_upper
  )
}

# compute a single GFI 'family' (point estimate + confidence interval, and the
# robust counterparts) for a given (unadjusted) test statistic. This is called
# once for the LR statistic and once for the RLS statistic.
lav_fit_gfi_family <- function(x2 = NULL, df = NULL, n = NULL, p = NULL,
                               ci_level = 0.90, robust_flag = FALSE,
                               xx3 = NULL, df3 = NULL, c_hat = NULL,
                               c_hat3 = NULL, xx3_scaled = NULL,
                               n3 = NULL) {
  # n3: the sample size to be used for the *robust* quantities (which may
  # be N rather than N-1 in the categorical case; see lav_fit_rmsea.R)
  if (is.null(n3)) {
    n3 <- n
  }
  out <- list(
    est = as.numeric(NA), ci.lower = as.numeric(NA), ci.upper = as.numeric(NA),
    robust = as.numeric(NA), ci.lower.robust = as.numeric(NA),
    ci.upper.robust = as.numeric(NA)
  )

  # point estimate (normal-theory, unbiased)
  out$est <- lav_fit_gfi(x2 = x2, df = df, n = n, p = p)

  # confidence interval
  ci <- lav_fit_gfi_ci(x2 = x2, df = df, n = n, p = p, level = ci_level)
  out$ci.lower <- ci$gfi.ci.lower
  out$ci.upper <- ci$gfi.ci.upper

  if (robust_flag) {
    out$robust <- lav_fit_gfi(
      x2 = xx3, df = df3, n = n3, p = p, c_hat = c_hat3
    )
    # note: input is the scaled test statistic!
    ci_robust <- lav_fit_gfi_ci(
      x2 = xx3_scaled, df = df3, n = n3, p = p,
      c_hat = c_hat, level = ci_level
    )
    out$ci.lower.robust <- ci_robust$gfi.ci.lower
    out$ci.upper.robust <- ci_robust$gfi.ci.upper
  }

  out
}

# ----------------------------------------------------------------------------
# higher-level function
# ----------------------------------------------------------------------------

lav_fit_gfi_lavobject <- function(lavobject = NULL, fit_measures = "gfi",
                                  standard_test = "standard",
                                  scaled_test = "none",
                                  ci_level = 0.90,
                                  robust = TRUE,
                                  cat_nonpd = "na") {
  # check lavobject
  stopifnot(inherits(lavobject, "lavaan"))

  # check for categorical
  categorical_flag <- lavobject@Model@categorical

  # tests
  test <- lavobject@test
  test_names <- sapply(lavobject@test, "[[", "test")
  if (test_names[1] == "none" || standard_test == "none") {
    return(list())
  }
  test_idx <- which(test_names == standard_test)[1]
  if (length(test_idx) == 0L || is.na(test_idx)) {
    return(list())
  }

  scaled_flag <- FALSE
  if (!scaled_test %in% c("none", "standard", "default")) {
    scaled_idx <- which(test_names == scaled_test)
    if (length(scaled_idx) > 0L) {
      scaled_idx <- scaled_idx[1] # only the first one
      scaled_flag <- TRUE
    }
  }

  # robust?
  robust_flag <- FALSE
  if (robust && scaled_flag &&
    scaled_test %in% c(
      "satorra.bentler", "yuan.bentler.mplus",
      "yuan.bentler", "yuan.chan", "scaled.shifted",
      "scaled.shifted.corrected"
    )) {
    robust_flag <- TRUE
  }

  # FIML? (the FIML-corrected robust indices are only available for
  # single-level models; for multilevel data, we fall back to the regular
  # scaled/robust computation above, which handles missing data too)
  fiml_flag <- FALSE
  if (robust && lavobject@Options$missing %in% c("ml", "ml.x") &&
    lavobject@Data@nlevels == 1L) {
    fiml_flag <- robust_flag <- TRUE
    # check if we can compute corrected values
    if (scaled_flag) {
      version <- "V3"
    } else {
      version <- "V6"
    }
    fiml <- try(
      lav_fit_fiml_corrected(lavobject,
        baseline_model = NULL,
        version = version
      ),
      silent = TRUE
    )
    if (inherits(fiml, "try-error")) {
      lav_msg_warn(gettext("computation of robust GFI failed."))
      fiml <- list(
        XX3 = as.numeric(NA), df3 = as.numeric(NA),
        c.hat3 = as.numeric(NA), XX3.scaled = as.numeric(NA)
      )
    } else if (anyNA(c(fiml$XX3, fiml$df3, fiml$c.hat3, fiml$XX3.scaled))) {
      lav_msg_warn(gettext(
        "computation of robust GFI resulted in NA values."
      ))
    }
  }

  # supported fit measures in this function
  # 1) the 'old' (LISREL) GFI and friends
  fit_gfi_lisrel <- c("gfi_lisrel", "agfi_lisrel", "pgfi")
  # 2) the 'new' GFI (with confidence interval); 'gfi' is the RLS version
  fit_gfi <- c("gfi", "gfi.ci.lower", "gfi.ci.upper", "gfi.ci.level")
  if (robust_flag) {
    fit_gfi <- c(
      fit_gfi, "gfi.robust",
      "gfi.ci.lower.robust", "gfi.ci.upper.robust"
    )
  }
  # 3) the explicit LR-based (gfi_lrt) and RLS-based (gfi_rls) versions
  fit_gfi_extra <- c(
    "gfi_lrt", "gfi_lrt.ci.lower", "gfi_lrt.ci.upper",
    "gfi_rls", "gfi_rls.ci.lower", "gfi_rls.ci.upper"
  )
  if (robust_flag) {
    fit_gfi_extra <- c(
      fit_gfi_extra,
      "gfi_lrt.robust", "gfi_lrt.ci.lower.robust", "gfi_lrt.ci.upper.robust",
      "gfi_rls.robust", "gfi_rls.ci.lower.robust", "gfi_rls.ci.upper.robust"
    )
  }
  fit_gfi_all <- c(fit_gfi_lisrel, fit_gfi, fit_gfi_extra)

  # which one do we need?
  if (missing(fit_measures)) {
    # default set
    fit_measures <- fit_gfi_all
  } else {
    # backward compatibility: "agfi" (<0.7-1) is an alias for "agfi_lisrel"
    fit_measures[fit_measures == "agfi"] <- "agfi_lisrel"
    # remove any not-GFI related index from fit.measures
    rm_idx <- which(!fit_measures %in% fit_gfi_all)
    if (length(rm_idx) > 0L) {
      fit_measures <- fit_measures[-rm_idx]
    }
    if (length(fit_measures) == 0L) {
      return(list())
    }
  }

  # container
  indices <- list()

  # ---------- the 'old' (LISREL) GFI and friends ----------
  if (any(fit_gfi_lisrel %in% fit_measures)) {
    # extract ingredients
    wls_obs <- lav_inspect_wls_obs(lavobject)
    wls_est <- lav_inspect_wls_est(lavobject)
    wls_v <- lav_inspect_wls_v(lavobject)
    nobs_1 <- lavobject@SampleStats@nobs

    # compute (LISREL) GFI
    gfi_lisrel <- lav_fit_gfi_lisrel(
      wls_obs = wls_obs, wls_est = wls_est,
      wls_v = wls_v, nobs_1 = nobs_1
    )

    # total number of modeled sample stats
    nel <- length(unlist(wls_obs))

    # degrees of freedom
    df <- lavobject@test[[1]]$df

    indices["gfi_lisrel"] <- gfi_lisrel
    indices["agfi_lisrel"] <-
      lav_fit_agfi_lisrel(gfi = gfi_lisrel, nel = nel, df = df)
    indices["pgfi"] <- lav_fit_pgfi(gfi = gfi_lisrel, nel = nel, df = df)
  }

  # ---------- the 'new' GFI (with confidence interval) ----------
  if (any(c(fit_gfi, fit_gfi_extra) %in% fit_measures)) {
    # basic ingredients
    x2 <- test[[test_idx]]$stat # LR (standard) chi-square statistic
    df <- test[[test_idx]]$df
    n <- lav_inspect_ntotal(object = lavobject) # N vs N-1
    # number of observed variables (per group)
    p <- lavobject@Model@nvar[[1]]

    # the RLS (reweighted least squares) statistic == 'browne.residual.nt.model'
    # (Maydeu-Olivares et al., 2024, recommend this over the LR statistic). Only
    # needed if an RLS-based measure (gfi, gfi_rls) is requested. There is no
    # separate RLS statistic in the categorical/FIML case (we reuse the LR one).
    rls_names <- c(fit_gfi, fit_gfi_extra[grepl("^gfi_rls", fit_gfi_extra)])
    x2_rls <- x2 # fallback
    if (any(rls_names %in% fit_measures) &&
      !categorical_flag && !fiml_flag) {
      rls <- try(lavTest(lavobject, test = "browne.residual.nt.model"),
        silent = TRUE
      )
      if (!inherits(rls, "try-error") &&
        !is.null(rls[["browne.residual.nt.model"]]) &&
        is.finite(rls[["browne.residual.nt.model"]]$stat)) {
        x2_rls <- rls[["browne.residual.nt.model"]]$stat
      }
    }

    # robust ingredients
    xx3 <- xx3_scaled <- df3 <- c_hat <- c_hat3 <- as.numeric(NA)
    xx3_rls <- xx3_scaled_rls <- as.numeric(NA)
    n3 <- n
    if (robust_flag) {
      if (categorical_flag) {
        out <- try(lav_fit_catml_dwls(lavobject, nonpd = cat_nonpd),
          silent = TRUE
        )
        if (inherits(out, "try-error")) {
          xx3 <- df3 <- c_hat3 <- xx3_scaled <- as.numeric(NA)
          c_hat <- as.numeric(NA)
        } else {
          xx3 <- out$XX3
          df3 <- out$df3
          c_hat3 <- c_hat <- out$c.hat3
          xx3_scaled <- out$XX3.scaled
        }
        # no separate RLS version for categorical data
        xx3_rls <- xx3
        xx3_scaled_rls <- xx3_scaled
        # the robust quantities are ML-based: use N, not N-1
        # (see lav_fit_rmsea.R)
        n3 <- lavobject@SampleStats@ntotal
      } else if (fiml_flag) {
        xx3 <- fiml$XX3
        df3 <- fiml$df3
        c_hat3 <- c_hat <- fiml$c.hat3
        xx3_scaled <- fiml$XX3.scaled
        # no separate RLS version for FIML
        xx3_rls <- xx3
        xx3_scaled_rls <- xx3_scaled
      } else {
        xx3 <- x2
        df3 <- df
        c_hat <- test[[scaled_idx]]$scaling.factor
        shift_par <- 0
        if (scaled_test %in% c(
          "scaled.shifted", "scaled.shifted.corrected"
        )) {
          # compute c.hat from a and b
          a <- test[[scaled_idx]]$scaling.factor
          b <- test[[scaled_idx]]$shift.parameter
          shift_par <- b
          c_hat3 <- a * (df - b) / df
        } else {
          c_hat3 <- c_hat
        }
        xx3_scaled <- test[[scaled_idx]]$stat
        # RLS analogues: the scaling constant c (and the shift b) are identical
        # for T_LR and T_RLS (Maydeu-Olivares et al., 2024, Eq. 19), so we only
        # need the unadjusted RLS statistic and reuse the same scaling factor(s)
        xx3_rls <- x2_rls
        xx3_scaled_rls <- x2_rls / c_hat + shift_par
      }
    }

    # GFI based on the LR statistic
    gfi_lrt <- lav_fit_gfi_family(
      x2 = x2, df = df, n = n, p = p, ci_level = ci_level,
      robust_flag = robust_flag, xx3 = xx3, df3 = df3,
      c_hat = c_hat, c_hat3 = c_hat3, xx3_scaled = xx3_scaled,
      n3 = n3
    )
    # GFI based on the RLS statistic (the recommended version)
    gfi_rls <- lav_fit_gfi_family(
      x2 = x2_rls, df = df, n = n, p = p, ci_level = ci_level,
      robust_flag = robust_flag, xx3 = xx3_rls, df3 = df3,
      c_hat = c_hat, c_hat3 = c_hat3, xx3_scaled = xx3_scaled_rls,
      n3 = n3
    )

    # the default (printed) 'gfi' is the RLS version
    indices["gfi"] <- gfi_rls$est
    indices["gfi.ci.level"] <- ci_level
    indices["gfi.ci.lower"] <- gfi_rls$ci.lower
    indices["gfi.ci.upper"] <- gfi_rls$ci.upper

    # explicit LR-based and RLS-based versions
    indices["gfi_lrt"] <- gfi_lrt$est
    indices["gfi_lrt.ci.lower"] <- gfi_lrt$ci.lower
    indices["gfi_lrt.ci.upper"] <- gfi_lrt$ci.upper
    indices["gfi_rls"] <- gfi_rls$est
    indices["gfi_rls.ci.lower"] <- gfi_rls$ci.lower
    indices["gfi_rls.ci.upper"] <- gfi_rls$ci.upper

    if (robust_flag) {
      indices["gfi.robust"] <- gfi_rls$robust
      indices["gfi.ci.lower.robust"] <- gfi_rls$ci.lower.robust
      indices["gfi.ci.upper.robust"] <- gfi_rls$ci.upper.robust

      indices["gfi_lrt.robust"] <- gfi_lrt$robust
      indices["gfi_lrt.ci.lower.robust"] <- gfi_lrt$ci.lower.robust
      indices["gfi_lrt.ci.upper.robust"] <- gfi_lrt$ci.upper.robust
      indices["gfi_rls.robust"] <- gfi_rls$robust
      indices["gfi_rls.ci.lower.robust"] <- gfi_rls$ci.lower.robust
      indices["gfi_rls.ci.upper.robust"] <- gfi_rls$ci.upper.robust
    }
  }

  # return only those that were requested
  indices[fit_measures]
}

# user-visible function to extract the fit measures
# output can be 1) vector (default), 2) list, 3) matrix, or 4) text
# in the latter case, the result will be of class "lavaan.fitMeasures"
# for which the printing is done by lav_fitmeasures_print()

# new in 0.6-13:
# the big families are computed in dedicated functions:
# - lav_fit_rmsea_lavobject
# - lav_fit_cfi_lavobject
# - lav_fit_aic_lavojbect
# - lav_residuals_summary

# Note: fitMeasures/fitmeasures are generic functions; they include a "..."
#       so lavaan.mi can add arguments to pass to lavTestLRT() and
#       lavTestLRT.mi() about how to pool chi-squared.

setMethod(
  "fitMeasures", signature(object = "lavaan"),
  function(object, fit_measures = "all", baseline_model = NULL, h1_model = NULL,
           fm_args = list(
             standard.test = "default",
             scaled.test = "default",
             rmsea.ci.level = 0.90,
             rmsea.close.h0 = 0.05,
             rmsea.notclose.h0 = 0.08,
             robust = TRUE,
             cat.nonpd = "na"
           ),
           output = "vector", level = NULL, ...) {
    dotdotdot <- list(...)
    lav_adapt_func(environment(), dotdotdot, NULL)
    if (!is.list(fit_measures))
        fit_measures <- list(fit.measures = fit_measures)
    if (!missing(fm_args)) fit_measures <- c(fit_measures, fm_args)
    lav_fit(
      object = object, fit_measures = fit_measures,
      baseline_model = baseline_model, h1_model = h1_model,
      output = output, level = level
    )
  }
)

setMethod(
  "fitmeasures", signature(object = "lavaan"),
  function(object, fit_measures = "all", baseline_model = NULL, h1_model = NULL,
           fm_args = list(
             standard.test = "default",
             scaled.test = "default",
             rmsea.ci.level = 0.90,
             rmsea.close.h0 = 0.05,
             rmsea.notclose.h0 = 0.08,
             robust = TRUE,
             cat.nonpd = "na"
           ),
           output = "vector", level = NULL, ...) {
    dotdotdot <- list(...)
    lav_adapt_func(environment(), dotdotdot, NULL)
    if (!is.list(fit_measures))
                  fit_measures <- list(fit.measures = fit_measures)
    if (!missing(fm_args)) fit_measures <- c(fit_measures, fm_args)
    lav_fit(
      object = object, fit_measures = fit_measures,
      baseline_model = baseline_model, h1_model = h1_model,
      output = output, level = level
    )
  }
)

# S3 method for efaList
lav_efalist_fitmeasures <- function(
    object,
    fit_measures = "all",
    baseline_model = NULL, h1_model = NULL,
    fm_args = list(
      standard.test = "default",
      scaled.test = "default",
      rmsea.ci.level = 0.90,
      rmsea.close.h0 = 0.05,
      rmsea.notclose.h0 = 0.08,
      robust = TRUE,
      cat.nonpd = "na"
    ),
    output = "list", ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)

  # kill object$loadings if present
  object[["loadings"]] <- NULL

  # get fit measures for each model
  if (!is.list(fit_measures)) fit_measures <- list(fit.measures = fit_measures)
  if (!missing(fm_args)) fit_measures <- c(fit_measures, fm_args)
  res <- simplify2array(lapply(
    object,
    function(x) {
      lav_fit(
        object = x,
        fit_measures = fit_measures, h1_model = h1_model,
        baseline_model = baseline_model,
        output = "vector"
      )
    }
  )
  )

  # check if res is a matrix
  if (!is.matrix(res)) {
    if (is.numeric(res)) {
      # fit_measures is just 1 element, or only one was correct
      name <- names(res)[1]
      res <- matrix(res, nrow = 1L)
      rownames(res) <- name
    } else { # wrong fit measures?
      # create empty matrix
      res <- matrix(0, nrow = 0L, ncol = length(object))
    }
  }

  # rownames
  nfactors <- sapply(object, function(x) x@pta$nfac[[1]])
  colnames(res) <- paste0("nfactors = ", nfactors)

  # class
  class(res) <- c("lavaan.matrix", "matrix")

  res
}


# complete catalog of fit measure names that fitMeasures() can ever return
# (across all estimators/test statistics/data types); used to catch
# misspelled or unknown measure names early -- keep in sync with the 'sets'
# defined in lav_fit() below
# note: a name being listed here does not imply the measure is available
# for every fitted model (eg the .scaled/.robust variants); requesting a
# valid-but-unavailable measure is silently ignored, as before
lav_fit_measures_names <- function() {
  c(
    # always
    "npar", "ntotal", "fmin",
    # chi-square test
    "chisq", "df", "pvalue",
    "chisq.scaled", "df.scaled", "pvalue.scaled", "chisq.scaling.factor",
    # baseline model
    "baseline.chisq", "baseline.df", "baseline.pvalue",
    "baseline.chisq.scaled", "baseline.df.scaled", "baseline.pvalue.scaled",
    "baseline.chisq.scaling.factor",
    # incremental fit indices
    "cfi", "tli", "cfi.scaled", "tli.scaled", "cfi.robust", "tli.robust",
    "nnfi", "rfi", "nfi", "pnfi", "ifi", "rni",
    "nnfi.scaled", "rfi.scaled", "nfi.scaled", "pnfi.scaled",
    "ifi.scaled", "rni.scaled",
    "nnfi.robust", "rni.robust",
    # likelihood based measures
    "logl", "unrestricted.logl", "aic", "bic", "bic2",
    "scaling.factor.h1", "scaling.factor.h0",
    # rmsea
    "rmsea", "rmsea.ci.lower", "rmsea.ci.upper", "rmsea.ci.level",
    "rmsea.pvalue", "rmsea.close.h0",
    "rmsea.notclose.pvalue", "rmsea.notclose.h0",
    "rmsea.scaled", "rmsea.ci.lower.scaled", "rmsea.ci.upper.scaled",
    "rmsea.pvalue.scaled", "rmsea.notclose.pvalue.scaled",
    "rmsea.robust", "rmsea.ci.lower.robust", "rmsea.ci.upper.robust",
    "rmsea.pvalue.robust", "rmsea.notclose.pvalue.robust",
    # gfi and friends
    "gfi", "gfi.ci.lower", "gfi.ci.upper", "gfi.ci.level",
    "gfi.robust", "gfi.ci.lower.robust", "gfi.ci.upper.robust",
    "gfi_lrt", "gfi_lrt.ci.lower", "gfi_lrt.ci.upper",
    "gfi_lrt.robust", "gfi_lrt.ci.lower.robust", "gfi_lrt.ci.upper.robust",
    "gfi_rls", "gfi_rls.ci.lower", "gfi_rls.ci.upper",
    "gfi_rls.robust", "gfi_rls.ci.lower.robust", "gfi_rls.ci.upper.robust",
    "gfi_lisrel", "agfi_lisrel", "pgfi",
    "agfi", # alias for agfi_lisrel (the only pre-0.7-1 name)
    # (s)rmr
    "rmr", "rmr_nomean", "srmr",
    "srmr_bentler", "srmr_bentler_nomean",
    "crmr", "crmr_nomean",
    "srmr_mplus", "srmr_mplus_nomean",
    "srmr_within", "srmr_between",
    # various
    "cn_05", "cn_01", "mfi", "ecvi", "wrmr"
  )
}

# if fit_measures is a list, the element holding the measures is classically
# named "fit.measures"; accept the snake_case spelling "fit_measures" as an
# alias, matching the name of the formal argument
lav_fit_measures_list_alias <- function(fit_measures) {
  if (is.list(fit_measures) && is.null(fit_measures[["fit.measures"]]) &&
      !is.null(fit_measures[["fit_measures"]])) {
    names(fit_measures)[names(fit_measures) == "fit_measures"] <-
      "fit.measures"
  }
  fit_measures
}

lav_fit <- function(object, fit_measures = "all",
                             baseline_model = NULL, h1_model = NULL,
                             fm_args = list(
                               standard.test = "default",
                               scaled.test = "default",
                               rmsea.ci.level = 0.90,
                               rmsea.close.h0 = 0.05,
                               rmsea.notclose.h0 = 0.08,
                               robust = TRUE,
                               cat.nonpd = "na"
                             ),
                             output = "vector", level = NULL) {
  # check object
  object <- lav_object_check_version(object)

  # level-specific fit measures for multilevel models: delegate to the
  # partially saturated model for that level (see lav_fit_by_level.R)
  if (!is.null(level)) {
    target <- lav_fit_by_level_target(object, level = level)
    fit_l <- lav_fit_by_level_get(object, level = level)
    if (missing(fm_args)) {
      out <- lav_fit(
        object = fit_l, fit_measures = fit_measures,
        baseline_model = baseline_model, h1_model = h1_model,
        output = output
      )
    } else {
      out <- lav_fit(
        object = fit_l, fit_measures = fit_measures,
        baseline_model = baseline_model, h1_model = h1_model,
        fm_args = fm_args, output = output
      )
    }
    if (output %in% c("vector", "text")) {
      attr(out, "header") <- gettextf(
        "Fit measures for the %s level only (the other level is saturated):",
        target)
    }
    return(out)
  }

  # sam objects (issue #517)
  if (!is.null(object@internal$sam.method)) {
    if (lav_sam_local_flag(object)) {
      # delegate to the stored step-2 structural fit: the only meaningful
      # fit measures for a local SAM model are those of the structural part,
      # conditional on the (fixed) step-1 measurement model
      fit_pa <- lav_sam_struc_object(object)
      if (is.null(fit_pa)) {
        lav_msg_stop(gettext(
          "no stored structural fit found: this sam object was created by an
           older version of lavaan; please rerun sam()."))
      }
      lav_msg_note(gettext(
        "the fit measures are computed for the structural part only,
         conditional on the (fixed) measurement model of step 1."))
      if (missing(fm_args)) {
        out <- lav_fit(
          object = fit_pa, fit_measures = fit_measures,
          baseline_model = baseline_model, h1_model = h1_model,
          output = output
        )
      } else {
        out <- lav_fit(
          object = fit_pa, fit_measures = fit_measures,
          baseline_model = baseline_model, h1_model = h1_model,
          fm_args = fm_args, output = output
        )
      }
      if (output %in% c("vector", "text")) {
        attr(out, "header") <- gettext(
          "Fit measures of the structural part (given the measurement model):")
      }
      return(out)
    } else {
      # sam.method = "global": the fit measures are computed from the joint
      # model, but the parameters were estimated in two steps
      lav_msg_warn(gettext(
        "the parameters of this model were estimated using a two-step (sam)
         procedure; the behavior of fit measures in this context has not
         been studied, and they should be interpreted with caution."))
    }
  }

  # default fm_args
  default_fm_args <- list(
    standard.test = "default",
    scaled.test = "default",
    rmsea.ci.level = 0.90,
    rmsea.close.h0 = 0.05,
    rmsea.notclose.h0 = 0.08,
    robust = TRUE,
    cat.nonpd = "na"
  )
  if (!missing(fm_args)) {
    lav_deprecated_args("fit_measures", "fm_args")
    fm_args <- modifyList(default_fm_args, fm_args)
  } else {
    fm_args <- default_fm_args
  }
  if (is.list(fit_measures)) {
    fit_measures <- lav_fit_measures_list_alias(fit_measures)
    if (is.null(names(fit_measures)) ||
        is.null(fit_measures$fit.measures)) {
      lav_msg_stop(gettextf(
        "If %s is a list, it must contain a named element %s.",
        "fit_measures", "fit.measures (or fit_measures)"
      ))
    }
    temp <- fit_measures$fit.measures
    fit_measures$fit.measures <- NULL
    fm_args <- modifyList(default_fm_args, fit_measures)
    fit_measures <- temp
  }

  # resolve the cat.nonpd option (new in 0.7-1): what to do when data are
  # categorical and an input correlation matrix is not positive definite?
  #    "na"     -> robust fit indices are NA (default)
  #    "refit"  -> smooth the matrix, and re-estimate the parameters
  #                using estimator = "catML"
  #    "smooth" -> smooth the matrix, but keep the (DWLS) parameter
  #                estimates (as in lavaan <= 0.6-13)
  # the older cat.check.pd = FALSE is a deprecated alias for "refit"
  if (!is.null(fm_args$cat.check.pd) && !isTRUE(fm_args$cat.check.pd) &&
      identical(fm_args$cat.nonpd, "na")) {
    fm_args$cat.nonpd <- "refit"
  }
  if (!is.character(fm_args$cat.nonpd) ||
      !fm_args$cat.nonpd %in% c("na", "refit", "smooth")) {
    lav_msg_warn(gettextf(
      "invalid cat.nonpd value [%s] set to default \"na\".",
      fm_args$cat.nonpd
    ))
    fm_args$cat.nonpd <- "na"
  }

  # standard test
  if (fm_args$standard.test == "default") {
    fm_args$standard.test <- object@Options$standard.test
    # usually "standard", but could have been changed
    if (is.null(fm_args$standard.test)) { # <older objects
      fm_args$standard.test <- "standard"
    }
  }

  # scaled test
  if (fm_args$scaled.test == "default") {
    fm_args$scaled.test <- object@Options$scaled.test
    # usually "standard", but could have been changed
    if (is.null(fm_args$scaled.test)) { # <older objects
      fm_args$scaled.test <- "standard"
    }
  }

  # do we have data? (yep, we had to include this check)
  if (object@Data@data.type == "none") {
    lav_msg_stop(gettext("fit measures not available if there is no data."))
  }

  # has the model converged?
  if (object@optim$npar > 0L && !object@optim$converged) {
    lav_msg_stop(gettext(
      "fit measures not available if model did not converge"))
  }

  # random slopes (rv() modifier): no test statistics (and no
  # comparable h1 loglikelihood) exist, but the loglikelihood-based
  # measures -- and the H0 scaling correction factor -- are available
  if (length(object@Model@rv.ov) > 0L ||
      length(object@Model@rv.lv) > 0L) {
    return(lav_fit_rv(object, fit_measures = fit_measures,
                      output = output))
  }

  # do we have a test statistic?
  test <- lavInspect(object, "test")
  test_names <- unname(sapply(test, "[[", "test"))
  if (test_names[1] == "none") {
    lav_msg_stop(gettext("fit measures not available if test = \"none\"."))
    #FIXME: allow RMRs, log.likelihoods, info criteria, npar, ntotal
  }

  standard_test <- fm_args$standard.test
  scaled_test <- fm_args$scaled.test

  # check standard.test
  standard_test <- lav_test_rename(standard_test, check = TRUE)[1] # only 1
  fmg_standard_test <- lav_test_fmg_is_fmg(standard_test)

  # check scaled.test
  if (!scaled_test %in% c("none", "default", "standard")) {
    scaled_test <- lav_test_rename(scaled_test, check = TRUE)[1] # only 1
  }

  # which test statistic do we need?
  rerun_lavtest_flag <- FALSE
  if (!standard_test %in% test_names) {
    rerun_lavtest_flag <- TRUE
  }
  if (!scaled_test %in% c("none", "default", "standard") &&
      !scaled_test %in% test_names) {
    rerun_lavtest_flag <- TRUE
  }

  # do we have a scaled test statistic? if so, which one?
  scaled_test_names <- c(
    "satorra.bentler",
    "yuan.bentler", "yuan.bentler.mplus", "yuan.chan",
    "mean.var.adjusted", "scaled.shifted",
    "mean.var.adjusted.corrected", "scaled.shifted.corrected"
  )
  scaled_flag <- FALSE
  if (!fmg_standard_test &&
      scaled_test != "none" &&
      any(test_names %in% scaled_test_names)) {
    scaled_flag <- TRUE
    # note: object@Options$scaled.test holds the *base* statistic that was
    # scaled (eg "browne.residual.nt.model"), not the name of a scaled
    # test; in that case (as for "standard"/"default"), use the first
    # scaled test that is available
    if (!scaled_test %in% scaled_test_names) {
      tmp_idx <- which(test_names %in% scaled_test_names)
      scaled_test <- test_names[tmp_idx[1]]
    }
  }

  # rerun lavTest?
  if (rerun_lavtest_flag) {
    this_test <- standard_test
    if (scaled_flag) {
      this_test <- unique(this_test, scaled_test)
    }
    lavtest_scaled_test <- standard_test
    if (fmg_standard_test) {
      lavtest_scaled_test <- scaled_test
    }
    test <- lavTest(object,
                    test = this_test, scaled_test = lavtest_scaled_test,
                    drop_list_single = FALSE
    )
    # replace in object, if we pass it to lav_fit_* functions
    object@test <- test
    test_names <- unname(sapply(test, "[[", "test"))
  }


  # TDJ: Check for user-supplied h1 model
  #      Similar to BASELINE model, use the following priority:
  #        1. user-provided h1 model
  #        2. h1 model in @external slot
  #        3. default h1 model (already in @h1 slot, no update necessary)

  user_h1_exists <- FALSE
  # 1. user-provided h1 model
  if (!is.null(h1_model)) {
    stopifnot(inherits(h1_model, "lavaan"))
    user_h1_exists <- TRUE

    # 2. h1 model in @external slot
  } else if (!is.null(object@external$h1.model)) {
    stopifnot(inherits(object@external$h1.model, "lavaan"))
    h1_model <- object@external$h1.model
    user_h1_exists <- TRUE
  }

  ## Update statistics in @test slot?
  if (user_h1_exists) {
    ## update @test slot
    fit <- lav_update_test_custom_h1(lav_obj_h0 = object, lav_obj_h1 = h1_model)

    ## re-assign TEST object that is used below
    object@test <- test <- fit@test
    test_names <- unname(sapply(test, "[[", "test"))
  }

  # get index of standard.test in TEST
  test_idx <- which(test_names == standard_test)[1]

  # get index of scaled test (if any) in TEST
  if (scaled_flag) {
    scaled_idx <- which(test_names == scaled_test)[1]
  }

  # check output argument
  if (output %in% c("vector", "horizontal")) {
    output <- "vector"
  } else if (output %in% c("list")) {
    output <- "list"
  } else if (output %in% c("matrix", "vertical")) {
    output <- "matrix"
  } else if (output %in% c("text", "pretty", "summary")) {
    output <- "text"
  } else {
    lav_msg_stop(gettextf("output should be %s.",
                          lav_msg_view(c("vector", "list", "matrix", "text"),
                          "none", FALSE)
    ))
  }

  # options
  categorical_flag <- object@Model@categorical
  fiml_flag <- (fm_args$robust &&
                  object@Options$missing %in% c("ml", "ml.x"))
  estimator <- object@Options$estimator

  # basic ingredients
  x2 <- test[[test_idx]]$stat
  df <- test[[test_idx]]$df
  if (scaled_flag) {
    x2_scaled <- test[[scaled_idx]]$stat
    df_scaled <- test[[scaled_idx]]$df
  }
  npar <- lav_inspect_npar(object = object, ceq = TRUE)
  n <- lav_inspect_ntotal(object = object) # N vs N-1


  # define 'sets' of fit measures:
  fit_always <- c("npar")

  # basic chi-square test
  fit_chisq <- c("fmin", "chisq", "df", "pvalue")
  if (scaled_flag) {
    fit_chisq <- c(
      fit_chisq, "chisq.scaled", "df.scaled", "pvalue.scaled",
      "chisq.scaling.factor"
    )
  }

  # baseline model
  fit_baseline <- c("baseline.chisq", "baseline.df", "baseline.pvalue")
  if (scaled_flag) {
    fit_baseline <- c(
      fit_baseline, "baseline.chisq.scaled",
      "baseline.df.scaled", "baseline.pvalue.scaled",
      "baseline.chisq.scaling.factor"
    )
  }

  fit_cfi_tli <- c("cfi", "tli")
  if (scaled_flag) {
    fit_cfi_tli <- c(fit_cfi_tli, "cfi.scaled", "tli.scaled")
  }
  if (fm_args$robust && (scaled_flag || categorical_flag || fiml_flag)) {
    fit_cfi_tli <- c(fit_cfi_tli, "cfi.robust", "tli.robust")
  }

  # other incremental fit indices
  fit_cfi_other <- c("nnfi", "rfi", "nfi", "pnfi", "ifi", "rni")
  if (scaled_flag) {
    fit_cfi_other <- c(
      fit_cfi_other, "nnfi.scaled", "rfi.scaled",
      "nfi.scaled", "pnfi.scaled", "ifi.scaled", "rni.scaled"
    )
  }
  if (fm_args$robust && (scaled_flag || categorical_flag || fiml_flag)) {
    fit_cfi_other <- c(fit_cfi_other, "nnfi.robust", "rni.robust")
  }
  fit_cfi <- c(fit_baseline, fit_cfi_tli, fit_cfi_other)

  # likelihood based measures
  if (estimator == "MML") {
    fit_logl <- c("logl", "aic", "bic", "ntotal", "bic2")
  } else {
    fit_logl <- c(
      "logl", "unrestricted.logl", "aic", "bic",
      "ntotal", "bic2"
    )
  }
  if (scaled_flag &&
      scaled_test %in% c("yuan.bentler", "yuan.bentler.mplus")) {
    fit_logl <- c(fit_logl, "scaling.factor.h1", "scaling.factor.h0")
  }

  # rmsea
  fit_rmsea <- c(
    "rmsea",
    "rmsea.ci.lower", "rmsea.ci.upper", "rmsea.ci.level",
    "rmsea.pvalue", "rmsea.close.h0",
    "rmsea.notclose.pvalue", "rmsea.notclose.h0"
  )
  if (scaled_flag) {
    fit_rmsea <- c(
      fit_rmsea, "rmsea.scaled", "rmsea.ci.lower.scaled",
      "rmsea.ci.upper.scaled", "rmsea.pvalue.scaled",
      "rmsea.notclose.pvalue.scaled"
    )
  }
  if (fm_args$robust && (scaled_flag || categorical_flag || fiml_flag)) {
    fit_rmsea <- c(
      fit_rmsea, "rmsea.robust", "rmsea.ci.lower.robust",
      "rmsea.ci.upper.robust", "rmsea.pvalue.robust",
      "rmsea.notclose.pvalue.robust"
    )
  }

  # gfi (the 'new' GFI of Maydeu-Olivares et al., 2024, with conf. interval)
  # note: the (printed) 'gfi' is the RLS-based version
  fit_gfi <- c("gfi", "gfi.ci.lower", "gfi.ci.upper", "gfi.ci.level")
  if (fm_args$robust && (scaled_flag || categorical_flag || fiml_flag)) {
    fit_gfi <- c(
      fit_gfi, "gfi.robust",
      "gfi.ci.lower.robust", "gfi.ci.upper.robust"
    )
  }
  # explicit LR-based (gfi_lrt) and RLS-based (gfi_rls) versions; these are not
  # part of the 'default' or 'all' sets, but can be requested by name
  fit_gfi_extra <- c(
    "gfi_lrt", "gfi_lrt.ci.lower", "gfi_lrt.ci.upper",
    "gfi_rls", "gfi_rls.ci.lower", "gfi_rls.ci.upper"
  )
  if (fm_args$robust && (scaled_flag || categorical_flag || fiml_flag)) {
    fit_gfi_extra <- c(
      fit_gfi_extra,
      "gfi_lrt.robust", "gfi_lrt.ci.lower.robust", "gfi_lrt.ci.upper.robust",
      "gfi_rls.robust", "gfi_rls.ci.lower.robust", "gfi_rls.ci.upper.robust"
    )
  }
  # the 'old' (LISREL) GFI and friends
  fit_gfi_lisrel <- c("gfi_lisrel", "agfi_lisrel", "pgfi")

  # srmr
  if (categorical_flag) {
    fit_srmr <- c("srmr")
    fit_srmr2 <- c(
      "rmr", "rmr_nomean",
      "srmr", # per default equal to srmr_bentler_nomean
      "srmr_bentler", "srmr_bentler_nomean",
      "crmr", "crmr_nomean",
      "srmr_mplus", "srmr_mplus_nomean"
    )
  } else {
    if (object@Data@nlevels > 1L) {
      fit_srmr <- c("srmr", "srmr_within", "srmr_between")
      fit_srmr2 <- c("srmr", "srmr_within", "srmr_between")
    } else {
      fit_srmr <- c("srmr")
      fit_srmr2 <- c(
        "rmr", "rmr_nomean",
        "srmr", # the default
        "srmr_bentler", "srmr_bentler_nomean",
        "crmr", "crmr_nomean",
        "srmr_mplus", "srmr_mplus_nomean"
      )
    }
  }

  # various
  if (object@Data@nlevels > 1L) {
    fit_other <- ""
  } else if (estimator == "PML") {
    fit_other <- c("cn_05", "cn_01", "mfi")
    if (!categorical_flag) { # really needed?
      fit_other <- c(fit_other, "ecvi")
    }
  } else {
    fit_other <- c("cn_05", "cn_01", fit_gfi_lisrel, "mfi")
    if (!categorical_flag) { # really needed?
      fit_other <- c(fit_other, "ecvi")
    } else {
      fit_other <- c(fit_other, "wrmr")
    }
  }

  # lower case
  fit_measures <- fit_measures_orig <- tolower(fit_measures)

  # select 'default' fit measures
  if (length(fit_measures) == 1L) {
    if (fit_measures == "default") {
      if (estimator == "ML" || estimator == "PML") {
        fit_measures <- c(
          fit_always, fit_chisq, fit_baseline,
          fit_cfi_tli, fit_logl,
          fit_rmsea, fit_srmr, fit_gfi
        )
      } else if (estimator == "MML") {
        fit_measures <- c(fit_always, fit_logl)
      } else {
        fit_measures <- c(
          fit_always,
          fit_chisq, fit_baseline, fit_cfi_tli,
          fit_rmsea, fit_srmr, fit_gfi
        )
      }
    } else if (fit_measures == "all") {
      if (estimator == "ML") {
        fit_measures <- c(
          fit_always, fit_chisq,
          fit_baseline, fit_cfi_tli, fit_cfi_other,
          fit_logl, fit_rmsea, fit_srmr2, fit_gfi, fit_other
        )
      } else {
        fit_measures <- c(
          fit_always, fit_chisq,
          fit_baseline, fit_cfi_tli, fit_cfi_other,
          fit_rmsea, fit_srmr2, fit_gfi, fit_other
        )
      }
    }
  }

  # if specific measures were requested (ie not one of the keywords
  # "all"/"default"/"none"), check that the names are known (new in 0.7-1);
  # note: we check against the complete catalog, not against what is
  # available for this particular fit
  # unknown names are dropped with a warning (not an error): several
  # packages pass their own non-lavaan measures along, relying on the
  # pre-0.7-1 behavior of silently ignoring them
  if (!(length(fit_measures_orig) == 1L &&
        fit_measures_orig %in% c("all", "default", "none"))) {
    bad_idx <- which(!fit_measures %in% lav_fit_measures_names())
    if (length(bad_idx) > 0L) {
      lav_msg_warn(ngettext(length(bad_idx),
        "unknown fit measure:", "unknown fit measures:"),
        lav_msg_view(fit_measures_orig[bad_idx], "none", FALSE)
      )
      fit_measures <- fit_measures[-bad_idx]
    }
  }

  # catch empty list
  if (length(fit_measures) == 0L) {
    return(list())
  }

  # main container
  indices <- list()
  indices["npar"] <- npar
  indices["ntotal"] <- object@SampleStats@ntotal
  indices["fmin"] <- object@optim$fx # note = 0.5 * fmin if ML

  # CHI-SQUARE TEST
  if (any(fit_chisq %in% fit_measures)) {
    indices["chisq"] <- x2
    indices["df"] <- df
    indices["pvalue"] <- test[[test_idx]]$pvalue
    if (scaled_flag) {
      indices["chisq.scaled"] <- x2_scaled
      indices["df.scaled"] <- df_scaled
      indices["chisq.scaling.factor"] <- test[[scaled_idx]]$scaling.factor
      indices["pvalue.scaled"] <- test[[scaled_idx]]$pvalue
    }
  }

  # BASELINE FAMILY
  if (any(fit_cfi %in% fit_measures)) {

    # rerun baseline?
    if (rerun_lavtest_flag) {
      object@Options$test <- this_test
      fit_indep <- try(lav_object_independence(object), silent = TRUE)
      # override object
      object@baseline$test <- fit_indep@test
    }

    indices <- c(
      indices,
      lav_fit_cfi_lavobject(
        lavobject = object,
        fit_measures = fit_measures,
        baseline_model = baseline_model,
        h1_model = h1_model,
        standard_test = standard_test,
        scaled_test = scaled_test,
        robust = fm_args$robust,
        cat_nonpd = fm_args$cat.nonpd
      )
    )
  }

  # INFORMATION CRITERIA
  if (any(fit_logl %in% fit_measures)) {
    indices <- c(
      indices,
      lav_fit_aic_lavobject(
        lavobject = object,
        fit_measures = fit_measures,
        standard_test = standard_test,
        scaled_test = scaled_test,
        estimator = estimator
      )
    )
  }

  # RMSEA and friends
  if (any(fit_rmsea %in% fit_measures)) {
    # check rmsea options
    rmsea_ci_level <- 0.90
    rmsea_close_h0 <- 0.05
    rmsea_notclose_h0 <- 0.08
    if (!is.null(fm_args$rmsea.ci.level) &&
        is.finite(fm_args$rmsea.ci.level)) {
      rmsea_ci_level <- fm_args$rmsea.ci.level
      if (rmsea_ci_level < 0 || rmsea_ci_level > 1.0) {
        lav_msg_warn(gettextf(
          "invalid rmsea.ci.level value [%s] set to default 0.90.",
          rmsea_ci_level))
        rmsea_ci_level <- 0.90
      }
    }
    if (!is.null(fm_args$rmsea.close.h0) &&
        is.finite(fm_args$rmsea.close.h0)) {
      rmsea_close_h0 <- fm_args$rmsea.close.h0
      if (rmsea_close_h0 < 0) {
        rmsea_close_h0 <- 0
      }
    }
    if (!is.null(fm_args$rmsea.notclose.h0) &&
        is.finite(fm_args$rmsea.notclose.h0)) {
      rmsea_notclose_h0 <- fm_args$rmsea.notclose.h0
      if (rmsea_notclose_h0 < 0) {
        rmsea_notclose_h0 <- 0
      }
    }
    indices <- c(
      indices,
      lav_fit_rmsea_lavobject(
        lavobject = object,
        fit_measures = fit_measures,
        standard_test = standard_test,
        scaled_test = scaled_test,
        ci_level = rmsea_ci_level,
        close_h0 = rmsea_close_h0,
        notclose_h0 = rmsea_notclose_h0,
        robust = fm_args$robust,
        cat_nonpd = fm_args$cat.nonpd
      )
    )
  }

  # SRMR and friends
  if (any(fit_srmr2 %in% fit_measures)) {
    indices <- c(
      indices,
      lav_fit_srmr_lavobject(
        lavobject = object,
        fit_measures = fit_measures
      )
    )
  }

  # GFI and friends
  # "agfi" is a backward-compatible alias for "agfi_lisrel" (<0.7-1)
  if (any(c(fit_gfi, fit_gfi_extra, fit_gfi_lisrel, "agfi")
          %in% fit_measures)) {
    # check gfi.ci.level option (default 0.90)
    gfi_ci_level <- 0.90
    if (!is.null(fm_args$gfi.ci.level) &&
        is.finite(fm_args$gfi.ci.level)) {
      gfi_ci_level <- fm_args$gfi.ci.level
      if (gfi_ci_level < 0 || gfi_ci_level > 1.0) {
        lav_msg_warn(gettextf(
          "invalid gfi.ci.level value [%s] set to default 0.90.",
          gfi_ci_level))
        gfi_ci_level <- 0.90
      }
    }
    indices <- c(
      indices,
      lav_fit_gfi_lavobject(
        lavobject = object,
        fit_measures = fit_measures,
        standard_test = standard_test,
        scaled_test = scaled_test,
        ci_level = gfi_ci_level,
        robust = fm_args$robust,
        cat_nonpd = fm_args$cat.nonpd
      )
    )
    # return the value under the old (requested) name
    if ("agfi" %in% fit_measures) {
      indices[["agfi"]] <- indices[["agfi_lisrel"]]
    }
  }

  # various: Hoelter Critical N (CN)
  if (any(c("cn_05", "cn_01") %in% fit_measures)) {
    indices["cn_05"] <- lav_fit_cn(x2 = x2, df = df, n = n, alpha = 0.05)
    indices["cn_01"] <- lav_fit_cn(x2 = x2, df = df, n = n, alpha = 0.01)
  }

  # various: WRMR
  if ("wrmr" %in% fit_measures) {
    nel <- length(object@SampleStats@WLS.obs[[1]])
    indices["wrmr"] <- lav_fit_wrmr(x2 = x2, nel = nel)
  }

  # various: MFI
  if ("mfi" %in% fit_measures) {
    indices["mfi"] <- lav_fit_mfi(x2 = x2, df = df, n = n)
  }

  # various: ECVI
  if ("ecvi" %in% fit_measures) {
    indices["ecvi"] <- lav_fit_ecvi(x2 = x2, npar = npar, n = n)
  }


  # keep only what we need
  out <- indices[fit_measures]

  if (all(is.na(names(out)))) { # perhaps, fit_measures = ""
    # nothing left
    return(numeric(0L))
  }

  # select output type
  if (output == "list") {
    # nothing to do
  } else if (output == "vector") {
    out <- unlist(out)
    class(out) <- c("lavaan.vector", "numeric")
  } else if (output == "matrix") {
    out <- as.matrix(unlist(out))
    colnames(out) <- ""
    class(out) <- c("lavaan.matrix", "matrix")
  } else if (output == "text") {
    out <- unlist(out)
    class(out) <- c("lavaan.fitMeasures", "lavaan.vector", "numeric")
  }

  # attributes?
  # only if fit_measures == "all" or "default"
  if (length(fit_measures_orig) == 1L &&
      fit_measures_orig %in% c("all", "default")) {
    x2_label <- test[[test_idx]]$label # NULL if "standard"
    x2_baseline_label <- object@baseline$test[[test_idx]]$label
    attr(out, "X2.label") <- x2_label
    attr(out, "X2.baseline.label") <- x2_baseline_label
    if (standard_test != "standard") {
      attr(out, "standard.test") <- standard_test
    }
    if (scaled_flag && scaled_test != "standard") {
      attr(out, "scaled.test") <- scaled_test
    }
  }

  out
}

# fit measures for models with random slopes (rv() modifier)
#
# no chi-square test statistic (or fit indices) exist for these
# models (the likelihood is conditional on the covariates carrying
# random slopes: there is no comparable saturated model); we provide
# the loglikelihood-based measures, and the H0 scaling correction
# factor (cfr. the Mplus MLR output) -- the latter is included in
# the default set only when se = "robust.huber.white", but can
# always be requested explicitly
lav_fit_rv <- function(object, fit_measures = "all",
                       output = "vector") {
  loglik <- object@loglik
  robust_flag <- object@Options$se == "robust.huber.white"

  indices <- list()
  indices["npar"] <- loglik$npar
  indices["ntotal"] <- loglik$ntotal
  indices["logl"] <- loglik$loglik
  indices["aic"] <- loglik$AIC
  indices["bic"] <- loglik$BIC
  indices["bic2"] <- loglik$BIC2

  available <- c(names(indices), "scaling.factor.h0")

  if (length(fit_measures) == 1L &&
      fit_measures %in% c("all", "default")) {
    fit_measures <- names(indices)
    if (robust_flag) {
      fit_measures <- c(fit_measures, "scaling.factor.h0")
    }
  } else {
    fit_measures <- tolower(fit_measures)
    bad <- fit_measures[!fit_measures %in% available]
    if (length(bad) > 0L) {
      lav_msg_stop(gettextf(
        "fit measure(s) not available for models with random slopes:
         %s.", paste(bad, collapse = " ")))
    }
  }

  # H0 scaling correction factor: c = tr(H^{-1} J) / npar
  if ("scaling.factor.h0" %in% fit_measures) {
    rs <- object@Cache[[1]]$rs
    if (is.null(rs)) {
      rs_info <- lav_mvn_cl_rs_info(
        lavmodel = object@Model, lavdata = object@Data
      )
      rs <- list(
        info = rs_info,
        stats = lav_mvn_cl_rs_stats(
          y1 = object@Data@X[[1]], lp = object@Data@Lp[[1]],
          rs_info = rs_info
        )
      )
    }
    indices["scaling.factor.h0"] <-
      lav_mvn_cl_rs_scaling_h0(lavobject = object, rs = rs)
  }

  # keep only what we need (in the requested order)
  out <- indices[fit_measures]

  # select output type
  if (output == "list") {
    # nothing to do
  } else if (output == "vector") {
    out <- unlist(out)
    class(out) <- c("lavaan.vector", "numeric")
  } else if (output == "matrix") {
    out <- as.matrix(unlist(out))
    colnames(out) <- ""
    class(out) <- c("lavaan.matrix", "matrix")
  } else if (output == "text") {
    out <- unlist(out)
    class(out) <- c("lavaan.fitMeasures", "lavaan.vector", "numeric")
  }

  out
}

# print a nice summary of the fit measures
lav_fitmeasures_print <- function(x, ..., nd = 3L, add_h0 = TRUE) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, FALSE)

  # optional header (eg for sam objects, where the fit measures refer to the
  # structural part only)
  if (!is.null(attr(x, "header"))) {
    cat("\n", attr(x, "header"), "\n", sep = "")
  }

  names_x <- names(x)

  # scaled?
  scaled_flag <- "chisq.scaled" %in% names_x

  # num format
  num_format <- paste("%", max(8L, nd + 5L), ".", nd, "f", sep = "")

  ## TDJ: optionally add h0 model's fit statistic, for lavaan.mi
  if (add_h0 && "chisq" %in% names_x) {
    cat("\nModel Test User Model:\n\n")

    # container three columns
    c1 <- c2 <- c3 <- character(0L)

    # TDJ: Add header used in summary() by lavaan.mi
    if (scaled_flag) {
      c1 <- c("", c1)
      c2 <- c("Standard", c2)
      c3 <- c("Scaled", c3)
    }

    # if test is not standard, add label
    if (!is.null(attr(x, "X2.label"))) {
      c1 <- c(c1, attr(x, "X2.label"))
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }

    c1 <- c(c1, "Test statistic")
    c2 <- c(c2, sprintf(num_format, x["chisq"]))
    c3 <- c(c3, ifelse(scaled_flag,
                       sprintf(num_format, x["chisq.scaled"]), ""
    ))

    c1 <- c(c1, "Degrees of freedom")
    c2 <- c(c2, x["df"])
    c3 <- c(c3, ifelse(scaled_flag,
                       ifelse(x["df.scaled"] %% 1 == 0, x["df.scaled"],
                              sprintf(num_format, x["df.scaled"])
                       ), ""
    ))

    c1 <- c(c1, "P-value")
    c2 <- c(c2, sprintf(num_format, x["pvalue"]))
    c3 <- c(c3, ifelse(scaled_flag,
                       sprintf(num_format, x["pvalue.scaled"]), ""
    ))

    if (scaled_flag && "chisq.scaling.factor" %in% names_x) {
      c1 <- c(c1, "Average scaling correction factor")
      c2 <- c(c2, "")
      c3 <- c(c3, sprintf(num_format, x["chisq.scaling.factor"]))

      ## check for shift parameter
      chisq_shift_parameter <- attr(attr(x, "header"), "shift")
      if (!is.null(chisq_shift_parameter)) {
        c1 <- c(c1, "Average shift parameter",
                "  simple second-order correction")
        c2 <- c(c2, "", "")
        ## This is only provided by the fitMeasures() method for lavaan.mi-class
        c3 <- c(c3, sprintf(num_format, chisq_shift_parameter),
                "")
      }
    }

    # check for lavaan.mi attributes "pool.method" and "pool.robust"
    if (!is.null(attr(x, "pool.method"))) {
      ## extract information from lavaan.mi object about
      ## the method used to pool the test statistic
      pool_method   <- attr(x, "pool.method")
      pool_robust   <- attr(x, "pool.robust")
      standard_test <- attr(attr(x, "header"), "standard.test")
      scaled_test   <- attr(attr(x, "header"),   "scaled.test")

      c1 <- c(c1, "Pooling method")
      c2 <- c(c2, pool_method)
      c3 <- c(c3, "")

      ## (conditionally for D2 method) add other pooling information
      if (scaled_flag) {
        c1 <- c(c1, "  Pooled statistic")
        c2 <- c(c2, ifelse(pool_robust, dQuote(scaled_test),
                                        dQuote(standard_test)))
        c3 <- c(c3, "")

        if (pool_robust && pool_method == "D2") {
          c1 <- c(c1, paste0("  ", dQuote(scaled_test), " correction applied"))
          c2 <- c(c2, "BEFORE")
          c3 <- c(c3, "pooling")
        } else {
          c1 <- c(c1, paste0("  ", dQuote(scaled_test), " correction applied"))
          c2 <- c(c2, "AFTER")
          c3 <- c(c3, "pooling")
        }

      }
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 35L)
    c2 <- format(c2, width = 16L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # print information about standard.test? (new in 0.6-21)
  # (only if "standard.test" is not "standard")
  if (!is.null(attr(x, "standard.test"))) {
    cat("\nNote: fit measures based on the chi-square test statistic\n",
        "     use", attr(x, "X2.label"), "\n")
  }

  # independence model
  if ("baseline.chisq" %in% names_x) {
    cat("\nModel Test Baseline Model:\n\n")

    c1 <- c2 <- c3 <- character(0L)
    # if test is not standard, add label
    if (!is.null(attr(x, "X2.baseline.label"))) {
      c1 <- c(c1, attr(x, "X2.label"))
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }

    c1 <- c(c1, "Test statistic")
    c2 <- c(c2, sprintf(num_format, x["baseline.chisq"]))
    c3 <- c(c3, ifelse(scaled_flag,
                       sprintf(num_format, x["baseline.chisq.scaled"]), ""
    ))

    c1 <- c(c1, "Degrees of freedom")
    c2 <- c(c2, x["baseline.df"])
    c3 <- c(c3, ifelse(scaled_flag,
                  ifelse(x["baseline.df.scaled"] %% 1 == 0,
                         x["baseline.df.scaled"],
                         sprintf(num_format, x["baseline.df.scaled"])
    ),
    ""
    ))

    c1 <- c(c1, "P-value")
    c2 <- c(c2, sprintf(num_format, x["baseline.pvalue"]))
    c3 <- c(c3, ifelse(scaled_flag,
                       sprintf(num_format, x["baseline.pvalue.scaled"]), ""
    ))

    if (scaled_flag && "baseline.chisq.scaling.factor" %in% names_x) {
      c1 <- c(c1, "Scaling correction factor")
      c2 <- c(c2, "")
      c3 <- c(c3, sprintf(num_format, x["baseline.chisq.scaling.factor"]))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 35L)
    c2 <- format(c2, width = 16L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # cfi/tli
  if (any(c("cfi", "tli", "nnfi", "rfi", "nfi", "ifi", "rni", "pnfi") %in%
                                                                 names_x)) {
    cat("\nUser Model versus Baseline Model:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    if ("cfi" %in% names_x) {
      c1 <- c(c1, "Comparative Fit Index (CFI)")
      c2 <- c(c2, sprintf(num_format, x["cfi"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["cfi.scaled"]), ""
      ))
    }

    if ("tli" %in% names_x) {
      c1 <- c(c1, "Tucker-Lewis Index (TLI)")
      c2 <- c(c2, sprintf(num_format, x["tli"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["tli.scaled"]), ""
      ))
    }

    if ("cfi.robust" %in% names_x) {
      c1 <- c(c1, "")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
      c1 <- c(c1, "Robust Comparative Fit Index (CFI)")
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["cfi.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["cfi.robust"]))
        c3 <- c(c3, "")
      }
    }

    if ("tli.robust" %in% names_x) {
      c1 <- c(c1, "Robust Tucker-Lewis Index (TLI)")
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["tli.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["tli.robust"]))
        c3 <- c(c3, "")
      }
    }

    if ("nnfi" %in% names_x) {
      c1 <- c(c1, "Bentler-Bonett Non-normed Fit Index (NNFI)")
      c2 <- c(c2, sprintf(num_format, x["nnfi"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["nnfi.robust"]), ""
      ))
    }

    if ("nfi" %in% names_x) {
      c1 <- c(c1, "Bentler-Bonett Normed Fit Index (NFI)")
      c2 <- c(c2, sprintf(num_format, x["nfi"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["nfi.scaled"]), ""
      ))
    }

    if ("pnfi" %in% names_x) {
      c1 <- c(c1, "Parsimony Normed Fit Index (PNFI)")
      c2 <- c(c2, sprintf(num_format, x["pnfi"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["pnfi.scaled"]), ""
      ))
    }

    if ("rfi" %in% names_x) {
      c1 <- c(c1, "Bollen's Relative Fit Index (RFI)")
      c2 <- c(c2, sprintf(num_format, x["rfi"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["rfi.scaled"]), ""
      ))
    }

    if ("ifi" %in% names_x) {
      c1 <- c(c1, "Bollen's Incremental Fit Index (IFI)")
      c2 <- c(c2, sprintf(num_format, x["ifi"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["ifi.scaled"]), ""
      ))
    }

    if ("rni" %in% names_x) {
      c1 <- c(c1, "Relative Noncentrality Index (RNI)")
      c2 <- c(c2, sprintf(num_format, x["rni"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["rni.robust"]), ""
      ))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # likelihood
  if ("logl" %in% names_x) {
    cat("\nLoglikelihood and Information Criteria:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    c1 <- c(c1, "Loglikelihood user model (H0)")
    c2 <- c(c2, sprintf(num_format, x["logl"]))
    c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["logl"]), ""))
    if (!is.na(x["scaling.factor.h0"])) {
      c1 <- c(c1, "Scaling correction factor")
      if (scaled_flag) {
        c2 <- c(c2, sprintf("  %10s", ""))
        c3 <- c(c3, sprintf(num_format, x["scaling.factor.h0"]))
      } else {
        # no scaled test statistic (eg random slopes): single column
        c2 <- c(c2, sprintf(num_format, x["scaling.factor.h0"]))
        c3 <- c(c3, "")
      }
      c1 <- c(c1, "    for the MLR correction")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }

    if ("unrestricted.logl" %in% names_x) {
      c1 <- c(c1, "Loglikelihood unrestricted model (H1)")
      c2 <- c(c2, sprintf(num_format, x["unrestricted.logl"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["unrestricted.logl"]), ""
      ))
      if (!is.na(x["scaling.factor.h1"])) {
        c1 <- c(c1, "Scaling correction factor")
        c2 <- c(c2, sprintf("  %10s", ""))
        c3 <- c(c3, sprintf(num_format, x["scaling.factor.h1"]))
        c1 <- c(c1, "    for the MLR correction")
        c2 <- c(c2, "")
        c3 <- c(c3, "")
      }
    }

    c1 <- c(c1, "")
    c2 <- c(c2, "")
    c3 <- c(c3, "")
    # c1 <- c(c1, "Number of free parameters")
    # c2 <- c(c2, sprintf("  %10i", x["npar"]))
    # c3 <- c(c3, ifelse(scaled, sprintf("  %10i", x["npar"]), ""))

    c1 <- c(c1, "Akaike (AIC)")
    c2 <- c(c2, sprintf(num_format, x["aic"]))
    c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["aic"]), ""))
    c1 <- c(c1, "Bayesian (BIC)")
    c2 <- c(c2, sprintf(num_format, x["bic"]))
    c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["bic"]), ""))
    if (!is.na(x["bic2"])) {
      c1 <- c(c1, "Sample-size adjusted Bayesian (SABIC)")
      c2 <- c(c2, sprintf(num_format, x["bic2"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["bic2"]), ""))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 39L)
    c2 <- format(c2, width = 12L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # RMSEA
  if ("rmsea" %in% names_x) {
    cat("\nRoot Mean Square Error of Approximation:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    c1 <- c(c1, "RMSEA")
    c2 <- c(c2, sprintf(num_format, x["rmsea"]))
    c3 <- c(c3, ifelse(scaled_flag,
                       sprintf(num_format, x["rmsea.scaled"]), ""
    ))

    ci_level <- NULL
    if ("rmsea.ci.level" %in% names_x) {
      ci_level <- x["rmsea.ci.level"]
    }
    if ("rmsea.ci.lower" %in% names_x) {
      if (is.null(ci_level)) {
        c1 <- c(c1, "Confidence interval - lower")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(ci_level * 100)),
          " Percent confidence interval - lower"
        ))
      }
      c2 <- c(c2, sprintf(num_format, x["rmsea.ci.lower"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["rmsea.ci.lower.scaled"]), ""
      ))
      if (is.null(ci_level)) {
        c1 <- c(c1, "Confidence interval - upper")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(ci_level * 100)),
          " Percent confidence interval - upper"
        ))
      }
      c2 <- c(c2, sprintf(num_format, x["rmsea.ci.upper"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["rmsea.ci.upper.scaled"]), ""
      ))
    }

    rmsea_close_h0 <- NULL
    if ("rmsea.close.h0" %in% names_x) {
      rmsea_close_h0 <- x["rmsea.close.h0"]
    }
    rmsea_notclose_h0 <- NULL
    if ("rmsea.notclose.h0" %in% names_x) {
      rmsea_notclose_h0 <- x["rmsea.notclose.h0"]
    }
    if ("rmsea.pvalue" %in% names_x) {
      if (is.null(rmsea_close_h0)) {
        c1 <- c(c1, "P-value H_0: RMSEA <= 0.05")
      } else {
        c1 <- c(c1, paste0(
          "P-value H_0: RMSEA <= ",
          sprintf("%4.3f", rmsea_close_h0)
        ))
      }
      c2 <- c(c2, sprintf(num_format, x["rmsea.pvalue"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["rmsea.pvalue.scaled"]), ""
      ))
    }
    if ("rmsea.notclose.pvalue" %in% names_x) {
      if (is.null(rmsea_notclose_h0)) {
        c1 <- c(c1, "P-value H_0: RMSEA >= 0.080")
      } else {
        c1 <- c(c1, paste0(
          "P-value H_0: RMSEA >= ",
          sprintf("%4.3f", rmsea_notclose_h0)
        ))
      }
      c2 <- c(c2, sprintf(num_format, x["rmsea.notclose.pvalue"]))
      c3 <- c(c3, ifelse(scaled_flag,
                    sprintf(num_format, x["rmsea.notclose.pvalue.scaled"]), ""
      ))
    }

    # robust
    if ("rmsea.robust" %in% names_x) {
      c1 <- c(c1, "")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
      c1 <- c(c1, "Robust RMSEA")
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["rmsea.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["rmsea.robust"]))
        c3 <- c(c3, "")
      }
    }
    if ("rmsea.ci.lower.robust" %in% names_x) {
      if (is.null(ci_level)) {
        c1 <- c(c1, "Confidence interval - lower")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(ci_level * 100)),
          " Percent confidence interval - lower"
        ))
      }
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["rmsea.ci.lower.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["rmsea.ci.lower.robust"]))
        c3 <- c(c3, "")
      }

      if (is.null(ci_level)) {
        c1 <- c(c1, "Confidence interval - upper")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(ci_level * 100)),
          " Percent confidence interval - upper"
        ))
      }
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["rmsea.ci.upper.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["rmsea.ci.upper.robust"]))
        c3 <- c(c3, "")
      }
    }
    if ("rmsea.pvalue.robust" %in% names_x) {
      if (is.null(rmsea_close_h0)) {
        c1 <- c(c1, "P-value H_0: Robust RMSEA <= 0.05")
      } else {
        c1 <- c(c1, paste0(
          "P-value H_0: Robust RMSEA <= ",
          sprintf("%4.3f", rmsea_close_h0)
        ))
      }
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["rmsea.pvalue.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["rmsea.pvalue.robust"]))
        c3 <- c(c3, "")
      }
    }
    if ("rmsea.notclose.pvalue.robust" %in% names_x) {
      if (is.null(rmsea_notclose_h0)) {
        c1 <- c(c1, "P-value H_0: Robust RMSEA >= 0.080")
      } else {
        c1 <- c(c1, paste0(
          "P-value H_0: Robust RMSEA >= ",
          sprintf("%4.3f", rmsea_notclose_h0)
        ))
      }
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(
          c3,
          sprintf(num_format, x["rmsea.notclose.pvalue.robust"])
        )
      } else {
        c2 <- c(
          c2,
          sprintf(num_format, x["rmsea.notclose.pvalue.robust"])
        )
        c3 <- c(c3, "")
      }
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # SRMR
  #TODO: add CRMR
  if (any(c("rmr", "srmr") %in% names_x) && !"srmr_within" %in% names_x) {
    cat("\nStandardized Root Mean Square Residual:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    if ("rmr" %in% names_x) {
      c1 <- c(c1, "RMR")
      c2 <- c(c2, sprintf(num_format, x["rmr"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["rmr"]), ""))
    }
    if ("rmr_nomean" %in% names_x) {
      c1 <- c(c1, "RMR (No Mean)")
      c2 <- c(c2, sprintf(num_format, x["rmr_nomean"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["rmr_nomean"]), ""
      ))
    }
    if ("srmr" %in% names_x) {
      c1 <- c(c1, "SRMR")
      c2 <- c(c2, sprintf(num_format, x["srmr"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["srmr"]), ""))
    }
    if ("srmr_nomean" %in% names_x) {
      c1 <- c(c1, "SRMR (No Mean)")
      c2 <- c(c2, sprintf(num_format, x["srmr_nomean"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["srmr_nomean"]), ""
      ))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # SRMR -- multilevel
  #TODO: add CRMR?
  if (any(c("srmr_within", "srmr_between") %in% names_x)) {
    cat("\nStandardized Root Mean Square Residual (corr metric):\n\n")

    c1 <- c2 <- c3 <- character(0L)

    if ("srmr_within" %in% names_x) {
      c1 <- c(c1, "SRMR (within covariance matrix)")
      c2 <- c(c2, sprintf(num_format, x["srmr_within"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["srmr_within"]), ""
      ))
    }
    if ("srmr_between" %in% names_x) {
      c1 <- c(c1, "SRMR (between covariance matrix)")
      c2 <- c(c2, sprintf(num_format, x["srmr_between"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["srmr_between"]), ""
      ))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # WRMR
  if ("wrmr" %in% names_x) {
    cat("\nWeighted Root Mean Square Residual:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    if ("wrmr" %in% names_x) {
      c1 <- c(c1, "WRMR")
      c2 <- c(c2, sprintf(num_format, x["wrmr"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["wrmr"]), ""))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # GFI (the 'new' GFI of Maydeu-Olivares et al., 2024)
  if ("gfi" %in% names_x) {
    cat("\nGoodness of Fit Index:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    c1 <- c(c1, "Goodness of Fit Index (GFI)")
    c2 <- c(c2, sprintf(num_format, x["gfi"]))
    c3 <- c(c3, "")

    gfi_ci_level <- NULL
    if ("gfi.ci.level" %in% names_x) {
      gfi_ci_level <- x["gfi.ci.level"]
    }
    if ("gfi.ci.lower" %in% names_x) {
      if (is.null(gfi_ci_level)) {
        c1 <- c(c1, "Confidence interval - lower")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(gfi_ci_level * 100)),
          " Percent confidence interval - lower"
        ))
      }
      c2 <- c(c2, sprintf(num_format, x["gfi.ci.lower"]))
      c3 <- c(c3, "")
      if (is.null(gfi_ci_level)) {
        c1 <- c(c1, "Confidence interval - upper")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(gfi_ci_level * 100)),
          " Percent confidence interval - upper"
        ))
      }
      c2 <- c(c2, sprintf(num_format, x["gfi.ci.upper"]))
      c3 <- c(c3, "")
    }

    # robust
    if ("gfi.robust" %in% names_x) {
      c1 <- c(c1, "")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
      c1 <- c(c1, "Robust GFI")
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["gfi.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["gfi.robust"]))
        c3 <- c(c3, "")
      }
    }
    if ("gfi.ci.lower.robust" %in% names_x) {
      if (is.null(gfi_ci_level)) {
        c1 <- c(c1, "Confidence interval - lower")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(gfi_ci_level * 100)),
          " Percent confidence interval - lower"
        ))
      }
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["gfi.ci.lower.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["gfi.ci.lower.robust"]))
        c3 <- c(c3, "")
      }
      if (is.null(gfi_ci_level)) {
        c1 <- c(c1, "Confidence interval - upper")
      } else {
        c1 <- c(c1, paste0(
          sprintf("%2d", round(gfi_ci_level * 100)),
          " Percent confidence interval - upper"
        ))
      }
      if (scaled_flag) {
        c2 <- c(c2, "")
        c3 <- c(c3, sprintf(num_format, x["gfi.ci.upper.robust"]))
      } else {
        c2 <- c(c2, sprintf(num_format, x["gfi.ci.upper.robust"]))
        c3 <- c(c3, "")
      }
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # Other
  if (any(c("cn_05", "cn_01", "gfi_lisrel", "agfi_lisrel", "pgfi", "mfi")
          %in% names_x)) {
    cat("\nOther Fit Indices:\n\n")

    c1 <- c2 <- c3 <- character(0L)

    if ("cn_05" %in% names_x) {
      c1 <- c(c1, "Hoelter Critical N (CN) alpha = 0.05")
      c2 <- c(c2, sprintf(num_format, x["cn_05"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["cn_05"]), ""
      ))
    }
    if ("cn_01" %in% names_x) {
      c1 <- c(c1, "Hoelter Critical N (CN) alpha = 0.01")
      c2 <- c(c2, sprintf(num_format, x["cn_01"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["cn_01"]), ""
      ))
    }
    if (any(c("cn_05", "cn_01") %in% names_x)) {
      c1 <- c(c1, "")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }
    if ("gfi_lisrel" %in% names_x) {
      c1 <- c(c1, "Goodness of Fit Index LISREL (GFI)")
      c2 <- c(c2, sprintf(num_format, x["gfi_lisrel"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["gfi_lisrel"]), ""
      ))
    }
    if ("agfi_lisrel" %in% names_x) {
      c1 <- c(c1, "Adjusted Goodness of Fit Index LISREL (AGFI)")
      c2 <- c(c2, sprintf(num_format, x["agfi_lisrel"]))
      c3 <- c(c3, ifelse(scaled_flag,
                         sprintf(num_format, x["agfi_lisrel"]), ""
      ))
    }
    if ("pgfi" %in% names_x) {
      c1 <- c(c1, "Parsimony Goodness of Fit Index (PGFI)")
      c2 <- c(c2, sprintf(num_format, x["pgfi"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["pgfi"]), ""))
    }
    if (any(c("gfi_lisrel", "agfi_lisrel", "pgfi") %in% names_x)) {
      c1 <- c(c1, "")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }
    if ("mfi" %in% names_x) {
      c1 <- c(c1, "McDonald Fit Index (MFI)")
      c2 <- c(c2, sprintf(num_format, x["mfi"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["mfi"]), ""))
    }
    if ("mfi" %in% names_x) {
      c1 <- c(c1, "")
      c2 <- c(c2, "")
      c3 <- c(c3, "")
    }
    if ("ecvi" %in% names_x) {
      c1 <- c(c1, "Expected Cross-Validation Index (ECVI)")
      c2 <- c(c2, sprintf(num_format, x["ecvi"]))
      c3 <- c(c3, ifelse(scaled_flag, sprintf(num_format, x["ecvi"]), ""))
    }

    # format c1/c2/c3
    c1 <- format(c1, width = 43L)
    c2 <- format(c2, width = 8L + max(0, (nd - 3L)) * 4L, justify = "right")
    c3 <- format(c3, width = 8L + nd, justify = "right")

    # create character matrix
    if (scaled_flag) {
      m <- cbind(c1, c2, c3, deparse.level = 0)
    } else {
      m <- cbind(c1, c2, deparse.level = 0)
    }
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  invisible(x)
}

#
# the lavData class describes how the data looks like
#  - do we have a full data frame, or only sample statistics?
#    (TODO: allow for patterns + freq, if data is categorical)
#  - variable type ("numeric", "ordered", ...)
#  - how many groups, how many observations, ...
#  - what about missing patterns?
#
# initial version: YR 14 April 2012

# YR 23 Feb 2017: blocks/levels/groups, but everything is group-based!

# FIXME: if nlevels > 1L, and ngroups > 1L, we should check that
# group is at the upper-level

# YR 08 May  2019: sampling weights normalization -> different options

# extract the data we need for this particular model
lav_lavdata <- function(data = NULL, # data.frame
                    group = NULL, # multiple groups?
                    cluster = NULL, # clusters?
                    ov_names = NULL, # variables in model
                    ov_names_x = character(0), # exo variables
                    ov_names_l = list(), # names per level
                    ov_names_aux = character(0), # auxiliary variables
                    ordered = NULL, # ordered variables
                    sampling_weights = NULL, # sampling weights
                    sample_cov = NULL, # sample covariance(s)
                    sample_mean = NULL, # sample mean vector(s)
                    sample_th = NULL, # sample thresholds
                    sample_nobs = NULL, # sample nobs
                    lavoptions = lavOptions(), # lavoptions
                    allow_single_case = FALSE # for newdata in predict
) {
  # get info from lavoptions

  # group.labels
  group_label <- lavoptions$group.label
  if (is.null(group_label)) {
    group_label <- character(0L)
  }

  # level.labels
  level_label <- lavoptions$level.label
  if (is.null(level_label)) {
    level_label <- character(0L)
  }

  # allow empty categories of ordinal variable
  allow_empty_cell <- lavoptions$allow.empty.cell

  # block.labels
  block_label <- character(0L)
  if (length(group_label) > 0L && length(level_label) == 0L) {
    block_label <- group_label
  } else if (length(level_label) > 0L && length(group_label) == 0L) {
    block_label <- level_label
  } else if (length(group_label) > 0L &&
    length(level_label) > 0L) {
    block_label <- paste(rep(group_label, each = length(level_label)),
      rep(level_label, times = length(group_label)),
      sep = "."
    )
  }

  # std.ov?
  std_ov <- lavoptions$std.ov
  if (is.null(std_ov)) {
    std_ov <- FALSE
  }
  # partial correlation structure: only these (observed) variables are
  # standardized to unit variance (empty -> all, the usual std.ov behavior)
  correlation_ov <- lavoptions$.correlation.ov
  if (is.null(correlation_ov)) {
    correlation_ov <- character(0L)
  }
  if (length(correlation_ov) > 0L && !is.null(ov_names)) {
    all_ov <- unique(unlist(ov_names))
    bad <- correlation_ov[!(correlation_ov %in% all_ov)]
    if (length(bad) > 0L) {
      lav_msg_warn(gettextf(
        "correlation= argument contains unknown observed variable(s): %s",
        paste(bad, collapse = ", ")))
    }
  }

  # missing? (lav_object_cor() does not parse options before calling lavdata...)
  missing <- tolower(lavoptions$missing)
  if (is.null(missing) || missing == "default") {
    missing <- "listwise"
  } else if (missing %in% c("ml", "fiml", "direct")) {
    missing <- "ml"
  } else if (missing %in% c("ml.x", "fiml.x", "direct.x")) {
    missing <- "ml.x"
  }

  # warn?
  if (allow_single_case) { # eg, in lavPredict
    current_warn <- lav_warn()
    if (lav_warn(FALSE))
        on.exit(lav_warn(current_warn), TRUE)
  }

  # four scenarios:
  #    0) data is already a lavData object: do nothing
  #    1) data is full data.frame (or a matrix)
  #    2) data are sample statistics only
  #    3) no data at all

  # 1) full data
  if (!is.null(data)) {
    # catch lavaan/lavData objects
    if (inherits(data, "lavData")) {
      return(data)
    } else if (inherits(data, "lavaan")) {
      return(data@Data)
    }

    # catch matrix
    if (!is.data.frame(data)) {
      # is it a matrix?
      if (is.matrix(data)) {
        if (nrow(data) == ncol(data)) {
          # perhaps it is a covariance matrix?
          if (isSymmetric(data)) {
            lav_msg_warn(
              gettext("data argument looks like a covariance matrix;
              please use the sample_cov= argument instead"))
          }
        }
        # or perhaps it is a data matrix?
        ### FIXME, we should avoid as.data.frame() and handle
        ### data matrices directly
        data <- as.data.frame(data, stringsAsFactors = FALSE)
      } else {
        lav_msg_stop(gettextf(
          "data= argument is not a data.frame, but of class %s",
          sQuote(class(data))))
      }
    }

    # no ov.names?
    if (is.null(ov_names)) {
      ov_names <- names(data)
      # remove group variable, if provided
      if (length(group) > 0L) {
        group_idx <- which(ov_names == group)
        ov_names <- ov_names[-group_idx]
      }
      # remove cluster variable, if provided
      if (length(cluster) > 0L) {
        cluster_idx <- which(ov_names == cluster)
        ov_names <- ov_names[-cluster_idx]
      }
    }

    lav_data <- lav_data_full(
      data = data,
      group = group,
      cluster = cluster,
      group_label = group_label,
      level_label = level_label,
      block_label = block_label,
      ov_names = ov_names,
      ordered = ordered,
      sampling_weights = sampling_weights,
      sampling_weights_normalization =
        lavoptions$sampling.weights.normalization,
      ov_names_x = ov_names_x,
      ov_names_l = ov_names_l,
      ov_names_aux = ov_names_aux,
      std_ov = std_ov,
      correlation_ov = correlation_ov,
      missing = missing,
      allow_single_case = allow_single_case,
      allow_empty_cell = allow_empty_cell
    )
    sample_cov <- NULL # not needed, but just in case
  }


  # 2) sample moments
  if (is.null(data) && !is.null(sample_cov)) {
    # for now: no levels!!
    nlevels <- 1L

    # we also need the number of observations (per group)
    if (is.null(sample_nobs)) {
      lav_msg_stop(gettext("please specify number of observations"))
    }

    # if a 'group' argument was provided, keep it -- new in 0.6-4
    if (is.null(group)) {
      group <- character(0L)
    } else if (is.character(group)) {
      # nothing to do, just store it
    } else {
      lav_msg_stop(gettext("group argument should be a string"))
    }

    # list?
    if (is.list(sample_cov)) {
      # multiple groups, multiple cov matrices
      if (!is.null(sample_mean) &&
          (!is.list(sample_mean) ||
           length(sample_mean) != length(sample_cov))) {
        lav_msg_stop(gettextf(
          "sample.cov is a list of %s matrices; sample.mean should be a
          list of the same length.", length(sample_cov)))
      }
      if (!is.null(sample_th) &&
          (!is.list(sample_th) ||
           length(sample_th) != length(sample_cov))) {
        lav_msg_stop(gettextf(
          "sample.cov is a list of %s matrices; sample.th should be a
          list of the same length.", length(sample_cov)))
      }
      # multiple groups, multiple cov matrices
      ngroups <- length(sample_cov)
      label <- names(sample_cov)
      if (is.null(group_label) || length(group_label) == 0L) {
        if (is.null(label)) {
          group_label <- paste("Group ", 1:ngroups, sep = "")
        } else {
          group_label <- label
        }
      } else {
        if (is.null(label)) {
          if (length(group_label) != ngroups) {
            lav_msg_stop(gettextf(
              "length(group.label) = %1$s but sample.cov implies
              ngroups = %2$s.", length(group_label), ngroups))
          }
        } else {
          # FIXME!!!!
          # check if they match
        }
      }
    } else {
      ngroups <- 1L
      group_label <- character(0)
      if (!is.matrix(sample_cov)) {
        lav_msg_stop(gettext(
          "sample_cov must be a matrix or a list of matrices"))
      }
      sample_cov <- list(sample_cov)
    }

    # check sample_nobs: we need one value per group
    sample_nobs <- as.list(sample_nobs)
    if (length(sample_nobs) != ngroups) {
      lav_msg_stop(gettextf(
        "length(sample.nobs) = %1$s but sample.cov implies ngroups = %2$s;
        please provide the number of observations for each group.",
        length(sample_nobs), ngroups))
    }

    # get ov.names
    if (is.null(ov_names)) {
      ov_names <- lapply(sample_cov, row.names)
    } else if (!is.list(ov_names)) {
      # duplicate ov.names for each group
      tmp <- ov_names
      ov_names <- vector("list", length = ngroups)
      ov_names[1:ngroups] <- list(tmp)
    } else {
      if (length(ov_names) != ngroups) {
        lav_msg_stop(gettextf(
          "ov.names assumes %1$s groups; data contains %2$s groups",
          length(ov_names), ngroups))
      }
      # nothing to do
    }

    # handle ov.names.x
    if (!is.list(ov_names_x)) {
      tmp <- ov_names_x
      ov_names_x <- vector("list", length = ngroups)
      ov_names_x[1:ngroups] <- list(tmp)
    } else {
      if (length(ov_names_x) != ngroups) {
        lav_msg_stop(gettextf(
          "ov.names.x assumes %1$s groups; data contains %2$s groups",
          length(ov_names_x), ngroups))
      }
    }

    ov <- list()
    ov$name <- unique(unlist(c(ov_names, ov_names_x)))
    nvar <- length(ov$name)
    ov$idx <- rep(NA, nvar)
    ov$nobs <- rep(sum(unlist(sample_nobs)), nvar)
    ov$type <- rep("numeric", nvar)
    ov$nlev <- rep(0, nvar)
    # check for categorical
    if (!is.null(sample_th)) {
      th_idx <- attr(sample_th, "th.idx")
      if (is.list(th_idx)) {
        th_idx <- th_idx[[1]] ## FIRST group only (assuming same ths!)
      }
      if (any(th_idx > 0)) {
        tab <- table(th_idx[th_idx > 0])
        ord_idx <- as.numeric(names(tab))
        nlev <- as.integer(unname(tab) + 1)
        ov$type[ord_idx] <- "ordered"
        ov$nlev[ord_idx] <- nlev
      }
    }

    # if std.ov = TRUE, give a warning (suggested by Peter Westfall)
    if (std_ov && !lavoptions$correlation) {
      lav_msg_warn(gettext(
        "std.ov argument is ignored if only sample statistics are provided."))
    }

    # check variances (new in 0.6-7)
    if (!allow_single_case) {
      for (g in seq_len(ngroups)) {
        var_1 <- diag(sample_cov[[g]])
        # 1. finite?
        if (!all(is.finite(var_1))) {
          lav_msg_stop(gettext(
          "at least one variance in the sample
          covariance matrix is not finite."))
          }
        # 2. near zero (or negative)?
        if (any(var_1 < .Machine$double.eps)) {
          lav_msg_stop(
            gettext("at least one variance in the sample covariance matrix is
            (near) zero or negative."))
        }
        # 3. very large?
        max_var <- max(var_1)
        if (max_var > 1000000) {
          lav_msg_warn(
            gettext("some observed variances in the sample covariance matrix
            are larger than 1000000."))
        }
      }
    }

    # block.labels
    block_label <- character(0L)
    if (length(group_label) > 0L && length(level_label) == 0L) {
      block_label <- group_label
    } else if (length(level_label) > 0L && length(group_label) == 0L) {
      block_label <- level_label
    } else if (length(group_label) > 0L &&
      length(level_label) > 0L) {
      block_label <- paste(rep(group_label, each = length(level_label)),
        rep(level_label, times = length(group_label)),
        sep = "."
      )
    }

    # construct lavData object
    lav_data <- new("lavData",
      data.type = "moment",
      ngroups = ngroups,
      group = group,
      nlevels = 1L, # for now
      cluster = character(0L),
      group.label = group_label,
      level.label = character(0L),
      block.label = block_label,
      nobs = as.list(sample_nobs),
      norig = as.list(sample_nobs),
      ov.names = ov_names,
      ov.names.x = ov_names_x,
      ov.names.l = ov_names_l,
      ordered = as.character(ordered),
      weights = vector("list", length = ngroups),
      sampling.weights = character(0L),
      ov = ov,
      std.ov = FALSE,
      missing = "listwise",
      case.idx = vector("list", length = ngroups),
      Mp = vector("list", length = ngroups),
      Rp = vector("list", length = ngroups),
      Lp = vector("list", length = ngroups),
      X = vector("list", length = ngroups),
      eXo = vector("list", length = ngroups)
    )
  }

  # 3) data.type = "none":  both data and sample_cov are NULL
  if (is.null(data) && is.null(sample_cov)) {
    # clustered/multilevel? --> ov.names.l should be filled in
    if (length(ov_names_l) > 0L) {
      nlevels <- length(ov_names_l[[1]]) # we assume the same number
      # of levels in each group!

      # do we have a cluster argument? if not, create one
      if (is.null(cluster)) {
        if (nlevels == 2L) {
          cluster <- "cluster"
        } else {
          cluster <- paste0("cluster", seq_len(nlevels - 1L))
        }
      }

      # default level.labels
      if (length(level_label) == 0L) {
        level_label <- c("within", cluster)
      } else {
        # check if length(level.label) = 1 + length(cluster)
        if (length(level_label) != length(cluster) + 1L) {
          lav_msg_stop(gettext("length(level.label) != length(cluster) + 1L"))
        }
        # nothing to do
      }
    } else {
      nlevels <- 1L
      cluster <- character(0L)
      level_label <- character(0L)
    }

    # ngroups: ov.names (when group: is used), or sample_nobs
    if (is.null(ov_names)) {
      lav_msg_warn(gettext("ov.names is NULL"))
      ov_names <- character(0L)
      if (is.null(sample_nobs)) {
        ngroups <- 1L
        sample_nobs <- rep(list(0L), ngroups)
      } else {
        sample_nobs <- as.list(sample_nobs)
        ngroups <- length(sample_nobs)
      }
    } else if (!is.list(ov_names)) {
      if (is.null(sample_nobs)) {
        ngroups <- 1L
        sample_nobs <- rep(list(0L), ngroups)
      } else {
        sample_nobs <- as.list(sample_nobs)
        ngroups <- length(sample_nobs)
      }
      ov_names <- rep(list(ov_names), ngroups)
    } else if (is.list(ov_names)) {
      ngroups <- length(ov_names)
      if (is.null(sample_nobs)) {
        sample_nobs <- rep(list(0L), ngroups)
      } else {
        sample_nobs <- as.list(sample_nobs)
        if (length(sample_nobs) != ngroups) {
          lav_msg_stop(gettextf(
            "length(sample_nobs) = %1$s but syntax implies ngroups = %2$s",
            length(sample_nobs), ngroups))
        }
      }
    }


    # group.label
    if (ngroups > 1L) {
      if (is.null(group)) {
        group <- "group"
      }
      group_label <- paste("Group", 1:ngroups, sep = "")
    } else {
      group <- character(0L)
      group_label <- character(0L)
    }

    # handle ov.names.x
    if (!is.list(ov_names_x)) {
      ov_names_x <- rep(list(ov_names_x), ngroups)
    }

    # handle ov.names.l: we need a model for each group
    if (length(ov_names_l) == 1L && ngroups > 1L) {
      lav_msg_stop(gettextf(
        "this analysis assumes %s groups, but the multilevel model syntax is
        specified for a single group only; combining multiple groups with
        multiple levels requires explicit group: blocks in the model
        syntax, one for each group (for now).", ngroups))
    }

    ov <- list()
    ov$name <- unique(unlist(c(ov_names, ov_names_x)))
    nvar <- length(ov$name)
    ov$idx <- rep(NA, nvar)
    ov$nobs <- rep(0L, nvar)
    ov$type <- rep("numeric", nvar)
    ov$nlev <- rep(0L, nvar)

    # collect information per upper-level group
    lp <- vector("list", length = ngroups)
    for (g in 1:ngroups) {
      if (nlevels > 1L) {
        # ALWAYS add ov.names.x at the end, even if conditional.x
        ov_names_1 <- unique(c(ov_names[[g]], ov_names_x[[g]]))
        lp[[g]] <- lav_data_cl_patterns(
          y = NULL, clus = NULL,
          cluster = cluster,
          multilevel = TRUE,
          ov_names = ov_names_1,
          ov_names_x = ov_names_x[[g]],
          ov_names_l = ov_names_l[[g]]
        )
      }
    } # g

    # block.labels
    block_label <- character(0L)
    if (length(group_label) > 0L && length(level_label) == 0L) {
      block_label <- group_label
    } else if (length(level_label) > 0L && length(group_label) == 0L) {
      block_label <- level_label
    } else if (length(group_label) > 0L &&
      length(level_label) > 0L) {
      block_label <- paste(rep(group_label, each = length(level_label)),
        rep(level_label, times = length(group_label)),
        sep = "."
      )
    }

    # construct lavData object
    lav_data <- new("lavData",
      data.type = "none",
      ngroups = ngroups,
      group = group,
      nlevels = nlevels,
      cluster = cluster,
      group.label = group_label,
      level.label = level_label,
      block.label = block_label,
      nobs = sample_nobs,
      norig = sample_nobs,
      ov.names = ov_names,
      ov.names.x = ov_names_x,
      ov.names.l = ov_names_l,
      ordered = as.character(ordered),
      weights = vector("list", length = ngroups),
      sampling.weights = character(0L),
      ov = ov,
      missing = "listwise",
      case.idx = vector("list", length = ngroups),
      Mp = vector("list", length = ngroups),
      Rp = vector("list", length = ngroups),
      Lp = lp,
      X = vector("list", length = ngroups),
      eXo = vector("list", length = ngroups)
    )
  }

  lav_data
}


# validate the aux= argument and return the names of the auxiliary variables
# that can actually be used in the current configuration
#
# auxiliary variables are observed variables that are NOT part of the model;
# they are never part of the model-implied summary statistics. For now, they
# are only used to improve the EM-based h1 (saturated) moments when
# missing = "ml" (continuous data). In any other configuration, they are
# ignored (with a note).
#
# binary or ordered/categorical auxiliary variables are not supported and are
# removed (with a warning).
lav_data_aux_check <- function(aux        = NULL,
                               data       = NULL,
                               ov_names   = NULL,
                               ordered    = NULL,
                               lavoptions = NULL) {
  # nothing to do?
  if (is.null(aux)) {
    return(character(0L))
  }
  if (!is.character(aux)) {
    lav_msg_stop(gettext("aux= argument must be a character vector."))
  }
  aux <- unique(aux[nchar(aux) > 0L])
  if (length(aux) == 0L) {
    return(character(0L))
  }

  # we can only use auxiliary variables if we have raw data, the model is
  # continuous, and missing is one of the EM-based options
  missing <- tolower(lavoptions$missing)
  aux_missing <- c("ml", "ml.x", "two.stage", "robust.two.stage")
  if (is.null(data) || !(is.data.frame(data) || is.matrix(data)) ||
      isTRUE(lavoptions$categorical) ||
      !missing %in% aux_missing) {
    lav_msg_note(gettext(
      "auxiliary (aux) variables are currently only used (to improve the
       EM-based h1/saturated summary statistics) when missing = \"ml\",
       \"two.stage\" or \"robust.two.stage\" with continuous data;
       they will be ignored here."))
    return(character(0L))
  }

  # drop aux names that are not usable (not in data, model variables, or
  # binary/ordered-categorical); shared with the FIML saturated-correlates path
  aux <- lav_aux_clean_names(
    aux = aux, data = data, ov_names = ov_names, ordered = ordered
  )
  if (length(aux) == 0L) {
    return(character(0L))
  }

  # note: for the two-stage (and robust two-stage) standard errors, the
  # auxiliary variables are accounted for via the augmented stage-1 ACOV
  # (Savalei & Bentler, 2009; see lav_model_nvcov_two_stage()), except when
  # fixed-x covariates are present, in which case the standard errors fall
  # back to the model-only ACOV

  aux
}


# handle full data
lav_data_full <- function(data = NULL, # data.frame
                          group = NULL, # multiple groups?
                          cluster = NULL, # clustered?
                          group_label = NULL, # custom group labels?
                          level_label = NULL,
                          block_label = NULL,
                          ov_names = NULL, # variables needed
                          # in model
                          ordered = NULL, # ordered variables
                          sampling_weights = NULL, # sampling weights
                          sampling_weights_normalization = "none",
                          ov_names_x = character(0L), # exo variables
                          ov_names_l = list(), # var per level
                          ov_names_aux = character(0L), # auxiliary variables
                          std_ov = FALSE, # standardize ov's?
                          correlation_ov = character(0L), # std only these
                          missing = "listwise", # remove missings?
                          allow_single_case = FALSE, # allow single case?
                          allow_empty_cell = FALSE
) {
  # number of groups and group labels
  if (!is.null(group) && length(group) > 0L) {
    if (length(group) > 1L) {
      lav_msg_stop(gettext(
        "lavaan supports only a single grouping variable; if you wish to
        combine several grouping variables, create a single new variable
        first (e.g., using interaction())."))
    }
    if (!(group %in% names(data))) {
      lav_msg_stop(gettextf(
        "grouping variable %1$s not found; variable names
        found in data frame are: %2$s",
        sQuote(group), paste(names(data), collapse = " ")))
    }
    # note: by default, we use the order as in the data;
    # not as in levels(data[,group])
    if (length(group_label) == 0L) {
      group_label <- unique(as.character(data[[group]]))
      if (any(is.na(group_label))) {
        lav_msg_warn(gettextf("group variable %s contains missing values",
                              sQuote(group)))
      }
      group_label <- group_label[!is.na(group_label)]
    } else {
      group_label <- unique(as.character(group_label))
      # check if user-provided group labels exist
      label <- unique(as.character(data[[group]]))
      idx <- match(group_label, label)
      if (any(is.na(idx))) {
        lav_msg_warn(gettextf(
          "some group.labels do not appear in the grouping variable: %s",
          lav_msg_view(group_label[which(is.na(idx))], log_sep = "none"))
        )
      }
      group_label <- group_label[!is.na(idx)]
      # any groups left?
      if (length(group_label) == 0L) {
        lav_msg_stop(gettext(
          "no group levels left; check the group.label argument"))
      }
    }
    ngroups <- length(group_label)
  } else {
    if (length(group_label) > 0L) {
      lav_msg_warn(gettext(
       "`group.label' argument will be ignored if `group' argument is missing"))
    }
    ngroups <- 1L
    group_label <- character(0L)
    group <- character(0L)
  }

  # ensure allow.empty.cell is logical
  if (is.null(allow_empty_cell)) allow_empty_cell <- FALSE

  # sampling weights
  if (!is.null(sampling_weights)) {
    if (is.character(sampling_weights)) {
      if (length(sampling_weights) > 1L) {
        lav_msg_stop(gettext(
          "lavaan supports only a single sampling weights variable."))
      }
      if (!(sampling_weights %in% names(data))) {
        lav_msg_stop(
          gettextf("sampling weights variable %1$s not found;
                   variable names found in data frame are: %2$s",
          sQuote(sampling_weights), paste(names(data), collapse = " ")))
      }
      # check for missing values in sampling weight variable
      if (any(is.na(data[[sampling_weights]]))) {
        lav_msg_stop(
          gettextf("sampling.weights variable %s contains missing values",
          sQuote(sampling_weights)))
      }
    } else {
      lav_msg_stop(gettext(
        "sampling weights argument should be a variable name in the data.frame"
        ))
    }
  }

  # clustered?
  if (!is.null(cluster) && length(cluster) > 0L) {
    # only a single clustering variable is supported (two-level models)
    if (length(cluster) > 1L) {
      lav_msg_stop(gettext(
        "lavaan supports only a single clustering variable; if you wish to
        combine several clustering variables, create a single new variable
        first (e.g., using interaction())."))
    }
    # cluster variable in data?
    if (!all(cluster %in% names(data))) {
      # which one did we not find?
      not_ok <- which(!cluster %in% names(data))
      lav_msg_stop(gettextf(
        "cluster variable(s) %1$s not found;
        variable names found in data frame are: %2$s",
        sQuote(cluster[not_ok]), paste(names(data), collapse = " ")))
    }

    # check for missing values in cluster variable(s)
    for (cl in seq_along(cluster)) {
      if (anyNA(data[[cluster[cl]]])) {
        lav_msg_warn(gettextf("cluster variable %s contains missing values",
          sQuote(cluster[cl])))
      }
    }

    # multilevel?
    if (length(ov_names_l) > 0L) {
      # default level.labels
      if (length(level_label) == 0L) {
        level_label <- c("within", cluster)
      } else {
        # check if length(level.label) = 1 + length(cluster)
        if (length(level_label) != length(cluster) + 1L) {
          lav_msg_stop(gettext("length(level.label) != length(cluster) + 1L"))
        }
        # nothing to do
      }
      nlevels <- length(level_label)
    } else {
      # just clustered data, but no random effects
      nlevels <- 1L
      level_label <- character(0L)
    }
  } else {
    if (length(level_label) > 0L) {
      lav_msg_warn(gettext(
       "`level.label' argument will be ignored if `cluster' argument is missing"
      ))
    }
    nlevels <- 1L
    level_label <- character(0L)
    cluster <- character(0L)
  }

  # check ov.names vs ngroups
  if (ngroups > 1L) {
    if (is.list(ov_names)) {
      if (length(ov_names) != ngroups) {
        lav_msg_stop(gettextf(
          "ov.names assumes %1$s groups; data contains %2$s groups",
          length(ov_names), ngroups))
      }
    } else {
      tmp <- ov_names
      ov_names <- vector("list", length = ngroups)
      ov_names[1:ngroups] <- list(tmp)
    }
    if (is.list(ov_names_x)) {
      if (length(ov_names_x) != ngroups) {
        lav_msg_stop(gettextf(
          "ov.names.x assumes %1$s groups; data contains %2$s groups",
          length(ov_names_x), ngroups))
      }
    } else {
      tmp <- ov_names_x
      ov_names_x <- vector("list", length = ngroups)
      ov_names_x[1:ngroups] <- list(tmp)
    }
    # multilevel: we need a model (with level: blocks) for each group
    if (length(ov_names_l) > 0L && length(ov_names_l) != ngroups) {
      if (length(ov_names_l) == 1L) {
        lav_msg_stop(gettextf(
          "the data contains %s groups, but the multilevel model syntax is
          specified for a single group only; combining multiple groups with
          multiple levels requires explicit group: blocks in the model
          syntax, one for each group (for now).", ngroups))
      } else {
        lav_msg_stop(gettextf(
          "ov.names.l assumes %1$s groups; data contains %2$s groups",
          length(ov_names_l), ngroups))
      }
    }
  } else {
    if (is.list(ov_names)) {
      if (length(ov_names) > 1L) {
        lav_msg_stop(gettext(
          "model syntax defines multiple groups; data suggests a single group"))
      }
    } else {
      ov_names <- list(ov_names)
    }
    if (is.list(ov_names_x)) {
      if (length(ov_names_x) > 1L) {
        lav_msg_stop(gettext(
          "model syntax defines multiple groups; data suggests a single group"))
      }
    } else {
      ov_names_x <- list(ov_names_x)
    }
  }

  # check if all ov.names can be found in the data.frame
  for (g in 1:ngroups) {
    # does the data contain all the observed variables
    # needed in the user-specified model for this group
    ov_all <- unique(c(ov_names[[g]], ov_names_x[[g]])) # no overlap if categ

    # handle interactions
    ov_int_names <- ov_all[grepl(":", ov_all, fixed = TRUE)]
    n_int <- length(ov_int_names)
    if (n_int > 0L) {
      ov_names_noint <- ov_all[!ov_all %in% ov_int_names]
      for (iv in seq_len(n_int)) {
        names_1 <- strsplit(ov_int_names[iv], ":", fixed = TRUE)[[1L]]
        if (all(names_1 %in% ov_names_noint)) {
          # add this interaction term to the data.frame, unless
          # it already exists
          if (is.null(data[[ov_int_names[iv]]])) {
            data[[ov_int_names[iv]]] <-
              data[[names_1[1L]]] * data[[names_1[2L]]]
          }
        }
      }
    }

    # check for missing observed variables
    idx_missing <- which(!(ov_all %in% names(data)))

    if (length(idx_missing)) {
      lav_msg_stop(
        gettextf("some (observed) variables specified in the model are
        not found in the dataset: %s",
        paste(ov_all[idx_missing], collapse = " ")))
    }
  }


  # here, we know for sure all ov.names exist in the data.frame
  # create varTable
  # FIXME: should we add the 'group'/'cluster' variable (no for now)
  ov <- lav_dataframe_vartable(
    frame = data, ov_names = ov_names,
    ov_names_x = ov_names_x, ordered = ordered,
    as_data_frame = FALSE, allow_empty_cell = allow_empty_cell
  )

  # variable table for the auxiliary/instrument variables (their type and
  # number of levels are needed, eg by the categorical IV estimator, to compute
  # the augmented polychoric sample statistics)
  ov_aux <- list()
  if (length(ov_names_aux) > 0L) {
    ov_aux <- lav_dataframe_vartable(
      frame = data, ov_names = ov_names_aux,
      ordered = ordered, as_data_frame = FALSE,
      allow_empty_cell = allow_empty_cell
    )
  }

  # do some checking
  # check for character variables that are not declared as ordered;
  # data.matrix() would silently convert them to (alphabetical) integer
  # codes and treat them as continuous (new in 0.7-1)
  if ("character" %in% ov$type) {
    c_names <- ov$name[ov$type == "character"]
    lav_msg_stop(gettextf(
      "Some model variables are of type character: %s. Please convert
      these variables to numeric values, or declare them as ordered
      categorical using the ordered= argument.",
      lav_msg_view(c_names, log_sep = "none")))
  }
  # check for unordered factors (but only if nlev > 2)
  if ("factor" %in% ov$type) {
    f_names <- ov$name[ov$type == "factor" & ov$nlev > 2L]
    f_names_all <- ov$name[ov$type == "factor"]
    ov_names_1 <- unlist(ov_names)
    ov_names_x_1 <- unlist(ov_names_x)
    ov_names_nox <- ov_names_1[!ov_names_1 %in% ov_names_x_1]
    if (any(f_names %in% ov_names_x_1)) {
      lav_msg_stop(
        gettext("unordered factor(s) with more than 2 levels detected
            as exogenous covariate(s): "),
        paste(f_names, collapse = " "))
    } else if (any(f_names_all %in% ov_names_nox)) {
      lav_msg_stop(
        gettext("unordered factor(s) detected; make them numeric or ordered:"),
        paste(f_names_all, collapse = " "))
    }
  }
  # check for ordered exogenous variables
  if ("ordered" %in% ov$type[ov$name %in% unlist(ov_names_x)]) {
    f_names <- ov$name[ov$type == "ordered" &
      ov$name %in% unlist(ov_names_x)]
    if (any(f_names %in% unlist(ov_names_x))) {
      lav_msg_warn(gettextf(
        "exogenous variable(s) declared as ordered in data: %s",
        lav_msg_view(f_names, log_sep = "none")))
    }
  }
  # check for ordered endogenous variables with more than 12 levels
  if ("ordered" %in% ov$type[!ov$name %in% unlist(ov_names_x)]) {
    f_names <- ov$name[ov$type == "ordered" &
      !ov$name %in% unlist(ov_names_x) &
      ov$nlev > 12L]
    if (length(f_names) > 0L) {
      lav_msg_warn(gettextf(
        "some ordered categorical variable(s) have more than 12 levels: %s",
        lav_msg_view(f_names, log_sep = "none")))
    }
  }
  # check for zero-cases
  idx <- which(ov$nobs == 0L | ov$var == 0)
  if (!allow_single_case && length(idx) > 0L) {
    ov_1 <- as.data.frame(ov)
    rn <- rownames(ov_1)
    rn[idx] <- paste(rn[idx], "***", sep = "")
    rownames(ov_1) <- rn
    print(ov_1)
    lav_msg_stop(gettext(
      "some variables have no values (only missings) or no variance"))
  }
  # check for single cases (no variance!)
  idx <- which(ov$nobs == 1L | (ov$type == "numeric" & !is.finite(ov$var)))
  if (!allow_single_case && length(idx) > 0L) {
    ov_1 <- as.data.frame(ov)
    rn <- rownames(ov_1)
    rn[idx] <- paste(rn[idx], "***", sep = "")
    rownames(ov_1) <- rn
    print(ov_1)
    lav_msg_stop(gettext(
      "some variables have only 1 observation or no finite variance"))
  }
  # check for ordered variables with only 1 level
  idx <- which(ov$type == "ordered" & ov$nlev == 1L)
  if (!allow_single_case && length(idx) > 0L) {
    ov_1 <- as.data.frame(ov)
    rn <- rownames(ov_1)
    rn[idx] <- paste(rn[idx], "***", sep = "")
    rownames(ov_1) <- rn
    print(ov_1)
    lav_msg_stop(gettext("ordered variable(s) has/have only 1 level"))
  }
  # check for mix small/large variances (NOT including exo variables)
  if (!std_ov && !allow_single_case && any(ov$type == "numeric")) {
    num_idx <- which(ov$type == "numeric" & ov$exo == 0L)
    if (length(num_idx) > 0L) {
      min_var <- min(ov$var[num_idx])
      max_var <- max(ov$var[num_idx])
      rel_var <- max_var / min_var
      if (rel_var > 1000) {
        lav_msg_warn(
          gettext("some observed variances are (at least) a factor 1000 times
           larger than others; use varTable(fit) to investigate"))
      }
    }
  }
  # check for really large variances (perhaps -999999 for missing?)
  if (!allow_single_case && !std_ov && any(ov$type == "numeric")) {
    num_idx <- which(ov$type == "numeric" & ov$exo == 0L)
    if (length(num_idx) > 0L) {
      max_var <- max(ov$var[num_idx])
      if (max_var > 1000000) {
        lav_msg_warn(
          gettext("some observed variances are larger than 1000000
          use varTable(fit) to investigate"))
      }
    }
  }
  # check for all-exogenous variables (eg in f <~ x1 + x2 + x3)
  if (all(ov$exo == 1L)) {
    lav_msg_warn(gettext(
      "all observed variables are exogenous; model may not be identified"))
  }
  # check for perfect correlations (NOT including exo variables)
  if (!allow_single_case && any(ov$type == "numeric")) {
    num_idx <- which(ov$type == "numeric" & ov$exo == 0L)
    cor_1 <- try(cor(data[, ov$idx[num_idx]], use = "pairwise.complete.obs"),
               silent = TRUE)
    # replace any NAs by 0 (as we only wish to detect perfect correlations)
    cor_1[is.na(cor_1)] <- 0
    if (!inherits(cor_1, "try-error") &&
        any(lav_mat_vech(cor_1, diagonal = FALSE) == 1)) {
      cor_1[upper.tri(cor_1, diag = TRUE)] <- 0
      idx <- which(cor_1 == 1)
      this_names <- ov$name[num_idx]
      bad_names <- this_names[sort(unique(c(row(cor_1)[idx],
                              col(cor_1)[idx]))) ]
      # should we make this a hard stop? things will most likely fail later...
      lav_msg_warn(gettextf(
      "some observed variables are perfectly correlated;
       please check your data; variables involved are: %s",
       paste(bad_names, collapse = " ")))
    }
  }

  # prepare empty lists

  # group-based
  case_idx <- vector("list", length = ngroups)
  mp <- vector("list", length = ngroups)
  rp <- vector("list", length = ngroups)
  norig <- vector("list", length = ngroups)
  nobs <- vector("list", length = ngroups)
  x <- vector("list", length = ngroups)
  exo <- vector("list", length = ngroups)
  aux <- vector("list", length = ngroups)
  lp <- vector("list", length = ngroups)
  weights <- vector("list", length = ngroups)

  # collect information per upper-level group
  # datam <- data.matrix(data) # YR: not yet, this breaks the stuart package!
  for (g in 1:ngroups) {
    # extract variables in correct order
    if (nlevels > 1L) {
      # keep 'joint' (Y,X) matrix in @X if multilevel (or always?)
      # yes for multilevel (for now); no for clustered only
      ov_names_2 <- unique(c(ov_names[[g]], ov_names_x[[g]]))
      ov_idx <- ov$idx[match(ov_names_2, ov$name)]
    } else {
      ov_idx <- ov$idx[match(ov_names[[g]], ov$name)]
    }
    exo_idx <- ov$idx[match(ov_names_x[[g]], ov$name)]
    all_idx <- unique(c(ov_idx, exo_idx))

    # extract cases per group
    if (ngroups > 1L || length(group_label) > 0L) {
      if (missing == "listwise") {
        case_idx[[g]] <- which(data[[group]] == group_label[g] &
          complete.cases(data[all_idx]))
        nobs[[g]] <- length(case_idx[[g]])
        norig[[g]] <- length(which(data[[group]] == group_label[g]))
        # check for empty data
        if (nobs[[g]] == 0L) {
          lav_msg_stop(gettextf(
            "all observations were deleted due to missing
             data after listwise deletion in group [%s];
             check your data or consider a different option
             for the missing= argument.", group_label[g]))
        }
        # } else if(missing == "pairwise" && length(exo.idx) > 0L) {
        #    case.idx[[g]] <- which(data[[group]] == group.label[g] &
        #                           complete.cases(data[exo.idx]))
        #    nobs[[g]] <- length(case.idx[[g]])
        #    norig[[g]] <- length(which(data[[group]] == group.label[g]))
      } else if (length(exo_idx) > 0L && missing != "ml.x") {
        case_idx[[g]] <- which(data[[group]] == group_label[g] &
          complete.cases(data[exo_idx]))
        nobs[[g]] <- length(case_idx[[g]])
        norig[[g]] <- length(which(data[[group]] == group_label[g]))
        if ((nobs[[g]] < norig[[g]])) {
          lav_msg_warn(gettextf(
              "%1$s cases were deleted in group %2$s  due to missing values
              in  exogenous variable(s), while fixed.x = TRUE.",
              (norig[[g]] - nobs[[g]]), group_label[g]))
        }
      } else {
        case_idx[[g]] <- which(data[[group]] == group_label[g])
        nobs[[g]] <- norig[[g]] <- length(case_idx[[g]])
      }
    } else {
      if (missing == "listwise") {
        case_idx[[g]] <- which(complete.cases(data[all_idx]))
        nobs[[g]] <- length(case_idx[[g]])
        norig[[g]] <- nrow(data)
        if (nobs[[g]] == 0L) {
          lav_msg_stop(gettext(
            "all observations were deleted due to missing
             data after listwise deletion;
             check your data or consider a different option
             for the missing= argument."))
        }
        # } else if(missing == "pairwise" && length(exo.idx) > 0L) {
        #    case.idx[[g]] <- which(complete.cases(data[exo.idx]))
        #    nobs[[g]] <- length(case.idx[[g]])
        #    norig[[g]] <- nrow(data)
      } else if (length(exo_idx) > 0L && missing != "ml.x") {
        case_idx[[g]] <- which(complete.cases(data[exo_idx]))
        nobs[[g]] <- length(case_idx[[g]])
        norig[[g]] <- nrow(data)
        if ((nobs[[g]] < norig[[g]])) {
          lav_msg_warn(
            gettextf("%s cases were deleted due to missing values in
                     exogenous variable(s), while fixed.x = TRUE.",
                     (norig[[g]] - nobs[[g]])))
        }
      } else {
        case_idx[[g]] <- seq_len(nrow(data))
        nobs[[g]] <- norig[[g]] <- length(case_idx[[g]])
      }
    }

    # extract data
    #X[[g]] <- datam[case.idx[[g]], ov.idx, drop = FALSE]
    x[[g]] <- data.matrix(data[case_idx[[g]], ov_idx, drop = FALSE])
    dimnames(x[[g]]) <- NULL ### copy?

    # sampling weights (but no normalization yet)
    if (!is.null(sampling_weights)) {
      wt <- data[[sampling_weights]][case_idx[[g]]]
      if (any(wt < 0)) {
        lav_msg_stop(gettext("some sampling weights are negative"))
      }

      # check for missing values in sampling weight variable
      if (any(is.na(wt))) {
        lav_msg_stop(gettextf(
          "sampling.weights variable %s contains missing values",
          sQuote(sampling_weights)))
      }

      # check for non-finite values (e.g., Inf) in sampling weight variable
      # (new in 0.7-1)
      if (any(!is.finite(wt))) {
        lav_msg_stop(gettextf(
          "Sampling.weights variable %s contains non-finite values.",
          sQuote(sampling_weights)))
      }

      # check that the sampling weights do not sum to zero (new in 0.7-1)
      if (sum(wt) == 0) {
        if (ngroups > 1L) {
          lav_msg_stop(gettextf(
            "Sampling weights sum to zero in group %s.", group_label[g]))
        } else {
          lav_msg_stop(gettext("Sampling weights sum to zero."))
        }
      }

      weights[[g]] <- wt
    }

    # construct integers for user-declared 'ordered' factors
    # FIXME: is this really (always) needed???
    #  (but still better than doing lapply(data[,idx], ordered) which
    #   generated even more copies)
    user_ordered_names <- ov$name[ov$type == "ordered" & ov$user == 1L]
    user_ordered_idx <- which(ov_names[[g]] %in% user_ordered_names)
    if (length(user_ordered_idx) > 0L) {
      for (i in user_ordered_idx) {
        x[[g]][, i][is.na(x[[g]][, i])] <- NA # change NaN to NA
        if (!allow_empty_cell) x[[g]][, i] <- as.numeric(as.factor(x[[g]][, i]))
        # possible alternative to the previous two lines:
        # X[[g]][,i] <- as.numeric(factor(X[[g]][,i], exclude = c(NA, NaN)))
      }
    }

    ## FIXME:
    ## - why also in X? (for samplestats, for now)
    if (length(exo_idx) > 0L) {
      exo[[g]] <- data.matrix(data[case_idx[[g]], exo_idx, drop = FALSE])
      dimnames(exo[[g]]) <- NULL
    } else {
      exo[g] <- list(NULL)
    }

    # auxiliary variables (not part of the model); kept here only so that
    # downstream code (eg the EM-based h1 moments) can use them; they never
    # enter the model-implied summary statistics. Rows are aligned with
    # case.idx[[g]], hence with x[[g]].
    if (length(ov_names_aux) > 0L) {
      aux[[g]] <- data.matrix(data[case_idx[[g]], ov_names_aux, drop = FALSE])
      dimnames(aux[[g]]) <- NULL
    } else {
      aux[g] <- list(NULL)
    }

    # standardize observed variables? numeric only!
    if (std_ov) {
      num_idx <- which(ov$name %in% ov_names[[g]] &
        ov$type == "numeric" & ov$exo == 0L)
      # partial correlation structure: only standardize the requested
      # variables, leave the others in their original metric
      if (length(correlation_ov) > 0L && length(num_idx) > 0L) {
        num_idx <- num_idx[ov$name[num_idx] %in% correlation_ov]
      }
      if (length(num_idx) > 0L) {
        x[[g]][, num_idx] <-
          scale(x[[g]][, num_idx, drop = FALSE])[, , drop = FALSE]
        # three copies are made!!!!!
      }
      if (length(exo_idx) > 0L) {
        exo[[g]] <- scale(exo[[g]])[, , drop = FALSE]
      }
    }

    # response patterns (ordered variables only)
    ord_idx <- which(ov_names[[g]] %in% ov$name[ov$type == "ordered"])
    if (length(ord_idx) > 0L) {
      rp[[g]] <- lav_data_resp_patterns(x[[g]][, ord_idx, drop = FALSE])
    }

    # warn if we have a small number of observations (but NO error!)
    if (!allow_single_case &&
      nobs[[g]] < (nvar <- length(ov_idx))) {
      txt <- ""
      if (ngroups > 1L) txt <- gettextf("in group %s", g)
      lav_msg_warn(
        gettextf("small number of observations (nobs < nvar) %1$s:
                 nobs = %2$s  nvar = %3$s", txt, nobs[[g]], nvar))
    }
    # check variances per group (if we have multiple groups)
    # to catch zero-variance variables within a group (new in 0.6-8)
    if (ngroups > 1L && !allow_empty_cell) {
      # X
      group_var <- apply(x[[g]], 2, var, na.rm = TRUE)
      zero_var <- which(group_var < .Machine$double.eps)
      if (length(zero_var) == 0L) {
        # all is good
      } else {
        # some zero variances!
        gtxt <- if (ngroups > 1L) {
          gettextf("in group %s", g)
        } else {
          ""
        }
       lav_msg_stop(
          gettext("some variables have no variance"), gtxt,
          ":", paste(ov_names[[g]][zero_var], collapse = " "))
      }

      # eXo (if conditional.x = TRUE)...
      if (length(exo_idx) > 0L) {
        group_var <- apply(exo[[g]], 2, var, na.rm = TRUE)
        zero_var <- which(group_var < .Machine$double.eps)
        if (length(zero_var) == 0L) {
          # all is good
        } else {
          # some zero variances!
          gtxt <- if (ngroups > 1L) {
            gettextf("in group %s", g)
          } else {
            ""
          }
          lav_msg_stop(
            gettext("some exogenous variables have no variance"), gtxt,
            ":", paste(ov_names_x[[g]][zero_var], collapse = " ")
          )
        }
      }
    }

    # cluster information
    if (length(cluster) > 0L) {
      # extract cluster variable(s), for this group
      clus <- data.matrix(data[case_idx[[g]], cluster])
      if (nlevels > 1L) {
        multilevel <- TRUE
      } else {
        multilevel <- FALSE
      }
      # ALWAYS add ov.names.x at the end, even if conditional.x (0.6-7)
      ov_names_2 <- unique(c(ov_names[[g]], ov_names_x[[g]]))
      lp[[g]] <- lav_data_cl_patterns(
        y = x[[g]], clus = clus,
        cluster = cluster,
        multilevel = multilevel,
        ov_names = ov_names_2,
        ov_names_x = ov_names_x[[g]],
        ov_names_l = ov_names_l[[g]]
      )

      # check that we have at least two clusters (per group); with a
      # single cluster, the between-level moments are undefined
      # (new in 0.7-1)
      for (l in 2:length(lp[[g]]$nclusters)) {
        if (lp[[g]]$nclusters[[l]] < 2L) {
          gtxt <- if (ngroups > 1L) {
            gettextf(" in group %s", dQuote(group_label[g]))
          } else {
            ""
          }
          lav_msg_stop(gettextf(
            "Cluster variable %1$s contains only %2$s cluster(s)%3$s;
            at least two clusters are required to fit a model to
            clustered data.",
            sQuote(cluster[l - 1L]), lp[[g]]$nclusters[[l]], gtxt))
        }
      }

      # new in 0.6-4
      # check for 'level-1' variables with zero within variance
      l1_idx <- c(
        lp[[g]]$within.idx[[2]], # within only
        lp[[g]]$both.idx[[2]]
      )
      for (v in l1_idx) {
        within_var <- tapply(x[[g]][, v], lp[[g]]$cluster.idx[[2]],
          FUN = var, na.rm = TRUE
        )
        # ignore singletons
        singleton_idx <- which(lp[[g]]$cluster.size[[2]] == 1L)
        if (length(singleton_idx) > 0L) {
          within_var[singleton_idx] <- 10 # non-zero variance
        }
        zero_var <- which(within_var < .Machine$double.eps)
        if (length(zero_var) == 0L) {
          # all is good
        } else if (length(zero_var) == length(within_var)) {
          # all zero! possibly a between-level variable
          gtxt <- if (ngroups > 1L) {
            gettextf("in group %s", g)
          } else {
            ""
          }
          lav_msg_warn(
            gettextf("Level-1 variable %1$s has no variance at the within
                     level %2$s. The variable appears to be a between-level
                     variable. Please remove this variable from the level 1
                     section in the model syntax.",
                     dQuote(ov_names[[g]][v]), gtxt))
        } else {
          # some zero variances!
          gtxt <- if (ngroups > 1L) {
            gettextf("in group %s", g)
          } else {
            ""
          }
          lav_msg_warn(gettextf(
          "Level-1 variable %1$s has no variance within some clusters %2$s.
          The cluster ids with zero within variance are: %3$s.",
          dQuote(ov_names[[g]][v]), gtxt,
          lav_msg_view(lp[[g]]$cluster.id[[2]][zero_var], "none")))
        }
      }

      # new in 0.6-4
      # check for 'level-2' only variables with non-zero within variance
      l2_idx <- lp[[g]]$between.idx[[2]] # between only
      error_flag <- FALSE
      for (v in l2_idx) {
        within_var <- tapply(x[[g]][, v], lp[[g]]$cluster.idx[[2]],
          FUN = var, na.rm = TRUE
        )
        non_zero_var <- which(unname(within_var) > .Machine$double.eps)
        if (length(non_zero_var) == 0L) {
          # all is good
        } else if (length(non_zero_var) == 1L) {
          # just one
          gtxt <- if (ngroups > 1L) {
            gettextf("in group %s.", g)
          } else {
            "."
          }
          lav_msg_warn(gettextf(
            "Level-2 variable %1$s has non-zero variance at the within
            level %2$s in one cluster with id: %3$s. Please double-check
            if this is a between only variable.",
            dQuote(ov_names[[g]][v]), gtxt,
            lp[[g]]$cluster.id[[2]][non_zero_var]))
        } else {
          error_flag <- TRUE
          # several
          gtxt <- if (ngroups > 1L) {
            gettextf("in group %s", g)
          } else {
            ""
          }
          lav_msg_warn(gettextf(
            "Level-2 variable %1$s has non-zero variance at the within level
            %2$s. The cluster ids with non-zero within variance are: %3$s",
            dQuote(ov_names[[g]][v]), gtxt,
            lav_msg_view(lp[[g]]$cluster.id[[2]][non_zero_var], "none")))
        }
      }
      if (error_flag) {
        lav_msg_stop(
          gettext("Some between-level (only) variables have
          non-zero variance at the within-level.
          Please double-check your data.")
        )
      }
    } # clustered data

    # missing data
    if (missing != "listwise") {
      if (length(cluster) > 0L) {
        # get missing patterns
        mp[[g]] <- lav_data_mi_patterns(x[[g]],
          sort_freq = TRUE, coverage = TRUE,
          lp = lp[[g]]
        )
      } else {
        # get missing patterns
        mp[[g]] <- lav_data_mi_patterns(x[[g]],
          sort_freq = TRUE, coverage = TRUE,
          lp = NULL
        )
      }

      # checking!
      if (length(mp[[g]]$empty.idx) > 0L) {
        # new in 0.6-4: return 'original' index in full data.frame
        empty_case_idx <- case_idx[[g]][mp[[g]]$empty.idx]
        lav_msg_warn(gettextf(
          "some cases are empty and will be ignored: %s.",
          paste(empty_case_idx, collapse = " ")))
      }
    if (any(mp[[g]]$coverage < 0.1)) {
      coverage_vech <- lav_mat_vech(mp[[g]]$coverage, diagonal = FALSE)
      small_idx <- which(coverage_vech < 0.1)
      if (all(coverage_vech[small_idx] == 0)) {
      # 0.6-18: no warning --> this could be due to missing by design
          # 0.6-20: give warning anyway (as EM is ignoring this)
          lav_msg_warn(gettext(
            "due to missing values, some pairwise combinations have zero
             coverage; the corresponding covariances are not identified;
             use lavInspect(fit, \"coverage\") to investigate."))
    } else {
          lav_msg_warn(gettext(
            "due to missing values, some pairwise combinations have less than
            10% coverage; use lavInspect(fit, \"coverage\") to investigate."))
      }
      }
      # in case we had observations with only missings
      nobs[[g]] <- NROW(x[[g]]) - length(mp[[g]]$empty.idx)
    } # missing
  } # groups, at first level

  # sampling weights, again
  if (is.null(sampling_weights)) {
    sampling_weights <- character(0L)
  } else {
    # check if we need normalization
    if (sampling_weights_normalization == "none") {
      # nothing to do
    } else if (sampling_weights_normalization == "total") {
      sum_weights <- sum(unlist(weights))
      ntotal <- sum(unlist(nobs))
      for (g in 1:ngroups) {
        wt <- weights[[g]]
        wt2 <- wt / sum_weights * ntotal
        weights[[g]] <- wt2
      }
    } else if (sampling_weights_normalization == "group") {
      for (g in 1:ngroups) {
        wt <- weights[[g]]
        wt2 <- wt / sum(wt) * nobs[[g]]
        weights[[g]] <- wt2
      }
    } else {
      lav_msg_stop(gettext(
        "sampling.weights.normalization should be total, group or none."))
    }
  }

  # block.labels
  block_label <- character(0L)
  if (length(group_label) > 0L && length(level_label) == 0L) {
    block_label <- group_label
  } else if (length(level_label) > 0L && length(group_label) == 0L) {
    block_label <- level_label
  } else if (length(group_label) > 0L &&
    length(level_label) > 0L) {
    block_label <- paste(rep(group_label, each = length(level_label)),
      rep(level_label, times = length(group_label)),
      sep = "."
    )
  }


  lav_data <- new("lavData",
    data.type = "full",
    ngroups = ngroups,
    group = group,
    nlevels = nlevels,
    cluster = cluster,
    group.label = group_label,
    level.label = level_label,
    block.label = block_label,
    std.ov = std_ov,
    nobs = nobs,
    norig = norig,
    ov.names = ov_names,
    ov.names.x = ov_names_x,
    ov.names.l = ov_names_l,
    ov.names.aux = if (length(ov_names_aux) > 0L) {
      rep(list(as.character(ov_names_aux)), ngroups)
    } else {
      vector("list", length = ngroups)
    },
    # ov.types        = ov.types,
    # ov.idx          = ov.idx,
    ordered = as.character(ordered),
    weights = weights,
    sampling.weights = sampling_weights,
    ov = ov,
    ov.aux = ov_aux,
    case.idx = case_idx,
    missing = missing,
    X = x,
    eXo = exo,
    aux = aux,
    Mp = mp,
    Rp = rp,
    Lp = lp
  )
  lav_data
}

# the multivariate normal distribution, unrestricted (h1)
# - everything is evaluated under the MLEs: Mu = ybar, Sigma = S

# 1) loglikelihood h1 (from raw data, or sample statistics)
# 4) hessian h1 around MLEs
# 5) information h1 (restricted Sigma/mu)
#    5a: (unit)    expected information h1 (A1 = gamma_nt^{-1})
#    5b: (unit)    observed information h1 (A1 = gamma_nt^{-1})
#    5c: (unit) first.order information h1 (B1 = A1 %*% Gamma %*% A1)


# YR 25 Mar 2016: first version
# YR 19 Jan 2017: added 6) + 7)
# YR 04 Jan 2020: adjust for sum(wt) != N
# YR 22 Jul 2022: adding correlation= argument for information_expected
#                 (only for catml; not used if correlation = TRUE!)
# YR 03 Jul 2026: use shared kernels from lav_mvnorm_kernels.R
# YR 08 Jul 2026: removed the unused 6) inverted-information and 7) ACOV
#                 families (see note at the end of this file)

# 1. log-likelihood h1

# 1a: input is raw data
lav_mvn_h1_loglik_data <- function(
    y = NULL,
    x_idx = integer(0L),
    casewise = FALSE,
    wt = NULL,
    sinv_method = "eigen") {

  n <- lav_mvn_nobs(y = y, wt = wt)
  p <- NCOL(y)

  # sample statistics
  samp <- lav_mvn_samp_stats(y = y, wt = wt)
  sample_mean <- samp$mean
  sample_cov <- samp$cov

  if (casewise) {
    log_2pi <- log(2 * pi)

    # invert sample_cov
    if (sinv_method == "chol") {
      c_s <- chol(sample_cov)
      ic_s <- backsolve(c_s, diag(p))
      yc <- t(t(y) - sample_mean)
      dist_1 <- rowSums((yc %*% ic_s)^2)
      logdet <- -2 * sum(log(diag(ic_s)))
    } else {
      sample_cov_inv <- lav_mat_sym_inverse(
        s = sample_cov,
        logdet = TRUE, sinv_method = sinv_method
      )
      logdet <- attr(sample_cov_inv, "logdet")
      # mahalanobis distance
      yc <- t(t(y) - sample_mean)
      dist_1 <- rowSums(yc %*% sample_cov_inv * yc)
    }

    loglik <- -(p * log_2pi + logdet + dist_1) / 2

    # weights
    if (!is.null(wt)) {
      loglik <- loglik * wt
    }
  } else {
    # invert sample_cov
    sample_cov_inv <- lav_mat_sym_inverse(
      s = sample_cov,
      logdet = TRUE, sinv_method = sinv_method
    )
    logdet <- attr(sample_cov_inv, "logdet")

    loglik <-
      lav_mvn_h1_loglik_samp(
        sample_cov_logdet = logdet,
        sample_nvar = p,
        sample_nobs = n
      )
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    loglik_x <- lav_mvn_h1_loglik_data(
      y = y[, x_idx, drop = FALSE],
      wt = wt, x_idx = integer(0L),
      casewise = casewise,
      sinv_method = sinv_method
    )
    # subtract logl.X
    loglik <- loglik - loglik_x
  }

  loglik
}



# 1b: input are sample statistics only (logdet, N and P)
lav_mvn_h1_loglik_samp <- function(
    sample_cov_logdet = NULL,
    sample_nvar = NULL,
    sample_nobs = NULL,
    # or
    sample_cov = NULL,
    x_idx = integer(0L),
    x_cov = NULL,
    sinv_method = "eigen") {

  if (is.null(sample_nvar)) {
    p <- NCOL(sample_cov)
  } else {
    p <- sample_nvar # number of variables
  }

  n <- sample_nobs
  stopifnot(!is.null(p), !is.null(n))

  log_2pi <- log(2 * pi)

  # all we need is the logdet
  if (is.null(sample_cov_logdet)) {
    sample_cov_inv <- lav_mat_sym_inverse(
      s = sample_cov,
      logdet = TRUE, sinv_method = sinv_method
    )
    logdet <- attr(sample_cov_inv, "logdet")
  } else {
    logdet <- sample_cov_logdet
  }

  loglik <- -n / 2 * (p * log_2pi + logdet + p)

  # fixed.x?
  if (length(x_idx) > 0L) {
    if (is.null(sample_cov)) {
      if (is.null(x_cov)) {
        lav_msg_stop(gettext(
          "when x_idx is not empty, we need sample_cov or x.cov"
        ))
      } else {
        sample_cov_x <- x_cov
      }
    } else {
      sample_cov_x <- sample_cov[x_idx, x_idx, drop = FALSE]
    }

    loglik_x <-
      lav_mvn_h1_loglik_samp(
        sample_cov = sample_cov_x,
        sample_nobs = sample_nobs,
        x_idx = integer(0L),
        sinv_method = sinv_method
      )
    # subtract logl.X
    loglik <- loglik - loglik_x
  }

  loglik
}


# 4. hessian of logl (around MLEs of Mu and Sigma)

# 4a: hessian logl Mu and vech(Sigma) from raw data
lav_mvn_h1_logl_hessian_data <- function(
    y = NULL,
    wt = NULL,
    x_idx = integer(0L),
    sinv_method = "eigen",
    sample_cov_inv = NULL,
    meanstructure = TRUE) {

  n <- lav_mvn_nobs(y = y, wt = wt)

  # observed information
  observed <- lav_mvn_h1_info_observed_data(
    y = y, wt = wt,
    x_idx = x_idx,
    sinv_method = sinv_method,
    sample_cov_inv = sample_cov_inv,
    meanstructure = meanstructure
  )

  -n * observed
}

# 4b: hessian Mu and vech(Sigma) from samplestats
lav_mvn_h1_logl_hessian_samp <- function(
    sample_mean = NULL, # unused!
    sample_cov = NULL,
    sample_nobs = NULL,
    x_idx = integer(0L),
    sinv_method = "eigen",
    sample_cov_inv = NULL,
    meanstructure = TRUE) {

  n <- sample_nobs

  # observed information
  observed <- lav_mvn_h1_info_observed_samp(
    sample_mean = sample_mean, sample_cov = sample_cov,
    x_idx = x_idx,
    sinv_method = sinv_method, sample_cov_inv = sample_cov_inv,
    meanstructure = meanstructure
  )

  -n * observed
}



# 5) Information h1 (note: expected == observed if data is complete!)

# 5a: unit expected information h1
lav_mvn_h1_info_expected <- function(
    y = NULL,
    wt = NULL,
    sample_cov = NULL,
    x_idx = integer(0L),
    sinv_method = "eigen",
    sample_cov_inv = NULL,
    meanstructure = TRUE,
  correlation = FALSE) {

  if (is.null(sample_cov_inv)) {
    if (is.null(sample_cov)) {
      sample_cov <- lav_mvn_samp_stats(y = y, wt = wt)$cov
    }

    # invert sample_cov
    sample_cov_inv <- lav_mat_sym_inverse(
      s = sample_cov,
      logdet = FALSE, sinv_method = sinv_method
    )
  }

  i22 <- lav_mvn_kron_dup_half(sample_cov_inv, correlation = correlation)

  if (meanstructure) {
    out <- lav_mat_bdiag(sample_cov_inv, i22)
  } else {
    out <- i22
  }

  # fixed.x?
  out <- lav_mvn_zero_x_idx(out,
    n = NCOL(sample_cov_inv), x_idx = x_idx,
    meanstructure = meanstructure
  )

  out
}

# 5b: unit observed information h1
#     (identical to the expected information, as the data is complete
#      and we evaluate at the MLEs Mu = ybar, Sigma = S;
#      note: the correlation= argument is NOT forwarded here)
lav_mvn_h1_info_observed_data <- function(
    y = NULL,
    wt = NULL,
    x_idx = integer(0L),
    sinv_method = "eigen",
    sample_cov_inv = NULL,
    meanstructure = TRUE) {

  lav_mvn_h1_info_expected(
    y = y, sinv_method = sinv_method,
    wt = wt, x_idx = x_idx,
    sample_cov_inv = sample_cov_inv,
    meanstructure = meanstructure
  )
}

# 5b-bis: observed information h1 from sample statistics
lav_mvn_h1_info_observed_samp <- function(
    sample_mean = NULL, # unused!
    sample_cov = NULL,
    x_idx = integer(0L),
    sinv_method = "eigen",
    sample_cov_inv = NULL,
    meanstructure = TRUE) {

  sample_cov_inv <- lav_mvn_sigma_inv(
    sigma_1 = sample_cov, sigma_inv = sample_cov_inv,
    sinv_method = sinv_method
  )

  i22 <- lav_mvn_kron_dup_half(sample_cov_inv)

  if (meanstructure) {
    out <- lav_mat_bdiag(sample_cov_inv, i22)
  } else {
    out <- i22
  }

  # fixed.x?
  out <- lav_mvn_zero_x_idx(out,
    n = NCOL(sample_cov_inv), x_idx = x_idx,
    meanstructure = meanstructure
  )

  out
}

# 5c: unit first-order information h1
#     note: first order information h1 == A1 %*% Gamma %*% A1
#           (where A1 = obs/exp information h1)
lav_mvn_h1_info_firstorder <- function(
    y = NULL,
    wt = NULL,
    sample_cov = NULL,
    x_idx = integer(0L),
    cluster_idx = NULL,
    sinv_method = "eigen",
    sample_cov_inv = NULL,
    gamma_1 = NULL,
    meanstructure = TRUE) {

  if (!is.null(wt)) {
    out <- stats::cov.wt(y, wt = wt, method = "ML")
    res <- lav_mvn_info_firstorder(
      y = y, wt = wt,
      cluster_idx = cluster_idx,
      mu = out$center, sigma_1 = out$cov, x_idx = x_idx,
      meanstructure = meanstructure
    )
    return(res)
  }

  # sample.cov.inv
  if (is.null(sample_cov_inv)) {
    # invert sample_cov
    if (is.null(sample_cov)) {
      sample_cov <- lav_mat_cov(y)
    }
    sample_cov_inv <- lav_mat_sym_inverse(
      s = sample_cov,
      logdet = FALSE, sinv_method = sinv_method
    )
  }

  # question: is there any benefit computing Gamma/A1 instead of just
  # calling lav_mvn_info_firstorder()?
  # answer (2014): probably not; it is just reassuring that the expression
  # J = A1 %*% gamma_1 %*% A1 seems to hold

  # Gamma
  # FIXME: what about the 'unbiased = TRUE' option?
  if (is.null(gamma_1)) {
    if (length(x_idx) > 0L) {
      gamma_1 <- lav_samp_gamma(y,
        x_idx = x_idx, fixed_x = TRUE,
        cluster_idx = cluster_idx,
        meanstructure = meanstructure
      )
    } else {
      gamma_1 <- lav_samp_gamma(y,
        meanstructure = meanstructure,
        cluster_idx = cluster_idx
      )
    }
  }

  # A1
  a1 <- lav_mvn_h1_info_expected(
    y = y, sinv_method = sinv_method,
    sample_cov_inv = sample_cov_inv,
    x_idx = x_idx,
    meanstructure = meanstructure
  )

  a1 %*% gamma_1 %*% a1
}

# NOTE: the former 6) inverted-information and 7) ACOV families (gamma_nt,
# B1.inv, 1/N * Gamma, ...) were unused reference implementations of the
# identity Gamma_ADF = A1^{-1} J1 A1^{-1}; they were removed in 0.7-1.
# The saturated-model Gamma itself lives in lav_samplestats_gamma.R
# (lav_samp_gamma / lav_samp_gamma_nt), and Gamma/N for a fitted model in
# lav_model_h1_acov() (lav_model_h1_information.R).

# utility functions for pairwise maximum likelihood

# stub for lav_pml_fml_dploglik_dimplied
lav_pml_fml_dploglik_dimplied <- function(
                       sigma_hat = NULL, # model-based var/cov/cor
                       th = NULL, # model-based thresholds + means
                       th_idx = NULL, # threshold idx per variable
                       num_idx = NULL, # which variables are numeric
                       x = NULL, # data
                       exo = NULL, # external covariates
                       lavcache = NULL, # housekeeping stuff
                       scores = FALSE, # return case-wise scores
                       negative = TRUE) {
  lav_msg_stop(gettext("not implemented"))
}

# the first derivative of the pairwise logLik function with respect to the
# thresholds/slopes/var/correlations; together with DELTA, we can use the
# chain rule to get the gradient
# this is adapted from code written by Myrsini Katsikatsou
# first attempt - YR 5 okt 2012
# HJ 18/10/23: Modification for complex design and completely observed data (no
# missing) with only ordinal indicators to get the right gradient for the
# optimisation and Hessian computation.
lav_pml_dploglik_dimplied <- function(
                       sigma_hat = NULL, # model-based var/cov/cor
                       mu_hat = NULL, # model-based means
                       th = NULL, # model-based thresholds + means
                       th_idx = NULL, # threshold idx per variable
                       num_idx = NULL, # which variables are numeric
                       x = NULL, # data
                       exo = NULL, # external covariates
                       wt = NULL, # case weights (not used yet)
                       lavcache = NULL, # housekeeping stuff
                       pi0 = NULL, # slopes
                       missing = "listwise", # how to deal with missings
                       scores = FALSE, # return case-wise scores
                       negative = TRUE) { # multiply by -1

  # diagonal of Sigma.hat is not necessarily 1, even for categorical vars
  sigma_hat2 <- sigma_hat
  if (length(num_idx) > 0L) {
    diag(sigma_hat2)[-num_idx] <- 1
  } else {
    diag(sigma_hat2) <- 1
  }
  cor_hat <- cov2cor(sigma_hat2) # to get correlations (rho!)
  cors <- lav_mat_vech(cor_hat, diagonal = FALSE)

  if (any(abs(cors) > 1)) {
    # what should we do now... force cov2cor?
    # cat("FFFFOOOORRRRRCEEE PD!\n")
    # Sigma.hat <- Matrix::nearPD(Sigma.hat)
    # Sigma.hat <- as.matrix(Sigma.hat$mat)
    # Sigma.hat <- cov2cor(Sigma.hat)
    # cors <- Sigma.hat[lower.tri(Sigma.hat)]
    idx <- which(abs(cors) > 0.99)
    cors[idx] <- 0.99 # clip
    # cat("CLIPPING!\n")
  }

  nvar <- nrow(sigma_hat)
  pstar <- nvar * (nvar - 1) / 2
  ov_types <- rep("ordered", nvar)
  if (length(num_idx) > 0L) ov_types[num_idx] <- "numeric"
  if (!is.null(exo)) {
    nexo <- ncol(exo)
  } else {
    nexo <- 0
  }


  if (all(ov_types == "numeric")) {
    n_th <- nvar
  } else {
    n_th <- length(th_idx)
  }
  n_sl <- nvar * nexo
  n_var <- length(num_idx)
  n_cor <- pstar

  # add num.idx to th.idx
  if (length(num_idx) > 0L) {
    th_idx[th_idx == 0] <- num_idx
  }

  # print(Sigma.hat); print(TH); print(th.idx); print(num.idx); print(str(X))

  # shortcut for ordinal-only/no-exo case
  if (!scores && all(ov_types == "ordered") && nexo == 0L) {
    # >>>>>>>> HJ/MK PML CODE >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    if (is.null(wt)) {
      n_xixj_vec <- lavcache$bifreq
    } else {
      n_xixj_vec <- lavcache$sum_obs_weights_xixj_ab_vec
    }
    gradient <- lav_pml_grad_tau_rho(
      no_x = nvar,
      all_thres = th,
      index_var_of_thres = th_idx,
      rho_xixj = cors,
      n_xixj_vec = n_xixj_vec,
      out_lav_pml_longvec_ind = lavcache$long
    )

    # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    if (missing == "available.cases") {
      uni_pi <- lav_pml_th_uni_prob(th = th, th_idx = th_idx)
      tmp <- lavcache$uniweights / uni_pi

      var_idx <- split(th_idx, th_idx)
      var_idx <- unlist(lapply(var_idx, function(x_1) {
        c(x_1, x_1[1])
      }))

      tmp_varwise <- split(tmp, var_idx)
      tmp1 <- unlist(lapply(
        tmp_varwise,
        function(x_1) {
          c(x_1[-length(x_1)])
        }
      ))
      tmp2 <- unlist(lapply(tmp_varwise, function(x_1) {
        c(x_1[-1])
      }))

      uni_der_tau <- dnorm(th) * (tmp1 - tmp2)
      n_th_1 <- length(th)
      gradient[1:n_th_1] <- gradient[1:n_th_1] + uni_der_tau
    }

    if (negative) {
      gradient <- -1 * gradient
    }
    return(gradient)
  }

  # in this order: TH/MEANS + SLOPES + VAR + COR
  grad_size <- n_th + n_sl + n_var + n_cor

  # scores or gradient?
  if (scores) {
    scores_1 <- matrix(0, nrow(x), grad_size) # we will sum up over all pairs
  } else {
    grad <- matrix(0, pstar, grad_size) # each pair is a row
  }
  pstar_1 <- matrix(0, nvar, nvar) # utility matrix, to get indices
  pstar_1[lav_mat_vech_idx(nvar, diagonal = FALSE)] <- 1:pstar
  # n <- length(x[, 1])

  for (j in seq_len(nvar - 1L)) {
    for (i in (j + 1L):nvar) {
      # cat(" i = ", i, " j = ", j, "\n") # debug only
      pstar_idx <- pstar_1[i, j]
      cor_idx <- n_th + n_sl + n_var + pstar_1[i, j]
      th_idx_i <- which(th_idx == i)
      th_idx_j <- which(th_idx == j)
      if (nexo > 0L) {
        sl_idx_i <- n_th + seq(i, by = nvar, length.out = nexo)
        sl_idx_j <- n_th + seq(j, by = nvar, length.out = nexo)

        if (length(num_idx) > 0L) {
          var_idx_i <- n_th + n_sl + match(i, num_idx)
          var_idx_j <- n_th + n_sl + match(j, num_idx)
        }
      } else {
        if (length(num_idx) > 0L) {
          var_idx_i <- n_th + match(i, num_idx)
          var_idx_j <- n_th + match(j, num_idx)
        }
      }
      if (ov_types[i] == "numeric" && ov_types[j] == "numeric") {
        if (nexo > 1L) {
          lav_msg_stop(gettext(
            "mixed + exo in PML not implemented;
            try optim.gradient = \"numerical\""))
        }

        sc <- lav_mvn_sc_mu_sigma(
          y = x[, c(i, j)],
          mu = mu_hat[c(i, j)], sigma_1 = sigma_hat[c(i, j), c(i, j)]
        )

        if (scores) {
          if (all(ov_types == "numeric") && nexo == 0L) {
            # MU1 + MU2
            scores_1[, c(i, j)] <- scores_1[, c(i, j)] + sc[, c(1, 2)]
            # VAR1 + COV_12 + VAR2
            var_idx <- (nvar +
              lav_mat_vech_match_idx(nvar, idx = c(i, j)))
            scores_1[, var_idx] <- scores_1[, var_idx] + sc[, c(3, 4, 5)]
          } else { # mixed ordered/continuous
            # MU
            mu_idx <- c(th_idx_i, th_idx_j)
            scores_1[, mu_idx] <- scores_1[, mu_idx] + (-1) * sc[, c(1, 2)]
            # VAR+COV
            var_idx <- c(var_idx_i, cor_idx, var_idx_j)
            scores_1[, var_idx] <- scores_1[, var_idx] + sc[, c(3, 4, 5)]
          }
        } else {
          if (all(ov_types == "numeric") && nexo == 0L) {
            mu_idx <- c(i, j)
            sigma_idx <- (nvar +
              lav_mat_vech_match_idx(nvar, idx = c(i, j)))
            # MU1 + MU2
            grad[pstar_idx, mu_idx] <-
              colSums(sc[, c(1, 2)], na.rm = TRUE)
          } else {
            mu_idx <- c(th_idx_i, th_idx_j)
            sigma_idx <- c(var_idx_i, cor_idx, var_idx_j)
            # MU (reverse sign!)
            grad[pstar_idx, mu_idx] <-
              -1 * colSums(sc[, c(1, 2)], na.rm = TRUE)
          }
          # SIGMA
          grad[pstar_idx, sigma_idx] <-
            colSums(sc[, c(3, 4, 5)], na.rm = TRUE)
        } # gradient only
      } else if (ov_types[i] == "numeric" && ov_types[j] == "ordered") {
        # polyserial correlation
        if (nexo > 1L) {
          lav_msg_stop(gettext(
            "mixed + exo in PML not implemented;
            try optim.gradient = \"numerical\""))
        }

        sc_cor_uni <- lav_bvmix_cor_sc(
          y1 = x[, i], y2 = x[, j],
          exo = NULL, wt = wt,
          evar_y1 = sigma_hat[i, i],
          beta_y1 = mu_hat[i],
          th_y2 = th[th_idx == j],
          sl_y2 = NULL,
          rho = cor_hat[i, j],
          sigma_correction = TRUE
        )

        if (scores) {
          # MU
          scores_1[, th_idx_i] <- (scores_1[, th_idx_i] +
            -1 * sc_cor_uni$dx_mu_y1)
          # TH
          scores_1[, th_idx_j] <- (scores_1[, th_idx_j] +
            sc_cor_uni$dx_th_y2)
          # VAR
          scores_1[, var_idx_i] <- (scores_1[, var_idx_i] +
            sc_cor_uni$dx_var_y1)
          # COR
          scores_1[, cor_idx] <- (scores_1[, cor_idx] +
            sc_cor_uni$dx_rho)
        } else {
          # MU
          grad[pstar_idx, th_idx_i] <-
            -1 * sum(sc_cor_uni$dx_mu_y1, na.rm = TRUE)
          # TH
          grad[pstar_idx, th_idx_j] <-
            colSums(sc_cor_uni$dx_th_y2, na.rm = TRUE)
          # VAR
          grad[pstar_idx, var_idx_i] <-
            sum(sc_cor_uni$dx_var_y1, na.rm = TRUE)
          # COR
          grad[pstar_idx, cor_idx] <-
            sum(sc_cor_uni$dx_rho, na.rm = TRUE)
        } # grad only
      } else if (ov_types[j] == "numeric" && ov_types[i] == "ordered") {
        # polyserial correlation
        if (nexo > 1L) {
          lav_msg_stop(gettext(
            "mixed + exo in PML not implemented;
            try optim.gradient = \"numerical\""))
        }

        sc_cor_uni <- lav_bvmix_cor_sc(
          y1 = x[, j], y2 = x[, i],
          exo = NULL, wt = wt,
          evar_y1 = sigma_hat[j, j],
          beta_y1 = mu_hat[j],
          th_y2 = th[th_idx == i],
          rho = cor_hat[i, j],
          sigma_correction = TRUE
        )

        if (scores) {
          # MU
          scores_1[, th_idx_j] <- (scores_1[, th_idx_j] +
            -1 * sc_cor_uni$dx_mu_y1)
          # TH
          scores_1[, th_idx_i] <- (scores_1[, th_idx_i] +
            sc_cor_uni$dx_th_y2)
          # VAR
          scores_1[, var_idx_j] <- (scores_1[, var_idx_j] +
            sc_cor_uni$dx_var_y1)
          # COR
          scores_1[, cor_idx] <- (scores_1[, cor_idx] +
            sc_cor_uni$dx_rho)
        } else {
          # MU
          grad[pstar_idx, th_idx_j] <-
            -1 * sum(sc_cor_uni$dx_mu_y1, na.rm = TRUE)
          # TH
          grad[pstar_idx, th_idx_i] <-
            colSums(sc_cor_uni$dx_th_y2, na.rm = TRUE)
          # VAR
          grad[pstar_idx, var_idx_j] <-
            sum(sc_cor_uni$dx_var_y1, na.rm = TRUE)
          # COR
          grad[pstar_idx, cor_idx] <-
            sum(sc_cor_uni$dx_rho, na.rm = TRUE)
        } # grad only
      } else if (ov_types[i] == "ordered" && ov_types[j] == "ordered") {
        # polychoric correlation
        if (nexo == 0L) {
          sc_cor_uni <-
            lav_bvord_cor_sc(
              y1 = x[, i], y2 = x[, j],
              exo = NULL, wt = wt,
              rho = sigma_hat[i, j],
              fit_y1 = NULL, # fixme
              fit_y2 = NULL, # fixme
              th_y1 = th[th_idx == i],
              th_y2 = th[th_idx == j],
              sl_y1 = NULL,
              sl_y2 = NULL,
              na_zero = TRUE
            )
        } else {
          sc_cor_uni <-
            lav_pml_dbilogl_dpar_x(
              y1 = x[, i],
              y2 = x[, j],
              exo = exo,
              rho = sigma_hat[i, j],
              th_y1 = th[th_idx == i],
              th_y2 = th[th_idx == j],
              sl_y1 = pi0[i, ],
              sl_y2 = pi0[j, ],
              missing_ind = missing
            )
        }

        if (scores) {
          # TH
          scores_1[, th_idx_i] <- scores_1[, th_idx_i] + sc_cor_uni$dx_th_y1
          scores_1[, th_idx_j] <- scores_1[, th_idx_j] + sc_cor_uni$dx_th_y2

          # SL
          if (nexo > 0L) {
            scores_1[, sl_idx_i] <- scores_1[, sl_idx_i] + sc_cor_uni$dx_sl_y1
            scores_1[, sl_idx_j] <- scores_1[, sl_idx_j] + sc_cor_uni$dx_sl_y2
          }
          # NO VAR
          # RHO
          scores_1[, cor_idx] <- scores_1[, cor_idx] + sc_cor_uni$dx_rho
        } else {
          # TH
          if (length(th_idx_i) > 1L) {
            grad[pstar_idx, th_idx_i] <-
              colSums(sc_cor_uni$dx_th_y1, na.rm = TRUE)
          } else {
            grad[pstar_idx, th_idx_i] <-
              sum(sc_cor_uni$dx_th_y1, na.rm = TRUE)
          }
          if (length(th_idx_j) > 1L) {
            grad[pstar_idx, th_idx_j] <-
              colSums(sc_cor_uni$dx_th_y2, na.rm = TRUE)
          } else {
            grad[pstar_idx, th_idx_j] <-
              sum(sc_cor_uni$dx_th_y2, na.rm = TRUE)
          }

          # SL
          if (nexo > 0L) {
            if (length(sl_idx_i) > 1L) {
              grad[pstar_idx, sl_idx_i] <-
                colSums(sc_cor_uni$dx_sl_y1, na.rm = TRUE)
            } else {
              grad[pstar_idx, sl_idx_i] <-
                sum(sc_cor_uni$dx_sl_y1, na.rm = TRUE)
            }
            if (length(sl_idx_j) > 1L) {
              grad[pstar_idx, sl_idx_j] <-
                colSums(sc_cor_uni$dx_sl_y2, na.rm = TRUE)
            } else {
              grad[pstar_idx, sl_idx_j] <-
                sum(sc_cor_uni$dx_sl_y2, na.rm = TRUE)
            }
          }
          # NO VAR

          # RHO
          grad[pstar_idx, cor_idx] <-
            sum(sc_cor_uni$dx_rho, na.rm = TRUE)
        }

        # GRAD2 <- numDeriv::grad(func = pc_logl_x,
        #                        x = c(Sigma.hat[i,j],
        #                              TH[ th.idx == i ],
        #                              TH[ th.idx == j]),
        #                        Y1  = X[,i],
        #                        Y2  = X[,j],
        #                        eXo = eXo,
        #                        nth.y1 = sum( th.idx == i ),
        #                        nth.y2 = sum( th.idx == j ))
      }
    }
  }

  if (missing == "available.cases" && all(ov_types == "ordered")) {
    # the available.cases objective adds a (weighted) UNIVARIATE likelihood
    # part to the pairwise likelihood; its scores/gradient must be added to
    # the threshold (and slope) entries. Note (fixed July 2026): the scores
    # branch used to (a) drop the univariate part entirely when nexo == 0,
    # and (b) overwrite scores_1 with only its first n_th + n_sl columns
    # when nexo > 0 (losing the correlation columns) -- which made the
    # casewise scores (and hence the first-order information) inconsistent
    # with the estimating function.
    if (nexo == 0L) {
      uni_scores <- matrix(0, nrow(x), n_th)
      for (i in seq_len(nvar)) {
        th_idx_i <- which(th_idx == i)
        der_y1 <- lav_pml_uni_sc(
          y1 = x[, i], th_y1 = th[th_idx == i],
          exo = NULL, sl_y1 = NULL,
          weights_casewise = lavcache$uniweights.casewise
        )
        uni_scores[, th_idx_i] <- der_y1$dx.th.y1
      }
      if (scores) {
        scores_1[, seq_len(n_th)] <-
          scores_1[, seq_len(n_th), drop = FALSE] + uni_scores
      } else {
        uni_gradient <- colSums(uni_scores)
      }
    } else {
      uni_scores <- matrix(0, nrow(x), ncol = (n_th + n_sl))
      for (i in seq_len(nvar)) {
        th_idx_i <- which(th_idx == i)
        sl_idx_i <- n_th + seq(i, by = nvar, length.out = nexo)
        der_y1 <- lav_pml_uni_sc(
          y1 = x[, i], th_y1 = th[th_idx == i],
          exo = exo, sl_y1 = pi0[i, ],
          weights_casewise = lavcache$uniweights.casewise
        )
        uni_scores[, th_idx_i] <- der_y1$dx.th.y1
        uni_scores[, sl_idx_i] <- der_y1$dx.sl.y1
      }
      if (scores) {
        scores_1[, seq_len(n_th + n_sl)] <-
          scores_1[, seq_len(n_th + n_sl), drop = FALSE] + uni_scores
      } else {
        uni_gradient <- colSums(uni_scores)
      }
    }
  }

  # do we need scores?
  if (scores) {
    return(scores_1)
  }

  # DEBUG
  # :print(GRAD)
  ###########



  # gradient is sum over all pairs
  gradient <- colSums(grad, na.rm = TRUE)

  if (missing == "available.cases" && all(ov_types == "ordered")) {
    # (fixed July 2026: the RHS used the FULL gradient vector instead of
    # its leading block, garbling the result through recycling; note that
    # the nexo == 0 case is normally unreachable here -- it is handled by
    # the shortcut at the top of this function)
    if (nexo == 0L) {
      gradient[seq_len(n_th)] <-
        gradient[seq_len(n_th)] + uni_gradient
    } else {
      gradient[seq_len(n_th + n_sl)] <-
        gradient[seq_len(n_th + n_sl)] + uni_gradient
    }
  }

  # we multiply by -1 because we minimize
  if (negative) {
    gradient <- -1 * gradient
  }

  gradient
}


### all code below written by Myrsini Katsikatsou


# The function lav_pml_grad_tau_rho
# input:
# no.x -  is scalar, the number of ordinal variables
# all.thres  - is vector containing the thresholds of all variables in the
#              following order: thres_var1, thres_var2,..., thres_var_p
#              within each variable the thresholds are in ascending order
#              Note that all.thres do NOT contain tau_0=-Inf and tau_last=Inf
#              for all variables.
# index.var.of.thres - a vector keeping track to which variable the thresholds
#                      in all.thres belongs to, it is of the form
#                      (1,1,1..., 2,2,2,...,  p,p,p,...)
# rho.xixj - is the vector of all correlations where j runs faster than i
#            i.e. the order is rho_12, rho_13, ..., rho_1p, rho_23, ..., rho_2p,
#            etc.
# n.xixj.vec -  a vector with the observed frequency for every combination
#               of categories and every pair. The frequencies are given in
#               the same order as the expected probabilities in the output of
#               lav_pml_expprob_vec output
# out.lav_pml_longvec_ind - it is the output of function lav_pml_longvec_ind
# the output: it gives the elements of der.L.to.tau and der.L.to.rho in this
#             order. The elements of der.L.to.tau where the elements are
#             ordered as follows: the thresholds of each variable with respect
#             to ascending order of the variable index (i.e. thres_var1,
#             thres_var2, etc.) and within each variable the thresholds in
#             ascending order.
#             The elements of vector der.L.to.rho are der.Lxixj.to.rho.xixj
#             where j runs faster than i.

# The function depends on four other functions: lav_pml_longvec_th_rho,
# lav_pml_expprob_vec, lav_pml_dl_drho, and lav_pml_dl_dtau, all given below.

# if n.xixj.ab is either an array or a list the following should be done
# n.xixj.vec <- if(is.array(n.xixj.ab)) {
#                 c(n.xixj.ab)
#              } else if(is.list(n.xixj.ab)){
#                 unlist(n.xixj.ab)
#              }


lav_pml_grad_tau_rho <- function(no_x, all_thres, index_var_of_thres, rho_xixj,
                         n_xixj_vec, out_lav_pml_longvec_ind) {
  out_lav_pml_longvec_th_rho <- lav_pml_longvec_th_rho(
    no_x = no_x, all_thres = all_thres,
    index_var_of_thres = index_var_of_thres,
    rho_xixj = rho_xixj
  )
  pi_xixj <- lav_pml_expprob_vec(
    ind_vec = out_lav_pml_longvec_ind,
    th_rho_vec = out_lav_pml_longvec_th_rho
  )

  out_lav_pml_dl_drho <- lav_pml_dl_drho(
    ind_vec = out_lav_pml_longvec_ind,
    th_rho_vec = out_lav_pml_longvec_th_rho,
    n_xixj = n_xixj_vec, pi_xixj = pi_xixj, no_x = no_x
  )

  out_lav_pml_dl_dtau <- lav_pml_dl_dtau(
    ind_vec = out_lav_pml_longvec_ind,
    th_rho_vec = out_lav_pml_longvec_th_rho,
    n_xixj = n_xixj_vec, pi_xixj = pi_xixj,
    no_x = no_x
  )

  grad <- c(out_lav_pml_dl_dtau, out_lav_pml_dl_drho)
  attr(grad, "pi.xixj") <- pi_xixj

  grad
}
################################################################################




# The input of the function  lav_pml_longvec_ind:

# no.x is scalar, the number of ordinal variables

# all.thres is vector containing the thresholds of all variables in the
# following order: thres_var1, thres_var2,..., thres_var_p
# within each variable the thresholds are in ascending order
# Note that all.thres does NOT contain the first and the last threshold of the
# variables, i.e. tau_0=-Inf and tau_last=Inf

# index.var.of.thres is a vector keeping track to which variable the thresholds
# in all.thres belongs to, it is of the form (1,1,1..., 2,2,2,...,  p,p,p,...)

# The output of the function:
# it is a list of vectors keeping track of the indices
# of thresholds, of variables, and of pairs, and two T/F vectors indicating
# if the threshold index corresponds to the last threshold of a variable; all
# these for all pairs of variables. All are needed for the
# computation of expected probabilities, der.L.to.rho, and der.L.to.tau

# all duplications of indices are done as follows: within each pair of
# variables,  xi-xj, if for example we want to duplicate the indices of
# the thresholds, tau^xi_a and tau^xj_b, then index a runs faster than b,
# i.e. for each b we take all different tau^xi's, and then we proceed to
# the next b and do the same. In other words if it was tabulated we fill
# the table columnwise.

# All pairs xi-xj are taken with index j running faster than i.

# Note that each variable may have a different number of categories, that's why
# for example we take lists below.

lav_pml_longvec_ind <- function(no_x, all_thres, index_var_of_thres) {
  no_thres_of_each_var <- tapply(all_thres, index_var_of_thres, length)
  index_pairs <- utils::combn(no_x, 2)
  no_pairs <- ncol(index_pairs)

  # index.thres.var1.of.pair and index.thres.var2.of.pair contain the indices of
  # of all thresholds (from tau_0 which is -Inf to tau_last which is Inf)
  # for any pair of variables appropriately duplicated so that the two vectors
  # together give all possible combinations of thresholds indices
  # Since here the threshold indices 0 and "last" are included, the vectors are
  # longer than the vectors thres.var1.of.pair and thres.var2.of.pair above.
  index_thres_var1_of_pair <- vector("list", no_pairs)
  index_thres_var2_of_pair <- vector("list", no_pairs)

  # index.var1.of.pair and index.var2.of.pair keep track the index of the
  # variable that the thresholds in index.thres.var1.of.pair and
  # index.thres.var2.of.pair belong to, respectively. So, these two variables
  # are of same length as that of index.thres.var1.of.pair and
  # index.thres.var2.of.pair
  index_var1_of_pair <- vector("list", no_pairs)
  index_var2_of_pair <- vector("list", no_pairs)

  # index.pairs.extended gives the index of the pair for each pair of variables
  # e.g. pair of variables 1-2 has index 1, variables 1-3 has index 2, etc.
  # The vector is of the same length as index.thres.var1.of.pair,
  # index.thres.var2.of.pair, index.var1.of.pair, and index.var2.of.pair
  index_pairs_extended <- vector("list", no_pairs)

  for (i in 1:no_pairs) {
    no_thres_var1_of_pair <- no_thres_of_each_var[index_pairs[1, i]]
    no_thres_var2_of_pair <- no_thres_of_each_var[index_pairs[2, i]]

    index_thres_var1_of_pair[[i]] <- rep(0:(no_thres_var1_of_pair + 1),
      times = (no_thres_var2_of_pair + 2)
    )
    index_thres_var2_of_pair[[i]] <- rep(0:(no_thres_var2_of_pair + 1),
      each = (no_thres_var1_of_pair + 2)
    )
    length_vec <- length(index_thres_var1_of_pair[[i]])
    index_var1_of_pair[[i]] <- rep(index_pairs[1, i], length_vec)
    index_var2_of_pair[[i]] <- rep(index_pairs[2, i], length_vec)
    index_pairs_extended[[i]] <- rep(i, length_vec)
  }

  index_thres_var1_of_pair <- unlist(index_thres_var1_of_pair)
  index_thres_var2_of_pair <- unlist(index_thres_var2_of_pair)
  index_var1_of_pair <- unlist(index_var1_of_pair)
  index_var2_of_pair <- unlist(index_var2_of_pair)
  index_pairs_extended <- unlist(index_pairs_extended)

  # indicator vector (T/F) showing which elements of index.thres.var1.of.pair
  # correspond to the last thresholds of variables. The length is the same as
  # that of index.thres.var1.of.pair.
  last_thres_var1_of_pair <- index_var1_of_pair == 1 &
    index_thres_var1_of_pair == (no_thres_of_each_var[1] + 1)
  # we consider up to variable (no.x-1) because in pairs xi-xj where j runs
  # faster than i, the last variable is not included in the column of xi's
  for (i in 2:(no_x - 1)) {
    new_condition <- index_var1_of_pair == i &
      index_thres_var1_of_pair == (no_thres_of_each_var[i] + 1)
    last_thres_var1_of_pair <- last_thres_var1_of_pair | new_condition
  }

  # indicator vector (T/F) showing which elements of index.thres.var2.of.pair
  # correspond to the last thresholds of variables. Notet that in pairs xi-xj
  # where j runs faster than i, the first variable is not included in the column
  # of xj's. That's why we start with variable 2. The length is the same as
  # that of index.thres.var1.of.pair.
  last_thres_var2_of_pair <- index_var2_of_pair == 2 &
    index_thres_var2_of_pair == (no_thres_of_each_var[2] + 1)
  for (i in 3:no_x) {
    new_condition <- index_var2_of_pair == i &
      index_thres_var2_of_pair == (no_thres_of_each_var[i] + 1)
    last_thres_var2_of_pair <- last_thres_var2_of_pair | new_condition
  }

  list(
    index.thres.var1.of.pair = index_thres_var1_of_pair,
    index.thres.var2.of.pair = index_thres_var2_of_pair,
    index.var1.of.pair = index_var1_of_pair,
    index.var2.of.pair = index_var2_of_pair,
    index.pairs.extended = index_pairs_extended,
    last.thres.var1.of.pair = last_thres_var1_of_pair,
    last.thres.var2.of.pair = last_thres_var2_of_pair
  )
}
################################################################################


# The input of the function  lav_pml_longvec_th_rho:

# no.x is scalar, the number of ordinal variables

# all.thres is vector containing the thresholds of all variables in the
# following order: thres_var1, thres_var2,..., thres_var_p
# within each variable the thresholds are in ascending order
# Note that all.thres does NOT contain the first and the last threshold of the
# variables, i.e. tau_0=-Inf and tau_last=Inf

# index.var.of.thres is a vector keeping track to which variable the thresholds
# in all.thres belongs to, it is of the form (1,1,1..., 2,2,2,...,  p,p,p,...)

# rho.xixj is the vector of all corrlations where j runs faster than i
# i.e. the order is rho_12, rho_13, ..., rho_1p, rho_23, ..., rho_2p, etc.

# The output of the function:
# it is a list of vectors with thresholds and rho's duplicated appropriately,
# all needed for the computation of expected probabilities,
# der.L.to.rho, and der.L.to.tau

# all duplications below are done as follows: within each pair of variables,
# xi-xj, if for example we want to duplicate their thresholds, tau^xi_a and
# tau^xj_b, then index a runs faster than b, i.e. for each b we take all
# different tau^xi's, and then we proceed to the next b and do the same.
# In other words if it was tabulated we fill the table columnwise.

# All pairs xi-xj are taken with index j running faster than i.

# Note that each variable may have a different number of categories, that's why
# for example we take lists below.

lav_pml_longvec_th_rho <- function(no_x, all_thres,
                                   index_var_of_thres, rho_xixj) {
  no_thres_of_each_var <- tapply(all_thres, index_var_of_thres, length)
  index_pairs <- utils::combn(no_x, 2)
  no_pairs <- ncol(index_pairs)

  # create the long vectors needed for the computation of expected probabilities
  # for each cell and each pair of variables. The vectors thres.var1.of.pair and
  # thres.var2.of.pair together give all the possible combinations of the
  # thresholds of any two variables. Note the combinations (-Inf, -Inf),
  # (-Inf, Inf), (Inf, -Inf), (Inf, Inf) are NOT included. Only the combinations
  # of the middle thresholds (tau_1 to tau_(last-1)).
  # thres.var1.of.pair and thres.var2.of.pair give the first and the second
  # argument, respectively, in functions pbivnorm and lav_dbinorm
  thres_var1_of_pair <- vector("list", no_pairs)
  thres_var2_of_pair <- vector("list", no_pairs)

  # Extending the rho.vector accordingly so that it will be the the third
  # argument in pbivnorm and lav_dbinorm functions. It is of same length as
  # thres.var1.of.pair and thres.var2.of.pair.
  rho_vector <- vector("list", no_pairs)

  # thres.var1.for.dnorm.in.der.pi.to.tau.xi and
  # thres.var2.for.dnorm.in.der.pi.to.tau.xj give the thresholds of almost
  # all variables appropriately duplicated so that the vectors can be used
  # as input in dnorm() to compute der.pi.xixj.to.tau.xi and
  # der.pi.xixj.to.tau.xj.
  # thres.var1.for.dnorm.in.der.pi.to.tau.xi does not contain the thresholds of
  # the last variable and thres.var2.for.dnorm.in.der.pi.to.tau.xj those of
  # the first variable
  thres_var1_for_dnorm_in_der_pi_to_tau_xi <- vector("list", no_pairs) # nolint
  thres_var2_for_dnorm_in_der_pi_to_tau_xj <- vector("list", no_pairs) # nolint

  for (i in 1:no_pairs) {
    single_thres_var1_of_pair <-
       all_thres[index_var_of_thres == index_pairs[1, i]]
    single_thres_var2_of_pair <-
       all_thres[index_var_of_thres == index_pairs[2, i]]
    # remember that the first (-Inf) and last (Inf) thresholds are not included
    # so no.thres.var1.of.pair is equal to number of categories of var1 minus 1
    # similarly for no.thres.var2.of.pair
    no_thres_var1_of_pair <- no_thres_of_each_var[index_pairs[1, i]]
    no_thres_var2_of_pair <- no_thres_of_each_var[index_pairs[2, i]]

    thres_var1_of_pair[[i]] <- rep(single_thres_var1_of_pair,
      times = no_thres_var2_of_pair
    )
    thres_var2_of_pair[[i]] <- rep(single_thres_var2_of_pair,
      each = no_thres_var1_of_pair
    )
    rho_vector[[i]] <- rep(rho_xixj[i], length(thres_var1_of_pair[[i]]))

    thres_var1_for_dnorm_in_der_pi_to_tau_xi[[i]] <-   # nolint
      rep(single_thres_var1_of_pair, times = (no_thres_var2_of_pair + 1))
    thres_var2_for_dnorm_in_der_pi_to_tau_xj[[i]] <-   # nolint
      rep(single_thres_var2_of_pair, each = (no_thres_var1_of_pair + 1))
  }

  thres_var1_of_pair <- unlist(thres_var1_of_pair)
  thres_var2_of_pair <- unlist(thres_var2_of_pair)
  rho_vector <- unlist(rho_vector)
  thres_var1_for_dnorm_in_der_pi_to_tau_xi <-          # nolint
    unlist(thres_var1_for_dnorm_in_der_pi_to_tau_xi)
  thres_var2_for_dnorm_in_der_pi_to_tau_xj <-          # nolint
    unlist(thres_var2_for_dnorm_in_der_pi_to_tau_xj)

  # thres.var2.for.last.cat.var1 and thres.var1.for.last.cat.var2 are needed
  # for the computation of expected probabilities. In the computation of
  # \Phi_2(tau1, tau2; rho) when either tau1 or tau2 are Inf then it is enought
  # to compute pnorm() with the non-infinite tau as an argument
  # In particular when the first variable of the pair has tau_last= Inf
  # and the second a non-infite threshold we compute
  # pnorm(thres.var2.for.last.cat.var1). Similarly, when the second variable of
  # the pair has tau_last=Inf and the first a non-infite threshold we compute
  # pnorm(thres.var1.for.last.cat.var2).
  thres_var2_for_last_cat_var1 <- vector("list", (no_x - 1))
  thres_var1_for_last_cat_var2 <- vector("list", (no_x - 1))
  for (i in 1:(no_x - 1)) {
    thres_var2_for_last_cat_var1[[i]] <-
      c(all_thres[index_var_of_thres %in% (i + 1):no_x])
    thres_var1_for_last_cat_var2[[i]] <- rep(all_thres[index_var_of_thres == i],
      times = (no_x - i)
    )
  }
  thres_var2_for_last_cat_var1 <- unlist(thres_var2_for_last_cat_var1)
  thres_var1_for_last_cat_var2 <- unlist(thres_var1_for_last_cat_var2)


  list(
    thres.var1.of.pair = thres_var1_of_pair, # these 3 of same length
    thres.var2.of.pair = thres_var2_of_pair,
    rho.vector = rho_vector,

    # the following of length dependning on the number of categories
    thres.var1.for.dnorm.in.der.pi.to.tau.xi =
      thres_var1_for_dnorm_in_der_pi_to_tau_xi,
    thres.var2.for.dnorm.in.der.pi.to.tau.xj =
      thres_var2_for_dnorm_in_der_pi_to_tau_xj,
    thres.var2.for.last.cat.var1 = thres_var2_for_last_cat_var1,
    thres.var1.for.last.cat.var2 = thres_var1_for_last_cat_var2
  )
}
#########################################################



#########################################################

# The function  lav_pml_expprob_vec
# input: ind.vec - the output of function lav_pml_longvec_ind
#        th.rho.vec - the output of function lav_pml_longvec_th_rho
# output: it gives the elements of pairwiseTablesExpected()$pi.tables
# table-wise and column-wise within each table. In other words if
# pi^xixj_ab is the expected probability for the pair of variables xi-xj
# and categories a and b, then index a runs the fastest of all, followed by b,
# then by j, and lastly by i.

lav_pml_expprob_vec <- function(ind_vec, th_rho_vec) {
  prob_vec <- rep(NA, length(ind_vec$index.thres.var1.of.pair))

  prob_vec[ind_vec$index.thres.var1.of.pair == 0 |
    ind_vec$index.thres.var2.of.pair == 0] <- 0

  prob_vec[ind_vec$last.thres.var1.of.pair &
    ind_vec$last.thres.var2.of.pair] <- 1

  prob_vec[ind_vec$last.thres.var1.of.pair &
    ind_vec$index.thres.var2.of.pair != 0 &
    !ind_vec$last.thres.var2.of.pair] <-
    pnorm(th_rho_vec$thres.var2.for.last.cat.var1)

  prob_vec[ind_vec$last.thres.var2.of.pair &
    ind_vec$index.thres.var1.of.pair != 0 &
    !ind_vec$last.thres.var1.of.pair] <-
    pnorm(th_rho_vec$thres.var1.for.last.cat.var2)

  prob_vec[is.na(prob_vec)] <- pbivnorm(
    th_rho_vec$thres.var1.of.pair,
    th_rho_vec$thres.var2.of.pair,
    th_rho_vec$rho.vector
  )

  cum_term1 <- prob_vec[ind_vec$index.thres.var1.of.pair != 0 &
    ind_vec$index.thres.var2.of.pair != 0]

  cum_term2 <- prob_vec[ind_vec$index.thres.var1.of.pair != 0 &
    !ind_vec$last.thres.var2.of.pair]

  cum_term3 <- prob_vec[ind_vec$index.thres.var2.of.pair != 0 &
    !ind_vec$last.thres.var1.of.pair]

  cum_term4 <- prob_vec[!ind_vec$last.thres.var1.of.pair &
    !ind_vec$last.thres.var2.of.pair]

  pi0 <- cum_term1 - cum_term2 - cum_term3 + cum_term4

  # added by YR 11 nov 2012 to avoid Nan/-Inf
  # log(.Machine$double.eps) = -36.04365
  # all elements should be strictly positive
  pi0[pi0 < .Machine$double.eps] <- .Machine$double.eps

  pi0
}


# lav_pml_dl_drho
# input: ind.vec - the output of function lav_pml_longvec_ind
#        th.rho.vec - the output of function lav_pml_longvec_th_rho
#        n.xixj - a vector with the observed frequency for every combination
#                 of categories and every pair. The frequencies are given in
#                 the same order as the expected probabilities in the output of
#                 lav_pml_expprob_vec output
#        pi.xixj - the output of lav_pml_expprob_vec function
#        no.x    - the number of ordinal variables
# output: the vector of der.L.to.rho, each element corresponds to
#         der.Lxixj.to.rho.xixj where j runs faster than i

lav_pml_dl_drho <- function(ind_vec, th_rho_vec, n_xixj, pi_xixj, no_x) {
  prob_vec <- rep(NA, length(ind_vec$index.thres.var1.of.pair))

  prob_vec[ind_vec$index.thres.var1.of.pair == 0 |
    ind_vec$index.thres.var2.of.pair == 0 |
    ind_vec$last.thres.var1.of.pair |
    ind_vec$last.thres.var2.of.pair] <- 0

  prob_vec[is.na(prob_vec)] <- lav_dbinorm(th_rho_vec$thres.var1.of.pair,
    th_rho_vec$thres.var2.of.pair,
    rho = th_rho_vec$rho.vector
  )

  den_term1 <- prob_vec[ind_vec$index.thres.var1.of.pair != 0 &
    ind_vec$index.thres.var2.of.pair != 0]

  den_term2 <- prob_vec[ind_vec$index.thres.var1.of.pair != 0 &
    !ind_vec$last.thres.var2.of.pair]

  den_term3 <- prob_vec[ind_vec$index.thres.var2.of.pair != 0 &
    !ind_vec$last.thres.var1.of.pair]

  den_term4 <- prob_vec[!ind_vec$last.thres.var1.of.pair &
    !ind_vec$last.thres.var2.of.pair]

  der_pi_xixj_to_rho_xixj <- den_term1 - den_term2 - den_term3 + den_term4
  prod_terms <- (n_xixj / pi_xixj) * der_pi_xixj_to_rho_xixj

  # to get der.Lxixj.to.rho.xixj we should all the elements of
  # der.pi.xixj.to.rho.xixj which correspond to the pair xi-xj, to do so:
  xnew <- lapply(
    ind_vec[c("index.pairs.extended")],
    function(y) {
      y[ind_vec$index.thres.var1.of.pair != 0 &
        ind_vec$index.thres.var2.of.pair != 0]
    }
  )
  # der.L.to.rho is:
  tapply(prod_terms, xnew$index.pairs.extended, sum)
}
###########################################################################


# lav_pml_dl_dtau
# input: ind.vec - the output of function lav_pml_longvec_ind
#        th.rho.vec - the output of function lav_pml_longvec_th_rho
#        n.xixj - a vector with the observed frequency for every combination
#                 of categories and every pair. The frequencies are given in
#                 the same order as the expected probabilities in the output of
#                 lav_pml_expprob_vec output
#        pi.xixj - the output of lav_pml_expprob_vec function
# output: the vector of der.L.to.tau where the elements are ordered as follows:
#         the thresholds of each variable with respect to ascending order of
#         the variable index (i.e. thres_var1, thres_var2, etc.) and within
#         each variable the thresholds in ascending order.


lav_pml_dl_dtau <- function(ind_vec, th_rho_vec, n_xixj, pi_xixj, no_x = 0L) {
  # to compute der.pi.xixj.to.tau.xi
  xi <- lapply(
    ind_vec[c(
      "index.thres.var2.of.pair",
      "last.thres.var2.of.pair"
    )],
    function(y) {
      y[!(ind_vec$index.thres.var1.of.pair == 0 |
        ind_vec$last.thres.var1.of.pair)]
    }
  )

  cum_prob_vec <- rep(NA, length(xi$index.thres.var2.of.pair))
  cum_prob_vec[xi$index.thres.var2.of.pair == 0] <- 0
  cum_prob_vec[xi$last.thres.var2.of.pair] <- 1
  denom <- sqrt(1 - (th_rho_vec$rho.vector * th_rho_vec$rho.vector))
  cum_prob_vec[is.na(cum_prob_vec)] <-
    pnorm((th_rho_vec$thres.var2.of.pair -
      th_rho_vec$rho.vector * th_rho_vec$thres.var1.of.pair) /
      denom)
  den_prob_vec <- dnorm(th_rho_vec$thres.var1.for.dnorm.in.der.pi.to.tau.xi)
  der_pi_xixj_to_tau_xi <- den_prob_vec *
    (cum_prob_vec[xi$index.thres.var2.of.pair != 0] -
      cum_prob_vec[!xi$last.thres.var2.of.pair])

  # to compute der.pi.xixj.to.tau.xj
  xj <- lapply(
    ind_vec[c(
      "index.thres.var1.of.pair",
      "last.thres.var1.of.pair"
    )],
    function(y) {
      y[!(ind_vec$index.thres.var2.of.pair == 0 |
        ind_vec$last.thres.var2.of.pair)]
    }
  )

  cum_prob_vec <- rep(NA, length(xj$index.thres.var1.of.pair))
  cum_prob_vec[xj$index.thres.var1.of.pair == 0] <- 0
  cum_prob_vec[xj$last.thres.var1.of.pair] <- 1
  denom <- sqrt(1 - (th_rho_vec$rho.vector * th_rho_vec$rho.vector))
  cum_prob_vec[is.na(cum_prob_vec)] <-
    pnorm((th_rho_vec$thres.var1.of.pair -
      th_rho_vec$rho.vector * th_rho_vec$thres.var2.of.pair) /
      denom)
  den_prob_vec <- dnorm(th_rho_vec$thres.var2.for.dnorm.in.der.pi.to.tau.xj)
  der_pi_xixj_to_tau_xj <- den_prob_vec *
    (cum_prob_vec[xj$index.thres.var1.of.pair != 0] -
      cum_prob_vec[!xj$last.thres.var1.of.pair])

  # to compute der.Lxixj.tau.xi and der.Lxixj.tau.xi
  n_over_pi <- n_xixj / pi_xixj
  # get the appropriate differences of n.over.pi for der.Lxixj.to.tau.xi  and
  # der.Lxixj.to.tau.xj
  x3a <- lapply(ind_vec, function(y) {
    y[!(ind_vec$index.thres.var1.of.pair == 0 |
      ind_vec$index.thres.var2.of.pair == 0)]
  })
  diff_n_over_pi_to_xi <- n_over_pi[!x3a$last.thres.var1.of.pair] -
    n_over_pi[x3a$index.thres.var1.of.pair != 1]
  diff_n_over_pi_to_xj <- n_over_pi[!x3a$last.thres.var2.of.pair] -
    n_over_pi[x3a$index.thres.var2.of.pair != 1]
  # terms.der.Lxixj.to.tau.xi  and terms.der.Lxixj.to.tau.xj
  terms_der_lxixj_to_tau_xi <- diff_n_over_pi_to_xi * der_pi_xixj_to_tau_xi
  terms_der_lxixj_to_tau_xj <- diff_n_over_pi_to_xj * der_pi_xixj_to_tau_xj
  # to add appropriately elements of terms.der.Lxixj.to.tau.xi
  x3b <- lapply(
    ind_vec[c(
      "index.pairs.extended",
      "index.thres.var1.of.pair"
    )],
    function(y) {
      y[!(ind_vec$index.thres.var1.of.pair == 0 |
        ind_vec$last.thres.var1.of.pair |
        ind_vec$index.thres.var2.of.pair == 0)]
    }
  )
  # to add appropriately elements of terms.der.Lxixj.to.tau.xj
  x4b <- lapply(
    ind_vec[c(
      "index.pairs.extended",
      "index.thres.var2.of.pair"
    )],
    function(y) {
      y[!(ind_vec$index.thres.var2.of.pair == 0 |
        ind_vec$last.thres.var2.of.pair |
        ind_vec$index.thres.var1.of.pair == 0)]
    }
  )
  ind_pairs <- utils::combn(no_x, 2)
  # der.Lxixj.to.tau.xi is a matrix, nrow=no.pairs, ncol=max(no.of.free.thres)
  # thus, there are NA's, similarly for der.Lxixj.to.tau.xj
  der_lxixj_to_tau_xi <- tapply(
    terms_der_lxixj_to_tau_xi,
    list(
      x3b$index.pairs.extended,
      x3b$index.thres.var1.of.pair
    ),
    sum
  )
  der_lxixj_to_tau_xj <- tapply(
    terms_der_lxixj_to_tau_xj,
    list(
      x4b$index.pairs.extended,
      x4b$index.thres.var2.of.pair
    ),
    sum
  )

  # to add appropriately the terms of der.Lxixj.to.tau.xi and
  # der.Lxixj.to.tau.xj
  split_der_lxixj_to_tau_xi <- split(
    as.data.frame(der_lxixj_to_tau_xi),
    ind_pairs[1, ]
  )
  sums_der_lxixj_to_tau_xi <- lapply(
    split_der_lxixj_to_tau_xi,
    function(x) {
      y <- apply(x, 2, sum)
      y[!is.na(y)]
    }
  )
  # Note: NA exist in the case where the ordinal variables have different
  # number of response categories
  split_der_lxixj_to_tau_xj <- split(
    as.data.frame(der_lxixj_to_tau_xj),
    ind_pairs[2, ]
  )
  sums_der_lxixj_to_tau_xj <- lapply(
    split_der_lxixj_to_tau_xj,
    function(x) {
      y <- apply(x, 2, sum)
      y[!is.na(y)]
    }
  )
  # to get der.L.to.tau
  c(
    sums_der_lxixj_to_tau_xi[[1]],
    c(unlist(sums_der_lxixj_to_tau_xi[2:(no_x - 1)]) +
      unlist(sums_der_lxixj_to_tau_xj[1:(no_x - 2)])),
    sums_der_lxixj_to_tau_xj[[no_x - 1]]
  )
}

# SAM step 2: estimate structural part

lav_sam_step2 <- function(step1 = NULL, fit = NULL,
                          sam_method = "local", struc_args = list()) {
  lavoptions <- fit@Options
  lavpta <- fit@pta
  nlevels <- lavpta$nlevels
  pt_1 <- step1$PT

  # Gamma available?
  gamma_flag <- (sam_method %in% c("local", "fsr", "cfsr") &&
                 !is.null(step1$Gamma.eta[[1]]))

  lv_names <- unique(unlist(fit@pta$vnames$lv.regular))

  # adjust options
  lavoptions_pa <- lavoptions
  # "yuan.chan" is a SAM-global test for the JOINT model, computed afterwards in
  # lav_sam_global_test(); the structural fit itself uses the ordinary test
  if (any(lavoptions_pa$test == "yuan.chan")) {
    lavoptions_pa$test <- "standard"
  }
  # the corrected two-step STRUCTURAL test (Satorra-Bentler, using Gamma.eta as
  # the NACOV of vech(VETA)) is the default test for sam.method = local/fsr/cfsr
  # whenever Gamma.eta is available -- INDEPENDENT of the requested SE. (For
  # se = "twostep"/"naive" the FIT.PA SEs below are not the final ones: twostep
  # SEs are recomputed in step 4, naive SEs are FIT.PA's plain vcov.)
  if (gamma_flag) {
    lavoptions_pa$test <- "satorra.bentler"
  }
  if (lavoptions_pa$se == "naive") {
    # naive SEs = FIT.PA's plain (standard) vcov
    lavoptions_pa$se <- "standard"
  } else if (lavoptions_pa$se %in% c("local", "local.nt")) {
    # local SEs ARE FIT.PA's robust.sem vcov (read back in lav_sam_step2_se())
    lavoptions_pa$se <- "robust.sem"
  } else if (gamma_flag) {
    # twostep / twostep.robust: the final SEs are recomputed in step 4
    # (lav_sam_step2_se). Use se = "standard" -- NOT "robust.sem" -- so that
    # FIT.PA's vcov stays the NAIVE (standard) one: the alpha.correction blend
    # in lav_sam_step2_se() reads it as 'vcov_naive'. The corrected structural
    # test is still computed (it is independent of the se).
    lavoptions_pa$se <- "standard"
  } else {
    # twostep or none, without Gamma.eta -> none
    lavoptions_pa$se <- "none"
  }
  if (!lavoptions_pa$conditional.x) {
    lavoptions_pa$fixed.x <- FALSE # until we fix this...
  }
  lavoptions_pa$categorical <- FALSE
  lavoptions_pa$.categorical <- FALSE
  lavoptions_pa$rotation <- "none"
  lavoptions_pa <- modifyList(lavoptions_pa, struc_args)

  if (gamma_flag) {
    lavoptions_pa$check.vcov <- FALSE # always non-pd
                                      # if interactions + fixed.x = FALSE
  }

  # override, no matter what
  lavoptions_pa$do.fit <- TRUE

  if (sam_method %in% c("local", "fsr", "cfsr")) {
    lavoptions_pa$missing <- "listwise"
    lavoptions_pa$sample.cov.rescale <- FALSE
    lavoptions_pa$loglik <- FALSE
    # first.order information (eg estimator = "MLF") needs raw data, which
    # the structural fit does not have (it is fitted from the estimated
    # latent moments VETA/EETA only); use the expected information instead
    if (any(lavoptions_pa$information == "first.order")) {
      lavoptions_pa$information <- rep.int("expected", 2L)
    }
  } else {
    lavoptions_pa$h1 <- FALSE
    lavoptions_pa$loglik <- FALSE
  }

  # construct PTS
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    # extract structural part
    pts <- lav_pt_subset_sm(pt_1,
      add_idx = TRUE,
      add_exo_cov = TRUE,
      fixed_x = lavoptions_pa$fixed.x,
      conditional_x = lavoptions_pa$conditional.x,
      free_fixed_var = TRUE,
      meanstructure = lavoptions_pa$meanstructure
    )

    # any 'extra' parameters: not (free) in PT, but free in PTS (user == 3)
    #  - fixed.x in PT, but fixed.x = FALSE is PTS
    #  - fixed-to-zero intercepts in PT, but free in PTS
    #  - add.exo.cov: absent/fixed-to-zero in PT, but add/free in PTS
    extra_id <- which(pts$user == 3L)

    # remove est/se/start columns
    pts$est <- NULL
    pts$se <- NULL
    pts$start <- NULL

    if (nlevels > 1L) {
      pts$level <- NULL
      pts$group <- NULL
      pts$group <- pts$block
      nobs_1 <- fit@Data@Lp[[1]]$nclusters
    } else {
      nobs_1 <- fit@Data@nobs
    }

    reg_idx <- attr(pts, "idx")
    attr(pts, "idx") <- NULL

    # edge case: conditional.x = TRUE, but the structural part contains no
    # exogenous covariates (eg they only affect indicators directly); the
    # conditional attributes of VETA (res.slopes/cov.x/mean.x) then have no
    # counterpart in the structural model -> drop them and fit the
    # structural part unconditionally
    if (lavoptions_pa$conditional.x &&
        length(unlist(lav_pt_vnames(pts, type = "ov.x"))) == 0L) {
      attr(step1$VETA, "res.slopes") <- NULL
      attr(step1$VETA, "cov.x") <- NULL
      attr(step1$VETA, "mean.x") <- NULL
      lavoptions_pa$conditional.x <- FALSE
      lavoptions_pa$fixed.x <- FALSE
    }
  } else {
    # global SAM

    # the measurement model parameters now become fixed ustart values
    pt_1$ustart[pt_1$free > 0] <- pt_1$est[pt_1$free > 0]

    reg_idx <- lav_pt_subset_sm(
      pt_1 = pt_1,
      idx_only = TRUE
    )

    # remove 'exogenous' factor variances (if any) from reg.idx
    lv_names_x <- lv_names[lv_names %in% unlist(lavpta$vnames$eqs.x) &
      !lv_names %in% unlist(lavpta$vnames$eqs.y)]
    if ((lavoptions_pa$fixed.x || lavoptions_pa$std.lv) &&
        length(lv_names_x) > 0L) {
      var_idx <- which(pt_1$lhs %in% lv_names_x &
        pt_1$op == "~~" &
        pt_1$lhs == pt_1$rhs)
      rm_idx <- which(reg_idx %in% var_idx)
      if (length(rm_idx) > 0L) {
        reg_idx <- reg_idx[-rm_idx]
      }
    }

    # adapt parameter table for structural part
    pts <- pt_1

    # remove constraints we don't need
    con_idx <- which(pts$op %in% c("==", "<", ">", ":="))
    if (length(con_idx) > 0L) {
      needed_idx <- which(con_idx %in% reg_idx)
      if (length(needed_idx) > 0L) {
        con_idx <- con_idx[-needed_idx]
      }
      if (length(con_idx) > 0L) {
        pts <- as.data.frame(pts, stringsAsFactors = FALSE)
        pts <- pts[-con_idx, ]
      }
    }
    pts$est <- NULL
    pts$se <- NULL

    # 'fix' step 1 parameters
    pts$free[!pts$id %in% reg_idx & pts$free > 0L] <- 0L

    # but free up residual variances if fixed (eg std.lv = TRUE) (new in 0.6-20)
    var_idx <- reg_idx[which(pt_1$free[reg_idx] == 0L &
                             pt_1$user[reg_idx] != 1L &
                             pt_1$op[reg_idx] == "~~")] # FIXME: more?
    pts$free[var_idx] <- max(pts$free) + seq_along(var_idx)
    # reset any stale bounds (a fixed parameter has lower == its fixed
    # value whenever the partable carries bounds columns; see the same
    # fix in lav_pt_subset_sm())
    if (!is.null(pts$lower)) {
      pts$lower[var_idx] <- -Inf
    }
    if (!is.null(pts$upper)) {
      pts$upper[var_idx] <- +Inf
    }

    # set 'ustart' values for free FIT.PA parameter to NA
    pts$ustart[pts$free > 0L] <- as.numeric(NA)

    pts <- lav_pt_complete(pts)

    extra_id <- integer(0L)
  } # global

  # fit structural model
  if (lav_verbose()) {
    cat("Fitting the structural part ... \n")
  }
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    if (gamma_flag) {
      nacov <- step1$Gamma.eta
    } else {
      nacov <- NULL
    }
    fit_pa <- tryCatch(
      lavaan::lavaan(pts,
        sample_cov  = step1$VETA,
        sample_mean = step1$EETA,
        sample_nobs = nobs_1,
        nacov       = nacov,
        slot_options = lavoptions_pa,
        verbose     = FALSE
      ),
      error = function(e) e
    )
    if (inherits(fit_pa, "error")) {
      # the corrected two-step STRUCTURAL test (Satorra-Bentler via Gamma.eta)
      # could not be computed for this structural model (eg a bi-factor
      # measurement model, whose VETA jacobian is not conformable here). For
      # se = twostep / twostep.robust / naive the FINAL SEs do not depend on
      # this FIT.PA fit, so degrade gracefully: refit with the standard
      # (uncorrected) structural test. For se = local / local.nt the SEs ARE
      # read from this fit, so we cannot silently degrade -> re-raise.
      if (gamma_flag &&
          lavoptions$se %in% c("twostep", "twostep.robust",
                               "twostep.huber.white", "naive")) {
        lavoptions_pa$test <- "standard"
        fit_pa <- lavaan::lavaan(pts,
          sample_cov  = step1$VETA,
          sample_mean = step1$EETA,
          sample_nobs = nobs_1,
          nacov       = nacov,
          slot_options = lavoptions_pa,
          verbose     = FALSE
        )
        lav_msg_warn(gettext(
          "the two-step corrected structural test could not be computed for
           this model (eg a bi-factor measurement model); the standard
           (uncorrected) structural test is reported instead."))
      } else {
        lav_msg_stop(gettextf(
          "the structural model could not be fitted: %s",
          conditionMessage(fit_pa)))
      }
    }
  } else {
    fit_pa <- lavaan::lavaan(
      model = pts,
      slot_data = fit@Data,
      slot_sample_stats = fit@SampleStats,
      slot_options = lavoptions_pa,
      verbose = FALSE
    )
  }
  if (lav_verbose()) {
    cat("Fitting the structural part ... done.\n")
  }

  # check that the structural part is identified from the latent moments
  # alone: in the SAM approach the structural model is estimated from the
  # (estimated) latent variable moments, so it cannot borrow identification
  # from the measurement part (eg a non-recursive system without
  # instruments may 'fit' in sem() but has more structural parameters than
  # latent moments). Without this check the step-2 vcov machinery fails
  # cryptically further down.
  if (sam_method %in% c("local", "fsr", "cfsr")) {
    pa_df <- fit_pa@test[[1]]$df
    if (!is.null(pa_df) && !is.na(pa_df) && pa_df < 0L) {
      lav_msg_stop(gettextf(
        "the structural part of the model is not identified: it has more
         free parameters than there are (estimated) latent variable moments
         (df = %d). In the SAM approach the structural model must be
         identified from the latent variable moments alone. Consider
         simplifying the structural part, or using sem() instead.", pa_df))
    }
  }

  # which parameters from PTS do we wish to fill in:
  # - all 'free' parameters
  # - :=, <, > (if any)
  # - and NOT element with user=3 (add.exo.cov = TRUE, extra.int.idx)
  pts_idx <- which((pts$free > 0L | (pts$op %in% c(":=", "<", ">"))) &
    !pts$user == 3L)

  # find corresponding rows in PT
  pts2 <- as.data.frame(pts, stringsAsFactors = FALSE)
  pt_idx <- lav_pt_map_id_p1_in_p2(pts2[pts_idx, ], pt_1,
    exclude_nonpar = FALSE
  )
  # fill in
  pt_1$est[pt_idx] <- fit_pa@ParTable$est[pts_idx]

  # create step2.free.idx
  p2_idx <- seq_along(pt_1$lhs) %in% pt_idx & pt_1$free > 0 # no def!
  step2_free_idx <- step1$PT.free[p2_idx]

  # add 'step' column in PT
  pt_1$step <- rep(1L, length(pt_1$lhs))
  pt_1$step[seq_along(pt_1$lhs) %in% reg_idx] <- 2L

  step2 <- list(
    FIT.PA = fit_pa, PT = pt_1, reg.idx = reg_idx,
    step2.free.idx = step2_free_idx, extra.id = extra_id,
    pt.idx = pt_idx, pts.idx = pts_idx
  )

  step2
}

# rotation algorithms for multiple groups
#
# YR 22 Dec 2025 -- initial version
#
# - rstarts only affects the 'initial' rotation

# main function to rotate a list of matrices 'Alist'
lav_mat_rotate_mg <- function(a_list = NULL, # original matrices
                                 orthogonal = FALSE, # default is oblique
                                 method = "geomin", # default rot method
                                 method_args = list(
                                   geomin_epsilon = 0.01,
                                   orthomax_gamma = 1,
                                   cf_gamma = 0,
                                   oblimin_gamma = 0,
                                   promax_kappa = 4,
                                   target = matrix(0, 0, 0),
                                   target_mask = matrix(0, 0, 0)
                                 ),
                                 init_rot_list = NULL, # initial rotation matrix
                                 init_rot_check = TRUE, #check if init ROT is ok
                                 rstarts = 30L, # number of random starts
                                 # row_weights = "default", # row weighting
                                 # std.ov = FALSE, # rescale ov
                                 # ov.var = NULL, # ov variances
                                 # algorithm = "gpa", # rotation algorithm
                                 mg_algorithm = "pairwise",
                                 mg_agreement_method = "procrustes",
                                 mg_agreement_weight = 0.5,
                                 reflect = TRUE, # reflect sign
                                 order_lv_by = "index", # how to order the lv's
                                 # gpa.tol = 0.00001, # stopping tol gpa
                                 tol = 1e-07, # stopping tol others
                                 keep_rep = FALSE, # store replications
                                 max_iter = 10000L) { # max iterations

  # check Alist
  if (!is.list(a_list)) {
    lav_msg_stop(gettext("a_list does not seem to be a list"))
  }
  ngroups <- length(a_list)

  # check A
  a <- a_list[[1]]
  if (!inherits(a, "matrix")) {
    lav_msg_stop(gettext("a does not seem to be a matrix"))
  }

  p <- nrow(a)
  m <- ncol(a)
  if (m < 2L) { # single dimension
    res <- list(
      lambdaList = a_list, rotList = NULL,
      orthogonal = orthogonal, method = "none",
      method.args = list(), row_weights = "none",
      algorithm = "none", iter = 0L, converged = TRUE,
      method.value = 0
    )
    return(res)
  }

  # method
  method <- tolower(method)

  # no promax (for now)
  if (method == "promax") {
    lav_msg_stop(gettext("mg-promax + agreement not supported"))
  }

  # check init.ROT per group
  if (!is.null(init_rot_list) && init_rot_check) {
    for (g in seq_len(ngroups)) {
      init_rot <- init_rot_list[[g]]
      if (!inherits(init_rot, "matrix")) {
        lav_msg_stop(gettext("init.ROT does not seem to be a matrix"))
      }
      if (nrow(init_rot) != m) {
        lav_msg_stop(gettextf(
          "nrow(init.ROT) = %1$s does not equal ncol(A) = %2$s",
          nrow(init_rot), m
        ))
      }
      if (nrow(init_rot) != ncol(init_rot)) {
        lav_msg_stop(gettextf(
          "nrow(init.ROT) = %1$s does not equal ncol(init.ROT) = %2$s",
          nrow(init_rot), ncol(init_rot)
        ))
      }
      # rotation matrix?
      if (!lav_mat_rotate_check(init_rot, orthogonal = orthogonal)) {
        lav_msg_stop(gettext("init.ROT does not look like a rotation matrix"))
      }
    } # group
  } # check init.rotList

  # determine method function name + validate target/target_mask
  # (multigroup always uses the first group's target/target_mask)
  tmp_resolve <- lav_mat_rotate_resolve_method(
    method = method, method_args = method_args, p = p, m = m, group = 1L
  )
  method <- tmp_resolve$method
  method_fname <- tmp_resolve$method_fname
  method_args <- tmp_resolve$method_args

  # set orthogonal option
  if (missing(orthogonal)) {
    # the default is oblique, except for varimax, entropy and a few others
    if (method %in% c(
      "varimax", "entropy", "mccammon",
      "tandem1", "tandem2"
    )) {
      orthogonal <- TRUE
    } else {
      orthogonal <- FALSE
    }
  } else {
    if (!orthogonal && method %in% c(
      "varimax", "entropy", "mccammon",
      "tandem1", "tandem2"
    )) {
      lav_msg_warn(gettextf(
        "rotation method %s may not work with oblique rotation.",
        dQuote(method)
      ))
    }
  }

  # 0. initialize: rotate each group separately + reorder
  alist_orig <- a_list
  order_list <- vector("list", length = ngroups)
  for (g in seq_len(ngroups)) {
    if (lav_verbose()) {
      cat("  group = ", g, "\n")
    }
    res0 <- lav_mat_rotate(
      a = a_list[[g]],
      orthogonal = orthogonal,
      method = method,
      method_args = method_args,
      init_rot = NULL,
      init_rot_check = FALSE,
      rstarts = rstarts,
      row_weights = "none",
      std_ov = FALSE,
      ov_var = NULL,
      algorithm = "pairwise",
      reflect = reflect,
      order_lv_by = order_lv_by,
      tol = tol,
      max_iter = max_iter,
      group = g
    )
    if (g > 1L) {
      # reorder factors to align with the first group
      order_idx <- lav_efa_find_best_order(
        lambda_ref = a_list[[1]],
        lambda_target = res0$mm_lambda
      )
      a_list[[g]] <- res0$mm_lambda[, order_idx, drop = FALSE]
      order_list[[g]] <- order_idx
    } else {
      a_list[[g]] <- res0$mm_lambda
      order_list[[g]] <- res0$order.idx
    }
  } # group

  # rotate: mg-pairwise + agreement
  # (note: rstarts only affects the per-group initial rotation above)
  if (lav_verbose()) {
    cat("  mg-pairwise + agreement rotation:\n")
  }
  rot_list <- lav_mat_rotate_pairwise_mg(
    a_list = a_list,
    orthogonal = orthogonal,
    random_rot = FALSE,
    method_fname = method_fname,
    method_args = method_args,
    mg_agreement_method = mg_agreement_method,
    mg_agreement_weight = mg_agreement_weight,
    tol = tol,
    max_iter = max_iter
  )
  info <- attr(rot_list, "info")
  attr(rot_list, "info") <- NULL
  lambda_list <- vector("list", length = ngroups)
  for (g in seq_len(ngroups)) {
    a <- a_list[[g]]
    rot <- rot_list[[g]]

    if (orthogonal) {
      # LAMBDA <- A %*% solve(t(ROT))
      # note: when ROT is orthogonal, solve(t(ROT)) == ROT
      mm_lambda <- a %*% rot
    } else {
      # LAMBDA <- A %*% solve(t(ROT))
      mm_lambda <- t(solve(rot, t(a)))
    }

    #  # 3. undo row weighting
    #  LAMBDA <- LAMBDA / weights

    #  # 3.b undo cov -> cor
    #  if (std.ov) {
    #    LAMBDA <- LAMBDA * sqrt(ov.var)
    #  }

    # put back in List
    lambda_list[[g]] <- mm_lambda
  } # group

  # re-compute final rotation matrix, using Alist.orig
  # because in step 0, we already rotated (per group)
  rot_list <- vector("list", length = ngroups)
  if (orthogonal) {
    for (g in seq_len(ngroups)) {
      rot_list[[g]] <- solve(
        crossprod(alist_orig[[g]]),
        crossprod(alist_orig[[g]], lambda_list[[g]])
      )
    }
  } else {
    for (g in seq_len(ngroups)) {
      # to be compatible with GPa
      rott_inv <- solve(
        crossprod(alist_orig[[g]]),
        crossprod(alist_orig[[g]], lambda_list[[g]])
      )
      rot_list[[g]] <- solve(t(rott_inv))
    }
  }

  # 6. return results as a list
  res <- list(
    lambdaList = lambda_list, rotList = rot_list,
    orderList = order_list,
    orthogonal = orthogonal, method = method,
    method_args = method_args, row_weights = "none"
  )

  # add method info
  res <- c(res, info)

  res
}


# pairwise rotation algorithm with direct line search
# (see lav_mat_rotate_pairwise()
#
# but adapted to handle multiple groups + an agreement criterion
#
lav_mat_rotate_pairwise_mg <- function(a_list = NULL, # original matrices
                                          orthogonal = FALSE,
                                          random_rot = FALSE,
                                          method_fname = NULL, # crit function
                                          method_args = list(), # method args
                                          tol = 1e-8,
                                          mg_agreement_weight = 0.50,
                                          mg_agreement_method = "procrustes",
                                          scale = TRUE,
                                          max_iter = 1000L) {
  # first group is 'A'
  a <- a_list[[1]]
  ngroups <- length(a_list)

  # number of columns
  m <- ncol(a)

  # initial LAMBDA + PSI
  lambda_list <- a_list
  psi_list <- rep(list(diag(m)), ngroups)

  # random initial rotation?
  if (random_rot) {
    for (g in seq_len(ngroups)) {
      init_rot <- lav_mat_rotate_gen(m = m, orthogonal = TRUE)
      if (orthogonal) {
        lambda_list[[g]] <- a_list[[g]] %*% init_rot
      } else {
        lambda_list[[g]] <- t(solve(init_rot, t(a_list[[g]])))
        psi_list[[g]] <- crossprod(init_rot)
      }
    }
  }

  # using the current LAMBDA, evaluate the user-specified
  # rotation criterion; return Q (the criterion) only
  q_current <- lav_mat_rotate_mg_agreement(
    lambda_list = lambda_list, method_fname = method_fname,
    method_args = method_args,
    mg_agreement_method = mg_agreement_method,
    w = mg_agreement_weight, scale = scale
  )

  # if verbose, print
  if (lav_verbose()) {
    q_mg <- attr(q_current, "Q.mg")
    a_mg <- attr(q_current, "A.mg")
    cat(
      "  iter = ", sprintf("%4d", 0),
      " Q = ", sprintf("%13.11f", q_current),
      " Q.rot = ", sprintf("%13.11f", q_mg),
      " Q.agr = ", sprintf("%13.11f", a_mg), "\n"
    )
  }

  # plane combinations
  if (orthogonal) {
    plane <- utils::combn(m, 2)
  } else {
    tmp <- utils::combn(m, 2)
    plane <- cbind(tmp, tmp[c(2, 1), , drop = FALSE])
  }

  # define objective function -- orthogonal
  objf_orth <- function(g = 1L, theta = 0, a = NULL, col1 = 0L, col2 = 0L) {
    # construct ROT
    rot <- lav_mat_givens_orth(m, col1, col2, theta)

    # rotate
    lambda_list[[g]] <- a %*% rot

    # evaluate criterion
    q_1 <- lav_mat_rotate_mg_agreement(
      lambda_list = lambda_list, method_fname = method_fname,
      method_args = method_args, mg_agreement_method = mg_agreement_method,
      w = mg_agreement_weight, scale = scale
    )

    q_1
  }

  # define objective function -- oblique
  # ('a' is the current loading matrix for group g, passed in by optimize())
  objf_obliq <- function(g = 1, delta = 0, a = NULL, col1 = 0L, col2 = 0L,
                         psi12 = 0) {
    # construct ROT
    rot <- lav_mat_givens_obliq(m, col1, col2, delta, psi12)

    # rotate
    lambda_list[[g]] <- a %*% rot

    # evaluate criterion
    q_1 <- lav_mat_rotate_mg_agreement(
      lambda_list = lambda_list, method_fname = method_fname,
      method_args = method_args, mg_agreement_method = mg_agreement_method,
      w = mg_agreement_weight, scale = scale
    )

    q_1
  }

  # start iterations
  converged <- FALSE
  q_old <- q_current
  for (iter in seq_len(max_iter)) {
    # for each group
    for (g in seq_len(ngroups)) {
      # rotate - one cycle
      for (pl in seq_len(ncol(plane))) {
        # choose plane
        col1 <- plane[1, pl]
        col2 <- plane[2, pl]

        # optimize
        if (orthogonal) {
          out <- optimize(
            f = objf_orth, interval = c(-pi / 4, +pi / 4),
            g = g,
            a = lambda_list[[g]], col1 = col1, col2 = col2,
            maximum = FALSE, tol = .Machine$double.eps^0.25
          )
          # best rotation - for this plane
          theta <- out$minimum

          # construct ROT
          rot <- lav_mat_givens_orth(m, col1, col2, theta)
        } else {
          psi12 <- psi_list[[g]][col1, col2]
          out <- optimize(
            f = objf_obliq, interval = c(-1, +1),
            g = g,
            a = lambda_list[[g]], col1 = col1, col2 = col2,
            psi12 = psi12,
            maximum = FALSE, tol = .Machine$double.eps^0.25
          )

          # best rotation - for this plane
          delta <- out$minimum

          # construct ROT (rot[col1, col1] == gamma, the PSI scaling factor)
          rot <- lav_mat_givens_obliq(m, col1, col2, delta, psi12)
          gamma <- rot[col1, col1]
        }

        # rotate
        lambda_list[[g]] <- lambda_list[[g]] %*% rot

        if (!orthogonal) {
          mm_psi <- psi_list[[g]]
          # rotate PSI
          mm_psi[col1, ] <- (1 / gamma) * mm_psi[col1, ] +
            (delta / gamma) * mm_psi[col2, ]
          mm_psi[, col1] <- mm_psi[col1, ]
          mm_psi[col1, col1] <- 1
          psi_list[[g]] <- mm_psi
        }
      } # all planes
    } # ngroups

    # check for convergence
    q_current <- lav_mat_rotate_mg_agreement(
      lambda_list = lambda_list, method_fname = method_fname,
      method_args = method_args, mg_agreement_method = mg_agreement_method,
      w = mg_agreement_weight, scale = scale
    )

    # absolute change in Q
    diff <- abs(q_old - q_current)

    # if verbose, print
    if (lav_verbose()) {
      q_mg <- attr(q_current, "Q.mg")
      a_mg <- attr(q_current, "A.mg")
      cat(
        "  iter = ", sprintf("%4d", iter),
        " Q = ", sprintf("%13.11f", q_current),
        " Q.rot = ", sprintf("%13.11f", q_mg),
        " Q.agr = ", sprintf("%13.11f", a_mg),
        " change = ", sprintf("%13.11f", diff), "\n"
      )
    }

    if (!is.finite(diff)) {
      # something went wrong ... bail out
      lav_msg_warn(gettextf("mg-pairwise rotation algorithm failed!"))
      break
    } else if (diff < tol) {
      converged <- TRUE
      break
    } else {
      q_old <- q_current
    }
  } # iter

  # warn if no convergence
  if (!converged) {
    lav_msg_warn(gettextf(
      "mg-pairwise rotation algorithm did not converge after %s iterations",
      max_iter
    ))
  }

  # compute final rotation matrix
  rot_list <- vector("list", length = ngroups)
  if (orthogonal) {
    for (g in seq_len(ngroups)) {
      rot_list[[g]] <- solve(
        crossprod(a_list[[g]]),
        crossprod(a_list[[g]], lambda_list[[g]])
      )
    }
  } else {
    for (g in seq_len(ngroups)) {
      # to be compatible with GPa
      rott_inv <- solve(
        crossprod(a_list[[g]]),
        crossprod(a_list[[g]], lambda_list[[g]])
      )
      rot_list[[g]] <- solve(t(rott_inv))
    }
  }

  # algorithm information
  info <- list(
    algorithm = "mg-pairwise",
    iter = iter,
    converged = converged,
    method.value = q_current
  )

  attr(rot_list, "info") <- info

  rot_list
}

# user visible function to add 'matrix' entries in the parameter table
lavMatrixRepresentation <- function(partable, representation = "LISREL", # nolint
                                    allow_composites = TRUE, # new in 0.6-20
                                    add_attributes = FALSE,
                                    as_data_frame = TRUE,
                                  ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)

  # check parameter table
  partable <- lav_pt_complete(partable)

  # get model matrices
  if (representation == "LISREL") {
    rep_1 <- lav_lisrel(partable, target = NULL, extra = add_attributes,
                      allow_composites = allow_composites)
  } else if (representation == "RAM") {
    rep_1 <- lav_ram(partable, target = NULL, extra = add_attributes)
  } else {
    lav_msg_stop(gettext("representation must be either \"LISREL\" or \"RAM\"."))
  }

  partable$mat <- rep_1$mat
  partable$row <- rep_1$row
  partable$col <- rep_1$col

  if (as_data_frame) {
    partable <- as.data.frame(partable, stringsAsFactors = FALSE)
    class(partable) <- c("lavaan.data.frame", "data.frame")
  }

  if (add_attributes) {
    if (representation == "LISREL") {
      attr(partable, "ov.dummy.names.nox") <- attr(rep_1, "ov.dummy.names.nox")
      attr(partable, "ov.dummy.names.x") <- attr(rep_1, "ov.dummy.names.x")
    } else if (representation == "RAM") {
      attr(partable, "ov.idx") <- attr(rep_1, "ov.idx")
    }
    attr(partable, "mmNames") <- attr(rep_1, "mmNames")
    attr(partable, "mmNumber") <- attr(rep_1, "mmNumber")
    attr(partable, "mmRows") <- attr(rep_1, "mmRows")
    attr(partable, "mmCols") <- attr(rep_1, "mmCols")
    attr(partable, "mmDimNames") <- attr(rep_1, "mmDimNames")
    attr(partable, "mmSymmetric") <- attr(rep_1, "mmSymmetric")
  }

  partable
}

# contributed by Myrsini Katsikatsou (March 2016)

# the function lav_pml_bi_lik_x gives the value of the bivariate likelihood
# for a specific pair of ordinal variables casewise when covariates are present
# and estimator=="PML" (the bivariate likelihood is essentially the bivariate
# probability of the observed response pattern of two ordinal variables)

# Input arguments:
# Y1 is a vector, includes the observed values for the first variable for all
#  cases/units, Y1 is ordinal
# Y2 similar to Y1
# Rho is the polychoric correlation of Y1 and Y2
# th.y1 is the vector of the thresholds for Y1* excluding the first and
# the last thresholds which are -Inf and Inf
# th.y2 is similar to th.y1
# eXo is the data for the covariates in a matrix format where
# nrows= no of cases, ncols= no of covariates
# PI.y1 is a vector, includes the regression coefficients of the covariates
# for the first variable, Y1, the length of the vector is the no of covariates;
# to obtain this vector apply the function
#                                   lav_model_pi()[row_correspondin_to_Y1, ]
# PI.y2 is similar to PI.y1
# missing.ind is of "character" value, taking the values listwise, pairwise,
# available_cases; to obtain a value use lavdata@missing

# Output:
# It is a vector, length= no of cases, giving the bivariate likelihood
#  for each case.
lav_pml_bi_lik_x <- function(y1, y2, rho,
                               th_y1, th_y2,
                               exo,
                               pi_y1, pi_y2,
                               missing_ind) {
  th_y1 <- c(-100, th_y1, 100)
  th_y2 <- c(-100, th_y2, 100)
  pred_y1 <- c(exo %*% pi_y1)
  pred_y2 <- c(exo %*% pi_y2)

  th_y1_upper <- th_y1[y1 + 1L] - pred_y1
  th_y1_lower <- th_y1[y1] - pred_y1
  th_y2_upper <- th_y2[y2 + 1L] - pred_y2
  th_y2_lower <- th_y2[y2] - pred_y2

  if (missing_ind == "listwise") { # I guess this is the default which
    # also handles the case of complete data
    biv_prob <- pbivnorm(th_y1_upper, th_y2_upper, rho = rho) -
      pbivnorm(th_y1_lower, th_y2_upper, rho = rho) -
      pbivnorm(th_y1_upper, th_y2_lower, rho = rho) +
      pbivnorm(th_y1_lower, th_y2_lower, rho = rho)
    lik <- biv_prob
  } else if (missing_ind %in% c(
    "pairwise",
    "available.cases",
    "available_cases"
  )) {
    # index of cases with complete pairs
    cp_idx <- which(complete.cases(cbind(y1, y2)))

    th_y1_upper <- th_y1_upper[cp_idx]
    th_y1_lower <- th_y1_lower[cp_idx]
    th_y2_upper <- th_y2_upper[cp_idx]
    th_y2_lower <- th_y2_lower[cp_idx]

    biv_prob <- pbivnorm(th_y1_upper, th_y2_upper, rho = rho) -
      pbivnorm(th_y1_lower, th_y2_upper, rho = rho) -
      pbivnorm(th_y1_upper, th_y2_lower, rho = rho) +
      pbivnorm(th_y1_lower, th_y2_lower, rho = rho)

    # lik <- numeric( length(Y1) )
    lik <- rep(as.numeric(NA), length(y1))
    lik[cp_idx] <- biv_prob
  }
  lik
}

#################################################################


# The function  lav_pml_uni_lik gives the value of the univariate likelihood
# for a specific ordinal variable, casewise (which is essentially the
# probability for the observed response category for each case).
# The input arguments are explained before the function lav_pml_bi_lik_x above.
# Output:
# It is a vector, length= no of cases, giving the univariate likelihoods
# for each case.

lav_pml_uni_lik <- function(y1, th_y1, exo = NULL, pi_y1 = NULL) {
  th_y1 <- c(-100, th_y1, 100)
  if (!is.null(exo)) {
    pred_y1 <- c(exo %*% pi_y1)
  }

  if (is.null(exo)) {
    th_y1_upper <- th_y1[y1 + 1L]
    th_y1_lower <- th_y1[y1]
  } else {
    th_y1_upper <- th_y1[y1 + 1L] - pred_y1
    th_y1_lower <- th_y1[y1] - pred_y1
  }

  pml_uni_lik <- pnorm(th_y1_upper) - pnorm(th_y1_lower)

  pml_uni_lik[is.na(pml_uni_lik)] <- 0

  pml_uni_lik
}

#################################################################

# The function lav_pml_th_uni_prob gives the model-based univariate
# probabilities for all ordinal indicators and for all of their response
# categories, i.e. pi(xi=a), where a=1,...,ci and i=1,...,p with a index
# running faster than i index.
# Input arguments:
# TH is a vector giving the thresholds for all variables, tau_ia, with a running
#    faster than i (the first and the last thresholds which are -Inf and Inf are
#    not included). TH can be given by the lavaan function lav_model_th .
# th.idx is a vector of same length as TH which gives the value of the i index,
#      namely which variable each thresholds refers to. This can be obtained by
#      lavmodel@th.idx .
# Output:
# It is a vector, length= Sum_i(ci), i.e. the sum of the response categories of
# all ordinal variables. The vector contains the model-based univariate
# probabilities pi(xi=a).

lav_pml_th_uni_prob <- function(th = th, th_idx = th_idx) {
  th_split <- split(th, th_idx)
  th_lower <- unlist(lapply(th_split, function(x) {
    c(-100, x)
  }), use.names = FALSE)
  th_upper <- unlist(lapply(th_split, function(x) {
    c(x, 100)
  }), use.names = FALSE)
  prob <- pnorm(th_upper) - pnorm(th_lower)
  # to avoid Nan/-Inf
  prob[prob < .Machine$double.eps] <- .Machine$double.eps
  prob
}

#############################################################################


# The function lav_pml_dbilogl_dpar_x computes the derivatives of a bivariate
# log-likelihood of two ordinal variables casewise with respect to thresholds,
# slopes (reduced-form regression coefficients for the covariates), and
# polychoric correlation.
# The function lav_dbinorm of lavaan is used.
# The function gives the right result for both listwise and pairwise deletion,
# and the case of complete data.
# Input arguments are explained before the function lav_pml_bi_lik_x
# defined above. The only difference is that PI.y1 and PI.y2 are (accidentally)
# renamed here as sl.y1 and sl.y2
# Output:
# It is a list containing the following
# a) the derivatives w.r.t. the thresholds of the first variable casewise.
#    This is a matrix, nrows=no of cases, ncols= no of thresholds of variable 1.
# b) the derivatives w.r.t. the thresholds of the second variable casewise.
#    This is a matrix, nrows=no of cases, ncols= no of thresholds of variable 2.
# c) the derivatives w.r.t slopes for variable 1.  This is a matrix, where
#    nrows=no of cases, ncols= no of covariates.
# d) the derivatives w.r.t slopes for variable 2.  This is a matrix, where
#    nrows=no of cases, ncols= no of covariates.
# e) the derivative w.r.t the polychoric correlation of the two variables.
#    This is a vector of length= no of cases.


lav_pml_dbilogl_dpar_x <- function(y1, y2, exo, rho,
                                      th_y1, th_y2,
                                      sl_y1, sl_y2,
                                      missing_ind) {
  nth_y1 <- length(th_y1)
  nth_y2 <- length(th_y2)

  start_th_y1 <- th_y1
  start_th_y2 <- th_y2

  nobs_1 <- length(y1)

  r <- sqrt(1 - rho * rho)
  th_y1 <- c(-100, th_y1, 100)
  th_y2 <- c(-100, th_y2, 100)
  pred_y1 <- c(exo %*% sl_y1)
  pred_y2 <- c(exo %*% sl_y2)

  th_y1_z1 <- th_y1[y1 + 1L] - pred_y1
  th_y1_z2 <- th_y1[y1] - pred_y1
  th_y2_z1 <- th_y2[y2 + 1L] - pred_y2
  th_y2_z2 <- th_y2[y2] - pred_y2

  # lik, i.e. the bivariate probability case-wise
  lik <- lav_pml_bi_lik_x(
    y1 = y1, y2 = y2,
    rho = rho,
    th_y1 = start_th_y1,
    th_y2 = start_th_y2,
    exo = exo,
    pi_y1 = sl_y1,
    pi_y2 = sl_y2,
    missing_ind = missing_ind
  )


  # w.r.t. th.y1, mean tau tilde
  # derivative bivariate prob w.r.t. tau^xi_ci, see formula in paper 2012
  y1_z1 <- dnorm(th_y1_z1) * (pnorm((th_y2_z1 - rho * th_y1_z1) / r) -
    pnorm((th_y2_z2 - rho * th_y1_z1) / r))
  # derivative bivariate prob w.r.t. tau^xi_(ci-1),
  y1_z2 <- (-1) * (dnorm(th_y1_z2) * (pnorm((th_y2_z1 - rho * th_y1_z2) / r) -
    pnorm((th_y2_z2 - rho * th_y1_z2) / r)))


  # allocate the derivatives at the right column casewise
  idx_y1_z1 <-
    matrix(1:nth_y1, nrow = nobs_1, ncol = nth_y1, byrow = TRUE) == y1
  idx_y1_z2 <-
    matrix(1:nth_y1, nrow = nobs_1, ncol = nth_y1, byrow = TRUE) == (y1 - 1L)
  der_table_y1 <- idx_y1_z1 * y1_z1 + idx_y1_z2 * y1_z2

  # der of pl w.r.t. th.y1
  dx_th_tilde_y1 <- der_table_y1 / lik
  dx_th_tilde_y1[is.na(dx_th_tilde_y1)] <- 0

  # w.r.t. th.y2, mean tau tilde
  # derivative bivariate prob w.r.t. tau^xi_ci, see formula in paper 2012
  y2_z1 <- dnorm(th_y2_z1) * (pnorm((th_y1_z1 - rho * th_y2_z1) / r) -
    pnorm((th_y1_z2 - rho * th_y2_z1) / r))
  # derivative bivariate prob w.r.t. tau^xi_(ci-1),
  y2_z2 <- (-1) * (dnorm(th_y2_z2) * (pnorm((th_y1_z1 - rho * th_y2_z2) / r) -
    pnorm((th_y1_z2 - rho * th_y2_z2) / r)))
  # allocate the derivatives at the right column casewise
  idx_y2_z1 <-
    matrix(1:nth_y2, nrow = nobs_1, ncol = nth_y2, byrow = TRUE) == y2
  idx_y2_z2 <-
    matrix(1:nth_y2, nrow = nobs_1, ncol = nth_y2, byrow = TRUE) == (y2 - 1L)
  der_table_y2 <- idx_y2_z1 * y2_z1 + idx_y2_z2 * y2_z2

  # der of pl w.r.t. th.y2
  dx_th_tilde_y2 <- der_table_y2 / lik
  dx_th_tilde_y2[is.na(dx_th_tilde_y2)] <- 0



  # w.r.t. rho
  # derivative bivariate prob w.r.t. rho, see formula in paper 2012
  dbivprob_wrt_rho <- (lav_dbinorm(th_y1_z1, th_y2_z1, rho) -
    lav_dbinorm(th_y1_z2, th_y2_z1, rho) -
    lav_dbinorm(th_y1_z1, th_y2_z2, rho) +
    lav_dbinorm(th_y1_z2, th_y2_z2, rho))
  # der of pl w.r.t. rho
  dx_rho <- dbivprob_wrt_rho / lik
  dx_rho[is.na(dx_rho)] <- 0


  # der of pl w.r.t. slopes (also referred to PI obtained by
  #     lav_model_pi function)
  row_sums_y1 <- rowSums(dx_th_tilde_y1)
  row_sums_y2 <- rowSums(dx_th_tilde_y2)
  dx_sl_y1 <- (-1) * exo * row_sums_y1
  dx_sl_y2 <- (-1) * exo * row_sums_y2


  list(
    dx_th_y1 = dx_th_tilde_y1, # note that dx.th.tilde=dx.th
    dx_th_y2 = dx_th_tilde_y2,
    dx_sl_y1 = dx_sl_y1,
    dx_sl_y2 = dx_sl_y2,
    dx_rho = dx_rho
  )
}

###############################################################


# The function lav_pml_uni_sc gives, casewise, the derivative of a
# univariate log-likelihood w.r.t. thresholds and slopes if present weighted by
# the casewise uni-weights as those defined in AC-PL (essentially the number of
# missing values per case).
# The function closely follows the "logic" of the function
# lav_pml_dbilogl_dpar_x defined above.
# Input arguments are as before plus: weights.casewise given by
# lavcavhe$uniweights.casewise .
# Output:
# A list including the following:
# a) the derivatives w.r.t. the thresholds of the variable. This is a matrix,
#    nrows=no of cases, ncols= no of thresholds of variable 1.
# b) the derivatives w.r.t slopes for the variable. If covariates are present,
#    this is a matrix, nrows=no of cases, ncols= no of covariates.
#    Otherwise it takes the value NULL.


lav_pml_uni_sc <- function(y1, th_y1, exo = NULL, sl_y1 = NULL,
                       weights_casewise) {
  nth_y1 <- length(th_y1)
  nobs_1 <- length(y1)
  th_y1_orig <- th_y1 # unpadded (lav_pml_uni_lik() pads itself)
  th_y1 <- c(-100, th_y1, 100)

  if (is.null(exo)) {
    th_y1_z1 <- th_y1[y1 + 1L]
    th_y1_z2 <- th_y1[y1]
  } else {
    pred_y1 <- c(exo %*% sl_y1)
    th_y1_z1 <- th_y1[y1 + 1L] - pred_y1
    th_y1_z2 <- th_y1[y1] - pred_y1
  }

  # lik, i.e. the univariate probability case-wise
  # (fixed July 2026: pass the UNPADDED thresholds -- lav_pml_uni_lik()
  # pads with -100/100 itself; passing the padded vector shifted every
  # category's probability and made the first category's likelihood zero,
  # producing Inf/garbage in the weights_casewise/lik factor below)
  lik <- lav_pml_uni_lik( # Y1 = X[,i],
    y1 = y1,
    # th_y1 = TH[th.idx==i],
    th_y1 = th_y1_orig,
    exo = exo,
    # pi_y1 = PI[i,])
    pi_y1 = sl_y1
  )

  # w.r.t. th.y1
  # derivative of the univariate prob w.r.t. to the upper limit threshold
  y1_z1 <- dnorm(th_y1_z1)
  # derivative of the univariate prob w.r.t. to the lower limit threshold
  y1_z2 <- (-1) * dnorm(th_y1_z2)

  # allocate the derivatives at the right column casewise
  idx_y1_z1 <-
    matrix(1:nth_y1, nrow = nobs_1, ncol = nth_y1, byrow = TRUE) == y1
  idx_y1_z2 <-
    matrix(1:nth_y1, nrow = nobs_1, ncol = nth_y1, byrow = TRUE) == (y1 - 1L)
  der_table_y1 <- idx_y1_z1 * y1_z1 + idx_y1_z2 * y1_z2

  # der of pl w.r.t. th.y1
  dx_th_tilde_y1 <- der_table_y1 * (weights_casewise / lik)
  dx_th_tilde_y1[is.na(dx_th_tilde_y1)] <- 0

  # der of pl w.r.t. slopes (also referred to PI obtained by
  #     lav_model_pi function)
  dx_sl_y1 <- NULL
  if (!is.null(exo)) {
    row_sums_y1 <- rowSums(dx_th_tilde_y1)
    dx_sl_y1 <- (-1) * exo * row_sums_y1
  }

  list(
    dx.th.y1 = dx_th_tilde_y1, # note that dx.th.tilde=dx.th
    dx.sl.y1 = dx_sl_y1
  )
}

# Standardize parameter estimates of a fitted lavaan model.
#
# Three "modes" of standardization are supported, all implemented on top of a
# single observed-variable engine (lav_standardize_all):
#   - std.lv  : only the latent variables are standardized
#   - std.all : latent *and* observed variables are standardized
#   - std.nox : like std.all, but the (observed) exogenous 'x' variables are
#               left unstandardized
# std.nox is in fact a special case of a more general mechanism: standardize
# all observed variables *except* a given set (here: the exogenous 'x'). The
# 'ov_std' argument of lav_standardize_all() exposes this generality, so that a
# caller can request that only a user-specified subset of observed variables is
# standardized.

# ---------------------------------------------------------------------------
# internal helpers
# ---------------------------------------------------------------------------

# Resolve the standard set of ingredients (lavmodel, lavpartable, partable,
# est, glist, cov_x) from either a fitted lavobject or the individual pieces.
lav_standardize_setup <- function(lavobject = NULL, lavmodel = NULL,
                                  lavpartable = NULL, partable = NULL,
                                  est = NULL, glist = NULL, cov_x = NULL,
                                  need_cov_x = FALSE) {
  if (is.null(lavobject)) {
    stopifnot(!is.null(lavmodel), !is.null(lavpartable))
    if (is.null(est)) {
      if (!is.null(lavpartable$est)) {
        est <- lavpartable$est
                # if this changes, tag @TDJorgensen in commit message
      } else {
        lav_msg_stop(gettext("could not find `est' in lavpartable"))
      }
    }
  } else {
    lavmodel <- lavobject@Model
    lavpartable <- lavobject@ParTable
    if (is.null(est)) {
      est <- lav_inspect_est(lavobject)
    }
    if (need_cov_x && lavmodel@conditional.x && is.null(cov_x)) {
      if (!is.null(lavobject@implied$cov.x[[1]])) {
        cov_x <- lavobject@implied$cov.x
                # if this changes, tag @TDJorgensen in commit message
      } else {
        # perhaps a lavaanList object: extract it from GLIST per block
        cov_x <- vector("list", length = lavmodel@nblocks)
        for (b in seq_len(lavmodel@nblocks)) {
          mm_in_block <- (seq_len(lavmodel@nmat[b]) +
            cumsum(c(0, lavmodel@nmat))[b])
          mlist <- lavmodel@GLIST[mm_in_block]
          cov_x[[b]] <- mlist[["cov.x"]]
        }
      }
    }
  }

  if (is.null(partable)) {
    partable <- lavpartable
  }
  if (is.null(glist)) {
    glist <- lavmodel@GLIST
  }

  list(lavmodel = lavmodel, lavpartable = lavpartable, partable = partable,
       est = est, glist = glist, cov_x = cov_x)
}

# Evaluate the standardized values of the defined (:=), equality (==) and
# inequality (< >) rows, given the (already standardized) 'out' vector.
lav_standardize_constraints <- function(out, partable, lavmodel) {
  x <- out[partable$free & !duplicated(partable$free)]

  idx <- which(partable$op == ":=")
  if (length(idx) > 0L) {
    out[idx] <- lavmodel@def.function(x)
  }
  idx <- which(partable$op == "==")
  if (length(idx) > 0L) {
    out[idx] <- lavmodel@ceq.function(x)
  }
  idx <- which(partable$op == "<" | partable$op == ">")
  if (length(idx) > 0L) {
    out[idx] <- lavmodel@cin.function(x)
  }

  out
}

# Observed-variable standard deviations (and variances) for block 'g'.
# Returns the (possibly conditional.x-extended) ov.names together with the
# vectors of SDs ('ov') and variances ('ov2').
lav_standardize_ov_sd <- function(lavpartable, g, vy = NULL, ov_var = NULL,
                                  conditional_x = FALSE, cov_x = NULL) {
  ov_names <- lav_pt_vnames(lavpartable, "ov", block = g) # not user

  if (is.null(ov_var)) {
    ov2 <- vy[[g]]
    # replace zero values by NA (but keep negative values)
    zero_idx <- which(abs(ov2) < .Machine$double.eps)
    if (length(zero_idx) > 0L) {
      ov2[zero_idx] <- as.numeric(NA)
    }
    # replace negative values by NA (for sqrt)
    tmp_ov2 <- ov2
    neg_idx <- which(tmp_ov2 < 0)
    if (length(neg_idx) > 0L) {
      tmp_ov2[neg_idx] <- as.numeric(NA)
    }
    ov <- sqrt(tmp_ov2)
  } else {
    ov2 <- ov_var[[g]]
    ov <- sqrt(ov2)
  }

  if (conditional_x) {
    # extend OV with ov.names.x
    ov_names_x <- lav_pt_vnames(lavpartable, "ov.x", block = g)
    ov_names_nox <- lav_pt_vnames(lavpartable, "ov.nox", block = g)
    ov_names <- c(ov_names_nox, ov_names_x)
    ov2 <- c(ov2, diag(cov_x[[g]]))
    ov <- c(ov, sqrt(diag(cov_x[[g]])))
  }

  list(ov_names = ov_names, ov = ov, ov2 = ov2)
}

# scaling corrections for two-way latent product terms (interactions
# X:Z, quadratic terms X:X) in block 'g' of lav_standardize_lv(), based on
# Brandt et al. (2015, Eq. 13) and Kelava & Brandt (2022). The standardized
# solution treats a product term as the product of the centered,
# standardized components, whereas the model parameters describe the raw
# product A:B. Three corrections are needed, all exact (no distributional
# assumptions):

# (1) scaling SD (FV): the SD of a standardized product term is
# SD(A)*SD(B), not SD(A:B). eta[A:B] is replaced accordingly, so that the
# generic scaling in the caller applies the right factor everywhere A:B
# appears.

# (2) simple main effects (KS, 08/05/26): with uncentered components, b1 in
# y ~ b1*x + b2*z + b3*x:z + b4*x:x is the slope at x = z = 0. The
# standardized solution reports the slope at the means, i.e. the partial
# derivative dy/dx = b1 + b3*z + 2*b4*x evaluated at E(x), E(z). Looping
# over the product terms, each term adds beta_ab * E(other component) to
# the main effect of each of its components.

# (3) product moments (FV, 05/07/26): the psi/alpha rows of A:B describe the
# uncentered product. Without centering them first, the standardized
# moments would depend on how the mean structure is parameterized. With
# ac = A - E(A) and bc = B - E(B) we have the exact identity
# ac*bc = A:B - E(B)*A - E(A)*B + E(A)*E(B), so the centered product is
# a linear combination of elements of eta. Collect the weights in a
# matrix K (one row per variable, regular lvs keep their identity row),
# e.g. for eta = (X, Z, X:Z, X:X) with mx = E(X), mz = E(Z):
# row X:Z = (-mz, -mx, 1, 0), row X:X = (-2*mx, 0, 0, 1). Then
# K %*% PSI %*% t(K) holds all centered (co)variances at once (constants
# do not affect them), and the centered means follow from
# E(ac*bc) = E(A:B) - E(A)*E(B). K acts on psi_g (not VETA): the "~~"
# rows are psi elements, and for a residual covariance with an
# endogenous lv, PSI[X, Y] = Cov(X, zeta_Y) is exactly what the
# centering needs. The corrected rows are also written into est_c,
# because the cov_std block in the caller derives its sqrt(variance)
# divisors from est_c (not out, which is scaled by then).

# Returns the corrected (out, est_c, eta)
lav_standardize_lv_int <- function(out, est_c, eta, est, partable, lv_names,
                                   lv_int_names, eeta, psi_g, g) {
  # component indices of each usable product term (int = A:B, a = A, b = B)
  int_names_ok <- character(0L)
  int_comp <- list()
  for (int_name in lv_int_names) {
    components <- strsplit(int_name, ":", fixed = TRUE)[[1L]]
    # only two-way terms: a three-way (or higher) product would need its own
    # decomposition, and the product of three SDs
    if (length(components) != 2L) next
    idx_int <- match(int_name, lv_names)
    idx_a   <- match(components[1L], lv_names)
    idx_b   <- match(components[2L], lv_names)
    # need a, b and the interaction term before moving on
    if (is.na(idx_a) || is.na(idx_b) || is.na(idx_int)) next
    int_names_ok <- c(int_names_ok, int_name)
    int_comp <- c(int_comp,
      list(list(int = idx_int, a = idx_a, b = idx_b)))
  }
  if (length(int_comp) == 0L) {
    return(list(out = out, est_c = est_c, eta = eta))
  }

  # only used by the shift in (2), which needs nonzero means, but guarding
  # it with any(eeta != 0) would cost as much as building it
  reg_row <- partable$op == "~" & partable$block == g
  for (i in seq_along(int_comp)) {
    p <- int_comp[[i]]

    # (1) replace ETA for the product term with SD(A)*SD(B)
    eta[p$int] <- eta[p$a] * eta[p$b]

    # (2) shift the simple main effects to the mean-evaluated slopes
    exp_a <- eeta[p$a]
    exp_b <- eeta[p$b]
    if (exp_a != 0 || exp_b != 0) {
      int_row <- reg_row & partable$rhs == int_names_ok[i]
      for (dep in unique(partable$lhs[int_row])) {
        # beta_ab is read from est (not partable$est): in the *_x wrappers
        # (delta method, bootstrap) est is the perturbed parameter vector, and
        # the gradient of the shifted main effects wrt beta_ab must not be lost
        beta_ab <- est[int_row & partable$lhs == dep]
        dep_row <- reg_row & partable$lhs == dep
        idx_beta_a <- which(dep_row & partable$rhs == lv_names[p$a])
        idx_beta_b <- which(dep_row & partable$rhs == lv_names[p$b])
        out[idx_beta_a] <- out[idx_beta_a] + beta_ab * exp_b
        out[idx_beta_b] <- out[idx_beta_b] + beta_ab * exp_a
      }
    }
  }

  # (3) center the (co)variances and means of the product terms
  if (!is.null(psi_g) && any(eeta != 0)) {
    kmat <- diag(nrow(psi_g))
    for (p in int_comp) {
      kmat[p$int, p$a] <- kmat[p$int, p$a] - eeta[p$b]
      kmat[p$int, p$b] <- kmat[p$int, p$b] - eeta[p$a]
    }
    # shift_mat holds, for every pair of latent variables, the mean-driven
    # part of their (co)variance (zero unless a product term is involved)
    shift_mat <- psi_g - kmat %*% psi_g %*% t(kmat)

    # "~~" rows involving at least one product term: this covers the
    # variance of a product term (X:Z ~~ X:Z), its covariance with a
    # regular lv (X:Z ~~ V), and product-product covariances (X:Z ~~ X:X).
    # Rows whose other side is not a lv are left alone (keep)
    idx_cov <- which(partable$op == "~~" & partable$block == g &
      (partable$lhs %in% int_names_ok |
       partable$rhs %in% int_names_ok))
    l_idx <- match(partable$lhs[idx_cov], lv_names)
    r_idx <- match(partable$rhs[idx_cov], lv_names)
    keep <- which(!is.na(l_idx) & !is.na(r_idx))
    shift <- shift_mat[cbind(l_idx[keep], r_idx[keep])]
    out[idx_cov[keep]] <- out[idx_cov[keep]] - shift
    # of est_c only the variance entries are read back later, but centering
    # all touched rows keeps est_c fully consistent and a diagonal-only
    # filter would cost as much as it saves
    est_c[idx_cov[keep]] <- est_c[idx_cov[keep]] - shift

    # "~1" rows of the product terms: E(ac*bc) = E(A:B) - E(A)*E(B). After
    # the generic scaling this equals, e.g., cor(X, Z) for X:Z
    idx_mean <- which(partable$op == "~1" & partable$block == g &
      partable$lhs %in% int_names_ok)
    for (row_i in idx_mean) {
      p <- int_comp[[match(partable$lhs[row_i], int_names_ok)]]
      out[row_i] <- out[row_i] - eeta[p$a] * eeta[p$b]
    }
  }

  list(out = out, est_c = est_c, eta = eta)
}

# Given a parameter vector 'x', set it as the model parameters and return the
# (glist, est) pair to feed into the standardization engines. Handles the
# (EFA) rotation case. Used by the bootstrap/Monte-Carlo machinery.
lav_standardize_prepare_x <- function(x, lavobject, rotation = FALSE) {
  lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x)

  if (rotation) {
    lavmodel@GLIST <- lavTech(lavobject, "est.unrotated") # unrotated!
    est_rot <- lav_model_efa_rotate_x(
      x = x,
      lavmodel = lavmodel, # unrotated!
      lavoptions = lavobject@Options,
      init_rot = lavmodel@H,
      type = "user",
      extra = TRUE
    )
    glist <- attr(est_rot, "extra")$glist
    attributes(est_rot) <- NULL
    est <- est_rot
  } else {
    glist <- lavmodel@GLIST
                    # if this changes, tag @TDJorgensen in commit message
    est <- lav_model_get_parameters(lavmodel, type = "user")
  }

  list(glist = glist, est = est)
}

# ---------------------------------------------------------------------------
# 'x' wrappers (parameter vector -> standardized values), used for bootstrap
# and Monte-Carlo standard errors.
# ---------------------------------------------------------------------------

lav_standardize_lv_x <- function(x, lavobject, partable = NULL, cov_std = TRUE,
                                 lv_var = NULL, rotation = FALSE) {
  ge <- lav_standardize_prepare_x(x, lavobject, rotation = rotation)
  lav_standardize_lv(
    lavobject = lavobject, partable = partable, est = ge$est,
    glist = ge$glist, cov_std = cov_std, lv_var = lv_var
  )
}

lav_standardize_all_x <- function(x, lavobject, partable = NULL, cov_std = TRUE,
                                  ov_std = NULL, nox = FALSE, rotation = FALSE) {
  ge <- lav_standardize_prepare_x(x, lavobject, rotation = rotation)
  lav_standardize_all(
    lavobject = lavobject, partable = partable, est = ge$est,
    est_std = NULL, glist = ge$glist, cov_std = cov_std,
    ov_std = ov_std, nox = nox
  )
}

lav_standardize_all_nox_x <- function(x, lavobject, partable = NULL,
                                      cov_std = TRUE, rotation = FALSE) {
  lav_standardize_all_x(
    x = x, lavobject = lavobject, partable = partable, cov_std = cov_std,
    nox = TRUE, rotation = rotation
  )
}

# ---------------------------------------------------------------------------
# std.lv : standardize the latent variables only
# ---------------------------------------------------------------------------

lav_standardize_lv <- function(lavobject = NULL,
                               partable = NULL, est = NULL, glist = NULL,
                               cov_std = TRUE, lv_var = NULL,
                               lavmodel = NULL, lavpartable = NULL) {
  s <- lav_standardize_setup(
    lavobject = lavobject, lavmodel = lavmodel, lavpartable = lavpartable,
    partable = partable, est = est, glist = glist
  )
  lavmodel <- s$lavmodel
  lavpartable <- s$lavpartable
  partable <- s$partable
  est <- s$est
  glist <- s$glist

  out <- est
  n <- length(est)
  stopifnot(n == length(partable$lhs))

  # working copy of est: for interaction terms (e.g., X:Z), the (co)variance
  # and mean rows are 'centered' below (the mean-driven part of the product
  # moments is removed) before any scaling takes place. The cov_std block
  # further down needs these centered -- but still unscaled -- variances to
  # compute its sqrt(variance) divisors, so we keep them in est_c.
  est_c <- est

  # any latent product terms (in any block)? only their corrections (see
  # lav_standardize_lv_int) need E(ETA), psi and the mm indices. Regular
  # models skip these, and E(ETA) is LISREL-only anyway (RAM would stop)
  lv_int_any <- length(lav_pt_vnames(lavpartable, "lv.interaction")) > 0L
  if (lv_int_any) {
    mm_idx <- lav_model_group_mm_indices(lavmodel@nmat)
  }

  # compute ETA
  if (is.null(lv_var)) {
    lv_eta <- lav_model_veta(
      lavmodel = lavmodel,
      glist = glist
    )

    if (lv_int_any) {
      lv_eeta <- lav_model_eeta(
        lavmodel = lavmodel,
        glist = glist,
        lavsamplestats = lavobject@SampleStats
      )
    }
  }

  for (g in 1:lavmodel@nblocks) {
    ov_names <- lav_pt_vnames(lavpartable, "ov", block = g) # not user,
    # which may be incomplete
    lv_names <- lav_pt_vnames(lavpartable, "lv", block = g)

    # shortcut: no latents in this block, nothing to do
    if (length(lv_names) == 0L) {
      next
    }

    if (is.null(lv_var)) {
      veta_g <- lv_eta[[g]] # full model-implied V(ETA), only kept as a
                            # psi fallback for the centering below (RAM)
      eta2 <- diag(veta_g)
    } else {
      veta_g <- NULL
      eta2 <- lv_var[[g]]
    }
    # change negative values to NA
    eta2[eta2 < 0] <- as.numeric(NA)
    eta <- sqrt(eta2)

    # two-way latent product terms (X:Z, X:X): shift the simple main
    # effects, scale by SD(A)*SD(B), and center the product moments (see
    # lav_standardize_lv_int() for the rationale)
    if (lv_int_any) {
      lv_int_names <- lav_pt_vnames(lavpartable, "lv.interaction",
        block = g
      )
    } else {
      lv_int_names <- character(0L)
    }
    if (length(lv_int_names) > 0L) {
      if (is.null(lv_var)) {
        eeta <- lv_eeta[[g]]
      } else {
        eeta <- numeric(length(eta2))
      }
      psi_g <- glist[mm_idx[[g]]]$psi
      if (is.null(psi_g)) { # e.g., RAM representation
        psi_g <- veta_g
      }
      tmp <- lav_standardize_lv_int(
        out = out, est_c = est_c, eta = eta, est = est, partable = partable,
        lv_names = lv_names, lv_int_names = lv_int_names, eeta = eeta,
        psi_g = psi_g, g = g
      )
      out <- tmp$out
      est_c <- tmp$est_c
      eta <- tmp$eta
    }

    # 1a. "=~" regular indicators
    idx <- which(partable$op == "=~" & !(partable$rhs %in% lv_names) &
      partable$block == g)
    out[idx] <- out[idx] * eta[match(partable$lhs[idx], lv_names)]

    # 1b. "=~" regular higher-order lv indicators
    idx <- which(partable$op == "=~" & !(partable$rhs %in% ov_names) &
      partable$block == g)
    out[idx] <- (out[idx] * eta[match(partable$lhs[idx], lv_names)]
      / eta[match(partable$rhs[idx], lv_names)])

    # 2. "~" regressions (and "<~")
    idx <- which((partable$op == "~" | partable$op == "<~") &
      partable$lhs %in% lv_names &
      partable$block == g)
    out[idx] <- out[idx] / eta[match(partable$lhs[idx], lv_names)]

    idx <- which((partable$op == "~" | partable$op == "<~") &
      partable$rhs %in% lv_names &
      partable$block == g)
    out[idx] <- out[idx] * eta[match(partable$rhs[idx], lv_names)]

    # 3b. "~~" lv variances
    # ATTENTION: in Mplus 4.1, the off-diagonal residual covariances
    #            were computed by the formula cov(i,j) / sqrt(i.var*j.var)
    #            where i.var and j.var were diagonal elements of ETA
    #
    #            in Mplus 6.1 (but also AMOS and EQS), the i.var and j.var
    #            elements are the 'PSI' diagonal elements!!
    rv_idx <- which(partable$op == "~~" & partable$rhs %in% lv_names &
      partable$lhs == partable$rhs &
      partable$block == g)
    out[rv_idx] <- (out[rv_idx] / eta[match(partable$lhs[rv_idx], lv_names)]
      / eta[match(partable$rhs[rv_idx], lv_names)])

    # covariances lv
    # three types:
    # - only lhs is LV (and fixed.x = FALSE)
    # - only rhs is LV (and fixed.x = FALSE)
    # - both lhs and rhs are LV (regular case)
    if (cov_std) {
      # est_c: est with the interaction-term variances centered (see above)
      if (!is.complex(est_c[rv_idx])) {
        rv <- sqrt(abs(est_c[rv_idx])) # abs in case of heywood cases
      } else {
        rv <- sqrt(est_c[rv_idx])
      }
      rv_names <- partable$lhs[rv_idx]
    }

    # left
    idx_lhs <- which(partable$op == "~~" &
      partable$lhs %in% lv_names &
      partable$lhs != partable$rhs &
      partable$block == g)
    if (length(idx_lhs) > 0L) {
      if (cov_std == FALSE) {
        out[idx_lhs] <-
          (out[idx_lhs] / eta[match(partable$lhs[idx_lhs], lv_names)])
      } else {
        out[idx_lhs] <-
          (out[idx_lhs] / rv[match(partable$lhs[idx_lhs], rv_names)])
      }
    }

    # right
    idx_rhs <- which(partable$op == "~~" &
      partable$rhs %in% lv_names &
      partable$lhs != partable$rhs &
      partable$block == g)
    if (length(idx_rhs) > 0L) {
      if (cov_std == FALSE) {
        out[idx_rhs] <-
          (out[idx_rhs] / eta[match(partable$rhs[idx_rhs], lv_names)])
      } else {
        out[idx_rhs] <-
          (out[idx_rhs] / rv[match(partable$rhs[idx_rhs], rv_names)])
      }
    }

    # 4b. "~1" lv
    idx <- which(partable$op == "~1" & partable$lhs %in% lv_names &
      partable$block == g)
    out[idx] <- out[idx] / eta[match(partable$lhs[idx], lv_names)]
  }

  # 5. defined/(in)equality constraints
  lav_standardize_constraints(out, partable, lavmodel)
}

# ---------------------------------------------------------------------------
# std.all / std.nox / user subset : standardize observed (and latent)
# variables.
#
# 'ov_std' and 'nox' control *which* observed variables are standardized:
#   - nox = FALSE, ov_std = NULL      -> all observed variables (std.all)
#   - nox = TRUE                      -> all but the exogenous 'x' (std.nox)
#   - nox = FALSE, ov_std = c(...)    -> only the observed variables listed in
#                                        'ov_std' (a generalization of std.nox)
# In all cases the latent variables are fully standardized (via
# lav_standardize_lv()).
# ---------------------------------------------------------------------------

lav_standardize_all <- function(lavobject = NULL,
                                partable = NULL, est = NULL, est_std = NULL,
                                glist = NULL, cov_std = TRUE, ov_var = NULL,
                                lv_var = NULL,
                                lavmodel = NULL, lavpartable = NULL,
                                cov_x = NULL, ov_std = NULL, nox = FALSE) {
  s <- lav_standardize_setup(
    lavobject = lavobject, lavmodel = lavmodel, lavpartable = lavpartable,
    partable = partable, est = est, glist = glist, cov_x = cov_x,
    need_cov_x = TRUE
  )
  lavmodel <- s$lavmodel
  lavpartable <- s$lavpartable
  partable <- s$partable
  est <- s$est
  glist <- s$glist
  cov_x <- s$cov_x

  if (is.null(est_std)) {
    est_std <- lav_standardize_lv(
      lavobject = lavobject,
      partable = partable, est = est, glist = glist,
      cov_std = cov_std, lv_var = lv_var, lavmodel = lavmodel,
      lavpartable = lavpartable
    )
  }

  out <- est_std
  n <- length(est_std)
  stopifnot(n == length(partable$lhs))

  vy <- lav_model_vy(
    lavmodel = lavmodel, glist = glist,
    diagonal_only = TRUE
  )

  for (g in 1:lavmodel@nblocks) {
    lv_names <- lav_pt_vnames(lavpartable, "lv", block = g)

    tmp <- lav_standardize_ov_sd(
      lavpartable = lavpartable, g = g, vy = vy, ov_var = ov_var,
      conditional_x = lavmodel@conditional.x, cov_x = cov_x
    )
    ov_names <- tmp$ov_names
    ov <- tmp$ov
    ov2 <- tmp$ov2

    # which observed variables should NOT be standardized in this block?
    # (std.all: none; std.nox: the exogenous 'x'; user: all but 'ov_std')
    if (nox) {
      if (lavmodel@fixed.x) {
        ov_nostd <- lav_pt_vnames(lavpartable, "ov.x", block = g)
      } else {
        # if fixed.x = FALSE, the exogenous 'x' are treated as random
        # variables like any other, and std.nox reduces to std.all
        ov_nostd <- character(0L)
      }
    } else if (is.null(ov_std)) {
      ov_nostd <- character(0L)
    } else {
      ov_nostd <- setdiff(ov_names, ov_std)
    }

    # 1a. "=~" regular indicators
    idx <- which(partable$op == "=~" & !(partable$rhs %in% lv_names) &
      !(partable$rhs %in% ov_nostd) &
      partable$block == g)
    out[idx] <- out[idx] / ov[match(partable$rhs[idx], ov_names)]

    # 1b. "=~" regular higher-order lv indicators -> handled in std.lv
    # 1c. "=~" indicators that are both in ov and lv -> not standardized here

    # 2. "~" regressions (and "<~")
    idx <- which((partable$op == "~" | partable$op == "<~") &
      partable$lhs %in% ov_names & !(partable$lhs %in% ov_nostd) &
      partable$block == g)
    out[idx] <- out[idx] / ov[match(partable$lhs[idx], ov_names)]

    idx <- which((partable$op == "~" | partable$op == "<~") &
      partable$rhs %in% ov_names & !(partable$rhs %in% ov_nostd) &
      partable$block == g)
    out[idx] <- out[idx] * ov[match(partable$rhs[idx], ov_names)]

    # 3a. "~~" ov
    # ATTENTION: in Mplus 4.1, the off-diagonal residual covariances
    #            were computed by the formula cov(i,j) / sqrt(i.var*j.var)
    #            where i.var and j.var were diagonal elements of OV
    #
    #            in Mplus 6.1 (but also AMOS and EQS), the i.var and j.var
    #            elements are the 'THETA' diagonal elements!!

    # variances
    rv_idx <- which(partable$op == "~~" & !(partable$lhs %in% lv_names) &
      !(partable$lhs %in% ov_nostd) &
      partable$lhs == partable$rhs &
      partable$block == g)
    out[rv_idx] <- (out[rv_idx] /
      ov2[match(partable$lhs[rv_idx], ov_names)])

    # covariances ov
    # three types:
    # - only lhs is OV (and fixed.x = FALSE)
    # - only rhs is OV (and fixed.x = FALSE)
    # - both lhs and rhs are OV (regular case)
    if (cov_std) {
      if (!is.complex(est[rv_idx])) {
        rv <- sqrt(abs(est[rv_idx]))
      } else {
        rv <- sqrt(est[rv_idx])
      }
      rv_names <- partable$lhs[rv_idx]
    }

    # left
    idx_lhs <- which(partable$op == "~~" &
      !(partable$lhs %in% lv_names) & !(partable$lhs %in% ov_nostd) &
      partable$lhs != partable$rhs &
      partable$block == g)
    if (length(idx_lhs) > 0L) {
      if (cov_std == FALSE) {
        out[idx_lhs] <-
          (out[idx_lhs] / ov[match(partable$lhs[idx_lhs], ov_names)])
      } else {
        out[idx_lhs] <-
          (out[idx_lhs] / rv[match(partable$lhs[idx_lhs], rv_names)])
      }
    }

    # right
    idx_rhs <- which(partable$op == "~~" &
      !(partable$rhs %in% lv_names) & !(partable$rhs %in% ov_nostd) &
      partable$lhs != partable$rhs &
      partable$block == g)
    if (length(idx_rhs) > 0L) {
      if (cov_std == FALSE) {
        out[idx_rhs] <-
          (out[idx_rhs] / ov[match(partable$rhs[idx_rhs], ov_names)])
      } else {
        out[idx_rhs] <-
          (out[idx_rhs] / rv[match(partable$rhs[idx_rhs], rv_names)])
      }
    }

    # 3b. "~~" lv -> handled in std.lv

    # 4a. "~1" ov
    idx <- which(partable$op == "~1" & !(partable$lhs %in% lv_names) &
      !(partable$lhs %in% ov_nostd) &
      partable$block == g)
    out[idx] <- out[idx] / ov[match(partable$lhs[idx], ov_names)]

    # 4b. "~1" lv -> handled in std.lv

    # 4c. "|" thresholds
    idx <- which(partable$op == "|" & !(partable$lhs %in% lv_names) &
      !(partable$lhs %in% ov_nostd) &
      partable$block == g)
    out[idx] <- out[idx] / ov[match(partable$lhs[idx], ov_names)]

    # 4d. "~*~" scales
    idx <- which(partable$op == "~*~" & !(partable$lhs %in% lv_names) &
      !(partable$lhs %in% ov_nostd) &
      partable$block == g)
    out[idx] <- 1.0
  }

  # 5. defined/(in)equality constraints
  lav_standardize_constraints(out, partable, lavmodel)
}

# std.nox: standardize all observed variables except the exogenous 'x'.
# Thin wrapper around lav_standardize_all().
lav_standardize_all_nox <- function(lavobject = NULL,
                                    partable = NULL, est = NULL, est_std = NULL,
                                    glist = NULL, cov_std = TRUE, ov_var = NULL,
                                    lv_var = NULL,
                                    lavmodel = NULL, lavpartable = NULL,
                                    cov_x = NULL) {
  lav_standardize_all(
    lavobject = lavobject, partable = partable, est = est, est_std = est_std,
    glist = glist, cov_std = cov_std, ov_var = ov_var, lv_var = lv_var,
    lavmodel = lavmodel, lavpartable = lavpartable, cov_x = cov_x,
    nox = TRUE
  )
}

# ---------------------------------------------------------------------------
# unstandardize observed variables (used by e.g. simulateData)
# ---------------------------------------------------------------------------

lav_unstandardize_ov <- function(partable, ov_var = NULL, cov_std = TRUE) {
  # check if ustart is missing; if so, look for est
  if (is.null(partable$ustart)) {
    partable$ustart <- partable$est
  }

  # check if block is missing
  if (is.null(partable$block)) {
    partable$block <- rep(1L, length(partable$ustart))
  }

  stopifnot(!any(is.na(partable$ustart)))
  est <- out <- partable$ustart

  # nblocks
  nblocks <- lav_pt_nblocks(partable)

  # if ov.var is NOT a list, make a list
  if (!is.list(ov_var)) {
    tmp <- ov_var
    ov_var <- vector("list", length = nblocks)
    ov_var[1:nblocks] <- list(tmp)
  }

  for (g in 1:nblocks) {
    ov_names <- lav_pt_vnames(partable, "ov", block = g) # not user
    lv_names <- lav_pt_vnames(partable, "lv", block = g)

    ov <- sqrt(ov_var[[g]])

    # 1a. "=~" regular indicators
    idx <- which(partable$op == "=~" & !(partable$rhs %in% lv_names) &
      partable$block == g)
    out[idx] <- out[idx] * ov[match(partable$rhs[idx], ov_names)]

    # 1b. "=~" regular higher-order lv indicators
    # 1c. "=~" indicators that are both in ov and lv

    # 2. "~" regressions (and "<~")
    idx <- which((partable$op == "~" | partable$op == "<~") &
      partable$lhs %in% ov_names &
      partable$block == g)
    out[idx] <- out[idx] * ov[match(partable$lhs[idx], ov_names)]

    idx <- which((partable$op == "~" | partable$op == "<~") &
      partable$rhs %in% ov_names &
      partable$block == g)
    out[idx] <- out[idx] / ov[match(partable$rhs[idx], ov_names)]

    # 3a. "~~" ov
    # ATTENTION: in Mplus 4.1, the off-diagonal residual covariances
    #            were computed by the formula cov(i,j) / sqrt(i.var*j.var)
    #            where i.var and j.var were diagonal elements of OV
    #
    #            in Mplus 6.1 (but also AMOS and EQS), the i.var and j.var
    #            elements are the 'THETA' diagonal elements!!

    # variances
    rv_idx <- which(partable$op == "~~" & !(partable$lhs %in% lv_names) &
      partable$lhs == partable$rhs &
      partable$block == g)
    out[rv_idx] <- (out[rv_idx] * ov[match(partable$lhs[rv_idx], ov_names)]
      * ov[match(partable$rhs[rv_idx], ov_names)])

    # covariances
    idx <- which(partable$op == "~~" & !(partable$lhs %in% lv_names) &
      partable$lhs != partable$rhs &
      partable$block == g)
    if (length(idx) > 0L) {
      if (cov_std == FALSE) {
        out[idx] <- (out[idx] * ov[match(partable$lhs[idx], ov_names)]
          * ov[match(partable$rhs[idx], ov_names)])
      } else {
        if (!is.complex(out[rv_idx])) {
          rv <- sqrt(abs(out[rv_idx]))
        } else {
          rv <- sqrt(out[rv_idx])
        }
        rv_names <- partable$lhs[rv_idx]
        out[idx] <- (out[idx] * rv[match(partable$lhs[idx], rv_names)]
          * rv[match(partable$rhs[idx], rv_names)])
      }
    }

    # 3b. "~~" lv

    # 4a. "~1" ov
    idx <- which(partable$op == "~1" & !(partable$lhs %in% lv_names) &
      partable$block == g)
    out[idx] <- out[idx] * ov[match(partable$lhs[idx], ov_names)]

    # 4b. "~1" lv
  }

  # 5a ":="
  # 5b "=="
  # 5c. "<" or ">"

  out
}

# functions related to the SRMR fit measures (single level only)

# lower-level functions:
# - lav_fit_srmr_mplus
# - lav_fit_srmr_twolevel

# higher-level functions:
# - lav_fit_srmr_lavobject

# Y.R. 22 July 2022

# Note: for rmrm/srmr/crmr, we use lav_residuals_summmary()

# SRMR for continuous data only
# see https://www.statmodel.com/download/SRMR.pdf
lav_fit_srmr_mplus <- function(lavobject) {

  lavsamplestats <- lavobject@SampleStats
  lavh1 <- lavobject@h1

  # ngroups
  g_1 <- lavobject@Data@ngroups

  # container per group
  srmr_mplus_group <- numeric(g_1)
  srmr_mplus_nomean_group <- numeric(g_1)

  # If you change how any of the observed/estimated moments below are retrieved,
  # please tag @TDJorgensen at the end of the commit message.
  for (g in 1:g_1) {
    # observed
    if (!lavsamplestats@missing.flag) {
      if (lavobject@Model@conditional.x) {
        s <- lavsamplestats@res.cov[[g]]
        m <- lavsamplestats@res.int[[g]]
      } else {
        s <- lavsamplestats@cov[[g]]
        m <- lavsamplestats@mean[[g]]
      }
    } else {
      # EM estimates
      # (note: exact [[ indexing -- under conditional.x, $cov/$mean would
      #  partially match the cov.x/mean.x elements of the implied list)
      if (lavobject@Model@conditional.x) {
        if (!is.null(lavh1$implied[["res.cov"]][[g]])) {
          s <- lavh1$implied[["res.cov"]][[g]]
          m <- lavh1$implied[["res.int"]][[g]]
        } else {
          s <- lavsamplestats@missing.h1[[g]]$res.cov
          m <- lavsamplestats@missing.h1[[g]]$res.int
        }
      } else {
        if (!is.null(lavh1$implied[["cov"]][[g]])) {
          s <- lavh1$implied[["cov"]][[g]]
        } else {
          s <- lavsamplestats@missing.h1[[g]]$sigma
        }
        if (!is.null(lavh1$implied[["mean"]][[g]])) {
          m <- lavh1$implied[["mean"]][[g]]
        } else {
          m <- lavsamplestats@missing.h1[[g]]$mu
        }
      }
    }
    nvar <- ncol(s)

    # estimated
    implied <- lavobject@implied
    lavmodel <- lavobject@Model
    sigma_hat <- if (lavmodel@conditional.x) {
      implied$res.cov[[g]]
    } else {
      implied$cov[[g]]
    }
    mu_hat <- if (lavmodel@conditional.x) {
      implied$res.int[[g]]
    } else {
      implied$mean[[g]]
    }

    # Bollen approach: simply using cov2cor ('correlation residuals')
    s_cor <- cov2cor(s)
    sigma_cor <- cov2cor(sigma_hat)
    r_cor <- (s_cor - sigma_cor)

    # meanstructure
    if (lavobject@Model@meanstructure) {
      # standardized residual mean vector
      r_cor_mean <- m / sqrt(diag(s)) - mu_hat / sqrt(diag(sigma_hat))

      e <- nvar * (nvar + 1) / 2 + nvar
      srmr_mplus_group[g] <-
        sqrt((sum(r_cor[lower.tri(r_cor, diag = FALSE)]^2) +
          sum(r_cor_mean^2) +
          sum(((diag(s) - diag(sigma_hat)) / diag(s))^2)) / e)

      e <- nvar * (nvar + 1) / 2
      srmr_mplus_nomean_group[g] <-
        sqrt((sum(r_cor[lower.tri(r_cor, diag = FALSE)]^2) +
          sum(((diag(s) - diag(sigma_hat)) / diag(s))^2)) / e)
    } else {
      e <- nvar * (nvar + 1) / 2
      srmr_mplus_nomean_group[g] <- srmr_mplus_group[g] <-
        sqrt((sum(r_cor[lower.tri(r_cor, diag = FALSE)]^2) +
          sum(((diag(s) - diag(sigma_hat)) / diag(s))^2)) / e)
    }
  } # G

  attr(srmr_mplus_group, "nomean") <- srmr_mplus_nomean_group
  srmr_mplus_group
}

lav_fit_srmr_twolevel <- function(lavobject = NULL) {
  nlevels <- lavobject@Data@nlevels
  g_1 <- lavobject@Data@ngroups

  srmr_within <- numeric(g_1)
  srmr_between <- numeric(g_1)
  for (g in 1:g_1) {
    b_within <- (g - 1L) * nlevels + 1L
    b_between <- (g - 1L) * nlevels + 2L

    # OBSERVED        # if these change, tag @TDJorgensen in commit message
    h1_implied <- lavobject@h1$implied
    if (!is.null(h1_implied$res.cov) &&
        !is.null(h1_implied$res.cov[[b_within]])) {
      # conditional.x with residual (y-only) unrestricted moments
      # (two-level (D)WLS with covariates): compare the residual
      # covariance matrices directly
      s_within <- h1_implied$res.cov[[b_within]]
      s_between <- h1_implied$res.cov[[b_between]]

      # ESTIMATED     # if these change, tag @TDJorgensen in commit message
      sigma_within <- lavobject@implied$res.cov[[b_within]]
      sigma_between <- lavobject@implied$res.cov[[b_between]]
    } else {
      s_within <- h1_implied$cov[[b_within]]
      # m_within <- h1_implied$mean[[b_within]]
      s_between <- h1_implied$cov[[b_between]]
      # m_between <- h1_implied$mean[[b_between]]

      # ESTIMATED     # if these change, tag @TDJorgensen in commit message
      implied <- lav_model_implied_cond2uncond(lavobject@implied)
      sigma_within <- implied$cov[[b_within]]
      # mu_within <- implied$mean[[b_within]]
      sigma_between <- implied$cov[[b_between]]
      # mu_between <- implied$mean[[b_between]]
    }

    # force pd for between
    #    S.between <- lav_mat_sym_force_pd(S.between)
    sigma_between <- lav_mat_sym_force_pd(sigma_between)

    # Bollen approach: simply using cov2cor ('residual correlations')
    s_within_cor <- cov2cor(s_within)
    s_between_cor <- cov2cor(s_between)
    sigma_within_cor <- cov2cor(sigma_within)
    if (all(diag(sigma_between) > 0)) {
      sigma_between_cor <- cov2cor(sigma_between)
    } else {
      sigma_between_cor <- matrix(as.numeric(NA),
        nrow = nrow(sigma_between),
        ncol = ncol(sigma_between)
      )
    }
    r_within_cor <- (s_within_cor - sigma_within_cor)
    r_between_cor <- (s_between_cor - sigma_between_cor)

    nvar_within <- NCOL(s_within)
    nvar_between <- NCOL(s_between)
    pstar_within <- nvar_within * (nvar_within + 1) / 2
    pstar_between <- nvar_between * (nvar_between + 1) / 2

    # SRMR
    srmr_within[g] <- sqrt(sum(lav_mat_vech(r_within_cor)^2) /
      pstar_within)
    srmr_between[g] <- sqrt(sum(lav_mat_vech(r_between_cor)^2) /
      pstar_between)
  }

  # adjust for group sizes
  ng <- unlist(lavobject@SampleStats@nobs)
              # if this changes, tag @TDJorgensen in commit message
  ntotal <- lavobject@SampleStats@ntotal
              # if this changes, tag @TDJorgensen in commit message
  srmr_within_1 <- sum(ng / ntotal * srmr_within)
  srmr_between_1 <- sum(ng / ntotal * srmr_between)
  srmr_total <- srmr_within_1 + srmr_between_1

  c(srmr_total, srmr_within_1, srmr_between_1)
}

lav_fit_srmr_lavobject <- function(lavobject = NULL, fit_measures = "rmsea") {
  # check lavobject
  stopifnot(inherits(lavobject, "lavaan"))

  # categorical?
  categorical <- lavobject@Model@categorical

  # supported fit measures in this function
  if (categorical) {
    fit_srmr <- c("srmr")
    fit_srmr2 <- c(
      "rmr", "rmr_nomean",
      "srmr", # per default equal to srmr_bentler_nomean
      "srmr_bentler", "srmr_bentler_nomean",
      "crmr", "crmr_nomean",
      "srmr_mplus", "srmr_mplus_nomean"
    )
  } else {
    if (lavobject@Data@nlevels > 1L) {
      fit_srmr <- c("srmr", "srmr_within", "srmr_between")
      fit_srmr2 <- c("srmr", "srmr_within", "srmr_between")
    } else {
      fit_srmr <- c("srmr")
      fit_srmr2 <- c(
        "rmr", "rmr_nomean",
        "srmr", # the default
        "srmr_bentler", "srmr_bentler_nomean",
        "crmr", "crmr_nomean",
        "srmr_mplus", "srmr_mplus_nomean"
      )
    }
  }

  # which one do we need?
  if (missing(fit_measures)) {
    # default set
    fit_measures <- fit_srmr
  } else {
    # remove any not-SRMR related index from fit.measures
    rm_idx <- which(!fit_measures %in% fit_srmr2)
    if (length(rm_idx) > 0L) {
      fit_measures <- fit_measures[-rm_idx]
    }
    if (length(fit_measures) == 0L) {
      return(list())
    }
  }

  # output container
  indices <- list()

  # 1. single level
  if (lavobject@Data@nlevels == 1L) {
    # RMR/SRMR/CRMR: we get it from lav_residuals_summary()
    out <- lav_residuals_summary(lavobject, se = FALSE, unbiased = FALSE)

    # name of the covariance-residual column in the summary tables
    cov_cor <- "cov"
    if (categorical) {
      # conditional.x categorical summary uses res.cov / res.th / res.slopes
      cov_cor <- if (lavobject@Model@conditional.x) "res.cov" else "cor"
    } else if (lavobject@Model@conditional.x) {
      # continuous conditional.x summary uses res.cov / res.int / res.slopes
      cov_cor <- "res.cov"
    }

    # only cov
    rmr_nomean_group <- sapply(lapply(out, "[[", "rmr"), "[[", cov_cor)
    srmr_nomean_group <- sapply(lapply(out, "[[", "srmr"), "[[", cov_cor)
    crmr_nomean_group <- sapply(lapply(out, "[[", "crmr"), "[[", cov_cor)

    # total
    if (lavobject@Model@meanstructure) {
      rmr_group <- sapply(lapply(out, "[[", "rmr"), "[[", "total")
      srmr_group <- sapply(lapply(out, "[[", "srmr"), "[[", "total")
      crmr_group <- sapply(lapply(out, "[[", "crmr"), "[[", "total")
    } else {
      # no 'total', only 'cov'
      rmr_group <- rmr_nomean_group
      srmr_group <- srmr_nomean_group
      crmr_group <- crmr_nomean_group
    }

    # the Mplus versions
    srmr_mplus_group <- lav_fit_srmr_mplus(lavobject = lavobject)
    srmr_mplus_nomean_group <- attr(srmr_mplus_group, "nomean")
    attr(srmr_mplus_group, "nomean") <- NULL

    # adjust for group sizes
    ng <- unlist(lavobject@SampleStats@nobs)
              # if this changes, tag @TDJorgensen in commit message
    ntotal <- lavobject@SampleStats@ntotal
              # if this changes, tag @TDJorgensen in commit message
    rmr <- sum(ng / ntotal * rmr_group)
    rmr_nomean <- sum(ng / ntotal * rmr_nomean_group)
    srmr_bentler <- sum(ng / ntotal * srmr_group)
    srmr_bentler_nomean <- sum(ng / ntotal * srmr_nomean_group)
    crmr <- sum(ng / ntotal * crmr_group)
    crmr_nomean <- sum(ng / ntotal * crmr_nomean_group)
    srmr_mplus <- sum(ng / ntotal * srmr_mplus_group)
    srmr_mplus_nomean <- sum(ng / ntotal * srmr_mplus_nomean_group)

    # srmr
    if (lavobject@Options$mimic %in% c("lavaan", "EQS")) {
      if (categorical) {
        indices["srmr"] <- srmr_bentler_nomean
      } else {
        indices["srmr"] <- srmr_bentler
      }
    } else if (lavobject@Options$mimic == "Mplus") {
      if (lavobject@Options$information[1] == "expected") {
        if (categorical) {
          indices["srmr"] <- srmr_bentler_nomean
        } else {
          indices["srmr"] <- srmr_bentler
        }
      } else {
        if (categorical) {
          indices["srmr"] <- srmr_mplus_nomean
        } else {
          indices["srmr"] <- srmr_mplus
        }
      }
    } # Mplus only

    # the others
    indices["srmr_bentler"] <- srmr_bentler
    indices["srmr_bentler_nomean"] <- srmr_bentler_nomean
    indices["crmr"] <- crmr
    indices["crmr_nomean"] <- crmr_nomean

    # only correct for non-categorical:
    if (lavobject@Model@categorical) {
      # FIXME! Compute Mplus 8.1 way to compute SRMR in the
      #        categorical setting
      #        See 'SRMR in Mplus (2018)' document on Mplus website
      indices["srmr_mplus"] <- as.numeric(NA)
      indices["srmr_mplus_nomean"] <- as.numeric(NA)
    } else {
      indices["srmr_mplus"] <- srmr_mplus
      indices["srmr_mplus_nomean"] <- srmr_mplus_nomean
    }
    indices["rmr"] <- rmr
    indices["rmr_nomean"] <- rmr_nomean
  } else {
    # 2. twolevel setting
    out <- lav_fit_srmr_twolevel(lavobject = lavobject)
    indices["srmr"] <- out[1]
    indices["srmr_within"] <- out[2]
    indices["srmr_between"] <- out[3]
  } # twolevel

  # return only those that were requested
  indices[fit_measures]
}

# ---------------------lav_parse_options ------------------------------------- #
# function to configurate settings which conduct the parser function           #
# can be used to add operators, modifiers and grouping classes                 #
# the function returns the status of the parser config                         #
# ---------------------------------------------------------------------------- #
lav_parse_options <- function(
    pkgname   = NULL, # must be given if any of the following arguments not NULL
    operators = NULL, # char vector
    modifiers = NULL, # data.frame(mod = ., side = ., expect = .)
    groupings  = NULL # char vector
) {
  stopifnot(is.null(pkgname) || (is.character(pkgname) && pkgname != "lavaan"))
  stopifnot(is.null(operators) ||
    (is.character(operators) && !is.null(pkgname)))
  stopifnot(is.null(modifiers) ||
    (is.data.frame(modifiers) && !is.null(pkgname)))
  stopifnot(is.null(groupings) ||
    (is.character(groupings) && !is.null(pkgname)))
  config <- get0("mdl_config", lavaan_cache_env)
  if (is.null(config)) {
    ops = data.frame(
      op = c("=~", "<~", "~*~", "~~", "~", "|~", "==",
      "<", ">", ":=", ":", "|", "%", ":~"),
      pkg = "lavaan"
    )
    mods = data.frame(
      mod = c("efa", "fixed", "label", "start",
      "lower", "upper", "rv", "equal"),
      side = c("l", "r", "r", "r",
        "r", "r", "r", "r"),
      expect = c("char", "num", "char", "num",
        "num", "num", "char", "char"),
      pkg = "lavaan"
    )
    groups = data.frame(
      group = c("group", "level", "class", "block"),
      pkg = "lavaan"
    )
    config <- list(operators = ops,
      modifiers = mods, groupings = groups)
    assign("mdl_config", config, envir = lavaan_cache_env)
  }
  if (is.null(pkgname)) return(config)
  if (!exists("ops", environment())) {
    ops    <- config$operators
    mods   <- config$modifiers
    groups <- config$groupings
  }
  # remove existing values
  ops <- ops[ops$pkg != pkgname, ]
  mods <- mods[mods$pkg != pkgname, ]
  groups <- groups[groups$pkg != pkgname, ]
  # add supplied values
  if (!is.null(operators)) {
    if (any(!grepl("^%.+%$", operators))) {
      lav_msg_stop(gettext("Operators must have the format '%.+%'."))
    }
    if (any(operators %in% ops$op)) {
      lav_msg_stop(gettext("Some operators already defined."))
    }
    ops <- rbind(ops,
    data.frame(
      op = operators,
      pkg = pkgname
    ))
  }
  if (!is.null(modifiers)) {
    if (is.null(modifiers$mod)) {
      lav_msg_stop(gettext("Modifiers must have a column 'mod'."))
    }
    if (any(!grepl("^[a-z_]+$", modifiers$mod))) {
      lav_msg_stop(gettext("Modifiers$mod must have the format '[a-z_]+'."))
    }
    if (is.null(modifiers$side)) {
      lav_msg_stop(gettext("Modifiers must have a column 'side'."))
    }
    if (any(!grepl("^l$|^r$", modifiers$side))) {
      lav_msg_stop(gettext("Modifiers$side must contain 'l' or 'r'."))
    }
    if (is.null(modifiers$expect)) {
      lav_msg_stop(gettext("Modifiers must have a column 'expect'."))
    }
    if (any(!grepl("^char$|^num$|^raw$", modifiers$expect))) {
      lav_msg_stop(gettext(
        "Modifiers$expect must contain 'char', 'num' or 'raw'."))
    }
    forbidden <- c("lhs", "op", "rhs", "block")
    if (any(modifiers %in% forbidden)) {
      lav_msg_stop(gettextf("Modifiers cannot be %s.",
                    lav_msg_view(forbidden, "or")))
    }
    if (any(modifiers %in% mods$mod)) {
      lav_msg_stop(gettext("Some modifiers already defined."))
    }
    mods <- rbind(mods,
    data.frame(
      mod = modifiers$mod,
      side = modifiers$side,
      expect = modifiers$expect,
      pkg = pkgname
    ))
  }
  if (!is.null(groupings)) {
    if (any(!grepl("^[a-z_]+$", groupings))) {
      lav_msg_stop(gettext("groupings must have the format '[a-z_]+'."))
    }
    if (any(groupings %in% groups$group)) {
      lav_msg_stop(gettext("Some groupings already defined."))
    }
    groups <- rbind(groups,
    data.frame(
      group = groupings,
      pkg = pkgname
    ))
  }
  config <- list(operators = ops,
    modifiers = mods, groupings = groups)
  assign("mdl_config", config, envir = lavaan_cache_env)
  config
}

# ------------------------ lav_parse_tokens_open ----------------------------- #
# function to split the model source in tokens.
# Returns a list with tokens with their attributes
#   elem_pos  : position in source
#   elem_type : type of token (cf. definition of types
#               in lav_parse_model_string)
#   elem_text : the text of the token
#   elem_formule.number : sequence number of the 'logical'
#                         formula where the token occurs
# the function returns the stored tokens in a list
# ---------------------------------------------------------------------------- #
lav_parse_tokens_open <- function(modelsrc, types) {
  nmax <- nchar(modelsrc)
  elem_pos <- vector("integer", nmax)
  elem_type <- elem_pos
  elem_text <- vector("character", nmax)
  elem_i <- 1L
  modelsrcw <- paste0(modelsrc, "\n") # working model, must end
                                      # with a newline for tests via regexpr
  stringliterals <- gregexpr("\"[^\"]*?[\"\n]", modelsrcw)[[1L]]
  if (stringliterals[1L] > -1L) {
    stringliteral_lengths <- attr(stringliterals, "match.length")
    for (i in seq_along(stringliterals)) {
      pfpos <- stringliterals[i]
      pflen <- stringliteral_lengths[i]
      substr(modelsrcw, pfpos + 1L, pfpos + pflen - 2L) <-
        strrep(" ", pflen - 2L)
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos + 1L, pfpos + pflen - 2L)
      elem_type[elem_i] <- types$stringliteral
      elem_i <- elem_i + 1L
    }
  }
  comments <- gregexpr("[#!].*?\n", modelsrcw)[[1L]]
  if (comments[1] > -1L) {
    comment_lengths <- attr(comments, "match.length")
    for (i in seq_along(comments)) {
      substr(modelsrcw, comments[i], comments[i] + comment_lengths[i] - 1L) <-
        strrep(" ", comment_lengths[i] - 1L)
      # check for stringliterals in comment
      str_in_comment <- (elem_pos > comments[i] &
                           elem_pos < comments[i] + comment_lengths[i])
      if (any(str_in_comment)) {
        elem_type[str_in_comment] <- 0
      }
    }
  }
  modelsrcw <- gsub("\t", " ", modelsrcw)
  newlines <- gregexpr("[;\n]", modelsrcw)[[1L]]
  if (newlines[1L] > -1L) {
    for (i in seq_along(newlines)) {
      pfpos <- newlines[i]
      substr(modelsrcw, pfpos, pfpos) <- "\n"
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- "\n"
      elem_type[elem_i] <- types$newline
      elem_i <- elem_i + 1L
    }
  }
  # --------------------- handling spaces in operators ----------------------- #
  config <- lav_parse_options()
  for (opr in config$operators$op[nchar(config$operators$op) == 2L]) {
    oprre <- paste0("([^=~%])[", substr(opr, 1L, 1L), "]( +)[",
                                            substr(opr, 2L, 2L), "]([^=~%])")
    oprsub <- paste0("\\1", opr, "\\2\\3")
    if (grepl(oprre, modelsrcw)) {
      waar <- regexpr(oprre, modelsrcw)[1] + 1L
      modelsrcw <- gsub(oprre, oprsub, modelsrcw)
      tl <- lav_parse_txtloc(modelsrc, waar)
      lav_msg_warn(gettextf("splitting of operator '%s' deprecated", opr),
                  tl[1L],
                  footer = tl[2L]
      )
    }
  }
  # -------------------------------------------------------------------------- #
  lavopsre <- paste(
    gsub("\t", "\\\\", gsub("([.|()[{^$*+?])", "\t\\1",
             config$operators$op)), collapse = "|")
  lavops <- gregexpr(lavopsre, modelsrcw)[[1]]
  if (lavops[1L] > -1L) {
    lavop_lengths <- attr(lavops, "match.length")
    for (i in seq_along(lavops)) {
      pfpos <- lavops[i]
      pflen <- lavop_lengths[i]
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrcw, pfpos, pfpos + pflen - 1L)
      elem_type[elem_i] <- types$lavaanoperator
      substr(modelsrcw, pfpos, pfpos + pflen - 1L) <- strrep(" ", pflen)
      elem_i <- elem_i + 1L
    }
  }
  symbols <- gregexpr("[,()/*?^']", modelsrcw)[[1L]] # f1=~x2 + 0.5 ? x3
  symbols1 <- gregexpr("[-+][^.0-9]", modelsrcw)[[1L]] # f1=~x2+x3
  symbols2 <- gregexpr("[._0-9a-df-zA-DF-Z)] *[-+][.0-9]", modelsrcw)[[1L]]
  # f1=~x2+2*x3, len-2 !
  symbols3 <- gregexpr("[^.0-9][eE] *[-+][.0-9]", modelsrcw)[[1L]]
  # f1=~xe+2*x3, len-2 !
  if (symbols1[1L] > -1L) {
    if (symbols[1L] == -1L) {
      symbols <- symbols1
    } else {
      symbols <- c(symbols, symbols1)
    }
  }
  if (symbols2[1L] > -1L) {
    symbols2_lengths <- attr(symbols2, "match.length")
    symbols2 <- symbols2 + symbols2_lengths - 2L
    if (symbols[1L] == -1L) {
      symbols <- symbols2
    } else {
      symbols <- c(symbols, symbols2)
    }
  }
  if (symbols3[1L] > -1L) {
    symbols3_lengths <- attr(symbols3, "match.length")
    symbols3 <- symbols3 + symbols3_lengths - 2L
    if (symbols[1L] == -1L) {
      symbols <- symbols3
    } else {
      symbols <- c(symbols, symbols3)
    }
  }
  if (symbols[1L] > -1L) {
    for (i in seq_along(symbols)) {
      pfpos <- symbols[i]
      substr(modelsrcw, pfpos, pfpos) <- " "
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos, pfpos)
      elem_type[elem_i] <- types$symbol
      elem_i <- elem_i + 1L
    }
  }

  numliterals <- gregexpr(
    "([ \n][-+][.0-9]|[ \n]\\.[0-9]|[ \n][0-9])[-+\\.0-9eE]*",
    paste0(" ", modelsrcw)
  )[[1]]
  if (numliterals[1L] > -1L) {
    numliteral_lengths <- attr(numliterals, "match.length") - 1L
    for (i in seq_along(numliterals)) {
      pfpos <- numliterals[i]
      pflen <- numliteral_lengths[i]
      substr(modelsrcw, pfpos, pfpos + pflen - 1L) <- strrep(" ", pflen)
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos, pfpos + pflen - 1L)
      elem_type[elem_i] <- types$numliteral
      elem_i <- elem_i + 1L
    }
  }
  identifiers <- gregexpr("[ \n][_.[:alpha:]][_.[:alnum:]]*",
                          paste0(" ", modelsrcw)
  )[[1]]
  if (identifiers[1L] > -1L) {
    identifier_lengths <- attr(identifiers, "match.length") - 1L
    for (i in seq_along(identifiers)) {
      pfpos <- identifiers[i]
      pflen <- identifier_lengths[i]
      substr(modelsrcw, pfpos, pfpos + pflen - 1L) <- strrep(" ", pflen)
      elem_pos[elem_i] <- pfpos
      elem_text[elem_i] <- substr(modelsrc, pfpos, pfpos + pflen - 1L)
      elem_type[elem_i] <- types$identifier
      elem_i <- elem_i + 1L
    }
  }
  # check for uninterpreted chars
  wrong <- regexpr("[^\"\n ]", modelsrcw)
  if (wrong != -1L) {
    tl <- lav_parse_txtloc(modelsrc, wrong)
    lav_msg_stop(gettext("unexpected character"),
                 tl[1L],
                 footer = tl[2L]
    )
  }
  # remove unused elements from vectors
  elements <- which(elem_type > 0L)
  elem_pos <- elem_pos[elements]
  elem_type <- elem_type[elements]
  elem_text <- elem_text[elements]
  # order tokens
  token_order <- order(elem_pos)
  elem_pos <- elem_pos[token_order]
  elem_type <- elem_type[token_order]
  elem_text <- elem_text[token_order]

  # concatenate identifiers with only spaces in between - LDW 22/4/2024
  elem_i <- length(elem_pos)
  concatenated <- FALSE
  while (elem_i > 1L) {
    if (any(elem_type[elem_i] == c(types$identifier, types$numliteral)) &&
        elem_type[elem_i - 1L] == types$identifier) {
      spaces_between <- elem_pos[elem_i] - elem_pos[elem_i - 1L] -
        length(elem_text[elem_i - 1L])
      elem_text[elem_i - 1L] <- paste0(
        elem_text[elem_i - 1L],
        strrep(" ", spaces_between),
        elem_text[elem_i]
      )
      elem_type[elem_i] <- 0L
      concatenated <- TRUE
    }
    elem_i <- elem_i - 1L
  }
  if (concatenated) { # remove items with type 0
    elements <- which(elem_type > 0L)
    elem_pos <- elem_pos[elements]
    elem_type <- elem_type[elements]
    elem_text <- elem_text[elements]
  }

  list_before_numbering <-  list(
    elem_pos = elem_pos, elem_type = elem_type,
    elem_text = elem_text
  )

  lav_parse_formula_numbers(list_before_numbering, types)
}


# ------------------------ lav_parse_formula_numbers ------------------------- #
# help function to add formula numbers to tokens
# ---------------------------------------------------------------------------- #
lav_parse_formula_numbers <- function(list_before_numbering, types) {
  elem_pos <- list_before_numbering$elem_pos
  elem_type <- list_before_numbering$elem_type
  elem_text <- list_before_numbering$elem_text
  # to set formula number
  elem_formula_number <- rep(0L, length(elem_type))
  frm_number <- 1L
  frm_hasefa <- FALSE
  frm_lastplus <- FALSE
  frm_incremented <- FALSE
  for (i in seq_along(elem_type)) {
  elem_formula_number[i] <- frm_number
  if (elem_type[i] == types$identifier && elem_text[i] == "efa") {
    frm_hasefa <- TRUE
  }
  if (any(elem_text[i] ==
      c("+", "*", "=~", "-", "<~", "~*~", "~~", "~", "|~", "|", "%", ":~"))) {
    if (frm_incremented) {
      frm_number <- frm_number - 1L
      elem_formula_number[i] <- frm_number
      frm_incremented <- FALSE
    }
    frm_lastplus <- TRUE
  } else {
    if (any(elem_type[i] == c(
      types$stringliteral, types$identifier, types$numliteral,
      types$stringliteral, types$symbol
    ))) {
      frm_lastplus <- FALSE
    }
    if (i > 1 && elem_type[i] != types$newline &&
        elem_type[i - 1L] == types$lavaanoperator) {
      frm_hasefa <- FALSE
    }
  }
  if (elem_type[i] == types$newline) {
    if (i > 1 && elem_type[i - 1L] != types$newline) { # ignore multiple nl's
      if (!frm_hasefa && !frm_lastplus) {
        frm_number <- frm_number + 1L
        frm_incremented <- TRUE
      } else {
        frm_hasefa <- FALSE
      }
    }
  } else {
    frm_incremented <- FALSE
  }
  }
  list(
  elem_pos = elem_pos, elem_type = elem_type,
  elem_text = elem_text, elem_formula_number = elem_formula_number
  )
 }

# ------------------------ lav_parse_formulas_open ------------------------- #
# function to group the modellist tokens in 'mono' formulas.
# mono means that the terms (for formulas other then blocks and constraints)
#   are split in seperate formula's, e.g.
#   a1 + a2 =~ b1 + b2  becomes
#     /  a1 =~ b1
#     |  a1 =~ b2
#     |  a2 =~ b1
#     \  a2 =~ b2
# newlines are removed
# the function returns a list of formulas
# ---------------------------------------------------------------------------- #
lav_parse_formulas_open <- function(modellist, modelsrc, types) {
  config <- lav_parse_options()
  group_operator <- ":"
  define_operator <- ":="
  conditional_operators <- c("==", "<", ">")
  non_relational <- config$operators$op %in%
                      c(group_operator, define_operator, conditional_operators)
  relational_operators <- config$operators$op[!non_relational]
  welke <- modellist$elem_type != types$newline
  formula_numbers <- unique(modellist$elem_formula_number[welke])
  formulas <- lapply(formula_numbers, function(s) {
    welkenu <- modellist$elem_formula_number == s & welke
    list(
      elem_pos = modellist$elem_pos[welkenu],
      elem_type = modellist$elem_type[welkenu],
      elem_text = modellist$elem_text[welkenu]
    )
  })
  maxnum <- length(formula_numbers) + sum(modellist$elem_text == "+")
  outval <- vector(mode = "list", length = maxnum)
  realnum <- 0L
  for (i in seq_along(formulas)) {
    formul1 <- formulas[[i]]
    opi <- which(formul1$elem_type == types$lavaanoperator)
    nelem <- length(formul1$elem_type)
    if (length(opi) == 0L) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[1])
      lav_msg_stop(gettext("formula without valid operator"),
                   tl[1L],
                   footer = tl[2L]
      )
    }
    if (length(opi) > 1L) opi <- opi[1] # only first operator taken
    if (any(formul1$elem_text[opi] == relational_operators) &&
        sum(formul1$elem_text == "+") > 0) {
      # check + symbols outside parentheses in left and right hand side
      lhplusjes <- integer(0)
      openparentheses <- 0L
      for (jj in seq.int(1L, opi - 1L)) {
        if (formul1$elem_text[jj] == "(") {
          openparentheses <- openparentheses + 1L
          next
        }
        if (formul1$elem_text[jj] == ")") {
          openparentheses <- openparentheses - 1L
          next
        }
        if (formul1$elem_text[jj] == "+" && openparentheses == 0L) {
          lhplusjes <- c(lhplusjes, jj)
        }
      }
      lhplusjes <- c(lhplusjes, opi)
      plusjes <- integer(0)
      openparentheses <- 0L
      for (jj in seq.int(opi + 1L, nelem)) {
        if (formul1$elem_text[jj] == "(") {
          openparentheses <- openparentheses + 1L
          next
        }
        if (formul1$elem_text[jj] == ")") {
          openparentheses <- openparentheses - 1L
          next
        }
        if (formul1$elem_text[jj] == "+" && openparentheses == 0L) {
          plusjes <- c(plusjes, jj)
        }
      }
      plusjes <- c(plusjes, nelem + 1)
      # splitting lhs and rhs on '+' signs
      for (j in seq_along(lhplusjes)) {
        j0 <- 1L
        if (j > 1L) j0 <- lhplusjes[j - 1L] + 1L
        j1 <- lhplusjes[j] - 1L
        if (j1 < j0) next # skip empty parts
        for (k in seq_along(plusjes)) {
          k0 <- opi + 1L
          k1 <- plusjes[k] - 1L
          if (k > 1L) k0 <- plusjes[k - 1L] + 1L
          if (k1 < k0) next # skip empty parts
          welke <- c(seq.int(j0, j1), opi, seq.int(k0, k1))
          realnum <- realnum + 1L
          outval[[realnum]] <- lav_parse_sublist(formul1, welke)
        }
      }
    } else {
      realnum <- realnum + 1L
      outval[[realnum]] <- formul1
    }
  }
  outval[seq_len(realnum)]
}

# ------------------------ lav_parse_modifier_open --------------------------- #
# The function takes a list with tokens belonging to a single 'mono' lavaan
# formula as input. The other arguments are:
#       lhs : check for lhs or rhs modifier
#       opi : index of the lavaan operator in the list-items
#  modelsrc : the model source string (for error messages and warnings)
#     types : the types of tokens
#       rme : index of last element of modifier in formula (*)
#   rmeprev : index of first element of modifier in formula - 1L (*)
# The function returns the modifier detected as element of a list
#  with name the modifier type and value an array of values
#   (length > 1 if vector via c(...)) for the modifier value.
# The lhs or rhs is limited to the elements with index
#   rmeprev+1:rme, this is to support multiple modifiers for the same element.
# An error message is produced when no modifier can be determined.
# ---------------------------------------------------------------------------- #
lav_parse_modifier_open <- function(formul, opi, modelsrc, types,
                                    rme, rmeprev, modenv, config) {
  if (rme < opi) { # lhs modifier
    lhs <- TRUE
    welke <- c(seq.int(rmeprev + 1L, rme),
               seq.int(opi - 1L, length(formul$elem_type)))
  } else {
    lhs <- FALSE
    welke <- c(seq.int(1L, opi), seq.int(rmeprev + 1L, rme),
               length(formul$elem_type))
  }
  formul1 <- lav_parse_sublist(formul, welke)
  nelem <- length(formul1$elem_type)
  opi <- which(formul1$elem_type == types$lavaanoperator)[1L]
  # remove unnecessary parentheses (one element between parentheses, previous
  # no identifier)
  check_more <- TRUE
  while (check_more && nelem > 4L) {
    check_more <- FALSE
    for (par.i in seq.int(3L, nelem - 1L)) {
      if (formul1$elem_text[par.i - 1L] == "(" &&
          formul1$elem_text[par.i + 1L] == ")" &&
          formul1$elem_type[par.i - 2L] != types$identifier) {
        formul1$elem_type[par.i - 1L] <- 0L
        formul1$elem_type[par.i + 1L] <- 0L
        check_more <- TRUE
      }
    }
    if (check_more) {
      formul1 <- lav_parse_sublist(formul1, which(formul1$elem_type > 0))
      nelem <- length(formul1$elem_type)
    }
  }
  getmodifier <- function(s) {
    if (s %in% c("c", "t")) {
      tmp <- s
      attr(tmp, "tiepe") <- "label"
      return(tmp)
    }
    v <- all.vars(str2expression(s))
    for (v1 in v) {
      if (v1 != "pi") assign(v1, v1, modenv)
    }
    tmp <- eval(str2expression(s), modenv)
    suppressWarnings(rm(list = v, pos = modenv))
    if (is.null(attr(tmp, "tiepe"))) {
      if (is.numeric(tmp) || is.logical(tmp)) {
        attr(tmp, "tiepe") <- "fixed"
      } else {
        attr(tmp, "tiepe") <- "label"
      }
    }
    tmp
  }
  strings <- vapply(seq.int(
    if (lhs) 1L else opi + 1L,
    if (lhs) opi - 3L else nelem - 2L
    ), function(x) {
    if (formul1$elem_type[x] == types$stringliteral) {
      paste0('"', formul1$elem_text[x], '"')
    } else {
      formul1$elem_text[x]
    }
  }, "")
  txt <- paste(strings, collapse = "")
  if (!lhs && formul1$elem_text[nelem - 1L] == "?") {
    formul1$elem_text[nelem - 1L] <- "*"
    txt <- paste0("start(", txt, ")")
  }
   modifier <- tryCatch(getmodifier(txt), error = function(e) "__error__")
  if (!is.na(modifier[1L]) && modifier[1L] == "__error__") {
    tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi + 1L])
    lav_msg_stop(gettext("invalid modifier specification"),
                  tl[1L],
                  footer = tl[2L]
    )
  }
  mdtype <- attr(modifier, "tiepe")
  mdind <- which(config$modifiers$mod == mdtype)[1L]
  if (config$modifiers$expect[mdind] == "char") {
    if (!is.character(modifier)) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi + 1L])
      lav_msg_stop(gettextf("invalid %s modifier (should be character)",
        mdtype), tl[1L], footer = tl[2L])
    }
  } else if (config$modifiers$expect[mdind] == "num") {
    if (!is.numeric(modifier) && !all(is.na(modifier))) {
      tl <- lav_parse_txtloc(modelsrc, formul1$elem_pos[opi + 1L])
      lav_msg_stop(gettextf("invalid %s modifier (should be numeric)",
        mdtype), tl[1L], footer = tl[2L])
    }
  }
  modifierlist <- list(as.vector(modifier))
  names(modifierlist) <- attr(modifier, "tiepe")
  modifierlist
}

# -------------------- create environment for evaluating modifiers ----------- #
lav_parse_modenv <- function(config, side) {
  modenv <- new.env()
  for (j in which(config$modifiers$side == side)) {
    mod1 <- config$modifiers$mod[j]
    src <- paste0("assign('", mod1, "', function(...) {\n",
    if (config$modifiers$expect[j] == "raw") {
      "x <- sub('^.*?\\\\(', '', sub('\\\\)$', '', deparse(sys.call())))\n"
    } else {
      "x <- do.call('c', args = list(...))\n"
    },
    "attr(x, 'tiepe') <- '", mod1, "'\n",
    "x\n}, modenv)")
    source(exprs = str2expression(src), local = TRUE)
  }
  # equal is synonym for label
  if (side == "r") assign("equal", modenv$label, modenv)
  modenv
}

# -------------------- main parsing function --------------------------------- #
lav_parse_model_string_open <- function(model_syntax = "",
                                        as_data_frame = FALSE) {
  # replace 'strange' tildes (in some locales) (new in 0.6-6)
  modelsrc <- gsub(
    pattern = "\u02dc",
    replacement = "~",
    paste(unlist(model_syntax), "", collapse = "\n"),
    fixed = TRUE
  )
  # check for empty, whitespace-only or comment-only model syntax
  if (length(model_syntax) == 0L ||
      !grepl("[^[:space:];]", gsub("[#!][^\n]*", "", modelsrc))) {
    lav_msg_stop(gettext("Model syntax is empty."))
  }
  hashstring <- paste0("mdl_o_", lav_char2hash(paste0(modelsrc, as_data_frame)))
  if (exists(hashstring, envir = lavaan_cache_env)) {
    return(get(hashstring, envir = lavaan_cache_env))
  }
  config <- lav_parse_options()
  modenvl <- lav_parse_modenv(config, "l")
  modenvr <- lav_parse_modenv(config, "r")
  types <- lav_create_enum(c(
    "identifier", "numliteral", "stringliteral",
    "symbol", "lavaanoperator", "newline"
  ))
  modellist <- lav_parse_tokens_open(modelsrc, types)
  if (lav_debug()) {
    print(data.frame(
      pos = modellist$elem_pos,
      type = types$enum_names[modellist$elem_type],
      text = modellist$elem_text,
      formula = modellist$elem_formula_number
    ))
  }
  formulalist <- lav_parse_formulas_open(modellist, modelsrc, types)
  #---- analyse syntax formulas and put in flat_-----
  max_mono_formulas <- length(formulalist)
  flat <- list()
  flat$lhs <- character(max_mono_formulas)
  flat$op <- character(max_mono_formulas)
  flat$rhs <- character(max_mono_formulas)
  flat$mod_idx <- integer(max_mono_formulas)
  flat$block <- integer(max_mono_formulas) # keep track of groups using ":" opr
  for (mod1 in config$modifiers$mod) {
    flat[[mod1]] <- character(max_mono_formulas) # only for display purposes!
  }
  tmplist <- list(
    flat = flat,
    flat_idx = 0L,
    mod_idx = 0L,
    constraints = list(),
    modlist = list(),
    block = 1L,
    block_op = FALSE
  )
  if (lav_debug()) {
    cat("formula to analyse:\n")
  }
  for (s in seq_along(formulalist)) {
    formule <- formulalist[[s]]
    tmplist <- lav_parse_handle_formule(formule, tmplist, types, modelsrc,
                                        modenvl, modenvr, config)
  }
  lav_parse_final_operations(tmplist$flat, tmplist$modlist, config,
    tmplist$constraints, as_data_frame, hashstring)
}

lav_parse_handle_formule <- function(formule, tmplist, types, modelsrc,
                                     modenvl, modenvr, config) {
  flat <- tmplist$flat
  flat_idx <- tmplist$flat_idx
  mod_idx <- tmplist$mod_idx
  constraints <- tmplist$constraints
  modlist <- tmplist$modlist
  block <- tmplist$block
  block_op <- tmplist$block_op
  constraint_operators <- c("==", "<", ">", ":=")
  if (lav_debug()) {
    cat(vapply(seq_along(formule$elem_type), function(j) {
      if (formule$elem_type[j] == types$stringliteral) {
        return(dQuote(formule$elem_text[j], FALSE))
      }
      formule$elem_text[j]
    }, ""), "\n")
  }
  nelem <- length(formule$elem_type)
  # where is the operator
  opi <- which(formule$elem_type == types$lavaanoperator)
  if (length(opi) > 1L) { # if more then 1 operator skip operators ':'
    opii <- 1L
    while (formule$elem_text[opi[opii]] == ":" && opii < length(opi)) {
      opii <- opii + 1L
    }
    opi <- opi[opii]
  }
  op <- formule$elem_text[opi]
  if (any(op == constraint_operators)) {         # ----- constraints ------- #
    # a constraint/definition needs a non-empty left- and right-hand side
    if (opi <= 1L || opi >= nelem) {
      tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[opi])
      lav_msg_stop(
        gettextf("Missing or empty %1$s of operator '%2$s'.",
                 if (opi <= 1L) gettext("left-hand side")
                 else gettext("right-hand side"),
                 op),
        tl[1L], footer = tl[2L]
      )
    }
    lhs <- paste(formule$elem_text[seq.int(1L, opi - 1L)], collapse = "")
    rhs <- paste(formule$elem_text[seq.int(opi + 1L, nelem)], collapse = "")
    constraints <- c(
      constraints,
      list(list(
        op = op,
        lhs = lhs,
        rhs = rhs,
        user = 1L
      ))
    )
    tmplist$constraints <- constraints
    return(tmplist)
  }
  if (op == ":") {           # --------------- block start ----------------- #
    correct_syntax <- gettextf("The correct syntax is: \"LHS : RHS\", where
    LHS is  block identifier (%s) and RHS is the block number or label.",
                             lav_msg_view(config$groupings$group, "or"))
    if (opi == 1L) {
      tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[1])
      lav_msg_stop(
        gettextf(
          "Missing block identifier. %s", correct_syntax
        ), tl[1L], footer = tl[2L]
      )
    }
    if (opi > 2L || all(tolower(formule$elem_text[1]) !=
                        config$groupings$group)) {
      tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[1])
      lav_msg_stop(
        gettextf(
          "Invalid block identifier. %s", correct_syntax
        ),
        tl[1L], footer = tl[2L]
      )
    }
    if (nelem != 3 || all(formule$elem_type[3] !=
            c(types$stringliteral, types$identifier, types$numliteral))) {
      tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[1])
      lav_msg_stop(
        gettextf("Syntax contains block identifier with missing or
              invalid number/label. %s", correct_syntax),
        tl[1L],
        footer = tl[2L]
      )
    }
    flat_idx <- flat_idx + 1L
    flat$lhs[flat_idx] <- formule$elem_text[1]
    flat$op[flat_idx] <- op
    flat$rhs[flat_idx] <- formule$elem_text[3]
    flat$mod_idx[flat_idx] <- 0L
    if (block_op) {
      block <- block + 1L
    } else {
      if (flat_idx != 1) {
        tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[1])
        lav_msg_warn(
          gettext("First block defined after other formula's"),
          tl[1L],
          footer = tl[2L]
        )
      }
    }
    flat$block[flat_idx] <- block
    tmplist$flat <- flat
    tmplist$flat_idx <- flat_idx
    tmplist$block <- block
    tmplist$block_op <- TRUE
    return(tmplist)
  }
  # ------------------ relational operators -------------------------------- #
  lav_parse_check_relational(formule, modelsrc, types, opi, nelem, op)
  tmpenv <- list2env(list(
    formule = formule,
    opi = opi,
    nelem = nelem
  ))
  lav_parse_update_relational(tmpenv, modelsrc, types, op)
  formule <- get("formule", tmpenv)
  opi <- get("opi", tmpenv)
  nelem <- get("nelem", tmpenv)
  rm(tmpenv)
  lhs <- formule$elem_text[opi - 1L]
  rhs <- formule$elem_text[nelem]
  already <- which(flat$lhs == lhs & flat$op == op & flat$block == block &
            (flat$rhs == rhs | (flat$rhs == "" &
              (op == "~" | op == "|~" | op == ":~") &
              formule$elem_type[nelem] == types$numliteral)))
  if (length(already) == 1L) {
    idx <- already
  } else {
    flat_idx <- flat_idx + 1L
    idx <- flat_idx
    flat$lhs[idx] <- lhs
    flat$op[idx] <- op
    flat$rhs[idx] <- rhs
    flat$block[idx] <- block
    if (formule$elem_type[nelem] == types$numliteral) {
      if (op == "~" || op == "|~"| op == ":~") flat$rhs[idx] <- ""
    }
  }
  # modifiers always come before an asterix or a question mark
  modelems <- which(formule$elem_type == types$symbol &
              (formule$elem_text == "*" | formule$elem_text == "?"))
  for (j in seq_along(modelems)) {
    if (sum(formule$elem_text[seq.int(opi, modelems[j])] == "(") !=
        sum(formule$elem_text[seq.int(opi, modelems[j])] == ")"))
      modelems[j] <- 0L
  }
  modelems <- modelems[modelems != 0L]
  if (length(modelems) != 0L) {
    modleftfirst <- TRUE
    modrightfirst <- TRUE
    for (rmei in seq_along(modelems)) {
      rme <- modelems[rmei]
      if (rme < opi) {
        rmeprev <- if (modleftfirst) 0L else modelems[rmei - 1L]
        modleftfirst <- FALSE
      } else {
        rmeprev <- if (modrightfirst) opi else modelems[rmei - 1L]
        modrightfirst <- FALSE
      }
      modnu <- lav_parse_modifier_open(
        formule,
        opi, modelsrc, types, rme, rmeprev,
        modenv = if (rme < opi) modenvl else modenvr,
        config
      )
      for (j in seq_along(config$modifiers$mod)) {
        mod1 <- config$modifiers$mod[j]
        flat[[mod1]][idx] <-
          if (is.null(modnu[[mod1]])) {
            flat[[mod1]][idx]
          } else {
            paste(modnu[[mod1]], collapse = ";")
          }
      }
      if (flat$mod_idx[idx] == 0) { # not yet modifier
        mod_idx <- mod_idx + 1L
        cur_mod_idx <- mod_idx
        modlist[[cur_mod_idx]] <- modnu
        flat$mod_idx[idx] <- cur_mod_idx
      } else { # use existing modifier index
        cur_mod_idx <- flat$mod_idx[idx]
        overwrite <- names(modnu)[names(modnu) %in%
                                    names(modlist[[cur_mod_idx]])]
        if (length(overwrite) > 0) {
          tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[rmeprev + 1L])
          lav_msg_warn(
            gettextf(
              "modifier %s specified multiple times, overwritten",
              overwrite[1L]
            ), tl[1L],
            footer = tl[2L]
          )
        }
        modlist[[cur_mod_idx]] <- modifyList(modlist[[cur_mod_idx]], modnu)
      }
    }
  } else {
    # length modelems == 0
    if (opi != nelem - 1L) {
      tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[opi + 1L])
      lav_msg_stop(gettext("invalid formula right hand side"),
        tl[1L],
        footer = tl[2L]
      )
    }
  }
  # check for variable regressed on itself
  if (formule$elem_text[opi] %in% c("~", ":~") &&
      formule$elem_text[opi - 1L] == formule$elem_text[nelem]) {
    if (!grepl("^0\\.?0*$", flat$fixed[idx])) {
      tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[opi])
      lav_msg_stop(
        gettext("a variable cannot be regressed on itself"),
        tl[1L],
        footer = tl[2L]
      )
    }
  }
  list(
    flat = flat,
    flat_idx = flat_idx,
    mod_idx = mod_idx,
    constraints = constraints,
    modlist = modlist,
    block = block,
    block_op = block_op
  )
}

lav_parse_check_relational <- function(formule, modelsrc, types, opi,
                                       nelem, op) {
  # warn if some identifiers contain spaces
  contsp <- which(formule$elem_type == types$identifier &
                    grepl(" ", formule$elem_text, fixed = TRUE))
  if (length(contsp) > 0L) {
    tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[contsp[1L]])
    lav_msg_warn(
      gettextf(
        "having identifiers with spaces ('%s') is deprecated",
        formule$elem_text[contsp[1]]
      ),
      tl[1L],
      footer = tl[2L]
    )
  }
  # checks for valid names in lhs and rhs
  lav_parse_check_name(formule, opi - 1L, modelsrc) # valid name lhs
  for (j in seq.int(opi + 1L, nelem)) { # valid names rhs
    if (formule$elem_type[j] == types$identifier &&
        formule$elem_text[j] != "NA") {
      lav_parse_check_name(formule, j, modelsrc)
    }
  }
  if (formule$elem_type[nelem] != types$identifier &&
      (formule$elem_type[nelem] != types$numliteral ||
        all(op != c("~", "|~", "=~", ":~")))) {
    tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[nelem])
    lav_msg_stop(
      gettext("Last element of rhs part expected to be an identifier or,
              for operator ~, |~, =~ or :~, a numeric literal!"),
      tl[1L],
      footer = tl[2L]
    )
  }
   if (nelem < opi + 1L &&
     !(formule$elem_text[nelem - 1L] %in% c("*", "?"))) {
    tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[nelem - 1L])
    lav_msg_stop(gettext("invalid modifier symbol (should be '*' or '?')"),
                  tl[1L],
                  footer = tl[2L]
    )
  }
  invisible(NULL)
}

lav_parse_update_relational <- function(tmpenv, modelsrc, types, op) {
  formule <- tmpenv$formule
  opi <- tmpenv$opi
  nelem <- tmpenv$nelem
  # intercept fixed on 0
  # replace 'lhs ~ 0' => 'lhs ~ 0 * 1' - intercept fixed on zero
  if (formule$elem_text[nelem] == "0" && op == "~" && opi == nelem - 1L) {
    formule$elem_type <- c(formule$elem_type, types$symbol, types$numliteral)
    formule$elem_text <- c(formule$elem_text, "*", "1")
    formule$elem_pos <- c(formule$elem_pos, rep(formule$elem_pos[nelem], 2))
    nelem <- length(formule$elem_type)
  }
  # phantom latent variable
  # replace 'lhs =~ 0' => 'lhs =~ fixed(0)*lhs', 0 can be other numliteral
  #          also, lhs is last element before '=~'
  if (formule$elem_type[nelem] == types$numliteral && op == "=~") {
    formule$elem_type <- c(
      formule$elem_type[seq.int(1L, nelem - 1L)], types$identifier,
      types$symbol, types$numliteral, types$symbol, types$symbol,
      types$identifier
    )
    formule$elem_text <- c(
      formule$elem_text[seq.int(1L, nelem - 1L)], "fixed", "(",
      formule$elem_text[nelem], ")", "*", formule$elem_text[opi - 1L]
    )
    formule$elem_pos <- c(
      formule$elem_pos[seq.int(1L, nelem - 1L)],
      rep(formule$elem_pos[nelem], 6)
    )
    nelem <- length(formule$elem_type)
  }
  # handling interaction variable types
  colons <- which(formule$elem_text[seq.int(1L, nelem - 1L)] == ":" &
                    formule$elem_type[seq.int(2L, nelem)] == types$identifier)
  # check at most 1 colon
  if (length(colons) > 2L ||
      (length(colons) == 2L && (colons[1L] > opi || colons[2L] < opi))) {
    tl <- lav_parse_txtloc(modelsrc, formule$elem_pos[colons[2]])
    lav_msg_stop(
      gettext(
        "Three-way or higher-order interaction terms (using multiple
        colons) are not supported in the lavaan syntax; please manually
        construct the product terms yourself in the data.frame, give
        them an appropriate name, and then you can use these interaction
        variables as any other (observed) variable in the model syntax."
      ), tl[1L], footer = tl[2L]
    )
  }
  if (length(colons) > 0L) {
    # collapse items around colon "a" ":" "b" => "a:b"
    formule$elem_text[colons[1L] - 1L] <-
      paste(formule$elem_text[seq.int(colons[1L] - 1L, colons[1L] + 1L)],
            collapse = ""
      )
    formule <- lav_parse_sublist(formule,
          setdiff(seq.int(1L, nelem), seq.int(colons[1L], colons[1L] + 1L)))
    nelem <- length(formule$elem_type)
    if (colons[1L] < opi) {
      opi <- opi - 2L # is in LHS
      if (length(colons) == 2L) colons[2L] <- colons[2L] - 2L
    }
  }
  if (length(colons) == 2L) {
    # collapse items around colon "a" ":" "b" => "a:b"
    formule$elem_text[colons[2L] - 1L] <-
      paste(formule$elem_text[seq.int(colons[2L] - 1L, colons[2L] + 1L)],
            collapse = ""
      )
    formule <- lav_parse_sublist(formule,
          setdiff(seq.int(1L, nelem), seq.int(colons[2L], colons[2L] + 1L)))
    nelem <- length(formule$elem_type)
  }
  assign("formule", formule, tmpenv)
  assign("opi", opi, tmpenv)
  assign("nelem", nelem, tmpenv)
  invisible(NULL)
}

lav_parse_final_operations <- function(flat, modlist, config,
                        constraints, as_data_frame, hashstring) {
  # update flat (omit items without operator)
  filled_ones <- which(flat$op != "")
  flat$lhs <- flat$lhs[filled_ones]
  flat$op <- flat$op[filled_ones]
  flat$rhs <- flat$rhs[filled_ones]
  flat$mod.idx <- flat$mod_idx[filled_ones]
  flat$mod_idx <- NULL
  flat$block <- flat$block[filled_ones]

  # change op for intercepts (for convenience only)
  int_idx <- which(flat$op == "~" & flat$rhs == "")
  if (length(int_idx) > 0L) {
    flat$op[int_idx] <- "~1"
  }
  # change op for instruments (for convenience only)
  int_idx <- which(flat$op == "|~" & flat$rhs == "")
  if (length(int_idx) > 0L) {
    flat$op[int_idx] <- "|~1"
  }
  # if there are constraints that are simple lower or upper limits, put
  # them in these members, add a modifier and remove constraint
  aantal <- length(constraints)
  if (aantal > 0) {
    for (j in aantal:1) {
      if (any(flat$label == constraints[[j]]$lhs) &&
          any(constraints[[j]]$op == c("<", ">"))) {
        rhslang <- str2lang(constraints[[j]]$rhs)
        numbound <- NA_real_
        if (mode(rhslang) == "numeric") {
          numbound <- as.numeric(constraints[[j]]$rhs)
        } else {
          if (mode(rhslang) == "call") {
            if (is.numeric(tryCatch(eval(rhslang),
                                    error = function(e) "error"))) {
              numbound <- eval(rhslang)
            }
          }
        }
        if (!is.na(numbound)) {
          nrs <- which(flat$label == constraints[[j]]$lhs)
          for (nr in nrs) {
            nrm <- length(modlist) + 1L
            if (flat$mod.idx[nr] > 0L) {
              nrm <- flat$mod.idx[nr]
            } else {
              flat$mod.idx[nr] <- nrm
              modlist <- c(modlist, list(label = constraints[[j]]$lhs))
            }
            if (constraints[[j]]$op == "<") {
              flat$upper[nr] <- as.character(numbound)
              modlist[[nrm]]$upper <- numbound
            } else {
              flat$lower[nr] <- as.character(numbound)
              modlist[[nrm]]$lower <- numbound
            }
          }
          constraints <- constraints[-j]
        }
      }
    }
  }
  # new in 0.6, reorder covariances here!
  flat <- lav_pt_covariance_reorder(flat)
  # remove empty modifier columns
  for (mod1 in config$modifiers$mod) {
    flat[[mod1]] <- flat[[mod1]][filled_ones]
    if (all(flat[[mod1]] == "")) flat[[mod1]] <- NULL
  }
  # transform to data.frame is asked
  if (as_data_frame) {
    flat <- as.data.frame(flat, stringsAsFactors = FALSE)
  }
  # new in 0.6-4: check for 'group' within 'level'
  if (any(flat$op == ":")) {
    op_idx <- which(flat$op == ":")
    if (length(op_idx) < 2L) {
      # only 1 block identifier? this is weird -> give warning
      lav_msg_warn(gettext("syntax contains only a single block identifier!"))
    } else {
      first_block <- flat$lhs[op_idx[1L]]
      second_block <- flat$lhs[op_idx[2L]]
      if (first_block == "level" && second_block == "group") {
        lav_msg_stop(gettext("groups can not be nested within levels!"))
      }
    }
  }
  # create output
  attr(flat, "modifiers") <- modlist
  attr(flat, "constraints") <- constraints
  assign(hashstring, flat, envir = lavaan_cache_env)
  flat
}

# store pta in attributes of partable
lav_pt_set_cache <- function(partable, pta = NULL, force = FALSE) {
  if (!force &&
      !is.null(attr(partable, "vnames")) &&
      !is.null(attr(partable, "nvar"))) {
        return(partable)                    # cache already OK
    }
  if (is.null(pta)) {
    if (force) attr(partable, "vnames") <- NULL
    pta <- lav_pt_attributes(partable)
  }

  for (n in names(pta)) {
    attr(partable, n) <- pta[[n]]
  }

  partable
}

lav_pt_remove_cache <- function(partable) {
  attributelist <- names(attributes(partable))

  for (n in attributelist) {
    if (n != "ovda" && n != "names") attr(partable, n) <- NULL
  }

  partable
}

# the multivariate normal distribution

# 1) loglikelihood (from raw data, or sample statistics)
# 2) derivatives with respect to mu, Sigma, vech(Sigma)
# 3) casewise scores with respect to mu, vech(Sigma), mu + vech(Sigma)
# 4) hessian mu + vech(Sigma)
# 5) information h0 mu + vech(Sigma)
#    5a: (unit)    expected information
#    5b: (unit)    observed information
#    5c: (unit) first.order information
# 6) inverted information h0 mu + vech(Sigma)
#    6a: (unit) inverted expected information
#    6b: /
#    6c: /
# 7) ACOV h0 mu + vech(Sigma)
#    7a: 1/N * inverted expected    information
#    7b: 1/N * inverted observed    information
#    7c: 1/N * inverted first-order information
#    7d: sandwich acov

# YR 07 Feb 2016: first version
# YR 24 Mar 2016: added firstorder information, hessian logl
# YR 19 Jan 2017: added lav_mvnorm_inverted_information_expected
# YR 27 Jun 2018: adding cluster.idx= argument for information_firstorder
# YR 04 Okt 2018: adding wt= argument, and missing meanstructure=
# YR 24 Jul 2022: adding correlation= argument for information_expected
#                 (only for catml; not for correlation = TRUE!)
# YR 03 Jul 2026: use shared kernels from lav_mvnorm_kernels.R

# 0. densities
lav_mvn_dmvnorm <- function(y = NULL,
                               wt = NULL,
                               mu = NULL,
                               sigma_1 = NULL,
                               sigma_inv = NULL,
                               sinv_method = "eigen",
                               x_idx = integer(0L),
                               x_mean = NULL,
                               x_cov = NULL,
                               log = TRUE) {
  if (is.matrix(y)) {
    if (is.null(mu) && is.null(sigma_1) && is.null(sigma_inv)) {
      out <- lav_mvn_loglik_data_z(y = y, casewise = TRUE)
    } else {
      out <- lav_mvn_loglik_data(
        y = y, mu = mu, sigma_1 = sigma_1,
        casewise = TRUE,
        sinv_method = sinv_method
      )
    }
  } else {
    # just one
    p <- length(y)
    log_2pi <- log(2 * pi)

    if (is.null(mu) && is.null(sigma_1) && is.null(sigma_inv)) {
      # mahalanobis distance
      dist_1 <- sum(y * y)
      out <- -(p * log_2pi + dist_1) / 2
    } else {
      sigma_inv <- lav_mvn_sigma_inv(
        sigma_1 = sigma_1, sigma_inv = sigma_inv,
        sinv_method = sinv_method, logdet = TRUE
      )
      logdet <- attr(sigma_inv, "logdet")

      # mahalanobis distance
      yc <- y - mu
      dist_1 <- sum(yc %*% sigma_inv * yc)
      out <- -(p * log_2pi + logdet + dist_1) / 2
    }
  }

  if (!is.null(wt)) {
    out <- out * wt
  }

  # x.idx?
  if (length(x_idx) > 0L) {
    if (is.null(sigma_1) && is.null(x_cov)) {
      lav_msg_stop(gettext("when x.idx is not empty, we need Sigma or x.cov"))
    }
    if (is.matrix(y)) {
      x <- y[, x_idx, drop = FALSE]
    } else {
      x <- y[x_idx]
    }

    mu_x <- x_mean
    sigma_x <- x_cov
    if (is.null(x_mean)) {
      mu_x <- as.numeric(mu)[x_idx]
    }
    if (is.null(x_cov)) {
      sigma_x <- sigma_1[x_idx, x_idx, drop = FALSE]
    }

    logl_x <- lav_mvn_dmvnorm(
      y = x, wt = wt, mu = mu_x, sigma_1 = sigma_x,
      sigma_inv = NULL,
      sinv_method = sinv_method,
      x_idx = integer(0L), log = TRUE
    )

    # subtract logl.X
    out <- out - logl_x
  }

  if (!log) {
    out <- exp(out)
  }

  out
}

# 1. likelihood

# 1a: input is raw data
# (note casewise = TRUE same as: dmvnorm(Y, mean, sigma, log = TRUE))
lav_mvn_loglik_data <- function(y = NULL,
                                   wt = NULL,
                                   mu = NULL,
                                   sigma_1 = NULL,
                                   x_idx = integer(0L),
                                   x_mean = NULL,
                                   x_cov = NULL,
                                   casewise = FALSE,
                                   sinv_method = "eigen") {
  # Y must be a matrix (use lav_mvn_dmvnorm() for non-matrix input)
  stopifnot(is.matrix(y))

  n <- lav_mvn_nobs(y = y, wt = wt)
  p <- NCOL(y)
  mu <- as.numeric(mu)

  if (casewise) {
    log_2pi <- log(2 * pi)

    # invert Sigma
    if (sinv_method == "chol") {
      c_s <- chol(sigma_1)
      ic_s <- backsolve(c_s, diag(p))
      yc <- t(t(y) - mu)
      dist_1 <- rowSums((yc %*% ic_s)^2)
      logdet <- -2 * sum(log(diag(ic_s)))
    } else {
      sigma_inv <- lav_mat_sym_inverse(
        s = sigma_1, logdet = TRUE,
        sinv_method = sinv_method
      )
      logdet <- attr(sigma_inv, "logdet")
      # mahalanobis distance
      yc <- t(t(y) - mu)
      dist_1 <- rowSums(yc %*% sigma_inv * yc)
    }

    loglik <- -(p * log_2pi + logdet + dist_1) / 2

    # weights
    if (!is.null(wt)) {
      loglik <- loglik * wt
    }
  } else {
    # invert Sigma
    sigma_inv <- lav_mat_sym_inverse(
      s = sigma_1, logdet = TRUE,
      sinv_method = sinv_method
    )
    samp <- lav_mvn_samp_stats(y = y, wt = wt)
    loglik <- lav_mvn_loglik_samp(
      sample_mean = samp$mean,
      sample_cov = samp$cov,
      sample_nobs = n,
      mu = mu,
      sigma_inv = sigma_inv
    )
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    mu_x <- x_mean
    sigma_x <- x_cov
    if (is.null(x_mean)) {
      mu_x <- as.numeric(mu)[x_idx]
    }
    if (is.null(x_cov)) {
      sigma_x <- sigma_1[x_idx, x_idx, drop = FALSE]
    }
    loglik_x <- lav_mvn_loglik_data(
      y = y[, x_idx, drop = FALSE],
      wt = wt, mu = mu_x, sigma_1 = sigma_x,
      x_idx = integer(0L), casewise = casewise,
      sinv_method = sinv_method
    )
    # subtract logl.X
    loglik <- loglik - loglik_x
  }

  loglik
}



# 1b: input are sample statistics (mean, cov, N) only
lav_mvn_loglik_samp <- function(sample_mean = NULL,
                                          sample_cov = NULL,
                                          sample_nobs = NULL,
                                          mu = NULL,
                                          sigma_1 = NULL,
                                          x_idx = integer(0L),
                                          x_mean = NULL,
                                          x_cov = NULL,
                                          sinv_method = "eigen",
                                          sigma_inv = NULL) {
  p <- length(sample_mean)
  n <- sample_nobs
  mu <- as.numeric(mu)
  sample_mean <- as.numeric(sample_mean)
  log_2pi <- log(2 * pi)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method, logdet = TRUE
  )
  logdet <- attr(sigma_inv, "logdet")

  # tr(Sigma^{-1} %*% S)
  dist1 <- sum(sigma_inv * sample_cov)
  # (ybar - mu)^T %*% Sigma.inv %*% (ybar - mu)
  diff_1 <- as.numeric(sample_mean - mu)
  dist2 <- sum(as.numeric(crossprod(diff_1, sigma_inv)) * diff_1)

  loglik <- -n / 2 * (p * log_2pi + logdet + dist1 + dist2)

  # fixed.x?
  if (length(x_idx) > 0L) {
    mu_x <- x_mean
    sigma_x <- x_cov
    if (is.null(x_mean)) {
      mu_x <- mu[x_idx]
    }
    if (is.null(x_cov)) {
      sigma_x <- sigma_1[x_idx, x_idx, drop = FALSE]
    }
    sample_mean_x <- sample_mean[x_idx]
    sample_cov_x <- sample_cov[x_idx, x_idx, drop = FALSE]
    loglik_x <-
      lav_mvn_loglik_samp(
        sample_mean = sample_mean_x,
        sample_cov = sample_cov_x,
        sample_nobs = sample_nobs,
        mu = mu_x, sigma_1 = sigma_x,
        x_idx = integer(0L),
        sinv_method = sinv_method
      )
    # subtract logl.X
    loglik <- loglik - loglik_x
  }

  loglik
}

# 1c special case: Mu = 0, Sigma = I
lav_mvn_loglik_data_z <- function(y = NULL,
                                     wt = NULL,
                                     casewise = FALSE) {
  n <- lav_mvn_nobs(y = y, wt = wt)
  p <- NCOL(y)
  log_2pi <- log(2 * pi)

  if (casewise) {
    dist_1 <- rowSums(y * y)
    loglik <- -(p * log_2pi + dist_1) / 2
    if (!is.null(wt)) {
      loglik <- loglik * wt
    }
  } else {
    samp <- lav_mvn_samp_stats(y = y, wt = wt)

    dist1 <- sum(diag(samp$cov))
    dist2 <- sum(samp$mean * samp$mean)

    loglik <- -n / 2 * (p * log_2pi + dist1 + dist2)
  }

  loglik
}





# 2. Derivatives

# 2a: derivative logl with respect to mu
lav_mvn_dlogl_dmu <- function(y = NULL,
                                 wt = NULL,
                                 mu = NULL,
                                 sigma_1 = NULL,
                                 x_idx = integer(0L),
                                 sinv_method = "eigen",
                                 sigma_inv = NULL) {
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # subtract 'Mu' from Y
  yc <- t(t(y) - mu)

  # weights
  if (!is.null(wt)) {
    yc <- yc * wt
  }

  # derivative
  dmu <- as.numeric(sigma_inv %*% colSums(yc))

  # fixed.x?
  if (length(x_idx) > 0L) {
    dmu[x_idx] <- 0
  }

  dmu
}

# 2b: derivative logl with respect to Sigma (full matrix, ignoring symmetry)
lav_mvn_dlogl_dsigma <- function(y = NULL,
                                    wt = NULL,
                                    mu = NULL,
                                    sigma_1 = NULL,
                                    x_idx = integer(0L),
                                    sinv_method = "eigen",
                                    sigma_inv = NULL) {
  n <- lav_mvn_nobs(y = y, wt = wt)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # W.tilde
  w_tilde <- lav_mvn_w_tilde(y = y, wt = wt, mu = mu)

  # derivative
  d_sigma <- -(n / 2) * (sigma_inv - (sigma_inv %*% w_tilde %*% sigma_inv))

  # fixed.x?
  if (length(x_idx) > 0L) {
    d_sigma[x_idx, x_idx] <- 0
  }

  d_sigma
}

# 2c: derivative logl with respect to vech(Sigma)
lav_mvn_dlogl_dvechsigma <- function(y = NULL,
                                        wt = NULL,
                                        mu = NULL,
                                        sigma_1 = NULL,
                                        x_idx = integer(0L),
                                        sinv_method = "eigen",
                                        sigma_inv = NULL) {
  d_sigma <- lav_mvn_dlogl_dsigma(
    y = y, wt = wt, mu = mu, sigma_1 = sigma_1,
    x_idx = x_idx, sinv_method = sinv_method, sigma_inv = sigma_inv
  )

  # vech
  as.numeric(lav_mat_dup_pre(as.matrix(lav_mat_vec(d_sigma))))
}

# 2d: : derivative logl with respect to Mu and vech(Sigma)
lav_mvn_dlogl_dmu_dvechsigma <- function(m_y = NULL,
                                            wt = NULL,
                                            m_mu = NULL,
                                            m_sigma = NULL,
                                            x_idx = integer(0L),
                                            sinv_method = "eigen",
                                            sigma_inv = NULL) {
  n <- lav_mvn_nobs(y = m_y, wt = wt)
  m_mu <- as.numeric(m_mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = m_sigma, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # subtract Mu
  m_yc <- t(t(m_y) - m_mu)

  # m_w_tilde
  m_w_tilde <- lav_mvn_w_tilde(y = m_y, wt = wt, mu = m_mu)
  if (!is.null(wt)) {
    dmu <- as.numeric(sigma_inv %*% colSums(m_yc * wt))
  } else {
    dmu <- as.numeric(sigma_inv %*% colSums(m_yc))
  }

  # derivative (avoiding kronecker product)
  m_dsigma <- -(n / 2) * (sigma_inv - (sigma_inv %*% m_w_tilde %*% sigma_inv))

  # fixed.x?
  if (length(x_idx) > 0L) {
    m_dsigma[x_idx, x_idx] <- 0
    dmu[x_idx] <- 0
  }

  # vech
  dvech_sigma <- as.numeric(lav_mat_dup_pre(
    as.matrix(lav_mat_vec(m_dsigma))
  ))

  c(dmu, dvech_sigma)
}

# 3. Casewise scores

# 3a: casewise scores with respect to mu
lav_mvn_sc_mu <- function(y = NULL,
                                 wt = NULL,
                                 mu = NULL,
                                 x_idx = integer(0L),
                                 sigma_1 = NULL,
                                 sinv_method = "eigen",
                                 sigma_inv = NULL) {
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # subtract Mu
  yc <- t(t(y) - mu)

  # postmultiply with Sigma.inv
  sc <- yc %*% sigma_inv

  # weights
  if (!is.null(wt)) {
    sc <- sc * wt
  }

  # fixed.x?
  if (length(x_idx) > 0L) {
    sc[, x_idx] <- 0
  }

  sc
}

# 3b: casewise scores with respect to vech(Sigma)
lav_mvn_sc_sigma <- function(y = NULL,
                                         wt = NULL,
                                         mu = NULL,
                                         sigma_1 = NULL,
                                         x_idx = integer(0L),
                                         sinv_method = "eigen",
                                         sigma_inv = NULL) {
  p <- NCOL(y)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # subtract Mu
  yc <- t(t(y) - mu)

  # score kernel
  sc <- lav_mvn_sc_kernel(yc = yc, sigma_inv = sigma_inv)$sigma

  # fixed.x?
  if (length(x_idx) > 0L) {
    sc[, lav_mat_vech_which_idx(n = p, idx = x_idx)] <- 0
  }

  # weights
  if (!is.null(wt)) {
    sc <- sc * wt
  }

  sc
}

# 3c: casewise scores with respect to mu + vech(Sigma)
lav_mvn_sc_mu_sigma <- function(y = NULL,
                                            wt = NULL,
                                            mu = NULL,
                                            sigma_1 = NULL,
                                            x_idx = integer(0L),
                                            sinv_method = "eigen",
                                            sigma_inv = NULL) {
  p <- NCOL(y)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  # subtract Mu
  yc <- t(t(y) - mu)

  # score kernel
  sc2 <- lav_mvn_sc_kernel(yc = yc, sigma_inv = sigma_inv)
  yc <- sc2$mu
  sc <- sc2$sigma

  # fixed.x?
  if (length(x_idx) > 0L) {
    yc[, x_idx] <- 0
    sc[, lav_mat_vech_which_idx(n = p, idx = x_idx)] <- 0
  }

  out <- cbind(yc, sc)

  # weights
  if (!is.null(wt)) {
    out <- out * wt
  }

  out
}


# 4. hessian of logl

# 4a: hessian logl Mu and vech(Sigma) from raw data
lav_mvn_logl_hessian_data <- function(y = NULL,
                                         wt = NULL,
                                         mu = NULL,
                                         sigma_1 = NULL,
                                         x_idx = integer(0L),
                                         sinv_method = "eigen",
                                         sigma_inv = NULL,
                                         meanstructure = TRUE) {
  n <- lav_mvn_nobs(y = y, wt = wt)

  # observed information
  observed <- lav_mvn_info_observed_data(
    y = y, wt = wt, mu = mu,
    sigma_1 = sigma_1, x_idx = x_idx,
    sinv_method = sinv_method, sigma_inv = sigma_inv,
    meanstructure = meanstructure
  )

  -n * observed
}

# 4b: hessian Mu and vech(Sigma) from samplestats
lav_mvnorm_logl_hessian_samplestats <- # nolint
  function(sample_mean = NULL,
           sample_cov = NULL,
           sample_nobs = NULL,
           mu = NULL,
           sigma_1 = NULL,
           x_idx = integer(0L),
           sinv_method = "eigen",
           sigma_inv = NULL,
           meanstructure = TRUE) {
    n <- sample_nobs

    # observed information
    observed <- lav_mvn_info_observed_samp(
      sample_mean = sample_mean, sample_cov = sample_cov,
      mu = mu, sigma_1 = sigma_1,
      x_idx = x_idx, sinv_method = sinv_method, sigma_inv = sigma_inv,
      meanstructure = meanstructure
    )

    -n * observed
  }

# 5) Information h0

# 5a: unit expected information h0 Mu and vech(Sigma)
lav_mvn_info_expected <- function(y = NULL, # unused!
                                            wt = NULL, # unused!
                                            mu = NULL, # unused!
                                            sigma_1 = NULL,
                                            x_idx = integer(0L),
                                            sinv_method = "eigen",
                                            sigma_inv = NULL,
                                            meanstructure = TRUE,
                      correlation = FALSE) {
  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  i22 <- lav_mvn_kron_dup_half(sigma_inv, correlation = correlation)

  if (meanstructure) {
    out <- lav_mat_bdiag(sigma_inv, i22)
  } else {
    out <- i22
  }

  # fixed.x?
  out <- lav_mvn_zero_x_idx(out,
    n = NCOL(sigma_inv), x_idx = x_idx,
    meanstructure = meanstructure
  )

  out
}

# 5b: unit observed information h0
lav_mvn_info_observed_data <- function(y = NULL,
                                                 wt = NULL,
                                                 mu = NULL,
                                                 sigma_1 = NULL,
                                                 x_idx = integer(0L),
                                                 sinv_method = "eigen",
                                                 sigma_inv = NULL,
                                                 meanstructure = TRUE) {
  # sample statistics
  samp <- lav_mvn_samp_stats(y = y, wt = wt)

  lav_mvn_info_observed_samp(
    sample_mean = samp$mean,
    sample_cov = samp$cov, mu = mu, sigma_1 = sigma_1,
    x_idx = x_idx, sinv_method = sinv_method, sigma_inv = sigma_inv,
    meanstructure = meanstructure
  )
}

# 5b-bis: observed information h0 from sample statistics
lav_mvn_info_observed_samp <- function(
    sample_mean = NULL,
    sample_cov = NULL,
    mu = NULL,
    sigma_1 = NULL,
    x_idx = integer(0L),
    sinv_method = "eigen",
    sigma_inv = NULL,
    meanstructure = TRUE) {

  sample_mean <- as.numeric(sample_mean)
  mu <- as.numeric(mu)

  sigma_inv <- lav_mvn_sigma_inv(
    sigma_1 = sigma_1, sigma_inv = sigma_inv,
    sinv_method = sinv_method
  )

  w_tilde <- lav_mvn_w_tilde_samp(
    sample_mean = sample_mean,
    sample_cov = sample_cov, mu = mu
  )

  info <- lav_mvn_info_obs_kernel(
    sigma_inv = sigma_inv, sigma_1 = sigma_1, w_tilde = w_tilde,
    mean_diff = if (meanstructure) sample_mean - mu else NULL
  )

  if (meanstructure) {
    out <- rbind(
      cbind(info$i11, t(info$i21)),
      cbind(info$i21, info$i22)
    )
  } else {
    out <- info$i22
  }

  # fixed.x?
  out <- lav_mvn_zero_x_idx(out,
    n = NCOL(sigma_inv), x_idx = x_idx,
    meanstructure = meanstructure
  )

  out
}

# 5c: unit first-order information h0
lav_mvn_info_firstorder <- function(y = NULL,
                                              wt = NULL,
                                              cluster_idx = NULL,
                                              mu = NULL,
                                              sigma_1 = NULL,
                                              x_idx = integer(0L),
                                              sinv_method = "eigen",
                                              sigma_inv = NULL,
                                              meanstructure = TRUE) {
  n <- lav_mvn_nobs(y = y, wt = wt)

  if (meanstructure) {
    sc <- lav_mvn_sc_mu_sigma(
      y = y, wt = wt,
      mu = mu, sigma_1 = sigma_1, x_idx = x_idx,
      sinv_method = sinv_method, sigma_inv = sigma_inv
    )
  } else {
    # the caller should use Mu = sample_mean
    sc <- lav_mvn_sc_sigma(
      y = y, wt = wt,
      mu = mu, sigma_1 = sigma_1,
      sinv_method = sinv_method, sigma_inv = sigma_inv
    )
  }

  # handle clustering
  if (!is.null(cluster_idx)) {
    # take the sum within each cluster
    sc <- rowsum(sc, group = cluster_idx, reorder = FALSE, na.rm = TRUE)

    # lower bias if number of clusters is not very high
    # FIXME: reference?
    n_c <- nrow(sc)
    correction_factor <- n_c / (n_c - 1)
    sc <- sc * sqrt(correction_factor)
  }

  # unit information
  out <- crossprod(sc) / n

  out
}


# 6: inverted information h0

# 6a: inverted unit expected information h0 Mu and vech(Sigma)
#
#     Note: this is the same as lav_samp_gamma_nt()
#           but where COV=Sigma and MEAN=Mu
#
lav_mvn_inv_info_expected <- function(y = NULL, # unused!
                                                     wt = NULL, # unused!
                                                     mu = NULL, # unused!
                                                     sigma_1 = NULL,
                                                     x_idx = integer(0L),
                                                     meanstructure = TRUE) {
  if (length(x_idx) > 0L) {
    # cov(Y|X) = A - B m_c^{-1} B'
    # where A = cov(Y), B = cov(Y,X), m_c = cov(X)
    a <- sigma_1[-x_idx, -x_idx, drop = FALSE]
    m_b <- sigma_1[-x_idx,  x_idx, drop = FALSE]
    m_c <- sigma_1[x_idx, x_idx, drop = FALSE]
    ybar_x <- a - m_b %*% solve(m_c, t(m_b))

    # reinsert YbarX in Y+X (residual) covariance matrix
    ybar_x_aug <- matrix(0, nrow = NROW(sigma_1), ncol = NCOL(sigma_1))
    ybar_x_aug[-x_idx, -x_idx] <- ybar_x

    # take difference
    r <- sigma_1 - ybar_x_aug

    ss <- lav_mvn_kron_dup_ginv2(sigma_1)
    rr <- lav_mvn_kron_dup_ginv2(r)
    i22 <- ss - rr

    if (meanstructure) {
      i11 <- ybar_x_aug
      out <- lav_mat_bdiag(i11, i22)
    } else {
      out <- i22
    }
  } else {
    i22 <- lav_mvn_kron_dup_ginv2(sigma_1)
    if (meanstructure) {
      i11 <- sigma_1
      out <- lav_mat_bdiag(i11, i22)
    } else {
      out <- i22
    }
  }

  out
}

# 6b:  inverted unit observed information h0

# one could use the inverse of a partitioned matrix, but that does not
# seem to help much... unless we can find an expression for solve(I22)

# 6c: inverted unit first-order information h0
# /


# 7) ACOV h0 mu + vech(Sigma)
# not implemented, as too trivial

#    7a: 1/N * inverted expected    information

#    7b: 1/N * inverted observed    information

#    7c: 1/N * inverted first-order information

#    7d: sandwich acov

# shared infrastructure for the univariate/bivariate ('uv'/'bv') modules:
#
#   lav_uvord.R  univariate ordinal probit (thresholds + slopes)
#   lav_uvreg.R  univariate linear regression (incl. residual variance)
#   lav_bvord.R  bivariate ordinal (polychoric/tetrachoric)
#   lav_bvmix.R  bivariate linear/ordinal (polyserial/biserial)
#   lav_bvreg.R  bivariate linear (pearson)
#
# YR/LDW 2026: extracted from the five modules (categorical-code refactor)
#
# All five modules share the same design:
#
#  - a 'cache' environment created by lav_XX_init_cache() holding the data,
#    fixed quantities, and the current parameter vector cache$theta;
#  - compute functions lav_XX_{loglik,logl,lik,sc,grad,hessian}_cache()
#    that evaluate in the cache environment via with(cache, ...);
#    NOTE: assignments made inside with(cache, ...) PERSIST in the cache.
#    The compute functions form a chain: the gradient assumes the
#    (log)lik function has been evaluated at cache$theta first, and the
#    hessian additionally assumes the gradient has been.  The nlminb
#    wrappers below maintain this invariant; call them in order if you
#    use the _cache functions directly.
#  - nlminb wrapper functions ('min fns') created once per module by
#    lav_uvbv_min_fns() below.

# check/normalize case weights (wt)
lav_uvbv_wt_check <- function(y = NULL, wt = NULL) {
  if (is.null(wt)) {
    wt <- rep(1, length(y))
  } else {
    if (length(y) != length(wt)) {
      lav_msg_stop(gettext("length y is not the same as length wt"))
    }
    if (any(wt < 0)) {
      lav_msg_stop(gettext("all weights should be positive"))
    }
  }
  wt
}

# select which derivative functions nlminb receives, given optim_method:
#   nlminb/nlminb2 = objective + gradient + hessian
#   nlminb1        = objective + gradient
#   nlminb0        = objective only
lav_uvbv_optim_fns <- function(optim_method = "nlminb",
                               min_fns = NULL) {
  out <- min_fns
  if (optim_method == "nlminb" || optim_method == "nlminb2") {
    # all three
  } else if (optim_method == "nlminb0") {
    out$gradient <- out$hessian <- NULL
  } else if (optim_method == "nlminb1") {
    out$hessian <- NULL
  }
  out
}

# create the standard nlminb objective/gradient/hessian wrappers for a
# module, implementing the state protocol described above:
#  - the objective stores 'x' in cache$theta and evaluates the (log)lik
#    (populating the intermediate quantities in the cache);
#  - gradient/hessian re-evaluate the earlier links of the chain first
#    whenever they are called at an 'x' that differs from cache$theta.
# All three return -1 * (quantity) / cache$n (minimization, per case).
#
# pre_objective (optional): function(x, cache) returning a replacement
# objective value (e.g. +Inf for an inadmissible 'x') or NULL to proceed.
lav_uvbv_min_fns <- function(logl_fun = NULL,
                             grad_fun = NULL,
                             hessian_fun = NULL,
                             pre_objective = NULL) {
  force(logl_fun); force(grad_fun); force(hessian_fun); force(pre_objective)

  objective <- function(x, cache = NULL) {
    if (!is.null(pre_objective)) {
      bad <- pre_objective(x, cache)
      if (!is.null(bad)) {
        return(bad)
      }
    }
    cache$theta <- x
    -1 * logl_fun(cache = cache) / cache$n
  }

  gradient <- function(x, cache = NULL) {
    # check if x has changed
    if (!all(x == cache$theta)) {
      cache$theta <- x
      tmp <- logl_fun(cache = cache)
      rm(tmp)
    }
    -1 * grad_fun(cache = cache) / cache$n
  }

  hessian <- function(x, cache = NULL) {
    # check if x has changed
    if (!all(x == cache$theta)) {
      cache$theta <- x
      tmp <- logl_fun(cache = cache)
      tmp <- grad_fun(cache = cache)
      rm(tmp)
    }
    -1 * hessian_fun(cache = cache) / cache$n
  }

  # a module may have no analytic gradient/hessian
  if (is.null(grad_fun)) {
    gradient <- NULL
  }
  if (is.null(hessian_fun)) {
    hessian <- NULL
  }

  list(objective = objective, gradient = gradient, hessian = hessian)
}

lav_step07_bounds <- function(lavoptions = NULL,
                                     lavh1 = NULL,
                                     lavdata = NULL,
                                     lavsamplestats = NULL,
                                     lavpartable = NULL) {
  # # # # # # # # # # # # # # #
  # #  7. parameter bounds  # #
  # # # # # # # # # # # # # # #

  # if lavoptions$optim.bounds not NULL and its members lower and upper
  #   have length > 0L
  #   modify lavpartable via lav_partable_add_bounds

  # automatic bounds (new in 0.6-6)
  if (!is.null(lavoptions$optim.bounds) ||
    length(lavoptions$optim.bounds$lower) > 0L ||
    length(lavoptions$optim.bounds$upper) > 0L) {
    if (lav_verbose()) {
      cat("lavpartable bounds ...")
    }
    lavpartable <- lav_pt_add_bounds(
      partable = lavpartable,
      lavh1 = lavh1, lavdata = lavdata, lavsamplestats = lavsamplestats,
      lavoptions = lavoptions
    )
    if (lav_verbose()) {
      cat(" done.\n")
    }
  }

  lavpartable
}

# RAM representation
#
# initial version: YR 2021-10-04

lav_ram <- function(partable = NULL,
                    target = NULL,
                    extra = FALSE,
                    remove_nonexisting = TRUE) {
  # prepare target list
  if (is.null(target)) target <- partable

  stopifnot(!is.null(target$block))

  # not for categorical data (yet)
  if (any(partable$op == "|")) {
    lav_msg_stop(gettext("RAM representation is not (yet) supported for
                         categorical endogenous variables."))
  }

  # not for conditional.x = TRUE yet
  conditional_x <- any(partable$exo > 0L & partable$op == "~")
  if (conditional_x) {
    lav_msg_stop(gettext("RAM representation is not (yet) supported
                         if conditional.x = TRUE"))
  }

  # prepare output
  n <- length(target$lhs)
  tmp_mat <- character(n)
  tmp_row <- integer(n)
  tmp_col <- integer(n)

  # global settings
  meanstructure <- any(partable$op == "~1")
  # categorical <- any(partable$op == "|")
  # group_w_free <- any(partable$lhs == "group" & partable$op == "%")

  # number of blocks
  nblocks <- lav_pt_nblocks(partable)

  # always return ov.idx
  ov_idx <- vector("list", nblocks)
  ov_dummy_names_nox <- vector("list", nblocks)
  ov_dummy_names_x <- vector("list", nblocks)
  if (extra) {
    rep_mm_names <- vector("list", nblocks)
    rep_mm_number <- vector("list", nblocks)
    rep_mm_rows <- vector("list", nblocks)
    rep_mm_cols <- vector("list", nblocks)
    rep_mm_dim_names <- vector("list", nblocks)
    rep_mm_symmetric <- vector("list", nblocks)
  }

  for (g in 1:nblocks) {
    # info from user model per block
    ov_names <- lav_pt_vnames(partable, "ov", block = g)
    nvar <- length(ov_names)
    ov_idx[[g]] <- seq_len(nvar)
    ov_dummy_names_nox[[g]] <- character(0)
    ov_dummy_names_x[[g]] <- character(0)

    lv_names <- lav_pt_vnames(partable, "lv", block = g)
    both_names <- c(ov_names, lv_names)
    nboth <- length(both_names)

    # 1. "=~" indicators
    idx <- which(target$block == g & target$op == "=~")
    tmp_mat[idx] <- "A"
    tmp_row[idx] <- match(target$rhs[idx], both_names)
    tmp_col[idx] <- match(target$lhs[idx], both_names)

    # 2. "~" regressions
    idx <- which(target$block == g & (target$op == "~" |
      target$op == "<~"))
    tmp_mat[idx] <- "A"
    tmp_row[idx] <- match(target$lhs[idx], both_names)
    tmp_col[idx] <- match(target$rhs[idx], both_names)

    # 3. "~~" variances/covariances
    idx <- which(target$block == g & target$op == "~~")
    tmp_mat[idx] <- "S"
    tmp_row[idx] <- match(target$lhs[idx], both_names)
    tmp_col[idx] <- match(target$rhs[idx], both_names)

    # catch lower-elements in theta/psi
    idx_lower <- which(tmp_mat == "S" & tmp_row > tmp_col)
    if (length(idx_lower) > 0L) {
      tmp <- tmp_row[idx_lower]
      tmp_row[idx_lower] <- tmp_col[idx_lower]
      tmp_col[idx_lower] <- tmp
    }

    # 4. "~1" means/intercepts
    idx <- which(target$block == g & target$op == "~1")
    tmp_mat[idx] <- "m"
    tmp_row[idx] <- match(target$lhs[idx], both_names)
    tmp_col[idx] <- 1L

    # 5. "|" th
    # not used yet

    # 6. "~*~" scales
    # not used yet

    # 7. group weights
    idx <- which(target$block == g & target$lhs == "group" &
      target$op == "%")
    tmp_mat[idx] <- "gw"
    tmp_row[idx] <- 1L
    tmp_col[idx] <- 1L

    # 8. instruments
    idx <- which(target$block == g & target$lhs == "group" &
      target$op == "|~")
    tmp_mat[idx] <- "miiv"
    tmp_row[idx] <- 0L
    tmp_col[idx] <- 0L

    if (extra) {
      # mRows
      mm_rows <- list(
        ov.idx = 1L,
        A = nboth,
        S = nboth,
        m = nboth,
        gw = 1L
      )

      # mCols
      mm_cols <- list(
        ov.idx = nvar,
        A = nboth,
        S = nboth,
        m = 1L,
        gw = 1L
      )

      # dimNames for LISREL model matrices
      mm_dim_names <- list(
        ov.idx = list("ov.idx", ov_names),
        A = list(both_names, both_names),
        S = list(both_names, both_names),
        m = list(both_names, "intercept"),
        gw = list("group", "weight")
      )
      # isSymmetric
      mm_symmetric <- list(
        ov.idx = FALSE,
        A = FALSE,
        S = TRUE,
        m = FALSE,
        gw = FALSE
      )

      # which mm's do we need? (always include ov.idx, A and S)
      idx_1 <- which(target$block == g)
      mm_names <- c("ov.idx", "A", "S")
      if (meanstructure) {
        mm_names <- c(mm_names, "m")
      }
      if ("gw" %in% tmp_mat[idx_1]) {
        mm_names <- c(mm_names, "gw")
      }

      rep_mm_names[[g]] <- mm_names
      rep_mm_number[[g]] <- length(mm_names)
      rep_mm_rows[[g]] <- unlist(mm_rows[mm_names])
      rep_mm_cols[[g]] <- unlist(mm_cols[mm_names])
      rep_mm_dim_names[[g]] <- mm_dim_names[mm_names]
      rep_mm_symmetric[[g]] <- unlist(mm_symmetric[mm_names])
    } # extra
  } # nblocks

  rep_1 <- list(
    mat = tmp_mat,
    row = tmp_row,
    col = tmp_col
  )

  # always return ov.idx attribute
  attr(rep_1, "ov.idx") <- ov_idx
  attr(rep_1, "ov.dummy.names.nox") <- ov_dummy_names_nox
  attr(rep_1, "ov.dummy.names.x") <- ov_dummy_names_x

  if (extra) {
    attr(rep_1, "mmNames") <- rep_mm_names
    attr(rep_1, "mmNumber") <- rep_mm_number
    attr(rep_1, "mmRows") <- rep_mm_rows
    attr(rep_1, "mmCols") <- rep_mm_cols
    attr(rep_1, "mmDimNames") <- rep_mm_dim_names
    attr(rep_1, "mmSymmetric") <- rep_mm_symmetric
  }

  rep_1
}

# the model-implied variance/covariance matrix of the observed variables
lav_ram_sigmahat <- function(mlist = NULL, delta = NULL) {
  ov_idx <- as.integer(mlist$ov.idx[1, ])
  a <- mlist$A
  s <- mlist$S

  # get (I-A)^{-1}
  ia_inv <- lav_mat_inverse_iminus(a)

  # compute Sigma for all ov and lv
  vyeta <- tcrossprod(ia_inv %*% s, ia_inv)

  # select only observed part
  vy <- vyeta[ov_idx, ov_idx, drop = FALSE]

  # if delta, scale
  if (!is.null(mlist$delta) && delta) {
    nvar <- ncol(vy)
    mm_delta <- diag(mlist$delta[, 1L], nrow = nvar, ncol = nvar)
    vy <- mm_delta %*% vy %*% mm_delta
  }

  vy
}

# VETA: the variance/covariance matrix of the latent variables only
lav_ram_veta <- function(mlist = NULL) {
  ov_idx <- as.integer(mlist$ov.idx[1, ])
  a <- mlist$A
  s <- mlist$S

  # get (I-A)^{-1}
  ia_inv <- lav_mat_inverse_iminus(a)

  # compute Sigma for all ov and lv
  vyeta <- tcrossprod(ia_inv %*% s, ia_inv)

  # select only latent part
  veta <- vyeta[-ov_idx, -ov_idx, drop = FALSE]

  veta
}

# MuHat: the model-implied means/intercepts
lav_ram_muhat <- function(mlist = NULL) {
  ov_idx <- as.integer(mlist$ov.idx[1, ])
  a <- mlist$A
  m <- mlist$m

  # shortcut
  if (is.null(m)) {
    return(matrix(0, nrow = length(ov_idx), 1L))
  }

  # get (I-A)^{-1}
  ia_inv <- lav_mat_inverse_iminus(a)

  # all means/intercepts
  eyeta <- ia_inv %*% m

  # select observed only
  muhat <- eyeta[ov_idx, , drop = FALSE]

  muhat
}

# derivative of 'Sigma' wrt the (freel) elements in A and/or S
lav_ram_dsigma <- function(m = "A",
                           idx = seq_along(mlist[[m]]),
                           mlist = NULL,
                           vech = TRUE) {
  ov_idx <- as.integer(mlist$ov.idx[1, ])
  a <- mlist$A
  s <- mlist$S

  nvar <- length(ov_idx)
  nboth <- nrow(a)

  # shortcut for ov.idx, m, ...
  if (!m %in% c("A", "S")) {
    pstar <- nvar * (nvar + 1) / 2
    return(matrix(0.0, nrow = pstar, ncol = length(idx)))
  }

  # get (I-A)^{-1}
  ia_inv <- lav_mat_inverse_iminus(a)

  if (m == "A") {
    l1 <- (ia_inv %*% s %*% t(ia_inv))[ov_idx, , drop = FALSE]
    kol_idx <- matrix(1:(nboth * nboth), nboth, nboth, byrow = TRUE)[idx]
    dx <- (l1 %x% ia_inv[ov_idx, , drop = FALSE])[, idx, drop = FALSE] +
      (ia_inv[ov_idx, , drop = FALSE] %x% l1)[, kol_idx, drop = FALSE]
    # this is not really needed (because we select idx=m.el.idx)
    # but just in case we need all elements of beta...
    dx[, which(idx %in% lav_mat_diag_idx(nboth))] <- 0.0
  } else if (m == "S") {
    dx <- (ia_inv[ov_idx, , drop = FALSE] %x% ia_inv[ov_idx, , drop = FALSE])
    # symmetry correction, but keeping all duplicated elements
    # since we depend on idx=m.el.idx
    lower_idx <- lav_mat_vech_idx(nboth, diagonal = FALSE)
    upper_idx <- lav_mat_vechru_idx(nboth, diagonal = FALSE)
    offdiag_sum <- dx[, lower_idx] + dx[, upper_idx]
    dx[, c(lower_idx, upper_idx)] <- cbind(offdiag_sum, offdiag_sum)
    dx <- dx[, idx, drop = FALSE]
  } else {
    lav_msg_stop(gettext("wrong model matrix names:"), m)
  }

  # vech?
  if (vech) {
    v_idx <- lav_mat_vech_idx(nvar)
    dx <- dx[v_idx, , drop = FALSE]
  }

  dx
}

# derivative of 'Mu' wrt the (free) elements in A and/or m
lav_ram_dmu <- function(m = "A",
                        idx = seq_along(mlist[[m]]),
                        mlist = NULL,
                        vech = TRUE) {
  ov_idx <- as.integer(mlist$ov.idx[1, ])
  a <- mlist$A
  # s <- mlist$S

  nvar <- length(ov_idx)
  # nboth <- nrow(a)

  # shortcut for ov.idx, m, ...
  if (!m %in% c("A", "m")) {
    return(matrix(0.0, nrow = nvar, ncol = length(idx)))
  }

  # get (I-A)^{-1}
  ia_inv <- lav_mat_inverse_iminus(a)

  if (m == "A") {
    dx <- (t(ia_inv %*% mlist$m) %x% ia_inv)[ov_idx, idx, drop = FALSE]
  } else if (m == "m") {
    dx <- ia_inv[ov_idx, idx, drop = FALSE]
  } else {
    lav_msg_stop(gettext("wrong model matrix names:"), m)
  }

  dx
}

# derivative of ML/GLS objective function F wrt the free parameters
lav_ram_df <- function(mlist = NULL, omega = NULL, omega_mu = NULL) {
  ov_idx <- as.integer(mlist$ov.idx[1, ])
  a <- mlist$A
  s <- mlist$S

  # nvar <- length(ov_idx)
  # nboth <- nrow(a)

  # get (I-A)^{-1}
  ia_inv <- lav_mat_inverse_iminus(a)

  # meanstructure?
  meanstructure <- FALSE
  if (!is.null(omega_mu)) meanstructure <- TRUE

  # pre-compute
  t_ia_inv <- t(ia_inv)
  omega__ia_inv__s__t_ia_inv <- (omega %*% ia_inv[ov_idx, , drop = FALSE] %*%
                                                             s %*% t(ia_inv))

  # 1. A
  if (meanstructure) {
    a_deriv <-
      -1.0 * ((t(ia_inv)[, ov_idx, drop = FALSE] %*%
                                    (omega_mu %*% t(mlist$m)) %*% t(ia_inv)) +
        (t_ia_inv[, ov_idx, drop = FALSE] %*% omega__ia_inv__s__t_ia_inv))
  } else {
    a_deriv <- -1.0 * (t_ia_inv[, ov_idx, drop = FALSE] %*%
                                                    omega__ia_inv__s__t_ia_inv)
  }

  # 2. S
  s_deriv <- -1.0 * (t_ia_inv[, ov_idx, drop = FALSE] %*% omega %*%
                                                   ia_inv[ov_idx, , drop = FALSE])
  diag(s_deriv) <- 0.5 * diag(s_deriv)

  if (meanstructure) {
    m_deriv <- -1.0 * t(t(omega_mu) %*% ia_inv[ov_idx, , drop = FALSE])
  } else {
    m_deriv <- NULL
  }

  list(
    A = a_deriv,
    S = s_deriv,
    m = m_deriv
  )
}


# Jacobian of the model-implied moments wrt the free parameters
# (single block, RAM representation, classical continuous case only:
# no categorical, no correlation structure, no conditional.x).
#
# Inputs:
#   MLIST         : named list of model matrices for the block
#                   (ov.idx, A, S, [m])
#   m.free.idx    : list of length(MLIST); m.free.idx[[mm]] holds the
#                   matrix-element indices of the free elements of MLIST[[mm]]
#   x.free.idx    : list of length(MLIST); x.free.idx[[mm]] holds the
#                   parameter-vector positions of those free elements
#   nx.free       : total number of free parameters (column dim of out)
#   meanstructure : logical
#   group.w.free  : logical; if TRUE, prepend a row for the free group weight
#
# Output: a matrix with nx.free columns; rows are
# [group weight (if group.w.free) | Mu (if meanstructure)] then [vech(Sigma)].
lav_ram_dimplied_dx <- function(mlist         = NULL,
                                m_free_idx    = NULL,
                                x_free_idx    = NULL,
                                nx_free       = NULL,
                                meanstructure = FALSE,
                                group_w_free  = FALSE) {

  ov_idx <- as.integer(mlist$ov.idx[1L, ])
  nvar   <- length(ov_idx)
  nboth  <- nrow(mlist$A)
  pstar  <- nvar * (nvar + 1L) / 2L

  mnames <- names(mlist)

  # A (asymmetric)
  mm_a_idx <- which(mnames == "A")
  x_a_idx  <- x_free_idx[[mm_a_idx]]
  m_a_idx  <- m_free_idx[[mm_a_idx]]
  n_a      <- length(m_a_idx)

  # S (symmetric)
  mm_s_idx <- which(mnames == "S")
  tmp      <- x_free_idx[[mm_s_idx]]
  x_s_idx  <- tmp[!duplicated(tmp)]
  m_s_idx  <- m_free_idx[[mm_s_idx]][!duplicated(tmp)]
  n_s      <- length(m_s_idx)

  # m (only if meanstructure)
  n_m     <- 0L
  x_m_idx <- integer(0L)
  m_m_idx <- integer(0L)
  if (meanstructure) {
    mm_m_idx <- which(mnames == "m")
    x_m_idx  <- x_free_idx[[mm_m_idx]]
    m_m_idx  <- m_free_idx[[mm_m_idx]]
    n_m      <- length(m_m_idx)
  }

  # precompute
  ia_inv <- lav_mat_inverse_iminus(mlist$A)            # nboth x nboth
  fobs   <- ia_inv[ov_idx, , drop = FALSE]                # nvar  x nboth
  m_w      <- ia_inv %*% mlist$S %*% t(fobs)                # nboth x nvar
  wt     <- t(m_w)                                          # nvar  x nboth
  if (meanstructure) {
    a <- as.vector(ia_inv %*% mlist$m)                    # nboth
  }

  # vech structure for Sigma
  r_s <- lav_mat_vech_row_idx(nvar)
  c_s <- lav_mat_vech_col_idx(nvar)

  # helper: vec index -> (row, col)
  vec2rc <- function(idx, nr) {
    cbind(row = (idx - 1L) %% nr + 1L,
          col = (idx - 1L) %/% nr + 1L)
  }

  # ---- jac_sigma ----
  n_free    <- n_a + n_s
  jac_sigma <- matrix(0, pstar, n_free)
  col       <- 1L

  # A[k,l]: dSigma[r,s] = Fobs[r,k]*W[l,s] + Fobs[s,k]*W[l,r]
  if (n_a > 0L) {
    rc  <- vec2rc(m_a_idx, nboth)
    k_v <- rc[, 1L]
    l_v <- rc[, 2L]

    t1 <- fobs[r_s, k_v, drop = FALSE] * wt[c_s, l_v, drop = FALSE]
    t2 <- fobs[c_s, k_v, drop = FALSE] * wt[r_s, l_v, drop = FALSE]

    jac_sigma[, col:(col + n_a - 1L)] <- t1 + t2
    col <- col + n_a
  }

  # S[k,l] symmetric: dSigma[r,s] = Fobs[r,k]*Fobs[s,l] + Fobs[r,l]*Fobs[s,k]
  # diagonal (k==l): halve (formula double-counts the single parameter)
  if (n_s > 0L) {
    rc  <- vec2rc(m_s_idx, nboth)
    k_v <- pmax(rc[, 1L], rc[, 2L])
    l_v <- pmin(rc[, 1L], rc[, 2L])

    t1 <- fobs[r_s, k_v, drop = FALSE] * fobs[c_s, l_v, drop = FALSE]
    t2 <- fobs[r_s, l_v, drop = FALSE] * fobs[c_s, k_v, drop = FALSE]
    dx <- t1 + t2

    diag_mask <- (k_v == l_v)
    if (any(diag_mask)) {
      dx[, diag_mask] <- dx[, diag_mask] * 0.5
    }

    jac_sigma[, col:(col + n_s - 1L)] <- dx
  }

  # ---- jac_mean (if meanstructure) ----
  if (meanstructure) {
    n_free_mu <- n_a + n_m
    jac_mean  <- matrix(0, nvar, n_free_mu)
    col       <- 1L

    # A[k,l]:  dMu[r] = Fobs[r,k] * a[l]
    if (n_a > 0L) {
      rc  <- vec2rc(m_a_idx, nboth)
      k_v <- rc[, 1L]
      l_v <- rc[, 2L]
      jac_mean[, col:(col + n_a - 1L)] <-
        fobs[, k_v, drop = FALSE] * rep(a[l_v], each = nvar)
      col <- col + n_a
    }

    # m[k]:  dMu[r] = Fobs[r, k]
    if (n_m > 0L) {
      jac_mean[, col:(col + n_m - 1L)] <- fobs[, m_m_idx, drop = FALSE]
    }
  }

  # ---- assemble output ----
  out    <- matrix(0, nrow = pstar, ncol = nx_free)
  el_idx <- c(x_a_idx, x_s_idx)
  out[, el_idx] <- jac_sigma

  if (meanstructure) {
    el_idx_mu      <- c(x_a_idx, x_m_idx)
    outm           <- matrix(0, nrow = nvar, ncol = nx_free)
    outm[, el_idx_mu] <- jac_mean
    out <- rbind(outm, out)
  }

  # group weight: prepend a row with 1.0 at the gw column
  if (group_w_free) {
    mm_gw_idx <- which(mnames == "gw")
    if (length(mm_gw_idx) > 0L) {
      x_gw_idx <- x_free_idx[[mm_gw_idx]]
      out_gw <- matrix(0, nrow = 1L, ncol = nx_free)
      out_gw[1L, x_gw_idx] <- 1.0
      out <- rbind(out_gw, out)
    }
  }

  out
}

## NOTE:
## round(1.2355, 3) = 1.236
## but
## round(1.2345, 3) = 1.234
##
## perhaps we should add 0.0005 or something to avoid this?

lav_dataframe_print <- function(x, ..., nd = 3L) {
  row_names <- rownames(x)
  y <- as.data.frame(lapply(x, function(x) {
    if (is.numeric(x)) round(x, nd) else x
  }))
  rownames(y) <- row_names

  if (!is.null(attr(x, "header"))) {
    cat("\n", attr(x, "header"), "\n\n", sep = "")
  }

  print(y, ...)

  if (!is.null(attr(x, "footer"))) {
    cat("\n", attr(x, "footer"), "\n\n", sep = "")
  }

  invisible(x)
}

lav_lavaanlist_print <- function(x, ...) {
  y <- unclass(x)
  attr(y, "header") <- NULL

  header <- attr(x, "header")
  if (!is.null(header)) {
    if (is.character(header)) {
      cat("\n", header, "\n\n", sep = "")
    } else {
      print(header)
      cat("\n")
    }
  }

  print(y, ...)
  invisible(x)
}


# prints only lower triangle of a symmetric matrix
lav_mat_sym_print <- function(x, ..., nd = 3L, shift = 0L,
                                          diag_na_dot = TRUE) {
  # print only lower triangle of a symmetric matrix
  # this function was inspired by the `print.correlation' function
  # in package nlme
  x <- as.matrix(x) # just in case
  y <- x
  y <- unclass(y)
  attributes(y)[c("header", "footer")] <- NULL
  ll <- lower.tri(x, diag = TRUE)
  y[ll] <- format(round(x[ll], digits = nd))
  y[!ll] <- ""
  if (diag_na_dot) {
    # print a "." instead of NA on the main diagonal (eg lav_efalist_summary)
    diag_idx <- lav_mat_diag_idx(ncol(x))
    tmp <- x[diag_idx]
    if (all(is.na(tmp))) {
      y[diag_idx] <- paste(strrep(" ", nd + 2L), ".", sep = "")
    }
  }
  if (!is.null(colnames(x))) {
    colnames(y) <- abbreviate(colnames(x), minlength = nd + 3L)
  }
  if (shift > 0L) {
    empty_string <- rep(strrep(x = " ", times = shift), times = nrow(x))
    if (!is.null(rownames(x))) {
      rownames(y) <- paste(empty_string, rownames(x), sep = "")
    } else {
      rownames(y) <- empty_string
    }
  }

  if (!is.null(attr(x, "header"))) {
    cat("\n", attr(x, "header"), "\n\n", sep = "")
  }

  print(y, ..., quote = FALSE, right = TRUE)

  if (!is.null(attr(x, "footer"))) {
    cat("\n", attr(x, "footer"), "\n\n", sep = "")
  }

  invisible(x)
}


lav_mat_print <- function(x, ..., nd = 3L, shift = 0L) {
  x <- as.matrix(x) # just in case
  y <- unclass(x)
  attributes(y)[c("header", "footer")] <- NULL
  if (!is.null(colnames(x))) {
    colnames(y) <- abbreviate(colnames(x), minlength = nd + 3L)
  }
  if (shift > 0L) {
    empty_string <- rep(strrep(x = " ", times = shift), times = nrow(x))
    if (!is.null(rownames(x))) {
      rownames(y) <- paste(empty_string, rownames(x), sep = "")
    } else {
      rownames(y) <- empty_string
    }
  }
  if (!is.null(attr(x, "header"))) {
    cat("\n", attr(x, "header"), "\n\n", sep = "")
  }

  print(round(y, nd), right = TRUE, ...)

  if (!is.null(attr(x, "footer"))) {
    cat("\n", attr(x, "footer"), "\n\n", sep = "")
  }

  invisible(x)
}

lav_vector_print <- function(x, ..., nd = 3L, shift = 0L) {
  y <- unclass(x)
  attributes(y)[c("header", "footer")] <- NULL
  # if(!is.null(names(x))) {
  #    names(y) <- abbreviate(names(x), minlength = nd + 3)
  # }
  if (!is.null(attr(x, "header"))) {
    cat("\n", attr(x, "header"), "\n\n", sep = "")
  }

  if (shift > 0L) {
    empty_string <- strrep(x = " ", times = shift)
    tmp <- format(y, digits = nd, width = 2L + nd)
    tmp[1] <- paste(empty_string, tmp[1], sep = "")
    print(tmp, quote = FALSE, ...)
  } else {
    print(round(y, nd), right = TRUE, ...)
  }

  if (!is.null(attr(x, "footer"))) {
    cat("\n", attr(x, "footer"), "\n\n", sep = "")
  }

  invisible(x)
}

lav_parameterestimates_print <- function(x, ..., nd = 3L) {
  # format for numeric values
  num_format <- paste("%", max(8L, nd + 5L), ".", nd, "f", sep = "")
  int_format <- paste("%", max(8L, nd + 5L), "d", sep = "")
  char_format <- paste("%", max(8L, nd + 5L), "s", sep = "")

  # output sections
  gsections <- c(
    "Latent Variables",
    "Composites",
    "Regressions",
    "Covariances",
    "Intercepts",
    "Thresholds",
    "Variances",
    "Scales y*",
    "Group Weight",
    "R-Square"
  )
  asections <- c(
    "Defined Parameters",
    "Constraints"
  )

  # header?
  header <- attr(x, "header")
  if (is.null(header)) {
    header <- FALSE
  }

  if (header) {
    cat("\nParameter Estimates:\n\n")

    # info about parameterization (if categorical only)
    categorical_flag <- attr(x, "categorical")
    if (categorical_flag) {
      # container
      c1 <- c2 <- character(0L)

      # which parameterization?
      c1 <- c(c1, "Parameterization")
      tmp_txt <- attr(x, "parameterization")
      c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
        substring(tmp_txt, 2),
        sep = ""
      ))

      # format c1/c2
      c1 <- format(c1, width = 37L)
      c2 <- format(c2,
        width = 14L + max(0, (nd - 3L)) * 4L, justify = "right"
      )

      # create character matrix
      m_1 <- cbind(c1, c2, deparse.level = 0)
      colnames(m_1) <- rep("", ncol(m_1))
      rownames(m_1) <- rep(" ", nrow(m_1))

      # print
      write.table(m_1, row.names = TRUE, col.names = FALSE, quote = FALSE)
    }

    # info about standard errors (if we have x$se only)
    # 1. information
    # 2. se
    # 3. bootstrap requested/successful draws
    if (!is.null(x$se)) {
      # container
      c1 <- c2 <- character(0L)

      # which type of standard errors?
      c1 <- c(c1, "Standard errors")
      if (attr(x, "se") == "robust.huber.white") {
        tmp_txt <- "sandwich" # since 0.6-6
      } else if (identical(attr(x, "estimator"), "IV") &&
                 attr(x, "se") != "bootstrap") {
        tmp_txt <- "two-stage" # estimator = "IV"
      } else {
        tmp_txt <- attr(x, "se")
      }
      c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
        substring(tmp_txt, 2),
        sep = ""
      ))
      # information
      if (attr(x, "se") != "bootstrap" &&
          identical(attr(x, "estimator"), "IV")) {
        # estimator = "IV": report the two-stage vcov methods and the type of
        # Gamma matrix instead of the information matrix
        c1 <- c(c1, "Standard errors stage 1")
        c2 <- c(c2, attr(x, "iv.vcov.stage1"))
        c1 <- c(c1, "Standard errors stage 2")
        c2 <- c(c2, attr(x, "iv.vcov.stage2"))
        if (!is.null(attr(x, "iv.gamma"))) {
          c1 <- c(c1, "Gamma matrix")
          c2 <- c(c2, attr(x, "iv.gamma"))
        }
      } else if (attr(x, "se") != "bootstrap") {
        # type for information
        if (attr(x, "se") == "robust.huber.white") {
          c1 <- c(c1, "Information bread")
        } else {
          c1 <- c(c1, "Information")
        }
        tmp_txt <- attr(x, "information")
        c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
          substring(tmp_txt, 2),
          sep = ""
        ))

        # if observed, which type? (hessian of h1)
        if (attr(x, "information") == "observed") {
          c1 <- c(c1, "Observed information based on")
          tmp_txt <- attr(x, "observed.information")
          c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
            substring(tmp_txt, 2),
            sep = ""
          ))
        }

        # if h1 is involved, structured or unstructured?
        if (attr(x, "information") %in% c("expected", "first.order") ||
          attr(x, "observed.information") == "h1") {
          if (attr(x, "se") == "robust.huber.white" &&
            attr(x, "h1.information") !=
              attr(x, "h1.information.meat")) {
            c1 <- c(c1, "Information bread saturated (h1) model")
          } else {
            c1 <- c(c1, "Information saturated (h1) model")
          }
          tmp_txt <- attr(x, "h1.information")
          c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
            substring(tmp_txt, 2),
            sep = ""
          ))
        }

        # if sandwich, which information for the meat? (first.order)
        # only print if it is NOT first.order
        if (attr(x, "se") == "robust.huber.white" &&
          attr(x, "information.meat") != "first.order") {
          c1 <- c(c1, "Information meat")
          tmp_txt <- attr(x, "information.meat")
          c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
            substring(tmp_txt, 2),
            sep = ""
          ))
        }

        # if sandwich, structured or unstructured for the meat?
        # only print if it is not the same as h1.information
        if (attr(x, "se") == "robust.huber.white" &&
          attr(x, "h1.information.meat") !=
            attr(x, "h1.information")) {
          c1 <- c(c1, "Information meat saturated (h1) model")
          tmp_txt <- attr(x, "h1.information.meat")
          c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
            substring(tmp_txt, 2),
            sep = ""
          ))
        }
      } # no bootstrap

      #TDJ: Pooling options for lavaan.mi-class objects (which NEVER bootstrap)
      if (isTRUE(attr(x, "pooled"))) {
        ## add an empty element for a space before pooling section
        c1 <- c(c1, "", "Pooled across imputations")
        c2 <- c(c2, "", "Rubin's (1987) rules")
      }
      if (!is.null(attr(x, "scale.W"))) {
        c1 <- c(c1, "Augment within-imputation variance")
        if (attr(x, "scale.W")) {
          c2 <- c(c2, "Scale by average RIV")
        } else {
          c2 <- c(c2, "Add between component")
        }
      }
      if (!is.null(attr(x, "asymptotic"))) {
        c1 <- c(c1, "Wald test for pooled parameters")
        if (attr(x, "asymptotic")) {
          c2 <- c(c2, "Normal (z) distribution")
        } else {
          c2 <- c(c2, "t(df) distribution")
        }
      }

      # 4.
      if (attr(x, "se") == "bootstrap" && !is.null(attr(x, "bootstrap"))) {
        c1 <- c(c1, "Number of requested bootstrap draws")
        c2 <- c(c2, attr(x, "bootstrap"))
        c1 <- c(c1, "Number of successful bootstrap draws")
        c2 <- c(c2, attr(x, "bootstrap.successful"))
      }

      # format c1/c2
      c1 <- format(c1, width = 37L)
      c2 <- format(c2,
        width = 14L + max(0, (nd - 3L)) * 4L, justify = "right"
      )

      # create character matrix
      m_1 <- cbind(c1, c2, deparse.level = 0)
      colnames(m_1) <- rep("", ncol(m_1))
      rownames(m_1) <- rep(" ", nrow(m_1))

      # print
      write.table(m_1, row.names = TRUE, col.names = FALSE, quote = FALSE)

      #TDJ: Message for lavaan.mi-class objects when df > 1000 for t test
      if (isTRUE(attr(x, "infDF"))) {
        cat(c("\n  Pooled t statistics with df >= 1000 are displayed with",
              "\n  df = Inf(inity) to save space. Although the t distribution",
              "\n  with large df closely approximates a standard normal",
              "\n  distribution, exact df for reporting these t tests can be",
              "\n  obtained from parameterEstimates.mi() \n\n"), sep = "")
      }

    }
  }

  # number of groups
  if (is.null(x$group)) {
    ngroups <- 1L
    x$group <- rep(1L, length(x$lhs))
  } else {
    ngroups <- lav_pt_ngroups(x)
  }

  # number of levels
  if (is.null(x$level)) {
    nlevels <- 1L
    x$level <- rep(1L, length(x$lhs))
  } else {
    nlevels <- lav_pt_nlevels(x)
  }

  # block column
  if (is.null(x$block)) {
    x$block <- rep(1L, length(x$lhs))
  }

  # step column (SAM)
  # if(!is.null(x$step)) {
  #    tmp.LABEL <- rep("", length(x$lhs))
  #    p1.idx <- which(x$step == 1L)
  #    p2.idx <- which(x$step == 2L)
  #    tmp.LABEL[p1.idx] <- "1"
  #    tmp.LABEL[p2.idx] <- "2"
  #
  #    if(is.null(x$label)) {
  #        x$label <- tmp.LABEL
  #    } else {
  #        x$label <- paste(x$label, tmp.LABEL, sep = "")
  #    }
  #
  #    x$step <- NULL
  # }

  # round to 3 digits after the decimal point
  y <- as.data.frame(
    lapply(x, function(x) {
      if (is.integer(x)) {
        sprintf(int_format, x)
      } else if (is.character(x)) { # perhaps plabel
        sprintf(char_format, x)
      } else if (is.numeric(x)) {
        sprintf(num_format, x)
      } else {
        x
      }
    }),
    stringsAsFactors = FALSE
  )

  # always remove /block/level/group/op/rhs/label/exo columns
  y$op <- y$group <- y$rhs <- y$label <- y$exo <- NULL
  y$block <- y$level <- NULL
  y$efa <- NULL

  # if standardized, remove std.nox column (space reasons only)
  # unless, std.all is already removed
  if (!is.null(y$std.all)) {
    y$std.nox <- NULL
  }

  # convert to character matrix
  m <- as.matrix(format.data.frame(y,
    na.encode = FALSE,
    justify = "right"
  ))

  # use empty row names
  rownames(m) <- rep("", nrow(m))

  # handle se == 0.0
  if (!is.null(x$se)) {
    se_idx <- which(x$se == 0)
    if (length(se_idx) > 0L) {
      m[se_idx, "se"] <- ""
      if (!is.null(x$z)) {
        m[se_idx, "z"] <- ""
      }
      if (!is.null(x$pvalue)) {
        m[se_idx, "pvalue"] <- ""
      }
      ## for lavaan.mi-class objects (semTools)
      if (!is.null(x$t)) {
        m[se_idx, "t"] <- ""
      }
      if (!is.null(x$df)) {
        m[se_idx, "df"] <- ""
      }
    }

    # handle se == NA
    se_idx <- which(is.na(x$se))
    if (length(se_idx) > 0L) {
      if (!is.null(x$z)) {
        m[se_idx, "z"] <- ""
      }
      if (!is.null(x$pvalue)) {
        m[se_idx, "pvalue"] <- ""
      }
      ## for lavaan.mi-class objects (semTools)
      if (!is.null(x$t)) {
        m[se_idx, "t"] <- ""
      }
      if (!is.null(x$df)) {
        m[se_idx, "df"] <- ""
      }
    }
  }

  # handle lower/upper boundary points
  # attention: to distinguish fixed values from free but bounded values,
  #            we have added a little bit of fuzz to the bounds of
  #            free parameters only
  if (!is.null(x$lower)) {
    diff_lower <- abs(x$lower - x$est)
    b_idx <- which(diff_lower > 0 & diff_lower < sqrt(.Machine$double.eps))
    if (length(b_idx) > 0L) {
      if (!is.null(x$pvalue)) {
        m[b_idx, "pvalue"] <- ""
      }
      if (is.null(x$label)) {
        x$label <- rep("", length(x$lhs))
      }
      x$label[b_idx] <- ifelse(nchar(x$label[b_idx]) > 0L,
        paste(x$label[b_idx], "+lb", sep = ""),
        "lb"
      )
    }
    # remove lower column
    m <- m[, colnames(m) != "lower"]
  }
  if (!is.null(x$upper)) {
    diff_upper <- abs(x$upper - x$est)
    b_idx <- which(diff_upper > 0 & diff_upper < sqrt(.Machine$double.eps))
    if (length(b_idx) > 0L) {
      if (!is.null(x$pvalue)) {
        m[b_idx, "pvalue"] <- ""
      }
      if (is.null(x$label)) {
        x$label <- rep("", length(x$lhs))
      }
      x$label[b_idx] <- ifelse(nchar(x$label[b_idx]) > 0L,
        paste(x$label[b_idx], "+ub", sep = ""),
        "ub"
      )
    }
    # remove upper column
    m <- m[, colnames(m) != "upper"]
  }


  # handle fmi
  if (!is.null(x$fmi)) {
    se_idx <- which(x$se == 0)
    if (length(se_idx) > 0L) {
      m[se_idx, "fmi"] <- ""
      ## for lavaan.mi-class objects (semTools)
      if (!is.null(x$riv)) m[se_idx, "riv"] <- ""
    }

    not_idx <- which(x$op %in% c(":=", "<", ">", "=="))
    if (length(not_idx) > 0L) {
      if (!is.null(x$fmi)) {
        m[not_idx, "fmi"] <- ""
        ## for lavaan.mi-class objects (semTools)
        if (!is.null(x$riv)) m[not_idx, "riv"] <- ""
      }
    }
  }

  # for blavaan, handle Post.SD and PSRF
  if (!is.null(x$Post.SD)) {
    se_idx <- which(x$Post.SD == 0)
    if (length(se_idx) > 0L) {
      m[se_idx, "Post.SD"] <- ""
      if (!is.null(x$psrf)) {
        m[se_idx, "psrf"] <- ""
      }
      if (!is.null(x$PSRF)) {
        m[se_idx, "PSRF"] <- ""
      }
    }

    # handle psrf for defined parameters
    not_idx <- which(x$op %in% c(":=", "<", ">", "=="))
    if (length(not_idx) > 0L) {
      if (!is.null(x$psrf)) {
        m[not_idx, "psrf"] <- ""
      }
      if (!is.null(x$PSRF)) {
        m[not_idx, "PSRF"] <- ""
      }
    }
  }

  # rename some column names
  colnames(m)[colnames(m) == "lhs"] <- ""
  colnames(m)[colnames(m) == "op"] <- ""
  colnames(m)[colnames(m) == "rhs"] <- ""
  colnames(m)[colnames(m) == "step"] <- "Step"
  colnames(m)[colnames(m) == "est"] <- "Estimate"
  colnames(m)[colnames(m) == "se"] <- "Std.Err"
  colnames(m)[colnames(m) == "z"] <- "z-value"
  colnames(m)[colnames(m) == "pvalue"] <- "P(>|z|)"
  colnames(m)[colnames(m) == "std.lv"] <- "Std.lv"
  colnames(m)[colnames(m) == "std.all"] <- "Std.all"
  colnames(m)[colnames(m) == "std.nox"] <- "Std.nox"
  colnames(m)[colnames(m) == "std.user"] <- "Std.usr"
  colnames(m)[colnames(m) == "prior"] <- "Prior"
  colnames(m)[colnames(m) == "fmi"] <- "FMI"
  ## for lavaan.mi-class objects (semTools)
  if ("t" %in% colnames(m)) {
    colnames(m)[colnames(m) == "t"] <- "t-value"
    colnames(m)[colnames(m) == "P(>|z|)"] <- "P(>|t|)"
    colnames(m)[colnames(m) == "riv"] <- "RIV"
  }

  # format column names
  colnames(m) <- sprintf(char_format, colnames(m))

  # exceptions for blavaan: Post.Mean (width = 9), Prior (width = 14)
  if (!is.null(x$Post.Mean)) {
    tmp <- gsub("[ \t]+", "", colnames(m), perl = TRUE)

    # reformat "Post.Mean" column
    col_idx <- which(tmp == "Post.Mean")
    if (length(col_idx) > 0L) {
      tmp_format <- paste("%", max(9, nd + 5), "s", sep = "")
      colnames(m)[col_idx] <- sprintf(tmp_format, colnames(m)[col_idx])
      m[, col_idx] <- sprintf(tmp_format, m[, col_idx])
    }

    # reformat "Prior" column
    col_idx <- which(tmp == "Prior")
    if (length(col_idx) > 0L) {
      max_1 <- max(nchar(m[, col_idx])) + 1L
      tmp_format <- paste("%", max(max_1, nd + 5), "s", sep = "")
      colnames(m)[col_idx] <- sprintf(tmp_format, colnames(m)[col_idx])
      m[, col_idx] <- sprintf(tmp_format, m[, col_idx])
    }
  }

  b <- 0L
  # group-specific sections
  for (g in 1:ngroups) {
    # group header
    if (ngroups > 1L) {
      group_label <- attr(x, "group.label")
      cat("\n\n")
      cat("Group ", g, " [", group_label[g], "]:\n", sep = "")
    }

    for (l in 1:nlevels) {
      # block number
      b <- b + 1L

      # ov/lv names
      ov_names <- lav_object_vnames(x, "ov", block = b)
      # lv_names <- lav_object_vnames(x, "lv", block = b)

      # level header
      if (nlevels > 1L) {
        level_label <- attr(x, "level.label")
        cat("\n\n")
        cat("Level ", l, " [", level_label[l], "]:\n", sep = "")
      }

      # group-specific sections
      for (s in gsections) {
        if (s == "Latent Variables") {
          row_idx <- which(x$op == "=~" & !x$lhs %in% ov_names &
            x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                                 x$label[row_idx])
        } else if (s == "Composites") {
          row_idx <- which(x$op == "<~" & x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                                 x$label[row_idx])
        } else if (s == "Regressions") {
          row_idx <- which(x$op == "~" & x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                                  x$label[row_idx])
        } else if (s == "Covariances") {
          row_idx <- which(x$op == "~~" & x$lhs != x$rhs & !x$exo &
            x$block == b)
          if (length(row_idx) == 0L) next
          # make distinction between residual and plain
          y_names <- unique(c(
            lav_object_vnames(x, "eqs.y"),
            lav_object_vnames(x, "ov.ind"),
            lav_object_vnames(x, "lv.ind")
          ))
          prefix <- rep("", length(row_idx))
          prefix[x$rhs[row_idx] %in% y_names] <- "  ."
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                      x$label[row_idx], prefix = prefix)
          # m[row.idx,1] <- lav_print_format_names(x$rhs[row.idx],
          #                                            x$label[row.idx])
        } else if (s == "Intercepts") {
          row_idx <- which(x$op == "~1" & !x$exo & x$block == b)
          if (length(row_idx) == 0L) next
          # make distinction between intercepts and means
          y_names <- unique(c(
            lav_object_vnames(x, "eqs.y"),
            lav_object_vnames(x, "ov.ind"),
            lav_object_vnames(x, "lv.ind")
          ))
          prefix <- rep("", length(row_idx))
          prefix[x$lhs[row_idx] %in% y_names] <- "  ."
          m[row_idx, 1] <- lav_print_format_names(x$lhs[row_idx],
                           x$label[row_idx], prefix = prefix)
          # m[row.idx,1] <- lav_print_format_names(x$lhs[row.idx],
          #                                             x$label[row.idx])
        } else if (s == "Thresholds") {
          row_idx <- which(x$op == "|" & x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(paste(x$lhs[row_idx], "|",
            x$rhs[row_idx],
            sep = ""
          ), x$label[row_idx])
        } else if (s == "Variances") {
          row_idx <- which(x$op == "~~" & x$lhs == x$rhs & !x$exo &
            x$block == b)
          if (length(row_idx) == 0L) next
          # make distinction between residual and plain
          y_names <- unique(c(
            lav_object_vnames(x, "eqs.y"),
            lav_object_vnames(x, "ov.ind"),
            lav_object_vnames(x, "lv.ind")
          ))
          prefix <- rep("", length(row_idx))
          prefix[x$rhs[row_idx] %in% y_names] <- "  ."
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                 x$label[row_idx], prefix = prefix)
        } else if (s == "Scales y*") {
          row_idx <- which(x$op == "~*~" & x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                                     x$label[row_idx])
        } else if (s == "Group Weight") {
          row_idx <- which(x$lhs == "group" & x$op == "%" & x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                                     x$label[row_idx])
        } else if (s == "R-Square") {
          row_idx <- which(x$op == "r2" & x$block == b)
          if (length(row_idx) == 0L) next
          m[row_idx, 1] <- lav_print_format_names(x$rhs[row_idx],
                                                     x$label[row_idx])
        } else {
          row_idx <- integer(0L)
        }

        # do we need special formatting for this section?
        # three types:
        #  - regular (nothing to do, except row/colnames)
        #  - R-square
        #  - Latent Variables (and Composites), Regressions and Covariances
        #    'bundle' the output per lhs element

        # bundling
        if (s %in% c(
          "Latent Variables", "Composites",
          "Regressions", "Covariances"
        )) {
          nel <- length(row_idx)
          m_1 <- matrix("", nrow = nel * 2, ncol = ncol(m))
          colnames(m_1) <- colnames(m)
          rownames(m_1) <- rep("", NROW(m_1))
          # colnames(M)[1] <- sprintf("%-17s", paste(s, ":", sep = ""))
          if (is.null(x$efa)) {
            lhs <- paste(x$lhs[row_idx], x$op[row_idx])
          } else {
            lhs <- paste(
              x$lhs[row_idx], x$op[row_idx],
              x$efa[row_idx]
            )
          }
          lhs_idx <- seq(1, nel * 2L, 2L)
          rhs_idx <- seq(1, nel * 2L, 2L) + 1L
          if (s == "Covariances") {
            # make distinction between residual and plain
            y_names <- unique(c(
              lav_object_vnames(x, "eqs.y"),
              lav_object_vnames(x, "ov.ind"),
              lav_object_vnames(x, "lv.ind")
            ))
            prefix <- rep("", length(row_idx))
            prefix[x$lhs[row_idx] %in% y_names] <- "."
          } else {
            prefix <- rep("", length(lhs))
          }
          m_1[lhs_idx, 1] <- sprintf("%1s%-15s", prefix, lhs)
          m_1[rhs_idx, ] <- m[row_idx, ]
          # avoid duplicated LHS labels
          if (nel > 1L) {
            del_idx <- integer(0)
            old_lhs <- ""
            for (i in 1:nel) {
              if (lhs[i] == old_lhs) {
                del_idx <- c(del_idx, lhs_idx[i])
              }
              old_lhs <- lhs[i]
            }
            if (length(del_idx) > 0L) {
              m_1 <- m_1[-del_idx, , drop = FALSE]
            }
          }
          cat("\n", s, ":\n", sep = "")
          # cat("\n")
          print(m_1, quote = FALSE)

          # R-square
        } else if (s == "R-Square") {
          m_1 <- m[row_idx, 1:2, drop = FALSE]
          colnames(m_1) <- colnames(m)[1:2]
          rownames(m_1) <- rep("", NROW(m_1))
          # colnames(M)[1] <- sprintf("%-17s", paste(s, ":", sep = ""))
          cat("\n", s, ":\n", sep = "")
          # cat("\n")
          print(m_1, quote = FALSE)

          # Regular
        } else {
          # M <- rbind(matrix("", nrow = 1L, ncol = ncol(m)),
          #           m[row.idx,])
          m_1 <- m[row_idx, , drop = FALSE]
          colnames(m_1) <- colnames(m)
          rownames(m_1) <- rep("", NROW(m_1))
          # colnames(M)[1] <- sprintf("%-17s", paste(s, ":", sep = ""))
          cat("\n", s, ":\n", sep = "")
          # cat("\n")
          print(m_1, quote = FALSE)
        }
      } # GSECTIONS
    } # levels
  } # groups

  # asections
  for (s in asections) {
    if (s == "Defined Parameters") {
      row_idx <- which(x$op == ":=")
      m[row_idx, 1] <- lav_print_format_names(x$lhs[row_idx], "")
      m_1 <- m[row_idx, , drop = FALSE]
      colnames(m_1) <- colnames(m)
    } else if (s == "Constraints") {
      row_idx <- which(x$op %in% c("==", "<", ">"))
      if (length(row_idx) == 0) next
      m[row_idx, 1] <- lav_print_format_con(x$lhs[row_idx],
                          x$op[row_idx], x$rhs[row_idx], nd = nd)
      m[row_idx, 2] <- sprintf(num_format, abs(x$est[row_idx]))
      m_1 <- m[row_idx, 1:2, drop = FALSE]
      colnames(m_1) <- c("", sprintf(char_format, "|Slack|"))
    } else {
      row_idx <- integer(0L)
    }

    if (length(row_idx) == 0L) {
      next
    }

    rownames(m_1) <- rep("", NROW(m_1))
    # colnames(M)[1] <- sprintf("%-17s", paste(s, ":", sep = ""))
    # cat("\n")
    cat("\n", s, ":\n", sep = "")
    print(m_1, quote = FALSE)
  }
  cat("\n")

  invisible(m)
}

lav_print_format_names <- function(names_1, labels_1, prefix = NULL) {
  m_w <- 14
  if (is.null(prefix)) {
    prefix <- rep("", length(names_1))
  }

  multi_b <- FALSE
  if (any(nchar(names_1) != nchar(names_1, "bytes"))) {
    multi_b <- TRUE
  }
  if (any(nchar(labels_1) != nchar(labels_1, "bytes"))) {
    multi_b <- TRUE
  }
  # labels?
  l_idx <- which(nchar(labels_1) > 0L)
  if (length(l_idx) > 0L) {
    if (!multi_b) {
      labels_1 <- abbreviate(labels_1, 4)
      labels_1[l_idx] <- paste(" (", labels_1[l_idx], ")", sep = "")
      max_l <- max(nchar(labels_1))
      names_1 <- abbreviate(names_1,
        minlength = (m_w - max_l),
        strict = TRUE
      )
    } else {
      # do not abbreviate anything (eg in multi-byte locales)
      max_l <- 4L
    }
    names_1 <- sprintf(paste("%-", (m_w - max_l), "s%", max_l, "s",
      sep = ""
    ), names_1, labels_1)
  } else {
    if (!multi_b) {
      names_1 <- abbreviate(names_1, minlength = m_w, strict = TRUE)
    } else {
      names_1 <- sprintf(paste("%-", m_w, "s", sep = ""), names_1)
    }
  }

  char_format <- paste("%3s%-", m_w, "s", sep = "")
  sprintf(char_format, prefix, names_1)
}

lav_print_format_con <- function(lhs, op, rhs, nd) {
  nd <- max(nd, 3)
  m_w <- 41 + (nd - 3) * 3

  nel <- length(lhs)
  if (length(nel) == 0L) {
    return(character(0))
  }
  names_1 <- character(nel)
  for (i in 1:nel) {
    if (rhs[i] == "0" && op[i] == ">") {
      con_string <- paste(lhs[i], " - 0", sep = "")
    } else if (rhs[i] == "0" && op[i] == "<") {
      con_string <- paste(rhs[i], " - (", lhs[i], ")", sep = "")
    } else if (rhs[i] != "0" && op[i] == ">") {
      con_string <- paste(lhs[i], " - (", rhs[i], ")", sep = "")
    } else if (rhs[i] != "0" && op[i] == "<") {
      con_string <- paste(rhs[i], " - (", lhs[i], ")", sep = "")
    } else if (rhs[i] == "0" && op[i] == "==") {
      con_string <- paste(lhs[i], " - 0", sep = "")
    } else if (rhs[i] != "0" && op[i] == "==") {
      con_string <- paste(lhs[i], " - (", rhs[i], ")", sep = "")
    }
    con_string <- abbreviate(con_string, m_w, strict = TRUE)
    char_format <- paste("   %-", m_w, "s", sep = "")
    names_1[i] <- sprintf(char_format, con_string)
  }

  names_1
}

# new in 0.6-12
lav_summary_print <- function(x, ..., nd = 3L) {
  y <- unclass(x) # change to ordinary list

  # get nd, if it is stored as an attribute
  nd_1 <- attr(y, "nd")
  if (!is.null(nd_1) && is.numeric(nd_1)) {
    nd <- as.integer(nd_1)
  }

  # header
  if (!is.null(y$header)) {
    lavaan_version <- y$header$lavaan.version
    sam_approach <- y$header$sam.approach
    optim_method <- y$header$optim.method
    optim_iterations <- y$header$optim.iterations
    optim_converged <- y$header$optim.converged

    # sam or sem?
    if (sam_approach) {
      cat("This is ",
        sprintf("lavaan %s", lavaan_version),
        " -- using the SAM approach to SEM\n",
        sep = ""
      )
    } else {
      cat(sprintf("lavaan %s ", lavaan_version))

      # Convergence or not?
      if (optim_method == "none") {
        cat("-- DRY RUN with 0 iterations --\n")
      } else if (optim_iterations > 0) {
        if (optim_converged) {
          if (optim_iterations == 1L) {
            cat("ended normally after 1 iteration\n")
          } else {
            cat(sprintf(
              "ended normally after %i iterations\n",
              optim_iterations
            ))
          }
        } else {
          if (optim_iterations == 1L) {
            cat("did NOT end normally after 1 iteration\n")
          } else {
            cat(sprintf(
              "did NOT end normally after %i iterations\n",
              optim_iterations
            ))
          }
          cat("** WARNING ** Estimates below are most likely unreliable\n")
        }
      } else {
        cat("did not run (perhaps do.fit = FALSE)?\n")
        cat("** WARNING ** Estimates below are simply the starting values\n")
      }
    }
  }

  #TDJ: print header for lavaan.mi object (or nothing when NULL)
  cat(y$top_of_lavaanmi)

  # optim
  if (!is.null(y$optim)) {
    estimator <- y$optim$estimator
    estimator_args <- y$optim$estimator.args
    optim_method <- y$optim$optim.method
    npar <- y$optim$npar
    eq_constraints <- y$optim$eq.constraints
    nrow_ceq_jac <- y$optim$nrow.ceq.jac
    nrow_cin_jac <- y$optim$nrow.cin.jac
    nrow_con_jac <- y$optim$nrow.con.jac
    con_jac_rank <- y$optim$con.jac.rank

    cat("\n")
    # cat("Optimization information:\n\n")

    c1 <- c("Estimator")
    # second column
    tmp_est <- toupper(estimator)
    if (tmp_est == "DLS") {
      dls_first_letter <- substr(
        estimator_args$dls.GammaNT,
        1L, 1L
      )
      tmp_est <- paste("DLS-", toupper(dls_first_letter), sep = "")
    }
    # reduced-bias M-estimation: show IRBM / ERBM (estimator is ML internally)
    if (!is.null(estimator_args$rbm.method)) {
      tmp_est <- switch(estimator_args$rbm.method,
        implicit = "IRB-ML",
        explicit = "ERB-ML",
        none = "ML",
        "RBM"
      )
    }
    c2 <- tmp_est

    # additional estimator args
    if (!is.null(estimator_args) &&
      length(estimator_args) > 0L) {
      if (estimator == "DLS") {
        c1 <- c(c1, "Estimator DLS value for a")
        c2 <- c(c2, estimator_args$dls.a)
      }
    }

    # optimization method + npar
    c1 <- c(c1, "Optimization method", "Number of model parameters")
    c2 <- c(c2, toupper(optim_method), npar)

    # optional output
    if (eq_constraints) {
      c1 <- c(c1, "Number of equality constraints")
      c2 <- c(c2, nrow_ceq_jac)
    }
    if (nrow_cin_jac > 0L) {
      c1 <- c(c1, "Number of inequality constraints")
      c2 <- c(c2, nrow_cin_jac)
    }
    if (nrow_con_jac > 0L) {
      if (con_jac_rank == (nrow_ceq_jac + nrow_cin_jac)) {
        # nothing to do (don't print, as this is redundant information)
      } else {
        c1 <- c(c1, "Row rank of the constraints matrix")
        c2 <- c(c2, con_jac_rank)
      }
    }

    # format
    c1 <- format(c1, width = 40L)
    c2 <- format(c2,
      width = 11L + max(0, (nd - 3L)) * 4L,
      justify = "right"
    )

    # character matrix
    m <- cbind(c1, c2, deparse.level = 0)
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # sam header
  if (!is.null(y$sam.header)) {
    cat("\n")
    sam_method <- y$sam.header$sam.method
    sam_local_options <- y$sam.header$sam.local.options
    sam_mm_list <- y$sam.header$sam.mm.list
    sam_mm_estimator <- y$sam.header$sam.mm.estimator
    sam_struc_estimator <- y$sam.header$sam.struc.estimator

    # sam method
    c1 <- c("SAM method")
    c2 <- toupper(sam_method)

    # options
    if (sam_method == "local") {
      c1 <- c(c1, "Mapping matrix M method")
      c2 <- c(c2, sam_local_options$M.method)
      # TODo: more!
    }

    # number of measurement blocks
    c1 <- c(c1, "Number of measurement blocks")
    c2 <- c(c2, length(sam_mm_list))

    # estimator measurement blocks
    c1 <- c(c1, "Estimator measurement part")
    c2 <- c(c2, sam_mm_estimator)

    # estimator structural part
    c1 <- c(c1, "Estimator  structural part")
    c2 <- c(c2, sam_struc_estimator)

    # format
    c1 <- format(c1, width = 40L)
    c2 <- format(c2,
      width = 11L + max(0, (nd - 3L)) * 4L,
      justify = "right"
    )

    # character matrix
    m <- cbind(c1, c2, deparse.level = 0)
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # efa/rotation
  if (!is.null(y$rotation)) {
    cat("\n")
    rotation <- y$rotation
    # rotation_args <- y$rotation.args

    # cat("Rotation information:\n\n")
    # container
    c1 <- c2 <- character(0L)

    # rotation method
    c1 <- c(c1, "Rotation method")
    if (rotation$rotation == "none") {
      mm <- toupper(rotation$rotation)
    } else if (rotation$rotation.args$orthogonal) {
      mm <- paste(toupper(rotation$rotation), " ", "ORTHOGONAL",
        sep = ""
      )
    } else {
      mm <- paste(toupper(rotation$rotation), " ", "OBLIQUE",
        sep = ""
      )
    }
    c2 <- c(c2, mm)

    if (rotation$rotation != "none") {
      # method options
      if (rotation$rotation == "geomin") {
        c1 <- c(c1, "Geomin epsilon")
        c2 <- c(c2, rotation$rotation.args$geomin_epsilon)
      } else if (rotation$rotation == "orthomax") {
        c1 <- c(c1, "Orthomax gamma")
        c2 <- c(c2, rotation$rotation.args$orthomax_gamma)
      } else if (rotation$rotation == "cf") {
        c1 <- c(c1, "Crawford-Ferguson gamma")
        c2 <- c(c2, rotation$rotation.args$cf_gamma)
      } else if (rotation$rotation == "oblimin") {
        c1 <- c(c1, "Oblimin gamma")
        c2 <- c(c2, rotation$rotation.args$oblimin_gamma)
      } else if (rotation$rotation == "promax") {
        c1 <- c(c1, "Promax kappa")
        c2 <- c(c2, rotation$rotation.args$promax_kappa)
      }

      # rotation algorithm
      c1 <- c(c1, "Rotation algorithm (rstarts)")
      tmp <- paste(toupper(rotation$rotation.args$algorithm),
        " (", rotation$rotation.args$rstarts, ")",
        sep = ""
      )
      c2 <- c(c2, tmp)

      # Standardized metric (or not)
      c1 <- c(c1, "Standardized metric")
      if (rotation$rotation.args$std_ov) {
        c2 <- c(c2, "TRUE")
      } else {
        c2 <- c(c2, "FALSE")
      }

      # Row weights
      c1 <- c(c1, "Row weights")
      tmp_txt <- rotation$rotation.args$row_weights
      c2 <- c(c2, paste(toupper(substring(tmp_txt, 1, 1)),
        substring(tmp_txt, 2),
        sep = ""
      ))
    }

    # format c1/c2
    c1 <- format(c1, width = 33L)
    c2 <- format(c2,
      width = 18L + max(0, (nd - 3L)) * 4L,
      justify = "right"
    )

    # create character matrix
    m <- cbind(c1, c2, deparse.level = 0)
    colnames(m) <- rep("", ncol(m))
    rownames(m) <- rep(" ", nrow(m))

    # print
    write.table(m, row.names = TRUE, col.names = FALSE, quote = FALSE)
  }

  # data object
  if (!is.null(y$data)) {
    cat("\n")
    lav_data_print_short(y$data, nd = nd)
  }

  # sam local stats: measurement blocks + structural part
  if (!is.null(y$sam)) {
    cat("\n")
    sam_method <- y$sam$sam.method
    sam_mm_table <- y$sam$sam.mm.table
    sam_mm_rel <- y$sam$sam.mm.rel
    sam_struc_fit <- y$sam$sam.struc.fit
    ngroups <- y$sam$ngroups
    nlevels <- y$sam$nlevels
    group_label <- y$sam$group.label
    level_label <- y$sam$level.label
    block_label <- y$sam$block.label

    # measurement
    tmp <- sam_mm_table
    if (sam_method == "global") {
      cat("Summary Information Measurement Part:\n\n")
    } else {
      cat("Summary Information Measurement + Structural:\n\n")
    }
    print(tmp, row.names = rep(" ", nrow(tmp)), nd = nd)

    if (sam_method == "local") {
      # reliability information
      c1 <- c2 <- character(0L)
      if (ngroups == 1L && nlevels == 1L) {
        cat("\n")
        cat("  Model-based reliability latent variables:\n\n")
        tmp <- data.frame(as.list(sam_mm_rel[[1]]))
        class(tmp) <- c("lavaan.data.frame", "data.frame")
        print(tmp, row.names = rep(" ", nrow(tmp)), nd = nd)
      } else if (ngroups > 1L && nlevels == 1L) {
        cat("\n")
        cat("  Model-based reliability latent variables (per group):\n")
        for (g in 1:ngroups) {
          cat("\n")
          cat("  Group ", g, " [", group_label[g], "]:\n\n",
            sep = ""
          )
          tmp <- data.frame(as.list(sam_mm_rel[[g]]))
          class(tmp) <- c("lavaan.data.frame", "data.frame")
          print(tmp, row.names = rep(" ", nrow(tmp)), nd = nd)
        }
      } else if (ngroups == 1L && nlevels > 1L) {
        cat("\n")
        cat("  Model-based reliability latent variables (per level):\n")
        for (g in 1:nlevels) {
          cat("\n")
          cat("  Level ", g, " [", level_label[g], "]:\n\n",
            sep = ""
          )
          tmp <- data.frame(as.list(sam_mm_rel[[g]]))
          class(tmp) <- c("lavaan.data.frame", "data.frame")
          print(tmp, row.names = rep(" ", nrow(tmp)), nd = nd)
        }
      } else if (ngroups > 1L && nlevels > 1L) {
        cat("\n")
        cat("  Model-based reliability latent variables (per group/level):\n")
        for (g in seq_along(block_label)) {
          cat("\n")
          cat("  Group/Level ", g, " [", block_label[g], "]:\n\n",
            sep = ""
          )
          tmp <- data.frame(as.list(sam_mm_rel[[g]]))
          class(tmp) <- c("lavaan.data.frame", "data.frame")
          print(tmp, row.names = rep(" ", nrow(tmp)), nd = nd)
        }
      }

      # skipped if fit.measures were requested: the (structural) test
      # statistics and fit measures are printed in full further below
      if (!is.null(sam_struc_fit)) {
        cat("\n")
        cat("  Summary Information Structural part:\n\n")
        tmp <- data.frame(as.list(sam_struc_fit))
        class(tmp) <- c("lavaan.data.frame", "data.frame")
        print(tmp, row.names = rep(" ", nrow(tmp)), nd = nd)
      }
    }
  }

  # test statistics
  if (!is.null(y$test)) {
    # optional header (eg for local sam objects, where the model test and
    # the fit measures concern the structural part only)
    if (!is.null(attr(y$test, "header"))) {
      cat("\n", attr(y$test, "header"), "\n", sep = "")
    }
    cat("\n")
    lav_test_print(y$test, nd = nd)
  }

  # extra fit measures (if present)
  if (!is.null(y$fit)) {
    add_h0 <- FALSE
    if (!is.null(attr(y$fit, "add.h0"))) {
      add_h0 <- isTRUE(attr(y$fit, "add.h0"))
    }
    lav_fitmeasures_print(y$fit, nd = nd, add_h0 = add_h0)
  }

  # level-specific fit measures (multilevel only, if present)
  if (!is.null(y$fit.by.level)) {
    lav_fit_by_level_print(y$fit.by.level, nd = nd)
  }

  # efa output
  if (!is.null(y$efa)) {
    # get cutoff, if it is stored as an attribute
    ct <- attr(y, "cutoff")
    if (!is.null(ct) && is.numeric(ct)) {
      cutoff <- ct
    } else {
      cutoff <- 0.3
    }
    # get dot.cutoff, if it is stored as an attribute
    dc <- attr(y, "dot.cutoff")
    if (!is.null(dc) && is.numeric(dc)) {
      dot_cutoff <- dc
    } else {
      dot_cutoff <- 0.1
    }
    # get alpha.level, if it is stored as an attribute
    al <- attr(y, "alpha.level")
    if (!is.null(al) && is.numeric(al)) {
      alpha_level <- al
    } else {
      alpha_level <- 0.01
    }

    for (b in seq_len(y$efa$nblocks)) {
      if (length(y$efa$block.label) > 0L) {
        cat("\n\n")
        cat(y$efa$block.label[[b]], ":\n\n", sep = "")
      }
      if (!is.null(y$efa$lambda[[b]])) {
        cat("\n")
        if (!is.null(y$efa$lambda.se[[b]]) && alpha_level > 0) {
          cat("Standardized loadings: (* = significant at ",
            round(alpha_level * 100),
            "% level)\n\n",
            sep = ""
          )
        } else {
          cat("Standardized loadings:\n\n")
        }
        mm_lambda <- unclass(y$efa$lambda[[b]])
        mm_theta <- unname(unclass(y$efa$theta[[b]]))
        lav_print_loadings(mm_lambda,
          nd = nd, cutoff = cutoff,
          dot_cutoff = dot_cutoff,
          alpha_level = alpha_level,
          resvar = mm_theta, # diag elements only
          x_se = y$efa$lambda.se[[b]]
        )
      }

      if (!is.null(y$efa$sumsq_table[[b]])) {
        cat("\n")
        print(y$efa$sumsq_table[[b]], nd = nd)
      }

      # factor correlations:
      if (!y$efa$orthogonal && !is.null(y$efa$psi[[b]]) &&
        ncol(y$efa$psi[[b]]) > 1L) {
        cat("\n")
        if (!is.null(y$efa$psi.se[[b]]) && alpha_level > 0) {
          cat("Factor correlations: (* = significant at ",
            round(alpha_level * 100),
            "% level)\n\n",
            sep = ""
          )
        } else {
          cat("Factor correlations:\n\n")
        }
        lav_print_psi(y$efa$psi[[b]],
          nd = nd,
          alpha_level = alpha_level,
          x_se = y$efa$psi.se[[b]]
        )
      }

      # factor score determinacy (for regression scores only!)
      if (!is.null(y$efa$fs_determinacy[[b]])) {
        cat("
Correlation regression factor scores and factors (determinacy):\n\n")
        print(y$efa$fs_determinacy[[b]], nd = nd)
        cat("
R2 regression factor scores (= squared correlations):\n\n")
        tmp <- y$efa$fs_determinacy[[b]]
        tmp2 <- tmp * tmp
        class(tmp2) <- c("lavaan.vector", "numeric")
        print(tmp2, nd = nd)
      }

      # lambda_structure
      if (!is.null(y$efa$lambda_structure[[b]])) {
        cat("\n")
        cat("Standardized structure (= LAMBDA %*% PSI):\n\n")
        print(y$efa$lambda_structure[[b]], nd = nd)
      }

      # standard errors lambda
      if (!is.null(y$efa$theta.se[[b]])) { # we check for theta.se
        # as lambda.se is needed for '*'
        cat("\n")
        cat("Standard errors standardized loadings:\n\n")
        print(y$efa$lambda.se[[b]], nd = nd)
      }

      # z-statistics lambda
      if (!is.null(y$efa$lambda.zstat[[b]])) {
        cat("\n")
        cat("Z-statistics standardized loadings:\n\n")
        print(y$efa$lambda.zstat[[b]], nd = nd)
      }

      # pvalues lambda
      if (!is.null(y$efa$lambda.pvalue[[b]])) {
        cat("\n")
        cat("P-values standardized loadings:\n\n")
        print(y$efa$lambda.pvalue[[b]], nd = nd)
      }

      # standard errors theta
      if (!is.null(y$efa$theta.se[[b]])) {
        cat("\n")
        cat("Standard errors unique variances:\n\n")
        print(y$efa$theta.se[[b]], nd = nd)
      }

      # z-statistics theta
      if (!is.null(y$efa$theta.zstat[[b]])) {
        cat("\n")
        cat("Z-statistics unique variances:\n\n")
        print(y$efa$theta.zstat[[b]], nd = nd)
      }

      # pvalues theta
      if (!is.null(y$efa$theta.pvalue[[b]])) {
        cat("\n")
        cat("P-values unique variances:\n\n")
        print(y$efa$theta.pvalue[[b]], nd = nd)
      }

      # standard errors psi
      if (!is.null(y$efa$theta.se[[b]])) { # we check for theta.se
        # as psi.se is needed for '*'
        cat("\n")
        cat("Standard errors factor correlations:\n\n")
        print(y$efa$psi.se[[b]], nd = nd)
      }

      # z-statistics psi
      if (!is.null(y$efa$psi.zstat[[b]])) {
        cat("\n")
        cat("Z-statistics factor correlations:\n\n")
        print(y$efa$psi.zstat[[b]], nd = nd)
      }

      # pvalues psi
      if (!is.null(y$efa$psi.pvalue[[b]])) {
        cat("\n")
        cat("P-values factor correlations:\n\n")
        print(y$efa$psi.pvalue[[b]], nd = nd)
      }
    } # blocks
    cat("\n")
  } # efa

  # residuals (printed in the lavResiduals(output = "text") style), after the
  # fit measures and just before the parameter table
  if (!is.null(y$residuals)) {
    res_text <- attr(y$residuals, "text")
    if (!is.null(res_text)) {
      lav_residuals_summary_print(res_text, nd = nd)
    }
  }

  # parameter table
  if (!is.null(y$pe) && is.null(y$efa)) {
    pe <- y$pe
    class(pe) <- c(
      "lavaan.parameterEstimates", "lavaan.data.frame",
      "data.frame"
    )
    print(pe, nd = nd)

    # note for nonlinear defined (:=) parameters under the delta method
    if (isTRUE(attr(pe, "delta.note"))) {
      lav_msg_note(gettext(
        "Standard errors and confidence intervals of the (nonlinear) defined
         (:=) parameters are based on the first-order delta method; for
         strongly nonlinear definitions, se.def = \"mc\" (Monte Carlo) or
         se = \"bootstrap\" may be more accurate."))
    }
  }

  # modification indices
  if (!is.null(y$mi)) {
    cat("Modification Indices:\n\n")
    mi <- y$mi
    rownames(mi) <- NULL
    print(mi, nd = nd)
  }

  invisible(y)
}

# helper function to print the loading matrix, masking small loadings
lav_print_loadings <- function(x, nd = 3L, cutoff = 0.3, dot_cutoff = 0.1,
                               alpha_level = 0.01, resvar = NULL, x_se = NULL) {
  # unclass
  y <- unclass(x)

  # round, and create a character matrix
  y <- format(round(y, nd), width = 3L + nd, justify = "right")

  # right-align column names
  colnames(y) <- format(colnames(y), width = 3L + nd, justify = "right")

  # create dot/empty string
  dot_string <- format(".", width = 3L + nd, justify = "right")
  empty_string <- format(" ", width = 3L + nd)

  # print a 'dot' if dot.cutoff < |loading| < cutoff
  if (dot_cutoff < cutoff) {
    y[abs(x) < cutoff & abs(x) > dot_cutoff] <- dot_string
  }

  # print nothing if |loading| < dot.cutoff
  y[abs(x) < min(dot_cutoff, cutoff)] <- empty_string

  # add 'star' for significant loadings (if provided) using alpha = 0.01
  if (!is.null(x_se) && !any(is.na(x_se))) {
    col_names <- colnames(y)
    row_names <- rownames(y)
    x_se[x_se < sqrt(.Machine$double.eps)] <- 1 # to avoid NA
    zstat <- x / x_se
    z_cutoff <- qnorm(1 - (alpha_level / 2))
    zstat_string <- ifelse(abs(zstat) > z_cutoff, "*", " ")
    y <- matrix(paste(y, zstat_string, sep = ""), nrow(y), ncol(y))
    colnames(y) <- col_names
    rownames(y) <- row_names
  }

  # add resvar
  if (!is.null(resvar)) {
    names_1 <- colnames(y)
    y <- cbind(y, format(round(cbind(resvar, 1 - resvar), nd),
      width = 12L + nd, justify = "right"
    ))
    resvar_names <- format(c("unique.var", "communalities"),
      width = 12L + nd, justify = "right"
    )
    colnames(y) <- c(names_1, resvar_names)
  }

  # print
  print(y, quote = FALSE)
}

# helper function to print the psi matrix, showing signif stars
lav_print_psi <- function(x, nd = 3L, alpha_level = 0.01, x_se = NULL) {
  # unclass
  y <- unclass(x)

  # round, and create a character matrix
  y <- format(round(y, nd), width = 3L + nd, justify = "right")

  # right-align column names
  colnames(y) <- format(colnames(y), width = 3L + nd, justify = "right")

  # add 'star' for significant loadings (if provided) using alpha = 0.01
  if (!is.null(x_se) && !any(is.na(x_se))) {
    col_names <- colnames(y)
    row_names <- rownames(y)
    x_se[x_se < sqrt(.Machine$double.eps)] <- 1 # to avoid NA
    zstat <- x / x_se
    z_cutoff <- qnorm(1 - (alpha_level / 2))
    zstat_string <- ifelse(abs(zstat) > z_cutoff, "*", " ")
    y <- matrix(paste(y, zstat_string, sep = ""), nrow(y), ncol(y))
    colnames(y) <- col_names
    rownames(y) <- row_names
  }

  # remove upper part
  ll <- upper.tri(x, diag = FALSE)
  y[ll] <- ""

  # print
  print(y, quote = FALSE)
}

# Reduced-bias M-estimation (RBM) for continuous-outcome SEM.
#
# RBM is a penalized maximum likelihood method (Firth/Kosmidis-style). The
# discrepancy function, implied moments, loglik, test statistic and information
# matrices are all the ML ones, so the estimator is kept as "ML" internally
# (see lav_options_est_rbm) and only the point estimate differs:
#
#   penalty  P(theta) = -1/2 * tr( J(theta)^-1 E(theta) )
#     J = observed information, E = first-order (outer-product) information.
#
#   implicit (iRBM): maximize  loglik(theta) + P(theta)
#   explicit (eRBM): maximize  loglik(theta), then one-step bias-correct
#                    theta <- theta - bias(theta), bias = -J^-1 dP/dtheta
#
# Standard errors use the sandwich estimator (se = "robust.huber.white"),
# computed by lav_model_vcov() in the usual way.
#
# Reference:
# - Jamil, H., Rosseel, Y., Kemp, O., & Kosmidis, I. (2026). Bias-Reduced
#   Estimation of Structural Equation Models. Structural Equation Modeling: A
#   Multidisciplinary Journal, 33(3), 376-392.
#   https://doi.org/10.1080/10705511.2025.2610462
# - ported from the brlavaan package (github.com/haziqj/brlavaan)

# check whether a matrix is non-positive-definite or has NAs
lav_model_rbm_check_mat <- function(mat) {
  if (anyNA(mat)) {
    return(TRUE)
  }
  eig <- eigen(mat, symmetric = TRUE, only.values = TRUE)$values
  any(eig < -1e-06 * eig[1])
}

# total (N-scaled) information matrix in the (free) parameter space
# kind: "observed" (J) or "first.order" (E)
lav_model_rbm_information <- function(x, lavmodel, lavsamplestats, lavdata,
                                      lavoptions, kind = "observed") {
  lavmodel <- lav_model_set_parameters(lavmodel, x)
  if (kind == "observed") {
    info <- lav_model_info_observed(
      lavmodel = lavmodel, lavsamplestats = lavsamplestats,
      lavdata = lavdata, lavoptions = lavoptions
    )
  } else {
    info <- lav_model_info_firstorder(
      lavmodel = lavmodel, lavsamplestats = lavsamplestats,
      lavdata = lavdata, lavoptions = lavoptions
    )
  }
  # the info functions return unit (per-observation) information
  info <- lavsamplestats@ntotal * info

  # reduce to the optimizer space for simple equality constraints
  if (lavmodel@ceq.simple.only) {
    k <- lavmodel@ceq.simple.K
    info <- t(k) %*% info %*% k
  }
  info
}

# RBM penalty P(theta) = -1/2 * tr( J^-1 E ), evaluated on the loglik scale
lav_model_rbm_penalty <- function(x, lavmodel, lavsamplestats, lavdata,
                                  lavoptions, fallback = -1e8) {
  e <- lav_model_rbm_information(x, lavmodel, lavsamplestats, lavdata,
    lavoptions,
    kind = "first.order"
  )
  j <- lav_model_rbm_information(x, lavmodel, lavsamplestats, lavdata,
    lavoptions,
    kind = "observed"
  )
  if (lav_model_rbm_check_mat(j)) {
    return(fallback)
  }
  jinv <- try(solve(j), silent = TRUE)
  if (inherits(jinv, "try-error")) {
    fallback
  } else {
    # sum(jinv * e) == tr(jinv %*% e) for symmetric jinv
    -sum(jinv * e) / 2
  }
}

# one-step bias for the explicit method: bias = -J^-1 dP/dtheta
lav_model_rbm_bias <- function(x, lavmodel, lavsamplestats, lavdata,
                               lavoptions) {
  j <- lav_model_rbm_information(x, lavmodel, lavsamplestats, lavdata,
    lavoptions,
    kind = "observed"
  )
  if (lav_model_rbm_check_mat(j)) {
    return(rep(as.numeric(NA), length(x)))
  }
  jinv <- try(solve(j), silent = TRUE)
  if (inherits(jinv, "try-error")) {
    return(rep(as.numeric(NA), length(x)))
  }
  a <- numDeriv::grad(
    func = lav_model_rbm_penalty, x = x,
    lavmodel = lavmodel, lavsamplestats = lavsamplestats,
    lavdata = lavdata, lavoptions = lavoptions
  )
  -drop(jinv %*% a)
}

# RBM estimation; returns the free parameter vector 'x' with the same
# attributes as lav_model_est() so that lav_step11_estoptim() can use it.
lav_model_est_rbm <- function(lavmodel = NULL,
                              lavpartable = NULL,
                              lavsamplestats = NULL,
                              lavdata = NULL,
                              lavoptions = NULL,
                              lavcache = list()) {
  rbm_method <- lavoptions$estimator.args$rbm.method
  if (is.null(rbm_method)) {
    rbm_method <- "implicit"
  }

  # general (non-simple) equality constraints are not supported yet
  # (nrow(ceq.JAC) > 0L also catches equality constraints coexisting with
  # inequality constraints/bounds, where @eq.constraints is FALSE)
  if (lavmodel@eq.constraints || nrow(lavmodel@ceq.JAC) > 0L) {
    lav_msg_stop(gettext(
      "estimator \"rbm\" does not support general equality constraints yet;
       use simple (label-based) equality constraints instead."))
  }

  ntotal <- lavsamplestats@ntotal

  # ----- explicit / none: reuse the standard ML optimizer -----
  if (rbm_method %in% c("explicit", "none")) {
    # lav_model_est() dispatches on optim.method; restore the ML optimizer
    lavoptions_ml <- lavoptions
    lavoptions_ml$optim.method <- "nlminb"
    x_ml <- lav_model_est(
      lavmodel = lavmodel, lavpartable = lavpartable,
      lavsamplestats = lavsamplestats, lavdata = lavdata,
      lavoptions = lavoptions_ml, lavcache = lavcache
    )
    if (rbm_method == "none") {
      return(x_ml)
    }

    # explicit: one-step bias correction
    b <- lav_model_rbm_bias(
      as.numeric(x_ml), lavmodel, lavsamplestats, lavdata, lavoptions
    )
    x <- as.numeric(x_ml) - b

    lavmodel_x <- lav_model_set_parameters(lavmodel, x)
    fx <- lav_model_objective(
      lavmodel = lavmodel_x, lavsamplestats = lavsamplestats,
      lavdata = lavdata, lavcache = lavcache
    )

    attr(x, "converged") <- attr(x_ml, "converged")
    attr(x, "iterations") <- attr(x_ml, "iterations")
    attr(x, "control") <- attr(x_ml, "control")
    attr(x, "warn.txt") <- attr(x_ml, "warn.txt")
    attr(x, "fx") <- fx
    attr(x, "dx") <- numeric(0L)
    attr(x, "parscale") <- attr(x_ml, "parscale")
    return(x)
  }

  # ----- implicit: penalized-ML optimization -----

  # starting values (in the optimizer/free space)
  start_x <- lav_model_get_parameters(lavmodel)

  # bounds (mirror lav_model_est; eq.constraints excluded above)
  if (is.null(lavpartable$lower)) {
    lower <- -Inf
  } else if (lavmodel@ceq.simple.only) {
    free_idx <- which(lavpartable$free > 0L & !duplicated(lavpartable$free))
    lower <- lavpartable$lower[free_idx]
  } else {
    lower <- lavpartable$lower[lavpartable$free > 0L]
  }
  if (is.null(lavpartable$upper)) {
    upper <- +Inf
  } else if (lavmodel@ceq.simple.only) {
    free_idx <- which(lavpartable$free > 0L & !duplicated(lavpartable$free))
    upper <- lavpartable$upper[free_idx]
  } else {
    upper <- lavpartable$upper[lavpartable$free > 0L]
  }
  bad_idx <- which(lower > upper)
  if (length(bad_idx) > 0L) {
    lower[bad_idx] <- -Inf
    upper[bad_idx] <- +Inf
  }

  # theta-independent part of the Gaussian loglik (loglik = const - N * fx_ML).
  # Optimizing the penalized negative loglik on its natural scale (rather than
  # the small N * fx_ML - P) is what lets nlminb reach 'relative convergence'
  # (matching brlavaan); the constant does not affect the estimate.
  log2pi <- log(2 * pi)
  loglik_const <- 0
  for (g in seq_len(lavsamplestats@ngroups)) {
    ng <- lavsamplestats@nobs[[g]]
    pg <- NROW(lavsamplestats@cov[[g]])
    loglik_const <- loglik_const -
      0.5 * ng * (pg * log2pi + lavsamplestats@cov.log.det[[g]] + pg)
  }
  if (!is.finite(loglik_const)) {
    loglik_const <- 0
  }

  # penalized objective: minimize  -(loglik + P) = N * fx_ML - P - const
  objective_function <- function(x) {
    lavmodel_x <- lav_model_set_parameters(lavmodel, x)
    fx <- lav_model_objective(
      lavmodel = lavmodel_x, lavsamplestats = lavsamplestats,
      lavdata = lavdata, lavcache = lavcache
    )
    if (!is.finite(fx)) {
      return(1e20)
    }
    pen <- lav_model_rbm_penalty(
      x, lavmodel, lavsamplestats, lavdata, lavoptions
    )
    val <- ntotal * as.numeric(fx) - pen - loglik_const
    if (!is.finite(val)) 1e20 else val
  }

  # nlminb control (mirror lav_model_est)
  control_nlminb <- list(
    eval.max = 20000L, iter.max = 10000L, trace = 0L,
    abs.tol = (.Machine$double.eps * 10), rel.tol = 1e-10,
    step.min = 1.0, step.max = 1.0, x.tol = 1.5e-8, xf.tol = 2.2e-14
  )
  if (!is.null(lavoptions$control)) {
    control_nlminb <- modifyList(control_nlminb, lavoptions$control)
  }
  control <- control_nlminb[c(
    "eval.max", "iter.max", "trace", "step.min", "step.max",
    "abs.tol", "rel.tol", "x.tol", "xf.tol"
  )]

  optim_out <- nlminb(
    start = start_x, objective = objective_function, gradient = NULL,
    lower = lower, upper = upper, control = control
  )

  x <- optim_out$par

  # ML objective (with fx.group attribute) at the solution, for steps 12-14
  lavmodel_x <- lav_model_set_parameters(lavmodel, x)
  fx <- lav_model_objective(
    lavmodel = lavmodel_x, lavsamplestats = lavsamplestats,
    lavdata = lavdata, lavcache = lavcache
  )

  converged <- optim_out$convergence == 0L
  attr(x, "converged") <- converged
  attr(x, "iterations") <- optim_out$iterations
  attr(x, "control") <- control
  attr(x, "warn.txt") <- if (converged) {
    ""
  } else {
    "the optimizer warns that a solution has NOT been found!"
  }
  attr(x, "fx") <- fx
  attr(x, "dx") <- numeric(0L)
  attr(x, "parscale") <- rep(1.0, length(x))
  x
}

# lav_model utility functions

# initial version: YR 25/03/2009: `methods' for the Model class
# - YR 14 Jan 2014: rename object -> lavmodel, all functions as lav_model_*
# - YR 20 Nov 2021: add lav_model_dmmdpar

lav_model_get_parameters <- function(lavmodel = NULL, glist = NULL,   # nolint start
                                     type = "free", extra = TRUE, ...) {   # nolint end
  # type == "free": only non-redundant free parameters (x)
  # type == "user": all parameters listed in User model
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # state or final?
  if (is.null(glist)) glist <- lavmodel@GLIST
                  # if this changes, tag @TDJorgensen in commit message

  if (type == "free") {
    n <- lavmodel@nx.free
    # } else if(type == "unco") {
    #    N <- lavmodel@nx.unco
  } else if (type == "user") {
    n <- lavmodel@nx.user
  }
  x <- numeric(n)

  for (mm in seq_along(lavmodel@GLIST)) {
    if (type == "free") {
      m_idx <- lavmodel@m.free.idx[[mm]]
      x_idx <- lavmodel@x.free.idx[[mm]]
      # } else if(type == "unco") {
      #    m.idx <- lavmodel@m.unco.idx[[mm]]
      #    x.idx <- lavmodel@x.unco.idx[[mm]]
    } else if (type == "user") {
      m_idx <- lavmodel@m.user.idx[[mm]]
      x_idx <- lavmodel@x.user.idx[[mm]]
    }
    x[x_idx] <- glist[[mm]][m_idx]
  }

  if (type == "user" && extra && sum(
    lavmodel@x.def.idx,
    lavmodel@x.ceq.idx,
    lavmodel@x.cin.idx
  ) > 0L) {
    # we need 'free' x
    x_free <- lav_model_get_parameters(
      lavmodel = lavmodel, glist = glist,
      type = "free"
    )
    if (length(lavmodel@x.def.idx) > 0L) {
      x[lavmodel@x.def.idx] <- lavmodel@def.function(x_free)
    }
    if (length(lavmodel@x.ceq.idx) > 0L) {
      x[lavmodel@x.ceq.idx] <- lavmodel@ceq.function(x_free)
    }
    if (length(lavmodel@x.cin.idx) > 0L) {
      tmp <- lavmodel@cin.function(x_free)
      # remove lower/upper bound values (if any)
      bound_idx <- attr(tmp, "bound.idx")
      if (length(bound_idx) > 0L) {
        tmp <- tmp[-bound_idx]
      }
      x[lavmodel@x.cin.idx] <- tmp
    }
  }

  x
}

# warning: this will make a copy of lavmodel
# warning: if categorical/correlation: 'delta' parameterization does
#          not work properly if we have 'mediators' (where x is not fixed)
#          that are observed (residuals are in PSI, and are not additive)
#          Note: fixed in 0.6-20 for recursive models
lav_model_set_parameters <- function(lavmodel = NULL, x = NULL) {
  tmp <- lavmodel@GLIST
  for (mm in seq_along(lavmodel@GLIST)) {
    m_free_idx <- lavmodel@m.free.idx[[mm]]
    x_free_idx <- lavmodel@x.free.idx[[mm]]
    tmp[[mm]][m_free_idx] <- x[x_free_idx]
  }

  correlation <- lavmodel@correlation

  # categorical? set categorical theta elements (if any)
  if (lavmodel@categorical || correlation) {
    nmat <- lavmodel@nmat
    if (lavmodel@representation == "LISREL") {
      for (g in 1:lavmodel@nblocks) {
        # which mm belong to group g?
        mm_in_group <- 1:nmat[g] + cumsum(c(0L, nmat))[g]

        if (lavmodel@estimator %in% c("MML", "FML")) {
          #  ttt <- diag(tmp[mm.in.group]$theta)
          #  diag(tmp[mm.in.group]$theta) <- as.numeric(NA)
          #  if(length(lavmodel@num.idx[[g]]) > 0L) {
          #      diag(tmp[mm.in.group]$theta)[ lavmodel@num.idx[[g]] ] <-
          #          ttt[ lavmodel@num.idx[[g]] ]
          #  }
        } else {
          if (lavmodel@parameterization == "delta") {
            tmp[mm_in_group] <-
              lav_lisrel_residual_variances(
                mlist = tmp[mm_in_group],
                num_idx = lavmodel@num.idx[[g]],
                ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
                ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]]
              )
          } else if (lavmodel@parameterization == "theta") {
            tmp[mm_in_group] <-
              lav_lisrel_delta(
                mlist = tmp[mm_in_group],
                num_idx = lavmodel@num.idx[[g]]
              )
          }
        }
      }
    } else {
      cat("FIXME: deal with theta elements in the categorical case (RAM)")
    }
  }

  if (lavmodel@composites) {
    # for package stdmod only! (vignette stdmod_lavaan uses old fit object)
    # if (.hasSlot(lavmodel, "composites") && lavmodel@composites) {
    nmat <- lavmodel@nmat
    if (lavmodel@representation == "LISREL") {
      for (g in 1:lavmodel@nblocks) {
        # which mm belong to group g?
        mm_in_group <- 1:nmat[g] + cumsum(c(0L, nmat))[g]

        tmp[mm_in_group] <-
          lav_lisrel_comp_set_intresvar(mlist = tmp[mm_in_group])
      }
    } else {
      cat("FIXME: deal with Composites if representation = RAM")
    }
  }

  lavmodel@GLIST <- tmp

  lavmodel
}

# create a standalone GLIST, filled with (new) x values
# (avoiding a copy of lavmodel)
lav_model_x2glist <- function(lavmodel = NULL, x = NULL,
                              type = "free", set_delta = TRUE,
                              m_el_idx = NULL, x_el_idx = NULL) {
  correlation <- lavmodel@correlation

  glist <- lavmodel@GLIST
  for (mm in seq_along(glist)) {
    # skip empty matrix
    if (nrow(glist[[mm]]) == 0L) {
      next
    }
    if (type == "free") {
      m_el_idx_1 <- lavmodel@m.free.idx[[mm]]
      x_el_idx_1 <- lavmodel@x.free.idx[[mm]]
    } else if (type == "unco") {
      m_el_idx_1 <- lavmodel@m.free.idx[[mm]]
      x_el_idx_1 <- lavmodel@x.unco.idx[[mm]]
    } else if (type == "full") {
      if (lavmodel@isSymmetric[mm]) {
        n <- ncol(glist[[mm]])
        m_el_idx_1 <- lav_mat_vech_idx(n)
      } else {
        m_el_idx_1 <- seq_along(glist[[mm]])
      }
      x_el_idx_1 <- seq_along(m_el_idx)
      if (mm > 1) x_el_idx_1 <- x_el_idx_1 + sum(lavmodel@mmSize[1:(mm - 1)])
    } else if (type == "custom") {
      # nothing to do, m.el.idx and x.el.idx should be given
      m_el_idx_1 <- m_el_idx[[mm]]
      x_el_idx_1 <- x_el_idx[[mm]]
    }

    # assign
    glist[[mm]][m_el_idx_1] <- x[x_el_idx_1]

    # make symmetric (if full)
    if (type == "full" && lavmodel@isSymmetric[mm]) {
      t_1 <- t(glist[[mm]])
      glist[[mm]][upper.tri(glist[[mm]])] <- t_1[upper.tri(t_1)]
    }
  }

  #    # theta parameterization: delta must be reset!
  #    if((lavmodel@categorical || correlation) && setDelta &&
  #       lavmodel@parameterization == "theta") {
  #        nmat <- lavmodel@nmat
  #        for(g in 1:lavmodel@nblocks) {
  #            # which mm belong to group g?
  #            mm.in.group <- 1:nmat[g] + cumsum(c(0L,nmat))[g]
  #            GLIST[mm.in.group] <-
  #                lav_lisrel_delta(MLIST = GLIST[mm.in.group],
  #                    num.idx = lavmodel@num.idx[[g]])
  #        }
  #    }

  # in 0.6-13: we always set theta/delta
  if ((lavmodel@categorical || correlation) && set_delta) {
    nmat <- lavmodel@nmat
    if (lavmodel@representation == "LISREL") {
      for (g in 1:lavmodel@nblocks) {
        # which mm belong to group g?
        mm_in_group <- 1:nmat[g] + cumsum(c(0L, nmat))[g]

        if (lavmodel@parameterization == "delta") {
          glist[mm_in_group] <-
            lav_lisrel_residual_variances(
              mlist = glist[mm_in_group],
              num_idx = lavmodel@num.idx[[g]],
              ov_y_dummy_ov_idx = lavmodel@ov.y.dummy.ov.idx[[g]],
              ov_y_dummy_lv_idx = lavmodel@ov.y.dummy.lv.idx[[g]]
            )
        } else if (lavmodel@parameterization == "theta") {
          glist[mm_in_group] <-
            lav_lisrel_delta(
              mlist = glist[mm_in_group],
              num_idx = lavmodel@num.idx[[g]]
            )
        }
      } # blocks
    } else {
      cat("FIXME: deal with theta elements in the categorical case (RAM)")
    }
  }

  if (lavmodel@composites) {
    nmat <- lavmodel@nmat
    if (lavmodel@representation == "LISREL") {
      for (g in 1:lavmodel@nblocks) {
        # which mm belong to group g?
        mm_in_group <- 1:nmat[g] + cumsum(c(0L, nmat))[g]

        glist[mm_in_group] <-
          lav_lisrel_comp_set_intresvar(mlist = glist[mm_in_group])
      }
    } else {
      cat("FIXME: deal with Composites when representation = RAM")
    }
  }

  glist
}

# backwards compatibility
# getModelParameters <- lav_model_get_parameters
# setModelParameters <- lav_model_set_parameters
# x2GLIST            <- lav_model_x2glist

# classic score test (= Lagrange Multiplier test)
#
# this function can run in two modes:
#
# MODE 1: 'add'
#   add new parameters that are currently not included in the model
#   (aka fixed to zero), but should be released
#
# MODE 2: 'release' (the default)
#   release existing "==" constraints
#
# - YR 8 Jan 2026: use ceq.JAC instead of con.jac (as we do not support
#                  inequality constraints anyway, and we do not want to
#                  release the EFA constraints)
# - YR 11 Jul 2026: add the scaled, adjusted and generalized ('robust')
#                   versions of the score test (Satorra, 2000); see
#                   lav_test_satorra2000.R

# For ceq.simple models, the simple equality constraints are absorbed (no '=='
# rows; ceq.JAC is empty). Reconstruct an explicit constraint jacobian in the
# 'unco' parameter space (one +1/-1 row per equality, ref == other), together
# with matching lhs/rhs labels, so that lavTestScore() can release them. The
# 'unco' parameter order matches which(partable$free > 0), i.e. the space used
# by lavTech(object, "gradient.logl") and "information.*".
lav_test_score_ceq_simple <- function(object) {
  pt <- object@ParTable
  free_rows <- which(pt$free > 0L) # one entry per 'unco' parameter
  nunco <- length(free_rows)
  comp_idx <- pt$free[free_rows]   # compact (nx.free) index per unco parameter
  # use plabels (unique per parameter) so the two sides of each equality are
  # distinguishable, exactly as for explicit '==' constraints (ceq.simple=FALSE)
  labels <- pt$plabel[free_rows]

  r_list <- list()
  lhs <- character(0L)
  rhs <- character(0L)
  dup_comp <- unique(comp_idx[duplicated(comp_idx)])
  for (k in dup_comp) {
    members <- which(comp_idx == k) # unco positions sharing free index k
    ref <- members[1L]
    for (o in members[-1L]) {
      row <- numeric(nunco)
      row[ref] <- 1
      row[o] <- -1
      r_list[[length(r_list) + 1L]] <- row
      lhs <- c(lhs, labels[ref])
      rhs <- c(rhs, labels[o])
    }
  }
  r_mat <- if (length(r_list) > 0L) {
    do.call(rbind, r_list)
  } else {
    matrix(0, 0L, nunco)
  }
  list(R = r_mat, lhs = lhs, op = rep("==", length(lhs)), rhs = rhs)
}

# the [1:npar, 1:npar] block of the inverse of the information matrix
# bordered with the constraint rows in r1 (the constraints that are KEPT,
# i.e. not released): the 'constrained' inverse that neutralizes the
# released constraints only.
#
# computed via the null-space identity
#   J.inv = Z solve(Z' I Z) Z',  Z = orthonormal basis of null(r1),
# which equals the [1:npar, 1:npar] block of
#   ginv(rbind(cbind(I, t(r1)), cbind(r1, 0)))
# (see also lav_model_info_augment_invert); one solve of order
# npar - nrow(r1) instead of the SVD of the (npar + nrow(r1))^2 bordered
# matrix -- this matters when the constraints are released one at a time.
# if Z'IZ cannot be solved (the model is not identified even with the
# kept constraints), we fall back to the explicit bordered Moore-Penrose
# route
lav_test_score_iinv <- function(information = NULL, r1 = NULL) {
  if (is.null(r1) || nrow(r1) == 0L) {
    return(MASS::ginv(information))
  }
  z <- lav_mat_ortho_complement(t(r1))
  ziz <- crossprod(z, information %*% z)
  out <- try(z %*% solve(ziz, t(z)), silent = TRUE)
  if (inherits(out, "try-error")) {
    z1 <- cbind(
      rbind(information, r1),
      rbind(t(r1), matrix(0, nrow(r1), nrow(r1)))
    )
    z1_plus <- MASS::ginv(z1)
    out <- z1_plus[seq_len(nrow(information)), seq_len(nrow(information))]
  } else {
    out <- (out + t(out)) / 2
  }
  out
}

lavTestScore <- function(object, add = NULL, release = NULL,       # nolint
                         univariate = TRUE, cumulative = FALSE,
                         epc = FALSE, standardized = epc, cov_std = epc,
                         verbose = FALSE, warn = TRUE,
                         information = "expected", ...) {
  dotdotdot <- list(...)
  lav_adapt_func(environment(), dotdotdot, NULL)
  # check object
  object <- lav_object_check_version(object)

  # sam objects: the score test needs the (joint) information/scores, which
  # a two-step (sam) procedure does not provide (issue #517)
  if (!is.null(object@internal$sam.method)) {
    lav_msg_stop(gettext(
      "lavTestScore() is not available for models fitted with sam();
       consider modindices() to inspect the structural part."))
  }

  if (!missing(warn)) {
    current_warn <- lav_warn()
    if (lav_warn(warn))
      on.exit(lav_warn(current_warn))
  }
  if (!missing(verbose)) {
    current_verbose <- lav_verbose()
    if (lav_verbose(verbose))
      on.exit(lav_verbose(current_verbose), TRUE)
  }
  # check object
  stopifnot(inherits(object, "lavaan"))
  lavoptions <- object@Options

  if (object@optim$npar > 0L && !object@optim$converged) {
    lav_msg_stop(gettext("model did not converge"))
  }

  # check for inequality constraints
  pt_1 <- object@ParTable
  if (any(pt_1$op == ">" | pt_1$op == "<")) {
    lav_msg_stop(gettext(
      "lavTestScore() does not handle inequality constraints (yet)"))
  }

  # check arguments
  if (cumulative) {
    univariate <- TRUE
  }

  # PML: the expected (h1) information is not available; use the
  # observed (pairwise) information instead
  if (object@Model@estimator == "PML" && information == "expected") {
    information <- "observed"
  }

  # Mode 1: ADDING new parameters
  if (!is.null(add) && all(nchar(add) > 0L)) {
    # check release argument
    if (!is.null(release)) {
      lav_msg_stop(gettext(
        "`add' and `release' arguments cannot be used together."))
    }

    # extend model with extra set of parameters
    fit <- lav_object_extended(object, add = add)

    # the object providing the ingredients for the robust versions
    # (Satorra, 2000): score/information/meat all live in the parameter
    # space of the extended model
    robust_object <- fit

    score <- lavTech(fit, "gradient.logl")
    information_1 <- lavTech(fit, paste("information", information, sep = "."))

    npar <- object@Model@nx.free
    nadd <- fit@Model@nx.free - npar

    # R
    r_model <- object@Model@ceq.JAC[, , drop = FALSE]
    if (nrow(r_model) > 0L) {
      r_model <- cbind(r_model, matrix(0, nrow(r_model), ncol = nadd))
      r_add <- cbind(matrix(0, nrow = nadd, ncol = npar), diag(nadd))
      r_1 <- rbind(r_model, r_add)

      # keep the model constraints
      j_inv <- lav_test_score_iinv(information_1, r_model)

      r_idx <- seq_len(nadd) + nrow(r_model)
    } else {
      r_1 <- cbind(matrix(0, nrow = nadd, ncol = npar), diag(nadd))
      j_inv <- lav_test_score_iinv(information_1)

      r_idx <- seq_len(nadd)
    }

    # lhs/rhs
    lhs <- lav_pt_labels(fit@ParTable)[fit@ParTable$user == 10L]
    op <- rep("==", nadd)
    rhs <- rep("0", nadd)
    table_1 <- data.frame(
      lhs = lhs, op = op, rhs = rhs,
      stringsAsFactors = FALSE
    )
    class(table_1) <- c("lavaan.data.frame", "data.frame")
  } else {
    # MODE 2: releasing constraints

    robust_object <- object

    r_1 <- object@Model@ceq.JAC[, , drop = FALSE]
    ceq_simple_con <- NULL
    if (nrow(r_1) == 0L && object@Model@ceq.simple.only) {
      # simple equality constraints (ceq.simple) are absorbed: ceq.JAC is empty.
      # reconstruct an explicit constraint jacobian in the 'unco' parameter
      # space (one +1/-1 row per equality), matching the space of the score and
      # information below, so the constraints can be released individually.
      ceq_simple_con <- lav_test_score_ceq_simple(object)
      r_1 <- ceq_simple_con$R
    }
    if (nrow(r_1) == 0L) {
      lav_msg_stop(gettext("no equality constraints found in model."))
    }

    score <- lavTech(object, "gradient.logl")
    information_1 <- lavTech(
      object,
      paste("information", information, sep = ".")
    )
    # R <- object@Model@con.jac[,]

    if (is.null(release)) {
      # ALL constraints
      r_idx <- seq_len(nrow(r_1))
      j_inv <- lav_test_score_iinv(information_1)
    } else if (is.numeric(release)) {
      r_idx <- release
      if (max(r_idx) > nrow(r_1)) {
        lav_msg_stop(gettextf(
          "maximum constraint number (%1$s) is larger than number of
          constraints (%2$s)", max(r_idx), nrow(r_1)))
      }

      # neutralize the non-needed constraints
      j_inv <- lav_test_score_iinv(
        information_1, r_1[-r_idx, , drop = FALSE]
      )
    } else if (is.character(release)) {
      lav_msg_stop(gettext("not implemented yet"))
    }

    # lhs/rhs
    if (!is.null(ceq_simple_con)) {
      lhs <- ceq_simple_con$lhs[r_idx]
      op <- rep("==", length(r_idx))
      rhs <- ceq_simple_con$rhs[r_idx]
    } else {
      eq_idx <- which(object@ParTable$op == "==")
      if (length(eq_idx) > 0L) {
        lhs <- object@ParTable$lhs[eq_idx][r_idx]
        op <- rep("==", length(r_idx))
        rhs <- object@ParTable$rhs[eq_idx][r_idx]
      }
    }
    table_1 <- data.frame(
      lhs = lhs, op = op, rhs = rhs,
      stringsAsFactors = FALSE
    )
    class(table_1) <- c("lavaan.data.frame", "data.frame")
  }

  # PML: lavTech(object, "gradient.logl") returns the TOTAL (summed over
  # cases) pairwise gradient, while the information matrices are on the
  # unit scale (see lav_model_info_firstorder()); rescale the score so
  # that both are on the same (unit) scale
  if (object@Model@estimator == "PML") {
    if (length(object@Data@sampling.weights) == 0L) {
      score <- score / object@SampleStats@ntotal
    } else {
      score <- score / sum(unlist(object@Data@weights))
    }
  }

  if (object@Data@nlevels == 1L) {
    n <- object@SampleStats@ntotal
    if (lavoptions$mimic == "EQS") {
      n <- n - 1
    }
  } else {
    # total number of clusters (over groups)
    n <- 0
    for (g in 1:object@SampleStats@ngroups) {
      n <- n + object@Data@Lp[[g]]$nclusters[[2]]
    }
    # score <- score * (2 * object@SampleStats@ntotal) / N
    score <- score / 2 # -2 * LRT
  }

  # the standard (normal-theory) score test statistic
  #
  # NOTE: we can NOT use VCOV here, because it reflects the constraints,
  # and the whole point is to test for these constraints...
  stat <- as.numeric(n * score %*% j_inv %*% score)

  # compute df, taking into account that some of the constraints may
  # be needed to identify the model (and hence Information is singular)
  # Information.plus <- Information + crossprod(R)
  # df <- qr(R[r_idx,,drop = FALSE])$rank +
  #          ( qr(Information)$rank - qr(Information.plus)$rank )
  df <- nrow(r_1[r_idx, , drop = FALSE])
  pvalue <- 1 - pchisq(stat, df = df)

  # total score test
  test <- data.frame(
    test = "score", X2 = stat, df = df, p.value = pvalue,
    stringsAsFactors = FALSE
  )

  # scaled, adjusted and generalized ('robust') versions (Satorra, 2000),
  # whenever the object provides the ingredients for the sandwich: the
  # (unit-scale) meat B, in the same parameter space as information_1
  b_meat <- lav_test_meat(robust_object)
  if (!is.null(b_meat)) {
    a_mat <- r_1[r_idx, , drop = FALSE]
    aj <- a_mat %*% j_inv
    m1 <- aj %*% t(a_mat)              # A J A'
    m2 <- aj %*% b_meat %*% t(aj)      # A J B J A'
    s2000 <- lav_test_satorra2000(
      stat = stat, df = df, m1 = m1, m2 = m2,
      v = as.numeric(aj %*% score), n = n
    )
    test <- rbind(test, data.frame(
      test = c("score.scaled", "score.adjusted", "score.robust"),
      X2 = c(
        s2000$stat.scaled, s2000$stat.adjusted,
        s2000$stat.robust
      ),
      df = c(s2000$df.scaled, s2000$df.adjusted, s2000$df.robust),
      p.value = c(
        s2000$p.value.scaled, s2000$p.value.adjusted,
        s2000$p.value.robust
      ),
      stringsAsFactors = FALSE
    ))
  } else if (!lavoptions$se %in% c("standard", "none")) {
    lav_msg_warn(gettextf(
      "the scaled, adjusted and robust versions of the score test are not
      available when se = %s; only the standard score test is computed.",
      dQuote(lavoptions$se)))
  }

  class(test) <- c("lavaan.data.frame", "data.frame")
  attr(test, "header") <- "total score test:"

  out <- list(test = test)

  if (univariate) {
    ts_1 <- numeric(nrow(r_1))
    ts_scaled <- rep(as.numeric(NA), nrow(r_1))
    epc_uni <- numeric(nrow(r_1)) # ignored in release= mode
    for (r in r_idx) {
      z1_plus1 <- lav_test_score_iinv(
        information_1, r_1[-r, , drop = FALSE]
      )
      ts_1[r] <- as.numeric(n * t(score) %*% z1_plus1 %*% score)
      if (!is.null(b_meat)) {
        # single restriction: the scaled, adjusted and generalized
        # versions all coincide (Satorra, 2000, eq. 25-28)
        a_r <- r_1[r, , drop = FALSE]
        aj_r <- a_r %*% z1_plus1
        m1_r <- as.numeric(aj_r %*% t(a_r))
        m2_r <- as.numeric(aj_r %*% b_meat %*% t(aj_r))
        if (is.finite(m1_r) && is.finite(m2_r) &&
          m1_r > 0 && m2_r > 0) {
          ts_scaled[r] <- ts_1[r] * m1_r / m2_r
        }
      }
      if (epc && !is.null(add)) {
        # EPC.uni[r] <- -1 * utils::tail(as.numeric(score %*%  Z1.plus1),
        #                               n = nrow(R))[r]
        # to keep the 'sign' consistent with modindices(), which
        # uses epc = 'new - old'
        epc_uni[r] <- +1 * utils::tail(as.numeric(score %*% z1_plus1),
          n = nrow(r_1)
        )[r]
      }
    }

    table2 <- table_1
    table2$X2 <- ts_1[r_idx]
    table2$df <- rep(1, length(r_idx))
    table2$p.value <- 1 - pchisq(table2$X2, df = table2$df)
    if (!is.null(b_meat)) {
      table2$X2.scaled <- ts_scaled[r_idx]
      table2$p.value.scaled <- 1 - pchisq(table2$X2.scaled, df = 1)
    }
    if (epc && !is.null(add)) {
      table2$epc <- epc_uni[r_idx]
    }
    attr(table2, "header") <- "univariate score tests:"
    out$uni <- table2
  }

  if (cumulative) {
    ts_order <- sort.int(ts_1, index.return = TRUE, decreasing = TRUE)$ix
    row_order <-
         sort.int(ts_1[r_idx], index.return = TRUE, decreasing = TRUE)$ix
    ts_1 <- numeric(length(r_idx))
    ts_scaled <- rep(as.numeric(NA), length(r_idx))
    for (r in seq_along(r_idx)) {
      rcumul_idx <- ts_order[1:r]

      z1_plus1 <- lav_test_score_iinv(
        information_1, r_1[-rcumul_idx, , drop = FALSE]
      )
      ts_1[r] <- as.numeric(n * t(score) %*% z1_plus1 %*% score)
      if (!is.null(b_meat)) {
        a_c <- r_1[rcumul_idx, , drop = FALSE]
        aj_c <- a_c %*% z1_plus1
        m1_c <- aj_c %*% t(a_c)
        m2_c <- aj_c %*% b_meat %*% t(aj_c)
        s2000 <- lav_test_satorra2000(
          stat = ts_1[r], df = r,
          m1 = m1_c, m2 = m2_c
        )
        ts_scaled[r] <- s2000$stat.scaled
      }
    }

    table3 <- table_1[row_order, ]
    table3$X2 <- ts_1
    table3$df <- seq_along(ts_1)
    table3$p.value <- 1 - pchisq(table3$X2, df = table3$df)
    if (!is.null(b_meat)) {
      table3$X2.scaled <- ts_scaled
      table3$p.value.scaled <- 1 - pchisq(table3$X2.scaled, df = table3$df)
    }
    attr(table3, "header") <- "cumulative score tests:"
    out$cumulative <- table3
  }

  if (epc) {
    # EPC <- vector("list", length = length(r_idx))
    # for(i in seq_along(r_idx)) {
    #    r <- r_idx[i]
    #    R1 <- R[-r,,drop = FALSE]
    #    Z1 <- cbind( rbind(Information, R1),
    #                 rbind(t(R1), matrix(0,nrow(R1),nrow(R1))) )
    #    Z1.plus <- MASS::ginv(Z1)
    #    Z1.plus1 <- Z1.plus[ 1:nrow(Information), 1:nrow(Information) ]
    #    EPC[[i]] <- -1 * as.numeric(score %*%  Z1.plus1)
    # }
    #
    # OUT$EPC <- EPC

    # alltogether
    z1_plus1 <- lav_test_score_iinv(
      information_1, r_1[-r_idx, , drop = FALSE]
    )
    # EPC.all <- -1 * as.numeric(score %*%  Z1.plus1)
    # to keep the 'sign' consistent with modindices(), which
    # uses epc = 'new - old'
    epc_all <- +1 * as.numeric(score %*% z1_plus1)

    # create epc table for the 'free' parameters
    if (!is.null(add) && all(nchar(add) > 0L)) {
      list_1 <- parTable(fit)
    } else {
      ## release mode
      list_1 <- parTable(object)
    }
    if (lav_pt_ngroups(list_1) == 1L) {
      list_1$group <- NULL
    }
    nonpar_idx <- which(list_1$op %in% c("==", ":=", "<", ">"))
    if (length(nonpar_idx) > 0L) {
      list_1 <- list_1[-nonpar_idx, ]
    }

    list_1$est[list_1$free > 0 & list_1$user != 10] <-
            lav_inspect_coef(object, type = "free")
    list_1$est[list_1$user == 10L] <- 0
    list_1$epc <- rep(as.numeric(NA), length(list_1$lhs))
    list_1$epc[list_1$free > 0] <- epc_all
    list_1$epv <- list_1$est + list_1$epc

    if (standardized) {
      epc_1 <- list_1$epc

      if (cov_std) {
        # replace epc values for variances by est values
        var_idx <- which(list_1$op == "~~" & list_1$lhs == list_1$rhs &
          list_1$exo == 0L)
        epc_1[var_idx] <- list_1$est[var_idx]
      }

      # two problems:
      #   - EPC of variances can be negative, and that is
      #     perfectly legal
      #   - EPC (of variances) can be tiny (near-zero), and we should
      #     not divide by tiny variables
      small_idx <- which(list_1$op == "~~" &
        list_1$lhs == list_1$rhs &
        abs(epc_1) < sqrt(.Machine$double.eps))
      if (length(small_idx) > 0L) {
        epc_1[small_idx] <- as.numeric(NA)
      }

      # get the sign
      epc_sign <- sign(list_1$epc)

      list_1$sepc.lv <- epc_sign * lav_standardize_lv(object,
        partable = list_1,
        est = abs(epc_1),
        cov_std = cov_std
      )
      if (length(small_idx) > 0L) {
        list_1$sepc.lv[small_idx] <- 0
      }
      list_1$sepc.all <- epc_sign * lav_standardize_all(object,
        partable = list_1,
        est = abs(epc_1),
        cov_std = cov_std
      )
      if (length(small_idx) > 0L) {
        list_1$sepc.all[small_idx] <- 0
      }
      list_1$sepc.nox <- epc_sign * lav_standardize_all_nox(object,
        partable = list_1,
        est = abs(epc_1),
        cov_std = cov_std
      )
      if (length(small_idx) > 0L) {
        list_1$sepc.nox[small_idx] <- 0
      }
    }

    list_1$free[list_1$user == 10L] <- 0
    list_1$user <- NULL
    # remove some more columns
    list_1$id <- list_1$ustart <- list_1$exo <-
           list_1$start <- list_1$se <- list_1$prior <- NULL
    if (lav_pt_nblocks(list_1) == 1L) {
      list_1$block <- NULL
      list_1$group <- NULL
      list_1$level <- NULL
    }

    attr(list_1, "header") <-
      "expected parameter changes (epc) and expected parameter values (epv):"

    out$epc <- list_1
  }

  out
}

\name{lavaan-class}
\docType{class}
\alias{lavaan-class}
\alias{coef,lavaan-method}
\alias{vcov,lavaan-method}
\alias{anova,lavaan-method}
\alias{predict,lavaan-method}
\alias{resid,lavaan-method}
\alias{residuals,lavaan-method}
\alias{fitted,lavaan-method}
\alias{fitted.values,lavaan-method}    %% NOTE: no white space here!
\alias{nobs} %% not imported in < 2.13
\alias{nobs,lavaan-method}
\alias{logLik,lavaan-method}
\alias{update,lavaan-method}
\alias{show,lavaan-method}
\alias{summary,lavaan-method}

\title{Class For Representing A (Fitted) Latent Variable Model}
\description{The \code{lavaan} class represents a (fitted) latent variable
  model. It contains a description of the model as specified by the user,
  a summary of the data, an internal matrix representation, and if the model
  was fitted, the fitting results.}
\section{Objects from the Class}{
Objects can be created via the
  \code{\link{cfa}}, \code{\link{sem}}, \code{\link{growth}} or
  \code{\link{lavaan}} functions.
}
\section{Slots}{
  \describe{
    \item{\code{version}:}{The lavaan package version used to create this object}
    \item{\code{call}:}{The function call as returned by \code{match.call()}.}
    \item{\code{timing}:}{The elapsed time (user+system) for various parts of
      the program as a list, including the total time.}
    \item{\code{Options}:}{Named list of options that were provided by
      the user, or filled-in automatically.}
    \item{\code{ParTable}:}{Named list describing the model parameters. Can be coerced to a data.frame. In the documentation, this is called the `parameter table'.}
    \item{\code{pta}:}{Named list containing parameter table attributes.}
    \item{\code{Data}:}{Object of internal class \code{"Data"}: information
about the data.}
    \item{\code{SampleStats}:}{Object of internal class \code{"SampleStats"}: sample
      statistics}
    \item{\code{Model}:}{Object of internal class \code{"Model"}: the
      internal (matrix) representation of the model}
    \item{\code{Cache}:}{List of objects that are computed only once and reused many times.}
    \item{\code{Fit}:}{Object of internal class \code{"Fit"}: the
      results of fitting the model. No longer used.}
    \item{\code{boot}:}{List. Results and information about the bootstrap.}
    \item{\code{optim}:}{List. Information about the optimization.}
    \item{\code{loglik}:}{List. Information about the loglikelihood of the model (if maximum likelihood was used).}
    \item{\code{implied}:}{List. Model implied statistics.}
    \item{\code{vcov}:}{List. Information about the variance matrix (vcov) of the model parameters.}
    \item{\code{test}:}{List. Different test statistics.}
    \item{\code{h1}:}{List. Information about the unrestricted h1 model (if available).}
    \item{\code{baseline}:}{List. Information about a baseline model (often
the independence model) (if available).}
    \item{\code{internal}:}{List. For internal use only.}
    \item{\code{external}:}{List. Empty slot to be used by add-on packages.}
  }
}
\section{Methods}{
  \describe{
    \item{coef}{\code{signature(object = "lavaan", type = "free")}: Returns
      the estimates of the parameters in the model as a named numeric vector.
      If \code{type="free"}, only the free parameters are returned.
      If \code{type="user"}, all parameters listed in the parameter table
      are returned, including constrained and fixed parameters.}
    \item{fitted.values}{\code{signature(object = "lavaan")}: Returns the
      implied moments of the model as a list with two elements (per group):
      \code{cov} for the implied covariance matrix,
      and \code{mean} for the implied mean
      vector. If only the covariance matrix was analyzed, the implied mean
      vector will be zero.}
    \item{fitted}{\code{signature(object = "lavaan")}: an alias for
        \code{fitted.values}.}
    \item{residuals}{\code{signature(object = "lavaan", type="raw")}:
	  If \code{type = "raw"}, this function returns the raw (= unscaled)
      difference between the observed and the expected (model-implied) summary
      statistics.
	  If \code{type = "cor"}, or \code{type = "cor.bollen"}, the observed and
      model implied covariance matrices are first transformed to a correlation
      matrix (using \code{cov2cor()}), before the residuals are computed.
	  If \code{type = "cor.bentler"}, both the observed and model implied
      covariance matrices are rescaled by dividing the elements by the square
      roots of the corresponding variances of the observed covariance matrix.
      If \code{type="normalized"}, the residuals are divided by the square
      root of the asymptotic variance of the corresponding summary statistic
      (the variance estimate depends on the choice for the \code{se} argument).
      Unfortunately, these normalized residuals are not entirely
      correct, and this option is retained only for historical interest.
      If \code{type="standardized"}, the residuals are divided by the square
      root of the asymptotic variance of these residuals. The resulting
      standardized residuals can be interpreted as z-scores.
      If \code{type="standardized.mplus"}, the residuals are divided by the
      square root of the asymptotic variance of these residuals. However, a
      simplified formula is used (see the Mplus reference below) which often
      results in negative estimates for the variances, resulting in many
      \code{NA} values for the standardized residuals.}
    \item{resid}{\code{signature(object = "lavaan")}: an alias
      for \code{residuals}}
    \item{vcov}{\code{signature(object = "lavaan", type = "free")}: returns
      the covariance matrix of the estimated parameters. If
      \code{type = "free"} (the default), only the free parameters are
      included. If \code{type = "user"} (or \code{"joint"}), the joint
      covariance matrix of the free and the user-defined (\code{:=})
      parameters is returned; the parameter names are the same as those
      used by \code{coef()}.}
    \item{predict}{\code{signature(object = "lavaan")}: compute
      factor scores for all cases that are provided in the data frame. For
      complete data only.}
    \item{anova}{\code{signature(object = "lavaan")}: returns
      model comparison statistics. This method is just a wrapper around
      the function \code{\link{lavTestLRT}}.
      If only a single argument (a fitted
      model) is provided, this model is compared to the unrestricted
      model. If two or more arguments (fitted models) are provided, the models
      are compared in a sequential order. Test statistics are based on the
      likelihood ratio test. For more details and
      further options, see the \code{\link{lavTestLRT}} page.}
    \item{update}{\code{signature(object = "lavaan", model, add, ...,
        evaluate = TRUE)}: update a fitted lavaan object and evaluate it
        (unless \code{evaluate = FALSE}). Note that we use the environment
        that is stored within the lavaan object, which is not necessarily
        the parent frame. The \code{add} argument is analogous to the one
        described in the \code{\link{lavTestScore}} page, and can be used to
        add parameters to the specified model rather than passing an entirely
        new \code{model} argument.}
    \item{nobs}{\code{signature(object = "lavaan")}: returns the effective
      number of observations used when fitting the model. In a multiple group
      analysis, this is the sum of all observations per group.}
    \item{logLik}{\code{signature(object = "lavaan")}:
      returns the log-likelihood of the fitted model, if maximum likelihood estimation
      was used. The \code{\link[stats]{AIC}} and \code{\link[stats]{BIC}}
      methods automatically work via \code{logLik()}.}
    \item{show}{\code{signature(object = "lavaan")}: Print a short summary
      of the model fit}
    \item{summary}{\code{signature(object = "lavaan", header = TRUE,
     fit_measures = FALSE, residuals = FALSE,
     estimates = TRUE, ci = FALSE, fmi = FALSE,
     standardized = FALSE, std = standardized, std_nox = FALSE,
     remove_system_eq = TRUE, remove_eq = TRUE, remove_ineq = TRUE,
     remove_def = FALSE, remove_nonfree = FALSE, remove_step1 = TRUE,
     remove_unused = TRUE, plabel = FALSE,
     cov_std = TRUE, rsquare = FALSE,
     baseline_model = NULL, h1_model = NULL,
     fm_args = list(standard.test = "default", scaled.test = "default",
     rmsea.ci.level = 0.90, rmsea.h0.closefit = 0.05,
     rmsea.h0.notclosefit = 0.08, robust = TRUE, cat.nonpd = "na"),
     modindices = FALSE, srmr_close_h0 = NULL,
     nd = 3L, cutoff = 0.3, dot_cutoff = 0.1)}:
      Print a nice summary of the model estimates.
      If \code{header = TRUE}, the header section (including fit measures) is
      printed.
      If \code{fit.measures = TRUE}, additional fit measures are added to the
      header section. \code{fit.measures} can also be a list, which allows one to set options
      related to the fit measures. See \code{\link{fitMeasures}}
      for more details.
      If \code{residuals = TRUE}, a residuals section is added (the largest
      residuals and the residual summary, printed as by
      \code{\link{lavResiduals}} with \code{output = "text"}); the underlying
      \code{lavResiduals(object, output = "list")} result is stored as the
      \code{residuals} element of the summary object. \code{residuals} can also
      be a list of arguments passed on to \code{\link{lavResiduals}} (for
      example \code{residuals = list(type = "raw", n.largest = 10)}).
      If both \code{fit.measures = TRUE} and \code{residuals = TRUE} and the
      residual summary is the SRMR (the default), the SRMR is removed from the
      fit measures section to avoid printing it twice.
      If \code{estimates = TRUE}, print the parameter estimates section.
      If \code{ci = TRUE}, add confidence intervals to the parameter estimates
      section.
      If \code{fmi = TRUE}, add the fmi (fraction of missing information)
      column, if it is available.
      If \code{standardized = TRUE} or a character vector, the standardized
      solution is also printed (see \code{\link{lavParameterEstimates}}).
      Note that \emph{SE}s and
      tests are still based on unstandardized estimates. Use
      \code{\link{standardizedSolution}} to obtain \emph{SE}s and test
      statistics for standardized estimates.
      The \code{std.nox} argument is deprecated; the \code{standardized}
      argument allows \code{"std.nox"} solution to be specifically requested.
      The \code{standardized} argument may also be a character vector of
      (observed) variable names (for example \code{c("x1", "x2")}); only the
      parameters involving these variables are then standardized, and the
      result is shown in a \code{Std.usr} column. This generalizes
      \code{"std.nox"}, where the exogenous \code{x} variables are the ones
      left unstandardized.
      If \code{remove_system_eq = TRUE}, the system-generated equality
      constraints (using plabels) are not shown.
      If \code{remove_eq = TRUE} or \code{remove_ineq = TRUE},
      the user-specified (in)equality constraints are not shown.
      If \code{remove_def = TRUE}, the user-specified parameter definitions are
      not shown.
      If \code{remove_nonfree = TRUE}, the nonfree parameters are not shown.
      If \code{remove_step1 = TRUE}, the parameters of the measurement part
      are not shown (only used when using \code{sam()}.)
      If \code{remove_unused = TRUE}, automatically added parameters that are
      fixed to their default (0 or 1) values are removed.
      If \code{rsquare = TRUE}, the R-Square values for the dependent variables
      in the model are printed.
      The \code{fm.args} list allows one to set options related to the fit
      measures (see \code{\link{fitMeasures}}).
      \code{baseline.model} and \code{h1.model} allow one to specify user-defined
      baseline and/or h1 models for the fit measures.
      If the model is an exploratory factor analysis (EFA) model, EFA related
      information is printed automatically. The \code{cutoff} and
      \code{dot.cutoff} arguments control how the factor loadings are displayed:
      loadings whose absolute value is at or above \code{cutoff} are shown as
      numbers; loadings whose absolute value falls between \code{dot.cutoff}
      and \code{cutoff} are replaced by a dot; loadings whose absolute value is
      below \code{dot.cutoff} are left blank.
      If \code{modindices = TRUE} or \code{modindices} is a list,
      modification indices are printed for all fixed parameters.
      The argument \code{nd} determines the number of digits after the
      decimal point to be printed (currently only in the parameter estimates
      section.) Historically, nothing was returned, but since 0.6-12, a
      list is returned of class \code{lavaan.summary} for which a print
      function is available.}
  }
}
\references{
Yves Rosseel (2012). lavaan: An R Package for Structural Equation
Modeling. Journal of Statistical Software, 48(2), 1-36.
\doi{10.18637/jss.v048.i02}

Standardized Residuals in Mplus. Document retrieved from URL
https://www.statmodel.com/download/StandardizedResiduals.pdf
}
\seealso{
\code{\link{cfa}}, \code{\link{sem}},
\code{\link{fitMeasures}}, \code{\link{standardizedSolution}},
\code{\link{lavParameterEstimates}}, \code{\link{lavInspect}},
\code{\link{modindices}}
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data = HolzingerSwineford1939)

summary(fit, standardized = TRUE, fit.measures = TRUE, rsquare = TRUE)
fitted(fit)
coef(fit)
resid(fit, type = "normalized")
}

\name{lavaan}
\alias{lavaan}
\title{Fit a Latent Variable Model}
\description{
Fit a latent variable model.}
\usage{
lavaan(model = NULL, data = NULL, ordered = NULL, aux = NULL,
       sampling_weights   = NULL,
       sample_cov = NULL, sample_mean = NULL, sample_th = NULL,
       sample_nobs = NULL,
       group = NULL, cluster = NULL, constraints = "", 
       wls_v = NULL, nacov = NULL, ov_order = "model",
       slot_options = NULL, slot_par_table = NULL, slot_sample_stats = NULL,
       slot_data = NULL, slot_model = NULL, slot_cache = NULL,
       sloth1 = NULL,
       ...)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See 
    \code{\link{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}
\item{data}{An optional data frame containing the observed variables used in
    the model. If some variables are declared as ordered factors, lavaan will
    treat them as ordinal variables.}
\item{ordered}{Character vector. Only used if the data is in a data.frame.
    Treat these variables as ordered (ordinal) variables, if they are
    endogenous in the model. Importantly, all other variables will be treated
    as numeric (unless they are declared as ordered in the data.frame.) 
    Since 0.6-4, ordered can also be logical. If TRUE, all
    observed endogenous variables are treated as ordered (ordinal).
    If FALSE, all observed endogenous variables are considered to be numeric
    (again, unless they are declared as ordered in the data.frame.)}
\item{aux}{Character vector. Names of auxiliary observed variables. Auxiliary
    variables are not part of the user-specified model; they are used to make
    the missing-at-random (MAR) assumption more plausible under missing data.
    Only available (for now) with continuous data and one of
    \code{missing = "ml"}, \code{"two.stage"} or \code{"robust.two.stage"}.
    With \code{missing = "ml"} (full-information ML), the auxiliary variables
    are added to the model as a block of `saturated correlates' (Graham, 2003):
    each auxiliary variable is freely correlated with every other auxiliary
    variable, with every model observed variable, and has a free mean. The
    (enlarged) model is estimated with casewise FIML, so the auxiliary
    variables affect both the point estimates and the standard errors, while
    leaving the model degrees of freedom unchanged. The baseline model is
    adjusted accordingly (so CFI/TLI remain correct), and the auxiliary
    (nuisance) parameters are hidden from the default output (use
    \code{remove_aux = FALSE} in \code{parameterEstimates()} to show them; they
    are always part of \code{coef()}). With \code{missing = "two.stage"} or
    \code{"robust.two.stage"}, the auxiliary variables are instead included in
    the first-stage EM run that estimates the saturated summary statistics the
    model is fitted to, so they also affect the model estimates; the two-stage
    standard errors account for the auxiliary variables via the augmented
    first-stage asymptotic covariance (Savalei & Bentler, 2009), except when
    fixed-x covariates are present (in which case they fall back to the
    model-only covariance). Binary or ordered/categorical auxiliary variables
    are not supported and are removed (with a warning).}
\item{sampling_weights}{A variable name in the data frame containing
    sampling weight information. Currently only available for non-clustered
    data. Depending on the \code{sampling_weights.normalization} option, these
    weights may be rescaled (or not) so that their sum equals the number of
    observations (total or per group).}
\item{sample_cov}{Numeric matrix. A sample variance-covariance matrix.
    The rownames and/or colnames must contain the observed variable names.
    For a multiple group analysis, a list with a variance-covariance matrix
    for each group. If \code{conditional.x = TRUE}, this is the
    \emph{residual} variance-covariance matrix (given the exogenous
    covariates), and the regression intercepts, slopes and covariate moments
    are provided through the attributes \code{res.slopes}, \code{cov.x} and
    \code{mean.x} of \code{sample_cov} (with \code{sample_mean} holding the
    residual intercepts). It is recommended to supply these attributes with
    (dim)names, so that they can be matched against the model's internal
    variable order; unnamed attributes are assumed to be in that internal
    order already.}
\item{sample_mean}{A sample mean vector. For a multiple group analysis,
    a list with a mean vector for each group.}
\item{sample_th}{Vector of sample-based thresholds. For a multiple group
    analysis, a list with a vector of thresholds for each group.}
\item{sample_nobs}{Number of observations if the full data frame is missing
    and only sample moments are given. For a multiple group analysis, a list
    or a vector with the number of observations for each group.}
\item{group}{Character. A variable name in the data frame defining the groups
in a multiple group analysis.}
\item{cluster}{Character. A (single) variable name in the data frame defining the clusters in a two-level dataset.}
\item{constraints}{Additional (in)equality constraints not yet included in the
    model syntax. See \code{\link{model.syntax}} for more information.}
\item{wls_v}{A user provided weight matrix to be used by estimator \code{"WLS"};
    if the estimator is \code{"DWLS"}, only the diagonal of this matrix will be
    used. For a multiple group analysis, a list with a weight matrix
    for each group. The elements of the weight matrix should be in the
    following order (if all data is continuous): first the means (if a
    meanstructure is involved), then the lower triangular elements of the
    covariance matrix including the diagonal, ordered column by column. In
    the categorical case: first the thresholds (including the means for
    continuous variables), then the slopes (if any), the variances of
    continuous variables (if any), and finally the lower triangular elements
    of the correlation/covariance matrix excluding the diagonal, ordered
    column by column.}
\item{nacov}{A user provided matrix containing the elements of (N times)
    the asymptotic variance-covariance matrix of the sample statistics.
    For a multiple group analysis, a list with an asymptotic
    variance-covariance matrix for each group. See the \code{wls_v}
    argument for information about the order of the elements.}
\item{ov_order}{Character. If \code{"model"} (the default), the order of
    the observed variable names (as reflected for example in the output of
    \code{lav_object_vnames()}) is determined by the model syntax. If
    \code{"data"}, the order is determined by the data (either the full
    data.frame or the sample (co)variance matrix). If the \code{wls_v} 
    and/or \code{nacov} matrices are provided, this argument is currently 
    set to \code{"data"}.}
\item{slot_options}{Options slot from a fitted lavaan object. If provided,
    no new Options slot will be created by this call.}
\item{slot_par_table}{ParTable slot from a fitted lavaan object. If provided,
    no new ParTable slot will be created by this call.}
\item{slot_sample_stats}{SampleStats slot from a fitted lavaan object. 
    If provided, no new SampleStats slot will be created by this call.}
\item{slot_data}{Data slot from a fitted lavaan object. If provided,
    no new Data slot will be created by this call.}
\item{slot_model}{Model slot from a fitted lavaan object. If provided,
    no new Model slot will be created by this call.}
\item{slot_cache}{Cache slot from a fitted lavaan object. If provided,
    no new Cache slot will be created by this call.}
\item{sloth1}{h1 slot from a fitted lavaan object. If provided,
    no new h1 slot will be created by this call.}
\item{...}{Many more options can be specified, using 'name = value'.
    See \code{\link{lavOptions}} for a complete list.}
}
\value{
  An object of class \code{\linkS4class{lavaan}}, for which several methods
  are available, including a \code{summary} method.
}
\references{
Yves Rosseel (2012). lavaan: An R Package for Structural Equation
Modeling. Journal of Statistical Software, 48(2), 1-36. 
\doi{10.18637/jss.v048.i02}}
\seealso{
\code{\link{cfa}}, \code{\link{sem}}, \code{\link{growth}}
}
\examples{
# The Holzinger and Swineford (1939) example
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- lavaan(HS.model, data=HolzingerSwineford1939,
              auto.var=TRUE, auto.fix.first=TRUE,
              auto.cov.lv.x=TRUE)
summary(fit, fit.measures=TRUE)
}

\name{lavTestLRT}
\alias{lavTestLRT}
\alias{lavtestLRT}
\alias{LRT}
\alias{lavLRTTest}
\alias{lavLRT}
\alias{anova}
\title{LRT test}
\description{
LRT test for comparing (nested) lavaan models.}
\usage{
lavTestLRT(object, ..., method = "default", test = "default",
           a_method = "delta", scaled_shifted = TRUE,
           type = "Chisq", model_names = NULL)
anova(object, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{...}{additional objects of class \code{\linkS4class{lavaan}}.}
\item{method}{Character string. The possible options are
  \code{"satorra.bentler.2001"}, \code{"satorra.bentler.2010"},
  \code{"satorra.2000"}, and \code{"standard"}. See details.
  Note that the robust options (\code{"satorra.bentler.2001"},
  \code{"satorra.bentler.2010"} and \code{"satorra.2000"}) require that the
  models were fitted with a robust (scaled) test statistic (for example, using
  \code{estimator = "MLM"} or \code{test = "satorra.bentler"}). If none of the
  models were fitted with a robust test statistic, but a robust \code{method=}
  is requested, a warning is issued, the \code{method=} argument is ignored,
  and a standard (regular) chi-squared difference test is computed instead.}
\item{test}{Character string specifying which scaled test statistics to use,
  in case multiple scaled \code{test=} options were requested when fitting the
  model(s). See details. FMG nested tests can be requested using names such as
  \code{"pall"}, \code{"all"}, \code{"peba4"}, or \code{"pols3"}; the
  convenience value \code{"fmg"} resolves to the recommended nested default.}
\item{a_method}{Character string. The possible options are \code{"exact"}
  and \code{"delta"}. This is only used when method = \code{"satorra.2000"}.
  It determines how the Jacobian of the constraint function (the matrix A)
  will be computed. Note that if \code{a_method = "exact"}, the models must
  be nested in the parameter sense, while if \code{a_method = "delta"}, they
  only need to be nested in the covariance matrix sense.}
\item{scaled_shifted}{Logical. Only used when method = \code{"satorra.2000"}.
  If \code{TRUE}, we use a scaled and shifted test statistic. If \code{FALSE},
  the statistic depends on the test that was used to fit the models: when the
  models were fitted with a mean-and-variance adjusted test
  (\code{test = "mean.var.adjusted"} or \code{"scaled_shifted"}), a mean and
  variance adjusted (Satterthwaite style) difference test is used; when they
  were fitted with a (Satorra-)Bentler scaled test
  (\code{test = "satorra.bentler"}, \code{"yuan.bentler"} or
  \code{"yuan.bentler.mplus"}), a scaled (non-shifted) difference test is used.
  The title printed above the anova table reflects the statistic that is
  actually used.}
\item{type}{Character. If \code{"Chisq"}, the test statistic for each
  model is the (scaled or unscaled) model fit test statistic. If \code{"Cf"},
  the test statistic for each model is computed by the
  \code{\link{lavTablesFitCf}} function. If \code{"browne.residual.adf"} (alias
  \code{"browne"}) or \code{"browne.residual.nt"}, the standard chi-squared
  difference is calculated from each model's residual-based statistic.}
\item{model_names}{Character vector. If provided, use these model names in
  the first column of the anova table.}
}
\value{
    An object of class anova. When given a single argument, it simply
    returns the test statistic of this model. When given a sequence of
    objects, this function tests the models against one another, after
    reordering the models according to their degrees of freedom.
}
\details{
    The \code{anova} function for lavaan objects simply calls the
    \code{lavTestLRT} function, which has a few additional arguments.

    The only \code{test=} options that currently have actual consequences are
    \code{"satorra.bentler"}, \code{"yuan.bentler"}, or \code{"yuan.bentler.mplus"}
    because \code{"mean.var.adjusted"} and \code{"scaled_shifted"} are
    currently distinguished by the \code{scaled_shifted} argument.
    See \code{\link{lavOptions}} for details about the \code{test=} options
    implied by robust \code{estimator=} options. The \code{"default"} is to
    select the first available scaled statistic, if any. To check which test(s)
    were computed when fitting your model(s), use
    \code{lavInspect(fit, "options")$test}.

    If \code{type = "Chisq"} and the test statistics are scaled, a
    special scaled difference test statistic is computed. If \code{method} is
    \code{"satorra.bentler.2001"}, a simple approximation is used
    described in Satorra & Bentler (2001). In some settings,
    this can lead to a negative test statistic. To ensure a positive
    test statistic, we can use the method proposed by
    Satorra & Bentler (2010). Alternatively, when \code{method="satorra.2000"},
    the original formulas of Satorra (2000) are used. The latter is used for
    model comparison when \code{...} contains additional (nested) models.
    Even when test statistics are scaled in \code{object} or \code{...},
    users may request the  \code{method="standard"} test statistic,
    without a robust adjustment.

    Conversely, a robust \code{method=} (\code{"satorra.bentler.2001"},
    \code{"satorra.bentler.2010"} or \code{"satorra.2000"}) can only be applied
    when the models were fitted with a robust (scaled) test statistic. If none
    of the models in \code{object} or \code{...} were fitted with a robust test
    statistic, a robust \code{method=} cannot be honored: in that case a warning
    is issued, the \code{method=} argument is ignored, and a standard (regular)
    chi-squared difference test is returned.

    FMG nested tests are available when \code{type = "Chisq"} and
    \code{test} is an FMG test name. The supported nested FMG methods are
    \code{"pall"}, \code{"all"}, \code{"peba"} and \code{"pols"} variants.
    The convenience value \code{"fmg"} resolves to \code{"pall_ug_rls"} for
    single-group comparisons and \code{"pall_ug_ml"} for multiple-group
    comparisons.
    They use the Satorra (2000) UGamma projection; \code{method} may be left
    at \code{"default"} or set to \code{"standard"} or
    \code{"satorra.2000"}. Other nested LRT methods are not used for FMG
    tests.
}
\note{
  If there is a \code{\linkS4class{lavaan}} model stored in
  \code{object@external$h1.model}, it will be added to \code{\dots}
}
\references{
Satorra, A. (2000). Scaled and adjusted restricted tests in multi-sample
analysis of moment structures. In Heijmans, R.D.H., Pollock, D.S.G. & Satorra,
A. (eds.), \emph{Innovations in multivariate statistical analysis}:
\emph{A Festschrift for Heinz Neudecker} (pp.233-247).
London, UK: Kluwer Academic Publishers.

Satorra, A., & Bentler, P. M. (2001). A scaled difference chi-square test
statistic for moment structure analysis. \emph{Psychometrika, 66}(4), 507-514.
\doi{10.1007/BF02296192}

Satorra, A., & Bentler, P. M. (2010). Ensuring positiveness of the scaled
difference chi-square test statistic. \emph{Psychometrika, 75}(2), 243-248.
\doi{10.1007/s11336-009-9135-y}

Foldnes, N., Moss, J., & Gronneberg, S. (2024). Improved goodness of fit
procedures for structural equation models. \emph{Structural Equation Modeling:
A Multidisciplinary Journal}, 1-13. \doi{10.1080/10705511.2024.2372028}

Foldnes, N., Gronneberg, S., & Moss, J. (2026). Penalized eigenvalue block
averaging: Extension to nested model comparison and Monte Carlo evaluations.
\emph{Behavior Research Methods, 58}(4). \doi{10.3758/s13428-026-02968-4}
}
\examples{
HS.model <- '
    visual  =~ x1 + b1*x2 + x3
    textual =~ x4 + b2*x5 + x6
    speed   =~ x7 + b3*x8 + x9
'
fit1 <- cfa(HS.model, data = HolzingerSwineford1939)
fit0 <- cfa(HS.model, data = HolzingerSwineford1939,
            orthogonal = TRUE)
lavTestLRT(fit1, fit0)


## When multiple test statistics are selected when the model is fitted,
## use the type= and test= arguments to select a test for comparison.

## refit models, requesting 6 test statistics (in addition to "standard")
t6.1 <- cfa(HS.model, data = HolzingerSwineford1939,
            test = c("browne.residual.adf","scaled.shifted","mean.var.adjusted",
                     "satorra.bentler", "yuan.bentler", "yuan.bentler.mplus"))
t6.0 <- cfa(HS.model, data = HolzingerSwineford1939, orthogonal = TRUE,
            test = c("browne.residual.adf","scaled.shifted","mean.var.adjusted",
                     "satorra.bentler", "yuan.bentler", "yuan.bentler.mplus"))

## By default (test="default", type="Chisq"), the first scaled statistic
## requested will be used. Here, that is "scaled.shifted"
lavTestLRT(t6.1, t6.0)
## But even if "satorra.bentler" were requested first, method="satorra.2000"
## provides the scaled-shifted chi-squared difference test:
lavTestLRT(t6.1, t6.0, method = "satorra.2000")
## == lavTestLRT(update(t6.1, test = "scaled.shifted"), update(t6.0, test = "scaled.shifted"))

## The mean- and variance-adjusted (Satterthwaite) statistic implies
## scaled_shifted = FALSE
lavTestLRT(t6.1, t6.0, method = "satorra.2000", scaled_shifted = FALSE)

## Because "satorra.bentler" is not the first scaled test in the list,
## we MUST request it explicitly:
lavTestLRT(t6.1, t6.0, test = "satorra.bentler") # method="satorra.bentler.2001"
## == lavTestLRT(update(t6.1, test = "satorra.bentler"),
##               update(t6.0, test = "satorra.bentler"))
## The "strictly-positive test" is necessary when the above test is < 0:
lavTestLRT(t6.1, t6.0, test = "satorra.bentler", method = "satorra.bentler.2010")

## Likewise, other scaled statistics can be selected:
lavTestLRT(t6.1, t6.0, test = "yuan.bentler")
## == lavTestLRT(update(t6.1, test = "yuan.bentler"),
##               update(t6.0, test = "yuan.bentler"))
lavTestLRT(t6.1, t6.0, test = "yuan.bentler.mplus")
## == lavTestLRT(update(t6.1, test = "yuan.bentler.mplus"),
##               update(t6.0, test = "yuan.bentler.mplus"))

## To request the difference between Browne's (1984) residual-based statistics,
## rather than statistics based on the fitted model's discrepancy function,
## use the type= argument:
lavTestLRT(t6.1, t6.0, type = "browne.residual.adf")

## Despite requesting multiple robust tests, it is still possible to obtain
## the standard chi-squared difference test (i.e., without a robust correction)
lavTestLRT(t6.1, t6.0, method = "standard")
## == lavTestLRT(update(t6.1, test = "standard"), update(t6.0, test = "standard"))

## FMG nested p-values use the Satorra (2000) UGamma projection
lavTestLRT(fit1, fit0, method = "satorra.2000", test = "pall")

}

\name{lavParameterEstimates}
\alias{lavParameterEstimates}
\alias{parameterEstimates}
\alias{parameterestimates}
\title{Parameter Estimates}
\description{
Parameter estimates of a latent variable model.}
\usage{
lavParameterEstimates(object, 
                   se = TRUE, zstat = TRUE, pvalue = TRUE, ci = TRUE, 
                   standardized = FALSE, 
                   fmi = FALSE, plabel = FALSE,
                   level = 0.95, boot_ci_type = "perc", 
                   cov_std = TRUE, fmi_options = list(), 
                   rsquare = FALSE, 
                   remove_system_eq = TRUE, remove_eq = TRUE, 
                   remove_ineq = TRUE, remove_def = FALSE, 
                   remove_nonfree = FALSE, remove_step1 = TRUE,
                   remove_unused = FALSE,
                   remove_aux = TRUE, add_attributes = FALSE, 
                   output = "data.frame", header = FALSE, ...)
parameterEstimates(object, 
                   se = TRUE, zstat = TRUE, pvalue = TRUE, ci = TRUE, 
                   standardized = FALSE, 
                   fmi = FALSE, plabel = FALSE,
                   level = 0.95, boot_ci_type = "perc", 
                   cov_std = TRUE, fmi_options = list(), 
                   rsquare = FALSE, 
                   remove_system_eq = TRUE, remove_eq = TRUE, 
                   remove_ineq = TRUE, remove_def = FALSE, 
                   remove_nonfree = FALSE, remove_step1 = TRUE,
                   remove_unused = FALSE,
                   remove_aux = TRUE, add_attributes = FALSE, 
                   output = "data.frame", header = FALSE, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{se}{Logical. If \code{TRUE}, include column containing the standard
errors. If \code{FALSE}, this implies that \code{zstat}, \code{pvalue}, and
\code{ci} are also \code{FALSE}.}
\item{zstat}{Logical. If \code{TRUE}, an extra column is added containing
the so-called z-statistic, which is simply the value of the estimate divided
by its standard error.}
\item{pvalue}{Logical. If \code{TRUE}, an extra column is added containing
the pvalues corresponding to the z-statistic, evaluated under a standard
normal distribution.}
\item{ci}{If \code{TRUE}, confidence intervals are added to the output.}
\item{level}{The confidence level required.}
\item{plabel}{Logical. If \code{TRUE}, show the plabel column of the
  parameter table in the output.}
\item{boot_ci_type}{If bootstrapping was used, the type of interval required.
  The value should be one of \code{"norm"}, \code{"basic"}, \code{"perc"},
  or \code{"bca.simple"}. For the first three options, see the help page of
  the \code{boot.ci} function in the boot package. The
  \code{"bca.simple"} option produces intervals using the adjusted bootstrap
  percentile (BCa) method, but with no correction for acceleration (only for
  bias). Note that the p-value is still computed assuming that the z-statistic
  follows a standard normal distribution.}
\item{standardized}{Logical or character. If \code{TRUE}, standardized estimates
are added to the output. Note that \emph{SE}s and tests are still based on
unstandardized estimates. Use \code{\link{standardizedSolution}} to obtain
\emph{SE}s and test statistics for standardized estimates. If a character
vector is passed with any of \code{c("std.lv","std.all","std.nox")}, only the
selected standardization methods are added. Note that \code{"std.nox"} only
differs from \code{"std.all"} if \code{fixed.x = TRUE}; if
\code{fixed.x = FALSE}, the exogenous covariates are treated as random
variables (and standardized) just like any other variable, and a warning
is issued. Alternatively, a character vector
of (observed) variable names may be passed (for example
\code{standardized = c("x1", "x2")}); in that case only the parameters
involving these variables are standardized (the other observed variables are
left unstandardized), and the result is added in a \code{std.user} column.
This is a generalization of \code{"std.nox"}, where the (observed) exogenous
\code{x} variables are the ones left unstandardized.}
\item{cov_std}{Logical. If TRUE, the (residual) observed
covariances are scaled by the square root of the `Theta' diagonal elements, and
the (residual) latent covariances are scaled by the square root of the `Psi'
diagonal elements. If FALSE, the (residual) observed covariances
are scaled by the square root of the diagonal elements of the observed
model-implied covariance matrix (Sigma), and the (residual) latent covariances
are scaled by the square root of diagonal elements of the model-implied
covariance matrix of the latent variables.}
\item{fmi}{Logical. If \code{TRUE}, an extra column is added containing the
fraction of missing information for each estimated parameter. Only
available if
\code{estimator="ML"}, \code{missing="(fi)ml"}, and \code{se="standard"}.
See references for more information.}
\item{fmi_options}{List. If non-empty, arguments can be provided to alter
the default options when the model is fitted with the complete(d) data;
otherwise, the same options are used as the original model.}
\item{remove_eq}{Logical. If \code{TRUE}, filter the output by removing all
rows containing user-specified equality constraints, if any.}
\item{remove_system_eq}{Logical. If \code{TRUE}, filter the output by
removing all rows containing system-generated equality constraints, if any.}
\item{remove_ineq}{Logical. If \code{TRUE}, filter the output by removing all
rows containing inequality constraints, if any.}
\item{remove_def}{Logical. If \code{TRUE}, filter the output by removing all
rows containing parameter definitions, if any.}
\item{remove_nonfree}{Logical. If \code{TRUE}, filter the output by removing 
all rows containing fixed (non-free) parameters.}
\item{remove_step1}{Logical. Only used by \code{sam()}. 
If \code{TRUE}, filter the 
output by removing all rows corresponding to the measurement parameters
that are part of the first step.}
\item{remove_unused}{Logical. If \code{TRUE}, filter the output by removing
all rows containing automatically added parameters (user == 0) that are 
nonfree, and with their final (est) values fixed to their default values 
(typically 1 or 0); currently only used for intercepts and
scaling-factors.}
\item{remove_aux}{Logical. If \code{TRUE} (the default), filter the output by
removing all rows corresponding to auxiliary (\code{aux=}) variables that were
added to the model as saturated correlates (only relevant when
\code{missing = "ml"} and the \code{aux=} argument was used).}
\item{rsquare}{Logical. If \code{TRUE}, add additional rows containing
the rsquare values (in the \code{est} column) of all endogenous variables
in the model. Both the \code{lhs} and \code{rhs} column contain the
name of the endogenous variable, while the \code{op} column contains \code{r2},
to indicate that the values in the \code{est} column are rsquare values.}
\item{add_attributes}{Deprecated argument. Please use output= instead.}
\item{output}{Character. If \code{"data.frame"}, the parameter table is
displayed as a standard (albeit lavaan-formatted) data.frame. 
If \code{"text"} (or alias \code{"pretty"}), the parameter table is
formatted and displayed with subsections (as used by the summary function).}
\item{header}{Logical. Only used if \code{output = "text"}. If
\code{TRUE}, print a header at the top of the parameter list. This header
contains information about the information matrix, whether the saturated (h1)
model is structured or unstructured, and which type of standard errors are
shown in the output.}
\item{...}{to support old argument names}
}
\value{
  A data.frame containing the estimated parameters,
  standard errors, and (by default) z-values, p-values, and
  the lower and upper values of the confidence intervals. If requested,
  extra columns are added with standardized versions of the parameter
  estimates.

  If the model was fitted with \code{se = "bootstrap"} and
  \code{boot_ci_type = "bca"}, the empirical-influence design matrix used to
  compute the acceleration constant is stored as the \code{"design"}
  attribute of the returned data.frame (retrieve with
  \code{attr(x, "design")}). The parameter estimates in the bootstrap samples
  themselves are available via \code{lavInspect(object, "coef.boot")}.
}
\references{
Savalei, V. & Rhemtulla, M. (2012). On obtaining estimates of the fraction of
missing information from FIML. Structural Equation Modeling: A
Multidisciplinary Journal, 19(3), 477-494.
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
lavParameterEstimates(fit)
lavParameterEstimates(fit, output = "text")
}

\name{lavPredict}
\alias{lavPredict}
\alias{lavpredict}
\title{Predict the values of latent variables (and their indicators)}
\description{
The main purpose of the \code{lavPredict()} function is to compute (or
`predict') individual scores for the latent variables in the model
(`factor scores'). NOTE: the goal of this
function is NOT to predict future values of dependent variables as in the
regression framework! (For models with only continuous observed variables, the function \code{lavPredictY()} supports this.)}
\usage{
lavPredict(object, newdata = NULL, type = "lv", method = "EBM",
           transform = FALSE, se = "none", acov = "none",
           label = TRUE, fsm = FALSE, mdist = FALSE, rel = FALSE,
           append_data = FALSE, assemble = FALSE,
           level = 1L, optim_method = "bfgs", eta = NULL,
           parallel = c("auto", "no", "multicore", "snow"),
           ncpus = NULL, cl = NULL,
           drop_list_single_group = TRUE,
           mdist_draws = 2000L,
           ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{newdata}{An optional data.frame, containing the same variables as
the data.frame used when fitting the model in object. For multilevel
models, the cluster variable(s) must be included; the cluster structure
and (for \code{missing = "ml"}) the missing-data patterns are rebuilt
from newdata, so the predictions are conditional on the newdata only
(the cluster labels need not match those of the original data).}
\item{type}{A character string. If \code{"lv"}, estimated values for the latent
variables in the model are computed. If \code{"ov"}, model predicted values for
the indicators of the latent variables in the model are computed. If
\code{"yhat"}, the estimated values for the observed indicators are computed,
given the user-specified values for the latent variables provided by the
\code{eta} argument.  If \code{"fy"}, densities (or probabilities) are computed
for each observed indicator, given the user-specified values for the latent
variables provided by the \code{eta} argument.}
\item{method}{A character string. In the linear case (when the indicators are
continuous), the possible options are \code{"regression"} or \code{"Bartlett"}.
In the categorical case, the two options are \code{"EBM"} for
the Empirical Bayes Modal approach, and \code{"ML"} for the maximum
likelihood approach. For higher-order factors (which have no observed
indicators of their own), the \code{"Bartlett"} scores are computed from the
collapsed measurement model (regressing the observed indicators of the
lower-order factors on the higher-order factors), as in \code{\link{sam}}.}
\item{transform}{Logical. If \code{TRUE}, transform the factor scores (per
group) so that their mean and variance-covariance matrix matches the
model-implied mean and variance-covariance matrix. This may be useful if the
individual factor scores will be used in a follow-up (regression) analysis.
Note: the standard errors (if requested) are not transformed (yet). The resulting
factor scores are often called correlation-preserving factor scores.}
\item{se}{Character. If \code{"none"}, no standard errors are computed.
If \code{"standard"}, standard errors are computed (assuming the
parameters of the measurement model are known). The standard errors are
returned as an attribute. For continuous data, the (naive) standard errors are
the same for every observation. For categorical data, where the factor scores
are obtained by numerical optimization (\code{method = "EBM"} or
\code{method = "ML"}), the standard errors are based on the curvature of the
objective at the optimum (the inverse of its Hessian, a Laplace/observed
-information approximation) and therefore differ from one response pattern to
the next; the result is an (nobs x nfactor) matrix per group.}
\item{acov}{Similar to the \code{"se"} argument, but optionally returns the full
sampling covariance matrix of the factor scores as an attribute. For
continuous data this is a single matrix per group; for categorical data it is
a list with one (nfactor x nfactor) matrix per observation.}
\item{label}{Logical. If TRUE, the columns in the output are labeled.}
\item{fsm}{Logical. If TRUE, return the factor score matrix as an attribute.
           Only for numeric data.}
\item{mdist}{Logical. If TRUE, the (squared)
Mahalanobis distances of the factor scores (if \code{type = "lv"}) or
the casewise residuals (if \code{type = "resid"}) are returned as an
attribute. If (some of) the data is ordered categorical, the distances
are generalized as in Mansolf and Reise (2017): the continuous distance is
replaced by its expected value over the region of the (multivariate
normally distributed) latent response vector that is consistent with the
observed response pattern; this expectation is approximated by Monte Carlo
integration (using \code{mdist_draws} random draws per distinct response
pattern; the result therefore depends on the state of the random number
generator, and \code{set.seed()} can be used for reproducibility).
Observed continuous variables are conditioned on, and missing values are
integrated out. In this categorical setting, \code{type = "resid"} is
allowed, and returns the expected casewise residuals of the latent
responses (a by-product of the same Monte Carlo integration).}
\item{rel}{Logical. Only used if \code{type = "lv"}. If TRUE,
the factor reliabilities are returned as an attribute. (The squared
values are often called the factor determinacies.)}
\item{append_data}{Logical. Only used when \code{type = "lv"}. If TRUE,
the original data (or the data provided
in the newdata argument) is appended to the factor scores.}
\item{assemble}{Logical. If TRUE,
the separate groups are reassembled into a single data.frame
with a group column, having the same dimensions as the original (or
newdata) dataset.}
\item{level}{Integer. Only used in a multilevel SEM.
If \code{level = 1}, only factor scores for latent variable
defined at the first (within) level are computed; if \code{level = 2},
only factor scores for latent variables defined at the second (between) level
are computed. For two-level models with \code{missing = "ml"}, the
posterior means of the cluster-level components and of the missing
values (given all the observed data; computed as in the E-step of the
EM algorithm) are used, so that the factor scores are the exact
posterior means given all the observed data (new in 0.7-1).}
\item{optim_method}{Character string. Only used in the categorical case.
If \code{"nlminb"} (the default in 0.5), the \code{"nlminb()"} function is used
for the optimization. If \code{"bfgs"} or \code{"BFGS"} (the default in 0.6),
the \code{"optim()"} function is used with the BFGS method.}
\item{eta}{An optional matrix or list, containing latent variable values
  for each observation. Used for computations when \code{type = "ov"}.}
\item{parallel}{Character. Only used in the categorical case, where factor
  scores are computed by a per-observation numerical optimization. The options
  are \code{"no"} (serial), \code{"multicore"} (fork-based, not available on
  Windows) and \code{"snow"} (a PSOCK cluster). The default, \code{"auto"},
  uses \code{"multicore"} automatically (on non-Windows platforms) when a
  large number of distinct optimizations is required, and runs serially
  otherwise. Because the computation is deterministic, the results do not
  depend on the value of this argument.}
\item{ncpus}{Integer. The number of processes to use in parallel computation.
  By default, this is 2 (following CRAN policy). You may set this to a higher
  value (for example, the number of available cores, minus two) if you wish to
  use more processes.}
\item{cl}{An optional \pkg{parallel} or \pkg{snow} cluster for use when
  \code{parallel = "snow"}. If not supplied, a cluster on the local machine is
  created for the duration of the call.}
\item{drop_list_single_group}{Logical. If \code{FALSE}, the results are
    returned as
    a list, where each element corresponds to a group (even if there is only
    a single group). If \code{TRUE}, the list will be unlisted if there is
    only a single group.}
\item{mdist_draws}{Integer. Only used if \code{mdist = TRUE} and (some of)
    the data is ordered categorical: the number of Monte Carlo draws per
    distinct response pattern used to approximate the expected Mahalanobis
    distances.}
\item{...}{To support old argument names.}
}
\details{
The \code{predict()} function calls the \code{lavPredict()} function
with its default options.

If there are no latent variables in the model, \code{type = "ov"} will
simply return the values of the observed variables. Note that this function
can not be used to `predict' values of dependent variables, given the
values of independent variables (in the regression sense). In other words,
the structural component is completely ignored (for now).
}
\references{
For an overview (and evaluation) of the various factor score methods, see:

Grice, J. W. (2001). Computing and evaluating factor scores.
\emph{Psychological Methods, 6}(4), 430-450. \doi{10.1037/1082-989X.6.4.430}

For the (continuous) regression and Bartlett methods, see:

Bartlett, M. S. (1937). The statistical conception of mental factors.
\emph{British Journal of Psychology, 28}, 97-104.
\doi{10.1111/j.2044-8295.1937.tb00863.x}

Bentler, P. M., & Yuan, K.-H. (1997). Optimal conditionally unbiased
equivariant factor score estimators. In M. Berkane (Ed.), \emph{Latent
variable modeling and applications to causality} (pp. 259-281). New York:
Springer-Verlag. \doi{10.1007/978-1-4612-1842-5_14}

For the (categorical) Empirical Bayes Modal (EBM) and Maximum Likelihood (ML)
methods, see:

Skrondal, A., & Rabe-Hesketh, S. (2004). \emph{Generalized latent variable
modeling: Multilevel, longitudinal, and structural equation models.} Boca
Raton, FL: Chapman & Hall/CRC.

For the Mahalanobis distances with ordered categorical data
(\code{mdist = TRUE}), see:

Mansolf, M., & Reise, S. P. (2017). Case diagnostics for factor analysis of
ordered categorical data with applications to person-fit measurement.
\emph{Structural Equation Modeling: A Multidisciplinary Journal, 25}(1),
86-100. \doi{10.1080/10705511.2017.1367926}

For the correlation-preserving factor scores (\code{transform = TRUE}), see:

ten Berge, J. M. F., Krijnen, W. P., Wansbeek, T., & Shapiro, A. (1999). Some
new results on correlation-preserving factor scores prediction methods.
\emph{Linear Algebra and its Applications, 289}(1-3), 311-318.
\doi{10.1016/S0024-3795(97)10007-6}
}
\seealso{
\code{\link{lavPredictY}} to predict y-variables given x-variables.
}
\examples{
data(HolzingerSwineford1939)

## fit model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data = HolzingerSwineford1939)
head(lavPredict(fit))
head(lavPredict(fit, type = "ov"))


## ---------------------------------------------
## standard errors for the factor scores (se =)
## ---------------------------------------------

## the standard errors are returned as the "se" attribute (a list, one
## (nobs x nfactor) matrix per group)

## for continuous indicators, the (naive) standard errors are the same
## for every observation
fscores <- lavPredict(fit, se = "standard")
attr(fscores, "se")[[1]]

## for categorical indicators, the factor scores are obtained by numerical
## optimization, and their standard errors differ from one response pattern
## (observation) to the next

## (a two-factor model is used here, as the numerical optimization and the
##  Monte Carlo integration below both get expensive as the number of
##  ordered indicators grows)
HS.model.ord <- ' visual  =~ x1 + x2 + x3
                  textual =~ x4 + x5 + x6 '
HS.ord <- HolzingerSwineford1939
HS.ord[ , paste0("x", 1:6)] <-
    lapply(HS.ord[ , paste0("x", 1:6)], function(x) ordered(cut(x, 3)))
fit.ord <- cfa(HS.model.ord, data = HS.ord, ordered = paste0("x", 1:6))
fscores <- lavPredict(fit.ord, se = "standard")
head(attr(fscores, "se")[[1]])


## -------------------------------------------------
## casewise Mahalanobis distances (case diagnostics)
## -------------------------------------------------

## continuous data: the squared distances are exact
fscores <- lavPredict(fit, mdist = TRUE)
head(attr(fscores, "mdist")[[1]])

## ordered categorical data: expected (squared) distances of the latent
## responses, obtained by Monte Carlo integration (Mansolf & Reise, 2017);
## set the seed for reproducibility
set.seed(123)
resid.ord <- lavPredict(fit.ord, type = "resid", mdist = TRUE)
## expected casewise residuals of the latent responses:
head(resid.ord)
## expected squared residual-based distances (person fit):
head(attr(resid.ord, "mdist")[[1]])


## ------------------------------------------
## merge factor scores to original data.frame
## ------------------------------------------

idx <- lavInspect(fit, "case.idx")
fscores <- lavPredict(fit)
## loop over factors
for (fs in colnames(fscores)) {
  HolzingerSwineford1939[idx, fs] <- fscores[ , fs]
}
head(HolzingerSwineford1939)


## multigroup models return a list of factor scores (one per group)
data(HolzingerSwineford1939)
mgfit <- update(fit, group = "school", group.equal = c("loadings","intercepts"))

idx <- lavInspect(mgfit, "case.idx") # list: 1 vector per group
fscores <- lavPredict(mgfit)         # list: 1 matrix per group
## loop over groups and factors
for (g in seq_along(fscores)) {
  for (fs in colnames(fscores[[g]])) {
    HolzingerSwineford1939[ idx[[g]], fs] <- fscores[[g]][ , fs]
  }
}
head(HolzingerSwineford1939)

## -------------------------------------
## Use factor scores in subsequent models
## -------------------------------------

## see Examples in semTools package: ?plausibleValues
}

\name{lavTablesFitCp}
\alias{lavTablesFit}
\alias{lavTablesFitCp}
\alias{lavTablesFitCf}
\alias{lavTablesFitCm}
\title{Pairwise maximum likelihood fit statistics}

\description{
Three measures of fit for the pairwise maximum likelihood estimation method that are based on likelihood ratios (LR) are defined: 
\eqn{C_F}, \eqn{C_M}, and \eqn{C_P}. Subscript \eqn{F} signifies a comparison of model-implied proportions of full response 
patterns with observed sample proportions, subscript \eqn{M} signifies a comparison of model-implied proportions of full response 
patterns with the proportions implied by the assumption of multivariate normality, and subscript \eqn{P} signifies
a comparison of model-implied proportions of pairs of item responses with the observed proportions of pairs of item responses. 
}

\usage{
lavTablesFitCf(object)
lavTablesFitCp(object, alpha = 0.05)
lavTablesFitCm(object)
}

\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{alpha}{The nominal level of significance of global fit.}
}

\references{
Barendse, M. T., Ligtvoet, R., Timmerman, M. E., & Oort, F. J. (2016). Structural Equation Modeling of Discrete data: 
Model Fit after Pairwise Maximum Likelihood. \emph{Frontiers in psychology, 7}, 1-8. 

Joreskog, K. G., & Moustaki, I. (2001). Factor analysis of ordinal variables: A comparison of three approaches. 
\emph{Multivariate Behavioral Research, 36}, 347-387. 
}

\details{
\subsection{\eqn{C_F}}{
The \eqn{C_F} statistic compares the log-likelihood of the model-implied proportions (\eqn{\pi_r}) with the observed proportions (\eqn{p_r})
of the full multivariate responses patterns:
\deqn{
C_F = 2N\sum_{r}p_{r}\ln[p_{r}/\hat{\pi}_{r}],
}
which asymptotically has a chi-square distribution with
\deqn{
df_F = m^k - n - 1,
}
where \eqn{k} denotes the number of items with discrete response scales, \eqn{m} denotes the number of response options, and \eqn{n} denotes 
the number of parameters to be estimated. Notice that \eqn{C_F} results may be biased because of large numbers of empty cells in the multivariate 
contingency table.
}

\subsection{\eqn{C_M}}{
The \eqn{C_M} statistic is based on the \eqn{C_F} statistic, and compares the proportions implied by the model of interest (Model 1) 
with proportions implied by the assumption of an underlying multivariate normal distribution (Model 0):
\deqn{
C_M = C_{F1} - C_{F0},
}
where \eqn{C_{F0}} is \eqn{C_F} for Model 0 and \eqn{C_{F1}} is \eqn{C_F} for Model 1. Statistic \eqn{C_M} has a chi-square distribution with 
degrees of freedom
\deqn{
df_M = k(k-1)/2 + k(m-1) - n_{1},
}
where \eqn{k} denotes the number of items with discrete response scales, \eqn{m} denotes the number of response options, and \eqn{k(k-1)/2} 
denotes the number of polychoric correlations, \eqn{k(m-1)} denotes the number of thresholds, and \eqn{n_1} is the number of parameters of the 
model of interest. Notice that \eqn{C_M}  results may be biased because of large numbers of empty cells in the multivariate contingency table. However, 
bias may cancel out as both Model 1 and Model 0 contain the same pattern of empty responses.
}

\subsection{\eqn{C_P}}{
With the \eqn{C_P} statistic we only consider pairs of responses, and compare observed sample proportions (\eqn{p}) with model-implied proportions 
of pairs of responses (\eqn{\pi}). For items \eqn{i} and \eqn{j} we obtain a pairwise likelihood ratio test statistic \eqn{C_{P_{ij}}}
\deqn{
C_{P_{ij}}=2N\sum_{c_i=1}^m \sum_{c_j=1}^m
p_{c_i,c_j}\ln[p_{c_i,c_j}/\hat{\pi}_{c_i,c_j}],
}
where \eqn{m} denotes the number of response options and \eqn{N} denotes sample size. The \eqn{C_P} statistic has an asymptotic chi-square distribution 
with degrees of freedom equal to the information \eqn{(m^2 -1)} minus the number of parameters (2(m-1) thresholds and 1 correlation),
\deqn{
df_P = m^{2} - 2(m - 1) - 2.
}
As \eqn{k} denotes the number of items, there are \eqn{k(k-1)/2} possible pairs of items. The \eqn{C_P} statistic should therefore be applied with 
a Bonferroni adjusted level of significance \eqn{\alpha^*}, with
\deqn{
\alpha^*= \alpha /(k(k-1)/2)),
}
to keep the family-wise error rate at \eqn{\alpha}. The hypothesis of overall goodness-of-fit is tested at \eqn{\alpha} and rejected as 
soon as \eqn{C_P} is significant at \eqn{\alpha^*} for at least one pair of items. Notice that with dichotomous items, \eqn{m = 2}, 
and \eqn{df_P = 0}, so that the hypothesis cannot be tested.
}
}

\seealso{
\code{\link{lavTables}, \link{lavaan}}
}

\examples{
# Data
HS9 <- HolzingerSwineford1939[,c("x1","x2","x3","x4","x5",
                                 "x6","x7","x8","x9")]
HSbinary <- as.data.frame( lapply(HS9, cut, 2, labels=FALSE) )

# Single group example with one latent factor
HS.model <- ' trait =~ x1 + x2 + x3 + x4 '
fit <- cfa(HS.model, data=HSbinary[,1:4], ordered=names(HSbinary[,1:4]),
           estimator="PML")
lavTablesFitCm(fit)
lavTablesFitCp(fit)
lavTablesFitCf(fit)
}

\name{lav_long_names}
\alias{lav_matrix_duplication_ginv_pre_post}
\alias{lav_matrix_orthogonal_complement}
\alias{lav_func_gradient_complex}
\alias{lav_matrix_duplication_ginv_post}
\alias{lav_func_gradient_simple}
\alias{lav_matrix_vech_row_idx}
\alias{lav_matrix_vech_col_idx}
\alias{lav_matrix_antidiag_idx}
\alias{lav_matrix_duplication_pre_post}
\alias{lav_matrix_duplication_ginv_pre}
\alias{lav_matrix_commutation_pre_post}
\alias{lav_partable_unrestricted}
\alias{lav_partable_independence}
\alias{lav_matrix_vechru_idx}
\alias{lav_matrix_upper2full}
\alias{lav_matrix_lower2full}
\alias{lav_matrix_commutation_mn_pre}
\alias{lav_samplestats_from_data}
\alias{lav_matrix_vechru_reverse}
\alias{lav_matrix_vechr_idx}
\alias{lav_matrix_vechu_idx}
\alias{lav_matrix_diagh_idx}
\alias{lav_partable_attributes}
\alias{lav_matrix_vechr_reverse}
\alias{lav_matrix_vechu_reverse}
\alias{lav_matrix_vech_idx}
\alias{lav_matrix_diag_idx}
\alias{lav_matrix_duplication_post}
\alias{lav_matrix_duplication_ginv}
\alias{lav_matrix_commutation_post}
\alias{lav_matrix_symmetric_sqrt}
\alias{lav_matrix_vech_reverse}
\alias{lav_matrix_duplication_pre}
\alias{lav_matrix_commutation_pre}
\alias{lav_partable_complete}
\alias{lav_matrix_vechru}
\alias{lav_partable_constraints_def}
\alias{lav_partable_constraints_ceq}
\alias{lav_partable_constraints_ciq}
\alias{lav_partable_from_lm}
\alias{lav_constraints_parse}
\alias{lav_matrix_vechr}
\alias{lav_matrix_vechu}
\alias{lav_matrix_bdiag}
\alias{lav_matrix_trace}
\alias{lav_partable_labels}
\alias{lav_matrix_print}
\alias{lav_matrix_vecr}
\alias{lav_matrix_vech}
\alias{lav_partable_merge}
\alias{lav_matrix_vec}
\alias{lav_matrix_duplication}
\alias{lav_matrix_commutation}
\alias{lav_matrix_cov}
\alias{lav_partable_ndat}
\alias{lav_partable_npar}
\alias{lav_partable_add}
\alias{lav_partable_df}

\title{Deprecated functions with long names}
\description{Functions that are deprecated following the adoption of shorter names in May 2026.}
\usage{
lav_matrix_duplication_ginv_pre_post(A = matrix(0, 0, 0))
lav_matrix_orthogonal_complement(A = matrix(0, 0, 0))
lav_func_gradient_complex(
                          func,
                          x,
                          h = .Machine$double.eps,
                          ...,
                          fallback.simple = TRUE)
lav_matrix_duplication_ginv_post(A = matrix(0, 0, 0))
lav_func_gradient_simple(
                         func,
                         x,
                         h = sqrt(.Machine$double.eps),
                         ...)
lav_matrix_vech_row_idx(n = 1L, diagonal = TRUE)
lav_matrix_vech_col_idx(n = 1L, diagonal = TRUE)
lav_matrix_antidiag_idx(n = 1L)
lav_matrix_duplication_pre_post(A = matrix(0, 0, 0))
lav_matrix_duplication_ginv_pre(A = matrix(0, 0, 0))
lav_matrix_commutation_pre_post(A = matrix(0, 0, 0))
lav_partable_unrestricted(
                          lavobject = NULL,
                          lavdata = NULL,
                          lavpta = NULL,
                          lavoptions = NULL,
                          lavsamplestats = NULL,
                          lavh1 = NULL,
                          sample.cov = NULL,
                          sample.mean = NULL,
                          sample.slopes = NULL,
                          sample.th = NULL,
                          sample.th.idx = NULL,
                          sample.cov.x = NULL,
                          sample.mean.x = NULL)
lav_partable_independence(
                          lavobject = NULL,
                          lavdata = NULL,
                          lavpta = NULL,
                          lavoptions = NULL,
                          lavsamplestats = NULL,
                          lavh1 = NULL,
                          sample.cov = NULL,
                          sample.mean = NULL,
                          sample.slopes = NULL,
                          sample.th = NULL,
                          sample.th.idx = NULL,
                          sample.cov.x = NULL,
                          sample.mean.x = NULL)
lav_matrix_vechru_idx(n = 1L, diagonal = TRUE)
lav_matrix_upper2full(x, diagonal = TRUE)
lav_matrix_lower2full(x, diagonal = TRUE)
lav_matrix_commutation_mn_pre(
                              A,
                              m = 1L,
                              n = 1L)
lav_samplestats_from_data(
                          lavdata = NULL,
                          lavoptions = NULL,
                          WLS.V = NULL,
                          NACOV = NULL)
lav_matrix_vechru_reverse(x, diagonal = TRUE)
lav_matrix_vechr_idx(n = 1L, diagonal = TRUE)
lav_matrix_vechu_idx(n = 1L, diagonal = TRUE)
lav_matrix_diagh_idx(n = 1L)
lav_partable_attributes(partable, pta = NULL)
lav_matrix_vechr_reverse(x, diagonal = TRUE)
lav_matrix_vechu_reverse(x, diagonal = TRUE)
lav_matrix_vech_idx(n = 1L, diagonal = TRUE)
lav_matrix_diag_idx(n = 1L)
lav_matrix_duplication_post(A = matrix(0, 0, 0))
lav_matrix_duplication_ginv(n = 1L)
lav_matrix_commutation_post(A = matrix(0, 0, 0))
lav_matrix_symmetric_sqrt(S = matrix(0, 0, 0))
lav_matrix_vech_reverse(x, diagonal = TRUE)
lav_matrix_duplication_pre(A = matrix(0, 0, 0))
lav_matrix_commutation_pre(A = matrix(0, 0, 0))
lav_partable_complete(partable = NULL, start = TRUE)
lav_matrix_vechru(S, diagonal = TRUE)
lav_partable_constraints_def(
                             partable,
                             con = NULL,
                             debug = FALSE,
                             txtOnly = FALSE,
                             warn = TRUE)
lav_partable_constraints_ceq(
                             partable,
                             con = NULL,
                             debug = FALSE,
                             txtOnly = FALSE)
lav_partable_constraints_ciq(
                             partable,
                             con = NULL,
                             debug = FALSE,
                             txtOnly = FALSE)
lav_partable_from_lm(
                     object,
                     est = FALSE,
                     label = FALSE,
                     as.data.frame. = FALSE)
lav_constraints_parse(
                      partable = NULL,
                      constraints = NULL,
                      theta = NULL,
                      debug = FALSE)
lav_matrix_vechr(S, diagonal = TRUE)
lav_matrix_vechu(S, diagonal = TRUE)
lav_matrix_bdiag(...)
lav_matrix_trace(..., check = TRUE)
lav_partable_labels(
                    partable,
                    blocks = c("group", "level"),
                    group.equal = "",
                    group.partial = "",
                    type = "user")
lav_matrix_vecr(A)
lav_matrix_vech(S, diagonal = TRUE)
lav_partable_merge(
                   pt1 = NULL,
                   pt2 = NULL,
                   remove.duplicated = FALSE,
                   fromLast = FALSE,
                   warn = TRUE)
lav_matrix_vec(A)
lav_matrix_duplication(n = 1L)
lav_matrix_commutation(m = 1L, n = 1L)
lav_matrix_cov(Y, Mu = NULL)
lav_partable_ndat(partable)
lav_partable_npar(partable)
lav_partable_add(partable = NULL, add = list())
lav_partable_df(partable)
}
\arguments{
\item{...}{See argument \code{...} in the corresponding function with shortened name.}
\item{A}{See argument \code{a} in the corresponding function with shortened name.}
\item{add}{See argument \code{add} in the corresponding function with shortened name.}
\item{as.data.frame.}{See argument \code{as_data_frame} in the corresponding function with shortened name.}
\item{blocks}{See argument \code{blocks} in the corresponding function with shortened name.}
\item{check}{See argument \code{check} in the corresponding function with shortened name.}
\item{con}{See argument \code{con} in the corresponding function with shortened name.}
\item{constraints}{See argument \code{constraints} in the corresponding function with shortened name.}
\item{debug}{See argument \code{debug} in the corresponding function with shortened name.}
\item{diagonal}{See argument \code{diagonal} in the corresponding function with shortened name.}
\item{est}{See argument \code{est} in the corresponding function with shortened name.}
\item{fallback.simple}{See argument \code{fallback_simple} in the corresponding function with shortened name.}
\item{fromLast}{See argument \code{from_last} in the corresponding function with shortened name.}
\item{func}{See argument \code{func} in the corresponding function with shortened name.}
\item{group.equal}{See argument \code{group_equal} in the corresponding function with shortened name.}
\item{group.partial}{See argument \code{group_partial} in the corresponding function with shortened name.}
\item{h}{See argument \code{h} in the corresponding function with shortened name.}
\item{label}{See argument \code{label} in the corresponding function with shortened name.}
\item{lavdata}{See argument \code{lavdata} in the corresponding function with shortened name.}
\item{lavh1}{See argument \code{lavh1} in the corresponding function with shortened name.}
\item{lavobject}{See argument \code{lavobject} in the corresponding function with shortened name.}
\item{lavoptions}{See argument \code{lavoptions} in the corresponding function with shortened name.}
\item{lavpta}{See argument \code{lavpta} in the corresponding function with shortened name.}
\item{lavsamplestats}{See argument \code{lavsamplestats} in the corresponding function with shortened name.}
\item{m}{See argument \code{m} in the corresponding function with shortened name.}
\item{Mu}{See argument \code{mu} in the corresponding function with shortened name.}
\item{n}{See argument \code{n} in the corresponding function with shortened name.}
\item{NACOV}{See argument \code{nacov} in the corresponding function with shortened name.}
\item{object}{See argument \code{object} in the corresponding function with shortened name.}
\item{partable}{See argument \code{partable} in the corresponding function with shortened name.}
\item{pt1}{See argument \code{pt1} in the corresponding function with shortened name.}
\item{pt2}{See argument \code{pt2} in the corresponding function with shortened name.}
\item{pta}{See argument \code{pta} in the corresponding function with shortened name.}
\item{remove.duplicated}{See argument \code{remove_duplicated} in the corresponding function with shortened name.}
\item{S}{See argument \code{s} in the corresponding function with shortened name.}
\item{sample.cov}{See argument \code{sample_cov} in the corresponding function with shortened name.}
\item{sample.cov.x}{See argument \code{sample_cov_x} in the corresponding function with shortened name.}
\item{sample.mean}{See argument \code{sample_mean} in the corresponding function with shortened name.}
\item{sample.mean.x}{See argument \code{sample_mean_x} in the corresponding function with shortened name.}
\item{sample.slopes}{See argument \code{sample_slopes} in the corresponding function with shortened name.}
\item{sample.th}{See argument \code{sample_th} in the corresponding function with shortened name.}
\item{sample.th.idx}{See argument \code{sample_th_idx} in the corresponding function with shortened name.}
\item{start}{See argument \code{start} in the corresponding function with shortened name.}
\item{theta}{See argument \code{theta} in the corresponding function with shortened name.}
\item{txtOnly}{See argument \code{txt_only} in the corresponding function with shortened name.}
\item{type}{See argument \code{type} in the corresponding function with shortened name.}
\item{warn}{See argument \code{warn} in the corresponding function with shortened name.}
\item{WLS.V}{See argument \code{wls_v} in the corresponding function with shortened name.}
\item{x}{See argument \code{x} in the corresponding function with shortened name.}
\item{Y}{See argument \code{y} in the corresponding function with shortened name.}
}
\details{
Function names are shortened by the following replacements:

\describe{
  \item{_mvnorm}{_mvn}
  \item{_cluster}{_cl}
  \item{_missing}{_mi}
  \item{_matrix}{_mat}
  \item{_duplication}{_dup}
  \item{_commutation}{_com}
  \item{_information}{_info}
  \item{_samplestats}{_samp}
  \item{_satorra_bentler}{_sb}
  \item{_SatorraBentler}{_sb}
  \item{_yuan_bentler}{_yb}
  \item{_inverted}{_inv}
  \item{_estimate}{_est}
  \item{_symmetric}{_sym}
  \item{_lavaan_step}{_step}
  \item{_univariate}{_uni}
  \item{_fit_measures}{_fit}
  \item{_constraints}{_con}
  \item{_scores}{_sc}
  \item{_reverse}{_rev}
  \item{_vech_sigma}{_sigma}
  \item{_vechsigma}{_sigma}
  \item{_object_inspect}{_inspect}
  \item{_orthogonal}{_ortho}
  \item{_gradient}{_grad}
  \item{_partable}{_pt}
}
}

\name{HolzingerSwineford1939}
\alias{HolzingerSwineford1939}
\docType{data}
\title{
Holzinger and Swineford Dataset (9 Variables)
}
\description{
The classic Holzinger and Swineford (1939) dataset consists of mental
ability test scores of seventh- and eighth-grade children from two
different schools (Pasteur and Grant-White). In the original dataset
(available in the \code{MBESS} package), there are scores for 26 tests.
However, a smaller subset with 9 variables is more widely used in the 
literature (for example in Joreskog's 1969 paper, which also uses the 145 
subjects from the Grant-White school only). 
}
\usage{data(HolzingerSwineford1939)}
\format{
  A data frame with 301 observations of 15 variables.
  \describe{
    \item{\code{id}}{Identifier}
    \item{\code{sex}}{Gender}
    \item{\code{ageyr}}{Age, year part}
    \item{\code{agemo}}{Age, month part}
    \item{\code{school}}{School (Pasteur or Grant-White)}
    \item{\code{grade}}{Grade}
    \item{\code{x1}}{Visual perception}
    \item{\code{x2}}{Cubes}
    \item{\code{x3}}{Lozenges}
    \item{\code{x4}}{Paragraph comprehension}
    \item{\code{x5}}{Sentence completion}
    \item{\code{x6}}{Word meaning}
    \item{\code{x7}}{Speeded addition}
    \item{\code{x8}}{Speeded counting of dots}
    \item{\code{x9}}{Speeded discrimination straight and curved capitals}
  }
}
\source{
This dataset was originally retrieved from \verb{http://web.missouri.edu/~kolenikovs/stata/hs-cfa.dta} (link no longer active) and converted to an R dataset.
}
\references{
Holzinger, K., and Swineford, F. (1939). A study in factor analysis: The 
stability of a bifactor solution. Supplementary Educational Monograph, no. 48. 
Chicago: University of Chicago Press. 

Joreskog, K. G. (1969). A general approach to confirmatory maximum likelihood
factor analysis. \emph{Psychometrika}, 34, 183-202.
}
\seealso{
\code{\link{cfa}}
}
\examples{
head(HolzingerSwineford1939)
}

\name{lav_mplus_lavaan}
\alias{lav_mplus_lavaan}
\alias{mplus2lavaan}
\alias{lavImport}
\title{mplus to lavaan converter}
\description{
Read in an Mplus input file, convert it to lavaan syntax, and fit
the model.}
\usage{
lav_mplus_lavaan(inpfile, run = TRUE)
mplus2lavaan(inpfile, run = TRUE)
}
\arguments{
\item{inpfile}{The filename (including a full path) of the Mplus input file.
The data (as referred to in the Mplus input file) should be in the same
directory as the Mplus input file.}
\item{run}{Whether to run the specified Mplus input syntax (\code{TRUE}) or
only to parse and convert the syntax (\code{FALSE}).} 
}
\value{
A \code{lavaan} object with the fitted results of the Mplus model. The parsed
and converted Mplus syntax is preserved in the \code{@external} slot of the \code{lavaan}
object in the \code{$mplus.inp} element. If \code{run} is \code{FALSE}, a \code{list} of converted
syntax is returned.
}
\author{Michael Hallquist}
\seealso{\code{\link{lavExport}}}
\examples{
\dontrun{
out <- lav_mplus_lavaan("ex5.1.inp")
summary(out)
}
}

\name{lav_model_plotinfo}
\alias{lav_model_plotinfo}
\title{Get model information}
\usage{
lav_model_plotinfo(model = NULL, infile = NULL, varlv = FALSE)
}
\arguments{
\item{model}{A character vector specifying the model in lavaan syntax or a list
(or data.frame) with at least members lhs, op, rhs, label and fixed or a fitted
lavaan object (in which case the \code{ParTable} object is extracted and column
\code{est} is used as value to show).
Should be \code{NULL} if infile is given.}
\item{infile}{A character string specifying the file that contains the model
syntax.}
\item{varlv}{A logical indicating that the (residual) variance of a variable
should be plotted as a separate latent variable (with a smaller circle than
ordinary latent variables). In this case, a covariance between two such variables
is plotted as a covariance between their variance latent variables.}
}
\value{
A structure 'plotinfo', which is a list with members nodes and edges. These
are data.frames containing the data needed to create a diagram.
\enumerate{
\item nodes
  \describe{
    \item{id}{character, identification of the node consisting of blok and naam.}
    \item{naam}{character, name of the node as specified in the model.
    For intercepts the name is "1vanXXXX", with XXXX the name
    of the regressed variable.}
    \item{tiepe}{character, type of node: ov (observed variable),
    lv (latent variable), varlv (variance as latent variable),
    cv (composite variable), wov (within level variable in multilevel model),
    bov (between level variable in multilevel model),
    const (intercept of regression).}
    \item{blok}{integer, level (0 if not a multilevel model).}
 }
\item edges
  \describe{
    \item{id}{integer, autoincrement identification of the edge.}
    \item{label}{character, label for the edge, made from the label specified in
    the model and the fixed (or estimated) value if present.}
    \item{van}{character, id of the starting node.}
    \item{naar}{character, id of the destination node.}
    \item{tiepe}{character, lavaan operator, with two exceptions: a (residual)
      variance is coded here as '~~~', and a regression introduced by
      varlv = TRUE is coded as '~.'.}
  }
}
}

\description{
Extracts the information from a model that is needed to produce a plot.
}

\examples{
model <- 'alpha  =~ 1 * x1 + x2 + x3        # latent variable
          beta <~ x4 + x5 + x6              # composite
          gamma =~ 1 * x7 + x8 + x9         # latent variable
          Xi =~ 1 * x10 + x11 + x12 + x13   # latent variable
          # regressions
          Xi ~ v * alpha + t * beta + cc * 1
          alpha ~ tt * beta + ss * gamma + yy * Theta1
          # variances and covariances
          x2 ~~ cc25 * x5
          x3 ~~ cc36 * x6
          x3 ~~ cc34 * x4
          gamma ~~ 0.55 * gamma
          '
(test <- lav_model_plotinfo(model))
}

\name{efa}
\alias{efa}
\alias{rotation}
\title{Exploratory Factor Analysis}
\description{
Fit one or more Exploratory Factor Analysis (EFA) model(s).}
\usage{
efa(data = NULL, nfactors = 1L, sample_cov = NULL, sample_nobs = NULL,
    rotation = "geomin", rotation_args = list(), ov_names = NULL,
    bounds = "pos.var", ..., output = "efa")
}
\arguments{
\item{data}{A data frame containing the observed variables we need for the
    EFA. If only a subset of the observed variables is needed, use the
    \code{ov.names} argument.}
\item{nfactors}{Integer, integer vector, or a list. The desired number of
    factors to extract. Can be a single number, or a vector of numbers
    (e.g., \code{nfactors = 1:4}); for each different number, a model is
    fitted. When the model has multiple blocks (because the data is
    multilevel and/or multigroup), the same number of factors is extracted in
    every block. To extract a different number of factors per block, supply a
    \emph{list}: each element of the list is one fitted model and contains a
    vector with the number of factors per block (or a single number that is
    recycled to all blocks). Blocks are ordered group-major, level-minor, ie
    block \eqn{= (g - 1) \times nlevels + level}. For example, for a twolevel
    model, \code{nfactors = list(c(1, 2))} extracts one factor at the within
    level and two factors at the between level.}
\item{sample_cov}{Numeric matrix. A sample variance-covariance matrix.
    The rownames and/or colnames must contain the observed variable names.
    Unlike sem and CFA, the matrix may be a correlation matrix.}
\item{sample_nobs}{Number of observations if the full data frame is missing
    and only the sample variance-covariance matrix is given.}
\item{rotation}{Character or a list with first element a character variable.
    The character variable is the rotation method to be used. Possible options
    are varimax, quartimax, orthomax, oblimin, quartimin, geomin, promax,
    entropy, mccammon, infomax, tandem1, tandem2, oblimax, bentler, simplimax, 
    target.strict, target (alias for pst), pst (=partially specified target), 
    cf, crawford-ferguson,
    cf-quartimax,  cf-varimax, cf-equamax,
    cf-parsimax, cf-facparsim, biquartimin, bigeomin. The latter two are
    for bifactor rotation only. The rotation algorithms (except promax
    and target) are similar to those from the GPArotation package, but have 
    been reimplemented for better control. The promax method is taken from the
    stats package. The target.strict method is equal to the target method in
    the GPArotation package. The target method is in fact the pst method where
    all non-zero elements (in the target matrix) are ignored.
    Options for the rotation algorithm can be provided in the list after the method.
    The default options (and their alternatives) are \code{orthogonal = FALSE},
   \code{row_weights = "default"} (or \code{"kaiser"},
   \code{"cureton.mulaik"} or \code{"none"}), \code{std_ov = TRUE},
   \code{algorithm = "gpa"} (or \code{"pairwise"}), \code{rstarts = 30},
   \code{gpa_tol = 1e-05}, \code{tol = 1e-08}, \code{max_iter = 10000L},
   \code{warn = FALSE}, \code{verbose = FALSE}, \code{reflect = TRUE},
   \code{order_lv_by = "index"} (or \code{"sumofsquares"} or \code{"none"}).
   Other options are specific for a particular rotation criterion:
   \code{geomin_epsilon = 0.001}, \code{orthomax_gamma = 1},
   \code{promax_kappa = 4},
   \code{cf_gamma = 0}, and \code{oblimin_gamma = 0}.

   For the target rotation methods (\code{target.strict}, \code{target} and
   \code{pst}), the \code{target} option (and, for \code{pst}, the optional
   \code{target_mask} option) can take three forms: 1) a single matrix,
   which is used for all groups and all efa blocks; 2) a \emph{nameless}
   list of matrices, one per group; or 3) a \emph{named} list of matrices,
   where the names are the efa block labels used in the model syntax (e.g.,
   \code{list(a = target.a, b = target.b)} if the model contains the efa
   blocks \code{efa("a")} and \code{efa("b")}), so that each efa block gets
   its own target matrix (the same targets are then used in all groups).
   The latter is only relevant when the model syntax (of \code{sem()} or
   \code{cfa()}) defines multiple efa blocks (set-ESEM). Note that a named
   list must provide a target matrix for \emph{every} efa block (with two
   or more factors) in the model; omitting a block results in an error, as
   the rotation criterion would be undefined for that block.}
\item{rotation_args}{List. Options related to the rotation algorithm. DEPRECATED.
   The options should now be given in the \code{rotation} argument as a list.}
\item{ov_names}{Character vector. The variables names that are needed for
   the EFA. Should be a subset of the variables names in the data.frame.
   By default (if NULL), all the variables in the data are used.}
\item{bounds}{Per default, \code{bounds = "pos.var"} forces all variances
   of both observed and latent variables to be strictly nonnegative. See
   the entry in \code{\link{lavOptions}} for more options.}
\item{...}{Additional options to be passed to lavaan, using 'name = value'.
    See \code{\link{lavOptions}} for a complete list.}
\item{output}{Character. If \code{"efa"} (the default), the output mimics
    the typical output of an EFA. If \code{"lavaan"}, a lavaan object is returned.
    The latter is only possible if nfactors contains a single (integer) number.}
}
\details{
    The \code{efa} function is essentially a wrapper around the
    \code{lavaan} function. It generates the model syntax (for a given number
    of factors) and then calls \code{lavaan()} treating the factors (in each
    block) as a set that should be rotated. Each block is rotated
    independently. Categorical data is handled as usual by first computing
    an appropriate (e.g., tetrachoric or polychoric) correlation matrix,
    which is then used as input for the EFA.
    Multiple groups (via the \code{group=} argument) and twolevel data
    (via the \code{cluster=} argument) are supported, and these may be
    combined. By default, the same number of factors is extracted in every
    block, but a different number of factors per block can be requested by
    supplying \code{nfactors} as a list (see the \code{nfactors} argument).
    The promax rotation method (taken from the stats package) is only 
    provided for convenience. Because promax is a two-step algorithm (first 
    varimax, then oblique rotation to get simple structure), it does not 
    use the gpa or pairwise rotation algorithms, and as a result, no
    standard errors are provided.
}
\value{
  If \code{output = "lavaan"}, an object of class 
  \code{\linkS4class{lavaan}}. If \code{output = "efa"},
  a list of class \code{efaList} for which a \code{print()}, 
  \code{summary()} and \code{fitMeasures()} method are available. Because
  we added the (standardized) loadings as an extra element, the \code{loadings}
  function (which is not a generic function) from the stats package will
  also work on \code{efaList} objects.
}
\seealso{
    \code{\link{lav_efalist_summary}} for a summary method if the output is
    of class \code{efaList}. 
}
\examples{
## The famous Holzinger and Swineford (1939) example
fit <- efa(data = HolzingerSwineford1939, 
           ov.names = paste("x", 1:9, sep = ""),
           nfactors = 1:3,
           rotation = list("geomin", geomin_epsilon = 0.01, rstarts = 1))
summary(fit, nd = 3L, cutoff = 0.2, dot.cutoff = 0.05)
fitMeasures(fit, fit.measures = "all")

# target rotation
target <- matrix(0, 9, 3)
target[1:3, 1] <- 1
target[4:6, 2] <- 1
target[7:9, 3] <- 1
fit2 <- efa(data = HolzingerSwineford1939,
            ov.names = paste("x", 1:9, sep = ""),
            nfactors = 3,
            rotation = list("target", target = target))
summary(fit2)

\dontrun{
# twolevel EFA with a different number of factors per level:
# one factor at the within level, two factors at the between level
fit3 <- efa(data = Demo.twolevel, ov.names = paste0("y", 1:6),
            cluster = "cluster", nfactors = list(c(1, 2)))
summary(fit3)
}
}

\name{lav_model}
\alias{lav_model_get_parameters}
\alias{lav_model_set_parameters}
\alias{lav_model_implied}
\alias{lav_model_vcov_se}
\title{lavaan model functions}
\description{Utility functions related to internal model representation (lavmodel)}
\usage{
# set/get free parameters
lav_model_set_parameters(lavmodel, x = NULL)
lav_model_get_parameters(lavmodel, glist = NULL, type = "free",
                         extra = TRUE, ...)

# compute model-implied statistics
lav_model_implied(lavmodel, glist = NULL, delta = TRUE, ...)

# compute standard errors
lav_model_vcov_se(lavmodel, lavpartable, vcov = NULL, boot = NULL,
                  mc = NULL, lavoptions = NULL, ...)
}
\arguments{
\item{lavmodel}{An internal representation of a lavaan model.}
\item{x}{Numeric. A vector containing the values of all the free model parameters.}
\item{glist}{List. A list of model matrices, similar to the output of
\code{lavInspect(object, "est")}.}
\item{type}{Character string. If \code{"free"}, only return the free model
parameters. If \code{"user"}, return all the parameters (free and fixed) as
they appear in the user-specified parameter table.}
\item{extra}{Logical. If \code{TRUE}, also include values for rows in
the parameter table where the operator is one of \code{":="}, \code{"=="},
\code{"<"} or \code{">"}.}
\item{delta}{Logical. If \code{TRUE}, and a Delta matrix is present in GLIST,
use the (diagonal) values of the Delta matrix to rescale the covariance matrix.
This is usually needed in the categorical setting to convert covariances to
correlations.}
\item{lavpartable}{A parameter table.}
\item{vcov}{Numeric matrix containing an estimate of the variance covariance matrix of the free model parameters.}
\item{boot}{Numeric matrix containing the bootstrap based parameter estimates (in the columns) for each bootstrap sample (in the rows).}
\item{mc}{Numeric matrix containing the Monte-Carlo based parameter estimates (in the columns) for each Monte-Carlo sample (in the rows).}
\item{lavoptions}{A named list. The Options slot from a lavaan object.}
\item{...}{To accept old argument names with dots or capitals. No other arguments are accepted.}
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
lavmodel <- fit@Model

est <- lav_model_get_parameters(lavmodel)
est
}

\name{lav_func}
\alias{lav_func_grad_complex}
\alias{lav_func_grad_simple}
\alias{lav_func_jacobian_complex}
\alias{lav_func_jacobian_simple}
\title{Utility Functions: Gradient and Jacobian}
\description{Utility functions for computing the gradient of a scalar-valued
function or the Jacobian of a vector-valued function by numerical approximation.}
\usage{
lav_func_grad_complex(func, x, h = .Machine$double.eps, ...,
                          fallback_simple = TRUE)
lav_func_jacobian_complex(func, x, h = .Machine$double.eps, ...,
                          fallback_simple = TRUE)

lav_func_grad_simple(func, x, h = sqrt(.Machine$double.eps), ...)
lav_func_jacobian_simple(func, x, h = sqrt(.Machine$double.eps), ...)
}
\arguments{
\item{func}{A real-valued function returning a numeric scalar or a
numeric vector.}
\item{x}{A numeric vector: the point(s) at which the gradient/Jacobian
of the function should be computed.}
\item{h}{Numeric value representing a small change in `x' when
computing the gradient/Jacobian.}
\item{...}{Additional arguments to be passed to the function `func'.}
\item{fallback_simple}{Logical. If TRUE, and the function evaluation fails,
we call the corresponding simple (non-complex) method instead.}
}
\details{
The complex versions use complex numbers to gain more precision, while
retaining the simplicity (and speed) of the simple forward method (see
references). These functions were added to lavaan (around 2012), when the
complex functionality was not yet part of the numDeriv package. They were
used internally, and made public in 0.5-17 at the request of other package
developers.
}
\references{
Squire, W. and Trapp, G. (1998). Using Complex Variables to Estimate
Derivatives of Real Functions. SIAM Review, 40(1), 110-112.
}
\examples{
# very accurate complex method
lav_func_grad_complex(func = exp, x = 1) - exp(1)

# less accurate forward method
lav_func_grad_simple(func = exp, x = 1) - exp(1)

# very accurate complex method
diag(lav_func_jacobian_complex(func = exp, x = c(1,2,3))) - exp(c(1,2,3))

# less accurate forward method
diag(lav_func_jacobian_simple(func = exp, x = c(1,2,3))) - exp(c(1,2,3))
}

\name{lav_plotinfo_positions}
\alias{lav_plotinfo_positions}
\title{Position the nodes in the diagram}
\usage{
lav_plotinfo_positions(plotinfo,
                       placenodes = NULL,
                       edgelabelsbelow = NULL,
                       group_covar_indicators = FALSE,
                       debug = FALSE)
}
\arguments{
\item{plotinfo}{The plotinfo structure as returned from \code{lav_model_plotinfo}.}
\item{placenodes}{optional list with members \code{\var{nodename} = c(row, column)},
\code{row} and \code{column} don't have to be integers.}
\item{edgelabelsbelow}{optional list with members \code{c(\var{nodename1},
\var{nodename2})}.}
\item{group_covar_indicators}{logical, should items with indicators which
have an explicit covariance link be placed in the same group, i.e. forced
to be on the same side of the diagram?}
\item{debug}{logical, print debug information?}
}
\value{
A plotinfo structure with modified \code{nodes} and \code{edges} 
data.frames, and an integer \code{mlrij} giving the position at which a line
should be drawn for multilevel models.

The data.frames have new columns defined as follows:
\enumerate{
\item \code{nodes}
  \describe{
    \item{rij}{index of the row where the node will be placed, initialized NA.}
    \item{kolom}{index of the column where the node will be placed,
    initialized NA.}
  }
\item \code{edges}
  \describe{
    \item{vananker}{character, anchor point for starting node, initialized NA.}
    \item{naaranker}{character, anchor point for destination node,
    initialized NA.}
    \item{controlpt.kol}{real, column position of control point if the
    edge has to be drawn as a quadratic Bezier curve.}
    \item{controlpt.row}{real, row position of control point if the
    edge has to be drawn as a quadratic Bezier curve.}
    \item{labelbelow}{logical, TRUE if label has to be positioned under 
    the line, initialized FALSE.}
  }
}
}
\description{
Computes the positions for the nodes and anchors and control points for the 
edges in the diagram.
}
\details{
This function \strong{tries} to arrange the nodes and the anchor points for the
edges in a way that keeps the diagram reasonably compact, while placing the
dependent variable of a regression to the right of the variables it depends on.
If the result is not what you want, you can inspect the plot with the
\code{lav_plotinfo_rgraph} or \code{lav_plot} functions and then place some
nodes at a different location (\code{placenodes}) and/or move some edge labels
to the other side of the edge (\code{labelsbelow}).

For multilevel models -- maximum two levels -- the nodes in block 2 are grouped
at the top of the diagram and those in block 1 at the bottom. The item 
\code{mlrij} gives the position of the separation between the blocks.
}

\seealso{
\code{\link{lav_model_plotinfo}}, \code{\link{lav_plotinfo_rgraph}},
\code{\link{lav_plot}}
}

\examples{
model <- 'alfa  =~ 1 * x1 + x2 + x3         # latent variable
          beta <~ x4 + x5 + x6              # composite
          gamma =~ 1 * x7 + x8 + x9         # latent variable
          Xi =~ 1 * x10 + x11 + x12 + x13   # latent variable
          # regressions
          Xi ~ v * alfa + t * beta + cc * 1
          alfa ~ tt * beta + ss * gamma + yy * Theta1
          # variances and covariances
          x2 ~~ cc25 * x5
          x3 ~~ cc36 * x6
          x3 ~~ cc34 * x4
          gamma ~~ 0.55 * gamma
          '
test <- lav_model_plotinfo(model)
(test_positioned <- lav_plotinfo_positions(test))
}

\name{lav_plotinfo_svgcode}
\alias{lav_plotinfo_svgcode}
\title{Creates code to show a diagram in svg format}
\usage{
lav_plotinfo_svgcode(plotinfo,
                     outfile = "",
                     sloped_labels = TRUE,
                     standalone = FALSE,
                     stroke_width = 2L,
                     font_size = 20L,
                     idx_font_size = 15L,
                     dy = 5L,
                     mlovcolors = c("lightgreen", "lightblue"),
                     lightness= 1,
                     font_family = "Latin Modern Math, arial, Arial, sans",
                     italic = TRUE,
                     auto_subscript = TRUE
                     )
}
\arguments{
\item{plotinfo}{A plotinfo structure as returned from \code{lav_plotinfo_positions}.}
\item{outfile}{A connection or a character string, the file to store the code.}
\item{sloped_labels}{logical, sloped labels for the edges.}
\item{standalone}{logical, should code be added to produce an html file with the
svg embedded in it? If FALSE (the default), the outfile (if specified) must have
the file extension 'svg'; if TRUE, the file extension must be 'htm' or 'html'.}
\item{stroke_width}{Value for stroke-width parameter in svg.}
\item{font_size}{An integer specifying the normal font size to use.}
\item{idx_font_size}{An integer specifying the font size to use for a subscript.}
\item{dy}{An integer specifying the distance by which to move the baseline of the subscript.}
\item{mlovcolors}{Array of two colors for distinguishing ov nodes with the same
name across the levels of a multilevel model.}
\item{lightness}{A scalar factor to modify the distances between nodes.}
\item{font_family}{Fonts to be tried for rendering the labels. The first one,
Latin Modern Math, is close to the default font used in tikz for mathematical
expressions and can be downloaded from
\href{https://ctan.org/tex-archive/fonts/lm-math}{CTAN}.}
\item{italic}{Should labels be in an italic font? If FALSE, the labels are
shown in mathrm font.}
\item{auto_subscript}{Logical, see \code{\link{lav_label_code}}.}
}
\value{
NULL (invisible)}
\description{
Creates svg code to show a diagram of the model.
}
\seealso{
\code{\link{lav_plotinfo_positions}}
}
\examples{
model <- 'alpha  =~ x1 + x2 + x3       # latent variable
          beta <~ x4 + x5 + x6         # composite
          gamma =~ x7 + x8 + x9        # latent variable
          Xi =~ x10 + x11 + x12 + x13  # latent variable
          # regressions
          Xi ~ v * alpha + t * beta + 1
          alpha ~ yy * Theta1 + tt * beta + ss * gamma
          '
test <- lav_model_plotinfo(model)
test <- lav_plotinfo_positions(test)
lav_plotinfo_svgcode(test) # no file given, so output to R console
modelml <- '
        level: 1
        fw =~ 1*y_1 + y_2 + y_3 + y_5
        level: 2
        fb =~ 1*y_1 + y_2 + y_3 + y_4
        y_2 ~~ cv24 * y_4
  '
testml <- lav_model_plotinfo(modelml)
testml <- lav_plotinfo_positions(testml)
# in the line hereunder no file is given, so output to R console
lav_plotinfo_svgcode(testml, sloped_labels = FALSE, standalone = TRUE,
            auto_subscript = FALSE)
\dontrun{
# example creating html file with the above diagrams
zz <- file("demosvg.html", open="w")
writeLines(c(
  '<!DOCTYPE html>',
  '<html>',
  '<body>',
  '<h2>SVG diagrams created by lav_plot R package</h2>'),
  zz)
lav_plotinfo_svgcode(test, outfile = "temp.svg")
tmp <- readLines("tmp.svg")
writeLines(tmp, zz)
writeLines("<br />", zz)
lav_plotinfo_svgcode(testml, outfile = "temp.svg", sloped_labels = FALSE,
             standalone = TRUE)
tmp <- readLines("tmp.svg")
writeLines(tmp, zz)
writeLines(c("</body>", "</html>"), zz)
close(zz)
browseURL("demosvg.html")
}
}

\name{lav_plot}
\alias{lav_plot}
\title{Create diagram code for tikz or svg, or produce an R plot}
\usage{
lav_plot(model = NULL,
        infile = NULL,
        varlv = FALSE,
        placenodes = NULL,
        edgelabelsbelow = NULL,
        group_covar_indicators = FALSE,
        common_opts = list(sloped_labels = TRUE,
                           mlovcolors = c("lightgreen", "lightblue"),
                           italic = TRUE,
                           lightness = 1,
                           auto_subscript = TRUE),
        rplot = list(outfile = "",
                     addgrid = TRUE),
        tikz = list(outfile = "",
                    cex = 1.3,
                    standalone = FALSE),
        svg = list(outfile = "",
                   stroke_width = 2L,
                   font_size = 20L,
                   idx_font_size = 15L,
                   dy = 5L,
                   font_family = "Latin Modern Math, arial, Arial, sans",
                   standalone = FALSE)
        )
}
\arguments{
\item{model}{A character vector specifying the model in lavaan syntax or a list
(or data.frame) with at least members lhs, op, rhs, label and fixed or a fitted
lavaan object (in which case the \code{ParTable} object is extracted and column
\code{est} is used as value to show).
Should be \code{NULL} if infile is given.}
\item{infile}{A character string specifying the file that contains the model
syntax.}
\item{varlv}{A logical indicating that the (residual) variance of a variable
should be plotted as a separate latent variable (with a smaller circle than
ordinary latent variables). In this case, a covariance between two such variables
is plotted as a covariance between their variance latent variables.}
\item{placenodes}{optional list with members \code{\var{nodename} = c(row, column)},
\code{row} and \code{column} don't have to be integers.}
\item{edgelabelsbelow}{optional list with members \code{c(\var{nodename1},
\var{nodename2})}.}
\item{group_covar_indicators}{logical, should items with indicators which
have an explicit covariance link be placed in the same group, i.e. forced
to be on the same side of the diagram?}
\item{common_opts}{options common to the three types of generated plots.}
\item{rplot}{options for creating Rplot,
see \code{\link{lav_plotinfo_rgraph}}.}
\item{tikz}{options for creating code for tikz plot,
see \code{\link{lav_plotinfo_tikzcode}}.}
\item{svg}{options for creating code for svg plot,
see \code{\link{lav_plotinfo_svgcode}}.}
}
\details{
If \code{rplot} is specified, or if neither \code{tikz} nor \code{svg} is
specified, an R plot is generated, and stored in a png file if the outfile
member of \code{rplot} is set. If \code{tikz} is specified, the code for a tikz
diagram is stored in the specified \code{outfile}; the same applies to svg.

The lav_plot command \strong{tries} to create a nice plot from the input model,
but the variable names should be kept short, and the nodes in the plot may
sometimes need to be rearranged. As an example, the (slightly modified) example
of the \code{sem} function in \code{lavaan} produces the first plot shown below.
Using the placenodes argument produces the second plot.
\if{html}{\figure{semex0.svg}{options: width=500 alt="sem example before"}}
\if{latex}{\figure{semex0.pdf}{options: width=3in}}
\if{html}{\figure{semex.svg}{options: width=500 alt="sem example after"}}
\if{latex}{\figure{semex.pdf}{options: width=3in}}

More details on the parameters can be found in the help for
the \code{lav_\dots} functions.
}
\value{
NULL, invisible
}
\description{
Creates code to draw a diagram in tikz or svg, plots the diagram, or stores
the diagram in a png file.
}
\seealso{
\code{\link{lav_model_plotinfo}}, \code{\link{lav_plotinfo_positions}},
\code{\link{lav_plotinfo_rgraph}}, \code{\link{lav_plotinfo_tikzcode}},
\code{\link{lav_plotinfo_svgcode}}
}
\examples{
model <- 'alpha11  =~ 1 * x1 + x2 + x3     # latent variable
          alpha12 <~ x4 + x5 + x6          # composite
          gamma =~ 1 * x7 + x8 + x9        # latent variable
          xi =~ 1 * x10 + x11 + x12 + x13  # latent variable
          x1 ~~ x3
          x2 ~~ epsilon2 * x2
          x12 ~~ epsilon12 * x12
          x4 ~~ epsilon4 * x4
          x7 ~~ x9
          x10 ~~ x11 + x13
          gamma ~~ 0.7 * xi
          # regressions
          xi ~ v * alpha11 + t * alpha12 + 1
          alpha11 ~ yy * Theta1 + tt1 * 0.12 * alpha12 + ss * gamma
          Theta1 ~~ alpha12
          '
lav_plot(model)
lav_plot(model,
        placenodes=list(Theta1 = c(2, 2.5)),
        tikz = list(outfile=stdout()))
modelml <- '
        level: 1
        fw =~ y_1 + y_2 + y_3 + y_4
        level: 2
        fb =~ y_1 + y_2 + y_3 + y_5
        y_2 ~~ cv24 * y_5
  '
tikzcodeml <- lav_plot(modelml,
              common_opts = list(auto_subscript = FALSE),
              svg = list(outfile=stdout())
              )

\dontrun{
# example creating tex file with the above models
zz <- file("testtikz.tex", open="w")
writeLines(c(
  '\\documentclass{article}',
  '\\usepackage{amsmath, amssymb}',
  '\\usepackage{amsfonts}',
  '\\usepackage[utf8]{inputenc}',
  '\\usepackage[english]{babel}',
  '\\usepackage{xcolor}',
  '\\usepackage{color}',
  '\\usepackage{tikz}',
  '\\usetikzlibrary {shapes.geometric}',
  '\\begin{document}'),
  zz)
lav_plot(model,
        tikz = list(outfile = "tmp.tex")
        )
tmp <- readLines("tmp.tex")
writeLines(tmp, zz)
writeLines(" ", zz)
lav_plot(modelml,
        common_opts = list(sloped_labels = FALSE,
                           mlovcolors = c("lightgreen", "lightblue")),
        tikz = list(outfile = "tmp.tex", cex = 1.4)
        )
tmp <- readLines("tmp.tex")
writeLines(tmp, zz)
writeLines("\\end{document}", zz)
close(zz)
openPDF <- function(f) {
  os <- .Platform$OS.type
  if (os=="windows")
    shell.exec(normalizePath(f))
  else {
    pdf <- getOption("pdfviewer", default='')
    if (nchar(pdf)==0)
      stop("The 'pdfviewer' option is not set. Use options(pdfviewer=...)")
    system2(pdf, args=c(f))
  }
}
tools::texi2dvi("testtikz.tex", pdf = TRUE, clean = TRUE)
openPDF("testtikz.pdf")

# example creating html file with the above diagrams
zz <- file("demosvg.html", open="w")
writeLines(c(
  '<!DOCTYPE html>',
  '<html>',
  '<body>',
  '<h2>SVG diagrams created by lav_plot R package</h2>'),
  zz)
lav_plot(model,
        svg =  list(outfile = "temp.svg")
        )
tmp <- readLines("tmp.svg")
writeLines(tmp, zz)
writeLines("<br />", zz)
lav_plot(modelml,
        common_opts = list(sloped_labels = FALSE),
        tikz = list(outfile = "tmp.svg")
        )
tmp <- readLines("tmp.svg")
writeLines(tmp, zz)
writeLines(c("</body>", "</html>"), zz)
close(zz)
browseURL("demosvg.html")
}
}

\name{lav_mplus_syntax_model}
\alias{lav_mplus_syntax_model}
\alias{mplus2lavaan.modelSyntax}
\title{Convert Mplus model syntax to lavaan}
\description{
Converts Mplus model syntax into lavaan model syntax.}
\usage{
lav_mplus_syntax_model(syntax)
mplus2lavaan.modelSyntax(syntax)
}
\arguments{
\item{syntax}{A character vector containing Mplus model syntax to be
  converted to lavaan model syntax. Note that parsing Mplus syntax often
  requires correct usage of newline characters. If \code{syntax} is a
  vector of multiple strings, these will be joined with newlines prior
  to conversion. Alternatively, \code{\\n} characters can be included
  inline in \code{syntax}.} 
}
\value{
A character string of converted \code{lavaan} model syntax.
}
\author{Michael Hallquist}
\seealso{\code{\link{lav_mplus_lavaan}}}
\examples{
\dontrun{
syntax <- '
    f1 BY x1*1 x2 x3;
    x1 WITH x2;
    x3 (1);
    x2 (1);
'
lavSyntax <- lav_mplus_syntax_model(syntax)
cat(lavSyntax)
}
}

\name{lavNames}
\alias{lavNames}
\alias{lavaanNames}
\title{lavaan Names}
\description{
Extract variable names from a fitted lavaan object.}
\usage{
lavNames(object, type = "ov", ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{type}{Character. The type of variables whose names should be extracted. 
See details for a complete list.}
\item{...}{Additional selection criteria. For example, \code{"group = 2L"}
(in a multiple-group analysis) restricts the selection to the variables
included in the model for the second group.}
}
\details{
The order of the variable names, as returned by \code{lav_object_vnames},
determines the order in which the variables are listed in the parameter
table, and therefore also in the summary output.


The following variable types are available:
\itemize{
  \item \code{"ov"}: observed variables

  \item \code{"ov.x"}: (pure) exogenous observed variables (no mediators)

  \item \code{"ov.nox"}: non-exogenous observed variables

  \item \code{"ov.model"}: modeled observed variables (joint vs conditional)

  \item \code{"ov.y"}: (pure) endogenous variables (dependent only) (no mediators)

  \item \code{"ov.num"}: numeric observed variables

  \item \code{"ov.ord"}: ordinal observed variables

  \item \code{"ov.ind"}: observed indicators of latent variables

  \item \code{"ov.orphan"}: isolated observed variables (only their intercepts or variances appear in the model syntax)

  \item \code{"ov.interaction"}: interaction terms (defined by the colon operator)

  \item \code{"th"}: threshold names (ordinal variables only)

  \item \code{"th.mean"}: threshold names (ordinal and numeric variables, if any)

  \item \code{"lv"}: latent variables

  \item \code{"lv.regular"}: measured latent variables (defined by =~ only)

  \item \code{"lv.composite"}: composites (defined by <~ only)

  \item \code{"lv.x"}: (pure) exogenous variables

  \item \code{"lv.y"}: (pure) endogenous variables

  \item \code{"lv.nox"}: non-exogenous latent variables

  \item \code{"lv.nonnormal"}: latent variables with non-normal indicators

  \item \code{"lv.interaction"}:  interaction terms at the latent level

  \item \code{"eqs.y"}: variables that appear as dependent variables in a
                        regression formula (but not indicators of latent
                        variables)
 
  \item \code{"eqs.x"}: variables that appear as independent variables in
                        a regression formula
}
}
\seealso{\code{\link{lavParTable}}, \code{\link{parTable}}}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
lavNames(fit, "ov")
}

\name{lavBootstrap}
\alias{bootstrapLavaan}
\alias{lavBootstrap}
\title{Bootstrapping a Lavaan Model}
\description{Bootstrap the LRT, or any other statistic (or vector
of statistics) you can extract from a fitted lavaan object.}
\usage{
lavBootstrap(object, r = 1000L, type = "ordinary", verbose = FALSE, 
                fun = "coef", keep_idx = FALSE,
                parallel = c("no", "multicore", "snow"),
                ncpus = min(2L, max(1L, parallel::detectCores() - 2L, na.rm = TRUE)),
                cl = NULL, iseed = NULL, h0_rmsea = NULL, ...)
bootstrapLavaan(object, r = 1000L, type = "ordinary", verbose = FALSE, 
                fun = "coef", keep_idx = FALSE,
                parallel = c("no", "multicore", "snow"),
                ncpus = min(2L, max(1L, parallel::detectCores() - 2L, na.rm = TRUE)),
                cl = NULL, iseed = NULL, h0_rmsea = NULL, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{r}{Integer. The number of bootstrap draws.}
\item{type}{If \code{"ordinary"} or \code{"nonparametric"}, the usual (naive)
bootstrap method is used. If \code{"bollen.stine"}, the data is first
transformed such that the null hypothesis holds exactly in the resampling
space. If \code{"yuan"}, the data is first transformed by combining data and
theory (model), such that the resampling space is closer to the population
space. Note that both \code{"bollen.stine"} and \code{"yuan"} require the data
to be continuous. They will not work with ordinal data.  If
\code{"parametric"}, the parametric bootstrap approach is used; currently, this
is only valid for continuous data following a multivariate normal distribution.
See references for more details. If the fitted model uses clustered or
multilevel data (that is, a \code{cluster=} argument was used, either for a
two-level model or to obtain cluster-robust standard errors), only the
\code{"ordinary"} (nonparametric) bootstrap is available, and whole clusters
(rather than individual observations) are resampled; see Details.}
\item{fun}{A function which when applied to the \code{\linkS4class{lavaan}} 
object returns a vector containing the statistic(s) of interest.
The default is \code{fun="coef"}, returning the estimated values of the
free parameters in the model.}
\item{\dots}{Other named arguments for \code{fun} which are passed
    unchanged each time it is called.}
\item{verbose}{If \code{TRUE}, show information for each bootstrap draw.}
\item{keep_idx}{If \code{TRUE}, store the indices of each bootstrap run (i.e.,
the observations that were used for this bootstrap run) as an attribute.}
\item{parallel}{The type of parallel operation to be used (if any).  If
missing, the default is \code{"no"}.  }
\item{ncpus}{Integer: number of processes to be used in parallel operation.
By default, this is 2 (following CRAN policy). You may set this to a higher
value (for example, the number of cores as detected by
\code{parallel::detectCores()} minus one) if you wish to use more processes.}
\item{cl}{An optional \pkg{parallel} or \pkg{snow} cluster for use if
\code{parallel = "snow"}.  If not supplied, a cluster on the local machine is
created for the duration of the \code{lavBootstrap} call.}
\item{iseed}{An integer to set the seed. Or NULL if no reproducible results are
  needed. This works for both serial (non-parallel) and parallel settings.
  Internally, \code{RNGkind()} is set to \code{"L'Ecuyer-CMRG"} if
  \code{parallel = "multicore"}. If \code{parallel = "snow"} (under windows),
  \code{parallel::clusterSetRNGStream()} is called which
  automatically switches to \code{"L'Ecuyer-CMRG"}. When \code{iseed} is not 
  NULL, \code{.Random.seed} (if it exists) in the global environment is 
  left untouched.}
\item{h0_rmsea}{Only used if \code{type="yuan"}. Allows one to do the Yuan
bootstrap under the hypothesis that the population RMSEA equals a specified
value.}
}
\author{Yves Rosseel and Leonard Vanbrabant. Ed Merkle contributed
Yuan's bootstrap. Improvements to Yuan's bootstrap were contributed
by Hao Wu and Chuchu Cheng. The handling of iseed was contributed by
Shu Fai Cheung.}
\value{For \code{lavBootstrap()}, the bootstrap distribution of the value(s)
returned by \code{fun}, when the object can be simplified to a vector.
}
\details{The fun function can return either a scalar or a numeric vector.
This function can be an existing function (for example \code{coef}) or
can be a custom defined function. For example:
\preformatted{
myFUN <- function(x) {
    # require(lavaan)
    modelImpliedCov <- fitted(x)$cov
    vech(modelImpliedCov)
}
}
If \code{parallel="snow"}, it is imperative that the \code{require(lavaan)}
is included in the custom function.

When the fitted model was estimated with clustered or multilevel data (i.e., a
\code{cluster=} argument was used), the bootstrap automatically switches to a
\emph{cluster bootstrap}: instead of resampling individual observations, whole
(level-2) clusters are resampled with replacement, keeping all level-1
observations within a selected cluster together. The number of clusters is held
fixed (equal to the number of clusters in the original data), while the total
number of observations may vary from one bootstrap sample to the next. A cluster
that is drawn more than once is treated as several distinct clusters. This
resampling scheme preserves the within-cluster dependence and is the
appropriate bootstrap both for two-level models and for single-level models with
cluster-robust standard errors. It is currently only available for a single
cluster (level-2) variable, and the \code{"bca"} confidence interval type
(in \code{\link{parameterEstimates}}) is not supported in this case.

When the fitted model contains rotated exploratory factors (EFA or ESEM; that
is, the model syntax uses the \code{efa()} modifier together with a
\code{rotation=} method other than \code{"none"}), bootstrap standard errors and
confidence intervals (requested via \code{se = "bootstrap"} when fitting the
model) are computed for the \emph{rotated} solution. Each bootstrap sample is
refit and rotated, and its factors are then aligned (reordered and reflected) to
the original (full-sample) rotated solution before their variability is used.
This alignment resolves the rotational indeterminacy (the sign and order of the
exploratory factors), which would otherwise inflate the standard errors. Note
that the raw bootstrap coefficients returned by \code{lavBootstrap(object,
fun = "coef")} are \emph{not} aligned to a reference solution in this way.}
\references{
Bollen, K. and Stine, r. (1992) Bootstrapping Goodness of Fit Measures in
        Structural Equation Models. Sociological Methods and Research, 21,
        205--229.

Yuan, K.-H., Hayashi, K., & Yanagihara, H. (2007). A class of population
covariance matrices in the bootstrap approach to covariance structure analysis.
Multivariate Behavioral Research, 42, 261--281.

Field, C. A., & Welsh, A. H. (2007). Bootstrapping clustered data. Journal of
the Royal Statistical Society: Series B, 69, 369--390.
}

\examples{
# fit the Holzinger and Swineford (1939) example
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939, se="none")

# get the test statistic for the original sample
T.orig <- fitMeasures(fit, "chisq")

# bootstrap to get bootstrap test statistics
# we only generate 10 bootstrap samples in this example; in practice
# you may wish to use a much higher number
T.boot <- lavBootstrap(fit, r=10, type="bollen.stine",
                          fun=fitMeasures, fit.measures="chisq")

# compute a bootstrap based p-value
pvalue.boot <- length(which(T.boot > T.orig))/length(T.boot)
}

\name{lav_matrix}
\alias{lav_mat_vec}
\alias{lav_mat_vecr}
\alias{lav_mat_vech}
\alias{lav_mat_vechr}
\alias{lav_mat_vechu}
\alias{lav_mat_vechru}
\alias{lav_mat_vech_idx}
\alias{lav_mat_vech_row_idx}
\alias{lav_mat_vech_col_idx}
\alias{lav_mat_vechr_idx}
\alias{lav_mat_vechu_idx}
\alias{lav_mat_vechru_idx}
\alias{lav_mat_vech_rev}
\alias{lav_mat_vechru_rev}
\alias{lav_mat_upper2full}
\alias{lav_mat_vechr_rev}
\alias{lav_mat_vechu_rev}
\alias{lav_mat_lower2full}
\alias{lav_mat_diag_idx}
\alias{lav_mat_diagh_idx}
\alias{lav_mat_antidiag_idx}
\alias{lav_mat_dup}
\alias{lav_mat_dup_pre}
\alias{lav_mat_dup_post}
\alias{lav_mat_dup_pre_post}
\alias{lav_mat_dup_ginv}
\alias{lav_mat_dup_ginv_pre}
\alias{lav_mat_dup_ginv_post}
\alias{lav_mat_dup_ginv_pre_post}
\alias{lav_mat_com}
\alias{lav_mat_com_pre}
\alias{lav_mat_com_post}
\alias{lav_mat_com_pre_post}
\alias{lav_mat_com_mn_pre}
\alias{lav_mat_sym_sqrt}
\alias{lav_mat_ortho_complement}
\alias{lav_mat_bdiag}
\alias{lav_mat_trace}
\alias{lav_mat_cov}

\title{Utility Functions: Matrices and Vectors}
\description{Utility functions for Matrix and Vector operations.}
\usage{
# matrix to vector
lav_mat_vec(a)
lav_mat_vecr(a)
lav_mat_vech(s, diagonal = TRUE)
lav_mat_vechr(s, diagonal = TRUE)

# matrix/vector indices
lav_mat_vech_idx(n = 1L, diagonal = TRUE)
lav_mat_vech_row_idx(n = 1L, diagonal = TRUE)
lav_mat_vech_col_idx(n = 1L, diagonal = TRUE)
lav_mat_vechr_idx(n = 1L, diagonal = TRUE)
lav_mat_vechru_idx(n = 1L, diagonal = TRUE)
lav_mat_diag_idx(n = 1L)
lav_mat_diagh_idx(n = 1L)
lav_mat_antidiag_idx(n = 1L)

# vector to matrix
lav_mat_vech_rev(x, diagonal = TRUE)
lav_mat_vechru_rev(x, diagonal = TRUE) 
lav_mat_upper2full(x, diagonal = TRUE)
lav_mat_vechr_rev(x, diagonal = TRUE)
lav_mat_vechu_rev(x, diagonal = TRUE)
lav_mat_lower2full(x, diagonal = TRUE)

# the duplication matrix
lav_mat_dup(n = 1L)
lav_mat_dup_pre(a = matrix(0,0,0))
lav_mat_dup_post(a = matrix(0,0,0))
lav_mat_dup_pre_post(a = matrix(0,0,0))
lav_mat_dup_ginv(n = 1L)
lav_mat_dup_ginv_pre(a = matrix(0,0,0))
lav_mat_dup_ginv_post(a = matrix(0,0,0))
lav_mat_dup_ginv_pre_post(a = matrix(0,0,0))

# the commutation matrix
lav_mat_com(m = 1L, n = 1L)
lav_mat_com_pre(a = matrix(0,0,0))
lav_mat_com_post(a = matrix(0,0,0))
lav_mat_com_pre_post(a = matrix(0,0,0))
lav_mat_com_mn_pre(a, m = 1L, n = 1L)

# sample statistics
lav_mat_cov(y, mu = NULL)

# other matrix operations
lav_mat_sym_sqrt(s = matrix(0,0,0))
lav_mat_ortho_complement(a = matrix(0,0,0))
lav_mat_bdiag(...)
lav_mat_trace(..., check = TRUE)
}
\arguments{
\item{a}{A general matrix.}
\item{s}{A symmetric matrix.}
\item{y}{A matrix representing a (numeric) dataset.}
\item{diagonal}{Logical. If TRUE, include the diagonal.}
\item{n}{Integer. When it is the only argument, the dimension of a square 
matrix. If m is also provided, the number of columns of the matrix.}
\item{m}{Integer. The number of rows of a matrix.}
\item{x}{Numeric. A vector.}
\item{mu}{Numeric. If given, use mu (instead of sample mean) to center, before 
 taking the crossproduct.}
\item{...}{One or more matrices, or a list of matrices.}
\item{check}{Logical. If \code{check = TRUE}, we check if the (final) matrix 
is square.}
}
\details{
These are a collection of lower-level matrix and vector functions
that are used throughout the lavaan code. They are made public at the
request of package developers. Below is a brief description of what
they do:

The \code{lav_mat_vec} function implements the vec operator (for
'vectorization') and transforms a matrix into a vector by stacking the
columns of the matrix one underneath the other.

The \code{lav_mat_vecr} function is similar to the \code{lav_mat_vec}
function but transforms a matrix into a vector by stacking the
rows of the matrix one underneath the other.

The \code{lav_mat_vech} function implements the vech operator
(for 'half vectorization') and transforms a symmetric matrix
into a vector by stacking the columns of the matrix one underneath the
other, but eliminating all supradiagonal elements. If diagonal = FALSE,
the diagonal elements are also eliminated.

The \code{lav_mat_vechr} function is similar to the \code{lav_mat_vech} 
function but transforms a matrix into a vector by stacking the
rows of the matrix one underneath the other, eliminating all 
supradiagonal elements.

The \code{lav_mat_vech_idx} function returns the vector indices of the lower
triangular elements of a symmetric matrix of size n, column by column.

The \code{lav_mat_vech_row_idx} function returns the row indices of the
lower triangular elements of a symmetric matrix of size n.

The \code{lav_mat_vech_col_idx} function returns the column indices of the
lower triangular elements of a symmetric matrix of size n.

The \code{lav_mat_vechr_idx} function returns the vector indices of the
lower triangular elements of a symmetric matrix of size n, row by row.

The \code{lav_mat_vechu_idx} function returns the vector indices of the
upper triangular elements of a symmetric matrix of size n, column by column.

The \code{lav_mat_vechru_idx} function returns the vector indices
of the upper triangular elements of a symmetric matrix of size n, row by row.

The \code{lav_mat_diag_idx} function returns the vector indices of the
diagonal elements of a symmetric matrix of size n.

The \code{lav_mat_diagh_idx} function returns the vector indices of
the lower part of a symmetric matrix of size n.

The \code{lav_mat_antidiag_idx} function returns the vector indices of
the anti diagonal elements of a symmetric matrix of size n.

The \code{lav_mat_vech_rev} function (alias:
\code{lav_mat_vechru_rev} and \code{lav_mat_upper2full}) creates a
symmetric matrix, given only upper triangular elements, row by row. If
diagonal = FALSE, a diagonal with zero elements is added.

The \code{lav_mat_vechr_rev} (alias: \code{lav_mat_vechu_rev} and
\code{lav_mat_lower2full}) creates a symmetric matrix, given only the lower
triangular elements, row by row. If diagonal = FALSE, a diagonal with zero
elements is added.

The \code{lav_mat_dup} function generates the duplication matrix
for a symmetric matrix of size n. This matrix duplicates the elements in
vech(S) to create vec(S) (where S is symmetric). This matrix is very
sparse, and should probably never be explicitly created. Use one of 
the functions below.

The \code{lav_mat_dup_pre} function computes the product of the
transpose of the duplication matrix and a matrix A. The A matrix should have
n*n rows, where n is an integer. The duplication matrix is not explicitly
created.

The \code{lav_mat_dup_post} function computes the product of a
matrix A with the duplication matrix. The A matrix should have n*n columns,
where n is an integer. The duplication matrix is not explicitly created.

The \code{lav_mat_dup_pre_post} function first pre-multiplies a
matrix A with the transpose of the duplication matrix, and then post multiplies
the result again with the duplication matrix. A must be square matrix with n*n
rows and columns, where n is an integer. The duplication matrix is not
explicitly created.

The \code{lav_mat_dup_ginv} function computes the generalized
inverse of the duplication matrix. The matrix removes the duplicated elements
in vec(S) to create vech(S). This matrix is very sparse, and should probably
never be explicitly created. Use one of the functions below.

The \code{lav_mat_dup_ginv_pre} function computes the product of the
generalized inverse of the duplication matrix and a matrix A with n*n rows,
where n is an integer. The generalized inverse of the duplication matrix
is not explicitly created.

The \code{lav_mat_dup_ginv_post} function computes the product of a
matrix A (with n*n columns, where n is an integer) and the transpose of the
generalized inverse of the duplication matrix. The generalized inverse of the
duplication matrix is not explicitly created.

The \code{lav_mat_dup_ginv_pre_post} function first pre-multiplies
a matrix A with the transpose of the generalized inverse of the duplication
matrix, and then post multiplies the result again with the transpose of the
generalized inverse matrix. The matrix A must be square with n*n rows and
columns, where n is an integer. The generalized inverse of the duplication
matrix is not explicitly created.

The \code{lav_mat_com} function computes the commutation matrix, a
permutation matrix that transforms vec(A) (with m rows and n columns)
into vec(t(A)).

The \code{lav_mat_com_pre} function computes the product of the
commutation matrix with a matrix A, without explicitly creating the commutation
matrix. The matrix A must have n*n rows, where n is an integer.

The \code{lav_mat_com_post} function computes the product of a 
matrix A with the commutation matrix, without explicitly creating the 
commutation matrix. The matrix A must have n*n rows, where n is an integer.

The \code{lav_mat_com_pre_post} function first pre-multiplies
a matrix A with the commutation matrix, and then post multiplies the result again with the commutation matrix, without explicitly creating the 
commutation matrix. The matrix A must have n*n rows, where n is an integer.

The \code{lav_mat_com_mn_pre} function computes the product of the
commutation matrix with a matrix A, without explicitly creating the commutation
matrix. The matrix A must have m*n rows, where m and n are integers.

The \code{lav_mat_cov} function computes the sample covariance matrix of
its input matrix, where the elements are divided by N (the number of rows).

The \code{lav_mat_sym_sqrt} function computes the square root of a
positive definite symmetric matrix (using an eigen decomposition).  If some of
the eigenvalues are negative, they are silently fixed to zero.

The \code{lav_mat_ortho_complement} function computes an orthogonal
complement of the matrix A, using a qr decomposition.

The \code{lav_mat_bdiag} function constructs a block diagonal matrix from
its arguments.

The \code{lav_mat_trace} function computes the trace (the sum of the
diagonal elements) of a single (square) matrix, or if multiple matrices are
provided (either as a list, or as multiple arguments), we first compute their
product (which must result in a square matrix), and then we compute the trace; 
if \code{check = TRUE}, we check if the (final) matrix is square.
}
\references{
Magnus, J. R. and H. Neudecker (1999). Matrix Differential Calculus
with Applications in Statistics and Econometrics, Second Edition, John Wiley.
}
\examples{
# upper elements of a 3 by 3 symmetric matrix (row by row)
x <- c(30, 16, 5, 10, 3, 1)
# construct full symmetric matrix
S <- lav_mat_upper2full(x)

# compute the normal theory `Gamma' matrix given a covariance
# matrix (S), using the formula: Gamma = 2 * D^{+} (S %x% S) t(D^{+})
Gamma.NT <- 2 * lav_mat_dup_ginv_pre_post(S \%x\% S)
Gamma.NT
}

\name{lavPredictY_cv}
\alias{lavPredictY_cv}
\title{Determine an optimal lambda penalty value through cross-validation}
\description{
This function can be used to determine an optimal lambda value for the
\code{lavPredictY} function, based on cross-validation.}
\usage{
lavPredictY_cv(object, data = NULL,
              xnames = lav_object_vnames(object, "ov.x"),
              ynames = lav_object_vnames(object, "ov.y"),
              n_folds = 10L,
              lambda_seq = seq(0, 1, 0.1),
              ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{data}{A data.frame, containing the same variables as the data.frame that
was used when fitting the model in \code{object}.}
\item{xnames}{The names of the observed variables that should be treated as the
x-variables. Can also be a list to allow for a separate set of variable names
per group (or block).}
\item{ynames}{The names of the observed variables that should be treated as the
y-variables. It is for these variables that the function will predict the
(model-based) values for each observation. Can also be a list to allow for a
separate set of variable names per group (or block).}
\item{n_folds}{Integer. The number of folds to be used during
cross-validation.}
\item{lambda_seq}{An R \code{seq()} containing the range of lambda penalty
values to be tested during cross-validation.}
\item{...}{To support old argument names.}
}
\details{
This function determines an optimal lambda value for use with
\code{lavPredictY}, in order to improve prediction accuracy.
}
\seealso{
\code{\link{lavPredictY}} to predict the values of (observed) y-variables given
the values of (observed) x-variables in a structural equation model.
}
\references{
de Rooij, M., Karch, J.D., Fokkema, M., Bakk, Z., Pratiwi, B.C, and
Kelderman, H. (2022) SEM-Based Out-of-Sample Predictions,
Structural Equation Modeling: A Multidisciplinary Journal.
\doi{10.1080/10705511.2022.2061494}

Molina, M. D., Molina, L., & Zappaterra, M. W. (2024). Aspects of Higher
Consciousness: A Psychometric Validation and Analysis of a New Model of
Mystical Experience. \doi{10.31219/osf.io/cgb6e}
}
\examples{
\dontrun{
colnames(PoliticalDemocracy) <- c("z1", "z2", "z3", "z4",
                                  "y1", "y2", "y3", "y4",
                                  "x1", "x2", "x3")

model <- '
  # latent variable definitions
  ind60 =~ x1 + x2 + x3
  dem60 =~ z1 + z2 + z3 + z4
  dem65 =~ y1 + y2 + y3 + y4
  # regressions
  dem60 ~ ind60
  dem65 ~ ind60 + dem60
  # residual correlations
  z1 ~~ y1
  z2 ~~ z4 + y2
  z3 ~~ y3
  z4 ~~ y4
  y2 ~~ y4
'
fit <- sem(model, data = PoliticalDemocracy, meanstructure = TRUE)

percent <- 0.5
nobs <- lavInspect(fit, "ntotal")
idx <- sort(sample(x = nobs, size = floor(percent*nobs)))

xnames = c("z1", "z2", "z3", "z4", "x1", "x2", "x3")
ynames = c("y1", "y2", "y3", "y4")

reg.results <- lavPredictY_cv(
    fit,
    PoliticalDemocracy[-idx, ],
    xnames = xnames,
    ynames = ynames,
    n_folds = 10L,
    lambda_seq = seq(from = .6, to = 2.5, by = .1)
)
lam <- reg.results$lambda.min

lavPredictY(fit, newdata = PoliticalDemocracy[idx,],
                 ynames  = ynames,
                 xnames  = xnames,
                 lambda  = lam)
}
}

\name{parTable}
\alias{parameterTable}
\alias{parametertable}
\alias{parTable}
\alias{partable}
\title{Parameter Table}
\description{
Show the parameter table of a fitted model.}
\usage{
parameterTable(object)
parTable(object)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
}
\value{
  A \code{data.frame} containing the model parameters. This is
  simply the output of the \code{\link{lavParTable}} function 
  coerced to a \code{data.frame} (with \code{stringsAsFactors = FALSE}).
}
\seealso{\code{\link{lavParTable}}.}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
parTable(fit)
}

\name{lavOptions}
\alias{lavOptions}
\alias{lavoptions}
\title{lavaan Options}
\description{
Show the default options used by the \code{lavaan()} function. The
options can be changed by passing 'name = value' arguments to the
\code{lavaan()} function call, where they are added to the '...'
argument.
}
\usage{
lavOptions(x = NULL, default = NULL, mimic = "lavaan")
}
\arguments{
\item{x}{Character. A character string holding an option name, or a character
string vector holding multiple option names. All option names are converted to
lower case.}
\item{default}{If a single option is specified but not available, this value
is returned.}
\item{mimic}{Not used for now.}
}
\details{
This is the full list of options that are accepted by the \code{lavaan()}
function, organized in several sections:


Model features:

\describe{
    \item{\code{meanstructure}:}{If \code{TRUE}, the means of the observed
      variables enter the model. If \code{"default"}, the value is set based
      on the user-specified model, and/or the values of other arguments.}
    \item{\code{int.ov.free}:}{If \code{FALSE}, the intercepts of the 
      observed variables are fixed to zero.}
    \item{\code{int.lv.free}:}{If \code{FALSE}, the intercepts of the latent 
      variables are fixed to zero.}
    \item{\code{conditional.x}:}{If \code{TRUE}, we set up the model 
      conditional on the exogenous `x' covariates; the model-implied sample 
      statistics only include the non-x variables. If \code{FALSE}, the 
      exogenous `x' variables are modeled jointly with the other variables, and 
      the model-implied statistics reflect both sets of variables. If
      \code{"default"}, the value is set depending on the estimator, and
      whether or not the model involves categorical endogenous variables.}
    \item{\code{fixed.x}:}{If \code{TRUE}, the exogenous `x' covariates are 
      considered fixed variables and the means, variances and covariances of 
      these variables are fixed to their sample values. If \code{FALSE}, they 
      are considered random, and the means, variances and covariances are free 
      parameters. If \code{"default"}, the value is set depending on the mimic 
      option.}
    \item{\code{orthogonal}:}{If \code{TRUE}, all covariances among
      latent variables are set to zero.}
    \item{\code{orthogonal.y}:}{If \code{TRUE}, all covariances among
      endogenous latent variables only are set to zero.}
    \item{\code{orthogonal.x}:}{If \code{TRUE}, all covariances among
      exogenous latent variables only are set to zero.}
    \item{\code{std.lv}:}{If \code{TRUE}, the metric of each latent variable 
      is determined by fixing their (residual) variances to 1.0. If 
      \code{FALSE}, the metric of each latent variable is determined by fixing 
      the factor loading of the first indicator to 1.0. If there are multiple
      groups, \code{std.lv = TRUE} and \code{"loadings"} is included in 
      the \code{group.equal} argument, then only the latent variances
      of the first group will be fixed to 1.0, while the latent
      variances of other groups are set free.}
    \item{\code{effect.coding}:}{Can be logical or character string. If
     logical and \code{TRUE}, this implies 
     \code{effect.coding = c("loadings", "intercepts")}. If logical and 
     \code{FALSE}, it is set equal to the empty string. 
     If \code{"loadings"} is included, equality
     constraints are used so that the average of the factor loadings (per
     latent variable) equals 1. Note that this should not be used 
     together with \code{std.lv = TRUE}. If \code{"intercepts"} is
     included, equality constraints are used so that the sum of the
     intercepts (belonging to the indicators of a single latent variable) 
     equals zero.
     As a result, the latent mean is freely estimated and usually
     equals the average of the means of the involved indicators.}
    \item{\code{ceq.simple}:}{Logical. If \code{TRUE}, and no other 
     general (equality or inequality) constraints are used in the model, 
     simple equality constraints
     are represented in the parameter table as duplicated free parameters
     (instead of extra rows with \code{op = "=="}). The default is
     \code{FALSE}.}
    \item{\code{parameterization}:}{Currently only used if data is 
      categorical. If \code{"delta"}, the delta parameterization is used. 
      If \code{"theta"}, the theta parameterization is used.}
    \item{\code{correlation}:}{Only used for (single-level)
     continuous data. If \code{TRUE}, analyze a correlation matrix (instead
     of a (co)variance matrix). This implies that the residual observed
     variances are no longer free parameters. Instead, they are set to
     values that ensure the model-implied variances are unity. This also
     affects the standard errors. The only available estimators are GLS and
     WLS, which produce correct standard errors and a correct test statistic
     under normal and non-normal conditions respectively. Both
     \code{fixed.x = FALSE} and \code{fixed.x = TRUE} are supported (the
     latter affects the way the standard errors are computed). For now, this
     option always assumes \code{conditional.x = FALSE}.
     Alternatively, \code{correlation} may be a character vector of
     (observed) variable names (for example
     \code{correlation = c("x1", "x2", "x3")}). In that case only the
     listed variables are standardized to unit variance (their residual
     variances are fixed accordingly), while all remaining observed
     variables keep their original metric (a 'partial' correlation
     structure). If the model contains observed product (interaction)
     terms (for example \code{x1:y1}) and \code{correlation = TRUE},
     the product terms are automatically excluded from the correlation
     set: the product of standardized variables does not have unit
     variance, so their (co)variances remain free parameters. Note that
     product terms are typically highly non-normal, so the normal-theory
     (GLS) standard errors for their (co)variances may be too small;
     consider \code{se = "robust.sem"}.}
    \item{\code{composites.cov}:}{Character string. Only used if the model
     contains composites. Controls whether the composite-indicator
     (co)variances (the elements of the \code{T} matrix) are fixed to their
     sample values or estimated as free parameters. If \code{"fixed"}, the
     \code{T} matrix is fixed to the sample (co)variances. If \code{"free"},
     the \code{T} matrix is estimated as free parameters. If
     \code{"default"}, the value is resolved to \code{"free"} for multilevel
     models (which makes composites at every level identified) and
     \code{"fixed"} otherwise. For a single-level model, freeing \code{T}
     is fit-invariant (the degrees of freedom are unchanged).}
    \item{\code{auto.fix.first}:}{If \code{TRUE}, the factor loading of the
      first indicator is set to 1.0 for every latent variable.}
    \item{\code{bad.marker.crit}:}{Only used if
      \code{auto.fix.first = TRUE}. A single number (default 0.1). If larger
      than zero, lavaan checks -- once the unrestricted (h1) sample statistics
      are available -- whether the first indicator of each latent variable is
      a poor item, in the sense that it correlates weakly with the other
      indicators of the same factor (more precisely: the absolute value of its
      corrected item-total correlation is below \code{bad.marker.crit}). If so
      (and a clearly better indicator is available), a warning is issued and
      the loading of that better indicator is fixed to 1.0 instead, in order
      to set the metric of that latent variable. The main purpose is to avoid
      convergence problems caused by a (very) poor marker item. The check is
      based on the (pooled) unrestricted (h1) covariance matrix, so it behaves
      consistently with or without missing data, categorical data, multiple
      groups, etc. Larger values make the switching more aggressive; a value
      of 0 disables the check and always keeps the first indicator as the
      marker.}
    \item{\code{auto.fix.single}:}{If \code{TRUE}, the residual variance (if
      included) of an observed indicator is set to zero if it is the only 
      indicator of a latent variable.}
    \item{auto.var}{If \code{TRUE}, the (residual) variances of both observed
      and latent variables are set free.}
    \item{\code{auto.cov.lv.x}:}{If \code{TRUE}, the covariances of exogenous
      latent variables are included in the model and set free.}
    \item{\code{auto.cov.x}:}{If \code{TRUE}, the covariances between
      exogenous latent variables and observed exogenous covariates are also
      included in the model and set free (this implies
      \code{auto.cov.lv.x = TRUE}). Ignored if \code{conditional.x = TRUE},
      as these covariances cannot be represented when the model is
      conditioned on the exogenous covariates.}
    \item{\code{auto.cov.y}:}{If \code{TRUE}, the covariances of dependent 
      variables (both observed and latent) are included in the model and set 
      free.}
    \item{\code{auto.th}:}{If \code{TRUE}, thresholds for limited dependent 
      variables are included in the model and set free.}
    \item{\code{auto.delta}:}{If \code{TRUE}, response scaling parameters 
      for limited dependent variables are included in the model and set free.}
    \item{\code{auto.efa}:}{If \code{TRUE}, the necessary constraints are
      imposed to make the (unrotated) exploratory factor analysis blocks
      identifiable: for each block, factor variances are set to 1, factor
      covariances are constrained to be zero, and factor loadings are
      constrained to follow an echelon pattern.}
}

Data options:

\describe{
    \item{\code{std.ov}:}{If \code{TRUE}, observed variables are 
      standardized before entering the analysis. By default, these are
      only the non-exogenous observed variables, unless \code{fixed.x = FALSE}.
      Use this option with caution; it can be used to test whether (for
      example) nonconvergence was due to scaling issues. Note that this is
      still a covariance-based analysis: no constraints are imposed to
      ensure the model-implied (co)variance matrix has unit variances, and
      the standard errors still assume that the input was unstandardized. See
      also the \code{correlation} option.}
    \item{\code{missing}:}{The default setting is \code{"listwise"}: all
         cases with missing values
      are removed listwise from the data before the analysis starts. This is
      only valid if the data are missing completely at random (MCAR).
      Therefore, it may not be the optimal choice, but
      it can be useful for a first run. If the estimator belongs to
      the ML family, another option is \code{"ml"} (alias: \code{"fiml"}
      or \code{"direct"}). This corresponds to the so-called full information
      maximum likelihood approach (fiml), where we compute the likelihood
      case by case, using all available data from that case. Note
      that if the model contains exogenous observed covariates, and
      \code{fixed.x = TRUE} (the default), all cases with any missing values
      on these covariates will be deleted first. The option \code{"ml.x"}
      (alias: \code{"fiml.x"} or \code{"direct.x"}) is similar to \code{"ml"},
      but does not delete any cases with missing values for the exogenous
      covariates, even if \code{fixed.x = TRUE}. (Note: all lavaan versions
      < 0.6 used \code{"ml.x"} instead of \code{"ml"}).
      Since version 0.7-2, \code{"ml"} can also be combined with
      \code{conditional.x = TRUE} (single-level, continuous data): the
      likelihood of the observed non-x variables is then computed case by
      case, conditional on the (complete) exogenous covariates of that
      case. Cases with missing values on the exogenous covariates are
      deleted first, and \code{"ml.x"} is not available in this setting.
      Note that the robust (missing-data corrected) versions of RMSEA/CFI
      are not (yet) available when \code{conditional.x = TRUE}.
      If you wish to use multiple
      imputation, you need to use an external package (e.g., mice) to
      generate imputed datasets, which can then be analyzed using
      the \code{\link{semList}} function. The semTools package contains
      several functions to do this automatically. Another option (with
      continuous data) is to use \code{"two.stage"}
      or \code{"robust.two.stage"}. In this approach, we first estimate
      the sample statistics (mean vector, variance-covariance matrix) using
      an EM algorithm. Then, we use these estimated sample statistics as
      input for a regular analysis (as if the data were complete). The
      standard errors and test statistics
      are adjusted correctly to reflect the two-step procedure. The
      \code{"robust.two.stage"} option produces standard errors and
      a test statistic that are robust against non-normality.
      Both options are available for the ML estimator (since version
      0.7-2 also in combination with \code{conditional.x = TRUE}), and
      also for the
      (continuous-data) least-squares estimators \code{"ULS"}, \code{"GLS"},
      \code{"WLS"} and \code{"DLS"}; for the latter, lavaan automatically
      switches to \code{se = "robust.sem"} and \code{test = "satorra.bentler"},
      using the two-stage variance-covariance matrix of the EM-based sample
      statistics. In addition, for these least-squares estimators, requesting
      \code{missing = "ml"} (or \code{"fiml"}) is interpreted as a generic
      request to handle the missing values, and is silently treated as
      \code{missing = "two.stage"}.
      If (part of) the data is categorical, and the estimator is
      from the (W)LS family, the only option (besides listwise deletion)
      is \code{"pairwise"}. In this three-step approach, missingness is
      only an issue in the first two steps. In the first step, we compute
      thresholds (for categorical variables) and means or intercepts
      (for continuous variables) using univariate information only.
      In this step, we simply ignore
	  the missing values just like in mean(x, na.rm = TRUE). In the second
      step, we compute polychoric/polyserial/pearson correlations using (only)
      two variables at a time. Here we use pairwise deletion: we only keep
      those observations for which both values are observed (not missing),
      and this set may change from pair to pair.
      By default, in the categorical case we use \code{conditional.x = TRUE}. 
      Therefore, any cases
      with missing values on the exogenous covariates will be deleted listwise
      from the data first.
      Finally, if the estimator is \code{"PML"}, the available options are
      \code{"pairwise"}, \code{"available.cases"} and
      \code{"doubly.robust"}. See the PML tutorial on the lavaan website for
      more information about these approaches.}
    \item{\code{sampling.weights.normalization}:}{If \code{"none"}, the
      sampling weights (if provided) will not be transformed. If \code{"total"},
      the sampling weights are normalized by dividing by the total sum of
      the weights, and multiplying again by the total sample size. 
      If \code{"group"}, the sampling weights are normalized per group:
      by dividing by the sum of the weights (in each group), and multiplying
      again by the group size. The default is \code{"group"}. (For a single
      group, \code{"group"} and \code{"total"} are identical; they only differ
      when there are multiple groups, in which case \code{"group"} keeps each
      group's relative contribution proportional to its sample size, matching
      the behavior of Mplus.)}
    \item{\code{sampling.weights.type}:}{Only used when sampling weights are
      provided and the estimator relies on a weight matrix or sandwich variance:
      the least-squares family (GLS/WLS/DWLS/ULS/DLS, both continuous and
      categorical) and PML. Determines how the sampling weights enter the
      asymptotic variance (the Gamma / NACOV matrix, or the PML first-order
      information), and hence the (robust) standard errors and the scaled test
      statistic. For GLS and ULS only the standard errors are affected; for
      WLS/DWLS/DLS the Gamma also feeds the weight matrix, so the point
      estimates shift as well. If
      \code{"design"} (the default), the weights are treated as sampling
      (design) weights and a design-based sandwich is used (the meat is
      weighted by the sum of the squared weights), matching the behavior of
      Mplus. If \code{"frequency"}, the weights are treated as frequency
      (replication) counts (the meat is weighted by the sum of the weights),
      so that the results match a fit on the row-replicated data. Without
      sampling weights the two are identical.}
    \item{\code{samplestats}:}{Logical. If \code{FALSE}, no sample statistics
      will be computed (and no estimation can take place). This can be useful
      when only a dummy lavaan object is requested, without any computations.
      The default is \code{TRUE}.}
}

Data summary options:

\describe{
    \item{\code{sample.cov.rescale}:}{If \code{TRUE}, the sample covariance 
      matrix provided by the user is internally rescaled by multiplying it 
      with a factor (N-1)/N.  If \code{"default"}, the value is set depending 
      on the estimator and the likelihood option: it is set to \code{TRUE} if 
      maximum likelihood estimation is used and \code{likelihood="normal"}, 
      and \code{FALSE} otherwise.}
    \item{\code{ridge}:}{Logical. If \code{TRUE}, a small constant value is
      added to the diagonal elements of the covariance (or correlation)
      matrix before analysis. The value can be set using the
      \code{ridge.constant} option.}
    \item{\code{ridge.constant}:}{Numeric. Small constant used for ridging.
      The default value is 1e-05.}
}

Multiple group options:

\describe{ 
    \item{\code{group.label}:}{A character vector. The user can specify 
      which group (or factor) levels need to be selected from the grouping 
      variable, and in which order. If missing, all grouping levels are 
      selected, in the order as they appear in the data.}
    \item{\code{group.equal}:}{A vector of character strings. Only used in
      a multiple group analysis. Can be one or more of the following:
      \code{"loadings"}, \code{"composite.weights"}, 
      \code{"intercepts"}, \code{"means"},
      \code{"thresholds"}, \code{"regressions"}, \code{"residuals"},
      \code{"residual.covariances"}, \code{"lv.variances"} or
      \code{"lv.covariances"}, specifying the pattern of equality
      constraints across multiple groups. As a shortcut, the single
      value \code{"all"} can be used to constrain all (free) parameters
      to be equal across groups; it is equivalent to listing all of the
      values above. When the model is not the same across groups (for
      example when the \code{group:} modifier is used to specify a
      different model per group), only parameters that are present in two
      or more groups are constrained to be equal; a parameter that occurs
      in a single group only is always left free. The constraints are
      symmetric and do not depend on the first group: a parameter that is
      shared by (say) the second and third group is constrained to be
      equal across those two groups, even if it does not appear in the
      first group.}
    \item{\code{group.partial}:}{A vector of character strings containing 
      the labels of the parameters which should be free in all groups (thereby
      overriding the group.equal argument for some specific parameters).}
    \item{\code{group.w.free}:}{Logical. If \code{TRUE}, the group 
      frequencies are considered to be free parameters in the model. In this 
      case, a Poisson model is fitted to estimate the group frequencies. If
      \code{FALSE} (the default), the group frequencies are fixed to their
      observed values.}
}

Estimation options:

\describe{
    \item{\code{estimator}:}{The estimator to be used. Can be one of the 
      following: \code{"ML"} for maximum likelihood, \code{"GLS"} for 
      (normal theory) generalized least squares, 
      \code{"WLS"} for weighted least squares 
      (sometimes called ADF estimation), \code{"ULS"} for unweighted least 
      squares, \code{"DWLS"} for diagonally weighted least squares,
      and \code{"DLS"} for distributionally-weighted least squares. These
      are the main options that affect the estimation. For convenience, the
      \code{"ML"} option can be extended as \code{"MLM"}, \code{"MLMV"},
      \code{"MLMVS"}, \code{"MLF"}, and \code{"MLR"}.
      The estimation will still be plain \code{"ML"}, but now
      with robust standard errors and a robust (scaled) test statistic. For
      \code{"MLM"}, \code{"MLMV"}, \code{"MLMVS"}, classic robust standard
      errors are used (\code{se="robust.sem"}); for \code{"MLF"}, standard
      errors are based on first-order derivatives 
      (\code{information = "first.order"});
      for \code{"MLR"}, `Huber-White' robust standard errors are used
      (\code{se="robust.huber.white"}). In addition, \code{"MLM"} will compute
      a Satorra-Bentler scaled (mean adjusted) test statistic
      (\code{test="satorra.bentler"}), \code{"MLMVS"} will compute a
      mean and variance adjusted test statistic (Satterthwaite style)
      (\code{test="mean.var.adjusted"}), \code{"MLMV"} will compute a mean
      and variance adjusted test statistic (scaled and shifted)
      (\code{test="scaled.shifted"}), and \code{"MLR"} will
      compute a test statistic which is asymptotically
      equivalent to the Yuan-Bentler T2-star test statistic
      (\code{test="yuan.bentler.mplus"}). Analogously,
      the estimators \code{"WLSM"} and \code{"WLSMV"} imply the \code{"DWLS"}
      estimator (not the \code{"WLS"} estimator) with robust standard errors
      and a mean or mean and variance adjusted test statistic. Estimators
      \code{"ULSM"} and \code{"ULSMV"} imply the \code{"ULS"}
      estimator with robust standard errors
      and a mean or mean and variance adjusted test statistic.
      Finally, \code{"RBM"} requests reduced-bias M-estimation
      (penalized maximum likelihood; continuous-outcome models for now). By
      default sandwich standard errors are used
      (\code{se = "robust.huber.white"}). The type of bias reduction is
      controlled by the \code{rbm.method} element of \code{estimator.args}
      (either \code{"implicit"} (default) or \code{"explicit"}), for example
      \code{estimator = list(estimator = "RBM", rbm.method = "explicit")}.
      The \code{"explicit"} method is much faster for models with many
      parameters (it requires a single ML fit plus a one-step correction),
      and is recommended in that case.}
    \item{\code{likelihood}:}{Only relevant for ML estimation. If 
      \code{"wishart"}, the Wishart likelihood approach is used. In this 
      approach, the covariance matrix has been divided by N-1, and both 
      standard errors and test statistics are based on N-1.
      If \code{"normal"}, the normal likelihood approach is used. Here,
      the covariance matrix has been divided by N, and both standard errors
      and test statistics are based on N. If \code{"default"}, it depends
      on the mimic option: if \code{mimic="lavaan"} or \code{mimic="Mplus"},
      normal likelihood is used; otherwise, Wishart likelihood is used.}
    \item{\code{link}:}{Not used yet. This is just a placeholder until
      the MML estimator is back.}
    \item{\code{information}:}{If \code{"expected"}, the expected 
      information matrix is used (to compute the standard errors). If 
      \code{"observed"}, the observed information matrix is used. 
      If \code{"first.order"}, the information matrix is based on the
      outer product of the casewise scores. See also the options
      \code{"h1.information"} and \code{"observed.information"} for
      further control. If \code{"default"}, the value is set depending 
      on the estimator, the missing argument, and the mimic option. If
      the argument is a vector with two elements, the first element
      is used for the computation of the standard errors, while the
      second element is used for the (robust) test statistic.}
   \item{\code{h1.missing.method}:}{Character. Only used for multilevel
      models with missing data (and \code{missing = "ml"}). If \code{"em"}
      (the default since 0.7-1), the means and (co)variances of the
      unrestricted (h1) model are estimated using the EM algorithm,
      integrating over both the cluster random effects and the missing
      values (this mimics Mplus). If \code{"fiml"}, the unrestricted model
      is estimated using quasi-Newton optimization (nlminb) with a cholesky
      parameterization of the covariance matrices (this was the default
      in versions < 0.7-1). See the \code{em.h1.args} option to control
      the EM algorithm.}
   \item{\code{em.h1.args}:}{List. Options related to the EM algorithm
      that is used to estimate the means and (co)variances of the
      unrestricted (h1) model. This is only relevant for multilevel models,
      and for single-level models with missing data (and
      \code{missing = "ml"} or \code{missing = "two.stage"}). The default
      values are \code{em.h1.args = list(max_iter = "default",
      tol = "default", min_variance = 1e-05, warn = TRUE,
      non_pd_action = "stop", non_pd_tol = 1e-05,
      acceleration = "squarem", fused = TRUE)}. The \code{max_iter}
      element sets the
      maximum number of EM iterations; the \code{tol} element sets the
      tolerance used to determine convergence. If \code{"default"}, they
      are set to 500 and 1e-05 respectively in the single-level setting
      (where the EM algorithm stops when the largest absolute change in
      the parameter values is smaller than \code{tol}), and to 5000 and
      1e-04 respectively in the multilevel setting (where the EM algorithm
      stops when the change in the loglikelihood is smaller than
      \code{tol}). The \code{min_variance} element sets the minimum value
      a variance can take during the EM iterations (multilevel only):
      whenever a within-level variance drops below this value, it is fixed
      to this value (and the corresponding covariances are set to zero).
      If the \code{warn} element is \code{TRUE} (the default),
      a warning is produced if the EM algorithm reaches \code{max_iter}
      iterations without converging. The \code{non_pd_action} element
      controls what happens if an EM estimated variance-covariance matrix
      is (near) singular: if any variance is (near) zero, or if the
      smallest eigenvalue of the matrix is smaller than the
      \code{non_pd_tol} element. This typically happens when a (within or
      between) variance of an observed variable approaches zero, but the
      full matrix may also be (near) singular. If \code{non_pd_action =
      "stop"} (the default), estimation stops with an error message; if
      \code{"warn"}, a warning is produced, and estimation continues; if
      \code{"none"}, estimation continues silently. The
      \code{acceleration} element controls whether the EM iterations are
      accelerated. If \code{"squarem"} (the default), the SQUAREM scheme
      (Varadhan & Roland, 2008) is used: every cycle takes two EM steps,
      extrapolates, and takes one stabilizing EM step from the
      extrapolated point, with a loglikelihood safeguard; this typically
      reduces the number of EM steps considerably (in particular when
      some variances approach zero), and converges to the same solution.
      If \code{"qn"}, a quasi-Newton scheme (Zhou, Alexander & Lange,
      2011; the \code{"qn"} method of the \pkg{turboEM} package) is used
      instead: a low-rank secant approximation of the EM map is used to
      take a quasi-Newton step towards the fixed point, again with a
      loglikelihood safeguard. If \code{"none"}, the plain (unaccelerated)
      EM iterations are used. If the \code{fused} element is \code{TRUE}
      (the default), the
      loglikelihood is computed as a cheap byproduct of the E-step
      quantities (only used in the two-level setting with missing data,
      for the plain EM iterations); if \code{FALSE}, a separate
      (stand-alone) loglikelihood evaluation is used in every iteration;
      both give identical results.}
   \item{\code{h1.information}:}{If \code{"structured"} (the default), the
      unrestricted (h1) information part of the (expected, first.order or 
      observed if h1 is used) information matrix is based on the structured, 
      or model-implied statistics (model-implied covariance matrix, 
      model-implied mean vector, etc.). 
      If \code{"unstructured"}, the unrestricted (h1) information part is
      based on sample-based statistics (observed covariance matrix, observed
      mean vector, etc.) If
      the argument is a vector with two elements, the first element
      is used for the computation of the standard errors, while the
      second element is used for the (robust) test statistic.}
   \item{\code{observed.information}:}{If \code{"hessian"}, the observed
      information matrix is based on the hessian of the objective function.
      If \code{"h1"}, an approximation is used that is based on 
      the observed information matrix of the unrestricted (h1) model. If
      the argument is a vector with two elements, the first element
      is used for the computation of the standard errors, while the
      second element is used for the (robust) test statistic.}
   \item{\code{se}:}{If \code{"standard"}, conventional standard errors
      are computed based on inverting the (expected, observed or first.order) 
      information matrix. If \code{"robust.sem"}, conventional robust
      standard errors are computed.  If \code{"robust.huber.white"},
      standard errors are computed based on the 'mlr' (aka pseudo ML,
      Huber-White) approach.
      If \code{"robust"}, either \code{"robust.sem"} or
      \code{"robust.huber.white"} is used depending on the estimator,
      the mimic option, and whether the data are complete or not.
      If \code{"boot"} or \code{"bootstrap"}, bootstrap standard errors are
      computed using standard bootstrapping (unless Bollen-Stine bootstrapping
      is requested for the test statistic; in this case bootstrap standard
      errors are computed using model-based bootstrapping).
      If \code{"none"}, no standard errors are computed.}
   \item{\code{test}:}{Character vector. See the documentation of
      the \code{\link{lavTest}} function for a full list. Multiple names
      of test statistics can be provided. If \code{"default"}, the value 
      depends on the values of other arguments. See also the 
      \code{\link{lavTest}} function to extract (alternative) 
      test statistics from a fitted lavaan object. FMG test names such as
      \code{"peba4"}, \code{"pols3"}, \code{"pall"}, and \code{"all"} can be
      requested here for supported ML-family models. The convenience value
      \code{"fmg"} resolves to \code{"peba4_rls"} for one-model tests.}
   \item{\code{standard.test}:}{Character. Choose the test statistic
    that will be used to compute fit measures (like CFI or RMSEA).
    The default is \code{"standard"}, but it could also be (for example)
    \code{"Browne.residual.nt"}, or an FMG test name such as
    \code{"peba4"} or \code{"fmg"}.}
   \item{\code{scaled.test}:}{Character. Choose the test statistic
    that will be scaled (if a scaled test statistic is requested).
    The default is \code{"standard"}, but it could also be (for example)
    \code{"Browne.residual.nt"}.}
   \item{\code{gamma.n.minus.one}}{Logical. If \code{TRUE}, we divide the
     Gamma matrix by N-1 (instead of the default N).}
   \item{\code{gamma.unbiased}}{Logical. If \code{TRUE}, we compute an
     unbiased version for the Gamma matrix. Only available for single-level
     complete data and when \code{conditional.x = FALSE} and
     \code{fixed.x = FALSE} (for now). Suffixless FMG test names use this
     option to choose between biased and unbiased Gamma.}
   \item{\code{se.delta.second.order}}{Logical. If \code{FALSE} (the
     default), the standard errors of defined parameters (introduced with
     the \code{:=} operator) and of standardized parameters (as returned
     by \code{standardizedSolution()} or \code{lavInspect(., "vcov.std.*")})
     are computed using the first-order delta method. If \code{TRUE}, the
     second-order delta method is used, which adds the term
     \eqn{\frac{1}{2}\textrm{tr}(H V H V)} (where \eqn{H} is the Hessian
     of the transformed parameter and \eqn{V} is the parameter covariance
     matrix) to the first-order variance approximation. This may improve
     accuracy when the transformation is a strongly non-linear function
     of the model parameters. Has no effect when standard errors are
     obtained by bootstrapping or by the Monte Carlo method.}
   \item{\code{se.def}}{Character. Method for computing standard errors
     (and confidence intervals) of defined parameters (introduced with
     the \code{:=} operator). The default is \code{"default"} (or
     equivalently \code{"delta"}), which uses the delta method (respecting
     \code{se.delta.second.order}). Set to \code{"monte.carlo"} to use
     the Monte Carlo method of Preacher and Selig (2012). Under this
     method, \code{R} draws are taken from
     \eqn{\textit{MVN}(\hat{\theta}, \widehat{\textit{Var}}(\hat{\theta}))},
     the defined parameter is evaluated for each draw, and standard
     errors are obtained as the standard deviation of these realizations.
     Confidence intervals (in \code{parameterEstimates()} and
     \code{standardizedSolution()}) are based on percentile bounds of
     the Monte Carlo distribution, and therefore can be asymmetric (just
     like bootstrap intervals). Has no effect when
     \code{se = "bootstrap"}.}
   \item{\code{monte.carlo}}{List of settings for the Monte Carlo method
     (used when \code{se.def = "monte.carlo"}). Currently recognizes:
     \code{R} (integer; number of Monte Carlo draws; default 20000) and
     \code{seed} (integer or \code{NULL}; optional seed for
     reproducibility).}
   \item{\code{bootstrap}:}{Number of bootstrap draws, if bootstrapping is 
     used.}
   \item{\code{do.fit}:}{If \code{FALSE}, the model is not fit, and the 
     current starting values of the model parameters are preserved.}
}

Optimization options:

\describe{
    \item{\code{integration.ngh}:}{Integer. The number of Gauss-Hermite
      quadrature points (per dimension) used for numerical integration.
      The default is 21. Numerical integration is used when
      \code{estimator = "MML"}, and for two-level models with random
      slopes for \emph{latent} covariates (the \code{rv()} modifier;
      only the latent-covariate slopes are integrated numerically, all
      other random effects are handled in closed form).}
    \item{\code{control}:}{A list containing control parameters passed to
      the external optimizer. By default, lavaan uses \code{"nlminb"}. 
      See the manpage of \code{\link{nlminb}} for an overview of the control 
      parameters. If another (external) optimizer is selected, see the
      manpage for that optimizer to see the possible control parameters.}
    \item{\code{optim.method}:}{Character. The optimizer that should be
      used. For unconstrained optimization or models with only linear
      equality constraints (i.e., the model syntax
      does not include any "==", ">" or "<" operators),
      the available options are \code{"nlminb"} (the default), \code{"BFGS"},
      \code{"L-BFGS-B"}. These are all quasi-newton methods. A
      Gauss-Newton (Fisher scoring) method with Levenberg-Marquardt
      damping is also available (\code{optim.method = "GN"}); it can be
      used with the \code{"ML"}, \code{"GLS"}, \code{"WLS"},
      \code{"DWLS"}, \code{"ULS"}, \code{"DLS"}, \code{"NTRLS"} and
      \code{"catML"} estimators, and is the default when
      \code{estimator = "DLS"}. For multilevel models, an EM algorithm
      is also available (\code{optim.method = "em"}); this is the
      default for two-level models with \code{missing = "ml"} (as the
      EM algorithm is more robust in that setting; this also mimics
      Mplus); see the \code{em.args} option to control the EM
      algorithm. For two-level models with random slopes (the
      \code{rv()} modifier), \code{optim.method = "em"} uses a
      dedicated EM algorithm. Note that for any
      \code{optim.method = "em"} fit, variances that are fixed to zero
      are replaced by the (small) value \code{em.args$zerovar_offset}
      (as in Mplus, which uses the same 'minimum variance' device); as
      a result, the reported loglikelihood may differ slightly from an
      \code{optim.method = "nlminb"} fit of the same model (which uses
      the exact zero values).
      For constrained
      optimization, the only available option is \code{"nlminb_constr"},
      which uses an augmented Lagrangian minimization algorithm.}
    \item{\code{optim.force.converged}:}{Logical. If \code{TRUE}, pretend
      the model has converged, no matter what.}
    \item{\code{optim.fix.saturated}:}{Logical. Only used for multilevel
      models. If \code{TRUE} (the default), and one or more blocks (levels)
      of the model are saturated (i.e., all variances/covariances and
      means/intercepts of the observed variables of that block are free
      and unconstrained), the free parameters of these blocks are
      (temporarily) fixed to their h1 (unrestricted model) estimates during
      optimization, as their estimated values are already known. This often
      helps convergence. After the optimization, they are treated as free
      parameters again (e.g., when standard errors are computed).}
    \item{\code{optim.dx.tol}}{Numeric. Tolerance used for checking if
      the elements of the (unscaled) gradient are all zero (in absolute 
      value). The default value is 0.001.}
    \item{\code{gn.args}:}{List. Options related to the Gauss-Newton
      optimizer (\code{optim.method = "GN"}). The default values are
      \code{gn.args = list(max_iter = 200, tol_x = 1e-05,
      tol_g = 1e-06)}. The \code{max_iter} element sets the maximum
      number of Gauss-Newton iterations. Optimization stops when either
      the (scaled) gradient of the discrepancy function is smaller than
      \code{tol_g}, or the root mean square of the difference between
      the old and new parameter values is smaller than \code{tol_x}.}
    \item{\code{bounds}:}{Only used if \code{optim.method = "nlminb"}.
      If logical: \code{FALSE} implies no bounds are imposed on the parameters.
      If \code{TRUE}, this implies \code{bounds = "wide"}. If character,
      possible options are \code{"none"} (the default), \code{"standard"},
      \code{"wide"}, \code{"pos.var"}, \code{"pos.ov.var"}, and 
      \code{"pos.lv.var"}. 
      If \code{bounds = "pos.ov.var"}, the observed variances are forced to be
      nonnegative. If \code{bounds = "pos.lv.var"}, the latent variances are
      forced to be nonnegative. If \code{bounds = "pos.var"}, both observed
      and latent variances are forced to be nonnegative. If 
      \code{bounds = "standard"}, lower and upper bounds are computed for
      observed and latent variances, and factor loadings. If
      \code{bounds = "wide"}, lower and upper bounds are computed for
      observed and latent variances, and factor loadings; but the range of
      the bounds is enlarged (allowing again for slightly negative variances).}
    \item{\code{optim.bounds}:}{List. This can be used instead of the 
      \code{bounds} argument to allow more control. Possible elements of the
      list are \code{lower}, \code{upper}, \code{lower.factor} and 
      \code{upper.factor}. All of these accept a vector. The \code{lower} and
      \code{upper} elements indicate for which type of parameters bounds
      should be computed. Possible choices are \code{"ov.var"}, \code{"lv.var"},
      \code{"loadings"} and \code{"covariances"}. The \code{lower.factor} and 
      \code{upper.factor} elements should have the same length as the 
      \code{lower} and \code{upper} elements respectively. They indicate the
      factor by which the range of the bounds should be enlarged (for 
      example, 1.1 or 1.2; the default is 1.0). Other elements are
      \code{min.reliability.marker} which sets the lower bound for the
      reliability of the marker indicator (if any) of each factor 
      (default is 0.1). Finally, the \code{min.var.lv.endo} element indicates
      the lower bound of the variance of any endogenous latent variable
      (default is 0.0).}
    \item{\code{em.args}:}{List. Options related to the EM algorithm that
      is used to estimate the model parameters when
      \code{optim.method = "em"} (currently only available for multilevel
      models, with complete or missing data). The default values are
      \code{em.args = list(max_iter =
      10000, fx_tol = 1e-08, dx_tol = 1e-04, zerovar_offset = 1e-04,
      acceleration = "squarem", fused = TRUE)}.
      The \code{max_iter} element sets the maximum number of EM
      iterations. The EM algorithm stops when either the change in the
      loglikelihood is smaller than \code{fx_tol}, or the largest absolute
      change in the parameter values is smaller than \code{dx_tol}.
      The \code{zerovar_offset} element is used as the starting value for
      variances that are fixed to zero, as the EM algorithm cannot handle
      exact zero variances. The \code{acceleration} element controls
      whether the EM iterations are accelerated by the SQUAREM scheme,
      and the \code{fused} element controls whether the loglikelihood is
      computed as a byproduct of the E-step (see the \code{em.h1.args}
      option for details on both). If the \code{nlminb_handoff} element
      is \code{TRUE} (the default), and the EM iterations end with a
      non-negligible gradient (larger than \code{dx_tol}; this typically
      happens when the loglikelihood stalls before the gradient has
      converged), the solution is refined by a quasi-Newton (nlminb)
      run, warm-started at the EM values; the refined solution is only
      accepted if nlminb converges to an (at least) equally good
      solution; otherwise, the EM solution is retained. If, in addition,
      the \code{early_handoff} element is \code{TRUE} (the default), the
      EM phase stops early (using a loosened tolerance,
      \code{max(fx_tol, 0.001)}), as it only needs to provide a reliable
      warm start for nlminb; this is usually considerably faster, and
      gives the same solution; if nlminb fails from the early start, a
      safety net resumes the EM iterations with the strict \code{fx_tol}
      (and tries the hand-off once more), reproducing the non-early
      behavior. If \code{verbose = TRUE}, only the EM (or SQUAREM) cycles
      are shown; set the \code{verbose} element to \code{TRUE} to also show
      the optimizer output of each (inner) M-step.}
}

Parallelization options (currently only used for bootstrapping):

\describe{
    \item{parallel}{The type of parallel operation to be used (if any).  If
      missing, the default is \code{"no"}.  }
    \item{ncpus}{Integer: number of processes to be used in parallel operation:
      typically one would set this to the number of available CPUs. By
      default this is 2 (following CRAN policy); you may set this to a higher
      value (for example, the number of cores as detected by
      \code{parallel::detectCores()} minus one) if you wish to use more
      processes.}
    \item{cl}{An optional \pkg{parallel} or \pkg{snow} cluster for use if
      \code{parallel = "snow"}. If not supplied, a cluster on the local 
      machine is created for the duration of the \code{lavBootstrap} 
      call.}
    \item{iseed}{An integer to set the seed. Or NULL if no reproducible 
      results are needed. This works for both serial (non-parallel) and 
      parallel settings. Internally, \code{RNGkind()} is set to 
      \code{"L'Ecuyer-CMRG"} if \code{parallel = "multicore"}. If 
      \code{parallel = "snow"} (under windows), 
      \code{parallel::clusterSetRNGStream()} is called which automatically 
      switches to \code{"L'Ecuyer-CMRG"}. When \code{iseed} is not
      NULL, \code{.Random.seed} (if it exists) in the global environment is
      left untouched.}
}

Categorical estimation options:

\describe{
    \item{\code{zero.add}:}{A numeric vector containing two values. These 
      values affect the calculation of polychoric correlations when some 
      frequencies in the bivariate table are zero. The first value only 
      applies for 2x2 tables. The second value for larger tables. This value 
      is added to the zero frequency in the bivariate table. If 
      \code{"default"}, the value is set depending on the \code{"mimic"}
      option. By default, lavaan uses \code{zero.add = c(0.5, 0.0)}.}
    \item{\code{zero.keep.margins}:}{Logical. This argument only affects 
      the computation of polychoric correlations for 2x2 tables with an empty 
      cell, and where a value is added to the empty cell. If \code{TRUE}, the 
      other values of the frequency table are adjusted so that all margins are 
      unaffected. If \code{"default"}, the value is set depending on the
      \code{"mimic"} option. The default is \code{TRUE}.}
    \item{\code{zero.cell.warn}:}{Logical. Only used if some observed 
      endogenous variables are categorical. If \code{TRUE}, give a warning if 
      one or more cells of a bivariate frequency table are empty.}
    \item{\code{allow.empty.cell}:}{Logical. If \code{TRUE}, ignore
      situations where an ordinal variable has fewer categories than
      expected, or where a category is empty in a specific group. This
    argument is currently used by the blavaan package (which
    uses lavaan to set up the model). The argument is not expected to
    salvage a lavaan model that results in errors.}
}

Starting values options:

\describe{
    \item{\code{start}:}{If it is a character string, the two options are 
      currently \code{"simple"} and \code{"default"}. In the first case, all 
      parameter values are set to zero, except the factor loadings and
      (residual) variances, which are set to 0.7 and 1.0 respectively.
      When \code{start} is \code{"default"}, the factor loadings are
      estimated using the fabin3 estimator (tsls) per factor. The
      residual variances of observed variables are set to half the
      observed variance, and all other (residual) variances are set to 0.05.
      The remaining parameters (regression coefficients, covariances) are
      set to zero.
      If \code{start} is a numerical vector, it should contain the starting
      values for all the free parameters (in the same order as the parameter
      table).
      If \code{start} is a fitted object of class \code{\linkS4class{lavaan}},
      the estimated values of the corresponding parameters will be extracted.
      If it is a parameter table, for example the output of the
      \code{parameterEstimates()} function, the values of the \code{est} or
      \code{start} or \code{ustart} column (whichever is found first) will be
      extracted.}
    \item{\code{rstarts}:}{Integer. The number of refits that lavaan should
      try with random starting values. Random starting values are computed
      by drawing random numbers from a uniform distribution. Correlations
      are drawn from the interval [-0.5, +0.5] and then converted to
      covariances. Lower and upper bounds for (residual) variances are
      computed just like the standard bounds in bounded estimation.
      Random starting values are not computed for regression coefficients
      (which are always zero) and factor loadings of higher-order constructs
      (which are always unity). From all the runs that converged, the final
      solution is the one that resulted in the smallest value for the
      discrepancy function.}
}

Check options:

\describe{
    \item{\code{check.start}:}{Logical. If \code{TRUE},
      the starting values are checked for possibly 
      inconsistent values (for example values implying correlations larger 
      than one). If needed, a warning is given.}
    \item{\code{check.sigma.pd}:}{Character. If \code{"chol"}, a Cholesky
      decomposition is used to check if the model implied covariance matrix
      (Sigma) is positive definite. If \code{"eigen"}, an eigen decomposition 
      is used to check if Sigma is positive definite. The latter was the
      default until 0.6-20. From 0.6-21 onwards, \code{"chol"} became the
      default (as it is much faster).}
    \item{\code{check.gradient}:}{Logical. If \code{TRUE}, and the model
      converged, a warning is given if the optimizer reported a (local)
      solution while not all elements of the (unscaled) gradient (as
      seen by the optimizer) are (near) zero, as they should be (the
      tolerance used is 0.001).}
    \item{\code{check.post}:}{Logical. If \code{TRUE}, and the model
      converged, a check is performed after (post) fitting, to verify if 
      the solution is admissible. This implies that all variances are
      non-negative, and all the model-implied covariance matrices are 
      positive (semi-)definite. For the latter test, we tolerate a tiny 
      negative eigenvalue that is smaller than .Machine$double.eps^(3/4), 
      treating it as being zero.}
    \item{\code{check.vcov}:}{Logical. If \code{TRUE}, and the model converged,
      we check if the variance-covariance matrix of the free parameters
      is positive definite. We take into account possible equality and
      active inequality constraints. If needed, a warning is given.}
    \item{\code{check.lv.names}:}{Logical. If \code{TRUE}, and latent variables
      are defined in the model, lavaan will stop with an error message if
      a latent variable name also occurs in the data (implying it is also
      an observed variable).}
}

Verbosity options:

\describe{
    \item{\code{verbose}:}{If \code{TRUE}, show what lavaan is doing. During
      estimation, the function value is printed out 
      during each iteration.}
    \item{\code{warn}:}{If \code{FALSE}, suppress all lavaan-specific 
      warning messages.}
    \item{\code{debug}:}{If \code{TRUE}, debugging information is printed 
      out.}
}

Miscellaneous:

\describe{
    \item{\code{model.type}:}{Set the model type: possible values
      are \code{"cfa"}, \code{"sem"} or \code{"growth"}. This may affect
      how starting values are computed, and may be used to alter the terminology
      used in the summary output, or the layout of path diagrams that are
      based on a fitted lavaan object.}
    \item{\code{mimic}:}{If \code{"Mplus"}, an attempt is made to mimic the
      Mplus program. If \code{"EQS"}, an attempt is made to mimic the EQS
      program. If \code{"default"}, the value is (currently) set to
      \code{"lavaan"}, which is very close to \code{"Mplus"}. See
      \code{\link{mimic}} for the list of options that are affected by each
      setting.}
    \item{\code{representation}:}{If \code{"LISREL"} the classical LISREL 
      matrix representation is used to represent the model (using the all-y 
      variant). No other options are available (for now).}
    \item{\code{implied}:}{Logical. If \code{TRUE}, compute the model-implied
      statistics, and store them in the implied slot.}
    \item{\code{h1}:}{Logical. If \code{TRUE}, compute the unrestricted model
      and store the unrestricted summary statistics (and perhaps a 
      loglikelihood) in the h1 slot.}
    \item{\code{baseline:}}{Logical. If \code{TRUE}, compute a baseline model
      (currently always the independence model, assuming all variables
      are uncorrelated) and store the results in the baseline slot.}
    \item{\code{baseline.conditional.x.free.slopes}:}{Logical. If \code{TRUE},
      and \code{conditional.x = TRUE}, the (default) baseline model will
      allow the slope structure to be unrestricted.}
    \item{\code{fit.by.level}:}{Logical. Only used for multilevel models.
      If \code{TRUE}, fit (and store) the partially saturated models that
      are needed to compute level-specific fit measures: for each level,
      an auxiliary model is fitted where the model at that level is
      retained, while the model at the other level is saturated (in
      addition, level-specific baseline models are fitted as well). See
      the \code{level} argument of \code{\link{fitMeasures}}. Set to
      \code{FALSE} to skip these additional model fits.}
    \item{\code{store.vcov}}{Logical. If \code{TRUE}, and \code{se=} is not
      set to \code{"none"}, store the full variance-covariance matrix of
      the model parameters in the vcov slot of the fitted lavaan object.}
    \item{\code{parser}}{Character. If \code{"new"} (the default), the new
      parser is used to parse the model syntax. If \code{"old"}, the original
      (pre 0.6-18) parser is used.}
}

}

\seealso{
\code{\link{lavaan}}, \code{\link{mimic}}, \code{\link{estimator_iv}}
}

\examples{
lavOptions()
lavOptions("std.lv")
lavOptions(c("std.lv", "orthogonal"))
}

\name{mimic}
\alias{mimic}
\title{The mimic= option}
\description{
The \code{mimic=} argument of the \code{\link{lavaan}} function (and its
wrappers \code{\link{cfa}}, \code{\link{sem}} and \code{\link{growth}})
controls a collection of primarily technical default settings, with the aim of
making lavaan's output as similar as possible to that of other SEM software.
This applies in particular to a few key results, such as the value of the
Satorra-Bentler scaled test statistic.

The \code{mimic=} option was originally introduced to demonstrate that
differences between lavaan and other SEM programs were often due to technical
implementation details rather than errors in lavaan. Nevertheless, the scope of
the \code{mimic=} option is intentionally limited. It is designed to reproduce
a small number of important results and does not attempt to replicate every
quantity reported by alternative SEM software packages.

This help page documents which options are affected by the
\code{mimic=} argument.
}
\details{
The \code{mimic=} argument does not change the model that is fitted, nor the
estimator that is used. Instead, it overwrites the default values of a number
of (mostly technical) options that are otherwise estimator-dependent. The
user can always override these defaults by setting the corresponding option
explicitly in the \code{\link{lavaan}} call.

The argument is case-insensitive and accepts a number of aliases. After
normalization, only four values remain:
\describe{
  \item{\code{"lavaan"}:}{the native lavaan defaults (also selected by the
    value \code{"default"}); this is the default.}
  \item{\code{"Mplus"}:}{mimic the \code{Mplus} program (also selected by
    the value \code{"mplus"}).}
  \item{\code{"EQS"}:}{mimic the \code{EQS} program (also selected by the
    values \code{"eqs"}, \code{"LISREL"} and \code{"lisrel"}).}
  \item{\code{"lm"}:}{a minimal setting used internally for plain regression
    (also selected by the value \code{"regression"}); not intended for
    general use.}
}

\strong{mimic = "lavaan" (the default)}

This is the reference setting. For maximum likelihood (ML) estimation, the
following defaults are used:
\describe{
  \item{\code{likelihood = "normal"}:}{the sample covariance matrix is
    rescaled by a factor \code{(N-1)/N}, so the loglikelihood and standard
    errors are based on a division by \code{N} (rather than \code{N-1}).}
  \item{\code{conditional.x = FALSE}:}{the exogenous \code{x} covariates are
    regressed out first, and a conditional (on x) model is fitted.}
  \item{\code{fixed.x = TRUE}:}{the means, variances and covariances of the
    exogenous \code{x} covariates are fixed to their sample values (for the
    ML, MML and IV estimator groups, unless \code{start = "simple"}).}
  \item{\code{zero.keep.margins = TRUE}:}{for categorical data, the (one-way)
    margins are preserved when adding a small constant to empty cells.}
}
In addition, the various Mplus-specific technical options
(\code{information.expected.mplus}, \code{gamma.vcov.mplus},
\code{gamma.wls.mplus}, \code{gls.v11.mplus} and
\code{rmsea.scaled.mplus}) are all set to \code{FALSE}.

\strong{mimic = "Mplus"}

Starting from the lavaan defaults, the following options are changed in an
attempt to reproduce the output of \code{Mplus}:
\describe{
  \item{\code{information.expected.mplus = TRUE}:}{use the Mplus variant of
    the expected information matrix. This affects the standard errors (when
    \code{information = "expected"}) and the scaled (Satorra-Bentler and
    Yuan-Bentler) test statistics.}
  \item{\code{gamma.vcov.mplus = TRUE}:}{use the Mplus way of computing the
    \eqn{\Gamma} matrix (the asymptotic covariance matrix of the sample
    statistics) that enters the robust (sandwich) standard errors.}
  \item{\code{gamma.wls.mplus = TRUE}:}{use the Mplus way of computing the
    \eqn{\Gamma} matrix used as the weight matrix in (D)WLS estimation.}
  \item{\code{gls.v11.mplus = TRUE}:}{use the Mplus variant of the weight
    matrix used in GLS estimation (when there is a meanstructure and
    \code{conditional.x = FALSE}).}
  \item{\code{rmsea.scaled.mplus = TRUE}:}{use the Mplus reference
    distribution for the \emph{scaled} RMSEA confidence interval and
    p-values. lavaan refers the \emph{unscaled} test statistic to a
    chi-square distribution with \eqn{tr[U\Gamma]} degrees of freedom, while
    Mplus refers the \emph{scaled} statistic to a chi-square distribution
    with the degrees of freedom of the scaled test. Both match the mean of
    the null distribution, but on a different scale, so they yield different
    intervals and p-values. This only affects the Satorra-Bentler and
    Yuan-Bentler family (\code{MLM}, \code{MLR}, \code{MLMVS},
    \code{WLSM}, ...); for the shifted family (\code{MLMV}, \code{WLSMV},
    \code{ULSMV}) lavaan already uses the Mplus convention.}
  \item{\code{missing = "ml"}:}{for continuous data and the ML or MLR
    estimator, the default missing-data handling becomes full information
    maximum likelihood (FIML), rather than listwise deletion.}
  \item{\code{meanstructure = TRUE}:}{a meanstructure is added by default
    (unless the estimator is PML).}
  \item{\code{group.equal}:}{in a multiple-group analysis, if no equality
    constraints are requested, a default set is imposed: \code{"loadings"}
    and \code{"thresholds"} for categorical data;
    \code{"loadings"} for continuous data without a meanstructure; and
    \code{"loadings"} and \code{"intercepts"} for continuous data with a
    meanstructure.}
  \item{\code{baseline.conditional.x.free.slopes = FALSE}:}{when
    \code{conditional.x = TRUE}, the baseline (independence) model does not
    free the slope structure.}
}
The \code{likelihood}, \code{conditional.x}, \code{fixed.x} and
\code{zero.keep.margins} defaults are the same as for
\code{mimic = "lavaan"}.

\strong{mimic = "EQS"}

Starting from the lavaan defaults, the following options are changed in an
attempt to reproduce the output of \code{EQS} (the value \code{"LISREL"} is
treated as an alias for \code{"EQS"}):
\describe{
  \item{\code{likelihood = "wishart"}:}{for ML estimation, the Wishart
    likelihood is used. The sample covariance matrix is divided by
    \code{N-1} (instead of \code{N}), which affects the chi-square test
    statistic, the loglikelihood and the standard errors.}
  \item{\code{baseline.fixed.x.free.cov = FALSE}:}{when \code{fixed.x = TRUE},
    the baseline (independence) model does not free the (co)variances of the
    exogenous covariates.}
}
In addition, for the MLR estimator, the robust test statistic defaults to
\code{"yuan.bentler"} (instead of the \code{"yuan.bentler.mplus"} variant
that is used for \code{mimic = "lavaan"} and \code{mimic = "Mplus"}).

Note that the \code{mimic=} argument only sets \emph{defaults}: any option
that is set explicitly in the \code{\link{lavaan}} call takes precedence. The
mimicking is also necessarily incomplete; small numerical differences with
the target program may remain.
}
\seealso{
\code{\link{lavOptions}}, \code{\link{lavaan}}
}
\examples{
# default (mimic = "lavaan")
fit1 <- cfa(" visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 ",
            test = "satorra.bentler",
            missing = "listwise",
            data = HolzingerSwineford1939)
fit1

# mimic the Mplus program
fit2 <- cfa(" visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 ",
            test = "satorra.bentler",
            missing = "listwise",
            data = HolzingerSwineford1939, mimic = "Mplus")
fit2

# mimic the EQS program
fit3 <- cfa(" visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 ",
            test = "satorra.bentler",
            missing = "listwise",
            data = HolzingerSwineford1939, mimic = "EQS")
fit3
}

\name{modificationIndices}
\alias{modificationIndices}
\alias{modificationindices}
\alias{modindices}
\title{Modification Indices}
\description{
Given a fitted lavaan object, compute the modification indices
(= univariate score tests) for a selected set of fixed-to-zero parameters.
}
\usage{
modificationIndices(object, standardized = TRUE, cov_std = TRUE,
                    information = "expected",
                    power = FALSE, delta = 0.1, alpha = 0.05,
                    high_power = 0.75, sort = FALSE, minimum_value = 0,
                    maximum_number = nrow(list_1), free_remove = TRUE,
                    na_remove = TRUE, op = NULL, ...)
modindices(object, standardized = TRUE, cov_std = TRUE, information = "expected",
                    power = FALSE, delta = 0.1, alpha = 0.05, high_power = 0.75,
                    sort = FALSE, minimum_value = 0,
                    maximum_number = nrow(list_1), free_remove = TRUE,
                    na_remove = TRUE, op = NULL, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{standardized}{If \code{TRUE}, two extra columns (sepc.lv and sepc.all)
will contain standardized values for the EPCs. In the first column (sepc.lv),
standardization is based on the variances of the (continuous) latent
variables. In the second column (sepc.all), standardization is based
on the variances of both the (continuous) observed and latent variables.
(Residual) covariances are standardized using (residual) variances.}
\item{cov_std}{Logical. See \code{\link{standardizedSolution}}.}
\item{information}{\code{character} indicating the type of information matrix
to use (check \code{\link{lavInspect}} for available options). \code{"expected"}
information is the default, which provides better control of Type I errors.}
\item{power}{If \code{TRUE}, the (post-hoc) power is computed for each
modification index, using the values of \code{delta} and \code{alpha}.}
\item{delta}{The value of the effect size, as used in the post-hoc power
computation, currently using the unstandardized metric of the epc column.}
\item{alpha}{The significance level used for deciding if the modification
index is statistically significant or not.}
\item{high_power}{If the computed power is higher than this cutoff value,
the power is considered `high'. If not, the power is considered `low'.
This affects the values in the 'decision' column in the output.}
\item{sort}{Logical. If TRUE, sort the output by the modification
index values. Higher values appear first.}
\item{minimum_value}{Numeric. Filter output and only show rows with a
modification index value equal to or higher than this minimum value.}
\item{maximum_number}{Integer. Filter output and only show the first
\code{maximum_number} rows. Most useful when combined with the \code{sort}
option.}
\item{free_remove}{Logical. If TRUE, filter output by removing all rows
corresponding to free (unconstrained) parameters in the original model.}
\item{na_remove}{Logical. If TRUE, filter output by removing all rows with
NA values for the modification indices.}
\item{op}{Character string. Filter the output by selecting only those rows with
operator \code{op}.}
\item{...}{To support old argument names.}
}
\value{
  A data.frame containing modification indices and EPCs.
}
\details{
  Modification indices are just 1-df (or univariate) score tests. The
  modification index (or score test) for a single parameter reflects
  (approximately) the improvement in model fit (in terms of the chi-square
  test statistic) if we were to refit the model with this parameter set
  free.
  This function is a convenience function in the sense that it produces a
  (hopefully sensible) table of currently fixed-to-zero (or fixed to another
  constant) parameters. For each of these parameters, a modification index
  is computed, together with an expected parameter change (epc) value.
  It is important to realize that this function will only consider
  fixed-to-zero parameters. If you have equality constraints in the model,
  and you wish to examine what happens if you release all (or some) of these
  equality constraints, use the \code{\link{lavTestScore}} function.
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
modindices(fit, minimum_value = 10, sort = TRUE)
}

\name{lav_samplestats}
\alias{lav_samp_from_data}
\title{lavaan samplestats functions}
\description{Utility functions related to the sample statistics}
\usage{
# generate samplestats object from full data
lav_samp_from_data(lavdata = NULL, lavoptions = NULL, 
                          wls_v = NULL, nacov = NULL)
}
\arguments{
\item{lavdata}{A lavdata object.}
\item{lavoptions}{A named list. The Options slot from a lavaan object.}
\item{wls_v}{A user provided weight matrix.}
\item{nacov}{A user provided matrix containing the elements of (N times)
    the asymptotic variance-covariance matrix of the sample statistics.
    For a multiple group analysis, a list with an asymptotic
    variance-covariance matrix for each group.}
}
\examples{
# generate syntax for an independence model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)

# extract data slot and options
lavdata <- fit@Data
lavoptions <- lavInspect(fit, "options")

# generate sample statistics object
sampleStats <- lav_samp_from_data(lavdata = lavdata, 
                                         lavoptions = lavoptions)
}

\name{lavResidualsY}
\alias{lavResidualsY}
\title{Compute residuals for the y-variables given the values of the x-variables}
\description{
This function can be used to compute the (case-wise) residuals for the
(observed) y-variables in a structural equation model, defined as the
difference between the observed values and the (model-based) predicted
values that are returned by \code{\link{lavPredictY}}.}
\usage{
lavResidualsY(object, newdata = NULL,
              ynames = lav_object_vnames(object, "ov.y"),
              xnames = lav_object_vnames(object, "ov.x"),
              method = "conditional.mean",
              label = TRUE, assemble = TRUE,
              force_zero_mean = FALSE,
              lambda = 0,
              ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{newdata}{An optional data.frame, containing the same variables as
the data.frame that was used when fitting the model in \code{object}.
This data.frame should also include
the y-variables (their observed values are used to compute the residuals).
Note that if no meanstructure was used in the original fit, we will
use the saturated sample means of the original fit as substitutes for the
model-implied means. Alternatively, refit the model using \code{meanstructure = TRUE}.
For a multiple-group model, \code{newdata} can also be a \emph{list} with one
data.frame per group (in the same order as the groups in \code{object}). An
element may be \code{NULL} to skip that group. This is useful when the groups
have a different set of variables, so that \code{ynames} and \code{xnames}
can be specified per group (as a list). In that case, the result contains
residuals only for the groups for which data were provided (a single matrix
if only one group was requested, otherwise a named list).}
\item{ynames}{The names of the observed variables that should be treated as the
y-variables. It is for these variables that the function will compute
the (model-based) residuals for each observation. Can also be a list to allow
for a separate set of variable names per group (or block).}
\item{xnames}{The names of the observed variables that should be treated as the
x-variables. Can also be a list to allow
for a separate set of variable names per group (or block).}
\item{method}{A character string. The only available option for now is
\code{"conditional.mean"}. See Details.}
\item{label}{Logical. If TRUE, the columns of the output are labeled.}
\item{assemble}{Logical. If TRUE,
the residuals for the separate groups in the output are
reassembled into a single data.frame
with a group column, having the same dimensions as the original
(or newdata) dataset.}
\item{force_zero_mean}{Logical. Only relevant if there is no mean structure.
If \code{TRUE}, the (model-implied) mean vector is set to the zero vector.
If \code{FALSE}, the (model-implied) mean vector is set to the (unrestricted)
sample mean vector.}
\item{lambda}{Numeric. A lambda regularization penalty term.}
\item{...}{To support old argument names.}
}
\details{
The residuals are computed as \code{observed - predicted}, where the
predicted values are obtained by \code{\link{lavPredictY}} using the same
arguments. See the help page of \code{\link{lavPredictY}} for more details
about how the predictions are computed.

These residuals can be used, for example, to diagnose the (SEM-based)
out-of-sample prediction of outcome (y) variables, or to check the
distributional assumptions (for example, multivariate normality) for the
y-variables given the x-variables.
}
\seealso{
\code{\link{lavPredictY}} to compute the (model-based) predicted values for
the y-variables given the x-variables.
}
\references{
de Rooij, M., Karch, J.D., Fokkema, M., Bakk, Z., Pratiwi, B.C, and
Kelderman, H. (2022) SEM-Based Out-of-Sample Predictions,
Structural Equation Modeling: A Multidisciplinary Journal.
\doi{10.1080/10705511.2022.2061494}
}
\examples{
model <- '
  # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + a*y2 + b*y3 + c*y4
     dem65 =~ y5 + a*y6 + b*y7 + c*y8

  # regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
'
fit <- sem(model, data = PoliticalDemocracy)

lavResidualsY(fit, ynames = c("y5", "y6", "y7", "y8"),
                   xnames = c("x1", "x2", "x3", "y1", "y2", "y3", "y4"))
}

\name{lav_plotinfo_rgraph}
\alias{lav_plotinfo_rgraph}
\title{Plots a graph in R}
\usage{
lav_plotinfo_rgraph(plotinfo,
                    sloped_labels = TRUE,
                    outfile = "",
                    addgrid = TRUE,
                    mlovcolors = c("lightgreen", "lightblue"),
                    lightness = 1,
                    italic = TRUE,
                    auto_subscript = TRUE
                   )
}
\arguments{
\item{plotinfo}{A plotinfo structure as returned from \code{lav_plotinfo_positions}.}
\item{sloped_labels}{Logical, should labels be sloped above (or below) the edges?}
\item{outfile}{Character string naming the file in which to store the diagram as
  a PNG, or NA to show the plot in R.}
\item{addgrid}{Logical, add a grid with indicated 'positions' to the graph?}
\item{mlovcolors}{Array of two colors for distinguishing ov nodes with the same
name across the levels of a multilevel model.}
\item{lightness}{A scalar factor to modify the distances between nodes.}
\item{italic}{Should labels be in an italic font? \strong{Attention:} switching
to an italic font is only possible if the label value is not an expression!}
\item{auto_subscript}{Logical, see \code{\link{lav_label_code}}.}
}
\value{
NULL (invisible)}
\description{
Creates a graph in R showing a simple diagram of the model.
}
\seealso{
\code{\link{lav_plotinfo_positions}}
}
\examples{
model <- 'alpha  =~ x1 + x2 + x_3     # latent variable
          beta <~ x4 + x5 + x6        # composite
          gamma =~ x7 + x8 + x9       # latent variable
          Xi =~ x10 + x11 + x12 + x13 # latent variable
          # regressions
          Xi ~ v * alpha + t * beta + c * 1
          alpha ~ yy * Theta1 + tt * beta + ss * gamma
          '
test <- lav_model_plotinfo(model)
test1 <- lav_plotinfo_positions(test)
lav_plotinfo_rgraph(test1, lightness = 1.1)
# better position for constant in regressen Xi, no sloped labels
test2 <- lav_plotinfo_positions(test, placenodes = list(`1vanXi` = c(2, 5)))
lav_plotinfo_rgraph(test2, FALSE, lightness = 1.1)
modelml <- '
        level: 1
        fw =~ 1*y_1 + y_2 + y_3 + y_5
        level: 2
        fb =~ 1*y_1 + y_2 + y_3 + y_4
        y_2 ~~ cv24 * y_4
  '
test <- lav_model_plotinfo(modelml)
test <- lav_plotinfo_positions(test)
lav_plotinfo_rgraph(test, sloped_labels = FALSE, addgrid = FALSE, 
                auto_subscript = FALSE)
\dontrun{
# example where plot is stored in a PNG file
lav_plotinfo_rgraph(test, sloped_labels = FALSE, addgrid = FALSE,
                auto_subscript = FALSE, outfile="demo_rplot.png")
}
}

\name{lavTables}
\alias{lavTables}
\title{lavaan frequency tables}
\description{
Frequency tables for categorical variables and related statistics.}
\usage{
lavTables(object, dimension = 2L, type = "cells", categorical = NULL,
    group = NULL, statistic = "default", g2_min = 3, x2_min = 3,
    p_value = FALSE, output = "data.frame", pattern_as_string = TRUE, ...)
}
\arguments{
\item{object}{Either a \code{data.frame}, or an object of class
\code{\linkS4class{lavaan}}.}
\item{dimension}{Integer. If 0L, display all response patterns. If 1L,
display one-dimensional (one-way) tables; if 2L, display two-dimensional
(two-way or pairwise) tables. For the latter, the information shown
per row can be changed: if \code{type = "cells"}, each row is a cell in a pairwise table;
if \code{type = "table"}, each row is a table.}
\item{type}{If \code{"cells"}, display information for each cell in the
(one-way or two-way) table. If \code{"table"}, display information per table.
If \code{"pattern"}, display response patterns (implying
\code{"dimension = 0L"}).}
\item{categorical}{Only used if \code{object} is a \code{data.frame}. Specify
variables that need to be treated as categorical.}
\item{group}{Only used if \code{object} is a \code{data.frame}. Specify
a grouping variable.}
\item{statistic}{Either a character string, or a vector of character strings
requesting one or more statistics for each cell, pattern or table.  Always
available are \code{X2} and \code{G2} for the Pearson and LRT based
goodness-of-fit statistics. A distinction is made between the unrestricted and
restricted model. The statistics based on the former carry the suffix
\code{*.un}, as in \code{X2.un} and \code{G2.un}.  If object is a
\code{data.frame}, only the unrestricted versions of the statistics are
available.  For one-way tables, additional statistics are the thresholds
(\code{th.un} and \code{th}). For two-way tables and \code{type = "table"}, the
following statistics are available: \code{X2}, \code{G2}, \code{cor}
(polychoric correlation), \code{RMSEA} and the corresponding unrestricted
versions (\code{X2.un} etc). Additional statistics are \code{G2.average},
\code{G2.nlarge} and \code{G2.plarge} statistics based on the cell values
\code{G2}:
\code{G2.average} is the average of the \code{G2} values in each cell of the
two-way table; \code{G2.nlarge} is the number of cells with a \code{G2} value
larger than \code{g2_min}, and \code{G2.plarge} is the proportion of cells with
a \code{G2} value larger than \code{g2_min}. A similar set of statistics based
on \code{X2} is also available. If \code{"default"}, the selection of
statistics (if any) depends on the \code{dim} and \code{type} arguments, and on
whether the object is a \code{data.frame} or a fitted lavaan object.}
\item{g2_min}{Numeric. All cells with a G2 statistic larger than this number
are considered `large', as reflected in the (optional) \code{"G2.plarge"} and
\code{"G2.nlarge"} columns.}
\item{x2_min}{Numeric. All cells with a X2 statistic larger than this number
are considered `large', as reflected in the (optional) \code{"X2.plarge"} and
\code{"X2.nlarge"} columns.}
\item{p_value}{Logical. If \code{"TRUE"}, p-values are computed for
requested statistics (eg G2 or X2) if possible.}
\item{output}{If \code{"data.frame"}, the output is presented as a data.frame
where each row is either a cell, a table, or a response pattern, depending
on the \code{"type"} argument. If \code{"table"}, the output is presented
as a table (or matrix) or a list of tables. Only a single statistic can be
shown in this case, and if the \code{statistic} is empty, the observed
frequencies are shown.}
\item{pattern_as_string}{Logical. Only used for response patterns (dimension = 0L). If \code{"TRUE"}, response patterns are displayed as a compact string.
If \code{"FALSE"}, as many columns as observed variables are displayed.}
\item{...}{To support old argument names.}
}
\value{
  If \code{output = "data.frame"}, the output is presented as a data.frame
  where each row is either a cell, a table, or a response pattern, depending on
  the \code{"type"} argument.
  If \code{output = "table"} (only for two-way tables), the output is
  a list of tables (if \code{type = "cells"}) in which each list element
  corresponds to a pairwise table, or a single table per group (if
  \code{type = "table"}). In both cases, the table entries are determined by the
  (single) \code{statistic} argument.
}
\references{
Joreskog, K.G. & Moustaki, I. (2001). Factor analysis of ordinal variables: A
comparison of three approaches. Multivariate Behavioral Research, 36, 347-387.
}
\seealso{\code{\link{varTable}}.}
\examples{
HS9 <- HolzingerSwineford1939[,c("x1","x2","x3","x4","x5",
                                 "x6","x7","x8","x9")]
HSbinary <- as.data.frame( lapply(HS9, cut, 2, labels=FALSE) )

# using the data only
lavTables(HSbinary, dim = 0L, categorical = names(HSbinary))
lavTables(HSbinary, dim = 1L, categorical = names(HSbinary), stat=c("th.un"))
lavTables(HSbinary, dim = 2L, categorical = names(HSbinary), type = "table")

# fit a model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HSbinary, ordered=names(HSbinary))


lavTables(fit, 1L)
lavTables(fit, 2L, type="cells")
lavTables(fit, 2L, type="table", stat=c("cor.un", "G2", "cor"))
lavTables(fit, 2L, type="table", output="table", stat="X2")
}

\name{lav_getcov}
\alias{lav_getcov}
\alias{getCov}
\alias{lav_cor2cov}
\alias{cor2cov}
\alias{lav_char2num}
\alias{char2num}
\title{Utility Functions For Covariance Matrices}
\description{Convenience functions to deal with covariance and correlation
matrices.}
\usage{
lav_getcov(x, lower = TRUE, diagonal = TRUE, sds = NULL,
       names = paste("V", 1:nvar, sep=""))
getCov(x, lower = TRUE, diagonal = TRUE, sds = NULL,
       names = paste("V", 1:nvar, sep=""))
lav_char2num(s)
char2num(s)
lav_cor2cov(r, sds, names = NULL, ...)
cor2cov(r, sds, names = NULL, ...)
}
\arguments{
\item{x}{The elements of the covariance matrix. Either inside a character 
  string or as a numeric vector. In the former case, the function
  \code{lav_char2num} is used to convert the numbers (inside the character
  string) to numeric values.}
\item{lower}{Logical. If \code{TRUE}, the numeric values in \code{x} are
the lower-triangular elements of the (symmetric) covariance matrix only.
If \code{FALSE}, \code{x} contains the upper triangular elements only.
Note we always assume the elements are provided row-wise!}
\item{diagonal}{Logical. If \code{TRUE}, the numeric values in \code{x}
include the diagonal elements. If \code{FALSE}, a unit diagonal is assumed.}
\item{sds}{A numeric vector containing the standard deviations to be 
used to scale the elements in \code{x} or the correlation matrix \code{r}
into a covariance matrix.}
\item{names}{The variable names of the observed variables.}
\item{s}{Character string containing numeric values; commas and semi-colons
are ignored.}
\item{r}{A correlation matrix, to be scaled into a covariance matrix.}
\item{...}{To accept old argument name \code{R} (is now \code{r}). No other arguments are accepted.} 
}
\details{
The \code{lav_getcov} function is typically used to input the lower 
(or upper) triangular elements of a (symmetric) covariance matrix. In many 
examples found in handbooks, only those elements are shown. However, lavaan 
needs a full matrix to proceed. 

The \code{lav_cor2cov} function is the inverse of the \code{\link[stats]{cov2cor}} 
function, and scales a correlation matrix into a covariance matrix given
the standard deviations of the variables. Optionally, variable names can
be given.
}
\examples{
# The classic Wheaton et. al. (1977) model 
# panel data on the stability of alienation
lower <- '
 11.834,
  6.947,    9.364,
  6.819,    5.091,   12.532,
  4.783,    5.028,    7.495,    9.986,
 -3.839,   -3.889,   -3.841,   -3.625,   9.610,
-21.899,  -18.831,  -21.748,  -18.775,  35.522,  450.288 '

# convert to a full symmetric covariance matrix with names
wheaton.cov <- lav_getcov(lower, names=c("anomia67","powerless67", "anomia71",
                                     "powerless71","education","sei"))

# the model
wheaton.model <- '
  # measurement model
    ses     =~ education + sei
    alien67 =~ anomia67 + powerless67
    alien71 =~ anomia71 + powerless71

  # equations
    alien71 ~ alien67 + ses
    alien67 ~ ses

  # correlated residuals
    anomia67 ~~ anomia71
    powerless67 ~~ powerless71
'

# fitting the model
fit <- sem(wheaton.model, sample_cov=wheaton.cov, sample_nobs=932)

# showing the results
summary(fit, standardized=TRUE)
}

\name{lavaan-deprecated}
\alias{lavaan-deprecated}
\title{Deprecated functions in lavaan package}
\details{
The functions listed in the first column below are deprecated; calls to them
should be replaced by calls to the corresponding function in the second column.
\tabular{ll}{
#  bootstrapLavaan \tab lavBootstrap \cr
  char2num \tab lav_char2num \cr
  cor2cov \tab lav_cor2cov \cr
  getCov \tab lav_getcov \cr
  inspectSampleCov \tab lavInspectSampleCov \cr
#  lavaanNames \tab lavNames \cr
  mplus2lavaan.modelSyntax \tab lav_mplus_syntax_model \cr
  mplus2lavaan \tab lav_Mplus_lavaan \cr
  parameterestimates \tab lavParameterEstimates \cr
#  simulateData \tab lavSimulateData \cr
}
}
\description{
Deprecated functions in lavaan.
}

\name{lavResiduals}
\alias{lavResiduals}
\alias{lavResidual}
\title{Residuals}
\description{
\sQuote{lavResiduals} provides model residuals and standardized residuals from
a fitted lavaan object, as well as various summaries of these residuals.

The \sQuote{residuals()} (and \sQuote{resid()}) methods are just shortcuts to
this function with a limited set of arguments.
}
\usage{
lavResiduals(object, type = "cor.bentler", h1 = NULL,
    se = FALSE, zstat = TRUE, summary = TRUE, elementwise = TRUE,
    combine = FALSE, usrmr_ci_level = 0.90, usrmr_close_h0 = 0.05,
    h1_acov = "unstructured",
    add_type = TRUE, add_labels = TRUE, add_class = TRUE,
    drop_list_single_group = TRUE,
    maximum_number = 0L, n_largest = 5L, output = "list", ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{type}{Character.
If \code{type = "raw"}, this function returns the raw (= unscaled)
difference between the observed and the expected (model-implied) summary
statistics, as well as the standardized version of these residuals.
If \code{type = "cor"}, or \code{type = "cor.bollen"}, the observed and model
implied covariance matrices are first transformed to a correlation matrix
(using \code{cov2cor()}), before the residuals are computed.
If \code{type = "cor.bentler"}, both the observed and model implied covariance
matrices are rescaled by dividing the elements by the square roots of the
corresponding variances of the observed covariance matrix.}
\item{h1}{Optional. A user-provided saturated (unrestricted) model supplying
the observed summary statistics against which the model-implied statistics are
compared. It can be a fitted \code{\linkS4class{lavaan}} object (its \code{h1}
slot is used), a \code{lavh1} list (a list with an \code{implied} element), or
an implied-moments list (a list with a \code{cov} or \code{res.cov} element,
per block). It must be compatible with \code{object} (same number of blocks
and same dimensions). If \code{NULL} (the default), the saturated statistics of
\code{object} itself are used.}
\item{se}{Logical. If \code{TRUE}, show the estimated standard errors
for the residuals.}
\item{zstat}{Logical. If \code{TRUE}, show the standardized residuals, which
are the raw residuals divided by the corresponding (estimated) standard
errors.}
\item{summary}{Logical. If \code{TRUE}, show various summaries of the
(possibly scaled) residuals. When \code{type = "raw"}, we compute the
RMR. When \code{type = "cor.bentler"},
we compute the SRMR. When \code{type = "cor.bollen"}, we compute the CRMR.
An unbiased version of these summaries (the URMR, USRMR and UCRMR) is also
computed (Maydeu-Olivares, 2017).

The two versions are reported with different inferential statistics, and this
asymmetry is intentional. The sample RMR/SRMR/CRMR overestimate (are positively
biased for) their population values. For these (biased) statistics, we report a
standard error and a test of \emph{exact} fit (the null hypothesis that the
population value equals zero). This test
relies on the known sampling distribution of the (biased) sample statistic
under exact fit (Maydeu-Olivares, 2017, Eqs. 28--30), where the bias is
explicitly accounted for. For the (approximately) unbiased URMR/USRMR/UCRMR, we
report a standard error, a confidence interval, and a test of \emph{close} fit
(the null hypothesis that the population value equals a small nonzero cutoff,
e.g. 0.05). The confidence interval and the close-fit test require an
(approximately) unbiased and normally distributed estimator, which is precisely
why the unbiased version is used here. Conversely, no test of exact fit is
reported for the unbiased estimator: because it is truncated at zero, its
sampling distribution is degenerate at the exact-fit boundary, so the test of
exact fit is (correctly) only based on the biased statistic.}
\item{elementwise}{Logical. If \code{TRUE} (the default), the element-wise
residuals are included: the (residualized) summary statistics when
\code{output = "list"}, or the largest-residuals table(s) when
\code{output = "text"}. If \code{FALSE}, these are omitted and only the summary
information is returned/shown (provided \code{summary = TRUE}). Combined with
\code{summary = FALSE}, this gives full control over which parts are produced.}
\item{combine}{Logical. Only relevant when multiple blocks are
involved (multiple groups, or multiple levels). If \code{FALSE} (the default),
a separate summary table is shown for each block. If \code{TRUE}, the
block-specific summary tables are replaced by a single overall summary table in
which the residual elements are pooled across all blocks (so, e.g., the SRMR is
computed from the residuals of all groups or all levels at once). The combined
table is returned as the \code{summary} element at the top level of the result.
The pooling correctly accounts for the (block-diagonal) independence of separate
groups and for the (joint) sampling covariance across the levels of a multilevel
model.}
\item{usrmr_ci_level}{Numeric. The confidence level of the confidence interval
that is reported (in the summary) for the unbiased URMR/USRMR/UCRMR. The default
is 0.90.}
\item{usrmr_close_h0}{Numeric. The value of the population URMR/USRMR/UCRMR under
the null hypothesis of the test of close fit that is reported (in the summary)
for the unbiased estimator. The default is 0.05.}
\item{h1_acov}{Character. If \code{"unstructured"}, the observed summary
statistics are used as consistent estimates of the corresponding (unrestricted)
population statistics. If \code{"structured"}, the model-implied summary
statistics are used as consistent estimates of the corresponding (unrestricted)
population statistics. This affects the way the asymptotic variance matrix
of the summary statistics is computed.}
\item{add_type}{Logical. If \code{TRUE}, show the type of residuals
in the output.}
\item{add_labels}{If \code{TRUE}, variable names are added to the vectors
and/or matrices.}
\item{add_class}{If \code{TRUE}, vectors are given the \sQuote{lavaan.vector}
class; matrices are given the \sQuote{lavaan.matrix} class, and symmetric
matrices are given the \sQuote{lavaan.matrix.symmetric} class.
This only affects the way they are printed on the screen.}
\item{drop_list_single_group}{If \code{FALSE}, the results are returned as
a list, where each element corresponds to a group (even if there is only
a single group). If \code{TRUE}, the list will be unlisted if there is
only a single group.}
\item{maximum_number}{Integer. Only used if \code{output = "table"}. If larger
than zero, show only the first maximum_number rows of the data.frame. The
default (\code{0L}) shows all rows.}
\item{n_largest}{Integer. Only used if \code{output = "text"}. The number of
largest (in absolute value) residuals that are shown, per element group
(covariances, variances, intercepts, thresholds, regression slopes) and per
block.}
\item{output}{Character. By default, \code{output = "list"}, and the output
is a list of elements. If \code{output = "table"}, the residuals are shown in
a single data.frame, one row per residual element, sorted from high (in
absolute value) to low. All residual elements are included: the (co)variances
(\code{lhs~~rhs}), the means/intercepts (\code{lhs~1}), the thresholds
(\code{lhs|t1}) and, when \code{conditional.x = TRUE}, the regression slopes
(\code{lhs~rhs}). Variances (the diagonal) are only shown for the raw-style
metrics (\code{type = "raw"}, \code{"normalized"} or \code{"standardized"}),
not for the correlation metrics where they are structurally zero. Structurally
fixed moments (e.g. saturated means or just-identified thresholds) are omitted.
If \code{se = TRUE}, a \code{se} column with the standard errors is added; if
\code{zstat = TRUE}, a \code{z} column with the z-statistics is added.
If \code{output = "text"} (or, equivalently, \code{"pretty"}), the output is
printed in the same style as the fit measures (e.g.\ the RMSEA). First, a table
with the \code{n_largest} largest residuals (per element group, and per block)
is shown, formatted like \code{parameterEstimates(., output = "text")} but with
a \sQuote{Residual} column (and, if available, \sQuote{Std.Err} and
\sQuote{Z-value} columns); its header reflects the residual \code{type} (e.g.\
\dQuote{Largest Standardized Residuals (Cor.Bentler)} or \dQuote{Largest Raw
Residuals}). Next, the residual summary table(s) are shown. When multiple blocks
are involved and \code{combine = FALSE}, one summary is printed per block;
otherwise a single (combined) summary is printed.}
\item{...}{To support old argument names.}
}
\value{
If \code{drop_list_single_group = TRUE}, a list of (residualized) summary
statistics, including type, standardized residuals, and summaries. If
\code{drop_list_single_group = FALSE}, the list of summary statistics is nested
within a list for each group.
}
\references{
Bentler, P.M. and Dijkstra, T. (1985). Efficient estimation via linearization
in structural models. In Krishnaiah, P.R. (Ed.),
\emph{Multivariate analysis - VI}, (pp. 9--42). New York, NY: Elsevier.

Ogasawara, H. (2001). Standard errors of fit indices using residuals in
structural equation modeling. \emph{Psychometrika, 66}(3), 421--436.
doi:10.1007/BF02294443

Maydeu-Olivares, A. (2017). Assessing the size of model misfit in structural
equation models. \emph{Psychometrika, 82}(3), 533--558.
doi:10.1007/s11336-016-9552-7

Standardized Residuals in M\emph{plus}. Document retrieved from URL
http://www.statmodel.com/download/StandardizedResiduals.pdf
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data = HolzingerSwineford1939)
lavResiduals(fit)
}

\name{lavInspectSampleCov}
\alias{lavInspectSampleCov}
\alias{inspectSampleCov}
\title{Observed Variable Covariance Matrix from a Model and Data}
\description{
The lavaan model syntax describes a latent variable model. Often, the user wants to see the sample covariance matrix of the variables in their model for diagnostic purposes. However, their data may contain many more columns than the variables used in the model.}
\usage{
lavInspectSampleCov(model, data, ...)
inspectSampleCov(model, data, ...)
}
\arguments{
\item{model}{The model that will be fit by lavaan.}
\item{data}{The data frame being used to fit the model.}
\item{...}{Other arguments to \code{\link{sem}} for how to deal with 
  multiple groups, missing values, etc.}
}
\author{Jarrett Byrnes}
\details{
One must supply both a model, coded with proper \code{\link{model.syntax}}, and
a data frame from which a covariance matrix will be calculated. This function
essentially calls \code{\link{sem}} without fitting the model, and then uses
\code{\link{lavInspect}} to obtain the sample covariance matrix and the
meanstructure.
}

\section{See also}{\code{\link{sem}}, \code{\link{lavInspect}} }

\name{lavEffects}
\alias{lavEffects}
\alias{laveffects}
\title{Indirect and Total Effects}
\description{
\sQuote{lavEffects} computes various \sQuote{effects} that are functions of
the estimated model parameters of a fitted lavaan object. In this version, the
focus is on the classic (LISREL-style) \emph{total}, \emph{indirect} and
\emph{direct} effects among the variables that appear in the structural part of
the model (that is, the variables that are involved in a regression). Both
observed and latent variables are supported.

This avoids the need to pre-specify all indirect (and total) effects manually
in the model syntax (using the \code{:=} operator), which can be tedious when
there are many of them.
}
\usage{
lavEffects(object, effects = c("total", "indirect"),
    se_def = NULL, level = 0.95, monte_carlo = NULL,
    boot_ci_type = "perc", zstat = TRUE, pvalue = TRUE, ci = TRUE,
    standardized = FALSE, cov_std = TRUE,
    add_class = TRUE, output = "data.frame", ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{effects}{Character vector. One or more of \code{"total"},
    \code{"indirect"} and \code{"direct"}. The default is
    \code{c("total", "indirect")}. The order of the elements does not matter;
    effects are always reported in the order total, indirect, direct.}
\item{se_def}{Character (or \code{NULL}). The method used to compute standard
    errors (and confidence intervals) for the effects. If \code{"monte_carlo"},
    the Monte Carlo method is used: parameters are drawn from a multivariate
    normal distribution with mean the parameter estimates and covariance matrix
    the estimated covariance matrix of the parameters, and the effects are
    recomputed for each draw (percentile confidence intervals). If
    \code{"delta"}, the delta method is used (based on a numerical Jacobian of
    the effects with respect to the free parameters; symmetric normal-theory
    confidence intervals). If \code{"bootstrap"}, the bootstrap draws that are
    stored in the fitted object (when the model was fitted with
    \code{se = "bootstrap"}) are used to compute bootstrap standard errors and
    bootstrap percentile confidence intervals. If \code{"none"}, no standard
    errors are computed.

    If \code{se_def = NULL} (the default), the choice is inherited from the
    fitted object: if the model was fitted with \code{se = "bootstrap"} (and
    bootstrap draws are available), the bootstrap method is used; else if the
    model was fitted with \code{se_def = "monte.carlo"}, the Monte Carlo method
    is used; otherwise the Monte Carlo method is used as the default. An
    explicit value always overrides this inheritance.}
\item{level}{Numeric. The confidence level for the confidence intervals.}
\item{monte_carlo}{List (or \code{NULL}). Settings for the Monte Carlo method
    (only used when \code{se_def = "monte.carlo"}), with elements \code{R} (the
    number of Monte Carlo draws) and \code{seed} (the random seed). If
    \code{NULL} (the default), the \code{monte_carlo} settings that were used
    when the model was fitted are inherited (which themselves default to
    \code{list(R = 20000L, seed = NULL)}). The state of the global random
    number generator is saved and restored, so the seed used here does not
    affect the user's random number stream.}
\item{boot_ci_type}{Character. Only used when \code{se_def = "bootstrap"}. The
    type of bootstrap confidence interval, as in \code{\link{parameterEstimates}}:
    \code{"norm"} (normal-theory using the bootstrap bias and standard error),
    \code{"basic"} (basic bootstrap), \code{"perc"} (the default; percentile
    method), \code{"bca.simple"} (bias-corrected percentile method, without the
    acceleration term) or \code{"bca"} (the adjusted bootstrap percentile
    method, with both bias and acceleration correction).}
\item{zstat}{Logical. If \code{TRUE}, an extra column is added with the
    \eqn{z}-statistic for each effect (the ratio of the estimate to its
    standard error).}
\item{pvalue}{Logical. If \code{TRUE}, an extra column is added with the
    (two-sided) p-value, computed using the standard normal distribution.}
\item{ci}{Logical. If \code{TRUE} (the default), two extra columns
    (\code{ci.lower} and \code{ci.upper}) are added with the lower and upper
    bounds of the confidence interval. If \code{FALSE}, the confidence interval
    bounds are not shown.}
\item{standardized}{Logical, character vector, or vector of (observed) variable
    names, behaving as in \code{\link{parameterEstimates}}. If \code{TRUE},
    standardized effects are added as extra columns (point estimates only):
    \code{std.lv} (only the latent variables are standardized), \code{std.all}
    (both observed and latent variables are standardized) and, if there are
    exogenous observed variables and \code{fixed.x = TRUE}, \code{std.nox}
    (like \code{std.all}, but the exogenous observed variables are not
    standardized). A character vector selects specific types (a subset of
    \code{"std.lv"}, \code{"std.all"}, \code{"std.nox"}); a vector of observed
    variable names standardizes only those observed variables (plus the latent
    variables), and the result is reported in a \code{std.user} column. The
    standardized effect of \eqn{j} on \eqn{i} equals the unstandardized effect
    times the ratio of the model-implied standard deviations
    \eqn{sd_j / sd_i}.}
\item{cov_std}{Logical. See \code{\link{parameterEstimates}}. (Has no influence
    on the standardized regression-type effects reported here.)}
\item{add_class}{Logical. If \code{TRUE}, the result (when
    \code{output = "data.frame"}) is given the class
    \code{c("lavaan.effects", "lavaan.data.frame", "data.frame")}, so that a
    dedicated print method is used.}
\item{output}{Character. If \code{"data.frame"} (the default), the result is a
    (classed) data.frame with one row per effect. If \code{"list"}, the result
    is a list of matrices (one per effect type, and one per block in the
    multilevel/multigroup case), where element \eqn{[i,j]} contains the effect
    of variable \eqn{j} on variable \eqn{i}.}
\item{...}{Only old names of arguments - with dots - are allowed here.}
}
\details{
All effects are derived from the reduced-form matrix
\eqn{(I - B)^{-1}}, where \eqn{B} is the matrix of (direct) regression
coefficients among the structural variables (the \code{beta} matrix in the
LISREL representation that lavaan uses internally). Writing \eqn{B[i,j]} for the
direct effect of variable \eqn{j} on variable \eqn{i}, we have:
\itemize{
  \item the \emph{direct} effect of \eqn{j} on \eqn{i} equals \eqn{B[i,j]};
  \item the \emph{total} effect of \eqn{j} on \eqn{i} equals
        \eqn{\left((I - B)^{-1} - I\right)[i,j]};
  \item the \emph{indirect} effect of \eqn{j} on \eqn{i} equals the total
        effect minus the direct effect.
}
Only effects that are structurally non-zero (that is, for which a directed path
from \eqn{j} to \eqn{i} exists) are reported. For indirect effects, this means
that at least one path of length two or more must exist.

The delta and Monte Carlo methods only require the parameter estimates and the
estimated covariance matrix of the parameters. The delta method produces
(symmetric) normal-theory confidence intervals; the Monte Carlo method produces
percentile-based confidence intervals, which need not be symmetric around the
point estimate. The Monte Carlo method typically behaves better than the delta
method for effects that are products of parameters (such as indirect effects),
in particular in smaller samples. When the model was fitted with
\code{se = "bootstrap"}, the bootstrap draws are reused to obtain bootstrap
standard errors and bootstrap percentile confidence intervals (this requires
no additional model fitting).

Models fitted with \code{conditional.x = TRUE} are also supported: in that
case the structural model is \eqn{eta = B eta + Gamma x + zeta}, where the
exogenous covariates \eqn{x} are stored in the \code{gamma} matrix. The effects
of these covariates on the endogenous variables are then computed as
\eqn{(I - B)^{-1} Gamma} (total), \eqn{Gamma} (direct) and
\eqn{((I - B)^{-1} - I) Gamma} (indirect), and reported alongside the effects
among the \code{beta} variables.
}
\value{
If \code{output = "data.frame"} (the default), a data.frame (of class
\code{lavaan.effects}) with the following columns: \code{effect} (the type of
effect: total, indirect or direct), \code{lhs} (the outcome variable),
\code{op} (always \code{"~"}), \code{rhs} (the predictor variable), and,
depending on the arguments, \code{group}/\code{level}/\code{block},
\code{est}, \code{se}, \code{z}, \code{pvalue}, \code{ci.lower},
\code{ci.upper} and (if \code{standardized} is requested) one or more of
\code{std.lv}, \code{std.all}, \code{std.nox} and \code{std.user}. The effect
in a given row is the effect of \code{rhs} on \code{lhs}.

If \code{output = "list"}, a (possibly nested) list of matrices, where element
\eqn{[i,j]} contains the effect of variable \eqn{j} on variable \eqn{i}.
}
\seealso{
\code{\link{parameterEstimates}}, \code{\link{standardizedSolution}}.
}
\examples{
# a simple mediation model
set.seed(1234)
X <- rnorm(300)
M <- 0.5 * X + rnorm(300)
Y <- 0.4 * M + 0.3 * X + rnorm(300)
Data <- data.frame(X = X, Y = Y, M = M)
model <- ' # direct effect
             Y ~ c*X
           # mediator
             M ~ a*X
             Y ~ b*M
         '
fit <- sem(model, data = Data)

# total and indirect effects, with Monte Carlo standard errors
lavEffects(fit)

# all effects, using the delta method
lavEffects(fit, effects = c("total", "indirect", "direct"), se_def = "delta")

# add standardized (total and indirect) effects
lavEffects(fit, se_def = "delta", standardized = TRUE)
}

\name{estimator_iv}
\alias{estimator_iv}
\alias{iv}
\alias{IV}
\alias{miiv}
\alias{MIIV}
\alias{miivsem}
\alias{MIIVsem}
\alias{instrument}
\alias{instruments}
\alias{sargan}
\alias{Sargan}
\title{Instrumental Variable (MIIV-2SLS) Estimation}
\description{
Estimate a structural equation model using model-implied instrumental
variables and two-stage least squares (MIIV-2SLS), a non-iterative,
equation-by-equation alternative to maximum likelihood (Bollen, 1996).
This page documents how to invoke the estimator, the options that can be
passed via the \code{estimator = list()} argument, how to specify
instruments manually with the \code{|~} operator, and the \code{lavInspect()}
fields that are relevant for IV estimation.
}
\details{
Instead of minimizing a single discrepancy function for the whole model,
MIIV-2SLS estimates each measurement and structural equation separately by
two-stage least squares (2SLS). For each equation, the instruments are other
observed variables in the model that are (under the hypothesized model)
uncorrelated with the equation's composite error term: the \emph{model-implied
instrumental variables} (MIIVs). Because the equations are estimated
separately, a misspecification in one equation does not necessarily propagate
to the others, the method is non-iterative, and no starting values are
required.

The estimator is selected by setting \code{estimator = "IV"} (the alias
\code{"MIIV"} is also accepted; names are case-insensitive), for example
\preformatted{    fit <- sem(model, data = Data, estimator = "IV")}

Estimation proceeds in two stages. In the first stage, the \emph{directed}
effects (factor loadings and regression coefficients) are estimated
equation-by-equation by 2SLS. In the second stage, the \emph{undirected}
effects (residual variances and covariances) are estimated by a weighted
least-squares step applied to the residual moments. Standard errors are
available for both stages (see the options below).

Both continuous and ordered-categorical data are supported (for ordered data,
the polychoric-based PIV estimator of Bollen & Maydeu-Olivares, 2007 is used).
Simple equality constraints (e.g. equal factor loadings) and general linear
equality constraints (e.g. \code{a == 2*b}) are honored. Multiple groups are
supported, with standard errors (for continuous, two-stage-missing and
categorical data), including the mean structure and cross-group equality
constraints for measurement invariance testing (configural, metric and scalar,
e.g. via \code{group.equal = c("loadings", "intercepts")}). For scalar
invariance the latent-mean (and intercept) estimates are obtained by a joint
GLS solve over the mean structure (see \code{iv_mean_structure}), which matches
the maximum likelihood mean estimates given the model-implied covariance
matrix. The IV estimator is restricted to single-level models.

\subsection{Estimator options}{

Options specific to the IV estimator are passed as named elements of a list
to the \code{estimator} argument, alongside the estimator name:
\preformatted{    fit <- sem(model, data = Data,
               estimator = list(estimator = "IV",
                                iv_varcov_method = "2RLS",
                                iv_sargan_adjust = "BH"))}

The option names use snake_case. For backward compatibility, dot.case names
(e.g. \code{iv.varcov.method}) are also accepted and silently converted.

\describe{
  \item{\code{iv_method}:}{Character. The instrumental-variable method.
    Currently only \code{"2SLS"} (two-stage least squares) is available.
    Default is \code{"2SLS"}.}
  \item{\code{iv_samplestats}:}{Logical. If \code{TRUE} (the default), the
    equations are estimated from the sample moments (covariances and means)
    rather than from the raw data. This is required for categorical data and
    when only sample moments are provided as input.}
  \item{\code{iv_varcov_method}:}{Character. The second-stage method used to
    estimate the residual variances and covariances. One of \code{"ULS"},
    \code{"GLS"}, \code{"2RLS"}, \code{"RLS"} or \code{"NONE"}. Default is
    \code{"RLS"} (reweighted least squares). For categorical data, \code{"ULS"}
    is used. If \code{"NONE"}, the variance/covariance parameters are not
    estimated.}
  \item{\code{iv_sargan}:}{Logical. If \code{TRUE} (the default),
    overidentification tests are computed for each overidentified equation. Two
    tests are reported side by side: the classical Sargan test (Sargan, 1958),
    which assumes normality (continuous data) and is not valid for polychoric
    correlations; and a robust, residual-based test that is valid more
    generally. The robust test is Browne's (1984) residual statistic applied to
    the single equation, computed with the asymptotic covariance matrix (ACOV)
    of the sample statistics: the distribution-free (ADF) ACOV of the sample
    covariances for continuous data (so it is robust to non-normality, and is
    then equivalent to Hansen's overidentification test), or the polychoric
    ACOV for categorical data. Both tests share the same degrees of freedom
    (number of instruments minus number of regressors). For continuous data
    with missing values the robust test is not available. The results can be
    retrieved with \code{lavInspect(fit, "sargan")} (or, equivalently,
    \code{lavInspect(fit, "hansen")}).}
  \item{\code{iv_sargan_adjust}:}{Character. A multiple-comparison adjustment
    applied to the per-equation overidentification p-values, using any method
    accepted by \code{\link[stats]{p.adjust}} (\code{"holm"}, \code{"hochberg"},
    \code{"hommel"}, \code{"bonferroni"}, \code{"BH"}, \code{"BY"},
    \code{"fdr"}, or \code{"none"}). Default is \code{"none"}. When a method
    other than \code{"none"} is requested, extra \code{sargan.pval.adj} and
    \code{browne.pval.adj} columns are added to the table returned by
    \code{lavInspect(fit, "sargan")}.}
  \item{\code{iv_weak}:}{Character. How to respond to weak instruments,
    diagnosed per equation by the first-stage F-statistic (Staiger & Stock,
    1997). One of:
    \describe{
      \item{\code{"warn"}:}{(the default) report the affected equations and
        their first-stage F, and suggest supplying stronger instruments
        manually; the estimates are not changed.}
      \item{\code{"prune"}:}{greedily drop the weakest excess instruments from
        weak, overidentified equations (never below just-identified) and report
        what was dropped.}
      \item{\code{"none"}:}{skip the weak-instrument check.}
    }}
  \item{\code{iv_weak_threshold}:}{Numeric. The first-stage F-statistic below
    which instruments are deemed weak. Default is \code{10}.}
  \item{\code{iv_vcov_stage1}:}{Character. The standard-error method for the
    first-stage (directed) coefficients. One of \code{"lm.vcov"},
    \code{"lm.vcov.dfres"}, \code{"gamma"} or \code{"none"}. Default is
    \code{"lm.vcov.dfres"} for continuous data and \code{"gamma"} for
    categorical data. The \code{"gamma"} method requires
    \code{iv_samplestats = TRUE}. When the model contains \emph{simple}
    equality constraints among the directed coefficients (e.g. equal factor
    loadings), the \code{"lm.vcov"}/\code{"lm.vcov.dfres"} methods use a
    variance-weighted restricted-2SLS covariance for the constrained
    coefficients (as in the MIIVsem package); set \code{iv_vcov_stage1 =
    "gamma"} to obtain the delta-method (moment-Jacobian) standard errors
    instead.}
  \item{\code{iv_vcov_stage2}:}{Character. The standard-error method for the
    second-stage (undirected) parameters. One of \code{"h2"}, \code{"delta"}
    or \code{"none"}. Default is \code{"h2"}. The \code{"h2"} method requires
    \code{iv_samplestats = TRUE}.}
  \item{\code{iv_vcov_gamma_modelbased}:}{Logical. If \code{TRUE} (the
    default), the normal-theory weight matrix (gamma) used for the standard
    errors is based on the model-implied covariance matrix; if \code{FALSE},
    it is based on the unrestricted (H1) sample covariance matrix.}
  \item{\code{iv_mean_structure}:}{Character. How the free mean-structure
    parameters (observed intercepts and latent means) are estimated. With
    \code{"wls"} (the default) they are obtained by a joint GLS solve that fits
    the model-implied means to the sample means across all groups, pooling
    parameters that are constrained equal across groups (e.g. intercepts for
    scalar invariance); this matches the maximum likelihood mean estimates
    (given the model-implied covariance matrix). With \code{"moments"} the
    intercepts are recomputed per equation and shared intercepts are pooled by
    a (simpler) nobs-weighted average. The two agree unless intercepts are
    constrained across groups (scalar invariance).}
  \item{\code{iv_vcov_jack_numerical}, \code{iv_vcov_jaca_numerical},
    \code{iv_vcov_jacb_numerical}:}{Logical. If \code{TRUE}, the corresponding
    Jacobians used in the standard-error computation are obtained numerically
    rather than from analytic expressions. The default is \code{FALSE}; these
    are mainly intended for checking the analytic derivatives.}
  \item{\code{iv_mimic_ml}:}{Logical. If \code{FALSE} (the default), the IV
    estimator uses unbiased divisors: the sample covariance matrix is divided
    by \eqn{N-1} (\code{sample.cov.rescale = FALSE}) and the per-equation
    residual variances (the RSS terms entering the standard errors and the
    Sargan test) use the \eqn{N-P} divisor. If \code{TRUE}, the maximum
    likelihood divisor \eqn{N} is used throughout: the sample covariance is
    rescaled to the ML divisor \eqn{N} and the residual variances use the
    \eqn{N} divisor as well. As a result, for just-identified models the IV
    point estimates and standard errors match the ML solution exactly.}
}
}

\subsection{Specifying instruments manually (the \code{|~} operator)}{

By default lavaan determines the model-implied instruments for each equation
automatically. The \code{|~} operator overrides this choice for a specific
equation: the left-hand side is the dependent variable of the equation (for a
latent variable, its scaling indicator), and the right-hand side lists the
instruments to use.

\preformatted{    model <- '
        ind60 =~ x1 + x2 + x3
        dem60 =~ y1 + y2 + y3 + y4
        dem60 ~ ind60
        # use only x2 and x3 as instruments for the 
        # dem60 ~ ind60 equation
        y1 |~ x2 + x3
    '
    fit <- sem(model, data = PoliticalDemocracy, estimator = "IV")}

This is useful when the automatically selected instruments are weak (see
\code{iv_weak} above) or when subject-matter knowledge suggests a different
set of instruments.

\emph{External instruments.} The instruments listed on the right-hand side of
\code{|~} need not be part of the model. If an instrument is an observed
variable that does not otherwise appear in the model (it is not an indicator,
a predictor, an outcome, or otherwise mentioned), it is treated as an
\emph{external} instrument: it is read from the data so that its covariances
with the model variables are available to the 2SLS estimator, but it never
enters the model-implied summary statistics. The model degrees of freedom, the
fit measures, and the model-implied moments are therefore exactly the same as
if the instrument had not been mentioned; only the equations for which it is
used are affected. External instruments require raw data (a \code{data=}
argument); they are not available when only sample moments are supplied. For
categorical data the external instruments must themselves be ordered
(categorical); the augmented polychoric correlations and the asymptotic
covariance of the sample statistics are then used for the estimates and the
standard errors. Single-group and multiple-group categorical models are
supported, but (for now) not equality constraints among the directed
coefficients (e.g. measurement invariance) together with external instruments.

\preformatted{    ## z is not part of the model; it is used only as an
    ## instrument for the (endogenous) regression y ~ x
    model <- '
        y ~ x
        y |~ z
    '
    fit <- sem(model, data = Data, estimator = "IV")}

For external instruments the point estimates and the directed-coefficient
(loading and regression) standard errors are the usual 2SLS quantities. The
second-stage (residual variance and covariance) standard errors also account
for the sampling variability of the instrument moments: the directed
coefficients depend on the instrument-model covariances, and this uncertainty
is propagated into the variance/covariance parameters through the full
(augmented) moment Jacobian.
}

\subsection{Missing data}{

For continuous data with missing values, the IV estimator defaults to a
two-stage (FIML) approach: the saturated mean vector and covariance matrix are
estimated by the EM algorithm, the model is then estimated from these moments,
and the standard errors are corrected for the additional uncertainty stemming
from the EM estimates (Savalei & Bentler, 2009). This default is used when the
data contain missing values and the user does not set the \code{missing}
argument. Set \code{missing = "robust.two.stage"} for sandwich-corrected
standard errors (Savalei & Falk, 2014), or \code{missing = "listwise"} to
delete incomplete cases instead. Two-stage missing-data estimation is also
available for multiple-group models.
}

\subsection{Multiple groups and measurement invariance}{

Multiple-group models are supported, with point estimates and standard errors
for continuous, two-stage-missing and categorical data. Cross-group equality
constraints are imposed in the usual way, either through shared parameter
labels in the model syntax or through the \code{group.equal} argument, so the
standard measurement-invariance sequence is available: configural
(\code{group.equal = NULL}), metric (\code{group.equal = "loadings"}) and
scalar (\code{group.equal = c("loadings", "intercepts")}).

A parameter that is constrained equal across groups is estimated by pooling the
information from all groups; for a configural model the per-group estimates (and
their standard errors) reproduce the separate single-group fits. The mean
structure (observed intercepts and latent means) is estimated by a joint GLS
solve over all groups (see \code{iv_mean_structure}), which reproduces the
maximum likelihood mean estimates given the model-implied covariance matrix.
}

\subsection{Inspecting the instruments and the Sargan test}{

Several \code{\link{lavInspect}} fields are specific to IV estimation:
\describe{
  \item{\code{lavInspect(fit, "iv")}:}{(aliases \code{"miiv"},
    \code{"instruments"}) a per-equation table with the dependent variable
    (\code{lhs}), the regressors (\code{rhs}), the observed variables actually
    used (\code{lhs_new}, \code{rhs_new}), the equation type (\code{type}:
    \code{"miiv"}, \code{"ols"} or \code{"user"}), and the instruments
    (\code{iv}) for that equation.}
  \item{\code{lavInspect(fit, "sargan")}:}{a per-equation table with the
    overidentification tests: the degrees of freedom (\code{df}), the classical
    Sargan statistic (\code{sargan.stat}) and its p-value (\code{sargan.pval}),
    and the robust Browne residual-based statistic (\code{browne.stat}) and its
    p-value (\code{browne.pval}); see \code{iv_sargan} above. For categorical
    data the classical Sargan p-value is \code{NA} (not valid), so use the
    \code{browne.*} columns. Adjusted p-values (\code{sargan.pval.adj} and
    \code{browne.pval.adj}) are added when \code{iv_sargan_adjust} is set.
    \code{lavInspect(fit, "hansen")} is an alias.}
  \item{\code{lavInspect(fit, "eqs")}:}{the underlying per-equation
    information.}
}
}
}
\references{
Bollen, K. A. (1996). An alternative two stage least squares (2SLS) estimator
for latent variable equations. \emph{Psychometrika}, 61(1), 109-121.

Bollen, K. A., & Maydeu-Olivares, A. (2007). A polychoric instrumental variable
(PIV) estimator for structural equation models with categorical variables.
\emph{Psychometrika}, 72(3), 309-326.

Browne, M. W. (1984). Asymptotically distribution-free methods for the analysis
of covariance structures. \emph{British Journal of Mathematical and Statistical
Psychology}, 37(1), 62-83.

Sargan, J. D. (1958). The estimation of economic relationships using
instrumental variables. \emph{Econometrica}, 26(3), 393-415.

Savalei, V., & Bentler, P. M. (2009). A two-stage approach to missing data:
Theory and application to auxiliary variables.
\emph{Structural Equation Modeling}, 16(3), 477-497.

Savalei, V., & Falk, C. F. (2014). Robust two-stage approach outperforms robust
full information maximum likelihood with incomplete nonnormal data.
\emph{Structural Equation Modeling}, 21(2), 280-302.

Staiger, D., & Stock, J. H. (1997). Instrumental variables regression with weak
instruments. \emph{Econometrica}, 65(3), 557-586.
}
\seealso{
\code{\link{lavaan}}, \code{\link{sem}}, \code{\link{lavOptions}},
\code{\link{lavInspect}}.
}
\examples{
## The classic Bollen (1989) Political Democracy example
model <- '
  # measurement model
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + y2 + y3 + y4
    dem65 =~ y5 + y6 + y7 + y8
  # regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60
'

## fit using MIIV-2SLS instead of maximum likelihood
fit <- sem(model, data = PoliticalDemocracy, estimator = "IV")
summary(fit)

## the model-implied instruments used for each equation
lavInspect(fit, "iv")

## the per-equation Sargan overidentification tests
lavInspect(fit, "sargan")

## pass estimator options via estimator = list(...)
fit2 <- sem(model, data = PoliticalDemocracy,
            estimator = list(estimator = "IV",
                             iv_varcov_method = "2RLS",
                             iv_sargan_adjust = "BH",
                             iv_weak = "warn"))
lavInspect(fit2, "sargan")

## specify instruments manually with the |~ operator:
## use only x2, x3 as instruments for the dem60 ~ ind60 equation
model.iv <- '
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + y2 + y3 + y4
    dem60 ~ ind60
    y1 |~ x2 + x3
'
fit3 <- sem(model.iv, data = PoliticalDemocracy, estimator = "IV")
lavInspect(fit3, "iv")

\dontrun{
## equality constraints are honored (here: equal loadings across factors)
model.eq <- '
    ind60 =~ x1 + x2 + x3
    dem60 =~ y1 + a*y2 + b*y3 + c*y4
    dem65 =~ y5 + a*y6 + b*y7 + c*y8
    dem60 ~ ind60
    dem65 ~ ind60 + dem60
'
fit4 <- sem(model.eq, data = PoliticalDemocracy, estimator = "IV")
coef(fit4)

## multiple groups: metric measurement invariance (equal loadings)
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '
fit.metric <- sem(HS.model, data = HolzingerSwineford1939, estimator = "IV",
                  group = "school", group.equal = "loadings")
summary(fit.metric)
}
}
\keyword{models}
\keyword{regression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lav_export_estimation.R
\name{lav_export_estimation}
\alias{lav_export_estimation}
\title{lav_export_estimation}
\usage{
lav_export_estimation(lavaan_model)
}
\arguments{
\item{lavaan_model}{a fitted lavaan model}
}
\value{
List with:
\itemize{
  \item get_coef - When equality constraints are present, lavaan applies
  internal transformations. get_coef is a function that reconstructs the
  output of the coef function for the parameters.
  \item starting_values - starting_values to be used in the optimization
  \item objective_function - objective function, expecting the current parameter
  values and the lavaan model
  \item gradient_function - gradient function, expecting the current parameter
  values and the lavaan model
  \item lower - lower bounds for parameters
  \item upper - upper bound for parameters
}
}
\description{
lavaan provides a range of optimization methods with the optim.method argument
(nlminb, BFGS, L-BFGS-B, GN, and nlminb_constr). `lav_export_estimation` 
allows exporting objects and functions necessary to pass a lavaan model into
any optimizer that takes a combination of (1) starting values, (2) fit-function,
(3) gradient-function, and (4) upper and lower bounds. This allows testing new
optimization frameworks.
}
\examples{
library(lavaan)
model <- ' 
  # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + y2 + y3 + y4
     dem65 =~ y5 + a*y6 + y7 + y8

  # regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60
'

fit <- sem(model, 
           data = PoliticalDemocracy, 
           do.fit = FALSE)

est <- lav_export_estimation(lavaan_model = fit)

# The starting values are:
est$starting_values
# Note that these do not have labels (and may also differ from coef(fit)
# in case of equality constraints):
coef(fit)
# To get the same parameters, use:
est$get_coef(parameter_values = est$starting_values, 
             lavaan_model = fit)

# The objective function can be used to compute the fit at the current estimates:
est$objective_function(parameter_values = est$starting_values, 
                       lavaan_model = fit)

# The gradient function can be used to compute the gradients at the current estimates:
est$gradient_function(parameter_values = est$starting_values, 
                      lavaan_model = fit)

# Together, these elements provide the means to estimate the parameters with a large
# range of optimizers. For simplicity, here is an example using optim:
est_fit <- optim(par = est$starting_values, 
                 fn = est$objective_function,
                 gr = est$gradient_function,
                 lavaan_model = fit,
                 method = "BFGS")
est$get_coef(parameter_values = est_fit$par,
             lavaan_model = fit)

# This is identical to
coef(sem(model, 
         data = PoliticalDemocracy))

# Example using ridge regularization for parameter a
fn_ridge <- function(parameter_values, lavaan_model, est, lambda){
  return(est$objective_function(parameter_values = parameter_values, 
                                lavaan_model = lavaan_model) + lambda * parameter_values[6]^2)
}
ridge_fit <- optim(par = est$get_coef(est$starting_values,
                                      lavaan_model = fit), 
                   fn = fn_ridge, 
                   lavaan_model = fit,
                   est = est,
                   lambda = 10)
est$get_coef(parameter_values = ridge_fit$par,
             lavaan_model = fit)
}

\name{sam}
\alias{sam}
\alias{fsr}
\title{Fit Structural Equation Models using the SAM approach}
\description{
Fit a Structural Equation Model (SEM) using the Structural After Measurement
(SAM) approach.}
\usage{
sam(model = NULL, data = NULL, aux = NULL, cmd = "sem", se = "twostep",
    mm_list = NULL, mm_args = list(bounds = "wide.zerovar"), 
    struc_args = list(estimator = "ML"), 
    sam_method = "local", ..., 
    local_options = list(M.method = "ML", lambda.correction = TRUE, 
                         alpha.correction = 0L, twolevel.method = "h1"), 
    global_options = list(), 
    bootstrap = list(R = 1000L, type = "ordinary", show.progress = FALSE),
    output = "lavaan",
    bootstrap_args = bootstrap
)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See 
    \code{\link{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}
\item{data}{A data frame containing the observed variables used in
    the model.}
\item{aux}{Character vector. Names of auxiliary observed variables, forwarded
    to the underlying measurement and structural model fits. See the
    \code{\link{lavaan}} function for more details.}
\item{cmd}{Character. Which command is used to run the sem models. The possible
    choices are \code{"sem"}, \code{"cfa"} or \code{"lavaan"}, determining how
    we deal with default options.}
\item{se}{Character. The type of standard errors that are used in the
    final (structural) model. If \code{"twostep"} (the default), the
    standard errors take the estimation uncertainty of the first (measurement)
    stage into account. If \code{"standard"}, this uncertainty is ignored,
    and we treat the measurement information as known. If \code{"none"},
    no standard errors are computed. The experimental
    \code{"twostep.huber.white"} variant computes the two-step correction
    as a stacked (Murphy-Topel type) sandwich with casewise-score based
    (first-order) meat -- the two-step analogue of
    \code{se = "robust.huber.white"} in \code{\link{sem}}; it is currently
    available for continuous complete or \code{missing = "ml"} data
    (single- or multigroup, single level, no equality constraints), and
    falls back to \code{"twostep.robust"} or \code{"twostep"} otherwise. If the data are clustered (a
    \code{cluster=} argument is provided, without multilevel model syntax),
    \code{se = "twostep"} is automatically replaced by
    \code{se = "twostep.robust"}, and the standard errors (also for
    \code{se = "local"} and \code{se = "bootstrap"}, which resamples whole
    clusters) account for the clustering. For two-level models,
    \code{se = "local"} is available for the single-group continuous
    complete-data setting; it uses the cluster-sandwich covariance of the
    saturated (h1) estimates, and also enables the corrected structural
    test statistic. For \code{estimator = "PML"}, \code{se = "local"} is
    the only available two-step correction (single group, complete data,
    all-ordinal indicators): it combines the casewise influences of the
    saturated thresholds/polychoric correlations with the casewise
    pairwise-likelihood influences of the measurement-block estimates,
    and also enables the corrected structural test statistic.}
\item{mm_list}{List. Define the measurement blocks. 
     Each element of the list should
     be either a single name of a latent variable, or a vector of latent
     variable names.  If omitted, a separate measurement
     block is used for each latent variable.}
\item{mm_args}{List. Optional arguments for the fitting
    function(s) of the measurement block(s) only. See \code{\link{lavOptions}} 
    for a complete list.}
\item{struc_args}{List. Optional arguments for the fitting function of the
    structural part only. See \code{\link{lavOptions}} for a complete list.}
\item{sam_method}{Character. Can be set to \code{"local"}, \code{"global"}
    or \code{"fsr"}. In the latter case, the results are the same as if
    Bartlett factor scores were used, without any bias correction.}
\item{...}{Many more options can be specified, using 'name = value'.
    See \code{\link{lavOptions}} for a complete list. These options affect
    both the measurement blocks and the structural part.}
\item{local_options}{List. Options specific for local SAM method (these
     options may change over time). If \code{lambda.correction = TRUE},
     we ensure that the variance matrix of the latent variables (VETA)
     is positive definite. The \code{alpha.correction} option must be an
     integer. Acceptable values are in the range 0 to
     N-1. If zero (the default), no small sample correction is performed,
     and the bias-correction is the same as with local SAM. When equal to N-1,
     the bias-correction is eliminated, and the results are the same as
     naive FSR. Typical values are 0, P+1 (where P is the number of
     predictors in the structural model), P+5, and (N-1)/2.}
\item{global_options}{List. Options specific for global SAM method (not
     used for now).}
\item{bootstrap}{List. Only used when \code{se = "bootstrap"}. Typical
     elements of this list are \code{R}: the number of bootstrap samples, and
     \code{type}, which can be set to \code{"ordinary"} (the default) or
     \code{"parametric"}. A single number is also accepted (as in
     \code{sem()}) and is interpreted as \code{R}, the number of bootstrap
     draws.}
\item{output}{Character. If \code{"lavaan"}, a lavaan object is returned. If
    \code{"list"}, a list is returned with all the ingredients from the 
    different stages.}
\item{bootstrap_args}{List. Deprecated, use \code{bootstrap} instead.}
}
\details{
    The \code{sam} function automates the SAM approach by first
    estimating the measurement part of the model and then the structural
    part of the model. See the reference for more details.

    Note that in the current implementation, all indicators of latent
    variables must be observed. As a result, second-order factor structures
    are not yet supported.

    If \code{sam.method} is \code{"local"}, \code{"fsr"} or \code{"cfsr"},
    the structural part of the model is fitted to the estimated latent
    (co)variances (and means) from step 1. This step-2 structural model is
    stored in the \code{@internal} slot of the returned object, and can be
    retrieved by \code{lavInspect(fit, "sam.struc.fit.object")}. The
    functions \code{\link{fitMeasures}} (and
    \code{summary(fit, fit.measures = TRUE)}), \code{\link{lavResiduals}}
    (and \code{summary(fit, residuals = TRUE)}), \code{\link{modindices}}
    and \code{\link{lavTestLRT}} (when all compared models are local SAM
    models) operate on this structural model: the reported quantities
    describe the fit of the structural part only, conditional on the (fixed)
    measurement model of step 1. When available, the reported (scaled) test
    statistic and the robust fit measures are corrected for the estimation
    of the measurement model in step 1.

    If \code{sam.method} is \code{"global"}, the returned object is an
    ordinary (joint) lavaan object, and \code{fitMeasures} returns the usual
    fit measures of the joint model. Note, however, that the parameters were
    estimated using a two-step procedure; the behavior of fit measures in
    this setting has not been studied, and they should be interpreted with
    caution.
}
\section{Exogenous covariates and conditional.x}{
    Observed exogenous covariates may enter the structural part of the
    model. With continuous data, they are simply treated as (pseudo) latent
    variables (the default \code{conditional.x = FALSE}), but
    \code{conditional.x = TRUE} is supported as well (single-group,
    single-level models): the structural part is then fitted to the latent
    intercepts, the latent-on-covariate slopes, and the residual latent
    (co)variances, all conditional on the covariates.

    With categorical (ordered) data, \code{conditional.x = TRUE} is the
    lavaan default (when exogenous covariates are present), and this is also
    the recommended setting: the covariates are conditioned on (as in a
    probit regression), rather than being absorbed into the joint polychoric
    correlation matrix -- which would (wrongly) treat them as jointly normal,
    a real concern for binary/dummy covariates. In this setting, the
    measurement blocks of step 1 are estimated with
    \code{conditional.x = TRUE} as well (each block latent variable is
    regressed on all exogenous covariates, MIMIC style), so that the
    measurement parameters live on the same conditional latent-response
    scale (\code{Var(y*|x) = 1}, delta parameterization) as the joint model.
    An \emph{ordered endogenous} variable in the structural part is
    supported too (with or without covariates): it enters the structural
    model on its latent-response scale, while its thresholds are fixed at
    their sample-based values.

    When \code{conditional.x = TRUE} is combined with a local
    \code{sam.method}, the default \code{se = "twostep"} is automatically
    replaced by \code{se = "twostep.robust"}: the classic two-step formula is
    built on the joint information matrix, which does not describe the local
    step-2 estimator in this setting (it underestimates the sampling
    variability, especially for binary covariates). The robust variant is
    computed as a structural-space sandwich here, and coincides with
    \code{se = "local"}.

    Not (yet) available in combination with \code{conditional.x = TRUE}:
    multiple groups (for the local standard errors and the corrected test
    statistic), multilevel models, latent interactions, and the Yuan-Chan
    scaled global test statistic of \code{sam_method = "global"} (the
    standard, unscaled test is reported instead, with a warning).
}
\value{
  If \code{output = "lavaan"}, an object of class
  \code{\linkS4class{lavaan}}, for which several methods
  are available, including a \code{summary} method. If \code{output = "list"},
  a list.
}
\references{
Rosseel and Loh (2021). A structural-after-measurement approach to Structural
Equation Modeling. Psychological Methods. Advance online publication.
https://dx.doi.org/10.1037/met0000503
}
\seealso{
\code{\link{lavaan}}
}
\examples{
## The industrialization and Political Democracy Example 
## Bollen (1989), page 332
model <- ' 
  # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + a*y2 + b*y3 + c*y4
     dem65 =~ y5 + a*y6 + b*y7 + c*y8

  # regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
'

fit.sam <- sam(model, data = PoliticalDemocracy,
               mm.list = list(ind = "ind60", dem = c("dem60", "dem65")))
summary(fit.sam)
}

\name{standardizedSolution}
\alias{standardizedSolution}
\alias{standardizedsolution}
\title{Standardized Solution}
\description{
Standardized solution of a latent variable model.}
\usage{
standardizedSolution(object, type = "std.all", se = TRUE, zstat = TRUE,
                     pvalue = TRUE, ci = TRUE, level = 0.95,
                     boot_ci_type = "perc", cov_std = TRUE,
                     remove_eq = TRUE, remove_ineq = TRUE, remove_def = FALSE,
                     remove_aux = TRUE,
                     partable = NULL, glist = NULL, est = NULL,
                     output = "data.frame", ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{type}{If \code{"std.lv"}, the standardized estimates are based
on the variances of the (continuous) latent variables only.
If \code{"std.all"}, the standardized estimates are based
on the variances of both (continuous) observed and latent variables.
If \code{"std.nox"}, the standardized estimates are based
on the variances of both (continuous) observed and latent variables,
but not the variances of exogenous covariates. Note that \code{"std.nox"}
only differs from \code{"std.all"} if \code{fixed.x = TRUE}; if
\code{fixed.x = FALSE}, the exogenous covariates are treated as random
variables (and standardized) just like any other variable, and a warning
is issued.
Alternatively, \code{type} may be a vector of (observed) variable names
(for example \code{type = c("x1", "x2")}); in that case only the parameters
involving these variables are standardized (the other observed variables are
left unstandardized). This is a generalization of \code{"std.nox"}, where the
(observed) exogenous \code{x} variables are the ones left unstandardized.}
\item{se}{Logical. If TRUE, standard errors for the standardized parameters
will be computed, together with a z-statistic and a p-value.}
\item{zstat}{Logical. If \code{TRUE}, an extra column is added containing
the so-called z-statistic, which is simply the value of the estimate divided
by its standard error.}
\item{pvalue}{Logical. If \code{TRUE}, an extra column is added containing
the pvalues corresponding to the z-statistic, evaluated under a standard
normal distribution.}
\item{ci}{If \code{TRUE}, confidence intervals are added to
the output.}
\item{level}{The confidence level required.}
\item{boot_ci_type}{Character. Only used if the model was fitted with
\code{se = "bootstrap"}. In that case, bootstrap standard errors and bootstrap
confidence intervals are computed for the standardized parameters (by
re-standardizing each bootstrap draw), and this argument selects the type of
bootstrap confidence interval, exactly as in
\code{\link{parameterEstimates}}: \code{"norm"}, \code{"basic"},
\code{"perc"} (the default), \code{"bca.simple"} or \code{"bca"}.}
\item{cov_std}{Logical. If TRUE, the (residual) observed
covariances are scaled by the square root of the `Theta' diagonal elements, and
the (residual) latent covariances are scaled by the square root of the `Psi'
diagonal elements. If FALSE, the (residual) observed covariances
are scaled by the square root of the diagonal elements of the observed
model-implied covariance matrix (Sigma), and the (residual) latent covariances
are scaled by the square root of diagonal elements of the model-implied
covariance matrix of the latent variables.}
\item{remove_eq}{Logical. If TRUE, filter the output by removing all rows
containing equality constraints, if any.}
\item{remove_ineq}{Logical. If TRUE, filter the output by removing all
rows containing inequality constraints, if any.}
\item{remove_def}{Logical. If TRUE, filter the output by removing all
rows containing parameter definitions, if any.}
\item{remove_aux}{Logical. If TRUE (the default), filter the output by
removing all rows corresponding to auxiliary (\code{aux=}) variables added to
the model as saturated correlates (only relevant when \code{missing = "ml"}
and the \code{aux=} argument was used).}
\item{glist}{List of model matrices. If provided, they will be used
instead of the GLIST inside the object@Model slot. Only works if the
\code{est} argument is also provided. See Note.}
\item{est}{Numeric. Parameter values (as in the `est' column of a
parameter table). If provided, they will be used instead of
the parameters that can be extracted from object. Only works if the \code{glist}
argument is also provided. See Note.}
\item{partable}{A custom \code{list} or \code{data.frame} in which to store
the standardized parameter values. If provided, it will be used instead of
the parameter table inside the object@ParTable slot.}
\item{output}{Character. If \code{"data.frame"}, the parameter table is
displayed as a standard (albeit lavaan-formatted) data.frame.
If \code{"text"} (or alias \code{"pretty"}), the parameter table is
prettified and displayed with subsections (as used by the summary function).}
\item{...}{To support old argument names.}
}
\details{
The standardized estimates are functions of the (unstandardized) free
parameters and of (a subset of) the model-implied (co)variances used for
scaling. The standard errors reported by \code{standardizedSolution} are
therefore standard errors \emph{for the standardized parameters}, and they
will in general differ from the standard errors of the unstandardized
parameters reported by \code{\link{parameterEstimates}} (or in the
\code{summary()} output). They are also not the same as the standard errors
that would be obtained by simply rescaling the unstandardized standard errors.

How the standard errors are computed depends on how the model was originally
fitted (in particular on the \code{se=} argument of \code{\link{lavaan}},
\code{\link{cfa}}, \code{\link{sem}}, \dots):
\itemize{
  \item By default (\code{se = "standard"}, \code{"robust.sem"},
    \code{"robust.huber.white"}, \dots), the standard errors are obtained with
    the \emph{delta method}: the Jacobian of the standardization function is
    computed (numerically) and combined with the variance-covariance matrix of
    the (unstandardized) parameter estimates. Any robustness present in that
    variance-covariance matrix (for example robust or sandwich-type standard
    errors) is automatically propagated to the standardized solution.
  \item If the model was fitted with \code{se = "bootstrap"} (and the bootstrap
    draws are available), bootstrap standard errors and bootstrap confidence
    intervals are reported instead. These are obtained by re-standardizing each
    bootstrap draw and computing the standard deviation (for the standard
    error) and the requested interval type (see \code{boot_ci_type}) of the
    resulting standardized values. Note that, as in
    \code{\link{parameterEstimates}}, the p-value is still computed by
    referring the z-statistic (standardized estimate divided by its bootstrap
    standard error) to a standard normal distribution.
}

There is no separate argument to choose the \emph{type} of standard error
within \code{standardizedSolution}: the type always follows the \code{se=}
setting that was used when the model was fitted. To obtain, say, robust or
bootstrap standard errors for the standardized solution, refit the model with
the corresponding \code{se=} argument.
}
\note{
The \code{est}, \code{glist}, and \code{partable} arguments are not meant for
everyday users, but for authors of external R packages that depend on
\code{lavaan}. Only to be used with great caution.
}
\value{
  A data.frame containing standardized model parameters.

  If the model was fitted with \code{se = "bootstrap"} (and the bootstrap
  draws are available), the standardized estimates in the bootstrap samples
  are stored as the \code{"boot_std"} attribute of the returned data.frame (a
  matrix with one row per bootstrap draw and one column per row of the
  returned solution). These can be retrieved with
  \code{attr(x, "boot_std")}, analogous to
  \code{lavInspect(object, "coef.boot")} for the unstandardized solution.

  In addition, if \code{boot_ci_type = "bca"}, the empirical-influence design
  matrix used to compute the acceleration constant is stored as the
  \code{"design"} attribute (retrieve with \code{attr(x, "design")}).
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
standardizedSolution(fit)
}

\name{lav_efalist_summary}
\alias{lav_efalist_summary}
\alias{lav_efalist_summary_print}
\title{Summarizing EFA Fits}
\description{
S3 summary and print methods for class \code{efaList}.}
\usage{
lav_efalist_summary(object,
        nd = 3L, cutoff = 0.3, dot_cutoff = 0.1, alpha_level = 0.01,
        lambda = TRUE, theta = TRUE, psi = TRUE, fit_table = TRUE,
        fs_determinacy = FALSE, eigenvalues = TRUE, sumsq_table = TRUE, 
        lambda_structure = FALSE, se = FALSE, zstat = FALSE,
        pvalue = FALSE, \dots)
lav_efalist_summary_print(x, nd = 3L, cutoff = 0.3, dot_cutoff = 0.1,
               alpha_level = 0.01, \dots)
}
\arguments{
\item{object}{An object of class \code{efaList}, usually, a result of a call
    to \code{\link{efa}} with (the default) \code{output = "efa"}.}
\item{x}{An object of class \code{lav_efalist_summary}, usually, a result of a 
    call to \code{lav_efalist_summary}.}
\item{nd}{Integer. The number of digits that are printed after the decimal 
    point in the output.}
\item{cutoff}{Numeric. Factor loadings smaller than this value (in absolute
    value) are not printed
    (even if they are significantly different from zero). The idea is that
    only medium to large factor loadings are printed, to better see the
    overall structure.}
\item{dot_cutoff}{Numeric. Factor loadings larger (in absolute value) 
    than this value, but smaller (in absolute value) than the cutoff value 
    are shown as a dot. They represent small loadings that may still need your
    attention.}
\item{alpha_level}{Numeric. If the p-value of a factor loading is smaller
    than this value, a significance star is printed to the right of the 
    factor loading. To switch this off, use \code{alpha_level = 0}.}
\item{lambda}{Logical. If \code{TRUE}, include the (standardized) factor 
    loadings in the summary.}
\item{theta}{Logical. If \code{TRUE}, include the unique variances and
    the communalities in the table of factor loadings.}
\item{psi}{Logical. If \code{TRUE}, include the factor correlations in the 
    summary. Ignored if only a single factor is used.}
\item{fit_table}{Logical. If \code{TRUE}, show fit information for each
    model.}
\item{fs_determinacy}{Logical. If \code{TRUE}, show the factor score
    determinacy values per factor (assuming regression factor scores are used)
    and their squared values.}
\item{eigenvalues}{Logical. If \code{TRUE}, include the eigenvalues of the
    sample variance-covariance matrix in the summary.}
\item{sumsq_table}{Logical. If \code{TRUE}, include a table including sums
    of squares of factor loadings (and related measures) in the summary.
    The sums of squares are computed as the
    diagonal elements of Lambda times Psi (where Psi is the matrix of
    factor correlations).
    If orthogonal rotation was used, Psi is diagonal and the sums of
    squares are identical to the sums of the squared column elements of the 
    Lambda matrix (i.e., the factor loadings). This is no longer the case
    when oblique rotation has been used. But in both cases (orthogonal or
    oblique), the (total) sum of the sums of squares equals the sum of the
    communalities. In the second row of the table (Proportion of total),
    the sums of squares are divided by the total. In the third row of the 
    table (Proportion var), the sums of squares are divided by the 
    number of items.}
\item{lambda_structure}{Logical. If \code{TRUE}, show the structure matrix
    (i.e., the factor loadings multiplied by the factor correlations).}
\item{se}{Logical. If \code{TRUE}, include the standard errors of the 
    standardized lambda, theta and psi elements in the summary.}
\item{zstat}{Logical. If \code{TRUE}, include the Z-statistics of the 
    standardized lambda, theta and psi elements in the summary.} 
\item{pvalue}{Logical. If \code{TRUE}, include the P-values of the 
    standardized lambda, theta and psi elements in the summary.}
\item{...}{Further arguments passed to or from other methods.}
}
\value{
    The function \code{lav_efalist_summary} computes and returns a list of 
    summary statistics for the list of EFA models in \code{object}.
}
\examples{
## The famous Holzinger and Swineford (1939) example
fit <- efa(data = HolzingerSwineford1939, 
           ov_names = paste("x", 1:9, sep = ""),
           nfactors = 1:3,
           rotation = list("geomin", geomin_epsilon = 0.01, rstarts = 1))
summary(fit, nd = 3L, cutoff = 0.2, dot_cutoff = 0.05,
        lambda_structure = TRUE, pvalue = TRUE)
}

\name{open_parser}
\alias{open_parser}
\alias{lav_parse_options}
\title{The open parser configuration}
\description{
The lavaan model syntax parser can be extended with new operators, modifiers and groupings.
These extensions are only used in the 'open' parser.}
\usage{
lav_parse_options(
    pkgname   = NULL,
    operators = NULL,
    modifiers = NULL,
    groupings  = NULL
  )
}
\arguments{
\item{pkgname}{The name of the package that extends the elements.}
\item{operators}{A character vector with new operators. These operators must start and end with
a percentage sign and be at least 3 characters long.}
\item{modifiers}{A data.frame with columns \code{mod}, \code{side} and \code{expect}. The elements in \code{mod}
must be character strings in snake_case. The elements in \code{side} must be \code{l} or \code{r}, indicating
whether the modifier is expected on the left-hand side or on the right-hand side of the operator. The elements in
\code{expect} must be \code{char}, \code{num} or \code{raw}, specifying the type of value the modifier expects, where
\code{raw} means the value is not evaluated by the parser but returned as an uninterpreted string.}
\item{groupings}{A character vector with new groupings; they must be in snake_case.}
}
\value{
  A \code{list} with \code{data.frames} containing the currently active parser configuration.
}
\examples{
curconfig <- lav_parse_options("testPKG",
    operators = c('\%w\%', '\%q\%'),
    modifiers = data.frame(mod = c("linksc", "linksn", "rechtsn"),
                           side = c("l", "l", "r"),
                           expect = c("char", "num", "raw")),
    groupings = c("color", "colour")
)
for (j in seq_along(curconfig)) {
    cat(names(curconfig)[j], ":\\n")
    print(curconfig[[j]])
}
model <- '
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + a*y2 + start(0.3)*b*y3 + upper(0.99)*c*y4
     dem65 =~ y5 + a*y6 + b*y7 + rechtsn(2 * pi / 3)*c*y8
     dem60 \%w\% aa *0.2*y4

    dem60 ~ ind60
    dem65 ~ ind60 + dem60
    linksc("klopt") * dem60 =~ aaa * y8
    linksn(sqrt(23.5)) * dem60 =~ bbb * y9
    efa("efatje") * dem65 =~ xyz * x3

    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
'
print(result <- lavParseModelString(model, TRUE, parser = "open"))
print(attributes(result))

model2 <- '
  color: green
    level: 1
        fw =~ y1 + y2 + y3
        fw ~ x1 + x2 + x3
    level: 2
        fb =~ y1 + y2 + y3
        fb ~ w1 + w2
  color: blue
    level: 1
        fw =~ y1 + y2 + y3
        fw ~ x1 + x2 + x3
    level: 2
        fb =~ y1 + y2 + y3
        fb ~ w1 + w2
'
print(result2 <- lavParseModelString(model2, TRUE, parser = "open"))
print(attributes(result2))
}

\name{PoliticalDemocracy}
\alias{PoliticalDemocracy}
\docType{data}
\title{
Industrialization And Political Democracy Dataset
}
\description{
The `famous' Industrialization and Political Democracy dataset. This dataset is
used throughout Bollen's 1989 book (see pages 12, 17, 36 in chapter 2, pages
228 and following in chapter 7, pages 321 and following in chapter 8).
The dataset contains various measures of political democracy and
industrialization in developing countries.
}
\usage{data(PoliticalDemocracy)}
\format{
A data frame of 75 observations of 11 variables.
\describe{
    \item{\code{y1}}{Expert ratings of the freedom of the press in 1960}
    \item{\code{y2}}{The freedom of political opposition in 1960}
    \item{\code{y3}}{The fairness of elections in 1960}
    \item{\code{y4}}{The effectiveness of the elected legislature in 1960}
    \item{\code{y5}}{Expert ratings of the freedom of the press in 1965}
    \item{\code{y6}}{The freedom of political opposition in 1965}
    \item{\code{y7}}{The fairness of elections in 1965}
    \item{\code{y8}}{The effectiveness of the elected legislature in 1965}
    \item{\code{x1}}{The gross national product (GNP) per capita in 1960}
    \item{\code{x2}}{The inanimate energy consumption per capita in 1960}
    \item{\code{x3}}{The percentage of the labor force in industry in 1960}
  }

}
\source{
The dataset was originally retrieved from \verb{http://web.missouri.edu/~kolenikovs/Stat9370/democindus.txt} (link no longer valid; see discussion on SEMNET 18 Jun 2009). The dataset is part of a larger (public) dataset (ICPSR 2532), see
\verb{https://www.icpsr.umich.edu/web/ICPSR/studies/2532}.
}
\references{
Bollen, K. A. (1989). \emph{Structural Equations with Latent Variables.} 
Wiley Series in Probability and Mathematical Statistics. New York: Wiley.

Bollen, K. A. (1979). Political democracy and the timing of development.
\emph{American Sociological Review}, 44, 572-587.

Bollen, K. A. (1980). Issues in the comparative measurement of political
democracy. \emph{American Sociological Review}, 45, 370-390.
}
\examples{
head(PoliticalDemocracy)
}

\name{lav_label_code}
\alias{lav_label_code}
\title{Provides a formatted label for use in tikz, svg or Rplot}
\usage{
lav_label_code(label = "", value = "", show =FALSE,
               idx_font_size = 20L, dy = 7L,
               italic = TRUE, auto_subscript = TRUE)
}
\arguments{
\item{label}{A character string in one of the formats
\itemize{
\item \var{name}
\item \var{name_index}
\item \var{name=value}
\item \var{name_index=value}
}
If \var{value} is specified in parameter \code{label} and parameter
\code{value} is empty, the value in \code{label} will be used.}
\item{value}{A character string specifying a value or empty.}
\item{show}{A logical indicating that the result should be plotted in an Rplot.}
\item{idx_font_size}{An integer specifying font size to use for the
subscript in svg.}
\item{dy}{An integer specifying the distance to move the baseline of
the subscript in svg.}
\item{italic}{A logical indicating whether the label's font should be italic.
Only used for tikz output and in the displayed Rplot when the result is not an expression.}
\item{auto_subscript}{Logical, if TRUE and \code{label} starts with one or more
alphabetic chars followed by one or more digits, an underscore will be
inserted between these parts.}
}
\value{
a list with members \code{svg}, \code{tikz} and \code{r}, giving the result.
}
\description{
Creates the code used to show the label in tikz, svg or R. The label is
plotted in an xy-plot if show = TRUE.
}
\details{
If both \code{label} and \code{value} are empty, the resulting codes are
also empty and nothing will be shown.

If \code{label} is empty and \code{value} is not, processing is done as if
\code{label} was specified and \code{value} was empty.

If \code{label} contains the string "1van", the label value is set to "1".
This allows distinct names for regression intercepts while still labeling
them as "1".

If \var{name} in \code{label} is a Greek character or \code{varepsilon},
the function attempts to generate code that shows the Greek symbol. If
\code{label} contains an \var{index} part, the function attempts to generate
code that displays this value as a subscript. In the svg code, the values
\code{idx_font_size} and \code{dy} are used for the subscript. If a
\var{value} is present, the function attempts to show this value after the
label, with an equal sign between the two.
}
\examples{
lav_label_code("x3")
lav_label_code("beta10", 0.65, show = TRUE)
lav_label_code("A_i,j=0.45", show = TRUE)
lav_label_code("Gamma")
lav_label_code(value="1.2345")
}

\name{lavTestWald}
\alias{lavTestWald}
\alias{lavtestwald}
\alias{wald}
\alias{Wald}
\alias{lavWaldTest}
\title{Wald test}
\description{
Wald test for testing a linear hypothesis about the parameters of
a fitted lavaan object.}
\usage{
lavTestWald(object, constraints = NULL, verbose = FALSE)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{constraints}{A character string (typically between single quotes) 
  containing one or more equality constraints. See examples for more details.}
\item{verbose}{Logical. If \code{TRUE}, print out the restriction matrix
  and the estimated restricted values.}
}
\details{
  The constraints are specified using the \code{"=="} operator. Both
  the left-hand side and the right-hand side of the equality can contain
  a linear combination of model parameters, or a constant (like zero).
  The model parameters must be specified by their user-specified labels.
  Names of defined parameters (using the \code{":="} operator) can be
  included too.

  The test statistic is based on the variance-covariance matrix of the
  estimated parameters (\code{vcov}) of the fitted object. Therefore, if
  robust standard errors were requested for the fitted object (e.g.,
  \code{se = "robust.sem"} or \code{se = "robust.huber.white"}), the main
  test statistic is already the generalized (\sQuote{robust}) Wald test
  described in Satorra (2000). In that case (or when a scaled test
  statistic was requested, so that the ingredients for a sandwich-type
  covariance matrix are available), additional versions of the Wald
  test statistic are computed as well: the standard (normal-theory)
  statistic (\code{stat.standard}), a mean-scaled statistic
  (\code{stat.scaled}), a mean-and-variance adjusted statistic with
  fractional degrees of freedom (\code{stat.adjusted}), and the
  generalized statistic (\code{stat.robust}); see Satorra (2000).
}
\value{
  A list containing at least three elements: the Wald test statistic
  (\code{stat}), the degrees of freedom (\code{df}), and a p-value under
  the chi-square distribution (\code{p.value}). If robust versions could
  be computed, the list also contains the standard, scaled, adjusted and
  generalized versions of the test statistic, together with their
  p-values (see Details).
}
\references{
Satorra, A. (2000). Scaled and adjusted restricted tests in multi-sample
analysis of moment structures. In Heijmans, R.D.H., Pollock, D.S.G. &
Satorra, A. (Eds.), Innovations in multivariate statistical analysis:
A festschrift for Heinz Neudecker (pp. 233-247). London: Kluwer Academic
Publishers.
}
\examples{
HS.model <- '
    visual  =~ x1 + b1*x2 + x3
    textual =~ x4 + b2*x5 + x6
    speed   =~ x7 + b3*x8 + x9
'

fit <- cfa(HS.model, data=HolzingerSwineford1939)

# test 1: test about a single parameter
# this is the 'chi-square' version of the 
# z-test from the summary() output
lavTestWald(fit, constraints = "b1 == 0")

# test 2: several constraints
con = '
   2*b1 == b3
   b2 - b3 == 0
'
lavTestWald(fit, constraints = con)
}

\name{lav_plotinfo_tikzcode}
\alias{lav_plotinfo_tikzcode}
\title{Creates code to show a diagram in tikz format}
\usage{
lav_plotinfo_tikzcode(plotinfo,
                      outfile = "",
                      cex = 1.3,
                      sloped_labels = TRUE,
                      standalone = FALSE,
                      mlovcolors = c("lightgreen", "lightblue"),
                      lightness = 1,
                      italic = TRUE,
                      auto_subscript = TRUE
                     )
}
\arguments{
\item{plotinfo}{A plotinfo structure as returned from \code{lav_plotinfo_positions}.}
\item{outfile}{A connection or a character string, the file to store the code.}
\item{cex}{Minimum distance between nodes in cm.}
\item{sloped_labels}{logical, sloped labels for the edges.}
\item{standalone}{logical, should code be added to make the TeX file standalone?}
\item{mlovcolors}{Array of two colors for distinguishing ov nodes with the same
name across the levels of a multilevel model.}
\item{lightness}{A scalar factor to modify the distances between nodes.}
\item{italic}{Should labels be in an italic font? If FALSE, the labels are
shown in mathrm font.}
\item{auto_subscript}{Logical, see \code{\link{lav_label_code}}.}
}
\value{
NULL, invisible
}
\description{
Creates the code to make a diagram in tikz.
}
\seealso{
\code{\link{lav_plotinfo_positions}}
}
\examples{
model <- 'alpha  =~ x1 + x2 + x3       # latent variable
          beta <~ x4 + x5 + x6         # composite
          gamma =~ x7 + x8 + x9        # latent variable
          Xi =~ x10 + x11 + x12 + x13 # latent variable
          # regressions
          Xi ~ v * alpha + t * beta + 1
          alpha ~ yy * Theta1 + tt * beta + ss * gamma
          '
test <- lav_model_plotinfo(model)
test <- lav_plotinfo_positions(test)
lav_plotinfo_tikzcode(test) # no file given, so output to R console
modelml <- '
        level: 1
        fw =~ 1*y_1 + y_2 + y_3 + y_5
        level: 2
        fb =~ 1*y_1 + y_2 + y_3 + y_4
        y_2 ~~ cv24 * y_4
  '
testml <- lav_model_plotinfo(modelml)
testml <- lav_plotinfo_positions(testml)
# in the line hereunder no file is given, so output to R console
lav_plotinfo_tikzcode(testml, cex = 1.4, sloped_labels = FALSE,
        standalone = TRUE, auto_subscript = FALSE)
\dontrun{
# example creating tex file with the above diagrams
zz <- file("testtikz.tex", open="w")
writeLines(c(
  '\\documentclass{article}',
  '\\usepackage{amsmath, amssymb}',
  '\\usepackage{amsfonts}',
  '\\usepackage[utf8]{inputenc}',
  '\\usepackage[english]{babel}',
  '\\usepackage{xcolor}',
  '\\usepackage{color}',
  '\\usepackage{tikz}',
  '\\usetikzlibrary {shapes.geometric}',
  '\\begin{document}'),
  zz)
lav_plotinfo_tikzcode(test, outfile = "temp.tex")
tmp <- readLines("tmp.tex")
writeLines(tmp, zz)
writeLines(" ", zz)
lav_plotinfo_tikzcode(testml, outfile = "temp.tex", cex = 1.4,
              sloped_labels = FALSE, auto_subscript = FALSE)
tmp <- readLines("tmp.tex")
writeLines(tmp, zz)
writeLines("\\end{document}", zz)
close(zz)
openPDF <- function(f) {
  os <- .Platform$OS.type
  if (os=="windows")
    shell.exec(normalizePath(f))
  else {
    pdf <- getOption("pdfviewer", default='')
    if (nchar(pdf)==0)
      stop("The 'pdfviewer' option is not set. Use options(pdfviewer=...)")
    system2(pdf, args=c(f))
  }
}
tools::texi2dvi("testtikz.tex", pdf = TRUE, clean = TRUE)
openPDF("testtikz.pdf")
}
}

\name{lavMatrixRepresentation}
\alias{lavMatrixRepresentation}
\title{lavaan matrix representation}
\description{
Extend the parameter table with a matrix representation.
}
\usage{
lavMatrixRepresentation(partable, representation = "LISREL",
                        allow_composites = TRUE,
                        add_attributes = FALSE, as_data_frame = TRUE, ...)
}
\arguments{
\item{partable}{A lavaan parameter table (as extracted by the
\code{\link{parTable}} function, or generated by the
\code{\link{lavParTable}} function).}
\item{representation}{Character. The matrix representation style.
Currently, only the all-y version of the LISREL representation
is supported.}
\item{allow_composites}{Logical. If \code{TRUE}, and the model syntax
contains the "<~" operator, handle the lhs as proper composites.
If \code{FALSE}, use the old (<0.6-20) approach.}
\item{add_attributes}{Logical. If \code{TRUE}, additional information
about the model matrix representation is added as attributes.}
\item{as_data_frame}{Logical. If \code{TRUE}, the extended parameter table is
returned as a data.frame.}
\item{...}{Used only to support old argument names.}
}
\value{
A list or a data.frame containing the original parameter table, plus
three columns: a \code{"mat"} column containing matrix names, and
a \code{"row"} and \code{"col"} column for the row and column indices
of the model parameters in the model matrices.
}
\seealso{\code{\link{lavParTable}}, \code{\link{parTable}}}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)

# extract partable
partable <- parTable(fit)

# add matrix representation (and show only a few columns)
lavMatrixRepresentation(partable)[,c("lhs","op","rhs","mat","row","col")]
}

\name{fitMeasures}
\alias{fitMeasures}
\alias{fitmeasures}
\alias{fitMeasures,lavaan-method}
\alias{fitmeasures,lavaan-method}
\alias{fitindices}
\title{Fit Measures for a Latent Variable Model}
\description{
This function computes a variety of fit measures to assess the global
fit of a latent variable model.}
\usage{
fitMeasures(object, fit_measures = "all",
            baseline_model = NULL, h1_model = NULL,
            fm_args = list(standard.test     = "default",
                           scaled.test       = "default",
                           rmsea.ci.level    = 0.90,
                           rmsea.close.h0    = 0.05,
                           rmsea.notclose.h0 = 0.08,
                           robust            = TRUE,
                           cat.nonpd         = "na"),
            output = "vector", level = NULL, ...)
fitmeasures(object, fit_measures = "all",
            baseline_model = NULL, h1_model = NULL,
            fm_args = list(standard.test     = "default",
                           scaled.test       = "default",
                           rmsea.ci.level    = 0.90,
                           rmsea.close.h0    = 0.05,
                           rmsea.notclose.h0 = 0.08,
                           robust            = TRUE,
                           cat.nonpd         = "na"),
            output = "vector", level = NULL, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{fit_measures}{If \code{"all"}, all fit measures available will be
  returned. If only a single or a few fit measures are specified by name,
  only those are computed and returned.
  If additional options for certain fit measures are needed, fit_measures
  should be a list whose \code{fit.measures} (or \code{fit_measures}) element
  holds the measures
  wanted, with the options supplied as further named elements of the list. The
  \code{standard.test} option determines the main test statistic (chi-square
  value) that will be used to compute all the fit measures that depend on this
  test statistic. Usually this is \code{"standard"}. FMG p-value tests can be
  selected here using names such as \code{"peba4"}, \code{"pols3"},
  \code{"pall"}, or \code{"all"}. The \code{scaled.test}
  option determines which scaling method is to be used for the scaled fit
  measures (in case multiple scaling methods were requested). The
  \code{rmsea.ci.level} option, default 0.9, determines the level of the confidence
  interval for the rmsea value. The \code{rmsea.close.h0} option, default 0.05,
  is the rmsea value that is used under the null hypothesis that \code{rmsea <= rmsea.close.h0}.
  The \code{rmsea.notclose.h0} option, default 0.08, is the rmsea value that is used under the
  null hypothesis that \code{rmsea >= rmsea.notclose.h0}. The
  \code{gfi.ci.level} option, default 0.9, determines the level of the
  confidence interval for the (new) goodness of fit index \code{gfi}. The
  \code{gfi} index (Maydeu-Olivares et al., 2024) is computed from the
  reweighted least squares (RLS) statistic; the version based on the
  (standard) likelihood ratio statistic is also available, and the two can
  be requested by name as \code{gfi_lrt} and \code{gfi_rls} (the latter being
  identical to \code{gfi}), together with their \code{.ci.lower},
  \code{.ci.upper} and \code{.robust} variants. The
  \code{robust}
  option, default \code{TRUE}, can be set to \code{FALSE} to avoid computing the so-called robust
  rmsea/cfi/gfi measures (for example if the computations take too long).
  The \code{cat.nonpd} option, default \code{"na"}, is only used when data is
  categorical, and determines what happens to the robust values of RMSEA, CFI
  and GFI when an input (polychoric/tetrachoric) correlation matrix is not
  positive-definite (for at least one group): if \code{"na"}, the robust
  values are \code{NA}; if \code{"refit"}, the input correlation matrix is
  smoothed to be positive-definite, and the model parameters are re-estimated
  using \code{estimator = "catML"}; if \code{"smooth"}, the input correlation
  matrix is smoothed, but the original (DWLS) parameter estimates are retained
  (this was the behavior in lavaan 0.6-13 and earlier). The older
  \code{cat.check.pd = FALSE} option is a deprecated alias for
  \code{cat.nonpd = "refit"}.}
\item{baseline_model}{If not NULL, an object of class
  \code{\linkS4class{lavaan}}, representing a user-specified baseline model.
  If a baseline model is provided, all fit indices relying on a
  baseline model (e.g., CFI or TLI) will use the test statistics from
  this user-specified baseline model, instead of the default baseline model.}
\item{h1_model}{If not NULL, an object of class \code{\linkS4class{lavaan}},
  representing a user-specified alternative to the default unrestricted model.
  If \code{h1_model} is provided, all fit indices calculated from chi-squared
  will use the chi-squared \emph{difference} test statistics from 
  \code{\link{lavTestLRT}}, which compare the user-provided \code{h1_model} 
  to \code{object}.}
\item{fm_args}{List. Additional options for certain fit measures. DEPRECATED.
  The options should now be specified in the \code{"fit_measures"} argument!}
\item{output}{Character. If \code{"vector"} (the default), display the
  output as a named (lavaan-formatted) vector. If \code{"matrix"}, display
  the output as a 1-column matrix. If \code{"text"}, display the
  output using subsections and verbose descriptions. The latter is used
  in the summary output, and does not print the chi-square test by default.
  In addition, \code{fit_measures} should contain the main ingredient
  (for example \code{"rmsea"}) if related fit measures are requested
  (for example \code{"rmsea.ci.lower"}). Otherwise, nothing will be 
  printed in that section.
  See the examples for how to add the chi-square test in the text output.}
\item{level}{Only used for multilevel models. If not \code{NULL}, either
  a level number (\code{1} or \code{2}) or a level name (\code{"within"} or
  \code{"between"}, or the level label used in the model syntax). The
  returned fit measures then reflect the fit of the model at that level
  only, while the model at the other level is saturated. This uses the
  so-called partially saturated model approach (Ryu & West, 2009): any
  misfit of the partially saturated model can be attributed to the target
  level. The incremental fit indices (e.g., CFI and TLI) are computed with
  respect to a level-specific baseline model: the independence model at
  the target level, again combined with a saturated model at the other
  level. By default, these partially saturated models are fitted (and
  stored) along with the original model; see the \code{fit.by.level}
  option in \code{\link{lavOptions}}. If they were not stored, they are
  re-fitted on the fly.}
\item{...}{Further arguments passed to or from other methods. Not currently
  used for \code{lavaan} objects.}
}
\value{
  A named numeric vector of fit measures.
}
\details{
When a scaled (or robust) test statistic is requested (for example, by using
\code{test = "satorra.bentler"}), the function will also return fit indices
based on the scaled chi-square statistic, rather than the standard version.
These scaled versions of fit measures, such as CFI and RMSEA, are calculated in
the same way as their standard counterparts, with the key difference being that
the scaled chi-square statistic is used in place of the regular one. In the
output of \code{fitMeasures()}, these appear with the \code{.scaled} suffix, 
or in the \code{Scaled} column of the \code{summary()} output.

However, this substitution-based approach---used in SEM software for many
years---has since been shown to be incorrect. Improved versions of robust fit
indices have been proposed, offering better theoretical properties. Although
still under development and not yet implemented for all estimation settings,
these improved robust fit measures are provided when available. They appear
with a \code{.robust} suffix in the output of \code{fitMeasures()}, or in the
\code{Scaled} column of the \code{summary()} output on a row labeled
\code{Robust}. As a general recommendation, these newer robust versions should
be used whenever available, in preference to the older scaled ones. See the
references below for more details.

It is also worth noting that, for models involving ordered categorical data,
robust fit indices are only computed if the underlying matrix of tetrachoric or
polychoric correlations is positive definite. If this condition is not
met---which is not uncommon in small samples---the robust measures are reported 
as \code{NA}.

Finally, in some situations (especially when the data contains missing values),
computing these robust fit indices may be computationally intensive. To avoid
long runtimes, the calculation of robust fit measures can be disabled by
setting the \code{robust} argument to \code{FALSE} in the \code{fit_measures} list.

FMG p-values can be selected through the existing \code{standard.test}
mechanism, for example
\code{fitMeasures(fit, "pvalue", fm_args = list(standard.test = "peba4"))}.
No separate FMG-specific fit-measure names are needed.

For models fitted with \code{\link{sam}} using a local \code{sam.method}
(\code{"local"}, \code{"fsr"} or \code{"cfsr"}), the fit measures are
computed for the step-2 structural model only, conditional on the (fixed)
measurement model of step 1; loglikelihood-based measures (aic, bic, ...)
are not available. If \code{sam.method = "global"}, the fit measures are
based on the joint model; since the parameters were estimated using a
two-step procedure, they should be interpreted with caution.
}
\references{
Ryu, E., & West, S. G. (2009). Level-specific evaluation of model fit in
multilevel structural equation modeling. Structural Equation Modeling, 16(4),
583--601. \doi{10.1080/10705510903203466}

Brosseau-Liard, P. E., Savalei, V., & Li, L. (2012). An investigation of the
sample performance of two nonnormality corrections for RMSEA. Multivariate
behavioral research, 47(6), 904-930.
\doi{10.1080/00273171.2012.715252}

Brosseau-Liard, P. E., & Savalei, V. (2014). Adjusting incremental fit
indices for nonnormality. Multivariate behavioral research, 49(5), 460-470.
\doi{10.1080/00273171.2014.933697}

Savalei, V. (2018). On the computation of the RMSEA and CFI from the
mean-and-variance corrected test statistic with nonnormal data in SEM.
Multivariate behavioral research, 53(3), 419-429.
\doi{10.1080/00273171.2018.1455142}

Savalei, V. (2021). Improving fit indices in structural equation modeling with
categorical data. Multivariate Behavioral Research, 56(3), 390-407.
\doi{10.1080/00273171.2020.1717922}

Savalei, V., Brace, J. C., & Fouladi, R. T. (2023). We need to change how we
compute RMSEA for nested model comparisons in structural equation modeling.
Psychological Methods.
\doi{10.1037/met0000537}

Zhang, X., & Savalei, V. (2023). New computations for RMSEA and CFI following
FIML and TS estimation with missing data. Psychological Methods, 28(2),
263-283. \doi{10.1037/met0000445}

Foldnes, N., Moss, J., & Gronneberg, S. (2024). Improved goodness of fit
procedures for structural equation models. \emph{Structural Equation Modeling:
A Multidisciplinary Journal}, 1-13. \doi{10.1080/10705511.2024.2372028}

Foldnes, N., Gronneberg, S., & Moss, J. (2026). Penalized eigenvalue block
averaging: Extension to nested model comparison and Monte Carlo evaluations.
\emph{Behavior Research Methods, 58}(4). \doi{10.3758/s13428-026-02968-4}
}

\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data = HolzingerSwineford1939)
fitMeasures(fit)
fitMeasures(fit, "cfi")
fitMeasures(fit, c("chisq", "df", "pvalue", "cfi", "rmsea"))
fitMeasures(fit, c("chisq", "df", "pvalue", "cfi", "rmsea"), 
            output = "matrix")
fitMeasures(fit, c("chisq", "df", "pvalue", "cfi", "rmsea"),
            output = "text")
fitMeasures(fit, "pvalue", fm_args = list(standard.test = "peba4"))
## specify another threshold for RMSEA confidence interval
fitMeasures(fit, list(
  fit.measures = c("cfi", "rmsea"),
  rmsea.ci.level = 0.95))

## fit a more restricted model
fit0 <- cfa(HS.model, data = HolzingerSwineford1939, orthogonal = TRUE)
## Calculate RMSEA_D (Savalei et al., 2023)
## See https://psycnet.apa.org/doi/10.1037/met0000537
fitMeasures(fit0, "rmsea", h1_model = fit)

## level-specific fit measures for a multilevel model (Ryu & West, 2009)
model <- '
    level: 1
        fw =~ y1 + y2 + y3
        fw ~ x1 + x2 + x3
    level: 2
        fb =~ y1 + y2 + y3
        fb ~ w1 + w2
'
fit2l <- sem(model, data = Demo.twolevel, cluster = "cluster")
fitMeasures(fit2l, c("chisq", "df", "pvalue", "cfi", "rmsea",
                     "srmr_within"), level = "within")
fitMeasures(fit2l, c("chisq", "df", "pvalue", "cfi", "rmsea",
                     "srmr_between"), level = "between")
}

\name{Demo.growth}
\alias{Demo.growth}
\docType{data}
\title{
Demo dataset for illustrating a linear growth model
}
\description{
A toy dataset containing measures on 4 time points (t1,t2, t3 and t4),
two predictors (x1 and x2) influencing the random intercept and slope, and 
a time-varying covariate (c1, c2, c3 and c4).
}
\usage{data(Demo.growth)}
\format{
A data frame of 400 observations of 10 variables.
\describe{
    \item{\code{t1}}{Measured value at time point 1}
    \item{\code{t2}}{Measured value at time point 2}
    \item{\code{t3}}{Measured value at time point 3}
    \item{\code{t4}}{Measured value at time point 4}
    \item{\code{x1}}{Predictor 1 influencing intercept and slope}
    \item{\code{x2}}{Predictor 2 influencing intercept and slope}
    \item{\code{c1}}{Time-varying covariate time point 1}
    \item{\code{c2}}{Time-varying covariate time point 2}
    \item{\code{c3}}{Time-varying covariate time point 3}
    \item{\code{c4}}{Time-varying covariate time point 4}
  }

}
\seealso{
\code{\link{growth}}
}
\examples{
head(Demo.growth)
}

\name{lavCor}
\alias{lavCor}
\title{Polychoric, polyserial and Pearson correlations}
\description{
Fit an unrestricted model to compute polychoric, polyserial and/or Pearson
correlations.}
\usage{
lavCor(object, ordered = NULL, group = NULL, missing = "listwise",
       ov_names_x = NULL, sampling_weights = NULL,
       se = "none", test = "none",
       estimator = "two.step", baseline = FALSE, ...,
       cor_smooth = FALSE, cor_smooth_tol = 1e-04, output = "cor")
}
\arguments{
\item{object}{Either a \code{data.frame}, or an object of class
\code{\linkS4class{lavaan}}. If the input is a \code{data.frame}, and
    some variables are declared as ordered factors, lavaan will
    treat them as ordinal variables.}
\item{ordered}{Character vector. Only used if \code{object} is a
    \code{data.frame}. Treat these variables as ordered (ordinal) variables.
    Importantly, all other variables will be treated as numeric (unless they
    are declared as ordered in the original data frame.)}
\item{group}{Only used if \code{object} is a \code{data.frame}. Specify
    a grouping variable.}
\item{missing}{If \code{"listwise"}, cases with missing values are removed
    listwise from the data frame. If \code{"direct"} or
    \code{"ml"} or \code{"fiml"} and the estimator is maximum likelihood,
    an EM algorithm is used to estimate the unrestricted covariance matrix
    (and mean vector). If \code{"pairwise"}, pairwise deletion is used. If
    \code{"default"}, the value is set depending on the estimator and the
    mimic option.}
\item{sampling_weights}{Only used if \code{object} is a \code{data.frame}.
    Specify a variable containing sampling weights.}
\item{ov_names_x}{Only used if \code{object} is a \code{data.frame}. Specify
    variables that need to be treated as exogenous. Only used if at least
    one variable is declared as ordered.}
\item{se}{Only used if \code{output} (see below) contains standard errors.
    See \code{\link{lavOptions}} for possible options.}
\item{test}{Only used if output is \code{"fit"} or \code{"lavaan"}. See
    \code{\link{lavOptions}} for possible options.}
\item{estimator}{If \code{"none"} or \code{"two.step"} or \code{"two.stage"},
   only starting values are computed for the correlations (and thresholds),
   without any further estimation. If all variables are continuous, the
   starting values are the sample covariances (converted to correlations
   if \code{output = "cor"}). If at least one variable is ordered,
   the thresholds are computed using univariate information only. The
   polychoric and/or polyserial correlations are computed in a second
   stage, keeping the values of the thresholds constant. If an estimator
   (other than \code{"two.step"} or \code{"two.stage"}) is specified
   (for example \code{estimator = "PML"}), these starting values are
   further updated by fitting the unrestricted model using the chosen
   estimator. See the \code{\link{lavaan}} function for alternative
   estimators.}
\item{baseline}{Only used if output is \code{"fit"} or \code{"lavaan"}. If
   \code{TRUE}, a baseline model is also estimated. Note that the \code{test}
   argument should also be set to a value other than \code{"none"}.}
\item{...}{Optional parameters that are passed to the \code{\link{lavaan}}
    function.}
\item{cor_smooth}{Logical. Only used if \code{output = "cor"}.
   If \code{TRUE}, ensure the resulting correlation matrix is positive
   definite. The following simple method is used: an eigenvalue decomposition
   is computed; then, eigenvalues smaller than \code{cor.smooth.tol}
   are set to be equal to \code{cor.smooth.tol}, before the  matrix
   is again reconstructed. Finally, the matrix (which may no longer have unit
   diagonal elements) is converted to a correlation matrix using
   \code{cov2cor}.}
\item{cor_smooth_tol}{Numeric. Smallest eigenvalue used when reconstructing
   the correlation matrix after an eigenvalue decomposition.}
\item{output}{If \code{"cor"}, the function returns the correlation matrix only.   If \code{"cov"}, the function returns the covariance matrix (this only makes
   a difference if at least one variable is numeric). If \code{"th"} or
   \code{"thresholds"}, only the thresholds are returned. If \code{"sampstat"},
   the output equals the result of \code{lavInspect(fit, "sampstat")} where
   fit is the unrestricted model. If \code{"est"} or \code{"pe"} or
   \code{"parameterEstimates"}, the output equals the result of
   \code{lavParameterEstimates(fit)}. Finally, if output is \code{"fit"} or
   \code{"lavaan"}, the function returns an object of class
   \code{\linkS4class{lavaan}}.}
}
\details{
    This function is a wrapper around the \code{\link{lavaan}} function,
    but where the model is defined as the unrestricted model. The
    following free parameters are included: all covariances/correlations among
    the variables, variances for continuous variables, means for continuous
    variables, thresholds for ordered variables, and if exogenous variables
    are included (\code{ov_names_x} is not empty) while some variables
    are ordered, also the regression slopes enter the model.
}
\value{
    By default, if \code{output = "cor"} or \code{output = "cov"}, a symmetric
    matrix (of class \code{"lavaan.matrix.symmetric"}, which only affects the
    way the matrix is printed). If \code{output = "th"}, a named vector of
    thresholds. If \code{output = "fit"} or \code{output = "lavaan"},
    an object of class  \code{\linkS4class{lavaan}}.
}
\references{
Olsson, U. (1979). Maximum likelihood estimation of the polychoric correlation
coefficient. Psychometrika, 44(4), 443-460.

Olsson, U., Drasgow, F., & Dorans, N. J. (1982). The polyserial correlation coefficient. Psychometrika, 47(3), 337-347.
}
\seealso{
\code{\link{lavaan}}
}
\examples{
# Holzinger and Swineford (1939) example
HS9 <- HolzingerSwineford1939[,c("x1","x2","x3","x4","x5",
                                 "x6","x7","x8","x9")]

# Pearson correlations
lavCor(HS9)

# ordinal version, with three categories
HS9ord <- as.data.frame( lapply(HS9, cut, 3, labels = FALSE) )

# polychoric correlations, two-stage estimation
lavCor(HS9ord, ordered=names(HS9ord))

# thresholds only
lavCor(HS9ord, ordered=names(HS9ord), output = "th")

# polychoric correlations, with standard errors
lavCor(HS9ord, ordered=names(HS9ord), se = "standard", output = "est")

# polychoric correlations, full output
fit.un <- lavCor(HS9ord, ordered=names(HS9ord), se = "standard",
                 output = "fit")
summary(fit.un)
}

\name{lavaanList}
\alias{lavaanList}
\alias{semList}
\alias{cfaList}
\title{Fit List of Latent Variable Models}
\description{
Fit the same latent variable model to a (potentially large) number of datasets.}
\usage{
lavaanList(model = NULL, data_list = NULL, data_function = NULL, 
    data_function_args = list(), ndat = length(data_list), cmd = "lavaan", 
    ..., store_slots = c("partable"), fun = NULL, show_progress = FALSE,
    store_failed = FALSE, parallel = c("no", "multicore", "snow"), 
    ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
    cl = NULL, iseed = NULL)

semList(model = NULL, data_list = NULL, data_function = NULL,
    data_function_args = list(), ndat = length(data_list),
    ..., store_slots = c("partable"), fun = NULL, show_progress = FALSE,
    store_failed = FALSE, parallel = c("no", "multicore", "snow"), 
    ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
    cl = NULL, iseed = NULL)

cfaList(model = NULL, data_list = NULL, data_function = NULL,
    data_function_args = list(), ndat = length(data_list),
    ..., store_slots = c("partable"), fun = NULL, show_progress = FALSE,
    store_failed = FALSE, parallel = c("no", "multicore", "snow"), 
    ncpus = min(2L, max(1L, parallel::detectCores() - 1L, na.rm = TRUE)),
    cl = NULL, iseed = NULL)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See 
    \code{\link{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}
\item{data_list}{List. Each element contains a full data frame containing
      the observed variables used in the model.}
\item{data_function}{Function. A function that generates a full data frame
      containing the observed variables used in the model. It can also be
      a matrix, if the columns are named.}
\item{data_function_args}{List. Optional list of arguments that are passed
    to the \code{data_function} function.}
\item{ndat}{Integer. The number of datasets that should be generated using
    the \code{data_function} function.}
\item{cmd}{Character. The command used to run the SEM models. The possible
    choices are \code{"sem"}, \code{"cfa"} or \code{"lavaan"}, which determine
    how the default options are handled.}
\item{\dots}{Other named arguments for the \code{lavaan} function.}
\item{store_slots}{Character vector. Which slots (from a lavaan object) 
    should be stored for each dataset? The possible choices are
    \code{"timing"}, \code{"partable"}, \code{"data"}, \code{"samplestats"},
    \code{"vcov"}, \code{"test"}, \code{"optim"}, \code{"h1"}, \code{"loglik"},
    or \code{"implied"}. Finally, \code{"all"} selects all slots.}
\item{fun}{Function. A function which when applied to the
    \code{\linkS4class{lavaan}} object returns the information of interest.}
\item{store_failed}{Logical. If \code{TRUE}, write (to tempdir()) the dataset
     and (if available) the fitted object when the estimation for a particular
     dataset failed for some reason. This allows post hoc inspection of the problem.}
\item{parallel}{The type of parallel operation to be used (if any). If
missing, the default is \code{"no"}.}
\item{ncpus}{Integer. The number of processes to be used in parallel operation:
By default, this is 2 (following CRAN policy). You may set this to a higher
value (for example, the number of available CPUs) if you wish to use more
processes.}
\item{cl}{An optional \pkg{parallel} or \pkg{snow} cluster for use if
\code{parallel = "snow"}.  If not supplied, a cluster on the local machine is
created for the duration of the \code{lavaanList} call.}
\item{iseed}{An integer to set the seed. Or NULL if no reproducible seeds are
  needed. To make this work, make sure the first
  RNGkind() element is \code{"L'Ecuyer-CMRG"}. You can check this by
  typing \code{RNGkind()} in the console. You can set it by typing
  \code{RNGkind("L'Ecuyer-CMRG")}, before the lavaanList functions are called.}
\item{show_progress}{If \code{TRUE}, show information for each dataset.}
}
\value{
  An object of class \code{\linkS4class{lavaanList}}, for which several methods
  are available, including a \code{summary} method.
}
\seealso{
class \code{\linkS4class{lavaanList}}
}
\examples{
# The Holzinger and Swineford (1939) example
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

# a data generating function
generateData <- function() lavSimulateData(HS.model, sample_nobs = 100)

set.seed(1234)
fit <- semList(HS.model, data_function = generateData, ndat = 5,
               store_slots = "partable")

# show parameter estimates, per dataset
coef(fit)
}

\name{lavInspect}
\alias{lavInspect}
\alias{inspect}
\alias{lavTech}
\title{Inspect or extract information from a fitted lavaan object}
\description{
The \code{lavInspect()} and \code{lavTech()} functions can be used to
inspect/extract information that is stored inside (or can be computed from) a
fitted lavaan object. Note: the (older) \code{inspect()} function is
now simply a shortcut for \code{lavInspect()} with default arguments.
}
\usage{
lavInspect(object, what = "free", add_labels = TRUE, add_class = TRUE,
           list_by_group = TRUE,
           drop_list_single_group = TRUE, ...)

lavTech(object, what = "free", add_labels = FALSE, add_class = FALSE,
        list_by_group = FALSE,
        drop_list_single_group = FALSE, ...)

inspect(object, what = "free", ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{what}{Character. What needs to be inspected/extracted? See Details for a
full list. Note: the \code{what} argument is not case-sensitive
(everything is converted to lower case.)}
\item{add_labels}{If \code{TRUE}, variable names are added to the vectors
and/or matrices.}
\item{add_class}{If \code{TRUE}, vectors are given the `lavaan.vector' class;
matrices are given the `lavaan.matrix' class, and symmetric matrices are
given the `lavaan.matrix.symmetric' class. This only affects the way they
are printed on the screen.}
\item{list_by_group}{Logical. Only used when the output are model matrices.
If \code{TRUE}, the model matrices are nested within groups. If \code{FALSE},
a flattened list is returned containing all model matrices, with repeated
names for multiple groups.}
\item{drop_list_single_group}{If \code{FALSE}, the results are returned as
    a list, where each element corresponds to a group (even if there is only
    a single group). If \code{TRUE}, the list will be unlisted if there is
    only a single group.}
\item{...}{For \code{inspect()} additional arguments can be specified, not used by lavaan, but by other packages.
For the functions \code{lavInspect()} and \code{lavTech()} this argument is used to support old names of arguments.}
}
\details{
The \code{lavInspect()} and \code{lavTech()} functions only differ in the way
they return the results. The \code{lavInspect()} function will prettify the
output by default, while the \code{lavTech()} will not attempt to prettify the
output by default. The (older) \code{inspect()} function is a simplified
version of \code{lavInspect()} with only the first two arguments.

Below is a list of possible values for the \code{what} argument, organized
in several sections:

Model matrices:

\describe{
    \item{\code{"free"}:}{A list of model matrices. The non-zero integers
        represent the free parameters. The numbers themselves correspond
        to the position of the free parameter in the parameter vector.
        This determines the order of the model parameters in the output
        of, for example, \code{coef()} and \code{vcov()}.}
    \item{\code{"partable"}:}{A list of model matrices. The non-zero integers
        represent both the fixed parameters (for example, factor loadings
        fixed at 1.0) and the free parameters (ignoring any equality
        constraints). They correspond to all entries (fixed or free)
        in the parameter table. See \code{\link{parTable}}.}
    \item{\code{"se"}:}{A list of model matrices. The non-zero numbers
        represent the standard errors for the free parameters in the model.
        If two parameters are constrained to be equal, they will share the
        same standard error.
        Aliases: \code{"std.err"} and \code{"standard.errors"}.}
    \item{\code{"se.std"}:}{A list of model matrices. The non-zero numbers
        represent the standard errors for the (completely) standardized free
        parameters in the model. Alias: \code{"std.se"}.}
    \item{\code{"start"}:}{A list of model matrices. The values represent
        the starting values for all model parameters.
        Alias: \code{"starting.values"}.}
    \item{\code{"est"}:}{A list of model matrices. The values represent
        the estimated model parameters. Aliases:
        \code{"estimates"}, and \code{"x"}.}
    \item{\code{"est.unrotated"}:}{A list of model matrices. The values
        represent the estimated model parameters before rotation was applied.
        Only relevant if rotation was used (for example in EFA); otherwise
        this is identical to \code{"est"}.}
    \item{\code{"dx.free"}:}{A list of model matrices. The values represent
        the gradient (first derivative) values of the model parameters.
        If two parameters are constrained to be equal, they will have the
        same gradient value.}
    \item{\code{"dx.all"}:}{A list of model matrices. The values represent
        the first derivative with respect to all possible matrix elements.
        Currently, this is only available when the estimator is \code{"ML"}
        or \code{"GLS"}.}
    \item{\code{"std"}:}{A list of model matrices. The values represent
        the (completely) standardized model parameters (the variances of
        both the observed and the latent variables are set to unity).
        Aliases: \code{"std.all"}, \code{"standardized"}.}
    \item{\code{"std.lv"}:}{A list of model matrices. The values represent
        the standardized model parameters (only the variances of the
        latent variables are set to unity.)}
    \item{\code{"std.nox"}:}{A list of model matrices. The values represent
        the (completely) standardized model parameters (the variances of
        both the observed and the latent variables are set to unity;
        however, the variances of any observed exogenous variables are not
        set to unity; hence no-x.)}
    \item{\code{"mm.lambda"}:}{The estimated factor loading matrix (per
        group), as extracted from the \code{@GLIST} slot.
        Alias: \code{"lambda"}.}
    \item{\code{"mm.delta"}:}{The estimated delta (scaling) matrix (per group),
        as extracted from the \code{@GLIST} slot. Only available for
        categorical data or when a correlation structure is used. (Note: this
        is the delta model matrix, not the delta \emph{jacobian} that is
        returned by \code{"delta"}.)}
}

To extract a single model matrix (from the \code{@GLIST} slot), use the
\code{"mm.*"} options (for example \code{"mm.lambda"} for the factor loading
matrix). For backward compatibility, the (older) bare names are still accepted
as aliases.

Information about the data:

\describe{
    \item{\code{"data"}:}{A matrix containing the observed variables
        that were used to fit the model. The matrix has no row or column
        names. The columns correspond to the output of
        \code{lav_object_vnames(object)}, while the rows correspond to the
        output of \code{lavInspect(object, "case.idx")}.}
    \item{\code{"ordered"}:}{A character vector. The ordered variables.}
    \item{\code{"nobs"}:}{Numeric vector. The effective number of observations
        in each group that were used in the analysis. If sampling weights were
        provided, this is the (per group) sum of the weights, consistent with
        the value returned by the \code{nobs()} method. Use \code{"norig"} to
        obtain the number of rows in the data.}
    \item{\code{"norig"}:}{Integer vector. The original number of observations
        (rows in the data) in each group.}
    \item{\code{"ntotal"}:}{Numeric. The total effective number of observations
        that were used in the analysis. If there is just a single group, this
        is the same as the \code{"nobs"} option; if there are multiple groups,
        this is the sum of the \code{"nobs"} numbers for each group. If sampling
        weights were provided, this is the sum of the weights.}
    \item{\code{"case.idx"}:}{Integer vector. The case/observation numbers
        that were used in the analysis.
        In the case of multiple groups: a list of numbers.}
    \item{\code{"empty.idx"}:}{The case/observation numbers of those
        cases/observations that contained missing values only
        (at least for the observed variables that were included in the model).
        In the case of multiple groups: a list of numbers.}
    \item{\code{"th.idx"}:}{Integer vector. For categorical data, the index of
        the observed variable that each threshold belongs to.
        In the case of multiple groups: a list of integer vectors.}
    \item{\code{"patterns"}:}{A binary matrix. The rows of the matrix
        are the missing data patterns where 1 and 0 denote non-missing
        and missing values for the corresponding observed variables
        respectively (or
        \code{TRUE} and \code{FALSE} if \code{lavTech()} is used.)
        If the data is complete (no missing values), there will be only
        a single pattern. In the case of multiple groups: a list of
        pattern matrices.}
    \item{\code{"coverage"}:}{A symmetric matrix in which each element
        contains the proportion of jointly observed data points for the
        corresponding pair of observed variables.
        In the case of multiple groups: a list of coverage matrices.}
    \item{\code{"group"}:}{A character string. The group variable in
        the data.frame (if any).}
    \item{\code{"ngroups"}:}{Integer. The number of groups.}
    \item{\code{"group.label"}:}{A character vector. The group labels.}
    \item{\code{"level.label"}:}{A character vector. The level labels.}
    \item{\code{"cluster"}:}{A character vector. The cluster variable(s)
        in the data.frame (if any).}
    \item{\code{"nlevels"}:}{Integer. The number of levels.}
    \item{\code{"nclusters"}:}{Integer. The number of clusters
        that were used in the analysis.}
    \item{\code{"ncluster.size"}:}{Integer. The number of different cluster
        sizes.}
    \item{\code{"cluster.size"}:}{Integer vector. The number of observations
        within each cluster. For multigroup multilevel models, a list of
        integer vectors, indicating cluster sizes within each group.}
    \item{\code{"cluster.id"}:}{Integer vector. The cluster IDs identifying
        the clusters. For multigroup multilevel models, a list of
        integer vectors, indicating cluster IDs within each group.}
    \item{\code{"cluster.idx"}:}{Integer vector. The cluster index for
        each observation. The cluster index ranges from 1 to the
        number of clusters. For multigroup multilevel models, a list of
        integer vectors, indicating cluster indices within each group.}
    \item{\code{"cluster.label"}:}{Integer vector. The cluster ID for
        each observation. For multigroup multilevel models, a list of
        integer vectors, indicating the cluster ID for each observation
        within each group.}
    \item{\code{"cluster.sizes"}:}{Integer vector. The different cluster
        sizes that were used in the analysis. For multigroup multilevel
        models, a list of integer vectors, indicating the different cluster
        sizes within each group.}
    \item{\code{"average.cluster.size"}:}{Integer. The average cluster
        size (using the formula
        \code{s = (N^2 - sum(cluster.size^2)) / (N*(nclusters - 1L))}).
        For multigroup multilevel
        models, a list containing the average cluster size per group.}
}

Observed sample statistics:

\describe{
    \item{\code{"sampstat"}:}{Observed sample statistics. Aliases:
         \code{"obs"}, \code{"observed"}, \code{"samp"}, \code{"sample"},
         \code{"samplestatistics"}. Since
         0.6-3, we always check if an h1 slot is available (the estimates
         for the unrestricted model); if present, we extract the sample
         statistics from this slot. This implies that if the variables are
         continuous and \code{missing = "ml"} (or \code{"fiml"}), we
         return the covariance matrix (and mean vector) as computed by
         the EM algorithm under the unrestricted (h1) model. If the h1 slot
         is not present (for example, because the model was fitted with
         \code{h1 = FALSE}), we return the sample statistics from the
         SampleStats slot. In that case, pairwise deletion is used for the
         elements of the covariance matrix (or correlation matrix), and
         listwise deletion for all univariate statistics (means, intercepts,
         and thresholds).}
    \item{\code{"sampstat.std"}:}{Standardized observed sample statistics.
         The covariance matrix is rescaled to a correlation matrix. Aliases:
         \code{"obs.std"}, \code{"observed.std"}, \code{"samp.std"},
         \code{"sample.std"}, \code{"samplestatistics.std"}.}
    \item{\code{"sampstat.h1"}:}{Deprecated. Do not use any longer.}
    \item{\code{"wls.obs"}:}{The observed sample statistics (covariance
        elements, intercepts/thresholds, etc.) in a single vector.}
    \item{\code{"wls.v"}:}{The weight vector as used in weighted least
        squares estimation.}
    \item{\code{"gamma"}:}{N times the asymptotic variance matrix of the
        sample statistics. Alias: \code{"sampstat.nacov"}.}
}

Model features:

\describe{
    \item{\code{"meanstructure"}:}{Logical. \code{TRUE} if a meanstructure
        was included in the model.}
    \item{\code{"categorical"}:}{Logical. \code{TRUE} if categorical endogenous
        variables were part of the model.}
    \item{\code{"fixed.x"}:}{Logical. \code{TRUE} if the exogenous x-covariates
        are treated as fixed.}
    \item{\code{"parameterization"}:}{Character. Either \code{"delta"} or
        \code{"theta"}.}
}


Model-implied sample statistics:

\describe{
    \item{\code{"implied"}:}{The model-implied summary statistics.
        Alias: \code{"fitted"}, \code{"expected"}, \code{"exp"}.}
    \item{\code{"resid"}:}{The difference between observed and model-implied
        summary statistics.
        Alias: \code{"residuals"}, \code{"residual"}, \code{"res"}.}
    \item{\code{"cov.lv"}:}{The model-implied variance-covariance matrix
        of the latent variables. Alias: \code{"veta"} [for V(eta)].}
    \item{\code{"cor.lv"}:}{The model-implied correlation matrix of the
        latent variables.}
    \item{\code{"mean.lv"}:}{The model-implied mean vector of the latent
        variables. Alias: \code{"eeta"} [for E(eta)].}
    \item{\code{"cov.ov"}:}{The model-implied variance-covariance matrix
        of the observed variables.
        Aliases: \code{"sigma"}, \code{"sigma.hat"}.}
    \item{\code{"cor.ov"}:}{The model-implied correlation matrix
        of the observed variables.}
    \item{\code{"mean.ov"}:}{The model-implied mean vector of the observed
        variables. Aliases: \code{"mu"}, \code{"mu.hat"}.}
    \item{\code{"cov.all"}:}{The model-implied variance-covariance matrix
        of both the observed and latent variables.}
    \item{\code{"cor.all"}:}{The model-implied correlation matrix
        of both the observed and latent variables.}
    \item{\code{"th"}:}{The model-implied thresholds.
        Alias: \code{"thresholds"}.}
    \item{\code{"mm.theta"}:}{The (residual) variance-covariance
        matrix of the observed variables (the theta model matrix).
        Aliases: \code{"theta"}, \code{"theta.cov"}.}
    \item{\code{"theta.cor"}:}{The (residual) variance-covariance
        matrix of the observed variables, expressed in correlation metric.}
    \item{\code{"wls.est"}:}{The model-implied sample statistics (covariance
        elements, intercepts/thresholds, etc.) in a single vector.}
    \item{\code{"vy"}:}{The model-implied unconditional variances of the
        observed variables.}
    \item{\code{"rsquare"}:}{The R-square value for all endogenous variables.
        Aliases: \code{"r-square"}, \code{"r2"}.}
	\item{\code{"fs_determinacy"}:}{The factor determinacies (based on
	   regression factor scores). They represent the (estimated) correlation
       between the factor scores and the latent variable scores.}
	\item{\code{"fs.reliability"}:}{The factor reliabilities (based on
       regression factor scores). They are the square of the factor
       determinacies.}
    \item{\code{"fs_determinacy_Bartlett"}:}{The factor determinacies (based on
       Bartlett factor scores). They represent the (estimated) correlation
       between the factor scores and the latent variable scores.}
    \item{\code{"fs_reliability_bartlett"}:}{The factor reliabilities (based on
       Bartlett factor scores). They are the square of the factor
       determinacies.}
    \item{\code{"icc"}:}{The intraclass correlation coefficients for clustered
       data with multilevel models. Computed from the model-implied (H1)
       within-group and between-group covariance matrices. Only available
       for models with clustered data and multiple levels.}
    \item{\code{"ranef"}:}{The random effects (empirical Bayes predictions of
       the cluster-level latent variables). Only available for clustered data
       (in the long format) with multiple levels.}
}

Diagnostics:

\describe{
    \item{\code{"mdist2.fs"}:}{The squared Mahalanobis distances for the (Bartlett) factor scores.}
    \item{\code{"mdist.fs"}:}{The Mahalanobis distances for the (Bartlett) factor scores.}
    \item{\code{"mdist2.resid"}:}{The squared Mahalanobis distances for the (Bartlett-based) casewise residuals.}
    \item{\code{"mdist.resid"}:}{The Mahalanobis distances for the (Bartlett-based) casewise residuals.}
}

If (some of) the data is ordered categorical, the Mahalanobis distances are
generalized as in Mansolf and Reise (2017): each distance is replaced by its
expected value over the region of the (multivariate normally distributed)
latent response vector that is consistent with the observed response pattern,
approximated by Monte Carlo integration (\code{set.seed()} can be used for
reproducibility). Observed continuous variables are conditioned on, and
missing values are integrated out. Note that in this case
\code{"mdist.fs"}/\code{"mdist.resid"} return the expected distance
\eqn{E[d]} (the person-fit indices \eqn{d_f^*} and \eqn{d_r^*} of Mansolf
and Reise), while \code{"mdist2.fs"}/\code{"mdist2.resid"} return the
expected squared distance \eqn{E[d^2]}, which is not the square of the
former.

Optimizer information:

\describe{
    \item{\code{"converged"}:}{Logical. \code{TRUE} if the optimizer has
        converged; \code{FALSE} otherwise.}
    \item{\code{"iterations"}:}{Integer. The number of iterations used by the
        optimizer.}
    \item{\code{"optim"}:}{List. All available information regarding the
        optimization results.}
    \item{\code{"npar"}:}{Integer. Number of free parameters (ignoring constraints).}
}

Gradient, Hessian, observed, expected and first.order information matrices:

\describe{
    \item{\code{"gradient"}:}{Numeric vector containing the first derivatives
        of the discrepancy function with respect to the (free) model
        parameters.}
    \item{\code{"gradient.logl"}:}{Numeric vector containing the first
        derivatives of the loglikelihood with respect to the (free) model
        parameters.}
    \item{\code{"optim.gradient"}:}{Numeric vector containing the first
        derivatives of the objective function (as seen by the optimizer)
        with respect to the (free) model parameters.}
    \item{\code{"hessian"}:}{Matrix containing the second derivatives
        of the discrepancy function with respect to the (free) model
        parameters.}
    \item{\code{"information"}:}{Matrix containing either the observed or
        the expected information matrix (depending on the information option
        of the fitted model). This is unit-information, not total-information.}
    \item{\code{"information.expected"}:}{Matrix containing the expected
        information matrix for the free model parameters.}
    \item{\code{"information.observed"}:}{Matrix containing the observed
        information matrix for the free model parameters.}
    \item{\code{"information.first.order"}:}{Matrix containing the first.order
        information matrix for the free model parameters. This is the
        outer product of the gradient elements (the first derivative of
        the discrepancy function with respect to the (free) model parameters).
        Alias: \code{"first.order"}.}
    \item{\code{"augmented.information"}:}{Matrix containing either the
        observed or the expected augmented (or bordered) information
        matrix (depending on the information option of the fitted model).
        Only relevant if constraints have been used in the model.}
    \item{\code{"augmented.information.expected"}:}{Matrix containing the
        expected augmented (or bordered) information matrix.
        Only relevant if constraints have been used in the model.}
    \item{\code{"augmented.information.observed"}:}{Matrix containing the
        observed augmented (or bordered) information matrix.
        Only relevant if constraints have been used in the model.}
    \item{\code{"augmented.information.first.order"}:}{Matrix containing
        the first.order augmented (or bordered) information matrix.
        Only relevant if constraints have been used in the model.}
    \item{\code{"inverted.information"}:}{Matrix containing either the
        observed or the expected inverted information matrix
        (depending on the information option of the fitted model).}
    \item{\code{"inverted.information.expected"}:}{Matrix containing the
        inverted expected information matrix for the free model parameters.}
    \item{\code{"inverted.information.observed"}:}{Matrix containing the
        inverted observed information matrix for the free model parameters.}
    \item{\code{"inverted.information.first.order"}:}{Matrix containing the
        inverted first.order information matrix for the free model parameters.}
   \item{\code{"h1.information"}:}{Matrix containing either the observed,
        expected or first.order information matrix (depending on the
        information option of the fitted model) of the unrestricted h1
        model. This is unit-information, not total-information.}
    \item{\code{"h1.information.expected"}:}{Matrix containing the expected
        information matrix for the unrestricted h1 model.}
    \item{\code{"h1.information.observed"}:}{Matrix containing the observed
        information matrix for the unrestricted h1 model.}
    \item{\code{"h1.information.first.order"}:}{Matrix containing the
        first.order information matrix for the unrestricted h1 model.
        Alias: \code{"h1.first.order"}.}
}

Variance covariance matrix of the model parameters:

\describe{
    \item{\code{"vcov"}:}{Matrix containing the variance covariance matrix
        of the estimated model parameters.}
    \item{\code{"vcov.std.all"}:}{Matrix containing the variance covariance
        matrix of the standardized estimated model parameters. Standardization
        is done with respect to both observed and latent variables.}
    \item{\code{"vcov.std.lv"}:}{Matrix containing the variance covariance
        matrix of the standardized estimated model parameters. Standardization
        is done with respect to the latent variables only.}
    \item{\code{"vcov.std.nox"}:}{Matrix containing the variance covariance
        matrix of the standardized estimated model parameters. Standardization
        is done with respect to both observed and latent variables, but
        ignoring any exogenous observed covariates.}
    \item{\code{"vcov.def"}:}{Matrix containing the variance covariance matrix
        of the user-defined (using the := operator) parameters.}
    \item{\code{"vcov.def.std.all"}:}{Matrix containing the variance covariance
        matrix of the standardized user-defined parameters. Standardization
        is done with respect to both observed and latent variables.}
    \item{\code{"vcov.def.std.lv"}:}{Matrix containing the variance covariance
        matrix of the standardized user-defined parameters. Standardization
        is done with respect to the latent variables only.}
    \item{\code{"vcov.def.std.nox"}:}{Matrix containing the variance covariance
        matrix of the standardized user-defined parameters. Standardization
        is done with respect to both observed and latent variables, but
        ignoring any exogenous observed covariates.}
    \item{\code{"vcov.def.joint"}:}{Matrix containing the joint variance
        covariance matrix of both the estimated model parameters and
        the defined (using the := operator) parameters.}
    \item{\code{"vcov.def.joint.std.all"}:}{Matrix containing the joint
        variance covariance matrix of both the standardized model parameters
        and the user-defined parameters. Standardization
        is done with respect to both observed and latent variables.}
    \item{\code{"vcov.def.joint.std.lv"}:}{Matrix containing the joint
        variance covariance matrix of both the standardized model parameters
        and the user-defined parameters. Standardization
        is done with respect to the latent variables only.}
    \item{\code{"vcov.def.joint.std.nox"}:}{Matrix containing the joint
        variance covariance matrix of both the standardized model parameters
        and the user-defined parameters. Standardization
        is done with respect to both observed and latent variables, but
        ignoring any exogenous observed covariates.}
}

Miscellaneous:

\describe{
    \item{\code{"coef.boot"}:}{Matrix containing estimated model parameters for each bootstrap sample. Only relevant when bootstrapping was used.}
    \item{\code{"monte.carlo"}:}{Matrix containing the Monte Carlo draws of the
        model parameters, as used for the Monte Carlo confidence intervals of
        defined parameters (Preacher & Selig, 2012). Only relevant when these
        draws were requested. Aliases: \code{"mc"}, \code{"mc.coef"},
        \code{"coef.mc"}.}
    \item{\code{"UGamma"}:}{Matrix containing the product of 'U' and 'Gamma'
        matrices as used by the Satorra-Bentler correction. The trace of
        this matrix, divided by the degrees of freedom, gives the scaling
        factor.}
    \item{\code{"UfromUGamma"}:}{Matrix containing the 'U' matrix
        as used by the Satorra-Bentler correction. Alias: \code{"U"}.}
    \item{\code{"delta"}:}{The delta matrix (per group): the Jacobian of the
        model-implied summary statistics (the rows: means, (co)variances,
        thresholds, \ldots) with respect to the free model parameters (the
        columns). This is the matrix used in the delta method. (Note: this is
        the delta \emph{Jacobian}, not the delta model matrix that is returned
        by \code{"mm.delta"}.) In the case of multiple groups: a list of
        matrices.}
    \item{\code{"delta.rownames"}:}{The row names of the delta matrix (see
        \code{"delta"}): the labels of the model-implied summary statistics.
        In the case of multiple groups: a list of character vectors.}
    \item{\code{"list"}:}{The parameter table. The same output as given
        by \code{parTable()}.}
    \item{\code{"fit"}:}{The fit measures. Aliases: \code{"fitmeasures"},
        \code{"fit.measures"}, \code{"fit.indices"}. The same output as
        given by \code{fitMeasures()}.}
    \item{\code{"mi"}:}{The modification indices. Alias: \code{"modindices"},
        \code{"modification.indices"}. The same output as given
        by \code{modindices()}.}
    \item{\code{"loglik.casewise"}:}{Vector containing the casewise
        loglikelihood contributions. Only available if estimator = \code{"ML"}.}
    \item{\code{"options"}:}{List. The option list.}
    \item{\code{"call"}:}{List. The call as returned by match.call, coerced to
        a list.}
    \item{\code{"timing"}:}{List. The timing (in milliseconds) of various
        lavaan subprocedures.}
    \item{\code{"test"}:}{List. All available information regarding the
        (goodness-of-fit) test statistic(s).}
    \item{\code{"baseline.test"}:}{List. All available information regarding the
        (goodness-of-fit) test statistic(s) of the baseline model.}
    \item{\code{"baseline.partable"}:}{Data.frame. The parameter table of
        the (internal) baseline model.}
    \item{\code{"post.check"}:}{Post-fitting check if the solution is
        admissible. A warning is raised if negative variances are found, or if
        either \code{lavInspect(fit, "cov.lv")} or
        \code{lavInspect(fit, "theta")} return a non-positive definite matrix.}
    \item{\code{"zero.cell.tables"}:}{List. List of bivariate frequency tables
        where at least one cell is empty.}
    \item{\code{"iv"}:}{Data.frame. One row per (transformed) equation, with
        the (model-implied) instrumental variables that were used for each
        equation. Columns are \code{lhs}, \code{rhs}, \code{lhs.new},
        \code{rhs.new}, \code{type}, and \code{instruments}. Only available
        when \code{estimator = "IV"}. Aliases: \code{"ivs"}, \code{"miiv"},
        \code{"miivs"}, \code{"instr"}, \code{"instruments"}. In the case of
        multiple groups: a list of data.frames.}
    \item{\code{"eqs"}:}{List. The (internal) list of equations that was used
        for the (MI)IV-based estimation, including the instruments, the
        coefficients, and the Sargan test per equation. Only available when
        \code{estimator = "IV"}.}
    \item{\code{"sargan"}:}{Data.frame. The per-equation overidentification
        tests, for each overidentified equation. Columns are
        \code{lhs}, \code{rhs}, \code{df}, \code{sargan.stat},
        \code{sargan.pval}, \code{browne.stat}, and \code{browne.pval}. The \code{sargan.*} columns are the classical
        Sargan test (assumes normality; the p-value is \code{NA} for
        categorical data, where it is not valid); the \code{browne.*} columns
        are a robust residual-based test computed with the distribution-free
        (ADF) ACOV for continuous data or the polychoric ACOV for categorical
        data. \code{"hansen"} is an alias. Only available when
        \code{estimator = "IV"}. In the case of multiple groups: a list of
        data.frames.}
    \item{\code{"sam.struc.fit.object"}:}{The (internal) step-2 structural
        model of a model fitted with \code{\link{sam}} using a local
        \code{sam.method} (\code{"local"}, \code{"fsr"} or \code{"cfsr"}):
        a lavaan object, fitted to the estimated latent (co)variances (and
        means) of step 1. \code{fitMeasures()}, \code{lavResiduals()} and
        \code{modindices()} operate on this object when they are called with
        the sam object itself. \code{"sam.struc"} is an alias.}
    \item{\code{"version"}:}{The lavaan version number that was used to
        construct the fitted lavaan object.}
}

}
\seealso{
\code{\link{lavaan}}, \code{\link{estimator_iv}}
}
\examples{
# fit model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '


fit <- cfa(HS.model, data = HolzingerSwineford1939, group = "school")

# extract information
lavInspect(fit, "sampstat")
lavTech(fit, "sampstat")
}

\name{lavListInspect}
\alias{lavListInspect}
\alias{lavListTech}
\title{Inspect or extract information from a lavaanList object}
\description{
The \code{lavListInspect()} and \code{lavListTech()} functions can be used to
inspect/extract information that is stored inside (or can be computed from) a
lavaanList object.
}
\usage{
lavListInspect(object, what = "free", add_labels = TRUE,
               add_class = TRUE, list_by_group = TRUE,
               drop_list_single_group = TRUE, ...)

lavListTech(object, what = "free", add_labels = FALSE,
            add_class = FALSE, list_by_group = FALSE,
            drop_list_single_group = FALSE, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaanList}}.}
\item{what}{Character. What needs to be inspected/extracted? See Details for a
full list. Note: the \code{what} argument is not case-sensitive
(everything is converted to lower case.)}
\item{add_labels}{If \code{TRUE}, variable names are added to the vectors
and/or matrices.}
\item{add_class}{If \code{TRUE}, vectors are given the `lavaan.vector' class;
matrices are given the `lavaan.matrix' class, and symmetric matrices are
given the `lavaan.matrix.symmetric' class. This only affects the way they
are printed on the screen.}
\item{list_by_group}{Logical. Only used when the output are model matrices.
If \code{TRUE}, the model matrices are nested within groups. If \code{FALSE},
a flattened list is returned containing all model matrices, with repeated
names for multiple groups.}
\item{drop_list_single_group}{If \code{FALSE}, the results are returned as
    a list, where each element corresponds to a group (even if there is only
    a single group.) If \code{TRUE}, the list will be unlisted if there is
    only a single group.}
\item{...}{To support old argument names.}
}
\details{
The \code{lavListInspect()} and \code{lavListTech()} functions only differ in
the way they return the results. The \code{lavListInspect()} function will
prettify the output by default, while the \code{lavListTech()} will not attempt
to prettify the output by default.

Below is a list of possible values for the \code{what} argument, organized
in several sections:

Model matrices:

\describe{
    \item{\code{"free"}:}{A list of model matrices. The non-zero integers
        represent the free parameters. The numbers themselves correspond
        to the position of the free parameter in the parameter vector.
        This determines the order of the model parameters in the output
        of, for example, \code{coef()} and \code{vcov()}.}
    \item{\code{"partable"}:}{A list of model matrices. The non-zero integers
        represent both the fixed parameters (for example, factor loadings
        fixed at 1.0) and the free parameters (ignoring any equality
        constraints). They correspond to all entries (fixed or free)
        in the parameter table. See \code{\link{parTable}}.}
    \item{\code{"start"}:}{A list of model matrices. The values represent
        the starting values for all model parameters.
        Alias: \code{"starting.values"}.}
}

Information about the data (including missing patterns):

\describe{
    \item{\code{"group"}:}{A character string. The group variable in
        the data.frame (if any).}
    \item{\code{"ngroups"}:}{Integer. The number of groups.}
    \item{\code{"group.label"}:}{A character vector. The group labels.}
    \item{\code{"level.label"}:}{A character vector. The level labels.}
    \item{\code{"cluster"}:}{A character vector. The cluster variable(s)
       in the data.frame (if any).}
    \item{\code{"nlevels"}:}{Integer. The number of levels.}
    \item{\code{"ordered"}:}{A character vector. The ordered variables.}
    \item{\code{"nobs"}:}{Integer vector. The number of observations
        in each group that were used in the analysis (in each dataset).}
    \item{\code{"norig"}:}{Integer vector. The original number of observations
        in each group (in each dataset).}
    \item{\code{"ntotal"}:}{Integer. The total number of observations that
        were used in the analysis. If there is just a single group, this
        is the same as the \code{"nobs"} option; if there are multiple groups,
        this is the sum of the \code{"nobs"} numbers for each group
        (in each dataset).}
}

Model features:

\describe{
    \item{\code{"meanstructure"}:}{Logical. \code{TRUE} if a meanstructure
        was included in the model.}
    \item{\code{"categorical"}:}{Logical. \code{TRUE} if categorical endogenous
        variables were part of the model.}
    \item{\code{"fixed.x"}:}{Logical. \code{TRUE} if the exogenous x-covariates
        are treated as fixed.}
    \item{\code{"parameterization"}:}{Character. Either \code{"delta"} or
        \code{"theta"}.}
}

\describe{
    \item{\code{"list"}:}{The parameter table. The same output as given
        by \code{parTable()}.}
    \item{\code{"options"}:}{List. The option list.}
    \item{\code{"call"}:}{List. The call as returned by match.call, coerced to
        a list.}
}

}
\seealso{
\code{\link{lavaanList}}
}
\examples{
# fit model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

# a data generating function
generateData <- function() lavSimulateData(HS.model, sample_nobs = 100)

set.seed(1234)
fit <- semList(HS.model, dataFunction = generateData, ndat = 5,
               store_slots = "partable")

# extract information
lavListInspect(fit, "free")
lavListTech(fit, "free")
}

\name{model.syntax}
\alias{model.syntax}
\alias{lavaanify}
\alias{lavParTable}
\alias{parseModelString}
\alias{lavParseModelString}
\title{The Lavaan Model Syntax}
\description{
The lavaan model syntax describes a latent variable model. The
function \code{lavParTable} turns it into a table that represents the full
model as specified by the user. We refer to this table as the parameter table.}
\usage{
lavaanify(model = NULL, meanstructure = FALSE, int_ov_free = FALSE,
    int_lv_free = FALSE, marker_int_zero = FALSE,
    orthogonal = FALSE, orthogonal_y = FALSE,
    orthogonal_x = FALSE, orthogonal_efa = FALSE, std_lv = FALSE,
    correlation = FALSE, composites = TRUE, composites_cov_free = FALSE,
    effect_coding = "", conditional_x = FALSE,
    fixed_x = FALSE, parameterization = "delta", constraints = NULL,
    ceq_simple = FALSE, auto = FALSE, model_type = "sem",
    auto_fix_first = FALSE, marker = NULL,
    auto_fix_single = FALSE, auto_var = FALSE,
    auto_cov_lv_x = FALSE, auto_cov_x = FALSE, auto_cov_y = FALSE,
    auto_th = FALSE, auto_delta = FALSE, auto_efa = FALSE,
    var_table = NULL, ngroups = 1L, nthresholds = NULL,
    group_equal = NULL, group_partial = NULL, group_w_free = FALSE,
    debug = FALSE, warn = TRUE, as_data_frame = TRUE, ...)

lavParTable(model = NULL, meanstructure = FALSE, int_ov_free = FALSE,
    int_lv_free = FALSE, marker_int_zero = FALSE,
    orthogonal = FALSE, orthogonal_y = FALSE,
    orthogonal_x = FALSE, orthogonal_efa = FALSE, std_lv = FALSE,
    correlation = FALSE, composites = TRUE, composites_cov_free = FALSE,
    effect_coding = "", conditional_x = FALSE,
    fixed_x = FALSE, parameterization = "delta", constraints = NULL,
    ceq_simple = FALSE, auto = FALSE, model_type = "sem",
    auto_fix_first = FALSE, marker = NULL,
    auto_fix_single = FALSE, auto_var = FALSE,
    auto_cov_lv_x = FALSE, auto_cov_x = FALSE, auto_cov_y = FALSE,
    auto_th = FALSE, auto_delta = FALSE, auto_efa = FALSE,
    var_table = NULL, ngroups = 1L, nthresholds = NULL,
    group_equal = NULL, group_partial = NULL, group_w_free = FALSE,
    debug = FALSE, warn = TRUE, as_data_frame = TRUE, ...)

lavParseModelString(model_syntax = '', as_data_frame = FALSE,
                    parser = "open", warn = TRUE, debug = FALSE, ...)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax; see details for more
    information. Alternatively, a parameter table (e.g., the output of
    \code{lavParseModelString}) is also accepted.}
\item{model_syntax}{The model syntax specifying the model. Must be a literal
string.}
\item{meanstructure}{If \code{TRUE}, intercepts/means will be added to
the model for both observed and latent variables.}
\item{int_ov_free}{If \code{FALSE}, the intercepts of the observed variables
    are fixed to zero.}
\item{int_lv_free}{If \code{FALSE}, the intercepts of the latent variables
    are fixed to zero.}
\item{marker_int_zero}{Logical. Only relevant if the metric of each latent
    variable is set by fixing the first factor loading to unity.
    If \code{TRUE}, it implies \code{meanstructure = TRUE} and
    \code{std_lv = FALSE}, and it fixes the intercepts of the marker
    indicators to zero, while freeing the means/intercepts of the latent
    variables. This only works correctly for single-group, single-level
    models.}
\item{orthogonal}{If \code{TRUE}, all covariances among
   latent variables are set to zero.}
\item{orthogonal_y}{If \code{TRUE}, all covariances among
   endogenous latent variables only are set to zero.}
\item{orthogonal_x}{If \code{TRUE}, all covariances among
   exogenous latent variables only are set to zero.}
\item{orthogonal_efa}{If \code{TRUE}, all covariances among
   latent variables involved in rotation only are set to zero.}
\item{std_lv}{If \code{TRUE}, the metric of each latent variable is
    determined by fixing their variances to 1.0. If \code{FALSE}, the metric
    of each latent variable is determined by fixing the factor loading of the
    first indicator to 1.0. If there are multiple
      groups, \code{std_lv = TRUE} and \code{"loadings"} is included in
      the \code{group.label} argument, then only the latent variances
      of the first group will be fixed to 1.0, while the latent
      variances of the other groups are freely estimated.}
\item{correlation}{If \code{TRUE}, a correlation structure is fitted. For
    continuous data, this implies that the (residual) variances are no
    longer parameters of the model.}
\item{composites}{Logical. If \code{TRUE}, use the new (0.6-20) approach
     to handle composites.}
\item{composites_cov_free}{Logical. Only relevant if the model contains
     composites. If \code{TRUE}, the composite-indicator (co)variances (the
     elements of the \code{T} matrix) are estimated as free parameters. If
     \code{FALSE} (the default), they are fixed to their sample values.}
\item{effect_coding}{Can be logical or character string. If
     logical and \code{TRUE}, this implies
     \code{effect_coding = c("loadings", "intercepts")}. If logical and
     \code{FALSE}, it is set equal to the empty string.
     If \code{"loadings"} is included, equality
     constraints are used so that the average of the factor loadings (per
     latent variable) equals 1. Note that this should not be used
     together with \code{std_lv = TRUE}. If \code{"intercepts"} is
     included, equality constraints are used so that the sum of the
     intercepts (belonging to the indicators of a single latent variable)
     equals zero.
     As a result, the latent mean will be freely estimated and will usually
     equal the average of the means of the indicators involved.}
\item{conditional_x}{If \code{TRUE}, we set up the model conditional on
    the exogenous `x' covariates; the model-implied sample statistics
    only include the non-x variables. If \code{FALSE}, the exogenous `x'
    variables are modeled jointly with the other variables, and the
    model-implied statistics reflect both sets of variables.}
\item{fixed_x}{If \code{TRUE}, the exogenous `x' covariates are considered
    fixed variables and the means, variances and covariances of these variables
    are fixed to their sample values. If \code{FALSE}, they are considered
    random, and the means, variances and covariances are free parameters.}
\item{parameterization}{Currently only used if data is categorical. If
    \code{"delta"}, the delta parameterization is used. If \code{"theta"},
    the theta parameterization is used.}
\item{constraints}{Additional (in)equality constraints. See details for
    more information.}
\item{ceq_simple}{If \code{TRUE}, and no other general constraints are
    used in the model, simple equality constraints are represented in the
    parameter table as duplicated free parameters (instead of extra rows
    with \code{op = "=="}).}
\item{auto}{If \code{TRUE}, the default values are used for the auto.*
    arguments, depending on the value of \code{model_type}.}
\item{model_type}{Either \code{"sem"} or \code{"growth"}; only used if
    \code{auto=TRUE}.}
\item{auto_fix_first}{If \code{TRUE}, the factor loading of the first indicator
    is set to 1.0 for every latent variable.}
\item{marker}{Optional named character vector mapping a latent variable
    (name) to the observed indicator (name) whose loading should be fixed to
    1.0 (instead of the first indicator) when \code{auto_fix_first = TRUE}.
    Latent variables that are not named in this vector keep the first
    indicator as their marker. This is used internally by the
    \code{bad.marker.crit} mechanism.}
\item{auto_fix_single}{If \code{TRUE}, the residual variance (if included)
    of an observed indicator is set to zero if it is the only indicator of a
    latent variable.}
\item{auto_var}{If \code{TRUE}, the (residual) variances of both observed
    and latent variables are set free.}
\item{auto_cov_lv_x}{If \code{TRUE}, the covariances of exogenous latent
    variables are included in the model and set free.}
\item{auto_cov_x}{If \code{TRUE}, the covariances between exogenous latent
    variables and observed exogenous covariates are also included in the
    model and set free (this implies \code{auto_cov_lv_x = TRUE}). Ignored
    (with a warning) if \code{conditional_x = TRUE}.}
\item{auto_cov_y}{If \code{TRUE}, the covariances of dependent variables
    (both observed and latent) are included in the model and set free.}
\item{auto_th}{If \code{TRUE}, thresholds for limited dependent variables
    are included in the model and set free.}
\item{auto_delta}{If \code{TRUE}, response scaling parameters for limited
    dependent variables are included in the model and set free.}
\item{auto_efa}{If \code{TRUE}, the necessary constraints are
    imposed to make the (unrotated) exploratory factor analysis blocks
    identifiable: for each block, factor variances are set to 1, factor
    covariances are constrained to be zero, and factor loadings are
    constrained to follow an echelon pattern.}
\item{var_table}{The variable table containing information about the
    observed variables in the model.}
\item{ngroups}{The number of (independent) groups.}
\item{nthresholds}{Either a single integer or a named vector of integers.
    If \code{nthresholds} is a single integer, all endogenous
    variables are assumed to be ordered with \code{nthresholds} indicating
    the number of thresholds needed in the model. If \code{nthresholds} is a
    named vector, it indicates the number of thresholds for these ordered
    variables only. This argument should not be used in combination with
    var_table.}
\item{group_equal}{A vector of character strings. Only used in
    a multiple group analysis. Can be one or more of the following:
    \code{"loadings"}, \code{"intercepts"},
    \code{"means"}, \code{"regressions"}, \code{"residuals"} or
    \code{"covariances"}, specifying the pattern of equality
    constraints across multiple groups. When (in the model syntax) a vector
    of labels is used as a modifier for a certain parameter, this will
    override the group_equal setting if it applies to this parameter.
    See also the Multiple groups section below for using modifiers in multiple
    groups.}
\item{group_partial}{A vector of character strings containing the labels
    of the parameters which should be free in all groups (thereby
    overriding the group_equal argument for some specific parameters).}
\item{group_w_free}{Logical. If \code{TRUE}, the group frequencies are
    considered to be free parameters in the model. In this case, a
    Poisson model is fitted to estimate the group frequencies. If
    \code{FALSE} (the default), the group frequencies are fixed to their
    observed values.}
\item{as_data_frame}{If \code{TRUE}, return the list of model parameters
    as a \code{data.frame}.}
\item{parser}{Character. If \code{"old"}, use the original/classic parser.
    If \code{"new"}, use the new/ldw parser.
    If \code{"open"}, use the newest, extensible parser.
    The default (as of version 0.7-1) is \code{"open"}.}
\item{warn}{If \code{TRUE}, some (possibly harmless) warnings are printed
    out.}
\item{debug}{If \code{TRUE}, debugging information is printed out.}
\item{...}{To accept old argument names with dots. No other arguments are accepted.}
}
\details{
The model syntax consists of one or more formula-like expressions, each one
describing a specific part of the model. The model syntax can be read from
a file (using \code{\link{readLines}}), or can be specified as a literal
string enclosed by single quotes as in the example below.
\preformatted{myModel <- '
  # 1. latent variable definitions
    f1 =~ y1 + y2 + y3
    f2 =~ y4 + y5 + y6
    f3 =~ y7 + y8 +
          y9 + y10
    f4 =~ y11 + y12 + y13

  ! this is also a comment

  # 2. regressions
    f1 ~ f3 + f4
    f2 ~ f4
    y1 + y2 ~ x1 + x2 + x3

  # 3. (co)variances
    y1 ~~ y1
    y2 ~~ y4 + y5
    f1 ~~ f2

  # 4. intercepts
    f1 ~ 1; y5 ~ 1

  # 5. thresholds
    y11 | t1 + t2 + t3
    y12 | t1
    y13 | t1 + t2

  # 6. scaling factors
    y11 ~*~ y11
    y12 ~*~ y12
    y13 ~*~ y13

  # 7. composites
    f5 <~ z1 + z2 + z3 + z4
'
}
Blank lines and comments can be used in between the formulas, and formulas can
be split over multiple lines. Both the sharp (#) and the exclamation (!)
characters can be used to start a comment. Multiple formulas can be placed
on a single line if they are separated by a semicolon (;).

There can be seven types of formula-like expressions in the model syntax:
\enumerate{

    \item Latent variable definitions: The \code{"=~"} operator can be
      used to define (continuous) latent variables. The name of the latent
      variable is on the left of the \code{"=~"} operator, while the terms
      on the right, separated by \code{"+"} operators, are the indicators
      of the latent variable.

      The operator \code{"=~"} can be read as ``is manifested by''.

    \item Regressions: The \code{"~"} operator specifies a regression.
      The dependent variable is on the left of a \code{"~"} operator and the
      independent variables, separated by \code{"+"} operators, are on the right.
      These regression formulas are similar to the way ordinary linear regression
      formulas are used in R, but they may include latent variables. Interaction
      terms are currently not supported.

    \item Variance-covariances: The \code{"~~"} (`double tilde') operator specifies
     (residual) variances of an observed or latent variable, or a set of
     covariances between one variable and several other variables (either
     observed or latent). Several variables, separated by \code{"+"}
     operators, can appear on the right. In this way, several pairwise
     (co)variances involving the same left-hand variable can be expressed in a
     single expression. The distinction between variances and residual variances
     is made automatically.

    \item Intercepts: A special case of a regression formula can be used to
      specify an intercept (or a mean) of either an observed or a latent variable.
     The variable name is on the left of a \code{"~"} operator, and the only
     term on the right is the number \code{"1"}, representing the intercept.
     Including an intercept
     formula in the model automatically implies \code{meanstructure = TRUE}. The
     distinction between intercepts and means is made automatically.

    \item Thresholds: The \code{"|"} operator can be used to define the
      thresholds of categorical endogenous variables (on the left-hand side
      of the operator). By convention, the
      thresholds (on the right-hand side, separated by the \code{"+"} operator)
      are named \code{"t1"}, \code{"t2"}, and so on.

    \item Scaling factors: The \code{"~*~"} operator defines a scale factor.
      The variable name on the left hand side must be the same as the variable
      name on the right hand side. Scale factors are used in the Delta
      parameterization, in a multiple group analysis when factor indicators
      are categorical.

    \item Composites: The \code{"<~"} operator can be used to define
      a composite (on the right hand side of the operator).
      A composite is a weighted linear combination of its composite indicators.
      The name of the composite variable is on the left of the \code{"<~"}
      operator, while the terms on the right, separated by \code{"+"}
      operators, are the indicators of the composite variable.
}
There are 4 additional operators, also with left- and right-hand sides, that can
be included in model syntax.  Three of them are used to specify (in)equality
constraints on estimated parameters (\code{==}, \code{>}, and \code{<}), and
those are demonstrated in a later section about
\bold{(In)equality constraints}.
The final additional operator (\code{:=}) can be used to define ``new'' parameters
that are functions of one or more other estimated parameters. The \code{:=}
operator is demonstrated in a section about \bold{User-defined parameters}.

Usually, only a single variable name appears on the left side of an
operator. However, if multiple variable names are specified,
separated by the \code{"+"} operator, the formula is repeated for each
element on the left side (as for example in the third regression
formula in the example above). The only exception is scaling factors, where
only a single element is allowed on the left-hand side.

In the right-hand side of these formula-like expressions, each element can be
modified (using the \code{"*"} operator) by either a numeric constant,
an expression resulting in a numeric constant, an expression resulting
in a character vector, or one
of three special functions: \code{start()}, \code{label()} and \code{equal()}.
This provides the user with a mechanism to fix parameters, to provide
alternative starting values, to label the parameters, and to define equality
constraints among model parameters. All \code{"*"} expressions are
referred to as \emph{modifiers}. They are explained in more detail in the
following sections.
}

\section{Fixing parameters}{
It is often desirable to fix a model parameter that is otherwise (by default)
free. Any parameter in a model can be fixed by using a modifier resulting
in a numerical constant. Here are some examples:
\itemize{

    \item Fixing the regression coefficient of the predictor
       \code{x2}: \preformatted{y ~ x1 + 2.4*x2 + x3}

    \item Specifying an orthogonal (zero) covariance between two latent
      variables: \preformatted{f1 ~~ 0*f2}

    \item Specifying an intercept and a linear slope in a growth
    model: \preformatted{i =~ 1*y11 + 1*y12 + 1*y13 + 1*y14
s =~ 0*y11 + 1*y12 + 2*y13 + 3*y14}
}
Instead of a numeric constant, one can use a mathematical function that returns
a numeric constant, for example \code{sqrt(10)}. Multiplying with \code{NA}
will force the corresponding parameter to be free.

Additionally, the \code{==} operator can be used to set a \emph{labeled} parameter
equal to a specific numeric value. This will be demonstrated in the section below
about \bold{(In)equality constraints}.
}

\section{Starting values}{
User-provided starting values can be given by using the special function
\code{start()}, containing a numeric constant. For example:
\preformatted{y ~ x1 + start(1.0)*x2 + x3}
Note that if a starting value is provided, the parameter is not
automatically considered to be free.
}

\section{Parameter labels and equality constraints}{
Each free parameter in a model is automatically given a name (or label).
The name given to a model
parameter consists of three parts, coerced to a single character vector.
The first part is the name of the variable in the left-hand side of the
formula where the parameter was
implied. The middle part is based on the special `operator' used in the
formula. This can be either one of \code{"=~"}, \code{"~"} or \code{"~~"}. The
third part is the name of the variable in the right-hand side of the formula
where the parameter was implied, or \code{"1"} if it is an intercept. The three
parts are pasted together in a single string. For example, the name of the
fixed regression coefficient in the regression formula
\code{y ~ x1 + 2.4*x2 + x3} is the string \code{"y~x2"}.
The name of the parameter
corresponding to the covariance between two latent variables in the
formula \code{f1 ~~ f2} is the string \code{"f1~~f2"}.

Although this automatic labeling of parameters is convenient, the user may
specify their own labels for specific parameters simply by pre-multiplying
the corresponding term (on the right hand side of the operator only) by
a character string (starting with a letter).
For example, in the formula \code{f1 =~ x1 + x2 + mylabel*x3}, the parameter
corresponding with the factor loading of
\code{x3} will be named \code{"mylabel"}.
An alternative way to specify the label is as follows:
\code{f1 =~ x1 + x2 + label("mylabel")*x3},
where the label is the argument of special function \code{label()};
this can be useful if the label contains a space, or an operator (like "~").

There are two ways to constrain a parameter
to be equal to another target parameter. If you
have specified your own labels, you can use the fact that
\emph{equal labels imply equal parameter values}.
If you rely on automatic parameter labels, you
can use the special function \code{equal()}. The argument of
\code{equal()} is the (automatic or user-specified) name of the target
parameter. For example, in the confirmatory factor analysis example below, the
intercepts of the three indicators of each latent variable are constrained to
be equal to each other. For the first three, we have used the default
names. For the last three, we have provided a custom label for the
\code{y2a} intercept.
\preformatted{model <- '
  # two latent variables with fixed loadings
    f1 =~ 1*y1a + 1*y1b + 1*y1c
    f2 =~ 1*y2a + 1*y2b + 1*y2c

  # intercepts constrained to be equal
  # using the default names
    y1a ~ 1
    y1b ~ equal("y1a~1") * 1
    y1c ~ equal("y1a~1") * 1

  # intercepts constrained to be equal
  # using a custom label
    y2a ~ int2*1
    y2b ~ int2*1
    y2c ~ int2*1
'
}
}

\section{Multiple groups}{
In a multiple group analysis, modifiers that contain a single element
should be replaced by a vector, having the same length as the number
of groups. If you provide a single element, it will be recycled
for all the groups. This may be dangerous, in particular when the modifier
is a label. In that case, the (same) label is copied across all groups,
and this would imply an equality constraint across groups.
Therefore, when using modifiers in a multiple group setting, it is always
safer (and cleaner) to specify the same number of
elements as the number of groups. Consider this example with two groups:
\preformatted{
HS.model <- ' visual  =~ x1 + 0.5*x2 + c(0.6, 0.8)*x3
              textual =~ x4 + start(c(1.2, 0.6))*x5 + x6
              speed   =~ x7 + x8 + c(x9.group1, x9.group2)*x9 '
}
In this example, the factor loading of the `x2' indicator is fixed to the
value 0.5 for both groups. However, the factor loadings of the `x3' indicator
are fixed to 0.6 and 0.8 for group 1 and group 2 respectively. The same
logic is used for all modifiers. Note that character vectors can contain
unquoted strings.
}

\section{Multiple modifiers}{
In the model syntax, you can specify a variable more than once on the right hand
side of an operator; therefore, several `modifiers' can be applied
simultaneously; for example, if you want to fix the value of a parameter and
also label that parameter, you can use something like:
\preformatted{ f1 =~ x1 + x2 + 4*x3 + x3.loading*x3}
}

\section{Random slopes (two-level models)}{
In a two-level model (with a \code{cluster=} argument), the \code{rv()}
modifier can be used in the level-1 (within) part of the model to turn a
regression coefficient into a \emph{random slope}: a latent variable at
the second (between) level. The (quoted) argument of \code{rv()} is the
name of this latent variable. At the between level, the random slope can
then be given a (residual) variance, covariances, regressions on
between-level covariates, and an intercept (which is the `fixed effect'
of the slope). For example (cfr. ex9.8 in the Mplus User's Guide):
\preformatted{model <- '
  level: 1
    fw =~ y1 + y2 + y3 + y4
    fw ~ rv("s1")*x1 + rv("s2")*x2
  level: 2
    fb =~ y1 + y2 + y3 + y4
    s1 + s2 + fb ~ w
'
}
Here, the within-level regression coefficients of \code{fw} on the
observed covariates \code{x1} and \code{x2} vary across clusters; the
random slopes \code{s1} and \code{s2} are regressed on the between-level
covariate \code{w}.

The covariate that carries a random slope may also be a \emph{latent}
(within-level) variable, as in
\preformatted{model <- '
  level: 1
    fxw =~ x1 + x2 + x3
    fyw =~ y1 + y2 + y3
    fyw ~ rv("s1")*fxw
  level: 2
    fxb =~ x1 + x2 + x3
    fyb =~ y1 + y2 + y3
    fyb ~ fxb
    fyb ~~ s1
'
}
Such a slope multiplies two random variables, and the marginal
loglikelihood is no longer available in closed form: it is computed by
(Gauss-Hermite) quadrature over the latent-covariate slopes only,
while all other random effects are still integrated out analytically
(Rockwood, 2020). The number of quadrature points (per dimension) can
be set via the \code{integration.ngh} option (default: 21).

The same route is taken when the covariate is a `split' both-level
observed variable (an observed variable that appears in both the
level-1 and the level-2 part of the model): the random slope then
multiplies its \emph{latent} within-cluster component (latent
centering), and the covariate is jointly modeled:
\preformatted{model <- '
  level: 1
    y ~ rv("s1")*x
  level: 2
    y ~ x
    y ~~ s1
'
}
This is the latent-covariate contextual model of Rockwood (2020,
section 4). Note the contrast with a within-only covariate (a
variable that does \emph{not} appear in the level-2 part): there, the
slope multiplies the observed values, which are conditioned upon.

The current implementation is limited to: two-level models, a single
group, continuous data, \code{estimator = "ML"} and \code{fixed.x =
TRUE}; the observed covariates that carry a random slope must be
within-only (level-1) variables. Missing data can be handled by
full-information maximum likelihood (\code{missing = "ml"}); the
exogenous covariates themselves must be complete (cases with missing
values on exogenous variables are removed). The (observed-data)
loglikelihood is \emph{conditional} on \emph{all} exogenous
covariates: the covariates carrying random slopes, any other
within-only exogenous covariates (as in Mplus
\code{TYPE = TWOLEVEL RANDOM}), and the between-level exogenous
covariates. Between-only \emph{endogenous} observed variables
(level-2 outcomes, e.g., indicators of a between-level factor) are
supported; they must be pure indicators (not involved in
regressions), with a free (or fixed to a nonzero value) residual
variance, and without missing values. No
chi-square test statistic (or fit indices) are available for models
with random slopes; \code{fitMeasures()} provides the
loglikelihood-based measures only (\code{npar}, \code{ntotal},
\code{logl}, \code{aic}, \code{bic}, \code{bic2}), plus (with
\code{estimator = "MLR"}) the H0 scaling correction factor
(\code{scaling.factor.h0}), which can be used for scaled
(loglikelihood) difference tests. Standard errors can be
\code{"standard"} (observed
information) or \code{"robust.huber.white"} (sandwich; this is also
what \code{estimator = "MLR"} implies). Both \code{optim.method =
"nlminb"} and \code{optim.method = "em"} are supported; with complete
data, \code{"nlminb"} is the default, while with \code{missing =
"ml"}, the EM algorithm is the default optimizer.
\code{lavPredict()} provides empirical Bayes (posterior mean)
predictions: with \code{level = 2}, the
between-level latent variables including the random slopes (and, with
\code{se = "standard"}, their posterior standard deviations); with
\code{level = 1} (the default), the within-level factor scores. For
latent-covariate models, these are EAP (expected a posteriori)
estimates: the posterior is a mixture over the quadrature nodes.

For random slopes with latent covariates, some additional restrictions
apply (for now): a random-slope label may not be attached to both an
observed and a latent covariate, and the optimizer is always
\code{"nlminb"}. The latent-covariate slopes may be regressed on
between-level covariates (e.g., \code{s1 ~ w} in the level-2 part),
as in the PISA example of Rockwood (2020).
}

\section{(In)equality constraints}{
The \code{==} operator can be used either to fix a parameter to a specific value,
or to set an estimated parameter equal to another parameter. Adapting the
example in the \bold{Parameter labels and equality constraints} section, we
could have used different labels for the second factor's intercepts:
\preformatted{
    y2a ~ int1*1
    y2b ~ int2*1
    y2c ~ int3*1
}
Then, we could fix the first intercept to zero by including in the syntax an
operation that indicates the parameter's label equals that value:
\preformatted{
    int1 == 0
}
Whereas we could still estimate the other two intercepts under an equality
constraint by setting their different labels equal to each other:
\preformatted{
    int2 == int3
}
Optimization can be less efficient when constraining parameters this way (see
the documentation linked under \bold{See also} for more information). But the
flexibility might be advantageous.  For example, the constraints could be
specified in a separate character-string object, which can be passed to the
\code{lavaan(..., constraints=)} argument, enabling users to compare results
with(out) the constraints.

Inequality constraints work in much the same way, using the \code{<} or \code{>}
operator to indicate which estimated parameter is hypothesized to be greater or
less than either a specific value or another estimated parameter. For example, a
variance can be constrained to be nonnegative:
\preformatted{
    y1a ~~ var1a*y1a
    ## hypothesized constraint:
    var1a > 0
}
Or the factor loading of a particular indicator might be expected to exceed
other indicators' loadings:
\preformatted{
    f1 =~ L1*y1a + L2*y1b + L3*y1c
    ## hypothesized constraints:
    L1 > L2
    L3 < L1
}
}

\section{User-defined parameters}{
Functions of parameters can be useful to test particular hypotheses. Following
from the \code{Multiple groups} example, we might be interested in which group's
factor loading is larger (i.e., an estimate of differential item functioning
(DIF) when the latent scales are linked by anchor items with equal loadings).
\preformatted{
  speed =~ c(L7, L7)*x7 + c(L8, L8)*x8 + c(L9.group1, L9.group2)*x9 '
  ## user-defined parameter:
  DIF_L9 := L9.group1 - L9.group2
}
Note that this hypothesis is easily tested without a user-defined parameter by
using the \code{lavTestWald()} function. However, a user-defined parameter
additionally provides an estimate of the parameter being tested.

User-defined parameters are particularly useful for specifying indirect effects
in models of mediation. For example:
\preformatted{
model <- ' # direct effect
             Y ~ c*X
           # mediator
             M ~ a*X
             Y ~ b*M

# user defined parameters:

    # indirect effect (a*b)
             ab := a*b
    # total effect (defined using another user-defined parameter)
             total := ab + c
'
}
}

\references{
Rockwood, N. J. (2020). Maximum likelihood estimation of multilevel
structural equation models with random slopes for latent covariates.
\emph{Psychometrika, 85}(2), 275--300.
\doi{10.1007/s11336-020-09702-9}

Rosseel, Y. (2012). \code{lavaan}: An R package for structural equation
modeling. \emph{Journal of Statistical Software, 48}(2), 1--36.
\doi{10.18637/jss.v048.i02}
}

\name{lavTestScore}
\alias{lavTestScore}
\alias{lavtestscore}
\alias{score}
\alias{Score}
\alias{lavScoreTest}
\title{Score test}
\description{
Score test (or Lagrange Multiplier test) for releasing one or more
fixed or constrained parameters in the model.}
\usage{
lavTestScore(object, add = NULL, release = NULL,
             univariate = TRUE, cumulative = FALSE,
             epc = FALSE, standardized = epc, cov_std = epc,
             verbose = FALSE, warn = TRUE, information = "expected", ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{add}{Either a character string (typically between single quotes) or
  a parameter table containing additional (currently fixed-to-zero) parameters
  for which the score test must be computed.}
\item{release}{Vector of integers. The indices of the constraints that
should be released. The indices correspond to the order in which the equality
constraints appear in the parameter table.}
\item{univariate}{Logical. If \code{TRUE}, compute the univariate score
statistics, one for each constraint.}
\item{cumulative}{Logical. If \code{TRUE}, order the univariate score
statistics from large to small, and compute a series of
multivariate score statistics, each time adding an additional constraint.}
\item{epc}{Logical. If \code{TRUE}, and we are releasing existing constraints,
compute the expected parameter changes for the existing (free) parameters,
for each released constraint.}
\item{standardized}{If \code{TRUE}, two extra columns (sepc.lv and sepc.all)
in the \code{$epc} table
will contain standardized values for the EPCs. In the first column (sepc.lv),
standardization is based on the variances of the (continuous) latent
variables. In the second column (sepc.all), standardization is based
on the variances of both (continuous) observed and latent variables.
(Residual) covariances are standardized using (residual) variances.}
\item{cov_std}{Logical. See \code{\link{standardizedSolution}}.}
\item{verbose}{Logical. Not used for now.}
\item{warn}{Logical. If \code{TRUE}, print out warnings if they occur.}
\item{information}{\code{character} indicating the type of information matrix
to use (check \code{\link{lavInspect}} for available options). \code{"expected"}
information is the default, which provides better control of Type I errors.}
\item{...}{to allow use of old argumentname cov_std}
}
\details{
    This function can be used to compute both multivariate and univariate
    score tests. There are two modes: 1) releasing fixed-to-zero parameters
    (using the \code{add} argument), and 2) releasing existing equality
    constraints (using the \code{release} argument). The two modes cannot
    be used simultaneously.

    When adding new parameters, they should not already be part of the model
    (i.e. not listed in the parameter table). If you want to test for
    a parameter that was explicitly fixed to a constant (say to zero),
    it is better to label the parameter, and use an explicit equality
    constraint.

    In addition to the standard (normal-theory) score test, robust
    versions are computed whenever the fitted object provides the
    ingredients for a sandwich-type covariance matrix: either because
    robust standard errors were requested (e.g.,
    \code{se = "robust.sem"} or \code{se = "robust.huber.white"}), or
    because a scaled test statistic was requested (e.g.,
    \code{test = "satorra.bentler"}). Following Satorra (2000), three
    robust versions are reported: a mean-scaled statistic
    (\code{score.scaled}), a mean-and-variance adjusted statistic with
    fractional degrees of freedom (\code{score.adjusted}), and the
    generalized, asymptotically distribution-free statistic
    (\code{score.robust}). For a single restriction, the three versions
    coincide. The univariate (and cumulative) score tests are then
    accompanied by scaled counterparts in the \code{X2.scaled} column.
    Note that for estimators that are not asymptotically efficient
    (e.g., \code{DWLS}, \code{ULS} and \code{PML}), only the robust
    versions have the correct asymptotic distribution.
}
\value{
  A list containing at least one \code{data.frame}:
  \itemize{
    \item \code{$test}: The total score test, with columns for the score
      test statistic (\code{X2}), the degrees of freedom (\code{df}), and
      a \emph{p} value under the \eqn{\chi^2} distribution (\code{p.value}).
      If robust versions could be computed, three additional rows are
      included: \code{score.scaled}, \code{score.adjusted} and
      \code{score.robust} (see Details).
    \item \code{$uni}: Optional (if \code{univariate=TRUE}).
      Each 1-\emph{df} score test, equivalent to modification indices.
      If robust versions could be computed, the scaled statistics are
      provided in the \code{X2.scaled} and \code{p.value.scaled} columns.
      If \code{epc=TRUE} when \code{add}ing parameters (not when releasing
      constraints), an unstandardized EPC is provided for each added
      parameter, as would be returned by \code{\link{modificationIndices}}.
    \item \code{$cumulative}: Optional (if \code{cumulative=TRUE}).
      Cumulative score tests, with scaled counterparts if robust versions
      could be computed.
    \item \code{$epc}: Optional (if \code{epc=TRUE}). Parameter estimates,
      expected parameter changes, and expected parameter values if all
      the tested constraints were freed.
  }
}
\references{
Bentler, P. M., & Chou, C. P. (1993). Some new covariance structure model
improvement statistics. Sage Focus Editions, 154, 235-255.

Satorra, A. (2000). Scaled and adjusted restricted tests in multi-sample
analysis of moment structures. In Heijmans, R.D.H., Pollock, D.S.G. &
Satorra, A. (Eds.), Innovations in multivariate statistical analysis:
A festschrift for Heinz Neudecker (pp. 233-247). London: Kluwer Academic
Publishers.
}

\examples{
HS.model <- '
    visual  =~ x1 + b1*x2 + x3
    textual =~ x4 + b2*x5 + x6
    speed   =~ x7 + b3*x8 + x9

    b1 == b2
    b2 == b3
'
fit <- cfa(HS.model, data=HolzingerSwineford1939)

# test 1: release both two equality constraints
lavTestScore(fit, cumulative = TRUE)

# test 2: the score test for adding two (currently fixed
# to zero) cross-loadings
newpar = '
    visual =~ x9
    textual =~ x3
'
lavTestScore(fit, add = newpar)

# equivalently, "add" can be a parameter table specifying parameters to free,
# but must include some additional information:
PT.add <- data.frame(lhs = c("visual","textual"),
                     op = c("=~","=~"),
                     rhs = c("x9","x3"),
                     user = 10L, # needed to identify new parameters
                     free = 1, # arbitrary numbers > 0
                     start = 0) # null-hypothesized value
PT.add
lavTestScore(fit, add = PT.add) # same result as above

}

\name{sem}
\alias{sem}
\title{Fit Structural Equation Models}
\description{
Fit a Structural Equation Model (SEM).}
\usage{
sem(model = NULL, data = NULL, ordered = NULL, aux = NULL, sampling_weights = NULL,
    sample_cov = NULL, sample_mean = NULL, sample_th = NULL,
    sample_nobs = NULL, group = NULL, cluster = NULL, 
    constraints = "", wls_v = NULL, nacov = NULL, ov_order = "model",
    ...)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See 
    \code{\link{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}
\item{data}{An optional data frame containing the observed variables used in
    the model. If some variables are declared as ordered factors, lavaan will
    treat them as ordinal variables.}
\item{ordered}{Character vector. Only used if the data is in a data.frame.
    Treat these variables as ordered (ordinal) variables, if they are
    endogenous in the model. Importantly, all other variables will be treated
    as numeric (unless they are declared as ordered in the data.frame.)
    Since 0.6-4, ordered can also be logical. If TRUE, all
    observed endogenous variables are treated as ordered (ordinal).
    If FALSE, all observed endogenous variables are considered to be numeric
    (again, unless they are declared as ordered in the data.frame.)}
\item{aux}{Character vector. Names of auxiliary observed variables, used to
    make the missing-at-random (MAR) assumption more plausible under missing
    data (continuous data only). With \code{missing = "ml"} they are added to
    the model as saturated correlates (Graham, 2003), affecting the estimates
    and standard errors but hidden from the default output; with
    \code{missing = "two.stage"}/\code{"robust.two.stage"} they enter the
    first-stage EM moments. See the \code{lavaan} function for more details.}
\item{sampling_weights}{A variable name in the data frame containing 
    sampling weight information. Currently only available for non-clustered
    data. Depending on the \code{sampling.weights.normalization} option, these
    weights may be rescaled (or not) so that their sum equals the number of
    observations (total or per group).}
\item{sample_cov}{Numeric matrix. A sample variance-covariance matrix.
    The rownames and/or colnames must contain the observed variable names.
    For a multiple group analysis, a list with a variance-covariance matrix
    for each group.}
\item{sample_mean}{A sample mean vector. For a multiple group analysis,
    a list with a mean vector for each group.}
\item{sample_th}{Vector of sample-based thresholds. For a multiple group 
    analysis, a list with a vector of thresholds for each group.}
\item{sample_nobs}{Number of observations if the full data frame is missing
    and only sample moments are given. For a multiple group analysis, a list
    or a vector with the number of observations for each group.}
\item{group}{Character. A variable name in the data frame defining the groups
in a multiple group analysis.}
\item{cluster}{Character. A (single) variable name in the data frame defining the clusters in a two-level dataset.}
\item{constraints}{Additional (in)equality constraints not yet included in the
    model syntax. See \code{\link{model.syntax}} for more information.}
\item{wls_v}{A user provided weight matrix to be used by estimator \code{"WLS"};
    if the estimator is \code{"DWLS"}, only the diagonal of this matrix will be
    used. For a multiple group analysis, a list with a weight matrix
    for each group. The elements of the weight matrix should be in the
    following order (if all data is continuous): first the means (if a
    meanstructure is involved), then the lower triangular elements of the
    covariance matrix including the diagonal, ordered column by column. In
    the categorical case: first the thresholds (including the means for
    continuous variables), then the slopes (if any), the variances of
    continuous variables (if any), and finally the lower triangular elements
    of the correlation/covariance matrix excluding the diagonal, ordered
    column by column.}
\item{nacov}{A user provided matrix containing the elements of (N times)
    the asymptotic variance-covariance matrix of the sample statistics.
    For a multiple group analysis, a list with an asymptotic
    variance-covariance matrix for each group. See the \code{wls_v}
    argument for information about the order of the elements.}
\item{ov_order}{Character. If \code{"model"} (the default), the order of
    the observed variable names (as reflected for example in the output of
    \code{lav_object_vnames()}) is determined by the model syntax. If
    \code{"data"}, the order is determined by the data (either the full
    data.frame or the sample (co)variance matrix). If the \code{wls_v}
    and/or \code{nacov} matrices are provided, this argument is currently
    set to \code{"data"}.}
\item{...}{Many more options can be specified, using 'name = value'.
    See \code{\link{lavOptions}} for a complete list.}
}
\details{
    The \code{sem} function is a wrapper for the more general
    \code{\link{lavaan}} function, using the following default arguments:
    \code{int.ov.free = TRUE}, \code{int.lv.free = FALSE},
    \code{auto.fix.first = TRUE} (unless \code{std.lv = TRUE}),
    \code{auto.fix.single = TRUE}, \code{auto.var = TRUE},
    \code{auto.cov.lv.x = TRUE}, \code{auto.efa = TRUE},
    \code{auto.th = TRUE}, \code{auto.delta = TRUE},
    and \code{auto.cov.y = TRUE}.
}
\value{
  An object of class \code{\linkS4class{lavaan}}, for which several methods
  are available, including a \code{summary} method.
}
\references{
Yves Rosseel (2012). lavaan: An R Package for Structural Equation
Modeling. Journal of Statistical Software, 48(2), 1-36. 
\doi{10.18637/jss.v048.i02}}
\seealso{
\code{\link{lavaan}}, \code{\link{estimator_iv}}
}
\examples{
## The industrialization and Political Democracy Example 
## Bollen (1989), page 332
model <- ' 
  # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + a*y2 + b*y3 + c*y4
     dem65 =~ y5 + a*y6 + b*y7 + c*y8

  # regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
'

fit <- sem(model, data = PoliticalDemocracy)
summary(fit, fit.measures = TRUE)
}

\name{varTable}
\alias{varTable}
\alias{vartable}
\alias{variableTable}
\alias{variabletable}
\title{Variable Table}
\description{
Summary information about the variables included in either a data.frame,
or a fitted lavaan object.}
\usage{
varTable(object, ov_names = names(object), ov_names_x = NULL,
    ordered = NULL, factor = NULL, as_data_frame = TRUE, ...)
}
\arguments{
\item{object}{Either a data.frame, or an object of class
\code{\linkS4class{lavaan}}.}
\item{ov_names}{Only used if object is a data.frame. A character vector containing the variables that need to be summarized.}
\item{ov_names_x}{Only used if object is a data.frame. A character vector containing additional variables that need to be summarized.}
\item{ordered}{Character vector. Which variables should be treated as ordered
factors?}
\item{factor}{Character vector. Which variables should be treated as (unordered) factors?}
\item{as_data_frame}{If TRUE, return the list as a data.frame.}
\item{...}{To support old argument names.}
}
\value{
  A \code{list} or \code{data.frame} containing summary information about
  variables in a data.frame. If \code{object} is a fitted lavaan object,
  it displays the summary information about the observed variables that are
  included in the model. The summary information includes
  variable type (numeric, ordered, \ldots), the number of non-missing values,
  the mean and variance for numeric variables, the number of levels of
  ordered variables, and the labels for ordered variables.
}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
varTable(fit)
}

\name{lav_data}
\alias{lav_data_update}
\title{lavaan data functions}
\description{Utility functions related to the Data slot}
\usage{
# update data slot with new data (of the same size)
lav_data_update(lavdata = NULL, new_x = NULL, boot_idx = NULL, boot_clus = NULL,
                lavoptions = NULL, ...)
}
\arguments{
\item{lavdata}{A lavdata object.}
\item{new_x}{A list of (new) data matrices (per group) of the same size.
They replace the data stored in the internal data slot.}
\item{boot_idx}{A list of integers. If bootstrapping was used to produce
the data in newX, use these indices to adapt the remaining slots.}
\item{boot_clus}{A list of integer matrices (per group), or NULL. For a
cluster bootstrap of two-level data, the (relabeled) cluster ids of the
resampled rows, used to adapt the multilevel (Lp) slots.}
\item{lavoptions}{A named list. The Options slot from a lavaan object.}
\item{...}{To accept old argument names with dots or capitals. No other arguments are accepted.}
}
\examples{
# generate syntax for an independence model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)

# extract data slot and options
lavdata <- fit@Data
lavoptions <- lavInspect(fit, "options")

# create bootstrap sample
boot.idx <- sample(x = nobs(fit), size = nobs(fit), replace = TRUE)
newX <- list(lavdata@X[[1]][boot.idx,])

# generate update lavdata object
newdata <- lav_data_update(lavdata = lavdata, new_x = newX,
                           lavoptions = lavoptions)
}

\name{Demo.twolevel}
\alias{Demo.twolevel}
\docType{data}
\title{
Demo dataset for illustrating a multilevel CFA
}
\description{
A toy dataset containing measures on 6 items (y1-y6), 3 within-level 
covariates (x1-x3) and 2 between-level covariates (w1-w2). The data
is clustered (200 clusters of size 5, 10, 15 and 20), and the cluster 
variable is \dQuote{cluster}.
}
\usage{data(Demo.twolevel)}
\format{
A data frame of 2500 observations of 12 variables.
\describe{
    \item{\code{y1}}{item 1}
    \item{\code{y2}}{item 2}
    \item{\code{y3}}{item 3}
    \item{\code{y4}}{item 4}
    \item{\code{y5}}{item 5}
    \item{\code{y6}}{item 6}
    \item{\code{x1}}{within-level covariate 1}
    \item{\code{x2}}{within-level covariate 2}
    \item{\code{x3}}{within-level covariate 3}
    \item{\code{w1}}{between-level covariate 1}
    \item{\code{w2}}{between-level covariate 2}
    \item{\code{cluster}}{cluster variable}
  }
}
\examples{
head(Demo.twolevel)

model <- '
    level: 1
        fw =~ y1 + y2 + y3
        fw ~ x1 + x2 + x3
    level: 2
        fb =~ y1 + y2 + y3
        fb ~ w1 + w2
'
fit <- sem(model, data = Demo.twolevel, cluster = "cluster")
summary(fit)
}

\name{lavSimulateData}
\alias{lavSimulateData}
\alias{lav_data_simulate_old}
\alias{simulateData}
\title{Simulate Data From a Lavaan Model Syntax}
\description{Simulate data starting from a lavaan model syntax.}
\usage{
lavSimulateData(model = NULL, model_type = "sem", meanstructure = FALSE,
    int_ov_free = TRUE, int_lv_free = FALSE,
    marker_int_zero = FALSE, conditional_x = FALSE,
    composites = TRUE, fixed_x = FALSE,
    orthogonal = FALSE, std_lv = TRUE, auto_fix_first = FALSE,
    auto_fix_single = FALSE, auto_var = TRUE, auto_cov_lv_x = TRUE,
    auto_cov_y = TRUE, ..., sample_nobs = 500L, ov_var = NULL,
    group_label = NULL, skewness = NULL,
    kurtosis = NULL, cluster_idx = NULL, seed = NULL, empirical = FALSE,
    mass = TRUE, ordered_center = TRUE, return_type = "data.frame",
    return_fit = FALSE, debug = FALSE, standardized = FALSE)
simulateData(model = NULL, model_type = "sem", meanstructure = FALSE,
    int_ov_free = TRUE, int_lv_free = FALSE,
    marker_int_zero = FALSE, conditional_x = FALSE,
    composites = TRUE, fixed_x = FALSE,
    orthogonal = FALSE, std_lv = TRUE, auto_fix_first = FALSE,
    auto_fix_single = FALSE, auto_var = TRUE, auto_cov_lv_x = TRUE,
    auto_cov_y = TRUE, ..., sample_nobs = 500L, ov_var = NULL,
    group_label = NULL, skewness = NULL,
    kurtosis = NULL, cluster_idx = NULL, seed = NULL, empirical = FALSE,
    mass = FALSE, ordered_center = TRUE, return_type = "data.frame",
    return_fit = FALSE, debug = FALSE, standardized = FALSE)
lav_data_simulate_old(..., ordered_center = FALSE)
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See
    \code{\link{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}
\item{model_type}{Set the model type: possible values
    are \code{"cfa"}, \code{"sem"} or \code{"growth"}. This may affect
    how starting values are computed, and may be used to alter the terminology
    used in the summary output, or the layout of path diagrams that are
    based on a fitted lavaan object.}
\item{meanstructure}{If \code{TRUE}, the means of the observed
    variables enter the model. If \code{"default"}, the value is set based
    on the user-specified model, and/or the values of other arguments.}
\item{int_ov_free}{If \code{FALSE}, the intercepts of the observed variables
    are fixed to zero.}
\item{int_lv_free}{If \code{FALSE}, the intercepts of the latent variables
    are fixed to zero.}
\item{marker_int_zero}{Logical. Only relevant if the metric of each latent
    variable is set by fixing the first factor loading to unity.
    If \code{TRUE}, it implies \code{meanstructure = TRUE} and 
    \code{std_lv = FALSE}, and it fixes the intercepts of the marker
    indicators to zero, while freeing the means/intercepts of the latent
    variables. Only works correctly for single group, single level models.}
\item{conditional_x}{If \code{TRUE}, we set up the model conditional on
    the exogenous `x' covariates; the model-implied sample statistics
    only include the non-x variables. If \code{FALSE}, the exogenous `x'
    variables are modeled jointly with the other variables, and the
    model-implied statistics reflect both sets of variables. If
    \code{"default"}, the value is set depending on the estimator, and
    whether or not the model involves categorical endogenous variables.}
\item{composites}{If \code{TRUE}, use the new (0.6-20) approach
    for handling composites.}
\item{fixed_x}{If \code{TRUE}, the exogenous `x' covariates are considered
    fixed variables and the means, variances and covariances of these variables
    are fixed to their sample values. If \code{FALSE}, they are considered
    random, and the means, variances and covariances are free parameters. If
    \code{"default"}, the value is set depending on the mimic option.}
\item{orthogonal}{If \code{TRUE}, the exogenous latent variables
    are assumed to be uncorrelated.}
\item{std_lv}{If \code{TRUE}, the metric of each latent variable is
    determined by fixing their variances to 1.0. If \code{FALSE}, the metric
    of each latent variable is determined by fixing the factor loading of the
    first indicator to 1.0.}
\item{auto_fix_first}{If \code{TRUE}, the factor loading of the first indicator
    is set to 1.0 for every latent variable.}
\item{auto_fix_single}{If \code{TRUE}, the residual variance (if included)
    of an observed indicator is set to zero if it is the only indicator of a
    latent variable.}
\item{auto_var}{If \code{TRUE}, the (residual) variances of both observed
    and latent variables are set free.}
\item{auto_cov_lv_x}{If \code{TRUE}, the covariances of exogenous latent
    variables are included in the model and set free.}
\item{auto_cov_y}{If \code{TRUE}, the covariances of dependent variables
    (both observed and latent) are included in the model and set free.}
\item{\dots}{additional arguments passed to the \code{\link{lavaan}}
    function.}
\item{sample_nobs}{Number of observations. If a vector, multiple datasets
    are created. If \code{return_type = "matrix"} or 
    \code{return_type = "cov"}, a list of \code{length(sample_nobs)} 
    is returned, with either the data or covariance matrices, each one
    based on the number of observations as specified in \code{sample_nobs}.
    If \code{return_type = "data.frame"}, all datasets are merged and 
    a \code{group} variable is added to mimic a multiple group dataset.}
\item{ov_var}{The user-specified variances of the observed variables.}
\item{group_label}{The group labels that should be used if multiple
    groups are created.}
\item{skewness}{Numeric vector. The skewness values for the observed variables. Defaults to zero.}
\item{kurtosis}{Numeric vector. The kurtosis values for the observed variables. Defaults to zero.}
\item{cluster_idx}{Optional. Only used (and only available via
    \code{lavSimulateData()} and \code{simulateData()}) for multilevel models.
    An integer vector (or a list of integer vectors, one per group) giving the
    cluster membership of each level-1 unit. If supplied, it determines both
    the number of level-1 units and the number of clusters. If not supplied for
    a multilevel model, balanced clusters are created based on \code{sample_nobs}.
    Providing this argument (or a multilevel model syntax with \code{level:}
    blocks) routes the request to the multilevel data-generation engine.}
\item{seed}{Set random seed.}
\item{empirical}{Logical. If \code{TRUE}, the implied moments (Mu and Sigma)
    specify the empirical not population mean and covariance matrix.}
\item{mass}{Logical. If \code{TRUE}, use the mvrnorm() function from the
    MASS package (the default before 0.7-1) instead of the internal
    lav_mvrnorm() function.}
\item{ordered_center}{Logical. Only relevant for categorical data, which is
    generated by the (new) multilevel-aware engine. If \code{TRUE} (the
    default), each ordered variable is centered (its sample mean is subtracted)
    before it is cut at the model-implied thresholds; this also works when the
    model-implied mean is nonzero. If \code{FALSE}, the variable is cut at the
    thresholds directly (the behaviour of the older engine).}
\item{return_type}{If \code{"data.frame"}, a data.frame is returned. If
    \code{"matrix"}, a numeric matrix is returned (without any variable names).
    If \code{"cov"}, a covariance matrix is returned (without any variable 
    names).}
\item{return_fit}{If \code{TRUE}, return the fitted model that has been used
    to generate the data as an attribute (called \code{"fit"}); this 
    may be useful for inspection.}
\item{debug}{If \code{TRUE}, debugging information is displayed.}
\item{standardized}{If \code{TRUE}, the residual variances of the observed
    variables are set in such a way that the model implied variances
    are unity. This allows regression coefficients and factor loadings 
    (involving observed variables) to be specified in a standardized metric.}
}
\details{Model parameters can be specified by fixed values in the lavaan model
syntax. If no fixed values are specified, the value zero will be assumed,
except for factor loadings and variances, which are set to 0.7 and 1.0
respectively. By default, multivariate normal data are generated. However, by
providing skewness and/or kurtosis values, nonnormal multivariate data can be
generated, using the Vale & Maurelli (1983) method.

There is a single data-simulation engine. Multilevel data (model syntax
containing \code{level:} blocks, or when the \code{cluster_idx} argument is
provided) are generated by an internal multilevel worker; all single-level data
(continuous and categorical, including the \code{ov_var}, \code{skewness},
\code{kurtosis}, \code{standardized} and \code{mass} options) are generated by
the historical single-level worker, so that the continuous, single-level output
remains byte-identical to previous versions. For multilevel data, the
\code{ov_var}, \code{skewness}, \code{kurtosis}, \code{standardized} and
\code{mass} arguments are not (yet) supported and are ignored (with a warning).

\code{lav_data_simulate_old()} is a deprecated wrapper kept for backward
compatibility; it forwards to the unified engine (with \code{ordered_center =
FALSE} by default, reproducing the historical threshold cut).}
\value{The generated data. Either as a data.frame 
(if \code{return_type="data.frame"}), 
a numeric matrix (if \code{return_type="matrix"}),
or a covariance matrix (if \code{return_type="cov"}).}
\examples{
# specify population model
population.model <- ' f1 =~ x1 + 0.8*x2 + 1.2*x3
                      f2 =~ x4 + 0.5*x5 + 1.5*x6
                      f3 =~ x7 + 0.1*x8 + 0.9*x9

                      f3 ~ 0.5*f1 + 0.6*f2
                    '

# generate data
set.seed(1234)
myData <- lavSimulateData(population.model, sample_nobs = 100L)

# population moments
fitted(sem(population.model))

# sample moments
round(cov(myData), 3)
round(colMeans(myData), 3)

# fit model
myModel <- ' f1 =~ x1 + x2 + x3
             f2 =~ x4 + x5 + x6
             f3 =~ x7 + x8 + x9
             f3 ~ f1 + f2 '
fit <- sem(myModel, data=myData)
summary(fit)
}

\name{lavTest}
\alias{lavTest}
\alias{lavtest}
\title{Test of exact fit}
\description{
Compute a variety of test statistics evaluating the global fit of
the model.}
\usage{
    lavTest(lavobject, test = "standard", scaled_test = "standard",
            output = "list", drop_list_single = TRUE, ...)
}
\arguments{
\item{lavobject}{An object of class \code{\linkS4class{lavaan}}.}
\item{test}{Character vector. Multiple names of test statistics can be provided.
    If \code{"standard"} is included, a conventional chi-square test
    is computed. If \code{"Browne.residual.adf"} is included,
    Browne's residual-based test statistic using ADF theory is computed.
    If \code{"Browne.residual.nt"} is included, Browne's residual-based
    test statistic using normal theory is computed.
    If \code{"Satorra.Bentler"} is included, a Satorra-Bentler scaled test
    statistic is computed. If \code{"Yuan.Bentler"} is included,
    a Yuan-Bentler scaled test statistic is computed.
    If \code{"Yuan.Bentler.Mplus"} is included, a
    test statistic is computed that is asymptotically equivalent to the
    Yuan-Bentler scaled test statistic. If \code{"mean.var.adjusted"} or
    \code{"Satterthwaite"} is included, a mean and variance adjusted test
    statistic is computed. If \code{"scaled.shifted"} is included,
    an alternative mean and variance adjusted test statistic is
    computed (as in Mplus version 6 or higher).
    If \code{"mean.var.adjusted.corrected"} (aliases: \code{"mvc"},
    \code{"hayakawa"}) or \code{"scaled.shifted.corrected"} (alias:
    \code{"ssc"}) is included, the corresponding test statistic is computed
    with the corrected (unbiased) estimator of the trace of the squared
    UGamma matrix (Hayakawa, 2018), which remains accurate when the number
    of observed variables is large relative to the sample size (where the
    naive estimator is severely biased, and the uncorrected adjusted tests
    underreject). These corrected tests are only available for single-group,
    single-level, complete, continuous data and \code{estimator = "ML"}.
    Following Hayakawa (2018), they are best combined with
    \code{scaled_test = "RLS"}.
    If \code{"boot"} or \code{"bootstrap"} or \code{"Bollen.Stine"} is
    included, the
    Bollen-Stine bootstrap is used to compute the bootstrap probability value
    of the (regular) test statistic. FMG p-value tests can be requested using
    names such as \code{"peba4"}, \code{"pols3"}, \code{"pall"}, or
    \code{"all"}. The convenience value \code{"fmg"} resolves to
    \code{"peba4_rls"} for one-model tests. Suffixless FMG names use the
    standard ML chi-square statistic by default; set
    \code{scaled_test = "Browne.residual.nt.model"} or
    \code{scaled_test = "RLS"} to use the Browne residual normal-theory
    model-based statistic. Suffixes such as \code{"_ml"}, \code{"_rls"},
    and \code{"_ug"} can be used for explicit control, for example
    \code{"peba4_ug_rls"} or \code{"pols3_ml"}.}
\item{scaled_test}{Character. Choose the test statistic
    that will be scaled (if a scaled test statistic is requested).
    The default is \code{"standard"}, but it could also be (for example)
    \code{"Browne.residual.nt"}.}
\item{output}{Character. If \code{"list"} (the default), return a list with
    all test statistics. If \code{"text"}, display the output as text with
    verbose descriptions (as in the summary output). If any scaled
    test statistics are included, they are printed first in a two-column
    format, followed by the remaining test statistics in a one-column format.}
\item{drop_list_single}{Logical. Only used when \code{output = "list"}.
    If \code{TRUE} and the list is of length one (i.e. only a single test
    statistic), drop the outer list. If \code{FALSE}, return a nested list
    with one element per test statistic.}
\item{...}{To support old argument names.}
}
\value{
    If \code{output = "list"}: a nested list with test statistics, or if
    only a single test statistic is requested (and
    \code{drop_list_single = TRUE}), a list with details for this test
    statistic. If \code{output = "text"}: the text is printed, and a
    nested list of test statistics (including an info attribute) is
    returned.
}
\references{
Hayakawa, K. (2019). Corrected goodness-of-fit test in covariance
structure analysis. \emph{Psychological Methods, 24}(3), 371--389.
\doi{10.1037/met0000180}

Himeno, T., & Yamada, T. (2014). Estimations for some functions of
covariance matrix in high dimension under non-normality and its
applications. \emph{Journal of Multivariate Analysis, 130}, 27--44.

Srivastava, M. S. (2005). Some tests concerning the covariance matrix in
high dimensional data. \emph{Journal of the Japan Statistical Society,
35}(2), 251--272.
}
\examples{
HS.model <- '
    visual  =~ x1 + x2 + x3
    textual =~ x4 + x5 + x6
    speed   =~ x7 + x8 + x9
'
fit <- cfa(HS.model, data = HolzingerSwineford1939)
lavTest(fit, test = "browne.residual.adf")
lavTest(fit, test = "peba4", output = "text")
lavTest(fit, test = "pols3", scaled_test = "RLS", output = "text")
# corrected adjusted test (Hayakawa, 2018), based on the RLS statistic
lavTest(fit, test = "mean.var.adjusted.corrected", scaled_test = "RLS",
        output = "text")
}

\name{growth}
\alias{growth}
\title{Fit Growth Curve Models}
\description{
Fit a Growth Curve model. Only useful if all the latent variables in the model
are growth factors. For more complex models, it may be better to use the
\code{\link{lavaan}} function.}
\usage{
growth(model = NULL, data = NULL, ordered = NULL, aux = NULL, sampling_weights = NULL,
       sample_cov = NULL, sample_mean = NULL, sample_th = NULL,
       sample_nobs = NULL, group = NULL, cluster = NULL, 
       constraints = "", wls_v = NULL, nacov = NULL, ov_order = "model",
       ...) 
}
\arguments{
\item{model}{A description of the user-specified model. Typically, the model
    is described using the lavaan model syntax. See 
    \code{\link{model.syntax}} for more information. Alternatively, a
    parameter table (e.g., the output of the \code{lavParTable()} function) is also
    accepted.}
\item{data}{An optional data frame containing the observed variables used in
    the model. If some variables are declared as ordered factors, lavaan will
    treat them as ordinal variables.}
\item{ordered}{Character vector. Only used if the data is in a data.frame.
    Treat these variables as ordered (ordinal) variables, if they are
    endogenous in the model. Importantly, all other variables will be treated
    as numeric (unless they are declared as ordered in the data.frame.) 
    Since 0.6-4, ordered can also be logical. If TRUE, all
    observed endogenous variables are treated as ordered (ordinal).
    If FALSE, all observed endogenous variables are considered to be numeric
    (again, unless they are declared as ordered in the data.frame.)}
\item{aux}{Character vector. Names of auxiliary observed variables, used to
    make the missing-at-random (MAR) assumption more plausible under missing
    data (continuous data only). With \code{missing = "ml"} they are added to
    the model as saturated correlates (Graham, 2003), affecting the estimates
    and standard errors but hidden from the default output; with
    \code{missing = "two.stage"}/\code{"robust.two.stage"} they enter the
    first-stage EM moments. See the \code{lavaan} function for more details.}
\item{sampling_weights}{A variable name in the data frame containing 
    sampling weight information. Currently only available for non-clustered
    data. Depending on the \code{sampling.weights.normalization} option, these
    weights may be rescaled (or not) so that their sum equals the number of
    observations (total or per group).}
\item{sample_cov}{Numeric matrix. A sample variance-covariance matrix.
    The rownames and/or colnames must contain the observed variable names.
    For a multiple group analysis, a list with a variance-covariance matrix
    for each group.}
\item{sample_mean}{A sample mean vector. For a multiple group analysis,
    a list with a mean vector for each group.}
\item{sample_th}{Vector of sample-based thresholds. For a multiple group 
    analysis, a list with a vector of thresholds for each group.}
\item{sample_nobs}{Number of observations if the full data frame is missing
    and only sample moments are given. For a multiple group analysis, a list
    or a vector with the number of observations for each group.}
\item{group}{Character. A variable name in the data frame defining the groups
in a multiple group analysis.}
\item{cluster}{Character. A (single) variable name in the data frame defining the clusters in a two-level dataset.}
\item{constraints}{Additional (in)equality constraints not yet included in the
    model syntax. See \code{\link{model.syntax}} for more information.}
\item{wls_v}{A user provided weight matrix to be used by estimator \code{"WLS"};
    if the estimator is \code{"DWLS"}, only the diagonal of this matrix will be
    used. For a multiple group analysis, a list with a weight matrix
    for each group. The elements of the weight matrix should be in the
    following order (if all data is continuous): first the means (if a
    meanstructure is involved), then the lower triangular elements of the
    covariance matrix including the diagonal, ordered column by column. In
    the categorical case: first the thresholds (including the means for
    continuous variables), then the slopes (if any), the variances of
    continuous variables (if any), and finally the lower triangular elements
    of the correlation/covariance matrix excluding the diagonal, ordered
    column by column.}
\item{nacov}{A user provided matrix containing the elements of (N times)
    the asymptotic variance-covariance matrix of the sample statistics.
    For a multiple group analysis, a list with an asymptotic
    variance-covariance matrix for each group. See the \code{wls_v}
    argument for information about the order of the elements.}
\item{ov_order}{Character. If \code{"model"} (the default), the order of
    the observed variable names (as reflected for example in the output of
    \code{lav_object_vnames()}) is determined by the model syntax. If
    \code{"data"}, the order is determined by the data (either the full
    data.frame or the sample (co)variance matrix). If the \code{wls_v}
    and/or \code{nacov} matrices are provided, this argument is currently
    set to \code{"data"}.}
\item{...}{Many more options can be specified, using 'name = value'.
    See \code{\link{lavOptions}} for a complete list.}
}
\details{
    The \code{growth} function is a wrapper for the more general
    \code{\link{lavaan}} function, using the following default arguments:
    \code{meanstructure = TRUE}, 
    \code{int.ov.free = FALSE}, \code{int.lv.free = TRUE},
    \code{auto.fix.first = TRUE} (unless \code{std.lv = TRUE}),
    \code{auto.fix.single = TRUE}, \code{auto.var = TRUE},
    \code{auto.cov.lv.x = TRUE}, \code{auto.efa = TRUE},
    \code{auto.th = TRUE}, \code{auto.delta = TRUE},
    and \code{auto.cov.y = TRUE}.
}
\value{
  An object of class \code{\linkS4class{lavaan}}, for which several methods
  are available, including a \code{summary} method.
}
\references{
Yves Rosseel (2012). lavaan: An R Package for Structural Equation
Modeling. Journal of Statistical Software, 48(2), 1-36. 
\doi{10.18637/jss.v048.i02}}
\seealso{
\code{\link{lavaan}}, \code{\link{estimator_iv}}
}
\examples{
## linear growth model with a time-varying covariate
model.syntax <- '
  # intercept and slope with fixed coefficients
    i =~ 1*t1 + 1*t2 + 1*t3 + 1*t4
    s =~ 0*t1 + 1*t2 + 2*t3 + 3*t4

  # regressions
    i ~ x1 + x2
    s ~ x1 + x2

  # time-varying covariates
    t1 ~ c1
    t2 ~ c2
    t3 ~ c3
    t4 ~ c4
'

fit <- growth(model.syntax, data = Demo.growth)
summary(fit)
}

\name{lav_partable}
\alias{lav_pt_independence}
\alias{lav_pt_unrestricted}
\alias{lav_pt_df}
\alias{lav_pt_ndat}
\alias{lav_pt_npar}
\alias{lav_pt_ngroups}
\alias{lav_pt_labels}
\alias{lav_pt_from_lm}
\alias{lav_pt_complete}
\alias{lav_pt_attributes}
\alias{lav_pt_merge}
\alias{lav_pt_add}
\title{lavaan partable functions}
\description{Utility functions related to the parameter table (partable)}
\usage{
# extract information from a parameter table
lav_pt_df(partable)
lav_pt_ndat(partable)
lav_pt_npar(partable)
lav_pt_ngroups(partable)
lav_pt_attributes(partable, pta = NULL)

# generate parameter labels
lav_pt_labels(partable, blocks = c("group", "level"), 
                    group_equal = "", group_partial = "", type = "user")

# generate parameter table for specific models
lav_pt_independence(lavobject = NULL, lavdata = NULL,
    lavpta = NULL, lavoptions = NULL, lavsamplestats = NULL,
    lavh1 = NULL,
    sample_cov = NULL, sample_mean = NULL, sample_slopes = NULL,
    sample_th = NULL, sample_th_idx = NULL,
    sample_cov_x = NULL, sample_mean_x = NULL)

lav_pt_unrestricted(lavobject = NULL, lavdata = NULL, 
    lavpta = NULL, lavoptions = NULL, lavsamplestats = NULL, 
    lavh1 = NULL,
    sample_cov = NULL, sample_mean = NULL, sample_slopes = NULL, 
    sample_th = NULL, sample_th_idx = NULL,
    sample_cov_x = NULL, sample_mean_x = NULL) 

lav_pt_from_lm(object, est = FALSE, label = FALSE, 
    as_data_frame = FALSE)

# complete a parameter table only containing a few columns (lhs,op,rhs)
lav_pt_complete(partable = NULL, start = TRUE)

# merge two parameter tables
lav_pt_merge(pt1 = NULL, pt2 = NULL, remove_duplicated = FALSE, 
    from_last = FALSE, warn = TRUE)

# add a single parameter to an existing parameter table
lav_pt_add(partable = NULL, add = list())
}
\arguments{
\item{partable}{A parameter table. See \code{\link{lavParTable}} 
for more information.}
\item{blocks}{Character vector. Which columns in the parameter table should be
  taken to distinguish between different blocks of parameters (and hence
  be given different labels)? If \code{"blocks"} includes \code{"group"}, a
  suffix \code{".g"} and the group number (or group label) is added for 
  the parameters of  all but the first group. If \code{"blocks"} includes 
  \code{"level"}, a suffix \code{".l"} and the level number is added for 
  the parameters of all but the first level. If \code{"blocks"} includes, 
  say \code{"foo"}, a suffix \code{".foo"} and the corresponding value of
  \code{"foo"} is added to all parameters.}
\item{group_equal}{The same options can be used here as in the fitting functions.
Parameters that are constrained to be equal across groups will be given the
same label.}
\item{group_partial}{A vector of character strings containing the labels
    of the parameters which should be free in all groups.}
\item{type}{Character string. Either `user' to select all entries, or
`free' to select only the free parameters from the parameter table.}
\item{lavobject}{An object of class `lavaan'. If this argument is provided,
it should be the only argument. All the values for the other arguments 
are extracted from this object.}
\item{lavdata}{An object of class `lavData'. The Data slot from a lavaan object.}
\item{lavoptions}{A named list. The Options slot from a lavaan object.}
\item{lavsamplestats}{An object of class `lavSampleStats'. The SampleStats
slot from a lavaan object.}
\item{lavh1}{A named list. The h1 slot from a lavaan object.}
\item{lavpta}{The pta (parameter table attributes) slot from a lavaan object.}
\item{sample_cov}{Optional list of numeric matrices. 
Each list element contains a sample variance-covariance matrix for this group.
If provided, these values will be used as starting values.}
\item{sample_mean}{Optional list of numeric vectors. 
Each list element contains a sample mean vector for this group. 
If provided, these values will be used as starting values.}
\item{sample_slopes}{Optional list of numeric matrices.
Each list element contains the sample slopes for this group (only used
when \code{conditional.x = TRUE}).
If provided, these values will be used as starting values.}
\item{sample_th}{Optional list of numeric vectors. 
Each list element contains a vector of sample thresholds for this group.
If provided (and also sample.th.idx is provided), 
these values will be used as starting values.}
\item{sample_th_idx}{Optional list of integers. Each list contains the
threshold indices for this group.}
\item{sample_cov_x}{Optional list of numeric matrices.  Each list element
contains a sample variance-covariance matrix for the exogenous variables
for this group (only used when \code{conditional.x = TRUE}).  If provided,
these values will be used as starting values.}
\item{sample_mean_x}{Optional list of numeric vectors.
Each list element contains a sample mean vector for the exogenous variables
for this group (only used when \code{conditional.x = TRUE}).
If provided, these values will be used as starting values.}
\item{est}{Logical. If TRUE, include the fitted estimates in the parameter table.}
\item{label}{Logical. If TRUE, include parameter labels in the parameter table.}
\item{as_data_frame}{Logical. If TRUE, return the parameter table as a data.frame.}
\item{object}{An object of class \code{lm}.}
\item{start}{Logical. If TRUE, include a start column, based on 
the simple method for generating starting values.}
\item{pta}{A list containing parameter attributes.}
\item{pt1}{A parameter table.}
\item{pt2}{A parameter table.}
\item{remove_duplicated}{Logical. If \code{TRUE}, remove duplicated elements
when merging two parameter tables.}
\item{from_last}{Logical. If \code{TRUE}, duplicated elements are considered
from the bottom of the merged parameter table.}
\item{warn}{Logical. If \code{TRUE}, a warning is produced when duplicated
elements are found, when merging two parameter tables.}
\item{add}{A named list.  A single row of a parameter table as a named list.}
}
\examples{
# generate syntax for an independence model
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
lav <- lav_pt_independence(fit)
as.data.frame(lav, stringsAsFactors = FALSE)

# how many free parameters?
lav_pt_npar(lav)

# how many sample statistics?
lav_pt_ndat(lav)
}

\name{lavScores}
\alias{lavScores}
\alias{estfun.lavaan}
\title{Extract Empirical Estimating Functions}
\description{
A function for extracting the empirical estimating functions of a fitted
lavaan model. This is the derivative of the objective function
with respect to the parameter vector, evaluated at the observed (case-wise)
data. In other words, this function returns the case-wise scores, evaluated
at the fitted model parameters.
}
\usage{
estfun.lavaan(object, scaling = FALSE, ignore_constraints = FALSE,
              remove_duplicated = TRUE, remove_empty_cases = TRUE, ...)
lavScores(object, scaling = FALSE, ignore_constraints = FALSE,
          remove_duplicated = TRUE, remove_empty_cases = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{scaling}{Only used for the ML estimator. If \code{TRUE}, the scores are
scaled to reflect the specific objective function used by lavaan. If
\code{FALSE} (the default), the objective function is the loglikelihood
function assuming multivariate normality.}
\item{ignore_constraints}{Logical. If \code{TRUE}, the scores do not reflect
the (equality or inequality) constraints. If \code{FALSE}, the scores are
computed by taking the unconstrained scores, and adding the term \code{t(R)
lambda}, where \code{lambda} are the (case-wise) Lagrange Multipliers, and
\code{R} is the Jacobian of the constraint function. Only in the latter case
will the sum of the columns be (almost) equal to zero.}
\item{remove_duplicated}{If \code{TRUE}, and all the equality constraints have
a simple form (e.g., a == b), the unconstrained scores are post-multiplied with a
transformation matrix in order to remove the duplicated parameters.}
\item{remove_empty_cases}{If \code{TRUE}, empty cases with only missing values
will be removed from the output.}
\item{...}{To accept old argument names with dots. No other arguments are accepted.}
}
\value{A n x k matrix corresponding to n observations and k parameters.}
\details{
Casewise scores are available for the ML, GLS, ULS and WLS estimators, and
(since 0.7-1) for the PML (pairwise maximum likelihood) estimator. For PML,
the scores are the casewise derivatives of the pairwise log-likelihood
(including the weighted univariate part when
\code{missing = "available.cases"}); \code{missing = "doubly.robust"} is not
supported (yet). Exogenous covariates (\code{conditional.x = TRUE}) are
supported for PML only.
}
\author{Ed Merkle for the ML case; the \code{remove.duplicated},
\code{ignore.constraints} and \code{remove.empty.cases} arguments were added by
Yves Rosseel; Franz Classe for the WLS case; Yves Rosseel for the PML case.}
\examples{
## The famous Holzinger and Swineford (1939) example
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data = HolzingerSwineford1939)
head(lavScores(fit))
}

\name{lavPredictY}
\alias{lavPredictY}
\title{Predict the values of y-variables given the values of x-variables}
\description{
This function can be used to predict the values of (observed) y-variables
given the values of (observed) x-variables in a structural equation model.}
\usage{
lavPredictY(object, newdata = NULL,
            ynames = lav_object_vnames(object, "ov.y"),
            xnames = lav_object_vnames(object, "ov.x"),
            method = "conditional.mean",
            label = TRUE, assemble = TRUE,
            force_zero_mean = FALSE,
            lambda = 0,
            ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{newdata}{An optional data.frame, containing the same variables as
the data.frame that was used when fitting the model in \code{object}.
This data.frame should also include
the y-variables (although their values will be ignored).
Note that if no meanstructure was used in the original fit, we will
use the saturated sample means of the original fit as substitutes for the
model-implied means. Alternatively, refit the model using \code{meanstructure = TRUE}.
For a multiple-group model, \code{newdata} can also be a \emph{list} with one
data.frame per group (in the same order as the groups in \code{object}). An
element may be \code{NULL} to skip that group. This is useful when the groups
have a different set of variables, so that \code{ynames} and \code{xnames}
can be specified per group (as a list). In that case, the result contains
predictions only for the groups for which data were provided (a single matrix
if only one group was requested, otherwise a named list).}
\item{ynames}{The names of the observed variables that should be treated as the
y-variables. It is for these variables that the function will predict
the (model-based) values for each observation. Can also be a list to allow
for a separate set of variable names per group (or block).}
\item{xnames}{The names of the observed variables that should be treated as the
x-variables. Can also be a list to allow
for a separate set of variable names per group (or block).}
\item{method}{A character string. The only available option for now is
\code{"conditional.mean"}. See Details.}
\item{label}{Logical. If TRUE, the columns of the output are labeled.}
\item{assemble}{Logical. If TRUE,
the predictions for the separate groups in the output are
reassembled into a single data.frame
with a group column, having the same dimensions as the original
(or newdata) dataset.}
\item{force_zero_mean}{Logical. Only relevant if there is no mean structure.
If \code{TRUE}, the (model-implied) mean vector is set to the zero vector.
If \code{FALSE}, the (model-implied) mean vector is set to the (unrestricted)
sample mean vector.}
\item{lambda}{Numeric. A lambda regularization penalty term.}
\item{...}{To support old argument name `force.zero.mean`}
}
\details{
This function can be used for (SEM-based) out-of-sample predictions of
outcome (y) variables, given the values of predictor (x) variables. This
is in contrast to the \code{lavPredict()} function which (historically)
only `predicts' the (factor) scores for latent variables, ignoring the
structural part of the model.

When \code{method = "conditional.mean"}, predictions (for y given x)
are based on the (joint y and x) model-implied variance-covariance (Sigma)
matrix and  mean vector (Mu), and the standard expression for the
conditional mean of a multivariate normal distribution. Note that if the
model is saturated (and hence df = 0), the SEM-based predictions are identical
to ordinary least squares predictions.

Lambda is a regularization penalty term to improve prediction accuracy that can
be determined using the \code{lavPredictY_cv} function.
}
\seealso{
\code{\link{lavPredict}} to compute scores for latent variables.

\code{\link{lavPredictY_cv}} to determine an optimal lambda to increase
prediction accuracy.

\code{\link{lavResidualsY}} to compute the residuals for the y-variables
(observed minus predicted).
}
\references{
de Rooij, M., Karch, J.D., Fokkema, M., Bakk, Z., Pratiwi, B.C, and
Kelderman, H. (2022) SEM-Based Out-of-Sample Predictions,
Structural Equation Modeling: A Multidisciplinary Journal.
\doi{10.1080/10705511.2022.2061494}

Molina, M. D., Molina, L., & Zappaterra, M. W. (2024). Aspects of Higher
Consciousness: A Psychometric Validation and Analysis of a New Model of Mystical
Experience. \doi{10.31219/osf.io/cgb6e}
}
\examples{
model <- '
  # latent variable definitions
     ind60 =~ x1 + x2 + x3
     dem60 =~ y1 + a*y2 + b*y3 + c*y4
     dem65 =~ y5 + a*y6 + b*y7 + c*y8

  # regressions
    dem60 ~ ind60
    dem65 ~ ind60 + dem60

  # residual correlations
    y1 ~~ y5
    y2 ~~ y4 + y6
    y3 ~~ y7
    y4 ~~ y8
    y6 ~~ y8
'
fit <- sem(model, data = PoliticalDemocracy)

lavPredictY(fit, ynames = c("y5", "y6", "y7", "y8"),
                 xnames = c("x1", "x2", "x3", "y1", "y2", "y3", "y4"))
}

\name{lavExport}
\alias{lavExport}
\title{lavaan Export}
\description{
Export a fitted lavaan object to an external program.}
\usage{
lavExport(object, target = "lavaan", prefix = "sem", dir_name = "lav_export",
          export = TRUE, ...)
}
\arguments{
\item{object}{An object of class \code{\linkS4class{lavaan}}.}
\item{target}{The target program. Current options are \code{"lavaan"} and
\code{"Mplus"}.}
\item{prefix}{The prefix used to create the input files; the name of the
input file has the pattern \sQuote{prefix dot target dot in}; the name of the
data file has the pattern \sQuote{prefix dot target dot raw}.}
\item{dir_name}{The directory name (including a full path) where the
input files will be written.}
\item{export}{If \code{TRUE}, the files are written to the output directory
(\code{dir_name}). If \code{FALSE}, only the syntax is generated as a
character string.}
\item{...}{Only to support old argument name 'dir.name'.}
}
\value{
If \code{export = TRUE}, a directory (called \code{lav_export} by default) will
be created, typically containing a data file, and an input file so that the
same analysis can be run using an external program. If \code{export = FALSE}, a
character string containing the model syntax only for the target program.
}
\details{
This function was mainly created to quickly generate an Mplus syntax file to
compare the results between Mplus and lavaan. The target \code{"lavaan"} can
be useful to create a full model syntax as needed for the \code{lavaan()}
function. More targets (perhaps for \code{LISREL} or \code{EQS}) will be added
in future releases.
}
\seealso{\code{\link{lavParTable}}, \code{\link{lav_mplus_lavaan}}}
\examples{
HS.model <- ' visual  =~ x1 + x2 + x3
              textual =~ x4 + x5 + x6
              speed   =~ x7 + x8 + x9 '

fit <- cfa(HS.model, data=HolzingerSwineford1939)
out <- lavExport(fit, target = "Mplus", export=FALSE)
cat(out)
}

\name{lavaanList-class}
\docType{class}
\alias{lavaanList-class}
\alias{coef,lavaanList-method}
\alias{summary,lavaanList-method}

\title{Class For Representing A List of (Fitted) Latent Variable Models}
\description{The \code{lavaanList} class represents a collection of (fitted)
  latent variable models, fitted to a (potentially large) number of datasets.
  It contains information about the model (which is always the same),
  and, for every dataset, a set of (user-specified) slots from a regular
  lavaan object.}
\section{Objects from the Class}{
Objects can be created via the
  \code{\link{cfaList}}, \code{\link{semList}}, or
  \code{\link{lavaanList}} functions.
}
\section{Slots}{
  \describe{
    \item{\code{version}:}{The lavaan package version used to create this object}
    \item{\code{call}:}{The function call as returned by \code{match.call()}.}
    \item{\code{Options}:}{Named list of options that were provided by
      the user, or filled-in automatically.}
    \item{\code{ParTable}:}{Named list describing the model parameters. Can be coerced to a data.frame. In the documentation, this is called the `parameter table'.}
    \item{\code{pta}:}{Named list containing parameter table attributes.}
    \item{\code{Data}:}{Object of internal class \code{"Data"}: information
about the data.}
    \item{\code{Model}:}{Object of internal class \code{"Model"}: the 
      internal (matrix) representation of the model}
    \item{\code{meta}:}{List containing additional flags. For internal use only.}
    \item{\code{timingList}:}{List. Timing slot per dataset.}
    \item{\code{ParTableList}:}{List. ParTable slot per dataset.}
    \item{\code{DataList}:}{List. Data slot per dataset.}
    \item{\code{SampleStatsList}:}{List. SampleStats slot per dataset.}
    \item{\code{CacheList}:}{List. Cache slot per dataset.}
    \item{\code{vcovList}:}{List. vcov slot per dataset.}
    \item{\code{testList}:}{List. test slot per dataset.}
    \item{\code{optimList}:}{List. optim slot per dataset.}
    \item{\code{impliedList}:}{List. implied slot per dataset.}
    \item{\code{h1List}:}{List. h1 slot per dataset.}
    \item{\code{loglikList}:}{List. loglik slot per dataset.}
    \item{\code{baselineList}:}{List. baseline slot per dataset.}
    \item{\code{funList}:}{List. fun slot per dataset.}
    \item{\code{internalList}:}{List. internal slot per dataset.}
    \item{\code{external}:}{List. Empty slot to be used by add-on packages.}
  }
}
\section{Methods}{
  \describe{
    \item{coef}{\code{signature(object = "lavaanList", type = "free")}: Returns 
      the estimates of the parameters in the model as the columns in a matrix;
      each column corresponds to a different dataset.
      If \code{type="free"}, only the free parameters are returned.
      If \code{type="user"}, all parameters listed in the parameter table
      are returned, including constrained and fixed parameters.}
	\item{summary}{\code{signature(object = "lavaanList", header = TRUE,
         estimates = TRUE, print = TRUE, nd = 3L,
         simulate.args = list(est.bias = TRUE, se.bias = TRUE,
         se.bias.med = TRUE, prop.sig = TRUE, coverage = TRUE,
         level = 0.95, trim = 0))}:
      Print a summary of the collection of fitted models.
      If \code{header = TRUE}, the header section is
      printed.
      If \code{estimates = TRUE}, print the parameter estimates section.
      If \code{print = TRUE}, the summary is printed to the screen.
      The argument \code{nd} determines the number of digits after the
      decimal point to be printed (currently only in the parameter estimates
      section.)
      The argument \code{simulate.args} is only used if the meta slot
      indicates that the parameter tables are obtained in the context of
      a simulation. The options switch on/off the columns that are printed,
      and the \code{trim} option determines the amount of trimming that is
      used when taking the average (or standard deviation) across all
      replications.
      If \code{se.bias.med = TRUE} (the default, only relevant when
      \code{se.bias = TRUE}), two additional columns are printed: \code{se.med}
      (the median, rather than the mean, of the computed standard errors across
      replications) and \code{se.bias.med} (the ratio \code{se.med/se.obs}).
      The median is a robust alternative to the mean-based \code{se.ave}/\code{se.bias},
      which can be dominated by a few outliers for nonlinear defined parameters
      (e.g. ratios with near-zero denominators in some replications).
    }
  }
}
\seealso{
\code{\link{cfaList}}, \code{\link{semList}}, \code{\link{lavaanList}}
}

\name{lav_constraints}
\alias{lav_con_parse}
\alias{lav_pt_con_ceq}
\alias{lav_pt_con_ciq}
\alias{lav_pt_con_def}
\title{Utility Functions: Constraints}
\description{Utility functions for equality and inequality constraints.}
\usage{
lav_con_parse(partable = NULL, constraints = NULL, theta = NULL,
                     debug = FALSE)
lav_pt_con_ceq(partable, con = NULL, debug = FALSE,
                             txt_only = FALSE)
lav_pt_con_ciq(partable, con = NULL, debug = FALSE,
                             txt_only = FALSE)
lav_pt_con_def(partable, con = NULL, debug = FALSE,
                             txt_only = FALSE, warn = TRUE)
}
\arguments{
\item{partable}{A lavaan parameter table.}
\item{constraints}{A character string containing the constraints.}
\item{theta}{A numeric vector. Optional vector with values for the
model parameters in the parameter table.}
\item{debug}{Logical. If TRUE, show debugging information.}
\item{con}{An optional partable where the operator is one of `==',
`>', `<' or `:='}
\item{txt_only}{Logical. If TRUE, only the body of the function is returned as a character string. If FALSE, a function is returned.}
\item{warn}{Logical. If FALSE, warnings are suppressed.}
}
\details{
This is a collection of lower-level constraint-related functions
that are used in the lavaan code. They are made public at the
request of package developers. Below is a brief description of what
they do:

The \code{lav_con_parse} function parses the constraints
specification (provided as a string, see example), and generates
a list with useful information about the constraints.

The \code{lav_pt_con_ceq} function creates a function
which takes the (unconstrained) parameter vector as input, and
returns the slack values for each equality constraint. If the equality
constraints hold perfectly, this function returns zeroes.

The \code{lav_pt_con_ciq} function creates a function
which takes the (unconstrained) parameter vector as input, and
returns the slack values for each inequality constraint.

The \code{lav_pt_con_def} function creates a function
which takes the (unconstrained) parameter vector as input, and
returns the computed values of the defined parameters.
}
\examples{
myModel <- 'x1 ~ a*x2 + b*x3 + c*x4'
myParTable <- lavParTable(myModel, as_data_frame = FALSE)
con <- ' a == 2*b
         b - c == 5 '
conInfo <- lav_con_parse(myParTable, constraints = con)

myModel2 <- 'x1 ~ a*x2 + b*x3 + c*x4
             a == 2*b
             b - c == 5 '
ceq <- lav_pt_con_ceq(partable = lavParTable(myModel2))
ceq( c(2,3,4) )
}
